#! /usr/bin/env python3
# Julian to Greenwich and vice versa
# with my own hvfTime class, works also for BC dates and leap years
# hvf 16.09.2021 26.12.2021 11.11.2022 13.04.2023
# make hvfTime() hashable -- hvf 02.10.2023

from datetime import datetime, timedelta
from math import floor
import sys
import re
import dateutil.parser

VersionJulian = "V1.8 hvf 2023-10-02 HVFTime.py based on Meeus"


# JulianDay from year, month, day -  hvf 16.11.2021
def getJulianDayfromYMD(year, month, day, hour=0, minute=0, second=0, 
                        verbose = False):
    if verbose: 
        print ("getJulianDayfromYMD: ", year, month, day, hour, minute, second)
    day  += hour/24. + \
            minute/1440. + \
            second/86400.
    isGregor = False    # 04.10.1582 is last day before Gregorian calendar
    if year > 1582 or year == 1582 and (month > 10 or month == 10 and day > 4):
        isGregor = True
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    # JD 0 is -4712-01-01.5 i.e. -4712-01-01 12h
    # N.B. JD 1524.5 is -4708-03-05 00h
    if verbose: print ("JD = ", JD)
    return JD


def roundhvfTime(hvfDT = None, roundTo = 3600):
    # Round a hvfTime object to anytime lapse in seconds
    # hvfDT: hvftime object, default now
    # roundTo: closest number of seconds to round to, default 1 hour.
    # Author: Harald von Fellenberg 2021
    # Problem: 2021-11-21 1223 returns 2021-11-21 12:22:60.000
    # hvf 2022-09-09
    verbose=False
    if not isinstance(hvfDT, hvfTime): hvfDT = hvfTime.now()
    h, m, s = hvfDT.getTime()
    seconds = h*3600 + m*60 + s
    fraction = seconds - floor(seconds)     # fractional seconds
    if verbose:
        print ("seconds, fraction, floor, rounding = ", seconds, fraction, floor(seconds))
    seconds = floor(seconds)
    rounding = (seconds+roundTo/2) // roundTo * roundTo
    if verbose: print (rounding)
    return hvfDT + timedelta(0, rounding-seconds, -fraction*1e6)

# returns yyyymmdd hh for a hvfTime object, rounded
def nicehvfTimehh (hvfDT):
    hvfDTrounded = roundhvfTime(hvfDT)
    yy, mm, dd = hvfDTrounded.getDate()
    hh, dummy, dummy = hvfDTrounded.getTime()

    s = ''
    if yy >= 0:
        s = " %04d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    else:
        s = "%05d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    return s

#######################################################################
# replaces astropy.time.Time since that does not support leap days 
# before 100BC
# values are represented in Julian days
# creator parameter can be:
# - a hvfTime object
# - a JD JulianDay object (float)
# - a datetime object
# - date and time as six values yy, mm, dd, [hh, MM, ss]
#   dd and ss may have fractions
# hvf 26.12.2021

class hvfTime:
    def __init__(self, *arg):
        # print ("__init__: arg = ", arg)
        # argument is a hvfTime object
        if isinstance(arg[0], hvfTime):
            JD = arg[0].getJD()
            self.__JD = JD
        # argument is a JD
        elif isinstance(arg[0], float) and len(arg) == 1:
            JD = arg[0]
            self.__JD = JD
        # argument is a datetime object
        elif isinstance(arg[0], datetime):
            dt = arg[0]
            yy = dt.year
            mm = dt.month
            dd = dt.day
            hh = dt.hour
            MM = dt.minute
            ss = dt.second + dt.microsecond*1e-6
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        # argument is yy mm dd [hh MM ss]
        elif (isinstance(arg[0], int) or isinstance(arg[0], float)) and \
            len(arg) >= 3:
            hh = 0; MM = 0; ss = 0
            yy = arg[0]; mm = arg[1]; dd = arg[2]
            if len(arg) > 3: hh = arg[3]
            if len(arg) > 4: MM = arg[4]
            if len(arg) > 5: ss = arg[5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD

        # argument is (yy mm dd [hh MM ss])
        elif isinstance(arg[0], tuple) or isinstance(arg[0], list):
            hh = 0; MM = 0; ss = 0
            yy = arg[0][0]; mm = arg[0][1]; dd = arg[0][2]
            if len(arg[0]) > 3: hh = arg[0][3]
            if len(arg[0]) > 4: MM = arg[0][4]
            if len(arg[0]) > 5: ss = arg[0][5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        else:
            print ("hvfTime instance not supported: ", arg[0])
            sys.exit(1)
            
    def getJD(self):
        return self.__JD

    def getDate(self, verbose = False):  # returns (yy, mm, dd), 
                                        # where dd may have fraction
        # works also for negative JDs! hvf 29.12.2021
        JD = self.__JD
        jd1 = JD + 0.5
        I = floor(jd1)      # integer part, may be negative
        F = jd1 - I         # fractional part, always positive
        if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
            A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
            B = I + A - int(A/4) + 1
        else:   B = I       # Julian Calendar for earlier dates
        C = B + 1524
        D = floor((C-122.1)/365.25)
        E = floor(365.25 * D)
        G = floor((C-E)/30.6001)
        d = C - E + F - floor(30.6001 * G)
        if G < 13.5:    m = G - 1
        else:           m = G - 13
        if m > 2.5:     y = D - 4716
        else:           y = D - 4715

        if verbose:
            print ("getDate: JD, jd1, I, F = ", JD, jd1, I, F)
            print ("getDate: B, C, D, E, G = ", B, C, D, E, G)
            print ("getDate: y, m, d       = ", y, m, d)
        return (y, m, d)

    def getTime(self):      # seconds have 3 digi fractions (millisec)
        dummy, dummy, t = self.getDate()
        t -= floor(t)
        # t is fraction of day - convert to sec and round
        daysec = round(t * 86400., 3)
        hms = daysec / 3600.
        #hms = t*24.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 3)
        #s = int((ms - m) * 60)
        #print ("daysec, hms, h, ms, m, s = ", daysec, hms, h, ms, m, s)
    
        return (h, m, s)

    def __str__(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, hh, mm, ss)

    def getStr(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        daysec = round(ss + mm*60 + hh*3600, 0)
        hms = daysec / 3600.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 0)

        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, h, m, s)

    # Ernesti, Python3, p. 380
    # hvfTime = hvfTime + timedelta
    # hvfTime = hvfTime + day   (float or int)
    def __add__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__add__: wrong type ", type(td))
            sys.exit(1)
        return hvfTime(self.__JD + sec/86400.)

    # hvfTime += timedelta
    def __iadd__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__iadd__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD += sec/86400.
        return self

    # hvfTime = hvfTime - timedelta
    # and   n = hvfTime - hvfTime       n in days
    def __sub__(self, td):
        # subtract a timedelta object 
        if isinstance (td, timedelta):
            sec = td.total_seconds()
            return hvfTime(self.__JD - sec/86400.)
        elif isinstance(td, hvfTime):
            # return JDs in days as difference between two hvfTime objects
            return self.__JD - td.getJD()
        else:
            print ("hvfTime.__sub__: wrong type ", type(td), td)
            sys.exit(1)

    # hvfTime -= timedelta
    def __isub__(self, td):
        # subtract a timedelta object 
        try:    
            sec = td.total_seconds()
        except:
            print ("hvfTime.__isub__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD -= sec/86400.
        return self

    # hvfTime < hvfTime
    def __lt__(self, other):
        return self.__JD < other.getJD()

    # hvfTime <= hvfTime
    def __le__(self, other):
        return self.__JD <= other.getJD()

    # hvfTime == hvfTime
    def __eq__(self, other):
        return self.__JD == other.getJD()

    # hash - makes hvfTime hashable, i.e. as key in dict or set 
    def __hash__(self):
        return hash(self.__JD)

    # hvfTime != hvfTime
    def __ne__(self, other):
        return self.__JD != other.getJD()

    # hvfTime > hvfTime
    def __gt__(self, other):
        return self.__JD > other.getJD()

    # hvfTime >= hvfTime
    def __ge__(self, other):
        return self.__JD >= other.getJD()

    def now():  # returns current time (h, m, s), uses datetime.now()
                # roundedd to integer seconds
        return hvfTime(datetime.now()).getTime()

    def today():    # returns current date (y, m, d), uses datetime.today()
                    # d has fractions
        return hvfTime(datetime.today()).getDate()
    
    def getDatetime(self):
        (yy, mm, dd) = self.getDate()
        dd = floor(dd)
        (HH, MM, SS) = self.getTime()
        SS = int(round(SS, 0))
        return datetime(yy, mm, dd, HH, MM, SS)
#######################################################################

# parse date time
# returns a hvfTime object
# hvf 19.11.2021 25.12.2021
# allowed patterns:
# 2021-11-21
# 2021-11-21 12:23:45.635
# 2021-11-21T12:23:45.635
# 2021-11-21 122345.635
# 2021-11-21 122345
# 2021-11-21 1223
# 2021-11-21 12
# -4712-1-1.5 
# 25.12.2021 12:34:56.789
# 1.1.2022
# 2451545.0 (a Julian Day)
# works for dates between -9999-01-01 and 9999-12-31

def parseDateTime(s, verbose = False):
    HH = MM = SS = 0
    # check if two parts separated by blank or T
    if verbose: print ("parseDateTime input: ", s)
    s = s.strip()
    dpart = s
    tpart = None
    parts = re.split (r"[\sT]", s)
    if len(parts) >= 2:
        dpart = parts[0]    # date part
        tpart = parts[1]    # time part
        if verbose:
            print ("date part = ", dpart)
            print ("time part = ", tpart)

    # parse date  2021-12-24 or  4712-1-1.5
    if re.fullmatch(r'\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        ymd = re.split(r'-', dpart)
        yy = ymd[0]
        mm = ymd[1]
        dd = ymd[2]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date 24.12.2021 or 1.1.2022
    elif re.fullmatch(r'\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        ymd = re.split(r'\.', dpart)
        yy = ymd[2]
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -4712-01-01.5
    elif re.fullmatch(r'-\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        symd = re.split(r'-', dpart)
        yy = -int(symd[1])
        mm = symd[2]
        dd = symd[3]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -24.12.2021 or -1.1.2022
    elif re.fullmatch(r'-\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        symd = re.split(r'-', dpart)
        ymd = re.split(r'\.', symd[1])
        yy = -int(ymd[2])
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    else:
        print ("parseDateTime format unknown: ", s)
    # parse date  2021-12-24 or  4712-1-1.5
    # parse date 24.12.2021 or 1.1.2022
    # parse date -4712-01-01.5
    # parse date -24.12.2021 or -1.1.2022
        print ("Allowed date formats include 2021-12-24, 4712-1-1.5,")
        print ("    24.12.2021, 1.1.2022, -4712-01-01.5, -24.12.2021, -1.1.2022")
        print ("Allowed range of years is from -9999 to 9999")
        sys.exit(1)

    # parse time using dateutil.parser.isoparse - this is easier
    # than checking all the different possible input formats
    if tpart:
        deltaT = dateutil.parser.isoparse("2021-12-24 " + tpart) \
                - datetime(2021, 12, 24)
        Secs = round(deltaT.total_seconds(), 3)    # avoid rounding error
        HH = Secs // 3600
        Secs -= HH * 3600
        MM = Secs // 60
        Secs -= MM * 60
        SS = Secs
        if verbose:
            print ("HH, MM, SS = ", HH, MM, SS)

    #return hvfTime object
    return hvfTime(int(yy), int(mm), float(dd), int(HH), int(MM), float(SS))

#######################################################################
# catculate weekday for any date from -9999-01-01 to 9999-12-31
# Meeus 1991 p. 65

def getDayofWeek(dt):
    name =["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday",
            "Friday", "Saturday"]
    hvfdt = hvfTime(dt)
    (yy, mm, dd) = hvfdt.getDate()
    hvfdtMidnight = hvfTime(yy, mm, int(dd), 0, 0, 0.0)
    JD0 = hvfdtMidnight.getJD()
    weekday = int(JD0 + 1.51) % 7
    return name[weekday]

#######################################################################

#!/usr/bin/env python3
# -*- coding: iso-8859-1 -*-

# Der folgende Python-Code fuer Sonne und Mond 
# basiert auf der Portierung des java-Scripts von Arnold Barmettler
# http://lexikon.astronomie.info/java/sunmoon/sunmoon.html
#
#     Entfernen Sie folgende Informationen auf keinen Fall: / 
#     Do not remove following text:
#     Source code by Arnold Barmettler, www.astronomie.info / www.CalSky.com
#     and Harald von Fellenberg, www.hvf.ch
#     based on algorithms by Peter Duffett-Smith's great and easy book
#     'Practical Astronomy with your Calculator or Spreadsheet'.
#     and Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Totale Ueberarbeitung durch Harald von Fellenberg, Dezember 2020
#   Portierung auf Python 3
#   Ergaenzung fuer Mitternachtssonne, Polarnacht
#   Ergaenzungen fuer Daemmerung, Hoehe der Sonne beim Transit tags und nachts
#   Ergaenzung mit Zeitgleichung, wahre und mittlere Ortszeit
#   Azimut für Sonnenaufgang, Sonnenuntergang
#   Liste einiger Städte mit ihren Koordinaten
#   Korrekte Berechnung in allen Zeitzonen
#   Behandlung aller Planeten
#
#   Die Behandlung von Sonne und Mond wurde ersetzt durch die Algorithmen in
#   Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Die Daten aller Planeten beruhen auf VSPO87
#   Pierre Bretagnon, Gerard Francou
#   Bureau des Longitudes
#   77, Avenue Denfert-Rochereau
#   F75014, Paris, France
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/
#
#   Die Resultate wurden mit denen auf https://www.timeanddate.com/ verglichen,
#   die Uebereinstimmung ist gut innerhalb von wenigen Minuten.
#   Beim Mond ist stets das Datum zu beachten!
#   Liegt die lokale Zeit zwischen Mitternacht und "zone", so wird für den Mond
#   zum Teil auf den Vortag zurückgegriffen. Dies ist besonders für Orte mit
#   grosser Zeitzonendifferenz wichtig, wie Sydney oder Honolulu.
#   Für Orte in den Polarregionen wird Mitternachtssonne und Polarnacht
#   korrekt ausgegeben, ausser am Tag, wo die Sonne unter- oder aufgeht.
#   Ebenso fuer den Mond.
#
#   Anmerkung: die Webseiten astronomie.info sowie calsky.com existieren
#   seit Oktober 2020 nicht mehr.
#
#   Alle 8 Planeten sind ebenfalls behandelt, gemaess den Algorithmen in
#   Jean Meeus, Astronomical Algorithms.
#   Die Daten zu den 8 Planeten stammen aus den VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Version 1.0
#   Version 2.0
#   Version 3.0
#   Version 3.1
#   Version 3.4
#   Version 4.0 inkl Sun-Events, Moon-Events
#   Version 5.1 
Version = "V5.3 hvf 2023-09-27 SunHead.py using VSOP87 ephemerides and HVFTime"
if 'VersionList' not in globals().keys(): VersionList = []
VersionList.append(Version)

import time
from datetime import datetime, date, timedelta
from math import sin, acos, cos, pi, radians, degrees, atan, atan2
from math import asin, tan, ceil, floor, sqrt, fabs
import argparse
import sys
import dateutil.parser
import os

# alle Konstanten sind auf die Epoche 2010.0, JD=2455196.5 bezogen
# falls nicht anders dokumentiert.

def getEpoch2010():
    return 2455196.5

# https://www.geeksforgeeks.org/how-to-print-to-stderr-and-stdout-in-python/
def print_to_stderr(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stderr) 

def print_to_stdout(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stdout) 

# produces a datetime object output string without microseconds
# seconds get rounded

def nicedatetime(dt):       # yyyy-mm-dd HH:MM:SS
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        string = dt.getStr()
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%Y-%m-%d %H:%M:%S")
    return string

def niceshortdatetime(dt):  # dd.mm.yy HH:MM:SS
    # only meaningful for known century
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        (yy, mm, dd) = dt.getDate()
        yy = int(abs(yy)%100)    # only last 2 ciphers, also for negative yyyy
        (hh, MM, ss) = dt.getTime()
        # round seconds! use %f fmt - hvf 13.04.2023
        #string = "%02d.%02d.%02d %02d:%02d:%02d" % (dd, mm, yy, hh, MM, ss)
        string = "%02d.%02d.%02d %02d:%02d:%02.0f" % (dd, mm, yy, hh, MM, ss)
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%d.%m.%y %H:%M:%S")
    return string

def nicedateshorttime(dt):  # yyyy-mm-dd HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=30)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        (h, m, s) = dt.getTime()
        string = "%4d-%02d-%02d %02d:%02d" % (y, mm, dd, h, m)
    else:
        string = dt.strftime("%Y-%m-%d %H:%M")
    return string

def niceshortdateshorttime(dt):  # dd.mm.yy HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        dtrounded = roundhvfTime(dt, 60)
        #print ("dt, dtrounded = ", dt, dtrounded)
        (y, mm, dd) = dtrounded.getDate()
        yy = abs(y)%100
        dd = floor(dd)
        (h, m, s) = dtrounded.getTime()
        string = "%02d.%02d.%02d %02d:%02d" % (dd, mm, yy, h, m)
    else:
        dt += timedelta(seconds=30)
        string = dt.strftime("%d.%m.%y %H:%M")
    return string

def nicedate(dt):           # yyyy-mm-dd
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=0.5)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        string = "%04d-%02d-%02d" % (y, mm, dd)
    else:
        string = dt.strftime("%Y-%m-%d")
    return string

def nicetime(dt):           # HH:MM:SS
    if isinstance(dt, type(None)): return None
    try:
        if isinstance(dt, hvfTime):
            (h, m, s) = dt.getTime()
            s =round(s, 0)
            string = "%02d:%02d:%02d" % (h, m, s)
        else:
            dt += timedelta(seconds=0.5)
            string = dt.strftime("%H:%M:%S")
    except:
        string = ' '*8
    return string

def nicetimedelta(dt):      # +-HH:MM:SS
    if isinstance(dt, type(None)): return None
    #dt += timedelta(seconds=0.5)
    # treat also -00:00:05 correctly! hvf 16.06.2021
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60

    string = "%1s%2d:%02d:%02d" % (sign, hours, minutes, seconds)
    return string

def nicedeltadayhhmm(dt):   # +-days HH.MM
    if isinstance(dt, type(None)): return None
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    days  = int(totalsecs/86400.); totalsecs -= days*86400.
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60:
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60
    if hours >= 24:
        days += 1; hours -= 24

    string = "%1s%d days %02d:%02d" % (sign, days, hours, minutes)
    return string

def niceminsec(secs):       # +- M:SS  for Time Equation - Zeitgleichung
    if isinstance(secs, type(None)): return None
    sign = ' '
    if secs < 0: sign = '-'; secs = abs(secs)
    minutes = int(secs/60.); secs -= minutes*60.
    seconds = round(secs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60

    string = "%1s%d:%02d" % (sign, minutes, seconds)
    return string

# printZeile(kopf, Liste, format, zeile=79)
# schreibt eine Zeile der max. Laenge zeile mit kopf und den
# Werten aus Liste, formattiert je mit format.
# Padding vor jedem Listenelement
# Die  formattierten Werte muessen gleich lang sein.
# hvf 31.12.2020

def printZeile(kopf, Liste, format, zeile=79):
    rest = zeile
    string = kopf
    rest -= len(string)
    n = len(Liste)      # Anzahl Werte
    test = format % Liste[0]
    testlen = len(test)
    padlen = int((rest - n*testlen)/n)
    padding = ' '*padlen

    for item in Liste:
        string += padding
        string += format % item

    print (string)
    return string

# modulus in Python has always the same sign as b
# Mod(-3, 24) gives 21
def Mod(a, b):
    return (a % b)

# return integer value, closer to 0
def Int(x): return int(x)

# return fractional part, closer to 0 (i.e. may be nagative)
def Frac(x): return (x - int(x))

# Winkelfunktionen fuer Grad Eingaben
def Sin(x):
    return sin(radians(Mod(x, 360.)))

def Cos(x):
    return cos(radians(Mod(x, 360.)))

def Tan(x):
    return tan(radians(Mod(x, 360.)))

def Deg(x):
    return degrees(x)

# calculate angular distance between two objects with (RA1, Decl1), (RA2, Decl2)
# all angles in degrees
# see Meeus, chapter 16, p. 105, Angular Separation
def AngularDistance(star1, star2, verbose=False):
    RA1 = star1[0]; Decl1 = star1[1]
    RA2 = star2[0]; Decl2 = star2[1]
    cosD = Sin(Decl1)*Sin(Decl2) + Cos(Decl1)*Cos(Decl2)*Cos(RA1-RA2)
    D = degrees(acos(cosD))
    # check if angular separation is small
    if D < 0.2:     # below 0.2 degrees, 12'
        Dmean = (Decl1+Decl2)/2.
        Ddelta = Decl1 - Decl2
        D = sqrt(((RA1-RA2)*Cos(Dmean))**2 + Ddelta**2)
    return D


#####################################################################
# Quelle: https://de.wikipedia.org/wiki/Julianisches_Datum
# abgerufen am 20.10.2019
# Name geaendert in getJulianDay (siehe Meeus, chapter 7) - hvf 24.03.2023

# Calculate Julian day for all Gregorian and Julian dates
# Greenwich date is supplied as a datetime object

def getJulianDay(dtGreenwichDate, isGregor=True ):
    '''
    # day may have fractions for h, m, s, us
    # Jan and Feb are months 13 and 14
    year  = dtGreenwichDate.year
    month = dtGreenwichDate.month
    day   = dtGreenwichDate.day
    day  += dtGreenwichDate.hour/24. + dtGreenwichDate.minute/1440. + \
            dtGreenwichDate.second/86400. + \
            dtGreenwichDate.microsecond/86400e6
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    return JD
    '''
    hvfdt = hvfTime(dtGreenwichDate)
    return hvfdt.getJD()

################################################################
# calculate Greenwich date from Julian date
# Ref. Practical Astronomy (see above), p.11
# returns a datetime object for midnight 

def getGreenwichDate(JD):
    '''
    jd1 = JD + 0.5
    I = int(jd1)        # integer part
    F = jd1 - I         # fractional part
    if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
        A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
        B = I + A - int(A/4) + 1
    else:   B = I       # Julian Calendar for earlier dates
    C = B + 1524
    D = int((C-122.1)/365.25)
    E = int(365.25 * D)
    G = int((C-E)/30.6001)
    d = C - E + F - int(30.6001 * G)
    if G < 13.5:    m = G - 1
    else:           m = G - 13
    if m > 2.5:     y = D - 4716
    else:           y = D - 4715

    # add fractional day - must use timedelta for this
    dtGreenwichDate = datetime(y, m, 1) + timedelta(days=d-1)
    return dtGreenwichDate
    '''

    return hvfTime(float(JD))

#   HMSDH convert hours, minutes, seconds to decimal hours
def HMSDH(hh, mm, ss):  # hh may be negative
    habs = abs(hh)
    habs += mm/60. + ss/3600.
    h = habs
    if hh < 0: h = -habs
    return h

#   DHHMS converts decimal hours into hours, minutes, seconds
#   seconds may have fractions, no rounding
def DHHMS(h):   # h may be negative, then only hours will be negative
    sign = 1
    if h < 0: sign = -1
    habs = abs(h)
    hh = int(habs)
    mm = int(Frac(habs)* 60.)
    ss = Frac(habs*60)*60.
    hh = hh * sign
    return (hh, mm, ss)

#   returns HH:MM and a rounded value of h
#   h must be positive or None or ''
def HHMM(hh):
    if (hh == None or hh == ''):
        return('--')
    m = Frac(hh) * 60.0
    h = Int(hh)
    if (m >= 59.5):
        h += 1
        m -= 60.0
    m = int(round(m))
    string = "%02d:%02d = " % (h, m) + str(round(hh, 4))
    return (string)

#   returns HH:MM:SS and a rounded value of h
#   h must be positive or None or ''
def HHMMSS(hh):
    if (hh == 0):
        return('--')
    m = Frac(hh) * 60
    h = Int(hh)
    s = Frac(m) * 60.0
    m = Int(m)
    if (s >= 59.5):
        m += 1
        s -= 60.0
    if (m >= 60):
        h += 1
        m -= 60
    s = int(round(s))
    string = "%02d:%02d:%02d = " % (h, m, s) + str(round(hh, 5))
    return (string)

def hmsstring(hour):
    # hour may also be negative, e.g. if used to convert declination
    (hh, mm, ss) = DHHMS(hour)
    str = "%d:%02d:%02d" % (hh, mm, ss)
    return str


#   local civic time to Universal Time
#   returns a datetime object 

def getUTfromCivicTime(dtLocalCivicTime, zone):    
    # dtLocalCivicTime is a datetime object
    # zone is time zone offset in hours, positive going East
    dtUT = dtLocalCivicTime - timedelta(hours=zone)
    return dtUT

def getCivicTimefromUT(dtUT, zone):
    dtLocalCivicTime = dtUT + timedelta(hours=zone)
    return dtLocalCivicTime

#   UTC to mean local time 
#   dtUT: Universal Time datetime object
#   longitude: Eastern longitude, degrees
#   returns a datetime object, rounded to seconds

def getMeanLocalTime(dtUT, longitude):
    dtmeanLT = dtUT + timedelta(seconds=round(240.*longitude, 0))
    return dtmeanLT

def getUTfromMeanLocalTime(dtmeanLT, longitude):
    dtUT =  dtmeanLT - timedelta(seconds=round(240.*longitude, 0))
    return dtUT

#   datetime.datetime or datetime.time to decimal hours

def getDecimalHour(dttime):
    '''
    hour = dttime.hour + dttime.minute/60. + dttime.second/3600. \
            + dttime.microsecond/3600.E6
    '''
    (hour, min, sec) = hvfTime(dttime).getTime()
    return hour    

# calculate horizontal coordinates (Azimut, Altitude)
# from ecliptic coordinates (Hour angle, Declination)
#   Parameters:
#   Hour    Hour angle counted from meridian to West, degrees
#   Decl    Declination, degrees
#   latitude    degrees
#   Returns:
#   azimut  counted from North, degrees
#   alt     altitude above horizon, degrees
#   x, y, z Cartesian coords in horizontal system, x pointing South, y West
#   https://de.wikipedia.org/wiki/Astronomische_Koordinatensysteme#Umrechnungen

def getAziAltfromHourDecl(Hour, Decl, latitude):
    # Hour angle and Declination into azimut and altitude
    azimut = atan2(Sin(Hour), Cos(Hour)*Sin(latitude)-Tan(Decl)*Cos(latitude))
    alt    = asin(Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour))
    azimut = Mod(degrees(azimut)+180., 360.)
    alt    = degrees(alt)
    # Cartesian coords in horizon system
    x = Cos(alt)*Cos(azimut-180.)
    y = Cos(alt)*Sin(azimut-180.)
    z = Sin(alt)
    # dito calculated from  Hour, Decl, latitude
    xx = -Cos(latitude)*Sin(Decl)+Sin(latitude)*Cos(Decl)*Cos(Hour)
    yy = Cos(Decl)*Sin(Hour)
    zz = Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour)
    xyzDict = {
        "x":    x,
        "y":    y,
        "z":    z, 
        "xx":   xx,
        "yy":   yy,
        "zz":   zz, }
    return (azimut, alt, xyzDict)

# Zodiac signs for Sun and Moon
#   lon ecliptic longitude, degrees
def getSign(lon):
    Signs_de = ("Widder", "Stier", "Zwillinge", "Krebs", "Löwe", "Jungfrau", \
        "Waage", "Skorpion", "Schütze", "Steinbock", "Wassermann", "Fische")
    Signs_en = ("Aries", "Taurus", "Gemini", "Cancer", "Leo", "Virgo", \
        "Libra", "Scorpio", "Sagittarius", "Capricorn", "Aquarius", "Pisces")
    return (Signs_en[int(floor(lon / 30))])

########################################################################
#
# getNutation(dtCivicTime, zone)
# Nutation in ecliptic longitude and ecliptic latitude
# see Meeus, Astronomical Algorithms, chapter 21

def getNutation(dtCivicTime, zone, verbose=False):
    # Epoch is J2000.0
    # JD 2451545
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    T = (JD - 2451545.0)/36525. # Centuries after the Epoch, 21.1

    # Moon Ascending Node
    Omega = 125.04452 - 1934.136261 * T + 0.0020708 * T*T + T*T*T/450000.
    Omega = Mod(Omega, 360.)
    O = radians(Omega)

    # accuracy of 0.5 arcsec in longitude, 0.1 arcsec in latitude
    # Sun Mean Longitude
    LmeanSun = 280.4665 + 36000.7698 * T
    LmeanSun = Mod(LmeanSun, 360.)
    LS = radians(LmeanSun)

    # Moon Mean Longitude
    LmeanMoon = 218.3165 + 481267.8813 * T
    LmeanMoon = Mod(LmeanMoon, 360.)
    LM = radians(LmeanMoon)

    # Effect on ecliptic longitude and latitude, arcsecs
    deltaELong = -17.20*sin(O) -1.32*sin(2.*LS) -0.23*sin(2*LM) +0.21*sin(2*O)
    deltaELat  = 9.20*cos(O) + 0.57*cos(2.*LS) + 0.10*cos(2*LM) -0.09*cos(2*O)

    # obliquity of the ecliptic, 21.2
    eps0 = 23+26/60.+21.448/3600. - 46.8150/3600.*T - 0.00059/3600.*T*T \
            +0.001813/3600*T*T*T
    eps = eps0 + deltaELat/3600.

    if verbose:
        print ("getNutation")
        print ("JD = ", JD)
        print ("T = ", T)
        print ("LmeanSun = ", LmeanSun)
        print ("LmeanMoon = ", LmeanMoon)
        print ("getNutation deltaELong (arcsec) = %.2f" % deltaELong)
        print ("getNutation deltaELat  (arcsec) = %.2f" % deltaELat )
        print ("eps0 (degrees) = %.6f" % eps0, ", ", HHMMSS(eps0))
        print ("eps  (degrees) = %.6f" % eps, ", ", HHMMSS(eps) )

    # return values in degrees
    return (deltaELong/3600., deltaELat/3600., eps)

#################################################################
#   Funktionen für die Berechnung der Zeitgleichung

def getSunStufffromTT(TT, verbose = False):
    # TT Tage nach Epoch2010 inkl. DeltaT Korrektur
    # returns RA, epsilon in degrees

    def atanLage(x, Lage):    # arctangens mit Lageparameter, alle Werte in Grad
        # siehe https://en.wikipedia.org/wiki/Equation_of_time
        # Resultat in Grad
        xrad = radians(x)
        Lagerad = radians(Lage)
        y = atan(xrad)
        valuerad = y + pi * round((Lagerad-y)/pi, 0)
        value = degrees(valuerad)
        return value
    
    def getE(t):    # numerische Exzentrizität zur Zeit t
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        e = 0.01670491 - 4.208E-5*T - 1.26E-7*T*T
        return e
    
    def getEps(t):  # Schiefe der Ekliptik in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
        return eps
    
    def getOmegabar(t):     # Länge des Perihels in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        omegabar = 283.11243767 + 1.720171*T + 0.000452778*T*T
        return omegabar
    
    def getL(t):    # mittlere Länge der Sonne in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        L = 279.557285 + 36000.769585*T + 0.0003025*T*T
        return L
    
    def getC(e, M):    # Mittelpunktsgleichung equation of the center
        # berechnet wahre minus mittlere Anomalie als Fkt der Exzentrizität
        # und der mittleren Anomalie in dritter Näherung in e
        # e numerische Exzentrizität
        # M mittlere Anomalie der Sonne in Grad
        # Resultat in Grad
        # Danby, Fundamentals of Celestial Mechanics, p.438
        Mrad = radians(M)
        Crad = (2*e-e*e*e/4)*sin(Mrad) + (1.25*e*e -11./24.*e**4)*sin(2*Mrad)\
             + 13./12.*e*e*e*sin(3*Mrad) + 103./96.*e**4*sin(4*Mrad)
        C = degrees(Crad)
        return C

    t = TT
    e = getE(t)                 # Exzentrizität der Erdbahn, Zeit t
    eps = getEps(t)             # Schiefe der Ekliptik in Grad, Zeit t
    epsRad = radians(eps)       # Schiefe der Ekliptik in Radians
    omegabar = getOmegabar(t)   # Länge des Perihels in Grad, Zeit t
    L = getL(t)                 # mittlere Länge der Sonne in Grad, Zeit t
    Lhaupt = Mod(L, 360)        # Hauptwert der Länge der Sonne
    M = Lhaupt - omegabar       # Mittlere Anomalie der Sonne in Grad, Zeit t
    Mhaupt = Mod(M, 360)        # Hauptwert der mittleren Anomalie
    C = getC(e, Mhaupt)         # Mittelpunktsgleichung, wahre minus mittlere A
    Lwahr = L + C               # wahre Länge der Sonne in Grad, Zeit t
    LwahrRad = radians(Lwahr)   # wahre Länge der Sonne in Radians
    Lwahrhaupt = Mod(Lwahr, 360)# Hauptwert der wahren Länge
    LwahrHauptRad = radians(Lwahrhaupt)  # dito, in Radians


    # Rektaszension der Sonne in Funktion der wahren Länge der Sonne und der 
    # Schiefe der Ekliptik
    alpha = atanLage(degrees(tan(LwahrRad)*cos(epsRad)), Lwahr)
    alphahaupt = atanLage(degrees(tan(LwahrHauptRad)*cos(epsRad)), Lwahrhaupt)

    Dict = { "Object"   : "SunStuff",
            "TT"        : t,            # Tage nach Epoch2010 inkl. deltaT
            "MPEq"      : C,            # wahre - mittlere Anomalie
            "omegabar"  : omegabar,     # Länge des perihels in Grad, Zeit t
            "M"         : M,            # mittlere Anomalie, Grad
            "nu"        : C+M,          # wahre Anomalie
            "Long"      : Lwahrhaupt,   # wahre Länge der Sonne, Grad
            "eps"       : eps,          # Schiefe der Ekliptik, Grad
            "exc"       : e,            # Exzentrizität der Erdbahn
            }
    if verbose:
        print ("Numerische Exzentrizitaet e = ", "%.6f" % e)
        print ("Omegabar, Laenge des Perihels = ", "%.3f Grad" % omegabar)
        print ("Lhaupt, Hauptwert Mittlere Laenge = ", "%.3f Grad" % Lhaupt)
        print ("M Mittlere Anomalie = ", "%.3f Grad" % M)
        print ("Mhaupt, Hauptwert Mittlere Anomalie = ", "%.3f Grad" % Mhaupt)
        print ("C Mittelpunktsgleichung = ", "%.3f Grad" % C)
        print ("Alpha Wahre Anomalie = ", "%.3f Grad" % alpha)
    return (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict)

#######################################################################    
def Zeitgleichung(JD, verbose=False):
    # berechnet die Zeitgleichung in ganzen Sekunden
    # JD Julianisches Datum
    # verbose aktiviert zusätzlichen Output
    # Resultat in Sekunden ohne Rundung
    # Epoche konsistent 2010.0, 31.12.2009 0h UT, JD 2455196.5
    dtEpoch2010 = datetime(2009, 12, 31)
    tbase = getJulianDay (dtEpoch2010)
    t = JD - tbase
    TT = JD - tbase + getDeltaT(JD)/86400.

    (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict) = getSunStufffromTT(TT, verbose)

    # Zeitgleichung erste Ursache Exzentrizität der Erdbahn (epsilon = 0)
    ZGstar0 = L - Lwahr         # in Grad
    ZG0 = 240. * ZGstar0        # in Sekunden

    # zweite Ursache Anteil Schiefe der Ekliptik
    ZGstareps = Lwahrhaupt - alphahaupt # in Grad
    ZGeps = 240. * ZGstareps            # in Sekunden

    # Summe der beiden Ursachen
    ZGstar = ZGstar0 + ZGstareps        # in Grad
    ZGsecs = ZG0 + ZGeps                # in Sekunden

    # ZGmin, ZGsec for printing purposes 
    ZGstr = niceminsec(ZGsecs)
    if verbose:
        print ("Epoch2010 = ", dtEpoch2010)
        print ("Epoch Julian Date = ", tbase)
        print ("TT = ", TT)
        print ("Schiefe der Ekliptik eps = ", "%.3f Grad" % eps)
        print ("L, Mittlere Laenge = ", "%.3f Grad" % L)
        print ("Lwahr Wahre Sonnenlaenge = ", "%.3f Grad" % Lwahr)
        print ("Lwahrhaupt Hauptwert Wahre Sonnenlaenge = ", "%.3f Grad" \
                    % Lwahrhaupt)
        print ("Alphahaupt Wahre Anomalie = ", "%.3f Grad" % alphahaupt)
        print ("ZGstar0 = ", "%.3f" % ZGstar0, " Grad")
        print ("ZG0 = ", "%.2f" % ZG0, " Sekunden")
        print ("ZGstareps = ", "%.3f" % ZGstareps, " Grad")
        print ("ZGeps = ", "%.2f" % ZGeps, " Sekunden")
        print ("ZGstar = ", "%.3f" % ZGstar, " Grad")
        print ("ZGsecs = ", "%.2f" % ZGsecs, " Sekunden")
        print ("ZG = %s" % ZGstr)

    return (ZGsecs, ZGstr, Dict)

def getZeitgleichung(JD, verbose=False):
    # returns a timedelta object and two int's for printing
    zgsecs, ZGstr, Dict = Zeitgleichung(JD, verbose)
    return (timedelta(seconds=zgsecs), ZGstr, Dict)

# Greenwich Mean Sidereal Time from Julian Date

def getGMSTfromJD(JD):
    '''
    # new Epoch2010, JD=2455196.5
    # JD usually float or int
    # returns a datetime object
    UT = Frac(JD-0.5) * 24.     # UT in hours
    JD = floor(JD-0.5) + 0.5    # JD at 0 hours UT
    # JD 2455196.5 is Epoch 2010.0
    # T is centuries after Epoch
    T  = (JD - 2455196.5) / 36525.
    T0 = 6.636798592 + T * (2400.051341 + T * 0.000025862)
    ST = Mod(T0 + UT * 1.002737909, 24.)
    Gdate = getGreenwichDate(JD)
    return Gdate + timedelta(hours=ST)
    '''

    # Meeus chapter 11 p. 83
    T = (JD - 2451545.0) / 36525.   # (11.1)
    theta0 = 280.46061837 + 360.98564736629 * (JD - 2451545.0) \
            + 0.000387933 * T*T - T*T*T/387110000   # (11.4)
    # normalize to 360 degrees
    theta = Mod(theta0, 360.)
    # add to midnight
    JD0 = floor(JD-0.5) + 0.5
    Gdate = hvfTime(JD0 + theta/360.)
    return Gdate

# Local Mean Sideral Time from GMST, longitude
# returns a datetime object

def getLMSTfromGMST(gmst, longitude):
    # gmst datetime object
    # longitude in degrees
    # returns a datetime object
    lmst = gmst + timedelta(hours=longitude/15.)
    return lmst

# Greenwich Mean Sideral Time from LMST, longitude
# returns a datetime object

def getGMSTfromLMST(lmst, longitude):
    # lmst datetime object
    # longitude in degrees
    # returns a datetime object
    #gmst = datetime.time(hour=lmst) - timedelta(hours=longitude/15.)
    gmst = lmst - timedelta(hours=longitude/15.)
    return gmst
    

# deltaT, the difference between Terrestrial Time TT and
# Universal Time 
# see https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
# for Julian Date JD

def getDeltaT(JD): 
    # convert JD into year, month, day
    '''
    dtymd = getGreenwichDate(JD)
    year = dtymd.year
    month = dtymd.month
    y = year + (month-0.5)/12.
    # print("getDeltaT year, month, y = %d %d %.2f" % (year, month, y))
    '''
    (year, month, day) = hvfTime(JD).getDate()
    # We define the decimal year "y" as follows:
    y = year + (month - 0.5)/12

    if year < -500:
        u = (year-1820.)/100.
        deltaT = -20 + 32*u*u
    elif year < 500:
        u = y/100.
        deltaT = 10583.6 - 1014.41*u + 33.78311*u*u - 5.952053*u**3 \
                - 0.1798452*u**4 + 0.022174192*u**5 + 0.0090316521*u**6
    elif year < 1600:
        u = (y-1000.)/100.
        deltaT = 1574.2 - 556.01*u + 71.23472*u*u + 0.319781*u**3 \
                - 0.8503463*u**4 - 0.005050998*u**5 + 0.0083572073*u**6
    elif year < 1700:
        t = y - 1600.
        deltaT = 120. -0.9808*t - 0.01532*t*t + t**3/7129
    elif year < 1800:
        t = y - 1700.
        deltaT = 8.83 + 0.1603*t - 0.0059285*t*t + 0.00013336*t**3 \
                - t**4/1174000.
    elif year < 1860:
        t = y - 1800.
        deltaT = 13.72 - 0.332447*t + 0.0068612*t*t + 0.0041116*t**3 \
                - 0.00037436*t**4 + 0.0000121272*t**5 \
                - 0.0000001699*t**6 + 0.000000000875*t**7
    elif year < 1900:
        t = y - 1860.
        deltaT = 7.62 + 0.5737*t - 0.251754*t*t + 0.01680668*t**3 \
                - 0.0004473624*t**4 + t**5/233174.
    elif year < 1920:
        t = y - 1900.
        deltaT = -2.79 + 1.494119*t - 0.0598939*t*t + 0.0061966*t**3 \
                -0.000197*t**4
    elif year < 1941:
        t = y - 1920.
        deltaT = 21.20 + 0.84493*t - 0.076100*t*t + 0.0020936*t**3
    elif year < 1961:
        t = y - 1950.
        deltaT = 29.07 + 0.407*t - t*t/233. + t**3/2547.
    elif year < 1986:
        t = y - 1975.
        deltaT = 45.45 + 1.067*t - t*t/260. - t**3/718.
    elif year < 2005:
        t = y - 2000.
        deltaT = 63.86 + 0.3345*t - 0.060374*t*t + 0.0017275*t**3 \
                + 0.000651814*t**4 + 0.00002373599*t**5
    elif year < 2050:
        t = y - 2000.
        deltaT = 62.92 + 0.32217*t + 0.005589*t*t
    elif year < 2150:
        deltaT = -20. + 32*((y-1820.)/100)**2 - 0.5628*(2150.-y)
    else:   # after 2150
        u = (year-1820.)/100.
        deltaT = -20. + 32.*u*u
        # https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
        # small correction beyond 2150 (only needed for Perihelion calc's)
        c = -0.000012932 * (y - 1955.)**2
        deltaT += c

    return deltaT

def getEps(t):  # Schiefe der Ekliptik in Grad
    # t in Tage nach Epoch2010
    # Resultat in Grad
    T = t / 36525
    eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
    return eps


#######################################################################
# Jean Meeus, Astronomical Algorithms, Willmann-Bell, Richmond VA, 1991
#
# hvf 03.12.2020

# quadratic interpolation
# xi 3 Stuetzstellen, yi 3 Funktionswerte
# Meeus p. 25
def interp(xi, yi, x):
    a10 = (yi[1]-yi[0])/(xi[1]-xi[0])
    a21 = (yi[2]-yi[1])/(xi[2]-xi[1])
    a210 = (a21-a10)/(xi[2]-xi[0])
    res = yi[0] + a10*(x-xi[0]) + a210*(x-xi[0])*(x-xi[1])
    return res

#############################################################################
#! /usr/bin/env python3
# Perigee and Apogee of the Moon, Meeus Chapter 48
# hvf 28.02.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87

from datetime import datetime, date, timedelta
from math import sin, acos, cos, pi, radians, degrees, atan, atan2
from math import asin, tan, ceil, floor, sqrt, fabs
import sys

VersionMoonPeriApo ="V3.1 hvf 2023-11-29 MoonPeriMain.py based on Meeus, Astronomical Algorithms"
VersionList.append(VersionMoonPeriApo)

def Usage(progname):
    print ("Usage: %s [UTC|TT] year [year ...]" % progname)
    print ("Usage: %s [UTC|TT] year-year ...]" % progname)
    sys.exit(0)

def isPerigee(k):   # k must be integer
    test = (abs(k - int(k)) < 0.001)
    #if verbose: print ("isPerigee k, test = ", k, test)
    return test     # True or False

def isApogee(k):  # k must be half-integer
    test = (abs(abs(k - int(k)) - 0.5) < 0.001)
    #if verbose: print ("isApogee k, test = ", k, test)
    return test     # True or False 

# getPeriJDEbyK (k, verbose = False)
# k counts the number of perigees since 1999-12-22
# Meeus chapter 48
# returns the Julian Ephemeris Days (JDE) time of the perigee, apogee
def getPeriJDEbyK(k, verbose = False):
    # k integer:        Perigee
    # k integer + 0.5 : Apogee

    T  = k / 1325.55            # 48.3, Julian centuries after Epoch2000

    # mean perigee, apogee (48.1)
    JDE = 2451534.6698 + 27.55454988 * k \
                        - 0.0006886 * T*T \
                        - 0.000001098 * T**3 \
                        + 0.0000000052 * T**4
    JDE0 = JDE

    # Moon's mean elongation at time JDE 
    D = 171.9179  + 335.9106046 * k \
                    - 0.0100250 * T*T \
                    - 0.00001156 * T**3 \
                    + 0.000000055 * T**4
    D = Mod(D, 360.)

    # Sun's Mean anomaly
    M = 347.3477   + 27.1577721 * k \
                    - 0.0008323 * T*T \
                    - 0.0000010 * T**3 
    M = Mod(M, 360.)

    # Moon's argument of Latitude
    F = 316.6109  + 364.5287911 * k \
                    - 0.0125131 * T*T \
                    - 0.0000148 *T**3 
    F = Mod(F, 360.)

    # sumP: perigee sum (in days)
    # sumA: apogee sum (in days)
    # sum HPP: perigee parallax (in arcsec), table 48.B
    # sum HPA: apogee parallax (in arcsec), table 48.B
    sumHPP = 3629.215;      sumHPA = 3245.251
    sumP = 0.;                                  sumA = 0.
    arg = 2*D;      sumP += -1.6769 * Sin(arg); sumA += 0.4392 * Sin(arg)
    sumHPP +=  63.224 * Cos(arg);    sumHPA += -9.147 * Cos(arg)
    arg = 4*D;      sumP +=  0.4589 * Sin(arg); sumA += 0.0684 * Sin(arg)
    sumHPP +=  -6.990 * Cos(arg);    sumHPA +=  0.355 * Cos(arg)
    arg = 6*D;      sumP += -0.1856 * Sin(arg); sumA += 0.0144 * Sin(arg)
    sumHPP +=   1.927 * Cos(arg);    sumHPA +=  0.052 * Cos(arg)
    arg = 8*D;      sumP +=  0.0883 * Sin(arg); sumA += 0.0035 * Sin(arg)
    sumHPP +=  -0.702 * Cos(arg);    sumHPA +=  0.010 * Cos(arg)
    arg = 2*D-M;    sumP += (-0.0773 + 0.00019 * T) * Sin(arg); \
                                sumA += (0.0426 - 0.00011 * T) * Sin(arg)
    sumHPP += (2.834 - 0.0071 * T) * Cos(arg);    sumHPA +=  0.159 * Cos(arg)
    arg = M;        sumP += ( 0.0502 - 0.00013 * T) * Sin(arg); \
                                sumA += (0.0456 - 0.00011 * T) * Sin(arg)
    sumHPP += ( 0.696 - 0.0017 * T) * Cos(arg);    
    sumHPA += (-0.656 + 0.0016 * T) * Cos(arg)
    arg = 10*D;     sumP += -0.0460 * Sin(arg); sumA +=  0.0009 * Sin(arg)
    sumHPP +=  0.297 * Cos(arg)
    arg = 4*D-M;    sumP += ( 0.0422 - 0.00011 * T) * Sin(arg); \
                                sumA +=  0.0113 * Sin(arg)
    sumHPP += (-0.629 + 0.0016 * T) * Cos(arg);    sumHPA += 0.065 * Cos(arg)
    arg = 6*D-M;    sumP += -0.0256 * Sin(arg); sumA +=  0.0034 * Sin(arg)
    sumHPP +=  0.260 * Cos(arg);    sumHPA +=  0.014 * Cos(arg)
    arg = 12*D;     sumP +=  0.0253 * Sin(arg); sumA +=  0.0003 * Sin(arg)
    sumHPP += -0.138 * Cos(arg)
    arg = D;        sumP +=  0.0237 * Sin(arg); sumA += -0.0189 * Sin(arg)
    sumHPP += -1.263 * Cos(arg);    sumHPA += -0.841 * Cos(arg)
    arg = 8*D-M;    sumP +=  0.0162 * Sin(arg); sumA +=  0.0011 * Sin(arg)
    sumHPP += -0.127 * Cos(arg)
    arg = 14*D;     sumP += -0.0145 * Sin(arg)
    sumHPP +=  0.068 * Cos(arg)
    arg = 2*F;      sumP +=  0.0129 * Sin(arg); sumA +=  0.0212 * Sin(arg)
    sumHPP += -0.690 * Cos(arg);    sumHPA +=  0.697 * Cos(arg)
    arg = 3*D;      sumP += -0.0112 * Sin(arg); sumA += -0.0017 * Sin(arg)
    sumHPP +=  0.201 * Cos(arg)
    arg = 10*D-M;   sumP += -0.0104 * Sin(arg); sumA +=  0.0004 * Sin(arg)
    sumHPP +=  0.067 * Cos(arg)
    arg = 16*D;     sumP +=  0.0086 * Sin(arg)
    sumHPP += -0.035 * Cos(arg)
    arg = 12*D-M;   sumP +=  0.0069 * Sin(arg)
    sumHPP += -0.038 * Cos(arg)
    arg = 5*D;      sumP +=  0.0066 * Sin(arg); sumA += -0.0004 * Sin(arg)
    sumHPP += -0.079 * Cos(arg)
    arg = 2*D+2*F;  sumP += -0.0053 * Sin(arg); sumA +=  0.0047 * Sin(arg)
    sumHPP +=  0.104 * Cos(arg);    sumHPA +=  0.031 * Cos(arg)
    arg = 18*D;     sumP += -0.0052 * Sin(arg)
    sumHPP +=  0.019 * Cos(arg)
    arg = 14*D-M;   sumP += -0.0046 * Sin(arg)
    sumHPP +=  0.023 * Cos(arg)
    arg = 7*D;      sumP += -0.0041 * Sin(arg)
    sumHPP +=  0.037 * Cos(arg)
    arg = 2*D+M;    sumP +=  0.0040 * Sin(arg); sumA +=  0.0005 * Sin(arg)
    sumHPP += -0.161 * Cos(arg);    sumHPA +=  0.043 * Cos(arg)
    arg = 20*D;     sumP +=  0.0032 * Sin(arg)
    sumHPP += -0.010 * Cos(arg)
    arg = D+M;      sumP += -0.0032 * Sin(arg); sumA +=  0.0036 * Sin(arg)
    sumHPP +=  0.157 * Cos(arg);    sumHPA +=  0.127 * Cos(arg)
    arg = 16*D-M;   sumP +=  0.0031 * Sin(arg)
    sumHPP += -0.014 * Cos(arg)
    arg = 4*D+M;    sumP += -0.0029 * Sin(arg)
    sumHPP +=  0.054 * Cos(arg)
    arg = 9*D;      sumP +=  0.0027 * Sin(arg)
    sumHPP += -0.020 * Cos(arg)
    arg = 4*D+2*F;  sumP +=  0.0027 * Sin(arg); sumA +=  0.0013 * Sin(arg)
    sumHPP += -0.037 * Cos(arg)

    arg = 2*D-2*M;  sumP += -0.0027 * Sin(arg); sumA +=  0.0022 * Sin(arg)
    sumHPP +=  0.104 * Cos(arg);    sumHPA +=  0.022 * Cos(arg)
    arg = 4*D-2*M;  sumP +=  0.0024 * Sin(arg); sumA +=  0.0010 * Sin(arg)
    sumHPP += -0.038 * Cos(arg)
    arg = 6*D-2*M;  sumP += -0.0021 * Sin(arg); sumA +=  0.0004 * Sin(arg)
    sumHPP +=  0.022 * Cos(arg)
    arg = 22*D;     sumP += -0.0021 * Sin(arg)
    arg = 18*D-M;   sumP += -0.0021 * Sin(arg)
    arg = 6*D+M;    sumP +=  0.0019 * Sin(arg)
    sumHPP += -0.025 * Cos(arg)
    arg = 11*D;     sumP += -0.0018 * Sin(arg)
    sumHPP +=  0.011 * Cos(arg)
    arg = 8*D+M;    sumP += -0.0014 * Sin(arg)
    sumHPP +=  0.012 * Cos(arg)
    arg = 4*D-2*F;  sumP += -0.0014 * Sin(arg); sumA += -0.0004 * Sin(arg)
    sumHPP +=  0.013 * Cos(arg)
    arg = 6*D+2*F;  sumP += -0.0014 * Sin(arg); sumA +=  0.0004 * Sin(arg)
    sumHPP +=  0.017 * Cos(arg)
    arg = 3*D+M;    sumP +=  0.0014 * Sin(arg); sumA +=  0.0007 * Sin(arg)
    sumHPP += -0.030 * Cos(arg)
    arg = 5*D+M;    sumP += -0.0014 * Sin(arg)
    sumHPP +=  0.010 * Cos(arg)
    arg = 13*D;     sumP +=  0.0013 * Sin(arg)
    arg = 20*D-M;   sumP +=  0.0013 * Sin(arg)
    arg = 3*D+2*M;  sumP +=  0.0011 * Sin(arg)
    arg = 2*D-2*F+M;sumP += -0.0011 * Sin(arg)
    arg = D+2*M;    sumP += -0.0010 * Sin(arg)
    arg = 2*F-M;    sumP +=  0.0008 * Sin(arg)
    sumHPP +=  0.014 * Cos(arg)
    arg = 2*D+4*F;  sumP +=  0.0007 * Sin(arg)
    arg = 2*F-2*M;  sumP += -0.0006 * Sin(arg)
    arg = 2*D-2*F+2*M;sumP += -0.0006 * Sin(arg)
    arg = 24*D;     sumP +=  0.0006 * Sin(arg)
    arg = 4*D-4*F;  sumP +=  0.0005 * Sin(arg)
    arg = 2*D+2*M;  sumP +=  0.0005 * Sin(arg); sumA +=  0.0005 * Sin(arg)
    sumHPP += -0.023 * Cos(arg);    sumHPA +=  0.019 * Cos(arg)
    arg = D-M;      sumP += -0.0004 * Sin(arg); sumA += -0.0003 * Sin(arg)
    sumHPP +=  0.029 * Cos(arg)
    arg = 2*D-2*F;                              sumA += -0.0034 * Sin(arg)
    sumHPP += -0.392 * Cos(arg);    sumHPA += -0.023 * Cos(arg)
    arg = 2*M;                                  sumA +=  0.0006 * Sin(arg)
    sumHPP +=  0.023 * Cos(arg);    sumHPA += -0.016 * Cos(arg)
    arg = 2*F+M;                                sumA +=  0.0003 * Sin(arg)
    arg = 2*D+2*F-M;                            sumA +=  0.0003 * Sin(arg)
    arg = 2*D-2*F-M;sumHPP += -0.021 * Sin(arg)

    MoonDistP = 6378.14 / Sin(sumHPP/3600.)
    MoonDistA = 6378.14 / Sin(sumHPA/3600.)

    if isPerigee(k):
        JDE += sumP
        MoonDist = MoonDistP
        if verbose: print ("Perigee")
    elif isApogee(k):
        JDE += sumA
        MoonDist = MoonDistA
        if verbose: print ("Apogee")
    else:
        print ("k not for Perigee or Apogee: ", k)
    

    if verbose:
        print ("getPeriJDEbyK: k = %.2f" % k)
        print ("JDE0 = %.6f" % JDE0)
        print ("M(Sun), D, F, sumP, sumA = %.7f, %.4f, %.4f, %.4f, %.4f" \
            % (M, D, F, sumP, sumA)) 
        print ("sumP = %.5f, sumA = %.5f" % (sumP, sumA))
        print ("sumHPP = %.5f, sumHPA = %.5f" % (sumHPP, sumHPA))
        print ("JDE = %.5f" % JDE)
    return (JDE, MoonDist)

# getMoonPeriApo(year, useUT)
#   returns:
#   Perigee
#   Apogee
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonPeriApo(year, useUT):
    # chapter 48 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-1999.97)*13.2555  # counting perigee zero 48.2
    if (k0 == int(k0)): kstart = k0
    else: kstart  = ceil(k0) - 1     # start with Perigee
    
    EventList = []
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    print (" Moon Perigees and Apogees for the Year %4d" % year, comment)
    print (" -------------------------------------------", "-"*len(comment))
    print ()
    print ("             Perigee                          Apogee")
    
    for kint in range(kstart, kstart+16):
        dateList = ['', '']
        for k in (kint, kint+0.5):
            T  = k / 1325.55            # Julian centuries
            #y  = year
            y = int(T * 100 + 1999.97)
            (JDE, MoonDist) = getPeriJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT
        
            # call getMoonData for distance of moon
            #(MoonRa, MoonDecl, MoonDict) = getMoonData(dt-deltaT, 0.0, False)
            #MoonDist = MoonDict["MoonDist"]
            # Super Moon, Mini Moon
            suffix = "   "
            if MoonDist <= 360000. : suffix = " ++" # Super Moon
            if MoonDist >= 405000. : suffix = " --" # Mini Moon

    
            (yyyy, dummy, dummy) = dt.getDate()
            if isPerigee(k) and yyyy == year: 
                #print ("Perigee  ", niceshortdatetime(dt))
                dateList[0] = niceshortdatetime(dt) + " %.0f"%MoonDist + suffix
            if isApogee(k): 
                #print ("Apogee   ", niceshortdatetime(dt))
                #if dt.year == year: 
                if yyyy == year: 
                    dateList[1] = niceshortdatetime(dt) + " %.0f"%MoonDist + suffix
                if len (dateList[0]) + len(dateList[1]) > 0: 
                    EventList.append(dateList)
                dateList = ['', '']
    
    return EventList
#########################################################      

# getMoonPeriApoxxx(dtCivicTime, zone, verbose=False)
#   returns:
#   Perigee
#   Apogee
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonPeriApoxxx(dtCivicTime, zone, verbose=False):
    # chapter 48 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 

    #k0 = (year-1999.97)*13.2555  # counting perigee zero 48.2
    k0 = (Ty+0.03)*13.2555  # counting perigee zero 48.2
    k  = ceil(k0) - 2     # start with Perigee
    kmin = k
    kmax = k + 4
    EventList = []
    useUT = True
    
    for kint in range(kmin, kmax):
        for k in (kint, kint+0.5):
            T  = k / 1325.55            # Julian centuries
            y = int(T * 100 + 1999.97)
            (JDE, MoonDist) = getPeriJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT
            dtCT = getCivicTimefromUT(dt, zone)

            # call getMoonData for declination of Moon
            (dummy, dummy, MoonDict) = getMoonData(dt, 0, False)
            MoonDecl = MoonDict["Decl"]
            # Super Moon, Mini Moon
            suffix = MoonDict["suffix"]
            mode = ''
            if isPerigee(k): 
                mode = 'Perigee'
            if isApogee(k): 
                mode = 'Apogee'
            Event = {'Object': 'Perigee',
                    'dt': dt,
                    'dtCivicTime' : dtCT,
                    'dist': MoonDist,
                    'decl': MoonDecl,
                    'mode': mode,
                    'suffix': suffix,
                    'Dict': MoonDict}
            EventList.append(Event)
    return EventList
#########################################################      
#########################################################33      
#! /usr/bin/env python3
# Perigee and Apogee of the Moon, Meeus Chapter 48
# hvf 28.02.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87

VersionMoonPeriApo ="1.2 hvf 23.09.2023 based on Meeus, Astronomical Algorithms"

print ()
print ("%s Version %s" % (sys.argv[0], VersionMoonPeriApo))
print ()

# read years
useUT = True
foundOne = False
while len(sys.argv) > 1:
    arg = sys.argv[1]
    if arg.upper() == 'UT' or arg.upper() == 'UTC':
        useUT = True
        del sys.argv[1]
    elif arg.upper() == 'TT':
        useUT = False
        del sys.argv[1]
    else:
        yearList = re.split(r'-', arg)
        # we have to consider negative years as well!
        if len(yearList) == 2:  # startYear-lastYear, both positive 
            try:
                if len(yearList[0]) > 0:
                    startYear = int(yearList[0])
                    lastYear  = int(yearList[1])
                else:           # -startyear, negative, e.g. -1234
                    startYear = -int(yearList[1])
                    lastYear  = startYear
            except:
                Usage(sys.argv[0])
        elif len(yearList) == 3:  # -0003-0003 from -3 to +3
            try:
                startYear = -int(yearList[1])
                lastYear  = int(yearList[2])
            except:
                Usage(sys.argv[0])
        elif len(yearList) == 4:  # -0003--0002 from -3 to -2
            try:
                startYear = -int(yearList[1])
                lastYear  = -int(yearList[3])
            except:
                Usage(sys.argv[0])
        else:
            try:
                startYear = int(yearList[0])
                lastYear  = startYear
            except:
                Usage(sys.argv[0])

        for year in range(startYear, lastYear+1):
            EventList = getMoonPeriApo(year, useUT)
            while len(EventList) > 0:
                printZeile("", EventList[0], "%30s", 66)
                del EventList[0]
            print()

            foundOne = True
        del sys.argv[1]
if not foundOne: Usage(sys.argv[0])

