#! /usr/bin/env python3
# Julian to Greenwich and vice versa
# with my own hvfTime class, works also for BC dates and leap years
# hvf 16.09.2021 26.12.2021 11.11.2022 13.04.2023
# make hvfTime() hashable -- hvf 02.10.2023

from datetime import datetime, timedelta
from math import floor
import sys
import re
import dateutil.parser

VersionJulian = "V1.8 hvf 2023-10-02 HVFTime.py based on Meeus"


# JulianDay from year, month, day -  hvf 16.11.2021
def getJulianDayfromYMD(year, month, day, hour=0, minute=0, second=0, 
                        verbose = False):
    if verbose: 
        print ("getJulianDayfromYMD: ", year, month, day, hour, minute, second)
    day  += hour/24. + \
            minute/1440. + \
            second/86400.
    isGregor = False    # 04.10.1582 is last day before Gregorian calendar
    if year > 1582 or year == 1582 and (month > 10 or month == 10 and day > 4):
        isGregor = True
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    # JD 0 is -4712-01-01.5 i.e. -4712-01-01 12h
    # N.B. JD 1524.5 is -4708-03-05 00h
    if verbose: print ("JD = ", JD)
    return JD


def roundhvfTime(hvfDT = None, roundTo = 3600):
    # Round a hvfTime object to anytime lapse in seconds
    # hvfDT: hvftime object, default now
    # roundTo: closest number of seconds to round to, default 1 hour.
    # Author: Harald von Fellenberg 2021
    # Problem: 2021-11-21 1223 returns 2021-11-21 12:22:60.000
    # hvf 2022-09-09
    verbose=False
    if not isinstance(hvfDT, hvfTime): hvfDT = hvfTime.now()
    h, m, s = hvfDT.getTime()
    seconds = h*3600 + m*60 + s
    fraction = seconds - floor(seconds)     # fractional seconds
    if verbose:
        print ("seconds, fraction, floor, rounding = ", seconds, fraction, floor(seconds))
    seconds = floor(seconds)
    rounding = (seconds+roundTo/2) // roundTo * roundTo
    if verbose: print (rounding)
    return hvfDT + timedelta(0, rounding-seconds, -fraction*1e6)

# returns yyyymmdd hh for a hvfTime object, rounded
def nicehvfTimehh (hvfDT):
    hvfDTrounded = roundhvfTime(hvfDT)
    yy, mm, dd = hvfDTrounded.getDate()
    hh, dummy, dummy = hvfDTrounded.getTime()

    s = ''
    if yy >= 0:
        s = " %04d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    else:
        s = "%05d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    return s

#######################################################################
# replaces astropy.time.Time since that does not support leap days 
# before 100BC
# values are represented in Julian days
# creator parameter can be:
# - a hvfTime object
# - a JD JulianDay object (float)
# - a datetime object
# - date and time as six values yy, mm, dd, [hh, MM, ss]
#   dd and ss may have fractions
# hvf 26.12.2021

class hvfTime:
    def __init__(self, *arg):
        # print ("__init__: arg = ", arg)
        # argument is a hvfTime object
        if isinstance(arg[0], hvfTime):
            JD = arg[0].getJD()
            self.__JD = JD
        # argument is a JD
        elif isinstance(arg[0], float) and len(arg) == 1:
            JD = arg[0]
            self.__JD = JD
        # argument is a datetime object
        elif isinstance(arg[0], datetime):
            dt = arg[0]
            yy = dt.year
            mm = dt.month
            dd = dt.day
            hh = dt.hour
            MM = dt.minute
            ss = dt.second + dt.microsecond*1e-6
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        # argument is yy mm dd [hh MM ss]
        elif (isinstance(arg[0], int) or isinstance(arg[0], float)) and \
            len(arg) >= 3:
            hh = 0; MM = 0; ss = 0
            yy = arg[0]; mm = arg[1]; dd = arg[2]
            if len(arg) > 3: hh = arg[3]
            if len(arg) > 4: MM = arg[4]
            if len(arg) > 5: ss = arg[5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD

        # argument is (yy mm dd [hh MM ss])
        elif isinstance(arg[0], tuple) or isinstance(arg[0], list):
            hh = 0; MM = 0; ss = 0
            yy = arg[0][0]; mm = arg[0][1]; dd = arg[0][2]
            if len(arg[0]) > 3: hh = arg[0][3]
            if len(arg[0]) > 4: MM = arg[0][4]
            if len(arg[0]) > 5: ss = arg[0][5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        else:
            print ("hvfTime instance not supported: ", arg[0])
            sys.exit(1)
            
    def getJD(self):
        return self.__JD

    def getDate(self, verbose = False):  # returns (yy, mm, dd), 
                                        # where dd may have fraction
        # works also for negative JDs! hvf 29.12.2021
        JD = self.__JD
        jd1 = JD + 0.5
        I = floor(jd1)      # integer part, may be negative
        F = jd1 - I         # fractional part, always positive
        if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
            A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
            B = I + A - int(A/4) + 1
        else:   B = I       # Julian Calendar for earlier dates
        C = B + 1524
        D = floor((C-122.1)/365.25)
        E = floor(365.25 * D)
        G = floor((C-E)/30.6001)
        d = C - E + F - floor(30.6001 * G)
        if G < 13.5:    m = G - 1
        else:           m = G - 13
        if m > 2.5:     y = D - 4716
        else:           y = D - 4715

        if verbose:
            print ("getDate: JD, jd1, I, F = ", JD, jd1, I, F)
            print ("getDate: B, C, D, E, G = ", B, C, D, E, G)
            print ("getDate: y, m, d       = ", y, m, d)
        return (y, m, d)

    def getTime(self):      # seconds have 3 digi fractions (millisec)
        dummy, dummy, t = self.getDate()
        t -= floor(t)
        # t is fraction of day - convert to sec and round
        daysec = round(t * 86400., 3)
        hms = daysec / 3600.
        #hms = t*24.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 3)
        #s = int((ms - m) * 60)
        #print ("daysec, hms, h, ms, m, s = ", daysec, hms, h, ms, m, s)
    
        return (h, m, s)

    def __str__(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, hh, mm, ss)

    def getStr(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        daysec = round(ss + mm*60 + hh*3600, 0)
        hms = daysec / 3600.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 0)

        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, h, m, s)

    # Ernesti, Python3, p. 380
    # hvfTime = hvfTime + timedelta
    # hvfTime = hvfTime + day   (float or int)
    def __add__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__add__: wrong type ", type(td))
            sys.exit(1)
        return hvfTime(self.__JD + sec/86400.)

    # hvfTime += timedelta
    def __iadd__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__iadd__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD += sec/86400.
        return self

    # hvfTime = hvfTime - timedelta
    # and   n = hvfTime - hvfTime       n in days
    def __sub__(self, td):
        # subtract a timedelta object 
        if isinstance (td, timedelta):
            sec = td.total_seconds()
            return hvfTime(self.__JD - sec/86400.)
        elif isinstance(td, hvfTime):
            # return JDs in days as difference between two hvfTime objects
            return self.__JD - td.getJD()
        else:
            print ("hvfTime.__sub__: wrong type ", type(td), td)
            sys.exit(1)

    # hvfTime -= timedelta
    def __isub__(self, td):
        # subtract a timedelta object 
        try:    
            sec = td.total_seconds()
        except:
            print ("hvfTime.__isub__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD -= sec/86400.
        return self

    # hvfTime < hvfTime
    def __lt__(self, other):
        return self.__JD < other.getJD()

    # hvfTime <= hvfTime
    def __le__(self, other):
        return self.__JD <= other.getJD()

    # hvfTime == hvfTime
    def __eq__(self, other):
        return self.__JD == other.getJD()

    # hash - makes hvfTime hashable, i.e. as key in dict or set 
    def __hash__(self):
        return hash(self.__JD)

    # hvfTime != hvfTime
    def __ne__(self, other):
        return self.__JD != other.getJD()

    # hvfTime > hvfTime
    def __gt__(self, other):
        return self.__JD > other.getJD()

    # hvfTime >= hvfTime
    def __ge__(self, other):
        return self.__JD >= other.getJD()

    def now():  # returns current time (h, m, s), uses datetime.now()
                # roundedd to integer seconds
        return hvfTime(datetime.now()).getTime()

    def today():    # returns current date (y, m, d), uses datetime.today()
                    # d has fractions
        return hvfTime(datetime.today()).getDate()
    
    def getDatetime(self):
        (yy, mm, dd) = self.getDate()
        dd = floor(dd)
        (HH, MM, SS) = self.getTime()
        SS = int(round(SS, 0))
        return datetime(yy, mm, dd, HH, MM, SS)
#######################################################################

# parse date time
# returns a hvfTime object
# hvf 19.11.2021 25.12.2021
# allowed patterns:
# 2021-11-21
# 2021-11-21 12:23:45.635
# 2021-11-21T12:23:45.635
# 2021-11-21 122345.635
# 2021-11-21 122345
# 2021-11-21 1223
# 2021-11-21 12
# -4712-1-1.5 
# 25.12.2021 12:34:56.789
# 1.1.2022
# 2451545.0 (a Julian Day)
# works for dates between -9999-01-01 and 9999-12-31

def parseDateTime(s, verbose = False):
    HH = MM = SS = 0
    # check if two parts separated by blank or T
    if verbose: print ("parseDateTime input: ", s)
    s = s.strip()
    dpart = s
    tpart = None
    parts = re.split (r"[\sT]", s)
    if len(parts) >= 2:
        dpart = parts[0]    # date part
        tpart = parts[1]    # time part
        if verbose:
            print ("date part = ", dpart)
            print ("time part = ", tpart)

    # parse date  2021-12-24 or  4712-1-1.5
    if re.fullmatch(r'\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        ymd = re.split(r'-', dpart)
        yy = ymd[0]
        mm = ymd[1]
        dd = ymd[2]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date 24.12.2021 or 1.1.2022
    elif re.fullmatch(r'\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        ymd = re.split(r'\.', dpart)
        yy = ymd[2]
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -4712-01-01.5
    elif re.fullmatch(r'-\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        symd = re.split(r'-', dpart)
        yy = -int(symd[1])
        mm = symd[2]
        dd = symd[3]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -24.12.2021 or -1.1.2022
    elif re.fullmatch(r'-\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        symd = re.split(r'-', dpart)
        ymd = re.split(r'\.', symd[1])
        yy = -int(ymd[2])
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    else:
        print ("parseDateTime format unknown: ", s)
    # parse date  2021-12-24 or  4712-1-1.5
    # parse date 24.12.2021 or 1.1.2022
    # parse date -4712-01-01.5
    # parse date -24.12.2021 or -1.1.2022
        print ("Allowed date formats include 2021-12-24, 4712-1-1.5,")
        print ("    24.12.2021, 1.1.2022, -4712-01-01.5, -24.12.2021, -1.1.2022")
        print ("Allowed range of years is from -9999 to 9999")
        sys.exit(1)

    # parse time using dateutil.parser.isoparse - this is easier
    # than checking all the different possible input formats
    if tpart:
        deltaT = dateutil.parser.isoparse("2021-12-24 " + tpart) \
                - datetime(2021, 12, 24)
        Secs = round(deltaT.total_seconds(), 3)    # avoid rounding error
        HH = Secs // 3600
        Secs -= HH * 3600
        MM = Secs // 60
        Secs -= MM * 60
        SS = Secs
        if verbose:
            print ("HH, MM, SS = ", HH, MM, SS)

    #return hvfTime object
    return hvfTime(int(yy), int(mm), float(dd), int(HH), int(MM), float(SS))

#######################################################################
# catculate weekday for any date from -9999-01-01 to 9999-12-31
# Meeus 1991 p. 65

def getDayofWeek(dt):
    name =["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday",
            "Friday", "Saturday"]
    hvfdt = hvfTime(dt)
    (yy, mm, dd) = hvfdt.getDate()
    hvfdtMidnight = hvfTime(yy, mm, int(dd), 0, 0, 0.0)
    JD0 = hvfdtMidnight.getJD()
    weekday = int(JD0 + 1.51) % 7
    return name[weekday]

#######################################################################

#!/usr/bin/env python3
# -*- coding: iso-8859-1 -*-

# Der folgende Python-Code fuer Sonne und Mond 
# basiert auf der Portierung des java-Scripts von Arnold Barmettler
# http://lexikon.astronomie.info/java/sunmoon/sunmoon.html
#
#     Entfernen Sie folgende Informationen auf keinen Fall: / 
#     Do not remove following text:
#     Source code by Arnold Barmettler, www.astronomie.info / www.CalSky.com
#     and Harald von Fellenberg, www.hvf.ch
#     based on algorithms by Peter Duffett-Smith's great and easy book
#     'Practical Astronomy with your Calculator or Spreadsheet'.
#     and Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Totale Ueberarbeitung durch Harald von Fellenberg, Dezember 2020
#   Portierung auf Python 3
#   Ergaenzung fuer Mitternachtssonne, Polarnacht
#   Ergaenzungen fuer Daemmerung, Hoehe der Sonne beim Transit tags und nachts
#   Ergaenzung mit Zeitgleichung, wahre und mittlere Ortszeit
#   Azimut für Sonnenaufgang, Sonnenuntergang
#   Liste einiger Städte mit ihren Koordinaten
#   Korrekte Berechnung in allen Zeitzonen
#   Behandlung aller Planeten
#
#   Die Behandlung von Sonne und Mond wurde ersetzt durch die Algorithmen in
#   Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Die Daten aller Planeten beruhen auf VSPO87
#   Pierre Bretagnon, Gerard Francou
#   Bureau des Longitudes
#   77, Avenue Denfert-Rochereau
#   F75014, Paris, France
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/
#
#   Die Resultate wurden mit denen auf https://www.timeanddate.com/ verglichen,
#   die Uebereinstimmung ist gut innerhalb von wenigen Minuten.
#   Beim Mond ist stets das Datum zu beachten!
#   Liegt die lokale Zeit zwischen Mitternacht und "zone", so wird für den Mond
#   zum Teil auf den Vortag zurückgegriffen. Dies ist besonders für Orte mit
#   grosser Zeitzonendifferenz wichtig, wie Sydney oder Honolulu.
#   Für Orte in den Polarregionen wird Mitternachtssonne und Polarnacht
#   korrekt ausgegeben, ausser am Tag, wo die Sonne unter- oder aufgeht.
#   Ebenso fuer den Mond.
#
#   Anmerkung: die Webseiten astronomie.info sowie calsky.com existieren
#   seit Oktober 2020 nicht mehr.
#
#   Alle 8 Planeten sind ebenfalls behandelt, gemaess den Algorithmen in
#   Jean Meeus, Astronomical Algorithms.
#   Die Daten zu den 8 Planeten stammen aus den VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Version 1.0
#   Version 2.0
#   Version 3.0
#   Version 3.1
#   Version 3.4
#   Version 4.0 inkl Sun-Events, Moon-Events
#   Version 5.1 
Version = "V5.3 hvf 2023-09-27 SunHead.py using VSOP87 ephemerides and HVFTime"
if 'VersionList' not in globals().keys(): VersionList = []
VersionList.append(Version)

import time
from datetime import datetime, date, timedelta
from math import sin, acos, cos, pi, radians, degrees, atan, atan2
from math import asin, tan, ceil, floor, sqrt, fabs
import argparse
import sys
import dateutil.parser
import os

# alle Konstanten sind auf die Epoche 2010.0, JD=2455196.5 bezogen
# falls nicht anders dokumentiert.

def getEpoch2010():
    return 2455196.5

# https://www.geeksforgeeks.org/how-to-print-to-stderr-and-stdout-in-python/
def print_to_stderr(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stderr) 

def print_to_stdout(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stdout) 

# produces a datetime object output string without microseconds
# seconds get rounded

def nicedatetime(dt):       # yyyy-mm-dd HH:MM:SS
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        string = dt.getStr()
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%Y-%m-%d %H:%M:%S")
    return string

def niceshortdatetime(dt):  # dd.mm.yy HH:MM:SS
    # only meaningful for known century
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        (yy, mm, dd) = dt.getDate()
        yy = int(abs(yy)%100)    # only last 2 ciphers, also for negative yyyy
        (hh, MM, ss) = dt.getTime()
        # round seconds! use %f fmt - hvf 13.04.2023
        #string = "%02d.%02d.%02d %02d:%02d:%02d" % (dd, mm, yy, hh, MM, ss)
        string = "%02d.%02d.%02d %02d:%02d:%02.0f" % (dd, mm, yy, hh, MM, ss)
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%d.%m.%y %H:%M:%S")
    return string

def nicedateshorttime(dt):  # yyyy-mm-dd HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=30)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        (h, m, s) = dt.getTime()
        string = "%4d-%02d-%02d %02d:%02d" % (y, mm, dd, h, m)
    else:
        string = dt.strftime("%Y-%m-%d %H:%M")
    return string

def niceshortdateshorttime(dt):  # dd.mm.yy HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        dtrounded = roundhvfTime(dt, 60)
        #print ("dt, dtrounded = ", dt, dtrounded)
        (y, mm, dd) = dtrounded.getDate()
        yy = abs(y)%100
        dd = floor(dd)
        (h, m, s) = dtrounded.getTime()
        string = "%02d.%02d.%02d %02d:%02d" % (dd, mm, yy, h, m)
    else:
        dt += timedelta(seconds=30)
        string = dt.strftime("%d.%m.%y %H:%M")
    return string

def nicedate(dt):           # yyyy-mm-dd
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=0.5)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        string = "%04d-%02d-%02d" % (y, mm, dd)
    else:
        string = dt.strftime("%Y-%m-%d")
    return string

def nicetime(dt):           # HH:MM:SS
    if isinstance(dt, type(None)): return None
    try:
        if isinstance(dt, hvfTime):
            (h, m, s) = dt.getTime()
            s =round(s, 0)
            string = "%02d:%02d:%02d" % (h, m, s)
        else:
            dt += timedelta(seconds=0.5)
            string = dt.strftime("%H:%M:%S")
    except:
        string = ' '*8
    return string

def nicetimedelta(dt):      # +-HH:MM:SS
    if isinstance(dt, type(None)): return None
    #dt += timedelta(seconds=0.5)
    # treat also -00:00:05 correctly! hvf 16.06.2021
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60

    string = "%1s%2d:%02d:%02d" % (sign, hours, minutes, seconds)
    return string

def nicedeltadayhhmm(dt):   # +-days HH.MM
    if isinstance(dt, type(None)): return None
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    days  = int(totalsecs/86400.); totalsecs -= days*86400.
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60:
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60
    if hours >= 24:
        days += 1; hours -= 24

    string = "%1s%d days %02d:%02d" % (sign, days, hours, minutes)
    return string

def niceminsec(secs):       # +- M:SS  for Time Equation - Zeitgleichung
    if isinstance(secs, type(None)): return None
    sign = ' '
    if secs < 0: sign = '-'; secs = abs(secs)
    minutes = int(secs/60.); secs -= minutes*60.
    seconds = round(secs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60

    string = "%1s%d:%02d" % (sign, minutes, seconds)
    return string

# printZeile(kopf, Liste, format, zeile=79)
# schreibt eine Zeile der max. Laenge zeile mit kopf und den
# Werten aus Liste, formattiert je mit format.
# Padding vor jedem Listenelement
# Die  formattierten Werte muessen gleich lang sein.
# hvf 31.12.2020

def printZeile(kopf, Liste, format, zeile=79):
    rest = zeile
    string = kopf
    rest -= len(string)
    n = len(Liste)      # Anzahl Werte
    test = format % Liste[0]
    testlen = len(test)
    padlen = int((rest - n*testlen)/n)
    padding = ' '*padlen

    for item in Liste:
        string += padding
        string += format % item

    print (string)
    return string

# modulus in Python has always the same sign as b
# Mod(-3, 24) gives 21
def Mod(a, b):
    return (a % b)

# return integer value, closer to 0
def Int(x): return int(x)

# return fractional part, closer to 0 (i.e. may be nagative)
def Frac(x): return (x - int(x))

# Winkelfunktionen fuer Grad Eingaben
def Sin(x):
    return sin(radians(Mod(x, 360.)))

def Cos(x):
    return cos(radians(Mod(x, 360.)))

def Tan(x):
    return tan(radians(Mod(x, 360.)))

def Deg(x):
    return degrees(x)

# calculate angular distance between two objects with (RA1, Decl1), (RA2, Decl2)
# all angles in degrees
# see Meeus, chapter 16, p. 105, Angular Separation
def AngularDistance(star1, star2, verbose=False):
    RA1 = star1[0]; Decl1 = star1[1]
    RA2 = star2[0]; Decl2 = star2[1]
    cosD = Sin(Decl1)*Sin(Decl2) + Cos(Decl1)*Cos(Decl2)*Cos(RA1-RA2)
    D = degrees(acos(cosD))
    # check if angular separation is small
    if D < 0.2:     # below 0.2 degrees, 12'
        Dmean = (Decl1+Decl2)/2.
        Ddelta = Decl1 - Decl2
        D = sqrt(((RA1-RA2)*Cos(Dmean))**2 + Ddelta**2)
    return D


#####################################################################
# Quelle: https://de.wikipedia.org/wiki/Julianisches_Datum
# abgerufen am 20.10.2019
# Name geaendert in getJulianDay (siehe Meeus, chapter 7) - hvf 24.03.2023

# Calculate Julian day for all Gregorian and Julian dates
# Greenwich date is supplied as a datetime object

def getJulianDay(dtGreenwichDate, isGregor=True ):
    '''
    # day may have fractions for h, m, s, us
    # Jan and Feb are months 13 and 14
    year  = dtGreenwichDate.year
    month = dtGreenwichDate.month
    day   = dtGreenwichDate.day
    day  += dtGreenwichDate.hour/24. + dtGreenwichDate.minute/1440. + \
            dtGreenwichDate.second/86400. + \
            dtGreenwichDate.microsecond/86400e6
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    return JD
    '''
    hvfdt = hvfTime(dtGreenwichDate)
    return hvfdt.getJD()

################################################################
# calculate Greenwich date from Julian date
# Ref. Practical Astronomy (see above), p.11
# returns a datetime object for midnight 

def getGreenwichDate(JD):
    '''
    jd1 = JD + 0.5
    I = int(jd1)        # integer part
    F = jd1 - I         # fractional part
    if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
        A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
        B = I + A - int(A/4) + 1
    else:   B = I       # Julian Calendar for earlier dates
    C = B + 1524
    D = int((C-122.1)/365.25)
    E = int(365.25 * D)
    G = int((C-E)/30.6001)
    d = C - E + F - int(30.6001 * G)
    if G < 13.5:    m = G - 1
    else:           m = G - 13
    if m > 2.5:     y = D - 4716
    else:           y = D - 4715

    # add fractional day - must use timedelta for this
    dtGreenwichDate = datetime(y, m, 1) + timedelta(days=d-1)
    return dtGreenwichDate
    '''

    return hvfTime(float(JD))

#   HMSDH convert hours, minutes, seconds to decimal hours
def HMSDH(hh, mm, ss):  # hh may be negative
    habs = abs(hh)
    habs += mm/60. + ss/3600.
    h = habs
    if hh < 0: h = -habs
    return h

#   DHHMS converts decimal hours into hours, minutes, seconds
#   seconds may have fractions, no rounding
def DHHMS(h):   # h may be negative, then only hours will be negative
    sign = 1
    if h < 0: sign = -1
    habs = abs(h)
    hh = int(habs)
    mm = int(Frac(habs)* 60.)
    ss = Frac(habs*60)*60.
    hh = hh * sign
    return (hh, mm, ss)

#   returns HH:MM and a rounded value of h
#   h must be positive or None or ''
def HHMM(hh):
    if (hh == None or hh == ''):
        return('--')
    m = Frac(hh) * 60.0
    h = Int(hh)
    if (m >= 59.5):
        h += 1
        m -= 60.0
    m = int(round(m))
    string = "%02d:%02d = " % (h, m) + str(round(hh, 4))
    return (string)

#   returns HH:MM:SS and a rounded value of h
#   h must be positive or None or ''
def HHMMSS(hh):
    if (hh == 0):
        return('--')
    m = Frac(hh) * 60
    h = Int(hh)
    s = Frac(m) * 60.0
    m = Int(m)
    if (s >= 59.5):
        m += 1
        s -= 60.0
    if (m >= 60):
        h += 1
        m -= 60
    s = int(round(s))
    string = "%02d:%02d:%02d = " % (h, m, s) + str(round(hh, 5))
    return (string)

def hmsstring(hour):
    # hour may also be negative, e.g. if used to convert declination
    (hh, mm, ss) = DHHMS(hour)
    str = "%d:%02d:%02d" % (hh, mm, ss)
    return str


#   local civic time to Universal Time
#   returns a datetime object 

def getUTfromCivicTime(dtLocalCivicTime, zone):    
    # dtLocalCivicTime is a datetime object
    # zone is time zone offset in hours, positive going East
    dtUT = dtLocalCivicTime - timedelta(hours=zone)
    return dtUT

def getCivicTimefromUT(dtUT, zone):
    dtLocalCivicTime = dtUT + timedelta(hours=zone)
    return dtLocalCivicTime

#   UTC to mean local time 
#   dtUT: Universal Time datetime object
#   longitude: Eastern longitude, degrees
#   returns a datetime object, rounded to seconds

def getMeanLocalTime(dtUT, longitude):
    dtmeanLT = dtUT + timedelta(seconds=round(240.*longitude, 0))
    return dtmeanLT

def getUTfromMeanLocalTime(dtmeanLT, longitude):
    dtUT =  dtmeanLT - timedelta(seconds=round(240.*longitude, 0))
    return dtUT

#   datetime.datetime or datetime.time to decimal hours

def getDecimalHour(dttime):
    '''
    hour = dttime.hour + dttime.minute/60. + dttime.second/3600. \
            + dttime.microsecond/3600.E6
    '''
    (hour, min, sec) = hvfTime(dttime).getTime()
    return hour    

# calculate horizontal coordinates (Azimut, Altitude)
# from ecliptic coordinates (Hour angle, Declination)
#   Parameters:
#   Hour    Hour angle counted from meridian to West, degrees
#   Decl    Declination, degrees
#   latitude    degrees
#   Returns:
#   azimut  counted from North, degrees
#   alt     altitude above horizon, degrees
#   x, y, z Cartesian coords in horizontal system, x pointing South, y West
#   https://de.wikipedia.org/wiki/Astronomische_Koordinatensysteme#Umrechnungen

def getAziAltfromHourDecl(Hour, Decl, latitude):
    # Hour angle and Declination into azimut and altitude
    azimut = atan2(Sin(Hour), Cos(Hour)*Sin(latitude)-Tan(Decl)*Cos(latitude))
    alt    = asin(Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour))
    azimut = Mod(degrees(azimut)+180., 360.)
    alt    = degrees(alt)
    # Cartesian coords in horizon system
    x = Cos(alt)*Cos(azimut-180.)
    y = Cos(alt)*Sin(azimut-180.)
    z = Sin(alt)
    # dito calculated from  Hour, Decl, latitude
    xx = -Cos(latitude)*Sin(Decl)+Sin(latitude)*Cos(Decl)*Cos(Hour)
    yy = Cos(Decl)*Sin(Hour)
    zz = Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour)
    xyzDict = {
        "x":    x,
        "y":    y,
        "z":    z, 
        "xx":   xx,
        "yy":   yy,
        "zz":   zz, }
    return (azimut, alt, xyzDict)

# Zodiac signs for Sun and Moon
#   lon ecliptic longitude, degrees
def getSign(lon):
    Signs_de = ("Widder", "Stier", "Zwillinge", "Krebs", "Löwe", "Jungfrau", \
        "Waage", "Skorpion", "Schütze", "Steinbock", "Wassermann", "Fische")
    Signs_en = ("Aries", "Taurus", "Gemini", "Cancer", "Leo", "Virgo", \
        "Libra", "Scorpio", "Sagittarius", "Capricorn", "Aquarius", "Pisces")
    return (Signs_en[int(floor(lon / 30))])

########################################################################
#
# getNutation(dtCivicTime, zone)
# Nutation in ecliptic longitude and ecliptic latitude
# see Meeus, Astronomical Algorithms, chapter 21

def getNutation(dtCivicTime, zone, verbose=False):
    # Epoch is J2000.0
    # JD 2451545
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    T = (JD - 2451545.0)/36525. # Centuries after the Epoch, 21.1

    # Moon Ascending Node
    Omega = 125.04452 - 1934.136261 * T + 0.0020708 * T*T + T*T*T/450000.
    Omega = Mod(Omega, 360.)
    O = radians(Omega)

    # accuracy of 0.5 arcsec in longitude, 0.1 arcsec in latitude
    # Sun Mean Longitude
    LmeanSun = 280.4665 + 36000.7698 * T
    LmeanSun = Mod(LmeanSun, 360.)
    LS = radians(LmeanSun)

    # Moon Mean Longitude
    LmeanMoon = 218.3165 + 481267.8813 * T
    LmeanMoon = Mod(LmeanMoon, 360.)
    LM = radians(LmeanMoon)

    # Effect on ecliptic longitude and latitude, arcsecs
    deltaELong = -17.20*sin(O) -1.32*sin(2.*LS) -0.23*sin(2*LM) +0.21*sin(2*O)
    deltaELat  = 9.20*cos(O) + 0.57*cos(2.*LS) + 0.10*cos(2*LM) -0.09*cos(2*O)

    # obliquity of the ecliptic, 21.2
    eps0 = 23+26/60.+21.448/3600. - 46.8150/3600.*T - 0.00059/3600.*T*T \
            +0.001813/3600*T*T*T
    eps = eps0 + deltaELat/3600.

    if verbose:
        print ("getNutation")
        print ("JD = ", JD)
        print ("T = ", T)
        print ("LmeanSun = ", LmeanSun)
        print ("LmeanMoon = ", LmeanMoon)
        print ("getNutation deltaELong (arcsec) = %.2f" % deltaELong)
        print ("getNutation deltaELat  (arcsec) = %.2f" % deltaELat )
        print ("eps0 (degrees) = %.6f" % eps0, ", ", HHMMSS(eps0))
        print ("eps  (degrees) = %.6f" % eps, ", ", HHMMSS(eps) )

    # return values in degrees
    return (deltaELong/3600., deltaELat/3600., eps)

#################################################################
#   Funktionen für die Berechnung der Zeitgleichung

def getSunStufffromTT(TT, verbose = False):
    # TT Tage nach Epoch2010 inkl. DeltaT Korrektur
    # returns RA, epsilon in degrees

    def atanLage(x, Lage):    # arctangens mit Lageparameter, alle Werte in Grad
        # siehe https://en.wikipedia.org/wiki/Equation_of_time
        # Resultat in Grad
        xrad = radians(x)
        Lagerad = radians(Lage)
        y = atan(xrad)
        valuerad = y + pi * round((Lagerad-y)/pi, 0)
        value = degrees(valuerad)
        return value
    
    def getE(t):    # numerische Exzentrizität zur Zeit t
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        e = 0.01670491 - 4.208E-5*T - 1.26E-7*T*T
        return e
    
    def getEps(t):  # Schiefe der Ekliptik in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
        return eps
    
    def getOmegabar(t):     # Länge des Perihels in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        omegabar = 283.11243767 + 1.720171*T + 0.000452778*T*T
        return omegabar
    
    def getL(t):    # mittlere Länge der Sonne in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        L = 279.557285 + 36000.769585*T + 0.0003025*T*T
        return L
    
    def getC(e, M):    # Mittelpunktsgleichung equation of the center
        # berechnet wahre minus mittlere Anomalie als Fkt der Exzentrizität
        # und der mittleren Anomalie in dritter Näherung in e
        # e numerische Exzentrizität
        # M mittlere Anomalie der Sonne in Grad
        # Resultat in Grad
        # Danby, Fundamentals of Celestial Mechanics, p.438
        Mrad = radians(M)
        Crad = (2*e-e*e*e/4)*sin(Mrad) + (1.25*e*e -11./24.*e**4)*sin(2*Mrad)\
             + 13./12.*e*e*e*sin(3*Mrad) + 103./96.*e**4*sin(4*Mrad)
        C = degrees(Crad)
        return C

    t = TT
    e = getE(t)                 # Exzentrizität der Erdbahn, Zeit t
    eps = getEps(t)             # Schiefe der Ekliptik in Grad, Zeit t
    epsRad = radians(eps)       # Schiefe der Ekliptik in Radians
    omegabar = getOmegabar(t)   # Länge des Perihels in Grad, Zeit t
    L = getL(t)                 # mittlere Länge der Sonne in Grad, Zeit t
    Lhaupt = Mod(L, 360)        # Hauptwert der Länge der Sonne
    M = Lhaupt - omegabar       # Mittlere Anomalie der Sonne in Grad, Zeit t
    Mhaupt = Mod(M, 360)        # Hauptwert der mittleren Anomalie
    C = getC(e, Mhaupt)         # Mittelpunktsgleichung, wahre minus mittlere A
    Lwahr = L + C               # wahre Länge der Sonne in Grad, Zeit t
    LwahrRad = radians(Lwahr)   # wahre Länge der Sonne in Radians
    Lwahrhaupt = Mod(Lwahr, 360)# Hauptwert der wahren Länge
    LwahrHauptRad = radians(Lwahrhaupt)  # dito, in Radians


    # Rektaszension der Sonne in Funktion der wahren Länge der Sonne und der 
    # Schiefe der Ekliptik
    alpha = atanLage(degrees(tan(LwahrRad)*cos(epsRad)), Lwahr)
    alphahaupt = atanLage(degrees(tan(LwahrHauptRad)*cos(epsRad)), Lwahrhaupt)

    Dict = { "Object"   : "SunStuff",
            "TT"        : t,            # Tage nach Epoch2010 inkl. deltaT
            "MPEq"      : C,            # wahre - mittlere Anomalie
            "omegabar"  : omegabar,     # Länge des perihels in Grad, Zeit t
            "M"         : M,            # mittlere Anomalie, Grad
            "nu"        : C+M,          # wahre Anomalie
            "Long"      : Lwahrhaupt,   # wahre Länge der Sonne, Grad
            "eps"       : eps,          # Schiefe der Ekliptik, Grad
            "exc"       : e,            # Exzentrizität der Erdbahn
            }
    if verbose:
        print ("Numerische Exzentrizitaet e = ", "%.6f" % e)
        print ("Omegabar, Laenge des Perihels = ", "%.3f Grad" % omegabar)
        print ("Lhaupt, Hauptwert Mittlere Laenge = ", "%.3f Grad" % Lhaupt)
        print ("M Mittlere Anomalie = ", "%.3f Grad" % M)
        print ("Mhaupt, Hauptwert Mittlere Anomalie = ", "%.3f Grad" % Mhaupt)
        print ("C Mittelpunktsgleichung = ", "%.3f Grad" % C)
        print ("Alpha Wahre Anomalie = ", "%.3f Grad" % alpha)
    return (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict)

#######################################################################    
def Zeitgleichung(JD, verbose=False):
    # berechnet die Zeitgleichung in ganzen Sekunden
    # JD Julianisches Datum
    # verbose aktiviert zusätzlichen Output
    # Resultat in Sekunden ohne Rundung
    # Epoche konsistent 2010.0, 31.12.2009 0h UT, JD 2455196.5
    dtEpoch2010 = datetime(2009, 12, 31)
    tbase = getJulianDay (dtEpoch2010)
    t = JD - tbase
    TT = JD - tbase + getDeltaT(JD)/86400.

    (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict) = getSunStufffromTT(TT, verbose)

    # Zeitgleichung erste Ursache Exzentrizität der Erdbahn (epsilon = 0)
    ZGstar0 = L - Lwahr         # in Grad
    ZG0 = 240. * ZGstar0        # in Sekunden

    # zweite Ursache Anteil Schiefe der Ekliptik
    ZGstareps = Lwahrhaupt - alphahaupt # in Grad
    ZGeps = 240. * ZGstareps            # in Sekunden

    # Summe der beiden Ursachen
    ZGstar = ZGstar0 + ZGstareps        # in Grad
    ZGsecs = ZG0 + ZGeps                # in Sekunden

    # ZGmin, ZGsec for printing purposes 
    ZGstr = niceminsec(ZGsecs)
    if verbose:
        print ("Epoch2010 = ", dtEpoch2010)
        print ("Epoch Julian Date = ", tbase)
        print ("TT = ", TT)
        print ("Schiefe der Ekliptik eps = ", "%.3f Grad" % eps)
        print ("L, Mittlere Laenge = ", "%.3f Grad" % L)
        print ("Lwahr Wahre Sonnenlaenge = ", "%.3f Grad" % Lwahr)
        print ("Lwahrhaupt Hauptwert Wahre Sonnenlaenge = ", "%.3f Grad" \
                    % Lwahrhaupt)
        print ("Alphahaupt Wahre Anomalie = ", "%.3f Grad" % alphahaupt)
        print ("ZGstar0 = ", "%.3f" % ZGstar0, " Grad")
        print ("ZG0 = ", "%.2f" % ZG0, " Sekunden")
        print ("ZGstareps = ", "%.3f" % ZGstareps, " Grad")
        print ("ZGeps = ", "%.2f" % ZGeps, " Sekunden")
        print ("ZGstar = ", "%.3f" % ZGstar, " Grad")
        print ("ZGsecs = ", "%.2f" % ZGsecs, " Sekunden")
        print ("ZG = %s" % ZGstr)

    return (ZGsecs, ZGstr, Dict)

def getZeitgleichung(JD, verbose=False):
    # returns a timedelta object and two int's for printing
    zgsecs, ZGstr, Dict = Zeitgleichung(JD, verbose)
    return (timedelta(seconds=zgsecs), ZGstr, Dict)

# Greenwich Mean Sidereal Time from Julian Date

def getGMSTfromJD(JD):
    '''
    # new Epoch2010, JD=2455196.5
    # JD usually float or int
    # returns a datetime object
    UT = Frac(JD-0.5) * 24.     # UT in hours
    JD = floor(JD-0.5) + 0.5    # JD at 0 hours UT
    # JD 2455196.5 is Epoch 2010.0
    # T is centuries after Epoch
    T  = (JD - 2455196.5) / 36525.
    T0 = 6.636798592 + T * (2400.051341 + T * 0.000025862)
    ST = Mod(T0 + UT * 1.002737909, 24.)
    Gdate = getGreenwichDate(JD)
    return Gdate + timedelta(hours=ST)
    '''

    # Meeus chapter 11 p. 83
    T = (JD - 2451545.0) / 36525.   # (11.1)
    theta0 = 280.46061837 + 360.98564736629 * (JD - 2451545.0) \
            + 0.000387933 * T*T - T*T*T/387110000   # (11.4)
    # normalize to 360 degrees
    theta = Mod(theta0, 360.)
    # add to midnight
    JD0 = floor(JD-0.5) + 0.5
    Gdate = hvfTime(JD0 + theta/360.)
    return Gdate

# Local Mean Sideral Time from GMST, longitude
# returns a datetime object

def getLMSTfromGMST(gmst, longitude):
    # gmst datetime object
    # longitude in degrees
    # returns a datetime object
    lmst = gmst + timedelta(hours=longitude/15.)
    return lmst

# Greenwich Mean Sideral Time from LMST, longitude
# returns a datetime object

def getGMSTfromLMST(lmst, longitude):
    # lmst datetime object
    # longitude in degrees
    # returns a datetime object
    #gmst = datetime.time(hour=lmst) - timedelta(hours=longitude/15.)
    gmst = lmst - timedelta(hours=longitude/15.)
    return gmst
    

# deltaT, the difference between Terrestrial Time TT and
# Universal Time 
# see https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
# for Julian Date JD

def getDeltaT(JD): 
    # convert JD into year, month, day
    '''
    dtymd = getGreenwichDate(JD)
    year = dtymd.year
    month = dtymd.month
    y = year + (month-0.5)/12.
    # print("getDeltaT year, month, y = %d %d %.2f" % (year, month, y))
    '''
    (year, month, day) = hvfTime(JD).getDate()
    # We define the decimal year "y" as follows:
    y = year + (month - 0.5)/12

    if year < -500:
        u = (year-1820.)/100.
        deltaT = -20 + 32*u*u
    elif year < 500:
        u = y/100.
        deltaT = 10583.6 - 1014.41*u + 33.78311*u*u - 5.952053*u**3 \
                - 0.1798452*u**4 + 0.022174192*u**5 + 0.0090316521*u**6
    elif year < 1600:
        u = (y-1000.)/100.
        deltaT = 1574.2 - 556.01*u + 71.23472*u*u + 0.319781*u**3 \
                - 0.8503463*u**4 - 0.005050998*u**5 + 0.0083572073*u**6
    elif year < 1700:
        t = y - 1600.
        deltaT = 120. -0.9808*t - 0.01532*t*t + t**3/7129
    elif year < 1800:
        t = y - 1700.
        deltaT = 8.83 + 0.1603*t - 0.0059285*t*t + 0.00013336*t**3 \
                - t**4/1174000.
    elif year < 1860:
        t = y - 1800.
        deltaT = 13.72 - 0.332447*t + 0.0068612*t*t + 0.0041116*t**3 \
                - 0.00037436*t**4 + 0.0000121272*t**5 \
                - 0.0000001699*t**6 + 0.000000000875*t**7
    elif year < 1900:
        t = y - 1860.
        deltaT = 7.62 + 0.5737*t - 0.251754*t*t + 0.01680668*t**3 \
                - 0.0004473624*t**4 + t**5/233174.
    elif year < 1920:
        t = y - 1900.
        deltaT = -2.79 + 1.494119*t - 0.0598939*t*t + 0.0061966*t**3 \
                -0.000197*t**4
    elif year < 1941:
        t = y - 1920.
        deltaT = 21.20 + 0.84493*t - 0.076100*t*t + 0.0020936*t**3
    elif year < 1961:
        t = y - 1950.
        deltaT = 29.07 + 0.407*t - t*t/233. + t**3/2547.
    elif year < 1986:
        t = y - 1975.
        deltaT = 45.45 + 1.067*t - t*t/260. - t**3/718.
    elif year < 2005:
        t = y - 2000.
        deltaT = 63.86 + 0.3345*t - 0.060374*t*t + 0.0017275*t**3 \
                + 0.000651814*t**4 + 0.00002373599*t**5
    elif year < 2050:
        t = y - 2000.
        deltaT = 62.92 + 0.32217*t + 0.005589*t*t
    elif year < 2150:
        deltaT = -20. + 32*((y-1820.)/100)**2 - 0.5628*(2150.-y)
    else:   # after 2150
        u = (year-1820.)/100.
        deltaT = -20. + 32.*u*u
        # https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
        # small correction beyond 2150 (only needed for Perihelion calc's)
        c = -0.000012932 * (y - 1955.)**2
        deltaT += c

    return deltaT

def getEps(t):  # Schiefe der Ekliptik in Grad
    # t in Tage nach Epoch2010
    # Resultat in Grad
    T = t / 36525
    eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
    return eps


#######################################################################
# Jean Meeus, Astronomical Algorithms, Willmann-Bell, Richmond VA, 1991
#
# hvf 03.12.2020

# quadratic interpolation
# xi 3 Stuetzstellen, yi 3 Funktionswerte
# Meeus p. 25
def interp(xi, yi, x):
    a10 = (yi[1]-yi[0])/(xi[1]-xi[0])
    a21 = (yi[2]-yi[1])/(xi[2]-xi[1])
    a210 = (a21-a10)/(xi[2]-xi[0])
    res = yi[0] + a10*(x-xi[0]) + a210*(x-xi[0])*(x-xi[1])
    return res

#############################################################################

from datetime import datetime
global EarthdtUTimes
EarthdtUTimes = {}  # empty dict of times
#!/usr/bin/env python
# coding: utf-8
#   getEarthLBRData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Earth"
#       Long  :  heliocentric Longitude, degrees
#       Lat   :  heliocentric Latitude,degrees
#       Rau   :  Radius, distaance Sun - Object (au)
#
#   Similarly for each planet
#   data from Meeus, Appendix II and VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.2 hvf 2023-04-06 getEarthLBRData.py created by mkPlanetLBRData.py"
VersionList.append(Version)

def getEarthLBRData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. (31.1) Julian Millennia

#######################################################################


    #strdtUT = dtUT.isoformat(sep=' ')     # convert to string and use as key to dict
    #strdtUT = str(dtUT)     # usually yyyy-mm-dd HH:MM:SS.sss
    strdtUT = dtUT.getStr()     # usually yyyy-mm-dd HH:MM:SS.sss
    # check if we have been here before
    if strdtUT in EarthdtUTimes:
        Data = EarthdtUTimes[strdtUT]
        # print ("getEarthLBRData using stored data %s" % strdtUT)
        return Data # L, B, R, Dict
    # print ("getEarthLBRData calculate fresh: dtUT = %s" % strdtUT)

    # EARTH Longitude
    sumL0  = 0
    sumL0  +=  175347045.673 * cos(0.00000000000  + 0.00000000000 *T)
    sumL0  +=  3341656.456 * cos(4.66925680417  + 6283.07584999140 *T)
    sumL0  +=   34894.275 * cos(4.62610241759  + 12566.15169998280 *T)
    sumL0  +=    3417.571 * cos(2.82886579606  + 3.52311834900 *T)
    sumL0  +=    3497.056 * cos(2.74411800971  + 5753.38488489680 *T)
    sumL0  +=    3135.896 * cos(3.62767041758  + 77713.77146812050 *T)
    sumL0  +=    2676.218 * cos(4.41808351397  + 7860.41939243920 *T)
    sumL0  +=    2342.687 * cos(6.13516237631  + 3930.20969621960 *T)
    sumL0  +=    1273.166 * cos(2.03709655772  + 529.69096509460 *T)
    sumL0  +=    1324.292 * cos(0.74246356352  + 11506.76976979360 *T)
    sumL0  +=     901.855 * cos(2.04505443513  + 26.29831979980 *T)
    sumL0  +=    1199.167 * cos(1.10962944315  + 1577.34354244780 *T)
    sumL0  +=     857.223 * cos(3.50849156957  + 398.14900340820 *T)
    sumL0  +=     779.786 * cos(1.17882652114  + 5223.69391980220 *T)
    sumL0  +=     990.250 * cos(5.23268129594  + 5884.92684658320 *T)
    sumL0  +=     753.141 * cos(2.53339053818  + 5507.55323866740 *T)
    sumL0  +=     505.264 * cos(4.58292563052  + 18849.22754997420 *T)
    sumL0  +=     492.379 * cos(4.20506639861  + 775.52261132400 *T)
    sumL0  +=     356.655 * cos(2.91954116867  + 0.06731030280 *T)
    sumL0  +=     284.125 * cos(1.89869034186  + 796.29800681640 *T)
    sumL0  +=     242.810 * cos(0.34481140906  + 5486.77784317500 *T)
    sumL0  +=     317.087 * cos(5.84901952218  + 11790.62908865880 *T)
    sumL0  +=     271.039 * cos(0.31488607649  + 10977.07880469900 *T)
    sumL0  +=     206.160 * cos(4.80646606059  + 2544.31441988340 *T)
    sumL0  +=     205.385 * cos(1.86947813692  + 5573.14280143310 *T)
    sumL0  +=     202.261 * cos(2.45767795458  + 6069.77675455340 *T)
    sumL0  +=     126.184 * cos(1.08302630210  + 20.77539549240 *T)
    sumL0  +=     155.516 * cos(0.83306073807  + 213.29909543800 *T)
    sumL0  +=     115.132 * cos(0.64544911683  + 0.98032106820 *T)
    sumL0  +=     102.851 * cos(0.63599846727  + 4694.00295470760 *T)
    sumL0  +=     101.724 * cos(4.26679821365  + 7.11354700080 *T)
    sumL0  +=      99.206 * cos(6.20992940258  + 2146.16541647520 *T)
    sumL0  +=     132.212 * cos(3.41118275555  + 2942.46342329160 *T)
    sumL0  +=      97.607 * cos(0.68101272270  + 155.42039943420 *T)
    sumL0  +=      85.128 * cos(1.29870743025  + 6275.96230299060 *T)
    sumL0  +=      74.651 * cos(1.75508916159  + 5088.62883976680 *T)
    sumL0  +=     101.895 * cos(0.97569221824  + 15720.83878487840 *T)
    sumL0  +=      84.711 * cos(3.67080093025  + 71430.69561812909 *T)
    sumL0  +=      73.547 * cos(4.67926565481  + 801.82093112380 *T)
    sumL0  +=      73.874 * cos(3.50319443167  + 3154.68708489560 *T)
    sumL0  +=      78.756 * cos(3.03698313141  + 12036.46073488820 *T)
    sumL0  +=      79.637 * cos(1.80791330700  + 17260.15465469040 *T)
    sumL0  +=      85.803 * cos(5.98322631256  + 161000.68573767410 *T)
    sumL0  +=      56.963 * cos(2.78430398043  + 6286.59896834040 *T)
    sumL0  +=      61.148 * cos(1.81839811024  + 7084.89678111520 *T)
    sumL0  +=      69.627 * cos(0.83297596966  + 9437.76293488700 *T)
    sumL0  +=      56.116 * cos(4.38694880779  + 14143.49524243060 *T)
    sumL0  +=      62.449 * cos(3.97763880587  + 8827.39026987480 *T)
    sumL0  +=      51.145 * cos(0.28306864501  + 5856.47765911540 *T)
    sumL0  +=      55.577 * cos(3.47006009062  + 6279.55273164240 *T)
    sumL0  +=      41.036 * cos(5.36817351402  + 8429.24126646660 *T)
    sumL0  +=      51.605 * cos(1.33282746983  + 1748.01641306700 *T)
    sumL0  +=      51.992 * cos(0.18914945834  + 12139.55350910680 *T)
    sumL0  +=      49.000 * cos(0.48735065033  + 1194.44701022460 *T)
    sumL0  +=      39.200 * cos(6.16832995016  + 10447.38783960440 *T)
    sumL0  +=      35.566 * cos(1.77597314691  + 6812.76681508600 *T)
    sumL0  +=      36.770 * cos(6.04133859347  + 10213.28554621100 *T)
    sumL0  +=      36.596 * cos(2.56955238628  + 1059.38193018920 *T)
    sumL0  +=      33.291 * cos(0.59309499459  + 17789.84561978500 *T)
    sumL0  +=      35.954 * cos(1.70876111898  + 2352.86615377180 *T)
    sumL0  +=      40.938 * cos(2.39850881707  + 19651.04848109800 *T)
    sumL0  +=      30.047 * cos(2.73975123935  + 1349.86740965880 *T)
    sumL0  +=      30.412 * cos(0.44294464135  + 83996.84731811189 *T)
    sumL0  +=      23.663 * cos(0.48473567763  + 8031.09226305840 *T)
    sumL0  +=      23.574 * cos(2.06527720049  + 3340.61242669980 *T)
    sumL0  +=      21.089 * cos(4.14825464101  + 951.71840625060 *T)
    sumL0  +=      24.738 * cos(0.21484762138  + 3.59042865180 *T)
    sumL0  +=      25.352 * cos(3.16470953405  + 4690.47983635860 *T)
    sumL0  +=      22.820 * cos(5.22197888032  + 4705.73230754360 *T)
    sumL0  +=      21.419 * cos(1.42563735525  + 16730.46368959580 *T)
    sumL0  +=      21.891 * cos(5.55594302562  + 553.56940284240 *T)
    sumL0  +=      17.481 * cos(4.56052900359  + 135.06508003540 *T)
    sumL0  +=      19.925 * cos(5.22208471269  + 12168.00269657460 *T)
    sumL0  +=      19.860 * cos(5.77470167653  + 6309.37416979120 *T)
    sumL0  +=      20.300 * cos(0.37133792946  + 283.85931886520 *T)
    sumL0  +=      14.421 * cos(4.19315332546  + 242.72860397400 *T)
    sumL0  +=      16.225 * cos(5.98837722564  + 11769.85369316640 *T)
    sumL0  +=      15.077 * cos(4.19567181073  + 6256.77753019160 *T)
    sumL0  +=      19.124 * cos(3.82219996949  + 23581.25817731760 *T)
    sumL0  +=      18.888 * cos(5.38626880969  + 149854.40013480789 *T)
    sumL0  +=      14.346 * cos(3.72355084422  + 38.02767263580 *T)
    sumL0  +=      17.898 * cos(2.21490735647  + 13367.97263110660 *T)
    sumL0  +=      12.054 * cos(2.62229588349  + 955.59974160860 *T)
    sumL0  +=      11.287 * cos(0.17739328092  + 4164.31198961300 *T)
    sumL0  +=      13.971 * cos(4.40138139996  + 6681.22485339960 *T)
    sumL0  +=      13.621 * cos(1.88934471407  + 7632.94325965020 *T)
    sumL0  +=      12.503 * cos(1.13052412208  + 5.52292430740 *T)
    sumL0  +=      10.498 * cos(5.35909518669  + 1592.59601363280 *T)
    sumL0  +=       9.803 * cos(0.99947478995  + 11371.70468975820 *T)
    sumL0  +=       9.220 * cos(4.57138609781  + 4292.33083295040 *T)
    sumL0  +=      10.327 * cos(6.19982566125  + 6438.49624942560 *T)
    sumL0  +=      12.003 * cos(1.00351456700  + 632.78373931320 *T)
    sumL0  +=      10.827 * cos(0.32734520222  + 103.09277421860 *T)
    sumL0  +=       8.356 * cos(4.53902685948  + 25132.30339996560 *T)
    sumL0  +=      10.005 * cos(6.02914963280  + 5746.27133789600 *T)
    sumL0  +=       8.409 * cos(3.29946744189  + 7234.79425624200 *T)
    sumL0  +=       8.006 * cos(5.82145271907  + 28.44918746780 *T)
    sumL0  +=      10.523 * cos(0.93871805506  + 11926.25441366880 *T)
    sumL0  +=       7.686 * cos(3.12142363172  + 7238.67559160000 *T)
    sumL0  +=       9.378 * cos(2.62414241032  + 5760.49843189760 *T)
    sumL0  +=       8.127 * cos(6.11228001785  + 4732.03062734340 *T)
    sumL0  +=       9.232 * cos(0.48343968736  + 522.57741809380 *T)
    sumL0  +=       9.802 * cos(5.24413991147  + 27511.46787353720 *T)
    sumL0  +=       7.871 * cos(0.99590177926  + 5643.17856367740 *T)
    sumL0  +=       8.123 * cos(6.27053013650  + 426.59819087600 *T)
    sumL0  +=       9.048 * cos(5.33686335897  + 6386.16862421000 *T)
    sumL0  +=       8.620 * cos(4.16538210888  + 7058.59846131540 *T)
    sumL0  +=       6.297 * cos(4.71724819317  + 6836.64525283380 *T)
    sumL0  +=       7.575 * cos(3.97382858911  + 11499.65622279280 *T)
    sumL0  +=       7.756 * cos(2.95729056763  + 23013.53953958720 *T)
    sumL0  +=       7.314 * cos(0.60652505806  + 11513.88331679440 *T)
    sumL0  +=       5.955 * cos(2.87641047971  + 6283.14316029419 *T)
    sumL0  +=       6.534 * cos(5.79072926033  + 18073.70493865020 *T)
    sumL0  +=       7.188 * cos(3.99831508699  + 74.78159856730 *T)
    sumL0  +=       7.346 * cos(4.38582365437  + 316.39186965660 *T)
    sumL0  +=       5.413 * cos(5.39199024641  + 419.48464387520 *T)
    sumL0  +=       5.127 * cos(2.36062848786  + 10973.55568635000 *T)
    sumL0  +=       7.056 * cos(0.32258441903  + 263.08392337280 *T)
    sumL0  +=       6.625 * cos(3.66475158672  + 17298.18232732620 *T)
    sumL0  +=       6.762 * cos(5.91132535899  + 90955.55169449610 *T)
    sumL0  +=       4.938 * cos(5.73672165674  + 9917.69687450980 *T)
    sumL0  +=       5.547 * cos(2.45152597661  + 12352.85260454480 *T)
    sumL0  +=       5.958 * cos(3.32051344676  + 6283.00853968860 *T)
    sumL0  +=       4.471 * cos(2.06385999536  + 7079.37385680780 *T)
    sumL0  +=       6.153 * cos(1.45823331144  + 233141.31440436147 *T)
    sumL0  +=       4.348 * cos(4.42342175480  + 5216.58037280140 *T)
    sumL0  +=       6.123 * cos(1.07494905258  + 19804.82729158280 *T)
    sumL0  +=       4.488 * cos(3.65285037150  + 206.18554843720 *T)
    sumL0  +=       4.020 * cos(0.83995823171  + 20.35531939880 *T)
    sumL0  +=       5.188 * cos(4.06503864016  + 6208.29425142410 *T)
    sumL0  +=       5.307 * cos(0.38217636096  + 31441.67756975680 *T)
    sumL0  +=       3.785 * cos(2.34369213733  + 3.88133535800 *T)
    sumL0  +=       4.497 * cos(3.27230796845  + 11015.10647733480 *T)
    sumL0  +=       4.132 * cos(0.92128915753  + 3738.76143010800 *T)
    sumL0  +=       3.521 * cos(5.97844807108  + 3894.18182954220 *T)
    sumL0  +=       4.215 * cos(1.90601120623  + 245.83164622940 *T)
    sumL0  +=       3.701 * cos(5.03069397926  + 536.80451209540 *T)
    sumL0  +=       3.865 * cos(1.82634360607  + 11856.21865142450 *T)
    sumL0  +=       3.652 * cos(1.01838584934  + 16200.77272450120 *T)
    sumL0  +=       3.390 * cos(0.97785123922  + 8635.94200376320 *T)
    sumL0  +=       3.737 * cos(2.95380107829  + 3128.38876509580 *T)
    sumL0  +=       3.507 * cos(3.71291946325  + 6290.18939699220 *T)
    sumL0  +=       3.086 * cos(3.64646921512  + 10.63666534980 *T)
    sumL0  +=       3.397 * cos(1.10590684017  + 14712.31711645800 *T)
    sumL0  +=       3.334 * cos(0.83684924911  + 6496.37494542940 *T)
    sumL0  +=       2.805 * cos(2.58504514144  + 14314.16811304980 *T)
    sumL0  +=       3.650 * cos(1.08344142571  + 88860.05707098669 *T)
    sumL0  +=       3.388 * cos(3.20185096055  + 5120.60114558360 *T)
    sumL0  +=       3.252 * cos(3.47859752062  + 6133.51265285680 *T)
    sumL0  +=       2.553 * cos(3.94869034189  + 1990.74501704100 *T)
    sumL0  +=       3.520 * cos(2.05559692878  + 244287.60000722768 *T)
    sumL0  +=       2.565 * cos(1.56071784900  + 23543.23050468179 *T)
    sumL0  +=       2.621 * cos(3.85639359951  + 266.60704172180 *T)
    sumL0  +=       2.955 * cos(3.39692949667  + 9225.53927328300 *T)
    sumL0  +=       2.876 * cos(6.02635617464  + 154717.60988768269 *T)
    sumL0  +=       2.395 * cos(1.16131956403  + 10984.19235169980 *T)
    sumL0  +=       3.161 * cos(1.32798718453  + 10873.98603048040 *T)
    sumL0  +=       3.163 * cos(5.08946464629  + 21228.39202354580 *T)
    sumL0  +=       2.361 * cos(4.27212906992  + 6040.34724601740 *T)
    sumL0  +=       3.030 * cos(1.80209931347  + 35371.88726597640 *T)
    sumL0  +=       2.343 * cos(3.57689860500  + 10969.96525769820 *T)
    sumL0  +=       2.618 * cos(2.57870156528  + 22483.84857449259 *T)
    sumL0  +=       2.113 * cos(3.71393780256  + 65147.61976813770 *T)
    sumL0  +=       2.019 * cos(0.81393923319  + 170.67287061920 *T)
    sumL0  +=       2.003 * cos(0.38091017375  + 6172.86952877200 *T)
    sumL0  +=       2.506 * cos(3.74379142438  + 10575.40668294180 *T)
    sumL0  +=       2.381 * cos(0.10581361289  + 7.04623669800 *T)
    sumL0  +=       1.949 * cos(4.86892513469  + 36.02786667740 *T)
    sumL0  +=       2.074 * cos(4.22794774570  + 5650.29211067820 *T)
    sumL0  +=       1.924 * cos(5.59460549860  + 6282.09552892320 *T)
    sumL0  +=       1.949 * cos(1.07002512703  + 5230.80746680300 *T)
    sumL0  +=       1.988 * cos(5.19736046771  + 6262.30045449900 *T)
    sumL0  +=       1.887 * cos(3.74365662683  + 23.87843774780 *T)
    sumL0  +=       1.787 * cos(1.25929682929  + 12559.03815298200 *T)
    sumL0  +=       1.883 * cos(1.90364058477  + 15.25247118500 *T)
    sumL0  +=       1.816 * cos(3.68083868442  + 15110.46611986620 *T)
    sumL0  +=       1.701 * cos(4.41105895380  + 110.20632121940 *T)
    sumL0  +=       1.990 * cos(3.93295788548  + 6206.80977871580 *T)
    sumL0  +=       2.103 * cos(0.75354917468  + 13521.75144159140 *T)
    sumL0  +=       1.774 * cos(0.48747535361  + 1551.04522264800 *T)
    sumL0  +=       1.882 * cos(0.86684493432  + 22003.91463486980 *T)
    sumL0  +=       1.924 * cos(1.22898324132  + 709.93304855830 *T)
    sumL0  +=       2.009 * cos(4.62850921980  + 6037.24420376200 *T)
    sumL0  +=       1.924 * cos(0.60231842508  + 6284.05617105960 *T)
    sumL0  +=       1.596 * cos(3.98332956992  + 13916.01910964160 *T)
    sumL0  +=       1.664 * cos(4.41939715469  + 8662.24032356300 *T)
    sumL0  +=       1.971 * cos(1.04560500503  + 18209.33026366019 *T)
    sumL0  +=       1.942 * cos(4.31335979989  + 6244.94281435360 *T)
    sumL0  +=       1.476 * cos(0.93271367331  + 2379.16447357160 *T)
    sumL0  +=       1.810 * cos(0.49112137707  + 1.48447270830 *T)
    sumL0  +=       1.346 * cos(1.51574702235  + 4136.91043351620 *T)
    sumL0  +=       1.528 * cos(5.61835711404  + 6127.65545055720 *T)
    sumL0  +=       1.791 * cos(3.22187270126  + 39302.09696219600 *T)
    sumL0  +=       1.747 * cos(3.05638656738  + 18319.53658487960 *T)
    sumL0  +=       1.431 * cos(4.51153808594  + 20426.57109242200 *T)
    sumL0  +=       1.695 * cos(0.22047718414  + 25158.60171976540 *T)
    sumL0  +=       1.242 * cos(4.46665769933  + 17256.63153634140 *T)
    sumL0  +=       1.463 * cos(4.69242679213  + 14945.31617355440 *T)
    sumL0  +=       1.205 * cos(1.86912144659  + 4590.91018048900 *T)
    sumL0  +=       1.192 * cos(2.74227166898  + 12569.67481833180 *T)
    sumL0  +=       1.222 * cos(5.18120087482  + 5333.90024102160 *T)
    sumL0  +=       1.390 * cos(5.42894648983  + 143571.32428481648 *T)
    sumL0  +=       1.473 * cos(1.70479245805  + 11712.95531823080 *T)
    sumL0  +=       1.362 * cos(2.61069503292  + 6062.66320755260 *T)
    sumL0  +=       1.148 * cos(6.03001800540  + 3634.62102451840 *T)
    sumL0  +=       1.198 * cos(5.15294130422  + 10177.25767953360 *T)
    sumL0  +=       1.266 * cos(0.11421493643  + 18422.62935909819 *T)
    sumL0  +=       1.411 * cos(1.09908857534  + 3496.03282613400 *T)
    sumL0  +=       1.349 * cos(2.99805109633  + 17654.78053974960 *T)
    sumL0  +=       1.253 * cos(2.79850152848  + 167283.76158766547 *T)
    sumL0  +=       1.311 * cos(1.60942984879  + 5481.25491886760 *T)
    sumL0  +=       1.079 * cos(6.20304501787  + 3.28635741780 *T)
    sumL0  +=       1.181 * cos(1.20653776978  + 131.54196168640 *T)
    sumL0  +=       1.254 * cos(5.45103277798  + 6076.89030155420 *T)
    sumL0  +=       1.035 * cos(2.32142722747  + 7342.45778018060 *T)
    sumL0  +=       1.117 * cos(0.38838354256  + 949.17560896980 *T)
    sumL0  +=       0.966 * cos(3.18341890851  + 11087.28512591840 *T)
    sumL0  +=       1.171 * cos(3.39635049962  + 12562.62858163380 *T)
    sumL0  +=       1.121 * cos(0.72627490378  + 220.41264243880 *T)
    sumL0  +=       1.024 * cos(2.19378315386  + 11403.67699557500 *T)
    sumL0  +=       0.888 * cos(3.91173199285  + 4686.88940770680 *T)
    sumL0  +=       0.910 * cos(1.98802695087  + 735.87651353180 *T)
    sumL0  +=       0.830 * cos(0.48984915507  + 24072.92146977640 *T)
    sumL0  +=       1.096 * cos(6.17377835617  + 5436.99301524020 *T)
    sumL0  +=       0.908 * cos(0.44959639433  + 7477.52286021600 *T)
    sumL0  +=       0.974 * cos(1.52996238356  + 9623.68827669120 *T)
    sumL0  +=       0.840 * cos(1.79543266333  + 5429.87946823940 *T)
    sumL0  +=       0.778 * cos(6.17699177946  + 38.13303563780 *T)
    sumL0  +=       0.776 * cos(4.09855402433  + 14.22709400160 *T)
    sumL0  +=       1.068 * cos(4.64200173735  + 43232.30665841560 *T)
    sumL0  +=       0.954 * cos(1.49988435748  + 1162.47470440780 *T)
    sumL0  +=       0.907 * cos(0.86986870809  + 10344.29506538580 *T)
    sumL0  +=       0.931 * cos(4.06044689031  + 28766.92442448400 *T)
    sumL0  +=       0.739 * cos(5.04368197372  + 639.89728631400 *T)
    sumL0  +=       0.937 * cos(3.46884698960  + 1589.07289528380 *T)
    sumL0  +=       0.763 * cos(5.86304932998  + 16858.48253293320 *T)
    sumL0  +=       0.953 * cos(4.20801492835  + 11190.37790013700 *T)
    sumL0  +=       0.708 * cos(1.72899988940  + 13095.84266507740 *T)
    sumL0  +=       0.969 * cos(1.64439522215  + 29088.81141598500 *T)
    sumL0  +=       0.717 * cos(0.16688678895  + 11.72935283600 *T)
    sumL0  +=       0.962 * cos(3.53092337542  + 12416.58850284820 *T)
    sumL0  +=       0.747 * cos(5.77866940346  + 12592.45001978260 *T)
    sumL0  +=       0.672 * cos(1.91095796194  + 3.93215326310 *T)
    sumL0  +=       0.671 * cos(5.46240843677  + 18052.92954315780 *T)
    sumL0  +=       0.675 * cos(6.28311558823  + 4535.05943692440 *T)
    sumL0  +=       0.684 * cos(0.39975012080  + 5849.36411211460 *T)
    sumL0  +=       0.799 * cos(0.29851185294  + 12132.43996210600 *T)
    sumL0  +=       0.758 * cos(0.96370823331  + 1052.26838318840 *T)
    sumL0  +=       0.782 * cos(5.33878339919  + 13517.87010623340 *T)
    sumL0  +=       0.730 * cos(1.70106160291  + 17267.26820169119 *T)
    sumL0  +=       0.749 * cos(2.59599901875  + 11609.86254401220 *T)
    sumL0  +=       0.734 * cos(2.78417782952  + 640.87760738220 *T)
    sumL0  +=       0.688 * cos(5.15048287468  + 16496.36139620240 *T)
    sumL0  +=       0.770 * cos(1.62469589333  + 4701.11650170840 *T)
    sumL0  +=       0.633 * cos(2.20587893893  + 25934.12433108940 *T)
    sumL0  +=       0.760 * cos(4.21317219403  + 377.37360791580 *T)
    sumL0  +=       0.584 * cos(2.13420121623  + 10557.59416082380 *T)
    sumL0  +=       0.574 * cos(0.24250054587  + 9779.10867612540 *T)
    sumL0  +=       0.573 * cos(3.16435264609  + 533.21408344360 *T)
    sumL0  +=       0.685 * cos(3.19344289472  + 12146.66705610760 *T)
    sumL0  +=       0.675 * cos(0.96179233959  + 10454.50138660520 *T)
    sumL0  +=       0.648 * cos(1.46327342555  + 6268.84875598980 *T)
    sumL0  +=       0.589 * cos(2.50543543638  + 3097.88382272579 *T)
    sumL0  +=       0.551 * cos(5.28099026956  + 9388.00590941520 *T)
    sumL0  +=       0.696 * cos(3.65342150016  + 4804.20927592700 *T)
    sumL0  +=       0.669 * cos(2.51030077026  + 2388.89402044920 *T)
    sumL0  +=       0.550 * cos(0.06883864342  + 20199.09495963300 *T)
    sumL0  +=       0.629 * cos(4.13350995675  + 45892.73043315699 *T)
    sumL0  +=       0.678 * cos(6.09190163533  + 135.62532501000 *T)
    sumL0  +=       0.593 * cos(1.50136257618  + 226858.23855437007 *T)
    sumL0  +=       0.542 * cos(3.58573645173  + 6148.01076995600 *T)
    sumL0  +=       0.682 * cos(5.02203067788  + 17253.04110768959 *T)
    sumL0  +=       0.565 * cos(4.29309238610  + 11933.36796066960 *T)
    sumL0  +=       0.486 * cos(0.77746204893  + 27.40155609680 *T)
    sumL0  +=       0.503 * cos(0.58963565969  + 15671.08175940660 *T)
    sumL0  +=       0.616 * cos(4.06539884128  + 227.47613278900 *T)
    sumL0  +=       0.583 * cos(6.12695541996  + 18875.52586977400 *T)
    sumL0  +=       0.537 * cos(2.15056440980  + 21954.15760939799 *T)
    sumL0  +=       0.669 * cos(6.06986269566  + 47162.51635463520 *T)
    sumL0  +=       0.475 * cos(0.40343842110  + 6915.85958930460 *T)
    sumL0  +=       0.540 * cos(2.83444222174  + 5326.78669402080 *T)
    sumL0  +=       0.530 * cos(5.26359885263  + 10988.80815753500 *T)
    sumL0  +=       0.582 * cos(3.24533095664  + 153.77881048480 *T)
    sumL0  +=       0.641 * cos(3.24711791371  + 2107.03450754240 *T)
    sumL0  +=       0.621 * cos(3.09698523779  + 33019.02111220460 *T)
    sumL0  +=       0.466 * cos(3.14982372198  + 10440.27429260360 *T)
    sumL0  +=       0.466 * cos(0.90708835657  + 5966.68398033480 *T)
    sumL0  +=       0.528 * cos(0.81926454470  + 813.55028395980 *T)
    sumL0  +=       0.603 * cos(3.81378921927  + 316428.22867391503 *T)
    sumL0  +=       0.559 * cos(1.81894804124  + 17996.03116822220 *T)
    sumL0  +=       0.437 * cos(2.28625594435  + 6303.85124548380 *T)
    sumL0  +=       0.518 * cos(4.86069178322  + 20597.24396304120 *T)
    sumL0  +=       0.424 * cos(6.23520018693  + 6489.26139842860 *T)
    sumL0  +=       0.518 * cos(6.17617826756  + 0.24381748350 *T)
    sumL0  +=       0.404 * cos(5.72804304258  + 5642.19824260920 *T)
    sumL0  +=       0.458 * cos(1.34117773915  + 6287.00800325450 *T)
    sumL0  +=       0.548 * cos(5.68454458320  + 155427.54293624099 *T)
    sumL0  +=       0.547 * cos(1.03391472061  + 3646.35037735440 *T)
    sumL0  +=       0.428 * cos(4.69800981138  + 846.08283475120 *T)
    sumL0  +=       0.413 * cos(6.02520699406  + 6279.48542133960 *T)
    sumL0  +=       0.534 * cos(3.03030638223  + 66567.48586525429 *T)
    sumL0  +=       0.383 * cos(1.49056949125  + 19800.94595622480 *T)
    sumL0  +=       0.410 * cos(5.28319622279  + 18451.07854656599 *T)
    sumL0  +=       0.352 * cos(4.68891600359  + 4907.30205014560 *T)
    sumL0  +=       0.480 * cos(5.36572651091  + 348.92442044800 *T)
    sumL0  +=       0.344 * cos(5.89157452896  + 6546.15977336420 *T)
    sumL0  +=       0.340 * cos(0.37557426440  + 13119.72110282519 *T)
    sumL0  +=       0.434 * cos(4.98417785901  + 6702.56049386660 *T)
    sumL0  +=       0.332 * cos(2.68902519126  + 29296.61538957860 *T)
    sumL0  +=       0.448 * cos(2.16478480251  + 5905.70224207560 *T)
    sumL0  +=       0.344 * cos(2.06546633735  + 49.75702547180 *T)
    sumL0  +=       0.315 * cos(1.24023811803  + 4061.21921539440 *T)
    sumL0  +=       0.324 * cos(2.30897526929  + 5017.50837136500 *T)
    sumL0  +=       0.413 * cos(0.17171692962  + 6286.66627864320 *T)
    sumL0  +=       0.431 * cos(3.86601101393  + 12489.88562870720 *T)
    sumL0  +=       0.349 * cos(4.55372342974  + 4933.20844033260 *T)
    sumL0  +=       0.323 * cos(0.41971136084  + 10770.89325626180 *T)
    sumL0  +=       0.341 * cos(2.68612860807  + 11.04570026390 *T)
    sumL0  +=       0.316 * cos(3.52936906658  + 17782.73207278420 *T)
    sumL0  +=       0.315 * cos(5.63357264999  + 568.82187402740 *T)
    sumL0  +=       0.340 * cos(3.83571212349  + 10660.68693504240 *T)
    sumL0  +=       0.297 * cos(0.62691416712  + 20995.39296644940 *T)
    sumL0  +=       0.405 * cos(1.00085779471  + 16460.33352952499 *T)
    sumL0  +=       0.414 * cos(1.21998752076  + 51092.72605085480 *T)
    sumL0  +=       0.336 * cos(4.71465945226  + 6179.98307577280 *T)
    sumL0  +=       0.361 * cos(3.71227508354  + 28237.23345938940 *T)
    sumL0  +=       0.385 * cos(6.21925225757  + 24356.78078864160 *T)
    sumL0  +=       0.327 * cos(1.05606504715  + 11919.14086666800 *T)
    sumL0  +=       0.327 * cos(6.14222420989  + 6254.62666252360 *T)
    sumL0  +=       0.268 * cos(2.47224339737  + 664.75604513000 *T)
    sumL0  +=       0.269 * cos(1.86207884109  + 23141.55838292460 *T)
    sumL0  +=       0.345 * cos(0.93461290184  + 6058.73105428950 *T)
    sumL0  +=       0.296 * cos(4.51687557180  + 6418.14093002680 *T)
    sumL0  +=       0.353 * cos(4.50033653082  + 36949.23080842420 *T)
    sumL0  +=       0.260 * cos(4.04963546305  + 6525.80445396540 *T)
    sumL0  +=       0.298 * cos(2.20046722622  + 156137.47598479927 *T)
    sumL0  +=       0.253 * cos(3.49900838384  + 29864.33402730900 *T)
    sumL0  +=       0.254 * cos(2.44901693835  + 5331.35744374080 *T)
    sumL0  +=       0.296 * cos(0.84347588787  + 5729.50644714900 *T)
    sumL0  +=       0.298 * cos(1.29194706125  + 22805.73556599360 *T)
    sumL0  +=       0.241 * cos(2.00721280805  + 16737.57723659660 *T)
    sumL0  +=       0.311 * cos(1.23668016334  + 6281.59137728310 *T)
    sumL0  +=       0.240 * cos(2.51650377121  + 6245.04817735560 *T)
    sumL0  +=       0.332 * cos(3.55576945724  + 7668.63742494250 *T)
    sumL0  +=       0.264 * cos(4.44052061202  + 12964.30070339100 *T)
    sumL0  +=       0.257 * cos(1.79654471948  + 11080.17157891760 *T)
    sumL0  +=       0.260 * cos(3.33077598420  + 5888.44996493220 *T)
    sumL0  +=       0.285 * cos(0.30886361430  + 11823.16163945020 *T)
    sumL0  +=       0.290 * cos(5.70141882483  + 77.67377042800 *T)
    sumL0  +=       0.255 * cos(4.00939664440  + 5881.40372823420 *T)
    sumL0  +=       0.253 * cos(4.73318493678  + 16723.35014259500 *T)
    sumL0  +=       0.228 * cos(0.95333661324  + 5540.08578945880 *T)
    sumL0  +=       0.319 * cos(1.38633229189  + 163096.18036118348 *T)
    sumL0  +=       0.224 * cos(1.65156322696  + 10027.90319572920 *T)
    sumL0  +=       0.226 * cos(0.34106460604  + 17796.95916678580 *T)
    sumL0  +=       0.236 * cos(4.19817431922  + 19.66976089979 *T)
    sumL0  +=       0.280 * cos(4.14080268970  + 12539.85338018300 *T)
    sumL0  +=       0.275 * cos(5.50306930248  + 32.53255079140 *T)
    sumL0  +=       0.223 * cos(5.23334210294  + 56.89837493560 *T)
    sumL0  +=       0.217 * cos(6.08587881787  + 6805.65326808520 *T)
    sumL0  +=       0.280 * cos(4.52472044653  + 6016.46880826960 *T)
    sumL0  +=       0.227 * cos(5.06509843737  + 6277.55292568400 *T)
    sumL0  +=       0.226 * cos(5.17755154305  + 11720.06886523160 *T)
    sumL0  +=       0.245 * cos(3.96486270306  + 22.77520145080 *T)
    sumL0  +=       0.220 * cos(4.72078081970  + 6.62855890001 *T)
    sumL0  +=       0.207 * cos(5.71701403951  + 41.55079098480 *T)
    sumL0  +=       0.204 * cos(3.91227411250  + 2699.73481931760 *T)
    sumL0  +=       0.209 * cos(0.86881969011  + 6321.10352262720 *T)
    sumL0  +=       0.200 * cos(2.11984445273  + 4274.51831083240 *T)
    sumL0  +=       0.200 * cos(5.39839888163  + 6019.99192661860 *T)
    sumL0  +=       0.209 * cos(5.67606291663  + 11293.47067435560 *T)
    sumL0  +=       0.252 * cos(1.64965729351  + 9380.95967271720 *T)
    sumL0  +=       0.275 * cos(5.04826903506  + 73.29712585900 *T)
    sumL0  +=       0.208 * cos(1.88207277133  + 11300.58422135640 *T)
    sumL0  +=       0.272 * cos(0.74640926842  + 1975.49254585600 *T)
    sumL0  +=       0.199 * cos(3.30836672397  + 22743.40937951640 *T)
    sumL0  +=       0.269 * cos(4.48560812155  + 64471.99124174489 *T)
    sumL0  +=       0.192 * cos(2.17464236325  + 5863.59120611620 *T)
    sumL0  +=       0.228 * cos(5.85373115869  + 128.01884333740 *T)
    sumL0  +=       0.261 * cos(2.64321183295  + 55022.93574707440 *T)
    sumL0  +=       0.220 * cos(5.75012110079  + 29.42950853600 *T)
    sumL0  +=       0.187 * cos(4.03230554718  + 467.96499035440 *T)
    sumL0  +=       0.200 * cos(5.60556112058  + 1066.49547719000 *T)
    sumL0  +=       0.231 * cos(1.09802712785  + 12341.80690428090 *T)
    sumL0  +=       0.199 * cos(0.29500625200  + 149.56319713460 *T)
    sumL0  +=       0.249 * cos(5.10473210814  + 7875.67186362420 *T)
    sumL0  +=       0.208 * cos(0.93013835019  + 14919.01785375460 *T)
    sumL0  +=       0.179 * cos(0.87104393079  + 12721.57209941700 *T)
    sumL0  +=       0.203 * cos(1.56920753653  + 28286.99048486120 *T)
    sumL0  +=       0.179 * cos(2.47036386443  + 16062.18452611680 *T)
    sumL0  +=       0.198 * cos(3.54061588502  + 30.91412563500 *T)
    sumL0  +=       0.171 * cos(3.45356518113  + 5327.47610838280 *T)
    sumL0  +=       0.183 * cos(0.72325421604  + 6272.03014972750 *T)
    sumL0  +=       0.216 * cos(2.97174580686  + 19402.79695281660 *T)
    sumL0  +=       0.168 * cos(2.51550550242  + 23937.85638974100 *T)
    sumL0  +=       0.195 * cos(0.09045393425  + 156.40072050240 *T)
    sumL0  +=       0.179 * cos(4.49471798090  + 31415.37924995700 *T)
    sumL0  +=       0.216 * cos(0.42177594328  + 23539.70738633280 *T)
    sumL0  +=       0.189 * cos(0.37542530191  + 9814.60410029120 *T)
    sumL0  +=       0.218 * cos(2.36835880025  + 16627.37091537720 *T)
    sumL0  +=       0.166 * cos(4.23182968446  + 16840.67001081519 *T)
    sumL0  +=       0.200 * cos(2.02153258098  + 16097.67995028260 *T)
    sumL0  +=       0.169 * cos(0.91318727000  + 95.97922721780 *T)
    sumL0  +=       0.211 * cos(5.73370637657  + 151.89728108520 *T)
    sumL0  +=       0.204 * cos(0.42643085174  + 515.46387109300 *T)
    sumL0  +=       0.212 * cos(3.00233538977  + 12043.57428188900 *T)
    sumL0  +=       0.192 * cos(5.46153589821  + 6379.05507720920 *T)
    sumL0  +=       0.165 * cos(1.38698167064  + 4171.42553661380 *T)
    sumL0  +=       0.160 * cos(6.23798383332  + 202.25339517410 *T)
    sumL0  +=       0.215 * cos(0.20889073407  + 5621.84292321040 *T)
    sumL0  +=       0.181 * cos(4.12439203622  + 13341.67431130680 *T)
    sumL0  +=       0.153 * cos(1.24460848836  + 29826.30635467320 *T)
    sumL0  +=       0.150 * cos(3.12999753018  + 799.82112516540 *T)
    sumL0  +=       0.175 * cos(4.55671604437  + 239424.39025435288 *T)
    sumL0  +=       0.192 * cos(1.33928820063  + 394.62588505920 *T)
    sumL0  +=       0.149 * cos(2.65697593276  + 21.33564046700 *T)
    sumL0  +=       0.146 * cos(5.58021191726  + 412.37109687440 *T)
    sumL0  +=       0.156 * cos(3.75650175503  + 12323.42309600880 *T)
    sumL0  +=       0.143 * cos(3.75708566606  + 58864.54391814630 *T)
    sumL0  +=       0.143 * cos(3.28248547724  + 29.82143814880 *T)
    sumL0  +=       0.144 * cos(1.07862546598  + 1265.56747862640 *T)
    sumL0  +=       0.148 * cos(0.23389236655  + 10021.83728009940 *T)
    sumL0  +=       0.193 * cos(5.92751083086  + 40879.44050464380 *T)
    sumL0  +=       0.140 * cos(4.97612440269  + 158.94351778320 *T)
    sumL0  +=       0.148 * cos(2.61640453469  + 17157.06188047180 *T)
    sumL0  +=       0.141 * cos(3.66871308723  + 26084.02180621620 *T)
    sumL0  +=       0.147 * cos(5.09968173403  + 661.23292678100 *T)
    sumL0  +=       0.146 * cos(4.96885605695  + 57375.80190084620 *T)
    sumL0  +=       0.142 * cos(0.78678347839  + 12779.45079542080 *T)
    sumL0  +=       0.134 * cos(4.79432636012  + 111.18664228760 *T)
    sumL0  +=       0.140 * cos(1.27748013377  + 107.66352393860 *T)
    sumL0  +=       0.169 * cos(2.74893543762  + 26735.94526221320 *T)
    sumL0  +=       0.165 * cos(3.95288000638  + 6357.85744855870 *T)
    sumL0  +=       0.183 * cos(5.43418358741  + 369.69981594040 *T)
    sumL0  +=       0.134 * cos(3.09132862833  + 17.81252211800 *T)
    sumL0  +=       0.132 * cos(3.05633896779  + 22490.96212149340 *T)
    sumL0  +=       0.134 * cos(4.09472795832  + 6599.46771964800 *T)
    sumL0  +=       0.181 * cos(4.22950689891  + 966.97087743560 *T)
    sumL0  +=       0.152 * cos(5.28885894415  + 12669.24447420140 *T)
    sumL0  +=       0.150 * cos(5.86819430908  + 97238.62754448748 *T)
    sumL0  +=       0.142 * cos(5.87266532526  + 22476.73502749179 *T)
    sumL0  +=       0.145 * cos(5.07330784304  + 87.30820453981 *T)
    sumL0  +=       0.133 * cos(5.65471067133  + 31.97230581680 *T)
    sumL0  +=       0.124 * cos(2.83326217072  + 12566.21901028560 *T)
    sumL0  +=       0.135 * cos(3.12861731644  + 32217.20018108080 *T)
    sumL0  +=       0.137 * cos(0.86487461904  + 9924.81042151060 *T)
    sumL0  +=       0.172 * cos(1.98369595114  + 174242.46596404971 *T)
    sumL0  +=       0.170 * cos(4.41115280254  + 327574.51427678124 *T)
    sumL0  +=       0.151 * cos(0.46542099527  + 39609.65458316560 *T)
    sumL0  +=       0.148 * cos(2.13439571118  + 491.66329245880 *T)
    sumL0  +=       0.153 * cos(3.78801830344  + 17363.24742890899 *T)
    sumL0  +=       0.165 * cos(5.31654110459  + 16943.76278503380 *T)
    sumL0  +=       0.165 * cos(4.06747587817  + 58953.14544329400 *T)
    sumL0  +=       0.118 * cos(0.63846333239  + 6.06591562980 *T)
    sumL0  +=       0.159 * cos(0.86086959274  + 221995.02880149524 *T)
    sumL0  +=       0.119 * cos(5.96432932413  + 1385.89527633620 *T)
    sumL0  +=       0.114 * cos(5.16516114595  + 25685.87280280800 *T)
    sumL0  +=       0.112 * cos(3.39403722178  + 21393.54196985760 *T)
    sumL0  +=       0.112 * cos(4.92889233335  + 56.80326216980 *T)
    sumL0  +=       0.119 * cos(2.40637635942  + 18635.92845453620 *T)
    sumL0  +=       0.115 * cos(0.23374479051  + 418.92439890060 *T)
    sumL0  +=       0.122 * cos(0.93575234049  + 24492.40611365159 *T)
    sumL0  +=       0.115 * cos(4.58880032176  + 26709.64694241340 *T)
    sumL0  +=       0.130 * cos(4.85539251000  + 22345.26037610820 *T)
    sumL0  +=       0.140 * cos(1.09413073202  + 44809.65020086340 *T)
    sumL0  +=       0.112 * cos(6.05401806281  + 433.71173787680 *T)
    sumL0  +=       0.104 * cos(1.54931540602  + 127.95153303460 *T)
    sumL0  +=       0.105 * cos(4.82620858888  + 33794.54372352860 *T)
    sumL0  +=       0.102 * cos(4.12448497391  + 15664.03552270859 *T)
    sumL0  +=       0.107 * cos(4.67919356465  + 77690.75950573849 *T)
    sumL0  +=       0.118 * cos(4.52320170120  + 19004.64794940840 *T)
    sumL0  +=       0.107 * cos(5.71774478555  + 77736.78343050249 *T)
    sumL0  +=       0.143 * cos(1.81201813018  + 4214.06901508480 *T)
    sumL0  +=       0.125 * cos(1.14419195615  + 625.67019231240 *T)
    sumL0  +=       0.124 * cos(3.27736514057  + 12566.08438968000 *T)
    sumL0  +=       0.110 * cos(1.08682570828  + 2787.04302385740 *T)
    sumL0  +=       0.105 * cos(1.78318141871  + 18139.29450141590 *T)
    sumL0  +=       0.102 * cos(4.75119578149  + 12242.64628332540 *T)
    sumL0  +=       0.137 * cos(1.43510636754  + 86464.61331683119 *T)
    sumL0  +=       0.101 * cos(4.91289409429  + 401.67212175720 *T)
    sumL0  +=       0.129 * cos(1.23567904485  + 12029.34718788740 *T)
    sumL0  +=       0.138 * cos(2.45654707999  + 7576.56007357400 *T)
    sumL0  +=       0.103 * cos(0.40004073416  + 90279.92316810328 *T)
    sumL0  +=       0.108 * cos(0.98989774940  + 5636.06501667660 *T)
    sumL0  +=       0.117 * cos(5.17362872063  + 34520.30930938080 *T)
    sumL0  +=       0.100 * cos(3.95534628189  + 5547.19933645960 *T)
    sumL0  +=       0.098 * cos(1.28118280598  + 21548.96236929180 *T)
    sumL0  +=       0.097 * cos(3.34717130592  + 16310.97904572060 *T)
    sumL0  +=       0.098 * cos(4.37041908717  + 34513.26307268280 *T)
    sumL0  +=       0.125 * cos(2.72164432960  + 24065.80792277559 *T)
    sumL0  +=       0.102 * cos(0.66938025772  + 10239.58386601080 *T)
    sumL0  +=       0.119 * cos(1.21689479331  + 1478.86657406440 *T)
    sumL0  +=       0.094 * cos(1.99595224256  + 13362.44970679920 *T)
    sumL0  +=       0.094 * cos(4.30965982872  + 26880.31981303260 *T)
    sumL0  +=       0.095 * cos(2.89807657534  + 34911.41207609100 *T)
    sumL0  +=       0.106 * cos(1.00156653590  + 16522.65971600220 *T)
    sumL0  +=       0.097 * cos(0.89642320201  + 71980.63357473118 *T)
    sumL0  +=       0.116 * cos(4.19967201116  + 206.70073729660 *T)
    sumL0  +=       0.099 * cos(1.37437847718  + 1039.02661079040 *T)
    sumL0  +=       0.126 * cos(3.21642544972  + 305281.94307104882 *T)
    sumL0  +=       0.094 * cos(0.68997876060  + 7834.12107263940 *T)
    sumL0  +=       0.094 * cos(5.58132218606  + 3104.93005942380 *T)
    sumL0  +=       0.095 * cos(3.03823741110  + 8982.81066930900 *T)
    sumL0  +=       0.108 * cos(0.52696637156  + 276.74577186440 *T)
    sumL0  +=       0.124 * cos(3.43899862683  + 172146.97134054027 *T)
    sumL0  +=       0.102 * cos(1.04031728553  + 95143.13292097810 *T)
    sumL0  +=       0.104 * cos(3.39218586218  + 290.97286586600 *T)
    sumL0  +=       0.110 * cos(3.68205877433  + 22380.75580027400 *T)
    sumL0  +=       0.117 * cos(0.78475956902  + 83286.91426955358 *T)
    sumL0  +=       0.083 * cos(0.18241793425  + 15141.39079431200 *T)
    sumL0  +=       0.089 * cos(4.45371820659  + 792.77488846740 *T)
    sumL0  +=       0.082 * cos(4.80703651241  + 6819.88036208680 *T)
    sumL0  +=       0.087 * cos(3.43122851097  + 27707.54249429480 *T)
    sumL0  +=       0.101 * cos(5.32081603011  + 2301.58581590939 *T)
    sumL0  +=       0.082 * cos(0.87060089842  + 10241.20229116720 *T)
    sumL0  +=       0.086 * cos(4.61919461931  + 36147.40987730040 *T)
    sumL0  +=       0.095 * cos(2.87032884659  + 23020.65308658799 *T)
    sumL0  +=       0.088 * cos(3.21133165690  + 33326.57873317420 *T)
    sumL0  +=       0.080 * cos(1.84900424847  + 21424.46664430340 *T)
    sumL0  +=       0.101 * cos(4.18796434479  + 30666.15495843280 *T)
    sumL0  +=       0.107 * cos(5.77864921649  + 34115.11406927460 *T)
    sumL0  +=       0.104 * cos(1.08739495962  + 6288.59877429880 *T)
    sumL0  +=       0.110 * cos(3.32898859416  + 72140.62866668739 *T)
    sumL0  +=       0.087 * cos(4.40657711727  + 142.17862703620 *T)
    sumL0  +=       0.109 * cos(1.94546030825  + 24279.10701821359 *T)
    sumL0  +=       0.087 * cos(4.32472045435  + 742.99006053260 *T)
    sumL0  +=       0.107 * cos(4.91580912547  + 277.03499374140 *T)
    sumL0  +=       0.088 * cos(2.10180220766  + 26482.17080962440 *T)
    sumL0  +=       0.086 * cos(4.01887374432  + 12491.37010141550 *T)
    sumL0  +=       0.106 * cos(5.49092372854  + 62883.35513951360 *T)
    sumL0  +=       0.080 * cos(6.19781316983  + 6709.67404086740 *T)
    sumL0  +=       0.088 * cos(2.09872810657  + 238004.52415723627 *T)
    sumL0  +=       0.083 * cos(4.90662164029  + 51.28033786241 *T)
    sumL0  +=       0.095 * cos(4.13387406591  + 18216.44381066100 *T)
    sumL0  +=       0.078 * cos(6.06949391680  + 148434.53403769128 *T)
    sumL0  +=       0.079 * cos(3.03048221644  + 838.96928775040 *T)
    sumL0  +=       0.074 * cos(5.49813051211  + 29026.48522950779 *T)
    sumL0  +=       0.073 * cos(3.05008665738  + 567.71863773040 *T)
    sumL0  +=       0.084 * cos(0.46604373274  + 45.14121963660 *T)
    sumL0  +=       0.093 * cos(2.52267536308  + 48739.85989708300 *T)
    sumL0  +=       0.076 * cos(1.76418124905  + 41654.96311596780 *T)
    sumL0  +=       0.067 * cos(5.77851227793  + 6311.52503745920 *T)
    sumL0  +=       0.062 * cos(3.32967880172  + 15508.61512327440 *T)
    sumL0  +=       0.079 * cos(5.59773841328  + 71960.38658322369 *T)
    sumL0  +=       0.057 * cos(3.90629505268  + 5999.21653112620 *T)
    sumL0  +=       0.061 * cos(0.05695043232  + 7856.89627409019 *T)
    sumL0  +=       0.061 * cos(5.63297958433  + 7863.94251078820 *T)
    sumL0  +=       0.065 * cos(3.72178394016  + 12573.26524698360 *T)
    sumL0  +=       0.057 * cos(4.18217219541  + 26087.90314157420 *T)
    sumL0  +=       0.066 * cos(3.92262333487  + 69853.35207568129 *T)
    sumL0  +=       0.053 * cos(5.51119362045  + 77710.24834977149 *T)
    sumL0  +=       0.053 * cos(4.88573986961  + 77717.29458646949 *T)
    sumL0  +=       0.062 * cos(2.88876342225  + 9411.46461508720 *T)
    sumL0  +=       0.051 * cos(1.12657183874  + 82576.98122099529 *T)
    sumL0  +=       0.045 * cos(2.95671076719  + 24602.61243487099 *T)
    sumL0  +=       0.040 * cos(5.55145719241  + 12565.17137891460 *T)
    sumL0  +=       0.039 * cos(1.20838190039  + 18842.11400297339 *T)
    sumL0  +=       0.045 * cos(3.18590558749  + 45585.17281218740 *T)
    sumL0  +=       0.049 * cos(2.44790934886  + 13613.80427733600 *T)

    # EARTH Longitude
    sumL1  = 0
    sumL1  +=  628331966747.491 * cos(0.00000000000  + 0.00000000000 *T)
    sumL1  +=  206058.863 * cos(2.67823455584  + 6283.07584999140 *T)
    sumL1  +=    4303.430 * cos(2.63512650414  + 12566.15169998280 *T)
    sumL1  +=     425.264 * cos(1.59046980729  + 3.52311834900 *T)
    sumL1  +=     108.977 * cos(2.96618001993  + 1577.34354244780 *T)
    sumL1  +=      93.478 * cos(2.59212835365  + 18849.22754997420 *T)
    sumL1  +=     119.261 * cos(5.79557487799  + 26.29831979980 *T)
    sumL1  +=      72.122 * cos(1.13846158196  + 529.69096509460 *T)
    sumL1  +=      67.768 * cos(1.87472304791  + 398.14900340820 *T)
    sumL1  +=      67.327 * cos(4.40918235168  + 5507.55323866740 *T)
    sumL1  +=      59.027 * cos(2.88797038460  + 5223.69391980220 *T)
    sumL1  +=      55.976 * cos(2.17471680261  + 155.42039943420 *T)
    sumL1  +=      45.407 * cos(0.39803079805  + 796.29800681640 *T)
    sumL1  +=      36.369 * cos(0.46624739835  + 775.52261132400 *T)
    sumL1  +=      28.958 * cos(2.64707383882  + 7.11354700080 *T)
    sumL1  +=      19.097 * cos(1.84628332577  + 5486.77784317500 *T)
    sumL1  +=      20.844 * cos(5.34138275149  + 0.98032106820 *T)
    sumL1  +=      18.508 * cos(4.96855124577  + 213.29909543800 *T)
    sumL1  +=      16.233 * cos(0.03216483047  + 2544.31441988340 *T)
    sumL1  +=      17.293 * cos(2.99116864949  + 6275.96230299060 *T)
    sumL1  +=      15.832 * cos(1.43049285325  + 2146.16541647520 *T)
    sumL1  +=      14.615 * cos(1.20532366323  + 10977.07880469900 *T)
    sumL1  +=      11.877 * cos(3.25804815607  + 5088.62883976680 *T)
    sumL1  +=      11.514 * cos(2.07502418155  + 4694.00295470760 *T)
    sumL1  +=       9.721 * cos(4.23925472239  + 1349.86740965880 *T)
    sumL1  +=       9.969 * cos(1.30262991097  + 6286.59896834040 *T)
    sumL1  +=       9.452 * cos(2.69957062864  + 242.72860397400 *T)
    sumL1  +=      12.461 * cos(2.83432285512  + 1748.01641306700 *T)
    sumL1  +=      11.808 * cos(5.27379790480  + 1194.44701022460 *T)
    sumL1  +=       8.577 * cos(5.64475868067  + 951.71840625060 *T)
    sumL1  +=      10.641 * cos(0.76614199202  + 553.56940284240 *T)
    sumL1  +=       7.576 * cos(5.30062664886  + 2352.86615377180 *T)
    sumL1  +=       5.834 * cos(1.76649917904  + 1059.38193018920 *T)
    sumL1  +=       6.385 * cos(2.65033984967  + 9437.76293488700 *T)
    sumL1  +=       5.223 * cos(5.66135767624  + 71430.69561812909 *T)
    sumL1  +=       5.305 * cos(0.90857521574  + 3154.68708489560 *T)
    sumL1  +=       6.101 * cos(4.66632584188  + 4690.47983635860 *T)
    sumL1  +=       4.330 * cos(0.24102555403  + 6812.76681508600 *T)
    sumL1  +=       5.041 * cos(1.42490103709  + 6438.49624942560 *T)
    sumL1  +=       4.259 * cos(0.77355900599  + 10447.38783960440 *T)
    sumL1  +=       5.198 * cos(1.85353197345  + 801.82093112380 *T)
    sumL1  +=       3.744 * cos(2.00119516488  + 8031.09226305840 *T)
    sumL1  +=       3.558 * cos(2.42901552681  + 14143.49524243060 *T)
    sumL1  +=       3.372 * cos(3.86210700128  + 1592.59601363280 *T)
    sumL1  +=       3.374 * cos(0.88776219727  + 12036.46073488820 *T)
    sumL1  +=       3.175 * cos(3.18785710594  + 4705.73230754360 *T)
    sumL1  +=       3.221 * cos(0.61599835472  + 8429.24126646660 *T)
    sumL1  +=       4.132 * cos(5.23992859705  + 7084.89678111520 *T)
    sumL1  +=       2.970 * cos(6.07026318493  + 4292.33083295040 *T)
    sumL1  +=       2.900 * cos(2.32464208411  + 20.35531939880 *T)
    sumL1  +=       3.504 * cos(4.79975694359  + 6279.55273164240 *T)
    sumL1  +=       2.950 * cos(1.43108874817  + 5746.27133789600 *T)
    sumL1  +=       2.697 * cos(4.80368225199  + 7234.79425624200 *T)
    sumL1  +=       2.531 * cos(6.22290682655  + 6836.64525283380 *T)
    sumL1  +=       2.745 * cos(0.93466065396  + 5760.49843189760 *T)
    sumL1  +=       3.250 * cos(3.39954640038  + 7632.94325965020 *T)
    sumL1  +=       2.277 * cos(5.00277837672  + 17789.84561978500 *T)
    sumL1  +=       2.075 * cos(3.95534978634  + 10213.28554621100 *T)
    sumL1  +=       2.061 * cos(2.22411683077  + 5856.47765911540 *T)
    sumL1  +=       2.252 * cos(5.67166499885  + 11499.65622279280 *T)
    sumL1  +=       2.148 * cos(5.20184578235  + 11513.88331679440 *T)
    sumL1  +=       1.886 * cos(0.53198320577  + 3340.61242669980 *T)
    sumL1  +=       1.875 * cos(4.73511970207  + 83996.84731811189 *T)
    sumL1  +=       2.060 * cos(2.54987293999  + 25132.30339996560 *T)
    sumL1  +=       1.794 * cos(1.47435409831  + 4164.31198961300 *T)
    sumL1  +=       1.778 * cos(3.02473091781  + 5.52292430740 *T)
    sumL1  +=       2.029 * cos(0.90960209983  + 6256.77753019160 *T)
    sumL1  +=       2.075 * cos(2.26767270157  + 522.57741809380 *T)
    sumL1  +=       1.772 * cos(3.02622802353  + 5753.38488489680 *T)
    sumL1  +=       1.569 * cos(6.12410242782  + 5216.58037280140 *T)
    sumL1  +=       1.590 * cos(4.63713748247  + 3.28635741780 *T)
    sumL1  +=       1.542 * cos(4.20004448567  + 13367.97263110660 *T)
    sumL1  +=       1.427 * cos(1.19088061711  + 3894.18182954220 *T)
    sumL1  +=       1.375 * cos(3.09301252193  + 135.06508003540 *T)
    sumL1  +=       1.359 * cos(4.24532506641  + 426.59819087600 *T)
    sumL1  +=       1.340 * cos(5.76511818622  + 6040.34724601740 *T)
    sumL1  +=       1.284 * cos(3.08524663344  + 5643.17856367740 *T)
    sumL1  +=       1.250 * cos(3.07748157144  + 11926.25441366880 *T)
    sumL1  +=       1.551 * cos(3.07665451458  + 6681.22485339960 *T)
    sumL1  +=       1.268 * cos(2.09196018331  + 6290.18939699220 *T)
    sumL1  +=       1.144 * cos(3.24444699514  + 12168.00269657460 *T)
    sumL1  +=       1.248 * cos(3.44504937285  + 536.80451209540 *T)
    sumL1  +=       1.118 * cos(2.31829670425  + 16730.46368959580 *T)
    sumL1  +=       1.105 * cos(5.31966001019  + 23.87843774780 *T)
    sumL1  +=       1.051 * cos(3.75015946014  + 7860.41939243920 *T)
    sumL1  +=       1.025 * cos(2.44688534235  + 1990.74501704100 *T)
    sumL1  +=       0.962 * cos(0.81771017882  + 3.88133535800 *T)
    sumL1  +=       0.910 * cos(0.41727865299  + 7079.37385680780 *T)
    sumL1  +=       0.883 * cos(5.16833917651  + 11790.62908865880 *T)
    sumL1  +=       0.957 * cos(4.07673573735  + 6127.65545055720 *T)
    sumL1  +=       1.110 * cos(3.90096793825  + 11506.76976979360 *T)
    sumL1  +=       0.802 * cos(3.88778875582  + 10973.55568635000 *T)
    sumL1  +=       0.780 * cos(2.39934293755  + 1589.07289528380 *T)
    sumL1  +=       0.758 * cos(1.30034364248  + 103.09277421860 *T)
    sumL1  +=       0.749 * cos(4.96275803300  + 6496.37494542940 *T)
    sumL1  +=       0.765 * cos(3.36312388424  + 36.02786667740 *T)
    sumL1  +=       0.915 * cos(5.41543742089  + 206.18554843720 *T)
    sumL1  +=       0.776 * cos(2.57589093871  + 11371.70468975820 *T)
    sumL1  +=       0.772 * cos(3.98369209464  + 955.59974160860 *T)
    sumL1  +=       0.749 * cos(5.17890001805  + 10969.96525769820 *T)
    sumL1  +=       0.806 * cos(0.34218864254  + 9917.69687450980 *T)
    sumL1  +=       0.728 * cos(5.20962563787  + 38.02767263580 *T)
    sumL1  +=       0.685 * cos(2.77592961854  + 20.77539549240 *T)
    sumL1  +=       0.636 * cos(4.28242193632  + 28.44918746780 *T)
    sumL1  +=       0.608 * cos(5.63278508906  + 10984.19235169980 *T)
    sumL1  +=       0.704 * cos(5.60738823665  + 3738.76143010800 *T)
    sumL1  +=       0.685 * cos(0.38876148682  + 15.25247118500 *T)
    sumL1  +=       0.601 * cos(0.73489602442  + 419.48464387520 *T)
    sumL1  +=       0.716 * cos(2.65279791438  + 6309.37416979120 *T)
    sumL1  +=       0.584 * cos(5.54502568227  + 17298.18232732620 *T)
    sumL1  +=       0.650 * cos(1.13379656406  + 7058.59846131540 *T)
    sumL1  +=       0.688 * cos(2.59683891779  + 3496.03282613400 *T)
    sumL1  +=       0.485 * cos(0.44467180946  + 12352.85260454480 *T)
    sumL1  +=       0.528 * cos(2.74936967681  + 3930.20969621960 *T)
    sumL1  +=       0.597 * cos(5.27668281777  + 10575.40668294180 *T)
    sumL1  +=       0.583 * cos(3.18929067810  + 4732.03062734340 *T)
    sumL1  +=       0.526 * cos(5.01697321546  + 5884.92684658320 *T)
    sumL1  +=       0.540 * cos(1.29175137075  + 640.87760738220 *T)
    sumL1  +=       0.473 * cos(5.49953306970  + 5230.80746680300 *T)
    sumL1  +=       0.406 * cos(5.21248452189  + 220.41264243880 *T)
    sumL1  +=       0.395 * cos(1.87474483222  + 16200.77272450120 *T)
    sumL1  +=       0.370 * cos(3.84921354713  + 18073.70493865020 *T)
    sumL1  +=       0.367 * cos(0.88533542778  + 6283.14316029419 *T)
    sumL1  +=       0.379 * cos(0.37983009325  + 10177.25767953360 *T)
    sumL1  +=       0.356 * cos(3.84145204913  + 11712.95531823080 *T)
    sumL1  +=       0.374 * cos(5.01577520608  + 7.04623669800 *T)
    sumL1  +=       0.381 * cos(4.30250406634  + 6062.66320755260 *T)
    sumL1  +=       0.471 * cos(0.86381834647  + 6069.77675455340 *T)
    sumL1  +=       0.367 * cos(1.32943839763  + 6283.00853968860 *T)
    sumL1  +=       0.460 * cos(5.19667219575  + 6284.05617105960 *T)
    sumL1  +=       0.333 * cos(5.54256205741  + 4686.88940770680 *T)
    sumL1  +=       0.341 * cos(4.36522989934  + 7238.67559160000 *T)
    sumL1  +=       0.336 * cos(4.00205876835  + 3097.88382272579 *T)
    sumL1  +=       0.359 * cos(6.22679790284  + 245.83164622940 *T)
    sumL1  +=       0.307 * cos(2.35299010924  + 170.67287061920 *T)
    sumL1  +=       0.343 * cos(3.77164927143  + 6076.89030155420 *T)
    sumL1  +=       0.296 * cos(5.44152227481  + 17260.15465469040 *T)
    sumL1  +=       0.328 * cos(0.13837875384  + 11015.10647733480 *T)
    sumL1  +=       0.268 * cos(1.13904550630  + 12569.67481833180 *T)
    sumL1  +=       0.263 * cos(0.00538633678  + 4136.91043351620 *T)
    sumL1  +=       0.282 * cos(5.04399837480  + 7477.52286021600 *T)
    sumL1  +=       0.288 * cos(3.13401177517  + 12559.03815298200 *T)
    sumL1  +=       0.259 * cos(0.93882269387  + 5642.19824260920 *T)
    sumL1  +=       0.292 * cos(1.98420020514  + 12132.43996210600 *T)
    sumL1  +=       0.247 * cos(3.84244798532  + 5429.87946823940 *T)
    sumL1  +=       0.245 * cos(5.70467521726  + 65147.61976813770 *T)
    sumL1  +=       0.241 * cos(0.99480969552  + 3634.62102451840 *T)
    sumL1  +=       0.246 * cos(3.06168069935  + 110.20632121940 *T)
    sumL1  +=       0.239 * cos(6.11855909114  + 11856.21865142450 *T)
    sumL1  +=       0.263 * cos(0.66348415419  + 21228.39202354580 *T)
    sumL1  +=       0.262 * cos(1.51070507866  + 12146.66705610760 *T)
    sumL1  +=       0.230 * cos(1.75927314884  + 9779.10867612540 *T)
    sumL1  +=       0.223 * cos(2.00967043606  + 6172.86952877200 *T)
    sumL1  +=       0.246 * cos(1.10411690865  + 6282.09552892320 *T)
    sumL1  +=       0.221 * cos(3.03945240854  + 8635.94200376320 *T)
    sumL1  +=       0.214 * cos(4.03840869663  + 14314.16811304980 *T)
    sumL1  +=       0.236 * cos(5.46915070580  + 13916.01910964160 *T)
    sumL1  +=       0.224 * cos(4.68408089456  + 24072.92146977640 *T)
    sumL1  +=       0.212 * cos(2.13695625494  + 5849.36411211460 *T)
    sumL1  +=       0.207 * cos(3.07724246401  + 11.72935283600 *T)
    sumL1  +=       0.207 * cos(6.10306282747  + 23543.23050468179 *T)
    sumL1  +=       0.266 * cos(1.00709566823  + 2388.89402044920 *T)
    sumL1  +=       0.217 * cos(6.27837036335  + 17267.26820169119 *T)
    sumL1  +=       0.204 * cos(2.34615348695  + 266.60704172180 *T)
    sumL1  +=       0.195 * cos(5.55015549753  + 6133.51265285680 *T)
    sumL1  +=       0.188 * cos(2.52667166175  + 6525.80445396540 *T)
    sumL1  +=       0.185 * cos(0.90960768344  + 18319.53658487960 *T)
    sumL1  +=       0.177 * cos(1.73429218289  + 154717.60988768269 *T)
    sumL1  +=       0.187 * cos(4.76483647432  + 4535.05943692440 *T)
    sumL1  +=       0.186 * cos(4.63080493407  + 10440.27429260360 *T)
    sumL1  +=       0.215 * cos(2.81255454560  + 7342.45778018060 *T)
    sumL1  +=       0.172 * cos(1.45551888559  + 9225.53927328300 *T)
    sumL1  +=       0.162 * cos(3.30661909388  + 639.89728631400 *T)
    sumL1  +=       0.168 * cos(2.17671416605  + 27.40155609680 *T)
    sumL1  +=       0.160 * cos(1.68164180475  + 15110.46611986620 *T)
    sumL1  +=       0.158 * cos(0.13519771874  + 13095.84266507740 *T)
    sumL1  +=       0.183 * cos(0.56281322071  + 13517.87010623340 *T)
    sumL1  +=       0.179 * cos(3.58450811616  + 87.30820453981 *T)
    sumL1  +=       0.152 * cos(2.84070476818  + 5650.29211067820 *T)
    sumL1  +=       0.182 * cos(0.44065530624  + 17253.04110768959 *T)
    sumL1  +=       0.160 * cos(5.95767264171  + 4701.11650170840 *T)
    sumL1  +=       0.142 * cos(1.46290137520  + 11087.28512591840 *T)
    sumL1  +=       0.142 * cos(2.04464036087  + 20426.57109242200 *T)
    sumL1  +=       0.131 * cos(5.40912137746  + 2699.73481931760 *T)
    sumL1  +=       0.144 * cos(2.07312090485  + 25158.60171976540 *T)
    sumL1  +=       0.147 * cos(6.15106982168  + 9623.68827669120 *T)
    sumL1  +=       0.141 * cos(5.55739979498  + 10454.50138660520 *T)
    sumL1  +=       0.135 * cos(0.06098110407  + 16723.35014259500 *T)
    sumL1  +=       0.124 * cos(5.81218025669  + 17256.63153634140 *T)
    sumL1  +=       0.124 * cos(2.36293551623  + 4933.20844033260 *T)
    sumL1  +=       0.126 * cos(3.47435905118  + 22483.84857449259 *T)
    sumL1  +=       0.159 * cos(5.63954754618  + 5729.50644714900 *T)
    sumL1  +=       0.123 * cos(3.92815963256  + 17996.03116822220 *T)
    sumL1  +=       0.148 * cos(3.02509280598  + 1551.04522264800 *T)
    sumL1  +=       0.120 * cos(5.91904349732  + 6206.80977871580 *T)
    sumL1  +=       0.134 * cos(3.11122937825  + 21954.15760939799 *T)
    sumL1  +=       0.119 * cos(5.52141123450  + 709.93304855830 *T)
    sumL1  +=       0.122 * cos(3.00813429479  + 19800.94595622480 *T)
    sumL1  +=       0.127 * cos(1.37618620001  + 14945.31617355440 *T)
    sumL1  +=       0.141 * cos(2.56889468729  + 1052.26838318840 *T)
    sumL1  +=       0.123 * cos(2.83671175442  + 11919.14086666800 *T)
    sumL1  +=       0.118 * cos(0.81934438215  + 5331.35744374080 *T)
    sumL1  +=       0.151 * cos(2.68731829165  + 11769.85369316640 *T)
    sumL1  +=       0.119 * cos(5.08835797638  + 5481.25491886760 *T)
    sumL1  +=       0.153 * cos(2.46021790779  + 11933.36796066960 *T)
    sumL1  +=       0.108 * cos(1.04936452145  + 11403.67699557500 *T)
    sumL1  +=       0.128 * cos(0.99794735107  + 8827.39026987480 *T)
    sumL1  +=       0.144 * cos(2.54869747042  + 227.47613278900 *T)
    sumL1  +=       0.150 * cos(4.50631437136  + 2379.16447357160 *T)
    sumL1  +=       0.107 * cos(1.79272017026  + 13119.72110282519 *T)
    sumL1  +=       0.107 * cos(4.43556814486  + 18422.62935909819 *T)
    sumL1  +=       0.109 * cos(0.29269062317  + 16737.57723659660 *T)
    sumL1  +=       0.141 * cos(3.18979826258  + 6262.30045449900 *T)
    sumL1  +=       0.122 * cos(4.23040027813  + 29.42950853600 *T)
    sumL1  +=       0.111 * cos(5.16954029551  + 17782.73207278420 *T)
    sumL1  +=       0.100 * cos(3.52213872761  + 18052.92954315780 *T)
    sumL1  +=       0.108 * cos(1.08514212991  + 16858.48253293320 *T)
    sumL1  +=       0.106 * cos(1.96085248410  + 74.78159856730 *T)
    sumL1  +=       0.110 * cos(2.30582372873  + 16460.33352952499 *T)
    sumL1  +=       0.097 * cos(3.50918940210  + 5333.90024102160 *T)
    sumL1  +=       0.099 * cos(3.56417337974  + 735.87651353180 *T)
    sumL1  +=       0.094 * cos(5.01857894228  + 3128.38876509580 *T)
    sumL1  +=       0.097 * cos(1.65579893894  + 533.21408344360 *T)
    sumL1  +=       0.092 * cos(0.89217162285  + 29296.61538957860 *T)
    sumL1  +=       0.123 * cos(3.16062050433  + 9380.95967271720 *T)
    sumL1  +=       0.102 * cos(1.20493500565  + 23020.65308658799 *T)
    sumL1  +=       0.088 * cos(2.21296088224  + 12721.57209941700 *T)
    sumL1  +=       0.089 * cos(1.54264720310  + 20199.09495963300 *T)
    sumL1  +=       0.113 * cos(4.83320707870  + 16496.36139620240 *T)
    sumL1  +=       0.121 * cos(6.19860353182  + 9388.00590941520 *T)
    sumL1  +=       0.089 * cos(4.08082274765  + 22805.73556599360 *T)
    sumL1  +=       0.098 * cos(1.09181832830  + 12043.57428188900 *T)
    sumL1  +=       0.086 * cos(1.13655027605  + 143571.32428481648 *T)
    sumL1  +=       0.088 * cos(5.96980472191  + 107.66352393860 *T)
    sumL1  +=       0.082 * cos(5.01340404594  + 22003.91463486980 *T)
    sumL1  +=       0.094 * cos(1.69615700473  + 23006.42599258639 *T)
    sumL1  +=       0.081 * cos(3.00657814365  + 2118.76386037840 *T)
    sumL1  +=       0.098 * cos(1.39215287161  + 8662.24032356300 *T)
    sumL1  +=       0.077 * cos(3.33555190840  + 15720.83878487840 *T)
    sumL1  +=       0.082 * cos(5.86880116464  + 2787.04302385740 *T)
    sumL1  +=       0.076 * cos(5.67183650604  + 14.22709400160 *T)
    sumL1  +=       0.081 * cos(6.16619455699  + 1039.02661079040 *T)
    sumL1  +=       0.076 * cos(3.21449884756  + 111.18664228760 *T)
    sumL1  +=       0.078 * cos(1.37531518377  + 21947.11137270000 *T)
    sumL1  +=       0.074 * cos(3.58814195051  + 11609.86254401220 *T)
    sumL1  +=       0.077 * cos(4.84846488388  + 22743.40937951640 *T)
    sumL1  +=       0.090 * cos(1.48869013606  + 15671.08175940660 *T)
    sumL1  +=       0.082 * cos(3.48618399109  + 29088.81141598500 *T)
    sumL1  +=       0.069 * cos(3.55746476593  + 4590.91018048900 *T)
    sumL1  +=       0.069 * cos(1.93625656075  + 135.62532501000 *T)
    sumL1  +=       0.070 * cos(2.66548322237  + 18875.52586977400 *T)
    sumL1  +=       0.069 * cos(5.41478093731  + 26735.94526221320 *T)
    sumL1  +=       0.079 * cos(5.15154513662  + 12323.42309600880 *T)
    sumL1  +=       0.094 * cos(3.62899392448  + 77713.77146812050 *T)
    sumL1  +=       0.078 * cos(4.17011182047  + 1066.49547719000 *T)
    sumL1  +=       0.071 * cos(3.89435637865  + 22779.43724619380 *T)
    sumL1  +=       0.063 * cos(4.53968787714  + 8982.81066930900 *T)
    sumL1  +=       0.069 * cos(0.96028230548  + 14919.01785375460 *T)
    sumL1  +=       0.076 * cos(3.29092216589  + 2942.46342329160 *T)
    sumL1  +=       0.063 * cos(4.09167842893  + 16062.18452611680 *T)
    sumL1  +=       0.065 * cos(3.34580407184  + 51.28033786241 *T)
    sumL1  +=       0.065 * cos(5.75757544877  + 52670.06959330260 *T)
    sumL1  +=       0.068 * cos(5.75884067555  + 21424.46664430340 *T)
    sumL1  +=       0.057 * cos(5.45122399850  + 12592.45001978260 *T)
    sumL1  +=       0.057 * cos(5.25043362558  + 20995.39296644940 *T)
    sumL1  +=       0.073 * cos(0.53299090807  + 2301.58581590939 *T)
    sumL1  +=       0.070 * cos(4.31243357502  + 19402.79695281660 *T)
    sumL1  +=       0.067 * cos(2.53852336668  + 377.37360791580 *T)
    sumL1  +=       0.056 * cos(3.20816844695  + 24889.57479599160 *T)
    sumL1  +=       0.053 * cos(3.17816599142  + 18451.07854656599 *T)
    sumL1  +=       0.053 * cos(3.61529270216  + 77.67377042800 *T)
    sumL1  +=       0.053 * cos(0.45467549335  + 30666.15495843280 *T)
    sumL1  +=       0.061 * cos(0.14807288453  + 23013.53953958720 *T)
    sumL1  +=       0.051 * cos(3.32803972907  + 56.89837493560 *T)
    sumL1  +=       0.052 * cos(3.41177624177  + 23141.55838292460 *T)
    sumL1  +=       0.058 * cos(3.13638677202  + 309.27832265580 *T)
    sumL1  +=       0.070 * cos(2.50592323465  + 31415.37924995700 *T)
    sumL1  +=       0.052 * cos(5.10673376738  + 17796.95916678580 *T)
    sumL1  +=       0.067 * cos(6.27917920454  + 22345.26037610820 *T)
    sumL1  +=       0.050 * cos(0.42577644151  + 25685.87280280800 *T)
    sumL1  +=       0.048 * cos(0.70204553333  + 1162.47470440780 *T)
    sumL1  +=       0.066 * cos(3.64350022359  + 15265.88651930040 *T)
    sumL1  +=       0.050 * cos(5.74382917440  + 19.66976089979 *T)
    sumL1  +=       0.050 * cos(4.69825387775  + 28237.23345938940 *T)
    sumL1  +=       0.047 * cos(5.74015846442  + 12139.55350910680 *T)
    sumL1  +=       0.054 * cos(1.97301333704  + 23581.25817731760 *T)
    sumL1  +=       0.049 * cos(4.98223579027  + 10021.83728009940 *T)
    sumL1  +=       0.046 * cos(5.41431705539  + 33019.02111220460 *T)
    sumL1  +=       0.051 * cos(1.23882053879  + 12539.85338018300 *T)
    sumL1  +=       0.046 * cos(2.41369976086  + 98068.53671630540 *T)
    sumL1  +=       0.044 * cos(0.80750593746  + 167283.76158766547 *T)
    sumL1  +=       0.045 * cos(4.39613584445  + 433.71173787680 *T)
    sumL1  +=       0.044 * cos(2.57358208785  + 12964.30070339100 *T)
    sumL1  +=       0.046 * cos(0.26142733448  + 11.04570026390 *T)
    sumL1  +=       0.045 * cos(2.46230645202  + 51868.24866217880 *T)
    sumL1  +=       0.048 * cos(0.89551707131  + 56600.27928952220 *T)
    sumL1  +=       0.057 * cos(1.86416707010  + 25287.72379939980 *T)
    sumL1  +=       0.042 * cos(5.26377513431  + 26084.02180621620 *T)
    sumL1  +=       0.049 * cos(3.17757670611  + 6303.85124548380 *T)
    sumL1  +=       0.052 * cos(3.65266055509  + 7872.14874527520 *T)
    sumL1  +=       0.040 * cos(1.81891629936  + 34596.36465465240 *T)
    sumL1  +=       0.043 * cos(1.94164978061  + 1903.43681250120 *T)
    sumL1  +=       0.041 * cos(0.74461854136  + 23937.85638974100 *T)
    sumL1  +=       0.048 * cos(6.26034008181  + 28286.99048486120 *T)
    sumL1  +=       0.045 * cos(5.45575017530  + 60530.48898574180 *T)
    sumL1  +=       0.040 * cos(2.92105728682  + 21548.96236929180 *T)
    sumL1  +=       0.040 * cos(0.04502010161  + 38526.57435087200 *T)
    sumL1  +=       0.053 * cos(3.64791042082  + 11925.27409260060 *T)
    sumL1  +=       0.041 * cos(5.04048954693  + 27832.03821928320 *T)
    sumL1  +=       0.042 * cos(5.19292937193  + 19004.64794940840 *T)
    sumL1  +=       0.040 * cos(2.57120233428  + 24356.78078864160 *T)
    sumL1  +=       0.038 * cos(3.49190341464  + 226858.23855437007 *T)
    sumL1  +=       0.039 * cos(4.61184303844  + 95.97922721780 *T)
    sumL1  +=       0.043 * cos(2.20648228147  + 13521.75144159140 *T)
    sumL1  +=       0.040 * cos(5.83461945819  + 16193.65917750039 *T)
    sumL1  +=       0.045 * cos(3.73714372195  + 7875.67186362420 *T)
    sumL1  +=       0.043 * cos(1.14078465002  + 49.75702547180 *T)
    sumL1  +=       0.037 * cos(1.29390383811  + 310.84079886840 *T)
    sumL1  +=       0.038 * cos(0.95970925950  + 664.75604513000 *T)
    sumL1  +=       0.037 * cos(4.27532649462  + 6709.67404086740 *T)
    sumL1  +=       0.038 * cos(2.20108541046  + 28628.33622609960 *T)
    sumL1  +=       0.039 * cos(0.85957361635  + 16522.65971600220 *T)
    sumL1  +=       0.040 * cos(4.35214003837  + 48739.85989708300 *T)
    sumL1  +=       0.036 * cos(1.68167662194  + 10344.29506538580 *T)
    sumL1  +=       0.040 * cos(5.13217319067  + 15664.03552270859 *T)
    sumL1  +=       0.036 * cos(3.72187132496  + 30774.50164257480 *T)
    sumL1  +=       0.036 * cos(3.32158458257  + 16207.88627150200 *T)
    sumL1  +=       0.045 * cos(3.94202418608  + 10988.80815753500 *T)
    sumL1  +=       0.039 * cos(1.51948786199  + 12029.34718788740 *T)
    sumL1  +=       0.026 * cos(3.87685883180  + 6262.72053059260 *T)
    sumL1  +=       0.024 * cos(4.91804163466  + 19651.04848109800 *T)
    sumL1  +=       0.023 * cos(0.29300197709  + 13362.44970679920 *T)
    sumL1  +=       0.021 * cos(3.18605672363  + 6277.55292568400 *T)
    sumL1  +=       0.021 * cos(6.07546891132  + 18139.29450141590 *T)
    sumL1  +=       0.022 * cos(2.31199937177  + 6303.43116939020 *T)
    sumL1  +=       0.021 * cos(3.58418394393  + 18209.33026366019 *T)
    sumL1  +=       0.026 * cos(2.06801296900  + 12573.26524698360 *T)
    sumL1  +=       0.021 * cos(1.56857722317  + 13341.67431130680 *T)
    sumL1  +=       0.024 * cos(5.72605158675  + 29864.33402730900 *T)
    sumL1  +=       0.024 * cos(1.40237993205  + 14712.31711645800 *T)
    sumL1  +=       0.025 * cos(5.71466092822  + 25934.12433108940 *T)

    # EARTH Longitude
    sumL2  = 0
    sumL2  +=   52918.870 * cos(0.00000000000  + 0.00000000000 *T)
    sumL2  +=    8719.837 * cos(1.07209665242  + 6283.07584999140 *T)
    sumL2  +=     309.125 * cos(0.86728818832  + 12566.15169998280 *T)
    sumL2  +=      27.339 * cos(0.05297871691  + 3.52311834900 *T)
    sumL2  +=      16.334 * cos(5.18826691036  + 26.29831979980 *T)
    sumL2  +=      15.752 * cos(3.68457889430  + 155.42039943420 *T)
    sumL2  +=       9.541 * cos(0.75742297675  + 18849.22754997420 *T)
    sumL2  +=       8.937 * cos(2.05705419118  + 77713.77146812050 *T)
    sumL2  +=       6.952 * cos(0.82673305410  + 775.52261132400 *T)
    sumL2  +=       5.064 * cos(4.66284525271  + 1577.34354244780 *T)
    sumL2  +=       4.061 * cos(1.03057162962  + 7.11354700080 *T)
    sumL2  +=       3.463 * cos(5.14074632811  + 796.29800681640 *T)
    sumL2  +=       3.169 * cos(6.05291851171  + 5507.55323866740 *T)
    sumL2  +=       3.020 * cos(1.19246506441  + 242.72860397400 *T)
    sumL2  +=       2.886 * cos(6.11652627155  + 529.69096509460 *T)
    sumL2  +=       3.810 * cos(3.44050803490  + 5573.14280143310 *T)
    sumL2  +=       2.714 * cos(0.30637881025  + 398.14900340820 *T)
    sumL2  +=       2.371 * cos(4.38118838167  + 5223.69391980220 *T)
    sumL2  +=       2.538 * cos(2.27992810679  + 553.56940284240 *T)
    sumL2  +=       2.079 * cos(3.75435330484  + 0.98032106820 *T)
    sumL2  +=       1.675 * cos(0.90216407959  + 951.71840625060 *T)
    sumL2  +=       1.534 * cos(5.75900462759  + 1349.86740965880 *T)
    sumL2  +=       1.224 * cos(2.97328088405  + 2146.16541647520 *T)
    sumL2  +=       1.449 * cos(4.36415913970  + 1748.01641306700 *T)
    sumL2  +=       1.341 * cos(3.72061130861  + 1194.44701022460 *T)
    sumL2  +=       1.254 * cos(2.94846826628  + 6438.49624942560 *T)
    sumL2  +=       0.999 * cos(5.98640014468  + 6286.59896834040 *T)
    sumL2  +=       0.917 * cos(4.79788687522  + 5088.62883976680 *T)
    sumL2  +=       0.828 * cos(3.31321076572  + 213.29909543800 *T)
    sumL2  +=       1.103 * cos(1.27104454479  + 161000.68573767410 *T)
    sumL2  +=       0.762 * cos(3.41582762988  + 5486.77784317500 *T)
    sumL2  +=       1.044 * cos(0.60409577691  + 3154.68708489560 *T)
    sumL2  +=       0.887 * cos(5.23465144638  + 7084.89678111520 *T)
    sumL2  +=       0.645 * cos(1.60096192515  + 2544.31441988340 *T)
    sumL2  +=       0.681 * cos(3.43155669169  + 4694.00295470760 *T)
    sumL2  +=       0.605 * cos(2.47806340546  + 10977.07880469900 *T)
    sumL2  +=       0.706 * cos(6.19393222575  + 4690.47983635860 *T)
    sumL2  +=       0.643 * cos(1.98042503148  + 801.82093112380 *T)
    sumL2  +=       0.502 * cos(1.44394375363  + 6836.64525283380 *T)
    sumL2  +=       0.490 * cos(2.34129524194  + 1592.59601363280 *T)
    sumL2  +=       0.458 * cos(1.30876448575  + 4292.33083295040 *T)
    sumL2  +=       0.431 * cos(0.03526421494  + 7234.79425624200 *T)
    sumL2  +=       0.379 * cos(3.17030522615  + 6309.37416979120 *T)
    sumL2  +=       0.348 * cos(0.99049550009  + 6040.34724601740 *T)
    sumL2  +=       0.386 * cos(1.57019797263  + 71430.69561812909 *T)
    sumL2  +=       0.347 * cos(0.67013291338  + 1059.38193018920 *T)
    sumL2  +=       0.458 * cos(3.81499443681  + 149854.40013480789 *T)
    sumL2  +=       0.302 * cos(1.91760044838  + 10447.38783960440 *T)
    sumL2  +=       0.307 * cos(3.55343347416  + 8031.09226305840 *T)
    sumL2  +=       0.395 * cos(4.93701776616  + 7632.94325965020 *T)
    sumL2  +=       0.314 * cos(3.18093696547  + 2352.86615377180 *T)
    sumL2  +=       0.282 * cos(4.41936437052  + 9437.76293488700 *T)
    sumL2  +=       0.276 * cos(2.71314254553  + 3894.18182954220 *T)
    sumL2  +=       0.298 * cos(2.52037474210  + 6127.65545055720 *T)
    sumL2  +=       0.230 * cos(1.37790215549  + 4705.73230754360 *T)
    sumL2  +=       0.252 * cos(0.55330133471  + 6279.55273164240 *T)
    sumL2  +=       0.255 * cos(5.26570187369  + 6812.76681508600 *T)
    sumL2  +=       0.275 * cos(0.67264264272  + 25132.30339996560 *T)
    sumL2  +=       0.178 * cos(0.92820785174  + 1990.74501704100 *T)
    sumL2  +=       0.221 * cos(0.63897368842  + 6256.77753019160 *T)
    sumL2  +=       0.155 * cos(0.77319790838  + 14143.49524243060 *T)
    sumL2  +=       0.150 * cos(2.40470465561  + 426.59819087600 *T)
    sumL2  +=       0.196 * cos(6.06877865012  + 640.87760738220 *T)
    sumL2  +=       0.137 * cos(2.21679460145  + 8429.24126646660 *T)
    sumL2  +=       0.127 * cos(3.26094223174  + 17789.84561978500 *T)
    sumL2  +=       0.128 * cos(5.47237279946  + 12036.46073488820 *T)
    sumL2  +=       0.122 * cos(2.16291082757  + 10213.28554621100 *T)
    sumL2  +=       0.118 * cos(0.45789822268  + 7058.59846131540 *T)
    sumL2  +=       0.141 * cos(2.34932647403  + 11506.76976979360 *T)
    sumL2  +=       0.100 * cos(0.85621569847  + 6290.18939699220 *T)
    sumL2  +=       0.092 * cos(5.10587476002  + 7079.37385680780 *T)
    sumL2  +=       0.126 * cos(2.65428307012  + 88860.05707098669 *T)
    sumL2  +=       0.106 * cos(5.85646710022  + 7860.41939243920 *T)
    sumL2  +=       0.084 * cos(3.57457554262  + 16730.46368959580 *T)
    sumL2  +=       0.089 * cos(4.21433259618  + 83996.84731811189 *T)
    sumL2  +=       0.097 * cos(5.57938280855  + 13367.97263110660 *T)
    sumL2  +=       0.102 * cos(2.05853060226  + 87.30820453981 *T)
    sumL2  +=       0.080 * cos(4.73792651816  + 11926.25441366880 *T)
    sumL2  +=       0.080 * cos(5.41418965044  + 10973.55568635000 *T)
    sumL2  +=       0.106 * cos(4.10978997399  + 3496.03282613400 *T)
    sumL2  +=       0.102 * cos(3.62650006043  + 244287.60000722768 *T)
    sumL2  +=       0.075 * cos(4.89483161769  + 5643.17856367740 *T)
    sumL2  +=       0.087 * cos(0.42863750683  + 11015.10647733480 *T)
    sumL2  +=       0.069 * cos(1.88908760720  + 10177.25767953360 *T)
    sumL2  +=       0.089 * cos(1.35567273119  + 6681.22485339960 *T)
    sumL2  +=       0.066 * cos(0.99455837265  + 6525.80445396540 *T)
    sumL2  +=       0.067 * cos(5.51240997070  + 3097.88382272579 *T)
    sumL2  +=       0.076 * cos(2.72016814799  + 4164.31198961300 *T)
    sumL2  +=       0.063 * cos(1.44349902540  + 9917.69687450980 *T)
    sumL2  +=       0.078 * cos(3.51469733747  + 11856.21865142450 *T)
    sumL2  +=       0.085 * cos(0.50956043858  + 10575.40668294180 *T)
    sumL2  +=       0.067 * cos(3.62043033405  + 16496.36139620240 *T)
    sumL2  +=       0.055 * cos(5.24637517308  + 3340.61242669980 *T)
    sumL2  +=       0.048 * cos(5.43966777314  + 20426.57109242200 *T)
    sumL2  +=       0.064 * cos(5.79535817813  + 2388.89402044920 *T)
    sumL2  +=       0.046 * cos(5.43499966519  + 6275.96230299060 *T)
    sumL2  +=       0.050 * cos(3.86263598617  + 5729.50644714900 *T)
    sumL2  +=       0.044 * cos(1.52269529228  + 12168.00269657460 *T)
    sumL2  +=       0.057 * cos(4.96352373486  + 14945.31617355440 *T)
    sumL2  +=       0.045 * cos(1.00861230160  + 8635.94200376320 *T)
    sumL2  +=       0.043 * cos(3.30685683359  + 9779.10867612540 *T)
    sumL2  +=       0.042 * cos(0.63481258930  + 2699.73481931760 *T)
    sumL2  +=       0.041 * cos(5.67996766641  + 11712.95531823080 *T)
    sumL2  +=       0.056 * cos(4.34024451468  + 90955.55169449610 *T)
    sumL2  +=       0.041 * cos(5.81722212845  + 709.93304855830 *T)
    sumL2  +=       0.053 * cos(6.17052087143  + 233141.31440436147 *T)
    sumL2  +=       0.037 * cos(3.12495025087  + 16200.77272450120 *T)
    sumL2  +=       0.035 * cos(5.76973458495  + 12569.67481833180 *T)
    sumL2  +=       0.037 * cos(0.31656444326  + 24356.78078864160 *T)
    sumL2  +=       0.035 * cos(0.96229051027  + 17298.18232732620 *T)
    sumL2  +=       0.033 * cos(5.23130355867  + 5331.35744374080 *T)
    sumL2  +=       0.035 * cos(0.62517020593  + 25158.60171976540 *T)
    sumL2  +=       0.035 * cos(0.80004512129  + 13916.01910964160 *T)
    sumL2  +=       0.037 * cos(2.89336088688  + 12721.57209941700 *T)
    sumL2  +=       0.030 * cos(4.50198402401  + 23543.23050468179 *T)
    sumL2  +=       0.030 * cos(5.31355708693  + 18319.53658487960 *T)
    sumL2  +=       0.029 * cos(3.47275229977  + 13119.72110282519 *T)
    sumL2  +=       0.029 * cos(3.11002782516  + 4136.91043351620 *T)
    sumL2  +=       0.032 * cos(5.52273255667  + 5753.38488489680 *T)
    sumL2  +=       0.035 * cos(3.79699996680  + 143571.32428481648 *T)
    sumL2  +=       0.026 * cos(1.50634201907  + 154717.60988768269 *T)
    sumL2  +=       0.030 * cos(3.53519084118  + 6284.05617105960 *T)
    sumL2  +=       0.023 * cos(4.41808025967  + 5884.92684658320 *T)
    sumL2  +=       0.025 * cos(1.38477355808  + 65147.61976813770 *T)
    sumL2  +=       0.023 * cos(3.49782549797  + 7477.52286021600 *T)
    sumL2  +=       0.019 * cos(3.14329413716  + 6496.37494542940 *T)
    sumL2  +=       0.019 * cos(2.20135125199  + 18073.70493865020 *T)
    sumL2  +=       0.019 * cos(4.95020255309  + 3930.20969621960 *T)
    sumL2  +=       0.019 * cos(0.57998702747  + 31415.37924995700 *T)
    sumL2  +=       0.021 * cos(1.75474323399  + 12139.55350910680 *T)
    sumL2  +=       0.019 * cos(3.92233070499  + 19651.04848109800 *T)
    sumL2  +=       0.014 * cos(0.98131213224  + 12559.03815298200 *T)
    sumL2  +=       0.019 * cos(4.93309333729  + 2942.46342329160 *T)
    sumL2  +=       0.016 * cos(5.55997534558  + 8827.39026987480 *T)
    sumL2  +=       0.013 * cos(1.68808165516  + 4535.05943692440 *T)
    sumL2  +=       0.013 * cos(0.33982116161  + 4933.20844033260 *T)
    sumL2  +=       0.012 * cos(1.85426309994  + 5856.47765911540 *T)
    sumL2  +=       0.010 * cos(4.82763996845  + 13095.84266507740 *T)
    sumL2  +=       0.011 * cos(5.38005490571  + 11790.62908865880 *T)
    sumL2  +=       0.010 * cos(1.40815507226  + 10988.80815753500 *T)
    sumL2  +=       0.011 * cos(3.05005267431  + 17260.15465469040 *T)
    sumL2  +=       0.010 * cos(4.93364992366  + 12352.85260454480 *T)

    # EARTH Longitude
    sumL3  = 0
    sumL3  +=     289.226 * cos(5.84384198723  + 6283.07584999140 *T)
    sumL3  +=      34.955 * cos(0.00000000000  + 0.00000000000 *T)
    sumL3  +=      16.819 * cos(5.48766912348  + 12566.15169998280 *T)
    sumL3  +=       2.962 * cos(5.19577265202  + 155.42039943420 *T)
    sumL3  +=       1.288 * cos(4.72200252235  + 3.52311834900 *T)
    sumL3  +=       0.635 * cos(5.96925937141  + 242.72860397400 *T)
    sumL3  +=       0.714 * cos(5.30045809128  + 18849.22754997420 *T)
    sumL3  +=       0.402 * cos(3.78682982419  + 553.56940284240 *T)
    sumL3  +=       0.072 * cos(4.29768126180  + 6286.59896834040 *T)
    sumL3  +=       0.067 * cos(0.90721687647  + 6127.65545055720 *T)
    sumL3  +=       0.036 * cos(5.24029648014  + 6438.49624942560 *T)
    sumL3  +=       0.024 * cos(5.16003960716  + 25132.30339996560 *T)
    sumL3  +=       0.023 * cos(3.01921570335  + 6309.37416979120 *T)
    sumL3  +=       0.017 * cos(5.82863573502  + 6525.80445396540 *T)
    sumL3  +=       0.017 * cos(3.67772863930  + 71430.69561812909 *T)
    sumL3  +=       0.009 * cos(4.58467294499  + 1577.34354244780 *T)
    sumL3  +=       0.008 * cos(1.40626662824  + 11856.21865142450 *T)
    sumL3  +=       0.008 * cos(5.07561257196  + 6256.77753019160 *T)
    sumL3  +=       0.007 * cos(2.82473374405  + 83996.84731811189 *T)
    sumL3  +=       0.005 * cos(2.71488713339  + 10977.07880469900 *T)
    sumL3  +=       0.005 * cos(3.76879847273  + 12036.46073488820 *T)
    sumL3  +=       0.005 * cos(4.28412873331  + 6275.96230299060 *T)

    # EARTH Longitude
    sumL4  = 0
    sumL4  +=     114.084 * cos(3.14159265359  + 0.00000000000 *T)
    sumL4  +=       7.717 * cos(4.13446589358  + 6283.07584999140 *T)
    sumL4  +=       0.765 * cos(3.83803776214  + 12566.15169998280 *T)
    sumL4  +=       0.420 * cos(0.41925861858  + 155.42039943420 *T)
    sumL4  +=       0.040 * cos(3.59847585840  + 18849.22754997420 *T)
    sumL4  +=       0.041 * cos(3.14398414077  + 3.52311834900 *T)
    sumL4  +=       0.035 * cos(5.00298940826  + 5573.14280143310 *T)
    sumL4  +=       0.013 * cos(0.48794833701  + 77713.77146812050 *T)
    sumL4  +=       0.010 * cos(5.64801766350  + 6127.65545055720 *T)
    sumL4  +=       0.008 * cos(2.84160570605  + 161000.68573767410 *T)
    sumL4  +=       0.002 * cos(0.54912904658  + 6438.49624942560 *T)

    # EARTH Longitude
    sumL5  = 0
    sumL5  +=       0.878 * cos(3.14159265359  + 0.00000000000 *T)
    sumL5  +=       0.172 * cos(2.76579069510  + 6283.07584999140 *T)
    sumL5  +=       0.050 * cos(2.01353298182  + 155.42039943420 *T)
    sumL5  +=       0.028 * cos(2.21496423926  + 12566.15169998280 *T)
    sumL5  +=       0.005 * cos(1.75600058765  + 18849.22754997420 *T)

    # EARTH Latitude
    sumB0  = 0
    sumB0  +=     279.620 * cos(3.19870156017  + 84334.66158130829 *T)
    sumB0  +=     101.643 * cos(5.42248619256  + 5507.55323866740 *T)
    sumB0  +=      80.445 * cos(3.88013204458  + 5223.69391980220 *T)
    sumB0  +=      43.806 * cos(3.70444689758  + 2352.86615377180 *T)
    sumB0  +=      31.933 * cos(4.00026369781  + 1577.34354244780 *T)
    sumB0  +=      22.724 * cos(3.98473831560  + 1047.74731175470 *T)
    sumB0  +=      16.392 * cos(3.56456119782  + 5856.47765911540 *T)
    sumB0  +=      18.141 * cos(4.98367470263  + 6283.07584999140 *T)
    sumB0  +=      14.443 * cos(3.70275614914  + 9437.76293488700 *T)
    sumB0  +=      14.304 * cos(3.41117857525  + 10213.28554621100 *T)
    sumB0  +=      11.246 * cos(4.82820690530  + 14143.49524243060 *T)
    sumB0  +=      10.900 * cos(2.08574562327  + 6812.76681508600 *T)
    sumB0  +=       9.714 * cos(3.47303947752  + 4694.00295470760 *T)
    sumB0  +=      10.367 * cos(4.05663927946  + 71092.88135493269 *T)
    sumB0  +=       8.775 * cos(4.44016515669  + 5753.38488489680 *T)
    sumB0  +=       8.366 * cos(4.99251512180  + 7084.89678111520 *T)
    sumB0  +=       6.921 * cos(4.32559054073  + 6275.96230299060 *T)
    sumB0  +=       9.145 * cos(1.14182646613  + 6620.89011318780 *T)
    sumB0  +=       7.194 * cos(3.60193205752  + 529.69096509460 *T)
    sumB0  +=       7.698 * cos(5.55425745881  + 167621.57585086187 *T)
    sumB0  +=       5.285 * cos(2.48446991566  + 4705.73230754360 *T)
    sumB0  +=       5.208 * cos(6.24992674537  + 18073.70493865020 *T)
    sumB0  +=       4.529 * cos(2.33827747356  + 6309.37416979120 *T)
    sumB0  +=       5.579 * cos(4.41023653738  + 7860.41939243920 *T)
    sumB0  +=       4.743 * cos(0.70995680136  + 5884.92684658320 *T)
    sumB0  +=       4.301 * cos(1.10255777773  + 6681.22485339960 *T)
    sumB0  +=       3.849 * cos(1.82229412531  + 5486.77784317500 *T)
    sumB0  +=       4.093 * cos(5.11700141207  + 13367.97263110660 *T)
    sumB0  +=       3.681 * cos(0.43793170356  + 3154.68708489560 *T)
    sumB0  +=       3.420 * cos(5.42034800952  + 6069.77675455340 *T)
    sumB0  +=       3.617 * cos(6.04641937526  + 3930.20969621960 *T)
    sumB0  +=       3.670 * cos(4.58210192227  + 12194.03291462090 *T)
    sumB0  +=       2.918 * cos(1.95463881126  + 10977.07880469900 *T)
    sumB0  +=       2.797 * cos(5.61259275048  + 11790.62908865880 *T)
    sumB0  +=       2.502 * cos(0.60499729367  + 6496.37494542940 *T)
    sumB0  +=       2.319 * cos(5.01648216014  + 1059.38193018920 *T)
    sumB0  +=       2.684 * cos(1.39470396488  + 22003.91463486980 *T)
    sumB0  +=       2.428 * cos(3.24183056052  + 78051.58573131690 *T)
    sumB0  +=       2.120 * cos(4.30691000285  + 5643.17856367740 *T)
    sumB0  +=       2.257 * cos(3.15557225618  + 90617.73743129970 *T)
    sumB0  +=       1.813 * cos(3.75574218285  + 3340.61242669980 *T)
    sumB0  +=       2.226 * cos(2.79699346659  + 12036.46073488820 *T)
    sumB0  +=       1.888 * cos(0.86991545823  + 8635.94200376320 *T)
    sumB0  +=       1.517 * cos(1.95852055701  + 398.14900340820 *T)
    sumB0  +=       1.581 * cos(3.19976230948  + 5088.62883976680 *T)
    sumB0  +=       1.421 * cos(6.25530883827  + 2544.31441988340 *T)
    sumB0  +=       1.595 * cos(0.25619915135  + 17298.18232732620 *T)
    sumB0  +=       1.391 * cos(4.69964175561  + 7058.59846131540 *T)
    sumB0  +=       1.478 * cos(2.81808207569  + 25934.12433108940 *T)
    sumB0  +=       1.481 * cos(3.65823554806  + 11506.76976979360 *T)
    sumB0  +=       1.693 * cos(4.95689385293  + 156475.29024799570 *T)
    sumB0  +=       1.183 * cos(1.29343061246  + 775.52261132400 *T)
    sumB0  +=       1.114 * cos(2.37889311846  + 3738.76143010800 *T)
    sumB0  +=       0.994 * cos(4.30088900425  + 9225.53927328300 *T)
    sumB0  +=       0.924 * cos(3.06451026812  + 4164.31198961300 *T)
    sumB0  +=       0.867 * cos(0.55606931068  + 8429.24126646660 *T)
    sumB0  +=       0.988 * cos(5.97286104208  + 7079.37385680780 *T)
    sumB0  +=       0.824 * cos(1.50984806173  + 10447.38783960440 *T)
    sumB0  +=       0.915 * cos(0.12635654592  + 11015.10647733480 *T)
    sumB0  +=       0.742 * cos(1.99159139281  + 26087.90314157420 *T)
    sumB0  +=       1.039 * cos(3.14159265359  + 0.00000000000 *T)
    sumB0  +=       0.850 * cos(4.24120016095  + 29864.33402730900 *T)
    sumB0  +=       0.755 * cos(2.89631873320  + 4732.03062734340 *T)
    sumB0  +=       0.714 * cos(1.37548118603  + 2146.16541647520 *T)
    sumB0  +=       0.708 * cos(1.91406542362  + 8031.09226305840 *T)
    sumB0  +=       0.746 * cos(0.57893808616  + 796.29800681640 *T)
    sumB0  +=       0.802 * cos(5.12339137230  + 2942.46342329160 *T)
    sumB0  +=       0.751 * cos(1.67479850166  + 21228.39202354580 *T)
    sumB0  +=       0.602 * cos(4.09976538826  + 64809.80550494129 *T)
    sumB0  +=       0.594 * cos(3.49580704962  + 16496.36139620240 *T)
    sumB0  +=       0.592 * cos(4.59481504319  + 4690.47983635860 *T)
    sumB0  +=       0.530 * cos(5.73979295200  + 8827.39026987480 *T)
    sumB0  +=       0.503 * cos(5.66433137112  + 33794.54372352860 *T)
    sumB0  +=       0.483 * cos(1.57106522411  + 801.82093112380 *T)
    sumB0  +=       0.438 * cos(0.06707733767  + 3128.38876509580 *T)
    sumB0  +=       0.423 * cos(2.86944595927  + 12566.15169998280 *T)
    sumB0  +=       0.504 * cos(3.26207669160  + 7632.94325965020 *T)
    sumB0  +=       0.552 * cos(1.02926440457  + 239762.20451754928 *T)
    sumB0  +=       0.427 * cos(3.67434378210  + 213.29909543800 *T)
    sumB0  +=       0.404 * cos(1.46193297142  + 15720.83878487840 *T)
    sumB0  +=       0.503 * cos(4.85802444134  + 6290.18939699220 *T)
    sumB0  +=       0.417 * cos(0.81920713533  + 5216.58037280140 *T)
    sumB0  +=       0.365 * cos(0.01002966162  + 12168.00269657460 *T)
    sumB0  +=       0.363 * cos(1.28376436579  + 6206.80977871580 *T)
    sumB0  +=       0.353 * cos(4.70059133110  + 7234.79425624200 *T)
    sumB0  +=       0.415 * cos(0.96862624175  + 4136.91043351620 *T)
    sumB0  +=       0.387 * cos(3.09145061418  + 25158.60171976540 *T)
    sumB0  +=       0.373 * cos(2.65119262792  + 7342.45778018060 *T)
    sumB0  +=       0.361 * cos(2.97762937739  + 9623.68827669120 *T)
    sumB0  +=       0.418 * cos(3.75759994446  + 5230.80746680300 *T)
    sumB0  +=       0.396 * cos(1.22507712354  + 6438.49624942560 *T)
    sumB0  +=       0.322 * cos(1.21162178805  + 8662.24032356300 *T)
    sumB0  +=       0.284 * cos(5.64170320068  + 1589.07289528380 *T)
    sumB0  +=       0.379 * cos(1.72248432748  + 14945.31617355440 *T)
    sumB0  +=       0.320 * cos(3.94161159962  + 7330.82316174610 *T)
    sumB0  +=       0.313 * cos(5.47602376446  + 1194.44701022460 *T)
    sumB0  +=       0.292 * cos(1.38971327603  + 11769.85369316640 *T)
    sumB0  +=       0.305 * cos(0.80429352049  + 37724.75341974820 *T)
    sumB0  +=       0.257 * cos(5.81382809757  + 426.59819087600 *T)
    sumB0  +=       0.265 * cos(6.10358507671  + 6836.64525283380 *T)
    sumB0  +=       0.250 * cos(4.56452895547  + 7477.52286021600 *T)
    sumB0  +=       0.266 * cos(2.62926282354  + 7238.67559160000 *T)
    sumB0  +=       0.263 * cos(6.22089501237  + 6133.51265285680 *T)
    sumB0  +=       0.306 * cos(2.79682380531  + 1748.01641306700 *T)
    sumB0  +=       0.236 * cos(2.46093023714  + 11371.70468975820 *T)
    sumB0  +=       0.316 * cos(1.62662805006  + 250908.49012041549 *T)
    sumB0  +=       0.216 * cos(3.68721275185  + 5849.36411211460 *T)
    sumB0  +=       0.230 * cos(0.36165162947  + 5863.59120611620 *T)
    sumB0  +=       0.233 * cos(5.03509933858  + 20426.57109242200 *T)
    sumB0  +=       0.200 * cos(5.86073159059  + 4535.05943692440 *T)
    sumB0  +=       0.277 * cos(4.65400292395  + 82239.16695779889 *T)
    sumB0  +=       0.209 * cos(3.72323200804  + 10973.55568635000 *T)
    sumB0  +=       0.199 * cos(5.05186622555  + 5429.87946823940 *T)
    sumB0  +=       0.256 * cos(2.40923279770  + 19651.04848109800 *T)
    sumB0  +=       0.210 * cos(4.50691909144  + 29088.81141598500 *T)
    sumB0  +=       0.181 * cos(6.00294783127  + 4292.33083295040 *T)
    sumB0  +=       0.249 * cos(0.12900984422  + 154379.79562448629 *T)
    sumB0  +=       0.209 * cos(3.87759458598  + 17789.84561978500 *T)
    sumB0  +=       0.225 * cos(3.18339652605  + 18875.52586977400 *T)
    sumB0  +=       0.191 * cos(4.53897489299  + 18477.10876461230 *T)
    sumB0  +=       0.172 * cos(2.09694183014  + 13095.84266507740 *T)
    sumB0  +=       0.182 * cos(3.16107943500  + 16730.46368959580 *T)
    sumB0  +=       0.188 * cos(2.22746128596  + 41654.96311596780 *T)
    sumB0  +=       0.164 * cos(5.18686275017  + 5481.25491886760 *T)
    sumB0  +=       0.160 * cos(2.49298855159  + 12592.45001978260 *T)
    sumB0  +=       0.155 * cos(1.59595438230  + 10021.83728009940 *T)
    sumB0  +=       0.135 * cos(0.21349051064  + 10988.80815753500 *T)
    sumB0  +=       0.178 * cos(3.80375177970  + 23581.25817731760 *T)
    sumB0  +=       0.123 * cos(1.66800739151  + 15110.46611986620 *T)
    sumB0  +=       0.122 * cos(2.72678272244  + 18849.22754997420 *T)
    sumB0  +=       0.126 * cos(1.17675512910  + 14919.01785375460 *T)
    sumB0  +=       0.142 * cos(3.95053441332  + 337.81426319640 *T)
    sumB0  +=       0.116 * cos(6.06340906229  + 6709.67404086740 *T)
    sumB0  +=       0.137 * cos(3.52143246757  + 12139.55350910680 *T)
    sumB0  +=       0.136 * cos(2.92179113542  + 32217.20018108080 *T)
    sumB0  +=       0.110 * cos(3.51203379263  + 18052.92954315780 *T)
    sumB0  +=       0.147 * cos(4.63371971408  + 22805.73556599360 *T)
    sumB0  +=       0.108 * cos(5.45280814878  + 7.11354700080 *T)
    sumB0  +=       0.148 * cos(0.65447253687  + 95480.94718417450 *T)
    sumB0  +=       0.119 * cos(5.92110458985  + 33019.02111220460 *T)
    sumB0  +=       0.110 * cos(5.34824206306  + 639.89728631400 *T)
    sumB0  +=       0.106 * cos(3.71081682629  + 14314.16811304980 *T)
    sumB0  +=       0.139 * cos(6.17607198418  + 24356.78078864160 *T)
    sumB0  +=       0.118 * cos(5.59738712670  + 161338.50000087050 *T)
    sumB0  +=       0.117 * cos(3.65065271640  + 45585.17281218740 *T)
    sumB0  +=       0.127 * cos(4.74596574209  + 49515.38250840700 *T)
    sumB0  +=       0.120 * cos(1.04211499785  + 6915.85958930460 *T)
    sumB0  +=       0.120 * cos(5.60638811846  + 5650.29211067820 *T)
    sumB0  +=       0.115 * cos(3.10668213289  + 14712.31711645800 *T)
    sumB0  +=       0.099 * cos(0.69018940049  + 12779.45079542080 *T)
    sumB0  +=       0.097 * cos(1.07908724794  + 9917.69687450980 *T)
    sumB0  +=       0.093 * cos(2.62295197319  + 17260.15465469040 *T)
    sumB0  +=       0.099 * cos(4.45774681732  + 4933.20844033260 *T)
    sumB0  +=       0.123 * cos(1.37488922089  + 28286.99048486120 *T)
    sumB0  +=       0.121 * cos(5.19767249813  + 27511.46787353720 *T)
    sumB0  +=       0.105 * cos(0.87192267806  + 77375.95720492408 *T)
    sumB0  +=       0.087 * cos(3.93637812950  + 17654.78053974960 *T)
    sumB0  +=       0.122 * cos(2.23956068680  + 83997.09113559539 *T)
    sumB0  +=       0.087 * cos(4.18201600952  + 22779.43724619380 *T)
    sumB0  +=       0.104 * cos(4.59580877295  + 1349.86740965880 *T)
    sumB0  +=       0.102 * cos(2.83545248411  + 12352.85260454480 *T)
    sumB0  +=       0.102 * cos(3.97386522171  + 10818.13528691580 *T)
    sumB0  +=       0.101 * cos(4.32892825857  + 36147.40987730040 *T)
    sumB0  +=       0.094 * cos(5.00001709261  + 150192.21439800429 *T)
    sumB0  +=       0.077 * cos(3.97199369296  + 1592.59601363280 *T)
    sumB0  +=       0.100 * cos(6.07733097102  + 26735.94526221320 *T)
    sumB0  +=       0.086 * cos(5.26029638250  + 28313.28880466100 *T)
    sumB0  +=       0.093 * cos(4.31900620254  + 44809.65020086340 *T)
    sumB0  +=       0.076 * cos(6.22743405935  + 13521.75144159140 *T)
    sumB0  +=       0.072 * cos(1.55820597747  + 6256.77753019160 *T)
    sumB0  +=       0.082 * cos(4.95202664555  + 10575.40668294180 *T)
    sumB0  +=       0.082 * cos(1.69647647075  + 1990.74501704100 *T)
    sumB0  +=       0.075 * cos(2.29836095644  + 3634.62102451840 *T)
    sumB0  +=       0.075 * cos(2.66367876557  + 16200.77272450120 *T)
    sumB0  +=       0.087 * cos(0.26630214764  + 31441.67756975680 *T)
    sumB0  +=       0.077 * cos(2.25530954137  + 5235.32853823670 *T)
    sumB0  +=       0.076 * cos(1.09869730846  + 12903.96596317920 *T)
    sumB0  +=       0.058 * cos(4.28246138307  + 12559.03815298200 *T)
    sumB0  +=       0.064 * cos(5.51112830114  + 173904.65170085328 *T)
    sumB0  +=       0.056 * cos(2.60133794851  + 73188.37597844210 *T)
    sumB0  +=       0.055 * cos(5.81483150022  + 143233.51002162008 *T)
    sumB0  +=       0.054 * cos(3.38482031504  + 323049.11878710287 *T)
    sumB0  +=       0.039 * cos(3.28500401343  + 71768.50988132549 *T)
    sumB0  +=       0.039 * cos(3.11239910690  + 96900.81328129108 *T)

    # EARTH Latitude
    sumB1  = 0
    sumB1  +=       9.030 * cos(3.89729061890  + 5507.55323866740 *T)
    sumB1  +=       6.177 * cos(1.73038850355  + 5223.69391980220 *T)
    sumB1  +=       3.800 * cos(5.24404145734  + 2352.86615377180 *T)
    sumB1  +=       2.834 * cos(2.47345037450  + 1577.34354244780 *T)
    sumB1  +=       1.817 * cos(0.41874743765  + 6283.07584999140 *T)
    sumB1  +=       1.499 * cos(1.83320979291  + 5856.47765911540 *T)
    sumB1  +=       1.466 * cos(5.69401926017  + 5753.38488489680 *T)
    sumB1  +=       1.301 * cos(2.18890066314  + 9437.76293488700 *T)
    sumB1  +=       1.233 * cos(4.95222451476  + 10213.28554621100 *T)
    sumB1  +=       1.021 * cos(0.12866660208  + 7860.41939243920 *T)
    sumB1  +=       0.982 * cos(0.09005453285  + 14143.49524243060 *T)
    sumB1  +=       0.865 * cos(1.73949953555  + 3930.20969621960 *T)
    sumB1  +=       0.581 * cos(2.26949174067  + 5884.92684658320 *T)
    sumB1  +=       0.524 * cos(5.65662503159  + 529.69096509460 *T)
    sumB1  +=       0.473 * cos(6.22750969242  + 6309.37416979120 *T)
    sumB1  +=       0.451 * cos(1.53288619213  + 18073.70493865020 *T)
    sumB1  +=       0.364 * cos(3.61614477374  + 13367.97263110660 *T)
    sumB1  +=       0.372 * cos(3.22470721320  + 6275.96230299060 *T)
    sumB1  +=       0.268 * cos(2.34341267879  + 11790.62908865880 *T)
    sumB1  +=       0.322 * cos(0.94084045832  + 6069.77675455340 *T)
    sumB1  +=       0.232 * cos(0.26781182579  + 7058.59846131540 *T)
    sumB1  +=       0.216 * cos(6.05952221329  + 10977.07880469900 *T)
    sumB1  +=       0.232 * cos(2.93325646109  + 22003.91463486980 *T)
    sumB1  +=       0.204 * cos(3.86264841382  + 6496.37494542940 *T)
    sumB1  +=       0.202 * cos(2.81892511133  + 15720.83878487840 *T)
    sumB1  +=       0.185 * cos(4.93512381859  + 12036.46073488820 *T)
    sumB1  +=       0.220 * cos(3.99305643742  + 6812.76681508600 *T)
    sumB1  +=       0.166 * cos(1.74970002999  + 11506.76976979360 *T)
    sumB1  +=       0.212 * cos(1.57166285369  + 4694.00295470760 *T)
    sumB1  +=       0.157 * cos(1.08259734788  + 5643.17856367740 *T)
    sumB1  +=       0.154 * cos(5.99434678412  + 5486.77784317500 *T)
    sumB1  +=       0.144 * cos(5.23285656085  + 78051.58573131690 *T)
    sumB1  +=       0.144 * cos(1.16454655948  + 90617.73743129970 *T)
    sumB1  +=       0.137 * cos(2.67760436027  + 6290.18939699220 *T)
    sumB1  +=       0.180 * cos(2.06509026215  + 7084.89678111520 *T)
    sumB1  +=       0.121 * cos(5.90212574947  + 9225.53927328300 *T)
    sumB1  +=       0.150 * cos(2.00175038718  + 5230.80746680300 *T)
    sumB1  +=       0.149 * cos(5.06157254516  + 17298.18232732620 *T)
    sumB1  +=       0.118 * cos(5.39979058038  + 3340.61242669980 *T)
    sumB1  +=       0.161 * cos(3.32421999691  + 6283.31966747490 *T)
    sumB1  +=       0.121 * cos(4.36722193162  + 19651.04848109800 *T)
    sumB1  +=       0.116 * cos(5.83462858507  + 4705.73230754360 *T)
    sumB1  +=       0.128 * cos(4.35489873365  + 25934.12433108940 *T)
    sumB1  +=       0.143 * cos(0.00000000000  + 0.00000000000 *T)
    sumB1  +=       0.109 * cos(2.52157834166  + 6438.49624942560 *T)
    sumB1  +=       0.099 * cos(2.70727488041  + 5216.58037280140 *T)
    sumB1  +=       0.103 * cos(0.93782340879  + 8827.39026987480 *T)
    sumB1  +=       0.082 * cos(4.29214680390  + 8635.94200376320 *T)
    sumB1  +=       0.079 * cos(2.24085737326  + 1059.38193018920 *T)
    sumB1  +=       0.097 * cos(5.50959692365  + 29864.33402730900 *T)
    sumB1  +=       0.072 * cos(0.21891639822  + 21228.39202354580 *T)
    sumB1  +=       0.071 * cos(2.86755026812  + 6681.22485339960 *T)
    sumB1  +=       0.074 * cos(2.20184828895  + 37724.75341974820 *T)
    sumB1  +=       0.063 * cos(4.45586625948  + 7079.37385680780 *T)
    sumB1  +=       0.061 * cos(0.63918772258  + 33794.54372352860 *T)
    sumB1  +=       0.047 * cos(2.09070235724  + 3128.38876509580 *T)
    sumB1  +=       0.047 * cos(3.32543843300  + 26087.90314157420 *T)
    sumB1  +=       0.049 * cos(1.60680905005  + 6702.56049386660 *T)
    sumB1  +=       0.057 * cos(0.11215813438  + 29088.81141598500 *T)
    sumB1  +=       0.056 * cos(5.47982934911  + 775.52261132400 *T)
    sumB1  +=       0.050 * cos(1.89396788463  + 12139.55350910680 *T)
    sumB1  +=       0.047 * cos(2.97214907240  + 20426.57109242200 *T)
    sumB1  +=       0.041 * cos(5.55329394890  + 11015.10647733480 *T)
    sumB1  +=       0.041 * cos(5.91861144924  + 23581.25817731760 *T)
    sumB1  +=       0.045 * cos(4.95273290181  + 5863.59120611620 *T)
    sumB1  +=       0.050 * cos(3.62740835096  + 41654.96311596780 *T)
    sumB1  +=       0.037 * cos(6.09033460601  + 64809.80550494129 *T)
    sumB1  +=       0.037 * cos(5.86153655431  + 12566.15169998280 *T)
    sumB1  +=       0.046 * cos(1.65798680284  + 25158.60171976540 *T)
    sumB1  +=       0.038 * cos(2.00673650251  + 426.59819087600 *T)
    sumB1  +=       0.036 * cos(6.24373396652  + 6283.14316029419 *T)
    sumB1  +=       0.036 * cos(0.40465162918  + 6283.00853968860 *T)
    sumB1  +=       0.032 * cos(6.03707103538  + 2942.46342329160 *T)
    sumB1  +=       0.041 * cos(4.86809570283  + 1592.59601363280 *T)
    sumB1  +=       0.028 * cos(4.38359423735  + 7632.94325965020 *T)
    sumB1  +=       0.028 * cos(6.03334294232  + 17789.84561978500 *T)
    sumB1  +=       0.026 * cos(3.88971333608  + 5331.35744374080 *T)
    sumB1  +=       0.026 * cos(5.94932724051  + 16496.36139620240 *T)
    sumB1  +=       0.031 * cos(1.44666331503  + 16730.46368959580 *T)
    sumB1  +=       0.026 * cos(6.26376705837  + 23543.23050468179 *T)
    sumB1  +=       0.033 * cos(0.93797239147  + 213.29909543800 *T)
    sumB1  +=       0.026 * cos(3.71858432944  + 13095.84266507740 *T)
    sumB1  +=       0.027 * cos(0.60565274405  + 10988.80815753500 *T)
    sumB1  +=       0.023 * cos(4.44388985550  + 18849.22754997420 *T)
    sumB1  +=       0.028 * cos(1.53862289477  + 6279.48542133960 *T)
    sumB1  +=       0.028 * cos(1.96831814872  + 6286.66627864320 *T)
    sumB1  +=       0.028 * cos(5.78094918529  + 15110.46611986620 *T)
    sumB1  +=       0.026 * cos(2.48165809843  + 5729.50644714900 *T)
    sumB1  +=       0.020 * cos(3.85655029499  + 9623.68827669120 *T)
    sumB1  +=       0.021 * cos(5.83006047147  + 7234.79425624200 *T)
    sumB1  +=       0.021 * cos(0.69628570421  + 398.14900340820 *T)
    sumB1  +=       0.022 * cos(5.02222806555  + 6127.65545055720 *T)
    sumB1  +=       0.020 * cos(3.47611265290  + 6148.01076995600 *T)
    sumB1  +=       0.020 * cos(0.90769829044  + 5481.25491886760 *T)
    sumB1  +=       0.020 * cos(0.03081589303  + 6418.14093002680 *T)
    sumB1  +=       0.020 * cos(3.74220084927  + 1589.07289528380 *T)
    sumB1  +=       0.021 * cos(4.00149269576  + 3154.68708489560 *T)
    sumB1  +=       0.018 * cos(1.58348238359  + 2118.76386037840 *T)
    sumB1  +=       0.019 * cos(0.85407021371  + 14712.31711645800 *T)

    # EARTH Latitude
    sumB2  = 0
    sumB2  +=       1.662 * cos(1.62703209173  + 84334.66158130829 *T)
    sumB2  +=       0.492 * cos(2.41382223971  + 1047.74731175470 *T)
    sumB2  +=       0.344 * cos(2.24353004539  + 5507.55323866740 *T)
    sumB2  +=       0.258 * cos(6.00906896311  + 5223.69391980220 *T)
    sumB2  +=       0.131 * cos(0.95447345240  + 6283.07584999140 *T)
    sumB2  +=       0.086 * cos(1.67530247303  + 7860.41939243920 *T)
    sumB2  +=       0.090 * cos(0.97606804452  + 1577.34354244780 *T)
    sumB2  +=       0.090 * cos(0.37899871725  + 2352.86615377180 *T)
    sumB2  +=       0.089 * cos(6.25807507963  + 10213.28554621100 *T)
    sumB2  +=       0.075 * cos(0.84213523741  + 167621.57585086187 *T)
    sumB2  +=       0.052 * cos(1.70501566089  + 14143.49524243060 *T)
    sumB2  +=       0.057 * cos(6.15295833679  + 12194.03291462090 *T)
    sumB2  +=       0.051 * cos(1.27616016740  + 5753.38488489680 *T)
    sumB2  +=       0.051 * cos(5.37229738682  + 6812.76681508600 *T)
    sumB2  +=       0.034 * cos(1.73672994279  + 7058.59846131540 *T)
    sumB2  +=       0.038 * cos(2.77761031485  + 10988.80815753500 *T)
    sumB2  +=       0.046 * cos(3.38617099014  + 156475.29024799570 *T)
    sumB2  +=       0.021 * cos(1.95248349228  + 8827.39026987480 *T)
    sumB2  +=       0.018 * cos(3.33419222028  + 8429.24126646660 *T)
    sumB2  +=       0.019 * cos(4.32945160287  + 17789.84561978500 *T)
    sumB2  +=       0.017 * cos(0.66191210656  + 6283.00853968860 *T)
    sumB2  +=       0.018 * cos(3.74885333072  + 11769.85369316640 *T)
    sumB2  +=       0.017 * cos(4.23058370776  + 10977.07880469900 *T)
    sumB2  +=       0.017 * cos(1.78116162721  + 5486.77784317500 *T)
    sumB2  +=       0.021 * cos(1.36972913918  + 12036.46073488820 *T)
    sumB2  +=       0.017 * cos(2.79601092529  + 796.29800681640 *T)
    sumB2  +=       0.015 * cos(0.43087848850  + 11790.62908865880 *T)
    sumB2  +=       0.017 * cos(1.35132152761  + 78051.58573131690 *T)
    sumB2  +=       0.015 * cos(1.17032155085  + 213.29909543800 *T)
    sumB2  +=       0.018 * cos(2.85221514199  + 5088.62883976680 *T)
    sumB2  +=       0.017 * cos(0.21780913672  + 6283.14316029419 *T)
    sumB2  +=       0.013 * cos(1.21201504386  + 25132.30339996560 *T)
    sumB2  +=       0.012 * cos(1.12953712197  + 90617.73743129970 *T)
    sumB2  +=       0.012 * cos(5.13714452592  + 7079.37385680780 *T)
    sumB2  +=       0.013 * cos(3.79842135217  + 4933.20844033260 *T)
    sumB2  +=       0.012 * cos(4.89407978213  + 3738.76143010800 *T)
    sumB2  +=       0.015 * cos(6.05682328852  + 398.14900340820 *T)
    sumB2  +=       0.014 * cos(4.81029291856  + 4694.00295470760 *T)
    sumB2  +=       0.011 * cos(0.61684523405  + 3128.38876509580 *T)
    sumB2  +=       0.011 * cos(5.32876538500  + 6040.34724601740 *T)
    sumB2  +=       0.014 * cos(5.27227350286  + 4535.05943692440 *T)
    sumB2  +=       0.011 * cos(2.39292099451  + 5331.35744374080 *T)
    sumB2  +=       0.010 * cos(4.45296532710  + 6525.80445396540 *T)
    sumB2  +=       0.014 * cos(4.66400985037  + 8031.09226305840 *T)
    sumB2  +=       0.010 * cos(3.22472385926  + 9437.76293488700 *T)
    sumB2  +=       0.011 * cos(3.80913404437  + 801.82093112380 *T)
    sumB2  +=       0.010 * cos(5.15032130575  + 11371.70468975820 *T)
    sumB2  +=       0.013 * cos(0.98720797401  + 5729.50644714900 *T)
    sumB2  +=       0.009 * cos(5.94191743597  + 7632.94325965020 *T)

    # EARTH Latitude
    sumB3  = 0
    sumB3  +=       0.011 * cos(0.23877262399  + 7860.41939243920 *T)
    sumB3  +=       0.009 * cos(1.16069982609  + 5507.55323866740 *T)
    sumB3  +=       0.008 * cos(1.65357552925  + 5884.92684658320 *T)
    sumB3  +=       0.008 * cos(2.86720038197  + 7058.59846131540 *T)
    sumB3  +=       0.007 * cos(3.04818741666  + 5486.77784317500 *T)
    sumB3  +=       0.007 * cos(2.59437103785  + 529.69096509460 *T)
    sumB3  +=       0.008 * cos(4.02863090524  + 6256.77753019160 *T)
    sumB3  +=       0.008 * cos(2.42003508927  + 5753.38488489680 *T)
    sumB3  +=       0.006 * cos(0.84181087594  + 6275.96230299060 *T)
    sumB3  +=       0.006 * cos(5.40160929468  + 1577.34354244780 *T)
    sumB3  +=       0.007 * cos(2.73399865247  + 6309.37416979120 *T)

    # EARTH Latitude
    sumB4  = 0
    sumB4  +=       0.004 * cos(0.79662198849  + 6438.49624942560 *T)
    sumB4  +=       0.005 * cos(0.84308705203  + 1047.74731175470 *T)
    sumB4  +=       0.005 * cos(0.05711572303  + 84334.66158130829 *T)
    sumB4  +=       0.003 * cos(3.46779895686  + 6279.55273164240 *T)
    sumB4  +=       0.003 * cos(2.89822201212  + 6127.65545055720 *T)

    # EARTH Radius
    sumR0  = 0
    sumR0  +=  100013988.799 * cos(0.00000000000  + 0.00000000000 *T)
    sumR0  +=  1670699.626 * cos(3.09846350771  + 6283.07584999140 *T)
    sumR0  +=   13956.023 * cos(3.05524609620  + 12566.15169998280 *T)
    sumR0  +=    3083.720 * cos(5.19846674381  + 77713.77146812050 *T)
    sumR0  +=    1628.461 * cos(1.17387749012  + 5753.38488489680 *T)
    sumR0  +=    1575.568 * cos(2.84685245825  + 7860.41939243920 *T)
    sumR0  +=     924.799 * cos(5.45292234084  + 11506.76976979360 *T)
    sumR0  +=     542.444 * cos(4.56409149777  + 3930.20969621960 *T)
    sumR0  +=     472.110 * cos(3.66100022149  + 5884.92684658320 *T)
    sumR0  +=     328.780 * cos(5.89983646482  + 5223.69391980220 *T)
    sumR0  +=     345.983 * cos(0.96368617687  + 5507.55323866740 *T)
    sumR0  +=     306.784 * cos(0.29867139512  + 5573.14280143310 *T)
    sumR0  +=     174.844 * cos(3.01193636534  + 18849.22754997420 *T)
    sumR0  +=     243.189 * cos(4.27349536153  + 11790.62908865880 *T)
    sumR0  +=     211.829 * cos(5.84714540314  + 1577.34354244780 *T)
    sumR0  +=     185.752 * cos(5.02194447178  + 10977.07880469900 *T)
    sumR0  +=     109.835 * cos(5.05510636285  + 5486.77784317500 *T)
    sumR0  +=      98.316 * cos(0.88681311277  + 6069.77675455340 *T)
    sumR0  +=      86.499 * cos(5.68959778254  + 15720.83878487840 *T)
    sumR0  +=      85.825 * cos(1.27083733351  + 161000.68573767410 *T)
    sumR0  +=      62.916 * cos(0.92177108832  + 529.69096509460 *T)
    sumR0  +=      57.056 * cos(2.01374292014  + 83996.84731811189 *T)
    sumR0  +=      64.903 * cos(0.27250613787  + 17260.15465469040 *T)
    sumR0  +=      49.384 * cos(3.24501240359  + 2544.31441988340 *T)
    sumR0  +=      55.736 * cos(5.24159798933  + 71430.69561812909 *T)
    sumR0  +=      42.515 * cos(6.01110242003  + 6275.96230299060 *T)
    sumR0  +=      46.963 * cos(2.57805070386  + 775.52261132400 *T)
    sumR0  +=      38.968 * cos(5.36071738169  + 4694.00295470760 *T)
    sumR0  +=      44.661 * cos(5.53715807302  + 9437.76293488700 *T)
    sumR0  +=      35.660 * cos(1.67468058995  + 12036.46073488820 *T)
    sumR0  +=      31.921 * cos(0.18368229781  + 5088.62883976680 *T)
    sumR0  +=      31.846 * cos(1.77775642085  + 398.14900340820 *T)
    sumR0  +=      33.193 * cos(0.24370300098  + 7084.89678111520 *T)
    sumR0  +=      38.245 * cos(2.39255343974  + 8827.39026987480 *T)
    sumR0  +=      28.464 * cos(1.21344868176  + 6286.59896834040 *T)
    sumR0  +=      37.490 * cos(0.82952922332  + 19651.04848109800 *T)
    sumR0  +=      36.957 * cos(4.90107591914  + 12139.55350910680 *T)
    sumR0  +=      34.537 * cos(1.84270693282  + 2942.46342329160 *T)
    sumR0  +=      26.275 * cos(4.58896850401  + 10447.38783960440 *T)
    sumR0  +=      24.596 * cos(3.78660875483  + 8429.24126646660 *T)
    sumR0  +=      23.587 * cos(0.26866117066  + 796.29800681640 *T)
    sumR0  +=      27.793 * cos(1.89934330904  + 6279.55273164240 *T)
    sumR0  +=      23.927 * cos(4.99598548138  + 5856.47765911540 *T)
    sumR0  +=      20.349 * cos(4.65267995431  + 2146.16541647520 *T)
    sumR0  +=      23.287 * cos(2.80783650928  + 14143.49524243060 *T)
    sumR0  +=      22.103 * cos(1.95004702988  + 3154.68708489560 *T)
    sumR0  +=      19.506 * cos(5.38227371393  + 2352.86615377180 *T)
    sumR0  +=      17.958 * cos(0.19871379385  + 6812.76681508600 *T)
    sumR0  +=      17.174 * cos(4.43315560735  + 10213.28554621100 *T)
    sumR0  +=      16.190 * cos(5.23160507859  + 17789.84561978500 *T)
    sumR0  +=      17.314 * cos(6.15200787916  + 16730.46368959580 *T)
    sumR0  +=      13.814 * cos(5.18962074032  + 8031.09226305840 *T)
    sumR0  +=      18.833 * cos(0.67306674027  + 149854.40013480789 *T)
    sumR0  +=      18.331 * cos(2.25348733734  + 23581.25817731760 *T)
    sumR0  +=      13.641 * cos(3.68516118804  + 4705.73230754360 *T)
    sumR0  +=      13.139 * cos(0.65289581324  + 13367.97263110660 *T)
    sumR0  +=      10.414 * cos(4.33285688538  + 11769.85369316640 *T)
    sumR0  +=       9.978 * cos(4.20126336355  + 6309.37416979120 *T)
    sumR0  +=      10.169 * cos(1.59390681369  + 4690.47983635860 *T)
    sumR0  +=       7.564 * cos(2.62560597390  + 6256.77753019160 *T)
    sumR0  +=       9.661 * cos(3.67586791220  + 27511.46787353720 *T)
    sumR0  +=       6.743 * cos(0.56270332741  + 3340.61242669980 *T)
    sumR0  +=       8.743 * cos(6.06359123461  + 1748.01641306700 *T)
    sumR0  +=       7.786 * cos(3.67371235637  + 12168.00269657460 *T)
    sumR0  +=       6.633 * cos(5.66149277792  + 11371.70468975820 *T)
    sumR0  +=       7.712 * cos(0.31242577789  + 7632.94325965020 *T)
    sumR0  +=       6.592 * cos(3.13576266188  + 801.82093112380 *T)
    sumR0  +=       7.460 * cos(5.64757188143  + 11926.25441366880 *T)
    sumR0  +=       6.933 * cos(2.92384586400  + 6681.22485339960 *T)
    sumR0  +=       6.802 * cos(1.42329806420  + 23013.53953958720 *T)
    sumR0  +=       6.115 * cos(5.13393615454  + 1194.44701022460 *T)
    sumR0  +=       6.477 * cos(2.64986648492  + 19804.82729158280 *T)
    sumR0  +=       5.233 * cos(4.62434053374  + 6438.49624942560 *T)
    sumR0  +=       6.147 * cos(3.02863936662  + 233141.31440436147 *T)
    sumR0  +=       4.608 * cos(1.72194702724  + 7234.79425624200 *T)
    sumR0  +=       4.221 * cos(1.55697533729  + 7238.67559160000 *T)
    sumR0  +=       5.314 * cos(2.40716580847  + 11499.65622279280 *T)
    sumR0  +=       5.128 * cos(5.32398965690  + 11513.88331679440 *T)
    sumR0  +=       4.770 * cos(0.25554312006  + 11856.21865142450 *T)
    sumR0  +=       5.519 * cos(2.09089154502  + 17298.18232732620 *T)
    sumR0  +=       5.625 * cos(4.34052903053  + 90955.55169449610 *T)
    sumR0  +=       4.578 * cos(4.46569641570  + 5746.27133789600 *T)
    sumR0  +=       3.788 * cos(4.90729383510  + 4164.31198961300 *T)
    sumR0  +=       5.337 * cos(5.09957905104  + 31441.67756975680 *T)
    sumR0  +=       3.967 * cos(1.20054555174  + 1349.86740965880 *T)
    sumR0  +=       4.008 * cos(3.03007204392  + 1059.38193018920 *T)
    sumR0  +=       3.476 * cos(0.76080277030  + 10973.55568635000 *T)
    sumR0  +=       4.232 * cos(1.05485713117  + 5760.49843189760 *T)
    sumR0  +=       4.582 * cos(3.76570026763  + 6386.16862421000 *T)
    sumR0  +=       3.335 * cos(3.13829943354  + 6836.64525283380 *T)
    sumR0  +=       3.418 * cos(3.00072390334  + 4292.33083295040 *T)
    sumR0  +=       3.598 * cos(5.70718084323  + 5643.17856367740 *T)
    sumR0  +=       3.237 * cos(4.16448773994  + 9917.69687450980 *T)
    sumR0  +=       4.154 * cos(2.59941292162  + 7058.59846131540 *T)
    sumR0  +=       3.362 * cos(4.54577697964  + 4732.03062734340 *T)
    sumR0  +=       2.978 * cos(1.30561268820  + 6283.14316029419 *T)
    sumR0  +=       2.765 * cos(0.51311975679  + 26.29831979980 *T)
    sumR0  +=       2.802 * cos(5.66263240521  + 8635.94200376320 *T)
    sumR0  +=       2.927 * cos(5.73787481548  + 16200.77272450120 *T)
    sumR0  +=       3.164 * cos(1.69140262657  + 11015.10647733480 *T)
    sumR0  +=       2.598 * cos(2.96244118586  + 25132.30339996560 *T)
    sumR0  +=       3.519 * cos(3.62639325753  + 244287.60000722768 *T)
    sumR0  +=       2.676 * cos(4.20725700850  + 18073.70493865020 *T)
    sumR0  +=       2.978 * cos(1.74971565805  + 6283.00853968860 *T)
    sumR0  +=       2.287 * cos(1.06975704977  + 14314.16811304980 *T)
    sumR0  +=       2.863 * cos(5.92838131397  + 14712.31711645800 *T)
    sumR0  +=       3.071 * cos(0.23793217002  + 35371.88726597640 *T)
    sumR0  +=       2.656 * cos(0.89959301780  + 12352.85260454480 *T)
    sumR0  +=       2.415 * cos(2.79975176257  + 709.93304855830 *T)
    sumR0  +=       2.814 * cos(3.51488206882  + 21228.39202354580 *T)
    sumR0  +=       1.977 * cos(2.61358297550  + 951.71840625060 *T)
    sumR0  +=       2.548 * cos(2.47684686575  + 6208.29425142410 *T)
    sumR0  +=       1.999 * cos(0.56090388160  + 7079.37385680780 *T)
    sumR0  +=       2.305 * cos(1.05376461628  + 22483.84857449259 *T)
    sumR0  +=       1.855 * cos(2.86090681163  + 5216.58037280140 *T)
    sumR0  +=       2.157 * cos(1.31396741861  + 154717.60988768269 *T)
    sumR0  +=       1.970 * cos(4.36929875289  + 167283.76158766547 *T)
    sumR0  +=       1.635 * cos(5.85571606764  + 10984.19235169980 *T)
    sumR0  +=       1.754 * cos(2.14452408833  + 6290.18939699220 *T)
    sumR0  +=       2.154 * cos(6.03828341543  + 10873.98603048040 *T)
    sumR0  +=       1.714 * cos(3.70157691113  + 1592.59601363280 *T)
    sumR0  +=       1.541 * cos(6.21598380732  + 23543.23050468179 *T)
    sumR0  +=       1.611 * cos(1.99824499377  + 10969.96525769820 *T)
    sumR0  +=       1.712 * cos(1.34295663542  + 3128.38876509580 *T)
    sumR0  +=       1.642 * cos(5.55026665339  + 6496.37494542940 *T)
    sumR0  +=       1.502 * cos(5.43948825854  + 155.42039943420 *T)
    sumR0  +=       1.827 * cos(5.91227480261  + 3738.76143010800 *T)
    sumR0  +=       1.726 * cos(2.16764983583  + 10575.40668294180 *T)
    sumR0  +=       1.532 * cos(5.35683107070  + 13521.75144159140 *T)
    sumR0  +=       1.829 * cos(1.66006148731  + 39302.09696219600 *T)
    sumR0  +=       1.605 * cos(1.90928637633  + 6133.51265285680 *T)
    sumR0  +=       1.282 * cos(2.46014880418  + 13916.01910964160 *T)
    sumR0  +=       1.211 * cos(4.41360631550  + 3894.18182954220 *T)
    sumR0  +=       1.394 * cos(1.77801929354  + 9225.53927328300 *T)
    sumR0  +=       1.571 * cos(4.95512957592  + 25158.60171976540 *T)
    sumR0  +=       1.205 * cos(1.19212540615  + 3.52311834900 *T)
    sumR0  +=       1.132 * cos(2.69830084955  + 6040.34724601740 *T)
    sumR0  +=       1.504 * cos(5.77002730341  + 18209.33026366019 *T)
    sumR0  +=       1.393 * cos(1.62621805428  + 5120.60114558360 *T)
    sumR0  +=       1.077 * cos(2.93931554233  + 17256.63153634140 *T)
    sumR0  +=       1.232 * cos(0.71655165307  + 143571.32428481648 *T)
    sumR0  +=       1.087 * cos(0.99769687939  + 955.59974160860 *T)
    sumR0  +=       1.068 * cos(5.28472576231  + 65147.61976813770 *T)
    sumR0  +=       0.980 * cos(5.10949204607  + 6172.86952877200 *T)
    sumR0  +=       1.169 * cos(3.11664290862  + 14945.31617355440 *T)
    sumR0  +=       1.202 * cos(4.02992510402  + 553.56940284240 *T)
    sumR0  +=       0.979 * cos(2.00000879212  + 15110.46611986620 *T)
    sumR0  +=       0.962 * cos(4.02380771400  + 6282.09552892320 *T)
    sumR0  +=       0.999 * cos(3.62643002790  + 6262.30045449900 *T)
    sumR0  +=       1.030 * cos(5.84989900289  + 213.29909543800 *T)
    sumR0  +=       1.014 * cos(2.84221578218  + 8662.24032356300 *T)
    sumR0  +=       1.185 * cos(1.51330541132  + 17654.78053974960 *T)
    sumR0  +=       0.967 * cos(2.67081017562  + 5650.29211067820 *T)
    sumR0  +=       1.222 * cos(2.65423784904  + 88860.05707098669 *T)
    sumR0  +=       0.981 * cos(2.36370360283  + 6206.80977871580 *T)
    sumR0  +=       1.033 * cos(0.13874927606  + 11712.95531823080 *T)
    sumR0  +=       1.103 * cos(3.08477302937  + 43232.30665841560 *T)
    sumR0  +=       0.781 * cos(2.53372735932  + 16496.36139620240 *T)
    sumR0  +=       1.019 * cos(3.04569392376  + 6037.24420376200 *T)
    sumR0  +=       0.795 * cos(5.80662989111  + 5230.80746680300 *T)
    sumR0  +=       0.813 * cos(3.57710279439  + 10177.25767953360 *T)
    sumR0  +=       0.962 * cos(5.31470594766  + 6284.05617105960 *T)
    sumR0  +=       0.721 * cos(5.96264301567  + 12559.03815298200 *T)
    sumR0  +=       0.966 * cos(2.74714939953  + 6244.94281435360 *T)
    sumR0  +=       0.921 * cos(0.10155275926  + 29088.81141598500 *T)
    sumR0  +=       0.692 * cos(3.89764447548  + 1589.07289528380 *T)
    sumR0  +=       0.719 * cos(5.91791450402  + 4136.91043351620 *T)
    sumR0  +=       0.772 * cos(4.05505682353  + 6127.65545055720 *T)
    sumR0  +=       0.712 * cos(5.49291532439  + 22003.91463486980 *T)
    sumR0  +=       0.672 * cos(1.60700490811  + 11087.28512591840 *T)
    sumR0  +=       0.690 * cos(4.50539825563  + 426.59819087600 *T)
    sumR0  +=       0.854 * cos(3.26104981596  + 20426.57109242200 *T)
    sumR0  +=       0.656 * cos(4.32410182940  + 16858.48253293320 *T)
    sumR0  +=       0.840 * cos(2.59572585222  + 28766.92442448400 *T)
    sumR0  +=       0.692 * cos(0.61650089011  + 11403.67699557500 *T)
    sumR0  +=       0.700 * cos(3.40901167143  + 7.11354700080 *T)
    sumR0  +=       0.726 * cos(0.04243053594  + 5481.25491886760 *T)
    sumR0  +=       0.557 * cos(4.78317696534  + 20199.09495963300 *T)
    sumR0  +=       0.649 * cos(1.04027912958  + 6062.66320755260 *T)
    sumR0  +=       0.633 * cos(5.70229959167  + 45892.73043315699 *T)
    sumR0  +=       0.592 * cos(6.11836729658  + 9623.68827669120 *T)
    sumR0  +=       0.523 * cos(3.62840021266  + 5333.90024102160 *T)
    sumR0  +=       0.604 * cos(5.57734696185  + 10344.29506538580 *T)
    sumR0  +=       0.496 * cos(2.21023499449  + 1990.74501704100 *T)
    sumR0  +=       0.691 * cos(1.96071732602  + 12416.58850284820 *T)
    sumR0  +=       0.640 * cos(1.59074172032  + 18319.53658487960 *T)
    sumR0  +=       0.625 * cos(3.82362791378  + 13517.87010623340 *T)
    sumR0  +=       0.663 * cos(5.08444996779  + 283.85931886520 *T)
    sumR0  +=       0.475 * cos(1.17025894287  + 12569.67481833180 *T)
    sumR0  +=       0.664 * cos(4.50029469969  + 47162.51635463520 *T)
    sumR0  +=       0.569 * cos(0.16310365162  + 17267.26820169119 *T)
    sumR0  +=       0.568 * cos(3.86100969474  + 6076.89030155420 *T)
    sumR0  +=       0.539 * cos(4.83282276086  + 18422.62935909819 *T)
    sumR0  +=       0.466 * cos(0.75872342878  + 7342.45778018060 *T)
    sumR0  +=       0.541 * cos(3.07212190507  + 226858.23855437007 *T)
    sumR0  +=       0.458 * cos(0.26774483096  + 4590.91018048900 *T)
    sumR0  +=       0.610 * cos(1.53597051291  + 33019.02111220460 *T)
    sumR0  +=       0.617 * cos(2.62356328726  + 11190.37790013700 *T)
    sumR0  +=       0.548 * cos(4.55798855791  + 18875.52586977400 *T)
    sumR0  +=       0.633 * cos(4.60110281228  + 66567.48586525429 *T)
    sumR0  +=       0.596 * cos(5.78202396722  + 632.78373931320 *T)
    sumR0  +=       0.533 * cos(5.01786882904  + 12132.43996210600 *T)
    sumR0  +=       0.603 * cos(5.38458554802  + 316428.22867391503 *T)
    sumR0  +=       0.469 * cos(0.59168241917  + 21954.15760939799 *T)
    sumR0  +=       0.548 * cos(3.50613163558  + 17253.04110768959 *T)
    sumR0  +=       0.502 * cos(0.98804327589  + 11609.86254401220 *T)
    sumR0  +=       0.568 * cos(1.98497313089  + 7668.63742494250 *T)
    sumR0  +=       0.482 * cos(1.62141803864  + 12146.66705610760 *T)
    sumR0  +=       0.391 * cos(3.68718382989  + 18052.92954315780 *T)
    sumR0  +=       0.457 * cos(3.77205737340  + 156137.47598479927 *T)
    sumR0  +=       0.401 * cos(5.28260651958  + 15671.08175940660 *T)
    sumR0  +=       0.469 * cos(1.80963184268  + 12562.62858163380 *T)
    sumR0  +=       0.508 * cos(3.36399024699  + 20597.24396304120 *T)
    sumR0  +=       0.450 * cos(5.66054299250  + 10454.50138660520 *T)
    sumR0  +=       0.375 * cos(4.98534633105  + 9779.10867612540 *T)
    sumR0  +=       0.523 * cos(0.97215560834  + 155427.54293624099 *T)
    sumR0  +=       0.403 * cos(5.13939866506  + 1551.04522264800 *T)
    sumR0  +=       0.372 * cos(3.69883738807  + 9388.00590941520 *T)
    sumR0  +=       0.367 * cos(4.43875659716  + 4535.05943692440 *T)
    sumR0  +=       0.406 * cos(4.20863156600  + 12592.45001978260 *T)
    sumR0  +=       0.360 * cos(2.53924644657  + 242.72860397400 *T)
    sumR0  +=       0.471 * cos(4.61907324819  + 5436.99301524020 *T)
    sumR0  +=       0.441 * cos(5.83872966262  + 3496.03282613400 *T)
    sumR0  +=       0.385 * cos(4.94496680973  + 24356.78078864160 *T)
    sumR0  +=       0.349 * cos(6.15018231784  + 19800.94595622480 *T)
    sumR0  +=       0.355 * cos(0.21895678106  + 5429.87946823940 *T)
    sumR0  +=       0.344 * cos(5.62993724928  + 2379.16447357160 *T)
    sumR0  +=       0.380 * cos(2.72105213143  + 11933.36796066960 *T)
    sumR0  +=       0.432 * cos(0.24221790536  + 17996.03116822220 *T)
    sumR0  +=       0.378 * cos(5.22517556974  + 7477.52286021600 *T)
    sumR0  +=       0.337 * cos(5.10888041439  + 5849.36411211460 *T)
    sumR0  +=       0.315 * cos(0.57827745123  + 10557.59416082380 *T)
    sumR0  +=       0.318 * cos(4.49953141399  + 3634.62102451840 *T)
    sumR0  +=       0.323 * cos(1.54274281393  + 10440.27429260360 *T)
    sumR0  +=       0.309 * cos(5.76839284397  + 20.77539549240 *T)
    sumR0  +=       0.301 * cos(2.34727604008  + 4686.88940770680 *T)
    sumR0  +=       0.414 * cos(5.93237602310  + 51092.72605085480 *T)
    sumR0  +=       0.361 * cos(2.16398609550  + 28237.23345938940 *T)
    sumR0  +=       0.288 * cos(0.18376252189  + 13095.84266507740 *T)
    sumR0  +=       0.277 * cos(5.12952205045  + 13119.72110282519 *T)
    sumR0  +=       0.327 * cos(6.19222146204  + 6268.84875598980 *T)
    sumR0  +=       0.273 * cos(0.30522428863  + 23141.55838292460 *T)
    sumR0  +=       0.267 * cos(5.76152585786  + 5966.68398033480 *T)
    sumR0  +=       0.308 * cos(5.99280509979  + 22805.73556599360 *T)
    sumR0  +=       0.345 * cos(2.92489919444  + 36949.23080842420 *T)
    sumR0  +=       0.253 * cos(5.20995219509  + 24072.92146977640 *T)
    sumR0  +=       0.342 * cos(5.72702586209  + 16460.33352952499 *T)
    sumR0  +=       0.261 * cos(2.00304796059  + 6148.01076995600 *T)
    sumR0  +=       0.238 * cos(5.08264392839  + 6915.85958930460 *T)
    sumR0  +=       0.249 * cos(2.94762789744  + 135.06508003540 *T)
    sumR0  +=       0.306 * cos(3.89764686987  + 10988.80815753500 *T)
    sumR0  +=       0.305 * cos(0.05827812117  + 4701.11650170840 *T)
    sumR0  +=       0.319 * cos(2.95712862064  + 163096.18036118348 *T)
    sumR0  +=       0.209 * cos(4.43768461442  + 6546.15977336420 *T)
    sumR0  +=       0.270 * cos(2.06643178717  + 4804.20927592700 *T)
    sumR0  +=       0.217 * cos(0.73691592312  + 6303.85124548380 *T)
    sumR0  +=       0.206 * cos(0.32075959415  + 25934.12433108940 *T)
    sumR0  +=       0.218 * cos(0.18428135264  + 28286.99048486120 *T)
    sumR0  +=       0.205 * cos(5.21312087405  + 20995.39296644940 *T)
    sumR0  +=       0.199 * cos(0.44384292491  + 16737.57723659660 *T)
    sumR0  +=       0.230 * cos(6.06567392849  + 6287.00800325450 *T)
    sumR0  +=       0.219 * cos(1.29194216300  + 5326.78669402080 *T)
    sumR0  +=       0.201 * cos(1.74700937253  + 22743.40937951640 *T)
    sumR0  +=       0.207 * cos(4.45440927276  + 6279.48542133960 *T)
    sumR0  +=       0.269 * cos(6.05640445030  + 64471.99124174489 *T)
    sumR0  +=       0.190 * cos(0.99256176518  + 29296.61538957860 *T)
    sumR0  +=       0.238 * cos(5.42471431221  + 39609.65458316560 *T)
    sumR0  +=       0.262 * cos(5.26961924198  + 522.57741809380 *T)
    sumR0  +=       0.210 * cos(4.68618183158  + 6254.62666252360 *T)
    sumR0  +=       0.197 * cos(2.80624554080  + 4933.20844033260 *T)
    sumR0  +=       0.252 * cos(4.36220154608  + 40879.44050464380 *T)
    sumR0  +=       0.261 * cos(1.07241516738  + 55022.93574707440 *T)
    sumR0  +=       0.189 * cos(3.82966734476  + 419.48464387520 *T)
    sumR0  +=       0.185 * cos(4.14324541379  + 5642.19824260920 *T)
    sumR0  +=       0.247 * cos(3.44855612987  + 6702.56049386660 *T)
    sumR0  +=       0.205 * cos(4.04424043223  + 536.80451209540 *T)
    sumR0  +=       0.191 * cos(3.14082686083  + 16723.35014259500 *T)
    sumR0  +=       0.222 * cos(5.16263907319  + 23539.70738633280 *T)
    sumR0  +=       0.180 * cos(4.56214752149  + 6489.26139842860 *T)
    sumR0  +=       0.219 * cos(0.80382553358  + 16627.37091537720 *T)
    sumR0  +=       0.227 * cos(0.60156339452  + 5905.70224207560 *T)
    sumR0  +=       0.168 * cos(0.88753528161  + 16062.18452611680 *T)
    sumR0  +=       0.158 * cos(0.92127725775  + 23937.85638974100 *T)
    sumR0  +=       0.157 * cos(4.69607868164  + 6805.65326808520 *T)
    sumR0  +=       0.207 * cos(4.88410451334  + 6286.66627864320 *T)
    sumR0  +=       0.160 * cos(4.95943826846  + 10021.83728009940 *T)
    sumR0  +=       0.166 * cos(0.97126433565  + 3097.88382272579 *T)
    sumR0  +=       0.209 * cos(5.75663411805  + 3646.35037735440 *T)
    sumR0  +=       0.175 * cos(6.12762824412  + 239424.39025435288 *T)
    sumR0  +=       0.173 * cos(3.13887234973  + 6179.98307577280 *T)
    sumR0  +=       0.157 * cos(3.62822058179  + 18451.07854656599 *T)
    sumR0  +=       0.157 * cos(4.67695912235  + 6709.67404086740 *T)
    sumR0  +=       0.146 * cos(3.09506069735  + 4907.30205014560 *T)
    sumR0  +=       0.165 * cos(2.27139128760  + 10660.68693504240 *T)
    sumR0  +=       0.201 * cos(1.67701267433  + 2107.03450754240 *T)
    sumR0  +=       0.144 * cos(3.96947747592  + 6019.99192661860 *T)
    sumR0  +=       0.171 * cos(5.91302216729  + 6058.73105428950 *T)
    sumR0  +=       0.144 * cos(2.13155655120  + 26084.02180621620 *T)
    sumR0  +=       0.151 * cos(0.67417383554  + 2388.89402044920 *T)
    sumR0  +=       0.189 * cos(5.07122281033  + 263.08392337280 *T)
    sumR0  +=       0.146 * cos(5.10373877968  + 10770.89325626180 *T)
    sumR0  +=       0.187 * cos(1.23915444627  + 19402.79695281660 *T)
    sumR0  +=       0.174 * cos(0.08407293391  + 9380.95967271720 *T)
    sumR0  +=       0.137 * cos(1.26247412309  + 12566.21901028560 *T)
    sumR0  +=       0.137 * cos(3.52826010842  + 639.89728631400 *T)
    sumR0  +=       0.148 * cos(1.76124372592  + 5888.44996493220 *T)
    sumR0  +=       0.164 * cos(2.39195095081  + 6357.85744855870 *T)
    sumR0  +=       0.146 * cos(2.43675816553  + 5881.40372823420 *T)
    sumR0  +=       0.161 * cos(1.15721259372  + 26735.94526221320 *T)
    sumR0  +=       0.131 * cos(2.51859277344  + 6599.46771964800 *T)
    sumR0  +=       0.153 * cos(5.85203687779  + 6281.59137728310 *T)
    sumR0  +=       0.151 * cos(3.72338532649  + 12669.24447420140 *T)
    sumR0  +=       0.132 * cos(2.38417741883  + 6525.80445396540 *T)
    sumR0  +=       0.129 * cos(0.75556744143  + 5017.50837136500 *T)
    sumR0  +=       0.127 * cos(0.00254936441  + 10027.90319572920 *T)
    sumR0  +=       0.148 * cos(2.85102145528  + 6418.14093002680 *T)
    sumR0  +=       0.143 * cos(5.74460279367  + 26087.90314157420 *T)
    sumR0  +=       0.172 * cos(0.41289962240  + 174242.46596404971 *T)
    sumR0  +=       0.136 * cos(4.15497742275  + 6311.52503745920 *T)
    sumR0  +=       0.170 * cos(5.98194913129  + 327574.51427678124 *T)
    sumR0  +=       0.124 * cos(1.65497607604  + 32217.20018108080 *T)
    sumR0  +=       0.136 * cos(2.48430783417  + 13341.67431130680 *T)
    sumR0  +=       0.165 * cos(2.49667924600  + 58953.14544329400 *T)
    sumR0  +=       0.123 * cos(3.45660563754  + 6277.55292568400 *T)
    sumR0  +=       0.117 * cos(0.86065134175  + 6245.04817735560 *T)
    sumR0  +=       0.149 * cos(5.61358280963  + 5729.50644714900 *T)
    sumR0  +=       0.153 * cos(0.26860029950  + 245.83164622940 *T)
    sumR0  +=       0.128 * cos(0.71204006588  + 103.09277421860 *T)
    sumR0  +=       0.159 * cos(2.43166592149  + 221995.02880149524 *T)
    sumR0  +=       0.130 * cos(2.80707316718  + 6016.46880826960 *T)
    sumR0  +=       0.137 * cos(1.70657709294  + 12566.08438968000 *T)
    sumR0  +=       0.111 * cos(1.56305648432  + 17782.73207278420 *T)
    sumR0  +=       0.113 * cos(3.58302904101  + 25685.87280280800 *T)
    sumR0  +=       0.109 * cos(3.26403795962  + 6819.88036208680 *T)
    sumR0  +=       0.122 * cos(0.34120688217  + 1162.47470440780 *T)
    sumR0  +=       0.119 * cos(5.84644718278  + 12721.57209941700 *T)
    sumR0  +=       0.144 * cos(2.28899679126  + 12489.88562870720 *T)
    sumR0  +=       0.137 * cos(5.82029768354  + 44809.65020086340 *T)
    sumR0  +=       0.107 * cos(2.42818544140  + 5547.19933645960 *T)
    sumR0  +=       0.134 * cos(1.26539982939  + 5331.35744374080 *T)
    sumR0  +=       0.103 * cos(5.96518130595  + 6321.10352262720 *T)
    sumR0  +=       0.109 * cos(0.33808549034  + 11300.58422135640 *T)
    sumR0  +=       0.129 * cos(5.89187277327  + 12029.34718788740 *T)
    sumR0  +=       0.122 * cos(5.77325634636  + 11919.14086666800 *T)
    sumR0  +=       0.107 * cos(6.24998989350  + 77690.75950573849 *T)
    sumR0  +=       0.107 * cos(1.00535580713  + 77736.78343050249 *T)
    sumR0  +=       0.143 * cos(0.24122178432  + 4214.06901508480 *T)
    sumR0  +=       0.143 * cos(0.88529649733  + 7576.56007357400 *T)
    sumR0  +=       0.107 * cos(2.92124030496  + 31415.37924995700 *T)
    sumR0  +=       0.099 * cos(5.70862227072  + 5540.08578945880 *T)
    sumR0  +=       0.110 * cos(0.37528037383  + 5863.59120611620 *T)
    sumR0  +=       0.104 * cos(4.44107178366  + 2118.76386037840 *T)
    sumR0  +=       0.098 * cos(5.95877916706  + 4061.21921539440 *T)
    sumR0  +=       0.113 * cos(1.24206857385  + 84672.47584450469 *T)
    sumR0  +=       0.124 * cos(2.55619029867  + 12539.85338018300 *T)
    sumR0  +=       0.110 * cos(3.66952094329  + 238004.52415723627 *T)
    sumR0  +=       0.112 * cos(4.32512422943  + 97238.62754448748 *T)
    sumR0  +=       0.097 * cos(3.70151541181  + 11720.06886523160 *T)
    sumR0  +=       0.120 * cos(1.26895630252  + 12043.57428188900 *T)
    sumR0  +=       0.094 * cos(2.56461130309  + 19004.64794940840 *T)
    sumR0  +=       0.117 * cos(3.65425622684  + 34520.30930938080 *T)
    sumR0  +=       0.098 * cos(0.13589994287  + 11080.17157891760 *T)
    sumR0  +=       0.097 * cos(5.38330115253  + 7834.12107263940 *T)
    sumR0  +=       0.097 * cos(2.46722096722  + 71980.63357473118 *T)
    sumR0  +=       0.095 * cos(5.36958330451  + 6288.59877429880 *T)
    sumR0  +=       0.111 * cos(5.01961920313  + 11823.16163945020 *T)
    sumR0  +=       0.090 * cos(2.72299804525  + 26880.31981303260 *T)
    sumR0  +=       0.099 * cos(0.90164266377  + 18635.92845453620 *T)
    sumR0  +=       0.126 * cos(4.78722177847  + 305281.94307104882 *T)
    sumR0  +=       0.093 * cos(0.21240380046  + 18139.29450141590 *T)
    sumR0  +=       0.124 * cos(5.00979495566  + 172146.97134054027 *T)
    sumR0  +=       0.099 * cos(5.67090026475  + 16522.65971600220 *T)
    sumR0  +=       0.092 * cos(2.28180963676  + 12491.37010141550 *T)
    sumR0  +=       0.090 * cos(4.50544881196  + 40077.61957352000 *T)
    sumR0  +=       0.100 * cos(2.00639461612  + 12323.42309600880 *T)
    sumR0  +=       0.095 * cos(5.68801979087  + 14919.01785375460 *T)
    sumR0  +=       0.087 * cos(1.86043406047  + 27707.54249429480 *T)
    sumR0  +=       0.105 * cos(3.02903468417  + 22345.26037610820 *T)
    sumR0  +=       0.087 * cos(5.43970168638  + 6272.03014972750 *T)
    sumR0  +=       0.089 * cos(1.63389387182  + 33326.57873317420 *T)
    sumR0  +=       0.082 * cos(5.58298993353  + 10241.20229116720 *T)
    sumR0  +=       0.094 * cos(5.47749711149  + 9924.81042151060 *T)
    sumR0  +=       0.082 * cos(4.71988314145  + 15141.39079431200 *T)
    sumR0  +=       0.097 * cos(5.61458778738  + 2787.04302385740 *T)
    sumR0  +=       0.096 * cos(3.89073946348  + 6379.05507720920 *T)
    sumR0  +=       0.081 * cos(3.13038482444  + 36147.40987730040 *T)
    sumR0  +=       0.110 * cos(4.89978492291  + 72140.62866668739 *T)
    sumR0  +=       0.097 * cos(5.20764563059  + 6303.43116939020 *T)
    sumR0  +=       0.082 * cos(5.26342716139  + 9814.60410029120 *T)
    sumR0  +=       0.109 * cos(2.35555589770  + 83286.91426955358 *T)
    sumR0  +=       0.097 * cos(2.58492958057  + 30666.15495843280 *T)
    sumR0  +=       0.093 * cos(1.32651591333  + 23020.65308658799 *T)
    sumR0  +=       0.078 * cos(3.99588630754  + 11293.47067435560 *T)
    sumR0  +=       0.090 * cos(0.57771932738  + 26482.17080962440 *T)
    sumR0  +=       0.106 * cos(3.92012705073  + 62883.35513951360 *T)
    sumR0  +=       0.098 * cos(2.94397773524  + 316.39186965660 *T)
    sumR0  +=       0.076 * cos(3.96310417608  + 29026.48522950779 *T)
    sumR0  +=       0.078 * cos(1.97068529306  + 90279.92316810328 *T)
    sumR0  +=       0.076 * cos(0.23027966596  + 21424.46664430340 *T)
    sumR0  +=       0.080 * cos(2.23099742212  + 266.60704172180 *T)
    sumR0  +=       0.079 * cos(1.46227790922  + 8982.81066930900 *T)
    sumR0  +=       0.102 * cos(4.92129953565  + 5621.84292321040 *T)
    sumR0  +=       0.100 * cos(0.39243148321  + 24279.10701821359 *T)
    sumR0  +=       0.071 * cos(1.52014858474  + 33794.54372352860 *T)
    sumR0  +=       0.076 * cos(0.22880641443  + 57375.80190084620 *T)
    sumR0  +=       0.091 * cos(0.96515913904  + 48739.85989708300 *T)
    sumR0  +=       0.075 * cos(2.77638585157  + 12964.30070339100 *T)
    sumR0  +=       0.077 * cos(5.18846946344  + 11520.99686379520 *T)
    sumR0  +=       0.068 * cos(0.50006599129  + 4274.51831083240 *T)
    sumR0  +=       0.075 * cos(2.07323762803  + 15664.03552270859 *T)
    sumR0  +=       0.074 * cos(1.01884134928  + 6393.28217121080 *T)
    sumR0  +=       0.077 * cos(0.46665178780  + 16207.88627150200 *T)
    sumR0  +=       0.081 * cos(4.10452219483  + 161710.61878623240 *T)
    sumR0  +=       0.067 * cos(3.83840630887  + 6262.72053059260 *T)
    sumR0  +=       0.071 * cos(3.91415523291  + 7875.67186362420 *T)
    sumR0  +=       0.081 * cos(0.91938383237  + 74.78159856730 *T)
    sumR0  +=       0.083 * cos(4.69916218791  + 23006.42599258639 *T)
    sumR0  +=       0.063 * cos(2.32556465878  + 6279.19451463340 *T)
    sumR0  +=       0.065 * cos(5.41938745446  + 28628.33622609960 *T)
    sumR0  +=       0.065 * cos(3.02336771694  + 5959.57043333400 *T)
    sumR0  +=       0.064 * cos(3.31033198370  + 2636.72547263700 *T)
    sumR0  +=       0.064 * cos(0.18375587519  + 1066.49547719000 *T)
    sumR0  +=       0.080 * cos(5.81239171612  + 12341.80690428090 *T)
    sumR0  +=       0.066 * cos(2.15105504851  + 38.02767263580 *T)
    sumR0  +=       0.062 * cos(2.43313614978  + 10138.10951694860 *T)
    sumR0  +=       0.060 * cos(3.16153906470  + 5490.30096152400 *T)
    sumR0  +=       0.069 * cos(0.30764736334  + 7018.95236352320 *T)
    sumR0  +=       0.068 * cos(2.24442548639  + 24383.07910844140 *T)
    sumR0  +=       0.078 * cos(1.39649386463  + 9411.46461508720 *T)
    sumR0  +=       0.063 * cos(0.72976362625  + 6286.95718534940 *T)
    sumR0  +=       0.073 * cos(4.95125917731  + 6453.74872061060 *T)
    sumR0  +=       0.078 * cos(0.32736023459  + 6528.90749622080 *T)
    sumR0  +=       0.059 * cos(4.95362151577  + 35707.71008290740 *T)
    sumR0  +=       0.070 * cos(2.37962727525  + 15508.61512327440 *T)
    sumR0  +=       0.073 * cos(1.35229143111  + 5327.47610838280 *T)
    sumR0  +=       0.072 * cos(5.91833527334  + 10881.09957748120 *T)
    sumR0  +=       0.059 * cos(5.36231868425  + 10239.58386601080 *T)
    sumR0  +=       0.059 * cos(1.63156134967  + 61306.01159706580 *T)
    sumR0  +=       0.054 * cos(4.29491690425  + 21947.11137270000 *T)
    sumR0  +=       0.057 * cos(5.89190132575  + 34513.26307268280 *T)
    sumR0  +=       0.074 * cos(1.38235845304  + 9967.45389998160 *T)
    sumR0  +=       0.053 * cos(3.86543309344  + 32370.97899156560 *T)
    sumR0  +=       0.055 * cos(4.51794544854  + 34911.41207609100 *T)
    sumR0  +=       0.063 * cos(5.41479412056  + 11502.83761653050 *T)
    sumR0  +=       0.063 * cos(2.34416220742  + 11510.70192305670 *T)
    sumR0  +=       0.068 * cos(0.77493931112  + 29864.33402730900 *T)
    sumR0  +=       0.060 * cos(5.57024703495  + 5756.90800324580 *T)
    sumR0  +=       0.072 * cos(2.80863088166  + 10866.87248347960 *T)
    sumR0  +=       0.061 * cos(2.69736991384  + 82576.98122099529 *T)
    sumR0  +=       0.063 * cos(5.32068807257  + 3116.65941225980 *T)
    sumR0  +=       0.052 * cos(1.02278758099  + 6272.43918464160 *T)
    sumR0  +=       0.069 * cos(5.00698550308  + 25287.72379939980 *T)
    sumR0  +=       0.066 * cos(6.12047940728  + 12074.48840752400 *T)
    sumR0  +=       0.051 * cos(2.59519527563  + 11396.56344857420 *T)
    sumR0  +=       0.056 * cos(2.57995973521  + 17892.93839400359 *T)
    sumR0  +=       0.059 * cos(0.44167237620  + 250570.67585721909 *T)
    sumR0  +=       0.059 * cos(3.84070143543  + 5483.25472482600 *T)
    sumR0  +=       0.049 * cos(0.54704693048  + 22594.05489571199 *T)
    sumR0  +=       0.065 * cos(2.38423614501  + 52670.06959330260 *T)
    sumR0  +=       0.069 * cos(5.34363738671  + 66813.56483573320 *T)
    sumR0  +=       0.057 * cos(5.42770501007  + 310145.15282392362 *T)
    sumR0  +=       0.053 * cos(1.17760296075  + 149.56319713460 *T)
    sumR0  +=       0.061 * cos(4.02090887211  + 34596.36465465240 *T)
    sumR0  +=       0.049 * cos(4.18361320516  + 18606.49894600020 *T)
    sumR0  +=       0.055 * cos(0.83886167974  + 20452.86941222180 *T)
    sumR0  +=       0.050 * cos(1.46327331958  + 37455.72649597440 *T)
    sumR0  +=       0.048 * cos(4.53854727167  + 29822.78323632420 *T)
    sumR0  +=       0.058 * cos(3.34847975377  + 33990.61834428620 *T)
    sumR0  +=       0.065 * cos(1.45522693982  + 76251.32777062019 *T)
    sumR0  +=       0.056 * cos(2.35650663692  + 37724.75341974820 *T)
    sumR0  +=       0.052 * cos(2.61551081496  + 5999.21653112620 *T)
    sumR0  +=       0.053 * cos(0.17334326094  + 77717.29458646949 *T)
    sumR0  +=       0.053 * cos(0.79879700631  + 77710.24834977149 *T)
    sumR0  +=       0.047 * cos(0.43240779709  + 735.87651353180 *T)
    sumR0  +=       0.053 * cos(4.58763261686  + 11616.97609101300 *T)
    sumR0  +=       0.048 * cos(6.20230111054  + 4171.42553661380 *T)
    sumR0  +=       0.052 * cos(1.09723616404  + 640.87760738220 *T)
    sumR0  +=       0.057 * cos(3.42008310383  + 50317.20343953080 *T)
    sumR0  +=       0.053 * cos(1.01528448581  + 149144.46708624958 *T)
    sumR0  +=       0.047 * cos(3.00924906195  + 52175.80628314840 *T)
    sumR0  +=       0.052 * cos(2.03254070404  + 6293.71251534120 *T)
    sumR0  +=       0.048 * cos(0.12356889734  + 13362.44970679920 *T)
    sumR0  +=       0.045 * cos(3.37963782356  + 10763.77970926100 *T)
    sumR0  +=       0.047 * cos(5.50981287869  + 12779.45079542080 *T)
    sumR0  +=       0.062 * cos(5.45209070099  + 949.17560896980 *T)
    sumR0  +=       0.061 * cos(2.93237974631  + 5791.41255753260 *T)
    sumR0  +=       0.044 * cos(2.87440620802  + 8584.66166590080 *T)
    sumR0  +=       0.046 * cos(4.03141796560  + 10667.80048204320 *T)
    sumR0  +=       0.047 * cos(3.89902931422  + 3903.91137641980 *T)
    sumR0  +=       0.046 * cos(2.75700467329  + 6993.00889854970 *T)
    sumR0  +=       0.045 * cos(1.93386293300  + 206.18554843720 *T)
    sumR0  +=       0.047 * cos(2.57670800912  + 11492.54267579200 *T)
    sumR0  +=       0.044 * cos(3.62570223167  + 63658.87775083760 *T)
    sumR0  +=       0.051 * cos(0.84536826273  + 12345.73905754400 *T)
    sumR0  +=       0.043 * cos(0.01524970172  + 37853.87549938260 *T)
    sumR0  +=       0.041 * cos(3.27146326065  + 8858.31494432060 *T)
    sumR0  +=       0.045 * cos(3.03765521215  + 65236.22129328540 *T)
    sumR0  +=       0.047 * cos(1.44447548944  + 21393.54196985760 *T)
    sumR0  +=       0.058 * cos(5.45843180927  + 1975.49254585600 *T)
    sumR0  +=       0.050 * cos(2.13285524146  + 12573.26524698360 *T)
    sumR0  +=       0.041 * cos(1.32190847146  + 2547.83753823240 *T)
    sumR0  +=       0.047 * cos(3.67579608544  + 28313.28880466100 *T)
    sumR0  +=       0.041 * cos(2.24013475126  + 8273.82086703240 *T)
    sumR0  +=       0.047 * cos(6.21438985953  + 10991.30589870060 *T)
    sumR0  +=       0.042 * cos(3.01631817350  + 853.19638175200 *T)
    sumR0  +=       0.056 * cos(1.09773690181  + 77376.20102240759 *T)
    sumR0  +=       0.040 * cos(2.35698541041  + 2699.73481931760 *T)
    sumR0  +=       0.043 * cos(5.28030898459  + 17796.95916678580 *T)
    sumR0  +=       0.054 * cos(2.59175932091  + 22910.44676536859 *T)
    sumR0  +=       0.054 * cos(0.88027764102  + 71960.38658322369 *T)
    sumR0  +=       0.055 * cos(0.07988899477  + 83467.15635301729 *T)
    sumR0  +=       0.039 * cos(1.12867321442  + 9910.58332750900 *T)
    sumR0  +=       0.040 * cos(1.35670430524  + 27177.85152920020 *T)
    sumR0  +=       0.039 * cos(4.39624220245  + 5618.31980486140 *T)
    sumR0  +=       0.042 * cos(4.78798367468  + 7856.89627409019 *T)
    sumR0  +=       0.047 * cos(2.75482175292  + 18202.21671665939 *T)
    sumR0  +=       0.039 * cos(1.97008298629  + 24491.42579258340 *T)
    sumR0  +=       0.042 * cos(4.04346599946  + 7863.94251078820 *T)
    sumR0  +=       0.038 * cos(0.49178679251  + 38650.17350619900 *T)
    sumR0  +=       0.036 * cos(4.86047906533  + 4157.19844261220 *T)
    sumR0  +=       0.043 * cos(5.64354880978  + 1062.90504853820 *T)
    sumR0  +=       0.036 * cos(3.98066313627  + 12565.17137891460 *T)
    sumR0  +=       0.042 * cos(2.30753932657  + 6549.68289171320 *T)
    sumR0  +=       0.040 * cos(5.39694918320  + 9498.21223063460 *T)
    sumR0  +=       0.040 * cos(3.30603243754  + 23536.11695768099 *T)
    sumR0  +=       0.050 * cos(6.15760345261  + 78051.34191383338 *T)

    # EARTH Radius
    sumR1  = 0
    sumR1  +=  103018.608 * cos(1.10748969588  + 6283.07584999140 *T)
    sumR1  +=    1721.238 * cos(1.06442301418  + 12566.15169998280 *T)
    sumR1  +=     702.215 * cos(3.14159265359  + 0.00000000000 *T)
    sumR1  +=      32.346 * cos(1.02169059149  + 18849.22754997420 *T)
    sumR1  +=      30.799 * cos(2.84353804832  + 5507.55323866740 *T)
    sumR1  +=      24.971 * cos(1.31906709482  + 5223.69391980220 *T)
    sumR1  +=      18.485 * cos(1.42429748614  + 1577.34354244780 *T)
    sumR1  +=      10.078 * cos(5.91378194648  + 10977.07880469900 *T)
    sumR1  +=       8.634 * cos(0.27146150602  + 5486.77784317500 *T)
    sumR1  +=       8.654 * cos(1.42046854427  + 6275.96230299060 *T)
    sumR1  +=       5.069 * cos(1.68613426734  + 5088.62883976680 *T)
    sumR1  +=       4.985 * cos(6.01401770704  + 6286.59896834040 *T)
    sumR1  +=       4.669 * cos(5.98724494073  + 529.69096509460 *T)
    sumR1  +=       4.395 * cos(0.51800238019  + 4694.00295470760 *T)
    sumR1  +=       3.872 * cos(4.74969833437  + 2544.31441988340 *T)
    sumR1  +=       3.750 * cos(5.07097685568  + 796.29800681640 *T)
    sumR1  +=       4.100 * cos(1.08424786092  + 9437.76293488700 *T)
    sumR1  +=       3.518 * cos(0.02290216272  + 83996.84731811189 *T)
    sumR1  +=       3.436 * cos(0.94937019624  + 71430.69561812909 *T)
    sumR1  +=       3.221 * cos(6.15628775313  + 2146.16541647520 *T)
    sumR1  +=       3.414 * cos(5.41218322538  + 775.52261132400 *T)
    sumR1  +=       2.863 * cos(5.48432847146  + 10447.38783960440 *T)
    sumR1  +=       2.520 * cos(0.24276941146  + 398.14900340820 *T)
    sumR1  +=       2.201 * cos(4.95216196651  + 6812.76681508600 *T)
    sumR1  +=       2.186 * cos(0.41991743105  + 8031.09226305840 *T)
    sumR1  +=       2.838 * cos(3.42034351366  + 2352.86615377180 *T)
    sumR1  +=       2.554 * cos(6.13241878525  + 6438.49624942560 *T)
    sumR1  +=       1.932 * cos(5.31374608366  + 8429.24126646660 *T)
    sumR1  +=       2.429 * cos(3.09164528262  + 4690.47983635860 *T)
    sumR1  +=       1.730 * cos(1.53686208550  + 4705.73230754360 *T)
    sumR1  +=       2.250 * cos(3.68863633842  + 7084.89678111520 *T)
    sumR1  +=       2.093 * cos(1.28191783032  + 1748.01641306700 *T)
    sumR1  +=       1.441 * cos(0.81656250862  + 14143.49524243060 *T)
    sumR1  +=       1.483 * cos(3.22225357771  + 7234.79425624200 *T)
    sumR1  +=       1.754 * cos(3.22883705112  + 6279.55273164240 *T)
    sumR1  +=       1.583 * cos(4.09702349428  + 11499.65622279280 *T)
    sumR1  +=       1.575 * cos(5.53890170575  + 3154.68708489560 *T)
    sumR1  +=       1.847 * cos(1.82040335363  + 7632.94325965020 *T)
    sumR1  +=       1.504 * cos(3.63293385726  + 11513.88331679440 *T)
    sumR1  +=       1.337 * cos(4.64440864339  + 6836.64525283380 *T)
    sumR1  +=       1.275 * cos(2.69341415363  + 1349.86740965880 *T)
    sumR1  +=       1.352 * cos(6.15101580257  + 5746.27133789600 *T)
    sumR1  +=       1.125 * cos(3.35673439497  + 17789.84561978500 *T)
    sumR1  +=       1.470 * cos(3.65282991755  + 1194.44701022460 *T)
    sumR1  +=       1.177 * cos(2.57676109092  + 13367.97263110660 *T)
    sumR1  +=       1.101 * cos(4.49748696552  + 4292.33083295040 *T)
    sumR1  +=       1.234 * cos(5.65036509521  + 5760.49843189760 *T)
    sumR1  +=       0.984 * cos(0.65517395136  + 5856.47765911540 *T)
    sumR1  +=       0.928 * cos(2.32420318751  + 10213.28554621100 *T)
    sumR1  +=       1.077 * cos(5.82812169132  + 12036.46073488820 *T)
    sumR1  +=       0.916 * cos(0.76613009583  + 16730.46368959580 *T)
    sumR1  +=       0.877 * cos(1.50137505051  + 11926.25441366880 *T)
    sumR1  +=       1.023 * cos(5.62076589825  + 6256.77753019160 *T)
    sumR1  +=       0.851 * cos(0.65709335533  + 155.42039943420 *T)
    sumR1  +=       0.802 * cos(4.10519132088  + 951.71840625060 *T)
    sumR1  +=       0.857 * cos(1.41661697538  + 5753.38488489680 *T)
    sumR1  +=       0.994 * cos(1.14418521187  + 1059.38193018920 *T)
    sumR1  +=       0.813 * cos(1.63948433322  + 6681.22485339960 *T)
    sumR1  +=       0.662 * cos(4.55200452260  + 5216.58037280140 *T)
    sumR1  +=       0.644 * cos(4.19478168733  + 6040.34724601740 *T)
    sumR1  +=       0.626 * cos(1.50767713598  + 5643.17856367740 *T)
    sumR1  +=       0.590 * cos(6.18277145205  + 4164.31198961300 *T)
    sumR1  +=       0.635 * cos(0.52413263542  + 6290.18939699220 *T)
    sumR1  +=       0.650 * cos(0.97935690350  + 25132.30339996560 *T)
    sumR1  +=       0.568 * cos(2.30125315873  + 10973.55568635000 *T)
    sumR1  +=       0.547 * cos(5.27256412213  + 3340.61242669980 *T)
    sumR1  +=       0.547 * cos(2.20144422886  + 1592.59601363280 *T)
    sumR1  +=       0.526 * cos(0.92464258226  + 11371.70468975820 *T)
    sumR1  +=       0.490 * cos(5.90951388655  + 3894.18182954220 *T)
    sumR1  +=       0.478 * cos(1.66857963179  + 12168.00269657460 *T)
    sumR1  +=       0.516 * cos(3.59803483887  + 10969.96525769820 *T)
    sumR1  +=       0.518 * cos(3.97914412373  + 17298.18232732620 *T)
    sumR1  +=       0.534 * cos(5.03740926442  + 9917.69687450980 *T)
    sumR1  +=       0.487 * cos(2.50545369269  + 6127.65545055720 *T)
    sumR1  +=       0.416 * cos(4.04828175503  + 10984.19235169980 *T)
    sumR1  +=       0.538 * cos(5.54081539805  + 553.56940284240 *T)
    sumR1  +=       0.402 * cos(2.16544019233  + 7860.41939243920 *T)
    sumR1  +=       0.553 * cos(2.32177369366  + 11506.76976979360 *T)
    sumR1  +=       0.367 * cos(3.39152532250  + 6496.37494542940 *T)
    sumR1  +=       0.360 * cos(5.34379853282  + 7079.37385680780 *T)
    sumR1  +=       0.337 * cos(3.61563704045  + 11790.62908865880 *T)
    sumR1  +=       0.456 * cos(0.30754294809  + 801.82093112380 *T)
    sumR1  +=       0.417 * cos(3.70009308674  + 10575.40668294180 *T)
    sumR1  +=       0.381 * cos(5.82033971802  + 7058.59846131540 *T)
    sumR1  +=       0.321 * cos(0.31988767355  + 16200.77272450120 *T)
    sumR1  +=       0.364 * cos(1.08414306177  + 6309.37416979120 *T)
    sumR1  +=       0.294 * cos(4.54798604957  + 11856.21865142450 *T)
    sumR1  +=       0.290 * cos(1.26473978562  + 8635.94200376320 *T)
    sumR1  +=       0.399 * cos(4.16998866302  + 26.29831979980 *T)
    sumR1  +=       0.262 * cos(5.08316906342  + 10177.25767953360 *T)
    sumR1  +=       0.243 * cos(2.25746091190  + 11712.95531823080 *T)
    sumR1  +=       0.237 * cos(1.05070575346  + 242.72860397400 *T)
    sumR1  +=       0.275 * cos(3.45319481756  + 5884.92684658320 *T)
    sumR1  +=       0.255 * cos(5.38496831087  + 21228.39202354580 *T)
    sumR1  +=       0.307 * cos(4.24313526604  + 3738.76143010800 *T)
    sumR1  +=       0.216 * cos(3.46037894728  + 213.29909543800 *T)
    sumR1  +=       0.196 * cos(0.69029243914  + 1990.74501704100 *T)
    sumR1  +=       0.198 * cos(5.16301829964  + 12352.85260454480 *T)
    sumR1  +=       0.214 * cos(3.91876200279  + 13916.01910964160 *T)
    sumR1  +=       0.212 * cos(4.00861198517  + 5230.80746680300 *T)
    sumR1  +=       0.184 * cos(5.59805976614  + 6283.14316029419 *T)
    sumR1  +=       0.184 * cos(2.85275392124  + 7238.67559160000 *T)
    sumR1  +=       0.179 * cos(2.54259058334  + 14314.16811304980 *T)
    sumR1  +=       0.225 * cos(1.64458698399  + 4732.03062734340 *T)
    sumR1  +=       0.236 * cos(5.58826125715  + 6069.77675455340 *T)
    sumR1  +=       0.187 * cos(2.72805985443  + 6062.66320755260 *T)
    sumR1  +=       0.184 * cos(6.04216273598  + 6283.00853968860 *T)
    sumR1  +=       0.230 * cos(3.62591335086  + 6284.05617105960 *T)
    sumR1  +=       0.163 * cos(2.19117396803  + 18073.70493865020 *T)
    sumR1  +=       0.172 * cos(0.97612950740  + 3930.20969621960 *T)
    sumR1  +=       0.215 * cos(1.04672844028  + 3496.03282613400 *T)
    sumR1  +=       0.169 * cos(4.75084479006  + 17267.26820169119 *T)
    sumR1  +=       0.152 * cos(0.19390712179  + 9779.10867612540 *T)
    sumR1  +=       0.182 * cos(5.16288118255  + 17253.04110768959 *T)
    sumR1  +=       0.149 * cos(0.80944184260  + 709.93304855830 *T)
    sumR1  +=       0.163 * cos(2.19209570390  + 6076.89030155420 *T)
    sumR1  +=       0.186 * cos(5.01159497089  + 11015.10647733480 *T)
    sumR1  +=       0.134 * cos(0.97765485759  + 65147.61976813770 *T)
    sumR1  +=       0.141 * cos(4.38421981312  + 4136.91043351620 *T)
    sumR1  +=       0.158 * cos(4.60974280627  + 9623.68827669120 *T)
    sumR1  +=       0.133 * cos(3.30508592837  + 154717.60988768269 *T)
    sumR1  +=       0.163 * cos(6.11782626245  + 3.52311834900 *T)
    sumR1  +=       0.174 * cos(1.58078542187  + 7.11354700080 *T)
    sumR1  +=       0.141 * cos(0.49976927274  + 25158.60171976540 *T)
    sumR1  +=       0.124 * cos(6.03440460031  + 9225.53927328300 *T)
    sumR1  +=       0.150 * cos(5.30166336812  + 13517.87010623340 *T)
    sumR1  +=       0.127 * cos(1.92389511438  + 22483.84857449259 *T)
    sumR1  +=       0.121 * cos(2.37813129011  + 167283.76158766547 *T)
    sumR1  +=       0.120 * cos(3.98423684853  + 4686.88940770680 *T)
    sumR1  +=       0.117 * cos(5.81072642211  + 12569.67481833180 *T)
    sumR1  +=       0.122 * cos(5.60973054224  + 5642.19824260920 *T)
    sumR1  +=       0.157 * cos(3.40236426002  + 16496.36139620240 *T)
    sumR1  +=       0.129 * cos(2.10705116371  + 1589.07289528380 *T)
    sumR1  +=       0.116 * cos(0.55839966736  + 5849.36411211460 *T)
    sumR1  +=       0.123 * cos(1.52961392771  + 12559.03815298200 *T)
    sumR1  +=       0.111 * cos(0.44848279675  + 6172.86952877200 *T)
    sumR1  +=       0.123 * cos(5.81645568991  + 6282.09552892320 *T)
    sumR1  +=       0.150 * cos(4.26278409223  + 3128.38876509580 *T)
    sumR1  +=       0.106 * cos(2.27437761356  + 5429.87946823940 *T)
    sumR1  +=       0.104 * cos(4.42743707728  + 23543.23050468179 *T)
    sumR1  +=       0.121 * cos(0.39459045915  + 12132.43996210600 *T)
    sumR1  +=       0.104 * cos(2.41842602527  + 426.59819087600 *T)
    sumR1  +=       0.110 * cos(5.80381480447  + 16858.48253293320 *T)
    sumR1  +=       0.100 * cos(2.93805577485  + 4535.05943692440 *T)
    sumR1  +=       0.097 * cos(3.97935904984  + 6133.51265285680 *T)
    sumR1  +=       0.110 * cos(6.22339014386  + 12146.66705610760 *T)
    sumR1  +=       0.098 * cos(0.87576563709  + 6525.80445396540 *T)
    sumR1  +=       0.098 * cos(3.15248421301  + 10440.27429260360 *T)
    sumR1  +=       0.095 * cos(2.46168411100  + 3097.88382272579 *T)
    sumR1  +=       0.088 * cos(0.23371480284  + 13119.72110282519 *T)
    sumR1  +=       0.098 * cos(5.77016493489  + 7342.45778018060 *T)
    sumR1  +=       0.092 * cos(6.03915555063  + 20426.57109242200 *T)
    sumR1  +=       0.096 * cos(5.56909292561  + 2388.89402044920 *T)
    sumR1  +=       0.081 * cos(1.32131147691  + 5650.29211067820 *T)
    sumR1  +=       0.086 * cos(3.94529200528  + 10454.50138660520 *T)
    sumR1  +=       0.076 * cos(2.70729716925  + 143571.32428481648 *T)
    sumR1  +=       0.091 * cos(5.64100034152  + 8827.39026987480 *T)
    sumR1  +=       0.076 * cos(1.80783856698  + 28286.99048486120 *T)
    sumR1  +=       0.081 * cos(1.90858992196  + 29088.81141598500 *T)
    sumR1  +=       0.075 * cos(3.40955892978  + 5481.25491886760 *T)
    sumR1  +=       0.069 * cos(4.49936170873  + 17256.63153634140 *T)
    sumR1  +=       0.088 * cos(1.10098454357  + 11769.85369316640 *T)
    sumR1  +=       0.066 * cos(2.78285801977  + 536.80451209540 *T)
    sumR1  +=       0.068 * cos(3.88179770758  + 17260.15465469040 *T)
    sumR1  +=       0.084 * cos(1.59303306354  + 9380.95967271720 *T)
    sumR1  +=       0.088 * cos(3.88076636762  + 7477.52286021600 *T)
    sumR1  +=       0.061 * cos(6.17558202197  + 11087.28512591840 *T)
    sumR1  +=       0.060 * cos(4.34824715818  + 6206.80977871580 *T)
    sumR1  +=       0.082 * cos(4.59843208943  + 9388.00590941520 *T)
    sumR1  +=       0.079 * cos(1.63131230601  + 4933.20844033260 *T)
    sumR1  +=       0.078 * cos(4.20905757484  + 5729.50644714900 *T)
    sumR1  +=       0.057 * cos(5.48157926651  + 18319.53658487960 *T)
    sumR1  +=       0.060 * cos(1.01261781084  + 12721.57209941700 *T)
    sumR1  +=       0.056 * cos(1.63031935692  + 15720.83878487840 *T)
    sumR1  +=       0.055 * cos(0.24926735018  + 15110.46611986620 *T)
    sumR1  +=       0.061 * cos(5.93059279661  + 12539.85338018300 *T)
    sumR1  +=       0.055 * cos(4.84298966314  + 13095.84266507740 *T)
    sumR1  +=       0.067 * cos(6.11690589247  + 8662.24032356300 *T)
    sumR1  +=       0.054 * cos(5.73750638571  + 3634.62102451840 *T)
    sumR1  +=       0.074 * cos(1.05466745829  + 16460.33352952499 *T)
    sumR1  +=       0.053 * cos(2.29084335688  + 16062.18452611680 *T)
    sumR1  +=       0.064 * cos(2.13513767927  + 7875.67186362420 *T)
    sumR1  +=       0.067 * cos(0.07096807518  + 14945.31617355440 *T)
    sumR1  +=       0.051 * cos(2.31511194429  + 6262.72053059260 *T)
    sumR1  +=       0.057 * cos(5.77055471237  + 12043.57428188900 *T)
    sumR1  +=       0.056 * cos(4.41980790431  + 4701.11650170840 *T)
    sumR1  +=       0.059 * cos(5.87963500073  + 5331.35744374080 *T)
    sumR1  +=       0.058 * cos(2.30546168628  + 955.59974160860 *T)
    sumR1  +=       0.049 * cos(1.93839278478  + 5333.90024102160 *T)
    sumR1  +=       0.048 * cos(2.69973662261  + 6709.67404086740 *T)
    sumR1  +=       0.064 * cos(1.64379897981  + 6262.30045449900 *T)
    sumR1  +=       0.046 * cos(3.98449608961  + 98068.53671630540 *T)
    sumR1  +=       0.050 * cos(3.68875893005  + 12323.42309600880 *T)
    sumR1  +=       0.045 * cos(3.30068569697  + 22003.91463486980 *T)
    sumR1  +=       0.047 * cos(1.26317154881  + 11919.14086666800 *T)
    sumR1  +=       0.045 * cos(0.89150445122  + 51868.24866217880 *T)
    sumR1  +=       0.043 * cos(1.61526242998  + 6277.55292568400 *T)
    sumR1  +=       0.043 * cos(5.74295325645  + 11403.67699557500 *T)
    sumR1  +=       0.044 * cos(3.43070646822  + 10021.83728009940 *T)
    sumR1  +=       0.056 * cos(0.02481833774  + 15671.08175940660 *T)
    sumR1  +=       0.055 * cos(3.14274403422  + 33019.02111220460 *T)
    sumR1  +=       0.045 * cos(3.00877289177  + 8982.81066930900 *T)
    sumR1  +=       0.046 * cos(0.73303568429  + 6303.43116939020 *T)
    sumR1  +=       0.049 * cos(1.60455690285  + 6303.85124548380 *T)
    sumR1  +=       0.045 * cos(0.40210030323  + 6805.65326808520 *T)
    sumR1  +=       0.053 * cos(0.94869680175  + 10988.80815753500 *T)
    sumR1  +=       0.041 * cos(1.61122384329  + 6819.88036208680 *T)
    sumR1  +=       0.055 * cos(0.89439119424  + 11933.36796066960 *T)
    sumR1  +=       0.045 * cos(3.88495384656  + 60530.48898574180 *T)
    sumR1  +=       0.040 * cos(4.75740908001  + 38526.57435087200 *T)
    sumR1  +=       0.040 * cos(1.49921251887  + 18451.07854656599 *T)
    sumR1  +=       0.040 * cos(3.77498297228  + 26087.90314157420 *T)
    sumR1  +=       0.051 * cos(1.70258603562  + 1551.04522264800 *T)
    sumR1  +=       0.039 * cos(2.97100699926  + 2118.76386037840 *T)
    sumR1  +=       0.053 * cos(5.19854123078  + 77713.77146812050 *T)
    sumR1  +=       0.047 * cos(4.26356628717  + 21424.46664430340 *T)
    sumR1  +=       0.037 * cos(0.62902722802  + 24356.78078864160 *T)
    sumR1  +=       0.036 * cos(0.11087914947  + 10344.29506538580 *T)
    sumR1  +=       0.036 * cos(0.77037556319  + 12029.34718788740 *T)
    sumR1  +=       0.035 * cos(3.30933994515  + 24072.92146977640 *T)
    sumR1  +=       0.035 * cos(5.93650887012  + 31570.79964939120 *T)
    sumR1  +=       0.036 * cos(2.15108874765  + 30774.50164257480 *T)
    sumR1  +=       0.036 * cos(1.75078825382  + 16207.88627150200 *T)
    sumR1  +=       0.033 * cos(5.06264177921  + 226858.23855437007 *T)
    sumR1  +=       0.034 * cos(6.16891378800  + 24491.42579258340 *T)
    sumR1  +=       0.035 * cos(3.19120695549  + 32217.20018108080 *T)
    sumR1  +=       0.034 * cos(2.31528650443  + 55798.45835839840 *T)
    sumR1  +=       0.032 * cos(4.21446357042  + 15664.03552270859 *T)
    sumR1  +=       0.039 * cos(1.24979117796  + 6418.14093002680 *T)
    sumR1  +=       0.037 * cos(4.11943655770  + 2787.04302385740 *T)
    sumR1  +=       0.032 * cos(1.62887710890  + 639.89728631400 *T)
    sumR1  +=       0.038 * cos(5.89832942685  + 640.87760738220 *T)
    sumR1  +=       0.032 * cos(1.72442327688  + 27433.88921587499 *T)
    sumR1  +=       0.031 * cos(2.78828943753  + 12139.55350910680 *T)
    sumR1  +=       0.035 * cos(4.44608896525  + 18202.21671665939 *T)
    sumR1  +=       0.034 * cos(3.96287980676  + 18216.44381066100 *T)
    sumR1  +=       0.033 * cos(4.73611335874  + 16723.35014259500 *T)
    sumR1  +=       0.034 * cos(1.43910280005  + 49515.38250840700 *T)
    sumR1  +=       0.031 * cos(0.23302920161  + 23581.25817731760 *T)
    sumR1  +=       0.029 * cos(2.02633840220  + 11609.86254401220 *T)
    sumR1  +=       0.030 * cos(2.54923230240  + 9924.81042151060 *T)
    sumR1  +=       0.032 * cos(4.91793198558  + 11300.58422135640 *T)
    sumR1  +=       0.028 * cos(0.26187189577  + 13521.75144159140 *T)
    sumR1  +=       0.028 * cos(3.84568936822  + 2699.73481931760 *T)
    sumR1  +=       0.029 * cos(1.83149729794  + 29822.78323632420 *T)
    sumR1  +=       0.033 * cos(4.60320094415  + 19004.64794940840 *T)
    sumR1  +=       0.027 * cos(4.46183450287  + 6702.56049386660 *T)
    sumR1  +=       0.030 * cos(4.46494072240  + 36147.40987730040 *T)
    sumR1  +=       0.027 * cos(0.03211931363  + 6279.78949257360 *T)
    sumR1  +=       0.026 * cos(5.46497324333  + 6245.04817735560 *T)
    sumR1  +=       0.035 * cos(4.52695674113  + 36949.23080842420 *T)
    sumR1  +=       0.027 * cos(3.52528177609  + 10770.89325626180 *T)
    sumR1  +=       0.026 * cos(1.48499438453  + 11080.17157891760 *T)
    sumR1  +=       0.035 * cos(2.82154380962  + 19402.79695281660 *T)
    sumR1  +=       0.025 * cos(2.46339998836  + 6279.48542133960 *T)
    sumR1  +=       0.026 * cos(4.97688894643  + 16737.57723659660 *T)
    sumR1  +=       0.026 * cos(2.36136541526  + 17996.03116822220 *T)
    sumR1  +=       0.029 * cos(4.15148654061  + 45892.73043315699 *T)
    sumR1  +=       0.026 * cos(4.50714272714  + 17796.95916678580 *T)
    sumR1  +=       0.027 * cos(4.72625223674  + 1066.49547719000 *T)
    sumR1  +=       0.025 * cos(2.89309528854  + 6286.66627864320 *T)
    sumR1  +=       0.027 * cos(0.37462444357  + 12964.30070339100 *T)
    sumR1  +=       0.029 * cos(4.94860010533  + 5863.59120611620 *T)
    sumR1  +=       0.031 * cos(3.93096113577  + 29864.33402730900 *T)
    sumR1  +=       0.024 * cos(6.14987193584  + 18606.49894600020 *T)
    sumR1  +=       0.024 * cos(3.74225964547  + 29026.48522950779 *T)
    sumR1  +=       0.025 * cos(5.70460621565  + 27707.54249429480 *T)
    sumR1  +=       0.025 * cos(5.33928840652  + 15141.39079431200 *T)
    sumR1  +=       0.027 * cos(3.02320897140  + 6286.36220740920 *T)
    sumR1  +=       0.023 * cos(0.28364955406  + 5327.47610838280 *T)
    sumR1  +=       0.026 * cos(1.34240461687  + 18875.52586977400 *T)
    sumR1  +=       0.024 * cos(1.33998410121  + 19800.94595622480 *T)
    sumR1  +=       0.025 * cos(6.00172494004  + 6489.26139842860 *T)
    sumR1  +=       0.022 * cos(1.81777974484  + 6288.59877429880 *T)
    sumR1  +=       0.022 * cos(3.58603606640  + 6915.85958930460 *T)
    sumR1  +=       0.029 * cos(2.09564449439  + 15265.88651930040 *T)
    sumR1  +=       0.022 * cos(1.02173599251  + 11925.27409260060 *T)
    sumR1  +=       0.022 * cos(4.74660932338  + 28230.18722269139 *T)
    sumR1  +=       0.021 * cos(2.30688751432  + 5999.21653112620 *T)
    sumR1  +=       0.021 * cos(3.22654944430  + 25934.12433108940 *T)
    sumR1  +=       0.021 * cos(3.04956726238  + 6566.93516885660 *T)
    sumR1  +=       0.027 * cos(5.35653084499  + 33794.54372352860 *T)
    sumR1  +=       0.028 * cos(3.91168324815  + 18208.34994259200 *T)
    sumR1  +=       0.020 * cos(1.52296293311  + 135.06508003540 *T)
    sumR1  +=       0.022 * cos(4.66462839521  + 13362.44970679920 *T)
    sumR1  +=       0.019 * cos(1.78121167862  + 156137.47598479927 *T)
    sumR1  +=       0.019 * cos(2.99969102221  + 19651.04848109800 *T)
    sumR1  +=       0.019 * cos(2.86664273362  + 18422.62935909819 *T)
    sumR1  +=       0.025 * cos(0.94995632141  + 31415.37924995700 *T)
    sumR1  +=       0.019 * cos(4.71432851499  + 77690.75950573849 *T)
    sumR1  +=       0.019 * cos(2.54227398241  + 77736.78343050249 *T)
    sumR1  +=       0.020 * cos(5.91915117116  + 48739.85989708300 *T)

    # EARTH Radius
    sumR2  = 0
    sumR2  +=    4359.385 * cos(5.78455133738  + 6283.07584999140 *T)
    sumR2  +=     123.633 * cos(5.57934722157  + 12566.15169998280 *T)
    sumR2  +=      12.341 * cos(3.14159265359  + 0.00000000000 *T)
    sumR2  +=       8.792 * cos(3.62777733395  + 77713.77146812050 *T)
    sumR2  +=       5.689 * cos(1.86958905084  + 5573.14280143310 *T)
    sumR2  +=       3.301 * cos(5.47027913302  + 18849.22754997420 *T)
    sumR2  +=       1.471 * cos(4.48028885617  + 5507.55323866740 *T)
    sumR2  +=       1.013 * cos(2.81456417694  + 5223.69391980220 *T)
    sumR2  +=       0.854 * cos(3.10878241236  + 1577.34354244780 *T)
    sumR2  +=       1.102 * cos(2.84173992403  + 161000.68573767410 *T)
    sumR2  +=       0.648 * cos(5.47349498544  + 775.52261132400 *T)
    sumR2  +=       0.609 * cos(1.37969434104  + 6438.49624942560 *T)
    sumR2  +=       0.499 * cos(4.41649242250  + 6286.59896834040 *T)
    sumR2  +=       0.417 * cos(0.90242451175  + 10977.07880469900 *T)
    sumR2  +=       0.402 * cos(3.20376585290  + 5088.62883976680 *T)
    sumR2  +=       0.351 * cos(1.81079227770  + 5486.77784317500 *T)
    sumR2  +=       0.467 * cos(3.65753702738  + 7084.89678111520 *T)
    sumR2  +=       0.458 * cos(5.38585314743  + 149854.40013480789 *T)
    sumR2  +=       0.304 * cos(3.51701098693  + 796.29800681640 *T)
    sumR2  +=       0.266 * cos(6.17413982699  + 6836.64525283380 *T)
    sumR2  +=       0.279 * cos(1.84120501086  + 4694.00295470760 *T)
    sumR2  +=       0.260 * cos(1.41629543251  + 2146.16541647520 *T)
    sumR2  +=       0.266 * cos(3.13832905677  + 71430.69561812909 *T)
    sumR2  +=       0.321 * cos(5.35313367048  + 3154.68708489560 *T)
    sumR2  +=       0.238 * cos(2.17720020018  + 155.42039943420 *T)
    sumR2  +=       0.293 * cos(4.61501268144  + 4690.47983635860 *T)
    sumR2  +=       0.229 * cos(4.75969588070  + 7234.79425624200 *T)
    sumR2  +=       0.211 * cos(0.21868065485  + 4705.73230754360 *T)
    sumR2  +=       0.201 * cos(4.21905743357  + 1349.86740965880 *T)
    sumR2  +=       0.195 * cos(4.57808285364  + 529.69096509460 *T)
    sumR2  +=       0.253 * cos(2.81496293039  + 1748.01641306700 *T)
    sumR2  +=       0.182 * cos(5.70454011389  + 6040.34724601740 *T)
    sumR2  +=       0.179 * cos(6.02897097053  + 4292.33083295040 *T)
    sumR2  +=       0.186 * cos(1.58690991244  + 6309.37416979120 *T)
    sumR2  +=       0.170 * cos(2.90220009715  + 9437.76293488700 *T)
    sumR2  +=       0.166 * cos(1.99984925026  + 8031.09226305840 *T)
    sumR2  +=       0.158 * cos(0.04783713552  + 2544.31441988340 *T)
    sumR2  +=       0.197 * cos(2.01083639502  + 1194.44701022460 *T)
    sumR2  +=       0.165 * cos(5.78372596778  + 83996.84731811189 *T)
    sumR2  +=       0.214 * cos(3.38285934319  + 7632.94325965020 *T)
    sumR2  +=       0.140 * cos(0.36401486094  + 10447.38783960440 *T)
    sumR2  +=       0.151 * cos(0.95153163031  + 6127.65545055720 *T)
    sumR2  +=       0.136 * cos(1.48426306582  + 2352.86615377180 *T)
    sumR2  +=       0.127 * cos(5.48475435134  + 951.71840625060 *T)
    sumR2  +=       0.126 * cos(5.26866506592  + 6279.55273164240 *T)
    sumR2  +=       0.125 * cos(3.75754889288  + 6812.76681508600 *T)
    sumR2  +=       0.101 * cos(4.95015746147  + 398.14900340820 *T)
    sumR2  +=       0.102 * cos(0.68468295277  + 1592.59601363280 *T)
    sumR2  +=       0.100 * cos(1.14568935785  + 3894.18182954220 *T)
    sumR2  +=       0.129 * cos(0.76540016965  + 553.56940284240 *T)
    sumR2  +=       0.109 * cos(5.41063597567  + 6256.77753019160 *T)
    sumR2  +=       0.075 * cos(5.84804322893  + 242.72860397400 *T)
    sumR2  +=       0.095 * cos(1.94452244083  + 11856.21865142450 *T)
    sumR2  +=       0.077 * cos(0.69373708195  + 8429.24126646660 *T)
    sumR2  +=       0.100 * cos(5.19725292131  + 244287.60000722768 *T)
    sumR2  +=       0.080 * cos(6.18440483705  + 1059.38193018920 *T)
    sumR2  +=       0.069 * cos(5.25699888595  + 14143.49524243060 *T)
    sumR2  +=       0.085 * cos(5.39484725499  + 25132.30339996560 *T)
    sumR2  +=       0.066 * cos(0.51779993906  + 801.82093112380 *T)
    sumR2  +=       0.055 * cos(5.16878202461  + 7058.59846131540 *T)
    sumR2  +=       0.051 * cos(3.88759155247  + 12036.46073488820 *T)
    sumR2  +=       0.050 * cos(5.57636570536  + 6290.18939699220 *T)
    sumR2  +=       0.061 * cos(2.24359003264  + 8635.94200376320 *T)
    sumR2  +=       0.050 * cos(5.54441900966  + 1990.74501704100 *T)
    sumR2  +=       0.056 * cos(4.00301078040  + 13367.97263110660 *T)
    sumR2  +=       0.052 * cos(4.13138898038  + 7860.41939243920 *T)
    sumR2  +=       0.052 * cos(3.90943054011  + 26.29831979980 *T)
    sumR2  +=       0.041 * cos(3.57128482780  + 7079.37385680780 *T)
    sumR2  +=       0.056 * cos(2.76959005761  + 90955.55169449610 *T)
    sumR2  +=       0.042 * cos(1.91461189199  + 7477.52286021600 *T)
    sumR2  +=       0.042 * cos(0.42728171713  + 10213.28554621100 *T)
    sumR2  +=       0.042 * cos(1.09413724455  + 709.93304855830 *T)
    sumR2  +=       0.039 * cos(3.93298068961  + 10973.55568635000 *T)
    sumR2  +=       0.038 * cos(6.17935925345  + 9917.69687450980 *T)
    sumR2  +=       0.049 * cos(0.83021145241  + 11506.76976979360 *T)
    sumR2  +=       0.053 * cos(1.45828359397  + 233141.31440436147 *T)
    sumR2  +=       0.047 * cos(6.21568666789  + 6681.22485339960 *T)
    sumR2  +=       0.037 * cos(0.36359309980  + 10177.25767953360 *T)
    sumR2  +=       0.035 * cos(3.33024911524  + 5643.17856367740 *T)
    sumR2  +=       0.034 * cos(5.63446915337  + 6525.80445396540 *T)
    sumR2  +=       0.035 * cos(5.36033855038  + 25158.60171976540 *T)
    sumR2  +=       0.034 * cos(5.36319798321  + 4933.20844033260 *T)
    sumR2  +=       0.033 * cos(4.24722336872  + 12569.67481833180 *T)
    sumR2  +=       0.043 * cos(5.26370903404  + 10575.40668294180 *T)
    sumR2  +=       0.042 * cos(5.08837645072  + 11015.10647733480 *T)
    sumR2  +=       0.040 * cos(1.98334703186  + 6284.05617105960 *T)
    sumR2  +=       0.042 * cos(4.22496037505  + 88860.05707098669 *T)
    sumR2  +=       0.029 * cos(3.19088628170  + 11926.25441366880 *T)
    sumR2  +=       0.029 * cos(0.15217616684  + 12168.00269657460 *T)
    sumR2  +=       0.030 * cos(1.61904744136  + 9779.10867612540 *T)
    sumR2  +=       0.027 * cos(0.76388991416  + 1589.07289528380 *T)
    sumR2  +=       0.036 * cos(2.74712003443  + 3738.76143010800 *T)
    sumR2  +=       0.033 * cos(3.08807829566  + 3930.20969621960 *T)
    sumR2  +=       0.031 * cos(5.34906619513  + 143571.32428481648 *T)
    sumR2  +=       0.025 * cos(0.10240267494  + 22483.84857449259 *T)
    sumR2  +=       0.030 * cos(3.47110495524  + 14945.31617355440 *T)
    sumR2  +=       0.024 * cos(1.10425016019  + 4535.05943692440 *T)
    sumR2  +=       0.024 * cos(1.58037259780  + 6496.37494542940 *T)
    sumR2  +=       0.023 * cos(3.87710321433  + 6275.96230299060 *T)
    sumR2  +=       0.025 * cos(3.94529778970  + 3128.38876509580 *T)
    sumR2  +=       0.023 * cos(3.44685609601  + 4136.91043351620 *T)
    sumR2  +=       0.023 * cos(3.83156029849  + 5753.38488489680 *T)
    sumR2  +=       0.022 * cos(1.86956128067  + 16730.46368959580 *T)
    sumR2  +=       0.025 * cos(2.42188933855  + 5729.50644714900 *T)
    sumR2  +=       0.020 * cos(1.78208352927  + 17789.84561978500 *T)
    sumR2  +=       0.021 * cos(4.30363087400  + 16858.48253293320 *T)
    sumR2  +=       0.021 * cos(0.49258939822  + 29088.81141598500 *T)
    sumR2  +=       0.025 * cos(1.33030250444  + 6282.09552892320 *T)
    sumR2  +=       0.027 * cos(2.54785812264  + 3496.03282613400 *T)
    sumR2  +=       0.022 * cos(1.11232521950  + 12721.57209941700 *T)
    sumR2  +=       0.021 * cos(5.97759081637  + 7.11354700080 *T)
    sumR2  +=       0.019 * cos(0.80292033311  + 16062.18452611680 *T)
    sumR2  +=       0.023 * cos(4.12454848769  + 2388.89402044920 *T)
    sumR2  +=       0.022 * cos(4.92663152168  + 18875.52586977400 *T)
    sumR2  +=       0.023 * cos(5.68902059771  + 16460.33352952499 *T)
    sumR2  +=       0.023 * cos(4.97346265647  + 17260.15465469040 *T)
    sumR2  +=       0.023 * cos(3.03021283729  + 66567.48586525429 *T)
    sumR2  +=       0.016 * cos(3.89740925257  + 5331.35744374080 *T)
    sumR2  +=       0.017 * cos(3.08268671348  + 154717.60988768269 *T)
    sumR2  +=       0.016 * cos(3.95085099736  + 3097.88382272579 *T)
    sumR2  +=       0.016 * cos(3.99041783945  + 6283.14316029419 *T)
    sumR2  +=       0.020 * cos(6.10644140189  + 167283.76158766547 *T)
    sumR2  +=       0.015 * cos(4.09775914607  + 11712.95531823080 *T)
    sumR2  +=       0.016 * cos(5.71769940700  + 17298.18232732620 *T)
    sumR2  +=       0.016 * cos(3.28894009404  + 5884.92684658320 *T)
    sumR2  +=       0.015 * cos(5.64785377164  + 12559.03815298200 *T)
    sumR2  +=       0.016 * cos(4.43452080930  + 6283.00853968860 *T)
    sumR2  +=       0.014 * cos(2.31721603062  + 5481.25491886760 *T)
    sumR2  +=       0.014 * cos(4.43479032305  + 13517.87010623340 *T)
    sumR2  +=       0.014 * cos(4.73209312936  + 7342.45778018060 *T)
    sumR2  +=       0.012 * cos(0.64705975463  + 18073.70493865020 *T)
    sumR2  +=       0.011 * cos(1.51443332200  + 16200.77272450120 *T)
    sumR2  +=       0.011 * cos(0.88708889185  + 21228.39202354580 *T)
    sumR2  +=       0.014 * cos(4.50116508534  + 640.87760738220 *T)
    sumR2  +=       0.011 * cos(4.64339996198  + 11790.62908865880 *T)
    sumR2  +=       0.011 * cos(1.31064298246  + 4164.31198961300 *T)
    sumR2  +=       0.009 * cos(3.02238989305  + 23543.23050468179 *T)
    sumR2  +=       0.009 * cos(2.04999402381  + 22003.91463486980 *T)
    sumR2  +=       0.009 * cos(4.91488110218  + 213.29909543800 *T)

    # EARTH Radius
    sumR3  = 0
    sumR3  +=     144.595 * cos(4.27319435148  + 6283.07584999140 *T)
    sumR3  +=       6.729 * cos(3.91697608662  + 12566.15169998280 *T)
    sumR3  +=       0.774 * cos(0.00000000000  + 0.00000000000 *T)
    sumR3  +=       0.247 * cos(3.73019298781  + 18849.22754997420 *T)
    sumR3  +=       0.036 * cos(2.80081409050  + 6286.59896834040 *T)
    sumR3  +=       0.033 * cos(5.62216602775  + 6127.65545055720 *T)
    sumR3  +=       0.019 * cos(3.71292621802  + 6438.49624942560 *T)
    sumR3  +=       0.016 * cos(4.26011484232  + 6525.80445396540 *T)
    sumR3  +=       0.016 * cos(3.50416887054  + 6256.77753019160 *T)
    sumR3  +=       0.014 * cos(3.62127621114  + 25132.30339996560 *T)
    sumR3  +=       0.011 * cos(4.39200958819  + 4705.73230754360 *T)
    sumR3  +=       0.011 * cos(5.22327127059  + 6040.34724601740 *T)
    sumR3  +=       0.010 * cos(4.28045254647  + 83996.84731811189 *T)
    sumR3  +=       0.009 * cos(1.56864096494  + 5507.55323866740 *T)
    sumR3  +=       0.011 * cos(1.37795688024  + 6309.37416979120 *T)
    sumR3  +=       0.010 * cos(5.19937959068  + 71430.69561812909 *T)
    sumR3  +=       0.009 * cos(0.47275199930  + 6279.55273164240 *T)
    sumR3  +=       0.009 * cos(0.74642756529  + 5729.50644714900 *T)
    sumR3  +=       0.007 * cos(2.97374891560  + 775.52261132400 *T)
    sumR3  +=       0.007 * cos(3.28615691021  + 7058.59846131540 *T)
    sumR3  +=       0.007 * cos(2.19184402142  + 6812.76681508600 *T)
    sumR3  +=       0.005 * cos(3.15419034438  + 529.69096509460 *T)
    sumR3  +=       0.006 * cos(4.54725567047  + 1059.38193018920 *T)
    sumR3  +=       0.005 * cos(1.51104406936  + 7079.37385680780 *T)
    sumR3  +=       0.007 * cos(2.98052059053  + 6681.22485339960 *T)
    sumR3  +=       0.005 * cos(2.30961231391  + 12036.46073488820 *T)
    sumR3  +=       0.005 * cos(3.71102966917  + 6290.18939699220 *T)

    # EARTH Radius
    sumR4  = 0
    sumR4  +=       3.858 * cos(2.56384387339  + 6283.07584999140 *T)
    sumR4  +=       0.306 * cos(2.26769501230  + 12566.15169998280 *T)
    sumR4  +=       0.053 * cos(3.44031471924  + 5573.14280143310 *T)
    sumR4  +=       0.015 * cos(2.04794573436  + 18849.22754997420 *T)
    sumR4  +=       0.013 * cos(2.05688873673  + 77713.77146812050 *T)
    sumR4  +=       0.007 * cos(4.41218854480  + 161000.68573767410 *T)
    sumR4  +=       0.005 * cos(5.26154653107  + 6438.49624942560 *T)
    sumR4  +=       0.005 * cos(4.07695126049  + 6127.65545055720 *T)
    sumR4  +=       0.006 * cos(3.81514213664  + 149854.40013480789 *T)
    sumR4  +=       0.003 * cos(1.28175749811  + 6286.59896834040 *T)

    # EARTH Radius
    sumR5  = 0
    sumR5  +=       0.086 * cos(1.21579741687  + 6283.07584999140 *T)
    sumR5  +=       0.012 * cos(0.65617264033  + 12566.15169998280 *T)
    sumR5  +=       0.001 * cos(0.38068797142  + 18849.22754997420 *T)

    # calculate and scale the sums
    L = sumL0 + sumL1*T**1 + sumL2*T**2 + sumL3*T**3 + sumL4*T**4 + sumL5*T**5
    L /= 1e8
    L = Mod(degrees(L), 360.)
    B = sumB0 + sumB1*T**1 + sumB2*T**2 + sumB3*T**3 + sumB4*T**4
    B /= 1e8
    B = degrees(B)
    R = sumR0 + sumR1*T**1 + sumR2*T**2 + sumR3*T**3 + sumR4*T**4 + sumR5*T**5
    R /= 1e8

    # calculate X, Y, Z
    X = R * Cos(B) * Cos(L)
    Y = R * Cos(B) * Sin(L)
    Z = R * Sin(B)

    # create dictionary
    Dict = {  "Object"  : "Earth",
              "Mode"    : "LBR",
              "Long"    : L,
              "Lat"     : B,
              "Rau"     : R,
              "X"       : X,
              "Y"       : Y,
              "Z"       : Z,
              "Alt"     : -34./60.
           }
    Data = (L, B, R, Dict)


    # save parameters and values
    EarthdtUTimes[strdtUT] = Data


    return Data
#######################################################################



def getEarthData(dtCivicTime, zone, verbose = False):
    L, B, R, Dict = getEarthLBRData(dtCivicTime, zone, verbose)
    X = Dict["X"]; Y = Dict["Y"]; Z = Dict["Z"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Le, Be, Re, Dicte = getEarthLBRData(dtCivicTime, zone, verbose)
        Xe = Dicte["X"]; Ye = Dicte["Y"]; Ze = Dicte["Z"]
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getJupiterLBRData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Jupiter"
#       Long  :  heliocentric Longitude, degrees
#       Lat   :  heliocentric Latitude,degrees
#       Rau   :  Radius, distaance Sun - Object (au)
#
#   Similarly for each planet
#   data from Meeus, Appendix II and VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.2 hvf 2023-04-06 getJupiterLBRData.py created by mkPlanetLBRData.py"
VersionList.append(Version)

def getJupiterLBRData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. (31.1) Julian Millennia

#######################################################################

    # JUPITER Longitude
    sumL0  = 0
    sumL0  +=  59954691.495 * cos(0.00000000000  + 0.00000000000 *T)
    sumL0  +=  9695898.711 * cos(5.06191793105  + 529.69096509460 *T)
    sumL0  +=  573610.145 * cos(1.44406205976  + 7.11354700080 *T)
    sumL0  +=  306389.180 * cos(5.41734729976  + 1059.38193018920 *T)
    sumL0  +=   97178.280 * cos(4.14264708819  + 632.78373931320 *T)
    sumL0  +=   72903.096 * cos(3.64042909255  + 522.57741809380 *T)
    sumL0  +=   64263.986 * cos(3.41145185203  + 103.09277421860 *T)
    sumL0  +=   39806.051 * cos(2.29376744855  + 419.48464387520 *T)
    sumL0  +=   38857.780 * cos(1.27231724860  + 316.39186965660 *T)
    sumL0  +=   27964.622 * cos(1.78454589485  + 536.80451209540 *T)
    sumL0  +=   13589.738 * cos(5.77481031590  + 1589.07289528380 *T)
    sumL0  +=    8246.362 * cos(3.58227961655  + 206.18554843720 *T)
    sumL0  +=    8768.686 * cos(3.63000324417  + 949.17560896980 *T)
    sumL0  +=    7368.057 * cos(5.08101125612  + 735.87651353180 *T)
    sumL0  +=    6263.171 * cos(0.02497643742  + 213.29909543800 *T)
    sumL0  +=    6114.050 * cos(4.51319531666  + 1162.47470440780 *T)
    sumL0  +=    4905.419 * cos(1.32084631684  + 110.20632121940 *T)
    sumL0  +=    5305.283 * cos(1.30671236848  + 14.22709400160 *T)
    sumL0  +=    5305.457 * cos(4.18625053495  + 1052.26838318840 *T)
    sumL0  +=    4647.249 * cos(4.69958109497  + 3.93215326310 *T)
    sumL0  +=    3045.009 * cos(4.31675960318  + 426.59819087600 *T)
    sumL0  +=    2610.001 * cos(1.56667594850  + 846.08283475120 *T)
    sumL0  +=    2028.191 * cos(1.06376547379  + 3.18139373770 *T)
    sumL0  +=    1764.768 * cos(2.14148077766  + 1066.49547719000 *T)
    sumL0  +=    1722.983 * cos(3.88036008872  + 1265.56747862640 *T)
    sumL0  +=    1920.959 * cos(0.97168928755  + 639.89728631400 *T)
    sumL0  +=    1633.217 * cos(3.58201089758  + 515.46387109300 *T)
    sumL0  +=    1431.997 * cos(4.29683690269  + 625.67019231240 *T)
    sumL0  +=     973.278 * cos(4.09764957065  + 95.97922721780 *T)
    sumL0  +=     884.439 * cos(2.43701426123  + 412.37109687440 *T)
    sumL0  +=     732.875 * cos(6.08534113239  + 838.96928775040 *T)
    sumL0  +=     731.072 * cos(3.80591233956  + 1581.95934828300 *T)
    sumL0  +=     691.928 * cos(6.13368222939  + 2118.76386037840 *T)
    sumL0  +=     709.190 * cos(1.29272573658  + 742.99006053260 *T)
    sumL0  +=     614.464 * cos(4.10853496756  + 1478.86657406440 *T)
    sumL0  +=     495.224 * cos(3.75567461379  + 323.50541665740 *T)
    sumL0  +=     581.902 * cos(4.53967717552  + 309.27832265580 *T)
    sumL0  +=     375.657 * cos(4.70299124833  + 1368.66025284500 *T)
    sumL0  +=     389.864 * cos(4.89716105852  + 1692.16566950240 *T)
    sumL0  +=     341.006 * cos(5.71452525783  + 533.62311835770 *T)
    sumL0  +=     330.458 * cos(4.74049819491  + 0.04818410980 *T)
    sumL0  +=     440.854 * cos(2.95818460943  + 454.90936652730 *T)
    sumL0  +=     417.266 * cos(1.03554430161  + 2.44768055480 *T)
    sumL0  +=     244.170 * cos(5.22020878900  + 728.76296653100 *T)
    sumL0  +=     261.540 * cos(1.87652461032  + 0.96320784650 *T)
    sumL0  +=     256.568 * cos(3.72410724159  + 199.07200143640 *T)
    sumL0  +=     261.009 * cos(0.82047246448  + 380.12776796000 *T)
    sumL0  +=     220.382 * cos(1.65115015995  + 543.91805909620 *T)
    sumL0  +=     201.996 * cos(1.80684574186  + 1375.77379984580 *T)
    sumL0  +=     207.327 * cos(1.85461666594  + 525.75881183150 *T)
    sumL0  +=     197.046 * cos(5.29252149016  + 1155.36115740700 *T)
    sumL0  +=     235.141 * cos(1.22693908124  + 909.81873305460 *T)
    sumL0  +=     174.809 * cos(5.90973505276  + 956.28915597060 *T)
    sumL0  +=     149.368 * cos(4.37745104275  + 1685.05212250160 *T)
    sumL0  +=     175.184 * cos(3.22634903433  + 1898.35121793960 *T)
    sumL0  +=     175.191 * cos(3.72966554761  + 942.06206196900 *T)
    sumL0  +=     157.909 * cos(4.36483921766  + 1795.25844372100 *T)
    sumL0  +=     137.871 * cos(1.31797920785  + 1169.58825140860 *T)
    sumL0  +=     117.495 * cos(2.50022140890  + 1596.18644228460 *T)
    sumL0  +=     150.502 * cos(3.90625022622  + 74.78159856730 *T)
    sumL0  +=     116.757 * cos(3.38920921041  + 0.52126486180 *T)
    sumL0  +=     105.895 * cos(4.55439798236  + 526.50957135690 *T)
    sumL0  +=     130.531 * cos(4.16867945489  + 1045.15483618760 *T)
    sumL0  +=     141.445 * cos(3.13568357861  + 491.55792945680 *T)
    sumL0  +=      99.511 * cos(1.42117395747  + 532.87235883230 *T)
    sumL0  +=      96.137 * cos(1.18156870005  + 117.31986822020 *T)
    sumL0  +=      91.758 * cos(0.85756633461  + 1272.68102562720 *T)
    sumL0  +=      87.695 * cos(1.21738140813  + 453.42489381900 *T)
    sumL0  +=      68.507 * cos(2.35242959478  + 2.92076130680 *T)
    sumL0  +=      66.098 * cos(5.34386149468  + 1471.75302706360 *T)
    sumL0  +=      77.401 * cos(4.42676337124  + 39.35687591520 *T)
    sumL0  +=      72.006 * cos(4.23834923691  + 2111.65031337760 *T)
    sumL0  +=      63.406 * cos(4.97665525033  + 0.75075952540 *T)
    sumL0  +=      59.427 * cos(4.11130498612  + 2001.44399215820 *T)
    sumL0  +=      62.481 * cos(0.51211384012  + 220.41264243880 *T)
    sumL0  +=      66.532 * cos(2.98864358135  + 2214.74308759620 *T)
    sumL0  +=      60.194 * cos(4.12628179571  + 4.19278569400 *T)
    sumL0  +=      56.012 * cos(1.15493222602  + 21.34064100240 *T)
    sumL0  +=      52.854 * cos(0.91207215543  + 10.29494073850 *T)
    sumL0  +=      70.297 * cos(5.14180555282  + 835.03713448730 *T)
    sumL0  +=      51.916 * cos(4.10048180020  + 1258.45393162560 *T)
    sumL0  +=      46.442 * cos(4.66531163524  + 5.62907429250 *T)
    sumL0  +=      58.190 * cos(5.86646380344  + 5753.38488489680 *T)
    sumL0  +=      40.103 * cos(4.68801114087  + 0.16005869440 *T)
    sumL0  +=      46.654 * cos(4.79394835282  + 305.34616939270 *T)
    sumL0  +=      39.298 * cos(4.25448423697  + 853.19638175200 *T)
    sumL0  +=      46.042 * cos(5.10983515150  + 4.66586644600 *T)
    sumL0  +=      54.459 * cos(1.57072704127  + 983.11585891360 *T)
    sumL0  +=      38.920 * cos(6.07592905580  + 518.64526483070 *T)
    sumL0  +=      38.450 * cos(2.43836870888  + 433.71173787680 *T)
    sumL0  +=      46.800 * cos(3.54640538283  + 5.41662597140 *T)
    sumL0  +=      41.830 * cos(4.67982493646  + 302.16477565500 *T)
    sumL0  +=      35.920 * cos(2.45088036239  + 430.53034413910 *T)
    sumL0  +=      37.888 * cos(0.21127448431  + 2648.45482547300 *T)
    sumL0  +=      39.190 * cos(1.71835571629  + 11.04570026390 *T)
    sumL0  +=      37.567 * cos(6.19481310233  + 831.85574074960 *T)
    sumL0  +=      35.828 * cos(4.61459907698  + 2008.55753915900 *T)
    sumL0  +=      43.402 * cos(0.14992289081  + 528.20649238630 *T)
    sumL0  +=      31.598 * cos(5.14073450755  + 1788.14489672020 *T)
    sumL0  +=      29.849 * cos(5.34441117167  + 2221.85663459700 *T)
    sumL0  +=      32.811 * cos(5.28907118836  + 88.86568021700 *T)
    sumL0  +=      27.686 * cos(1.85227036207  + 0.21244832110 *T)
    sumL0  +=      25.820 * cos(3.85920882494  + 2317.83586181480 *T)
    sumL0  +=      24.705 * cos(2.63495214991  + 114.13847448250 *T)
    sumL0  +=      33.844 * cos(1.00563073268  + 9683.59458111640 *T)
    sumL0  +=      24.266 * cos(3.82355417268  + 1574.84580128220 *T)
    sumL0  +=      27.111 * cos(2.80845435102  + 18.15924726470 *T)
    sumL0  +=      26.837 * cos(1.77586123775  + 532.13864564940 *T)
    sumL0  +=      26.064 * cos(2.74361318804  + 2531.13495725280 *T)
    sumL0  +=      30.765 * cos(0.42330537728  + 1.48447270830 *T)
    sumL0  +=      30.476 * cos(3.66677894407  + 508.35032409220 *T)
    sumL0  +=      23.282 * cos(3.24372142416  + 984.60033162190 *T)
    sumL0  +=      19.445 * cos(0.52370214471  + 14.97785352700 *T)
    sumL0  +=      19.332 * cos(4.86314494382  + 1361.54670584420 *T)
    sumL0  +=      22.910 * cos(3.84914895064  + 2428.04218303420 *T)
    sumL0  +=      21.617 * cos(6.01696940024  + 1063.31408345230 *T)
    sumL0  +=      20.155 * cos(5.59582008789  + 527.24328453980 *T)
    sumL0  +=      23.732 * cos(2.52766031921  + 494.26624244250 *T)
    sumL0  +=      20.189 * cos(1.01560227681  + 628.85158605010 *T)
    sumL0  +=      15.994 * cos(5.09003530653  + 529.73914920440 *T)
    sumL0  +=      16.134 * cos(5.27095037302  + 142.44965013380 *T)
    sumL0  +=      20.697 * cos(4.03443281612  + 355.74874557180 *T)
    sumL0  +=      21.479 * cos(1.28668134295  + 35.42472265210 *T)
    sumL0  +=      14.964 * cos(4.86039684390  + 2104.53676637680 *T)
    sumL0  +=      17.242 * cos(1.59187913206  + 1439.50969814920 *T)
    sumL0  +=      15.994 * cos(1.89222417794  + 529.64278098480 *T)
    sumL0  +=      17.958 * cos(4.30178016003  + 6.15033915430 *T)
    sumL0  +=      13.279 * cos(2.18943981644  + 1055.44977692610 *T)
    sumL0  +=      14.148 * cos(2.71597731671  + 0.26063243090 *T)
    sumL0  +=      14.689 * cos(0.87944553412  + 99.16062095550 *T)
    sumL0  +=      14.202 * cos(2.41335693735  + 530.65417294110 *T)
    sumL0  +=      15.320 * cos(6.07703092728  + 149.56319713460 *T)
    sumL0  +=      15.832 * cos(4.11682440678  + 636.71589257630 *T)
    sumL0  +=      12.398 * cos(2.61042299578  + 405.25754987360 *T)
    sumL0  +=      16.199 * cos(2.77035044582  + 760.25553592000 *T)
    sumL0  +=      13.665 * cos(3.56039678310  + 217.23124870110 *T)
    sumL0  +=      15.261 * cos(2.81824770887  + 621.73803904930 *T)
    sumL0  +=      14.681 * cos(6.26423732742  + 569.04784100980 *T)
    sumL0  +=      12.529 * cos(1.39077179081  + 7.06536289100 *T)
    sumL0  +=      11.677 * cos(3.60447374272  + 2634.22773147140 *T)
    sumL0  +=      11.603 * cos(4.60461756191  + 7.16173111060 *T)
    sumL0  +=      12.152 * cos(0.24540531919  + 1485.98012106520 *T)
    sumL0  +=      11.347 * cos(2.00818458261  + 1073.60902419080 *T)
    sumL0  +=      11.242 * cos(2.48000947870  + 423.41679713830 *T)
    sumL0  +=      10.942 * cos(5.03602448252  + 458.84151979040 *T)
    sumL0  +=      11.117 * cos(4.04973271023  + 519.39602435610 *T)
    sumL0  +=      12.256 * cos(4.30153222783  + 604.47256366190 *T)
    sumL0  +=      13.149 * cos(2.72189077702  + 1364.72809958190 *T)
    sumL0  +=      10.604 * cos(3.11518747072  + 1.27202438720 *T)
    sumL0  +=       9.874 * cos(1.70200068743  + 1699.27921650320 *T)
    sumL0  +=      10.851 * cos(5.08554552028  + 2324.94940881560 *T)
    sumL0  +=      10.692 * cos(2.51401681528  + 2847.52682690940 *T)
    sumL0  +=      12.640 * cos(4.75572797691  + 528.72775724810 *T)
    sumL0  +=      10.084 * cos(4.05599810206  + 38.13303563780 *T)
    sumL0  +=      11.536 * cos(2.35034215745  + 643.82943957710 *T)
    sumL0  +=      10.247 * cos(3.63479911496  + 2744.43405269080 *T)
    sumL0  +=      10.105 * cos(3.65845333837  + 107.02492748170 *T)
    sumL0  +=      10.121 * cos(1.31482648275  + 1905.46476494040 *T)
    sumL0  +=       9.341 * cos(5.92176693887  + 1148.24761040620 *T)
    sumL0  +=       8.796 * cos(2.77421822809  + 6.59228213900 *T)
    sumL0  +=       8.420 * cos(4.52537756809  + 1677.93857550080 *T)
    sumL0  +=      10.128 * cos(2.09034472544  + 511.53171782990 *T)
    sumL0  +=       8.272 * cos(2.98682673354  + 540.73666535850 *T)
    sumL0  +=       9.753 * cos(1.22438911827  + 32.24332891440 *T)
    sumL0  +=      10.630 * cos(2.07777800288  + 92.04707395470 *T)
    sumL0  +=       7.850 * cos(0.98996894618  + 408.43894361130 *T)
    sumL0  +=       8.811 * cos(3.46911754939  + 1021.24889455140 *T)
    sumL0  +=       7.946 * cos(2.86682926070  + 2125.87740737920 *T)
    sumL0  +=       8.575 * cos(5.29590411702  + 415.55249061210 *T)
    sumL0  +=       7.841 * cos(6.08025056721  + 70.84944530420 *T)
    sumL0  +=       7.706 * cos(1.69832954219  + 8.07675484730 *T)
    sumL0  +=       7.265 * cos(4.65503563919  + 629.60234557550 *T)
    sumL0  +=       7.164 * cos(4.93400217968  + 1056.20053645150 *T)
    sumL0  +=       7.247 * cos(4.61607677560  + 2420.92863603340 *T)
    sumL0  +=       7.753 * cos(2.12871653382  + 33.94024994380 *T)
    sumL0  +=       6.645 * cos(0.45647460873  + 635.96513305090 *T)
    sumL0  +=       9.377 * cos(4.03158388202  + 2810.92146160520 *T)
    sumL0  +=       8.263 * cos(1.23558676139  + 1802.37199072180 *T)
    sumL0  +=       6.341 * cos(0.07278001580  + 202.25339517410 *T)
    sumL0  +=       6.383 * cos(3.54310669809  + 1891.23767093880 *T)
    sumL0  +=       7.902 * cos(2.32510002614  + 230.56457082540 *T)
    sumL0  +=       6.214 * cos(4.54560345237  + 2.70831298570 *T)
    sumL0  +=       7.347 * cos(1.24457237337  + 24.37902238820 *T)
    sumL0  +=       7.451 * cos(3.02719199239  + 330.61896365820 *T)
    sumL0  +=       6.220 * cos(1.77687561489  + 1062.56332392690 *T)
    sumL0  +=       5.674 * cos(5.14132196367  + 746.92221379570 *T)
    sumL0  +=       5.855 * cos(5.42130172896  + 28.31117565130 *T)
    sumL0  +=       5.629 * cos(3.24348217277  + 529.16970023280 *T)
    sumL0  +=       7.652 * cos(0.52813391052  + 672.14061522840 *T)
    sumL0  +=       5.456 * cos(3.34716871364  + 2950.61960112800 *T)
    sumL0  +=       7.127 * cos(1.43497795005  + 6.21977512350 *T)
    sumL0  +=       5.388 * cos(4.90175095580  + 69.15252427480 *T)
    sumL0  +=       5.618 * cos(4.97903783721  + 2641.34127847220 *T)
    sumL0  +=       5.844 * cos(2.95364118152  + 490.33408917940 *T)
    sumL0  +=       4.943 * cos(5.37597740579  + 721.64941953020 *T)
    sumL0  +=       5.062 * cos(4.84282906467  + 31.01948863700 *T)
    sumL0  +=       5.163 * cos(5.07410777073  + 67.66805156650 *T)
    sumL0  +=       4.739 * cos(6.10248862834  + 106.27416795630 *T)
    sumL0  +=       4.879 * cos(0.07095292379  + 78.71375183040 *T)
    sumL0  +=       4.854 * cos(5.63875661096  + 1.69692102940 *T)
    sumL0  +=       5.629 * cos(3.73871604865  + 530.21222995640 *T)
    sumL0  +=       4.471 * cos(4.49152882547  + 505.31194270640 *T)
    sumL0  +=       4.313 * cos(4.79369370451  + 535.10759106600 *T)
    sumL0  +=       4.280 * cos(0.54783823710  + 1.43628859850 *T)
    sumL0  +=       4.453 * cos(0.50551854591  + 524.06189080210 *T)
    sumL0  +=       4.936 * cos(4.82992988255  + 422.66603761290 *T)
    sumL0  +=       4.701 * cos(3.41634869046  + 3060.82592234740 *T)
    sumL0  +=       4.261 * cos(2.67044686458  + 561.93429400900 *T)
    sumL0  +=       4.156 * cos(4.00660658688  + 99.91138048090 *T)
    sumL0  +=       4.561 * cos(2.29650164054  + 3163.91869656600 *T)
    sumL0  +=       4.414 * cos(5.67224020329  + 1464.63948006280 *T)
    sumL0  +=       5.345 * cos(0.31513851830  + 1289.94650101460 *T)
    sumL0  +=       5.269 * cos(3.89116469022  + 191.95845443560 *T)
    sumL0  +=       3.855 * cos(4.28942301453  + 1994.33044515740 *T)
    sumL0  +=       4.210 * cos(5.32763589447  + 2538.24850425360 *T)
    sumL0  +=       3.949 * cos(4.56507101172  + 1382.88734684660 *T)
    sumL0  +=       3.885 * cos(1.56778786810  + 647.01083331480 *T)
    sumL0  +=       4.227 * cos(5.51697599030  + 5223.69391980220 *T)
    sumL0  +=       4.129 * cos(2.81119457666  + 416.30325013750 *T)
    sumL0  +=       3.663 * cos(4.35187510477  + 2737.32050569000 *T)
    sumL0  +=       3.566 * cos(5.48243943375  + 750.10360753340 *T)
    sumL0  +=       4.330 * cos(0.84941756640  + 531.17543780290 *T)
    sumL0  +=       4.093 * cos(0.19980340452  + 525.02509864860 *T)
    sumL0  +=       4.022 * cos(1.92293311337  + 1512.80682400820 *T)
    sumL0  +=       3.400 * cos(6.00302355875  + 1.22384027740 *T)
    sumL0  +=       3.496 * cos(0.31252921473  + 597.35901666110 *T)
    sumL0  +=       3.299 * cos(4.27596694481  + 526.77020378780 *T)
    sumL0  +=       3.226 * cos(2.90455264496  + 963.40270297140 *T)
    sumL0  +=       3.150 * cos(3.81061764181  + 280.96714700450 *T)
    sumL0  +=       4.129 * cos(4.74946631331  + 0.89377187730 *T)
    sumL0  +=       3.840 * cos(1.91064405186  + 378.64329525170 *T)
    sumL0  +=       3.057 * cos(1.65589659685  + 528.94020556920 *T)
    sumL0  +=       3.011 * cos(1.59276337369  + 224.34479570190 *T)
    sumL0  +=       3.196 * cos(5.86588452873  + 4.14460158420 *T)
    sumL0  +=       3.628 * cos(0.07930225897  + 558.00214074590 *T)
    sumL0  +=       2.932 * cos(0.41424445089  + 7.86430652620 *T)
    sumL0  +=       3.316 * cos(2.70211697795  + 532.61172640140 *T)
    sumL0  +=       2.925 * cos(4.47580363425  + 533.88375078860 *T)
    sumL0  +=       3.690 * cos(0.39897023849  + 685.47393735270 *T)
    sumL0  +=       3.223 * cos(2.45833032883  + 960.22130923370 *T)
    sumL0  +=       3.059 * cos(5.32616140812  + 530.44172462000 *T)
    sumL0  +=       3.383 * cos(4.42170370028  + 312.45971639350 *T)
    sumL0  +=       3.320 * cos(2.71417812514  + 495.75071515080 *T)
    sumL0  +=       2.697 * cos(5.23146633437  + 739.80866679490 *T)
    sumL0  +=       3.590 * cos(2.30999595873  + 908.33426034630 *T)
    sumL0  +=       3.677 * cos(5.07337955976  + 73.29712585900 *T)
    sumL0  +=       2.618 * cos(3.09118499149  + 3267.01147078460 *T)
    sumL0  +=       2.796 * cos(2.98942316119  + 483.22054217860 *T)
    sumL0  +=       3.398 * cos(3.29598270278  + 911.30320576290 *T)
    sumL0  +=       3.352 * cos(1.44391979336  + 593.42686339800 *T)
    sumL0  +=       2.563 * cos(3.35080110279  + 2207.62954059540 *T)
    sumL0  +=       2.553 * cos(0.36892288645  + 1048.33622992530 *T)
    sumL0  +=       2.620 * cos(3.82769874340  + 520.12973753900 *T)
    sumL0  +=       3.356 * cos(1.08315053878  + 46.47042291600 *T)
    sumL0  +=       3.035 * cos(5.52230028113  + 618.55664531160 *T)
    sumL0  +=       3.397 * cos(3.83084746522  + 210.11770170030 *T)
    sumL0  +=       2.497 * cos(0.47917884538  + 945.24345570670 *T)
    sumL0  +=       2.341 * cos(5.87941292649  + 2751.54759969160 *T)
    sumL0  +=       2.656 * cos(0.49713061045  + 1057.89745748090 *T)
    sumL0  +=       2.581 * cos(0.03759881914  + 1.64453140270 *T)
    sumL0  +=       2.900 * cos(2.50019054587  + 525.49817940060 *T)
    sumL0  +=       3.153 * cos(2.30900986177  + 457.61767951300 *T)
    sumL0  +=       2.201 * cos(3.94367109739  + 31.49256938900 *T)
    sumL0  +=       2.381 * cos(6.19252134885  + 327.43756992050 *T)
    sumL0  +=       2.458 * cos(0.65614291954  + 9153.90361602180 *T)
    sumL0  +=       2.111 * cos(5.61905648764  + 16.46232623530 *T)
    sumL0  +=       2.130 * cos(3.75880734109  + 724.83081326790 *T)
    sumL0  +=       2.406 * cos(2.29315649755  + 195.13984817330 *T)
    sumL0  +=       2.166 * cos(5.43262641046  + 534.35683154060 *T)
    sumL0  +=       2.057 * cos(1.49875151278  + 551.03160609700 *T)
    sumL0  +=       2.676 * cos(5.06374981112  + 456.39383923560 *T)
    sumL0  +=       2.078 * cos(5.28920097886  + 76.26607127560 *T)
    sumL0  +=       2.261 * cos(5.38117230692  + 1781.03134971940 *T)
    sumL0  +=       2.356 * cos(0.67392574097  + 227.52618943960 *T)
    sumL0  +=       2.240 * cos(3.18006978517  + 3377.21779200400 *T)
    sumL0  +=       2.183 * cos(3.08384250950  + 524.27433912320 *T)
    sumL0  +=       2.119 * cos(2.70107659927  + 387.24131496080 *T)
    sumL0  +=       2.056 * cos(4.82779196994  + 2957.73314812880 *T)
    sumL0  +=       2.116 * cos(6.20263841494  + 209.36694217490 *T)
    sumL0  +=       2.712 * cos(3.18157754631  + 1474.67378837040 *T)
    sumL0  +=       2.127 * cos(1.24424012514  + 539.98590583310 *T)
    sumL0  +=       2.424 * cos(3.57595925853  + 953.10776223290 *T)
    sumL0  +=       1.947 * cos(1.94468082546  + 529.53090640020 *T)
    sumL0  +=       1.896 * cos(4.01406242800  + 2310.72231481400 *T)
    sumL0  +=       1.935 * cos(4.10051493950  + 3053.71237534660 *T)
    sumL0  +=       2.056 * cos(6.27074148550  + 245.54242435240 *T)
    sumL0  +=       2.108 * cos(3.22886474225  + 252.65597135320 *T)
    sumL0  +=       2.596 * cos(2.77467278614  + 177.87437278590 *T)
    sumL0  +=       1.919 * cos(3.14834694111  + 381.61224066830 *T)
    sumL0  +=       2.217 * cos(1.92368906925  + 535.91074021810 *T)
    sumL0  +=       1.947 * cos(5.03751780002  + 529.85102378900 *T)
    sumL0  +=       2.025 * cos(4.82814272957  + 17.26547538740 *T)
    sumL0  +=       1.945 * cos(2.10611582568  + 3480.31056622260 *T)
    sumL0  +=       1.899 * cos(0.05104263891  + 560.71045373160 *T)
    sumL0  +=       2.221 * cos(0.58365090630  + 3178.14579056760 *T)
    sumL0  +=       2.271 * cos(1.67360565619  + 731.94436026870 *T)
    sumL0  +=       1.706 * cos(5.40277333462  + 20.44686912510 *T)
    sumL0  +=       2.295 * cos(4.20863103004  + 1038.04128918680 *T)
    sumL0  +=       2.218 * cos(3.65982280555  + 282.45161971280 *T)
    sumL0  +=       2.181 * cos(4.87369503022  + 535.32003938710 *T)
    sumL0  +=       1.745 * cos(1.34021867874  + 25.12978191360 *T)
    sumL0  +=       1.601 * cos(3.92730015840  + 17.52610781830 *T)
    sumL0  +=       1.651 * cos(0.63598292839  + 17.40848773930 *T)
    sumL0  +=       1.826 * cos(0.31592311031  + 124.43341522100 *T)
    sumL0  +=       2.041 * cos(0.15617294873  + 598.84348936940 *T)
    sumL0  +=       1.494 * cos(3.81418025130  + 319.57326339430 *T)
    sumL0  +=       1.551 * cos(5.25201528605  + 437.64389113990 *T)
    sumL0  +=       1.852 * cos(2.36130812462  + 37.87240320690 *T)
    sumL0  +=       1.466 * cos(1.72926380881  + 59.80374504030 *T)
    sumL0  +=       1.417 * cos(5.82273267086  + 81.75213321620 *T)
    sumL0  +=       1.430 * cos(1.17528806260  + 440.82528487760 *T)
    sumL0  +=       1.906 * cos(4.06896022692  + 1819.63746610920 *T)
    sumL0  +=       1.397 * cos(0.26383366743  + 50.40257617910 *T)
    sumL0  +=       1.756 * cos(2.32977483716  + 938.12990870590 *T)
    sumL0  +=       1.487 * cos(2.24866746540  + 10.03430830760 *T)
    sumL0  +=       1.368 * cos(3.56691602771  + 1514.29129671650 *T)
    sumL0  +=       1.400 * cos(4.84502200703  + 295.05122865420 *T)
    sumL0  +=       1.344 * cos(2.20177702122  + 529.90341341570 *T)
    sumL0  +=       1.464 * cos(1.42648716568  + 1158.54255114470 *T)
    sumL0  +=       1.341 * cos(1.15693423225  + 2435.15573003500 *T)
    sumL0  +=       1.786 * cos(5.44716330146  + 2854.64037391020 *T)
    sumL0  +=       1.677 * cos(6.22875777048  + 833.55266177900 *T)
    sumL0  +=       1.471 * cos(4.80574535807  + 696.51963761660 *T)
    sumL0  +=       1.436 * cos(1.45810957330  + 537.76771994190 *T)
    sumL0  +=       1.657 * cos(0.02890651793  + 138.51749687070 *T)
    sumL0  +=       1.300 * cos(3.14074420421  + 547.85021235930 *T)
    sumL0  +=       1.343 * cos(6.14827138025  + 988.53248488500 *T)
    sumL0  +=       1.344 * cos(4.78042160426  + 529.47851677350 *T)
    sumL0  +=       1.234 * cos(2.83294330979  + 3583.40334044120 *T)
    sumL0  +=       1.651 * cos(2.12056447005  + 1061.82961074400 *T)
    sumL0  +=       1.479 * cos(0.24646493075  + 1593.00504854690 *T)
    sumL0  +=       1.413 * cos(3.07444632745  + 6283.07584999140 *T)
    sumL0  +=       1.246 * cos(5.94882321661  + 1056.93424963440 *T)
    sumL0  +=       1.225 * cos(1.95642397635  + 1969.20066324380 *T)
    sumL0  +=       1.388 * cos(2.87749576073  + 1023.95720753710 *T)
    sumL0  +=       1.263 * cos(3.46181945031  + 40.84134862350 *T)
    sumL0  +=       1.325 * cos(4.15429781246  + 916.93228005540 *T)
    sumL0  +=       1.477 * cos(5.26691818477  + 810.65811209910 *T)
    sumL0  +=       1.165 * cos(4.65528125418  + 944.98282327580 *T)
    sumL0  +=       1.137 * cos(2.48561382158  + 2.00573757010 *T)
    sumL0  +=       1.118 * cos(3.80747957482  + 7.00167241620 *T)
    sumL0  +=       1.138 * cos(5.11611532241  + 885.43971066640 *T)
    sumL0  +=       1.131 * cos(1.54599459004  + 775.23338944700 *T)
    sumL0  +=       1.477 * cos(4.69742954455  + 630.33605875840 *T)
    sumL0  +=       1.252 * cos(1.34316620527  + 739.05790726950 *T)
    sumL0  +=       1.273 * cos(5.19070939905  + 2097.42321937600 *T)
    sumL0  +=       1.446 * cos(5.54999644374  + 43.28902917830 *T)
    sumL0  +=       1.344 * cos(4.75897665313  + 1166.40685767090 *T)
    sumL0  +=       1.101 * cos(4.56997613488  + 3274.12501778540 *T)
    sumL0  +=       1.376 * cos(3.60998729004  + 415.29185818120 *T)
    sumL0  +=       1.437 * cos(6.22410093972  + 155.78297225810 *T)
    sumL0  +=       1.167 * cos(4.09497264272  + 203.00415469950 *T)
    sumL0  +=       1.237 * cos(4.41132627005  + 292.01284726840 *T)
    sumL0  +=       1.077 * cos(2.57045229823  + 25.27279426550 *T)
    sumL0  +=       1.341 * cos(0.49262296655  + 635.23141986800 *T)
    sumL0  +=       1.209 * cos(3.36289125536  + 521.61421024730 *T)
    sumL0  +=       1.030 * cos(1.81822316284  + 465.95506679120 *T)
    sumL0  +=       1.002 * cos(3.21720955284  + 2524.02141025200 *T)
    sumL0  +=       1.338 * cos(1.26054917773  + 902.70518605380 *T)
    sumL0  +=       1.037 * cos(3.87858871885  + 3370.10424500320 *T)
    sumL0  +=       1.224 * cos(0.09219976028  + 824.74219374880 *T)
    sumL0  +=       1.255 * cos(3.04675952762  + 447.79581952650 *T)
    sumL0  +=       0.991 * cos(4.16587903812  + 632.83192342300 *T)
    sumL0  +=       0.975 * cos(3.80216680539  + 2627.11418447060 *T)
    sumL0  +=       1.061 * cos(5.60184374277  + 732.69511979410 *T)
    sumL0  +=       1.049 * cos(2.94931080683  + 3693.60966166060 *T)
    sumL0  +=       0.984 * cos(0.98260254313  + 632.73555520340 *T)
    sumL0  +=       1.050 * cos(2.20935815967  + 7.22542158540 *T)
    sumL0  +=       0.996 * cos(5.41921062583  + 1059.43011429900 *T)
    sumL0  +=       0.961 * cos(0.87315283361  + 544.66881862160 *T)
    sumL0  +=       1.175 * cos(3.09093466406  + 1894.41906467650 *T)
    sumL0  +=       1.049 * cos(5.81616384906  + 26.82670294300 *T)
    sumL0  +=       1.161 * cos(0.01274801567  + 850.01498801430 *T)
    sumL0  +=       1.109 * cos(3.63294273717  + 306.83064210100 *T)
    sumL0  +=       1.077 * cos(0.95716576092  + 608.40471692500 *T)
    sumL0  +=       1.288 * cos(4.23019288942  + 1215.16490244730 *T)
    sumL0  +=       1.060 * cos(3.85856787901  + 631.82053146670 *T)
    sumL0  +=       1.251 * cos(6.15889818604  + 462.02291352810 *T)
    sumL0  +=       1.165 * cos(3.50653563773  + 8.59801970910 *T)
    sumL0  +=       0.933 * cos(4.62559759882  + 1049.08698945070 *T)
    sumL0  +=       1.035 * cos(1.30805283339  + 633.74694715970 *T)
    sumL0  +=       1.238 * cos(2.21195391602  + 25558.21217647960 *T)
    sumL0  +=       1.240 * cos(2.27960685992  + 6.90109867970 *T)
    sumL0  +=       0.942 * cos(4.14526324371  + 945.99421523210 *T)
    sumL0  +=       0.927 * cos(6.10893117637  + 514.71311156760 *T)
    sumL0  +=       0.914 * cos(6.17656044376  + 952.35700270750 *T)
    sumL0  +=       0.893 * cos(4.27448748055  + 0.63313944640 *T)
    sumL0  +=       1.045 * cos(1.64682770236  + 565.11568774670 *T)
    sumL0  +=       0.903 * cos(1.94250156640  + 3796.70243587920 *T)
    sumL0  +=       1.162 * cos(5.51229668479  + 2.96894541660 *T)
    sumL0  +=       0.901 * cos(3.03568112112  + 460.53844081980 *T)
    sumL0  +=       0.903 * cos(2.24012822393  + 523.54062594030 *T)
    sumL0  +=       1.060 * cos(5.28027224466  + 3171.03224356680 *T)
    sumL0  +=       1.064 * cos(0.99330150801  + 320.32402291970 *T)
    sumL0  +=       0.970 * cos(4.56607888439  + 429.04587143080 *T)
    sumL0  +=       1.071 * cos(4.33203090957  + 610.69233878540 *T)
    sumL0  +=       0.865 * cos(0.21831429230  + 1098.73880610440 *T)
    sumL0  +=       0.865 * cos(2.82123742108  + 1060.34513803570 *T)
    sumL0  +=       0.882 * cos(4.80076824948  + 384.05992122310 *T)
    sumL0  +=       0.959 * cos(5.45468005818  + 451.94042111070 *T)
    sumL0  +=       1.042 * cos(5.79270325150  + 303.86169668440 *T)
    sumL0  +=       0.784 * cos(1.85150700827  + 313.21047591890 *T)
    sumL0  +=       1.083 * cos(1.40526460812  + 72.07328558160 *T)
    sumL0  +=       0.782 * cos(3.03559242565  + 5.84152261360 *T)
    sumL0  +=       0.854 * cos(1.22236205478  + 611.44309831080 *T)
    sumL0  +=       0.996 * cos(2.22139794743  + 1059.33374607940 *T)
    sumL0  +=       0.719 * cos(4.92550252164  + 421.93232443000 *T)
    sumL0  +=       0.953 * cos(3.98347050083  + 836.52160719560 *T)
    sumL0  +=       0.822 * cos(4.49679856387  + 10213.28554621100 *T)
    sumL0  +=       0.707 * cos(2.16473400319  + 2228.97018159780 *T)
    sumL0  +=       0.715 * cos(4.62515255534  + 385.54439393140 *T)
    sumL0  +=       0.737 * cos(4.63776694324  + 1134.16352875650 *T)
    sumL0  +=       0.730 * cos(1.87179326186  + 153.49535039770 *T)
    sumL0  +=       0.709 * cos(2.93132115910  + 417.03696332040 *T)
    sumL0  +=       0.926 * cos(1.77006317007  + 2332.06295581640 *T)
    sumL0  +=       0.864 * cos(3.03246275970  + 1041.22268292450 *T)
    sumL0  +=       0.708 * cos(6.01601101389  + 395.10562148700 *T)
    sumL0  +=       0.935 * cos(6.01864676296  + 173.94221952280 *T)
    sumL0  +=       0.695 * cos(1.39408383356  + 432.01481684740 *T)
    sumL0  +=       0.687 * cos(3.06548397586  + 529.95159752550 *T)
    sumL0  +=       0.677 * cos(3.58357527210  + 244.31858407500 *T)
    sumL0  +=       0.850 * cos(5.46114025921  + 41.05379694460 *T)
    sumL0  +=       0.817 * cos(4.65315342412  + 535.84130424890 *T)
    sumL0  +=       0.652 * cos(0.44173759183  + 1201.83158032300 *T)
    sumL0  +=       0.711 * cos(0.96283289310  + 373.01422095920 *T)
    sumL0  +=       0.665 * cos(1.03244633471  + 623.22251175760 *T)
    sumL0  +=       0.643 * cos(5.05335060049  + 522.62560220360 *T)
    sumL0  +=       0.639 * cos(4.22718483639  + 25.86349509650 *T)
    sumL0  +=       0.718 * cos(5.07576900710  + 1058.41872234270 *T)
    sumL0  +=       0.664 * cos(2.43728454444  + 1585.14074202070 *T)
    sumL0  +=       0.833 * cos(1.49468440213  + 563.63121503840 *T)
    sumL0  +=       0.760 * cos(4.34849823663  + 100.64509366380 *T)
    sumL0  +=       0.633 * cos(4.31796718640  + 3590.51688744200 *T)
    sumL0  +=       0.629 * cos(6.23431126402  + 679.25416222920 *T)
    sumL0  +=       0.617 * cos(2.68075016456  + 3899.79521009780 *T)
    sumL0  +=       0.646 * cos(2.88581188015  + 13.49338081870 *T)
    sumL0  +=       0.768 * cos(3.18498076120  + 1151.42900414390 *T)
    sumL0  +=       0.731 * cos(5.86653168561  + 501.37978944330 *T)
    sumL0  +=       0.652 * cos(0.82865771780  + 2015.67108615980 *T)
    sumL0  +=       0.796 * cos(5.36663489938  + 420.96911658350 *T)
    sumL0  +=       0.647 * cos(4.74965662438  + 567.82400073240 *T)
    sumL0  +=       0.845 * cos(1.69406147722  + 1744.85586754190 *T)
    sumL0  +=       0.802 * cos(5.79824707751  + 981.63138620530 *T)
    sumL0  +=       0.764 * cos(5.05232933368  + 827.92358748650 *T)
    sumL0  +=       0.604 * cos(5.11265182908  + 1159.29331067010 *T)
    sumL0  +=       0.682 * cos(3.68248136835  + 2281.23049651060 *T)
    sumL0  +=       0.740 * cos(0.74512356954  + 1261.63532536330 *T)
    sumL0  +=       0.666 * cos(2.06624389616  + 27.08733537390 *T)
    sumL0  +=       0.652 * cos(4.92932795958  + 2413.81508903260 *T)
    sumL0  +=       0.559 * cos(0.17558868481  + 63.73589830340 *T)
    sumL0  +=       0.577 * cos(3.82752312276  + 1550.93985964600 *T)
    sumL0  +=       0.727 * cos(1.05835550856  + 490.07345674850 *T)
    sumL0  +=       0.574 * cos(3.61492119092  + 3686.49611465980 *T)
    sumL0  +=       0.732 * cos(5.93179840659  + 42.53826965290 *T)
    sumL0  +=       0.606 * cos(2.71411884300  + 1173.52040467170 *T)
    sumL0  +=       0.633 * cos(4.21720828607  + 166.82867252200 *T)
    sumL0  +=       0.687 * cos(3.91671464962  + 529.43033266370 *T)
    sumL0  +=       0.570 * cos(2.73551750122  + 4010.00153131720 *T)
    sumL0  +=       0.552 * cos(2.36967119362  + 1603.29998928540 *T)
    sumL0  +=       0.600 * cos(1.82659364395  + 522.52923398400 *T)
    sumL0  +=       0.558 * cos(5.09099246601  + 1354.43315884340 *T)
    sumL0  +=       0.519 * cos(6.11952999304  + 366.79444583570 *T)
    sumL0  +=       0.719 * cos(0.85722557905  + 362.86229257260 *T)
    sumL0  +=       0.518 * cos(2.03954064144  + 418.52143602870 *T)
    sumL0  +=       0.515 * cos(3.51750445111  + 528.41894070740 *T)
    sumL0  +=       0.515 * cos(3.47930063838  + 103.14095832840 *T)
    sumL0  +=       0.550 * cos(5.77676837730  + 420.44785172170 *T)
    sumL0  +=       0.702 * cos(3.67952126446  + 1279.79457262800 *T)
    sumL0  +=       0.550 * cos(0.61451088395  + 104.05598206510 *T)
    sumL0  +=       0.495 * cos(2.41738205536  + 179.35884549420 *T)
    sumL0  +=       0.513 * cos(0.29823688044  + 103.04459010880 *T)
    sumL0  +=       0.537 * cos(5.47946238724  + 771.30123618390 *T)
    sumL0  +=       0.507 * cos(3.08777345288  + 1357.61455258110 *T)
    sumL0  +=       0.495 * cos(4.95362659160  + 536.85269620520 *T)
    sumL0  +=       0.681 * cos(4.56294416261  + 112.65400177420 *T)
    sumL0  +=       0.500 * cos(3.15631977489  + 1070.42763045310 *T)
    sumL0  +=       0.484 * cos(0.79038835602  + 28.45418800320 *T)
    sumL0  +=       0.529 * cos(5.46978501034  + 419.43645976540 *T)
    sumL0  +=       0.597 * cos(4.98058295172  + 1251.34038462480 *T)
    sumL0  +=       0.492 * cos(3.96066546484  + 1269.49963188950 *T)
    sumL0  +=       0.482 * cos(3.60167662490  + 2943.50605412720 *T)
    sumL0  +=       0.630 * cos(6.16496640092  + 105.54045477340 *T)
    sumL0  +=       0.480 * cos(0.86786400621  + 35.21227433100 *T)
    sumL0  +=       0.516 * cos(5.97528782923  + 3067.93946934820 *T)
    sumL0  +=       0.586 * cos(5.48467997697  + 56.62235130260 *T)
    sumL0  +=       0.502 * cos(1.43671788959  + 469.88722005430 *T)
    sumL0  +=       0.473 * cos(2.28007170041  + 2042.49778910280 *T)
    sumL0  +=       0.565 * cos(1.90952569252  + 107.28555991260 *T)
    sumL0  +=       0.452 * cos(3.13938145287  + 934.94851496820 *T)
    sumL0  +=       0.605 * cos(1.65413715574  + 761.74000862830 *T)
    sumL0  +=       0.443 * cos(5.46282223686  + 135.33610313300 *T)
    sumL0  +=       0.580 * cos(2.06327501551  + 493.04240216510 *T)
    sumL0  +=       0.540 * cos(1.73777995970  + 536.75632798560 *T)
    sumL0  +=       0.432 * cos(0.27167052107  + 93.53154666300 *T)
    sumL0  +=       0.515 * cos(3.46469417437  + 530.96298948180 *T)
    sumL0  +=       0.440 * cos(5.28884782489  + 497.44763618020 *T)
    sumL0  +=       0.487 * cos(5.78767525063  + 12036.46073488820 *T)
    sumL0  +=       0.452 * cos(2.57855172248  + 1254.52177836250 *T)
    sumL0  +=       0.427 * cos(3.21032629463  + 2840.41327990860 *T)
    sumL0  +=       0.414 * cos(1.54298025443  + 115.62294719080 *T)
    sumL0  +=       0.424 * cos(0.12699448931  + 1268.74887236410 *T)
    sumL0  +=       0.411 * cos(3.12424023238  + 536.28324723360 *T)
    sumL0  +=       0.452 * cos(1.00194596383  + 113.38771495710 *T)
    sumL0  +=       0.419 * cos(0.81834479225  + 1165.65609814550 *T)
    sumL0  +=       0.490 * cos(4.72785081986  + 277.03499374140 *T)
    sumL0  +=       0.434 * cos(0.36146539146  + 1304.92435454160 *T)
    sumL0  +=       0.401 * cos(5.70326543719  + 1127.04998175570 *T)
    sumL0  +=       0.461 * cos(3.26462894820  + 102.12956637210 *T)
    sumL0  +=       0.533 * cos(2.54951615753  + 141.22580985640 *T)
    sumL0  +=       0.413 * cos(4.38801694479  + 6151.53388830500 *T)
    sumL0  +=       0.415 * cos(1.68861617902  + 391.17346822390 *T)
    sumL0  +=       0.385 * cos(1.69092319074  + 4113.09430553580 *T)
    sumL0  +=       0.450 * cos(5.49339192735  + 602.98809095360 *T)
    sumL0  +=       0.499 * cos(3.80738617353  + 81.00137369080 *T)
    sumL0  +=       0.454 * cos(0.10952919733  + 600.54041039880 *T)
    sumL0  +=       0.377 * cos(6.25375060718  + 913.75088631770 *T)
    sumL0  +=       0.453 * cos(3.86104865567  + 758.77106321170 *T)
    sumL0  +=       0.401 * cos(4.44475618337  + 990.22940591440 *T)
    sumL0  +=       0.407 * cos(5.13442416563  + 3487.42411322340 *T)
    sumL0  +=       0.435 * cos(3.76103358490  + 523.09868295560 *T)
    sumL0  +=       0.425 * cos(3.22287851959  + 2655.56837247380 *T)
    sumL0  +=       0.365 * cos(5.16456645463  + 4694.00295470760 *T)
    sumL0  +=       0.454 * cos(1.63325197950  + 976.00231191280 *T)
    sumL0  +=       0.406 * cos(2.72102389267  + 1438.02522544090 *T)
    sumL0  +=       0.349 * cos(3.59598366422  + 1058.86066532740 *T)
    sumL0  +=       0.354 * cos(0.62136331420  + 498.67147645760 *T)
    sumL0  +=       0.383 * cos(5.09229089574  + 539.25219265020 *T)
    sumL0  +=       0.380 * cos(3.92653231573  + 561.18353448360 *T)
    sumL0  +=       0.339 * cos(4.12175871949  + 3906.90875709860 *T)
    sumL0  +=       0.458 * cos(3.42556794767  + 121.25202148330 *T)
    sumL0  +=       0.427 * cos(3.61285264910  + 860.30992875280 *T)
    sumL0  +=       0.424 * cos(4.72757252331  + 1366.21257229020 *T)
    sumL0  +=       0.328 * cos(4.55286002816  + 1696.09782276550 *T)
    sumL0  +=       0.324 * cos(4.23685005210  + 642.34496686880 *T)
    sumL0  +=       0.395 * cos(3.26282558955  + 484.44438245600 *T)
    sumL0  +=       0.330 * cos(6.05223507989  + 215.74677599280 *T)
    sumL0  +=       0.318 * cos(2.02072800070  + 2964.84669512960 *T)
    sumL0  +=       0.417 * cos(0.20173093597  + 842.90144101350 *T)
    sumL0  +=       0.408 * cos(0.45800247268  + 1578.02719501990 *T)
    sumL0  +=       0.342 * cos(6.15347077985  + 1371.84164658270 *T)
    sumL0  +=       0.310 * cos(1.97259286255  + 754.03576079650 *T)
    sumL0  +=       0.340 * cos(2.77813018312  + 3.52311834900 *T)
    sumL0  +=       0.333 * cos(2.91352254678  + 576.16138801060 *T)
    sumL0  +=       0.324 * cos(0.32544817254  + 586.31331639720 *T)
    sumL0  +=       0.302 * cos(2.08708848849  + 526.98265210890 *T)
    sumL0  +=       0.363 * cos(4.70567113230  + 2730.20695868920 *T)
    sumL0  +=       0.300 * cos(0.94464473068  + 1432.39615114840 *T)
    sumL0  +=       0.352 * cos(5.75013621801  + 806.72595883600 *T)
    sumL0  +=       0.296 * cos(3.97807312133  + 2043.98226181110 *T)
    sumL0  +=       0.295 * cos(2.35257797599  + 4216.18707975440 *T)
    sumL0  +=       0.309 * cos(2.49768755925  + 4326.39340097380 *T)
    sumL0  +=       0.306 * cos(3.35876843257  + 2424.11002977110 *T)
    sumL0  +=       0.300 * cos(4.94288858368  + 1379.70595310890 *T)
    sumL0  +=       0.336 * cos(4.49193455535  + 1585.89150154610 *T)
    sumL0  +=       0.402 * cos(2.04684001796  + 842.15068148810 *T)
    sumL0  +=       0.312 * cos(4.59043534747  + 188.92007304980 *T)
    sumL0  +=       0.346 * cos(5.19792097706  + 523.47118997110 *T)
    sumL0  +=       0.380 * cos(1.67961600066  + 36.64856292950 *T)
    sumL0  +=       0.338 * cos(1.32014513725  + 148.07872442630 *T)
    sumL0  +=       0.391 * cos(4.82224015188  + 1012.91150727320 *T)
    sumL0  +=       0.285 * cos(3.43655052437  + 1053.96530421780 *T)
    sumL0  +=       0.332 * cos(2.02575636311  + 1091.62525910360 *T)
    sumL0  +=       0.282 * cos(5.78865321890  + 1064.04779663520 *T)
    sumL0  +=       0.282 * cos(0.39153852422  + 207.67002114550 *T)
    sumL0  +=       0.280 * cos(3.80196391678  + 298.23262239190 *T)
    sumL0  +=       0.387 * cos(6.26819309990  + 1141.13406340540 *T)
    sumL0  +=       0.349 * cos(4.09121908199  + 1059.90319505100 *T)
    sumL0  +=       0.320 * cos(0.39871942000  + 2122.69601364150 *T)
    sumL0  +=       0.327 * cos(4.76503823073  + 134.58534360760 *T)
    sumL0  +=       0.283 * cos(3.90409016441  + 127.47179660680 *T)
    sumL0  +=       0.301 * cos(4.30291951219  + 299.12639426920 *T)
    sumL0  +=       0.322 * cos(2.48251052680  + 1065.60170531270 *T)
    sumL0  +=       0.297 * cos(2.40814103509  + 1591.52057583860 *T)
    sumL0  +=       0.286 * cos(5.85849626574  + 172.24529849340 *T)
    sumL0  +=       0.285 * cos(4.55845472479  + 1446.62324515000 *T)
    sumL0  +=       0.270 * cos(4.08342186112  + 1578.77795454530 *T)
    sumL0  +=       0.362 * cos(1.06148806683  + 181.80652604900 *T)
    sumL0  +=       0.335 * cos(4.51094500655  + 2349.32843120380 *T)
    sumL0  +=       0.347 * cos(0.62281394535  + 1542.60247236780 *T)
    sumL0  +=       0.275 * cos(3.38473403113  + 4002.88798431640 *T)
    sumL0  +=       0.255 * cos(1.52357936497  + 1688.23351623930 *T)
    sumL0  +=       0.276 * cos(4.32192160071  + 1912.57831194120 *T)
    sumL0  +=       0.253 * cos(2.40482338279  + 97.67614824720 *T)
    sumL0  +=       0.248 * cos(4.45058246237  + 1688.98427576470 *T)
    sumL0  +=       0.300 * cos(3.07435583442  + 1902.28337120270 *T)
    sumL0  +=       0.257 * cos(4.79180478086  + 1670.82502850000 *T)
    sumL0  +=       0.319 * cos(1.34244222683  + 1288.46202830630 *T)
    sumL0  +=       0.245 * cos(4.01852686769  + 1567.73225428140 *T)
    sumL0  +=       0.278 * cos(0.25406312148  + 874.39401040250 *T)
    sumL0  +=       0.324 * cos(5.57824969423  + 1670.07426897460 *T)
    sumL0  +=       0.300 * cos(4.67161812947  + 1329.30337692980 *T)
    sumL0  +=       0.241 * cos(0.01789818312  + 1586.62521472900 *T)
    sumL0  +=       0.295 * cos(5.86996114913  + 2804.23779773110 *T)
    sumL0  +=       0.317 * cos(3.17967272487  + 1020.02505427400 *T)
    sumL0  +=       0.238 * cos(4.97765946754  + 351.81659230870 *T)
    sumL0  +=       0.302 * cos(1.20236375616  + 232.04904353370 *T)
    sumL0  +=       0.301 * cos(5.53432687957  + 2274.54683263650 *T)
    sumL0  +=       0.286 * cos(2.41008592059  + 2545.36205125440 *T)
    sumL0  +=       0.294 * cos(2.01783542485  + 313.94418910180 *T)
    sumL0  +=       0.292 * cos(2.12690999284  + 1592.25428902150 *T)
    sumL0  +=       0.250 * cos(2.31712163679  + 632.26247445140 *T)
    sumL0  +=       0.238 * cos(5.06557054569  + 3803.81598288000 *T)
    sumL0  +=       0.226 * cos(0.05916712753  + 1518.22344997960 *T)
    sumL0  +=       0.235 * cos(0.16574304942  + 137.03302416240 *T)
    sumL0  +=       0.298 * cos(2.99720233431  + 1467.82087380050 *T)
    sumL0  +=       0.286 * cos(5.08357076653  + 774.00954916960 *T)
    sumL0  +=       0.246 * cos(2.81685822336  + 633.30500417500 *T)
    sumL0  +=       0.269 * cos(4.93023426152  + 151.04766984290 *T)
    sumL0  +=       0.228 * cos(6.13118739321  + 3281.23856478620 *T)
    sumL0  +=       0.228 * cos(1.22066024988  + 700.45179087970 *T)
    sumL0  +=       0.239 * cos(0.71695698501  + 1276.61317889030 *T)
    sumL0  +=       0.289 * cos(6.08263862565  + 3384.33133900480 *T)
    sumL0  +=       0.218 * cos(2.90308501961  + 85.82729883120 *T)
    sumL0  +=       0.283 * cos(6.28058228271  + 71.81265315070 *T)
    sumL0  +=       0.271 * cos(6.01605074549  + 170.76082578510 *T)
    sumL0  +=       0.221 * cos(0.99914179141  + 1053.75285589670 *T)
    sumL0  +=       0.218 * cos(1.50681393471  + 1087.69310584050 *T)
    sumL0  +=       0.223 * cos(3.39126063354  + 3259.89792378380 *T)
    sumL0  +=       0.229 * cos(1.19373202707  + 1060.86640289750 *T)
    sumL0  +=       0.264 * cos(3.93467945263  + 1363.24362687360 *T)
    sumL0  +=       0.228 * cos(5.04188376116  + 1064.79855616060 *T)
    sumL0  +=       0.295 * cos(2.15253086390  + 6386.16862421000 *T)
    sumL0  +=       0.214 * cos(3.85961180377  + 4223.30062675520 *T)
    sumL0  +=       0.218 * cos(0.79681703388  + 1909.39691820350 *T)
    sumL0  +=       0.212 * cos(4.11706418218  + 269.92144674060 *T)
    sumL0  +=       0.264 * cos(5.81676406517  + 77.96299230500 *T)
    sumL0  +=       0.256 * cos(5.65978708108  + 799.61241183520 *T)
    sumL0  +=       0.242 * cos(6.25078283449  + 1621.31622419820 *T)
    sumL0  +=       0.235 * cos(2.20668997852  + 1570.91364801910 *T)
    sumL0  +=       0.212 * cos(2.88214546012  + 1674.00642223770 *T)
    sumL0  +=       0.206 * cos(1.59586787037  + 4429.48617519240 *T)
    sumL0  +=       0.208 * cos(2.31366614282  + 878.32616366560 *T)
    sumL0  +=       0.213 * cos(0.30373338388  + 8624.21265092720 *T)
    sumL0  +=       0.223 * cos(4.88419887133  + 1035.00290780100 *T)
    sumL0  +=       0.279 * cos(3.65173543621  + 84.93352695390 *T)
    sumL0  +=       0.210 * cos(4.08825553401  + 203.73786788240 *T)
    sumL0  +=       0.214 * cos(4.63498396475  + 812.14258480740 *T)
    sumL0  +=       0.258 * cos(1.73501688450  + 1887.30551767570 *T)
    sumL0  +=       0.210 * cos(4.51798082710  + 1262.38608488870 *T)
    sumL0  +=       0.252 * cos(5.69246905091  + 104.57724692690 *T)
    sumL0  +=       0.205 * cos(4.62946016431  + 1056.46116888240 *T)
    sumL0  +=       0.263 * cos(3.04951219565  + 1493.09366806600 *T)
    sumL0  +=       0.222 * cos(5.54424082649  + 5216.58037280140 *T)
    sumL0  +=       0.244 * cos(0.91026645686  + 3707.83675566220 *T)
    sumL0  +=       0.204 * cos(0.90117975859  + 1408.01712876020 *T)
    sumL0  +=       0.225 * cos(1.23997048012  + 3340.61242669980 *T)
    sumL0  +=       0.258 * cos(2.35906183505  + 2861.75392091100 *T)
    sumL0  +=       0.267 * cos(3.27705002283  + 5120.60114558360 *T)
    sumL0  +=       0.214 * cos(0.66988779149  + 9146.79006902100 *T)
    sumL0  +=       0.235 * cos(4.93761209111  + 1443.44185141230 *T)
    sumL0  +=       0.194 * cos(1.60798828275  + 102.57150935680 *T)
    sumL0  +=       0.215 * cos(0.97603524747  + 479.28838891550 *T)
    sumL0  +=       0.205 * cos(5.23642605904  + 4649.89881763120 *T)
    sumL0  +=       0.257 * cos(4.70227260707  + 9050.81084180320 *T)
    sumL0  +=       0.228 * cos(6.23410921116  + 64.95973858080 *T)
    sumL0  +=       0.180 * cos(4.21309134581  + 143.93412284210 *T)
    sumL0  +=       0.180 * cos(4.82870451226  + 1063.57471588320 *T)
    sumL0  +=       0.180 * cos(5.06126965624  + 52.69019803950 *T)
    sumL0  +=       0.226 * cos(0.55334952097  + 554.06998748280 *T)
    sumL0  +=       0.209 * cos(5.67975843693  + 48.75804477640 *T)
    sumL0  +=       0.186 * cos(3.66368928017  + 108.72184851110 *T)
    sumL0  +=       0.190 * cos(2.00852986549  + 1058.63117066380 *T)
    sumL0  +=       0.183 * cos(3.17358464220  + 140.96517742550 *T)
    sumL0  +=       0.198 * cos(5.49816579454  + 4333.50694797460 *T)
    sumL0  +=       0.240 * cos(6.06602357868  + 1821.12193881750 *T)
    sumL0  +=       0.172 * cos(3.04802064781  + 54.33472944220 *T)
    sumL0  +=       0.170 * cos(4.66520291204  + 1372.59240610810 *T)
    sumL0  +=       0.173 * cos(4.72884056307  + 77204.32749453338 *T)
    sumL0  +=       0.174 * cos(0.85370421252  + 1587.58842257550 *T)
    sumL0  +=       0.215 * cos(0.68219980704  + 1054.71606374320 *T)
    sumL0  +=       0.170 * cos(1.52204803308  + 5591.96087960020 *T)
    sumL0  +=       0.200 * cos(1.60275092073  + 6681.22485339960 *T)
    sumL0  +=       0.193 * cos(2.13003479280  + 103.61403908040 *T)
    sumL0  +=       0.231 * cos(4.69962389031  + 1966.23171782720 *T)
    sumL0  +=       0.179 * cos(5.57395905447  + 1457.52593306200 *T)
    sumL0  +=       0.205 * cos(3.65507571128  + 906.84978763800 *T)
    sumL0  +=       0.181 * cos(4.52272934666  + 24498.83024629040 *T)
    sumL0  +=       0.223 * cos(0.11650319998  + 67.88049988760 *T)
    sumL0  +=       0.172 * cos(5.68083885227  + 1884.12412393800 *T)
    sumL0  +=       0.219 * cos(0.60964963735  + 2729.45619916380 *T)
    sumL0  +=       0.164 * cos(1.06675279755  + 594.65070367540 *T)
    sumL0  +=       0.176 * cos(2.36848603898  + 977.48678462110 *T)
    sumL0  +=       0.170 * cos(2.43036684800  + 4532.57894941100 *T)
    sumL0  +=       0.191 * cos(3.64255924842  + 1440.99417085750 *T)
    sumL0  +=       0.207 * cos(0.49276008455  + 71.60020482960 *T)
    sumL0  +=       0.157 * cos(4.26888100582  + 5069.38346150640 *T)
    sumL0  +=       0.157 * cos(5.14847227422  + 451.72797278960 *T)
    sumL0  +=       0.158 * cos(5.00063628575  + 650.94298657790 *T)
    sumL0  +=       0.159 * cos(5.37530499642  + 20426.57109242200 *T)
    sumL0  +=       0.218 * cos(0.27875408082  + 175.16605980020 *T)
    sumL0  +=       0.155 * cos(0.83696849428  + 1474.93442080130 *T)
    sumL0  +=       0.154 * cos(2.62839957291  + 683.18631549230 *T)
    sumL0  +=       0.171 * cos(1.79511736017  + 1123.11782849260 *T)
    sumL0  +=       0.188 * cos(5.24747110812  + 25565.32572348040 *T)
    sumL0  +=       0.168 * cos(4.14907553818  + 946.72792841500 *T)
    sumL0  +=       0.203 * cos(2.83699715530  + 1489.91227432830 *T)
    sumL0  +=       0.173 * cos(4.34546063838  + 3046.59882834580 *T)
    sumL0  +=       0.190 * cos(5.67865607835  + 1060.13268971460 *T)
    sumL0  +=       0.201 * cos(2.38524182920  + 419.53282798500 *T)
    sumL0  +=       0.152 * cos(5.89088685790  + 208.63322899200 *T)
    sumL0  +=       0.206 * cos(4.46933127349  + 2654.67460059650 *T)
    sumL0  +=       0.156 * cos(2.37819796438  + 2758.66114669240 *T)
    sumL0  +=       0.203 * cos(0.70565514297  + 498.19839570560 *T)
    sumL0  +=       0.205 * cos(3.05468636546  + 1062.30269149600 *T)
    sumL0  +=       0.174 * cos(3.50824761708  + 2004.36475346500 *T)
    sumL0  +=       0.148 * cos(4.73961194393  + 1799.19059698410 *T)
    sumL0  +=       0.188 * cos(3.62315953725  + 3156.80514956520 *T)
    sumL0  +=       0.183 * cos(2.35011338194  + 25551.09862947879 *T)
    sumL0  +=       0.162 * cos(1.58053710589  + 628.59095361920 *T)
    sumL0  +=       0.162 * cos(3.99983876824  + 1482.79872732750 *T)
    sumL0  +=       0.181 * cos(2.85489861839  + 1055.18914449520 *T)
    sumL0  +=       0.151 * cos(3.43198157222  + 629.86297800640 *T)
    sumL0  +=       0.157 * cos(3.15195826490  + 1025.44168024540 *T)
    sumL0  +=       0.194 * cos(5.13049187783  + 1818.15299340090 *T)
    sumL0  +=       0.193 * cos(1.92287052164  + 1140.38330388000 *T)
    sumL0  +=       0.137 * cos(4.22335221970  + 1049.82070263360 *T)
    sumL0  +=       0.167 * cos(2.85163087563  + 5746.27133789600 *T)
    sumL0  +=       0.167 * cos(5.73970282991  + 5760.49843189760 *T)
    sumL0  +=       0.138 * cos(2.23519776527  + 1176.70179840940 *T)
    sumL0  +=       0.151 * cos(4.89507270899  + 532.39927808030 *T)
    sumL0  +=       0.147 * cos(2.65931838448  + 987.30864460760 *T)
    sumL0  +=       0.135 * cos(0.12836417770  + 991.71387862270 *T)
    sumL0  +=       0.166 * cos(3.12682515439  + 580.09354127370 *T)
    sumL0  +=       0.118 * cos(5.98810576300  + 531.38788612400 *T)
    sumL0  +=       0.135 * cos(5.26601313643  + 1065.01100448170 *T)
    sumL0  +=       0.138 * cos(3.18511244397  + 707.56533788050 *T)
    sumL0  +=       0.122 * cos(1.34377059565  + 446.31134681820 *T)
    sumL0  +=       0.120 * cos(2.29717714347  + 1059.22187149480 *T)
    sumL0  +=       0.121 * cos(0.58145552537  + 5621.84292321040 *T)
    sumL0  +=       0.103 * cos(4.75645235023  + 1226.21060271120 *T)
    sumL0  +=       0.104 * cos(6.08481630139  + 528.25467649610 *T)
    sumL0  +=       0.119 * cos(1.06475523307  + 527.99404406520 *T)
    sumL0  +=       0.104 * cos(0.89730746841  + 531.12725369310 *T)
    sumL0  +=       0.120 * cos(5.39001411803  + 1059.54198888360 *T)
    sumL0  +=       0.104 * cos(0.44849170648  + 1128.53445446400 *T)
    sumL0  +=       0.117 * cos(5.42449214711  + 986.08480433020 *T)
    sumL0  +=       0.101 * cos(5.09893554462  + 530.58473697190 *T)
    sumL0  +=       0.102 * cos(0.26948040239  + 450.97721326420 *T)
    sumL0  +=       0.107 * cos(1.58724086516  + 1069.67687092770 *T)
    sumL0  +=       0.086 * cos(2.28711702506  + 2498.89162833840 *T)
    sumL0  +=       0.101 * cos(1.88318822518  + 528.79719321730 *T)
    sumL0  +=       0.086 * cos(1.37568728263  + 970.51624997220 *T)
    sumL0  +=       0.083 * cos(0.06930748288  + 530.91480537200 *T)
    sumL0  +=       0.085 * cos(3.22094000094  + 1553.64817263170 *T)
    sumL0  +=       0.083 * cos(0.62963097974  + 528.46712481720 *T)
    sumL0  +=       0.083 * cos(4.16314675511  + 849.26422848890 *T)
    sumL0  +=       0.079 * cos(3.46688102340  + 1077.54117745390 *T)
    sumL0  +=       0.097 * cos(0.87886975916  + 9690.70812811720 *T)
    sumL0  +=       0.097 * cos(4.27398311206  + 9676.48103411560 *T)
    sumL0  +=       0.101 * cos(0.29639798579  + 857.12853501510 *T)
    sumL0  +=       0.083 * cos(2.55427333923  + 1059.59437851030 *T)
    sumL0  +=       0.078 * cos(0.06461496210  + 521.82665856840 *T)
    sumL0  +=       0.078 * cos(0.76677000862  + 525.54636351040 *T)
    sumL0  +=       0.096 * cos(0.33631035749  + 1090.40141882620 *T)
    sumL0  +=       0.098 * cos(1.42815294497  + 757.21715453420 *T)
    sumL0  +=       0.077 * cos(0.85066773729  + 537.55527162080 *T)
    sumL0  +=       0.084 * cos(5.04765104413  + 1160.02702385300 *T)
    sumL0  +=       0.076 * cos(3.62264327413  + 782.34693644780 *T)
    sumL0  +=       0.085 * cos(1.86831145784  + 25028.52121138500 *T)
    sumL0  +=       0.079 * cos(2.90602202890  + 2114.83170711530 *T)

    # JUPITER Longitude
    sumL1  = 0
    sumL1  +=  52993480757.497 * cos(0.00000000000  + 0.00000000000 *T)
    sumL1  +=  489741.194 * cos(4.22066689928  + 529.69096509460 *T)
    sumL1  +=  228918.538 * cos(6.02647464016  + 7.11354700080 *T)
    sumL1  +=   27655.380 * cos(4.57265956824  + 1059.38193018920 *T)
    sumL1  +=   20720.943 * cos(5.45938936295  + 522.57741809380 *T)
    sumL1  +=   12105.732 * cos(0.16985765041  + 536.80451209540 *T)
    sumL1  +=    6068.051 * cos(4.42419502005  + 103.09277421860 *T)
    sumL1  +=    5433.924 * cos(3.98478382565  + 419.48464387520 *T)
    sumL1  +=    4237.795 * cos(5.89009351271  + 14.22709400160 *T)
    sumL1  +=    2211.854 * cos(5.26771446618  + 206.18554843720 *T)
    sumL1  +=    1295.769 * cos(5.55132765087  + 3.18139373770 *T)
    sumL1  +=    1745.919 * cos(4.92669378486  + 1589.07289528380 *T)
    sumL1  +=    1163.411 * cos(0.51450895328  + 3.93215326310 *T)
    sumL1  +=    1007.216 * cos(0.46478398551  + 735.87651353180 *T)
    sumL1  +=    1173.129 * cos(5.85647304350  + 1052.26838318840 *T)
    sumL1  +=     847.678 * cos(5.75805850450  + 110.20632121940 *T)
    sumL1  +=     827.329 * cos(4.80312015734  + 213.29909543800 *T)
    sumL1  +=    1003.574 * cos(3.15040301822  + 426.59819087600 *T)
    sumL1  +=    1098.735 * cos(5.30704981594  + 515.46387109300 *T)
    sumL1  +=     816.397 * cos(0.58643054886  + 1066.49547719000 *T)
    sumL1  +=     725.447 * cos(5.51827471473  + 639.89728631400 *T)
    sumL1  +=     567.845 * cos(5.98867049451  + 625.67019231240 *T)
    sumL1  +=     474.181 * cos(4.13245269168  + 412.37109687440 *T)
    sumL1  +=     412.930 * cos(5.73652891261  + 95.97922721780 *T)
    sumL1  +=     335.817 * cos(3.73248749046  + 1162.47470440780 *T)
    sumL1  +=     345.249 * cos(4.24159565410  + 632.78373931320 *T)
    sumL1  +=     234.066 * cos(6.24302226646  + 309.27832265580 *T)
    sumL1  +=     194.784 * cos(2.21879010911  + 323.50541665740 *T)
    sumL1  +=     234.340 * cos(4.03469970332  + 949.17560896980 *T)
    sumL1  +=     183.938 * cos(6.27963588822  + 543.91805909620 *T)
    sumL1  +=     198.525 * cos(1.50458442825  + 838.96928775040 *T)
    sumL1  +=     186.899 * cos(6.08620565908  + 742.99006053260 *T)
    sumL1  +=     171.380 * cos(5.41655983845  + 199.07200143640 *T)
    sumL1  +=     130.771 * cos(0.62643377351  + 728.76296653100 *T)
    sumL1  +=     107.575 * cos(4.49282760117  + 956.28915597060 *T)
    sumL1  +=     115.393 * cos(0.68019050174  + 846.08283475120 *T)
    sumL1  +=     115.047 * cos(5.28641699144  + 2118.76386037840 *T)
    sumL1  +=      66.824 * cos(5.73365126533  + 21.34064100240 *T)
    sumL1  +=      69.618 * cos(5.97263450278  + 532.87235883230 *T)
    sumL1  +=      64.850 * cos(6.08803490288  + 1581.95934828300 *T)
    sumL1  +=      79.686 * cos(5.82412400273  + 1045.15483618760 *T)
    sumL1  +=      57.939 * cos(0.99453087342  + 1596.18644228460 *T)
    sumL1  +=      65.635 * cos(0.12924191430  + 526.50957135690 *T)
    sumL1  +=      58.509 * cos(0.58626971028  + 1155.36115740700 *T)
    sumL1  +=      56.600 * cos(1.41198438841  + 533.62311835770 *T)
    sumL1  +=      71.643 * cos(5.34162650321  + 942.06206196900 *T)
    sumL1  +=      57.368 * cos(5.96851304799  + 1169.58825140860 *T)
    sumL1  +=      54.935 * cos(5.42806383723  + 10.29494073850 *T)
    sumL1  +=      52.016 * cos(0.22981299129  + 1368.66025284500 *T)
    sumL1  +=      52.309 * cos(5.72661448388  + 117.31986822020 *T)
    sumL1  +=      50.418 * cos(6.08075147811  + 525.75881183150 *T)
    sumL1  +=      47.418 * cos(3.62611843241  + 1478.86657406440 *T)
    sumL1  +=      39.888 * cos(4.16158013600  + 1692.16566950240 *T)
    sumL1  +=      46.678 * cos(0.51144073175  + 1265.56747862640 *T)
    sumL1  +=      32.827 * cos(5.03596689455  + 220.41264243880 *T)
    sumL1  +=      33.558 * cos(0.09913904872  + 302.16477565500 *T)
    sumL1  +=      29.379 * cos(3.35927241533  + 4.66586644600 *T)
    sumL1  +=      29.307 * cos(0.75907909735  + 88.86568021700 *T)
    sumL1  +=      32.449 * cos(5.37492530697  + 508.35032409220 *T)
    sumL1  +=      29.483 * cos(5.42208897099  + 1272.68102562720 *T)
    sumL1  +=      21.802 * cos(6.15054054070  + 1685.05212250160 *T)
    sumL1  +=      25.195 * cos(1.60723063387  + 831.85574074960 *T)
    sumL1  +=      21.133 * cos(5.86346824200  + 1258.45393162560 *T)
    sumL1  +=      19.747 * cos(2.17205957814  + 316.39186965660 *T)
    sumL1  +=      17.871 * cos(0.82841413516  + 433.71173787680 *T)
    sumL1  +=      17.703 * cos(5.95527049039  + 5.41662597140 *T)
    sumL1  +=      17.230 * cos(2.76395560958  + 853.19638175200 *T)
    sumL1  +=      17.453 * cos(0.70749901224  + 1471.75302706360 *T)
    sumL1  +=      17.508 * cos(0.49799925173  + 1375.77379984580 *T)
    sumL1  +=      14.368 * cos(0.91459831140  + 18.15924726470 *T)
    sumL1  +=      14.107 * cos(0.63031082833  + 2.92076130680 *T)
    sumL1  +=      11.559 * cos(4.30379009964  + 405.25754987360 *T)
    sumL1  +=      11.728 * cos(1.76426582357  + 380.12776796000 *T)
    sumL1  +=      11.054 * cos(5.56735602213  + 1574.84580128220 *T)
    sumL1  +=      10.425 * cos(0.31355034390  + 1361.54670584420 *T)
    sumL1  +=       9.804 * cos(5.90363777277  + 519.39602435610 *T)
    sumL1  +=       9.805 * cos(0.38648727979  + 1073.60902419080 *T)
    sumL1  +=       9.285 * cos(3.21842287530  + 1795.25844372100 *T)
    sumL1  +=       8.864 * cos(0.53776257958  + 1788.14489672020 *T)
    sumL1  +=       8.370 * cos(5.88484552222  + 2001.44399215820 *T)
    sumL1  +=       8.148 * cos(5.10162311410  + 1485.98012106520 *T)
    sumL1  +=       7.658 * cos(5.64890060131  + 2648.45482547300 *T)
    sumL1  +=       6.690 * cos(2.41093459420  + 4.19278569400 *T)
    sumL1  +=       5.840 * cos(4.22347896053  + 2008.55753915900 *T)
    sumL1  +=       7.256 * cos(6.19384525651  + 11.04570026390 *T)
    sumL1  +=       6.266 * cos(1.36137786945  + 1148.24761040620 *T)
    sumL1  +=       5.141 * cos(5.23083932012  + 628.85158605010 *T)
    sumL1  +=       5.140 * cos(2.92955981951  + 518.64526483070 *T)
    sumL1  +=       4.765 * cos(0.16838181862  + 629.60234557550 *T)
    sumL1  +=       4.603 * cos(0.78529559911  + 721.64941953020 *T)
    sumL1  +=       4.575 * cos(6.24794935732  + 1677.93857550080 *T)
    sumL1  +=       4.537 * cos(4.95096707833  + 635.96513305090 *T)
    sumL1  +=       4.518 * cos(2.06523915453  + 453.42489381900 *T)
    sumL1  +=       4.414 * cos(0.15381186059  + 1699.27921650320 *T)
    sumL1  +=       5.593 * cos(5.57489981207  + 191.95845443560 *T)
    sumL1  +=       5.403 * cos(1.46004886198  + 330.61896365820 *T)
    sumL1  +=       4.285 * cos(0.23949868127  + 2104.53676637680 *T)
    sumL1  +=       4.223 * cos(1.44087555881  + 2125.87740737920 *T)
    sumL1  +=       4.101 * cos(6.19274358942  + 636.71589257630 *T)
    sumL1  +=       4.432 * cos(4.35811524051  + 423.41679713830 *T)
    sumL1  +=       4.132 * cos(0.50170694173  + 1056.20053645150 *T)
    sumL1  +=       4.398 * cos(4.14280286969  + 511.53171782990 *T)
    sumL1  +=       5.406 * cos(4.40429493698  + 2221.85663459700 *T)
    sumL1  +=       4.467 * cos(0.08534650684  + 1062.56332392690 *T)
    sumL1  +=       3.569 * cos(5.66540477010  + 2317.83586181480 *T)
    sumL1  +=       4.007 * cos(2.54845549248  + 74.78159856730 *T)
    sumL1  +=       3.515 * cos(0.25495124831  + 1055.44977692610 *T)
    sumL1  +=       3.687 * cos(2.93378008847  + 32.24332891440 *T)
    sumL1  +=       2.883 * cos(5.72793010505  + 99.91138048090 *T)
    sumL1  +=       2.969 * cos(5.50054720569  + 107.02492748170 *T)
    sumL1  +=       2.720 * cos(1.25222590925  + 540.73666535850 *T)
    sumL1  +=       2.808 * cos(3.30714813896  + 0.75075952540 *T)
    sumL1  +=       2.768 * cos(1.61339487804  + 1063.31408345230 *T)
    sumL1  +=       2.666 * cos(4.28662288102  + 106.27416795630 *T)
    sumL1  +=       2.704 * cos(3.03615556153  + 422.66603761290 *T)
    sumL1  +=       3.290 * cos(5.89081682150  + 1802.37199072180 *T)
    sumL1  +=       2.578 * cos(3.60390367979  + 750.10360753340 *T)
    sumL1  +=       2.661 * cos(0.35249312659  + 1898.35121793960 *T)
    sumL1  +=       2.486 * cos(5.28950877719  + 1891.23767093880 *T)
    sumL1  +=       2.936 * cos(1.09052029450  + 1464.63948006280 *T)
    sumL1  +=       3.190 * cos(4.60740643547  + 416.30325013750 *T)
    sumL1  +=       2.390 * cos(6.01779736611  + 551.03160609700 *T)
    sumL1  +=       2.214 * cos(5.24450923180  + 621.73803904930 *T)
    sumL1  +=       2.319 * cos(5.82920300130  + 305.34616939270 *T)
    sumL1  +=       2.089 * cos(5.99310370434  + 1994.33044515740 *T)
    sumL1  +=       2.042 * cos(0.75008788531  + 142.44965013380 *T)
    sumL1  +=       2.121 * cos(0.01537599023  + 2420.92863603340 *T)
    sumL1  +=       2.114 * cos(6.25308371567  + 647.01083331480 *T)
    sumL1  +=       2.020 * cos(4.17560390841  + 569.04784100980 *T)
    sumL1  +=       2.109 * cos(5.18682321403  + 227.52618943960 *T)
    sumL1  +=       2.283 * cos(5.80043809222  + 539.98590583310 *T)
    sumL1  +=       1.977 * cos(3.99197009651  + 24.37902238820 *T)
    sumL1  +=       1.960 * cos(1.35288793079  + 963.40270297140 *T)
    sumL1  +=       1.903 * cos(2.78349628184  + 2428.04218303420 *T)
    sumL1  +=       1.915 * cos(4.22134509685  + 2324.94940881560 *T)
    sumL1  +=       1.971 * cos(5.88715684267  + 217.23124870110 *T)
    sumL1  +=       1.917 * cos(3.03728154374  + 1382.88734684660 *T)
    sumL1  +=       2.026 * cos(3.08606488714  + 408.43894361130 *T)
    sumL1  +=       1.834 * cos(5.61474110217  + 430.53034413910 *T)
    sumL1  +=       1.838 * cos(1.25467410218  + 81.75213321620 *T)
    sumL1  +=       2.460 * cos(4.63268678998  + 1905.46476494040 *T)
    sumL1  +=       1.820 * cos(5.97497926120  + 114.13847448250 *T)
    sumL1  +=       2.043 * cos(4.34047514845  + 70.84944530420 *T)
    sumL1  +=       1.959 * cos(4.03116026306  + 92.04707395470 *T)
    sumL1  +=       1.768 * cos(0.33097462499  + 35.42472265210 *T)
    sumL1  +=       2.334 * cos(5.87042638470  + 1038.04128918680 *T)
    sumL1  +=       1.835 * cos(4.81326127892  + 124.43341522100 *T)
    sumL1  +=       2.269 * cos(1.02549350754  + 618.55664531160 *T)
    sumL1  +=       1.919 * cos(5.01297395549  + 99.16062095550 *T)
    sumL1  +=       1.923 * cos(0.28688549585  + 31.01948863700 *T)
    sumL1  +=       1.878 * cos(5.69299116574  + 210.11770170030 *T)
    sumL1  +=       1.679 * cos(0.25635730278  + 295.05122865420 *T)
    sumL1  +=       1.656 * cos(5.46039280732  + 2634.22773147140 *T)
    sumL1  +=       1.675 * cos(6.15609073315  + 643.82943957710 *T)
    sumL1  +=       1.953 * cos(5.09846435548  + 17.40848773930 *T)
    sumL1  +=       1.539 * cos(2.75316078346  + 415.55249061210 *T)
    sumL1  +=       1.467 * cos(0.54812675158  + 458.84151979040 *T)
    sumL1  +=       1.482 * cos(3.76736278426  + 534.35683154060 *T)
    sumL1  +=       1.446 * cos(3.15802770791  + 25.12978191360 *T)
    sumL1  +=       1.667 * cos(0.26406950755  + 835.03713448730 *T)
    sumL1  +=       1.472 * cos(0.83054329617  + 28.31117565130 *T)
    sumL1  +=       1.655 * cos(0.88908548504  + 1781.03134971940 *T)
    sumL1  +=       1.294 * cos(5.76241191046  + 440.82528487760 *T)
    sumL1  +=       1.348 * cos(2.49823510924  + 984.60033162190 *T)
    sumL1  +=       1.352 * cos(5.10869562455  + 149.56319713460 *T)
    sumL1  +=       1.344 * cos(0.01942249067  + 2214.74308759620 *T)
    sumL1  +=       1.188 * cos(2.24279457878  + 31.49256938900 *T)
    sumL1  +=       1.166 * cos(0.80686346228  + 739.80866679490 *T)
    sumL1  +=       1.322 * cos(4.25691184168  + 2538.24850425360 *T)
    sumL1  +=       1.094 * cos(6.02985819406  + 2737.32050569000 *T)
    sumL1  +=       1.112 * cos(4.38204360670  + 561.93429400900 *T)
    sumL1  +=       1.346 * cos(3.20575848870  + 525.02509864860 *T)
    sumL1  +=       1.056 * cos(5.76507115032  + 2310.72231481400 *T)
    sumL1  +=       1.159 * cos(0.46189564970  + 67.66805156650 *T)
    sumL1  +=       1.027 * cos(0.20709586018  + 7.86430652620 *T)
    sumL1  +=       1.143 * cos(5.56626418636  + 46.47042291600 *T)
    sumL1  +=       1.012 * cos(0.54293005597  + 532.13864564940 *T)
    sumL1  +=       0.978 * cos(5.13939194101  + 2207.62954059540 *T)
    sumL1  +=       0.993 * cos(2.03698185233  + 319.57326339430 *T)
    sumL1  +=       1.035 * cos(2.90231353535  + 611.44309831080 *T)
    sumL1  +=       1.021 * cos(4.75651217048  + 527.24328453980 *T)
    sumL1  +=       1.308 * cos(1.78809336431  + 824.74219374880 *T)
    sumL1  +=       0.964 * cos(2.82269601958  + 2111.65031337760 *T)
    sumL1  +=       0.896 * cos(2.54505998806  + 2744.43405269080 *T)
    sumL1  +=       0.890 * cos(5.41036782817  + 28.45418800320 *T)
    sumL1  +=       0.906 * cos(0.76565238554  + 1439.50969814920 *T)
    sumL1  +=       0.985 * cos(0.88687623770  + 5760.49843189760 *T)
    sumL1  +=       0.983 * cos(1.42102343372  + 5746.27133789600 *T)
    sumL1  +=       0.892 * cos(5.87250060663  + 203.00415469950 *T)
    sumL1  +=       0.942 * cos(2.31049430734  + 9690.70812811720 *T)
    sumL1  +=       0.941 * cos(2.84331157527  + 9676.48103411560 *T)
    sumL1  +=       0.867 * cos(0.81020362547  + 524.27433912320 *T)
    sumL1  +=       0.829 * cos(2.35178495412  + 312.45971639350 *T)
    sumL1  +=       0.912 * cos(2.80494184378  + 6.21977512350 *T)
    sumL1  +=       0.809 * cos(1.05148218513  + 529.64278098480 *T)
    sumL1  +=       0.779 * cos(4.80009242059  + 945.24345570670 *T)
    sumL1  +=       0.878 * cos(5.76532521399  + 1.64453140270 *T)
    sumL1  +=       0.953 * cos(4.30945738629  + 209.36694217490 *T)
    sumL1  +=       0.772 * cos(5.25607113566  + 2950.61960112800 *T)
    sumL1  +=       0.745 * cos(0.03810558502  + 535.10759106600 *T)
    sumL1  +=       0.744 * cos(0.58381523987  + 25.27279426550 *T)
    sumL1  +=       0.734 * cos(0.20800485100  + 1049.08698945070 *T)
    sumL1  +=       0.747 * cos(2.71772840871  + 38.13303563780 *T)
    sumL1  +=       0.728 * cos(5.97210358938  + 945.99421523210 *T)
    sumL1  +=       0.769 * cos(4.51394016967  + 952.35700270750 *T)
    sumL1  +=       0.710 * cos(0.38016353553  + 69.15252427480 *T)
    sumL1  +=       0.760 * cos(3.07033779824  + 39.35687591520 *T)
    sumL1  +=       0.802 * cos(1.14191463412  + 532.61172640140 *T)
    sumL1  +=       0.704 * cos(1.25447308120  + 547.85021235930 *T)
    sumL1  +=       0.721 * cos(0.73855379162  + 2228.97018159780 *T)
    sumL1  +=       0.794 * cos(4.25051539085  + 2641.34127847220 *T)
    sumL1  +=       0.795 * cos(3.20588363820  + 604.47256366190 *T)
    sumL1  +=       0.818 * cos(1.05229815343  + 909.81873305460 *T)
    sumL1  +=       0.724 * cos(5.68281830264  + 953.10776223290 *T)
    sumL1  +=       0.836 * cos(0.60410469174  + 2097.42321937600 *T)
    sumL1  +=       0.669 * cos(5.75757140051  + 2015.67108615980 *T)
    sumL1  +=       0.682 * cos(1.19994890339  + 387.24131496080 *T)
    sumL1  +=       0.640 * cos(3.91546675664  + 528.72775724810 *T)
    sumL1  +=       0.809 * cos(4.24929331276  + 529.73914920440 *T)
    sumL1  +=       0.819 * cos(4.91540072376  + 2751.54759969160 *T)
    sumL1  +=       0.692 * cos(2.51162384766  + 916.93228005540 *T)
    sumL1  +=       0.784 * cos(4.23651511312  + 195.13984817330 *T)
    sumL1  +=       0.762 * cos(1.12201139619  + 732.69511979410 *T)
    sumL1  +=       0.617 * cos(5.80920925081  + 739.05790726950 *T)
    sumL1  +=       0.727 * cos(4.24401822698  + 760.25553592000 *T)
    sumL1  +=       0.591 * cos(3.26075006572  + 202.25339517410 *T)
    sumL1  +=       0.552 * cos(5.83533550039  + 526.77020378780 *T)
    sumL1  +=       0.640 * cos(1.38530872949  + 530.65417294110 *T)
    sumL1  +=       0.577 * cos(6.09100925678  + 2531.13495725280 *T)
    sumL1  +=       0.620 * cos(3.01917904435  + 902.70518605380 *T)
    sumL1  +=       0.722 * cos(5.18171159557  + 1.48447270830 *T)
    sumL1  +=       0.540 * cos(3.78809230820  + 2957.73314812880 *T)
    sumL1  +=       0.523 * cos(3.63882376000  + 437.64389113990 *T)
    sumL1  +=       0.527 * cos(5.80796427555  + 3053.71237534660 *T)
    sumL1  +=       0.488 * cos(4.99103190309  + 483.22054217860 *T)
    sumL1  +=       0.557 * cos(4.11381202161  + 2854.64037391020 *T)
    sumL1  +=       0.492 * cos(0.76371083106  + 1603.29998928540 *T)
    sumL1  +=       0.487 * cos(5.55383951779  + 2627.11418447060 *T)
    sumL1  +=       0.487 * cos(5.86510858429  + 724.83081326790 *T)
    sumL1  +=       0.453 * cos(0.61375011101  + 1159.29331067010 *T)
    sumL1  +=       0.450 * cos(2.28121042355  + 3060.82592234740 *T)
    sumL1  +=       0.515 * cos(4.78126059280  + 447.79581952650 *T)
    sumL1  +=       0.449 * cos(4.70231576312  + 934.94851496820 *T)
    sumL1  +=       0.450 * cos(1.91049508739  + 597.35901666110 *T)
    sumL1  +=       0.438 * cos(6.01178917646  + 3178.14579056760 *T)
    sumL1  +=       0.494 * cos(0.53844942275  + 1354.43315884340 *T)
    sumL1  +=       0.501 * cos(5.51752195462  + 2435.15573003500 *T)
    sumL1  +=       0.432 * cos(3.64903264921  + 313.21047591890 *T)
    sumL1  +=       0.435 * cos(3.02449828967  + 533.88375078860 *T)
    sumL1  +=       0.426 * cos(5.07945534339  + 2524.02141025200 *T)
    sumL1  +=       0.491 * cos(3.59286364200  + 230.56457082540 *T)
    sumL1  +=       0.547 * cos(0.34432090949  + 1251.34038462480 *T)
    sumL1  +=       0.503 * cos(1.57454509207  + 454.90936652730 *T)
    sumL1  +=       0.486 * cos(4.39351469958  + 462.02291352810 *T)
    sumL1  +=       0.524 * cos(2.03003740296  + 1279.79457262800 *T)
    sumL1  +=       0.388 * cos(5.58318013074  + 731.94436026870 *T)
    sumL1  +=       0.449 * cos(1.11025492739  + 56.62235130260 *T)
    sumL1  +=       0.398 * cos(5.19943284273  + 3267.01147078460 *T)
    sumL1  +=       0.416 * cos(1.70821917336  + 245.54242435240 *T)
    sumL1  +=       0.379 * cos(1.80234948769  + 2655.56837247380 *T)
    sumL1  +=       0.355 * cos(1.65214516751  + 78.71375183040 *T)
    sumL1  +=       0.404 * cos(1.72647262603  + 1141.13406340540 *T)
    sumL1  +=       0.335 * cos(6.01254286794  + 960.22130923370 *T)
    sumL1  +=       0.331 * cos(1.74086938716  + 490.33408917940 *T)
    sumL1  +=       0.401 * cos(0.30034336462  + 2332.06295581640 *T)
    sumL1  +=       0.336 * cos(2.64385574909  + 1021.24889455140 *T)
    sumL1  +=       0.389 * cos(0.31259289221  + 2413.81508903260 *T)
    sumL1  +=       0.314 * cos(5.73833529708  + 1158.54255114470 *T)
    sumL1  +=       0.313 * cos(4.74363791106  + 938.12990870590 *T)
    sumL1  +=       0.333 * cos(0.80112437148  + 1585.14074202070 *T)
    sumL1  +=       0.323 * cos(3.52656245280  + 3274.12501778540 *T)
    sumL1  +=       0.395 * cos(1.73181407631  + 1593.00504854690 *T)
    sumL1  +=       0.302 * cos(4.64184749164  + 1261.63532536330 *T)
    sumL1  +=       0.325 * cos(0.54991590409  + 43.28902917830 *T)
    sumL1  +=       0.293 * cos(0.97977818746  + 1585.89150154610 *T)
    sumL1  +=       0.341 * cos(2.80833606944  + 1514.29129671650 *T)
    sumL1  +=       0.304 * cos(6.12522825214  + 1262.38608488870 *T)
    sumL1  +=       0.286 * cos(2.89800423081  + 530.21222995640 *T)
    sumL1  +=       0.387 * cos(0.46648572639  + 1592.25428902150 *T)
    sumL1  +=       0.285 * cos(4.56394598052  + 1268.74887236410 *T)
    sumL1  +=       0.310 * cos(4.69102289591  + 76.26607127560 *T)
    sumL1  +=       0.278 * cos(5.49867187248  + 280.96714700450 *T)
    sumL1  +=       0.358 * cos(5.45926487831  + 113.38771495710 *T)
    sumL1  +=       0.283 * cos(1.09230506350  + 1061.82961074400 *T)
    sumL1  +=       0.326 * cos(0.60265259639  + 827.92358748650 *T)
    sumL1  +=       0.284 * cos(5.36580034539  + 1165.65609814550 *T)
    sumL1  +=       0.281 * cos(5.54635461050  + 3370.10424500320 *T)
    sumL1  +=       0.269 * cos(3.92616563946  + 42.53826965290 *T)
    sumL1  +=       0.275 * cos(2.58465453365  + 373.01422095920 *T)
    sumL1  +=       0.357 * cos(1.39391983207  + 1493.09366806600 *T)
    sumL1  +=       0.258 * cos(5.96670694140  + 1269.49963188950 *T)
    sumL1  +=       0.259 * cos(2.56026216784  + 9146.79006902100 *T)
    sumL1  +=       0.281 * cos(2.74823090198  + 4694.00295470760 *T)
    sumL1  +=       0.281 * cos(3.01324655940  + 320.32402291970 *T)
    sumL1  +=       0.272 * cos(4.18504958920  + 8624.21265092720 *T)
    sumL1  +=       0.245 * cos(1.24462798353  + 252.65597135320 *T)
    sumL1  +=       0.244 * cos(2.02892764690  + 3377.21779200400 *T)
    sumL1  +=       0.324 * cos(1.84851618413  + 1289.94650101460 *T)
    sumL1  +=       0.221 * cos(6.22167997496  + 3281.23856478620 *T)
    sumL1  +=       0.238 * cos(3.93371505401  + 3171.03224356680 *T)
    sumL1  +=       0.226 * cos(5.94296271326  + 224.34479570190 *T)
    sumL1  +=       0.213 * cos(3.68264234750  + 1048.33622992530 *T)
    sumL1  +=       0.216 * cos(5.82941334164  + 1567.73225428140 *T)
    sumL1  +=       0.295 * cos(4.70194747095  + 3067.93946934820 *T)
    sumL1  +=       0.206 * cos(4.98184230959  + 1357.61455258110 *T)
    sumL1  +=       0.202 * cos(1.32439444045  + 4326.39340097380 *T)
    sumL1  +=       0.227 * cos(0.78540105705  + 59.80374504030 *T)
    sumL1  +=       0.237 * cos(5.56926897693  + 2943.50605412720 *T)
    sumL1  +=       0.207 * cos(0.07907015398  + 5223.69391980220 *T)
    sumL1  +=       0.199 * cos(3.30501818656  + 4120.20785253660 *T)
    sumL1  +=       0.194 * cos(5.95526916809  + 84.93352695390 *T)
    sumL1  +=       0.266 * cos(1.58032565718  + 983.11585891360 *T)
    sumL1  +=       0.198 * cos(4.31078641704  + 4017.11507831800 *T)
    sumL1  +=       0.198 * cos(0.30166351366  + 1166.40685767090 *T)
    sumL1  +=       0.188 * cos(0.90738705875  + 135.33610313300 *T)
    sumL1  +=       0.186 * cos(0.69289672485  + 92.79783348010 *T)
    sumL1  +=       0.182 * cos(1.18931462257  + 1512.80682400820 *T)
    sumL1  +=       0.191 * cos(1.04146023518  + 1884.12412393800 *T)
    sumL1  +=       0.174 * cos(6.13734594396  + 3597.63043444280 *T)
    sumL1  +=       0.189 * cos(0.35191512844  + 1372.59240610810 *T)
    sumL1  +=       0.172 * cos(4.35250972697  + 1578.02719501990 *T)
    sumL1  +=       0.173 * cos(2.30241719278  + 1176.70179840940 *T)
    sumL1  +=       0.220 * cos(1.06991056825  + 2200.51599359460 *T)
    sumL1  +=       0.186 * cos(4.90511103807  + 3583.40334044120 *T)
    sumL1  +=       0.189 * cos(0.24160744024  + 1670.82502850000 *T)
    sumL1  +=       0.206 * cos(0.01485146863  + 2730.20695868920 *T)
    sumL1  +=       0.174 * cos(1.83997277029  + 746.92221379570 *T)
    sumL1  +=       0.225 * cos(3.13108099660  + 630.33605875840 *T)
    sumL1  +=       0.206 * cos(5.22730929781  + 3995.77443731560 *T)
    sumL1  +=       0.169 * cos(2.57956682688  + 9161.01716302260 *T)
    sumL1  +=       0.165 * cos(1.51795928301  + 4010.00153131720 *T)
    sumL1  +=       0.181 * cos(2.05055200822  + 842.90144101350 *T)
    sumL1  +=       0.181 * cos(5.96554625357  + 1578.77795454530 *T)
    sumL1  +=       0.166 * cos(1.55114863100  + 1070.42763045310 *T)
    sumL1  +=       0.157 * cos(5.87839958880  + 3914.02230409940 *T)
    sumL1  +=       0.160 * cos(0.43729819176  + 2545.36205125440 *T)
    sumL1  +=       0.168 * cos(5.73975661792  + 2847.52682690940 *T)
    sumL1  +=       0.157 * cos(2.25764581068  + 850.01498801430 *T)
    sumL1  +=       0.187 * cos(0.64918748618  + 842.15068148810 *T)
    sumL1  +=       0.180 * cos(1.88055488803  + 685.47393735270 *T)
    sumL1  +=       0.153 * cos(4.15259684562  + 4333.50694797460 *T)
    sumL1  +=       0.154 * cos(3.65536637158  + 77734.01845962799 *T)
    sumL1  +=       0.151 * cos(3.17795437121  + 3590.51688744200 *T)
    sumL1  +=       0.155 * cos(3.87623547990  + 327.43756992050 *T)
    sumL1  +=       0.171 * cos(3.33647878498  + 1912.57831194120 *T)
    sumL1  +=       0.188 * cos(4.53005359421  + 1041.22268292450 *T)
    sumL1  +=       0.134 * cos(4.09921613445  + 530.44172462000 *T)
    sumL1  +=       0.123 * cos(4.79543460218  + 1098.73880610440 *T)
    sumL1  +=       0.161 * cos(2.02006564218  + 860.30992875280 *T)
    sumL1  +=       0.143 * cos(2.40197278329  + 529.16970023280 *T)
    sumL1  +=       0.115 * cos(1.55831212007  + 9153.90361602180 *T)
    sumL1  +=       0.106 * cos(5.94313244357  + 1057.89745748090 *T)
    sumL1  +=       0.119 * cos(5.10578428676  + 1056.93424963440 *T)
    sumL1  +=       0.100 * cos(5.74974781049  + 501.23677709140 *T)
    sumL1  +=       0.094 * cos(1.40134175492  + 1059.33374607940 *T)
    sumL1  +=       0.098 * cos(3.79115318281  + 497.44763618020 *T)
    sumL1  +=       0.090 * cos(4.09610113044  + 1064.04779663520 *T)
    sumL1  +=       0.102 * cos(1.10442899544  + 1969.20066324380 *T)
    sumL1  +=       0.087 * cos(0.58218477838  + 1173.52040467170 *T)
    sumL1  +=       0.109 * cos(3.83745968299  + 525.49817940060 *T)
    sumL1  +=       0.094 * cos(4.59915291355  + 1059.43011429900 *T)
    sumL1  +=       0.118 * cos(6.11701561559  + 1069.67687092770 *T)
    sumL1  +=       0.107 * cos(5.40509332689  + 679.25416222920 *T)
    sumL1  +=       0.089 * cos(5.90037690244  + 757.21715453420 *T)
    sumL1  +=       0.078 * cos(6.06217863109  + 970.51624997220 *T)
    sumL1  +=       0.080 * cos(5.45470236239  + 3163.91869656600 *T)
    sumL1  +=       0.072 * cos(5.65789862232  + 1151.42900414390 *T)
    sumL1  +=       0.080 * cos(0.04539720100  + 1080.72257119160 *T)
    sumL1  +=       0.075 * cos(4.26526686574  + 1058.41872234270 *T)

    # JUPITER Longitude
    sumL2  = 0
    sumL2  +=   47233.598 * cos(4.32148323554  + 7.11354700080 *T)
    sumL2  +=   30629.053 * cos(2.93021440216  + 529.69096509460 *T)
    sumL2  +=   38965.550 * cos(0.00000000000  + 0.00000000000 *T)
    sumL2  +=    3189.317 * cos(1.05504615595  + 522.57741809380 *T)
    sumL2  +=    2723.358 * cos(3.41411526638  + 1059.38193018920 *T)
    sumL2  +=    2729.292 * cos(4.84545481351  + 536.80451209540 *T)
    sumL2  +=    1721.069 * cos(4.18734385158  + 14.22709400160 *T)
    sumL2  +=     383.258 * cos(5.76790714387  + 419.48464387520 *T)
    sumL2  +=     367.498 * cos(6.05509120409  + 103.09277421860 *T)
    sumL2  +=     377.524 * cos(0.76048964872  + 515.46387109300 *T)
    sumL2  +=     337.386 * cos(3.78644384244  + 3.18139373770 *T)
    sumL2  +=     308.200 * cos(0.69356654052  + 206.18554843720 *T)
    sumL2  +=     218.408 * cos(3.81389191353  + 1589.07289528380 *T)
    sumL2  +=     198.883 * cos(5.33996443444  + 1066.49547719000 *T)
    sumL2  +=     197.445 * cos(2.48356402053  + 3.93215326310 *T)
    sumL2  +=     146.230 * cos(3.81373196838  + 639.89728631400 *T)
    sumL2  +=     155.862 * cos(1.40642426467  + 1052.26838318840 *T)
    sumL2  +=     129.570 * cos(5.83738872525  + 412.37109687440 *T)
    sumL2  +=     141.932 * cos(1.63435169016  + 426.59819087600 *T)
    sumL2  +=     117.327 * cos(1.41435462588  + 625.67019231240 *T)
    sumL2  +=      96.733 * cos(4.03383427887  + 110.20632121940 *T)
    sumL2  +=      90.823 * cos(1.10630629042  + 95.97922721780 *T)
    sumL2  +=      78.769 * cos(4.63726131329  + 543.91805909620 *T)
    sumL2  +=      72.392 * cos(2.21716670026  + 735.87651353180 *T)
    sumL2  +=      87.292 * cos(2.52235174825  + 632.78373931320 *T)
    sumL2  +=      56.910 * cos(3.12292059854  + 213.29909543800 *T)
    sumL2  +=      48.622 * cos(1.67283791618  + 309.27832265580 *T)
    sumL2  +=      58.475 * cos(0.83216317444  + 199.07200143640 *T)
    sumL2  +=      40.150 * cos(4.02485444740  + 21.34064100240 *T)
    sumL2  +=      39.784 * cos(0.62416945827  + 323.50541665740 *T)
    sumL2  +=      35.718 * cos(2.32581247002  + 728.76296653100 *T)
    sumL2  +=      25.620 * cos(2.51240623862  + 1162.47470440780 *T)
    sumL2  +=      29.255 * cos(3.60838327799  + 10.29494073850 *T)
    sumL2  +=      23.591 * cos(3.00532139306  + 956.28915597060 *T)
    sumL2  +=      27.814 * cos(3.23992013743  + 838.96928775040 *T)
    sumL2  +=      25.993 * cos(4.50118298290  + 742.99006053260 *T)
    sumL2  +=      25.194 * cos(1.21868110687  + 1045.15483618760 *T)
    sumL2  +=      19.458 * cos(4.29028644674  + 532.87235883230 *T)
    sumL2  +=      17.660 * cos(0.80953941560  + 508.35032409220 *T)
    sumL2  +=      15.355 * cos(5.81037986941  + 1596.18644228460 *T)
    sumL2  +=      17.058 * cos(4.20001977723  + 2118.76386037840 *T)
    sumL2  +=      17.040 * cos(1.83402146640  + 526.50957135690 *T)
    sumL2  +=      14.661 * cos(3.99989622586  + 117.31986822020 *T)
    sumL2  +=      13.639 * cos(1.80336677963  + 302.16477565500 *T)
    sumL2  +=      13.230 * cos(2.51856643603  + 88.86568021700 *T)
    sumL2  +=      12.756 * cos(4.36856232414  + 1169.58825140860 *T)
    sumL2  +=      15.292 * cos(0.68174165476  + 942.06206196900 *T)
    sumL2  +=      10.986 * cos(4.43586634639  + 525.75881183150 *T)
    sumL2  +=      13.920 * cos(5.95169568482  + 316.39186965660 *T)
    sumL2  +=       9.437 * cos(2.17684563456  + 1155.36115740700 *T)
    sumL2  +=       8.812 * cos(3.29452783338  + 220.41264243880 *T)
    sumL2  +=       7.823 * cos(5.75672228354  + 846.08283475120 *T)
    sumL2  +=       7.549 * cos(2.70955516779  + 533.62311835770 *T)
    sumL2  +=       9.681 * cos(1.71563161051  + 1581.95934828300 *T)
    sumL2  +=       8.690 * cos(3.31924493607  + 831.85574074960 *T)
    sumL2  +=       6.285 * cos(0.49939863541  + 949.17560896980 *T)
    sumL2  +=       6.685 * cos(2.17560093281  + 1265.56747862640 *T)
    sumL2  +=       5.381 * cos(6.00510875948  + 405.25754987360 *T)
    sumL2  +=       4.676 * cos(1.40846192799  + 1258.45393162560 *T)
    sumL2  +=       4.421 * cos(3.02360159274  + 1692.16566950240 *T)
    sumL2  +=       4.403 * cos(5.47737266160  + 433.71173787680 *T)
    sumL2  +=       4.286 * cos(5.07139951645  + 1073.60902419080 *T)
    sumL2  +=       4.201 * cos(5.28560721767  + 18.15924726470 *T)
    sumL2  +=       3.933 * cos(1.26665387164  + 853.19638175200 *T)
    sumL2  +=       5.351 * cos(3.65320121089  + 1272.68102562720 *T)
    sumL2  +=       4.392 * cos(2.27325303667  + 1368.66025284500 *T)
    sumL2  +=       3.482 * cos(1.53983001273  + 519.39602435610 *T)
    sumL2  +=       2.745 * cos(2.09685315627  + 1478.86657406440 *T)
    sumL2  +=       2.737 * cos(1.06017230524  + 1574.84580128220 *T)
    sumL2  +=       2.897 * cos(2.05128453665  + 1361.54670584420 *T)
    sumL2  +=       3.075 * cos(0.99085727534  + 191.95845443560 *T)
    sumL2  +=       2.462 * cos(2.37173605635  + 1471.75302706360 *T)
    sumL2  +=       2.203 * cos(2.47960567714  + 721.64941953020 *T)
    sumL2  +=       2.096 * cos(3.71482580504  + 1485.98012106520 *T)
    sumL2  +=       1.984 * cos(1.88475229557  + 1685.05212250160 *T)
    sumL2  +=       2.274 * cos(3.03360234351  + 1148.24761040620 *T)
    sumL2  +=       2.041 * cos(6.17114556019  + 330.61896365820 *T)
    sumL2  +=       1.451 * cos(4.72055072637  + 32.24332891440 *T)
    sumL2  +=       1.454 * cos(5.14703918585  + 1375.77379984580 *T)
    sumL2  +=       1.447 * cos(3.18833439444  + 635.96513305090 *T)
    sumL2  +=       1.403 * cos(4.26712075104  + 551.03160609700 *T)
    sumL2  +=       1.420 * cos(1.99288040133  + 629.60234557550 *T)
    sumL2  +=       1.269 * cos(0.03300387779  + 2125.87740737920 *T)
    sumL2  +=       1.276 * cos(2.26356919237  + 1788.14489672020 *T)
    sumL2  +=       1.189 * cos(1.70223550488  + 1677.93857550080 *T)
    sumL2  +=       1.182 * cos(2.18142313946  + 1795.25844372100 *T)
    sumL2  +=       1.366 * cos(1.27629917215  + 1038.04128918680 *T)
    sumL2  +=       1.306 * cos(4.76302079847  + 1062.56332392690 *T)
    sumL2  +=       1.109 * cos(2.97787130235  + 81.75213321620 *T)
    sumL2  +=       1.027 * cos(1.99236027398  + 295.05122865420 *T)
    sumL2  +=       1.349 * cos(4.01621534182  + 539.98590583310 *T)
    sumL2  +=       1.025 * cos(3.75336759986  + 28.45418800320 *T)
    sumL2  +=       0.977 * cos(3.01355125761  + 124.43341522100 *T)
    sumL2  +=       1.290 * cos(4.62594234857  + 2648.45482547300 *T)
    sumL2  +=       1.065 * cos(5.06153058155  + 1699.27921650320 *T)
    sumL2  +=       0.965 * cos(1.17716405513  + 99.91138048090 *T)
    sumL2  +=       1.021 * cos(1.90712102660  + 750.10360753340 *T)
    sumL2  +=       0.923 * cos(3.53450109212  + 227.52618943960 *T)
    sumL2  +=       1.059 * cos(0.13532061468  + 416.30325013750 *T)
    sumL2  +=       0.836 * cos(2.07492422755  + 1056.20053645150 *T)
    sumL2  +=       0.889 * cos(1.75177808106  + 1898.35121793960 *T)
    sumL2  +=       0.772 * cos(2.89217715561  + 2008.55753915900 *T)
    sumL2  +=       1.014 * cos(2.80847772922  + 1464.63948006280 *T)
    sumL2  +=       0.820 * cos(1.99735697577  + 2111.65031337760 *T)
    sumL2  +=       0.787 * cos(4.91912237671  + 1055.44977692610 *T)
    sumL2  +=       0.743 * cos(2.65209650690  + 106.27416795630 *T)
    sumL2  +=       0.705 * cos(0.08006443278  + 963.40270297140 *T)
    sumL2  +=       0.724 * cos(3.29664246938  + 628.85158605010 *T)
    sumL2  +=       0.791 * cos(1.64655202110  + 2001.44399215820 *T)
    sumL2  +=       0.822 * cos(2.74067639972  + 618.55664531160 *T)
    sumL2  +=       0.761 * cos(1.26393500358  + 1382.88734684660 *T)
    sumL2  +=       0.650 * cos(1.19590511216  + 422.66603761290 *T)
    sumL2  +=       0.677 * cos(1.88476058357  + 2104.53676637680 *T)
    sumL2  +=       0.681 * cos(5.47481665606  + 5760.49843189760 *T)
    sumL2  +=       0.681 * cos(3.11621209674  + 5746.27133789600 *T)
    sumL2  +=       0.644 * cos(4.68385640894  + 611.44309831080 *T)
    sumL2  +=       0.752 * cos(3.03497138894  + 2221.85663459700 *T)
    sumL2  +=       0.641 * cos(1.86274530783  + 636.71589257630 *T)
    sumL2  +=       0.614 * cos(3.07677356670  + 380.12776796000 *T)
    sumL2  +=       0.635 * cos(4.53916684689  + 9676.48103411560 *T)
    sumL2  +=       0.635 * cos(0.61458805483  + 9690.70812811720 *T)
    sumL2  +=       0.822 * cos(6.25170365084  + 423.41679713830 *T)
    sumL2  +=       0.762 * cos(4.32362906505  + 1802.37199072180 *T)
    sumL2  +=       0.582 * cos(0.84137872868  + 1891.23767093880 *T)
    sumL2  +=       0.558 * cos(3.96171840325  + 440.82528487760 *T)
    sumL2  +=       0.624 * cos(2.83657771014  + 1905.46476494040 *T)
    sumL2  +=       0.711 * cos(3.43538032357  + 824.74219374880 *T)
    sumL2  +=       0.517 * cos(1.10660016329  + 107.02492748170 *T)
    sumL2  +=       0.535 * cos(1.55761050176  + 1994.33044515740 *T)
    sumL2  +=       0.501 * cos(4.44389802599  + 647.01083331480 *T)
    sumL2  +=       0.414 * cos(5.37130370397  + 2228.97018159780 *T)
    sumL2  +=       0.533 * cos(2.54756313371  + 1781.03134971940 *T)
    sumL2  +=       0.393 * cos(1.26351262287  + 210.11770170030 *T)
    sumL2  +=       0.433 * cos(2.90103969634  + 1063.31408345230 *T)
    sumL2  +=       0.384 * cos(1.36194621083  + 203.00415469950 *T)
    sumL2  +=       0.440 * cos(1.46934545869  + 2214.74308759620 *T)
    sumL2  +=       0.424 * cos(4.98974282486  + 3178.14579056760 *T)
    sumL2  +=       0.338 * cos(2.72210106345  + 2324.94940881560 *T)
    sumL2  +=       0.332 * cos(0.37505564414  + 2655.56837247380 *T)
    sumL2  +=       0.318 * cos(6.11024720065  + 934.94851496820 *T)
    sumL2  +=       0.405 * cos(3.51005860013  + 2751.54759969160 *T)
    sumL2  +=       0.388 * cos(5.00609647265  + 2015.67108615980 *T)
    sumL2  +=       0.424 * cos(4.29668654117  + 5753.38488489680 *T)
    sumL2  +=       0.328 * cos(2.35571531981  + 1251.34038462480 *T)
    sumL2  +=       0.316 * cos(0.16949503062  + 1279.79457262800 *T)
    sumL2  +=       0.345 * cos(2.89328206121  + 2957.73314812880 *T)
    sumL2  +=       0.303 * cos(1.63964826684  + 2428.04218303420 *T)
    sumL2  +=       0.328 * cos(3.36132375845  + 1141.13406340540 *T)
    sumL2  +=       0.294 * cos(2.48947693371  + 2641.34127847220 *T)
    sumL2  +=       0.350 * cos(1.50537240918  + 2317.83586181480 *T)
    sumL2  +=       0.287 * cos(1.69638214958  + 2420.92863603340 *T)
    sumL2  +=       0.272 * cos(0.27466529753  + 319.57326339430 *T)
    sumL2  +=       0.303 * cos(2.43034117616  + 70.84944530420 *T)
    sumL2  +=       0.251 * cos(0.43544711316  + 3259.89792378380 *T)
    sumL2  +=       0.224 * cos(4.49752269293  + 5223.69391980220 *T)
    sumL2  +=       0.272 * cos(2.98590404673  + 1457.52593306200 *T)
    sumL2  +=       0.228 * cos(5.47896916415  + 1603.29998928540 *T)
    sumL2  +=       0.288 * cos(2.30146999217  + 2854.64037391020 *T)
    sumL2  +=       0.207 * cos(5.94297320087  + 9153.90361602180 *T)
    sumL2  +=       0.243 * cos(1.58604251447  + 2744.43405269080 *T)
    sumL2  +=       0.228 * cos(1.28182702946  + 2310.72231481400 *T)
    sumL2  +=       0.224 * cos(1.28623905132  + 3060.82592234740 *T)
    sumL2  +=       0.222 * cos(0.63265553397  + 3163.91869656600 *T)
    sumL2  +=       0.242 * cos(2.52382905368  + 3274.12501778540 *T)
    sumL2  +=       0.188 * cos(6.00513627145  + 92.04707395470 *T)
    sumL2  +=       0.239 * cos(1.93897157244  + 2413.81508903260 *T)
    sumL2  +=       0.214 * cos(1.14529237568  + 2531.13495725280 *T)
    sumL2  +=       0.200 * cos(3.42280996072  + 99.16062095550 *T)
    sumL2  +=       0.179 * cos(0.53892926207  + 2207.62954059540 *T)
    sumL2  +=       0.177 * cos(5.56545270243  + 2332.06295581640 *T)
    sumL2  +=       0.172 * cos(1.38604067808  + 945.99421523210 *T)
    sumL2  +=       0.203 * cos(0.41899069603  + 2840.41327990860 *T)
    sumL2  +=       0.231 * cos(2.26353330460  + 2097.42321937600 *T)
    sumL2  +=       0.228 * cos(3.82701076821  + 113.38771495710 *T)
    sumL2  +=       0.165 * cos(4.08776703733  + 6283.07584999140 *T)
    sumL2  +=       0.202 * cos(3.30429764992  + 3067.93946934820 *T)
    sumL2  +=       0.224 * cos(3.69285208525  + 2435.15573003500 *T)
    sumL2  +=       0.214 * cos(2.55756944911  + 2538.24850425360 *T)
    sumL2  +=       0.203 * cos(2.24205059922  + 67.66805156650 *T)
    sumL2  +=       0.152 * cos(5.48122906518  + 10213.28554621100 *T)
    sumL2  +=       0.191 * cos(2.68685722531  + 1773.91780271860 *T)
    sumL2  +=       0.189 * cos(2.95184620359  + 732.69511979410 *T)
    sumL2  +=       0.149 * cos(1.98737542735  + 1049.08698945070 *T)
    sumL2  +=       0.163 * cos(1.24084734609  + 3053.71237534660 *T)
    sumL2  +=       0.171 * cos(2.34210749987  + 1354.43315884340 *T)
    sumL2  +=       0.112 * cos(5.77407285790  + 547.85021235930 *T)
    sumL2  +=       0.124 * cos(0.14001204498  + 860.30992875280 *T)
    sumL2  +=       0.086 * cos(1.26924601636  + 511.53171782990 *T)
    sumL2  +=       0.114 * cos(5.15982838070  + 1592.25428902150 *T)
    sumL2  +=       0.091 * cos(1.48896790758  + 1567.73225428140 *T)
    sumL2  +=       0.086 * cos(4.34444949905  + 1069.67687092770 *T)

    # JUPITER Longitude
    sumL3  = 0
    sumL3  +=    6501.665 * cos(2.59862880482  + 7.11354700080 *T)
    sumL3  +=    1356.524 * cos(1.34635886411  + 529.69096509460 *T)
    sumL3  +=     470.716 * cos(2.47503977883  + 14.22709400160 *T)
    sumL3  +=     416.960 * cos(3.24451243214  + 536.80451209540 *T)
    sumL3  +=     352.851 * cos(2.97360159003  + 522.57741809380 *T)
    sumL3  +=     154.880 * cos(2.07565585817  + 1059.38193018920 *T)
    sumL3  +=      86.771 * cos(2.51431584316  + 515.46387109300 *T)
    sumL3  +=      33.538 * cos(3.82633794497  + 1066.49547719000 *T)
    sumL3  +=      44.378 * cos(0.00000000000  + 0.00000000000 *T)
    sumL3  +=      22.644 * cos(2.98231326774  + 543.91805909620 *T)
    sumL3  +=      23.737 * cos(1.27667172313  + 412.37109687440 *T)
    sumL3  +=      28.457 * cos(2.44754756058  + 206.18554843720 *T)
    sumL3  +=      19.798 * cos(2.10099934005  + 639.89728631400 *T)
    sumL3  +=      19.740 * cos(1.40255938973  + 419.48464387520 *T)
    sumL3  +=      18.768 * cos(1.59368403500  + 103.09277421860 *T)
    sumL3  +=      17.033 * cos(2.30214681202  + 21.34064100240 *T)
    sumL3  +=      16.774 * cos(2.59821460673  + 1589.07289528380 *T)
    sumL3  +=      16.214 * cos(3.14521117299  + 625.67019231240 *T)
    sumL3  +=      16.055 * cos(3.36030126297  + 1052.26838318840 *T)
    sumL3  +=      13.392 * cos(2.75973892202  + 95.97922721780 *T)
    sumL3  +=      13.234 * cos(2.53862244340  + 199.07200143640 *T)
    sumL3  +=      12.611 * cos(6.26578110400  + 426.59819087600 *T)
    sumL3  +=       8.637 * cos(2.26563256289  + 110.20632121940 *T)
    sumL3  +=       6.725 * cos(3.42566433316  + 309.27832265580 *T)
    sumL3  +=       8.701 * cos(1.76334960737  + 10.29494073850 *T)
    sumL3  +=       6.527 * cos(4.03869562907  + 728.76296653100 *T)
    sumL3  +=       5.368 * cos(5.25196153539  + 323.50541665740 *T)
    sumL3  +=       5.675 * cos(2.52096417685  + 508.35032409220 *T)
    sumL3  +=       5.399 * cos(2.91184687105  + 1045.15483618760 *T)
    sumL3  +=       3.996 * cos(4.30290261177  + 88.86568021700 *T)
    sumL3  +=       3.857 * cos(3.52381361552  + 302.16477565500 *T)
    sumL3  +=       3.774 * cos(4.09125315146  + 735.87651353180 *T)
    sumL3  +=       3.269 * cos(1.43175991274  + 956.28915597060 *T)
    sumL3  +=       2.783 * cos(4.35817507670  + 1596.18644228460 *T)
    sumL3  +=       2.661 * cos(1.25276590759  + 213.29909543800 *T)
    sumL3  +=       2.553 * cos(2.23785673285  + 117.31986822020 *T)
    sumL3  +=       2.371 * cos(2.89662409244  + 742.99006053260 *T)
    sumL3  +=       2.656 * cos(5.01505839848  + 838.96928775040 *T)
    sumL3  +=       1.948 * cos(2.77248294666  + 1169.58825140860 *T)
    sumL3  +=       2.279 * cos(2.35581871230  + 942.06206196900 *T)
    sumL3  +=       1.474 * cos(1.61011468581  + 220.41264243880 *T)
    sumL3  +=       1.457 * cos(3.09381959396  + 2118.76386037840 *T)
    sumL3  +=       1.937 * cos(5.01388256693  + 831.85574074960 *T)
    sumL3  +=       1.585 * cos(1.40097680805  + 405.25754987360 *T)
    sumL3  +=       1.257 * cos(3.97811260358  + 1155.36115740700 *T)
    sumL3  +=       1.227 * cos(3.45959919972  + 1073.60902419080 *T)
    sumL3  +=       0.986 * cos(3.39209446167  + 532.87235883230 *T)
    sumL3  +=       0.942 * cos(2.70200385825  + 191.95845443560 *T)
    sumL3  +=       0.828 * cos(1.48348768286  + 632.78373931320 *T)
    sumL3  +=       0.797 * cos(1.10706688850  + 1162.47470440780 *T)
    sumL3  +=       0.822 * cos(3.30295824153  + 1258.45393162560 *T)
    sumL3  +=       0.710 * cos(5.89798771980  + 853.19638175200 *T)
    sumL3  +=       0.766 * cos(3.66351539483  + 1581.95934828300 *T)
    sumL3  +=       0.722 * cos(3.74673245797  + 433.71173787680 *T)
    sumL3  +=       0.663 * cos(2.93063953915  + 1574.84580128220 *T)
    sumL3  +=       0.658 * cos(3.52797311863  + 525.75881183150 *T)
    sumL3  +=       0.609 * cos(4.14881313523  + 721.64941953020 *T)
    sumL3  +=       0.598 * cos(4.69454609357  + 81.75213321620 *T)
    sumL3  +=       0.668 * cos(1.96442971289  + 1272.68102562720 *T)
    sumL3  +=       0.515 * cos(1.57251270902  + 949.17560896980 *T)
    sumL3  +=       0.658 * cos(2.02329201466  + 526.50957135690 *T)
    sumL3  +=       0.517 * cos(4.35827478516  + 1368.66025284500 *T)
    sumL3  +=       0.510 * cos(4.95846155301  + 1148.24761040620 *T)
    sumL3  +=       0.507 * cos(4.31396370095  + 330.61896365820 *T)
    sumL3  +=       0.567 * cos(2.27813343743  + 551.03160609700 *T)
    sumL3  +=       0.480 * cos(3.86758235988  + 1361.54670584420 *T)
    sumL3  +=       0.383 * cos(0.24287136454  + 611.44309831080 *T)
    sumL3  +=       0.434 * cos(2.95461755540  + 1038.04128918680 *T)
    sumL3  +=       0.377 * cos(1.42957648215  + 124.43341522100 *T)
    sumL3  +=       0.391 * cos(4.07770324592  + 1471.75302706360 *T)
    sumL3  +=       0.385 * cos(4.70295179800  + 519.39602435610 *T)
    sumL3  +=       0.428 * cos(2.22472522305  + 539.98590583310 *T)
    sumL3  +=       0.343 * cos(4.83463725823  + 2125.87740737920 *T)
    sumL3  +=       0.394 * cos(4.52891996323  + 1464.63948006280 *T)
    sumL3  +=       0.305 * cos(2.02797683648  + 1485.98012106520 *T)
    sumL3  +=       0.283 * cos(0.97461612169  + 1905.46476494040 *T)
    sumL3  +=       0.276 * cos(3.83552772064  + 1062.56332392690 *T)
    sumL3  +=       0.351 * cos(2.06334334462  + 533.62311835770 *T)
    sumL3  +=       0.304 * cos(3.93228052293  + 1685.05212250160 *T)
    sumL3  +=       0.322 * cos(3.54763044791  + 846.08283475120 *T)
    sumL3  +=       0.345 * cos(4.18332148409  + 1788.14489672020 *T)
    sumL3  +=       0.253 * cos(3.12703531516  + 1994.33044515740 *T)
    sumL3  +=       0.257 * cos(1.05361498985  + 1478.86657406440 *T)
    sumL3  +=       0.232 * cos(1.69999081817  + 1692.16566950240 *T)
    sumL3  +=       0.225 * cos(2.51624149780  + 1891.23767093880 *T)
    sumL3  +=       0.217 * cos(4.58512911216  + 963.40270297140 *T)
    sumL3  +=       0.277 * cos(3.63353707701  + 1677.93857550080 *T)
    sumL3  +=       0.242 * cos(2.90163762388  + 2310.72231481400 *T)
    sumL3  +=       0.211 * cos(3.96419403991  + 295.05122865420 *T)
    sumL3  +=       0.199 * cos(5.17046500750  + 618.55664531160 *T)
    sumL3  +=       0.256 * cos(4.19052619061  + 1781.03134971940 *T)
    sumL3  +=       0.192 * cos(0.81556540966  + 2221.85663459700 *T)
    sumL3  +=       0.187 * cos(3.49895198981  + 2648.45482547300 *T)
    sumL3  +=       0.208 * cos(4.11838429822  + 2097.42321937600 *T)
    sumL3  +=       0.183 * cos(3.30680692414  + 1699.27921650320 *T)
    sumL3  +=       0.231 * cos(2.54516792766  + 1375.77379984580 *T)
    sumL3  +=       0.189 * cos(5.74277274755  + 2627.11418447060 *T)
    sumL3  +=       0.214 * cos(5.48031974537  + 1354.43315884340 *T)
    sumL3  +=       0.220 * cos(3.87471989410  + 2104.53676637680 *T)
    sumL3  +=       0.171 * cos(6.10827209399  + 1382.88734684660 *T)
    sumL3  +=       0.184 * cos(5.98415847544  + 750.10360753340 *T)
    sumL3  +=       0.171 * cos(5.25744961028  + 824.74219374880 *T)
    sumL3  +=       0.151 * cos(4.30799091626  + 2001.44399215820 *T)
    sumL3  +=       0.140 * cos(4.27089466070  + 1265.56747862640 *T)
    sumL3  +=       0.097 * cos(4.67188056608  + 647.01083331480 *T)
    sumL3  +=       0.088 * cos(2.43775210355  + 440.82528487760 *T)
    sumL3  +=       0.075 * cos(3.93105183253  + 1055.44977692610 *T)
    sumL3  +=       0.079 * cos(1.88533153220  + 934.94851496820 *T)
    sumL3  +=       0.077 * cos(3.80503143236  + 1603.29998928540 *T)

    # JUPITER Longitude
    sumL4  = 0
    sumL4  +=     669.483 * cos(0.85282421090  + 7.11354700080 *T)
    sumL4  +=      99.961 * cos(0.74258947751  + 14.22709400160 *T)
    sumL4  +=     114.019 * cos(3.14159265359  + 0.00000000000 *T)
    sumL4  +=      50.024 * cos(1.65346208248  + 536.80451209540 *T)
    sumL4  +=      43.585 * cos(5.82026386621  + 529.69096509460 *T)
    sumL4  +=      31.813 * cos(4.85829986650  + 522.57741809380 *T)
    sumL4  +=      14.742 * cos(4.29061635784  + 515.46387109300 *T)
    sumL4  +=       8.899 * cos(0.71478520741  + 1059.38193018920 *T)
    sumL4  +=       4.957 * cos(1.29502259434  + 543.91805909620 *T)
    sumL4  +=       4.484 * cos(2.31715516627  + 1066.49547719000 *T)
    sumL4  +=       4.251 * cos(0.48326797501  + 21.34064100240 *T)
    sumL4  +=       3.100 * cos(3.00245542678  + 412.37109687440 *T)
    sumL4  +=       2.055 * cos(0.39858940218  + 639.89728631400 *T)
    sumL4  +=       1.762 * cos(4.90536207307  + 625.67019231240 *T)
    sumL4  +=       1.902 * cos(4.25925620271  + 199.07200143640 *T)
    sumL4  +=       1.695 * cos(4.26147580803  + 206.18554843720 *T)
    sumL4  +=       1.375 * cos(5.25546955667  + 1052.26838318840 *T)
    sumL4  +=       1.203 * cos(4.71614633845  + 95.97922721780 *T)
    sumL4  +=       1.086 * cos(1.28604571172  + 1589.07289528380 *T)
    sumL4  +=       0.982 * cos(4.77990073662  + 1045.15483618760 *T)
    sumL4  +=       0.935 * cos(6.05847062188  + 88.86568021700 *T)
    sumL4  +=       0.916 * cos(5.77537499431  + 728.76296653100 *T)
    sumL4  +=       0.890 * cos(4.55299189579  + 426.59819087600 *T)
    sumL4  +=       0.784 * cos(3.40161567950  + 419.48464387520 *T)
    sumL4  +=       0.768 * cos(3.54672049322  + 103.09277421860 *T)
    sumL4  +=       0.670 * cos(0.52223307700  + 110.20632121940 *T)
    sumL4  +=       0.415 * cos(5.22809480633  + 302.16477565500 *T)
    sumL4  +=       0.393 * cos(6.24184621807  + 956.28915597060 *T)
    sumL4  +=       0.381 * cos(5.25466966040  + 309.27832265580 *T)
    sumL4  +=       0.421 * cos(0.59561318533  + 117.31986822020 *T)
    sumL4  +=       0.346 * cos(4.78348312106  + 508.35032409220 *T)
    sumL4  +=       0.319 * cos(3.47979828725  + 323.50541665740 *T)
    sumL4  +=       0.331 * cos(2.95893485883  + 1596.18644228460 *T)
    sumL4  +=       0.295 * cos(4.32713459459  + 942.06206196900 *T)
    sumL4  +=       0.319 * cos(0.47990052824  + 831.85574074960 *T)
    sumL4  +=       0.251 * cos(1.79898001222  + 1073.60902419080 *T)
    sumL4  +=       0.212 * cos(0.43917684084  + 220.41264243880 *T)
    sumL4  +=       0.188 * cos(1.12654974776  + 1169.58825140860 *T)
    sumL4  +=       0.188 * cos(2.16135407548  + 1361.54670584420 *T)
    sumL4  +=       0.180 * cos(3.43266428069  + 1148.24761040620 *T)
    sumL4  +=       0.164 * cos(1.92864127211  + 2118.76386037840 *T)
    sumL4  +=       0.157 * cos(3.02963907392  + 1272.68102562720 *T)
    sumL4  +=       0.093 * cos(5.60436000012  + 1581.95934828300 *T)
    sumL4  +=       0.085 * cos(5.02317256200  + 1155.36115740700 *T)
    sumL4  +=       0.075 * cos(3.13198879608  + 632.78373931320 *T)

    # JUPITER Longitude
    sumL5  = 0
    sumL5  +=      49.577 * cos(5.25658966184  + 7.11354700080 *T)
    sumL5  +=      15.761 * cos(5.25126837478  + 14.22709400160 *T)
    sumL5  +=       4.343 * cos(0.01461869263  + 536.80451209540 *T)
    sumL5  +=       1.526 * cos(1.09739911439  + 522.57741809380 *T)
    sumL5  +=       0.728 * cos(5.85949047619  + 543.91805909620 *T)
    sumL5  +=       0.694 * cos(0.87382487754  + 515.46387109300 *T)
    sumL5  +=       0.845 * cos(3.14159265359  + 0.00000000000 *T)
    sumL5  +=       0.456 * cos(0.81521692852  + 1066.49547719000 *T)
    sumL5  +=       0.293 * cos(5.62909357048  + 1059.38193018920 *T)
    sumL5  +=       0.090 * cos(0.21178119710  + 529.69096509460 *T)

    # JUPITER Latitude
    sumB0  = 0
    sumB0  +=  2268615.703 * cos(3.55852606718  + 529.69096509460 *T)
    sumB0  +=  109971.634 * cos(3.90809347389  + 1059.38193018920 *T)
    sumB0  +=  110090.358 * cos(0.00000000000  + 0.00000000000 *T)
    sumB0  +=    8101.427 * cos(3.60509573368  + 522.57741809380 *T)
    sumB0  +=    6043.996 * cos(4.25883108794  + 1589.07289528380 *T)
    sumB0  +=    6437.782 * cos(0.30627121409  + 536.80451209540 *T)
    sumB0  +=    1106.880 * cos(2.98534421928  + 1162.47470440780 *T)
    sumB0  +=     941.651 * cos(2.93619072405  + 1052.26838318840 *T)
    sumB0  +=     894.088 * cos(1.75447429921  + 7.11354700080 *T)
    sumB0  +=     767.280 * cos(2.15473594060  + 632.78373931320 *T)
    sumB0  +=     944.328 * cos(1.67522288396  + 426.59819087600 *T)
    sumB0  +=     684.220 * cos(3.67808770098  + 213.29909543800 *T)
    sumB0  +=     629.223 * cos(0.64343282328  + 1066.49547719000 *T)
    sumB0  +=     835.861 * cos(5.17881973234  + 103.09277421860 *T)
    sumB0  +=     531.670 * cos(2.70305954352  + 110.20632121940 *T)
    sumB0  +=     558.524 * cos(0.01354830508  + 846.08283475120 *T)
    sumB0  +=     464.449 * cos(1.17337249185  + 949.17560896980 *T)
    sumB0  +=     431.072 * cos(2.60825000494  + 419.48464387520 *T)
    sumB0  +=     351.433 * cos(4.61062990714  + 2118.76386037840 *T)
    sumB0  +=     123.148 * cos(3.34968181384  + 1692.16566950240 *T)
    sumB0  +=     115.038 * cos(5.04892295442  + 316.39186965660 *T)
    sumB0  +=     132.160 * cos(4.77816990670  + 742.99006053260 *T)
    sumB0  +=     103.402 * cos(2.31878999565  + 1478.86657406440 *T)
    sumB0  +=     116.379 * cos(1.38688232033  + 323.50541665740 *T)
    sumB0  +=     102.420 * cos(3.15293785436  + 1581.95934828300 *T)
    sumB0  +=     103.762 * cos(3.70103838110  + 515.46387109300 *T)
    sumB0  +=      78.650 * cos(3.98318653238  + 1265.56747862640 *T)
    sumB0  +=      69.935 * cos(2.56006216424  + 956.28915597060 *T)
    sumB0  +=      55.597 * cos(0.37500753017  + 1375.77379984580 *T)
    sumB0  +=      51.986 * cos(0.99007119033  + 1596.18644228460 *T)
    sumB0  +=      55.194 * cos(0.40176412035  + 525.75881183150 *T)
    sumB0  +=      63.456 * cos(4.50073574333  + 735.87651353180 *T)
    sumB0  +=      49.691 * cos(0.18649893085  + 543.91805909620 *T)
    sumB0  +=      48.831 * cos(3.57260550671  + 533.62311835770 *T)
    sumB0  +=      28.353 * cos(1.53532744749  + 625.67019231240 *T)
    sumB0  +=      29.209 * cos(5.43145863011  + 206.18554843720 *T)
    sumB0  +=      23.255 * cos(5.95197992848  + 838.96928775040 *T)
    sumB0  +=      22.841 * cos(6.19262787685  + 532.87235883230 *T)
    sumB0  +=      23.202 * cos(4.06473368575  + 526.50957135690 *T)
    sumB0  +=      24.436 * cos(6.10947656959  + 1169.58825140860 *T)
    sumB0  +=      21.116 * cos(4.96322972735  + 2648.45482547300 *T)
    sumB0  +=      17.879 * cos(3.08704395969  + 1795.25844372100 *T)
    sumB0  +=      16.234 * cos(4.83515727869  + 1368.66025284500 *T)
    sumB0  +=      21.314 * cos(2.69476951059  + 1045.15483618760 *T)
    sumB0  +=      15.740 * cos(1.15130330106  + 942.06206196900 *T)
    sumB0  +=      17.325 * cos(1.61550009206  + 14.22709400160 *T)
    sumB0  +=      13.396 * cos(2.30539585502  + 853.19638175200 *T)
    sumB0  +=      11.904 * cos(3.09811974536  + 2111.65031337760 *T)
    sumB0  +=      11.734 * cos(2.83006431723  + 2008.55753915900 *T)
    sumB0  +=      11.291 * cos(0.98957560201  + 433.71173787680 *T)
    sumB0  +=      11.830 * cos(4.76527836803  + 309.27832265580 *T)
    sumB0  +=      10.702 * cos(3.70181397065  + 2221.85663459700 *T)
    sumB0  +=      10.815 * cos(5.81958878617  + 1272.68102562720 *T)
    sumB0  +=      13.505 * cos(3.28126975760  + 1155.36115740700 *T)
    sumB0  +=      10.179 * cos(2.58691128827  + 117.31986822020 *T)
    sumB0  +=      10.632 * cos(5.23487936086  + 95.97922721780 *T)
    sumB0  +=       8.771 * cos(0.40456546655  + 220.41264243880 *T)
    sumB0  +=       7.439 * cos(2.94638292086  + 412.37109687440 *T)
    sumB0  +=       6.151 * cos(2.69100382247  + 380.12776796000 *T)
    sumB0  +=       5.028 * cos(0.72750312028  + 1055.44977692610 *T)
    sumB0  +=       4.939 * cos(0.73756716762  + 1905.46476494040 *T)
    sumB0  +=       5.421 * cos(4.08612438558  + 1685.05212250160 *T)
    sumB0  +=       5.936 * cos(4.32059910537  + 1063.31408345230 *T)
    sumB0  +=       4.737 * cos(4.09303016850  + 527.24328453980 *T)
    sumB0  +=       4.010 * cos(0.51530008355  + 1073.60902419080 *T)
    sumB0  +=       4.709 * cos(1.84067645204  + 984.60033162190 *T)
    sumB0  +=       3.974 * cos(1.33608029246  + 2125.87740737920 *T)
    sumB0  +=       3.762 * cos(3.58647039394  + 529.73914920440 *T)
    sumB0  +=       4.731 * cos(6.16377350841  + 532.13864564940 *T)
    sumB0  +=       4.666 * cos(5.88762905802  + 639.89728631400 *T)
    sumB0  +=       3.763 * cos(0.38865925413  + 529.64278098480 *T)
    sumB0  +=       3.409 * cos(4.05398247269  + 1898.35121793960 *T)
    sumB0  +=       3.457 * cos(3.43865563497  + 1485.98012106520 *T)
    sumB0  +=       4.229 * cos(2.23767157901  + 74.78159856730 *T)
    sumB0  +=       3.091 * cos(0.16470256025  + 1699.27921650320 *T)
    sumB0  +=       2.975 * cos(0.72268908074  + 530.65417294110 *T)
    sumB0  +=       3.162 * cos(1.25048416420  + 330.61896365820 *T)
    sumB0  +=       2.727 * cos(4.37679213321  + 149.56319713460 *T)
    sumB0  +=       2.837 * cos(0.05987107395  + 1439.50969814920 *T)
    sumB0  +=       2.983 * cos(3.25251207220  + 528.72775724810 *T)
    sumB0  +=       2.232 * cos(0.26149880534  + 1062.56332392690 *T)
    sumB0  +=       2.464 * cos(1.16913304420  + 453.42489381900 *T)
    sumB0  +=       2.596 * cos(3.30510149086  + 2324.94940881560 *T)
    sumB0  +=       1.988 * cos(2.85269577619  + 1574.84580128220 *T)
    sumB0  +=       2.527 * cos(5.94458202950  + 909.81873305460 *T)
    sumB0  +=       2.269 * cos(1.30379329597  + 3.93215326310 *T)
    sumB0  +=       1.742 * cos(4.49909767044  + 1258.45393162560 *T)
    sumB0  +=       1.714 * cos(4.12945878208  + 2001.44399215820 *T)
    sumB0  +=       2.029 * cos(3.97938086639  + 1056.20053645150 *T)
    sumB0  +=       1.667 * cos(0.36037092553  + 10213.28554621100 *T)
    sumB0  +=       1.579 * cos(6.11640144795  + 1802.37199072180 *T)
    sumB0  +=       1.393 * cos(3.69324470827  + 2214.74308759620 *T)
    sumB0  +=       1.604 * cos(1.98841031703  + 38.13303563780 *T)
    sumB0  +=       1.325 * cos(1.74025919863  + 529.16970023280 *T)
    sumB0  +=       1.451 * cos(2.39804501178  + 2428.04218303420 *T)
    sumB0  +=       1.594 * cos(2.07556780757  + 1021.24889455140 *T)
    sumB0  +=       1.320 * cos(1.33770977126  + 618.55664531160 *T)
    sumB0  +=       1.346 * cos(3.27591492540  + 2641.34127847220 *T)
    sumB0  +=       1.230 * cos(0.19552728220  + 305.34616939270 *T)
    sumB0  +=       1.223 * cos(2.86681556337  + 1382.88734684660 *T)
    sumB0  +=       1.324 * cos(2.23549334986  + 530.21222995640 *T)
    sumB0  +=       1.056 * cos(3.80579750957  + 76.26607127560 *T)
    sumB0  +=       1.050 * cos(4.68011652614  + 1788.14489672020 *T)
    sumB0  +=       1.226 * cos(5.34003255221  + 3178.14579056760 *T)
    sumB0  +=       1.009 * cos(3.19608028376  + 2538.24850425360 *T)
    sumB0  +=       1.266 * cos(3.04704446731  + 604.47256366190 *T)
    sumB0  +=       0.954 * cos(3.86932544808  + 728.76296653100 *T)
    sumB0  +=       1.124 * cos(1.59560367480  + 3.18139373770 *T)
    sumB0  +=       0.978 * cos(0.25223689838  + 983.11585891360 *T)
    sumB0  +=       0.948 * cos(0.21552742733  + 750.10360753340 *T)
    sumB0  +=       0.946 * cos(3.93927748120  + 508.35032409220 *T)
    sumB0  +=       0.920 * cos(1.14672086939  + 963.40270297140 *T)
    sumB0  +=       0.817 * cos(5.93809619876  + 831.85574074960 *T)
    sumB0  +=       0.770 * cos(2.96062737592  + 526.77020378780 *T)
    sumB0  +=       1.017 * cos(5.55711112145  + 199.07200143640 *T)
    sumB0  +=       0.761 * cos(1.38163787157  + 532.61172640140 *T)
    sumB0  +=       0.726 * cos(3.98337964395  + 2317.83586181480 *T)
    sumB0  +=       0.862 * cos(0.87975657414  + 490.33408917940 *T)
    sumB0  +=       0.868 * cos(3.44331872364  + 569.04784100980 *T)
    sumB0  +=       0.711 * cos(4.11107052823  + 2751.54759969160 *T)
    sumB0  +=       0.708 * cos(0.33555577415  + 528.94020556920 *T)
    sumB0  +=       0.708 * cos(4.00539820601  + 530.44172462000 *T)
    sumB0  +=       0.656 * cos(4.39568451439  + 519.39602435610 *T)
    sumB0  +=       0.801 * cos(4.03984430862  + 1364.72809958190 *T)
    sumB0  +=       0.679 * cos(1.18645749024  + 525.49817940060 *T)
    sumB0  +=       0.645 * cos(5.10510349996  + 1361.54670584420 *T)
    sumB0  +=       0.668 * cos(3.15607509055  + 533.88375078860 *T)
    sumB0  +=       0.663 * cos(0.73722024843  + 5223.69391980220 *T)
    sumB0  +=       0.663 * cos(1.57092786811  + 6283.07584999140 *T)
    sumB0  +=       0.543 * cos(0.26376529935  + 227.52618943960 *T)
    sumB0  +=       0.525 * cos(6.22318693939  + 539.98590583310 *T)
    sumB0  +=       0.513 * cos(4.98337900151  + 302.16477565500 *T)
    sumB0  +=       0.544 * cos(2.22227019273  + 2744.43405269080 *T)
    sumB0  +=       0.532 * cos(2.62425372687  + 99.16062095550 *T)
    sumB0  +=       0.602 * cos(1.56074089013  + 454.90936652730 *T)
    sumB0  +=       0.518 * cos(0.26343805959  + 551.03160609700 *T)
    sumB0  +=       0.516 * cos(1.09376390349  + 934.94851496820 *T)
    sumB0  +=       0.659 * cos(0.62560671589  + 1512.80682400820 *T)
    sumB0  +=       0.524 * cos(0.64710955846  + 524.06189080210 *T)
    sumB0  +=       0.516 * cos(3.69478866795  + 535.32003938710 *T)
    sumB0  +=       0.491 * cos(3.63039940597  + 2531.13495725280 *T)
    sumB0  +=       0.570 * cos(0.61976758791  + 540.73666535850 *T)
    sumB0  +=       0.496 * cos(2.19398015038  + 1514.29129671650 *T)
    sumB0  +=       0.532 * cos(0.20040217534  + 525.02509864860 *T)
    sumB0  +=       0.493 * cos(0.39160693598  + 224.34479570190 *T)
    sumB0  +=       0.449 * cos(0.62392433691  + 529.53090640020 *T)
    sumB0  +=       0.449 * cos(3.71676131146  + 529.85102378900 *T)
    sumB0  +=       0.450 * cos(5.02467015031  + 1048.33622992530 *T)
    sumB0  +=       0.428 * cos(5.44804660290  + 11.04570026390 *T)
    sumB0  +=       0.499 * cos(4.13924061941  + 534.35683154060 *T)
    sumB0  +=       0.528 * cos(1.76471074936  + 524.27433912320 *T)
    sumB0  +=       0.454 * cos(4.53321742354  + 1056.93424963440 *T)
    sumB0  +=       0.520 * cos(2.57406093768  + 535.10759106600 *T)
    sumB0  +=       0.398 * cos(1.40345870113  + 960.22130923370 *T)
    sumB0  +=       0.457 * cos(4.17708652827  + 2104.53676637680 *T)
    sumB0  +=       0.505 * cos(5.36536256321  + 1057.89745748090 *T)
    sumB0  +=       0.535 * cos(4.80455380313  + 1593.00504854690 *T)
    sumB0  +=       0.415 * cos(0.96548127237  + 2435.15573003500 *T)
    sumB0  +=       0.519 * cos(0.54543519483  + 1061.82961074400 *T)
    sumB0  +=       0.359 * cos(4.02704454075  + 1059.43011429900 *T)
    sumB0  +=       0.356 * cos(2.66818105522  + 835.03713448730 *T)
    sumB0  +=       0.443 * cos(5.27513700376  + 1.48447270830 *T)
    sumB0  +=       0.358 * cos(5.94423960514  + 440.82528487760 *T)
    sumB0  +=       0.471 * cos(6.05791940453  + 1471.75302706360 *T)
    sumB0  +=       0.386 * cos(2.15984900214  + 9153.90361602180 *T)
    sumB0  +=       0.424 * cos(2.70929670030  + 1038.04128918680 *T)
    sumB0  +=       0.359 * cos(0.82922836987  + 1059.33374607940 *T)
    sumB0  +=       0.310 * cos(0.88102053266  + 529.90341341570 *T)
    sumB0  +=       0.310 * cos(3.45966511571  + 529.47851677350 *T)
    sumB0  +=       0.300 * cos(3.70331799503  + 2634.22773147140 *T)
    sumB0  +=       0.292 * cos(2.63594456361  + 415.55249061210 *T)
    sumB0  +=       0.279 * cos(1.60669121578  + 643.82943957710 *T)
    sumB0  +=       0.291 * cos(5.83134071820  + 1148.24761040620 *T)
    sumB0  +=       0.370 * cos(5.71572992274  + 531.17543780290 *T)
    sumB0  +=       0.268 * cos(5.39275891813  + 1891.23767093880 *T)
    sumB0  +=       0.275 * cos(3.34108666036  + 518.64526483070 *T)
    sumB0  +=       0.269 * cos(1.06051406954  + 1585.14074202070 *T)
    sumB0  +=       0.306 * cos(2.50289017370  + 511.53171782990 *T)
    sumB0  +=       0.295 * cos(1.84394223501  + 547.85021235930 *T)
    sumB0  +=       0.254 * cos(2.98312992496  + 1134.16352875650 *T)
    sumB0  +=       0.289 * cos(1.86070918711  + 21.34064100240 *T)
    sumB0  +=       0.265 * cos(4.93075479744  + 679.25416222920 *T)
    sumB0  +=       0.250 * cos(0.42860925124  + 1969.20066324380 *T)
    sumB0  +=       0.308 * cos(2.67237933272  + 2957.73314812880 *T)
    sumB0  +=       0.313 * cos(4.88085697819  + 528.20649238630 *T)
    sumB0  +=       0.222 * cos(4.78828764413  + 514.71311156760 *T)
    sumB0  +=       0.221 * cos(4.32763468981  + 1677.93857550080 *T)
    sumB0  +=       0.217 * cos(3.46278526461  + 2950.61960112800 *T)
    sumB0  +=       0.216 * cos(0.52207667980  + 2228.97018159780 *T)
    sumB0  +=       0.214 * cos(5.83569926578  + 544.66881862160 *T)
    sumB0  +=       0.283 * cos(2.88709716090  + 35.42472265210 *T)
    sumB0  +=       0.272 * cos(1.65708415457  + 3060.82592234740 *T)
    sumB0  +=       0.234 * cos(1.68821537711  + 2655.56837247380 *T)
    sumB0  +=       0.205 * cos(3.36186888290  + 2847.52682690940 *T)
    sumB0  +=       0.264 * cos(3.62722625694  + 2420.92863603340 *T)
    sumB0  +=       0.191 * cos(4.26821147044  + 430.53034413910 *T)
    sumB0  +=       0.179 * cos(3.91470663005  + 3340.61242669980 *T)
    sumB0  +=       0.180 * cos(0.04531671003  + 387.24131496080 *T)
    sumB0  +=       0.241 * cos(4.03927631611  + 494.26624244250 *T)
    sumB0  +=       0.176 * cos(4.26298906325  + 672.14061522840 *T)
    sumB0  +=       0.187 * cos(2.72587420586  + 299.12639426920 *T)
    sumB0  +=       0.234 * cos(1.34474827450  + 173.94221952280 *T)
    sumB0  +=       0.171 * cos(0.85473611718  + 1603.29998928540 *T)
    sumB0  +=       0.224 * cos(0.33130232434  + 565.11568774670 *T)
    sumB0  +=       0.200 * cos(1.27632489123  + 39.35687591520 *T)
    sumB0  +=       0.170 * cos(4.96479470273  + 1464.63948006280 *T)
    sumB0  +=       0.211 * cos(1.00937080256  + 523.54062594030 *T)
    sumB0  +=       0.210 * cos(3.75793720248  + 2854.64037391020 *T)
    sumB0  +=       0.162 * cos(5.87784787295  + 3480.31056622260 *T)
    sumB0  +=       0.163 * cos(4.62850343495  + 2015.67108615980 *T)
    sumB0  +=       0.191 * cos(3.33159283750  + 535.84130424890 *T)
    sumB0  +=       0.151 * cos(1.17096741034  + 1060.34513803570 *T)
    sumB0  +=       0.160 * cos(1.81852636004  + 312.45971639350 *T)
    sumB0  +=       0.158 * cos(2.59595816107  + 529.43033266370 *T)
    sumB0  +=       0.158 * cos(1.74472748730  + 529.95159752550 *T)
    sumB0  +=       0.173 * cos(3.62399350412  + 230.56457082540 *T)
    sumB0  +=       0.142 * cos(0.70435921398  + 522.52923398400 *T)
    sumB0  +=       0.144 * cos(5.35763122430  + 107.02492748170 *T)
    sumB0  +=       0.144 * cos(6.13954848857  + 1158.54255114470 *T)
    sumB0  +=       0.178 * cos(0.27566275049  + 3906.90875709860 *T)
    sumB0  +=       0.126 * cos(5.14832919826  + 2207.62954059540 *T)
    sumB0  +=       0.126 * cos(3.41994798109  + 2.44768055480 *T)
    sumB0  +=       0.127 * cos(0.39825164051  + 70.84944530420 *T)
    sumB0  +=       0.123 * cos(4.77865550523  + 2524.02141025200 *T)
    sumB0  +=       0.123 * cos(0.46184813516  + 647.01083331480 *T)
    sumB0  +=       0.144 * cos(3.60261852727  + 1058.41872234270 *T)
    sumB0  +=       0.158 * cos(3.76231915252  + 92.04707395470 *T)
    sumB0  +=       0.119 * cos(4.08266911415  + 1585.89150154610 *T)
    sumB0  +=       0.125 * cos(2.35496721797  + 3163.91869656600 *T)
    sumB0  +=       0.122 * cos(3.21027426317  + 3377.21779200400 *T)
    sumB0  +=       0.121 * cos(3.39770381916  + 18.15924726470 *T)
    sumB0  +=       0.131 * cos(1.67926417552  + 1289.94650101460 *T)
    sumB0  +=       0.115 * cos(2.35735471566  + 1550.93985964600 *T)
    sumB0  +=       0.126 * cos(2.40833814513  + 106.27416795630 *T)
    sumB0  +=       0.131 * cos(1.37610474529  + 1023.95720753710 *T)
    sumB0  +=       0.121 * cos(1.60252617273  + 10.29494073850 *T)
    sumB0  +=       0.121 * cos(0.61420823557  + 1592.25428902150 *T)
    sumB0  +=       0.135 * cos(3.60177675518  + 124.43341522100 *T)
    sumB0  +=       0.137 * cos(2.41724947062  + 3274.12501778540 *T)
    sumB0  +=       0.129 * cos(0.09702914345  + 2332.06295581640 *T)
    sumB0  +=       0.093 * cos(4.88949890397  + 1098.73880610440 *T)
    sumB0  +=       0.106 * cos(5.18592950792  + 2281.23049651060 *T)
    sumB0  +=       0.114 * cos(2.96523316419  + 1166.40685767090 *T)
    sumB0  +=       0.092 * cos(1.65166124027  + 860.30992875280 *T)
    sumB0  +=       0.102 * cos(3.64093193142  + 3171.03224356680 *T)
    sumB0  +=       0.103 * cos(1.63066232967  + 1894.41906467650 *T)
    sumB0  +=       0.080 * cos(0.38766601876  + 4694.00295470760 *T)
    sumB0  +=       0.074 * cos(3.86865238736  + 3067.93946934820 *T)
    sumB0  +=       0.095 * cos(1.66362447044  + 1151.42900414390 *T)

    # JUPITER Latitude
    sumB1  = 0
    sumB1  +=  177351.787 * cos(5.70166488486  + 529.69096509460 *T)
    sumB1  +=    3230.171 * cos(5.77941619340  + 1059.38193018920 *T)
    sumB1  +=    3081.364 * cos(5.47464296527  + 522.57741809380 *T)
    sumB1  +=    2211.914 * cos(4.73477480209  + 536.80451209540 *T)
    sumB1  +=    1694.232 * cos(3.14159265359  + 0.00000000000 *T)
    sumB1  +=     346.445 * cos(4.74595174109  + 1052.26838318840 *T)
    sumB1  +=     234.264 * cos(5.18856099929  + 1066.49547719000 *T)
    sumB1  +=     196.154 * cos(6.18554286642  + 7.11354700080 *T)
    sumB1  +=     150.468 * cos(3.92721226087  + 1589.07289528380 *T)
    sumB1  +=     114.128 * cos(3.43897271830  + 632.78373931320 *T)
    sumB1  +=      96.667 * cos(2.91426304090  + 949.17560896980 *T)
    sumB1  +=      76.599 * cos(2.50522188662  + 103.09277421860 *T)
    sumB1  +=      81.671 * cos(5.07666097497  + 1162.47470440780 *T)
    sumB1  +=      76.572 * cos(0.61288981445  + 419.48464387520 *T)
    sumB1  +=      73.875 * cos(5.49958292155  + 515.46387109300 *T)
    sumB1  +=      49.915 * cos(3.94799616572  + 735.87651353180 *T)
    sumB1  +=      60.544 * cos(5.44740084359  + 213.29909543800 *T)
    sumB1  +=      36.561 * cos(4.69828392839  + 543.91805909620 *T)
    sumB1  +=      46.032 * cos(0.53850360901  + 110.20632121940 *T)
    sumB1  +=      45.123 * cos(1.89516645239  + 846.08283475120 *T)
    sumB1  +=      36.019 * cos(6.10952578764  + 316.39186965660 *T)
    sumB1  +=      31.975 * cos(4.92452714629  + 1581.95934828300 *T)
    sumB1  +=      21.015 * cos(5.62957731410  + 1596.18644228460 *T)
    sumB1  +=      23.156 * cos(5.84829490183  + 323.50541665740 *T)
    sumB1  +=      24.719 * cos(3.94107395247  + 2118.76386037840 *T)
    sumB1  +=      17.274 * cos(5.65310656429  + 533.62311835770 *T)
    sumB1  +=      16.521 * cos(5.89840100621  + 526.50957135690 *T)
    sumB1  +=      16.698 * cos(5.66663034948  + 1265.56747862640 *T)
    sumB1  +=      15.815 * cos(4.43314786393  + 1045.15483618760 *T)
    sumB1  +=      13.398 * cos(4.30179033605  + 532.87235883230 *T)
    sumB1  +=      11.744 * cos(1.80990486955  + 956.28915597060 *T)
    sumB1  +=      11.925 * cos(4.30094564154  + 525.75881183150 *T)
    sumB1  +=       9.514 * cos(2.02589667166  + 206.18554843720 *T)
    sumB1  +=      10.542 * cos(6.15533910933  + 14.22709400160 *T)
    sumB1  +=       8.414 * cos(3.92910450340  + 1478.86657406440 *T)
    sumB1  +=       8.099 * cos(4.20152809071  + 1169.58825140860 *T)
    sumB1  +=       7.712 * cos(2.99160389601  + 942.06206196900 *T)
    sumB1  +=       8.825 * cos(1.55897920307  + 426.59819087600 *T)
    sumB1  +=       8.884 * cos(4.87430124264  + 1155.36115740700 *T)
    sumB1  +=       7.793 * cos(3.84684930196  + 625.67019231240 *T)
    sumB1  +=       5.646 * cos(3.40915964493  + 639.89728631400 *T)
    sumB1  +=       4.615 * cos(0.83374662294  + 117.31986822020 *T)
    sumB1  +=       4.020 * cos(5.50502127885  + 433.71173787680 *T)
    sumB1  +=       3.704 * cos(0.90226777963  + 95.97922721780 *T)
    sumB1  +=       3.859 * cos(0.69640284662  + 853.19638175200 *T)
    sumB1  +=       3.091 * cos(5.09115860882  + 1073.60902419080 *T)
    sumB1  +=       3.360 * cos(5.10133284081  + 1692.16566950240 *T)
    sumB1  +=       2.892 * cos(4.90418916660  + 220.41264243880 *T)
    sumB1  +=       2.772 * cos(5.09066125724  + 2111.65031337760 *T)
    sumB1  +=       2.425 * cos(3.74438653232  + 742.99006053260 *T)
    sumB1  +=       2.558 * cos(5.46955948791  + 1795.25844372100 *T)
    sumB1  +=       2.466 * cos(4.22278355430  + 2648.45482547300 *T)
    sumB1  +=       1.968 * cos(0.57192251841  + 309.27832265580 *T)
    sumB1  +=       1.794 * cos(4.60765219417  + 1272.68102562720 *T)
    sumB1  +=       1.822 * cos(1.98842964323  + 1375.77379984580 *T)
    sumB1  +=       1.703 * cos(6.12660562937  + 2125.87740737920 *T)
    sumB1  +=       2.011 * cos(5.00936865256  + 412.37109687440 *T)
    sumB1  +=       1.645 * cos(0.08830372958  + 1063.31408345230 *T)
    sumB1  +=       1.875 * cos(5.81006158403  + 330.61896365820 *T)
    sumB1  +=       1.741 * cos(4.58650290431  + 1574.84580128220 *T)
    sumB1  +=       1.529 * cos(5.81660291389  + 1258.45393162560 *T)
    sumB1  +=       1.916 * cos(0.85150399517  + 1368.66025284500 *T)
    sumB1  +=       1.614 * cos(4.36839107221  + 728.76296653100 *T)
    sumB1  +=       1.510 * cos(2.79374165455  + 1485.98012106520 *T)
    sumB1  +=       1.333 * cos(4.84260898693  + 1062.56332392690 *T)
    sumB1  +=       1.359 * cos(5.16511980864  + 838.96928775040 *T)
    sumB1  +=       1.165 * cos(5.66275740881  + 508.35032409220 *T)
    sumB1  +=       1.092 * cos(4.68797557406  + 1699.27921650320 *T)
    sumB1  +=       1.438 * cos(5.78105679279  + 1056.20053645150 *T)
    sumB1  +=       1.083 * cos(3.99886917926  + 1471.75302706360 *T)
    sumB1  +=       1.002 * cos(4.79949608524  + 1055.44977692610 *T)
    sumB1  +=       0.749 * cos(6.14400862030  + 519.39602435610 *T)
    sumB1  +=       0.657 * cos(5.63765568876  + 1898.35121793960 *T)
    sumB1  +=       0.702 * cos(5.04126574492  + 1685.05212250160 *T)
    sumB1  +=       0.607 * cos(3.15707515246  + 618.55664531160 *T)
    sumB1  +=       0.587 * cos(1.37658820775  + 199.07200143640 *T)
    sumB1  +=       0.552 * cos(4.80657729450  + 551.03160609700 *T)
    sumB1  +=       0.494 * cos(4.43417307482  + 539.98590583310 *T)
    sumB1  +=       0.517 * cos(0.05161181997  + 3.18139373770 *T)
    sumB1  +=       0.469 * cos(3.81715950042  + 2008.55753915900 *T)
    sumB1  +=       0.415 * cos(1.34693184108  + 1382.88734684660 *T)
    sumB1  +=       0.382 * cos(4.86764073919  + 227.52618943960 *T)
    sumB1  +=       0.473 * cos(1.72405831407  + 532.13864564940 *T)
    sumB1  +=       0.458 * cos(4.44604993015  + 1038.04128918680 *T)
    sumB1  +=       0.376 * cos(2.23190744786  + 529.64278098480 *T)
    sumB1  +=       0.451 * cos(3.75869883836  + 984.60033162190 *T)
    sumB1  +=       0.376 * cos(5.42971857629  + 529.73914920440 *T)
    sumB1  +=       0.389 * cos(1.92698506631  + 525.02509864860 *T)
    sumB1  +=       0.364 * cos(3.35456685746  + 2221.85663459700 *T)
    sumB1  +=       0.476 * cos(5.93625415892  + 527.24328453980 *T)
    sumB1  +=       0.383 * cos(6.12255867339  + 149.56319713460 *T)
    sumB1  +=       0.301 * cos(4.09378934049  + 440.82528487760 *T)
    sumB1  +=       0.310 * cos(5.58150418981  + 2428.04218303420 *T)
    sumB1  +=       0.282 * cos(4.85996662231  + 1788.14489672020 *T)
    sumB1  +=       0.298 * cos(5.09589374634  + 528.72775724810 *T)
    sumB1  +=       0.340 * cos(4.56537070220  + 750.10360753340 *T)
    sumB1  +=       0.272 * cos(2.35346960340  + 534.35683154060 *T)
    sumB1  +=       0.360 * cos(3.91050161665  + 74.78159856730 *T)
    sumB1  +=       0.299 * cos(1.43093538841  + 909.81873305460 *T)
    sumB1  +=       0.297 * cos(2.56584512211  + 530.65417294110 *T)
    sumB1  +=       0.235 * cos(4.81644489422  + 535.10759106600 *T)
    sumB1  +=       0.306 * cos(0.68420442848  + 380.12776796000 *T)
    sumB1  +=       0.236 * cos(4.63162956792  + 526.77020378780 *T)
    sumB1  +=       0.270 * cos(0.18549916939  + 21.34064100240 *T)
    sumB1  +=       0.288 * cos(4.26655874393  + 1802.37199072180 *T)
    sumB1  +=       0.196 * cos(5.35950443033  + 2214.74308759620 *T)
    sumB1  +=       0.190 * cos(4.54615193260  + 2104.53676637680 *T)
    sumB1  +=       0.193 * cos(4.35426216497  + 511.53171782990 *T)
    sumB1  +=       0.178 * cos(4.51895208036  + 3178.14579056760 *T)
    sumB1  +=       0.194 * cos(0.57050756837  + 1361.54670584420 *T)
    sumB1  +=       0.200 * cos(1.48040474749  + 302.16477565500 *T)
    sumB1  +=       0.168 * cos(5.40141749419  + 524.27433912320 *T)
    sumB1  +=       0.152 * cos(0.68077486546  + 1905.46476494040 *T)
    sumB1  +=       0.149 * cos(1.06678990744  + 831.85574074960 *T)
    sumB1  +=       0.182 * cos(3.62401009613  + 38.13303563780 *T)
    sumB1  +=       0.176 * cos(5.64331384323  + 963.40270297140 *T)
    sumB1  +=       0.184 * cos(4.48850356629  + 604.47256366190 *T)
    sumB1  +=       0.133 * cos(5.45026366125  + 2641.34127847220 *T)
    sumB1  +=       0.143 * cos(2.21577268292  + 1439.50969814920 *T)
    sumB1  +=       0.130 * cos(4.88155705493  + 2531.13495725280 *T)
    sumB1  +=       0.129 * cos(6.15206333598  + 547.85021235930 *T)
    sumB1  +=       0.133 * cos(5.43193972385  + 1603.29998928540 *T)
    sumB1  +=       0.133 * cos(3.49297492409  + 529.16970023280 *T)
    sumB1  +=       0.132 * cos(3.98820790955  + 530.21222995640 *T)
    sumB1  +=       0.118 * cos(5.38352943814  + 1891.23767093880 *T)
    sumB1  +=       0.133 * cos(5.65694269884  + 76.26607127560 *T)
    sumB1  +=       0.145 * cos(2.94976686191  + 454.90936652730 *T)
    sumB1  +=       0.115 * cos(3.29206553804  + 3.93215326310 *T)
    sumB1  +=       0.102 * cos(4.48856749557  + 2001.44399215820 *T)
    sumB1  +=       0.106 * cos(6.08434275898  + 10.29494073850 *T)
    sumB1  +=       0.093 * cos(5.84737771840  + 2324.94940881560 *T)
    sumB1  +=       0.101 * cos(0.15815934254  + 2655.56837247380 *T)
    sumB1  +=       0.115 * cos(3.59221021604  + 2015.67108615980 *T)
    sumB1  +=       0.103 * cos(4.70399583323  + 305.34616939270 *T)
    sumB1  +=       0.084 * cos(0.44180206332  + 1593.00504854690 *T)
    sumB1  +=       0.092 * cos(2.44863388631  + 490.33408917940 *T)
    sumB1  +=       0.087 * cos(6.23817512863  + 6283.07584999140 *T)
    sumB1  +=       0.095 * cos(3.30154605532  + 2317.83586181480 *T)
    sumB1  +=       0.072 * cos(1.90578907085  + 528.94020556920 *T)
    sumB1  +=       0.072 * cos(5.57619428876  + 530.44172462000 *T)
    sumB1  +=       0.078 * cos(5.97323507836  + 1585.89150154610 *T)

    # JUPITER Latitude
    sumB2  = 0
    sumB2  +=    8094.051 * cos(1.46322843658  + 529.69096509460 *T)
    sumB2  +=     742.415 * cos(0.95691639003  + 522.57741809380 *T)
    sumB2  +=     813.244 * cos(3.14159265359  + 0.00000000000 *T)
    sumB2  +=     398.951 * cos(2.89888666447  + 536.80451209540 *T)
    sumB2  +=     342.226 * cos(1.44683789727  + 1059.38193018920 *T)
    sumB2  +=      73.948 * cos(0.40724675866  + 1052.26838318840 *T)
    sumB2  +=      46.151 * cos(3.48036895772  + 1066.49547719000 *T)
    sumB2  +=      29.314 * cos(0.99088831805  + 515.46387109300 *T)
    sumB2  +=      29.717 * cos(1.92504171329  + 1589.07289528380 *T)
    sumB2  +=      22.753 * cos(4.27124052435  + 7.11354700080 *T)
    sumB2  +=      13.916 * cos(2.92242387338  + 543.91805909620 *T)
    sumB2  +=      12.067 * cos(5.22168932482  + 632.78373931320 *T)
    sumB2  +=      10.703 * cos(4.88024222475  + 949.17560896980 *T)
    sumB2  +=       6.078 * cos(6.21089108431  + 1045.15483618760 *T)
    sumB2  +=       5.935 * cos(0.52977760072  + 1581.95934828300 *T)
    sumB2  +=       5.037 * cos(1.43444929374  + 526.50957135690 *T)
    sumB2  +=       4.564 * cos(0.91811732585  + 1162.47470440780 *T)
    sumB2  +=       4.547 * cos(4.01953745202  + 1596.18644228460 *T)
    sumB2  +=       5.098 * cos(6.03169795231  + 735.87651353180 *T)
    sumB2  +=       3.593 * cos(4.54080164408  + 110.20632121940 *T)
    sumB2  +=       3.443 * cos(1.38618954572  + 533.62311835770 *T)
    sumB2  +=       3.277 * cos(4.39650286553  + 14.22709400160 *T)
    sumB2  +=       3.407 * cos(0.42275631534  + 419.48464387520 *T)
    sumB2  +=       2.904 * cos(2.06041641723  + 316.39186965660 *T)
    sumB2  +=       2.541 * cos(3.98323842017  + 323.50541665740 *T)
    sumB2  +=       3.113 * cos(2.48079280193  + 2118.76386037840 *T)
    sumB2  +=       3.061 * cos(2.39880866911  + 532.87235883230 *T)
    sumB2  +=       2.155 * cos(4.77990637140  + 942.06206196900 *T)
    sumB2  +=       2.143 * cos(3.88727338786  + 426.59819087600 *T)
    sumB2  +=       2.252 * cos(0.37196434120  + 1155.36115740700 *T)
    sumB2  +=       2.019 * cos(3.89985000464  + 846.08283475120 *T)
    sumB2  +=       1.857 * cos(1.19658907851  + 103.09277421860 *T)
    sumB2  +=       1.683 * cos(1.42264195434  + 1265.56747862640 *T)
    sumB2  +=       2.313 * cos(0.87671613055  + 213.29909543800 *T)
    sumB2  +=       1.443 * cos(2.38565505909  + 1169.58825140860 *T)
    sumB2  +=       1.823 * cos(5.80106463776  + 625.67019231240 *T)
    sumB2  +=       1.728 * cos(2.24114678267  + 525.75881183150 *T)
    sumB2  +=       1.198 * cos(0.03252059731  + 956.28915597060 *T)
    sumB2  +=       1.138 * cos(3.46420904745  + 1073.60902419080 *T)
    sumB2  +=       1.086 * cos(5.35279146700  + 117.31986822020 *T)
    sumB2  +=       0.840 * cos(2.89946334223  + 95.97922721780 *T)
    sumB2  +=       0.746 * cos(5.53017890231  + 1478.86657406440 *T)
    sumB2  +=       0.944 * cos(4.05587053500  + 206.18554843720 *T)
    sumB2  +=       0.758 * cos(3.74770617289  + 433.71173787680 *T)
    sumB2  +=       0.673 * cos(1.26396626349  + 508.35032409220 *T)
    sumB2  +=       0.889 * cos(6.07878453176  + 728.76296653100 *T)
    sumB2  +=       0.600 * cos(1.82954494089  + 639.89728631400 *T)
    sumB2  +=       0.589 * cos(1.23625943417  + 1258.45393162560 *T)
    sumB2  +=       0.619 * cos(0.67923057477  + 838.96928775040 *T)
    sumB2  +=       0.566 * cos(5.36336098734  + 742.99006053260 *T)
    sumB2  +=       0.648 * cos(5.32990375008  + 853.19638175200 *T)
    sumB2  +=       0.553 * cos(3.15511946637  + 220.41264243880 *T)
    sumB2  +=       0.432 * cos(1.03719283016  + 1692.16566950240 *T)
    sumB2  +=       0.435 * cos(1.65056479007  + 519.39602435610 *T)
    sumB2  +=       0.430 * cos(1.41830384501  + 412.37109687440 *T)
    sumB2  +=       0.431 * cos(2.20986254651  + 1368.66025284500 *T)
    sumB2  +=       0.415 * cos(4.35372561905  + 330.61896365820 *T)
    sumB2  +=       0.438 * cos(0.16552277290  + 1574.84580128220 *T)
    sumB2  +=       0.312 * cos(4.50639455819  + 2125.87740737920 *T)
    sumB2  +=       0.280 * cos(3.01441283033  + 551.03160609700 *T)
    sumB2  +=       0.309 * cos(0.67399908949  + 2111.65031337760 *T)
    sumB2  +=       0.301 * cos(3.06868080871  + 1062.56332392690 *T)
    sumB2  +=       0.236 * cos(1.94696842200  + 1485.98012106520 *T)
    sumB2  +=       0.235 * cos(3.41850395941  + 199.07200143640 *T)
    sumB2  +=       0.246 * cos(2.61803442505  + 309.27832265580 *T)
    sumB2  +=       0.238 * cos(2.56643737684  + 539.98590583310 *T)
    sumB2  +=       0.248 * cos(2.96997778167  + 2648.45482547300 *T)
    sumB2  +=       0.209 * cos(5.82481690851  + 1471.75302706360 *T)
    sumB2  +=       0.205 * cos(1.20202002469  + 1056.20053645150 *T)
    sumB2  +=       0.188 * cos(0.97113663101  + 1685.05212250160 *T)
    sumB2  +=       0.137 * cos(2.91203499563  + 1699.27921650320 *T)
    sumB2  +=       0.131 * cos(1.79274504072  + 1063.31408345230 *T)
    sumB2  +=       0.161 * cos(1.05926568614  + 1795.25844372100 *T)
    sumB2  +=       0.112 * cos(2.62660288825  + 440.82528487760 *T)
    sumB2  +=       0.110 * cos(3.56263668146  + 227.52618943960 *T)
    sumB2  +=       0.114 * cos(6.13907482464  + 1038.04128918680 *T)
    sumB2  +=       0.103 * cos(4.64287101040  + 3.18139373770 *T)
    sumB2  +=       0.123 * cos(4.81268110532  + 21.34064100240 *T)
    sumB2  +=       0.102 * cos(4.27603827970  + 1375.77379984580 *T)
    sumB2  +=       0.089 * cos(1.22926014128  + 1898.35121793960 *T)
    sumB2  +=       0.080 * cos(0.62129648755  + 831.85574074960 *T)

    # JUPITER Latitude
    sumB3  = 0
    sumB3  +=     251.624 * cos(3.38087923084  + 529.69096509460 *T)
    sumB3  +=     121.738 * cos(2.73311837200  + 522.57741809380 *T)
    sumB3  +=      48.694 * cos(1.03689996685  + 536.80451209540 *T)
    sumB3  +=      10.988 * cos(2.31463561347  + 1052.26838318840 *T)
    sumB3  +=       8.067 * cos(2.76729757621  + 515.46387109300 *T)
    sumB3  +=       6.205 * cos(1.78115827370  + 1066.49547719000 *T)
    sumB3  +=       7.287 * cos(4.25268318975  + 1059.38193018920 *T)
    sumB3  +=       3.627 * cos(1.13028917221  + 543.91805909620 *T)
    sumB3  +=       2.798 * cos(3.14159265359  + 0.00000000000 *T)
    sumB3  +=       1.898 * cos(2.28934054087  + 7.11354700080 *T)
    sumB3  +=       1.643 * cos(1.77507208483  + 1045.15483618760 *T)
    sumB3  +=       0.945 * cos(0.45261136388  + 632.78373931320 *T)
    sumB3  +=       0.758 * cos(0.30577920142  + 949.17560896980 *T)
    sumB3  +=       0.731 * cos(2.63748223583  + 14.22709400160 *T)
    sumB3  +=       0.876 * cos(0.32927768725  + 1589.07289528380 *T)
    sumB3  +=       0.678 * cos(2.36909615348  + 1581.95934828300 *T)
    sumB3  +=       0.623 * cos(2.48056213600  + 1596.18644228460 *T)
    sumB3  +=       0.736 * cos(1.52532370632  + 735.87651353180 *T)
    sumB3  +=       0.499 * cos(3.67985494258  + 419.48464387520 *T)
    sumB3  +=       0.454 * cos(0.26977404624  + 942.06206196900 *T)
    sumB3  +=       0.453 * cos(3.18232334886  + 526.50957135690 *T)
    sumB3  +=       0.409 * cos(2.88147337106  + 110.20632121940 *T)
    sumB3  +=       0.347 * cos(5.76244285870  + 103.09277421860 *T)
    sumB3  +=       0.310 * cos(2.98017326384  + 508.35032409220 *T)
    sumB3  +=       0.321 * cos(4.40642025933  + 532.87235883230 *T)
    sumB3  +=       0.300 * cos(1.66936571536  + 625.67019231240 *T)
    sumB3  +=       0.295 * cos(1.75924202728  + 1073.60902419080 *T)
    sumB3  +=       0.282 * cos(3.11087801399  + 533.62311835770 *T)
    sumB3  +=       0.263 * cos(0.55255030187  + 426.59819087600 *T)
    sumB3  +=       0.208 * cos(2.17540496886  + 1155.36115740700 *T)
    sumB3  +=       0.183 * cos(4.34670868038  + 525.75881183150 *T)
    sumB3  +=       0.180 * cos(6.07777744541  + 639.89728631400 *T)
    sumB3  +=       0.159 * cos(2.60843864402  + 1162.47470440780 *T)
    sumB3  +=       0.117 * cos(4.70141431381  + 95.97922721780 *T)
    sumB3  +=       0.107 * cos(5.48942805114  + 433.71173787680 *T)
    sumB3  +=       0.105 * cos(3.75192101775  + 316.39186965660 *T)
    sumB3  +=       0.130 * cos(1.37897716939  + 323.50541665740 *T)
    sumB3  +=       0.094 * cos(3.05797832024  + 1265.56747862640 *T)
    sumB3  +=       0.114 * cos(3.75170981478  + 117.31986822020 *T)
    sumB3  +=       0.095 * cos(0.54905691533  + 1169.58825140860 *T)
    sumB3  +=       0.088 * cos(3.26874502411  + 213.29909543800 *T)
    sumB3  +=       0.098 * cos(2.00704668688  + 1574.84580128220 *T)

    # JUPITER Latitude
    sumB4  = 0
    sumB4  +=      15.050 * cos(4.52956999637  + 522.57741809380 *T)
    sumB4  +=       5.370 * cos(4.47427159142  + 529.69096509460 *T)
    sumB4  +=       4.456 * cos(5.43908581047  + 536.80451209540 *T)
    sumB4  +=       3.422 * cos(0.00000000000  + 0.00000000000 *T)
    sumB4  +=       1.833 * cos(4.51807036227  + 515.46387109300 *T)
    sumB4  +=       1.322 * cos(4.20117611581  + 1052.26838318840 *T)
    sumB4  +=       0.755 * cos(5.59451554966  + 543.91805909620 *T)
    sumB4  +=       0.512 * cos(0.05803177475  + 1066.49547719000 *T)
    sumB4  +=       0.282 * cos(3.66807771223  + 1059.38193018920 *T)
    sumB4  +=       0.147 * cos(3.56490986181  + 1045.15483618760 *T)
    sumB4  +=       0.142 * cos(5.69936472988  + 7.11354700080 *T)
    sumB4  +=       0.112 * cos(1.16718383135  + 14.22709400160 *T)

    # JUPITER Latitude
    sumB5  = 0
    sumB5  +=       1.445 * cos(0.09198554072  + 522.57741809380 *T)
    sumB5  +=       0.368 * cos(0.00874408003  + 515.46387109300 *T)
    sumB5  +=       0.304 * cos(3.27902945138  + 536.80451209540 *T)
    sumB5  +=       0.129 * cos(0.33959775247  + 529.69096509460 *T)
    sumB5  +=       0.095 * cos(1.29305954542  + 543.91805909620 *T)

    # JUPITER Radius
    sumR0  = 0
    sumR0  +=  520887429.471 * cos(0.00000000000  + 0.00000000000 *T)
    sumR0  +=  25209327.020 * cos(3.49108640015  + 529.69096509460 *T)
    sumR0  +=  610599.902 * cos(3.84115365602  + 1059.38193018920 *T)
    sumR0  +=  282029.465 * cos(2.57419879933  + 632.78373931320 *T)
    sumR0  +=  187647.391 * cos(2.07590380082  + 522.57741809380 *T)
    sumR0  +=   86792.941 * cos(0.71001090609  + 419.48464387520 *T)
    sumR0  +=   72062.869 * cos(0.21465694745  + 536.80451209540 *T)
    sumR0  +=   65517.227 * cos(5.97995850843  + 316.39186965660 *T)
    sumR0  +=   29134.620 * cos(1.67759243710  + 103.09277421860 *T)
    sumR0  +=   30135.275 * cos(2.16132058449  + 949.17560896980 *T)
    sumR0  +=   23453.209 * cos(3.54023147303  + 735.87651353180 *T)
    sumR0  +=   22283.710 * cos(4.19362773546  + 1589.07289528380 *T)
    sumR0  +=   23947.340 * cos(0.27457854894  + 7.11354700080 *T)
    sumR0  +=   13032.600 * cos(2.96043055741  + 1162.47470440780 *T)
    sumR0  +=    9703.346 * cos(1.90669572402  + 206.18554843720 *T)
    sumR0  +=   12749.004 * cos(2.71550102862  + 1052.26838318840 *T)
    sumR0  +=    9161.431 * cos(4.41352618935  + 213.29909543800 *T)
    sumR0  +=    7894.539 * cos(2.47907551404  + 426.59819087600 *T)
    sumR0  +=    7057.978 * cos(2.18184753111  + 1265.56747862640 *T)
    sumR0  +=    6137.755 * cos(6.26417542514  + 846.08283475120 *T)
    sumR0  +=    5477.093 * cos(5.65729325169  + 639.89728631400 *T)
    sumR0  +=    3502.519 * cos(0.56531297394  + 1066.49547719000 *T)
    sumR0  +=    4136.890 * cos(2.72219979684  + 625.67019231240 *T)
    sumR0  +=    4170.012 * cos(2.01605033912  + 515.46387109300 *T)
    sumR0  +=    2499.966 * cos(4.55182055941  + 838.96928775040 *T)
    sumR0  +=    2616.955 * cos(2.00993967129  + 1581.95934828300 *T)
    sumR0  +=    1911.876 * cos(0.85621927419  + 412.37109687440 *T)
    sumR0  +=    2127.644 * cos(6.12751461750  + 742.99006053260 *T)
    sumR0  +=    1610.549 * cos(3.08867789275  + 1368.66025284500 *T)
    sumR0  +=    1479.484 * cos(2.68026191372  + 1478.86657406440 *T)
    sumR0  +=    1230.708 * cos(1.89042979701  + 323.50541665740 *T)
    sumR0  +=    1216.810 * cos(1.80171561024  + 110.20632121940 *T)
    sumR0  +=     961.072 * cos(4.54876989805  + 2118.76386037840 *T)
    sumR0  +=     885.708 * cos(4.14785948471  + 533.62311835770 *T)
    sumR0  +=     776.700 * cos(3.67696954690  + 728.76296653100 *T)
    sumR0  +=     998.579 * cos(2.87208940110  + 309.27832265580 *T)
    sumR0  +=    1014.959 * cos(1.38673237666  + 454.90936652730 *T)
    sumR0  +=     727.162 * cos(3.98824686402  + 1155.36115740700 *T)
    sumR0  +=     655.289 * cos(2.79065604219  + 1685.05212250160 *T)
    sumR0  +=     821.465 * cos(1.59342534396  + 1898.35121793960 *T)
    sumR0  +=     620.798 * cos(4.82284338962  + 956.28915597060 *T)
    sumR0  +=     653.981 * cos(3.38150775269  + 1692.16566950240 *T)
    sumR0  +=     812.036 * cos(5.94091899141  + 909.81873305460 *T)
    sumR0  +=     562.120 * cos(0.08095987241  + 543.91805909620 *T)
    sumR0  +=     542.221 * cos(0.28360266386  + 525.75881183150 *T)
    sumR0  +=     457.859 * cos(0.12722694510  + 1375.77379984580 *T)
    sumR0  +=     614.784 * cos(2.27624915604  + 942.06206196900 *T)
    sumR0  +=     435.805 * cos(2.60272129748  + 95.97922721780 *T)
    sumR0  +=     496.066 * cos(5.53005947761  + 380.12776796000 *T)
    sumR0  +=     469.965 * cos(2.81896276101  + 1795.25844372100 *T)
    sumR0  +=     445.003 * cos(0.14623567024  + 14.22709400160 *T)
    sumR0  +=     290.869 * cos(3.89339143564  + 1471.75302706360 *T)
    sumR0  +=     276.627 * cos(2.52238450687  + 2001.44399215820 *T)
    sumR0  +=     275.084 * cos(2.98863518924  + 526.50957135690 *T)
    sumR0  +=     293.875 * cos(2.04938438861  + 199.07200143640 *T)
    sumR0  +=     290.985 * cos(6.03131226226  + 1169.58825140860 *T)
    sumR0  +=     338.342 * cos(2.79873192583  + 1045.15483618760 *T)
    sumR0  +=     257.482 * cos(6.13395478303  + 532.87235883230 *T)
    sumR0  +=     319.013 * cos(1.34803130803  + 2214.74308759620 *T)
    sumR0  +=     309.352 * cos(5.36855804945  + 1272.68102562720 *T)
    sumR0  +=     345.804 * cos(1.56404293688  + 491.55792945680 *T)
    sumR0  +=     303.364 * cos(1.15407454372  + 5753.38488489680 *T)
    sumR0  +=     192.325 * cos(0.91996333387  + 1596.18644228460 *T)
    sumR0  +=     215.398 * cos(2.63572815848  + 2111.65031337760 *T)
    sumR0  +=     200.738 * cos(2.37259566683  + 1258.45393162560 *T)
    sumR0  +=     239.036 * cos(3.57397189838  + 835.03713448730 *T)
    sumR0  +=     197.073 * cos(5.92859096863  + 453.42489381900 *T)
    sumR0  +=     139.440 * cos(3.63960322318  + 1788.14489672020 *T)
    sumR0  +=     191.373 * cos(6.28251311870  + 983.11585891360 *T)
    sumR0  +=     176.551 * cos(2.57669991654  + 9683.59458111640 *T)
    sumR0  +=     123.567 * cos(2.26158186345  + 2317.83586181480 *T)
    sumR0  +=     128.176 * cos(4.66585907670  + 831.85574074960 *T)
    sumR0  +=     112.430 * cos(0.85604150812  + 433.71173787680 *T)
    sumR0  +=     128.817 * cos(1.10567106595  + 2531.13495725280 *T)
    sumR0  +=      99.390 * cos(4.50312054049  + 518.64526483070 *T)
    sumR0  +=      93.870 * cos(2.72553879990  + 853.19638175200 *T)
    sumR0  +=     106.481 * cos(5.81462222290  + 220.41264243880 *T)
    sumR0  +=     120.188 * cos(2.95156363556  + 3.93215326310 *T)
    sumR0  +=     104.002 * cos(2.22221906187  + 74.78159856730 *T)
    sumR0  +=      81.655 * cos(3.23481337678  + 1361.54670584420 *T)
    sumR0  +=     112.513 * cos(4.86216964016  + 528.20649238630 *T)
    sumR0  +=      79.539 * cos(0.88542246830  + 430.53034413910 *T)
    sumR0  +=      85.801 * cos(2.11458386763  + 1574.84580128220 *T)
    sumR0  +=      85.685 * cos(2.33823884827  + 2428.04218303420 *T)
    sumR0  +=      68.311 * cos(3.35727048905  + 2104.53676637680 *T)
    sumR0  +=      69.570 * cos(3.04164697156  + 302.16477565500 *T)
    sumR0  +=      69.775 * cos(3.22402404312  + 305.34616939270 *T)
    sumR0  +=      69.570 * cos(0.20494979941  + 532.13864564940 *T)
    sumR0  +=      56.991 * cos(2.00204191909  + 2634.22773147140 *T)
    sumR0  +=      77.062 * cos(2.09816000231  + 508.35032409220 *T)
    sumR0  +=      56.716 * cos(3.91743976711  + 2221.85663459700 *T)
    sumR0  +=      58.325 * cos(5.72360355252  + 628.85158605010 *T)
    sumR0  +=      52.485 * cos(4.02485010492  + 527.24328453980 *T)
    sumR0  +=      63.645 * cos(1.09973563964  + 1364.72809958190 *T)
    sumR0  +=      53.607 * cos(0.87425992614  + 2847.52682690940 *T)
    sumR0  +=      59.598 * cos(0.95822471775  + 494.26624244250 *T)
    sumR0  +=      57.960 * cos(3.45779497978  + 2008.55753915900 *T)
    sumR0  +=      41.512 * cos(3.51955526735  + 529.73914920440 *T)
    sumR0  +=      44.666 * cos(1.62313786651  + 984.60033162190 *T)
    sumR0  +=      44.883 * cos(4.90091959557  + 2648.45482547300 *T)
    sumR0  +=      53.206 * cos(1.19800364308  + 760.25553592000 *T)
    sumR0  +=      44.393 * cos(4.42623747662  + 1063.31408345230 *T)
    sumR0  +=      37.566 * cos(2.93021095213  + 1677.93857550080 *T)
    sumR0  +=      41.516 * cos(0.32174409278  + 529.64278098480 *T)
    sumR0  +=      42.855 * cos(0.03093594081  + 1439.50969814920 *T)
    sumR0  +=      45.963 * cos(2.54342106514  + 636.71589257630 *T)
    sumR0  +=      40.181 * cos(4.39381642864  + 1148.24761040620 *T)
    sumR0  +=      38.770 * cos(4.31675565025  + 149.56319713460 *T)
    sumR0  +=      40.348 * cos(2.10140891053  + 2744.43405269080 *T)
    sumR0  +=      48.851 * cos(5.60297777544  + 2810.92146160520 *T)
    sumR0  +=      37.085 * cos(5.07828164301  + 1905.46476494040 *T)
    sumR0  +=      43.875 * cos(1.24536971083  + 621.73803904930 *T)
    sumR0  +=      34.005 * cos(3.09360167248  + 2420.92863603340 *T)
    sumR0  +=      36.782 * cos(0.84232174637  + 530.65417294110 *T)
    sumR0  +=      31.139 * cos(5.35811251334  + 1485.98012106520 *T)
    sumR0  +=      39.295 * cos(4.70800489067  + 569.04784100980 *T)
    sumR0  +=      39.700 * cos(2.46163878814  + 355.74874557180 *T)
    sumR0  +=      31.527 * cos(6.19284070863  + 3.18139373770 *T)
    sumR0  +=      28.399 * cos(2.48456666067  + 519.39602435610 *T)
    sumR0  +=      32.432 * cos(2.73281750275  + 604.47256366190 *T)
    sumR0  +=      27.119 * cos(3.92341697086  + 2324.94940881560 *T)
    sumR0  +=      26.753 * cos(1.74975198417  + 2950.61960112800 *T)
    sumR0  +=      28.986 * cos(1.83535862643  + 1891.23767093880 *T)
    sumR0  +=      26.493 * cos(0.60380196895  + 1055.44977692610 *T)
    sumR0  +=      33.525 * cos(0.76068430639  + 643.82943957710 *T)
    sumR0  +=      26.568 * cos(1.03594610835  + 405.25754987360 *T)
    sumR0  +=      25.534 * cos(3.46320665375  + 458.84151979040 *T)
    sumR0  +=      24.421 * cos(0.88181836930  + 423.41679713830 *T)
    sumR0  +=      32.949 * cos(3.18597137308  + 528.72775724810 *T)
    sumR0  +=      22.456 * cos(0.43129919683  + 1073.60902419080 *T)
    sumR0  +=      21.599 * cos(1.41820425091  + 540.73666535850 *T)
    sumR0  +=      25.673 * cos(0.52358194760  + 511.53171782990 *T)
    sumR0  +=      21.115 * cos(3.08023522766  + 629.60234557550 *T)
    sumR0  +=      22.713 * cos(0.65234613144  + 3163.91869656600 *T)
    sumR0  +=      19.189 * cos(5.16589014963  + 635.96513305090 *T)
    sumR0  +=      26.042 * cos(1.33629471285  + 330.61896365820 *T)
    sumR0  +=      18.263 * cos(3.59973446951  + 746.92221379570 *T)
    sumR0  +=      18.210 * cos(2.66819439927  + 1994.33044515740 *T)
    sumR0  +=      19.724 * cos(4.13552133321  + 1464.63948006280 *T)
    sumR0  +=      19.480 * cos(1.85656428109  + 3060.82592234740 *T)
    sumR0  +=      23.927 * cos(4.99826361784  + 1289.94650101460 *T)
    sumR0  +=      21.886 * cos(5.91718683551  + 1802.37199072180 *T)
    sumR0  +=      17.482 * cos(2.82161612542  + 2737.32050569000 *T)
    sumR0  +=      16.608 * cos(5.67394889755  + 408.43894361130 *T)
    sumR0  +=      22.892 * cos(5.26731352093  + 672.14061522840 *T)
    sumR0  +=      18.349 * cos(1.89869734949  + 1021.24889455140 *T)
    sumR0  +=      19.123 * cos(3.65882402977  + 415.55249061210 *T)
    sumR0  +=      15.735 * cos(3.34772676006  + 1056.20053645150 *T)
    sumR0  +=      16.373 * cos(0.18094878053  + 1699.27921650320 *T)
    sumR0  +=      18.899 * cos(3.69120638874  + 88.86568021700 *T)
    sumR0  +=      18.655 * cos(1.97327300097  + 38.13303563780 *T)
    sumR0  +=      15.542 * cos(3.82204881010  + 721.64941953020 *T)
    sumR0  +=      16.780 * cos(1.90976657921  + 217.23124870110 *T)
    sumR0  +=      15.313 * cos(1.05907174619  + 114.13847448250 *T)
    sumR0  +=      15.190 * cos(1.32317039042  + 117.31986822020 *T)
    sumR0  +=      15.080 * cos(3.74469077216  + 2641.34127847220 *T)
    sumR0  +=      19.836 * cos(2.73184571324  + 39.35687591520 *T)
    sumR0  +=      14.708 * cos(1.67270454473  + 529.16970023280 *T)
    sumR0  +=      14.036 * cos(3.54305270022  + 142.44965013380 *T)
    sumR0  +=      12.931 * cos(1.48829749349  + 3267.01147078460 *T)
    sumR0  +=      14.924 * cos(1.32546085940  + 490.33408917940 *T)
    sumR0  +=      14.753 * cos(4.64530618027  + 6283.07584999140 *T)
    sumR0  +=      14.672 * cos(0.80451954754  + 5223.69391980220 *T)
    sumR0  +=      12.085 * cos(3.67072510553  + 750.10360753340 *T)
    sumR0  +=      11.954 * cos(2.97127390765  + 505.31194270640 *T)
    sumR0  +=      14.650 * cos(2.16792930250  + 530.21222995640 *T)
    sumR0  +=      11.869 * cos(1.66551754962  + 2207.62954059540 *T)
    sumR0  +=      12.273 * cos(0.20690014405  + 1062.56332392690 *T)
    sumR0  +=      11.460 * cos(1.11906683214  + 561.93429400900 *T)
    sumR0  +=      11.083 * cos(3.22049096074  + 535.10759106600 *T)
    sumR0  +=      11.567 * cos(5.22625628971  + 524.06189080210 *T)
    sumR0  +=      11.161 * cos(3.82945634036  + 76.26607127560 *T)
    sumR0  +=      10.918 * cos(1.27796962818  + 2125.87740737920 *T)
    sumR0  +=      12.685 * cos(3.96848605476  + 2538.24850425360 *T)
    sumR0  +=      11.230 * cos(3.23092119889  + 422.66603761290 *T)
    sumR0  +=      12.645 * cos(0.73670428580  + 908.33426034630 *T)
    sumR0  +=      11.330 * cos(5.56127247007  + 531.17543780290 *T)
    sumR0  +=       9.509 * cos(5.00507284204  + 597.35901666110 *T)
    sumR0  +=      10.291 * cos(3.84159025239  + 1781.03134971940 *T)
    sumR0  +=      10.762 * cos(4.91380719453  + 525.02509864860 *T)
    sumR0  +=      11.786 * cos(5.11863653538  + 685.47393735270 *T)
    sumR0  +=      11.980 * cos(1.72470898635  + 911.30320576290 *T)
    sumR0  +=       8.937 * cos(2.40338241992  + 2310.72231481400 *T)
    sumR0  +=       9.253 * cos(2.57670338148  + 3053.71237534660 *T)
    sumR0  +=       9.488 * cos(2.95089828501  + 1382.88734684660 *T)
    sumR0  +=       9.889 * cos(0.43758517388  + 3480.31056622260 *T)
    sumR0  +=       8.781 * cos(3.66562388594  + 739.80866679490 *T)
    sumR0  +=       8.664 * cos(2.70398612383  + 526.77020378780 *T)
    sumR0  +=       9.505 * cos(1.61249870019  + 3377.21779200400 *T)
    sumR0  +=      11.540 * cos(1.59520481029  + 1474.67378837040 *T)
    sumR0  +=       9.533 * cos(0.35468711552  + 1512.80682400820 *T)
    sumR0  +=       9.980 * cos(4.80984684596  + 558.00214074590 *T)
    sumR0  +=       9.014 * cos(1.21458362718  + 416.30325013750 *T)
    sumR0  +=       7.969 * cos(0.08480602718  + 528.94020556920 *T)
    sumR0  +=       8.668 * cos(5.29060005706  + 945.24345570670 *T)
    sumR0  +=       7.851 * cos(1.46751861875  + 963.40270297140 *T)
    sumR0  +=       8.611 * cos(1.13232641062  + 532.61172640140 *T)
    sumR0  +=       7.838 * cos(6.26933498027  + 647.01083331480 *T)
    sumR0  +=       7.581 * cos(2.90608705954  + 533.88375078860 *T)
    sumR0  +=       8.583 * cos(6.06634530166  + 10213.28554621100 *T)
    sumR0  +=      10.198 * cos(2.48743123636  + 1819.63746610920 *T)
    sumR0  +=       8.536 * cos(2.22700701790  + 9153.90361602180 *T)
    sumR0  +=       9.759 * cos(6.15593336218  + 593.42686339800 *T)
    sumR0  +=       7.968 * cos(3.75535355212  + 530.44172462000 *T)
    sumR0  +=       7.142 * cos(3.58836120327  + 2957.73314812880 *T)
    sumR0  +=       7.122 * cos(0.11970048938  + 224.34479570190 *T)
    sumR0  +=       8.731 * cos(0.75302913970  + 960.22130923370 *T)
    sumR0  +=       7.063 * cos(2.16793037690  + 724.83081326790 *T)
    sumR0  +=       7.263 * cos(2.29499675875  + 520.12973753900 *T)
    sumR0  +=       6.418 * cos(1.25058991868  + 3583.40334044120 *T)
    sumR0  +=       8.270 * cos(1.24806288317  + 495.75071515080 *T)
    sumR0  +=       6.483 * cos(4.74567772640  + 202.25339517410 *T)
    sumR0  +=       7.197 * cos(3.84169279666  + 618.55664531160 *T)
    sumR0  +=       8.146 * cos(0.73147060302  + 230.56457082540 *T)
    sumR0  +=       6.165 * cos(5.50124418381  + 11.04570026390 *T)
    sumR0  +=       7.946 * cos(2.07754951174  + 953.10776223290 *T)
    sumR0  +=       7.675 * cos(0.92400307662  + 525.49817940060 *T)
    sumR0  +=       6.210 * cos(1.45641362115  + 483.22054217860 *T)
    sumR0  +=       7.359 * cos(0.31355650764  + 378.64329525170 *T)
    sumR0  +=       6.707 * cos(2.92071167098  + 1038.04128918680 *T)
    sumR0  +=       7.143 * cos(0.18218134889  + 731.94436026870 *T)
    sumR0  +=       7.309 * cos(6.27084533477  + 21.34064100240 *T)
    sumR0  +=       6.135 * cos(2.67651237303  + 312.45971639350 *T)
    sumR0  +=       5.558 * cos(3.83419160288  + 534.35683154060 *T)
    sumR0  +=       5.344 * cos(5.25294750019  + 1048.33622992530 *T)
    sumR0  +=       7.504 * cos(0.74281415471  + 457.61767951300 *T)
    sumR0  +=       5.335 * cos(6.23059924424  + 551.03160609700 *T)
    sumR0  +=       5.613 * cos(1.51210605952  + 524.27433912320 *T)
    sumR0  +=       5.284 * cos(2.18579185671  + 280.96714700450 *T)
    sumR0  +=       5.475 * cos(5.95864753605  + 539.98590583310 *T)
    sumR0  +=       5.056 * cos(0.37387972537  + 529.53090640020 *T)
    sumR0  +=       6.202 * cos(5.53813122743  + 2.44768055480 *T)
    sumR0  +=       5.490 * cos(5.97692444199  + 227.52618943960 *T)
    sumR0  +=       6.266 * cos(0.76632858238  + 938.12990870590 *T)
    sumR0  +=       5.750 * cos(2.13496323512  + 191.95845443560 *T)
    sumR0  +=       5.218 * cos(4.69335266854  + 560.71045373160 *T)
    sumR0  +=       5.480 * cos(5.21157595558  + 1057.89745748090 *T)
    sumR0  +=       5.738 * cos(0.34249718209  + 535.91074021810 *T)
    sumR0  +=       4.816 * cos(1.51326236835  + 2524.02141025200 *T)
    sumR0  +=       5.056 * cos(3.46671669992  + 529.85102378900 *T)
    sumR0  +=       4.710 * cos(2.27813830550  + 3370.10424500320 *T)
    sumR0  +=       5.228 * cos(3.61776977584  + 2097.42321937600 *T)
    sumR0  +=       4.878 * cos(1.39829798223  + 3693.60966166060 *T)
    sumR0  +=       5.727 * cos(4.80120381106  + 598.84348936940 *T)
    sumR0  +=       5.707 * cos(3.94177950323  + 2854.64037391020 *T)
    sumR0  +=       4.988 * cos(4.87244187719  + 1.48447270830 *T)
    sumR0  +=       5.424 * cos(3.53268613904  + 456.39383923560 *T)
    sumR0  +=       4.288 * cos(4.84438067847  + 70.84944530420 *T)
    sumR0  +=       5.944 * cos(3.79180483544  + 25558.21217647960 *T)
    sumR0  +=       4.195 * cos(2.09136830994  + 2627.11418447060 *T)
    sumR0  +=       4.582 * cos(5.61707254513  + 2435.15573003500 *T)
    sumR0  +=       4.268 * cos(6.20250525415  + 775.23338944700 *T)
    sumR0  +=       4.521 * cos(0.20049967962  + 92.04707395470 *T)
    sumR0  +=       5.405 * cos(4.66492781581  + 833.55266177900 *T)
    sumR0  +=       5.607 * cos(3.30226645638  + 535.32003938710 *T)
    sumR0  +=       4.171 * cos(3.14873010832  + 944.98282327580 *T)
    sumR0  +=       4.108 * cos(5.84489743779  + 440.82528487760 *T)
    sumR0  +=       4.367 * cos(4.68363584557  + 327.43756992050 *T)
    sumR0  +=       4.033 * cos(3.30883782817  + 3274.12501778540 *T)
    sumR0  +=       4.292 * cos(0.20604269202  + 3796.70243587920 *T)
    sumR0  +=       4.270 * cos(0.98941708997  + 387.24131496080 *T)
    sumR0  +=       4.259 * cos(3.21120589971  + 696.51963761660 *T)
    sumR0  +=       4.673 * cos(1.96606729969  + 107.02492748170 *T)
    sumR0  +=       4.031 * cos(4.62854606236  + 2751.54759969160 *T)
    sumR0  +=       5.115 * cos(2.66416451377  + 1215.16490244730 *T)
    sumR0  +=       4.181 * cos(4.74527698816  + 988.53248488500 *T)
    sumR0  +=       4.374 * cos(1.50010561403  + 1894.41906467650 *T)
    sumR0  +=       3.803 * cos(3.59911687954  + 437.64389113990 *T)
    sumR0  +=       3.761 * cos(3.96903199782  + 732.69511979410 *T)
    sumR0  +=       3.620 * cos(1.57847427805  + 381.61224066830 *T)
    sumR0  +=       3.490 * cos(0.63097592112  + 529.90341341570 *T)
    sumR0  +=       4.019 * cos(2.57664165720  + 916.93228005540 *T)
    sumR0  +=       4.133 * cos(4.78417930217  + 824.74219374880 *T)
    sumR0  +=       4.411 * cos(3.13179382423  + 630.33605875840 *T)
    sumR0  +=       4.099 * cos(3.63702212253  + 810.65811209910 *T)
    sumR0  +=       3.704 * cos(6.17243801274  + 537.76771994190 *T)
    sumR0  +=       4.124 * cos(2.14248285449  + 210.11770170030 *T)
    sumR0  +=       3.490 * cos(3.20962050417  + 529.47851677350 *T)
    sumR0  +=       3.281 * cos(1.53106243317  + 547.85021235930 *T)
    sumR0  +=       3.554 * cos(6.03787799174  + 739.05790726950 *T)
    sumR0  +=       4.101 * cos(6.00406226999  + 902.70518605380 *T)
    sumR0  +=       3.267 * cos(3.49354065789  + 1166.40685767090 *T)
    sumR0  +=       3.286 * cos(2.55966870530  + 945.99421523210 *T)
    sumR0  +=       4.041 * cos(4.78735413707  + 850.01498801430 *T)
    sumR0  +=       4.304 * cos(0.11406117717  + 1744.85586754190 *T)
    sumR0  +=       4.043 * cos(5.20417093600  + 635.23141986800 *T)
    sumR0  +=       3.115 * cos(4.61986265585  + 952.35700270750 *T)
    sumR0  +=       3.016 * cos(0.95126220905  + 3899.79521009780 *T)
    sumR0  +=       3.017 * cos(2.59699501992  + 632.83192342300 *T)
    sumR0  +=       3.219 * cos(1.83594791142  + 18.15924726470 *T)
    sumR0  +=       3.203 * cos(6.12597544496  + 10.29494073850 *T)
    sumR0  +=       3.220 * cos(6.14213423140  + 1158.54255114470 *T)
    sumR0  +=       3.000 * cos(5.69509924353  + 632.73555520340 *T)
    sumR0  +=       3.226 * cos(5.59910267099  + 608.40471692500 *T)
    sumR0  +=       3.118 * cos(5.64998934505  + 99.16062095550 *T)
    sumR0  +=       3.745 * cos(2.08111521615  + 282.45161971280 *T)
    sumR0  +=       2.837 * cos(4.60175594220  + 245.54242435240 *T)
    sumR0  +=       3.093 * cos(6.02049413961  + 633.74694715970 *T)
    sumR0  +=       3.120 * cos(2.29047945342  + 631.82053146670 *T)
    sumR0  +=       2.662 * cos(3.69016679729  + 885.43971066640 *T)
    sumR0  +=       3.150 * cos(1.79784999553  + 521.61421024730 *T)
    sumR0  +=       2.822 * cos(3.14927418161  + 295.05122865420 *T)
    sumR0  +=       2.615 * cos(0.20732170653  + 35.42472265210 *T)
    sumR0  +=       2.971 * cos(1.28795094653  + 1023.95720753710 *T)
    sumR0  +=       2.571 * cos(2.01817133502  + 1514.29129671650 *T)
    sumR0  +=       2.592 * cos(0.48790221200  + 195.13984817330 *T)
    sumR0  +=       3.263 * cos(2.38820607343  + 836.52160719560 *T)
    sumR0  +=       2.501 * cos(0.21653750027  + 465.95506679120 *T)
    sumR0  +=       2.451 * cos(5.58559489768  + 544.66881862160 *T)
    sumR0  +=       2.535 * cos(1.44414086617  + 460.53844081980 *T)
    sumR0  +=       2.666 * cos(3.30350145485  + 2413.81508903260 *T)
    sumR0  +=       2.412 * cos(4.36756580310  + 1056.93424963440 *T)
    sumR0  +=       2.452 * cos(4.53818816565  + 514.71311156760 *T)
    sumR0  +=       3.239 * cos(1.17022488774  + 177.87437278590 *T)
    sumR0  +=       3.218 * cos(0.60551913257  + 1061.82961074400 *T)
    sumR0  +=       2.408 * cos(0.65423523810  + 523.54062594030 *T)
    sumR0  +=       2.299 * cos(2.15247752560  + 319.57326339430 *T)
    sumR0  +=       2.791 * cos(2.71505085086  + 610.69233878540 *T)
    sumR0  +=       2.729 * cos(1.77685979153  + 252.65597135320 *T)
    sumR0  +=       2.666 * cos(3.77750458842  + 3171.03224356680 *T)
    sumR0  +=       2.303 * cos(0.36676453766  + 1969.20066324380 *T)
    sumR0  +=       2.664 * cos(0.09674841214  + 565.11568774670 *T)
    sumR0  +=       2.312 * cos(2.07210502831  + 3686.49611465980 *T)
    sumR0  +=       2.680 * cos(4.94445888050  + 1593.00504854690 *T)
    sumR0  +=       2.193 * cos(0.55645982205  + 2228.97018159780 *T)
    sumR0  +=       2.526 * cos(1.07528597373  + 12036.46073488820 *T)
    sumR0  +=       2.778 * cos(1.48379350517  + 447.79581952650 *T)
    sumR0  +=       2.235 * cos(5.95475282699  + 6151.53388830500 *T)
    sumR0  +=       2.759 * cos(4.63976153480  + 462.02291352810 *T)
    sumR0  +=       2.175 * cos(4.53588570240  + 501.37978944330 *T)
    sumR0  +=       2.323 * cos(5.93670041006  + 611.44309831080 *T)
    sumR0  +=       2.384 * cos(2.81746622971  + 3340.61242669980 *T)
    sumR0  +=       2.087 * cos(3.10716079675  + 1049.08698945070 *T)
    sumR0  +=       1.994 * cos(2.02500860064  + 1058.86066532740 *T)
    sumR0  +=       2.199 * cos(2.20937490997  + 1269.49963188950 *T)
    sumR0  +=       2.705 * cos(1.97665276677  + 415.29185818120 *T)
    sumR0  +=       2.787 * cos(1.31053438756  + 1041.22268292450 *T)
    sumR0  +=       2.003 * cos(4.66904374443  + 679.25416222920 *T)
    sumR0  +=       1.962 * cos(1.82999730674  + 2943.50605412720 *T)
    sumR0  +=       2.289 * cos(2.96480800939  + 69.15252427480 *T)
    sumR0  +=       2.192 * cos(4.47837196209  + 209.36694217490 *T)
    sumR0  +=       2.020 * cos(0.04621364490  + 4113.09430553580 *T)
    sumR0  +=       2.082 * cos(1.11203059170  + 4010.00153131720 *T)
    sumR0  +=       1.991 * cos(3.20108648275  + 3590.51688744200 *T)
    sumR0  +=       1.900 * cos(3.32227077969  + 421.93232443000 *T)
    sumR0  +=       2.193 * cos(2.82218305362  + 292.01284726840 *T)
    sumR0  +=       2.288 * cos(1.94695631885  + 1279.79457262800 *T)
    sumR0  +=       1.843 * cos(5.23293634337  + 14.97785352700 *T)
    sumR0  +=       1.932 * cos(5.46684252030  + 2281.23049651060 *T)
    sumR0  +=       2.177 * cos(2.93031976617  + 429.04587143080 *T)
    sumR0  +=       2.125 * cos(0.06224847826  + 24.37902238820 *T)
    sumR0  +=       2.464 * cos(5.39581078430  + 1261.63532536330 *T)
    sumR0  +=       1.938 * cos(3.79908004671  + 1059.43011429900 *T)
    sumR0  +=       2.029 * cos(3.95461157815  + 771.30123618390 *T)
    sumR0  +=       1.841 * cos(4.74905354737  + 78.71375183040 *T)
    sumR0  +=       1.922 * cos(2.21862085389  + 99.91138048090 *T)
    sumR0  +=       1.836 * cos(5.75449805175  + 623.22251175760 *T)
    sumR0  +=       2.145 * cos(3.87052575546  + 451.94042111070 *T)
    sumR0  +=       1.782 * cos(0.40860352236  + 754.03576079650 *T)
    sumR0  +=       1.784 * cos(1.49468287576  + 529.95159752550 *T)
    sumR0  +=       1.842 * cos(3.49726261337  + 1354.43315884340 *T)
    sumR0  +=       1.748 * cos(3.48730020953  + 522.62560220360 *T)
    sumR0  +=       1.816 * cos(1.24334711210  + 417.03696332040 *T)
    sumR0  +=       1.752 * cos(1.15500390019  + 1060.34513803570 *T)
    sumR0  +=       1.729 * cos(2.69831073799  + 642.34496686880 *T)
    sumR0  +=       1.985 * cos(1.99916658759  + 934.94851496820 *T)
    sumR0  +=       1.828 * cos(5.44095029767  + 1201.83158032300 *T)
    sumR0  +=       2.158 * cos(3.45672748590  + 827.92358748650 *T)
    sumR0  +=       1.959 * cos(1.06033047373  + 33.94024994380 *T)
    sumR0  +=       1.751 * cos(3.13572498964  + 384.05992122310 *T)
    sumR0  +=       1.781 * cos(5.02895146997  + 1098.73880610440 *T)
    sumR0  +=       2.074 * cos(3.18582065441  + 1366.21257229020 *T)
    sumR0  +=       1.757 * cos(5.02778552877  + 586.31331639720 *T)
    sumR0  +=       2.045 * cos(3.08816627459  + 535.84130424890 *T)
    sumR0  +=       2.273 * cos(5.17998505813  + 3178.14579056760 *T)
    sumR0  +=       1.617 * cos(3.16674916201  + 67.66805156650 *T)
    sumR0  +=       1.627 * cos(6.10603469594  + 432.01481684740 *T)
    sumR0  +=       1.930 * cos(1.63968957659  + 5.41662597140 *T)
    sumR0  +=       1.741 * cos(0.99408274736  + 1254.52177836250 *T)
    sumR0  +=       1.607 * cos(5.65498642076  + 1165.65609814550 *T)
    sumR0  +=       1.676 * cos(3.06138410273  + 1134.16352875650 *T)
    sumR0  +=       1.821 * cos(3.05183555090  + 567.82400073240 *T)
    sumR0  +=       1.677 * cos(3.09175084930  + 1251.34038462480 *T)
    sumR0  +=       1.994 * cos(2.52023134712  + 1059.90319505100 *T)
    sumR0  +=       2.204 * cos(6.15376698510  + 563.63121503840 *T)
    sumR0  +=       1.692 * cos(4.19142612803  + 106.27416795630 *T)
    sumR0  +=       1.906 * cos(5.58417395051  + 32.24332891440 *T)
    sumR0  +=       2.206 * cos(1.75883974012  + 1151.42900414390 *T)
    sumR0  +=       1.552 * cos(3.04262360186  + 385.54439393140 *T)
    sumR0  +=       1.508 * cos(0.42002830727  + 313.21047591890 *T)
    sumR0  +=       1.494 * cos(1.43672345922  + 2840.41327990860 *T)
    sumR0  +=       1.678 * cos(2.17255433434  + 306.83064210100 *T)
    sumR0  +=       1.511 * cos(4.44377608685  + 395.10562148700 *T)
    sumR0  +=       1.958 * cos(0.05215107058  + 761.74000862830 *T)
    sumR0  +=       1.760 * cos(1.27045286501  + 1173.52040467170 *T)
    sumR0  +=       1.463 * cos(6.07810373103  + 0.96320784650 *T)
    sumR0  +=       1.498 * cos(2.79408561759  + 277.03499374140 *T)
    sumR0  +=       1.636 * cos(0.26199351490  + 522.52923398400 *T)
    sumR0  +=       1.507 * cos(0.48961801593  + 4216.18707975440 *T)
    sumR0  +=       1.530 * cos(3.42953827550  + 1159.29331067010 *T)
    sumR0  +=       1.744 * cos(2.39637837261  + 203.00415469950 *T)
    sumR0  +=       1.569 * cos(2.55719070621  + 4.19278569400 *T)
    sumR0  +=       1.576 * cos(3.45039607104  + 1058.41872234270 *T)
    sumR0  +=       1.466 * cos(2.24427539934  + 1550.93985964600 *T)
    sumR0  +=       1.784 * cos(2.34591354953  + 529.43033266370 *T)
    sumR0  +=       1.939 * cos(4.73685428610  + 3067.93946934820 *T)
    sumR0  +=       1.938 * cos(0.60126164334  + 1059.33374607940 *T)
    sumR0  +=       1.523 * cos(2.98744673443  + 2730.20695868920 *T)
    sumR0  +=       1.834 * cos(3.78099298791  + 420.96911658350 *T)
    sumR0  +=       1.372 * cos(3.53997115825  + 5.62907429250 *T)
    sumR0  +=       1.361 * cos(0.45533257707  + 418.52143602870 *T)
    sumR0  +=       1.833 * cos(5.12743628215  + 1578.02719501990 *T)
    sumR0  +=       1.839 * cos(4.24616044210  + 981.63138620530 *T)
    sumR0  +=       1.567 * cos(3.32429870195  + 532.39927808030 *T)
    sumR0  +=       1.340 * cos(1.94668282270  + 528.41894070740 *T)
    sumR0  +=       1.422 * cos(1.83191577465  + 4002.88798431640 *T)
    sumR0  +=       1.745 * cos(5.76913240451  + 490.07345674850 *T)
    sumR0  +=       1.437 * cos(4.19470227783  + 420.44785172170 *T)
    sumR0  +=       1.419 * cos(0.74849005330  + 632.26247445140 *T)
    sumR0  +=       1.447 * cos(5.65611888743  + 373.01422095920 *T)
    sumR0  +=       1.578 * cos(3.90273683089  + 602.98809095360 *T)
    sumR0  +=       1.385 * cos(3.88479835656  + 419.43645976540 *T)
    sumR0  +=       1.352 * cos(0.81697905853  + 1585.14074202070 *T)
    sumR0  +=       1.399 * cos(1.24785452243  + 633.30500417500 *T)
    sumR0  +=       1.297 * cos(5.57914023189  + 1276.61317889030 *T)
    sumR0  +=       1.491 * cos(1.66541781223  + 2655.56837247380 *T)
    sumR0  +=       1.252 * cos(0.72155670765  + 173.94221952280 *T)
    sumR0  +=       1.658 * cos(5.60924662850  + 362.86229257260 *T)
    sumR0  +=       1.606 * cos(3.95301396173  + 2274.54683263650 *T)
    sumR0  +=       1.213 * cos(4.55264289565  + 366.79444583570 *T)
    sumR0  +=       1.521 * cos(0.55773831071  + 1592.25428902150 *T)
    sumR0  +=       1.220 * cos(3.63029788040  + 497.44763618020 *T)
    sumR0  +=       1.215 * cos(4.42854185903  + 531.38788612400 *T)
    sumR0  +=       1.549 * cos(5.73765962068  + 320.32402291970 *T)
    sumR0  +=       1.480 * cos(4.29779032931  + 303.86169668440 *T)
    sumR0  +=       1.507 * cos(2.27998567874  + 758.77106321170 *T)
    sumR0  +=       1.212 * cos(3.38335836048  + 536.85269620520 *T)
    sumR0  +=       1.245 * cos(4.21639959154  + 4.66586644600 *T)
    sumR0  +=       1.507 * cos(3.52136655355  + 774.00954916960 *T)
    sumR0  +=       1.481 * cos(3.06156044618  + 1585.89150154610 *T)
    sumR0  +=       1.462 * cos(2.30628702634  + 1363.24362687360 *T)
    sumR0  +=       1.180 * cos(3.52708055024  + 1064.79855616060 *T)
    sumR0  +=       1.193 * cos(5.88284733845  + 1060.86640289750 *T)
    sumR0  +=       1.398 * cos(4.99456521692  + 842.90144101350 *T)
    sumR0  +=       1.406 * cos(1.53799746944  + 1020.02505427400 *T)
    sumR0  +=       1.367 * cos(4.10254739443  + 799.61241183520 *T)
    sumR0  +=       1.336 * cos(1.89387272380  + 530.96298948180 *T)
    sumR0  +=       1.238 * cos(3.62226383331  + 3487.42411322340 *T)
    sumR0  +=       1.306 * cos(3.39985119727  + 539.25219265020 *T)
    sumR0  +=       1.156 * cos(0.77127511567  + 1603.29998928540 *T)
    sumR0  +=       1.482 * cos(0.48451915093  + 493.04240216510 *T)
    sumR0  +=       1.247 * cos(5.64344659992  + 479.28838891550 *T)
    sumR0  +=       1.195 * cos(2.39909893341  + 561.18353448360 *T)
    sumR0  +=       1.106 * cos(0.89453807282  + 2.92076130680 *T)
    sumR0  +=       1.227 * cos(2.76231244946  + 299.12639426920 *T)
    sumR0  +=       1.128 * cos(4.72319873338  + 124.43341522100 *T)
    sumR0  +=       1.086 * cos(5.66180289525  + 1053.75285589670 *T)
    sumR0  +=       1.329 * cos(0.16664094530  + 536.75632798560 *T)
    sumR0  +=       1.082 * cos(4.51407359350  + 528.25467649610 *T)
    sumR0  +=       1.105 * cos(1.93890691771  + 244.31858407500 *T)
    sumR0  +=       1.446 * cos(0.65096230619  + 1091.62525910360 *T)
    sumR0  +=       1.071 * cos(4.67974963103  + 521.82665856840 *T)
    sumR0  +=       1.413 * cos(4.72936311016  + 1141.13406340540 *T)
    sumR0  +=       1.086 * cos(2.88721124443  + 1262.38608488870 *T)
    sumR0  +=       1.254 * cos(5.74156595137  + 527.99404406520 *T)
    sumR0  +=       1.082 * cos(5.60975006771  + 531.12725369310 *T)
    sumR0  +=       1.148 * cos(3.27410230525  + 1035.00290780100 *T)
    sumR0  +=       1.224 * cos(3.68807537150  + 81.75213321620 *T)
    sumR0  +=       1.072 * cos(0.48068438564  + 1058.63117066380 *T)
    sumR0  +=       1.036 * cos(1.68789163831  + 1070.42763045310 *T)
    sumR0  +=       1.052 * cos(4.72763208332  + 913.75088631770 *T)
    sumR0  +=       1.166 * cos(4.97812626679  + 450.97721326420 *T)
    sumR0  +=       1.042 * cos(2.90894542321  + 3906.90875709860 *T)
    sumR0  +=       0.997 * cos(1.65967703856  + 3259.89792378380 *T)
    sumR0  +=       1.113 * cos(3.06502453809  + 1482.79872732750 *T)
    sumR0  +=       0.991 * cos(0.91568114148  + 576.16138801060 *T)
    sumR0  +=       0.987 * cos(0.91349590742  + 2332.06295581640 *T)
    sumR0  +=       1.003 * cos(6.17381204883  + 391.17346822390 *T)
    sumR0  +=       1.087 * cos(3.19260020877  + 151.04766984290 *T)
    sumR0  +=       0.987 * cos(2.48065918834  + 1912.57831194120 *T)
    sumR0  +=       0.975 * cos(1.55458771092  + 536.28324723360 *T)
    sumR0  +=       1.193 * cos(2.19383228000  + 523.09868295560 *T)
    sumR0  +=       0.979 * cos(3.28693620660  + 1379.70595310890 *T)
    sumR0  +=       0.963 * cos(2.29845109892  + 1467.82087380050 *T)
    sumR0  +=       1.279 * cos(4.73978455573  + 600.54041039880 *T)
    sumR0  +=       1.269 * cos(1.77171706595  + 5120.60114558360 *T)
    sumR0  +=       0.938 * cos(3.13636271584  + 1372.59240610810 *T)
    sumR0  +=       0.956 * cos(0.94045126791  + 429.77958461370 *T)
    sumR0  +=       1.130 * cos(4.87259620358  + 874.39401040250 *T)
    sumR0  +=       1.044 * cos(3.52819283674  + 530.58473697190 *T)
    sumR0  +=       1.244 * cos(0.80634178279  + 419.53282798500 *T)
    sumR0  +=       0.914 * cos(4.34324212455  + 1127.04998175570 *T)
    sumR0  +=       1.095 * cos(3.17513475763  + 6681.22485339960 *T)
    sumR0  +=       0.926 * cos(5.53099018797  + 537.55527162080 *T)
    sumR0  +=       1.025 * cos(6.08315999637  + 469.88722005430 *T)
    sumR0  +=       0.928 * cos(2.64064849636  + 31.01948863700 *T)
    sumR0  +=       0.887 * cos(5.53922649066  + 498.67147645760 *T)
    sumR0  +=       1.153 * cos(5.20213407651  + 554.06998748280 *T)
    sumR0  +=       0.976 * cos(4.26047885490  + 806.72595883600 *T)
    sumR0  +=       0.871 * cos(5.79751110150  + 594.65070367540 *T)
    sumR0  +=       1.044 * cos(0.31244551729  + 528.79719321730 *T)
    sumR0  +=       0.911 * cos(0.94039205468  + 337.73251065900 *T)
    sumR0  +=       1.197 * cos(3.12884590029  + 1966.23171782720 *T)
    sumR0  +=       0.930 * cos(2.88178471518  + 1056.46116888240 *T)
    sumR0  +=       1.052 * cos(1.69484089706  + 484.44438245600 *T)
    sumR0  +=       0.862 * cos(0.67309397482  + 20426.57109242200 *T)
    sumR0  +=       1.152 * cos(1.16751621652  + 1489.91227432830 *T)
    sumR0  +=       0.847 * cos(3.25831322825  + 1063.57471588320 *T)
    sumR0  +=       0.884 * cos(0.71487680084  + 2042.49778910280 *T)
    sumR0  +=       0.888 * cos(5.38714907441  + 5621.84292321040 *T)
    sumR0  +=       1.137 * cos(4.02029739425  + 1670.07426897460 *T)
    sumR0  +=       0.844 * cos(3.31846798590  + 812.14258480740 *T)
    sumR0  +=       0.860 * cos(4.78175008217  + 530.91480537200 *T)
    sumR0  +=       0.835 * cos(3.63117401608  + 451.72797278960 *T)
    sumR0  +=       0.931 * cos(2.27352189963  + 100.64509366380 *T)
    sumR0  +=       0.939 * cos(3.51238251326  + 523.47118997110 *T)
    sumR0  +=       0.860 * cos(5.34207357904  + 528.46712481720 *T)
    sumR0  +=       0.875 * cos(0.87775537110  + 4326.39340097380 *T)
    sumR0  +=       0.961 * cos(5.69327275886  + 498.19839570560 *T)
    sumR0  +=       0.966 * cos(6.25512226434  + 700.45179087970 *T)
    sumR0  +=       0.842 * cos(3.20535945596  + 1670.82502850000 *T)
    sumR0  +=       0.808 * cos(1.09148925587  + 683.18631549230 *T)
    sumR0  +=       0.810 * cos(5.47935192896  + 525.54636351040 *T)
    sumR0  +=       0.855 * cos(6.06969867736  + 446.31134681820 *T)
    sumR0  +=       0.989 * cos(1.55623875216  + 1493.09366806600 *T)
    sumR0  +=       0.837 * cos(1.49510080792  + 1025.44168024540 *T)
    sumR0  +=       0.974 * cos(3.67667471757  + 25565.32572348040 *T)
    sumR0  +=       0.788 * cos(0.51622458293  + 526.98265210890 *T)
    sumR0  +=       0.820 * cos(1.86002542644  + 629.86297800640 *T)
    sumR0  +=       0.813 * cos(0.45441968195  + 4694.00295470760 *T)
    sumR0  +=       0.953 * cos(0.58786779132  + 627.36711334180 *T)
    sumR0  +=       0.908 * cos(2.82093327912  + 3046.59882834580 *T)
    sumR0  +=       0.912 * cos(2.69124310451  + 946.72792841500 *T)
    sumR0  +=       0.820 * cos(4.14947931572  + 1884.12412393800 *T)
    sumR0  +=       0.948 * cos(0.77931728039  + 25551.09862947879 *T)
    sumR0  +=       0.844 * cos(0.00976249584  + 628.59095361920 *T)
    sumR0  +=       0.910 * cos(0.99542530366  + 5760.49843189760 *T)
    sumR0  +=       0.844 * cos(0.22630964490  + 1123.11782849260 *T)
    sumR0  +=       0.924 * cos(4.41952345708  + 5746.27133789600 *T)
    sumR0  +=       0.967 * cos(3.20618313117  + 9050.81084180320 *T)
    sumR0  +=       0.800 * cos(0.10663079153  + 4532.57894941100 *T)
    sumR0  +=       0.748 * cos(3.01376405927  + 5481.75455838080 *T)
    sumR0  +=       0.752 * cos(5.82360472890  + 701.93626358800 *T)
    sumR0  +=       0.771 * cos(0.12101982692  + 635.70450062000 *T)
    sumR0  +=       0.725 * cos(2.81220410314  + 3597.63043444280 *T)
    sumR0  +=       0.944 * cos(0.40327408174  + 1140.38330388000 *T)
    sumR0  +=       0.726 * cos(5.28930472464  + 1304.92435454160 *T)
    sumR0  +=       0.994 * cos(5.16391370100  + 10316.37832042960 *T)
    sumR0  +=       0.890 * cos(4.10819809692  + 1060.13268971460 *T)
    sumR0  +=       0.962 * cos(1.48376004549  + 1062.30269149600 *T)
    sumR0  +=       0.883 * cos(5.26813169286  + 1542.60247236780 *T)
    sumR0  +=       0.916 * cos(6.02908368648  + 7.86430652620 *T)
    sumR0  +=       0.725 * cos(2.18773773010  + 1176.70179840940 *T)
    sumR0  +=       0.808 * cos(5.81725174908  + 1087.69310584050 *T)
    sumR0  +=       0.757 * cos(0.77440414330  + 977.48678462110 *T)
    sumR0  +=       0.838 * cos(3.81585420192  + 986.08480433020 *T)
    sumR0  +=       0.888 * cos(1.89634795578  + 707.56533788050 *T)
    sumR0  +=       0.854 * cos(5.47701506544  + 2818.03500860600 *T)
    sumR0  +=       0.796 * cos(1.08794807212  + 987.30864460760 *T)
    sumR0  +=       0.856 * cos(2.58042139486  + 2803.80791460440 *T)
    sumR0  +=       0.708 * cos(1.09492310353  + 248.72381809010 *T)
    sumR0  +=       0.811 * cos(3.23726191865  + 121.25202148330 *T)
    sumR0  +=       0.727 * cos(1.56150632966  + 4319.27985397300 *T)
    sumR0  +=       0.687 * cos(2.65457835371  + 1567.73225428140 *T)
    sumR0  +=       0.675 * cos(1.78690909614  + 103.14095832840 *T)
    sumR0  +=       0.853 * cos(4.74476428852  + 951.62328952460 *T)
    sumR0  +=       0.832 * cos(5.14362789810  + 1054.71606374320 *T)
    sumR0  +=       0.846 * cos(1.47557828604  + 898.77303279070 *T)
    sumR0  +=       0.701 * cos(1.72139817505  + 5230.80746680300 *T)
    sumR0  +=       0.863 * cos(3.98700238575  + 686.95841006100 *T)
    sumR0  +=       0.703 * cos(2.89202252444  + 63.73589830340 *T)
    sumR0  +=       0.673 * cos(6.11618580510  + 738.32419408660 *T)
    sumR0  +=       0.806 * cos(4.64475158248  + 533.83556667880 *T)
    sumR0  +=       0.670 * cos(2.67625974048  + 1012.91150727320 *T)
    sumR0  +=       0.668 * cos(4.93815253692  + 5172.47623572500 *T)
    sumR0  +=       0.818 * cos(1.41973280302  + 580.09354127370 *T)
    sumR0  +=       0.652 * cos(3.41422919445  + 650.94298657790 *T)
    sumR0  +=       0.643 * cos(2.46566726278  + 1049.82070263360 *T)
    sumR0  +=       0.859 * cos(2.50530106631  + 782.34693644780 *T)
    sumR0  +=       0.662 * cos(4.13533996643  + 733.42883297700 *T)
    sumR0  +=       0.812 * cos(1.30325352179  + 1055.18914449520 *T)
    sumR0  +=       0.638 * cos(4.21760246824  + 1064.04779663520 *T)
    sumR0  +=       0.637 * cos(6.13121700151  + 4752.99159184980 *T)
    sumR0  +=       0.636 * cos(0.83411828974  + 711.49749114360 *T)
    sumR0  +=       0.642 * cos(1.86741704507  + 1053.96530421780 *T)
    sumR0  +=       0.795 * cos(4.54081089118  + 1457.52593306200 *T)
    sumR0  +=       0.783 * cos(4.37652961667  + 105.54045477340 *T)
    sumR0  +=       0.640 * cos(5.44039474349  + 632.03297978780 *T)
    sumR0  +=       0.651 * cos(5.02431301146  + 528.04643369190 *T)
    sumR0  +=       0.686 * cos(0.27079898498  + 11.77941344680 *T)
    sumR0  +=       0.644 * cos(5.36935176134  + 835.78789401270 *T)
    sumR0  +=       0.639 * cos(1.86699974431  + 6172.86952877200 *T)
    sumR0  +=       0.630 * cos(2.86895754523  + 633.53449883860 *T)
    sumR0  +=       0.826 * cos(1.46026926041  + 2199.76523406920 *T)
    sumR0  +=       0.687 * cos(3.81221717134  + 73.29712585900 *T)
    sumR0  +=       0.697 * cos(4.18082589322  + 1.69692102940 *T)
    sumR0  +=       0.788 * cos(0.21278801649  + 313.94418910180 *T)
    sumR0  +=       0.686 * cos(2.51807576494  + 638.41281360570 *T)
    sumR0  +=       0.847 * cos(5.56263749391  + 4429.48617519240 *T)
    sumR0  +=       0.673 * cos(4.87494072856  + 103.04459010880 *T)
    sumR0  +=       0.663 * cos(4.80713895807  + 991.71387862270 *T)
    sumR0  +=       0.614 * cos(3.87231597482  + 767.36908292080 *T)
    sumR0  +=       0.666 * cos(5.71697262323  + 661.09491496450 *T)
    sumR0  +=       0.681 * cos(2.33844767741  + 501.23677709140 *T)
    sumR0  +=       0.597 * cos(3.03921014345  + 6.95348830640 *T)
    sumR0  +=       0.777 * cos(3.08786050361  + 441.57604440300 *T)
    sumR0  +=       0.588 * cos(0.08236113246  + 4164.31198961300 *T)
    sumR0  +=       0.693 * cos(4.66190836234  + 3384.33133900480 *T)
    sumR0  +=       0.810 * cos(1.97701084490  + 860.30992875280 *T)
    sumR0  +=       0.602 * cos(5.56403449542  + 1587.58842257550 *T)
    sumR0  +=       0.622 * cos(6.11554348965  + 7.06536289100 *T)
    sumR0  +=       0.592 * cos(3.29013906024  + 10103.07922499160 *T)
    sumR0  +=       0.692 * cos(6.10931942233  + 12.74262129330 *T)
    sumR0  +=       0.597 * cos(6.13204711801  + 7.27360569520 *T)
    sumR0  +=       0.594 * cos(2.58839673551  + 849.26422848890 *T)
    sumR0  +=       0.728 * cos(2.73732195088  + 6.15033915430 *T)
    sumR0  +=       0.602 * cos(5.28816527514  + 949.12742486000 *T)
    sumR0  +=       0.568 * cos(1.75508433865  + 1077.54117745390 *T)
    sumR0  +=       0.575 * cos(4.50676079721  + 1230.14275597430 *T)
    sumR0  +=       0.588 * cos(0.65827893998  + 4642.78527063040 *T)
    sumR0  +=       0.561 * cos(3.87565914360  + 135.33610313300 *T)
    sumR0  +=       0.558 * cos(3.36094471852  + 24498.83024629040 *T)
    sumR0  +=       0.557 * cos(3.45629457197  + 19896.88012732740 *T)
    sumR0  +=       0.558 * cos(1.17103892689  + 3576.28979344040 *T)
    sumR0  +=       0.574 * cos(5.19235074140  + 104.05598206510 *T)
    sumR0  +=       0.560 * cos(3.57141429379  + 5333.90024102160 *T)
    sumR0  +=       0.555 * cos(0.18349908409  + 512.42548970720 *T)
    sumR0  +=       0.571 * cos(0.83070148820  + 1570.91364801910 *T)
    sumR0  +=       0.632 * cos(3.67893818442  + 1065.01100448170 *T)
    sumR0  +=       0.744 * cos(2.33083237537  + 620.25356634100 *T)
    sumR0  +=       0.540 * cos(5.15775909675  + 1751.53953141600 *T)
    sumR0  +=       0.592 * cos(3.07238123875  + 1446.62324515000 *T)
    sumR0  +=       0.537 * cos(1.52803865425  + 8094.52168583260 *T)
    sumR0  +=       0.550 * cos(5.50701003577  + 1432.39615114840 *T)
    sumR0  +=       0.546 * cos(2.34388967045  + 949.22379307960 *T)
    sumR0  +=       0.534 * cos(3.04076654796  + 7.16173111060 *T)
    sumR0  +=       0.619 * cos(6.07865159203  + 46.47042291600 *T)
    sumR0  +=       0.562 * cos(0.96641974928  + 1438.02522544090 *T)
    sumR0  +=       0.531 * cos(1.06695547390  + 100.17201291180 *T)
    sumR0  +=       0.599 * cos(3.59295739143  + 1144.31545714310 *T)
    sumR0  +=       0.526 * cos(3.51641923371  + 0.75075952540 *T)
    sumR0  +=       0.564 * cos(0.72677136494  + 1059.22187149480 *T)
    sumR0  +=       0.537 * cos(5.72603965787  + 513.22863885930 *T)
    sumR0  +=       0.630 * cos(2.31183143900  + 2729.45619916380 *T)
    sumR0  +=       0.530 * cos(4.99510636441  + 9264.10993724120 *T)
    sumR0  +=       0.649 * cos(0.95666735852  + 920.86443331850 *T)
    sumR0  +=       0.547 * cos(1.18801926149  + 11506.76976979360 *T)
    sumR0  +=       0.516 * cos(3.28562070858  + 734.91330568530 *T)
    sumR0  +=       0.567 * cos(5.13926871155  + 288.08069400530 *T)
    sumR0  +=       0.538 * cos(0.28159637680  + 153.49535039770 *T)
    sumR0  +=       0.718 * cos(0.48326672359  + 842.15068148810 *T)
    sumR0  +=       0.526 * cos(4.39778401928  + 546.15329132990 *T)
    sumR0  +=       0.695 * cos(2.44235086902  + 657.16276170140 *T)
    sumR0  +=       0.697 * cos(4.99042365686  + 12.53017297220 *T)
    sumR0  +=       0.519 * cos(6.27847163164  + 59.80374504030 *T)
    sumR0  +=       0.504 * cos(2.58550284000  + 5378.66178416220 *T)
    sumR0  +=       0.496 * cos(2.43659402827  + 990.22940591440 *T)
    sumR0  +=       0.617 * cos(5.73284985700  + 745.43774108740 *T)
    sumR0  +=       0.519 * cos(3.10157097770  + 9161.01716302260 *T)
    sumR0  +=       0.654 * cos(1.31181453784  + 878.32616366560 *T)
    sumR0  +=       0.619 * cos(3.71554817226  + 2090.30967237520 *T)
    sumR0  +=       0.500 * cos(4.28937439066  + 5216.58037280140 *T)
    sumR0  +=       0.621 * cos(3.98893673383  + 409.92341631960 *T)
    sumR0  +=       0.685 * cos(1.95310431695  + 3156.80514956520 *T)
    sumR0  +=       0.552 * cos(2.81774132958  + 344.70304530790 *T)
    sumR0  +=       0.551 * cos(1.91969778405  + 113.38771495710 *T)
    sumR0  +=       0.682 * cos(0.87321578326  + 6069.77675455340 *T)
    sumR0  +=       0.651 * cos(5.09951064975  + 531.33549649730 *T)
    sumR0  +=       0.537 * cos(3.67357440226  + 605.95703637020 *T)
    sumR0  +=       0.525 * cos(0.74584814988  + 736.83972137830 *T)
    sumR0  +=       0.505 * cos(3.12494814307  + 1475.68518032670 *T)
    sumR0  +=       0.622 * cos(3.00013939606  + 2349.32843120380 *T)
    sumR0  +=       0.644 * cos(3.00156986335  + 298.23262239190 *T)
    sumR0  +=       0.564 * cos(3.81960833949  + 1059.54198888360 *T)
    sumR0  +=       0.468 * cos(3.50348554992  + 4841.85727206680 *T)
    sumR0  +=       0.491 * cos(1.28535573072  + 247.23934538180 *T)
    sumR0  +=       0.458 * cos(0.45056377876  + 1065.60170531270 *T)
    sumR0  +=       0.543 * cos(2.39704308320  + 9690.70812811720 *T)
    sumR0  +=       0.459 * cos(5.29870259698  + 1474.93442080130 *T)
    sumR0  +=       0.483 * cos(3.63649121244  + 131.40394986990 *T)
    sumR0  +=       0.632 * cos(2.75028345792  + 334.55111692130 *T)
    sumR0  +=       0.483 * cos(0.42979609421  + 735.82832942200 *T)
    sumR0  +=       0.540 * cos(0.54791737146  + 51646.11531805379 *T)
    sumR0  +=       0.531 * cos(0.30026207053  + 912.78767847120 *T)
    sumR0  +=       0.449 * cos(3.02583472996  + 5901.23920225600 *T)
    sumR0  +=       0.544 * cos(2.98747240952  + 4223.30062675520 *T)
    sumR0  +=       0.557 * cos(5.83542572008  + 9676.48103411560 *T)
    sumR0  +=       0.501 * cos(0.03408180117  + 1080.72257119160 *T)
    sumR0  +=       0.517 * cos(4.40400852026  + 2545.36205125440 *T)
    sumR0  +=       0.481 * cos(3.63292807076  + 5584.84733259940 *T)
    sumR0  +=       0.557 * cos(6.11443978190  + 976.00231191280 *T)
    sumR0  +=       0.481 * cos(3.41035583659  + 3803.81598288000 *T)
    sumR0  +=       0.622 * cos(2.29597570837  + 9999.98645077300 *T)
    sumR0  +=       0.454 * cos(2.88584538455  + 1987.21689815660 *T)
    sumR0  +=       0.439 * cos(4.83198101064  + 50.40257617910 *T)
    sumR0  +=       0.475 * cos(2.69994471394  + 491.81856188770 *T)
    sumR0  +=       0.618 * cos(0.72471290082  + 1291.43097372290 *T)
    sumR0  +=       0.503 * cos(0.13449993622  + 2015.67108615980 *T)
    sumR0  +=       0.551 * cos(2.13418546604  + 1440.99417085750 *T)
    sumR0  +=       0.595 * cos(3.78181802545  + 6386.16862421000 *T)
    sumR0  +=       0.434 * cos(2.64411689486  + 748.40668650400 *T)
    sumR0  +=       0.592 * cos(0.32587740408  + 737.36098624010 *T)
    sumR0  +=       0.490 * cos(2.37988828800  + 2225.78878786010 *T)
    sumR0  +=       0.439 * cos(1.33582802018  + 995.64603188580 *T)
    sumR0  +=       0.543 * cos(2.05067702505  + 906.84978763800 *T)
    sumR0  +=       0.466 * cos(2.43707405011  + 3362.99069800240 *T)
    sumR0  +=       0.481 * cos(2.32223226419  + 1357.61455258110 *T)
    sumR0  +=       0.566 * cos(0.59740900184  + 350.33211960040 *T)
    sumR0  +=       0.429 * cos(2.46287580628  + 3914.02230409940 *T)
    sumR0  +=       0.429 * cos(1.01299906509  + 4333.50694797460 *T)
    sumR0  +=       0.425 * cos(1.67255823369  + 148.07872442630 *T)
    sumR0  +=       0.412 * cos(3.29630633921  + 7.32599532190 *T)
    sumR0  +=       0.508 * cos(1.16158524676  + 9.56122755560 *T)
    sumR0  +=       0.524 * cos(5.02562926120  + 1090.40141882620 *T)
    sumR0  +=       0.409 * cos(5.80053072411  + 9146.79006902100 *T)
    sumR0  +=       0.497 * cos(0.01579913593  + 1069.67687092770 *T)
    sumR0  +=       0.548 * cos(6.03429743373  + 9367.20271145980 *T)
    sumR0  +=       0.433 * cos(5.93688350840  + 1688.23351623930 *T)
    sumR0  +=       0.424 * cos(4.18150111530  + 550.13783421970 *T)
    sumR0  +=       0.401 * cos(0.11519846139  + 970.51624997220 *T)
    sumR0  +=       0.503 * cos(5.28212300854  + 668.20846196530 *T)
    sumR0  +=       0.555 * cos(1.00328633255  + 141.22580985640 *T)
    sumR0  +=       0.404 * cos(2.48633976473  + 519.65665678700 *T)
    sumR0  +=       0.441 * cos(6.06185501734  + 25.12978191360 *T)
    sumR0  +=       0.412 * cos(5.87495245826  + 6.90109867970 *T)
    sumR0  +=       0.478 * cos(0.71264950607  + 1094.80665284130 *T)
    sumR0  +=       0.446 * cos(2.71248183031  + 31.49256938900 *T)
    sumR0  +=       0.404 * cos(5.49462012486  + 447.93883187840 *T)
    sumR0  +=       0.391 * cos(1.26105612700  + 8.07675484730 *T)
    sumR0  +=       0.463 * cos(1.93535321271  + 6275.96230299060 *T)
    sumR0  +=       0.507 * cos(3.61089992782  + 546.95644048200 *T)
    sumR0  +=       0.402 * cos(5.86200127054  + 927.83496796740 *T)
    sumR0  +=       0.481 * cos(6.21043578332  + 683.98946464440 *T)
    sumR0  +=       0.483 * cos(5.02142924458  + 857.12853501510 *T)
    sumR0  +=       0.444 * cos(0.84873092377  + 1371.84164658270 *T)
    sumR0  +=       0.391 * cos(2.81753436573  + 5798.14642803740 *T)
    sumR0  +=       0.395 * cos(0.22367886581  + 51116.42435295920 *T)
    sumR0  +=       0.378 * cos(6.03765733432  + 1268.74887236410 *T)
    sumR0  +=       0.471 * cos(6.24506463249  + 946.46729598410 *T)
    sumR0  +=       0.405 * cos(0.57785207581  + 107.28555991260 *T)
    sumR0  +=       0.371 * cos(6.15750793727  + 509.24409596950 *T)
    sumR0  +=       0.370 * cos(4.90330687618  + 1436.54075273260 *T)
    sumR0  +=       0.448 * cos(4.76565111029  + 284.14854074220 *T)
    sumR0  +=       0.474 * cos(0.71146352197  + 2108.46891963990 *T)
    sumR0  +=       0.509 * cos(5.53328407404  + 1128.53445446400 *T)

    # JUPITER Radius
    sumR1  = 0
    sumR1  +=  1271801.596 * cos(2.64937511122  + 529.69096509460 *T)
    sumR1  +=   61661.771 * cos(3.00076251018  + 1059.38193018920 *T)
    sumR1  +=   53443.592 * cos(3.89717644226  + 522.57741809380 *T)
    sumR1  +=   31185.167 * cos(4.88276663526  + 536.80451209540 *T)
    sumR1  +=   41390.257 * cos(0.00000000000  + 0.00000000000 *T)
    sumR1  +=   11847.190 * cos(2.41329588176  + 419.48464387520 *T)
    sumR1  +=    9166.360 * cos(4.75979408587  + 7.11354700080 *T)
    sumR1  +=    3175.763 * cos(2.79297987071  + 103.09277421860 *T)
    sumR1  +=    3203.446 * cos(5.21083285476  + 735.87651353180 *T)
    sumR1  +=    3403.605 * cos(3.34688537997  + 1589.07289528380 *T)
    sumR1  +=    2600.003 * cos(3.63435101622  + 206.18554843720 *T)
    sumR1  +=    2412.207 * cos(1.46947308304  + 426.59819087600 *T)
    sumR1  +=    2806.064 * cos(3.74223693580  + 515.46387109300 *T)
    sumR1  +=    2676.575 * cos(4.33052878699  + 1052.26838318840 *T)
    sumR1  +=    2100.507 * cos(3.92762682306  + 639.89728631400 *T)
    sumR1  +=    1646.182 * cos(5.30953510947  + 1066.49547719000 *T)
    sumR1  +=    1641.257 * cos(4.41628669824  + 625.67019231240 *T)
    sumR1  +=    1049.866 * cos(3.16113622955  + 213.29909543800 *T)
    sumR1  +=    1024.802 * cos(2.55432643018  + 412.37109687440 *T)
    sumR1  +=     740.996 * cos(2.17094630558  + 1162.47470440780 *T)
    sumR1  +=     806.404 * cos(2.67750801380  + 632.78373931320 *T)
    sumR1  +=     676.928 * cos(6.24953479790  + 838.96928775040 *T)
    sumR1  +=     468.895 * cos(4.70973463481  + 543.91805909620 *T)
    sumR1  +=     444.683 * cos(0.40281181402  + 323.50541665740 *T)
    sumR1  +=     567.076 * cos(4.57655414712  + 742.99006053260 *T)
    sumR1  +=     415.894 * cos(5.36836018215  + 728.76296653100 *T)
    sumR1  +=     484.689 * cos(2.46882793186  + 949.17560896980 *T)
    sumR1  +=     337.555 * cos(3.16781951120  + 956.28915597060 *T)
    sumR1  +=     401.738 * cos(4.60528841541  + 309.27832265580 *T)
    sumR1  +=     347.378 * cos(4.68148808722  + 14.22709400160 *T)
    sumR1  +=     260.753 * cos(5.34290306101  + 846.08283475120 *T)
    sumR1  +=     220.084 * cos(4.84210964963  + 1368.66025284500 *T)
    sumR1  +=     203.217 * cos(5.59995425432  + 1155.36115740700 *T)
    sumR1  +=     246.603 * cos(3.92313823537  + 942.06206196900 *T)
    sumR1  +=     183.504 * cos(4.26526769703  + 95.97922721780 *T)
    sumR1  +=     180.134 * cos(4.40165491159  + 532.87235883230 *T)
    sumR1  +=     197.134 * cos(3.70551461394  + 2118.76386037840 *T)
    sumR1  +=     196.005 * cos(3.75877587139  + 199.07200143640 *T)
    sumR1  +=     200.190 * cos(4.43888814441  + 1045.15483618760 *T)
    sumR1  +=     170.225 * cos(4.84647488867  + 526.50957135690 *T)
    sumR1  +=     146.335 * cos(6.12958365535  + 533.62311835770 *T)
    sumR1  +=     133.483 * cos(1.32245735855  + 110.20632121940 *T)
    sumR1  +=     132.076 * cos(4.51187950811  + 525.75881183150 *T)
    sumR1  +=     123.851 * cos(2.04290370696  + 1478.86657406440 *T)
    sumR1  +=     121.861 * cos(4.40581788491  + 1169.58825140860 *T)
    sumR1  +=     115.313 * cos(4.46741278152  + 1581.95934828300 *T)
    sumR1  +=      98.527 * cos(5.72833991647  + 1596.18644228460 *T)
    sumR1  +=      91.608 * cos(4.52965592121  + 1685.05212250160 *T)
    sumR1  +=     110.638 * cos(3.62504147403  + 1272.68102562720 *T)
    sumR1  +=      80.536 * cos(4.11311699583  + 1258.45393162560 *T)
    sumR1  +=      79.552 * cos(2.71898473954  + 1692.16566950240 *T)
    sumR1  +=     100.164 * cos(5.24693885858  + 1265.56747862640 *T)
    sumR1  +=      77.854 * cos(5.56722651753  + 1471.75302706360 *T)
    sumR1  +=      85.766 * cos(0.07906707372  + 831.85574074960 *T)
    sumR1  +=      82.132 * cos(3.80763015979  + 508.35032409220 *T)
    sumR1  +=      55.319 * cos(0.35180851191  + 316.39186965660 *T)
    sumR1  +=      52.338 * cos(5.53074272117  + 433.71173787680 *T)
    sumR1  +=      55.769 * cos(4.75141241141  + 302.16477565500 *T)
    sumR1  +=      50.597 * cos(4.85603161770  + 1375.77379984580 *T)
    sumR1  +=      43.554 * cos(4.94441642712  + 1361.54670584420 *T)
    sumR1  +=      42.172 * cos(1.22404278447  + 853.19638175200 *T)
    sumR1  +=      37.695 * cos(4.26767539209  + 2001.44399215820 *T)
    sumR1  +=      49.395 * cos(4.01422828967  + 220.41264243880 *T)
    sumR1  +=      38.263 * cos(5.33025236797  + 1788.14489672020 *T)
    sumR1  +=      35.611 * cos(1.76205571128  + 1795.25844372100 *T)
    sumR1  +=      36.296 * cos(3.84995284393  + 1574.84580128220 *T)
    sumR1  +=      29.332 * cos(5.16619257786  + 3.93215326310 *T)
    sumR1  +=      25.180 * cos(4.33777727362  + 519.39602435610 *T)
    sumR1  +=      24.778 * cos(2.72907897410  + 405.25754987360 *T)
    sumR1  +=      27.025 * cos(6.09669947903  + 1148.24761040620 *T)
    sumR1  +=      22.604 * cos(0.19173890105  + 380.12776796000 *T)
    sumR1  +=      20.499 * cos(4.32881495378  + 3.18139373770 *T)
    sumR1  +=      19.925 * cos(4.62967500111  + 1677.93857550080 *T)
    sumR1  +=      19.528 * cos(5.10596326232  + 1073.60902419080 *T)
    sumR1  +=      18.427 * cos(3.76522178300  + 1485.98012106520 *T)
    sumR1  +=      18.869 * cos(5.05259402407  + 2104.53676637680 *T)
    sumR1  +=      17.031 * cos(4.01843356903  + 2317.83586181480 *T)
    sumR1  +=      16.671 * cos(5.42931676507  + 88.86568021700 *T)
    sumR1  +=      15.337 * cos(2.92700926091  + 2008.55753915900 *T)
    sumR1  +=      14.499 * cos(3.63339836845  + 628.85158605010 *T)
    sumR1  +=      14.575 * cos(5.50832843322  + 721.64941953020 *T)
    sumR1  +=      13.728 * cos(4.87623389735  + 629.60234557550 *T)
    sumR1  +=      18.481 * cos(6.03032762264  + 330.61896365820 *T)
    sumR1  +=      13.499 * cos(1.38539534821  + 518.64526483070 *T)
    sumR1  +=      15.740 * cos(2.93038271684  + 1905.46476494040 *T)
    sumR1  +=      12.459 * cos(1.58587053146  + 2111.65031337760 *T)
    sumR1  +=      12.272 * cos(3.37671053917  + 635.96513305090 *T)
    sumR1  +=      11.836 * cos(4.08486322993  + 2648.45482547300 *T)
    sumR1  +=      11.166 * cos(4.62623267608  + 636.71589257630 *T)
    sumR1  +=      14.348 * cos(2.74177797727  + 2221.85663459700 *T)
    sumR1  +=      11.221 * cos(3.55311861205  + 1891.23767093880 *T)
    sumR1  +=      13.121 * cos(5.83845065644  + 1464.63948006280 *T)
    sumR1  +=      11.351 * cos(2.57606886230  + 511.53171782990 *T)
    sumR1  +=      10.487 * cos(0.49850799841  + 453.42489381900 *T)
    sumR1  +=       9.728 * cos(4.38837468002  + 1994.33044515740 *T)
    sumR1  +=      10.131 * cos(2.76432756215  + 423.41679713830 *T)
    sumR1  +=       8.620 * cos(5.16374493158  + 1056.20053645150 *T)
    sumR1  +=       8.952 * cos(4.79407952752  + 2420.92863603340 *T)
    sumR1  +=       8.126 * cos(3.72977106954  + 2634.22773147140 *T)
    sumR1  +=       8.078 * cos(1.29246272894  + 2428.04218303420 *T)
    sumR1  +=       8.867 * cos(1.85684753622  + 750.10360753340 *T)
    sumR1  +=       8.912 * cos(4.80973516711  + 1062.56332392690 *T)
    sumR1  +=       8.552 * cos(4.53818617984  + 21.34064100240 *T)
    sumR1  +=       9.468 * cos(4.33472161983  + 1802.37199072180 *T)
    sumR1  +=       6.904 * cos(5.96616555709  + 540.73666535850 *T)
    sumR1  +=       7.293 * cos(4.97763580465  + 1699.27921650320 *T)
    sumR1  +=       7.083 * cos(4.99096728816  + 1055.44977692610 *T)
    sumR1  +=       7.226 * cos(4.97823884383  + 1898.35121793960 *T)
    sumR1  +=       6.464 * cos(1.39173466879  + 422.66603761290 *T)
    sumR1  +=       6.214 * cos(4.46490158256  + 551.03160609700 *T)
    sumR1  +=       6.794 * cos(2.90878831415  + 2324.94940881560 *T)
    sumR1  +=       6.173 * cos(3.65617162985  + 621.73803904930 *T)
    sumR1  +=       6.243 * cos(6.13691919694  + 2125.87740737920 *T)
    sumR1  +=       5.936 * cos(2.58312235120  + 569.04784100980 *T)
    sumR1  +=       6.504 * cos(4.56908431757  + 1038.04128918680 *T)
    sumR1  +=       7.305 * cos(3.02062127734  + 416.30325013750 *T)
    sumR1  +=       6.598 * cos(5.55348005731  + 1781.03134971940 *T)
    sumR1  +=       5.133 * cos(6.21646917980  + 963.40270297140 *T)
    sumR1  +=       5.876 * cos(4.23153077453  + 539.98590583310 *T)
    sumR1  +=       5.119 * cos(0.06942832171  + 1063.31408345230 *T)
    sumR1  +=       5.460 * cos(4.91084384602  + 835.03713448730 *T)
    sumR1  +=       4.989 * cos(1.35153694680  + 1382.88734684660 *T)
    sumR1  +=       5.224 * cos(0.18468411116  + 117.31986822020 *T)
    sumR1  +=       6.187 * cos(3.87193497099  + 191.95845443560 *T)
    sumR1  +=       4.681 * cos(4.61057119508  + 643.82943957710 *T)
    sumR1  +=       4.627 * cos(3.34644534691  + 2207.62954059540 *T)
    sumR1  +=       4.526 * cos(4.07729737127  + 2310.72231481400 *T)
    sumR1  +=       4.718 * cos(4.55578336947  + 2737.32050569000 *T)
    sumR1  +=       4.471 * cos(1.47603161897  + 408.43894361130 *T)
    sumR1  +=       4.073 * cos(1.13014903180  + 415.55249061210 *T)
    sumR1  +=       5.476 * cos(5.63198569698  + 618.55664531160 *T)
    sumR1  +=       4.034 * cos(4.09631702747  + 430.53034413910 *T)
    sumR1  +=       4.304 * cos(4.60536378943  + 647.01083331480 *T)
    sumR1  +=       3.765 * cos(3.42751259825  + 2950.61960112800 *T)
    sumR1  +=       4.559 * cos(4.23723998745  + 227.52618943960 *T)
    sumR1  +=       3.695 * cos(1.03127824978  + 2744.43405269080 *T)
    sumR1  +=       3.667 * cos(4.12268925541  + 440.82528487760 *T)
    sumR1  +=       3.677 * cos(2.19480200527  + 534.35683154060 *T)
    sumR1  +=       3.818 * cos(1.14800596289  + 74.78159856730 *T)
    sumR1  +=       4.221 * cos(2.37721579949  + 2538.24850425360 *T)
    sumR1  +=       3.488 * cos(5.33792561596  + 458.84151979040 *T)
    sumR1  +=       3.437 * cos(4.26164443643  + 10.29494073850 *T)
    sumR1  +=       4.394 * cos(0.18808423412  + 824.74219374880 *T)
    sumR1  +=       3.339 * cos(4.85708402591  + 295.05122865420 *T)
    sumR1  +=       3.329 * cos(5.50043586719  + 739.80866679490 *T)
    sumR1  +=       3.623 * cos(4.64011531952  + 2214.74308759620 *T)
    sumR1  +=       3.185 * cos(2.69708590442  + 561.93429400900 *T)
    sumR1  +=       3.421 * cos(3.38512615384  + 149.56319713460 *T)
    sumR1  +=       3.442 * cos(4.34217280083  + 305.34616939270 *T)
    sumR1  +=       3.580 * cos(5.29481665335  + 2097.42321937600 *T)
    sumR1  +=       3.401 * cos(2.74761862893  + 2641.34127847220 *T)
    sumR1  +=       2.901 * cos(0.91012525424  + 984.60033162190 *T)
    sumR1  +=       3.566 * cos(1.63400343968  + 525.02509864860 *T)
    sumR1  +=       2.869 * cos(1.31799241974  + 611.44309831080 *T)
    sumR1  +=       2.635 * cos(5.25517910535  + 532.13864564940 *T)
    sumR1  +=       2.683 * cos(4.24641945773  + 3053.71237534660 *T)
    sumR1  +=       2.614 * cos(3.17862099921  + 527.24328453980 *T)
    sumR1  +=       2.251 * cos(4.21598247360  + 739.05790726950 *T)
    sumR1  +=       2.268 * cos(5.52248110560  + 524.27433912320 *T)
    sumR1  +=       2.372 * cos(4.19741177512  + 217.23124870110 *T)
    sumR1  +=       2.623 * cos(5.82647427958  + 732.69511979410 *T)
    sumR1  +=       2.666 * cos(3.92538056951  + 210.11770170030 *T)
    sumR1  +=       2.036 * cos(4.84043420813  + 1049.08698945070 *T)
    sumR1  +=       2.441 * cos(2.63840901843  + 760.25553592000 *T)
    sumR1  +=       2.095 * cos(5.76269812349  + 529.64278098480 *T)
    sumR1  +=       2.021 * cos(3.81308146017  + 2627.11418447060 *T)
    sumR1  +=       2.089 * cos(4.18463193132  + 945.99421523210 *T)
    sumR1  +=       2.305 * cos(1.61220665690  + 604.47256366190 *T)
    sumR1  +=       1.969 * cos(5.37427735384  + 142.44965013380 *T)
    sumR1  +=       1.923 * cos(4.75088270631  + 535.10759106600 *T)
    sumR1  +=       1.955 * cos(5.49000238006  + 1439.50969814920 *T)
    sumR1  +=       1.877 * cos(3.26978877187  + 3267.01147078460 *T)
    sumR1  +=       2.286 * cos(2.93885172004  + 76.26607127560 *T)
    sumR1  +=       2.074 * cos(5.85386852879  + 532.61172640140 *T)
    sumR1  +=       2.121 * cos(3.92430797099  + 2435.15573003500 *T)
    sumR1  +=       1.807 * cos(3.17208959472  + 2524.02141025200 *T)
    sumR1  +=       1.712 * cos(4.02986641257  + 731.94436026870 *T)
    sumR1  +=       2.119 * cos(0.41049593984  + 1279.79457262800 *T)
    sumR1  +=       1.660 * cos(2.34370903423  + 528.72775724810 *T)
    sumR1  +=       1.655 * cos(0.78809717175  + 3060.82592234740 *T)
    sumR1  +=       1.729 * cos(4.26127896267  + 724.83081326790 *T)
    sumR1  +=       2.060 * cos(5.04785330873  + 2413.81508903260 *T)
    sumR1  +=       2.095 * cos(2.67732367556  + 529.73914920440 *T)
    sumR1  +=       1.933 * cos(2.49162437046  + 2957.73314812880 *T)
    sumR1  +=       1.898 * cos(2.71948262975  + 952.35700270750 *T)
    sumR1  +=       1.634 * cos(2.98113068812  + 945.24345570670 *T)
    sumR1  +=       1.582 * cos(5.84373095005  + 547.85021235930 *T)
    sumR1  +=       1.662 * cos(0.27359627181  + 454.90936652730 *T)
    sumR1  +=       1.595 * cos(1.18530167095  + 38.13303563780 *T)
    sumR1  +=       1.550 * cos(0.64264572959  + 312.45971639350 *T)
    sumR1  +=       1.525 * cos(4.08789824989  + 1158.54255114470 *T)
    sumR1  +=       1.542 * cos(1.12520322326  + 1021.24889455140 *T)
    sumR1  +=       1.539 * cos(0.37324921979  + 319.57326339430 *T)
    sumR1  +=       1.628 * cos(5.24285773388  + 1354.43315884340 *T)
    sumR1  +=       1.897 * cos(3.79973291113  + 953.10776223290 *T)
    sumR1  +=       1.440 * cos(4.37872256685  + 3178.14579056760 *T)
    sumR1  +=       1.439 * cos(4.26513521887  + 526.77020378780 *T)
    sumR1  +=       1.557 * cos(5.43779802371  + 81.75213321620 *T)
    sumR1  +=       1.656 * cos(6.09667089740  + 530.65417294110 *T)
    sumR1  +=       1.548 * cos(3.48799710267  + 934.94851496820 *T)
    sumR1  +=       1.772 * cos(5.82549274759  + 909.81873305460 *T)
    sumR1  +=       1.615 * cos(1.45018725033  + 902.70518605380 *T)
    sumR1  +=       1.387 * cos(2.52840497309  + 530.44172462000 *T)
    sumR1  +=       1.574 * cos(1.89565809136  + 437.64389113990 *T)
    sumR1  +=       1.459 * cos(3.32546061506  + 1041.22268292450 *T)
    sumR1  +=       1.377 * cos(0.10015418633  + 490.33408917940 *T)
    sumR1  +=       1.460 * cos(4.00706825185  + 3370.10424500320 *T)
    sumR1  +=       1.605 * cos(4.27993020192  + 2531.13495725280 *T)
    sumR1  +=       1.707 * cos(6.28253681644  + 18.15924726470 *T)
    sumR1  +=       1.802 * cos(2.23019296374  + 2854.64037391020 *T)
    sumR1  +=       1.390 * cos(3.76737324192  + 1165.65609814550 *T)
    sumR1  +=       1.498 * cos(0.17285954362  + 1141.13406340540 *T)
    sumR1  +=       1.401 * cos(4.81225317549  + 1251.34038462480 *T)
    sumR1  +=       1.244 * cos(2.83383980283  + 124.43341522100 *T)
    sumR1  +=       1.320 * cos(5.80675430384  + 387.24131496080 *T)
    sumR1  +=       1.329 * cos(0.88314574243  + 916.93228005540 *T)
    sumR1  +=       1.558 * cos(6.17808619637  + 983.11585891360 *T)
    sumR1  +=       1.243 * cos(0.29239666059  + 597.35901666110 *T)
    sumR1  +=       1.541 * cos(3.51095241498  + 2751.54759969160 *T)
    sumR1  +=       1.482 * cos(0.83066678204  + 529.16970023280 *T)
    sumR1  +=       1.149 * cos(3.91142023857  + 99.91138048090 *T)
    sumR1  +=       1.114 * cos(3.53339637290  + 483.22054217860 *T)
    sumR1  +=       1.195 * cos(4.16301075999  + 203.00415469950 *T)
    sumR1  +=       1.100 * cos(1.74769285223  + 497.44763618020 *T)
    sumR1  +=       1.458 * cos(5.19315120878  + 1592.25428902150 *T)
    sumR1  +=       1.123 * cos(1.45270581179  + 533.88375078860 *T)
    sumR1  +=       1.078 * cos(5.23991792940  + 1159.29331067010 *T)
    sumR1  +=       1.083 * cos(3.57026506855  + 2943.50605412720 *T)
    sumR1  +=       1.072 * cos(0.07132659992  + 1070.42763045310 *T)
    sumR1  +=       1.037 * cos(5.48955598976  + 1585.89150154610 *T)
    sumR1  +=       1.343 * cos(0.29600445633  + 860.30992875280 *T)
    sumR1  +=       1.361 * cos(3.46603373194  + 107.02492748170 *T)
    sumR1  +=       1.061 * cos(2.44580706826  + 1048.33622992530 *T)
    sumR1  +=       1.002 * cos(5.55216117410  + 337.73251065900 *T)
    sumR1  +=       0.981 * cos(3.15500987023  + 70.84944530420 *T)
    sumR1  +=       1.007 * cos(4.11504050436  + 501.23677709140 *T)
    sumR1  +=       0.965 * cos(5.63719524421  + 1603.29998928540 *T)
    sumR1  +=       1.083 * cos(4.88373909810  + 1166.40685767090 *T)
    sumR1  +=       0.953 * cos(2.83352026342  + 3583.40334044120 *T)
    sumR1  +=       1.060 * cos(3.18542176646  + 447.79581952650 *T)
    sumR1  +=       1.136 * cos(2.26568590950  + 525.49817940060 *T)
    sumR1  +=       1.191 * cos(2.25249961404  + 106.27416795630 *T)
    sumR1  +=       0.884 * cos(4.69777781327  + 960.22130923370 *T)
    sumR1  +=       1.165 * cos(1.56030440737  + 630.33605875840 *T)
    sumR1  +=       0.947 * cos(0.50856414717  + 842.90144101350 *T)
    sumR1  +=       1.011 * cos(0.30814674949  + 1593.00504854690 *T)
    sumR1  +=       0.924 * cos(2.31939900786  + 327.43756992050 *T)
    sumR1  +=       0.896 * cos(0.22222521202  + 746.92221379570 *T)
    sumR1  +=       1.078 * cos(4.78329116086  + 2730.20695868920 *T)
    sumR1  +=       0.938 * cos(5.42471506763  + 1585.14074202070 *T)
    sumR1  +=       0.923 * cos(4.44469169065  + 9676.48103411560 *T)
    sumR1  +=       0.894 * cos(0.26940821870  + 2655.56837247380 *T)
    sumR1  +=       1.131 * cos(5.46382510304  + 224.34479570190 *T)
    sumR1  +=       0.808 * cos(0.48295590141  + 3377.21779200400 *T)
    sumR1  +=       0.809 * cos(4.14122746067  + 114.13847448250 *T)
    sumR1  +=       0.864 * cos(1.83217006136  + 4.66586644600 *T)
    sumR1  +=       1.106 * cos(2.60444312553  + 209.36694217490 *T)
    sumR1  +=       0.790 * cos(0.11493626208  + 460.53844081980 *T)
    sumR1  +=       0.799 * cos(1.60426497590  + 5223.69391980220 *T)
    sumR1  +=       0.933 * cos(0.30976125598  + 685.47393735270 *T)
    sumR1  +=       1.053 * cos(5.23433104008  + 842.15068148810 *T)
    sumR1  +=       0.846 * cos(3.02878393490  + 5746.27133789600 *T)
    sumR1  +=       0.799 * cos(2.08457026425  + 77734.01845962799 *T)
    sumR1  +=       0.820 * cos(0.99821486743  + 373.01422095920 *T)
    sumR1  +=       0.892 * cos(5.36446426391  + 827.92358748650 *T)
    sumR1  +=       0.821 * cos(3.53889274951  + 498.67147645760 *T)
    sumR1  +=       0.741 * cos(1.32379374647  + 530.21222995640 *T)
    sumR1  +=       0.790 * cos(2.88034567513  + 938.12990870590 *T)
    sumR1  +=       0.842 * cos(3.39449778904  + 484.44438245600 *T)
    sumR1  +=       0.785 * cos(0.57841470897  + 850.01498801430 *T)
    sumR1  +=       0.759 * cos(3.82014112009  + 6283.07584999140 *T)
    sumR1  +=       0.954 * cos(2.94534072982  + 462.02291352810 *T)
    sumR1  +=       0.767 * cos(3.33725133157  + 99.16062095550 *T)
    sumR1  +=       0.810 * cos(4.69425300466  + 2228.97018159780 *T)
    sumR1  +=       0.700 * cos(1.72050221502  + 775.23338944700 *T)
    sumR1  +=       0.764 * cos(4.91747674296  + 1670.82502850000 *T)
    sumR1  +=       0.724 * cos(6.08692841992  + 2281.23049651060 *T)
    sumR1  +=       0.711 * cos(4.82250918143  + 11.77941344680 *T)
    sumR1  +=       0.692 * cos(2.63705354662  + 6.59228213900 *T)
    sumR1  +=       0.771 * cos(3.87410612014  + 9690.70812811720 *T)
    sumR1  +=       0.906 * cos(2.47189948442  + 3274.12501778540 *T)
    sumR1  +=       0.781 * cos(1.25357484582  + 202.25339517410 *T)
    sumR1  +=       0.757 * cos(3.78079814332  + 2818.03500860600 *T)
    sumR1  +=       0.756 * cos(4.28312053897  + 2803.80791460440 *T)
    sumR1  +=       0.663 * cos(5.27704405712  + 4532.57894941100 *T)
    sumR1  +=       0.759 * cos(5.45358686570  + 9683.59458111640 *T)
    sumR1  +=       0.698 * cos(5.43712520216  + 565.11568774670 *T)
    sumR1  +=       0.709 * cos(3.71117647887  + 3686.49611465980 *T)
    sumR1  +=       0.677 * cos(4.27891183416  + 25028.52121138500 *T)
    sumR1  +=       0.643 * cos(1.40239510103  + 9161.01716302260 *T)
    sumR1  +=       0.656 * cos(0.60909845504  + 835.78789401270 *T)
    sumR1  +=       0.635 * cos(5.75373871128  + 429.77958461370 *T)
    sumR1  +=       0.702 * cos(6.10412979847  + 4635.67172362960 *T)
    sumR1  +=       0.627 * cos(3.03666956129  + 2840.41327990860 *T)
    sumR1  +=       0.802 * cos(4.18688054701  + 5753.38488489680 *T)
    sumR1  +=       0.838 * cos(4.51386507097  + 1069.67687092770 *T)
    sumR1  +=       0.633 * cos(4.37183361444  + 5.41662597140 *T)
    sumR1  +=       0.652 * cos(5.79409889124  + 1061.82961074400 *T)
    sumR1  +=       0.638 * cos(2.18896270346  + 313.21047591890 *T)
    sumR1  +=       0.827 * cos(5.94231186039  + 1457.52593306200 *T)
    sumR1  +=       0.678 * cos(2.45013730979  + 5760.49843189760 *T)
    sumR1  +=       0.814 * cos(4.89578791170  + 1567.73225428140 *T)
    sumR1  +=       0.624 * cos(0.61631100566  + 1176.70179840940 *T)
    sumR1  +=       0.600 * cos(3.20918322285  + 1098.73880610440 *T)
    sumR1  +=       0.717 * cos(1.82349064490  + 3171.03224356680 *T)
    sumR1  +=       0.651 * cos(4.14419317491  + 2847.52682690940 *T)
    sumR1  +=       0.629 * cos(1.75272560843  + 92.04707395470 *T)
    sumR1  +=       0.626 * cos(3.53146082217  + 3067.93946934820 *T)
    sumR1  +=       0.667 * cos(4.22974611158  + 4539.69249641180 *T)
    sumR1  +=       0.565 * cos(0.99416346033  + 1894.41906467650 *T)
    sumR1  +=       0.752 * cos(0.46063700150  + 635.23141986800 *T)
    sumR1  +=       0.622 * cos(1.98136818407  + 25565.32572348040 *T)
    sumR1  +=       0.614 * cos(2.48275371627  + 25551.09862947879 *T)
    sumR1  +=       0.560 * cos(1.40733893388  + 446.31134681820 *T)
    sumR1  +=       0.558 * cos(4.37217796469  + 1057.89745748090 *T)
    sumR1  +=       0.628 * cos(4.65037810102  + 6275.96230299060 *T)
    sumR1  +=       0.659 * cos(2.41470950463  + 195.13984817330 *T)
    sumR1  +=       0.616 * cos(2.08837621877  + 10.03430830760 *T)
    sumR1  +=       0.692 * cos(3.13229025530  + 7.63481186260 *T)
    sumR1  +=       0.685 * cos(4.18539472904  + 46.47042291600 *T)
    sumR1  +=       0.624 * cos(0.02693303471  + 1493.09366806600 *T)
    sumR1  +=       0.594 * cos(2.13375704438  + 121.25202148330 *T)
    sumR1  +=       0.508 * cos(2.13584300710  + 1.69692102940 *T)
    sumR1  +=       0.674 * cos(1.47570122611  + 4694.00295470760 *T)
    sumR1  +=       0.559 * cos(4.48852017557  + 531.17543780290 *T)
    sumR1  +=       0.640 * cos(3.10239233469  + 11.04570026390 *T)
    sumR1  +=       0.496 * cos(1.29000001439  + 927.83496796740 *T)
    sumR1  +=       0.587 * cos(3.30651435298  + 600.54041039880 *T)
    sumR1  +=       0.582 * cos(0.44540948860  + 113.38771495710 *T)
    sumR1  +=       0.492 * cos(4.83275232000  + 9492.14631500480 *T)
    sumR1  +=       0.549 * cos(4.34579166146  + 3046.59882834580 *T)
    sumR1  +=       0.576 * cos(1.22846846364  + 1514.29129671650 *T)
    sumR1  +=       0.593 * cos(5.86079640612  + 524.06189080210 *T)
    sumR1  +=       0.510 * cos(2.62557031270  + 529.85102378900 *T)
    sumR1  +=       0.489 * cos(6.26855707323  + 3693.60966166060 *T)
    sumR1  +=       0.480 * cos(0.30754294369  + 528.94020556920 *T)
    sumR1  +=       0.582 * cos(3.51934668795  + 1056.93424963440 *T)
    sumR1  +=       0.493 * cos(5.52699906925  + 512.28247735530 *T)
    sumR1  +=       0.481 * cos(2.99681040149  + 9153.90361602180 *T)
    sumR1  +=       0.562 * cos(3.73437025868  + 2015.67108615980 *T)
    sumR1  +=       0.458 * cos(3.86646994292  + 11.30633269480 *T)
    sumR1  +=       0.457 * cos(1.80238019931  + 3281.23856478620 *T)
    sumR1  +=       0.453 * cos(6.17995938655  + 1059.33374607940 *T)
    sumR1  +=       0.551 * cos(0.13794958618  + 1912.57831194120 *T)
    sumR1  +=       0.446 * cos(5.53828660924  + 2332.06295581640 *T)
    sumR1  +=       0.444 * cos(5.06219342598  + 7.86430652620 *T)
    sumR1  +=       0.461 * cos(0.16951411708  + 26087.90314157420 *T)
    sumR1  +=       0.439 * cos(4.14986379679  + 1151.42900414390 *T)
    sumR1  +=       0.614 * cos(5.42289673768  + 2090.30967237520 *T)
    sumR1  +=       0.488 * cos(3.71681959056  + 447.93883187840 *T)
    sumR1  +=       0.592 * cos(2.91424148255  + 8624.21265092720 *T)
    sumR1  +=       0.433 * cos(2.55336268329  + 1064.04779663520 *T)
    sumR1  +=       0.449 * cos(5.24955106938  + 10213.28554621100 *T)
    sumR1  +=       0.510 * cos(5.81591864532  + 529.53090640020 *T)
    sumR1  +=       0.435 * cos(5.34355963629  + 560.71045373160 *T)
    sumR1  +=       0.449 * cos(0.72330388784  + 2758.66114669240 *T)
    sumR1  +=       0.430 * cos(0.94519103478  + 6.36278747540 *T)
    sumR1  +=       0.563 * cos(6.19175228344  + 1884.12412393800 *T)
    sumR1  +=       0.443 * cos(3.39246520261  + 1152.17976366930 *T)
    sumR1  +=       0.430 * cos(1.28652623263  + 505.31194270640 *T)
    sumR1  +=       0.422 * cos(5.12631540623  + 944.98282327580 *T)
    sumR1  +=       0.464 * cos(2.90444584145  + 398.14400287280 *T)
    sumR1  +=       0.410 * cos(1.24248975309  + 5069.38346150640 *T)
    sumR1  +=       0.411 * cos(2.95117124177  + 4326.39340097380 *T)
    sumR1  +=       0.418 * cos(5.15499986314  + 1173.52040467170 *T)
    sumR1  +=       0.412 * cos(2.98125446330  + 554.06998748280 *T)
    sumR1  +=       0.403 * cos(0.34381388674  + 32.24332891440 *T)
    sumR1  +=       0.402 * cos(5.88926765351  + 1570.91364801910 *T)
    sumR1  +=       0.505 * cos(1.49028912471  + 3782.47534187760 *T)
    sumR1  +=       0.447 * cos(0.03952029309  + 245.54242435240 *T)
    sumR1  +=       0.453 * cos(3.09458004153  + 1059.43011429900 *T)
    sumR1  +=       0.411 * cos(3.21727542472  + 1475.68518032670 *T)
    sumR1  +=       0.426 * cos(3.12237794195  + 12566.15169998280 *T)
    sumR1  +=       0.434 * cos(3.59362426939  + 3259.89792378380 *T)
    sumR1  +=       0.398 * cos(4.91510709622  + 4120.20785253660 *T)
    sumR1  +=       0.399 * cos(4.67075122011  + 234.63973644040 *T)
    sumR1  +=       0.386 * cos(4.81320787761  + 970.51624997220 *T)
    sumR1  +=       0.427 * cos(3.21176085113  + 977.48678462110 *T)
    sumR1  +=       0.411 * cos(4.31566962034  + 757.21715453420 *T)
    sumR1  +=       0.392 * cos(1.86527946688  + 885.43971066640 *T)
    sumR1  +=       0.416 * cos(3.81408093105  + 3156.80514956520 *T)

    # JUPITER Radius
    sumR2  = 0
    sumR2  +=   79644.833 * cos(1.35865896596  + 529.69096509460 *T)
    sumR2  +=    8251.618 * cos(5.77773935444  + 522.57741809380 *T)
    sumR2  +=    7029.864 * cos(3.27476965833  + 536.80451209540 *T)
    sumR2  +=    5314.006 * cos(1.83835109712  + 1059.38193018920 *T)
    sumR2  +=    1860.833 * cos(2.97682139367  + 7.11354700080 *T)
    sumR2  +=     836.267 * cos(4.19889881718  + 419.48464387520 *T)
    sumR2  +=     964.466 * cos(5.48031822015  + 515.46387109300 *T)
    sumR2  +=     406.453 * cos(3.78250730354  + 1066.49547719000 *T)
    sumR2  +=     426.570 * cos(2.22753101795  + 639.89728631400 *T)
    sumR2  +=     377.316 * cos(2.24248352873  + 1589.07289528380 *T)
    sumR2  +=     497.920 * cos(3.14159265359  + 0.00000000000 *T)
    sumR2  +=     339.043 * cos(6.12690864038  + 625.67019231240 *T)
    sumR2  +=     362.943 * cos(5.36761847267  + 206.18554843720 *T)
    sumR2  +=     342.048 * cos(6.09922969324  + 1052.26838318840 *T)
    sumR2  +=     279.920 * cos(4.26162555827  + 412.37109687440 *T)
    sumR2  +=     332.578 * cos(0.00328961161  + 426.59819087600 *T)
    sumR2  +=     229.777 * cos(0.70530766213  + 735.87651353180 *T)
    sumR2  +=     200.783 * cos(3.06850623368  + 543.91805909620 *T)
    sumR2  +=     199.807 * cos(4.42884165317  + 103.09277421860 *T)
    sumR2  +=     257.290 * cos(0.96295364983  + 632.78373931320 *T)
    sumR2  +=     138.606 * cos(2.93235671606  + 14.22709400160 *T)
    sumR2  +=     113.535 * cos(0.78713911289  + 728.76296653100 *T)
    sumR2  +=      86.025 * cos(5.14434751994  + 323.50541665740 *T)
    sumR2  +=      94.565 * cos(1.70498041073  + 838.96928775040 *T)
    sumR2  +=      83.469 * cos(0.05834873484  + 309.27832265580 *T)
    sumR2  +=      75.198 * cos(1.60495195911  + 956.28915597060 *T)
    sumR2  +=      70.451 * cos(1.50988357484  + 213.29909543800 *T)
    sumR2  +=      80.328 * cos(2.98122361797  + 742.99006053260 *T)
    sumR2  +=      56.203 * cos(0.95534810533  + 1162.47470440780 *T)
    sumR2  +=      61.649 * cos(6.10137889854  + 1045.15483618760 *T)
    sumR2  +=      66.572 * cos(5.47307178077  + 199.07200143640 *T)
    sumR2  +=      50.057 * cos(2.72063162317  + 532.87235883230 *T)
    sumR2  +=      51.904 * cos(5.58435625607  + 942.06206196900 *T)
    sumR2  +=      39.833 * cos(5.94566506227  + 95.97922721780 *T)
    sumR2  +=      44.548 * cos(5.52445621411  + 508.35032409220 *T)
    sumR2  +=      44.282 * cos(0.27118152557  + 526.50957135690 *T)
    sumR2  +=      29.944 * cos(0.93641735919  + 1155.36115740700 *T)
    sumR2  +=      28.412 * cos(2.87835720211  + 525.75881183150 *T)
    sumR2  +=      26.330 * cos(4.26891877269  + 1596.18644228460 *T)
    sumR2  +=      27.039 * cos(2.80607741398  + 1169.58825140860 *T)
    sumR2  +=      27.477 * cos(2.64841266238  + 2118.76386037840 *T)
    sumR2  +=      22.705 * cos(0.17830004133  + 302.16477565500 *T)
    sumR2  +=      29.347 * cos(1.78589692350  + 831.85574074960 *T)
    sumR2  +=      19.991 * cos(0.04328951895  + 949.17560896980 *T)
    sumR2  +=      19.906 * cos(1.16072627347  + 533.62311835770 *T)
    sumR2  +=      21.714 * cos(1.88820231818  + 1272.68102562720 *T)
    sumR2  +=      17.581 * cos(4.14974757919  + 846.08283475120 *T)
    sumR2  +=      17.085 * cos(5.89188996975  + 1258.45393162560 *T)
    sumR2  +=      21.407 * cos(4.35468497204  + 316.39186965660 *T)
    sumR2  +=      21.295 * cos(0.54429472455  + 1265.56747862640 *T)
    sumR2  +=      19.859 * cos(0.06453825800  + 1581.95934828300 *T)
    sumR2  +=      17.025 * cos(0.53383755278  + 1368.66025284500 *T)
    sumR2  +=      12.804 * cos(3.90044242142  + 433.71173787680 *T)
    sumR2  +=      13.072 * cos(0.79468040717  + 110.20632121940 *T)
    sumR2  +=      11.945 * cos(0.40671403646  + 1361.54670584420 *T)
    sumR2  +=      11.695 * cos(4.44394618065  + 405.25754987360 *T)
    sumR2  +=      11.979 * cos(2.22872778682  + 220.41264243880 *T)
    sumR2  +=       9.633 * cos(6.01002272123  + 853.19638175200 *T)
    sumR2  +=      10.163 * cos(0.99504635158  + 1471.75302706360 *T)
    sumR2  +=       8.977 * cos(1.60328709409  + 1692.16566950240 *T)
    sumR2  +=       8.701 * cos(3.52167876799  + 1073.60902419080 *T)
    sumR2  +=       8.314 * cos(5.60169732564  + 1574.84580128220 *T)
    sumR2  +=       8.958 * cos(6.26708748901  + 519.39602435610 *T)
    sumR2  +=       7.828 * cos(0.65241611799  + 1478.86657406440 *T)
    sumR2  +=       7.833 * cos(0.17920601344  + 1685.05212250160 *T)
    sumR2  +=       7.451 * cos(0.88421084942  + 88.86568021700 *T)
    sumR2  +=       7.320 * cos(0.89341249264  + 721.64941953020 *T)
    sumR2  +=       9.135 * cos(1.51210840939  + 1148.24761040620 *T)
    sumR2  +=       6.110 * cos(2.50080005128  + 3.18139373770 *T)
    sumR2  +=       7.037 * cos(4.44127496638  + 330.61896365820 *T)
    sumR2  +=       5.163 * cos(2.79219166952  + 21.34064100240 *T)
    sumR2  +=       5.079 * cos(2.97991736844  + 1375.77379984580 *T)
    sumR2  +=       4.930 * cos(0.04683167622  + 1677.93857550080 *T)
    sumR2  +=       4.664 * cos(2.28007273876  + 1485.98012106520 *T)
    sumR2  +=       4.692 * cos(0.86220230505  + 3.93215326310 *T)
    sumR2  +=       5.307 * cos(0.85008578245  + 1788.14489672020 *T)
    sumR2  +=       4.239 * cos(0.40758287124  + 629.60234557550 *T)
    sumR2  +=       4.230 * cos(1.61046658091  + 635.96513305090 *T)
    sumR2  +=       3.627 * cos(2.71151441113  + 551.03160609700 *T)
    sumR2  +=       3.314 * cos(0.55067236587  + 1795.25844372100 *T)
    sumR2  +=       4.409 * cos(1.28127751050  + 1464.63948006280 *T)
    sumR2  +=       3.270 * cos(1.18744032691  + 1905.46476494040 *T)
    sumR2  +=       3.226 * cos(6.18716071251  + 1038.04128918680 *T)
    sumR2  +=       3.103 * cos(6.22971614425  + 2001.44399215820 *T)
    sumR2  +=       3.410 * cos(2.44624067925  + 539.98590583310 *T)
    sumR2  +=       3.174 * cos(5.54870592599  + 191.95845443560 *T)
    sumR2  +=       2.590 * cos(3.24430559059  + 1062.56332392690 *T)
    sumR2  +=       2.614 * cos(0.55149554149  + 2104.53676637680 *T)
    sumR2  +=       2.174 * cos(5.32613824409  + 1891.23767093880 *T)
    sumR2  +=       2.659 * cos(4.82459974220  + 416.30325013750 *T)
    sumR2  +=       2.187 * cos(1.71707514653  + 628.85158605010 *T)
    sumR2  +=       2.263 * cos(6.19233486371  + 1994.33044515740 *T)
    sumR2  +=       2.328 * cos(4.28236795066  + 963.40270297140 *T)
    sumR2  +=       2.579 * cos(0.03256542251  + 1898.35121793960 *T)
    sumR2  +=       2.077 * cos(3.32602157426  + 1699.27921650320 *T)
    sumR2  +=       2.529 * cos(2.39697505835  + 227.52618943960 *T)
    sumR2  +=       2.468 * cos(0.06551346218  + 750.10360753340 *T)
    sumR2  +=       1.989 * cos(0.29206371261  + 636.71589257630 *T)
    sumR2  +=       1.927 * cos(0.32286661566  + 295.05122865420 *T)
    sumR2  +=       1.904 * cos(3.43534792123  + 647.01083331480 *T)
    sumR2  +=       1.940 * cos(0.29170673525  + 2111.65031337760 *T)
    sumR2  +=       1.880 * cos(3.14403615586  + 611.44309831080 *T)
    sumR2  +=       2.324 * cos(1.94960720763  + 824.74219374880 *T)
    sumR2  +=       1.854 * cos(4.71794950485  + 2125.87740737920 *T)
    sumR2  +=       2.547 * cos(1.23908353090  + 2221.85663459700 *T)
    sumR2  +=       1.814 * cos(1.60250861074  + 2008.55753915900 *T)
    sumR2  +=       1.611 * cos(5.83466560322  + 422.66603761290 *T)
    sumR2  +=       1.667 * cos(2.32455940876  + 440.82528487760 *T)
    sumR2  +=       1.622 * cos(0.36650974375  + 1056.20053645150 *T)
    sumR2  +=       1.624 * cos(2.42139677881  + 10.29494073850 *T)
    sumR2  +=       1.622 * cos(3.51892791175  + 1055.44977692610 *T)
    sumR2  +=       1.606 * cos(5.76205763975  + 117.31986822020 *T)
    sumR2  +=       1.646 * cos(5.88662636573  + 2317.83586181480 *T)
    sumR2  +=       2.026 * cos(4.61781314145  + 423.41679713830 *T)
    sumR2  +=       2.098 * cos(1.04559231028  + 1781.03134971940 *T)
    sumR2  +=       1.868 * cos(1.12487729469  + 618.55664531160 *T)
    sumR2  +=       1.885 * cos(2.78775930564  + 1802.37199072180 *T)
    sumR2  +=       1.445 * cos(0.08308050305  + 1382.88734684660 *T)
    sumR2  +=       1.797 * cos(3.00776822706  + 2648.45482547300 *T)
    sumR2  +=       1.422 * cos(0.17649746278  + 2420.92863603340 *T)
    sumR2  +=       1.129 * cos(1.59030291320  + 380.12776796000 *T)
    sumR2  +=       1.126 * cos(4.19989673600  + 547.85021235930 *T)
    sumR2  +=       1.186 * cos(5.98943062173  + 2310.72231481400 *T)
    sumR2  +=       1.108 * cos(4.22655117757  + 934.94851496820 *T)
    sumR2  +=       1.259 * cos(1.19687222266  + 1063.31408345230 *T)
    sumR2  +=       1.072 * cos(3.86169004168  + 1603.29998928540 *T)
    sumR2  +=       0.946 * cos(5.59968097387  + 99.91138048090 *T)
    sumR2  +=       0.937 * cos(1.03083276760  + 81.75213321620 *T)
    sumR2  +=       0.938 * cos(6.18136092771  + 945.99421523210 *T)
    sumR2  +=       0.908 * cos(2.54355964041  + 6283.07584999140 *T)
    sumR2  +=       0.874 * cos(5.21903196047  + 2207.62954059540 *T)
    sumR2  +=       0.874 * cos(6.01240284465  + 511.53171782990 *T)
    sumR2  +=       1.188 * cos(0.75698357968  + 2097.42321937600 *T)
    sumR2  +=       0.789 * cos(3.91035208173  + 10213.28554621100 *T)
    sumR2  +=       1.000 * cos(1.34667100304  + 732.69511979410 *T)
    sumR2  +=       0.952 * cos(1.55355777420  + 2324.94940881560 *T)
    sumR2  +=       0.811 * cos(5.00475553271  + 319.57326339430 *T)
    sumR2  +=       0.763 * cos(3.98527559630  + 337.73251065900 *T)
    sumR2  +=       0.880 * cos(1.14789972199  + 952.35700270750 *T)
    sumR2  +=       0.780 * cos(4.69463316930  + 5746.27133789600 *T)
    sumR2  +=       0.910 * cos(0.08774541571  + 2737.32050569000 *T)
    sumR2  +=       0.773 * cos(0.77131695762  + 5760.49843189760 *T)
    sumR2  +=       0.764 * cos(6.11686539353  + 9676.48103411560 *T)
    sumR2  +=       0.758 * cos(2.19350719860  + 9690.70812811720 *T)
    sumR2  +=       0.671 * cos(1.19532387143  + 124.43341522100 *T)
    sumR2  +=       0.661 * cos(5.99578306627  + 501.23677709140 *T)
    sumR2  +=       0.729 * cos(0.65312263578  + 2538.24850425360 *T)
    sumR2  +=       0.825 * cos(2.70770030205  + 3370.10424500320 *T)
    sumR2  +=       0.670 * cos(5.44169923277  + 107.02492748170 *T)
    sumR2  +=       0.739 * cos(1.14609907817  + 2641.34127847220 *T)
    sumR2  +=       0.866 * cos(3.02831268213  + 3046.59882834580 *T)
    sumR2  +=       0.718 * cos(4.83684196454  + 860.30992875280 *T)
    sumR2  +=       0.813 * cos(6.01229270247  + 2214.74308759620 *T)
    sumR2  +=       0.746 * cos(1.12371143332  + 739.80866679490 *T)
    sumR2  +=       0.741 * cos(5.93171662010  + 2634.22773147140 *T)
    sumR2  +=       0.667 * cos(0.89885058003  + 106.27416795630 *T)
    sumR2  +=       0.573 * cos(2.42701822581  + 739.05790726950 *T)
    sumR2  +=       0.734 * cos(0.72837704619  + 1354.43315884340 *T)
    sumR2  +=       0.662 * cos(2.21768976390  + 2015.67108615980 *T)
    sumR2  +=       0.782 * cos(2.52401202862  + 3679.38256765900 *T)
    sumR2  +=       0.779 * cos(2.38608991574  + 3267.01147078460 *T)
    sumR2  +=       0.553 * cos(1.85211127676  + 453.42489381900 *T)
    sumR2  +=       0.701 * cos(4.23431087374  + 9683.59458111640 *T)
    sumR2  +=       0.571 * cos(2.98435419019  + 1262.38608488870 *T)
    sumR2  +=       0.621 * cos(1.24462887440  + 3803.81598288000 *T)
    sumR2  +=       0.563 * cos(5.99845316446  + 1049.08698945070 *T)
    sumR2  +=       0.538 * cos(4.92334194042  + 447.79581952650 *T)
    sumR2  +=       0.534 * cos(0.99911551571  + 462.02291352810 *T)
    sumR2  +=       0.541 * cos(6.19275150397  + 1987.21689815660 *T)
    sumR2  +=       0.511 * cos(3.28553278370  + 4.66586644600 *T)
    sumR2  +=       0.539 * cos(5.33214565622  + 2751.54759969160 *T)
    sumR2  +=       0.651 * cos(5.12199308959  + 3156.80514956520 *T)
    sumR2  +=       0.483 * cos(3.03782387056  + 3281.23856478620 *T)
    sumR2  +=       0.476 * cos(2.17592053936  + 149.56319713460 *T)
    sumR2  +=       0.510 * cos(5.35664230912  + 9.56122755560 *T)
    sumR2  +=       0.490 * cos(1.57324553106  + 1251.34038462480 *T)
    sumR2  +=       0.467 * cos(5.92343423840  + 203.00415469950 *T)
    sumR2  +=       0.528 * cos(5.81786945766  + 2627.11418447060 *T)
    sumR2  +=       0.447 * cos(3.51498961805  + 18.15924726470 *T)
    sumR2  +=       0.429 * cos(0.16627197188  + 74.78159856730 *T)
    sumR2  +=       0.497 * cos(0.30985248432  + 2428.04218303420 *T)
    sumR2  +=       0.516 * cos(3.89424540015  + 2516.90786325120 *T)
    sumR2  +=       0.519 * cos(2.43126348834  + 3686.49611465980 *T)
    sumR2  +=       0.404 * cos(2.77840802846  + 7.16173111060 *T)
    sumR2  +=       0.533 * cos(4.77083438961  + 3473.19701922180 *T)
    sumR2  +=       0.515 * cos(3.54549816613  + 3178.14579056760 *T)
    sumR2  +=       0.533 * cos(5.61415688189  + 2524.02141025200 *T)
    sumR2  +=       0.458 * cos(4.91616403047  + 3067.93946934820 *T)
    sumR2  +=       0.400 * cos(3.13887720912  + 540.73666535850 *T)
    sumR2  +=       0.378 * cos(0.86122450940  + 525.02509864860 *T)

    # JUPITER Radius
    sumR3  = 0
    sumR3  +=    3519.257 * cos(6.05800633846  + 529.69096509460 *T)
    sumR3  +=    1073.239 * cos(1.67321345760  + 536.80451209540 *T)
    sumR3  +=     915.666 * cos(1.41329676116  + 522.57741809380 *T)
    sumR3  +=     341.593 * cos(0.52296542656  + 1059.38193018920 *T)
    sumR3  +=     254.893 * cos(1.19625473533  + 7.11354700080 *T)
    sumR3  +=     221.512 * cos(0.95225226237  + 515.46387109300 *T)
    sumR3  +=      69.078 * cos(2.26885282314  + 1066.49547719000 *T)
    sumR3  +=      89.729 * cos(3.14159265359  + 0.00000000000 *T)
    sumR3  +=      57.827 * cos(1.41389745339  + 543.91805909620 *T)
    sumR3  +=      57.653 * cos(0.52580117593  + 639.89728631400 *T)
    sumR3  +=      51.079 * cos(5.98016364677  + 412.37109687440 *T)
    sumR3  +=      46.935 * cos(1.57864237959  + 625.67019231240 *T)
    sumR3  +=      42.824 * cos(6.11689609099  + 419.48464387520 *T)
    sumR3  +=      37.477 * cos(1.18262762330  + 14.22709400160 *T)
    sumR3  +=      33.816 * cos(1.66671706951  + 1052.26838318840 *T)
    sumR3  +=      31.195 * cos(1.04290245896  + 1589.07289528380 *T)
    sumR3  +=      30.023 * cos(4.63236245032  + 426.59819087600 *T)
    sumR3  +=      33.531 * cos(0.84784977903  + 206.18554843720 *T)
    sumR3  +=      20.804 * cos(2.50071243814  + 728.76296653100 *T)
    sumR3  +=      14.466 * cos(0.96040197071  + 508.35032409220 *T)
    sumR3  +=      12.969 * cos(1.50233788550  + 1045.15483618760 *T)
    sumR3  +=      11.654 * cos(3.55513510121  + 323.50541665740 *T)
    sumR3  +=      12.319 * cos(2.60952614503  + 735.87651353180 *T)
    sumR3  +=      15.023 * cos(0.89136998434  + 199.07200143640 *T)
    sumR3  +=      11.160 * cos(1.79041437555  + 309.27832265580 *T)
    sumR3  +=      10.554 * cos(6.27845112678  + 956.28915597060 *T)
    sumR3  +=       9.812 * cos(6.26016859519  + 103.09277421860 *T)
    sumR3  +=       9.301 * cos(3.45126812476  + 838.96928775040 *T)
    sumR3  +=       6.672 * cos(1.87004905364  + 302.16477565500 *T)
    sumR3  +=       7.442 * cos(1.28047007623  + 742.99006053260 *T)
    sumR3  +=       7.178 * cos(0.92022189637  + 942.06206196900 *T)
    sumR3  +=       5.577 * cos(1.37980792905  + 95.97922721780 *T)
    sumR3  +=       6.834 * cos(3.45228722967  + 831.85574074960 *T)
    sumR3  +=       4.632 * cos(2.82934545414  + 1596.18644228460 *T)
    sumR3  +=       3.969 * cos(1.21290005054  + 1169.58825140860 *T)
    sumR3  +=       3.869 * cos(5.99495313698  + 213.29909543800 *T)
    sumR3  +=       3.551 * cos(6.10714791535  + 405.25754987360 *T)
    sumR3  +=       2.943 * cos(2.32831075458  + 1155.36115740700 *T)
    sumR3  +=       2.442 * cos(1.86965213405  + 532.87235883230 *T)
    sumR3  +=       2.410 * cos(0.42627205128  + 220.41264243880 *T)
    sumR3  +=       2.289 * cos(1.94941487274  + 1073.60902419080 *T)
    sumR3  +=       2.274 * cos(0.09211517505  + 632.78373931320 *T)
    sumR3  +=       2.189 * cos(1.58907745204  + 2118.76386037840 *T)
    sumR3  +=       2.387 * cos(5.97080671477  + 1162.47470440780 *T)
    sumR3  +=       2.104 * cos(1.06751462671  + 21.34064100240 *T)
    sumR3  +=       2.128 * cos(1.51119399925  + 1258.45393162560 *T)
    sumR3  +=       2.491 * cos(0.35125020737  + 1272.68102562720 *T)
    sumR3  +=       2.006 * cos(5.94487388360  + 110.20632121940 *T)
    sumR3  +=       1.980 * cos(2.54989377864  + 88.86568021700 *T)
    sumR3  +=       2.040 * cos(2.16463966964  + 433.71173787680 *T)
    sumR3  +=       1.955 * cos(2.70341589777  + 721.64941953020 *T)
    sumR3  +=       1.670 * cos(4.46255717328  + 853.19638175200 *T)
    sumR3  +=       1.910 * cos(2.25964760758  + 1361.54670584420 *T)
    sumR3  +=       1.710 * cos(1.98372066321  + 525.75881183150 *T)
    sumR3  +=       1.520 * cos(0.11641358425  + 949.17560896980 *T)
    sumR3  +=       2.003 * cos(3.16520599208  + 1148.24761040620 *T)
    sumR3  +=       1.710 * cos(2.70850417287  + 330.61896365820 *T)
    sumR3  +=       1.629 * cos(0.47376028854  + 526.50957135690 *T)
    sumR3  +=       1.229 * cos(3.01987279595  + 963.40270297140 *T)
    sumR3  +=       1.671 * cos(0.44352103086  + 533.62311835770 *T)
    sumR3  +=       1.207 * cos(1.15774089269  + 1574.84580128220 *T)
    sumR3  +=       1.146 * cos(2.54505851138  + 846.08283475120 *T)
    sumR3  +=       1.355 * cos(1.17462112647  + 1038.04128918680 *T)
    sumR3  +=       1.001 * cos(2.70272799283  + 519.39602435610 *T)
    sumR3  +=       1.372 * cos(0.67467128629  + 551.03160609700 *T)
    sumR3  +=       0.983 * cos(4.17198081351  + 2627.11418447060 *T)
    sumR3  +=       1.084 * cos(1.07011164067  + 227.52618943960 *T)
    sumR3  +=       0.892 * cos(2.92543286761  + 1368.66025284500 *T)
    sumR3  +=       0.823 * cos(4.86559196955  + 611.44309831080 *T)
    sumR3  +=       1.136 * cos(1.78981738432  + 1581.95934828300 *T)
    sumR3  +=       0.897 * cos(4.91073630270  + 1670.82502850000 *T)
    sumR3  +=       0.908 * cos(3.68804047330  + 824.74219374880 *T)
    sumR3  +=       0.789 * cos(3.23380893250  + 2125.87740737920 *T)
    sumR3  +=       0.771 * cos(2.39070707004  + 2317.83586181480 *T)
    sumR3  +=       0.891 * cos(0.59692950778  + 539.98590583310 *T)
    sumR3  +=       0.876 * cos(4.52127091462  + 750.10360753340 *T)
    sumR3  +=       0.802 * cos(0.20759322884  + 1141.13406340540 *T)
    sumR3  +=       0.850 * cos(0.94145487094  + 191.95845443560 *T)
    sumR3  +=       0.762 * cos(2.25149516048  + 2538.24850425360 *T)
    sumR3  +=       0.694 * cos(0.67080348659  + 440.82528487760 *T)
    sumR3  +=       0.741 * cos(5.79934203525  + 1485.98012106520 *T)
    sumR3  +=       0.643 * cos(2.48127580335  + 1265.56747862640 *T)
    sumR3  +=       0.575 * cos(6.13756590872  + 1279.79457262800 *T)
    sumR3  +=       0.636 * cos(5.51001645505  + 2413.81508903260 *T)
    sumR3  +=       0.636 * cos(4.40777238491  + 1382.88734684660 *T)
    sumR3  +=       0.555 * cos(2.18233983981  + 1062.56332392690 *T)
    sumR3  +=       0.564 * cos(1.92775967119  + 2634.22773147140 *T)
    sumR3  +=       0.531 * cos(2.04824376019  + 295.05122865420 *T)
    sumR3  +=       0.541 * cos(2.32424368689  + 1471.75302706360 *T)
    sumR3  +=       0.697 * cos(2.27179476322  + 1699.27921650320 *T)
    sumR3  +=       0.546 * cos(1.95774905730  + 1677.93857550080 *T)
    sumR3  +=       0.465 * cos(4.35550844067  + 1692.16566950240 *T)
    sumR3  +=       0.508 * cos(2.50298248836  + 2207.62954059540 *T)
    sumR3  +=       0.496 * cos(5.77087043616  + 1478.86657406440 *T)
    sumR3  +=       0.440 * cos(5.98661963879  + 934.94851496820 *T)
    sumR3  +=       0.424 * cos(2.80194129521  + 81.75213321620 *T)
    sumR3  +=       0.406 * cos(3.93940190897  + 316.39186965660 *T)
    sumR3  +=       0.506 * cos(0.18719982992  + 10.29494073850 *T)

    # JUPITER Radius
    sumR4  = 0
    sumR4  +=     128.628 * cos(0.08419309557  + 536.80451209540 *T)
    sumR4  +=     113.458 * cos(4.24858855779  + 529.69096509460 *T)
    sumR4  +=      82.650 * cos(3.29754909408  + 522.57741809380 *T)
    sumR4  +=      37.883 * cos(2.73326611144  + 515.46387109300 *T)
    sumR4  +=      26.694 * cos(5.69142588558  + 7.11354700080 *T)
    sumR4  +=      17.650 * cos(5.40012536918  + 1059.38193018920 *T)
    sumR4  +=      12.612 * cos(6.01560416057  + 543.91805909620 *T)
    sumR4  +=       9.287 * cos(0.76813946494  + 1066.49547719000 *T)
    sumR4  +=       8.107 * cos(5.68228065707  + 14.22709400160 *T)
    sumR4  +=       6.271 * cos(5.12286932534  + 639.89728631400 *T)
    sumR4  +=       6.978 * cos(1.42751292055  + 412.37109687440 *T)
    sumR4  +=       5.377 * cos(3.33501947275  + 625.67019231240 *T)
    sumR4  +=       2.911 * cos(3.40334805052  + 1052.26838318840 *T)
    sumR4  +=       2.593 * cos(4.16090412984  + 728.76296653100 *T)
    sumR4  +=       2.562 * cos(2.89802035072  + 426.59819087600 *T)
    sumR4  +=       2.268 * cos(6.22195938856  + 1589.07289528380 *T)
    sumR4  +=       2.114 * cos(3.11758855774  + 1045.15483618760 *T)
    sumR4  +=       1.673 * cos(2.81399290364  + 206.18554843720 *T)
    sumR4  +=       1.805 * cos(2.60030006919  + 199.07200143640 *T)
    sumR4  +=       1.823 * cos(1.89432426038  + 419.48464387520 *T)
    sumR4  +=       1.522 * cos(1.33432648232  + 1596.18644228460 *T)
    sumR4  +=       1.697 * cos(0.00000000000  + 0.00000000000 *T)
    sumR4  +=       1.039 * cos(4.41904942302  + 956.28915597060 *T)
    sumR4  +=       1.161 * cos(5.16181311538  + 831.85574074960 *T)
    sumR4  +=       0.916 * cos(3.17245716108  + 508.35032409220 *T)
    sumR4  +=       0.870 * cos(5.79387813500  + 1169.58825140860 *T)
    sumR4  +=       0.916 * cos(1.87129662931  + 1148.24761040620 *T)
    sumR4  +=       0.955 * cos(0.66801367802  + 1361.54670584420 *T)
    sumR4  +=       0.788 * cos(1.47515450553  + 1272.68102562720 *T)
    sumR4  +=       0.966 * cos(5.47457968043  + 220.41264243880 *T)
    sumR4  +=       0.788 * cos(2.42252866885  + 117.31986822020 *T)
    sumR4  +=       0.712 * cos(0.49655897030  + 1073.60902419080 *T)
    sumR4  +=       0.656 * cos(3.53022740783  + 302.16477565500 *T)
    sumR4  +=       0.681 * cos(2.84507174340  + 191.95845443560 *T)
    sumR4  +=       0.771 * cos(2.19893222018  + 942.06206196900 *T)
    sumR4  +=       0.765 * cos(5.31147257700  + 551.03160609700 *T)
    sumR4  +=       0.667 * cos(3.72432305249  + 88.86568021700 *T)
    sumR4  +=       0.534 * cos(1.83172084748  + 647.01083331480 *T)
    sumR4  +=       0.553 * cos(0.85896003802  + 330.61896365820 *T)
    sumR4  +=       0.543 * cos(5.26057584439  + 21.34064100240 *T)
    sumR4  +=       0.584 * cos(3.82243061802  + 618.55664531160 *T)
    sumR4  +=       0.512 * cos(4.44485521707  + 110.20632121940 *T)
    sumR4  +=       0.612 * cos(1.59320941864  + 3.18139373770 *T)
    sumR4  +=       0.631 * cos(1.83863158533  + 10.29494073850 *T)
    sumR4  +=       0.491 * cos(1.52912023181  + 405.25754987360 *T)
    sumR4  +=       0.521 * cos(0.24011424451  + 433.71173787680 *T)

    # JUPITER Radius
    sumR5  = 0
    sumR5  +=      11.188 * cos(4.75249399945  + 536.80451209540 *T)
    sumR5  +=       4.255 * cos(5.91516229170  + 522.57741809380 *T)
    sumR5  +=       2.079 * cos(5.56781555864  + 515.46387109300 *T)
    sumR5  +=       1.908 * cos(4.29659647286  + 543.91805909620 *T)
    sumR5  +=       1.875 * cos(3.69357495838  + 7.11354700080 *T)
    sumR5  +=       1.590 * cos(5.49312796166  + 1066.49547719000 *T)
    sumR5  +=       1.612 * cos(4.13222808529  + 1059.38193018920 *T)
    sumR5  +=       1.240 * cos(3.77981722506  + 14.22709400160 *T)
    sumR5  +=       1.033 * cos(4.50671820436  + 529.69096509460 *T)

    # calculate and scale the sums
    L = sumL0 + sumL1*T**1 + sumL2*T**2 + sumL3*T**3 + sumL4*T**4 + sumL5*T**5
    L /= 1e8
    L = Mod(degrees(L), 360.)
    B = sumB0 + sumB1*T**1 + sumB2*T**2 + sumB3*T**3 + sumB4*T**4 + sumB5*T**5
    B /= 1e8
    B = degrees(B)
    R = sumR0 + sumR1*T**1 + sumR2*T**2 + sumR3*T**3 + sumR4*T**4 + sumR5*T**5
    R /= 1e8

    # calculate X, Y, Z
    X = R * Cos(B) * Cos(L)
    Y = R * Cos(B) * Sin(L)
    Z = R * Sin(B)

    # create dictionary
    Dict = {  "Object"  : "Jupiter",
              "Mode"    : "LBR",
              "Long"    : L,
              "Lat"     : B,
              "Rau"     : R,
              "X"       : X,
              "Y"       : Y,
              "Z"       : Z,
              "Alt"     : -34./60.
           }
    Data = (L, B, R, Dict)


    return Data
#######################################################################



def getJupiterData(dtCivicTime, zone, verbose = False):
    L, B, R, Dict = getJupiterLBRData(dtCivicTime, zone, verbose)
    X = Dict["X"]; Y = Dict["Y"]; Z = Dict["Z"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Le, Be, Re, Dicte = getEarthLBRData(dtCivicTime, zone, verbose)
        Xe = Dicte["X"]; Ye = Dicte["Y"]; Ze = Dicte["Z"]
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getMarsLBRData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Mars"
#       Long  :  heliocentric Longitude, degrees
#       Lat   :  heliocentric Latitude,degrees
#       Rau   :  Radius, distaance Sun - Object (au)
#
#   Similarly for each planet
#   data from Meeus, Appendix II and VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.2 hvf 2023-04-06 getMarsLBRData.py created by mkPlanetLBRData.py"
VersionList.append(Version)

def getMarsLBRData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. (31.1) Julian Millennia

#######################################################################

    # MARS Longitude
    sumL0  = 0
    sumL0  +=  620347711.583 * cos(0.00000000000  + 0.00000000000 *T)
    sumL0  +=  18656368.100 * cos(5.05037100303  + 3340.61242669980 *T)
    sumL0  +=  1108216.792 * cos(5.40099836958  + 6681.22485339960 *T)
    sumL0  +=   91798.394 * cos(5.75478745111  + 10021.83728009940 *T)
    sumL0  +=   27744.987 * cos(5.97049512942  + 3.52311834900 *T)
    sumL0  +=   10610.230 * cos(2.93958524973  + 2281.23049651060 *T)
    sumL0  +=   12315.897 * cos(0.84956081238  + 2810.92146160520 *T)
    sumL0  +=    8926.772 * cos(4.15697845939  + 0.01725365220 *T)
    sumL0  +=    8715.688 * cos(6.11005159792  + 13362.44970679920 *T)
    sumL0  +=    6797.552 * cos(0.36462243626  + 398.14900340820 *T)
    sumL0  +=    7774.867 * cos(3.33968655074  + 5621.84292321040 *T)
    sumL0  +=    3575.079 * cos(1.66186540141  + 2544.31441988340 *T)
    sumL0  +=    4161.101 * cos(0.22814975330  + 2942.46342329160 *T)
    sumL0  +=    3075.250 * cos(0.85696597082  + 191.44826611160 *T)
    sumL0  +=    2628.122 * cos(0.64806143570  + 3337.08930835080 *T)
    sumL0  +=    2937.543 * cos(6.07893711408  + 0.06731030280 *T)
    sumL0  +=    2389.420 * cos(5.03896401349  + 796.29800681640 *T)
    sumL0  +=    2579.842 * cos(0.02996706197  + 3344.13554504880 *T)
    sumL0  +=    1528.140 * cos(1.14979306228  + 6151.53388830500 *T)
    sumL0  +=    1798.808 * cos(0.65634026844  + 529.69096509460 *T)
    sumL0  +=    1264.356 * cos(3.62275092231  + 5092.15195811580 *T)
    sumL0  +=    1286.232 * cos(3.06795924626  + 2146.16541647520 *T)
    sumL0  +=    1546.408 * cos(2.91579633392  + 1751.53953141600 *T)
    sumL0  +=    1024.907 * cos(3.69334293555  + 8962.45534991020 *T)
    sumL0  +=     891.567 * cos(0.18293899090  + 16703.06213349900 *T)
    sumL0  +=     858.760 * cos(2.40093704204  + 2914.01423582380 *T)
    sumL0  +=     832.718 * cos(2.46418591282  + 3340.59517304760 *T)
    sumL0  +=     832.724 * cos(4.49495753458  + 3340.62968035200 *T)
    sumL0  +=     712.899 * cos(3.66336014788  + 1059.38193018920 *T)
    sumL0  +=     748.724 * cos(3.82248399468  + 155.42039943420 *T)
    sumL0  +=     723.863 * cos(0.67497565801  + 3738.76143010800 *T)
    sumL0  +=     635.557 * cos(2.92182704275  + 8432.76438481560 *T)
    sumL0  +=     655.163 * cos(0.48864075176  + 3127.31333126180 *T)
    sumL0  +=     550.472 * cos(3.81001205408  + 0.98032106820 *T)
    sumL0  +=     552.746 * cos(4.47478863016  + 1748.01641306700 *T)
    sumL0  +=     425.972 * cos(0.55365138172  + 6283.07584999140 *T)
    sumL0  +=     415.132 * cos(0.49662314774  + 213.29909543800 *T)
    sumL0  +=     472.164 * cos(3.62547819410  + 1194.44701022460 *T)
    sumL0  +=     306.552 * cos(0.38052862973  + 6684.74797174860 *T)
    sumL0  +=     312.141 * cos(0.99853322843  + 6677.70173505060 *T)
    sumL0  +=     293.199 * cos(4.22131277914  + 20.77539549240 *T)
    sumL0  +=     302.377 * cos(4.48618150321  + 3532.06069281140 *T)
    sumL0  +=     274.028 * cos(0.54222141841  + 3340.54511639700 *T)
    sumL0  +=     281.073 * cos(5.88163372945  + 1349.86740965880 *T)
    sumL0  +=     231.185 * cos(1.28240685294  + 3870.30339179440 *T)
    sumL0  +=     283.600 * cos(5.76885494123  + 3149.16416058820 *T)
    sumL0  +=     236.114 * cos(5.75504515576  + 3333.49887969900 *T)
    sumL0  +=     274.035 * cos(0.13372501211  + 3340.67973700260 *T)
    sumL0  +=     299.396 * cos(2.78323705697  + 6254.62666252360 *T)
    sumL0  +=     204.161 * cos(2.82133266185  + 1221.84856632140 *T)
    sumL0  +=     238.857 * cos(5.37155471672  + 4136.91043351620 *T)
    sumL0  +=     188.639 * cos(1.49103016486  + 9492.14631500480 *T)
    sumL0  +=     221.225 * cos(3.50466672203  + 382.89653222320 *T)
    sumL0  +=     179.196 * cos(1.00561112574  + 951.71840625060 *T)
    sumL0  +=     172.110 * cos(0.43943041719  + 5486.77784317500 *T)
    sumL0  +=     193.126 * cos(3.35715137745  + 3.59042865180 *T)
    sumL0  +=     144.305 * cos(1.41874193418  + 135.06508003540 *T)
    sumL0  +=     160.011 * cos(3.94854735192  + 4562.46099302120 *T)
    sumL0  +=     174.068 * cos(2.41360332576  + 553.56940284240 *T)
    sumL0  +=     130.993 * cos(4.04491720264  + 12303.06777661000 *T)
    sumL0  +=     138.245 * cos(4.30145176915  + 7.11354700080 *T)
    sumL0  +=     128.062 * cos(1.80665643332  + 5088.62883976680 *T)
    sumL0  +=     139.897 * cos(3.32592516164  + 2700.71514038580 *T)
    sumL0  +=     128.102 * cos(2.20806651008  + 1592.59601363280 *T)
    sumL0  +=     116.945 * cos(3.12805282207  + 7903.07341972100 *T)
    sumL0  +=     110.375 * cos(1.05195079687  + 242.72860397400 *T)
    sumL0  +=     113.486 * cos(3.70070798123  + 1589.07289528380 *T)
    sumL0  +=     100.090 * cos(3.24343740861  + 11773.37681151540 *T)
    sumL0  +=      95.592 * cos(0.53954181149  + 20043.67456019880 *T)
    sumL0  +=      98.947 * cos(4.84558294740  + 6681.24210705180 *T)
    sumL0  +=     104.541 * cos(0.78535382076  + 8827.39026987480 *T)
    sumL0  +=      84.187 * cos(3.98970720730  + 4399.99435688900 *T)
    sumL0  +=      86.931 * cos(2.20186740523  + 11243.68584642080 *T)
    sumL0  +=      71.437 * cos(2.80307550016  + 3185.19202726560 *T)
    sumL0  +=      72.091 * cos(5.84672102525  + 5884.92684658320 *T)
    sumL0  +=      73.476 * cos(2.18428012567  + 8429.24126646660 *T)
    sumL0  +=      98.946 * cos(2.81481140371  + 6681.20759974740 *T)
    sumL0  +=      68.414 * cos(2.73834914412  + 2288.34404351140 *T)
    sumL0  +=      86.751 * cos(1.02092221563  + 7079.37385680780 *T)
    sumL0  +=      65.320 * cos(2.68118597578  + 28.44918746780 *T)
    sumL0  +=      83.749 * cos(3.20256130990  + 4690.47983635860 *T)
    sumL0  +=      75.034 * cos(0.76643418252  + 6467.92575796160 *T)
    sumL0  +=      68.984 * cos(3.76399731788  + 6041.32756708560 *T)
    sumL0  +=      66.706 * cos(0.73630620766  + 3723.50895892300 *T)
    sumL0  +=      63.314 * cos(4.52771470470  + 426.59819087600 *T)
    sumL0  +=      61.683 * cos(6.16831509419  + 2274.11694950980 *T)
    sumL0  +=      52.256 * cos(0.89941531307  + 9623.68827669120 *T)
    sumL0  +=      55.488 * cos(4.60625467020  + 4292.33083295040 *T)
    sumL0  +=      51.332 * cos(4.14823636534  + 3341.59274776800 *T)
    sumL0  +=      56.629 * cos(5.06250410206  + 15.25247118500 *T)
    sumL0  +=      63.376 * cos(0.91296240798  + 3553.91152213780 *T)
    sumL0  +=      45.829 * cos(0.78784235062  + 1990.74501704100 *T)
    sumL0  +=      48.542 * cos(3.95670418719  + 4535.05943692440 *T)
    sumL0  +=      41.223 * cos(6.02019329922  + 3894.18182954220 *T)
    sumL0  +=      41.939 * cos(3.58326425115  + 8031.09226305840 *T)
    sumL0  +=      56.396 * cos(1.68727150304  + 6872.67311951120 *T)
    sumL0  +=      55.909 * cos(3.46260833495  + 263.08392337280 *T)
    sumL0  +=      51.678 * cos(2.81307492682  + 3339.63210563160 *T)
    sumL0  +=      40.671 * cos(3.13832621829  + 9595.23908922340 *T)
    sumL0  +=      38.107 * cos(0.73401946320  + 10025.36039844840 *T)
    sumL0  +=      39.495 * cos(5.63225392160  + 3097.88382272579 *T)
    sumL0  +=      44.174 * cos(3.19529736702  + 5628.95647021120 *T)
    sumL0  +=      36.716 * cos(2.63720775102  + 692.15760122680 *T)
    sumL0  +=      45.905 * cos(0.28718981497  + 5614.72937620960 *T)
    sumL0  +=      38.352 * cos(5.82880707426  + 3191.04922956520 *T)
    sumL0  +=      38.206 * cos(2.34835984063  + 162.46663613220 *T)
    sumL0  +=      32.562 * cos(0.48400659333  + 6681.29216370240 *T)
    sumL0  +=      37.135 * cos(0.68508150774  + 2818.03500860600 *T)
    sumL0  +=      31.168 * cos(3.98160912982  + 20.35531939880 *T)
    sumL0  +=      32.561 * cos(0.89250316888  + 6681.15754309680 *T)
    sumL0  +=      37.752 * cos(4.15482955299  + 2803.80791460440 *T)
    sumL0  +=      33.626 * cos(6.11992401052  + 6489.77658728800 *T)
    sumL0  +=      29.007 * cos(2.42707385674  + 3319.83703120740 *T)
    sumL0  +=      38.790 * cos(1.35198498795  + 10018.31416175040 *T)
    sumL0  +=      33.148 * cos(1.14023770004  + 5.52292430740 *T)
    sumL0  +=      27.584 * cos(1.59691203058  + 7210.91581849420 *T)
    sumL0  +=      28.686 * cos(5.72055456734  + 7477.52286021600 *T)
    sumL0  +=      34.031 * cos(2.59544082509  + 11769.85369316640 *T)
    sumL0  +=      25.380 * cos(0.52093116112  + 10.63666534980 *T)
    sumL0  +=      26.357 * cos(1.34532646574  + 3496.03282613400 *T)
    sumL0  +=      24.554 * cos(4.00323183088  + 11371.70468975820 *T)
    sumL0  +=      25.637 * cos(0.24963523420  + 522.57741809380 *T)
    sumL0  +=      27.278 * cos(4.55645328122  + 3361.38782219220 *T)
    sumL0  +=      23.764 * cos(1.84058377256  + 12832.75874170460 *T)
    sumL0  +=      22.816 * cos(3.52628212106  + 1648.44675719740 *T)
    sumL0  +=      22.274 * cos(0.72106133721  + 266.60704172180 *T)
    sumL0  +=      21.202 * cos(3.11824472284  + 2957.71589447660 *T)
    sumL0  +=      20.158 * cos(3.67131504946  + 1758.65307841680 *T)
    sumL0  +=      21.530 * cos(6.15388757177  + 3264.34635542420 *T)
    sumL0  +=      20.093 * cos(1.08247416065  + 7064.12138562280 *T)
    sumL0  +=      21.343 * cos(4.28218757863  + 4032.77002792660 *T)
    sumL0  +=      27.540 * cos(6.08389942337  + 6674.11130639880 *T)
    sumL0  +=      19.849 * cos(2.37668920745  + 10713.99488132620 *T)
    sumL0  +=      25.512 * cos(3.43242352804  + 3443.70520091840 *T)
    sumL0  +=      22.542 * cos(5.64861703438  + 2388.89402044920 *T)
    sumL0  +=      24.378 * cos(0.96994696413  + 632.78373931320 *T)
    sumL0  +=      23.079 * cos(4.74990214223  + 3347.72597370060 *T)
    sumL0  +=      17.709 * cos(3.69742343974  + 3344.20285535160 *T)
    sumL0  +=      22.662 * cos(3.95446324417  + 4989.05918389720 *T)
    sumL0  +=      22.604 * cos(5.24082917494  + 3205.54734666440 *T)
    sumL0  +=      16.811 * cos(5.48619684111  + 3.88133535800 *T)
    sumL0  +=      18.422 * cos(4.22535881468  + 2787.04302385740 *T)
    sumL0  +=      22.737 * cos(4.98520896596  + 7632.94325965020 *T)
    sumL0  +=      16.648 * cos(2.52823633184  + 14584.29827312060 *T)
    sumL0  +=      20.963 * cos(4.27878216453  + 5099.26550511660 *T)
    sumL0  +=      16.042 * cos(1.76786752521  + 3475.67750673520 *T)
    sumL0  +=      15.816 * cos(3.13240869691  + 59.37386191360 *T)
    sumL0  +=      18.113 * cos(3.25756020453  + 3337.02199804800 *T)
    sumL0  +=      19.295 * cos(3.23911854642  + 7.04623669800 *T)
    sumL0  +=      16.772 * cos(4.39731507110  + 15643.68020330980 *T)
    sumL0  +=      17.555 * cos(4.09197396097  + 74.78159856730 *T)
    sumL0  +=      13.704 * cos(2.54117018160  + 4933.20844033260 *T)
    sumL0  +=      16.011 * cos(1.54669633224  + 14054.60730802600 *T)
    sumL0  +=      13.547 * cos(4.04152185347  + 4929.68532198360 *T)
    sumL0  +=      14.566 * cos(3.45210993051  + 7373.38245462640 *T)
    sumL0  +=      13.926 * cos(5.40797129468  + 10973.55568635000 *T)
    sumL0  +=      14.246 * cos(0.59808746067  + 23.87843774780 *T)
    sumL0  +=      14.023 * cos(1.44218648988  + 10404.73381232260 *T)
    sumL0  +=      16.051 * cos(3.79409950488  + 2118.76386037840 *T)
    sumL0  +=      13.714 * cos(3.59050634457  + 15113.98923821520 *T)
    sumL0  +=      18.038 * cos(4.25391532000  + 2487.41604494780 *T)
    sumL0  +=      15.846 * cos(0.56901288692  + 103.09277421860 *T)
    sumL0  +=      13.403 * cos(5.16920432994  + 10213.28554621100 *T)
    sumL0  +=      16.069 * cos(2.36895958451  + 3265.83082813250 *T)
    sumL0  +=      12.773 * cos(0.10483085657  + 7234.79425624200 *T)
    sumL0  +=      12.199 * cos(1.73079687044  + 36.02786667740 *T)
    sumL0  +=      12.283 * cos(5.19940030635  + 10021.85453375160 *T)
    sumL0  +=      11.945 * cos(5.47997890162  + 2921.12778282460 *T)
    sumL0  +=      11.890 * cos(4.76593905835  + 5828.02847164760 *T)
    sumL0  +=      12.283 * cos(3.16862882612  + 10021.82002644720 *T)
    sumL0  +=      13.274 * cos(6.17806905340  + 1744.42598441520 *T)
    sumL0  +=      11.777 * cos(5.72731550900  + 0.42007609361 *T)
    sumL0  +=      12.340 * cos(2.52146766331  + 2906.90068882300 *T)
    sumL0  +=      14.458 * cos(4.38010658432  + 316.39186965660 *T)
    sumL0  +=      10.639 * cos(3.45016942919  + 639.89728631400 *T)
    sumL0  +=      10.925 * cos(0.60397688999  + 5085.03841111500 *T)
    sumL0  +=      10.645 * cos(5.47696728127  + 419.48464387520 *T)
    sumL0  +=      10.797 * cos(1.37191539718  + 10419.98628350760 *T)
    sumL0  +=      10.565 * cos(1.09061610786  + 12168.00269657460 *T)
    sumL0  +=       9.578 * cos(4.89489266197  + 3230.40610548040 *T)
    sumL0  +=       9.808 * cos(5.83632873998  + 14314.16811304980 *T)
    sumL0  +=       9.720 * cos(6.28282606427  + 9225.53927328300 *T)
    sumL0  +=       9.146 * cos(1.10220667397  + 9808.53818466140 *T)
    sumL0  +=      12.733 * cos(1.79883375851  + 13745.34623902240 *T)
    sumL0  +=       9.779 * cos(3.60056884868  + 206.18554843720 *T)
    sumL0  +=      12.156 * cos(4.42295240981  + 14712.31711645800 *T)
    sumL0  +=       8.801 * cos(3.97218590685  + 170.67287061920 *T)
    sumL0  +=      10.685 * cos(4.33894776374  + 7740.60678358880 *T)
    sumL0  +=      10.041 * cos(1.38294666830  + 3583.34103067380 *T)
    sumL0  +=       8.499 * cos(4.29275471153  + 0.42988312670 *T)
    sumL0  +=       9.882 * cos(2.69148019691  + 36.60536530420 *T)
    sumL0  +=      10.585 * cos(0.89641284928  + 23384.28698689860 *T)
    sumL0  +=       8.478 * cos(2.86886131803  + 9381.93999378540 *T)
    sumL0  +=       9.560 * cos(4.33382353581  + 131.54196168640 *T)
    sumL0  +=       8.433 * cos(3.15239733970  + 6525.80445396540 *T)
    sumL0  +=       7.517 * cos(1.24476310635  + 6894.52394883760 *T)
    sumL0  +=       6.753 * cos(1.58869332894  + 6836.64525283380 *T)
    sumL0  +=       6.666 * cos(4.38910103043  + 66.48740891440 *T)
    sumL0  +=       6.697 * cos(5.77315870484  + 5202.35827933520 *T)
    sumL0  +=       6.188 * cos(1.57847967850  + 3325.35995551480 *T)
    sumL0  +=       8.392 * cos(2.90734956761  + 43.71891230500 *T)
    sumL0  +=       6.427 * cos(6.03333185624  + 574.34479833480 *T)
    sumL0  +=       6.296 * cos(0.06976040737  + 1964.83862685400 *T)
    sumL0  +=       6.144 * cos(5.43713363624  + 1861.74585263540 *T)
    sumL0  +=       8.180 * cos(0.43967386003  + 2707.82868738660 *T)
    sumL0  +=       6.532 * cos(1.24907069366  + 12964.30070339100 *T)
    sumL0  +=       7.341 * cos(4.95693808975  + 3767.21061757580 *T)
    sumL0  +=       6.134 * cos(3.16322441707  + 6680.24453233140 *T)
    sumL0  +=       5.937 * cos(4.22752447794  + 4459.36821880260 *T)
    sumL0  +=       5.848 * cos(4.30379700916  + 7875.67186362420 *T)
    sumL0  +=       5.895 * cos(0.77226207039  + 2699.73481931760 *T)
    sumL0  +=       5.962 * cos(3.01318151415  + 3369.06161416760 *T)
    sumL0  +=       6.307 * cos(1.77830983983  + 2178.13772229200 *T)
    sumL0  +=       6.103 * cos(4.49866000852  + 6682.20517446780 *T)
    sumL0  +=       5.770 * cos(0.96127853878  + 13916.01910964160 *T)
    sumL0  +=       5.725 * cos(4.93042706875  + 2384.32327072920 *T)
    sumL0  +=       7.576 * cos(6.16308742022  + 6531.66165626500 *T)
    sumL0  +=       7.285 * cos(0.52756973131  + 10575.40668294180 *T)
    sumL0  +=       5.212 * cos(3.48999909365  + 12935.85151592320 *T)
    sumL0  +=       5.266 * cos(4.51201042373  + 6144.42034130420 *T)
    sumL0  +=       6.632 * cos(3.48100908925  + 1118.75579210280 *T)
    sumL0  +=       5.182 * cos(3.25459432228  + 2391.43681773000 *T)
    sumL0  +=       5.214 * cos(0.01023839897  + 533.21408344360 *T)
    sumL0  +=       5.436 * cos(6.18510045571  + 8425.65083781480 *T)
    sumL0  +=       5.491 * cos(3.42235890731  + 3134.42687826260 *T)
    sumL0  +=       5.773 * cos(3.55190490896  + 8969.56889691100 *T)
    sumL0  +=       5.093 * cos(0.60873962999  + 8955.34180290940 *T)
    sumL0  +=       4.798 * cos(4.63144694279  + 4569.57454002200 *T)
    sumL0  +=       5.773 * cos(5.92316959013  + 640.87760738220 *T)
    sumL0  +=       5.661 * cos(0.75892958537  + 3120.19978426100 *T)
    sumL0  +=       4.912 * cos(1.70762118141  + 13358.92658845020 *T)
    sumL0  +=       5.438 * cos(4.29256867866  + 3503.07906283200 *T)
    sumL0  +=       4.631 * cos(3.34058594066  + 3116.26763099790 *T)
    sumL0  +=       5.823 * cos(2.39009621568  + 3302.47939106200 *T)
    sumL0  +=       5.095 * cos(1.12497131307  + 5331.35744374080 *T)
    sumL0  +=       4.409 * cos(1.38708732950  + 17256.63153634140 *T)
    sumL0  +=       4.193 * cos(0.19116539271  + 9830.38901398780 *T)
    sumL0  +=       4.164 * cos(0.43798301495  + 1066.49547719000 *T)
    sumL0  +=       4.220 * cos(1.69427060019  + 13524.91634293140 *T)
    sumL0  +=       5.192 * cos(3.63500646439  + 536.80451209540 *T)
    sumL0  +=       5.664 * cos(0.50446281443  + 5305.45105355380 *T)
    sumL0  +=       5.264 * cos(5.40031180025  + 3355.86489788480 *T)
    sumL0  +=       5.403 * cos(5.96867227489  + 3074.00538497800 *T)
    sumL0  +=       4.046 * cos(0.83826342101  + 10021.90459040220 *T)
    sumL0  +=       4.153 * cos(3.14497570817  + 8439.87793181640 *T)
    sumL0  +=       4.820 * cos(1.09025983917  + 13365.97282514820 *T)
    sumL0  +=       4.906 * cos(3.73078406569  + 1228.96211332220 *T)
    sumL0  +=       4.576 * cos(0.99344843456  + 6158.64743530580 *T)
    sumL0  +=       4.211 * cos(3.87091723842  + 3312.16323923200 *T)
    sumL0  +=       4.917 * cos(6.25051681717  + 17654.78053974960 *T)
    sumL0  +=       3.597 * cos(6.07298181151  + 10818.13528691580 *T)
    sumL0  +=       3.547 * cos(1.87663674277  + 17395.21973472580 *T)
    sumL0  +=       3.734 * cos(1.37011069213  + 3973.39616601300 *T)
    sumL0  +=       4.422 * cos(2.89974680697  + 6247.51311552280 *T)
    sumL0  +=       3.975 * cos(4.03441621102  + 1052.26838318840 *T)
    sumL0  +=       4.466 * cos(4.59188422424  + 5518.75014899180 *T)
    sumL0  +=       3.445 * cos(4.26610076208  + 3178.14579056760 *T)
    sumL0  +=       3.607 * cos(4.24527056399  + 8329.67161059700 *T)
    sumL0  +=       3.511 * cos(1.95133364438  + 10177.25767953360 *T)
    sumL0  +=       3.300 * cos(0.68865737747  + 149.56319713460 *T)
    sumL0  +=       3.235 * cos(3.90171358031  + 27.40155609680 *T)
    sumL0  +=       3.946 * cos(5.81982010903  + 6261.74020952440 *T)
    sumL0  +=       3.387 * cos(0.14377701728  + 10014.72373309860 *T)
    sumL0  +=       3.374 * cos(0.74722238154  + 6048.44111408640 *T)
    sumL0  +=       3.090 * cos(3.98540766861  + 2648.45482547300 *T)
    sumL0  +=       3.821 * cos(5.23987859516  + 5724.93569742900 *T)
    sumL0  +=       3.026 * cos(0.34314724795  + 220.41264243880 *T)
    sumL0  +=       4.052 * cos(1.24671617482  + 10021.76996979660 *T)
    sumL0  +=       2.937 * cos(0.73100893561  + 2.75151061100 *T)
    sumL0  +=       3.084 * cos(3.79925632319  + 169.58018313300 *T)
    sumL0  +=       4.040 * cos(2.91258200196  + 22747.29071487440 *T)
    sumL0  +=       3.246 * cos(4.90760526525  + 6702.00024889200 *T)
    sumL0  +=       3.223 * cos(0.92640008610  + 16865.52876963120 *T)
    sumL0  +=       2.879 * cos(1.47180601483  + 3346.13535100720 *T)
    sumL0  +=       3.723 * cos(0.49978317761  + 1.48447270830 *T)
    sumL0  +=       2.798 * cos(3.26910698548  + 9168.64089834740 *T)
    sumL0  +=       3.347 * cos(0.68791690253  + 3863.18984479360 *T)
    sumL0  +=       3.440 * cos(2.77706064625  + 6660.44945790720 *T)
    sumL0  +=       2.798 * cos(2.79673379094  + 16858.48253293320 *T)
    sumL0  +=       2.753 * cos(4.41190782119  + 3415.39402526710 *T)
    sumL0  +=       2.703 * cos(0.19222683184  + 3237.51965248120 *T)
    sumL0  +=       2.719 * cos(3.26628341377  + 3914.95722503460 *T)
    sumL0  +=       2.941 * cos(3.76169133143  + 6784.31762761820 *T)
    sumL0  +=       3.078 * cos(5.48104322207  + 3335.08950239240 *T)
    sumL0  +=       2.701 * cos(5.08537226206  + 6688.33840040040 *T)
    sumL0  +=       2.606 * cos(4.83032556909  + 4672.66731424060 *T)
    sumL0  +=       2.883 * cos(2.64068086362  + 3320.25710730100 *T)
    sumL0  +=       3.089 * cos(4.39619824300  + 1332.05488754080 *T)
    sumL0  +=       3.438 * cos(1.93467064449  + 10551.52824519400 *T)
    sumL0  +=       3.433 * cos(0.20933510815  + 6604.95878212400 *T)
    sumL0  +=       2.513 * cos(2.88151502604  + 17924.91069982040 *T)
    sumL0  +=       2.683 * cos(3.68680362407  + 3603.69635007260 *T)
    sumL0  +=       2.696 * cos(2.66931776730  + 10184.30391623160 *T)
    sumL0  +=       2.386 * cos(1.05710815775  + 3607.21946842160 *T)
    sumL0  +=       2.444 * cos(3.46901444423  + 6298.32832117640 *T)
    sumL0  +=       2.566 * cos(5.55941956468  + 6546.15977336420 *T)
    sumL0  +=       2.354 * cos(0.85896789940  + 3351.24909204960 *T)
    sumL0  +=       2.304 * cos(6.07210985790  + 1214.73501932060 *T)
    sumL0  +=       2.387 * cos(4.30419979225  + 3360.96774609859 *T)
    sumL0  +=       3.109 * cos(2.18099805904  + 16173.37116840440 *T)
    sumL0  +=       2.907 * cos(3.43894993113  + 2693.60159338500 *T)
    sumL0  +=       2.550 * cos(4.18354174372  + 3546.79797513700 *T)
    sumL0  +=       2.991 * cos(2.37500894603  + 13517.87010623340 *T)
    sumL0  +=       2.508 * cos(5.99249607455  + 5729.50644714900 *T)
    sumL0  +=       2.202 * cos(5.58486973955  + 664.75604513000 *T)
    sumL0  +=       2.467 * cos(2.81052575497  + 15110.46611986620 *T)
    sumL0  +=       2.154 * cos(3.59696831702  + 6677.63442474780 *T)
    sumL0  +=       2.166 * cos(3.35965682842  + 589.06482700820 *T)
    sumL0  +=       2.112 * cos(4.57048853017  + 6127.65545055720 *T)
    sumL0  +=       2.178 * cos(3.21663279493  + 20199.09495963300 *T)
    sumL0  +=       2.260 * cos(3.62776666288  + 7799.98064550240 *T)
    sumL0  +=       2.703 * cos(5.89441421026  + 6438.49624942560 *T)
    sumL0  +=       2.201 * cos(4.69972051344  + 17277.40693183380 *T)
    sumL0  +=       2.131 * cos(2.51801264340  + 1545.35398297880 *T)
    sumL0  +=       2.093 * cos(4.03688419630  + 6684.81528205140 *T)
    sumL0  +=       2.249 * cos(5.26171743929  + 5618.31980486140 *T)
    sumL0  +=       2.117 * cos(4.80404492675  + 3657.00429635640 *T)
    sumL0  +=       2.265 * cos(3.87401620754  + 110.20632121940 *T)
    sumL0  +=       2.521 * cos(4.21237950687  + 2494.52959194860 *T)
    sumL0  +=       2.492 * cos(6.10452779411  + 3329.97576135000 *T)
    sumL0  +=       2.154 * cos(4.59093274516  + 5625.36604155940 *T)
    sumL0  +=       1.929 * cos(1.21920197307  + 21.85082932640 *T)
    sumL0  +=       2.343 * cos(0.90613584736  + 227.47613278900 *T)
    sumL0  +=       1.871 * cos(2.03707390080  + 56.89837493560 *T)
    sumL0  +=       1.894 * cos(4.12432529517  + 3399.98628861340 *T)
    sumL0  +=       2.271 * cos(3.78951734652  + 7910.18696672180 *T)
    sumL0  +=       1.876 * cos(3.02034887080  + 4885.96640967860 *T)
    sumL0  +=       2.145 * cos(4.74972617629  + 18984.29263000960 *T)
    sumL0  +=       2.035 * cos(2.75110234296  + 128.01884333740 *T)
    sumL0  +=       1.770 * cos(2.73496011079  + 6606.44325483230 *T)
    sumL0  +=       1.733 * cos(1.70571779606  + 6.68366387410 *T)
    sumL0  +=       1.743 * cos(2.22986352012  + 17085.95866572220 *T)
    sumL0  +=       1.723 * cos(1.98207274526  + 735.87651353180 *T)
    sumL0  +=       1.814 * cos(0.92405242552  + 4555.34744602040 *T)
    sumL0  +=       1.661 * cos(3.07319305915  + 1692.16566950240 *T)
    sumL0  +=       2.332 * cos(5.05223613332  + 20618.01935853360 *T)
    sumL0  +=       1.653 * cos(3.86316179052  + 699.27114822760 *T)
    sumL0  +=       1.890 * cos(4.14080522607  + 5459.37628707820 *T)
    sumL0  +=       1.680 * cos(3.49750704538  + 3347.65866339780 *T)
    sumL0  +=       1.909 * cos(2.11478314309  + 6816.28993343500 *T)
    sumL0  +=       1.918 * cos(3.31232891649  + 3333.56619000180 *T)
    sumL0  +=       1.707 * cos(4.88386665097  + 3407.09983561420 *T)
    sumL0  +=       1.577 * cos(3.51777131580  + 13362.43245314700 *T)
    sumL0  +=       1.827 * cos(1.14937084769  + 2807.39834325620 *T)
    sumL0  +=       2.094 * cos(1.78938870686  + 20597.24396304120 *T)
    sumL0  +=       1.577 * cos(5.54854302324  + 13362.46696045140 *T)
    sumL0  +=       1.677 * cos(4.39073937265  + 8270.29774868340 *T)
    sumL0  +=       1.561 * cos(0.80246904730  + 3017.10701004240 *T)
    sumL0  +=       1.552 * cos(2.66876850182  + 3024.22055704320 *T)
    sumL0  +=       1.708 * cos(2.09721350898  + 2814.44457995420 *T)
    sumL0  +=       2.109 * cos(1.48733908496  + 2679.37949991880 *T)
    sumL0  +=       1.547 * cos(2.12956399169  + 661.23292678100 *T)
    sumL0  +=       1.764 * cos(5.90517079295  + 3326.38533269820 *T)
    sumL0  +=       1.503 * cos(4.34389183170  + 13936.79450513400 *T)
    sumL0  +=       1.504 * cos(0.83367652471  + 4775.76008845920 *T)
    sumL0  +=       1.543 * cos(5.82417982389  + 3344.49376205780 *T)
    sumL0  +=       1.458 * cos(1.42630589706  + 15508.61512327440 *T)
    sumL0  +=       1.648 * cos(2.88801518612  + 8273.82086703240 *T)
    sumL0  +=       1.527 * cos(1.10261249454  + 2675.85638156980 *T)
    sumL0  +=       1.800 * cos(5.17144672709  + 38.13303563780 *T)
    sumL0  +=       1.439 * cos(0.48751113425  + 76.26607127560 *T)
    sumL0  +=       1.478 * cos(4.63008666184  + 19800.94595622480 *T)
    sumL0  +=       1.770 * cos(2.18315009072  + 2301.58581590939 *T)
    sumL0  +=       1.712 * cos(2.60291779744  + 29.42950853600 *T)
    sumL0  +=       1.416 * cos(2.46912016889  + 2597.62236616720 *T)
    sumL0  +=       1.835 * cos(3.93223068469  + 6843.69148953180 *T)
    sumL0  +=       1.437 * cos(3.48241890037  + 3281.23856478620 *T)
    sumL0  +=       1.385 * cos(4.18004525683  + 2284.75361485960 *T)
    sumL0  +=       1.479 * cos(3.88994194685  + 6034.21402008480 *T)
    sumL0  +=       1.469 * cos(2.91808856545  + 12722.55242048520 *T)
    sumL0  +=       1.824 * cos(3.94139541752  + 18454.60166491500 *T)
    sumL0  +=       1.368 * cos(1.71899681607  + 13760.59871020740 *T)
    sumL0  +=       1.406 * cos(2.42916192473  + 853.19638175200 *T)
    sumL0  +=       1.361 * cos(4.69126219798  + 11081.21921028860 *T)
    sumL0  +=       1.375 * cos(4.90787845983  + 3304.58456002240 *T)
    sumL0  +=       1.544 * cos(3.23251603238  + 156.40072050240 *T)
    sumL0  +=       1.543 * cos(1.13497136095  + 3336.73109134180 *T)
    sumL0  +=       1.385 * cos(2.90281983822  + 1581.95934828300 *T)
    sumL0  +=       1.311 * cos(6.21748423079  + 2547.83753823240 *T)
    sumL0  +=       1.377 * cos(2.56537541792  + 568.82187402740 *T)
    sumL0  +=       1.474 * cos(4.65215247212  + 394.62588505920 *T)
    sumL0  +=       1.619 * cos(0.07481583409  + 1435.14766175940 *T)
    sumL0  +=       1.369 * cos(0.08979359617  + 7895.95987272020 *T)
    sumL0  +=       1.272 * cos(5.84659079053  + 21.33564046700 *T)
    sumL0  +=       1.331 * cos(3.80007277718  + 13119.72110282519 *T)
    sumL0  +=       1.202 * cos(5.90681878458  + 187.92514776260 *T)
    sumL0  +=       1.284 * cos(5.30205682052  + 6571.01853218020 *T)
    sumL0  +=       1.468 * cos(5.91503888421  + 3339.12795399150 *T)
    sumL0  +=       1.497 * cos(2.57629850836  + 151.89728108520 *T)
    sumL0  +=       1.270 * cos(4.08996241400  + 4356.27544458400 *T)
    sumL0  +=       1.159 * cos(0.00546309207  + 799.82112516540 *T)
    sumL0  +=       1.154 * cos(1.81024985603  + 158.94351778320 *T)
    sumL0  +=       1.170 * cos(4.15567553953  + 14.22709400160 *T)
    sumL0  +=       1.349 * cos(0.39391022745  + 2540.79130153440 *T)
    sumL0  +=       1.403 * cos(4.89956428847  + 4039.88357492740 *T)
    sumL0  +=       1.188 * cos(3.46422365287  + 1015.66301788420 *T)
    sumL0  +=       1.136 * cos(5.54132891775  + 13553.89797291080 *T)
    sumL0  +=       1.282 * cos(4.52288421340  + 3929.67725370800 *T)
    sumL0  +=       1.124 * cos(1.40102701439  + 13149.15061136120 *T)
    sumL0  +=       1.254 * cos(3.84775964741  + 3980.50971301380 *T)
    sumL0  +=       1.121 * cos(5.04552583589  + 23141.55838292460 *T)
    sumL0  +=       1.088 * cos(0.89388884633  + 3340.19235060619 *T)
    sumL0  +=       1.476 * cos(4.20852908107  + 16460.33352952499 *T)
    sumL0  +=       1.084 * cos(5.70551481838  + 3760.09707057500 *T)
    sumL0  +=       1.197 * cos(1.24543578245  + 26724.89941359840 *T)
    sumL0  +=       1.253 * cos(5.82286965998  + 3344.54457996290 *T)
    sumL0  +=       1.078 * cos(6.24742453183  + 17.81252211800 *T)
    sumL0  +=       1.462 * cos(2.27467510211  + 369.69981594040 *T)
    sumL0  +=       1.044 * cos(4.34158344243  + 2277.70737816160 *T)
    sumL0  +=       1.099 * cos(4.08747611117  + 802.36392244620 *T)
    sumL0  +=       1.122 * cos(4.31726854418  + 107.66352393860 *T)
    sumL0  +=       1.148 * cos(0.42666214138  + 949.17560896980 *T)
    sumL0  +=       1.183 * cos(0.31560034120  + 1162.47470440780 *T)
    sumL0  +=       1.130 * cos(3.46061330657  + 5732.04924442980 *T)
    sumL0  +=       1.059 * cos(1.34075999078  + 2149.68853482420 *T)
    sumL0  +=       1.126 * cos(5.22902222299  + 194.97138446060 *T)
    sumL0  +=       0.977 * cos(3.44426658417  + 9779.10867612540 *T)
    sumL0  +=       1.134 * cos(0.13507173479  + 12566.15169998280 *T)
    sumL0  +=       1.185 * cos(3.65682936623  + 6456.88005769770 *T)
    sumL0  +=       0.962 * cos(4.85370284827  + 3510.19260983280 *T)
    sumL0  +=       1.033 * cos(2.47380728164  + 31.97230581680 *T)
    sumL0  +=       1.172 * cos(1.72927931309  + 6923.95345737360 *T)
    sumL0  +=       0.986 * cos(1.05152205629  + 16335.83780453660 *T)
    sumL0  +=       1.292 * cos(6.02388659607  + 3316.73398895200 *T)
    sumL0  +=       0.932 * cos(1.56812178735  + 111.18664228760 *T)
    sumL0  +=       1.130 * cos(2.06879370733  + 3376.64029337720 *T)
    sumL0  +=       0.924 * cos(0.82106390115  + 3364.49086444760 *T)
    sumL0  +=       1.005 * cos(5.48594935121  + 11236.57229942000 *T)
    sumL0  +=       0.936 * cos(1.58837554752  + 10235.13637553740 *T)
    sumL0  +=       1.063 * cos(1.93321644244  + 87.30820453981 *T)
    sumL0  +=       0.945 * cos(5.10393817688  + 8013.27974094040 *T)
    sumL0  +=       0.957 * cos(4.41421752360  + 433.71173787680 *T)
    sumL0  +=       0.921 * cos(1.71021166461  + 2067.93140107260 *T)
    sumL0  +=       0.910 * cos(1.91562889347  + 401.67212175720 *T)
    sumL0  +=       0.931 * cos(4.61327789373  + 5415.65737477320 *T)
    sumL0  +=       1.033 * cos(3.98512631089  + 10596.18207843420 *T)
    sumL0  +=       0.908 * cos(5.30549216988  + 7107.82304427560 *T)
    sumL0  +=       0.945 * cos(0.21610469082  + 9872.27408296480 *T)
    sumL0  +=       0.991 * cos(2.76931703924  + 5938.23479286700 *T)
    sumL0  +=       0.918 * cos(4.37217510741  + 1854.63230563460 *T)
    sumL0  +=       0.889 * cos(1.81025662091  + 2409.24933984800 *T)
    sumL0  +=       1.058 * cos(3.63303046818  + 7255.56965173440 *T)
    sumL0  +=       0.860 * cos(2.78826625605  + 5621.86017686260 *T)
    sumL0  +=       0.860 * cos(0.75749461239  + 5621.82566955820 *T)
    sumL0  +=       0.850 * cos(3.82228424021  + 272.67295735160 *T)
    sumL0  +=       0.851 * cos(0.25287875945  + 11766.26326451460 *T)
    sumL0  +=       0.853 * cos(1.84243320985  + 2142.64229812620 *T)
    sumL0  +=       1.030 * cos(0.42255009123  + 1596.18644228460 *T)
    sumL0  +=       0.826 * cos(1.26069501589  + 20206.14119633100 *T)
    sumL0  +=       0.835 * cos(4.64394754967  + 127.95153303460 *T)
    sumL0  +=       1.046 * cos(3.50232998647  + 9866.41688066520 *T)
    sumL0  +=       0.824 * cos(2.42928138519  + 11.04570026390 *T)
    sumL0  +=       1.090 * cos(6.06536924506  + 3341.03250279340 *T)
    sumL0  +=       0.808 * cos(1.17345081945  + 6460.81221096080 *T)
    sumL0  +=       0.821 * cos(0.19617131185  + 22743.40937951640 *T)
    sumL0  +=       0.804 * cos(5.03144428114  + 2.54279728080 *T)
    sumL0  +=       1.074 * cos(2.80025232939  + 3077.52850332700 *T)
    sumL0  +=       0.794 * cos(4.57067003887  + 5223.69391980220 *T)
    sumL0  +=       0.822 * cos(0.90150471199  + 8646.06348025360 *T)
    sumL0  +=       0.884 * cos(1.71007411939  + 3.93215326310 *T)
    sumL0  +=       0.850 * cos(1.24977903471  + 12295.95422960920 *T)
    sumL0  +=       0.926 * cos(3.02847055739  + 3377.21779200400 *T)
    sumL0  +=       0.779 * cos(1.97930055918  + 5408.54382777240 *T)
    sumL0  +=       0.960 * cos(6.16414205869  + 8542.97070603500 *T)
    sumL0  +=       0.785 * cos(4.63077378309  + 3341.04230982650 *T)
    sumL0  +=       0.752 * cos(3.51125131831  + 10020.85695903120 *T)
    sumL0  +=       0.749 * cos(2.50286751020  + 2295.45759051220 *T)
    sumL0  +=       0.801 * cos(4.54004277501  + 1039.02661079040 *T)
    sumL0  +=       0.735 * cos(3.91703254169  + 12310.18132361080 *T)
    sumL0  +=       0.784 * cos(3.24368310213  + 3384.33133900480 *T)
    sumL0  +=       0.719 * cos(3.35536995801  + 11780.49035851620 *T)
    sumL0  +=       0.940 * cos(3.58974561600  + 23539.70738633280 *T)
    sumL0  +=       0.865 * cos(1.12444569157  + 206.70073729660 *T)
    sumL0  +=       0.731 * cos(5.61232905415  + 16062.18452611680 *T)
    sumL0  +=       0.823 * cos(2.65316808033  + 3169.93955608060 *T)
    sumL0  +=       0.903 * cos(4.86743346013  + 931.36308685180 *T)
    sumL0  +=       0.705 * cos(1.46305394620  + 792.77488846740 *T)
    sumL0  +=       0.754 * cos(2.89691411536  + 3296.89351439480 *T)
    sumL0  +=       0.736 * cos(2.21038016464  + 146.81168652360 *T)
    sumL0  +=       0.785 * cos(2.32844966721  + 3340.18254357310 *T)
    sumL0  +=       0.685 * cos(2.34948834397  + 1.18064265210 *T)
    sumL0  +=       0.677 * cos(4.33803507050  + 3877.41693879520 *T)
    sumL0  +=       0.729 * cos(3.56726385957  + 485.97205278960 *T)
    sumL0  +=       0.668 * cos(5.16798891078  + 12721.57209941700 *T)
    sumL0  +=       0.713 * cos(4.80558699772  + 4142.97634914600 *T)
    sumL0  +=       0.842 * cos(1.59708732155  + 16304.91313009080 *T)
    sumL0  +=       0.698 * cos(1.91829605833  + 6665.97238221460 *T)
    sumL0  +=       0.698 * cos(4.64681188157  + 11216.28429032400 *T)
    sumL0  +=       0.684 * cos(3.88514563357  + 846.08283475120 *T)
    sumL0  +=       0.666 * cos(2.23927960017  + 20735.83216142559 *T)
    sumL0  +=       0.638 * cos(2.05380863176  + 16699.53901514999 *T)
    sumL0  +=       0.726 * cos(5.04566216294  + 3329.56672643590 *T)
    sumL0  +=       0.697 * cos(0.30680314664  + 19676.45023123640 *T)
    sumL0  +=       0.650 * cos(3.23944472757  + 9588.12554222260 *T)
    sumL0  +=       0.736 * cos(2.24443728739  + 11250.79939342160 *T)
    sumL0  +=       0.692 * cos(1.76465916332  + 51.28033786241 *T)
    sumL0  +=       0.645 * cos(0.07661067442  + 52.26031491280 *T)
    sumL0  +=       0.685 * cos(2.72118140264  + 6643.09181776180 *T)
    sumL0  +=       0.684 * cos(1.26157751115  + 142.17862703620 *T)
    sumL0  +=       0.635 * cos(3.85231847116  + 6.06591562980 *T)
    sumL0  +=       0.819 * cos(6.04967230806  + 19402.79695281660 *T)
    sumL0  +=       0.664 * cos(3.60223930010  + 1903.43681250120 *T)
    sumL0  +=       0.613 * cos(4.96450118774  + 8859.36257569160 *T)
    sumL0  +=       0.615 * cos(5.74467417840  + 6696.47732458460 *T)
    sumL0  +=       0.622 * cos(1.64154132852  + 24150.08005134500 *T)
    sumL0  +=       0.659 * cos(0.12506816965  + 8116.37251515900 *T)
    sumL0  +=       0.679 * cos(1.94943016061  + 12995.22537783680 *T)
    sumL0  +=       0.616 * cos(4.94607504516  + 9485.03276800400 *T)
    sumL0  +=       0.793 * cos(2.84922599989  + 29.49181830340 *T)
    sumL0  +=       0.597 * cos(5.72067567055  + 3113.13629391080 *T)
    sumL0  +=       0.763 * cos(0.42513533173  + 3873.82651014340 *T)
    sumL0  +=       0.587 * cos(1.31684217558  + 9499.25986200560 *T)
    sumL0  +=       0.592 * cos(4.34796024028  + 3826.58447948940 *T)
    sumL0  +=       0.652 * cos(5.53658742492  + 9065.54812412880 *T)
    sumL0  +=       0.583 * cos(0.59190068342  + 26084.02180621620 *T)
    sumL0  +=       0.603 * cos(0.77135566630  + 3342.09689940810 *T)
    sumL0  +=       0.701 * cos(3.35897774328  + 6709.67404086740 *T)
    sumL0  +=       0.574 * cos(1.08976979801  + 6357.71943674220 *T)
    sumL0  +=       0.750 * cos(4.84737198179  + 10022.81760116760 *T)
    sumL0  +=       0.588 * cos(5.65927128604  + 2171.02417529120 *T)
    sumL0  +=       0.632 * cos(3.84852695407  + 16276.46394262300 *T)
    sumL0  +=       0.632 * cos(1.43230388645  + 16706.58525184800 *T)
    sumL0  +=       0.780 * cos(1.15181502540  + 377.37360791580 *T)
    sumL0  +=       0.554 * cos(4.18192049239  + 382.87927857100 *T)
    sumL0  +=       0.568 * cos(0.03816333919  + 6414.61781167780 *T)
    sumL0  +=       0.569 * cos(2.57704345512  + 35.56273446860 *T)
    sumL0  +=       0.566 * cos(0.62804830762  + 3189.56475685690 *T)
    sumL0  +=       0.687 * cos(5.61456944029  + 155.35308913140 *T)
    sumL0  +=       0.563 * cos(2.57358138188  + 19406.67828817460 *T)
    sumL0  +=       0.693 * cos(1.15645091892  + 966.97087743560 *T)
    sumL0  +=       0.576 * cos(1.60357663736  + 3192.53370227350 *T)
    sumL0  +=       0.621 * cos(2.23214771591  + 3274.12501778540 *T)
    sumL0  +=       0.586 * cos(6.17266280012  + 9602.35263622420 *T)
    sumL0  +=       0.541 * cos(0.53968808391  + 13171.00144068760 *T)
    sumL0  +=       0.570 * cos(3.11852290115  + 2221.85663459700 *T)
    sumL0  +=       0.542 * cos(0.41889651002  + 1641.33321019660 *T)
    sumL0  +=       0.532 * cos(0.16606105669  + 5511.63660199100 *T)
    sumL0  +=       0.530 * cos(3.78205188174  + 7270.28968040780 *T)
    sumL0  +=       0.575 * cos(5.48929613719  + 2075.04494807340 *T)
    sumL0  +=       0.670 * cos(3.67090656417  + 6475.03930496240 *T)
    sumL0  +=       0.509 * cos(4.21526585284  + 7380.49600162720 *T)
    sumL0  +=       0.510 * cos(1.59587338243  + 13362.38239649640 *T)
    sumL0  +=       0.494 * cos(2.13711598060  + 2604.73591316800 *T)
    sumL0  +=       0.534 * cos(1.03109772656  + 1478.86657406440 *T)
    sumL0  +=       0.506 * cos(5.92145494356  + 685.04405422600 *T)
    sumL0  +=       0.469 * cos(0.83917541691  + 3041.48603243060 *T)
    sumL0  +=       0.468 * cos(1.97135671591  + 2412.77245819700 *T)
    sumL0  +=       0.468 * cos(4.57635781664  + 11670.28403729680 *T)
    sumL0  +=       0.559 * cos(1.47641018288  + 8671.96987044060 *T)
    sumL0  +=       0.475 * cos(2.07517976424  + 5835.14201864840 *T)
    sumL0  +=       0.466 * cos(2.92801596791  + 2277.29834324750 *T)
    sumL0  +=       0.478 * cos(6.13550464878  + 4825.54491639400 *T)
    sumL0  +=       0.502 * cos(0.77928275348  + 4407.10790388980 *T)
    sumL0  +=       0.462 * cos(3.80503324215  + 11140.59307220220 *T)
    sumL0  +=       0.499 * cos(3.25557381873  + 1744.49329471800 *T)
    sumL0  +=       0.525 * cos(0.74022979976  + 1265.56747862640 *T)
    sumL0  +=       0.521 * cos(6.27577138682  + 3981.49003408200 *T)
    sumL0  +=       0.482 * cos(0.14820337530  + 14158.74771361560 *T)
    sumL0  +=       0.460 * cos(2.03348517376  + 25685.87280280800 *T)
    sumL0  +=       0.510 * cos(4.27704405425  + 3472.15438838620 *T)
    sumL0  +=       0.444 * cos(1.95486223268  + 3226.21331978640 *T)
    sumL0  +=       0.575 * cos(0.59007504383  + 2766.26762836500 *T)
    sumL0  +=       0.437 * cos(5.63861950911  + 8958.93223156120 *T)
    sumL0  +=       0.492 * cos(6.10071747922  + 2285.16264977370 *T)
    sumL0  +=       0.572 * cos(5.10308842466  + 8564.30634650200 *T)
    sumL0  +=       0.430 * cos(4.12851356192  + 13.24177239800 *T)
    sumL0  +=       0.544 * cos(4.78254843814  + 9380.95967271720 *T)
    sumL0  +=       0.430 * cos(3.83861916944  + 1765.76662541760 *T)
    sumL0  +=       0.499 * cos(4.26055736809  + 6652.77566593180 *T)
    sumL0  +=       0.541 * cos(0.84245625839  + 4981.94563689640 *T)
    sumL0  +=       0.591 * cos(2.01874828234  + 27490.69247804480 *T)
    sumL0  +=       0.438 * cos(1.36437546581  + 12509.25332504720 *T)
    sumL0  +=       0.470 * cos(4.70961176845  + 3723.49170527080 *T)
    sumL0  +=       0.421 * cos(1.76100140625  + 956.28915597060 *T)
    sumL0  +=       0.413 * cos(0.49206034039  + 13355.33615979840 *T)
    sumL0  +=       0.412 * cos(0.29353415123  + 2810.93871525740 *T)
    sumL0  +=       0.412 * cos(3.24865417750  + 18451.07854656599 *T)
    sumL0  +=       0.412 * cos(1.70721395764  + 7314.00859271280 *T)
    sumL0  +=       0.412 * cos(4.54594776716  + 2810.90420795300 *T)
    sumL0  +=       0.448 * cos(3.04254429984  + 5636.07001721200 *T)
    sumL0  +=       0.510 * cos(1.18739936388  + 13362.51701710200 *T)
    sumL0  +=       0.408 * cos(1.32068176489  + 20809.46762464520 *T)
    sumL0  +=       0.427 * cos(1.16100742040  + 19004.64794940840 *T)
    sumL0  +=       0.420 * cos(3.49180180953  + 1655.56030419820 *T)
    sumL0  +=       0.433 * cos(0.55429134487  + 9945.57120882380 *T)
    sumL0  +=       0.424 * cos(4.60854671688  + 6518.75821726740 *T)
    sumL0  +=       0.415 * cos(5.67120343816  + 99.56965586960 *T)
    sumL0  +=       0.476 * cos(5.94537443289  + 48835.19385644859 *T)
    sumL0  +=       0.405 * cos(5.63206200287  + 3450.81874791920 *T)
    sumL0  +=       0.495 * cos(4.38317490235  + 2480.30249794700 *T)
    sumL0  +=       0.415 * cos(3.61905205961  + 2089.78223039900 *T)
    sumL0  +=       0.395 * cos(0.25208772249  + 1375.77379984580 *T)
    sumL0  +=       0.421 * cos(2.26337694295  + 13892.14067189380 *T)
    sumL0  +=       0.391 * cos(5.25566087245  + 10042.61267559180 *T)
    sumL0  +=       0.504 * cos(4.91016850189  + 8965.97846825920 *T)
    sumL0  +=       0.417 * cos(3.45088934666  + 279.78650435240 *T)
    sumL0  +=       0.389 * cos(0.98703891513  + 7203.80227149340 *T)
    sumL0  +=       0.431 * cos(2.11881719623  + 56.80326216980 *T)
    sumL0  +=       0.497 * cos(1.63121523594  + 22345.26037610820 *T)
    sumL0  +=       0.377 * cos(4.06263841184  + 10124.93005431800 *T)
    sumL0  +=       0.377 * cos(3.16019395941  + 224.34479570190 *T)
    sumL0  +=       0.378 * cos(5.82532048605  + 6675.70192909220 *T)
    sumL0  +=       0.377 * cos(0.05833226835  + 905.45669666480 *T)
    sumL0  +=       0.445 * cos(0.33560120738  + 515.46387109300 *T)
    sumL0  +=       0.426 * cos(2.66458973231  + 4076.48894023160 *T)
    sumL0  +=       0.489 * cos(0.75760372851  + 3561.02506913860 *T)
    sumL0  +=       0.389 * cos(0.50283826580  + 2825.14855560680 *T)
    sumL0  +=       0.396 * cos(5.14198726837  + 5195.24473233440 *T)
    sumL0  +=       0.414 * cos(3.12530321804  + 10001.06188460700 *T)
    sumL0  +=       0.502 * cos(5.02356488339  + 73.29712585900 *T)
    sumL0  +=       0.432 * cos(3.31538359140  + 5617.91076994730 *T)
    sumL0  +=       0.359 * cos(4.97622870666  + 5820.91492464680 *T)
    sumL0  +=       0.359 * cos(5.42072227300  + 6019.99192661860 *T)
    sumL0  +=       0.449 * cos(3.79544854612  + 4996.17273089800 *T)
    sumL0  +=       0.374 * cos(2.85563965649  + 2111.65031337760 *T)
    sumL0  +=       0.426 * cos(0.24129917209  + 5625.77507647350 *T)
    sumL0  +=       0.423 * cos(3.60507971235  + 8226.57883637840 *T)
    sumL0  +=       0.402 * cos(4.51813719510  + 4392.88080988820 *T)
    sumL0  +=       0.391 * cos(4.26714089799  + 21795.21409161479 *T)
    sumL0  +=       0.447 * cos(4.25776540974  + 18052.92954315780 *T)
    sumL0  +=       0.355 * cos(4.17584780659  + 6740.59871531320 *T)
    sumL0  +=       0.353 * cos(1.81574804066  + 6686.74777770700 *T)
    sumL0  +=       0.462 * cos(3.22304237134  + 2011.10033643980 *T)
    sumL0  +=       0.341 * cos(3.22071023668  + 4253.18267036540 *T)
    sumL0  +=       0.475 * cos(4.28822688035  + 367.22432896240 *T)
    sumL0  +=       0.462 * cos(2.91312544527  + 418.92439890060 *T)
    sumL0  +=       0.416 * cos(4.83203726375  + 8535.85715903420 *T)
    sumL0  +=       0.431 * cos(3.22440150226  + 21265.52312652020 *T)
    sumL0  +=       0.333 * cos(3.99300815003  + 1353.39052800780 *T)
    sumL0  +=       0.340 * cos(0.71636465254  + 15664.03552270859 *T)
    sumL0  +=       0.416 * cos(2.58443168474  + 5753.38488489680 *T)
    sumL0  +=       0.356 * cos(2.35639379021  + 4.57074972000 *T)
    sumL0  +=       0.330 * cos(4.04831945983  + 3.10304225540 *T)
    sumL0  +=       0.381 * cos(4.06283076724  + 1062.90504853820 *T)
    sumL0  +=       0.431 * cos(5.42246026973  + 26482.17080962440 *T)
    sumL0  +=       0.328 * cos(4.06276954550  + 6944.30877677240 *T)
    sumL0  +=       0.334 * cos(5.10221163477  + 22324.90505670940 *T)
    sumL0  +=       0.400 * cos(2.28211698182  + 3209.07046501340 *T)
    sumL0  +=       0.420 * cos(2.44159662565  + 6155.05700665400 *T)
    sumL0  +=       0.375 * cos(1.09229714548  + 4246.06912336460 *T)
    sumL0  +=       0.449 * cos(4.38183542571  + 5106.37905211740 *T)
    sumL0  +=       0.330 * cos(2.55172725319  + 19513.98359510420 *T)
    sumL0  +=       0.319 * cos(4.05704496382  + 5430.39465709880 *T)
    sumL0  +=       0.310 * cos(4.99907184482  + 5095.67507646480 *T)
    sumL0  +=       0.344 * cos(0.62691832180  + 9389.05354078620 *T)
    sumL0  +=       0.319 * cos(3.31673970253  + 596.17837400900 *T)
    sumL0  +=       0.305 * cos(4.66942910910  + 6756.00645196690 *T)
    sumL0  +=       0.409 * cos(6.07568936266  + 50.83245930580 *T)
    sumL0  +=       0.377 * cos(5.42211267415  + 1190.92389187560 *T)
    sumL0  +=       0.342 * cos(2.67821304845  + 6148.01076995600 *T)
    sumL0  +=       0.298 * cos(5.91093444215  + 9886.77220006400 *T)
    sumL0  +=       0.299 * cos(5.44077050156  + 10028.95082710020 *T)
    sumL0  +=       0.315 * cos(1.03304445564  + 3490.17562383440 *T)
    sumL0  +=       0.369 * cos(3.93279262125  + 1879.55837475340 *T)
    sumL0  +=       0.304 * cos(4.56372485787  + 7483.58877584580 *T)
    sumL0  +=       0.330 * cos(3.29725079066  + 286.96236112060 *T)
    sumL0  +=       0.290 * cos(3.43850024020  + 6418.14093002680 *T)
    sumL0  +=       0.289 * cos(2.82766045300  + 3171.03224356680 *T)
    sumL0  +=       0.317 * cos(4.13345374602  + 1883.08149310240 *T)
    sumL0  +=       0.294 * cos(0.01906732590  + 202.25339517410 *T)
    sumL0  +=       0.310 * cos(4.37565854379  + 2796.69436760360 *T)
    sumL0  +=       0.364 * cos(0.54935210242  + 290.48547946960 *T)
    sumL0  +=       0.291 * cos(2.39413982848  + 29026.48522950779 *T)
    sumL0  +=       0.286 * cos(5.76562894312  + 5642.19824260920 *T)
    sumL0  +=       0.392 * cos(3.01171327788  + 10721.10842832700 *T)
    sumL0  +=       0.318 * cos(4.93376925851  + 1197.97012857360 *T)
    sumL0  +=       0.315 * cos(2.84437840100  + 10610.90210710760 *T)
    sumL0  +=       0.299 * cos(1.18299169304  + 550.04628449340 *T)
    sumL0  +=       0.336 * cos(4.30769818760  + 5989.06725217280 *T)
    sumL0  +=       0.315 * cos(1.41407863340  + 6947.83189512140 *T)
    sumL0  +=       0.294 * cos(6.16453752963  + 8982.81066930900 *T)
    sumL0  +=       0.285 * cos(2.82644088669  + 9654.61295113700 *T)
    sumL0  +=       0.281 * cos(0.27739627997  + 8166.15734309380 *T)
    sumL0  +=       0.279 * cos(4.08648927643  + 5355.23588148860 *T)
    sumL0  +=       0.330 * cos(3.11968019451  + 41.55079098480 *T)
    sumL0  +=       0.275 * cos(5.89019272334  + 3337.86091608880 *T)
    sumL0  +=       0.275 * cos(4.00268079937  + 3620.39893105220 *T)
    sumL0  +=       0.292 * cos(3.06830662617  + 4.36203638980 *T)
    sumL0  +=       0.317 * cos(0.01919037405  + 2267.00340250900 *T)
    sumL0  +=       0.272 * cos(1.70373580224  + 3742.28454845700 *T)
    sumL0  +=       0.294 * cos(0.16019381973  + 6670.58818804980 *T)
    sumL0  +=       0.264 * cos(2.07967576148  + 3735.23831175900 *T)
    sumL0  +=       0.262 * cos(4.49437120405  + 6887.41040183680 *T)
    sumL0  +=       0.307 * cos(3.03375977801  + 6660.86953400080 *T)
    sumL0  +=       0.280 * cos(0.47728086796  + 5401.43028077160 *T)
    sumL0  +=       0.268 * cos(0.46331887242  + 6578.13207918100 *T)
    sumL0  +=       0.312 * cos(3.62016422039  + 255.97037637200 *T)
    sumL0  +=       0.278 * cos(5.62032972822  + 3378.74546233760 *T)
    sumL0  +=       0.268 * cos(2.97528422526  + 21947.11137270000 *T)
    sumL0  +=       0.303 * cos(2.71946492092  + 3313.21087060300 *T)
    sumL0  +=       0.263 * cos(3.85883671333  + 28628.33622609960 *T)
    sumL0  +=       0.257 * cos(0.25607724431  + 19146.75926614180 *T)
    sumL0  +=       0.282 * cos(2.29169514758  + 2008.55753915900 *T)
    sumL0  +=       0.301 * cos(0.83275594655  + 2806.98930834210 *T)
    sumL0  +=       0.264 * cos(3.78635346388  + 3936.79080070880 *T)
    sumL0  +=       0.254 * cos(1.28062508539  + 95.97922721780 *T)
    sumL0  +=       0.250 * cos(5.58493858560  + 412.37109687440 *T)
    sumL0  +=       0.284 * cos(4.71065509252  + 5621.91023351320 *T)
    sumL0  +=       0.286 * cos(5.11854366190  + 5621.77561290760 *T)
    sumL0  +=       0.254 * cos(5.77068701321  + 5813.29118932200 *T)
    sumL0  +=       0.311 * cos(2.67573060947  + 912.57024366560 *T)
    sumL0  +=       0.252 * cos(4.19548440341  + 24.85875881600 *T)
    sumL0  +=       0.255 * cos(1.80023727978  + 3193.80074017620 *T)
    sumL0  +=       0.247 * cos(2.94199704738  + 310.84079886840 *T)
    sumL0  +=       0.261 * cos(5.68285399692  + 3133.91168940320 *T)
    sumL0  +=       0.255 * cos(4.46589323429  + 5490.30096152400 *T)
    sumL0  +=       0.247 * cos(2.28207487665  + 246.25172232300 *T)
    sumL0  +=       0.257 * cos(5.64335704456  + 10706.88133432540 *T)
    sumL0  +=       0.246 * cos(0.26910281348  + 22.76849660940 *T)
    sumL0  +=       0.241 * cos(0.03931394254  + 9070.11887384880 *T)
    sumL0  +=       0.283 * cos(1.30585259585  + 4025.65648092580 *T)
    sumL0  +=       0.272 * cos(1.28327911416  + 1442.26120876020 *T)
    sumL0  +=       0.294 * cos(4.02082336381  + 2814.85361486830 *T)
    sumL0  +=       0.249 * cos(4.92144829429  + 31022.75317085620 *T)
    sumL0  +=       0.237 * cos(3.07683243726  + 9947.05568153210 *T)
    sumL0  +=       0.236 * cos(6.05766380360  + 948.19528790160 *T)
    sumL0  +=       0.309 * cos(3.45153658664  + 25287.72379939980 *T)
    sumL0  +=       0.232 * cos(4.99538565545  + 1505.28780909299 *T)
    sumL0  +=       0.275 * cos(1.20294105507  + 6691.86151874940 *T)
    sumL0  +=       0.305 * cos(2.73902403412  + 176.69373013380 *T)
    sumL0  +=       0.231 * cos(5.13610567818  + 6997.61672305620 *T)
    sumL0  +=       0.237 * cos(0.90339496046  + 8.09386806900 *T)
    sumL0  +=       0.229 * cos(0.63250047970  + 23017.06265793620 *T)
    sumL0  +=       0.249 * cos(6.05357589804  + 2060.81785407180 *T)
    sumL0  +=       0.270 * cos(2.14769161882  + 2973.38809773740 *T)
    sumL0  +=       0.241 * cos(4.63019900075  + 1612.95133303160 *T)
    sumL0  +=       0.227 * cos(5.94537053649  + 2942.48067694380 *T)
    sumL0  +=       0.269 * cos(5.93687481153  + 4005.36847182980 *T)
    sumL0  +=       0.227 * cos(3.91459879933  + 2942.44616963940 *T)
    sumL0  +=       0.238 * cos(5.14570745178  + 721.13923120620 *T)
    sumL0  +=       0.275 * cos(1.06902528739  + 3343.36393731080 *T)
    sumL0  +=       0.232 * cos(3.52110290510  + 6674.17861670160 *T)
    sumL0  +=       0.277 * cos(0.19619249083  + 270.19747037360 *T)
    sumL0  +=       0.274 * cos(0.30235979008  + 5607.61582920880 *T)
    sumL0  +=       0.233 * cos(5.75237138576  + 604.84974070480 *T)
    sumL0  +=       0.214 * cos(3.38626276995  + 647.01083331480 *T)
    sumL0  +=       0.282 * cos(0.29694635625  + 12323.42309600880 *T)
    sumL0  +=       0.282 * cos(3.18114716742  + 6364.83298374300 *T)
    sumL0  +=       0.264 * cos(4.64751563064  + 1346.34429130980 *T)
    sumL0  +=       0.280 * cos(4.64995184564  + 6701.58017279840 *T)
    sumL0  +=       0.246 * cos(3.03036675631  + 3304.00706139560 *T)
    sumL0  +=       0.213 * cos(4.80555095681  + 8830.91338822380 *T)
    sumL0  +=       0.220 * cos(1.54642908481  + 23546.75362303080 *T)
    sumL0  +=       0.214 * cos(1.35475764936  + 184.84990797020 *T)
    sumL0  +=       0.228 * cos(4.29420876593  + 2970.91261075940 *T)
    sumL0  +=       0.229 * cos(2.53361265404  + 3710.31224264020 *T)
    sumL0  +=       0.273 * cos(1.89943168433  + 270.13016007080 *T)
    sumL0  +=       0.284 * cos(3.28345980607  + 16063.16484718500 *T)
    sumL0  +=       0.211 * cos(5.84341192825  + 1971.95217385480 *T)
    sumL0  +=       0.214 * cos(0.22905754741  + 5.19919116580 *T)
    sumL0  +=       0.246 * cos(4.55971876123  + 6040.34724601740 *T)
    sumL0  +=       0.229 * cos(1.24432891752  + 3568.08855948880 *T)
    sumL0  +=       0.216 * cos(3.20098680120  + 362.12113673080 *T)
    sumL0  +=       0.201 * cos(2.45025935972  + 20426.57109242200 *T)
    sumL0  +=       0.223 * cos(2.07804482295  + 17101.21113690720 *T)
    sumL0  +=       0.271 * cos(3.49210901141  + 3436.59165391760 *T)
    sumL0  +=       0.226 * cos(0.42945734871  + 557.09252119140 *T)
    sumL0  +=       0.264 * cos(5.05794940070  + 2938.94030494260 *T)
    sumL0  +=       0.207 * cos(6.03491870748  + 22487.37169284160 *T)
    sumL0  +=       0.197 * cos(5.62923956977  + 5209.47182633600 *T)
    sumL0  +=       0.211 * cos(2.26445749553  + 765.79306444640 *T)
    sumL0  +=       0.202 * cos(0.88670673933  + 767.84881934860 *T)
    sumL0  +=       0.212 * cos(1.85965510753  + 4782.87363546000 *T)
    sumL0  +=       0.194 * cos(4.77420671032  + 6850.80503653260 *T)
    sumL0  +=       0.192 * cos(0.51565572072  + 323.50541665740 *T)
    sumL0  +=       0.209 * cos(4.79794980231  + 14047.49376102520 *T)
    sumL0  +=       0.193 * cos(2.56057288512  + 18606.49894600020 *T)
    sumL0  +=       0.209 * cos(1.46088434217  + 5.85720229960 *T)
    sumL0  +=       0.214 * cos(0.97578687412  + 17468.85519794540 *T)
    sumL0  +=       0.196 * cos(4.08789826861  + 3262.86188271590 *T)
    sumL0  +=       0.198 * cos(3.99111216230  + 10018.24685144760 *T)
    sumL0  +=       0.235 * cos(1.11133081576  + 625.67019231240 *T)
    sumL0  +=       0.197 * cos(0.48692287381  + 15106.87569121440 *T)
    sumL0  +=       0.214 * cos(5.33581901309  + 955.24152459960 *T)
    sumL0  +=       0.199 * cos(0.40088144456  + 7586.68155006440 *T)
    sumL0  +=       0.251 * cos(5.57906426342  + 138.58819838440 *T)
    sumL0  +=       0.208 * cos(4.11480858445  + 1755.06264976500 *T)
    sumL0  +=       0.210 * cos(2.30499661227  + 53.30794628380 *T)
    sumL0  +=       0.253 * cos(1.27099771143  + 309.27832265580 *T)
    sumL0  +=       0.242 * cos(3.71426781511  + 3212.59358336240 *T)
    sumL0  +=       0.178 * cos(3.06961783443  + 1437.17561419860 *T)
    sumL0  +=       0.203 * cos(0.28410501881  + 582.99891137840 *T)
    sumL0  +=       0.190 * cos(0.10457198807  + 26087.90314157420 *T)
    sumL0  +=       0.178 * cos(0.53068485806  + 7366.26890762560 *T)
    sumL0  +=       0.245 * cos(4.20195166994  + 8436.28750316460 *T)
    sumL0  +=       0.176 * cos(3.66697456425  + 6688.27109009760 *T)
    sumL0  +=       0.177 * cos(4.48104044431  + 8799.98871377800 *T)
    sumL0  +=       0.181 * cos(3.57461078885  + 15121.10278521600 *T)
    sumL0  +=       0.198 * cos(3.55065962909  + 24606.13555322000 *T)
    sumL0  +=       0.187 * cos(3.45440079747  + 1329.51209026000 *T)
    sumL0  +=       0.176 * cos(5.81532365473  + 12406.16055082860 *T)
    sumL0  +=       0.235 * cos(1.28568802052  + 4427.39591298580 *T)
    sumL0  +=       0.179 * cos(1.55718466444  + 3362.46325602620 *T)
    sumL0  +=       0.174 * cos(1.61086801253  + 30065.51184029820 *T)
    sumL0  +=       0.199 * cos(0.07164714815  + 375.76573157020 *T)
    sumL0  +=       0.174 * cos(4.58412775793  + 3283.71405176420 *T)
    sumL0  +=       0.171 * cos(5.82189798695  + 23937.85638974100 *T)
    sumL0  +=       0.182 * cos(2.57146189845  + 418.50432280700 *T)
    sumL0  +=       0.168 * cos(5.13131619552  + 21393.54196985760 *T)
    sumL0  +=       0.183 * cos(4.93161962050  + 9468.26787725700 *T)
    sumL0  +=       0.167 * cos(1.48091400654  + 2619.47319549360 *T)
    sumL0  +=       0.169 * cos(5.12437031125  + 3223.29255847960 *T)
    sumL0  +=       0.188 * cos(3.41823914376  + 5032.77809620220 *T)
    sumL0  +=       0.181 * cos(0.50010974122  + 2125.87740737920 *T)
    sumL0  +=       0.164 * cos(1.71077130702  + 15849.86575174700 *T)
    sumL0  +=       0.202 * cos(6.21085922593  + 3909.43430072720 *T)
    sumL0  +=       0.171 * cos(5.86158194602  + 625.62513619720 *T)
    sumL0  +=       0.165 * cos(3.85569118220  + 13207.02930736500 *T)
    sumL0  +=       0.163 * cos(2.04373367430  + 3347.29609057390 *T)
    sumL0  +=       0.160 * cos(2.83784244321  + 5888.44996493220 *T)
    sumL0  +=       0.212 * cos(2.32801112252  + 3232.94890276120 *T)
    sumL0  +=       0.163 * cos(4.23488695195  + 31968.94865279940 *T)
    sumL0  +=       0.159 * cos(1.48046671186  + 249.90446074220 *T)
    sumL0  +=       0.160 * cos(0.21960307161  + 12942.96506292400 *T)
    sumL0  +=       0.220 * cos(3.90787704883  + 9638.94074787620 *T)
    sumL0  +=       0.161 * cos(2.94093367568  + 3370.04193523580 *T)
    sumL0  +=       0.159 * cos(5.59017475732  + 1442.27846241240 *T)
    sumL0  +=       0.221 * cos(3.82548751198  + 1954.71715036360 *T)
    sumL0  +=       0.192 * cos(4.62595272276  + 3336.68027343670 *T)
    sumL0  +=       0.164 * cos(1.59200641542  + 386.41965057220 *T)
    sumL0  +=       0.169 * cos(5.45784867095  + 259.56080502380 *T)
    sumL0  +=       0.201 * cos(0.95077053594  + 29424.63423291600 *T)
    sumL0  +=       0.174 * cos(1.15216485688  + 6382.09845913040 *T)
    sumL0  +=       0.163 * cos(1.16606667991  + 4289.78803566960 *T)
    sumL0  +=       0.187 * cos(3.22135956460  + 2751.54759969160 *T)
    sumL0  +=       0.162 * cos(4.79372588575  + 24889.57479599160 *T)
    sumL0  +=       0.192 * cos(0.88040946364  + 5244.04923920100 *T)
    sumL0  +=       0.160 * cos(2.85034784370  + 9374.82644678460 *T)
    sumL0  +=       0.172 * cos(2.38601254063  + 2281.24775016280 *T)
    sumL0  +=       0.202 * cos(4.12164786769  + 7321.12213971360 *T)
    sumL0  +=       0.156 * cos(4.55171204694  + 27682.14074415640 *T)
    sumL0  +=       0.170 * cos(4.62851491273  + 10824.20120254560 *T)
    sumL0  +=       0.157 * cos(0.91363725609  + 4503.08713110760 *T)
    sumL0  +=       0.156 * cos(5.58255618318  + 10448.43547097540 *T)
    sumL0  +=       0.157 * cos(0.92229160815  + 15636.56665630900 *T)
    sumL0  +=       0.171 * cos(5.43820628341  + 11904.91877320180 *T)
    sumL0  +=       0.180 * cos(1.89091405241  + 13575.74880223720 *T)
    sumL0  +=       0.198 * cos(4.42539692212  + 10025.42770875120 *T)
    sumL0  +=       0.158 * cos(4.12087914642  + 11240.16272807180 *T)
    sumL0  +=       0.211 * cos(2.42218772392  + 8749.15625447220 *T)
    sumL0  +=       0.183 * cos(1.47934835951  + 6677.34351804160 *T)
    sumL0  +=       0.197 * cos(1.43469282909  + 14061.72085502680 *T)
    sumL0  +=       0.170 * cos(4.32424842659  + 742.99006053260 *T)
    sumL0  +=       0.170 * cos(2.92905324873  + 9093.99731159660 *T)
    sumL0  +=       0.160 * cos(2.79797608932  + 1461.05405194640 *T)
    sumL0  +=       0.174 * cos(5.37738922471  + 3318.76159737340 *T)
    sumL0  +=       0.172 * cos(0.35524089578  + 2281.21324285840 *T)
    sumL0  +=       0.147 * cos(2.76602235522  + 4193.80880845180 *T)
    sumL0  +=       0.151 * cos(4.23962231148  + 3368.01398279660 *T)
    sumL0  +=       0.157 * cos(1.01295201512  + 24336.00539314920 *T)
    sumL0  +=       0.165 * cos(0.58335652806  + 15906.76412668260 *T)
    sumL0  +=       0.149 * cos(5.69553079999  + 31570.79964939120 *T)
    sumL0  +=       0.154 * cos(5.28504381510  + 1481.40937134520 *T)
    sumL0  +=       0.163 * cos(1.81288628510  + 18849.22754997420 *T)
    sumL0  +=       0.145 * cos(3.57053080979  + 3497.01314720220 *T)
    sumL0  +=       0.156 * cos(3.56978076425  + 21791.69097326579 *T)
    sumL0  +=       0.143 * cos(0.56704903096  + 13212.88650966460 *T)
    sumL0  +=       0.155 * cos(0.08429188155  + 6657.34641565180 *T)
    sumL0  +=       0.184 * cos(2.54999403339  + 24076.44458812540 *T)
    sumL0  +=       0.146 * cos(1.16594990123  + 526.16784674560 *T)
    sumL0  +=       0.144 * cos(4.19229335185  + 2771.79055267240 *T)
    sumL0  +=       0.169 * cos(3.10771037057  + 239.20548562500 *T)
    sumL0  +=       0.144 * cos(6.26688878164  + 6679.74038069130 *T)
    sumL0  +=       0.174 * cos(2.37509025280  + 3397.51080163540 *T)
    sumL0  +=       0.163 * cos(3.88047919138  + 16703.04487984680 *T)
    sumL0  +=       0.161 * cos(1.73958633312  + 2185.25126929280 *T)
    sumL0  +=       0.164 * cos(0.62146994730  + 1538.24043597800 *T)
    sumL0  +=       0.182 * cos(6.16856014864  + 6685.10618875760 *T)
    sumL0  +=       0.140 * cos(4.08200595943  + 4186.69526145100 *T)
    sumL0  +=       0.192 * cos(4.47709998867  + 57.87869600380 *T)
    sumL0  +=       0.151 * cos(3.06451266512  + 838.96928775040 *T)
    sumL0  +=       0.139 * cos(1.12767399649  + 6682.70932610790 *T)
    sumL0  +=       0.187 * cos(4.97174626997  + 6681.65473652630 *T)
    sumL0  +=       0.150 * cos(5.69209290362  + 9360.60435331840 *T)
    sumL0  +=       0.160 * cos(5.66336382679  + 10927.29397676420 *T)
    sumL0  +=       0.137 * cos(3.46860408347  + 5562.46906129680 *T)
    sumL0  +=       0.192 * cos(5.27908098216  + 34363.36559755600 *T)
    sumL0  +=       0.137 * cos(5.87023088486  + 2945.98654164060 *T)
    sumL0  +=       0.180 * cos(1.07686767816  + 15806.14683944200 *T)
    sumL0  +=       0.135 * cos(0.34109799474  + 7322.10246078180 *T)
    sumL0  +=       0.169 * cos(2.27408583150  + 379.37341387420 *T)
    sumL0  +=       0.172 * cos(3.83276777655  + 6621.85099148600 *T)
    sumL0  +=       0.134 * cos(0.89486611090  + 13286.18363552360 *T)
    sumL0  +=       0.168 * cos(3.24434245132  + 11614.43329373220 *T)
    sumL0  +=       0.135 * cos(6.03727673544  + 1214.80232962340 *T)
    sumL0  +=       0.150 * cos(0.81912738038  + 3416.87849797540 *T)
    sumL0  +=       0.136 * cos(5.06627117467  + 14421.83163698840 *T)
    sumL0  +=       0.138 * cos(5.19336961955  + 13363.43002786740 *T)
    sumL0  +=       0.163 * cos(5.91124428824  + 16703.07938715119 *T)
    sumL0  +=       0.181 * cos(3.31978654659  + 139.69813952280 *T)
    sumL0  +=       0.162 * cos(4.05905033891  + 1795.25844372100 *T)
    sumL0  +=       0.132 * cos(0.40901506100  + 1083.26036793700 *T)
    sumL0  +=       0.156 * cos(0.12304144640  + 1107.13880568480 *T)
    sumL0  +=       0.158 * cos(6.25022915072  + 6666.99775939800 *T)
    sumL0  +=       0.131 * cos(3.95250083817  + 3.45580804620 *T)
    sumL0  +=       0.147 * cos(4.94673462086  + 15010.89646399660 *T)
    sumL0  +=       0.140 * cos(6.13853404667  + 12729.66596748600 *T)
    sumL0  +=       0.131 * cos(5.43398849473  + 25665.51748340920 *T)
    sumL0  +=       0.135 * cos(5.06017183369  + 11876.46958573400 *T)
    sumL0  +=       0.128 * cos(2.86969332730  + 6549.68289171320 *T)
    sumL0  +=       0.174 * cos(5.18169887171  + 28230.18722269139 *T)
    sumL0  +=       0.179 * cos(3.25589354429  + 4922.57177498280 *T)
    sumL0  +=       0.139 * cos(5.33624615245  + 23958.63178523340 *T)
    sumL0  +=       0.127 * cos(5.78406495652  + 14577.18472611980 *T)
    sumL0  +=       0.128 * cos(1.27276688747  + 8584.66166590080 *T)
    sumL0  +=       0.133 * cos(0.39052474466  + 12410.73130054860 *T)
    sumL0  +=       0.134 * cos(1.09641189843  + 8962.43809625800 *T)
    sumL0  +=       0.125 * cos(3.20445201519  + 29.88204361020 *T)
    sumL0  +=       0.136 * cos(3.62250122667  + 3511.28529731900 *T)
    sumL0  +=       0.162 * cos(5.36465149996  + 2472.67876262220 *T)
    sumL0  +=       0.150 * cos(0.40179894287  + 1111.64224510200 *T)
    sumL0  +=       0.129 * cos(1.94909076932  + 19645.52555679060 *T)
    sumL0  +=       0.172 * cos(5.21328160366  + 6747.71226231400 *T)
    sumL0  +=       0.126 * cos(4.39378462616  + 21548.96236929180 *T)
    sumL0  +=       0.146 * cos(3.85987749053  + 13361.46938573100 *T)
    sumL0  +=       0.156 * cos(5.24740844581  + 6645.19698672220 *T)
    sumL0  +=       0.137 * cos(4.94588047257  + 20995.39296644940 *T)
    sumL0  +=       0.131 * cos(1.72431469184  + 11925.27409260060 *T)
    sumL0  +=       0.136 * cos(2.04268940441  + 10654.62101941260 *T)
    sumL0  +=       0.140 * cos(1.54439148652  + 8219.46528937760 *T)
    sumL0  +=       0.127 * cos(6.16695075115  + 10016.31435579200 *T)
    sumL0  +=       0.155 * cos(5.47145482987  + 2.96894541660 *T)
    sumL0  +=       0.132 * cos(6.25626202928  + 13227.38462676380 *T)
    sumL0  +=       0.121 * cos(3.74135485334  + 10294.51023745100 *T)
    sumL0  +=       0.134 * cos(3.12718351732  + 8962.47260356240 *T)
    sumL0  +=       0.122 * cos(0.77282907794  + 4936.79886898440 *T)
    sumL0  +=       0.125 * cos(4.23213631052  + 15650.79375031060 *T)
    sumL0  +=       0.157 * cos(1.79175545524  + 16489.76303806100 *T)
    sumL0  +=       0.155 * cos(4.15655698197  + 56.38318607620 *T)
    sumL0  +=       0.134 * cos(1.53268827347  + 708.98980227659 *T)
    sumL0  +=       0.136 * cos(4.01025697673  + 9797.49248439750 *T)
    sumL0  +=       0.156 * cos(4.61556268820  + 3354.83952070140 *T)
    sumL0  +=       0.131 * cos(4.47476808853  + 11776.89992986440 *T)
    sumL0  +=       0.118 * cos(1.17466010141  + 4606.17990532620 *T)
    sumL0  +=       0.128 * cos(0.12648796247  + 6681.64492949320 *T)
    sumL0  +=       0.122 * cos(0.13511377940  + 8322.55806359620 *T)
    sumL0  +=       0.119 * cos(1.29139116350  + 10544.41469819320 *T)
    sumL0  +=       0.128 * cos(4.44966050740  + 13465.54248101780 *T)
    sumL0  +=       0.119 * cos(4.01709629990  + 14481.20549890200 *T)
    sumL0  +=       0.147 * cos(1.81590279650  + 685.11136452880 *T)
    sumL0  +=       0.137 * cos(4.78402836071  + 3613.28538405140 *T)
    sumL0  +=       0.151 * cos(1.48428558337  + 9698.33186344200 *T)
    sumL0  +=       0.136 * cos(2.47757608387  + 10156.90236013480 *T)
    sumL0  +=       0.123 * cos(2.42353206298  + 6.13322593260 *T)
    sumL0  +=       0.117 * cos(5.40635440024  + 688.63448287780 *T)
    sumL0  +=       0.153 * cos(5.00372030984  + 14556.89671702380 *T)
    sumL0  +=       0.133 * cos(3.15773785434  + 1125.86933910360 *T)
    sumL0  +=       0.153 * cos(1.94292660454  + 6533.14612897330 *T)
    sumL0  +=       0.124 * cos(4.94608245253  + 12825.64519470380 *T)
    sumL0  +=       0.117 * cos(1.13528750738  + 4452.25467180180 *T)
    sumL0  +=       0.115 * cos(3.34121813462  + 10001.48196070061 *T)
    sumL0  +=       0.115 * cos(3.44586362144  + 7696.88787128380 *T)
    sumL0  +=       0.117 * cos(0.72489390402  + 27.46886639960 *T)
    sumL0  +=       0.132 * cos(2.41318250916  + 6717.25272007700 *T)
    sumL0  +=       0.118 * cos(0.35995031424  + 27832.03821928320 *T)
    sumL0  +=       0.157 * cos(1.73929012934  + 11560.07771607740 *T)
    sumL0  +=       0.148 * cos(2.14023976341  + 15265.88651930040 *T)
    sumL0  +=       0.113 * cos(1.29183863745  + 26880.31981303260 *T)
    sumL0  +=       0.151 * cos(0.95890610457  + 6530.17718355670 *T)
    sumL0  +=       0.153 * cos(5.62679784282  + 9911.63095888000 *T)
    sumL0  +=       0.157 * cos(0.09791378159  + 138.51749687070 *T)
    sumL0  +=       0.119 * cos(2.16883964744  + 7082.89697515680 *T)
    sumL0  +=       0.113 * cos(5.43102427558  + 422.02744115600 *T)
    sumL0  +=       0.130 * cos(2.13493506322  + 117.81280289200 *T)
    sumL0  +=       0.140 * cos(2.51194591630  + 28.98162997940 *T)
    sumL0  +=       0.128 * cos(1.23822751200  + 6680.80477730600 *T)
    sumL0  +=       0.111 * cos(4.60294776614  + 25135.82651831460 *T)
    sumL0  +=       0.118 * cos(0.34226497704  + 4140.43355186520 *T)
    sumL0  +=       0.112 * cos(4.02233737541  + 6314.00052443720 *T)
    sumL0  +=       0.134 * cos(3.35926214730  + 11247.20896476980 *T)
    sumL0  +=       0.120 * cos(6.09430372868  + 4106.40549114620 *T)
    sumL0  +=       0.117 * cos(0.03006245187  + 7.67379197540 *T)
    sumL0  +=       0.109 * cos(3.62630036582  + 3344.11829139661 *T)
    sumL0  +=       0.109 * cos(1.94277745743  + 16702.99482319620 *T)
    sumL0  +=       0.114 * cos(4.87315937546  + 6151.51663465280 *T)
    sumL0  +=       0.109 * cos(5.65707270595  + 3344.15279870100 *T)
    sumL0  +=       0.108 * cos(0.72637519191  + 10.56935504700 *T)
    sumL0  +=       0.128 * cos(5.77326416186  + 13369.56325380000 *T)
    sumL0  +=       0.147 * cos(4.73848506543  + 11610.91017538320 *T)
    sumL0  +=       0.145 * cos(4.95495312886  + 9859.37064396720 *T)
    sumL0  +=       0.111 * cos(3.28808788953  + 257.01800774300 *T)
    sumL0  +=       0.112 * cos(4.79777726412  + 12140.60114047780 *T)
    sumL0  +=       0.109 * cos(3.73199687068  + 13635.12266415080 *T)
    sumL0  +=       0.117 * cos(1.20057736014  + 6705.10329114740 *T)
    sumL0  +=       0.149 * cos(4.91688353237  + 277.03499374140 *T)
    sumL0  +=       0.108 * cos(3.37299798972  + 6717.83021870380 *T)
    sumL0  +=       0.106 * cos(0.53379407701  + 5085.10572141780 *T)
    sumL0  +=       0.125 * cos(4.62770076269  + 9329.67967887260 *T)
    sumL0  +=       0.107 * cos(4.60288570375  + 1957.72507985320 *T)
    sumL0  +=       0.113 * cos(4.04786119569  + 6953.89781075120 *T)
    sumL0  +=       0.106 * cos(0.75378922686  + 149.89747512680 *T)
    sumL0  +=       0.114 * cos(0.62074562387  + 6151.55114195720 *T)
    sumL0  +=       0.108 * cos(5.22204692781  + 1556.56814695540 *T)
    sumL0  +=       0.104 * cos(1.02502614223  + 4878.85286267780 *T)
    sumL0  +=       0.133 * cos(1.15811543786  + 816.65332621520 *T)
    sumL0  +=       0.134 * cos(1.69834857217  + 183.86958690200 *T)
    sumL0  +=       0.104 * cos(2.58832774275  + 14591.41182012140 *T)
    sumL0  +=       0.107 * cos(4.42117075795  + 7.13080065300 *T)
    sumL0  +=       0.126 * cos(0.03750399181  + 16.83220104979 *T)
    sumL0  +=       0.102 * cos(1.20624870972  + 11986.67590695340 *T)
    sumL0  +=       0.102 * cos(2.97579561878  + 29822.78323632420 *T)
    sumL0  +=       0.116 * cos(4.73049067619  + 13366.04013545100 *T)
    sumL0  +=       0.130 * cos(4.10484632841  + 35.60779058380 *T)
    sumL0  +=       0.113 * cos(3.69200525056  + 44.65383324020 *T)
    sumL0  +=       0.111 * cos(5.12900291511  + 146.79443287140 *T)
    sumL0  +=       0.126 * cos(1.82187459534  + 26887.36604973060 *T)
    sumL0  +=       0.121 * cos(6.08309355316  + 7100.70949727480 *T)
    sumL0  +=       0.105 * cos(6.16606590276  + 9175.75444534820 *T)
    sumL0  +=       0.132 * cos(2.97955203077  + 1250.83019630080 *T)
    sumL0  +=       0.106 * cos(5.08419485009  + 8742.04270747140 *T)
    sumL0  +=       0.120 * cos(0.59116561391  + 31172.65064598300 *T)
    sumL0  +=       0.112 * cos(3.70888430091  + 10050.28646756720 *T)
    sumL0  +=       0.100 * cos(1.82077956276  + 12012.58229714040 *T)
    sumL0  +=       0.107 * cos(4.86100312387  + 7899.55030137200 *T)
    sumL0  +=       0.099 * cos(0.72997085174  + 18977.17908300880 *T)
    sumL0  +=       0.107 * cos(5.25025570307  + 63.73589830340 *T)
    sumL0  +=       0.099 * cos(2.79139996703  + 17932.02424682120 *T)
    sumL0  +=       0.097 * cos(4.94370173613  + 28109.21461605200 *T)
    sumL0  +=       0.122 * cos(4.18676567644  + 19617.07636932279 *T)
    sumL0  +=       0.121 * cos(3.14207351411  + 9278.84721956680 *T)
    sumL0  +=       0.112 * cos(1.43537110436  + 16232.74503031800 *T)
    sumL0  +=       0.096 * cos(2.57929371627  + 12715.43887348440 *T)
    sumL0  +=       0.096 * cos(2.51396484677  + 3535.58381116040 *T)
    sumL0  +=       0.112 * cos(6.19137312195  + 6685.15700666270 *T)
    sumL0  +=       0.109 * cos(1.53428110338  + 16703.12944380180 *T)
    sumL0  +=       0.096 * cos(4.29245810274  + 7906.59653807000 *T)
    sumL0  +=       0.096 * cos(3.08847844641  + 3468.63127003720 *T)
    sumL0  +=       0.095 * cos(6.09186084700  + 10037.08975128440 *T)
    sumL0  +=       0.094 * cos(1.79812841802  + 20047.19767854780 *T)
    sumL0  +=       0.110 * cos(4.24859410294  + 3337.07205469860 *T)
    sumL0  +=       0.093 * cos(2.17945548408  + 16066.06586147480 *T)
    sumL0  +=       0.115 * cos(5.44548079869  + 11353.89216764020 *T)
    sumL0  +=       0.098 * cos(1.89605054783  + 33406.12426699800 *T)
    sumL0  +=       0.101 * cos(3.00793005829  + 3528.53757446240 *T)
    sumL0  +=       0.096 * cos(5.40659912420  + 5749.86176654780 *T)
    sumL0  +=       0.113 * cos(3.93443805059  + 7167.19690618920 *T)
    sumL0  +=       0.092 * cos(3.58985964414  + 6724.94376570460 *T)
    sumL0  +=       0.119 * cos(3.90561570260  + 11346.77862063940 *T)
    sumL0  +=       0.097 * cos(2.57280790230  + 1747.60737815290 *T)
    sumL0  +=       0.118 * cos(3.82111723459  + 9072.66167112960 *T)
    sumL0  +=       0.096 * cos(0.72646224606  + 3152.68727893721 *T)
    sumL0  +=       0.092 * cos(6.23928508055  + 24491.42579258340 *T)
    sumL0  +=       0.099 * cos(1.86991802881  + 72.05573335480 *T)
    sumL0  +=       0.094 * cos(1.51091806861  + 9801.42463766060 *T)
    sumL0  +=       0.091 * cos(4.94377681711  + 8756.26980147300 *T)
    sumL0  +=       0.095 * cos(2.58771497236  + 11663.17049029600 *T)
    sumL0  +=       0.093 * cos(0.66092443320  + 30.99198474860 *T)
    sumL0  +=       0.088 * cos(2.47386826588  + 17232.75309859360 *T)
    sumL0  +=       0.100 * cos(0.19616913512  + 11883.58313273480 *T)
    sumL0  +=       0.094 * cos(0.61277732908  + 73.63546321960 *T)
    sumL0  +=       0.091 * cos(6.04139177526  + 153.92523352440 *T)
    sumL0  +=       0.095 * cos(0.30865798871  + 29168.59654624120 *T)
    sumL0  +=       0.110 * cos(6.27936643935  + 3337.10656200300 *T)
    sumL0  +=       0.087 * cos(2.09287965409  + 10264.56588407340 *T)
    sumL0  +=       0.089 * cos(3.27633365389  + 3002.36972771680 *T)
    sumL0  +=       0.091 * cos(1.89622049447  + 37895.42629036740 *T)
    sumL0  +=       0.118 * cos(3.88268323672  + 1236.07566032300 *T)
    sumL0  +=       0.094 * cos(2.27819061990  + 4845.90023579280 *T)
    sumL0  +=       0.085 * cos(3.07438636612  + 9983.70424446160 *T)
    sumL0  +=       0.086 * cos(3.94463425808  + 16.76489074700 *T)
    sumL0  +=       0.099 * cos(4.58890225582  + 3774.32416457660 *T)
    sumL0  +=       0.086 * cos(1.27758598252  + 34513.26307268280 *T)
    sumL0  +=       0.118 * cos(6.10625309129  + 454.90936652730 *T)
    sumL0  +=       0.096 * cos(6.08772700197  + 1755.47168467910 *T)
    sumL0  +=       0.085 * cos(4.73193631802  + 4576.68808702280 *T)
    sumL0  +=       0.116 * cos(0.88573303872  + 16511.61386738740 *T)
    sumL0  +=       0.083 * cos(3.13948373813  + 10191.41746323240 *T)
    sumL0  +=       0.083 * cos(0.36972674664  + 9755.23023837760 *T)
    sumL0  +=       0.112 * cos(2.26524635842  + 10006.58480891440 *T)
    sumL0  +=       0.087 * cos(4.25461508452  + 8.16813658240 *T)
    sumL0  +=       0.087 * cos(5.31716927554  + 1868.85939963620 *T)
    sumL0  +=       0.115 * cos(3.30507830050  + 14157.70008224460 *T)
    sumL0  +=       0.087 * cos(2.90507350729  + 36.99559061100 *T)
    sumL0  +=       0.083 * cos(3.58400049244  + 12928.73796892240 *T)
    sumL0  +=       0.086 * cos(2.46839611271  + 113.72943956840 *T)
    sumL0  +=       0.096 * cos(2.99829695985  + 6510.55198278040 *T)
    sumL0  +=       0.084 * cos(1.34333260788  + 4133.38731516720 *T)
    sumL0  +=       0.093 * cos(2.41895388985  + 20040.15144184979 *T)
    sumL0  +=       0.085 * cos(5.92093268818  + 12299.54465826100 *T)
    sumL0  +=       0.082 * cos(4.71321236549  + 7733.49323658800 *T)
    sumL0  +=       0.081 * cos(5.26971110105  + 12199.97500239140 *T)
    sumL0  +=       0.080 * cos(4.34410099751  + 7218.02936549500 *T)
    sumL0  +=       0.095 * cos(0.02613500045  + 2.90101428980 *T)
    sumL0  +=       0.094 * cos(6.22018441621  + 20043.69181385099 *T)
    sumL0  +=       0.085 * cos(4.66573809999  + 8638.94993325280 *T)
    sumL0  +=       0.079 * cos(2.89839604813  + 22854.59602180400 *T)
    sumL0  +=       0.104 * cos(0.94882888784  + 2735.76268599500 *T)
    sumL0  +=       0.080 * cos(1.68097745917  + 12839.87228870540 *T)
    sumL0  +=       0.085 * cos(1.95515305721  + 4686.95671800959 *T)
    sumL0  +=       0.094 * cos(0.58147434179  + 14128.24277124560 *T)
    sumL0  +=       0.091 * cos(1.16317999954  + 3495.96551583120 *T)
    sumL0  +=       0.084 * cos(4.04104644774  + 9815.65173166220 *T)
    sumL0  +=       0.082 * cos(1.83662951627  + 19190.47817844680 *T)
    sumL0  +=       0.080 * cos(4.04733274699  + 3188.71514561460 *T)
    sumL0  +=       0.084 * cos(2.71620952175  + 12089.76868117200 *T)
    sumL0  +=       0.080 * cos(2.77363158210  + 9495.66943335380 *T)
    sumL0  +=       0.078 * cos(5.77285430206  + 10397.60301166960 *T)
    sumL0  +=       0.084 * cos(0.78188080031  + 7214.43893684320 *T)
    sumL0  +=       0.094 * cos(2.14762315943  + 13532.02988993220 *T)
    sumL0  +=       0.101 * cos(5.76349087526  + 1484.93248969420 *T)
    sumL0  +=       0.082 * cos(3.11421789241  + 8446.99147881720 *T)
    sumL0  +=       0.096 * cos(5.85152670536  + 3492.50970778500 *T)
    sumL0  +=       0.079 * cos(4.69625309523  + 5298.33750655300 *T)
    sumL0  +=       0.097 * cos(3.97960395898  + 18208.34994259200 *T)
    sumL0  +=       0.105 * cos(1.74376904536  + 36.64856292950 *T)
    sumL0  +=       0.094 * cos(4.18941238607  + 20043.65730654660 *T)
    sumL0  +=       0.075 * cos(3.46629875063  + 13341.67431130680 *T)
    sumL0  +=       0.092 * cos(2.67278833288  + 6680.79497027290 *T)
    sumL0  +=       0.075 * cos(4.59723160021  + 9993.38809263160 *T)
    sumL0  +=       0.087 * cos(0.91979096638  + 8425.71814811760 *T)
    sumL0  +=       0.074 * cos(0.42228487063  + 48429.28218232440 *T)
    sumL0  +=       0.074 * cos(0.48470953224  + 11456.98494185880 *T)
    sumL0  +=       0.077 * cos(0.10707484151  + 1371.41176345600 *T)
    sumL0  +=       0.093 * cos(1.38010528069  + 3145.64104223920 *T)
    sumL0  +=       0.075 * cos(5.28061257070  + 12306.59089495900 *T)
    sumL0  +=       0.079 * cos(3.78666386040  + 7.57867920960 *T)
    sumL0  +=       0.073 * cos(4.42902162523  + 14.73728232560 *T)
    sumL0  +=       0.084 * cos(3.41538432898  + 6637.50594109460 *T)
    sumL0  +=       0.093 * cos(3.41302342683  + 1596.11913198180 *T)
    sumL0  +=       0.071 * cos(3.13537213242  + 9488.62319665580 *T)
    sumL0  +=       0.079 * cos(2.33442828498  + 12465.53441274220 *T)
    sumL0  +=       0.078 * cos(2.21525134547  + 2810.98877190800 *T)
    sumL0  +=       0.099 * cos(0.48801369293  + 17499.36014031540 *T)
    sumL0  +=       0.075 * cos(4.10299895268  + 5525.86369599260 *T)
    sumL0  +=       0.071 * cos(4.04186021414  + 259.49349472100 *T)
    sumL0  +=       0.073 * cos(2.20102876718  + 30774.50164257480 *T)
    sumL0  +=       0.074 * cos(4.93439968296  + 13517.80279593060 *T)
    sumL0  +=       0.071 * cos(5.32625652676  + 6670.17915313570 *T)
    sumL0  +=       0.078 * cos(2.17664185326  + 13421.82356871280 *T)
    sumL0  +=       0.088 * cos(5.38469204503  + 17762.44406368820 *T)
    sumL0  +=       0.081 * cos(2.48353174238  + 73.60095591520 *T)
    sumL0  +=       0.092 * cos(1.13088448190  + 7747.72033058960 *T)
    sumL0  +=       0.070 * cos(1.16361323015  + 5220.17080145320 *T)
    sumL0  +=       0.084 * cos(5.83220387910  + 22.37827130260 *T)
    sumL0  +=       0.076 * cos(5.21065604639  + 4271.97551355160 *T)
    sumL0  +=       0.080 * cos(4.01817623627  + 3311.18291816379 *T)
    sumL0  +=       0.068 * cos(2.45937697940  + 6279.55273164240 *T)
    sumL0  +=       0.067 * cos(4.31723563048  + 2281.29780681340 *T)
    sumL0  +=       0.067 * cos(4.88776321797  + 4379.63903749020 *T)
    sumL0  +=       0.069 * cos(3.57006764536  + 14955.04572043200 *T)
    sumL0  +=       0.078 * cos(2.62374773232  + 2810.85415130240 *T)
    sumL0  +=       0.067 * cos(4.76301301456  + 6812.76681508600 *T)
    sumL0  +=       0.070 * cos(5.60319851277  + 13383.22510229160 *T)
    sumL0  +=       0.070 * cos(3.05647570196  + 10081.21114201300 *T)
    sumL0  +=       0.081 * cos(4.91537623765  + 3333.92876282570 *T)
    sumL0  +=       0.082 * cos(5.49204987110  + 5483.25472482600 *T)
    sumL0  +=       0.065 * cos(3.50466797067  + 9161.52735134660 *T)
    sumL0  +=       0.065 * cos(2.79212460026  + 13497.51478683460 *T)
    sumL0  +=       0.070 * cos(0.93640431668  + 6106.88005506480 *T)
    sumL0  +=       0.073 * cos(2.57583496978  + 6614.73744448520 *T)
    sumL0  +=       0.070 * cos(0.84170070580  + 16695.94858649819 *T)
    sumL0  +=       0.069 * cos(1.10040627005  + 2544.33167353560 *T)
    sumL0  +=       0.063 * cos(6.05841309176  + 6453.74872061060 *T)
    sumL0  +=       0.069 * cos(2.67053540192  + 2641.34127847220 *T)
    sumL0  +=       0.086 * cos(1.73112264739  + 6286.59896834040 *T)
    sumL0  +=       0.062 * cos(5.49002783256  + 4957.08687808040 *T)
    sumL0  +=       0.072 * cos(3.38857910283  + 3184.21170619740 *T)
    sumL0  +=       0.069 * cos(5.35281990892  + 2544.29716623120 *T)
    sumL0  +=       0.064 * cos(4.79175268505  + 3181.66890891660 *T)
    sumL0  +=       0.060 * cos(6.18449904353  + 3361.94806716680 *T)
    sumL0  +=       0.065 * cos(1.48236627574  + 4819.47900076420 *T)
    sumL0  +=       0.054 * cos(3.66018290396  + 2917.53735417280 *T)
    sumL0  +=       0.064 * cos(4.72286914683  + 2281.16318620780 *T)
    sumL0  +=       0.053 * cos(4.65521663671  + 3448.27595063840 *T)
    sumL0  +=       0.060 * cos(0.77450276964  + 3319.27678623280 *T)
    sumL0  +=       0.051 * cos(1.45142985881  + 3547.31316399640 *T)
    sumL0  +=       0.051 * cos(4.81258158513  + 4403.51747523800 *T)
    sumL0  +=       0.050 * cos(0.37366902835  + 3322.79990458180 *T)
    sumL0  +=       0.050 * cos(4.68787711774  + 3253.30422216000 *T)
    sumL0  +=       0.067 * cos(4.63578431535  + 4694.00295470760 *T)
    sumL0  +=       0.049 * cos(3.17471347671  + 5452.26274007740 *T)
    sumL0  +=       0.051 * cos(2.54791773443  + 4001.84535348080 *T)
    sumL0  +=       0.048 * cos(2.81171553464  + 3372.58473251660 *T)
    sumL0  +=       0.054 * cos(4.45945188023  + 10284.92120347220 *T)
    sumL0  +=       0.053 * cos(3.85131213520  + 9758.75335672660 *T)
    sumL0  +=       0.051 * cos(1.11244863072  + 6901.63749583840 *T)
    sumL0  +=       0.062 * cos(1.42198009630  + 6830.78805053420 *T)
    sumL0  +=       0.055 * cos(5.61672988180  + 45494.58142974879 *T)
    sumL0  +=       0.054 * cos(0.61701959453  + 4716.38622654560 *T)
    sumL0  +=       0.054 * cos(4.28556326526  + 2910.49111747480 *T)
    sumL0  +=       0.045 * cos(3.14626081312  + 13951.51453380740 *T)
    sumL0  +=       0.050 * cos(4.16235480519  + 12979.55317457600 *T)
    sumL0  +=       0.046 * cos(1.66182208107  + 7768.00833968560 *T)
    sumL0  +=       0.045 * cos(5.07701425888  + 14951.52260208300 *T)
    sumL0  +=       0.043 * cos(4.31723459530  + 5881.40372823420 *T)
    sumL0  +=       0.045 * cos(2.39713581676  + 8852.24902869080 *T)
    sumL0  +=       0.043 * cos(2.10202660495  + 3337.15661865360 *T)
    sumL0  +=       0.048 * cos(0.19934255781  + 1905.46476494040 *T)
    sumL0  +=       0.056 * cos(5.25687695855  + 12808.88030395680 *T)
    sumL0  +=       0.049 * cos(2.06320966659  + 4296.90158267040 *T)
    sumL0  +=       0.042 * cos(6.07282126953  + 4295.85395129940 *T)
    sumL0  +=       0.043 * cos(1.74571613780  + 3130.83644961080 *T)
    sumL0  +=       0.050 * cos(2.27112469544  + 3427.92063123960 *T)
    sumL0  +=       0.050 * cos(0.30214747765  + 3358.42494881780 *T)
    sumL0  +=       0.043 * cos(5.05297177283  + 3229.42578441220 *T)
    sumL0  +=       0.054 * cos(6.11998016428  + 16894.51039961060 *T)
    sumL0  +=       0.043 * cos(1.90603004035  + 3451.79906898740 *T)
    sumL0  +=       0.039 * cos(0.50727886431  + 10011.20061474960 *T)
    sumL0  +=       0.052 * cos(3.17851184154  + 2324.94940881560 *T)
    sumL0  +=       0.038 * cos(4.88249540246  + 3499.55594448300 *T)
    sumL0  +=       0.039 * cos(5.28113083211  + 3343.15522398060 *T)
    sumL0  +=       0.039 * cos(4.98185580267  + 3468.56395973440 *T)
    sumL0  +=       0.047 * cos(2.16244218985  + 10027.36020440680 *T)
    sumL0  +=       0.037 * cos(5.84953090935  + 6791.43117461900 *T)
    sumL0  +=       0.048 * cos(5.03779546898  + 7064.10413197060 *T)
    sumL0  +=       0.038 * cos(0.58076516432  + 8966.38750317330 *T)
    sumL0  +=       0.051 * cos(3.36181117037  + 8976.68244391180 *T)
    sumL0  +=       0.042 * cos(4.13725767158  + 8336.78515759780 *T)
    sumL0  +=       0.038 * cos(5.75425648893  + 4531.53631857540 *T)
    sumL0  +=       0.038 * cos(3.55681323215  + 5351.71276313960 *T)
    sumL0  +=       0.040 * cos(2.79414141035  + 3067.93946934820 *T)
    sumL0  +=       0.040 * cos(4.99798468494  + 10042.19259949820 *T)
    sumL0  +=       0.037 * cos(1.82286411009  + 10017.95594474140 *T)
    sumL0  +=       0.036 * cos(6.02115369796  + 6474.52411610300 *T)
    sumL0  +=       0.044 * cos(3.01299253859  + 7417.10136693140 *T)
    sumL0  +=       0.045 * cos(2.33470159894  + 3123.79021291280 *T)
    sumL0  +=       0.039 * cos(4.91767933978  + 8034.61538140740 *T)

    # MARS Longitude
    sumL1  = 0
    sumL1  +=  334085627474.342 * cos(0.00000000000  + 0.00000000000 *T)
    sumL1  +=  1458227.051 * cos(3.60426053609  + 3340.61242669980 *T)
    sumL1  +=  164901.343 * cos(3.92631250962  + 6681.22485339960 *T)
    sumL1  +=   19963.338 * cos(4.26594061030  + 10021.83728009940 *T)
    sumL1  +=    3452.399 * cos(4.73210386365  + 3.52311834900 *T)
    sumL1  +=    2485.480 * cos(4.61277567318  + 13362.44970679920 *T)
    sumL1  +=     841.551 * cos(4.45858256765  + 2281.23049651060 *T)
    sumL1  +=     537.566 * cos(5.01589727492  + 398.14900340820 *T)
    sumL1  +=     521.041 * cos(4.99422678175  + 3344.13554504880 *T)
    sumL1  +=     432.614 * cos(2.56066402860  + 191.44826611160 *T)
    sumL1  +=     429.656 * cos(5.31646162367  + 155.42039943420 *T)
    sumL1  +=     381.747 * cos(3.53881289437  + 796.29800681640 *T)
    sumL1  +=     314.129 * cos(4.96335266049  + 16703.06213349900 *T)
    sumL1  +=     282.804 * cos(3.15967518204  + 2544.31441988340 *T)
    sumL1  +=     205.664 * cos(4.56891455660  + 2146.16541647520 *T)
    sumL1  +=     168.805 * cos(1.32894813366  + 3337.08930835080 *T)
    sumL1  +=     157.587 * cos(4.18501035954  + 1751.53953141600 *T)
    sumL1  +=     133.686 * cos(2.23325104196  + 0.98032106820 *T)
    sumL1  +=     116.561 * cos(2.21347652545  + 1059.38193018920 *T)
    sumL1  +=     117.591 * cos(6.02407213861  + 6151.53388830500 *T)
    sumL1  +=     113.595 * cos(5.42803224317  + 3738.76143010800 *T)
    sumL1  +=     133.563 * cos(5.97421903927  + 1748.01641306700 *T)
    sumL1  +=      91.098 * cos(1.09627836591  + 1349.86740965880 *T)
    sumL1  +=      83.301 * cos(5.29636626272  + 6684.74797174860 *T)
    sumL1  +=     113.876 * cos(2.12869455089  + 1194.44701022460 *T)
    sumL1  +=      80.776 * cos(4.42813405865  + 529.69096509460 *T)
    sumL1  +=      79.531 * cos(2.24864266330  + 8962.45534991020 *T)
    sumL1  +=      72.505 * cos(5.84208163240  + 242.72860397400 *T)
    sumL1  +=      72.946 * cos(2.50189460554  + 951.71840625060 *T)
    sumL1  +=      71.487 * cos(3.85636094435  + 2914.01423582380 *T)
    sumL1  +=      85.342 * cos(3.90854841008  + 553.56940284240 *T)
    sumL1  +=      67.582 * cos(5.02327686473  + 382.89653222320 *T)
    sumL1  +=      65.089 * cos(1.01802439311  + 3340.59517304760 *T)
    sumL1  +=      65.089 * cos(3.04879603978  + 3340.62968035200 *T)
    sumL1  +=      61.508 * cos(4.15183159800  + 3149.16416058820 *T)
    sumL1  +=      48.477 * cos(4.87362121538  + 213.29909543800 *T)
    sumL1  +=      46.584 * cos(1.31452419914  + 3185.19202726560 *T)
    sumL1  +=      56.520 * cos(3.88813699320  + 4136.91043351620 *T)
    sumL1  +=      47.613 * cos(1.18238046057  + 3333.49887969900 *T)
    sumL1  +=      41.343 * cos(0.71385375517  + 1592.59601363280 *T)
    sumL1  +=      40.055 * cos(5.31611875491  + 20043.67456019880 *T)
    sumL1  +=      40.272 * cos(2.72542480614  + 7.11354700080 *T)
    sumL1  +=      32.886 * cos(5.41067411968  + 6283.07584999140 *T)
    sumL1  +=      28.244 * cos(0.04534124888  + 9492.14631500480 *T)
    sumL1  +=      22.294 * cos(5.88516997273  + 3870.30339179440 *T)
    sumL1  +=      22.431 * cos(5.46592525433  + 20.35531939880 *T)
    sumL1  +=      22.612 * cos(0.83775884934  + 3097.88382272579 *T)
    sumL1  +=      21.418 * cos(5.37934044204  + 3340.54511639700 *T)
    sumL1  +=      23.335 * cos(6.16762213077  + 3532.06069281140 *T)
    sumL1  +=      26.579 * cos(3.88960724782  + 1221.84856632140 *T)
    sumL1  +=      22.797 * cos(1.54504711003  + 2274.11694950980 *T)
    sumL1  +=      20.431 * cos(2.36353950189  + 1589.07289528380 *T)
    sumL1  +=      20.186 * cos(3.36375535766  + 5088.62883976680 *T)
    sumL1  +=      26.554 * cos(5.11271747607  + 2700.71514038580 *T)
    sumL1  +=      19.675 * cos(2.57805423988  + 12303.06777661000 *T)
    sumL1  +=      19.468 * cos(0.49216434489  + 6677.70173505060 *T)
    sumL1  +=      21.104 * cos(3.52525428062  + 15.25247118500 *T)
    sumL1  +=      21.425 * cos(4.97081508139  + 3340.67973700260 *T)
    sumL1  +=      18.505 * cos(5.57863503922  + 1990.74501704100 *T)
    sumL1  +=      17.811 * cos(6.12537931996  + 4292.33083295040 *T)
    sumL1  +=      16.472 * cos(2.60291845066  + 3341.59274776800 *T)
    sumL1  +=      16.599 * cos(1.25519718278  + 3894.18182954220 *T)
    sumL1  +=      19.455 * cos(2.53112676345  + 4399.99435688900 *T)
    sumL1  +=      15.000 * cos(1.03464802434  + 2288.34404351140 *T)
    sumL1  +=      20.029 * cos(4.73119428749  + 4690.47983635860 *T)
    sumL1  +=      15.381 * cos(2.47009470350  + 4535.05943692440 *T)
    sumL1  +=      19.964 * cos(5.78652958398  + 7079.37385680780 *T)
    sumL1  +=      15.307 * cos(2.26515985343  + 3723.50895892300 *T)
    sumL1  +=      14.705 * cos(3.36979890389  + 6681.24210705180 *T)
    sumL1  +=      13.535 * cos(2.12334410410  + 5486.77784317500 *T)
    sumL1  +=      12.950 * cos(5.61929676688  + 10025.36039844840 *T)
    sumL1  +=      12.682 * cos(2.95022113262  + 3496.03282613400 *T)
    sumL1  +=      13.644 * cos(1.97739547259  + 5614.72937620960 *T)
    sumL1  +=      13.013 * cos(1.51424752315  + 5628.95647021120 *T)
    sumL1  +=      14.705 * cos(1.33902715586  + 6681.20759974740 *T)
    sumL1  +=      11.353 * cos(6.23438193885  + 135.06508003540 *T)
    sumL1  +=      13.275 * cos(3.42243595774  + 5621.84292321040 *T)
    sumL1  +=      10.867 * cos(5.28184140482  + 2818.03500860600 *T)
    sumL1  +=      11.850 * cos(3.12701832949  + 426.59819087600 *T)
    sumL1  +=      10.472 * cos(2.73581537999  + 2787.04302385740 *T)
    sumL1  +=      11.132 * cos(5.84178807242  + 2803.80791460440 *T)
    sumL1  +=      11.764 * cos(2.58551521265  + 8432.76438481560 *T)
    sumL1  +=      11.854 * cos(5.47630686910  + 3553.91152213780 *T)
    sumL1  +=       8.490 * cos(1.91378007528  + 11773.37681151540 *T)
    sumL1  +=       9.708 * cos(4.52957217749  + 6489.77658728800 *T)
    sumL1  +=       8.562 * cos(3.16141186861  + 162.46663613220 *T)
    sumL1  +=      10.958 * cos(4.15771850822  + 2388.89402044920 *T)
    sumL1  +=       8.133 * cos(1.61295625304  + 2957.71589447660 *T)
    sumL1  +=       8.840 * cos(4.23294294197  + 7477.52286021600 *T)
    sumL1  +=       8.034 * cos(5.69983564288  + 6041.32756708560 *T)
    sumL1  +=       8.344 * cos(2.18273563186  + 23.87843774780 *T)
    sumL1  +=       7.696 * cos(5.71877332978  + 9623.68827669120 *T)
    sumL1  +=       8.695 * cos(4.43542512603  + 5092.15195811580 *T)
    sumL1  +=       8.434 * cos(3.16292250873  + 3347.72597370060 *T)
    sumL1  +=       6.664 * cos(5.07517838003  + 8031.09226305840 *T)
    sumL1  +=       8.650 * cos(4.33256981793  + 3339.63210563160 *T)
    sumL1  +=       7.372 * cos(6.17831593269  + 3583.34103067380 *T)
    sumL1  +=       5.726 * cos(3.68120120299  + 8429.24126646660 *T)
    sumL1  +=       6.186 * cos(3.54165967734  + 692.15760122680 *T)
    sumL1  +=       5.438 * cos(1.05129689580  + 4933.20844033260 *T)
    sumL1  +=       6.108 * cos(1.66240879939  + 6525.80445396540 *T)
    sumL1  +=       5.154 * cos(1.14703246368  + 28.44918746780 *T)
    sumL1  +=       4.850 * cos(5.29254832907  + 6681.29216370240 *T)
    sumL1  +=       5.467 * cos(6.12511022569  + 2487.41604494780 *T)
    sumL1  +=       4.866 * cos(3.10475368803  + 5.52292430740 *T)
    sumL1  +=       6.360 * cos(2.11896608283  + 5884.92684658320 *T)
    sumL1  +=       5.223 * cos(0.37446264120  + 12832.75874170460 *T)
    sumL1  +=       4.710 * cos(0.23326120326  + 36.02786667740 *T)
    sumL1  +=       4.954 * cos(2.44806818502  + 5099.26550511660 *T)
    sumL1  +=       4.861 * cos(5.60505298870  + 6467.92575796160 *T)
    sumL1  +=       4.706 * cos(0.02998416568  + 7210.91581849420 *T)
    sumL1  +=       4.845 * cos(5.70115105957  + 6681.15754309680 *T)
    sumL1  +=       5.496 * cos(2.01006612503  + 522.57741809380 *T)
    sumL1  +=       4.964 * cos(1.51006845561  + 1744.42598441520 *T)
    sumL1  +=       4.443 * cos(0.31208413867  + 10018.31416175040 *T)
    sumL1  +=       5.381 * cos(0.18359380473  + 2942.46342329160 *T)
    sumL1  +=       4.075 * cos(3.95582108330  + 3.88133535800 *T)
    sumL1  +=       5.462 * cos(0.19274227117  + 7632.94325965020 *T)
    sumL1  +=       4.110 * cos(1.59535768711  + 7234.79425624200 *T)
    sumL1  +=       4.287 * cos(2.87635993968  + 2810.92146160520 *T)
    sumL1  +=       5.276 * cos(2.22638595594  + 3127.31333126180 *T)
    sumL1  +=       4.450 * cos(4.17005729081  + 2906.90068882300 *T)
    sumL1  +=       5.144 * cos(5.66878565669  + 23384.28698689860 *T)
    sumL1  +=       3.844 * cos(2.26442183160  + 2699.73481931760 *T)
    sumL1  +=       3.514 * cos(1.76463961051  + 1758.65307841680 *T)
    sumL1  +=       3.351 * cos(2.66194137496  + 4929.68532198360 *T)
    sumL1  +=       4.299 * cos(4.43057446968  + 640.87760738220 *T)
    sumL1  +=       3.140 * cos(1.75866226873  + 9595.23908922340 *T)
    sumL1  +=       3.716 * cos(2.91969220147  + 15643.68020330980 *T)
    sumL1  +=       3.249 * cos(6.13937134379  + 10419.98628350760 *T)
    sumL1  +=       3.077 * cos(2.56115174488  + 7064.12138562280 *T)
    sumL1  +=       3.208 * cos(2.32519453080  + 5085.03841111500 *T)
    sumL1  +=       2.930 * cos(1.27797225349  + 574.34479833480 *T)
    sumL1  +=       2.771 * cos(1.75664216142  + 639.89728631400 *T)
    sumL1  +=       3.325 * cos(2.58945297384  + 2118.76386037840 *T)
    sumL1  +=       3.187 * cos(2.86646751510  + 7740.60678358880 *T)
    sumL1  +=       2.780 * cos(0.43157089331  + 5828.02847164760 *T)
    sumL1  +=       2.824 * cos(0.98500544471  + 3191.04922956520 *T)
    sumL1  +=       3.016 * cos(1.86555882509  + 7.04623669800 *T)
    sumL1  +=       3.364 * cos(1.52847138842  + 6674.11130639880 *T)
    sumL1  +=       2.672 * cos(3.70855172347  + 10021.85453375160 *T)
    sumL1  +=       2.636 * cos(3.11790581052  + 6836.64525283380 *T)
    sumL1  +=       2.672 * cos(1.67778079449  + 10021.82002644720 *T)
    sumL1  +=       2.563 * cos(3.77294986894  + 2921.12778282460 *T)
    sumL1  +=       2.509 * cos(0.30454165124  + 3475.67750673520 *T)
    sumL1  +=       2.400 * cos(0.96972421975  + 3319.83703120740 *T)
    sumL1  +=       2.262 * cos(2.81394314950  + 7875.67186362420 *T)
    sumL1  +=       2.395 * cos(2.96002707485  + 6682.20517446780 *T)
    sumL1  +=       2.210 * cos(0.61263930586  + 10973.55568635000 *T)
    sumL1  +=       2.248 * cos(4.12382007742  + 59.37386191360 *T)
    sumL1  +=       2.426 * cos(5.91508357946  + 5331.35744374080 *T)
    sumL1  +=       2.158 * cos(2.17583545077  + 15113.98923821520 *T)
    sumL1  +=       1.941 * cos(5.47668312685  + 11371.70468975820 *T)
    sumL1  +=       1.903 * cos(5.11165653855  + 1066.49547719000 *T)
    sumL1  +=       2.370 * cos(3.87889340214  + 3355.86489788480 *T)
    sumL1  +=       2.299 * cos(1.15914205086  + 3320.25710730100 *T)
    sumL1  +=       1.944 * cos(5.89081872133  + 6894.52394883760 *T)
    sumL1  +=       1.843 * cos(3.07643314617  + 3325.35995551480 *T)
    sumL1  +=       1.809 * cos(4.97905218276  + 1648.44675719740 *T)
    sumL1  +=       2.136 * cos(1.91364787635  + 8969.56889691100 *T)
    sumL1  +=       2.099 * cos(3.00410255642  + 6254.62666252360 *T)
    sumL1  +=       1.915 * cos(3.55907431740  + 3767.21061757580 *T)
    sumL1  +=       1.991 * cos(5.37274107053  + 206.18554843720 *T)
    sumL1  +=       1.685 * cos(5.49701299817  + 266.60704172180 *T)
    sumL1  +=       1.646 * cos(1.31923405548  + 3264.34635542420 *T)
    sumL1  +=       1.732 * cos(1.81361103995  + 536.80451209540 *T)
    sumL1  +=       1.723 * cos(3.25900379342  + 7903.07341972100 *T)
    sumL1  +=       1.564 * cos(5.75428852012  + 3360.96774609859 *T)
    sumL1  +=       1.589 * cos(1.73273563259  + 3134.42687826260 *T)
    sumL1  +=       1.690 * cos(2.43213510013  + 3120.19978426100 *T)
    sumL1  +=       1.549 * cos(1.54016426558  + 8425.65083781480 *T)
    sumL1  +=       1.536 * cos(5.88431472627  + 20.77539549240 *T)
    sumL1  +=       1.460 * cos(4.89733072879  + 9830.38901398780 *T)
    sumL1  +=       2.023 * cos(5.94808387002  + 13365.97282514820 *T)
    sumL1  +=       1.991 * cos(3.11613326265  + 3361.38782219220 *T)
    sumL1  +=       1.401 * cos(2.24482184868  + 3344.20285535160 *T)
    sumL1  +=       1.365 * cos(4.58006320751  + 10818.13528691580 *T)
    sumL1  +=       1.392 * cos(5.48931017516  + 170.67287061920 *T)
    sumL1  +=       1.360 * cos(3.07974035205  + 6127.65545055720 *T)
    sumL1  +=       1.345 * cos(1.18653158091  + 14584.29827312060 *T)
    sumL1  +=       1.717 * cos(5.62501515015  + 6158.64743530580 *T)
    sumL1  +=       1.408 * cos(1.82072980335  + 3337.02199804800 *T)
    sumL1  +=       1.736 * cos(2.01921900546  + 10575.40668294180 *T)
    sumL1  +=       1.402 * cos(4.50079374387  + 5729.50644714900 *T)
    sumL1  +=       1.266 * cos(5.91088435118  + 9808.53818466140 *T)
    sumL1  +=       1.433 * cos(6.05024653324  + 12964.30070339100 *T)
    sumL1  +=       1.223 * cos(0.82796258263  + 419.48464387520 *T)
    sumL1  +=       1.393 * cos(1.05117949107  + 6438.49624942560 *T)
    sumL1  +=       1.272 * cos(1.50116723856  + 8439.87793181640 *T)
    sumL1  +=       1.143 * cos(4.89747373731  + 220.41264243880 *T)
    sumL1  +=       1.183 * cos(3.52587190041  + 6688.33840040040 *T)
    sumL1  +=       1.132 * cos(6.19236255633  + 6144.42034130420 *T)
    sumL1  +=       1.154 * cos(2.23058485970  + 8955.34180290940 *T)
    sumL1  +=       1.129 * cos(3.44264300692  + 10177.25767953360 *T)
    sumL1  +=       1.152 * cos(5.29913300616  + 27.40155609680 *T)
    sumL1  +=       1.274 * cos(4.58421238440  + 6247.51311552280 *T)
    sumL1  +=       1.093 * cos(2.82623332360  + 4569.57454002200 *T)
    sumL1  +=       1.303 * cos(0.44350560735  + 87.30820453981 *T)
    sumL1  +=       1.335 * cos(2.14204457730  + 11243.68584642080 *T)
    sumL1  +=       1.102 * cos(1.96260837539  + 6298.32832117640 *T)
    sumL1  +=       1.066 * cos(2.89865914321  + 10404.73381232260 *T)
    sumL1  +=       1.027 * cos(4.79269049654  + 3914.95722503460 *T)
    sumL1  +=       1.015 * cos(0.22847818730  + 3230.40610548040 *T)
    sumL1  +=       1.041 * cos(3.73274497451  + 10213.28554621100 *T)
    sumL1  +=       1.154 * cos(4.14326179244  + 6261.74020952440 *T)
    sumL1  +=       1.306 * cos(3.67645557029  + 2301.58581590939 *T)
    sumL1  +=       1.145 * cos(5.12292846748  + 4562.46099302120 *T)
    sumL1  +=       1.275 * cos(5.14333847164  + 2693.60159338500 *T)
    sumL1  +=       0.914 * cos(1.22398892152  + 103.09277421860 *T)
    sumL1  +=       0.955 * cos(1.52875141393  + 3369.06161416760 *T)
    sumL1  +=       0.908 * cos(0.48223420834  + 13358.92658845020 *T)
    sumL1  +=       0.892 * cos(1.35161136807  + 1214.73501932060 *T)
    sumL1  +=       0.998 * cos(3.29665881950  + 3178.14579056760 *T)
    sumL1  +=       0.980 * cos(1.69212466625  + 43.71891230500 *T)
    sumL1  +=       0.919 * cos(2.45445889752  + 13916.01910964160 *T)
    sumL1  +=       0.899 * cos(5.94990531529  + 12168.00269657460 *T)
    sumL1  +=       0.891 * cos(5.63297246408  + 10021.90459040220 *T)
    sumL1  +=       0.922 * cos(3.91483430303  + 9225.53927328300 *T)
    sumL1  +=       0.930 * cos(0.24073004700  + 6923.95345737360 *T)
    sumL1  +=       0.923 * cos(1.10396074450  + 29.42950853600 *T)
    sumL1  +=       0.816 * cos(4.65198282005  + 2707.82868738660 *T)
    sumL1  +=       0.791 * cos(6.08595583868  + 2384.32327072920 *T)
    sumL1  +=       0.820 * cos(4.80968546763  + 533.21408344360 *T)
    sumL1  +=       0.758 * cos(2.06012386134  + 12935.85151592320 *T)
    sumL1  +=       0.738 * cos(1.03564559078  + 14314.16811304980 *T)
    sumL1  +=       0.908 * cos(0.69862047595  + 16173.37116840440 *T)
    sumL1  +=       0.700 * cos(4.08788234610  + 5202.35827933520 *T)
    sumL1  +=       0.829 * cos(2.01062325398  + 9866.41688066520 *T)
    sumL1  +=       0.887 * cos(6.04145842617  + 10021.76996979660 *T)
    sumL1  +=       0.768 * cos(1.39532897827  + 8273.82086703240 *T)
    sumL1  +=       0.873 * cos(4.45446600602  + 3316.73398895200 *T)
    sumL1  +=       0.673 * cos(3.02440642752  + 1039.02661079040 *T)
    sumL1  +=       0.660 * cos(2.83410276989  + 107.66352393860 *T)
    sumL1  +=       0.838 * cos(0.39195370222  + 10551.52824519400 *T)
    sumL1  +=       0.708 * cos(3.27560955336  + 18984.29263000960 *T)
    sumL1  +=       0.658 * cos(6.01853128902  + 26724.89941359840 *T)
    sumL1  +=       0.685 * cos(1.98132615912  + 1228.96211332220 *T)
    sumL1  +=       0.595 * cos(0.10260171285  + 111.18664228760 *T)
    sumL1  +=       0.563 * cos(1.59138368358  + 2391.43681773000 *T)
    sumL1  +=       0.555 * cos(2.70801962190  + 4555.34744602040 *T)
    sumL1  +=       0.696 * cos(2.89276686038  + 2648.45482547300 *T)
    sumL1  +=       0.587 * cos(4.56017988729  + 6680.24453233140 *T)
    sumL1  +=       0.540 * cos(2.86002662919  + 5459.37628707820 *T)
    sumL1  +=       0.530 * cos(5.64877399946  + 6034.21402008480 *T)
    sumL1  +=       0.520 * cos(0.20012848836  + 13760.59871020740 *T)
    sumL1  +=       0.552 * cos(5.08766140543  + 1903.43681250120 *T)
    sumL1  +=       0.512 * cos(3.21411265909  + 11081.21921028860 *T)
    sumL1  +=       0.641 * cos(5.19459033638  + 6048.44111408640 *T)
    sumL1  +=       0.677 * cos(3.87723948458  + 13517.87010623340 *T)
    sumL1  +=       0.534 * cos(0.23224033336  + 51.28033786241 *T)
    sumL1  +=       0.670 * cos(3.69368226469  + 3335.08950239240 *T)
    sumL1  +=       0.486 * cos(2.41879628327  + 3364.49086444760 *T)
    sumL1  +=       0.500 * cos(4.31447859057  + 3344.49376205780 *T)
    sumL1  +=       0.481 * cos(1.56481992611  + 1964.83862685400 *T)
    sumL1  +=       0.504 * cos(2.47456295599  + 3863.18984479360 *T)
    sumL1  +=       0.523 * cos(0.65856269237  + 853.19638175200 *T)
    sumL1  +=       0.481 * cos(0.12971954679  + 66.48740891440 *T)
    sumL1  +=       0.535 * cos(2.98601678918  + 8270.29774868340 *T)
    sumL1  +=       0.450 * cos(2.02303462834  + 13362.43245314700 *T)
    sumL1  +=       0.448 * cos(5.59827312967  + 149.56319713460 *T)
    sumL1  +=       0.519 * cos(2.75931838722  + 3503.07906283200 *T)
    sumL1  +=       0.534 * cos(4.77352933347  + 1118.75579210280 *T)
    sumL1  +=       0.450 * cos(4.05380888708  + 13362.46696045140 *T)
    sumL1  +=       0.439 * cos(4.83194205477  + 3116.26763099790 *T)
    sumL1  +=       0.567 * cos(5.67483490268  + 227.47613278900 *T)
    sumL1  +=       0.459 * cos(3.44555998004  + 6702.00024889200 *T)
    sumL1  +=       0.545 * cos(2.01193901951  + 7910.18696672180 *T)
    sumL1  +=       0.425 * cos(2.79854459343  + 433.71173787680 *T)
    sumL1  +=       0.429 * cos(4.30113040289  + 16858.48253293320 *T)
    sumL1  +=       0.409 * cos(0.05448009540  + 3304.58456002240 *T)
    sumL1  +=       0.434 * cos(5.70806855136  + 21.85082932640 *T)
    sumL1  +=       0.399 * cos(4.93233684937  + 9779.10867612540 *T)
    sumL1  +=       0.532 * cos(1.31038986189  + 6660.44945790720 *T)
    sumL1  +=       0.398 * cos(5.31345458361  + 13119.72110282519 *T)
    sumL1  +=       0.459 * cos(2.53671963587  + 74.78159856730 *T)
    sumL1  +=       0.384 * cos(2.29906801437  + 12310.18132361080 *T)
    sumL1  +=       0.467 * cos(5.12562716972  + 1596.18644228460 *T)
    sumL1  +=       0.516 * cos(5.84767782422  + 1052.26838318840 *T)
    sumL1  +=       0.414 * cos(4.75409582610  + 3981.49003408200 *T)
    sumL1  +=       0.365 * cos(3.73271671549  + 5518.75014899180 *T)
    sumL1  +=       0.367 * cos(0.13506394328  + 56.89837493560 *T)
    sumL1  +=       0.459 * cos(0.15582180531  + 9381.93999378540 *T)
    sumL1  +=       0.392 * cos(2.15845463651  + 3980.50971301380 *T)
    sumL1  +=       0.396 * cos(1.48538591462  + 17924.91069982040 *T)
    sumL1  +=       0.456 * cos(0.64517343174  + 6816.28993343500 *T)
    sumL1  +=       0.358 * cos(5.87219240658  + 3607.21946842160 *T)
    sumL1  +=       0.490 * cos(0.65766946042  + 3376.64029337720 *T)
    sumL1  +=       0.365 * cos(1.91816243676  + 3347.65866339780 *T)
    sumL1  +=       0.397 * cos(1.80006148744  + 7895.95987272020 *T)
    sumL1  +=       0.336 * cos(2.14687780119  + 6677.63442474780 *T)
    sumL1  +=       0.340 * cos(2.88185925998  + 17256.63153634140 *T)
    sumL1  +=       0.334 * cos(6.13670038311  + 5724.93569742900 *T)
    sumL1  +=       0.339 * cos(4.08527025169  + 664.75604513000 *T)
    sumL1  +=       0.432 * cos(2.52188285182  + 18454.60166491500 *T)
    sumL1  +=       0.336 * cos(4.22863444521  + 6696.47732458460 *T)
    sumL1  +=       0.342 * cos(5.96724705923  + 3546.79797513700 *T)
    sumL1  +=       0.326 * cos(4.02557052581  + 6872.67311951120 *T)
    sumL1  +=       0.323 * cos(5.05444843838  + 3237.51965248120 *T)
    sumL1  +=       0.324 * cos(2.89151245241  + 8329.67161059700 *T)
    sumL1  +=       0.321 * cos(6.25886976298  + 10235.13637553740 *T)
    sumL1  +=       0.333 * cos(2.57725424455  + 6684.81528205140 *T)
    sumL1  +=       0.356 * cos(6.27424874986  + 8671.96987044060 *T)
    sumL1  +=       0.319 * cos(5.05665355586  + 36.60536530420 *T)
    sumL1  +=       0.305 * cos(3.88755666972  + 7107.82304427560 *T)
    sumL1  +=       0.322 * cos(6.28125601341  + 16706.58525184800 *T)
    sumL1  +=       0.334 * cos(3.15240620873  + 11216.28429032400 *T)
    sumL1  +=       0.287 * cos(6.16467002771  + 3973.39616601300 *T)
    sumL1  +=       0.283 * cos(2.67802456636  + 3877.41693879520 *T)
    sumL1  +=       0.283 * cos(1.66293157090  + 1692.16566950240 *T)
    sumL1  +=       0.276 * cos(2.94210551399  + 3415.39402526710 *T)
    sumL1  +=       0.275 * cos(0.53418048945  + 17395.21973472580 *T)
    sumL1  +=       0.355 * cos(3.31406527401  + 10022.81760116760 *T)
    sumL1  +=       0.311 * cos(1.50310910269  + 6660.86953400080 *T)
    sumL1  +=       0.269 * cos(1.84517097065  + 11780.49035851620 *T)
    sumL1  +=       0.270 * cos(4.42425307819  + 310.84079886840 *T)
    sumL1  +=       0.275 * cos(3.58464612058  + 128.01884333740 *T)
    sumL1  +=       0.275 * cos(2.22523539580  + 3017.10701004240 *T)
    sumL1  +=       0.312 * cos(5.15950395287  + 7255.56965173440 *T)
    sumL1  +=       0.299 * cos(0.72552273097  + 155.35308913140 *T)
    sumL1  +=       0.353 * cos(5.70047798350  + 16460.33352952499 *T)
    sumL1  +=       0.267 * cos(5.97864271046  + 9499.25986200560 *T)
    sumL1  +=       0.270 * cos(0.77063210836  + 11236.57229942000 *T)
    sumL1  +=       0.339 * cos(3.36092148900  + 5625.36604155940 *T)
    sumL1  +=       0.315 * cos(2.33795159922  + 3281.23856478620 *T)
    sumL1  +=       0.247 * cos(3.71002922076  + 7373.38245462640 *T)
    sumL1  +=       0.328 * cos(0.18162415648  + 5618.31980486140 *T)
    sumL1  +=       0.247 * cos(6.27486009856  + 15508.61512327440 *T)
    sumL1  +=       0.292 * cos(0.14989609091  + 16304.91313009080 *T)
    sumL1  +=       0.326 * cos(4.53606745007  + 2178.13772229200 *T)
    sumL1  +=       0.286 * cos(5.47710043383  + 9168.64089834740 *T)
    sumL1  +=       0.246 * cos(1.49838712480  + 15110.46611986620 *T)
    sumL1  +=       0.262 * cos(2.58821936465  + 3336.73109134180 *T)
    sumL1  +=       0.244 * cos(0.84015413449  + 16062.18452611680 *T)
    sumL1  +=       0.245 * cos(0.37772563756  + 12721.57209941700 *T)
    sumL1  +=       0.250 * cos(2.26824758119  + 6784.31762761820 *T)
    sumL1  +=       0.248 * cos(6.22740483254  + 13149.15061136120 *T)
    sumL1  +=       0.255 * cos(4.93078809107  + 14158.74771361560 *T)
    sumL1  +=       0.240 * cos(6.15843594225  + 19800.94595622480 *T)
    sumL1  +=       0.249 * cos(5.47044926479  + 4407.10790388980 *T)
    sumL1  +=       0.235 * cos(5.38750866169  + 76.26607127560 *T)
    sumL1  +=       0.258 * cos(6.10384464886  + 2480.30249794700 *T)
    sumL1  +=       0.306 * cos(5.35546231697  + 2766.26762836500 *T)
    sumL1  +=       0.236 * cos(5.25670707064  + 13171.00144068760 *T)
    sumL1  +=       0.224 * cos(4.52466909993  + 12566.15169998280 *T)
    sumL1  +=       0.220 * cos(5.83694256642  + 13936.79450513400 *T)
    sumL1  +=       0.271 * cos(1.42460945147  + 14054.60730802600 *T)
    sumL1  +=       0.213 * cos(0.21127914063  + 1505.28780909299 *T)
    sumL1  +=       0.267 * cos(5.16501015011  + 3205.54734666440 *T)
    sumL1  +=       0.212 * cos(4.26202838353  + 6546.15977336420 *T)
    sumL1  +=       0.211 * cos(6.21401684263  + 3253.30422216000 *T)
    sumL1  +=       0.233 * cos(3.72007597749  + 3346.13535100720 *T)
    sumL1  +=       0.274 * cos(2.91986569135  + 10713.99488132620 *T)
    sumL1  +=       0.201 * cos(3.36695295492  + 6.68366387410 *T)
    sumL1  +=       0.223 * cos(3.08788599159  + 401.67212175720 *T)
    sumL1  +=       0.234 * cos(2.24268269202  + 110.20632121940 *T)
    sumL1  +=       0.264 * cos(2.08178742740  + 6475.03930496240 *T)
    sumL1  +=       0.231 * cos(4.53806384480  + 9602.35263622420 *T)
    sumL1  +=       0.213 * cos(2.85452302656  + 5415.65737477320 *T)
    sumL1  +=       0.195 * cos(0.99589439506  + 5642.19824260920 *T)
    sumL1  +=       0.259 * cos(0.00464351114  + 9380.95967271720 *T)
    sumL1  +=       0.197 * cos(3.32573550633  + 3657.00429635640 *T)
    sumL1  +=       0.228 * cos(5.33299975472  + 3561.02506913860 *T)
    sumL1  +=       0.193 * cos(1.25502846507  + 6606.44325483230 *T)
    sumL1  +=       0.199 * cos(1.13665869139  + 685.04405422600 *T)
    sumL1  +=       0.227 * cos(4.49610509002  + 589.06482700820 *T)
    sumL1  +=       0.217 * cos(5.48740879816  + 10596.18207843420 *T)
    sumL1  +=       0.192 * cos(4.26501800444  + 3333.56619000180 *T)
    sumL1  +=       0.188 * cos(1.44301618203  + 4885.96640967860 *T)
    sumL1  +=       0.178 * cos(4.82506490541  + 9070.11887384880 *T)
    sumL1  +=       0.184 * cos(5.69637552141  + 3351.24909204960 *T)
    sumL1  +=       0.187 * cos(0.76021337348  + 16699.53901514999 *T)
    sumL1  +=       0.226 * cos(0.82767654373  + 3265.83082813250 *T)
    sumL1  +=       0.204 * cos(6.20933387021  + 394.62588505920 *T)
    sumL1  +=       0.176 * cos(3.89567349231  + 10028.95082710020 *T)
    sumL1  +=       0.174 * cos(3.68843293982  + 735.87651353180 *T)
    sumL1  +=       0.173 * cos(2.44269377255  + 3603.69635007260 *T)
    sumL1  +=       0.177 * cos(1.24154853329  + 12722.55242048520 *T)
    sumL1  +=       0.184 * cos(4.77203925989  + 286.96236112060 *T)
    sumL1  +=       0.171 * cos(4.67140116008  + 20199.09495963300 *T)
    sumL1  +=       0.170 * cos(5.13753345526  + 1332.05488754080 *T)
    sumL1  +=       0.201 * cos(2.37863157745  + 16276.46394262300 *T)
    sumL1  +=       0.209 * cos(0.57156268506  + 11250.79939342160 *T)
    sumL1  +=       0.164 * cos(1.98441291396  + 10014.72373309860 *T)
    sumL1  +=       0.191 * cos(0.60250751218  + 56.80326216980 *T)
    sumL1  +=       0.171 * cos(6.22556266993  + 17277.40693183380 *T)
    sumL1  +=       0.166 * cos(1.05948008727  + 19513.98359510420 *T)
    sumL1  +=       0.163 * cos(1.59661610701  + 1437.17561419860 *T)
    sumL1  +=       0.165 * cos(3.36308723589  + 6665.97238221460 *T)
    sumL1  +=       0.184 * cos(3.20554894393  + 263.08392337280 *T)
    sumL1  +=       0.212 * cos(3.10485836003  + 4039.88357492740 *T)
    sumL1  +=       0.176 * cos(3.41768939214  + 9468.26787725700 *T)
    sumL1  +=       0.163 * cos(1.39275730949  + 8982.81066930900 *T)
    sumL1  +=       0.213 * cos(3.39734274482  + 931.36308685180 *T)
    sumL1  +=       0.189 * cos(4.54004144896  + 8542.97070603500 *T)
    sumL1  +=       0.191 * cos(1.15555618959  + 3169.93955608060 *T)
    sumL1  +=       0.155 * cos(1.41249963094  + 22.76849660940 *T)
    sumL1  +=       0.153 * cos(5.14168081601  + 156.40072050240 *T)
    sumL1  +=       0.159 * cos(3.64996617906  + 8013.27974094040 *T)
    sumL1  +=       0.151 * cos(1.93804487507  + 3384.33133900480 *T)
    sumL1  +=       0.157 * cos(0.58554505759  + 158.94351778320 *T)
    sumL1  +=       0.173 * cos(2.72517427493  + 2807.39834325620 *T)
    sumL1  +=       0.159 * cos(0.67192454133  + 13892.14067189380 *T)
    sumL1  +=       0.150 * cos(2.66045714174  + 19004.64794940840 *T)
    sumL1  +=       0.192 * cos(5.73782632783  + 206.70073729660 *T)
    sumL1  +=       0.143 * cos(3.19213280913  + 6843.69148953180 *T)
    sumL1  +=       0.194 * cos(1.32358882667  + 19402.79695281660 *T)
    sumL1  +=       0.143 * cos(2.36478163720  + 13207.02930736500 *T)
    sumL1  +=       0.140 * cos(1.88800568840  + 11766.26326451460 *T)
    sumL1  +=       0.144 * cos(0.69018080218  + 17085.95866572220 *T)
    sumL1  +=       0.183 * cos(5.98085295555  + 13362.51701710200 *T)
    sumL1  +=       0.161 * cos(2.92764155222  + 5.85720229960 *T)
    sumL1  +=       0.162 * cos(6.07051064413  + 6701.58017279840 *T)
    sumL1  +=       0.192 * cos(0.86266150575  + 2814.44457995420 *T)
    sumL1  +=       0.182 * cos(5.26446797092  + 3873.82651014340 *T)
    sumL1  +=       0.137 * cos(0.41563614709  + 5820.91492464680 *T)
    sumL1  +=       0.144 * cos(3.02314051168  + 708.98980227659 *T)
    sumL1  +=       0.184 * cos(4.61314496499  + 3329.97576135000 *T)
    sumL1  +=       0.131 * cos(3.48156082643  + 367.22432896240 *T)
    sumL1  +=       0.173 * cos(3.09922849765  + 12295.95422960920 *T)
    sumL1  +=       0.135 * cos(2.23311632892  + 15664.03552270859 *T)
    sumL1  +=       0.147 * cos(1.95810911154  + 5732.04924442980 *T)
    sumL1  +=       0.158 * cos(1.48909254724  + 29.49181830340 *T)
    sumL1  +=       0.127 * cos(5.55534080040  + 3368.01398279660 *T)
    sumL1  +=       0.129 * cos(1.78002583252  + 22743.40937951640 *T)
    sumL1  +=       0.132 * cos(2.81496895377  + 21795.21409161479 *T)
    sumL1  +=       0.127 * cos(5.73090203501  + 3340.19235060619 *T)
    sumL1  +=       0.164 * cos(1.87613918877  + 6709.67404086740 *T)
    sumL1  +=       0.123 * cos(3.61238958991  + 22324.90505670940 *T)
    sumL1  +=       0.129 * cos(4.92064308735  + 2540.79130153440 *T)
    sumL1  +=       0.121 * cos(6.16922638434  + 20206.14119633100 *T)
    sumL1  +=       0.122 * cos(5.79901866314  + 1854.63230563460 *T)
    sumL1  +=       0.133 * cos(0.50941998058  + 3274.12501778540 *T)
    sumL1  +=       0.151 * cos(1.61342807879  + 1107.13880568480 *T)
    sumL1  +=       0.165 * cos(2.02795177586  + 290.48547946960 *T)
    sumL1  +=       0.125 * cos(0.52719797619  + 2604.73591316800 *T)
    sumL1  +=       0.144 * cos(5.68526782434  + 8827.39026987480 *T)
    sumL1  +=       0.126 * cos(3.80246508251  + 765.79306444640 *T)
    sumL1  +=       0.116 * cos(1.79450246249  + 647.01083331480 *T)
    sumL1  +=       0.126 * cos(2.00195272473  + 699.27114822760 *T)
    sumL1  +=       0.147 * cos(6.22619740782  + 6040.34724601740 *T)
    sumL1  +=       0.119 * cos(2.05840518265  + 15121.10278521600 *T)
    sumL1  +=       0.114 * cos(2.74877091470  + 6460.81221096080 *T)
    sumL1  +=       0.155 * cos(1.78154091696  + 21265.52312652020 *T)
    sumL1  +=       0.146 * cos(3.37351237411  + 1861.74585263540 *T)
    sumL1  +=       0.118 * cos(4.07281676691  + 418.50432280700 *T)
    sumL1  +=       0.116 * cos(0.10434606071  + 13362.38239649640 *T)
    sumL1  +=       0.129 * cos(0.78419803719  + 3427.92063123960 *T)
    sumL1  +=       0.152 * cos(0.32620694442  + 3443.70520091840 *T)
    sumL1  +=       0.110 * cos(0.56398082486  + 661.23292678100 *T)
    sumL1  +=       0.111 * cos(4.05380946072  + 568.82187402740 *T)
    sumL1  +=       0.108 * cos(3.17700641574  + 3448.27595063840 *T)
    sumL1  +=       0.138 * cos(4.47698517191  + 3326.38533269820 *T)
    sumL1  +=       0.108 * cos(4.89922372003  + 9588.12554222260 *T)
    sumL1  +=       0.114 * cos(4.80828825403  + 6657.34641565180 *T)
    sumL1  +=       0.108 * cos(4.10637483972  + 13553.89797291080 *T)
    sumL1  +=       0.125 * cos(0.33573243959  + 18849.22754997420 *T)
    sumL1  +=       0.115 * cos(3.18885465852  + 2409.24933984800 *T)
    sumL1  +=       0.104 * cos(3.23074163851  + 3472.15438838620 *T)
    sumL1  +=       0.104 * cos(0.09799515047  + 30065.51184029820 *T)
    sumL1  +=       0.112 * cos(1.64487733528  + 10001.06188460700 *T)
    sumL1  +=       0.143 * cos(3.53781769283  + 6518.75821726740 *T)
    sumL1  +=       0.113 * cos(5.20979306912  + 2125.87740737920 *T)
    sumL1  +=       0.104 * cos(2.77582098882  + 38.13303563780 *T)
    sumL1  +=       0.133 * cos(5.88513337452  + 5835.14201864840 *T)
    sumL1  +=       0.105 * cos(4.11662579413  + 6675.70192909220 *T)
    sumL1  +=       0.102 * cos(0.60100887043  + 10264.56588407340 *T)
    sumL1  +=       0.101 * cos(3.78636130664  + 10042.61267559180 *T)
    sumL1  +=       0.139 * cos(1.80936944447  + 12323.42309600880 *T)
    sumL1  +=       0.101 * cos(2.47217208753  + 7380.49600162720 *T)
    sumL1  +=       0.130 * cos(2.53454569863  + 11769.85369316640 *T)
    sumL1  +=       0.100 * cos(5.72291104291  + 14.22709400160 *T)
    sumL1  +=       0.135 * cos(4.20237564510  + 4672.66731424060 *T)
    sumL1  +=       0.133 * cos(0.34413768012  + 16489.76303806100 *T)
    sumL1  +=       0.098 * cos(1.44874403589  + 3370.04193523580 *T)
    sumL1  +=       0.131 * cos(1.31336606248  + 3313.21087060300 *T)
    sumL1  +=       0.111 * cos(3.12463539337  + 309.27832265580 *T)
    sumL1  +=       0.101 * cos(3.15369992044  + 24150.08005134500 *T)
    sumL1  +=       0.102 * cos(6.13479937096  + 2277.70737816160 *T)
    sumL1  +=       0.099 * cos(0.10085261274  + 12839.87228870540 *T)
    sumL1  +=       0.134 * cos(2.91637947295  + 57.87869600380 *T)
    sumL1  +=       0.104 * cos(3.30283052330  + 3399.98628861340 *T)
    sumL1  +=       0.108 * cos(4.92699760221  + 802.36392244620 *T)
    sumL1  +=       0.106 * cos(2.89298330043  + 7799.98064550240 *T)
    sumL1  +=       0.112 * cos(3.12761163915  + 5989.06725217280 *T)
    sumL1  +=       0.094 * cos(3.42562596561  + 3510.19260983280 *T)
    sumL1  +=       0.102 * cos(0.94285421551  + 3209.07046501340 *T)
    sumL1  +=       0.096 * cos(0.79636181668  + 3024.22055704320 *T)
    sumL1  +=       0.093 * cos(1.08979608844  + 14577.18472611980 *T)
    sumL1  +=       0.097 * cos(3.56551535742  + 14421.83163698840 *T)
    sumL1  +=       0.092 * cos(5.02279637510  + 2494.52959194860 *T)
    sumL1  +=       0.123 * cos(5.83481108101  + 7747.72033058960 *T)
    sumL1  +=       0.111 * cos(3.31406538379  + 11610.91017538320 *T)
    sumL1  +=       0.118 * cos(5.00845322960  + 1581.95934828300 *T)
    sumL1  +=       0.091 * cos(2.36839307589  + 11140.59307220220 *T)
    sumL1  +=       0.099 * cos(2.93269536697  + 2067.93140107260 *T)
    sumL1  +=       0.124 * cos(3.15220420912  + 22345.26037610820 *T)
    sumL1  +=       0.090 * cos(0.85462370851  + 17232.75309859360 *T)
    sumL1  +=       0.125 * cos(1.54866979468  + 1435.14766175940 *T)
    sumL1  +=       0.113 * cos(4.57220387105  + 14712.31711645800 *T)
    sumL1  +=       0.103 * cos(0.11158194413  + 11.04570026390 *T)
    sumL1  +=       0.113 * cos(2.41110151532  + 5244.04923920100 *T)
    sumL1  +=       0.091 * cos(1.77184730280  + 2221.85663459700 *T)
    sumL1  +=       0.114 * cos(1.92696878615  + 8226.57883637840 *T)
    sumL1  +=       0.118 * cos(0.29322259611  + 2945.98654164060 *T)
    sumL1  +=       0.089 * cos(4.49875865671  + 21947.11137270000 *T)
    sumL1  +=       0.104 * cos(1.36766858693  + 4032.77002792660 *T)
    sumL1  +=       0.116 * cos(4.67476277278  + 272.67295735160 *T)
    sumL1  +=       0.105 * cos(3.73327973850  + 4989.05918389720 *T)
    sumL1  +=       0.095 * cos(0.34653808205  + 13517.80279593060 *T)
    sumL1  +=       0.086 * cos(2.70925448214  + 7218.02936549500 *T)
    sumL1  +=       0.112 * cos(5.27572276726  + 17499.36014031540 *T)
    sumL1  +=       0.114 * cos(3.37335971932  + 2938.94030494260 *T)
    sumL1  +=       0.090 * cos(0.26623024836  + 9485.03276800400 *T)
    sumL1  +=       0.086 * cos(3.66786582491  + 6997.61672305620 *T)
    sumL1  +=       0.087 * cos(4.07823608680  + 18606.49894600020 *T)
    sumL1  +=       0.117 * cos(5.61863859185  + 16511.61386738740 *T)
    sumL1  +=       0.084 * cos(2.80171829534  + 8584.66166590080 *T)
    sumL1  +=       0.086 * cos(5.87459621301  + 73.60095591520 *T)
    sumL1  +=       0.090 * cos(1.07789160729  + 10721.10842832700 *T)
    sumL1  +=       0.084 * cos(0.78729386249  + 146.81168652360 *T)
    sumL1  +=       0.112 * cos(2.31690430881  + 9638.94074787620 *T)
    sumL1  +=       0.083 * cos(5.90575201511  + 16858.41522263039 *T)
    sumL1  +=       0.082 * cos(4.31690557700  + 9374.82644678460 *T)
    sumL1  +=       0.083 * cos(1.41315204958  + 22854.59602180400 *T)
    sumL1  +=       0.084 * cos(0.90320571725  + 2171.02417529120 *T)
    sumL1  +=       0.082 * cos(1.59442952041  + 9947.05568153210 *T)
    sumL1  +=       0.082 * cos(0.44163602941  + 4782.87363546000 *T)
    sumL1  +=       0.086 * cos(5.18335054161  + 12410.73130054860 *T)
    sumL1  +=       0.085 * cos(3.61669636863  + 8965.97846825920 *T)
    sumL1  +=       0.097 * cos(5.35511765140  + 4996.17273089800 *T)
    sumL1  +=       0.085 * cos(2.36814442737  + 1062.90504853820 *T)
    sumL1  +=       0.078 * cos(1.75580354602  + 2060.81785407180 *T)
    sumL1  +=       0.087 * cos(0.98702744399  + 10156.90236013480 *T)
    sumL1  +=       0.098 * cos(0.35294347682  + 23546.75362303080 *T)
    sumL1  +=       0.101 * cos(1.03298143418  + 20040.15144184979 *T)
    sumL1  +=       0.088 * cos(0.56201084357  + 17101.21113690720 *T)
    sumL1  +=       0.082 * cos(2.65765057749  + 12979.55317457600 *T)
    sumL1  +=       0.085 * cos(1.07007237991  + 10706.88133432540 *T)
    sumL1  +=       0.082 * cos(0.80588396320  + 20735.83216142559 *T)
    sumL1  +=       0.092 * cos(4.87180501294  + 9389.05354078620 *T)
    sumL1  +=       0.078 * cos(3.22869264518  + 11925.27409260060 *T)
    sumL1  +=       0.074 * cos(5.59171946181  + 8535.85715903420 *T)
    sumL1  +=       0.080 * cos(0.42260849968  + 956.28915597060 *T)
    sumL1  +=       0.082 * cos(1.47379060963  + 16.83220104979 *T)
    sumL1  +=       0.073 * cos(3.76950560688  + 8859.36257569160 *T)
    sumL1  +=       0.073 * cos(4.59130182630  + 362.12113673080 *T)
    sumL1  +=       0.072 * cos(5.15373872266  + 9872.27408296480 *T)
    sumL1  +=       0.098 * cos(2.47740242208  + 5401.43028077160 *T)
    sumL1  +=       0.071 * cos(0.08395696279  + 15849.86575174700 *T)
    sumL1  +=       0.078 * cos(0.04212599783  + 16703.12944380180 *T)
    sumL1  +=       0.075 * cos(3.77172360793  + 4845.90023579280 *T)
    sumL1  +=       0.072 * cos(2.01742377451  + 1329.51209026000 *T)
    sumL1  +=       0.076 * cos(0.30413402871  + 72.05573335480 *T)
    sumL1  +=       0.097 * cos(5.15109318550  + 1.48447270830 *T)
    sumL1  +=       0.070 * cos(0.33648335823  + 23141.55838292460 *T)
    sumL1  +=       0.068 * cos(0.34828385806  + 20047.19767854780 *T)
    sumL1  +=       0.078 * cos(1.42772075938  + 16063.16484718500 *T)
    sumL1  +=       0.081 * cos(2.37498404818  + 16703.04487984680 *T)
    sumL1  +=       0.069 * cos(3.66727984195  + 13363.43002786740 *T)
    sumL1  +=       0.067 * cos(1.77194706681  + 18451.07854656599 *T)
    sumL1  +=       0.073 * cos(0.41181711796  + 33406.12426699800 *T)
    sumL1  +=       0.067 * cos(3.15946840380  + 8799.98871377800 *T)
    sumL1  +=       0.067 * cos(4.07602260745  + 10448.43547097540 *T)
    sumL1  +=       0.068 * cos(1.78981361818  + 224.34479570190 *T)
    sumL1  +=       0.066 * cos(1.01449371817  + 4193.80880845180 *T)
    sumL1  +=       0.067 * cos(4.89249339125  + 12082.65513417120 *T)
    sumL1  +=       0.081 * cos(0.61914094848  + 949.17560896980 *T)
    sumL1  +=       0.066 * cos(3.39914635235  + 10184.30391623160 *T)
    sumL1  +=       0.064 * cos(5.50528849889  + 4936.79886898440 *T)
    sumL1  +=       0.064 * cos(1.82608235490  + 3077.52850332700 *T)
    sumL1  +=       0.069 * cos(2.48571574894  + 10018.24685144760 *T)
    sumL1  +=       0.064 * cos(2.80447322320  + 20809.46762464520 *T)
    sumL1  +=       0.068 * cos(0.97907335430  + 6717.25272007700 *T)
    sumL1  +=       0.087 * cos(2.45966764758  + 7321.12213971360 *T)
    sumL1  +=       0.062 * cos(1.69131771765  + 632.78373931320 *T)
    sumL1  +=       0.081 * cos(4.40575713075  + 16703.07938715119 *T)
    sumL1  +=       0.063 * cos(3.63625395496  + 25685.87280280800 *T)
    sumL1  +=       0.075 * cos(5.54907590704  + 16872.64231663200 *T)
    sumL1  +=       0.061 * cos(0.33159827734  + 12012.58229714040 *T)
    sumL1  +=       0.063 * cos(3.16314818302  + 11670.28403729680 *T)
    sumL1  +=       0.062 * cos(0.23148800541  + 7314.00859271280 *T)
    sumL1  +=       0.078 * cos(1.65377731167  + 11614.43329373220 *T)
    sumL1  +=       0.068 * cos(1.10866475394  + 6155.05700665400 *T)
    sumL1  +=       0.068 * cos(4.45713540461  + 9175.75444534820 *T)
    sumL1  +=       0.081 * cos(3.66553577428  + 15265.88651930040 *T)
    sumL1  +=       0.068 * cos(5.59792852411  + 377.37360791580 *T)
    sumL1  +=       0.059 * cos(6.25689995147  + 41427.48698317880 *T)
    sumL1  +=       0.072 * cos(3.39739853142  + 685.11136452880 *T)
    sumL1  +=       0.062 * cos(0.11206359088  + 14047.49376102520 *T)
    sumL1  +=       0.057 * cos(4.31819630987  + 95.97922721780 *T)
    sumL1  +=       0.070 * cos(5.98596344975  + 6531.66165626500 *T)
    sumL1  +=       0.066 * cos(5.93804470886  + 3490.17562383440 *T)
    sumL1  +=       0.056 * cos(5.11205539684  + 7322.10246078180 *T)
    sumL1  +=       0.056 * cos(3.36788837326  + 4379.63903749020 *T)
    sumL1  +=       0.056 * cos(2.20908914878  + 6688.27109009760 *T)
    sumL1  +=       0.057 * cos(4.57828186723  + 10037.08975128440 *T)
    sumL1  +=       0.057 * cos(0.04695703833  + 24889.57479599160 *T)
    sumL1  +=       0.065 * cos(2.70973517401  + 19617.07636932279 *T)
    sumL1  +=       0.068 * cos(3.51423189318  + 14556.89671702380 *T)
    sumL1  +=       0.056 * cos(5.94452825242  + 21548.96236929180 *T)
    sumL1  +=       0.056 * cos(3.94541346495  + 25665.51748340920 *T)
    sumL1  +=       0.055 * cos(0.45166365461  + 16702.99482319620 *T)
    sumL1  +=       0.062 * cos(0.05229160039  + 14061.72085502680 *T)
    sumL1  +=       0.055 * cos(1.14774475393  + 4459.36821880260 *T)
    sumL1  +=       0.065 * cos(6.21442820689  + 6947.83189512140 *T)
    sumL1  +=       0.056 * cos(5.01712943722  + 485.97205278960 *T)
    sumL1  +=       0.065 * cos(1.51615558729  + 5511.63660199100 *T)
    sumL1  +=       0.072 * cos(4.90742373357  + 25287.72379939980 *T)
    sumL1  +=       0.057 * cos(5.31542907454  + 28628.33622609960 *T)
    sumL1  +=       0.055 * cos(4.02023620788  + 4005.36847182980 *T)
    sumL1  +=       0.054 * cos(5.51196184505  + 44.65383324020 *T)
    sumL1  +=       0.053 * cos(0.46840976995  + 19645.52555679060 *T)
    sumL1  +=       0.053 * cos(4.51539970261  + 9886.77220006400 *T)
    sumL1  +=       0.069 * cos(3.37092084432  + 41.55079098480 *T)
    sumL1  +=       0.072 * cos(4.66520155151  + 6685.10618875760 *T)
    sumL1  +=       0.052 * cos(2.05209599225  + 7366.26890762560 *T)
    sumL1  +=       0.052 * cos(6.25882226723  + 12509.25332504720 *T)
    sumL1  +=       0.060 * cos(0.33350021257  + 13575.74880223720 *T)
    sumL1  +=       0.059 * cos(5.94433755684  + 625.67019231240 *T)
    sumL1  +=       0.050 * cos(2.65624501709  + 10124.93005431800 *T)
    sumL1  +=       0.068 * cos(2.92105834159  + 10025.42770875120 *T)
    sumL1  +=       0.051 * cos(2.88289068067  + 39601.89191244960 *T)
    sumL1  +=       0.069 * cos(1.09637075565  + 24076.44458812540 *T)
    sumL1  +=       0.052 * cos(2.65290577481  + 15650.79375031060 *T)
    sumL1  +=       0.052 * cos(3.28683313010  + 6756.00645196690 *T)
    sumL1  +=       0.054 * cos(5.34683925496  + 6578.13207918100 *T)
    sumL1  +=       0.054 * cos(2.77410607136  + 6705.10329114740 *T)
    sumL1  +=       0.049 * cos(2.65700209900  + 6944.30877677240 *T)
    sumL1  +=       0.053 * cos(5.67436244967  + 1883.08149310240 *T)
    sumL1  +=       0.048 * cos(5.68336050679  + 8646.06348025360 *T)
    sumL1  +=       0.051 * cos(0.11390643300  + 17402.33328172660 *T)
    sumL1  +=       0.065 * cos(1.03099992649  + 4106.40549114620 *T)
    sumL1  +=       0.060 * cos(1.00159365247  + 151.89728108520 *T)
    sumL1  +=       0.050 * cos(3.84651247899  + 45494.58142974879 *T)
    sumL1  +=       0.047 * cos(3.03959709242  + 5408.54382777240 *T)
    sumL1  +=       0.063 * cos(4.16165369755  + 8186.51266249260 *T)
    sumL1  +=       0.046 * cos(2.69368087378  + 16547.64173406480 *T)
    sumL1  +=       0.051 * cos(2.99576014378  + 3774.32416457660 *T)
    sumL1  +=       0.044 * cos(2.00664763411  + 6418.14093002680 *T)
    sumL1  +=       0.045 * cos(4.01853755929  + 19406.67828817460 *T)
    sumL1  +=       0.058 * cos(3.14474753550  + 4025.65648092580 *T)
    sumL1  +=       0.050 * cos(2.59881540437  + 6621.85099148600 *T)
    sumL1  +=       0.043 * cos(4.87912487459  + 6414.61781167780 *T)
    sumL1  +=       0.042 * cos(5.20400092044  + 4447.75123238460 *T)
    sumL1  +=       0.051 * cos(1.99634375899  + 5032.77809620220 *T)
    sumL1  +=       0.043 * cos(1.28813888865  + 6643.09181776180 *T)
    sumL1  +=       0.040 * cos(0.96801618560  + 14591.41182012140 *T)
    sumL1  +=       0.039 * cos(1.84985100829  + 10001.48196070061 *T)
    sumL1  +=       0.039 * cos(5.69967200167  + 6106.88005506480 *T)
    sumL1  +=       0.038 * cos(3.27498743518  + 18052.92954315780 *T)
    sumL1  +=       0.039 * cos(2.84167905068  + 6652.77566593180 *T)
    sumL1  +=       0.044 * cos(0.57891618854  + 16865.52876963120 *T)
    sumL1  +=       0.043 * cos(4.61937364869  + 3341.03250279340 *T)
    sumL1  +=       0.042 * cos(6.02555835659  + 6691.86151874940 *T)
    sumL1  +=       0.034 * cos(4.97734992350  + 6670.58818804980 *T)
    sumL1  +=       0.033 * cos(1.39167727215  + 4825.54491639400 *T)
    sumL1  +=       0.035 * cos(6.02955363644  + 3568.08855948880 *T)
    sumL1  +=       0.035 * cos(0.31961016732  + 6645.19698672220 *T)
    sumL1  +=       0.032 * cos(5.63043769073  + 3511.28529731900 *T)
    sumL1  +=       0.031 * cos(5.42978464210  + 9945.57120882380 *T)
    sumL1  +=       0.038 * cos(5.66461657503  + 3416.87849797540 *T)
    sumL1  +=       0.030 * cos(0.98518793666  + 20426.57109242200 *T)
    sumL1  +=       0.038 * cos(0.12870962242  + 6604.95878212400 *T)
    sumL1  +=       0.037 * cos(5.48374357342  + 3311.18291816379 *T)
    sumL1  +=       0.032 * cos(6.11106979810  + 4392.88080988820 *T)
    sumL1  +=       0.031 * cos(3.18481282781  + 3341.04230982650 *T)
    sumL1  +=       0.034 * cos(2.32358226279  + 9072.66167112960 *T)
    sumL1  +=       0.039 * cos(4.11042361929  + 3312.16323923200 *T)
    sumL1  +=       0.027 * cos(0.57810321636  + 3391.89276456221 *T)
    sumL1  +=       0.029 * cos(2.48646403166  + 9815.65173166220 *T)
    sumL1  +=       0.031 * cos(0.44265747667  + 3451.79906898740 *T)
    sumL1  +=       0.027 * cos(6.13498177783  + 3362.46325602620 *T)
    sumL1  +=       0.027 * cos(6.21846173482  + 5223.69391980220 *T)
    sumL1  +=       0.027 * cos(2.94945830517  + 7203.80227149340 *T)
    sumL1  +=       0.027 * cos(3.26179855800  + 8756.26980147300 *T)
    sumL1  +=       0.027 * cos(3.94385271700  + 23958.63178523340 *T)
    sumL1  +=       0.033 * cos(3.77237326006  + 12808.88030395680 *T)
    sumL1  +=       0.030 * cos(4.75096367323  + 15906.76412668260 *T)
    sumL1  +=       0.031 * cos(0.88248871193  + 3340.18254357310 *T)
    sumL1  +=       0.025 * cos(0.31303295413  + 6571.01853218020 *T)
    sumL1  +=       0.031 * cos(4.29076841627  + 10020.85695903120 *T)
    sumL1  +=       0.026 * cos(2.22427360058  + 10050.28646756720 *T)
    sumL1  +=       0.025 * cos(0.67881122439  + 23937.85638974100 *T)
    sumL1  +=       0.031 * cos(1.72899093511  + 13745.34623902240 *T)
    sumL1  +=       0.024 * cos(0.20355912395  + 3229.42578441220 *T)
    sumL1  +=       0.032 * cos(3.37195631109  + 2284.75361485960 *T)

    # MARS Longitude
    sumL2  = 0
    sumL2  +=   58015.791 * cos(2.04979463279  + 3340.61242669980 *T)
    sumL2  +=   54187.645 * cos(0.00000000000  + 0.00000000000 *T)
    sumL2  +=   13908.426 * cos(2.45742359888  + 6681.22485339960 *T)
    sumL2  +=    2465.104 * cos(2.80000020929  + 10021.83728009940 *T)
    sumL2  +=     398.379 * cos(3.14118428289  + 13362.44970679920 *T)
    sumL2  +=     222.022 * cos(3.19436080019  + 3.52311834900 *T)
    sumL2  +=     120.957 * cos(0.54325292454  + 155.42039943420 *T)
    sumL2  +=      61.517 * cos(3.48529427371  + 16703.06213349900 *T)
    sumL2  +=      53.638 * cos(3.54191121461  + 3344.13554504880 *T)
    sumL2  +=      34.268 * cos(6.00188499119  + 2281.23049651060 *T)
    sumL2  +=      31.665 * cos(4.14015171788  + 191.44826611160 *T)
    sumL2  +=      29.839 * cos(1.99870679845  + 796.29800681640 *T)
    sumL2  +=      23.168 * cos(4.33403365928  + 242.72860397400 *T)
    sumL2  +=      21.659 * cos(3.44532466378  + 398.14900340820 *T)
    sumL2  +=      16.044 * cos(6.11000472441  + 2146.16541647520 *T)
    sumL2  +=      20.370 * cos(5.42191375400  + 553.56940284240 *T)
    sumL2  +=      14.927 * cos(6.09541783564  + 3185.19202726560 *T)
    sumL2  +=      16.227 * cos(0.65678953303  + 0.98032106820 *T)
    sumL2  +=      14.317 * cos(2.61851897591  + 1349.86740965880 *T)
    sumL2  +=      14.416 * cos(4.01923812101  + 951.71840625060 *T)
    sumL2  +=      11.934 * cos(3.86122163021  + 6684.74797174860 *T)
    sumL2  +=      15.648 * cos(1.22086121940  + 1748.01641306700 *T)
    sumL2  +=      11.260 * cos(4.71822363671  + 2544.31441988340 *T)
    sumL2  +=      13.352 * cos(0.60189008414  + 1194.44701022460 *T)
    sumL2  +=      10.396 * cos(0.25038714677  + 382.89653222320 *T)
    sumL2  +=       9.468 * cos(0.68170713564  + 1059.38193018920 *T)
    sumL2  +=       9.229 * cos(3.83209092321  + 20043.67456019880 *T)
    sumL2  +=       9.005 * cos(3.88271826102  + 3738.76143010800 *T)
    sumL2  +=       7.501 * cos(5.46498630412  + 1751.53953141600 *T)
    sumL2  +=       6.497 * cos(5.47773072872  + 1592.59601363280 *T)
    sumL2  +=       6.311 * cos(2.34104793674  + 3097.88382272579 *T)
    sumL2  +=       6.859 * cos(2.57522504136  + 3149.16416058820 *T)
    sumL2  +=       5.870 * cos(1.14783576679  + 7.11354700080 *T)
    sumL2  +=       6.681 * cos(2.37843690339  + 4136.91043351620 *T)
    sumL2  +=       4.647 * cos(4.42957708526  + 6151.53388830500 *T)
    sumL2  +=       4.166 * cos(3.68631477611  + 5614.72937620960 *T)
    sumL2  +=       4.764 * cos(2.89684755585  + 3333.49887969900 *T)
    sumL2  +=       4.045 * cos(6.12493402657  + 5628.95647021120 *T)
    sumL2  +=       3.653 * cos(4.06679068397  + 1990.74501704100 *T)
    sumL2  +=       3.618 * cos(2.46868561769  + 529.69096509460 *T)
    sumL2  +=       3.277 * cos(0.68101740787  + 8962.45534991020 *T)
    sumL2  +=       3.253 * cos(2.79565340390  + 3894.18182954220 *T)
    sumL2  +=       3.091 * cos(4.56861203364  + 3496.03282613400 *T)
    sumL2  +=       2.921 * cos(5.41458945995  + 2914.01423582380 *T)
    sumL2  +=       2.921 * cos(1.23050883841  + 2787.04302385740 *T)
    sumL2  +=       2.784 * cos(1.38911141844  + 4292.33083295040 *T)
    sumL2  +=       2.620 * cos(1.04061894134  + 3341.59274776800 *T)
    sumL2  +=       2.888 * cos(3.41062353663  + 3337.08930835080 *T)
    sumL2  +=       2.418 * cos(0.96341462666  + 4535.05943692440 *T)
    sumL2  +=       2.357 * cos(4.84628239765  + 9492.14631500480 *T)
    sumL2  +=       2.593 * cos(5.74934234498  + 3340.59517304760 *T)
    sumL2  +=       2.191 * cos(3.26449527357  + 213.29909543800 *T)
    sumL2  +=       2.594 * cos(1.49510566123  + 3340.62968035200 *T)
    sumL2  +=       2.344 * cos(4.18104725028  + 10025.36039844840 *T)
    sumL2  +=       2.630 * cos(4.67640929857  + 3583.34103067380 *T)
    sumL2  +=       2.602 * cos(2.64911714813  + 2388.89402044920 *T)
    sumL2  +=       1.830 * cos(0.97181050149  + 1589.07289528380 *T)
    sumL2  +=       2.416 * cos(1.04749173375  + 4399.99435688900 *T)
    sumL2  +=       2.386 * cos(4.27072575550  + 7079.37385680780 *T)
    sumL2  +=       2.187 * cos(0.16036551231  + 6525.80445396540 *T)
    sumL2  +=       2.344 * cos(0.01425578204  + 4690.47983635860 *T)
    sumL2  +=       1.617 * cos(4.95614491689  + 5088.62883976680 *T)
    sumL2  +=       1.633 * cos(1.10703599922  + 12303.06777661000 *T)
    sumL2  +=       2.126 * cos(0.48290227706  + 2700.71514038580 *T)
    sumL2  +=       1.629 * cos(4.94267977718  + 1221.84856632140 *T)
    sumL2  +=       1.504 * cos(0.11031912519  + 2957.71589447660 *T)
    sumL2  +=       1.759 * cos(3.81170701376  + 3723.50895892300 *T)
    sumL2  +=       1.401 * cos(3.85907867678  + 6283.07584999140 *T)
    sumL2  +=       1.338 * cos(5.29685392418  + 6677.70173505060 *T)
    sumL2  +=       1.763 * cos(2.51660121293  + 2810.92146160520 *T)
    sumL2  +=       1.392 * cos(2.73498041122  + 7477.52286021600 *T)
    sumL2  +=       1.431 * cos(2.97747408389  + 6489.77658728800 *T)
    sumL2  +=       1.236 * cos(3.77245965590  + 2699.73481931760 *T)
    sumL2  +=       1.234 * cos(1.88931735265  + 6681.24210705180 *T)
    sumL2  +=       1.513 * cos(2.92614134711  + 640.87760738220 *T)
    sumL2  +=       1.234 * cos(6.14168429036  + 6681.20759974740 *T)
    sumL2  +=       1.408 * cos(1.54395721611  + 3347.72597370060 *T)
    sumL2  +=       1.038 * cos(5.82880072933  + 4933.20844033260 *T)
    sumL2  +=       1.156 * cos(1.50825464304  + 426.59819087600 *T)
    sumL2  +=       1.362 * cos(4.17794297520  + 23384.28698689860 *T)
    sumL2  +=       1.135 * cos(3.77506455273  + 3870.30339179440 *T)
    sumL2  +=       0.916 * cos(3.81726339298  + 5092.15195811580 *T)
    sumL2  +=       0.853 * cos(3.82520490669  + 3340.54511639700 *T)
    sumL2  +=       1.077 * cos(5.05062828760  + 5621.84292321040 *T)
    sumL2  +=       1.074 * cos(3.81446920470  + 3553.91152213780 *T)
    sumL2  +=       0.847 * cos(3.41702696402  + 3340.67973700260 *T)
    sumL2  +=       0.920 * cos(1.91108056416  + 3532.06069281140 *T)
    sumL2  +=       0.738 * cos(4.25786145387  + 9623.68827669120 *T)
    sumL2  +=       0.908 * cos(4.12911006922  + 162.46663613220 *T)
    sumL2  +=       0.647 * cos(3.10301033831  + 7234.79425624200 *T)
    sumL2  +=       0.800 * cos(5.20674574801  + 87.30820453981 *T)
    sumL2  +=       0.657 * cos(1.57895580467  + 2487.41604494780 *T)
    sumL2  +=       0.650 * cos(2.78932995437  + 574.34479833480 *T)
    sumL2  +=       0.659 * cos(5.16655918817  + 12832.75874170460 *T)
    sumL2  +=       0.712 * cos(5.79288230676  + 3339.63210563160 *T)
    sumL2  +=       0.660 * cos(0.25125103909  + 8969.56889691100 *T)
    sumL2  +=       0.527 * cos(4.62218528897  + 10419.98628350760 *T)
    sumL2  +=       0.607 * cos(3.84724721085  + 5486.77784317500 *T)
    sumL2  +=       0.500 * cos(4.66323134619  + 6836.64525283380 *T)
    sumL2  +=       0.509 * cos(0.32548381735  + 8031.09226305840 *T)
    sumL2  +=       0.613 * cos(1.72250879737  + 7632.94325965020 *T)
    sumL2  +=       0.516 * cos(1.37906978509  + 7740.60678358880 *T)
    sumL2  +=       0.469 * cos(1.31324778369  + 7875.67186362420 *T)
    sumL2  +=       0.442 * cos(5.34515135225  + 10018.31416175040 *T)
    sumL2  +=       0.496 * cos(4.49656852602  + 692.15760122680 *T)
    sumL2  +=       0.544 * cos(1.44465369200  + 15643.68020330980 *T)
    sumL2  +=       0.467 * cos(1.40928870138  + 6682.20517446780 *T)
    sumL2  +=       0.547 * cos(4.42021065522  + 5331.35744374080 *T)
    sumL2  +=       0.433 * cos(1.25048504108  + 4929.68532198360 *T)
    sumL2  +=       0.409 * cos(3.80689273098  + 6681.29216370240 *T)
    sumL2  +=       0.403 * cos(1.57190239310  + 6127.65545055720 *T)
    sumL2  +=       0.401 * cos(4.59976459753  + 7210.91581849420 *T)
    sumL2  +=       0.406 * cos(0.60945437905  + 11773.37681151540 *T)
    sumL2  +=       0.367 * cos(0.97726583907  + 6041.32756708560 *T)
    sumL2  +=       0.361 * cos(0.13117924893  + 639.89728631400 *T)
    sumL2  +=       0.411 * cos(4.21323421517  + 6681.15754309680 *T)
    sumL2  +=       0.441 * cos(4.49050100878  + 13365.97282514820 *T)
    sumL2  +=       0.474 * cos(5.18562524130  + 2301.58581590939 *T)
    sumL2  +=       0.334 * cos(2.21001303889  + 10021.85453375160 *T)
    sumL2  +=       0.370 * cos(5.02880065186  + 6923.95345737360 *T)
    sumL2  +=       0.333 * cos(0.18229887483  + 10021.82002644720 *T)
    sumL2  +=       0.342 * cos(2.49988747611  + 6438.49624942560 *T)
    sumL2  +=       0.320 * cos(1.32374476890  + 2118.76386037840 *T)
    sumL2  +=       0.330 * cos(2.98027481579  + 5729.50644714900 *T)
    sumL2  +=       0.340 * cos(0.51409045792  + 9866.41688066520 *T)
    sumL2  +=       0.283 * cos(4.03219455446  + 8955.34180290940 *T)
    sumL2  +=       0.281 * cos(3.33891891806  + 9830.38901398780 *T)
    sumL2  +=       0.271 * cos(1.53144358045  + 1039.02661079040 *T)
    sumL2  +=       0.252 * cos(3.04356928941  + 10818.13528691580 *T)
    sumL2  +=       0.270 * cos(2.22972724035  + 5828.02847164760 *T)
    sumL2  +=       0.259 * cos(4.24406546278  + 6894.52394883760 *T)
    sumL2  +=       0.237 * cos(5.07818982743  + 8429.24126646660 *T)
    sumL2  +=       0.263 * cos(2.00552313665  + 3767.21061757580 *T)
    sumL2  +=       0.241 * cos(4.06396704332  + 7064.12138562280 *T)
    sumL2  +=       0.250 * cos(0.45422818547  + 6298.32832117640 *T)
    sumL2  +=       0.240 * cos(1.94655459341  + 6688.33840040040 *T)
    sumL2  +=       0.209 * cos(0.03043017984  + 3914.95722503460 *T)
    sumL2  +=       0.218 * cos(4.53201818250  + 26724.89941359840 *T)
    sumL2  +=       0.246 * cos(3.85272742042  + 4562.46099302120 *T)
    sumL2  +=       0.254 * cos(1.19313236635  + 8432.76438481560 *T)
    sumL2  +=       0.231 * cos(1.70340106651  + 2942.46342329160 *T)
    sumL2  +=       0.209 * cos(3.82345999055  + 6467.92575796160 *T)
    sumL2  +=       0.175 * cos(3.26444055581  + 3981.49003408200 *T)
    sumL2  +=       0.210 * cos(3.55052707697  + 10575.40668294180 *T)
    sumL2  +=       0.162 * cos(5.95176683701  + 310.84079886840 *T)
    sumL2  +=       0.165 * cos(5.04267055142  + 10177.25767953360 *T)
    sumL2  +=       0.198 * cos(0.80464315638  + 15113.98923821520 *T)
    sumL2  +=       0.188 * cos(1.99007233842  + 3127.31333126180 *T)
    sumL2  +=       0.188 * cos(4.55406803143  + 12964.30070339100 *T)
    sumL2  +=       0.163 * cos(2.14874886056  + 10973.55568635000 *T)
    sumL2  +=       0.154 * cos(4.35708331036  + 3360.96774609859 *T)
    sumL2  +=       0.169 * cos(2.40504327781  + 3355.86489788480 *T)
    sumL2  +=       0.164 * cos(5.53875620970  + 16173.37116840440 *T)
    sumL2  +=       0.144 * cos(0.30186831602  + 1903.43681250120 *T)
    sumL2  +=       0.161 * cos(3.18977924032  + 6674.11130639880 *T)
    sumL2  +=       0.135 * cos(1.65570006128  + 3325.35995551480 *T)
    sumL2  +=       0.143 * cos(6.25825818399  + 9595.23908922340 *T)
    sumL2  +=       0.158 * cos(6.14255608340  + 8273.82086703240 *T)
    sumL2  +=       0.170 * cos(5.98413937993  + 3320.25710730100 *T)
    sumL2  +=       0.124 * cos(1.74571336137  + 11081.21921028860 *T)
    sumL2  +=       0.146 * cos(2.22851709304  + 3178.14579056760 *T)
    sumL2  +=       0.126 * cos(5.02927593525  + 3475.67750673520 *T)
    sumL2  +=       0.123 * cos(4.99342648375  + 13760.59871020740 *T)
    sumL2  +=       0.141 * cos(4.56735815490  + 10021.76996979660 *T)
    sumL2  +=       0.118 * cos(5.42945437851  + 13358.92658845020 *T)
    sumL2  +=       0.116 * cos(4.22080571309  + 5884.92684658320 *T)
    sumL2  +=       0.119 * cos(4.09962692144  + 10021.90459040220 *T)
    sumL2  +=       0.124 * cos(1.79723243306  + 18984.29263000960 *T)
    sumL2  +=       0.135 * cos(1.00085140609  + 18454.60166491500 *T)
    sumL2  +=       0.106 * cos(5.76342924005  + 2288.34404351140 *T)
    sumL2  +=       0.103 * cos(0.63812052706  + 12310.18132361080 *T)
    sumL2  +=       0.100 * cos(4.37946945260  + 9808.53818466140 *T)
    sumL2  +=       0.104 * cos(4.01151451052  + 13916.01910964160 *T)
    sumL2  +=       0.125 * cos(4.94825577002  + 10551.52824519400 *T)
    sumL2  +=       0.098 * cos(5.91086068785  + 14584.29827312060 *T)
    sumL2  +=       0.126 * cos(1.55752034021  + 8270.29774868340 *T)
    sumL2  +=       0.098 * cos(2.52869810530  + 3191.04922956520 *T)
    sumL2  +=       0.095 * cos(5.50808311620  + 3319.83703120740 *T)
    sumL2  +=       0.108 * cos(1.67115668669  + 11216.28429032400 *T)
    sumL2  +=       0.090 * cos(1.71281301325  + 1505.28780909299 *T)
    sumL2  +=       0.088 * cos(0.42472333366  + 11371.70468975820 *T)
    sumL2  +=       0.089 * cos(4.83867235269  + 16706.58525184800 *T)
    sumL2  +=       0.106 * cos(1.78998252916  + 10022.81760116760 *T)
    sumL2  +=       0.087 * cos(5.89602688150  + 16858.48253293320 *T)
    sumL2  +=       0.085 * cos(0.55423657166  + 13362.43245314700 *T)
    sumL2  +=       0.091 * cos(1.61117547913  + 24093.27678917520 *T)
    sumL2  +=       0.083 * cos(0.55328893146  + 3344.20285535160 *T)
    sumL2  +=       0.083 * cos(1.06830368571  + 3364.49086444760 *T)
    sumL2  +=       0.084 * cos(6.22980173043  + 3369.06161416760 *T)
    sumL2  +=       0.085 * cos(2.58442459869  + 13362.46696045140 *T)
    sumL2  +=       0.079 * cos(3.08636079495  + 2818.03500860600 *T)
    sumL2  +=       0.077 * cos(1.53791408869  + 5459.37628707820 *T)
    sumL2  +=       0.075 * cos(3.34890581175  + 9070.11887384880 *T)
    sumL2  +=       0.097 * cos(2.99973987655  + 3316.73398895200 *T)
    sumL2  +=       0.084 * cos(5.55309637085  + 3427.92063123960 *T)
    sumL2  +=       0.089 * cos(4.16432726564  + 6158.64743530580 *T)
    sumL2  +=       0.078 * cos(3.72375730390  + 13171.00144068760 *T)
    sumL2  +=       0.083 * cos(0.11917380846  + 3337.02199804800 *T)
    sumL2  +=       0.073 * cos(4.91632010974  + 13362.38239649640 *T)
    sumL2  +=       0.076 * cos(4.52961122356  + 708.98980227659 *T)
    sumL2  +=       0.071 * cos(5.37905772348  + 10264.56588407340 *T)
    sumL2  +=       0.084 * cos(1.93007660929  + 9468.26787725700 *T)
    sumL2  +=       0.073 * cos(5.28498987702  + 4845.90023579280 *T)
    sumL2  +=       0.080 * cos(4.81193020727  + 13149.15061136120 *T)
    sumL2  +=       0.076 * cos(4.77873149036  + 8671.96987044060 *T)
    sumL2  +=       0.069 * cos(0.86312560839  + 13207.02930736500 *T)
    sumL2  +=       0.070 * cos(1.22001285505  + 2274.11694950980 *T)
    sumL2  +=       0.070 * cos(0.53977528244  + 13119.72110282519 *T)
    sumL2  +=       0.070 * cos(0.09801356525  + 1437.17561419860 *T)
    sumL2  +=       0.069 * cos(2.23597403243  + 7107.82304427560 *T)
    sumL2  +=       0.074 * cos(0.67876411085  + 16460.33352952499 *T)
    sumL2  +=       0.068 * cos(0.54393617190  + 12935.85151592320 *T)
    sumL2  +=       0.084 * cos(5.11617356640  + 13892.14067189380 *T)
    sumL2  +=       0.078 * cos(5.27206373031  + 853.19638175200 *T)
    sumL2  +=       0.065 * cos(4.74504626032  + 6144.42034130420 *T)
    sumL2  +=       0.065 * cos(2.89744951086  + 7903.07341972100 *T)
    sumL2  +=       0.089 * cos(5.44755326514  + 13517.87010623340 *T)
    sumL2  +=       0.073 * cos(3.81554166604  + 3503.07906283200 *T)
    sumL2  +=       0.066 * cos(0.47312197978  + 3980.50971301380 *T)
    sumL2  +=       0.065 * cos(1.38901912957  + 3253.30422216000 *T)
    sumL2  +=       0.086 * cos(5.44050231013  + 6816.28993343500 *T)
    sumL2  +=       0.072 * cos(3.44697351738  + 14158.74771361560 *T)
    sumL2  +=       0.086 * cos(4.50213985772  + 13362.51701710200 *T)
    sumL2  +=       0.074 * cos(1.34701853675  + 3361.38782219220 *T)
    sumL2  +=       0.076 * cos(6.04002959430  + 5085.03841111500 *T)
    sumL2  +=       0.065 * cos(0.07623776004  + 5099.26550511660 *T)
    sumL2  +=       0.066 * cos(2.09160717290  + 14421.83163698840 *T)
    sumL2  +=       0.077 * cos(2.75740817982  + 19402.79695281660 *T)
    sumL2  +=       0.059 * cos(3.61679189501  + 7322.10246078180 *T)
    sumL2  +=       0.060 * cos(4.67079289372  + 10235.13637553740 *T)
    sumL2  +=       0.064 * cos(6.17220244155  + 9381.93999378540 *T)
    sumL2  +=       0.066 * cos(0.35824154106  + 7255.56965173440 *T)
    sumL2  +=       0.059 * cos(4.52818219212  + 9499.25986200560 *T)
    sumL2  +=       0.065 * cos(1.86765892646  + 11610.91017538320 *T)
    sumL2  +=       0.074 * cos(3.12276411101  + 1107.13880568480 *T)
    sumL2  +=       0.060 * cos(1.35069679183  + 19800.94595622480 *T)
    sumL2  +=       0.055 * cos(0.64581579779  + 6677.63442474780 *T)
    sumL2  +=       0.055 * cos(0.65206957868  + 7373.38245462640 *T)
    sumL2  +=       0.055 * cos(0.41979092764  + 9779.10867612540 *T)
    sumL2  +=       0.074 * cos(3.85276242820  + 2766.26762836500 *T)
    sumL2  +=       0.056 * cos(2.49261586879  + 5642.19824260920 *T)
    sumL2  +=       0.055 * cos(2.62555306434  + 14314.16811304980 *T)
    sumL2  +=       0.054 * cos(3.95212835807  + 4032.77002792660 *T)
    sumL2  +=       0.055 * cos(0.51062350773  + 15110.46611986620 *T)
    sumL2  +=       0.053 * cos(5.88757767795  + 21265.52312652020 *T)
    sumL2  +=       0.051 * cos(5.99253934388  + 21947.11137270000 *T)
    sumL2  +=       0.049 * cos(5.16723732129  + 20047.19767854780 *T)
    sumL2  +=       0.067 * cos(1.53758821512  + 9380.95967271720 *T)
    sumL2  +=       0.057 * cos(5.34893894346  + 17101.21113690720 *T)
    sumL2  +=       0.049 * cos(1.92678602770  + 12721.57209941700 *T)
    sumL2  +=       0.046 * cos(2.31590320480  + 10028.95082710020 *T)
    sumL2  +=       0.052 * cos(4.92463460288  + 16304.91313009080 *T)
    sumL2  +=       0.050 * cos(1.69803302925  + 9225.53927328300 *T)
    sumL2  +=       0.045 * cos(2.91271003676  + 8982.81066930900 *T)
    sumL2  +=       0.044 * cos(4.38728556203  + 12168.00269657460 *T)
    sumL2  +=       0.047 * cos(4.44933635857  + 6872.67311951120 *T)
    sumL2  +=       0.043 * cos(2.13295656057  + 22324.90505670940 *T)
    sumL2  +=       0.050 * cos(1.34759191325  + 8439.87793181640 *T)
    sumL2  +=       0.044 * cos(0.89037464728  + 20752.66436247540 *T)
    sumL2  +=       0.041 * cos(2.76164156557  + 6696.47732458460 *T)
    sumL2  +=       0.042 * cos(4.45031185062  + 10404.73381232260 *T)
    sumL2  +=       0.044 * cos(3.20262780268  + 22743.40937951640 *T)
    sumL2  +=       0.042 * cos(3.54559408987  + 30376.35263916660 *T)
    sumL2  +=       0.042 * cos(0.82118836477  + 20597.24396304120 *T)
    sumL2  +=       0.052 * cos(4.59037852162  + 22345.26037610820 *T)
    sumL2  +=       0.042 * cos(3.79872109079  + 28628.33622609960 *T)
    sumL2  +=       0.041 * cos(2.36253602080  + 16062.18452611680 *T)
    sumL2  +=       0.050 * cos(1.48377570574  + 6040.34724601740 *T)
    sumL2  +=       0.041 * cos(5.33088551342  + 765.79306444640 *T)
    sumL2  +=       0.039 * cos(1.57984331116  + 24889.57479599160 *T)
    sumL2  +=       0.038 * cos(0.21571600948  + 9168.64089834740 *T)
    sumL2  +=       0.038 * cos(3.47667842127  + 1066.49547719000 *T)
    sumL2  +=       0.052 * cos(0.81427285458  + 9638.94074787620 *T)
    sumL2  +=       0.041 * cos(0.04356010851  + 6660.86953400080 *T)
    sumL2  +=       0.046 * cos(0.89345529755  + 16703.04487984680 *T)
    sumL2  +=       0.044 * cos(0.85335841824  + 17468.85519794540 *T)
    sumL2  +=       0.039 * cos(0.27098916103  + 11614.43329373220 *T)
    sumL2  +=       0.036 * cos(3.93388136028  + 33716.96506586640 *T)
    sumL2  +=       0.047 * cos(3.90896957151  + 5244.04923920100 *T)
    sumL2  +=       0.034 * cos(2.37310468308  + 10213.28554621100 *T)
    sumL2  +=       0.034 * cos(4.85454495742  + 14469.58851248400 *T)
    sumL2  +=       0.035 * cos(1.96497348634  + 6702.00024889200 *T)
    sumL2  +=       0.033 * cos(2.14488758889  + 13363.43002786740 *T)
    sumL2  +=       0.046 * cos(2.92422266239  + 16703.07938715119 *T)
    sumL2  +=       0.035 * cos(1.78692835372  + 2648.45482547300 *T)
    sumL2  +=       0.033 * cos(4.24484480826  + 8584.66166590080 *T)
    sumL2  +=       0.043 * cos(6.09391349922  + 6660.44945790720 *T)
    sumL2  +=       0.035 * cos(5.72426647579  + 16699.53901514999 *T)
    sumL2  +=       0.031 * cos(1.95254273311  + 6665.97238221460 *T)
    sumL2  +=       0.028 * cos(5.88548900893  + 19513.98359510420 *T)
    sumL2  +=       0.030 * cos(4.38908125588  + 12295.95422960920 *T)
    sumL2  +=       0.029 * cos(6.25276686131  + 17924.91069982040 *T)
    sumL2  +=       0.025 * cos(4.89000247975  + 30065.51184029820 *T)
    sumL2  +=       0.026 * cos(2.43119321236  + 6518.75821726740 *T)
    sumL2  +=       0.028 * cos(3.97032105354  + 4407.10790388980 *T)
    sumL2  +=       0.025 * cos(1.88631752737  + 4379.63903749020 *T)
    sumL2  +=       0.028 * cos(3.06196380356  + 6247.51311552280 *T)
    sumL2  +=       0.029 * cos(5.98057498931  + 6680.24453233140 *T)
    sumL2  +=       0.032 * cos(4.65556810177  + 6701.58017279840 *T)
    sumL2  +=       0.026 * cos(0.76018385758  + 3863.18984479360 *T)
    sumL2  +=       0.023 * cos(0.31107548102  + 8827.39026987480 *T)
    sumL2  +=       0.022 * cos(1.11553014451  + 6684.81528205140 *T)
    sumL2  +=       0.020 * cos(4.81193192299  + 15508.61512327440 *T)
    sumL2  +=       0.019 * cos(5.11656455993  + 12012.58229714040 *T)
    sumL2  +=       0.026 * cos(1.07254469525  + 8425.65083781480 *T)
    sumL2  +=       0.021 * cos(2.68436255141  + 6261.74020952440 *T)
    sumL2  +=       0.022 * cos(2.15631095909  + 3335.08950239240 *T)
    sumL2  +=       0.017 * cos(0.03684125273  + 3877.41693879520 *T)
    sumL2  +=       0.017 * cos(2.84467149903  + 3344.49376205780 *T)
    sumL2  +=       0.020 * cos(5.41519706836  + 3205.54734666440 *T)

    # MARS Longitude
    sumL3  = 0
    sumL3  +=    1482.423 * cos(0.44434694876  + 3340.61242669980 *T)
    sumL3  +=     662.095 * cos(0.88469178686  + 6681.22485339960 *T)
    sumL3  +=     188.268 * cos(1.28799982497  + 10021.83728009940 *T)
    sumL3  +=      41.474 * cos(1.64850786997  + 13362.44970679920 *T)
    sumL3  +=      22.661 * cos(2.05267665262  + 155.42039943420 *T)
    sumL3  +=      25.994 * cos(0.00000000000  + 0.00000000000 *T)
    sumL3  +=       8.024 * cos(1.99858757687  + 16703.06213349900 *T)
    sumL3  +=      10.454 * cos(1.58006906385  + 3.52311834900 *T)
    sumL3  +=       4.900 * cos(2.82452457966  + 242.72860397400 *T)
    sumL3  +=       3.782 * cos(2.01914272515  + 3344.13554504880 *T)
    sumL3  +=       3.176 * cos(4.59144897927  + 3185.19202726560 *T)
    sumL3  +=       3.134 * cos(0.65044714325  + 553.56940284240 *T)
    sumL3  +=       1.684 * cos(5.53835848782  + 951.71840625060 *T)
    sumL3  +=       1.511 * cos(5.71795850828  + 191.44826611160 *T)
    sumL3  +=       1.448 * cos(0.45869142895  + 796.29800681640 *T)
    sumL3  +=       1.442 * cos(2.34368495577  + 20043.67456019880 *T)
    sumL3  +=       1.302 * cos(5.36284013048  + 0.98032106820 *T)
    sumL3  +=       1.169 * cos(4.14601161433  + 1349.86740965880 *T)
    sumL3  +=       1.133 * cos(2.38180830662  + 6684.74797174860 *T)
    sumL3  +=       1.037 * cos(1.76892750558  + 382.89653222320 *T)
    sumL3  +=       0.894 * cos(5.33688328934  + 1194.44701022460 *T)
    sumL3  +=       0.807 * cos(2.74798886181  + 1748.01641306700 *T)
    sumL3  +=       0.640 * cos(6.10665147849  + 3496.03282613400 *T)
    sumL3  +=       0.558 * cos(1.85212342360  + 398.14900340820 *T)
    sumL3  +=       0.567 * cos(5.85922384979  + 7.11354700080 *T)
    sumL3  +=       0.647 * cos(3.17645475605  + 3583.34103067380 *T)
    sumL3  +=       0.452 * cos(5.98109989317  + 2787.04302385740 *T)
    sumL3  +=       0.508 * cos(1.01139298015  + 3149.16416058820 *T)
    sumL3  +=       0.519 * cos(4.93376176788  + 6525.80445396540 *T)
    sumL3  +=       0.405 * cos(1.27295444059  + 2281.23049651060 *T)
    sumL3  +=       0.399 * cos(2.32888685659  + 3738.76143010800 *T)
    sumL3  +=       0.472 * cos(0.84411483892  + 4136.91043351620 *T)
    sumL3  +=       0.333 * cos(5.42704539231  + 1059.38193018920 *T)
    sumL3  +=       0.313 * cos(3.70599897858  + 3097.88382272579 *T)
    sumL3  +=       0.281 * cos(5.74581724084  + 3341.59274776800 *T)
    sumL3  +=       0.287 * cos(2.69304799864  + 10025.36039844840 *T)
    sumL3  +=       0.237 * cos(5.68519881994  + 4535.05943692440 *T)
    sumL3  +=       0.215 * cos(2.49494803822  + 1990.74501704100 *T)
    sumL3  +=       0.260 * cos(2.67996877129  + 23384.28698689860 *T)
    sumL3  +=       0.269 * cos(1.10390153866  + 2388.89402044920 *T)
    sumL3  +=       0.186 * cos(4.86408411823  + 2957.71589447660 *T)
    sumL3  +=       0.176 * cos(3.86618588087  + 1592.59601363280 *T)
    sumL3  +=       0.177 * cos(0.85374318134  + 3894.18182954220 *T)
    sumL3  +=       0.191 * cos(4.58805692093  + 3337.08930835080 *T)
    sumL3  +=       0.185 * cos(5.80899316550  + 4399.99435688900 *T)
    sumL3  +=       0.143 * cos(1.15592612974  + 7477.52286021600 *T)
    sumL3  +=       0.173 * cos(2.70622920014  + 7079.37385680780 *T)
    sumL3  +=       0.146 * cos(1.37569151302  + 6489.77658728800 *T)
    sumL3  +=       0.121 * cos(3.24333934982  + 9492.14631500480 *T)
    sumL3  +=       0.116 * cos(5.71639819530  + 12303.06777661000 *T)
    sumL3  +=       0.111 * cos(2.69566947038  + 6151.53388830500 *T)
    sumL3  +=       0.103 * cos(3.53587741373  + 6923.95345737360 *T)
    sumL3  +=       0.097 * cos(5.23712981002  + 8962.45534991020 *T)
    sumL3  +=       0.089 * cos(0.05951324220  + 6127.65545055720 *T)
    sumL3  +=       0.093 * cos(5.29818155587  + 9866.41688066520 *T)
    sumL3  +=       0.084 * cos(5.68544044325  + 4292.33083295040 *T)
    sumL3  +=       0.091 * cos(6.16223680965  + 6682.20517446780 *T)
    sumL3  +=       0.081 * cos(4.32723330620  + 4933.20844033260 *T)
    sumL3  +=       0.077 * cos(5.63189824170  + 1589.07289528380 *T)
    sumL3  +=       0.102 * cos(6.26937663026  + 3347.72597370060 *T)
    sumL3  +=       0.076 * cos(3.05724276091  + 13365.97282514820 *T)
    sumL3  +=       0.063 * cos(3.12770931753  + 10419.98628350760 *T)
    sumL3  +=       0.062 * cos(1.45647168723  + 7234.79425624200 *T)
    sumL3  +=       0.069 * cos(4.82266605869  + 6677.70173505060 *T)
    sumL3  +=       0.064 * cos(1.73713448446  + 3870.30339179440 *T)
    sumL3  +=       0.058 * cos(4.85723088492  + 5486.77784317500 *T)
    sumL3  +=       0.058 * cos(6.09849516790  + 7875.67186362420 *T)
    sumL3  +=       0.063 * cos(4.14716431158  + 3340.59517304760 *T)
    sumL3  +=       0.065 * cos(6.17932665318  + 3340.62968035200 *T)
    sumL3  +=       0.057 * cos(1.48905992502  + 5729.50644714900 *T)
    sumL3  +=       0.060 * cos(6.08710604997  + 7740.60678358880 *T)
    sumL3  +=       0.056 * cos(1.96253427165  + 3553.91152213780 *T)
    sumL3  +=       0.059 * cos(0.89874385953  + 3339.63210563160 *T)
    sumL3  +=       0.053 * cos(3.04776975379  + 26724.89941359840 *T)
    sumL3  +=       0.061 * cos(3.63390789623  + 12832.75874170460 *T)
    sumL3  +=       0.059 * cos(0.22989746040  + 6681.24210705180 *T)
    sumL3  +=       0.053 * cos(3.97996929188  + 5092.15195811580 *T)
    sumL3  +=       0.059 * cos(4.48255178273  + 6681.20759974740 *T)
    sumL3  +=       0.045 * cos(3.88055745121  + 10264.56588407340 *T)
    sumL3  +=       0.058 * cos(4.85371375265  + 5621.84292321040 *T)
    sumL3  +=       0.049 * cos(4.03623343126  + 16173.37116840440 *T)
    sumL3  +=       0.061 * cos(2.90942510134  + 5331.35744374080 *T)
    sumL3  +=       0.043 * cos(1.80990718340  + 9830.38901398780 *T)
    sumL3  +=       0.044 * cos(4.10828944542  + 4690.47983635860 *T)
    sumL3  +=       0.044 * cos(3.49332765553  + 13760.59871020740 *T)
    sumL3  +=       0.048 * cos(1.81207342615  + 3723.50895892300 *T)
    sumL3  +=       0.041 * cos(1.59906754314  + 10818.13528691580 *T)
    sumL3  +=       0.041 * cos(2.05797417369  + 6681.29216370240 *T)
    sumL3  +=       0.038 * cos(3.38619280993  + 16706.58525184800 *T)
    sumL3  +=       0.041 * cos(2.53336616026  + 6681.15754309680 *T)
    sumL3  +=       0.041 * cos(1.79026132252  + 3340.54511639700 *T)
    sumL3  +=       0.050 * cos(4.70324784090  + 8273.82086703240 *T)
    sumL3  +=       0.037 * cos(3.37898394417  + 30065.51184029820 *T)
    sumL3  +=       0.039 * cos(2.65566008587  + 6894.52394883760 *T)
    sumL3  +=       0.035 * cos(2.58475025674  + 9623.68827669120 *T)
    sumL3  +=       0.037 * cos(6.07442797404  + 4929.68532198360 *T)
    sumL3  +=       0.042 * cos(3.04642561189  + 12964.30070339100 *T)
    sumL3  +=       0.037 * cos(5.23784488646  + 6298.32832117640 *T)
    sumL3  +=       0.043 * cos(6.17835198533  + 15643.68020330980 *T)
    sumL3  +=       0.034 * cos(2.44806511268  + 7210.91581849420 *T)
    sumL3  +=       0.038 * cos(0.26295105909  + 10022.81760116760 *T)
    sumL3  +=       0.028 * cos(0.74281728305  + 10021.85453375160 *T)
    sumL3  +=       0.028 * cos(4.99536302577  + 10021.82002644720 *T)
    sumL3  +=       0.034 * cos(5.04118058624  + 8969.56889691100 *T)
    sumL3  +=       0.025 * cos(2.59849002421  + 8955.34180290940 *T)
    sumL3  +=       0.030 * cos(1.14354290938  + 2146.16541647520 *T)
    sumL3  +=       0.027 * cos(0.26980783576  + 18984.29263000960 *T)
    sumL3  +=       0.025 * cos(0.39810308575  + 6688.33840040040 *T)
    sumL3  +=       0.024 * cos(5.64558695441  + 13207.02930736500 *T)
    sumL3  +=       0.027 * cos(6.24398988438  + 5088.62883976680 *T)
    sumL3  +=       0.024 * cos(0.25455613132  + 11081.21921028860 *T)
    sumL3  +=       0.023 * cos(0.53673547304  + 3333.49887969900 *T)
    sumL3  +=       0.024 * cos(1.80155161992  + 6674.11130639880 *T)
    sumL3  +=       0.029 * cos(4.39953507920  + 10018.31416175040 *T)
    sumL3  +=       0.023 * cos(2.91040060956  + 10551.52824519400 *T)
    sumL3  +=       0.022 * cos(1.20895727897  + 529.69096509460 *T)
    sumL3  +=       0.026 * cos(0.08254324904  + 7632.94325965020 *T)
    sumL3  +=       0.026 * cos(1.81668969835  + 6283.07584999140 *T)
    sumL3  +=       0.021 * cos(0.67232850824  + 6836.64525283380 *T)
    sumL3  +=       0.017 * cos(5.19518401928  + 2914.01423582380 *T)
    sumL3  +=       0.017 * cos(5.16488977775  + 8031.09226305840 *T)
    sumL3  +=       0.021 * cos(3.62973082412  + 5884.92684658320 *T)
    sumL3  +=       0.018 * cos(5.06870872024  + 2544.31441988340 *T)
    sumL3  +=       0.014 * cos(2.50522181917  + 7064.12138562280 *T)
    sumL3  +=       0.017 * cos(0.41835598950  + 9468.26787725700 *T)
    sumL3  +=       0.011 * cos(2.21363101654  + 6438.49624942560 *T)
    sumL3  +=       0.013 * cos(1.38424462832  + 3340.67973700260 *T)
    sumL3  +=       0.012 * cos(3.28248484262  + 8671.96987044060 *T)
    sumL3  +=       0.014 * cos(0.23027665815  + 3767.21061757580 *T)

    # MARS Longitude
    sumL4  = 0
    sumL4  +=     113.969 * cos(3.14159265359  + 0.00000000000 *T)
    sumL4  +=      28.725 * cos(5.63662412043  + 6681.22485339960 *T)
    sumL4  +=      24.447 * cos(5.13868481454  + 3340.61242669980 *T)
    sumL4  +=      11.187 * cos(6.03161074431  + 10021.83728009940 *T)
    sumL4  +=       3.190 * cos(3.56267988299  + 155.42039943420 *T)
    sumL4  +=       3.252 * cos(0.13228350651  + 13362.44970679920 *T)
    sumL4  +=       0.787 * cos(0.49340783377  + 16703.06213349900 *T)
    sumL4  +=       0.776 * cos(1.31734531594  + 242.72860397400 *T)
    sumL4  +=       0.494 * cos(3.06356214498  + 3185.19202726560 *T)
    sumL4  +=       0.374 * cos(2.15785846355  + 553.56940284240 *T)
    sumL4  +=       0.331 * cos(6.23159792887  + 3.52311834900 *T)
    sumL4  +=       0.197 * cos(0.44350153983  + 3344.13554504880 *T)
    sumL4  +=       0.181 * cos(0.81531283571  + 20043.67456019880 *T)
    sumL4  +=       0.168 * cos(3.73509781785  + 3496.03282613400 *T)
    sumL4  +=       0.086 * cos(0.79259553758  + 6684.74797174860 *T)
    sumL4  +=       0.115 * cos(1.66898531261  + 3583.34103067380 *T)
    sumL4  +=       0.092 * cos(3.40530361815  + 6525.80445396540 *T)
    sumL4  +=       0.064 * cos(4.47443580658  + 2787.04302385740 *T)
    sumL4  +=       0.045 * cos(5.17216217058  + 3097.88382272579 *T)
    sumL4  +=       0.041 * cos(1.21875027733  + 23384.28698689860 *T)
    sumL4  +=       0.036 * cos(5.53975653407  + 3149.16416058820 *T)
    sumL4  +=       0.039 * cos(5.40966345885  + 4136.91043351620 *T)
    sumL4  +=       0.028 * cos(4.85378781404  + 6127.65545055720 *T)
    sumL4  +=       0.027 * cos(0.42811091036  + 3738.76143010800 *T)
    sumL4  +=       0.032 * cos(1.22337093927  + 10025.36039844840 *T)
    sumL4  +=       0.025 * cos(1.56348878811  + 26724.89941359840 *T)
    sumL4  +=       0.029 * cos(2.40178626971  + 3894.18182954220 *T)
    sumL4  +=       0.023 * cos(4.16371253947  + 3341.59274776800 *T)
    sumL4  +=       0.022 * cos(5.98429191667  + 7477.52286021600 *T)
    sumL4  +=       0.019 * cos(2.03228980685  + 6923.95345737360 *T)
    sumL4  +=       0.020 * cos(3.79275573433  + 9866.41688066520 *T)
    sumL4  +=       0.021 * cos(4.27532111122  + 6836.64525283380 *T)
    sumL4  +=       0.015 * cos(1.56315497374  + 13365.97282514820 *T)
    sumL4  +=       0.011 * cos(1.33261955275  + 2281.23049651060 *T)
    sumL4  +=       0.009 * cos(6.15390464542  + 6489.77658728800 *T)
    sumL4  +=       0.007 * cos(4.23894194106  + 4535.05943692440 *T)

    # MARS Longitude
    sumL5  = 0
    sumL5  +=       0.710 * cos(4.04089996521  + 6681.22485339960 *T)
    sumL5  +=       0.868 * cos(3.14159265359  + 0.00000000000 *T)
    sumL5  +=       0.510 * cos(4.49214901625  + 10021.83728009940 *T)
    sumL5  +=       0.357 * cos(5.07435505061  + 155.42039943420 *T)
    sumL5  +=       0.223 * cos(3.51351884241  + 3340.61242669980 *T)
    sumL5  +=       0.199 * cos(4.85313666795  + 13362.44970679920 *T)
    sumL5  +=       0.100 * cos(6.09089356066  + 242.72860397400 *T)
    sumL5  +=       0.069 * cos(5.19017483537  + 16703.06213349900 *T)
    sumL5  +=       0.064 * cos(1.55783055571  + 3185.19202726560 *T)
    sumL5  +=       0.035 * cos(3.68246171643  + 553.56940284240 *T)
    sumL5  +=       0.025 * cos(5.22079788019  + 3496.03282613400 *T)
    sumL5  +=       0.024 * cos(0.17861991485  + 3583.34103067380 *T)
    sumL5  +=       0.019 * cos(5.59184714460  + 20043.67456019880 *T)
    sumL5  +=       0.014 * cos(5.95565787085  + 23384.28698689860 *T)
    sumL5  +=       0.012 * cos(1.93859256739  + 6525.80445396540 *T)

    # MARS Latitude
    sumB0  = 0
    sumB0  +=  3197134.986 * cos(3.76832042432  + 3340.61242669980 *T)
    sumB0  +=  298033.234 * cos(4.10616996243  + 6681.22485339960 *T)
    sumB0  +=  289104.742 * cos(0.00000000000  + 0.00000000000 *T)
    sumB0  +=   31365.538 * cos(4.44651052853  + 10021.83728009940 *T)
    sumB0  +=    3484.100 * cos(4.78812547889  + 13362.44970679920 *T)
    sumB0  +=     442.999 * cos(5.65233015876  + 3337.08930835080 *T)
    sumB0  +=     443.401 * cos(5.02642620491  + 3344.13554504880 *T)
    sumB0  +=     399.109 * cos(5.13056814700  + 16703.06213349900 *T)
    sumB0  +=     292.506 * cos(3.79290644595  + 2281.23049651060 *T)
    sumB0  +=     181.982 * cos(6.13648011704  + 6151.53388830500 *T)
    sumB0  +=     163.159 * cos(4.26399626634  + 529.69096509460 *T)
    sumB0  +=     159.678 * cos(2.23194610246  + 1059.38193018920 *T)
    sumB0  +=     139.323 * cos(2.41796344238  + 8962.45534991020 *T)
    sumB0  +=     149.297 * cos(2.16501209917  + 5621.84292321040 *T)
    sumB0  +=     142.686 * cos(1.18215016110  + 3340.59517304760 *T)
    sumB0  +=     142.685 * cos(3.21292180820  + 3340.62968035200 *T)
    sumB0  +=      82.544 * cos(5.36667872319  + 6684.74797174860 *T)
    sumB0  +=      73.640 * cos(5.09187524843  + 398.14900340820 *T)
    sumB0  +=      72.660 * cos(5.53775710437  + 6283.07584999140 *T)
    sumB0  +=      86.377 * cos(5.74429648412  + 3738.76143010800 *T)
    sumB0  +=      83.276 * cos(5.98866315739  + 6677.70173505060 *T)
    sumB0  +=      60.116 * cos(3.67960808826  + 796.29800681640 *T)
    sumB0  +=      63.111 * cos(0.73049113369  + 5884.92684658320 *T)
    sumB0  +=      62.338 * cos(4.85071999184  + 2942.46342329160 *T)
    sumB0  +=      46.951 * cos(5.54339723804  + 3340.54511639700 *T)
    sumB0  +=      46.953 * cos(5.13486627234  + 3340.67973700260 *T)
    sumB0  +=      46.630 * cos(5.47361665459  + 20043.67456019880 *T)
    sumB0  +=      45.588 * cos(2.13262507507  + 2810.92146160520 *T)
    sumB0  +=      41.269 * cos(0.20003189001  + 9492.14631500480 *T)
    sumB0  +=      47.199 * cos(4.52184736343  + 3149.16416058820 *T)
    sumB0  +=      38.540 * cos(4.08008443274  + 4136.91043351620 *T)
    sumB0  +=      33.069 * cos(4.06581918329  + 1751.53953141600 *T)
    sumB0  +=      29.694 * cos(5.92218297386  + 3532.06069281140 *T)
    sumB0  +=      32.736 * cos(2.62071056958  + 2914.01423582380 *T)
    sumB0  +=      29.521 * cos(2.75342566734  + 12303.06777661000 *T)
    sumB0  +=      28.169 * cos(2.06282533993  + 5486.77784317500 *T)
    sumB0  +=      28.618 * cos(4.94710527914  + 3870.30339179440 *T)
    sumB0  +=      26.603 * cos(3.55085844020  + 6681.24210705180 *T)
    sumB0  +=      26.603 * cos(1.52008675291  + 6681.20759974740 *T)
    sumB0  +=      23.336 * cos(2.27624532707  + 1589.07289528380 *T)
    sumB0  +=      26.052 * cos(2.60064548916  + 4399.99435688900 *T)
    sumB0  +=      22.637 * cos(2.27507466406  + 1194.44701022460 *T)
    sumB0  +=      18.887 * cos(6.04416196149  + 7079.37385680780 *T)
    sumB0  +=      14.846 * cos(3.41358603159  + 5088.62883976680 *T)
    sumB0  +=      19.947 * cos(2.67365368471  + 8432.76438481560 *T)
    sumB0  +=      14.682 * cos(5.89211938785  + 9623.68827669120 *T)
    sumB0  +=      14.152 * cos(2.42512744356  + 3333.49887969900 *T)
    sumB0  +=      13.310 * cos(2.62839773036  + 426.59819087600 *T)
    sumB0  +=      14.008 * cos(1.67425558329  + 6254.62666252360 *T)
    sumB0  +=      15.104 * cos(2.81013535571  + 3496.03282613400 *T)
    sumB0  +=      13.011 * cos(5.70759434129  + 10025.36039844840 *T)
    sumB0  +=      12.080 * cos(1.51804981987  + 3185.19202726560 *T)
    sumB0  +=      13.183 * cos(0.04521207632  + 10018.31416175040 *T)
    sumB0  +=      11.554 * cos(5.57418971820  + 191.44826611160 *T)
    sumB0  +=      11.196 * cos(0.55829576311  + 5092.15195811580 *T)
    sumB0  +=      11.530 * cos(2.13314819584  + 11773.37681151540 *T)
    sumB0  +=      10.435 * cos(5.72414012635  + 6467.92575796160 *T)
    sumB0  +=       9.846 * cos(0.86942245495  + 1592.59601363280 *T)
    sumB0  +=       9.761 * cos(1.09342288660  + 2544.31441988340 *T)
    sumB0  +=       8.754 * cos(5.47281660412  + 6681.29216370240 *T)
    sumB0  +=       8.937 * cos(4.83790385610  + 6489.77658728800 *T)
    sumB0  +=       8.652 * cos(4.72118876809  + 213.29909543800 *T)
    sumB0  +=       8.797 * cos(2.86597575792  + 3341.59274776800 *T)
    sumB0  +=       8.384 * cos(2.65895230037  + 4535.05943692440 *T)
    sumB0  +=       8.213 * cos(4.82608134710  + 3553.91152213780 *T)
    sumB0  +=       8.799 * cos(1.52910882795  + 3339.63210563160 *T)
    sumB0  +=       8.103 * cos(1.00994045172  + 9225.53927328300 *T)
    sumB0  +=       8.754 * cos(5.88131160438  + 6681.15754309680 *T)
    sumB0  +=       7.209 * cos(4.41679446871  + 7477.52286021600 *T)
    sumB0  +=       8.560 * cos(4.79003986337  + 4690.47983635860 *T)
    sumB0  +=       6.087 * cos(1.89071507516  + 9595.23908922340 *T)
    sumB0  +=       6.974 * cos(0.53246284561  + 12832.75874170460 *T)
    sumB0  +=       5.584 * cos(6.18909308524  + 4292.33083295040 *T)
    sumB0  +=       5.038 * cos(6.06393121994  + 7210.91581849420 *T)
    sumB0  +=       5.126 * cos(0.11856805000  + 4562.46099302120 *T)
    sumB0  +=       4.863 * cos(1.33051301524  + 3894.18182954220 *T)
    sumB0  +=       5.592 * cos(3.97792577165  + 3127.31333126180 *T)
    sumB0  +=       4.965 * cos(5.74589127183  + 1990.74501704100 *T)
    sumB0  +=       5.530 * cos(5.81711987294  + 23384.28698689860 *T)
    sumB0  +=       4.184 * cos(3.88899446923  + 10021.85453375160 *T)
    sumB0  +=       4.184 * cos(1.85821885540  + 10021.82002644720 *T)
    sumB0  +=       4.685 * cos(1.11750235736  + 3319.83703120740 *T)
    sumB0  +=       3.954 * cos(2.65573703692  + 4929.68532198360 *T)
    sumB0  +=       5.174 * cos(3.62343554596  + 2700.71514038580 *T)
    sumB0  +=       4.639 * cos(6.14089337232  + 7.11354700080 *T)
    sumB0  +=       4.686 * cos(3.27731664264  + 3361.38782219220 *T)
    sumB0  +=       4.999 * cos(3.09121636067  + 15643.68020330980 *T)
    sumB0  +=       3.919 * cos(2.96552252075  + 7740.60678358880 *T)
    sumB0  +=       3.361 * cos(1.85016770821  + 2957.71589447660 *T)
    sumB0  +=       4.246 * cos(4.19639278630  + 2146.16541647520 *T)
    sumB0  +=       4.101 * cos(6.22532797325  + 10213.28554621100 *T)
    sumB0  +=       3.124 * cos(0.08536218671  + 10419.98628350760 *T)
    sumB0  +=       3.079 * cos(2.41163389513  + 3344.20285535160 *T)
    sumB0  +=       3.093 * cos(1.98041729732  + 3337.02199804800 *T)
    sumB0  +=       3.044 * cos(4.38894419912  + 1221.84856632140 *T)
    sumB0  +=       2.879 * cos(2.93292197885  + 2787.04302385740 *T)
    sumB0  +=       3.372 * cos(4.45558202843  + 6674.11130639880 *T)
    sumB0  +=       2.510 * cos(4.06249795775  + 6872.67311951120 *T)
    sumB0  +=       2.616 * cos(1.23659420897  + 4933.20844033260 *T)
    sumB0  +=       3.011 * cos(4.33965766608  + 2388.89402044920 *T)
    sumB0  +=       3.094 * cos(2.40385309455  + 2118.76386037840 *T)
    sumB0  +=       2.516 * cos(2.30275910483  + 3723.50895892300 *T)
    sumB0  +=       2.192 * cos(1.68877598700  + 639.89728631400 *T)
    sumB0  +=       2.157 * cos(6.07600403318  + 9808.53818466140 *T)
    sumB0  +=       2.686 * cos(2.38172448359  + 15113.98923821520 *T)
    sumB0  +=       2.156 * cos(2.87771528495  + 6041.32756708560 *T)
    sumB0  +=       2.495 * cos(3.87395320804  + 7903.07341972100 *T)
    sumB0  +=       2.253 * cos(0.47469626036  + 3475.67750673520 *T)
    sumB0  +=       2.443 * cos(3.02632351975  + 3767.21061757580 *T)
    sumB0  +=       1.937 * cos(0.38000997629  + 13358.92658845020 *T)
    sumB0  +=       2.213 * cos(3.91982326360  + 3205.54734666440 *T)
    sumB0  +=       2.385 * cos(6.23883456815  + 12964.30070339100 *T)
    sumB0  +=       1.806 * cos(2.38408432339  + 11243.68584642080 *T)
    sumB0  +=       2.265 * cos(1.85773885180  + 6525.80445396540 *T)
    sumB0  +=       1.642 * cos(1.86812936557  + 6680.24453233140 *T)
    sumB0  +=       1.616 * cos(3.39954231101  + 8429.24126646660 *T)
    sumB0  +=       1.646 * cos(0.73261192032  + 2274.11694950980 *T)
    sumB0  +=       1.910 * cos(6.04319894051  + 13365.97282514820 *T)
    sumB0  +=       1.492 * cos(2.99360929050  + 7875.67186362420 *T)
    sumB0  +=       1.567 * cos(1.76860259924  + 382.89653222320 *T)
    sumB0  +=       1.459 * cos(1.40571286543  + 14584.29827312060 *T)
    sumB0  +=       1.527 * cos(5.79343536555  + 6894.52394883760 *T)
    sumB0  +=       1.409 * cos(5.17099876287  + 9830.38901398780 *T)
    sumB0  +=       1.665 * cos(3.60183854165  + 2288.34404351140 *T)
    sumB0  +=       1.616 * cos(2.19447775480  + 8827.39026987480 *T)
    sumB0  +=       1.746 * cos(0.10485504333  + 3583.34103067380 *T)
    sumB0  +=       1.637 * cos(3.20410489018  + 6682.20517446780 *T)
    sumB0  +=       1.269 * cos(2.02253791623  + 155.42039943420 *T)
    sumB0  +=       1.507 * cos(5.80795976424  + 1748.01641306700 *T)
    sumB0  +=       1.296 * cos(2.92899986522  + 9381.93999378540 *T)
    sumB0  +=       1.291 * cos(2.98932404463  + 6836.64525283380 *T)
    sumB0  +=       1.377 * cos(5.80980821450  + 10021.90459040220 *T)
    sumB0  +=       1.357 * cos(0.32856318286  + 4032.77002792660 *T)
    sumB0  +=       1.377 * cos(6.21830533138  + 10021.76996979660 *T)
    sumB0  +=       1.227 * cos(2.20421067322  + 12935.85151592320 *T)
    sumB0  +=       1.080 * cos(4.75502761725  + 10818.13528691580 *T)
    sumB0  +=       1.016 * cos(1.33950348904  + 12566.15169998280 *T)
    sumB0  +=       1.083 * cos(0.87622049125  + 16173.37116840440 *T)
    sumB0  +=       0.957 * cos(5.29505019432  + 1066.49547719000 *T)
    sumB0  +=       1.035 * cos(2.44286811558  + 1052.26838318840 *T)
    sumB0  +=       1.153 * cos(4.73786864168  + 8031.09226305840 *T)
    sumB0  +=       1.042 * cos(1.73541825729  + 3369.06161416760 *T)
    sumB0  +=       0.899 * cos(1.87265511538  + 3077.52850332700 *T)
    sumB0  +=       0.892 * cos(2.51347636401  + 3603.69635007260 *T)
    sumB0  +=       0.904 * cos(0.27361540197  + 3325.35995551480 *T)
    sumB0  +=       0.883 * cos(5.46537936740  + 5614.72937620960 *T)
    sumB0  +=       0.949 * cos(0.04595688771  + 7064.12138562280 *T)
    sumB0  +=       0.872 * cos(3.61205859852  + 6702.00024889200 *T)
    sumB0  +=       0.842 * cos(5.61713543212  + 8955.34180290940 *T)
    sumB0  +=       0.842 * cos(0.57536275896  + 3191.04922956520 *T)
    sumB0  +=       0.904 * cos(4.11832249537  + 3355.86489788480 *T)
    sumB0  +=       1.044 * cos(2.65889542110  + 3312.16323923200 *T)
    sumB0  +=       0.807 * cos(2.26443259820  + 8969.56889691100 *T)
    sumB0  +=       1.046 * cos(6.11317298058  + 5331.35744374080 *T)
    sumB0  +=       0.933 * cos(0.82551054416  + 3503.07906283200 *T)
    sumB0  +=       0.946 * cos(1.92368414869  + 5628.95647021120 *T)
    sumB0  +=       0.978 * cos(4.55697167317  + 3097.88382272579 *T)
    sumB0  +=       0.846 * cos(4.65096242323  + 6438.49624942560 *T)
    sumB0  +=       0.901 * cos(1.46788821922  + 6660.44945790720 *T)
    sumB0  +=       0.739 * cos(4.97716008834  + 10014.72373309860 *T)
    sumB0  +=       0.762 * cos(3.42576079312  + 18984.29263000960 *T)
    sumB0  +=       0.934 * cos(0.35014113954  + 3264.34635542420 *T)
    sumB0  +=       0.894 * cos(0.34991139502  + 10551.52824519400 *T)
    sumB0  +=       0.665 * cos(6.05561837558  + 7632.94325965020 *T)
    sumB0  +=       0.605 * cos(4.68480399923  + 5729.50644714900 *T)
    sumB0  +=       0.593 * cos(2.97207045977  + 3178.14579056760 *T)
    sumB0  +=       0.617 * cos(2.19529512598  + 13362.43245314700 *T)
    sumB0  +=       0.579 * cos(3.32303439858  + 11081.21921028860 *T)
    sumB0  +=       0.579 * cos(2.31217959727  + 6677.63442474780 *T)
    sumB0  +=       0.572 * cos(4.27984088783  + 6546.15977336420 *T)
    sumB0  +=       0.584 * cos(4.51625727923  + 1648.44675719740 *T)
    sumB0  +=       0.626 * cos(4.60715188203  + 2487.41604494780 *T)
    sumB0  +=       0.649 * cos(2.17894376907  + 6298.32832117640 *T)
    sumB0  +=       0.617 * cos(4.22635718375  + 13362.46696045140 *T)
    sumB0  +=       0.550 * cos(3.11161314141  + 6144.42034130420 *T)
    sumB0  +=       0.551 * cos(3.27081871281  + 6127.65545055720 *T)
    sumB0  +=       0.536 * cos(0.18973312148  + 3346.13535100720 *T)
    sumB0  +=       0.557 * cos(6.13686769377  + 26724.89941359840 *T)
    sumB0  +=       0.571 * cos(5.65341463660  + 2921.12778282460 *T)
    sumB0  +=       0.552 * cos(5.75724124958  + 149.56319713460 *T)
    sumB0  +=       0.638 * cos(4.49647258906  + 3.52311834900 *T)
    sumB0  +=       0.552 * cos(1.07795672495  + 536.80451209540 *T)
    sumB0  +=       0.537 * cos(4.19207200801  + 3335.08950239240 *T)
    sumB0  +=       0.567 * cos(2.62806466971  + 2648.45482547300 *T)
    sumB0  +=       0.641 * cos(3.24534699403  + 3347.72597370060 *T)
    sumB0  +=       0.546 * cos(2.76744357149  + 5459.37628707820 *T)
    sumB0  +=       0.556 * cos(4.98610303145  + 6531.66165626500 *T)
    sumB0  +=       0.578 * cos(2.74445077242  + 6684.81528205140 *T)
    sumB0  +=       0.613 * cos(3.01718810256  + 8270.29774868340 *T)
    sumB0  +=       0.558 * cos(5.89236366633  + 6158.64743530580 *T)
    sumB0  +=       0.489 * cos(4.59082197201  + 522.57741809380 *T)
    sumB0  +=       0.445 * cos(3.64293918046  + 10713.99488132620 *T)
    sumB0  +=       0.539 * cos(0.59284523316  + 553.56940284240 *T)
    sumB0  +=       0.434 * cos(1.99641894213  + 26087.90314157420 *T)
    sumB0  +=       0.459 * cos(4.84216944554  + 1758.65307841680 *T)
    sumB0  +=       0.423 * cos(3.77854586358  + 6688.33840040040 *T)
    sumB0  +=       0.421 * cos(3.10245838137  + 8329.67161059700 *T)
    sumB0  +=       0.421 * cos(4.43540354818  + 103.09277421860 *T)
    sumB0  +=       0.502 * cos(3.03141512630  + 3360.96774609859 *T)
    sumB0  +=       0.422 * cos(5.99158563064  + 242.72860397400 *T)
    sumB0  +=       0.401 * cos(2.06012045711  + 3980.50971301380 *T)
    sumB0  +=       0.410 * cos(5.85255605682  + 3351.24909204960 *T)
    sumB0  +=       0.375 * cos(2.50803340920  + 6784.31762761820 *T)
    sumB0  +=       0.407 * cos(0.30117328390  + 16304.91313009080 *T)
    sumB0  +=       0.507 * cos(1.35016417800  + 3320.25710730100 *T)
    sumB0  +=       0.385 * cos(1.06248122739  + 7234.79425624200 *T)
    sumB0  +=       0.519 * cos(4.61705823333  + 7373.38245462640 *T)
    sumB0  +=       0.401 * cos(0.51161484292  + 17085.95866572220 *T)
    sumB0  +=       0.371 * cos(0.41707432281  + 13760.59871020740 *T)
    sumB0  +=       0.476 * cos(0.81142637639  + 6816.28993343500 *T)
    sumB0  +=       0.348 * cos(3.36497054127  + 8439.87793181640 *T)
    sumB0  +=       0.366 * cos(6.05186692577  + 3607.21946842160 *T)
    sumB0  +=       0.402 * cos(2.67375675473  + 18454.60166491500 *T)
    sumB0  +=       0.439 * cos(0.41122589061  + 10404.73381232260 *T)
    sumB0  +=       0.412 * cos(4.81039654306  + 3329.97576135000 *T)
    sumB0  +=       0.413 * cos(0.21011211689  + 13149.15061136120 *T)
    sumB0  +=       0.363 * cos(2.19017574725  + 9866.41688066520 *T)
    sumB0  +=       0.337 * cos(5.46398740226  + 6604.95878212400 *T)
    sumB0  +=       0.340 * cos(1.73392662280  + 1692.16566950240 *T)
    sumB0  +=       0.344 * cos(3.58310197017  + 7107.82304427560 *T)
    sumB0  +=       0.420 * cos(2.77274750272  + 2906.90068882300 *T)
    sumB0  +=       0.335 * cos(2.02339633221  + 3333.56619000180 *T)
    sumB0  +=       0.336 * cos(2.78173647754  + 4989.05918389720 *T)
    sumB0  +=       0.354 * cos(5.41265456299  + 951.71840625060 *T)
    sumB0  +=       0.328 * cos(0.43464916253  + 6923.95345737360 *T)
    sumB0  +=       0.422 * cos(1.57591737080  + 8273.82086703240 *T)
    sumB0  +=       0.324 * cos(4.99165215532  + 11371.70468975820 *T)
    sumB0  +=       0.351 * cos(4.60766539695  + 3074.00538497800 *T)
    sumB0  +=       0.333 * cos(4.86991808664  + 3863.18984479360 *T)
    sumB0  +=       0.319 * cos(1.44510282827  + 11769.85369316640 *T)
    sumB0  +=       0.296 * cos(5.25637576704  + 76.26607127560 *T)
    sumB0  +=       0.349 * cos(1.62483962938  + 17924.91069982040 *T)
    sumB0  +=       0.293 * cos(2.20643287367  + 3347.65866339780 *T)
    sumB0  +=       0.315 * cos(4.55227306770  + 5828.02847164760 *T)
    sumB0  +=       0.280 * cos(1.42134291540  + 6606.44325483230 *T)
    sumB0  +=       0.336 * cos(1.56854533505  + 1581.95934828300 *T)
    sumB0  +=       0.343 * cos(0.27605726543  + 10973.55568635000 *T)
    sumB0  +=       0.271 * cos(0.08314924409  + 10235.13637553740 *T)
    sumB0  +=       0.280 * cos(0.76436434398  + 853.19638175200 *T)
    sumB0  +=       0.317 * cos(4.91466081381  + 3443.70520091840 *T)
    sumB0  +=       0.257 * cos(0.15709629188  + 12168.00269657460 *T)
    sumB0  +=       0.237 * cos(3.13380421805  + 74.78159856730 *T)
    sumB0  +=       0.271 * cos(1.83008759256  + 3399.98628861340 *T)
    sumB0  +=       0.233 * cos(3.20163702613  + 10177.25767953360 *T)
    sumB0  +=       0.312 * cos(0.28436027987  + 13745.34623902240 *T)
    sumB0  +=       0.273 * cos(4.53535816659  + 3344.49376205780 *T)
    sumB0  +=       0.235 * cos(0.67980074869  + 17395.21973472580 *T)
    sumB0  +=       0.262 * cos(3.84436620806  + 9168.64089834740 *T)
    sumB0  +=       0.271 * cos(4.94509131052  + 692.15760122680 *T)
    sumB0  +=       0.254 * cos(2.19491751584  + 3281.23856478620 *T)
    sumB0  +=       0.220 * cos(1.65400798925  + 3546.79797513700 *T)
    sumB0  +=       0.283 * cos(5.57884389098  + 3973.39616601300 *T)
    sumB0  +=       0.229 * cos(5.93153003758  + 3364.49086444760 *T)
    sumB0  +=       0.223 * cos(4.82744958322  + 9779.10867612540 *T)
    sumB0  +=       0.222 * cos(0.71788606092  + 16699.53901514999 *T)
    sumB0  +=       0.219 * cos(0.09831940529  + 16706.58525184800 *T)
    sumB0  +=       0.221 * cos(1.66596576205  + 10575.40668294180 *T)
    sumB0  +=       0.259 * cos(1.10208840059  + 1744.42598441520 *T)
    sumB0  +=       0.243 * cos(1.52425310144  + 14054.60730802600 *T)
    sumB0  +=       0.273 * cos(6.12932898811  + 3336.73109134180 *T)
    sumB0  +=       0.232 * cos(4.73422268172  + 3316.73398895200 *T)
    sumB0  +=       0.210 * cos(5.97305256642  + 419.48464387520 *T)
    sumB0  +=       0.213 * cos(5.35378111313  + 1596.18644228460 *T)
    sumB0  +=       0.197 * cos(3.32400761027  + 5085.03841111500 *T)
    sumB0  +=       0.193 * cos(0.90522037315  + 14314.16811304980 *T)
    sumB0  +=       0.192 * cos(2.63243342989  + 433.71173787680 *T)
    sumB0  +=       0.188 * cos(1.14983778034  + 13892.14067189380 *T)
    sumB0  +=       0.200 * cos(4.79567916236  + 6261.74020952440 *T)
    sumB0  +=       0.204 * cos(2.32272167710  + 13916.01910964160 *T)
    sumB0  +=       0.191 * cos(5.08454560792  + 2178.13772229200 *T)
    sumB0  +=       0.191 * cos(5.88821293073  + 3340.19235060619 *T)
    sumB0  +=       0.236 * cos(0.97341490472  + 3230.40610548040 *T)
    sumB0  +=       0.200 * cos(3.60041329896  + 3304.58456002240 *T)
    sumB0  +=       0.165 * cos(2.91701234796  + 3415.39402526710 *T)
    sumB0  +=       0.191 * cos(4.77647338120  + 3341.03250279340 *T)
    sumB0  +=       0.174 * cos(2.20549134205  + 10020.85695903120 *T)
    sumB0  +=       0.161 * cos(2.64795288862  + 3304.00706139560 *T)
    sumB0  +=       0.173 * cos(3.54115285115  + 10022.81760116760 *T)
    sumB0  +=       0.154 * cos(2.45916882835  + 12722.55242048520 *T)
    sumB0  +=       0.160 * cos(1.73968020026  + 3377.21779200400 *T)
    sumB0  +=       0.196 * cos(0.78013533938  + 3376.64029337720 *T)
    sumB0  +=       0.153 * cos(0.27328628499  + 13362.38239649640 *T)
    sumB0  +=       0.190 * cos(2.65143864216  + 3657.00429635640 *T)
    sumB0  +=       0.172 * cos(5.71092117488  + 2384.32327072920 *T)
    sumB0  +=       0.157 * cos(1.00235389487  + 3209.07046501340 *T)
    sumB0  +=       0.150 * cos(3.39592748537  + 3472.15438838620 *T)
    sumB0  +=       0.165 * cos(5.86551673354  + 2803.80791460440 *T)
    sumB0  +=       0.154 * cos(0.32136196834  + 6665.97238221460 *T)
    sumB0  +=       0.178 * cos(3.34059693754  + 11216.28429032400 *T)
    sumB0  +=       0.192 * cos(0.66079944629  + 3134.42687826260 *T)
    sumB0  +=       0.161 * cos(2.87915391245  + 4392.88080988820 *T)
    sumB0  +=       0.159 * cos(3.72249746113  + 13517.87010623340 *T)
    sumB0  +=       0.168 * cos(0.55746622745  + 110.20632121940 *T)
    sumB0  +=       0.158 * cos(5.67248197213  + 4407.10790388980 *T)
    sumB0  +=       0.161 * cos(1.76789928464  + 6709.67404086740 *T)
    sumB0  +=       0.158 * cos(5.50643845256  + 13171.00144068760 *T)
    sumB0  +=       0.129 * cos(1.21297315367  + 19513.98359510420 *T)
    sumB0  +=       0.173 * cos(0.17070148373  + 3024.22055704320 *T)
    sumB0  +=       0.132 * cos(4.25407099272  + 16858.48253293320 *T)
    sumB0  +=       0.125 * cos(2.88510294667  + 17256.63153634140 *T)
    sumB0  +=       0.124 * cos(2.59724598646  + 12310.18132361080 *T)
    sumB0  +=       0.170 * cos(4.05945259741  + 2818.03500860600 *T)
    sumB0  +=       0.121 * cos(5.23997785185  + 9872.27408296480 *T)
    sumB0  +=       0.142 * cos(3.02798835989  + 3511.28529731900 *T)
    sumB0  +=       0.165 * cos(2.53171951288  + 16276.46394262300 *T)
    sumB0  +=       0.153 * cos(6.14783670557  + 13362.51701710200 *T)
    sumB0  +=       0.119 * cos(4.15694365082  + 3760.09707057500 *T)
    sumB0  +=       0.120 * cos(0.64287725481  + 4459.36821880260 *T)
    sumB0  +=       0.130 * cos(4.95002309460  + 13553.89797291080 *T)
    sumB0  +=       0.120 * cos(0.17087854222  + 8671.96987044060 *T)
    sumB0  +=       0.112 * cos(0.16822264326  + 135.06508003540 *T)
    sumB0  +=       0.137 * cos(3.34809361979  + 3341.04230982650 *T)
    sumB0  +=       0.125 * cos(1.32195559043  + 1349.86740965880 *T)
    sumB0  +=       0.111 * cos(3.14151030451  + 13524.91634293140 *T)
    sumB0  +=       0.119 * cos(5.95361348050  + 12295.95422960920 *T)
    sumB0  +=       0.131 * cos(5.09769375731  + 14158.74771361560 *T)
    sumB0  +=       0.141 * cos(1.37128440708  + 3169.93955608060 *T)
    sumB0  +=       0.112 * cos(3.35831868034  + 5989.06725217280 *T)
    sumB0  +=       0.104 * cos(5.00696041032  + 13119.72110282519 *T)
    sumB0  +=       0.110 * cos(5.23317664736  + 1375.77379984580 *T)
    sumB0  +=       0.105 * cos(2.72692368303  + 1162.47470440780 *T)
    sumB0  +=       0.104 * cos(1.73769165705  + 2221.85663459700 *T)
    sumB0  +=       0.137 * cos(1.04576950390  + 3340.18254357310 *T)
    sumB0  +=       0.106 * cos(6.13415161313  + 162.46663613220 *T)
    sumB0  +=       0.119 * cos(2.63312561442  + 7321.12213971360 *T)
    sumB0  +=       0.105 * cos(3.09551802365  + 20618.01935853360 *T)
    sumB0  +=       0.099 * cos(4.25515697974  + 23539.70738633280 *T)
    sumB0  +=       0.108 * cos(1.01854506729  + 3265.83082813250 *T)
    sumB0  +=       0.119 * cos(4.07277528003  + 10184.30391623160 *T)
    sumB0  +=       0.096 * cos(1.81122023425  + 10001.06188460700 *T)
    sumB0  +=       0.093 * cos(3.58905885066  + 5099.26550511660 *T)
    sumB0  +=       0.095 * cos(4.94756054764  + 3981.49003408200 *T)
    sumB0  +=       0.094 * cos(5.37493368020  + 13355.33615979840 *T)
    sumB0  +=       0.095 * cos(0.13037485775  + 15508.61512327440 *T)
    sumB0  +=       0.103 * cos(0.43484130196  + 1861.74585263540 *T)
    sumB0  +=       0.090 * cos(3.76370412628  + 22324.90505670940 *T)
    sumB0  +=       0.091 * cos(3.95041101283  + 10042.61267559180 *T)
    sumB0  +=       0.106 * cos(4.30186500383  + 640.87760738220 *T)
    sumB0  +=       0.109 * cos(6.18873749839  + 1478.86657406440 *T)
    sumB0  +=       0.088 * cos(1.79608901332  + 6247.51311552280 *T)
    sumB0  +=       0.102 * cos(5.58754073056  + 2766.26762836500 *T)
    sumB0  +=       0.110 * cos(0.94707767481  + 3274.12501778540 *T)
    sumB0  +=       0.084 * cos(4.45487801845  + 6696.47732458460 *T)
    sumB0  +=       0.085 * cos(2.74791518135  + 3407.09983561420 *T)
    sumB0  +=       0.087 * cos(4.51145821088  + 220.41264243880 *T)
    sumB0  +=       0.101 * cos(5.94930983227  + 8425.65083781480 *T)
    sumB0  +=       0.082 * cos(0.01837230371  + 9499.25986200560 *T)
    sumB0  +=       0.080 * cos(0.42550989980  + 18052.92954315780 *T)
    sumB0  +=       0.083 * cos(2.96589752213  + 6652.77566593180 *T)
    sumB0  +=       0.080 * cos(4.61446168762  + 3914.95722503460 *T)
    sumB0  +=       0.079 * cos(1.50228636499  + 2111.65031337760 *T)
    sumB0  +=       0.089 * cos(3.52977975496  + 9485.03276800400 *T)
    sumB0  +=       0.086 * cos(0.41976545794  + 956.28915597060 *T)
    sumB0  +=       0.088 * cos(5.46013317934  + 16460.33352952499 *T)
    sumB0  +=       0.091 * cos(2.09965252231  + 949.17560896980 *T)
    sumB0  +=       0.104 * cos(1.72206104768  + 3296.89351439480 *T)
    sumB0  +=       0.103 * cos(1.25691413032  + 3384.33133900480 *T)
    sumB0  +=       0.084 * cos(5.78647729498  + 5518.75014899180 *T)
    sumB0  +=       0.079 * cos(1.79313426804  + 38.13303563780 *T)
    sumB0  +=       0.073 * cos(0.10667695992  + 29822.78323632420 *T)
    sumB0  +=       0.087 * cos(2.11654357529  + 3450.81874791920 *T)
    sumB0  +=       0.072 * cos(3.89476829327  + 9380.95967271720 *T)
    sumB0  +=       0.075 * cos(2.59340305340  + 1964.83862685400 *T)
    sumB0  +=       0.098 * cos(4.01577665825  + 6843.69148953180 *T)
    sumB0  +=       0.074 * cos(5.32032289064  + 11766.26326451460 *T)
    sumB0  +=       0.068 * cos(0.04775525953  + 2125.87740737920 *T)
    sumB0  +=       0.069 * cos(6.07427052412  + 26482.17080962440 *T)
    sumB0  +=       0.069 * cos(2.05018999200  + 29424.63423291600 *T)
    sumB0  +=       0.084 * cos(0.16960920719  + 263.08392337280 *T)
    sumB0  +=       0.068 * cos(5.03013252197  + 9070.11887384880 *T)
    sumB0  +=       0.076 * cos(2.00296087293  + 224.34479570190 *T)
    sumB0  +=       0.078 * cos(2.17362706851  + 30220.93223973240 *T)
    sumB0  +=       0.066 * cos(3.85497672006  + 19406.67828817460 *T)
    sumB0  +=       0.066 * cos(5.70059718737  + 33561.54466643220 *T)
    sumB0  +=       0.067 * cos(0.16600936321  + 22743.40937951640 *T)
    sumB0  +=       0.065 * cos(4.65423392949  + 2807.39834325620 *T)
    sumB0  +=       0.069 * cos(3.34387224268  + 11670.28403729680 *T)
    sumB0  +=       0.087 * cos(4.97838021880  + 1118.75579210280 *T)
    sumB0  +=       0.063 * cos(0.18907106180  + 30065.51184029820 *T)
    sumB0  +=       0.064 * cos(4.61909647015  + 9886.77220006400 *T)
    sumB0  +=       0.073 * cos(0.93706647938  + 20735.83216142559 *T)
    sumB0  +=       0.060 * cos(5.83757395809  + 8646.06348025360 *T)
    sumB0  +=       0.062 * cos(4.81389895867  + 20199.09495963300 *T)
    sumB0  +=       0.059 * cos(5.00150762621  + 6414.61781167780 *T)
    sumB0  +=       0.068 * cos(3.84252763135  + 6571.01853218020 *T)
    sumB0  +=       0.062 * cos(2.81689634717  + 6944.30877677240 *T)
    sumB0  +=       0.065 * cos(4.49078808776  + 632.78373931320 *T)
    sumB0  +=       0.058 * cos(5.64889513615  + 9945.57120882380 *T)
    sumB0  +=       0.070 * cos(2.51605694403  + 9638.94074787620 *T)
    sumB0  +=       0.057 * cos(3.28105791201  + 206.18554843720 *T)
    sumB0  +=       0.057 * cos(2.97448265957  + 21795.21409161479 *T)
    sumB0  +=       0.056 * cos(2.23565630779  + 20995.39296644940 *T)
    sumB0  +=       0.057 * cos(1.88614831237  + 18451.07854656599 *T)
    sumB0  +=       0.071 * cos(4.82445647307  + 8542.97070603500 *T)
    sumB0  +=       0.061 * cos(3.65945073900  + 14421.83163698840 *T)
    sumB0  +=       0.056 * cos(3.13789031275  + 8799.98871377800 *T)
    sumB0  +=       0.057 * cos(4.89927831599  + 9602.35263622420 *T)
    sumB0  +=       0.065 * cos(3.37109873211  + 11610.91017538320 *T)
    sumB0  +=       0.067 * cos(1.92945007459  + 21265.52312652020 *T)
    sumB0  +=       0.055 * cos(1.95164531764  + 9588.12554222260 *T)
    sumB0  +=       0.057 * cos(2.82240075154  + 10124.93005431800 *T)
    sumB0  +=       0.057 * cos(6.10407356832  + 19800.94595622480 *T)
    sumB0  +=       0.055 * cos(5.20976473824  + 3237.51965248120 *T)
    sumB0  +=       0.057 * cos(4.12235760406  + 10028.95082710020 *T)
    sumB0  +=       0.055 * cos(1.41700952855  + 15906.76412668260 *T)
    sumB0  +=       0.053 * cos(2.16328741039  + 6418.14093002680 *T)
    sumB0  +=       0.060 * cos(2.64683840328  + 10018.24685144760 *T)
    sumB0  +=       0.068 * cos(5.36539876845  + 1228.96211332220 *T)
    sumB0  +=       0.051 * cos(5.73824213507  + 6048.44111408640 *T)
    sumB0  +=       0.053 * cos(0.31937174553  + 12721.57209941700 *T)
    sumB0  +=       0.051 * cos(0.06312524105  + 20206.14119633100 *T)
    sumB0  +=       0.049 * cos(4.53401402385  + 6675.70192909220 *T)
    sumB0  +=       0.051 * cos(1.15475560534  + 10156.90236013480 *T)
    sumB0  +=       0.064 * cos(4.56332268770  + 16703.07938715119 *T)
    sumB0  +=       0.060 * cos(3.61007443614  + 9468.26787725700 *T)
    sumB0  +=       0.059 * cos(3.08413561767  + 10025.42770875120 *T)
    sumB0  +=       0.064 * cos(2.53229538141  + 16703.04487984680 *T)
    sumB0  +=       0.056 * cos(3.31988072467  + 6518.75821726740 *T)
    sumB0  +=       0.047 * cos(1.44559165677  + 6643.09181776180 *T)
    sumB0  +=       0.050 * cos(1.92342238827  + 11614.43329373220 *T)
    sumB0  +=       0.047 * cos(4.03794177027  + 23958.63178523340 *T)
    sumB0  +=       0.046 * cos(3.70927352724  + 8859.36257569160 *T)
    sumB0  +=       0.060 * cos(2.55506470511  + 11780.49035851620 *T)
    sumB0  +=       0.047 * cos(1.69256878711  + 6660.86953400080 *T)
    sumB0  +=       0.044 * cos(6.09481217162  + 6460.81221096080 *T)
    sumB0  +=       0.044 * cos(2.63040622140  + 13936.79450513400 *T)
    sumB0  +=       0.053 * cos(0.77878945764  + 16865.52876963120 *T)
    sumB0  +=       0.049 * cos(1.83368544550  + 17654.78053974960 *T)
    sumB0  +=       0.048 * cos(0.52828042378  + 6686.74777770700 *T)
    sumB0  +=       0.042 * cos(4.30347553493  + 9065.54812412880 *T)
    sumB0  +=       0.042 * cos(5.71964550673  + 7203.80227149340 *T)
    sumB0  +=       0.041 * cos(0.98427208931  + 20426.57109242200 *T)
    sumB0  +=       0.051 * cos(3.54335413699  + 20597.24396304120 *T)
    sumB0  +=       0.041 * cos(0.21219617682  + 7314.00859271280 *T)
    sumB0  +=       0.038 * cos(2.53074981011  + 13207.02930736500 *T)
    sumB0  +=       0.039 * cos(5.15577369902  + 6670.58818804980 *T)
    sumB0  +=       0.051 * cos(3.25271478667  + 7799.98064550240 *T)
    sumB0  +=       0.049 * cos(0.77060706107  + 17101.21113690720 *T)
    sumB0  +=       0.038 * cos(6.06684699984  + 9389.05354078620 *T)
    sumB0  +=       0.043 * cos(0.51983815091  + 16489.76303806100 *T)
    sumB0  +=       0.036 * cos(0.84102576439  + 23937.85638974100 *T)

    # MARS Latitude
    sumB1  = 0
    sumB1  +=  350068.845 * cos(5.36847836211  + 3340.61242669980 *T)
    sumB1  +=   14116.030 * cos(3.14159265359  + 0.00000000000 *T)
    sumB1  +=    9670.755 * cos(5.47877786506  + 6681.22485339960 *T)
    sumB1  +=    1471.918 * cos(3.20205766795  + 10021.83728009940 *T)
    sumB1  +=     425.864 * cos(3.40843812875  + 13362.44970679920 *T)
    sumB1  +=     102.039 * cos(0.77617286189  + 3337.08930835080 *T)
    sumB1  +=      78.848 * cos(3.71768293865  + 16703.06213349900 *T)
    sumB1  +=      26.171 * cos(2.48293558065  + 2281.23049651060 *T)
    sumB1  +=      32.708 * cos(3.45803723682  + 5621.84292321040 *T)
    sumB1  +=      20.712 * cos(1.44120802297  + 6151.53388830500 *T)
    sumB1  +=      18.294 * cos(6.03102943125  + 529.69096509460 *T)
    sumB1  +=      15.680 * cos(3.93075566599  + 8962.45534991020 *T)
    sumB1  +=      16.975 * cos(4.81115186866  + 3344.13554504880 *T)
    sumB1  +=      13.067 * cos(0.97324736181  + 6677.70173505060 *T)
    sumB1  +=      15.622 * cos(2.78241383265  + 3340.59517304760 *T)
    sumB1  +=      15.622 * cos(4.81318636318  + 3340.62968035200 *T)
    sumB1  +=      13.771 * cos(1.67983063909  + 3532.06069281140 *T)
    sumB1  +=      12.711 * cos(4.04546734935  + 20043.67456019880 *T)
    sumB1  +=      14.268 * cos(0.24640247719  + 2942.46342329160 *T)
    sumB1  +=      12.493 * cos(2.25620513522  + 5884.92684658320 *T)
    sumB1  +=       8.800 * cos(0.34079528233  + 398.14900340820 *T)
    sumB1  +=       8.637 * cos(1.75213704409  + 2544.31441988340 *T)
    sumB1  +=       8.903 * cos(5.95437916504  + 2810.92146160520 *T)
    sumB1  +=       8.102 * cos(0.84279830287  + 6283.07584999140 *T)
    sumB1  +=       9.250 * cos(4.35071778619  + 3496.03282613400 *T)
    sumB1  +=       8.085 * cos(4.29614034209  + 6684.74797174860 *T)
    sumB1  +=       5.811 * cos(3.55479498415  + 5092.15195811580 *T)
    sumB1  +=       5.864 * cos(3.69652093329  + 5486.77784317500 *T)
    sumB1  +=       5.618 * cos(0.03475872145  + 3185.19202726560 *T)
    sumB1  +=       5.148 * cos(0.85886443528  + 3340.54511639700 *T)
    sumB1  +=       5.145 * cos(0.45169791514  + 3340.67973700260 *T)
    sumB1  +=       4.898 * cos(4.12922007874  + 1059.38193018920 *T)
    sumB1  +=       3.471 * cos(5.07253485267  + 5088.62883976680 *T)
    sumB1  +=       3.102 * cos(3.11487063736  + 3339.63210563160 *T)
    sumB1  +=       3.013 * cos(1.00910437496  + 2914.01423582380 *T)
    sumB1  +=       3.171 * cos(1.24730004776  + 3738.76143010800 *T)
    sumB1  +=       3.194 * cos(2.86899092087  + 1751.53953141600 *T)
    sumB1  +=       2.304 * cos(1.38460565999  + 3870.30339179440 *T)
    sumB1  +=       2.233 * cos(2.91238180064  + 3894.18182954220 *T)
    sumB1  +=       2.792 * cos(4.27117916474  + 8432.76438481560 *T)
    sumB1  +=       2.431 * cos(0.14615904873  + 4690.47983635860 *T)
    sumB1  +=       2.577 * cos(2.78651098185  + 6254.62666252360 *T)
    sumB1  +=       1.996 * cos(1.51200589223  + 4292.33083295040 *T)
    sumB1  +=       1.860 * cos(5.15222219319  + 796.29800681640 *T)
    sumB1  +=       2.014 * cos(5.48528849094  + 3127.31333126180 *T)
    sumB1  +=       1.922 * cos(4.37668623732  + 23384.28698689860 *T)
    sumB1  +=       2.248 * cos(4.46756441051  + 10025.36039844840 *T)
    sumB1  +=       1.782 * cos(2.70229954651  + 4136.91043351620 *T)
    sumB1  +=       1.621 * cos(2.26742355334  + 3149.16416058820 *T)
    sumB1  +=       1.452 * cos(1.53948251192  + 9492.14631500480 *T)
    sumB1  +=       1.373 * cos(3.20232312108  + 2146.16541647520 *T)
    sumB1  +=       1.332 * cos(5.67966837010  + 1592.59601363280 *T)
    sumB1  +=       1.378 * cos(4.12087865376  + 3723.50895892300 *T)
    sumB1  +=       1.215 * cos(0.33914405698  + 6674.11130639880 *T)
    sumB1  +=       1.168 * cos(3.52407129935  + 11773.37681151540 *T)
    sumB1  +=       1.179 * cos(6.15475441355  + 213.29909543800 *T)
    sumB1  +=       1.173 * cos(1.26036608440  + 3341.59274776800 *T)
    sumB1  +=       1.211 * cos(0.97241747335  + 6467.92575796160 *T)
    sumB1  +=       1.204 * cos(0.95200561837  + 10018.31416175040 *T)
    sumB1  +=       1.107 * cos(1.44142157852  + 2787.04302385740 *T)
    sumB1  +=       1.194 * cos(0.84501638145  + 1194.44701022460 *T)
    sumB1  +=       1.043 * cos(3.98123209815  + 12303.06777661000 *T)
    sumB1  +=       1.029 * cos(1.21951732572  + 4535.05943692440 *T)
    sumB1  +=       0.939 * cos(2.99716248257  + 7477.52286021600 *T)
    sumB1  +=       1.303 * cos(1.12409937702  + 4399.99435688900 *T)
    sumB1  +=       1.032 * cos(4.26592917807  + 1990.74501704100 *T)
    sumB1  +=       1.230 * cos(0.37904885593  + 6525.80445396540 *T)
    sumB1  +=       0.989 * cos(2.37220445455  + 9225.53927328300 *T)
    sumB1  +=       0.946 * cos(3.14940196742  + 6489.77658728800 *T)
    sumB1  +=       1.013 * cos(5.90131661122  + 3097.88382272579 *T)
    sumB1  +=       0.913 * cos(2.86098924372  + 2388.89402044920 *T)
    sumB1  +=       0.954 * cos(4.90448639106  + 3583.34103067380 *T)
    sumB1  +=       0.866 * cos(4.92421595837  + 6681.24210705180 *T)
    sumB1  +=       0.797 * cos(4.53307543814  + 7079.37385680780 *T)
    sumB1  +=       0.866 * cos(2.89344915945  + 6681.20759974740 *T)
    sumB1  +=       0.659 * cos(0.33901474348  + 2957.71589447660 *T)
    sumB1  +=       0.670 * cos(3.23650405278  + 9595.23908922340 *T)
    sumB1  +=       0.782 * cos(1.56739177814  + 3333.49887969900 *T)
    sumB1  +=       0.641 * cos(4.46918721250  + 6836.64525283380 *T)
    sumB1  +=       0.611 * cos(3.97393774087  + 6041.32756708560 *T)
    sumB1  +=       0.738 * cos(3.78715624780  + 4562.46099302120 *T)
    sumB1  +=       0.558 * cos(2.92718433901  + 1589.07289528380 *T)
    sumB1  +=       0.513 * cos(0.96852780537  + 9623.68827669120 *T)
    sumB1  +=       0.508 * cos(4.07262127921  + 7.11354700080 *T)
    sumB1  +=       0.512 * cos(1.53960708348  + 7740.60678358880 *T)
    sumB1  +=       0.481 * cos(3.51023225206  + 155.42039943420 *T)
    sumB1  +=       0.528 * cos(6.06190721770  + 4933.20844033260 *T)
    sumB1  +=       0.558 * cos(2.09491968451  + 2288.34404351140 *T)
    sumB1  +=       0.445 * cos(3.42439886280  + 6680.24453233140 *T)
    sumB1  +=       0.482 * cos(4.73460866232  + 13365.97282514820 *T)
    sumB1  +=       0.534 * cos(4.88536251621  + 3361.38782219220 *T)
    sumB1  +=       0.520 * cos(6.19100285186  + 6438.49624942560 *T)
    sumB1  +=       0.441 * cos(1.82408206251  + 3325.35995551480 *T)
    sumB1  +=       0.487 * cos(2.72879191049  + 3319.83703120740 *T)
    sumB1  +=       0.417 * cos(4.48716253276  + 7903.07341972100 *T)
    sumB1  +=       0.430 * cos(0.62845892205  + 5614.72937620960 *T)
    sumB1  +=       0.395 * cos(4.89021898045  + 10419.98628350760 *T)
    sumB1  +=       0.380 * cos(1.91138845097  + 3553.91152213780 *T)
    sumB1  +=       0.358 * cos(4.21584518245  + 426.59819087600 *T)
    sumB1  +=       0.409 * cos(4.53421956625  + 3360.96774609859 *T)
    sumB1  +=       0.426 * cos(5.47172830908  + 3205.54734666440 *T)
    sumB1  +=       0.330 * cos(1.00078638665  + 8955.34180290940 *T)
    sumB1  +=       0.322 * cos(2.42607399606  + 2274.11694950980 *T)
    sumB1  +=       0.316 * cos(3.22064089430  + 1221.84856632140 *T)
    sumB1  +=       0.312 * cos(3.66073476074  + 3337.02199804800 *T)
    sumB1  +=       0.310 * cos(4.09100416919  + 3344.20285535160 *T)
    sumB1  +=       0.295 * cos(2.87635162282  + 7210.91581849420 *T)
    sumB1  +=       0.329 * cos(5.64440592793  + 12832.75874170460 *T)
    sumB1  +=       0.306 * cos(1.52567836482  + 7875.67186362420 *T)
    sumB1  +=       0.284 * cos(0.56373187712  + 6681.29216370240 *T)
    sumB1  +=       0.358 * cos(4.21258737082  + 1052.26838318840 *T)
    sumB1  +=       0.347 * cos(1.52200237477  + 10213.28554621100 *T)
    sumB1  +=       0.268 * cos(3.57749600251  + 9830.38901398780 *T)
    sumB1  +=       0.346 * cos(1.65247014273  + 6682.20517446780 *T)
    sumB1  +=       0.269 * cos(4.45182417080  + 9381.93999378540 *T)
    sumB1  +=       0.285 * cos(0.97118840317  + 6681.15754309680 *T)
    sumB1  +=       0.256 * cos(3.67744143301  + 1066.49547719000 *T)
    sumB1  +=       0.269 * cos(1.78448639785  + 6127.65545055720 *T)
    sumB1  +=       0.312 * cos(4.63751658715  + 5331.35744374080 *T)
    sumB1  +=       0.317 * cos(6.15274242841  + 3320.25710730100 *T)
    sumB1  +=       0.245 * cos(4.71030710599  + 26724.89941359840 *T)
    sumB1  +=       0.249 * cos(3.31328995337  + 10818.13528691580 *T)
    sumB1  +=       0.240 * cos(4.87364672707  + 6144.42034130420 *T)
    sumB1  +=       0.239 * cos(1.35709001659  + 7064.12138562280 *T)
    sumB1  +=       0.237 * cos(3.19837233399  + 5729.50644714900 *T)
    sumB1  +=       0.278 * cos(1.50842060340  + 4929.68532198360 *T)
    sumB1  +=       0.274 * cos(3.61160906396  + 191.44826611160 *T)
    sumB1  +=       0.248 * cos(3.93789974497  + 11243.68584642080 *T)
    sumB1  +=       0.242 * cos(2.57630671866  + 3355.86489788480 *T)
    sumB1  +=       0.241 * cos(1.89683861728  + 15643.68020330980 *T)
    sumB1  +=       0.228 * cos(0.71217112323  + 9866.41688066520 *T)
    sumB1  +=       0.246 * cos(0.16913226579  + 2700.71514038580 *T)
    sumB1  +=       0.219 * cos(0.03864812268  + 522.57741809380 *T)
    sumB1  +=       0.215 * cos(3.44260776071  + 8827.39026987480 *T)
    sumB1  +=       0.206 * cos(5.23615052385  + 6923.95345737360 *T)
    sumB1  +=       0.202 * cos(0.60277862639  + 10021.82002644720 *T)
    sumB1  +=       0.226 * cos(2.43332503270  + 8429.24126646660 *T)
    sumB1  +=       0.202 * cos(2.63373646725  + 10021.85453375160 *T)
    sumB1  +=       0.199 * cos(3.26545613445  + 382.89653222320 *T)
    sumB1  +=       0.198 * cos(4.50685315424  + 242.72860397400 *T)
    sumB1  +=       0.215 * cos(2.12744795327  + 553.56940284240 *T)
    sumB1  +=       0.177 * cos(3.92599588022  + 8031.09226305840 *T)
    sumB1  +=       0.162 * cos(3.94498519498  + 2921.12778282460 *T)
    sumB1  +=       0.199 * cos(4.22948972595  + 3312.16323923200 *T)
    sumB1  +=       0.188 * cos(0.67072289565  + 6298.32832117640 *T)
    sumB1  +=       0.160 * cos(2.90951395323  + 6872.67311951120 *T)
    sumB1  +=       0.155 * cos(1.24529852403  + 3364.49086444760 *T)
    sumB1  +=       0.186 * cos(2.44723831367  + 3503.07906283200 *T)
    sumB1  +=       0.139 * cos(5.75493557040  + 7632.94325965020 *T)
    sumB1  +=       0.149 * cos(0.49665393273  + 8969.56889691100 *T)
    sumB1  +=       0.132 * cos(4.97828413367  + 1748.01641306700 *T)
    sumB1  +=       0.130 * cos(0.66881298338  + 10014.72373309860 *T)
    sumB1  +=       0.128 * cos(2.80753554051  + 14584.29827312060 *T)
    sumB1  +=       0.176 * cos(2.37291314099  + 2118.76386037840 *T)
    sumB1  +=       0.127 * cos(1.76223414864  + 11081.21921028860 *T)
    sumB1  +=       0.127 * cos(0.26234206855  + 13358.92658845020 *T)
    sumB1  +=       0.163 * cos(5.83356697025  + 639.89728631400 *T)
    sumB1  +=       0.129 * cos(2.02607662846  + 3346.13535100720 *T)
    sumB1  +=       0.155 * cos(1.57189307640  + 3767.21061757580 *T)
    sumB1  +=       0.133 * cos(4.20989922795  + 11371.70468975820 *T)
    sumB1  +=       0.114 * cos(2.12636383988  + 6688.33840040040 *T)
    sumB1  +=       0.130 * cos(1.33114943655  + 3347.72597370060 *T)
    sumB1  +=       0.110 * cos(5.78893316282  + 16173.37116840440 *T)
    sumB1  +=       0.105 * cos(3.27202438053  + 15113.98923821520 *T)
    sumB1  +=       0.117 * cos(5.66812806862  + 536.80451209540 *T)
    sumB1  +=       0.106 * cos(1.21268139587  + 7234.79425624200 *T)
    sumB1  +=       0.108 * cos(0.94652521237  + 9808.53818466140 *T)
    sumB1  +=       0.121 * cos(4.22204758443  + 6158.64743530580 *T)
    sumB1  +=       0.124 * cos(4.92495255567  + 12964.30070339100 *T)
    sumB1  +=       0.142 * cos(3.85764234922  + 6894.52394883760 *T)
    sumB1  +=       0.099 * cos(5.62543167000  + 10973.55568635000 *T)
    sumB1  +=       0.100 * cos(2.56456804207  + 3178.14579056760 *T)
    sumB1  +=       0.131 * cos(0.08971075365  + 8273.82086703240 *T)
    sumB1  +=       0.112 * cos(3.14240135508  + 3316.73398895200 *T)
    sumB1  +=       0.095 * cos(0.64723185059  + 2487.41604494780 *T)
    sumB1  +=       0.087 * cos(4.14322802003  + 3603.69635007260 *T)
    sumB1  +=       0.087 * cos(3.50260474148  + 3077.52850332700 *T)
    sumB1  +=       0.096 * cos(5.13010621713  + 3304.58456002240 *T)
    sumB1  +=       0.082 * cos(4.87880877413  + 5828.02847164760 *T)
    sumB1  +=       0.083 * cos(2.04400907684  + 18984.29263000960 *T)
    sumB1  +=       0.085 * cos(3.33671991814  + 1758.65307841680 *T)
    sumB1  +=       0.099 * cos(3.28955694324  + 17654.78053974960 *T)
    sumB1  +=       0.078 * cos(0.89238228021  + 2699.73481931760 *T)
    sumB1  +=       0.095 * cos(0.32948273562  + 3863.18984479360 *T)
    sumB1  +=       0.083 * cos(0.80327899268  + 5085.03841111500 *T)
    sumB1  +=       0.091 * cos(1.48623539837  + 8270.29774868340 *T)
    sumB1  +=       0.075 * cos(4.91191036053  + 10021.76996979660 *T)
    sumB1  +=       0.079 * cos(4.25171159471  + 7373.38245462640 *T)
    sumB1  +=       0.084 * cos(2.16662156833  + 12566.15169998280 *T)
    sumB1  +=       0.088 * cos(2.78893554858  + 11769.85369316640 *T)
    sumB1  +=       0.095 * cos(0.29428946155  + 5628.95647021120 *T)
    sumB1  +=       0.076 * cos(5.23354729169  + 13760.59871020740 *T)
    sumB1  +=       0.092 * cos(1.40883592065  + 3336.73109134180 *T)
    sumB1  +=       0.069 * cos(3.50477916910  + 10713.99488132620 *T)
    sumB1  +=       0.079 * cos(0.61833410504  + 951.71840625060 *T)
    sumB1  +=       0.070 * cos(1.93458925881  + 3475.67750673520 *T)
    sumB1  +=       0.066 * cos(5.01996258363  + 16706.58525184800 *T)
    sumB1  +=       0.067 * cos(3.68731475523  + 4032.77002792660 *T)
    sumB1  +=       0.071 * cos(1.71691035008  + 10575.40668294180 *T)
    sumB1  +=       0.086 * cos(1.82073461894  + 10404.73381232260 *T)
    sumB1  +=       0.064 * cos(2.43149744819  + 2648.45482547300 *T)
    sumB1  +=       0.062 * cos(0.25250107112  + 3980.50971301380 *T)
    sumB1  +=       0.062 * cos(3.16420520010  + 13517.87010623340 *T)
    sumB1  +=       0.078 * cos(1.41762772678  + 3.52311834900 *T)
    sumB1  +=       0.086 * cos(3.34935714534  + 1581.95934828300 *T)
    sumB1  +=       0.064 * cos(3.75812808116  + 1596.18644228460 *T)
    sumB1  +=       0.061 * cos(1.37421861116  + 419.48464387520 *T)
    sumB1  +=       0.055 * cos(1.11154560464  + 433.71173787680 *T)
    sumB1  +=       0.061 * cos(5.33515994167  + 853.19638175200 *T)
    sumB1  +=       0.059 * cos(0.80382380790  + 13362.43245314700 *T)
    sumB1  +=       0.055 * cos(0.37478855406  + 149.56319713460 *T)
    sumB1  +=       0.053 * cos(0.55329112916  + 14314.16811304980 *T)
    sumB1  +=       0.059 * cos(0.07234691744  + 6531.66165626500 *T)
    sumB1  +=       0.059 * cos(2.83458587563  + 13362.46696045140 *T)
    sumB1  +=       0.059 * cos(4.13409540555  + 4407.10790388980 *T)
    sumB1  +=       0.057 * cos(5.57913927474  + 3376.64029337720 *T)
    sumB1  +=       0.049 * cos(4.58036235824  + 10551.52824519400 *T)
    sumB1  +=       0.050 * cos(2.01769015785  + 10022.81760116760 *T)
    sumB1  +=       0.055 * cos(1.36081433917  + 21393.54196985760 *T)
    sumB1  +=       0.048 * cos(1.59939595487  + 18052.92954315780 *T)
    sumB1  +=       0.045 * cos(4.97928157994  + 8671.96987044060 *T)
    sumB1  +=       0.044 * cos(2.58542985685  + 13916.01910964160 *T)
    sumB1  +=       0.047 * cos(1.40206686464  + 8425.65083781480 *T)
    sumB1  +=       0.043 * cos(0.21864618144  + 9779.10867612540 *T)
    sumB1  +=       0.042 * cos(6.22365857047  + 3914.95722503460 *T)
    sumB1  +=       0.048 * cos(3.44104044998  + 692.15760122680 *T)
    sumB1  +=       0.043 * cos(3.21191030055  + 6247.51311552280 *T)
    sumB1  +=       0.055 * cos(4.01321481700  + 10177.25767953360 *T)
    sumB1  +=       0.053 * cos(2.93285711320  + 3981.49003408200 *T)
    sumB1  +=       0.039 * cos(5.54634215618  + 6702.00024889200 *T)
    sumB1  +=       0.041 * cos(2.46188112193  + 6660.44945790720 *T)
    sumB1  +=       0.049 * cos(3.04236065330  + 640.87760738220 *T)
    sumB1  +=       0.039 * cos(1.62689158397  + 3335.08950239240 *T)
    sumB1  +=       0.038 * cos(3.31198341709  + 1744.42598441520 *T)
    sumB1  +=       0.038 * cos(6.16317067723  + 2818.03500860600 *T)
    sumB1  +=       0.038 * cos(4.03534957207  + 103.09277421860 *T)
    sumB1  +=       0.044 * cos(0.91040525278  + 5459.37628707820 *T)
    sumB1  +=       0.049 * cos(1.88325571471  + 11216.28429032400 *T)
    sumB1  +=       0.046 * cos(5.62762941168  + 9168.64089834740 *T)
    sumB1  +=       0.035 * cos(1.65976191407  + 12168.00269657460 *T)
    sumB1  +=       0.047 * cos(4.52334272666  + 10021.90459040220 *T)
    sumB1  +=       0.039 * cos(2.08367076229  + 13745.34623902240 *T)
    sumB1  +=       0.039 * cos(4.76186135572  + 20597.24396304120 *T)
    sumB1  +=       0.039 * cos(3.00951950244  + 3344.49376205780 *T)
    sumB1  +=       0.037 * cos(1.33065895739  + 12295.95422960920 *T)
    sumB1  +=       0.040 * cos(3.96893429107  + 13171.00144068760 *T)
    sumB1  +=       0.032 * cos(0.91212661789  + 12310.18132361080 *T)
    sumB1  +=       0.032 * cos(2.22618836835  + 7107.82304427560 *T)
    sumB1  +=       0.032 * cos(4.96078703650  + 24734.15439655740 *T)
    sumB1  +=       0.037 * cos(3.30085609301  + 20995.39296644940 *T)
    sumB1  +=       0.033 * cos(3.70858397913  + 10020.85695903120 *T)
    sumB1  +=       0.031 * cos(4.18674524484  + 17256.63153634140 *T)
    sumB1  +=       0.034 * cos(3.25657016622  + 14054.60730802600 *T)
    sumB1  +=       0.032 * cos(5.75360661734  + 6546.15977336420 *T)
    sumB1  +=       0.031 * cos(3.16209740784  + 12935.85151592320 *T)
    sumB1  +=       0.037 * cos(2.14901622101  + 6665.97238221460 *T)
    sumB1  +=       0.030 * cos(0.21926429217  + 6660.86953400080 *T)
    sumB1  +=       0.030 * cos(3.55459460301  + 9070.11887384880 *T)
    sumB1  +=       0.029 * cos(2.97927335664  + 2766.26762836500 *T)
    sumB1  +=       0.029 * cos(4.65051930186  + 10235.13637553740 *T)
    sumB1  +=       0.030 * cos(5.16882435400  + 16304.91313009080 *T)
    sumB1  +=       0.032 * cos(5.01760544488  + 3191.04922956520 *T)
    sumB1  +=       0.029 * cos(2.93988761068  + 6696.47732458460 *T)
    sumB1  +=       0.029 * cos(5.06447742493  + 5099.26550511660 *T)
    sumB1  +=       0.029 * cos(5.68073325802  + 3329.97576135000 *T)
    sumB1  +=       0.029 * cos(0.43949269085  + 3351.24909204960 *T)
    sumB1  +=       0.032 * cos(1.37056688814  + 6040.34724601740 *T)
    sumB1  +=       0.032 * cos(1.42334455474  + 76.26607127560 *T)
    sumB1  +=       0.033 * cos(0.36423446435  + 6604.95878212400 *T)
    sumB1  +=       0.033 * cos(4.85478789063  + 6701.58017279840 *T)
    sumB1  +=       0.030 * cos(3.33348460172  + 3369.06161416760 *T)
    sumB1  +=       0.028 * cos(0.60182098585  + 11766.26326451460 *T)
    sumB1  +=       0.026 * cos(1.05197848649  + 13207.02930736500 *T)
    sumB1  +=       0.026 * cos(5.03535225585  + 30065.51184029820 *T)
    sumB1  +=       0.034 * cos(3.65553061498  + 14158.74771361560 *T)
    sumB1  +=       0.025 * cos(4.60925601393  + 8329.67161059700 *T)
    sumB1  +=       0.025 * cos(2.05970140678  + 17395.21973472580 *T)
    sumB1  +=       0.030 * cos(5.22285260441  + 9485.03276800400 *T)
    sumB1  +=       0.025 * cos(4.84964369679  + 1648.44675719740 *T)
    sumB1  +=       0.026 * cos(5.47626814223  + 3074.00538497800 *T)
    sumB1  +=       0.024 * cos(2.45239543931  + 17085.95866572220 *T)
    sumB1  +=       0.029 * cos(5.36428737888  + 2707.82868738660 *T)
    sumB1  +=       0.024 * cos(5.57488544800  + 10264.56588407340 *T)
    sumB1  +=       0.025 * cos(4.22744782935  + 1692.16566950240 *T)
    sumB1  +=       0.023 * cos(6.09244870438  + 20199.09495963300 *T)
    sumB1  +=       0.028 * cos(1.30366587075  + 8439.87793181640 *T)
    sumB1  +=       0.032 * cos(2.13597148493  + 9468.26787725700 *T)

    # MARS Latitude
    sumB2  = 0
    sumB2  +=   16726.690 * cos(0.60221392419  + 3340.61242669980 *T)
    sumB2  +=    4986.799 * cos(3.14159265359  + 0.00000000000 *T)
    sumB2  +=     302.141 * cos(5.55871276021  + 6681.22485339960 *T)
    sumB2  +=      25.767 * cos(1.89662673499  + 13362.44970679920 *T)
    sumB2  +=      21.452 * cos(0.91749968618  + 10021.83728009940 *T)
    sumB2  +=      11.820 * cos(2.24240738700  + 3337.08930835080 *T)
    sumB2  +=       7.985 * cos(2.24892866611  + 16703.06213349900 *T)
    sumB2  +=       2.960 * cos(5.89425825808  + 3496.03282613400 *T)
    sumB2  +=       2.445 * cos(5.18770525274  + 5621.84292321040 *T)
    sumB2  +=       1.428 * cos(1.25238140580  + 2281.23049651060 *T)
    sumB2  +=       1.779 * cos(2.58759968520  + 20043.67456019880 *T)
    sumB2  +=       1.501 * cos(3.18533003542  + 3532.06069281140 *T)
    sumB2  +=       1.259 * cos(4.80695172904  + 3185.19202726560 *T)
    sumB2  +=       1.029 * cos(2.35029907056  + 6677.70173505060 *T)
    sumB2  +=       1.109 * cos(3.80982317372  + 5884.92684658320 *T)
    sumB2  +=       0.928 * cos(0.29719160927  + 3344.13554504880 *T)
    sumB2  +=       0.860 * cos(3.11698318930  + 6151.53388830500 *T)
    sumB2  +=       0.853 * cos(1.33003321402  + 529.69096509460 *T)
    sumB2  +=       0.883 * cos(5.18681316017  + 5486.77784317500 *T)
    sumB2  +=       1.100 * cos(1.82962075794  + 2942.46342329160 *T)
    sumB2  +=       0.815 * cos(3.40910567373  + 2544.31441988340 *T)
    sumB2  +=       0.702 * cos(5.49076132554  + 8962.45534991020 *T)
    sumB2  +=       0.644 * cos(0.31400306761  + 5088.62883976680 *T)
    sumB2  +=       0.745 * cos(4.30248377111  + 3340.59517304760 *T)
    sumB2  +=       0.586 * cos(4.48071459693  + 3894.18182954220 *T)
    sumB2  +=       0.681 * cos(5.40951261308  + 2810.92146160520 *T)
    sumB2  +=       0.745 * cos(0.05006463495  + 3340.62968035200 *T)
    sumB2  +=       0.542 * cos(4.68895461946  + 3339.63210563160 *T)
    sumB2  +=       0.497 * cos(4.65665197749  + 5092.15195811580 *T)
    sumB2  +=       0.460 * cos(1.06670343610  + 3097.88382272579 *T)
    sumB2  +=       0.423 * cos(3.08034568777  + 4292.33083295040 *T)
    sumB2  +=       0.381 * cos(2.24933115190  + 6283.07584999140 *T)
    sumB2  +=       0.429 * cos(1.70668733378  + 4690.47983635860 *T)
    sumB2  +=       0.339 * cos(2.92195933142  + 23384.28698689860 *T)
    sumB2  +=       0.310 * cos(5.63324402105  + 3723.50895892300 *T)
    sumB2  +=       0.287 * cos(1.98619629868  + 398.14900340820 *T)
    sumB2  +=       0.298 * cos(2.58990636818  + 6684.74797174860 *T)
    sumB2  +=       0.323 * cos(5.15230144390  + 6525.80445396540 *T)
    sumB2  +=       0.247 * cos(2.36923535528  + 3340.54511639700 *T)
    sumB2  +=       0.247 * cos(1.96071951597  + 3340.67973700260 *T)
    sumB2  +=       0.223 * cos(4.75562842441  + 2146.16541647520 *T)
    sumB2  +=       0.240 * cos(3.53331653789  + 3583.34103067380 *T)
    sumB2  +=       0.238 * cos(4.69268478666  + 6254.62666252360 *T)
    sumB2  +=       0.198 * cos(6.21548965568  + 2787.04302385740 *T)
    sumB2  +=       0.202 * cos(4.25439775032  + 3333.49887969900 *T)
    sumB2  +=       0.158 * cos(1.89216973387  + 6674.11130639880 *T)
    sumB2  +=       0.185 * cos(2.99350590687  + 10025.36039844840 *T)
    sumB2  +=       0.177 * cos(1.41002572971  + 6438.49624942560 *T)
    sumB2  +=       0.138 * cos(1.23172102079  + 3127.31333126180 *T)
    sumB2  +=       0.153 * cos(5.93528616595  + 6836.64525283380 *T)
    sumB2  +=       0.106 * cos(5.35414400998  + 3738.76143010800 *T)
    sumB2  +=       0.102 * cos(3.67040445208  + 1059.38193018920 *T)
    sumB2  +=       0.136 * cos(1.06039656714  + 2388.89402044920 *T)
    sumB2  +=       0.093 * cos(5.04434801864  + 155.42039943420 *T)
    sumB2  +=       0.093 * cos(2.70817394910  + 8955.34180290940 *T)
    sumB2  +=       0.108 * cos(6.13917250674  + 1748.01641306700 *T)
    sumB2  +=       0.116 * cos(5.97739483530  + 2914.01423582380 *T)
    sumB2  +=       0.099 * cos(1.50221319099  + 1751.53953141600 *T)
    sumB2  +=       0.092 * cos(2.68470906437  + 1990.74501704100 *T)
    sumB2  +=       0.077 * cos(4.01966292109  + 1592.59601363280 *T)
    sumB2  +=       0.076 * cos(4.49660101731  + 4562.46099302120 *T)
    sumB2  +=       0.076 * cos(5.67354102576  + 6041.32756708560 *T)
    sumB2  +=       0.078 * cos(0.79220883728  + 2288.34404351140 *T)
    sumB2  +=       0.074 * cos(5.84795427965  + 3341.59274776800 *T)
    sumB2  +=       0.080 * cos(5.48556941416  + 8432.76438481560 *T)
    sumB2  +=       0.065 * cos(2.11705931744  + 10018.31416175040 *T)
    sumB2  +=       0.066 * cos(5.08073805943  + 2957.71589447660 *T)
    sumB2  +=       0.059 * cos(3.79695096075  + 6923.95345737360 *T)
    sumB2  +=       0.061 * cos(4.51116100631  + 4933.20844033260 *T)
    sumB2  +=       0.059 * cos(0.27765984344  + 6127.65545055720 *T)
    sumB2  +=       0.052 * cos(5.84885902302  + 4535.05943692440 *T)
    sumB2  +=       0.051 * cos(1.45495904398  + 7477.52286021600 *T)
    sumB2  +=       0.054 * cos(1.50278239577  + 6489.77658728800 *T)
    sumB2  +=       0.056 * cos(5.23212313891  + 9866.41688066520 *T)
    sumB2  +=       0.057 * cos(4.73711357046  + 2274.11694950980 *T)
    sumB2  +=       0.051 * cos(2.48654805044  + 6467.92575796160 *T)
    sumB2  +=       0.047 * cos(3.25899585469  + 26724.89941359840 *T)
    sumB2  +=       0.048 * cos(1.12255878532  + 1349.86740965880 *T)
    sumB2  +=       0.048 * cos(2.87122326236  + 242.72860397400 *T)
    sumB2  +=       0.045 * cos(1.61358524585  + 5729.50644714900 *T)
    sumB2  +=       0.043 * cos(2.24334731490  + 7.11354700080 *T)
    sumB2  +=       0.044 * cos(4.61065704096  + 11773.37681151540 *T)
    sumB2  +=       0.042 * cos(3.32038536161  + 13365.97282514820 *T)
    sumB2  +=       0.047 * cos(5.10736479739  + 8969.56889691100 *T)
    sumB2  +=       0.044 * cos(4.64194663616  + 4399.99435688900 *T)
    sumB2  +=       0.039 * cos(5.45779680648  + 8031.09226305840 *T)
    sumB2  +=       0.044 * cos(5.21409542420  + 4136.91043351620 *T)
    sumB2  +=       0.045 * cos(4.26206063790  + 796.29800681640 *T)
    sumB2  +=       0.048 * cos(3.04342098072  + 5331.35744374080 *T)
    sumB2  +=       0.042 * cos(2.12882700518  + 213.29909543800 *T)
    sumB2  +=       0.038 * cos(2.49365585516  + 951.71840625060 *T)
    sumB2  +=       0.037 * cos(3.44816851943  + 3325.35995551480 *T)
    sumB2  +=       0.038 * cos(3.13981385980  + 3149.16416058820 *T)
    sumB2  +=       0.039 * cos(3.64984946180  + 553.56940284240 *T)
    sumB2  +=       0.036 * cos(0.54858124109  + 7632.94325965020 *T)
    sumB2  +=       0.032 * cos(4.29985057106  + 3355.86489788480 *T)
    sumB2  +=       0.037 * cos(6.26349600634  + 6682.20517446780 *T)
    sumB2  +=       0.037 * cos(3.83188032092  + 3870.30339179440 *T)
    sumB2  +=       0.031 * cos(4.94959395405  + 6680.24453233140 *T)
    sumB2  +=       0.035 * cos(6.14813219827  + 3360.96774609859 *T)
    sumB2  +=       0.029 * cos(0.54115054572  + 6681.29216370240 *T)
    sumB2  +=       0.035 * cos(1.49899662719  + 3320.25710730100 *T)
    sumB2  +=       0.033 * cos(0.19183878029  + 3347.72597370060 *T)
    sumB2  +=       0.036 * cos(3.28055425527  + 1589.07289528380 *T)
    sumB2  +=       0.031 * cos(5.66015228291  + 1066.49547719000 *T)
    sumB2  +=       0.028 * cos(4.18178455540  + 7210.91581849420 *T)
    sumB2  +=       0.025 * cos(1.21627749818  + 9492.14631500480 *T)
    sumB2  +=       0.029 * cos(2.02283118033  + 7234.79425624200 *T)
    sumB2  +=       0.024 * cos(2.96777681837  + 7064.12138562280 *T)
    sumB2  +=       0.024 * cos(2.16823248221  + 10419.98628350760 *T)
    sumB2  +=       0.022 * cos(4.67496434823  + 1194.44701022460 *T)
    sumB2  +=       0.022 * cos(5.01797528499  + 9595.23908922340 *T)
    sumB2  +=       0.028 * cos(4.19843924702  + 8429.24126646660 *T)
    sumB2  +=       0.023 * cos(6.21195010050  + 7740.60678358880 *T)
    sumB2  +=       0.025 * cos(1.45100528915  + 6872.67311951120 *T)
    sumB2  +=       0.024 * cos(1.93342510858  + 2699.73481931760 *T)
    sumB2  +=       0.028 * cos(5.87009898150  + 1052.26838318840 *T)
    sumB2  +=       0.025 * cos(2.49707182259  + 2118.76386037840 *T)
    sumB2  +=       0.020 * cos(0.15682384975  + 3205.54734666440 *T)
    sumB2  +=       0.019 * cos(3.82087005916  + 12303.06777661000 *T)
    sumB2  +=       0.020 * cos(6.26534330674  + 7875.67186362420 *T)
    sumB2  +=       0.019 * cos(0.99282533197  + 522.57741809380 *T)
    sumB2  +=       0.018 * cos(4.11668144665  + 426.59819087600 *T)
    sumB2  +=       0.020 * cos(3.28531167827  + 9225.53927328300 *T)
    sumB2  +=       0.019 * cos(5.72169822815  + 11371.70468975820 *T)
    sumB2  +=       0.018 * cos(4.76871968197  + 382.89653222320 *T)
    sumB2  +=       0.017 * cos(0.96519535826  + 191.44826611160 *T)
    sumB2  +=       0.017 * cos(0.86184324841  + 10973.55568635000 *T)
    sumB2  +=       0.018 * cos(6.22706341047  + 9381.93999378540 *T)
    sumB2  +=       0.015 * cos(0.54135050421  + 640.87760738220 *T)

    # MARS Latitude
    sumB3  = 0
    sumB3  +=     606.506 * cos(1.98050633529  + 3340.61242669980 *T)
    sumB3  +=      42.611 * cos(0.00000000000  + 0.00000000000 *T)
    sumB3  +=      13.652 * cos(1.79588228800  + 6681.22485339960 *T)
    sumB3  +=       2.730 * cos(3.45377082121  + 10021.83728009940 *T)
    sumB3  +=       0.929 * cos(3.75226159072  + 3337.08930835080 *T)
    sumB3  +=       0.607 * cos(0.10618486408  + 13362.44970679920 *T)
    sumB3  +=       0.617 * cos(1.14471772765  + 3496.03282613400 *T)
    sumB3  +=       0.479 * cos(0.70504966293  + 16703.06213349900 *T)
    sumB3  +=       0.185 * cos(3.28778562029  + 3185.19202726560 *T)
    sumB3  +=       0.169 * cos(0.29980532608  + 5621.84292321040 *T)
    sumB3  +=       0.158 * cos(1.09025317222  + 20043.67456019880 *T)
    sumB3  +=       0.123 * cos(2.55664973413  + 3097.88382272579 *T)
    sumB3  +=       0.126 * cos(4.74517022983  + 3532.06069281140 *T)
    sumB3  +=       0.093 * cos(6.03607685759  + 3894.18182954220 *T)
    sumB3  +=       0.099 * cos(4.92257049901  + 2544.31441988340 *T)
    sumB3  +=       0.106 * cos(3.41315845439  + 2942.46342329160 *T)
    sumB3  +=       0.062 * cos(6.25738949900  + 3339.63210563160 *T)
    sumB3  +=       0.058 * cos(4.64867983361  + 4292.33083295040 *T)
    sumB3  +=       0.050 * cos(3.75358626972  + 6677.70173505060 *T)
    sumB3  +=       0.050 * cos(1.89408668049  + 5088.62883976680 *T)
    sumB3  +=       0.058 * cos(3.65295480755  + 6525.80445396540 *T)
    sumB3  +=       0.045 * cos(0.38295906298  + 5486.77784317500 *T)
    sumB3  +=       0.052 * cos(3.25535694335  + 4690.47983635860 *T)
    sumB3  +=       0.044 * cos(1.89342588822  + 3583.34103067380 *T)
    sumB3  +=       0.037 * cos(0.04871920725  + 2146.16541647520 *T)
    sumB3  +=       0.039 * cos(0.91787211117  + 3723.50895892300 *T)
    sumB3  +=       0.051 * cos(5.24063514394  + 5884.92684658320 *T)
    sumB3  +=       0.036 * cos(2.79633276263  + 529.69096509460 *T)
    sumB3  +=       0.032 * cos(1.44641701752  + 23384.28698689860 *T)
    sumB3  +=       0.024 * cos(5.41362501363  + 3340.59517304760 *T)
    sumB3  +=       0.023 * cos(4.70639896508  + 2787.04302385740 *T)
    sumB3  +=       0.021 * cos(1.12581894224  + 3340.62968035200 *T)
    sumB3  +=       0.021 * cos(0.51744721589  + 8962.45534991020 *T)
    sumB3  +=       0.018 * cos(0.36433365806  + 155.42039943420 *T)
    sumB3  +=       0.018 * cos(5.53170427847  + 3333.49887969900 *T)
    sumB3  +=       0.019 * cos(4.34401282378  + 6151.53388830500 *T)
    sumB3  +=       0.021 * cos(2.94765501590  + 398.14900340820 *T)
    sumB3  +=       0.020 * cos(5.96181389299  + 5092.15195811580 *T)
    sumB3  +=       0.017 * cos(0.80891817293  + 1194.44701022460 *T)
    sumB3  +=       0.011 * cos(0.43817745316  + 1059.38193018920 *T)
    sumB3  +=       0.014 * cos(1.06654064248  + 6836.64525283380 *T)

    # MARS Latitude
    sumB4  = 0
    sumB4  +=      11.334 * cos(3.45724352586  + 3340.61242669980 *T)
    sumB4  +=      13.369 * cos(0.00000000000  + 0.00000000000 *T)
    sumB4  +=       0.744 * cos(0.50445805257  + 6681.22485339960 *T)
    sumB4  +=       0.148 * cos(1.05056602649  + 10021.83728009940 *T)
    sumB4  +=       0.102 * cos(2.66185835593  + 3496.03282613400 *T)
    sumB4  +=       0.053 * cos(5.27888218929  + 3337.08930835080 *T)
    sumB4  +=       0.022 * cos(4.09971603267  + 3097.88382272579 *T)
    sumB4  +=       0.023 * cos(1.72422342130  + 3185.19202726560 *T)
    sumB4  +=       0.013 * cos(5.41704779112  + 16703.06213349900 *T)
    sumB4  +=       0.013 * cos(2.06957065662  + 13362.44970679920 *T)
    sumB4  +=       0.010 * cos(0.95989595600  + 3894.18182954220 *T)

    # MARS Latitude
    sumB5  = 0
    sumB5  +=       0.457 * cos(4.86794125358  + 3340.61242669980 *T)
    sumB5  +=       0.053 * cos(5.30547050586  + 6681.22485339960 *T)
    sumB5  +=       0.012 * cos(5.75114070583  + 10021.83728009940 *T)
    sumB5  +=       0.013 * cos(4.17736925293  + 3496.03282613400 *T)
    sumB5  +=       0.007 * cos(0.00000000000  + 0.00000000000 *T)

    # MARS Radius
    sumR0  = 0
    sumR0  +=  153033488.276 * cos(0.00000000000  + 0.00000000000 *T)
    sumR0  +=  14184953.153 * cos(3.47971283519  + 3340.61242669980 *T)
    sumR0  +=  660776.357 * cos(3.81783442097  + 6681.22485339960 *T)
    sumR0  +=   46179.117 * cos(4.15595316284  + 10021.83728009940 *T)
    sumR0  +=    8109.738 * cos(5.55958460165  + 2810.92146160520 *T)
    sumR0  +=    7485.315 * cos(1.77238998069  + 5621.84292321040 *T)
    sumR0  +=    5523.193 * cos(1.36436318880  + 2281.23049651060 *T)
    sumR0  +=    3825.160 * cos(4.49407182408  + 13362.44970679920 *T)
    sumR0  +=    2306.539 * cos(0.09081742493  + 2544.31441988340 *T)
    sumR0  +=    1999.399 * cos(5.36059605227  + 3337.08930835080 *T)
    sumR0  +=    2484.385 * cos(4.92545577893  + 2942.46342329160 *T)
    sumR0  +=    1960.198 * cos(4.74249386323  + 3344.13554504880 *T)
    sumR0  +=    1167.115 * cos(2.11261501155  + 5092.15195811580 *T)
    sumR0  +=    1102.828 * cos(5.00908264160  + 398.14900340820 *T)
    sumR0  +=     899.077 * cos(4.40790433994  + 529.69096509460 *T)
    sumR0  +=     992.252 * cos(5.83862401067  + 6151.53388830500 *T)
    sumR0  +=     807.348 * cos(2.10216647104  + 1059.38193018920 *T)
    sumR0  +=     797.910 * cos(3.44839026172  + 796.29800681640 *T)
    sumR0  +=     740.980 * cos(1.49906336892  + 2146.16541647520 *T)
    sumR0  +=     692.340 * cos(2.13378814785  + 8962.45534991020 *T)
    sumR0  +=     633.144 * cos(0.89353285018  + 3340.59517304760 *T)
    sumR0  +=     725.583 * cos(1.24516913473  + 8432.76438481560 *T)
    sumR0  +=     633.140 * cos(2.92430448169  + 3340.62968035200 *T)
    sumR0  +=     574.352 * cos(0.82896196337  + 2914.01423582380 *T)
    sumR0  +=     526.187 * cos(5.38292276228  + 3738.76143010800 *T)
    sumR0  +=     629.976 * cos(1.28738135858  + 1751.53953141600 *T)
    sumR0  +=     472.776 * cos(5.19850457873  + 3127.31333126180 *T)
    sumR0  +=     348.095 * cos(4.83219198908  + 16703.06213349900 *T)
    sumR0  +=     283.702 * cos(2.90692294913  + 3532.06069281140 *T)
    sumR0  +=     279.552 * cos(5.25749247548  + 6283.07584999140 *T)
    sumR0  +=     233.827 * cos(5.10546492529  + 5486.77784317500 *T)
    sumR0  +=     219.428 * cos(5.58340248784  + 191.44826611160 *T)
    sumR0  +=     269.891 * cos(3.76394728622  + 5884.92684658320 *T)
    sumR0  +=     208.333 * cos(5.25476080773  + 3340.54511639700 *T)
    sumR0  +=     275.224 * cos(2.90818883832  + 1748.01641306700 *T)
    sumR0  +=     275.501 * cos(1.21767967781  + 6254.62666252360 *T)
    sumR0  +=     239.133 * cos(2.03669896238  + 1194.44701022460 *T)
    sumR0  +=     223.190 * cos(4.19861593779  + 3149.16416058820 *T)
    sumR0  +=     182.686 * cos(5.08062683355  + 6684.74797174860 *T)
    sumR0  +=     186.213 * cos(5.69871555748  + 6677.70173505060 *T)
    sumR0  +=     175.995 * cos(5.95341786369  + 3870.30339179440 *T)
    sumR0  +=     178.613 * cos(4.18423025538  + 3333.49887969900 *T)
    sumR0  +=     208.336 * cos(4.84626442122  + 3340.67973700260 *T)
    sumR0  +=     228.128 * cos(3.25529020620  + 6872.67311951120 *T)
    sumR0  +=     144.286 * cos(0.21296012258  + 5088.62883976680 *T)
    sumR0  +=     163.534 * cos(3.79889068111  + 4136.91043351620 *T)
    sumR0  +=     133.120 * cos(1.53910106710  + 7903.07341972100 *T)
    sumR0  +=     141.759 * cos(2.47790321309  + 4562.46099302120 *T)
    sumR0  +=     114.941 * cos(4.31745088059  + 1349.86740965880 *T)
    sumR0  +=     118.781 * cos(2.12178071222  + 1589.07289528380 *T)
    sumR0  +=     102.096 * cos(6.18138550087  + 9492.14631500480 *T)
    sumR0  +=     128.555 * cos(5.49883294915  + 8827.39026987480 *T)
    sumR0  +=     111.538 * cos(0.55339169625  + 11243.68584642080 *T)
    sumR0  +=      82.498 * cos(1.62227044590  + 11773.37681151540 *T)
    sumR0  +=      83.212 * cos(0.61553380568  + 8429.24126646660 *T)
    sumR0  +=      84.470 * cos(0.62274593110  + 1592.59601363280 *T)
    sumR0  +=      86.659 * cos(1.74988330093  + 2700.71514038580 *T)
    sumR0  +=      71.826 * cos(2.47489899385  + 12303.06777661000 *T)
    sumR0  +=      85.312 * cos(1.61621097912  + 4690.47983635860 *T)
    sumR0  +=      63.641 * cos(2.67334126661  + 426.59819087600 *T)
    sumR0  +=      68.599 * cos(2.40197828418  + 4399.99435688900 *T)
    sumR0  +=      58.559 * cos(4.72052787516  + 213.29909543800 *T)
    sumR0  +=      62.015 * cos(1.10065866221  + 1221.84856632140 *T)
    sumR0  +=      66.509 * cos(2.21307705185  + 6041.32756708560 *T)
    sumR0  +=      55.811 * cos(1.23288325946  + 3185.19202726560 *T)
    sumR0  +=      54.989 * cos(5.72691385306  + 951.71840625060 *T)
    sumR0  +=      52.418 * cos(3.02366828926  + 4292.33083295040 *T)
    sumR0  +=      55.686 * cos(5.44686699242  + 3723.50895892300 *T)
    sumR0  +=      58.959 * cos(3.26242666052  + 6681.24210705180 *T)
    sumR0  +=      44.629 * cos(2.01473640390  + 8031.09226305840 *T)
    sumR0  +=      58.959 * cos(1.23165502899  + 6681.20759974740 *T)
    sumR0  +=      42.444 * cos(2.26551590902  + 155.42039943420 *T)
    sumR0  +=      38.956 * cos(2.57760416009  + 3341.59274776800 *T)
    sumR0  +=      51.561 * cos(5.72326937712  + 7079.37385680780 *T)
    sumR0  +=      48.939 * cos(5.61614696751  + 3553.91152213780 *T)
    sumR0  +=      45.414 * cos(5.43290921705  + 6467.92575796160 *T)
    sumR0  +=      36.435 * cos(4.43921812388  + 3894.18182954220 *T)
    sumR0  +=      35.980 * cos(1.15966567007  + 2288.34404351140 *T)
    sumR0  +=      35.265 * cos(5.49029710802  + 1990.74501704100 *T)
    sumR0  +=      42.191 * cos(1.63253742760  + 5628.95647021120 *T)
    sumR0  +=      44.292 * cos(5.00341366850  + 5614.72937620960 *T)
    sumR0  +=      33.623 * cos(5.17029029766  + 20043.67456019880 *T)
    sumR0  +=      43.256 * cos(1.03732072925  + 11769.85369316640 *T)
    sumR0  +=      39.237 * cos(1.24237122859  + 3339.63210563160 *T)
    sumR0  +=      31.943 * cos(4.59258406791  + 2274.11694950980 *T)
    sumR0  +=      30.345 * cos(2.44177670130  + 11371.70468975820 *T)
    sumR0  +=      32.259 * cos(2.38215172582  + 4535.05943692440 *T)
    sumR0  +=      31.870 * cos(4.37521442752  + 3.52311834900 *T)
    sumR0  +=      29.350 * cos(4.06034813442  + 3097.88382272579 *T)
    sumR0  +=      31.972 * cos(1.93970478412  + 382.89653222320 *T)
    sumR0  +=      26.166 * cos(5.58466944895  + 9623.68827669120 *T)
    sumR0  +=      27.904 * cos(4.25805969214  + 3191.04922956520 *T)
    sumR0  +=      33.065 * cos(0.85467740581  + 553.56940284240 *T)
    sumR0  +=      27.543 * cos(1.57668567401  + 9595.23908922340 *T)
    sumR0  +=      25.159 * cos(0.81355213242  + 10713.99488132620 *T)
    sumR0  +=      22.070 * cos(0.85747723964  + 3319.83703120740 *T)
    sumR0  +=      24.772 * cos(5.38970742761  + 2818.03500860600 *T)
    sumR0  +=      23.359 * cos(6.01453778225  + 3496.03282613400 *T)
    sumR0  +=      24.732 * cos(2.58034797703  + 2803.80791460440 *T)
    sumR0  +=      19.365 * cos(5.18528750472  + 6681.29216370240 *T)
    sumR0  +=      19.122 * cos(5.41968559451  + 10025.36039844840 *T)
    sumR0  +=      19.364 * cos(5.59378382138  + 6681.15754309680 *T)
    sumR0  +=      18.330 * cos(5.79567324240  + 7064.12138562280 *T)
    sumR0  +=      18.193 * cos(5.61307426173  + 7.11354700080 *T)
    sumR0  +=      20.392 * cos(4.53637816869  + 6489.77658728800 *T)
    sumR0  +=      21.260 * cos(6.19160142215  + 14054.60730802600 *T)
    sumR0  +=      17.094 * cos(1.55004739305  + 2957.71589447660 *T)
    sumR0  +=      22.791 * cos(3.41709388606  + 7632.94325965020 *T)
    sumR0  +=      20.585 * cos(2.98697279083  + 3361.38782219220 *T)
    sumR0  +=      18.005 * cos(2.81431094394  + 4032.77002792660 *T)
    sumR0  +=      17.049 * cos(6.15528099726  + 10404.73381232260 *T)
    sumR0  +=      16.488 * cos(3.84534700818  + 10973.55568635000 *T)
    sumR0  +=      16.052 * cos(0.92823508003  + 14584.29827312060 *T)
    sumR0  +=      21.027 * cos(2.38474290907  + 4989.05918389720 *T)
    sumR0  +=      16.267 * cos(1.92321585819  + 7373.38245462640 *T)
    sumR0  +=      16.291 * cos(6.28233085307  + 7210.91581849420 *T)
    sumR0  +=      18.585 * cos(4.07325116588  + 2388.89402044920 *T)
    sumR0  +=      15.977 * cos(4.58368417141  + 3264.34635542420 *T)
    sumR0  +=      19.913 * cos(2.73518445950  + 5099.26550511660 *T)
    sumR0  +=      19.661 * cos(1.86285979000  + 3443.70520091840 *T)
    sumR0  +=      16.500 * cos(4.14061745086  + 7477.52286021600 *T)
    sumR0  +=      19.495 * cos(6.03778234182  + 10018.31416175040 *T)
    sumR0  +=      15.104 * cos(2.65433427561  + 2787.04302385740 *T)
    sumR0  +=      19.099 * cos(0.22623441108  + 13745.34623902240 *T)
    sumR0  +=      17.163 * cos(3.18825562972  + 3347.72597370060 *T)
    sumR0  +=      13.423 * cos(2.12818658793  + 3344.20285535160 *T)
    sumR0  +=      15.410 * cos(2.20773507960  + 2118.76386037840 *T)
    sumR0  +=      17.238 * cos(3.67067776368  + 3205.54734666440 *T)
    sumR0  +=      13.113 * cos(4.27490214998  + 14314.16811304980 *T)
    sumR0  +=      16.451 * cos(2.86641622696  + 14712.31711645800 *T)
    sumR0  +=      13.734 * cos(1.68629769646  + 3337.02199804800 *T)
    sumR0  +=      16.659 * cos(4.52130808861  + 6674.11130639880 *T)
    sumR0  +=      11.830 * cos(0.19684525299  + 3475.67750673520 *T)
    sumR0  +=      11.767 * cos(3.22897247987  + 5828.02847164760 *T)
    sumR0  +=      11.886 * cos(4.82057654742  + 7234.79425624200 *T)
    sumR0  +=      10.609 * cos(1.73997337551  + 639.89728631400 *T)
    sumR0  +=      11.154 * cos(0.23859830185  + 12832.75874170460 *T)
    sumR0  +=      11.030 * cos(0.44541706440  + 10213.28554621100 *T)
    sumR0  +=      10.240 * cos(5.74758340632  + 242.72860397400 *T)
    sumR0  +=      10.051 * cos(2.45102946726  + 4929.68532198360 *T)
    sumR0  +=      10.061 * cos(0.78907665448  + 9381.93999378540 *T)
    sumR0  +=      10.065 * cos(5.37506605762  + 5085.03841111500 *T)
    sumR0  +=      11.902 * cos(0.79897698904  + 3265.83082813250 *T)
    sumR0  +=       8.990 * cos(0.96463418322  + 4933.20844033260 *T)
    sumR0  +=       8.970 * cos(4.18326774405  + 9225.53927328300 *T)
    sumR0  +=       8.982 * cos(1.98501418026  + 15113.98923821520 *T)
    sumR0  +=       8.324 * cos(1.93694866513  + 1648.44675719740 *T)
    sumR0  +=       7.833 * cos(2.04997447879  + 1758.65307841680 *T)
    sumR0  +=       7.957 * cos(3.92331072722  + 2921.12778282460 *T)
    sumR0  +=      10.224 * cos(2.66497189021  + 2487.41604494780 *T)
    sumR0  +=       8.277 * cos(0.94860765546  + 2906.90068882300 *T)
    sumR0  +=       7.371 * cos(0.84378341481  + 692.15760122680 *T)
    sumR0  +=       7.545 * cos(5.68031160782  + 13916.01910964160 *T)
    sumR0  +=       7.912 * cos(2.81294761885  + 15643.68020330980 *T)
    sumR0  +=       6.958 * cos(3.32193127272  + 3230.40610548040 *T)
    sumR0  +=       7.426 * cos(6.09656283295  + 3583.34103067380 *T)
    sumR0  +=       6.402 * cos(4.19806996774  + 5202.35827933520 *T)
    sumR0  +=       6.529 * cos(6.11929252710  + 135.06508003540 *T)
    sumR0  +=       6.130 * cos(0.00117252232  + 6836.64525283380 *T)
    sumR0  +=       6.226 * cos(6.10683955669  + 17256.63153634140 *T)
    sumR0  +=       8.193 * cos(5.24811458833  + 10575.40668294180 *T)
    sumR0  +=       6.168 * cos(3.60069207439  + 10021.85453375160 *T)
    sumR0  +=       6.169 * cos(1.56992114335  + 10021.82002644720 *T)
    sumR0  +=       5.671 * cos(0.13650306660  + 13524.91634293140 *T)
    sumR0  +=       6.257 * cos(4.50450768937  + 8425.65083781480 *T)
    sumR0  +=       5.249 * cos(2.70122358603  + 4459.36821880260 *T)
    sumR0  +=       6.479 * cos(2.74267498287  + 7740.60678358880 *T)
    sumR0  +=       5.530 * cos(6.06408145092  + 10419.98628350760 *T)
    sumR0  +=       5.536 * cos(5.74996063243  + 12168.00269657460 *T)
    sumR0  +=       6.826 * cos(4.69327545839  + 17654.78053974960 *T)
    sumR0  +=       5.005 * cos(4.68382632238  + 522.57741809380 *T)
    sumR0  +=       6.328 * cos(3.31896726895  + 3767.21061757580 *T)
    sumR0  +=       4.734 * cos(0.00755678425  + 3325.35995551480 *T)
    sumR0  +=       5.032 * cos(2.33696685608  + 1052.26838318840 *T)
    sumR0  +=       4.661 * cos(5.15051711401  + 1066.49547719000 *T)
    sumR0  +=       4.724 * cos(5.77956037309  + 9808.53818466140 *T)
    sumR0  +=       5.136 * cos(1.57259709878  + 6525.80445396540 *T)
    sumR0  +=       4.522 * cos(1.44218620352  + 3369.06161416760 *T)
    sumR0  +=       6.189 * cos(4.58996159356  + 6531.66165626500 *T)
    sumR0  +=       6.205 * cos(4.48164856516  + 22747.29071487440 *T)
    sumR0  +=       5.329 * cos(4.55145599182  + 1744.42598441520 *T)
    sumR0  +=       4.511 * cos(5.94511266539  + 6894.52394883760 *T)
    sumR0  +=       4.330 * cos(3.10901365758  + 4569.57454002200 *T)
    sumR0  +=       5.366 * cos(5.08043436437  + 2707.82868738660 *T)
    sumR0  +=       5.134 * cos(1.28568358496  + 8439.87793181640 *T)
    sumR0  +=       4.127 * cos(5.48538052912  + 2699.73481931760 *T)
    sumR0  +=       5.394 * cos(5.21695066244  + 5305.45105355380 *T)
    sumR0  +=       4.449 * cos(5.56764082611  + 16865.52876963120 *T)
    sumR0  +=       3.898 * cos(1.48782434790  + 9168.64089834740 *T)
    sumR0  +=       3.863 * cos(1.23050803930  + 16858.48253293320 *T)
    sumR0  +=       3.766 * cos(0.27090392616  + 17395.21973472580 *T)
    sumR0  +=       4.683 * cos(3.05668892586  + 5518.75014899180 *T)
    sumR0  +=       4.258 * cos(2.79090145790  + 3503.07906283200 *T)
    sumR0  +=       3.864 * cos(0.37957291785  + 10177.25767953360 *T)
    sumR0  +=       3.992 * cos(1.84424768030  + 3134.42687826260 *T)
    sumR0  +=       3.643 * cos(2.95318600206  + 6144.42034130420 *T)
    sumR0  +=       3.654 * cos(1.58063207414  + 6680.24453233140 *T)
    sumR0  +=       3.931 * cos(1.98436570971  + 8969.56889691100 *T)
    sumR0  +=       3.350 * cos(2.72637081667  + 7875.67186362420 *T)
    sumR0  +=       3.623 * cos(2.91544991631  + 6682.20517446780 *T)
    sumR0  +=       4.392 * cos(0.81931997702  + 3302.47939106200 *T)
    sumR0  +=       4.062 * cos(5.46935175827  + 3120.19978426100 *T)
    sumR0  +=       3.319 * cos(1.77193665786  + 3116.26763099790 *T)
    sumR0  +=       3.510 * cos(1.18027333874  + 10184.30391623160 *T)
    sumR0  +=       4.008 * cos(1.33675167812  + 6247.51311552280 *T)
    sumR0  +=       3.603 * cos(0.15469852619  + 2178.13772229200 *T)
    sumR0  +=       3.313 * cos(3.12853663982  + 17277.40693183380 *T)
    sumR0  +=       4.138 * cos(4.39568891039  + 3074.00538497800 *T)
    sumR0  +=       3.201 * cos(3.36492925275  + 2384.32327072920 *T)
    sumR0  +=       3.991 * cos(3.82867836910  + 3355.86489788480 *T)
    sumR0  +=       4.205 * cos(1.90532279170  + 263.08392337280 *T)
    sumR0  +=       3.743 * cos(4.25458024187  + 6261.74020952440 *T)
    sumR0  +=       3.111 * cos(1.65445053349  + 20199.09495963300 *T)
    sumR0  +=       3.634 * cos(5.55063049451  + 632.78373931320 *T)
    sumR0  +=       2.892 * cos(1.91514920700  + 12935.85151592320 *T)
    sumR0  +=       3.373 * cos(5.50812408522  + 23384.28698689860 *T)
    sumR0  +=       3.312 * cos(5.83174680245  + 5331.35744374080 *T)
    sumR0  +=       3.124 * cos(5.44069658195  + 6048.44111408640 *T)
    sumR0  +=       3.813 * cos(0.80274487430  + 13517.87010623340 *T)
    sumR0  +=       3.612 * cos(3.68140265686  + 5724.93569742900 *T)
    sumR0  +=       2.813 * cos(1.68598843422  + 2391.43681773000 *T)
    sumR0  +=       2.902 * cos(5.30666239739  + 8955.34180290940 *T)
    sumR0  +=       3.225 * cos(2.29832592489  + 3312.16323923200 *T)
    sumR0  +=       3.863 * cos(3.48188264725  + 20618.01935853360 *T)
    sumR0  +=       2.738 * cos(5.49768261369  + 149.56319713460 *T)
    sumR0  +=       2.809 * cos(4.76933217397  + 1964.83862685400 *T)
    sumR0  +=       2.711 * cos(2.69244730345  + 3178.14579056760 *T)
    sumR0  +=       2.711 * cos(2.38275660721  + 2648.45482547300 *T)
    sumR0  +=       2.743 * cos(1.09492569851  + 536.80451209540 *T)
    sumR0  +=       2.821 * cos(5.91845472246  + 12964.30070339100 *T)
    sumR0  +=       2.716 * cos(6.10377796874  + 3973.39616601300 *T)
    sumR0  +=       2.487 * cos(3.87703916286  + 1861.74585263540 *T)
    sumR0  +=       2.617 * cos(2.65526535627  + 8329.67161059700 *T)
    sumR0  +=       2.938 * cos(5.68402398615  + 6158.64743530580 *T)
    sumR0  +=       2.336 * cos(3.24847913311  + 4672.66731424060 *T)
    sumR0  +=       2.318 * cos(1.69214259672  + 3914.95722503460 *T)
    sumR0  +=       2.371 * cos(4.75067664712  + 103.09277421860 *T)
    sumR0  +=       2.963 * cos(0.23381699914  + 20597.24396304120 *T)
    sumR0  +=       2.190 * cos(6.18344448099  + 3346.13535100720 *T)
    sumR0  +=       2.444 * cos(1.92547995169  + 7799.98064550240 *T)
    sumR0  +=       2.121 * cos(4.87491216115  + 9830.38901398780 *T)
    sumR0  +=       2.532 * cos(5.39550087270  + 3863.18984479360 *T)
    sumR0  +=       2.101 * cos(2.84309138388  + 3415.39402526710 *T)
    sumR0  +=       2.176 * cos(0.58632570025  + 162.46663613220 *T)
    sumR0  +=       2.106 * cos(3.06229353931  + 19800.94595622480 *T)
    sumR0  +=       2.347 * cos(3.90795942709  + 3335.08950239240 *T)
    sumR0  +=       2.031 * cos(5.52057907797  + 10021.90459040220 *T)
    sumR0  +=       1.997 * cos(2.77243710569  + 13936.79450513400 *T)
    sumR0  +=       2.139 * cos(5.40620646615  + 266.60704172180 *T)
    sumR0  +=       2.147 * cos(0.08966987600  + 13358.92658845020 *T)
    sumR0  +=       1.996 * cos(2.62541669265  + 20.77539549240 *T)
    sumR0  +=       1.961 * cos(4.88521794174  + 3237.51965248120 *T)
    sumR0  +=       2.216 * cos(1.06829128652  + 3320.25710730100 *T)
    sumR0  +=       2.131 * cos(3.02112533027  + 5625.36604155940 *T)
    sumR0  +=       2.124 * cos(3.68620121537  + 5618.31980486140 *T)
    sumR0  +=       1.938 * cos(1.29006691721  + 17924.91069982040 *T)
    sumR0  +=       2.555 * cos(4.91826220321  + 6604.95878212400 *T)
    sumR0  +=       2.561 * cos(2.10055088914  + 7910.18696672180 *T)
    sumR0  +=       1.820 * cos(5.57528712663  + 3351.24909204960 *T)
    sumR0  +=       1.786 * cos(5.77310414452  + 3607.21946842160 *T)
    sumR0  +=       1.780 * cos(4.48010071981  + 10818.13528691580 *T)
    sumR0  +=       2.106 * cos(5.75526661975  + 13365.97282514820 *T)
    sumR0  +=       1.987 * cos(2.61151965233  + 3546.79797513700 *T)
    sumR0  +=       1.799 * cos(2.73192475257  + 3360.96774609859 *T)
    sumR0  +=       1.715 * cos(1.50805385053  + 1692.16566950240 *T)
    sumR0  +=       1.752 * cos(2.21455466761  + 13119.72110282519 *T)
    sumR0  +=       1.913 * cos(3.32230688971  + 6702.00024889200 *T)
    sumR0  +=       1.724 * cos(1.43449979531  + 4885.96640967860 *T)
    sumR0  +=       2.050 * cos(1.19293239093  + 6660.44945790720 *T)
    sumR0  +=       1.774 * cos(2.18404386388  + 6784.31762761820 *T)
    sumR0  +=       1.722 * cos(4.86031154305  + 10014.72373309860 *T)
    sumR0  +=       1.773 * cos(2.09448668554  + 3603.69635007260 *T)
    sumR0  +=       1.606 * cos(3.48105136801  + 23141.55838292460 *T)
    sumR0  +=       1.621 * cos(5.73820120882  + 4555.34744602040 *T)
    sumR0  +=       1.579 * cos(1.88769198841  + 6298.32832117640 *T)
    sumR0  +=       1.530 * cos(5.16381564230  + 76.26607127560 *T)
    sumR0  +=       1.615 * cos(3.24110713658  + 3657.00429635640 *T)
    sumR0  +=       1.576 * cos(3.52622401575  + 6688.33840040040 *T)
    sumR0  +=       2.034 * cos(2.63620520451  + 16460.33352952499 *T)
    sumR0  +=       2.025 * cos(5.92907541624  + 10021.76996979660 *T)
    sumR0  +=       1.689 * cos(4.41053057494  + 5729.50644714900 *T)
    sumR0  +=       1.878 * cos(4.53291044847  + 3329.97576135000 *T)
    sumR0  +=       1.530 * cos(4.76331644411  + 7895.95987272020 *T)
    sumR0  +=       1.529 * cos(1.35289110986  + 1581.95934828300 *T)
    sumR0  +=       1.807 * cos(1.86212004697  + 2693.60159338500 *T)
    sumR0  +=       1.855 * cos(2.38561742394  + 6843.69148953180 *T)
    sumR0  +=       1.518 * cos(3.98476157750  + 6546.15977336420 *T)
    sumR0  +=       1.389 * cos(1.82099537095  + 9779.10867612540 *T)
    sumR0  +=       1.447 * cos(2.35649936427  + 6034.21402008480 *T)
    sumR0  +=       1.386 * cos(5.55304113895  + 4775.76008845920 *T)
    sumR0  +=       1.372 * cos(1.07224580315  + 12722.55242048520 *T)
    sumR0  +=       1.423 * cos(4.46530428193  + 574.34479833480 *T)
    sumR0  +=       1.424 * cos(2.57162391016  + 3399.98628861340 *T)
    sumR0  +=       1.380 * cos(5.76156315252  + 16335.83780453660 *T)
    sumR0  +=       1.338 * cos(2.97604558638  + 6127.65545055720 *T)
    sumR0  +=       1.479 * cos(4.74310691166  + 12566.15169998280 *T)
    sumR0  +=       1.706 * cos(0.30579918494  + 10551.52824519400 *T)
    sumR0  +=       1.281 * cos(2.00285974432  + 6677.63442474780 *T)
    sumR0  +=       1.350 * cos(0.78892333409  + 853.19638175200 *T)
    sumR0  +=       1.534 * cos(4.33326399444  + 640.87760738220 *T)
    sumR0  +=       1.247 * cos(1.02503908468  + 3024.22055704320 *T)
    sumR0  +=       1.289 * cos(1.92786975543  + 3347.65866339780 *T)
    sumR0  +=       1.243 * cos(2.44217806237  + 6684.81528205140 *T)
    sumR0  +=       1.453 * cos(1.74218016403  + 3333.56619000180 *T)
    sumR0  +=       1.675 * cos(1.79693456330  + 1118.75579210280 *T)
    sumR0  +=       1.491 * cos(2.59386711806  + 2494.52959194860 *T)
    sumR0  +=       1.293 * cos(3.31710472549  + 3407.09983561420 *T)
    sumR0  +=       1.188 * cos(4.92989260576  + 22743.40937951640 *T)
    sumR0  +=       1.329 * cos(1.99426530402  + 1228.96211332220 *T)
    sumR0  +=       1.373 * cos(2.53354987340  + 5459.37628707820 *T)
    sumR0  +=       1.183 * cos(4.25338096667  + 3344.49376205780 *T)
    sumR0  +=       1.231 * cos(2.50206227837  + 4356.27544458400 *T)
    sumR0  +=       1.243 * cos(2.65176267860  + 74.78159856730 *T)
    sumR0  +=       1.285 * cos(4.34087881585  + 3326.38533269820 *T)
    sumR0  +=       1.119 * cos(1.91321862491  + 3281.23856478620 *T)
    sumR0  +=       1.094 * cos(5.50748655535  + 3017.10701004240 *T)
    sumR0  +=       1.259 * cos(3.77654662830  + 11236.57229942000 *T)
    sumR0  +=       1.285 * cos(1.38335267684  + 3077.52850332700 *T)
    sumR0  +=       1.100 * cos(1.17130732373  + 6606.44325483230 *T)
    sumR0  +=       1.115 * cos(5.81275569652  + 2675.85638156980 *T)
    sumR0  +=       1.380 * cos(5.70641426169  + 2807.39834325620 *T)
    sumR0  +=       1.256 * cos(3.35479933251  + 4039.88357492740 *T)
    sumR0  +=       1.187 * cos(2.41348693872  + 10596.18207843420 *T)
    sumR0  +=       1.052 * cos(3.33521939538  + 3304.58456002240 *T)
    sumR0  +=       1.188 * cos(5.84735836632  + 3336.73109134180 *T)
    sumR0  +=       1.072 * cos(2.78383184435  + 8270.29774868340 *T)
    sumR0  +=       1.105 * cos(3.03463252672  + 3929.67725370800 *T)
    sumR0  +=       1.013 * cos(3.52026711847  + 8013.27974094040 *T)
    sumR0  +=       1.079 * cos(0.51857999039  + 2814.44457995420 *T)
    sumR0  +=       0.999 * cos(4.72734008760  + 533.21408344360 *T)
    sumR0  +=       1.131 * cos(0.52584038526  + 6816.28993343500 *T)
    sumR0  +=       1.191 * cos(0.60874292520  + 2301.58581590939 *T)
    sumR0  +=       1.313 * cos(2.07273299121  + 23539.70738633280 *T)
    sumR0  +=       0.996 * cos(4.03971126547  + 16062.18452611680 *T)
    sumR0  +=       0.954 * cos(5.90340414098  + 20206.14119633100 *T)
    sumR0  +=       0.993 * cos(0.07132588892  + 24150.08005134500 *T)
    sumR0  +=       1.051 * cos(2.22096534870  + 3980.50971301380 *T)
    sumR0  +=       1.089 * cos(1.25512213569  + 5938.23479286700 *T)
    sumR0  +=       0.912 * cos(2.54221161167  + 433.71173787680 *T)
    sumR0  +=       1.249 * cos(0.60003625956  + 16173.37116840440 *T)
    sumR0  +=       1.027 * cos(4.95999945094  + 19676.45023123640 *T)
    sumR0  +=       1.108 * cos(4.34209448160  + 3339.12795399150 *T)
    sumR0  +=       1.188 * cos(6.21563747433  + 2679.37949991880 *T)
    sumR0  +=       0.849 * cos(0.82548606454  + 2597.62236616720 *T)
    sumR0  +=       1.145 * cos(4.48151980872  + 19402.79695281660 *T)
    sumR0  +=       0.948 * cos(1.30280088857  + 8273.82086703240 *T)
    sumR0  +=       1.016 * cos(5.14464815830  + 1596.18644228460 *T)
    sumR0  +=       0.832 * cos(5.60623652030  + 3340.19235060619 *T)
    sumR0  +=       1.035 * cos(4.71893106874  + 419.48464387520 *T)
    sumR0  +=       0.903 * cos(0.45419000582  + 12995.22537783680 *T)
    sumR0  +=       1.089 * cos(0.51294377637  + 11250.79939342160 *T)
    sumR0  +=       0.840 * cos(5.30858028008  + 26084.02180621620 *T)
    sumR0  +=       0.990 * cos(2.06776368865  + 7255.56965173440 *T)
    sumR0  +=       0.808 * cos(6.25630819993  + 15508.61512327440 *T)
    sumR0  +=       0.806 * cos(3.09007612135  + 5415.65737477320 *T)
    sumR0  +=       0.782 * cos(4.62274599734  + 2547.83753823240 *T)
    sumR0  +=       0.963 * cos(2.10680539916  + 6456.88005769770 *T)
    sumR0  +=       0.778 * cos(3.56602161857  + 12721.57209941700 *T)
    sumR0  +=       0.873 * cos(5.09097164784  + 2540.79130153440 *T)
    sumR0  +=       0.772 * cos(3.08101797047  + 11081.21921028860 *T)
    sumR0  +=       0.965 * cos(2.33106703115  + 18454.60166491500 *T)
    sumR0  +=       0.859 * cos(4.14788214122  + 6438.49624942560 *T)
    sumR0  +=       1.012 * cos(4.45011664540  + 3316.73398895200 *T)
    sumR0  +=       0.906 * cos(4.29336078401  + 3344.54457996290 *T)
    sumR0  +=       0.741 * cos(2.61446403580  + 2284.75361485960 *T)
    sumR0  +=       0.790 * cos(6.03436225041  + 12509.25332504720 *T)
    sumR0  +=       0.738 * cos(0.52092422137  + 18052.92954315780 *T)
    sumR0  +=       0.737 * cos(4.11165247543  + 3760.09707057500 *T)
    sumR0  +=       0.727 * cos(3.28066632751  + 3510.19260983280 *T)
    sumR0  +=       1.006 * cos(0.45037465289  + 27490.69247804480 *T)
    sumR0  +=       0.923 * cos(2.78717931388  + 1332.05488754080 *T)
    sumR0  +=       0.756 * cos(0.86881841787  + 1545.35398297880 *T)
    sumR0  +=       0.774 * cos(3.71535541900  + 6571.01853218020 *T)
    sumR0  +=       0.900 * cos(2.74944190055  + 316.39186965660 *T)
    sumR0  +=       0.704 * cos(1.89617185328  + 13362.43245314700 *T)
    sumR0  +=       0.701 * cos(2.21328293796  + 20995.39296644940 *T)
    sumR0  +=       0.701 * cos(3.92689438700  + 13362.46696045140 *T)
    sumR0  +=       0.695 * cos(5.52658147215  + 3364.49086444760 *T)
    sumR0  +=       0.764 * cos(1.88253040972  + 5732.04924442980 *T)
    sumR0  +=       0.693 * cos(0.34849213821  + 5835.14201864840 *T)
    sumR0  +=       0.698 * cos(1.79132650081  + 206.18554843720 *T)
    sumR0  +=       0.687 * cos(0.11649928911  + 13760.59871020740 *T)
    sumR0  +=       0.827 * cos(0.49766945172  + 3376.64029337720 *T)
    sumR0  +=       0.686 * cos(0.23965591265  + 2409.24933984800 *T)
    sumR0  +=       0.672 * cos(0.51352450554  + 25685.87280280800 *T)
    sumR0  +=       0.699 * cos(5.90167595080  + 19004.64794940840 *T)
    sumR0  +=       0.647 * cos(3.01091875955  + 5223.69391980220 *T)
    sumR0  +=       0.643 * cos(4.88507402785  + 11766.26326451460 *T)
    sumR0  +=       0.723 * cos(3.28296530537  + 4142.97634914600 *T)
    sumR0  +=       0.660 * cos(4.27334393830  + 1214.73501932060 *T)
    sumR0  +=       0.675 * cos(0.48874492682  + 5408.54382777240 *T)
    sumR0  +=       0.809 * cos(3.13310075522  + 18984.29263000960 *T)
    sumR0  +=       0.646 * cos(3.63004914186  + 7107.82304427560 *T)
    sumR0  +=       0.831 * cos(4.49449658957  + 3341.03250279340 *T)
    sumR0  +=       0.680 * cos(0.16949964513  + 110.20632121940 *T)
    sumR0  +=       0.592 * cos(6.03272224596  + 20809.46762464520 *T)
    sumR0  +=       0.627 * cos(4.93438097728  + 9872.27408296480 *T)
    sumR0  +=       0.698 * cos(1.45709305452  + 3377.21779200400 *T)
    sumR0  +=       0.583 * cos(4.01887095237  + 664.75604513000 *T)
    sumR0  +=       0.591 * cos(4.01381288194  + 19406.67828817460 *T)
    sumR0  +=       0.571 * cos(1.20823284619  + 5621.86017686260 *T)
    sumR0  +=       0.680 * cos(4.72905586557  + 8116.37251515900 *T)
    sumR0  +=       0.739 * cos(3.19765996917  + 9380.95967271720 *T)
    sumR0  +=       0.571 * cos(5.46064659057  + 5621.82566955820 *T)
    sumR0  +=       0.702 * cos(0.14107083764  + 6923.95345737360 *T)
    sumR0  +=       0.598 * cos(3.05986754375  + 3341.04230982650 *T)
    sumR0  +=       0.601 * cos(2.17944973698  + 7270.28968040780 *T)
    sumR0  +=       0.572 * cos(0.42001145821  + 15110.46611986620 *T)
    sumR0  +=       0.655 * cos(1.08706089359  + 3169.93955608060 *T)
    sumR0  +=       0.582 * cos(1.68224967199  + 26087.90314157420 *T)
    sumR0  +=       0.543 * cos(2.42716241058  + 170.67287061920 *T)
    sumR0  +=       0.543 * cos(1.68460678113  + 11780.49035851620 *T)
    sumR0  +=       0.533 * cos(2.63982684952  + 2277.70737816160 *T)
    sumR0  +=       0.550 * cos(2.54220679987  + 7380.49600162720 *T)
    sumR0  +=       0.576 * cos(1.69042028754  + 3384.33133900480 *T)
    sumR0  +=       0.535 * cos(5.01279223460  + 1375.77379984580 *T)
    sumR0  +=       0.579 * cos(5.99182453047  + 2149.68853482420 *T)
    sumR0  +=       0.598 * cos(0.75754342787  + 3340.18254357310 *T)
    sumR0  +=       0.575 * cos(1.33793171924  + 3296.89351439480 *T)
    sumR0  +=       0.689 * cos(0.02993986536  + 22345.26037610820 *T)
    sumR0  +=       0.526 * cos(6.14212516186  + 13149.15061136120 *T)
    sumR0  +=       0.540 * cos(5.54137118955  + 8646.06348025360 *T)
    sumR0  +=       0.528 * cos(2.74183495775  + 3826.58447948940 *T)
    sumR0  +=       0.689 * cos(6.28205551764  + 12295.95422960920 *T)
    sumR0  +=       0.499 * cos(0.25665224668  + 2142.64229812620 *T)
    sumR0  +=       0.498 * cos(5.91370528026  + 6460.81221096080 *T)
    sumR0  +=       0.588 * cos(3.39977690249  + 3329.56672643590 *T)
    sumR0  +=       0.507 * cos(0.63367870048  + 20735.83216142559 *T)
    sumR0  +=       0.590 * cos(3.87951788923  + 9065.54812412880 *T)
    sumR0  +=       0.509 * cos(2.59333214746  + 3877.41693879520 *T)
    sumR0  +=       0.486 * cos(4.01113864377  + 23937.85638974100 *T)
    sumR0  +=       0.541 * cos(1.55072392536  + 2221.85663459700 *T)
    sumR0  +=       0.482 * cos(0.00200643783  + 10235.13637553740 *T)
    sumR0  +=       0.474 * cos(4.14503805378  + 3113.13629391080 *T)
    sumR0  +=       0.523 * cos(2.89171162726  + 6040.34724601740 *T)
    sumR0  +=       0.458 * cos(4.08131432844  + 1162.47470440780 *T)
    sumR0  +=       0.493 * cos(2.98933582802  + 1039.02661079040 *T)
    sumR0  +=       0.489 * cos(4.88201647585  + 5511.63660199100 *T)
    sumR0  +=       0.537 * cos(1.24870344441  + 48835.19385644859 *T)
    sumR0  +=       0.585 * cos(5.53384919340  + 4981.94563689640 *T)
    sumR0  +=       0.462 * cos(5.38999453893  + 15664.03552270859 *T)
    sumR0  +=       0.531 * cos(1.22287288928  + 10721.10842832700 *T)
    sumR0  +=       0.440 * cos(4.69966067585  + 799.82112516540 *T)
    sumR0  +=       0.588 * cos(5.15076051189  + 3873.82651014340 *T)
    sumR0  +=       0.465 * cos(5.49236978535  + 3342.09689940810 *T)
    sumR0  +=       0.545 * cos(5.91813092143  + 1478.86657406440 *T)
    sumR0  +=       0.440 * cos(2.28287871244  + 12310.18132361080 *T)
    sumR0  +=       0.436 * cos(0.12478564862  + 956.28915597060 *T)
    sumR0  +=       0.573 * cos(3.86388572602  + 26482.17080962440 *T)
    sumR0  +=       0.429 * cos(1.20474317285  + 28.44918746780 *T)
    sumR0  +=       0.435 * cos(3.75437730899  + 5195.24473233440 *T)
    sumR0  +=       0.457 * cos(1.45606713663  + 5636.07001721200 *T)
    sumR0  +=       0.553 * cos(5.30628175145  + 2766.26762836500 *T)
    sumR0  +=       0.494 * cos(0.63303483918  + 3274.12501778540 *T)
    sumR0  +=       0.530 * cos(3.18458612574  + 8535.85715903420 *T)
    sumR0  +=       0.535 * cos(1.89165024367  + 9866.41688066520 *T)
    sumR0  +=       0.433 * cos(0.33270291162  + 6665.97238221460 *T)
    sumR0  +=       0.404 * cos(4.67939964590  + 4825.54491639400 *T)
    sumR0  +=       0.402 * cos(3.25727658426  + 5820.91492464680 *T)
    sumR0  +=       0.415 * cos(1.41915166067  + 21947.11137270000 *T)
    sumR0  +=       0.414 * cos(3.32843373219  + 8859.36257569160 *T)
    sumR0  +=       0.403 * cos(1.68400791606  + 9588.12554222260 *T)
    sumR0  +=       0.399 * cos(0.93465630912  + 2295.45759051220 *T)
    sumR0  +=       0.402 * cos(0.85364636108  + 29026.48522950779 *T)
    sumR0  +=       0.494 * cos(4.57792960199  + 8542.97070603500 *T)
    sumR0  +=       0.413 * cos(2.60545132222  + 38.13303563780 *T)
    sumR0  +=       0.413 * cos(5.44081623497  + 4407.10790388980 *T)
    sumR0  +=       0.404 * cos(5.33168348188  + 3189.56475685690 *T)
    sumR0  +=       0.385 * cos(1.90313951519  + 10020.85695903120 *T)
    sumR0  +=       0.454 * cos(1.18358809702  + 10610.90210710760 *T)
    sumR0  +=       0.381 * cos(4.90212003667  + 19146.75926614180 *T)
    sumR0  +=       0.415 * cos(3.07892141478  + 11216.28429032400 *T)
    sumR0  +=       0.378 * cos(0.19461412344  + 2067.93140107260 *T)
    sumR0  +=       0.418 * cos(1.14626975196  + 6643.09181776180 *T)
    sumR0  +=       0.426 * cos(4.33863084563  + 220.41264243880 *T)
    sumR0  +=       0.412 * cos(0.02314567339  + 3192.53370227350 *T)
    sumR0  +=       0.391 * cos(4.75902617217  + 6414.61781167780 *T)
    sumR0  +=       0.399 * cos(1.37708573980  + 9654.61295113700 *T)
    sumR0  +=       0.381 * cos(2.74120772781  + 1854.63230563460 *T)
    sumR0  +=       0.378 * cos(4.16366866018  + 13553.89797291080 *T)
    sumR0  +=       0.409 * cos(3.12117409635  + 3723.49170527080 *T)
    sumR0  +=       0.379 * cos(5.87578711682  + 6357.71943674220 *T)
    sumR0  +=       0.375 * cos(3.29175394061  + 31022.75317085620 *T)
    sumR0  +=       0.358 * cos(3.99134138048  + 8958.93223156120 *T)
    sumR0  +=       0.390 * cos(2.01175211992  + 1903.43681250120 *T)
    sumR0  +=       0.373 * cos(1.27018251531  + 2111.65031337760 *T)
    sumR0  +=       0.493 * cos(4.66294641806  + 1435.14766175940 *T)
    sumR0  +=       0.379 * cos(4.08720452061  + 10706.88133432540 *T)
    sumR0  +=       0.461 * cos(1.90122301470  + 25287.72379939980 *T)
    sumR0  +=       0.364 * cos(3.25231136840  + 5095.67507646480 *T)
    sumR0  +=       0.382 * cos(2.18684874688  + 16276.46394262300 *T)
    sumR0  +=       0.359 * cos(2.29096599931  + 28628.33622609960 *T)
    sumR0  +=       0.362 * cos(4.15084528423  + 6696.47732458460 *T)
    sumR0  +=       0.374 * cos(5.78780163439  + 3490.17562383440 *T)
    sumR0  +=       0.386 * cos(4.73230339383  + 3981.49003408200 *T)
    sumR0  +=       0.344 * cos(5.79156714790  + 26880.31981303260 *T)
    sumR0  +=       0.368 * cos(3.44405073824  + 15.25247118500 *T)
    sumR0  +=       0.410 * cos(1.76294940448  + 6709.67404086740 *T)
    sumR0  +=       0.339 * cos(1.71242500987  + 224.34479570190 *T)
    sumR0  +=       0.373 * cos(2.71286157688  + 3472.15438838620 *T)
    sumR0  +=       0.334 * cos(1.10569460106  + 59.37386191360 *T)
    sumR0  +=       0.394 * cos(4.02046774116  + 2075.04494807340 *T)
    sumR0  +=       0.402 * cos(3.27520695853  + 931.36308685180 *T)
    sumR0  +=       0.318 * cos(3.28892524780  + 9485.03276800400 *T)
    sumR0  +=       0.319 * cos(4.05344391333  + 3450.81874791920 *T)
    sumR0  +=       0.341 * cos(6.08567978892  + 17085.95866572220 *T)
    sumR0  +=       0.378 * cos(0.71609301686  + 3209.07046501340 *T)
    sumR0  +=       0.326 * cos(4.58257191106  + 8982.81066930900 *T)
    sumR0  +=       0.369 * cos(6.17228742886  + 8671.96987044060 *T)
    sumR0  +=       0.308 * cos(1.01924467785  + 18606.49894600020 *T)
    sumR0  +=       0.370 * cos(2.84011243059  + 4392.88080988820 *T)
    sumR0  +=       0.303 * cos(5.00239879717  + 2810.93871525740 *T)
    sumR0  +=       0.303 * cos(2.97162741204  + 2810.90420795300 *T)
    sumR0  +=       0.419 * cos(2.79830562400  + 5106.37905211740 *T)
    sumR0  +=       0.358 * cos(1.15955325186  + 4076.48894023160 *T)
    sumR0  +=       0.360 * cos(5.83226123310  + 26724.89941359840 *T)
    sumR0  +=       0.298 * cos(5.57160706753  + 3041.48603243060 *T)
    sumR0  +=       0.298 * cos(3.06970285022  + 11670.28403729680 *T)
    sumR0  +=       0.388 * cos(5.01197738185  + 12323.42309600880 *T)
    sumR0  +=       0.336 * cos(4.60208332758  + 9602.35263622420 *T)
    sumR0  +=       0.371 * cos(2.23508025241  + 4996.17273089800 *T)
    sumR0  +=       0.300 * cos(6.04069385215  + 9499.25986200560 *T)
    sumR0  +=       0.345 * cos(1.74260482164  + 5617.91076994730 *T)
    sumR0  +=       0.302 * cos(3.01127349940  + 7483.58877584580 *T)
    sumR0  +=       0.289 * cos(0.39479288685  + 2412.77245819700 *T)
    sumR0  +=       0.289 * cos(2.21430640344  + 11140.59307220220 *T)
    sumR0  +=       0.330 * cos(5.81605457596  + 4246.06912336460 *T)
    sumR0  +=       0.394 * cos(2.12229107240  + 6475.03930496240 *T)
    sumR0  +=       0.301 * cos(5.26147877814  + 9945.57120882380 *T)
    sumR0  +=       0.338 * cos(4.94717046909  + 5625.77507647350 *T)
    sumR0  +=       0.372 * cos(5.46968662800  + 3561.02506913860 *T)
    sumR0  +=       0.279 * cos(0.54063870001  + 3226.21331978640 *T)
    sumR0  +=       0.291 * cos(5.21021494024  + 13171.00144068760 *T)
    sumR0  +=       0.384 * cos(3.23921380878  + 10022.81760116760 *T)
    sumR0  +=       0.309 * cos(3.17514941939  + 14047.49376102520 *T)
    sumR0  +=       0.273 * cos(1.68203034215  + 4253.18267036540 *T)
    sumR0  +=       0.272 * cos(0.11218647217  + 7314.00859271280 *T)
    sumR0  +=       0.281 * cos(5.15132055967  + 2825.14855560680 *T)
    sumR0  +=       0.292 * cos(3.33720586058  + 9468.26787725700 *T)
    sumR0  +=       0.316 * cos(1.41719074976  + 589.06482700820 *T)
    sumR0  +=       0.264 * cos(0.48845594730  + 16699.53901514999 *T)
    sumR0  +=       0.266 * cos(1.69694779915  + 647.01083331480 *T)
    sumR0  +=       0.318 * cos(2.47072726153  + 8436.28750316460 *T)
    sumR0  +=       0.260 * cos(2.54459932529  + 20.35531939880 *T)
    sumR0  +=       0.275 * cos(2.78608579994  + 2970.91261075940 *T)
    sumR0  +=       0.295 * cos(5.96367554548  + 4025.65648092580 *T)
    sumR0  +=       0.268 * cos(3.01034973031  + 6518.75821726740 *T)
    sumR0  +=       0.259 * cos(5.20888482695  + 7366.26890762560 *T)
    sumR0  +=       0.298 * cos(2.71010678192  + 6652.77566593180 *T)
    sumR0  +=       0.276 * cos(0.78545108300  + 3735.23831175900 *T)
    sumR0  +=       0.305 * cos(6.19137255377  + 6677.34351804160 *T)
    sumR0  +=       0.265 * cos(3.25145629239  + 24889.57479599160 *T)
    sumR0  +=       0.260 * cos(3.99612605351  + 2171.02417529120 *T)
    sumR0  +=       0.252 * cos(4.14773813625  + 5642.19824260920 *T)
    sumR0  +=       0.254 * cos(1.38470256851  + 846.08283475120 *T)
    sumR0  +=       0.258 * cos(2.03261985834  + 2089.78223039900 *T)
    sumR0  +=       0.298 * cos(3.81212222628  + 28230.18722269139 *T)
    sumR0  +=       0.241 * cos(2.96550398155  + 27682.14074415640 *T)
    sumR0  +=       0.259 * cos(4.79545870271  + 6657.34641565180 *T)
    sumR0  +=       0.238 * cos(1.18977479528  + 3171.03224356680 *T)
    sumR0  +=       0.256 * cos(1.01427800277  + 568.82187402740 *T)
    sumR0  +=       0.236 * cos(5.56425829084  + 14.22709400160 *T)
    sumR0  +=       0.304 * cos(3.81556245925  + 1190.92389187560 *T)
    sumR0  +=       0.237 * cos(1.37222961867  + 2277.29834324750 *T)
    sumR0  +=       0.239 * cos(2.47752611026  + 5430.39465709880 *T)
    sumR0  +=       0.234 * cos(4.34929504798  + 6675.70192909220 *T)
    sumR0  +=       0.239 * cos(0.14012746335  + 3742.28454845700 *T)
    sumR0  +=       0.286 * cos(5.04045301355  + 5607.61582920880 *T)
    sumR0  +=       0.305 * cos(4.59739079660  + 6685.10618875760 *T)
    sumR0  +=       0.254 * cos(5.03693878366  + 1905.46476494040 *T)
    sumR0  +=       0.323 * cos(1.67390215145  + 4922.57177498280 *T)
    sumR0  +=       0.232 * cos(4.82565548677  + 9070.11887384880 *T)
    sumR0  +=       0.236 * cos(2.40662610715  + 3620.39893105220 *T)
    sumR0  +=       0.260 * cos(5.72282468723  + 17468.85519794540 *T)
    sumR0  +=       0.259 * cos(6.15179402778  + 16706.58525184800 *T)
    sumR0  +=       0.263 * cos(0.63922292958  + 2008.55753915900 *T)
    sumR0  +=       0.300 * cos(3.78527265088  + 34363.36559755600 *T)
    sumR0  +=       0.226 * cos(1.86970344963  + 6418.14093002680 *T)
    sumR0  +=       0.239 * cos(0.04616997400  + 13362.38239649640 *T)
    sumR0  +=       0.241 * cos(4.85896907298  + 14158.74771361560 *T)
    sumR0  +=       0.225 * cos(1.70179250908  + 18451.07854656599 *T)
    sumR0  +=       0.288 * cos(2.26316945288  + 6621.85099148600 *T)
    sumR0  +=       0.231 * cos(2.19861265305  + 3936.79080070880 *T)
    sumR0  +=       0.251 * cos(5.51232121883  + 3416.87849797540 *T)
    sumR0  +=       0.245 * cos(3.30613942274  + 1197.97012857360 *T)
    sumR0  +=       0.253 * cos(4.54308131689  + 2285.16264977370 *T)
    sumR0  +=       0.225 * cos(5.50822507089  + 4936.79886898440 *T)
    sumR0  +=       0.249 * cos(1.06089727346  + 3313.21087060300 *T)
    sumR0  +=       0.309 * cos(6.21936675838  + 16304.91313009080 *T)
    sumR0  +=       0.244 * cos(1.94855224181  + 3.59042865180 *T)
    sumR0  +=       0.287 * cos(5.70461951656  + 792.77488846740 *T)
    sumR0  +=       0.254 * cos(5.34446995416  + 5401.43028077160 *T)
    sumR0  +=       0.263 * cos(1.49663212332  + 6364.83298374300 *T)
    sumR0  +=       0.223 * cos(2.66825139116  + 31968.94865279940 *T)
    sumR0  +=       0.222 * cos(2.48370132727  + 5355.23588148860 *T)
    sumR0  +=       0.220 * cos(5.20799024654  + 23017.06265793620 *T)
    sumR0  +=       0.215 * cos(2.72743996418  + 6740.59871531320 *T)
    sumR0  +=       0.218 * cos(1.30797197521  + 29822.78323632420 *T)
    sumR0  +=       0.270 * cos(0.90714939427  + 6155.05700665400 *T)
    sumR0  +=       0.216 * cos(4.73975263349  + 6679.74038069130 *T)
    sumR0  +=       0.211 * cos(3.72756562629  + 10042.61267559180 *T)
    sumR0  +=       0.211 * cos(2.61999755641  + 10124.93005431800 *T)
    sumR0  +=       0.293 * cos(6.07059383381  + 14061.72085502680 *T)
    sumR0  +=       0.219 * cos(2.98472846458  + 131.54196168640 *T)
    sumR0  +=       0.210 * cos(5.27496906319  + 13355.33615979840 *T)
    sumR0  +=       0.259 * cos(1.25267305830  + 2641.34127847220 *T)
    sumR0  +=       0.208 * cos(3.30241021109  + 6850.80503653260 *T)
    sumR0  +=       0.226 * cos(5.48438086246  + 7203.80227149340 *T)
    sumR0  +=       0.243 * cos(2.44748800604  + 3311.18291816379 *T)
    sumR0  +=       0.208 * cos(1.13500579457  + 5888.44996493220 *T)
    sumR0  +=       0.251 * cos(4.67012983729  + 6666.99775939800 *T)
    sumR0  +=       0.227 * cos(1.59926413307  + 10001.06188460700 *T)
    sumR0  +=       0.264 * cos(3.72622435628  + 6747.71226231400 *T)
    sumR0  +=       0.216 * cos(0.34122804918  + 6686.74777770700 *T)
    sumR0  +=       0.260 * cos(3.67749190896  + 6645.19698672220 *T)
    sumR0  +=       0.209 * cos(4.31928920378  + 3337.86091608880 *T)
    sumR0  +=       0.218 * cos(4.08068730999  + 3378.74546233760 *T)
    sumR0  +=       0.275 * cos(1.64274205426  + 2011.10033643980 *T)
    sumR0  +=       0.204 * cos(0.73237459784  + 3.93215326310 *T)
    sumR0  +=       0.219 * cos(0.88584017263  + 19513.98359510420 *T)
    sumR0  +=       0.205 * cos(2.60851826933  + 2771.79055267240 *T)
    sumR0  +=       0.215 * cos(2.99299817517  + 10824.20120254560 *T)
    sumR0  +=       0.249 * cos(0.99914444136  + 5753.38488489680 *T)
    sumR0  +=       0.225 * cos(0.23309143434  + 4782.87363546000 *T)
    sumR0  +=       0.275 * cos(0.86215660461  + 8749.15625447220 *T)
    sumR0  +=       0.201 * cos(2.87118854159  + 21548.96236929180 *T)
    sumR0  +=       0.246 * cos(3.34468800742  + 3333.92876282570 *T)
    sumR0  +=       0.203 * cos(4.11410443276  + 31570.79964939120 *T)
    sumR0  +=       0.202 * cos(4.96805650734  + 8166.15734309380 *T)
    sumR0  +=       0.232 * cos(2.69372584349  + 5989.06725217280 *T)
    sumR0  +=       0.214 * cos(4.83852070026  + 6681.64492949320 *T)
    sumR0  +=       0.258 * cos(2.66551831451  + 1062.90504853820 *T)
    sumR0  +=       0.197 * cos(0.55202541352  + 735.87651353180 *T)
    sumR0  +=       0.256 * cos(2.76274941586  + 2480.30249794700 *T)
    sumR0  +=       0.216 * cos(4.02506717011  + 3133.91168940320 *T)
    sumR0  +=       0.193 * cos(1.52645682146  + 949.17560896980 *T)
    sumR0  +=       0.209 * cos(0.67348618655  + 13892.14067189380 *T)
    sumR0  +=       0.220 * cos(1.52502617699  + 6660.86953400080 *T)
    sumR0  +=       0.223 * cos(1.09348882524  + 6148.01076995600 *T)
    sumR0  +=       0.192 * cos(2.90571322264  + 8799.98871377800 *T)
    sumR0  +=       0.243 * cos(3.36420301442  + 8965.97846825920 *T)
    sumR0  +=       0.206 * cos(5.11175800472  + 4140.43355186520 *T)
    sumR0  +=       0.189 * cos(1.06641624209  + 9374.82644678460 *T)
    sumR0  +=       0.240 * cos(5.92078519032  + 13362.51701710200 *T)
    sumR0  +=       0.248 * cos(5.79997873732  + 15806.14683944200 *T)
    sumR0  +=       0.193 * cos(3.19008521814  + 6756.00645196690 *T)
    sumR0  +=       0.237 * cos(4.11979030463  + 22487.37169284160 *T)
    sumR0  +=       0.218 * cos(0.84212090761  + 6717.25272007700 *T)
    sumR0  +=       0.200 * cos(2.43100846630  + 10018.24685144760 *T)
    sumR0  +=       0.199 * cos(5.81238461796  + 4289.78803566960 *T)
    sumR0  +=       0.214 * cos(5.95026024979  + 6680.80477730600 *T)
    sumR0  +=       0.192 * cos(3.06285109033  + 32765.24665961580 *T)
    sumR0  +=       0.229 * cos(1.68884404323  + 11614.43329373220 *T)
    sumR0  +=       0.185 * cos(3.13072183492  + 3253.30422216000 *T)
    sumR0  +=       0.246 * cos(2.58151525126  + 1795.25844372100 *T)
    sumR0  +=       0.187 * cos(4.06797969837  + 14577.18472611980 *T)
    sumR0  +=       0.185 * cos(0.96747889626  + 2604.73591316800 *T)
    sumR0  +=       0.184 * cos(1.46731725339  + 1437.17561419860 *T)
    sumR0  +=       0.186 * cos(2.55094700069  + 3188.71514561460 *T)
    sumR0  +=       0.211 * cos(4.23522784526  + 16703.07938715119 *T)
    sumR0  +=       0.196 * cos(2.80582160764  + 2796.69436760360 *T)
    sumR0  +=       0.198 * cos(5.92372067560  + 4133.38731516720 *T)
    sumR0  +=       0.238 * cos(2.30957041178  + 1879.55837475340 *T)
    sumR0  +=       0.224 * cos(3.13550652378  + 3613.28538405140 *T)
    sumR0  +=       0.223 * cos(1.81801741599  + 3184.21170619740 *T)
    sumR0  +=       0.199 * cos(3.64983703995  + 4271.97551355160 *T)
    sumR0  +=       0.181 * cos(2.71721121316  + 4186.69526145100 *T)
    sumR0  +=       0.184 * cos(2.03051030984  + 6674.17861670160 *T)
    sumR0  +=       0.182 * cos(1.54431442810  + 9947.05568153210 *T)
    sumR0  +=       0.179 * cos(4.38762678121  + 9886.77220006400 *T)
    sumR0  +=       0.181 * cos(1.80044971979  + 6717.83021870380 *T)
    sumR0  +=       0.190 * cos(4.97714900529  + 6670.58818804980 *T)
    sumR0  +=       0.247 * cos(5.77972128678  + 29424.63423291600 *T)
    sumR0  +=       0.194 * cos(5.90745451439  + 6705.10329114740 *T)
    sumR0  +=       0.224 * cos(5.61893896091  + 227.47613278900 *T)
    sumR0  +=       0.182 * cos(2.77233344458  + 6887.41040183680 *T)
    sumR0  +=       0.224 * cos(1.94094757956  + 8226.57883637840 *T)
    sumR0  +=       0.181 * cos(4.61381081330  + 3361.94806716680 *T)
    sumR0  +=       0.203 * cos(5.24659800837  + 7586.68155006440 *T)
    sumR0  +=       0.183 * cos(5.23826533455  + 6578.13207918100 *T)
    sumR0  +=       0.207 * cos(0.88292186428  + 10156.90236013480 *T)
    sumR0  +=       0.232 * cos(0.75375986797  + 3232.94890276120 *T)
    sumR0  +=       0.179 * cos(5.89794151802  + 4452.25467180180 *T)
    sumR0  +=       0.177 * cos(6.16223629155  + 3547.31316399640 *T)
    sumR0  +=       0.177 * cos(0.19292898030  + 401.67212175720 *T)
    sumR0  +=       0.173 * cos(3.89814064100  + 6019.99192661860 *T)
    sumR0  +=       0.172 * cos(2.42933586666  + 6944.30877677240 *T)
    sumR0  +=       0.201 * cos(6.18392570315  + 6947.83189512140 *T)
    sumR0  +=       0.183 * cos(5.82189570737  + 8962.43809625800 *T)
    sumR0  +=       0.189 * cos(2.71231990816  + 742.99006053260 *T)
    sumR0  +=       0.172 * cos(3.86810071059  + 10028.95082710020 *T)
    sumR0  +=       0.188 * cos(5.22462530948  + 2125.87740737920 *T)
    sumR0  +=       0.196 * cos(1.88388569900  + 5032.77809620220 *T)
    sumR0  +=       0.209 * cos(5.78130707501  + 3343.36393731080 *T)
    sumR0  +=       0.167 * cos(6.26758992296  + 15849.86575174700 *T)
    sumR0  +=       0.189 * cos(2.38894304278  + 3212.59358336240 *T)
    sumR0  +=       0.208 * cos(5.62485439431  + 31172.65064598300 *T)
    sumR0  +=       0.170 * cos(1.07346401060  + 20426.57109242200 *T)
    sumR0  +=       0.183 * cos(5.21160560654  + 27832.03821928320 *T)
    sumR0  +=       0.181 * cos(2.19756653553  + 16703.04487984680 *T)
    sumR0  +=       0.214 * cos(1.66101921272  + 21265.52312652020 *T)
    sumR0  +=       0.165 * cos(2.10517924574  + 6688.27109009760 *T)
    sumR0  +=       0.172 * cos(0.85495655009  + 14591.41182012140 *T)
    sumR0  +=       0.164 * cos(1.68898542913  + 3468.63127003720 *T)
    sumR0  +=       0.183 * cos(1.56948199057  + 8962.47260356240 *T)
    sumR0  +=       0.191 * cos(5.54553650068  + 2806.98930834210 *T)
    sumR0  +=       0.162 * cos(4.06699160265  + 5209.47182633600 *T)
    sumR0  +=       0.165 * cos(3.44003333876  + 14421.83163698840 *T)
    sumR0  +=       0.163 * cos(2.42152585351  + 6314.00052443720 *T)
    sumR0  +=       0.189 * cos(0.57681608522  + 2973.38809773740 *T)
    sumR0  +=       0.169 * cos(2.86190807726  + 5490.30096152400 *T)
    sumR0  +=       0.174 * cos(1.69333072403  + 1744.49329471800 *T)
    sumR0  +=       0.224 * cos(2.35886332922  + 9638.94074787620 *T)
    sumR0  +=       0.207 * cos(3.13737248079  + 3336.68027343670 *T)
    sumR0  +=       0.221 * cos(3.14807587279  + 11610.91017538320 *T)
    sumR0  +=       0.158 * cos(5.87599422932  + 4606.17990532620 *T)
    sumR0  +=       0.166 * cos(4.61834424167  + 9175.75444534820 *T)
    sumR0  +=       0.162 * cos(5.40045489036  + 23.87843774780 *T)
    sumR0  +=       0.158 * cos(5.57047764661  + 4503.08713110760 *T)
    sumR0  +=       0.156 * cos(3.08339691033  + 3448.27595063840 *T)
    sumR0  +=       0.202 * cos(2.56469307457  + 7321.12213971360 *T)
    sumR0  +=       0.164 * cos(1.10347159072  + 3710.31224264020 *T)
    sumR0  +=       0.158 * cos(3.65995745934  + 6997.61672305620 *T)
    sumR0  +=       0.169 * cos(1.85985317117  + 1329.51209026000 *T)
    sumR0  +=       0.211 * cos(1.66624816725  + 3304.00706139560 *T)
    sumR0  +=       0.206 * cos(1.81572746145  + 1596.11913198180 *T)
    sumR0  +=       0.169 * cos(0.46310427277  + 17101.21113690720 *T)
    sumR0  +=       0.159 * cos(5.73384451559  + 1265.56747862640 *T)
    sumR0  +=       0.159 * cos(0.92643196778  + 4001.84535348080 *T)
    sumR0  +=       0.189 * cos(2.74163850920  + 2910.49111747480 *T)
    sumR0  +=       0.181 * cos(5.48699984659  + 3319.27678623280 *T)
    sumR0  +=       0.156 * cos(5.95462118470  + 8584.66166590080 *T)
    sumR0  +=       0.158 * cos(3.38732507447  + 8742.04270747140 *T)
    sumR0  +=       0.171 * cos(3.11406750626  + 5621.91023351320 *T)
    sumR0  +=       0.199 * cos(2.86370198218  + 10025.42770875120 *T)
    sumR0  +=       0.173 * cos(4.73384838991  + 9389.05354078620 *T)
    sumR0  +=       0.160 * cos(3.06861532268  + 1612.95133303160 *T)
    sumR0  +=       0.203 * cos(1.71629299113  + 699.27114822760 *T)
    sumR0  +=       0.198 * cos(0.56267868345  + 15265.88651930040 *T)
    sumR0  +=       0.186 * cos(2.44573687199  + 2814.85361486830 *T)
    sumR0  +=       0.167 * cos(0.44417942260  + 37895.42629036740 *T)
    sumR0  +=       0.152 * cos(5.08616610530  + 3322.79990458180 *T)
    sumR0  +=       0.162 * cos(2.58982908935  + 1883.08149310240 *T)
    sumR0  +=       0.178 * cos(4.58774508034  + 6685.15700666270 *T)
    sumR0  +=       0.191 * cos(4.12771141768  + 9911.63095888000 *T)
    sumR0  +=       0.185 * cos(4.32284898739  + 323.50541665740 *T)
    sumR0  +=       0.199 * cos(1.83524172935  + 3436.59165391760 *T)
    sumR0  +=       0.199 * cos(6.00292619279  + 4427.39591298580 *T)
    sumR0  +=       0.153 * cos(3.46447669590  + 22324.90505670940 *T)
    sumR0  +=       0.171 * cos(3.52256446633  + 5621.77561290760 *T)
    sumR0  +=       0.165 * cos(3.32970937124  + 1272.68102562720 *T)
    sumR0  +=       0.149 * cos(0.13370088791  + 36.02786667740 *T)
    sumR0  +=       0.149 * cos(1.23280447264  + 4193.80880845180 *T)
    sumR0  +=       0.156 * cos(0.13018524921  + 11925.27409260060 *T)
    sumR0  +=       0.146 * cos(1.24102730441  + 3372.58473251660 *T)
    sumR0  +=       0.154 * cos(5.63537863362  + 3185.25933756840 *T)
    sumR0  +=       0.197 * cos(3.23364188440  + 3774.32416457660 *T)
    sumR0  +=       0.180 * cos(3.47485722477  + 8564.30634650200 *T)
    sumR0  +=       0.185 * cos(2.18248214016  + 802.36392244620 *T)
    sumR0  +=       0.145 * cos(2.62035460481  + 3368.01398279660 *T)
    sumR0  +=       0.180 * cos(6.01236697222  + 6691.86151874940 *T)
    sumR0  +=       0.148 * cos(2.51710178830  + 3262.86188271590 *T)
    sumR0  +=       0.151 * cos(4.17118896078  + 16894.51039961060 *T)
    sumR0  +=       0.185 * cos(3.03939668350  + 3354.83952070140 *T)
    sumR0  +=       0.143 * cos(3.37267915445  + 3468.56395973440 *T)
    sumR0  +=       0.160 * cos(2.16705959730  + 0.98032106820 *T)
    sumR0  +=       0.190 * cos(1.63240538938  + 2324.94940881560 *T)
    sumR0  +=       0.140 * cos(1.27592043896  + 6549.68289171320 *T)
    sumR0  +=       0.186 * cos(1.44849264278  + 1015.66301788420 *T)
    sumR0  +=       0.141 * cos(4.19208049402  + 5813.29118932200 *T)
    sumR0  +=       0.153 * cos(5.97405255778  + 3568.08855948880 *T)
    sumR0  +=       0.136 * cos(0.45455359025  + 3193.80074017620 *T)
    sumR0  +=       0.188 * cos(0.98977442882  + 2538.24850425360 *T)
    sumR0  +=       0.157 * cos(3.32117477144  + 3181.66890891660 *T)
    sumR0  +=       0.164 * cos(1.67705754742  + 2751.54759969160 *T)
    sumR0  +=       0.134 * cos(5.26327877742  + 13212.88650966460 *T)
    sumR0  +=       0.184 * cos(0.73082474043  + 369.69981594040 *T)
    sumR0  +=       0.145 * cos(5.54086212771  + 3152.68727893721 *T)
    sumR0  +=       0.139 * cos(3.30464120245  + 6151.51663465280 *T)
    sumR0  +=       0.132 * cos(0.96753793729  + 3487.42411322340 *T)
    sumR0  +=       0.148 * cos(1.55321724063  + 7696.88787128380 *T)
    sumR0  +=       0.140 * cos(5.39000757800  + 2945.98654164060 *T)
    sumR0  +=       0.155 * cos(4.35207037526  + 4005.36847182980 *T)
    sumR0  +=       0.142 * cos(5.13441372991  + 1641.33321019660 *T)
    sumR0  +=       0.131 * cos(2.48606821423  + 1353.39052800780 *T)
    sumR0  +=       0.131 * cos(2.18754536881  + 1765.76662541760 *T)
    sumR0  +=       0.133 * cos(3.01300582040  + 3283.71405176420 *T)
    sumR0  +=       0.177 * cos(2.37558492362  + 11346.77862063940 *T)
    sumR0  +=       0.145 * cos(0.79525636457  + 2281.24775016280 *T)
    sumR0  +=       0.133 * cos(5.96263652624  + 34513.26307268280 *T)
    sumR0  +=       0.165 * cos(1.32559153826  + 6510.55198278040 *T)
    sumR0  +=       0.132 * cos(4.31118469609  + 1971.95217385480 *T)
    sumR0  +=       0.176 * cos(3.17539133688  + 6701.58017279840 *T)
    sumR0  +=       0.140 * cos(1.18723558164  + 1461.05405194640 *T)
    sumR0  +=       0.128 * cos(6.07261046225  + 3362.46325602620 *T)
    sumR0  +=       0.139 * cos(5.33541181199  + 6151.55114195720 *T)
    sumR0  +=       0.143 * cos(1.04060603784  + 3351.65812696370 *T)
    sumR0  +=       0.138 * cos(1.96009952022  + 2014.62345478880 *T)
    sumR0  +=       0.127 * cos(2.12310260285  + 2917.53735417280 *T)
    sumR0  +=       0.142 * cos(0.29094317859  + 3212.66089366520 *T)
    sumR0  +=       0.128 * cos(0.49974456440  + 10264.56588407340 *T)
    sumR0  +=       0.125 * cos(1.62548810323  + 10191.41746323240 *T)
    sumR0  +=       0.128 * cos(3.55358985291  + 3223.29255847960 *T)
    sumR0  +=       0.140 * cos(5.87379732521  + 6382.09845913040 *T)
    sumR0  +=       0.129 * cos(2.90428606882  + 6812.76681508600 *T)
    sumR0  +=       0.127 * cos(4.03101399507  + 10448.43547097540 *T)
    sumR0  +=       0.128 * cos(3.23265469000  + 4576.68808702280 *T)
    sumR0  +=       0.124 * cos(2.02283056331  + 6724.94376570460 *T)
    sumR0  +=       0.134 * cos(0.19295362623  + 3130.83644961080 *T)
    sumR0  +=       0.145 * cos(5.04767049759  + 2281.21324285840 *T)
    sumR0  +=       0.143 * cos(4.45744983982  + 7100.70949727480 *T)
    sumR0  +=       0.133 * cos(5.00404182123  + 11883.58313273480 *T)
    sumR0  +=       0.124 * cos(0.73925895836  + 1485.98012106520 *T)
    sumR0  +=       0.123 * cos(0.47283015475  + 3347.29609057390 *T)
    sumR0  +=       0.137 * cos(5.73370922615  + 5244.04923920100 *T)
    sumR0  +=       0.152 * cos(5.01464455460  + 3358.42494881780 *T)
    sumR0  +=       0.122 * cos(1.37003015612  + 3370.04193523580 *T)
    sumR0  +=       0.143 * cos(1.95740753086  + 2921.68802779920 *T)
    sumR0  +=       0.123 * cos(3.94139290312  + 4957.08687808040 *T)
    sumR0  +=       0.130 * cos(0.90844304292  + 3427.92063123960 *T)
    sumR0  +=       0.135 * cos(3.56996218538  + 3229.42578441220 *T)
    sumR0  +=       0.143 * cos(5.30225767870  + 14128.24277124560 *T)
    sumR0  +=       0.134 * cos(2.47836558803  + 7167.19690618920 *T)
    sumR0  +=       0.150 * cos(5.89116199733  + 966.97087743560 *T)
    sumR0  +=       0.123 * cos(3.58787674774  + 3318.76159737340 *T)
    sumR0  +=       0.153 * cos(0.37199746202  + 6533.14612897330 *T)
    sumR0  +=       0.145 * cos(1.73326841661  + 6637.50594109460 *T)
    sumR0  +=       0.133 * cos(3.27553306809  + 11876.46958573400 *T)
    sumR0  +=       0.119 * cos(5.15087396729  + 15636.56665630900 *T)
    sumR0  +=       0.117 * cos(3.70999224131  + 3343.15522398060 *T)
    sumR0  +=       0.115 * cos(5.99541570797  + 11986.67590695340 *T)
    sumR0  +=       0.152 * cos(5.66662339960  + 6530.17718355670 *T)
    sumR0  +=       0.123 * cos(0.05231448642  + 9801.42463766060 *T)
    sumR0  +=       0.130 * cos(1.42377926040  + 16063.16484718500 *T)
    sumR0  +=       0.120 * cos(4.19579234337  + 604.84974070480 *T)
    sumR0  +=       0.126 * cos(1.38788085609  + 169.58018313300 *T)
    sumR0  +=       0.122 * cos(0.18785107006  + 18849.22754997420 *T)
    sumR0  +=       0.113 * cos(4.67731144918  + 24491.42579258340 *T)
    sumR0  +=       0.122 * cos(1.21803689218  + 3067.93946934820 *T)
    sumR0  +=       0.118 * cos(3.92389195187  + 2963.23881878400 *T)
    sumR0  +=       0.137 * cos(0.80076862963  + 3123.79021291280 *T)
    sumR0  +=       0.124 * cos(0.77274469912  + 12465.53441274220 *T)
    sumR0  +=       0.132 * cos(4.73888023892  + 2267.00340250900 *T)
    sumR0  +=       0.122 * cos(3.66231014052  + 8756.26980147300 *T)
    sumR0  +=       0.113 * cos(3.74955221887  + 21393.54196985760 *T)
    sumR0  +=       0.153 * cos(5.00596002145  + 15906.76412668260 *T)
    sumR0  +=       0.109 * cos(1.46394078440  + 2854.64037391020 *T)
    sumR0  +=       0.125 * cos(6.00744459186  + 1442.26120876020 *T)
    sumR0  +=       0.140 * cos(0.64532646558  + 13532.02988993220 *T)
    sumR0  +=       0.108 * cos(2.11691804278  + 43.71891230500 *T)
    sumR0  +=       0.130 * cos(1.43185911960  + 3370.10424500320 *T)
    sumR0  +=       0.133 * cos(2.31239626151  + 9072.66167112960 *T)
    sumR0  +=       0.107 * cos(5.68864833782  + 4878.85286267780 *T)
    sumR0  +=       0.116 * cos(3.91262234002  + 11904.91877320180 *T)
    sumR0  +=       0.106 * cos(1.99989592271  + 3497.01314720220 *T)
    sumR0  +=       0.106 * cos(5.79020088316  + 1.48447270830 *T)
    sumR0  +=       0.105 * cos(1.79751699456  + 3517.30615683360 *T)
    sumR0  +=       0.130 * cos(0.80396347034  + 3397.51080163540 *T)
    sumR0  +=       0.107 * cos(5.29449090921  + 10.63666534980 *T)
    sumR0  +=       0.137 * cos(0.47037516401  + 13575.74880223720 *T)
    sumR0  +=       0.104 * cos(5.03145022037  + 1751.52227776380 *T)
    sumR0  +=       0.146 * cos(2.57659804831  + 3308.64012088300 *T)
    sumR0  +=       0.103 * cos(3.29589452622  + 4379.63903749020 *T)
    sumR0  +=       0.104 * cos(0.77903634693  + 1751.55678506820 *T)
    sumR0  +=       0.118 * cos(4.11892224522  + 10927.29397676420 *T)
    sumR0  +=       0.106 * cos(2.25714927822  + 13207.02930736500 *T)
    sumR0  +=       0.126 * cos(1.10289326962  + 6680.79497027290 *T)
    sumR0  +=       0.105 * cos(1.41449747957  + 10081.21114201300 *T)
    sumR0  +=       0.110 * cos(6.16072219971  + 3338.06962941900 *T)
    sumR0  +=       0.113 * cos(0.61947991056  + 13421.82356871280 *T)
    sumR0  +=       0.139 * cos(2.21345603147  + 1954.71715036360 *T)
    sumR0  +=       0.102 * cos(3.81767069895  + 6670.17915313570 *T)
    sumR0  +=       0.101 * cos(3.44583422545  + 1505.28780909299 *T)
    sumR0  +=       0.105 * cos(2.46134952447  + 5525.86369599260 *T)
    sumR0  +=       0.119 * cos(5.64454862995  + 8425.71814811760 *T)
    sumR0  +=       0.101 * cos(1.87446179141  + 9161.52735134660 *T)
    sumR0  +=       0.111 * cos(5.46183253107  + 7214.43893684320 *T)
    sumR0  +=       0.123 * cos(0.14056284625  + 16489.76303806100 *T)
    sumR0  +=       0.128 * cos(0.00452525697  + 2185.25126929280 *T)
    sumR0  +=       0.106 * cos(6.20225179330  + 3717.98603461560 *T)
    sumR0  +=       0.099 * cos(1.70554147455  + 1655.56030419820 *T)
    sumR0  +=       0.100 * cos(1.76373681853  + 3002.36972771680 *T)
    sumR0  +=       0.101 * cos(0.55693458655  + 3337.15661865360 *T)
    sumR0  +=       0.101 * cos(0.56441794685  + 30774.50164257480 *T)
    sumR0  +=       0.098 * cos(1.93983799659  + 5562.46906129680 *T)
    sumR0  +=       0.130 * cos(5.65255471249  + 2735.76268599500 *T)
    sumR0  +=       0.125 * cos(6.18575288406  + 9698.33186344200 *T)
    sumR0  +=       0.100 * cos(2.35767934317  + 3337.07205469860 *T)
    sumR0  +=       0.100 * cos(3.76317527562  + 3344.15279870100 *T)
    sumR0  +=       0.105 * cos(5.33897491215  + 5096.08411137890 *T)
    sumR0  +=       0.103 * cos(1.24189801298  + 1385.89527633620 *T)
    sumR0  +=       0.101 * cos(1.42727612221  + 37455.72649597440 *T)
    sumR0  +=       0.096 * cos(2.55328478399  + 4466.48176580340 *T)
    sumR0  +=       0.097 * cos(5.86779127410  + 5220.17080145320 *T)
    sumR0  +=       0.101 * cos(2.61513508379  + 16858.41522263039 *T)
    sumR0  +=       0.107 * cos(4.23943588615  + 8823.86715152580 *T)
    sumR0  +=       0.100 * cos(4.38845221965  + 3337.10656200300 *T)
    sumR0  +=       0.100 * cos(1.73240240188  + 3344.11829139661 *T)
    sumR0  +=       0.101 * cos(3.51764169063  + 4694.00295470760 *T)
    sumR0  +=       0.096 * cos(3.16895015703  + 12825.64519470380 *T)
    sumR0  +=       0.101 * cos(6.24060604565  + 2942.53073359440 *T)
    sumR0  +=       0.103 * cos(1.09500325425  + 12089.76868117200 *T)
    sumR0  +=       0.106 * cos(5.57931664756  + 39048.32250960720 *T)
    sumR0  +=       0.126 * cos(3.40521738550  + 6681.65473652630 *T)
    sumR0  +=       0.105 * cos(2.52404371984  + 1755.06264976500 *T)
    sumR0  +=       0.115 * cos(6.10031030713  + 3145.64104223920 *T)
    sumR0  +=       0.092 * cos(5.06385191401  + 5085.10572141780 *T)
    sumR0  +=       0.097 * cos(1.11670836674  + 3341.79306935190 *T)
    sumR0  +=       0.092 * cos(5.83209800268  + 6682.70932610790 *T)
    sumR0  +=       0.121 * cos(2.39411154886  + 18208.34994259200 *T)
    sumR0  +=       0.110 * cos(2.51677822824  + 1045.15483618760 *T)
    sumR0  +=       0.091 * cos(1.07173725022  + 927.83996850280 *T)
    sumR0  +=       0.096 * cos(3.60155667418  + 13363.43002786740 *T)
    sumR0  +=       0.097 * cos(5.34224904711  + 7636.46637799920 *T)
    sumR0  +=       0.094 * cos(1.20811080231  + 9495.66943335380 *T)
    sumR0  +=       0.090 * cos(4.17568982529  + 0.06731030280 *T)
    sumR0  +=       0.097 * cos(3.20862890174  + 5298.33750655300 *T)
    sumR0  +=       0.092 * cos(3.20029831258  + 4403.51747523800 *T)
    sumR0  +=       0.090 * cos(3.79900428724  + 685.04405422600 *T)
    sumR0  +=       0.101 * cos(2.12778329280  + 10050.28646756720 *T)
    sumR0  +=       0.098 * cos(3.89390262307  + 2373.64154926420 *T)
    sumR0  +=       0.090 * cos(4.49046504054  + 6453.74872061060 *T)
    sumR0  +=       0.096 * cos(5.13129666936  + 7792.86709850160 *T)
    sumR0  +=       0.089 * cos(4.51228904794  + 10037.08975128440 *T)
    sumR0  +=       0.098 * cos(0.65689906709  + 2810.98877190800 *T)
    sumR0  +=       0.114 * cos(4.15908543927  + 9090.47419324760 *T)
    sumR0  +=       0.094 * cos(4.78960145866  + 35707.71008290740 *T)
    sumR0  +=       0.087 * cos(6.26525215858  + 34554.81386366760 *T)
    sumR0  +=       0.087 * cos(1.50181194995  + 9983.70424446160 *T)
    sumR0  +=       0.087 * cos(1.43790822626  + 9488.62319665580 *T)
    sumR0  +=       0.090 * cos(2.79573736864  + 47477.56377607380 *T)
    sumR0  +=       0.094 * cos(5.68919291448  + 6106.88005506480 *T)
    sumR0  +=       0.087 * cos(0.74049208991  + 41990.78593289880 *T)
    sumR0  +=       0.101 * cos(0.99308133443  + 6614.73744448520 *T)
    sumR0  +=       0.115 * cos(3.49732055752  + 4.50343941720 *T)
    sumR0  +=       0.086 * cos(0.18628425275  + 216.82221378700 *T)
    sumR0  +=       0.099 * cos(1.06539589813  + 2810.85415130240 *T)
    sumR0  +=       0.087 * cos(3.02856210899  + 7218.02936549500 *T)
    sumR0  +=       0.085 * cos(4.58476007089  + 1699.27921650320 *T)
    sumR0  +=       0.088 * cos(4.47996444588  + 21957.68072774700 *T)
    sumR0  +=       0.106 * cos(6.12011880730  + 4819.47900076420 *T)
    sumR0  +=       0.089 * cos(3.88600719587  + 12199.97500239140 *T)
    sumR0  +=       0.097 * cos(2.70091698598  + 3339.43178404770 *T)
    sumR0  +=       0.118 * cos(1.02739527200  + 34115.11406927460 *T)
    sumR0  +=       0.101 * cos(0.36591831115  + 2942.39611298880 *T)
    sumR0  +=       0.084 * cos(5.66081869654  + 1169.58825140860 *T)
    sumR0  +=       0.086 * cos(1.91214909013  + 55516.41870984820 *T)
    sumR0  +=       0.084 * cos(4.30909434949  + 4150.08989614680 *T)
    sumR0  +=       0.094 * cos(1.36645151323  + 9093.99731159660 *T)
    sumR0  +=       0.097 * cos(2.26544245706  + 13361.46938573100 *T)
    sumR0  +=       0.099 * cos(1.16054562056  + 128.01884333740 *T)
    sumR0  +=       0.094 * cos(1.10573952457  + 4591.44262300060 *T)
    sumR0  +=       0.092 * cos(0.74386387542  + 4845.90023579280 *T)
    sumR0  +=       0.101 * cos(1.51570299040  + 7807.09419250320 *T)
    sumR0  +=       0.096 * cos(1.02187746571  + 2899.78714182220 *T)
    sumR0  +=       0.086 * cos(0.66787658068  + 2957.73314812880 *T)
    sumR0  +=       0.087 * cos(3.49676189640  + 4061.75165790600 *T)
    sumR0  +=       0.095 * cos(3.35452781355  + 661.23292678100 *T)
    sumR0  +=       0.087 * cos(0.04840436087  + 23546.75362303080 *T)
    sumR0  +=       0.114 * cos(2.83270741566  + 394.62588505920 *T)
    sumR0  +=       0.081 * cos(3.44420073782  + 148.07872442630 *T)
    sumR0  +=       0.081 * cos(1.27062968985  + 4893.07995667940 *T)
    sumR0  +=       0.083 * cos(5.48462687348  + 4349.16189758320 *T)
    sumR0  +=       0.105 * cos(3.47842036130  + 3413.90955255880 *T)
    sumR0  +=       0.084 * cos(4.33209852801  + 51449.32426327820 *T)
    sumR0  +=       0.093 * cos(1.59775582035  + 5088.21980485270 *T)
    sumR0  +=       0.088 * cos(3.29950478012  + 2655.50106217100 *T)
    sumR0  +=       0.081 * cos(1.65543714666  + 78263.70942472259 *T)
    sumR0  +=       0.086 * cos(1.19342535692  + 4029.24690957760 *T)
    sumR0  +=       0.079 * cos(4.41964678009  + 151.04766984290 *T)
    sumR0  +=       0.084 * cos(1.96289182788  + 14867.73751589219 *T)
    sumR0  +=       0.091 * cos(5.76229906142  + 2544.33167353560 *T)
    sumR0  +=       0.086 * cos(3.78659802914  + 1481.40937134520 *T)
    sumR0  +=       0.082 * cos(2.85633413428  + 14164.81362924540 *T)
    sumR0  +=       0.078 * cos(4.29426812598  + 187.92514776260 *T)
    sumR0  +=       0.078 * cos(5.38566529072  + 2655.56837247380 *T)
    sumR0  +=       0.084 * cos(2.15121245434  + 3311.12060839640 *T)
    sumR0  +=       0.078 * cos(2.01651247942  + 9124.92198604240 *T)
    sumR0  +=       0.083 * cos(4.28280129037  + 74923.09699802278 *T)
    sumR0  +=       0.077 * cos(1.45903147715  + 2544.38173018620 *T)
    sumR0  +=       0.081 * cos(3.62204400197  + 12306.59089495900 *T)
    sumR0  +=       0.100 * cos(1.33498371324  + 9278.84721956680 *T)
    sumR0  +=       0.089 * cos(2.62080312892  + 24341.52831745660 *T)
    sumR0  +=       0.098 * cos(3.78877505378  + 3198.43379966360 *T)
    sumR0  +=       0.081 * cos(4.34306418828  + 10177.19036923080 *T)
    sumR0  +=       0.080 * cos(5.80054348103  + 2751.01515718000 *T)
    sumR0  +=       0.092 * cos(2.80228448795  + 3707.83675566220 *T)
    sumR0  +=       0.082 * cos(2.45812896240  + 9815.65173166220 *T)
    sumR0  +=       0.076 * cos(3.03616461584  + 9993.38809263160 *T)
    sumR0  +=       0.076 * cos(3.92674885488  + 2910.08208256070 *T)
    sumR0  +=       0.076 * cos(3.20004040720  + 7733.49323658800 *T)
    sumR0  +=       0.098 * cos(3.28574616663  + 3289.33208883739 *T)
    sumR0  +=       0.077 * cos(2.05948930010  + 12928.73796892240 *T)
    sumR0  +=       0.075 * cos(0.85784191736  + 28638.90558114660 *T)
    sumR0  +=       0.078 * cos(0.29690749181  + 10654.62101941260 *T)
    sumR0  +=       0.101 * cos(0.72388999452  + 48827.43118573260 *T)
    sumR0  +=       0.085 * cos(5.19438010130  + 2146.14816282300 *T)
    sumR0  +=       0.083 * cos(4.42475849926  + 2060.81785407180 *T)
    sumR0  +=       0.075 * cos(5.19097428778  + 1055.44977692610 *T)
    sumR0  +=       0.096 * cos(6.24132708357  + 11.04570026390 *T)
    sumR0  +=       0.074 * cos(3.37882325472  + 8006.16619393960 *T)
    sumR0  +=       0.101 * cos(0.95332927487  + 8186.51266249260 *T)
    sumR0  +=       0.092 * cos(0.96173005100  + 151.89728108520 *T)
    sumR0  +=       0.074 * cos(2.22716630210  + 272.67295735160 *T)
    sumR0  +=       0.092 * cos(5.56705883046  + 27873.58901026800 *T)
    sumR0  +=       0.091 * cos(3.73152787182  + 2544.29716623120 *T)
    sumR0  +=       0.084 * cos(0.41493230617  + 4296.90158267040 *T)
    sumR0  +=       0.074 * cos(5.09669613096  + 9755.23023837760 *T)
    sumR0  +=       0.078 * cos(1.79565224435  + 3564.95722240170 *T)
    sumR0  +=       0.075 * cos(4.58345774133  + 9153.90361602180 *T)
    sumR0  +=       0.074 * cos(3.88658218247  + 25298.29315444680 *T)
    sumR0  +=       0.085 * cos(5.31472029744  + 4716.38622654560 *T)
    sumR0  +=       0.074 * cos(1.46907693289  + 11527.12508919240 *T)
    sumR0  +=       0.101 * cos(2.94847804254  + 12032.93761653920 *T)
    sumR0  +=       0.072 * cos(0.24413748615  + 1324.94134054000 *T)
    sumR0  +=       0.078 * cos(0.17721609024  + 3391.89276456221 *T)
    sumR0  +=       0.073 * cos(1.65072742864  + 5452.26274007740 *T)
    sumR0  +=       0.073 * cos(4.91730083470  + 10787.63034454580 *T)
    sumR0  +=       0.073 * cos(3.28285909319  + 8830.91338822380 *T)
    sumR0  +=       0.081 * cos(5.46340742050  + 3856.07629779280 *T)
    sumR0  +=       0.073 * cos(4.61913155968  + 3288.35211178700 *T)
    sumR0  +=       0.096 * cos(5.71767221046  + 206.70073729660 *T)
    sumR0  +=       0.080 * cos(4.47918417138  + 12299.54465826100 *T)
    sumR0  +=       0.071 * cos(4.98130328330  + 12942.96506292400 *T)
    sumR0  +=       0.073 * cos(2.86057567032  + 4960.60999642940 *T)
    sumR0  +=       0.072 * cos(5.14120362338  + 4665.55376723980 *T)
    sumR0  +=       0.076 * cos(0.00928867041  + 6534.41316687600 *T)
    sumR0  +=       0.090 * cos(5.38224594245  + 2125.81009707640 *T)
    sumR0  +=       0.085 * cos(0.94196641948  + 2146.18267012740 *T)
    sumR0  +=       0.096 * cos(0.68069708134  + 10006.58480891440 *T)
    sumR0  +=       0.073 * cos(1.06048379839  + 3866.78027344540 *T)
    sumR0  +=       0.072 * cos(0.84342376295  + 3334.54651107000 *T)
    sumR0  +=       0.091 * cos(5.12701600437  + 3392.87274161260 *T)
    sumR0  +=       0.076 * cos(5.96038992872  + 3482.79105373600 *T)
    sumR0  +=       0.090 * cos(1.46230169383  + 5092.16921176800 *T)
    sumR0  +=       0.071 * cos(1.16745785331  + 1574.84580128220 *T)
    sumR0  +=       0.072 * cos(3.86137692169  + 16762.43599541260 *T)
    sumR0  +=       0.070 * cos(0.07310763691  + 12839.87228870540 *T)
    sumR0  +=       0.073 * cos(0.11725123446  + 10507.80933288900 *T)
    sumR0  +=       0.081 * cos(5.86817748181  + 3495.96551583120 *T)
    sumR0  +=       0.071 * cos(2.07963275673  + 1063.31408345230 *T)
    sumR0  +=       0.068 * cos(3.30993028662  + 4106.40549114620 *T)
    sumR0  +=       0.069 * cos(0.33467954461  + 5.52292430740 *T)
    sumR0  +=       0.068 * cos(1.56404561711  + 2398.55036473080 *T)
    sumR0  +=       0.090 * cos(2.61953849393  + 3346.67834232960 *T)
    sumR0  +=       0.068 * cos(1.06108185791  + 191.38095580880 *T)
    sumR0  +=       0.077 * cos(0.87794040199  + 2917.94638908690 *T)
    sumR0  +=       0.067 * cos(0.29712068423  + 12012.58229714040 *T)
    sumR0  +=       0.067 * cos(2.55451787747  + 11567.19126307820 *T)
    sumR0  +=       0.083 * cos(1.82578570364  + 3511.28529731900 *T)
    sumR0  +=       0.068 * cos(2.68247854175  + 20213.25474333179 *T)
    sumR0  +=       0.086 * cos(2.79273636804  + 367.22432896240 *T)
    sumR0  +=       0.086 * cos(0.81171860574  + 209.77597708900 *T)
    sumR0  +=       0.067 * cos(5.28035928877  + 8966.38750317330 *T)
    sumR0  +=       0.067 * cos(0.20258020424  + 27433.88921587499 *T)
    sumR0  +=       0.071 * cos(2.76410148218  + 15650.79375031060 *T)
    sumR0  +=       0.090 * cos(0.71593793107  + 3451.79906898740 *T)
    sumR0  +=       0.067 * cos(6.18148420146  + 3966.28261901220 *T)
    sumR0  +=       0.074 * cos(3.61959120765  + 7064.10413197060 *T)
    sumR0  +=       0.068 * cos(2.69550405031  + 12725.45343477500 *T)
    sumR0  +=       0.074 * cos(1.24854309783  + 8027.56914470940 *T)
    sumR0  +=       0.084 * cos(2.47006028699  + 3738.74417645580 *T)
    sumR0  +=       0.077 * cos(5.89874180572  + 4307.58330413540 *T)
    sumR0  +=       0.067 * cos(0.33285041428  + 19645.52555679060 *T)
    sumR0  +=       0.083 * cos(3.79134493696  + 2472.67876262220 *T)
    sumR0  +=       0.064 * cos(0.43216125634  + 1994.26813539000 *T)
    sumR0  +=       0.071 * cos(0.65707804543  + 860.30992875280 *T)
    sumR0  +=       0.064 * cos(0.98960038205  + 3376.17516116840 *T)
    sumR0  +=       0.070 * cos(0.96227192685  + 8428.83223155250 *T)
    sumR0  +=       0.066 * cos(0.65425879867  + 8219.46528937760 *T)
    sumR0  +=       0.070 * cos(4.36060455306  + 8436.69653807870 *T)
    sumR0  +=       0.084 * cos(4.50083253887  + 3738.77868376020 *T)
    sumR0  +=       0.073 * cos(5.69670871544  + 14687.39104733920 *T)
    sumR0  +=       0.071 * cos(0.36334330253  + 2619.47319549360 *T)
    sumR0  +=       0.080 * cos(5.85771861744  + 7747.72033058960 *T)
    sumR0  +=       0.068 * cos(3.06409889136  + 8446.99147881720 *T)
    sumR0  +=       0.066 * cos(1.07137364271  + 1987.22189869200 *T)
    sumR0  +=       0.061 * cos(3.71171429905  + 1055.85881184020 *T)
    sumR0  +=       0.079 * cos(0.06487001849  + 6286.59896834040 *T)
    sumR0  +=       0.075 * cos(1.54177744365  + 1059.39918384140 *T)
    sumR0  +=       0.061 * cos(1.47350010728  + 1685.05212250160 *T)
    sumR0  +=       0.061 * cos(0.46897959303  + 1727.66109366820 *T)
    sumR0  +=       0.064 * cos(2.11869877898  + 3305.04969223120 *T)
    sumR0  +=       0.060 * cos(1.17940250396  + 97670.38771289718 *T)
    sumR0  +=       0.060 * cos(5.13953821352  + 1898.33396428740 *T)
    sumR0  +=       0.064 * cos(4.04058406722  + 3945.46216740460 *T)
    sumR0  +=       0.063 * cos(5.20562384814  + 24606.13555322000 *T)
    sumR0  +=       0.061 * cos(0.55428968913  + 7082.89697515680 *T)
    sumR0  +=       0.075 * cos(4.81208819076  + 27299.24421193320 *T)
    sumR0  +=       0.059 * cos(1.75511446129  + 12082.65513417120 *T)
    sumR0  +=       0.061 * cos(5.88333855610  + 6901.63749583840 *T)
    sumR0  +=       0.074 * cos(3.46408580786  + 14556.89671702380 *T)
    sumR0  +=       0.074 * cos(4.40649822049  + 9360.60435331840 *T)
    sumR0  +=       0.058 * cos(4.22827613192  + 4531.53631857540 *T)
    sumR0  +=       0.080 * cos(2.14103357856  + 7917.30051372260 *T)
    sumR0  +=       0.075 * cos(1.85618011420  + 52566.19261584059 *T)
    sumR0  +=       0.075 * cos(5.15830659680  + 45884.96776244100 *T)
    sumR0  +=       0.068 * cos(2.06915254754  + 8958.52319664710 *T)
    sumR0  +=       0.056 * cos(3.77847713307  + 299.12639426920 *T)
    sumR0  +=       0.062 * cos(2.25564130818  + 9758.75335672660 *T)
    sumR0  +=       0.059 * cos(2.68657637673  + 6283.05859633920 *T)
    sumR0  +=       0.056 * cos(5.54148283448  + 13286.18363552360 *T)
    sumR0  +=       0.056 * cos(3.89200811600  + 8962.38803960740 *T)
    sumR0  +=       0.063 * cos(2.07254352069  + 7593.79509706520 *T)
    sumR0  +=       0.061 * cos(4.31994402891  + 4709.27267954480 *T)
    sumR0  +=       0.075 * cos(5.79419108898  + 1059.36467653700 *T)
    sumR0  +=       0.056 * cos(0.18569912415  + 7768.00833968560 *T)
    sumR0  +=       0.059 * cos(5.49627068572  + 7322.10246078180 *T)
    sumR0  +=       0.064 * cos(3.95992112220  + 3050.12694723020 *T)
    sumR0  +=       0.055 * cos(2.19506741266  + 7.04623669800 *T)
    sumR0  +=       0.057 * cos(4.05252695300  + 3930.20969621960 *T)
    sumR0  +=       0.075 * cos(3.99788100837  + 5483.25472482600 *T)
    sumR0  +=       0.060 * cos(4.61703903739  + 6990.50317605540 *T)
    sumR0  +=       0.073 * cos(4.20362525534  + 21150.81336588360 *T)
    sumR0  +=       0.066 * cos(2.82009591985  + 8771.00708379860 *T)
    sumR0  +=       0.054 * cos(0.38782719267  + 360.11078196160 *T)
    sumR0  +=       0.059 * cos(4.71734799803  + 6283.09310364360 *T)
    sumR0  +=       0.068 * cos(5.74655335863  + 987.74627292800 *T)

    # MARS Radius
    sumR1  = 0
    sumR1  +=  1107433.340 * cos(2.03250524950  + 3340.61242669980 *T)
    sumR1  +=  103175.886 * cos(2.37071845682  + 6681.22485339960 *T)
    sumR1  +=   12877.200 * cos(0.00000000000  + 0.00000000000 *T)
    sumR1  +=   10815.880 * cos(2.70888093803  + 10021.83728009940 *T)
    sumR1  +=    1194.550 * cos(3.04702182503  + 13362.44970679920 *T)
    sumR1  +=     438.579 * cos(2.88835072628  + 2281.23049651060 *T)
    sumR1  +=     395.698 * cos(3.42324611291  + 3344.13554504880 *T)
    sumR1  +=     182.572 * cos(1.58428644001  + 2544.31441988340 *T)
    sumR1  +=     135.850 * cos(3.38507017993  + 16703.06213349900 *T)
    sumR1  +=     128.204 * cos(0.62991220570  + 1059.38193018920 *T)
    sumR1  +=     127.068 * cos(1.95389775740  + 796.29800681640 *T)
    sumR1  +=     118.443 * cos(2.99761345074  + 2146.16541647520 *T)
    sumR1  +=     128.362 * cos(6.04343360441  + 3337.08930835080 *T)
    sumR1  +=      87.537 * cos(3.42052758979  + 398.14900340820 *T)
    sumR1  +=      83.026 * cos(3.85574986653  + 3738.76143010800 *T)
    sumR1  +=      75.598 * cos(4.45101839349  + 6151.53388830500 *T)
    sumR1  +=      71.999 * cos(2.76442180680  + 529.69096509460 *T)
    sumR1  +=      66.542 * cos(2.54892602695  + 1751.53953141600 *T)
    sumR1  +=      54.314 * cos(0.67750943459  + 8962.45534991020 *T)
    sumR1  +=      51.035 * cos(3.72585409207  + 6684.74797174860 *T)
    sumR1  +=      66.430 * cos(4.40597549957  + 1748.01641306700 *T)
    sumR1  +=      47.863 * cos(2.28527896843  + 2914.01423582380 *T)
    sumR1  +=      49.428 * cos(5.72959428364  + 3340.59517304760 *T)
    sumR1  +=      49.424 * cos(1.47717922226  + 3340.62968035200 *T)
    sumR1  +=      57.518 * cos(0.54354327916  + 1194.44701022460 *T)
    sumR1  +=      48.318 * cos(2.58061691301  + 3149.16416058820 *T)
    sumR1  +=      36.384 * cos(6.02728752344  + 3185.19202726560 *T)
    sumR1  +=      37.176 * cos(5.81439911546  + 1349.86740965880 *T)
    sumR1  +=      36.036 * cos(5.89508336048  + 3333.49887969900 *T)
    sumR1  +=      31.115 * cos(0.97832506960  + 191.44826611160 *T)
    sumR1  +=      38.953 * cos(2.31900090554  + 4136.91043351620 *T)
    sumR1  +=      27.244 * cos(5.41367977087  + 1592.59601363280 *T)
    sumR1  +=      24.300 * cos(3.75843924498  + 155.42039943420 *T)
    sumR1  +=      22.804 * cos(1.74830773908  + 5088.62883976680 *T)
    sumR1  +=      22.324 * cos(0.93932040730  + 951.71840625060 *T)
    sumR1  +=      21.708 * cos(3.83571581352  + 6283.07584999140 *T)
    sumR1  +=      21.304 * cos(0.78049229782  + 1589.07289528380 *T)
    sumR1  +=      21.631 * cos(4.56895741061  + 3532.06069281140 *T)
    sumR1  +=      17.956 * cos(4.21930481803  + 3870.30339179440 *T)
    sumR1  +=      18.237 * cos(0.41328624131  + 5486.77784317500 *T)
    sumR1  +=      16.251 * cos(3.80760134974  + 3340.54511639700 *T)
    sumR1  +=      16.803 * cos(5.54857987615  + 3097.88382272579 *T)
    sumR1  +=      16.850 * cos(4.53690440252  + 4292.33083295040 *T)
    sumR1  +=      15.755 * cos(4.75736730681  + 9492.14631500480 *T)
    sumR1  +=      15.746 * cos(3.72356090283  + 20043.67456019880 *T)
    sumR1  +=      20.428 * cos(3.13540712557  + 4690.47983635860 *T)
    sumR1  +=      14.699 * cos(5.95325006816  + 3894.18182954220 *T)
    sumR1  +=      16.251 * cos(3.39910907599  + 3340.67973700260 *T)
    sumR1  +=      14.259 * cos(3.99897353022  + 1990.74501704100 *T)
    sumR1  +=      16.528 * cos(0.96752074938  + 4399.99435688900 *T)
    sumR1  +=      13.010 * cos(5.14230107067  + 6677.70173505060 *T)
    sumR1  +=      12.492 * cos(1.03211063742  + 3341.59274776800 *T)
    sumR1  +=      16.463 * cos(3.53882915214  + 2700.71514038580 *T)
    sumR1  +=      16.171 * cos(2.34870953554  + 553.56940284240 *T)
    sumR1  +=      13.169 * cos(0.41461716663  + 5614.72937620960 *T)
    sumR1  +=      11.272 * cos(1.02375627844  + 12303.06777661000 *T)
    sumR1  +=      12.408 * cos(6.23142869816  + 5628.95647021120 *T)
    sumR1  +=      12.747 * cos(0.69046314049  + 3723.50895892300 *T)
    sumR1  +=      11.827 * cos(6.25283898676  + 2274.11694950980 *T)
    sumR1  +=      10.384 * cos(1.23257236014  + 426.59819087600 *T)
    sumR1  +=      11.208 * cos(1.31750963435  + 3496.03282613400 *T)
    sumR1  +=      10.345 * cos(0.90062464690  + 4535.05943692440 *T)
    sumR1  +=      12.215 * cos(4.22316056098  + 7079.37385680780 *T)
    sumR1  +=       9.765 * cos(3.45310940204  + 382.89653222320 *T)
    sumR1  +=       8.586 * cos(1.16471901139  + 2787.04302385740 *T)
    sumR1  +=       7.879 * cos(5.73808861790  + 2288.34404351140 *T)
    sumR1  +=       9.195 * cos(1.81689739851  + 6681.24210705180 *T)
    sumR1  +=       7.751 * cos(4.15046998466  + 6041.32756708560 *T)
    sumR1  +=       9.195 * cos(6.06945250657  + 6681.20759974740 *T)
    sumR1  +=       9.026 * cos(2.58210941053  + 2388.89402044920 *T)
    sumR1  +=       6.771 * cos(0.23987737854  + 11773.37681151540 *T)
    sumR1  +=       7.087 * cos(3.51414944377  + 8031.09226305840 *T)
    sumR1  +=       9.159 * cos(3.90158590550  + 3553.91152213780 *T)
    sumR1  +=       6.702 * cos(4.25572879119  + 242.72860397400 *T)
    sumR1  +=       7.232 * cos(3.70287400141  + 2818.03500860600 *T)
    sumR1  +=       6.546 * cos(0.04353472459  + 2957.71589447660 *T)
    sumR1  +=       8.781 * cos(2.19735028572  + 1221.84856632140 *T)
    sumR1  +=       6.540 * cos(2.11834687923  + 8429.24126646660 *T)
    sumR1  +=       6.836 * cos(4.04512263654  + 10025.36039844840 *T)
    sumR1  +=       7.280 * cos(4.26943100715  + 2803.80791460440 *T)
    sumR1  +=       7.676 * cos(1.00782250264  + 8432.76438481560 *T)
    sumR1  +=       5.732 * cos(3.13956241764  + 213.29909543800 *T)
    sumR1  +=       5.344 * cos(3.78220670098  + 5092.15195811580 *T)
    sumR1  +=       5.991 * cos(2.96408254428  + 6489.77658728800 *T)
    sumR1  +=       5.132 * cos(3.98288453952  + 7.11354700080 *T)
    sumR1  +=       6.261 * cos(1.90346469972  + 5621.84292321040 *T)
    sumR1  +=       5.241 * cos(2.67093795050  + 7477.52286021600 *T)
    sumR1  +=       6.261 * cos(1.60056695152  + 3347.72597370060 *T)
    sumR1  +=       6.526 * cos(2.76281601349  + 3339.63210563160 *T)
    sumR1  +=       4.591 * cos(1.81986278078  + 2810.92146160520 *T)
    sumR1  +=       5.460 * cos(4.60874445963  + 3583.34103067380 *T)
    sumR1  +=       4.733 * cos(0.90560026340  + 5099.26550511660 *T)
    sumR1  +=       5.486 * cos(4.91408093456  + 7632.94325965020 *T)
    sumR1  +=       4.004 * cos(4.14067157211  + 9623.68827669120 *T)
    sumR1  +=       3.837 * cos(0.03343295627  + 7234.79425624200 *T)
    sumR1  +=       3.621 * cos(5.76532393672  + 4933.20844033260 *T)
    sumR1  +=       3.747 * cos(0.08778985966  + 6525.80445396540 *T)
    sumR1  +=       3.016 * cos(3.73838855125  + 6681.29216370240 *T)
    sumR1  +=       3.975 * cos(4.91288592965  + 2942.46342329160 *T)
    sumR1  +=       3.911 * cos(0.67451768877  + 3127.31333126180 *T)
    sumR1  +=       3.923 * cos(3.07703531632  + 3.52311834900 *T)
    sumR1  +=       3.944 * cos(0.53974754515  + 5884.92684658320 *T)
    sumR1  +=       2.901 * cos(4.66281989264  + 7210.91581849420 *T)
    sumR1  +=       2.803 * cos(1.00530177454  + 7064.12138562280 *T)
    sumR1  +=       3.153 * cos(4.54673175664  + 2487.41604494780 *T)
    sumR1  +=       2.784 * cos(0.05495331967  + 639.89728631400 *T)
    sumR1  +=       2.759 * cos(5.17099857257  + 5828.02847164760 *T)
    sumR1  +=       3.017 * cos(4.14668877243  + 6681.15754309680 *T)
    sumR1  +=       2.999 * cos(0.82918667275  + 5085.03841111500 *T)
    sumR1  +=       2.677 * cos(0.69427974692  + 2699.73481931760 *T)
    sumR1  +=       3.023 * cos(2.59489020298  + 2906.90068882300 *T)
    sumR1  +=       2.596 * cos(1.08693013783  + 4929.68532198360 *T)
    sumR1  +=       3.126 * cos(1.00027692782  + 2118.76386037840 *T)
    sumR1  +=       2.598 * cos(5.01195749912  + 10018.31416175040 *T)
    sumR1  +=       2.606 * cos(5.34420770679  + 10973.55568635000 *T)
    sumR1  +=       2.779 * cos(3.98358744953  + 6467.92575796160 *T)
    sumR1  +=       2.460 * cos(1.52653571249  + 6836.64525283380 *T)
    sumR1  +=       2.382 * cos(3.93610586965  + 11371.70468975820 *T)
    sumR1  +=       2.590 * cos(5.08892664109  + 12832.75874170460 *T)
    sumR1  +=       2.201 * cos(0.18880589605  + 9595.23908922340 *T)
    sumR1  +=       2.128 * cos(5.69450469171  + 3191.04922956520 *T)
    sumR1  +=       2.065 * cos(3.54133174600  + 1066.49547719000 *T)
    sumR1  +=       1.868 * cos(5.01592531810  + 3475.67750673520 *T)
    sumR1  +=       2.081 * cos(1.30274459220  + 7740.60678358880 *T)
    sumR1  +=       1.830 * cos(5.68365327697  + 3319.83703120740 *T)
    sumR1  +=       2.022 * cos(6.16092328060  + 1744.42598441520 *T)
    sumR1  +=       1.708 * cos(2.21675931288  + 2921.12778282460 *T)
    sumR1  +=       1.836 * cos(6.16477009621  + 8425.65083781480 *T)
    sumR1  +=       2.060 * cos(6.24041853265  + 6674.11130639880 *T)
    sumR1  +=       1.735 * cos(4.58243571826  + 10419.98628350760 *T)
    sumR1  +=       1.852 * cos(1.36883022935  + 15643.68020330980 *T)
    sumR1  +=       1.689 * cos(1.92829590736  + 3767.21061757580 *T)
    sumR1  +=       1.954 * cos(0.46215988899  + 10575.40668294180 *T)
    sumR1  +=       1.800 * cos(2.30913343660  + 3355.86489788480 *T)
    sumR1  +=       1.444 * cos(2.15042616899  + 10021.85453375160 *T)
    sumR1  +=       1.760 * cos(5.87267582299  + 3320.25710730100 *T)
    sumR1  +=       1.845 * cos(4.06123235448  + 23384.28698689860 *T)
    sumR1  +=       1.423 * cos(4.65085713203  + 4562.46099302120 *T)
    sumR1  +=       1.410 * cos(1.50438410845  + 3325.35995551480 *T)
    sumR1  +=       1.444 * cos(0.11969993808  + 10021.82002644720 *T)
    sumR1  +=       1.443 * cos(0.55828572929  + 15113.98923821520 *T)
    sumR1  +=       1.334 * cos(1.25483731308  + 7875.67186362420 *T)
    sumR1  +=       1.442 * cos(1.38958943728  + 6682.20517446780 *T)
    sumR1  +=       1.606 * cos(4.34241260905  + 5331.35744374080 *T)
    sumR1  +=       1.377 * cos(0.14791683820  + 1758.65307841680 *T)
    sumR1  +=       1.282 * cos(5.78533014319  + 14584.29827312060 *T)
    sumR1  +=       1.281 * cos(1.35125896423  + 10404.73381232260 *T)
    sumR1  +=       1.471 * cos(0.34649469321  + 8969.56889691100 *T)
    sumR1  +=       1.245 * cos(6.02681709446  + 3264.34635542420 *T)
    sumR1  +=       1.234 * cos(1.88341938354  + 10177.25767953360 *T)
    sumR1  +=       1.552 * cos(5.92727420332  + 8439.87793181640 *T)
    sumR1  +=       1.182 * cos(4.18100226016  + 3360.96774609859 *T)
    sumR1  +=       1.333 * cos(1.86551437099  + 692.15760122680 *T)
    sumR1  +=       1.189 * cos(0.89892514263  + 13916.01910964160 *T)
    sumR1  +=       1.192 * cos(4.30473818946  + 6894.52394883760 *T)
    sumR1  +=       1.145 * cos(0.15197504252  + 3134.42687826260 *T)
    sumR1  +=       1.299 * cos(1.44631688592  + 6254.62666252360 *T)
    sumR1  +=       1.494 * cos(1.54417907271  + 3361.38782219220 *T)
    sumR1  +=       1.054 * cos(0.67591855339  + 3344.20285535160 *T)
    sumR1  +=       1.210 * cos(0.85117132607  + 3120.19978426100 *T)
    sumR1  +=       1.061 * cos(0.13258232364  + 522.57741809380 *T)
    sumR1  +=       1.003 * cos(1.28021784912  + 4569.57454002200 *T)
    sumR1  +=       1.012 * cos(5.77497169905  + 14314.16811304980 *T)
    sumR1  +=       1.169 * cos(2.99767730172  + 6247.51311552280 *T)
    sumR1  +=       1.062 * cos(0.25467691252  + 3337.02199804800 *T)
    sumR1  +=       1.032 * cos(5.69833683620  + 536.80451209540 *T)
    sumR1  +=       1.082 * cos(2.58016159997  + 6261.74020952440 *T)
    sumR1  +=       0.867 * cos(1.50283829131  + 6127.65545055720 *T)
    sumR1  +=       0.953 * cos(2.92377030638  + 5729.50644714900 *T)
    sumR1  +=       0.887 * cos(5.28743574972  + 10213.28554621100 *T)
    sumR1  +=       1.108 * cos(2.84339171890  + 640.87760738220 *T)
    sumR1  +=       0.861 * cos(3.21491381593  + 3914.95722503460 *T)
    sumR1  +=       1.069 * cos(0.49400181869  + 11243.68584642080 *T)
    sumR1  +=       0.800 * cos(4.64154002129  + 6144.42034130420 *T)
    sumR1  +=       0.799 * cos(3.33215212691  + 9830.38901398780 *T)
    sumR1  +=       1.046 * cos(4.05907877700  + 6158.64743530580 *T)
    sumR1  +=       0.743 * cos(4.93337997290  + 3230.40610548040 *T)
    sumR1  +=       0.740 * cos(3.01511102229  + 10818.13528691580 *T)
    sumR1  +=       0.716 * cos(2.52567242762  + 5202.35827933520 *T)
    sumR1  +=       0.705 * cos(4.39344403231  + 9808.53818466140 *T)
    sumR1  +=       0.786 * cos(1.72357089999  + 3178.14579056760 *T)
    sumR1  +=       0.719 * cos(0.38924465445  + 6298.32832117640 *T)
    sumR1  +=       0.717 * cos(6.23927283856  + 3369.06161416760 *T)
    sumR1  +=       0.704 * cos(1.96763488855  + 6688.33840040040 *T)
    sumR1  +=       0.882 * cos(2.10339285493  + 2301.58581590939 *T)
    sumR1  +=       0.913 * cos(2.30662326829  + 13517.87010623340 *T)
    sumR1  +=       0.914 * cos(4.36610748804  + 13365.97282514820 *T)
    sumR1  +=       0.659 * cos(3.40905557071  + 1648.44675719740 *T)
    sumR1  +=       0.828 * cos(2.13305589137  + 7903.07341972100 *T)
    sumR1  +=       0.652 * cos(5.97050449428  + 574.34479833480 *T)
    sumR1  +=       0.688 * cos(4.02232587894  + 1052.26838318840 *T)
    sumR1  +=       0.709 * cos(2.24658423912  + 9225.53927328300 *T)
    sumR1  +=       0.590 * cos(0.67418318613  + 8955.34180290940 *T)
    sumR1  +=       0.604 * cos(2.73550395152  + 16858.48253293320 *T)
    sumR1  +=       0.785 * cos(3.58704322751  + 2693.60159338500 *T)
    sumR1  +=       0.570 * cos(4.11146135269  + 6034.21402008480 *T)
    sumR1  +=       0.624 * cos(4.45415856101  + 12964.30070339100 *T)
    sumR1  +=       0.563 * cos(3.72383942144  + 13119.72110282519 *T)
    sumR1  +=       0.547 * cos(3.04142188931  + 2707.82868738660 *T)
    sumR1  +=       0.536 * cos(4.65546776421  + 135.06508003540 *T)
    sumR1  +=       0.680 * cos(3.62640859214  + 6048.44111408640 *T)
    sumR1  +=       0.535 * cos(3.32503801413  + 9779.10867612540 *T)
    sumR1  +=       0.615 * cos(0.31792697641  + 7910.18696672180 *T)
    sumR1  +=       0.506 * cos(1.24339056369  + 4555.34744602040 *T)
    sumR1  +=       0.633 * cos(1.32124830132  + 2648.45482547300 *T)
    sumR1  +=       0.671 * cos(2.88173567087  + 3316.73398895200 *T)
    sumR1  +=       0.570 * cos(4.94522781693  + 6923.95345737360 *T)
    sumR1  +=       0.489 * cos(0.48479532469  + 12935.85151592320 *T)
    sumR1  +=       0.490 * cos(1.57712746955  + 162.46663613220 *T)
    sumR1  +=       0.492 * cos(4.08592823251  + 10021.90459040220 *T)
    sumR1  +=       0.468 * cos(4.50628939729  + 2384.32327072920 *T)
    sumR1  +=       0.462 * cos(1.30653882308  + 17256.63153634140 *T)
    sumR1  +=       0.444 * cos(5.20556841675  + 13358.92658845020 *T)
    sumR1  +=       0.511 * cos(5.29436707935  + 853.19638175200 *T)
    sumR1  +=       0.458 * cos(5.52316111365  + 6438.49624942560 *T)
    sumR1  +=       0.443 * cos(4.37626774329  + 12168.00269657460 *T)
    sumR1  +=       0.429 * cos(4.06329082027  + 149.56319713460 *T)
    sumR1  +=       0.502 * cos(6.12188881856  + 8273.82086703240 *T)
    sumR1  +=       0.420 * cos(3.51990468754  + 9168.64089834740 *T)
    sumR1  +=       0.452 * cos(0.15931716927  + 7895.95987272020 *T)
    sumR1  +=       0.401 * cos(0.93276882765  + 433.71173787680 *T)
    sumR1  +=       0.491 * cos(4.49372643902  + 10021.76996979660 *T)
    sumR1  +=       0.390 * cos(1.26790524143  + 5459.37628707820 *T)
    sumR1  +=       0.460 * cos(3.56765356005  + 1596.18644228460 *T)
    sumR1  +=       0.514 * cos(4.14617297678  + 16460.33352952499 *T)
    sumR1  +=       0.510 * cos(2.12338136542  + 3335.08950239240 *T)
    sumR1  +=       0.365 * cos(0.84035244694  + 3364.49086444760 *T)
    sumR1  +=       0.444 * cos(0.42154996651  + 9866.41688066520 *T)
    sumR1  +=       0.380 * cos(2.74315160117  + 3344.49376205780 *T)
    sumR1  +=       0.355 * cos(2.20831807719  + 5518.75014899180 *T)
    sumR1  +=       0.374 * cos(0.93920520997  + 3863.18984479360 *T)
    sumR1  +=       0.380 * cos(0.60825828556  + 3980.50971301380 *T)
    sumR1  +=       0.366 * cos(1.45297694305  + 1039.02661079040 *T)
    sumR1  +=       0.458 * cos(5.01867719021  + 10551.52824519400 *T)
    sumR1  +=       0.352 * cos(5.59661824792  + 16062.18452611680 *T)
    sumR1  +=       0.361 * cos(5.43698827477  + 11236.57229942000 *T)
    sumR1  +=       0.323 * cos(3.25474887566  + 3116.26763099790 *T)
    sumR1  +=       0.396 * cos(5.43635200834  + 16173.37116840440 *T)
    sumR1  +=       0.316 * cos(5.10862851689  + 17395.21973472580 *T)
    sumR1  +=       0.325 * cos(4.59103320065  + 19800.94595622480 *T)
    sumR1  +=       0.320 * cos(0.10022132590  + 1692.16566950240 *T)
    sumR1  +=       0.313 * cos(4.76509182263  + 3304.58456002240 *T)
    sumR1  +=       0.329 * cos(3.51515201342  + 1903.43681250120 *T)
    sumR1  +=       0.361 * cos(1.40179572820  + 8270.29774868340 *T)
    sumR1  +=       0.380 * cos(1.11278812152  + 3503.07906283200 *T)
    sumR1  +=       0.309 * cos(1.63585786252  + 11081.21921028860 *T)
    sumR1  +=       0.304 * cos(6.27804875103  + 2391.43681773000 *T)
    sumR1  +=       0.336 * cos(2.95777936427  + 6680.24453233140 *T)
    sumR1  +=       0.299 * cos(0.48751113813  + 3347.65866339780 *T)
    sumR1  +=       0.327 * cos(3.19951479738  + 3981.49003408200 *T)
    sumR1  +=       0.298 * cos(4.26967988388  + 13936.79450513400 *T)
    sumR1  +=       0.309 * cos(5.05775078180  + 12721.57209941700 *T)
    sumR1  +=       0.310 * cos(6.10243075443  + 17924.91069982040 *T)
    sumR1  +=       0.290 * cos(4.52880846817  + 5724.93569742900 *T)
    sumR1  +=       0.279 * cos(4.30091602167  + 3607.21946842160 *T)
    sumR1  +=       0.293 * cos(1.90646493379  + 6702.00024889200 *T)
    sumR1  +=       0.375 * cos(5.37146136739  + 3376.64029337720 *T)
    sumR1  +=       0.268 * cos(4.93796347898  + 13760.59871020740 *T)
    sumR1  +=       0.269 * cos(5.63342755549  + 4032.77002792660 *T)
    sumR1  +=       0.361 * cos(4.70448724123  + 9380.95967271720 *T)
    sumR1  +=       0.285 * cos(4.37639190605  + 6040.34724601740 *T)
    sumR1  +=       0.305 * cos(3.60878132297  + 7255.56965173440 *T)
    sumR1  +=       0.265 * cos(5.82002998533  + 1214.73501932060 *T)
    sumR1  +=       0.246 * cos(3.48844723382  + 3237.51965248120 *T)
    sumR1  +=       0.339 * cos(6.05128485092  + 6660.44945790720 *T)
    sumR1  +=       0.325 * cos(5.12798981834  + 11250.79939342160 *T)
    sumR1  +=       0.243 * cos(0.48346700024  + 13362.43245314700 *T)
    sumR1  +=       0.246 * cos(4.41923388458  + 3546.79797513700 *T)
    sumR1  +=       0.239 * cos(3.73073204843  + 76.26607127560 *T)
    sumR1  +=       0.239 * cos(6.27709788173  + 1964.83862685400 *T)
    sumR1  +=       0.242 * cos(2.51411096474  + 13362.46696045140 *T)
    sumR1  +=       0.238 * cos(4.61745673345  + 17277.40693183380 *T)
    sumR1  +=       0.232 * cos(3.13987850882  + 20199.09495963300 *T)
    sumR1  +=       0.245 * cos(3.88310965561  + 4407.10790388980 *T)
    sumR1  +=       0.242 * cos(3.07232693388  + 12566.15169998280 *T)
    sumR1  +=       0.277 * cos(3.88618268628  + 10596.18207843420 *T)
    sumR1  +=       0.277 * cos(6.14459469980  + 19402.79695281660 *T)
    sumR1  +=       0.222 * cos(2.20253799065  + 7107.82304427560 *T)
    sumR1  +=       0.218 * cos(1.37621606096  + 3415.39402526710 *T)
    sumR1  +=       0.297 * cos(3.78408680173  + 2766.26762836500 *T)
    sumR1  +=       0.264 * cos(5.35345322905  + 6816.28993343500 *T)
    sumR1  +=       0.213 * cos(1.10915354432  + 19004.64794940840 *T)
    sumR1  +=       0.213 * cos(4.60032427636  + 3973.39616601300 *T)
    sumR1  +=       0.230 * cos(4.39356021431  + 26724.89941359840 *T)
    sumR1  +=       0.202 * cos(0.60981942427  + 6677.63442474780 *T)
    sumR1  +=       0.268 * cos(5.05058237041  + 9381.93999378540 *T)
    sumR1  +=       0.199 * cos(1.34402137469  + 8329.67161059700 *T)
    sumR1  +=       0.198 * cos(5.08527311506  + 5820.91492464680 *T)
    sumR1  +=       0.201 * cos(0.05002970566  + 11780.49035851620 *T)
    sumR1  +=       0.199 * cos(2.66187512344  + 6696.47732458460 *T)
    sumR1  +=       0.227 * cos(1.69939961119  + 18984.29263000960 *T)
    sumR1  +=       0.217 * cos(4.71565260831  + 8671.96987044060 *T)
    sumR1  +=       0.199 * cos(0.66532566162  + 15664.03552270859 *T)
    sumR1  +=       0.201 * cos(3.90133282111  + 266.60704172180 *T)
    sumR1  +=       0.242 * cos(0.76562027191  + 3281.23856478620 *T)
    sumR1  +=       0.254 * cos(6.07712146530  + 14054.60730802600 *T)
    sumR1  +=       0.193 * cos(6.08798590746  + 8982.81066930900 *T)
    sumR1  +=       0.202 * cos(1.12506176909  + 6684.81528205140 *T)
    sumR1  +=       0.212 * cos(0.41763591084  + 5732.04924442980 *T)
    sumR1  +=       0.191 * cos(0.63315794517  + 3017.10701004240 *T)
    sumR1  +=       0.184 * cos(5.68898048651  + 5642.19824260920 *T)
    sumR1  +=       0.180 * cos(4.60366258431  + 10235.13637553740 *T)
    sumR1  +=       0.198 * cos(1.01742381306  + 3336.73109134180 *T)
    sumR1  +=       0.207 * cos(0.80702499658  + 3877.41693879520 *T)
    sumR1  +=       0.202 * cos(1.61753574776  + 11216.28429032400 *T)
    sumR1  +=       0.177 * cos(0.20305783491  + 22743.40937951640 *T)
    sumR1  +=       0.181 * cos(6.17864681592  + 4885.96640967860 *T)
    sumR1  +=       0.228 * cos(0.90779655425  + 18454.60166491500 *T)
    sumR1  +=       0.165 * cos(3.97299091423  + 20.35531939880 *T)
    sumR1  +=       0.206 * cos(3.58770221682  + 3205.54734666440 *T)
    sumR1  +=       0.185 * cos(2.77088436060  + 220.41264243880 *T)
    sumR1  +=       0.163 * cos(0.00569068383  + 1437.17561419860 *T)
    sumR1  +=       0.185 * cos(6.23128875160  + 6660.86953400080 *T)
    sumR1  +=       0.164 * cos(4.64548469502  + 3253.30422216000 *T)
    sumR1  +=       0.210 * cos(4.85590304986  + 5618.31980486140 *T)
    sumR1  +=       0.226 * cos(1.76563729092  + 5625.36604155940 *T)
    sumR1  +=       0.163 * cos(0.67931390118  + 12310.18132361080 *T)
    sumR1  +=       0.211 * cos(1.51278040364  + 4039.88357492740 *T)
    sumR1  +=       0.177 * cos(2.14779710666  + 3346.13535100720 *T)
    sumR1  +=       0.154 * cos(3.33525296567  + 9070.11887384880 *T)
    sumR1  +=       0.178 * cos(1.65162116131  + 10713.99488132620 *T)
    sumR1  +=       0.156 * cos(6.12783235584  + 103.09277421860 *T)
    sumR1  +=       0.155 * cos(3.51854579269  + 7373.38245462640 *T)
    sumR1  +=       0.157 * cos(4.91614155659  + 15508.61512327440 *T)
    sumR1  +=       0.191 * cos(1.74459210202  + 10022.81760116760 *T)
    sumR1  +=       0.178 * cos(3.53665589741  + 533.21408344360 *T)
    sumR1  +=       0.154 * cos(2.64640687885  + 3333.56619000180 *T)
    sumR1  +=       0.162 * cos(1.27235392680  + 5415.65737477320 *T)
    sumR1  +=       0.144 * cos(1.30886254982  + 6460.81221096080 *T)
    sumR1  +=       0.173 * cos(3.76291627342  + 3561.02506913860 *T)
    sumR1  +=       0.144 * cos(0.86629692094  + 7380.49600162720 *T)
    sumR1  +=       0.150 * cos(0.23529641094  + 1228.96211332220 *T)
    sumR1  +=       0.144 * cos(0.12285484149  + 647.01083331480 *T)
    sumR1  +=       0.160 * cos(0.24157324936  + 11614.43329373220 *T)
    sumR1  +=       0.146 * cos(4.76067886088  + 16706.58525184800 *T)
    sumR1  +=       0.188 * cos(2.96688973732  + 2178.13772229200 *T)
    sumR1  +=       0.175 * cos(1.66915420767  + 22345.26037610820 *T)
    sumR1  +=       0.141 * cos(1.75571255457  + 3657.00429635640 *T)
    sumR1  +=       0.145 * cos(0.71777671857  + 6784.31762761820 *T)
    sumR1  +=       0.138 * cos(2.74151377335  + 6546.15977336420 *T)
    sumR1  +=       0.139 * cos(4.12056692213  + 3351.24909204960 *T)
    sumR1  +=       0.152 * cos(4.88528890547  + 110.20632121940 *T)
    sumR1  +=       0.134 * cos(0.86682710306  + 3603.69635007260 *T)
    sumR1  +=       0.170 * cos(5.54401695608  + 3265.83082813250 *T)
    sumR1  +=       0.165 * cos(2.74557093704  + 4672.66731424060 *T)
    sumR1  +=       0.129 * cos(0.94520493983  + 419.48464387520 *T)
    sumR1  +=       0.167 * cos(3.43007767272  + 1581.95934828300 *T)
    sumR1  +=       0.136 * cos(3.68877945956  + 13171.00144068760 *T)
    sumR1  +=       0.160 * cos(0.68321924763  + 6475.03930496240 *T)
    sumR1  +=       0.129 * cos(4.46548223743  + 9499.25986200560 *T)
    sumR1  +=       0.144 * cos(1.88085102951  + 9468.26787725700 *T)
    sumR1  +=       0.149 * cos(3.49118810274  + 14158.74771361560 *T)
    sumR1  +=       0.138 * cos(1.92770959208  + 15.25247118500 *T)
    sumR1  +=       0.172 * cos(0.23189971635  + 12323.42309600880 *T)
    sumR1  +=       0.121 * cos(1.75813513715  + 4379.63903749020 *T)
    sumR1  +=       0.120 * cos(6.04299490267  + 6606.44325483230 *T)
    sumR1  +=       0.134 * cos(1.98845453946  + 8965.97846825920 *T)
    sumR1  +=       0.118 * cos(0.22641710036  + 11766.26326451460 *T)
    sumR1  +=       0.122 * cos(2.92219895926  + 21947.11137270000 *T)
    sumR1  +=       0.147 * cos(5.86576524955  + 3169.93955608060 *T)
    sumR1  +=       0.121 * cos(5.75886629308  + 19513.98359510420 *T)
    sumR1  +=       0.119 * cos(3.93419985914  + 4936.79886898440 *T)
    sumR1  +=       0.123 * cos(4.65853055862  + 13149.15061136120 *T)
    sumR1  +=       0.151 * cos(4.38870879458  + 13362.51701710200 *T)
    sumR1  +=       0.134 * cos(4.49208622602  + 2480.30249794700 *T)
    sumR1  +=       0.115 * cos(5.78268772131  + 12722.55242048520 *T)
    sumR1  +=       0.126 * cos(4.43971496486  + 3568.08855948880 *T)
    sumR1  +=       0.114 * cos(0.36704570747  + 3384.33133900480 *T)
    sumR1  +=       0.117 * cos(0.67607782416  + 13207.02930736500 *T)
    sumR1  +=       0.124 * cos(5.80386208894  + 10721.10842832700 *T)
    sumR1  +=       0.115 * cos(0.23672592056  + 224.34479570190 *T)
    sumR1  +=       0.110 * cos(0.26402316132  + 3077.52850332700 *T)
    sumR1  +=       0.112 * cos(3.50979462224  + 1375.77379984580 *T)
    sumR1  +=       0.108 * cos(5.30086016687  + 10264.56588407340 *T)
    sumR1  +=       0.146 * cos(2.36972522432  + 4989.05918389720 *T)
    sumR1  +=       0.143 * cos(0.92798026058  + 5401.43028077160 *T)
    sumR1  +=       0.111 * cos(2.47771615302  + 18606.49894600020 *T)
    sumR1  +=       0.124 * cos(1.13870545502  + 2807.39834325620 *T)
    sumR1  +=       0.105 * cos(5.87122793977  + 4193.80880845180 *T)
    sumR1  +=       0.115 * cos(3.01754571849  + 1118.75579210280 *T)
    sumR1  +=       0.107 * cos(2.03274473347  + 25685.87280280800 *T)
    sumR1  +=       0.107 * cos(4.82345245872  + 13362.38239649640 *T)
    sumR1  +=       0.105 * cos(1.21018340855  + 8584.66166590080 *T)
    sumR1  +=       0.104 * cos(5.07527834174  + 23141.55838292460 *T)
    sumR1  +=       0.100 * cos(4.08698733114  + 3511.28529731900 *T)
    sumR1  +=       0.105 * cos(1.13833234280  + 1162.47470440780 *T)
    sumR1  +=       0.102 * cos(2.55243872323  + 664.75604513000 *T)
    sumR1  +=       0.138 * cos(3.04361079523  + 3329.97576135000 *T)
    sumR1  +=       0.130 * cos(3.04769575195  + 3341.03250279340 *T)
    sumR1  +=       0.117 * cos(2.97177407391  + 9602.35263622420 *T)
    sumR1  +=       0.098 * cos(4.54292546461  + 3362.46325602620 *T)
    sumR1  +=       0.097 * cos(2.21618504638  + 10042.61267559180 *T)
    sumR1  +=       0.100 * cos(5.45541648580  + 3024.22055704320 *T)
    sumR1  +=       0.097 * cos(1.55448890675  + 8799.98871377800 *T)
    sumR1  +=       0.107 * cos(5.15001764390  + 956.28915597060 *T)
    sumR1  +=       0.100 * cos(1.80839732601  + 6665.97238221460 *T)
    sumR1  +=       0.104 * cos(0.04641040148  + 10001.06188460700 *T)
    sumR1  +=       0.097 * cos(0.30502193916  + 10014.72373309860 *T)
    sumR1  +=       0.125 * cos(3.71144870102  + 3873.82651014340 *T)
    sumR1  +=       0.102 * cos(5.38933965862  + 13892.14067189380 *T)
    sumR1  +=       0.108 * cos(4.33438895790  + 3490.17562383440 *T)
    sumR1  +=       0.095 * cos(2.32591017464  + 10028.95082710020 *T)
    sumR1  +=       0.113 * cos(1.74323945160  + 11610.91017538320 *T)
    sumR1  +=       0.119 * cos(4.94578003013  + 3443.70520091840 *T)
    sumR1  +=       0.096 * cos(1.63239207663  + 11925.27409260060 *T)
    sumR1  +=       0.124 * cos(0.82697196609  + 9638.94074787620 *T)
    sumR1  +=       0.101 * cos(1.51065535085  + 1062.90504853820 *T)
    sumR1  +=       0.107 * cos(4.36653716080  + 6531.66165626500 *T)
    sumR1  +=       0.090 * cos(2.01534774170  + 22324.90505670940 *T)
    sumR1  +=       0.094 * cos(4.85202869277  + 14047.49376102520 *T)
    sumR1  +=       0.089 * cos(1.23714034226  + 20809.46762464520 *T)
    sumR1  +=       0.095 * cos(5.69916785631  + 6717.25272007700 *T)
    sumR1  +=       0.095 * cos(2.15038299943  + 8013.27974094040 *T)
    sumR1  +=       0.111 * cos(3.91313035726  + 3311.18291816379 *T)
    sumR1  +=       0.089 * cos(1.37925538521  + 74.78159856730 *T)
    sumR1  +=       0.094 * cos(3.24872875047  + 6657.34641565180 *T)
    sumR1  +=       0.089 * cos(0.70324117212  + 23.87843774780 *T)
    sumR1  +=       0.087 * cos(0.40376031248  + 2221.85663459700 *T)
    sumR1  +=       0.092 * cos(3.33165786190  + 2540.79130153440 *T)
    sumR1  +=       0.120 * cos(3.09344469205  + 6685.10618875760 *T)
    sumR1  +=       0.087 * cos(4.92694309070  + 1505.28780909299 *T)
    sumR1  +=       0.091 * cos(1.66121998822  + 24150.08005134500 *T)
    sumR1  +=       0.090 * cos(5.47687082598  + 16699.53901514999 *T)
    sumR1  +=       0.119 * cos(2.53709230647  + 3312.16323923200 *T)
    sumR1  +=       0.118 * cos(5.56523997231  + 2814.44457995420 *T)
    sumR1  +=       0.093 * cos(1.61394240489  + 3341.04230982650 *T)
    sumR1  +=       0.084 * cos(1.19439840758  + 3450.81874791920 *T)
    sumR1  +=       0.085 * cos(4.95060410388  + 7314.00859271280 *T)
    sumR1  +=       0.084 * cos(3.92244493446  + 170.67287061920 *T)
    sumR1  +=       0.096 * cos(5.49205832489  + 3427.92063123960 *T)
    sumR1  +=       0.106 * cos(1.61703818023  + 3774.32416457660 *T)
    sumR1  +=       0.101 * cos(6.18535119914  + 5511.63660199100 *T)
    sumR1  +=       0.081 * cos(3.16562241132  + 52175.80628314840 *T)
    sumR1  +=       0.093 * cos(5.15512549422  + 3451.79906898740 *T)
    sumR1  +=       0.108 * cos(0.33314004399  + 6709.67404086740 *T)
    sumR1  +=       0.078 * cos(2.38080208142  + 13553.89797291080 *T)
    sumR1  +=       0.085 * cos(1.18588430640  + 6705.10329114740 *T)
    sumR1  +=       0.098 * cos(3.24856024758  + 25287.72379939980 *T)
    sumR1  +=       0.080 * cos(2.54327778068  + 6675.70192909220 *T)
    sumR1  +=       0.078 * cos(4.70533248161  + 24889.57479599160 *T)
    sumR1  +=       0.095 * cos(1.59916172961  + 4271.97551355160 *T)
    sumR1  +=       0.076 * cos(0.90799171644  + 2675.85638156980 *T)
    sumR1  +=       0.091 * cos(3.80728044706  + 2125.87740737920 *T)
    sumR1  +=       0.093 * cos(5.59480359619  + 3340.18254357310 *T)
    sumR1  +=       0.095 * cos(2.06299921822  + 6518.75821726740 *T)
    sumR1  +=       0.100 * cos(6.02550501045  + 3313.21087060300 *T)
    sumR1  +=       0.075 * cos(6.16095890090  + 3370.04193523580 *T)
    sumR1  +=       0.077 * cos(4.92444879515  + 3229.42578441220 *T)
    sumR1  +=       0.090 * cos(4.55063359845  + 6701.58017279840 *T)
    sumR1  +=       0.078 * cos(2.01202524407  + 14421.83163698840 *T)
    sumR1  +=       0.079 * cos(5.92142974185  + 6155.05700665400 *T)
    sumR1  +=       0.101 * cos(2.07923317336  + 15265.88651930040 *T)
    sumR1  +=       0.082 * cos(6.04078010866  + 3760.09707057500 *T)
    sumR1  +=       0.073 * cos(3.01106366205  + 206.18554843720 *T)
    sumR1  +=       0.078 * cos(1.71095661489  + 3399.98628861340 *T)
    sumR1  +=       0.073 * cos(3.32824316241  + 9588.12554222260 *T)
    sumR1  +=       0.076 * cos(1.74030672711  + 6756.00645196690 *T)
    sumR1  +=       0.073 * cos(1.44980946324  + 5408.54382777240 *T)
    sumR1  +=       0.077 * cos(5.64501410813  + 3209.07046501340 *T)
    sumR1  +=       0.080 * cos(5.70272305344  + 10156.90236013480 *T)
    sumR1  +=       0.075 * cos(3.83222347813  + 39048.32250960720 *T)
    sumR1  +=       0.072 * cos(0.82629826575  + 7366.26890762560 *T)
    sumR1  +=       0.083 * cos(4.71804515025  + 14061.72085502680 *T)
    sumR1  +=       0.072 * cos(4.35767091893  + 21548.96236929180 *T)
    sumR1  +=       0.086 * cos(1.50107043029  + 5989.06725217280 *T)
    sumR1  +=       0.073 * cos(3.77377582806  + 28628.33622609960 *T)
    sumR1  +=       0.082 * cos(0.78740181090  + 16276.46394262300 *T)
    sumR1  +=       0.070 * cos(2.69225539304  + 3296.89351439480 *T)
    sumR1  +=       0.069 * cos(1.01941228196  + 12295.95422960920 *T)
    sumR1  +=       0.081 * cos(4.64227979982  + 6947.83189512140 *T)
    sumR1  +=       0.094 * cos(0.74502278458  + 5244.04923920100 *T)
    sumR1  +=       0.067 * cos(4.47328871650  + 20206.14119633100 *T)
    sumR1  +=       0.091 * cos(2.95968633697  + 3326.38533269820 *T)
    sumR1  +=       0.076 * cos(2.86128350856  + 8542.97070603500 *T)
    sumR1  +=       0.065 * cos(1.31079130285  + 23546.75362303080 *T)
    sumR1  +=       0.065 * cos(3.31422850411  + 6414.61781167780 *T)
    sumR1  +=       0.064 * cos(3.98415414793  + 3368.01398279660 *T)
    sumR1  +=       0.065 * cos(4.15810281695  + 3340.19235060619 *T)
    sumR1  +=       0.071 * cos(3.82044323862  + 6578.13207918100 *T)
    sumR1  +=       0.080 * cos(1.02128994578  + 7321.12213971360 *T)
    sumR1  +=       0.064 * cos(1.07965886113  + 6944.30877677240 *T)
    sumR1  +=       0.064 * cos(1.20215520018  + 38.13303563780 *T)
    sumR1  +=       0.070 * cos(5.11166911478  + 9485.03276800400 *T)
    sumR1  +=       0.067 * cos(2.22216014849  + 4845.90023579280 *T)
    sumR1  +=       0.083 * cos(1.86067842148  + 931.36308685180 *T)
    sumR1  +=       0.063 * cos(0.45029141698  + 6418.14093002680 *T)
    sumR1  +=       0.062 * cos(5.68207205878  + 4459.36821880260 *T)
    sumR1  +=       0.062 * cos(0.62454636835  + 6688.27109009760 *T)
    sumR1  +=       0.076 * cos(0.41820625869  + 5032.77809620220 *T)
    sumR1  +=       0.062 * cos(5.72811519809  + 2008.55753915900 *T)
    sumR1  +=       0.060 * cos(1.98790201827  + 12199.97500239140 *T)
    sumR1  +=       0.078 * cos(2.18035217615  + 3232.94890276120 *T)
    sumR1  +=       0.060 * cos(1.08971396071  + 10124.93005431800 *T)
    sumR1  +=       0.072 * cos(4.08555524166  + 3416.87849797540 *T)
    sumR1  +=       0.061 * cos(2.49066420507  + 4005.36847182980 *T)
    sumR1  +=       0.059 * cos(2.93755144370  + 9886.77220006400 *T)
    sumR1  +=       0.069 * cos(1.02371083242  + 6621.85099148600 *T)
    sumR1  +=       0.061 * cos(4.57446273568  + 2277.70737816160 *T)
    sumR1  +=       0.059 * cos(4.74745481720  + 16865.52876963120 *T)
    sumR1  +=       0.069 * cos(3.62576981720  + 15906.76412668260 *T)
    sumR1  +=       0.058 * cos(4.15993379336  + 1854.63230563460 *T)
    sumR1  +=       0.060 * cos(4.92798256747  + 3391.89276456221 *T)
    sumR1  +=       0.070 * cos(4.51699340219  + 16304.91313009080 *T)
    sumR1  +=       0.056 * cos(3.59354639280  + 7322.10246078180 *T)
    sumR1  +=       0.068 * cos(0.81217091970  + 16703.04487984680 *T)
    sumR1  +=       0.056 * cos(3.19997567622  + 4140.43355186520 *T)
    sumR1  +=       0.058 * cos(1.73479637157  + 2409.24933984800 *T)
    sumR1  +=       0.054 * cos(5.54310937490  + 2604.73591316800 *T)
    sumR1  +=       0.059 * cos(2.02922256380  + 45486.81875903279 *T)
    sumR1  +=       0.063 * cos(5.29505507594  + 17101.21113690720 *T)
    sumR1  +=       0.054 * cos(4.88230660750  + 48300.16010269000 *T)
    sumR1  +=       0.071 * cos(4.84381041200  + 3289.33208883739 *T)
    sumR1  +=       0.054 * cos(1.76917054114  + 3510.19260983280 *T)
    sumR1  +=       0.063 * cos(2.80175814276  + 16894.51039961060 *T)
    sumR1  +=       0.058 * cos(0.91631357427  + 10018.24685144760 *T)
    sumR1  +=       0.054 * cos(1.06826755543  + 8948.22825590860 *T)
    sumR1  +=       0.072 * cos(1.36121375974  + 10025.42770875120 *T)
    sumR1  +=       0.061 * cos(0.92427393171  + 17468.85519794540 *T)
    sumR1  +=       0.068 * cos(2.84271358352  + 16703.07938715119 *T)
    sumR1  +=       0.053 * cos(3.42679182927  + 14867.73751589219 *T)
    sumR1  +=       0.054 * cos(0.41070672545  + 18451.07854656599 *T)
    sumR1  +=       0.060 * cos(5.86922287544  + 7447.01791784600 *T)
    sumR1  +=       0.051 * cos(1.64684042106  + 3472.15438838620 *T)
    sumR1  +=       0.057 * cos(6.13250399703  + 17085.95866572220 *T)
    sumR1  +=       0.053 * cos(4.16969434701  + 6106.88005506480 *T)
    sumR1  +=       0.050 * cos(3.17196717347  + 2111.65031337760 *T)
    sumR1  +=       0.055 * cos(5.72607545560  + 10706.88133432540 *T)
    sumR1  +=       0.062 * cos(0.57333106196  + 3171.03224356680 *T)
    sumR1  +=       0.053 * cos(1.29007833025  + 6652.77566593180 *T)
    sumR1  +=       0.050 * cos(2.96441858632  + 10037.08975128440 *T)
    sumR1  +=       0.065 * cos(5.11165617529  + 3113.13629391080 *T)
    sumR1  +=       0.053 * cos(4.87447079015  + 1485.98012106520 *T)
    sumR1  +=       0.050 * cos(2.52848298042  + 10448.43547097540 *T)
    sumR1  +=       0.059 * cos(1.69623653028  + 4025.65648092580 *T)
    sumR1  +=       0.047 * cos(5.15559748285  + 12012.58229714040 *T)
    sumR1  +=       0.049 * cos(1.19473362353  + 3448.27595063840 *T)
    sumR1  +=       0.063 * cos(1.01191427607  + 11769.85369316640 *T)
    sumR1  +=       0.064 * cos(3.86219786087  + 18208.34994259200 *T)
    sumR1  +=       0.059 * cos(2.96261275062  + 14071.43950907579 *T)
    sumR1  +=       0.049 * cos(1.41657179779  + 2938.94030494260 *T)
    sumR1  +=       0.048 * cos(2.95878690387  + 11527.12508919240 *T)
    sumR1  +=       0.052 * cos(0.01971915447  + 8226.57883637840 *T)
    sumR1  +=       0.045 * cos(5.07966377852  + 3318.76159737340 *T)
    sumR1  +=       0.043 * cos(1.23879381294  + 7218.02936549500 *T)
    sumR1  +=       0.058 * cos(5.58121433163  + 6643.09181776180 *T)
    sumR1  +=       0.048 * cos(5.02446939402  + 6645.19698672220 *T)
    sumR1  +=       0.043 * cos(0.69492704598  + 20995.39296644940 *T)
    sumR1  +=       0.044 * cos(4.02272101657  + 9389.05354078620 *T)
    sumR1  +=       0.055 * cos(4.38138154697  + 1478.86657406440 *T)
    sumR1  +=       0.051 * cos(4.24292455428  + 792.77488846740 *T)
    sumR1  +=       0.042 * cos(2.74826708762  + 14577.18472611980 *T)
    sumR1  +=       0.044 * cos(4.18397905503  + 8535.85715903420 *T)
    sumR1  +=       0.047 * cos(1.33588473182  + 632.78373931320 *T)
    sumR1  +=       0.042 * cos(5.05676915852  + 3397.51080163540 *T)
    sumR1  +=       0.042 * cos(0.28204510006  + 10001.48196070061 *T)
    sumR1  +=       0.042 * cos(0.75310918544  + 6357.71943674220 *T)
    sumR1  +=       0.042 * cos(4.94532732982  + 18052.92954315780 *T)
    sumR1  +=       0.052 * cos(4.09912687749  + 5835.14201864840 *T)
    sumR1  +=       0.054 * cos(2.46533302314  + 8186.51266249260 *T)
    sumR1  +=       0.043 * cos(4.77713978044  + 32124.36905223359 *T)
    sumR1  +=       0.053 * cos(6.08293348275  + 3377.21779200400 *T)
    sumR1  +=       0.041 * cos(2.51168269556  + 4186.69526145100 *T)
    sumR1  +=       0.041 * cos(1.24482327948  + 3212.59358336240 *T)
    sumR1  +=       0.041 * cos(5.42003026893  + 685.04405422600 *T)
    sumR1  +=       0.041 * cos(5.04768364997  + 6571.01853218020 *T)
    sumR1  +=       0.042 * cos(2.08904552145  + 13363.43002786740 *T)
    sumR1  +=       0.048 * cos(4.82888746454  + 1835.32461760680 *T)
    sumR1  +=       0.043 * cos(4.17203713456  + 43340.65334255760 *T)
    sumR1  +=       0.046 * cos(0.81640935106  + 45884.96776244100 *T)
    sumR1  +=       0.042 * cos(2.26773068307  + 12729.66596748600 *T)
    sumR1  +=       0.039 * cos(5.85791936573  + 846.08283475120 *T)
    sumR1  +=       0.043 * cos(2.90976420757  + 6872.67311951120 *T)
    sumR1  +=       0.045 * cos(1.98725045807  + 1861.74585263540 *T)
    sumR1  +=       0.045 * cos(0.50053853542  + 14128.24277124560 *T)
    sumR1  +=       0.046 * cos(2.86512929328  + 38650.17350619900 *T)
    sumR1  +=       0.038 * cos(3.65846461938  + 29698.28751133580 *T)
    sumR1  +=       0.039 * cos(4.57679716458  + 6901.63749583840 *T)
    sumR1  +=       0.039 * cos(3.85504465583  + 9945.57120882380 *T)
    sumR1  +=       0.040 * cos(0.06127203284  + 9947.05568153210 *T)
    sumR1  +=       0.043 * cos(5.28854105201  + 3274.12501778540 *T)
    sumR1  +=       0.047 * cos(6.25707790441  + 24606.13555322000 *T)
    sumR1  +=       0.037 * cos(5.02115296017  + 11128.97608578420 *T)
    sumR1  +=       0.039 * cos(1.71421919870  + 7696.88787128380 *T)
    sumR1  +=       0.037 * cos(4.34652985120  + 3283.71405176420 *T)
    sumR1  +=       0.037 * cos(0.05572748092  + 21150.81336588360 *T)
    sumR1  +=       0.042 * cos(4.97872041460  + 13575.74880223720 *T)
    sumR1  +=       0.050 * cos(4.24170332288  + 7747.72033058960 *T)
    sumR1  +=       0.037 * cos(4.07496312186  + 8646.06348025360 *T)
    sumR1  +=       0.038 * cos(0.44080908793  + 24491.42579258340 *T)
    sumR1  +=       0.036 * cos(1.73681874925  + 3468.63127003720 *T)
    sumR1  +=       0.041 * cos(5.69294900686  + 26087.90314157420 *T)
    sumR1  +=       0.036 * cos(1.80256389689  + 8756.26980147300 *T)
    sumR1  +=       0.036 * cos(3.37374689465  + 48429.28218232440 *T)
    sumR1  +=       0.035 * cos(0.10555289345  + 8742.04270747140 *T)
    sumR1  +=       0.041 * cos(4.26832466355  + 21000.91589075680 *T)
    sumR1  +=       0.038 * cos(0.73199792046  + 26084.02180621620 *T)

    # MARS Radius
    sumR2  = 0
    sumR2  +=   44242.247 * cos(0.47930603943  + 3340.61242669980 *T)
    sumR2  +=    8138.042 * cos(0.86998398093  + 6681.22485339960 *T)
    sumR2  +=    1274.915 * cos(1.22594050809  + 10021.83728009940 *T)
    sumR2  +=     187.387 * cos(1.57298991982  + 13362.44970679920 *T)
    sumR2  +=      40.744 * cos(1.97080175060  + 3344.13554504880 *T)
    sumR2  +=      52.396 * cos(3.14159265359  + 0.00000000000 *T)
    sumR2  +=      26.616 * cos(1.91665615762  + 16703.06213349900 *T)
    sumR2  +=      17.825 * cos(4.43499505333  + 2281.23049651060 *T)
    sumR2  +=      11.713 * cos(4.52510453730  + 3185.19202726560 *T)
    sumR2  +=      10.209 * cos(5.39143469548  + 1059.38193018920 *T)
    sumR2  +=       9.950 * cos(0.41870577185  + 796.29800681640 *T)
    sumR2  +=       9.237 * cos(4.53579272961  + 2146.16541647520 *T)
    sumR2  +=       7.299 * cos(3.14218509183  + 2544.31441988340 *T)
    sumR2  +=       7.217 * cos(2.29300859074  + 6684.74797174860 *T)
    sumR2  +=       6.808 * cos(5.26702580055  + 155.42039943420 *T)
    sumR2  +=       6.528 * cos(2.30781369329  + 3738.76143010800 *T)
    sumR2  +=       7.785 * cos(5.93369079547  + 1748.01641306700 *T)
    sumR2  +=       5.840 * cos(1.05191350362  + 1349.86740965880 *T)
    sumR2  +=       6.749 * cos(5.30194395749  + 1194.44701022460 *T)
    sumR2  +=       4.695 * cos(0.76880586144  + 3097.88382272579 *T)
    sumR2  +=       5.391 * cos(1.00223256041  + 3149.16416058820 *T)
    sumR2  +=       4.406 * cos(2.45556303355  + 951.71840625060 *T)
    sumR2  +=       4.286 * cos(3.89643520638  + 1592.59601363280 *T)
    sumR2  +=       3.514 * cos(1.85168391963  + 398.14900340820 *T)
    sumR2  +=       3.699 * cos(2.26043707772  + 20043.67456019880 *T)
    sumR2  +=       3.377 * cos(3.81683532672  + 1751.53953141600 *T)
    sumR2  +=       4.585 * cos(0.80787441740  + 4136.91043351620 *T)
    sumR2  +=       3.201 * cos(2.11657635165  + 5614.72937620960 *T)
    sumR2  +=       3.622 * cos(1.32395191387  + 3333.49887969900 *T)
    sumR2  +=       2.916 * cos(1.19337460559  + 529.69096509460 *T)
    sumR2  +=       2.979 * cos(2.86481008776  + 6151.53388830500 *T)
    sumR2  +=       3.057 * cos(4.55276793064  + 5628.95647021120 *T)
    sumR2  +=       2.906 * cos(1.20295377623  + 3894.18182954220 *T)
    sumR2  +=       3.850 * cos(3.86055626689  + 553.56940284240 *T)
    sumR2  +=       2.820 * cos(2.48683324916  + 1990.74501704100 *T)
    sumR2  +=       2.657 * cos(6.07411629964  + 4292.33083295040 *T)
    sumR2  +=       2.700 * cos(2.92139773350  + 3496.03282613400 *T)
    sumR2  +=       2.395 * cos(5.94175921617  + 2787.04302385740 *T)
    sumR2  +=       2.264 * cos(2.56219866409  + 191.44826611160 *T)
    sumR2  +=       2.167 * cos(5.36812435483  + 8962.45534991020 *T)
    sumR2  +=       2.149 * cos(2.74950075397  + 242.72860397400 *T)
    sumR2  +=       2.217 * cos(1.85265984462  + 3337.08930835080 *T)
    sumR2  +=       1.996 * cos(5.76429928131  + 3341.59274776800 *T)
    sumR2  +=       1.999 * cos(3.82349238481  + 2914.01423582380 *T)
    sumR2  +=       1.835 * cos(5.68592723044  + 1589.07289528380 *T)
    sumR2  +=       1.812 * cos(3.32042068028  + 5088.62883976680 *T)
    sumR2  +=       2.413 * cos(4.68291336853  + 4690.47983635860 *T)
    sumR2  +=       1.970 * cos(4.17480610904  + 3340.59517304760 *T)
    sumR2  +=       1.970 * cos(6.20643855008  + 3340.62968035200 *T)
    sumR2  +=       1.627 * cos(5.67733051452  + 4535.05943692440 *T)
    sumR2  +=       2.160 * cos(1.07452599834  + 2388.89402044920 *T)
    sumR2  +=       1.964 * cos(3.10805316088  + 3583.34103067380 *T)
    sumR2  +=       1.985 * cos(5.75850351840  + 4399.99435688900 *T)
    sumR2  +=       1.507 * cos(4.95936409838  + 382.89653222320 *T)
    sumR2  +=       1.278 * cos(4.82232889938  + 2957.71589447660 *T)
    sumR2  +=       1.475 * cos(2.22707926559  + 3723.50895892300 *T)
    sumR2  +=       1.196 * cos(3.26724458920  + 9492.14631500480 *T)
    sumR2  +=       1.349 * cos(4.87573224485  + 6525.80445396540 *T)
    sumR2  +=       1.433 * cos(2.69734916443  + 7079.37385680780 *T)
    sumR2  +=       1.224 * cos(2.62012336714  + 10025.36039844840 *T)
    sumR2  +=       1.404 * cos(5.19056026479  + 2700.71514038580 *T)
    sumR2  +=       1.202 * cos(0.93472783088  + 2810.92146160520 *T)
    sumR2  +=       0.869 * cos(5.81340811635  + 12303.06777661000 *T)
    sumR2  +=       0.867 * cos(2.20046640409  + 2699.73481931760 *T)
    sumR2  +=       0.830 * cos(2.01484544773  + 5092.15195811580 *T)
    sumR2  +=       0.855 * cos(5.96220147975  + 426.59819087600 *T)
    sumR2  +=       0.848 * cos(2.26407047301  + 6283.07584999140 *T)
    sumR2  +=       0.917 * cos(1.40295785881  + 6489.77658728800 *T)
    sumR2  +=       0.833 * cos(1.17384197174  + 7477.52286021600 *T)
    sumR2  +=       1.041 * cos(6.27171470048  + 3347.72597370060 *T)
    sumR2  +=       0.965 * cos(3.39855816541  + 5621.84292321040 *T)
    sumR2  +=       0.722 * cos(4.26304776331  + 4933.20844033260 *T)
    sumR2  +=       0.706 * cos(2.34131594714  + 7.11354700080 *T)
    sumR2  +=       0.768 * cos(2.06208352904  + 5486.77784317500 *T)
    sumR2  +=       0.953 * cos(2.11123497948  + 3870.30339179440 *T)
    sumR2  +=       0.844 * cos(2.23931963240  + 3553.91152213780 *T)
    sumR2  +=       0.646 * cos(2.24669034469  + 3340.54511639700 *T)
    sumR2  +=       0.653 * cos(3.99043329363  + 6677.70173505060 *T)
    sumR2  +=       0.714 * cos(0.29739480601  + 6681.24210705180 *T)
    sumR2  +=       0.828 * cos(0.22863617670  + 3532.06069281140 *T)
    sumR2  +=       0.612 * cos(1.55388376751  + 7234.79425624200 *T)
    sumR2  +=       0.714 * cos(4.54969883976  + 6681.20759974740 *T)
    sumR2  +=       0.586 * cos(3.30118433303  + 1221.84856632140 *T)
    sumR2  +=       0.646 * cos(1.83853693231  + 3340.67973700260 *T)
    sumR2  +=       0.560 * cos(5.05848353328  + 8031.09226305840 *T)
    sumR2  +=       0.651 * cos(0.15897472160  + 7632.94325965020 *T)
    sumR2  +=       0.488 * cos(3.08086378649  + 6836.64525283380 *T)
    sumR2  +=       0.534 * cos(4.25727954634  + 3339.63210563160 *T)
    sumR2  +=       0.504 * cos(2.59990772815  + 23384.28698689860 *T)
    sumR2  +=       0.372 * cos(6.27186983164  + 2487.41604494780 *T)
    sumR2  +=       0.343 * cos(4.71059403854  + 639.89728631400 *T)
    sumR2  +=       0.449 * cos(1.34123846614  + 640.87760738220 *T)
    sumR2  +=       0.342 * cos(5.96338023755  + 4929.68532198360 *T)
    sumR2  +=       0.329 * cos(2.65955258770  + 9623.68827669120 *T)
    sumR2  +=       0.352 * cos(6.09567212646  + 7740.60678358880 *T)
    sumR2  +=       0.326 * cos(6.05498346093  + 2118.76386037840 *T)
    sumR2  +=       0.403 * cos(4.96465831103  + 8969.56889691100 *T)
    sumR2  +=       0.402 * cos(2.84624835384  + 5331.35744374080 *T)
    sumR2  +=       0.319 * cos(1.44545457475  + 5729.50644714900 *T)
    sumR2  +=       0.303 * cos(0.01918776551  + 6127.65545055720 *T)
    sumR2  +=       0.297 * cos(5.74070606294  + 6041.32756708560 *T)
    sumR2  +=       0.297 * cos(3.06763224975  + 10419.98628350760 *T)
    sumR2  +=       0.282 * cos(6.04069352017  + 7875.67186362420 *T)
    sumR2  +=       0.278 * cos(5.11654962157  + 11773.37681151540 *T)
    sumR2  +=       0.259 * cos(2.89786343627  + 7210.91581849420 *T)
    sumR2  +=       0.273 * cos(3.83181794308  + 8429.24126646660 *T)
    sumR2  +=       0.256 * cos(1.08862226474  + 2288.34404351140 *T)
    sumR2  +=       0.288 * cos(6.13572687680  + 6682.20517446780 *T)
    sumR2  +=       0.302 * cos(3.61490647715  + 12832.75874170460 *T)
    sumR2  +=       0.235 * cos(2.21413387615  + 6681.29216370240 *T)
    sumR2  +=       0.317 * cos(3.60634897209  + 2301.58581590939 *T)
    sumR2  +=       0.289 * cos(5.08223311018  + 2274.11694950980 *T)
    sumR2  +=       0.254 * cos(0.36603904729  + 3767.21061757580 *T)
    sumR2  +=       0.235 * cos(2.62434481550  + 6681.15754309680 *T)
    sumR2  +=       0.226 * cos(3.45305949713  + 6923.95345737360 *T)
    sumR2  +=       0.205 * cos(3.88943452829  + 10018.31416175040 *T)
    sumR2  +=       0.206 * cos(3.44453113177  + 10177.25767953360 *T)
    sumR2  +=       0.210 * cos(0.57926734389  + 10973.55568635000 *T)
    sumR2  +=       0.194 * cos(1.57870686859  + 10818.13528691580 *T)
    sumR2  +=       0.190 * cos(2.42312911773  + 8955.34180290940 *T)
    sumR2  +=       0.194 * cos(2.45887376159  + 7064.12138562280 *T)
    sumR2  +=       0.193 * cos(0.58793888724  + 5828.02847164760 *T)
    sumR2  +=       0.183 * cos(1.68247271963  + 213.29909543800 *T)
    sumR2  +=       0.236 * cos(1.98863774162  + 10575.40668294180 *T)
    sumR2  +=       0.245 * cos(2.17761779228  + 4562.46099302120 *T)
    sumR2  +=       0.177 * cos(0.66379705293  + 10021.85453375160 *T)
    sumR2  +=       0.176 * cos(4.91607557961  + 10021.82002644720 *T)
    sumR2  +=       0.217 * cos(2.93968443649  + 13365.97282514820 *T)
    sumR2  +=       0.167 * cos(4.74764434327  + 3914.95722503460 *T)
    sumR2  +=       0.213 * cos(0.15734917857  + 2942.46342329160 *T)
    sumR2  +=       0.155 * cos(1.78776991092  + 9830.38901398780 *T)
    sumR2  +=       0.180 * cos(5.21088564787  + 9866.41688066520 *T)
    sumR2  +=       0.161 * cos(5.17683861380  + 6298.32832117640 *T)
    sumR2  +=       0.159 * cos(2.68980684850  + 6894.52394883760 *T)
    sumR2  +=       0.149 * cos(6.26524493469  + 1039.02661079040 *T)
    sumR2  +=       0.147 * cos(0.36939961702  + 6688.33840040040 *T)
    sumR2  +=       0.151 * cos(1.22272892290  + 574.34479833480 *T)
    sumR2  +=       0.137 * cos(2.09492296067  + 1066.49547719000 *T)
    sumR2  +=       0.165 * cos(6.15952050903  + 15643.68020330980 *T)
    sumR2  +=       0.163 * cos(4.66507592292  + 8273.82086703240 *T)
    sumR2  +=       0.124 * cos(0.67855210339  + 6438.49624942560 *T)
    sumR2  +=       0.128 * cos(1.69206367749  + 3981.49003408200 *T)
    sumR2  +=       0.116 * cos(2.78191248964  + 3360.96774609859 *T)
    sumR2  +=       0.128 * cos(0.84461423001  + 3355.86489788480 *T)
    sumR2  +=       0.124 * cos(1.63253732557  + 6674.11130639880 *T)
    sumR2  +=       0.120 * cos(0.31762150109  + 3127.31333126180 *T)
    sumR2  +=       0.129 * cos(2.26678583645  + 6467.92575796160 *T)
    sumR2  +=       0.104 * cos(0.07355376701  + 3325.35995551480 *T)
    sumR2  +=       0.130 * cos(4.41777190830  + 3320.25710730100 *T)
    sumR2  +=       0.095 * cos(1.31957248342  + 7903.07341972100 *T)
    sumR2  +=       0.114 * cos(0.65681602343  + 3178.14579056760 *T)
    sumR2  +=       0.092 * cos(4.98392367746  + 9779.10867612540 *T)
    sumR2  +=       0.101 * cos(4.10670787374  + 536.80451209540 *T)
    sumR2  +=       0.106 * cos(5.41622566807  + 15113.98923821520 *T)
    sumR2  +=       0.096 * cos(3.52561014696  + 3475.67750673520 *T)
    sumR2  +=       0.101 * cos(2.91169549546  + 692.15760122680 *T)
    sumR2  +=       0.093 * cos(3.21464368291  + 8671.96987044060 *T)
    sumR2  +=       0.087 * cos(5.24375873119  + 13119.72110282519 *T)
    sumR2  +=       0.106 * cos(3.76804681559  + 13517.87010623340 *T)
    sumR2  +=       0.089 * cos(5.73652101746  + 11371.70468975820 *T)
    sumR2  +=       0.104 * cos(5.90078964406  + 8432.76438481560 *T)
    sumR2  +=       0.099 * cos(3.61425346453  + 853.19638175200 *T)
    sumR2  +=       0.078 * cos(0.95802015551  + 3191.04922956520 *T)
    sumR2  +=       0.089 * cos(2.50872076312  + 1052.26838318840 *T)
    sumR2  +=       0.078 * cos(2.98115354616  + 26724.89941359840 *T)
    sumR2  +=       0.079 * cos(5.91514993930  + 6040.34724601740 *T)
    sumR2  +=       0.090 * cos(3.00950470302  + 12964.30070339100 *T)
    sumR2  +=       0.075 * cos(1.31512531841  + 17924.91069982040 *T)
    sumR2  +=       0.073 * cos(5.01153593776  + 1903.43681250120 *T)
    sumR2  +=       0.075 * cos(2.57969241717  + 10021.90459040220 *T)
    sumR2  +=       0.073 * cos(4.09954161905  + 3319.83703120740 *T)
    sumR2  +=       0.070 * cos(5.37678396256  + 12310.18132361080 *T)
    sumR2  +=       0.071 * cos(0.98422998240  + 13362.46696045140 *T)
    sumR2  +=       0.068 * cos(4.69140248943  + 522.57741809380 *T)
    sumR2  +=       0.068 * cos(3.24217441291  + 16706.58525184800 *T)
    sumR2  +=       0.069 * cos(2.53964193741  + 5884.92684658320 *T)
    sumR2  +=       0.070 * cos(2.38237948321  + 13916.01910964160 *T)
    sumR2  +=       0.066 * cos(3.48443797546  + 13760.59871020740 *T)
    sumR2  +=       0.068 * cos(5.27435069681  + 3980.50971301380 *T)
    sumR2  +=       0.076 * cos(0.12288413214  + 11216.28429032400 *T)
    sumR2  +=       0.073 * cos(3.94785759909  + 16173.37116840440 *T)
    sumR2  +=       0.062 * cos(0.15518975647  + 11081.21921028860 *T)
    sumR2  +=       0.074 * cos(2.99926779692  + 10021.76996979660 *T)
    sumR2  +=       0.080 * cos(6.18812323827  + 9380.95967271720 *T)
    sumR2  +=       0.064 * cos(3.99294283669  + 13358.92658845020 *T)
    sumR2  +=       0.061 * cos(2.56817782412  + 149.56319713460 *T)
    sumR2  +=       0.059 * cos(3.80850908229  + 10264.56588407340 *T)
    sumR2  +=       0.060 * cos(4.79879360562  + 1437.17561419860 *T)
    sumR2  +=       0.068 * cos(2.07515658228  + 1596.18644228460 *T)
    sumR2  +=       0.060 * cos(2.26503107747  + 28628.33622609960 *T)
    sumR2  +=       0.058 * cos(6.10782258899  + 5459.37628707820 *T)
    sumR2  +=       0.063 * cos(0.24407077742  + 18984.29263000960 *T)
    sumR2  +=       0.058 * cos(5.22738430747  + 13362.43245314700 *T)
    sumR2  +=       0.056 * cos(1.88708253797  + 9070.11887384880 *T)
    sumR2  +=       0.055 * cos(5.72587156499  + 433.71173787680 *T)
    sumR2  +=       0.056 * cos(1.18238857561  + 6696.47732458460 *T)
    sumR2  +=       0.062 * cos(3.98214740104  + 3427.92063123960 *T)
    sumR2  +=       0.056 * cos(5.37129176649  + 5099.26550511660 *T)
    sumR2  +=       0.057 * cos(0.86204712032  + 5642.19824260920 *T)
    sumR2  +=       0.055 * cos(1.31787605222  + 8982.81066930900 *T)
    sumR2  +=       0.056 * cos(5.49808236951  + 24093.27678917520 *T)
    sumR2  +=       0.054 * cos(3.77562154746  + 10213.28554621100 *T)
    sumR2  +=       0.054 * cos(2.36294643031  + 33716.96506586640 *T)
    sumR2  +=       0.063 * cos(3.53402427573  + 3205.54734666440 *T)
    sumR2  +=       0.066 * cos(6.25810929981  + 8270.29774868340 *T)
    sumR2  +=       0.067 * cos(0.58592813080  + 3335.08950239240 *T)
    sumR2  +=       0.055 * cos(4.74995645767  + 6660.86953400080 *T)
    sumR2  +=       0.053 * cos(0.31930032532  + 12721.57209941700 *T)
    sumR2  +=       0.069 * cos(2.42905593362  + 6158.64743530580 *T)
    sumR2  +=       0.055 * cos(5.03192783121  + 7255.56965173440 *T)
    sumR2  +=       0.051 * cos(5.45351947711  + 13207.02930736500 *T)
    sumR2  +=       0.050 * cos(0.23368526162  + 4379.63903749020 *T)
    sumR2  +=       0.053 * cos(2.12826344020  + 13171.00144068760 *T)
    sumR2  +=       0.049 * cos(2.35512151368  + 4032.77002792660 *T)
    sumR2  +=       0.049 * cos(2.81773050890  + 9808.53818466140 *T)
    sumR2  +=       0.067 * cos(2.26295786556  + 2766.26762836500 *T)
    sumR2  +=       0.049 * cos(0.72597680486  + 10028.95082710020 *T)
    sumR2  +=       0.050 * cos(6.10119450350  + 3253.30422216000 *T)
    sumR2  +=       0.059 * cos(0.27861398526  + 9225.53927328300 *T)
    sumR2  +=       0.053 * cos(1.27330874549  + 3344.49376205780 *T)
    sumR2  +=       0.049 * cos(2.91499389416  + 10404.73381232260 *T)
    sumR2  +=       0.047 * cos(0.48598973231  + 17395.21973472580 *T)
    sumR2  +=       0.049 * cos(3.40240207389  + 3304.58456002240 *T)
    sumR2  +=       0.060 * cos(3.40554368741  + 10551.52824519400 *T)
    sumR2  +=       0.050 * cos(1.83471955355  + 30376.35263916660 *T)
    sumR2  +=       0.046 * cos(3.04980996832  + 6144.42034130420 *T)
    sumR2  +=       0.049 * cos(0.33200904496  + 6702.00024889200 *T)
    sumR2  +=       0.053 * cos(0.35820094732  + 9468.26787725700 *T)
    sumR2  +=       0.050 * cos(0.15378507249  + 2648.45482547300 *T)
    sumR2  +=       0.044 * cos(0.79762157297  + 7107.82304427560 *T)
    sumR2  +=       0.050 * cos(2.41332661328  + 4407.10790388980 *T)
    sumR2  +=       0.054 * cos(0.21682940856  + 10022.81760116760 *T)
    sumR2  +=       0.044 * cos(5.78502737591  + 3364.49086444760 *T)
    sumR2  +=       0.044 * cos(0.75508636274  + 16062.18452611680 *T)
    sumR2  +=       0.043 * cos(3.10581037685  + 10235.13637553740 *T)
    sumR2  +=       0.044 * cos(5.21252098878  + 7373.38245462640 *T)
    sumR2  +=       0.058 * cos(4.51680312444  + 6660.44945790720 *T)
    sumR2  +=       0.043 * cos(0.15523355839  + 1505.28780909299 *T)
    sumR2  +=       0.055 * cos(5.93371026829  + 8439.87793181640 *T)
    sumR2  +=       0.042 * cos(5.40417232308  + 3344.20285535160 *T)
    sumR2  +=       0.056 * cos(3.87534582911  + 6816.28993343500 *T)
    sumR2  +=       0.049 * cos(5.57153056447  + 17468.85519794540 *T)
    sumR2  +=       0.046 * cos(3.07783299049  + 12722.55242048520 *T)
    sumR2  +=       0.045 * cos(1.49303542795  + 6247.51311552280 *T)
    sumR2  +=       0.055 * cos(6.26098836404  + 3361.38782219220 *T)
    sumR2  +=       0.042 * cos(4.74455545491  + 3369.06161416760 *T)
    sumR2  +=       0.040 * cos(1.58676666294  + 2818.03500860600 *T)
    sumR2  +=       0.043 * cos(4.97295184315  + 3337.02199804800 *T)
    sumR2  +=       0.055 * cos(3.80215638520  + 3376.64029337720 *T)
    sumR2  +=       0.047 * cos(1.97367775421  + 1581.95934828300 *T)
    sumR2  +=       0.045 * cos(2.03598705348  + 5244.04923920100 *T)
    sumR2  +=       0.049 * cos(4.43182954044  + 9595.23908922340 *T)
    sumR2  +=       0.043 * cos(0.38914582094  + 6665.97238221460 *T)
    sumR2  +=       0.040 * cos(1.06953099245  + 23695.12778576700 *T)
    sumR2  +=       0.038 * cos(0.59996214864  + 10713.99488132620 *T)
    sumR2  +=       0.039 * cos(3.79055778605  + 33318.81606245820 *T)
    sumR2  +=       0.041 * cos(5.54542347834  + 3863.18984479360 *T)
    sumR2  +=       0.037 * cos(5.27208151144  + 20752.66436247540 *T)
    sumR2  +=       0.039 * cos(3.15412760412  + 14469.58851248400 *T)
    sumR2  +=       0.041 * cos(0.55705920174  + 3346.13535100720 *T)
    sumR2  +=       0.048 * cos(1.42210748234  + 3316.73398895200 *T)
    sumR2  +=       0.049 * cos(5.44942247991  + 9638.94074787620 *T)
    sumR2  +=       0.038 * cos(2.11846271943  + 15664.03552270859 *T)
    sumR2  +=       0.035 * cos(5.30782933148  + 6677.63442474780 *T)
    sumR2  +=       0.036 * cos(2.68151450016  + 8584.66166590080 *T)
    sumR2  +=       0.046 * cos(3.09450395763  + 6701.58017279840 *T)
    sumR2  +=       0.038 * cos(1.81885395671  + 14158.74771361560 *T)
    sumR2  +=       0.037 * cos(5.80483071722  + 6684.81528205140 *T)
    sumR2  +=       0.047 * cos(2.60317702777  + 162.46663613220 *T)
    sumR2  +=       0.038 * cos(2.22559377525  + 3503.07906283200 *T)
    sumR2  +=       0.035 * cos(0.84358942718  + 6518.75821726740 *T)
    sumR2  +=       0.033 * cos(4.92302322062  + 23141.55838292460 *T)
    sumR2  +=       0.043 * cos(0.95148893985  + 34115.11406927460 *T)
    sumR2  +=       0.046 * cos(4.99457196207  + 9381.93999378540 *T)
    sumR2  +=       0.043 * cos(1.73832102244  + 12323.42309600880 *T)
    sumR2  +=       0.032 * cos(2.92741580696  + 9499.25986200560 *T)
    sumR2  +=       0.043 * cos(1.50296327105  + 27035.74021246680 *T)
    sumR2  +=       0.033 * cos(5.41731109688  + 20597.24396304120 *T)
    sumR2  +=       0.032 * cos(4.13881058425  + 31570.79964939120 *T)
    sumR2  +=       0.040 * cos(5.84463069180  + 16460.33352952499 *T)
    sumR2  +=       0.035 * cos(5.50083743557  + 8425.65083781480 *T)
    sumR2  +=       0.038 * cos(5.91991985929  + 3336.73109134180 *T)
    sumR2  +=       0.030 * cos(5.09155042811  + 5085.03841111500 *T)
    sumR2  +=       0.031 * cos(4.27211180916  + 36659.42848915800 *T)
    sumR2  +=       0.030 * cos(4.87389489592  + 1692.16566950240 *T)
    sumR2  +=       0.030 * cos(3.75494646937  + 4845.90023579280 *T)
    sumR2  +=       0.034 * cos(3.71030605417  + 8827.39026987480 *T)
    sumR2  +=       0.030 * cos(0.01599456850  + 11243.68584642080 *T)
    sumR2  +=       0.029 * cos(2.66198696023  + 12295.95422960920 *T)
    sumR2  +=       0.028 * cos(3.12114207206  + 11925.27409260060 *T)
    sumR2  +=       0.036 * cos(1.16568750904  + 18052.92954315780 *T)
    sumR2  +=       0.032 * cos(1.72704226918  + 12566.15169998280 *T)
    sumR2  +=       0.028 * cos(2.80254457895  + 3607.21946842160 *T)
    sumR2  +=       0.032 * cos(4.04390839531  + 6680.24453233140 *T)
    sumR2  +=       0.027 * cos(2.53565904393  + 19004.64794940840 *T)
    sumR2  +=       0.036 * cos(3.33518515689  + 29978.20363575840 *T)
    sumR2  +=       0.028 * cos(0.58104982217  + 20354.51535906720 *T)
    sumR2  +=       0.029 * cos(5.85207825749  + 31172.65064598300 *T)
    sumR2  +=       0.026 * cos(2.22598820379  + 3.52311834900 *T)
    sumR2  +=       0.033 * cos(5.15903548241  + 10596.18207843420 *T)
    sumR2  +=       0.028 * cos(1.89792803028  + 3333.56619000180 *T)
    sumR2  +=       0.027 * cos(2.25129832716  + 76.26607127560 *T)
    sumR2  +=       0.026 * cos(4.32703172263  + 17085.95866572220 *T)
    sumR2  +=       0.026 * cos(5.12847780702  + 11766.26326451460 *T)
    sumR2  +=       0.028 * cos(0.80926596905  + 20735.83216142559 *T)
    sumR2  +=       0.026 * cos(5.06814803493  + 17412.05193577560 *T)
    sumR2  +=       0.024 * cos(4.47399155600  + 3134.42687826260 *T)
    sumR2  +=       0.024 * cos(4.35557152695  + 14584.29827312060 *T)
    sumR2  +=       0.027 * cos(4.98717355201  + 3877.41693879520 *T)
    sumR2  +=       0.025 * cos(5.63712701265  + 14054.60730802600 *T)
    sumR2  +=       0.024 * cos(4.12573657295  + 4193.80880845180 *T)
    sumR2  +=       0.025 * cos(1.40880282063  + 8186.51266249260 *T)
    sumR2  +=       0.024 * cos(4.96463811748  + 26087.90314157420 *T)

    # MARS Radius
    sumR3  = 0
    sumR3  +=    1113.107 * cos(5.14987350142  + 3340.61242669980 *T)
    sumR3  +=     424.446 * cos(5.61343766478  + 6681.22485339960 *T)
    sumR3  +=     100.044 * cos(5.99726827028  + 10021.83728009940 *T)
    sumR3  +=      19.606 * cos(0.07633062094  + 13362.44970679920 *T)
    sumR3  +=       3.477 * cos(0.42951907576  + 16703.06213349900 *T)
    sumR3  +=       4.693 * cos(3.14159265359  + 0.00000000000 *T)
    sumR3  +=       2.869 * cos(0.44711842697  + 3344.13554504880 *T)
    sumR3  +=       2.428 * cos(3.02115527957  + 3185.19202726560 *T)
    sumR3  +=       0.688 * cos(0.80693359444  + 6684.74797174860 *T)
    sumR3  +=       0.540 * cos(3.86836515672  + 1059.38193018920 *T)
    sumR3  +=       0.577 * cos(0.77853275120  + 20043.67456019880 *T)
    sumR3  +=       0.468 * cos(4.52450786544  + 3496.03282613400 *T)
    sumR3  +=       0.487 * cos(1.60862391345  + 3583.34103067380 *T)
    sumR3  +=       0.362 * cos(4.42397903418  + 2787.04302385740 *T)
    sumR3  +=       0.397 * cos(5.71967986581  + 3149.16416058820 *T)
    sumR3  +=       0.299 * cos(0.75640033535  + 3738.76143010800 *T)
    sumR3  +=       0.350 * cos(5.55691984889  + 4136.91043351620 *T)
    sumR3  +=       0.314 * cos(3.37632898783  + 6525.80445396540 *T)
    sumR3  +=       0.233 * cos(2.13965262306  + 3097.88382272579 *T)
    sumR3  +=       0.212 * cos(4.20487494011  + 3341.59274776800 *T)
    sumR3  +=       0.214 * cos(0.56779556947  + 155.42039943420 *T)
    sumR3  +=       0.201 * cos(2.35284464564  + 1592.59601363280 *T)
    sumR3  +=       0.204 * cos(0.96723315302  + 1990.74501704100 *T)
    sumR3  +=       0.230 * cos(5.82959057146  + 2388.89402044920 *T)
    sumR3  +=       0.160 * cos(4.14974194757  + 4535.05943692440 *T)
    sumR3  +=       0.156 * cos(5.13345210285  + 796.29800681640 *T)
    sumR3  +=       0.146 * cos(3.29541339704  + 2957.71589447660 *T)
    sumR3  +=       0.156 * cos(1.13805508457  + 10025.36039844840 *T)
    sumR3  +=       0.159 * cos(3.76174286126  + 1194.44701022460 *T)
    sumR3  +=       0.136 * cos(5.58293924642  + 3894.18182954220 *T)
    sumR3  +=       0.145 * cos(3.02112488974  + 3337.08930835080 *T)
    sumR3  +=       0.146 * cos(4.25359617225  + 4399.99435688900 *T)
    sumR3  +=       0.100 * cos(4.26740096115  + 1589.07289528380 *T)
    sumR3  +=       0.091 * cos(6.12108227231  + 6489.77658728800 *T)
    sumR3  +=       0.101 * cos(1.17242141914  + 7079.37385680780 *T)
    sumR3  +=       0.098 * cos(1.12613498307  + 23384.28698689860 *T)
    sumR3  +=       0.075 * cos(6.02182689533  + 529.69096509460 *T)
    sumR3  +=       0.078 * cos(5.93418069229  + 7477.52286021600 *T)
    sumR3  +=       0.068 * cos(1.70775041911  + 9492.14631500480 *T)
    sumR3  +=       0.070 * cos(3.67253155189  + 8962.45534991020 *T)
    sumR3  +=       0.062 * cos(4.14954627063  + 4292.33083295040 *T)
    sumR3  +=       0.067 * cos(0.15092246550  + 3870.30339179440 *T)
    sumR3  +=       0.066 * cos(1.12578506413  + 6151.53388830500 *T)
    sumR3  +=       0.062 * cos(1.93984222091  + 6923.95345737360 *T)
    sumR3  +=       0.060 * cos(4.30641513096  + 12303.06777661000 *T)
    sumR3  +=       0.074 * cos(4.72156358149  + 3347.72597370060 *T)
    sumR3  +=       0.055 * cos(2.75742802464  + 4933.20844033260 *T)
    sumR3  +=       0.053 * cos(4.77894621871  + 6127.65545055720 *T)
    sumR3  +=       0.050 * cos(2.57807944338  + 3340.59517304760 *T)
    sumR3  +=       0.048 * cos(3.72532068729  + 9866.41688066520 *T)
    sumR3  +=       0.042 * cos(6.15391373101  + 7234.79425624200 *T)
    sumR3  +=       0.043 * cos(3.43955584179  + 5486.77784317500 *T)
    sumR3  +=       0.050 * cos(4.60885064189  + 3340.62968035200 *T)
    sumR3  +=       0.044 * cos(4.47943289816  + 426.59819087600 *T)
    sumR3  +=       0.045 * cos(6.24485207500  + 2274.11694950980 *T)
    sumR3  +=       0.040 * cos(0.73805316471  + 7.11354700080 *T)
    sumR3  +=       0.038 * cos(1.06497784748  + 9623.68827669120 *T)
    sumR3  +=       0.041 * cos(4.60954640493  + 6682.20517446780 *T)
    sumR3  +=       0.038 * cos(4.60963924082  + 7740.60678358880 *T)
    sumR3  +=       0.036 * cos(3.15129246169  + 639.89728631400 *T)
    sumR3  +=       0.037 * cos(6.14629970622  + 5729.50644714900 *T)
    sumR3  +=       0.040 * cos(1.47395544850  + 13365.97282514820 *T)
    sumR3  +=       0.035 * cos(2.60772245492  + 4690.47983635860 *T)
    sumR3  +=       0.035 * cos(2.45077931900  + 2288.34404351140 *T)
    sumR3  +=       0.034 * cos(4.50544497582  + 7875.67186362420 *T)
    sumR3  +=       0.042 * cos(2.11383330882  + 12832.75874170460 *T)
    sumR3  +=       0.033 * cos(3.97655860802  + 191.44826611160 *T)
    sumR3  +=       0.032 * cos(4.06047154083  + 3532.06069281140 *T)
    sumR3  +=       0.034 * cos(5.06671994554  + 6681.24210705180 *T)
    sumR3  +=       0.042 * cos(4.67024017441  + 15643.68020330980 *T)
    sumR3  +=       0.036 * cos(0.29140871479  + 3723.50895892300 *T)
    sumR3  +=       0.033 * cos(4.52588149629  + 4929.68532198360 *T)
    sumR3  +=       0.034 * cos(2.74499072890  + 6677.70173505060 *T)
    sumR3  +=       0.030 * cos(1.57743012295  + 10419.98628350760 *T)
    sumR3  +=       0.040 * cos(1.31713176335  + 5331.35744374080 *T)
    sumR3  +=       0.034 * cos(3.03490000148  + 6681.20759974740 *T)
    sumR3  +=       0.031 * cos(1.25306164576  + 242.72860397400 *T)
    sumR3  +=       0.029 * cos(5.44118048011  + 10021.85453375160 *T)
    sumR3  +=       0.029 * cos(3.41037097794  + 10021.82002644720 *T)
    sumR3  +=       0.028 * cos(4.81855060968  + 6836.64525283380 *T)
    sumR3  +=       0.028 * cos(0.72745302325  + 6283.07584999140 *T)
    sumR3  +=       0.030 * cos(3.06267495632  + 2281.23049651060 *T)
    sumR3  +=       0.027 * cos(0.62535799761  + 6681.29216370240 *T)
    sumR3  +=       0.030 * cos(3.56389476096  + 8969.56889691100 *T)
    sumR3  +=       0.032 * cos(5.31078691958  + 553.56940284240 *T)
    sumR3  +=       0.033 * cos(2.00753799963  + 5884.92684658320 *T)
    sumR3  +=       0.027 * cos(1.18001479303  + 8955.34180290940 *T)
    sumR3  +=       0.025 * cos(2.94122791628  + 2544.31441988340 *T)
    sumR3  +=       0.025 * cos(0.17808889827  + 9830.38901398780 *T)
    sumR3  +=       0.031 * cos(5.73384204593  + 3339.63210563160 *T)
    sumR3  +=       0.028 * cos(3.70436008855  + 6298.32832117640 *T)
    sumR3  +=       0.024 * cos(0.07739730466  + 10818.13528691580 *T)
    sumR3  +=       0.027 * cos(1.03401737006  + 6681.15754309680 *T)
    sumR3  +=       0.024 * cos(0.09987383242  + 382.89653222320 *T)
    sumR3  +=       0.027 * cos(0.60053652251  + 3553.91152213780 *T)
    sumR3  +=       0.027 * cos(4.97107900859  + 3767.21061757580 *T)
    sumR3  +=       0.024 * cos(3.36621347824  + 8031.09226305840 *T)
    sumR3  +=       0.026 * cos(4.80509833364  + 7632.94325965020 *T)
    sumR3  +=       0.022 * cos(1.43966107077  + 26724.89941359840 *T)
    sumR3  +=       0.020 * cos(0.36901153144  + 398.14900340820 *T)
    sumR3  +=       0.019 * cos(1.00325306041  + 7064.12138562280 *T)
    sumR3  +=       0.020 * cos(0.19821371186  + 6674.11130639880 *T)
    sumR3  +=       0.018 * cos(1.07169297658  + 7210.91581849420 *T)
    sumR3  +=       0.021 * cos(2.69512140340  + 10018.31416175040 *T)
    sumR3  +=       0.023 * cos(3.25670295081  + 5621.84292321040 *T)
    sumR3  +=       0.018 * cos(0.24542781929  + 3981.49003408200 *T)
    sumR3  +=       0.018 * cos(5.14428605713  + 6688.33840040040 *T)
    sumR3  +=       0.020 * cos(0.97539014916  + 6894.52394883760 *T)
    sumR3  +=       0.024 * cos(4.42156401638  + 1748.01641306700 *T)
    sumR3  +=       0.016 * cos(0.65389503440  + 3340.54511639700 *T)
    sumR3  +=       0.016 * cos(5.28427106012  + 3333.49887969900 *T)

    # MARS Radius
    sumR4  = 0
    sumR4  +=      19.552 * cos(3.58211650473  + 3340.61242669980 *T)
    sumR4  +=      16.323 * cos(4.05116076923  + 6681.22485339960 *T)
    sumR4  +=       5.848 * cos(4.46383962094  + 10021.83728009940 *T)
    sumR4  +=       1.532 * cos(4.84374321619  + 13362.44970679920 *T)
    sumR4  +=       0.375 * cos(1.50962233608  + 3185.19202726560 *T)
    sumR4  +=       0.339 * cos(5.20684967613  + 16703.06213349900 *T)
    sumR4  +=       0.151 * cos(5.16472931648  + 3344.13554504880 *T)
    sumR4  +=       0.125 * cos(2.19233532803  + 3496.03282613400 *T)
    sumR4  +=       0.148 * cos(0.00000000000  + 0.00000000000 *T)
    sumR4  +=       0.087 * cos(0.10275067375  + 3583.34103067380 *T)
    sumR4  +=       0.070 * cos(5.55303619365  + 20043.67456019880 *T)
    sumR4  +=       0.048 * cos(2.91797786512  + 2787.04302385740 *T)
    sumR4  +=       0.049 * cos(5.57038514830  + 6684.74797174860 *T)
    sumR4  +=       0.057 * cos(1.86867280743  + 6525.80445396540 *T)
    sumR4  +=       0.034 * cos(3.63370917313  + 3097.88382272579 *T)
    sumR4  +=       0.021 * cos(2.30677637991  + 1059.38193018920 *T)
    sumR4  +=       0.021 * cos(4.03682122005  + 4136.91043351620 *T)
    sumR4  +=       0.018 * cos(4.17438755890  + 3149.16416058820 *T)
    sumR4  +=       0.013 * cos(1.97805475719  + 155.42039943420 *T)
    sumR4  +=       0.017 * cos(5.94856414690  + 10025.36039844840 *T)
    sumR4  +=       0.012 * cos(0.51097435253  + 6923.95345737360 *T)
    sumR4  +=       0.012 * cos(3.27945299261  + 6127.65545055720 *T)
    sumR4  +=       0.013 * cos(2.22938285522  + 9866.41688066520 *T)
    sumR4  +=       0.016 * cos(2.64206771227  + 4399.99435688900 *T)
    sumR4  +=       0.013 * cos(2.75813969974  + 6836.64525283380 *T)
    sumR4  +=       0.013 * cos(4.51988349676  + 6489.77658728800 *T)
    sumR4  +=       0.014 * cos(0.76591550754  + 3894.18182954220 *T)
    sumR4  +=       0.011 * cos(5.47381242145  + 3738.76143010800 *T)

    # MARS Radius
    sumR5  = 0
    sumR5  +=       0.476 * cos(2.47617204701  + 6681.22485339960 *T)
    sumR5  +=       0.268 * cos(2.91510547706  + 10021.83728009940 *T)
    sumR5  +=       0.115 * cos(1.76888962311  + 3340.61242669980 *T)
    sumR5  +=       0.096 * cos(3.31378377179  + 13362.44970679920 *T)
    sumR5  +=       0.048 * cos(6.27457828287  + 3185.19202726560 *T)
    sumR5  +=       0.027 * cos(3.69380877066  + 16703.06213349900 *T)
    sumR5  +=       0.013 * cos(3.72005807206  + 3496.03282613400 *T)
    sumR5  +=       0.012 * cos(4.88179002689  + 3583.34103067380 *T)
    sumR5  +=       0.012 * cos(3.14159265359  + 0.00000000000 *T)

    # calculate and scale the sums
    L = sumL0 + sumL1*T**1 + sumL2*T**2 + sumL3*T**3 + sumL4*T**4 + sumL5*T**5
    L /= 1e8
    L = Mod(degrees(L), 360.)
    B = sumB0 + sumB1*T**1 + sumB2*T**2 + sumB3*T**3 + sumB4*T**4 + sumB5*T**5
    B /= 1e8
    B = degrees(B)
    R = sumR0 + sumR1*T**1 + sumR2*T**2 + sumR3*T**3 + sumR4*T**4 + sumR5*T**5
    R /= 1e8

    # calculate X, Y, Z
    X = R * Cos(B) * Cos(L)
    Y = R * Cos(B) * Sin(L)
    Z = R * Sin(B)

    # create dictionary
    Dict = {  "Object"  : "Mars",
              "Mode"    : "LBR",
              "Long"    : L,
              "Lat"     : B,
              "Rau"     : R,
              "X"       : X,
              "Y"       : Y,
              "Z"       : Z,
              "Alt"     : -34./60.
           }
    Data = (L, B, R, Dict)


    return Data
#######################################################################



def getMarsData(dtCivicTime, zone, verbose = False):
    L, B, R, Dict = getMarsLBRData(dtCivicTime, zone, verbose)
    X = Dict["X"]; Y = Dict["Y"]; Z = Dict["Z"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Le, Be, Re, Dicte = getEarthLBRData(dtCivicTime, zone, verbose)
        Xe = Dicte["X"]; Ye = Dicte["Y"]; Ze = Dicte["Z"]
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getMercuryLBRData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Mercury"
#       Long  :  heliocentric Longitude, degrees
#       Lat   :  heliocentric Latitude,degrees
#       Rau   :  Radius, distaance Sun - Object (au)
#
#   Similarly for each planet
#   data from Meeus, Appendix II and VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.2 hvf 2023-04-06 getMercuryLBRData.py created by mkPlanetLBRData.py"
VersionList.append(Version)

def getMercuryLBRData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. (31.1) Julian Millennia

#######################################################################

    # MERCURY Longitude
    sumL0  = 0
    sumL0  +=  440250710.144 * cos(0.00000000000  + 0.00000000000 *T)
    sumL0  +=  40989414.976 * cos(1.48302034194  + 26087.90314157420 *T)
    sumL0  +=  5046294.199 * cos(4.47785489540  + 52175.80628314840 *T)
    sumL0  +=  855346.843 * cos(1.16520322351  + 78263.70942472259 *T)
    sumL0  +=  165590.362 * cos(4.11969163181  + 104351.61256629678 *T)
    sumL0  +=   34561.897 * cos(0.77930765817  + 130439.51570787100 *T)
    sumL0  +=    7583.476 * cos(3.71348400510  + 156527.41884944518 *T)
    sumL0  +=    3559.740 * cos(1.51202669419  + 1109.37855209340 *T)
    sumL0  +=    1726.012 * cos(0.35832239908  + 182615.32199101936 *T)
    sumL0  +=    1803.463 * cos(4.10333178410  + 5661.33204915220 *T)
    sumL0  +=    1364.682 * cos(4.59918318745  + 27197.28169366760 *T)
    sumL0  +=    1589.923 * cos(2.99510417815  + 25028.52121138500 *T)
    sumL0  +=    1017.332 * cos(0.88031439040  + 31749.23519072640 *T)
    sumL0  +=     714.182 * cos(1.54144865265  + 24978.52458948080 *T)
    sumL0  +=     643.759 * cos(5.30266110787  + 21535.94964451540 *T)
    sumL0  +=     404.200 * cos(3.28228847025  + 208703.22513259359 *T)
    sumL0  +=     352.441 * cos(5.24156297101  + 20426.57109242200 *T)
    sumL0  +=     343.313 * cos(5.76531885335  + 955.59974160860 *T)
    sumL0  +=     339.214 * cos(5.86327765000  + 25558.21217647960 *T)
    sumL0  +=     451.137 * cos(6.04989275289  + 51116.42435295920 *T)
    sumL0  +=     325.335 * cos(1.33674334780  + 53285.18483524180 *T)
    sumL0  +=     259.587 * cos(0.98732428184  + 4551.95349705880 *T)
    sumL0  +=     345.212 * cos(2.79211901539  + 15874.61759536320 *T)
    sumL0  +=     272.947 * cos(2.49451163975  + 529.69096509460 *T)
    sumL0  +=     234.830 * cos(0.26672118900  + 11322.66409830440 *T)
    sumL0  +=     238.793 * cos(0.11343953378  + 1059.38193018920 *T)
    sumL0  +=     264.336 * cos(3.91705094013  + 57837.13833230060 *T)
    sumL0  +=     216.645 * cos(0.65987207348  + 13521.75144159140 *T)
    sumL0  +=     183.359 * cos(2.62878670784  + 27043.50288318280 *T)
    sumL0  +=     175.965 * cos(4.53636829858  + 51066.42773105500 *T)
    sumL0  +=     181.629 * cos(2.43413502466  + 25661.30495069820 *T)
    sumL0  +=     208.995 * cos(2.09178234008  + 47623.85278608960 *T)
    sumL0  +=     172.643 * cos(2.45200164173  + 24498.83024629040 *T)
    sumL0  +=     142.316 * cos(3.36003948842  + 37410.56723987860 *T)
    sumL0  +=     137.942 * cos(0.29098447849  + 10213.28554621100 *T)
    sumL0  +=     118.233 * cos(2.78149786369  + 77204.32749453338 *T)
    sumL0  +=      96.860 * cos(6.20398202740  + 234791.12827416777 *T)
    sumL0  +=     125.219 * cos(3.72079804425  + 39609.65458316560 *T)
    sumL0  +=      86.819 * cos(2.64219349385  + 51646.11531805379 *T)
    sumL0  +=      86.723 * cos(1.95953042650  + 46514.47423399620 *T)
    sumL0  +=      88.329 * cos(5.41338795963  + 26617.59410666880 *T)
    sumL0  +=     106.422 * cos(4.20572116254  + 19804.82729158280 *T)
    sumL0  +=      89.987 * cos(5.85243631094  + 41962.52073693740 *T)
    sumL0  +=      84.971 * cos(4.33100364958  + 79373.08797681599 *T)
    sumL0  +=      69.247 * cos(4.19446437496  + 19.66976089979 *T)
    sumL0  +=      63.463 * cos(3.14700877722  + 7238.67559160000 *T)
    sumL0  +=      68.493 * cos(0.63424819267  + 83925.04147387479 *T)
    sumL0  +=      69.729 * cos(3.57201709671  + 25132.30339996560 *T)
    sumL0  +=      59.481 * cos(2.74692752000  + 16983.99614745660 *T)
    sumL0  +=      64.830 * cos(0.04762925810  + 33326.57873317420 *T)
    sumL0  +=      55.376 * cos(4.05312663019  + 30639.85663863300 *T)
    sumL0  +=      54.442 * cos(3.14331542453  + 27147.28507176339 *T)
    sumL0  +=      47.560 * cos(5.49722099211  + 3.88133535800 *T)
    sumL0  +=      49.567 * cos(3.98985863874  + 6770.71060124560 *T)
    sumL0  +=      56.531 * cos(5.11920557675  + 73711.75592766379 *T)
    sumL0  +=      41.764 * cos(5.64185159566  + 53131.40602475700 *T)
    sumL0  +=      51.458 * cos(5.47786463494  + 50586.73338786459 *T)
    sumL0  +=      44.744 * cos(1.22366857463  + 77154.33087262919 *T)
    sumL0  +=      41.882 * cos(5.19309298528  + 6283.07584999140 *T)
    sumL0  +=      38.045 * cos(2.43117327523  + 12566.15169998280 *T)
    sumL0  +=      35.627 * cos(0.81390126585  + 32858.61374281979 *T)
    sumL0  +=      48.007 * cos(5.49260554912  + 51749.20809227239 *T)
    sumL0  +=      35.392 * cos(3.36964859355  + 36301.18868778519 *T)
    sumL0  +=      33.951 * cos(2.78618091049  + 14765.23904326980 *T)
    sumL0  +=      30.560 * cos(5.84045074182  + 43071.89928903080 *T)
    sumL0  +=      35.964 * cos(1.42380838630  + 2218.75710418680 *T)
    sumL0  +=      34.044 * cos(0.47470299167  + 65697.55772473979 *T)
    sumL0  +=      30.800 * cos(5.77017310191  + 103292.23063610760 *T)
    sumL0  +=      28.496 * cos(0.65048992658  + 426.59819087600 *T)
    sumL0  +=      26.215 * cos(5.24158618719  + 22645.32819660879 *T)
    sumL0  +=      26.253 * cos(0.64807043102  + 1589.07289528380 *T)
    sumL0  +=      29.538 * cos(0.69771244088  + 213.29909543800 *T)
    sumL0  +=      27.504 * cos(0.98010127839  + 45892.73043315699 *T)
    sumL0  +=      22.347 * cos(5.65335125838  + 77734.01845962799 *T)
    sumL0  +=      22.047 * cos(4.93398225193  + 72602.37737557039 *T)
    sumL0  +=      22.275 * cos(2.17909842576  + 52705.49724824299 *T)
    sumL0  +=      24.252 * cos(4.39994170609  + 7.11354700080 *T)
    sumL0  +=      26.751 * cos(1.06145361792  + 3442.57494496540 *T)
    sumL0  +=      23.656 * cos(2.84168536986  + 260879.03141574195 *T)
    sumL0  +=      22.908 * cos(2.58462026514  + 68050.42387851159 *T)
    sumL0  +=      27.086 * cos(0.08501738669  + 63498.47038145279 *T)
    sumL0  +=      22.247 * cos(3.22418265191  + 25448.00585526019 *T)
    sumL0  +=      17.803 * cos(3.61202297483  + 110012.94461544900 *T)
    sumL0  +=      22.407 * cos(1.02520094825  + 105460.99111839020 *T)
    sumL0  +=      17.576 * cos(4.71742326981  + 25874.60404613620 *T)
    sumL0  +=      18.586 * cos(4.52709871258  + 28306.66024576099 *T)
    sumL0  +=      14.176 * cos(6.12394176563  + 53235.18821333759 *T)
    sumL0  +=      14.186 * cos(5.14246797066  + 26068.23338067440 *T)
    sumL0  +=      17.244 * cos(0.28394746813  + 51220.20654153979 *T)
    sumL0  +=      17.176 * cos(3.26084092971  + 153.77881048480 *T)
    sumL0  +=      14.938 * cos(1.83542009339  + 99799.65906923798 *T)
    sumL0  +=      13.387 * cos(0.76564655407  + 56727.75978020720 *T)
    sumL0  +=      13.978 * cos(2.30193139916  + 76674.63652943878 *T)
    sumL0  +=      14.428 * cos(0.96646356501  + 26107.57290247399 *T)
    sumL0  +=      11.990 * cos(6.20492907598  + 18849.22754997420 *T)
    sumL0  +=      14.381 * cos(1.90956715654  + 23969.13928119580 *T)
    sumL0  +=      11.233 * cos(2.04817126136  + 32370.97899156560 *T)
    sumL0  +=      13.392 * cos(4.51750784605  + 26080.78959457339 *T)
    sumL0  +=      11.632 * cos(2.38496860260  + 79219.30916633119 *T)
    sumL0  +=      12.412 * cos(2.22280944169  + 77837.11123384659 *T)
    sumL0  +=       9.950 * cos(2.04594448880  + 48733.23133818299 *T)
    sumL0  +=       9.803 * cos(2.26706433546  + 26091.78447693220 *T)
    sumL0  +=       9.362 * cos(5.44291958209  + 38654.05484155699 *T)
    sumL0  +=       9.747 * cos(3.83976857418  + 26084.02180621620 *T)
    sumL0  +=       9.264 * cos(4.02987000812  + 467.96499035440 *T)
    sumL0  +=       8.961 * cos(0.11062526114  + 62389.09182935939 *T)
    sumL0  +=      11.543 * cos(4.17789167759  + 103242.23401420339 *T)
    sumL0  +=      11.146 * cos(3.78292300417  + 26301.20223701220 *T)
    sumL0  +=       9.677 * cos(2.98527809776  + 59414.48187474840 *T)
    sumL0  +=       8.977 * cos(3.47888073089  + 91785.46086631398 *T)
    sumL0  +=       9.664 * cos(5.77941968495  + 25938.33994443960 *T)
    sumL0  +=       8.855 * cos(2.84672636028  + 25035.63475838580 *T)
    sumL0  +=       8.181 * cos(5.77857196635  + 40853.14218484400 *T)
    sumL0  +=       8.033 * cos(2.45692824195  + 129380.13377768178 *T)
    sumL0  +=       8.343 * cos(5.34499871294  + 19317.19254032860 *T)
    sumL0  +=       7.425 * cos(4.71160330390  + 6.62855890001 *T)
    sumL0  +=       9.001 * cos(6.23396256413  + 25021.40766438420 *T)
    sumL0  +=       8.126 * cos(1.12294634635  + 26095.01668857500 *T)
    sumL0  +=       6.568 * cos(3.66248946629  + 26514.50133245020 *T)
    sumL0  +=       7.038 * cos(3.99035923761  + 71980.63357473118 *T)
    sumL0  +=       6.956 * cos(1.62821260299  + 23869.14603738740 *T)
    sumL0  +=       7.595 * cos(0.18334396433  + 12432.04265039780 *T)
    sumL0  +=       6.061 * cos(3.67044794062  + 27676.97603685800 *T)
    sumL0  +=       5.896 * cos(5.57171141866  + 94138.32702008578 *T)
    sumL0  +=       5.716 * cos(5.18204203484  + 78793.40038981718 *T)
    sumL0  +=       5.855 * cos(2.14311779301  + 20760.42703319140 *T)
    sumL0  +=       5.681 * cos(1.60727624525  + 98690.28051714458 *T)
    sumL0  +=       5.788 * cos(2.35467492390  + 103821.92160120218 *T)
    sumL0  +=       5.121 * cos(3.77832929907  + 58946.51688439399 *T)
    sumL0  +=       5.873 * cos(5.76210244486  + 286966.93455731618 *T)
    sumL0  +=       5.858 * cos(6.12538452806  + 26011.63707029860 *T)
    sumL0  +=       5.215 * cos(3.29186833997  + 38519.94579197200 *T)
    sumL0  +=       6.000 * cos(0.00057044073  + 51535.90899683439 *T)
    sumL0  +=       4.647 * cos(0.29020584575  + 136100.84775702321 *T)
    sumL0  +=       5.787 * cos(4.44783057272  + 19406.67828817460 *T)
    sumL0  +=       5.908 * cos(4.12195491631  + 29530.47808653960 *T)
    sumL0  +=       5.918 * cos(3.98930701135  + 131548.89425996438 *T)
    sumL0  +=       5.728 * cos(3.02314979708  + 89586.37352302698 *T)
    sumL0  +=       3.928 * cos(4.81437933690  + 125887.56221081220 *T)
    sumL0  +=       3.929 * cos(2.48449041501  + 69159.80243060499 *T)
    sumL0  +=       3.681 * cos(2.80180999964  + 79323.09135491178 *T)
    sumL0  +=       4.498 * cos(1.50325539137  + 51962.50718771040 *T)
    sumL0  +=       3.687 * cos(5.33717753698  + 102762.53967101300 *T)
    sumL0  +=       3.497 * cos(1.85400531491  + 52156.13652224860 *T)
    sumL0  +=       3.867 * cos(1.25354714671  + 54394.56338733519 *T)
    sumL0  +=       4.459 * cos(4.88911997687  + 50057.04242277000 *T)
    sumL0  +=       3.396 * cos(3.73870967348  + 82815.66292178139 *T)
    sumL0  +=       3.525 * cos(6.07665337319  + 25934.12433108940 *T)
    sumL0  +=       3.492 * cos(1.28206984744  + 52168.69273614759 *T)
    sumL0  +=       3.411 * cos(6.13976263434  + 639.89728631400 *T)
    sumL0  +=       4.395 * cos(3.25475914760  + 77308.10968311399 *T)
    sumL0  +=       2.990 * cos(5.27323635392  + 46848.33017476560 *T)
    sumL0  +=       3.218 * cos(5.20938751579  + 103925.01437542078 *T)
    sumL0  +=       2.899 * cos(5.08451495112  + 58458.88213313979 *T)
    sumL0  +=       3.708 * cos(0.02713701028  + 26241.68195205900 *T)
    sumL0  +=       2.991 * cos(2.92278339368  + 44937.13069154840 *T)
    sumL0  +=       3.177 * cos(0.05753403864  + 22747.29071487440 *T)
    sumL0  +=       3.163 * cos(5.38713552769  + 105307.21230790540 *T)
    sumL0  +=       3.556 * cos(3.96231142071  + 52195.47604404819 *T)
    sumL0  +=       2.787 * cos(0.54704419913  + 52389.10537858640 *T)
    sumL0  +=       2.458 * cos(1.14904830408  + 2333.19639287200 *T)
    sumL0  +=       2.492 * cos(5.24922078791  + 25668.41849769900 *T)
    sumL0  +=       2.409 * cos(2.14208355028  + 64741.95798313119 *T)
    sumL0  +=       2.509 * cos(0.61934630416  + 85034.42002596818 *T)
    sumL0  +=       2.416 * cos(5.26204437192  + 52179.68761850640 *T)
    sumL0  +=       2.296 * cos(3.09900451546  + 88476.99497093359 *T)
    sumL0  +=       2.357 * cos(0.17742385045  + 117873.36400788820 *T)
    sumL0  +=       3.012 * cos(0.83689268367  + 129330.13715577760 *T)
    sumL0  +=       2.421 * cos(3.95913346049  + 49.99662190420 *T)
    sumL0  +=       2.197 * cos(2.56129546358  + 3340.61242669980 *T)
    sumL0  +=       2.402 * cos(0.55121112633  + 52171.92494779040 *T)
    sumL0  +=       2.106 * cos(5.41271503337  + 155468.03691925600 *T)
    sumL0  +=       2.566 * cos(3.00606194044  + 51109.31080595839 *T)
    sumL0  +=       2.392 * cos(0.17898534184  + 103.09277421860 *T)
    sumL0  +=       2.737 * cos(1.08192567668  + 632.78373931320 *T)
    sumL0  +=       2.523 * cos(5.90356438675  + 51123.53789995999 *T)
    sumL0  +=       2.098 * cos(2.47132069140  + 66941.04532641819 *T)
    sumL0  +=       2.003 * cos(3.75422326828  + 16066.06586147480 *T)
    sumL0  +=       2.188 * cos(1.18654376903  + 2118.76386037840 *T)
    sumL0  +=       2.593 * cos(2.55122032446  + 25654.19140369739 *T)
    sumL0  +=       2.559 * cos(1.18101453315  + 45494.58142974879 *T)
    sumL0  +=       1.918 * cos(0.79899605678  + 26555.86813192860 *T)
    sumL0  +=       1.819 * cos(0.69178943674  + 98068.53671630540 *T)
    sumL0  +=       2.471 * cos(2.55057179214  + 52026.24308601380 *T)
    sumL0  +=       1.704 * cos(0.36723577740  + 52602.40447402440 *T)
    sumL0  +=       2.129 * cos(2.66251948472  + 18093.37469954999 *T)
    sumL0  +=       1.715 * cos(4.62177464710  + 49957.04917896160 *T)
    sumL0  +=       1.916 * cos(5.83726462050  + 85502.38501632259 *T)
    sumL0  +=       1.976 * cos(3.57365896777  + 24395.73747207180 *T)
    sumL0  +=       1.581 * cos(0.36987184257  + 53764.87917843220 *T)
    sumL0  +=       1.589 * cos(1.33531241950  + 114.43928868521 *T)
    sumL0  +=       2.042 * cos(2.05385564074  + 45405.09568190280 *T)
    sumL0  +=       1.539 * cos(2.25468618977  + 120226.23016166000 *T)
    sumL0  +=       1.525 * cos(1.48231179113  + 26094.53170047421 *T)
    sumL0  +=       2.055 * cos(4.16315644175  + 52182.91983014920 *T)
    sumL0  +=       1.829 * cos(0.38846130335  + 522.57741809380 *T)
    sumL0  +=       2.105 * cos(4.04128506778  + 25234.70675982219 *T)
    sumL0  +=       1.480 * cos(1.87644604156  + 104881.30353139140 *T)
    sumL0  +=       1.460 * cos(1.01790185557  + 25455.11940226100 *T)
    sumL0  +=       1.453 * cos(0.34125222190  + 1052.26838318840 *T)
    sumL0  +=       1.449 * cos(3.39742463494  + 24505.94379329119 *T)
    sumL0  +=       1.480 * cos(4.55094492082  + 124778.18365871880 *T)
    sumL0  +=       1.481 * cos(3.62242942443  + 149.56319713460 *T)
    sumL0  +=       1.516 * cos(5.32327802440  + 129909.82474277640 *T)
    sumL0  +=       1.594 * cos(3.00887674765  + 77623.81213840858 *T)
    sumL0  +=       1.927 * cos(3.90378108197  + 74.78159856730 *T)
    sumL0  +=       1.373 * cos(4.32419381493  + 80482.46652890938 *T)
    sumL0  +=       1.419 * cos(3.90109225299  + 7880.08915333899 *T)
    sumL0  +=       1.525 * cos(4.62451312025  + 26081.27458267419 *T)
    sumL0  +=       1.567 * cos(0.65981123945  + 157636.79740153858 *T)
    sumL0  +=       1.725 * cos(4.40112128069  + 316.39186965660 *T)
    sumL0  +=       1.263 * cos(4.96320745251  + 74821.13447975718 *T)
    sumL0  +=       1.472 * cos(5.56006656586  + 95247.70557217918 *T)
    sumL0  +=       1.315 * cos(1.83254115004  + 76144.94556434419 *T)
    sumL0  +=       1.439 * cos(0.83286166370  + 55618.38122811380 *T)
    sumL0  +=       1.355 * cos(3.15837061982  + 1066.49547719000 *T)
    sumL0  +=       1.224 * cos(3.24032680768  + 162188.75089859738 *T)
    sumL0  +=       1.390 * cos(5.13745290383  + 419.48464387520 *T)
    sumL0  +=       1.310 * cos(5.28250636680  + 25619.93815121980 *T)
    sumL0  +=       1.477 * cos(2.40227455027  + 313054.83769889042 *T)
    sumL0  +=       1.252 * cos(5.65227031449  + 26013.12154300690 *T)
    sumL0  +=       1.143 * cos(0.38707356937  + 42153.96900304900 *T)
    sumL0  +=       1.082 * cos(2.56550728704  + 26190.99591579279 *T)
    sumL0  +=       1.493 * cos(2.89973504649  + 52099.54021187280 *T)
    sumL0  +=       1.468 * cos(5.71507791250  + 24491.71669928959 *T)
    sumL0  +=       1.034 * cos(1.49544294440  + 151975.46535238638 *T)
    sumL0  +=       1.314 * cos(5.91505351846  + 115674.27666460120 *T)
    sumL0  +=       1.065 * cos(3.61081835123  + 206.18554843720 *T)
    sumL0  +=       1.053 * cos(5.17996026320  + 23754.70674870219 *T)
    sumL0  +=       0.964 * cos(2.05264296353  + 128850.44281258718 *T)
    sumL0  +=       1.032 * cos(3.39114761061  + 25551.09862947879 *T)
    sumL0  +=       1.065 * cos(3.38225524038  + 25863.55834587229 *T)
    sumL0  +=       1.153 * cos(5.76498787983  + 44181.27784112419 *T)
    sumL0  +=       0.960 * cos(5.74774250678  + 105410.99449648600 *T)
    sumL0  +=       1.137 * cos(6.20919445956  + 103396.01282468818 *T)
    sumL0  +=       0.914 * cos(4.28855434634  + 78256.59587772179 *T)
    sumL0  +=       0.917 * cos(3.27543718918  + 76.26607127560 *T)
    sumL0  +=       0.892 * cos(4.82397571946  + 78244.03966382280 *T)
    sumL0  +=       0.976 * cos(4.86442946387  + 25131.61398560359 *T)
    sumL0  +=       1.086 * cos(0.73424633411  + 33967.99229491319 *T)
    sumL0  +=       1.162 * cos(4.52128441898  + 78050.41032928458 *T)
    sumL0  +=       0.868 * cos(2.78848256837  + 52022.02747266360 *T)
    sumL0  +=       0.943 * cos(4.09736853870  + 26720.68688088739 *T)
    sumL0  +=       0.973 * cos(4.54018615747  + 24925.42843716640 *T)
    sumL0  +=       0.845 * cos(5.05414707498  + 25977.69682035479 *T)
    sumL0  +=       0.841 * cos(1.89412486884  + 130012.91751699500 *T)
    sumL0  +=       0.920 * cos(3.02504494542  + 52329.58509363319 *T)
    sumL0  +=       0.814 * cos(3.15330668609  + 48835.19385644859 *T)
    sumL0  +=       0.876 * cos(0.41318857467  + 108903.56606335558 *T)
    sumL0  +=       0.818 * cos(1.61477924287  + 40565.25432477420 *T)
    sumL0  +=       1.074 * cos(1.37296024678  + 23439.44831610119 *T)
    sumL0  +=       0.812 * cos(2.82156935191  + 25984.81036735560 *T)
    sumL0  +=       0.775 * cos(2.86911510061  + 26727.80042788820 *T)
    sumL0  +=       0.755 * cos(5.44067645183  + 24609.03656750980 *T)
    sumL0  +=       0.921 * cos(4.72557901430  + 14477.35118320000 *T)
    sumL0  +=       0.914 * cos(4.70826349904  + 6681.22485339960 *T)
    sumL0  +=       0.749 * cos(1.80279541026  + 84546.78527471398 *T)
    sumL0  +=       0.907 * cos(0.64974692551  + 78283.37918562238 *T)
    sumL0  +=       0.854 * cos(5.05058435764  + 307.55762096960 *T)
    sumL0  +=       0.766 * cos(5.89942349817  + 71025.03383312259 *T)
    sumL0  +=       0.713 * cos(3.55192633098  + 78477.00852016058 *T)
    sumL0  +=       0.718 * cos(1.85061030321  + 100909.03762133140 *T)
    sumL0  +=       0.858 * cos(5.43959629234  + 536.80451209540 *T)
    sumL0  +=       0.794 * cos(3.77076518156  + 155418.04029735178 *T)
    sumL0  +=       0.848 * cos(2.08790526575  + 131395.11544947961 *T)
    sumL0  +=       0.622 * cos(3.14583461131  + 143961.26714946239 *T)
    sumL0  +=       0.623 * cos(4.58424507963  + 9103.90699411760 *T)
    sumL0  +=       0.623 * cos(5.10535773746  + 90829.86112470538 *T)
    sumL0  +=       0.706 * cos(4.19554833142  + 71582.48457132299 *T)
    sumL0  +=       0.738 * cos(0.15942404038  + 24712.12934172840 *T)
    sumL0  +=       0.621 * cos(1.32834420218  + 3328.13565628019 *T)
    sumL0  +=       0.714 * cos(1.88052234658  + 2199.08734328700 *T)
    sumL0  +=       0.678 * cos(6.01392909264  + 77197.21394753258 *T)
    sumL0  +=       0.650 * cos(5.51227348491  + 29428.51556827400 *T)
    sumL0  +=       0.658 * cos(2.03299256393  + 51756.32163927320 *T)
    sumL0  +=       0.593 * cos(6.06807779961  + 114564.89811250778 *T)
    sumL0  +=       0.666 * cos(0.70706334861  + 26162.68474014150 *T)
    sumL0  +=       0.597 * cos(5.15771478155  + 333.85594076940 *T)
    sumL0  +=       0.616 * cos(1.94902289932  + 78267.59076008058 *T)
    sumL0  +=       0.774 * cos(2.84206828086  + 13655.86049117640 *T)
    sumL0  +=       0.698 * cos(5.60747842278  + 51742.09454527159 *T)
    sumL0  +=       0.557 * cos(2.07529602574  + 181555.94006083015 *T)
    sumL0  +=       0.547 * cos(3.17747066781  + 90695.75207512038 *T)
    sumL0  +=       0.541 * cos(4.20816350943  + 28421.09953444620 *T)
    sumL0  +=       0.639 * cos(5.55908134374  + 78114.14622758799 *T)
    sumL0  +=       0.529 * cos(3.60209913515  + 1911.19948321720 *T)
    sumL0  +=       0.620 * cos(4.87816246784  + 25771.51127191760 *T)
    sumL0  +=       0.569 * cos(3.56250435364  + 111122.32316754240 *T)
    sumL0  +=       0.510 * cos(4.20331823730  + 28206.66700195260 *T)
    sumL0  +=       0.539 * cos(5.42639519551  + 93028.94846799241 *T)
    sumL0  +=       0.659 * cos(3.45407389663  + 26610.48055966799 *T)
    sumL0  +=       0.613 * cos(3.52133069078  + 78259.82808936459 *T)
    sumL0  +=       0.594 * cos(1.16299449628  + 26404.29501123080 *T)
    sumL0  +=       0.576 * cos(0.81617266730  + 51322.60990139639 *T)
    sumL0  +=       0.497 * cos(5.37540191935  + 26037.90651966999 *T)
    sumL0  +=       0.663 * cos(2.63246066036  + 77211.44104153418 *T)
    sumL0  +=       0.487 * cos(1.96691664683  + 26507.38778544939 *T)
    sumL0  +=       0.474 * cos(3.65243004234  + 124156.43985787958 *T)
    sumL0  +=       0.550 * cos(5.81040334783  + 1731.12235293260 *T)
    sumL0  +=       0.472 * cos(3.79351701945  + 52643.77127350280 *T)
    sumL0  +=       0.467 * cos(4.50175920356  + 110.20632121940 *T)
    sumL0  +=       0.593 * cos(3.04818872924  + 12725.45343477500 *T)
    sumL0  +=       0.552 * cos(0.68733268561  + 26137.89976347840 *T)
    sumL0  +=       0.525 * cos(0.87988157932  + 78270.82297172339 *T)
    sumL0  +=       0.600 * cos(0.33902061394  + 50483.64061364600 *T)
    sumL0  +=       0.443 * cos(3.33385484899  + 78690.30761559859 *T)
    sumL0  +=       0.441 * cos(1.31034191525  + 76044.95232053580 *T)
    sumL0  +=       0.588 * cos(2.37074288411  + 121335.60871375341 *T)
    sumL0  +=       0.427 * cos(0.14810668871  + 50593.84693486539 *T)
    sumL0  +=       0.413 * cos(3.33346664325  + 79852.78232000639 *T)
    sumL0  +=       0.403 * cos(5.20705536949  + 146314.13330323418 *T)
    sumL0  +=       0.524 * cos(5.02853960393  + 71492.99882347698 *T)
    sumL0  +=       0.418 * cos(5.99113644745  + 103711.71527998280 *T)
    sumL0  +=       0.388 * cos(4.44142037364  + 29416.03879785439 *T)
    sumL0  +=       0.390 * cos(4.02400157761  + 51543.02254383520 *T)
    sumL0  +=       0.442 * cos(2.37158178341  + 111590.28815789679 *T)
    sumL0  +=       0.375 * cos(4.47565026408  + 52182.43484204840 *T)
    sumL0  +=       0.387 * cos(4.83640155029  + 130969.20667296558 *T)
    sumL0  +=       0.390 * cos(1.20415517897  + 150866.08680029298 *T)
    sumL0  +=       0.416 * cos(3.60399872842  + 183724.70054311279 *T)
    sumL0  +=       0.407 * cos(0.52938530500  + 433.71173787680 *T)
    sumL0  +=       0.439 * cos(2.45401498182  + 50579.61984086379 *T)
    sumL0  +=       0.412 * cos(0.50186886239  + 27999.10262479140 *T)
    sumL0  +=       0.391 * cos(0.68517124695  + 134.10904958500 *T)
    sumL0  +=       0.398 * cos(1.99594942261  + 155997.72788435058 *T)
    sumL0  +=       0.346 * cos(3.40308148193  + 27140.17152476259 *T)
    sumL0  +=       0.422 * cos(5.83293462268  + 38813.35657634920 *T)
    sumL0  +=       0.342 * cos(0.83679406818  + 25764.39772491679 *T)
    sumL0  +=       0.368 * cos(3.80659045949  + 81706.28436968799 *T)
    sumL0  +=       0.355 * cos(1.99286044781  + 1089.70879119360 *T)
    sumL0  +=       0.462 * cos(3.33451630664  + 25440.89230825939 *T)
    sumL0  +=       0.330 * cos(0.37013669364  + 26237.46633870879 *T)
    sumL0  +=       0.330 * cos(6.19331560111  + 188276.65404017159 *T)
    sumL0  +=       0.407 * cos(2.17141736652  + 641.41356173899 *T)
    sumL0  +=       0.321 * cos(5.02532002650  + 102232.84870591838 *T)
    sumL0  +=       0.330 * cos(5.53086966080  + 10021.83728009940 *T)
    sumL0  +=       0.336 * cos(2.49163362426  + 25344.91308104160 *T)
    sumL0  +=       0.381 * cos(5.90928705861  + 78187.44335344699 *T)
    sumL0  +=       0.316 * cos(2.42532836233  + 52101.02468458109 *T)
    sumL0  +=       0.331 * cos(4.38897286971  + 26202.34243025941 *T)
    sumL0  +=       0.401 * cos(1.99380933624  + 49842.60989027639 *T)
    sumL0  +=       0.422 * cos(0.46556669940  + 1.48447270830 *T)
    sumL0  +=       0.310 * cos(6.18819197073  + 27154.39861876420 *T)
    sumL0  +=       0.302 * cos(4.68575942793  + 25565.32572348040 *T)
    sumL0  +=       0.296 * cos(1.70818196194  + 25973.46385288896 *T)
    sumL0  +=       0.375 * cos(5.31085497894  + 339142.74084046454 *T)
    sumL0  +=       0.329 * cos(5.32609571265  + 853.19638175200 *T)
    sumL0  +=       0.299 * cos(5.81101886868  + 1692.16566950240 *T)
    sumL0  +=       0.378 * cos(1.33745475273  + 52169.17772424839 *T)
    sumL0  +=       0.293 * cos(4.43993559835  + 579.68758699880 *T)
    sumL0  +=       0.276 * cos(4.45147472194  + 178063.36849396059 *T)
    sumL0  +=       0.283 * cos(6.19860164015  + 25881.71759313700 *T)
    sumL0  +=       0.315 * cos(1.95956643364  + 51707.84129279399 *T)
    sumL0  +=       0.337 * cos(4.21457120705  + 34282.17847478280 *T)
    sumL0  +=       0.334 * cos(3.00354887525  + 23866.04650697719 *T)
    sumL0  +=       0.268 * cos(1.34437493520  + 35191.81013569180 *T)
    sumL0  +=       0.306 * cos(5.94445231410  + 64607.84893354619 *T)
    sumL0  +=       0.283 * cos(0.96234220412  + 1639.06951718800 *T)
    sumL0  +=       0.287 * cos(1.28099267608  + 51013.33157874059 *T)
    sumL0  +=       0.269 * cos(0.55818949975  + 3462.24470586520 *T)
    sumL0  +=       0.274 * cos(5.61047457506  + 52278.89905736699 *T)
    sumL0  +=       0.352 * cos(0.49071599781  + 26294.08869001139 *T)
    sumL0  +=       0.358 * cos(1.23360990528  + 1223.81784077861 *T)
    sumL0  +=       0.343 * cos(4.29256939583  + 49527.35145767539 *T)
    sumL0  +=       0.255 * cos(5.03019684368  + 154938.34595416137 *T)
    sumL0  +=       0.297 * cos(2.87172991159  + 129483.91596626240 *T)
    sumL0  +=       0.252 * cos(0.09044630682  + 13541.42120249119 *T)
    sumL0  +=       0.301 * cos(1.22789205624  + 104138.31347085880 *T)
    sumL0  +=       0.274 * cos(3.67792692372  + 68241.87214462319 *T)
    sumL0  +=       0.321 * cos(2.49914111769  + 141762.17980617538 *T)
    sumL0  +=       0.290 * cos(1.02241335068  + 8194.27533320860 *T)
    sumL0  +=       0.240 * cos(0.99284920524  + 104344.49901929598 *T)
    sumL0  +=       0.278 * cos(0.53151576617  + 949.17560896980 *T)
    sumL0  +=       0.249 * cos(2.39787134441  + 131498.89763806018 *T)
    sumL0  +=       0.274 * cos(2.09560778105  + 26624.70765366959 *T)
    sumL0  +=       0.240 * cos(3.14668052954  + 23976.25282819660 *T)
    sumL0  +=       0.239 * cos(0.85198559240  + 52808.59002246159 *T)
    sumL0  +=       0.266 * cos(0.18324709351  + 2168.76048228260 *T)
    sumL0  +=       0.226 * cos(4.85701184113  + 156100.82065856917 *T)
    sumL0  +=       0.228 * cos(2.19799888298  + 50696.93970908399 *T)
    sumL0  +=       0.231 * cos(1.49522976308  + 104331.94280539700 *T)
    sumL0  +=       0.245 * cos(4.68642861263  + 5327.47610838280 *T)
    sumL0  +=       0.281 * cos(1.63579312063  + 51219.51712717779 *T)
    sumL0  +=       0.219 * cos(5.75893553103  + 78109.93061423779 *T)
    sumL0  +=       0.233 * cos(5.99853185125  + 78417.48823520739 *T)
    sumL0  +=       0.227 * cos(6.02396548637  + 1581.95934828300 *T)
    sumL0  +=       0.218 * cos(5.42819370607  + 18207.81398823521 *T)
    sumL0  +=       0.220 * cos(3.72017611382  + 7994.52844202420 *T)
    sumL0  +=       0.271 * cos(0.20833686608  + 51639.00177105299 *T)
    sumL0  +=       0.244 * cos(2.33639123341  + 25138.72753260440 *T)
    sumL0  +=       0.254 * cos(5.46143104532  + 621.74380083920 *T)
    sumL0  +=       0.209 * cos(0.01908528531  + 26164.16921284980 *T)
    sumL0  +=       0.208 * cos(6.18336687237  + 74923.09699802278 *T)
    sumL0  +=       0.274 * cos(0.15423898790  + 51951.46148744649 *T)
    sumL0  +=       0.212 * cos(1.95557421945  + 735.87651353180 *T)
    sumL0  +=       0.236 * cos(4.28368070837  + 24079.34560241519 *T)
    sumL0  +=       0.206 * cos(5.86972946528  + 52072.71350892979 *T)
    sumL0  +=       0.201 * cos(1.61226877070  + 25241.82030682300 *T)
    sumL0  +=       0.235 * cos(3.60446251088  + 104371.28232719658 *T)
    sumL0  +=       0.229 * cos(3.66187363672  + 1596.18644228460 *T)
    sumL0  +=       0.198 * cos(5.85298126352  + 52815.70356946240 *T)
    sumL0  +=       0.220 * cos(1.39541547612  + 32769.12799497380 *T)
    sumL0  +=       0.229 * cos(3.36265291664  + 134991.46920492980 *T)
    sumL0  +=       0.192 * cos(1.03510456065  + 25867.49049913539 *T)
    sumL0  +=       0.199 * cos(4.78126289270  + 110634.68841628821 *T)
    sumL0  +=       0.191 * cos(4.24110997660  + 14.22709400160 *T)
    sumL0  +=       0.210 * cos(5.44710702022  + 147423.51185532758 *T)
    sumL0  +=       0.229 * cos(0.76266531820  + 220.41264243880 *T)
    sumL0  +=       0.185 * cos(0.24752751791  + 104564.91166173480 *T)
    sumL0  +=       0.211 * cos(1.81826670820  + 52065.59996192899 *T)
    sumL0  +=       0.182 * cos(2.31946183497  + 9384.84100807520 *T)
    sumL0  +=       0.207 * cos(0.77329376239  + 917.93028598180 *T)
    sumL0  +=       0.190 * cos(0.91354241432  + 97670.38771289718 *T)
    sumL0  +=       0.184 * cos(3.32299343484  + 30.32686100440 *T)
    sumL0  +=       0.209 * cos(0.41150840785  + 181505.94343892601 *T)
    sumL0  +=       0.225 * cos(5.06158596891  + 157483.01859105378 *T)
    sumL0  +=       0.184 * cos(2.71529935065  + 103285.11708910680 *T)
    sumL0  +=       0.191 * cos(1.96528922573  + 314.18617986960 *T)
    sumL0  +=       0.182 * cos(4.30249539122  + 1884.90116341740 *T)
    sumL0  +=       0.192 * cos(4.82097876700  + 126996.94076290558 *T)
    sumL0  +=       0.188 * cos(0.98150502881  + 26049.77010593640 *T)
    sumL0  +=       0.184 * cos(0.67028697590  + 106570.36967048360 *T)
    sumL0  +=       0.167 * cos(5.71084692524  + 96357.08412427260 *T)
    sumL0  +=       0.184 * cos(2.33710037381  + 77829.99768684579 *T)
    sumL0  +=       0.201 * cos(2.57648669602  + 97112.93697469680 *T)
    sumL0  +=       0.176 * cos(3.20466399658  + 39629.32434406539 *T)
    sumL0  +=       0.169 * cos(2.30516098722  + 101.96251826560 *T)
    sumL0  +=       0.163 * cos(6.11134623776  + 170049.17029103660 *T)
    sumL0  +=       0.195 * cos(5.18276528052  + 99024.13645791401 *T)
    sumL0  +=       0.165 * cos(6.00583599609  + 224.34479570190 *T)
    sumL0  +=       0.167 * cos(3.75645242744  + 52250.58788171570 *T)
    sumL0  +=       0.172 * cos(4.28474997892  + 25780.34552060460 *T)
    sumL0  +=       0.216 * cos(3.20425967752  + 50800.03248330259 *T)
    sumL0  +=       0.157 * cos(1.61810971101  + 24815.22211594700 *T)
    sumL0  +=       0.163 * cos(1.77694235279  + 116917.76426627958 *T)
    sumL0  +=       0.156 * cos(1.94644057334  + 26421.75908234360 *T)
    sumL0  +=       0.164 * cos(3.81487619480  + 77410.51304297059 *T)
    sumL0  +=       0.158 * cos(4.95279229085  + 35472.74414964940 *T)
    sumL0  +=       0.150 * cos(5.03580916221  + 207643.84320240439 *T)
    sumL0  +=       0.208 * cos(2.59008696784  + 27819.02549450680 *T)
    sumL0  +=       0.180 * cos(5.62518388437  + 103299.34418310841 *T)
    sumL0  +=       0.165 * cos(2.25945126594  + 104202.04936916218 *T)
    sumL0  +=       0.148 * cos(3.94625893131  + 24601.92302050899 *T)
    sumL0  +=       0.149 * cos(4.49766341520  + 956.28915597060 *T)
    sumL0  +=       0.157 * cos(1.62995234425  + 51859.41441349179 *T)
    sumL0  +=       0.166 * cos(0.21228205353  + 52698.38370124219 *T)
    sumL0  +=       0.151 * cos(2.75061488385  + 140652.80125408198 *T)
    sumL0  +=       0.161 * cos(1.72952408524  + 2648.45482547300 *T)
    sumL0  +=       0.169 * cos(2.61298183547  + 26086.41866886590 *T)
    sumL0  +=       0.170 * cos(0.78905798191  + 72936.23331633979 *T)
    sumL0  +=       0.177 * cos(5.06009562295  + 77844.22478084739 *T)
    sumL0  +=       0.142 * cos(2.61984020842  + 24292.64469785319 *T)
    sumL0  +=       0.150 * cos(4.93124201089  + 28286.99048486120 *T)
    sumL0  +=       0.175 * cos(5.79468120941  + 39743.76363275060 *T)
    sumL0  +=       0.163 * cos(3.37032055762  + 76571.54375522019 *T)
    sumL0  +=       0.150 * cos(1.79967850988  + 3.93215326310 *T)
    sumL0  +=       0.159 * cos(4.90360163470  + 104355.49390165480 *T)
    sumL0  +=       0.149 * cos(4.20242344258  + 52492.19815280499 *T)
    sumL0  +=       0.146 * cos(1.31203861446  + 26198.10946279360 *T)
    sumL0  +=       0.162 * cos(2.25640253895  + 55516.41870984820 *T)
    sumL0  +=       0.178 * cos(1.81205070225  + 26395.46076254379 *T)
    sumL0  +=       0.132 * cos(0.90685777764  + 54294.57014352679 *T)
    sumL0  +=       0.142 * cos(1.70641940798  + 92741.06060792258 *T)
    sumL0  +=       0.138 * cos(0.64298043300  + 86143.79857806159 *T)
    sumL0  +=       0.125 * cos(0.32592941945  + 150244.34299945380 *T)
    sumL0  +=       0.131 * cos(0.91963253985  + 54509.00267602040 *T)
    sumL0  +=       0.144 * cos(4.17124383409  + 24822.33566294780 *T)
    sumL0  +=       0.135 * cos(3.86821614788  + 104358.72611329758 *T)
    sumL0  +=       0.159 * cos(4.72016316702  + 4083.98850670440 *T)
    sumL0  +=       0.122 * cos(2.08133407370  + 52125.80966124419 *T)
    sumL0  +=       0.141 * cos(2.08765477131  + 119116.85160956658 *T)
    sumL0  +=       0.122 * cos(5.01211051965  + 52595.29092702359 *T)
    sumL0  +=       0.149 * cos(0.38744219837  + 487.63475125420 *T)
    sumL0  +=       0.121 * cos(3.30215904574  + 1263.15736257819 *T)
    sumL0  +=       0.138 * cos(1.04671626049  + 1083.08023229360 *T)
    sumL0  +=       0.157 * cos(5.17018916622  + 23962.02573419499 *T)
    sumL0  +=       0.163 * cos(3.26593103179  + 26089.38761428249 *T)
    sumL0  +=       0.130 * cos(0.56779128434  + 25936.85547173129 *T)
    sumL0  +=       0.160 * cos(0.19385295261  + 104347.73123093881 *T)
    sumL0  +=       0.116 * cos(2.69818513122  + 129799.61842155698 *T)
    sumL0  +=       0.113 * cos(4.26323214388  + 102132.85546211000 *T)
    sumL0  +=       0.115 * cos(1.35196569742  + 75615.25459924959 *T)
    sumL0  +=       0.135 * cos(1.70032124143  + 97580.90196505121 *T)
    sumL0  +=       0.121 * cos(0.20372434486  + 38.13303563780 *T)
    sumL0  +=       0.118 * cos(0.49434134122  + 78731.67441507700 *T)
    sumL0  +=       0.119 * cos(5.54562725890  + 76667.52298243798 *T)
    sumL0  +=       0.133 * cos(4.67446352847  + 2014.98167179780 *T)
    sumL0  +=       0.126 * cos(6.24694138116  + 636.99627202420 *T)
    sumL0  +=       0.120 * cos(3.20442216756  + 76681.75007643958 *T)
    sumL0  +=       0.125 * cos(3.65255045892  + 29550.14784743939 *T)
    sumL0  +=       0.118 * cos(5.26052652437  + 26729.31670331319 *T)
    sumL0  +=       0.113 * cos(6.26688921593  + 104778.21075717278 *T)
    sumL0  +=       0.134 * cos(1.56524046563  + 25939.82441714789 *T)
    sumL0  +=       0.114 * cos(3.52566774494  + 24356.78078864160 *T)
    sumL0  +=       0.120 * cos(1.72596426067  + 22759.76748529401 *T)
    sumL0  +=       0.145 * cos(3.61638421570  + 52225.80290505260 *T)
    sumL0  +=       0.107 * cos(5.72285490406  + 24176.70365835700 *T)
    sumL0  +=       0.112 * cos(0.28882268029  + 209812.60368468691 *T)
    sumL0  +=       0.121 * cos(0.11234419157  + 51528.79544983359 *T)
    sumL0  +=       0.120 * cos(4.17769370414  + 25754.04720080480 *T)
    sumL0  +=       0.105 * cos(1.88137155058  + 172402.03644480839 *T)
    sumL0  +=       0.106 * cos(6.27133163237  + 105940.68546158058 *T)
    sumL0  +=       0.104 * cos(3.85061395100  + 1478.86657406440 *T)
    sumL0  +=       0.108 * cos(2.96292200098  + 1375.77379984580 *T)
    sumL0  +=       0.132 * cos(4.60210460022  + 11610.55195837420 *T)
    sumL0  +=       0.114 * cos(5.20612177122  + 137678.19129947099 *T)
    sumL0  +=       0.107 * cos(0.00557717441  + 13362.44970679920 *T)
    sumL0  +=       0.101 * cos(0.73934409276  + 77630.92568540938 *T)
    sumL0  +=       0.097 * cos(1.19155686364  + 78270.33798362259 *T)
    sumL0  +=       0.100 * cos(0.25974575144  + 160.40736938481 *T)
    sumL0  +=       0.105 * cos(0.03739853961  + 180.07713028460 *T)
    sumL0  +=       0.114 * cos(5.13609864972  + 1135.67687189320 *T)
    sumL0  +=       0.101 * cos(4.14323258440  + 176953.98994186718 *T)
    sumL0  +=       0.095 * cos(0.49488385937  + 107794.18751126219 *T)
    sumL0  +=       0.095 * cos(1.54869651436  + 25973.50403466079 *T)
    sumL0  +=       0.100 * cos(1.51228088427  + 157057.10981453978 *T)
    sumL0  +=       0.106 * cos(4.57237246234  + 26073.67604757259 *T)
    sumL0  +=       0.094 * cos(3.53697338405  + 26521.61487945100 *T)
    sumL0  +=       0.091 * cos(3.13017596287  + 64901.25971792339 *T)
    sumL0  +=       0.088 * cos(5.20234335205  + 1485.29070670320 *T)
    sumL0  +=       0.116 * cos(5.96693811684  + 24182.43837663380 *T)
    sumL0  +=       0.099 * cos(2.12002549240  + 66653.15746634839 *T)
    sumL0  +=       0.105 * cos(4.04764282865  + 27726.97265876220 *T)
    sumL0  +=       0.094 * cos(2.62198655221  + 104275.34649502118 *T)
    sumL0  +=       0.106 * cos(2.06629215696  + 125112.03959948818 *T)
    sumL0  +=       0.088 * cos(3.88240703310  + 51852.30086649099 *T)
    sumL0  +=       0.085 * cos(2.83285123847  + 214364.55718174577 *T)
    sumL0  +=       0.083 * cos(5.44599307753  + 78188.92782615528 *T)
    sumL0  +=       0.083 * cos(3.35497157412  + 52325.36948028299 *T)
    sumL0  +=       0.104 * cos(4.95116515346  + 182085.63102592481 *T)
    sumL0  +=       0.110 * cos(6.04814955763  + 49953.94964855139 *T)
    sumL0  +=       0.093 * cos(1.00014028766  + 9745.32055585660 *T)
    sumL0  +=       0.081 * cos(4.03821735964  + 132658.27281205778 *T)
    sumL0  +=       0.084 * cos(0.11234538521  + 53228.07466633679 *T)
    sumL0  +=       0.094 * cos(5.55962830464  + 51432.81622261579 *T)
    sumL0  +=       0.080 * cos(6.13595673095  + 137210.22630911658 *T)
    sumL0  +=       0.080 * cos(2.72256616799  + 25042.74830538660 *T)
    sumL0  +=       0.080 * cos(4.29137248560  + 77101.23472031478 *T)
    sumL0  +=       0.085 * cos(0.97100574775  + 25788.77674730500 *T)
    sumL0  +=       0.082 * cos(1.89501710792  + 128320.75184749260 *T)
    sumL0  +=       0.077 * cos(1.16185579398  + 204151.27163553479 *T)
    sumL0  +=       0.092 * cos(4.44154858356  + 26091.83529483729 *T)
    sumL0  +=       0.078 * cos(3.73813140716  + 26222.01219115920 *T)
    sumL0  +=       0.076 * cos(4.28770981047  + 86457.98475793119 *T)
    sumL0  +=       0.076 * cos(2.88868549481  + 53242.30176033840 *T)
    sumL0  +=       0.104 * cos(4.32540467122  + 23888.81579828719 *T)
    sumL0  +=       0.094 * cos(4.32453647728  + 78257.08086582259 *T)
    sumL0  +=       0.074 * cos(4.96325610801  + 26823.77965510599 *T)
    sumL0  +=       0.077 * cos(1.07428657729  + 50167.24874398939 *T)
    sumL0  +=       0.075 * cos(2.56662960973  + 19202.75325164339 *T)
    sumL0  +=       0.099 * cos(6.00141798836  + 19958.60610206760 *T)
    sumL0  +=       0.087 * cos(3.55579419373  + 52381.99183158559 *T)
    sumL0  +=       0.078 * cos(1.00822618263  + 25352.02662804239 *T)
    sumL0  +=       0.088 * cos(3.52269656514  + 60055.89543648739 *T)
    sumL0  +=       0.078 * cos(5.82066892741  + 155571.81910783658 *T)
    sumL0  +=       0.076 * cos(2.96046756402  + 25024.58905812189 *T)
    sumL0  +=       0.076 * cos(4.24421235710  + 130226.21661243298 *T)
    sumL0  +=       0.076 * cos(2.22645638378  + 36109.74042167360 *T)
    sumL0  +=       0.072 * cos(2.23348468194  + 173511.41499690176 *T)
    sumL0  +=       0.079 * cos(1.13204602827  + 52290.24557183361 *T)
    sumL0  +=       0.074 * cos(1.55328520138  + 51653.22886505459 *T)
    sumL0  +=       0.071 * cos(2.26106298981  + 159.30173479220 *T)
    sumL0  +=       0.073 * cos(2.05345178568  + 26941.09952332620 *T)
    sumL0  +=       0.095 * cos(1.93550514722  + 365230.64398203877 *T)
    sumL0  +=       0.070 * cos(4.74315846161  + 52061.36699446317 *T)
    sumL0  +=       0.085 * cos(1.31607641314  + 26667.59072857300 *T)
    sumL0  +=       0.076 * cos(5.58726680703  + 13675.53025207620 *T)
    sumL0  +=       0.092 * cos(3.24599684677  + 28256.66362385679 *T)
    sumL0  +=       0.067 * cos(1.15071488774  + 55503.94193942859 *T)
    sumL0  +=       0.068 * cos(2.92993581219  + 23919.14265929160 *T)
    sumL0  +=       0.067 * cos(1.77035470445  + 181026.24909573561 *T)
    sumL0  +=       0.066 * cos(2.59648990707  + 27780.06881107659 *T)
    sumL0  +=       0.082 * cos(5.36282814327  + 167850.08294774959 *T)
    sumL0  +=       0.064 * cos(3.99165913279  + 130432.40216087016 *T)
    sumL0  +=       0.069 * cos(5.13777502156  + 52712.61079524379 *T)
    sumL0  +=       0.063 * cos(0.54837259436  + 34082.43158359840 *T)
    sumL0  +=       0.065 * cos(3.89976847376  + 27005.83342755599 *T)
    sumL0  +=       0.073 * cos(6.19252091852  + 50064.15596977079 *T)
    sumL0  +=       0.066 * cos(1.91671959572  + 102018.41617342478 *T)
    sumL0  +=       0.082 * cos(0.87824651640  + 25446.48957983520 *T)
    sumL0  +=       0.063 * cos(3.87624799927  + 78896.49316403578 *T)
    sumL0  +=       0.062 * cos(1.55377407031  + 182188.72380014337 *T)
    sumL0  +=       0.062 * cos(4.71223563105  + 103.78218858060 *T)
    sumL0  +=       0.087 * cos(3.57434964589  + 27037.07875054399 *T)
    sumL0  +=       0.077 * cos(3.43626093365  + 54087.00576636560 *T)
    sumL0  +=       0.078 * cos(4.92073893807  + 77795.74443436819 *T)
    sumL0  +=       0.071 * cos(2.76139796167  + 11.04570026390 *T)
    sumL0  +=       0.064 * cos(5.37974575072  + 51226.63067417859 *T)
    sumL0  +=       0.082 * cos(3.78003355754  + 26308.31578401300 *T)
    sumL0  +=       0.062 * cos(2.70733081732  + 104505.39137678158 *T)
    sumL0  +=       0.071 * cos(5.05097183202  + 27177.61193276780 *T)
    sumL0  +=       0.070 * cos(4.29342266024  + 27311.72098235281 *T)
    sumL0  +=       0.068 * cos(2.34136362165  + 78366.80219894118 *T)
    sumL0  +=       0.060 * cos(4.90888421377  + 25508.21555457540 *T)
    sumL0  +=       0.068 * cos(6.04917764171  + 647.01083331480 *T)
    sumL0  +=       0.058 * cos(2.46745404038  + 104197.83375581198 *T)
    sumL0  +=       0.067 * cos(2.96573339768  + 51969.62073471119 *T)
    sumL0  +=       0.066 * cos(5.33871332672  + 157586.80077963436 *T)
    sumL0  +=       0.058 * cos(2.17528608366  + 323.50541665740 *T)
    sumL0  +=       0.062 * cos(4.44991435917  + 130419.84594697120 *T)
    sumL0  +=       0.063 * cos(0.27396211581  + 130459.18546877075 *T)
    sumL0  +=       0.063 * cos(5.22327184571  + 76784.84285065818 *T)
    sumL0  +=       0.064 * cos(4.64108529425  + 61279.71327726600 *T)
    sumL0  +=       0.062 * cos(1.84001159422  + 183570.92173262799 *T)
    sumL0  +=       0.055 * cos(1.66692917947  + 31415.37924995700 *T)
    sumL0  +=       0.058 * cos(4.06522187846  + 32132.13172294960 *T)
    sumL0  +=       0.057 * cos(2.07569387080  + 17893.62780836560 *T)
    sumL0  +=       0.055 * cos(2.49038254003  + 22625.65843570900 *T)
    sumL0  +=       0.077 * cos(6.17301946895  + 25032.45336464809 *T)
    sumL0  +=       0.068 * cos(4.68765980625  + 77307.42026875199 *T)
    sumL0  +=       0.054 * cos(4.63708190151  + 51329.72344839720 *T)
    sumL0  +=       0.055 * cos(0.21233757772  + 2221.85663459700 *T)
    sumL0  +=       0.068 * cos(2.36358626958  + 25953.79409198919 *T)
    sumL0  +=       0.053 * cos(2.88564903323  + 101011.00013959700 *T)
    sumL0  +=       0.062 * cos(1.19998944516  + 26083.97098831109 *T)
    sumL0  +=       0.054 * cos(3.90599349467  + 123758.29085447140 *T)
    sumL0  +=       0.052 * cos(2.60595997150  + 78160.61665050399 *T)
    sumL0  +=       0.072 * cos(0.83460657204  + 22909.75735100660 *T)
    sumL0  +=       0.066 * cos(3.17005690509  + 78039.36462902068 *T)
    sumL0  +=       0.052 * cos(2.13323028898  + 44295.71712980940 *T)
    sumL0  +=       0.061 * cos(1.63440457571  + 61.72597474020 *T)
    sumL0  +=       0.053 * cos(2.00748586733  + 24448.83362438620 *T)
    sumL0  +=       0.051 * cos(2.50978165702  + 78903.60671103658 *T)
    sumL0  +=       0.051 * cos(3.22110952745  + 130652.81480330898 *T)
    sumL0  +=       0.060 * cos(6.23105854697  + 76887.93562487679 *T)
    sumL0  +=       0.066 * cos(3.23596712512  + 77726.90491262719 *T)
    sumL0  +=       0.059 * cos(1.53485505891  + 6044.22858137540 *T)
    sumL0  +=       0.059 * cos(1.10559326249  + 26102.13023557580 *T)
    sumL0  +=       0.060 * cos(1.08408977503  + 27972.80430499159 *T)
    sumL0  +=       0.049 * cos(2.24548271996  + 425.90877651400 *T)
    sumL0  +=       0.055 * cos(3.33645161611  + 207593.84658050016 *T)
    sumL0  +=       0.067 * cos(0.59579109206  + 94329.77528619740 *T)
    sumL0  +=       0.056 * cos(4.05652507832  + 26010.15259759030 *T)
    sumL0  +=       0.056 * cos(1.49670493057  + 136722.59155786238 *T)
    sumL0  +=       0.049 * cos(4.72236375377  + 5193.36705879780 *T)
    sumL0  +=       0.049 * cos(2.99400480872  + 52252.07235442399 *T)
    sumL0  +=       0.049 * cos(4.15104806247  + 51955.39364070959 *T)
    sumL0  +=       0.050 * cos(5.31916782876  + 103917.90082841998 *T)
    sumL0  +=       0.051 * cos(4.32173710361  + 58857.03113654799 *T)
    sumL0  +=       0.050 * cos(0.42233074233  + 27684.08958385880 *T)
    sumL0  +=       0.066 * cos(5.06498799279  + 73.29712585900 *T)
    sumL0  +=       0.048 * cos(0.80745705976  + 2703.61615467560 *T)
    sumL0  +=       0.048 * cos(0.16882113823  + 24918.31489016559 *T)
    sumL0  +=       0.049 * cos(5.71984539931  + 50380.54783942739 *T)
    sumL0  +=       0.061 * cos(0.05575706789  + 161079.37234650398 *T)
    sumL0  +=       0.047 * cos(5.70016549134  + 129373.02023068098 *T)
    sumL0  +=       0.048 * cos(4.71754734372  + 50903.12525752120 *T)
    sumL0  +=       0.050 * cos(0.77120384679  + 50689.82616208319 *T)
    sumL0  +=       0.050 * cos(4.75557775921  + 143005.66740785379 *T)
    sumL0  +=       0.046 * cos(1.03065291695  + 51868.24866217880 *T)
    sumL0  +=       0.046 * cos(4.07474771884  + 52137.67324751059 *T)
    sumL0  +=       0.045 * cos(2.75496032307  + 196137.07343261081 *T)
    sumL0  +=       0.056 * cos(1.48667124221  + 30171.89164827860 *T)
    sumL0  +=       0.047 * cos(3.49884159195  + 26575.53789282839 *T)
    sumL0  +=       0.048 * cos(2.04489119557  + 27573.19384827740 *T)
    sumL0  +=       0.046 * cos(1.96812600979  + 27223.58001346740 *T)
    sumL0  +=       0.044 * cos(0.58466186909  + 1243.48760167840 *T)
    sumL0  +=       0.043 * cos(0.48582749664  + 78338.49102328988 *T)
    sumL0  +=       0.046 * cos(0.10663541022  + 102659.44689679438 *T)
    sumL0  +=       0.049 * cos(1.92223535597  + 50049.92887576919 *T)
    sumL0  +=       0.047 * cos(5.34811633125  + 846.08283475120 *T)
    sumL0  +=       0.046 * cos(2.32911930278  + 129387.24732468258 *T)
    sumL0  +=       0.056 * cos(1.46049477175  + 28102.88481337200 *T)
    sumL0  +=       0.050 * cos(5.62967300375  + 53906.92863608099 *T)
    sumL0  +=       0.046 * cos(5.09770472011  + 151199.94274106238 *T)
    sumL0  +=       0.043 * cos(2.30154744958  + 26709.64694241340 *T)
    sumL0  +=       0.051 * cos(2.88082546361  + 27669.86248985719 *T)
    sumL0  +=       0.055 * cos(5.55849793732  + 123200.84011627098 *T)
    sumL0  +=       0.044 * cos(0.48527820908  + 80174.90890793978 *T)
    sumL0  +=       0.048 * cos(2.59296696271  + 6720.71397934140 *T)
    sumL0  +=       0.044 * cos(0.24824938054  + 742.99006053260 *T)
    sumL0  +=       0.045 * cos(2.48949498300  + 65831.66677432480 *T)
    sumL0  +=       0.045 * cos(4.86503115391  + 52483.36390411799 *T)
    sumL0  +=       0.041 * cos(5.23419163545  + 130289.95251073640 *T)
    sumL0  +=       0.041 * cos(0.91885509326  + 50910.23880452200 *T)
    sumL0  +=       0.044 * cos(3.12424163301  + 26126.03617721200 *T)
    sumL0  +=       0.042 * cos(5.66855218492  + 166740.70439565618 *T)
    sumL0  +=       0.039 * cos(5.01739570944  + 26402.08932144380 *T)
    sumL0  +=       0.038 * cos(0.02313645928  + 2111.65031337760 *T)
    sumL0  +=       0.049 * cos(4.11697361923  + 25227.59321282139 *T)
    sumL0  +=       0.038 * cos(0.09191461136  + 26118.23000257860 *T)
    sumL0  +=       0.038 * cos(5.03146092380  + 24203.00197815680 *T)
    sumL0  +=       0.051 * cos(4.86305906640  + 78153.50310350319 *T)
    sumL0  +=       0.038 * cos(0.13003640396  + 27351.06050415239 *T)
    sumL0  +=       0.038 * cos(1.34984785206  + 70269.18098269838 *T)
    sumL0  +=       0.038 * cos(0.52032944033  + 103498.41618454478 *T)
    sumL0  +=       0.040 * cos(5.66117959398  + 52174.32181044009 *T)
    sumL0  +=       0.038 * cos(0.92511339052  + 112231.70171963578 *T)
    sumL0  +=       0.051 * cos(4.15223165568  + 27170.98337386779 *T)
    sumL0  +=       0.037 * cos(1.67308781547  + 54374.89362643540 *T)
    sumL0  +=       0.038 * cos(6.26592309401  + 202.25339517410 *T)
    sumL0  +=       0.036 * cos(3.23911910889  + 176332.24614102798 *T)
    sumL0  +=       0.037 * cos(1.90730795656  + 3308.46589538040 *T)
    sumL0  +=       0.037 * cos(3.03203144287  + 76041.85279012559 *T)
    sumL0  +=       0.038 * cos(3.17519743409  + 78786.28684281638 *T)
    sumL0  +=       0.035 * cos(3.90697621183  + 80382.47328510099 *T)
    sumL0  +=       0.038 * cos(2.00924199028  + 1162.47470440780 *T)
    sumL0  +=       0.036 * cos(1.80080871248  + 19336.86230122839 *T)
    sumL0  +=       0.035 * cos(4.66254825701  + 77947.31755506598 *T)
    sumL0  +=       0.043 * cos(1.57972686428  + 130443.39704322898 *T)
    sumL0  +=       0.035 * cos(0.74694549648  + 25985.94062330859 *T)
    sumL0  +=       0.040 * cos(2.11415671365  + 3492.57156686960 *T)
    sumL0  +=       0.038 * cos(4.65738726744  + 22003.91463486980 *T)
    sumL0  +=       0.037 * cos(0.77571838110  + 16703.06213349900 *T)
    sumL0  +=       0.035 * cos(5.00128829727  + 52509.66222391780 *T)
    sumL0  +=       0.047 * cos(0.02275392308  + 25014.29411738339 *T)
    sumL0  +=       0.043 * cos(3.15273137814  + 130435.63437251298 *T)
    sumL0  +=       0.034 * cos(3.05681901713  + 2820.83114412620 *T)
    sumL0  +=       0.037 * cos(4.68248751145  + 123668.80510662538 *T)
    sumL0  +=       0.034 * cos(5.73615227262  + 45290.65639321759 *T)
    sumL0  +=       0.040 * cos(2.76893017076  + 50270.34151820800 *T)
    sumL0  +=       0.038 * cos(6.01229835605  + 26057.57628056979 *T)
    sumL0  +=       0.034 * cos(6.25090722416  + 21716.02677480000 *T)
    sumL0  +=       0.034 * cos(0.95587257158  + 78580.10129437919 *T)
    sumL0  +=       0.033 * cos(3.35217476663  + 235900.50682626120 *T)
    sumL0  +=       0.032 * cos(2.35469869139  + 2383.19301477620 *T)
    sumL0  +=       0.037 * cos(3.80938441515  + 8989.46770543239 *T)
    sumL0  +=       0.043 * cos(2.27476214717  + 25169.97285559240 *T)
    sumL0  +=       0.032 * cos(4.14291864821  + 2686.72209454120 *T)
    sumL0  +=       0.037 * cos(0.04571185840  + 52177.29075585669 *T)
    sumL0  +=       0.032 * cos(1.53610678036  + 233731.74634397856 *T)
    sumL0  +=       0.038 * cos(3.98892541543  + 1300.82681820500 *T)
    sumL0  +=       0.042 * cos(5.10993958059  + 145204.75475114080 *T)
    sumL0  +=       0.033 * cos(2.91530565997  + 130866.11389874700 *T)
    sumL0  +=       0.043 * cos(1.77793563831  + 103932.12792242158 *T)
    sumL0  +=       0.035 * cos(5.35726747093  + 26189.86565983980 *T)
    sumL0  +=       0.035 * cos(0.65119403092  + 1535.97674296940 *T)
    sumL0  +=       0.044 * cos(3.08976502022  + 8014.19820292400 *T)
    sumL0  +=       0.031 * cos(0.90638189649  + 128220.75860368418 *T)
    sumL0  +=       0.042 * cos(3.37905134636  + 23549.65463732060 *T)
    sumL0  +=       0.038 * cos(0.38543000653  + 515.46387109300 *T)
    sumL0  +=       0.039 * cos(1.08681725797  + 25773.71696170459 *T)
    sumL0  +=       0.036 * cos(4.30448026071  + 52286.01260436779 *T)
    sumL0  +=       0.031 * cos(5.08183330827  + 78213.71280281838 *T)
    sumL0  +=       0.038 * cos(3.16432629306  + 91805.13062721380 *T)
    sumL0  +=       0.039 * cos(1.47494002110  + 4022.26253196420 *T)
    sumL0  +=       0.042 * cos(0.60192138814  + 24388.62392507099 *T)
    sumL0  +=       0.037 * cos(3.23606285260  + 28791.51929624980 *T)
    sumL0  +=       0.040 * cos(0.26646931277  + 78313.70604662679 *T)
    sumL0  +=       0.036 * cos(4.38588497247  + 81591.84508100279 *T)
    sumL0  +=       0.035 * cos(0.17056359345  + 6129.29703950660 *T)
    sumL0  +=       0.030 * cos(2.50459470137  + 102755.42612401220 *T)
    sumL0  +=       0.040 * cos(4.58870970073  + 26076.85744131030 *T)
    sumL0  +=       0.040 * cos(4.09419634913  + 35833.22369743080 *T)
    sumL0  +=       0.030 * cos(4.79837659477  + 28736.35796704720 *T)
    sumL0  +=       0.032 * cos(3.42931668217  + 24402.85101907260 *T)
    sumL0  +=       0.029 * cos(2.52212486456  + 1353.74874501680 *T)
    sumL0  +=       0.030 * cos(4.69352116507  + 24072.92146977640 *T)
    sumL0  +=       0.029 * cos(5.31070299201  + 199599.31813847597 *T)
    sumL0  +=       0.030 * cos(3.20350587499  + 23762.95373275860 *T)
    sumL0  +=       0.036 * cos(2.44362166097  + 23336.35554188260 *T)
    sumL0  +=       0.028 * cos(2.77231585227  + 26312.24793727610 *T)
    sumL0  +=       0.028 * cos(0.26775880889  + 50444.68393021580 *T)
    sumL0  +=       0.029 * cos(3.50869792127  + 15406.65260500879 *T)
    sumL0  +=       0.030 * cos(2.95993198433  + 132028.58860315481 *T)
    sumL0  +=       0.028 * cos(3.41609673663  + 133882.09065283637 *T)
    sumL0  +=       0.033 * cos(2.39929042753  + 61560.64729122359 *T)
    sumL0  +=       0.032 * cos(1.34796958829  + 37698.45509994840 *T)
    sumL0  +=       0.028 * cos(1.75026246832  + 78683.19406859778 *T)
    sumL0  +=       0.033 * cos(4.51080770987  + 3776.43088573480 *T)
    sumL0  +=       0.032 * cos(5.00665307820  + 29396.36903695460 *T)
    sumL0  +=       0.037 * cos(5.23066729388  + 81604.32185142238 *T)
    sumL0  +=       0.028 * cos(0.99927277404  + 1322.67764753140 *T)
    sumL0  +=       0.029 * cos(2.00159770770  + 25004.82290928060 *T)
    sumL0  +=       0.035 * cos(4.69198477732  + 22065.64060961000 *T)
    sumL0  +=       0.029 * cos(3.80697601044  + 24513.05734029200 *T)
    sumL0  +=       0.032 * cos(4.69401466159  + 27250.37784598199 *T)
    sumL0  +=       0.031 * cos(5.86715902954  + 5815.11085963699 *T)
    sumL0  +=       0.027 * cos(1.26702311366  + 27044.19229754480 *T)
    sumL0  +=       0.030 * cos(3.44351888035  + 104819.57755665120 *T)
    sumL0  +=       0.028 * cos(5.36078058708  + 412.37109687440 *T)
    sumL0  +=       0.028 * cos(5.73893599832  + 240452.46032332000 *T)
    sumL0  +=       0.034 * cos(5.19151430726  + 2008.55753915900 *T)
    sumL0  +=       0.030 * cos(1.73241634839  + 163766.09444104519 *T)
    sumL0  +=       0.027 * cos(5.30425762252  + 24864.08530079559 *T)
    sumL0  +=       0.032 * cos(1.12362303179  + 625.67019231240 *T)
    sumL0  +=       0.030 * cos(2.04171272941  + 193937.98608932376 *T)
    sumL0  +=       0.032 * cos(4.78663079749  + 75930.51303185058 *T)
    sumL0  +=       0.030 * cos(4.46891226150  + 25450.90686955000 *T)
    sumL0  +=       0.032 * cos(4.73906176949  + 48847.67062686820 *T)
    sumL0  +=       0.027 * cos(4.12514822062  + 104358.24112519680 *T)
    sumL0  +=       0.026 * cos(0.67609297065  + 35077.37084700659 *T)
    sumL0  +=       0.033 * cos(2.69651735555  + 25647.07785669659 *T)
    sumL0  +=       0.026 * cos(4.94767789631  + 209658.82487420217 *T)
    sumL0  +=       0.035 * cos(3.86705432105  + 25466.15934073500 *T)
    sumL0  +=       0.026 * cos(4.79912409089  + 198489.93958638256 *T)
    sumL0  +=       0.026 * cos(1.59971503257  + 31281.27020037200 *T)
    sumL0  +=       0.026 * cos(1.20008223797  + 153084.84390447981 *T)
    sumL0  +=       0.026 * cos(3.02443658098  + 26724.89941359840 *T)
    sumL0  +=       0.025 * cos(2.03567222767  + 52817.21984488739 *T)
    sumL0  +=       0.027 * cos(1.23585314301  + 53.09615231440 *T)
    sumL0  +=       0.026 * cos(5.53069670975  + 25657.37279743509 *T)
    sumL0  +=       0.028 * cos(0.93845909288  + 158746.17595363196 *T)
    sumL0  +=       0.025 * cos(1.46977800361  + 1272.68102562720 *T)
    sumL0  +=       0.026 * cos(2.45961936761  + 181659.72224941081 *T)
    sumL0  +=       0.034 * cos(2.39774146758  + 39763.43339365039 *T)
    sumL0  +=       0.029 * cos(6.23821353511  + 102769.65321801380 *T)
    sumL0  +=       0.029 * cos(4.90111965589  + 116783.65521669458 *T)
    sumL0  +=       0.025 * cos(3.02186465541  + 39450.35284837340 *T)
    sumL0  +=       0.027 * cos(3.77575599741  + 103718.82882698360 *T)
    sumL0  +=       0.027 * cos(3.92125046850  + 80596.90581759460 *T)
    sumL0  +=       0.023 * cos(0.26970481255  + 52609.51802102519 *T)
    sumL0  +=       0.023 * cos(5.98846059571  + 27463.67694142000 *T)
    sumL0  +=       0.025 * cos(1.22968381596  + 49976.71893986139 *T)
    sumL0  +=       0.023 * cos(5.57437029749  + 155887.52156313116 *T)
    sumL0  +=       0.023 * cos(5.99507985020  + 50007.04580086580 *T)
    sumL0  +=       0.023 * cos(1.44496012077  + 5687.63036895200 *T)
    sumL0  +=       0.024 * cos(0.46516912647  + 130446.62925487180 *T)
    sumL0  +=       0.023 * cos(0.94331115897  + 28774.62523611540 *T)
    sumL0  +=       0.027 * cos(0.80529908381  + 2698.45144737720 *T)
    sumL0  +=       0.027 * cos(0.27124610516  + 1265.56747862640 *T)
    sumL0  +=       0.023 * cos(0.89381228793  + 156314.11975400720 *T)
    sumL0  +=       0.023 * cos(2.96257779899  + 12546.48193908300 *T)
    sumL0  +=       0.022 * cos(5.42940490576  + 1485.98012106520 *T)
    sumL0  +=       0.022 * cos(1.03839316349  + 103189.13786188900 *T)
    sumL0  +=       0.023 * cos(1.35379542887  + 52161.57918914679 *T)
    sumL0  +=       0.026 * cos(6.01331218379  + 122444.98726584678 *T)
    sumL0  +=       0.030 * cos(2.66518277698  + 25600.26839032000 *T)
    sumL0  +=       0.023 * cos(4.07521268525  + 230239.17477710897 *T)
    sumL0  +=       0.024 * cos(0.64355332403  + 203041.89308344136 *T)
    sumL0  +=       0.022 * cos(1.39463873909  + 952.07662325960 *T)
    sumL0  +=       0.023 * cos(1.98852683178  + 7392.45440208480 *T)
    sumL0  +=       0.028 * cos(1.37381287166  + 309.27832265580 *T)
    sumL0  +=       0.022 * cos(2.15165551798  + 104276.83096772947 *T)
    sumL0  +=       0.025 * cos(4.93854060665  + 101703.15774082378 *T)
    sumL0  +=       0.023 * cos(3.12039946936  + 77616.69859140778 *T)
    sumL0  +=       0.028 * cos(0.40745832322  + 55638.05098901359 *T)
    sumL0  +=       0.024 * cos(0.44199264440  + 24285.53115085240 *T)
    sumL0  +=       0.023 * cos(1.33496997363  + 163298.12945069079 *T)
    sumL0  +=       0.021 * cos(4.01469381260  + 120417.67842777158 *T)
    sumL0  +=       0.024 * cos(0.63196904603  + 77940.20400806518 *T)
    sumL0  +=       0.022 * cos(5.70851933792  + 2840.50090502600 *T)
    sumL0  +=       0.024 * cos(4.45840068641  + 183145.01295611399 *T)
    sumL0  +=       0.023 * cos(3.62152465067  + 5635.03372935240 *T)
    sumL0  +=       0.021 * cos(3.04515069387  + 9123.57675501740 *T)
    sumL0  +=       0.021 * cos(1.96886020684  + 177287.84588263658 *T)
    sumL0  +=       0.023 * cos(0.59040250365  + 25672.35065096209 *T)
    sumL0  +=       0.021 * cos(3.40810269373  + 129586.31932611900 *T)
    sumL0  +=       0.021 * cos(3.09046433472  + 26267.98027185880 *T)
    sumL0  +=       0.024 * cos(1.94276984966  + 183674.70392120857 *T)
    sumL0  +=       0.021 * cos(3.98455358233  + 6191.02301424680 *T)
    sumL0  +=       0.028 * cos(0.93414213433  + 51841.95034237900 *T)
    sumL0  +=       0.020 * cos(4.59348701559  + 52061.40717623499 *T)
    sumL0  +=       0.027 * cos(4.61439897742  + 52027.72755872209 *T)
    sumL0  +=       0.020 * cos(0.48128544481  + 52309.91533273340 *T)
    sumL0  +=       0.020 * cos(3.31281154655  + 26248.31051095900 *T)
    sumL0  +=       0.023 * cos(1.67970146036  + 23113.29318696320 *T)
    sumL0  +=       0.020 * cos(2.44057473344  + 50264.60679993120 *T)
    sumL0  +=       0.025 * cos(5.43422956232  + 2.96894541660 *T)
    sumL0  +=       0.026 * cos(3.61410457829  + 52024.75861330549 *T)
    sumL0  +=       0.020 * cos(0.60321200025  + 138319.60486120995 *T)
    sumL0  +=       0.020 * cos(2.62628682595  + 164.43591058941 *T)
    sumL0  +=       0.020 * cos(0.65324774200  + 156520.30530244438 *T)
    sumL0  +=       0.027 * cos(4.74395008878  + 391318.54712361295 *T)
    sumL0  +=       0.021 * cos(4.45035214874  + 25862.07387316400 *T)
    sumL0  +=       0.021 * cos(1.54947465846  + 148.07872442630 *T)
    sumL0  +=       0.027 * cos(5.13202499307  + 12382.04602849360 *T)
    sumL0  +=       0.020 * cos(0.68567124507  + 42.88307490340 *T)
    sumL0  +=       0.019 * cos(1.92072938625  + 78800.51393681798 *T)
    sumL0  +=       0.027 * cos(2.48188330716  + 25665.23710396129 *T)
    sumL0  +=       0.027 * cos(1.27457874648  + 7860.41939243920 *T)
    sumL0  +=       0.022 * cos(2.29809961839  + 25865.04281858059 *T)
    sumL0  +=       0.019 * cos(4.09473921906  + 76255.15188556358 *T)
    sumL0  +=       0.020 * cos(4.67174071386  + 207114.15223730975 *T)
    sumL0  +=       0.022 * cos(2.63315058351  + 7768.36655669460 *T)
    sumL0  +=       0.020 * cos(1.24271716091  + 52179.73843641149 *T)
    sumL0  +=       0.020 * cos(6.11431848045  + 24824.74577899600 *T)
    sumL0  +=       0.023 * cos(4.18071159026  + 2125.87740737920 *T)
    sumL0  +=       0.021 * cos(0.74169140667  + 1795.25844372100 *T)
    sumL0  +=       0.019 * cos(2.66558529381  + 191.44826611160 *T)
    sumL0  +=       0.020 * cos(3.22150044915  + 156547.08861034497 *T)
    sumL0  +=       0.021 * cos(6.18753985884  + 104127.26777059487 *T)
    sumL0  +=       0.020 * cos(5.52295053123  + 130363.24963659538 *T)
    sumL0  +=       0.021 * cos(3.90684382108  + 51534.39272140940 *T)
    sumL0  +=       0.020 * cos(3.10019509540  + 838.96928775040 *T)
    sumL0  +=       0.025 * cos(5.79466888187  + 26098.94884183810 *T)
    sumL0  +=       0.021 * cos(6.23594952039  + 103814.80805420138 *T)
    sumL0  +=       0.019 * cos(5.79715725254  + 26279.35140768579 *T)
    sumL0  +=       0.022 * cos(0.71256272418  + 959.12285995760 *T)
    sumL0  +=       0.021 * cos(1.39589766697  + 103395.32341032618 *T)
    sumL0  +=       0.018 * cos(1.52837590084  + 107.30530692960 *T)
    sumL0  +=       0.019 * cos(0.60262102760  + 27566.76971563859 *T)
    sumL0  +=       0.025 * cos(0.78252166479  + 53814.87580033639 *T)
    sumL0  +=       0.025 * cos(2.29282750485  + 77520.71936418998 *T)
    sumL0  +=       0.018 * cos(2.11606037402  + 26.29831979980 *T)
    sumL0  +=       0.024 * cos(1.60264230919  + 208173.53416749896 *T)
    sumL0  +=       0.021 * cos(6.27983567062  + 54344.56676543099 *T)
    sumL0  +=       0.018 * cos(2.13288203799  + 77314.53381575279 *T)
    sumL0  +=       0.019 * cos(5.02228884604  + 26114.20146137400 *T)
    sumL0  +=       0.020 * cos(5.52520867267  + 9591.54174537180 *T)
    sumL0  +=       0.017 * cos(4.12183264175  + 24602.61243487099 *T)
    sumL0  +=       0.022 * cos(0.51026067819  + 95.97922721780 *T)
    sumL0  +=       0.017 * cos(4.48889832500  + 208276.62694171761 *T)
    sumL0  +=       0.017 * cos(3.94937058994  + 25459.05155552409 *T)
    sumL0  +=       0.017 * cos(5.29323037427  + 76137.83201734339 *T)
    sumL0  +=       0.019 * cos(3.67294881327  + 6885.14988993081 *T)
    sumL0  +=       0.018 * cos(1.02433476262  + 53399.62412392700 *T)
    sumL0  +=       0.020 * cos(2.79332004367  + 25927.49577218939 *T)
    sumL0  +=       0.018 * cos(5.83348216244  + 293.41078437720 *T)
    sumL0  +=       0.024 * cos(3.00078431660  + 23446.56186310200 *T)
    sumL0  +=       0.020 * cos(1.11172116756  + 156507.74908854539 *T)
    sumL0  +=       0.017 * cos(0.56645607004  + 104984.39630560996 *T)
    sumL0  +=       0.020 * cos(4.48412602835  + 52755.49387014719 *T)
    sumL0  +=       0.019 * cos(5.18601674976  + 2544.67263689240 *T)
    sumL0  +=       0.023 * cos(4.19936868759  + 24952.22626968100 *T)
    sumL0  +=       0.017 * cos(2.06473217893  + 130005.80396999416 *T)
    sumL0  +=       0.019 * cos(0.55214472899  + 52396.21892558720 *T)
    sumL0  +=       0.019 * cos(1.79653943961  + 89.48574784600 *T)
    sumL0  +=       0.022 * cos(3.72989530019  + 48997.66049258080 *T)
    sumL0  +=       0.020 * cos(5.93006048006  + 11852.35506339900 *T)
    sumL0  +=       0.018 * cos(6.24870732707  + 233681.74972207440 *T)
    sumL0  +=       0.021 * cos(3.01566725550  + 25907.82601128960 *T)
    sumL0  +=       0.016 * cos(1.73390726659  + 52911.68279668019 *T)
    sumL0  +=       0.017 * cos(3.23716624064  + 117893.03376878797 *T)
    sumL0  +=       0.022 * cos(0.46092942812  + 19367.18916223280 *T)
    sumL0  +=       0.016 * cos(5.06429474773  + 43981.53094993980 *T)
    sumL0  +=       0.021 * cos(1.13950009268  + 25650.25925043430 *T)
    sumL0  +=       0.019 * cos(3.19311280972  + 47803.92991637420 *T)
    sumL0  +=       0.021 * cos(4.64218889935  + 112545.88789950538 *T)
    sumL0  +=       0.017 * cos(2.96216633831  + 102975.83876645098 *T)
    sumL0  +=       0.016 * cos(0.60226326404  + 149846.19399604559 *T)
    sumL0  +=       0.016 * cos(5.86803075725  + 127098.90328117118 *T)
    sumL0  +=       0.018 * cos(1.20391834667  + 87367.61641884019 *T)
    sumL0  +=       0.016 * cos(4.53670173929  + 2667.05233364140 *T)
    sumL0  +=       0.017 * cos(5.77652308057  + 51130.65144696079 *T)
    sumL0  +=       0.018 * cos(5.87706432040  + 78057.52387628538 *T)
    sumL0  +=       0.019 * cos(5.24206436649  + 60370.08161635699 *T)
    sumL0  +=       0.016 * cos(1.31713012946  + 78843.39701172139 *T)
    sumL0  +=       0.019 * cos(5.33584429695  + 104454.70534051540 *T)
    sumL0  +=       0.015 * cos(4.52789928708  + 25984.12095299360 *T)
    sumL0  +=       0.016 * cos(4.12162585450  + 189386.03259226491 *T)
    sumL0  +=       0.017 * cos(3.51115348265  + 45455.09230380700 *T)
    sumL0  +=       0.015 * cos(1.65943772958  + 51596.11869614960 *T)
    sumL0  +=       0.015 * cos(4.28577331083  + 117.31986822020 *T)
    sumL0  +=       0.016 * cos(0.28814169191  + 78469.89497315978 *T)
    sumL0  +=       0.015 * cos(0.64079475058  + 53093.73656913019 *T)
    sumL0  +=       0.020 * cos(5.58915267196  + 1423.56473196300 *T)
    sumL0  +=       0.015 * cos(6.19871258212  + 156740.71794488319 *T)
    sumL0  +=       0.015 * cos(2.05066243688  + 16342.58258571760 *T)
    sumL0  +=       0.018 * cos(0.93517825293  + 104344.98400739678 *T)
    sumL0  +=       0.015 * cos(4.64013294295  + 2974.60995461100 *T)
    sumL0  +=       0.015 * cos(2.61010767947  + 7830.09253143480 *T)
    sumL0  +=       0.015 * cos(2.60585262136  + 151.04766984290 *T)
    sumL0  +=       0.018 * cos(0.92309429794  + 1577.34354244780 *T)
    sumL0  +=       0.017 * cos(1.36952124854  + 78115.63070029629 *T)
    sumL0  +=       0.018 * cos(1.95659679786  + 102872.74599223241 *T)
    sumL0  +=       0.017 * cos(5.37350443158  + 25384.26995695679 *T)
    sumL0  +=       0.015 * cos(2.12952532360  + 128106.31931499895 *T)
    sumL0  +=       0.018 * cos(4.06290374391  + 51439.92976961659 *T)
    sumL0  +=       0.015 * cos(4.32808623679  + 77741.13200662879 *T)
    sumL0  +=       0.014 * cos(3.46389680879  + 104426.39416486408 *T)
    sumL0  +=       0.020 * cos(5.15870331048  + 154408.65498906677 *T)
    sumL0  +=       0.014 * cos(4.53089771582  + 57503.28239153120 *T)
    sumL0  +=       0.016 * cos(3.64720066701  + 24072.23205541439 *T)
    sumL0  +=       0.015 * cos(2.96793644807  + 24719.24288872919 *T)
    sumL0  +=       0.014 * cos(0.79165737765  + 25928.60140678200 *T)
    sumL0  +=       0.014 * cos(4.36570321274  + 78249.48233072099 *T)
    sumL0  +=       0.014 * cos(5.35774038541  + 896.07945665540 *T)
    sumL0  +=       0.014 * cos(5.57146329055  + 104248.51979207818 *T)
    sumL0  +=       0.014 * cos(6.22061869408  + 104874.18998439058 *T)
    sumL0  +=       0.015 * cos(0.66056760946  + 77622.29586298359 *T)
    sumL0  +=       0.017 * cos(4.39404084250  + 162810.49469943656 *T)
    sumL0  +=       0.017 * cos(4.01700926317  + 51876.67988887919 *T)
    sumL0  +=       0.014 * cos(5.96074966457  + 4371.87636677420 *T)
    sumL0  +=       0.014 * cos(5.51412243126  + 48713.56157728320 *T)
    sumL0  +=       0.014 * cos(5.04161806976  + 53029.00266490040 *T)
    sumL0  +=       0.014 * cos(3.09979574700  + 128747.35003836860 *T)
    sumL0  +=       0.019 * cos(4.61702547610  + 16028.39640584800 *T)
    sumL0  +=       0.017 * cos(2.09526572835  + 52277.76880141400 *T)
    sumL0  +=       0.014 * cos(1.51516908270  + 52164.76058288449 *T)
    sumL0  +=       0.014 * cos(2.76589411039  + 26734.91397488900 *T)
    sumL0  +=       0.017 * cos(5.31477308011  + 103608.62250576420 *T)
    sumL0  +=       0.013 * cos(6.04881581356  + 52812.80255517260 *T)
    sumL0  +=       0.015 * cos(0.92065367569  + 256327.07791868312 *T)
    sumL0  +=       0.016 * cos(2.76936523753  + 46046.50924364180 *T)
    sumL0  +=       0.019 * cos(4.91561037107  + 277.03499374140 *T)
    sumL0  +=       0.015 * cos(1.87874008695  + 26161.20026743319 *T)
    sumL0  +=       0.013 * cos(1.48593089360  + 54824.26110862140 *T)
    sumL0  +=       0.015 * cos(1.37228117638  + 18043.37807764579 *T)
    sumL0  +=       0.018 * cos(2.95472168569  + 76152.05911134499 *T)
    sumL0  +=       0.017 * cos(1.57020520015  + 104241.40624507738 *T)
    sumL0  +=       0.018 * cos(2.97721991956  + 187167.27548807816 *T)
    sumL0  +=       0.013 * cos(3.49902133309  + 78397.81847430758 *T)
    sumL0  +=       0.018 * cos(2.18386661425  + 149288.74325784517 *T)
    sumL0  +=       0.017 * cos(5.38851170070  + 52041.69723356339 *T)
    sumL0  +=       0.013 * cos(6.26956660770  + 90989.16285949761 *T)
    sumL0  +=       0.013 * cos(1.13674691740  + 237009.88537835455 *T)
    sumL0  +=       0.014 * cos(2.18322393546  + 225687.22128005017 *T)
    sumL0  +=       0.013 * cos(6.15204536524  + 35211.47989659159 *T)
    sumL0  +=       0.017 * cos(1.32178768936  + 106262.81204951400 *T)
    sumL0  +=       0.014 * cos(1.73651187343  + 2912.88397987080 *T)
    sumL0  +=       0.013 * cos(0.84907711504  + 102343.05502713780 *T)
    sumL0  +=       0.013 * cos(2.39970220380  + 155460.92337225520 *T)
    sumL0  +=       0.017 * cos(4.15374414668  + 26014.60601571519 *T)
    sumL0  +=       0.013 * cos(0.72484082297  + 153878.96402397219 *T)
    sumL0  +=       0.013 * cos(3.04907288197  + 80432.46990700519 *T)
    sumL0  +=       0.013 * cos(1.37316946035  + 104035.22069664020 *T)
    sumL0  +=       0.015 * cos(3.12808384217  + 173.44857138459 *T)
    sumL0  +=       0.013 * cos(2.93774353095  + 533.62311835770 *T)
    sumL0  +=       0.013 * cos(2.24535401962  + 246113.79237247215 *T)
    sumL0  +=       0.016 * cos(0.39665846969  + 78112.66175487968 *T)
    sumL0  +=       0.015 * cos(4.75053341616  + 130020.03106399580 *T)
    sumL0  +=       0.016 * cos(0.76587038276  + 557.45073820040 *T)
    sumL0  +=       0.015 * cos(6.01045796437  + 51112.49219969609 *T)
    sumL0  +=       0.015 * cos(0.93774708323  + 461.33643145440 *T)
    sumL0  +=       0.012 * cos(4.03988032603  + 77956.15180375299 *T)
    sumL0  +=       0.015 * cos(1.43833460468  + 169093.57054942797 *T)
    sumL0  +=       0.014 * cos(2.14749108820  + 78129.60037513758 *T)
    sumL0  +=       0.012 * cos(4.57098987537  + 6751.04084034580 *T)
    sumL0  +=       0.017 * cos(0.64086933600  + 5131.64108405760 *T)
    sumL0  +=       0.013 * cos(5.02936180992  + 144916.86689107097 *T)
    sumL0  +=       0.016 * cos(3.65387335184  + 104028.10714963940 *T)
    sumL0  +=       0.013 * cos(3.37789792164  + 79212.88503369238 *T)
    sumL0  +=       0.013 * cos(1.38408689559  + 4885.80943782820 *T)
    sumL0  +=       0.012 * cos(4.95862705668  + 50536.73676596039 *T)
    sumL0  +=       0.014 * cos(2.05563043425  + 107692.22499299661 *T)
    sumL0  +=       0.012 * cos(3.93772391530  + 104668.00443595338 *T)
    sumL0  +=       0.014 * cos(3.26407683636  + 104401.60918820098 *T)
    sumL0  +=       0.014 * cos(2.94492731309  + 51120.35650622229 *T)
    sumL0  +=       0.013 * cos(5.30554859990  + 155475.15046625675 *T)
    sumL0  +=       0.014 * cos(5.14050840336  + 726.48201987020 *T)
    sumL0  +=       0.012 * cos(2.41433841278  + 24551.92639860479 *T)
    sumL0  +=       0.013 * cos(3.02673254226  + 182828.62108645737 *T)
    sumL0  +=       0.012 * cos(2.94598982599  + 5039.58824831300 *T)
    sumL0  +=       0.013 * cos(3.57478000552  + 78484.12206716138 *T)
    sumL0  +=       0.013 * cos(2.94343799803  + 79315.97780791098 *T)
    sumL0  +=       0.012 * cos(5.17024351037  + 103402.43695732698 *T)
    sumL0  +=       0.012 * cos(3.19152649116  + 28309.75977617119 *T)
    sumL0  +=       0.012 * cos(6.16852058989  + 78413.27262185719 *T)
    sumL0  +=       0.013 * cos(4.35939831008  + 133767.65136415116 *T)
    sumL0  +=       0.013 * cos(5.96217383792  + 2538.24850425360 *T)
    sumL0  +=       0.012 * cos(2.97455363125  + 26290.15653674829 *T)
    sumL0  +=       0.012 * cos(1.94777069809  + 156377.85565231059 *T)
    sumL0  +=       0.012 * cos(1.81926611664  + 70069.43409151399 *T)
    sumL0  +=       0.013 * cos(1.24212367708  + 4437.51420837359 *T)
    sumL0  +=       0.012 * cos(0.53775842570  + 106470.37642667518 *T)
    sumL0  +=       0.012 * cos(4.06029507610  + 79487.52726550119 *T)
    sumL0  +=       0.012 * cos(1.39982040688  + 77417.62658997139 *T)
    sumL0  +=       0.012 * cos(5.70094378358  + 79330.20490191258 *T)
    sumL0  +=       0.012 * cos(2.36617272843  + 79994.83177765518 *T)
    sumL0  +=       0.012 * cos(4.77094162628  + 78999.58593825439 *T)
    sumL0  +=       0.012 * cos(5.23189289937  + 26411.40855823160 *T)
    sumL0  +=       0.014 * cos(2.99800402366  + 1073.60902419080 *T)
    sumL0  +=       0.012 * cos(6.15146193785  + 103704.60173298200 *T)
    sumL0  +=       0.012 * cos(1.58617824624  + 78571.26704569219 *T)
    sumL0  +=       0.013 * cos(0.37894193721  + 53124.98189211819 *T)
    sumL0  +=       0.012 * cos(3.86549077079  + 189853.99758261937 *T)
    sumL0  +=       0.013 * cos(4.70785515154  + 234261.43730907320 *T)
    sumL0  +=       0.012 * cos(4.54097648940  + 56259.79478985279 *T)
    sumL0  +=       0.012 * cos(1.32257018476  + 149756.70824819960 *T)
    sumL0  +=       0.011 * cos(5.69779744921  + 222224.97657418496 *T)
    sumL0  +=       0.012 * cos(0.60900120711  + 2751.54759969160 *T)
    sumL0  +=       0.011 * cos(0.79876219873  + 78225.57638908479 *T)
    sumL0  +=       0.013 * cos(2.57664086838  + 1116.49209909420 *T)
    sumL0  +=       0.012 * cos(0.86411055123  + 78043.29678228378 *T)
    sumL0  +=       0.011 * cos(4.02869363295  + 263097.78851992881 *T)
    sumL0  +=       0.011 * cos(3.97587600566  + 76998.14194609619 *T)
    sumL0  +=       0.011 * cos(5.29433199334  + 52797.55008398759 *T)
    sumL0  +=       0.013 * cos(5.69550284497  + 51688.17153189420 *T)
    sumL0  +=       0.015 * cos(3.91946731118  + 27665.24668402200 *T)
    sumL0  +=       0.013 * cos(4.89334198524  + 45424.76544280259 *T)
    sumL0  +=       0.012 * cos(2.33066141026  + 130336.42293365240 *T)
    sumL0  +=       0.012 * cos(2.97893637259  + 25885.64974640009 *T)
    sumL0  +=       0.012 * cos(0.33532325982  + 41494.55574658299 *T)
    sumL0  +=       0.011 * cos(1.81390783997  + 104301.61594439260 *T)
    sumL0  +=       0.011 * cos(3.67540306066  + 79181.63971070439 *T)
    sumL0  +=       0.013 * cos(1.92207364074  + 2962.88060177500 *T)
    sumL0  +=       0.012 * cos(4.51335455691  + 54190.78795494619 *T)
    sumL0  +=       0.012 * cos(5.89140397220  + 156954.01704032117 *T)
    sumL0  +=       0.012 * cos(5.80412470555  + 72134.41238521598 *T)
    sumL0  +=       0.011 * cos(2.54446138738  + 26183.88236879199 *T)
    sumL0  +=       0.011 * cos(0.57810940674  + 142871.55835826878 *T)
    sumL0  +=       0.014 * cos(1.49797486148  + 103883.64757594238 *T)
    sumL0  +=       0.011 * cos(0.25563634033  + 52663.44103440259 *T)
    sumL0  +=       0.011 * cos(4.34226157432  + 63786.35824152260 *T)
    sumL0  +=       0.011 * cos(5.79546155779  + 76358.24465978218 *T)
    sumL0  +=       0.011 * cos(5.43581617959  + 4398.17468657401 *T)
    sumL0  +=       0.010 * cos(3.45435285202  + 26830.89320210680 *T)
    sumL0  +=       0.010 * cos(1.05961891051  + 216897.50046580212 *T)
    sumL0  +=       0.010 * cos(5.54078372768  + 53867.97195265079 *T)
    sumL0  +=       0.013 * cos(4.09291929409  + 78378.14871340781 *T)
    sumL0  +=       0.012 * cos(6.09214378282  + 156523.53751408719 *T)
    sumL0  +=       0.012 * cos(2.30793560033  + 192828.60753723039 *T)
    sumL0  +=       0.011 * cos(5.49984194986  + 130593.29451835580 *T)
    sumL0  +=       0.010 * cos(5.02934387609  + 63.73589830340 *T)
    sumL0  +=       0.010 * cos(2.56161075764  + 3178.14579056760 *T)
    sumL0  +=       0.010 * cos(4.51229030686  + 80912.16425019559 *T)
    sumL0  +=       0.010 * cos(3.26964460988  + 143980.93691036216 *T)
    sumL0  +=       0.014 * cos(0.09133406732  + 138.51749687070 *T)
    sumL0  +=       0.014 * cos(4.26704901561  + 52171.87412988529 *T)
    sumL0  +=       0.010 * cos(2.12874423735  + 181563.05360783095 *T)
    sumL0  +=       0.010 * cos(2.68594312022  + 53551.58008299420 *T)
    sumL0  +=       0.010 * cos(5.01052160566  + 53311.48315504160 *T)
    sumL0  +=       0.013 * cos(2.21304135803  + 287.88786006980 *T)
    sumL0  +=       0.010 * cos(5.99285759925  + 129063.74190802520 *T)
    sumL0  +=       0.010 * cos(0.23580667035  + 159969.99379441061 *T)
    sumL0  +=       0.014 * cos(4.15846442712  + 52190.03337714999 *T)
    sumL0  +=       0.014 * cos(2.17544501087  + 2506.64496425660 *T)
    sumL0  +=       0.010 * cos(2.39055885342  + 78262.22495201428 *T)
    sumL0  +=       0.010 * cos(5.33589950489  + 128843.32926558638 *T)
    sumL0  +=       0.010 * cos(3.69795961564  + 175934.09713761977 *T)
    sumL0  +=       0.010 * cos(5.50982370855  + 181548.82651382935 *T)
    sumL0  +=       0.012 * cos(0.91293309387  + 53258.88651544199 *T)
    sumL0  +=       0.010 * cos(0.21546965297  + 155674.22246769318 *T)
    sumL0  +=       0.010 * cos(5.05408550952  + 182465.75879388480 *T)
    sumL0  +=       0.011 * cos(1.47055363263  + 76991.02839909539 *T)
    sumL0  +=       0.014 * cos(1.74381856487  + 36.64856292950 *T)
    sumL0  +=       0.014 * cos(1.73080789777  + 171292.65789271498 *T)
    sumL0  +=       0.010 * cos(1.79047706298  + 235746.72801577640 *T)
    sumL0  +=       0.012 * cos(2.86918517100  + 2045.30853280220 *T)
    sumL0  +=       0.013 * cos(5.94639118711  + 168959.46149984296 *T)
    sumL0  +=       0.010 * cos(5.24213703651  + 130285.73689738620 *T)
    sumL0  +=       0.014 * cos(6.10459624713  + 454.90936652730 *T)
    sumL0  +=       0.010 * cos(5.39397583397  + 77.75054398390 *T)
    sumL0  +=       0.011 * cos(1.99827759086  + 1088.60315660100 *T)
    sumL0  +=       0.010 * cos(1.76064600598  + 52489.99246301800 *T)
    sumL0  +=       0.010 * cos(5.11611480967  + 113455.51956041438 *T)
    sumL0  +=       0.014 * cos(4.11983704037  + 54060.70744656579 *T)
    sumL0  +=       0.012 * cos(2.84049561669  + 183041.92018189537 *T)
    sumL0  +=       0.012 * cos(3.18297085790  + 228508.05242417639 *T)
    sumL0  +=       0.010 * cos(1.62042818189  + 231348.55332920240 *T)
    sumL0  +=       0.010 * cos(2.38308200520  + 181975.42470470537 *T)
    sumL0  +=       0.010 * cos(5.17901157458  + 203375.74902421079 *T)
    sumL0  +=       0.010 * cos(1.70609651468  + 78597.56536549199 *T)
    sumL0  +=       0.010 * cos(3.01880618190  + 104501.17576343138 *T)
    sumL0  +=       0.011 * cos(1.42631059185  + 58220.03486452380 *T)
    sumL0  +=       0.010 * cos(3.45269160846  + 99979.73619952259 *T)
    sumL0  +=       0.011 * cos(4.42716378223  + 235320.81923926240 *T)
    sumL0  +=       0.010 * cos(4.45423293249  + 5469.88378304060 *T)
    sumL0  +=       0.009 * cos(4.68983089958  + 80462.79676800959 *T)
    sumL0  +=       0.009 * cos(1.77635894319  + 50290.90511973100 *T)
    sumL0  +=       0.010 * cos(1.41464391522  + 48091.81777644400 *T)
    sumL0  +=       0.011 * cos(5.93477819724  + 158116.49174472899 *T)
    sumL0  +=       0.010 * cos(2.55614008171  + 105418.10804348681 *T)
    sumL0  +=       0.009 * cos(3.11835021508  + 52206.13314415280 *T)
    sumL0  +=       0.009 * cos(2.24380470375  + 299.12639426920 *T)
    sumL0  +=       0.009 * cos(3.22171579815  + 51006.21803173979 *T)
    sumL0  +=       0.009 * cos(3.24808968579  + 78697.42116259939 *T)
    sumL0  +=       0.011 * cos(0.73383746517  + 17605.73994829580 *T)
    sumL0  +=       0.010 * cos(2.45523846147  + 233835.52853255920 *T)
    sumL0  +=       0.009 * cos(4.35363918141  + 190809.59732422800 *T)
    sumL0  +=       0.009 * cos(2.77090630961  + 76094.94894243999 *T)
    sumL0  +=       0.012 * cos(0.23038883503  + 130907.48069822539 *T)
    sumL0  +=       0.009 * cos(5.05551915235  + 70383.62027138360 *T)
    sumL0  +=       0.009 * cos(5.11051597902  + 156093.70711156839 *T)
    sumL0  +=       0.011 * cos(5.44617170556  + 56777.75640211140 *T)
    sumL0  +=       0.009 * cos(4.67339642631  + 104771.09721017198 *T)
    sumL0  +=       0.009 * cos(4.56456351898  + 259819.64948555280 *T)
    sumL0  +=       0.009 * cos(0.73352390338  + 75085.56363415498 *T)
    sumL0  +=       0.010 * cos(2.64743106367  + 153186.80642274540 *T)
    sumL0  +=       0.010 * cos(5.18315351919  + 1382.19793248460 *T)
    sumL0  +=       0.012 * cos(4.75714453262  + 250665.74586953095 *T)
    sumL0  +=       0.011 * cos(5.79093643324  + 292628.26660646836 *T)
    sumL0  +=       0.009 * cos(6.24971585432  + 202420.14928260216 *T)
    sumL0  +=       0.012 * cos(5.83966176266  + 11790.62908865880 *T)
    sumL0  +=       0.011 * cos(2.84614041718  + 184204.39488630320 *T)
    sumL0  +=       0.010 * cos(3.41187941424  + 1905.46476494040 *T)
    sumL0  +=       0.010 * cos(3.00240265080  + 118828.96374949680 *T)
    sumL0  +=       0.009 * cos(3.90754065548  + 154308.66174525840 *T)
    sumL0  +=       0.010 * cos(0.88848120961  + 51315.49635439559 *T)
    sumL0  +=       0.011 * cos(4.49431301230  + 156531.30018480320 *T)
    sumL0  +=       0.009 * cos(0.37279337773  + 182622.43553802016 *T)
    sumL0  +=       0.009 * cos(5.94479236467  + 38634.38508065720 *T)
    sumL0  +=       0.009 * cos(0.28434529807  + 5019.91848741320 *T)
    sumL0  +=       0.011 * cos(5.30507160437  + 51257.87599716660 *T)
    sumL0  +=       0.009 * cos(5.05598171209  + 62197.64356324779 *T)
    sumL0  +=       0.009 * cos(2.00743011270  + 154194.22245657316 *T)
    sumL0  +=       0.010 * cos(2.29548861266  + 1211.34107035900 *T)
    sumL0  +=       0.009 * cos(1.40840962373  + 78149.27013603736 *T)
    sumL0  +=       0.011 * cos(2.69818187519  + 195047.36464141720 *T)
    sumL0  +=       0.009 * cos(3.77338110020  + 52073.84376488279 *T)
    sumL0  +=       0.010 * cos(4.16632261841  + 1130.15394758580 *T)
    sumL0  +=       0.009 * cos(4.71851534374  + 257436.45647077647 *T)
    sumL0  +=       0.009 * cos(1.56697202296  + 156107.93420556997 *T)
    sumL0  +=       0.010 * cos(3.84491070936  + 42790.96527507320 *T)
    sumL0  +=       0.010 * cos(0.84911221571  + 61921.12683900499 *T)
    sumL0  +=       0.011 * cos(3.43504498990  + 81725.95413058778 *T)
    sumL0  +=       0.008 * cos(3.46841587447  + 156534.53239644598 *T)
    sumL0  +=       0.009 * cos(3.07023951195  + 78265.19389743089 *T)
    sumL0  +=       0.011 * cos(5.33240591310  + 49424.25868345679 *T)
    sumL0  +=       0.009 * cos(3.80703237983  + 184834.07909520617 *T)
    sumL0  +=       0.010 * cos(5.24984657637  + 860.30992875280 *T)
    sumL0  +=       0.008 * cos(2.32846981012  + 156451.15277816958 *T)
    sumL0  +=       0.009 * cos(5.86810139691  + 221135.26778299137 *T)
    sumL0  +=       0.009 * cos(0.07267483428  + 54879.42243782400 *T)
    sumL0  +=       0.009 * cos(3.53847001127  + 1169.58825140860 *T)
    sumL0  +=       0.008 * cos(0.40730399000  + 5852.78031526380 *T)
    sumL0  +=       0.010 * cos(5.08938165465  + 53661.09698985160 *T)
    sumL0  +=       0.009 * cos(0.09227292796  + 285857.55600522278 *T)
    sumL0  +=       0.009 * cos(1.53921610006  + 171004.77003264520 *T)
    sumL0  +=       0.010 * cos(6.02369999880  + 105403.88094948516 *T)
    sumL0  +=       0.009 * cos(1.41005383919  + 215473.93573383920 *T)
    sumL0  +=       0.008 * cos(2.95064025568  + 182611.44065566137 *T)
    sumL0  +=       0.011 * cos(0.82592745776  + 52098.05573916449 *T)
    sumL0  +=       0.011 * cos(1.89268014069  + 53265.51507434200 *T)
    sumL0  +=       0.009 * cos(0.91534494504  + 130446.14426677101 *T)
    sumL0  +=       0.011 * cos(5.68651428373  + 97466.46267636596 *T)
    sumL0  +=       0.008 * cos(6.09336539202  + 65538.25598994759 *T)
    sumL0  +=       0.010 * cos(4.71864870979  + 1865.23140251760 *T)
    sumL0  +=       0.011 * cos(0.33942008091  + 2352.86615377180 *T)
    sumL0  +=       0.010 * cos(0.58149039679  + 51554.06248230919 *T)
    sumL0  +=       0.009 * cos(4.63908561414  + 130329.30938665158 *T)
    sumL0  +=       0.008 * cos(3.29340479484  + 76532.58707178998 *T)
    sumL0  +=       0.008 * cos(0.39996086139  + 31722.93687092660 *T)
    sumL0  +=       0.009 * cos(2.07892055945  + 84307.93800609799 *T)
    sumL0  +=       0.008 * cos(0.10251061291  + 130489.51232977520 *T)
    sumL0  +=       0.009 * cos(4.86488540298  + 197380.56103428919 *T)
    sumL0  +=       0.008 * cos(2.84319355109  + 247223.17092456561 *T)
    sumL0  +=       0.009 * cos(5.40644601602  + 205260.65018762811 *T)
    sumL0  +=       0.008 * cos(2.99554948925  + 28199.55345495179 *T)
    sumL0  +=       0.008 * cos(2.31903661034  + 50951.98844236979 *T)
    sumL0  +=       0.009 * cos(3.67243760379  + 132558.27956824939 *T)
    sumL0  +=       0.010 * cos(3.84482406254  + 199.07200143640 *T)
    sumL0  +=       0.008 * cos(5.13311255182  + 133780.12813457078 *T)
    sumL0  +=       0.008 * cos(4.50659262629  + 31775.53351052620 *T)
    sumL0  +=       0.010 * cos(3.05711078565  + 12098.18670962840 *T)
    sumL0  +=       0.009 * cos(5.43609233651  + 65851.33653522459 *T)
    sumL0  +=       0.008 * cos(3.16471275120  + 53438.96364572659 *T)
    sumL0  +=       0.008 * cos(5.49628419906  + 208063.32784627960 *T)
    sumL0  +=       0.009 * cos(0.80806156177  + 84944.93427812219 *T)
    sumL0  +=       0.008 * cos(1.45936166924  + 50160.82461135059 *T)
    sumL0  +=       0.009 * cos(3.83369994945  + 79902.77894191058 *T)
    sumL0  +=       0.008 * cos(5.05169260225  + 220025.88923089800 *T)
    sumL0  +=       0.010 * cos(0.99908522442  + 78373.91574594198 *T)
    sumL0  +=       0.008 * cos(1.78435325528  + 4743.40176317040 *T)
    sumL0  +=       0.009 * cos(2.44642878457  + 76468.45098100159 *T)
    sumL0  +=       0.010 * cos(0.64041079500  + 106684.80895916880 *T)
    sumL0  +=       0.008 * cos(0.22706755866  + 130514.29730643828 *T)
    sumL0  +=       0.008 * cos(4.45568539518  + 175844.61138977378 *T)
    sumL0  +=       0.008 * cos(5.30875425569  + 104991.50985261080 *T)
    sumL0  +=       0.008 * cos(4.26578657244  + 78267.64157798568 *T)
    sumL0  +=       0.010 * cos(3.66900856638  + 22595.33157470459 *T)
    sumL0  +=       0.008 * cos(5.39530906170  + 1168.89883704660 *T)
    sumL0  +=       0.007 * cos(2.97217552646  + 130962.09312596478 *T)
    sumL0  +=       0.008 * cos(5.41659156640  + 106082.73491922941 *T)
    sumL0  +=       0.007 * cos(6.14449914187  + 154835.25317994278 *T)
    sumL0  +=       0.009 * cos(3.77109166248  + 76777.72930365738 *T)
    sumL0  +=       0.010 * cos(2.68699198523  + 31903.01400121119 *T)
    sumL0  +=       0.010 * cos(4.11325286170  + 51861.62010327879 *T)
    sumL0  +=       0.010 * cos(6.22285311837  + 52213.93931878620 *T)
    sumL0  +=       0.008 * cos(4.05691998652  + 130432.88714897096 *T)
    sumL0  +=       0.008 * cos(5.44182392744  + 24844.41553989579 *T)
    sumL0  +=       0.008 * cos(5.31551782104  + 26247.20487636640 *T)
    sumL0  +=       0.007 * cos(3.34982243781  + 53771.99272543299 *T)
    sumL0  +=       0.007 * cos(0.76435484795  + 103829.03514820298 *T)
    sumL0  +=       0.009 * cos(0.48166296597  + 4571.62325795859 *T)
    sumL0  +=       0.010 * cos(2.53857628850  + 23735.03698780239 *T)
    sumL0  +=       0.009 * cos(3.06932982640  + 51102.19725895759 *T)
    sumL0  +=       0.009 * cos(0.81667861584  + 180396.56488683255 *T)
    sumL0  +=       0.008 * cos(1.37381137878  + 182619.20332637735 *T)
    sumL0  +=       0.010 * cos(4.53462580754  + 195181.47369100215 *T)
    sumL0  +=       0.007 * cos(5.92229713446  + 102129.75593169978 *T)
    sumL0  +=       0.009 * cos(1.70197819412  + 114.39910691340 *T)
    sumL0  +=       0.009 * cos(2.75554865259  + 52145.47942214399 *T)
    sumL0  +=       0.007 * cos(5.03002929900  + 78905.12298646157 *T)
    sumL0  +=       0.008 * cos(4.70162992466  + 77684.02183772379 *T)
    sumL0  +=       0.007 * cos(2.16016333042  + 252884.50297371775 *T)
    sumL0  +=       0.007 * cos(4.70104811610  + 197092.67317421935 *T)
    sumL0  +=       0.008 * cos(3.34882273958  + 92.05283574460 *T)
    sumL0  +=       0.007 * cos(4.64287583099  + 104659.17018726638 *T)
    sumL0  +=       0.007 * cos(1.74834665027  + 79116.90580647459 *T)
    sumL0  +=       0.007 * cos(1.40803395997  + 78149.31031780918 *T)
    sumL0  +=       0.007 * cos(5.96603423152  + 1685.05212250160 *T)
    sumL0  +=       0.008 * cos(5.74025115791  + 27441.65188659100 *T)
    sumL0  +=       0.007 * cos(3.32172728837  + 186057.89693598481 *T)
    sumL0  +=       0.007 * cos(5.77399157694  + 52822.81711646319 *T)
    sumL0  +=       0.009 * cos(5.40969007280  + 25970.58327335399 *T)
    sumL0  +=       0.007 * cos(5.03097379670  + 51092.72605085480 *T)
    sumL0  +=       0.008 * cos(4.85500497006  + 79353.41821591619 *T)
    sumL0  +=       0.009 * cos(2.91790460578  + 128857.55635958798 *T)
    sumL0  +=       0.010 * cos(2.62831495455  + 104145.42701785961 *T)
    sumL0  +=       0.008 * cos(3.65717048924  + 50476.52706664519 *T)
    sumL0  +=       0.010 * cos(4.97375703081  + 128960.64913380658 *T)
    sumL0  +=       0.007 * cos(0.63731291763  + 132350.71519108818 *T)
    sumL0  +=       0.007 * cos(3.70718359263  + 61165.27398858079 *T)
    sumL0  +=       0.007 * cos(1.52409890800  + 224577.84272795680 *T)
    sumL0  +=       0.007 * cos(5.83295581556  + 78339.97549599818 *T)
    sumL0  +=       0.007 * cos(1.28982657880  + 127791.06088239800 *T)
    sumL0  +=       0.007 * cos(4.38077281912  + 87648.55043279778 *T)
    sumL0  +=       0.007 * cos(0.78772735446  + 104044.05494532720 *T)
    sumL0  +=       0.007 * cos(5.55018242223  + 148532.89040742096 *T)
    sumL0  +=       0.008 * cos(3.94353800260  + 77929.85348395319 *T)
    sumL0  +=       0.007 * cos(4.68484570125  + 129971.55071751658 *T)
    sumL0  +=       0.007 * cos(5.46082675395  + 76352.50994150538 *T)
    sumL0  +=       0.007 * cos(0.89727780818  + 28213.78054895339 *T)
    sumL0  +=       0.006 * cos(2.25546242417  + 79955.87509422498 *T)
    sumL0  +=       0.007 * cos(1.85784871452  + 28096.46068073319 *T)
    sumL0  +=       0.006 * cos(6.22318977780  + 261988.40996783532 *T)
    sumL0  +=       0.006 * cos(3.92166942434  + 182402.02289558135 *T)
    sumL0  +=       0.006 * cos(4.54553590954  + 26235.98186600050 *T)
    sumL0  +=       0.006 * cos(0.04385251847  + 182634.99175191921 *T)
    sumL0  +=       0.006 * cos(3.27000773736  + 104557.79811473396 *T)
    sumL0  +=       0.006 * cos(1.48316940910  + 26191.68533015479 *T)
    sumL0  +=       0.006 * cos(1.42098831764  + 26026.17716683400 *T)
    sumL0  +=       0.006 * cos(4.63624566094  + 57369.17334194619 *T)
    sumL0  +=       0.006 * cos(2.24614788518  + 131079.41299418497 *T)
    sumL0  +=       0.008 * cos(1.46359871781  + 74935.57376844239 *T)
    sumL0  +=       0.007 * cos(0.19039890992  + 117077.06600107178 *T)
    sumL0  +=       0.006 * cos(2.30839350284  + 27353.47062020059 *T)
    sumL0  +=       0.008 * cos(4.05701479596  + 60170.33472517260 *T)
    sumL0  +=       0.006 * cos(2.25676540184  + 118007.47305747319 *T)
    sumL0  +=       0.007 * cos(5.19369769843  + 91919.56991589900 *T)
    sumL0  +=       0.006 * cos(2.17221286567  + 27573.88326263939 *T)
    sumL0  +=       0.006 * cos(3.04358438526  + 83591.18553310538 *T)
    sumL0  +=       0.006 * cos(5.50261387027  + 26238.95081141709 *T)
    sumL0  +=       0.006 * cos(5.55141392261  + 207747.62539098496 *T)
    sumL0  +=       0.007 * cos(3.67760971466  + 229129.79622501560 *T)
    sumL0  +=       0.006 * cos(5.14998820697  + 130364.73410930368 *T)
    sumL0  +=       0.006 * cos(2.59291082745  + 52186.85198341230 *T)
    sumL0  +=       0.006 * cos(3.70620257597  + 182608.20844401856 *T)
    sumL0  +=       0.007 * cos(5.89987227798  + 53757.76563143139 *T)
    sumL0  +=       0.007 * cos(1.17954845926  + 209232.91609768817 *T)
    sumL0  +=       0.007 * cos(1.99199973963  + 88285.54670482199 *T)
    sumL0  +=       0.006 * cos(5.29985761538  + 175376.64639941938 *T)
    sumL0  +=       0.007 * cos(1.76338847974  + 55484.27217852879 *T)
    sumL0  +=       0.008 * cos(0.55724966367  + 129806.73196855778 *T)
    sumL0  +=       0.005 * cos(4.19687137282  + 111032.83741969640 *T)
    sumL0  +=       0.006 * cos(1.86021237763  + 96471.52341295780 *T)
    sumL0  +=       0.006 * cos(1.44535366888  + 20043.67456019880 *T)
    sumL0  +=       0.006 * cos(1.48244229587  + 260349.34045064732 *T)
    sumL0  +=       0.006 * cos(1.00031747475  + 51538.81001112419 *T)
    sumL0  +=       0.005 * cos(0.08842088266  + 79859.89586700719 *T)
    sumL0  +=       0.006 * cos(2.17116807817  + 156373.64003896038 *T)
    sumL0  +=       0.007 * cos(2.57540094308  + 77218.55458853499 *T)
    sumL0  +=       0.006 * cos(4.66124811472  + 65717.22748563958 *T)
    sumL0  +=       0.005 * cos(5.69376389474  + 52400.15107885030 *T)
    sumL0  +=       0.006 * cos(2.47972848111  + 266540.36346489412 *T)
    sumL0  +=       0.005 * cos(4.47086994598  + 27360.58416720139 *T)
    sumL0  +=       0.005 * cos(3.20880924837  + 259769.65286364852 *T)
    sumL0  +=       0.005 * cos(3.98361684334  + 54862.52837768960 *T)
    sumL0  +=       0.007 * cos(2.86077141334  + 179172.74704605396 *T)
    sumL0  +=       0.005 * cos(2.46087822928  + 156681.19765992998 *T)
    sumL0  +=       0.007 * cos(5.00361119157  + 209762.60706278280 *T)
    sumL0  +=       0.005 * cos(5.35670104895  + 26513.81191808820 *T)
    sumL0  +=       0.006 * cos(3.62607025643  + 27883.16158529519 *T)
    sumL0  +=       0.005 * cos(0.80991017339  + 25661.99436506020 *T)
    sumL0  +=       0.006 * cos(0.20069428083  + 50490.75416064679 *T)
    sumL0  +=       0.005 * cos(3.77056538505  + 164407.50800278416 *T)
    sumL0  +=       0.005 * cos(2.71005052987  + 104427.87863757240 *T)
    sumL0  +=       0.006 * cos(0.84754795471  + 104466.05185498200 *T)
    sumL0  +=       0.005 * cos(2.69389382491  + 180496.55813064097 *T)
    sumL0  +=       0.005 * cos(4.04820376371  + 129277.04100346318 *T)
    sumL0  +=       0.006 * cos(0.57829057010  + 50600.96048186620 *T)
    sumL0  +=       0.006 * cos(4.21216601811  + 182595.65223011959 *T)
    sumL0  +=       0.005 * cos(3.59670550074  + 131072.29944718417 *T)
    sumL0  +=       0.005 * cos(0.08842684725  + 27993.36790651460 *T)
    sumL0  +=       0.006 * cos(0.00392034365  + 73891.83305794839 *T)
    sumL0  +=       0.006 * cos(4.68979746906  + 26149.62911631439 *T)
    sumL0  +=       0.005 * cos(4.61789756581  + 20894.53608277640 *T)
    sumL0  +=       0.005 * cos(3.57093232780  + 53654.67285721279 *T)
    sumL0  +=       0.005 * cos(0.36682731238  + 146505.58156934578 *T)
    sumL0  +=       0.005 * cos(1.40545824640  + 233202.05537888399 *T)
    sumL0  +=       0.006 * cos(2.35439456378  + 139543.42270198860 *T)
    sumL0  +=       0.004 * cos(1.28935961583  + 234364.53008329176 *T)
    sumL0  +=       0.004 * cos(1.10749672173  + 51086.09749195479 *T)
    sumL0  +=       0.005 * cos(3.64294402198  + 27331.39074325260 *T)
    sumL0  +=       0.004 * cos(0.00216060134  + 213255.17862965239 *T)
    sumL0  +=       0.004 * cos(4.40507427126  + 129483.22655190038 *T)
    sumL0  +=       0.004 * cos(2.18529936793  + 23384.28698689860 *T)
    sumL0  +=       0.005 * cos(0.93391333656  + 52102.50915728939 *T)
    sumL0  +=       0.004 * cos(2.33941435272  + 23389.45169419700 *T)
    sumL0  +=       0.004 * cos(3.59370682931  + 46564.47085590040 *T)
    sumL0  +=       0.004 * cos(2.90835831307  + 61299.38303816579 *T)
    sumL0  +=       0.004 * cos(6.27884818175  + 170068.84005193636 *T)
    sumL0  +=       0.004 * cos(4.90170477242  + 104888.41707839220 *T)
    sumL0  +=       0.004 * cos(4.56426377076  + 46131.57770177300 *T)
    sumL0  +=       0.005 * cos(2.12658067178  + 130542.60848208958 *T)
    sumL0  +=       0.005 * cos(2.50519016528  + 79845.66877300559 *T)
    sumL0  +=       0.004 * cos(6.24881318031  + 18073.70493865020 *T)
    sumL0  +=       0.004 * cos(1.10868110225  + 52072.02409456780 *T)
    sumL0  +=       0.004 * cos(6.14515791357  + 34102.10134449819 *T)
    sumL0  +=       0.004 * cos(2.94097893970  + 129902.71119577558 *T)
    sumL0  +=       0.004 * cos(2.91542426579  + 130215.17091216908 *T)
    sumL0  +=       0.005 * cos(2.20146513310  + 82865.65954368559 *T)
    sumL0  +=       0.005 * cos(4.46570663623  + 104237.17327761155 *T)
    sumL0  +=       0.004 * cos(5.13334797430  + 26387.02953584340 *T)
    sumL0  +=       0.004 * cos(5.34965278511  + 24787.07632336920 *T)
    sumL0  +=       0.004 * cos(4.17772291197  + 53132.09543911899 *T)
    sumL0  +=       0.004 * cos(2.66996288682  + 25572.43927048119 *T)
    sumL0  +=       0.005 * cos(1.14175711692  + 188898.39784101080 *T)

    # MERCURY Longitude
    sumL1  = 0
    sumL1  +=  2608814706222.746 * cos(0.00000000000  + 0.00000000000 *T)
    sumL1  +=  1126007.832 * cos(6.21703970996  + 26087.90314157420 *T)
    sumL1  +=  303471.395 * cos(3.05565472363  + 52175.80628314840 *T)
    sumL1  +=   80538.452 * cos(6.10454743366  + 78263.70942472259 *T)
    sumL1  +=   21245.035 * cos(2.83531934452  + 104351.61256629678 *T)
    sumL1  +=    5592.094 * cos(5.82675673328  + 130439.51570787100 *T)
    sumL1  +=    1472.233 * cos(2.51845458395  + 156527.41884944518 *T)
    sumL1  +=     352.244 * cos(3.05238094403  + 1109.37855209340 *T)
    sumL1  +=     388.318 * cos(5.48039225891  + 182615.32199101936 *T)
    sumL1  +=      93.540 * cos(6.11791163931  + 27197.28169366760 *T)
    sumL1  +=      90.579 * cos(0.00045481669  + 24978.52458948080 *T)
    sumL1  +=     102.743 * cos(2.14879173777  + 208703.22513259359 *T)
    sumL1  +=      51.941 * cos(5.62107554052  + 5661.33204915220 *T)
    sumL1  +=      44.370 * cos(4.57348500464  + 25028.52121138500 *T)
    sumL1  +=      28.070 * cos(3.04195430989  + 51066.42773105500 *T)
    sumL1  +=      22.003 * cos(0.86475371243  + 955.59974160860 *T)
    sumL1  +=      27.295 * cos(5.09210138837  + 234791.12827416777 *T)
    sumL1  +=      20.425 * cos(3.71509622702  + 20426.57109242200 *T)
    sumL1  +=      20.221 * cos(0.51934047295  + 21535.94964451540 *T)
    sumL1  +=      17.496 * cos(5.72663608620  + 4551.95349705880 *T)
    sumL1  +=      16.680 * cos(1.35134428173  + 529.69096509460 *T)
    sumL1  +=      15.306 * cos(1.79184360652  + 11322.66409830440 *T)
    sumL1  +=      15.398 * cos(5.74263453989  + 19.66976089979 *T)
    sumL1  +=      13.964 * cos(3.59426938083  + 24498.83024629040 *T)
    sumL1  +=      12.822 * cos(2.69591798562  + 53285.18483524180 *T)
    sumL1  +=      12.621 * cos(3.89530641889  + 3.88133535800 *T)
    sumL1  +=      12.566 * cos(4.70537436663  + 1059.38193018920 *T)
    sumL1  +=       7.974 * cos(4.17682324505  + 26617.59410666880 *T)
    sumL1  +=       7.929 * cos(0.50426804318  + 46514.47423399620 *T)
    sumL1  +=       8.024 * cos(3.92723313702  + 27043.50288318280 *T)
    sumL1  +=       7.665 * cos(2.48158305355  + 57837.13833230060 *T)
    sumL1  +=       8.639 * cos(6.06360282793  + 77154.33087262919 *T)
    sumL1  +=       6.838 * cos(2.77387065312  + 7.11354700080 *T)
    sumL1  +=       6.554 * cos(5.53478998989  + 6770.71060124560 *T)
    sumL1  +=       5.846 * cos(4.28139659800  + 16983.99614745660 *T)
    sumL1  +=       7.285 * cos(1.74581868954  + 260879.03141574195 *T)
    sumL1  +=       7.179 * cos(2.97766079692  + 2218.75710418680 *T)
    sumL1  +=       6.364 * cos(2.13805619380  + 25132.30339996560 *T)
    sumL1  +=       5.883 * cos(2.19545230409  + 13521.75144159140 *T)
    sumL1  +=       4.969 * cos(2.47752134180  + 30639.85663863300 *T)
    sumL1  +=       4.578 * cos(1.55845855383  + 27147.28507176339 *T)
    sumL1  +=       4.857 * cos(4.84757952324  + 37410.56723987860 *T)
    sumL1  +=       5.022 * cos(3.94124675992  + 25661.30495069820 *T)
    sumL1  +=       4.381 * cos(4.94471456712  + 213.29909543800 *T)
    sumL1  +=       4.211 * cos(5.53886788148  + 83925.04147387479 *T)
    sumL1  +=       4.299 * cos(5.09070560802  + 10213.28554621100 *T)
    sumL1  +=       4.593 * cos(0.81920676333  + 25558.21217647960 *T)
    sumL1  +=       3.571 * cos(2.34341884629  + 32858.61374281979 *T)
    sumL1  +=       4.454 * cos(5.79159156180  + 3442.57494496540 *T)
    sumL1  +=       3.547 * cos(3.59413896239  + 26068.23338067440 *T)
    sumL1  +=       3.664 * cos(1.39952931281  + 77204.32749453338 *T)
    sumL1  +=       3.417 * cos(0.49855041138  + 22645.32819660879 *T)
    sumL1  +=       3.195 * cos(1.25743024246  + 14765.23904326980 *T)
    sumL1  +=       3.031 * cos(4.45207030477  + 7238.67559160000 *T)
    sumL1  +=       3.284 * cos(5.22996707635  + 25448.00585526019 *T)
    sumL1  +=       2.974 * cos(0.14234349726  + 50586.73338786459 *T)
    sumL1  +=       2.855 * cos(0.67083991526  + 26091.78447693220 *T)
    sumL1  +=       2.773 * cos(3.54351198199  + 72602.37737557039 *T)
    sumL1  +=       3.132 * cos(6.21365344759  + 26080.78959457339 *T)
    sumL1  +=       2.732 * cos(5.64171761904  + 1589.07289528380 *T)
    sumL1  +=       2.652 * cos(0.88219699366  + 52705.49724824299 *T)
    sumL1  +=       3.125 * cos(6.07160751766  + 28306.66024576099 *T)
    sumL1  +=       2.799 * cos(2.51152391376  + 26107.57290247399 *T)
    sumL1  +=       2.403 * cos(1.15515130360  + 25035.63475838580 *T)
    sumL1  +=       2.274 * cos(1.85167073193  + 36301.18868778519 *T)
    sumL1  +=       2.501 * cos(4.34723465048  + 41962.52073693740 *T)
    sumL1  +=       2.321 * cos(5.44262378304  + 26084.02180621620 *T)
    sumL1  +=       2.535 * cos(5.80101392346  + 26095.01668857500 *T)
    sumL1  +=       2.872 * cos(1.63503968499  + 25021.40766438420 *T)
    sumL1  +=       2.066 * cos(1.07231845630  + 43071.89928903080 *T)
    sumL1  +=       2.616 * cos(2.78322243283  + 103242.23401420339 *T)
    sumL1  +=       1.987 * cos(2.95923214706  + 23969.13928119580 *T)
    sumL1  +=       1.979 * cos(4.44092591248  + 103292.23063610760 *T)
    sumL1  +=       1.974 * cos(0.83681704901  + 12566.15169998280 *T)
    sumL1  +=       1.700 * cos(2.28603732736  + 110012.94461544900 *T)
    sumL1  +=       1.677 * cos(4.63366122462  + 53235.18821333759 *T)
    sumL1  +=       1.954 * cos(4.67733727976  + 286966.93455731618 *T)
    sumL1  +=       1.653 * cos(1.25518276011  + 33326.57873317420 *T)
    sumL1  +=       1.785 * cos(1.80830270808  + 26301.20223701220 *T)
    sumL1  +=       1.784 * cos(5.17926273269  + 426.59819087600 *T)
    sumL1  +=       2.143 * cos(5.16312278245  + 51220.20654153979 *T)
    sumL1  +=       1.597 * cos(5.53494401703  + 56727.75978020720 *T)
    sumL1  +=       1.590 * cos(0.07660134798  + 23869.14603738740 *T)
    sumL1  +=       1.574 * cos(3.75870586370  + 73711.75592766379 *T)
    sumL1  +=       1.362 * cos(1.13830507985  + 68050.42387851159 *T)
    sumL1  +=       1.358 * cos(1.45317020242  + 51646.11531805379 *T)
    sumL1  +=       1.586 * cos(4.65681044624  + 79373.08797681599 *T)
    sumL1  +=       1.347 * cos(3.80254397382  + 19317.19254032860 *T)
    sumL1  +=       1.242 * cos(3.79389911401  + 25874.60404613620 *T)
    sumL1  +=       1.124 * cos(1.28920904043  + 31749.23519072640 *T)
    sumL1  +=       1.065 * cos(4.31416942177  + 40853.14218484400 *T)
    sumL1  +=       1.014 * cos(3.58126441973  + 48733.23133818299 *T)
    sumL1  +=       1.259 * cos(1.73130656223  + 12432.04265039780 *T)
    sumL1  +=       0.989 * cos(0.33276224293  + 52156.13652224860 *T)
    sumL1  +=       1.057 * cos(4.35251408825  + 77734.01845962799 *T)
    sumL1  +=       1.033 * cos(3.72503880786  + 6283.07584999140 *T)
    sumL1  +=       0.894 * cos(0.29768279333  + 105460.99111839020 *T)
    sumL1  +=       0.910 * cos(0.49823258962  + 99799.65906923798 *T)
    sumL1  +=       0.999 * cos(5.29812267938  + 65697.55772473979 *T)
    sumL1  +=       1.138 * cos(2.56391312022  + 29530.47808653960 *T)
    sumL1  +=       0.876 * cos(4.91112286703  + 62389.09182935939 *T)
    sumL1  +=       0.912 * cos(0.27449857787  + 98690.28051714458 *T)
    sumL1  +=       0.879 * cos(4.78847722011  + 18849.22754997420 *T)
    sumL1  +=       0.869 * cos(3.87957425954  + 78793.40038981718 *T)
    sumL1  +=       0.831 * cos(2.35459129726  + 27676.97603685800 *T)
    sumL1  +=       0.791 * cos(1.17444825871  + 129380.13377768178 *T)
    sumL1  +=       0.782 * cos(3.92653768734  + 38654.05484155699 *T)
    sumL1  +=       0.782 * cos(3.69412482234  + 52179.68761850640 *T)
    sumL1  +=       0.745 * cos(2.09648894863  + 51535.90899683439 *T)
    sumL1  +=       0.736 * cos(6.05470921559  + 114.43928868521 *T)
    sumL1  +=       0.733 * cos(3.23019978320  + 47623.85278608960 *T)
    sumL1  +=       0.698 * cos(2.99026308247  + 52168.69273614759 *T)
    sumL1  +=       0.780 * cos(5.77004504059  + 129330.13715577760 *T)
    sumL1  +=       0.752 * cos(4.22525357653  + 51123.53789995999 *T)
    sumL1  +=       0.752 * cos(4.70099207936  + 51109.31080595839 *T)
    sumL1  +=       0.651 * cos(5.87383050095  + 2333.19639287200 *T)
    sumL1  +=       0.699 * cos(4.82823185979  + 38519.94579197200 *T)
    sumL1  +=       0.656 * cos(3.65375726729  + 25668.41849769900 *T)
    sumL1  +=       0.660 * cos(1.98037689941  + 26514.50133245020 *T)
    sumL1  +=       0.808 * cos(6.27898737551  + 53131.40602475700 *T)
    sumL1  +=       0.753 * cos(5.64446551994  + 45892.73043315699 *T)
    sumL1  +=       0.610 * cos(1.21194007471  + 153.77881048480 *T)
    sumL1  +=       0.604 * cos(5.29267380518  + 136100.84775702321 *T)
    sumL1  +=       0.811 * cos(4.22885181263  + 25654.19140369739 *T)
    sumL1  +=       0.697 * cos(1.89331617498  + 77308.10968311399 *T)
    sumL1  +=       0.581 * cos(2.61707271637  + 76674.63652943878 *T)
    sumL1  +=       0.549 * cos(4.18661776549  + 94138.32702008578 *T)
    sumL1  +=       0.567 * cos(6.08785859254  + 25234.70675982219 *T)
    sumL1  +=       0.549 * cos(2.08003708209  + 91785.46086631398 *T)
    sumL1  +=       0.563 * cos(1.38660458000  + 79323.09135491178 *T)
    sumL1  +=       0.527 * cos(5.83448542494  + 50057.04242277000 *T)
    sumL1  +=       0.524 * cos(2.45745967369  + 467.96499035440 *T)
    sumL1  +=       0.570 * cos(3.10425513710  + 15874.61759536320 *T)
    sumL1  +=       0.699 * cos(2.57322536922  + 52182.91983014920 *T)
    sumL1  +=       0.501 * cos(0.39578180187  + 32370.97899156560 *T)
    sumL1  +=       0.542 * cos(2.77109053528  + 54394.56338733519 *T)
    sumL1  +=       0.525 * cos(1.31791713133  + 313054.83769889042 *T)
    sumL1  +=       0.513 * cos(2.28640740711  + 82815.66292178139 *T)
    sumL1  +=       0.579 * cos(5.45103833735  + 52195.47604404819 *T)
    sumL1  +=       0.617 * cos(4.18395572588  + 639.89728631400 *T)
    sumL1  +=       0.449 * cos(3.10044122691  + 49957.04917896160 *T)
    sumL1  +=       0.495 * cos(2.11807671869  + 52171.92494779040 *T)
    sumL1  +=       0.431 * cos(3.06714836720  + 131548.89425996438 *T)
    sumL1  +=       0.427 * cos(5.35367311100  + 49.99662190420 *T)
    sumL1  +=       0.470 * cos(1.46585019834  + 1066.49547719000 *T)
    sumL1  +=       0.470 * cos(1.07485595760  + 103821.92160120218 *T)
    sumL1  +=       0.415 * cos(2.50109134151  + 71980.63357473118 *T)
    sumL1  +=       0.395 * cos(5.24320159822  + 58946.51688439399 *T)
    sumL1  +=       0.408 * cos(0.92030240581  + 77837.11123384659 *T)
    sumL1  +=       0.523 * cos(4.92390082310  + 52389.10537858640 *T)
    sumL1  +=       0.377 * cos(3.50882025297  + 125887.56221081220 *T)
    sumL1  +=       0.508 * cos(1.03838633466  + 24491.71669928959 *T)
    sumL1  +=       0.418 * cos(0.36439379673  + 51962.50718771040 *T)
    sumL1  +=       0.400 * cos(3.51557589073  + 20760.42703319140 *T)
    sumL1  +=       0.359 * cos(0.03091967859  + 2118.76386037840 *T)
    sumL1  +=       0.356 * cos(2.03727281128  + 1052.26838318840 *T)
    sumL1  +=       0.423 * cos(4.21437546240  + 18093.37469954999 *T)
    sumL1  +=       0.340 * cos(1.64349471632  + 79219.30916633119 *T)
    sumL1  +=       0.336 * cos(1.55276786593  + 24505.94379329119 *T)
    sumL1  +=       0.348 * cos(1.06528227311  + 66941.04532641819 *T)
    sumL1  +=       0.327 * cos(5.45093694594  + 7880.08915333899 *T)
    sumL1  +=       0.303 * cos(1.67537340603  + 88476.99497093359 *T)
    sumL1  +=       0.323 * cos(5.59204872298  + 55618.38122811380 *T)
    sumL1  +=       0.281 * cos(0.83478570176  + 63498.47038145279 *T)
    sumL1  +=       0.393 * cos(0.55222046684  + 45405.09568190280 *T)
    sumL1  +=       0.280 * cos(0.69995179262  + 64741.95798313119 *T)
    sumL1  +=       0.381 * cos(2.16540447546  + 522.57741809380 *T)
    sumL1  +=       0.279 * cos(4.17319236859  + 155468.03691925600 *T)
    sumL1  +=       0.314 * cos(1.52755129973  + 44937.13069154840 *T)
    sumL1  +=       0.280 * cos(3.33549459014  + 78244.03966382280 *T)
    sumL1  +=       0.272 * cos(5.34763989826  + 53764.87917843220 *T)
    sumL1  +=       0.287 * cos(4.98400171694  + 25551.09862947879 *T)
    sumL1  +=       0.288 * cos(3.26773464056  + 124778.18365871880 *T)
    sumL1  +=       0.275 * cos(0.59220058467  + 104881.30353139140 *T)
    sumL1  +=       0.229 * cos(5.07700202235  + 52602.40447402440 *T)
    sumL1  +=       0.265 * cos(1.04033681288  + 25938.33994443960 *T)
    sumL1  +=       0.224 * cos(5.10726819180  + 117873.36400788820 *T)
    sumL1  +=       0.242 * cos(0.44760602599  + 23754.70674870219 *T)
    sumL1  +=       0.214 * cos(3.50460190598  + 58458.88213313979 *T)
    sumL1  +=       0.230 * cos(2.45937674101  + 155418.04029735178 *T)
    sumL1  +=       0.211 * cos(3.92377004339  + 103925.01437542078 *T)
    sumL1  +=       0.282 * cos(3.81265243625  + 536.80451209540 *T)
    sumL1  +=       0.218 * cos(4.88963285925  + 103396.01282468818 *T)
    sumL1  +=       0.200 * cos(5.41573374322  + 16066.06586147480 *T)
    sumL1  +=       0.218 * cos(0.41409439248  + 78267.59076008058 *T)
    sumL1  +=       0.201 * cos(2.38235635877  + 23439.44831610119 *T)
    sumL1  +=       0.203 * cos(2.69458032262  + 59414.48187474840 *T)
    sumL1  +=       0.196 * cos(0.92807787406  + 120226.23016166000 *T)
    sumL1  +=       0.201 * cos(2.00560473829  + 162188.75089859738 *T)
    sumL1  +=       0.183 * cos(3.83128433173  + 69159.80243060499 *T)
    sumL1  +=       0.180 * cos(2.20428238305  + 89586.37352302698 *T)
    sumL1  +=       0.215 * cos(5.35652096610  + 206.18554843720 *T)
    sumL1  +=       0.178 * cos(1.24568538055  + 103.09277421860 *T)
    sumL1  +=       0.196 * cos(0.44564203205  + 51756.32163927320 *T)
    sumL1  +=       0.183 * cos(2.88003872828  + 3328.13565628019 *T)
    sumL1  +=       0.216 * cos(4.48180965712  + 26241.68195205900 *T)
    sumL1  +=       0.215 * cos(2.26752450953  + 33967.99229491319 *T)
    sumL1  +=       0.193 * cos(4.34407767816  + 105307.21230790540 *T)
    sumL1  +=       0.200 * cos(1.01165813281  + 51742.09454527159 *T)
    sumL1  +=       0.170 * cos(5.57003076622  + 98068.53671630540 *T)
    sumL1  +=       0.224 * cos(0.97932137082  + 77211.44104153418 *T)
    sumL1  +=       0.168 * cos(5.96360976147  + 157636.79740153858 *T)
    sumL1  +=       0.178 * cos(1.42319344949  + 77197.21394753258 *T)
    sumL1  +=       0.196 * cos(5.59654497318  + 78270.82297172339 *T)
    sumL1  +=       0.161 * cos(5.51918174079  + 26555.86813192860 *T)
    sumL1  +=       0.197 * cos(4.79975858562  + 25455.11940226100 *T)
    sumL1  +=       0.181 * cos(4.39987679970  + 105410.99449648600 *T)
    sumL1  +=       0.159 * cos(2.63928704121  + 28421.09953444620 *T)
    sumL1  +=       0.158 * cos(0.75224521142  + 51116.42435295920 *T)
    sumL1  +=       0.196 * cos(4.92493608061  + 102762.53967101300 *T)
    sumL1  +=       0.162 * cos(2.82561928047  + 26202.34243025941 *T)
    sumL1  +=       0.194 * cos(1.02167898855  + 44181.27784112419 *T)
    sumL1  +=       0.179 * cos(5.11561177302  + 25440.89230825939 *T)
    sumL1  +=       0.152 * cos(5.21934955501  + 77623.81213840858 *T)
    sumL1  +=       0.150 * cos(5.98491689335  + 78256.59587772179 *T)
    sumL1  +=       0.184 * cos(0.26375114604  + 419.48464387520 *T)
    sumL1  +=       0.144 * cos(2.90657294970  + 51322.60990139639 *T)
    sumL1  +=       0.163 * cos(1.01728531879  + 26727.80042788820 *T)
    sumL1  +=       0.160 * cos(1.39225289207  + 26011.63707029860 *T)
    sumL1  +=       0.145 * cos(5.76503611587  + 80482.46652890938 *T)
    sumL1  +=       0.140 * cos(0.27227514908  + 46848.33017476560 *T)
    sumL1  +=       0.136 * cos(0.22641179921  + 151975.46535238638 *T)
    sumL1  +=       0.142 * cos(4.24343923379  + 339142.74084046454 *T)
    sumL1  +=       0.178 * cos(4.08547287819  + 129909.82474277640 *T)
    sumL1  +=       0.142 * cos(3.31130337074  + 78050.41032928458 *T)
    sumL1  +=       0.131 * cos(0.70125733168  + 3340.61242669980 *T)
    sumL1  +=       0.154 * cos(1.72826405606  + 78477.00852016058 *T)
    sumL1  +=       0.159 * cos(5.29643521725  + 108903.56606335558 *T)
    sumL1  +=       0.129 * cos(3.26743532785  + 25973.46385288896 *T)
    sumL1  +=       0.128 * cos(6.10775258657  + 76044.95232053580 *T)
    sumL1  +=       0.144 * cos(4.06091644657  + 50579.61984086379 *T)
    sumL1  +=       0.164 * cos(6.17696810226  + 19406.67828817460 *T)
    sumL1  +=       0.139 * cos(5.11979950242  + 433.71173787680 *T)
    sumL1  +=       0.153 * cos(1.30046862259  + 13655.86049117640 *T)
    sumL1  +=       0.132 * cos(5.23445066335  + 26610.48055966799 *T)
    sumL1  +=       0.115 * cos(4.61845793770  + 50593.84693486539 *T)
    sumL1  +=       0.121 * cos(2.52027376421  + 76144.94556434419 *T)
    sumL1  +=       0.114 * cos(4.51555533683  + 27154.39861876420 *T)
    sumL1  +=       0.106 * cos(5.18068866237  + 85502.38501632259 *T)
    sumL1  +=       0.110 * cos(3.45757703411  + 25565.32572348040 *T)
    sumL1  +=       0.100 * cos(0.11496600894  + 74821.13447975718 *T)
    sumL1  +=       0.100 * cos(3.95989644984  + 26037.90651966999 *T)
    sumL1  +=       0.100 * cos(3.01535382150  + 28206.66700195260 *T)
    sumL1  +=       0.099 * cos(5.99257260255  + 29416.03879785439 *T)
    sumL1  +=       0.117 * cos(3.56597928882  + 71492.99882347698 *T)
    sumL1  +=       0.131 * cos(5.96070215452  + 1223.81784077861 *T)
    sumL1  +=       0.095 * cos(1.84065750300  + 85034.42002596818 *T)
    sumL1  +=       0.102 * cos(4.69459571654  + 114564.89811250778 *T)
    sumL1  +=       0.097 * cos(0.87353840441  + 181555.94006083015 *T)
    sumL1  +=       0.118 * cos(2.02298794976  + 78283.37918562238 *T)
    sumL1  +=       0.111 * cos(4.07316227960  + 93028.94846799241 *T)
    sumL1  +=       0.090 * cos(5.01726237192  + 115674.27666460120 *T)
    sumL1  +=       0.094 * cos(2.31806507832  + 81706.28436968799 *T)
    sumL1  +=       0.089 * cos(5.94413487088  + 25131.61398560359 *T)
    sumL1  +=       0.095 * cos(3.72299894590  + 90829.86112470538 *T)
    sumL1  +=       0.120 * cos(5.44384753792  + 51749.20809227239 *T)
    sumL1  +=       0.105 * cos(5.01803826552  + 78259.82808936459 *T)
    sumL1  +=       0.086 * cos(1.83551544255  + 143961.26714946239 *T)
    sumL1  +=       0.108 * cos(4.53825443085  + 71025.03383312259 *T)
    sumL1  +=       0.083 * cos(0.64015893446  + 130012.91751699500 *T)
    sumL1  +=       0.099 * cos(0.52218556269  + 26624.70765366959 *T)
    sumL1  +=       0.094 * cos(2.65386396812  + 9103.90699411760 *T)
    sumL1  +=       0.098 * cos(3.31480172250  + 853.19638175200 *T)
    sumL1  +=       0.080 * cos(2.13424472361  + 42153.96900304900 *T)
    sumL1  +=       0.086 * cos(2.05760695844  + 79852.78232000639 *T)
    sumL1  +=       0.084 * cos(0.97023030623  + 131395.11544947961 *T)
    sumL1  +=       0.081 * cos(0.28916970825  + 84546.78527471398 *T)
    sumL1  +=       0.081 * cos(5.18614814481  + 220.41264243880 *T)
    sumL1  +=       0.081 * cos(1.98383157756  + 149.56319713460 *T)
    sumL1  +=       0.082 * cos(0.05817684667  + 104331.94280539700 *T)
    sumL1  +=       0.074 * cos(1.62836007935  + 25984.81036735560 *T)
    sumL1  +=       0.075 * cos(1.84245188133  + 78690.30761559859 *T)
    sumL1  +=       0.087 * cos(3.58634855360  + 130969.20667296558 *T)
    sumL1  +=       0.074 * cos(5.06505616563  + 27140.17152476259 *T)
    sumL1  +=       0.075 * cos(5.47760610876  + 134.10904958500 *T)
    sumL1  +=       0.089 * cos(6.25587801019  + 150866.08680029298 *T)
    sumL1  +=       0.098 * cos(4.71698591369  + 12725.45343477500 *T)
    sumL1  +=       0.069 * cos(5.05822197567  + 1911.19948321720 *T)
    sumL1  +=       0.081 * cos(2.05307563846  + 1596.18644228460 *T)
    sumL1  +=       0.075 * cos(2.05688075943  + 26137.89976347840 *T)
    sumL1  +=       0.069 * cos(3.67576176321  + 24609.03656750980 *T)
    sumL1  +=       0.081 * cos(1.38060133126  + 52329.58509363319 *T)
    sumL1  +=       0.069 * cos(1.60181551296  + 129483.91596626240 *T)
    sumL1  +=       0.067 * cos(5.43668176629  + 181505.94343892601 *T)
    sumL1  +=       0.066 * cos(1.56953438995  + 19804.82729158280 *T)
    sumL1  +=       0.066 * cos(3.92553131581  + 146314.13330323418 *T)
    sumL1  +=       0.062 * cos(3.46967694133  + 24925.42843716640 *T)
    sumL1  +=       0.079 * cos(3.56674606057  + 49842.60989027639 *T)
    sumL1  +=       0.067 * cos(4.76959706409  + 25881.71759313700 *T)
    sumL1  +=       0.061 * cos(2.61995550094  + 183724.70054311279 *T)
    sumL1  +=       0.061 * cos(1.63344766876  + 13541.42120249119 *T)
    sumL1  +=       0.061 * cos(2.31153135385  + 124156.43985787958 *T)
    sumL1  +=       0.058 * cos(3.88678278617  + 18207.81398823521 *T)
    sumL1  +=       0.082 * cos(1.36242531494  + 128850.44281258718 *T)
    sumL1  +=       0.058 * cos(2.98990179150  + 110.20632121940 *T)
    sumL1  +=       0.064 * cos(4.98131448153  + 188276.65404017159 *T)
    sumL1  +=       0.067 * cos(2.32849722473  + 26294.08869001139 *T)
    sumL1  +=       0.057 * cos(3.72454115593  + 29428.51556827400 *T)
    sumL1  +=       0.066 * cos(1.69869503126  + 51543.02254383520 *T)
    sumL1  +=       0.055 * cos(2.30093263701  + 52643.77127350280 *T)
    sumL1  +=       0.069 * cos(1.78774826271  + 51639.00177105299 *T)
    sumL1  +=       0.067 * cos(4.00863968986  + 103299.34418310841 *T)
    sumL1  +=       0.074 * cos(6.14089486388  + 14477.35118320000 *T)
    sumL1  +=       0.054 * cos(5.27646510692  + 49527.35145767539 *T)
    sumL1  +=       0.056 * cos(2.33495620660  + 104358.72611329758 *T)
    sumL1  +=       0.062 * cos(3.41967311766  + 104355.49390165480 *T)
    sumL1  +=       0.050 * cos(1.20854390978  + 23976.25282819660 *T)
    sumL1  +=       0.064 * cos(0.58512976749  + 307.55762096960 *T)
    sumL1  +=       0.062 * cos(3.69245785160  + 641.41356173899 *T)
    sumL1  +=       0.062 * cos(4.47317710621  + 636.99627202420 *T)
    sumL1  +=       0.049 * cos(4.09291926847  + 52815.70356946240 *T)
    sumL1  +=       0.061 * cos(0.78849475911  + 155997.72788435058 *T)
    sumL1  +=       0.047 * cos(2.14659254610  + 7994.52844202420 *T)
    sumL1  +=       0.047 * cos(0.00507866085  + 104138.31347085880 *T)
    sumL1  +=       0.045 * cos(2.28588893693  + 25764.39772491679 *T)
    sumL1  +=       0.051 * cos(2.98348805265  + 40565.25432477420 *T)
    sumL1  +=       0.047 * cos(4.03957254582  + 77829.99768684579 *T)
    sumL1  +=       0.047 * cos(3.21862563103  + 178063.36849396059 *T)
    sumL1  +=       0.045 * cos(4.41401819607  + 103285.11708910680 *T)
    sumL1  +=       0.054 * cos(3.47879330948  + 77844.22478084739 *T)
    sumL1  +=       0.054 * cos(2.83659707386  + 6681.22485339960 *T)
    sumL1  +=       0.056 * cos(1.11360360738  + 131498.89763806018 *T)
    sumL1  +=       0.046 * cos(3.38384624666  + 26507.38778544939 *T)
    sumL1  +=       0.055 * cos(2.68637168389  + 25934.12433108940 *T)
    sumL1  +=       0.046 * cos(4.78696197215  + 104564.91166173480 *T)
    sumL1  +=       0.059 * cos(1.29879100690  + 38813.35657634920 *T)
    sumL1  +=       0.043 * cos(3.86025201636  + 10021.83728009940 *T)
    sumL1  +=       0.044 * cos(1.90320971624  + 51528.79544983359 *T)
    sumL1  +=       0.055 * cos(0.48335079540  + 23962.02573419499 *T)
    sumL1  +=       0.042 * cos(4.11693586607  + 26190.99591579279 *T)
    sumL1  +=       0.044 * cos(1.64247485835  + 76.26607127560 *T)
    sumL1  +=       0.042 * cos(6.21848377968  + 26094.53170047421 *T)
    sumL1  +=       0.041 * cos(2.73996884659  + 25867.49049913539 *T)
    sumL1  +=       0.043 * cos(0.38981425968  + 1089.70879119360 *T)
    sumL1  +=       0.043 * cos(4.89046790101  + 24395.73747207180 *T)
    sumL1  +=       0.041 * cos(0.44351702671  + 25619.93815121980 *T)
    sumL1  +=       0.041 * cos(1.47439583353  + 76681.75007643958 *T)
    sumL1  +=       0.039 * cos(1.71519867865  + 111590.28815789679 *T)
    sumL1  +=       0.041 * cos(5.67594181048  + 54509.00267602040 *T)
    sumL1  +=       0.049 * cos(1.65739950579  + 1581.95934828300 *T)
    sumL1  +=       0.040 * cos(1.85008425708  + 27999.10262479140 *T)
    sumL1  +=       0.038 * cos(3.99668572981  + 9384.84100807520 *T)
    sumL1  +=       0.037 * cos(2.83147987268  + 102132.85546211000 *T)
    sumL1  +=       0.039 * cos(0.84546566819  + 365230.64398203877 *T)
    sumL1  +=       0.037 * cos(4.65200082367  + 90695.75207512038 *T)
    sumL1  +=       0.037 * cos(0.94736176192  + 76667.52298243798 *T)
    sumL1  +=       0.040 * cos(0.16806354899  + 22759.76748529401 *T)
    sumL1  +=       0.049 * cos(2.02480943113  + 134991.46920492980 *T)
    sumL1  +=       0.041 * cos(5.85592543833  + 52290.24557183361 *T)
    sumL1  +=       0.040 * cos(1.21491127400  + 64607.84893354619 *T)
    sumL1  +=       0.041 * cos(3.07842612799  + 26081.27458267419 *T)
    sumL1  +=       0.035 * cos(3.89800430981  + 157483.01859105378 *T)
    sumL1  +=       0.037 * cos(0.29068345666  + 97580.90196505121 *T)
    sumL1  +=       0.035 * cos(4.74965057267  + 39629.32434406539 *T)
    sumL1  +=       0.035 * cos(1.88777425098  + 26521.61487945100 *T)
    sumL1  +=       0.033 * cos(3.60227423859  + 156100.82065856917 *T)
    sumL1  +=       0.035 * cos(1.96114191570  + 74.78159856730 *T)
    sumL1  +=       0.045 * cos(3.72467663437  + 39609.65458316560 *T)
    sumL1  +=       0.038 * cos(4.29218444251  + 39743.76363275060 *T)
    sumL1  +=       0.032 * cos(5.95900501016  + 77410.51304297059 *T)
    sumL1  +=       0.032 * cos(0.64794221724  + 2648.45482547300 *T)
    sumL1  +=       0.039 * cos(5.08510962281  + 25863.55834587229 *T)
    sumL1  +=       0.031 * cos(1.62486752727  + 141762.17980617538 *T)
    sumL1  +=       0.031 * cos(1.25910229689  + 53242.30176033840 *T)
    sumL1  +=       0.031 * cos(2.65918037491  + 104344.49901929598 *T)
    sumL1  +=       0.032 * cos(0.00169740803  + 52061.36699446317 *T)
    sumL1  +=       0.031 * cos(6.12210016938  + 25977.69682035479 *T)
    sumL1  +=       0.030 * cos(5.72239820341  + 579.68758699880 *T)
    sumL1  +=       0.030 * cos(0.70553319870  + 52125.80966124419 *T)
    sumL1  +=       0.039 * cos(1.60801806885  + 333.85594076940 *T)
    sumL1  +=       0.030 * cos(5.28578152795  + 102232.84870591838 *T)
    sumL1  +=       0.029 * cos(0.53270948692  + 50696.93970908399 *T)
    sumL1  +=       0.029 * cos(0.40281274949  + 35472.74414964940 *T)
    sumL1  +=       0.028 * cos(2.96171859812  + 100909.03762133140 *T)
    sumL1  +=       0.034 * cos(1.41428136641  + 140652.80125408198 *T)
    sumL1  +=       0.029 * cos(6.00401578551  + 54294.57014352679 *T)
    sumL1  +=       0.032 * cos(0.45221016093  + 116917.76426627958 *T)
    sumL1  +=       0.028 * cos(5.07088961710  + 26237.46633870879 *T)
    sumL1  +=       0.028 * cos(5.33382835025  + 107794.18751126219 *T)
    sumL1  +=       0.030 * cos(3.86080757411  + 207643.84320240439 *T)
    sumL1  +=       0.028 * cos(4.81897403543  + 170049.17029103660 *T)
    sumL1  +=       0.031 * cos(1.90989851759  + 26308.31578401300 *T)
    sumL1  +=       0.035 * cos(1.26114015069  + 97112.93697469680 *T)
    sumL1  +=       0.028 * cos(2.01146788563  + 52698.38370124219 *T)
    sumL1  +=       0.033 * cos(4.31496976407  + 154938.34595416137 *T)
    sumL1  +=       0.028 * cos(4.46781875730  + 78417.48823520739 *T)
    sumL1  +=       0.033 * cos(0.25034815279  + 51653.22886505459 *T)
    sumL1  +=       0.026 * cos(0.07835049738  + 51013.33157874059 *T)
    sumL1  +=       0.029 * cos(3.32582529322  + 110634.68841628821 *T)
    sumL1  +=       0.027 * cos(6.01993388409  + 25227.59321282139 *T)
    sumL1  +=       0.036 * cos(0.79025288269  + 119116.85160956658 *T)
    sumL1  +=       0.026 * cos(4.68624143375  + 632.78373931320 *T)
    sumL1  +=       0.027 * cos(1.14521834701  + 121335.60871375341 *T)
    sumL1  +=       0.025 * cos(3.42095174931  + 956.28915597060 *T)
    sumL1  +=       0.028 * cos(2.73386270717  + 27311.72098235281 *T)
    sumL1  +=       0.024 * cos(0.82215620694  + 314.18617986960 *T)
    sumL1  +=       0.025 * cos(5.21681600178  + 25241.82030682300 *T)
    sumL1  +=       0.025 * cos(3.57372833543  + 52712.61079524379 *T)
    sumL1  +=       0.027 * cos(5.05183520711  + 105940.68546158058 *T)
    sumL1  +=       0.031 * cos(2.25462520789  + 26720.68688088739 *T)
    sumL1  +=       0.025 * cos(4.86890931082  + 104778.21075717278 *T)
    sumL1  +=       0.023 * cos(3.01659387998  + 52182.43484204840 *T)
    sumL1  +=       0.026 * cos(2.13973174244  + 103711.71527998280 *T)
    sumL1  +=       0.023 * cos(4.60383496965  + 52072.71350892979 *T)
    sumL1  +=       0.023 * cos(4.55605691468  + 155571.81910783658 *T)
    sumL1  +=       0.022 * cos(3.69622699918  + 735.87651353180 *T)
    sumL1  +=       0.023 * cos(2.40544682432  + 8194.27533320860 *T)
    sumL1  +=       0.022 * cos(2.44094101414  + 2703.61615467560 *T)
    sumL1  +=       0.027 * cos(5.29221944987  + 26162.68474014150 *T)
    sumL1  +=       0.022 * cos(4.11517024183  + 19202.75325164339 *T)
    sumL1  +=       0.029 * cos(4.25385019536  + 647.01083331480 *T)
    sumL1  +=       0.021 * cos(2.11851790816  + 207593.84658050016 *T)
    sumL1  +=       0.029 * cos(2.91003553544  + 176953.98994186718 *T)
    sumL1  +=       0.026 * cos(5.68398596586  + 95247.70557217918 *T)
    sumL1  +=       0.021 * cos(5.29289519906  + 150244.34299945380 *T)
    sumL1  +=       0.024 * cos(3.05210539174  + 130419.84594697120 *T)
    sumL1  +=       0.020 * cos(2.54144132983  + 51219.51712717779 *T)
    sumL1  +=       0.020 * cos(6.16401959180  + 35191.81013569180 *T)
    sumL1  +=       0.022 * cos(3.22920342008  + 1083.08023229360 *T)
    sumL1  +=       0.028 * cos(0.30709453091  + 157057.10981453978 *T)
    sumL1  +=       0.023 * cos(1.65280537431  + 214364.55718174577 *T)
    sumL1  +=       0.027 * cos(3.72484040647  + 621.74380083920 *T)
    sumL1  +=       0.023 * cos(2.59065413623  + 71582.48457132299 *T)
    sumL1  +=       0.022 * cos(0.94588004362  + 52022.02747266360 *T)
    sumL1  +=       0.020 * cos(3.84776796820  + 5327.47610838280 *T)
    sumL1  +=       0.024 * cos(4.82779429778  + 104371.28232719658 *T)
    sumL1  +=       0.019 * cos(1.94368317235  + 22747.29071487440 *T)
    sumL1  +=       0.023 * cos(3.19739923216  + 6044.22858137540 *T)
    sumL1  +=       0.020 * cos(0.58967836529  + 172402.03644480839 *T)
    sumL1  +=       0.019 * cos(2.73169091081  + 55503.94193942859 *T)
    sumL1  +=       0.018 * cos(1.63099638721  + 48835.19385644859 *T)
    sumL1  +=       0.022 * cos(1.62646807429  + 104347.73123093881 *T)
    sumL1  +=       0.018 * cos(4.90435895585  + 77630.92568540938 *T)
    sumL1  +=       0.018 * cos(0.59687432455  + 183570.92173262799 *T)
    sumL1  +=       0.024 * cos(5.58828197916  + 34282.17847478280 *T)
    sumL1  +=       0.018 * cos(5.54856348293  + 209812.60368468691 *T)
    sumL1  +=       0.019 * cos(0.53136209334  + 55516.41870984820 *T)
    sumL1  +=       0.020 * cos(0.73777252741  + 129387.24732468258 *T)
    sumL1  +=       0.021 * cos(4.62446809035  + 13362.44970679920 *T)
    sumL1  +=       0.021 * cos(1.94768946886  + 487.63475125420 *T)
    sumL1  +=       0.019 * cos(4.18946427203  + 78114.14622758799 *T)
    sumL1  +=       0.023 * cos(2.43977160852  + 24822.33566294780 *T)
    sumL1  +=       0.017 * cos(2.24086275048  + 26222.01219115920 *T)
    sumL1  +=       0.018 * cos(5.08348808758  + 27684.08958385880 *T)
    sumL1  +=       0.017 * cos(4.31062701282  + 50064.15596977079 *T)
    sumL1  +=       0.017 * cos(0.62713581550  + 44295.71712980940 *T)
    sumL1  +=       0.022 * cos(0.09847423078  + 26941.09952332620 *T)
    sumL1  +=       0.019 * cos(2.09224104001  + 86143.79857806159 *T)
    sumL1  +=       0.016 * cos(0.43713509764  + 101.96251826560 *T)
    sumL1  +=       0.016 * cos(4.27219175799  + 24176.70365835700 *T)
    sumL1  +=       0.018 * cos(5.24040005379  + 68241.87214462319 *T)
    sumL1  +=       0.020 * cos(5.09682447452  + 60055.89543648739 *T)
    sumL1  +=       0.016 * cos(4.08590900542  + 147423.51185532758 *T)
    sumL1  +=       0.018 * cos(2.12913862198  + 106570.36967048360 *T)
    sumL1  +=       0.016 * cos(1.00668387150  + 25138.72753260440 *T)
    sumL1  +=       0.016 * cos(5.76419504418  + 14.22709400160 *T)
    sumL1  +=       0.021 * cos(6.16511987091  + 52169.17772424839 *T)
    sumL1  +=       0.016 * cos(0.82798074394  + 96357.08412427260 *T)
    sumL1  +=       0.016 * cos(3.75399093257  + 159.30173479220 *T)
    sumL1  +=       0.017 * cos(4.95698248546  + 52225.80290505260 *T)
    sumL1  +=       0.017 * cos(6.11461465994  + 26198.10946279360 *T)
    sumL1  +=       0.017 * cos(6.19302565011  + 204151.27163553479 *T)
    sumL1  +=       0.017 * cos(5.34573358099  + 78731.67441507700 *T)
    sumL1  +=       0.016 * cos(1.90468795194  + 128320.75184749260 *T)
    sumL1  +=       0.017 * cos(3.73180335896  + 182085.63102592481 *T)
    sumL1  +=       0.017 * cos(5.86222137701  + 32132.13172294960 *T)
    sumL1  +=       0.019 * cos(1.56034038551  + 51969.62073471119 *T)
    sumL1  +=       0.018 * cos(2.54086242120  + 25780.34552060460 *T)
    sumL1  +=       0.015 * cos(4.29267057949  + 1692.16566950240 *T)
    sumL1  +=       0.016 * cos(1.75865740707  + 53228.07466633679 *T)
    sumL1  +=       0.020 * cos(2.35700883433  + 75615.25459924959 *T)
    sumL1  +=       0.015 * cos(5.29296109539  + 130446.62925487180 *T)
    sumL1  +=       0.014 * cos(0.85174293693  + 78903.60671103658 *T)
    sumL1  +=       0.015 * cos(5.50776870695  + 52808.59002246159 *T)
    sumL1  +=       0.017 * cos(5.96622721781  + 32769.12799497380 *T)
    sumL1  +=       0.014 * cos(5.40589812318  + 52381.99183158559 *T)
    sumL1  +=       0.014 * cos(2.92783624754  + 130226.21661243298 *T)
    sumL1  +=       0.014 * cos(4.73334676783  + 26164.16921284980 *T)
    sumL1  +=       0.018 * cos(4.09540263999  + 157586.80077963436 *T)
    sumL1  +=       0.015 * cos(5.00568645183  + 28791.51929624980 *T)
    sumL1  +=       0.013 * cos(0.88297787282  + 24292.64469785319 *T)
    sumL1  +=       0.018 * cos(0.07544318179  + 2199.08734328700 *T)
    sumL1  +=       0.014 * cos(0.39198838648  + 1162.47470440780 *T)
    sumL1  +=       0.012 * cos(4.87283902225  + 64901.25971792339 *T)
    sumL1  +=       0.013 * cos(4.28417448099  + 23866.04650697719 *T)
    sumL1  +=       0.013 * cos(3.77892895408  + 391318.54712361295 *T)
    sumL1  +=       0.012 * cos(0.90287358387  + 104202.04936916218 *T)
    sumL1  +=       0.012 * cos(1.68139724364  + 72936.23331633979 *T)
    sumL1  +=       0.015 * cos(3.76108194990  + 111122.32316754240 *T)
    sumL1  +=       0.012 * cos(1.82629728829  + 22909.75735100660 *T)
    sumL1  +=       0.013 * cos(1.25451832397  + 26724.89941359840 *T)
    sumL1  +=       0.012 * cos(5.74844907374  + 97670.38771289718 *T)
    sumL1  +=       0.012 * cos(1.52451201886  + 130652.81480330898 *T)
    sumL1  +=       0.011 * cos(5.30694474936  + 24601.92302050899 *T)
    sumL1  +=       0.014 * cos(2.82065667408  + 1639.06951718800 *T)
    sumL1  +=       0.015 * cos(4.79144654769  + 77726.90491262719 *T)
    sumL1  +=       0.013 * cos(5.37369194634  + 8989.46770543239 *T)
    sumL1  +=       0.013 * cos(3.84042504901  + 25953.79409198919 *T)
    sumL1  +=       0.015 * cos(0.08558342266  + 130443.39704322898 *T)
    sumL1  +=       0.011 * cos(5.16135808879  + 233681.74972207440 *T)
    sumL1  +=       0.012 * cos(0.99700192092  + 65831.66677432480 *T)
    sumL1  +=       0.012 * cos(4.51508367315  + 78187.44335344699 *T)
    sumL1  +=       0.012 * cos(0.62022460619  + 36109.74042167360 *T)
    sumL1  +=       0.011 * cos(1.08182792000  + 129373.02023068098 *T)
    sumL1  +=       0.014 * cos(5.59428027773  + 25446.48957983520 *T)
    sumL1  +=       0.011 * cos(4.68045355937  + 74923.09699802278 *T)
    sumL1  +=       0.014 * cos(4.68060965317  + 27669.86248985719 *T)
    sumL1  +=       0.012 * cos(4.91061349082  + 54087.00576636560 *T)
    sumL1  +=       0.010 * cos(0.73914692842  + 103917.90082841998 *T)
    sumL1  +=       0.013 * cos(4.96538793287  + 161079.37234650398 *T)
    sumL1  +=       0.011 * cos(0.31330958005  + 182188.72380014337 *T)
    sumL1  +=       0.013 * cos(3.74649756573  + 145204.75475114080 *T)
    sumL1  +=       0.010 * cos(1.79124533717  + 949.17560896980 *T)
    sumL1  +=       0.010 * cos(4.24347093551  + 78109.93061423779 *T)
    sumL1  +=       0.010 * cos(1.22394899357  + 104505.39137678158 *T)
    sumL1  +=       0.012 * cos(3.21001019103  + 102018.41617342478 *T)
    sumL1  +=       0.011 * cos(4.48294709049  + 102769.65321801380 *T)
    sumL1  +=       0.010 * cos(4.33936275875  + 95.97922721780 *T)
    sumL1  +=       0.010 * cos(3.73409843943  + 199599.31813847597 *T)
    sumL1  +=       0.012 * cos(6.08780023115  + 25450.90686955000 *T)
    sumL1  +=       0.012 * cos(0.52219299955  + 125112.03959948818 *T)
    sumL1  +=       0.012 * cos(3.12928292612  + 11610.55195837420 *T)
    sumL1  +=       0.013 * cos(2.59745313501  + 78378.14871340781 *T)
    sumL1  +=       0.010 * cos(0.95664142687  + 45290.65639321759 *T)
    sumL1  +=       0.011 * cos(3.04942272888  + 309.27832265580 *T)
    sumL1  +=       0.012 * cos(3.42470973506  + 143005.66740785379 *T)
    sumL1  +=       0.009 * cos(5.24571653547  + 34082.43158359840 *T)
    sumL1  +=       0.012 * cos(1.94632930172  + 1135.67687189320 *T)
    sumL1  +=       0.010 * cos(2.75654212398  + 103.78218858060 *T)
    sumL1  +=       0.010 * cos(6.06089846081  + 78270.33798362259 *T)
    sumL1  +=       0.009 * cos(1.42295924162  + 12546.48193908300 *T)
    sumL1  +=       0.009 * cos(5.23889079020  + 160.40736938481 *T)
    sumL1  +=       0.010 * cos(1.77339719648  + 25754.04720080480 *T)
    sumL1  +=       0.011 * cos(3.45444324976  + 27177.61193276780 *T)
    sumL1  +=       0.010 * cos(2.40808781707  + 3462.24470586520 *T)
    sumL1  +=       0.012 * cos(0.21115577900  + 103932.12792242158 *T)
    sumL1  +=       0.010 * cos(0.44397528833  + 26729.31670331319 *T)
    sumL1  +=       0.009 * cos(4.39355955971  + 224.34479570190 *T)
    sumL1  +=       0.009 * cos(5.76319803259  + 86457.98475793119 *T)
    sumL1  +=       0.009 * cos(3.21350384534  + 123668.80510662538 *T)
    sumL1  +=       0.009 * cos(5.50134612867  + 25352.02662804239 *T)
    sumL1  +=       0.009 * cos(3.61240341796  + 76784.84285065818 *T)
    sumL1  +=       0.009 * cos(3.71346537596  + 78213.71280281838 *T)
    sumL1  +=       0.011 * cos(0.04658635952  + 136722.59155786238 *T)
    sumL1  +=       0.009 * cos(3.16339317765  + 24079.34560241519 *T)
    sumL1  +=       0.010 * cos(1.52307057829  + 196137.07343261081 *T)
    sumL1  +=       0.009 * cos(5.78277172101  + 137210.22630911658 *T)
    sumL1  +=       0.009 * cos(5.34709321166  + 16703.06213349900 *T)
    sumL1  +=       0.010 * cos(1.10603431719  + 76152.05911134499 *T)
    sumL1  +=       0.010 * cos(2.92130709301  + 78257.08086582259 *T)
    sumL1  +=       0.009 * cos(4.92591389280  + 52609.51802102519 *T)
    sumL1  +=       0.008 * cos(4.47915729665  + 167850.08294774959 *T)
    sumL1  +=       0.010 * cos(3.52532046928  + 50049.92887576919 *T)
    sumL1  +=       0.009 * cos(5.74701495675  + 128220.75860368418 *T)
    sumL1  +=       0.009 * cos(3.39636103723  + 25771.51127191760 *T)
    sumL1  +=       0.009 * cos(0.40885845708  + 173511.41499690176 *T)
    sumL1  +=       0.009 * cos(0.05214771460  + 209232.91609768817 *T)
    sumL1  +=       0.011 * cos(4.39473500560  + 166740.70439565618 *T)
    sumL1  +=       0.008 * cos(0.83788793173  + 52101.02468458109 *T)
    sumL1  +=       0.009 * cos(0.39740145698  + 208173.53416749896 *T)
    sumL1  +=       0.009 * cos(2.14648409917  + 52250.58788171570 *T)
    sumL1  +=       0.011 * cos(0.77019715443  + 181026.24909573561 *T)
    sumL1  +=       0.008 * cos(0.46458507594  + 11.04570026390 *T)
    sumL1  +=       0.009 * cos(3.75726239516  + 26395.46076254379 *T)
    sumL1  +=       0.011 * cos(3.88183253605  + 99024.13645791401 *T)
    sumL1  +=       0.009 * cos(5.91589628233  + 26404.29501123080 *T)
    sumL1  +=       0.008 * cos(3.62038830716  + 28736.35796704720 *T)
    sumL1  +=       0.009 * cos(5.53218561478  + 50910.23880452200 *T)
    sumL1  +=       0.008 * cos(3.98222002302  + 51226.63067417859 *T)
    sumL1  +=       0.008 * cos(2.20332298794  + 35077.37084700659 *T)
    sumL1  +=       0.010 * cos(1.71606704784  + 132028.58860315481 *T)
    sumL1  +=       0.009 * cos(4.19828805921  + 123200.84011627098 *T)
    sumL1  +=       0.008 * cos(4.94949508683  + 1263.15736257819 *T)
    sumL1  +=       0.008 * cos(0.38255024740  + 78800.51393681798 *T)
    sumL1  +=       0.008 * cos(3.55931433878  + 25508.21555457540 *T)
    sumL1  +=       0.008 * cos(2.06623542364  + 133882.09065283637 *T)
    sumL1  +=       0.008 * cos(2.76004606287  + 103498.41618454478 *T)
    sumL1  +=       0.009 * cos(6.07146960892  + 52595.29092702359 *T)
    sumL1  +=       0.010 * cos(4.64768763301  + 240452.46032332000 *T)
    sumL1  +=       0.010 * cos(4.75089417351  + 158116.49174472899 *T)
    sumL1  +=       0.009 * cos(3.16830318341  + 156531.30018480320 *T)
    sumL1  +=       0.009 * cos(0.51442466727  + 233731.74634397856 *T)
    sumL1  +=       0.008 * cos(1.84170519176  + 52325.36948028299 *T)
    sumL1  +=       0.009 * cos(1.58020929410  + 130866.11389874700 *T)
    sumL1  +=       0.008 * cos(1.31021834242  + 78160.61665050399 *T)
    sumL1  +=       0.007 * cos(3.52394232228  + 209658.82487420217 *T)
    sumL1  +=       0.007 * cos(5.30361639156  + 51852.30086649099 *T)
    sumL1  +=       0.007 * cos(1.68129390597  + 103704.60173298200 *T)
    sumL1  +=       0.007 * cos(4.93711462539  + 77616.69859140778 *T)
    sumL1  +=       0.007 * cos(4.26547182656  + 130363.24963659538 *T)
    sumL1  +=       0.008 * cos(5.88262362064  + 51955.39364070959 *T)
    sumL1  +=       0.009 * cos(2.85041915486  + 52286.01260436779 *T)
    sumL1  +=       0.009 * cos(4.54751921819  + 222224.97657418496 *T)
    sumL1  +=       0.007 * cos(2.95164397426  + 58857.03113654799 *T)
    sumL1  +=       0.007 * cos(5.09932324366  + 26421.75908234360 *T)
    sumL1  +=       0.008 * cos(5.95900851407  + 625.67019231240 *T)
    sumL1  +=       0.008 * cos(2.87611784406  + 182595.65223011959 *T)
    sumL1  +=       0.007 * cos(2.95245442922  + 70269.18098269838 *T)
    sumL1  +=       0.009 * cos(1.85768285044  + 2168.76048228260 *T)
    sumL1  +=       0.009 * cos(5.87894263333  + 203041.89308344136 *T)
    sumL1  +=       0.007 * cos(5.76376514627  + 53399.62412392700 *T)
    sumL1  +=       0.007 * cos(3.19368974135  + 77947.31755506598 *T)
    sumL1  +=       0.007 * cos(0.58513513972  + 17893.62780836560 *T)
    sumL1  +=       0.007 * cos(1.25969543054  + 104275.34649502118 *T)
    sumL1  +=       0.007 * cos(1.78997113742  + 187167.27548807816 *T)
    sumL1  +=       0.007 * cos(3.04234073392  + 77101.23472031478 *T)
    sumL1  +=       0.007 * cos(5.13452237557  + 77307.42026875199 *T)
    sumL1  +=       0.008 * cos(2.39389370291  + 51534.39272140940 *T)
    sumL1  +=       0.007 * cos(0.36315569295  + 179172.74704605396 *T)
    sumL1  +=       0.009 * cos(4.65044386389  + 156954.01704032117 *T)
    sumL1  +=       0.007 * cos(5.75015377153  + 78580.10129437919 *T)
    sumL1  +=       0.008 * cos(4.90310056069  + 101703.15774082378 *T)
    sumL1  +=       0.007 * cos(2.81152781596  + 92741.06060792258 *T)
    sumL1  +=       0.007 * cos(4.53777069646  + 78339.97549599818 *T)
    sumL1  +=       0.008 * cos(3.72793390212  + 77795.74443436819 *T)
    sumL1  +=       0.007 * cos(4.99225219771  + 102659.44689679438 *T)
    sumL1  +=       0.008 * cos(1.24101149187  + 128857.55635958798 *T)
    sumL1  +=       0.006 * cos(1.24806705522  + 181659.72224941081 *T)
    sumL1  +=       0.007 * cos(1.41908776548  + 39450.35284837340 *T)
    sumL1  +=       0.008 * cos(5.07207947299  + 202420.14928260216 *T)
    sumL1  +=       0.006 * cos(3.88021248158  + 104991.50985261080 *T)
    sumL1  +=       0.007 * cos(0.44811727260  + 104301.61594439260 *T)
    sumL1  +=       0.007 * cos(6.19019459405  + 37698.45509994840 *T)
    sumL1  +=       0.008 * cos(3.27628873745  + 183145.01295611399 *T)
    sumL1  +=       0.006 * cos(1.04156861985  + 104197.83375581198 *T)
    sumL1  +=       0.007 * cos(6.04804574225  + 103189.13786188900 *T)
    sumL1  +=       0.007 * cos(6.15588549143  + 75930.51303185058 *T)
    sumL1  +=       0.007 * cos(2.01982955946  + 176332.24614102798 *T)
    sumL1  +=       0.007 * cos(4.68299502093  + 838.96928775040 *T)
    sumL1  +=       0.007 * cos(4.32694078051  + 104248.51979207818 *T)
    sumL1  +=       0.006 * cos(0.30098706976  + 54824.26110862140 *T)
    sumL1  +=       0.007 * cos(2.39268679107  + 80596.90581759460 *T)
    sumL1  +=       0.006 * cos(0.45279066685  + 225687.22128005017 *T)
    sumL1  +=       0.006 * cos(4.93617566760  + 215473.93573383920 *T)
    sumL1  +=       0.006 * cos(4.22528261713  + 79330.20490191258 *T)
    sumL1  +=       0.007 * cos(5.53062936347  + 917.93028598180 *T)
    sumL1  +=       0.006 * cos(0.27837413876  + 76137.83201734339 *T)
    sumL1  +=       0.007 * cos(1.50054882304  + 78313.70604662679 *T)
    sumL1  +=       0.006 * cos(4.84865195534  + 213255.17862965239 *T)
    sumL1  +=       0.007 * cos(0.37969316028  + 102872.74599223241 *T)
    sumL1  +=       0.006 * cos(4.82915128356  + 79859.89586700719 *T)
    sumL1  +=       0.007 * cos(2.36435599398  + 58220.03486452380 *T)
    sumL1  +=       0.007 * cos(0.76243719900  + 183674.70392120857 *T)
    sumL1  +=       0.006 * cos(3.48561717935  + 61560.64729122359 *T)
    sumL1  +=       0.007 * cos(3.15542325331  + 53029.00266490040 *T)
    sumL1  +=       0.006 * cos(2.04066112896  + 156534.53239644598 *T)
    sumL1  +=       0.006 * cos(4.08906102427  + 91919.56991589900 *T)
    sumL1  +=       0.006 * cos(3.64988511750  + 151199.94274106238 *T)
    sumL1  +=       0.006 * cos(4.09613031145  + 53867.97195265079 *T)
    sumL1  +=       0.006 * cos(3.49486002612  + 66653.15746634839 *T)
    sumL1  +=       0.007 * cos(2.89745929648  + 230239.17477710897 *T)
    sumL1  +=       0.006 * cos(3.99809593088  + 50380.54783942739 *T)
    sumL1  +=       0.006 * cos(3.65663654813  + 70383.62027138360 *T)
    sumL1  +=       0.006 * cos(6.23319937932  + 182619.20332637735 *T)
    sumL1  +=       0.006 * cos(0.15202166498  + 51859.41441349179 *T)
    sumL1  +=       0.006 * cos(0.36162525319  + 26823.77965510599 *T)
    sumL1  +=       0.006 * cos(5.03417122369  + 78786.28684281638 *T)
    sumL1  +=       0.006 * cos(2.61401898803  + 80382.47328510099 *T)
    sumL1  +=       0.006 * cos(5.06939782090  + 51841.95034237900 *T)
    sumL1  +=       0.008 * cos(4.96432797037  + 52396.21892558720 *T)
    sumL1  +=       0.006 * cos(3.74091556543  + 155475.15046625675 *T)
    sumL1  +=       0.006 * cos(4.29792825666  + 52812.80255517260 *T)
    sumL1  +=       0.007 * cos(6.02219529486  + 156507.74908854539 *T)
    sumL1  +=       0.005 * cos(6.28085632767  + 31415.37924995700 *T)
    sumL1  +=       0.006 * cos(2.70517720720  + 52492.19815280499 *T)
    sumL1  +=       0.007 * cos(6.25185953525  + 74935.57376844239 *T)
    sumL1  +=       0.005 * cos(6.03800330793  + 84944.93427812219 *T)
    sumL1  +=       0.007 * cos(2.08385155566  + 104819.57755665120 *T)
    sumL1  +=       0.005 * cos(2.79143401592  + 51315.49635439559 *T)
    sumL1  +=       0.007 * cos(0.15300257128  + 2014.98167179780 *T)
    sumL1  +=       0.006 * cos(5.45690475475  + 1243.48760167840 *T)
    sumL1  +=       0.005 * cos(1.72310829290  + 103718.82882698360 *T)
    sumL1  +=       0.005 * cos(0.05343175320  + 27044.19229754480 *T)
    sumL1  +=       0.005 * cos(1.35980682609  + 94329.77528619740 *T)
    sumL1  +=       0.007 * cos(5.94496795198  + 256327.07791868312 *T)
    sumL1  +=       0.006 * cos(1.88665916947  + 80174.90890793978 *T)
    sumL1  +=       0.005 * cos(2.15640465605  + 235900.50682626120 *T)
    sumL1  +=       0.005 * cos(5.97271372078  + 156314.11975400720 *T)
    sumL1  +=       0.006 * cos(2.66620904510  + 229129.79622501560 *T)
    sumL1  +=       0.005 * cos(2.52891655111  + 123758.29085447140 *T)
    sumL1  +=       0.006 * cos(1.20953605164  + 130459.18546877075 *T)
    sumL1  +=       0.005 * cos(4.06443360712  + 130285.73689738620 *T)
    sumL1  +=       0.007 * cos(1.34355308244  + 1731.12235293260 *T)
    sumL1  +=       0.005 * cos(2.36158299954  + 78896.49316403578 *T)
    sumL1  +=       0.005 * cos(0.96987247797  + 149288.74325784517 *T)
    sumL1  +=       0.005 * cos(0.76610110351  + 24864.08530079559 *T)
    sumL1  +=       0.005 * cos(1.95647355963  + 76571.54375522019 *T)
    sumL1  +=       0.005 * cos(1.13701088228  + 26312.24793727610 *T)
    sumL1  +=       0.005 * cos(1.05826639118  + 27780.06881107659 *T)
    sumL1  +=       0.005 * cos(4.54130791881  + 156740.71794488319 *T)
    sumL1  +=       0.006 * cos(3.58276567647  + 198489.93958638256 *T)
    sumL1  +=       0.005 * cos(0.51603916276  + 171292.65789271498 *T)
    sumL1  +=       0.006 * cos(3.03452226064  + 77741.13200662879 *T)
    sumL1  +=       0.005 * cos(1.46111595893  + 101011.00013959700 *T)
    sumL1  +=       0.005 * cos(0.02805624095  + 149756.70824819960 *T)
    sumL1  +=       0.005 * cos(5.10195955921  + 112231.70171963578 *T)
    sumL1  +=       0.005 * cos(4.87459278742  + 78413.27262185719 *T)
    sumL1  +=       0.007 * cos(3.69898590349  + 81604.32185142238 *T)
    sumL1  +=       0.005 * cos(0.76195152302  + 417406.45026518719 *T)
    sumL1  +=       0.005 * cos(0.55904094194  + 2383.19301477620 *T)
    sumL1  +=       0.005 * cos(3.95096384085  + 78188.92782615528 *T)
    sumL1  +=       0.005 * cos(1.76549372525  + 53771.99272543299 *T)
    sumL1  +=       0.005 * cos(3.35356708410  + 205260.65018762811 *T)
    sumL1  +=       0.006 * cos(5.57902902188  + 87253.17713015496 *T)
    sumL1  +=       0.005 * cos(0.26610630357  + 235746.72801577640 *T)
    sumL1  +=       0.005 * cos(2.90596877294  + 154194.22245657316 *T)
    sumL1  +=       0.004 * cos(1.16974176650  + 193937.98608932376 *T)
    sumL1  +=       0.004 * cos(3.93130462476  + 130289.95251073640 *T)
    sumL1  +=       0.004 * cos(4.33627478046  + 89.48574784600 *T)
    sumL1  +=       0.005 * cos(2.80022335340  + 4437.51420837359 *T)
    sumL1  +=       0.006 * cos(4.70286246928  + 91805.13062721380 *T)
    sumL1  +=       0.005 * cos(1.82284966431  + 1884.90116341740 *T)
    sumL1  +=       0.005 * cos(2.02580304988  + 259769.65286364852 *T)
    sumL1  +=       0.004 * cos(4.65165864815  + 195047.36464141720 *T)
    sumL1  +=       0.004 * cos(0.58858147700  + 323.50541665740 *T)
    sumL1  +=       0.004 * cos(4.00623668740  + 175376.64639941938 *T)
    sumL1  +=       0.004 * cos(0.64833718455  + 156377.85565231059 *T)
    sumL1  +=       0.004 * cos(2.71436739189  + 182402.02289558135 *T)
    sumL1  +=       0.005 * cos(4.68929437146  + 79315.97780791098 *T)
    sumL1  +=       0.004 * cos(5.71176695959  + 81591.84508100279 *T)
    sumL1  +=       0.004 * cos(1.42876132655  + 52252.07235442399 *T)
    sumL1  +=       0.005 * cos(2.55069425936  + 154308.66174525840 *T)
    sumL1  +=       0.005 * cos(1.42813778372  + 266540.36346489412 *T)
    sumL1  +=       0.004 * cos(2.00085806148  + 189386.03259226491 *T)
    sumL1  +=       0.005 * cos(4.29231489707  + 220025.88923089800 *T)
    sumL1  +=       0.004 * cos(6.20316213523  + 79116.90580647459 *T)
    sumL1  +=       0.004 * cos(0.85257120719  + 118007.47305747319 *T)
    sumL1  +=       0.004 * cos(5.77579934605  + 129799.61842155698 *T)
    sumL1  +=       0.004 * cos(4.47492094862  + 137678.19129947099 *T)
    sumL1  +=       0.004 * cos(1.64566631180  + 78697.42116259939 *T)
    sumL1  +=       0.005 * cos(2.17514536840  + 51329.72344839720 *T)
    sumL1  +=       0.006 * cos(4.78710650697  + 102755.42612401220 *T)
    sumL1  +=       0.004 * cos(1.46002058771  + 935.92998070881 *T)
    sumL1  +=       0.005 * cos(5.13889133658  + 130907.48069822539 *T)
    sumL1  +=       0.004 * cos(1.89321187110  + 2111.65031337760 *T)
    sumL1  +=       0.004 * cos(0.10839681347  + 38.13303563780 *T)
    sumL1  +=       0.005 * cos(4.42201446083  + 78057.52387628538 *T)
    sumL1  +=       0.005 * cos(5.65870377214  + 1300.82681820500 *T)
    sumL1  +=       0.005 * cos(5.90579064813  + 1478.86657406440 *T)
    sumL1  +=       0.004 * cos(5.24580055005  + 52309.91533273340 *T)
    sumL1  +=       0.004 * cos(2.74756161638  + 144916.86689107097 *T)
    sumL1  +=       0.005 * cos(3.31232177694  + 48847.67062686820 *T)
    sumL1  +=       0.004 * cos(4.43048849829  + 118828.96374949680 *T)
    sumL1  +=       0.004 * cos(0.51369916437  + 181563.05360783095 *T)
    sumL1  +=       0.004 * cos(5.22571179686  + 261988.40996783532 *T)
    sumL1  +=       0.004 * cos(0.80000960641  + 1485.29070670320 *T)
    sumL1  +=       0.004 * cos(3.25642623421  + 195181.47369100215 *T)
    sumL1  +=       0.005 * cos(5.23264312237  + 126996.94076290558 *T)
    sumL1  +=       0.004 * cos(3.86524951456  + 209762.60706278280 *T)
    sumL1  +=       0.004 * cos(3.43760746038  + 234261.43730907320 *T)
    sumL1  +=       0.004 * cos(0.67741502331  + 131079.41299418497 *T)
    sumL1  +=       0.004 * cos(5.40862384292  + 104984.39630560996 *T)
    sumL1  +=       0.004 * cos(0.30384598125  + 285907.55262712692 *T)
    sumL1  +=       0.004 * cos(5.65734972357  + 51868.24866217880 *T)
    sumL1  +=       0.004 * cos(5.55015227967  + 149846.19399604559 *T)
    sumL1  +=       0.004 * cos(1.51752118872  + 103814.80805420138 *T)
    sumL1  +=       0.004 * cos(2.53135799952  + 79487.52726550119 *T)
    sumL1  +=       0.004 * cos(1.88512485652  + 54879.42243782400 *T)
    sumL1  +=       0.004 * cos(5.12386500502  + 182622.43553802016 *T)
    sumL1  +=       0.004 * cos(4.36228014188  + 207747.62539098496 *T)
    sumL1  +=       0.004 * cos(3.24471199066  + 130020.03106399580 *T)
    sumL1  +=       0.004 * cos(0.51045647492  + 107692.22499299661 *T)
    sumL1  +=       0.004 * cos(4.26267840541  + 132658.27281205778 *T)
    sumL1  +=       0.004 * cos(3.08522628552  + 175844.61138977378 *T)
    sumL1  +=       0.004 * cos(3.38866894732  + 104888.41707839220 *T)
    sumL1  +=       0.004 * cos(1.21404057369  + 139543.42270198860 *T)
    sumL1  +=       0.005 * cos(3.75130732000  + 62197.64356324779 *T)
    sumL1  +=       0.004 * cos(0.30056239753  + 233202.05537888399 *T)
    sumL1  +=       0.004 * cos(4.20598588282  + 218916.51067880457 *T)
    sumL1  +=       0.004 * cos(1.34568004641  + 182828.62108645737 *T)
    sumL1  +=       0.004 * cos(0.68519128486  + 104276.83096772947 *T)
    sumL1  +=       0.004 * cos(0.20412928769  + 169093.57054942797 *T)
    sumL1  +=       0.003 * cos(0.98973904852  + 105418.10804348681 *T)
    sumL1  +=       0.004 * cos(3.42019897755  + 61279.71327726600 *T)
    sumL1  +=       0.003 * cos(5.22780381742  + 78338.49102328988 *T)
    sumL1  +=       0.003 * cos(2.67114319601  + 24998.19435038059 *T)
    sumL1  +=       0.004 * cos(1.65814724695  + 50800.03248330259 *T)
    sumL1  +=       0.004 * cos(3.24764679291  + 28286.99048486120 *T)
    sumL1  +=       0.004 * cos(6.01723954923  + 104237.17327761155 *T)
    sumL1  +=       0.003 * cos(3.99050522839  + 71378.55953479178 *T)
    sumL1  +=       0.003 * cos(3.46632047879  + 45494.58142974879 *T)
    sumL1  +=       0.003 * cos(4.27304181959  + 130593.29451835580 *T)
    sumL1  +=       0.003 * cos(4.05932083484  + 52822.81711646319 *T)
    sumL1  +=       0.003 * cos(5.89324219451  + 50483.64061364600 *T)
    sumL1  +=       0.003 * cos(2.88091404327  + 78149.27013603736 *T)
    sumL1  +=       0.003 * cos(4.09573256493  + 155460.92337225520 *T)
    sumL1  +=       0.003 * cos(0.40695406256  + 52278.89905736699 *T)
    sumL1  +=       0.003 * cos(0.53381434741  + 26709.64694241340 *T)
    sumL1  +=       0.003 * cos(1.00008179165  + 50264.60679993120 *T)
    sumL1  +=       0.004 * cos(5.67939124426  + 106470.37642667518 *T)
    sumL1  +=       0.003 * cos(0.77610404005  + 142871.55835826878 *T)
    sumL1  +=       0.003 * cos(3.80375047171  + 26402.08932144380 *T)
    sumL1  +=       0.003 * cos(2.44134109777  + 156520.30530244438 *T)
    sumL1  +=       0.003 * cos(0.98764974659  + 128843.32926558638 *T)
    sumL1  +=       0.003 * cos(3.62903815930  + 130435.63437251298 *T)
    sumL1  +=       0.003 * cos(2.12651302584  + 78469.89497315978 *T)
    sumL1  +=       0.003 * cos(4.29158350083  + 154408.65498906677 *T)
    sumL1  +=       0.003 * cos(1.06499703474  + 26734.91397488900 *T)
    sumL1  +=       0.003 * cos(1.86490629185  + 51707.84129279399 *T)
    sumL1  +=       0.003 * cos(5.95559848753  + 104344.98400739678 *T)
    sumL1  +=       0.004 * cos(3.60184879954  + 207114.15223730975 *T)
    sumL1  +=       0.003 * cos(3.28684847821  + 208276.62694171761 *T)
    sumL1  +=       0.003 * cos(2.80509981839  + 104358.24112519680 *T)
    sumL1  +=       0.003 * cos(5.02808401414  + 130432.40216087016 *T)
    sumL1  +=       0.003 * cos(6.16601650050  + 20043.67456019880 *T)
    sumL1  +=       0.003 * cos(5.29581547615  + 57503.28239153120 *T)
    sumL1  +=       0.003 * cos(1.35036448550  + 23919.14265929160 *T)
    sumL1  +=       0.002 * cos(3.75076268715  + 130005.80396999416 *T)
    sumL1  +=       0.003 * cos(3.05479754348  + 162810.49469943656 *T)
    sumL1  +=       0.003 * cos(2.06087055668  + 52065.59996192899 *T)
    sumL1  +=       0.003 * cos(1.11245383788  + 192828.60753723039 *T)
    sumL1  +=       0.002 * cos(1.90218644384  + 24356.78078864160 *T)
    sumL1  +=       0.002 * cos(4.74203919229  + 76887.93562487679 *T)
    sumL1  +=       0.002 * cos(0.63548334553  + 23384.28698689860 *T)
    sumL1  +=       0.002 * cos(2.51564214512  + 26667.59072857300 *T)
    sumL1  +=       0.002 * cos(6.08909078783  + 116783.65521669458 *T)
    sumL1  +=       0.002 * cos(3.56733241522  + 26073.67604757259 *T)
    sumL1  +=       0.003 * cos(5.60392923976  + 104466.05185498200 *T)
    sumL1  +=       0.002 * cos(3.56126103854  + 26189.86565983980 *T)
    sumL1  +=       0.002 * cos(5.97797555746  + 50167.24874398939 *T)
    sumL1  +=       0.002 * cos(4.41481991835  + 65538.25598994759 *T)
    sumL1  +=       0.002 * cos(5.91297170155  + 87367.61641884019 *T)
    sumL1  +=       0.002 * cos(5.17508169630  + 163298.12945069079 *T)
    sumL1  +=       0.003 * cos(1.10832319418  + 25600.26839032000 *T)
    sumL1  +=       0.002 * cos(2.11477268878  + 25773.71696170459 *T)
    sumL1  +=       0.002 * cos(2.15515419441  + 51951.46148744649 *T)
    sumL1  +=       0.003 * cos(5.43138498093  + 106684.80895916880 *T)
    sumL1  +=       0.002 * cos(0.11511311275  + 27170.98337386779 *T)
    sumL1  +=       0.002 * cos(5.22579437397  + 6885.14988993081 *T)
    sumL1  +=       0.002 * cos(2.93412437896  + 63786.35824152260 *T)
    sumL1  +=       0.003 * cos(6.19860078008  + 25004.82290928060 *T)
    sumL1  +=       0.002 * cos(0.23150025826  + 24448.83362438620 *T)
    sumL1  +=       0.002 * cos(0.54898566845  + 153084.84390447981 *T)
    sumL1  +=       0.002 * cos(0.63373893542  + 1265.56747862640 *T)
    sumL1  +=       0.002 * cos(4.49873870855  + 50007.04580086580 *T)
    sumL1  +=       0.002 * cos(5.57750243319  + 25928.60140678200 *T)
    sumL1  +=       0.002 * cos(5.66467419812  + 25466.15934073500 *T)
    sumL1  +=       0.002 * cos(5.10682926241  + 159969.99379441061 *T)
    sumL1  +=       0.002 * cos(3.49751807681  + 259819.64948555280 *T)

    # MERCURY Longitude
    sumL2  = 0
    sumL2  +=   53049.845 * cos(0.00000000000  + 0.00000000000 *T)
    sumL2  +=   16903.658 * cos(4.69072300649  + 26087.90314157420 *T)
    sumL2  +=    7396.711 * cos(1.34735624669  + 52175.80628314840 *T)
    sumL2  +=    3018.297 * cos(4.45643539705  + 78263.70942472259 *T)
    sumL2  +=    1107.419 * cos(1.26226537554  + 104351.61256629678 *T)
    sumL2  +=     378.173 * cos(4.31998055900  + 130439.51570787100 *T)
    sumL2  +=     122.998 * cos(1.06868541052  + 156527.41884944518 *T)
    sumL2  +=      38.663 * cos(4.08011610182  + 182615.32199101936 *T)
    sumL2  +=      14.898 * cos(4.63343085810  + 1109.37855209340 *T)
    sumL2  +=      11.861 * cos(0.79187646439  + 208703.22513259359 *T)
    sumL2  +=       5.243 * cos(4.71799772791  + 24978.52458948080 *T)
    sumL2  +=       3.575 * cos(3.77317513032  + 234791.12827416777 *T)
    sumL2  +=       2.566 * cos(1.44059109766  + 27197.28169366760 *T)
    sumL2  +=       2.045 * cos(1.49570544876  + 51066.42773105500 *T)
    sumL2  +=       1.064 * cos(0.46023695675  + 260879.03141574195 *T)
    sumL2  +=       0.972 * cos(1.80344701358  + 955.59974160860 *T)
    sumL2  +=       0.628 * cos(6.18483168149  + 529.69096509460 *T)
    sumL2  +=       0.628 * cos(4.84930105320  + 24498.83024629040 *T)
    sumL2  +=       0.763 * cos(4.54299337366  + 77154.33087262919 *T)
    sumL2  +=       0.654 * cos(0.98170349539  + 5661.33204915220 *T)
    sumL2  +=       0.572 * cos(5.92841937309  + 25028.52121138500 *T)
    sumL2  +=       0.552 * cos(2.13228646570  + 20426.57109242200 *T)
    sumL2  +=       0.507 * cos(4.11466890786  + 4551.95349705880 *T)
    sumL2  +=       0.608 * cos(4.54152310086  + 2218.75710418680 *T)
    sumL2  +=       0.439 * cos(3.40705996719  + 11322.66409830440 *T)
    sumL2  +=       0.381 * cos(2.83442391504  + 26617.59410666880 *T)
    sumL2  +=       0.370 * cos(0.82360646560  + 6770.71060124560 *T)
    sumL2  +=       0.312 * cos(5.18785829630  + 46514.47423399620 *T)
    sumL2  +=       0.314 * cos(3.42002895816  + 286966.93455731618 *T)
    sumL2  +=       0.307 * cos(3.20507351217  + 1059.38193018920 *T)
    sumL2  +=       0.332 * cos(0.97416155325  + 25132.30339996560 *T)
    sumL2  +=       0.327 * cos(2.22989850002  + 21535.94964451540 *T)
    sumL2  +=       0.303 * cos(4.68482058244  + 27043.50288318280 *T)
    sumL2  +=       0.247 * cos(5.87612469398  + 16983.99614745660 *T)
    sumL2  +=       0.311 * cos(4.22422008005  + 3442.57494496540 *T)
    sumL2  +=       0.237 * cos(4.37101224231  + 114.43928868521 *T)
    sumL2  +=       0.234 * cos(0.93917232374  + 30639.85663863300 *T)
    sumL2  +=       0.273 * cos(1.29086264556  + 103242.23401420339 *T)
    sumL2  +=       0.209 * cos(3.30158370895  + 213.29909543800 *T)
    sumL2  +=       0.197 * cos(2.07792605062  + 22645.32819660879 *T)
    sumL2  +=       0.182 * cos(0.62534354989  + 25448.00585526019 *T)
    sumL2  +=       0.219 * cos(1.34188738340  + 28306.66024576099 *T)
    sumL2  +=       0.168 * cos(6.20806483822  + 27147.28507176339 *T)
    sumL2  +=       0.156 * cos(3.91999448441  + 32858.61374281979 *T)
    sumL2  +=       0.160 * cos(5.79377449410  + 52705.49724824299 *T)
    sumL2  +=       0.153 * cos(4.33936015634  + 1589.07289528380 *T)
    sumL2  +=       0.161 * cos(4.80105233203  + 23869.14603738740 *T)
    sumL2  +=       0.148 * cos(1.97575825359  + 72602.37737557039 *T)
    sumL2  +=       0.145 * cos(5.96024756685  + 14765.23904326980 *T)
    sumL2  +=       0.129 * cos(4.16213277137  + 23969.13928119580 *T)
    sumL2  +=       0.135 * cos(3.86775437070  + 51220.20654153979 *T)
    sumL2  +=       0.140 * cos(3.66642468738  + 25558.21217647960 *T)
    sumL2  +=       0.108 * cos(1.39723983534  + 50586.73338786459 *T)
    sumL2  +=       0.115 * cos(0.15124159675  + 26301.20223701220 *T)
    sumL2  +=       0.113 * cos(4.98137485250  + 79373.08797681599 *T)
    sumL2  +=       0.115 * cos(3.05230359467  + 53285.18483524180 *T)
    sumL2  +=       0.099 * cos(3.99635236807  + 56727.75978020720 *T)
    sumL2  +=       0.099 * cos(3.84242768306  + 83925.04147387479 *T)
    sumL2  +=       0.093 * cos(2.04998627084  + 26068.23338067440 *T)
    sumL2  +=       0.092 * cos(0.08628005998  + 313054.83769889042 *T)
    sumL2  +=       0.094 * cos(4.30968259383  + 129330.13715577760 *T)
    sumL2  +=       0.085 * cos(3.01984486981  + 53235.18821333759 *T)
    sumL2  +=       0.105 * cos(5.32363189985  + 7238.67559160000 *T)
    sumL2  +=       0.095 * cos(2.24093904831  + 19317.19254032860 *T)
    sumL2  +=       0.099 * cos(1.00218248773  + 29530.47808653960 *T)
    sumL2  +=       0.075 * cos(5.40269356209  + 26091.78447693220 *T)
    sumL2  +=       0.074 * cos(4.31017375756  + 2333.19639287200 *T)
    sumL2  +=       0.086 * cos(0.78555390902  + 57837.13833230060 *T)
    sumL2  +=       0.081 * cos(0.95889229790  + 26107.57290247399 *T)
    sumL2  +=       0.085 * cos(3.28561166885  + 12432.04265039780 *T)
    sumL2  +=       0.070 * cos(3.55533072657  + 13521.75144159140 *T)
    sumL2  +=       0.078 * cos(0.27707431029  + 37410.56723987860 *T)
    sumL2  +=       0.063 * cos(0.65828525440  + 110012.94461544900 *T)
    sumL2  +=       0.064 * cos(2.68437002997  + 43071.89928903080 *T)
    sumL2  +=       0.067 * cos(3.89216578780  + 26084.02180621620 *T)
    sumL2  +=       0.064 * cos(2.75958867988  + 40853.14218484400 *T)
    sumL2  +=       0.061 * cos(5.77836694481  + 12566.15169998280 *T)
    sumL2  +=       0.069 * cos(3.44813862795  + 10213.28554621100 *T)
    sumL2  +=       0.060 * cos(0.79452201416  + 53131.40602475700 *T)
    sumL2  +=       0.063 * cos(5.02106928248  + 98690.28051714458 *T)
    sumL2  +=       0.063 * cos(2.49160532665  + 78793.40038981718 *T)
    sumL2  +=       0.059 * cos(0.15541098170  + 36301.18868778519 *T)
    sumL2  +=       0.056 * cos(1.01839356017  + 27676.97603685800 *T)
    sumL2  +=       0.053 * cos(1.55958841442  + 49957.04917896160 *T)
    sumL2  +=       0.052 * cos(5.14041028675  + 52156.13652224860 *T)
    sumL2  +=       0.052 * cos(1.15062443564  + 26202.34243025941 *T)
    sumL2  +=       0.052 * cos(0.53687576880  + 77308.10968311399 *T)
    sumL2  +=       0.048 * cos(1.60851837486  + 25234.70675982219 *T)
    sumL2  +=       0.048 * cos(5.19507775798  + 48733.23133818299 *T)
    sumL2  +=       0.053 * cos(5.49113976101  + 25661.30495069820 *T)
    sumL2  +=       0.046 * cos(5.18676728087  + 77204.32749453338 *T)
    sumL2  +=       0.044 * cos(4.97145613800  + 25973.46385288896 *T)
    sumL2  +=       0.054 * cos(0.32708522231  + 25874.60404613620 *T)
    sumL2  +=       0.042 * cos(2.21900478384  + 52179.68761850640 *T)
    sumL2  +=       0.045 * cos(2.52716095523  + 103292.23063610760 *T)
    sumL2  +=       0.048 * cos(0.48944876530  + 51646.11531805379 *T)
    sumL2  +=       0.039 * cos(6.25275173960  + 47623.85278608960 *T)
    sumL2  +=       0.041 * cos(2.81015791586  + 41962.52073693740 *T)
    sumL2  +=       0.037 * cos(5.01445357044  + 26080.78959457339 *T)
    sumL2  +=       0.038 * cos(6.08585317961  + 79323.09135491178 *T)
    sumL2  +=       0.043 * cos(3.24574920597  + 52389.10537858640 *T)
    sumL2  +=       0.041 * cos(2.43000457106  + 639.89728631400 *T)
    sumL2  +=       0.039 * cos(0.76157461206  + 82815.66292178139 *T)
    sumL2  +=       0.034 * cos(3.28090243752  + 62389.09182935939 *T)
    sumL2  +=       0.037 * cos(0.05738918168  + 38519.94579197200 *T)
    sumL2  +=       0.035 * cos(5.78457148656  + 68050.42387851159 *T)
    sumL2  +=       0.040 * cos(3.14856042352  + 426.59819087600 *T)
    sumL2  +=       0.042 * cos(4.09553132731  + 52195.47604404819 *T)
    sumL2  +=       0.032 * cos(2.45518077222  + 38654.05484155699 *T)
    sumL2  +=       0.039 * cos(1.80165134294  + 33326.57873317420 *T)
    sumL2  +=       0.041 * cos(4.32141055637  + 26095.01668857500 *T)
    sumL2  +=       0.032 * cos(3.72358129034  + 136100.84775702321 *T)
    sumL2  +=       0.040 * cos(5.69015360764  + 31749.23519072640 *T)
    sumL2  +=       0.032 * cos(1.65828846012  + 105460.99111839020 *T)
    sumL2  +=       0.032 * cos(1.02549984810  + 155418.04029735178 *T)
    sumL2  +=       0.030 * cos(1.57915654792  + 52168.69273614759 *T)
    sumL2  +=       0.031 * cos(0.69335015231  + 50057.04242277000 *T)
    sumL2  +=       0.033 * cos(4.04678616141  + 55618.38122811380 *T)
    sumL2  +=       0.029 * cos(3.72464718175  + 51535.90899683439 *T)
    sumL2  +=       0.029 * cos(5.73654447860  + 129380.13377768178 *T)
    sumL2  +=       0.038 * cos(3.74472554242  + 18849.22754997420 *T)
    sumL2  +=       0.028 * cos(3.05059885517  + 339142.74084046454 *T)
    sumL2  +=       0.035 * cos(0.74194291507  + 52171.92494779040 *T)
    sumL2  +=       0.028 * cos(0.73918545262  + 7880.08915333899 *T)
    sumL2  +=       0.036 * cos(1.26742819742  + 51116.42435295920 *T)
    sumL2  +=       0.035 * cos(5.29298143160  + 45405.09568190280 *T)
    sumL2  +=       0.032 * cos(5.79438282818  + 18093.37469954999 *T)
    sumL2  +=       0.026 * cos(5.79318346346  + 66941.04532641819 *T)
    sumL2  +=       0.028 * cos(4.27104078603  + 54394.56338733519 *T)
    sumL2  +=       0.024 * cos(5.06304791610  + 2118.76386037840 *T)
    sumL2  +=       0.030 * cos(1.05420470769  + 52182.91983014920 *T)
    sumL2  +=       0.025 * cos(1.75996910240  + 124778.18365871880 *T)
    sumL2  +=       0.022 * cos(4.43631488516  + 3328.13565628019 *T)
    sumL2  +=       0.022 * cos(5.09106835544  + 99799.65906923798 *T)
    sumL2  +=       0.022 * cos(1.88199362566  + 78244.03966382280 *T)
    sumL2  +=       0.021 * cos(2.56123211200  + 94138.32702008578 *T)
    sumL2  +=       0.025 * cos(5.46150069597  + 104881.30353139140 *T)
    sumL2  +=       0.025 * cos(0.05236467599  + 26514.50133245020 *T)
    sumL2  +=       0.025 * cos(3.02983070039  + 77734.01845962799 *T)
    sumL2  +=       0.020 * cos(3.48966334599  + 103396.01282468818 *T)
    sumL2  +=       0.021 * cos(3.97495995227  + 53764.87917843220 *T)
    sumL2  +=       0.018 * cos(1.07895331282  + 28421.09953444620 *T)
    sumL2  +=       0.019 * cos(1.98308917333  + 23754.70674870219 *T)
    sumL2  +=       0.018 * cos(4.58631365960  + 76044.95232053580 *T)
    sumL2  +=       0.019 * cos(2.90042907299  + 6283.07584999140 *T)
    sumL2  +=       0.019 * cos(3.18954246101  + 51109.31080595839 *T)
    sumL2  +=       0.018 * cos(2.11231697626  + 73711.75592766379 *T)
    sumL2  +=       0.017 * cos(0.06118735536  + 88476.99497093359 *T)
    sumL2  +=       0.018 * cos(6.26984143774  + 15874.61759536320 *T)
    sumL2  +=       0.018 * cos(2.58820345674  + 51123.53789995999 *T)
    sumL2  +=       0.017 * cos(5.82864656129  + 32370.97899156560 *T)
    sumL2  +=       0.018 * cos(5.22519288335  + 78267.59076008058 *T)
    sumL2  +=       0.015 * cos(5.43053381026  + 64741.95798313119 *T)
    sumL2  +=       0.015 * cos(2.53107093357  + 155468.03691925600 *T)
    sumL2  +=       0.020 * cos(4.40959528115  + 1223.81784077861 *T)
    sumL2  +=       0.015 * cos(4.05461375332  + 51962.50718771040 *T)
    sumL2  +=       0.015 * cos(3.52769852262  + 79219.30916633119 *T)
    sumL2  +=       0.015 * cos(0.45988103836  + 162188.75089859738 *T)
    sumL2  +=       0.015 * cos(0.87152359342  + 78283.37918562238 *T)
    sumL2  +=       0.014 * cos(1.87990160627  + 125887.56221081220 *T)
    sumL2  +=       0.016 * cos(2.82759858554  + 105410.99449648600 *T)
    sumL2  +=       0.013 * cos(4.67219046898  + 78256.59587772179 *T)
    sumL2  +=       0.016 * cos(0.04901690153  + 78477.00852016058 *T)
    sumL2  +=       0.014 * cos(4.21146699511  + 52290.24557183361 *T)
    sumL2  +=       0.014 * cos(4.08726245298  + 78270.82297172339 *T)
    sumL2  +=       0.015 * cos(3.80902990348  + 33967.99229491319 *T)
    sumL2  +=       0.016 * cos(0.28761005157  + 44937.13069154840 *T)
    sumL2  +=       0.014 * cos(5.84330605810  + 103821.92160120218 *T)
    sumL2  +=       0.015 * cos(6.03738264520  + 13655.86049117640 *T)
    sumL2  +=       0.012 * cos(0.81685952408  + 81706.28436968799 *T)
    sumL2  +=       0.015 * cos(3.79894620695  + 108903.56606335558 *T)
    sumL2  +=       0.011 * cos(3.26651828162  + 52602.40447402440 *T)
    sumL2  +=       0.011 * cos(1.28755195884  + 29416.03879785439 *T)
    sumL2  +=       0.011 * cos(4.03185618461  + 181505.94343892601 *T)
    sumL2  +=       0.011 * cos(5.58448046140  + 120226.23016166000 *T)
    sumL2  +=       0.012 * cos(0.32316787548  + 91785.46086631398 *T)
    sumL2  +=       0.013 * cos(2.02636565149  + 71492.99882347698 *T)
    sumL2  +=       0.015 * cos(3.56357453844  + 76674.63652943878 *T)
    sumL2  +=       0.013 * cos(3.80590035776  + 78259.82808936459 *T)
    sumL2  +=       0.010 * cos(0.04109363791  + 77197.21394753258 *T)
    sumL2  +=       0.010 * cos(6.00753285175  + 365230.64398203877 *T)
    sumL2  +=       0.010 * cos(1.74642821988  + 52061.36699446317 *T)
    sumL2  +=       0.013 * cos(2.79229719351  + 636.99627202420 *T)
    sumL2  +=       0.011 * cos(5.52780687723  + 26727.80042788820 *T)
    sumL2  +=       0.011 * cos(4.04728142335  + 45892.73043315699 *T)
    sumL2  +=       0.008 * cos(4.96985130540  + 151975.46535238638 *T)
    sumL2  +=       0.008 * cos(3.44547998787  + 23439.44831610119 *T)
    sumL2  +=       0.011 * cos(2.54429302352  + 44181.27784112419 *T)
    sumL2  +=       0.010 * cos(3.50274229100  + 65697.55772473979 *T)
    sumL2  +=       0.009 * cos(2.45809063705  + 93028.94846799241 *T)
    sumL2  +=       0.009 * cos(4.92568932060  + 104331.94280539700 *T)
    sumL2  +=       0.008 * cos(0.80535713418  + 71980.63357473118 *T)
    sumL2  +=       0.010 * cos(2.20026830808  + 130969.20667296558 *T)
    sumL2  +=       0.007 * cos(2.30885367448  + 18207.81398823521 *T)
    sumL2  +=       0.010 * cos(5.72020107388  + 77211.44104153418 *T)
    sumL2  +=       0.007 * cos(0.15536656581  + 129483.91596626240 *T)
    sumL2  +=       0.008 * cos(2.31255848253  + 58458.88213313979 *T)
    sumL2  +=       0.007 * cos(2.07713892797  + 90829.86112470538 *T)
    sumL2  +=       0.007 * cos(3.50101836609  + 117873.36400788820 *T)
    sumL2  +=       0.007 * cos(3.20275960823  + 64607.84893354619 *T)
    sumL2  +=       0.007 * cos(4.73829767381  + 51322.60990139639 *T)
    sumL2  +=       0.007 * cos(5.65553946700  + 181555.94006083015 *T)
    sumL2  +=       0.009 * cos(2.56870006051  + 129909.82474277640 *T)
    sumL2  +=       0.008 * cos(0.67171808844  + 79852.78232000639 *T)
    sumL2  +=       0.007 * cos(4.06204756163  + 39609.65458316560 *T)
    sumL2  +=       0.007 * cos(1.46575838897  + 853.19638175200 *T)
    sumL2  +=       0.007 * cos(2.57400610109  + 69159.80243060499 *T)
    sumL2  +=       0.006 * cos(3.86283147263  + 59414.48187474840 *T)
    sumL2  +=       0.006 * cos(1.33545681950  + 102132.85546211000 *T)
    sumL2  +=       0.007 * cos(2.80506785787  + 39743.76363275060 *T)
    sumL2  +=       0.006 * cos(1.66711581835  + 28206.66700195260 *T)
    sumL2  +=       0.008 * cos(4.73806237855  + 150866.08680029298 *T)
    sumL2  +=       0.006 * cos(0.81325444169  + 207593.84658050016 *T)
    sumL2  +=       0.006 * cos(0.06925555540  + 58946.51688439399 *T)
    sumL2  +=       0.006 * cos(3.09908354288  + 114564.89811250778 *T)
    sumL2  +=       0.006 * cos(0.88473087003  + 104358.72611329758 *T)
    sumL2  +=       0.006 * cos(3.51715962887  + 188276.65404017159 *T)
    sumL2  +=       0.006 * cos(4.13083556162  + 54509.00267602040 *T)
    sumL2  +=       0.006 * cos(3.14680331401  + 104564.91166173480 *T)
    sumL2  +=       0.007 * cos(5.89465282471  + 131498.89763806018 *T)
    sumL2  +=       0.005 * cos(0.27174681491  + 143961.26714946239 *T)
    sumL2  +=       0.005 * cos(5.25316838887  + 3340.61242669980 *T)
    sumL2  +=       0.007 * cos(1.99860349328  + 104355.49390165480 *T)
    sumL2  +=       0.005 * cos(1.47362369989  + 9103.90699411760 *T)
    sumL2  +=       0.007 * cos(0.60725013988  + 60055.89543648739 *T)
    sumL2  +=       0.005 * cos(3.20081630256  + 13541.42120249119 *T)
    sumL2  +=       0.006 * cos(0.50137409605  + 63498.47038145279 *T)
    sumL2  +=       0.005 * cos(5.07739475360  + 97580.90196505121 *T)
    sumL2  +=       0.005 * cos(1.35632842517  + 104344.49901929598 *T)
    sumL2  +=       0.005 * cos(2.16611734526  + 103925.01437542078 *T)
    sumL2  +=       0.005 * cos(4.04581717475  + 98068.53671630540 *T)
    sumL2  +=       0.005 * cos(4.90166806916  + 22759.76748529401 *T)
    sumL2  +=       0.005 * cos(3.87246722733  + 89586.37352302698 *T)
    sumL2  +=       0.005 * cos(3.01510200596  + 103285.11708910680 *T)
    sumL2  +=       0.006 * cos(1.44791474080  + 78050.41032928458 *T)
    sumL2  +=       0.005 * cos(2.47831218420  + 26241.68195205900 *T)
    sumL2  +=       0.005 * cos(3.85799454501  + 107794.18751126219 *T)
    sumL2  +=       0.005 * cos(0.57300969993  + 134991.46920492980 *T)
    sumL2  +=       0.005 * cos(2.38724530268  + 146314.13330323418 *T)
    sumL2  +=       0.004 * cos(6.11121027497  + 90695.75207512038 *T)
    sumL2  +=       0.004 * cos(1.83181231470  + 74821.13447975718 *T)
    sumL2  +=       0.004 * cos(5.23938469806  + 130012.91751699500 *T)
    sumL2  +=       0.005 * cos(1.18020888905  + 27311.72098235281 *T)
    sumL2  +=       0.005 * cos(3.00001411726  + 85034.42002596818 *T)
    sumL2  +=       0.005 * cos(5.56662713374  + 155997.72788435058 *T)
    sumL2  +=       0.005 * cos(2.35342507828  + 145204.75475114080 *T)
    sumL2  +=       0.004 * cos(4.57119707855  + 157636.79740153858 *T)
    sumL2  +=       0.004 * cos(5.78384635962  + 51742.09454527159 *T)
    sumL2  +=       0.004 * cos(2.41778026187  + 157483.01859105378 *T)
    sumL2  +=       0.005 * cos(2.53226316949  + 103299.34418310841 *T)
    sumL2  +=       0.004 * cos(3.71458526250  + 26555.86813192860 *T)
    sumL2  +=       0.004 * cos(3.83639975710  + 150244.34299945380 *T)
    sumL2  +=       0.004 * cos(3.30201863644  + 76144.94556434419 *T)
    sumL2  +=       0.004 * cos(0.68904721853  + 124156.43985787958 *T)
    sumL2  +=       0.004 * cos(1.73206522989  + 178063.36849396059 *T)
    sumL2  +=       0.005 * cos(3.20991489892  + 71025.03383312259 *T)
    sumL2  +=       0.005 * cos(3.69408876581  + 105940.68546158058 *T)
    sumL2  +=       0.004 * cos(2.75867944254  + 391318.54712361295 *T)
    sumL2  +=       0.005 * cos(0.53586068200  + 80482.46652890938 *T)
    sumL2  +=       0.004 * cos(5.07565982849  + 51756.32163927320 *T)
    sumL2  +=       0.004 * cos(0.12462379137  + 78690.30761559859 *T)
    sumL2  +=       0.004 * cos(5.25512859919  + 157057.10981453978 *T)
    sumL2  +=       0.004 * cos(5.66367904256  + 9384.84100807520 *T)
    sumL2  +=       0.004 * cos(4.70366039546  + 54294.57014352679 *T)
    sumL2  +=       0.004 * cos(4.74767635950  + 204151.27163553479 *T)
    sumL2  +=       0.003 * cos(3.24177385660  + 155571.81910783658 *T)
    sumL2  +=       0.003 * cos(0.85377908455  + 16066.06586147480 *T)
    sumL2  +=       0.003 * cos(1.58566666033  + 130226.21661243298 *T)
    sumL2  +=       0.004 * cos(5.58218314089  + 119116.85160956658 *T)
    sumL2  +=       0.004 * cos(0.94898112028  + 78378.14871340781 *T)
    sumL2  +=       0.004 * cos(6.15781348253  + 97112.93697469680 *T)
    sumL2  +=       0.004 * cos(2.31787867435  + 52815.70356946240 *T)
    sumL2  +=       0.004 * cos(2.73401660736  + 78417.48823520739 *T)
    sumL2  +=       0.004 * cos(5.67762508473  + 183674.70392120857 *T)
    sumL2  +=       0.004 * cos(0.21138526735  + 214364.55718174577 *T)
    sumL2  +=       0.004 * cos(5.24311464709  + 116917.76426627958 *T)
    sumL2  +=       0.003 * cos(3.86634858823  + 104371.28232719658 *T)
    sumL2  +=       0.004 * cos(5.42881221134  + 233731.74634397856 *T)
    sumL2  +=       0.003 * cos(1.97251042531  + 183145.01295611399 *T)
    sumL2  +=       0.003 * cos(0.60740481915  + 133882.09065283637 *T)
    sumL2  +=       0.003 * cos(6.23399386415  + 181659.72224941081 *T)
    sumL2  +=       0.004 * cos(6.22336284288  + 140652.80125408198 *T)
    sumL2  +=       0.003 * cos(5.94876130841  + 25035.63475838580 *T)
    sumL2  +=       0.003 * cos(3.83428144879  + 233681.74972207440 *T)
    sumL2  +=       0.003 * cos(2.32902857669  + 182085.63102592481 *T)
    sumL2  +=       0.003 * cos(2.84354681160  + 123200.84011627098 *T)
    sumL2  +=       0.003 * cos(1.53933182823  + 176953.98994186718 *T)
    sumL2  +=       0.004 * cos(2.45893471155  + 207643.84320240439 *T)
    sumL2  +=       0.003 * cos(5.72001402121  + 102762.53967101300 *T)
    sumL2  +=       0.004 * cos(1.40880562474  + 80382.47328510099 *T)
    sumL2  +=       0.003 * cos(4.90489886883  + 51749.20809227239 *T)
    sumL2  +=       0.003 * cos(4.76799924569  + 104138.31347085880 *T)
    sumL2  +=       0.004 * cos(6.26291139255  + 25021.40766438420 *T)
    sumL2  +=       0.003 * cos(0.40352872921  + 132028.58860315481 *T)
    sumL2  +=       0.003 * cos(2.66696150741  + 157586.80077963436 *T)
    sumL2  +=       0.003 * cos(5.23202292846  + 84546.78527471398 *T)
    sumL2  +=       0.003 * cos(4.59428408981  + 131548.89425996438 *T)
    sumL2  +=       0.003 * cos(2.53649995681  + 50579.61984086379 *T)
    sumL2  +=       0.003 * cos(2.55542640702  + 77829.99768684579 *T)
    sumL2  +=       0.004 * cos(1.67304824627  + 130419.84594697120 *T)
    sumL2  +=       0.003 * cos(6.21862713072  + 95247.70557217918 *T)
    sumL2  +=       0.003 * cos(5.37878648081  + 44295.71712980940 *T)
    sumL2  +=       0.003 * cos(4.68342323732  + 156507.74908854539 *T)
    sumL2  +=       0.003 * cos(3.36164307862  + 104778.21075717278 *T)
    sumL2  +=       0.003 * cos(2.91632245418  + 166740.70439565618 *T)
    sumL2  +=       0.002 * cos(0.72039482289  + 7994.52844202420 *T)
    sumL2  +=       0.002 * cos(5.43309839263  + 78903.60671103658 *T)
    sumL2  +=       0.002 * cos(1.12849064703  + 183724.70054311279 *T)
    sumL2  +=       0.003 * cos(1.94668130835  + 143005.66740785379 *T)
    sumL2  +=       0.003 * cos(0.56289392960  + 104347.73123093881 *T)
    sumL2  +=       0.002 * cos(1.78482957305  + 156531.30018480320 *T)
    sumL2  +=       0.003 * cos(5.03542424850  + 130443.39704322898 *T)
    sumL2  +=       0.003 * cos(4.72735160431  + 78149.27013603736 *T)
    sumL2  +=       0.002 * cos(5.31651763274  + 208173.53416749896 *T)
    sumL2  +=       0.003 * cos(5.36470448658  + 131395.11544947961 *T)
    sumL2  +=       0.002 * cos(0.06581745526  + 196137.07343261081 *T)
    sumL2  +=       0.002 * cos(2.93678015526  + 167850.08294774959 *T)
    sumL2  +=       0.002 * cos(5.84781134233  + 65831.66677432480 *T)
    sumL2  +=       0.002 * cos(5.56634055777  + 70269.18098269838 *T)
    sumL2  +=       0.002 * cos(5.42145220392  + 183570.92173262799 *T)
    sumL2  +=       0.002 * cos(6.06636386319  + 53242.30176033840 *T)
    sumL2  +=       0.002 * cos(4.37154457896  + 128220.75860368418 *T)
    sumL2  +=       0.002 * cos(0.35903300236  + 187167.27548807816 *T)
    sumL2  +=       0.002 * cos(2.78347515438  + 79330.20490191258 *T)
    sumL2  +=       0.002 * cos(4.51952536661  + 203041.89308344136 *T)
    sumL2  +=       0.002 * cos(3.36865001265  + 170049.17029103660 *T)
    sumL2  +=       0.002 * cos(3.08184453618  + 50593.84693486539 *T)
    sumL2  +=       0.002 * cos(4.32534428501  + 55503.94193942859 *T)
    sumL2  +=       0.002 * cos(1.81636961816  + 110634.68841628821 *T)
    sumL2  +=       0.002 * cos(5.58793955981  + 103917.90082841998 *T)
    sumL2  +=       0.002 * cos(5.79409172557  + 417406.45026518719 *T)
    sumL2  +=       0.002 * cos(0.85923934181  + 235900.50682626120 *T)
    sumL2  +=       0.002 * cos(4.60084058009  + 26941.09952332620 *T)
    sumL2  +=       0.002 * cos(5.14995676460  + 182188.72380014337 *T)
    sumL2  +=       0.002 * cos(3.91756419842  + 130446.62925487180 *T)
    sumL2  +=       0.002 * cos(2.88997273266  + 27154.39861876420 *T)
    sumL2  +=       0.003 * cos(5.85721944439  + 52329.58509363319 *T)
    sumL2  +=       0.002 * cos(6.10145799164  + 77623.81213840858 *T)
    sumL2  +=       0.002 * cos(1.98246165585  + 77844.22478084739 *T)
    sumL2  +=       0.002 * cos(0.11071759480  + 130866.11389874700 *T)
    sumL2  +=       0.002 * cos(0.64849171165  + 52643.77127350280 *T)
    sumL2  +=       0.002 * cos(3.98410950717  + 104466.05185498200 *T)
    sumL2  +=       0.002 * cos(1.83715639115  + 123668.80510662538 *T)
    sumL2  +=       0.002 * cos(5.61305711026  + 76667.52298243798 *T)
    sumL2  +=       0.002 * cos(5.83992902277  + 19804.82729158280 *T)
    sumL2  +=       0.002 * cos(0.54331921743  + 130459.18546877075 *T)
    sumL2  +=       0.002 * cos(0.00997429525  + 141762.17980617538 *T)
    sumL2  +=       0.002 * cos(5.59377744943  + 35191.81013569180 *T)
    sumL2  +=       0.002 * cos(0.39746661127  + 26521.61487945100 *T)
    sumL2  +=       0.002 * cos(6.16192704020  + 130652.81480330898 *T)
    sumL2  +=       0.002 * cos(0.85416063806  + 25934.12433108940 *T)
    sumL2  +=       0.002 * cos(2.18339110714  + 156100.82065856917 *T)
    sumL2  +=       0.002 * cos(4.39281517355  + 130432.40216087016 *T)
    sumL2  +=       0.002 * cos(5.50117730009  + 172402.03644480839 *T)
    sumL2  +=       0.001 * cos(3.49234867738  + 130435.63437251298 *T)
    sumL2  +=       0.001 * cos(4.80995492080  + 77837.11123384659 *T)
    sumL2  +=       0.002 * cos(6.26921260007  + 87367.61641884019 *T)
    sumL2  +=       0.002 * cos(2.79561490409  + 115674.27666460120 *T)
    sumL2  +=       0.002 * cos(4.92122664455  + 24864.08530079559 *T)
    sumL2  +=       0.001 * cos(2.41576832330  + 26037.90651966999 *T)
    sumL2  +=       0.001 * cos(6.04596837149  + 129373.02023068098 *T)
    sumL2  +=       0.002 * cos(0.94020572434  + 80596.90581759460 *T)
    sumL2  +=       0.002 * cos(0.30073776192  + 51639.00177105299 *T)
    sumL2  +=       0.001 * cos(0.18846827752  + 34082.43158359840 *T)
    sumL2  +=       0.001 * cos(3.62844435085  + 161079.37234650398 *T)
    sumL2  +=       0.002 * cos(5.13133615893  + 49842.60989027639 *T)
    sumL2  +=       0.001 * cos(0.30832555285  + 53228.07466633679 *T)
    sumL2  +=       0.001 * cos(5.56519625382  + 129387.24732468258 *T)
    sumL2  +=       0.001 * cos(5.63110777945  + 46848.33017476560 *T)
    sumL2  +=       0.001 * cos(1.38420566741  + 26011.63707029860 *T)
    sumL2  +=       0.001 * cos(5.86450783334  + 26724.89941359840 *T)
    sumL2  +=       0.001 * cos(6.14517818644  + 76681.75007643958 *T)
    sumL2  +=       0.001 * cos(3.19299019423  + 25654.19140369739 *T)
    sumL2  +=       0.001 * cos(0.34495954566  + 128850.44281258718 *T)
    sumL2  +=       0.001 * cos(3.60230008239  + 20760.42703319140 *T)
    sumL2  +=       0.001 * cos(5.80015495751  + 19406.67828817460 *T)
    sumL2  +=       0.001 * cos(4.42116052212  + 100909.03762133140 *T)
    sumL2  +=       0.001 * cos(0.48535289275  + 52698.38370124219 *T)
    sumL2  +=       0.001 * cos(6.22940855233  + 111590.28815789679 *T)
    sumL2  +=       0.001 * cos(0.16055219324  + 105307.21230790540 *T)
    sumL2  +=       0.001 * cos(2.19795772042  + 29428.51556827400 *T)
    sumL2  +=       0.001 * cos(5.09783503526  + 102232.84870591838 *T)
    sumL2  +=       0.001 * cos(0.61014627166  + 25668.41849769900 *T)
    sumL2  +=       0.001 * cos(5.62459774786  + 25565.32572348040 *T)
    sumL2  +=       0.001 * cos(0.90991571866  + 103711.71527998280 *T)
    sumL2  +=       0.001 * cos(0.28037611495  + 25450.90686955000 *T)
    sumL2  +=       0.001 * cos(0.54824728173  + 26137.89976347840 *T)
    sumL2  +=       0.001 * cos(4.25324476801  + 209812.60368468691 *T)
    sumL2  +=       0.001 * cos(0.08089899001  + 51543.02254383520 *T)
    sumL2  +=       0.001 * cos(0.33953378597  + 51528.79544983359 *T)
    sumL2  +=       0.001 * cos(4.15667461533  + 25551.09862947879 *T)
    sumL2  +=       0.001 * cos(1.59597814656  + 72936.23331633979 *T)
    sumL2  +=       0.001 * cos(5.59362222240  + 25619.93815121980 *T)
    sumL2  +=       0.001 * cos(3.26520760555  + 240452.46032332000 *T)
    sumL2  +=       0.001 * cos(5.52286605071  + 52125.80966124419 *T)

    # MERCURY Longitude
    sumL3  = 0
    sumL3  +=     188.077 * cos(0.03466830117  + 52175.80628314840 *T)
    sumL3  +=     142.152 * cos(3.12505452600  + 26087.90314157420 *T)
    sumL3  +=      96.877 * cos(3.00378171915  + 78263.70942472259 *T)
    sumL3  +=      43.669 * cos(6.01867965826  + 104351.61256629678 *T)
    sumL3  +=      35.395 * cos(0.00000000000  + 0.00000000000 *T)
    sumL3  +=      18.045 * cos(2.77538373991  + 130439.51570787100 *T)
    sumL3  +=       6.971 * cos(5.81808665742  + 156527.41884944518 *T)
    sumL3  +=       2.556 * cos(2.57014364454  + 182615.32199101936 *T)
    sumL3  +=       0.900 * cos(5.59308888939  + 208703.22513259359 *T)
    sumL3  +=       0.307 * cos(2.32189002493  + 234791.12827416777 *T)
    sumL3  +=       0.117 * cos(3.16567324160  + 24978.52458948080 *T)
    sumL3  +=       0.102 * cos(5.32134064237  + 260879.03141574195 *T)
    sumL3  +=       0.077 * cos(6.24155593431  + 51066.42773105500 *T)
    sumL3  +=       0.057 * cos(6.11222330831  + 27197.28169366760 *T)
    sumL3  +=       0.034 * cos(2.03244612467  + 286966.93455731618 *T)
    sumL3  +=       0.038 * cos(2.99912912226  + 77154.33087262919 *T)
    sumL3  +=       0.031 * cos(3.11483471984  + 53285.18483524180 *T)
    sumL3  +=       0.017 * cos(6.02950226714  + 103242.23401420339 *T)
    sumL3  +=       0.011 * cos(5.00109293882  + 313054.83769889042 *T)
    sumL3  +=       0.009 * cos(0.59396483404  + 20426.57109242200 *T)
    sumL3  +=       0.009 * cos(3.65190477342  + 46514.47423399620 *T)
    sumL3  +=       0.009 * cos(6.28305510644  + 79373.08797681599 *T)
    sumL3  +=       0.008 * cos(2.74138109305  + 129330.13715577760 *T)
    sumL3  +=       0.007 * cos(5.79018632644  + 25132.30339996560 *T)
    sumL3  +=       0.005 * cos(0.33554769835  + 72602.37737557039 *T)
    sumL3  +=       0.005 * cos(1.31115810350  + 26617.59410666880 *T)
    sumL3  +=       0.004 * cos(2.46414914425  + 51220.20654153979 *T)
    sumL3  +=       0.004 * cos(1.74528964721  + 339142.74084046454 *T)
    sumL3  +=       0.004 * cos(4.31141963105  + 52705.49724824299 *T)
    sumL3  +=       0.004 * cos(3.07093349506  + 27043.50288318280 *T)
    sumL3  +=       0.004 * cos(5.68013701250  + 30639.85663863300 *T)
    sumL3  +=       0.003 * cos(5.80662640937  + 155418.04029735178 *T)
    sumL3  +=       0.003 * cos(4.64519088060  + 27147.28507176339 *T)
    sumL3  +=       0.004 * cos(3.02893592955  + 23869.14603738740 *T)
    sumL3  +=       0.003 * cos(4.35199345075  + 79323.09135491178 *T)
    sumL3  +=       0.003 * cos(1.29040630287  + 53235.18821333759 *T)
    sumL3  +=       0.003 * cos(4.34017676059  + 14765.23904326980 *T)
    sumL3  +=       0.003 * cos(2.86652251194  + 105460.99111839020 *T)
    sumL3  +=       0.003 * cos(6.23875368963  + 28306.66024576099 *T)
    sumL3  +=       0.003 * cos(3.37815322354  + 98690.28051714458 *T)
    sumL3  +=       0.002 * cos(5.37466119262  + 110012.94461544900 *T)
    sumL3  +=       0.002 * cos(0.16040583556  + 50586.73338786459 *T)
    sumL3  +=       0.002 * cos(0.35038889636  + 19317.19254032860 *T)
    sumL3  +=       0.002 * cos(6.23271649650  + 49957.04917896160 *T)
    sumL3  +=       0.002 * cos(2.52219274566  + 181505.94343892601 *T)
    sumL3  +=       0.002 * cos(5.05358903232  + 162188.75089859738 *T)
    sumL3  +=       0.002 * cos(0.15071040297  + 124778.18365871880 *T)
    sumL3  +=       0.002 * cos(2.29024350361  + 56727.75978020720 *T)
    sumL3  +=       0.002 * cos(0.42411179784  + 71492.99882347698 *T)
    sumL3  +=       0.002 * cos(6.11262694468  + 57837.13833230060 *T)
    sumL3  +=       0.002 * cos(2.17198531948  + 108903.56606335558 *T)
    sumL3  +=       0.002 * cos(2.92723030315  + 54394.56338733519 *T)
    sumL3  +=       0.002 * cos(2.40411188609  + 83925.04147387479 *T)
    sumL3  +=       0.002 * cos(3.43613336205  + 78267.59076008058 *T)
    sumL3  +=       0.002 * cos(4.77873510695  + 365230.64398203877 *T)
    sumL3  +=       0.002 * cos(3.54043808451  + 64741.95798313119 *T)
    sumL3  +=       0.002 * cos(1.60502923885  + 77734.01845962799 *T)
    sumL3  +=       0.002 * cos(3.11800510932  + 150866.08680029298 *T)
    sumL3  +=       0.002 * cos(1.02098746090  + 78793.40038981718 *T)
    sumL3  +=       0.002 * cos(4.39439085682  + 88476.99497093359 *T)
    sumL3  +=       0.002 * cos(1.49912329404  + 43071.89928903080 *T)
    sumL3  +=       0.002 * cos(1.96277160647  + 25558.21217647960 *T)
    sumL3  +=       0.002 * cos(4.76774165204  + 77204.32749453338 *T)
    sumL3  +=       0.002 * cos(0.61066658116  + 155468.03691925600 *T)
    sumL3  +=       0.002 * cos(5.26875308161  + 77308.10968311399 *T)
    sumL3  +=       0.002 * cos(0.52736095020  + 94138.32702008578 *T)
    sumL3  +=       0.002 * cos(2.05793854229  + 103396.01282468818 *T)
    sumL3  +=       0.002 * cos(4.12698349577  + 66941.04532641819 *T)
    sumL3  +=       0.002 * cos(5.43336870762  + 78283.37918562238 *T)
    sumL3  +=       0.002 * cos(1.97520131105  + 136100.84775702321 *T)
    sumL3  +=       0.002 * cos(5.54343111585  + 81706.28436968799 *T)
    sumL3  +=       0.002 * cos(1.25444219651  + 40853.14218484400 *T)
    sumL3  +=       0.002 * cos(1.07684180123  + 103292.23063610760 *T)
    sumL3  +=       0.001 * cos(3.23979322492  + 52156.13652224860 *T)
    sumL3  +=       0.002 * cos(5.33834891475  + 52182.91983014920 *T)
    sumL3  +=       0.002 * cos(3.36867866204  + 99799.65906923798 *T)
    sumL3  +=       0.002 * cos(5.40697906670  + 82815.66292178139 *T)
    sumL3  +=       0.001 * cos(5.80853057947  + 4551.95349705880 *T)
    sumL3  +=       0.001 * cos(4.95909385120  + 129483.91596626240 *T)
    sumL3  +=       0.001 * cos(3.17481245588  + 76674.63652943878 *T)
    sumL3  +=       0.001 * cos(1.53803169600  + 52389.10537858640 *T)
    sumL3  +=       0.001 * cos(2.48052961456  + 53764.87917843220 *T)
    sumL3  +=       0.001 * cos(1.20843095195  + 105410.99449648600 *T)
    sumL3  +=       0.002 * cos(3.64576833526  + 45405.09568190280 *T)
    sumL3  +=       0.002 * cos(4.00710217385  + 104881.30353139140 *T)
    sumL3  +=       0.001 * cos(3.09532094520  + 76044.95232053580 *T)
    sumL3  +=       0.001 * cos(5.52119023864  + 58946.51688439399 *T)
    sumL3  +=       0.002 * cos(5.53036712466  + 131548.89425996438 *T)
    sumL3  +=       0.001 * cos(0.38352295948  + 52179.68761850640 *T)
    sumL3  +=       0.001 * cos(0.08902944955  + 51116.42435295920 *T)
    sumL3  +=       0.001 * cos(1.47514166670  + 62389.09182935939 *T)
    sumL3  +=       0.001 * cos(4.20864838814  + 131498.89763806018 *T)
    sumL3  +=       0.001 * cos(2.88161012078  + 79219.30916633119 *T)
    sumL3  +=       0.001 * cos(5.03155341199  + 51646.11531805379 *T)
    sumL3  +=       0.001 * cos(0.05873279909  + 53131.40602475700 *T)
    sumL3  +=       0.001 * cos(0.60647620600  + 38654.05484155699 *T)
    sumL3  +=       0.001 * cos(1.58007357234  + 391318.54712361295 *T)
    sumL3  +=       0.001 * cos(0.68687557696  + 130969.20667296558 *T)
    sumL3  +=       0.001 * cos(3.90367213586  + 129380.13377768178 *T)
    sumL3  +=       0.001 * cos(2.11138180302  + 32858.61374281979 *T)
    sumL3  +=       0.001 * cos(0.18542211423  + 78244.03966382280 *T)
    sumL3  +=       0.001 * cos(1.24063083423  + 9103.90699411760 *T)
    sumL3  +=       0.001 * cos(1.78547815249  + 188276.65404017159 *T)
    sumL3  +=       0.001 * cos(0.94125502564  + 93028.94846799241 *T)
    sumL3  +=       0.001 * cos(2.41309932759  + 55618.38122811380 *T)
    sumL3  +=       0.001 * cos(2.29683748973  + 52195.47604404819 *T)
    sumL3  +=       0.001 * cos(3.39284109553  + 25028.52121138500 *T)
    sumL3  +=       0.001 * cos(4.29717383023  + 26301.20223701220 *T)
    sumL3  +=       0.001 * cos(6.10639166059  + 102132.85546211000 *T)
    sumL3  +=       0.001 * cos(4.58189329524  + 37410.56723987860 *T)
    sumL3  +=       0.001 * cos(6.16145880062  + 125887.56221081220 *T)
    sumL3  +=       0.001 * cos(5.22010441376  + 52171.92494779040 *T)
    sumL3  +=       0.001 * cos(5.95645110567  + 80482.46652890938 *T)
    sumL3  +=       0.001 * cos(4.81981872221  + 26107.57290247399 *T)
    sumL3  +=       0.001 * cos(4.27546069271  + 36301.18868778519 *T)
    sumL3  +=       0.001 * cos(2.43374991272  + 51535.90899683439 *T)
    sumL3  +=       0.001 * cos(4.03088051207  + 1109.37855209340 *T)
    sumL3  +=       0.001 * cos(5.60495524983  + 207593.84658050016 *T)
    sumL3  +=       0.001 * cos(4.89918476796  + 35191.81013569180 *T)
    sumL3  +=       0.001 * cos(6.07747011982  + 27676.97603685800 *T)
    sumL3  +=       0.001 * cos(5.48361013686  + 29530.47808653960 *T)
    sumL3  +=       0.001 * cos(0.15288602386  + 25448.00585526019 *T)
    sumL3  +=       0.001 * cos(2.18342825402  + 63498.47038145279 *T)
    sumL3  +=       0.001 * cos(3.67054321453  + 21535.94964451540 *T)
    sumL3  +=       0.001 * cos(1.49815422727  + 26084.02180621620 *T)
    sumL3  +=       0.001 * cos(1.75551273491  + 26095.01668857500 *T)
    sumL3  +=       0.001 * cos(2.60061927559  + 12566.15169998280 *T)
    sumL3  +=       0.001 * cos(2.69537843932  + 52290.24557183361 *T)
    sumL3  +=       0.001 * cos(4.87760109673  + 25874.60404613620 *T)
    sumL3  +=       0.001 * cos(2.82170820105  + 26091.78447693220 *T)
    sumL3  +=       0.001 * cos(0.20736391933  + 52061.36699446317 *T)
    sumL3  +=       0.001 * cos(3.82330817663  + 68050.42387851159 *T)
    sumL3  +=       0.001 * cos(4.42501613287  + 22645.32819660879 *T)
    sumL3  +=       0.001 * cos(6.01128227925  + 52168.69273614759 *T)
    sumL3  +=       0.001 * cos(0.28595674172  + 11322.66409830440 *T)
    sumL3  +=       0.001 * cos(5.86280830053  + 26202.34243025941 *T)
    sumL3  +=       0.001 * cos(3.42109354799  + 25973.46385288896 *T)
    sumL3  +=       0.001 * cos(0.30050065300  + 41962.52073693740 *T)
    sumL3  +=       0.001 * cos(2.73414310032  + 25035.63475838580 *T)
    sumL3  +=       0.001 * cos(1.72346523407  + 51962.50718771040 *T)
    sumL3  +=       0.001 * cos(4.77297529316  + 78477.00852016058 *T)
    sumL3  +=       0.001 * cos(5.84491832575  + 50057.04242277000 *T)
    sumL3  +=       0.001 * cos(3.22072290870  + 25021.40766438420 *T)
    sumL3  +=       0.001 * cos(4.74388045372  + 26514.50133245020 *T)
    sumL3  +=       0.001 * cos(4.31760630574  + 69159.80243060499 *T)
    sumL3  +=       0.001 * cos(0.65529515413  + 426.59819087600 *T)
    sumL3  +=       0.001 * cos(5.58694327783  + 79852.78232000639 *T)
    sumL3  +=       0.001 * cos(6.07111696987  + 102762.53967101300 *T)
    sumL3  +=       0.001 * cos(5.84438975852  + 105307.21230790540 *T)
    sumL3  +=       0.001 * cos(2.82906439989  + 24498.83024629040 *T)

    # MERCURY Longitude
    sumL4  = 0
    sumL4  +=     114.078 * cos(3.14159265359  + 0.00000000000 *T)
    sumL4  +=       3.247 * cos(2.02848007619  + 26087.90314157420 *T)
    sumL4  +=       1.914 * cos(1.41731803758  + 78263.70942472259 *T)
    sumL4  +=       1.727 * cos(4.50137643801  + 52175.80628314840 *T)
    sumL4  +=       1.237 * cos(4.49970181057  + 104351.61256629678 *T)
    sumL4  +=       0.645 * cos(1.26591776986  + 130439.51570787100 *T)
    sumL4  +=       0.298 * cos(4.30600984981  + 156527.41884944518 *T)
    sumL4  +=       0.128 * cos(1.05702505039  + 182615.32199101936 *T)
    sumL4  +=       0.051 * cos(4.08566191934  + 208703.22513259359 *T)
    sumL4  +=       0.020 * cos(0.82130394681  + 234791.12827416777 *T)
    sumL4  +=       0.008 * cos(3.82199615635  + 260879.03141574195 *T)
    sumL4  +=       0.003 * cos(0.54824060574  + 286966.93455731618 *T)
    sumL4  +=       0.002 * cos(3.52134138964  + 313054.83769889042 *T)
    sumL4  +=       0.001 * cos(1.67435118540  + 24978.52458948080 *T)
    sumL4  +=       0.002 * cos(4.60718203612  + 51066.42773105500 *T)
    sumL4  +=       0.001 * cos(1.11396201203  + 129330.13715577760 *T)
    sumL4  +=       0.001 * cos(4.35040700322  + 103242.23401420339 *T)
    sumL4  +=       0.001 * cos(4.17434865844  + 79373.08797681599 *T)
    sumL4  +=       0.001 * cos(1.20388630609  + 77154.33087262919 *T)
    sumL4  +=       0.001 * cos(4.84677084740  + 27197.28169366760 *T)
    sumL4  +=       0.001 * cos(0.29913093643  + 339142.74084046454 *T)

    # MERCURY Longitude
    sumL5  = 0
    sumL5  +=       0.877 * cos(3.14159265359  + 0.00000000000 *T)
    sumL5  +=       0.059 * cos(3.37513289692  + 52175.80628314840 *T)
    sumL5  +=       0.042 * cos(0.02433099382  + 78263.70942472259 *T)
    sumL5  +=       0.043 * cos(4.65071406046  + 26087.90314157420 *T)
    sumL5  +=       0.031 * cos(2.98112204944  + 104351.61256629678 *T)
    sumL5  +=       0.019 * cos(6.00992467582  + 130439.51570787100 *T)
    sumL5  +=       0.010 * cos(2.75699791920  + 156527.41884944518 *T)
    sumL5  +=       0.005 * cos(5.79082359724  + 182615.32199101936 *T)
    sumL5  +=       0.002 * cos(2.51435448230  + 208703.22513259359 *T)
    sumL5  +=       0.001 * cos(5.49411470932  + 234791.12827416777 *T)
    sumL5  +=       0.001 * cos(2.23688504668  + 260879.03141574195 *T)
    sumL5  +=       0.001 * cos(5.18718589133  + 286966.93455731618 *T)

    # MERCURY Latitude
    sumB0  = 0
    sumB0  +=  11737528.962 * cos(1.98357498767  + 26087.90314157420 *T)
    sumB0  +=  2388076.996 * cos(5.03738959685  + 52175.80628314840 *T)
    sumB0  +=  1222839.532 * cos(3.14159265359  + 0.00000000000 *T)
    sumB0  +=  543251.810 * cos(1.79644363963  + 78263.70942472259 *T)
    sumB0  +=  129778.770 * cos(4.83232503961  + 104351.61256629678 *T)
    sumB0  +=   31866.927 * cos(1.58088495667  + 130439.51570787100 *T)
    sumB0  +=    7963.301 * cos(4.60972126348  + 156527.41884944518 *T)
    sumB0  +=    2014.189 * cos(1.35324164694  + 182615.32199101936 *T)
    sumB0  +=     513.953 * cos(4.37835409309  + 208703.22513259359 *T)
    sumB0  +=     207.674 * cos(4.91772564073  + 27197.28169366760 *T)
    sumB0  +=     208.584 * cos(2.02020294153  + 24978.52458948080 *T)
    sumB0  +=     132.013 * cos(1.11908492283  + 234791.12827416777 *T)
    sumB0  +=     100.454 * cos(5.65684734206  + 20426.57109242200 *T)
    sumB0  +=     121.395 * cos(1.81271752059  + 53285.18483524180 *T)
    sumB0  +=      91.566 * cos(2.28163128692  + 25028.52121138500 *T)
    sumB0  +=      99.214 * cos(0.09391887097  + 51116.42435295920 *T)
    sumB0  +=      94.574 * cos(1.24184909234  + 31749.23519072640 *T)
    sumB0  +=      78.785 * cos(4.40725880000  + 57837.13833230060 *T)
    sumB0  +=      77.747 * cos(0.52557061749  + 1059.38193018920 *T)
    sumB0  +=      84.264 * cos(5.08510388314  + 51066.42773105500 *T)
    sumB0  +=      49.948 * cos(3.49752993688  + 5661.33204915220 *T)
    sumB0  +=      46.454 * cos(3.23739270829  + 77204.32749453338 *T)
    sumB0  +=      44.767 * cos(4.87849816734  + 79373.08797681599 *T)
    sumB0  +=      40.766 * cos(2.46558332165  + 46514.47423399620 *T)
    sumB0  +=      37.378 * cos(4.45768797944  + 4551.95349705880 *T)
    sumB0  +=      34.082 * cos(4.14209210575  + 260879.03141574195 *T)
    sumB0  +=      35.911 * cos(1.09057317869  + 1109.37855209340 *T)
    sumB0  +=      31.953 * cos(1.18516389747  + 83925.04147387479 *T)
    sumB0  +=      30.954 * cos(3.50328027210  + 21535.94964451540 *T)
    sumB0  +=      31.808 * cos(2.41474588439  + 47623.85278608960 *T)
    sumB0  +=      28.691 * cos(1.84828614269  + 77154.33087262919 *T)
    sumB0  +=      25.765 * cos(2.77593370583  + 27043.50288318280 *T)
    sumB0  +=      25.199 * cos(3.59062266460  + 27147.28507176339 *T)
    sumB0  +=      20.244 * cos(3.06833797229  + 51646.11531805379 *T)
    sumB0  +=      18.591 * cos(5.58427274440  + 73711.75592766379 *T)
    sumB0  +=      16.971 * cos(0.02791276551  + 103292.23063610760 *T)
    sumB0  +=      20.099 * cos(4.06593040301  + 25132.30339996560 *T)
    sumB0  +=      17.002 * cos(6.13739392193  + 41962.52073693740 *T)
    sumB0  +=      14.984 * cos(1.64717994813  + 105460.99111839020 *T)
    sumB0  +=      14.186 * cos(0.33074185469  + 10213.28554621100 *T)
    sumB0  +=      15.577 * cos(6.07693643204  + 53131.40602475700 *T)
    sumB0  +=      15.795 * cos(3.79629547258  + 529.69096509460 *T)
    sumB0  +=      14.011 * cos(5.52786452723  + 72602.37737557039 *T)
    sumB0  +=      12.309 * cos(3.16626298867  + 14765.23904326980 *T)
    sumB0  +=      11.261 * cos(0.11326534696  + 13521.75144159140 *T)
    sumB0  +=      12.448 * cos(4.05109331029  + 39609.65458316560 *T)
    sumB0  +=      13.044 * cos(3.48016433624  + 37410.56723987860 *T)
    sumB0  +=      11.042 * cos(4.23192662377  + 110012.94461544900 *T)
    sumB0  +=      11.152 * cos(0.55658461820  + 63498.47038145279 *T)
    sumB0  +=      10.717 * cos(1.53686240986  + 25661.30495069820 *T)
    sumB0  +=      10.213 * cos(2.87881017166  + 12566.15169998280 *T)
    sumB0  +=       9.924 * cos(0.94824584866  + 65697.55772473979 *T)
    sumB0  +=      11.047 * cos(5.79741510309  + 51749.20809227239 *T)
    sumB0  +=      10.460 * cos(5.82962163777  + 50586.73338786459 *T)
    sumB0  +=       9.774 * cos(1.66271762650  + 24498.83024629040 *T)
    sumB0  +=       9.412 * cos(1.82364886045  + 15874.61759536320 *T)
    sumB0  +=      12.866 * cos(4.81650804018  + 30639.85663863300 *T)
    sumB0  +=       9.123 * cos(4.88604394110  + 103242.23401420339 *T)
    sumB0  +=       9.011 * cos(1.04262074744  + 426.59819087600 *T)
    sumB0  +=       8.735 * cos(3.04132552652  + 68050.42387851159 *T)
    sumB0  +=       8.491 * cos(1.05130498445  + 1589.07289528380 *T)
    sumB0  +=       8.835 * cos(0.88128343813  + 286966.93455731618 *T)
    sumB0  +=       8.823 * cos(5.81343662067  + 11322.66409830440 *T)
    sumB0  +=       8.196 * cos(0.84015129448  + 51220.20654153979 *T)
    sumB0  +=       7.640 * cos(3.43584035231  + 36301.18868778519 *T)
    sumB0  +=       9.175 * cos(6.16059719071  + 77734.01845962799 *T)
    sumB0  +=       8.897 * cos(0.39691733779  + 53235.18821333759 *T)
    sumB0  +=       7.253 * cos(2.37179970828  + 99799.65906923798 *T)
    sumB0  +=       7.104 * cos(5.73557153523  + 26617.59410666880 *T)
    sumB0  +=       8.322 * cos(5.27157128800  + 25558.21217647960 *T)
    sumB0  +=       6.578 * cos(2.66741925057  + 52705.49724824299 *T)
    sumB0  +=       5.825 * cos(6.25072627414  + 33326.57873317420 *T)
    sumB0  +=       5.625 * cos(3.08030000230  + 129380.13377768178 *T)
    sumB0  +=       5.295 * cos(1.35863387771  + 45892.73043315699 *T)
    sumB0  +=       5.116 * cos(2.67576280851  + 76674.63652943878 *T)
    sumB0  +=       5.035 * cos(2.66614676257  + 77837.11123384659 *T)
    sumB0  +=       4.762 * cos(4.68983196238  + 131548.89425996438 *T)
    sumB0  +=       5.714 * cos(2.87339047644  + 79219.30916633119 *T)
    sumB0  +=       5.601 * cos(5.37190885150  + 955.59974160860 *T)
    sumB0  +=       4.965 * cos(3.47619946892  + 59414.48187474840 *T)
    sumB0  +=       5.468 * cos(1.43781223143  + 56727.75978020720 *T)
    sumB0  +=       4.080 * cos(4.01667499000  + 91785.46086631398 *T)
    sumB0  +=       3.971 * cos(5.44452771841  + 6283.07584999140 *T)
    sumB0  +=       4.063 * cos(5.64291749144  + 26068.23338067440 *T)
    sumB0  +=       3.877 * cos(3.60887633874  + 89586.37352302698 *T)
    sumB0  +=       4.047 * cos(5.95235957644  + 38654.05484155699 *T)
    sumB0  +=       4.064 * cos(1.46283079734  + 26107.57290247399 *T)
    sumB0  +=       3.681 * cos(0.47365215122  + 62389.09182935939 *T)
    sumB0  +=       4.247 * cos(6.23163402945  + 40853.14218484400 *T)
    sumB0  +=       3.694 * cos(2.75543227024  + 7238.67559160000 *T)
    sumB0  +=       3.348 * cos(5.91868911857  + 43071.89928903080 *T)
    sumB0  +=       4.475 * cos(2.28937952702  + 98690.28051714458 *T)
    sumB0  +=       3.115 * cos(0.74943919509  + 32858.61374281979 *T)
    sumB0  +=       3.100 * cos(3.35203735852  + 19804.82729158280 *T)
    sumB0  +=       3.215 * cos(6.11374633777  + 94138.32702008578 *T)
    sumB0  +=       3.558 * cos(0.98820289555  + 136100.84775702321 *T)
    sumB0  +=       2.908 * cos(4.06284172838  + 26514.50133245020 *T)
    sumB0  +=       2.799 * cos(1.63602212056  + 129330.13715577760 *T)
    sumB0  +=       2.800 * cos(3.88358619611  + 77308.10968311399 *T)
    sumB0  +=       2.680 * cos(4.48235268618  + 71980.63357473118 *T)
    sumB0  +=       2.767 * cos(4.12700796923  + 27676.97603685800 *T)
    sumB0  +=       3.285 * cos(0.39398081520  + 18849.22754997420 *T)
    sumB0  +=       3.290 * cos(2.93487892542  + 103821.92160120218 *T)
    sumB0  +=       3.048 * cos(0.67698362024  + 9103.90699411760 *T)
    sumB0  +=       2.484 * cos(5.42332204960  + 125887.56221081220 *T)
    sumB0  +=       2.663 * cos(1.70280488847  + 16983.99614745660 *T)
    sumB0  +=       2.791 * cos(2.76839978418  + 26091.78447693220 *T)
    sumB0  +=       2.747 * cos(5.75135788288  + 19317.19254032860 *T)
    sumB0  +=       2.676 * cos(5.73635682216  + 78793.40038981718 *T)
    sumB0  +=       2.710 * cos(4.31274902720  + 58946.51688439399 *T)
    sumB0  +=       2.877 * cos(3.45410696457  + 79323.09135491178 *T)
    sumB0  +=       2.787 * cos(4.34025817469  + 26084.02180621620 *T)
    sumB0  +=       2.097 * cos(2.09090732812  + 23869.14603738740 *T)
    sumB0  +=       2.297 * cos(3.90321251530  + 313054.83769889042 *T)
    sumB0  +=       2.535 * cos(3.05170623483  + 69159.80243060499 *T)
    sumB0  +=       2.211 * cos(4.65120210900  + 28306.66024576099 *T)
    sumB0  +=       1.912 * cos(5.78403850812  + 102762.53967101300 *T)
    sumB0  +=       2.265 * cos(3.34478937379  + 22645.32819660879 *T)
    sumB0  +=       1.817 * cos(5.74251490625  + 103925.01437542078 *T)
    sumB0  +=       1.961 * cos(5.94577427093  + 105307.21230790540 *T)
    sumB0  +=       1.922 * cos(4.43763124338  + 82815.66292178139 *T)
    sumB0  +=       1.667 * cos(4.51887419928  + 52195.47604404819 *T)
    sumB0  +=       1.771 * cos(6.12278757535  + 155468.03691925600 *T)
    sumB0  +=       1.515 * cos(1.72458115830  + 54394.56338733519 *T)
    sumB0  +=       1.585 * cos(0.23900687506  + 85502.38501632259 *T)
    sumB0  +=       1.477 * cos(4.73359545019  + 29530.47808653960 *T)
    sumB0  +=       1.469 * cos(1.44574704924  + 157636.79740153858 *T)
    sumB0  +=       1.653 * cos(2.41345886386  + 52156.13652224860 *T)
    sumB0  +=       1.377 * cos(0.59981573116  + 26080.78959457339 *T)
    sumB0  +=       1.430 * cos(0.78416943990  + 117873.36400788820 *T)
    sumB0  +=       1.680 * cos(3.42684715134  + 6770.71060124560 *T)
    sumB0  +=       1.604 * cos(4.52048871587  + 3442.57494496540 *T)
    sumB0  +=       1.345 * cos(0.31857167988  + 51535.90899683439 *T)
    sumB0  +=       1.357 * cos(3.17349203273  + 25874.60404613620 *T)
    sumB0  +=       1.250 * cos(2.33380858079  + 25448.00585526019 *T)
    sumB0  +=       1.196 * cos(2.09277609512  + 48733.23133818299 *T)
    sumB0  +=       1.194 * cos(0.36376300747  + 115674.27666460120 *T)
    sumB0  +=       1.443 * cos(2.73186009497  + 64741.95798313119 *T)
    sumB0  +=       1.375 * cos(5.32672767568  + 124778.18365871880 *T)
    sumB0  +=       1.325 * cos(3.57805811914  + 88476.99497093359 *T)
    sumB0  +=       1.115 * cos(0.26092087309  + 639.89728631400 *T)
    sumB0  +=       1.094 * cos(2.22755234371  + 32370.97899156560 *T)
    sumB0  +=       1.433 * cos(3.03965981318  + 66941.04532641819 *T)
    sumB0  +=       1.370 * cos(3.45924813827  + 44937.13069154840 *T)
    sumB0  +=       1.003 * cos(0.28769164324  + 25934.12433108940 *T)
    sumB0  +=       1.136 * cos(1.11057140999  + 52171.92494779040 *T)
    sumB0  +=       1.000 * cos(0.52457514920  + 26241.68195205900 *T)
    sumB0  +=       1.033 * cos(2.55563840213  + 45405.09568190280 *T)
    sumB0  +=       1.096 * cos(5.01257889902  + 7.11354700080 *T)
    sumB0  +=       1.146 * cos(5.40631829059  + 213.29909543800 *T)
    sumB0  +=       1.080 * cos(5.97835772937  + 129909.82474277640 *T)
    sumB0  +=       0.988 * cos(1.25815251981  + 98068.53671630540 *T)
    sumB0  +=       1.062 * cos(2.88132742088  + 120226.23016166000 *T)
    sumB0  +=       0.892 * cos(0.63710475446  + 103396.01282468818 *T)
    sumB0  +=       1.041 * cos(0.89365572198  + 52602.40447402440 *T)
    sumB0  +=       0.978 * cos(0.93067667928  + 53764.87917843220 *T)
    sumB0  +=       1.136 * cos(5.82023159264  + 52179.68761850640 *T)
    sumB0  +=       0.838 * cos(4.66556603681  + 155418.04029735178 *T)
    sumB0  +=       0.910 * cos(5.53568417425  + 58458.88213313979 *T)
    sumB0  +=       0.834 * cos(5.16159172473  + 49957.04917896160 *T)
    sumB0  +=       1.100 * cos(4.02157761190  + 162188.75089859738 *T)
    sumB0  +=       0.805 * cos(1.03425281191  + 23969.13928119580 *T)
    sumB0  +=       0.835 * cos(5.28205659749  + 50057.04242277000 *T)
    sumB0  +=       0.891 * cos(0.21542628779  + 105410.99449648600 *T)
    sumB0  +=       0.743 * cos(0.86945390288  + 52389.10537858640 *T)
    sumB0  +=       0.783 * cos(3.32304812554  + 26301.20223701220 *T)
    sumB0  +=       0.711 * cos(1.57780048955  + 2118.76386037840 *T)
    sumB0  +=       0.801 * cos(2.18220012579  + 151975.46535238638 *T)
    sumB0  +=       0.915 * cos(1.71247948367  + 51962.50718771040 *T)
    sumB0  +=       0.832 * cos(5.53506450006  + 74821.13447975718 *T)
    sumB0  +=       0.750 * cos(2.16183196338  + 52168.69273614759 *T)
    sumB0  +=       0.923 * cos(2.50170936928  + 104881.30353139140 *T)
    sumB0  +=       0.727 * cos(6.05394635764  + 95247.70557217918 *T)
    sumB0  +=       0.640 * cos(2.71738158564  + 131395.11544947961 *T)
    sumB0  +=       0.628 * cos(1.43868691098  + 55618.38122811380 *T)
    sumB0  +=       0.622 * cos(3.46102298068  + 77623.81213840858 *T)
    sumB0  +=       0.825 * cos(1.00561540075  + 85034.42002596818 *T)
    sumB0  +=       0.611 * cos(0.99677911598  + 2218.75710418680 *T)
    sumB0  +=       0.605 * cos(4.68164610499  + 52182.91983014920 *T)
    sumB0  +=       0.646 * cos(2.57803910757  + 128850.44281258718 *T)
    sumB0  +=       0.621 * cos(1.16576338077  + 108903.56606335558 *T)
    sumB0  +=       0.651 * cos(1.60040223342  + 26095.01668857500 *T)
    sumB0  +=       0.599 * cos(2.51123620954  + 130012.91751699500 *T)
    sumB0  +=       0.570 * cos(1.27714994259  + 78283.37918562238 *T)
    sumB0  +=       0.616 * cos(2.88990323867  + 52026.24308601380 *T)
    sumB0  +=       0.553 * cos(5.17654422678  + 25938.33994443960 *T)
    sumB0  +=       0.560 * cos(3.32834155275  + 51109.31080595839 *T)
    sumB0  +=       0.538 * cos(4.74592892110  + 80482.46652890938 *T)
    sumB0  +=       0.514 * cos(5.52495783925  + 25021.40766438420 *T)
    sumB0  +=       0.512 * cos(2.12021253064  + 25035.63475838580 *T)
    sumB0  +=       0.545 * cos(1.31029342517  + 26555.86813192860 *T)
    sumB0  +=       0.475 * cos(4.05390077061  + 149.56319713460 *T)
    sumB0  +=       0.464 * cos(3.26222495894  + 111590.28815789679 *T)
    sumB0  +=       0.460 * cos(5.34048307893  + 46848.33017476560 *T)
    sumB0  +=       0.550 * cos(5.84636069235  + 25619.93815121980 *T)
    sumB0  +=       0.447 * cos(3.57688571640  + 1066.49547719000 *T)
    sumB0  +=       0.550 * cos(6.22077946049  + 51123.53789995999 *T)
    sumB0  +=       0.596 * cos(0.64492352990  + 339142.74084046454 *T)
    sumB0  +=       0.484 * cos(3.16177806139  + 38519.94579197200 *T)
    sumB0  +=       0.441 * cos(0.67395966884  + 1052.26838318840 *T)
    sumB0  +=       0.477 * cos(0.22606306742  + 71025.03383312259 *T)
    sumB0  +=       0.423 * cos(2.08215220948  + 76144.94556434419 *T)
    sumB0  +=       0.472 * cos(3.35508518254  + 35191.81013569180 *T)
    sumB0  +=       0.418 * cos(3.57908956341  + 52329.58509363319 *T)
    sumB0  +=       0.441 * cos(4.91196768491  + 78050.41032928458 *T)
    sumB0  +=       0.440 * cos(4.47866581814  + 183724.70054311279 *T)
    sumB0  +=       0.435 * cos(1.98197259533  + 26094.53170047421 *T)
    sumB0  +=       0.564 * cos(5.45416215614  + 78244.03966382280 *T)
    sumB0  +=       0.540 * cos(2.87570725847  + 181555.94006083015 *T)
    sumB0  +=       0.450 * cos(6.09063658976  + 93028.94846799241 *T)
    sumB0  +=       0.461 * cos(3.82361162048  + 143961.26714946239 *T)
    sumB0  +=       0.408 * cos(3.34314470563  + 52022.02747266360 *T)
    sumB0  +=       0.412 * cos(2.07516355558  + 150866.08680029298 *T)
    sumB0  +=       0.435 * cos(5.12507754577  + 26081.27458267419 *T)
    sumB0  +=       0.409 * cos(0.56822912556  + 64607.84893354619 *T)
    sumB0  +=       0.388 * cos(4.16039183226  + 78259.82808936459 *T)
    sumB0  +=       0.393 * cos(3.23715146834  + 13655.86049117640 *T)
    sumB0  +=       0.375 * cos(5.33297275541  + 26011.63707029860 *T)
    sumB0  +=       0.467 * cos(5.77388878977  + 90829.86112470538 *T)
    sumB0  +=       0.350 * cos(5.62301652558  + 71492.99882347698 *T)
    sumB0  +=       0.352 * cos(3.27694758804  + 26727.80042788820 *T)
    sumB0  +=       0.350 * cos(3.39302729957  + 141762.17980617538 *T)
    sumB0  +=       0.390 * cos(2.58800166779  + 78267.59076008058 *T)
    sumB0  +=       0.365 * cos(3.19925143022  + 52099.54021187280 *T)
    sumB0  +=       0.433 * cos(1.55185163390  + 20760.42703319140 *T)
    sumB0  +=       0.327 * cos(5.69228054184  + 12432.04265039780 *T)
    sumB0  +=       0.339 * cos(1.58088572820  + 45494.58142974879 *T)
    sumB0  +=       0.338 * cos(2.73161573173  + 155997.72788435058 *T)
    sumB0  +=       0.315 * cos(4.00061444247  + 78477.00852016058 *T)
    sumB0  +=       0.435 * cos(0.36408526333  + 114564.89811250778 *T)
    sumB0  +=       0.344 * cos(5.02616143626  + 78256.59587772179 *T)
    sumB0  +=       0.346 * cos(4.14121333484  + 111122.32316754240 *T)
    sumB0  +=       0.286 * cos(5.40513646014  + 6681.22485339960 *T)
    sumB0  +=       0.283 * cos(3.65516534367  + 76.26607127560 *T)
    sumB0  +=       0.325 * cos(4.30412080460  + 124156.43985787958 *T)
    sumB0  +=       0.364 * cos(2.32771554672  + 84546.78527471398 *T)
    sumB0  +=       0.331 * cos(0.76882559900  + 188276.65404017159 *T)
    sumB0  +=       0.267 * cos(3.81778097058  + 19406.67828817460 *T)
    sumB0  +=       0.270 * cos(6.02010732963  + 78114.14622758799 *T)
    sumB0  +=       0.273 * cos(3.66726180240  + 129483.91596626240 *T)
    sumB0  +=       0.263 * cos(0.18743596408  + 77197.21394753258 *T)
    sumB0  +=       0.268 * cos(3.25051884036  + 131498.89763806018 *T)
    sumB0  +=       0.284 * cos(1.93848296863  + 76044.95232053580 *T)
    sumB0  +=       0.295 * cos(5.53951004419  + 130969.20667296558 *T)
    sumB0  +=       0.335 * cos(5.92048177497  + 146314.13330323418 *T)
    sumB0  +=       0.246 * cos(1.45343047775  + 78270.82297172339 *T)
    sumB0  +=       0.259 * cos(3.08241704780  + 77211.44104153418 *T)
    sumB0  +=       0.253 * cos(2.86398164134  + 121335.60871375341 *T)
    sumB0  +=       0.338 * cos(3.95823396010  + 78690.30761559859 *T)
    sumB0  +=       0.317 * cos(3.98138369220  + 79852.78232000639 *T)
    sumB0  +=       0.232 * cos(4.66776441683  + 28206.66700195260 *T)
    sumB0  +=       0.247 * cos(1.40414696612  + 181505.94343892601 *T)
    sumB0  +=       0.228 * cos(0.25677390582  + 103711.71527998280 *T)
    sumB0  +=       0.219 * cos(4.69799359762  + 71582.48457132299 *T)
    sumB0  +=       0.221 * cos(4.46394113791  + 81706.28436968799 *T)
    sumB0  +=       0.254 * cos(1.43735157085  + 5327.47610838280 *T)
    sumB0  +=       0.199 * cos(5.76674433704  + 157483.01859105378 *T)
    sumB0  +=       0.253 * cos(2.47481215159  + 72936.23331633979 *T)
    sumB0  +=       0.191 * cos(4.17749403686  + 134991.46920492980 *T)
    sumB0  +=       0.246 * cos(5.21552773120  + 178063.36849396059 *T)
    sumB0  +=       0.182 * cos(4.31488794636  + 104371.28232719658 *T)
    sumB0  +=       0.206 * cos(1.62004139707  + 106570.36967048360 *T)
    sumB0  +=       0.234 * cos(2.18979921440  + 100909.03762133140 *T)
    sumB0  +=       0.176 * cos(3.58325574282  + 48835.19385644859 *T)
    sumB0  +=       0.188 * cos(2.58568452037  + 51707.84129279399 *T)
    sumB0  +=       0.177 * cos(1.89461095130  + 52169.17772424839 *T)
    sumB0  +=       0.168 * cos(3.01264496377  + 23754.70674870219 *T)
    sumB0  +=       0.222 * cos(4.35573235377  + 52643.77127350280 *T)
    sumB0  +=       0.169 * cos(5.25194313253  + 102232.84870591838 *T)
    sumB0  +=       0.188 * cos(5.55333225508  + 156100.82065856917 *T)
    sumB0  +=       0.163 * cos(1.72109994467  + 104138.31347085880 *T)
    sumB0  +=       0.205 * cos(5.64507146979  + 154938.34595416137 *T)
    sumB0  +=       0.161 * cos(0.06514575292  + 78187.44335344699 *T)
    sumB0  +=       0.155 * cos(3.29540644622  + 97112.93697469680 *T)
    sumB0  +=       0.152 * cos(1.66425291176  + 25654.19140369739 *T)
    sumB0  +=       0.179 * cos(2.20248310353  + 104331.94280539700 *T)
    sumB0  +=       0.157 * cos(5.91615237188  + 51742.09454527159 *T)
    sumB0  +=       0.176 * cos(5.03409571975  + 52182.43484204840 *T)
    sumB0  +=       0.145 * cos(0.34814763869  + 78417.48823520739 *T)
    sumB0  +=       0.151 * cos(0.84016182050  + 26237.46633870879 *T)
    sumB0  +=       0.143 * cos(0.36102957131  + 27154.39861876420 *T)
    sumB0  +=       0.142 * cos(5.87490895177  + 26037.90651966999 *T)
    sumB0  +=       0.167 * cos(4.43520444137  + 25668.41849769900 *T)
    sumB0  +=       0.137 * cos(2.85089329350  + 119116.85160956658 *T)
    sumB0  +=       0.161 * cos(5.91020381915  + 207643.84320240439 *T)
    sumB0  +=       0.130 * cos(6.26801612444  + 137678.19129947099 *T)
    sumB0  +=       0.130 * cos(0.93376066695  + 433.71173787680 *T)
    sumB0  +=       0.140 * cos(1.22522976768  + 26137.89976347840 *T)
    sumB0  +=       0.144 * cos(4.56326656247  + 419.48464387520 *T)
    sumB0  +=       0.145 * cos(2.53141564902  + 116917.76426627958 *T)
    sumB0  +=       0.145 * cos(3.74606280935  + 27140.17152476259 *T)
    sumB0  +=       0.155 * cos(3.65976494728  + 365230.64398203877 *T)
    sumB0  +=       0.124 * cos(0.92716868075  + 104347.73123093881 *T)
    sumB0  +=       0.140 * cos(0.11287334170  + 78109.93061423779 *T)
    sumB0  +=       0.126 * cos(1.13823138141  + 51322.60990139639 *T)
    sumB0  +=       0.146 * cos(3.25675289634  + 3340.61242669980 *T)
    sumB0  +=       0.122 * cos(5.10611624534  + 176953.98994186718 *T)
    sumB0  +=       0.125 * cos(5.63666753628  + 104355.49390165480 *T)
    sumB0  +=       0.155 * cos(2.30861381490  + 51756.32163927320 *T)
    sumB0  +=       0.136 * cos(3.41914767255  + 140652.80125408198 *T)
    sumB0  +=       0.115 * cos(6.25299482333  + 39743.76363275060 *T)
    sumB0  +=       0.113 * cos(0.79353194356  + 7994.52844202420 *T)
    sumB0  +=       0.126 * cos(0.12477487369  + 52815.70356946240 *T)
    sumB0  +=       0.121 * cos(0.68099507149  + 50483.64061364600 *T)
    sumB0  +=       0.118 * cos(0.87294905519  + 137210.22630911658 *T)
    sumB0  +=       0.112 * cos(2.41390426449  + 97580.90196505121 *T)
    sumB0  +=       0.130 * cos(1.22267912068  + 209812.60368468691 *T)
    sumB0  +=       0.110 * cos(3.15687226733  + 25234.70675982219 *T)
    sumB0  +=       0.109 * cos(2.85465805569  + 24395.73747207180 *T)
    sumB0  +=       0.108 * cos(5.73587160710  + 853.19638175200 *T)
    sumB0  +=       0.141 * cos(4.76875897743  + 28421.09953444620 *T)
    sumB0  +=       0.143 * cos(0.57659923971  + 170049.17029103660 *T)
    sumB0  +=       0.127 * cos(5.39590556187  + 110634.68841628821 *T)
    sumB0  +=       0.110 * cos(0.79294568581  + 104564.91166173480 *T)
    sumB0  +=       0.103 * cos(5.77115680934  + 182085.63102592481 *T)
    sumB0  +=       0.108 * cos(0.78147749534  + 42153.96900304900 *T)
    sumB0  +=       0.101 * cos(6.24176236571  + 1692.16566950240 *T)
    sumB0  +=       0.125 * cos(1.72822604057  + 104344.49901929598 *T)
    sumB0  +=       0.112 * cos(1.53691309669  + 18093.37469954999 *T)
    sumB0  +=       0.100 * cos(6.03656277681  + 147423.51185532758 *T)
    sumB0  +=       0.102 * cos(2.16777222106  + 32769.12799497380 *T)
    sumB0  +=       0.127 * cos(5.69552946334  + 44181.27784112419 *T)
    sumB0  +=       0.096 * cos(2.81662615895  + 104202.04936916218 *T)
    sumB0  +=       0.094 * cos(6.16378269144  + 103299.34418310841 *T)
    sumB0  +=       0.095 * cos(3.08582674985  + 70269.18098269838 *T)
    sumB0  +=       0.096 * cos(3.26682732105  + 103285.11708910680 *T)
    sumB0  +=       0.101 * cos(1.49832461293  + 40565.25432477420 *T)
    sumB0  +=       0.099 * cos(0.11817727117  + 167850.08294774959 *T)
    sumB0  +=       0.099 * cos(3.80154992347  + 214364.55718174577 *T)
    sumB0  +=       0.093 * cos(3.27120904242  + 90695.75207512038 *T)
    sumB0  +=       0.104 * cos(5.35405285854  + 126996.94076290558 *T)
    sumB0  +=       0.105 * cos(0.73008875321  + 104778.21075717278 *T)
    sumB0  +=       0.089 * cos(2.82753100547  + 50579.61984086379 *T)
    sumB0  +=       0.098 * cos(0.55393078868  + 33967.99229491319 *T)
    sumB0  +=       0.087 * cos(1.47928882834  + 97670.38771289718 *T)
    sumB0  +=       0.117 * cos(0.17401840597  + 61279.71327726600 *T)
    sumB0  +=       0.089 * cos(0.41552735762  + 26164.16921284980 *T)
    sumB0  +=       0.091 * cos(2.28447351962  + 157057.10981453978 *T)
    sumB0  +=       0.103 * cos(2.67784306575  + 172402.03644480839 *T)
    sumB0  +=       0.084 * cos(4.31852113132  + 60055.89543648739 *T)
    sumB0  +=       0.085 * cos(4.50666554965  + 104358.72611329758 *T)
    sumB0  +=       0.098 * cos(0.74967522191  + 105940.68546158058 *T)
    sumB0  +=       0.082 * cos(2.80078395016  + 52101.02468458109 *T)
    sumB0  +=       0.086 * cos(0.38069361609  + 74923.09699802278 *T)
    sumB0  +=       0.077 * cos(4.91602955543  + 24491.71669928959 *T)
    sumB0  +=       0.104 * cos(6.08692240129  + 10021.83728009940 *T)
    sumB0  +=       0.094 * cos(2.21754829180  + 25973.46385288896 *T)
    sumB0  +=       0.079 * cos(4.63574888508  + 2333.19639287200 *T)
    sumB0  +=       0.076 * cos(0.06676607630  + 157586.80077963436 *T)
    sumB0  +=       0.076 * cos(4.09119863738  + 1596.18644228460 *T)
    sumB0  +=       0.089 * cos(0.51926466594  + 50593.84693486539 *T)
    sumB0  +=       0.078 * cos(5.85412082904  + 18207.81398823521 *T)
    sumB0  +=       0.083 * cos(1.48347794862  + 54294.57014352679 *T)
    sumB0  +=       0.102 * cos(1.11648927836  + 150244.34299945380 *T)
    sumB0  +=       0.081 * cos(0.43168464249  + 155571.81910783658 *T)
    sumB0  +=       0.071 * cos(2.77560894834  + 77829.99768684579 *T)
    sumB0  +=       0.098 * cos(3.04617186486  + 16066.06586147480 *T)
    sumB0  +=       0.089 * cos(3.91487126164  + 68241.87214462319 *T)
    sumB0  +=       0.076 * cos(3.32839538947  + 129799.61842155698 *T)
    sumB0  +=       0.091 * cos(4.99146967087  + 102132.85546211000 *T)
    sumB0  +=       0.088 * cos(4.54401955709  + 51543.02254383520 *T)
    sumB0  +=       0.092 * cos(4.88825714860  + 26202.34243025941 *T)
    sumB0  +=       0.066 * cos(1.97272013972  + 632.78373931320 *T)
    sumB0  +=       0.066 * cos(1.34921841483  + 52808.59002246159 *T)
    sumB0  +=       0.071 * cos(1.21551241620  + 107794.18751126219 *T)
    sumB0  +=       0.088 * cos(3.90671333341  + 74.78159856730 *T)
    sumB0  +=       0.076 * cos(1.11684262920  + 78731.67441507700 *T)
    sumB0  +=       0.060 * cos(4.94665488683  + 78257.08086582259 *T)
    sumB0  +=       0.067 * cos(4.24719146625  + 14477.35118320000 *T)
    sumB0  +=       0.069 * cos(5.46721744006  + 77844.22478084739 *T)
    sumB0  +=       0.059 * cos(2.80000354604  + 25863.55834587229 *T)
    sumB0  +=       0.060 * cos(4.23924716288  + 52225.80290505260 *T)
    sumB0  +=       0.059 * cos(5.90340687048  + 103.09277421860 *T)
    sumB0  +=       0.059 * cos(3.81127888964  + 76571.54375522019 *T)
    sumB0  +=       0.060 * cos(5.62072757879  + 77795.74443436819 *T)
    sumB0  +=       0.069 * cos(4.36673018646  + 207593.84658050016 *T)
    sumB0  +=       0.066 * cos(0.49890308433  + 51951.46148744649 *T)
    sumB0  +=       0.058 * cos(1.63810168022  + 51013.33157874059 *T)
    sumB0  +=       0.058 * cos(4.28286858594  + 77410.51304297059 *T)
    sumB0  +=       0.066 * cos(0.61313044187  + 51639.00177105299 *T)
    sumB0  +=       0.057 * cos(3.15657420172  + 104275.34649502118 *T)
    sumB0  +=       0.058 * cos(4.41262101447  + 26720.68688088739 *T)
    sumB0  +=       0.075 * cos(0.13973480485  + 1581.95934828300 *T)
    sumB0  +=       0.057 * cos(0.41985749035  + 23439.44831610119 *T)
    sumB0  +=       0.069 * cos(6.04001555481  + 29428.51556827400 *T)
    sumB0  +=       0.055 * cos(0.87548003173  + 161079.37234650398 *T)
    sumB0  +=       0.054 * cos(2.10504641371  + 2648.45482547300 *T)
    sumB0  +=       0.070 * cos(1.90064683191  + 204151.27163553479 *T)
    sumB0  +=       0.054 * cos(3.73359360369  + 24925.42843716640 *T)
    sumB0  +=       0.055 * cos(4.74493087072  + 49527.35145767539 *T)
    sumB0  +=       0.071 * cos(2.67007207774  + 24505.94379329119 *T)
    sumB0  +=       0.057 * cos(1.10422154948  + 130459.18546877075 *T)
    sumB0  +=       0.068 * cos(5.29164108644  + 26013.12154300690 *T)
    sumB0  +=       0.054 * cos(0.06047544079  + 52072.71350892979 *T)
    sumB0  +=       0.053 * cos(0.06419360550  + 536.80451209540 *T)
    sumB0  +=       0.053 * cos(5.00852687358  + 25977.69682035479 *T)
    sumB0  +=       0.052 * cos(4.17403886040  + 25131.61398560359 *T)
    sumB0  +=       0.052 * cos(3.31036935415  + 7880.08915333899 *T)
    sumB0  +=       0.070 * cos(6.10702300469  + 52278.89905736699 *T)
    sumB0  +=       0.059 * cos(1.97971323674  + 51219.51712717779 *T)
    sumB0  +=       0.052 * cos(0.17883872581  + 224.34479570190 *T)
    sumB0  +=       0.057 * cos(2.64616122722  + 52125.80966124419 *T)
    sumB0  +=       0.051 * cos(2.51127480556  + 183570.92173262799 *T)
    sumB0  +=       0.050 * cos(4.67814206741  + 130226.21661243298 *T)
    sumB0  +=       0.050 * cos(4.77326024770  + 25780.34552060460 *T)
    sumB0  +=       0.053 * cos(2.81579181250  + 55516.41870984820 *T)
    sumB0  +=       0.057 * cos(2.13355985608  + 128320.75184749260 *T)
    sumB0  +=       0.049 * cos(2.32162107938  + 26395.46076254379 *T)
    sumB0  +=       0.049 * cos(4.96049794937  + 956.28915597060 *T)
    sumB0  +=       0.047 * cos(4.20636709156  + 52250.58788171570 *T)
    sumB0  +=       0.058 * cos(2.23088591593  + 52065.59996192899 *T)
    sumB0  +=       0.048 * cos(5.31209962934  + 1162.47470440780 *T)
    sumB0  +=       0.062 * cos(6.11793881000  + 22747.29071487440 *T)
    sumB0  +=       0.051 * cos(5.16823572193  + 130419.84594697120 *T)
    sumB0  +=       0.058 * cos(1.50494628165  + 54509.00267602040 *T)
    sumB0  +=       0.060 * cos(0.43836944868  + 116783.65521669458 *T)
    sumB0  +=       0.059 * cos(1.50474668065  + 26507.38778544939 *T)
    sumB0  +=       0.047 * cos(3.28426374672  + 25565.32572348040 *T)
    sumB0  +=       0.056 * cos(2.45288584577  + 181026.24909573561 *T)
    sumB0  +=       0.051 * cos(3.46813929041  + 53242.30176033840 *T)
    sumB0  +=       0.045 * cos(0.64303440167  + 52698.38370124219 *T)
    sumB0  +=       0.049 * cos(0.43622928396  + 27999.10262479140 *T)
    sumB0  +=       0.046 * cos(0.93537805397  + 26162.68474014150 *T)
    sumB0  +=       0.043 * cos(5.95311422864  + 76667.52298243798 *T)
    sumB0  +=       0.060 * cos(1.79764830651  + 78270.33798362259 *T)
    sumB0  +=       0.056 * cos(2.27809233545  + 182188.72380014337 *T)
    sumB0  +=       0.043 * cos(3.95909386803  + 26521.61487945100 *T)
    sumB0  +=       0.043 * cos(2.06740816154  + 51859.41441349179 *T)
    sumB0  +=       0.051 * cos(5.31772214752  + 66653.15746634839 *T)
    sumB0  +=       0.044 * cos(1.68470111077  + 23888.81579828719 *T)
    sumB0  +=       0.054 * cos(3.94217571152  + 52325.36948028299 *T)
    sumB0  +=       0.043 * cos(5.13864955940  + 11610.55195837420 *T)
    sumB0  +=       0.042 * cos(4.61363107691  + 52492.19815280499 *T)
    sumB0  +=       0.046 * cos(3.39633739377  + 104505.39137678158 *T)
    sumB0  +=       0.046 * cos(4.83387098498  + 110.20632121940 *T)
    sumB0  +=       0.045 * cos(3.54092987543  + 50800.03248330259 *T)
    sumB0  +=       0.041 * cos(5.69167231584  + 24712.12934172840 *T)
    sumB0  +=       0.040 * cos(0.05181171325  + 123200.84011627098 *T)
    sumB0  +=       0.053 * cos(0.61042891791  + 53228.07466633679 *T)
    sumB0  +=       0.039 * cos(3.11194295827  + 25984.81036735560 *T)
    sumB0  +=       0.046 * cos(2.53528410360  + 50696.93970908399 *T)
    sumB0  +=       0.038 * cos(4.44293961717  + 29416.03879785439 *T)
    sumB0  +=       0.041 * cos(2.66264705448  + 233731.74634397856 *T)
    sumB0  +=       0.037 * cos(3.37388547918  + 1375.77379984580 *T)
    sumB0  +=       0.038 * cos(5.61891042604  + 143005.66740785379 *T)
    sumB0  +=       0.037 * cos(5.42828179354  + 28286.99048486120 *T)
    sumB0  +=       0.049 * cos(5.33653478407  + 99024.13645791401 *T)
    sumB0  +=       0.040 * cos(5.63484527365  + 522.57741809380 *T)
    sumB0  +=       0.039 * cos(3.40419666656  + 25551.09862947879 *T)
    sumB0  +=       0.036 * cos(5.94925007197  + 145204.75475114080 *T)
    sumB0  +=       0.038 * cos(2.07740095188  + 153084.84390447981 *T)
    sumB0  +=       0.039 * cos(3.20021878147  + 163766.09444104519 *T)
    sumB0  +=       0.037 * cos(5.22659089943  + 52595.29092702359 *T)
    sumB0  +=       0.042 * cos(4.90625897857  + 25771.51127191760 *T)
    sumB0  +=       0.041 * cos(2.99128318097  + 26190.99591579279 *T)
    sumB0  +=       0.044 * cos(1.11685231827  + 26404.29501123080 *T)
    sumB0  +=       0.045 * cos(4.61987657347  + 132658.27281205778 *T)
    sumB0  +=       0.037 * cos(0.21788276252  + 166740.70439565618 *T)
    sumB0  +=       0.034 * cos(1.94222980062  + 203041.89308344136 *T)
    sumB0  +=       0.045 * cos(3.16157322891  + 104197.83375581198 *T)
    sumB0  +=       0.046 * cos(4.46372826281  + 316.39186965660 *T)
    sumB0  +=       0.039 * cos(2.02681719442  + 136722.59155786238 *T)
    sumB0  +=       0.034 * cos(1.34191646454  + 467.96499035440 *T)
    sumB0  +=       0.043 * cos(3.34582658464  + 78903.60671103658 *T)
    sumB0  +=       0.039 * cos(3.65232351757  + 196137.07343261081 *T)
    sumB0  +=       0.037 * cos(0.37289111380  + 391318.54712361295 *T)
    sumB0  +=       0.037 * cos(2.88378512797  + 36109.74042167360 *T)
    sumB0  +=       0.037 * cos(1.66197775716  + 52290.24557183361 *T)
    sumB0  +=       0.036 * cos(3.91856090884  + 54087.00576636560 *T)
    sumB0  +=       0.034 * cos(0.07777074766  + 26294.08869001139 *T)
    sumB0  +=       0.033 * cos(2.57658065167  + 26941.09952332620 *T)
    sumB0  +=       0.034 * cos(1.66706631543  + 49842.60989027639 *T)
    sumB0  +=       0.042 * cos(3.65143192554  + 76681.75007643958 *T)
    sumB0  +=       0.032 * cos(4.02685079656  + 24356.78078864160 *T)
    sumB0  +=       0.041 * cos(3.48117117726  + 26610.48055966799 *T)
    sumB0  +=       0.040 * cos(1.30798263894  + 77630.92568540938 *T)
    sumB0  +=       0.033 * cos(3.31591322468  + 65831.66677432480 *T)
    sumB0  +=       0.042 * cos(4.68305225464  + 24609.03656750980 *T)
    sumB0  +=       0.034 * cos(2.43308912753  + 26421.75908234360 *T)
    sumB0  +=       0.031 * cos(0.55207634036  + 13362.44970679920 *T)
    sumB0  +=       0.031 * cos(3.97168176907  + 130435.63437251298 *T)
    sumB0  +=       0.035 * cos(5.90195011358  + 78188.92782615528 *T)
    sumB0  +=       0.032 * cos(5.58275762493  + 25881.71759313700 *T)
    sumB0  +=       0.032 * cos(3.03394602316  + 27780.06881107659 *T)
    sumB0  +=       0.038 * cos(4.61602478200  + 130432.40216087016 *T)
    sumB0  +=       0.031 * cos(5.37268049479  + 75930.51303185058 *T)
    sumB0  +=       0.034 * cos(4.68006477569  + 25754.04720080480 *T)
    sumB0  +=       0.030 * cos(2.76917399150  + 173511.41499690176 *T)
    sumB0  +=       0.037 * cos(4.37537497770  + 1478.86657406440 *T)
    sumB0  +=       0.030 * cos(2.69970929614  + 208173.53416749896 *T)
    sumB0  +=       0.029 * cos(2.28199342452  + 130443.39704322898 *T)
    sumB0  +=       0.039 * cos(4.04390928499  + 34282.17847478280 *T)
    sumB0  +=       0.033 * cos(1.47153495142  + 55503.94193942859 *T)
    sumB0  +=       0.029 * cos(3.09374594552  + 27819.02549450680 *T)
    sumB0  +=       0.030 * cos(2.30968283835  + 26624.70765366959 *T)
    sumB0  +=       0.035 * cos(2.18534594257  + 22759.76748529401 *T)
    sumB0  +=       0.027 * cos(0.63624513172  + 240452.46032332000 *T)
    sumB0  +=       0.035 * cos(5.40588312340  + 58857.03113654799 *T)
    sumB0  +=       0.028 * cos(5.46904052697  + 123668.80510662538 *T)
    sumB0  +=       0.028 * cos(3.82629644574  + 130866.11389874700 *T)
    sumB0  +=       0.027 * cos(3.91273567728  + 163298.12945069079 *T)
    sumB0  +=       0.032 * cos(3.55783772742  + 52252.07235442399 *T)
    sumB0  +=       0.036 * cos(5.26852179816  + 52061.36699446317 *T)
    sumB0  +=       0.027 * cos(0.81868301401  + 25455.11940226100 *T)
    sumB0  +=       0.030 * cos(6.23608007707  + 24176.70365835700 *T)
    sumB0  +=       0.025 * cos(1.48874402537  + 60370.08161635699 *T)
    sumB0  +=       0.027 * cos(4.41957053501  + 78896.49316403578 *T)
    sumB0  +=       0.025 * cos(3.33404428554  + 206.18554843720 *T)
    sumB0  +=       0.025 * cos(2.22580879047  + 12725.45343477500 *T)
    sumB0  +=       0.025 * cos(5.37774707105  + 183145.01295611399 *T)
    sumB0  +=       0.026 * cos(3.83674753452  + 132028.58860315481 *T)
    sumB0  +=       0.025 * cos(2.93605355840  + 129387.24732468258 *T)
    sumB0  +=       0.025 * cos(5.86161454740  + 130289.95251073640 *T)
    sumB0  +=       0.024 * cos(4.03572089813  + 26089.38761428249 *T)
    sumB0  +=       0.024 * cos(3.09138745629  + 26086.41866886590 *T)
    sumB0  +=       0.024 * cos(1.40452977850  + 25446.48957983520 *T)
    sumB0  +=       0.033 * cos(3.35253473958  + 193937.98608932376 *T)
    sumB0  +=       0.024 * cos(4.81342275570  + 27311.72098235281 *T)
    sumB0  +=       0.024 * cos(2.36840229716  + 25440.89230825939 *T)
    sumB0  +=       0.024 * cos(2.60590286552  + 125112.03959948818 *T)
    sumB0  +=       0.027 * cos(2.64006708171  + 44295.71712980940 *T)
    sumB0  +=       0.029 * cos(3.85906422319  + 130652.81480330898 *T)
    sumB0  +=       0.033 * cos(4.26007262410  + 235900.50682626120 *T)
    sumB0  +=       0.026 * cos(3.37356285012  + 87367.61641884019 *T)
    sumB0  +=       0.024 * cos(3.83432680564  + 52381.99183158559 *T)
    sumB0  +=       0.027 * cos(4.14811634858  + 176332.24614102798 *T)
    sumB0  +=       0.023 * cos(0.19227535404  + 647.01083331480 *T)
    sumB0  +=       0.025 * cos(0.03897225418  + 129373.02023068098 *T)
    sumB0  +=       0.023 * cos(5.73072631591  + 26729.31670331319 *T)
    sumB0  +=       0.026 * cos(2.78110886808  + 64901.25971792339 *T)
    sumB0  +=       0.028 * cos(6.14191568069  + 38813.35657634920 *T)
    sumB0  +=       0.025 * cos(5.64677901976  + 198489.93958638256 *T)
    sumB0  +=       0.022 * cos(2.47489263310  + 24864.08530079559 *T)
    sumB0  +=       0.023 * cos(4.26682241390  + 26222.01219115920 *T)
    sumB0  +=       0.024 * cos(4.52962335246  + 123758.29085447140 *T)
    sumB0  +=       0.023 * cos(2.88962909157  + 25953.79409198919 *T)
    sumB0  +=       0.028 * cos(4.67408679361  + 80382.47328510099 *T)
    sumB0  +=       0.021 * cos(2.19925857864  + 23866.04650697719 *T)
    sumB0  +=       0.025 * cos(0.81737451667  + 94329.77528619740 *T)
    sumB0  +=       0.021 * cos(5.56863005528  + 52712.61079524379 *T)
    sumB0  +=       0.021 * cos(0.94657470314  + 27684.08958385880 *T)
    sumB0  +=       0.023 * cos(1.24923754668  + 130446.62925487180 *T)
    sumB0  +=       0.022 * cos(1.55521175306  + 24998.19435038059 *T)
    sumB0  +=       0.024 * cos(3.25930776298  + 27669.86248985719 *T)
    sumB0  +=       0.021 * cos(3.16657819785  + 183674.70392120857 *T)
    sumB0  +=       0.020 * cos(1.45970914613  + 75615.25459924959 *T)
    sumB0  +=       0.028 * cos(2.65124448902  + 78366.80219894118 *T)
    sumB0  +=       0.021 * cos(2.85719729554  + 78160.61665050399 *T)
    sumB0  +=       0.022 * cos(0.14164137879  + 51528.79544983359 *T)
    sumB0  +=       0.022 * cos(1.54435105177  + 158746.17595363196 *T)
    sumB0  +=       0.019 * cos(1.95606380894  + 25344.91308104160 *T)
    sumB0  +=       0.019 * cos(3.62491058376  + 78039.36462902068 *T)
    sumB0  +=       0.019 * cos(2.59809360557  + 17893.62780836560 *T)
    sumB0  +=       0.019 * cos(5.86487152937  + 103917.90082841998 *T)
    sumB0  +=       0.019 * cos(5.07516493363  + 230239.17477710897 *T)
    sumB0  +=       0.023 * cos(1.73809221823  + 128220.75860368418 *T)
    sumB0  +=       0.018 * cos(3.70453311537  + 77726.90491262719 *T)
    sumB0  +=       0.020 * cos(5.58951443277  + 27177.61193276780 *T)
    sumB0  +=       0.019 * cos(2.26298959398  + 103932.12792242158 *T)
    sumB0  +=       0.018 * cos(5.26103373994  + 28736.35796704720 *T)
    sumB0  +=       0.018 * cos(4.77075669731  + 77101.23472031478 *T)
    sumB0  +=       0.019 * cos(4.18613339997  + 104819.57755665120 *T)
    sumB0  +=       0.019 * cos(3.72825596269  + 31415.37924995700 *T)
    sumB0  +=       0.018 * cos(5.11985094325  + 77307.42026875199 *T)
    sumB0  +=       0.019 * cos(1.61447515906  + 51653.22886505459 *T)
    sumB0  +=       0.022 * cos(3.45569006244  + 101011.00013959700 *T)
    sumB0  +=       0.017 * cos(3.34823797519  + 23919.14265929160 *T)
    sumB0  +=       0.017 * cos(0.84289136609  + 86143.79857806159 *T)
    sumB0  +=       0.018 * cos(0.81997628030  + 26073.67604757259 *T)
    sumB0  +=       0.020 * cos(0.96632907921  + 52609.51802102519 *T)
    sumB0  +=       0.016 * cos(1.27839838984  + 51868.24866217880 *T)
    sumB0  +=       0.020 * cos(0.07187923898  + 155887.52156313116 *T)
    sumB0  +=       0.018 * cos(5.44524309894  + 25508.21555457540 *T)
    sumB0  +=       0.016 * cos(0.62722455150  + 102659.44689679438 *T)
    sumB0  +=       0.016 * cos(0.52102795891  + 2221.85663459700 *T)
    sumB0  +=       0.017 * cos(3.93009669878  + 51852.30086649099 *T)
    sumB0  +=       0.016 * cos(5.74180318220  + 96357.08412427260 *T)
    sumB0  +=       0.019 * cos(5.09029341511  + 118828.96374949680 *T)
    sumB0  +=       0.017 * cos(2.59947793429  + 24448.83362438620 *T)
    sumB0  +=       0.016 * cos(5.52244950928  + 2199.08734328700 *T)
    sumB0  +=       0.015 * cos(3.36667960581  + 1911.19948321720 *T)
    sumB0  +=       0.021 * cos(3.53223071778  + 181659.72224941081 *T)
    sumB0  +=       0.019 * cos(1.29524947399  + 233681.74972207440 *T)
    sumB0  +=       0.016 * cos(1.11236515889  + 103498.41618454478 *T)
    sumB0  +=       0.020 * cos(3.19274792861  + 51969.62073471119 *T)
    sumB0  +=       0.017 * cos(3.27645001526  + 26312.24793727610 *T)
    sumB0  +=       0.016 * cos(5.54159968247  + 207114.15223730975 *T)
    sumB0  +=       0.015 * cos(2.75565944373  + 26709.64694241340 *T)
    sumB0  +=       0.015 * cos(1.72732606714  + 104344.98400739678 *T)
    sumB0  +=       0.016 * cos(5.31828785394  + 78153.50310350319 *T)
    sumB0  +=       0.016 * cos(1.13565154505  + 742.99006053260 *T)
    sumB0  +=       0.015 * cos(6.19247438386  + 130363.24963659538 *T)
    sumB0  +=       0.015 * cos(1.16646378812  + 92741.06060792258 *T)
    sumB0  +=       0.015 * cos(4.39941321864  + 25466.15934073500 *T)
    sumB0  +=       0.014 * cos(1.03341872997  + 78313.70604662679 *T)
    sumB0  +=       0.016 * cos(1.77081422797  + 27044.19229754480 *T)
    sumB0  +=       0.016 * cos(1.77854347809  + 26667.59072857300 *T)
    sumB0  +=       0.014 * cos(2.39782698608  + 103883.64757594238 *T)
    sumB0  +=       0.019 * cos(6.27226219255  + 25764.39772491679 *T)
    sumB0  +=       0.014 * cos(1.89621592736  + 37698.45509994840 *T)
    sumB0  +=       0.014 * cos(5.86393829627  + 81604.32185142238 *T)
    sumB0  +=       0.015 * cos(2.41952075336  + 27250.37784598199 *T)
    sumB0  +=       0.017 * cos(5.62378615870  + 51432.81622261579 *T)
    sumB0  +=       0.014 * cos(0.57479381854  + 8194.27533320860 *T)
    sumB0  +=       0.014 * cos(2.01915873640  + 156507.74908854539 *T)
    sumB0  +=       0.014 * cos(4.00033644851  + 187167.27548807816 *T)
    sumB0  +=       0.016 * cos(3.05516572588  + 22625.65843570900 *T)
    sumB0  +=       0.014 * cos(0.41333257896  + 76887.93562487679 *T)
    sumB0  +=       0.014 * cos(3.81333935444  + 28256.66362385679 *T)
    sumB0  +=       0.016 * cos(5.63829774136  + 209658.82487420217 *T)
    sumB0  +=       0.017 * cos(2.79674558169  + 323.50541665740 *T)
    sumB0  +=       0.015 * cos(4.52292533775  + 27726.97265876220 *T)
    sumB0  +=       0.017 * cos(1.61686538851  + 26198.10946279360 *T)
    sumB0  +=       0.017 * cos(4.28628662370  + 133882.09065283637 *T)
    sumB0  +=       0.013 * cos(1.02936352404  + 25867.49049913539 *T)
    sumB0  +=       0.014 * cos(5.57651447296  + 2008.55753915900 *T)
    sumB0  +=       0.014 * cos(1.55161726953  + 156314.11975400720 *T)
    sumB0  +=       0.017 * cos(6.07989986315  + 49953.94964855139 *T)
    sumB0  +=       0.015 * cos(4.87081069542  + 104358.24112519680 *T)
    sumB0  +=       0.013 * cos(3.77228393975  + 19.66976089979 *T)
    sumB0  +=       0.015 * cos(5.41429733370  + 208276.62694171761 *T)
    sumB0  +=       0.013 * cos(1.15273910706  + 949.17560896980 *T)
    sumB0  +=       0.017 * cos(4.64614224273  + 112231.70171963578 *T)
    sumB0  +=       0.015 * cos(5.32485053265  + 154408.65498906677 *T)
    sumB0  +=       0.013 * cos(4.02426203972  + 417406.45026518719 *T)
    sumB0  +=       0.013 * cos(2.83441937048  + 25169.97285559240 *T)
    sumB0  +=       0.016 * cos(3.55004561005  + 86457.98475793119 *T)
    sumB0  +=       0.014 * cos(3.36717179008  + 12546.48193908300 *T)
    sumB0  +=       0.013 * cos(0.99900287294  + 78338.49102328988 *T)
    sumB0  +=       0.012 * cos(3.72935261349  + 78786.28684281638 *T)
    sumB0  +=       0.013 * cos(4.55769261775  + 80596.90581759460 *T)
    sumB0  +=       0.013 * cos(3.68794178346  + 79315.97780791098 *T)
    sumB0  +=       0.014 * cos(4.16918945992  + 156547.08861034497 *T)
    sumB0  +=       0.012 * cos(3.12295185577  + 149288.74325784517 *T)
    sumB0  +=       0.012 * cos(2.69822851627  + 153.77881048480 *T)
    sumB0  +=       0.012 * cos(4.38117483432  + 27005.83342755599 *T)
    sumB0  +=       0.012 * cos(0.19322533903  + 27463.67694142000 *T)
    sumB0  +=       0.014 * cos(4.15127082280  + 29550.14784743939 *T)
    sumB0  +=       0.012 * cos(2.80411772823  + 102755.42612401220 *T)
    sumB0  +=       0.011 * cos(5.82734448081  + 259819.64948555280 *T)
    sumB0  +=       0.011 * cos(2.40607503106  + 169093.57054942797 *T)
    sumB0  +=       0.012 * cos(0.49737747696  + 102769.65321801380 *T)
    sumB0  +=       0.014 * cos(5.69923121834  + 76784.84285065818 *T)
    sumB0  +=       0.012 * cos(5.56466695759  + 26402.08932144380 *T)
    sumB0  +=       0.012 * cos(1.48301102812  + 78580.10129437919 *T)
    sumB0  +=       0.011 * cos(5.15388067065  + 77947.31755506598 *T)
    sumB0  +=       0.011 * cos(1.64586759073  + 25773.71696170459 *T)
    sumB0  +=       0.011 * cos(2.45779334771  + 34082.43158359840 *T)
    sumB0  +=       0.014 * cos(2.00354520885  + 846.08283475120 *T)
    sumB0  +=       0.013 * cos(0.71980668578  + 78413.27262185719 *T)
    sumB0  +=       0.011 * cos(5.17906120193  + 162810.49469943656 *T)
    sumB0  +=       0.011 * cos(2.73596940194  + 171292.65789271498 *T)
    sumB0  +=       0.011 * cos(5.59141622457  + 24203.00197815680 *T)
    sumB0  +=       0.011 * cos(0.45608202666  + 2111.65031337760 *T)
    sumB0  +=       0.010 * cos(3.29404880642  + 192828.60753723039 *T)
    sumB0  +=       0.010 * cos(0.45857052500  + 222224.97657418496 *T)
    sumB0  +=       0.013 * cos(5.70684493376  + 78213.71280281838 *T)
    sumB0  +=       0.011 * cos(4.37474459591  + 103718.82882698360 *T)
    sumB0  +=       0.011 * cos(0.63918585951  + 26118.23000257860 *T)
    sumB0  +=       0.013 * cos(5.83092856998  + 26411.40855823160 *T)
    sumB0  +=       0.010 * cos(5.45367935399  + 52483.36390411799 *T)
    sumB0  +=       0.010 * cos(4.70344741563  + 2125.87740737920 *T)
    sumB0  +=       0.011 * cos(0.28816338163  + 26057.57628056979 *T)
    sumB0  +=       0.010 * cos(5.57764903480  + 151199.94274106238 *T)
    sumB0  +=       0.011 * cos(0.03362968403  + 104991.50985261080 *T)
    sumB0  +=       0.010 * cos(3.71565881071  + 148532.89040742096 *T)
    sumB0  +=       0.013 * cos(0.25119201649  + 79330.20490191258 *T)
    sumB0  +=       0.010 * cos(6.23763004539  + 128106.31931499895 *T)
    sumB0  +=       0.011 * cos(6.23138757669  + 130285.73689738620 *T)
    sumB0  +=       0.013 * cos(3.60703035316  + 39629.32434406539 *T)
    sumB0  +=       0.011 * cos(2.08623018213  + 78683.19406859778 *T)
    sumB0  +=       0.011 * cos(2.89478268807  + 122444.98726584678 *T)
    sumB0  +=       0.011 * cos(1.47224457688  + 156520.30530244438 *T)
    sumB0  +=       0.012 * cos(0.18308102701  + 130593.29451835580 *T)
    sumB0  +=       0.009 * cos(5.24611723729  + 138319.60486120995 *T)
    sumB0  +=       0.009 * cos(5.64920582852  + 1265.56747862640 *T)
    sumB0  +=       0.009 * cos(6.25551575071  + 53906.92863608099 *T)
    sumB0  +=       0.011 * cos(3.36985744471  + 142871.55835826878 *T)
    sumB0  +=       0.009 * cos(5.47085206361  + 156531.30018480320 *T)
    sumB0  +=       0.009 * cos(4.79150958411  + 101703.15774082378 *T)
    sumB0  +=       0.009 * cos(2.73233240797  + 38.13303563780 *T)
    sumB0  +=       0.009 * cos(5.91954900221  + 62197.64356324779 *T)
    sumB0  +=       0.009 * cos(4.73154570336  + 78378.14871340781 *T)
    sumB0  +=       0.011 * cos(0.80389582741  + 3328.13565628019 *T)
    sumB0  +=       0.009 * cos(6.25489842872  + 91919.56991589900 *T)
    sumB0  +=       0.009 * cos(5.87197255624  + 199599.31813847597 *T)
    sumB0  +=       0.009 * cos(0.76033087382  + 156523.53751408719 *T)
    sumB0  +=       0.008 * cos(0.63159267832  + 80174.90890793978 *T)
    sumB0  +=       0.009 * cos(5.02352759781  + 229129.79622501560 *T)
    sumB0  +=       0.008 * cos(2.63272452796  + 25138.72753260440 *T)
    sumB0  +=       0.008 * cos(0.59154768044  + 156954.01704032117 *T)
    sumB0  +=       0.008 * cos(3.95681232703  + 120417.67842777158 *T)
    sumB0  +=       0.008 * cos(3.46882638388  + 77616.69859140778 *T)
    sumB0  +=       0.008 * cos(0.20349823136  + 113455.51956041438 *T)
    sumB0  +=       0.008 * cos(0.21724719953  + 13541.42120249119 *T)
    sumB0  +=       0.008 * cos(1.77719128692  + 26076.85744131030 *T)
    sumB0  +=       0.009 * cos(6.21878690440  + 189853.99758261937 *T)
    sumB0  +=       0.007 * cos(5.88768146200  + 51226.63067417859 *T)
    sumB0  +=       0.007 * cos(0.61161547244  + 158116.49174472899 *T)
    sumB0  +=       0.008 * cos(2.05860519043  + 78149.27013603736 *T)
    sumB0  +=       0.007 * cos(0.09500184933  + 90989.16285949761 *T)
    sumB0  +=       0.007 * cos(0.56541937862  + 25985.94062330859 *T)
    sumB0  +=       0.007 * cos(0.73212945216  + 78469.89497315978 *T)
    sumB0  +=       0.007 * cos(0.44088896603  + 104127.26777059487 *T)
    sumB0  +=       0.008 * cos(5.86653927002  + 104454.70534051540 *T)
    sumB0  +=       0.007 * cos(0.70585294359  + 52177.29075585669 *T)
    sumB0  +=       0.009 * cos(1.96929942879  + 1272.68102562720 *T)
    sumB0  +=       0.010 * cos(2.72173324175  + 104276.83096772947 *T)
    sumB0  +=       0.007 * cos(5.16315798030  + 26189.86565983980 *T)
    sumB0  +=       0.007 * cos(5.17079740730  + 179172.74704605396 *T)
    sumB0  +=       0.008 * cos(5.26981396126  + 48847.67062686820 *T)
    sumB0  +=       0.008 * cos(0.19880797043  + 1.48447270830 *T)
    sumB0  +=       0.007 * cos(6.26371276121  + 52174.32181044009 *T)
    sumB0  +=       0.008 * cos(0.98956493578  + 16703.06213349900 *T)
    sumB0  +=       0.007 * cos(0.08448723866  + 26098.94884183810 *T)
    sumB0  +=       0.008 * cos(5.67030179581  + 53029.00266490040 *T)
    sumB0  +=       0.007 * cos(1.15333561308  + 261988.40996783532 *T)
    sumB0  +=       0.007 * cos(0.70724943755  + 65717.22748563958 *T)
    sumB0  +=       0.007 * cos(0.52800551865  + 103814.80805420138 *T)
    sumB0  +=       0.008 * cos(2.25202792521  + 149756.70824819960 *T)
    sumB0  +=       0.007 * cos(1.45586274187  + 51841.95034237900 *T)
    sumB0  +=       0.007 * cos(2.28027029168  + 54374.89362643540 *T)
    sumB0  +=       0.007 * cos(1.96795602434  + 103395.32341032618 *T)
    sumB0  +=       0.007 * cos(0.91100826202  + 27972.80430499159 *T)
    sumB0  +=       0.007 * cos(6.04142381509  + 155475.15046625675 *T)
    sumB0  +=       0.008 * cos(0.66187102958  + 156740.71794488319 *T)
    sumB0  +=       0.007 * cos(0.94700216301  + 202420.14928260216 *T)
    sumB0  +=       0.008 * cos(0.34397341638  + 78339.97549599818 *T)
    sumB0  +=       0.008 * cos(1.99164711035  + 84944.93427812219 *T)
    sumB0  +=       0.006 * cos(3.16647388859  + 76041.85279012559 *T)
    sumB0  +=       0.007 * cos(1.18920934641  + 104984.39630560996 *T)
    sumB0  +=       0.007 * cos(6.12812258419  + 53867.97195265079 *T)
    sumB0  +=       0.007 * cos(2.68832841048  + 156377.85565231059 *T)
    sumB0  +=       0.007 * cos(4.82261895829  + 52286.01260436779 *T)
    sumB0  +=       0.006 * cos(2.12110075547  + 148.07872442630 *T)
    sumB0  +=       0.006 * cos(3.05937939252  + 151.04766984290 *T)
    sumB0  +=       0.007 * cos(0.09315209734  + 220025.88923089800 *T)
    sumB0  +=       0.008 * cos(5.56250511075  + 52509.66222391780 *T)
    sumB0  +=       0.006 * cos(2.57622376266  + 224577.84272795680 *T)
    sumB0  +=       0.007 * cos(2.65171386195  + 220.41264243880 *T)
    sumB0  +=       0.006 * cos(2.00091220299  + 102018.41617342478 *T)
    sumB0  +=       0.007 * cos(0.86158158633  + 50444.68393021580 *T)
    sumB0  +=       0.007 * cos(1.61990482140  + 103189.13786188900 *T)
    sumB0  +=       0.006 * cos(2.68512587119  + 77520.71936418998 *T)
    sumB0  +=       0.006 * cos(4.82671721903  + 154308.66174525840 *T)
    sumB0  +=       0.006 * cos(2.40727316051  + 78800.51393681798 *T)
    sumB0  +=       0.006 * cos(5.66286619587  + 234261.43730907320 *T)
    sumB0  +=       0.007 * cos(1.38238075345  + 106470.37642667518 *T)
    sumB0  +=       0.007 * cos(4.62193558950  + 81591.84508100279 *T)
    sumB0  +=       0.007 * cos(0.26927581545  + 127098.90328117118 *T)
    sumB0  +=       0.006 * cos(0.92172283144  + 57503.28239153120 *T)
    sumB0  +=       0.006 * cos(5.39808870469  + 130020.03106399580 *T)
    sumB0  +=       0.006 * cos(2.52761413373  + 52817.21984488739 *T)
    sumB0  +=       0.006 * cos(0.06817978613  + 53757.76563143139 *T)
    sumB0  +=       0.007 * cos(3.13547312140  + 155460.92337225520 *T)
    sumB0  +=       0.005 * cos(2.17501494391  + 209232.91609768817 *T)
    sumB0  +=       0.005 * cos(2.16759656006  + 104241.40624507738 *T)
    sumB0  +=       0.006 * cos(2.68409858369  + 130005.80396999416 *T)
    sumB0  +=       0.005 * cos(1.58914721681  + 50167.24874398939 *T)
    sumB0  +=       0.005 * cos(1.01687539089  + 55638.05098901359 *T)
    sumB0  +=       0.007 * cos(1.31888655444  + 149846.19399604559 *T)
    sumB0  +=       0.005 * cos(2.48117306579  + 9384.84100807520 *T)
    sumB0  +=       0.005 * cos(3.70948655428  + 266540.36346489412 *T)
    sumB0  +=       0.007 * cos(5.71513815276  + 70383.62027138360 *T)
    sumB0  +=       0.006 * cos(4.55015130253  + 51955.39364070959 *T)
    sumB0  +=       0.005 * cos(3.51602263966  + 102975.83876645098 *T)
    sumB0  +=       0.005 * cos(3.69945557546  + 39450.35284837340 *T)
    sumB0  +=       0.005 * cos(3.98232048303  + 53771.99272543299 *T)
    sumB0  +=       0.006 * cos(1.01394190869  + 77940.20400806518 *T)
    sumB0  +=       0.005 * cos(2.50588363651  + 102872.74599223241 *T)
    sumB0  +=       0.007 * cos(3.01244944080  + 50264.60679993120 *T)
    sumB0  +=       0.006 * cos(0.45316437598  + 50064.15596977079 *T)
    sumB0  +=       0.007 * cos(6.03933985106  + 1485.98012106520 *T)
    sumB0  +=       0.005 * cos(1.35734487293  + 53814.87580033639 *T)
    sumB0  +=       0.005 * cos(4.43939776623  + 51534.39272140940 *T)
    sumB0  +=       0.005 * cos(2.29075654159  + 61560.64729122359 *T)
    sumB0  +=       0.005 * cos(0.07896316759  + 78057.52387628538 *T)
    sumB0  +=       0.005 * cos(0.95484155225  + 130907.48069822539 *T)
    sumB0  +=       0.005 * cos(1.56485117164  + 53399.62412392700 *T)
    sumB0  +=       0.005 * cos(4.55586232099  + 52137.67324751059 *T)
    sumB0  +=       0.006 * cos(4.34778590755  + 156534.53239644598 *T)
    sumB0  +=       0.005 * cos(3.72691818970  + 128747.35003836860 *T)
    sumB0  +=       0.004 * cos(0.56386565513  + 189386.03259226491 *T)
    sumB0  +=       0.004 * cos(2.67910077904  + 107692.22499299661 *T)
    sumB0  +=       0.006 * cos(6.10947724529  + 104248.51979207818 *T)
    sumB0  +=       0.004 * cos(3.99859010561  + 104401.60918820098 *T)
    sumB0  +=       0.004 * cos(0.51280078907  + 54344.56676543099 *T)
    sumB0  +=       0.005 * cos(4.19917665740  + 129586.31932611900 *T)
    sumB0  +=       0.004 * cos(5.91763059229  + 52041.69723356339 *T)
    sumB0  +=       0.005 * cos(3.18633834950  + 181975.42470470537 *T)
    sumB0  +=       0.005 * cos(0.99354989216  + 1795.25844372100 *T)
    sumB0  +=       0.004 * cos(6.20030341798  + 209762.60706278280 *T)
    sumB0  +=       0.004 * cos(4.57601261459  + 184834.07909520617 *T)
    sumB0  +=       0.006 * cos(4.81632935419  + 77741.13200662879 *T)
    sumB0  +=       0.004 * cos(2.32553366031  + 53265.51507434200 *T)
    sumB0  +=       0.004 * cos(4.83867556795  + 52755.49387014719 *T)
    sumB0  +=       0.004 * cos(0.47379105299  + 105403.88094948516 *T)
    sumB0  +=       0.004 * cos(2.62000942355  + 50049.92887576919 *T)
    sumB0  +=       0.005 * cos(1.01847704961  + 52309.91533273340 *T)
    sumB0  +=       0.004 * cos(5.12416312887  + 52027.72755872209 *T)
    sumB0  +=       0.004 * cos(0.26524339921  + 50007.04580086580 *T)
    sumB0  +=       0.005 * cos(0.30725353996  + 207747.62539098496 *T)
    sumB0  +=       0.004 * cos(4.93794243881  + 63786.35824152260 *T)
    sumB0  +=       0.004 * cos(4.51168915835  + 77956.15180375299 *T)
    sumB0  +=       0.005 * cos(1.16733741695  + 27566.76971563859 *T)
    sumB0  +=       0.004 * cos(3.77703754152  + 104501.17576343138 *T)
    sumB0  +=       0.004 * cos(5.44859511930  + 129971.55071751658 *T)
    sumB0  +=       0.005 * cos(5.22593130125  + 51329.72344839720 *T)
    sumB0  +=       0.004 * cos(1.64237114391  + 130446.14426677101 *T)
    sumB0  +=       0.004 * cos(1.28513103275  + 112545.88789950538 *T)
    sumB0  +=       0.004 * cos(3.50115780256  + 76152.05911134499 *T)
    sumB0  +=       0.004 * cos(4.79032272131  + 130432.88714897096 *T)
    sumB0  +=       0.004 * cos(5.62786746084  + 43981.53094993980 *T)
    sumB0  +=       0.004 * cos(4.12687975050  + 52024.75861330549 *T)
    sumB0  +=       0.004 * cos(5.13202982684  + 735.87651353180 *T)
    sumB0  +=       0.004 * cos(4.64327471216  + 76255.15188556358 *T)
    sumB0  +=       0.004 * cos(1.34781477964  + 106684.80895916880 *T)
    sumB0  +=       0.004 * cos(2.34997151211  + 177287.84588263658 *T)

    # MERCURY Latitude
    sumB1  = 0
    sumB1  +=  429151.362 * cos(3.50169780393  + 26087.90314157420 *T)
    sumB1  +=  146233.668 * cos(3.14159265359  + 0.00000000000 *T)
    sumB1  +=   22675.295 * cos(0.01515366880  + 52175.80628314840 *T)
    sumB1  +=   10894.981 * cos(0.48540174006  + 78263.70942472259 *T)
    sumB1  +=    6353.462 * cos(3.42943919982  + 104351.61256629678 *T)
    sumB1  +=    2495.743 * cos(0.16051210665  + 130439.51570787100 *T)
    sumB1  +=     859.585 * cos(3.18452433647  + 156527.41884944518 *T)
    sumB1  +=     277.503 * cos(6.21020774184  + 182615.32199101936 *T)
    sumB1  +=      86.233 * cos(2.95244391822  + 208703.22513259359 *T)
    sumB1  +=      26.133 * cos(5.97708962692  + 234791.12827416777 *T)
    sumB1  +=      27.696 * cos(0.29068938889  + 27197.28169366760 *T)
    sumB1  +=      12.831 * cos(3.37744320558  + 53285.18483524180 *T)
    sumB1  +=      12.720 * cos(0.53792661684  + 24978.52458948080 *T)
    sumB1  +=       7.781 * cos(2.71768609268  + 260879.03141574195 *T)
    sumB1  +=       7.527 * cos(3.58305121268  + 51066.42773105500 *T)
    sumB1  +=       6.183 * cos(2.92383205004  + 31749.23519072640 *T)
    sumB1  +=       5.453 * cos(1.97318763801  + 51116.42435295920 *T)
    sumB1  +=       3.394 * cos(0.34761695275  + 77154.33087262919 *T)
    sumB1  +=       3.481 * cos(0.10739761667  + 79373.08797681599 *T)
    sumB1  +=       2.932 * cos(5.95430013169  + 57837.13833230060 *T)
    sumB1  +=       2.742 * cos(0.98758439378  + 25028.52121138500 *T)
    sumB1  +=       2.126 * cos(4.16352818018  + 47623.85278608960 *T)
    sumB1  +=       2.286 * cos(5.74036496396  + 286966.93455731618 *T)
    sumB1  +=       2.380 * cos(4.00213666550  + 21535.94964451540 *T)
    sumB1  +=       1.651 * cos(4.43606584681  + 27043.50288318280 *T)
    sumB1  +=       1.523 * cos(1.97926797181  + 5661.33204915220 *T)
    sumB1  +=       1.518 * cos(5.01636479848  + 77204.32749453338 *T)
    sumB1  +=       1.398 * cos(0.66375029294  + 20426.57109242200 *T)
    sumB1  +=       1.345 * cos(3.39193943388  + 103242.23401420339 *T)
    sumB1  +=       1.297 * cos(5.34251327744  + 37410.56723987860 *T)
    sumB1  +=       1.076 * cos(0.95624460335  + 50586.73338786459 *T)
    sumB1  +=       1.050 * cos(3.00944151152  + 26107.57290247399 *T)
    sumB1  +=       0.918 * cos(1.89988871166  + 1059.38193018920 *T)
    sumB1  +=       1.063 * cos(1.28486292967  + 53131.40602475700 *T)
    sumB1  +=       0.905 * cos(3.34776562702  + 25558.21217647960 *T)
    sumB1  +=       0.929 * cos(4.81159652409  + 51646.11531805379 *T)
    sumB1  +=       0.782 * cos(6.08817839271  + 529.69096509460 *T)
    sumB1  +=       0.783 * cos(1.50962699410  + 41962.52073693740 *T)
    sumB1  +=       0.794 * cos(2.16868396183  + 63498.47038145279 *T)
    sumB1  +=       0.731 * cos(0.92148918837  + 73711.75592766379 *T)
    sumB1  +=       0.755 * cos(4.10910637390  + 26068.23338067440 *T)
    sumB1  +=       0.703 * cos(2.22466530621  + 10213.28554621100 *T)
    sumB1  +=       0.922 * cos(2.30268247092  + 1109.37855209340 *T)
    sumB1  +=       0.739 * cos(3.05842748947  + 105460.99111839020 *T)
    sumB1  +=       0.681 * cos(1.40695567526  + 51749.20809227239 *T)
    sumB1  +=       0.841 * cos(5.93206787810  + 26084.02180621620 *T)
    sumB1  +=       0.616 * cos(5.95368982430  + 39609.65458316560 *T)
    sumB1  +=       0.610 * cos(2.96796433797  + 25132.30339996560 *T)
    sumB1  +=       0.721 * cos(1.16472614778  + 46514.47423399620 *T)
    sumB1  +=       0.638 * cos(1.16941585214  + 26091.78447693220 *T)
    sumB1  +=       0.633 * cos(3.07427024064  + 30639.85663863300 *T)
    sumB1  +=       0.665 * cos(2.47994200156  + 313054.83769889042 *T)
    sumB1  +=       0.669 * cos(4.10310179631  + 72602.37737557039 *T)
    sumB1  +=       0.495 * cos(0.14776350456  + 129330.13715577760 *T)
    sumB1  +=       0.474 * cos(2.61466823958  + 32858.61374281979 *T)
    sumB1  +=       0.470 * cos(5.04436148211  + 13521.75144159140 *T)
    sumB1  +=       0.494 * cos(0.07430556450  + 28306.66024576099 *T)
    sumB1  +=       0.459 * cos(5.81611579004  + 51220.20654153979 *T)
    sumB1  +=       0.467 * cos(5.56622114735  + 4551.95349705880 *T)
    sumB1  +=       0.433 * cos(2.10197184422  + 24498.83024629040 *T)
    sumB1  +=       0.430 * cos(3.91046377635  + 76674.63652943878 *T)
    sumB1  +=       0.436 * cos(1.46213785727  + 43071.89928903080 *T)
    sumB1  +=       0.383 * cos(6.05537830853  + 52195.47604404819 *T)
    sumB1  +=       0.335 * cos(0.55996605330  + 23869.14603738740 *T)
    sumB1  +=       0.369 * cos(4.43217995206  + 22645.32819660879 *T)
    sumB1  +=       0.329 * cos(2.59508442112  + 65697.55772473979 *T)
    sumB1  +=       0.346 * cos(2.03170085305  + 27147.28507176339 *T)
    sumB1  +=       0.424 * cos(0.29775934661  + 1589.07289528380 *T)
    sumB1  +=       0.378 * cos(5.09810178577  + 53235.18821333759 *T)
    sumB1  +=       0.308 * cos(3.30225015955  + 54394.56338733519 *T)
    sumB1  +=       0.314 * cos(2.40932519602  + 83925.04147387479 *T)
    sumB1  +=       0.364 * cos(5.87784809361  + 58946.51688439399 *T)
    sumB1  +=       0.352 * cos(0.87939481900  + 52156.13652224860 *T)
    sumB1  +=       0.352 * cos(5.92675293084  + 25661.30495069820 *T)
    sumB1  +=       0.310 * cos(2.69873876637  + 52171.92494779040 *T)
    sumB1  +=       0.345 * cos(0.84406679212  + 98690.28051714458 *T)
    sumB1  +=       0.358 * cos(6.14890038263  + 56727.75978020720 *T)
    sumB1  +=       0.248 * cos(2.34396840060  + 51535.90899683439 *T)
    sumB1  +=       0.286 * cos(2.39648280163  + 19804.82729158280 *T)
    sumB1  +=       0.338 * cos(1.55189200325  + 52705.49724824299 *T)
    sumB1  +=       0.239 * cos(3.35169081927  + 52168.69273614759 *T)
    sumB1  +=       0.234 * cos(2.51108507067  + 77308.10968311399 *T)
    sumB1  +=       0.229 * cos(5.16737598091  + 36301.18868778519 *T)
    sumB1  +=       0.245 * cos(4.26996032400  + 19317.19254032860 *T)
    sumB1  +=       0.210 * cos(3.16281965425  + 29530.47808653960 *T)
    sumB1  +=       0.212 * cos(5.33861521612  + 15874.61759536320 *T)
    sumB1  +=       0.291 * cos(4.22929240183  + 52179.68761850640 *T)
    sumB1  +=       0.263 * cos(4.64138649646  + 69159.80243060499 *T)
    sumB1  +=       0.235 * cos(5.08060846802  + 59414.48187474840 *T)
    sumB1  +=       0.251 * cos(1.90047504790  + 14765.23904326980 *T)
    sumB1  +=       0.253 * cos(4.21581970504  + 79219.30916633119 *T)
    sumB1  +=       0.194 * cos(5.05904557922  + 51109.31080595839 *T)
    sumB1  +=       0.188 * cos(3.90664116889  + 48733.23133818299 *T)
    sumB1  +=       0.207 * cos(2.89949362192  + 7.11354700080 *T)
    sumB1  +=       0.179 * cos(4.43018837604  + 77837.11123384659 *T)
    sumB1  +=       0.223 * cos(2.96569885227  + 110012.94461544900 *T)
    sumB1  +=       0.176 * cos(5.15204624857  + 89586.37352302698 *T)
    sumB1  +=       0.215 * cos(4.50903357312  + 78793.40038981718 *T)
    sumB1  +=       0.173 * cos(3.18496771460  + 155418.04029735178 *T)
    sumB1  +=       0.205 * cos(1.89888752382  + 79323.09135491178 *T)
    sumB1  +=       0.175 * cos(2.91183047213  + 82815.66292178139 *T)
    sumB1  +=       0.165 * cos(0.49123694380  + 25035.63475838580 *T)
    sumB1  +=       0.183 * cos(1.42148796836  + 77734.01845962799 *T)
    sumB1  +=       0.158 * cos(3.63030783882  + 49957.04917896160 *T)
    sumB1  +=       0.201 * cos(4.78711687782  + 40853.14218484400 *T)
    sumB1  +=       0.157 * cos(3.02451795123  + 52182.91983014920 *T)
    sumB1  +=       0.192 * cos(5.50218209050  + 339142.74084046454 *T)
    sumB1  +=       0.145 * cos(3.39803928124  + 45892.73043315699 *T)
    sumB1  +=       0.140 * cos(2.64826898848  + 16983.99614745660 *T)
    sumB1  +=       0.190 * cos(3.05043439775  + 27676.97603685800 *T)
    sumB1  +=       0.169 * cos(3.40273824751  + 25874.60404613620 *T)
    sumB1  +=       0.148 * cos(3.87588943419  + 124778.18365871880 *T)
    sumB1  +=       0.134 * cos(0.94930028730  + 25021.40766438420 *T)
    sumB1  +=       0.174 * cos(5.88326736492  + 136100.84775702321 *T)
    sumB1  +=       0.138 * cos(2.34393108627  + 1052.26838318840 *T)
    sumB1  +=       0.163 * cos(6.05484015332  + 26080.78959457339 *T)
    sumB1  +=       0.142 * cos(3.07209103150  + 3442.57494496540 *T)
    sumB1  +=       0.137 * cos(2.10984857727  + 426.59819087600 *T)
    sumB1  +=       0.126 * cos(1.89432609260  + 1066.49547719000 *T)
    sumB1  +=       0.122 * cos(1.04298624291  + 45405.09568190280 *T)
    sumB1  +=       0.135 * cos(4.66414339767  + 25448.00585526019 *T)
    sumB1  +=       0.115 * cos(2.81365851520  + 78283.37918562238 *T)
    sumB1  +=       0.114 * cos(4.47315411813  + 68050.42387851159 *T)
    sumB1  +=       0.134 * cos(5.35393682799  + 213.29909543800 *T)
    sumB1  +=       0.125 * cos(1.61827845372  + 129380.13377768178 *T)
    sumB1  +=       0.158 * cos(5.53788931475  + 9103.90699411760 *T)
    sumB1  +=       0.116 * cos(0.50075235780  + 102762.53967101300 *T)
    sumB1  +=       0.136 * cos(0.24559658100  + 50057.04242277000 *T)
    sumB1  +=       0.124 * cos(2.20590858584  + 33326.57873317420 *T)
    sumB1  +=       0.106 * cos(6.15493608726  + 11322.66409830440 *T)
    sumB1  +=       0.123 * cos(4.43265735200  + 639.89728631400 *T)
    sumB1  +=       0.137 * cos(3.93191153638  + 78244.03966382280 *T)
    sumB1  +=       0.138 * cos(4.46503235172  + 51123.53789995999 *T)
    sumB1  +=       0.119 * cos(3.64440622006  + 26617.59410666880 *T)
    sumB1  +=       0.099 * cos(6.18313188027  + 55618.38122811380 *T)
    sumB1  +=       0.101 * cos(5.52733987069  + 103396.01282468818 *T)
    sumB1  +=       0.107 * cos(4.79994153405  + 6770.71060124560 *T)
    sumB1  +=       0.096 * cos(0.01853371395  + 80482.46652890938 *T)
    sumB1  +=       0.115 * cos(1.57423656512  + 66941.04532641819 *T)
    sumB1  +=       0.111 * cos(1.00588061631  + 78267.59076008058 *T)
    sumB1  +=       0.121 * cos(5.90839328752  + 26095.01668857500 *T)
    sumB1  +=       0.095 * cos(5.74549458142  + 78259.82808936459 *T)
    sumB1  +=       0.100 * cos(1.90676485528  + 103292.23063610760 *T)
    sumB1  +=       0.107 * cos(5.73194510449  + 131548.89425996438 *T)
    sumB1  +=       0.098 * cos(5.49064814924  + 77623.81213840858 *T)
    sumB1  +=       0.112 * cos(0.87682090108  + 74821.13447975718 *T)
    sumB1  +=       0.090 * cos(5.82373560803  + 955.59974160860 *T)
    sumB1  +=       0.093 * cos(0.15279056716  + 78256.59587772179 *T)
    sumB1  +=       0.100 * cos(1.21759895953  + 104881.30353139140 *T)
    sumB1  +=       0.087 * cos(4.20722631983  + 12566.15169998280 *T)
    sumB1  +=       0.088 * cos(5.00176994115  + 38519.94579197200 *T)
    sumB1  +=       0.086 * cos(1.23375034376  + 26301.20223701220 *T)
    sumB1  +=       0.085 * cos(1.88852891739  + 77197.21394753258 *T)
    sumB1  +=       0.082 * cos(3.75652571856  + 99799.65906923798 *T)
    sumB1  +=       0.089 * cos(4.95684147986  + 105410.99449648600 *T)
    sumB1  +=       0.096 * cos(5.10490831351  + 52389.10537858640 *T)
    sumB1  +=       0.075 * cos(2.50722139185  + 2218.75710418680 *T)
    sumB1  +=       0.093 * cos(6.00229475799  + 53764.87917843220 *T)
    sumB1  +=       0.073 * cos(6.11603054445  + 78270.82297172339 *T)
    sumB1  +=       0.091 * cos(4.67762120545  + 155468.03691925600 *T)
    sumB1  +=       0.080 * cos(1.98564256640  + 7238.67559160000 *T)
    sumB1  +=       0.089 * cos(1.62820970873  + 6283.07584999140 *T)
    sumB1  +=       0.073 * cos(5.93728826025  + 108903.56606335558 *T)
    sumB1  +=       0.087 * cos(2.61787684648  + 162188.75089859738 *T)
    sumB1  +=       0.092 * cos(2.55298032665  + 85034.42002596818 *T)
    sumB1  +=       0.076 * cos(4.44230028342  + 38654.05484155699 *T)
    sumB1  +=       0.070 * cos(2.07147354168  + 25934.12433108940 *T)
    sumB1  +=       0.078 * cos(1.79664677749  + 23969.13928119580 *T)
    sumB1  +=       0.068 * cos(2.15267768357  + 64607.84893354619 *T)
    sumB1  +=       0.069 * cos(1.37976390686  + 77211.44104153418 *T)
    sumB1  +=       0.080 * cos(0.75440631101  + 2118.76386037840 *T)
    sumB1  +=       0.060 * cos(5.95179539253  + 18849.22754997420 *T)
    sumB1  +=       0.058 * cos(0.98126270514  + 51962.50718771040 *T)
    sumB1  +=       0.057 * cos(0.63495768344  + 150866.08680029298 *T)
    sumB1  +=       0.063 * cos(0.40577774917  + 76044.95232053580 *T)
    sumB1  +=       0.061 * cos(3.20728040856  + 76144.94556434419 *T)
    sumB1  +=       0.053 * cos(1.34424134476  + 51742.09454527159 *T)
    sumB1  +=       0.052 * cos(4.11944860444  + 71492.99882347698 *T)
    sumB1  +=       0.054 * cos(2.23964604878  + 44937.13069154840 *T)
    sumB1  +=       0.064 * cos(1.19785741966  + 64741.95798313119 *T)
    sumB1  +=       0.058 * cos(6.22316473811  + 181505.94343892601 *T)
    sumB1  +=       0.060 * cos(2.12727479557  + 88476.99497093359 *T)
    sumB1  +=       0.049 * cos(1.76854754615  + 13655.86049117640 *T)
    sumB1  +=       0.049 * cos(1.27506119902  + 26727.80042788820 *T)
    sumB1  +=       0.048 * cos(2.04581990152  + 78477.00852016058 *T)
    sumB1  +=       0.050 * cos(4.62948267835  + 93028.94846799241 *T)
    sumB1  +=       0.055 * cos(5.47526225139  + 62389.09182935939 *T)
    sumB1  +=       0.047 * cos(4.99113735496  + 27154.39861876420 *T)
    sumB1  +=       0.055 * cos(4.79762177368  + 94138.32702008578 *T)
    sumB1  +=       0.051 * cos(5.56459812422  + 52602.40447402440 *T)
    sumB1  +=       0.055 * cos(1.20462687751  + 95247.70557217918 *T)
    sumB1  +=       0.050 * cos(2.21530140501  + 365230.64398203877 *T)
    sumB1  +=       0.049 * cos(0.68636780387  + 104331.94280539700 *T)
    sumB1  +=       0.047 * cos(4.30403626352  + 23754.70674870219 *T)
    sumB1  +=       0.042 * cos(3.32657201754  + 25654.19140369739 *T)
    sumB1  +=       0.051 * cos(1.08082309375  + 25619.93815121980 *T)
    sumB1  +=       0.047 * cos(2.79966035046  + 26514.50133245020 *T)
    sumB1  +=       0.040 * cos(2.91932668980  + 81706.28436968799 *T)
    sumB1  +=       0.040 * cos(4.65660434525  + 129909.82474277640 *T)
    sumB1  +=       0.041 * cos(4.21875403274  + 130969.20667296558 *T)
    sumB1  +=       0.050 * cos(1.46644208786  + 120226.23016166000 *T)
    sumB1  +=       0.047 * cos(3.77053215629  + 25973.46385288896 *T)
    sumB1  +=       0.047 * cos(2.82867670887  + 25668.41849769900 *T)
    sumB1  +=       0.046 * cos(4.14427477722  + 125887.56221081220 *T)
    sumB1  +=       0.039 * cos(4.06156914485  + 104355.49390165480 *T)
    sumB1  +=       0.037 * cos(3.22245380356  + 51322.60990139639 *T)
    sumB1  +=       0.039 * cos(2.26439106521  + 129483.91596626240 *T)
    sumB1  +=       0.040 * cos(0.54598734169  + 105307.21230790540 *T)
    sumB1  +=       0.035 * cos(5.53895897820  + 433.71173787680 *T)
    sumB1  +=       0.038 * cos(6.05329467990  + 419.48464387520 *T)
    sumB1  +=       0.039 * cos(2.72082598885  + 79852.78232000639 *T)
    sumB1  +=       0.035 * cos(1.72937228501  + 131498.89763806018 *T)
    sumB1  +=       0.034 * cos(0.83611265968  + 12432.04265039780 *T)
    sumB1  +=       0.032 * cos(4.45254278754  + 50579.61984086379 *T)
    sumB1  +=       0.042 * cos(1.41457022130  + 181555.94006083015 *T)
    sumB1  +=       0.032 * cos(4.73349986920  + 52026.24308601380 *T)
    sumB1  +=       0.036 * cos(5.60766535694  + 188276.65404017159 *T)
    sumB1  +=       0.033 * cos(3.98525990131  + 78050.41032928458 *T)
    sumB1  +=       0.032 * cos(5.86059322955  + 104371.28232719658 *T)
    sumB1  +=       0.040 * cos(5.45287901295  + 27140.17152476259 *T)
    sumB1  +=       0.031 * cos(5.17982917253  + 71025.03383312259 *T)
    sumB1  +=       0.039 * cos(3.34206786809  + 26202.34243025941 *T)
    sumB1  +=       0.029 * cos(1.94010215318  + 103821.92160120218 *T)
    sumB1  +=       0.036 * cos(0.65969730203  + 51756.32163927320 *T)
    sumB1  +=       0.030 * cos(0.82864967511  + 46848.33017476560 *T)
    sumB1  +=       0.038 * cos(0.80809536333  + 151975.46535238638 *T)
    sumB1  +=       0.029 * cos(2.89227792530  + 104358.72611329758 *T)
    sumB1  +=       0.029 * cos(4.49728060011  + 103299.34418310841 *T)
    sumB1  +=       0.035 * cos(1.48121158024  + 85502.38501632259 *T)
    sumB1  +=       0.028 * cos(5.48224379273  + 91785.46086631398 *T)
    sumB1  +=       0.030 * cos(3.11318420566  + 106570.36967048360 *T)
    sumB1  +=       0.034 * cos(4.26548581082  + 90829.86112470538 *T)
    sumB1  +=       0.031 * cos(5.67743476614  + 117873.36400788820 *T)
    sumB1  +=       0.028 * cos(4.97598935775  + 103285.11708910680 *T)
    sumB1  +=       0.033 * cos(5.18111128318  + 114564.89811250778 *T)
    sumB1  +=       0.027 * cos(0.88054842818  + 157636.79740153858 *T)
    sumB1  +=       0.026 * cos(2.57652351573  + 134991.46920492980 *T)
    sumB1  +=       0.029 * cos(2.64144900778  + 26137.89976347840 *T)
    sumB1  +=       0.028 * cos(3.23650761988  + 128850.44281258718 *T)
    sumB1  +=       0.030 * cos(6.03255187776  + 32370.97899156560 *T)
    sumB1  +=       0.029 * cos(3.23415692694  + 104344.49901929598 *T)
    sumB1  +=       0.032 * cos(3.21019810469  + 28421.09953444620 *T)
    sumB1  +=       0.024 * cos(2.52484264487  + 1596.18644228460 *T)
    sumB1  +=       0.024 * cos(0.27171495756  + 24491.71669928959 *T)
    sumB1  +=       0.026 * cos(1.20902208843  + 44181.27784112419 *T)
    sumB1  +=       0.027 * cos(5.71639245109  + 111122.32316754240 *T)
    sumB1  +=       0.031 * cos(3.68314656803  + 28206.66700195260 *T)
    sumB1  +=       0.030 * cos(2.36500429653  + 103711.71527998280 *T)
    sumB1  +=       0.022 * cos(4.47547965260  + 77829.99768684579 *T)
    sumB1  +=       0.022 * cos(5.93080138405  + 98068.53671630540 *T)
    sumB1  +=       0.021 * cos(4.45887879737  + 52815.70356946240 *T)
    sumB1  +=       0.022 * cos(1.51276188597  + 35191.81013569180 *T)
    sumB1  +=       0.025 * cos(3.67349275753  + 853.19638175200 *T)
    sumB1  +=       0.021 * cos(5.21045664362  + 104564.91166173480 *T)
    sumB1  +=       0.023 * cos(3.52400873624  + 100909.03762133140 *T)
    sumB1  +=       0.021 * cos(0.91705070968  + 97580.90196505121 *T)
    sumB1  +=       0.021 * cos(3.93681460045  + 25455.11940226100 *T)
    sumB1  +=       0.020 * cos(2.51210823023  + 104347.73123093881 *T)
    sumB1  +=       0.027 * cos(2.35592427305  + 78690.30761559859 *T)
    sumB1  +=       0.020 * cos(5.09828683951  + 52099.54021187280 *T)
    sumB1  +=       0.019 * cos(3.37799607043  + 45494.58142974879 *T)
    sumB1  +=       0.021 * cos(1.47540794459  + 115674.27666460120 *T)
    sumB1  +=       0.020 * cos(1.17802900406  + 155997.72788435058 *T)
    sumB1  +=       0.025 * cos(4.46628547605  + 146314.13330323418 *T)
    sumB1  +=       0.023 * cos(2.29934592297  + 33967.99229491319 *T)
    sumB1  +=       0.020 * cos(0.67974525638  + 24505.94379329119 *T)
    sumB1  +=       0.020 * cos(5.32443861216  + 25234.70675982219 *T)
    sumB1  +=       0.018 * cos(4.83011056299  + 50593.84693486539 *T)
    sumB1  +=       0.020 * cos(2.19515561332  + 51639.00177105299 *T)
    sumB1  +=       0.020 * cos(4.54362105039  + 26037.90651966999 *T)
    sumB1  +=       0.018 * cos(3.66945372389  + 176953.98994186718 *T)
    sumB1  +=       0.022 * cos(2.36457126914  + 143961.26714946239 *T)
    sumB1  +=       0.020 * cos(1.91605933726  + 53242.30176033840 *T)
    sumB1  +=       0.017 * cos(3.65797302958  + 25938.33994443960 *T)
    sumB1  +=       0.017 * cos(4.72356636284  + 39743.76363275060 *T)
    sumB1  +=       0.017 * cos(4.63385029370  + 70269.18098269838 *T)
    sumB1  +=       0.023 * cos(3.47296935942  + 102132.85546211000 *T)
    sumB1  +=       0.018 * cos(3.89075237232  + 77844.22478084739 *T)
    sumB1  +=       0.016 * cos(1.39550514982  + 119116.85160956658 *T)
    sumB1  +=       0.017 * cos(3.19413997784  + 2333.19639287200 *T)
    sumB1  +=       0.018 * cos(0.10505670252  + 52290.24557183361 *T)
    sumB1  +=       0.015 * cos(5.66501892458  + 7994.52844202420 *T)
    sumB1  +=       0.017 * cos(5.59546671313  + 60055.89543648739 *T)
    sumB1  +=       0.015 * cos(3.35297565546  + 58458.88213313979 *T)
    sumB1  +=       0.019 * cos(1.88696722328  + 1581.95934828300 *T)
    sumB1  +=       0.016 * cos(3.51339563149  + 26094.53170047421 *T)
    sumB1  +=       0.015 * cos(5.87796844640  + 105940.68546158058 *T)
    sumB1  +=       0.014 * cos(4.90677373859  + 632.78373931320 *T)
    sumB1  +=       0.017 * cos(2.94915136654  + 18093.37469954999 *T)
    sumB1  +=       0.016 * cos(3.30923893089  + 26507.38778544939 *T)
    sumB1  +=       0.017 * cos(2.79032831551  + 124156.43985787958 *T)
    sumB1  +=       0.017 * cos(2.32026007183  + 53228.07466633679 *T)
    sumB1  +=       0.015 * cos(1.27735280723  + 103925.01437542078 *T)
    sumB1  +=       0.014 * cos(0.47578456087  + 104138.31347085880 *T)
    sumB1  +=       0.014 * cos(4.33486658309  + 18207.81398823521 *T)
    sumB1  +=       0.014 * cos(6.25912081869  + 102232.84870591838 *T)
    sumB1  +=       0.016 * cos(5.19489448771  + 52022.02747266360 *T)
    sumB1  +=       0.015 * cos(3.36079341013  + 26241.68195205900 *T)
    sumB1  +=       0.016 * cos(0.62153007569  + 71980.63357473118 *T)
    sumB1  +=       0.016 * cos(0.34240403974  + 26081.27458267419 *T)
    sumB1  +=       0.015 * cos(4.43868814661  + 207643.84320240439 *T)
    sumB1  +=       0.017 * cos(3.44704905407  + 183724.70054311279 *T)
    sumB1  +=       0.016 * cos(0.08608453345  + 77410.51304297059 *T)
    sumB1  +=       0.013 * cos(4.45469253351  + 90695.75207512038 *T)
    sumB1  +=       0.014 * cos(5.26281612633  + 391318.54712361295 *T)
    sumB1  +=       0.013 * cos(2.28140033187  + 26521.61487945100 *T)
    sumB1  +=       0.013 * cos(4.75564434373  + 536.80451209540 *T)
    sumB1  +=       0.013 * cos(3.72895039092  + 130419.84594697120 *T)
    sumB1  +=       0.015 * cos(6.26388997196  + 54509.00267602040 *T)
    sumB1  +=       0.014 * cos(1.32802909705  + 20760.42703319140 *T)
    sumB1  +=       0.012 * cos(0.96182853080  + 157057.10981453978 *T)
    sumB1  +=       0.017 * cos(3.00168933222  + 207593.84658050016 *T)
    sumB1  +=       0.016 * cos(0.54167157981  + 52061.36699446317 *T)
    sumB1  +=       0.016 * cos(3.76004077423  + 178063.36849396059 *T)
    sumB1  +=       0.012 * cos(0.50356687265  + 5327.47610838280 *T)
    sumB1  +=       0.014 * cos(0.65739903621  + 84546.78527471398 *T)
    sumB1  +=       0.014 * cos(2.64166191335  + 42153.96900304900 *T)
    sumB1  +=       0.011 * cos(1.87375624710  + 97112.93697469680 *T)
    sumB1  +=       0.012 * cos(5.28144640817  + 26610.48055966799 *T)
    sumB1  +=       0.012 * cos(1.02715379165  + 130012.91751699500 *T)
    sumB1  +=       0.011 * cos(4.76973198116  + 157586.80077963436 *T)
    sumB1  +=       0.011 * cos(4.03645617358  + 26011.63707029860 *T)
    sumB1  +=       0.012 * cos(1.03774094920  + 116917.76426627958 *T)
    sumB1  +=       0.011 * cos(2.34762914711  + 214364.55718174577 *T)
    sumB1  +=       0.013 * cos(5.96105286469  + 49527.35145767539 *T)
    sumB1  +=       0.013 * cos(0.38784055589  + 54294.57014352679 *T)
    sumB1  +=       0.011 * cos(0.80279403389  + 130443.39704322898 *T)
    sumB1  +=       0.012 * cos(1.91910219792  + 140652.80125408198 *T)
    sumB1  +=       0.011 * cos(4.82050468552  + 647.01083331480 *T)
    sumB1  +=       0.012 * cos(3.95784286024  + 72936.23331633979 *T)
    sumB1  +=       0.010 * cos(2.38363926963  + 52698.38370124219 *T)
    sumB1  +=       0.012 * cos(5.21962000846  + 25551.09862947879 *T)
    sumB1  +=       0.011 * cos(1.24041943634  + 51543.02254383520 *T)
    sumB1  +=       0.010 * cos(1.23610986711  + 76667.52298243798 *T)
    sumB1  +=       0.011 * cos(5.27116774881  + 155571.81910783658 *T)
    sumB1  +=       0.011 * cos(5.98419309923  + 107794.18751126219 *T)
    sumB1  +=       0.010 * cos(0.55390559777  + 26941.09952332620 *T)
    sumB1  +=       0.010 * cos(1.49161844344  + 52329.58509363319 *T)
    sumB1  +=       0.011 * cos(4.68946178137  + 26190.99591579279 *T)
    sumB1  +=       0.009 * cos(5.63092803389  + 27684.08958385880 *T)
    sumB1  +=       0.011 * cos(6.09445403088  + 26555.86813192860 *T)
    sumB1  +=       0.010 * cos(0.09698975387  + 29416.03879785439 *T)
    sumB1  +=       0.010 * cos(0.72339727974  + 22759.76748529401 *T)
    sumB1  +=       0.009 * cos(1.96066446481  + 51528.79544983359 *T)
    sumB1  +=       0.009 * cos(4.28124553384  + 121335.60871375341 *T)
    sumB1  +=       0.008 * cos(4.26127907796  + 182085.63102592481 *T)
    sumB1  +=       0.009 * cos(3.34775259107  + 22747.29071487440 *T)
    sumB1  +=       0.008 * cos(1.25238578145  + 129387.24732468258 *T)
    sumB1  +=       0.008 * cos(3.86388656969  + 24864.08530079559 *T)
    sumB1  +=       0.010 * cos(1.34737181255  + 2648.45482547300 *T)
    sumB1  +=       0.008 * cos(5.72234642822  + 161079.37234650398 *T)
    sumB1  +=       0.009 * cos(5.45351693096  + 104778.21075717278 *T)
    sumB1  +=       0.011 * cos(2.85637796333  + 52643.77127350280 *T)
    sumB1  +=       0.008 * cos(3.27339488052  + 27311.72098235281 *T)
    sumB1  +=       0.010 * cos(1.31510855953  + 52125.80966124419 *T)
    sumB1  +=       0.009 * cos(1.17621493154  + 172402.03644480839 *T)
    sumB1  +=       0.008 * cos(2.43297081646  + 131395.11544947961 *T)
    sumB1  +=       0.009 * cos(0.46231104626  + 26624.70765366959 *T)
    sumB1  +=       0.007 * cos(4.93485551094  + 48835.19385644859 *T)
    sumB1  +=       0.007 * cos(5.92438953036  + 130446.62925487180 *T)
    sumB1  +=       0.007 * cos(5.15107638670  + 149.56319713460 *T)
    sumB1  +=       0.008 * cos(0.00476652316  + 209812.60368468691 *T)
    sumB1  +=       0.008 * cos(5.06558822410  + 61279.71327726600 *T)
    sumB1  +=       0.009 * cos(0.56618846666  + 522.57741809380 *T)
    sumB1  +=       0.007 * cos(2.63009300708  + 130459.18546877075 *T)
    sumB1  +=       0.008 * cos(2.08959717350  + 25565.32572348040 *T)
    sumB1  +=       0.009 * cos(5.37562306264  + 170049.17029103660 *T)
    sumB1  +=       0.008 * cos(4.05554156200  + 156100.82065856917 *T)
    sumB1  +=       0.007 * cos(5.58541192959  + 129799.61842155698 *T)
    sumB1  +=       0.007 * cos(1.76385898252  + 129373.02023068098 *T)
    sumB1  +=       0.006 * cos(4.86066957244  + 7880.08915333899 *T)
    sumB1  +=       0.008 * cos(4.47849277972  + 10021.83728009940 *T)
    sumB1  +=       0.006 * cos(5.30561035943  + 77726.90491262719 *T)
    sumB1  +=       0.006 * cos(1.27631388714  + 103917.90082841998 *T)
    sumB1  +=       0.006 * cos(4.61189217781  + 157483.01859105378 *T)
    sumB1  +=       0.006 * cos(3.96170038596  + 123668.80510662538 *T)
    sumB1  +=       0.006 * cos(1.33523648266  + 78903.60671103658 *T)
    sumB1  +=       0.006 * cos(0.55702386570  + 204151.27163553479 *T)
    sumB1  +=       0.005 * cos(1.55562437906  + 26294.08869001139 *T)
    sumB1  +=       0.006 * cos(2.05839183005  + 130652.81480330898 *T)
    sumB1  +=       0.005 * cos(5.58094407959  + 130435.63437251298 *T)
    sumB1  +=       0.005 * cos(1.47580092631  + 78114.14622758799 *T)
    sumB1  +=       0.006 * cos(4.10195998301  + 51707.84129279399 *T)
    sumB1  +=       0.005 * cos(5.42356096080  + 3340.61242669980 *T)
    sumB1  +=       0.005 * cos(0.71252148739  + 103932.12792242158 *T)
    sumB1  +=       0.005 * cos(3.31021043385  + 111590.28815789679 *T)
    sumB1  +=       0.005 * cos(5.18242316734  + 13362.44970679920 *T)
    sumB1  +=       0.007 * cos(0.02670007799  + 130432.40216087016 *T)
    sumB1  +=       0.006 * cos(5.64961767520  + 52225.80290505260 *T)
    sumB1  +=       0.006 * cos(0.24412111381  + 128220.75860368418 *T)
    sumB1  +=       0.005 * cos(4.45623339313  + 145204.75475114080 *T)
    sumB1  +=       0.006 * cos(5.59932602145  + 154938.34595416137 *T)
    sumB1  +=       0.005 * cos(0.44843767302  + 203041.89308344136 *T)
    sumB1  +=       0.006 * cos(0.61515143577  + 126996.94076290558 *T)
    sumB1  +=       0.006 * cos(3.76267562514  + 110634.68841628821 *T)
    sumB1  +=       0.005 * cos(3.87958020217  + 25440.89230825939 *T)
    sumB1  +=       0.006 * cos(6.12898547290  + 132658.27281205778 *T)
    sumB1  +=       0.006 * cos(1.78589858446  + 76681.75007643958 *T)
    sumB1  +=       0.006 * cos(3.31942003464  + 25881.71759313700 *T)
    sumB1  +=       0.005 * cos(4.83610476376  + 78417.48823520739 *T)
    sumB1  +=       0.005 * cos(1.62128192664  + 65831.66677432480 *T)
    sumB1  +=       0.006 * cos(5.79832326006  + 150244.34299945380 *T)
    sumB1  +=       0.004 * cos(4.90210493212  + 123200.84011627098 *T)
    sumB1  +=       0.004 * cos(0.04822157393  + 71582.48457132299 *T)
    sumB1  +=       0.004 * cos(1.25993980368  + 233731.74634397856 *T)
    sumB1  +=       0.004 * cos(4.09490053686  + 143005.66740785379 *T)
    sumB1  +=       0.004 * cos(0.60263195052  + 6681.22485339960 *T)
    sumB1  +=       0.005 * cos(3.53504145685  + 130226.21661243298 *T)
    sumB1  +=       0.004 * cos(2.52659732174  + 132028.58860315481 *T)
    sumB1  +=       0.004 * cos(5.09490146272  + 76.26607127560 *T)
    sumB1  +=       0.004 * cos(1.41880070183  + 31415.37924995700 *T)
    sumB1  +=       0.005 * cos(5.67226040777  + 68241.87214462319 *T)
    sumB1  +=       0.006 * cos(6.07125845396  + 1478.86657406440 *T)
    sumB1  +=       0.004 * cos(2.90387262796  + 128320.75184749260 *T)
    sumB1  +=       0.004 * cos(3.16756186361  + 78378.14871340781 *T)
    sumB1  +=       0.004 * cos(0.32589274840  + 52595.29092702359 *T)
    sumB1  +=       0.005 * cos(1.78096081490  + 116783.65521669458 *T)
    sumB1  +=       0.004 * cos(4.99518857973  + 166740.70439565618 *T)
    sumB1  +=       0.005 * cos(4.90494970331  + 79330.20490191258 *T)
    sumB1  +=       0.004 * cos(5.23437839920  + 206.18554843720 *T)
    sumB1  +=       0.004 * cos(2.64302747396  + 75615.25459924959 *T)
    sumB1  +=       0.004 * cos(0.81407144130  + 75930.51303185058 *T)
    sumB1  +=       0.004 * cos(0.53239895977  + 156507.74908854539 *T)
    sumB1  +=       0.004 * cos(4.27504704332  + 956.28915597060 *T)
    sumB1  +=       0.004 * cos(3.01307667234  + 80596.90581759460 *T)
    sumB1  +=       0.003 * cos(4.15170143655  + 25867.49049913539 *T)
    sumB1  +=       0.003 * cos(4.02158764895  + 52712.61079524379 *T)
    sumB1  +=       0.004 * cos(5.97032773544  + 78731.67441507700 *T)
    sumB1  +=       0.004 * cos(5.13753166239  + 137678.19129947099 *T)
    sumB1  +=       0.003 * cos(5.38378225170  + 79315.97780791098 *T)
    sumB1  +=       0.003 * cos(4.90402266218  + 77630.92568540938 *T)
    sumB1  +=       0.004 * cos(3.35681065144  + 80382.47328510099 *T)
    sumB1  +=       0.004 * cos(3.34724303278  + 55503.94193942859 *T)
    sumB1  +=       0.003 * cos(4.39477038963  + 102755.42612401220 *T)
    sumB1  +=       0.004 * cos(2.19880101386  + 137210.22630911658 *T)
    sumB1  +=       0.003 * cos(1.92191274920  + 87367.61641884019 *T)
    sumB1  +=       0.004 * cos(3.62687237484  + 78149.27013603736 *T)
    sumB1  +=       0.003 * cos(5.34278285975  + 77616.69859140778 *T)
    sumB1  +=       0.003 * cos(4.04927793779  + 183145.01295611399 *T)
    sumB1  +=       0.003 * cos(2.33740424416  + 86143.79857806159 *T)
    sumB1  +=       0.003 * cos(0.29051326311  + 66653.15746634839 *T)
    sumB1  +=       0.004 * cos(3.19530187424  + 103498.41618454478 *T)
    sumB1  +=       0.003 * cos(1.83670250940  + 78187.44335344699 *T)
    sumB1  +=       0.004 * cos(6.08402133763  + 233681.74972207440 *T)
    sumB1  +=       0.003 * cos(5.44799723236  + 52609.51802102519 *T)
    sumB1  +=       0.003 * cos(3.88568044923  + 156531.30018480320 *T)
    sumB1  +=       0.003 * cos(1.56467339056  + 183674.70392120857 *T)
    sumB1  +=       0.003 * cos(2.37223694472  + 11610.55195837420 *T)
    sumB1  +=       0.004 * cos(2.36349454401  + 3328.13565628019 *T)
    sumB1  +=       0.003 * cos(2.21015804942  + 130866.11389874700 *T)
    sumB1  +=       0.003 * cos(1.00111385284  + 12546.48193908300 *T)
    sumB1  +=       0.003 * cos(5.04736992413  + 167850.08294774959 *T)
    sumB1  +=       0.004 * cos(2.74224804458  + 133882.09065283637 *T)
    sumB1  +=       0.003 * cos(6.04124419781  + 16066.06586147480 *T)
    sumB1  +=       0.003 * cos(2.00390511344  + 25984.81036735560 *T)
    sumB1  +=       0.003 * cos(3.58451231520  + 19406.67828817460 *T)
    sumB1  +=       0.003 * cos(5.58865591583  + 52381.99183158559 *T)
    sumB1  +=       0.003 * cos(5.49839860031  + 78786.28684281638 *T)
    sumB1  +=       0.003 * cos(2.35251877810  + 50483.64061364600 *T)
    sumB1  +=       0.003 * cos(2.23638545873  + 417406.45026518719 *T)
    sumB1  +=       0.003 * cos(0.89820153238  + 182188.72380014337 *T)
    sumB1  +=       0.003 * cos(2.18659555288  + 196137.07343261081 *T)
    sumB1  +=       0.003 * cos(1.23213629886  + 183570.92173262799 *T)
    sumB1  +=       0.003 * cos(0.65362427623  + 1162.47470440780 *T)
    sumB1  +=       0.003 * cos(2.37265737062  + 141762.17980617538 *T)
    sumB1  +=       0.003 * cos(2.37167138660  + 103.09277421860 *T)
    sumB1  +=       0.003 * cos(3.43071585489  + 51219.51712717779 *T)
    sumB1  +=       0.002 * cos(5.43620933898  + 240452.46032332000 *T)
    sumB1  +=       0.002 * cos(4.79210725186  + 64901.25971792339 *T)
    sumB1  +=       0.003 * cos(0.58543714627  + 136722.59155786238 *T)
    sumB1  +=       0.002 * cos(3.65273296014  + 53029.00266490040 *T)
    sumB1  +=       0.003 * cos(1.40305085463  + 52278.89905736699 *T)
    sumB1  +=       0.003 * cos(2.91888451759  + 235900.50682626120 *T)
    sumB1  +=       0.002 * cos(5.24862580724  + 32769.12799497380 *T)
    sumB1  +=       0.003 * cos(0.92825301327  + 29428.51556827400 *T)
    sumB1  +=       0.002 * cos(1.28026650043  + 96357.08412427260 *T)
    sumB1  +=       0.003 * cos(3.47745075021  + 49842.60989027639 *T)
    sumB1  +=       0.003 * cos(1.27522510736  + 44295.71712980940 *T)
    sumB1  +=       0.003 * cos(2.06356044829  + 181659.72224941081 *T)
    sumB1  +=       0.002 * cos(1.82577636633  + 181026.24909573561 *T)
    sumB1  +=       0.002 * cos(3.73653346143  + 48847.67062686820 *T)
    sumB1  +=       0.002 * cos(4.29258038531  + 198489.93958638256 *T)
    sumB1  +=       0.003 * cos(1.30423082531  + 26237.46633870879 *T)
    sumB1  +=       0.002 * cos(5.62405074860  + 112231.70171963578 *T)
    sumB1  +=       0.003 * cos(4.34969907580  + 78213.71280281838 *T)
    sumB1  +=       0.002 * cos(4.40950391344  + 155475.15046625675 *T)
    sumB1  +=       0.002 * cos(5.10859116787  + 27669.86248985719 *T)
    sumB1  +=       0.002 * cos(2.40342983563  + 53771.99272543299 *T)
    sumB1  +=       0.002 * cos(1.76898437388  + 104505.39137678158 *T)
    sumB1  +=       0.002 * cos(1.13321786137  + 208173.53416749896 *T)
    sumB1  +=       0.003 * cos(4.95012664609  + 102769.65321801380 *T)
    sumB1  +=       0.002 * cos(2.37593594042  + 51951.46148744649 *T)
    sumB1  +=       0.003 * cos(0.65556179120  + 1692.16566950240 *T)
    sumB1  +=       0.002 * cos(1.65727778641  + 163766.09444104519 *T)

    # MERCURY Latitude
    sumB2  = 0
    sumB2  +=   11830.934 * cos(4.79065585784  + 26087.90314157420 *T)
    sumB2  +=    1913.516 * cos(0.00000000000  + 0.00000000000 *T)
    sumB2  +=    1044.801 * cos(1.21216540536  + 52175.80628314840 *T)
    sumB2  +=     266.213 * cos(4.43418336532  + 78263.70942472259 *T)
    sumB2  +=     170.280 * cos(1.62255638714  + 104351.61256629678 *T)
    sumB2  +=      96.300 * cos(4.80023692017  + 130439.51570787100 *T)
    sumB2  +=      44.692 * cos(1.60758267772  + 156527.41884944518 *T)
    sumB2  +=      18.316 * cos(4.66904655377  + 182615.32199101936 *T)
    sumB2  +=       6.927 * cos(1.43404888930  + 208703.22513259359 *T)
    sumB2  +=       2.479 * cos(4.47495202955  + 234791.12827416777 *T)
    sumB2  +=       1.739 * cos(1.83080039600  + 27197.28169366760 *T)
    sumB2  +=       0.852 * cos(1.22749255198  + 260879.03141574195 *T)
    sumB2  +=       0.641 * cos(4.87358642253  + 53285.18483524180 *T)
    sumB2  +=       0.301 * cos(1.96098924430  + 51066.42773105500 *T)
    sumB2  +=       0.306 * cos(5.03912693671  + 24978.52458948080 *T)
    sumB2  +=       0.284 * cos(4.25874901943  + 286966.93455731618 *T)
    sumB2  +=       0.235 * cos(3.14549432371  + 51116.42435295920 *T)
    sumB2  +=       0.232 * cos(4.22847849119  + 31749.23519072640 *T)
    sumB2  +=       0.186 * cos(5.05207772442  + 77154.33087262919 *T)
    sumB2  +=       0.143 * cos(5.71074961492  + 21535.94964451540 *T)
    sumB2  +=       0.114 * cos(5.62232460770  + 27043.50288318280 *T)
    sumB2  +=       0.123 * cos(1.51066137419  + 79373.08797681599 *T)
    sumB2  +=       0.093 * cos(1.83776226489  + 103242.23401420339 *T)
    sumB2  +=       0.081 * cos(0.85802178989  + 57837.13833230060 *T)
    sumB2  +=       0.079 * cos(5.50027152752  + 47623.85278608960 *T)
    sumB2  +=       0.093 * cos(1.01040170383  + 313054.83769889042 *T)
    sumB2  +=       0.058 * cos(2.35944147105  + 53131.40602475700 *T)
    sumB2  +=       0.074 * cos(2.28645155343  + 50586.73338786459 *T)
    sumB2  +=       0.049 * cos(1.78193753560  + 28306.66024576099 *T)
    sumB2  +=       0.058 * cos(5.88387132512  + 77204.32749453338 *T)
    sumB2  +=       0.060 * cos(0.54679388547  + 37410.56723987860 *T)
    sumB2  +=       0.046 * cos(5.77182292982  + 529.69096509460 *T)
    sumB2  +=       0.052 * cos(3.58488737294  + 25558.21217647960 *T)
    sumB2  +=       0.061 * cos(1.64891174657  + 20426.57109242200 *T)
    sumB2  +=       0.043 * cos(6.15538236294  + 51646.11531805379 *T)
    sumB2  +=       0.042 * cos(4.89875940528  + 129330.13715577760 *T)
    sumB2  +=       0.036 * cos(5.61094806248  + 22645.32819660879 *T)
    sumB2  +=       0.033 * cos(4.34189175312  + 32858.61374281979 *T)
    sumB2  +=       0.037 * cos(4.50512068847  + 26107.57290247399 *T)
    sumB2  +=       0.032 * cos(1.99466664427  + 25132.30339996560 *T)
    sumB2  +=       0.030 * cos(0.76607738630  + 5661.33204915220 *T)
    sumB2  +=       0.027 * cos(4.87180838160  + 54394.56338733519 *T)
    sumB2  +=       0.029 * cos(5.66168505554  + 26068.23338067440 *T)
    sumB2  +=       0.025 * cos(3.31115218239  + 1059.38193018920 *T)
    sumB2  +=       0.029 * cos(2.70280425915  + 41962.52073693740 *T)
    sumB2  +=       0.029 * cos(3.56940606994  + 63498.47038145279 *T)
    sumB2  +=       0.029 * cos(4.01947170286  + 339142.74084046454 *T)
    sumB2  +=       0.028 * cos(0.80200569507  + 39609.65458316560 *T)
    sumB2  +=       0.023 * cos(5.30820497885  + 23869.14603738740 *T)
    sumB2  +=       0.024 * cos(4.86369445330  + 46514.47423399620 *T)
    sumB2  +=       0.027 * cos(3.32081057071  + 10213.28554621100 *T)
    sumB2  +=       0.030 * cos(1.15631768661  + 26084.02180621620 *T)
    sumB2  +=       0.024 * cos(2.71284750937  + 26091.78447693220 *T)
    sumB2  +=       0.021 * cos(5.85119251529  + 25874.60404613620 *T)
    sumB2  +=       0.026 * cos(3.11971794265  + 43071.89928903080 *T)
    sumB2  +=       0.023 * cos(1.14258266850  + 58946.51688439399 *T)
    sumB2  +=       0.021 * cos(4.78965226125  + 51220.20654153979 *T)
    sumB2  +=       0.025 * cos(5.17153436907  + 76674.63652943878 *T)
    sumB2  +=       0.020 * cos(6.05659966832  + 25028.52121138500 *T)
    sumB2  +=       0.017 * cos(4.06375348164  + 51535.90899683439 *T)
    sumB2  +=       0.019 * cos(2.26589689692  + 72602.37737557039 *T)
    sumB2  +=       0.019 * cos(2.02802084878  + 73711.75592766379 *T)
    sumB2  +=       0.020 * cos(2.44092663198  + 51749.20809227239 *T)
    sumB2  +=       0.017 * cos(1.67702029587  + 155418.04029735178 *T)
    sumB2  +=       0.018 * cos(2.11191340126  + 26617.59410666880 *T)
    sumB2  +=       0.019 * cos(0.32685450395  + 4551.95349705880 *T)
    sumB2  +=       0.016 * cos(4.02105643770  + 105460.99111839020 *T)
    sumB2  +=       0.014 * cos(2.35860855509  + 26080.78959457339 *T)
    sumB2  +=       0.013 * cos(3.28426450224  + 25661.30495069820 *T)
    sumB2  +=       0.016 * cos(2.09223320909  + 103292.23063610760 *T)
    sumB2  +=       0.013 * cos(2.08475100066  + 49957.04917896160 *T)
    sumB2  +=       0.014 * cos(5.66400816218  + 48733.23133818299 *T)
    sumB2  +=       0.013 * cos(1.65023710844  + 26095.01668857500 *T)
    sumB2  +=       0.015 * cos(5.43792291298  + 25973.46385288896 *T)
    sumB2  +=       0.012 * cos(5.15541566845  + 79219.30916633119 *T)
    sumB2  +=       0.012 * cos(2.93766606728  + 33326.57873317420 *T)
    sumB2  +=       0.011 * cos(5.16599347976  + 15874.61759536320 *T)
    sumB2  +=       0.015 * cos(3.49521469034  + 83925.04147387479 *T)
    sumB2  +=       0.012 * cos(1.33980027148  + 77308.10968311399 *T)
    sumB2  +=       0.012 * cos(2.74476655782  + 19317.19254032860 *T)
    sumB2  +=       0.014 * cos(5.48457159802  + 98690.28051714458 *T)
    sumB2  +=       0.012 * cos(6.17629016531  + 36301.18868778519 *T)
    sumB2  +=       0.012 * cos(5.16549673414  + 1589.07289528380 *T)
    sumB2  +=       0.010 * cos(3.19718718931  + 24498.83024629040 *T)
    sumB2  +=       0.010 * cos(3.10957426655  + 16983.99614745660 *T)
    sumB2  +=       0.010 * cos(2.67818215946  + 77734.01845962799 *T)
    sumB2  +=       0.012 * cos(1.64438976704  + 26202.34243025941 *T)
    sumB2  +=       0.011 * cos(6.12421691815  + 52705.49724824299 *T)
    sumB2  +=       0.010 * cos(5.42117419251  + 14765.23904326980 *T)
    sumB2  +=       0.011 * cos(1.59577581724  + 50057.04242277000 *T)
    sumB2  +=       0.012 * cos(2.44342322463  + 213.29909543800 *T)
    sumB2  +=       0.011 * cos(4.59333229552  + 56727.75978020720 *T)
    sumB2  +=       0.008 * cos(2.22267415318  + 6283.07584999140 *T)
    sumB2  +=       0.010 * cos(3.63881306668  + 65697.55772473979 *T)
    sumB2  +=       0.008 * cos(0.85708473896  + 365230.64398203877 *T)
    sumB2  +=       0.010 * cos(3.56346926379  + 53235.18821333759 *T)
    sumB2  +=       0.009 * cos(3.03935867498  + 78793.40038981718 *T)
    sumB2  +=       0.008 * cos(2.05186284069  + 30639.85663863300 *T)
    sumB2  +=       0.008 * cos(6.07959531374  + 59414.48187474840 *T)
    sumB2  +=       0.007 * cos(1.48753095657  + 29530.47808653960 *T)
    sumB2  +=       0.007 * cos(0.54370332724  + 38519.94579197200 *T)
    sumB2  +=       0.008 * cos(5.99792913268  + 7.11354700080 *T)
    sumB2  +=       0.009 * cos(6.08426293157  + 69159.80243060499 *T)
    sumB2  +=       0.007 * cos(1.35513014510  + 82815.66292178139 *T)
    sumB2  +=       0.008 * cos(3.00335450338  + 40853.14218484400 *T)
    sumB2  +=       0.007 * cos(3.40544433462  + 1066.49547719000 *T)
    sumB2  +=       0.007 * cos(5.80782632115  + 45405.09568190280 *T)
    sumB2  +=       0.007 * cos(3.82318715536  + 1052.26838318840 *T)
    sumB2  +=       0.006 * cos(4.57200111034  + 55618.38122811380 *T)
    sumB2  +=       0.006 * cos(4.85942618671  + 52168.69273614759 *T)
    sumB2  +=       0.008 * cos(0.34211699258  + 79323.09135491178 *T)
    sumB2  +=       0.006 * cos(0.67823675060  + 110012.94461544900 *T)
    sumB2  +=       0.007 * cos(4.12285655199  + 45892.73043315699 *T)
    sumB2  +=       0.007 * cos(5.41918774642  + 1109.37855209340 *T)
    sumB2  +=       0.007 * cos(3.14407932401  + 51962.50718771040 *T)
    sumB2  +=       0.006 * cos(5.13879522396  + 76044.95232053580 *T)
    sumB2  +=       0.007 * cos(2.06447590782  + 124778.18365871880 *T)
    sumB2  +=       0.006 * cos(1.61443478349  + 3442.57494496540 *T)
    sumB2  +=       0.005 * cos(2.01359728100  + 27676.97603685800 *T)
    sumB2  +=       0.007 * cos(5.49596353318  + 23754.70674870219 *T)
    sumB2  +=       0.006 * cos(0.71445272050  + 27147.28507176339 *T)
    sumB2  +=       0.005 * cos(5.58153663720  + 68050.42387851159 *T)
    sumB2  +=       0.006 * cos(4.72387999717  + 52290.24557183361 *T)
    sumB2  +=       0.005 * cos(1.65565332809  + 80482.46652890938 *T)
    sumB2  +=       0.005 * cos(4.67655631038  + 181505.94343892601 *T)
    sumB2  +=       0.005 * cos(5.35730277715  + 12566.15169998280 *T)
    sumB2  +=       0.005 * cos(2.24953488234  + 52061.36699446317 *T)
    sumB2  +=       0.004 * cos(1.93085125724  + 52156.13652224860 *T)
    sumB2  +=       0.006 * cos(5.23350745087  + 129380.13377768178 *T)
    sumB2  +=       0.005 * cos(6.27908880897  + 66941.04532641819 *T)
    sumB2  +=       0.006 * cos(0.86124028508  + 52195.47604404819 *T)
    sumB2  +=       0.004 * cos(0.23100295125  + 89586.37352302698 *T)
    sumB2  +=       0.005 * cos(4.23944686799  + 103396.01282468818 *T)
    sumB2  +=       0.004 * cos(0.29398633221  + 51109.31080595839 *T)
    sumB2  +=       0.005 * cos(0.38451698643  + 955.59974160860 *T)
    sumB2  +=       0.005 * cos(5.75812965170  + 104881.30353139140 *T)
    sumB2  +=       0.005 * cos(3.78671774559  + 52171.92494779040 *T)
    sumB2  +=       0.004 * cos(5.32828531774  + 25448.00585526019 *T)
    sumB2  +=       0.005 * cos(3.20016226905  + 426.59819087600 *T)
    sumB2  +=       0.005 * cos(3.86489045633  + 9103.90699411760 *T)
    sumB2  +=       0.004 * cos(0.62543334322  + 11322.66409830440 *T)
    sumB2  +=       0.004 * cos(6.07508617692  + 2118.76386037840 *T)
    sumB2  +=       0.004 * cos(2.43638238426  + 74821.13447975718 *T)
    sumB2  +=       0.004 * cos(5.29735530203  + 52179.68761850640 *T)
    sumB2  +=       0.003 * cos(3.19112063073  + 13521.75144159140 *T)
    sumB2  +=       0.004 * cos(5.09850398144  + 77837.11123384659 *T)
    sumB2  +=       0.004 * cos(1.81128771043  + 102762.53967101300 *T)
    sumB2  +=       0.004 * cos(2.92685125689  + 639.89728631400 *T)
    sumB2  +=       0.004 * cos(4.37998582536  + 52182.91983014920 *T)
    sumB2  +=       0.004 * cos(0.90562468300  + 77623.81213840858 *T)
    sumB2  +=       0.004 * cos(3.40653389629  + 105410.99449648600 *T)
    sumB2  +=       0.004 * cos(5.56139758889  + 26301.20223701220 *T)
    sumB2  +=       0.003 * cos(3.84817111044  + 2218.75710418680 *T)
    sumB2  +=       0.004 * cos(3.40685398189  + 52389.10537858640 *T)
    sumB2  +=       0.003 * cos(3.64497587259  + 64607.84893354619 *T)
    sumB2  +=       0.003 * cos(4.90907522926  + 99799.65906923798 *T)
    sumB2  +=       0.003 * cos(0.33551297325  + 13655.86049117640 *T)
    sumB2  +=       0.003 * cos(4.38738526963  + 108903.56606335558 *T)
    sumB2  +=       0.003 * cos(2.71996542810  + 38654.05484155699 *T)
    sumB2  +=       0.003 * cos(5.21639569130  + 150866.08680029298 *T)
    sumB2  +=       0.003 * cos(5.79669813464  + 131548.89425996438 *T)
    sumB2  +=       0.003 * cos(4.24481036893  + 136100.84775702321 *T)
    sumB2  +=       0.003 * cos(4.74285528385  + 53764.87917843220 *T)
    sumB2  +=       0.003 * cos(5.12230469501  + 32370.97899156560 *T)
    sumB2  +=       0.003 * cos(4.91979299090  + 18849.22754997420 *T)
    sumB2  +=       0.003 * cos(1.21305125502  + 105307.21230790540 *T)
    sumB2  +=       0.002 * cos(2.61741407230  + 71492.99882347698 *T)
    sumB2  +=       0.002 * cos(5.29927753161  + 64741.95798313119 *T)
    sumB2  +=       0.002 * cos(2.59250638038  + 155468.03691925600 *T)
    sumB2  +=       0.002 * cos(5.60083003625  + 26514.50133245020 *T)
    sumB2  +=       0.003 * cos(4.14284031224  + 85034.42002596818 *T)
    sumB2  +=       0.002 * cos(3.04958336441  + 93028.94846799241 *T)
    sumB2  +=       0.003 * cos(6.16142033746  + 78477.00852016058 *T)
    sumB2  +=       0.003 * cos(4.49779099000  + 76144.94556434419 *T)
    sumB2  +=       0.002 * cos(2.74114399659  + 130969.20667296558 *T)
    sumB2  +=       0.002 * cos(6.18606575651  + 51123.53789995999 *T)
    sumB2  +=       0.003 * cos(1.68659744145  + 28421.09953444620 *T)
    sumB2  +=       0.002 * cos(1.39891525246  + 81706.28436968799 *T)
    sumB2  +=       0.002 * cos(1.36487993355  + 79852.78232000639 *T)
    sumB2  +=       0.002 * cos(0.86961235954  + 129483.91596626240 *T)
    sumB2  +=       0.002 * cos(0.16105302705  + 25021.40766438420 *T)
    sumB2  +=       0.002 * cos(0.20969999844  + 19804.82729158280 *T)
    sumB2  +=       0.002 * cos(0.99320491085  + 162188.75089859738 *T)
    sumB2  +=       0.002 * cos(3.94039306142  + 391318.54712361295 *T)
    sumB2  +=       0.002 * cos(1.80783236802  + 23969.13928119580 *T)
    sumB2  +=       0.002 * cos(5.12099356220  + 51322.60990139639 *T)
    sumB2  +=       0.002 * cos(2.80140642811  + 62389.09182935939 *T)
    sumB2  +=       0.002 * cos(2.60965333457  + 52602.40447402440 *T)
    sumB2  +=       0.002 * cos(2.72257468038  + 7994.52844202420 *T)
    sumB2  +=       0.002 * cos(5.84222491614  + 103821.92160120218 *T)
    sumB2  +=       0.002 * cos(6.04515878720  + 25035.63475838580 *T)
    sumB2  +=       0.001 * cos(2.34997564292  + 85502.38501632259 *T)
    sumB2  +=       0.001 * cos(0.17409967102  + 131498.89763806018 *T)
    sumB2  +=       0.001 * cos(3.11496551907  + 25934.12433108940 *T)
    sumB2  +=       0.002 * cos(5.12532278095  + 78267.59076008058 *T)
    sumB2  +=       0.001 * cos(5.49175475674  + 104331.94280539700 *T)
    sumB2  +=       0.001 * cos(0.44808516059  + 78050.41032928458 *T)
    sumB2  +=       0.002 * cos(1.92656028088  + 102132.85546211000 *T)
    sumB2  +=       0.002 * cos(1.93198050022  + 853.19638175200 *T)
    sumB2  +=       0.001 * cos(2.42972717822  + 94138.32702008578 *T)
    sumB2  +=       0.002 * cos(5.13550087739  + 6770.71060124560 *T)
    sumB2  +=       0.001 * cos(4.65441500870  + 78259.82808936459 *T)
    sumB2  +=       0.001 * cos(3.99687455490  + 188276.65404017159 *T)
    sumB2  +=       0.001 * cos(4.39069692330  + 104371.28232719658 *T)
    sumB2  +=       0.001 * cos(1.64083661387  + 78283.37918562238 *T)
    sumB2  +=       0.001 * cos(1.50558734911  + 78378.14871340781 *T)
    sumB2  +=       0.002 * cos(6.27597485867  + 91785.46086631398 *T)
    sumB2  +=       0.001 * cos(4.00786092689  + 78244.03966382280 *T)
    sumB2  +=       0.001 * cos(1.70741856509  + 28206.66700195260 *T)
    sumB2  +=       0.001 * cos(5.31900403501  + 78149.27013603736 *T)
    sumB2  +=       0.001 * cos(1.40205887414  + 44937.13069154840 *T)
    sumB2  +=       0.001 * cos(1.03752853909  + 104347.73123093881 *T)
    sumB2  +=       0.001 * cos(0.66520680194  + 419.48464387520 *T)
    sumB2  +=       0.001 * cos(2.46437105095  + 104355.49390165480 *T)
    sumB2  +=       0.001 * cos(0.81835364761  + 35191.81013569180 *T)
    sumB2  +=       0.001 * cos(0.33780670840  + 88476.99497093359 *T)
    sumB2  +=       0.001 * cos(6.08346231930  + 70269.18098269838 *T)
    sumB2  +=       0.001 * cos(1.86810476133  + 125887.56221081220 *T)
    sumB2  +=       0.001 * cos(2.10468974660  + 25619.93815121980 *T)
    sumB2  +=       0.001 * cos(3.99494842792  + 71025.03383312259 *T)
    sumB2  +=       0.001 * cos(0.57714874874  + 71980.63357473118 *T)
    sumB2  +=       0.001 * cos(5.66312942513  + 97580.90196505121 *T)
    sumB2  +=       0.001 * cos(3.06688416266  + 129909.82474277640 *T)
    sumB2  +=       0.001 * cos(5.66709359706  + 433.71173787680 *T)
    sumB2  +=       0.001 * cos(0.91634007063  + 103925.01437542078 *T)
    sumB2  +=       0.001 * cos(6.02359186416  + 181555.94006083015 *T)
    sumB2  +=       0.001 * cos(1.44980129997  + 26727.80042788820 *T)
    sumB2  +=       0.001 * cos(1.45442920390  + 207593.84658050016 *T)
    sumB2  +=       0.001 * cos(1.12941846578  + 60055.89543648739 *T)
    sumB2  +=       0.001 * cos(1.16964771504  + 134991.46920492980 *T)

    # MERCURY Latitude
    sumB3  = 0
    sumB3  +=     235.423 * cos(0.35387524604  + 26087.90314157420 *T)
    sumB3  +=     160.537 * cos(0.00000000000  + 0.00000000000 *T)
    sumB3  +=      18.904 * cos(4.36275460261  + 52175.80628314840 *T)
    sumB3  +=       6.376 * cos(2.50715381439  + 78263.70942472259 *T)
    sumB3  +=       4.580 * cos(6.14257817571  + 104351.61256629678 *T)
    sumB3  +=       3.061 * cos(3.12497552681  + 130439.51570787100 *T)
    sumB3  +=       1.732 * cos(6.26642412058  + 156527.41884944518 *T)
    sumB3  +=       0.857 * cos(3.07673166705  + 182615.32199101936 *T)
    sumB3  +=       0.384 * cos(6.14815319932  + 208703.22513259359 *T)
    sumB3  +=       0.159 * cos(2.92437378320  + 234791.12827416777 *T)
    sumB3  +=       0.062 * cos(5.97292432097  + 260879.03141574195 *T)
    sumB3  +=       0.054 * cos(3.31612529961  + 27197.28169366760 *T)
    sumB3  +=       0.023 * cos(2.74287679452  + 286966.93455731618 *T)
    sumB3  +=       0.017 * cos(0.77798463435  + 24978.52458948080 *T)
    sumB3  +=       0.012 * cos(6.24116133415  + 53285.18483524180 *T)
    sumB3  +=       0.008 * cos(5.79551081392  + 313054.83769889042 *T)
    sumB3  +=       0.006 * cos(5.81555175073  + 31749.23519072640 *T)
    sumB3  +=       0.004 * cos(4.67212195693  + 51116.42435295920 *T)
    sumB3  +=       0.004 * cos(3.36833718999  + 77154.33087262919 *T)
    sumB3  +=       0.003 * cos(0.26522113546  + 103242.23401420339 *T)
    sumB3  +=       0.003 * cos(0.79794705473  + 27043.50288318280 *T)
    sumB3  +=       0.004 * cos(0.82832725003  + 21535.94964451540 *T)
    sumB3  +=       0.002 * cos(2.57068697582  + 339142.74084046454 *T)
    sumB3  +=       0.002 * cos(1.82309397040  + 1109.37855209340 *T)
    sumB3  +=       0.002 * cos(0.74986942696  + 47623.85278608960 *T)
    sumB3  +=       0.002 * cos(3.61692083154  + 50586.73338786459 *T)
    sumB3  +=       0.002 * cos(4.70766241226  + 51066.42773105500 *T)
    sumB3  +=       0.001 * cos(3.33688549090  + 129330.13715577760 *T)
    sumB3  +=       0.001 * cos(2.64641192612  + 57837.13833230060 *T)
    sumB3  +=       0.001 * cos(4.93247555987  + 25558.21217647960 *T)
    sumB3  +=       0.002 * cos(5.00344393497  + 20426.57109242200 *T)
    sumB3  +=       0.002 * cos(4.80841415889  + 25028.52121138500 *T)
    sumB3  +=       0.002 * cos(5.85039338716  + 1059.38193018920 *T)
    sumB3  +=       0.001 * cos(1.85116662745  + 37410.56723987860 *T)
    sumB3  +=       0.002 * cos(3.97317842825  + 25132.30339996560 *T)
    sumB3  +=       0.001 * cos(2.14494291300  + 79373.08797681599 *T)
    sumB3  +=       0.001 * cos(2.53223701416  + 46514.47423399620 *T)
    sumB3  +=       0.001 * cos(5.84921632430  + 5661.33204915220 *T)
    sumB3  +=       0.001 * cos(3.81958470276  + 53131.40602475700 *T)
    sumB3  +=       0.001 * cos(1.05410669937  + 1589.07289528380 *T)
    sumB3  +=       0.001 * cos(1.39436798574  + 51646.11531805379 *T)
    sumB3  +=       0.001 * cos(0.36358097174  + 22645.32819660879 *T)
    sumB3  +=       0.001 * cos(2.76609972270  + 28306.66024576099 *T)
    sumB3  +=       0.001 * cos(3.67689055309  + 4551.95349705880 *T)
    sumB3  +=       0.001 * cos(5.39398540975  + 32858.61374281979 *T)
    sumB3  +=       0.001 * cos(3.73579127170  + 24498.83024629040 *T)
    sumB3  +=       0.001 * cos(0.60289607514  + 26068.23338067440 *T)
    sumB3  +=       0.001 * cos(5.78167324673  + 26107.57290247399 *T)
    sumB3  +=       0.001 * cos(3.81913361795  + 26091.78447693220 *T)
    sumB3  +=       0.001 * cos(5.64455351094  + 365230.64398203877 *T)
    sumB3  +=       0.001 * cos(4.87981157471  + 10213.28554621100 *T)
    sumB3  +=       0.001 * cos(2.44816565966  + 26084.02180621620 *T)

    # MERCURY Latitude
    sumB4  = 0
    sumB4  +=       4.276 * cos(1.74579932115  + 26087.90314157420 *T)
    sumB4  +=       1.023 * cos(3.14159265359  + 0.00000000000 *T)
    sumB4  +=       0.425 * cos(4.03419509143  + 52175.80628314840 *T)
    sumB4  +=       0.257 * cos(0.20643590425  + 78263.70942472259 *T)
    sumB4  +=       0.116 * cos(3.75237354024  + 104351.61256629678 *T)
    sumB4  +=       0.073 * cos(1.18210375402  + 130439.51570787100 *T)
    sumB4  +=       0.051 * cos(4.54581086194  + 156527.41884944518 *T)
    sumB4  +=       0.031 * cos(1.44226942756  + 182615.32199101936 *T)
    sumB4  +=       0.016 * cos(4.56372679459  + 208703.22513259359 *T)
    sumB4  +=       0.008 * cos(1.34684622635  + 234791.12827416777 *T)
    sumB4  +=       0.003 * cos(4.40269422669  + 260879.03141574195 *T)
    sumB4  +=       0.001 * cos(4.69096566687  + 27197.28169366760 *T)
    sumB4  +=       0.001 * cos(1.16082903756  + 286966.93455731618 *T)

    # MERCURY Latitude
    sumB5  = 0
    sumB5  +=       0.106 * cos(3.94555784256  + 26087.90314157420 *T)
    sumB5  +=       0.075 * cos(3.14159265359  + 0.00000000000 *T)
    sumB5  +=       0.022 * cos(1.30514874546  + 52175.80628314840 *T)
    sumB5  +=       0.007 * cos(4.99717136857  + 78263.70942472259 *T)
    sumB5  +=       0.004 * cos(2.05662545278  + 104351.61256629678 *T)
    sumB5  +=       0.002 * cos(5.43162058632  + 130439.51570787100 *T)
    sumB5  +=       0.001 * cos(2.68052517331  + 156527.41884944518 *T)
    sumB5  +=       0.001 * cos(6.00841870284  + 182615.32199101936 *T)

    # MERCURY Radius
    sumR0  = 0
    sumR0  +=  39528271.652 * cos(0.00000000000  + 0.00000000000 *T)
    sumR0  +=  7834131.817 * cos(6.19233722599  + 26087.90314157420 *T)
    sumR0  +=  795525.557 * cos(2.95989690096  + 52175.80628314840 *T)
    sumR0  +=  121281.763 * cos(6.01064153805  + 78263.70942472259 *T)
    sumR0  +=   21921.969 * cos(2.77820093975  + 104351.61256629678 *T)
    sumR0  +=    4354.065 * cos(5.82894543257  + 130439.51570787100 *T)
    sumR0  +=     918.228 * cos(2.59650562598  + 156527.41884944518 *T)
    sumR0  +=     260.033 * cos(3.02817753482  + 27197.28169366760 *T)
    sumR0  +=     289.955 * cos(1.42441936951  + 25028.52121138500 *T)
    sumR0  +=     201.855 * cos(5.64725040350  + 182615.32199101936 *T)
    sumR0  +=     201.499 * cos(5.59227724202  + 31749.23519072640 *T)
    sumR0  +=     141.980 * cos(6.25264202645  + 24978.52458948080 *T)
    sumR0  +=     100.144 * cos(3.73435608689  + 21535.94964451540 *T)
    sumR0  +=      77.561 * cos(3.66972526976  + 20426.57109242200 *T)
    sumR0  +=      63.277 * cos(4.29905918105  + 25558.21217647960 *T)
    sumR0  +=      62.951 * cos(4.76588899933  + 1059.38193018920 *T)
    sumR0  +=      66.754 * cos(2.52520309182  + 5661.33204915220 *T)
    sumR0  +=      75.500 * cos(4.47428642962  + 51116.42435295920 *T)
    sumR0  +=      48.266 * cos(6.06824478778  + 53285.18483524180 *T)
    sumR0  +=      45.748 * cos(2.41480951648  + 208703.22513259359 *T)
    sumR0  +=      35.224 * cos(1.05917802674  + 27043.50288318280 *T)
    sumR0  +=      40.815 * cos(2.35882016415  + 57837.13833230060 *T)
    sumR0  +=      44.234 * cos(1.21957314874  + 15874.61759536320 *T)
    sumR0  +=      33.873 * cos(0.86381554651  + 25661.30495069820 *T)
    sumR0  +=      37.203 * cos(0.51733821470  + 47623.85278608960 *T)
    sumR0  +=      30.092 * cos(1.79500530627  + 37410.56723987860 *T)
    sumR0  +=      28.417 * cos(3.02063625668  + 51066.42773105500 *T)
    sumR0  +=      30.903 * cos(0.88366335532  + 24498.83024629040 *T)
    sumR0  +=      26.105 * cos(2.15021963174  + 39609.65458316560 *T)
    sumR0  +=      18.699 * cos(4.96496008403  + 11322.66409830440 *T)
    sumR0  +=      21.270 * cos(5.36857139841  + 13521.75144159140 *T)
    sumR0  +=      19.422 * cos(4.98378647655  + 10213.28554621100 *T)
    sumR0  +=      16.941 * cos(3.88765393402  + 26617.59410666880 *T)
    sumR0  +=      15.109 * cos(0.44510589948  + 46514.47423399620 *T)
    sumR0  +=      17.087 * cos(1.24077764194  + 77204.32749453338 *T)
    sumR0  +=      13.940 * cos(1.62573946865  + 27147.28507176339 *T)
    sumR0  +=      13.382 * cos(1.07657890477  + 51646.11531805379 *T)
    sumR0  +=      15.012 * cos(4.28173463507  + 41962.52073693740 *T)
    sumR0  +=      13.977 * cos(4.77056848793  + 33326.57873317420 *T)
    sumR0  +=      12.794 * cos(6.06437138766  + 1109.37855209340 *T)
    sumR0  +=      13.938 * cos(1.99984876578  + 25132.30339996560 *T)
    sumR0  +=      16.297 * cos(2.63293587817  + 19804.82729158280 *T)
    sumR0  +=      11.933 * cos(2.36500939134  + 4551.95349705880 *T)
    sumR0  +=      10.612 * cos(5.46555460932  + 234791.12827416777 *T)
    sumR0  +=      12.754 * cos(2.07613721222  + 529.69096509460 *T)
    sumR0  +=      12.069 * cos(2.84997619452  + 79373.08797681599 *T)
    sumR0  +=       9.069 * cos(1.21263611811  + 14765.23904326980 *T)
    sumR0  +=       9.491 * cos(0.83697007534  + 12566.15169998280 *T)
    sumR0  +=       9.379 * cos(5.41195286503  + 83925.04147387479 *T)
    sumR0  +=       7.499 * cos(2.44636811119  + 30639.85663863300 *T)
    sumR0  +=       7.463 * cos(5.53233943198  + 32858.61374281979 *T)
    sumR0  +=       7.215 * cos(1.17101960691  + 16983.99614745660 *T)
    sumR0  +=       8.492 * cos(3.56622930635  + 73711.75592766379 *T)
    sumR0  +=       7.109 * cos(5.32625264624  + 426.59819087600 *T)
    sumR0  +=       6.863 * cos(1.82313992707  + 36301.18868778519 *T)
    sumR0  +=       6.560 * cos(4.27818149213  + 43071.89928903080 *T)
    sumR0  +=       8.853 * cos(3.87737694074  + 50586.73338786459 *T)
    sumR0  +=       6.885 * cos(5.31108526630  + 1589.07289528380 *T)
    sumR0  +=       5.943 * cos(4.06893157254  + 53131.40602475700 *T)
    sumR0  +=       7.653 * cos(3.91505031889  + 51749.20809227239 *T)
    sumR0  +=       6.460 * cos(6.07127449283  + 77154.33087262919 *T)
    sumR0  +=       5.415 * cos(5.20028107807  + 65697.55772473979 *T)
    sumR0  +=       5.186 * cos(3.56743214904  + 6283.07584999140 *T)
    sumR0  +=       4.090 * cos(3.67593658710  + 22645.32819660879 *T)
    sumR0  +=       4.075 * cos(4.29142111073  + 103292.23063610760 *T)
    sumR0  +=       4.428 * cos(5.69109527379  + 45892.73043315699 *T)
    sumR0  +=       3.426 * cos(0.64911736234  + 52705.49724824299 *T)
    sumR0  +=       3.354 * cos(3.49345105494  + 72602.37737557039 *T)
    sumR0  +=       4.120 * cos(1.65386255382  + 25448.00585526019 *T)
    sumR0  +=       3.261 * cos(1.03195890028  + 68050.42387851159 *T)
    sumR0  +=       3.278 * cos(3.13863529552  + 25874.60404613620 *T)
    sumR0  +=       3.054 * cos(4.12578324522  + 77734.01845962799 *T)
    sumR0  +=       3.614 * cos(2.95861599353  + 28306.66024576099 *T)
    sumR0  +=       2.820 * cos(4.62628436074  + 18849.22754997420 *T)
    sumR0  +=       2.746 * cos(4.67880529205  + 53235.18821333759 *T)
    sumR0  +=       2.713 * cos(3.56874913980  + 26068.23338067440 *T)
    sumR0  +=       2.507 * cos(2.23312085627  + 260879.03141574195 *T)
    sumR0  +=       3.044 * cos(5.90613310181  + 105460.99111839020 *T)
    sumR0  +=       2.755 * cos(5.67587087632  + 26107.57290247399 *T)
    sumR0  +=       2.267 * cos(2.17941340037  + 110012.94461544900 *T)
    sumR0  +=       2.274 * cos(0.50173671332  + 32370.97899156560 *T)
    sumR0  +=       2.741 * cos(4.75427920262  + 63498.47038145279 *T)
    sumR0  +=       2.792 * cos(5.05148206806  + 51220.20654153979 *T)
    sumR0  +=       2.557 * cos(2.95114696617  + 26080.78959457339 *T)
    sumR0  +=       2.135 * cos(0.84621449019  + 76674.63652943878 *T)
    sumR0  +=       1.971 * cos(0.47407494310  + 48733.23133818299 *T)
    sumR0  +=       1.965 * cos(1.37420902653  + 7238.67559160000 *T)
    sumR0  +=       2.526 * cos(0.34535290139  + 23969.13928119580 *T)
    sumR0  +=       2.022 * cos(0.33828477567  + 99799.65906923798 *T)
    sumR0  +=       1.872 * cos(3.93046425037  + 38654.05484155699 *T)
    sumR0  +=       1.873 * cos(0.69327393635  + 26091.78447693220 *T)
    sumR0  +=       1.945 * cos(2.37383894182  + 6770.71060124560 *T)
    sumR0  +=       1.863 * cos(2.26591720786  + 26084.02180621620 *T)
    sumR0  +=       1.750 * cos(5.50651903992  + 56727.75978020720 *T)
    sumR0  +=       1.971 * cos(3.77455887616  + 19317.19254032860 *T)
    sumR0  +=       1.672 * cos(2.15275897581  + 26514.50133245020 *T)
    sumR0  +=       1.703 * cos(4.28965990785  + 40853.14218484400 *T)
    sumR0  +=       1.657 * cos(4.98021773372  + 9103.90699411760 *T)
    sumR0  +=       2.084 * cos(2.19427123968  + 26301.20223701220 *T)
    sumR0  +=       1.818 * cos(4.20870869718  + 25938.33994443960 *T)
    sumR0  +=       1.780 * cos(0.91702846577  + 79219.30916633119 *T)
    sumR0  +=       1.525 * cos(2.15720744047  + 27676.97603685800 *T)
    sumR0  +=       1.533 * cos(1.54503054415  + 955.59974160860 *T)
    sumR0  +=       1.614 * cos(1.27565104562  + 25035.63475838580 *T)
    sumR0  +=       1.743 * cos(0.68269002122  + 77837.11123384659 *T)
    sumR0  +=       1.389 * cos(4.85102048256  + 62389.09182935939 *T)
    sumR0  +=       1.436 * cos(0.05742216761  + 23869.14603738740 *T)
    sumR0  +=       1.563 * cos(5.84795233948  + 26095.01668857500 *T)
    sumR0  +=       1.640 * cos(4.66281337712  + 25021.40766438420 *T)
    sumR0  +=       1.552 * cos(2.83910580545  + 103242.23401420339 *T)
    sumR0  +=       1.249 * cos(1.97225274067  + 91785.46086631398 *T)
    sumR0  +=       1.160 * cos(1.73113341823  + 38519.94579197200 *T)
    sumR0  +=       1.005 * cos(1.05898314860  + 129380.13377768178 *T)
    sumR0  +=       1.041 * cos(2.78036322434  + 213.29909543800 *T)
    sumR0  +=       1.105 * cos(4.55512109515  + 26011.63707029860 *T)
    sumR0  +=       0.979 * cos(2.43870419667  + 71980.63357473118 *T)
    sumR0  +=       0.853 * cos(0.56334001923  + 20760.42703319140 *T)
    sumR0  +=       0.874 * cos(4.53679970936  + 639.89728631400 *T)
    sumR0  +=       0.783 * cos(3.69782403005  + 78793.40038981718 *T)
    sumR0  +=       0.772 * cos(4.08714349549  + 94138.32702008578 *T)
    sumR0  +=       0.797 * cos(0.26237215917  + 98690.28051714458 *T)
    sumR0  +=       0.965 * cos(4.70784179954  + 51535.90899683439 *T)
    sumR0  +=       0.952 * cos(2.53604676495  + 29530.47808653960 *T)
    sumR0  +=       0.773 * cos(2.67893222158  + 131548.89425996438 *T)
    sumR0  +=       0.734 * cos(0.89243669490  + 103821.92160120218 *T)
    sumR0  +=       0.668 * cos(3.73113394924  + 46848.33017476560 *T)
    sumR0  +=       0.828 * cos(2.88184821542  + 19406.67828817460 *T)
    sumR0  +=       0.622 * cos(1.05579535631  + 59414.48187474840 *T)
    sumR0  +=       0.676 * cos(4.50357599360  + 25934.12433108940 *T)
    sumR0  +=       0.602 * cos(5.28399057704  + 286966.93455731618 *T)
    sumR0  +=       0.619 * cos(1.44855037685  + 79323.09135491178 *T)
    sumR0  +=       0.691 * cos(6.19733262608  + 51962.50718771040 *T)
    sumR0  +=       0.563 * cos(5.22994296186  + 136100.84775702321 *T)
    sumR0  +=       0.782 * cos(3.25257254691  + 50057.04242277000 *T)
    sumR0  +=       0.551 * cos(0.33644229836  + 52156.13652224860 *T)
    sumR0  +=       0.705 * cos(4.73705371920  + 26241.68195205900 *T)
    sumR0  +=       0.577 * cos(5.85654734429  + 2118.76386037840 *T)
    sumR0  +=       0.517 * cos(3.93195379429  + 102762.53967101300 *T)
    sumR0  +=       0.635 * cos(4.87720838965  + 12432.04265039780 *T)
    sumR0  +=       0.635 * cos(1.81903685898  + 77308.10968311399 *T)
    sumR0  +=       0.543 * cos(1.40877130839  + 44937.13069154840 *T)
    sumR0  +=       0.556 * cos(5.44324390449  + 85034.42002596818 *T)
    sumR0  +=       0.496 * cos(3.39004132624  + 125887.56221081220 *T)
    sumR0  +=       0.546 * cos(4.78005317483  + 22747.29071487440 *T)
    sumR0  +=       0.520 * cos(6.00172746972  + 52168.69273614759 *T)
    sumR0  +=       0.463 * cos(3.67549187956  + 25668.41849769900 *T)
    sumR0  +=       0.452 * cos(3.53009000682  + 58458.88213313979 *T)
    sumR0  +=       0.474 * cos(5.94330512040  + 54394.56338733519 *T)
    sumR0  +=       0.559 * cos(2.44355554028  + 52195.47604404819 *T)
    sumR0  +=       0.418 * cos(3.73369494476  + 103925.01437542078 *T)
    sumR0  +=       0.397 * cos(0.69014608020  + 64741.95798313119 *T)
    sumR0  +=       0.511 * cos(5.88243734524  + 45494.58142974879 *T)
    sumR0  +=       0.412 * cos(2.26754404446  + 82815.66292178139 *T)
    sumR0  +=       0.425 * cos(5.24511927496  + 52389.10537858640 *T)
    sumR0  +=       0.482 * cos(0.98165400059  + 25654.19140369739 *T)
    sumR0  +=       0.380 * cos(3.74413132157  + 52179.68761850640 *T)
    sumR0  +=       0.359 * cos(5.50961077408  + 26555.86813192860 *T)
    sumR0  +=       0.375 * cos(1.97920050124  + 149.56319713460 *T)
    sumR0  +=       0.385 * cos(5.88991948790  + 129330.13715577760 *T)
    sumR0  +=       0.348 * cos(4.91064019959  + 1052.26838318840 *T)
    sumR0  +=       0.417 * cos(4.32916053867  + 51123.53789995999 *T)
    sumR0  +=       0.425 * cos(1.43228996047  + 51109.31080595839 *T)
    sumR0  +=       0.464 * cos(3.98651628073  + 105307.21230790540 *T)
    sumR0  +=       0.329 * cos(5.20636278221  + 52602.40447402440 *T)
    sumR0  +=       0.354 * cos(1.04383146533  + 66941.04532641819 *T)
    sumR0  +=       0.324 * cos(1.62975761811  + 88476.99497093359 *T)
    sumR0  +=       0.378 * cos(5.31676765847  + 52171.92494779040 *T)
    sumR0  +=       0.440 * cos(1.46381183715  + 89586.37352302698 *T)
    sumR0  +=       0.355 * cos(1.52732265492  + 1066.49547719000 *T)
    sumR0  +=       0.302 * cos(5.21113786435  + 53764.87917843220 *T)
    sumR0  +=       0.301 * cos(5.02488829134  + 117873.36400788820 *T)
    sumR0  +=       0.351 * cos(2.00433341730  + 24395.73747207180 *T)
    sumR0  +=       0.289 * cos(6.19161799918  + 26094.53170047421 *T)
    sumR0  +=       0.286 * cos(3.77170690304  + 7.11354700080 *T)
    sumR0  +=       0.284 * cos(3.10942616808  + 49957.04917896160 *T)
    sumR0  +=       0.385 * cos(2.47184729453  + 25234.70675982219 *T)
    sumR0  +=       0.382 * cos(0.97683830518  + 52026.24308601380 *T)
    sumR0  +=       0.271 * cos(5.74710334372  + 25455.11940226100 *T)
    sumR0  +=       0.375 * cos(0.55187466535  + 45405.09568190280 *T)
    sumR0  +=       0.267 * cos(2.86410554301  + 80482.46652890938 *T)
    sumR0  +=       0.304 * cos(6.00190488272  + 2218.75710418680 *T)
    sumR0  +=       0.259 * cos(1.82921035765  + 24505.94379329119 *T)
    sumR0  +=       0.251 * cos(4.10979574073  + 155468.03691925600 *T)
    sumR0  +=       0.289 * cos(3.05373259745  + 26081.27458267419 *T)
    sumR0  +=       0.242 * cos(5.05501089774  + 42153.96900304900 *T)
    sumR0  +=       0.244 * cos(2.15621668050  + 3442.57494496540 *T)
    sumR0  +=       0.315 * cos(2.61613386386  + 52182.91983014920 *T)
    sumR0  +=       0.256 * cos(3.71117332461  + 25619.93815121980 *T)
    sumR0  +=       0.232 * cos(2.21576531181  + 16066.06586147480 *T)
    sumR0  +=       0.233 * cos(5.49033130906  + 98068.53671630540 *T)
    sumR0  +=       0.236 * cos(4.08612620637  + 26013.12154300690 *T)
    sumR0  +=       0.225 * cos(0.44449338810  + 76144.94556434419 *T)
    sumR0  +=       0.242 * cos(3.29851820234  + 6681.22485339960 *T)
    sumR0  +=       0.227 * cos(1.63236736805  + 76.26607127560 *T)
    sumR0  +=       0.268 * cos(1.09037236588  + 18093.37469954999 *T)
    sumR0  +=       0.255 * cos(4.20635818545  + 44181.27784112419 *T)
    sumR0  +=       0.205 * cos(1.00234895959  + 26190.99591579279 *T)
    sumR0  +=       0.228 * cos(1.46077047931  + 77623.81213840858 *T)
    sumR0  +=       0.197 * cos(1.36693360783  + 26727.80042788820 *T)
    sumR0  +=       0.259 * cos(1.28296039354  + 13655.86049117640 *T)
    sumR0  +=       0.197 * cos(5.73307583430  + 157636.79740153858 *T)
    sumR0  +=       0.196 * cos(0.06606260924  + 40565.25432477420 *T)
    sumR0  +=       0.187 * cos(0.86187283408  + 120226.23016166000 *T)
    sumR0  +=       0.262 * cos(4.14850032709  + 24491.71669928959 *T)
    sumR0  +=       0.244 * cos(4.14399890076  + 95247.70557217918 *T)
    sumR0  +=       0.232 * cos(5.45165695316  + 33967.99229491319 *T)
    sumR0  +=       0.206 * cos(5.59196913034  + 55618.38122811380 *T)
    sumR0  +=       0.203 * cos(1.64252462466  + 58946.51688439399 *T)
    sumR0  +=       0.187 * cos(0.46291089182  + 104881.30353139140 *T)
    sumR0  +=       0.192 * cos(1.82761158300  + 25551.09862947879 *T)
    sumR0  +=       0.200 * cos(1.81376901527  + 25863.55834587229 *T)
    sumR0  +=       0.191 * cos(5.57428963071  + 5327.47610838280 *T)
    sumR0  +=       0.195 * cos(3.31297083154  + 124778.18365871880 *T)
    sumR0  +=       0.169 * cos(1.69270064342  + 90695.75207512038 *T)
    sumR0  +=       0.226 * cos(1.32354627258  + 52099.54021187280 *T)
    sumR0  +=       0.177 * cos(3.29620848300  + 25131.61398560359 *T)
    sumR0  +=       0.178 * cos(2.54823795050  + 26720.68688088739 *T)
    sumR0  +=       0.167 * cos(3.61983109002  + 23754.70674870219 *T)
    sumR0  +=       0.159 * cos(3.48396967776  + 25977.69682035479 *T)
    sumR0  +=       0.150 * cos(5.20072476078  + 69159.80243060499 *T)
    sumR0  +=       0.179 * cos(3.94252428767  + 129909.82474277640 *T)
    sumR0  +=       0.166 * cos(0.40320638452  + 100909.03762133140 *T)
    sumR0  +=       0.176 * cos(2.97410942402  + 24925.42843716640 *T)
    sumR0  +=       0.154 * cos(1.27206049137  + 25984.81036735560 *T)
    sumR0  +=       0.144 * cos(1.89319156163  + 64607.84893354619 *T)
    sumR0  +=       0.145 * cos(2.05409839424  + 313054.83769889042 *T)
    sumR0  +=       0.152 * cos(4.87194153709  + 103396.01282468818 *T)
    sumR0  +=       0.137 * cos(3.86782217557  + 24609.03656750980 *T)
    sumR0  +=       0.141 * cos(1.99770625298  + 162188.75089859738 *T)
    sumR0  +=       0.185 * cos(6.10038742063  + 23439.44831610119 *T)
    sumR0  +=       0.135 * cos(1.27363012793  + 52022.02747266360 *T)
    sumR0  +=       0.127 * cos(0.70464976569  + 128850.44281258718 *T)
    sumR0  +=       0.142 * cos(1.50696203402  + 52329.58509363319 *T)
    sumR0  +=       0.126 * cos(2.70555128398  + 28206.66700195260 *T)
    sumR0  +=       0.126 * cos(1.61963252183  + 48835.19385644859 *T)
    sumR0  +=       0.123 * cos(0.15629905349  + 151975.46535238638 *T)
    sumR0  +=       0.137 * cos(1.77306412615  + 35191.81013569180 *T)
    sumR0  +=       0.147 * cos(4.50311642422  + 105410.99449648600 *T)
    sumR0  +=       0.125 * cos(3.38994834526  + 78244.03966382280 *T)
    sumR0  +=       0.119 * cos(2.77644375904  + 78256.59587772179 *T)
    sumR0  +=       0.156 * cos(2.96616703196  + 78050.41032928458 *T)
    sumR0  +=       0.145 * cos(1.35569389554  + 3340.61242669980 *T)
    sumR0  +=       0.126 * cos(3.97006576134  + 29428.51556827400 *T)
    sumR0  +=       0.128 * cos(5.41891989692  + 26162.68474014150 *T)
    sumR0  +=       0.132 * cos(4.86901248209  + 24712.12934172840 *T)
    sumR0  +=       0.110 * cos(3.87954702440  + 85502.38501632259 *T)
    sumR0  +=       0.127 * cos(5.49703075555  + 78283.37918562238 *T)
    sumR0  +=       0.126 * cos(0.94870979828  + 121335.60871375341 *T)
    sumR0  +=       0.129 * cos(4.00691218396  + 72936.23331633979 *T)
    sumR0  +=       0.102 * cos(0.50487650429  + 130012.91751699500 *T)
    sumR0  +=       0.104 * cos(0.29849788751  + 84546.78527471398 *T)
    sumR0  +=       0.118 * cos(0.76511176346  + 131395.11544947961 *T)
    sumR0  +=       0.120 * cos(4.45261989231  + 71025.03383312259 *T)
    sumR0  +=       0.117 * cos(3.31644134053  + 25771.51127191760 *T)
    sumR0  +=       0.110 * cos(2.63249934300  + 71582.48457132299 *T)
    sumR0  +=       0.097 * cos(2.66105435933  + 155418.04029735178 *T)
    sumR0  +=       0.113 * cos(5.88733575390  + 26404.29501123080 *T)
    sumR0  +=       0.098 * cos(2.01417457095  + 78477.00852016058 *T)
    sumR0  +=       0.103 * cos(0.46416881502  + 51756.32163927320 *T)
    sumR0  +=       0.093 * cos(2.37389467504  + 419.48464387520 *T)
    sumR0  +=       0.125 * cos(1.87932536282  + 26610.48055966799 *T)
    sumR0  +=       0.095 * cos(0.34221817875  + 26507.38778544939 *T)
    sumR0  +=       0.093 * cos(3.80346806552  + 26037.90651966999 *T)
    sumR0  +=       0.091 * cos(2.63082094391  + 28421.09953444620 *T)
    sumR0  +=       0.101 * cos(5.20997240587  + 433.71173787680 *T)
    sumR0  +=       0.096 * cos(4.47724179825  + 77197.21394753258 *T)
    sumR0  +=       0.108 * cos(4.03679713919  + 51742.09454527159 *T)
    sumR0  +=       0.086 * cos(5.14039556935  + 26237.46633870879 *T)
    sumR0  +=       0.092 * cos(3.74335845403  + 90829.86112470538 *T)
    sumR0  +=       0.086 * cos(1.86276413658  + 27140.17152476259 *T)
    sumR0  +=       0.102 * cos(5.40465219833  + 26137.89976347840 *T)
    sumR0  +=       0.086 * cos(4.01653667598  + 10021.83728009940 *T)
    sumR0  +=       0.100 * cos(5.32142188276  + 108903.56606335558 *T)
    sumR0  +=       0.079 * cos(4.68108496987  + 27154.39861876420 *T)
    sumR0  +=       0.092 * cos(5.52103093635  + 51322.60990139639 *T)
    sumR0  +=       0.093 * cos(4.20604773511  + 38813.35657634920 *T)
    sumR0  +=       0.078 * cos(2.88001982141  + 29416.03879785439 *T)
    sumR0  +=       0.095 * cos(4.44018421307  + 115674.27666460120 *T)
    sumR0  +=       0.104 * cos(5.05004528676  + 50483.64061364600 *T)
    sumR0  +=       0.075 * cos(1.78927060530  + 143961.26714946239 *T)
    sumR0  +=       0.086 * cos(0.51423355089  + 78267.59076008058 *T)
    sumR0  +=       0.078 * cos(4.66831173790  + 114564.89811250778 *T)
    sumR0  +=       0.078 * cos(4.20137623264  + 1692.16566950240 *T)
    sumR0  +=       0.086 * cos(4.03030391951  + 78114.14622758799 *T)
    sumR0  +=       0.073 * cos(2.27328552013  + 52643.77127350280 *T)
    sumR0  +=       0.082 * cos(3.73510592110  + 853.19638175200 *T)
    sumR0  +=       0.071 * cos(4.89150553141  + 50593.84693486539 *T)
    sumR0  +=       0.082 * cos(5.22677693918  + 27999.10262479140 *T)
    sumR0  +=       0.093 * cos(1.09569161440  + 77211.44104153418 *T)
    sumR0  +=       0.074 * cos(1.97833720876  + 78690.30761559859 *T)
    sumR0  +=       0.082 * cos(3.59482991001  + 71492.99882347698 *T)
    sumR0  +=       0.083 * cos(4.10355702051  + 93028.94846799241 *T)
    sumR0  +=       0.083 * cos(3.12647584943  + 14477.35118320000 *T)
    sumR0  +=       0.067 * cos(0.24182046727  + 66653.15746634839 *T)
    sumR0  +=       0.074 * cos(5.66833060422  + 78270.82297172339 *T)
    sumR0  +=       0.068 * cos(1.98236355905  + 79852.78232000639 *T)
    sumR0  +=       0.064 * cos(5.55115814500  + 25764.39772491679 *T)
    sumR0  +=       0.064 * cos(6.15935356751  + 76044.95232053580 *T)
    sumR0  +=       0.087 * cos(1.77296830923  + 25440.89230825939 *T)
    sumR0  +=       0.073 * cos(2.06220281759  + 111122.32316754240 *T)
    sumR0  +=       0.066 * cos(2.28151075383  + 7880.08915333899 *T)
    sumR0  +=       0.076 * cos(2.65553220848  + 34282.17847478280 *T)
    sumR0  +=       0.078 * cos(0.40338105541  + 49842.60989027639 *T)
    sumR0  +=       0.064 * cos(0.87558639332  + 181555.94006083015 *T)
    sumR0  +=       0.086 * cos(2.08697748680  + 78259.82808936459 *T)
    sumR0  +=       0.074 * cos(0.86501751700  + 50579.61984086379 *T)
    sumR0  +=       0.082 * cos(1.02659883956  + 74821.13447975718 *T)
    sumR0  +=       0.059 * cos(2.45952644668  + 51543.02254383520 *T)
    sumR0  +=       0.062 * cos(0.94273311929  + 25344.91308104160 *T)
    sumR0  +=       0.058 * cos(2.95861781205  + 52182.43484204840 *T)
    sumR0  +=       0.058 * cos(4.20402726578  + 96357.08412427260 *T)
    sumR0  +=       0.061 * cos(2.81500265978  + 26202.34243025941 *T)
    sumR0  +=       0.057 * cos(2.27689377996  + 124156.43985787958 *T)
    sumR0  +=       0.057 * cos(4.37167017012  + 1581.95934828300 *T)
    sumR0  +=       0.055 * cos(0.12465494356  + 25973.46385288896 *T)
    sumR0  +=       0.054 * cos(4.79303212221  + 26164.16921284980 *T)
    sumR0  +=       0.054 * cos(3.05622204718  + 25565.32572348040 *T)
    sumR0  +=       0.055 * cos(4.51497762833  + 103711.71527998280 *T)
    sumR0  +=       0.054 * cos(3.87143037686  + 18207.81398823521 *T)
    sumR0  +=       0.054 * cos(4.62784653733  + 25881.71759313700 *T)
    sumR0  +=       0.053 * cos(4.32548272270  + 70269.18098269838 *T)
    sumR0  +=       0.050 * cos(2.50066877650  + 183724.70054311279 *T)
    sumR0  +=       0.058 * cos(2.06211216656  + 1596.18644228460 *T)
    sumR0  +=       0.066 * cos(5.20298814290  + 26294.08869001139 *T)
    sumR0  +=       0.047 * cos(0.85039730966  + 52101.02468458109 *T)
    sumR0  +=       0.058 * cos(1.43409045174  + 23866.04650697719 *T)
    sumR0  +=       0.048 * cos(2.35364103056  + 81706.28436968799 *T)
    sumR0  +=       0.052 * cos(0.19368783267  + 92741.06060792258 *T)
    sumR0  +=       0.053 * cos(0.58517949906  + 26624.70765366959 *T)
    sumR0  +=       0.061 * cos(2.61493039721  + 49527.35145767539 *T)
    sumR0  +=       0.047 * cos(3.88384847335  + 146314.13330323418 *T)
    sumR0  +=       0.049 * cos(6.20308247475  + 32769.12799497380 *T)
    sumR0  +=       0.053 * cos(4.37486529196  + 78187.44335344699 *T)
    sumR0  +=       0.048 * cos(5.97568482401  + 51013.33157874059 *T)
    sumR0  +=       0.048 * cos(3.84102043742  + 102232.84870591838 *T)
    sumR0  +=       0.049 * cos(5.05464216653  + 7994.52844202420 *T)
    sumR0  +=       0.044 * cos(5.42890224840  + 86143.79857806159 *T)
    sumR0  +=       0.058 * cos(3.80571138237  + 99024.13645791401 *T)
    sumR0  +=       0.044 * cos(4.02792920142  + 147423.51185532758 *T)
    sumR0  +=       0.058 * cos(6.09746566424  + 52169.17772424839 *T)
    sumR0  +=       0.042 * cos(4.38843836266  + 224.34479570190 *T)
    sumR0  +=       0.043 * cos(1.57859396874  + 23976.25282819660 *T)
    sumR0  +=       0.049 * cos(0.46973350644  + 51707.84129279399 *T)
    sumR0  +=       0.040 * cos(4.41764731583  + 52815.70356946240 *T)
    sumR0  +=       0.046 * cos(3.51985131014  + 130969.20667296558 *T)
    sumR0  +=       0.045 * cos(4.35527612249  + 39743.76363275060 *T)
    sumR0  +=       0.045 * cos(0.78163192715  + 25138.72753260440 *T)
    sumR0  +=       0.049 * cos(0.90057831755  + 632.78373931320 *T)
    sumR0  +=       0.041 * cos(4.04225615670  + 52278.89905736699 *T)
    sumR0  +=       0.039 * cos(2.91829595120  + 956.28915597060 *T)
    sumR0  +=       0.043 * cos(0.13178694356  + 2648.45482547300 *T)
    sumR0  +=       0.038 * cos(0.62377784752  + 50696.93970908399 *T)
    sumR0  +=       0.042 * cos(2.75028269027  + 24079.34560241519 *T)
    sumR0  +=       0.040 * cos(1.64358052897  + 39629.32434406539 *T)
    sumR0  +=       0.037 * cos(5.75760879379  + 25867.49049913539 *T)
    sumR0  +=       0.051 * cos(1.57829002915  + 12725.45343477500 *T)
    sumR0  +=       0.037 * cos(1.62981695072  + 129483.91596626240 *T)
    sumR0  +=       0.045 * cos(0.04645406242  + 51219.51712717779 *T)
    sumR0  +=       0.036 * cos(0.09834373600  + 25241.82030682300 *T)
    sumR0  +=       0.049 * cos(0.07681157050  + 150866.08680029298 *T)
    sumR0  +=       0.034 * cos(4.56091607046  + 536.80451209540 *T)
    sumR0  +=       0.036 * cos(5.59483746723  + 52808.59002246159 *T)
    sumR0  +=       0.039 * cos(5.99500252470  + 104138.31347085880 *T)
    sumR0  +=       0.046 * cos(0.70772608543  + 155997.72788435058 *T)
    sumR0  +=       0.033 * cos(4.54716206885  + 78417.48823520739 *T)
    sumR0  +=       0.041 * cos(4.92673567168  + 51639.00177105299 *T)
    sumR0  +=       0.032 * cos(4.31372208361  + 78109.93061423779 *T)
    sumR0  +=       0.036 * cos(5.69791953426  + 26049.77010593640 *T)
    sumR0  +=       0.034 * cos(3.32168809248  + 35472.74414964940 *T)
    sumR0  +=       0.042 * cos(4.86530735726  + 51951.46148744649 *T)
    sumR0  +=       0.033 * cos(3.73733919231  + 154938.34595416137 *T)
    sumR0  +=       0.041 * cos(2.38147812869  + 68241.87214462319 *T)
    sumR0  +=       0.035 * cos(5.09571047220  + 339142.74084046454 *T)
    sumR0  +=       0.031 * cos(4.31506703539  + 52072.71350892979 *T)
    sumR0  +=       0.032 * cos(3.21248509301  + 178063.36849396059 *T)
    sumR0  +=       0.035 * cos(5.04278529087  + 188276.65404017159 *T)
    sumR0  +=       0.041 * cos(3.22838990589  + 1162.47470440780 *T)
    sumR0  +=       0.030 * cos(2.33061092140  + 1478.86657406440 *T)
    sumR0  +=       0.032 * cos(2.71244280066  + 25780.34552060460 *T)
    sumR0  +=       0.032 * cos(3.80411460420  + 157483.01859105378 *T)
    sumR0  +=       0.030 * cos(0.37624991562  + 26421.75908234360 *T)
    sumR0  +=       0.029 * cos(5.82513433576  + 104344.49901929598 *T)
    sumR0  +=       0.031 * cos(6.02215690021  + 26198.10946279360 *T)
    sumR0  +=       0.040 * cos(1.02201252788  + 27819.02549450680 *T)
    sumR0  +=       0.030 * cos(2.25849995020  + 104371.28232719658 *T)
    sumR0  +=       0.029 * cos(2.67851330746  + 110.20632121940 *T)
    sumR0  +=       0.030 * cos(1.34446903292  + 1375.77379984580 *T)
    sumR0  +=       0.028 * cos(4.62845152557  + 74923.09699802278 *T)
    sumR0  +=       0.037 * cos(3.03899000872  + 11610.55195837420 *T)
    sumR0  +=       0.032 * cos(0.26062549755  + 52065.59996192899 *T)
    sumR0  +=       0.038 * cos(1.25704652400  + 131498.89763806018 *T)
    sumR0  +=       0.032 * cos(1.04832153595  + 26086.41866886590 *T)
    sumR0  +=       0.037 * cos(1.63595908508  + 50800.03248330259 *T)
    sumR0  +=       0.030 * cos(0.15550783905  + 104331.94280539700 *T)
    sumR0  +=       0.026 * cos(0.03549300809  + 24815.22211594700 *T)
    sumR0  +=       0.026 * cos(1.03871337432  + 24292.64469785319 *T)
    sumR0  +=       0.026 * cos(3.54557412535  + 156100.82065856917 *T)
    sumR0  +=       0.027 * cos(0.60333270859  + 111590.28815789679 *T)
    sumR0  +=       0.031 * cos(3.35540556110  + 126996.94076290558 *T)
    sumR0  +=       0.026 * cos(5.73937805822  + 54294.57014352679 *T)
    sumR0  +=       0.026 * cos(2.79743885101  + 86457.98475793119 *T)
    sumR0  +=       0.026 * cos(5.70579806927  + 97670.38771289718 *T)
    sumR0  +=       0.025 * cos(2.34928986187  + 24601.92302050899 *T)
    sumR0  +=       0.025 * cos(5.70613852018  + 181505.94343892601 *T)
    sumR0  +=       0.034 * cos(0.23651097126  + 26395.46076254379 *T)
    sumR0  +=       0.025 * cos(2.19241735868  + 52250.58788171570 *T)
    sumR0  +=       0.025 * cos(1.82880416247  + 74.78159856730 *T)
    sumR0  +=       0.028 * cos(0.57661358158  + 125112.03959948818 *T)
    sumR0  +=       0.024 * cos(2.03938254170  + 26521.61487945100 *T)
    sumR0  +=       0.024 * cos(5.05213893429  + 104564.91166173480 *T)
    sumR0  +=       0.023 * cos(6.23874249128  + 75615.25459924959 *T)
    sumR0  +=       0.026 * cos(3.34519516519  + 110634.68841628821 *T)
    sumR0  +=       0.026 * cos(4.74840245512  + 13362.44970679920 *T)
    sumR0  +=       0.026 * cos(0.15419927428  + 22759.76748529401 *T)
    sumR0  +=       0.027 * cos(2.39568978424  + 522.57741809380 *T)
    sumR0  +=       0.028 * cos(0.25542748453  + 103.09277421860 *T)
    sumR0  +=       0.025 * cos(3.32961707416  + 28286.99048486120 *T)
    sumR0  +=       0.026 * cos(2.07987634823  + 29550.14784743939 *T)
    sumR0  +=       0.032 * cos(1.70464933285  + 26089.38761428249 *T)
    sumR0  +=       0.023 * cos(4.16670746048  + 24176.70365835700 *T)
    sumR0  +=       0.024 * cos(0.08582796512  + 51859.41441349179 *T)
    sumR0  +=       0.025 * cos(5.27240695394  + 25936.85547173129 *T)
    sumR0  +=       0.024 * cos(0.80363857551  + 77829.99768684579 *T)
    sumR0  +=       0.023 * cos(2.30360039127  + 77410.51304297059 *T)
    sumR0  +=       0.030 * cos(1.21286011206  + 97112.93697469680 *T)
    sumR0  +=       0.022 * cos(4.82094558003  + 13541.42120249119 *T)
    sumR0  +=       0.023 * cos(2.65250267545  + 52492.19815280499 *T)
    sumR0  +=       0.026 * cos(6.26936948934  + 25939.82441714789 *T)
    sumR0  +=       0.024 * cos(4.92454390318  + 52698.38370124219 *T)
    sumR0  +=       0.022 * cos(1.23470921550  + 103285.11708910680 *T)
    sumR0  +=       0.023 * cos(3.68813881444  + 26729.31670331319 *T)
    sumR0  +=       0.022 * cos(1.95831083558  + 24356.78078864160 *T)
    sumR0  +=       0.022 * cos(0.78346634338  + 104202.04936916218 *T)
    sumR0  +=       0.021 * cos(2.52931432670  + 467.96499035440 *T)
    sumR0  +=       0.024 * cos(1.84656751628  + 76571.54375522019 *T)
    sumR0  +=       0.025 * cos(2.55626379808  + 24822.33566294780 *T)
    sumR0  +=       0.027 * cos(3.61440862477  + 23962.02573419499 *T)
    sumR0  +=       0.023 * cos(1.18953680418  + 141762.17980617538 *T)
    sumR0  +=       0.020 * cos(5.49778171394  + 112231.70171963578 *T)
    sumR0  +=       0.023 * cos(2.60719608050  + 25754.04720080480 *T)
    sumR0  +=       0.026 * cos(0.75788900266  + 55516.41870984820 *T)
    sumR0  +=       0.026 * cos(2.09375500363  + 134991.46920492980 *T)
    sumR0  +=       0.020 * cos(3.35091650179  + 52595.29092702359 *T)
    sumR0  +=       0.021 * cos(0.36570195077  + 97580.90196505121 *T)
    sumR0  +=       0.024 * cos(0.51085875774  + 116917.76426627958 *T)
    sumR0  +=       0.021 * cos(5.71018496974  + 60055.89543648739 *T)
    sumR0  +=       0.019 * cos(0.56236491686  + 52125.80966124419 *T)
    sumR0  +=       0.019 * cos(2.85801633295  + 60370.08161635699 *T)
    sumR0  +=       0.024 * cos(3.50657176712  + 77844.22478084739 *T)
    sumR0  +=       0.020 * cos(0.53533403977  + 26941.09952332620 *T)
    sumR0  +=       0.019 * cos(6.25214103463  + 25973.50403466079 *T)
    sumR0  +=       0.022 * cos(5.13145226592  + 104347.73123093881 *T)
    sumR0  +=       0.021 * cos(4.13407823619  + 103299.34418310841 *T)
    sumR0  +=       0.019 * cos(0.33492578967  + 102018.41617342478 *T)
    sumR0  +=       0.019 * cos(5.67148210198  + 54509.00267602040 *T)
    sumR0  +=       0.018 * cos(1.90051892916  + 52325.36948028299 *T)
    sumR0  +=       0.019 * cos(4.27364751897  + 122444.98726584678 *T)
    sumR0  +=       0.018 * cos(2.45675888957  + 104358.72611329758 *T)
    sumR0  +=       0.018 * cos(1.05990359493  + 27780.06881107659 *T)
    sumR0  +=       0.018 * cos(1.43297513339  + 53242.30176033840 *T)
    sumR0  +=       0.021 * cos(2.99876487527  + 26073.67604757259 *T)
    sumR0  +=       0.023 * cos(1.86389686355  + 116783.65521669458 *T)
    sumR0  +=       0.018 * cos(4.87152831033  + 170049.17029103660 *T)
    sumR0  +=       0.019 * cos(4.89400475242  + 53228.07466633679 *T)
    sumR0  +=       0.021 * cos(2.47295566818  + 27726.97265876220 *T)
    sumR0  +=       0.020 * cos(4.98762998978  + 104778.21075717278 *T)
    sumR0  +=       0.021 * cos(1.46468188731  + 140652.80125408198 *T)
    sumR0  +=       0.017 * cos(4.05153553008  + 76667.52298243798 *T)
    sumR0  +=       0.020 * cos(4.80985268979  + 51528.79544983359 *T)
    sumR0  +=       0.022 * cos(0.86617331302  + 119116.85160956658 *T)
    sumR0  +=       0.023 * cos(2.08281937549  + 52225.80290505260 *T)
    sumR0  +=       0.018 * cos(1.69333480892  + 76681.75007643958 *T)
    sumR0  +=       0.017 * cos(2.09406764538  + 64901.25971792339 *T)
    sumR0  +=       0.019 * cos(5.00011023502  + 105940.68546158058 *T)
    sumR0  +=       0.019 * cos(0.74020534937  + 36109.74042167360 *T)
    sumR0  +=       0.017 * cos(2.93246868693  + 102132.85546211000 *T)
    sumR0  +=       0.022 * cos(3.55902023067  + 104355.49390165480 *T)
    sumR0  +=       0.021 * cos(3.54390310557  + 106570.36967048360 *T)
    sumR0  +=       0.018 * cos(4.65006394461  + 61279.71327726600 *T)
    sumR0  +=       0.019 * cos(4.43905256177  + 647.01083331480 *T)
    sumR0  +=       0.017 * cos(2.83124480600  + 132658.27281205778 *T)
    sumR0  +=       0.018 * cos(2.89808596608  + 26091.83529483729 *T)
    sumR0  +=       0.017 * cos(5.67044617806  + 25788.77674730500 *T)
    sumR0  +=       0.017 * cos(5.37178978632  + 78731.67441507700 *T)
    sumR0  +=       0.015 * cos(5.25910515843  + 150244.34299945380 *T)
    sumR0  +=       0.019 * cos(1.72681089737  + 153.77881048480 *T)
    sumR0  +=       0.015 * cos(2.16472550978  + 26222.01219115920 *T)
    sumR0  +=       0.015 * cos(1.13029685380  + 25042.74830538660 *T)
    sumR0  +=       0.015 * cos(2.31335544105  + 51852.30086649099 *T)
    sumR0  +=       0.019 * cos(4.38254962804  + 24182.43837663380 *T)
    sumR0  +=       0.015 * cos(5.16465400786  + 27684.08958385880 *T)
    sumR0  +=       0.017 * cos(3.97954718583  + 207643.84320240439 *T)
    sumR0  +=       0.016 * cos(5.74057898460  + 25352.02662804239 *T)
    sumR0  +=       0.014 * cos(3.42441202713  + 26823.77965510599 *T)
    sumR0  +=       0.014 * cos(5.30168284837  + 742.99006053260 *T)
    sumR0  +=       0.014 * cos(2.50431306810  + 32132.13172294960 *T)
    sumR0  +=       0.014 * cos(1.36632464908  + 23919.14265929160 *T)
    sumR0  +=       0.015 * cos(6.07490514430  + 78270.33798362259 *T)
    sumR0  +=       0.019 * cos(4.40557078960  + 49953.94964855139 *T)
    sumR0  +=       0.014 * cos(0.65475293554  + 323.50541665740 *T)
    sumR0  +=       0.014 * cos(0.47093305165  + 17893.62780836560 *T)
    sumR0  +=       0.014 * cos(1.33029112501  + 129799.61842155698 *T)
    sumR0  +=       0.017 * cos(6.02915345940  + 26667.59072857300 *T)
    sumR0  +=       0.014 * cos(1.17588054181  + 104275.34649502118 *T)
    sumR0  +=       0.014 * cos(4.84032377923  + 2221.85663459700 *T)
    sumR0  +=       0.014 * cos(5.53694881682  + 77630.92568540938 *T)
    sumR0  +=       0.013 * cos(5.59791238730  + 209812.60368468691 *T)
    sumR0  +=       0.013 * cos(0.84042148682  + 173511.41499690176 *T)
    sumR0  +=       0.015 * cos(3.96416547461  + 51432.81622261579 *T)
    sumR0  +=       0.015 * cos(3.17020350946  + 176953.98994186718 *T)
    sumR0  +=       0.014 * cos(2.91048815444  + 78257.08086582259 *T)
    sumR0  +=       0.013 * cos(5.34915944924  + 80174.90890793978 *T)
    sumR0  +=       0.018 * cos(1.65795375577  + 28256.66362385679 *T)
    sumR0  +=       0.013 * cos(3.34629064183  + 25508.21555457540 *T)
    sumR0  +=       0.016 * cos(1.32228458025  + 27669.86248985719 *T)
    sumR0  +=       0.013 * cos(1.38990649917  + 25024.58905812189 *T)
    sumR0  +=       0.017 * cos(2.01249037695  + 27037.07875054399 *T)
    sumR0  +=       0.015 * cos(1.99208068531  + 52381.99183158559 *T)
    sumR0  +=       0.013 * cos(2.32883800921  + 27005.83342755599 *T)
    sumR0  +=       0.014 * cos(3.86914390591  + 182085.63102592481 *T)
    sumR0  +=       0.012 * cos(3.97054224711  + 78188.92782615528 *T)
    sumR0  +=       0.016 * cos(5.59174105029  + 25446.48957983520 *T)
    sumR0  +=       0.014 * cos(0.70490117614  + 172402.03644480839 *T)
    sumR0  +=       0.012 * cos(1.54422590258  + 52252.07235442399 *T)
    sumR0  +=       0.014 * cos(3.47518139481  + 27177.61193276780 *T)
    sumR0  +=       0.012 * cos(5.46223056321  + 107794.18751126219 *T)
    sumR0  +=       0.012 * cos(5.76552608362  + 50167.24874398939 *T)
    sumR0  +=       0.016 * cos(2.18648993322  + 26308.31578401300 *T)
    sumR0  +=       0.014 * cos(2.72148386686  + 27311.72098235281 *T)
    sumR0  +=       0.013 * cos(0.55524977232  + 75930.51303185058 *T)
    sumR0  +=       0.012 * cos(2.66601602132  + 137210.22630911658 *T)
    sumR0  +=       0.011 * cos(0.61596543052  + 44295.71712980940 *T)
    sumR0  +=       0.011 * cos(4.76091715459  + 155571.81910783658 *T)
    sumR0  +=       0.014 * cos(5.93896401871  + 52290.24557183361 *T)
    sumR0  +=       0.011 * cos(0.92596625175  + 22625.65843570900 *T)
    sumR0  +=       0.012 * cos(6.19273795549  + 51653.22886505459 *T)
    sumR0  +=       0.014 * cos(0.79969658546  + 25953.79409198919 *T)
    sumR0  +=       0.011 * cos(2.82093111356  + 77101.23472031478 *T)
    sumR0  +=       0.011 * cos(1.26994666311  + 26312.24793727610 *T)
    sumR0  +=       0.013 * cos(2.88923696437  + 112545.88789950538 *T)
    sumR0  +=       0.011 * cos(6.05528757595  + 27044.19229754480 *T)
    sumR0  +=       0.012 * cos(3.29626000869  + 52061.36699446317 *T)
    sumR0  +=       0.011 * cos(3.58945976808  + 52712.61079524379 *T)
    sumR0  +=       0.014 * cos(2.79036507500  + 23888.81579828719 *T)
    sumR0  +=       0.011 * cos(4.15810059578  + 45290.65639321759 *T)
    sumR0  +=       0.011 * cos(2.94979930726  + 58857.03113654799 *T)
    sumR0  +=       0.012 * cos(2.20589728666  + 31415.37924995700 *T)
    sumR0  +=       0.012 * cos(1.34841658603  + 51969.62073471119 *T)
    sumR0  +=       0.011 * cos(3.29585795787  + 28736.35796704720 *T)
    sumR0  +=       0.011 * cos(0.43538573957  + 24448.83362438620 *T)
    sumR0  +=       0.010 * cos(2.85999162474  + 130226.21661243298 *T)
    sumR0  +=       0.013 * cos(4.41880429822  + 19958.60610206760 *T)
    sumR0  +=       0.012 * cos(5.96394058241  + 26083.97098831109 *T)
    sumR0  +=       0.010 * cos(5.44297146700  + 163298.12945069079 *T)
    sumR0  +=       0.012 * cos(1.57767582029  + 91805.13062721380 *T)
    sumR0  +=       0.010 * cos(3.83536418151  + 51226.63067417859 *T)
    sumR0  +=       0.013 * cos(0.34187342138  + 157057.10981453978 *T)
    sumR0  +=       0.010 * cos(1.12787053523  + 78903.60671103658 *T)
    sumR0  +=       0.010 * cos(4.67818130014  + 2111.65031337760 *T)
    sumR0  +=       0.012 * cos(3.56391619799  + 77795.74443436819 *T)
    sumR0  +=       0.011 * cos(5.80720491261  + 26102.13023557580 *T)
    sumR0  +=       0.012 * cos(5.80252382200  + 27972.80430499159 *T)
    sumR0  +=       0.011 * cos(2.47026109919  + 26010.15259759030 *T)
    sumR0  +=       0.013 * cos(4.60245036657  + 25032.45336464809 *T)
    sumR0  +=       0.010 * cos(1.14221106164  + 9384.84100807520 *T)
    sumR0  +=       0.013 * cos(5.56953201474  + 22909.75735100660 *T)
    sumR0  +=       0.010 * cos(4.35806585483  + 157586.80077963436 *T)
    sumR0  +=       0.010 * cos(1.11553298844  + 65831.66677432480 *T)
    sumR0  +=       0.011 * cos(4.67108617661  + 50064.15596977079 *T)
    sumR0  +=       0.011 * cos(2.80726281395  + 81591.84508100279 *T)
    sumR0  +=       0.009 * cos(0.47353753060  + 128106.31931499895 *T)
    sumR0  +=       0.009 * cos(1.42459829123  + 104505.39137678158 *T)
    sumR0  +=       0.009 * cos(2.37565722096  + 78896.49316403578 *T)
    sumR0  +=       0.009 * cos(4.84820485208  + 24918.31489016559 *T)
    sumR0  +=       0.011 * cos(6.23834008870  + 30171.89164827860 *T)
    sumR0  +=       0.011 * cos(1.65565660037  + 78039.36462902068 *T)
    sumR0  +=       0.009 * cos(3.65734666897  + 151199.94274106238 *T)
    sumR0  +=       0.009 * cos(5.40966654859  + 130459.18546877075 *T)
    sumR0  +=       0.010 * cos(0.87714716556  + 78366.80219894118 *T)
    sumR0  +=       0.009 * cos(3.04701773654  + 51329.72344839720 *T)
    sumR0  +=       0.009 * cos(1.96005885441  + 26575.53789282839 *T)
    sumR0  +=       0.010 * cos(1.18004009296  + 50270.34151820800 *T)
    sumR0  +=       0.009 * cos(5.48583841069  + 138319.60486120995 *T)
    sumR0  +=       0.010 * cos(3.21154218694  + 77307.42026875199 *T)
    sumR0  +=       0.009 * cos(5.55407965409  + 8194.27533320860 *T)
    sumR0  +=       0.008 * cos(1.35584545211  + 12546.48193908300 *T)
    sumR0  +=       0.009 * cos(0.42023516699  + 27223.58001346740 *T)
    sumR0  +=       0.010 * cos(1.76704509660  + 77726.90491262719 *T)
    sumR0  +=       0.009 * cos(1.60501838348  + 77616.69859140778 *T)
    sumR0  +=       0.009 * cos(5.46388323313  + 16703.06213349900 *T)
    sumR0  +=       0.009 * cos(1.60520912676  + 117893.03376878797 *T)
    sumR0  +=       0.009 * cos(6.23528739890  + 846.08283475120 *T)
    sumR0  +=       0.009 * cos(5.66619352199  + 50910.23880452200 *T)
    sumR0  +=       0.009 * cos(1.18753574623  + 104197.83375581198 *T)
    sumR0  +=       0.009 * cos(0.69327774851  + 26709.64694241340 *T)
    sumR0  +=       0.010 * cos(5.48870132496  + 94329.77528619740 *T)
    sumR0  +=       0.008 * cos(1.17006056902  + 78160.61665050399 *T)
    sumR0  +=       0.009 * cos(0.49965065220  + 27573.19384827740 *T)
    sumR0  +=       0.011 * cos(6.20705720585  + 28102.88481337200 *T)
    sumR0  +=       0.011 * cos(0.75657533913  + 128320.75184749260 *T)
    sumR0  +=       0.008 * cos(3.86430794820  + 103917.90082841998 *T)
    sumR0  +=       0.009 * cos(3.64580324894  + 76784.84285065818 *T)
    sumR0  +=       0.008 * cos(4.52722085618  + 27463.67694142000 *T)
    sumR0  +=       0.009 * cos(1.73726961641  + 28791.51929624980 *T)
    sumR0  +=       0.011 * cos(1.86762532326  + 214364.55718174577 *T)
    sumR0  +=       0.008 * cos(3.17547134507  + 50903.12525752120 *T)
    sumR0  +=       0.008 * cos(6.23140491980  + 148.07872442630 *T)
    sumR0  +=       0.008 * cos(0.97646930359  + 19202.75325164339 *T)
    sumR0  +=       0.008 * cos(0.87651427964  + 151.04766984290 *T)
    sumR0  +=       0.009 * cos(1.55645109653  + 26126.03617721200 *T)
    sumR0  +=       0.008 * cos(3.12232090039  + 22003.91463486980 *T)
    sumR0  +=       0.008 * cos(3.45204671122  + 24203.00197815680 *T)
    sumR0  +=       0.008 * cos(2.50177739941  + 51955.39364070959 *T)
    sumR0  +=       0.008 * cos(4.20326276911  + 50380.54783942739 *T)
    sumR0  +=       0.008 * cos(2.66681931912  + 130432.40216087016 *T)
    sumR0  +=       0.010 * cos(2.61302801987  + 27170.98337386779 *T)
    sumR0  +=       0.008 * cos(6.15791792940  + 37698.45509994840 *T)
    sumR0  +=       0.008 * cos(5.36718741692  + 35077.37084700659 *T)
    sumR0  +=       0.008 * cos(0.39542025081  + 183570.92173262799 *T)
    sumR0  +=       0.008 * cos(4.37576068847  + 123200.84011627098 *T)
    sumR0  +=       0.008 * cos(3.44870239791  + 26402.08932144380 *T)
    sumR0  +=       0.007 * cos(5.88380837195  + 51868.24866217880 *T)
    sumR0  +=       0.007 * cos(4.80640660702  + 26118.23000257860 *T)
    sumR0  +=       0.008 * cos(5.82773614091  + 3328.13565628019 *T)
    sumR0  +=       0.009 * cos(2.53209611805  + 25227.59321282139 *T)
    sumR0  +=       0.010 * cos(2.52793236150  + 35833.22369743080 *T)
    sumR0  +=       0.008 * cos(4.11204744654  + 53906.92863608099 *T)
    sumR0  +=       0.008 * cos(1.58745297967  + 101011.00013959700 *T)
    sumR0  +=       0.009 * cos(2.57384727244  + 2125.87740737920 *T)
    sumR0  +=       0.007 * cos(2.46084982092  + 52137.67324751059 *T)
    sumR0  +=       0.008 * cos(5.56577535390  + 50689.82616208319 *T)
    sumR0  +=       0.007 * cos(5.98112141065  + 52179.73843641149 *T)
    sumR0  +=       0.010 * cos(5.16567712694  + 55638.05098901359 *T)
    sumR0  +=       0.008 * cos(4.67643518278  + 76887.93562487679 *T)
    sumR0  +=       0.007 * cos(4.36322743599  + 155887.52156313116 *T)
    sumR0  +=       0.009 * cos(0.68598593831  + 38.13303563780 *T)
    sumR0  +=       0.008 * cos(0.46194680162  + 181026.24909573561 *T)
    sumR0  +=       0.008 * cos(0.42442705032  + 182188.72380014337 *T)
    sumR0  +=       0.009 * cos(3.57254516611  + 1265.56747862640 *T)
    sumR0  +=       0.009 * cos(1.65553917926  + 65717.22748563958 *T)
    sumR0  +=       0.010 * cos(3.53066047636  + 2008.55753915900 *T)
    sumR0  +=       0.007 * cos(3.89373397550  + 1485.98012106520 *T)
    sumR0  +=       0.010 * cos(0.82550269560  + 39763.43339365039 *T)
    sumR0  +=       0.007 * cos(4.33817871844  + 129373.02023068098 *T)
    sumR0  +=       0.007 * cos(3.41340083585  + 52483.36390411799 *T)
    sumR0  +=       0.008 * cos(1.23992323880  + 76041.85279012559 *T)
    sumR0  +=       0.007 * cos(0.04567921269  + 206.18554843720 *T)
    sumR0  +=       0.007 * cos(2.56169701158  + 207593.84658050016 *T)
    sumR0  +=       0.009 * cos(1.97148507623  + 365230.64398203877 *T)
    sumR0  +=       0.007 * cos(3.14813276839  + 52061.40717623499 *T)
    sumR0  +=       0.007 * cos(4.86954611488  + 27351.06050415239 *T)
    sumR0  +=       0.008 * cos(6.27520424489  + 204151.27163553479 *T)
    sumR0  +=       0.007 * cos(5.46143749214  + 25985.94062330859 *T)
    sumR0  +=       0.008 * cos(4.44335543148  + 26057.57628056979 *T)
    sumR0  +=       0.007 * cos(1.00677483843  + 129387.24732468258 *T)
    sumR0  +=       0.008 * cos(0.22314041177  + 50049.92887576919 *T)
    sumR0  +=       0.007 * cos(3.98224449185  + 13675.53025207620 *T)
    sumR0  +=       0.007 * cos(4.34489404354  + 167850.08294774959 *T)
    sumR0  +=       0.009 * cos(0.71471851401  + 233731.74634397856 *T)
    sumR0  +=       0.007 * cos(1.69935809006  + 54087.00576636560 *T)
    sumR0  +=       0.007 * cos(6.21044393135  + 203041.89308344136 *T)
    sumR0  +=       0.007 * cos(2.53524305485  + 80382.47328510099 *T)
    sumR0  +=       0.007 * cos(0.03218225075  + 31281.27020037200 *T)
    sumR0  +=       0.007 * cos(5.26037977949  + 78338.49102328988 *T)
    sumR0  +=       0.007 * cos(3.52744190481  + 52509.66222391780 *T)
    sumR0  +=       0.007 * cos(5.24771876203  + 1795.25844372100 *T)
    sumR0  +=       0.007 * cos(5.20875856726  + 161079.37234650398 *T)
    sumR0  +=       0.007 * cos(3.19471628666  + 24072.92146977640 *T)
    sumR0  +=       0.006 * cos(1.58096793180  + 143980.93691036216 *T)
    sumR0  +=       0.009 * cos(0.70414754528  + 25169.97285559240 *T)
    sumR0  +=       0.006 * cos(0.15264531683  + 52911.68279668019 *T)
    sumR0  +=       0.007 * cos(4.28578455720  + 51130.65144696079 *T)
    sumR0  +=       0.007 * cos(3.42650023556  + 29396.36903695460 *T)
    sumR0  +=       0.007 * cos(4.12959254769  + 52174.32181044009 *T)
    sumR0  +=       0.008 * cos(3.79812131071  + 137678.19129947099 *T)
    sumR0  +=       0.007 * cos(5.70602822567  + 9745.32055585660 *T)
    sumR0  +=       0.008 * cos(0.14316432919  + 136722.59155786238 *T)
    sumR0  +=       0.007 * cos(3.78832454636  + 26189.86565983980 *T)
    sumR0  +=       0.007 * cos(5.44664616173  + 27566.76971563859 *T)
    sumR0  +=       0.006 * cos(0.53269408847  + 208173.53416749896 *T)
    sumR0  +=       0.007 * cos(3.95192199993  + 145204.75475114080 *T)
    sumR0  +=       0.006 * cos(1.80610265750  + 26830.89320210680 *T)
    sumR0  +=       0.008 * cos(5.80105964058  + 25773.71696170459 *T)
    sumR0  +=       0.006 * cos(1.20705629404  + 26734.91397488900 *T)
    sumR0  +=       0.008 * cos(3.87508349031  + 52041.69723356339 *T)
    sumR0  +=       0.007 * cos(1.87429126900  + 1905.46476494040 *T)
    sumR0  +=       0.007 * cos(5.68938243001  + 158746.17595363196 *T)
    sumR0  +=       0.008 * cos(3.39884199213  + 78153.50310350319 *T)
    sumR0  +=       0.007 * cos(1.20867598891  + 61560.64729122359 *T)
    sumR0  +=       0.008 * cos(3.30580293572  + 130419.84594697120 *T)
    sumR0  +=       0.007 * cos(0.39338960799  + 2199.08734328700 *T)
    sumR0  +=       0.006 * cos(3.17982806365  + 77947.31755506598 *T)
    sumR0  +=       0.008 * cos(0.88585179650  + 220.41264243880 *T)
    sumR0  +=       0.008 * cos(1.46508855917  + 87367.61641884019 *T)
    sumR0  +=       0.006 * cos(0.60809363482  + 118828.96374949680 *T)
    sumR0  +=       0.008 * cos(5.53182088929  + 53814.87580033639 *T)
    sumR0  +=       0.007 * cos(1.65646088833  + 196137.07343261081 *T)
    sumR0  +=       0.006 * cos(5.74552748087  + 55503.94193942859 *T)
    sumR0  +=       0.006 * cos(1.43996863707  + 2333.19639287200 *T)
    sumR0  +=       0.006 * cos(1.91186990644  + 130652.81480330898 *T)
    sumR0  +=       0.006 * cos(5.06448767269  + 50444.68393021580 *T)
    sumR0  +=       0.006 * cos(1.09838300878  + 183674.70392120857 *T)
    sumR0  +=       0.007 * cos(0.35338852938  + 103932.12792242158 *T)
    sumR0  +=       0.006 * cos(0.28924808590  + 54374.89362643540 *T)
    sumR0  +=       0.006 * cos(2.72829193454  + 76255.15188556358 *T)
    sumR0  +=       0.006 * cos(3.92290484480  + 130289.95251073640 *T)
    sumR0  +=       0.007 * cos(1.27402494790  + 46046.50924364180 *T)
    sumR0  +=       0.007 * cos(2.96149281662  + 52755.49387014719 *T)
    sumR0  +=       0.008 * cos(2.23784693584  + 25466.15934073500 *T)
    sumR0  +=       0.006 * cos(1.78075029765  + 78786.28684281638 *T)
    sumR0  +=       0.008 * cos(4.71972114899  + 25014.29411738339 *T)
    sumR0  +=       0.006 * cos(5.52798470860  + 103498.41618454478 *T)
    sumR0  +=       0.008 * cos(5.43345902063  + 949.17560896980 *T)
    sumR0  +=       0.006 * cos(5.55571041746  + 130446.62925487180 *T)
    sumR0  +=       0.007 * cos(5.36488899884  + 77940.20400806518 *T)
    sumR0  +=       0.006 * cos(0.52080930820  + 25004.82290928060 *T)
    sumR0  +=       0.006 * cos(3.48878406062  + 123668.80510662538 *T)
    sumR0  +=       0.008 * cos(3.02023688443  + 26076.85744131030 *T)
    sumR0  +=       0.008 * cos(5.31586228888  + 24388.62392507099 *T)
    sumR0  +=       0.006 * cos(5.06257578673  + 78469.89497315978 *T)
    sumR0  +=       0.006 * cos(4.79504045120  + 52177.29075585669 *T)
    sumR0  +=       0.006 * cos(4.57296057068  + 35211.47989659159 *T)
    sumR0  +=       0.006 * cos(5.37176342301  + 53093.73656913019 *T)
    sumR0  +=       0.006 * cos(4.26973755729  + 130363.24963659538 *T)
    sumR0  +=       0.006 * cos(5.75364706874  + 78580.10129437919 *T)
    sumR0  +=       0.006 * cos(1.99314462977  + 130435.63437251298 *T)
    sumR0  +=       0.006 * cos(1.55918658153  + 39450.35284837340 *T)
    sumR0  +=       0.008 * cos(4.77256284659  + 142871.55835826878 *T)
    sumR0  +=       0.006 * cos(6.09530056324  + 103395.32341032618 *T)
    sumR0  +=       0.006 * cos(1.29134609458  + 102975.83876645098 *T)
    sumR0  +=       0.006 * cos(1.83917095242  + 24402.85101907260 *T)
    sumR0  +=       0.005 * cos(3.63216834248  + 43981.53094993980 *T)
    sumR0  +=       0.005 * cos(5.64850202903  + 28774.62523611540 *T)
    sumR0  +=       0.007 * cos(4.22124824964  + 26098.94884183810 *T)
    sumR0  +=       0.005 * cos(5.85936922514  + 156314.11975400720 *T)
    sumR0  +=       0.005 * cos(4.97283396704  + 102659.44689679438 *T)
    sumR0  +=       0.005 * cos(1.46473339833  + 181659.72224941081 *T)
    sumR0  +=       0.006 * cos(0.71584342921  + 104276.83096772947 *T)
    sumR0  +=       0.007 * cos(2.49362242764  + 51439.92976961659 *T)
    sumR0  +=       0.006 * cos(0.02944876240  + 20043.67456019880 *T)
    sumR0  +=       0.006 * cos(2.19480801460  + 12412.37288949800 *T)
    sumR0  +=       0.007 * cos(3.28347851449  + 183145.01295611399 *T)
    sumR0  +=       0.005 * cos(3.46807375357  + 50536.73676596039 *T)
    sumR0  +=       0.006 * cos(6.20804958899  + 114.39910691340 *T)
    sumR0  +=       0.006 * cos(2.26035174588  + 103718.82882698360 *T)
    sumR0  +=       0.007 * cos(3.65318062648  + 143005.66740785379 *T)
    sumR0  +=       0.005 * cos(0.12859263642  + 51596.11869614960 *T)
    sumR0  +=       0.007 * cos(0.75531928973  + 77520.71936418998 *T)
    sumR0  +=       0.005 * cos(1.52974136650  + 26724.89941359840 *T)
    sumR0  +=       0.007 * cos(3.81960920375  + 62197.64356324779 *T)
    sumR0  +=       0.005 * cos(5.08212104486  + 52609.51802102519 *T)
    sumR0  +=       0.005 * cos(4.49337901115  + 50007.04580086580 *T)
    sumR0  +=       0.005 * cos(5.03210861040  + 1045.15483618760 *T)
    sumR0  +=       0.005 * cos(6.27674019656  + 1272.68102562720 *T)
    sumR0  +=       0.006 * cos(3.84313039676  + 81604.32185142238 *T)
    sumR0  +=       0.006 * cos(3.27974922227  + 144916.86689107097 *T)
    sumR0  +=       0.005 * cos(0.50987626702  + 299.12639426920 *T)
    sumR0  +=       0.006 * cos(5.10596879808  + 78313.70604662679 *T)
    sumR0  +=       0.005 * cos(4.14575201906  + 53867.97195265079 *T)
    sumR0  +=       0.007 * cos(4.53402175332  + 166740.70439565618 *T)
    sumR0  +=       0.005 * cos(5.57648480152  + 106262.81204951400 *T)
    sumR0  +=       0.007 * cos(3.08562485832  + 52027.72755872209 *T)
    sumR0  +=       0.006 * cos(2.81224506925  + 25450.90686955000 *T)
    sumR0  +=       0.005 * cos(1.51400872659  + 28199.55345495179 *T)
    sumR0  +=       0.006 * cos(1.58495517428  + 47803.92991637420 *T)
    sumR0  +=       0.006 * cos(2.79561551205  + 52286.01260436779 *T)
    sumR0  +=       0.007 * cos(1.80630878142  + 23549.65463732060 *T)
    sumR0  +=       0.007 * cos(3.72396318065  + 198489.93958638256 *T)
    sumR0  +=       0.005 * cos(5.74129147264  + 103189.13786188900 *T)
    sumR0  +=       0.006 * cos(2.94825789213  + 56259.79478985279 *T)
    sumR0  +=       0.005 * cos(4.96546514085  + 78413.27262185719 *T)
    sumR0  +=       0.005 * cos(1.78518845730  + 129586.31932611900 *T)
    sumR0  +=       0.005 * cos(0.31701812829  + 78800.51393681798 *T)
    sumR0  +=       0.005 * cos(0.55478931159  + 52817.21984488739 *T)
    sumR0  +=       0.006 * cos(1.18448496437  + 25600.26839032000 *T)
    sumR0  +=       0.005 * cos(0.42862127579  + 179172.74704605396 *T)
    sumR0  +=       0.006 * cos(3.74336258475  + 26247.20487636640 *T)
    sumR0  +=       0.005 * cos(2.46583763315  + 123758.29085447140 *T)
    sumR0  +=       0.006 * cos(5.09461941443  + 53124.98189211819 *T)
    sumR0  +=       0.006 * cos(0.42320281908  + 130443.39704322898 *T)
    sumR0  +=       0.006 * cos(2.57896380153  + 26014.60601571519 *T)
    sumR0  +=       0.005 * cos(2.23635663946  + 24513.05734029200 *T)
    sumR0  +=       0.005 * cos(5.98752668018  + 104344.98400739678 *T)
    sumR0  +=       0.006 * cos(0.32131472009  + 102872.74599223241 *T)
    sumR0  +=       0.005 * cos(1.95885559875  + 53771.99272543299 *T)
    sumR0  +=       0.007 * cos(2.16257501519  + 52024.75861330549 *T)
    sumR0  +=       0.006 * cos(1.37111815532  + 51120.35650622229 *T)
    sumR0  +=       0.005 * cos(0.15128494504  + 78683.19406859778 *T)
    sumR0  +=       0.005 * cos(4.00810087423  + 48713.56157728320 *T)
    sumR0  +=       0.006 * cos(1.45883115780  + 76152.05911134499 *T)
    sumR0  +=       0.005 * cos(1.72421399237  + 28309.75977617119 *T)
    sumR0  +=       0.005 * cos(0.71424864998  + 102755.42612401220 *T)
    sumR0  +=       0.006 * cos(0.03283607980  + 78149.27013603736 *T)
    sumR0  +=       0.006 * cos(2.69429372458  + 78378.14871340781 *T)
    sumR0  +=       0.006 * cos(1.12570261398  + 25647.07785669659 *T)
    sumR0  +=       0.005 * cos(3.60259399819  + 11.04570026390 *T)
    sumR0  +=       0.006 * cos(4.83171247280  + 240452.46032332000 *T)
    sumR0  +=       0.005 * cos(3.71949948652  + 26411.40855823160 *T)
    sumR0  +=       0.005 * cos(4.55172986205  + 79330.20490191258 *T)
    sumR0  +=       0.005 * cos(0.84232799985  + 24551.92639860479 *T)
    sumR0  +=       0.005 * cos(3.74067330742  + 52797.55008398759 *T)
    sumR0  +=       0.005 * cos(2.92277012561  + 54190.78795494619 *T)
    sumR0  +=       0.005 * cos(3.96038779086  + 25657.37279743509 *T)
    sumR0  +=       0.005 * cos(4.54502031861  + 102769.65321801380 *T)
    sumR0  +=       0.006 * cos(2.35342867622  + 27665.24668402200 *T)
    sumR0  +=       0.005 * cos(3.72877345305  + 78213.71280281838 *T)
    sumR0  +=       0.005 * cos(6.19447408823  + 26191.68533015479 *T)
    sumR0  +=       0.004 * cos(0.53733592588  + 77314.53381575279 *T)
    sumR0  +=       0.004 * cos(1.39917225370  + 26290.15653674829 *T)
    sumR0  +=       0.005 * cos(4.85332036328  + 54344.56676543099 *T)
    sumR0  +=       0.006 * cos(0.92030946706  + 23336.35554188260 *T)
    sumR0  +=       0.006 * cos(0.37487337241  + 53265.51507434200 *T)
    sumR0  +=       0.004 * cos(2.01983013209  + 156547.08861034497 *T)
    sumR0  +=       0.005 * cos(6.13113074115  + 26026.17716683400 *T)
    sumR0  +=       0.004 * cos(4.94928690334  + 52663.44103440259 *T)
    sumR0  +=       0.005 * cos(0.40443457223  + 28096.46068073319 *T)
    sumR0  +=       0.004 * cos(4.40405080911  + 130593.29451835580 *T)
    sumR0  +=       0.004 * cos(4.17440698686  + 91919.56991589900 *T)
    sumR0  +=       0.005 * cos(3.71777436660  + 70383.62027138360 *T)
    sumR0  +=       0.005 * cos(1.52885825328  + 51102.19725895759 *T)
    sumR0  +=       0.004 * cos(1.71609047005  + 23762.95373275860 *T)
    sumR0  +=       0.004 * cos(0.55287089413  + 26933.98597632540 *T)
    sumR0  +=       0.005 * cos(6.02225706023  + 128220.75860368418 *T)
    sumR0  +=       0.005 * cos(1.78226900229  + 187167.27548807816 *T)
    sumR0  +=       0.005 * cos(3.32000517354  + 45424.76544280259 *T)
    sumR0  +=       0.005 * cos(4.64166221826  + 24824.74577899600 *T)
    sumR0  +=       0.005 * cos(3.45997061965  + 48847.67062686820 *T)
    sumR0  +=       0.006 * cos(2.62272580920  + 52171.87412988529 *T)
    sumR0  +=       0.004 * cos(3.62789650971  + 53029.00266490040 *T)
    sumR0  +=       0.006 * cos(2.53652345505  + 52190.03337714999 *T)
    sumR0  +=       0.004 * cos(0.08235213515  + 58220.03486452380 *T)
    sumR0  +=       0.006 * cos(2.56621666334  + 51876.67988887919 *T)
    sumR0  +=       0.005 * cos(5.04323059321  + 41494.55574658299 *T)
    sumR0  +=       0.004 * cos(3.42433134364  + 53311.48315504160 *T)
    sumR0  +=       0.005 * cos(5.60899913389  + 1911.19948321720 *T)
    sumR0  +=       0.004 * cos(1.13990612191  + 26183.88236879199 *T)
    sumR0  +=       0.004 * cos(4.24142693734  + 189853.99758261937 *T)
    sumR0  +=       0.004 * cos(0.21521651200  + 19336.86230122839 *T)
    sumR0  +=       0.005 * cos(0.30306745520  + 103883.64757594238 *T)
    sumR0  +=       0.004 * cos(6.09719517639  + 77417.62658997139 *T)
    sumR0  +=       0.004 * cos(3.38148734662  + 316.39186965660 *T)
    sumR0  +=       0.004 * cos(1.40344660688  + 25885.64974640009 *T)
    sumR0  +=       0.006 * cos(2.54556610763  + 54060.70744656579 *T)
    sumR0  +=       0.005 * cos(4.62289027765  + 78339.97549599818 *T)
    sumR0  +=       0.004 * cos(6.06868163269  + 84944.93427812219 *T)
    sumR0  +=       0.005 * cos(0.64970018860  + 3178.14579056760 *T)
    sumR0  +=       0.004 * cos(5.48109256360  + 78225.57638908479 *T)
    sumR0  +=       0.004 * cos(3.76404568854  + 153084.84390447981 *T)
    sumR0  +=       0.004 * cos(0.18748530762  + 52489.99246301800 *T)
    sumR0  +=       0.005 * cos(1.74044114280  + 79315.97780791098 *T)
    sumR0  +=       0.004 * cos(6.06406108213  + 52161.57918914679 *T)
    sumR0  +=       0.005 * cos(0.28595106407  + 26161.20026743319 *T)
    sumR0  +=       0.004 * cos(2.11074347112  + 189386.03259226491 *T)
    sumR0  +=       0.004 * cos(0.59715530740  + 16342.58258571760 *T)
    sumR0  +=       0.004 * cos(3.91091220587  + 101703.15774082378 *T)
    sumR0  +=       0.004 * cos(1.16130290140  + 50264.60679993120 *T)
    sumR0  +=       0.004 * cos(5.51188515445  + 78043.29678228378 *T)
    sumR0  +=       0.005 * cos(4.39350322600  + 53757.76563143139 *T)
    sumR0  +=       0.004 * cos(5.15437140585  + 24285.53115085240 *T)
    sumR0  +=       0.005 * cos(2.15280343086  + 48997.66049258080 *T)
    sumR0  +=       0.004 * cos(5.35316514819  + 104984.39630560996 *T)
    sumR0  +=       0.005 * cos(5.75390640070  + 51841.95034237900 *T)
    sumR0  +=       0.004 * cos(1.52206374281  + 26267.98027185880 *T)
    sumR0  +=       0.004 * cos(4.66563868021  + 104127.26777059487 *T)
    sumR0  +=       0.004 * cos(2.36671017385  + 19.66976089979 *T)
    sumR0  +=       0.004 * cos(5.10954286080  + 31722.93687092660 *T)
    sumR0  +=       0.004 * cos(1.54518951672  + 52206.13314415280 *T)
    sumR0  +=       0.004 * cos(4.16791904155  + 130285.73689738620 *T)
    sumR0  +=       0.004 * cos(2.93299901703  + 31775.53351052620 *T)
    sumR0  +=       0.005 * cos(4.47247592301  + 51112.49219969609 *T)
    sumR0  +=       0.004 * cos(2.41568769208  + 51534.39272140940 *T)
    sumR0  +=       0.004 * cos(1.74441095464  + 26248.31051095900 *T)
    sumR0  +=       0.004 * cos(5.23856038893  + 52396.21892558720 *T)
    sumR0  +=       0.004 * cos(4.12254586219  + 27441.65188659100 *T)
    sumR0  +=       0.005 * cos(5.17529407091  + 19367.18916223280 *T)
    sumR0  +=       0.004 * cos(4.21064422396  + 26279.35140768579 *T)
    sumR0  +=       0.005 * cos(4.41570881449  + 78057.52387628538 *T)
    sumR0  +=       0.004 * cos(5.30327887066  + 25672.35065096209 *T)
    sumR0  +=       0.005 * cos(1.10382058499  + 31903.01400121119 *T)
    sumR0  +=       0.005 * cos(2.71619626622  + 24952.22626968100 *T)
    sumR0  +=       0.004 * cos(2.28793023586  + 235900.50682626120 *T)
    sumR0  +=       0.004 * cos(1.58884724433  + 128747.35003836860 *T)
    sumR0  +=       0.004 * cos(5.78862899204  + 22779.43724619380 *T)
    sumR0  +=       0.004 * cos(4.22021430561  + 1731.12235293260 *T)
    sumR0  +=       0.004 * cos(4.68318371027  + 103814.80805420138 *T)
    sumR0  +=       0.004 * cos(3.29575966354  + 76137.83201734339 *T)
    sumR0  +=       0.004 * cos(5.47487283809  + 132350.71519108818 *T)
    sumR0  +=       0.004 * cos(2.99265295426  + 77741.13200662879 *T)
    sumR0  +=       0.004 * cos(3.47889610610  + 53661.09698985160 *T)
    sumR0  +=       0.004 * cos(4.76216346874  + 156740.71794488319 *T)
    sumR0  +=       0.004 * cos(1.47799681769  + 120417.67842777158 *T)
    sumR0  +=       0.004 * cos(2.78305398087  + 133767.65136415116 *T)
    sumR0  +=       0.004 * cos(2.87953962922  + 25862.07387316400 *T)
    sumR0  +=       0.004 * cos(1.92972990240  + 15406.65260500879 *T)
    sumR0  +=       0.005 * cos(1.81394930430  + 130866.11389874700 *T)
    sumR0  +=       0.004 * cos(2.62752187758  + 24602.61243487099 *T)
    sumR0  +=       0.004 * cos(3.12567603415  + 20894.53608277640 *T)
    sumR0  +=       0.004 * cos(1.00432915295  + 149288.74325784517 *T)
    sumR0  +=       0.004 * cos(3.47489986602  + 26114.20146137400 *T)
    sumR0  +=       0.004 * cos(4.27251739299  + 104991.50985261080 *T)
    sumR0  +=       0.004 * cos(2.20022040184  + 52073.84376488279 *T)
    sumR0  +=       0.004 * cos(3.77568584773  + 14.22709400160 *T)
    sumR0  +=       0.004 * cos(1.65179392234  + 51006.21803173979 *T)
    sumR0  +=       0.004 * cos(0.72729563393  + 25865.04281858059 *T)
    sumR0  +=       0.004 * cos(5.61140287619  + 53258.88651544199 *T)
    sumR0  +=       0.004 * cos(1.55115346866  + 53438.96364572659 *T)
    sumR0  +=       0.005 * cos(0.91157457367  + 25665.23710396129 *T)
    sumR0  +=       0.005 * cos(6.11073147567  + 156507.74908854539 *T)
    sumR0  +=       0.003 * cos(3.12711312875  + 80462.79676800959 *T)
    sumR0  +=       0.004 * cos(2.21068358470  + 76777.72930365738 *T)
    sumR0  +=       0.004 * cos(0.82472101949  + 30689.85326053720 *T)
    sumR0  +=       0.004 * cos(3.85651603357  + 26513.81191808820 *T)
    sumR0  +=       0.003 * cos(3.76311835707  + 77.75054398390 *T)
    sumR0  +=       0.004 * cos(0.76897521016  + 23384.28698689860 *T)
    sumR0  +=       0.003 * cos(4.58555781046  + 57503.28239153120 *T)
    sumR0  +=       0.004 * cos(3.82967402925  + 104454.70534051540 *T)
    sumR0  +=       0.003 * cos(2.47448927310  + 77956.15180375299 *T)
    sumR0  +=       0.004 * cos(4.93160467297  + 391318.54712361295 *T)
    sumR0  +=       0.004 * cos(1.56118199279  + 1055.44977692610 *T)
    sumR0  +=       0.004 * cos(0.77257538538  + 79994.83177765518 *T)
    sumR0  +=       0.005 * cos(3.11675458540  + 26149.62911631439 *T)
    sumR0  +=       0.003 * cos(4.05641745083  + 104248.51979207818 *T)
    sumR0  +=       0.004 * cos(2.97133561512  + 30110.16567353840 *T)
    sumR0  +=       0.004 * cos(5.60420829330  + 156520.30530244438 *T)
    sumR0  +=       0.004 * cos(5.59378669038  + 51315.49635439559 *T)
    sumR0  +=       0.003 * cos(5.33733887891  + 52309.91533273340 *T)
    sumR0  +=       0.003 * cos(5.47155908699  + 149846.19399604559 *T)
    sumR0  +=       0.003 * cos(0.19039417498  + 50290.90511973100 *T)
    sumR0  +=       0.003 * cos(0.55626990901  + 130005.80396999416 *T)
    sumR0  +=       0.004 * cos(1.22491945176  + 25927.49577218939 *T)
    sumR0  +=       0.004 * cos(1.83119228357  + 132028.58860315481 *T)
    sumR0  +=       0.004 * cos(3.78910193450  + 25970.58327335399 *T)
    sumR0  +=       0.004 * cos(4.12310730501  + 33856.26969826880 *T)
    sumR0  +=       0.004 * cos(3.29497466474  + 208276.62694171761 *T)
    sumR0  +=       0.003 * cos(1.23800163682  + 193937.98608932376 *T)
    sumR0  +=       0.004 * cos(3.81093761011  + 49424.25868345679 *T)
    sumR0  +=       0.004 * cos(6.17742034131  + 76991.02839909539 *T)
    sumR0  +=       0.003 * cos(2.37855052298  + 25459.05155552409 *T)
    sumR0  +=       0.003 * cos(0.06662548789  + 54824.26110862140 *T)
    sumR0  +=       0.003 * cos(0.00882668027  + 78571.26704569219 *T)
    sumR0  +=       0.004 * cos(3.50355655289  + 33480.35754365900 *T)
    sumR0  +=       0.004 * cos(1.42998120813  + 23446.56186310200 *T)
    sumR0  +=       0.004 * cos(4.57967744099  + 52213.93931878620 *T)
    sumR0  +=       0.004 * cos(4.30793307856  + 2538.24850425360 *T)
    sumR0  +=       0.004 * cos(2.64175466237  + 80596.90581759460 *T)
    sumR0  +=       0.004 * cos(2.23067570586  + 104819.57755665120 *T)
    sumR0  +=       0.004 * cos(2.10867686930  + 27331.39074325260 *T)
    sumR0  +=       0.003 * cos(3.74321552731  + 28471.09615635040 *T)
    sumR0  +=       0.003 * cos(5.35951703976  + 233681.74972207440 *T)
    sumR0  +=       0.004 * cos(1.44726666359  + 25907.82601128960 *T)
    sumR0  +=       0.003 * cos(0.77055249403  + 23389.45169419700 *T)
    sumR0  +=       0.004 * cos(5.53731910534  + 52098.05573916449 *T)
    sumR0  +=       0.004 * cos(2.71065844730  + 107679.74822257696 *T)
    sumR0  +=       0.003 * cos(4.85884222947  + 1063.31408345230 *T)
    sumR0  +=       0.003 * cos(6.10322230645  + 48091.81777644400 *T)
    sumR0  +=       0.003 * cos(0.57378510632  + 154194.22245657316 *T)
    sumR0  +=       0.003 * cos(4.65504051632  + 25896.45487546260 *T)
    sumR0  +=       0.003 * cos(2.07685199898  + 3.93215326310 *T)
    sumR0  +=       0.004 * cos(4.67601599622  + 18073.70493865020 *T)
    sumR0  +=       0.003 * cos(4.62237776277  + 21716.02677480000 *T)
    sumR0  +=       0.003 * cos(2.40811934972  + 76998.14194609619 *T)
    sumR0  +=       0.004 * cos(1.45941637429  + 1073.60902419080 *T)
    sumR0  +=       0.003 * cos(4.06965227350  + 32808.61712091560 *T)
    sumR0  +=       0.003 * cos(3.54974279869  + 23401.18104703300 *T)
    sumR0  +=       0.003 * cos(5.94935193305  + 6751.04084034580 *T)
    sumR0  +=       0.003 * cos(4.32590129932  + 52400.15107885030 *T)
    sumR0  +=       0.003 * cos(1.93788940079  + 45455.09230380700 *T)
    sumR0  +=       0.003 * cos(6.08303870194  + 53399.62412392700 *T)
    sumR0  +=       0.003 * cos(4.45999574391  + 127098.90328117118 *T)
    sumR0  +=       0.003 * cos(5.45176028820  + 26061.60482177439 *T)
    sumR0  +=       0.003 * cos(2.26429696537  + 27047.02600153180 *T)
    sumR0  +=       0.004 * cos(1.93735592711  + 26603.36701266720 *T)
    sumR0  +=       0.003 * cos(2.94612910423  + 27039.97976483379 *T)
    sumR0  +=       0.004 * cos(3.64314833794  + 860.30992875280 *T)
    sumR0  +=       0.004 * cos(2.04162116492  + 27623.87988454360 *T)
    sumR0  +=       0.004 * cos(1.02413610685  + 25991.92391435639 *T)
    sumR0  +=       0.004 * cos(3.72190941745  + 51257.87599716660 *T)
    sumR0  +=       0.003 * cos(0.83533616968  + 78262.22495201428 *T)
    sumR0  +=       0.004 * cos(5.84174581609  + 25650.25925043430 *T)
    sumR0  +=       0.003 * cos(3.86698088130  + 56777.75640211140 *T)
    sumR0  +=       0.004 * cos(0.94091162922  + 23735.03698780239 *T)
    sumR0  +=       0.003 * cos(2.87479269049  + 159855.55450572536 *T)
    sumR0  +=       0.004 * cos(2.92372103776  + 104358.24112519680 *T)
    sumR0  +=       0.003 * cos(5.50683793531  + 25928.60140678200 *T)
    sumR0  +=       0.003 * cos(1.50282981980  + 4371.87636677420 *T)
    sumR0  +=       0.003 * cos(2.10941795435  + 60170.33472517260 *T)
    sumR0  +=       0.003 * cos(1.71064218785  + 170068.84005193636 *T)
    sumR0  +=       0.003 * cos(1.97431548371  + 104426.39416486408 *T)
    sumR0  +=       0.003 * cos(2.96711706073  + 25984.12095299360 *T)
    sumR0  +=       0.003 * cos(1.86795145608  + 126067.63934109680 *T)
    sumR0  +=       0.003 * cos(4.74630310296  + 54879.42243782400 *T)
    sumR0  +=       0.004 * cos(3.71365034796  + 29580.47470844380 *T)
    sumR0  +=       0.003 * cos(4.20767161544  + 128843.32926558638 *T)
    sumR0  +=       0.003 * cos(0.91166163253  + 76468.45098100159 *T)
    sumR0  +=       0.003 * cos(6.20945940475  + 50160.82461135059 *T)
    sumR0  +=       0.003 * cos(6.10137777265  + 52164.76058288449 *T)
    sumR0  +=       0.003 * cos(0.31852065143  + 28632.57577846660 *T)
    sumR0  +=       0.003 * cos(2.88816579373  + 27360.58416720139 *T)
    sumR0  +=       0.003 * cos(5.29083364884  + 2751.54759969160 *T)
    sumR0  +=       0.003 * cos(0.87901414893  + 1699.27921650320 *T)
    sumR0  +=       0.003 * cos(0.12942378280  + 203375.74902421079 *T)
    sumR0  +=       0.003 * cos(1.88640533644  + 27353.47062020059 *T)
    sumR0  +=       0.003 * cos(3.77660990144  + 24787.07632336920 *T)
    sumR0  +=       0.003 * cos(5.26339958526  + 289185.69166150293 *T)
    sumR0  +=       0.003 * cos(1.34477523795  + 1169.58825140860 *T)
    sumR0  +=       0.003 * cos(0.07928565452  + 104241.40624507738 *T)
    sumR0  +=       0.003 * cos(4.47916732925  + 38634.38508065720 *T)
    sumR0  +=       0.004 * cos(1.37420794262  + 128857.55635958798 *T)
    sumR0  +=       0.003 * cos(2.08343776186  + 50476.52706664519 *T)
    sumR0  +=       0.003 * cos(4.23656246406  + 76358.24465978218 *T)
    sumR0  +=       0.003 * cos(5.54498269246  + 61921.12683900499 *T)
    sumR0  +=       0.003 * cos(3.80268685478  + 25384.26995695679 *T)
    sumR0  +=       0.003 * cos(5.22182880098  + 3462.24470586520 *T)
    sumR0  +=       0.003 * cos(0.61479556097  + 171292.65789271498 *T)
    sumR0  +=       0.003 * cos(3.12135348664  + 162810.49469943656 *T)
    sumR0  +=       0.003 * cos(4.22351187737  + 67608.75720795698 *T)
    sumR0  +=       0.003 * cos(5.30037747699  + 51554.06248230919 *T)
    sumR0  +=       0.003 * cos(4.58024970278  + 34102.10134449819 *T)
    sumR0  +=       0.003 * cos(2.07639441349  + 24072.23205541439 *T)
    sumR0  +=       0.003 * cos(4.65503825059  + 104874.18998439058 *T)
    sumR0  +=       0.003 * cos(4.53866780815  + 28908.73428570040 *T)
    sumR0  +=       0.003 * cos(1.00191312140  + 155460.92337225520 *T)
    sumR0  +=       0.003 * cos(2.05553254051  + 176332.24614102798 *T)
    sumR0  +=       0.003 * cos(3.94870291510  + 155475.15046625675 *T)
    sumR0  +=       0.003 * cos(1.43273740893  + 24719.24288872919 *T)
    sumR0  +=       0.003 * cos(0.23823986548  + 24510.55959912640 *T)
    sumR0  +=       0.003 * cos(5.21326346293  + 1639.06951718800 *T)
    sumR0  +=       0.003 * cos(2.53461972944  + 51861.62010327879 *T)
    sumR0  +=       0.003 * cos(4.13978656372  + 33917.99567300899 *T)
    sumR0  +=       0.003 * cos(4.34295741397  + 16964.32638655680 *T)
    sumR0  +=       0.003 * cos(0.28015371448  + 169093.57054942797 *T)
    sumR0  +=       0.003 * cos(2.49492450438  + 636.99627202420 *T)
    sumR0  +=       0.003 * cos(4.20183911478  + 149642.26895951436 *T)
    sumR0  +=       0.003 * cos(0.73100534966  + 35679.44488694599 *T)
    sumR0  +=       0.003 * cos(1.18977971170  + 143164.96914264598 *T)
    sumR0  +=       0.003 * cos(1.17691552034  + 52145.47942214399 *T)
    sumR0  +=       0.003 * cos(4.94963823249  + 49637.55777889479 *T)
    sumR0  +=       0.003 * cos(3.00732745961  + 24705.01579472759 *T)
    sumR0  +=       0.003 * cos(5.76323456196  + 49976.71893986139 *T)
    sumR0  +=       0.003 * cos(4.83245208446  + 156523.53751408719 *T)
    sumR0  +=       0.002 * cos(3.69719041336  + 24616.15011451060 *T)
    sumR0  +=       0.003 * cos(0.18439316147  + 78597.56536549199 *T)
    sumR0  +=       0.003 * cos(4.98373933018  + 102129.75593169978 *T)
    sumR0  +=       0.003 * cos(0.02228487734  + 1.48447270830 *T)
    sumR0  +=       0.003 * cos(4.06250318984  + 97466.46267636596 *T)
    sumR0  +=       0.003 * cos(2.65726234287  + 13705.85711308060 *T)
    sumR0  +=       0.002 * cos(0.12883868680  + 149756.70824819960 *T)
    sumR0  +=       0.003 * cos(2.82299563379  + 53132.09543911899 *T)
    sumR0  +=       0.003 * cos(5.99061509510  + 29864.33402730900 *T)
    sumR0  +=       0.003 * cos(3.93018711070  + 17098.43543614181 *T)
    sumR0  +=       0.002 * cos(4.22698039366  + 10681.25053656540 *T)
    sumR0  +=       0.002 * cos(1.31889013785  + 25572.43927048119 *T)
    sumR0  +=       0.002 * cos(3.49816744164  + 51092.72605085480 *T)
    sumR0  +=       0.003 * cos(1.52361691660  + 78265.19389743089 *T)
    sumR0  +=       0.003 * cos(0.52188463522  + 52277.76880141400 *T)
    sumR0  +=       0.003 * cos(3.80600862869  + 25124.50043860279 *T)
    sumR0  +=       0.003 * cos(6.14366757538  + 48153.54375118419 *T)
    sumR0  +=       0.002 * cos(0.58190501461  + 27573.88326263939 *T)
    sumR0  +=       0.003 * cos(5.63901497713  + 106470.37642667518 *T)
    sumR0  +=       0.002 * cos(3.36639066096  + 42430.48572729180 *T)
    sumR0  +=       0.003 * cos(3.08628517069  + 4083.98850670440 *T)
    sumR0  +=       0.002 * cos(6.13024035669  + 104035.22069664020 *T)
    sumR0  +=       0.003 * cos(1.81280963372  + 104401.60918820098 *T)
    sumR0  +=       0.003 * cos(3.34404708133  + 130020.03106399580 *T)
    sumR0  +=       0.002 * cos(1.22333771717  + 76094.94894243999 *T)
    sumR0  +=       0.002 * cos(0.63542334090  + 1485.29070670320 *T)
    sumR0  +=       0.003 * cos(5.24568241255  + 6885.14988993081 *T)
    sumR0  +=       0.003 * cos(2.46305012048  + 133882.09065283637 *T)
    sumR0  +=       0.002 * cos(0.84094631125  + 163766.09444104519 *T)
    sumR0  +=       0.003 * cos(2.12660879166  + 42790.96527507320 *T)
    sumR0  +=       0.002 * cos(2.98805690719  + 26235.98186600050 *T)
    sumR0  +=       0.003 * cos(3.26197855217  + 156531.30018480320 *T)
    sumR0  +=       0.002 * cos(4.50734435605  + 25248.93385382380 *T)
    sumR0  +=       0.002 * cos(4.32859791317  + 1685.05212250160 *T)
    sumR0  +=       0.002 * cos(5.62335584014  + 28213.78054895339 *T)
    sumR0  +=       0.002 * cos(3.87492323552  + 199599.31813847597 *T)
    sumR0  +=       0.002 * cos(2.60597304148  + 26713.57333388659 *T)
    sumR0  +=       0.003 * cos(4.15733644316  + 51688.17153189420 *T)
    sumR0  +=       0.003 * cos(4.33101389874  + 220025.88923089800 *T)
    sumR0  +=       0.002 * cos(4.46567523081  + 31127.49138988719 *T)
    sumR0  +=       0.002 * cos(3.66265576771  + 26397.18146422999 *T)
    sumR0  +=       0.002 * cos(4.68312774052  + 49850.85687433279 *T)
    sumR0  +=       0.002 * cos(4.51215350981  + 52812.80255517260 *T)
    sumR0  +=       0.003 * cos(3.56621882535  + 154408.65498906677 *T)
    sumR0  +=       0.002 * cos(3.95708432187  + 1574.84580128220 *T)
    sumR0  +=       0.002 * cos(2.41940194318  + 104668.00443595338 *T)
    sumR0  +=       0.003 * cos(0.24921211552  + 25778.62481891839 *T)
    sumR0  +=       0.002 * cos(3.18370791629  + 207114.15223730975 *T)
    sumR0  +=       0.002 * cos(4.17768592199  + 51734.98099827079 *T)
    sumR0  +=       0.002 * cos(1.74283332135  + 76532.58707178998 *T)
    sumR0  +=       0.002 * cos(1.36339547692  + 6044.22858137540 *T)
    sumR0  +=       0.002 * cos(2.82219230315  + 63786.35824152260 *T)
    sumR0  +=       0.002 * cos(5.27165023098  + 75085.56363415498 *T)
    sumR0  +=       0.002 * cos(5.76735720320  + 25661.99436506020 *T)
    sumR0  +=       0.002 * cos(5.24666183170  + 24734.15439655740 *T)
    sumR0  +=       0.002 * cos(5.31625926462  + 24998.19435038059 *T)
    sumR0  +=       0.002 * cos(0.17868115875  + 177287.84588263658 *T)
    sumR0  +=       0.003 * cos(5.47127542511  + 27388.72995977919 *T)
    sumR0  +=       0.002 * cos(2.48055036253  + 25543.98508247800 *T)
    sumR0  +=       0.002 * cos(0.16253031708  + 23113.29318696320 *T)
    sumR0  +=       0.003 * cos(1.68453793975  + 2703.61615467560 *T)
    sumR0  +=       0.002 * cos(1.17069889009  + 53551.58008299420 *T)
    sumR0  +=       0.003 * cos(3.06893058902  + 230239.17477710897 *T)
    sumR0  +=       0.003 * cos(5.00145362139  + 180496.55813064097 *T)
    sumR0  +=       0.002 * cos(1.90098213386  + 20272.79228193720 *T)
    sumR0  +=       0.003 * cos(5.06877735444  + 440.82528487760 *T)
    sumR0  +=       0.002 * cos(1.33626564472  + 196156.74319351057 *T)
    sumR0  +=       0.002 * cos(4.50940380454  + 65538.25598994759 *T)
    sumR0  +=       0.002 * cos(4.55891488212  + 2324.94940881560 *T)
    sumR0  +=       0.002 * cos(0.21025125051  + 18227.48374913500 *T)
    sumR0  +=       0.002 * cos(0.58359043648  + 156377.85565231059 *T)
    sumR0  +=       0.002 * cos(4.57189351312  + 26926.87242932459 *T)
    sumR0  +=       0.002 * cos(3.79747475767  + 65851.33653522459 *T)
    sumR0  +=       0.002 * cos(0.30139015500  + 50951.98844236979 *T)
    sumR0  +=       0.002 * cos(3.94570007959  + 225687.22128005017 *T)
    sumR0  +=       0.002 * cos(0.99415397716  + 2064.97829370200 *T)
    sumR0  +=       0.002 * cos(3.49827019018  + 78905.12298646157 *T)
    sumR0  +=       0.002 * cos(4.16512398869  + 24864.08530079559 *T)
    sumR0  +=       0.002 * cos(0.47605198546  + 107692.22499299661 *T)
    sumR0  +=       0.002 * cos(0.02213914405  + 197092.67317421935 *T)
    sumR0  +=       0.002 * cos(0.35795159068  + 104301.61594439260 *T)
    sumR0  +=       0.002 * cos(2.54426644231  + 249268.47945736776 *T)
    sumR0  +=       0.002 * cos(0.35872697789  + 25462.23294926180 *T)
    sumR0  +=       0.002 * cos(0.03956368481  + 24189.55192363460 *T)
    sumR0  +=       0.002 * cos(2.00682967230  + 28786.35458895140 *T)
    sumR0  +=       0.002 * cos(1.64612572499  + 40738.70289615879 *T)
    sumR0  +=       0.002 * cos(1.77509184801  + 22595.33157470459 *T)
    sumR0  +=       0.002 * cos(1.93881159123  + 78697.42116259939 *T)
    sumR0  +=       0.002 * cos(1.06567957268  + 21562.24796431520 *T)
    sumR0  +=       0.003 * cos(4.58782698256  + 156954.01704032117 *T)
    sumR0  +=       0.002 * cos(0.84740706727  + 24494.89809302729 *T)
    sumR0  +=       0.002 * cos(5.98697403459  + 283524.35961235076 *T)
    sumR0  +=       0.002 * cos(4.02135788869  + 26238.95081141709 *T)
    sumR0  +=       0.002 * cos(3.98649649262  + 24808.10856894620 *T)
    sumR0  +=       0.002 * cos(3.06252327958  + 57369.17334194619 *T)
    sumR0  +=       0.002 * cos(6.10258403438  + 78115.63070029629 *T)
    sumR0  +=       0.002 * cos(0.19431714321  + 55484.27217852879 *T)
    sumR0  +=       0.002 * cos(4.26819419359  + 90989.16285949761 *T)
    sumR0  +=       0.002 * cos(2.94941610280  + 22065.64060961000 *T)
    sumR0  +=       0.002 * cos(4.53722544295  + 52355.88341343300 *T)
    sumR0  +=       0.002 * cos(1.97399498955  + 430.53034413910 *T)
    sumR0  +=       0.002 * cos(1.99872081555  + 24932.54198416719 *T)
    sumR0  +=       0.002 * cos(0.37032643194  + 26500.27423844859 *T)
    sumR0  +=       0.002 * cos(3.83480132336  + 307.55762096960 *T)
    sumR0  +=       0.002 * cos(5.81356382744  + 51538.81001112419 *T)
    sumR0  +=       0.002 * cos(6.20773606719  + 49.99662190420 *T)
    sumR0  +=       0.002 * cos(2.40589345998  + 77929.85348395319 *T)
    sumR0  +=       0.002 * cos(4.51719783127  + 21067.98465416099 *T)
    sumR0  +=       0.002 * cos(2.06311358499  + 23233.26276766400 *T)
    sumR0  +=       0.002 * cos(0.29430089395  + 24765.22549404280 *T)
    sumR0  +=       0.002 * cos(2.21513587213  + 25675.53204469980 *T)
    sumR0  +=       0.002 * cos(2.29291278322  + 27410.58078910560 *T)
    sumR0  +=       0.002 * cos(2.39356701559  + 242985.40360737639 *T)
    sumR0  +=       0.002 * cos(4.28957020884  + 71378.55953479178 *T)
    sumR0  +=       0.002 * cos(5.41199778568  + 23704.71012679799 *T)
    sumR0  +=       0.002 * cos(0.54252133289  + 3.88133535800 *T)
    sumR0  +=       0.002 * cos(5.88236623932  + 52072.02409456780 *T)
    sumR0  +=       0.002 * cos(5.78749542412  + 78373.91574594198 *T)
    sumR0  +=       0.002 * cos(5.69226787602  + 51040.12941125520 *T)
    sumR0  +=       0.002 * cos(4.42185841709  + 93696.66034953120 *T)
    sumR0  +=       0.002 * cos(5.15491529148  + 78112.66175487968 *T)
    sumR0  +=       0.002 * cos(2.30493903824  + 27883.16158529519 *T)
    sumR0  +=       0.002 * cos(5.53948896377  + 87253.17713015496 *T)
    sumR0  +=       0.002 * cos(4.75957265478  + 52336.21365253320 *T)
    sumR0  +=       0.002 * cos(6.08199781417  + 18043.37807764579 *T)
    sumR0  +=       0.002 * cos(0.73324052336  + 25757.28417791599 *T)
    sumR0  +=       0.002 * cos(0.91928644358  + 52186.85198341230 *T)
    sumR0  +=       0.002 * cos(3.22142261212  + 104771.09721017198 *T)
    sumR0  +=       0.002 * cos(2.01671301400  + 46564.47085590040 *T)
    sumR0  +=       0.002 * cos(4.39545799671  + 41520.85406638280 *T)
    sumR0  +=       0.002 * cos(2.52858429504  + 26140.99929388859 *T)
    sumR0  +=       0.002 * cos(2.07561943546  + 53654.67285721279 *T)
    sumR0  +=       0.002 * cos(0.08977124652  + 25135.82651831460 *T)
    sumR0  +=       0.002 * cos(1.67350036401  + 66826.60603773296 *T)
    sumR0  +=       0.002 * cos(4.57268426145  + 158116.49174472899 *T)
    sumR0  +=       0.002 * cos(2.84194927894  + 78249.48233072099 *T)
    sumR0  +=       0.002 * cos(3.07002924674  + 23432.33476910039 *T)
    sumR0  +=       0.002 * cos(4.11559657607  + 123554.36581794015 *T)
    sumR0  +=       0.002 * cos(0.05687800902  + 26034.80698925980 *T)
    sumR0  +=       0.002 * cos(4.28354458896  + 24484.60315228879 *T)
    sumR0  +=       0.002 * cos(3.24222429199  + 21509.65132471559 *T)
    sumR0  +=       0.002 * cos(2.52315832958  + 210614.42461581080 *T)
    sumR0  +=       0.002 * cos(4.44247371827  + 148532.89040742096 *T)
    sumR0  +=       0.002 * cos(0.11737800376  + 146505.58156934578 *T)
    sumR0  +=       0.002 * cos(2.41371362390  + 54862.52837768960 *T)
    sumR0  +=       0.002 * cos(4.66022680664  + 247223.17092456561 *T)
    sumR0  +=       0.002 * cos(3.79260774499  + 3308.46589538040 *T)
    sumR0  +=       0.002 * cos(5.83454304626  + 2974.60995461100 *T)
    sumR0  +=       0.002 * cos(5.37878500160  + 77622.29586298359 *T)
    sumR0  +=       0.002 * cos(1.63153524958  + 99979.73619952259 *T)
    sumR0  +=       0.002 * cos(5.18613735503  + 50600.96048186620 *T)
    sumR0  +=       0.002 * cos(2.55420547959  + 76064.62208143558 *T)
    sumR0  +=       0.002 * cos(4.05996888958  + 24502.76239955349 *T)
    sumR0  +=       0.002 * cos(2.47809526116  + 179966.86716554637 *T)
    sumR0  +=       0.002 * cos(1.25528265056  + 50912.64892057020 *T)
    sumR0  +=       0.002 * cos(0.75306774592  + 235746.72801577640 *T)
    sumR0  +=       0.002 * cos(2.97464643769  + 31219.54422563180 *T)
    sumR0  +=       0.001 * cos(0.17722772219  + 158438.61833266239 *T)
    sumR0  +=       0.002 * cos(4.94589808424  + 168959.46149984296 *T)
    sumR0  +=       0.002 * cos(0.63110575282  + 415.55249061210 *T)
    sumR0  +=       0.001 * cos(0.77799418557  + 299398.97720771394 *T)
    sumR0  +=       0.002 * cos(5.97867896248  + 164721.69418265377 *T)
    sumR0  +=       0.002 * cos(5.62955025847  + 21819.80896338059 *T)
    sumR0  +=       0.001 * cos(4.93638931294  + 27993.36790651460 *T)
    sumR0  +=       0.002 * cos(6.22578827437  + 74935.57376844239 *T)
    sumR0  +=       0.002 * cos(1.03965665113  + 1802.37199072180 *T)
    sumR0  +=       0.001 * cos(5.61994219678  + 50690.51557644519 *T)
    sumR0  +=       0.002 * cos(5.02281046549  + 422.66603761290 *T)
    sumR0  +=       0.001 * cos(3.57619837497  + 209658.82487420217 *T)
    sumR0  +=       0.002 * cos(0.11249539090  + 38469.94917006779 *T)
    sumR0  +=       0.002 * cos(0.62470424533  + 51745.27593900929 *T)
    sumR0  +=       0.001 * cos(0.71806518278  + 28928.40404660019 *T)
    sumR0  +=       0.002 * cos(3.91669753023  + 76352.50994150538 *T)
    sumR0  +=       0.002 * cos(3.05277164813  + 225.82926841020 *T)
    sumR0  +=       0.001 * cos(4.33602746737  + 4601.95011896300 *T)

    # MERCURY Radius
    sumR1  = 0
    sumR1  +=  217347.739 * cos(4.65617158663  + 26087.90314157420 *T)
    sumR1  +=   44141.826 * cos(1.42385543975  + 52175.80628314840 *T)
    sumR1  +=   10094.479 * cos(4.47466326316  + 78263.70942472259 *T)
    sumR1  +=    2432.804 * cos(1.24226083435  + 104351.61256629678 *T)
    sumR1  +=    1624.367 * cos(0.00000000000  + 0.00000000000 *T)
    sumR1  +=     603.996 * cos(4.29303116561  + 130439.51570787100 *T)
    sumR1  +=     152.851 * cos(1.06060779810  + 156527.41884944518 *T)
    sumR1  +=      39.202 * cos(4.11136751416  + 182615.32199101936 *T)
    sumR1  +=      17.760 * cos(4.54424653085  + 27197.28169366760 *T)
    sumR1  +=      17.999 * cos(4.71193725810  + 24978.52458948080 *T)
    sumR1  +=      10.154 * cos(0.87893548494  + 208703.22513259359 *T)
    sumR1  +=       8.086 * cos(3.00540854890  + 25028.52121138500 *T)
    sumR1  +=       4.444 * cos(2.13639058123  + 20426.57109242200 *T)
    sumR1  +=       4.393 * cos(1.48074475165  + 51066.42773105500 *T)
    sumR1  +=       3.510 * cos(3.21171223697  + 1059.38193018920 *T)
    sumR1  +=       3.133 * cos(5.23846336855  + 21535.94964451540 *T)
    sumR1  +=       2.650 * cos(3.92968881423  + 234791.12827416777 *T)
    sumR1  +=       2.498 * cos(2.02623950395  + 24498.83024629040 *T)
    sumR1  +=       2.011 * cos(1.23911360588  + 53285.18483524180 *T)
    sumR1  +=       1.963 * cos(4.04524902962  + 5661.33204915220 *T)
    sumR1  +=       1.546 * cos(2.61849590442  + 26617.59410666880 *T)
    sumR1  +=       1.542 * cos(2.35659537465  + 27043.50288318280 *T)
    sumR1  +=       1.417 * cos(1.37876323533  + 1109.37855209340 *T)
    sumR1  +=       1.295 * cos(5.19094850935  + 46514.47423399620 *T)
    sumR1  +=       1.241 * cos(0.21246226135  + 11322.66409830440 *T)
    sumR1  +=       1.090 * cos(6.23733263925  + 27147.28507176339 *T)
    sumR1  +=       1.271 * cos(0.56437938715  + 25132.30339996560 *T)
    sumR1  +=       1.120 * cos(0.86374628388  + 57837.13833230060 *T)
    sumR1  +=       1.005 * cos(3.28272284427  + 37410.56723987860 *T)
    sumR1  +=       1.177 * cos(4.53194022227  + 77154.33087262919 *T)
    sumR1  +=       0.841 * cos(5.96261035419  + 14765.23904326980 *T)
    sumR1  +=       0.935 * cos(2.37277497611  + 25661.30495069820 *T)
    sumR1  +=       0.742 * cos(0.77751594919  + 32858.61374281979 *T)
    sumR1  +=       0.736 * cos(4.06885296570  + 1589.07289528380 *T)
    sumR1  +=       0.712 * cos(2.70884051645  + 16983.99614745660 *T)
    sumR1  +=       0.863 * cos(5.55308682281  + 25558.21217647960 *T)
    sumR1  +=       0.707 * cos(0.92992556568  + 30639.85663863300 *T)
    sumR1  +=       0.696 * cos(0.69734796212  + 260879.03141574195 *T)
    sumR1  +=       0.678 * cos(2.02193885930  + 26068.23338067440 *T)
    sumR1  +=       0.608 * cos(0.85708468483  + 4551.95349705880 *T)
    sumR1  +=       0.546 * cos(5.38138962861  + 26091.78447693220 *T)
    sumR1  +=       0.608 * cos(3.65993973133  + 25448.00585526019 *T)
    sumR1  +=       0.597 * cos(4.64867954048  + 26080.78959457339 *T)
    sumR1  +=       0.572 * cos(3.50125649521  + 10213.28554621100 *T)
    sumR1  +=       0.515 * cos(4.83662922562  + 50586.73338786459 *T)
    sumR1  +=       0.532 * cos(5.21855381213  + 22645.32819660879 *T)
    sumR1  +=       0.572 * cos(0.62751016416  + 13521.75144159140 *T)
    sumR1  +=       0.516 * cos(3.89421000089  + 83925.04147387479 *T)
    sumR1  +=       0.507 * cos(5.55466020276  + 12566.15169998280 *T)
    sumR1  +=       0.606 * cos(4.50215566301  + 28306.66024576099 *T)
    sumR1  +=       0.533 * cos(0.93587795381  + 26107.57290247399 *T)
    sumR1  +=       0.435 * cos(5.79043985706  + 43071.89928903080 *T)
    sumR1  +=       0.436 * cos(5.86659160961  + 25035.63475838580 *T)
    sumR1  +=       0.429 * cos(0.26260280461  + 36301.18868778519 *T)
    sumR1  +=       0.472 * cos(6.00507778421  + 77204.32749453338 *T)
    sumR1  +=       0.443 * cos(3.86758570847  + 26084.02180621620 *T)
    sumR1  +=       0.487 * cos(4.23950195756  + 26095.01668857500 *T)
    sumR1  +=       0.406 * cos(5.60324652352  + 52705.49724824299 *T)
    sumR1  +=       0.523 * cos(0.06439941262  + 25021.40766438420 *T)
    sumR1  +=       0.379 * cos(1.95647626585  + 72602.37737557039 *T)
    sumR1  +=       0.390 * cos(3.81996185881  + 426.59819087600 *T)
    sumR1  +=       0.418 * cos(2.77521418660  + 41962.52073693740 *T)
    sumR1  +=       0.453 * cos(0.72803137250  + 529.69096509460 *T)
    sumR1  +=       0.358 * cos(6.00165335616  + 33326.57873317420 *T)
    sumR1  +=       0.347 * cos(1.39524259509  + 23969.13928119580 *T)
    sumR1  +=       0.340 * cos(0.24111380317  + 26301.20223701220 *T)
    sumR1  +=       0.325 * cos(4.78747749132  + 23869.14603738740 *T)
    sumR1  +=       0.290 * cos(3.04044859601  + 53235.18821333759 *T)
    sumR1  +=       0.326 * cos(1.30033833957  + 103242.23401420339 *T)
    sumR1  +=       0.316 * cos(2.22791434675  + 19317.19254032860 *T)
    sumR1  +=       0.264 * cos(3.93160763696  + 6770.71060124560 *T)
    sumR1  +=       0.332 * cos(3.59235758173  + 51220.20654153979 *T)
    sumR1  +=       0.223 * cos(2.76528831516  + 103292.23063610760 *T)
    sumR1  +=       0.215 * cos(3.97996009773  + 56727.75978020720 *T)
    sumR1  +=       0.235 * cos(2.24050503260  + 25874.60404613620 *T)
    sumR1  +=       0.215 * cos(6.16660008496  + 51646.11531805379 *T)
    sumR1  +=       0.208 * cos(2.74551899929  + 40853.14218484400 *T)
    sumR1  +=       0.204 * cos(3.44654782597  + 9103.90699411760 *T)
    sumR1  +=       0.197 * cos(0.80026289238  + 27676.97603685800 *T)
    sumR1  +=       0.205 * cos(3.20844012088  + 18849.22754997420 *T)
    sumR1  +=       0.196 * cos(1.99924288764  + 48733.23133818299 *T)
    sumR1  +=       0.218 * cos(2.11865566044  + 73711.75592766379 *T)
    sumR1  +=       0.189 * cos(0.65681038131  + 110012.94461544900 *T)
    sumR1  +=       0.207 * cos(5.89082708705  + 31749.23519072640 *T)
    sumR1  +=       0.184 * cos(3.74837387583  + 286966.93455731618 *T)
    sumR1  +=       0.174 * cos(5.81264188702  + 68050.42387851159 *T)
    sumR1  +=       0.163 * cos(4.00380715262  + 79373.08797681599 *T)
    sumR1  +=       0.152 * cos(5.07389967152  + 52156.13652224860 *T)
    sumR1  +=       0.151 * cos(2.83369443311  + 955.59974160860 *T)
    sumR1  +=       0.151 * cos(0.40953650210  + 26514.50133245020 *T)
    sumR1  +=       0.147 * cos(2.32191248197  + 38654.05484155699 *T)
    sumR1  +=       0.154 * cos(3.26437373895  + 38519.94579197200 *T)
    sumR1  +=       0.185 * cos(0.98966099528  + 29530.47808653960 *T)
    sumR1  +=       0.150 * cos(3.69230644689  + 65697.55772473979 *T)
    sumR1  +=       0.132 * cos(2.71837382155  + 77734.01845962799 *T)
    sumR1  +=       0.140 * cos(2.05736044462  + 6283.07584999140 *T)
    sumR1  +=       0.122 * cos(3.31361750281  + 62389.09182935939 *T)
    sumR1  +=       0.123 * cos(2.07840148125  + 25668.41849769900 *T)
    sumR1  +=       0.118 * cos(0.51913709271  + 51535.90899683439 *T)
    sumR1  +=       0.121 * cos(2.15433834534  + 52179.68761850640 *T)
    sumR1  +=       0.151 * cos(2.66028539824  + 25654.19140369739 *T)
    sumR1  +=       0.113 * cos(2.33243126122  + 78793.40038981718 *T)
    sumR1  +=       0.124 * cos(3.13058661995  + 51109.31080595839 *T)
    sumR1  +=       0.150 * cos(2.60376479044  + 639.89728631400 *T)
    sumR1  +=       0.123 * cos(2.65137961558  + 51123.53789995999 *T)
    sumR1  +=       0.106 * cos(1.31393069594  + 47623.85278608960 *T)
    sumR1  +=       0.112 * cos(5.00798069297  + 98690.28051714458 *T)
    sumR1  +=       0.107 * cos(5.12738291017  + 99799.65906923798 *T)
    sumR1  +=       0.123 * cos(6.13795219131  + 1066.49547719000 *T)
    sumR1  +=       0.106 * cos(1.42997322881  + 52168.69273614759 *T)
    sumR1  +=       0.104 * cos(4.51477063494  + 25234.70675982219 *T)
    sumR1  +=       0.120 * cos(4.06567300860  + 45892.73043315699 *T)
    sumR1  +=       0.092 * cos(4.72811907093  + 2118.76386037840 *T)
    sumR1  +=       0.112 * cos(4.78983902693  + 53131.40602475700 *T)
    sumR1  +=       0.096 * cos(5.03438502363  + 32370.97899156560 *T)
    sumR1  +=       0.091 * cos(4.22099920434  + 50057.04242277000 *T)
    sumR1  +=       0.091 * cos(4.35481864546  + 129330.13715577760 *T)
    sumR1  +=       0.106 * cos(0.14583281261  + 12432.04265039780 *T)
    sumR1  +=       0.084 * cos(0.32848825294  + 1052.26838318840 *T)
    sumR1  +=       0.084 * cos(5.81442139422  + 129380.13377768178 *T)
    sumR1  +=       0.093 * cos(0.34539006564  + 77308.10968311399 *T)
    sumR1  +=       0.078 * cos(1.17633266287  + 76674.63652943878 *T)
    sumR1  +=       0.107 * cos(1.01316358023  + 52182.91983014920 *T)
    sumR1  +=       0.082 * cos(6.11445014523  + 79323.09135491178 *T)
    sumR1  +=       0.085 * cos(2.50332044483  + 7238.67559160000 *T)
    sumR1  +=       0.072 * cos(1.56170832500  + 49957.04917896160 *T)
    sumR1  +=       0.076 * cos(1.62577226881  + 15874.61759536320 *T)
    sumR1  +=       0.091 * cos(5.75408440781  + 24491.71669928959 *T)
    sumR1  +=       0.092 * cos(3.97747935513  + 52195.47604404819 *T)
    sumR1  +=       0.068 * cos(0.44378826429  + 91785.46086631398 *T)
    sumR1  +=       0.078 * cos(0.63585741399  + 52171.92494779040 *T)
    sumR1  +=       0.064 * cos(2.57311321703  + 94138.32702008578 *T)
    sumR1  +=       0.065 * cos(2.12942233842  + 7.11354700080 *T)
    sumR1  +=       0.062 * cos(3.70751240431  + 136100.84775702321 *T)
    sumR1  +=       0.068 * cos(1.18946361248  + 54394.56338733519 *T)
    sumR1  +=       0.079 * cos(3.35596277474  + 52389.10537858640 *T)
    sumR1  +=       0.064 * cos(5.09677055426  + 51962.50718771040 *T)
    sumR1  +=       0.060 * cos(6.25938912295  + 24505.94379329119 *T)
    sumR1  +=       0.061 * cos(0.73867692725  + 82815.66292178139 *T)
    sumR1  +=       0.063 * cos(1.29883580487  + 2218.75710418680 *T)
    sumR1  +=       0.070 * cos(5.28692696943  + 45405.09568190280 *T)
    sumR1  +=       0.057 * cos(1.92365363632  + 20760.42703319140 *T)
    sumR1  +=       0.054 * cos(0.89448628916  + 71980.63357473118 *T)
    sumR1  +=       0.051 * cos(4.71710385051  + 105460.99111839020 *T)
    sumR1  +=       0.054 * cos(6.22009952451  + 44937.13069154840 *T)
    sumR1  +=       0.052 * cos(3.43365769233  + 25551.09862947879 *T)
    sumR1  +=       0.052 * cos(5.77930503293  + 66941.04532641819 *T)
    sumR1  +=       0.048 * cos(0.51937745535  + 313054.83769889042 *T)
    sumR1  +=       0.052 * cos(5.71476972233  + 103821.92160120218 *T)
    sumR1  +=       0.047 * cos(3.82203240163  + 53764.87917843220 *T)
    sumR1  +=       0.050 * cos(5.76150579489  + 25938.33994443960 *T)
    sumR1  +=       0.047 * cos(5.52705163618  + 77837.11123384659 *T)
    sumR1  +=       0.043 * cos(5.84994474092  + 63498.47038145279 *T)
    sumR1  +=       0.052 * cos(2.63638299982  + 18093.37469954999 *T)
    sumR1  +=       0.042 * cos(5.39178421925  + 64741.95798313119 *T)
    sumR1  +=       0.041 * cos(1.87113963554  + 125887.56221081220 *T)
    sumR1  +=       0.045 * cos(4.04603530632  + 55618.38122811380 *T)
    sumR1  +=       0.039 * cos(3.50934954924  + 52602.40447402440 *T)
    sumR1  +=       0.048 * cos(0.72400634326  + 33967.99229491319 *T)
    sumR1  +=       0.051 * cos(6.02427997984  + 13655.86049117640 *T)
    sumR1  +=       0.038 * cos(0.09359107033  + 88476.99497093359 *T)
    sumR1  +=       0.038 * cos(1.84297739511  + 78244.03966382280 *T)
    sumR1  +=       0.041 * cos(5.74927951755  + 26727.80042788820 *T)
    sumR1  +=       0.045 * cos(5.73633413733  + 44181.27784112419 *T)
    sumR1  +=       0.035 * cos(0.81793649503  + 23439.44831610119 *T)
    sumR1  +=       0.038 * cos(5.17914668407  + 23754.70674870219 *T)
    sumR1  +=       0.036 * cos(4.81782090538  + 51116.42435295920 *T)
    sumR1  +=       0.041 * cos(2.91943000674  + 26241.68195205900 *T)
    sumR1  +=       0.032 * cos(1.91760062435  + 58458.88213313979 *T)
    sumR1  +=       0.033 * cos(1.27507180487  + 131548.89425996438 *T)
    sumR1  +=       0.032 * cos(5.01820692941  + 46848.33017476560 *T)
    sumR1  +=       0.030 * cos(3.95444605179  + 26555.86813192860 *T)
    sumR1  +=       0.031 * cos(1.17911028239  + 59414.48187474840 *T)
    sumR1  +=       0.030 * cos(0.55604674983  + 3442.57494496540 *T)
    sumR1  +=       0.034 * cos(3.55745328485  + 433.71173787680 *T)
    sumR1  +=       0.033 * cos(5.36229697065  + 104881.30353139140 *T)
    sumR1  +=       0.030 * cos(1.24209931344  + 26202.34243025941 *T)
    sumR1  +=       0.033 * cos(3.55673853222  + 25440.89230825939 *T)
    sumR1  +=       0.029 * cos(2.98847901150  + 27154.39861876420 *T)
    sumR1  +=       0.035 * cos(3.23888704617  + 25455.11940226100 *T)
    sumR1  +=       0.033 * cos(1.77322555881  + 124778.18365871880 *T)
    sumR1  +=       0.028 * cos(4.35951220773  + 80482.46652890938 *T)
    sumR1  +=       0.030 * cos(5.14555074991  + 51756.32163927320 *T)
    sumR1  +=       0.029 * cos(0.03524298861  + 79219.30916633119 *T)
    sumR1  +=       0.027 * cos(2.59555966553  + 155468.03691925600 *T)
    sumR1  +=       0.027 * cos(1.05480537311  + 28421.09953444620 *T)
    sumR1  +=       0.030 * cos(5.72087736840  + 51742.09454527159 *T)
    sumR1  +=       0.029 * cos(5.20529460359  + 78267.59076008058 *T)
    sumR1  +=       0.026 * cos(1.11551342052  + 155418.04029735178 *T)
    sumR1  +=       0.030 * cos(6.10884155937  + 26011.63707029860 *T)
    sumR1  +=       0.027 * cos(3.38666053174  + 103396.01282468818 *T)
    sumR1  +=       0.025 * cos(3.48592449926  + 117873.36400788820 *T)
    sumR1  +=       0.025 * cos(1.69043243445  + 25973.46385288896 *T)
    sumR1  +=       0.026 * cos(6.19577940400  + 77197.21394753258 *T)
    sumR1  +=       0.024 * cos(1.47972864008  + 28206.66700195260 *T)
    sumR1  +=       0.023 * cos(3.86767325148  + 16066.06586147480 *T)
    sumR1  +=       0.027 * cos(4.06858086478  + 78270.82297172339 *T)
    sumR1  +=       0.022 * cos(4.03528313026  + 419.48464387520 *T)
    sumR1  +=       0.023 * cos(3.38426107540  + 58946.51688439399 *T)
    sumR1  +=       0.022 * cos(1.33669336918  + 51322.60990139639 *T)
    sumR1  +=       0.023 * cos(2.23954141042  + 103925.01437542078 *T)
    sumR1  +=       0.024 * cos(3.65673385174  + 26610.48055966799 *T)
    sumR1  +=       0.021 * cos(4.49565014938  + 78256.59587772179 *T)
    sumR1  +=       0.021 * cos(3.69266624882  + 77623.81213840858 *T)
    sumR1  +=       0.028 * cos(5.71652736884  + 77211.44104153418 *T)
    sumR1  +=       0.021 * cos(3.97119771059  + 98068.53671630540 *T)
    sumR1  +=       0.021 * cos(5.61140929496  + 120226.23016166000 *T)
    sumR1  +=       0.025 * cos(1.73460075205  + 853.19638175200 *T)
    sumR1  +=       0.024 * cos(2.89897553229  + 105410.99449648600 *T)
    sumR1  +=       0.020 * cos(6.15988447425  + 213.29909543800 *T)
    sumR1  +=       0.021 * cos(0.42126764910  + 149.56319713460 *T)
    sumR1  +=       0.020 * cos(4.41588880947  + 29416.03879785439 *T)
    sumR1  +=       0.024 * cos(2.43220296378  + 50579.61984086379 *T)
    sumR1  +=       0.020 * cos(0.47739908961  + 162188.75089859738 *T)
    sumR1  +=       0.021 * cos(0.18110422167  + 35191.81013569180 *T)
    sumR1  +=       0.019 * cos(2.38756251160  + 26037.90651966999 *T)
    sumR1  +=       0.020 * cos(0.61997291259  + 85034.42002596818 *T)
    sumR1  +=       0.018 * cos(4.60331561584  + 76044.95232053580 *T)
    sumR1  +=       0.018 * cos(3.06180224259  + 50593.84693486539 *T)
    sumR1  +=       0.022 * cos(0.45787269756  + 1596.18644228460 *T)
    sumR1  +=       0.019 * cos(1.76795970776  + 78050.41032928458 *T)
    sumR1  +=       0.019 * cos(3.58333941723  + 27140.17152476259 *T)
    sumR1  +=       0.021 * cos(0.15585183693  + 78477.00852016058 *T)
    sumR1  +=       0.018 * cos(1.81330190957  + 25565.32572348040 *T)
    sumR1  +=       0.021 * cos(3.45385724015  + 102762.53967101300 *T)
    sumR1  +=       0.022 * cos(4.62891777236  + 19406.67828817460 *T)
    sumR1  +=       0.019 * cos(5.28998871939  + 26624.70765366959 *T)
    sumR1  +=       0.016 * cos(4.37908924360  + 25131.61398560359 *T)
    sumR1  +=       0.019 * cos(2.69598918032  + 105307.21230790540 *T)
    sumR1  +=       0.018 * cos(4.13274028026  + 5327.47610838280 *T)
    sumR1  +=       0.018 * cos(1.10819559505  + 76144.94556434419 *T)
    sumR1  +=       0.018 * cos(2.04678515653  + 71492.99882347698 *T)
    sumR1  +=       0.015 * cos(0.52406345582  + 42153.96900304900 *T)
    sumR1  +=       0.018 * cos(0.71897799959  + 78283.37918562238 *T)
    sumR1  +=       0.020 * cos(3.87105632280  + 51749.20809227239 *T)
    sumR1  +=       0.014 * cos(0.06732488000  + 25984.81036735560 *T)
    sumR1  +=       0.019 * cos(2.51337817026  + 129909.82474277640 *T)
    sumR1  +=       0.018 * cos(3.79705759661  + 108903.56606335558 *T)
    sumR1  +=       0.014 * cos(2.30467577173  + 18207.81398823521 *T)
    sumR1  +=       0.014 * cos(4.91192145857  + 151975.46535238638 *T)
    sumR1  +=       0.014 * cos(4.25365079026  + 157636.79740153858 *T)
    sumR1  +=       0.014 * cos(3.82494343251  + 7880.08915333899 *T)
    sumR1  +=       0.014 * cos(0.47442483304  + 26137.89976347840 *T)
    sumR1  +=       0.015 * cos(1.41820144369  + 40565.25432477420 *T)
    sumR1  +=       0.016 * cos(1.92985195149  + 49842.60989027639 *T)
    sumR1  +=       0.016 * cos(2.97241905198  + 71025.03383312259 *T)
    sumR1  +=       0.013 * cos(0.07086712207  + 76.26607127560 *T)
    sumR1  +=       0.013 * cos(2.19019855275  + 90829.86112470538 *T)
    sumR1  +=       0.014 * cos(1.68395491426  + 6681.22485339960 *T)
    sumR1  +=       0.012 * cos(2.16242752229  + 29428.51556827400 *T)
    sumR1  +=       0.013 * cos(0.55373605944  + 79852.78232000639 *T)
    sumR1  +=       0.012 * cos(2.10236188465  + 24609.03656750980 *T)
    sumR1  +=       0.012 * cos(3.49372354820  + 64607.84893354619 *T)
    sumR1  +=       0.013 * cos(3.16312334997  + 25881.71759313700 *T)
    sumR1  +=       0.012 * cos(0.32301447482  + 78690.30761559859 *T)
    sumR1  +=       0.016 * cos(3.67591976897  + 78259.82808936459 *T)
    sumR1  +=       0.012 * cos(0.81060544329  + 81706.28436968799 *T)
    sumR1  +=       0.015 * cos(2.57802360273  + 93028.94846799241 *T)
    sumR1  +=       0.013 * cos(3.55704873434  + 339142.74084046454 *T)
    sumR1  +=       0.012 * cos(3.14169564714  + 114564.89811250778 *T)
    sumR1  +=       0.011 * cos(4.96251639039  + 84546.78527471398 *T)
    sumR1  +=       0.011 * cos(2.93160493628  + 536.80451209540 *T)
    sumR1  +=       0.011 * cos(0.04474275203  + 19804.82729158280 *T)
    sumR1  +=       0.011 * cos(2.39391294657  + 10021.83728009940 *T)
    sumR1  +=       0.011 * cos(3.22530161123  + 90695.75207512038 *T)
    sumR1  +=       0.012 * cos(0.71785378975  + 26294.08869001139 *T)
    sumR1  +=       0.011 * cos(3.51130712506  + 7994.52844202420 *T)
    sumR1  +=       0.011 * cos(1.78813511249  + 69159.80243060499 *T)
    sumR1  +=       0.010 * cos(1.89377082198  + 24925.42843716640 *T)
    sumR1  +=       0.012 * cos(6.06775243662  + 52329.58509363319 *T)
    sumR1  +=       0.012 * cos(0.11656433296  + 51543.02254383520 *T)
    sumR1  +=       0.012 * cos(0.01462307671  + 1581.95934828300 *T)
    sumR1  +=       0.010 * cos(3.60476399042  + 49527.35145767539 *T)
    sumR1  +=       0.009 * cos(5.35498571113  + 2648.45482547300 *T)
    sumR1  +=       0.010 * cos(2.54352513591  + 52815.70356946240 *T)
    sumR1  +=       0.009 * cos(5.25498849885  + 130012.91751699500 *T)
    sumR1  +=       0.010 * cos(5.63770719377  + 181555.94006083015 *T)
    sumR1  +=       0.011 * cos(1.65385601347  + 26507.38778544939 *T)
    sumR1  +=       0.011 * cos(4.93008131424  + 104331.94280539700 *T)
    sumR1  +=       0.010 * cos(0.35445064092  + 26521.61487945100 *T)
    sumR1  +=       0.010 * cos(2.79303258728  + 39743.76363275060 *T)
    sumR1  +=       0.009 * cos(0.71144667758  + 25764.39772491679 *T)
    sumR1  +=       0.008 * cos(4.22951187793  + 181505.94343892601 *T)
    sumR1  +=       0.008 * cos(0.17514303190  + 129483.91596626240 *T)
    sumR1  +=       0.010 * cos(1.16786157884  + 25934.12433108940 *T)
    sumR1  +=       0.010 * cos(2.50335300453  + 103299.34418310841 *T)
    sumR1  +=       0.009 * cos(3.18692590406  + 39629.32434406539 *T)
    sumR1  +=       0.010 * cos(3.98242351264  + 522.57741809380 *T)
    sumR1  +=       0.011 * cos(0.23158217376  + 51639.00177105299 *T)
    sumR1  +=       0.009 * cos(0.30008658313  + 27999.10262479140 *T)
    sumR1  +=       0.011 * cos(4.84801040284  + 150866.08680029298 *T)
    sumR1  +=       0.011 * cos(5.94731028211  + 38813.35657634920 *T)
    sumR1  +=       0.008 * cos(4.66199260294  + 26094.53170047421 *T)
    sumR1  +=       0.008 * cos(0.71995165086  + 52643.77127350280 *T)
    sumR1  +=       0.008 * cos(3.31825277790  + 24395.73747207180 *T)
    sumR1  +=       0.008 * cos(2.52932369592  + 26190.99591579279 *T)
    sumR1  +=       0.008 * cos(5.11541081184  + 25619.93815121980 *T)
    sumR1  +=       0.008 * cos(1.18625312267  + 25867.49049913539 *T)
    sumR1  +=       0.008 * cos(0.31593558687  + 51528.79544983359 *T)
    sumR1  +=       0.010 * cos(2.12019897039  + 130969.20667296558 *T)
    sumR1  +=       0.008 * cos(0.24131547257  + 143961.26714946239 *T)
    sumR1  +=       0.007 * cos(5.90865595918  + 23976.25282819660 *T)
    sumR1  +=       0.008 * cos(4.88221803127  + 22759.76748529401 *T)
    sumR1  +=       0.007 * cos(3.17202081415  + 632.78373931320 *T)
    sumR1  +=       0.008 * cos(2.40875907837  + 146314.13330323418 *T)
    sumR1  +=       0.007 * cos(6.03640433343  + 53242.30176033840 *T)
    sumR1  +=       0.007 * cos(0.86406939215  + 104358.72611329758 *T)
    sumR1  +=       0.007 * cos(3.45761762182  + 26237.46633870879 *T)
    sumR1  +=       0.009 * cos(3.30628597530  + 12725.45343477500 *T)
    sumR1  +=       0.007 * cos(0.74174993099  + 124156.43985787958 *T)
    sumR1  +=       0.008 * cos(5.62600154280  + 131395.11544947961 *T)
    sumR1  +=       0.007 * cos(2.52679818243  + 77829.99768684579 *T)
    sumR1  +=       0.008 * cos(1.93255904016  + 77844.22478084739 *T)
    sumR1  +=       0.007 * cos(3.69439677930  + 85502.38501632259 *T)
    sumR1  +=       0.007 * cos(5.14299105203  + 35472.74414964940 *T)
    sumR1  +=       0.008 * cos(1.52116219771  + 26081.27458267419 *T)
    sumR1  +=       0.008 * cos(3.50246944455  + 25863.55834587229 *T)
    sumR1  +=       0.008 * cos(5.17598738878  + 23962.02573419499 *T)
    sumR1  +=       0.009 * cos(2.01720755803  + 104355.49390165480 *T)
    sumR1  +=       0.007 * cos(0.57777174711  + 86143.79857806159 *T)
    sumR1  +=       0.007 * cos(2.03534300720  + 100909.03762133140 *T)
    sumR1  +=       0.006 * cos(3.30402834266  + 104564.91166173480 *T)
    sumR1  +=       0.006 * cos(5.26635186541  + 50696.93970908399 *T)
    sumR1  +=       0.006 * cos(3.60993601694  + 27684.08958385880 *T)
    sumR1  +=       0.008 * cos(5.97157930225  + 131498.89763806018 *T)
    sumR1  +=       0.006 * cos(4.43204136420  + 54294.57014352679 *T)
    sumR1  +=       0.006 * cos(4.16183734880  + 54509.00267602040 *T)
    sumR1  +=       0.007 * cos(4.35729742194  + 52290.24557183361 *T)
    sumR1  +=       0.007 * cos(4.88364528875  + 26941.09952332620 *T)
    sumR1  +=       0.007 * cos(3.84735933689  + 34282.17847478280 *T)
    sumR1  +=       0.006 * cos(0.71889591582  + 89586.37352302698 *T)
    sumR1  +=       0.006 * cos(4.77410201490  + 104138.31347085880 *T)
    sumR1  +=       0.006 * cos(5.81016662275  + 70269.18098269838 *T)
    sumR1  +=       0.007 * cos(3.59164938122  + 188276.65404017159 *T)
    sumR1  +=       0.006 * cos(1.42397911026  + 102132.85546211000 *T)
    sumR1  +=       0.006 * cos(5.15043168056  + 97580.90196505121 *T)
    sumR1  +=       0.008 * cos(2.67404385604  + 647.01083331480 *T)
    sumR1  +=       0.005 * cos(4.58227258685  + 25977.69682035479 *T)
    sumR1  +=       0.007 * cos(6.18437842544  + 128850.44281258718 *T)
    sumR1  +=       0.005 * cos(0.05338946228  + 13541.42120249119 *T)
    sumR1  +=       0.006 * cos(5.55589010148  + 155997.72788435058 *T)
    sumR1  +=       0.005 * cos(5.54731935938  + 96357.08412427260 *T)
    sumR1  +=       0.006 * cos(0.36201698821  + 26308.31578401300 *T)
    sumR1  +=       0.005 * cos(3.05584811656  + 103285.11708910680 *T)
    sumR1  +=       0.005 * cos(4.80319811501  + 51013.33157874059 *T)
    sumR1  +=       0.006 * cos(3.04691542636  + 13362.44970679920 *T)
    sumR1  +=       0.005 * cos(5.59905054438  + 76667.52298243798 *T)
    sumR1  +=       0.005 * cos(3.15313951574  + 26164.16921284980 *T)
    sumR1  +=       0.006 * cos(1.15194666645  + 27311.72098235281 *T)
    sumR1  +=       0.005 * cos(6.16463589751  + 76681.75007643958 *T)
    sumR1  +=       0.005 * cos(2.77667883666  + 224.34479570190 *T)
    sumR1  +=       0.005 * cos(4.27775147717  + 14477.35118320000 *T)
    sumR1  +=       0.005 * cos(5.50743820134  + 52125.80966124419 *T)
    sumR1  +=       0.006 * cos(0.68838699976  + 26720.68688088739 *T)
    sumR1  +=       0.005 * cos(4.44878671847  + 25227.59321282139 *T)
    sumR1  +=       0.005 * cos(4.88707969684  + 52061.36699446317 *T)
    sumR1  +=       0.005 * cos(3.66186576821  + 25241.82030682300 *T)
    sumR1  +=       0.005 * cos(1.76383923429  + 178063.36849396059 *T)
    sumR1  +=       0.005 * cos(3.09083817171  + 115674.27666460120 *T)
    sumR1  +=       0.005 * cos(1.73277038612  + 956.28915597060 *T)
    sumR1  +=       0.006 * cos(0.59128664983  + 134991.46920492980 *T)
    sumR1  +=       0.005 * cos(1.39435703220  + 104344.49901929598 *T)
    sumR1  +=       0.005 * cos(5.66016557162  + 45290.65639321759 *T)
    sumR1  +=       0.005 * cos(2.69362119172  + 50064.15596977079 *T)
    sumR1  +=       0.005 * cos(4.11526724305  + 209812.60368468691 *T)
    sumR1  +=       0.006 * cos(3.72633598860  + 26162.68474014150 *T)
    sumR1  +=       0.004 * cos(1.01205356829  + 183724.70054311279 *T)
    sumR1  +=       0.005 * cos(2.02484630372  + 52712.61079524379 *T)
    sumR1  +=       0.005 * cos(4.47460304849  + 77410.51304297059 *T)
    sumR1  +=       0.005 * cos(0.42971081835  + 52698.38370124219 *T)
    sumR1  +=       0.005 * cos(1.56442459236  + 11610.55195837420 *T)
    sumR1  +=       0.004 * cos(0.62582613574  + 22747.29071487440 *T)
    sumR1  +=       0.004 * cos(1.51592730997  + 66653.15746634839 *T)
    sumR1  +=       0.005 * cos(6.10277243405  + 97112.93697469680 *T)
    sumR1  +=       0.004 * cos(1.00562353481  + 51219.51712717779 *T)
    sumR1  +=       0.004 * cos(0.31906550066  + 48835.19385644859 *T)
    sumR1  +=       0.005 * cos(4.98644771495  + 51653.22886505459 *T)
    sumR1  +=       0.005 * cos(4.34429010560  + 32132.13172294960 *T)
    sumR1  +=       0.004 * cos(2.92125579020  + 78417.48823520739 *T)
    sumR1  +=       0.004 * cos(5.99601868842  + 121335.60871375341 *T)
    sumR1  +=       0.005 * cos(3.13172791066  + 61279.71327726600 *T)
    sumR1  +=       0.005 * cos(2.13824746365  + 39609.65458316560 *T)
    sumR1  +=       0.005 * cos(5.68604378667  + 119116.85160956658 *T)
    sumR1  +=       0.004 * cos(2.73202591375  + 1692.16566950240 *T)
    sumR1  +=       0.004 * cos(5.78251523942  + 26734.91397488900 *T)
    sumR1  +=       0.005 * cos(0.77777359038  + 103711.71527998280 *T)
    sumR1  +=       0.005 * cos(0.91021434517  + 71582.48457132299 *T)
    sumR1  +=       0.004 * cos(3.03035393128  + 52072.71350892979 *T)
    sumR1  +=       0.004 * cos(0.45156085724  + 31415.37924995700 *T)
    sumR1  +=       0.004 * cos(3.93028686443  + 104371.28232719658 *T)
    sumR1  +=       0.004 * cos(4.54919541216  + 26198.10946279360 *T)
    sumR1  +=       0.004 * cos(2.66986199617  + 24176.70365835700 *T)
    sumR1  +=       0.004 * cos(3.50110688244  + 74821.13447975718 *T)
    sumR1  +=       0.004 * cos(5.75284691398  + 25138.72753260440 *T)
    sumR1  +=       0.004 * cos(5.44367911308  + 72936.23331633979 *T)
    sumR1  +=       0.004 * cos(1.60493978664  + 176953.98994186718 *T)
    sumR1  +=       0.004 * cos(1.93554854399  + 50049.92887576919 *T)
    sumR1  +=       0.003 * cos(3.90053130991  + 107794.18751126219 *T)
    sumR1  +=       0.003 * cos(5.45429683058  + 44295.71712980940 *T)
    sumR1  +=       0.004 * cos(0.60054840458  + 104347.73123093881 *T)
    sumR1  +=       0.004 * cos(0.88305624488  + 60055.89543648739 *T)
    sumR1  +=       0.004 * cos(6.20681068925  + 140652.80125408198 *T)
    sumR1  +=       0.003 * cos(1.44572212087  + 130226.21661243298 *T)
    sumR1  +=       0.004 * cos(4.37750242770  + 32769.12799497380 *T)
    sumR1  +=       0.004 * cos(1.75525459469  + 110634.68841628821 *T)
    sumR1  +=       0.003 * cos(0.66980197410  + 26222.01219115920 *T)
    sumR1  +=       0.004 * cos(2.20364417845  + 182085.63102592481 *T)
    sumR1  +=       0.003 * cos(5.60615263810  + 129387.24732468258 *T)
    sumR1  +=       0.003 * cos(0.96551693568  + 207593.84658050016 *T)
    sumR1  +=       0.004 * cos(4.51472339809  + 52169.17772424839 *T)
    sumR1  +=       0.004 * cos(0.86790293541  + 24822.33566294780 *T)
    sumR1  +=       0.003 * cos(2.05745330337  + 156100.82065856917 *T)
    sumR1  +=       0.004 * cos(0.27139822027  + 214364.55718174577 *T)
    sumR1  +=       0.003 * cos(3.63958954831  + 102232.84870591838 *T)
    sumR1  +=       0.003 * cos(3.33824382272  + 170049.17029103660 *T)
    sumR1  +=       0.003 * cos(6.15959485239  + 12546.48193908300 *T)
    sumR1  +=       0.003 * cos(0.54712514364  + 1911.19948321720 *T)
    sumR1  +=       0.003 * cos(5.53215106225  + 323.50541665740 *T)
    sumR1  +=       0.003 * cos(5.78740917576  + 27780.06881107659 *T)
    sumR1  +=       0.004 * cos(3.60062257669  + 105940.68546158058 *T)
    sumR1  +=       0.003 * cos(5.23668974633  + 17893.62780836560 *T)
    sumR1  +=       0.003 * cos(3.21536557308  + 155571.81910783658 *T)
    sumR1  +=       0.004 * cos(3.78439308193  + 68241.87214462319 *T)
    sumR1  +=       0.004 * cos(0.35026324282  + 53228.07466633679 *T)
    sumR1  +=       0.003 * cos(1.39462627029  + 110.20632121940 *T)
    sumR1  +=       0.003 * cos(2.98585442540  + 102769.65321801380 *T)
    sumR1  +=       0.003 * cos(5.54914760118  + 52101.02468458109 *T)
    sumR1  +=       0.004 * cos(5.31358771393  + 116917.76426627958 *T)
    sumR1  +=       0.004 * cos(5.02170245911  + 130443.39704322898 *T)
    sumR1  +=       0.003 * cos(3.39708771017  + 104778.21075717278 *T)
    sumR1  +=       0.003 * cos(1.35333080979  + 52182.43484204840 *T)
    sumR1  +=       0.003 * cos(3.53872943291  + 28791.51929624980 *T)
    sumR1  +=       0.003 * cos(0.32151031474  + 75615.25459924959 *T)
    sumR1  +=       0.003 * cos(2.81736551277  + 78187.44335344699 *T)
    sumR1  +=       0.003 * cos(3.41370464721  + 77616.69859140778 *T)
    sumR1  +=       0.003 * cos(2.72102083462  + 157586.80077963436 *T)
    sumR1  +=       0.004 * cos(0.37990458940  + 365230.64398203877 *T)
    sumR1  +=       0.003 * cos(5.29398694468  + 183570.92173262799 *T)
    sumR1  +=       0.003 * cos(0.53039476474  + 35077.37084700659 *T)
    sumR1  +=       0.003 * cos(0.96990585284  + 25780.34552060460 *T)
    sumR1  +=       0.004 * cos(5.03799187214  + 103932.12792242158 *T)
    sumR1  +=       0.003 * cos(0.57063116288  + 52250.58788171570 *T)
    sumR1  +=       0.003 * cos(4.90694398172  + 27044.19229754480 *T)
    sumR1  +=       0.003 * cos(5.64342324025  + 78903.60671103658 *T)
    sumR1  +=       0.003 * cos(3.90829945174  + 130446.62925487180 *T)
    sumR1  +=       0.004 * cos(5.39121442451  + 52022.02747266360 *T)
    sumR1  +=       0.003 * cos(5.38325956945  + 172402.03644480839 *T)
    sumR1  +=       0.003 * cos(3.62725946038  + 161079.37234650398 *T)
    sumR1  +=       0.003 * cos(0.02159890719  + 130652.81480330898 *T)
    sumR1  +=       0.003 * cos(5.26052772975  + 55516.41870984820 *T)
    sumR1  +=       0.003 * cos(1.45533607551  + 77101.23472031478 *T)
    sumR1  +=       0.003 * cos(2.45621582046  + 207643.84320240439 *T)
    sumR1  +=       0.003 * cos(3.04264756378  + 74923.09699802278 *T)
    sumR1  +=       0.003 * cos(3.27797530359  + 27669.86248985719 *T)
    sumR1  +=       0.004 * cos(6.20795181771  + 3340.61242669980 *T)
    sumR1  +=       0.003 * cos(2.86818821459  + 79330.20490191258 *T)
    sumR1  +=       0.003 * cos(5.42245453158  + 36109.74042167360 *T)
    sumR1  +=       0.003 * cos(1.11238018002  + 2125.87740737920 *T)
    sumR1  +=       0.003 * cos(4.70325749301  + 52595.29092702359 *T)
    sumR1  +=       0.003 * cos(4.14290639481  + 25446.48957983520 *T)
    sumR1  +=       0.003 * cos(4.11874155864  + 1478.86657406440 *T)
    sumR1  +=       0.003 * cos(4.41727880469  + 128220.75860368418 *T)
    sumR1  +=       0.003 * cos(5.84301330450  + 65831.66677432480 *T)
    sumR1  +=       0.003 * cos(0.63458746349  + 112231.70171963578 *T)
    sumR1  +=       0.003 * cos(5.20358469660  + 157057.10981453978 *T)
    sumR1  +=       0.003 * cos(4.22446997246  + 53399.62412392700 *T)
    sumR1  +=       0.003 * cos(6.08145059717  + 26724.89941359840 *T)
    sumR1  +=       0.003 * cos(3.32246989691  + 111122.32316754240 *T)
    sumR1  +=       0.003 * cos(1.11208497752  + 78378.14871340781 *T)
    sumR1  +=       0.003 * cos(1.18205388535  + 80382.47328510099 *T)
    sumR1  +=       0.003 * cos(3.98183946807  + 25352.02662804239 *T)
    sumR1  +=       0.003 * cos(5.17537238750  + 225687.22128005017 *T)
    sumR1  +=       0.002 * cos(3.19970010923  + 64901.25971792339 *T)
    sumR1  +=       0.003 * cos(5.90368046078  + 104505.39137678158 *T)
    sumR1  +=       0.003 * cos(2.15712281571  + 28736.35796704720 *T)
    sumR1  +=       0.002 * cos(0.27988699718  + 52325.36948028299 *T)
    sumR1  +=       0.002 * cos(1.09392137734  + 24998.19435038059 *T)
    sumR1  +=       0.003 * cos(1.72654636472  + 130419.84594697120 *T)
    sumR1  +=       0.003 * cos(6.24477199702  + 51969.62073471119 *T)
    sumR1  +=       0.003 * cos(3.27154613105  + 77630.92568540938 *T)
    sumR1  +=       0.002 * cos(1.80684819652  + 123668.80510662538 *T)
    sumR1  +=       0.002 * cos(3.75304905922  + 150244.34299945380 *T)
    sumR1  +=       0.003 * cos(2.04107819596  + 76784.84285065818 *T)
    sumR1  +=       0.003 * cos(4.07589828594  + 97670.38771289718 *T)
    sumR1  +=       0.002 * cos(3.43439078149  + 52609.51802102519 *T)
    sumR1  +=       0.002 * cos(0.51530246334  + 133882.09065283637 *T)
    sumR1  +=       0.003 * cos(4.75031746198  + 204151.27163553479 *T)
    sumR1  +=       0.002 * cos(5.57618391474  + 24292.64469785319 *T)
    sumR1  +=       0.002 * cos(2.22207407093  + 78213.71280281838 *T)
    sumR1  +=       0.002 * cos(6.08590562625  + 39450.35284837340 *T)
    sumR1  +=       0.003 * cos(4.42541702442  + 51955.39364070959 *T)
    sumR1  +=       0.003 * cos(5.48323931966  + 220.41264243880 *T)
    sumR1  +=       0.003 * cos(5.24728665819  + 26709.64694241340 *T)
    sumR1  +=       0.003 * cos(4.52126286883  + 20043.67456019880 *T)
    sumR1  +=       0.002 * cos(1.69429787369  + 53029.00266490040 *T)
    sumR1  +=       0.003 * cos(0.98781948776  + 80596.90581759460 *T)
    sumR1  +=       0.002 * cos(2.72291536990  + 23866.04650697719 *T)
    sumR1  +=       0.002 * cos(2.23117481869  + 26402.08932144380 *T)
    sumR1  +=       0.003 * cos(2.32561622864  + 145204.75475114080 *T)
    sumR1  +=       0.003 * cos(3.55406019317  + 52225.80290505260 *T)
    sumR1  +=       0.003 * cos(0.82939608505  + 87253.17713015496 *T)
    sumR1  +=       0.003 * cos(3.59850686303  + 52396.21892558720 *T)
    sumR1  +=       0.002 * cos(0.25549281668  + 22909.75735100660 *T)
    sumR1  +=       0.002 * cos(4.18932113589  + 81591.84508100279 *T)
    sumR1  +=       0.002 * cos(5.58984236569  + 104202.04936916218 *T)
    sumR1  +=       0.002 * cos(1.59325428260  + 206.18554843720 *T)
    sumR1  +=       0.002 * cos(2.33240745085  + 157483.01859105378 *T)
    sumR1  +=       0.002 * cos(3.78499822852  + 78731.67441507700 *T)
    sumR1  +=       0.003 * cos(1.73139148586  + 77741.13200662879 *T)
    sumR1  +=       0.002 * cos(1.98761216006  + 102018.41617342478 *T)
    sumR1  +=       0.002 * cos(3.82447809654  + 1162.47470440780 *T)
    sumR1  +=       0.002 * cos(3.73613869026  + 24601.92302050899 *T)
    sumR1  +=       0.002 * cos(4.87002033008  + 51859.41441349179 *T)
    sumR1  +=       0.002 * cos(0.10623356446  + 111590.28815789679 *T)
    sumR1  +=       0.002 * cos(4.43911582586  + 78270.33798362259 *T)
    sumR1  +=       0.002 * cos(2.47821765975  + 78114.14622758799 *T)
    sumR1  +=       0.002 * cos(5.83346201740  + 181026.24909573561 *T)
    sumR1  +=       0.002 * cos(2.27341561267  + 25953.79409198919 *T)
    sumR1  +=       0.002 * cos(2.73852323809  + 123200.84011627098 *T)
    sumR1  +=       0.002 * cos(2.35063735347  + 78109.93061423779 *T)
    sumR1  +=       0.002 * cos(0.45168929332  + 2111.65031337760 *T)
    sumR1  +=       0.002 * cos(4.46926491505  + 25450.90686955000 *T)
    sumR1  +=       0.002 * cos(3.97278244291  + 52808.59002246159 *T)
    sumR1  +=       0.002 * cos(0.58667782465  + 128320.75184749260 *T)
    sumR1  +=       0.002 * cos(3.05065649587  + 102755.42612401220 *T)
    sumR1  +=       0.002 * cos(1.13486054740  + 52492.19815280499 *T)
    sumR1  +=       0.002 * cos(4.64642529286  + 60370.08161635699 *T)
    sumR1  +=       0.002 * cos(3.68035435205  + 52309.91533273340 *T)
    sumR1  +=       0.002 * cos(5.93910800197  + 78160.61665050399 *T)
    sumR1  +=       0.002 * cos(1.26706232000  + 58857.03113654799 *T)
    sumR1  +=       0.002 * cos(5.09637370849  + 26823.77965510599 *T)
    sumR1  +=       0.002 * cos(3.88808196008  + 16703.06213349900 *T)
    sumR1  +=       0.002 * cos(0.00715741623  + 130866.11389874700 *T)
    sumR1  +=       0.002 * cos(3.63338095338  + 51852.30086649099 *T)
    sumR1  +=       0.002 * cos(5.67927933196  + 103917.90082841998 *T)
    sumR1  +=       0.002 * cos(1.35964203538  + 3328.13565628019 *T)
    sumR1  +=       0.002 * cos(6.13531665590  + 129373.02023068098 *T)
    sumR1  +=       0.002 * cos(3.74140255139  + 52381.99183158559 *T)
    sumR1  +=       0.002 * cos(0.37783717379  + 53771.99272543299 *T)
    sumR1  +=       0.002 * cos(0.70779949763  + 26667.59072857300 *T)
    sumR1  +=       0.002 * cos(2.05372686428  + 70383.62027138360 *T)
    sumR1  +=       0.002 * cos(6.06999395703  + 23919.14265929160 *T)
    sumR1  +=       0.002 * cos(1.22493486348  + 103498.41618454478 *T)
    sumR1  +=       0.002 * cos(4.00058813480  + 129799.61842155698 *T)
    sumR1  +=       0.002 * cos(1.62085169147  + 78149.27013603736 *T)
    sumR1  +=       0.002 * cos(1.81898501228  + 25771.51127191760 *T)
    sumR1  +=       0.002 * cos(6.22451796390  + 52252.07235442399 *T)
    sumR1  +=       0.002 * cos(0.20849484282  + 25754.04720080480 *T)
    sumR1  +=       0.002 * cos(2.38916713246  + 51226.63067417859 *T)
    sumR1  +=       0.002 * cos(4.34697979051  + 26404.29501123080 *T)
    sumR1  +=       0.002 * cos(2.87638496001  + 167850.08294774959 *T)
    sumR1  +=       0.002 * cos(4.01638193690  + 51868.24866217880 *T)
    sumR1  +=       0.002 * cos(3.38565068956  + 77726.90491262719 *T)
    sumR1  +=       0.002 * cos(5.73512271075  + 189853.99758261937 *T)
    sumR1  +=       0.002 * cos(3.93444553575  + 50910.23880452200 *T)
    sumR1  +=       0.002 * cos(3.62039658007  + 6885.14988993081 *T)
    sumR1  +=       0.002 * cos(1.76492939350  + 92741.06060792258 *T)
    sumR1  +=       0.002 * cos(2.58190149030  + 154938.34595416137 *T)
    sumR1  +=       0.002 * cos(5.37279792857  + 52278.89905736699 *T)
    sumR1  +=       0.002 * cos(3.37039086197  + 2221.85663459700 *T)
    sumR1  +=       0.002 * cos(6.07737780217  + 94329.77528619740 *T)
    sumR1  +=       0.002 * cos(5.16106622649  + 26729.31670331319 *T)
    sumR1  +=       0.002 * cos(1.58212922198  + 26013.12154300690 *T)
    sumR1  +=       0.002 * cos(5.77714525422  + 25600.26839032000 *T)
    sumR1  +=       0.002 * cos(1.88481234244  + 27177.61193276780 *T)
    sumR1  +=       0.002 * cos(5.68695965343  + 50264.60679993120 *T)
    sumR1  +=       0.002 * cos(6.07601531670  + 141762.17980617538 *T)
    sumR1  +=       0.002 * cos(4.33545610317  + 8194.27533320860 *T)
    sumR1  +=       0.002 * cos(1.10189365852  + 1223.81784077861 *T)
    sumR1  +=       0.002 * cos(1.68048369897  + 27250.37784598199 *T)
    sumR1  +=       0.002 * cos(5.32044002117  + 151199.94274106238 *T)
    sumR1  +=       0.002 * cos(1.33411828477  + 78257.08086582259 *T)
    sumR1  +=       0.002 * cos(5.40842531044  + 95247.70557217918 *T)
    sumR1  +=       0.002 * cos(4.55349965747  + 37698.45509994840 *T)
    sumR1  +=       0.002 * cos(2.71889214634  + 173511.41499690176 *T)
    sumR1  +=       0.002 * cos(1.95076646341  + 143005.66740785379 *T)
    sumR1  +=       0.002 * cos(2.00346690661  + 26073.67604757259 *T)
    sumR1  +=       0.002 * cos(1.78781584986  + 48847.67062686820 *T)
    sumR1  +=       0.002 * cos(5.89586766332  + 26312.24793727610 *T)
    sumR1  +=       0.002 * cos(1.99160321093  + 26189.86565983980 *T)
    sumR1  +=       0.002 * cos(0.40598906242  + 25773.71696170459 *T)
    sumR1  +=       0.002 * cos(3.36946627953  + 26575.53789282839 *T)
    sumR1  +=       0.002 * cos(4.77845331095  + 136722.59155786238 *T)
    sumR1  +=       0.002 * cos(1.11920130054  + 51315.49635439559 *T)
    sumR1  +=       0.002 * cos(4.99966961564  + 78800.51393681798 *T)
    sumR1  +=       0.002 * cos(2.50164995664  + 19202.75325164339 *T)
    sumR1  +=       0.002 * cos(5.31192385841  + 23384.28698689860 *T)
    sumR1  +=       0.002 * cos(2.93563622378  + 137678.19129947099 *T)
    sumR1  +=       0.002 * cos(5.00974264854  + 182188.72380014337 *T)
    sumR1  +=       0.002 * cos(2.82066370220  + 118828.96374949680 *T)
    sumR1  +=       0.002 * cos(1.23355180586  + 52286.01260436779 *T)
    sumR1  +=       0.002 * cos(0.27866926000  + 24356.78078864160 *T)
    sumR1  +=       0.002 * cos(2.19163720361  + 26395.46076254379 *T)
    sumR1  +=       0.002 * cos(2.41579406356  + 147423.51185532758 *T)
    sumR1  +=       0.002 * cos(1.80661945912  + 860.30992875280 *T)
    sumR1  +=       0.001 * cos(3.75687703582  + 34082.43158359840 *T)
    sumR1  +=       0.001 * cos(5.85918351804  + 163766.09444104519 *T)
    sumR1  +=       0.001 * cos(4.84627694564  + 27170.98337386779 *T)
    sumR1  +=       0.002 * cos(3.12958128871  + 195047.36464141720 *T)
    sumR1  +=       0.002 * cos(0.59313712363  + 74.78159856730 *T)
    sumR1  +=       0.002 * cos(4.60111717878  + 25004.82290928060 *T)
    sumR1  +=       0.001 * cos(1.14751201091  + 55503.94193942859 *T)
    sumR1  +=       0.002 * cos(4.55507467591  + 132658.27281205778 *T)
    sumR1  +=       0.001 * cos(1.61844283111  + 24079.34560241519 *T)
    sumR1  +=       0.002 * cos(1.99970840056  + 17098.43543614181 *T)
    sumR1  +=       0.002 * cos(4.95650007002  + 24448.83362438620 *T)
    sumR1  +=       0.001 * cos(6.20512620421  + 2333.19639287200 *T)
    sumR1  +=       0.002 * cos(1.27464400058  + 144916.86689107097 *T)
    sumR1  +=       0.002 * cos(6.17677089224  + 87367.61641884019 *T)
    sumR1  +=       0.001 * cos(1.29983905478  + 154194.22245657316 *T)
    sumR1  +=       0.002 * cos(4.47232796046  + 130432.40216087016 *T)
    sumR1  +=       0.001 * cos(4.48574397522  + 101703.15774082378 *T)
    sumR1  +=       0.002 * cos(2.19277171383  + 103.09277421860 *T)
    sumR1  +=       0.001 * cos(1.80796273878  + 25508.21555457540 *T)
    sumR1  +=       0.001 * cos(5.48664013700  + 22625.65843570900 *T)
    sumR1  +=       0.001 * cos(0.74294839995  + 78896.49316403578 *T)
    sumR1  +=       0.001 * cos(3.51165251141  + 26421.75908234360 *T)
    sumR1  +=       0.001 * cos(5.55668995710  + 220025.88923089800 *T)
    sumR1  +=       0.002 * cos(4.08809950869  + 25466.15934073500 *T)
    sumR1  +=       0.002 * cos(2.89128653047  + 166740.70439565618 *T)
    sumR1  +=       0.001 * cos(3.92791860433  + 189386.03259226491 *T)
    sumR1  +=       0.001 * cos(0.33128467520  + 76571.54375522019 *T)
    sumR1  +=       0.002 * cos(1.98187462633  + 81604.32185142238 *T)
    sumR1  +=       0.001 * cos(3.57210475329  + 78786.28684281638 *T)
    sumR1  +=       0.002 * cos(1.13879447241  + 61165.27398858079 *T)
    sumR1  +=       0.001 * cos(5.62721081889  + 122444.98726584678 *T)
    sumR1  +=       0.001 * cos(0.51173156424  + 113341.08027172917 *T)
    sumR1  +=       0.001 * cos(0.63238768568  + 130459.18546877075 *T)
    sumR1  +=       0.001 * cos(0.62172861766  + 138319.60486120995 *T)
    sumR1  +=       0.002 * cos(0.35135879351  + 132028.58860315481 *T)
    sumR1  +=       0.001 * cos(1.68473271541  + 28286.99048486120 *T)
    sumR1  +=       0.002 * cos(2.61859360479  + 61560.64729122359 *T)
    sumR1  +=       0.001 * cos(3.42727473387  + 130435.63437251298 *T)
    sumR1  +=       0.002 * cos(3.42836265136  + 65717.22748563958 *T)
    sumR1  +=       0.001 * cos(2.98248327032  + 846.08283475120 *T)
    sumR1  +=       0.001 * cos(5.35146960556  + 1265.56747862640 *T)
    sumR1  +=       0.001 * cos(4.88379701920  + 205260.65018762811 *T)
    sumR1  +=       0.001 * cos(1.38681194099  + 467.96499035440 *T)
    sumR1  +=       0.001 * cos(5.78967137689  + 76152.05911134499 *T)
    sumR1  +=       0.001 * cos(1.64625172020  + 75930.51303185058 *T)
    sumR1  +=       0.001 * cos(0.08732855602  + 103718.82882698360 *T)
    sumR1  +=       0.001 * cos(0.74831854153  + 51534.39272140940 *T)
    sumR1  +=       0.001 * cos(1.98744571247  + 26411.40855823160 *T)
    sumR1  +=       0.002 * cos(0.58716824437  + 52065.59996192899 *T)
    sumR1  +=       0.001 * cos(5.83388253183  + 104275.34649502118 *T)
    sumR1  +=       0.001 * cos(2.07307900948  + 78188.92782615528 *T)
    sumR1  +=       0.001 * cos(4.54379329814  + 203041.89308344136 *T)
    sumR1  +=       0.001 * cos(3.67961974137  + 116783.65521669458 *T)
    sumR1  +=       0.001 * cos(2.76419827660  + 52812.80255517260 *T)
    sumR1  +=       0.001 * cos(5.76246998806  + 1375.77379984580 *T)
    sumR1  +=       0.001 * cos(3.49542841313  + 79315.97780791098 *T)
    sumR1  +=       0.001 * cos(3.09408892186  + 9384.84100807520 *T)
    sumR1  +=       0.002 * cos(2.99230815235  + 91805.13062721380 *T)
    sumR1  +=       0.001 * cos(2.31818985780  + 58220.03486452380 *T)
    sumR1  +=       0.001 * cos(2.52061585089  + 91919.56991589900 *T)
    sumR1  +=       0.001 * cos(0.20823485438  + 54879.42243782400 *T)
    sumR1  +=       0.001 * cos(4.44457464944  + 6044.22858137540 *T)
    sumR1  +=       0.001 * cos(0.49331056411  + 104819.57755665120 *T)
    sumR1  +=       0.001 * cos(5.98062907857  + 106570.36967048360 *T)
    sumR1  +=       0.001 * cos(3.43578051874  + 51841.95034237900 *T)
    sumR1  +=       0.001 * cos(1.50478057782  + 137210.22630911658 *T)
    sumR1  +=       0.001 * cos(0.85871789181  + 25985.94062330859 *T)
    sumR1  +=       0.001 * cos(0.00464768227  + 196137.07343261081 *T)
    sumR1  +=       0.001 * cos(4.31601888279  + 50483.64061364600 *T)
    sumR1  +=       0.001 * cos(3.82192552665  + 25928.60140678200 *T)
    sumR1  +=       0.001 * cos(2.13422641639  + 62197.64356324779 *T)
    sumR1  +=       0.001 * cos(0.26342054765  + 1795.25844372100 *T)
    sumR1  +=       0.001 * cos(3.33069384141  + 19.66976089979 *T)
    sumR1  +=       0.001 * cos(5.04672208460  + 26247.20487636640 *T)
    sumR1  +=       0.001 * cos(3.69091550110  + 77307.42026875199 *T)
    sumR1  +=       0.001 * cos(0.68629642431  + 26102.13023557580 *T)
    sumR1  +=       0.001 * cos(5.61963025660  + 104197.83375581198 *T)
    sumR1  +=       0.001 * cos(0.41026701894  + 636.99627202420 *T)
    sumR1  +=       0.001 * cos(2.73690252229  + 170068.84005193636 *T)
    sumR1  +=       0.001 * cos(2.97981003623  + 78057.52387628538 *T)
    sumR1  +=       0.001 * cos(2.84174115238  + 168959.46149984296 *T)
    sumR1  +=       0.001 * cos(6.19695864555  + 241561.83887541335 *T)
    sumR1  +=       0.001 * cos(2.31427972623  + 179172.74704605396 *T)
    sumR1  +=       0.001 * cos(0.82072277702  + 51329.72344839720 *T)
    sumR1  +=       0.001 * cos(5.23954201903  + 99024.13645791401 *T)
    sumR1  +=       0.001 * cos(1.86269361791  + 45494.58142974879 *T)
    sumR1  +=       0.001 * cos(3.54450921353  + 78338.49102328988 *T)
    sumR1  +=       0.001 * cos(4.60745432488  + 156507.74908854539 *T)
    sumR1  +=       0.001 * cos(1.62346464256  + 128106.31931499895 *T)
    sumR1  +=       0.001 * cos(0.72186155433  + 27005.83342755599 *T)
    sumR1  +=       0.001 * cos(2.30344666604  + 50380.54783942739 *T)
    sumR1  +=       0.001 * cos(1.18658609813  + 8989.46770543239 *T)
    sumR1  +=       0.001 * cos(6.25857682168  + 50800.03248330259 *T)
    sumR1  +=       0.001 * cos(0.99962008973  + 163298.12945069079 *T)
    sumR1  +=       0.001 * cos(2.86790873835  + 38634.38508065720 *T)
    sumR1  +=       0.001 * cos(3.61827980669  + 106262.81204951400 *T)
    sumR1  +=       0.001 * cos(0.61515728295  + 78469.89497315978 *T)
    sumR1  +=       0.001 * cos(4.60636743786  + 126996.94076290558 *T)
    sumR1  +=       0.001 * cos(4.48187053152  + 231348.55332920240 *T)
    sumR1  +=       0.001 * cos(5.19829793680  + 142871.55835826878 *T)
    sumR1  +=       0.001 * cos(5.68864830775  + 3178.14579056760 *T)
    sumR1  +=       0.001 * cos(4.89753750946  + 53265.51507434200 *T)
    sumR1  +=       0.001 * cos(3.68051221186  + 2703.61615467560 *T)
    sumR1  +=       0.001 * cos(0.01737531330  + 103814.80805420138 *T)
    sumR1  +=       0.001 * cos(1.79348974875  + 183145.01295611399 *T)
    sumR1  +=       0.001 * cos(6.16564988887  + 199599.31813847597 *T)
    sumR1  +=       0.001 * cos(0.63301837081  + 51439.92976961659 *T)
    sumR1  +=       0.001 * cos(2.29427904310  + 1485.98012106520 *T)
    sumR1  +=       0.001 * cos(1.44184009822  + 112545.88789950538 *T)
    sumR1  +=       0.001 * cos(0.55411127132  + 42790.96527507320 *T)
    sumR1  +=       0.001 * cos(5.07507945312  + 54824.26110862140 *T)
    sumR1  +=       0.001 * cos(5.10858625008  + 107679.74822257696 *T)
    sumR1  +=       0.001 * cos(5.50068575093  + 27819.02549450680 *T)
    sumR1  +=       0.001 * cos(5.60579347269  + 153084.84390447981 *T)
    sumR1  +=       0.001 * cos(4.92352598929  + 51596.11869614960 *T)
    sumR1  +=       0.001 * cos(6.10538330268  + 181659.72224941081 *T)

    # MERCURY Radius
    sumR2  = 0
    sumR2  +=    3117.867 * cos(3.08231840296  + 26087.90314157420 *T)
    sumR2  +=    1245.396 * cos(6.15183317423  + 52175.80628314840 *T)
    sumR2  +=     424.822 * cos(2.92583352960  + 78263.70942472259 *T)
    sumR2  +=     136.130 * cos(5.97983925842  + 104351.61256629678 *T)
    sumR2  +=      42.175 * cos(2.74936980629  + 130439.51570787100 *T)
    sumR2  +=      21.759 * cos(3.14159265359  + 0.00000000000 *T)
    sumR2  +=      12.793 * cos(5.80143162209  + 156527.41884944518 *T)
    sumR2  +=       3.825 * cos(2.56993599584  + 182615.32199101936 *T)
    sumR2  +=       1.042 * cos(3.14648120079  + 24978.52458948080 *T)
    sumR2  +=       1.131 * cos(5.62142196970  + 208703.22513259359 *T)
    sumR2  +=       0.483 * cos(6.14307654520  + 27197.28169366760 *T)
    sumR2  +=       0.332 * cos(2.38992575670  + 234791.12827416777 *T)
    sumR2  +=       0.320 * cos(6.20674766565  + 51066.42773105500 *T)
    sumR2  +=       0.120 * cos(0.56745598887  + 20426.57109242200 *T)
    sumR2  +=       0.113 * cos(3.28048907943  + 24498.83024629040 *T)
    sumR2  +=       0.105 * cos(4.36663468239  + 25028.52121138500 *T)
    sumR2  +=       0.097 * cos(5.44128884026  + 260879.03141574195 *T)
    sumR2  +=       0.095 * cos(1.70276410782  + 1059.38193018920 *T)
    sumR2  +=       0.103 * cos(2.98038190305  + 77154.33087262919 *T)
    sumR2  +=       0.074 * cos(1.28630381800  + 26617.59410666880 *T)
    sumR2  +=       0.057 * cos(2.96753362980  + 1109.37855209340 *T)
    sumR2  +=       0.066 * cos(5.68201233964  + 25132.30339996560 *T)
    sumR2  +=       0.053 * cos(3.63699398162  + 46514.47423399620 *T)
    sumR2  +=       0.059 * cos(3.10922979724  + 27043.50288318280 *T)
    sumR2  +=       0.053 * cos(0.66673109494  + 21535.94964451540 *T)
    sumR2  +=       0.043 * cos(4.69270605698  + 27147.28507176339 *T)
    sumR2  +=       0.042 * cos(2.77372510028  + 1589.07289528380 *T)
    sumR2  +=       0.036 * cos(4.39741248058  + 14765.23904326980 *T)
    sumR2  +=       0.035 * cos(1.80468389323  + 11322.66409830440 *T)
    sumR2  +=       0.043 * cos(6.07426254990  + 28306.66024576099 *T)
    sumR2  +=       0.033 * cos(5.34572581657  + 25448.00585526019 *T)
    sumR2  +=       0.031 * cos(5.66489075070  + 30639.85663863300 *T)
    sumR2  +=       0.031 * cos(2.36104565226  + 32858.61374281979 *T)
    sumR2  +=       0.031 * cos(4.27403143366  + 16983.99614745660 *T)
    sumR2  +=       0.033 * cos(3.22976052744  + 23869.14603738740 *T)
    sumR2  +=       0.031 * cos(0.50041076120  + 22645.32819660879 *T)
    sumR2  +=       0.028 * cos(2.21390393577  + 286966.93455731618 *T)
    sumR2  +=       0.033 * cos(6.03682885493  + 103242.23401420339 *T)
    sumR2  +=       0.025 * cos(4.23934105457  + 52705.49724824299 *T)
    sumR2  +=       0.023 * cos(2.59711556958  + 23969.13928119580 *T)
    sumR2  +=       0.027 * cos(2.10665521901  + 25558.21217647960 *T)
    sumR2  +=       0.021 * cos(0.40892383580  + 72602.37737557039 *T)
    sumR2  +=       0.025 * cos(5.69513499802  + 5661.33204915220 *T)
    sumR2  +=       0.022 * cos(4.86064608327  + 26301.20223701220 *T)
    sumR2  +=       0.024 * cos(0.66522740624  + 19317.19254032860 *T)
    sumR2  +=       0.020 * cos(5.61847810688  + 4551.95349705880 *T)
    sumR2  +=       0.021 * cos(2.30083904827  + 51220.20654153979 *T)
    sumR2  +=       0.018 * cos(0.48839037250  + 26068.23338067440 *T)
    sumR2  +=       0.016 * cos(6.07779022342  + 50586.73338786459 *T)
    sumR2  +=       0.016 * cos(1.47218845921  + 53235.18821333759 *T)
    sumR2  +=       0.017 * cos(2.94277588966  + 53285.18483524180 *T)
    sumR2  +=       0.015 * cos(2.39097980793  + 83925.04147387479 *T)
    sumR2  +=       0.015 * cos(3.85502950028  + 26091.78447693220 *T)
    sumR2  +=       0.014 * cos(4.13753968603  + 12566.15169998280 *T)
    sumR2  +=       0.015 * cos(5.51545801304  + 6770.71060124560 *T)
    sumR2  +=       0.014 * cos(5.67708505049  + 57837.13833230060 *T)
    sumR2  +=       0.014 * cos(5.73775080815  + 27676.97603685800 *T)
    sumR2  +=       0.016 * cos(5.69298300679  + 26107.57290247399 *T)
    sumR2  +=       0.014 * cos(5.00837046884  + 37410.56723987860 *T)
    sumR2  +=       0.015 * cos(5.72321978212  + 29530.47808653960 *T)
    sumR2  +=       0.012 * cos(1.18827026407  + 40853.14218484400 *T)
    sumR2  +=       0.013 * cos(2.34101240702  + 26084.02180621620 *T)
    sumR2  +=       0.012 * cos(2.43920574243  + 56727.75978020720 *T)
    sumR2  +=       0.012 * cos(1.12086684687  + 43071.89928903080 *T)
    sumR2  +=       0.012 * cos(1.86975778356  + 9103.90699411760 *T)
    sumR2  +=       0.012 * cos(4.90236416950  + 36301.18868778519 *T)
    sumR2  +=       0.010 * cos(2.33297570685  + 426.59819087600 *T)
    sumR2  +=       0.011 * cos(2.80390774003  + 129330.13715577760 *T)
    sumR2  +=       0.010 * cos(5.86382886189  + 26202.34243025941 *T)
    sumR2  +=       0.009 * cos(0.03545904481  + 25234.70675982219 *T)
    sumR2  +=       0.011 * cos(0.92571990717  + 639.89728631400 *T)
    sumR2  +=       0.010 * cos(3.94827815051  + 25661.30495069820 *T)
    sumR2  +=       0.009 * cos(0.00145195768  + 49957.04917896160 *T)
    sumR2  +=       0.008 * cos(5.42702003399  + 110012.94461544900 *T)
    sumR2  +=       0.008 * cos(3.39306818321  + 25973.46385288896 *T)
    sumR2  +=       0.009 * cos(4.83499870418  + 38519.94579197200 *T)
    sumR2  +=       0.008 * cos(5.65148718768  + 53131.40602475700 *T)
    sumR2  +=       0.009 * cos(5.73812314204  + 51116.42435295920 *T)
    sumR2  +=       0.008 * cos(1.98050481337  + 13521.75144159140 *T)
    sumR2  +=       0.009 * cos(0.95963913886  + 78793.40038981718 *T)
    sumR2  +=       0.008 * cos(5.24647237181  + 313054.83769889042 *T)
    sumR2  +=       0.008 * cos(3.52212788936  + 52156.13652224860 *T)
    sumR2  +=       0.007 * cos(3.57060093012  + 48733.23133818299 *T)
    sumR2  +=       0.008 * cos(3.47359025801  + 98690.28051714458 *T)
    sumR2  +=       0.008 * cos(5.32679971128  + 51646.11531805379 *T)
    sumR2  +=       0.010 * cos(5.07319693095  + 25874.60404613620 *T)
    sumR2  +=       0.008 * cos(4.34411026261  + 31749.23519072640 *T)
    sumR2  +=       0.007 * cos(3.44556423722  + 26080.78959457339 *T)
    sumR2  +=       0.007 * cos(5.28088749622  + 77308.10968311399 *T)
    sumR2  +=       0.007 * cos(3.46667099767  + 2118.76386037840 *T)
    sumR2  +=       0.007 * cos(1.96360039314  + 10213.28554621100 *T)
    sumR2  +=       0.007 * cos(1.18631489114  + 103292.23063610760 *T)
    sumR2  +=       0.007 * cos(4.28693259279  + 955.59974160860 *T)
    sumR2  +=       0.007 * cos(0.88520158736  + 38654.05484155699 *T)
    sumR2  +=       0.009 * cos(2.12100150030  + 18849.22754997420 *T)
    sumR2  +=       0.008 * cos(2.75301666365  + 26095.01668857500 *T)
    sumR2  +=       0.007 * cos(4.98530328971  + 529.69096509460 *T)
    sumR2  +=       0.007 * cos(0.20928215626  + 33326.57873317420 *T)
    sumR2  +=       0.006 * cos(4.54286088482  + 79323.09135491178 *T)
    sumR2  +=       0.006 * cos(0.60060875587  + 52179.68761850640 *T)
    sumR2  +=       0.006 * cos(1.21940434381  + 41962.52073693740 *T)
    sumR2  +=       0.006 * cos(1.74585091288  + 62389.09182935939 *T)
    sumR2  +=       0.006 * cos(2.13481309441  + 51535.90899683439 *T)
    sumR2  +=       0.006 * cos(4.92478078408  + 26514.50133245020 *T)
    sumR2  +=       0.007 * cos(4.18249532118  + 77204.32749453338 *T)
    sumR2  +=       0.007 * cos(1.68449319925  + 52389.10537858640 *T)
    sumR2  +=       0.005 * cos(4.28220819106  + 68050.42387851159 *T)
    sumR2  +=       0.007 * cos(1.69076988754  + 12432.04265039780 *T)
    sumR2  +=       0.006 * cos(3.73803069606  + 45405.09568190280 *T)
    sumR2  +=       0.005 * cos(2.61320374120  + 79373.08797681599 *T)
    sumR2  +=       0.005 * cos(2.82458540907  + 54394.56338733519 *T)
    sumR2  +=       0.004 * cos(6.27439443684  + 52168.69273614759 *T)
    sumR2  +=       0.005 * cos(4.50767453332  + 13655.86049117640 *T)
    sumR2  +=       0.006 * cos(2.41451327257  + 52195.47604404819 *T)
    sumR2  +=       0.005 * cos(5.35396003655  + 52171.92494779040 *T)
    sumR2  +=       0.005 * cos(5.48187638039  + 82815.66292178139 *T)
    sumR2  +=       0.004 * cos(2.46695073802  + 55618.38122811380 *T)
    sumR2  +=       0.004 * cos(2.40470389725  + 53764.87917843220 *T)
    sumR2  +=       0.004 * cos(4.24591997877  + 66941.04532641819 *T)
    sumR2  +=       0.004 * cos(3.26730161531  + 7238.67559160000 *T)
    sumR2  +=       0.004 * cos(4.24517841791  + 129380.13377768178 *T)
    sumR2  +=       0.004 * cos(5.46706523938  + 50057.04242277000 *T)
    sumR2  +=       0.004 * cos(2.17208918484  + 136100.84775702321 *T)
    sumR2  +=       0.004 * cos(5.86573287261  + 155418.04029735178 *T)
    sumR2  +=       0.004 * cos(4.15256069197  + 32370.97899156560 *T)
    sumR2  +=       0.004 * cos(2.92358717546  + 2218.75710418680 *T)
    sumR2  +=       0.004 * cos(2.22141015932  + 33967.99229491319 *T)
    sumR2  +=       0.005 * cos(5.74788858603  + 52182.91983014920 *T)
    sumR2  +=       0.004 * cos(0.97295961632  + 44181.27784112419 *T)
    sumR2  +=       0.003 * cos(5.78273085505  + 28421.09953444620 *T)
    sumR2  +=       0.003 * cos(3.65975480266  + 99799.65906923798 *T)
    sumR2  +=       0.004 * cos(3.98623592167  + 26727.80042788820 *T)
    sumR2  +=       0.003 * cos(2.65056124435  + 76674.63652943878 *T)
    sumR2  +=       0.003 * cos(4.91152657563  + 47623.85278608960 *T)
    sumR2  +=       0.004 * cos(1.60135504353  + 77734.01845962799 *T)
    sumR2  +=       0.003 * cos(0.26868606771  + 78244.03966382280 *T)
    sumR2  +=       0.003 * cos(4.58298191655  + 1066.49547719000 *T)
    sumR2  +=       0.003 * cos(4.04696774318  + 18093.37469954999 *T)
    sumR2  +=       0.003 * cos(1.04981780528  + 94138.32702008578 *T)
    sumR2  +=       0.003 * cos(0.82425633309  + 73711.75592766379 *T)
    sumR2  +=       0.003 * cos(5.12180734718  + 1052.26838318840 *T)
    sumR2  +=       0.003 * cos(1.58544193974  + 51109.31080595839 *T)
    sumR2  +=       0.003 * cos(0.22337858409  + 124778.18365871880 *T)
    sumR2  +=       0.003 * cos(3.08207297479  + 76044.95232053580 *T)
    sumR2  +=       0.003 * cos(6.26000429251  + 853.19638175200 *T)
    sumR2  +=       0.003 * cos(3.89747679105  + 64741.95798313119 *T)
    sumR2  +=       0.003 * cos(1.86869273254  + 52602.40447402440 *T)
    sumR2  +=       0.004 * cos(0.01480665337  + 7.11354700080 *T)
    sumR2  +=       0.003 * cos(1.95420622178  + 103396.01282468818 *T)
    sumR2  +=       0.003 * cos(6.00022344367  + 29416.03879785439 *T)
    sumR2  +=       0.003 * cos(1.05827389660  + 51123.53789995999 *T)
    sumR2  +=       0.003 * cos(3.93336187090  + 104881.30353139140 *T)
    sumR2  +=       0.002 * cos(4.80666683935  + 88476.99497093359 *T)
    sumR2  +=       0.002 * cos(2.56919394798  + 51962.50718771040 *T)
    sumR2  +=       0.002 * cos(5.19328838214  + 91785.46086631398 *T)
    sumR2  +=       0.003 * cos(4.94448744333  + 44937.13069154840 *T)
    sumR2  +=       0.002 * cos(0.39272598961  + 23754.70674870219 *T)
    sumR2  +=       0.003 * cos(3.64926786888  + 78267.59076008058 *T)
    sumR2  +=       0.002 * cos(1.93776267237  + 117873.36400788820 *T)
    sumR2  +=       0.002 * cos(0.16310879863  + 28206.66700195260 *T)
    sumR2  +=       0.002 * cos(2.74232958068  + 52290.24557183361 *T)
    sumR2  +=       0.002 * cos(1.94114905421  + 7994.52844202420 *T)
    sumR2  +=       0.002 * cos(0.39437121176  + 125887.56221081220 *T)
    sumR2  +=       0.002 * cos(1.03418813348  + 155468.03691925600 *T)
    sumR2  +=       0.002 * cos(2.06866763231  + 339142.74084046454 *T)
    sumR2  +=       0.002 * cos(1.29912216629  + 105410.99449648600 *T)
    sumR2  +=       0.002 * cos(0.54124189090  + 74821.13447975718 *T)
    sumR2  +=       0.002 * cos(2.12346357366  + 65697.55772473979 *T)
    sumR2  +=       0.002 * cos(2.54143985834  + 78270.82297172339 *T)
    sumR2  +=       0.002 * cos(4.87063627681  + 35191.81013569180 *T)
    sumR2  +=       0.002 * cos(0.51631216786  + 71492.99882347698 *T)
    sumR2  +=       0.002 * cos(0.73976887550  + 18207.81398823521 *T)
    sumR2  +=       0.002 * cos(5.50150244414  + 78283.37918562238 *T)
    sumR2  +=       0.002 * cos(4.78661791721  + 78477.00852016058 *T)
    sumR2  +=       0.002 * cos(4.40683761496  + 103821.92160120218 *T)
    sumR2  +=       0.002 * cos(3.07092516027  + 78256.59587772179 *T)
    sumR2  +=       0.002 * cos(5.01301614244  + 3442.57494496540 *T)
    sumR2  +=       0.002 * cos(2.58799752737  + 181505.94343892601 *T)
    sumR2  +=       0.002 * cos(4.95243490606  + 129483.91596626240 *T)
    sumR2  +=       0.002 * cos(4.12821833516  + 77211.44104153418 *T)
    sumR2  +=       0.002 * cos(4.05475453577  + 120226.23016166000 *T)
    sumR2  +=       0.002 * cos(3.34041291655  + 151975.46535238638 *T)
    sumR2  +=       0.002 * cos(3.27812728356  + 51322.60990139639 *T)
    sumR2  +=       0.002 * cos(4.06092192484  + 77837.11123384659 *T)
    sumR2  +=       0.002 * cos(5.22546741404  + 162188.75089859738 *T)
    sumR2  +=       0.002 * cos(5.57497617017  + 71980.63357473118 *T)
    sumR2  +=       0.002 * cos(1.86655156569  + 23439.44831610119 *T)
    sumR2  +=       0.002 * cos(4.79020891265  + 15874.61759536320 *T)
    sumR2  +=       0.002 * cos(2.43202419279  + 45892.73043315699 *T)
    sumR2  +=       0.001 * cos(4.91443985635  + 78690.30761559859 *T)
    sumR2  +=       0.002 * cos(2.55186486587  + 39609.65458316560 *T)
    sumR2  +=       0.002 * cos(2.04705904795  + 433.71173787680 *T)
    sumR2  +=       0.001 * cos(6.06582167254  + 102132.85546211000 *T)
    sumR2  +=       0.002 * cos(0.36512406083  + 6283.07584999140 *T)
    sumR2  +=       0.002 * cos(5.96418661485  + 78050.41032928458 *T)
    sumR2  +=       0.001 * cos(0.66730245858  + 58458.88213313979 *T)
    sumR2  +=       0.001 * cos(2.46652537811  + 98068.53671630540 *T)
    sumR2  +=       0.002 * cos(3.24884087108  + 150866.08680029298 *T)
    sumR2  +=       0.002 * cos(0.33626385981  + 52061.36699446317 *T)
    sumR2  +=       0.001 * cos(1.58214319162  + 114564.89811250778 *T)
    sumR2  +=       0.001 * cos(1.86505965946  + 27140.17152476259 *T)
    sumR2  +=       0.001 * cos(4.70954784660  + 77197.21394753258 *T)
    sumR2  +=       0.001 * cos(3.30080224324  + 104331.94280539700 *T)
    sumR2  +=       0.002 * cos(2.15014442693  + 78259.82808936459 *T)
    sumR2  +=       0.001 * cos(4.16909637294  + 51742.09454527159 *T)
    sumR2  +=       0.002 * cos(0.61993180902  + 130969.20667296558 *T)
    sumR2  +=       0.001 * cos(5.64166202504  + 81706.28436968799 *T)
    sumR2  +=       0.001 * cos(1.28449131177  + 27154.39861876420 *T)
    sumR2  +=       0.001 * cos(0.20716724934  + 131548.89425996438 *T)
    sumR2  +=       0.001 * cos(5.01586042695  + 105460.99111839020 *T)
    sumR2  +=       0.001 * cos(2.20142137730  + 104371.28232719658 *T)
    sumR2  +=       0.001 * cos(0.71547903080  + 103925.01437542078 *T)
    sumR2  +=       0.001 * cos(1.69298852699  + 71025.03383312259 *T)
    sumR2  +=       0.002 * cos(1.01857337375  + 93028.94846799241 *T)
    sumR2  +=       0.002 * cos(0.99564225222  + 129909.82474277640 *T)
    sumR2  +=       0.001 * cos(3.04953117213  + 26941.09952332620 *T)
    sumR2  +=       0.001 * cos(1.54774252742  + 104564.91166173480 *T)
    sumR2  +=       0.001 * cos(1.29736654025  + 26241.68195205900 *T)
    sumR2  +=       0.002 * cos(0.39879309478  + 104355.49390165480 *T)
    sumR2  +=       0.001 * cos(1.73442408485  + 64607.84893354619 *T)
    sumR2  +=       0.001 * cos(3.52887677711  + 97580.90196505121 *T)
    sumR2  +=       0.001 * cos(3.55488474055  + 84546.78527471398 *T)
    sumR2  +=       0.001 * cos(2.68991734615  + 59414.48187474840 *T)
    sumR2  +=       0.001 * cos(2.23301107736  + 107794.18751126219 *T)
    sumR2  +=       0.001 * cos(6.04185199127  + 104344.49901929598 *T)
    sumR2  +=       0.001 * cos(2.33178230942  + 108903.56606335558 *T)
    sumR2  +=       0.001 * cos(5.23291332097  + 134991.46920492980 *T)
    sumR2  +=       0.001 * cos(5.42420842294  + 79852.78232000639 *T)
    sumR2  +=       0.001 * cos(5.56485897392  + 104358.72611329758 *T)
    sumR2  +=       0.001 * cos(2.03042058117  + 76144.94556434419 *T)
    sumR2  +=       0.001 * cos(4.42403586918  + 25021.40766438420 *T)
    sumR2  +=       0.001 * cos(4.07362401058  + 181555.94006083015 *T)
    sumR2  +=       0.001 * cos(4.39544806298  + 131498.89763806018 *T)
    sumR2  +=       0.001 * cos(3.37653548744  + 22759.76748529401 *T)
    sumR2  +=       0.001 * cos(3.34598726101  + 24864.08530079559 *T)
    sumR2  +=       0.001 * cos(0.81949463309  + 52815.70356946240 *T)
    sumR2  +=       0.001 * cos(4.13026377404  + 2648.45482547300 *T)
    sumR2  +=       0.001 * cos(1.40701146170  + 103285.11708910680 *T)
    sumR2  +=       0.001 * cos(2.72338212355  + 419.48464387520 *T)
    sumR2  +=       0.001 * cos(2.86239729930  + 63498.47038145279 *T)
    sumR2  +=       0.001 * cos(4.60118276053  + 19804.82729158280 *T)
    sumR2  +=       0.001 * cos(3.76488909896  + 51749.20809227239 *T)
    sumR2  +=       0.001 * cos(5.71167880995  + 78378.14871340781 *T)
    sumR2  +=       0.001 * cos(3.11068504003  + 54294.57014352679 *T)
    sumR2  +=       0.001 * cos(5.17191010522  + 1596.18644228460 *T)
    sumR2  +=       0.001 * cos(0.84666659977  + 26037.90651966999 *T)
    sumR2  +=       0.001 * cos(1.27214241387  + 39743.76363275060 *T)
    sumR2  +=       0.001 * cos(3.62572271343  + 51756.32163927320 *T)
    sumR2  +=       0.001 * cos(0.99401491673  + 50579.61984086379 *T)
    sumR2  +=       0.001 * cos(2.65940437112  + 79219.30916633119 *T)
    sumR2  +=       0.001 * cos(2.78122688424  + 157636.79740153858 *T)
    sumR2  +=       0.001 * cos(5.02865728622  + 365230.64398203877 *T)
    sumR2  +=       0.001 * cos(4.26629771518  + 26724.89941359840 *T)
    sumR2  +=       0.001 * cos(0.69459147389  + 90829.86112470538 *T)
    sumR2  +=       0.001 * cos(3.74834736873  + 44295.71712980940 *T)
    sumR2  +=       0.001 * cos(2.67181600104  + 5327.47610838280 *T)
    sumR2  +=       0.001 * cos(5.10769130141  + 58946.51688439399 *T)
    sumR2  +=       0.001 * cos(0.55503171074  + 69159.80243060499 *T)
    sumR2  +=       0.001 * cos(4.68391610322  + 90695.75207512038 *T)
    sumR2  +=       0.001 * cos(5.17372763616  + 149.56319713460 *T)
    sumR2  +=       0.001 * cos(5.26044215633  + 25934.12433108940 *T)
    sumR2  +=       0.001 * cos(5.13679156205  + 104347.73123093881 *T)
    sumR2  +=       0.001 * cos(0.91197645609  + 103299.34418310841 *T)
    sumR2  +=       0.001 * cos(5.89466975855  + 27311.72098235281 *T)
    sumR2  +=       0.001 * cos(0.91206378451  + 77829.99768684579 *T)
    sumR2  +=       0.001 * cos(0.82686052448  + 146314.13330323418 *T)
    sumR2  +=       0.001 * cos(5.01041490382  + 25450.90686955000 *T)
    sumR2  +=       0.001 * cos(0.74891931364  + 10021.83728009940 *T)
    sumR2  +=       0.001 * cos(3.75910629606  + 25619.93815121980 *T)
    sumR2  +=       0.001 * cos(1.57855828108  + 61279.71327726600 *T)
    sumR2  +=       0.001 * cos(5.22196769252  + 77623.81213840858 *T)
    sumR2  +=       0.001 * cos(5.44280337360  + 52643.77127350280 *T)
    sumR2  +=       0.001 * cos(3.38760463430  + 49842.60989027639 *T)
    sumR2  +=       0.001 * cos(4.04176416141  + 25565.32572348040 *T)
    sumR2  +=       0.001 * cos(1.99169557218  + 105940.68546158058 *T)
    sumR2  +=       0.001 * cos(2.40126368211  + 26555.86813192860 *T)
    sumR2  +=       0.001 * cos(3.28153684541  + 78149.27013603736 *T)
    sumR2  +=       0.001 * cos(1.50606534242  + 50593.84693486539 *T)
    sumR2  +=       0.001 * cos(4.99390846744  + 143961.26714946239 *T)
    sumR2  +=       0.001 * cos(4.36279289081  + 70269.18098269838 *T)
    sumR2  +=       0.001 * cos(0.56330446870  + 29428.51556827400 *T)
    sumR2  +=       0.001 * cos(1.96827343473  + 188276.65404017159 *T)
    sumR2  +=       0.001 * cos(6.14024093350  + 3328.13565628019 *T)
    sumR2  +=       0.001 * cos(4.38278455115  + 52329.58509363319 *T)
    sumR2  +=       0.001 * cos(5.32677211586  + 26137.89976347840 *T)
    sumR2  +=       0.001 * cos(5.49001686286  + 7880.08915333899 *T)
    sumR2  +=       0.001 * cos(2.08096369099  + 26610.48055966799 *T)
    sumR2  +=       0.001 * cos(2.81808749734  + 54509.00267602040 *T)
    sumR2  +=       0.001 * cos(3.85620334673  + 26624.70765366959 *T)
    sumR2  +=       0.001 * cos(1.15157832957  + 536.80451209540 *T)
    sumR2  +=       0.001 * cos(3.62974919098  + 130012.91751699500 *T)
    sumR2  +=       0.001 * cos(0.35977888560  + 105307.21230790540 *T)
    sumR2  +=       0.001 * cos(2.75970257539  + 55503.94193942859 *T)
    sumR2  +=       0.001 * cos(3.95087871071  + 119116.85160956658 *T)
    sumR2  +=       0.001 * cos(2.02746284465  + 85034.42002596818 *T)
    sumR2  +=       0.001 * cos(0.01631363612  + 25668.41849769900 *T)
    sumR2  +=       0.001 * cos(4.69497491223  + 1581.95934828300 *T)
    sumR2  +=       0.001 * cos(4.43589881789  + 53242.30176033840 *T)
    sumR2  +=       0.001 * cos(0.02851648548  + 213.29909543800 *T)
    sumR2  +=       0.001 * cos(2.42155941603  + 25551.09862947879 *T)
    sumR2  +=       0.001 * cos(4.51335885923  + 12546.48193908300 *T)
    sumR2  +=       0.001 * cos(3.07012190290  + 104138.31347085880 *T)
    sumR2  +=       0.001 * cos(4.25105785914  + 65831.66677432480 *T)
    sumR2  +=       0.001 * cos(0.26591217377  + 89586.37352302698 *T)
    sumR2  +=       0.001 * cos(3.25491459524  + 46848.33017476560 *T)
    sumR2  +=       0.001 * cos(5.00501348827  + 51639.00177105299 *T)
    sumR2  +=       0.001 * cos(3.55161443787  + 116917.76426627958 *T)
    sumR2  +=       0.001 * cos(0.12176395772  + 6681.22485339960 *T)
    sumR2  +=       0.001 * cos(0.06932929209  + 48847.67062686820 *T)
    sumR2  +=       0.001 * cos(0.08695556653  + 26507.38778544939 *T)
    sumR2  +=       0.001 * cos(4.47600641528  + 97112.93697469680 *T)
    sumR2  +=       0.001 * cos(0.40812953108  + 77844.22478084739 *T)

    # MERCURY Radius
    sumR3  = 0
    sumR3  +=      32.676 * cos(1.67971635359  + 26087.90314157420 *T)
    sumR3  +=      24.166 * cos(4.63403168997  + 52175.80628314840 *T)
    sumR3  +=      12.133 * cos(1.38983781545  + 78263.70942472259 *T)
    sumR3  +=       5.140 * cos(4.43915386930  + 104351.61256629678 *T)
    sumR3  +=       1.981 * cos(1.20733880274  + 130439.51570787100 *T)
    sumR3  +=       1.460 * cos(3.14159265359  + 0.00000000000 *T)
    sumR3  +=       0.719 * cos(4.25913631362  + 156527.41884944518 *T)
    sumR3  +=       0.250 * cos(1.02794425848  + 182615.32199101936 *T)
    sumR3  +=       0.084 * cos(4.08008618813  + 208703.22513259359 *T)
    sumR3  +=       0.028 * cos(0.85051616237  + 234791.12827416777 *T)
    sumR3  +=       0.023 * cos(1.60029451014  + 24978.52458948080 *T)
    sumR3  +=       0.011 * cos(4.58074493182  + 27197.28169366760 *T)
    sumR3  +=       0.012 * cos(4.65878660390  + 51066.42773105500 *T)
    sumR3  +=       0.009 * cos(3.90622332667  + 260879.03141574195 *T)
    sumR3  +=       0.005 * cos(1.43247327249  + 77154.33087262919 *T)
    sumR3  +=       0.004 * cos(1.37106678794  + 53285.18483524180 *T)
    sumR3  +=       0.003 * cos(0.67217979275  + 286966.93455731618 *T)
    sumR3  +=       0.002 * cos(5.31135121818  + 20426.57109242200 *T)
    sumR3  +=       0.002 * cos(1.54812921017  + 1109.37855209340 *T)
    sumR3  +=       0.002 * cos(4.48222146022  + 103242.23401420339 *T)
    sumR3  +=       0.002 * cos(0.20057520755  + 1059.38193018920 *T)
    sumR3  +=       0.002 * cos(4.21865197056  + 25132.30339996560 *T)
    sumR3  +=       0.001 * cos(2.11997334018  + 46514.47423399620 *T)
    sumR3  +=       0.001 * cos(3.24267833975  + 27147.28507176339 *T)
    sumR3  +=       0.001 * cos(1.33238537892  + 1589.07289528380 *T)
    sumR3  +=       0.001 * cos(5.14405498793  + 72602.37737557039 *T)
    sumR3  +=       0.001 * cos(6.14595652541  + 26617.59410666880 *T)
    sumR3  +=       0.001 * cos(3.90494988936  + 4551.95349705880 *T)
    sumR3  +=       0.001 * cos(2.85134614467  + 14765.23904326980 *T)
    sumR3  +=       0.001 * cos(3.74339959279  + 313054.83769889042 *T)
    sumR3  +=       0.001 * cos(2.82234410581  + 52705.49724824299 *T)
    sumR3  +=       0.001 * cos(1.70132625273  + 23869.14603738740 *T)
    sumR3  +=       0.001 * cos(1.44102444460  + 27043.50288318280 *T)
    sumR3  +=       0.001 * cos(1.25358423340  + 129330.13715577760 *T)
    sumR3  +=       0.001 * cos(0.32054002273  + 9103.90699411760 *T)
    sumR3  +=       0.001 * cos(0.90273607509  + 51220.20654153979 *T)
    sumR3  +=       0.001 * cos(4.28935605425  + 27676.97603685800 *T)
    sumR3  +=       0.001 * cos(4.41206223998  + 79373.08797681599 *T)
    sumR3  +=       0.001 * cos(3.23518929162  + 26301.20223701220 *T)
    sumR3  +=       0.001 * cos(2.25667255668  + 21535.94964451540 *T)
    sumR3  +=       0.001 * cos(4.70414203873  + 50586.73338786459 *T)
    sumR3  +=       0.001 * cos(1.88608259832  + 98690.28051714458 *T)
    sumR3  +=       0.001 * cos(5.90420893056  + 40853.14218484400 *T)
    sumR3  +=       0.001 * cos(4.46892853356  + 53131.40602475700 *T)
    sumR3  +=       0.001 * cos(3.83684899837  + 77308.10968311399 *T)
    sumR3  +=       0.001 * cos(1.31974603498  + 54394.56338733519 *T)
    sumR3  +=       0.001 * cos(0.74728939443  + 32858.61374281979 *T)
    sumR3  +=       0.001 * cos(6.27673970277  + 53235.18821333759 *T)
    sumR3  +=       0.001 * cos(4.14814578610  + 29530.47808653960 *T)
    sumR3  +=       0.001 * cos(0.87302135775  + 83925.04147387479 *T)
    sumR3  +=       0.001 * cos(5.20339566427  + 26068.23338067440 *T)
    sumR3  +=       0.001 * cos(2.88631079597  + 79323.09135491178 *T)
    sumR3  +=       0.001 * cos(5.39853672849  + 19317.19254032860 *T)
    sumR3  +=       0.001 * cos(5.73721596586  + 78793.40038981718 *T)
    sumR3  +=       0.001 * cos(3.87802524387  + 82815.66292178139 *T)
    sumR3  +=       0.001 * cos(3.35625397389  + 37410.56723987860 *T)
    sumR3  +=       0.001 * cos(4.13825019881  + 3442.57494496540 *T)
    sumR3  +=       0.001 * cos(4.46169102731  + 28306.66024576099 *T)
    sumR3  +=       0.001 * cos(2.16986592809  + 45405.09568190280 *T)
    sumR3  +=       0.001 * cos(3.54716896556  + 51646.11531805379 *T)

    # MERCURY Radius
    sumR4  = 0
    sumR4  +=       0.394 * cos(0.36735403840  + 26087.90314157420 *T)
    sumR4  +=       0.387 * cos(3.18568771507  + 52175.80628314840 *T)
    sumR4  +=       0.270 * cos(6.16983616444  + 78263.70942472259 *T)
    sumR4  +=       0.149 * cos(2.91591904641  + 104351.61256629678 *T)
    sumR4  +=       0.071 * cos(5.95867889641  + 130439.51570787100 *T)
    sumR4  +=       0.031 * cos(2.72386700044  + 156527.41884944518 *T)
    sumR4  +=       0.017 * cos(0.00000000000  + 0.00000000000 *T)
    sumR4  +=       0.012 * cos(5.77775334056  + 182615.32199101936 *T)
    sumR4  +=       0.005 * cos(2.54506442647  + 208703.22513259359 *T)
    sumR4  +=       0.002 * cos(5.59978443851  + 234791.12827416777 *T)
    sumR4  +=       0.001 * cos(2.33785160871  + 260879.03141574195 *T)

    # MERCURY Radius
    sumR5  = 0
    sumR5  +=       0.006 * cos(3.98900269603  + 26087.90314157420 *T)
    sumR5  +=       0.006 * cos(1.55248278782  + 52175.80628314840 *T)
    sumR5  +=       0.005 * cos(4.65461721116  + 78263.70942472259 *T)
    sumR5  +=       0.004 * cos(1.40238366492  + 104351.61256629678 *T)
    sumR5  +=       0.002 * cos(4.44231488663  + 130439.51570787100 *T)
    sumR5  +=       0.001 * cos(3.14159265359  + 0.00000000000 *T)
    sumR5  +=       0.001 * cos(1.21278177664  + 156527.41884944518 *T)
    sumR5  +=       0.001 * cos(4.26149026340  + 182615.32199101936 *T)

    # calculate and scale the sums
    L = sumL0 + sumL1*T**1 + sumL2*T**2 + sumL3*T**3 + sumL4*T**4 + sumL5*T**5
    L /= 1e8
    L = Mod(degrees(L), 360.)
    B = sumB0 + sumB1*T**1 + sumB2*T**2 + sumB3*T**3 + sumB4*T**4 + sumB5*T**5
    B /= 1e8
    B = degrees(B)
    R = sumR0 + sumR1*T**1 + sumR2*T**2 + sumR3*T**3 + sumR4*T**4 + sumR5*T**5
    R /= 1e8

    # calculate X, Y, Z
    X = R * Cos(B) * Cos(L)
    Y = R * Cos(B) * Sin(L)
    Z = R * Sin(B)

    # create dictionary
    Dict = {  "Object"  : "Mercury",
              "Mode"    : "LBR",
              "Long"    : L,
              "Lat"     : B,
              "Rau"     : R,
              "X"       : X,
              "Y"       : Y,
              "Z"       : Z,
              "Alt"     : -34./60.
           }
    Data = (L, B, R, Dict)


    return Data
#######################################################################



def getMercuryData(dtCivicTime, zone, verbose = False):
    L, B, R, Dict = getMercuryLBRData(dtCivicTime, zone, verbose)
    X = Dict["X"]; Y = Dict["Y"]; Z = Dict["Z"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Le, Be, Re, Dicte = getEarthLBRData(dtCivicTime, zone, verbose)
        Xe = Dicte["X"]; Ye = Dicte["Y"]; Ze = Dicte["Z"]
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getNeptuneLBRData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Neptune"
#       Long  :  heliocentric Longitude, degrees
#       Lat   :  heliocentric Latitude,degrees
#       Rau   :  Radius, distaance Sun - Object (au)
#
#   Similarly for each planet
#   data from Meeus, Appendix II and VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.2 hvf 2023-04-06 getNeptuneLBRData.py created by mkPlanetLBRData.py"
VersionList.append(Version)

def getNeptuneLBRData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. (31.1) Julian Millennia

#######################################################################

    # NEPTUNE Longitude
    sumL0  = 0
    sumL0  +=  531188633.047 * cos(0.00000000000  + 0.00000000000 *T)
    sumL0  +=  1798475.509 * cos(2.90101273050  + 38.13303563780 *T)
    sumL0  +=  1019727.662 * cos(0.48580923660  + 1.48447270830 *T)
    sumL0  +=  124531.845 * cos(4.83008090682  + 36.64856292950 *T)
    sumL0  +=   42064.450 * cos(5.41054991607  + 2.96894541660 *T)
    sumL0  +=   37714.589 * cos(6.09221834946  + 35.16409022120 *T)
    sumL0  +=   33784.734 * cos(1.24488865578  + 76.26607127560 *T)
    sumL0  +=   16482.741 * cos(0.00007729261  + 491.55792945680 *T)
    sumL0  +=    9198.582 * cos(4.93747059924  + 39.61750834610 *T)
    sumL0  +=    8994.249 * cos(0.27462142569  + 175.16605980020 *T)
    sumL0  +=    4216.235 * cos(1.98711914364  + 73.29712585900 *T)
    sumL0  +=    3364.818 * cos(1.03590121818  + 33.67961751290 *T)
    sumL0  +=    2284.800 * cos(4.20606932559  + 4.45341812490 *T)
    sumL0  +=    1433.512 * cos(2.78340432711  + 74.78159856730 *T)
    sumL0  +=     900.240 * cos(2.07606702418  + 109.94568878850 *T)
    sumL0  +=     744.996 * cos(3.19032530145  + 71.81265315070 *T)
    sumL0  +=     506.206 * cos(5.74785370252  + 114.39910691340 *T)
    sumL0  +=     399.552 * cos(0.34972342569  + 1021.24889455140 *T)
    sumL0  +=     345.195 * cos(3.46186210169  + 41.10198105440 *T)
    sumL0  +=     306.338 * cos(0.49684039897  + 0.52126486180 *T)
    sumL0  +=     287.322 * cos(4.50523446022  + 0.04818410980 *T)
    sumL0  +=     323.004 * cos(2.24815188609  + 32.19514480460 *T)
    sumL0  +=     340.323 * cos(3.30369900416  + 77.75054398390 *T)
    sumL0  +=     266.605 * cos(4.88932609483  + 0.96320784650 *T)
    sumL0  +=     227.079 * cos(1.79713054538  + 453.42489381900 *T)
    sumL0  +=     244.722 * cos(1.24693337933  + 9.56122755560 *T)
    sumL0  +=     232.887 * cos(2.50459795017  + 137.03302416240 *T)
    sumL0  +=     282.170 * cos(2.24565579693  + 146.59425171800 *T)
    sumL0  +=     251.941 * cos(5.78166597292  + 388.46515523820 *T)
    sumL0  +=     150.180 * cos(2.99706110414  + 5.93789083320 *T)
    sumL0  +=     170.404 * cos(3.32390630650  + 108.46121608020 *T)
    sumL0  +=     151.401 * cos(2.19153094280  + 33.94024994380 *T)
    sumL0  +=     148.295 * cos(0.85948986145  + 111.43016149680 *T)
    sumL0  +=     118.672 * cos(3.67706204305  + 2.44768055480 *T)
    sumL0  +=     101.821 * cos(5.70539236951  + 0.11187458460 *T)
    sumL0  +=      97.873 * cos(2.80518260528  + 8.07675484730 *T)
    sumL0  +=     103.054 * cos(4.40441222000  + 70.32818044240 *T)
    sumL0  +=     103.305 * cos(0.04078966679  + 0.26063243090 *T)
    sumL0  +=     109.300 * cos(2.41599378049  + 183.24281464750 *T)
    sumL0  +=      73.938 * cos(1.32805041516  + 529.69096509460 *T)
    sumL0  +=      77.725 * cos(4.16446516424  + 4.19278569400 *T)
    sumL0  +=      86.379 * cos(4.22834506045  + 490.07345674850 *T)
    sumL0  +=      81.536 * cos(5.19908046216  + 493.04240216510 *T)
    sumL0  +=      71.503 * cos(5.29530386579  + 350.33211960040 *T)
    sumL0  +=      64.418 * cos(3.54541016050  + 168.05251279940 *T)
    sumL0  +=      62.570 * cos(0.15028731465  + 182.27960680100 *T)
    sumL0  +=      58.488 * cos(3.50106873945  + 145.10977900970 *T)
    sumL0  +=      48.276 * cos(1.11259925628  + 112.91463420510 *T)
    sumL0  +=      47.229 * cos(4.57373229818  + 46.20979048510 *T)
    sumL0  +=      39.124 * cos(1.66569356050  + 213.29909543800 *T)
    sumL0  +=      47.728 * cos(0.12906212461  + 484.44438245600 *T)
    sumL0  +=      46.858 * cos(3.01699530327  + 498.67147645760 *T)
    sumL0  +=      38.659 * cos(2.38685706479  + 2.92076130680 *T)
    sumL0  +=      47.046 * cos(4.49844660400  + 173.68158709190 *T)
    sumL0  +=      47.565 * cos(2.58404814824  + 219.89137757700 *T)
    sumL0  +=      44.714 * cos(5.47302733614  + 176.65053250850 *T)
    sumL0  +=      32.279 * cos(3.45759151220  + 30.71067209630 *T)
    sumL0  +=      28.249 * cos(4.13282446716  + 6.59228213900 *T)
    sumL0  +=      24.433 * cos(4.55736848232  + 106.97674337190 *T)
    sumL0  +=      24.661 * cos(3.67822620786  + 181.75834193920 *T)
    sumL0  +=      24.505 * cos(1.55095867965  + 7.11354700080 *T)
    sumL0  +=      21.848 * cos(1.04366818343  + 39.09624348430 *T)
    sumL0  +=      16.936 * cos(6.10896452834  + 44.72531777680 *T)
    sumL0  +=      22.169 * cos(2.74932970271  + 256.53994050650 *T)
    sumL0  +=      16.614 * cos(4.98188930613  + 37.61177077600 *T)
    sumL0  +=      17.728 * cos(3.55049134167  + 1.37259812370 *T)
    sumL0  +=      17.347 * cos(2.14069234880  + 42.58645376270 *T)
    sumL0  +=      14.953 * cos(3.36405649131  + 98.89998852460 *T)
    sumL0  +=      14.566 * cos(0.69857991985  + 1550.93985964600 *T)
    sumL0  +=      15.676 * cos(6.22010212025  + 454.90936652730 *T)
    sumL0  +=      13.243 * cos(5.61712542227  + 68.84370773410 *T)
    sumL0  +=      14.837 * cos(3.52557245517  + 25.60286266560 *T)
    sumL0  +=      12.757 * cos(0.04509743861  + 11.04570026390 *T)
    sumL0  +=      11.988 * cos(4.81687553351  + 24.11838995730 *T)
    sumL0  +=      11.060 * cos(1.78958277553  + 7.42236354150 *T)
    sumL0  +=      12.108 * cos(1.87022663714  + 79.23501669220 *T)
    sumL0  +=      11.698 * cos(0.49005698002  + 1.59634729290 *T)
    sumL0  +=      10.459 * cos(2.38743199893  + 381.35160823740 *T)
    sumL0  +=      11.681 * cos(3.85151357766  + 218.40690486870 *T)
    sumL0  +=       8.744 * cos(0.14168568610  + 148.07872442630 *T)
    sumL0  +=       9.196 * cos(1.00274090619  + 72.07328558160 *T)
    sumL0  +=      11.343 * cos(0.81432278263  + 525.49817940060 *T)
    sumL0  +=      10.097 * cos(5.03383557061  + 601.76425067620 *T)
    sumL0  +=       8.035 * cos(1.77685723010  + 0.21244832110 *T)
    sumL0  +=       8.382 * cos(3.07534786987  + 1.27202438720 *T)
    sumL0  +=      10.803 * cos(2.92081211459  + 293.18850343600 *T)
    sumL0  +=       7.666 * cos(1.52223325105  + 115.88357962170 *T)
    sumL0  +=       7.531 * cos(5.37537256533  + 5.10780943070 *T)
    sumL0  +=       8.691 * cos(4.74352784364  + 143.62530630140 *T)
    sumL0  +=      10.183 * cos(1.15395455831  + 6244.94281435360 *T)
    sumL0  +=       8.283 * cos(0.35956716764  + 138.51749687070 *T)
    sumL0  +=       9.544 * cos(4.02452832984  + 152.53214255120 *T)
    sumL0  +=       7.274 * cos(4.10937535938  + 251.43213107580 *T)
    sumL0  +=       7.465 * cos(1.72131945843  + 31.01948863700 *T)
    sumL0  +=       6.902 * cos(4.62452068308  + 2.70831298570 *T)
    sumL0  +=       7.094 * cos(5.11528393609  + 312.19908396260 *T)
    sumL0  +=       7.929 * cos(2.10765101655  + 27.08733537390 *T)
    sumL0  +=       6.156 * cos(3.50746507109  + 28.57180808220 *T)
    sumL0  +=       7.134 * cos(2.05292376023  + 278.25883401880 *T)
    sumL0  +=       8.193 * cos(2.58588219154  + 141.22580985640 *T)
    sumL0  +=       5.499 * cos(2.09250039025  + 1.69692102940 *T)
    sumL0  +=       5.279 * cos(4.09390686798  + 983.11585891360 *T)
    sumL0  +=       6.947 * cos(3.48041784595  + 415.29185818120 *T)
    sumL0  +=       5.916 * cos(0.68957324226  + 62.25142559510 *T)
    sumL0  +=       5.925 * cos(4.02504592620  + 255.05546779820 *T)
    sumL0  +=       4.606 * cos(1.17779101436  + 43.24084506850 *T)
    sumL0  +=       5.357 * cos(3.63061058987  + 5.41662597140 *T)
    sumL0  +=       5.918 * cos(2.57693824084  + 10175.15251057320 *T)
    sumL0  +=       5.482 * cos(3.07979737280  + 329.83706636550 *T)
    sumL0  +=       3.956 * cos(5.00418696742  + 184.72728735580 *T)
    sumL0  +=       5.408 * cos(3.31313295602  + 528.20649238630 *T)
    sumL0  +=       4.767 * cos(4.91981150665  + 456.39383923560 *T)
    sumL0  +=       3.770 * cos(1.57277409442  + 32.71640966640 *T)
    sumL0  +=       3.924 * cos(4.92763242635  + 180.27386923090 *T)
    sumL0  +=       3.707 * cos(4.82965453201  + 221.37585028530 *T)
    sumL0  +=       3.802 * cos(4.96279204998  + 594.65070367540 *T)
    sumL0  +=       4.014 * cos(1.63905164030  + 40.58071619260 *T)
    sumL0  +=       3.061 * cos(0.39713858313  + 1.43628859850 *T)
    sumL0  +=       3.261 * cos(4.65478978469  + 29.22619938800 *T)
    sumL0  +=       3.474 * cos(5.65891305944  + 395.57870223900 *T)
    sumL0  +=       2.918 * cos(5.91079083895  + 1.22384027740 *T)
    sumL0  +=       3.649 * cos(3.88114678609  + 494.52687487340 *T)
    sumL0  +=       3.225 * cos(5.57423738665  + 1014.13534755060 *T)
    sumL0  +=       2.845 * cos(0.56009386585  + 144.14657116320 *T)
    sumL0  +=       2.848 * cos(0.55423029727  + 567.82400073240 *T)
    sumL0  +=       3.440 * cos(1.70887250883  + 12.53017297220 *T)
    sumL0  +=       3.267 * cos(5.63287799820  + 488.58898404020 *T)
    sumL0  +=       3.107 * cos(5.79335949207  + 105.49227066360 *T)
    sumL0  +=       2.712 * cos(2.43726364359  + 60.76695288680 *T)
    sumL0  +=       3.202 * cos(2.21483496593  + 41.05379694460 *T)
    sumL0  +=       3.134 * cos(4.69665220513  + 82.85835341460 *T)
    sumL0  +=       3.590 * cos(5.69939670162  + 1124.34166877000 *T)
    sumL0  +=       2.967 * cos(0.54448940101  + 135.54855145410 *T)
    sumL0  +=       3.211 * cos(4.19927605853  + 291.70403072770 *T)
    sumL0  +=       2.899 * cos(5.99669788291  + 22.63391724900 *T)
    sumL0  +=       3.143 * cos(2.93495725805  + 31.23193695810 *T)
    sumL0  +=       2.729 * cos(4.62707721219  + 5.62907429250 *T)
    sumL0  +=       2.513 * cos(5.60391563025  + 19.12245511120 *T)
    sumL0  +=       2.690 * cos(5.32070128202  + 2.00573757010 *T)
    sumL0  +=       2.630 * cos(6.00855841124  + 37.16982779130 *T)
    sumL0  +=       2.296 * cos(6.06934502789  + 451.94042111070 *T)
    sumL0  +=       2.858 * cos(4.88677262419  + 258.02441321480 *T)
    sumL0  +=       2.879 * cos(5.12239168488  + 38.65430049960 *T)
    sumL0  +=       2.270 * cos(2.08634524182  + 30.05628079050 *T)
    sumL0  +=       2.301 * cos(3.35951602914  + 1028.36244155220 *T)
    sumL0  +=       3.001 * cos(3.59143817947  + 211.81462272970 *T)
    sumL0  +=       2.237 * cos(0.38455553470  + 3.62333672240 *T)
    sumL0  +=       2.901 * cos(3.24755614136  + 366.48562929500 *T)
    sumL0  +=       2.592 * cos(1.36262641469  + 35.42472265210 *T)
    sumL0  +=       2.418 * cos(4.93467056526  + 47.69426319340 *T)
    sumL0  +=       2.089 * cos(5.79838063413  + 4.66586644600 *T)
    sumL0  +=       2.586 * cos(2.69392971321  + 38.18121974760 *T)
    sumL0  +=       1.913 * cos(5.53560681085  + 149.56319713460 *T)
    sumL0  +=       1.971 * cos(6.00790964671  + 34.20088237470 *T)
    sumL0  +=       2.586 * cos(6.24984047544  + 38.08485152800 *T)
    sumL0  +=       2.098 * cos(4.57819744766  + 1019.76442184310 *T)
    sumL0  +=       1.869 * cos(3.85907708723  + 911.04257333200 *T)
    sumL0  +=       2.486 * cos(5.21235809332  + 140.00196957900 *T)
    sumL0  +=       1.795 * cos(1.68012868451  + 1059.38193018920 *T)
    sumL0  +=       2.326 * cos(2.82664069146  + 807.94979911340 *T)
    sumL0  +=       1.984 * cos(5.54763522932  + 1022.73336725970 *T)
    sumL0  +=       1.919 * cos(5.10717766499  + 216.92243216040 *T)
    sumL0  +=       2.004 * cos(5.47811228948  + 63.73589830340 *T)
    sumL0  +=       2.021 * cos(4.15631916516  + 178.13500521680 *T)
    sumL0  +=       1.760 * cos(6.00927149342  + 172.19711438360 *T)
    sumL0  +=       2.140 * cos(2.65037925793  + 700.66423920080 *T)
    sumL0  +=       1.988 * cos(3.35850272780  + 186.21176006410 *T)
    sumL0  +=       1.956 * cos(5.01527508588  + 294.67297614430 *T)
    sumL0  +=       1.966 * cos(4.07957525462  + 20.60692781950 *T)
    sumL0  +=       1.637 * cos(0.53823942149  + 67.35923502580 *T)
    sumL0  +=       1.540 * cos(2.62327849119  + 41.75637236020 *T)
    sumL0  +=       1.810 * cos(5.81430038477  + 129.91947716160 *T)
    sumL0  +=       1.776 * cos(4.37047808449  + 328.35259365720 *T)
    sumL0  +=       1.460 * cos(2.63664516309  + 2.85707083200 *T)
    sumL0  +=       1.388 * cos(2.10598045632  + 3.93215326310 *T)
    sumL0  +=       1.352 * cos(0.55618245459  + 0.65439130580 *T)
    sumL0  +=       1.668 * cos(2.77543377384  + 16.15350969460 *T)
    sumL0  +=       1.338 * cos(0.37643611305  + 14.01464568050 *T)
    sumL0  +=       1.218 * cos(0.73456434750  + 426.59819087600 *T)
    sumL0  +=       1.531 * cos(4.54891769768  + 526.72201967800 *T)
    sumL0  +=       1.610 * cos(3.40993944436  + 403.13419222450 *T)
    sumL0  +=       1.361 * cos(4.48227243414  + 17.63798240290 *T)
    sumL0  +=       1.589 * cos(5.59323020112  + 3302.47939106200 *T)
    sumL0  +=       1.132 * cos(5.64520725360  + 151.04766984290 *T)
    sumL0  +=       1.357 * cos(4.06399031430  + 26.82670294300 *T)
    sumL0  +=       1.494 * cos(4.98692049495  + 666.72398925700 *T)
    sumL0  +=       1.077 * cos(4.30911470250  + 0.63313944640 *T)
    sumL0  +=       1.042 * cos(6.02756893581  + 106.01353552540 *T)
    sumL0  +=       1.060 * cos(0.74679491358  + 487.36514376280 *T)
    sumL0  +=       1.310 * cos(3.78526380930  + 386.98068252990 *T)
    sumL0  +=       1.342 * cos(4.52685061062  + 563.63121503840 *T)
    sumL0  +=       0.986 * cos(0.00600924269  + 81.37388070630 *T)
    sumL0  +=       1.232 * cos(5.17443930901  + 331.32153907380 *T)
    sumL0  +=       0.929 * cos(4.51267465978  + 38.39366806870 *T)
    sumL0  +=       0.956 * cos(3.50447791020  + 64.95973858080 *T)
    sumL0  +=       0.929 * cos(4.43109514438  + 37.87240320690 *T)
    sumL0  +=       0.926 * cos(6.09803297747  + 4.14460158420 *T)
    sumL0  +=       0.972 * cos(0.59038366513  + 8.90683624980 *T)
    sumL0  +=       1.246 * cos(4.69840351226  + 389.94962794650 *T)
    sumL0  +=       1.009 * cos(5.98451242784  + 142.14083359310 *T)
    sumL0  +=       1.020 * cos(0.83233892300  + 39.35687591520 *T)
    sumL0  +=       1.013 * cos(0.37845630298  + 36.90919536040 *T)
    sumL0  +=       0.940 * cos(2.42688145966  + 343.21857259960 *T)
    sumL0  +=       0.974 * cos(5.23958752786  + 253.57099508990 *T)
    sumL0  +=       0.964 * cos(5.09748190218  + 357.44566660120 *T)
    sumL0  +=       0.835 * cos(1.45568626670  + 35.21227433100 *T)
    sumL0  +=       1.077 * cos(0.71409061316  + 44.07092647100 *T)
    sumL0  +=       1.083 * cos(2.27578897621  + 6.90109867970 *T)
    sumL0  +=       0.938 * cos(5.03471583911  + 69.36497259590 *T)
    sumL0  +=       1.078 * cos(1.20253141912  + 35.68535508300 *T)
    sumL0  +=       1.027 * cos(0.18243183397  + 84.34282612290 *T)
    sumL0  +=       0.764 * cos(4.62720907712  + 0.83008140250 *T)
    sumL0  +=       1.013 * cos(0.42234855022  + 32.45577723550 *T)
    sumL0  +=       0.939 * cos(4.50445799766  + 365.00115658670 *T)
    sumL0  +=       0.756 * cos(0.82872484717  + 17.52610781830 *T)
    sumL0  +=       0.916 * cos(3.89409205418  + 38.24491022240 *T)
    sumL0  +=       0.736 * cos(4.78125743795  + 5.36844186160 *T)
    sumL0  +=       0.762 * cos(0.01897337130  + 189.39315380180 *T)
    sumL0  +=       0.738 * cos(2.31770478416  + 42.32582133180 *T)
    sumL0  +=       0.860 * cos(4.82440483506  + 210.33015002140 *T)
    sumL0  +=       0.888 * cos(3.20360339895  + 348.84764689210 *T)
    sumL0  +=       0.916 * cos(5.04967792934  + 38.02116105320 *T)
    sumL0  +=       0.638 * cos(0.63267396269  + 244.31858407500 *T)
    sumL0  +=       0.636 * cos(1.02615137352  + 2080.63082474060 *T)
    sumL0  +=       0.774 * cos(5.44432678139  + 367.97010200330 *T)
    sumL0  +=       0.644 * cos(1.94044989547  + 446.31134681820 *T)
    sumL0  +=       0.631 * cos(4.82928491724  + 460.53844081980 *T)
    sumL0  +=       0.855 * cos(3.57592750113  + 439.78275515400 *T)
    sumL0  +=       0.678 * cos(4.48687912809  + 351.81659230870 *T)
    sumL0  +=       0.724 * cos(4.89141609280  + 119.50691634410 *T)
    sumL0  +=       0.594 * cos(0.59315717529  + 491.03666459500 *T)
    sumL0  +=       0.655 * cos(1.99014093000  + 19.01058052660 *T)
    sumL0  +=       0.580 * cos(2.57189536188  + 492.07919431860 *T)
    sumL0  +=       0.694 * cos(0.08328521209  + 5.67725840230 *T)
    sumL0  +=       0.733 * cos(5.81485239057  + 29.74746424980 *T)
    sumL0  +=       0.666 * cos(3.42196897591  + 179.09821306330 *T)
    sumL0  +=       0.678 * cos(0.29428615814  + 171.23390653710 *T)
    sumL0  +=       0.635 * cos(2.13805182663  + 164.12035953630 *T)
    sumL0  +=       0.623 * cos(5.61454940380  + 285.37238101960 *T)
    sumL0  +=       0.529 * cos(1.88063108785  + 416.77633088950 *T)
    sumL0  +=       0.529 * cos(5.13250788030  + 697.74347789400 *T)
    sumL0  +=       0.500 * cos(1.49548514415  + 704.85702489480 *T)
    sumL0  +=       0.487 * cos(4.97772067947  + 274.06604832480 *T)
    sumL0  +=       0.666 * cos(6.26456825266  + 1474.67378837040 *T)
    sumL0  +=       0.532 * cos(0.25784352716  + 477.33083545520 *T)
    sumL0  +=       0.557 * cos(0.71378452161  + 80.71948940050 *T)
    sumL0  +=       0.556 * cos(2.60791360513  + 418.26080359780 *T)
    sumL0  +=       0.584 * cos(4.29064541383  + 16.67477455640 *T)
    sumL0  +=       0.524 * cos(5.42759392280  + 290.21955801940 *T)
    sumL0  +=       0.524 * cos(0.29054995359  + 247.23934538180 *T)
    sumL0  +=       0.541 * cos(4.36400580938  + 815.06334611420 *T)
    sumL0  +=       0.526 * cos(1.66512720297  + 97.41551581630 *T)
    sumL0  +=       0.497 * cos(4.72640318293  + 401.64971951620 *T)
    sumL0  +=       0.432 * cos(2.98481475894  + 100.38446123290 *T)
    sumL0  +=       0.382 * cos(0.28067758468  + 8.38557138800 *T)
    sumL0  +=       0.424 * cos(6.16774845481  + 178.78939652260 *T)
    sumL0  +=       0.484 * cos(0.01535318279  + 738.79727483860 *T)
    sumL0  +=       0.518 * cos(4.48916591410  + 875.83029900100 *T)
    sumL0  +=       0.506 * cos(5.38611121207  + 404.61866493280 *T)
    sumL0  +=       0.396 * cos(4.62747640832  + 6.15033915430 *T)
    sumL0  +=       0.466 * cos(0.23340415764  + 120.99138905240 *T)
    sumL0  +=       0.409 * cos(3.08849480895  + 59.28248017850 *T)
    sumL0  +=       0.470 * cos(5.01853200224  + 313.68355667090 *T)
    sumL0  +=       0.442 * cos(3.68919475089  + 457.87831194390 *T)
    sumL0  +=       0.384 * cos(3.69499925394  + 160.93896579860 *T)
    sumL0  +=       0.364 * cos(0.76192181046  + 104.00779795530 *T)
    sumL0  +=       0.416 * cos(0.26652109651  + 103.09277421860 *T)
    sumL0  +=       0.401 * cos(4.06530055968  + 14.66903698630 *T)
    sumL0  +=       0.454 * cos(3.72767803715  + 476.43131808350 *T)
    sumL0  +=       0.434 * cos(0.33533802200  + 984.60033162190 *T)
    sumL0  +=       0.340 * cos(0.99915726716  + 31.54075349880 *T)
    sumL0  +=       0.420 * cos(3.65147769268  + 20.49505323490 *T)
    sumL0  +=       0.334 * cos(0.35121412008  + 1227.43444298860 *T)
    sumL0  +=       0.323 * cos(5.45836731979  + 918.15612033280 *T)
    sumL0  +=       0.407 * cos(4.19457842203  + 309.79958751760 *T)
    sumL0  +=       0.381 * cos(0.01364856960  + 495.49008271990 *T)
    sumL0  +=       0.334 * cos(4.05924071124  + 8.33738727820 *T)
    sumL0  +=       0.380 * cos(3.17063415023  + 487.62577619370 *T)
    sumL0  +=       0.309 * cos(0.48352303405  + 118.02244363580 *T)
    sumL0  +=       0.380 * cos(2.70238752925  + 134.11226285560 *T)
    sumL0  +=       0.362 * cos(4.88985810610  + 438.29828244570 *T)
    sumL0  +=       0.327 * cos(2.91090790412  + 505.78502345840 *T)
    sumL0  +=       0.308 * cos(0.96082817124  + 21.14944454070 *T)
    sumL0  +=       0.288 * cos(1.48123872077  + 220.41264243880 *T)
    sumL0  +=       0.293 * cos(2.56582281789  + 662.53120356300 *T)
    sumL0  +=       0.331 * cos(4.37715965811  + 180.79513409270 *T)
    sumL0  +=       0.326 * cos(2.46104924164  + 169.53698550770 *T)
    sumL0  +=       0.289 * cos(2.63591886391  + 55.77101804070 *T)
    sumL0  +=       0.288 * cos(5.02487283285  + 1440.73353842660 *T)
    sumL0  +=       0.344 * cos(1.48930997270  + 166.56804009110 *T)
    sumL0  +=       0.266 * cos(0.63672427386  + 79.18683258240 *T)
    sumL0  +=       0.268 * cos(5.02354540478  + 377.41945497430 *T)
    sumL0  +=       0.308 * cos(1.50185265748  + 77.22927912210 *T)
    sumL0  +=       0.324 * cos(5.30240189273  + 457.61767951300 *T)
    sumL0  +=       0.265 * cos(1.08736632800  + 450.45594840240 *T)
    sumL0  +=       0.264 * cos(0.83337660655  + 488.37653571910 *T)
    sumL0  +=       0.290 * cos(1.80003152563  + 101.86893394120 *T)
    sumL0  +=       0.262 * cos(2.30390003360  + 494.73932319450 *T)
    sumL0  +=       0.325 * cos(5.52669889053  + 441.26722786230 *T)
    sumL0  +=       0.254 * cos(0.02963623277  + 117.36805233000 *T)
    sumL0  +=       0.300 * cos(0.17435705540  + 252.91660378410 *T)
    sumL0  +=       0.315 * cos(5.34885013040  + 183.76407950930 *T)
    sumL0  +=       0.313 * cos(5.45945846595  + 13.49338081870 *T)
    sumL0  +=       0.306 * cos(5.23085809622  + 45.24658263860 *T)
    sumL0  +=       0.237 * cos(0.32676889138  + 208.84567731310 *T)
    sumL0  +=       0.263 * cos(2.66670785888  + 464.73122651380 *T)
    sumL0  +=       0.234 * cos(1.82700149824  + 52175.80628314840 *T)
    sumL0  +=       0.275 * cos(5.04385701142  + 156.15547927360 *T)
    sumL0  +=       0.265 * cos(5.64967127743  + 326.86812094890 *T)
    sumL0  +=       0.247 * cos(1.74540930625  + 65.87476231750 *T)
    sumL0  +=       0.269 * cos(6.09827783249  + 1654.03263386460 *T)
    sumL0  +=       0.229 * cos(2.25832077914  + 190.66517818900 *T)
    sumL0  +=       0.294 * cos(5.45249564193  + 206.18554843720 *T)
    sumL0  +=       0.238 * cos(1.55647021369  + 79.88940799800 *T)
    sumL0  +=       0.230 * cos(6.13158632762  + 178.34745353790 *T)
    sumL0  +=       0.274 * cos(4.10829870815  + 518.38463239980 *T)
    sumL0  +=       0.225 * cos(3.86300359251  + 171.98466606250 *T)
    sumL0  +=       0.228 * cos(2.48511565618  + 12566.15169998280 *T)
    sumL0  +=       0.272 * cos(5.61149862463  + 148.33935685720 *T)
    sumL0  +=       0.214 * cos(1.45987216039  + 522.57741809380 *T)
    sumL0  +=       0.211 * cos(4.04791980901  + 6205.32530600750 *T)
    sumL0  +=       0.266 * cos(0.99036038827  + 209.10630974400 *T)
    sumL0  +=       0.230 * cos(0.54049951530  + 532.61172640140 *T)
    sumL0  +=       0.226 * cos(3.84152961620  + 283.62727588040 *T)
    sumL0  +=       0.243 * cos(5.32730346969  + 485.92885516430 *T)
    sumL0  +=       0.209 * cos(4.35051470487  + 536.80451209540 *T)
    sumL0  +=       0.232 * cos(3.01948719112  + 10.93382567930 *T)
    sumL0  +=       0.264 * cos(5.70536379124  + 490.33408917940 *T)
    sumL0  +=       0.280 * cos(3.99993658196  + 674.80074410430 *T)
    sumL0  +=       0.246 * cos(0.37698964335  + 157.63995198190 *T)
    sumL0  +=       0.219 * cos(5.67679857772  + 52099.54021187280 *T)
    sumL0  +=       0.251 * cos(1.52353965506  + 6.85291456990 *T)
    sumL0  +=       0.203 * cos(5.44328656642  + 145.63104387150 *T)
    sumL0  +=       0.238 * cos(0.96169723853  + 497.18700374930 *T)
    sumL0  +=       0.219 * cos(4.52300776062  + 1615.89959822680 *T)
    sumL0  +=       0.275 * cos(2.37619210741  + 2118.76386037840 *T)
    sumL0  +=       0.258 * cos(5.12448148780  + 608.87779767700 *T)
    sumL0  +=       0.260 * cos(3.88543008475  + 513.07988101300 *T)
    sumL0  +=       0.191 * cos(3.72574595369  + 65.22037101170 *T)
    sumL0  +=       0.211 * cos(0.06484535455  + 215.43795945210 *T)
    sumL0  +=       0.236 * cos(3.95835282821  + 141.48644228730 *T)
    sumL0  +=       0.189 * cos(5.28135043909  + 377.15882254340 *T)
    sumL0  +=       0.243 * cos(4.35559878377  + 482.95990974770 *T)
    sumL0  +=       0.243 * cos(6.06808644973  + 154.01661525950 *T)
    sumL0  +=       0.249 * cos(1.57215637373  + 14.22709400160 *T)
    sumL0  +=       0.238 * cos(1.93340192445  + 500.15594916590 *T)
    sumL0  +=       0.209 * cos(5.02893682321  + 364.55921360200 *T)
    sumL0  +=       0.227 * cos(5.72984298540  + 1543.82631264520 *T)
    sumL0  +=       0.217 * cos(2.45036922991  + 187.17496791060 *T)
    sumL0  +=       0.181 * cos(1.65699502247  + 1627.20593092160 *T)
    sumL0  +=       0.214 * cos(1.60213179145  + 11.30633269480 *T)
    sumL0  +=       0.203 * cos(0.74638490279  + 14.55716240170 *T)
    sumL0  +=       0.192 * cos(3.17719161639  + 343.47920503050 *T)
    sumL0  +=       0.177 * cos(1.50027795761  + 9.44935297100 *T)
    sumL0  +=       0.177 * cos(0.03038098292  + 165.60483224460 *T)
    sumL0  +=       0.176 * cos(4.64462444674  + 315.16802937920 *T)
    sumL0  +=       0.208 * cos(2.65835778368  + 496.01134758170 *T)
    sumL0  +=       0.174 * cos(2.76155855705  + 49.17873590170 *T)
    sumL0  +=       0.196 * cos(1.95549714182  + 335.77495719870 *T)
    sumL0  +=       0.200 * cos(4.16839394758  + 285.11174858870 *T)
    sumL0  +=       0.199 * cos(0.06168021293  + 73.55775828990 *T)
    sumL0  +=       0.188 * cos(6.17288913873  + 535.32003938710 *T)
    sumL0  +=       0.215 * cos(1.92414563346  + 552.69738935910 *T)
    sumL0  +=       0.166 * cos(5.49038139690  + 10135.53500222710 *T)
    sumL0  +=       0.192 * cos(0.96973434120  + 304.23420369990 *T)
    sumL0  +=       0.209 * cos(5.34065233845  + 13.64213866500 *T)
    sumL0  +=       0.203 * cos(5.11234865419  + 324.72925693480 *T)
    sumL0  +=       0.177 * cos(3.50680841790  + 207.36120460480 *T)
    sumL0  +=       0.174 * cos(1.95010708561  + 319.31263096340 *T)
    sumL0  +=       0.187 * cos(5.57685931698  + 266.10116806210 *T)
    sumL0  +=       0.181 * cos(1.43525075751  + 279.74330672710 *T)
    sumL0  +=       0.165 * cos(4.00537112057  + 493.56366702690 *T)
    sumL0  +=       0.191 * cos(1.68313683465  + 563.37058260750 *T)
    sumL0  +=       0.173 * cos(3.93200456456  + 238.90195810360 *T)
    sumL0  +=       0.161 * cos(5.96143146317  + 36.12729806770 *T)
    sumL0  +=       0.194 * cos(2.37664231450  + 944.98282327580 *T)
    sumL0  +=       0.165 * cos(0.97421918976  + 556.51766803760 *T)
    sumL0  +=       0.189 * cos(1.11279570541  + 1127.26243007680 *T)
    sumL0  +=       0.172 * cos(0.75085513952  + 267.58564077040 *T)
    sumL0  +=       0.193 * cos(2.12636756833  + 20350.30502114640 *T)
    sumL0  +=       0.181 * cos(2.10814562080  + 113.87784205160 *T)
    sumL0  +=       0.194 * cos(1.13504964219  + 57.25549074900 *T)
    sumL0  +=       0.181 * cos(6.23699820519  + 355.96119389290 *T)
    sumL0  +=       0.198 * cos(5.68125942959  + 6280.10690457480 *T)
    sumL0  +=       0.173 * cos(5.15083799917  + 474.94684537520 *T)
    sumL0  +=       0.151 * cos(1.66981962338  + 116.53797092750 *T)
    sumL0  +=       0.150 * cos(5.42593657173  + 526.98265210890 *T)
    sumL0  +=       0.205 * cos(4.16096717573  + 711.44930703380 *T)
    sumL0  +=       0.177 * cos(3.49360697678  + 421.22974901440 *T)
    sumL0  +=       0.168 * cos(0.52839230204  + 487.10451133190 *T)
    sumL0  +=       0.160 * cos(4.77712663799  + 524.01370669230 *T)
    sumL0  +=       0.145 * cos(2.81448128781  + 1512.80682400820 *T)
    sumL0  +=       0.146 * cos(4.99570112660  + 142.66209845490 *T)
    sumL0  +=       0.188 * cos(0.82104161550  + 10210.31660079440 *T)
    sumL0  +=       0.145 * cos(4.96888131586  + 1189.30140735080 *T)
    sumL0  +=       0.181 * cos(2.99704790590  + 75.74480641380 *T)
    sumL0  +=       0.176 * cos(0.41626373842  + 222.86032299360 *T)
    sumL0  +=       0.137 * cos(2.96534226337  + 6206.80977871580 *T)
    sumL0  +=       0.138 * cos(1.22260849471  + 187.69623277240 *T)
    sumL0  +=       0.128 * cos(2.53394068407  + 276.77436131050 *T)
    sumL0  +=       0.130 * cos(3.04810765699  + 310.71461125430 *T)
    sumL0  +=       0.122 * cos(3.01323006886  + 70.84944530420 *T)
    sumL0  +=       0.111 * cos(0.77449448649  + 179.35884549420 *T)
    sumL0  +=       0.141 * cos(0.18423889807  + 131.40394986990 *T)
    sumL0  +=       0.126 * cos(5.77648809669  + 525.23754696970 *T)
    sumL0  +=       0.124 * cos(2.93225731024  + 179.61947792510 *T)
    sumL0  +=       0.111 * cos(6.18471578216  + 981.63138620530 *T)
    sumL0  +=       0.141 * cos(2.63342951123  + 381.61224066830 *T)
    sumL0  +=       0.110 * cos(5.25053027081  + 986.08480433020 *T)
    sumL0  +=       0.096 * cos(3.86591534559  + 240.12579838100 *T)
    sumL0  +=       0.120 * cos(3.78755085035  + 1057.89745748090 *T)
    sumL0  +=       0.093 * cos(4.54014016637  + 36.69674703930 *T)
    sumL0  +=       0.109 * cos(1.53327585900  + 419.74527630610 *T)
    sumL0  +=       0.094 * cos(4.21870300178  + 1024.21783996800 *T)
    sumL0  +=       0.109 * cos(2.15905156247  + 289.56516671360 *T)
    sumL0  +=       0.104 * cos(0.20665642552  + 564.85505531580 *T)
    sumL0  +=       0.081 * cos(1.89134135215  + 36.60037881970 *T)
    sumL0  +=       0.080 * cos(4.38832594589  + 10137.01947493540 *T)
    sumL0  +=       0.080 * cos(1.73940577376  + 39.50563376150 *T)
    sumL0  +=       0.084 * cos(0.81316746605  + 170.71264167530 *T)
    sumL0  +=       0.090 * cos(0.60145818457  + 36.76043751410 *T)
    sumL0  +=       0.074 * cos(4.92511651321  + 1549.45538693770 *T)
    sumL0  +=       0.072 * cos(5.06852406179  + 249.94765836750 *T)

    # NEPTUNE Longitude
    sumL1  = 0
    sumL1  +=  3837687716.731 * cos(0.00000000000  + 0.00000000000 *T)
    sumL1  +=   16604.187 * cos(4.86319129565  + 1.48447270830 *T)
    sumL1  +=   15807.148 * cos(2.27923488532  + 38.13303563780 *T)
    sumL1  +=    3334.701 * cos(3.68199676020  + 76.26607127560 *T)
    sumL1  +=    1305.840 * cos(3.67320813491  + 2.96894541660 *T)
    sumL1  +=     604.832 * cos(1.50477747549  + 35.16409022120 *T)
    sumL1  +=     178.623 * cos(3.45318524147  + 39.61750834610 *T)
    sumL1  +=     106.537 * cos(2.45126138334  + 4.45341812490 *T)
    sumL1  +=     105.747 * cos(2.75479326550  + 33.67961751290 *T)
    sumL1  +=      72.684 * cos(5.48724732699  + 36.64856292950 *T)
    sumL1  +=      57.069 * cos(5.21649804970  + 0.52126486180 *T)
    sumL1  +=      57.355 * cos(1.85767603384  + 114.39910691340 *T)
    sumL1  +=      35.368 * cos(4.51676827545  + 74.78159856730 *T)
    sumL1  +=      32.216 * cos(5.90411489680  + 77.75054398390 *T)
    sumL1  +=      29.871 * cos(3.67043294114  + 388.46515523820 *T)
    sumL1  +=      28.866 * cos(5.16877529164  + 9.56122755560 *T)
    sumL1  +=      28.742 * cos(5.16732589024  + 2.44768055480 *T)
    sumL1  +=      25.507 * cos(5.24526281928  + 168.05251279940 *T)
    sumL1  +=      24.869 * cos(4.73193067810  + 182.27960680100 *T)
    sumL1  +=      20.205 * cos(5.78945415677  + 1021.24889455140 *T)
    sumL1  +=      19.022 * cos(1.82981144269  + 484.44438245600 *T)
    sumL1  +=      18.661 * cos(1.31606255521  + 498.67147645760 *T)
    sumL1  +=      15.063 * cos(4.95003893760  + 137.03302416240 *T)
    sumL1  +=      15.094 * cos(3.98705254940  + 32.19514480460 *T)
    sumL1  +=      10.720 * cos(2.44148149225  + 4.19278569400 *T)
    sumL1  +=      11.725 * cos(4.89255650674  + 71.81265315070 *T)
    sumL1  +=       9.581 * cos(1.23188039594  + 5.93789083320 *T)
    sumL1  +=       9.606 * cos(1.88534821556  + 41.10198105440 *T)
    sumL1  +=       8.968 * cos(0.01758559103  + 8.07675484730 *T)
    sumL1  +=       9.882 * cos(6.08165628679  + 7.11354700080 *T)
    sumL1  +=       7.632 * cos(5.51307048241  + 73.29712585900 *T)
    sumL1  +=       6.992 * cos(0.61688864282  + 2.92076130680 *T)
    sumL1  +=       5.543 * cos(2.24141557794  + 46.20979048510 *T)
    sumL1  +=       4.845 * cos(3.71055823750  + 112.91463420510 *T)
    sumL1  +=       3.700 * cos(5.25713252333  + 111.43016149680 *T)
    sumL1  +=       3.233 * cos(6.10303038418  + 70.32818044240 *T)
    sumL1  +=       2.939 * cos(4.86520586648  + 98.89998852460 *T)
    sumL1  +=       2.403 * cos(2.90637675099  + 601.76425067620 *T)
    sumL1  +=       2.398 * cos(1.04343654629  + 6.59228213900 *T)
    sumL1  +=       2.784 * cos(4.95821114677  + 108.46121608020 *T)
    sumL1  +=       2.894 * cos(4.20148844767  + 381.35160823740 *T)
    sumL1  +=       2.111 * cos(5.93089610785  + 25.60286266560 *T)
    sumL1  +=       2.075 * cos(5.20632201951  + 30.71067209630 *T)
    sumL1  +=       2.126 * cos(0.54976393136  + 41.05379694460 *T)
    sumL1  +=       2.235 * cos(2.38045158073  + 453.42489381900 *T)
    sumL1  +=       1.859 * cos(0.89409373259  + 24.11838995730 *T)
    sumL1  +=       2.018 * cos(3.42245274178  + 31.01948863700 *T)
    sumL1  +=       1.700 * cos(3.91715254287  + 11.04570026390 *T)
    sumL1  +=       1.776 * cos(3.86571077241  + 395.57870223900 *T)
    sumL1  +=       1.644 * cos(0.15855999051  + 152.53214255120 *T)
    sumL1  +=       1.646 * cos(3.34591387314  + 44.72531777680 *T)
    sumL1  +=       1.876 * cos(2.59784179105  + 33.94024994380 *T)
    sumL1  +=       1.614 * cos(0.42137145545  + 175.16605980020 *T)
    sumL1  +=       1.468 * cos(6.12983933526  + 1550.93985964600 *T)
    sumL1  +=       1.408 * cos(6.13722948564  + 490.07345674850 *T)
    sumL1  +=       1.207 * cos(0.59525736062  + 312.19908396260 *T)
    sumL1  +=       1.336 * cos(3.28611928206  + 493.04240216510 *T)
    sumL1  +=       1.176 * cos(5.87266726996  + 5.41662597140 *T)
    sumL1  +=       1.517 * cos(3.12967210501  + 491.55792945680 *T)
    sumL1  +=       1.053 * cos(4.60375516830  + 79.23501669220 *T)
    sumL1  +=       1.037 * cos(4.89007314395  + 1.27202438720 *T)
    sumL1  +=       1.034 * cos(5.93741289103  + 32.71640966640 *T)
    sumL1  +=       1.038 * cos(1.13470380744  + 1014.13534755060 *T)
    sumL1  +=       1.002 * cos(1.85850922283  + 5.10780943070 *T)
    sumL1  +=       0.983 * cos(0.05345050384  + 7.42236354150 *T)
    sumL1  +=       0.998 * cos(1.73689827444  + 1028.36244155220 *T)
    sumL1  +=       1.193 * cos(4.63176675581  + 60.76695288680 *T)
    sumL1  +=       0.940 * cos(3.09103721222  + 62.25142559510 *T)
    sumL1  +=       0.994 * cos(4.11489180313  + 4.66586644600 *T)
    sumL1  +=       0.890 * cos(0.87049255398  + 31.23193695810 *T)
    sumL1  +=       0.852 * cos(5.35508394316  + 144.14657116320 *T)
    sumL1  +=       0.922 * cos(5.12373360511  + 145.10977900970 *T)
    sumL1  +=       0.789 * cos(0.37496785039  + 26.82670294300 *T)
    sumL1  +=       0.828 * cos(4.06035194600  + 115.88357962170 *T)
    sumL1  +=       0.711 * cos(3.14189997439  + 278.25883401880 *T)
    sumL1  +=       0.727 * cos(1.39718382835  + 213.29909543800 *T)
    sumL1  +=       0.781 * cos(0.10946327923  + 173.68158709190 *T)
    sumL1  +=       0.793 * cos(6.13086312116  + 567.82400073240 *T)
    sumL1  +=       0.669 * cos(4.50554989443  + 27.08733537390 *T)
    sumL1  +=       0.825 * cos(1.35568908148  + 129.91947716160 *T)
    sumL1  +=       0.738 * cos(3.56766018960  + 176.65053250850 *T)
    sumL1  +=       0.714 * cos(6.24797992301  + 106.97674337190 *T)
    sumL1  +=       0.654 * cos(1.13177751192  + 68.84370773410 *T)
    sumL1  +=       0.624 * cos(0.01567750666  + 28.57180808220 *T)
    sumL1  +=       0.608 * cos(4.60180625368  + 189.39315380180 *T)
    sumL1  +=       0.595 * cos(0.00857468445  + 42.58645376270 *T)
    sumL1  +=       0.530 * cos(5.61201247153  + 12.53017297220 *T)
    sumL1  +=       0.521 * cos(1.02371768017  + 415.29185818120 *T)
    sumL1  +=       0.639 * cos(0.68930265745  + 529.69096509460 *T)
    sumL1  +=       0.526 * cos(3.02138731705  + 5.62907429250 *T)
    sumL1  +=       0.456 * cos(4.44331571392  + 43.24084506850 *T)
    sumL1  +=       0.524 * cos(3.43316448349  + 38.65430049960 *T)
    sumL1  +=       0.436 * cos(2.41630174435  + 82.85835341460 *T)
    sumL1  +=       0.424 * cos(1.95736011325  + 477.33083545520 *T)
    sumL1  +=       0.443 * cos(3.39350946329  + 357.44566660120 *T)
    sumL1  +=       0.383 * cos(1.90232196422  + 22.63391724900 *T)
    sumL1  +=       0.479 * cos(5.55141744216  + 37.61177077600 *T)
    sumL1  +=       0.462 * cos(3.80436154644  + 343.21857259960 *T)
    sumL1  +=       0.384 * cos(5.60377408953  + 594.65070367540 *T)
    sumL1  +=       0.369 * cos(4.45577410338  + 6.90109867970 *T)
    sumL1  +=       0.358 * cos(3.69126616347  + 3.93215326310 *T)
    sumL1  +=       0.352 * cos(3.10952926034  + 135.54855145410 *T)
    sumL1  +=       0.368 * cos(3.53577440355  + 40.58071619260 *T)
    sumL1  +=       0.424 * cos(5.27159202779  + 181.75834193920 *T)
    sumL1  +=       0.361 * cos(0.29018303419  + 72.07328558160 *T)
    sumL1  +=       0.390 * cos(5.49512204296  + 350.33211960040 *T)
    sumL1  +=       0.378 * cos(2.74122401337  + 488.37653571910 *T)
    sumL1  +=       0.372 * cos(0.39980033572  + 494.73932319450 *T)
    sumL1  +=       0.353 * cos(1.10614174053  + 20.60692781950 *T)
    sumL1  +=       0.296 * cos(0.86351261285  + 149.56319713460 *T)
    sumL1  +=       0.307 * cos(5.39420288683  + 160.93896579860 *T)
    sumL1  +=       0.395 * cos(1.93577214824  + 10137.01947493540 *T)
    sumL1  +=       0.288 * cos(2.28755739359  + 47.69426319340 *T)
    sumL1  +=       0.295 * cos(2.48737537240  + 19.12245511120 *T)
    sumL1  +=       0.290 * cos(0.18636083306  + 143.62530630140 *T)
    sumL1  +=       0.266 * cos(3.09977370364  + 69.36497259590 *T)
    sumL1  +=       0.266 * cos(1.21002824826  + 505.78502345840 *T)
    sumL1  +=       0.252 * cos(3.12745026026  + 460.53844081980 *T)
    sumL1  +=       0.328 * cos(0.50849285663  + 6206.80977871580 *T)
    sumL1  +=       0.257 * cos(3.64119914774  + 446.31134681820 *T)
    sumL1  +=       0.239 * cos(5.54080102299  + 911.04257333200 *T)
    sumL1  +=       0.265 * cos(0.62702473701  + 253.57099508990 *T)
    sumL1  +=       0.287 * cos(2.44403568436  + 16.67477455640 *T)
    sumL1  +=       0.231 * cos(2.47026250085  + 454.90936652730 *T)
    sumL1  +=       0.230 * cos(3.24571542922  + 1066.49547719000 *T)
    sumL1  +=       0.282 * cos(1.48595620175  + 983.11585891360 *T)
    sumL1  +=       0.212 * cos(5.41931177641  + 64.95973858080 *T)
    sumL1  +=       0.213 * cos(1.64175339637  + 1089.12939443900 *T)
    sumL1  +=       0.238 * cos(2.69801319489  + 882.94384600180 *T)
    sumL1  +=       0.210 * cos(4.53976756699  + 1093.32218013300 *T)
    sumL1  +=       0.220 * cos(2.30038816175  + 1052.26838318840 *T)
    sumL1  +=       0.256 * cos(0.42073598460  + 23.90594163620 *T)
    sumL1  +=       0.216 * cos(5.44225918870  + 39.09624348430 *T)
    sumL1  +=       0.201 * cos(2.58746514605  + 119.50691634410 *T)
    sumL1  +=       0.224 * cos(4.43751392203  + 639.89728631400 *T)
    sumL1  +=       0.186 * cos(2.50651218075  + 487.36514376280 *T)
    sumL1  +=       0.189 * cos(4.05785534221  + 120.99138905240 *T)
    sumL1  +=       0.184 * cos(2.24245977278  + 815.06334611420 *T)
    sumL1  +=       0.202 * cos(3.43517732411  + 45.24658263860 *T)
    sumL1  +=       0.175 * cos(4.49165234532  + 171.23390653710 *T)
    sumL1  +=       0.171 * cos(5.50633466316  + 179.09821306330 *T)
    sumL1  +=       0.200 * cos(6.12663205401  + 14.22709400160 *T)
    sumL1  +=       0.173 * cos(2.61090344107  + 389.94962794650 *T)
    sumL1  +=       0.167 * cos(3.94754384833  + 77.22927912210 *T)
    sumL1  +=       0.166 * cos(3.41009128748  + 81.37388070630 *T)
    sumL1  +=       0.163 * cos(3.88198848446  + 556.51766803760 *T)
    sumL1  +=       0.164 * cos(1.49614763046  + 63.73589830340 *T)
    sumL1  +=       0.176 * cos(3.86129425367  + 148.33935685720 *T)
    sumL1  +=       0.161 * cos(2.22215642318  + 574.93754773320 *T)
    sumL1  +=       0.171 * cos(0.66899426684  + 179.31066138440 *T)
    sumL1  +=       0.161 * cos(1.21480182441  + 1024.43028828910 *T)
    sumL1  +=       0.155 * cos(3.25842414799  + 10251.41858184880 *T)
    sumL1  +=       0.183 * cos(5.45168150656  + 218.40690486870 *T)
    sumL1  +=       0.152 * cos(3.35145509017  + 285.37238101960 *T)
    sumL1  +=       0.152 * cos(0.42398786475  + 274.06604832480 *T)
    sumL1  +=       0.146 * cos(5.70714579127  + 419.48464387520 *T)
    sumL1  +=       0.156 * cos(0.64321524870  + 1029.84691426050 *T)
    sumL1  +=       0.147 * cos(4.30958930740  + 157.63995198190 *T)
    sumL1  +=       0.147 * cos(1.80689177510  + 377.41945497430 *T)
    sumL1  +=       0.140 * cos(1.49826604627  + 386.98068252990 *T)
    sumL1  +=       0.137 * cos(2.14480243915  + 563.63121503840 *T)
    sumL1  +=       0.127 * cos(3.98726599710  + 84.34282612290 *T)
    sumL1  +=       0.134 * cos(4.16039455079  + 169.53698550770 *T)
    sumL1  +=       0.121 * cos(0.29300927469  + 206.18554843720 *T)
    sumL1  +=       0.129 * cos(2.67625057010  + 180.79513409270 *T)
    sumL1  +=       0.134 * cos(3.18868986487  + 166.56804009110 *T)
    sumL1  +=       0.135 * cos(5.07517561780  + 426.59819087600 *T)
    sumL1  +=       0.136 * cos(1.81672451740  + 151.04766984290 *T)
    sumL1  +=       0.129 * cos(3.64795525602  + 183.76407950930 *T)
    sumL1  +=       0.116 * cos(6.06435563172  + 220.41264243880 *T)
    sumL1  +=       0.123 * cos(4.46641157829  + 1022.73336725970 *T)
    sumL1  +=       0.112 * cos(4.34485256988  + 138.51749687070 *T)
    sumL1  +=       0.116 * cos(5.58946529961  + 35.68535508300 *T)
    sumL1  +=       0.108 * cos(1.03796693383  + 488.58898404020 *T)
    sumL1  +=       0.108 * cos(2.10378485880  + 494.52687487340 *T)
    sumL1  +=       0.106 * cos(0.87068583107  + 1059.38193018920 *T)
    sumL1  +=       0.097 * cos(0.74486741478  + 485.92885516430 *T)
    sumL1  +=       0.095 * cos(5.54259914856  + 497.18700374930 *T)
    sumL1  +=       0.085 * cos(3.16062141266  + 522.57741809380 *T)
    sumL1  +=       0.097 * cos(6.05634803604  + 482.95990974770 *T)
    sumL1  +=       0.095 * cos(0.23111852730  + 500.15594916590 *T)
    sumL1  +=       0.084 * cos(2.64687252518  + 536.80451209540 *T)
    sumL1  +=       0.074 * cos(3.90678924318  + 1019.76442184310 *T)

    # NEPTUNE Longitude
    sumL2  = 0
    sumL2  +=   53892.649 * cos(0.00000000000  + 0.00000000000 *T)
    sumL2  +=     281.251 * cos(1.19084538887  + 38.13303563780 *T)
    sumL2  +=     295.693 * cos(1.85520292248  + 1.48447270830 *T)
    sumL2  +=     270.190 * cos(5.72143228148  + 76.26607127560 *T)
    sumL2  +=      23.023 * cos(1.21035596452  + 2.96894541660 *T)
    sumL2  +=       7.333 * cos(0.54033306830  + 2.44768055480 *T)
    sumL2  +=       9.057 * cos(4.42544992035  + 35.16409022120 *T)
    sumL2  +=       5.223 * cos(0.67427930044  + 168.05251279940 *T)
    sumL2  +=       5.201 * cos(3.02338671812  + 182.27960680100 *T)
    sumL2  +=       4.288 * cos(3.84351844003  + 114.39910691340 *T)
    sumL2  +=       3.925 * cos(3.53214557374  + 484.44438245600 *T)
    sumL2  +=       3.741 * cos(5.90238217874  + 498.67147645760 *T)
    sumL2  +=       2.966 * cos(0.31002477611  + 4.45341812490 *T)
    sumL2  +=       3.415 * cos(0.55971639038  + 74.78159856730 *T)
    sumL2  +=       3.255 * cos(1.84921884906  + 175.16605980020 *T)
    sumL2  +=       2.157 * cos(1.89135758747  + 388.46515523820 *T)
    sumL2  +=       2.211 * cos(4.37997092240  + 7.11354700080 *T)
    sumL2  +=       1.847 * cos(3.48574435762  + 9.56122755560 *T)
    sumL2  +=       2.451 * cos(4.68586840176  + 491.55792945680 *T)
    sumL2  +=       1.844 * cos(5.12281562096  + 33.67961751290 *T)
    sumL2  +=       2.204 * cos(1.69321574906  + 77.75054398390 *T)
    sumL2  +=       1.652 * cos(2.55859494053  + 36.64856292950 *T)
    sumL2  +=       1.309 * cos(4.52400192922  + 1021.24889455140 *T)
    sumL2  +=       1.124 * cos(0.38710602242  + 137.03302416240 *T)
    sumL2  +=       0.664 * cos(0.88101734307  + 4.19278569400 *T)
    sumL2  +=       0.497 * cos(2.24615784762  + 395.57870223900 *T)
    sumL2  +=       0.512 * cos(6.22609200672  + 381.35160823740 *T)
    sumL2  +=       0.582 * cos(5.25716719826  + 31.01948863700 *T)
    sumL2  +=       0.446 * cos(0.36647221351  + 98.89998852460 *T)
    sumL2  +=       0.383 * cos(5.48585528762  + 5.93789083320 *T)
    sumL2  +=       0.375 * cos(4.61250246774  + 8.07675484730 *T)
    sumL2  +=       0.354 * cos(1.30783918287  + 601.76425067620 *T)
    sumL2  +=       0.259 * cos(5.66033623678  + 112.91463420510 *T)
    sumL2  +=       0.247 * cos(2.89695614593  + 189.39315380180 *T)
    sumL2  +=       0.245 * cos(4.26572913391  + 220.41264243880 *T)
    sumL2  +=       0.200 * cos(0.52604535784  + 64.95973858080 *T)
    sumL2  +=       0.191 * cos(4.88786653062  + 39.61750834610 *T)
    sumL2  +=       0.233 * cos(3.16423779113  + 41.10198105440 *T)
    sumL2  +=       0.248 * cos(5.85877831382  + 1059.38193018920 *T)
    sumL2  +=       0.194 * cos(2.37949641473  + 73.29712585900 *T)
    sumL2  +=       0.227 * cos(0.20028518978  + 60.76695288680 *T)
    sumL2  +=       0.184 * cos(3.01962045713  + 1014.13534755060 *T)
    sumL2  +=       0.190 * cos(5.57500985081  + 343.21857259960 *T)
    sumL2  +=       0.172 * cos(3.66036463613  + 477.33083545520 *T)
    sumL2  +=       0.172 * cos(0.59550457102  + 46.20979048510 *T)
    sumL2  +=       0.182 * cos(1.92429384025  + 183.76407950930 *T)
    sumL2  +=       0.171 * cos(1.61368476689  + 357.44566660120 *T)
    sumL2  +=       0.173 * cos(6.23717119485  + 493.04240216510 *T)
    sumL2  +=       0.217 * cos(1.46218158211  + 71.81265315070 *T)
    sumL2  +=       0.178 * cos(0.34928799031  + 1028.36244155220 *T)
    sumL2  +=       0.169 * cos(4.91086673212  + 166.56804009110 *T)
    sumL2  +=       0.157 * cos(5.89200571154  + 169.53698550770 *T)
    sumL2  +=       0.182 * cos(2.33457064554  + 152.53214255120 *T)
    sumL2  +=       0.151 * cos(3.81621340568  + 146.59425171800 *T)
    sumL2  +=       0.136 * cos(2.75150881988  + 144.14657116320 *T)
    sumL2  +=       0.104 * cos(6.03262825314  + 529.69096509460 *T)
    sumL2  +=       0.076 * cos(0.20932812381  + 453.42489381900 *T)

    # NEPTUNE Longitude
    sumL3  = 0
    sumL3  +=      31.254 * cos(0.00000000000  + 0.00000000000 *T)
    sumL3  +=      12.461 * cos(6.04431418812  + 1.48447270830 *T)
    sumL3  +=      14.541 * cos(1.35337075856  + 76.26607127560 *T)
    sumL3  +=      11.547 * cos(6.11257808366  + 38.13303563780 *T)
    sumL3  +=       1.351 * cos(4.93951495175  + 2.96894541660 *T)
    sumL3  +=       0.741 * cos(2.35936954597  + 168.05251279940 *T)
    sumL3  +=       0.715 * cos(1.27409542804  + 182.27960680100 *T)
    sumL3  +=       0.537 * cos(5.23632185196  + 484.44438245600 *T)
    sumL3  +=       0.523 * cos(4.16769839601  + 498.67147645760 *T)
    sumL3  +=       0.664 * cos(0.55871435877  + 31.01948863700 *T)
    sumL3  +=       0.301 * cos(2.69253200796  + 7.11354700080 *T)
    sumL3  +=       0.194 * cos(2.05904114139  + 137.03302416240 *T)
    sumL3  +=       0.206 * cos(2.51012178002  + 74.78159856730 *T)
    sumL3  +=       0.160 * cos(5.63111039032  + 114.39910691340 *T)
    sumL3  +=       0.149 * cos(3.09327713923  + 35.16409022120 *T)

    # NEPTUNE Longitude
    sumL4  = 0
    sumL4  +=     113.998 * cos(3.14159265359  + 0.00000000000 *T)
    sumL4  +=       0.605 * cos(3.18211885677  + 76.26607127560 *T)

    # NEPTUNE Longitude
    sumL5  = 0
    sumL5  +=       0.874 * cos(3.14159265359  + 0.00000000000 *T)

    # NEPTUNE Latitude
    sumB0  = 0
    sumB0  +=  3088622.933 * cos(1.44104372626  + 38.13303563780 *T)
    sumB0  +=   27780.087 * cos(5.91271882843  + 76.26607127560 *T)
    sumB0  +=   27623.609 * cos(0.00000000000  + 0.00000000000 *T)
    sumB0  +=   15355.490 * cos(2.52123799481  + 36.64856292950 *T)
    sumB0  +=   15448.133 * cos(3.50877080888  + 39.61750834610 *T)
    sumB0  +=    1999.919 * cos(1.50998669505  + 74.78159856730 *T)
    sumB0  +=    1967.540 * cos(4.37778195768  + 1.48447270830 *T)
    sumB0  +=    1015.137 * cos(3.21561035875  + 35.16409022120 *T)
    sumB0  +=     605.767 * cos(2.80246601405  + 73.29712585900 *T)
    sumB0  +=     594.878 * cos(2.12892708114  + 41.10198105440 *T)
    sumB0  +=     588.805 * cos(3.18655882497  + 2.96894541660 *T)
    sumB0  +=     401.830 * cos(4.16883287237  + 114.39910691340 *T)
    sumB0  +=     254.333 * cos(3.27120499438  + 453.42489381900 *T)
    sumB0  +=     261.647 * cos(3.76722704749  + 213.29909543800 *T)
    sumB0  +=     279.964 * cos(1.68165309699  + 77.75054398390 *T)
    sumB0  +=     205.590 * cos(4.25652348864  + 529.69096509460 *T)
    sumB0  +=     140.455 * cos(3.52969556376  + 137.03302416240 *T)
    sumB0  +=      98.530 * cos(4.16774829927  + 33.67961751290 *T)
    sumB0  +=      51.257 * cos(1.95121181203  + 4.45341812490 *T)
    sumB0  +=      67.971 * cos(4.66970781659  + 71.81265315070 *T)
    sumB0  +=      41.931 * cos(5.41783694467  + 111.43016149680 *T)
    sumB0  +=      41.822 * cos(5.94832001477  + 112.91463420510 *T)
    sumB0  +=      30.637 * cos(0.93620571932  + 42.58645376270 *T)
    sumB0  +=      11.084 * cos(5.88898793049  + 108.46121608020 *T)
    sumB0  +=       9.620 * cos(0.03944255108  + 70.32818044240 *T)
    sumB0  +=       9.664 * cos(0.22455797403  + 79.23501669220 *T)
    sumB0  +=       9.728 * cos(5.30069593532  + 32.19514480460 *T)
    sumB0  +=       7.386 * cos(3.00684933642  + 426.59819087600 *T)
    sumB0  +=       7.087 * cos(0.12535040656  + 109.94568878850 *T)
    sumB0  +=       6.021 * cos(6.20514068152  + 115.88357962170 *T)
    sumB0  +=       6.169 * cos(3.62098109648  + 983.11585891360 *T)
    sumB0  +=       4.777 * cos(0.75210194972  + 5.93789083320 *T)
    sumB0  +=       6.391 * cos(5.84646101060  + 148.07872442630 *T)
    sumB0  +=       6.251 * cos(2.41678769385  + 152.53214255120 *T)
    sumB0  +=       4.539 * cos(5.58182098700  + 175.16605980020 *T)
    sumB0  +=       5.006 * cos(4.60815664851  + 1059.38193018920 *T)
    sumB0  +=       4.289 * cos(4.19647392821  + 47.69426319340 *T)
    sumB0  +=       5.795 * cos(5.07516716087  + 415.29185818120 *T)
    sumB0  +=       4.749 * cos(2.51605725604  + 37.61177077600 *T)
    sumB0  +=       4.119 * cos(1.72779509865  + 28.57180808220 *T)
    sumB0  +=       4.076 * cos(6.00252170354  + 145.10977900970 *T)
    sumB0  +=       4.429 * cos(5.65995321659  + 98.89998852460 *T)
    sumB0  +=       3.950 * cos(2.74104636753  + 350.33211960040 *T)
    sumB0  +=       4.091 * cos(1.61787956945  + 39.09624348430 *T)
    sumB0  +=       4.131 * cos(4.40682554313  + 37.16982779130 *T)
    sumB0  +=       4.710 * cos(3.50929350767  + 38.65430049960 *T)
    sumB0  +=       4.440 * cos(4.78977105547  + 38.08485152800 *T)
    sumB0  +=       4.433 * cos(1.23386935925  + 38.18121974760 *T)
    sumB0  +=       3.762 * cos(4.83940791709  + 491.55792945680 *T)
    sumB0  +=       2.606 * cos(1.20956732792  + 451.94042111070 *T)
    sumB0  +=       2.537 * cos(2.18628045751  + 454.90936652730 *T)
    sumB0  +=       2.328 * cos(5.19779918719  + 72.07328558160 *T)
    sumB0  +=       2.502 * cos(0.85987904350  + 106.97674337190 *T)
    sumB0  +=       2.342 * cos(0.81387240947  + 4.19278569400 *T)
    sumB0  +=       1.981 * cos(0.46617960831  + 184.72728735580 *T)
    sumB0  +=       1.963 * cos(6.01909114576  + 44.07092647100 *T)
    sumB0  +=       2.180 * cos(0.70099749844  + 206.18554843720 *T)
    sumB0  +=       1.811 * cos(0.40456996647  + 40.58071619260 *T)
    sumB0  +=       1.814 * cos(3.64699555185  + 220.41264243880 *T)
    sumB0  +=       1.705 * cos(6.13551142362  + 181.75834193920 *T)
    sumB0  +=       1.855 * cos(5.61635630213  + 35.68535508300 *T)
    sumB0  +=       1.595 * cos(2.97147156093  + 37.87240320690 *T)
    sumB0  +=       1.785 * cos(2.42154818096  + 388.46515523820 *T)
    sumB0  +=       1.595 * cos(3.05266110075  + 38.39366806870 *T)
    sumB0  +=       1.437 * cos(1.48678704605  + 135.54855145410 *T)
    sumB0  +=       1.387 * cos(2.46149266117  + 138.51749687070 *T)
    sumB0  +=       1.366 * cos(1.52026779665  + 68.84370773410 *T)
    sumB0  +=       1.575 * cos(3.58964541604  + 38.02116105320 *T)
    sumB0  +=       1.297 * cos(5.06156596196  + 33.94024994380 *T)
    sumB0  +=       1.487 * cos(0.20211121607  + 30.05628079050 *T)
    sumB0  +=       1.504 * cos(5.80298577327  + 46.20979048510 *T)
    sumB0  +=       1.192 * cos(0.87275514483  + 42.32582133180 *T)
    sumB0  +=       1.569 * cos(2.43405967107  + 38.24491022240 *T)
    sumB0  +=       1.207 * cos(1.84658687853  + 251.43213107580 *T)
    sumB0  +=       1.015 * cos(0.53439848924  + 129.91947716160 *T)
    sumB0  +=       0.999 * cos(2.47463873948  + 312.19908396260 *T)
    sumB0  +=       0.990 * cos(3.41514319052  + 144.14657116320 *T)
    sumB0  +=       0.963 * cos(4.31733242907  + 151.04766984290 *T)
    sumB0  +=       1.020 * cos(0.98226686775  + 143.62530630140 *T)
    sumB0  +=       0.941 * cos(1.02993053785  + 221.37585028530 *T)
    sumB0  +=       0.938 * cos(2.43648356625  + 567.82400073240 *T)
    sumB0  +=       1.111 * cos(0.65175024456  + 146.59425171800 *T)
    sumB0  +=       0.777 * cos(0.00175975222  + 218.40690486870 *T)
    sumB0  +=       0.895 * cos(0.25123869620  + 30.71067209630 *T)
    sumB0  +=       0.795 * cos(5.80519741659  + 149.56319713460 *T)
    sumB0  +=       0.737 * cos(3.40060492866  + 446.31134681820 *T)
    sumB0  +=       0.719 * cos(1.43795191278  + 8.07675484730 *T)
    sumB0  +=       0.720 * cos(0.00651007550  + 460.53844081980 *T)
    sumB0  +=       0.766 * cos(4.03399506246  + 522.57741809380 *T)
    sumB0  +=       0.666 * cos(1.39457824982  + 84.34282612290 *T)
    sumB0  +=       0.584 * cos(1.01405548136  + 536.80451209540 *T)
    sumB0  +=       0.596 * cos(0.62390100715  + 35.21227433100 *T)
    sumB0  +=       0.598 * cos(5.39946724188  + 41.05379694460 *T)
    sumB0  +=       0.475 * cos(5.80072248338  + 7.42236354150 *T)
    sumB0  +=       0.510 * cos(1.34478579740  + 258.02441321480 *T)
    sumB0  +=       0.458 * cos(5.25325523118  + 80.71948940050 *T)
    sumB0  +=       0.421 * cos(3.24496387889  + 416.77633088950 *T)
    sumB0  +=       0.446 * cos(1.19167306357  + 180.27386923090 *T)
    sumB0  +=       0.471 * cos(0.92632922375  + 44.72531777680 *T)
    sumB0  +=       0.387 * cos(1.68488418788  + 183.24281464750 *T)
    sumB0  +=       0.375 * cos(0.15223869165  + 255.05546779820 *T)
    sumB0  +=       0.354 * cos(4.21526988674  + 0.96320784650 *T)
    sumB0  +=       0.379 * cos(2.16947487177  + 105.49227066360 *T)
    sumB0  +=       0.341 * cos(4.79194051680  + 110.20632121940 *T)
    sumB0  +=       0.427 * cos(5.15774894584  + 31.54075349880 *T)
    sumB0  +=       0.302 * cos(3.45706306280  + 100.38446123290 *T)
    sumB0  +=       0.298 * cos(2.26790695187  + 639.89728631400 *T)
    sumB0  +=       0.279 * cos(0.25689162963  + 39.50563376150 *T)
    sumB0  +=       0.320 * cos(3.58085653166  + 45.24658263860 *T)
    sumB0  +=       0.269 * cos(5.72024180826  + 36.76043751410 *T)
    sumB0  +=       0.247 * cos(0.61040148804  + 186.21176006410 *T)
    sumB0  +=       0.245 * cos(0.64173616273  + 419.48464387520 *T)
    sumB0  +=       0.235 * cos(0.73189197665  + 10213.28554621100 *T)
    sumB0  +=       0.232 * cos(0.37399822852  + 490.07345674850 *T)
    sumB0  +=       0.230 * cos(5.76570492457  + 12.53017297220 *T)
    sumB0  +=       0.240 * cos(4.13447692727  + 0.52126486180 *T)
    sumB0  +=       0.279 * cos(1.62614865256  + 294.67297614430 *T)
    sumB0  +=       0.238 * cos(2.18528916550  + 219.89137757700 *T)
    sumB0  +=       0.262 * cos(3.08384135298  + 6.59228213900 *T)
    sumB0  +=       0.217 * cos(2.93214905312  + 27.08733537390 *T)
    sumB0  +=       0.217 * cos(4.69210602828  + 406.10313764110 *T)
    sumB0  +=       0.219 * cos(1.35212712560  + 216.92243216040 *T)
    sumB0  +=       0.200 * cos(2.35215465744  + 605.95703637020 *T)
    sumB0  +=       0.232 * cos(3.92583619589  + 1512.80682400820 *T)
    sumB0  +=       0.223 * cos(5.52392277606  + 187.69623277240 *T)
    sumB0  +=       0.190 * cos(0.29169556516  + 291.70403072770 *T)
    sumB0  +=       0.236 * cos(3.12464145036  + 563.63121503840 *T)
    sumB0  +=       0.193 * cos(0.53675942386  + 60.76695288680 *T)
    sumB0  +=       0.215 * cos(3.78391259001  + 103.09277421860 *T)
    sumB0  +=       0.172 * cos(5.63262770743  + 7.11354700080 *T)
    sumB0  +=       0.164 * cos(4.14700645532  + 77.22927912210 *T)
    sumB0  +=       0.162 * cos(0.72021213236  + 11.04570026390 *T)
    sumB0  +=       0.160 * cos(4.23490438166  + 487.36514376280 *T)
    sumB0  +=       0.191 * cos(0.37651439206  + 31.01948863700 *T)
    sumB0  +=       0.157 * cos(1.02419759383  + 6283.07584999140 *T)
    sumB0  +=       0.157 * cos(4.42530429545  + 6206.80977871580 *T)
    sumB0  +=       0.178 * cos(6.24797160202  + 316.39186965660 *T)
    sumB0  +=       0.161 * cos(5.65988283675  + 343.21857259960 *T)
    sumB0  +=       0.153 * cos(5.58405022784  + 252.08652238160 *T)
    sumB0  +=       0.189 * cos(4.80791039970  + 641.12112659140 *T)
    sumB0  +=       0.166 * cos(5.50438043692  + 662.53120356300 *T)
    sumB0  +=       0.146 * cos(5.08949604858  + 286.59622129700 *T)
    sumB0  +=       0.145 * cos(2.13015521881  + 2042.49778910280 *T)
    sumB0  +=       0.156 * cos(2.19452173251  + 274.06604832480 *T)
    sumB0  +=       0.148 * cos(4.85696640135  + 442.75170057060 *T)
    sumB0  +=       0.187 * cos(4.96121139073  + 1589.07289528380 *T)
    sumB0  +=       0.155 * cos(2.28260574227  + 142.14083359310 *T)
    sumB0  +=       0.134 * cos(1.29277093566  + 456.39383923560 *T)
    sumB0  +=       0.126 * cos(5.59769497652  + 179.35884549420 *T)
    sumB0  +=       0.146 * cos(2.53359213478  + 256.53994050650 *T)
    sumB0  +=       0.140 * cos(1.57962199954  + 75.74480641380 *T)
    sumB0  +=       0.123 * cos(0.05442220184  + 944.98282327580 *T)
    sumB0  +=       0.122 * cos(1.90676379802  + 418.26080359780 *T)
    sumB0  +=       0.154 * cos(1.86865302773  + 331.32153907380 *T)
    sumB0  +=       0.144 * cos(5.52229258454  + 14.01464568050 *T)
    sumB0  +=       0.138 * cos(2.80728175526  + 82.85835341460 *T)
    sumB0  +=       0.107 * cos(0.66995358132  + 190.66517818900 *T)
    sumB0  +=       0.114 * cos(1.48894980280  + 253.57099508990 *T)
    sumB0  +=       0.110 * cos(5.32587573069  + 240.12579838100 *T)
    sumB0  +=       0.105 * cos(0.65548440578  + 173.68158709190 *T)
    sumB0  +=       0.102 * cos(2.58735617801  + 450.45594840240 *T)
    sumB0  +=       0.098 * cos(0.44044795266  + 328.35259365720 *T)
    sumB0  +=       0.101 * cos(4.71267656829  + 117.36805233000 *T)
    sumB0  +=       0.094 * cos(0.54938580474  + 293.18850343600 *T)
    sumB0  +=       0.095 * cos(2.17636214523  + 101.86893394120 *T)
    sumB0  +=       0.093 * cos(0.63687810471  + 377.15882254340 *T)
    sumB0  +=       0.091 * cos(5.84828809934  + 10137.01947493540 *T)
    sumB0  +=       0.089 * cos(1.02830167997  + 1021.24889455140 *T)
    sumB0  +=       0.094 * cos(1.79320597168  + 493.04240216510 *T)
    sumB0  +=       0.080 * cos(1.58140274465  + 69.15252427480 *T)
    sumB0  +=       0.075 * cos(0.23453373368  + 63.73589830340 *T)
    sumB0  +=       0.071 * cos(1.51961989690  + 488.58898404020 *T)

    # NEPTUNE Latitude
    sumB1  = 0
    sumB1  +=  227279.214 * cos(3.80793089870  + 38.13303563780 *T)
    sumB1  +=    1803.120 * cos(1.97576485377  + 76.26607127560 *T)
    sumB1  +=    1385.733 * cos(4.82555548018  + 36.64856292950 *T)
    sumB1  +=    1433.300 * cos(3.14159265359  + 0.00000000000 *T)
    sumB1  +=    1073.298 * cos(6.08054240712  + 39.61750834610 *T)
    sumB1  +=     147.903 * cos(3.85766231348  + 74.78159856730 *T)
    sumB1  +=     136.448 * cos(0.47764957338  + 1.48447270830 *T)
    sumB1  +=      70.285 * cos(6.18782052139  + 35.16409022120 *T)
    sumB1  +=      51.899 * cos(5.05221791891  + 73.29712585900 *T)
    sumB1  +=      37.273 * cos(4.89476629246  + 41.10198105440 *T)
    sumB1  +=      42.568 * cos(0.30721737205  + 114.39910691340 *T)
    sumB1  +=      37.104 * cos(5.75999349109  + 2.96894541660 *T)
    sumB1  +=      26.399 * cos(5.21566335936  + 213.29909543800 *T)
    sumB1  +=      16.949 * cos(4.26463671859  + 77.75054398390 *T)
    sumB1  +=      18.747 * cos(0.90426522185  + 453.42489381900 *T)
    sumB1  +=      12.951 * cos(6.17709713139  + 529.69096509460 *T)
    sumB1  +=      10.502 * cos(1.20336443465  + 137.03302416240 *T)
    sumB1  +=       4.416 * cos(1.25478204684  + 111.43016149680 *T)
    sumB1  +=       4.383 * cos(6.14147099615  + 71.81265315070 *T)
    sumB1  +=       3.694 * cos(0.94837702528  + 33.67961751290 *T)
    sumB1  +=       2.957 * cos(4.77532871210  + 4.45341812490 *T)
    sumB1  +=       2.698 * cos(1.92435531119  + 112.91463420510 *T)
    sumB1  +=       1.989 * cos(3.96637567224  + 42.58645376270 *T)
    sumB1  +=       1.150 * cos(4.30568700024  + 37.61177077600 *T)
    sumB1  +=       0.871 * cos(4.81775882249  + 152.53214255120 *T)
    sumB1  +=       0.944 * cos(2.21777772050  + 109.94568878850 *T)
    sumB1  +=       0.936 * cos(1.17054983940  + 148.07872442630 *T)
    sumB1  +=       0.925 * cos(2.40329074000  + 206.18554843720 *T)
    sumB1  +=       0.690 * cos(1.57381082857  + 38.65430049960 *T)
    sumB1  +=       0.624 * cos(2.79466003645  + 79.23501669220 *T)
    sumB1  +=       0.726 * cos(4.13829519132  + 28.57180808220 *T)
    sumB1  +=       0.640 * cos(2.46161252327  + 115.88357962170 *T)
    sumB1  +=       0.531 * cos(2.96991530500  + 98.89998852460 *T)
    sumB1  +=       0.537 * cos(1.95986772922  + 220.41264243880 *T)
    sumB1  +=       0.539 * cos(2.06690307827  + 40.58071619260 *T)
    sumB1  +=       0.716 * cos(0.55781847010  + 350.33211960040 *T)
    sumB1  +=       0.563 * cos(1.84072805158  + 983.11585891360 *T)
    sumB1  +=       0.533 * cos(1.34787677940  + 47.69426319340 *T)
    sumB1  +=       0.566 * cos(1.80111775954  + 175.16605980020 *T)
    sumB1  +=       0.449 * cos(1.62191691011  + 144.14657116320 *T)
    sumB1  +=       0.371 * cos(2.74239666472  + 415.29185818120 *T)
    sumB1  +=       0.381 * cos(6.11910193382  + 426.59819087600 *T)
    sumB1  +=       0.366 * cos(2.39752585360  + 129.91947716160 *T)
    sumB1  +=       0.456 * cos(3.19611413854  + 108.46121608020 *T)
    sumB1  +=       0.327 * cos(3.62341506247  + 38.18121974760 *T)
    sumB1  +=       0.328 * cos(0.89613145346  + 38.08485152800 *T)
    sumB1  +=       0.341 * cos(3.87265469070  + 35.68535508300 *T)
    sumB1  +=       0.331 * cos(4.48858774501  + 460.53844081980 *T)
    sumB1  +=       0.414 * cos(1.03543720726  + 70.32818044240 *T)
    sumB1  +=       0.310 * cos(0.51297445145  + 37.16982779130 *T)
    sumB1  +=       0.287 * cos(2.18351651800  + 491.55792945680 *T)
    sumB1  +=       0.274 * cos(6.11504724934  + 522.57741809380 *T)
    sumB1  +=       0.281 * cos(3.81657117512  + 5.93789083320 *T)
    sumB1  +=       0.298 * cos(4.00532631258  + 39.09624348430 *T)
    sumB1  +=       0.265 * cos(5.26569823181  + 446.31134681820 *T)
    sumB1  +=       0.319 * cos(1.34097217817  + 184.72728735580 *T)
    sumB1  +=       0.203 * cos(6.02944475303  + 149.56319713460 *T)
    sumB1  +=       0.205 * cos(5.53935732020  + 536.80451209540 *T)
    sumB1  +=       0.226 * cos(6.17710997862  + 454.90936652730 *T)
    sumB1  +=       0.186 * cos(3.24302117645  + 4.19278569400 *T)
    sumB1  +=       0.179 * cos(4.91458426239  + 451.94042111070 *T)
    sumB1  +=       0.198 * cos(2.30775852880  + 146.59425171800 *T)
    sumB1  +=       0.166 * cos(1.16793600058  + 72.07328558160 *T)
    sumB1  +=       0.147 * cos(2.10574339673  + 44.07092647100 *T)
    sumB1  +=       0.123 * cos(1.98250467171  + 46.20979048510 *T)
    sumB1  +=       0.159 * cos(3.46955908364  + 145.10977900970 *T)
    sumB1  +=       0.116 * cos(5.88971113590  + 38.02116105320 *T)
    sumB1  +=       0.115 * cos(4.73412534395  + 38.24491022240 *T)
    sumB1  +=       0.125 * cos(3.42713474801  + 251.43213107580 *T)
    sumB1  +=       0.128 * cos(1.51108932026  + 221.37585028530 *T)
    sumB1  +=       0.127 * cos(0.17176461812  + 138.51749687070 *T)
    sumB1  +=       0.124 * cos(5.85160407534  + 1059.38193018920 *T)
    sumB1  +=       0.091 * cos(2.38273591235  + 30.05628079050 *T)
    sumB1  +=       0.118 * cos(5.27114846878  + 37.87240320690 *T)
    sumB1  +=       0.117 * cos(5.35267669439  + 38.39366806870 *T)
    sumB1  +=       0.099 * cos(5.19920708255  + 135.54855145410 *T)
    sumB1  +=       0.114 * cos(4.37452353441  + 388.46515523820 *T)
    sumB1  +=       0.093 * cos(4.64183693718  + 106.97674337190 *T)
    sumB1  +=       0.084 * cos(1.35269684746  + 33.94024994380 *T)
    sumB1  +=       0.111 * cos(3.56226463770  + 181.75834193920 *T)
    sumB1  +=       0.082 * cos(3.18401661435  + 42.32582133180 *T)
    sumB1  +=       0.084 * cos(5.51669920239  + 8.07675484730 *T)

    # NEPTUNE Latitude
    sumB2  = 0
    sumB2  +=    9690.766 * cos(5.57123750291  + 38.13303563780 *T)
    sumB2  +=      78.815 * cos(3.62705474219  + 76.26607127560 *T)
    sumB2  +=      71.523 * cos(0.45476688580  + 36.64856292950 *T)
    sumB2  +=      58.646 * cos(3.14159265359  + 0.00000000000 *T)
    sumB2  +=      29.915 * cos(1.60671721861  + 39.61750834610 *T)
    sumB2  +=       6.472 * cos(5.60736756575  + 74.78159856730 *T)
    sumB2  +=       5.800 * cos(2.25341847151  + 1.48447270830 *T)
    sumB2  +=       4.309 * cos(1.68126737666  + 35.16409022120 *T)
    sumB2  +=       3.502 * cos(2.39142672984  + 114.39910691340 *T)
    sumB2  +=       2.649 * cos(0.65061457644  + 73.29712585900 *T)
    sumB2  +=       1.518 * cos(0.37600329684  + 213.29909543800 *T)
    sumB2  +=       1.223 * cos(1.23116043030  + 2.96894541660 *T)
    sumB2  +=       0.766 * cos(5.45279753249  + 453.42489381900 *T)
    sumB2  +=       0.779 * cos(2.07081431472  + 529.69096509460 *T)
    sumB2  +=       0.496 * cos(0.26552533921  + 41.10198105440 *T)
    sumB2  +=       0.469 * cos(5.87866293959  + 77.75054398390 *T)
    sumB2  +=       0.482 * cos(5.63056237954  + 137.03302416240 *T)
    sumB2  +=       0.345 * cos(1.80085651594  + 71.81265315070 *T)
    sumB2  +=       0.274 * cos(2.86650141006  + 33.67961751290 *T)
    sumB2  +=       0.158 * cos(4.63868656467  + 206.18554843720 *T)
    sumB2  +=       0.166 * cos(1.24877330835  + 220.41264243880 *T)
    sumB2  +=       0.153 * cos(2.87376446497  + 111.43016149680 *T)
    sumB2  +=       0.116 * cos(3.63838544843  + 112.91463420510 *T)
    sumB2  +=       0.085 * cos(0.43712705655  + 4.45341812490 *T)
    sumB2  +=       0.104 * cos(6.12597614674  + 144.14657116320 *T)

    # NEPTUNE Latitude
    sumB3  = 0
    sumB3  +=     273.423 * cos(1.01688979072  + 38.13303563780 *T)
    sumB3  +=       2.274 * cos(2.36805657126  + 36.64856292950 *T)
    sumB3  +=       2.029 * cos(5.33364321342  + 76.26607127560 *T)
    sumB3  +=       2.393 * cos(0.00000000000  + 0.00000000000 *T)
    sumB3  +=       0.538 * cos(3.21934211365  + 39.61750834610 *T)
    sumB3  +=       0.242 * cos(4.52650721578  + 114.39910691340 *T)
    sumB3  +=       0.185 * cos(1.04913770083  + 74.78159856730 *T)
    sumB3  +=       0.155 * cos(3.62376309338  + 35.16409022120 *T)
    sumB3  +=       0.157 * cos(3.94195369610  + 1.48447270830 *T)

    # NEPTUNE Latitude
    sumB4  = 0
    sumB4  +=       5.728 * cos(2.66872693322  + 38.13303563780 *T)

    # NEPTUNE Latitude
    sumB5  = 0
    sumB5  +=       0.113 * cos(4.70646877989  + 38.13303563780 *T)

    # NEPTUNE Radius
    sumR0  = 0
    sumR0  +=  3007013206.102 * cos(0.00000000000  + 0.00000000000 *T)
    sumR0  +=  27062259.490 * cos(1.32999458930  + 38.13303563780 *T)
    sumR0  +=  1691764.281 * cos(3.25186138896  + 36.64856292950 *T)
    sumR0  +=  807830.737 * cos(5.18592836167  + 1.48447270830 *T)
    sumR0  +=  537760.613 * cos(4.52113902845  + 35.16409022120 *T)
    sumR0  +=  495725.642 * cos(1.57105654815  + 491.55792945680 *T)
    sumR0  +=  274571.970 * cos(1.84552256801  + 175.16605980020 *T)
    sumR0  +=  135134.095 * cos(3.37220607384  + 39.61750834610 *T)
    sumR0  +=  121801.825 * cos(5.79754444303  + 76.26607127560 *T)
    sumR0  +=  100895.397 * cos(0.37702748681  + 73.29712585900 *T)
    sumR0  +=   69791.722 * cos(3.79617226928  + 2.96894541660 *T)
    sumR0  +=   46687.838 * cos(5.74937810094  + 33.67961751290 *T)
    sumR0  +=   24593.778 * cos(0.50801728204  + 109.94568878850 *T)
    sumR0  +=   16939.242 * cos(1.59422166991  + 71.81265315070 *T)
    sumR0  +=   14229.686 * cos(1.07786112902  + 74.78159856730 *T)
    sumR0  +=   12011.825 * cos(1.92062131635  + 1021.24889455140 *T)
    sumR0  +=    8394.731 * cos(0.67816895547  + 146.59425171800 *T)
    sumR0  +=    7571.800 * cos(1.07149263431  + 388.46515523820 *T)
    sumR0  +=    5720.852 * cos(2.59059512267  + 4.45341812490 *T)
    sumR0  +=    4839.672 * cos(1.90685991070  + 41.10198105440 *T)
    sumR0  +=    4483.492 * cos(2.90573457534  + 529.69096509460 *T)
    sumR0  +=    4270.202 * cos(3.41343865825  + 453.42489381900 *T)
    sumR0  +=    4353.790 * cos(0.67985662370  + 32.19514480460 *T)
    sumR0  +=    4420.804 * cos(1.74993796503  + 108.46121608020 *T)
    sumR0  +=    2881.063 * cos(1.98600105123  + 137.03302416240 *T)
    sumR0  +=    2635.535 * cos(3.09755943422  + 213.29909543800 *T)
    sumR0  +=    3380.930 * cos(0.84810683275  + 183.24281464750 *T)
    sumR0  +=    2878.942 * cos(3.67415901855  + 350.33211960040 *T)
    sumR0  +=    2306.293 * cos(2.80962935724  + 70.32818044240 *T)
    sumR0  +=    2530.149 * cos(5.79839567009  + 490.07345674850 *T)
    sumR0  +=    2523.132 * cos(0.48630800015  + 493.04240216510 *T)
    sumR0  +=    2087.303 * cos(0.61858378281  + 33.94024994380 *T)
    sumR0  +=    1976.522 * cos(5.11703044560  + 168.05251279940 *T)
    sumR0  +=    1905.254 * cos(1.72186472126  + 182.27960680100 *T)
    sumR0  +=    1654.039 * cos(1.92782545887  + 145.10977900970 *T)
    sumR0  +=    1435.072 * cos(1.70005157785  + 484.44438245600 *T)
    sumR0  +=    1403.029 * cos(4.58914203187  + 498.67147645760 *T)
    sumR0  +=    1499.193 * cos(1.01623299513  + 219.89137757700 *T)
    sumR0  +=    1398.860 * cos(0.76220317620  + 176.65053250850 *T)
    sumR0  +=    1403.377 * cos(6.07659416908  + 173.68158709190 *T)
    sumR0  +=    1128.560 * cos(5.96661179805  + 9.56122755560 *T)
    sumR0  +=    1228.304 * cos(1.59881465324  + 77.75054398390 *T)
    sumR0  +=     835.414 * cos(3.97066884218  + 114.39910691340 *T)
    sumR0  +=     811.186 * cos(3.00258880870  + 46.20979048510 *T)
    sumR0  +=     731.925 * cos(2.10447054189  + 181.75834193920 *T)
    sumR0  +=     615.781 * cos(2.97874625677  + 106.97674337190 *T)
    sumR0  +=     704.778 * cos(1.18738210880  + 256.53994050650 *T)
    sumR0  +=     502.040 * cos(1.38657803368  + 5.93789083320 *T)
    sumR0  +=     530.357 * cos(4.24059166485  + 111.43016149680 *T)
    sumR0  +=     437.096 * cos(2.27029212923  + 1550.93985964600 *T)
    sumR0  +=     400.250 * cos(1.25609325435  + 8.07675484730 *T)
    sumR0  +=     421.011 * cos(1.89084929506  + 30.71067209630 *T)
    sumR0  +=     382.457 * cos(3.29965259685  + 983.11585891360 *T)
    sumR0  +=     422.485 * cos(5.53186169605  + 525.49817940060 *T)
    sumR0  +=     355.389 * cos(2.27847846648  + 218.40690486870 *T)
    sumR0  +=     280.062 * cos(1.54129714238  + 98.89998852460 *T)
    sumR0  +=     314.499 * cos(3.95932948594  + 381.35160823740 *T)
    sumR0  +=     280.556 * cos(4.54238271682  + 44.72531777680 *T)
    sumR0  +=     267.738 * cos(5.13323364247  + 112.91463420510 *T)
    sumR0  +=     333.311 * cos(5.75067616021  + 39.09624348430 *T)
    sumR0  +=     291.625 * cos(4.02398326341  + 68.84370773410 *T)
    sumR0  +=     321.429 * cos(1.50625025822  + 454.90936652730 *T)
    sumR0  +=     309.196 * cos(2.85452752153  + 72.07328558160 *T)
    sumR0  +=     345.094 * cos(1.35905860594  + 293.18850343600 *T)
    sumR0  +=     307.439 * cos(0.31964571332  + 601.76425067620 *T)
    sumR0  +=     251.356 * cos(3.53992782846  + 312.19908396260 *T)
    sumR0  +=     248.152 * cos(3.41078346726  + 37.61177077600 *T)
    sumR0  +=     306.000 * cos(2.72475094464  + 6244.94281435360 *T)
    sumR0  +=     293.532 * cos(4.89079857814  + 528.20649238630 *T)
    sumR0  +=     234.479 * cos(0.59231043427  + 42.58645376270 *T)
    sumR0  +=     239.628 * cos(3.16441455173  + 143.62530630140 *T)
    sumR0  +=     214.523 * cos(3.62480283040  + 278.25883401880 *T)
    sumR0  +=     246.198 * cos(1.01506302015  + 141.22580985640 *T)
    sumR0  +=     174.089 * cos(5.55011789988  + 567.82400073240 *T)
    sumR0  +=     163.934 * cos(2.10166491786  + 2.44768055480 *T)
    sumR0  +=     162.897 * cos(2.48946521653  + 4.19278569400 *T)
    sumR0  +=     193.455 * cos(1.58425287580  + 138.51749687070 *T)
    sumR0  +=     155.323 * cos(3.28425127954  + 31.01948863700 *T)
    sumR0  +=     182.469 * cos(2.45244890571  + 255.05546779820 *T)
    sumR0  +=     177.846 * cos(4.14773474853  + 10175.15251057320 *T)
    sumR0  +=     174.413 * cos(1.53042999914  + 329.83706636550 *T)
    sumR0  +=     137.649 * cos(3.34900537767  + 0.96320784650 *T)
    sumR0  +=     161.011 * cos(5.16655038482  + 211.81462272970 *T)
    sumR0  +=     113.473 * cos(4.96286007991  + 148.07872442630 *T)
    sumR0  +=     128.823 * cos(3.25521535448  + 24.11838995730 *T)
    sumR0  +=     107.363 * cos(3.26457701792  + 1059.38193018920 *T)
    sumR0  +=     122.732 * cos(5.39399536941  + 62.25142559510 *T)
    sumR0  +=     120.529 * cos(3.08050145518  + 184.72728735580 *T)
    sumR0  +=      99.356 * cos(1.92888554099  + 28.57180808220 *T)
    sumR0  +=      97.713 * cos(2.59474415429  + 6.59228213900 *T)
    sumR0  +=     124.095 * cos(3.11516750340  + 221.37585028530 *T)
    sumR0  +=     124.693 * cos(2.97042405451  + 251.43213107580 *T)
    sumR0  +=     114.252 * cos(0.25039919123  + 594.65070367540 *T)
    sumR0  +=     111.006 * cos(3.34276426767  + 180.27386923090 *T)
    sumR0  +=     120.939 * cos(1.92914010593  + 25.60286266560 *T)
    sumR0  +=     104.667 * cos(0.94883561775  + 395.57870223900 *T)
    sumR0  +=     109.779 * cos(5.43147520571  + 494.52687487340 *T)
    sumR0  +=      96.919 * cos(0.86184760695  + 1014.13534755060 *T)
    sumR0  +=      98.685 * cos(0.89577952710  + 488.58898404020 *T)
    sumR0  +=      88.968 * cos(4.78109764779  + 144.14657116320 *T)
    sumR0  +=     107.888 * cos(0.98700578434  + 1124.34166877000 *T)
    sumR0  +=      97.067 * cos(2.62667400276  + 291.70403072770 *T)
    sumR0  +=      75.131 * cos(5.88936524779  + 43.24084506850 *T)
    sumR0  +=      93.718 * cos(6.09873565184  + 526.72201967800 *T)
    sumR0  +=      94.822 * cos(0.20662943940  + 456.39383923560 *T)
    sumR0  +=      70.036 * cos(2.39683345663  + 426.59819087600 *T)
    sumR0  +=      77.187 * cos(4.21076753240  + 105.49227066360 *T)
    sumR0  +=      89.874 * cos(3.25100749923  + 258.02441321480 *T)
    sumR0  +=      69.133 * cos(4.93031154435  + 1028.36244155220 *T)
    sumR0  +=      90.657 * cos(1.69466970587  + 366.48562929500 *T)
    sumR0  +=      74.242 * cos(3.14479101276  + 82.85835341460 *T)
    sumR0  +=      57.995 * cos(0.86159785905  + 60.76695288680 *T)
    sumR0  +=      78.695 * cos(1.09307575550  + 700.66423920080 *T)
    sumR0  +=      57.230 * cos(0.81331949225  + 2.92076130680 *T)
    sumR0  +=      63.443 * cos(4.39590123005  + 149.56319713460 *T)
    sumR0  +=      55.698 * cos(3.89047249911  + 47.69426319340 *T)
    sumR0  +=      56.430 * cos(5.15003563302  + 0.52126486180 *T)
    sumR0  +=      56.174 * cos(5.42986960794  + 911.04257333200 *T)
    sumR0  +=      61.746 * cos(6.16453667559  + 1019.76442184310 *T)
    sumR0  +=      70.503 * cos(0.08077330612  + 40.58071619260 *T)
    sumR0  +=      74.677 * cos(4.85904499980  + 186.21176006410 *T)
    sumR0  +=      61.861 * cos(4.78702599861  + 11.04570026390 *T)
    sumR0  +=      61.135 * cos(0.83712253227  + 1022.73336725970 *T)
    sumR0  +=      61.268 * cos(5.70228826765  + 178.13500521680 *T)
    sumR0  +=      52.887 * cos(0.37458943972  + 27.08733537390 *T)
    sumR0  +=      56.722 * cos(3.52318112447  + 216.92243216040 *T)
    sumR0  +=      48.819 * cos(5.10789123481  + 64.95973858080 *T)
    sumR0  +=      63.290 * cos(4.39424910030  + 807.94979911340 *T)
    sumR0  +=      64.062 * cos(6.28297531806  + 7.11354700080 *T)
    sumR0  +=      46.356 * cos(1.34735469284  + 451.94042111070 *T)
    sumR0  +=      60.540 * cos(3.40316162416  + 294.67297614430 *T)
    sumR0  +=      46.900 * cos(0.17048203552  + 7.42236354150 *T)
    sumR0  +=      56.766 * cos(0.45048868231  + 140.00196957900 *T)
    sumR0  +=      55.887 * cos(1.06815733757  + 172.19711438360 *T)
    sumR0  +=      53.761 * cos(2.79644687008  + 328.35259365720 *T)
    sumR0  +=      43.828 * cos(6.04655696644  + 135.54855145410 *T)
    sumR0  +=      49.549 * cos(0.64106656292  + 41.05379694460 *T)
    sumR0  +=      53.960 * cos(2.91774494436  + 563.63121503840 *T)
    sumR0  +=      42.961 * cos(5.40175361431  + 487.36514376280 *T)
    sumR0  +=      51.508 * cos(0.09105540708  + 210.33015002140 *T)
    sumR0  +=      41.889 * cos(3.12343223889  + 29.22619938800 *T)
    sumR0  +=      47.655 * cos(3.90701760087  + 63.73589830340 *T)
    sumR0  +=      41.639 * cos(6.26847783513  + 32.71640966640 *T)
    sumR0  +=      41.429 * cos(4.45464156759  + 37.16982779130 *T)
    sumR0  +=      40.745 * cos(0.16043648294  + 79.23501669220 *T)
    sumR0  +=      48.205 * cos(1.84198373010  + 403.13419222450 *T)
    sumR0  +=      36.912 * cos(0.44771386183  + 30.05628079050 *T)
    sumR0  +=      47.762 * cos(0.88083849566  + 3302.47939106200 *T)
    sumR0  +=      39.465 * cos(3.50565484069  + 357.44566660120 *T)
    sumR0  +=      42.139 * cos(0.63375113663  + 343.21857259960 *T)
    sumR0  +=      41.275 * cos(1.36370496322  + 31.23193695810 *T)
    sumR0  +=      42.612 * cos(3.55270845713  + 38.65430049960 *T)
    sumR0  +=      38.931 * cos(5.26691753270  + 415.29185818120 *T)
    sumR0  +=      38.967 * cos(5.25866056502  + 386.98068252990 *T)
    sumR0  +=      33.734 * cos(5.24400184426  + 67.35923502580 *T)
    sumR0  +=      40.879 * cos(3.55292279438  + 331.32153907380 *T)
    sumR0  +=      38.768 * cos(1.12288359393  + 38.18121974760 *T)
    sumR0  +=      37.500 * cos(6.08687972441  + 35.42472265210 *T)
    sumR0  +=      38.831 * cos(4.67876780698  + 38.08485152800 *T)
    sumR0  +=      38.231 * cos(6.26491054328  + 389.94962794650 *T)
    sumR0  +=      29.976 * cos(4.45759985804  + 22.63391724900 *T)
    sumR0  +=      31.356 * cos(0.07746010366  + 12.53017297220 *T)
    sumR0  +=      26.341 * cos(4.59559782754  + 106.01353552540 *T)
    sumR0  +=      27.465 * cos(5.99541587890  + 206.18554843720 *T)
    sumR0  +=      25.152 * cos(4.49867760320  + 34.20088237470 *T)
    sumR0  +=      24.122 * cos(5.17089441917  + 129.91947716160 *T)
    sumR0  +=      28.997 * cos(3.64927210210  + 253.57099508990 *T)
    sumR0  +=      27.173 * cos(4.37944546475  + 142.14083359310 *T)
    sumR0  +=      30.634 * cos(1.59348806560  + 348.84764689210 *T)
    sumR0  +=      31.464 * cos(1.05065113524  + 100.38446123290 *T)
    sumR0  +=      24.056 * cos(1.02801635413  + 41.75637236020 *T)
    sumR0  +=      22.632 * cos(4.72511111292  + 81.37388070630 *T)
    sumR0  +=      21.942 * cos(3.48416607882  + 69.15252427480 *T)
    sumR0  +=      26.333 * cos(3.01556008632  + 365.00115658670 *T)
    sumR0  +=      22.355 * cos(3.92220883921  + 5.10780943070 *T)
    sumR0  +=      22.498 * cos(4.03487494425  + 19.12245511120 *T)
    sumR0  +=      22.885 * cos(1.58977064672  + 189.39315380180 *T)
    sumR0  +=      26.520 * cos(3.61427038042  + 367.97010200330 *T)
    sumR0  +=      25.496 * cos(2.43810518614  + 351.81659230870 *T)
    sumR0  +=      19.111 * cos(2.59694457001  + 2080.63082474060 *T)
    sumR0  +=      19.640 * cos(6.15701741238  + 35.21227433100 *T)
    sumR0  +=      25.688 * cos(2.00512719767  + 439.78275515400 *T)
    sumR0  +=      21.613 * cos(3.32354204724  + 119.50691634410 *T)
    sumR0  +=      25.389 * cos(4.74025836522  + 1474.67378837040 *T)
    sumR0  +=      18.107 * cos(5.35129342595  + 244.31858407500 *T)
    sumR0  +=      23.295 * cos(5.93767742799  + 316.39186965660 *T)
    sumR0  +=      22.087 * cos(4.81594755148  + 84.34282612290 *T)
    sumR0  +=      16.972 * cos(3.05105149940  + 220.41264243880 *T)
    sumR0  +=      20.022 * cos(4.99276451168  + 179.09821306330 *T)
    sumR0  +=      20.370 * cos(1.86508317889  + 171.23390653710 *T)
    sumR0  +=      19.426 * cos(2.04829970231  + 5.41662597140 *T)
    sumR0  +=      22.628 * cos(0.27205783433  + 666.72398925700 *T)
    sumR0  +=      19.072 * cos(3.70882976684  + 164.12035953630 *T)
    sumR0  +=      17.969 * cos(3.40425338171  + 69.36497259590 *T)
    sumR0  +=      18.716 * cos(0.90215956591  + 285.37238101960 *T)
    sumR0  +=      15.889 * cos(0.42011285882  + 697.74347789400 *T)
    sumR0  +=      14.988 * cos(3.08544843665  + 704.85702489480 *T)
    sumR0  +=      14.774 * cos(3.36129613309  + 274.06604832480 *T)
    sumR0  +=      15.972 * cos(1.82864185268  + 477.33083545520 *T)
    sumR0  +=      13.892 * cos(2.94161501165  + 38.39366806870 *T)
    sumR0  +=      13.922 * cos(2.85574364078  + 37.87240320690 *T)
    sumR0  +=      15.481 * cos(4.94982954853  + 101.86893394120 *T)
    sumR0  +=      17.571 * cos(5.82317632469  + 35.68535508300 *T)
    sumR0  +=      15.856 * cos(5.04973561582  + 36.90919536040 *T)
    sumR0  +=      16.414 * cos(3.63049397028  + 45.24658263860 *T)
    sumR0  +=      17.158 * cos(2.51251149482  + 20.60692781950 *T)
    sumR0  +=      12.941 * cos(3.03041555329  + 522.57741809380 *T)
    sumR0  +=      15.752 * cos(5.00292909214  + 247.23934538180 *T)
    sumR0  +=      12.679 * cos(0.20331109568  + 460.53844081980 *T)
    sumR0  +=      16.260 * cos(5.93480347217  + 815.06334611420 *T)
    sumR0  +=      12.903 * cos(3.51141502996  + 446.31134681820 *T)
    sumR0  +=      13.891 * cos(5.51064697670  + 31.54075349880 *T)
    sumR0  +=      13.668 * cos(5.45576135320  + 39.35687591520 *T)
    sumR0  +=      13.418 * cos(3.95805150079  + 290.21955801940 *T)
    sumR0  +=      15.368 * cos(2.45783892707  + 26.82670294300 *T)
    sumR0  +=      14.246 * cos(3.18588280921  + 401.64971951620 *T)
    sumR0  +=      12.222 * cos(4.94370170146  + 14.01464568050 *T)
    sumR0  +=      15.484 * cos(3.79703715637  + 404.61866493280 *T)
    sumR0  +=      13.427 * cos(3.79527836573  + 151.04766984290 *T)
    sumR0  +=      14.450 * cos(4.93940408761  + 120.99138905240 *T)
    sumR0  +=      14.331 * cos(4.71117327722  + 738.79727483860 *T)
    sumR0  +=      11.566 * cos(5.91003539239  + 536.80451209540 *T)
    sumR0  +=      15.578 * cos(2.91836788254  + 875.83029900100 *T)
    sumR0  +=      13.124 * cos(2.16056013419  + 152.53214255120 *T)
    sumR0  +=      11.744 * cos(2.94770244071  + 2.70831298570 *T)
    sumR0  +=      12.793 * cos(1.97868575679  + 1.37259812370 *T)
    sumR0  +=      12.969 * cos(0.00535826017  + 97.41551581630 *T)
    sumR0  +=      13.891 * cos(4.76435441820  + 0.26063243090 *T)
    sumR0  +=      13.729 * cos(2.32306473850  + 38.24491022240 *T)
    sumR0  +=      10.714 * cos(6.18129683877  + 115.88357962170 *T)
    sumR0  +=      11.610 * cos(4.61712859898  + 178.78939652260 *T)
    sumR0  +=      11.257 * cos(0.79300245838  + 42.32582133180 *T)
    sumR0  +=      14.500 * cos(5.44690193314  + 44.07092647100 *T)
    sumR0  +=      11.534 * cos(5.26580538005  + 160.93896579860 *T)
    sumR0  +=      13.355 * cos(5.20849186729  + 32.45577723550 *T)
    sumR0  +=      13.658 * cos(2.15687632802  + 476.43131808350 *T)
    sumR0  +=      13.782 * cos(3.47865209163  + 38.02116105320 *T)
    sumR0  +=      12.714 * cos(2.09462988855  + 20.49505323490 *T)
    sumR0  +=      13.257 * cos(5.15138524813  + 103.09277421860 *T)
    sumR0  +=       9.715 * cos(0.74597883480  + 918.15612033280 *T)
    sumR0  +=      10.340 * cos(5.38977407079  + 222.86032299360 *T)
    sumR0  +=      13.357 * cos(5.89635739027  + 748.09786996330 *T)
    sumR0  +=      12.632 * cos(1.20306997433  + 16.15350969460 *T)
    sumR0  +=      11.437 * cos(1.58444114292  + 495.49008271990 *T)
    sumR0  +=      11.424 * cos(4.74142930795  + 487.62577619370 *T)
    sumR0  +=       9.098 * cos(5.19932138822  + 118.02244363580 *T)
    sumR0  +=       9.336 * cos(0.97313630925  + 662.53120356300 *T)
    sumR0  +=       9.827 * cos(4.48170250645  + 505.78502345840 *T)
    sumR0  +=       8.585 * cos(0.20375451897  + 944.98282327580 *T)
    sumR0  +=       8.875 * cos(5.53111742265  + 17.52610781830 *T)
    sumR0  +=       9.957 * cos(4.03258125243  + 169.53698550770 *T)
    sumR0  +=      11.506 * cos(3.11649121817  + 17.63798240290 *T)
    sumR0  +=       9.818 * cos(5.20376439002  + 1.59634729290 *T)
    sumR0  +=      10.160 * cos(3.74441320429  + 457.61767951300 *T)
    sumR0  +=       8.661 * cos(0.31247523804  + 1440.73353842660 *T)
    sumR0  +=       8.496 * cos(1.06445636872  + 55.77101804070 *T)
    sumR0  +=      11.162 * cos(1.92907800408  + 564.85505531580 *T)
    sumR0  +=       8.057 * cos(0.31116345866  + 377.41945497430 *T)
    sumR0  +=       9.851 * cos(4.23328578127  + 418.26080359780 *T)
    sumR0  +=       7.938 * cos(2.40417397694  + 488.37653571910 *T)
    sumR0  +=       9.894 * cos(0.63707319139  + 183.76407950930 *T)
    sumR0  +=       9.913 * cos(3.94049519088  + 441.26722786230 *T)
    sumR0  +=       7.867 * cos(3.87469522964  + 494.73932319450 *T)
    sumR0  +=       7.589 * cos(3.15909316566  + 416.77633088950 *T)
    sumR0  +=       8.496 * cos(5.38968698704  + 104.00779795530 *T)
    sumR0  +=       9.716 * cos(3.06038536864  + 166.56804009110 *T)
    sumR0  +=       9.377 * cos(0.56416645296  + 673.31627139600 *T)
    sumR0  +=       8.771 * cos(5.24534141981  + 1057.89745748090 *T)
    sumR0  +=       7.990 * cos(1.55726966638  + 59.28248017850 *T)
    sumR0  +=       9.090 * cos(4.32953439022  + 29.74746424980 *T)
    sumR0  +=       9.667 * cos(5.89033222679  + 358.40887444770 *T)
    sumR0  +=       7.209 * cos(2.29464803358  + 79.18683258240 *T)
    sumR0  +=       8.062 * cos(0.44458003524  + 19.01058052660 *T)
    sumR0  +=       8.254 * cos(3.47304582051  + 156.15547927360 *T)
    sumR0  +=       9.804 * cos(6.06393995615  + 784.74643289280 *T)
    sumR0  +=       8.516 * cos(5.99060386955  + 180.79513409270 *T)
    sumR0  +=       8.090 * cos(1.38588221442  + 1654.03263386460 *T)
    sumR0  +=       9.074 * cos(4.03971490460  + 1017.05610885740 *T)
    sumR0  +=       6.908 * cos(1.41919832926  + 178.34745353790 *T)
    sumR0  +=       8.230 * cos(2.53750470473  + 518.38463239980 *T)
    sumR0  +=       8.594 * cos(5.29104206063  + 457.87831194390 *T)
    sumR0  +=       6.769 * cos(5.43380191356  + 171.98466606250 *T)
    sumR0  +=       8.571 * cos(0.35876828441  + 636.66770846650 *T)
    sumR0  +=       8.995 * cos(1.36992508507  + 6209.77872413240 *T)
    sumR0  +=       6.641 * cos(2.92327140872  + 0.04818410980 *T)
    sumR0  +=       9.278 * cos(3.80308677009  + 25558.21217647960 *T)
    sumR0  +=       6.567 * cos(4.01934954352  + 0.11187458460 *T)
    sumR0  +=       6.441 * cos(4.28250687347  + 36.12729806770 *T)
    sumR0  +=       7.257 * cos(4.09776235307  + 326.86812094890 *T)
    sumR0  +=       8.384 * cos(5.49363770202  + 532.61172640140 *T)
    sumR0  +=       7.471 * cos(4.62144262894  + 526.98265210890 *T)
    sumR0  +=       7.500 * cos(0.61545750834  + 485.92885516430 *T)
    sumR0  +=       7.716 * cos(1.04880632264  + 525.23754696970 *T)
    sumR0  +=       8.504 * cos(2.79350586429  + 10139.98842035200 *T)
    sumR0  +=       7.466 * cos(5.07942174095  + 157.63995198190 *T)
    sumR0  +=       7.186 * cos(6.22833818429  + 77.22927912210 *T)
    sumR0  +=       7.784 * cos(1.89308880453  + 984.60033162190 *T)
    sumR0  +=       6.513 * cos(0.07498932215  + 79.88940799800 *T)
    sumR0  +=       6.077 * cos(2.96673519667  + 36.69674703930 *T)
    sumR0  +=       7.706 * cos(5.70632580790  + 209.10630974400 *T)
    sumR0  +=       7.265 * cos(4.94483532589  + 131.40394986990 *T)
    sumR0  +=       6.984 * cos(2.53239305821  + 497.18700374930 *T)
    sumR0  +=       7.824 * cos(2.31462643851  + 513.07988101300 *T)
    sumR0  +=       7.175 * cos(3.69203633127  + 524.01370669230 *T)
    sumR0  +=       6.855 * cos(0.14076801572  + 283.62727588040 *T)
    sumR0  +=       6.922 * cos(3.36515011915  + 438.29828244570 *T)
    sumR0  +=       7.349 * cos(3.50406958122  + 500.15594916590 *T)
    sumR0  +=       6.301 * cos(0.14776691217  + 608.87779767700 *T)
    sumR0  +=       5.892 * cos(4.24403528888  + 4.66586644600 *T)
    sumR0  +=       7.613 * cos(5.14905171677  + 259.50888592310 *T)
    sumR0  +=       7.128 * cos(5.92696788834  + 482.95990974770 *T)
    sumR0  +=       6.829 * cos(1.01745137848  + 1543.82631264520 *T)
    sumR0  +=       5.981 * cos(4.79954091087  + 215.43795945210 *T)
    sumR0  +=       5.526 * cos(2.34003154732  + 65.22037101170 *T)
    sumR0  +=       6.817 * cos(6.12162829690  + 395.05743737720 *T)
    sumR0  +=       5.369 * cos(3.76855960849  + 52099.54021187280 *T)
    sumR0  +=       5.776 * cos(5.61434462641  + 987.56927703850 *T)
    sumR0  +=       7.523 * cos(5.60432148128  + 2810.92146160520 *T)
    sumR0  +=       7.329 * cos(3.76815551582  + 1512.80682400820 *T)
    sumR0  +=       5.616 * cos(2.13872867116  + 145.63104387150 *T)
    sumR0  +=       5.258 * cos(0.30850836910  + 36.60037881970 *T)
    sumR0  +=       5.688 * cos(1.82274388581  + 1227.43444298860 *T)
    sumR0  +=       5.658 * cos(2.35049199704  + 5.62907429250 *T)
    sumR0  +=       6.135 * cos(4.23390561816  + 496.01134758170 *T)
    sumR0  +=       5.128 * cos(2.89050864873  + 313.68355667090 *T)
    sumR0  +=       6.472 * cos(3.49494191669  + 552.69738935910 *T)
    sumR0  +=       4.983 * cos(3.91958511552  + 10135.53500222710 *T)
    sumR0  +=       5.217 * cos(0.40052635702  + 319.31263096340 *T)
    sumR0  +=       4.952 * cos(1.42482088612  + 49.17873590170 *T)
    sumR0  +=       5.964 * cos(5.70758449643  + 309.79958751760 *T)
    sumR0  +=       5.091 * cos(6.00974510144  + 1409.71404978960 *T)
    sumR0  +=       5.205 * cos(5.50271334510  + 238.90195810360 *T)
    sumR0  +=       4.800 * cos(1.13450310670  + 134.06407874580 *T)
    sumR0  +=       4.943 * cos(1.43051344597  + 422.40540518200 *T)
    sumR0  +=       5.604 * cos(2.05669305961  + 207.36120460480 *T)
    sumR0  +=       6.310 * cos(5.22966882627  + 139.74133714810 *T)
    sumR0  +=       4.772 * cos(3.06668713747  + 464.73122651380 *T)
    sumR0  +=       4.919 * cos(3.57280542629  + 52175.80628314840 *T)
    sumR0  +=       4.762 * cos(5.90654311203  + 838.96928775040 *T)
    sumR0  +=       4.848 * cos(0.77467099227  + 1.69692102940 *T)
    sumR0  +=       5.694 * cos(0.77313415569  + 709.96483432550 *T)
    sumR0  +=       5.455 * cos(0.90289242792  + 208.84567731310 *T)
    sumR0  +=       4.901 * cos(3.79986913631  + 15.49911838880 *T)
    sumR0  +=       4.772 * cos(0.15755140037  + 39.50563376150 *T)
    sumR0  +=       5.673 * cos(2.68359159067  + 1127.26243007680 *T)
    sumR0  +=       5.477 * cos(0.53123497431  + 113.87784205160 *T)
    sumR0  +=       5.077 * cos(1.59268428609  + 1547.97091422940 *T)
    sumR0  +=       4.981 * cos(1.44584050478  + 1.27202438720 *T)
    sumR0  +=       5.813 * cos(5.85024085408  + 57.25549074900 *T)
    sumR0  +=       5.520 * cos(5.06396698257  + 421.22974901440 *T)
    sumR0  +=       5.938 * cos(0.96886308551  + 6280.10690457480 *T)
    sumR0  +=       5.206 * cos(3.58003819370  + 474.94684537520 *T)
    sumR0  +=       5.256 * cos(0.61005270999  + 95.97922721780 *T)
    sumR0  +=       5.531 * cos(5.28764137194  + 36.76043751410 *T)
    sumR0  +=       6.158 * cos(5.73176703797  + 711.44930703380 *T)
    sumR0  +=       5.003 * cos(2.19048397989  + 501.64042187420 *T)
    sumR0  +=       5.150 * cos(5.58407480282  + 26049.77010593640 *T)
    sumR0  +=       5.138 * cos(4.55234158942  + 670.91677495100 *T)
    sumR0  +=       5.609 * cos(4.37272759780  + 52.80207262410 *T)
    sumR0  +=       5.636 * cos(2.39183054397  + 10210.31660079440 *T)
    sumR0  +=       4.512 * cos(2.59978208967  + 1234.54798998940 *T)
    sumR0  +=       5.412 * cos(4.58813638089  + 179.61947792510 *T)
    sumR0  +=       4.314 * cos(3.38846714337  + 142.66209845490 *T)
    sumR0  +=       4.708 * cos(5.23537414423  + 3.62333672240 *T)
    sumR0  +=       4.471 * cos(3.94378336812  + 12566.15169998280 *T)
    sumR0  +=       5.296 * cos(1.12249063176  + 134.11226285560 *T)
    sumR0  +=       4.188 * cos(2.52490407427  + 6205.32530600750 *T)
    sumR0  +=       4.645 * cos(1.90644271528  + 13324.31667116140 *T)
    sumR0  +=       4.502 * cos(2.01956920977  + 315.16802937920 *T)
    sumR0  +=       5.346 * cos(2.94804816223  + 353.04043258610 *T)
    sumR0  +=       4.177 * cos(2.09489065926  + 803.75701341940 *T)
    sumR0  +=       5.296 * cos(3.88249567974  + 2118.76386037840 *T)
    sumR0  +=       5.325 * cos(4.28221258353  + 477.91579079180 *T)
    sumR0  +=       5.519 * cos(0.09960891963  + 600.01914553700 *T)
    sumR0  +=       5.169 * cos(0.59948596687  + 6.90109867970 *T)
    sumR0  +=       4.179 * cos(0.14619703083  + 6644.57629047010 *T)
    sumR0  +=       4.490 * cos(1.07042724999  + 52139.15772021889 *T)
    sumR0  +=       3.970 * cos(6.13227798578  + 1553.90880506260 *T)
    sumR0  +=       3.970 * cos(4.69887237362  + 91.78644152380 *T)
    sumR0  +=       4.234 * cos(0.14478458924  + 65.87476231750 *T)
    sumR0  +=       5.183 * cos(3.52837189306  + 110.20632121940 *T)
    sumR0  +=       5.259 * cos(6.20809827528  + 142.71028256470 *T)
    sumR0  +=       3.869 * cos(5.25125030487  + 1558.05340664680 *T)
    sumR0  +=       4.457 * cos(2.10248126544  + 487.10451133190 *T)
    sumR0  +=       4.890 * cos(1.83606790269  + 46.51860702580 *T)
    sumR0  +=       3.875 * cos(5.60269278935  + 385.49620982160 *T)
    sumR0  +=       3.826 * cos(1.30946706974  + 2176.61005195840 *T)
    sumR0  +=       4.591 * cos(4.84657580441  + 1337.64076420800 *T)
    sumR0  +=       5.111 * cos(1.18808079775  + 981.63138620530 *T)
    sumR0  +=       4.709 * cos(1.40878215308  + 52213.93931878620 *T)
    sumR0  +=       3.891 * cos(5.43661875415  + 154.67100656530 *T)
    sumR0  +=       4.145 * cos(4.32505910718  + 363.51668387840 *T)
    sumR0  +=       4.441 * cos(3.50158424570  + 187.69623277240 *T)
    sumR0  +=       3.703 * cos(2.48768949613  + 67.88049988760 *T)
    sumR0  +=       4.094 * cos(1.42347047260  + 310.71461125430 *T)
    sumR0  +=       3.681 * cos(5.70552661143  + 491.66980404140 *T)
    sumR0  +=       4.787 * cos(3.65822147476  + 589.34595228860 *T)
    sumR0  +=       4.020 * cos(5.45643059988  + 6641.60734505350 *T)
    sumR0  +=       3.656 * cos(0.57790726599  + 491.44605487220 *T)
    sumR0  +=       4.288 * cos(3.35265955957  + 203.21660302060 *T)
    sumR0  +=       3.843 * cos(4.61508898119  + 1025.70231267630 *T)
    sumR0  +=       3.767 * cos(0.05292047125  + 320.27583880990 *T)
    sumR0  +=       4.632 * cos(0.82011276589  + 3265.83082813250 *T)
    sumR0  +=       4.609 * cos(5.25443775917  + 296.15744885260 *T)
    sumR0  +=       4.555 * cos(5.30391170376  + 26013.12154300690 *T)
    sumR0  +=       3.556 * cos(4.80267245336  + 224.34479570190 *T)
    sumR0  +=       4.859 * cos(5.52756242256  + 487.41332787260 *T)
    sumR0  +=       3.626 * cos(1.44624342082  + 70.84944530420 *T)
    sumR0  +=       4.302 * cos(1.60914544159  + 12529.50313705330 *T)
    sumR0  +=       3.493 * cos(4.75315651083  + 12489.88562870720 *T)
    sumR0  +=       3.722 * cos(0.27433061822  + 949.43624140070 *T)
    sumR0  +=       4.234 * cos(5.25112033465  + 194.28851491140 *T)
    sumR0  +=       3.451 * cos(2.97409317928  + 499.63468430410 *T)
    sumR0  +=       4.796 * cos(6.21059766333  + 491.81856188770 *T)
    sumR0  +=       3.639 * cos(1.25605018211  + 2603.20824283440 *T)
    sumR0  +=       4.646 * cos(5.71392540144  + 321.76031151820 *T)
    sumR0  +=       3.702 * cos(2.08952561657  + 491.03666459500 *T)
    sumR0  +=       3.672 * cos(2.87489628704  + 497.49582029000 *T)
    sumR0  +=       3.965 * cos(1.05484988240  + 75.74480641380 *T)
    sumR0  +=       3.416 * cos(0.68584132933  + 305.08553696180 *T)
    sumR0  +=       4.513 * cos(4.38927002490  + 425.11371816770 *T)
    sumR0  +=       3.853 * cos(0.61321572401  + 12526.53419163670 *T)
    sumR0  +=       3.788 * cos(3.32221995840  + 3140.01275492980 *T)
    sumR0  +=       3.781 * cos(5.58125317044  + 1652.54816115630 *T)
    sumR0  +=       3.903 * cos(5.31609723466  + 408.17831118040 *T)
    sumR0  +=       3.945 * cos(3.60558877407  + 1589.07289528380 *T)
    sumR0  +=       4.084 * cos(0.83813879869  + 52.36012963940 *T)
    sumR0  +=       4.084 * cos(3.50290269471  + 23.90594163620 *T)
    sumR0  +=       3.694 * cos(1.03218855688  + 481.47543703940 *T)
    sumR0  +=       3.636 * cos(5.31068934607  + 141.48644228730 *T)
    sumR0  +=       3.345 * cos(3.94392179077  + 20389.92252949249 *T)
    sumR0  +=       4.639 * cos(6.24618220184  + 821.39499582230 *T)
    sumR0  +=       3.934 * cos(0.26992234338  + 1655.51710657290 *T)
    sumR0  +=       4.431 * cos(2.48647437800  + 549.72844394250 *T)
    sumR0  +=       4.168 * cos(5.39993754642  + 236.50246165860 *T)
    sumR0  +=       4.020 * cos(0.07393243012  + 52136.18877480229 *T)
    sumR0  +=       4.055 * cos(1.34004288978  + 1054.92851206430 *T)
    sumR0  +=       3.275 * cos(0.98533127454  + 1344.75431120880 *T)
    sumR0  +=       3.213 * cos(2.97105590703  + 20386.95358407589 *T)
    sumR0  +=       4.428 * cos(0.06728869735  + 491.29729702590 *T)
    sumR0  +=       4.063 * cos(0.06192838570  + 6168.67674307800 *T)
    sumR0  +=       3.804 * cos(5.34897033476  + 523.75307426140 *T)
    sumR0  +=       3.917 * cos(5.67905809516  + 1131.19458333990 *T)
    sumR0  +=       3.833 * cos(0.87811168267  + 52.69019803950 *T)
    sumR0  +=       4.020 * cos(2.69209723289  + 1439.46151403940 *T)
    sumR0  +=       4.373 * cos(1.86209663434  + 73.55775828990 *T)
    sumR0  +=       3.159 * cos(1.04693380342  + 703.37255218650 *T)
    sumR0  +=       3.116 * cos(5.20159166840  + 449.23210812500 *T)
    sumR0  +=       3.258 * cos(4.65131076542  + 696.25900518570 *T)
    sumR0  +=       3.427 * cos(0.27003884843  + 2389.90914739640 *T)
    sumR0  +=       4.349 * cos(0.07531141761  + 20426.57109242200 *T)
    sumR0  +=       3.383 * cos(5.61838426864  + 699.22795060230 *T)
    sumR0  +=       3.305 * cos(1.41666877290  + 562.14674233010 *T)
    sumR0  +=       3.297 * cos(5.46677712589  + 1442.21801113490 *T)
    sumR0  +=       3.277 * cos(2.71815883511  + 980.14691349700 *T)
    sumR0  +=       3.171 * cos(4.49510885866  + 1439.24906571830 *T)
    sumR0  +=       4.175 * cos(4.24327707038  + 381.61224066830 *T)
    sumR0  +=       3.155 * cos(3.40776789576  + 39.72938293070 *T)
    sumR0  +=       4.112 * cos(0.90309319273  + 1087.69310584050 *T)
    sumR0  +=       3.350 * cos(5.27474671017  + 80.71948940050 *T)
    sumR0  +=       3.725 * cos(1.52448613082  + 1058.10990580200 *T)
    sumR0  +=       3.650 * cos(3.59798316565  + 192.80404220310 *T)
    sumR0  +=       3.837 * cos(1.48519528444  + 10098.88643929760 *T)
    sumR0  +=       2.959 * cos(1.23012121982  + 2500.11546861580 *T)
    sumR0  +=       3.330 * cos(6.12470287875  + 10172.18356515660 *T)
    sumR0  +=       3.361 * cos(4.31837298696  + 492.07919431860 *T)
    sumR0  +=       3.288 * cos(3.14692435376  + 347.36317418380 *T)
    sumR0  +=       2.992 * cos(5.01304660316  + 175.21424391000 *T)
    sumR0  +=       3.294 * cos(2.52694043155  + 1692.16566950240 *T)
    sumR0  +=       2.984 * cos(1.81780659890  + 175.11787569040 *T)
    sumR0  +=       3.013 * cos(0.92957285991  + 1515.77576942480 *T)
    sumR0  +=       3.863 * cos(5.46044928570  + 332.80601178210 *T)
    sumR0  +=       3.403 * cos(1.10932483984  + 987.30864460760 *T)
    sumR0  +=       3.312 * cos(0.67710158807  + 977.48678462110 *T)
    sumR0  +=       3.030 * cos(1.77996261146  + 156489.28581380739 *T)
    sumR0  +=       3.605 * cos(4.89955108152  + 1043.88281180040 *T)
    sumR0  +=       2.937 * cos(0.60469671230  + 990.22940591440 *T)
    sumR0  +=       3.276 * cos(4.26765608367  + 1189.30140735080 *T)
    sumR0  +=       2.966 * cos(5.29808076929  + 31.98269648350 *T)
    sumR0  +=       2.994 * cos(2.58599359402  + 178.08682110700 *T)
    sumR0  +=       3.905 * cos(1.87748122254  + 1158.28191871380 *T)
    sumR0  +=       3.110 * cos(3.09203517638  + 235.93301268700 *T)
    sumR0  +=       3.313 * cos(2.70308129756  + 604.47256366190 *T)
    sumR0  +=       3.276 * cos(1.24440460327  + 874.65464283340 *T)
    sumR0  +=       3.276 * cos(5.58544609667  + 950.92071410900 *T)
    sumR0  +=       3.746 * cos(0.33859914037  + 913.96333463880 *T)
    sumR0  +=       3.552 * cos(3.07180917863  + 240.38643081190 *T)
    sumR0  +=       2.885 * cos(6.01130634957  + 1097.51496582700 *T)
    sumR0  +=       3.643 * cos(5.11977873355  + 452.20105354160 *T)
    sumR0  +=       2.768 * cos(4.38396269009  + 391.43410065480 *T)
    sumR0  +=       2.776 * cos(5.01821594830  + 8.90683624980 *T)
    sumR0  +=       2.990 * cos(5.62911695857  + 140.65636088480 *T)
    sumR0  +=       2.761 * cos(4.05534163807  + 6283.07584999140 *T)
    sumR0  +=       3.226 * cos(4.76711354367  + 6241.97386893700 *T)
    sumR0  +=       3.748 * cos(4.84009347869  + 341.73409989130 *T)
    sumR0  +=       2.752 * cos(4.53621078796  + 6206.80977871580 *T)
    sumR0  +=       3.847 * cos(2.40982343643  + 26086.41866886590 *T)
    sumR0  +=       2.727 * cos(3.28234198801  + 483.48117460950 *T)
    sumR0  +=       2.884 * cos(4.05452029151  + 1.22384027740 *T)
    sumR0  +=       2.702 * cos(3.72061244391  + 946.46729598410 *T)
    sumR0  +=       2.723 * cos(4.37517047024  + 15.19030184810 *T)
    sumR0  +=       2.847 * cos(5.22951186538  + 661.04673085470 *T)
    sumR0  +=       2.680 * cos(4.19379121323  + 13.18456427800 *T)
    sumR0  +=       3.269 * cos(0.43119778520  + 496.97455542820 *T)
    sumR0  +=       3.489 * cos(3.82213189319  + 625.99451521810 *T)
    sumR0  +=       3.757 * cos(3.88223872147  + 495.70253104100 *T)
    sumR0  +=       2.872 * cos(5.00345974886  + 252.08652238160 *T)
    sumR0  +=       3.742 * cos(2.03372773652  + 8.59801970910 *T)
    sumR0  +=       3.172 * cos(1.11135762382  + 260.99335863140 *T)
    sumR0  +=       3.341 * cos(2.91360557418  + 304.23420369990 *T)
    sumR0  +=       2.915 * cos(2.63627684599  + 6681.22485339960 *T)
    sumR0  +=       2.915 * cos(1.43773625890  + 6604.95878212400 *T)
    sumR0  +=       2.629 * cos(2.09824407450  + 2713.41456405380 *T)
    sumR0  +=       2.901 * cos(3.33924800230  + 515.46387109300 *T)
    sumR0  +=       2.803 * cos(1.23584865903  + 6643.09181776180 *T)
    sumR0  +=       3.045 * cos(3.33515866438  + 921.07688163960 *T)
    sumR0  +=       2.699 * cos(5.42597794650  + 925.26966733360 *T)
    sumR0  +=       2.808 * cos(5.77870303237  + 1024.21783996800 *T)
    sumR0  +=       3.028 * cos(3.75501312393  + 511.59540830470 *T)
    sumR0  +=       3.090 * cos(2.49453093252  + 14.66903698630 *T)
    sumR0  +=       2.913 * cos(4.83296711477  + 515.93695184500 *T)
    sumR0  +=       3.139 * cos(5.99134254710  + 570.74476203920 *T)
    sumR0  +=       2.752 * cos(3.08268180744  + 853.19638175200 *T)
    sumR0  +=       2.779 * cos(3.74527347899  + 494.00561001160 *T)
    sumR0  +=       2.643 * cos(1.99093797444  + 470.21728845440 *T)
    sumR0  +=       2.763 * cos(4.01095972177  + 448.97147569410 *T)
    sumR0  +=       2.643 * cos(5.24970673655  + 249.94765836750 *T)
    sumR0  +=       3.426 * cos(4.73955481174  + 1050.99635880120 *T)
    sumR0  +=       2.573 * cos(2.01267457287  + 1514.29129671650 *T)
    sumR0  +=       2.633 * cos(1.63640090603  + 170.71264167530 *T)
    sumR0  +=       3.034 * cos(4.48979734509  + 560.71045373160 *T)
    sumR0  +=       3.025 * cos(5.51446170055  + 369.45457471160 *T)
    sumR0  +=       3.095 * cos(4.01459691667  + 1615.89959822680 *T)
    sumR0  +=       2.490 * cos(0.15301603966  + 78187.44335344699 *T)
    sumR0  +=       2.589 * cos(0.79196093766  + 1228.91891569690 *T)
    sumR0  +=       3.143 * cos(5.33170343283  + 1542.34183993690 *T)
    sumR0  +=       3.138 * cos(4.50785484172  + 461.76228109720 *T)
    sumR0  +=       2.812 * cos(3.74246594120  + 2.00573757010 *T)
    sumR0  +=       3.062 * cos(4.88018345098  + 227.96813242430 *T)
    sumR0  +=       2.553 * cos(4.85437812287  + 488.84961647110 *T)
    sumR0  +=       2.971 * cos(1.27359129352  + 530.91480537200 *T)
    sumR0  +=       2.646 * cos(3.64828423565  + 335.77495719870 *T)
    sumR0  +=       3.329 * cos(2.71693827722  + 171.02145821600 *T)
    sumR0  +=       2.648 * cos(0.60243117586  + 70.58881287330 *T)
    sumR0  +=       3.061 * cos(5.05044834864  + 378.64329525170 *T)
    sumR0  +=       2.738 * cos(4.75405645015  + 151.26011816400 *T)
    sumR0  +=       2.728 * cos(5.89052930055  + 213.95348674380 *T)
    sumR0  +=       3.411 * cos(2.24137878065  + 734.45573129830 *T)
    sumR0  +=       2.623 * cos(0.54340876464  + 1586.10394986720 *T)
    sumR0  +=       3.169 * cos(5.84871429991  + 1049.51188609290 *T)
    sumR0  +=       2.430 * cos(2.34595493263  + 450.45594840240 *T)
    sumR0  +=       2.907 * cos(5.58085498481  + 597.57146498220 *T)
    sumR0  +=       3.300 * cos(0.94221473935  + 58.17051448570 *T)
    sumR0  +=       2.543 * cos(5.30426930256  + 419.48464387520 *T)
    sumR0  +=       3.175 * cos(2.32600231924  + 339.28641933650 *T)
    sumR0  +=       2.858 * cos(2.36621678719  + 32.50396134530 *T)
    sumR0  +=       2.712 * cos(5.79983621237  + 1587.58842257550 *T)
    sumR0  +=       3.340 * cos(1.36950315448  + 384.27236954420 *T)
    sumR0  +=       3.301 * cos(5.83023910521  + 51.77517430280 *T)
    sumR0  +=       2.415 * cos(0.69446923670  + 489.55219188670 *T)
    sumR0  +=       2.736 * cos(5.74320864965  + 1167.84314626940 *T)
    sumR0  +=       2.956 * cos(5.22962139507  + 199.85389872910 *T)
    sumR0  +=       3.262 * cos(0.01501002027  + 1545.31078535350 *T)
    sumR0  +=       2.506 * cos(4.84043333582  + 943.49835056750 *T)
    sumR0  +=       3.240 * cos(2.46676155925  + 1016.79547642650 *T)
    sumR0  +=       3.148 * cos(4.62079057738  + 233.53351624200 *T)
    sumR0  +=       2.327 * cos(4.10421417326  + 70.11573212130 *T)
    sumR0  +=       2.371 * cos(4.79963943424  + 271.14528701800 *T)
    sumR0  +=       3.006 * cos(3.66877796077  + 1476.15826107870 *T)
    sumR0  +=       2.537 * cos(5.66681769885  + 21.14944454070 *T)
    sumR0  +=       3.006 * cos(0.93048909480  + 21.97952594320 *T)
    sumR0  +=       3.033 * cos(0.67157488690  + 292.48592802040 *T)
    sumR0  +=       2.344 * cos(1.83547256266  + 492.30868898220 *T)
    sumR0  +=       3.117 * cos(2.76268894894  + 1473.18931566210 *T)
    sumR0  +=       2.323 * cos(2.88799980853  + 533.62311835770 *T)
    sumR0  +=       2.340 * cos(4.44862573253  + 490.80716993140 *T)
    sumR0  +=       2.511 * cos(0.99467349084  + 266.10116806210 *T)
    sumR0  +=       2.919 * cos(4.75889516601  + 1511.32235129990 *T)
    sumR0  +=       2.493 * cos(6.10541658597  + 1225.94997028030 *T)
    sumR0  +=       2.798 * cos(3.06162629894  + 419.74527630610 *T)
    sumR0  +=       2.691 * cos(3.20679023131  + 463.50738623640 *T)
    sumR0  +=       2.291 * cos(5.81534758547  + 246.97871295090 *T)
    sumR0  +=       2.319 * cos(6.05514281470  + 525.75881183150 *T)
    sumR0  +=       3.112 * cos(0.89712836583  + 314.90739694830 *T)
    sumR0  +=       3.085 * cos(5.84605938859  + 1192.22216865760 *T)
    sumR0  +=       2.897 * cos(0.54747024257  + 20350.30502114640 *T)
    sumR0  +=       3.067 * cos(2.22206306288  + 248.46318565920 *T)
    sumR0  +=       2.252 * cos(0.87483094907  + 61.02758531770 *T)
    sumR0  +=       2.392 * cos(3.62837597194  + 439.19779981740 *T)
    sumR0  +=       2.817 * cos(2.73562306571  + 16.67477455640 *T)
    sumR0  +=       2.379 * cos(6.17876088396  + 467.65198782060 *T)
    sumR0  +=       2.598 * cos(4.82643304253  + 384.58118608490 *T)
    sumR0  +=       2.718 * cos(1.01823841209  + 215.95922431390 *T)
    sumR0  +=       2.998 * cos(1.09755715300  + 1964.74724511890 *T)
    sumR0  +=       2.884 * cos(2.97813466834  + 383.09671337660 *T)
    sumR0  +=       2.231 * cos(4.48841493844  + 4.14460158420 *T)
    sumR0  +=       2.203 * cos(2.23336308907  + 481.26298871830 *T)
    sumR0  +=       2.260 * cos(2.35404913660  + 659.61044225620 *T)
    sumR0  +=       2.491 * cos(1.70236357070  + 445.34813897170 *T)
    sumR0  +=       3.041 * cos(5.55577674116  + 674.80074410430 *T)
    sumR0  +=       2.289 * cos(1.18497528002  + 1552.42433235430 *T)
    sumR0  +=       2.975 * cos(0.48272389481  + 1052.48083150950 *T)
    sumR0  +=       2.339 * cos(0.75318738767  + 478.81530816350 *T)
    sumR0  +=       3.011 * cos(0.16359500858  + 54.28654533240 *T)
    sumR0  +=       2.820 * cos(6.18522693724  + 556.51766803760 *T)
    sumR0  +=       2.266 * cos(5.91286000054  + 3.49021027840 *T)
    sumR0  +=       2.231 * cos(1.45038594906  + 196.50670080260 *T)

    # NEPTUNE Radius
    sumR1  = 0
    sumR1  +=  236338.502 * cos(0.70498011235  + 38.13303563780 *T)
    sumR1  +=   13220.279 * cos(3.32015499895  + 1.48447270830 *T)
    sumR1  +=    8621.863 * cos(6.21628951630  + 35.16409022120 *T)
    sumR1  +=    2701.740 * cos(1.88140666779  + 39.61750834610 *T)
    sumR1  +=    2153.150 * cos(5.16873840979  + 76.26607127560 *T)
    sumR1  +=    2154.735 * cos(2.09431198086  + 2.96894541660 *T)
    sumR1  +=    1463.924 * cos(1.18417031047  + 33.67961751290 *T)
    sumR1  +=    1603.165 * cos(0.00000000000  + 0.00000000000 *T)
    sumR1  +=    1135.773 * cos(3.91891199655  + 36.64856292950 *T)
    sumR1  +=     897.650 * cos(5.24122933533  + 388.46515523820 *T)
    sumR1  +=     789.908 * cos(0.53315484580  + 168.05251279940 *T)
    sumR1  +=     760.030 * cos(0.02051033644  + 182.27960680100 *T)
    sumR1  +=     607.183 * cos(1.07706500350  + 1021.24889455140 *T)
    sumR1  +=     571.622 * cos(3.40060785432  + 484.44438245600 *T)
    sumR1  +=     560.790 * cos(2.88685815667  + 498.67147645760 *T)
    sumR1  +=     490.190 * cos(3.46830928696  + 137.03302416240 *T)
    sumR1  +=     264.093 * cos(0.86220057976  + 4.45341812490 *T)
    sumR1  +=     270.526 * cos(3.27355867939  + 71.81265315070 *T)
    sumR1  +=     203.524 * cos(2.41820674409  + 32.19514480460 *T)
    sumR1  +=     155.438 * cos(0.36537064534  + 41.10198105440 *T)
    sumR1  +=     132.766 * cos(3.60157672619  + 9.56122755560 *T)
    sumR1  +=      93.626 * cos(0.66670888163  + 46.20979048510 *T)
    sumR1  +=      83.317 * cos(3.25992461673  + 98.89998852460 *T)
    sumR1  +=      72.205 * cos(4.47717435693  + 601.76425067620 *T)
    sumR1  +=      68.983 * cos(1.46326969479  + 74.78159856730 *T)
    sumR1  +=      86.953 * cos(5.77228651853  + 381.35160823740 *T)
    sumR1  +=      68.717 * cos(4.52563942435  + 70.32818044240 *T)
    sumR1  +=      64.724 * cos(3.85477388838  + 73.29712585900 *T)
    sumR1  +=      68.377 * cos(3.39509945953  + 108.46121608020 *T)
    sumR1  +=      53.375 * cos(5.43650770516  + 395.57870223900 *T)
    sumR1  +=      44.453 * cos(3.61409723545  + 2.44768055480 *T)
    sumR1  +=      41.243 * cos(4.73866592865  + 8.07675484730 *T)
    sumR1  +=      48.331 * cos(1.98568593981  + 175.16605980020 *T)
    sumR1  +=      41.744 * cos(4.94257598763  + 31.01948863700 *T)
    sumR1  +=      44.102 * cos(1.41744904844  + 1550.93985964600 *T)
    sumR1  +=      41.170 * cos(1.41999374753  + 490.07345674850 *T)
    sumR1  +=      41.099 * cos(4.86312637841  + 493.04240216510 *T)
    sumR1  +=      36.267 * cos(5.30764043577  + 312.19908396260 *T)
    sumR1  +=      36.284 * cos(0.38187812797  + 77.75054398390 *T)
    sumR1  +=      40.619 * cos(2.27237172464  + 529.69096509460 *T)
    sumR1  +=      32.360 * cos(5.91123007786  + 5.93789083320 *T)
    sumR1  +=      31.197 * cos(2.70549944134  + 1014.13534755060 *T)
    sumR1  +=      32.730 * cos(5.22147683115  + 41.05379694460 *T)
    sumR1  +=      36.079 * cos(4.87817494829  + 491.55792945680 *T)
    sumR1  +=      30.181 * cos(3.63273193845  + 30.71067209630 *T)
    sumR1  +=      29.991 * cos(3.30769367603  + 1028.36244155220 *T)
    sumR1  +=      27.048 * cos(1.77647060739  + 44.72531777680 *T)
    sumR1  +=      27.756 * cos(4.55583165091  + 7.11354700080 *T)
    sumR1  +=      27.475 * cos(0.97228280623  + 33.94024994380 *T)
    sumR1  +=      24.944 * cos(3.10083391185  + 144.14657116320 *T)
    sumR1  +=      25.958 * cos(2.99724758632  + 60.76695288680 *T)
    sumR1  +=      21.369 * cos(4.71270048898  + 278.25883401880 *T)
    sumR1  +=      21.283 * cos(0.68957829113  + 251.43213107580 *T)
    sumR1  +=      23.727 * cos(5.12044184469  + 176.65053250850 *T)
    sumR1  +=      21.392 * cos(0.86286397645  + 4.19278569400 *T)
    sumR1  +=      23.373 * cos(1.64955088447  + 173.68158709190 *T)
    sumR1  +=      24.163 * cos(3.56602004577  + 145.10977900970 *T)
    sumR1  +=      20.238 * cos(5.61479765982  + 24.11838995730 *T)
    sumR1  +=      26.958 * cos(4.14294870704  + 453.42489381900 *T)
    sumR1  +=      24.048 * cos(1.00718363213  + 213.29909543800 *T)
    sumR1  +=      18.322 * cos(1.98028683488  + 72.07328558160 *T)
    sumR1  +=      18.266 * cos(6.17260374467  + 189.39315380180 *T)
    sumR1  +=      19.201 * cos(4.65162168927  + 106.97674337190 *T)
    sumR1  +=      17.606 * cos(1.60307551767  + 62.25142559510 *T)
    sumR1  +=      16.545 * cos(1.69931816587  + 357.44566660120 *T)
    sumR1  +=      20.132 * cos(3.29520553529  + 114.39910691340 *T)
    sumR1  +=      15.425 * cos(4.38812302799  + 25.60286266560 *T)
    sumR1  +=      19.173 * cos(2.20014267311  + 343.21857259960 *T)
    sumR1  +=      15.077 * cos(3.66802659382  + 0.52126486180 *T)
    sumR1  +=      14.029 * cos(0.55336333290  + 129.91947716160 *T)
    sumR1  +=      13.361 * cos(5.85751083720  + 68.84370773410 *T)
    sumR1  +=      15.357 * cos(4.20731277007  + 567.82400073240 *T)
    sumR1  +=      12.746 * cos(3.52815836608  + 477.33083545520 *T)
    sumR1  +=      11.724 * cos(5.57647263460  + 31.23193695810 *T)
    sumR1  +=      11.533 * cos(0.89138506506  + 594.65070367540 *T)
    sumR1  +=      10.508 * cos(4.35552732772  + 32.71640966640 *T)
    sumR1  +=      10.826 * cos(5.21826226871  + 26.82670294300 *T)
    sumR1  +=      10.085 * cos(1.98102855874  + 40.58071619260 *T)
    sumR1  +=      10.518 * cos(5.27281360238  + 2.92076130680 *T)
    sumR1  +=       9.207 * cos(0.50092534158  + 64.95973858080 *T)
    sumR1  +=       9.231 * cos(0.68180977710  + 160.93896579860 *T)
    sumR1  +=       8.735 * cos(5.80657503476  + 6.59228213900 *T)
    sumR1  +=      10.114 * cos(4.51164596694  + 28.57180808220 *T)
    sumR1  +=      10.392 * cos(5.18877536013  + 42.58645376270 *T)
    sumR1  +=       9.873 * cos(3.76512158080  + 181.75834193920 *T)
    sumR1  +=       8.350 * cos(2.82449631025  + 43.24084506850 *T)
    sumR1  +=       9.838 * cos(1.49438763600  + 47.69426319340 *T)
    sumR1  +=       7.645 * cos(4.07503370297  + 389.94962794650 *T)
    sumR1  +=       8.004 * cos(2.78082277326  + 505.78502345840 *T)
    sumR1  +=       7.440 * cos(2.35731983047  + 11.04570026390 *T)
    sumR1  +=       7.342 * cos(1.62279119952  + 135.54855145410 *T)
    sumR1  +=       9.450 * cos(0.27241261915  + 426.59819087600 *T)
    sumR1  +=       7.192 * cos(0.82841201068  + 911.04257333200 *T)
    sumR1  +=       6.979 * cos(1.86753914872  + 206.18554843720 *T)
    sumR1  +=       6.874 * cos(0.83802906828  + 82.85835341460 *T)
    sumR1  +=       7.897 * cos(1.86554246391  + 38.65430049960 *T)
    sumR1  +=       6.729 * cos(3.98338053636  + 12.53017297220 *T)
    sumR1  +=       6.357 * cos(0.90093123522  + 487.36514376280 *T)
    sumR1  +=       6.720 * cos(1.33936040700  + 220.41264243880 *T)
    sumR1  +=       7.695 * cos(5.13312500855  + 23.90594163620 *T)
    sumR1  +=       7.059 * cos(5.99832463494  + 639.89728631400 *T)
    sumR1  +=       8.302 * cos(3.85960902325  + 37.61177077600 *T)
    sumR1  +=       6.412 * cos(2.41743702679  + 1059.38193018920 *T)
    sumR1  +=       6.751 * cos(1.96860894470  + 45.24658263860 *T)
    sumR1  +=       6.431 * cos(4.07813226506  + 35.68535508300 *T)
    sumR1  +=       5.517 * cos(3.81325790890  + 815.06334611420 *T)
    sumR1  +=       5.562 * cos(0.41619602150  + 563.63121503840 *T)
    sumR1  +=       6.115 * cos(2.10934525342  + 697.74347789400 *T)
    sumR1  +=       6.216 * cos(4.79301628209  + 143.62530630140 *T)
    sumR1  +=       5.346 * cos(3.13071964722  + 386.98068252990 *T)
    sumR1  +=       5.245 * cos(6.06245070403  + 171.23390653710 *T)
    sumR1  +=       5.129 * cos(0.79394555531  + 179.09821306330 *T)
    sumR1  +=       5.168 * cos(4.73765992885  + 522.57741809380 *T)
    sumR1  +=       6.422 * cos(0.64684316894  + 350.33211960040 *T)
    sumR1  +=       5.006 * cos(2.37645082899  + 77.22927912210 *T)
    sumR1  +=       5.005 * cos(4.70632786971  + 460.53844081980 *T)
    sumR1  +=       5.167 * cos(5.20246616570  + 446.31134681820 *T)
    sumR1  +=       5.119 * cos(2.17338058771  + 494.73932319450 *T)
    sumR1  +=       5.025 * cos(4.21265519856  + 536.80451209540 *T)
    sumR1  +=       4.722 * cos(6.22814313946  + 63.73589830340 *T)
    sumR1  +=       5.125 * cos(5.38138329172  + 179.31066138440 *T)
    sumR1  +=       4.918 * cos(4.09031782903  + 488.37653571910 *T)
    sumR1  +=       4.652 * cos(5.10765073368  + 274.06604832480 *T)
    sumR1  +=       4.711 * cos(5.56542374115  + 42.32582133180 *T)
    sumR1  +=       4.459 * cos(1.30784829830  + 69.36497259590 *T)
    sumR1  +=       5.485 * cos(3.88088464259  + 218.40690486870 *T)
    sumR1  +=       4.416 * cos(3.05353893868  + 27.08733537390 *T)
    sumR1  +=       4.559 * cos(4.92224120952  + 285.37238101960 *T)
    sumR1  +=       4.393 * cos(4.18047835584  + 5.41662597140 *T)
    sumR1  +=       4.687 * cos(2.21401153210  + 1029.84691426050 *T)
    sumR1  +=       4.644 * cos(1.87902594973  + 1433.61999142580 *T)
    sumR1  +=       5.639 * cos(3.05596737234  + 983.11585891360 *T)
    sumR1  +=       6.045 * cos(5.68817982786  + 351.81659230870 *T)
    sumR1  +=       4.430 * cos(3.37768805833  + 377.41945497430 *T)
    sumR1  +=       4.683 * cos(2.14346624864  + 97.41551581630 *T)
    sumR1  +=       5.845 * cos(4.62301099402  + 1024.21783996800 *T)
    sumR1  +=       4.536 * cos(2.45860473853  + 496.01134758170 *T)
    sumR1  +=       4.398 * cos(5.65312496227  + 3.93215326310 *T)
    sumR1  +=       4.287 * cos(0.66340266603  + 1012.65087484230 *T)
    sumR1  +=       4.086 * cos(0.14551174994  + 385.28376150050 *T)
    sumR1  +=       4.029 * cos(5.98399329775  + 178.34745353790 *T)
    sumR1  +=       4.276 * cos(3.68205082970  + 348.84764689210 *T)
    sumR1  +=       5.257 * cos(3.75263242432  + 379.86713552910 *T)
    sumR1  +=       4.012 * cos(0.42559540783  + 104313.47953065898 *T)
    sumR1  +=       4.025 * cos(2.40645188238  + 84.34282612290 *T)
    sumR1  +=       3.957 * cos(0.86846121055  + 171.98466606250 *T)
    sumR1  +=       3.961 * cos(3.04953080906  + 1017.31674128830 *T)
    sumR1  +=       5.559 * cos(0.77714806229  + 1447.84708542740 *T)
    sumR1  +=       5.071 * cos(2.61075526868  + 1536.71276564440 *T)
    sumR1  +=       4.052 * cos(5.00014006312  + 391.64654897590 *T)
    sumR1  +=       5.182 * cos(4.73444634983  + 382.83608094570 *T)
    sumR1  +=       3.763 * cos(4.29449373755  + 313.68355667090 *T)
    sumR1  +=       4.038 * cos(2.82857942788  + 1661.14618086540 *T)
    sumR1  +=       4.067 * cos(5.73169928960  + 169.53698550770 *T)
    sumR1  +=       3.841 * cos(1.62580928420  + 0.96320784650 *T)
    sumR1  +=       3.901 * cos(2.70874386576  + 14.01464568050 *T)
    sumR1  +=       3.721 * cos(1.20062375429  + 1026.87796884390 *T)
    sumR1  +=       3.911 * cos(3.01809123569  + 100.38446123290 *T)
    sumR1  +=       3.489 * cos(4.28865448963  + 1025.18104781450 *T)
    sumR1  +=       3.714 * cos(5.05021268365  + 292.48592802040 *T)
    sumR1  +=       3.816 * cos(3.93084933114  + 39.09624348430 *T)
    sumR1  +=       3.988 * cos(2.82832650224  + 134.11226285560 *T)
    sumR1  +=       3.745 * cos(4.24728135115  + 180.79513409270 *T)
    sumR1  +=       3.836 * cos(1.02685786071  + 1018.27994913480 *T)
    sumR1  +=       3.941 * cos(5.21895739331  + 183.76407950930 *T)
    sumR1  +=       4.669 * cos(4.38080962573  + 1066.49547719000 *T)
    sumR1  +=       3.780 * cos(6.03723468132  + 1022.73336725970 *T)
    sumR1  +=       3.647 * cos(3.98130320367  + 608.87779767700 *T)
    sumR1  +=       3.456 * cos(5.54052355058  + 846.08283475120 *T)
    sumR1  +=       4.047 * cos(3.71041480907  + 1018.06750081370 *T)
    sumR1  +=       3.865 * cos(4.76002199091  + 166.56804009110 *T)
    sumR1  +=       3.629 * cos(3.29053233846  + 447.79581952650 *T)
    sumR1  +=       3.564 * cos(4.36703678321  + 397.06317494730 *T)
    sumR1  +=       3.304 * cos(1.49289552229  + 1505.69327700740 *T)
    sumR1  +=       3.976 * cos(2.42476188945  + 106.01353552540 *T)
    sumR1  +=       4.217 * cos(4.21677652639  + 1052.26838318840 *T)
    sumR1  +=       3.294 * cos(0.42088065654  + 22.63391724900 *T)
    sumR1  +=       3.615 * cos(3.68096122231  + 494.52687487340 *T)
    sumR1  +=       3.230 * cos(5.10786091356  + 69.15252427480 *T)
    sumR1  +=       3.280 * cos(3.62226152032  + 531.17543780290 *T)
    sumR1  +=       3.337 * cos(2.72502876320  + 481.47543703940 *T)
    sumR1  +=       3.187 * cos(0.08677634706  + 399.51085550210 *T)
    sumR1  +=       3.389 * cos(1.79454271219  + 1519.92037100900 *T)
    sumR1  +=       3.179 * cos(3.40418030121  + 423.62924545940 *T)
    sumR1  +=       3.154 * cos(3.69356460843  + 470.21728845440 *T)
    sumR1  +=       3.706 * cos(2.79048710497  + 462.02291352810 *T)
    sumR1  +=       3.136 * cos(4.38015969606  + 385.49620982160 *T)
    sumR1  +=       3.122 * cos(0.48346644637  + 79.18683258240 *T)
    sumR1  +=       3.392 * cos(0.48037804731  + 521.09294538550 *T)
    sumR1  +=       3.465 * cos(0.93152295589  + 2183.72359895920 *T)
    sumR1  +=       3.735 * cos(0.98809808606  + 487.41332787260 *T)
    sumR1  +=       3.998 * cos(3.38773325131  + 6283.07584999140 *T)
    sumR1  +=       2.998 * cos(2.61728063127  + 487.62577619370 *T)
    sumR1  +=       3.295 * cos(2.53821501556  + 4.66586644600 *T)
    sumR1  +=       2.964 * cos(3.66274645375  + 495.49008271990 *T)
    sumR1  +=       3.901 * cos(1.65463523144  + 210.33015002140 *T)
    sumR1  +=       2.950 * cos(1.99904237956  + 872.90953769420 *T)
    sumR1  +=       2.948 * cos(2.90769224206  + 391.43410065480 *T)
    sumR1  +=       2.971 * cos(0.31626092637  + 5.10780943070 *T)
    sumR1  +=       3.085 * cos(0.95725590904  + 109.94568878850 *T)
    sumR1  +=       2.995 * cos(3.34433305798  + 394.09422953070 *T)
    sumR1  +=       3.126 * cos(5.89472116854  + 105.49227066360 *T)
    sumR1  +=       3.904 * cos(3.01022809543  + 556.51766803760 *T)
    sumR1  +=       3.388 * cos(6.24936444215  + 535.32003938710 *T)
    sumR1  +=       2.930 * cos(6.15005257333  + 164.12035953630 *T)
    sumR1  +=       3.267 * cos(4.19718045293  + 518.38463239980 *T)
    sumR1  +=       3.946 * cos(2.88842759670  + 151.26011816400 *T)
    sumR1  +=       3.076 * cos(6.04134449219  + 142.14083359310 *T)
    sumR1  +=       2.823 * cos(0.60712626756  + 214.78356814630 *T)
    sumR1  +=       2.917 * cos(2.74502617182  + 138.51749687070 *T)
    sumR1  +=       3.347 * cos(6.09373507569  + 6246.42728706190 *T)
    sumR1  +=       3.659 * cos(5.12211619716  + 79.23501669220 *T)
    sumR1  +=       3.010 * cos(0.24656411754  + 91.78644152380 *T)
    sumR1  +=       2.861 * cos(6.17465663902  + 422.40540518200 *T)
    sumR1  +=       2.989 * cos(2.31620917965  + 485.92885516430 *T)
    sumR1  +=       3.088 * cos(2.29186342974  + 110.20632121940 *T)
    sumR1  +=       3.030 * cos(3.69866149100  + 532.61172640140 *T)
    sumR1  +=       3.020 * cos(2.36422658177  + 290.21955801940 *T)
    sumR1  +=       3.170 * cos(1.23078934548  + 10176.63698328150 *T)
    sumR1  +=       2.652 * cos(3.35836234807  + 148.07872442630 *T)
    sumR1  +=       2.673 * cos(6.03366372927  + 196.50670080260 *T)
    sumR1  +=       2.630 * cos(0.46957619348  + 1970.42450352120 *T)
    sumR1  +=       2.599 * cos(4.86022081674  + 439.19779981740 *T)
    sumR1  +=       2.878 * cos(2.61946597178  + 488.58898404020 *T)
    sumR1  +=       2.720 * cos(1.71836225398  + 364.55921360200 *T)
    sumR1  +=       3.333 * cos(3.25126857354  + 30.05628079050 *T)
    sumR1  +=       3.053 * cos(2.49346960035  + 6243.45834164530 *T)
    sumR1  +=       3.062 * cos(6.23776299963  + 419.48464387520 *T)
    sumR1  +=       2.786 * cos(0.83078219939  + 497.18700374930 *T)
    sumR1  +=       2.834 * cos(3.52926079424  + 457.87831194390 *T)
    sumR1  +=       2.932 * cos(1.80245810977  + 500.15594916590 *T)
    sumR1  +=       3.030 * cos(5.10152500393  + 367.97010200330 *T)
    sumR1  +=       2.956 * cos(5.76230870725  + 986.08480433020 *T)
    sumR1  +=       3.116 * cos(2.20042242739  + 495.70253104100 *T)
    sumR1  +=       2.554 * cos(0.65945973992  + 67.35923502580 *T)
    sumR1  +=       2.901 * cos(3.91891656185  + 10173.66803786490 *T)
    sumR1  +=       2.840 * cos(1.34453183591  + 482.95990974770 *T)
    sumR1  +=       2.458 * cos(1.20012815574  + 489.11024890200 *T)
    sumR1  +=       2.556 * cos(3.86921927085  + 487.10451133190 *T)
    sumR1  +=       2.614 * cos(1.51881085312  + 463.50738623640 *T)
    sumR1  +=       2.386 * cos(4.58400538443  + 615.99134467780 *T)
    sumR1  +=       2.438 * cos(5.19827220476  + 501.11915701240 *T)
    sumR1  +=       2.537 * cos(1.64802783144  + 519.60847267720 *T)
    sumR1  +=       2.444 * cos(3.87859489652  + 185.24855221760 *T)
    sumR1  +=       2.795 * cos(4.04265752580  + 255.05546779820 *T)
    sumR1  +=       2.895 * cos(3.26202698812  + 1646.91908686380 *T)
    sumR1  +=       2.225 * cos(5.75197574692  + 605.95703637020 *T)
    sumR1  +=       2.324 * cos(3.99503920129  + 481.26298871830 *T)
    sumR1  +=       2.962 * cos(1.74151265966  + 2080.63082474060 *T)
    sumR1  +=       2.621 * cos(1.74442251671  + 35.21227433100 *T)

    # NEPTUNE Radius
    sumR2  = 0
    sumR2  +=    4247.412 * cos(5.89910679117  + 38.13303563780 *T)
    sumR2  +=     217.570 * cos(0.34581829080  + 1.48447270830 *T)
    sumR2  +=     163.025 * cos(2.23872947130  + 168.05251279940 *T)
    sumR2  +=     156.285 * cos(4.59414467342  + 182.27960680100 *T)
    sumR2  +=     117.940 * cos(5.10295026024  + 484.44438245600 *T)
    sumR2  +=     112.429 * cos(1.19000583596  + 498.67147645760 *T)
    sumR2  +=     127.141 * cos(2.84786298079  + 35.16409022120 *T)
    sumR2  +=      99.467 * cos(3.41578558739  + 175.16605980020 *T)
    sumR2  +=      64.814 * cos(3.46214064840  + 388.46515523820 *T)
    sumR2  +=      77.286 * cos(0.01659281785  + 491.55792945680 *T)
    sumR2  +=      49.509 * cos(4.06995509133  + 76.26607127560 *T)
    sumR2  +=      39.330 * cos(6.09521855958  + 1021.24889455140 *T)
    sumR2  +=      36.450 * cos(5.17130059988  + 137.03302416240 *T)
    sumR2  +=      37.080 * cos(5.97288967681  + 2.96894541660 *T)
    sumR2  +=      30.484 * cos(3.58259801313  + 33.67961751290 *T)
    sumR2  +=      21.099 * cos(0.76843555176  + 36.64856292950 *T)
    sumR2  +=      13.886 * cos(3.59248623971  + 395.57870223900 *T)
    sumR2  +=      13.117 * cos(5.09263515697  + 98.89998852460 *T)
    sumR2  +=      11.379 * cos(1.18060018898  + 381.35160823740 *T)
    sumR2  +=       9.132 * cos(2.34787658568  + 601.76425067620 *T)
    sumR2  +=       8.527 * cos(5.25134685897  + 2.44768055480 *T)
    sumR2  +=       8.136 * cos(4.96270726986  + 4.45341812490 *T)
    sumR2  +=       7.417 * cos(4.46775409796  + 189.39315380180 *T)
    sumR2  +=       7.225 * cos(1.92287508629  + 9.56122755560 *T)
    sumR2  +=       7.289 * cos(1.65519525780  + 1028.36244155220 *T)
    sumR2  +=       8.076 * cos(5.84268048311  + 220.41264243880 *T)
    sumR2  +=       9.654 * cos(0.00000000000  + 0.00000000000 *T)
    sumR2  +=       6.554 * cos(0.69397520733  + 144.14657116320 *T)
    sumR2  +=       7.782 * cos(1.14341656235  + 1059.38193018920 *T)
    sumR2  +=       5.665 * cos(6.25378258571  + 74.78159856730 *T)
    sumR2  +=       5.628 * cos(5.23383764266  + 46.20979048510 *T)
    sumR2  +=       5.523 * cos(4.59041448911  + 1014.13534755060 *T)
    sumR2  +=       5.177 * cos(5.23116646157  + 477.33083545520 *T)
    sumR2  +=       5.503 * cos(3.49522319102  + 183.76407950930 *T)
    sumR2  +=       4.878 * cos(3.52934357721  + 39.61750834610 *T)
    sumR2  +=       4.787 * cos(2.08260524745  + 41.10198105440 *T)
    sumR2  +=       5.055 * cos(0.19949888617  + 166.56804009110 *T)
    sumR2  +=       4.751 * cos(1.18054948270  + 169.53698550770 *T)
    sumR2  +=       4.747 * cos(1.50608965076  + 73.29712585900 *T)
    sumR2  +=       6.113 * cos(6.18326155595  + 71.81265315070 *T)
    sumR2  +=       4.606 * cos(3.91970908886  + 587.53715667460 *T)
    sumR2  +=       5.756 * cos(2.23667359233  + 176.65053250850 *T)
    sumR2  +=       4.536 * cos(2.84337336954  + 7.11354700080 *T)
    sumR2  +=       4.338 * cos(0.51553847388  + 446.31134681820 *T)
    sumR2  +=       3.891 * cos(0.26338839265  + 1550.93985964600 *T)
    sumR2  +=       4.465 * cos(3.01487041298  + 129.91947716160 *T)
    sumR2  +=       3.727 * cos(2.37977930658  + 160.93896579860 *T)
    sumR2  +=       3.840 * cos(3.79290381880  + 111.43016149680 *T)
    sumR2  +=       4.142 * cos(1.70293820961  + 983.11585891360 *T)
    sumR2  +=       3.296 * cos(1.07748822909  + 505.78502345840 *T)
    sumR2  +=       4.008 * cos(0.30663868827  + 494.73932319450 *T)
    sumR2  +=       3.974 * cos(5.97351783840  + 488.37653571910 *T)
    sumR2  +=       3.925 * cos(4.85736421123  + 60.76695288680 *T)
    sumR2  +=       2.966 * cos(2.01608546009  + 822.17689311500 *T)
    sumR2  +=       3.972 * cos(1.07780371834  + 374.23806123660 *T)
    sumR2  +=       3.843 * cos(5.23002047199  + 350.33211960040 *T)
    sumR2  +=       2.848 * cos(6.17799253802  + 704.85702489480 *T)
    sumR2  +=       3.527 * cos(0.79317138165  + 274.06604832480 *T)
    sumR2  +=       2.828 * cos(1.32275775835  + 386.98068252990 *T)
    sumR2  +=       2.773 * cos(5.37132330836  + 251.43213107580 *T)
    sumR2  +=       3.113 * cos(5.12622288690  + 426.59819087600 *T)
    sumR2  +=       3.344 * cos(5.61433537548  + 1124.34166877000 *T)
    sumR2  +=       2.597 * cos(0.67759426519  + 312.19908396260 *T)
    sumR2  +=       2.581 * cos(3.55847612121  + 567.82400073240 *T)
    sumR2  +=       2.578 * cos(1.45603792456  + 1035.47598855300 *T)
    sumR2  +=       2.541 * cos(5.19427579702  + 1227.43444298860 *T)
    sumR2  +=       2.510 * cos(4.12148891512  + 171.23390653710 *T)
    sumR2  +=       2.511 * cos(2.71606957319  + 179.09821306330 *T)
    sumR2  +=       2.342 * cos(0.96469916587  + 1019.76442184310 *T)
    sumR2  +=       2.500 * cos(0.70282276030  + 707.77778620160 *T)
    sumR2  +=       2.480 * cos(4.59623030219  + 693.55069220000 *T)
    sumR2  +=       2.253 * cos(0.74334306011  + 976.00231191280 *T)

    # NEPTUNE Radius
    sumR3  = 0
    sumR3  +=     166.297 * cos(4.55243893489  + 38.13303563780 *T)
    sumR3  +=      22.380 * cos(3.94830879358  + 168.05251279940 *T)
    sumR3  +=      21.348 * cos(2.86296778794  + 182.27960680100 *T)
    sumR3  +=      16.233 * cos(0.54226725872  + 484.44438245600 *T)
    sumR3  +=      15.623 * cos(5.75702251906  + 498.67147645760 *T)
    sumR3  +=      11.867 * cos(4.40280192710  + 1.48447270830 *T)
    sumR3  +=       6.448 * cos(5.19003066847  + 31.01948863700 *T)
    sumR3  +=       3.655 * cos(5.91335292846  + 1007.02180054980 *T)
    sumR3  +=       3.681 * cos(1.62865545676  + 388.46515523820 *T)
    sumR3  +=       3.198 * cos(0.70197118575  + 1558.05340664680 *T)
    sumR3  +=       3.243 * cos(1.88035665980  + 522.57741809380 *T)
    sumR3  +=       3.269 * cos(2.94301808574  + 76.26607127560 *T)
    sumR3  +=       2.688 * cos(1.87062743473  + 402.69224923980 *T)
    sumR3  +=       3.246 * cos(0.79381356193  + 536.80451209540 *T)
    sumR3  +=       2.650 * cos(5.76858449026  + 343.21857259960 *T)
    sumR3  +=       2.644 * cos(4.64542905401  + 500.15594916590 *T)
    sumR3  +=       2.541 * cos(4.79217120822  + 482.95990974770 *T)
    sumR3  +=       2.523 * cos(1.72869889780  + 395.57870223900 *T)
    sumR3  +=       2.690 * cos(2.21096415618  + 446.31134681820 *T)
    sumR3  +=       2.355 * cos(5.77381398401  + 485.92885516430 *T)
    sumR3  +=       2.874 * cos(6.19643340540  + 815.06334611420 *T)
    sumR3  +=       2.278 * cos(3.66579603119  + 497.18700374930 *T)

    # NEPTUNE Radius
    sumR4  = 0
    sumR4  +=       4.227 * cos(2.40375758563  + 477.33083545520 *T)
    sumR4  +=       4.333 * cos(0.10459484545  + 395.57870223900 *T)
    sumR4  +=       3.545 * cos(4.78431259422  + 1028.36244155220 *T)
    sumR4  +=       3.154 * cos(3.88192942366  + 505.78502345840 *T)
    sumR4  +=       3.016 * cos(1.03609346831  + 189.39315380180 *T)
    sumR4  +=       2.294 * cos(1.10879658603  + 182.27960680100 *T)
    sumR4  +=       2.295 * cos(5.67776133184  + 168.05251279940 *T)

    # calculate and scale the sums
    L = sumL0 + sumL1*T**1 + sumL2*T**2 + sumL3*T**3 + sumL4*T**4 + sumL5*T**5
    L /= 1e8
    L = Mod(degrees(L), 360.)
    B = sumB0 + sumB1*T**1 + sumB2*T**2 + sumB3*T**3 + sumB4*T**4 + sumB5*T**5
    B /= 1e8
    B = degrees(B)
    R = sumR0 + sumR1*T**1 + sumR2*T**2 + sumR3*T**3 + sumR4*T**4
    R /= 1e8

    # calculate X, Y, Z
    X = R * Cos(B) * Cos(L)
    Y = R * Cos(B) * Sin(L)
    Z = R * Sin(B)

    # create dictionary
    Dict = {  "Object"  : "Neptune",
              "Mode"    : "LBR",
              "Long"    : L,
              "Lat"     : B,
              "Rau"     : R,
              "X"       : X,
              "Y"       : Y,
              "Z"       : Z,
              "Alt"     : -34./60.
           }
    Data = (L, B, R, Dict)


    return Data
#######################################################################



def getNeptuneData(dtCivicTime, zone, verbose = False):
    L, B, R, Dict = getNeptuneLBRData(dtCivicTime, zone, verbose)
    X = Dict["X"]; Y = Dict["Y"]; Z = Dict["Z"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Le, Be, Re, Dicte = getEarthLBRData(dtCivicTime, zone, verbose)
        Xe = Dicte["X"]; Ye = Dicte["Y"]; Ze = Dicte["Z"]
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getSaturnLBRData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Saturn"
#       Long  :  heliocentric Longitude, degrees
#       Lat   :  heliocentric Latitude,degrees
#       Rau   :  Radius, distaance Sun - Object (au)
#
#   Similarly for each planet
#   data from Meeus, Appendix II and VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.2 hvf 2023-04-06 getSaturnLBRData.py created by mkPlanetLBRData.py"
VersionList.append(Version)

def getSaturnLBRData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. (31.1) Julian Millennia

#######################################################################

    # SATURN Longitude
    sumL0  = 0
    sumL0  +=  87401354.029 * cos(0.00000000000  + 0.00000000000 *T)
    sumL0  +=  11107659.780 * cos(3.96205090194  + 213.29909543800 *T)
    sumL0  +=  1414150.958 * cos(4.58581515873  + 7.11354700080 *T)
    sumL0  +=  398379.386 * cos(0.52112025957  + 206.18554843720 *T)
    sumL0  +=  350769.223 * cos(3.30329903015  + 426.59819087600 *T)
    sumL0  +=  206816.296 * cos(0.24658366938  + 103.09277421860 *T)
    sumL0  +=   79271.288 * cos(3.84007078530  + 220.41264243880 *T)
    sumL0  +=   23990.338 * cos(4.66976934860  + 110.20632121940 *T)
    sumL0  +=   16573.583 * cos(0.43719123541  + 419.48464387520 *T)
    sumL0  +=   14906.995 * cos(5.76903283845  + 316.39186965660 *T)
    sumL0  +=   15820.300 * cos(0.93808953760  + 632.78373931320 *T)
    sumL0  +=   14609.562 * cos(1.56518573691  + 3.93215326310 *T)
    sumL0  +=   13160.308 * cos(4.44891180176  + 14.22709400160 *T)
    sumL0  +=   15053.509 * cos(2.71670027883  + 639.89728631400 *T)
    sumL0  +=   13005.305 * cos(5.98119067061  + 11.04570026390 *T)
    sumL0  +=   10725.066 * cos(3.12939596466  + 202.25339517410 *T)
    sumL0  +=    5863.207 * cos(0.23657028777  + 529.69096509460 *T)
    sumL0  +=    5227.771 * cos(4.20783162380  + 3.18139373770 *T)
    sumL0  +=    6126.308 * cos(1.76328499656  + 277.03499374140 *T)
    sumL0  +=    5019.658 * cos(3.17787919533  + 433.71173787680 *T)
    sumL0  +=    4592.541 * cos(0.61976424374  + 199.07200143640 *T)
    sumL0  +=    4005.862 * cos(2.24479893937  + 63.73589830340 *T)
    sumL0  +=    2953.815 * cos(0.98280385206  + 95.97922721780 *T)
    sumL0  +=    3873.696 * cos(3.22282692566  + 138.51749687070 *T)
    sumL0  +=    2461.172 * cos(2.03163631205  + 735.87651353180 *T)
    sumL0  +=    3269.490 * cos(0.77491895787  + 949.17560896980 *T)
    sumL0  +=    1758.143 * cos(3.26580514774  + 522.57741809380 *T)
    sumL0  +=    1640.183 * cos(5.50504966218  + 846.08283475120 *T)
    sumL0  +=    1391.336 * cos(4.02331978116  + 323.50541665740 *T)
    sumL0  +=    1580.641 * cos(4.37266314120  + 309.27832265580 *T)
    sumL0  +=    1123.515 * cos(2.83726793572  + 415.55249061210 *T)
    sumL0  +=    1017.258 * cos(3.71698151814  + 227.52618943960 *T)
    sumL0  +=     848.643 * cos(3.19149825839  + 209.36694217490 *T)
    sumL0  +=    1087.237 * cos(4.18343232481  + 2.44768055480 *T)
    sumL0  +=     956.752 * cos(0.50740889886  + 1265.56747862640 *T)
    sumL0  +=     789.205 * cos(5.00745123149  + 0.96320784650 *T)
    sumL0  +=     686.965 * cos(1.74714407827  + 1052.26838318840 *T)
    sumL0  +=     654.470 * cos(1.59889331515  + 0.04818410980 *T)
    sumL0  +=     748.811 * cos(2.14398149298  + 853.19638175200 *T)
    sumL0  +=     633.980 * cos(2.29889903023  + 412.37109687440 *T)
    sumL0  +=     743.584 * cos(5.25276954625  + 224.34479570190 *T)
    sumL0  +=     852.677 * cos(3.42141350697  + 175.16605980020 *T)
    sumL0  +=     579.857 * cos(3.09259007048  + 74.78159856730 *T)
    sumL0  +=     624.904 * cos(0.97046831256  + 210.11770170030 *T)
    sumL0  +=     529.861 * cos(4.44938897119  + 117.31986822020 *T)
    sumL0  +=     542.643 * cos(1.51824320514  + 9.56122755560 *T)
    sumL0  +=     474.279 * cos(5.47527185987  + 742.99006053260 *T)
    sumL0  +=     448.542 * cos(1.28990416161  + 127.47179660680 *T)
    sumL0  +=     546.358 * cos(2.12678554211  + 350.33211960040 *T)
    sumL0  +=     478.054 * cos(2.96488054338  + 137.03302416240 *T)
    sumL0  +=     354.944 * cos(3.01286483030  + 838.96928775040 *T)
    sumL0  +=     451.827 * cos(1.04436664241  + 490.33408917940 *T)
    sumL0  +=     347.413 * cos(1.53928227764  + 340.77089204480 *T)
    sumL0  +=     343.475 * cos(0.24604039134  + 0.52126486180 *T)
    sumL0  +=     309.001 * cos(3.49486734909  + 216.48048917570 *T)
    sumL0  +=     322.185 * cos(0.96137456104  + 203.73786788240 *T)
    sumL0  +=     372.308 * cos(2.27819108625  + 217.23124870110 *T)
    sumL0  +=     321.543 * cos(2.57182354537  + 647.01083331480 *T)
    sumL0  +=     330.196 * cos(0.24715617844  + 1581.95934828300 *T)
    sumL0  +=     249.116 * cos(1.47010534421  + 1368.66025284500 *T)
    sumL0  +=     286.688 * cos(2.37043745859  + 351.81659230870 *T)
    sumL0  +=     220.225 * cos(4.20422424873  + 200.76892246580 *T)
    sumL0  +=     277.775 * cos(0.40020408926  + 211.81462272970 *T)
    sumL0  +=     204.500 * cos(6.01082206600  + 265.98929347750 *T)
    sumL0  +=     207.663 * cos(0.48349820488  + 1162.47470440780 *T)
    sumL0  +=     208.655 * cos(1.34516255304  + 625.67019231240 *T)
    sumL0  +=     182.454 * cos(5.49122292426  + 2.92076130680 *T)
    sumL0  +=     226.609 * cos(4.91003163138  + 12.53017297220 *T)
    sumL0  +=     207.659 * cos(1.28302218900  + 39.35687591520 *T)
    sumL0  +=     173.914 * cos(1.86305806814  + 0.75075952540 *T)
    sumL0  +=     184.690 * cos(3.50344404958  + 149.56319713460 *T)
    sumL0  +=     183.511 * cos(0.97254952728  + 4.19278569400 *T)
    sumL0  +=     146.068 * cos(6.23102544071  + 195.13984817330 *T)
    sumL0  +=     164.541 * cos(0.44005517520  + 5.41662597140 *T)
    sumL0  +=     147.526 * cos(1.53529320509  + 5.62907429250 *T)
    sumL0  +=     139.666 * cos(4.29450260069  + 21.34064100240 *T)
    sumL0  +=     131.283 * cos(4.06828961903  + 10.29494073850 *T)
    sumL0  +=     117.283 * cos(2.67920400584  + 1155.36115740700 *T)
    sumL0  +=     149.299 * cos(5.73594349789  + 52.69019803950 *T)
    sumL0  +=     122.373 * cos(1.97588777199  + 4.66586644600 *T)
    sumL0  +=     113.747 * cos(5.59427544714  + 1059.38193018920 *T)
    sumL0  +=     102.702 * cos(1.19748124058  + 1685.05212250160 *T)
    sumL0  +=     118.156 * cos(5.34072933900  + 554.06998748280 *T)
    sumL0  +=     109.275 * cos(3.43812715686  + 536.80451209540 *T)
    sumL0  +=     110.399 * cos(0.16604024090  + 1.48447270830 *T)
    sumL0  +=     124.969 * cos(6.27737805832  + 1898.35121793960 *T)
    sumL0  +=      89.949 * cos(5.80392934702  + 114.13847448250 *T)
    sumL0  +=     103.956 * cos(2.19210363069  + 88.86568021700 *T)
    sumL0  +=     112.437 * cos(1.10502663534  + 191.20769491020 *T)
    sumL0  +=     106.570 * cos(4.01156608514  + 956.28915597060 *T)
    sumL0  +=      91.430 * cos(1.87521577510  + 38.13303563780 *T)
    sumL0  +=      83.791 * cos(5.48810655641  + 0.11187458460 *T)
    sumL0  +=      83.461 * cos(2.28972767279  + 628.85158605010 *T)
    sumL0  +=      96.987 * cos(4.53666595763  + 302.16477565500 *T)
    sumL0  +=     100.631 * cos(4.96513666539  + 269.92144674060 *T)
    sumL0  +=      75.491 * cos(2.18045274099  + 728.76296653100 *T)
    sumL0  +=      96.330 * cos(2.83319189210  + 275.55052103310 *T)
    sumL0  +=      82.363 * cos(3.05469876064  + 440.82528487760 *T)
    sumL0  +=      73.888 * cos(5.08914205084  + 1375.77379984580 *T)
    sumL0  +=      71.633 * cos(5.10940743430  + 65.22037101170 *T)
    sumL0  +=      70.409 * cos(4.86846451411  + 0.21244832110 *T)
    sumL0  +=      69.760 * cos(3.71029022489  + 14.97785352700 *T)
    sumL0  +=      88.772 * cos(3.86334563977  + 278.51946644970 *T)
    sumL0  +=      68.090 * cos(0.73415460990  + 1478.86657406440 *T)
    sumL0  +=      66.501 * cos(0.02677580336  + 70.84944530420 *T)
    sumL0  +=      65.682 * cos(2.02165559602  + 142.44965013380 *T)
    sumL0  +=      75.765 * cos(1.61410487792  + 284.14854074220 *T)
    sumL0  +=      63.153 * cos(3.49493353034  + 479.28838891550 *T)
    sumL0  +=      62.539 * cos(2.58713611532  + 422.66603761290 *T)
    sumL0  +=      69.313 * cos(3.43979731402  + 515.46387109300 *T)
    sumL0  +=      79.021 * cos(4.45154941586  + 35.42472265210 *T)
    sumL0  +=      63.664 * cos(3.31749528708  + 62.25142559510 *T)
    sumL0  +=      52.939 * cos(5.51392725227  + 0.26063243090 *T)
    sumL0  +=      53.011 * cos(3.18480701697  + 8.07675484730 *T)
    sumL0  +=      54.492 * cos(2.45674090515  + 22.09140052780 *T)
    sumL0  +=      50.514 * cos(4.26749346978  + 99.16062095550 *T)
    sumL0  +=      55.170 * cos(0.96797446150  + 942.06206196900 *T)
    sumL0  +=      49.288 * cos(2.38641424063  + 1471.75302706360 *T)
    sumL0  +=      47.199 * cos(2.02515248245  + 312.19908396260 *T)
    sumL0  +=      61.080 * cos(1.50295092063  + 210.85141488320 *T)
    sumL0  +=      45.126 * cos(0.93109376473  + 2001.44399215820 *T)
    sumL0  +=      60.556 * cos(2.68715551585  + 388.46515523820 *T)
    sumL0  +=      43.452 * cos(2.52602011714  + 288.08069400530 *T)
    sumL0  +=      42.544 * cos(3.81793980322  + 330.61896365820 *T)
    sumL0  +=      39.915 * cos(5.71378652900  + 408.43894361130 *T)
    sumL0  +=      50.145 * cos(6.03164759907  + 2214.74308759620 *T)
    sumL0  +=      45.860 * cos(0.54229721801  + 212.33588759150 *T)
    sumL0  +=      54.165 * cos(0.78154835399  + 191.95845443560 *T)
    sumL0  +=      47.016 * cos(4.59934671151  + 437.64389113990 *T)
    sumL0  +=      42.362 * cos(1.90070070955  + 430.53034413910 *T)
    sumL0  +=      39.722 * cos(1.63259419913  + 1066.49547719000 *T)
    sumL0  +=      36.345 * cos(0.84756992711  + 213.34727954780 *T)
    sumL0  +=      35.468 * cos(4.18603772925  + 215.74677599280 *T)
    sumL0  +=      36.344 * cos(3.93295730315  + 213.25091132820 *T)
    sumL0  +=      38.005 * cos(0.31313803095  + 423.41679713830 *T)
    sumL0  +=      44.746 * cos(1.12488341174  + 6.15033915430 *T)
    sumL0  +=      37.902 * cos(1.19795851115  + 2.70831298570 *T)
    sumL0  +=      43.402 * cos(1.37363944007  + 563.63121503840 *T)
    sumL0  +=      43.764 * cos(3.93043802956  + 525.49817940060 *T)
    sumL0  +=      34.825 * cos(1.01566605408  + 203.00415469950 *T)
    sumL0  +=      31.755 * cos(1.69273634405  + 0.16005869440 *T)
    sumL0  +=      30.880 * cos(6.13525703832  + 417.03696332040 *T)
    sumL0  +=      36.388 * cos(6.00586032647  + 18.15924726470 *T)
    sumL0  +=      29.032 * cos(1.19660544505  + 404.50679034820 *T)
    sumL0  +=      32.812 * cos(0.53649479713  + 107.02492748170 *T)
    sumL0  +=      30.433 * cos(0.72335287989  + 222.86032299360 *T)
    sumL0  +=      32.644 * cos(0.81204701486  + 1795.25844372100 *T)
    sumL0  +=      37.769 * cos(3.69666903716  + 1272.68102562720 *T)
    sumL0  +=      27.679 * cos(1.45663979401  + 7.16173111060 *T)
    sumL0  +=      27.187 * cos(1.89731951902  + 1045.15483618760 *T)
    sumL0  +=      37.699 * cos(4.51997049537  + 24.37902238820 *T)
    sumL0  +=      34.885 * cos(4.46095761791  + 214.26230328450 *T)
    sumL0  +=      32.650 * cos(0.66372395761  + 692.58748435350 *T)
    sumL0  +=      30.324 * cos(5.30369950147  + 33.94024994380 *T)
    sumL0  +=      27.480 * cos(6.22702216249  + 1.27202438720 *T)
    sumL0  +=      26.657 * cos(4.56713198392  + 7.06536289100 *T)
    sumL0  +=      31.745 * cos(5.49798599565  + 56.62235130260 *T)
    sumL0  +=      28.050 * cos(5.64447420566  + 128.95626931510 *T)
    sumL0  +=      24.277 * cos(3.93966553574  + 414.06801790380 *T)
    sumL0  +=      32.017 * cos(5.22260660455  + 92.04707395470 *T)
    sumL0  +=      26.976 * cos(0.06705123981  + 205.22234059070 *T)
    sumL0  +=      22.974 * cos(3.65817751770  + 207.67002114550 *T)
    sumL0  +=      31.775 * cos(5.59198119173  + 6069.77675455340 *T)
    sumL0  +=      23.153 * cos(2.10054506119  + 1788.14489672020 *T)
    sumL0  +=      31.025 * cos(0.37190053329  + 703.63318461740 *T)
    sumL0  +=      29.376 * cos(0.14742155778  + 131.40394986990 *T)
    sumL0  +=      22.562 * cos(5.24009182383  + 212.77783057620 *T)
    sumL0  +=      26.185 * cos(5.41311252822  + 140.00196957900 *T)
    sumL0  +=      25.673 * cos(4.36038885283  + 32.24332891440 *T)
    sumL0  +=      20.392 * cos(2.82413909260  + 429.77958461370 *T)
    sumL0  +=      20.659 * cos(0.67091805084  + 2317.83586181480 *T)
    sumL0  +=      24.397 * cos(3.08740396398  + 145.63104387150 *T)
    sumL0  +=      23.735 * cos(2.54365387567  + 76.26607127560 *T)
    sumL0  +=      20.157 * cos(5.06708675157  + 617.80588578620 *T)
    sumL0  +=      23.307 * cos(3.97357729211  + 483.22054217860 *T)
    sumL0  +=      22.878 * cos(6.10452832642  + 177.87437278590 *T)
    sumL0  +=      22.978 * cos(3.20140795404  + 208.63322899200 *T)
    sumL0  +=      20.638 * cos(5.22128727027  + 6.59228213900 *T)
    sumL0  +=      21.446 * cos(0.72034565528  + 1258.45393162560 *T)
    sumL0  +=      18.034 * cos(6.11382719947  + 210.37833413120 *T)
    sumL0  +=      22.380 * cos(5.92299908546  + 173.94221952280 *T)
    sumL0  +=      19.128 * cos(5.77772013766  + 213.82036029980 *T)
    sumL0  +=      20.871 * cos(5.79126331864  + 2531.13495725280 *T)
    sumL0  +=      19.327 * cos(1.64147367403  + 565.11568774670 *T)
    sumL0  +=      16.806 * cos(3.27953583323  + 98.89998852460 *T)
    sumL0  +=      20.833 * cos(2.01655935909  + 860.30992875280 *T)
    sumL0  +=      17.939 * cos(3.14329498012  + 831.85574074960 *T)
    sumL0  +=      15.653 * cos(3.10137669623  + 106.27416795630 *T)
    sumL0  +=      18.235 * cos(5.22595172482  + 73.29712585900 *T)
    sumL0  +=      19.302 * cos(5.93947114050  + 425.11371816770 *T)
    sumL0  +=      14.514 * cos(2.75049388379  + 1.22384027740 *T)
    sumL0  +=      14.562 * cos(5.18795088579  + 305.34616939270 *T)
    sumL0  +=      14.254 * cos(3.88079504939  + 54.17467074780 *T)
    sumL0  +=      14.594 * cos(3.25016810034  + 78.71375183040 *T)
    sumL0  +=      13.637 * cos(2.55486219141  + 405.25754987360 *T)
    sumL0  +=      13.914 * cos(1.72356993808  + 69.15252427480 *T)
    sumL0  +=      13.689 * cos(2.37430586272  + 125.98732389850 *T)
    sumL0  +=      13.496 * cos(0.82683590985  + 99.91138048090 *T)
    sumL0  +=      18.483 * cos(0.73171264866  + 9999.98645077300 *T)
    sumL0  +=      13.542 * cos(3.58584380924  + 234.63973644040 *T)
    sumL0  +=      13.741 * cos(6.18458356845  + 245.54242435240 *T)
    sumL0  +=      16.944 * cos(0.72200792996  + 2111.65031337760 *T)
    sumL0  +=      17.441 * cos(0.23803796878  + 134.58534360760 *T)
    sumL0  +=      14.181 * cos(4.51963935804  + 59.80374504030 *T)
    sumL0  +=      13.598 * cos(2.53776983965  + 1.69692102940 *T)
    sumL0  +=      12.240 * cos(2.11973445754  + 28.31117565130 *T)
    sumL0  +=      11.988 * cos(1.62114832786  + 1361.54670584420 *T)
    sumL0  +=      11.974 * cos(4.07378735120  + 280.96714700450 *T)
    sumL0  +=      12.758 * cos(5.31146919749  + 344.70304530790 *T)
    sumL0  +=      16.051 * cos(3.97093160336  + 355.74874557180 *T)
    sumL0  +=      11.427 * cos(5.51123470805  + 192.69216761850 *T)
    sumL0  +=      13.133 * cos(4.69168003518  + 767.36908292080 *T)
    sumL0  +=      14.746 * cos(3.28998910617  + 1589.07289528380 *T)
    sumL0  +=      11.417 * cos(1.81615681635  + 2104.53676637680 *T)
    sumL0  +=      11.626 * cos(2.79410384978  + 362.86229257260 *T)
    sumL0  +=      13.234 * cos(4.16642914717  + 225.82926841020 *T)
    sumL0  +=      10.599 * cos(5.50554288376  + 199.28444975750 *T)
    sumL0  +=      10.558 * cos(3.57501718639  + 1.43628859850 *T)
    sumL0  +=      10.485 * cos(2.84462532686  + 85.82729883120 *T)
    sumL0  +=      10.296 * cos(0.22225264071  + 198.32124191100 *T)
    sumL0  +=      10.552 * cos(0.18716643576  + 217.49188113200 *T)
    sumL0  +=      11.853 * cos(0.11584857323  + 7.63481186260 *T)
    sumL0  +=      10.248 * cos(0.21904154170  + 144.14657116320 *T)
    sumL0  +=      10.403 * cos(1.68776321208  + 31.01948863700 *T)
    sumL0  +=      10.313 * cos(4.72132701805  + 216.21985674480 *T)
    sumL0  +=      10.719 * cos(2.60869377832  + 339.28641933650 *T)
    sumL0  +=       9.636 * cos(3.66746262954  + 212.54833591260 *T)
    sumL0  +=       9.631 * cos(3.34275630477  + 223.59403617650 *T)
    sumL0  +=       9.684 * cos(0.41556436593  + 2634.22773147140 *T)
    sumL0  +=       9.885 * cos(4.01798130416  + 207.14875628370 *T)
    sumL0  +=      13.212 * cos(6.00683506785  + 214.78356814630 *T)
    sumL0  +=      11.346 * cos(2.61898383052  + 7.86430652620 *T)
    sumL0  +=       9.158 * cos(5.39855118256  + 342.25536475310 *T)
    sumL0  +=      11.882 * cos(4.00188476744  + 267.47376618580 *T)
    sumL0  +=      12.054 * cos(3.59904816676  + 124.43341522100 *T)
    sumL0  +=       8.921 * cos(4.22716773496  + 6.36278747540 *T)
    sumL0  +=      10.142 * cos(3.60807025662  + 14.01464568050 *T)
    sumL0  +=       9.350 * cos(0.72255756005  + 347.88443904560 *T)
    sumL0  +=      10.529 * cos(2.36779614951  + 831.10498122420 *T)
    sumL0  +=       8.587 * cos(4.48439552745  + 1692.16566950240 *T)
    sumL0  +=      10.142 * cos(3.93620624488  + 207.88246946660 *T)
    sumL0  +=       9.147 * cos(4.28032835242  + 312.45971639350 *T)
    sumL0  +=       8.088 * cos(0.81225752596  + 264.50482076920 *T)
    sumL0  +=       9.241 * cos(4.26402650779  + 20.60692781950 *T)
    sumL0  +=       9.614 * cos(0.64291347187  + 218.92816973050 *T)
    sumL0  +=       8.537 * cos(0.48756672382  + 1574.84580128220 *T)
    sumL0  +=       7.986 * cos(4.71088791079  + 333.65734504400 *T)
    sumL0  +=       8.951 * cos(0.90641577433  + 497.44763618020 *T)
    sumL0  +=       7.959 * cos(2.73277594136  + 4.14460158420 *T)
    sumL0  +=       9.133 * cos(5.08250578843  + 241.61027108930 *T)
    sumL0  +=       9.669 * cos(1.60623316904  + 0.89377187730 *T)
    sumL0  +=       8.883 * cos(5.55491009279  + 2847.52682690940 *T)
    sumL0  +=       8.926 * cos(5.80857835271  + 329.72519178090 *T)
    sumL0  +=       7.226 * cos(0.60164771281  + 206.23373254700 *T)
    sumL0  +=       7.655 * cos(5.53676341721  + 116.42609634290 *T)
    sumL0  +=       7.118 * cos(0.18747501525  + 209.10630974400 *T)
    sumL0  +=       7.507 * cos(5.43555636173  + 621.73803904930 *T)
    sumL0  +=       8.885 * cos(5.36210591059  + 343.21857259960 *T)
    sumL0  +=       7.056 * cos(0.41911130648  + 756.32338265690 *T)
    sumL0  +=       8.124 * cos(4.05571025939  + 237.67811782620 *T)
    sumL0  +=       8.964 * cos(1.65023927130  + 210.33015002140 *T)
    sumL0  +=       6.961 * cos(3.17855200943  + 543.02428721890 *T)
    sumL0  +=       8.916 * cos(0.56503620503  + 2428.04218303420 *T)
    sumL0  +=       6.926 * cos(3.66869171435  + 247.23934538180 *T)
    sumL0  +=       8.982 * cos(4.25046722481  + 46.47042291600 *T)
    sumL0  +=       7.089 * cos(5.14399672225  + 231.45834270270 *T)
    sumL0  +=       7.381 * cos(1.25092810119  + 217.96496188400 *T)
    sumL0  +=       7.134 * cos(2.83090354854  + 1148.24761040620 *T)
    sumL0  +=       6.353 * cos(0.82582711056  + 31.49256938900 *T)
    sumL0  +=       7.558 * cos(5.62617378543  + 518.64526483070 *T)
    sumL0  +=       6.383 * cos(3.54809945181  + 244.31858407500 *T)
    sumL0  +=       6.914 * cos(3.70012837706  + 206.13736432740 *T)
    sumL0  +=       6.286 * cos(5.79144749096  + 179.35884549420 *T)
    sumL0  +=       6.639 * cos(4.55197585824  + 120.35824960600 *T)
    sumL0  +=       5.823 * cos(1.40737990571  + 214.04985496340 *T)
    sumL0  +=       5.850 * cos(4.86725483749  + 320.32402291970 *T)
    sumL0  +=       6.213 * cos(1.07959478499  + 251.43213107580 *T)
    sumL0  +=       7.730 * cos(3.82244175824  + 380.12776796000 *T)
    sumL0  +=       5.716 * cos(1.34909972549  + 1677.93857550080 *T)
    sumL0  +=       6.469 * cos(1.34776801494  + 188.92007304980 *T)
    sumL0  +=       5.668 * cos(2.28643368177  + 20.44686912510 *T)
    sumL0  +=       6.092 * cos(3.62275289839  + 1169.58825140860 *T)
    sumL0  +=       5.711 * cos(0.51687421521  + 148.07872442630 *T)
    sumL0  +=       5.804 * cos(1.54831552984  + 2420.92863603340 *T)
    sumL0  +=       5.703 * cos(5.05993483230  + 2.96894541660 *T)
    sumL0  +=       5.913 * cos(1.66225477547  + 842.15068148810 *T)
    sumL0  +=       7.449 * cos(1.36195943673  + 166.82867252200 *T)
    sumL0  +=       6.482 * cos(1.94032041024  + 357.44566660120 *T)
    sumL0  +=       6.368 * cos(2.44556930837  + 654.12438031560 *T)
    sumL0  +=       6.327 * cos(0.40654591365  + 168.05251279940 *T)
    sumL0  +=       5.573 * cos(2.69383455663  + 750.10360753340 *T)
    sumL0  +=       7.216 * cos(2.22547711392  + 488.84961647110 *T)
    sumL0  +=       6.701 * cos(6.03737590382  + 160.60889739850 *T)
    sumL0  +=       6.938 * cos(5.78362034410  + 700.66423920080 *T)
    sumL0  +=       6.701 * cos(3.14738404371  + 491.81856188770 *T)
    sumL0  +=       5.684 * cos(2.59531540359  + 1.64453140270 *T)
    sumL0  +=       4.900 * cos(2.03902856851  + 0.80314915210 *T)
    sumL0  +=       5.147 * cos(4.10182033298  + 196.62432088160 *T)
    sumL0  +=       4.985 * cos(2.96765983996  + 258.87574647670 *T)
    sumL0  +=       5.911 * cos(1.81507526918  + 252.65597135320 *T)
    sumL0  +=       6.056 * cos(3.33431010543  + 182.27960680100 *T)
    sumL0  +=       6.195 * cos(5.01900871714  + 273.10284047830 *T)
    sumL0  +=       6.316 * cos(5.49053160191  + 206.70681329900 *T)
    sumL0  +=       5.529 * cos(3.31498938717  + 1905.46476494040 *T)
    sumL0  +=       5.102 * cos(3.98171453610  + 254.94359321360 *T)
    sumL0  +=       4.762 * cos(2.24463685255  + 635.96513305090 *T)
    sumL0  +=       5.213 * cos(0.53609344278  + 135.54855145410 *T)
    sumL0  +=       4.639 * cos(0.04466373027  + 213.18722085340 *T)
    sumL0  +=       5.951 * cos(0.54565487490  + 51.20572533120 *T)
    sumL0  +=       4.535 * cos(0.16088614438  + 2950.61960112800 *T)
    sumL0  +=       4.639 * cos(4.73769153591  + 213.41097002260 *T)
    sumL0  +=       4.716 * cos(3.13636467789  + 28.57180808220 *T)
    sumL0  +=       4.748 * cos(1.12156952989  + 6.21977512350 *T)
    sumL0  +=       5.735 * cos(0.04425142145  + 348.84764689210 *T)
    sumL0  +=       4.334 * cos(2.68814219154  + 81.75213321620 *T)
    sumL0  +=       4.538 * cos(3.83676888638  + 487.36514376280 *T)
    sumL0  +=       5.582 * cos(3.63486861028  + 248.72381809010 *T)
    sumL0  +=       4.106 * cos(3.39164360376  + 50.40257617910 *T)
    sumL0  +=       5.657 * cos(3.59967787362  + 282.45161971280 *T)
    sumL0  +=       5.145 * cos(1.33329458239  + 173.68158709190 *T)
    sumL0  +=       3.898 * cos(4.11804949361  + 213.51154375910 *T)
    sumL0  +=       3.898 * cos(0.66430577257  + 213.08664711690 *T)
    sumL0  +=       4.418 * cos(0.10784811796  + 905.88657979150 *T)
    sumL0  +=       4.935 * cos(2.19060382431  + 189.72322220190 *T)
    sumL0  +=       3.799 * cos(2.60752583205  + 546.95644048200 *T)
    sumL0  +=       3.960 * cos(1.60339889010  + 218.71572140940 *T)
    sumL0  +=       3.740 * cos(3.30724497407  + 274.06604832480 *T)
    sumL0  +=       3.778 * cos(0.26606330942  + 636.71589257630 *T)
    sumL0  +=       4.657 * cos(0.37532078548  + 2744.43405269080 *T)
    sumL0  +=       3.682 * cos(5.11587898667  + 458.84151979040 *T)
    sumL0  +=       4.230 * cos(5.18313062329  + 27.08733537390 *T)
    sumL0  +=       5.181 * cos(3.75590784411  + 3127.31333126180 *T)
    sumL0  +=       3.904 * cos(2.21738744557  + 358.93013930950 *T)
    sumL0  +=       4.784 * cos(4.60666675927  + 72.07328558160 *T)
    sumL0  +=       3.552 * cos(3.23789349146  + 543.91805909620 *T)
    sumL0  +=       3.502 * cos(3.68869576093  + 41.64449777560 *T)
    sumL0  +=       4.803 * cos(4.73553427126  + 240.38643081190 *T)
    sumL0  +=       3.768 * cos(3.86077796242  + 2008.55753915900 *T)
    sumL0  +=       3.680 * cos(5.36657425183  + 10.03430830760 *T)
    sumL0  +=       4.298 * cos(3.15595944154  + 738.79727483860 *T)
    sumL0  +=       3.388 * cos(0.73176365772  + 11.30633269480 *T)
    sumL0  +=       3.507 * cos(2.62508475661  + 13.33332212430 *T)
    sumL0  +=       3.552 * cos(0.28967392251  + 1891.23767093880 *T)
    sumL0  +=       3.604 * cos(4.69324090480  + 295.05122865420 *T)
    sumL0  +=       3.621 * cos(6.25264336426  + 129.91947716160 *T)
    sumL0  +=       3.334 * cos(5.04221806054  + 153.49535039770 *T)
    sumL0  +=       3.837 * cos(5.31732096284  + 3163.91869656600 *T)
    sumL0  +=       3.281 * cos(5.59031570352  + 2.00573757010 *T)
    sumL0  +=       4.042 * cos(2.37081308090  + 176.65053250850 *T)
    sumL0  +=       3.500 * cos(2.54744268360  + 1464.63948006280 *T)
    sumL0  +=       4.144 * cos(5.46982520458  + 6.90109867970 *T)
    sumL0  +=       3.691 * cos(4.07518441665  + 969.62247809490 *T)
    sumL0  +=       3.947 * cos(4.27108449197  + 181.80652604900 *T)
    sumL0  +=       3.867 * cos(5.48643386310  + 37.87240320690 *T)
    sumL0  +=       3.339 * cos(6.05372370584  + 9.40116886120 *T)
    sumL0  +=       3.484 * cos(5.81097824751  + 13.49338081870 *T)
    sumL0  +=       3.033 * cos(2.38897886651  + 221.37585028530 *T)
    sumL0  +=       2.990 * cos(4.13995939326  + 66.70484372000 *T)
    sumL0  +=       3.746 * cos(5.29902286106  + 561.18353448360 *T)
    sumL0  +=       3.233 * cos(4.27743802321  + 593.42686339800 *T)
    sumL0  +=       3.170 * cos(1.75400477770  + 235.39049596580 *T)
    sumL0  +=       4.114 * cos(2.01006788412  + 601.76425067620 *T)
    sumL0  +=       2.937 * cos(4.76351448561  + 213.55972786890 *T)
    sumL0  +=       2.932 * cos(1.83671373509  + 501.37978944330 *T)
    sumL0  +=       2.937 * cos(0.01884528825  + 213.03846300710 *T)
    sumL0  +=       3.268 * cos(4.44653949711  + 60.76695288680 *T)
    sumL0  +=       3.608 * cos(0.14307251176  + 552.58551477450 *T)
    sumL0  +=       2.947 * cos(0.74753671556  + 17.52610781830 *T)
    sumL0  +=       3.979 * cos(0.76931722276  + 424.15051032120 *T)
    sumL0  +=       2.803 * cos(1.07518176128  + 1994.33044515740 *T)
    sumL0  +=       2.905 * cos(1.27201007426  + 2737.32050569000 *T)
    sumL0  +=       3.610 * cos(0.22394084000  + 121.25202148330 *T)
    sumL0  +=       2.846 * cos(5.11748545179  + 205.66428357540 *T)
    sumL0  +=       3.156 * cos(2.74955723696  + 494.26624244250 *T)
    sumL0  +=       3.576 * cos(4.49826302447  + 167.08930495290 *T)
    sumL0  +=       2.746 * cos(0.66908290712  + 7.00167241620 *T)
    sumL0  +=       2.780 * cos(2.10066625279  + 894.84087952760 *T)
    sumL0  +=       2.875 * cos(2.39009721774  + 151.04766984290 *T)
    sumL0  +=       3.020 * cos(0.25475826890  + 40.84134862350 *T)
    sumL0  +=       2.731 * cos(3.74814908509  + 429.04587143080 *T)
    sumL0  +=       2.793 * cos(4.17938837230  + 292.01284726840 *T)
    sumL0  +=       2.706 * cos(5.34438894925  + 327.43756992050 *T)
    sumL0  +=       2.965 * cos(0.61653881148  + 643.82943957710 *T)
    sumL0  +=       2.616 * cos(4.81901387560  + 681.54178408960 *T)
    sumL0  +=       2.548 * cos(3.78162580820  + 1485.98012106520 *T)
    sumL0  +=       3.483 * cos(5.76091147029  + 141.22580985640 *T)
    sumL0  +=       3.257 * cos(0.75722680616  + 555.55446019110 *T)
    sumL0  +=       2.887 * cos(6.15899159727  + 425.63498302950 *T)
    sumL0  +=       2.450 * cos(1.29619859767  + 193.65537546500 *T)
    sumL0  +=       3.401 * cos(2.48137843009  + 43.28902917830 *T)
    sumL0  +=       3.208 * cos(0.66002842340  + 776.93031047640 *T)
    sumL0  +=       2.435 * cos(4.58097103726  + 477.80391620720 *T)
    sumL0  +=       2.577 * cos(1.41538858001  + 100.64509366380 *T)
    sumL0  +=       2.600 * cos(3.73139519973  + 17.40848773930 *T)
    sumL0  +=       2.428 * cos(1.04400815278  + 1279.79457262800 *T)
    sumL0  +=       2.569 * cos(5.36004101928  + 7.22542158540 *T)
    sumL0  +=       2.844 * cos(2.47228767650  + 280.00393915800 *T)
    sumL0  +=       2.847 * cos(1.52706408796  + 17.26547538740 *T)
    sumL0  +=       2.461 * cos(2.73899140465  + 172.24529849340 *T)
    sumL0  +=       3.228 * cos(4.10258705369  + 618.55664531160 *T)
    sumL0  +=       2.288 * cos(0.18365494079  + 426.64637498580 *T)
    sumL0  +=       2.952 * cos(3.97748947007  + 650.94298657790 *T)
    sumL0  +=       2.653 * cos(0.14255829255  + 162.89651925890 *T)
    sumL0  +=       2.291 * cos(3.26940117011  + 426.55000676620 *T)
    sumL0  +=       3.118 * cos(2.80941831445  + 2221.85663459700 *T)
    sumL0  +=       2.343 * cos(4.24349768377  + 113.38771495710 *T)
    sumL0  +=       2.780 * cos(4.36271946528  + 130.44074202340 *T)
    sumL0  +=       2.539 * cos(5.58396427573  + 381.35160823740 *T)
    sumL0  +=       2.673 * cos(2.74210116623  + 45.57665103870 *T)
    sumL0  +=       3.017 * cos(3.72208070740  + 228.27694896500 *T)
    sumL0  +=       2.781 * cos(0.36312756349  + 8.59801970910 *T)
    sumL0  +=       2.377 * cos(4.49193242045  + 25.12978191360 *T)
    sumL0  +=       2.140 * cos(5.43424670725  + 630.33605875840 *T)
    sumL0  +=       2.456 * cos(1.71617205116  + 313.68355667090 *T)
    sumL0  +=       2.071 * cos(2.40453395841  + 16.46232623530 *T)
    sumL0  +=       2.050 * cos(6.19704773331  + 3267.01147078460 *T)
    sumL0  +=       2.764 * cos(0.40107063007  + 219.44943459230 *T)
    sumL0  +=       2.307 * cos(2.61462153778  + 26.82670294300 *T)
    sumL0  +=       2.650 * cos(0.05892373791  + 5856.47765911540 *T)
    sumL0  +=       1.974 * cos(2.15890150781  + 746.92221379570 *T)
    sumL0  +=       1.949 * cos(3.13157993205  + 226.63241756230 *T)
    sumL0  +=       2.063 * cos(0.75916097286  + 472.17484191470 *T)
    sumL0  +=       2.172 * cos(1.41622302638  + 23.57587323610 *T)
    sumL0  +=       2.378 * cos(3.45446288811  + 241.87090352020 *T)
    sumL0  +=       2.314 * cos(2.92766120608  + 170.76082578510 *T)
    sumL0  +=       2.409 * cos(1.55291842382  + 112.65400177420 *T)
    sumL0  +=       2.092 * cos(4.33481587531  + 210.59078245230 *T)
    sumL0  +=       1.883 * cos(4.75777119721  + 115.62294719080 *T)
    sumL0  +=       1.963 * cos(5.63940648232  + 454.90936652730 *T)
    sumL0  +=       1.871 * cos(2.14579836453  + 135.33610313300 *T)
    sumL0  +=       2.304 * cos(0.11816226543  + 3060.82592234740 *T)
    sumL0  +=       2.221 * cos(4.34506511014  + 556.51766803760 *T)
    sumL0  +=       1.867 * cos(5.70943358261  + 19.12245511120 *T)
    sumL0  +=       2.269 * cos(3.36100653157  + 696.51963761660 *T)
    sumL0  +=       2.127 * cos(0.44754929310  + 216.00740842370 *T)
    sumL0  +=       1.807 * cos(6.15427316170  + 5.84152261360 *T)
    sumL0  +=       2.213 * cos(3.42223891884  + 533.62311835770 *T)
    sumL0  +=       1.866 * cos(3.90535444843  + 220.36445832900 *T)
    sumL0  +=       1.767 * cos(0.94232357739  + 213.45915413240 *T)
    sumL0  +=       1.767 * cos(3.84003619647  + 213.13903674360 *T)
    sumL0  +=       1.910 * cos(3.72504487558  + 104.05598206510 *T)
    sumL0  +=       1.750 * cos(0.82378244287  + 220.46082654860 *T)
    sumL0  +=       1.838 * cos(0.06310147657  + 436.15941843160 *T)
    sumL0  +=       2.146 * cos(4.41415180481  + 184.09414790940 *T)
    sumL0  +=       1.730 * cos(2.21039276178  + 416.30325013750 *T)
    sumL0  +=       1.715 * cos(0.26601715797  + 103.14095832840 *T)
    sumL0  +=       1.710 * cos(0.63515407580  + 181.05576652360 *T)
    sumL0  +=       2.307 * cos(3.29544714308  + 569.04784100980 *T)
    sumL0  +=       1.906 * cos(5.30639447218  + 405.99126305650 *T)
    sumL0  +=       1.863 * cos(4.68613642432  + 286.59622129700 *T)
    sumL0  +=       1.873 * cos(2.26516020863  + 1781.03134971940 *T)
    sumL0  +=       2.035 * cos(3.85188859267  + 672.14061522840 *T)
    sumL0  +=       2.236 * cos(3.01959133214  + 105.54045477340 *T)
    sumL0  +=       1.767 * cos(1.45800271562  + 16.67477455640 *T)
    sumL0  +=       1.633 * cos(0.16030477876  + 18.91000679010 *T)
    sumL0  +=       2.116 * cos(2.90186702031  + 486.40193591630 *T)
    sumL0  +=       2.202 * cos(3.88125957017  + 427.56139872250 *T)
    sumL0  +=       1.706 * cos(3.35213628354  + 103.04459010880 *T)
    sumL0  +=       1.604 * cos(2.48973273967  + 55.65914345610 *T)
    sumL0  +=       1.744 * cos(1.83791106739  + 1044.40407666220 *T)
    sumL0  +=       1.569 * cos(6.10089581118  + 106.01353552540 *T)
    sumL0  +=       2.081 * cos(6.03810192844  + 916.93228005540 *T)
    sumL0  +=       1.799 * cos(5.01592570405  + 731.94436026870 *T)
    sumL0  +=       1.737 * cos(1.49651330833  + 25.86349509650 *T)
    sumL0  +=       1.695 * cos(3.53314158403  + 627.36711334180 *T)
    sumL0  +=       1.543 * cos(0.81384993001  + 2310.72231481400 *T)
    sumL0  +=       1.896 * cos(3.38169845451  + 2324.94940881560 *T)
    sumL0  +=       1.926 * cos(4.66519027283  + 353.30106501700 *T)
    sumL0  +=       1.765 * cos(5.14740716994  + 107.28555991260 *T)
    sumL0  +=       1.556 * cos(1.12431826916  + 230.82520325630 *T)
    sumL0  +=       1.843 * cos(0.02435960281  + 102.12956637210 *T)
    sumL0  +=       1.501 * cos(4.18415120927  + 194.17664032680 *T)
    sumL0  +=       1.528 * cos(1.00328674046  + 3053.71237534660 *T)
    sumL0  +=       1.529 * cos(5.58893570479  + 212.02707105080 *T)
    sumL0  +=       1.684 * cos(5.08547245125  + 3480.31056622260 *T)
    sumL0  +=       1.461 * cos(2.31020597821  + 721.64941953020 *T)
    sumL0  +=       1.480 * cos(5.34331643017  + 418.52143602870 *T)
    sumL0  +=       1.601 * cos(5.53623000915  + 391.17346822390 *T)
    sumL0  +=       1.893 * cos(3.62340803433  + 204.70107572890 *T)
    sumL0  +=       1.529 * cos(6.06535432009  + 77.96299230500 *T)
    sumL0  +=       1.529 * cos(5.47660937625  + 214.57111982520 *T)
    sumL0  +=       1.552 * cos(2.06693539836  + 36.64856292950 *T)
    sumL0  +=       1.453 * cos(6.04709831442  + 165.60483224460 *T)
    sumL0  +=       1.393 * cos(2.28253369060  + 403.02231763990 *T)
    sumL0  +=       1.444 * cos(2.90650214018  + 447.93883187840 *T)
    sumL0  +=       1.924 * cos(1.37028714759  + 468.24268865160 *T)
    sumL0  +=       1.426 * cos(0.13255011458  + 2207.62954059540 *T)
    sumL0  +=       1.389 * cos(2.21739183113  + 643.07868005170 *T)
    sumL0  +=       1.365 * cos(1.63853880518  + 629.60234557550 *T)
    sumL0  +=       1.362 * cos(3.35131049142  + 93.53154666300 *T)
    sumL0  +=       1.376 * cos(5.36989538450  + 180.16199464630 *T)
    sumL0  +=       1.584 * cos(0.85642767335  + 271.40591944890 *T)
    sumL0  +=       1.405 * cos(5.69231057947  + 25.27279426550 *T)
    sumL0  +=       1.681 * cos(5.30308110734  + 835.03713448730 *T)
    sumL0  +=       1.598 * cos(3.04233449432  + 42.53826965290 *T)
    sumL0  +=       1.759 * cos(3.59043066940  + 508.35032409220 *T)
    sumL0  +=       1.394 * cos(4.55070863290  + 426.07692601420 *T)
    sumL0  +=       1.314 * cos(1.81147178081  + 1382.88734684660 *T)
    sumL0  +=       1.281 * cos(4.26508388040  + 123.53964334370 *T)
    sumL0  +=       1.742 * cos(5.71133189432  + 22.89454967990 *T)
    sumL0  +=       1.483 * cos(1.84687831602  + 289.56516671360 *T)
    sumL0  +=       1.257 * cos(3.01131200921  + 409.92341631960 *T)
    sumL0  +=       1.285 * cos(4.41168551011  + 558.00214074590 *T)
    sumL0  +=       1.355 * cos(3.87115897452  + 1802.37199072180 *T)
    sumL0  +=       1.333 * cos(0.08474224795  + 411.62033734900 *T)
    sumL0  +=       1.235 * cos(4.08060394635  + 28.45418800320 *T)
    sumL0  +=       1.373 * cos(5.06955106471  + 427.11945573780 *T)
    sumL0  +=       1.565 * cos(2.32953532704  + 41.05379694460 *T)
    sumL0  +=       1.656 * cos(6.06169130804  + 268.43697403230 *T)
    sumL0  +=       1.212 * cos(3.05966957556  + 420.96911658350 *T)
    sumL0  +=       1.238 * cos(5.25936700679  + 412.58354519550 *T)
    sumL0  +=       1.220 * cos(3.92987038126  + 2.66012887590 *T)
    sumL0  +=       1.552 * cos(1.48184004773  + 9786.68735533500 *T)
    sumL0  +=       1.240 * cos(1.46716327302  + 291.26208774300 *T)
    sumL0  +=       1.133 * cos(5.39046583617  + 423.67742956920 *T)
    sumL0  +=       1.319 * cos(5.79905891015  + 1108.13997496560 *T)
    sumL0  +=       1.329 * cos(0.92291650117  + 778.41478318470 *T)
    sumL0  +=       1.399 * cos(2.55906860098  + 421.93232443000 *T)
    sumL0  +=       1.120 * cos(3.86777259232  + 1033.35837639830 *T)
    sumL0  +=       1.164 * cos(4.10048660918  + 685.47393735270 *T)
    sumL0  +=       1.321 * cos(1.45843550806  + 1073.60902419080 *T)
    sumL0  +=       1.313 * cos(0.11761534168  + 71.81265315070 *T)
    sumL0  +=       1.438 * cos(2.57741975416  + 100.38446123290 *T)
    sumL0  +=       1.190 * cos(5.63379509659  + 5.10780943070 *T)
    sumL0  +=       1.289 * cos(5.20604565993  + 278.25883401880 *T)
    sumL0  +=       1.157 * cos(5.00101860101  + 230.56457082540 *T)
    sumL0  +=       1.233 * cos(2.70207317014  + 282.66406803390 *T)
    sumL0  +=       1.209 * cos(4.02230498135  + 980.66817835880 *T)
    sumL0  +=       1.070 * cos(5.17569455055  + 313.21047591890 *T)
    sumL0  +=       1.292 * cos(4.30946655209  + 219.89137757700 *T)
    sumL0  +=       1.399 * cos(2.58476795858  + 2538.24850425360 *T)
    sumL0  +=       1.038 * cos(0.14212199680  + 820.05928096030 *T)
    sumL0  +=       1.245 * cos(4.08278897130  + 525.75881183150 *T)
    sumL0  +=       1.254 * cos(2.46275017735  + 457.61767951300 *T)
    sumL0  +=       1.021 * cos(1.11239421009  + 69.36497259590 *T)
    sumL0  +=       1.009 * cos(1.01709171385  + 143.93412284210 *T)
    sumL0  +=       1.075 * cos(2.39196853318  + 48.75804477640 *T)
    sumL0  +=       1.180 * cos(6.18938910429  + 3377.21779200400 *T)
    sumL0  +=       0.989 * cos(5.94928603657  + 3583.40334044120 *T)
    sumL0  +=       0.972 * cos(4.25434114756  + 397.39324334740 *T)
    sumL0  +=       0.983 * cos(0.04442608551  + 140.96517742550 *T)
    sumL0  +=       0.972 * cos(5.67683107883  + 422.40540518200 *T)
    sumL0  +=       1.298 * cos(1.34524469231  + 875.83029900100 *T)
    sumL0  +=       1.190 * cos(0.67933974618  + 699.70103135430 *T)
    sumL0  +=       0.950 * cos(2.66964340700  + 92.30770638560 *T)
    sumL0  +=       0.933 * cos(0.63000656580  + 406.95447090300 *T)
    sumL0  +=       0.959 * cos(1.77556884452  + 67.66805156650 *T)
    sumL0  +=       1.185 * cos(3.70140604185  + 285.63301345050 *T)
    sumL0  +=       0.956 * cos(5.18928530992  + 319.31263096340 *T)
    sumL0  +=       1.014 * cos(1.97449310063  + 2097.42321937600 *T)
    sumL0  +=       1.048 * cos(3.69659410655  + 117.91056905120 *T)
    sumL0  +=       1.153 * cos(2.53320305623  + 104.57724692690 *T)
    sumL0  +=       1.258 * cos(2.51536062507  + 694.07195706180 *T)
    sumL0  +=       0.971 * cos(5.19147635849  + 240.12579838100 *T)
    sumL0  +=       0.940 * cos(3.94701776697  + 35.21227433100 *T)
    sumL0  +=       1.047 * cos(6.12360979460  + 238.90195810360 *T)
    sumL0  +=       1.185 * cos(5.28289734361  + 638.41281360570 *T)
    sumL0  +=       0.893 * cos(0.95364488395  + 14.81779483260 *T)
    sumL0  +=       1.094 * cos(3.93009679240  + 945.24345570670 *T)
    sumL0  +=       0.949 * cos(3.46451925897  + 443.86366626340 *T)
    sumL0  +=       1.002 * cos(3.18639902867  + 337.73251065900 *T)
    sumL0  +=       1.017 * cos(2.87111101661  + 211.60217440860 *T)
    sumL0  +=       0.875 * cos(0.58638080067  + 2.28762186040 *T)
    sumL0  +=       0.925 * cos(1.54981519784  + 19.64371997300 *T)
    sumL0  +=       1.152 * cos(1.68528608590  + 691.10301164520 *T)
    sumL0  +=       0.832 * cos(2.64637256467  + 436.89313161450 *T)
    sumL0  +=       0.834 * cos(3.88913521570  + 331.20966448920 *T)
    sumL0  +=       0.825 * cos(2.15437872210  + 739.80866679490 *T)
    sumL0  +=       0.848 * cos(3.18263239100  + 196.03362005060 *T)
    sumL0  +=       1.044 * cos(3.87842686803  + 532.61172640140 *T)
    sumL0  +=       0.846 * cos(5.38853773752  + 97.67614824720 *T)
    sumL0  +=       1.021 * cos(2.93075488512  + 184.98791978670 *T)
    sumL0  +=       0.843 * cos(6.12012061227  + 616.32141307790 *T)
    sumL0  +=       0.820 * cos(1.01380400969  + 480.77286162380 *T)
    sumL0  +=       0.842 * cos(3.56523575381  + 421.18156490460 *T)
    sumL0  +=       1.076 * cos(3.28234305253  + 5.67725840230 *T)
    sumL0  +=       0.808 * cos(2.78227865672  + 212.07525516060 *T)
    sumL0  +=       0.812 * cos(0.94281737163  + 108.72184851110 *T)
    sumL0  +=       0.808 * cos(2.21202653278  + 610.69233878540 *T)
    sumL0  +=       0.808 * cos(2.00008111713  + 214.52293571540 *T)
    sumL0  +=       0.875 * cos(2.14897461363  + 114.39910691340 *T)
    sumL0  +=       0.791 * cos(2.35474255596  + 1.37259812370 *T)
    sumL0  +=       0.960 * cos(0.25496742364  + 710.74673161820 *T)
    sumL0  +=       1.001 * cos(2.34471240227  + 16.04163511000 *T)
    sumL0  +=       0.994 * cos(0.63700664871  + 84.93352695390 *T)
    sumL0  +=       0.985 * cos(2.63664920104  + 395.57870223900 *T)
    sumL0  +=       0.874 * cos(2.59112594967  + 418.00017116690 *T)
    sumL0  +=       0.758 * cos(0.58117487362  + 2627.11418447060 *T)
    sumL0  +=       0.894 * cos(1.48976897396  + 760.25553592000 *T)
    sumL0  +=       0.768 * cos(5.25095392845  + 305.08553696180 *T)
    sumL0  +=       1.020 * cos(2.73153988233  + 268.95823889410 *T)
    sumL0  +=       0.903 * cos(0.13224671457  + 238.42887735160 *T)
    sumL0  +=       0.750 * cos(0.76128043194  + 724.83081326790 *T)
    sumL0  +=       0.903 * cos(3.37105323370  + 526.50957135690 *T)
    sumL0  +=       0.930 * cos(2.83622594110  + 2641.34127847220 *T)
    sumL0  +=       0.808 * cos(5.23759255053  + 216.26804085460 *T)
    sumL0  +=       0.864 * cos(1.22059443823  + 570.74476203920 *T)
    sumL0  +=       0.798 * cos(3.72388187653  + 124.50285119020 *T)
    sumL0  +=       0.753 * cos(0.72747041757  + 3370.10424500320 *T)
    sumL0  +=       0.844 * cos(2.03251767810  + 511.53171782990 *T)
    sumL0  +=       0.827 * cos(4.49936223096  + 444.75743814070 *T)
    sumL0  +=       0.848 * cos(3.74330244183  + 2118.76386037840 *T)
    sumL0  +=       0.795 * cos(0.27939057139  + 101.86893394120 *T)
    sumL0  +=       0.754 * cos(5.68583497533  + 662.53120356300 *T)
    sumL0  +=       0.750 * cos(4.78778128003  + 102.57150935680 *T)
    sumL0  +=       0.709 * cos(1.64518562815  + 159.12442469020 *T)
    sumL0  +=       0.770 * cos(4.02404991950  + 909.81873305460 *T)
    sumL0  +=       0.765 * cos(1.62693133597  + 465.95506679120 *T)
    sumL0  +=       0.911 * cos(5.04635658282  + 913.96333463880 *T)
    sumL0  +=       0.861 * cos(2.78971410809  + 495.75071515080 *T)
    sumL0  +=       0.688 * cos(1.10207467005  + 1.53686233500 *T)
    sumL0  +=       0.803 * cos(4.33043919090  + 453.42489381900 *T)
    sumL0  +=       0.673 * cos(0.03439333853  + 2524.02141025200 *T)
    sumL0  +=       0.856 * cos(3.50639182375  + 439.12836384820 *T)
    sumL0  +=       0.684 * cos(3.93906807606  + 337.80194662820 *T)
    sumL0  +=       0.716 * cos(6.18909854987  + 310.71461125430 *T)
    sumL0  +=       0.922 * cos(1.70634200320  + 125.18417474640 *T)
    sumL0  +=       0.885 * cos(1.69955870744  + 6283.07584999140 *T)
    sumL0  +=       0.656 * cos(1.28102954508  + 432.01481684740 *T)
    sumL0  +=       0.808 * cos(1.64410808383  + 299.12639426920 *T)
    sumL0  +=       0.656 * cos(1.49449164620  + 849.26422848890 *T)
    sumL0  +=       0.679 * cos(4.02962984490  + 429.51895218280 *T)
    sumL0  +=       0.854 * cos(3.04068731741  + 298.23262239190 *T)
    sumL0  +=       0.676 * cos(2.17631477883  + 576.16138801060 *T)
    sumL0  +=       0.881 * cos(5.47733557925  + 220.93390730060 *T)
    sumL0  +=       0.739 * cos(5.85330901725  + 938.12990870590 *T)
    sumL0  +=       0.637 * cos(2.96294462433  + 425.84743135060 *T)
    sumL0  +=       0.665 * cos(1.62998758015  + 221.16340196420 *T)
    sumL0  +=       0.693 * cos(3.53871697600  + 1182.92157353290 *T)
    sumL0  +=       0.659 * cos(1.85319023888  + 72.33391801250 *T)
    sumL0  +=       0.631 * cos(2.01234919422  + 58.10682401090 *T)
    sumL0  +=       0.806 * cos(5.21763933753  + 428.08266358430 *T)
    sumL0  +=       0.804 * cos(5.94281804567  + 26.02355379090 *T)
    sumL0  +=       0.679 * cos(2.11784460940  + 256.42806592190 *T)
    sumL0  +=       0.692 * cos(1.88791537515  + 214.99601646740 *T)
    sumL0  +=       0.731 * cos(1.95762888351  + 19.01058052660 *T)
    sumL0  +=       0.786 * cos(0.91252523635  + 518.38463239980 *T)
    sumL0  +=       0.684 * cos(4.89288171806  + 3796.70243587920 *T)
    sumL0  +=       0.612 * cos(2.08511492036  + 1038.04128918680 *T)
    sumL0  +=       0.598 * cos(3.48814927085  + 219.66188291340 *T)
    sumL0  +=       0.828 * cos(0.31689472200  + 25.60286266560 *T)
    sumL0  +=       0.773 * cos(4.57524006328  + 624.91943278700 *T)
    sumL0  +=       0.775 * cos(6.12920077021  + 432.22726516850 *T)
    sumL0  +=       0.820 * cos(4.11320326130  + 141.48644228730 *T)
    sumL0  +=       0.588 * cos(1.95775535412  + 211.86280683950 *T)
    sumL0  +=       0.661 * cos(5.30100397707  + 103.61403908040 *T)
    sumL0  +=       0.588 * cos(2.82460441973  + 214.73538403650 *T)
    sumL0  +=       0.651 * cos(5.56732715834  + 393.46109008430 *T)
    sumL0  +=       0.564 * cos(4.01666572198  + 850.01498801430 *T)
    sumL0  +=       0.657 * cos(2.58166087726  + 526.98265210890 *T)
    sumL0  +=       0.722 * cos(0.68494219480  + 953.10776223290 *T)
    sumL0  +=       0.592 * cos(2.37190662490  + 205.43478891180 *T)
    sumL0  +=       0.738 * cos(1.07981019512  + 239.16259053450 *T)
    sumL0  +=       0.590 * cos(6.03587790219  + 188.02630117250 *T)
    sumL0  +=       0.559 * cos(5.76010635813  + 430.79097657000 *T)
    sumL0  +=       0.535 * cos(5.80499883199  + 100.17201291180 *T)
    sumL0  +=       0.611 * cos(5.95439360100  + 3693.60966166060 *T)
    sumL0  +=       0.591 * cos(4.22379888536  + 505.31194270640 *T)
    sumL0  +=       0.691 * cos(2.96568305933  + 606.76018552230 *T)
    sumL0  +=       0.648 * cos(2.33387623043  + 30.75885620610 *T)
    sumL0  +=       0.544 * cos(2.21686115865  + 92.79783348010 *T)
    sumL0  +=       0.517 * cos(2.68282421083  + 262.05714021440 *T)
    sumL0  +=       0.563 * cos(1.69735688719  + 2413.81508903260 *T)
    sumL0  +=       0.569 * cos(2.85075508949  + 227.31374111850 *T)
    sumL0  +=       0.531 * cos(2.17713708433  + 263.02034806090 *T)
    sumL0  +=       0.504 * cos(4.44947885193  + 343.73983746140 *T)
    sumL0  +=       0.578 * cos(3.31462999801  + 33.72780162270 *T)
    sumL0  +=       0.619 * cos(1.83409636127  + 867.42347575360 *T)
    sumL0  +=       0.546 * cos(4.82201187196  + 1048.33622992530 *T)
    sumL0  +=       0.517 * cos(1.42016110098  + 1246.65747183630 *T)
    sumL0  +=       0.681 * cos(1.94124532036  + 25874.60404613620 *T)
    sumL0  +=       0.551 * cos(5.70617358907  + 1119.18567522950 *T)
    sumL0  +=       0.523 * cos(5.78878978812  + 366.79444583570 *T)
    sumL0  +=       0.486 * cos(1.90063955671  + 1063.31408345230 *T)
    sumL0  +=       0.552 * cos(3.64325031166  + 256.58812461630 *T)
    sumL0  +=       0.612 * cos(2.39349965241  + 2854.64037391020 *T)
    sumL0  +=       0.495 * cos(3.46833581240  + 597.35901666110 *T)
    sumL0  +=       0.622 * cos(1.86539391351  + 524.01370669230 *T)
    sumL0  +=       0.480 * cos(5.33557742428  + 29.20494752860 *T)
    sumL0  +=       0.492 * cos(4.64109549618  + 384.05992122310 *T)
    sumL0  +=       0.520 * cos(2.32171681836  + 2957.73314812880 *T)
    sumL0  +=       0.545 * cos(0.53274778710  + 431.26405732200 *T)
    sumL0  +=       0.479 * cos(2.13325177240  + 319.57326339430 *T)
    sumL0  +=       0.526 * cos(4.17771910249  + 136.06981631590 *T)
    sumL0  +=       0.612 * cos(4.56148986681  + 774.48262992160 *T)
    sumL0  +=       0.642 * cos(3.25195912708  + 67.88049988760 *T)
    sumL0  +=       0.527 * cos(4.00299045889  + 2435.15573003500 *T)
    sumL0  +=       0.524 * cos(4.69817741494  + 336.83873878170 *T)
    sumL0  +=       0.495 * cos(5.95703962647  + 765.88461021250 *T)
    sumL0  +=       0.463 * cos(6.09780322600  + 54.33472944220 *T)
    sumL0  +=       0.538 * cos(0.22116216124  + 450.97721326420 *T)
    sumL0  +=       0.465 * cos(1.87487942848  + 958.57677783100 *T)
    sumL0  +=       0.500 * cos(1.54084756342  + 572.22923474750 *T)
    sumL0  +=       0.528 * cos(3.54764543325  + 233.90602325750 *T)
    sumL0  +=       0.451 * cos(5.72664397006  + 3899.79521009780 *T)
    sumL0  +=       0.514 * cos(5.00509609437  + 273.85360000370 *T)
    sumL0  +=       0.471 * cos(2.96871670899  + 306.83064210100 *T)
    sumL0  +=       0.447 * cos(1.87279400375  + 62.03897727400 *T)
    sumL0  +=       0.456 * cos(5.45521998520  + 1171.87587326900 *T)
    sumL0  +=       0.499 * cos(1.92821778530  + 217.44369702220 *T)
    sumL0  +=       0.548 * cos(3.21873307934  + 824.74219374880 *T)
    sumL0  +=       0.579 * cos(2.29424247924  + 810.65811209910 *T)
    sumL0  +=       0.550 * cos(0.67875196590  + 315.16802937920 *T)
    sumL0  +=       0.556 * cos(1.30277646899  + 133.10087089930 *T)
    sumL0  +=       0.449 * cos(6.05954557724  + 141.69889060840 *T)
    sumL0  +=       0.447 * cos(5.83737433218  + 823.99143422340 *T)
    sumL0  +=       0.517 * cos(3.62832879827  + 934.94851496820 *T)
    sumL0  +=       0.482 * cos(1.04855231956  + 1055.44977692610 *T)
    sumL0  +=       0.482 * cos(4.07207792722  + 195.89060769870 *T)
    sumL0  +=       0.428 * cos(0.65142455407  + 427.34895040140 *T)
    sumL0  +=       0.585 * cos(5.48406138684  + 376.19561469690 *T)
    sumL0  +=       0.469 * cos(3.87344294455  + 320.58465535060 *T)
    sumL0  +=       0.488 * cos(2.83523964260  + 460.53844081980 *T)
    sumL0  +=       0.450 * cos(4.30419652064  + 88.11492069160 *T)
    sumL0  +=       0.537 * cos(0.85582040238  + 214.19286731530 *T)
    sumL0  +=       0.438 * cos(1.32216133929  + 963.40270297140 *T)
    sumL0  +=       0.560 * cos(2.50374228728  + 952.09637027660 *T)
    sumL0  +=       0.442 * cos(2.80002649649  + 209.15449385380 *T)
    sumL0  +=       0.443 * cos(4.31062007978  + 9992.87290377220 *T)
    sumL0  +=       0.469 * cos(0.45235276525  + 464.73122651380 *T)
    sumL0  +=       0.488 * cos(0.35817443686  + 36.90919536040 *T)
    sumL0  +=       0.418 * cos(4.81080887468  + 775.23338944700 *T)
    sumL0  +=       0.417 * cos(4.93943593579  + 306.09692891810 *T)
    sumL0  +=       0.483 * cos(3.92847922420  + 39.61750834610 *T)
    sumL0  +=       0.467 * cos(1.89153069484  + 30.05628079050 *T)
    sumL0  +=       0.410 * cos(5.52148731635  + 118.07062774560 *T)
    sumL0  +=       0.406 * cos(1.35991757653  + 945.99421523210 *T)
    sumL0  +=       0.446 * cos(4.06656112480  + 380.38840039090 *T)
    sumL0  +=       0.466 * cos(3.65954736664  + 988.53248488500 *T)
    sumL0  +=       0.418 * cos(1.40185532806  + 313.94418910180 *T)
    sumL0  +=       0.481 * cos(1.80873987903  + 43.12897048390 *T)
    sumL0  +=       0.437 * cos(0.86746182155  + 170.97327410620 *T)
    sumL0  +=       0.483 * cos(4.49894122772  + 46.20979048510 *T)
    sumL0  +=       0.398 * cos(2.90977731924  + 131.54696222180 *T)
    sumL0  +=       0.529 * cos(3.74604329884  + 699.17976649250 *T)
    sumL0  +=       0.396 * cos(0.34033987778  + 2943.50605412720 *T)
    sumL0  +=       0.545 * cos(2.97400965609  + 305.60680182360 *T)
    sumL0  +=       0.412 * cos(3.81935995126  + 84.34282612290 *T)
    sumL0  +=       0.425 * cos(2.60672101181  + 121.84272231430 *T)
    sumL0  +=       0.474 * cos(2.41769418569  + 838.21852822500 *T)
    sumL0  +=       0.457 * cos(1.27246488727  + 107.75864066460 *T)
    sumL0  +=       0.519 * cos(3.12247974037  + 10213.28554621100 *T)
    sumL0  +=       0.495 * cos(4.63705386984  + 301.41401612960 *T)
    sumL0  +=       0.537 * cos(3.92653937147  + 212.40532356070 *T)
    sumL0  +=       0.385 * cos(3.33476325050  + 806.72595883600 *T)
    sumL0  +=       0.477 * cos(1.66001855277  + 175.42669223110 *T)
    sumL0  +=       0.378 * cos(0.47722247042  + 200.55647414470 *T)
    sumL0  +=       0.459 * cos(5.14821844812  + 960.22130923370 *T)
    sumL0  +=       0.401 * cos(4.36420932544  + 739.05790726950 *T)
    sumL0  +=       0.467 * cos(2.96423984091  + 170.01006625970 *T)
    sumL0  +=       0.457 * cos(4.45829983338  + 33.13710079170 *T)
    sumL0  +=       0.476 * cos(3.63257697985  + 20.49505323490 *T)
    sumL0  +=       0.462 * cos(3.57494442856  + 71.60020482960 *T)
    sumL0  +=       0.376 * cos(2.94770389653  + 6062.66320755260 *T)
    sumL0  +=       0.473 * cos(5.32759318114  + 373.90799283650 *T)
    sumL0  +=       0.383 * cos(1.84111991862  + 346.39996633730 *T)
    sumL0  +=       0.366 * cos(2.05039240297  + 87.31177153950 *T)
    sumL0  +=       0.383 * cos(2.00608370504  + 3274.12501778540 *T)
    sumL0  +=       0.374 * cos(5.65444305031  + 540.73666535850 *T)
    sumL0  +=       0.410 * cos(5.62629715297  + 58.31927233200 *T)
    sumL0  +=       0.441 * cos(6.26992749567  + 378.90392768260 *T)
    sumL0  +=       0.451 * cos(3.05920369737  + 898.77303279070 *T)
    sumL0  +=       0.371 * cos(5.65787287655  + 89.75945209430 *T)
    sumL0  +=       0.367 * cos(5.71998148487  + 96.87299909510 *T)
    sumL0  +=       0.410 * cos(1.06290837961  + 1257.70317210020 *T)
    sumL0  +=       0.418 * cos(2.03053179312  + 146.59425171800 *T)
    sumL0  +=       0.492 * cos(0.89529424356  + 423.62924545940 *T)
    sumL0  +=       0.474 * cos(2.43080822444  + 705.11765732570 *T)
    sumL0  +=       0.457 * cos(3.42347802916  + 829.62050851590 *T)
    sumL0  +=       0.419 * cos(0.32183412086  + 90.56260124640 *T)
    sumL0  +=       0.347 * cos(3.60116807440  + 449.23210812500 *T)
    sumL0  +=       0.397 * cos(2.29159647723  + 782.34693644780 *T)
    sumL0  +=       0.433 * cos(3.40938468811  + 32.45577723550 *T)
    sumL0  +=       0.343 * cos(4.30952656038  + 401.32539661050 *T)
    sumL0  +=       0.351 * cos(0.42160026295  + 3686.49611465980 *T)
    sumL0  +=       0.360 * cos(5.83204569879  + 491.55792945680 *T)
    sumL0  +=       0.389 * cos(2.73429108352  + 36.17548217750 *T)
    sumL0  +=       0.434 * cos(0.33664392528  + 55.13787859430 *T)
    sumL0  +=       0.346 * cos(5.09191837323  + 392.65794093220 *T)
    sumL0  +=       0.336 * cos(2.38696934868  + 295.19424100610 *T)
    sumL0  +=       0.336 * cos(1.56832822829  + 233.74596456310 *T)
    sumL0  +=       0.400 * cos(3.08774286568  + 745.91082183940 *T)
    sumL0  +=       0.468 * cos(4.21298903633  + 832.58945393250 *T)
    sumL0  +=       0.400 * cos(1.32415028336  + 551.10104206620 *T)
    sumL0  +=       0.344 * cos(1.52193307438  + 754.83890994860 *T)
    sumL0  +=       0.433 * cos(3.06696455642  + 885.43971066640 *T)
    sumL0  +=       0.346 * cos(4.76971433870  + 4113.09430553580 *T)
    sumL0  +=       0.329 * cos(3.31034285904  + 952.35700270750 *T)
    sumL0  +=       0.425 * cos(2.90590905341  + 462.02291352810 *T)
    sumL0  +=       0.342 * cos(5.87738962422  + 561.93429400900 *T)
    sumL0  +=       0.439 * cos(0.20791179302  + 768.85355562910 *T)
    sumL0  +=       0.330 * cos(2.08599193524  + 614.62449204850 *T)
    sumL0  +=       0.394 * cos(3.59805431851  + 1261.63532536330 *T)
    sumL0  +=       0.369 * cos(6.03075127328  + 199.80571461930 *T)
    sumL0  +=       0.375 * cos(2.56075851331  + 732.69511979410 *T)
    sumL0  +=       0.408 * cos(0.69233617038  + 328.24071907260 *T)
    sumL0  +=       0.331 * cos(4.28333309370  + 541.53981451060 *T)
    sumL0  +=       0.323 * cos(0.06694179304  + 433.75992198660 *T)
    sumL0  +=       0.431 * cos(4.50603340920  + 2914.01423582380 *T)
    sumL0  +=       0.343 * cos(2.18898146246  + 80.19822453870 *T)
    sumL0  +=       0.437 * cos(5.34124436008  + 387.24131496080 *T)
    sumL0  +=       0.371 * cos(4.82569229712  + 103.35340664950 *T)
    sumL0  +=       0.318 * cos(3.75720880396  + 749.20983565610 *T)
    sumL0  +=       0.344 * cos(0.70749566894  + 229.97386999440 *T)
    sumL0  +=       0.312 * cos(6.02741985422  + 361.37781986430 *T)
    sumL0  +=       0.340 * cos(0.64358874512  + 303.86169668440 *T)
    sumL0  +=       0.394 * cos(0.49801994278  + 248.46318565920 *T)
    sumL0  +=       0.309 * cos(5.83535487659  + 236.19364511790 *T)
    sumL0  +=       0.325 * cos(2.65292455786  + 757.21715453420 *T)
    sumL0  +=       0.365 * cos(0.56788400592  + 402.21916848780 *T)
    sumL0  +=       0.342 * cos(3.83450185886  + 519.39602435610 *T)
    sumL0  +=       0.306 * cos(0.35126074897  + 354.99798604640 *T)
    sumL0  +=       0.314 * cos(5.42086935152  + 1151.42900414390 *T)
    sumL0  +=       0.384 * cos(0.09367760642  + 201.51968199120 *T)
    sumL0  +=       0.307 * cos(5.62165090342  + 426.48631629140 *T)
    sumL0  +=       0.301 * cos(1.78550205039  + 1354.43315884340 *T)
    sumL0  +=       0.300 * cos(5.23426324539  + 190.40454575810 *T)
    sumL0  +=       0.296 * cos(3.15801280224  + 192.85222631290 *T)
    sumL0  +=       0.381 * cos(1.98600461808  + 109.94568878850 *T)
    sumL0  +=       0.370 * cos(5.57659190517  + 562.14674233010 *T)
    sumL0  +=       0.305 * cos(2.73187068494  + 840.66620877980 *T)
    sumL0  +=       0.374 * cos(3.73375594662  + 420.44785172170 *T)
    sumL0  +=       0.307 * cos(4.03149340189  + 426.71006546060 *T)
    sumL0  +=       0.320 * cos(1.42665180100  + 2730.20695868920 *T)
    sumL0  +=       0.392 * cos(0.14343294900  + 206.39799675830 *T)
    sumL0  +=       0.288 * cos(2.44787565138  + 623.22251175760 *T)
    sumL0  +=       0.344 * cos(1.57825843162  + 6290.18939699220 *T)
    sumL0  +=       0.317 * cos(1.84490263693  + 214.94362684070 *T)
    sumL0  +=       0.346 * cos(2.16849893508  + 3171.03224356680 *T)
    sumL0  +=       0.286 * cos(1.01623455966  + 315.64111013120 *T)
    sumL0  +=       0.361 * cos(3.44807605339  + 259.76951835400 *T)
    sumL0  +=       0.280 * cos(4.64272946177  + 254.14044406150 *T)
    sumL0  +=       0.285 * cos(2.09430258387  + 335.14181775230 *T)
    sumL0  +=       0.281 * cos(4.72136141979  + 317.14262918200 *T)
    sumL0  +=       0.289 * cos(4.70265740009  + 29.74746424980 *T)
    sumL0  +=       0.290 * cos(2.74360609522  + 551.03160609700 *T)
    sumL0  +=       0.300 * cos(5.23308974173  + 1321.43907040360 *T)
    sumL0  +=       0.283 * cos(1.21193797828  + 1699.27921650320 *T)
    sumL0  +=       0.280 * cos(0.45877292284  + 38.60611638980 *T)
    sumL0  +=       0.273 * cos(1.81952809851  + 1056.20053645150 *T)
    sumL0  +=       0.336 * cos(5.63115066542  + 95.22846769240 *T)
    sumL0  +=       0.309 * cos(3.33676739908  + 1193.96727379680 *T)
    sumL0  +=       0.288 * cos(2.57603349418  + 1166.40685767090 *T)
    sumL0  +=       0.277 * cos(3.35359995343  + 532.87235883230 *T)
    sumL0  +=       0.287 * cos(0.37229115993  + 114.94162363460 *T)
    sumL0  +=       0.274 * cos(4.21164027953  + 90.82323367730 *T)
    sumL0  +=       0.267 * cos(5.61367132586  + 870.46185713940 *T)
    sumL0  +=       0.376 * cos(6.14391456675  + 913.00012679230 *T)
    sumL0  +=       0.296 * cos(5.75705165982  + 4010.00153131720 *T)
    sumL0  +=       0.304 * cos(1.97982468634  + 495.96316347190 *T)
    sumL0  +=       0.308 * cos(5.08911712262  + 481.73606947030 *T)
    sumL0  +=       0.265 * cos(0.02893016561  + 172.45774681450 *T)
    sumL0  +=       0.291 * cos(2.10106037916  + 619.29035849450 *T)
    sumL0  +=       0.273 * cos(4.78343050572  + 771.30123618390 *T)
    sumL0  +=       0.358 * cos(0.04093867464  + 637.44960575920 *T)
    sumL0  +=       0.272 * cos(5.86505586319  + 332.17287233570 *T)
    sumL0  +=       0.266 * cos(3.28626000731  + 560.71045373160 *T)
    sumL0  +=       0.312 * cos(3.91213951915  + 1226.21060271120 *T)
    sumL0  +=       0.258 * cos(3.41185135958  + 426.81063919710 *T)
    sumL0  +=       0.257 * cos(1.06772454181  + 714.67888488130 *T)
    sumL0  +=       0.258 * cos(6.24129294573  + 426.38574255490 *T)
    sumL0  +=       0.256 * cos(0.75289096697  + 103.84353374400 *T)
    sumL0  +=       0.254 * cos(3.21116652124  + 102.34201469320 *T)
    sumL0  +=       0.254 * cos(6.14632777985  + 620.25356634100 *T)
    sumL0  +=       0.266 * cos(2.54280196709  + 132.88842257820 *T)
    sumL0  +=       0.314 * cos(0.31303204249  + 991.71387862270 *T)
    sumL0  +=       0.317 * cos(2.93589163442  + 357.23321828010 *T)
    sumL0  +=       0.266 * cos(6.12280636670  + 57.51612317990 *T)
    sumL0  +=       0.254 * cos(2.99730079627  + 642.34496686880 *T)
    sumL0  +=       0.267 * cos(5.54663413439  + 628.59095361920 *T)
    sumL0  +=       0.348 * cos(1.33319249154  + 815.06334611420 *T)
    sumL0  +=       0.278 * cos(5.59573741920  + 334.55111692130 *T)
    sumL0  +=       0.303 * cos(3.22789148979  + 409.18970313670 *T)
    sumL0  +=       0.246 * cos(3.90430259983  + 441.57604440300 *T)
    sumL0  +=       0.260 * cos(3.86355293530  + 639.37602145220 *T)
    sumL0  +=       0.250 * cos(0.07635434166  + 2840.41327990860 *T)
    sumL0  +=       0.246 * cos(5.71009371698  + 476.31944349890 *T)
    sumL0  +=       0.301 * cos(6.15272106984  + 559.69906177530 *T)
    sumL0  +=       0.268 * cos(3.73809606804  + 658.05653357870 *T)
    sumL0  +=       0.316 * cos(4.63035287047  + 745.27768239300 *T)
    sumL0  +=       0.311 * cos(3.51827727424  + 2751.54759969160 *T)
    sumL0  +=       0.239 * cos(0.51133317457  + 1041.22268292450 *T)
    sumL0  +=       0.238 * cos(5.46347279420  + 4216.18707975440 *T)
    sumL0  +=       0.262 * cos(3.20254407166  + 1251.34038462480 *T)
    sumL0  +=       0.238 * cos(1.02679111743  + 1262.38608488870 *T)
    sumL0  +=       0.277 * cos(3.32996713394  + 545.47196777370 *T)
    sumL0  +=       0.271 * cos(0.49229846069  + 419.53282798500 *T)
    sumL0  +=       0.303 * cos(3.77087511317  + 285.37238101960 *T)
    sumL0  +=       0.234 * cos(3.64328189368  + 407.47573576480 *T)
    sumL0  +=       0.270 * cos(3.93080966194  + 313.47110834980 *T)
    sumL0  +=       0.302 * cos(3.15201610429  + 915.23535902600 *T)
    sumL0  +=       0.264 * cos(0.46127639727  + 720.89866000480 *T)
    sumL0  +=       0.235 * cos(1.74238582338  + 369.08206769610 *T)
    sumL0  +=       0.284 * cos(5.20345848078  + 395.10562148700 *T)
    sumL0  +=       0.295 * cos(5.03016292492  + 594.65070367540 *T)
    sumL0  +=       0.290 * cos(1.94941528794  + 907.37105249980 *T)
    sumL0  +=       0.229 * cos(0.11049722694  + 3259.89792378380 *T)
    sumL0  +=       0.304 * cos(1.81516929363  + 49.72125262290 *T)
    sumL0  +=       0.268 * cos(5.54944666204  + 12352.85260454480 *T)
    sumL0  +=       0.248 * cos(4.39993603541  + 385.54439393140 *T)
    sumL0  +=       0.234 * cos(1.63365119249  + 3590.51688744200 *T)
    sumL0  +=       0.317 * cos(4.74907646809  + 420.00590873700 *T)
    sumL0  +=       0.228 * cos(4.89094697746  + 1181.43710082460 *T)
    sumL0  +=       0.258 * cos(2.41510535278  + 550.13783421970 *T)
    sumL0  +=       0.236 * cos(4.10002766188  + 6467.92575796160 *T)
    sumL0  +=       0.249 * cos(1.97767956861  + 589.49471013490 *T)
    sumL0  +=       0.226 * cos(1.60308230318  + 316.27999507200 *T)
    sumL0  +=       0.286 * cos(6.10513234736  + 484.70501488690 *T)
    sumL0  +=       0.299 * cos(3.71741328977  + 1123.11782849260 *T)
    sumL0  +=       0.263 * cos(1.85502475341  + 608.87779767700 *T)
    sumL0  +=       0.220 * cos(0.94624116595  + 316.50374424120 *T)
    sumL0  +=       0.292 * cos(3.12099513976  + 47.69426319340 *T)
    sumL0  +=       0.217 * cos(4.93010394323  + 281.17959532560 *T)
    sumL0  +=       0.295 * cos(3.18346450625  + 1050.99635880120 *T)
    sumL0  +=       0.248 * cos(5.48523310249  + 638.93407846750 *T)
    sumL0  +=       0.276 * cos(1.09325899753  + 544.50875992720 *T)
    sumL0  +=       0.279 * cos(2.65446123759  + 134.11226285560 *T)
    sumL0  +=       0.247 * cos(4.17284927352  + 950.13881681630 *T)
    sumL0  +=       0.212 * cos(2.45582771491  + 1164.76232626820 *T)
    sumL0  +=       0.266 * cos(2.30028827109  + 314.90739694830 *T)
    sumL0  +=       0.215 * cos(3.55148746518  + 1097.09427470170 *T)
    sumL0  +=       0.223 * cos(0.36862624591  + 81.89514556810 *T)
    sumL0  +=       0.283 * cos(0.35013012349  + 1269.49963188950 *T)
    sumL0  +=       0.263 * cos(0.58255768951  + 386.98068252990 *T)
    sumL0  +=       0.208 * cos(4.61303945066  + 668.20846196530 *T)
    sumL0  +=       0.222 * cos(4.51639912193  + 304.12232911530 *T)
    sumL0  +=       0.274 * cos(0.63572336701  + 679.25416222920 *T)
    sumL0  +=       0.215 * cos(0.74685271552  + 1008.97935401010 *T)
    sumL0  +=       0.268 * cos(3.43326489941  + 598.84348936940 *T)
    sumL0  +=       0.271 * cos(3.98364990268  + 453.68552624990 *T)
    sumL0  +=       0.215 * cos(2.24139383226  + 661.23792731640 *T)
    sumL0  +=       0.264 * cos(2.58516335032  + 2527.20280398970 *T)
    sumL0  +=       0.205 * cos(1.94472028389  + 650.19222705250 *T)
    sumL0  +=       0.258 * cos(1.50724315598  + 1759.83372106890 *T)
    sumL0  +=       0.272 * cos(6.27135287672  + 990.22940591440 *T)
    sumL0  +=       0.201 * cos(1.11891338895  + 97.41551581630 *T)
    sumL0  +=       0.236 * cos(4.13547185065  + 348.63519857100 *T)
    sumL0  +=       0.201 * cos(2.67485193508  + 1546.53462563090 *T)
    sumL0  +=       0.250 * cos(5.44391862722  + 1254.52177836250 *T)
    sumL0  +=       0.203 * cos(5.48660442144  + 557.03893289940 *T)
    sumL0  +=       0.250 * cos(0.32142312427  + 25448.00585526019 *T)
    sumL0  +=       0.198 * cos(0.91019549387  + 1310.39337013970 *T)
    sumL0  +=       0.200 * cos(0.90438804926  + 47.06112374700 *T)
    sumL0  +=       0.194 * cos(4.05730813129  + 426.85882330690 *T)
    sumL0  +=       0.197 * cos(0.58292199999  + 156.67674413540 *T)
    sumL0  +=       0.197 * cos(2.59384188520  + 639.84910220420 *T)
    sumL0  +=       0.201 * cos(1.49980256114  + 827.92358748650 *T)
    sumL0  +=       0.197 * cos(5.79126360809  + 639.94547042380 *T)
    sumL0  +=       0.266 * cos(1.45514683392  + 109.24311337290 *T)
    sumL0  +=       0.239 * cos(4.63640382863  + 868.71675200020 *T)
    sumL0  +=       0.194 * cos(5.59582424111  + 426.33755844510 *T)
    sumL0  +=       0.247 * cos(2.91348766248  + 689.61853893690 *T)
    sumL0  +=       0.205 * cos(1.04616771894  + 448.68959140380 *T)
    sumL0  +=       0.232 * cos(1.76503818816  + 354.26427286350 *T)
    sumL0  +=       0.224 * cos(3.45391027762  + 1190.03512053370 *T)
    sumL0  +=       0.221 * cos(6.27951545913  + 1596.18644228460 *T)
    sumL0  +=       0.237 * cos(1.24961141369  + 882.94384600180 *T)
    sumL0  +=       0.206 * cos(5.17010664247  + 253.45912050530 *T)
    sumL0  +=       0.203 * cos(0.25518217825  + 4002.88798431640 *T)
    sumL0  +=       0.245 * cos(3.94368773869  + 769.81676347560 *T)
    sumL0  +=       0.255 * cos(6.11790751550  + 763.43692965770 *T)
    sumL0  +=       0.194 * cos(2.83197251801  + 263.70167161710 *T)
    sumL0  +=       0.237 * cos(2.14007373880  + 2700.71514038580 *T)
    sumL0  +=       0.203 * cos(3.00078001542  + 1385.17496870700 *T)
    sumL0  +=       0.203 * cos(3.88000057282  + 419.43645976540 *T)
    sumL0  +=       0.186 * cos(4.79530535895  + 843.63515419640 *T)
    sumL0  +=       0.203 * cos(0.16707173895  + 535.91074021810 *T)
    sumL0  +=       0.245 * cos(0.76762638475  + 5643.17856367740 *T)
    sumL0  +=       0.183 * cos(2.20151176434  + 35.16409022120 *T)
    sumL0  +=       0.242 * cos(3.41613986919  + 864.24208201590 *T)
    sumL0  +=       0.249 * cos(3.47840802421  + 1045.88854937050 *T)
    sumL0  +=       0.199 * cos(4.70077509959  + 1276.61317889030 *T)
    sumL0  +=       0.189 * cos(1.84247610390  + 434.67494572330 *T)
    sumL0  +=       0.192 * cos(3.83148077309  + 666.72398925700 *T)
    sumL0  +=       0.201 * cos(1.34290804373  + 1012.91150727320 *T)
    sumL0  +=       0.210 * cos(5.12097279511  + 3494.53766022420 *T)
    sumL0  +=       0.192 * cos(1.36630036429  + 904.40210708320 *T)
    sumL0  +=       0.195 * cos(1.62629576534  + 364.34676528090 *T)
    sumL0  +=       0.179 * cos(1.66502999050  + 244.79166482700 *T)
    sumL0  +=       0.186 * cos(4.55663748319  + 347.41135829360 *T)
    sumL0  +=       0.201 * cos(0.50219680740  + 36.38793049860 *T)
    sumL0  +=       0.179 * cos(4.55153097409  + 97.46369992610 *T)
    sumL0  +=       0.200 * cos(0.68398141887  + 2015.67108615980 *T)
    sumL0  +=       0.195 * cos(2.20599546209  + 66.18357885820 *T)
    sumL0  +=       0.186 * cos(4.48925201018  + 611.44309831080 *T)
    sumL0  +=       0.186 * cos(3.13663606153  + 433.66355376700 *T)
    sumL0  +=       0.177 * cos(3.62811326217  + 326.68681039510 *T)
    sumL0  +=       0.206 * cos(0.38552475035  + 857.12853501510 *T)
    sumL0  +=       0.229 * cos(1.22330638216  + 2906.90068882300 *T)
    sumL0  +=       0.213 * cos(6.08436923028  + 271.61836777000 *T)
    sumL0  +=       0.174 * cos(6.18833529233  + 3576.28979344040 *T)
    sumL0  +=       0.220 * cos(2.89866380776  + 322.61164478010 *T)
    sumL0  +=       0.202 * cos(2.43755817264  + 812.14258480740 *T)
    sumL0  +=       0.211 * cos(4.27999470527  + 1127.26243007680 *T)
    sumL0  +=       0.188 * cos(0.97173474238  + 1080.72257119160 *T)
    sumL0  +=       0.168 * cos(0.65008064023  + 1493.09366806600 *T)
    sumL0  +=       0.208 * cos(4.68942071563  + 5429.87946823940 *T)
    sumL0  +=       0.182 * cos(0.76058628096  + 504.56118318100 *T)
    sumL0  +=       0.167 * cos(6.22608046965  + 108.50940019000 *T)
    sumL0  +=       0.176 * cos(2.08816807487  + 670.91677495100 *T)
    sumL0  +=       0.170 * cos(1.55680702386  + 1670.82502850000 *T)
    sumL0  +=       0.187 * cos(4.12687876136  + 9985.75935677140 *T)
    sumL0  +=       0.166 * cos(1.58596354370  + 1379.70595310890 *T)
    sumL0  +=       0.170 * cos(1.30946662341  + 837.69726336320 *T)
    sumL0  +=       0.166 * cos(0.02413278965  + 224.60542813280 *T)
    sumL0  +=       0.204 * cos(6.12072939658  + 9360.08916445900 *T)
    sumL0  +=       0.189 * cos(0.28191962964  + 1175.80802653210 *T)
    sumL0  +=       0.184 * cos(2.87461093640  + 398.14400287280 *T)
    sumL0  +=       0.172 * cos(5.41410693088  + 2306.79016155090 *T)
    sumL0  +=       0.174 * cos(0.94052578814  + 632.03297978780 *T)
    sumL0  +=       0.181 * cos(2.20017945285  + 1049.08698945070 *T)
    sumL0  +=       0.176 * cos(1.09857632660  + 531.97858695500 *T)
    sumL0  +=       0.201 * cos(2.68411933635  + 795.68025857210 *T)
    sumL0  +=       0.197 * cos(1.43444932583  + 347.36317418380 *T)
    sumL0  +=       0.200 * cos(4.33954193534  + 1364.72809958190 *T)
    sumL0  +=       0.166 * cos(2.83936872840  + 3553.91152213780 *T)
    sumL0  +=       0.160 * cos(1.28778451384  + 962.50893109410 *T)
    sumL0  +=       0.206 * cos(3.02518737424  + 1141.13406340540 *T)
    sumL0  +=       0.200 * cos(5.54644533685  + 308.31511480930 *T)
    sumL0  +=       0.158 * cos(4.75462437610  + 1534.73816584160 *T)
    sumL0  +=       0.156 * cos(3.42359004405  + 241.75328344120 *T)
    sumL0  +=       0.182 * cos(5.26192506263  + 968.13800538660 *T)
    sumL0  +=       0.168 * cos(0.37055108740  + 10007.09999777380 *T)
    sumL0  +=       0.174 * cos(4.61279044571  + 223.33340374560 *T)
    sumL0  +=       0.158 * cos(4.23116519562  + 821.70381236300 *T)
    sumL0  +=       0.212 * cos(5.98406058023  + 432.74853003030 *T)
    sumL0  +=       0.160 * cos(4.08093018212  + 632.73555520340 *T)
    sumL0  +=       0.179 * cos(5.74376301842  + 924.04582705620 *T)
    sumL0  +=       0.185 * cos(3.01715709315  + 55.87159177720 *T)
    sumL0  +=       0.185 * cos(0.75718598244  + 1286.90811962880 *T)
    sumL0  +=       0.170 * cos(3.46903106526  + 1304.92435454160 *T)
    sumL0  +=       0.167 * cos(3.19767851189  + 635.23141986800 *T)
    sumL0  +=       0.153 * cos(3.83006208210  + 318.67949151700 *T)
    sumL0  +=       0.156 * cos(1.61558494091  + 110.25450532920 *T)
    sumL0  +=       0.206 * cos(1.62702696825  + 389.94962794650 *T)
    sumL0  +=       0.184 * cos(3.31730160238  + 1578.02719501990 *T)
    sumL0  +=       0.184 * cos(6.19913208096  + 731.68372783780 *T)
    sumL0  +=       0.158 * cos(5.58628906712  + 42.32582133180 *T)
    sumL0  +=       0.159 * cos(2.00727485442  + 702.14871190910 *T)
    sumL0  +=       0.155 * cos(0.02889842368  + 1357.61455258110 *T)
    sumL0  +=       0.151 * cos(5.80616069064  + 680.05731138130 *T)
    sumL0  +=       0.157 * cos(5.67235483479  + 77837.11123384659 *T)
    sumL0  +=       0.156 * cos(2.78001360754  + 1567.73225428140 *T)
    sumL0  +=       0.203 * cos(5.91820261629  + 971.10695080320 *T)
    sumL0  +=       0.156 * cos(5.68482948401  + 649.45851386960 *T)
    sumL0  +=       0.186 * cos(5.72093247348  + 664.27630870220 *T)
    sumL0  +=       0.184 * cos(4.05878767739  + 976.73602509570 *T)
    sumL0  +=       0.159 * cos(3.10978439502  + 230.70758317730 *T)
    sumL0  +=       0.154 * cos(4.54479813612  + 1239.54392483550 *T)
    sumL0  +=       0.177 * cos(4.68148789757  + 2921.12778282460 *T)
    sumL0  +=       0.158 * cos(4.30854435851  + 633.74694715970 *T)
    sumL0  +=       0.195 * cos(5.33376523453  + 1130.23137549340 *T)
    sumL0  +=       0.188 * cos(2.04023570497  + 1127.04998175570 *T)
    sumL0  +=       0.174 * cos(2.88817750136  + 25668.41849769900 *T)
    sumL0  +=       0.161 * cos(2.96300008890  + 152.74459087230 *T)
    sumL0  +=       0.170 * cos(1.70779560770  + 493.30303459600 *T)
    sumL0  +=       0.156 * cos(3.05331862578  + 913.75088631770 *T)
    sumL0  +=       0.169 * cos(3.18325334507  + 757.80785536520 *T)
    sumL0  +=       0.145 * cos(0.89672198690  + 632.83192342300 *T)
    sumL0  +=       0.149 * cos(2.69519361988  + 203.26478713040 *T)
    sumL0  +=       0.167 * cos(2.96547549589  + 1201.83158032300 *T)
    sumL0  +=       0.187 * cos(3.17592113403  + 842.90144101350 *T)
    sumL0  +=       0.173 * cos(1.91528836350  + 3487.42411322340 *T)
    sumL0  +=       0.154 * cos(0.37169915171  + 285.11174858870 *T)
    sumL0  +=       0.198 * cos(3.14620903981  + 640.86049416050 *T)
    sumL0  +=       0.143 * cos(3.42760427364  + 520.12973753900 *T)
    sumL0  +=       0.157 * cos(2.34913118792  + 5959.57043333400 *T)
    sumL0  +=       0.178 * cos(5.34566184657  + 272.58157561650 *T)
    sumL0  +=       0.171 * cos(3.34185265551  + 3067.93946934820 *T)
    sumL0  +=       0.171 * cos(4.87848878292  + 354.52490529440 *T)
    sumL0  +=       0.189 * cos(0.07069084691  + 1585.89150154610 *T)
    sumL0  +=       0.137 * cos(1.28925665885  + 214.10224459010 *T)
    sumL0  +=       0.137 * cos(3.49438482900  + 212.49594628590 *T)
    sumL0  +=       0.174 * cos(3.07450555579  + 64.95973858080 *T)
    sumL0  +=       0.181 * cos(0.78652249647  + 657.16276170140 *T)
    sumL0  +=       0.158 * cos(2.93809799392  + 211.65456403530 *T)
    sumL0  +=       0.133 * cos(5.70724173596  + 469.72716135990 *T)
    sumL0  +=       0.140 * cos(0.35530248121  + 219.51887056150 *T)
    sumL0  +=       0.132 * cos(1.60614108449  + 1372.59240610810 *T)
    sumL0  +=       0.164 * cos(1.68236348359  + 707.56533788050 *T)
    sumL0  +=       0.128 * cos(1.49742950959  + 45.24658263860 *T)
    sumL0  +=       0.132 * cos(4.98075995751  + 238.57188970350 *T)
    sumL0  +=       0.140 * cos(3.62845435388  + 423.88987789030 *T)
    sumL0  +=       0.134 * cos(1.41837795083  + 3906.90875709860 *T)
    sumL0  +=       0.127 * cos(1.48966646069  + 856.37777548970 *T)
    sumL0  +=       0.158 * cos(2.57990141197  + 369.97583957340 *T)
    sumL0  +=       0.132 * cos(2.39662455993  + 184.72728735580 *T)
    sumL0  +=       0.140 * cos(4.36785639279  + 207.07932031450 *T)
    sumL0  +=       0.157 * cos(4.56655940808  + 251.17149864490 *T)
    sumL0  +=       0.125 * cos(4.60433583965  + 6076.89030155420 *T)
    sumL0  +=       0.121 * cos(1.59513269211  + 184.84490743480 *T)
    sumL0  +=       0.145 * cos(5.85121885906  + 221.89711514710 *T)
    sumL0  +=       0.137 * cos(0.76499603885  + 476.10699517780 *T)
    sumL0  +=       0.139 * cos(6.02467582226  + 429.30650386170 *T)
    sumL0  +=       0.117 * cos(0.23611722307  + 426.75824957040 *T)
    sumL0  +=       0.117 * cos(3.13382984215  + 426.43813218160 *T)
    sumL0  +=       0.116 * cos(4.34801448720  + 418.96337901340 *T)
    sumL0  +=       0.135 * cos(2.69789181467  + 455.16999895820 *T)
    sumL0  +=       0.134 * cos(1.19058772772  + 502.86426215160 *T)
    sumL0  +=       0.123 * cos(3.97560160218  + 499.89531673500 *T)
    sumL0  +=       0.110 * cos(2.37032413718  + 439.93151300030 *T)
    sumL0  +=       0.109 * cos(6.20303291896  + 220.30076785420 *T)
    sumL0  +=       0.109 * cos(1.38979633175  + 325.95309721220 *T)
    sumL0  +=       0.115 * cos(0.05845336809  + 631.82053146670 *T)
    sumL0  +=       0.139 * cos(2.21086387259  + 9573.38825989700 *T)
    sumL0  +=       0.142 * cos(6.14666228712  + 3340.61242669980 *T)
    sumL0  +=       0.122 * cos(4.77182119272  + 604.47256366190 *T)
    sumL0  +=       0.119 * cos(3.03461367061  + 528.20649238630 *T)
    sumL0  +=       0.127 * cos(2.97263950543  + 498.93210888850 *T)
    sumL0  +=       0.111 * cos(5.01192320232  + 220.20019411770 *T)
    sumL0  +=       0.126 * cos(3.95743516129  + 566.60016045500 *T)
    sumL0  +=       0.098 * cos(2.36210526249  + 634.26821202150 *T)
    sumL0  +=       0.108 * cos(2.46872857126  + 83.37961827640 *T)
    sumL0  +=       0.101 * cos(4.88259474274  + 425.32616648880 *T)
    sumL0  +=       0.127 * cos(4.89006771254  + 162.09337010680 *T)
    sumL0  +=       0.099 * cos(0.90073463816  + 586.31331639720 *T)
    sumL0  +=       0.099 * cos(0.16310526294  + 394.35486196160 *T)
    sumL0  +=       0.115 * cos(0.49160291262  + 517.16079212240 *T)
    sumL0  +=       0.101 * cos(3.86043866859  + 198.10879358990 *T)
    sumL0  +=       0.105 * cos(3.48222097813  + 5863.59120611620 *T)
    sumL0  +=       0.104 * cos(4.61148354671  + 220.52451702340 *T)
    sumL0  +=       0.101 * cos(4.77041950285  + 427.87021526320 *T)
    sumL0  +=       0.098 * cos(2.90784721214  + 199.96577331370 *T)
    sumL0  +=       0.127 * cos(0.14136936897  + 2332.06295581640 *T)
    sumL0  +=       0.091 * cos(6.22628300117  + 211.29335786790 *T)
    sumL0  +=       0.100 * cos(5.14847283436  + 226.79247625670 *T)
    sumL0  +=       0.091 * cos(4.84054379386  + 215.30483300810 *T)
    sumL0  +=       0.099 * cos(4.37583492400  + 640.41855117580 *T)
    sumL0  +=       0.093 * cos(5.30395179617  + 222.70026429920 *T)
    sumL0  +=       0.086 * cos(4.57481701854  + 636.97652500720 *T)
    sumL0  +=       0.111 * cos(0.61102669309  + 1089.12939443900 *T)
    sumL0  +=       0.086 * cos(4.50969293872  + 625.88264063350 *T)
    sumL0  +=       0.088 * cos(0.36828594935  + 444.12429869430 *T)
    sumL0  +=       0.093 * cos(5.76287499885  + 203.89792657680 *T)
    sumL0  +=       0.082 * cos(2.85558603378  + 318.83955021140 *T)
    sumL0  +=       0.082 * cos(5.19618475111  + 1467.82087380050 *T)
    sumL0  +=       0.086 * cos(0.97633784601  + 200.03520928290 *T)
    sumL0  +=       0.082 * cos(4.78557953017  + 195.77298761970 *T)
    sumL0  +=       0.106 * cos(2.29520624233  + 799.61241183520 *T)
    sumL0  +=       0.081 * cos(3.57727166766  + 205.97310011610 *T)
    sumL0  +=       0.078 * cos(5.50343512580  + 262.80789973980 *T)
    sumL0  +=       0.087 * cos(0.76830756075  + 201.99276274320 *T)
    sumL0  +=       0.102 * cos(2.11516755277  + 206.93630796260 *T)
    sumL0  +=       0.081 * cos(5.29639775054  + 111.16952906590 *T)
    sumL0  +=       0.075 * cos(2.77117107886  + 255.83736509090 *T)
    sumL0  +=       0.074 * cos(5.81587984729  + 316.44005376640 *T)
    sumL0  +=       0.076 * cos(1.78743197928  + 171.65459766240 *T)
    sumL0  +=       0.094 * cos(4.99996904753  + 378.64329525170 *T)
    sumL0  +=       0.089 * cos(5.85818860151  + 807.94979911340 *T)
    sumL0  +=       0.072 * cos(0.99858616883  + 280.21638747910 *T)
    sumL0  +=       0.099 * cos(0.15018241445  + 186.21176006410 *T)
    sumL0  +=       0.072 * cos(5.15715918322  + 110.15813710960 *T)

    # SATURN Longitude
    sumL1  = 0
    sumL1  +=  21354295595.986 * cos(0.00000000000  + 0.00000000000 *T)
    sumL1  +=  1296855.005 * cos(1.82820544701  + 213.29909543800 *T)
    sumL1  +=  564347.566 * cos(2.88500136429  + 7.11354700080 *T)
    sumL1  +=   98323.030 * cos(1.08070061328  + 426.59819087600 *T)
    sumL1  +=  107678.770 * cos(2.27769911872  + 206.18554843720 *T)
    sumL1  +=   40254.586 * cos(2.04128257090  + 220.41264243880 *T)
    sumL1  +=   19941.734 * cos(1.27954662736  + 103.09277421860 *T)
    sumL1  +=   10511.706 * cos(2.74880392800  + 14.22709400160 *T)
    sumL1  +=    6939.233 * cos(0.40493079985  + 639.89728631400 *T)
    sumL1  +=    4803.325 * cos(2.44194097666  + 419.48464387520 *T)
    sumL1  +=    4056.325 * cos(2.92166618776  + 110.20632121940 *T)
    sumL1  +=    3768.630 * cos(3.64965631460  + 3.93215326310 *T)
    sumL1  +=    3384.684 * cos(2.41694251653  + 3.18139373770 *T)
    sumL1  +=    3302.200 * cos(1.26256486715  + 433.71173787680 *T)
    sumL1  +=    3071.382 * cos(2.32739317750  + 199.07200143640 *T)
    sumL1  +=    1953.036 * cos(3.56394683300  + 11.04570026390 *T)
    sumL1  +=    1249.348 * cos(2.62803737519  + 95.97922721780 *T)
    sumL1  +=     921.683 * cos(1.96089834250  + 227.52618943960 *T)
    sumL1  +=     705.587 * cos(4.41689249330  + 529.69096509460 *T)
    sumL1  +=     649.654 * cos(6.17418093659  + 202.25339517410 *T)
    sumL1  +=     627.603 * cos(6.11088227167  + 309.27832265580 *T)
    sumL1  +=     486.843 * cos(6.03998200305  + 853.19638175200 *T)
    sumL1  +=     468.377 * cos(4.61707843907  + 63.73589830340 *T)
    sumL1  +=     478.501 * cos(4.98776987984  + 522.57741809380 *T)
    sumL1  +=     417.010 * cos(2.11708169277  + 323.50541665740 *T)
    sumL1  +=     407.630 * cos(1.29949556676  + 209.36694217490 *T)
    sumL1  +=     343.826 * cos(3.95854178574  + 412.37109687440 *T)
    sumL1  +=     339.724 * cos(3.63396398752  + 316.39186965660 *T)
    sumL1  +=     335.936 * cos(3.77173072712  + 735.87651353180 *T)
    sumL1  +=     331.933 * cos(2.86077699882  + 210.11770170030 *T)
    sumL1  +=     352.489 * cos(2.31707079463  + 632.78373931320 *T)
    sumL1  +=     289.429 * cos(2.73263080235  + 117.31986822020 *T)
    sumL1  +=     265.801 * cos(0.54344631312  + 647.01083331480 *T)
    sumL1  +=     230.493 * cos(1.64428879621  + 216.48048917570 *T)
    sumL1  +=     280.911 * cos(5.74398845416  + 2.44768055480 *T)
    sumL1  +=     191.667 * cos(2.96512946582  + 224.34479570190 *T)
    sumL1  +=     172.891 * cos(4.07695221044  + 846.08283475120 *T)
    sumL1  +=     167.131 * cos(2.59745202658  + 21.34064100240 *T)
    sumL1  +=     136.328 * cos(2.28580246629  + 10.29494073850 *T)
    sumL1  +=     131.364 * cos(3.44108355646  + 742.99006053260 *T)
    sumL1  +=     127.838 * cos(4.09533471247  + 217.23124870110 *T)
    sumL1  +=     108.862 * cos(6.16141072262  + 415.55249061210 *T)
    sumL1  +=      93.909 * cos(3.48397279899  + 1052.26838318840 *T)
    sumL1  +=      92.482 * cos(3.94755499926  + 88.86568021700 *T)
    sumL1  +=      97.584 * cos(4.72845436677  + 838.96928775040 *T)
    sumL1  +=      86.600 * cos(1.21951325061  + 440.82528487760 *T)
    sumL1  +=      83.463 * cos(3.11269504725  + 625.67019231240 *T)
    sumL1  +=      77.588 * cos(6.24408938835  + 302.16477565500 *T)
    sumL1  +=      61.557 * cos(1.82789612597  + 195.13984817330 *T)
    sumL1  +=      61.900 * cos(4.29344363385  + 127.47179660680 *T)
    sumL1  +=      67.106 * cos(0.28961738595  + 4.66586644600 *T)
    sumL1  +=      56.919 * cos(5.01889578112  + 137.03302416240 *T)
    sumL1  +=      54.160 * cos(5.12628572382  + 490.33408917940 *T)
    sumL1  +=      54.585 * cos(0.28356341456  + 74.78159856730 *T)
    sumL1  +=      51.425 * cos(1.45766406064  + 536.80451209540 *T)
    sumL1  +=      65.843 * cos(5.64757042732  + 9.56122755560 *T)
    sumL1  +=      57.780 * cos(2.47630552035  + 191.95845443560 *T)
    sumL1  +=      44.444 * cos(2.70873627665  + 5.41662597140 *T)
    sumL1  +=      46.799 * cos(1.17721211050  + 149.56319713460 *T)
    sumL1  +=      40.380 * cos(3.88870105683  + 728.76296653100 *T)
    sumL1  +=      37.768 * cos(2.53379013859  + 12.53017297220 *T)
    sumL1  +=      46.649 * cos(5.14818326902  + 515.46387109300 *T)
    sumL1  +=      45.891 * cos(2.23198878761  + 956.28915597060 *T)
    sumL1  +=      40.400 * cos(0.41281520440  + 269.92144674060 *T)
    sumL1  +=      37.191 * cos(3.78239026411  + 2.92076130680 *T)
    sumL1  +=      33.778 * cos(3.21070688046  + 1368.66025284500 *T)
    sumL1  +=      37.969 * cos(0.64665967180  + 422.66603761290 *T)
    sumL1  +=      32.857 * cos(0.30063884563  + 351.81659230870 *T)
    sumL1  +=      33.050 * cos(5.43038091186  + 1066.49547719000 *T)
    sumL1  +=      30.276 * cos(2.84067004928  + 203.00415469950 *T)
    sumL1  +=      35.116 * cos(6.08421794089  + 5.62907429250 *T)
    sumL1  +=      29.667 * cos(3.39052569135  + 1059.38193018920 *T)
    sumL1  +=      33.217 * cos(4.64063092111  + 277.03499374140 *T)
    sumL1  +=      31.876 * cos(4.38622923770  + 1155.36115740700 *T)
    sumL1  +=      28.913 * cos(2.02614760507  + 330.61896365820 *T)
    sumL1  +=      28.264 * cos(2.74178953996  + 265.98929347750 *T)
    sumL1  +=      30.089 * cos(6.18684614308  + 284.14854074220 *T)
    sumL1  +=      31.329 * cos(2.43455855525  + 52.69019803950 *T)
    sumL1  +=      26.493 * cos(4.51214170121  + 340.77089204480 *T)
    sumL1  +=      21.983 * cos(5.14437352579  + 4.19278569400 *T)
    sumL1  +=      22.230 * cos(1.96481952451  + 203.73786788240 *T)
    sumL1  +=      20.824 * cos(6.16048095923  + 860.30992875280 *T)
    sumL1  +=      21.690 * cos(2.67578768862  + 942.06206196900 *T)
    sumL1  +=      22.552 * cos(5.88579123000  + 210.85141488320 *T)
    sumL1  +=      19.807 * cos(2.31345263487  + 437.64389113990 *T)
    sumL1  +=      19.447 * cos(4.76573277668  + 70.84944530420 *T)
    sumL1  +=      19.310 * cos(4.10209060369  + 18.15924726470 *T)
    sumL1  +=      22.662 * cos(4.13732273379  + 191.20769491020 *T)
    sumL1  +=      18.209 * cos(0.90310796389  + 429.77958461370 *T)
    sumL1  +=      17.667 * cos(1.84954766042  + 234.63973644040 *T)
    sumL1  +=      17.547 * cos(2.44735118493  + 423.41679713830 *T)
    sumL1  +=      15.428 * cos(4.23790088205  + 1162.47470440780 *T)
    sumL1  +=      14.608 * cos(3.59713247857  + 1045.15483618760 *T)
    sumL1  +=      14.111 * cos(2.94262468353  + 1685.05212250160 *T)
    sumL1  +=      16.328 * cos(4.05665272725  + 949.17560896980 *T)
    sumL1  +=      13.348 * cos(6.24509592240  + 38.13303563780 *T)
    sumL1  +=      15.918 * cos(1.06434204938  + 56.62235130260 *T)
    sumL1  +=      14.059 * cos(1.43503954068  + 408.43894361130 *T)
    sumL1  +=      13.093 * cos(5.75815864257  + 138.51749687070 *T)
    sumL1  +=      15.772 * cos(5.59350835225  + 6.15033915430 *T)
    sumL1  +=      14.962 * cos(5.77192239389  + 22.09140052780 *T)
    sumL1  +=      16.024 * cos(1.93900586533  + 1272.68102562720 *T)
    sumL1  +=      16.751 * cos(5.96673627422  + 628.85158605010 *T)
    sumL1  +=      12.843 * cos(4.24658666814  + 405.25754987360 *T)
    sumL1  +=      13.628 * cos(4.09892958087  + 1471.75302706360 *T)
    sumL1  +=      15.067 * cos(0.74142807591  + 200.76892246580 *T)
    sumL1  +=      10.961 * cos(1.55022573283  + 223.59403617650 *T)
    sumL1  +=      11.695 * cos(1.81237511034  + 124.43341522100 *T)
    sumL1  +=      10.346 * cos(3.46814088412  + 1375.77379984580 *T)
    sumL1  +=      12.056 * cos(1.85655834555  + 131.40394986990 *T)
    sumL1  +=      10.123 * cos(2.38221133049  + 107.02492748170 *T)
    sumL1  +=       9.855 * cos(3.95166998848  + 430.53034413910 *T)
    sumL1  +=       9.803 * cos(2.55389483994  + 99.91138048090 *T)
    sumL1  +=      10.614 * cos(5.36692189034  + 215.74677599280 *T)
    sumL1  +=      12.080 * cos(4.84549317054  + 831.85574074960 *T)
    sumL1  +=      10.210 * cos(6.07692961370  + 32.24332891440 *T)
    sumL1  +=       9.245 * cos(3.65417467270  + 142.44965013380 *T)
    sumL1  +=       8.984 * cos(1.23808405498  + 106.27416795630 *T)
    sumL1  +=       9.336 * cos(5.81062768434  + 7.16173111060 *T)
    sumL1  +=       9.717 * cos(1.38703872827  + 145.63104387150 *T)
    sumL1  +=       8.394 * cos(4.42341211111  + 703.63318461740 *T)
    sumL1  +=       8.370 * cos(5.64015188458  + 62.25142559510 *T)
    sumL1  +=       8.244 * cos(2.42225929772  + 1258.45393162560 *T)
    sumL1  +=       7.784 * cos(0.52562994711  + 654.12438031560 *T)
    sumL1  +=       7.626 * cos(3.75258725596  + 312.19908396260 *T)
    sumL1  +=       7.222 * cos(0.28429555677  + 0.75075952540 *T)
    sumL1  +=       8.236 * cos(6.22250515902  + 14.97785352700 *T)
    sumL1  +=       7.054 * cos(0.53177810740  + 388.46515523820 *T)
    sumL1  +=       6.567 * cos(3.48657341701  + 35.42472265210 *T)
    sumL1  +=       9.011 * cos(4.94919626910  + 208.63322899200 *T)
    sumL1  +=       8.980 * cos(0.08138173719  + 288.08069400530 *T)
    sumL1  +=       6.421 * cos(3.32905264657  + 1361.54670584420 *T)
    sumL1  +=       6.489 * cos(2.89389587598  + 114.13847448250 *T)
    sumL1  +=       6.244 * cos(0.54973852782  + 65.22037101170 *T)
    sumL1  +=       6.154 * cos(2.67885860584  + 2001.44399215820 *T)
    sumL1  +=       6.742 * cos(0.23586769279  + 8.07675484730 *T)
    sumL1  +=       7.297 * cos(4.85321224483  + 222.86032299360 *T)
    sumL1  +=       6.302 * cos(3.80651124694  + 1788.14489672020 *T)
    sumL1  +=       5.824 * cos(4.39327457448  + 81.75213321620 *T)
    sumL1  +=       6.102 * cos(0.88585782895  + 92.04707395470 *T)
    sumL1  +=       6.914 * cos(2.04631426723  + 99.16062095550 *T)
    sumL1  +=       5.363 * cos(5.47995103139  + 563.63121503840 *T)
    sumL1  +=       5.172 * cos(2.11968421583  + 214.26230328450 *T)
    sumL1  +=       5.117 * cos(5.76987684107  + 565.11568774670 *T)
    sumL1  +=       6.197 * cos(1.62553688800  + 1589.07289528380 *T)
    sumL1  +=       4.970 * cos(0.41949366126  + 76.26607127560 *T)
    sumL1  +=       6.640 * cos(5.82582210639  + 483.22054217860 *T)
    sumL1  +=       5.277 * cos(4.57975789757  + 134.58534360760 *T)
    sumL1  +=       4.974 * cos(4.20243895902  + 404.50679034820 *T)
    sumL1  +=       5.150 * cos(4.67582673243  + 212.33588759150 *T)
    sumL1  +=       4.764 * cos(4.59303997414  + 554.06998748280 *T)
    sumL1  +=       4.573 * cos(3.24875415786  + 231.45834270270 *T)
    sumL1  +=       4.811 * cos(0.46206327592  + 362.86229257260 *T)
    sumL1  +=       5.148 * cos(3.33570646174  + 1.48447270830 *T)
    sumL1  +=       4.654 * cos(5.80233066659  + 217.96496188400 *T)
    sumL1  +=       4.509 * cos(5.37581684215  + 497.44763618020 *T)
    sumL1  +=       4.443 * cos(0.11349392292  + 295.05122865420 *T)
    sumL1  +=       4.943 * cos(3.78020789259  + 1265.56747862640 *T)
    sumL1  +=       4.211 * cos(4.88306021960  + 98.89998852460 *T)
    sumL1  +=       4.252 * cos(5.00120115113  + 213.34727954780 *T)
    sumL1  +=       4.774 * cos(4.53259894142  + 1148.24761040620 *T)
    sumL1  +=       3.911 * cos(0.58582192963  + 750.10360753340 *T)
    sumL1  +=       5.069 * cos(2.20305668335  + 207.88246946660 *T)
    sumL1  +=       3.553 * cos(0.35374030841  + 333.65734504400 *T)
    sumL1  +=       3.771 * cos(0.98542435766  + 24.37902238820 *T)
    sumL1  +=       3.458 * cos(1.84990273999  + 225.82926841020 *T)
    sumL1  +=       3.401 * cos(5.31342401626  + 347.88443904560 *T)
    sumL1  +=       3.347 * cos(0.21414641376  + 635.96513305090 *T)
    sumL1  +=       3.637 * cos(1.61315058382  + 245.54242435240 *T)
    sumL1  +=       3.416 * cos(2.19551489078  + 1574.84580128220 *T)
    sumL1  +=       3.655 * cos(0.80544245690  + 343.21857259960 *T)
    sumL1  +=       4.260 * cos(1.80258750109  + 213.25091132820 *T)
    sumL1  +=       3.110 * cos(3.03815175282  + 1677.93857550080 *T)
    sumL1  +=       3.052 * cos(1.33858964447  + 543.91805909620 *T)
    sumL1  +=       3.694 * cos(0.81606028298  + 344.70304530790 *T)
    sumL1  +=       3.016 * cos(3.36219319026  + 7.86430652620 *T)
    sumL1  +=       2.937 * cos(4.86927342776  + 144.14657116320 *T)
    sumL1  +=       2.768 * cos(2.42707131609  + 2317.83586181480 *T)
    sumL1  +=       3.059 * cos(4.30820099442  + 6062.66320755260 *T)
    sumL1  +=       3.650 * cos(5.12802531219  + 218.92816973050 *T)
    sumL1  +=       2.963 * cos(3.53480751374  + 2104.53676637680 *T)
    sumL1  +=       3.230 * cos(2.88057019783  + 216.21985674480 *T)
    sumL1  +=       2.984 * cos(2.52971310583  + 1692.16566950240 *T)
    sumL1  +=       2.897 * cos(5.73256482240  + 9992.87290377220 *T)
    sumL1  +=       2.591 * cos(3.79880285744  + 17.26547538740 *T)
    sumL1  +=       3.495 * cos(5.29902525443  + 350.33211960040 *T)
    sumL1  +=       2.859 * cos(3.72804950659  + 6076.89030155420 *T)
    sumL1  +=       2.775 * cos(0.23549396237  + 357.44566660120 *T)
    sumL1  +=       2.976 * cos(2.48769315964  + 46.47042291600 *T)
    sumL1  +=       2.487 * cos(4.37868078530  + 217.49188113200 *T)
    sumL1  +=       2.711 * cos(5.15376840150  + 10007.09999777380 *T)
    sumL1  +=       3.127 * cos(1.92343235583  + 17.40848773930 *T)
    sumL1  +=       3.181 * cos(1.72419900322  + 1169.58825140860 *T)
    sumL1  +=       2.348 * cos(0.77373103004  + 414.06801790380 *T)
    sumL1  +=       2.606 * cos(3.42836913440  + 31.01948863700 *T)
    sumL1  +=       2.556 * cos(0.91735028377  + 479.28838891550 *T)
    sumL1  +=       2.399 * cos(4.82440545738  + 1279.79457262800 *T)
    sumL1  +=       2.245 * cos(3.76323995584  + 425.11371816770 *T)
    sumL1  +=       3.020 * cos(0.25310250109  + 120.35824960600 *T)
    sumL1  +=       2.503 * cos(2.10679832121  + 168.05251279940 *T)
    sumL1  +=       2.564 * cos(1.63158205055  + 182.27960680100 *T)
    sumL1  +=       2.221 * cos(3.15472373256  + 212.77783057620 *T)
    sumL1  +=       2.357 * cos(2.33145013714  + 218.71572140940 *T)
    sumL1  +=       2.510 * cos(4.51903989011  + 198.32124191100 *T)
    sumL1  +=       2.715 * cos(5.76330259543  + 618.55664531160 *T)
    sumL1  +=       2.204 * cos(3.35952557362  + 160.60889739850 *T)
    sumL1  +=       2.648 * cos(0.71962005233  + 85.82729883120 *T)
    sumL1  +=       2.029 * cos(5.28642331696  + 125.98732389850 *T)
    sumL1  +=       2.497 * cos(1.36671447252  + 1905.46476494040 *T)
    sumL1  +=       2.017 * cos(1.11498225426  + 447.93883187840 *T)
    sumL1  +=       2.052 * cos(1.27587874735  + 14.01464568050 *T)
    sumL1  +=       2.254 * cos(3.22447674190  + 273.10284047830 *T)
    sumL1  +=       2.014 * cos(0.39787014152  + 358.93013930950 *T)
    sumL1  +=       1.981 * cos(2.33696859021  + 28.45418800320 *T)
    sumL1  +=       2.197 * cos(5.93386789705  + 13.33332212430 *T)
    sumL1  +=       2.237 * cos(3.64433751164  + 213.82036029980 *T)
    sumL1  +=       1.930 * cos(1.85671740340  + 1.27202438720 *T)
    sumL1  +=       2.037 * cos(5.05300562628  + 424.15051032120 *T)
    sumL1  +=       1.994 * cos(1.35690802366  + 20.60692781950 *T)
    sumL1  +=       1.911 * cos(3.44106886717  + 69.15252427480 *T)
    sumL1  +=       1.925 * cos(3.75243031545  + 28.31117565130 *T)
    sumL1  +=       2.297 * cos(4.24557050896  + 1464.63948006280 *T)
    sumL1  +=       2.117 * cos(2.25897766314  + 116.42609634290 *T)
    sumL1  +=       1.847 * cos(5.40631472802  + 31.49256938900 *T)
    sumL1  +=       1.841 * cos(1.56916484272  + 650.94298657790 *T)
    sumL1  +=       1.884 * cos(6.27233535258  + 25.12978191360 *T)
    sumL1  +=       1.960 * cos(4.89484014840  + 275.55052103310 *T)
    sumL1  +=       2.016 * cos(5.45791785675  + 842.15068148810 *T)
    sumL1  +=       2.282 * cos(4.96276947440  + 258.87574647670 *T)
    sumL1  +=       1.709 * cos(3.99098237135  + 416.30325013750 *T)
    sumL1  +=       2.176 * cos(0.00746756006  + 0.89377187730 *T)
    sumL1  +=       1.634 * cos(5.30978165487  + 251.43213107580 *T)
    sumL1  +=       1.687 * cos(0.41586020065  + 54.17467074780 *T)
    sumL1  +=       1.910 * cos(2.59825755790  + 329.72519178090 *T)
    sumL1  +=       2.113 * cos(2.56582292726  + 59.80374504030 *T)
    sumL1  +=       1.921 * cos(2.42279051938  + 113.38771495710 *T)
    sumL1  +=       1.658 * cos(5.47323651540  + 1073.60902419080 *T)
    sumL1  +=       1.590 * cos(2.77545297350  + 1994.33044515740 *T)
    sumL1  +=       1.936 * cos(3.47558926847  + 1581.95934828300 *T)
    sumL1  +=       1.649 * cos(1.82779010589  + 128.95626931510 *T)
    sumL1  +=       1.598 * cos(1.71806465300  + 129.91947716160 *T)
    sumL1  +=       1.967 * cos(1.25160413795  + 621.73803904930 *T)
    sumL1  +=       1.702 * cos(1.91076102800  + 278.51946644970 *T)
    sumL1  +=       1.569 * cos(0.16491194947  + 643.07868005170 *T)
    sumL1  +=       1.989 * cos(5.28799230992  + 508.35032409220 *T)
    sumL1  +=       1.520 * cos(0.56950979689  + 320.32402291970 *T)
    sumL1  +=       1.501 * cos(1.99815894193  + 1891.23767093880 *T)
    sumL1  +=       1.532 * cos(3.27362317849  + 2420.92863603340 *T)
    sumL1  +=       1.701 * cos(2.72041261115  + 767.36908292080 *T)
    sumL1  +=       1.561 * cos(6.09424459628  + 280.96714700450 *T)
    sumL1  +=       1.331 * cos(4.20944443790  + 546.95644048200 *T)
    sumL1  +=       1.381 * cos(2.06768100830  + 192.69216761850 *T)
    sumL1  +=       1.368 * cos(6.28049502257  + 1795.25844372100 *T)
    sumL1  +=       1.519 * cos(2.20299556153  + 2008.55753915900 *T)
    sumL1  +=       1.356 * cos(4.01521042413  + 721.64941953020 *T)
    sumL1  +=       1.296 * cos(4.84815978742  + 45.57665103870 *T)
    sumL1  +=       1.267 * cos(5.28146654999  + 173.94221952280 *T)
    sumL1  +=       1.402 * cos(6.12951551550  + 39.35687591520 *T)
    sumL1  +=       1.252 * cos(2.19169926554  + 2634.22773147140 *T)
    sumL1  +=       1.466 * cos(4.16354845643  + 26.82670294300 *T)
    sumL1  +=       1.285 * cos(3.76170874847  + 2.28762186040 *T)
    sumL1  +=       1.500 * cos(5.41022492529  + 214.04985496340 *T)
    sumL1  +=       1.396 * cos(4.78595583428  + 219.44943459230 *T)
    sumL1  +=       1.430 * cos(0.70934745161  + 254.94359321360 *T)
    sumL1  +=       1.195 * cos(3.71281085322  + 264.50482076920 *T)
    sumL1  +=       1.181 * cos(0.42635230882  + 41.64449777560 *T)
    sumL1  +=       1.190 * cos(2.02079286787  + 1485.98012106520 *T)
    sumL1  +=       1.160 * cos(5.23649231796  + 181.05576652360 *T)
    sumL1  +=       1.535 * cos(3.62746990294  + 561.18353448360 *T)
    sumL1  +=       1.120 * cos(1.09127922130  + 6.59228213900 *T)
    sumL1  +=       1.100 * cos(0.27844612141  + 184.09414790940 *T)
    sumL1  +=       1.227 * cos(1.39969681270  + 209.10630974400 *T)
    sumL1  +=       1.353 * cos(6.12903657666  + 207.67002114550 *T)
    sumL1  +=       1.124 * cos(6.05105541765  + 291.26208774300 *T)
    sumL1  +=       1.194 * cos(4.79565407023  + 1478.86657406440 *T)
    sumL1  +=       1.082 * cos(4.73602931755  + 78.71375183040 *T)
    sumL1  +=       1.202 * cos(3.47301104146  + 51.20572533120 *T)
    sumL1  +=       1.298 * cos(2.34761557822  + 210.37833413120 *T)
    sumL1  +=       1.166 * cos(4.20037524355  + 417.03696332040 *T)
    sumL1  +=       1.228 * cos(3.94985981275  + 1781.03134971940 *T)
    sumL1  +=       1.401 * cos(2.41318931513  + 636.71589257630 *T)
    sumL1  +=       1.009 * cos(6.17414889934  + 2111.65031337760 *T)
    sumL1  +=       1.084 * cos(3.68958647346  + 274.06604832480 *T)
    sumL1  +=       1.068 * cos(0.80258823981  + 436.89313161450 *T)
    sumL1  +=       1.007 * cos(3.42792508860  + 629.60234557550 *T)
    sumL1  +=       0.998 * cos(5.57130056835  + 205.22234059070 *T)
    sumL1  +=       1.058 * cos(1.05742945779  + 237.67811782620 *T)
    sumL1  +=       1.020 * cos(3.33667290300  + 166.82867252200 *T)
    sumL1  +=       0.965 * cos(6.08359503243  + 601.76425067620 *T)
    sumL1  +=       1.005 * cos(3.56310748091  + 643.82943957710 *T)
    sumL1  +=       0.987 * cos(0.97129012811  + 305.34616939270 *T)
    sumL1  +=       0.927 * cos(3.87717400791  + 135.33610313300 *T)
    sumL1  +=       1.129 * cos(5.94840103961  + 196.62432088160 *T)
    sumL1  +=       1.118 * cos(5.25415059584  + 189.72322220190 *T)
    sumL1  +=       1.200 * cos(1.16671933467  + 2221.85663459700 *T)
    sumL1  +=       0.909 * cos(2.14001565047  + 617.80588578620 *T)
    sumL1  +=       0.899 * cos(2.31811625712  + 312.45971639350 *T)
    sumL1  +=       1.081 * cos(0.91006048421  + 313.21047591890 *T)
    sumL1  +=       0.891 * cos(3.74923531791  + 916.93228005540 *T)
    sumL1  +=       0.886 * cos(4.76066858907  + 776.93031047640 *T)
    sumL1  +=       0.912 * cos(0.99592540858  + 491.81856188770 *T)
    sumL1  +=       0.880 * cos(3.67349449376  + 25.27279426550 *T)
    sumL1  +=       1.203 * cos(1.39749267410  + 337.73251065900 *T)
    sumL1  +=       0.867 * cos(0.11684071625  + 267.47376618580 *T)
    sumL1  +=       0.879 * cos(6.12222682852  + 867.42347575360 *T)
    sumL1  +=       1.080 * cos(0.15038819285  + 175.16605980020 *T)
    sumL1  +=       0.988 * cos(3.12456192471  + 214.78356814630 *T)
    sumL1  +=       0.889 * cos(4.70508769146  + 148.07872442630 *T)
    sumL1  +=       0.827 * cos(6.08977582217  + 488.84961647110 *T)
    sumL1  +=       0.889 * cos(5.05124166027  + 220.46082654860 *T)
    sumL1  +=       0.828 * cos(6.27262544155  + 1382.88734684660 *T)
    sumL1  +=       1.040 * cos(5.76735098196  + 501.37978944330 *T)
    sumL1  +=       1.103 * cos(0.48706477230  + 692.58748435350 *T)
    sumL1  +=       0.810 * cos(2.50362385080  + 2310.72231481400 *T)
    sumL1  +=       0.850 * cos(4.55410385197  + 77.96299230500 *T)
    sumL1  +=       1.108 * cos(5.31792012163  + 235.39049596580 *T)
    sumL1  +=       0.790 * cos(0.89213206336  + 342.25536475310 *T)
    sumL1  +=       0.775 * cos(2.85873930879  + 211.81462272970 *T)
    sumL1  +=       0.842 * cos(2.99884993009  + 2737.32050569000 *T)
    sumL1  +=       0.784 * cos(0.05748459240  + 543.02428721890 *T)
    sumL1  +=       0.754 * cos(5.18317747668  + 244.31858407500 *T)
    sumL1  +=       0.969 * cos(1.31760425414  + 486.40193591630 *T)
    sumL1  +=       0.943 * cos(5.48641674428  + 339.28641933650 *T)
    sumL1  +=       0.759 * cos(6.25347177163  + 151.04766984290 *T)
    sumL1  +=       0.710 * cos(2.41619968810  + 247.23934538180 *T)
    sumL1  +=       0.794 * cos(2.59522645936  + 1.64453140270 *T)
    sumL1  +=       0.857 * cos(1.99318788624  + 248.72381809010 *T)
    sumL1  +=       0.717 * cos(4.56798357445  + 121.25202148330 *T)
    sumL1  +=       0.671 * cos(2.50955477476  + 444.75743814070 *T)
    sumL1  +=       0.683 * cos(5.51033310275  + 487.36514376280 *T)
    sumL1  +=       0.684 * cos(0.01892628603  + 228.27694896500 *T)
    sumL1  +=       0.665 * cos(1.47172657769  + 427.56139872250 *T)
    sumL1  +=       0.761 * cos(4.61079808671  + 23.57587323610 *T)
    sumL1  +=       0.807 * cos(3.21513718120  + 1898.35121793960 *T)
    sumL1  +=       0.645 * cos(1.92436523628  + 2950.61960112800 *T)
    sumL1  +=       0.624 * cos(6.05830190539  + 241.61027108930 *T)
    sumL1  +=       0.699 * cos(4.02804515616  + 425.63498302950 *T)
    sumL1  +=       0.624 * cos(5.85966148394  + 696.51963761660 *T)
    sumL1  +=       0.620 * cos(1.86426453489  + 2207.62954059540 *T)
    sumL1  +=       0.641 * cos(5.69868017561  + 319.57326339430 *T)
    sumL1  +=       0.646 * cos(3.78920578728  + 1038.04128918680 *T)
    sumL1  +=       0.672 * cos(2.54160055954  + 271.40591944890 *T)
    sumL1  +=       0.768 * cos(1.80484245332  + 2324.94940881560 *T)
    sumL1  +=       0.737 * cos(1.50539891226  + 268.43697403230 *T)
    sumL1  +=       0.836 * cos(1.26583811010  + 212.54833591260 *T)
    sumL1  +=       0.753 * cos(5.27536166240  + 204.70107572890 *T)
    sumL1  +=       0.633 * cos(2.19920009577  + 1802.37199072180 *T)
    sumL1  +=       0.720 * cos(2.58587107868  + 472.17484191470 *T)
    sumL1  +=       0.683 * cos(3.83223866420  + 43.28902917830 *T)
    sumL1  +=       0.740 * cos(6.21601938401  + 556.51766803760 *T)
    sumL1  +=       0.795 * cos(1.14460330178  + 381.35160823740 *T)
    sumL1  +=       0.678 * cos(3.65930963429  + 2097.42321937600 *T)
    sumL1  +=       0.568 * cos(5.92158661090  + 2428.04218303420 *T)
    sumL1  +=       0.570 * cos(1.18024241664  + 131.54696222180 *T)
    sumL1  +=       0.566 * cos(4.74157739398  + 380.12776796000 *T)
    sumL1  +=       0.586 * cos(5.71168743146  + 570.74476203920 *T)
    sumL1  +=       0.550 * cos(4.92413290959  + 188.92007304980 *T)
    sumL1  +=       0.712 * cos(2.69456114358  + 16.67477455640 *T)
    sumL1  +=       0.545 * cos(5.38725529600  + 206.23373254700 *T)
    sumL1  +=       0.572 * cos(5.79167804981  + 195.89060769870 *T)
    sumL1  +=       0.602 * cos(5.81756794592  + 963.40270297140 *T)
    sumL1  +=       0.588 * cos(4.25026865253  + 426.64637498580 *T)
    sumL1  +=       0.563 * cos(3.28295055824  + 193.65537546500 *T)
    sumL1  +=       0.583 * cos(5.44099997963  + 526.50957135690 *T)
    sumL1  +=       0.679 * cos(4.45748326743  + 105.54045477340 *T)
    sumL1  +=       0.516 * cos(5.99843937287  + 289.56516671360 *T)
    sumL1  +=       0.520 * cos(2.19322568805  + 180.16199464630 *T)
    sumL1  +=       0.543 * cos(4.19333695628  + 213.18722085340 *T)
    sumL1  +=       0.586 * cos(3.03470168346  + 6275.96230299060 *T)
    sumL1  +=       0.572 * cos(3.96788877624  + 140.00196957900 *T)
    sumL1  +=       0.611 * cos(4.15392239870  + 436.15941843160 *T)
    sumL1  +=       0.505 * cos(2.95739392583  + 135.54855145410 *T)
    sumL1  +=       0.587 * cos(4.55320395537  + 5863.59120611620 *T)
    sumL1  +=       0.492 * cos(2.71595874382  + 84.93352695390 *T)
    sumL1  +=       0.576 * cos(5.98300938454  + 9793.80090233580 *T)
    sumL1  +=       0.489 * cos(5.68450383182  + 533.62311835770 *T)
    sumL1  +=       0.519 * cos(3.09688510923  + 327.43756992050 *T)
    sumL1  +=       0.486 * cos(5.24220804875  + 5849.36411211460 *T)
    sumL1  +=       0.475 * cos(4.51295931678  + 411.62033734900 *T)
    sumL1  +=       0.540 * cos(4.44843952768  + 10206.17199921020 *T)
    sumL1  +=       0.479 * cos(0.87707794164  + 207.14875628370 *T)
    sumL1  +=       0.468 * cos(0.46572028197  + 306.09692891810 *T)
    sumL1  +=       0.586 * cos(0.86387928244  + 2538.24850425360 *T)
    sumL1  +=       0.475 * cos(6.19152982788  + 397.39324334740 *T)
    sumL1  +=       0.541 * cos(1.47958133221  + 42.53826965290 *T)
    sumL1  +=       0.496 * cos(6.07879620658  + 576.16138801060 *T)
    sumL1  +=       0.447 * cos(2.59259132013  + 7.22542158540 *T)
    sumL1  +=       0.445 * cos(5.06827300470  + 778.41478318470 *T)
    sumL1  +=       0.560 * cos(0.00461017471  + 221.37585028530 *T)
    sumL1  +=       0.456 * cos(4.60143715337  + 710.74673161820 *T)
    sumL1  +=       0.449 * cos(5.79223649465  + 685.47393735270 *T)
    sumL1  +=       0.501 * cos(1.91370965325  + 831.10498122420 *T)
    sumL1  +=       0.595 * cos(4.90329839607  + 824.74219374880 *T)
    sumL1  +=       0.447 * cos(4.88662794571  + 429.04587143080 *T)
    sumL1  +=       0.445 * cos(1.74764943142  + 525.75881183150 *T)
    sumL1  +=       0.457 * cos(0.80892712530  + 458.84151979040 *T)
    sumL1  +=       0.543 * cos(2.60317945475  + 213.41097002260 *T)
    sumL1  +=       0.493 * cos(0.61947189193  + 41.05379694460 *T)
    sumL1  +=       0.455 * cos(2.69847252264  + 3053.71237534660 *T)
    sumL1  +=       0.429 * cos(3.89071982978  + 92.79783348010 *T)
    sumL1  +=       0.411 * cos(1.34981168865  + 27.08733537390 *T)
    sumL1  +=       0.448 * cos(1.84775051361  + 980.66817835880 *T)
    sumL1  +=       0.445 * cos(4.21745990439  + 905.88657979150 *T)
    sumL1  +=       0.403 * cos(2.33067250642  + 2627.11418447060 *T)
    sumL1  +=       0.404 * cos(5.00179215709  + 431.26405732200 *T)
    sumL1  +=       0.384 * cos(1.65634584042  + 241.75328344120 *T)
    sumL1  +=       0.410 * cos(0.76907037678  + 395.57870223900 *T)
    sumL1  +=       0.456 * cos(1.98353741244  + 213.51154375910 *T)
    sumL1  +=       0.459 * cos(2.04878772547  + 285.63301345050 *T)
    sumL1  +=       0.396 * cos(5.04141834913  + 298.23262239190 *T)
    sumL1  +=       0.377 * cos(5.68073822097  + 2744.43405269080 *T)
    sumL1  +=       0.415 * cos(4.41600504868  + 179.35884549420 *T)
    sumL1  +=       0.396 * cos(4.29872851950  + 206.70681329900 *T)
    sumL1  +=       0.389 * cos(5.69091953122  + 849.26422848890 *T)
    sumL1  +=       0.369 * cos(1.36192003466  + 835.03713448730 *T)
    sumL1  +=       0.374 * cos(0.41402282126  + 9779.57380833420 *T)
    sumL1  +=       0.379 * cos(1.72255764532  + 184.98791978670 *T)
    sumL1  +=       0.365 * cos(5.88205574821  + 19.64371997300 *T)
    sumL1  +=       0.456 * cos(4.81297899859  + 213.08664711690 *T)
    sumL1  +=       0.359 * cos(1.06819138836  + 206.13736432740 *T)
    sumL1  +=       0.367 * cos(1.14184327929  + 569.04784100980 *T)
    sumL1  +=       0.352 * cos(3.04388401587  + 638.41281360570 *T)
    sumL1  +=       0.463 * cos(1.55834877017  + 421.18156490460 *T)
    sumL1  +=       0.459 * cos(5.34648461645  + 699.70103135430 *T)
    sumL1  +=       0.383 * cos(4.05921035379  + 739.80866679490 *T)
    sumL1  +=       0.354 * cos(1.09760553168  + 738.79727483860 *T)
    sumL1  +=       0.382 * cos(0.05348541587  + 252.65597135320 *T)
    sumL1  +=       0.344 * cos(1.18536656224  + 439.12836384820 *T)
    sumL1  +=       0.382 * cos(2.10483762147  + 532.61172640140 *T)
    sumL1  +=       0.361 * cos(0.50215018154  + 50.40257617910 *T)
    sumL1  +=       0.351 * cos(3.49546336297  + 1354.43315884340 *T)
    sumL1  +=       0.395 * cos(4.26278871560  + 432.22726516850 *T)
    sumL1  +=       0.345 * cos(2.38455893509  + 426.07692601420 *T)
    sumL1  +=       0.350 * cos(1.51541607946  + 259.76951835400 *T)
    sumL1  +=       0.426 * cos(5.29998227949  + 934.94851496820 *T)
    sumL1  +=       0.339 * cos(5.59774645356  + 519.39602435610 *T)
    sumL1  +=       0.388 * cos(3.40083809779  + 2413.81508903260 *T)
    sumL1  +=       0.324 * cos(3.68352014131  + 72.07328558160 *T)
    sumL1  +=       0.323 * cos(1.79597508586  + 405.99126305650 *T)
    sumL1  +=       0.366 * cos(3.56764349139  + 1119.18567522950 *T)
    sumL1  +=       0.358 * cos(4.11241839677  + 37.87240320690 *T)
    sumL1  +=       0.423 * cos(1.45116702108  + 2641.34127847220 *T)
    sumL1  +=       0.314 * cos(0.68465789313  + 757.21715453420 *T)
    sumL1  +=       0.320 * cos(3.12697568936  + 945.99421523210 *T)
    sumL1  +=       0.338 * cos(4.89782013581  + 898.77303279070 *T)
    sumL1  +=       0.319 * cos(5.76881401291  + 69.36497259590 *T)
    sumL1  +=       0.310 * cos(5.35598720822  + 815.06334611420 *T)
    sumL1  +=       0.369 * cos(4.46143610142  + 421.93232443000 *T)
    sumL1  +=       0.311 * cos(2.19275640712  + 5856.47765911540 *T)
    sumL1  +=       0.306 * cos(2.99917010799  + 1130.23137549340 *T)
    sumL1  +=       0.330 * cos(0.64102961163  + 558.00214074590 *T)
    sumL1  +=       0.305 * cos(0.40963115602  + 661.23792731640 *T)
    sumL1  +=       0.320 * cos(3.29267319940  + 760.25553592000 *T)
    sumL1  +=       0.298 * cos(5.48693246086  + 702.14871190910 *T)
    sumL1  +=       0.352 * cos(2.18179692198  + 2118.76386037840 *T)
    sumL1  +=       0.299 * cos(5.94980651345  + 572.22923474750 *T)
    sumL1  +=       0.343 * cos(2.62900083650  + 213.55972786890 *T)
    sumL1  +=       0.296 * cos(4.12563821701  + 73.29712585900 *T)
    sumL1  +=       0.360 * cos(2.94387423457  + 2214.74308759620 *T)
    sumL1  +=       0.293 * cos(5.71837797264  + 60.76695288680 *T)
    sumL1  +=       0.326 * cos(1.93806509331  + 480.77286162380 *T)
    sumL1  +=       0.335 * cos(2.60120542851  + 518.38463239980 *T)
    sumL1  +=       0.322 * cos(2.89685459163  + 427.11945573780 *T)
    sumL1  +=       0.367 * cos(2.20489848330  + 518.64526483070 *T)
    sumL1  +=       0.361 * cos(3.31464351282  + 630.33605875840 *T)
    sumL1  +=       0.288 * cos(0.87760478150  + 887.72733252680 *T)
    sumL1  +=       0.290 * cos(0.24071300709  + 705.11765732570 *T)
    sumL1  +=       0.332 * cos(5.96464701829  + 100.64509366380 *T)
    sumL1  +=       0.284 * cos(1.58760551116  + 681.54178408960 *T)
    sumL1  +=       0.281 * cos(1.68339116394  + 3267.01147078460 *T)
    sumL1  +=       0.287 * cos(3.54730637851  + 756.32338265690 *T)
    sumL1  +=       0.331 * cos(2.74250642576  + 22.89454967990 *T)
    sumL1  +=       0.281 * cos(4.79802388453  + 409.92341631960 *T)
    sumL1  +=       0.372 * cos(1.08754087151  + 426.55000676620 *T)
    sumL1  +=       0.340 * cos(0.59629116557  + 627.36711334180 *T)
    sumL1  +=       0.325 * cos(4.07319450014  + 511.53171782990 *T)
    sumL1  +=       0.273 * cos(0.71334827688  + 305.08553696180 *T)
    sumL1  +=       0.272 * cos(1.76124839309  + 945.24345570670 *T)
    sumL1  +=       0.295 * cos(4.00327005783  + 432.74853003030 *T)
    sumL1  +=       0.271 * cos(5.28903262032  + 1080.72257119160 *T)
    sumL1  +=       0.276 * cos(3.89192411657  + 610.69233878540 *T)
    sumL1  +=       0.294 * cos(2.80121651058  + 724.83081326790 *T)
    sumL1  +=       0.319 * cos(5.24824059915  + 229.97386999440 *T)
    sumL1  +=       0.264 * cos(2.36406383589  + 731.94436026870 *T)
    sumL1  +=       0.288 * cos(4.67818844930  + 170.76082578510 *T)
    sumL1  +=       0.326 * cos(3.81328980623  + 525.49817940060 *T)
    sumL1  +=       0.283 * cos(3.52027709716  + 319.31263096340 *T)
    sumL1  +=       0.264 * cos(0.25871603855  + 494.26624244250 *T)
    sumL1  +=       0.261 * cos(4.08135671345  + 25.86349509650 *T)
    sumL1  +=       0.296 * cos(4.49129913731  + 693.55069220000 *T)
    sumL1  +=       0.292 * cos(0.65370180027  + 25867.49049913539 *T)
    sumL1  +=       0.292 * cos(0.12510953311  + 25881.71759313700 *T)
    sumL1  +=       0.254 * cos(4.03912322565  + 990.22940591440 *T)
    sumL1  +=       0.288 * cos(3.98604904657  + 707.77778620160 *T)
    sumL1  +=       0.285 * cos(1.92328297431  + 3134.42687826260 *T)
    sumL1  +=       0.284 * cos(2.45411523294  + 3120.19978426100 *T)
    sumL1  +=       0.256 * cos(3.63282757780  + 430.79097657000 *T)
    sumL1  +=       0.283 * cos(2.51091647682  + 286.59622129700 *T)
    sumL1  +=       0.325 * cos(4.33261281211  + 732.69511979410 *T)
    sumL1  +=       0.264 * cos(0.05450228136  + 650.19222705250 *T)
    sumL1  +=       0.273 * cos(4.90735780421  + 409.18970313670 *T)
    sumL1  +=       0.304 * cos(4.61759348542  + 468.24268865160 *T)
    sumL1  +=       0.285 * cos(5.72467903890  + 33.94024994380 *T)
    sumL1  +=       0.242 * cos(5.28336514054  + 403.02231763990 *T)
    sumL1  +=       0.270 * cos(0.51583145648  + 263.70167161710 *T)
    sumL1  +=       0.263 * cos(4.81670787366  + 1055.44977692610 *T)
    sumL1  +=       0.237 * cos(2.92617048443  + 913.96333463880 *T)
    sumL1  +=       0.246 * cos(2.19675150666  + 2943.50605412720 *T)
    sumL1  +=       0.278 * cos(4.58404840578  + 398.14400287280 *T)
    sumL1  +=       0.234 * cos(2.64374114605  + 739.05790726950 *T)
    sumL1  +=       0.229 * cos(3.80445074468  + 58.10682401090 *T)
    sumL1  +=       0.300 * cos(2.06111081979  + 429.51895218280 *T)
    sumL1  +=       0.223 * cos(3.39888651505  + 188.02630117250 *T)
    sumL1  +=       0.301 * cos(2.96411385108  + 624.91943278700 *T)
    sumL1  +=       0.221 * cos(1.79137414078  + 2524.02141025200 *T)
    sumL1  +=       0.220 * cos(0.95686592581  + 1894.41906467650 *T)
    sumL1  +=       0.225 * cos(4.30669421945  + 637.44960575920 *T)
    sumL1  +=       0.214 * cos(1.70442143644  + 658.05653357870 *T)
    sumL1  +=       0.227 * cos(3.22613053351  + 638.93407846750 *T)
    sumL1  +=       0.220 * cos(2.66798936385  + 953.10776223290 *T)
    sumL1  +=       0.253 * cos(3.09377787768  + 29.20494752860 *T)
    sumL1  +=       0.244 * cos(3.15828383212  + 7.00167241620 *T)
    sumL1  +=       0.295 * cos(4.95843934543  + 714.67888488130 *T)
    sumL1  +=       0.209 * cos(0.94525938634  + 864.24208201590 *T)
    sumL1  +=       0.216 * cos(0.12221236180  + 28.57180808220 *T)
    sumL1  +=       0.214 * cos(2.80190604605  + 373.90799283650 *T)
    sumL1  +=       0.212 * cos(2.07343849515  + 1357.61455258110 *T)
    sumL1  +=       0.216 * cos(1.25531205533  + 477.80391620720 *T)
    sumL1  +=       0.206 * cos(5.35971491902  + 3060.82592234740 *T)
    sumL1  +=       0.204 * cos(3.08579410460  + 67.66805156650 *T)
    sumL1  +=       0.210 * cos(1.91489853604  + 938.12990870590 *T)
    sumL1  +=       0.209 * cos(1.46554109301  + 952.35700270750 *T)
    sumL1  +=       0.202 * cos(3.57670882297  + 334.55111692130 *T)
    sumL1  +=       0.228 * cos(5.66209641464  + 1699.27921650320 *T)
    sumL1  +=       0.197 * cos(4.61055255182  + 464.73122651380 *T)
    sumL1  +=       0.193 * cos(4.24606721746  + 141.69889060840 *T)
    sumL1  +=       0.266 * cos(0.69665031373  + 2854.64037391020 *T)
    sumL1  +=       0.227 * cos(1.31845358943  + 230.70758317730 *T)
    sumL1  +=       0.192 * cos(5.26739976413  + 504.56118318100 *T)
    sumL1  +=       0.187 * cos(0.85537192230  + 273.85360000370 *T)
    sumL1  +=       0.199 * cos(3.91291687807  + 418.52143602870 *T)
    sumL1  +=       0.192 * cos(6.15674105214  + 611.44309831080 *T)
    sumL1  +=       0.210 * cos(1.47873602747  + 205.43478891180 *T)
    sumL1  +=       0.194 * cos(2.37167703302  + 3370.10424500320 *T)
    sumL1  +=       0.228 * cos(2.15266015145  + 55.13787859430 *T)
    sumL1  +=       0.201 * cos(2.71380671608  + 586.31331639720 *T)
    sumL1  +=       0.194 * cos(3.29560033731  + 1670.82502850000 *T)
    sumL1  +=       0.201 * cos(4.23447633663  + 1493.09366806600 *T)
    sumL1  +=       0.181 * cos(3.61567262848  + 9786.68735533500 *T)
    sumL1  +=       0.181 * cos(2.83211558346  + 1262.38608488870 *T)
    sumL1  +=       0.242 * cos(4.69869158516  + 1141.13406340540 *T)
    sumL1  +=       0.184 * cos(4.66807336402  + 1251.34038462480 *T)
    sumL1  +=       0.221 * cos(2.25887876254  + 355.74874557180 *T)
    sumL1  +=       0.200 * cos(1.17340443616  + 4952.06359328620 *T)
    sumL1  +=       0.222 * cos(2.23360866067  + 2435.15573003500 *T)
    sumL1  +=       0.175 * cos(0.04701598422  + 107.75864066460 *T)
    sumL1  +=       0.171 * cos(5.02500742690  + 93.53154666300 *T)
    sumL1  +=       0.184 * cos(5.19723697138  + 835.78789401270 *T)
    sumL1  +=       0.221 * cos(4.49141283681  + 913.00012679230 *T)
    sumL1  +=       0.195 * cos(0.92088046109  + 551.03160609700 *T)
    sumL1  +=       0.166 * cos(5.01778115937  + 354.99798604640 *T)
    sumL1  +=       0.165 * cos(2.26267552932  + 406.95447090300 *T)
    sumL1  +=       0.189 * cos(0.31221126958  + 420.96911658350 *T)
    sumL1  +=       0.196 * cos(2.70333585839  + 774.48262992160 *T)
    sumL1  +=       0.176 * cos(6.12029409039  + 181.80652604900 *T)
    sumL1  +=       0.172 * cos(1.94132177757  + 3259.89792378380 *T)
    sumL1  +=       0.160 * cos(0.55319954265  + 5429.87946823940 *T)
    sumL1  +=       0.161 * cos(2.88623631474  + 184.84490743480 *T)
    sumL1  +=       0.192 * cos(0.26639534884  + 295.19424100610 *T)
    sumL1  +=       0.167 * cos(3.71345214172  + 1056.20053645150 *T)
    sumL1  +=       0.195 * cos(4.83926717598  + 1596.18644228460 *T)
    sumL1  +=       0.156 * cos(2.81916058733  + 428.08266358430 *T)
    sumL1  +=       0.215 * cos(1.88276472005  + 220.36445832900 *T)
    sumL1  +=       0.167 * cos(2.68872854428  + 423.67742956920 *T)
    sumL1  +=       0.154 * cos(1.66553954375  + 115.62294719080 *T)
    sumL1  +=       0.175 * cos(0.20216461467  + 384.05992122310 *T)
    sumL1  +=       0.201 * cos(4.38095931887  + 418.00017116690 *T)
    sumL1  +=       0.167 * cos(1.86485857353  + 393.46109008430 *T)
    sumL1  +=       0.155 * cos(0.92480392431  + 282.66406803390 *T)
    sumL1  +=       0.146 * cos(1.97663966745  + 9360.08916445900 *T)
    sumL1  +=       0.160 * cos(2.62483919699  + 353.30106501700 *T)
    sumL1  +=       0.186 * cos(1.37307151419  + 292.01284726840 *T)
    sumL1  +=       0.198 * cos(1.15631374887  + 2957.73314812880 *T)
    sumL1  +=       0.144 * cos(4.82956915076  + 453.42489381900 *T)
    sumL1  +=       0.149 * cos(3.60682821788  + 205.66428357540 *T)
    sumL1  +=       0.147 * cos(4.48377791879  + 81.89514556810 *T)
    sumL1  +=       0.147 * cos(5.74795037748  + 856.37777548970 *T)
    sumL1  +=       0.142 * cos(3.53823120158  + 212.02707105080 *T)
    sumL1  +=       0.140 * cos(0.70476909062  + 640.86049416050 *T)
    sumL1  +=       0.139 * cos(1.39047667205  + 1261.63532536330 *T)
    sumL1  +=       0.153 * cos(3.29559426243  + 391.17346822390 *T)
    sumL1  +=       0.158 * cos(1.79872341304  + 326.68681039510 *T)
    sumL1  +=       0.174 * cos(3.98677435872  + 1049.08698945070 *T)
    sumL1  +=       0.171 * cos(4.16825100469  + 213.03846300710 *T)
    sumL1  +=       0.133 * cos(4.74095454922  + 0.04818410980 *T)
    sumL1  +=       0.155 * cos(5.32313618730  + 2015.67108615980 *T)
    sumL1  +=       0.158 * cos(2.67557086253  + 2531.13495725280 *T)
    sumL1  +=       0.158 * cos(4.64622526567  + 427.34895040140 *T)
    sumL1  +=       0.123 * cos(2.20103444636  + 210.59078245230 *T)
    sumL1  +=       0.160 * cos(1.85888551524  + 201.51968199120 *T)
    sumL1  +=       0.119 * cos(3.12572799769  + 238.57188970350 *T)
    sumL1  +=       0.120 * cos(4.62897224203  + 203.26478713040 *T)
    sumL1  +=       0.129 * cos(4.92592016162  + 1286.90811962880 *T)
    sumL1  +=       0.132 * cos(3.44682160054  + 156.67674413540 *T)
    sumL1  +=       0.143 * cos(0.67951827513  + 425.84743135060 *T)
    sumL1  +=       0.114 * cos(5.46519773276  + 552.58551477450 *T)
    sumL1  +=       0.132 * cos(1.76335093671  + 432.01481684740 *T)
    sumL1  +=       0.113 * cos(0.68933513038  + 450.97721326420 *T)
    sumL1  +=       0.128 * cos(2.13986068877  + 2751.54759969160 *T)
    sumL1  +=       0.123 * cos(4.59695145319  + 216.00740842370 *T)
    sumL1  +=       0.119 * cos(1.04688666457  + 462.02291352810 *T)
    sumL1  +=       0.108 * cos(5.36873170289  + 3377.21779200400 *T)
    sumL1  +=       0.142 * cos(6.24626256472  + 299.12639426920 *T)
    sumL1  +=       0.118 * cos(0.63448253510  + 369.97583957340 *T)
    sumL1  +=       0.105 * cos(2.31570619675  + 200.55647414470 *T)
    sumL1  +=       0.124 * cos(1.87110815140  + 850.01498801430 *T)
    sumL1  +=       0.106 * cos(0.55623662570  + 114.39910691340 *T)
    sumL1  +=       0.102 * cos(3.95315219638  + 361.37781986430 *T)
    sumL1  +=       0.095 * cos(4.10658529323  + 10213.28554621100 *T)
    sumL1  +=       0.097 * cos(1.13534710734  + 387.24131496080 *T)
    sumL1  +=       0.096 * cos(4.46689094543  + 401.32539661050 *T)
    sumL1  +=       0.119 * cos(2.33636675091  + 318.83955021140 *T)
    sumL1  +=       0.115 * cos(3.37508073115  + 313.94418910180 *T)
    sumL1  +=       0.106 * cos(3.73586211650  + 220.93390730060 *T)
    sumL1  +=       0.090 * cos(0.59788492023  + 227.31374111850 *T)
    sumL1  +=       0.103 * cos(5.09172929383  + 213.45915413240 *T)
    sumL1  +=       0.097 * cos(5.95268532215  + 1044.40407666220 *T)
    sumL1  +=       0.103 * cos(1.70625660572  + 213.13903674360 *T)
    sumL1  +=       0.080 * cos(0.86872596168  + 233.90602325750 *T)
    sumL1  +=       0.089 * cos(5.35990932230  + 214.19286731530 *T)
    sumL1  +=       0.080 * cos(2.69565238975  + 540.73666535850 *T)
    sumL1  +=       0.095 * cos(1.19504849611  + 460.53844081980 *T)
    sumL1  +=       0.105 * cos(0.58624363205  + 481.73606947030 *T)
    sumL1  +=       0.099 * cos(2.68841109007  + 219.89137757700 *T)
    sumL1  +=       0.098 * cos(1.59923557478  + 484.70501488690 *T)
    sumL1  +=       0.081 * cos(1.12279793521  + 420.44785172170 *T)
    sumL1  +=       0.075 * cos(4.58892231446  + 394.35486196160 *T)
    sumL1  +=       0.099 * cos(4.68895851750  + 448.68959140380 *T)
    sumL1  +=       0.076 * cos(1.66929798365  + 196.03362005060 *T)
    sumL1  +=       0.087 * cos(3.12477195090  + 857.12853501510 *T)
    sumL1  +=       0.078 * cos(5.59819387460  + 364.34676528090 *T)
    sumL1  +=       0.079 * cos(3.53267171729  + 969.62247809490 *T)

    # SATURN Longitude
    sumL2  = 0
    sumL2  +=  116441.181 * cos(1.17987850633  + 7.11354700080 *T)
    sumL2  +=   91920.844 * cos(0.07425261094  + 213.29909543800 *T)
    sumL2  +=   90592.251 * cos(0.00000000000  + 0.00000000000 *T)
    sumL2  +=   15276.909 * cos(4.06492007503  + 206.18554843720 *T)
    sumL2  +=   10631.396 * cos(0.25778277414  + 220.41264243880 *T)
    sumL2  +=   10604.979 * cos(5.40963595885  + 426.59819087600 *T)
    sumL2  +=    4265.368 * cos(1.04595556630  + 14.22709400160 *T)
    sumL2  +=    1215.527 * cos(2.91860042123  + 103.09277421860 *T)
    sumL2  +=    1164.684 * cos(4.60942128971  + 639.89728631400 *T)
    sumL2  +=    1081.967 * cos(5.69130351670  + 433.71173787680 *T)
    sumL2  +=    1020.079 * cos(0.63369182642  + 3.18139373770 *T)
    sumL2  +=    1044.754 * cos(4.04206453611  + 199.07200143640 *T)
    sumL2  +=     633.582 * cos(4.38825410036  + 419.48464387520 *T)
    sumL2  +=     549.329 * cos(5.57303134242  + 3.93215326310 *T)
    sumL2  +=     456.914 * cos(1.26840971349  + 110.20632121940 *T)
    sumL2  +=     425.100 * cos(0.20935499279  + 227.52618943960 *T)
    sumL2  +=     273.739 * cos(4.28841011784  + 95.97922721780 *T)
    sumL2  +=     161.571 * cos(1.38139149420  + 11.04570026390 *T)
    sumL2  +=     129.494 * cos(1.56586884170  + 309.27832265580 *T)
    sumL2  +=     117.008 * cos(3.88120915956  + 853.19638175200 *T)
    sumL2  +=     105.415 * cos(4.90003203599  + 647.01083331480 *T)
    sumL2  +=     100.967 * cos(0.89270493100  + 21.34064100240 *T)
    sumL2  +=      95.227 * cos(5.62561150598  + 412.37109687440 *T)
    sumL2  +=      81.948 * cos(1.02477558315  + 117.31986822020 *T)
    sumL2  +=      74.857 * cos(4.76178468163  + 210.11770170030 *T)
    sumL2  +=      82.727 * cos(6.05030934786  + 216.48048917570 *T)
    sumL2  +=      95.659 * cos(2.91093561539  + 316.39186965660 *T)
    sumL2  +=      63.696 * cos(0.35179804917  + 323.50541665740 *T)
    sumL2  +=      84.860 * cos(5.73472777961  + 209.36694217490 *T)
    sumL2  +=      60.647 * cos(4.87517850190  + 632.78373931320 *T)
    sumL2  +=      66.459 * cos(0.48297940601  + 10.29494073850 *T)
    sumL2  +=      67.184 * cos(0.45648612616  + 522.57741809380 *T)
    sumL2  +=      53.281 * cos(2.74730541387  + 529.69096509460 *T)
    sumL2  +=      45.827 * cos(5.69296621745  + 440.82528487760 *T)
    sumL2  +=      45.293 * cos(1.66856699796  + 202.25339517410 *T)
    sumL2  +=      42.330 * cos(5.70768187703  + 88.86568021700 *T)
    sumL2  +=      32.140 * cos(0.07050050346  + 63.73589830340 *T)
    sumL2  +=      31.573 * cos(1.67190022213  + 302.16477565500 *T)
    sumL2  +=      31.150 * cos(4.16379537691  + 191.95845443560 *T)
    sumL2  +=      24.631 * cos(5.65564728570  + 735.87651353180 *T)
    sumL2  +=      26.558 * cos(0.83256214407  + 224.34479570190 *T)
    sumL2  +=      20.108 * cos(5.94364609981  + 217.23124870110 *T)
    sumL2  +=      17.511 * cos(4.90014736798  + 625.67019231240 *T)
    sumL2  +=      17.130 * cos(1.62593421274  + 742.99006053260 *T)
    sumL2  +=      13.744 * cos(3.76497167300  + 195.13984817330 *T)
    sumL2  +=      12.236 * cos(4.71789723976  + 203.00415469950 *T)
    sumL2  +=      11.940 * cos(0.12620714199  + 234.63973644040 *T)
    sumL2  +=      16.040 * cos(0.57886320845  + 515.46387109300 *T)
    sumL2  +=      11.154 * cos(5.92216844780  + 536.80451209540 *T)
    sumL2  +=      14.068 * cos(0.20675293700  + 838.96928775040 *T)
    sumL2  +=      11.013 * cos(5.60207982774  + 728.76296653100 *T)
    sumL2  +=      11.718 * cos(3.12098483554  + 846.08283475120 *T)
    sumL2  +=       9.962 * cos(4.15472049127  + 860.30992875280 *T)
    sumL2  +=      10.601 * cos(3.20327613035  + 1066.49547719000 *T)
    sumL2  +=      10.072 * cos(0.25709351996  + 330.61896365820 *T)
    sumL2  +=       9.490 * cos(0.46379969328  + 956.28915597060 *T)
    sumL2  +=      10.240 * cos(4.98736656070  + 422.66603761290 *T)
    sumL2  +=       8.287 * cos(2.13990364272  + 269.92144674060 *T)
    sumL2  +=       7.238 * cos(5.39724715258  + 1052.26838318840 *T)
    sumL2  +=       7.730 * cos(5.24602742309  + 429.77958461370 *T)
    sumL2  +=       6.353 * cos(4.46211130731  + 284.14854074220 *T)
    sumL2  +=       5.935 * cos(5.40967847103  + 149.56319713460 *T)
    sumL2  +=       7.550 * cos(4.03401153929  + 9.56122755560 *T)
    sumL2  +=       5.779 * cos(4.29380891110  + 415.55249061210 *T)
    sumL2  +=       6.082 * cos(5.93416924841  + 405.25754987360 *T)
    sumL2  +=       5.711 * cos(0.01824076994  + 124.43341522100 *T)
    sumL2  +=       5.676 * cos(6.02235682150  + 223.59403617650 *T)
    sumL2  +=       4.757 * cos(4.92804854717  + 654.12438031560 *T)
    sumL2  +=       4.727 * cos(2.27461984667  + 18.15924726470 *T)
    sumL2  +=       4.509 * cos(4.40688707557  + 942.06206196900 *T)
    sumL2  +=       5.621 * cos(0.29694719379  + 127.47179660680 *T)
    sumL2  +=       5.453 * cos(5.53868222772  + 949.17560896980 *T)
    sumL2  +=       4.130 * cos(4.68673560379  + 74.78159856730 *T)
    sumL2  +=       4.098 * cos(5.30851262200  + 1045.15483618760 *T)
    sumL2  +=       4.223 * cos(2.89014939299  + 56.62235130260 *T)
    sumL2  +=       4.887 * cos(3.20022991216  + 277.03499374140 *T)
    sumL2  +=       3.905 * cos(3.30270187305  + 490.33408917940 *T)
    sumL2  +=       3.923 * cos(6.09732996823  + 81.75213321620 *T)
    sumL2  +=       3.755 * cos(4.93065184796  + 52.69019803950 *T)
    sumL2  +=       4.602 * cos(6.13908576681  + 1155.36115740700 *T)
    sumL2  +=       3.714 * cos(0.40648076787  + 137.03302416240 *T)
    sumL2  +=       3.407 * cos(4.28514461015  + 99.91138048090 *T)
    sumL2  +=       3.579 * cos(0.20402442077  + 1272.68102562720 *T)
    sumL2  +=       3.946 * cos(0.36500928968  + 12.53017297220 *T)
    sumL2  +=       3.246 * cos(1.56761884227  + 1059.38193018920 *T)
    sumL2  +=       4.063 * cos(0.29084229143  + 831.85574074960 *T)
    sumL2  +=       3.688 * cos(0.15467406177  + 437.64389113990 *T)
    sumL2  +=       2.895 * cos(3.13473183482  + 70.84944530420 *T)
    sumL2  +=       2.800 * cos(0.32727938074  + 191.20769491020 *T)
    sumL2  +=       2.672 * cos(1.87612402267  + 295.05122865420 *T)
    sumL2  +=       3.454 * cos(4.77197610696  + 423.41679713830 *T)
    sumL2  +=       2.623 * cos(5.15237415384  + 1368.66025284500 *T)
    sumL2  +=       2.457 * cos(3.89612890177  + 210.85141488320 *T)
    sumL2  +=       2.461 * cos(1.58522876760  + 32.24332891440 *T)
    sumL2  +=       2.595 * cos(3.59007068361  + 131.40394986990 *T)
    sumL2  +=       2.289 * cos(4.76825865118  + 351.81659230870 *T)
    sumL2  +=       2.357 * cos(5.83099000562  + 106.27416795630 *T)
    sumL2  +=       2.221 * cos(5.98277491515  + 6062.66320755260 *T)
    sumL2  +=       2.221 * cos(2.05930402282  + 6076.89030155420 *T)
    sumL2  +=       2.183 * cos(5.94985336393  + 145.63104387150 *T)
    sumL2  +=       2.718 * cos(3.37801252354  + 408.43894361130 *T)
    sumL2  +=       2.288 * cos(3.14000619320  + 22.09140052780 *T)
    sumL2  +=       2.090 * cos(1.12304173562  + 9992.87290377220 *T)
    sumL2  +=       2.089 * cos(3.48276230686  + 10007.09999777380 *T)
    sumL2  +=       2.570 * cos(5.12167203704  + 265.98929347750 *T)
    sumL2  +=       1.835 * cos(4.15379879659  + 1258.45393162560 *T)
    sumL2  +=       1.820 * cos(5.05340615445  + 1361.54670584420 *T)
    sumL2  +=       1.760 * cos(4.13532689228  + 107.02492748170 *T)
    sumL2  +=       1.921 * cos(4.51790997496  + 138.51749687070 *T)
    sumL2  +=       1.707 * cos(1.35864593280  + 231.45834270270 *T)
    sumL2  +=       1.956 * cos(5.87006093798  + 1471.75302706360 *T)
    sumL2  +=       2.133 * cos(5.23409848720  + 1265.56747862640 *T)
    sumL2  +=       1.595 * cos(5.61962698786  + 447.93883187840 *T)
    sumL2  +=       1.609 * cos(3.74893709671  + 628.85158605010 *T)
    sumL2  +=       1.490 * cos(0.48352404940  + 340.77089204480 *T)
    sumL2  +=       1.560 * cos(5.97095003614  + 430.53034413910 *T)
    sumL2  +=       1.352 * cos(0.71405348653  + 28.45418800320 *T)
    sumL2  +=       1.355 * cos(2.91219493604  + 215.74677599280 *T)
    sumL2  +=       1.298 * cos(5.84254169775  + 543.91805909620 *T)
    sumL2  +=       1.664 * cos(6.23834873469  + 1148.24761040620 *T)
    sumL2  +=       1.205 * cos(2.83373725021  + 200.76892246580 *T)
    sumL2  +=       1.192 * cos(3.52219428945  + 497.44763618020 *T)
    sumL2  +=       1.122 * cos(2.60571030270  + 1279.79457262800 *T)
    sumL2  +=       1.217 * cos(6.23528359211  + 1589.07289528380 *T)
    sumL2  +=       1.420 * cos(0.85079202155  + 6069.77675455340 *T)
    sumL2  +=       1.120 * cos(4.95656566453  + 1685.05212250160 *T)
    sumL2  +=       1.010 * cos(3.39689646619  + 1073.60902419080 *T)
    sumL2  +=       1.352 * cos(2.27575429523  + 9999.98645077300 *T)
    sumL2  +=       0.979 * cos(1.58571463442  + 1375.77379984580 *T)
    sumL2  +=       1.159 * cos(0.71823181781  + 508.35032409220 *T)
    sumL2  +=       1.014 * cos(2.40759054741  + 703.63318461740 *T)
    sumL2  +=       0.956 * cos(2.66256831556  + 134.58534360760 *T)
    sumL2  +=       1.110 * cos(1.19713920197  + 618.55664531160 *T)
    sumL2  +=       0.945 * cos(4.68155456977  + 362.86229257260 *T)
    sumL2  +=       0.953 * cos(4.20749172571  + 288.08069400530 *T)
    sumL2  +=       1.033 * cos(1.08781255146  + 184.84490743480 *T)
    sumL2  +=       0.942 * cos(2.43465223460  + 222.86032299360 *T)
    sumL2  +=       0.909 * cos(4.51769385360  + 38.13303563780 *T)
    sumL2  +=       1.002 * cos(1.38543153271  + 483.22054217860 *T)
    sumL2  +=       1.082 * cos(4.52832816548  + 635.96513305090 *T)
    sumL2  +=       1.008 * cos(4.91325851448  + 750.10360753340 *T)
    sumL2  +=       0.862 * cos(4.79998518474  + 1677.93857550080 *T)
    sumL2  +=       0.828 * cos(2.21940849017  + 333.65734504400 *T)
    sumL2  +=       0.745 * cos(3.97279299984  + 1574.84580128220 *T)
    sumL2  +=       0.903 * cos(5.58963782799  + 1788.14489672020 *T)
    sumL2  +=       0.735 * cos(2.28191723259  + 1162.47470440780 *T)
    sumL2  +=       0.773 * cos(5.82270096882  + 416.30325013750 *T)
    sumL2  +=       0.734 * cos(2.35356586018  + 120.35824960600 *T)
    sumL2  +=       0.745 * cos(4.84266000843  + 76.26607127560 *T)
    sumL2  +=       0.765 * cos(2.50840146722  + 343.21857259960 *T)
    sumL2  +=       0.908 * cos(5.01046293458  + 1581.95934828300 *T)
    sumL2  +=       0.707 * cos(3.66631544506  + 347.88443904560 *T)
    sumL2  +=       0.870 * cos(0.77106152694  + 113.38771495710 *T)
    sumL2  +=       0.686 * cos(2.88543836068  + 92.04707395470 *T)
    sumL2  +=       0.673 * cos(3.75650667651  + 203.73786788240 *T)
    sumL2  +=       0.656 * cos(3.77718582702  + 217.96496188400 *T)
    sumL2  +=       0.675 * cos(5.62875135263  + 17.26547538740 *T)
    sumL2  +=       0.691 * cos(0.21330089609  + 99.16062095550 *T)
    sumL2  +=       0.786 * cos(4.49318079175  + 643.07868005170 *T)
    sumL2  +=       0.641 * cos(0.67588390141  + 46.47042291600 *T)
    sumL2  +=       0.663 * cos(5.74837848383  + 721.64941953020 *T)
    sumL2  +=       0.809 * cos(5.94893988352  + 1464.63948006280 *T)
    sumL2  +=       0.638 * cos(4.86195439622  + 357.44566660120 *T)
    sumL2  +=       0.740 * cos(6.00053422445  + 337.73251065900 *T)
    sumL2  +=       0.555 * cos(4.95858934298  + 358.93013930950 *T)
    sumL2  +=       0.581 * cos(3.87669679805  + 565.11568774670 *T)
    sumL2  +=       0.541 * cos(1.22296838713  + 62.25142559510 *T)
    sumL2  +=       0.697 * cos(0.00715950269  + 1169.58825140860 *T)
    sumL2  +=       0.524 * cos(1.53830423608  + 195.89060769870 *T)
    sumL2  +=       0.518 * cos(5.41992758537  + 312.19908396260 *T)
    sumL2  +=       0.626 * cos(5.26580317026  + 436.89313161450 *T)
    sumL2  +=       0.537 * cos(6.17031657600  + 182.27960680100 *T)
    sumL2  +=       0.574 * cos(5.98607898826  + 1905.46476494040 *T)
    sumL2  +=       0.541 * cos(0.30589337713  + 98.89998852460 *T)
    sumL2  +=       0.603 * cos(3.26888470585  + 208.63322899200 *T)
    sumL2  +=       0.504 * cos(3.80930996688  + 168.05251279940 *T)
    sumL2  +=       0.477 * cos(3.56642391994  + 563.63121503840 *T)
    sumL2  +=       0.511 * cos(4.70719837179  + 2001.44399215820 *T)
    sumL2  +=       0.475 * cos(1.06025557585  + 5856.47765911540 *T)
    sumL2  +=       0.540 * cos(0.87230551412  + 1692.16566950240 *T)
    sumL2  +=       0.454 * cos(2.48128029368  + 9786.68735533500 *T)
    sumL2  +=       0.456 * cos(3.18303484133  + 218.92816973050 *T)
    sumL2  +=       0.462 * cos(0.71358186864  + 258.87574647670 *T)
    sumL2  +=       0.424 * cos(4.89778948357  + 636.71589257630 *T)
    sumL2  +=       0.537 * cos(2.59376221736  + 313.21047591890 *T)
    sumL2  +=       0.410 * cos(4.22147787617  + 867.42347575360 *T)
    sumL2  +=       0.408 * cos(3.06057772788  + 424.15051032120 *T)
    sumL2  +=       0.407 * cos(3.79376013938  + 24.37902238820 *T)
    sumL2  +=       0.569 * cos(3.68547825941  + 350.33211960040 *T)
    sumL2  +=       0.404 * cos(0.91401255827  + 114.13847448250 *T)
    sumL2  +=       0.395 * cos(3.50478374207  + 129.91947716160 *T)
    sumL2  +=       0.395 * cos(2.86309689622  + 212.33588759150 *T)
    sumL2  +=       0.386 * cos(5.00762729432  + 388.46515523820 *T)
    sumL2  +=       0.393 * cos(6.26835522096  + 241.75328344120 *T)
    sumL2  +=       0.401 * cos(4.60258908692  + 1994.33044515740 *T)
    sumL2  +=       0.385 * cos(0.91582119643  + 160.60889739850 *T)
    sumL2  +=       0.467 * cos(0.54876489832  + 404.50679034820 *T)
    sumL2  +=       0.368 * cos(0.35674031808  + 214.26230328450 *T)
    sumL2  +=       0.471 * cos(0.67360047481  + 207.88246946660 *T)
    sumL2  +=       0.379 * cos(0.92901327825  + 767.36908292080 *T)
    sumL2  +=       0.420 * cos(5.69797398044  + 225.82926841020 *T)
    sumL2  +=       0.356 * cos(3.10092792842  + 842.15068148810 *T)
    sumL2  +=       0.428 * cos(5.35375368944  + 2104.53676637680 *T)
    sumL2  +=       0.422 * cos(2.67975581832  + 77.96299230500 *T)
    sumL2  +=       0.370 * cos(5.46144813372  + 1038.04128918680 *T)
    sumL2  +=       0.379 * cos(5.56429091578  + 131.54696222180 *T)
    sumL2  +=       0.441 * cos(5.68196668399  + 1781.03134971940 *T)
    sumL2  +=       0.361 * cos(5.20616019966  + 629.60234557550 *T)
    sumL2  +=       0.341 * cos(5.92928351979  + 26.82670294300 *T)
    sumL2  +=       0.419 * cos(5.26851686707  + 85.82729883120 *T)
    sumL2  +=       0.322 * cos(0.80223983857  + 6283.07584999140 *T)
    sumL2  +=       0.323 * cos(3.86700993914  + 576.16138801060 *T)
    sumL2  +=       0.321 * cos(2.17186032970  + 10213.28554621100 *T)
    sumL2  +=       0.355 * cos(2.80560859177  + 344.70304530790 *T)
    sumL2  +=       0.311 * cos(3.77477255556  + 1891.23767093880 *T)
    sumL2  +=       0.318 * cos(5.22020784209  + 142.44965013380 *T)
    sumL2  +=       0.315 * cos(0.52272202855  + 5849.36411211460 *T)
    sumL2  +=       0.428 * cos(4.63722058283  + 1898.35121793960 *T)
    sumL2  +=       0.337 * cos(0.68198429948  + 45.57665103870 *T)
    sumL2  +=       0.316 * cos(0.54074780109  + 444.75743814070 *T)
    sumL2  +=       0.310 * cos(1.41032075652  + 273.10284047830 *T)
    sumL2  +=       0.311 * cos(3.53744556230  + 251.43213107580 *T)
    sumL2  +=       0.295 * cos(1.93253015677  + 436.15941843160 *T)
    sumL2  +=       0.296 * cos(1.97705648834  + 9779.57380833420 *T)
    sumL2  +=       0.326 * cos(3.67854047003  + 963.40270297140 *T)
    sumL2  +=       0.389 * cos(5.76841276132  + 39.35687591520 *T)
    sumL2  +=       0.277 * cos(5.73995694175  + 92.79783348010 *T)
    sumL2  +=       0.315 * cos(4.96371610197  + 757.21715453420 *T)
    sumL2  +=       0.295 * cos(1.81638833900  + 1493.09366806600 *T)
    sumL2  +=       0.287 * cos(0.97698377929  + 685.47393735270 *T)
    sumL2  +=       0.281 * cos(2.66463042095  + 1286.90811962880 *T)
    sumL2  +=       0.330 * cos(5.79776922760  + 650.94298657790 *T)
    sumL2  +=       0.292 * cos(3.97858181479  + 472.17484191470 *T)
    sumL2  +=       0.266 * cos(4.13716111320  + 601.76425067620 *T)
    sumL2  +=       0.262 * cos(0.91887592474  + 245.54242435240 *T)
    sumL2  +=       0.278 * cos(3.08964256591  + 778.41478318470 *T)
    sumL2  +=       0.277 * cos(3.08750916880  + 621.73803904930 *T)
    sumL2  +=       0.255 * cos(3.93981592051  + 181.05576652360 *T)
    sumL2  +=       0.333 * cos(2.04835822938  + 561.18353448360 *T)
    sumL2  +=       0.247 * cos(2.92754257675  + 219.44943459230 *T)
    sumL2  +=       0.306 * cos(0.36127922606  + 824.74219374880 *T)
    sumL2  +=       0.253 * cos(1.80130756458  + 5643.17856367740 *T)
    sumL2  +=       0.337 * cos(4.97764462199  + 175.16605980020 *T)
    sumL2  +=       0.273 * cos(0.66599369335  + 2008.55753915900 *T)
    sumL2  +=       0.227 * cos(4.87285356383  + 661.23792731640 *T)
    sumL2  +=       0.249 * cos(3.14202895058  + 144.14657116320 *T)
    sumL2  +=       0.220 * cos(3.93526603081  + 319.57326339430 *T)
    sumL2  +=       0.212 * cos(5.85248164087  + 546.95644048200 *T)
    sumL2  +=       0.234 * cos(1.65314711167  + 554.06998748280 *T)
    sumL2  +=       0.204 * cos(0.88373842674  + 31.49256938900 *T)
    sumL2  +=       0.205 * cos(2.93169866171  + 1596.18644228460 *T)
    sumL2  +=       0.201 * cos(3.36504567824  + 1080.72257119160 *T)
    sumL2  +=       0.224 * cos(4.34612745705  + 1382.88734684660 *T)
    sumL2  +=       0.192 * cos(5.13697232918  + 329.72519178090 *T)
    sumL2  +=       0.208 * cos(3.08549771485  + 41.64449777560 *T)
    sumL2  +=       0.236 * cos(0.07998742860  + 1141.13406340540 *T)
    sumL2  +=       0.203 * cos(4.13011580915  + 2627.11418447060 *T)
    sumL2  +=       0.203 * cos(0.13969067385  + 1485.98012106520 *T)
    sumL2  +=       0.204 * cos(3.38137545713  + 699.70103135430 *T)
    sumL2  +=       0.212 * cos(4.52370676085  + 2310.72231481400 *T)
    sumL2  +=       0.218 * cos(5.79277335862  + 2221.85663459700 *T)
    sumL2  +=       0.213 * cos(0.50441377637  + 934.94851496820 *T)
    sumL2  +=       0.210 * cos(5.04017633795  + 2420.92863603340 *T)
    sumL2  +=       0.214 * cos(4.64286758581  + 2317.83586181480 *T)
    sumL2  +=       0.178 * cos(0.84588580004  + 128.36556848410 *T)
    sumL2  +=       0.170 * cos(2.75006619605  + 710.74673161820 *T)
    sumL2  +=       0.171 * cos(4.32615182967  + 291.26208774300 *T)
    sumL2  +=       0.172 * cos(3.46971306920  + 501.37978944330 *T)
    sumL2  +=       0.170 * cos(1.05408992106  + 526.50957135690 *T)
    sumL2  +=       0.162 * cos(1.15683042950  + 519.39602435610 *T)
    sumL2  +=       0.180 * cos(4.96266204107  + 1670.82502850000 *T)
    sumL2  +=       0.172 * cos(1.65385549578  + 916.93228005540 *T)
    sumL2  +=       0.170 * cos(2.30821101766  + 429.04587143080 *T)
    sumL2  +=       0.170 * cos(5.98716489326  + 643.82943957710 *T)
    sumL2  +=       0.173 * cos(5.19933564968  + 1354.43315884340 *T)
    sumL2  +=       0.195 * cos(4.50165508529  + 2214.74308759620 *T)
    sumL2  +=       0.156 * cos(4.16290662749  + 572.22923474750 *T)
    sumL2  +=       0.153 * cos(1.23776248578  + 2413.81508903260 *T)
    sumL2  +=       0.150 * cos(0.63076983213  + 1478.86657406440 *T)
    sumL2  +=       0.169 * cos(4.28090123029  + 305.34616939270 *T)
    sumL2  +=       0.174 * cos(6.23077892653  + 3384.33133900480 *T)
    sumL2  +=       0.149 * cos(3.13274908516  + 9573.38825989700 *T)
    sumL2  +=       0.162 * cos(6.25601818345  + 213.25091132820 *T)
    sumL2  +=       0.149 * cos(4.81749019484  + 945.99421523210 *T)
    sumL2  +=       0.162 * cos(0.88610129190  + 216.21985674480 *T)
    sumL2  +=       0.133 * cos(2.31915371262  + 156.67674413540 *T)
    sumL2  +=       0.165 * cos(6.06456216591  + 732.69511979410 *T)
    sumL2  +=       0.141 * cos(6.14293754333  + 1795.25844372100 *T)
    sumL2  +=       0.133 * cos(0.06530337135  + 218.71572140940 *T)
    sumL2  +=       0.162 * cos(3.17058130506  + 213.34727954780 *T)
    sumL2  +=       0.125 * cos(2.07143636845  + 425.11371816770 *T)
    sumL2  +=       0.146 * cos(1.88627500632  + 211.81462272970 *T)
    sumL2  +=       0.113 * cos(2.79541965778  + 235.39049596580 *T)
    sumL2  +=       0.117 * cos(0.76464798684  + 479.28838891550 *T)
    sumL2  +=       0.108 * cos(3.95650672786  + 570.74476203920 *T)
    sumL2  +=       0.106 * cos(0.12820734703  + 188.02630117250 *T)
    sumL2  +=       0.134 * cos(3.58244908862  + 849.26422848890 *T)
    sumL2  +=       0.114 * cos(0.25990388555  + 398.14400287280 *T)
    sumL2  +=       0.112 * cos(2.39181495831  + 217.49188113200 *T)
    sumL2  +=       0.091 * cos(2.50716605179  + 121.25202148330 *T)
    sumL2  +=       0.091 * cos(1.75367948574  + 213.82036029980 *T)
    sumL2  +=       0.088 * cos(5.26121947108  + 395.57870223900 *T)
    sumL2  +=       0.096 * cos(3.98832609364  + 289.56516671360 *T)
    sumL2  +=       0.091 * cos(0.35318362186  + 312.45971639350 *T)
    sumL2  +=       0.112 * cos(1.14387590923  + 1802.37199072180 *T)
    sumL2  +=       0.082 * cos(3.73605217214  + 207.67002114550 *T)
    sumL2  +=       0.082 * cos(6.06283262812  + 210.37833413120 *T)
    sumL2  +=       0.084 * cos(3.34470673492  + 67.66805156650 *T)
    sumL2  +=       0.086 * cos(2.73917300180  + 5863.59120611620 *T)
    sumL2  +=       0.083 * cos(2.81499116485  + 776.93031047640 *T)
    sumL2  +=       0.091 * cos(1.26160093170  + 212.77783057620 *T)
    sumL2  +=       0.090 * cos(2.08722491981  + 2111.65031337760 *T)
    sumL2  +=       0.080 * cos(2.13136842916  + 421.93232443000 *T)
    sumL2  +=       0.082 * cos(4.16358350281  + 9793.80090233580 *T)
    sumL2  +=       0.077 * cos(2.96973341607  + 431.26405732200 *T)
    sumL2  +=       0.079 * cos(3.42790361067  + 417.03696332040 *T)
    sumL2  +=       0.079 * cos(3.18693585419  + 320.32402291970 *T)
    sumL2  +=       0.080 * cos(0.78975763683  + 204.70107572890 *T)
    sumL2  +=       0.077 * cos(1.89354243952  + 556.51766803760 *T)
    sumL2  +=       0.073 * cos(4.85923277221  + 2118.76386037840 *T)
    sumL2  +=       0.071 * cos(3.64551577433  + 198.32124191100 *T)

    # SATURN Longitude
    sumL3  = 0
    sumL3  +=   16038.734 * cos(5.73945377424  + 7.11354700080 *T)
    sumL3  +=    4249.793 * cos(4.58539675603  + 213.29909543800 *T)
    sumL3  +=    1906.524 * cos(4.76082050205  + 220.41264243880 *T)
    sumL3  +=    1465.687 * cos(5.91326678323  + 206.18554843720 *T)
    sumL3  +=    1162.041 * cos(5.61973132428  + 14.22709400160 *T)
    sumL3  +=    1066.581 * cos(3.60816533142  + 426.59819087600 *T)
    sumL3  +=     239.377 * cos(3.86088273439  + 433.71173787680 *T)
    sumL3  +=     236.975 * cos(5.76826451465  + 199.07200143640 *T)
    sumL3  +=     165.641 * cos(5.11641150216  + 3.18139373770 *T)
    sumL3  +=     131.409 * cos(4.74327544615  + 227.52618943960 *T)
    sumL3  +=     151.352 * cos(2.73594641861  + 639.89728631400 *T)
    sumL3  +=      61.630 * cos(4.74287052463  + 103.09277421860 *T)
    sumL3  +=      63.365 * cos(0.22850089497  + 419.48464387520 *T)
    sumL3  +=      40.437 * cos(5.47298059144  + 21.34064100240 *T)
    sumL3  +=      40.205 * cos(5.96420266720  + 95.97922721780 *T)
    sumL3  +=      38.746 * cos(5.83386199529  + 110.20632121940 *T)
    sumL3  +=      28.025 * cos(3.01235311514  + 647.01083331480 *T)
    sumL3  +=      25.029 * cos(0.98808170740  + 3.93215326310 *T)
    sumL3  +=      18.101 * cos(1.02506397063  + 412.37109687440 *T)
    sumL3  +=      17.879 * cos(3.31913418974  + 309.27832265580 *T)
    sumL3  +=      16.208 * cos(3.89825272754  + 440.82528487760 *T)
    sumL3  +=      15.763 * cos(5.61667809625  + 117.31986822020 *T)
    sumL3  +=      19.014 * cos(1.91614237463  + 853.19638175200 *T)
    sumL3  +=      18.262 * cos(4.96738415934  + 10.29494073850 *T)
    sumL3  +=      12.947 * cos(1.18068953942  + 88.86568021700 *T)
    sumL3  +=      17.919 * cos(4.20376505349  + 216.48048917570 *T)
    sumL3  +=      11.453 * cos(5.57520615096  + 11.04570026390 *T)
    sumL3  +=      10.548 * cos(5.92906266269  + 191.95845443560 *T)
    sumL3  +=      10.389 * cos(3.94838736947  + 209.36694217490 *T)
    sumL3  +=       8.650 * cos(3.39335369698  + 302.16477565500 *T)
    sumL3  +=       7.580 * cos(4.87736913157  + 323.50541665740 *T)
    sumL3  +=       6.697 * cos(0.38198725552  + 632.78373931320 *T)
    sumL3  +=       5.864 * cos(1.05621157685  + 210.11770170030 *T)
    sumL3  +=       5.449 * cos(4.64268475485  + 234.63973644040 *T)
    sumL3  +=       6.327 * cos(2.25492722762  + 522.57741809380 *T)
    sumL3  +=       3.602 * cos(2.30677010956  + 515.46387109300 *T)
    sumL3  +=       3.229 * cos(2.20309400066  + 860.30992875280 *T)
    sumL3  +=       3.701 * cos(3.14159265359  + 0.00000000000 *T)
    sumL3  +=       2.583 * cos(4.93447677059  + 224.34479570190 *T)
    sumL3  +=       2.543 * cos(0.42393884183  + 625.67019231240 *T)
    sumL3  +=       2.213 * cos(3.19814958289  + 202.25339517410 *T)
    sumL3  +=       2.421 * cos(4.76621391814  + 330.61896365820 *T)
    sumL3  +=       2.850 * cos(0.58604395010  + 529.69096509460 *T)
    sumL3  +=       1.965 * cos(4.39525359412  + 124.43341522100 *T)
    sumL3  +=       2.154 * cos(1.35488209144  + 405.25754987360 *T)
    sumL3  +=       2.296 * cos(3.34809165905  + 429.77958461370 *T)
    sumL3  +=       2.018 * cos(3.06693569701  + 654.12438031560 *T)
    sumL3  +=       1.979 * cos(1.02981005658  + 728.76296653100 *T)
    sumL3  +=       1.868 * cos(3.09383546177  + 422.66603761290 *T)
    sumL3  +=       1.846 * cos(4.15225985450  + 536.80451209540 *T)
    sumL3  +=       2.194 * cos(1.18918501013  + 1066.49547719000 *T)
    sumL3  +=       2.090 * cos(4.15631351317  + 223.59403617650 *T)
    sumL3  +=       1.481 * cos(0.37916705169  + 316.39186965660 *T)
    sumL3  +=       1.720 * cos(5.82865773356  + 195.13984817330 *T)
    sumL3  +=       1.460 * cos(1.57663426355  + 81.75213321620 *T)
    sumL3  +=       1.623 * cos(6.03706764648  + 742.99006053260 *T)
    sumL3  +=       1.286 * cos(1.66154726117  + 63.73589830340 *T)
    sumL3  +=       1.304 * cos(5.02409881054  + 956.28915597060 *T)
    sumL3  +=       1.446 * cos(2.10575519127  + 838.96928775040 *T)
    sumL3  +=       1.245 * cos(3.88109752770  + 269.92144674060 *T)
    sumL3  +=       1.018 * cos(3.72599601656  + 295.05122865420 *T)
    sumL3  +=       1.323 * cos(1.38492882986  + 735.87651353180 *T)
    sumL3  +=       1.318 * cos(2.33460998999  + 217.23124870110 *T)
    sumL3  +=       0.943 * cos(2.75813531246  + 284.14854074220 *T)
    sumL3  +=       0.906 * cos(0.71155526266  + 846.08283475120 *T)
    sumL3  +=       0.886 * cos(3.83754799777  + 447.93883187840 *T)
    sumL3  +=       0.943 * cos(3.31480217015  + 18.15924726470 *T)
    sumL3  +=       0.800 * cos(4.71386673963  + 56.62235130260 *T)
    sumL3  +=       0.908 * cos(2.02119147951  + 831.85574074960 *T)
    sumL3  +=       0.787 * cos(0.80410269937  + 1045.15483618760 *T)
    sumL3  +=       0.709 * cos(4.27064410504  + 437.64389113990 *T)
    sumL3  +=       0.651 * cos(6.17565900032  + 942.06206196900 *T)
    sumL3  +=       0.785 * cos(2.40767785311  + 203.00415469950 *T)
    sumL3  +=       0.702 * cos(1.64585301418  + 423.41679713830 *T)
    sumL3  +=       0.543 * cos(2.86326941725  + 184.84490743480 *T)
    sumL3  +=       0.532 * cos(6.25762144463  + 1059.38193018920 *T)
    sumL3  +=       0.521 * cos(3.43013038466  + 149.56319713460 *T)
    sumL3  +=       0.484 * cos(4.88366060720  + 1272.68102562720 *T)
    sumL3  +=       0.437 * cos(5.40220619672  + 408.43894361130 *T)
    sumL3  +=       0.388 * cos(2.57589594168  + 508.35032409220 *T)
    sumL3  +=       0.421 * cos(4.05836524024  + 543.91805909620 *T)
    sumL3  +=       0.375 * cos(1.22747948298  + 2324.94940881560 *T)
    sumL3  +=       0.347 * cos(0.59237194522  + 22.09140052780 *T)
    sumL3  +=       0.433 * cos(1.69090148012  + 1155.36115740700 *T)
    sumL3  +=       0.389 * cos(1.46170367972  + 1073.60902419080 *T)
    sumL3  +=       0.307 * cos(1.82185086955  + 628.85158605010 *T)
    sumL3  +=       0.409 * cos(1.21858750514  + 1052.26838318840 *T)
    sumL3  +=       0.309 * cos(0.33610530663  + 6076.89030155420 *T)
    sumL3  +=       0.309 * cos(1.42279282226  + 6062.66320755260 *T)
    sumL3  +=       0.340 * cos(1.83325770310  + 1141.13406340540 *T)
    sumL3  +=       0.303 * cos(2.41584747330  + 127.47179660680 *T)
    sumL3  +=       0.305 * cos(5.34154702988  + 131.40394986990 *T)
    sumL3  +=       0.298 * cos(2.28594631393  + 635.96513305090 *T)
    sumL3  +=       0.372 * cos(1.03723911390  + 313.21047591890 *T)
    sumL3  +=       0.338 * cos(0.69100012338  + 1361.54670584420 *T)
    sumL3  +=       0.325 * cos(1.78816356937  + 1148.24761040620 *T)
    sumL3  +=       0.322 * cos(1.18628805010  + 721.64941953020 *T)
    sumL3  +=       0.271 * cos(2.45663156460  + 415.55249061210 *T)
    sumL3  +=       0.251 * cos(3.12046701975  + 1382.88734684660 *T)
    sumL3  +=       0.254 * cos(3.00353256829  + 618.55664531160 *T)
    sumL3  +=       0.295 * cos(0.35280179538  + 2730.20695868920 *T)
    sumL3  +=       0.242 * cos(1.52154324392  + 70.84944530420 *T)
    sumL3  +=       0.296 * cos(0.89576757167  + 2104.53676637680 *T)
    sumL3  +=       0.264 * cos(3.00987438634  + 661.23792731640 *T)
    sumL3  +=       0.267 * cos(0.31623829657  + 1677.93857550080 *T)
    sumL3  +=       0.270 * cos(2.56774718753  + 643.07868005170 *T)
    sumL3  +=       0.261 * cos(1.55058302472  + 1457.52593306200 *T)
    sumL3  +=       0.246 * cos(2.29214585472  + 867.42347575360 *T)
    sumL3  +=       0.269 * cos(3.18157515051  + 750.10360753340 *T)
    sumL3  +=       0.272 * cos(1.12208982319  + 1788.14489672020 *T)
    sumL3  +=       0.256 * cos(0.37673546414  + 1279.79457262800 *T)
    sumL3  +=       0.206 * cos(1.81129778306  + 497.44763618020 *T)
    sumL3  +=       0.251 * cos(0.61933213502  + 2413.81508903260 *T)
    sumL3  +=       0.237 * cos(3.35941544147  + 436.89313161450 *T)
    sumL3  +=       0.247 * cos(0.10102936687  + 99.91138048090 *T)
    sumL3  +=       0.247 * cos(0.93125798111  + 52.69019803950 *T)
    sumL3  +=       0.221 * cos(2.07880035795  + 824.74219374880 *T)
    sumL3  +=       0.197 * cos(6.16682223437  + 1258.45393162560 *T)
    sumL3  +=       0.229 * cos(5.57917534840  + 2943.50605412720 *T)
    sumL3  +=       0.227 * cos(0.43324651601  + 2737.32050569000 *T)
    sumL3  +=       0.203 * cos(4.12623986247  + 337.73251065900 *T)
    sumL3  +=       0.214 * cos(3.57607524509  + 934.94851496820 *T)
    sumL3  +=       0.212 * cos(1.25688162158  + 1773.91780271860 *T)
    sumL3  +=       0.215 * cos(0.88867647880  + 1038.04128918680 *T)
    sumL3  +=       0.244 * cos(5.51572084570  + 231.45834270270 *T)
    sumL3  +=       0.181 * cos(2.13821830481  + 416.30325013750 *T)
    sumL3  +=       0.210 * cos(4.19139167658  + 2221.85663459700 *T)
    sumL3  +=       0.178 * cos(2.91685344537  + 74.78159856730 *T)
    sumL3  +=       0.201 * cos(0.46214583002  + 2854.64037391020 *T)
    sumL3  +=       0.236 * cos(4.64388694899  + 1905.46476494040 *T)
    sumL3  +=       0.199 * cos(1.54991619669  + 1471.75302706360 *T)
    sumL3  +=       0.199 * cos(0.70725247497  + 2420.92863603340 *T)
    sumL3  +=       0.162 * cos(2.51488345020  + 430.53034413910 *T)
    sumL3  +=       0.160 * cos(1.23508694599  + 1596.18644228460 *T)
    sumL3  +=       0.175 * cos(4.14605894816  + 2090.30967237520 *T)
    sumL3  +=       0.152 * cos(0.05796022559  + 32.24332891440 *T)
    sumL3  +=       0.176 * cos(1.29002070623  + 490.33408917940 *T)
    sumL3  +=       0.154 * cos(3.60622857548  + 650.94298657790 *T)
    sumL3  +=       0.185 * cos(4.74969742128  + 319.57326339430 *T)
    sumL3  +=       0.154 * cos(1.54587199996  + 1464.63948006280 *T)
    sumL3  +=       0.108 * cos(4.25125786500  + 145.63104387150 *T)
    sumL3  +=       0.106 * cos(1.04047809351  + 1162.47470440780 *T)
    sumL3  +=       0.114 * cos(2.64055737100  + 362.86229257260 *T)
    sumL3  +=       0.093 * cos(3.36746275886  + 483.22054217860 *T)
    sumL3  +=       0.091 * cos(2.05796248692  + 210.85141488320 *T)
    sumL3  +=       0.091 * cos(4.53336314765  + 241.75328344120 *T)
    sumL3  +=       0.072 * cos(3.74361312157  + 1485.98012106520 *T)
    sumL3  +=       0.076 * cos(3.33892447677  + 195.89060769870 *T)

    # SATURN Longitude
    sumL4  = 0
    sumL4  +=    1661.894 * cos(3.99826248978  + 7.11354700080 *T)
    sumL4  +=     257.107 * cos(2.98436499013  + 220.41264243880 *T)
    sumL4  +=     236.344 * cos(3.90241428075  + 14.22709400160 *T)
    sumL4  +=     149.418 * cos(2.74110824208  + 213.29909543800 *T)
    sumL4  +=     109.598 * cos(1.51515739251  + 206.18554843720 *T)
    sumL4  +=     113.953 * cos(3.14159265359  + 0.00000000000 *T)
    sumL4  +=      68.390 * cos(1.72120953337  + 426.59819087600 *T)
    sumL4  +=      37.699 * cos(1.23795458356  + 199.07200143640 *T)
    sumL4  +=      40.060 * cos(2.04644897412  + 433.71173787680 *T)
    sumL4  +=      31.219 * cos(3.01094184090  + 227.52618943960 *T)
    sumL4  +=      15.111 * cos(0.82897064529  + 639.89728631400 *T)
    sumL4  +=       9.444 * cos(3.71485300868  + 21.34064100240 *T)
    sumL4  +=       5.690 * cos(2.41995290633  + 419.48464387520 *T)
    sumL4  +=       4.470 * cos(1.45120818748  + 95.97922721780 *T)
    sumL4  +=       5.608 * cos(1.15607095740  + 647.01083331480 *T)
    sumL4  +=       4.463 * cos(2.11783225176  + 440.82528487760 *T)
    sumL4  +=       3.229 * cos(4.09278077834  + 110.20632121940 *T)
    sumL4  +=       2.871 * cos(2.77203153866  + 412.37109687440 *T)
    sumL4  +=       2.796 * cos(3.00730249564  + 88.86568021700 *T)
    sumL4  +=       2.638 * cos(0.00255721254  + 853.19638175200 *T)
    sumL4  +=       2.574 * cos(0.39246854091  + 103.09277421860 *T)
    sumL4  +=       1.862 * cos(5.07955457727  + 309.27832265580 *T)
    sumL4  +=       2.225 * cos(3.77689198137  + 117.31986822020 *T)
    sumL4  +=       1.769 * cos(5.19176876406  + 302.16477565500 *T)
    sumL4  +=       1.921 * cos(2.82884328662  + 234.63973644040 *T)
    sumL4  +=       1.805 * cos(2.23816036743  + 216.48048917570 *T)
    sumL4  +=       1.211 * cos(1.54685246534  + 191.95845443560 *T)
    sumL4  +=       0.765 * cos(3.44501766503  + 323.50541665740 *T)
    sumL4  +=       0.763 * cos(4.83197222448  + 210.11770170030 *T)
    sumL4  +=       0.613 * cos(4.19052656353  + 515.46387109300 *T)
    sumL4  +=       0.648 * cos(2.28591710303  + 209.36694217490 *T)
    sumL4  +=       0.616 * cos(4.03194472161  + 522.57741809380 *T)
    sumL4  +=       0.630 * cos(2.37952532019  + 632.78373931320 *T)
    sumL4  +=       0.639 * cos(0.29772678242  + 860.30992875280 *T)
    sumL4  +=       0.559 * cos(2.17110060530  + 124.43341522100 *T)
    sumL4  +=       0.442 * cos(2.23500083592  + 447.93883187840 *T)
    sumL4  +=       0.407 * cos(5.44515970990  + 1066.49547719000 *T)
    sumL4  +=       0.469 * cos(1.26889429317  + 654.12438031560 *T)
    sumL4  +=       0.488 * cos(3.20329778617  + 405.25754987360 *T)
    sumL4  +=       0.415 * cos(3.12435410343  + 330.61896365820 *T)
    sumL4  +=       0.442 * cos(3.38933498625  + 81.75213321620 *T)
    sumL4  +=       0.332 * cos(4.12464206608  + 838.96928775040 *T)
    sumL4  +=       0.320 * cos(3.18332026736  + 529.69096509460 *T)
    sumL4  +=       0.312 * cos(1.40962796637  + 429.77958461370 *T)
    sumL4  +=       0.291 * cos(3.18885372262  + 1464.63948006280 *T)
    sumL4  +=       0.333 * cos(2.94355912397  + 728.76296653100 *T)
    sumL4  +=       0.235 * cos(3.67049647573  + 1148.24761040620 *T)
    sumL4  +=       0.286 * cos(2.57895004576  + 1045.15483618760 *T)
    sumL4  +=       0.223 * cos(3.57980034401  + 1155.36115740700 *T)
    sumL4  +=       0.261 * cos(2.04564143519  + 1677.93857550080 *T)
    sumL4  +=       0.218 * cos(2.61967125327  + 536.80451209540 *T)
    sumL4  +=       0.262 * cos(2.48322150677  + 625.67019231240 *T)
    sumL4  +=       0.191 * cos(4.39064160974  + 1574.84580128220 *T)
    sumL4  +=       0.176 * cos(1.26161895188  + 422.66603761290 *T)
    sumL4  +=       0.190 * cos(2.32693171200  + 223.59403617650 *T)
    sumL4  +=       0.185 * cos(1.08713469614  + 742.99006053260 *T)
    sumL4  +=       0.168 * cos(0.69946458053  + 824.74219374880 *T)
    sumL4  +=       0.177 * cos(5.02663339078  + 203.00415469950 *T)
    sumL4  +=       0.218 * cos(0.40426546037  + 867.42347575360 *T)
    sumL4  +=       0.178 * cos(3.67593243311  + 831.85574074960 *T)
    sumL4  +=       0.175 * cos(5.75326979098  + 1073.60902419080 *T)
    sumL4  +=       0.156 * cos(3.02120117572  + 1781.03134971940 *T)
    sumL4  +=       0.148 * cos(2.28313808274  + 295.05122865420 *T)
    sumL4  +=       0.150 * cos(3.48436135302  + 956.28915597060 *T)
    sumL4  +=       0.152 * cos(1.91404443241  + 942.06206196900 *T)
    sumL4  +=       0.146 * cos(6.16519696640  + 316.39186965660 *T)
    sumL4  +=       0.096 * cos(2.93247663741  + 224.34479570190 *T)
    sumL4  +=       0.088 * cos(4.48383632427  + 423.41679713830 *T)

    # SATURN Longitude
    sumL5  = 0
    sumL5  +=     123.615 * cos(2.25923345732  + 7.11354700080 *T)
    sumL5  +=      34.190 * cos(2.16250652689  + 14.22709400160 *T)
    sumL5  +=      27.546 * cos(1.19868150215  + 220.41264243880 *T)
    sumL5  +=       5.818 * cos(1.21584270184  + 227.52618943960 *T)
    sumL5  +=       5.318 * cos(0.23550400093  + 433.71173787680 *T)
    sumL5  +=       3.677 * cos(6.22669694355  + 426.59819087600 *T)
    sumL5  +=       3.057 * cos(2.97372046322  + 199.07200143640 *T)
    sumL5  +=       2.861 * cos(4.28710932685  + 206.18554843720 *T)
    sumL5  +=       1.617 * cos(6.25265362286  + 213.29909543800 *T)
    sumL5  +=       1.279 * cos(5.27612561266  + 639.89728631400 *T)
    sumL5  +=       0.932 * cos(5.56741549127  + 647.01083331480 *T)
    sumL5  +=       0.756 * cos(6.17716234487  + 191.95845443560 *T)
    sumL5  +=       0.760 * cos(0.69475544472  + 302.16477565500 *T)
    sumL5  +=       1.038 * cos(0.23516951637  + 440.82528487760 *T)
    sumL5  +=       1.007 * cos(3.14159265359  + 0.00000000000 *T)
    sumL5  +=       0.549 * cos(4.87733288264  + 88.86568021700 *T)
    sumL5  +=       0.504 * cos(4.77955496203  + 419.48464387520 *T)
    sumL5  +=       0.346 * cos(4.31847547394  + 853.19638175200 *T)
    sumL5  +=       0.392 * cos(5.69922389094  + 654.12438031560 *T)
    sumL5  +=       0.242 * cos(2.05052677361  + 323.50541665740 *T)
    sumL5  +=       0.266 * cos(1.11384528244  + 234.63973644040 *T)
    sumL5  +=       0.199 * cos(0.88505901097  + 309.27832265580 *T)
    sumL5  +=       0.258 * cos(5.10074489186  + 95.97922721780 *T)
    sumL5  +=       0.166 * cos(2.40063312194  + 515.46387109300 *T)
    sumL5  +=       0.155 * cos(4.70433216164  + 860.30992875280 *T)
    sumL5  +=       0.089 * cos(1.36371070380  + 412.37109687440 *T)
    sumL5  +=       0.102 * cos(0.49450039082  + 117.31986822020 *T)

    # SATURN Latitude
    sumB0  = 0
    sumB0  +=  4330678.040 * cos(3.60284428399  + 213.29909543800 *T)
    sumB0  +=  240348.303 * cos(2.85238489390  + 426.59819087600 *T)
    sumB0  +=   84745.939 * cos(0.00000000000  + 0.00000000000 *T)
    sumB0  +=   30863.357 * cos(3.48441504465  + 220.41264243880 *T)
    sumB0  +=   34116.063 * cos(0.57297307844  + 206.18554843720 *T)
    sumB0  +=   14734.070 * cos(2.11846597870  + 639.89728631400 *T)
    sumB0  +=    9916.668 * cos(5.79003189405  + 419.48464387520 *T)
    sumB0  +=    6993.564 * cos(4.73604689179  + 7.11354700080 *T)
    sumB0  +=    4807.587 * cos(5.43305315602  + 316.39186965660 *T)
    sumB0  +=    4788.392 * cos(4.96512927420  + 110.20632121940 *T)
    sumB0  +=    3432.125 * cos(2.73255752123  + 433.71173787680 *T)
    sumB0  +=    1506.129 * cos(6.01304536144  + 103.09277421860 *T)
    sumB0  +=    1060.298 * cos(5.63099292414  + 529.69096509460 *T)
    sumB0  +=     969.071 * cos(5.20434966103  + 632.78373931320 *T)
    sumB0  +=     942.050 * cos(1.39646678088  + 853.19638175200 *T)
    sumB0  +=     707.645 * cos(3.80302329547  + 323.50541665740 *T)
    sumB0  +=     552.313 * cos(5.13149109045  + 202.25339517410 *T)
    sumB0  +=     399.675 * cos(3.35891413961  + 227.52618943960 *T)
    sumB0  +=     316.063 * cos(1.99716764199  + 647.01083331480 *T)
    sumB0  +=     319.380 * cos(3.62571550980  + 209.36694217490 *T)
    sumB0  +=     284.494 * cos(4.88648481625  + 224.34479570190 *T)
    sumB0  +=     314.225 * cos(0.46510272410  + 217.23124870110 *T)
    sumB0  +=     236.442 * cos(2.13887472281  + 11.04570026390 *T)
    sumB0  +=     215.354 * cos(5.94982610103  + 846.08283475120 *T)
    sumB0  +=     208.522 * cos(2.12003893769  + 415.55249061210 *T)
    sumB0  +=     178.958 * cos(2.95361514672  + 63.73589830340 *T)
    sumB0  +=     207.213 * cos(0.73021462851  + 199.07200143640 *T)
    sumB0  +=     139.140 * cos(1.99821990940  + 735.87651353180 *T)
    sumB0  +=     134.884 * cos(5.24500819605  + 742.99006053260 *T)
    sumB0  +=     140.585 * cos(0.64417620299  + 490.33408917940 *T)
    sumB0  +=     121.669 * cos(3.11537140876  + 522.57741809380 *T)
    sumB0  +=     139.240 * cos(4.59535168021  + 14.22709400160 *T)
    sumB0  +=     115.524 * cos(3.10891547171  + 216.48048917570 *T)
    sumB0  +=     114.218 * cos(0.96261442133  + 210.11770170030 *T)
    sumB0  +=      96.376 * cos(4.48164339766  + 117.31986822020 *T)
    sumB0  +=      80.593 * cos(1.31692750150  + 277.03499374140 *T)
    sumB0  +=      72.952 * cos(3.05988482370  + 536.80451209540 *T)
    sumB0  +=      69.261 * cos(4.92378633635  + 309.27832265580 *T)
    sumB0  +=      74.302 * cos(2.89376539620  + 149.56319713460 *T)
    sumB0  +=      68.040 * cos(2.18002263974  + 351.81659230870 *T)
    sumB0  +=      61.734 * cos(0.67728106562  + 1066.49547719000 *T)
    sumB0  +=      56.598 * cos(2.60963391288  + 440.82528487760 *T)
    sumB0  +=      48.864 * cos(5.78725874107  + 95.97922721780 *T)
    sumB0  +=      48.243 * cos(2.18211837430  + 74.78159856730 *T)
    sumB0  +=      38.304 * cos(5.29151303843  + 1059.38193018920 *T)
    sumB0  +=      36.323 * cos(1.63348365121  + 628.85158605010 *T)
    sumB0  +=      35.055 * cos(1.71279210041  + 1052.26838318840 *T)
    sumB0  +=      34.270 * cos(2.45740470599  + 422.66603761290 *T)
    sumB0  +=      34.313 * cos(5.97994514798  + 412.37109687440 *T)
    sumB0  +=      33.787 * cos(1.14073392951  + 949.17560896980 *T)
    sumB0  +=      31.633 * cos(4.14722153007  + 437.64389113990 *T)
    sumB0  +=      36.833 * cos(6.27769966148  + 1162.47470440780 *T)
    sumB0  +=      26.980 * cos(1.27154816810  + 860.30992875280 *T)
    sumB0  +=      23.516 * cos(2.74936525342  + 838.96928775040 *T)
    sumB0  +=      23.460 * cos(0.98962849901  + 210.85141488320 *T)
    sumB0  +=      23.600 * cos(4.11386961467  + 3.93215326310 *T)
    sumB0  +=      23.631 * cos(3.07427204313  + 215.74677599280 *T)
    sumB0  +=      20.813 * cos(3.51084686918  + 330.61896365820 *T)
    sumB0  +=      19.509 * cos(2.81857577372  + 127.47179660680 *T)
    sumB0  +=      17.103 * cos(3.89784279922  + 214.26230328450 *T)
    sumB0  +=      17.635 * cos(6.19715516746  + 703.63318461740 *T)
    sumB0  +=      17.824 * cos(2.28524493886  + 388.46515523820 *T)
    sumB0  +=      20.935 * cos(0.14356167048  + 430.53034413910 *T)
    sumB0  +=      16.551 * cos(1.66649120724  + 38.13303563780 *T)
    sumB0  +=      19.100 * cos(2.97699096081  + 137.03302416240 *T)
    sumB0  +=      15.517 * cos(4.54798410406  + 956.28915597060 *T)
    sumB0  +=      17.065 * cos(0.16611115812  + 212.33588759150 *T)
    sumB0  +=      14.169 * cos(0.48937283445  + 213.34727954780 *T)
    sumB0  +=      19.027 * cos(6.27326062836  + 423.41679713830 *T)
    sumB0  +=      13.344 * cos(2.37136126257  + 429.77958461370 *T)
    sumB0  +=      12.565 * cos(1.03178071173  + 563.63121503840 *T)
    sumB0  +=      14.173 * cos(3.57477564831  + 213.25091132820 *T)
    sumB0  +=      11.374 * cos(1.45300927024  + 1368.66025284500 *T)
    sumB0  +=      10.585 * cos(6.17633425930  + 200.76892246580 *T)
    sumB0  +=      10.600 * cos(3.84358958373  + 138.51749687070 *T)
    sumB0  +=      10.263 * cos(2.17423692422  + 76.26607127560 *T)
    sumB0  +=      10.072 * cos(1.33197220789  + 565.11568774670 *T)
    sumB0  +=      12.058 * cos(0.44149242700  + 222.86032299360 *T)
    sumB0  +=      10.367 * cos(1.85278552549  + 350.33211960040 *T)
    sumB0  +=       8.706 * cos(2.58144528603  + 1155.36115740700 *T)
    sumB0  +=       8.470 * cos(1.97890349826  + 625.67019231240 *T)
    sumB0  +=       8.518 * cos(4.51649648578  + 3.18139373770 *T)
    sumB0  +=       7.439 * cos(4.92597321442  + 212.77783057620 *T)
    sumB0  +=       7.409 * cos(2.03506679104  + 288.08069400530 *T)
    sumB0  +=       8.137 * cos(3.98500592467  + 85.82729883120 *T)
    sumB0  +=       7.985 * cos(2.20794292064  + 362.86229257260 *T)
    sumB0  +=       6.610 * cos(6.14944028835  + 417.03696332040 *T)
    sumB0  +=       7.753 * cos(6.23664549070  + 1478.86657406440 *T)
    sumB0  +=       6.318 * cos(1.87388481013  + 654.12438031560 *T)
    sumB0  +=       6.319 * cos(1.17328438271  + 1265.56747862640 *T)
    sumB0  +=       5.841 * cos(2.35829915285  + 750.10360753340 *T)
    sumB0  +=       5.808 * cos(5.01602242794  + 479.28838891550 *T)
    sumB0  +=       8.079 * cos(0.42574715104  + 554.06998748280 *T)
    sumB0  +=       6.014 * cos(5.58952234348  + 425.11371816770 *T)
    sumB0  +=       7.444 * cos(5.41859596459  + 213.82036029980 *T)
    sumB0  +=       7.567 * cos(2.68446523795  + 191.20769491020 *T)
    sumB0  +=       7.421 * cos(4.19354269508  + 9.56122755560 *T)
    sumB0  +=       5.466 * cos(3.21737829505  + 234.63973644040 *T)
    sumB0  +=       5.661 * cos(1.46982713550  + 265.98929347750 *T)
    sumB0  +=       5.851 * cos(4.81776629912  + 1.48447270830 *T)
    sumB0  +=       5.341 * cos(3.45755372717  + 203.73786788240 *T)
    sumB0  +=       4.960 * cos(1.04628615559  + 12.53017297220 *T)
    sumB0  +=       4.920 * cos(3.85622235967  + 173.94221952280 *T)
    sumB0  +=       4.883 * cos(1.94823282939  + 195.13984817330 *T)
    sumB0  +=       5.621 * cos(0.81869581274  + 52.69019803950 *T)
    sumB0  +=       5.200 * cos(3.32827437636  + 515.46387109300 *T)
    sumB0  +=       4.927 * cos(3.81806549732  + 225.82926841020 *T)
    sumB0  +=       5.033 * cos(0.10756875163  + 252.65597135320 *T)
    sumB0  +=       4.416 * cos(5.45506938037  + 408.43894361130 *T)
    sumB0  +=       4.169 * cos(1.21145214135  + 1685.05212250160 *T)
    sumB0  +=       4.066 * cos(6.24213578122  + 1279.79457262800 *T)
    sumB0  +=       3.972 * cos(6.13850317719  + 217.49188113200 *T)
    sumB0  +=       5.398 * cos(5.67212179194  + 1375.77379984580 *T)
    sumB0  +=       3.916 * cos(5.96105725915  + 210.37833413120 *T)
    sumB0  +=       4.017 * cos(0.99840226682  + 842.15068148810 *T)
    sumB0  +=       3.899 * cos(4.58983662507  + 1272.68102562720 *T)
    sumB0  +=       3.764 * cos(3.30663337976  + 212.54833591260 *T)
    sumB0  +=       4.345 * cos(3.18562241830  + 414.06801790380 *T)
    sumB0  +=       3.565 * cos(4.75262007127  + 207.88246946660 *T)
    sumB0  +=       3.542 * cos(2.30814954338  + 1471.75302706360 *T)
    sumB0  +=       3.732 * cos(1.61040235688  + 635.96513305090 *T)
    sumB0  +=       3.709 * cos(2.97082943086  + 223.59403617650 *T)
    sumB0  +=       3.576 * cos(3.83436862558  + 483.22054217860 *T)
    sumB0  +=       4.053 * cos(3.72105017218  + 942.06206196900 *T)
    sumB0  +=       3.756 * cos(0.74987556308  + 214.04985496340 *T)
    sumB0  +=       3.162 * cos(3.64550741000  + 207.67002114550 *T)
    sumB0  +=       3.149 * cos(2.27647454229  + 728.76296653100 *T)
    sumB0  +=       3.971 * cos(4.37874143597  + 216.21985674480 *T)
    sumB0  +=       3.541 * cos(5.62281936827  + 218.71572140940 *T)
    sumB0  +=       2.965 * cos(3.40117024932  + 650.94298657790 *T)
    sumB0  +=       3.949 * cos(4.19728912450  + 209.10630974400 *T)
    sumB0  +=       2.853 * cos(4.81077453523  + 231.45834270270 *T)
    sumB0  +=       2.826 * cos(0.86682282341  + 217.96496188400 *T)
    sumB0  +=       2.970 * cos(5.75162134301  + 160.60889739850 *T)
    sumB0  +=       2.724 * cos(0.47941267764  + 497.44763618020 *T)
    sumB0  +=       2.787 * cos(4.03144791896  + 62.25142559510 *T)
    sumB0  +=       2.605 * cos(5.04152791794  + 65.22037101170 *T)
    sumB0  +=       2.652 * cos(0.30602610654  + 424.15051032120 *T)
    sumB0  +=       2.543 * cos(2.76499056123  + 543.91805909620 *T)
    sumB0  +=       2.485 * cos(5.78396049817  + 99.16062095550 *T)
    sumB0  +=       3.209 * cos(0.42853440759  + 218.92816973050 *T)
    sumB0  +=       2.509 * cos(2.94704589100  + 70.84944530420 *T)
    sumB0  +=       2.308 * cos(0.63861650866  + 251.43213107580 *T)
    sumB0  +=       2.869 * cos(4.31959346745  + 767.36908292080 *T)
    sumB0  +=       2.232 * cos(0.56929937568  + 1073.60902419080 *T)
    sumB0  +=       2.245 * cos(1.69945964547  + 488.84961647110 *T)
    sumB0  +=       2.283 * cos(1.55589787463  + 601.76425067620 *T)
    sumB0  +=       2.384 * cos(4.44583493310  + 21.34064100240 *T)
    sumB0  +=       2.096 * cos(5.77425542767  + 88.86568021700 *T)
    sumB0  +=       2.363 * cos(3.35786310868  + 124.43341522100 *T)
    sumB0  +=       2.162 * cos(6.24029269257  + 1795.25844372100 *T)
    sumB0  +=       2.452 * cos(3.19804814047  + 208.63322899200 *T)
    sumB0  +=       2.033 * cos(4.87029603776  + 327.43756992050 *T)
    sumB0  +=       1.950 * cos(5.56004000293  + 18.15924726470 *T)
    sumB0  +=       2.283 * cos(4.22375355881  + 22.09140052780 *T)
    sumB0  +=       2.217 * cos(4.61433094630  + 302.16477565500 *T)
    sumB0  +=       1.888 * cos(5.45600788064  + 142.44965013380 *T)
    sumB0  +=       2.075 * cos(3.55622165076  + 1169.58825140860 *T)
    sumB0  +=       2.069 * cos(2.75786819366  + 491.81856188770 *T)
    sumB0  +=       1.810 * cos(5.96568495526  + 213.18722085340 *T)
    sumB0  +=       1.813 * cos(1.39785500313  + 211.81462272970 *T)
    sumB0  +=       1.843 * cos(1.15001484281  + 203.00415469950 *T)
    sumB0  +=       1.854 * cos(1.41350087782  + 1581.95934828300 *T)
    sumB0  +=       1.697 * cos(3.23613719814  + 427.56139872250 *T)
    sumB0  +=       1.736 * cos(5.45933992115  + 916.93228005540 *T)
    sumB0  +=       1.714 * cos(6.14729146384  + 643.82943957710 *T)
    sumB0  +=       1.948 * cos(5.70817363392  + 425.63498302950 *T)
    sumB0  +=       1.770 * cos(3.36194411768  + 248.72381809010 *T)
    sumB0  +=       1.611 * cos(0.97888081762  + 2001.44399215820 *T)
    sumB0  +=       1.971 * cos(2.59654430358  + 429.04587143080 *T)
    sumB0  +=       1.628 * cos(0.74011617198  + 177.87437278590 *T)
    sumB0  +=       1.564 * cos(2.04342011485  + 1788.14489672020 *T)
    sumB0  +=       1.574 * cos(6.01995224314  + 426.64637498580 *T)
    sumB0  +=       1.939 * cos(5.48134669380  + 636.71589257630 *T)
    sumB0  +=       1.651 * cos(4.61629429952  + 621.73803904930 *T)
    sumB0  +=       1.552 * cos(2.55542734908  + 692.58748435350 *T)
    sumB0  +=       1.484 * cos(6.17173980637  + 56.62235130260 *T)
    sumB0  +=       1.782 * cos(3.26122906302  + 175.16605980020 *T)
    sumB0  +=       1.503 * cos(2.59953333916  + 228.27694896500 *T)
    sumB0  +=       1.559 * cos(0.36281050773  + 776.93031047640 *T)
    sumB0  +=       1.799 * cos(3.46395976970  + 1258.45393162560 *T)
    sumB0  +=       1.521 * cos(3.75588462293  + 213.51154375910 *T)
    sumB0  +=       1.810 * cos(4.37552745264  + 213.41097002260 *T)
    sumB0  +=       1.521 * cos(0.30214462385  + 213.08664711690 *T)
    sumB0  +=       1.608 * cos(4.66724132818  + 269.92144674060 *T)
    sumB0  +=       1.525 * cos(1.47939329423  + 198.32124191100 *T)
    sumB0  +=       1.408 * cos(1.38491846750  + 501.37978944330 *T)
    sumB0  +=       1.327 * cos(0.23760037979  + 148.07872442630 *T)
    sumB0  +=       1.305 * cos(2.41201772023  + 275.55052103310 *T)
    sumB0  +=       1.578 * cos(2.82443242444  + 426.55000676620 *T)
    sumB0  +=       1.203 * cos(1.36928065935  + 235.39049596580 *T)
    sumB0  +=       1.296 * cos(5.75203277385  + 1692.16566950240 *T)
    sumB0  +=       1.295 * cos(3.04090959062  + 831.85574074960 *T)
    sumB0  +=       1.283 * cos(1.62400181159  + 643.07868005170 *T)
    sumB0  +=       1.617 * cos(1.31181209458  + 214.78356814630 *T)
    sumB0  +=       1.346 * cos(4.01262069353  + 278.51946644970 *T)
    sumB0  +=       1.166 * cos(0.09590019561  + 340.77089204480 *T)
    sumB0  +=       1.115 * cos(2.20460481017  + 221.37585028530 *T)
    sumB0  +=       1.176 * cos(1.07528227869  + 312.19908396260 *T)
    sumB0  +=       1.107 * cos(1.50329021421  + 289.56516671360 *T)
    sumB0  +=       1.143 * cos(4.40125874383  + 213.55972786890 *T)
    sumB0  +=       1.253 * cos(0.21632769953  + 404.50679034820 *T)
    sumB0  +=       1.074 * cos(3.17412275025  + 98.89998852460 *T)
    sumB0  +=       1.103 * cos(0.23626839162  + 617.80588578620 *T)
    sumB0  +=       1.077 * cos(3.51670933532  + 312.45971639350 *T)
    sumB0  +=       1.039 * cos(0.53953974796  + 778.41478318470 *T)
    sumB0  +=       1.195 * cos(2.11232088496  + 205.22234059070 *T)
    sumB0  +=       1.093 * cos(5.16243153571  + 630.33605875840 *T)
    sumB0  +=       1.143 * cos(5.93977485365  + 213.03846300710 *T)
    sumB0  +=       1.040 * cos(4.79631324365  + 106.27416795630 *T)
    sumB0  +=       1.100 * cos(0.02509241739  + 219.44943459230 *T)
    sumB0  +=       1.311 * cos(5.93900415785  + 436.15941843160 *T)
    sumB0  +=       1.260 * cos(0.72481995446  + 355.74874557180 *T)
    sumB0  +=       0.950 * cos(2.00801292252  + 1045.15483618760 *T)
    sumB0  +=       1.186 * cos(1.84906064486  + 151.04766984290 *T)
    sumB0  +=       0.974 * cos(3.01092368346  + 696.51963761660 *T)
    sumB0  +=       0.924 * cos(4.88158186437  + 39.35687591520 *T)
    sumB0  +=       0.961 * cos(2.80113869315  + 738.79727483860 *T)
    sumB0  +=       1.011 * cos(6.27004359435  + 121.25202148330 *T)
    sumB0  +=       0.904 * cos(4.15218356485  + 426.07692601420 *T)
    sumB0  +=       0.904 * cos(4.24232505252  + 10.29494073850 *T)
    sumB0  +=       0.895 * cos(2.47587956888  + 447.93883187840 *T)
    sumB0  +=       1.186 * cos(0.85270715988  + 525.49817940060 *T)
    sumB0  +=       0.824 * cos(3.97540449663  + 210.59078245230 *T)
    sumB0  +=       0.831 * cos(4.05299295705  + 207.14875628370 *T)
    sumB0  +=       0.937 * cos(5.44353432307  + 344.70304530790 *T)
    sumB0  +=       0.823 * cos(2.08766677969  + 358.93013930950 *T)
    sumB0  +=       0.971 * cos(5.09512804595  + 1589.07289528380 *T)
    sumB0  +=       1.037 * cos(1.04152859909  + 2.44768055480 *T)
    sumB0  +=       0.816 * cos(0.62175655307  + 188.92007304980 *T)
    sumB0  +=       0.798 * cos(3.36396062989  + 237.67811782620 *T)
    sumB0  +=       0.755 * cos(5.90983584858  + 284.14854074220 *T)
    sumB0  +=       0.870 * cos(1.62765846893  + 114.13847448250 *T)
    sumB0  +=       0.734 * cos(6.23523714922  + 2111.65031337760 *T)
    sumB0  +=       0.767 * cos(2.73651219269  + 627.36711334180 *T)
    sumB0  +=       0.701 * cos(0.72526525525  + 10213.28554621100 *T)
    sumB0  +=       0.801 * cos(5.84130533519  + 905.88657979150 *T)
    sumB0  +=       0.710 * cos(1.78740818763  + 2104.53676637680 *T)
    sumB0  +=       0.670 * cos(0.75839374890  + 2317.83586181480 *T)
    sumB0  +=       0.720 * cos(4.95645178898  + 638.41281360570 *T)
    sumB0  +=       0.796 * cos(4.87623914638  + 342.25536475310 *T)
    sumB0  +=       0.703 * cos(0.35096991676  + 220.46082654860 *T)
    sumB0  +=       0.835 * cos(3.19492825640  + 1574.84580128220 *T)
    sumB0  +=       0.824 * cos(0.08813665925  + 216.00740842370 *T)
    sumB0  +=       0.653 * cos(4.19599635390  + 247.23934538180 *T)
    sumB0  +=       0.826 * cos(4.66845240923  + 427.11945573780 *T)
    sumB0  +=       0.690 * cos(0.41873449575  + 5856.47765911540 *T)
    sumB0  +=       0.690 * cos(1.34023204821  + 6283.07584999140 *T)
    sumB0  +=       0.690 * cos(0.58291546593  + 213.45915413240 *T)
    sumB0  +=       0.639 * cos(1.20304559619  + 867.42347575360 *T)
    sumB0  +=       0.830 * cos(1.57233214789  + 1898.35121793960 *T)
    sumB0  +=       0.753 * cos(1.51187970880  + 576.16138801060 *T)
    sumB0  +=       0.629 * cos(2.83598833891  + 420.96911658350 *T)
    sumB0  +=       0.690 * cos(3.48062808501  + 213.13903674360 *T)
    sumB0  +=       0.600 * cos(5.22938546212  + 212.02707105080 *T)
    sumB0  +=       0.565 * cos(5.28099337758  + 423.67742956920 *T)
    sumB0  +=       0.552 * cos(5.83265103738  + 84.34282612290 *T)
    sumB0  +=       0.546 * cos(3.56588711151  + 1485.98012106520 *T)
    sumB0  +=       0.642 * cos(6.15007145598  + 179.35884549420 *T)
    sumB0  +=       0.697 * cos(1.91977327925  + 134.58534360760 *T)
    sumB0  +=       0.648 * cos(5.15917450752  + 8.07675484730 *T)
    sumB0  +=       0.551 * cos(3.33109164145  + 980.66817835880 *T)
    sumB0  +=       0.511 * cos(3.87073988213  + 125.98732389850 *T)
    sumB0  +=       0.506 * cos(0.80261216855  + 181.05576652360 *T)
    sumB0  +=       0.564 * cos(4.05871107615  + 831.10498122420 *T)
    sumB0  +=       0.681 * cos(3.45804290093  + 220.36445832900 *T)
    sumB0  +=       0.541 * cos(2.39252901431  + 421.93232443000 *T)
    sumB0  +=       0.498 * cos(3.11515053568  + 439.12836384820 *T)
    sumB0  +=       0.517 * cos(2.75430004800  + 1148.24761040620 *T)
    sumB0  +=       0.491 * cos(4.47199498503  + 558.00214074590 *T)
    sumB0  +=       0.467 * cos(4.18144797677  + 444.75743814070 *T)
    sumB0  +=       0.458 * cos(0.60848440253  + 206.23373254700 *T)
    sumB0  +=       0.542 * cos(3.28613020157  + 245.54242435240 *T)
    sumB0  +=       0.477 * cos(5.48556902348  + 35.42472265210 *T)
    sumB0  +=       0.474 * cos(2.42545073874  + 436.89313161450 *T)
    sumB0  +=       0.481 * cos(3.72498040536  + 206.13736432740 *T)
    sumB0  +=       0.435 * cos(1.61732308614  + 191.95845443560 *T)
    sumB0  +=       0.488 * cos(0.26059969650  + 416.30325013750 *T)
    sumB0  +=       0.493 * cos(4.13699180247  + 518.64526483070 *T)
    sumB0  +=       0.595 * cos(5.11706007934  + 214.57111982520 *T)
    sumB0  +=       0.486 * cos(5.17710069856  + 67.66805156650 *T)
    sumB0  +=       0.463 * cos(5.53192185211  + 418.52143602870 *T)
    sumB0  +=       0.421 * cos(5.57377685121  + 430.79097657000 *T)
    sumB0  +=       0.446 * cos(6.20735049659  + 73.29712585900 *T)
    sumB0  +=       0.421 * cos(4.65837340438  + 543.02428721890 *T)
    sumB0  +=       0.435 * cos(2.95256554350  + 5.41662597140 *T)
    sumB0  +=       0.416 * cos(4.36391909218  + 113.38771495710 *T)
    sumB0  +=       0.495 * cos(5.38121485133  + 391.17346822390 *T)
    sumB0  +=       0.520 * cos(3.99939347071  + 618.55664531160 *T)
    sumB0  +=       0.429 * cos(3.26903461513  + 144.14657116320 *T)
    sumB0  +=       0.435 * cos(1.60816661416  + 2214.74308759620 *T)
    sumB0  +=       0.398 * cos(2.38329818919  + 299.12639426920 *T)
    sumB0  +=       0.399 * cos(0.06301179341  + 206.70681329900 *T)
    sumB0  +=       0.394 * cos(2.75496219113  + 425.84743135060 *T)
    sumB0  +=       0.508 * cos(2.86873328929  + 337.73251065900 *T)
    sumB0  +=       0.475 * cos(5.68530292289  + 320.32402291970 *T)
    sumB0  +=       0.432 * cos(0.68132398291  + 116.42609634290 *T)
    sumB0  +=       0.404 * cos(1.84249441289  + 9786.68735533500 *T)
    sumB0  +=       0.371 * cos(5.83382962844  + 2008.55753915900 *T)
    sumB0  +=       0.382 * cos(4.24995364794  + 387.24131496080 *T)
    sumB0  +=       0.416 * cos(3.81986946256  + 429.51895218280 *T)
    sumB0  +=       0.491 * cos(4.18623117082  + 219.89137757700 *T)
    sumB0  +=       0.352 * cos(1.65610545221  + 963.40270297140 *T)
    sumB0  +=       0.353 * cos(5.50209003460  + 305.34616939270 *T)
    sumB0  +=       0.431 * cos(4.39380503963  + 353.30106501700 *T)
    sumB0  +=       0.375 * cos(2.67828133567  + 319.57326339430 *T)
    sumB0  +=       0.359 * cos(3.54801032661  + 421.18156490460 *T)
    sumB0  +=       0.339 * cos(5.19074405462  + 69.15252427480 *T)
    sumB0  +=       0.358 * cos(1.11857595997  + 1044.40407666220 *T)
    sumB0  +=       0.334 * cos(1.84260994740  + 1361.54670584420 *T)
    sumB0  +=       0.328 * cos(6.07106596408  + 710.74673161820 *T)
    sumB0  +=       0.328 * cos(1.48618893585  + 2420.92863603340 *T)
    sumB0  +=       0.359 * cos(5.62797136991  + 78.71375183040 *T)
    sumB0  +=       0.405 * cos(2.91366762549  + 1891.23767093880 *T)
    sumB0  +=       0.398 * cos(2.08900381937  + 4.66586644600 *T)
    sumB0  +=       0.326 * cos(1.62373774313  + 5.62907429250 *T)
    sumB0  +=       0.405 * cos(1.87470341223  + 114.39910691340 *T)
    sumB0  +=       0.393 * cos(0.56487847337  + 128.95626931510 *T)
    sumB0  +=       0.312 * cos(5.29291124448  + 347.88443904560 *T)
    sumB0  +=       0.310 * cos(0.31232452686  + 427.34895040140 *T)
    sumB0  +=       0.308 * cos(2.84912656825  + 487.36514376280 *T)
    sumB0  +=       0.291 * cos(3.25977762780  + 494.26624244250 *T)
    sumB0  +=       0.303 * cos(4.93962873690  + 373.90799283650 *T)
    sumB0  +=       0.289 * cos(2.83591185309  + 212.07525516060 *T)
    sumB0  +=       0.289 * cos(2.05371431350  + 214.52293571540 *T)
    sumB0  +=       0.398 * cos(6.03822674845  + 432.22726516850 *T)
    sumB0  +=       0.288 * cos(6.16001418475  + 969.62247809490 *T)
    sumB0  +=       0.296 * cos(0.30332524090  + 1055.44977692610 *T)
    sumB0  +=       0.280 * cos(1.29728455136  + 241.61027108930 *T)
    sumB0  +=       0.280 * cos(5.41630221077  + 1493.09366806600 *T)
    sumB0  +=       0.315 * cos(6.24003908326  + 465.95506679120 *T)
    sumB0  +=       0.274 * cos(5.03981944861  + 458.84151979040 *T)
    sumB0  +=       0.296 * cos(3.04457317761  + 211.60217440860 *T)
    sumB0  +=       0.274 * cos(2.72607352851  + 145.63104387150 *T)
    sumB0  +=       0.293 * cos(1.32452002382  + 159.12442469020 *T)
    sumB0  +=       0.263 * cos(6.11559198968  + 2428.04218303420 *T)
    sumB0  +=       0.323 * cos(1.17502395659  + 815.06334611420 *T)
    sumB0  +=       0.345 * cos(5.37083374878  + 428.08266358430 *T)
    sumB0  +=       0.258 * cos(0.43205106363  + 2634.22773147140 *T)
    sumB0  +=       0.275 * cos(0.91628212149  + 849.26422848890 *T)
    sumB0  +=       0.340 * cos(1.29378813067  + 329.72519178090 *T)
    sumB0  +=       0.294 * cos(4.29399534634  + 4.19278569400 *T)
    sumB0  +=       0.339 * cos(1.03883773894  + 32.24332891440 *T)
    sumB0  +=       0.244 * cos(3.52504227332  + 184.98791978670 *T)
    sumB0  +=       0.243 * cos(3.13047989401  + 525.75881183150 *T)
    sumB0  +=       0.260 * cos(5.99216785208  + 20.60692781950 *T)
    sumB0  +=       0.303 * cos(3.96772261614  + 934.94851496820 *T)
    sumB0  +=       0.285 * cos(5.69474711283  + 220.93390730060 *T)
    sumB0  +=       0.239 * cos(2.09516779457  + 292.01284726840 *T)
    sumB0  +=       0.242 * cos(0.98744748894  + 282.45161971280 *T)
    sumB0  +=       0.278 * cos(2.81667003542  + 87.31177153950 *T)
    sumB0  +=       0.285 * cos(4.76303256917  + 54.17467074780 *T)
    sumB0  +=       0.236 * cos(5.79560286324  + 280.96714700450 *T)
    sumB0  +=       0.246 * cos(1.83689902078  + 214.99601646740 *T)
    sumB0  +=       0.221 * cos(0.48302467341  + 153.49535039770 *T)
    sumB0  +=       0.238 * cos(3.52738705554  + 267.47376618580 *T)
    sumB0  +=       0.293 * cos(5.91401607974  + 14.97785352700 *T)
    sumB0  +=       0.235 * cos(0.29884419224  + 14.01464568050 *T)
    sumB0  +=       0.229 * cos(3.91975580405  + 182.27960680100 *T)
    sumB0  +=       0.217 * cos(3.96328561940  + 894.84087952760 *T)
    sumB0  +=       0.218 * cos(1.46057992688  + 2531.13495725280 *T)
    sumB0  +=       0.210 * cos(2.01138855049  + 211.86280683950 *T)
    sumB0  +=       0.210 * cos(2.87823761610  + 214.73538403650 *T)
    sumB0  +=       0.223 * cos(2.49651288358  + 1464.63948006280 *T)
    sumB0  +=       0.217 * cos(4.03234048538  + 835.03713448730 *T)
    sumB0  +=       0.209 * cos(2.97542058241  + 273.10284047830 *T)
    sumB0  +=       0.210 * cos(2.56849303008  + 593.42686339800 *T)
    sumB0  +=       0.218 * cos(1.71859101510  + 0.96320784650 *T)
    sumB0  +=       0.232 * cos(1.86083014117  + 221.16340196420 *T)
    sumB0  +=       0.199 * cos(3.53454927143  + 219.66188291340 *T)
    sumB0  +=       0.197 * cos(2.65338829617  + 864.24208201590 *T)
    sumB0  +=       0.219 * cos(3.46267185338  + 1182.92157353290 *T)
    sumB0  +=       0.199 * cos(2.56046317223  + 264.50482076920 *T)
    sumB0  +=       0.199 * cos(2.03508708900  + 757.21715453420 *T)
    sumB0  +=       0.237 * cos(5.05443109284  + 254.94359321360 *T)
    sumB0  +=       0.191 * cos(2.07106909876  + 756.32338265690 *T)
    sumB0  +=       0.192 * cos(1.67985944172  + 1677.93857550080 *T)
    sumB0  +=       0.191 * cos(1.05067348453  + 702.14871190910 *T)
    sumB0  +=       0.181 * cos(1.89151852263  + 6.15033915430 *T)
    sumB0  +=       0.205 * cos(1.98151360584  + 199.28444975750 *T)
    sumB0  +=       0.181 * cos(1.25381796494  + 2737.32050569000 *T)
    sumB0  +=       0.186 * cos(2.81416016738  + 569.04784100980 *T)
    sumB0  +=       0.215 * cos(4.52373060846  + 3060.82592234740 *T)
    sumB0  +=       0.222 * cos(2.90133577623  + 205.43478891180 *T)
    sumB0  +=       0.246 * cos(3.55891849574  + 1251.34038462480 *T)
    sumB0  +=       0.191 * cos(4.20221553993  + 556.51766803760 *T)
    sumB0  +=       0.217 * cos(2.64509967170  + 2207.62954059540 *T)
    sumB0  +=       0.225 * cos(0.14271906959  + 131.40394986990 *T)
    sumB0  +=       0.189 * cos(1.27260556263  + 192.69216761850 *T)
    sumB0  +=       0.179 * cos(6.15189171649  + 2.92076130680 *T)
    sumB0  +=       0.178 * cos(2.01622328964  + 705.11765732570 *T)
    sumB0  +=       0.181 * cos(3.62483757675  + 233.90602325750 *T)
    sumB0  +=       0.188 * cos(2.92836809840  + 227.31374111850 *T)
    sumB0  +=       0.164 * cos(3.50682537694  + 1382.88734684660 *T)
    sumB0  +=       0.166 * cos(5.85452227121  + 637.44960575920 *T)
    sumB0  +=       0.160 * cos(0.13309484488  + 431.26405732200 *T)
    sumB0  +=       0.158 * cos(5.92242110049  + 96.87299909510 *T)
    sumB0  +=       0.178 * cos(4.55557913565  + 46.47042291600 *T)
    sumB0  +=       0.157 * cos(1.35908014451  + 51.20572533120 *T)
    sumB0  +=       0.155 * cos(6.24092514222  + 464.73122651380 *T)
    sumB0  +=       0.155 * cos(6.02684189458  + 1286.90811962880 *T)
    sumB0  +=       0.155 * cos(1.36669731999  + 206.93630796260 *T)
    sumB0  +=       0.175 * cos(4.95121713507  + 1905.46476494040 *T)
    sumB0  +=       0.153 * cos(6.06094547271  + 561.18353448360 *T)
    sumB0  +=       0.208 * cos(4.50537355579  + 24.37902238820 *T)
    sumB0  +=       0.185 * cos(5.49802440713  + 205.66428357540 *T)
    sumB0  +=       0.160 * cos(4.18196878816  + 3340.61242669980 *T)
    sumB0  +=       0.160 * cos(2.85370771355  + 209.15449385380 *T)
    sumB0  +=       0.161 * cos(4.98020340619  + 2648.45482547300 *T)
    sumB0  +=       0.152 * cos(0.85558875667  + 570.74476203920 *T)
    sumB0  +=       0.156 * cos(2.03601440129  + 217.44369702220 *T)
    sumB0  +=       0.192 * cos(3.98017256784  + 212.40532356070 *T)
    sumB0  +=       0.192 * cos(0.90945359875  + 214.19286731530 *T)
    sumB0  +=       0.198 * cos(3.54289000289  + 533.62311835770 *T)
    sumB0  +=       0.160 * cos(2.51522796187  + 3127.31333126180 *T)
    sumB0  +=       0.145 * cos(0.93414637377  + 1994.33044515740 *T)
    sumB0  +=       0.141 * cos(5.66801998888  + 120.35824960600 *T)
    sumB0  +=       0.141 * cos(3.88995778619  + 454.90936652730 *T)
    sumB0  +=       0.152 * cos(1.01453902153  + 2840.41327990860 *T)
    sumB0  +=       0.172 * cos(4.15145223592  + 2.96894541660 *T)
    sumB0  +=       0.146 * cos(5.26789260159  + 7.06536289100 *T)
    sumB0  +=       0.177 * cos(0.43196690516  + 140.00196957900 *T)
    sumB0  +=       0.144 * cos(3.95680110579  + 300.61086697750 *T)
    sumB0  +=       0.152 * cos(4.29475572258  + 555.55446019110 *T)
    sumB0  +=       0.143 * cos(4.15264164139  + 31.01948863700 *T)
    sumB0  +=       0.138 * cos(4.20096561019  + 731.94436026870 *T)
    sumB0  +=       0.139 * cos(0.79924371385  + 166.82867252200 *T)
    sumB0  +=       0.135 * cos(1.27192121638  + 92.94084583200 *T)
    sumB0  +=       0.165 * cos(1.94881873062  + 107.02492748170 *T)
    sumB0  +=       0.153 * cos(3.07590434707  + 3480.31056622260 *T)
    sumB0  +=       0.125 * cos(3.41796878361  + 1802.37199072180 *T)
    sumB0  +=       0.128 * cos(5.83700968658  + 2324.94940881560 *T)
    sumB0  +=       0.129 * cos(2.75851443754  + 480.77286162380 *T)
    sumB0  +=       0.122 * cos(4.75728514255  + 2854.64037391020 *T)
    sumB0  +=       0.129 * cos(4.67730374872  + 913.96333463880 *T)
    sumB0  +=       0.121 * cos(1.19957548726  + 572.22923474750 *T)
    sumB0  +=       0.146 * cos(3.71232877850  + 546.95644048200 *T)
    sumB0  +=       0.120 * cos(4.59083886034  + 339.28641933650 *T)
    sumB0  +=       0.127 * cos(6.19372294369  + 59.80374504030 *T)
    sumB0  +=       0.123 * cos(5.98484393970  + 477.80391620720 *T)
    sumB0  +=       0.122 * cos(5.82973501131  + 990.22940591440 *T)
    sumB0  +=       0.151 * cos(2.39413061881  + 2524.02141025200 *T)
    sumB0  +=       0.147 * cos(4.98199291770  + 850.01498801430 *T)
    sumB0  +=       0.127 * cos(0.47350572907  + 6.59228213900 *T)
    sumB0  +=       0.116 * cos(4.71488890981  + 1130.23137549340 *T)
    sumB0  +=       0.129 * cos(5.42665311725  + 2538.24850425360 *T)
    sumB0  +=       0.111 * cos(4.51520219898  + 1699.27921650320 *T)
    sumB0  +=       0.114 * cos(2.06120434865  + 952.09637027660 *T)
    sumB0  +=       0.112 * cos(4.14736642579  + 422.40540518200 *T)
    sumB0  +=       0.111 * cos(5.53230411085  + 857.12853501510 *T)
    sumB0  +=       0.109 * cos(3.00005651288  + 420.44785172170 *T)
    sumB0  +=       0.112 * cos(2.20667724213  + 395.57870223900 *T)
    sumB0  +=       0.147 * cos(1.52324472511  + 552.58551477450 *T)
    sumB0  +=       0.118 * cos(5.47495367121  + 2957.73314812880 *T)
    sumB0  +=       0.113 * cos(2.57036693965  + 462.02291352810 *T)
    sumB0  +=       0.122 * cos(4.96246567897  + 638.93407846750 *T)
    sumB0  +=       0.104 * cos(1.91139383428  + 472.17484191470 *T)
    sumB0  +=       0.116 * cos(2.82742160564  + 450.97721326420 *T)
    sumB0  +=       0.115 * cos(2.26043201622  + 1781.03134971940 *T)
    sumB0  +=       0.110 * cos(4.86686492403  + 2914.01423582380 *T)
    sumB0  +=       0.109 * cos(4.43848727280  + 405.99126305650 *T)
    sumB0  +=       0.102 * cos(5.90112611078  + 99.91138048090 *T)
    sumB0  +=       0.101 * cos(2.53392410330  + 640.86049416050 *T)
    sumB0  +=       0.112 * cos(5.53802838779  + 381.35160823740 *T)
    sumB0  +=       0.099 * cos(5.92199927896  + 411.62033734900 *T)
    sumB0  +=       0.100 * cos(5.21941099517  + 426.48631629140 *T)
    sumB0  +=       0.137 * cos(2.20269111622  + 7.16173111060 *T)
    sumB0  +=       0.097 * cos(1.27914551364  + 2847.52682690940 *T)
    sumB0  +=       0.115 * cos(5.22953781515  + 1119.18567522950 *T)
    sumB0  +=       0.095 * cos(4.26135357007  + 540.73666535850 *T)
    sumB0  +=       0.098 * cos(5.27107833435  + 639.94547042380 *T)
    sumB0  +=       0.107 * cos(4.38879925113  + 412.58354519550 *T)
    sumB0  +=       0.093 * cos(5.35954173624  + 334.55111692130 *T)
    sumB0  +=       0.094 * cos(1.16749092536  + 5643.17856367740 *T)
    sumB0  +=       0.106 * cos(4.19443004843  + 486.40193591630 *T)
    sumB0  +=       0.096 * cos(0.59816870672  + 714.67888488130 *T)
    sumB0  +=       0.094 * cos(0.54205024076  + 423.62924545940 *T)
    sumB0  +=       0.109 * cos(2.82817225044  + 468.24268865160 *T)
    sumB0  +=       0.083 * cos(6.12100285205  + 380.12776796000 *T)
    sumB0  +=       0.084 * cos(2.20217125255  + 909.81873305460 *T)
    sumB0  +=       0.085 * cos(5.20920130934  + 562.14674233010 *T)
    sumB0  +=       0.105 * cos(2.66415710279  + 460.53844081980 *T)
    sumB0  +=       0.084 * cos(0.14646013561  + 681.54178408960 *T)
    sumB0  +=       0.080 * cos(3.03551986945  + 409.92341631960 *T)
    sumB0  +=       0.097 * cos(5.09373549436  + 92.04707395470 *T)
    sumB0  +=       0.110 * cos(2.03622569317  + 642.34496686880 *T)
    sumB0  +=       0.080 * cos(5.71035549752  + 361.37781986430 *T)
    sumB0  +=       0.084 * cos(3.00961145133  + 426.81063919710 *T)
    sumB0  +=       0.085 * cos(4.28770375688  + 135.54855145410 *T)
    sumB0  +=       0.093 * cos(5.32943472274  + 432.01481684740 *T)
    sumB0  +=       0.086 * cos(1.51247258028  + 760.25553592000 *T)
    sumB0  +=       0.084 * cos(5.83905303748  + 426.38574255490 *T)
    sumB0  +=       0.100 * cos(3.62925349363  + 426.71006546060 *T)
    sumB0  +=       0.094 * cos(4.30151510535  + 3377.21779200400 *T)
    sumB0  +=       0.098 * cos(2.07334671974  + 639.84910220420 *T)
    sumB0  +=       0.080 * cos(4.11576173565  + 774.48262992160 *T)
    sumB0  +=       0.075 * cos(2.89122656610  + 806.72595883600 *T)
    sumB0  +=       0.080 * cos(0.88468467902  + 856.37777548970 *T)
    sumB0  +=       0.072 * cos(4.85259171933  + 392.65794093220 *T)
    sumB0  +=       0.083 * cos(0.11133738383  + 402.21916848780 *T)

    # SATURN Latitude
    sumB1  = 0
    sumB1  +=  397554.998 * cos(5.33289992556  + 213.29909543800 *T)
    sumB1  +=   49478.641 * cos(3.14159265359  + 0.00000000000 *T)
    sumB1  +=   18571.607 * cos(6.09919206378  + 426.59819087600 *T)
    sumB1  +=   14800.587 * cos(2.30586060520  + 206.18554843720 *T)
    sumB1  +=    9643.981 * cos(1.69674660120  + 220.41264243880 *T)
    sumB1  +=    3757.161 * cos(1.25429514018  + 419.48464387520 *T)
    sumB1  +=    2716.647 * cos(5.91166664787  + 639.89728631400 *T)
    sumB1  +=    1455.309 * cos(0.85161616532  + 433.71173787680 *T)
    sumB1  +=    1290.595 * cos(2.91770857090  + 7.11354700080 *T)
    sumB1  +=     852.630 * cos(0.43572078997  + 316.39186965660 *T)
    sumB1  +=     284.386 * cos(1.61881754773  + 227.52618943960 *T)
    sumB1  +=     292.185 * cos(5.31574251270  + 853.19638175200 *T)
    sumB1  +=     275.090 * cos(3.88864137336  + 103.09277421860 *T)
    sumB1  +=     297.726 * cos(0.91909206723  + 632.78373931320 *T)
    sumB1  +=     172.359 * cos(0.05215146556  + 647.01083331480 *T)
    sumB1  +=     127.731 * cos(1.20711452525  + 529.69096509460 *T)
    sumB1  +=     166.237 * cos(2.44351613165  + 199.07200143640 *T)
    sumB1  +=     158.220 * cos(5.20850125766  + 110.20632121940 *T)
    sumB1  +=     109.839 * cos(2.45695551627  + 217.23124870110 *T)
    sumB1  +=      81.759 * cos(2.75839171353  + 210.11770170030 *T)
    sumB1  +=      81.010 * cos(2.86038377187  + 14.22709400160 *T)
    sumB1  +=      68.658 * cos(1.65537623146  + 202.25339517410 *T)
    sumB1  +=      59.281 * cos(1.82410768234  + 323.50541665740 *T)
    sumB1  +=      65.161 * cos(1.25527521313  + 216.48048917570 *T)
    sumB1  +=      61.024 * cos(1.25273412095  + 209.36694217490 *T)
    sumB1  +=      46.386 * cos(0.81534705304  + 440.82528487760 *T)
    sumB1  +=      36.163 * cos(1.81851057689  + 224.34479570190 *T)
    sumB1  +=      34.041 * cos(2.83971297997  + 117.31986822020 *T)
    sumB1  +=      32.164 * cos(1.18676132343  + 846.08283475120 *T)
    sumB1  +=      33.114 * cos(1.30557080010  + 412.37109687440 *T)
    sumB1  +=      27.282 * cos(4.64744847591  + 1066.49547719000 *T)
    sumB1  +=      22.805 * cos(4.12923703368  + 415.55249061210 *T)
    sumB1  +=      27.128 * cos(4.44228739187  + 11.04570026390 *T)
    sumB1  +=      18.100 * cos(5.56392353608  + 860.30992875280 *T)
    sumB1  +=      20.851 * cos(1.40999273740  + 309.27832265580 *T)
    sumB1  +=      14.947 * cos(1.34302610607  + 95.97922721780 *T)
    sumB1  +=      15.316 * cos(1.22393617996  + 63.73589830340 *T)
    sumB1  +=      14.601 * cos(1.00753704970  + 536.80451209540 *T)
    sumB1  +=      12.842 * cos(2.27059911053  + 742.99006053260 *T)
    sumB1  +=      12.832 * cos(4.88898877901  + 522.57741809380 *T)
    sumB1  +=      13.137 * cos(2.45991904379  + 490.33408917940 *T)
    sumB1  +=      11.883 * cos(1.87308666696  + 423.41679713830 *T)
    sumB1  +=      13.027 * cos(3.21731634178  + 277.03499374140 *T)
    sumB1  +=       9.946 * cos(3.11650057543  + 625.67019231240 *T)
    sumB1  +=      12.710 * cos(0.29501589197  + 422.66603761290 *T)
    sumB1  +=       9.644 * cos(1.74586356703  + 330.61896365820 *T)
    sumB1  +=       8.079 * cos(2.41931187953  + 430.53034413910 *T)
    sumB1  +=       8.245 * cos(4.68121931659  + 215.74677599280 *T)
    sumB1  +=       8.958 * cos(0.46482448501  + 429.77958461370 *T)
    sumB1  +=       6.547 * cos(3.01351967549  + 949.17560896980 *T)
    sumB1  +=       7.251 * cos(5.97098186912  + 149.56319713460 *T)
    sumB1  +=       6.056 * cos(1.49115011100  + 234.63973644040 *T)
    sumB1  +=       5.791 * cos(5.36720639912  + 735.87651353180 *T)
    sumB1  +=       5.994 * cos(0.02442871989  + 654.12438031560 *T)
    sumB1  +=       6.647 * cos(3.90879134581  + 351.81659230870 *T)
    sumB1  +=       6.824 * cos(1.52456408861  + 437.64389113990 *T)
    sumB1  +=       5.134 * cos(3.81149834833  + 74.78159856730 *T)
    sumB1  +=       3.959 * cos(5.63505813057  + 210.85141488320 *T)
    sumB1  +=       3.811 * cos(2.63992803111  + 3.18139373770 *T)
    sumB1  +=       3.643 * cos(1.73267151007  + 1059.38193018920 *T)
    sumB1  +=       3.554 * cos(4.98621474362  + 3.93215326310 *T)
    sumB1  +=       4.568 * cos(4.33599514584  + 628.85158605010 *T)
    sumB1  +=       3.145 * cos(2.51404811765  + 1162.47470440780 *T)
    sumB1  +=       3.522 * cos(1.16093567319  + 223.59403617650 *T)
    sumB1  +=       2.933 * cos(2.06057834252  + 956.28915597060 *T)
    sumB1  +=       2.644 * cos(5.62559379305  + 203.73786788240 *T)
    sumB1  +=       2.992 * cos(5.06312015437  + 515.46387109300 *T)
    sumB1  +=       2.304 * cos(2.73123930535  + 21.34064100240 *T)
    sumB1  +=       2.168 * cos(2.91805928238  + 203.00415469950 *T)
    sumB1  +=       2.398 * cos(3.99421633537  + 1279.79457262800 *T)
    sumB1  +=       2.146 * cos(0.87500689888  + 408.43894361130 *T)
    sumB1  +=       2.074 * cos(1.65731069687  + 137.03302416240 *T)
    sumB1  +=       1.797 * cos(1.56879308343  + 124.43341522100 *T)
    sumB1  +=       2.088 * cos(1.85721384366  + 138.51749687070 *T)
    sumB1  +=       1.769 * cos(4.82294294946  + 1073.60902419080 *T)
    sumB1  +=       1.635 * cos(1.20387813348  + 88.86568021700 *T)
    sumB1  +=       2.202 * cos(5.93027042684  + 1052.26838318840 *T)
    sumB1  +=       1.843 * cos(0.22126910774  + 750.10360753340 *T)
    sumB1  +=       1.851 * cos(2.45470409290  + 340.77089204480 *T)
    sumB1  +=       1.890 * cos(0.41025631859  + 127.47179660680 *T)
    sumB1  +=       1.582 * cos(5.63360832372  + 214.26230328450 *T)
    sumB1  +=       1.920 * cos(3.77935901504  + 350.33211960040 *T)
    sumB1  +=       1.786 * cos(5.78644477326  + 635.96513305090 *T)
    sumB1  +=       1.497 * cos(3.13026893210  + 703.63318461740 *T)
    sumB1  +=       1.583 * cos(3.46882532865  + 38.13303563780 *T)
    sumB1  +=       1.577 * cos(4.02973226017  + 388.46515523820 *T)
    sumB1  +=       1.645 * cos(5.59115773632  + 483.22054217860 *T)
    sumB1  +=       1.405 * cos(4.07880624509  + 728.76296653100 *T)
    sumB1  +=       1.498 * cos(5.87094430469  + 362.86229257260 *T)
    sumB1  +=       1.317 * cos(2.22386203585  + 213.34727954780 *T)
    sumB1  +=       1.321 * cos(2.91534782718  + 1265.56747862640 *T)
    sumB1  +=       1.307 * cos(5.41748323885  + 217.96496188400 *T)
    sumB1  +=       1.483 * cos(0.91111666841  + 543.91805909620 *T)
    sumB1  +=       1.291 * cos(2.62333801070  + 554.06998748280 *T)
    sumB1  +=       1.406 * cos(0.34582712649  + 85.82729883120 *T)
    sumB1  +=       1.287 * cos(2.82247279651  + 231.45834270270 *T)
    sumB1  +=       1.563 * cos(4.88438049382  + 208.63322899200 *T)
    sumB1  +=       1.316 * cos(5.30963570131  + 213.25091132820 *T)
    sumB1  +=       1.164 * cos(1.39531381032  + 210.37833413120 *T)
    sumB1  +=       1.295 * cos(2.46089213219  + 200.76892246580 *T)
    sumB1  +=       1.236 * cos(3.03659580871  + 838.96928775040 *T)
    sumB1  +=       1.449 * cos(4.00934078371  + 195.13984817330 *T)
    sumB1  +=       1.251 * cos(1.46674521697  + 218.71572140940 *T)
    sumB1  +=       1.568 * cos(1.89939852487  + 212.33588759150 *T)
    sumB1  +=       1.067 * cos(5.34734443894  + 207.67002114550 *T)
    sumB1  +=       1.111 * cos(0.70962013461  + 447.93883187840 *T)
    sumB1  +=       1.012 * cos(1.37217220640  + 636.71589257630 *T)
    sumB1  +=       1.163 * cos(6.00108996618  + 191.20769491020 *T)
    sumB1  +=       0.887 * cos(2.84483069917  + 191.95845443560 *T)
    sumB1  +=       1.005 * cos(2.72373040634  + 1478.86657406440 *T)
    sumB1  +=       0.879 * cos(1.19585734916  + 417.03696332040 *T)
    sumB1  +=       0.829 * cos(4.94182950387  + 497.44763618020 *T)
    sumB1  +=       0.878 * cos(6.24981924813  + 265.98929347750 *T)
    sumB1  +=       0.781 * cos(4.61973017912  + 424.15051032120 *T)
    sumB1  +=       0.924 * cos(3.64210508536  + 222.86032299360 *T)
    sumB1  +=       0.971 * cos(2.89404568581  + 563.63121503840 *T)
    sumB1  +=       0.946 * cos(5.72987725592  + 1368.66025284500 *T)
    sumB1  +=       0.834 * cos(6.12384852532  + 209.10630974400 *T)
    sumB1  +=       0.911 * cos(1.06795057723  + 650.94298657790 *T)
    sumB1  +=       0.731 * cos(0.81660632103  + 142.44965013380 *T)
    sumB1  +=       0.795 * cos(5.54574074566  + 76.26607127560 *T)
    sumB1  +=       1.012 * cos(5.97140626297  + 643.07868005170 *T)
    sumB1  +=       0.703 * cos(2.41479303782  + 10.29494073850 *T)
    sumB1  +=       0.726 * cos(4.52598413209  + 565.11568774670 *T)
    sumB1  +=       0.691 * cos(2.31682364985  + 160.60889739850 *T)
    sumB1  +=       0.695 * cos(0.37889317398  + 212.77783057620 *T)
    sumB1  +=       0.676 * cos(4.43606270900  + 842.15068148810 *T)
    sumB1  +=       0.769 * cos(4.47368582271  + 52.69019803950 *T)
    sumB1  +=       0.658 * cos(0.32331118921  + 621.73803904930 *T)
    sumB1  +=       0.838 * cos(4.68035046143  + 288.08069400530 *T)
    sumB1  +=       0.674 * cos(5.63961963995  + 867.42347575360 *T)
    sumB1  +=       0.695 * cos(1.32062509205  + 1169.58825140860 *T)
    sumB1  +=       0.841 * cos(2.46995220838  + 1375.77379984580 *T)
    sumB1  +=       0.633 * cos(3.46145143241  + 18.15924726470 *T)
    sumB1  +=       0.722 * cos(0.04119071457  + 269.92144674060 *T)
    sumB1  +=       0.757 * cos(2.15030650611  + 207.88246946660 *T)
    sumB1  +=       0.692 * cos(0.87072324976  + 213.82036029980 *T)
    sumB1  +=       0.602 * cos(0.85288769518  + 225.82926841020 *T)
    sumB1  +=       0.635 * cos(4.76109030475  + 831.85574074960 *T)
    sumB1  +=       0.701 * cos(1.20410854661  + 479.28838891550 *T)
    sumB1  +=       0.624 * cos(2.30585779534  + 643.82943957710 *T)
    sumB1  +=       0.582 * cos(0.18811617696  + 1.48447270830 *T)
    sumB1  +=       0.513 * cos(2.56525967840  + 404.50679034820 *T)
    sumB1  +=       0.502 * cos(4.97423814367  + 212.54833591260 *T)
    sumB1  +=       0.644 * cos(2.10955154583  + 1272.68102562720 *T)
    sumB1  +=       0.467 * cos(4.47217820662  + 235.39049596580 *T)
    sumB1  +=       0.566 * cos(4.44740324446  + 429.04587143080 *T)
    sumB1  +=       0.448 * cos(0.57491120802  + 22.09140052780 *T)
    sumB1  +=       0.520 * cos(1.15131866397  + 337.73251065900 *T)
    sumB1  +=       0.476 * cos(4.78513362967  + 218.92816973050 *T)
    sumB1  +=       0.519 * cos(0.61616177345  + 436.89313161450 *T)
    sumB1  +=       0.442 * cos(2.10204008144  + 416.30325013750 *T)
    sumB1  +=       0.536 * cos(1.08779067908  + 344.70304530790 *T)
    sumB1  +=       0.477 * cos(2.42483193385  + 216.21985674480 *T)
    sumB1  +=       0.469 * cos(5.22622034028  + 942.06206196900 *T)
    sumB1  +=       0.392 * cos(0.41137926465  + 302.16477565500 *T)
    sumB1  +=       0.400 * cos(5.17216478470  + 414.06801790380 *T)
    sumB1  +=       0.383 * cos(3.58419227076  + 1045.15483618760 *T)
    sumB1  +=       0.443 * cos(1.11051326413  + 425.11371816770 *T)
    sumB1  +=       0.517 * cos(3.44547026103  + 12.53017297220 *T)
    sumB1  +=       0.369 * cos(1.60095273908  + 56.62235130260 *T)
    sumB1  +=       0.354 * cos(2.79123486392  + 1581.95934828300 *T)
    sumB1  +=       0.405 * cos(5.93402105921  + 173.94221952280 *T)
    sumB1  +=       0.319 * cos(4.31850876624  + 219.44943459230 *T)
    sumB1  +=       0.330 * cos(0.62529198264  + 358.93013930950 *T)
    sumB1  +=       0.305 * cos(0.82404423420  + 1485.98012106520 *T)
    sumB1  +=       0.391 * cos(2.59385552893  + 1795.25844372100 *T)
    sumB1  +=       0.291 * cos(3.12019266878  + 217.49188113200 *T)
    sumB1  +=       0.294 * cos(2.18552193901  + 444.75743814070 *T)
    sumB1  +=       0.281 * cos(0.77791302266  + 757.21715453420 *T)
    sumB1  +=       0.355 * cos(5.44570491928  + 1685.05212250160 *T)
    sumB1  +=       0.305 * cos(2.65927043884  + 355.74874557180 *T)
    sumB1  +=       0.269 * cos(6.00323720265  + 934.94851496820 *T)
    sumB1  +=       0.287 * cos(2.60486363576  + 113.38771495710 *T)
    sumB1  +=       0.348 * cos(0.98872551635  + 70.84944530420 *T)
    sumB1  +=       0.331 * cos(5.62133883922  + 9.56122755560 *T)
    sumB1  +=       0.255 * cos(4.14086605030  + 284.14854074220 *T)
    sumB1  +=       0.311 * cos(6.27145060602  + 207.14875628370 *T)
    sumB1  +=       0.267 * cos(4.72606312146  + 696.51963761660 *T)
    sumB1  +=       0.319 * cos(4.68828119248  + 1155.36115740700 *T)
    sumB1  +=       0.227 * cos(3.10352674343  + 1361.54670584420 *T)
    sumB1  +=       0.228 * cos(1.19253095837  + 1589.07289528380 *T)
    sumB1  +=       0.244 * cos(5.36327976010  + 245.54242435240 *T)
    sumB1  +=       0.218 * cos(2.16069250901  + 177.87437278590 *T)
    sumB1  +=       0.254 * cos(4.51652534648  + 1148.24761040620 *T)
    sumB1  +=       0.211 * cos(2.82699627326  + 106.27416795630 *T)
    sumB1  +=       0.230 * cos(4.63171743406  + 107.02492748170 *T)
    sumB1  +=       0.201 * cos(4.52152562223  + 508.35032409220 *T)
    sumB1  +=       0.230 * cos(5.93560798508  + 618.55664531160 *T)
    sumB1  +=       0.253 * cos(2.08074949572  + 252.65597135320 *T)
    sumB1  +=       0.234 * cos(2.43423283339  + 1692.16566950240 *T)
    sumB1  +=       0.196 * cos(1.01284131123  + 114.39910691340 *T)
    sumB1  +=       0.196 * cos(2.73629728926  + 214.04985496340 *T)
    sumB1  +=       0.191 * cos(1.51642829677  + 6069.77675455340 *T)
    sumB1  +=       0.252 * cos(5.10097595426  + 1258.45393162560 *T)
    sumB1  +=       0.240 * cos(2.93712394928  + 916.93228005540 *T)
    sumB1  +=       0.224 * cos(4.42406538277  + 251.43213107580 *T)
    sumB1  +=       0.223 * cos(2.34400676548  + 1677.93857550080 *T)
    sumB1  +=       0.228 * cos(5.00073557208  + 1574.84580128220 *T)
    sumB1  +=       0.183 * cos(5.09056895026  + 220.46082654860 *T)
    sumB1  +=       0.178 * cos(6.05669760153  + 206.13736432740 *T)
    sumB1  +=       0.185 * cos(3.73859309582  + 114.13847448250 *T)
    sumB1  +=       0.200 * cos(1.89409546254  + 2420.92863603340 *T)
    sumB1  +=       0.158 * cos(2.78517362162  + 2008.55753915900 *T)
    sumB1  +=       0.191 * cos(2.28724146195  + 2435.15573003500 *T)
    sumB1  +=       0.188 * cos(4.29343910228  + 1471.75302706360 *T)
    sumB1  +=       0.155 * cos(5.35194123420  + 576.16138801060 *T)
    sumB1  +=       0.174 * cos(2.80423114755  + 1781.03134971940 *T)
    sumB1  +=       0.166 * cos(5.32531835813  + 2001.44399215820 *T)
    sumB1  +=       0.165 * cos(2.62993712087  + 525.49817940060 *T)
    sumB1  +=       0.141 * cos(4.73916921092  + 501.37978944330 *T)
    sumB1  +=       0.161 * cos(5.83368523750  + 181.05576652360 *T)
    sumB1  +=       0.137 * cos(1.45135867137  + 131.54696222180 *T)
    sumB1  +=       0.157 * cos(3.22870773657  + 1493.09366806600 *T)
    sumB1  +=       0.136 * cos(4.20279658293  + 710.74673161820 *T)
    sumB1  +=       0.144 * cos(4.61003572124  + 121.25202148330 *T)
    sumB1  +=       0.131 * cos(5.85409245557  + 175.16605980020 *T)
    sumB1  +=       0.144 * cos(0.96056164245  + 214.78356814630 *T)
    sumB1  +=       0.146 * cos(3.95687956474  + 421.93232443000 *T)
    sumB1  +=       0.138 * cos(1.03080573610  + 4.66586644600 *T)
    sumB1  +=       0.167 * cos(1.97508934614  + 62.25142559510 *T)
    sumB1  +=       0.123 * cos(2.28640485589  + 1898.35121793960 *T)
    sumB1  +=       0.168 * cos(4.63122081226  + 1891.23767093880 *T)
    sumB1  +=       0.129 * cos(0.05327999225  + 211.81462272970 *T)
    sumB1  +=       0.134 * cos(3.49720535944  + 488.84961647110 *T)
    sumB1  +=       0.117 * cos(3.43819501459  + 436.15941843160 *T)
    sumB1  +=       0.125 * cos(5.87007326241  + 963.40270297140 *T)
    sumB1  +=       0.120 * cos(0.70795300239  + 81.75213321620 *T)
    sumB1  +=       0.125 * cos(4.50999471095  + 2317.83586181480 *T)
    sumB1  +=       0.116 * cos(6.11600926571  + 558.00214074590 *T)
    sumB1  +=       0.117 * cos(4.78666549046  + 601.76425067620 *T)
    sumB1  +=       0.108 * cos(0.45464469172  + 1802.37199072180 *T)
    sumB1  +=       0.111 * cos(1.44669239244  + 2531.13495725280 *T)
    sumB1  +=       0.109 * cos(6.14289264597  + 151.04766984290 *T)
    sumB1  +=       0.113 * cos(4.05600865495  + 1286.90811962880 *T)
    sumB1  +=       0.129 * cos(5.13936946160  + 849.26422848890 *T)
    sumB1  +=       0.127 * cos(3.88189432056  + 98.89998852460 *T)
    sumB1  +=       0.133 * cos(2.38290634070  + 2111.65031337760 *T)
    sumB1  +=       0.122 * cos(4.40757611742  + 778.41478318470 *T)
    sumB1  +=       0.095 * cos(0.07909774752  + 213.41097002260 *T)
    sumB1  +=       0.095 * cos(1.66925524906  + 213.18722085340 *T)
    sumB1  +=       0.103 * cos(1.88058957173  + 99.16062095550 *T)
    sumB1  +=       0.119 * cos(3.62785705509  + 248.72381809010 *T)
    sumB1  +=       0.090 * cos(4.63029999228  + 228.27694896500 *T)
    sumB1  +=       0.092 * cos(5.48700119144  + 767.36908292080 *T)
    sumB1  +=       0.089 * cos(4.61331934339  + 431.26405732200 *T)
    sumB1  +=       0.099 * cos(3.60670326134  + 776.93031047640 *T)
    sumB1  +=       0.085 * cos(4.93878023673  + 2.44768055480 *T)
    sumB1  +=       0.089 * cos(6.24541644164  + 661.23792731640 *T)
    sumB1  +=       0.085 * cos(0.45896349060  + 1382.88734684660 *T)
    sumB1  +=       0.088 * cos(3.81144552178  + 1788.14489672020 *T)
    sumB1  +=       0.103 * cos(3.20558404998  + 312.19908396260 *T)
    sumB1  +=       0.080 * cos(2.28889729136  + 213.08664711690 *T)
    sumB1  +=       0.080 * cos(5.74264101239  + 213.51154375910 *T)
    sumB1  +=       0.082 * cos(3.23546757052  + 198.32124191100 *T)
    sumB1  +=       0.078 * cos(6.03841191050  + 835.03713448730 *T)
    sumB1  +=       0.080 * cos(0.22601918692  + 427.56139872250 *T)
    sumB1  +=       0.072 * cos(2.05164614795  + 2634.22773147140 *T)
    sumB1  +=       0.091 * cos(5.97938003596  + 556.51766803760 *T)
    sumB1  +=       0.087 * cos(2.71469794199  + 617.80588578620 *T)

    # SATURN Latitude
    sumB2  = 0
    sumB2  +=   20629.977 * cos(0.50482422817  + 213.29909543800 *T)
    sumB2  +=    3719.555 * cos(3.99833475829  + 206.18554843720 *T)
    sumB2  +=    1627.158 * cos(6.18189939500  + 220.41264243880 *T)
    sumB2  +=    1346.067 * cos(0.00000000000  + 0.00000000000 *T)
    sumB2  +=     705.842 * cos(3.03914308836  + 419.48464387520 *T)
    sumB2  +=     365.042 * cos(5.09928680706  + 426.59819087600 *T)
    sumB2  +=     329.632 * cos(5.27899210039  + 433.71173787680 *T)
    sumB2  +=     219.335 * cos(3.82841533795  + 639.89728631400 *T)
    sumB2  +=     139.393 * cos(1.04272623499  + 7.11354700080 *T)
    sumB2  +=     103.980 * cos(6.15730992966  + 227.52618943960 *T)
    sumB2  +=      92.961 * cos(1.97994412845  + 316.39186965660 *T)
    sumB2  +=      71.242 * cos(4.14754353431  + 199.07200143640 *T)
    sumB2  +=      51.927 * cos(2.88364833898  + 632.78373931320 *T)
    sumB2  +=      48.961 * cos(4.43390206741  + 647.01083331480 *T)
    sumB2  +=      41.373 * cos(3.15927770079  + 853.19638175200 *T)
    sumB2  +=      28.602 * cos(4.52978327558  + 210.11770170030 *T)
    sumB2  +=      23.969 * cos(1.11595912146  + 14.22709400160 *T)
    sumB2  +=      20.511 * cos(4.35095844197  + 217.23124870110 *T)
    sumB2  +=      19.532 * cos(5.30779711223  + 440.82528487760 *T)
    sumB2  +=      18.263 * cos(0.85391476786  + 110.20632121940 *T)
    sumB2  +=      15.742 * cos(4.25767226302  + 103.09277421860 *T)
    sumB2  +=      16.840 * cos(5.68112084135  + 216.48048917570 *T)
    sumB2  +=      13.613 * cos(2.99904334066  + 412.37109687440 *T)
    sumB2  +=      11.567 * cos(2.52679928410  + 529.69096509460 *T)
    sumB2  +=       7.963 * cos(3.31512423920  + 202.25339517410 *T)
    sumB2  +=       6.599 * cos(0.28766025146  + 323.50541665740 *T)
    sumB2  +=       6.312 * cos(1.16121321336  + 117.31986822020 *T)
    sumB2  +=       5.891 * cos(3.58260177246  + 309.27832265580 *T)
    sumB2  +=       6.648 * cos(5.55714129949  + 209.36694217490 *T)
    sumB2  +=       5.590 * cos(2.47783944511  + 1066.49547719000 *T)
    sumB2  +=       6.192 * cos(3.61231886519  + 860.30992875280 *T)
    sumB2  +=       4.231 * cos(3.02212363572  + 846.08283475120 *T)
    sumB2  +=       3.612 * cos(4.79935735435  + 625.67019231240 *T)
    sumB2  +=       3.398 * cos(3.76732731354  + 423.41679713830 *T)
    sumB2  +=       3.387 * cos(6.04222745633  + 234.63973644040 *T)
    sumB2  +=       2.578 * cos(5.63610668746  + 735.87651353180 *T)
    sumB2  +=       2.831 * cos(4.81642822334  + 429.77958461370 *T)
    sumB2  +=       2.817 * cos(4.47516563908  + 654.12438031560 *T)
    sumB2  +=       2.573 * cos(0.22467245054  + 522.57741809380 *T)
    sumB2  +=       2.610 * cos(3.29126967191  + 95.97922721780 *T)
    sumB2  +=       2.419 * cos(0.02986335489  + 415.55249061210 *T)
    sumB2  +=       2.112 * cos(4.55964179603  + 422.66603761290 *T)
    sumB2  +=       2.304 * cos(6.25081073546  + 330.61896365820 *T)
    sumB2  +=       1.758 * cos(5.53430456858  + 536.80451209540 *T)
    sumB2  +=       1.814 * cos(5.05675881426  + 277.03499374140 *T)
    sumB2  +=       1.550 * cos(5.60375604692  + 223.59403617650 *T)
    sumB2  +=       1.457 * cos(4.47767649852  + 430.53034413910 *T)
    sumB2  +=       1.607 * cos(5.53599550100  + 224.34479570190 *T)
    sumB2  +=       1.172 * cos(4.71017775994  + 203.00415469950 *T)
    sumB2  +=       1.231 * cos(0.25115931880  + 3.93215326310 *T)
    sumB2  +=       1.105 * cos(1.01595427676  + 21.34064100240 *T)
    sumB2  +=       0.868 * cos(4.84623483952  + 949.17560896980 *T)
    sumB2  +=       0.939 * cos(1.35429452093  + 742.99006053260 *T)
    sumB2  +=       0.693 * cos(6.03599130692  + 124.43341522100 *T)
    sumB2  +=       0.712 * cos(4.45550701473  + 191.95845443560 *T)
    sumB2  +=       0.690 * cos(5.44243765037  + 437.64389113990 *T)
    sumB2  +=       0.810 * cos(0.46198177342  + 515.46387109300 *T)
    sumB2  +=       0.694 * cos(5.23748122403  + 447.93883187840 *T)
    sumB2  +=       0.604 * cos(2.95749705544  + 88.86568021700 *T)
    sumB2  +=       0.669 * cos(0.08457977809  + 215.74677599280 *T)
    sumB2  +=       0.579 * cos(0.65329445948  + 3.18139373770 *T)
    sumB2  +=       0.712 * cos(6.05964117622  + 11.04570026390 *T)
    sumB2  +=       0.698 * cos(2.91371419321  + 1073.60902419080 *T)
    sumB2  +=       0.526 * cos(2.24947851818  + 1059.38193018920 *T)
    sumB2  +=       0.511 * cos(2.86838724347  + 408.43894361130 *T)
    sumB2  +=       0.589 * cos(5.79268515755  + 63.73589830340 *T)
    sumB2  +=       0.519 * cos(1.76641574095  + 1279.79457262800 *T)
    sumB2  +=       0.503 * cos(5.73762297081  + 728.76296653100 *T)
    sumB2  +=       0.482 * cos(4.68234512154  + 838.96928775040 *T)
    sumB2  +=       0.494 * cos(4.04363805503  + 490.33408917940 *T)
    sumB2  +=       0.458 * cos(1.17998315936  + 210.85141488320 *T)
    sumB2  +=       0.380 * cos(5.28045750432  + 1052.26838318840 *T)
    sumB2  +=       0.404 * cos(4.58953258519  + 302.16477565500 *T)
    sumB2  +=       0.377 * cos(5.20131800999  + 74.78159856730 *T)
    sumB2  +=       0.328 * cos(0.11893501088  + 956.28915597060 *T)
    sumB2  +=       0.290 * cos(3.99300398632  + 1162.47470440780 *T)
    sumB2  +=       0.262 * cos(2.04320741578  + 1471.75302706360 *T)
    sumB2  +=       0.259 * cos(3.76206113036  + 635.96513305090 *T)
    sumB2  +=       0.254 * cos(0.16694559092  + 195.13984817330 *T)
    sumB2  +=       0.309 * cos(5.44921175960  + 543.91805909620 *T)
    sumB2  +=       0.237 * cos(1.27761853769  + 231.45834270270 *T)
    sumB2  +=       0.288 * cos(1.32449995239  + 203.73786788240 *T)
    sumB2  +=       0.229 * cos(4.19748765966  + 1265.56747862640 *T)
    sumB2  +=       0.238 * cos(4.02925601887  + 643.07868005170 *T)
    sumB2  +=       0.238 * cos(0.49997895983  + 10.29494073850 *T)
    sumB2  +=       0.257 * cos(3.69107889837  + 867.42347575360 *T)
    sumB2  +=       0.191 * cos(0.17807919948  + 628.85158605010 *T)
    sumB2  +=       0.246 * cos(5.62469599682  + 351.81659230870 *T)
    sumB2  +=       0.183 * cos(3.38184740572  + 636.71589257630 *T)
    sumB2  +=       0.172 * cos(3.83173494030  + 1581.95934828300 *T)
    sumB2  +=       0.220 * cos(1.03443668151  + 483.22054217860 *T)
    sumB2  +=       0.217 * cos(4.65210162713  + 750.10360753340 *T)
    sumB2  +=       0.143 * cos(2.31969979791  + 18.15924726470 *T)
    sumB2  +=       0.137 * cos(5.50046852846  + 1169.58825140860 *T)
    sumB2  +=       0.120 * cos(3.70151294359  + 416.30325013750 *T)
    sumB2  +=       0.136 * cos(3.38453909352  + 1155.36115740700 *T)
    sumB2  +=       0.149 * cos(0.85459831932  + 1375.77379984580 *T)
    sumB2  +=       0.150 * cos(5.71949902293  + 618.55664531160 *T)
    sumB2  +=       0.125 * cos(4.82446274394  + 436.89313161450 *T)
    sumB2  +=       0.120 * cos(3.26968058035  + 1478.86657406440 *T)
    sumB2  +=       0.131 * cos(0.11496484259  + 1898.35121793960 *T)
    sumB2  +=       0.099 * cos(4.57241894541  + 643.82943957710 *T)
    sumB2  +=       0.095 * cos(4.92115458463  + 650.94298657790 *T)
    sumB2  +=       0.090 * cos(2.09300085806  + 621.73803904930 *T)
    sumB2  +=       0.111 * cos(0.11975665259  + 831.85574074960 *T)
    sumB2  +=       0.089 * cos(2.54351587616  + 85.82729883120 *T)
    sumB2  +=       0.080 * cos(5.09103451442  + 340.77089204480 *T)
    sumB2  +=       0.078 * cos(3.17395501851  + 497.44763618020 *T)
    sumB2  +=       0.085 * cos(0.18997660997  + 1258.45393162560 *T)
    sumB2  +=       0.081 * cos(1.16732337173  + 217.96496188400 *T)
    sumB2  +=       0.072 * cos(5.47328223678  + 337.73251065900 *T)

    # SATURN Latitude
    sumB3  = 0
    sumB3  +=     666.252 * cos(1.99006340181  + 213.29909543800 *T)
    sumB3  +=     632.350 * cos(5.69778316807  + 206.18554843720 *T)
    sumB3  +=     398.051 * cos(0.00000000000  + 0.00000000000 *T)
    sumB3  +=     187.838 * cos(4.33779804809  + 220.41264243880 *T)
    sumB3  +=      91.884 * cos(4.84104208217  + 419.48464387520 *T)
    sumB3  +=      42.369 * cos(2.38073239056  + 426.59819087600 *T)
    sumB3  +=      51.548 * cos(3.42149490328  + 433.71173787680 *T)
    sumB3  +=      25.661 * cos(4.40167213109  + 227.52618943960 *T)
    sumB3  +=      20.551 * cos(5.85313509872  + 199.07200143640 *T)
    sumB3  +=      18.081 * cos(1.99321433229  + 639.89728631400 *T)
    sumB3  +=      10.874 * cos(5.37344546547  + 7.11354700080 *T)
    sumB3  +=       9.590 * cos(2.54901825866  + 647.01083331480 *T)
    sumB3  +=       7.085 * cos(3.45518372721  + 316.39186965660 *T)
    sumB3  +=       6.002 * cos(4.80055225135  + 632.78373931320 *T)
    sumB3  +=       5.778 * cos(0.01680378777  + 210.11770170030 *T)
    sumB3  +=       4.881 * cos(5.63719730884  + 14.22709400160 *T)
    sumB3  +=       4.501 * cos(1.22424419010  + 853.19638175200 *T)
    sumB3  +=       5.542 * cos(3.51756747774  + 440.82528487760 *T)
    sumB3  +=       3.548 * cos(4.71299370890  + 412.37109687440 *T)
    sumB3  +=       2.851 * cos(0.62679207578  + 103.09277421860 *T)
    sumB3  +=       2.173 * cos(3.71982274459  + 216.48048917570 *T)
    sumB3  +=       1.991 * cos(6.10867071657  + 217.23124870110 *T)
    sumB3  +=       1.435 * cos(1.69177141453  + 860.30992875280 *T)
    sumB3  +=       1.217 * cos(4.30778838827  + 234.63973644040 *T)
    sumB3  +=       1.157 * cos(5.75027789902  + 309.27832265580 *T)
    sumB3  +=       0.795 * cos(5.69026441157  + 117.31986822020 *T)
    sumB3  +=       0.733 * cos(0.59842720676  + 1066.49547719000 *T)
    sumB3  +=       0.713 * cos(0.21700311697  + 625.67019231240 *T)
    sumB3  +=       0.773 * cos(5.48361981990  + 202.25339517410 *T)
    sumB3  +=       0.897 * cos(2.65577866867  + 654.12438031560 *T)
    sumB3  +=       0.509 * cos(2.86079833766  + 429.77958461370 *T)
    sumB3  +=       0.462 * cos(4.17742567173  + 529.69096509460 *T)
    sumB3  +=       0.390 * cos(6.11288036049  + 191.95845443560 *T)
    sumB3  +=       0.505 * cos(4.51905764563  + 323.50541665740 *T)
    sumB3  +=       0.379 * cos(3.74436004151  + 223.59403617650 *T)
    sumB3  +=       0.332 * cos(5.49370890570  + 21.34064100240 *T)
    sumB3  +=       0.377 * cos(5.25624813434  + 95.97922721780 *T)
    sumB3  +=       0.384 * cos(4.48187414769  + 330.61896365820 *T)
    sumB3  +=       0.367 * cos(5.03190929680  + 846.08283475120 *T)
    sumB3  +=       0.281 * cos(1.14133888637  + 735.87651353180 *T)
    sumB3  +=       0.245 * cos(5.81618253250  + 423.41679713830 *T)
    sumB3  +=       0.241 * cos(1.70335120180  + 522.57741809380 *T)
    sumB3  +=       0.258 * cos(3.69110118716  + 447.93883187840 *T)
    sumB3  +=       0.231 * cos(4.15697626494  + 110.20632121940 *T)
    sumB3  +=       0.305 * cos(5.97746884029  + 302.16477565500 *T)
    sumB3  +=       0.284 * cos(0.66224572127  + 203.00415469950 *T)
    sumB3  +=       0.204 * cos(1.54683820621  + 209.36694217490 *T)
    sumB3  +=       0.194 * cos(4.21193801453  + 124.43341522100 *T)
    sumB3  +=       0.145 * cos(4.79689259614  + 88.86568021700 *T)
    sumB3  +=       0.151 * cos(3.82010884134  + 536.80451209540 *T)
    sumB3  +=       0.100 * cos(0.03596545368  + 949.17560896980 *T)
    sumB3  +=       0.097 * cos(0.91303450276  + 1073.60902419080 *T)
    sumB3  +=       0.110 * cos(2.21197473966  + 515.46387109300 *T)
    sumB3  +=       0.084 * cos(2.53842533109  + 422.66603761290 *T)
    sumB3  +=       0.085 * cos(5.11102520704  + 3.93215326310 *T)
    sumB3  +=       0.077 * cos(6.04074586787  + 838.96928775040 *T)
    sumB3  +=       0.085 * cos(1.18898817378  + 728.76296653100 *T)
    sumB3  +=       0.084 * cos(4.10158366806  + 224.34479570190 *T)

    # SATURN Latitude
    sumB4  = 0
    sumB4  +=      80.384 * cos(1.11918414679  + 206.18554843720 *T)
    sumB4  +=      31.660 * cos(3.12218745098  + 213.29909543800 *T)
    sumB4  +=      17.143 * cos(2.48073200414  + 220.41264243880 *T)
    sumB4  +=      11.844 * cos(3.14159265359  + 0.00000000000 *T)
    sumB4  +=       9.005 * cos(0.38441424927  + 419.48464387520 *T)
    sumB4  +=       6.164 * cos(1.56186379537  + 433.71173787680 *T)
    sumB4  +=       4.660 * cos(1.28235639570  + 199.07200143640 *T)
    sumB4  +=       4.775 * cos(2.63498295487  + 227.52618943960 *T)
    sumB4  +=       1.487 * cos(1.43096671616  + 426.59819087600 *T)
    sumB4  +=       1.424 * cos(0.66988083613  + 647.01083331480 *T)
    sumB4  +=       1.075 * cos(6.18092274059  + 639.89728631400 *T)
    sumB4  +=       1.145 * cos(1.72041928134  + 440.82528487760 *T)
    sumB4  +=       0.682 * cos(3.84841098180  + 14.22709400160 *T)
    sumB4  +=       0.655 * cos(3.49486258327  + 7.11354700080 *T)
    sumB4  +=       0.456 * cos(0.47338193402  + 632.78373931320 *T)
    sumB4  +=       0.509 * cos(0.31432285584  + 412.37109687440 *T)
    sumB4  +=       0.343 * cos(5.86413875355  + 853.19638175200 *T)
    sumB4  +=       0.270 * cos(2.50125594913  + 234.63973644040 *T)
    sumB4  +=       0.197 * cos(5.39156324804  + 316.39186965660 *T)
    sumB4  +=       0.236 * cos(2.11084590211  + 210.11770170030 *T)
    sumB4  +=       0.172 * cos(6.09682874401  + 860.30992875280 *T)
    sumB4  +=       0.159 * cos(5.95049154821  + 216.48048917570 *T)
    sumB4  +=       0.100 * cos(1.98534903594  + 625.67019231240 *T)
    sumB4  +=       0.112 * cos(0.85526419268  + 654.12438031560 *T)
    sumB4  +=       0.115 * cos(5.03884718594  + 117.31986822020 *T)
    sumB4  +=       0.115 * cos(0.44589613974  + 110.20632121940 *T)

    # SATURN Latitude
    sumB5  = 0
    sumB5  +=       7.895 * cos(2.81927558645  + 206.18554843720 *T)
    sumB5  +=       1.014 * cos(0.51187210270  + 220.41264243880 *T)
    sumB5  +=       0.772 * cos(2.99484124049  + 199.07200143640 *T)
    sumB5  +=       0.967 * cos(3.14159265359  + 0.00000000000 *T)
    sumB5  +=       0.583 * cos(5.96456944075  + 433.71173787680 *T)
    sumB5  +=       0.588 * cos(0.78008666397  + 227.52618943960 *T)
    sumB5  +=       0.445 * cos(2.38630799074  + 419.48464387520 *T)
    sumB5  +=       0.098 * cos(5.10622131539  + 647.01083331480 *T)
    sumB5  +=       0.091 * cos(5.81659714144  + 7.11354700080 *T)
    sumB5  +=       0.088 * cos(6.17828532308  + 440.82528487760 *T)
    sumB5  +=       0.089 * cos(0.58396864530  + 213.29909543800 *T)

    # SATURN Radius
    sumR0  = 0
    sumR0  +=  955758135.801 * cos(0.00000000000  + 0.00000000000 *T)
    sumR0  +=  52921382.465 * cos(2.39226219733  + 213.29909543800 *T)
    sumR0  +=  1873679.934 * cos(5.23549605091  + 206.18554843720 *T)
    sumR0  +=  1464663.959 * cos(1.64763045468  + 426.59819087600 *T)
    sumR0  +=  821891.059 * cos(5.93520025371  + 316.39186965660 *T)
    sumR0  +=  547506.899 * cos(5.01532628454  + 103.09277421860 *T)
    sumR0  +=  371684.449 * cos(2.27114833428  + 220.41264243880 *T)
    sumR0  +=  361778.433 * cos(3.13904303264  + 7.11354700080 *T)
    sumR0  +=  140617.548 * cos(5.70406652991  + 632.78373931320 *T)
    sumR0  +=  108974.737 * cos(3.29313595577  + 110.20632121940 *T)
    sumR0  +=   69007.015 * cos(5.94099622447  + 419.48464387520 *T)
    sumR0  +=   61053.350 * cos(0.94037761156  + 639.89728631400 *T)
    sumR0  +=   48913.044 * cos(1.55733388472  + 202.25339517410 *T)
    sumR0  +=   34143.794 * cos(0.19518550682  + 277.03499374140 *T)
    sumR0  +=   32401.718 * cos(5.47084606947  + 949.17560896980 *T)
    sumR0  +=   20936.573 * cos(0.46349163993  + 735.87651353180 *T)
    sumR0  +=   20839.118 * cos(1.52102590640  + 433.71173787680 *T)
    sumR0  +=   20746.678 * cos(5.33255667599  + 199.07200143640 *T)
    sumR0  +=   15298.457 * cos(3.05943652881  + 529.69096509460 *T)
    sumR0  +=   14296.479 * cos(2.60433537909  + 323.50541665740 *T)
    sumR0  +=   11993.314 * cos(5.98051421881  + 846.08283475120 *T)
    sumR0  +=   11380.261 * cos(1.73105746566  + 522.57741809380 *T)
    sumR0  +=   12884.128 * cos(1.64892310393  + 138.51749687070 *T)
    sumR0  +=    7752.769 * cos(5.85191318903  + 95.97922721780 *T)
    sumR0  +=    9796.061 * cos(5.20475863996  + 1265.56747862640 *T)
    sumR0  +=    6465.967 * cos(0.17733160145  + 1052.26838318840 *T)
    sumR0  +=    6770.621 * cos(3.00433479284  + 14.22709400160 *T)
    sumR0  +=    5850.443 * cos(1.45519636076  + 415.55249061210 *T)
    sumR0  +=    5307.481 * cos(0.59737534050  + 63.73589830340 *T)
    sumR0  +=    4695.746 * cos(2.14919036956  + 227.52618943960 *T)
    sumR0  +=    4043.988 * cos(1.64010323863  + 209.36694217490 *T)
    sumR0  +=    3688.132 * cos(0.78016133170  + 412.37109687440 *T)
    sumR0  +=    3376.457 * cos(3.69528478828  + 224.34479570190 *T)
    sumR0  +=    2885.348 * cos(1.38764077631  + 838.96928775040 *T)
    sumR0  +=    2976.033 * cos(5.68467931117  + 210.11770170030 *T)
    sumR0  +=    3419.551 * cos(4.94549148887  + 1581.95934828300 *T)
    sumR0  +=    3460.943 * cos(1.85088802878  + 175.16605980020 *T)
    sumR0  +=    3400.616 * cos(0.55386747515  + 350.33211960040 *T)
    sumR0  +=    2507.630 * cos(3.53851863255  + 742.99006053260 *T)
    sumR0  +=    2448.325 * cos(6.18412386316  + 1368.66025284500 *T)
    sumR0  +=    2406.138 * cos(2.96559220267  + 117.31986822020 *T)
    sumR0  +=    2881.181 * cos(0.17960757891  + 853.19638175200 *T)
    sumR0  +=    2173.959 * cos(0.01508587396  + 340.77089204480 *T)
    sumR0  +=    2024.483 * cos(5.05411271271  + 11.04570026390 *T)
    sumR0  +=    1740.254 * cos(2.34657043464  + 309.27832265580 *T)
    sumR0  +=    1861.397 * cos(5.93361638244  + 625.67019231240 *T)
    sumR0  +=    1888.436 * cos(0.02968443389  + 3.93215326310 *T)
    sumR0  +=    1610.859 * cos(1.17302463549  + 74.78159856730 *T)
    sumR0  +=    1462.631 * cos(1.92588134017  + 216.48048917570 *T)
    sumR0  +=    1474.547 * cos(5.67670461130  + 203.73786788240 *T)
    sumR0  +=    1395.109 * cos(5.93669404929  + 127.47179660680 *T)
    sumR0  +=    1781.165 * cos(0.76314388077  + 217.23124870110 *T)
    sumR0  +=    1817.186 * cos(5.77713225779  + 490.33408917940 *T)
    sumR0  +=    1472.392 * cos(1.40064915651  + 137.03302416240 *T)
    sumR0  +=    1304.089 * cos(0.77235613966  + 647.01083331480 *T)
    sumR0  +=    1149.773 * cos(5.74021249703  + 1162.47470440780 *T)
    sumR0  +=    1126.667 * cos(4.46707803791  + 265.98929347750 *T)
    sumR0  +=    1277.489 * cos(2.98412586423  + 1059.38193018920 *T)
    sumR0  +=    1207.053 * cos(0.75285933160  + 351.81659230870 *T)
    sumR0  +=    1071.399 * cos(1.13567265104  + 1155.36115740700 *T)
    sumR0  +=    1020.922 * cos(5.91233512844  + 1685.05212250160 *T)
    sumR0  +=    1315.042 * cos(5.11202572637  + 211.81462272970 *T)
    sumR0  +=    1295.553 * cos(4.69184139933  + 1898.35121793960 *T)
    sumR0  +=    1099.037 * cos(1.81765118601  + 149.56319713460 *T)
    sumR0  +=     998.462 * cos(2.63131596867  + 200.76892246580 *T)
    sumR0  +=     985.869 * cos(2.25992849742  + 956.28915597060 *T)
    sumR0  +=     932.434 * cos(3.66980793184  + 554.06998748280 *T)
    sumR0  +=     664.481 * cos(0.60297724821  + 728.76296653100 *T)
    sumR0  +=     659.850 * cos(4.66635439533  + 195.13984817330 *T)
    sumR0  +=     617.740 * cos(5.62092000007  + 942.06206196900 *T)
    sumR0  +=     626.382 * cos(5.94208232590  + 1478.86657406440 *T)
    sumR0  +=     482.230 * cos(1.84070179496  + 479.28838891550 *T)
    sumR0  +=     487.689 * cos(2.79373616806  + 3.18139373770 *T)
    sumR0  +=     470.086 * cos(0.83847755040  + 1471.75302706360 *T)
    sumR0  +=     451.817 * cos(5.64468459871  + 2001.44399215820 *T)
    sumR0  +=     553.128 * cos(3.41088600844  + 269.92144674060 *T)
    sumR0  +=     534.397 * cos(1.26443331367  + 275.55052103310 *T)
    sumR0  +=     472.572 * cos(1.88198584660  + 515.46387109300 *T)
    sumR0  +=     405.434 * cos(1.64001413521  + 536.80451209540 *T)
    sumR0  +=     517.196 * cos(4.44310450526  + 2214.74308759620 *T)
    sumR0  +=     452.848 * cos(3.00349117198  + 302.16477565500 *T)
    sumR0  +=     494.340 * cos(2.28626675074  + 278.51946644970 *T)
    sumR0  +=     489.825 * cos(5.80631420383  + 191.20769491020 *T)
    sumR0  +=     427.459 * cos(0.05741344372  + 284.14854074220 *T)
    sumR0  +=     339.763 * cos(1.40198657693  + 440.82528487760 *T)
    sumR0  +=     340.627 * cos(0.89091104306  + 628.85158605010 *T)
    sumR0  +=     385.974 * cos(1.99700402508  + 1272.68102562720 *T)
    sumR0  +=     288.298 * cos(1.12160250272  + 422.66603761290 *T)
    sumR0  +=     294.444 * cos(0.42577061903  + 312.19908396260 *T)
    sumR0  +=     262.490 * cos(0.31753439818  + 1045.15483618760 *T)
    sumR0  +=     295.331 * cos(0.67144493789  + 88.86568021700 *T)
    sumR0  +=     342.968 * cos(5.85600322299  + 1795.25844372100 *T)
    sumR0  +=     341.117 * cos(2.37585247250  + 525.49817940060 *T)
    sumR0  +=     234.018 * cos(4.22756813216  + 114.13847448250 *T)
    sumR0  +=     223.729 * cos(2.28129446763  + 330.61896365820 *T)
    sumR0  +=     275.814 * cos(0.47832439352  + 38.13303563780 *T)
    sumR0  +=     224.592 * cos(0.54754005675  + 1788.14489672020 *T)
    sumR0  +=     303.300 * cos(0.87946670205  + 6069.77675455340 *T)
    sumR0  +=     292.103 * cos(6.21420611920  + 210.85141488320 *T)
    sumR0  +=     226.121 * cos(0.37495223398  + 142.44965013380 *T)
    sumR0  +=     277.257 * cos(5.31917702012  + 692.58748435350 *T)
    sumR0  +=     242.911 * cos(5.37187983246  + 1258.45393162560 *T)
    sumR0  +=     205.571 * cos(0.95755250527  + 288.08069400530 *T)
    sumR0  +=     207.567 * cos(5.38126259725  + 2317.83586181480 *T)
    sumR0  +=     186.835 * cos(6.03591766061  + 404.50679034820 *T)
    sumR0  +=     218.536 * cos(5.25607043545  + 212.33588759150 *T)
    sumR0  +=     222.155 * cos(5.94588016768  + 39.35687591520 *T)
    sumR0  +=     179.673 * cos(4.41045924362  + 408.43894361130 *T)
    sumR0  +=     241.440 * cos(1.12525868110  + 388.46515523820 *T)
    sumR0  +=     197.093 * cos(3.90141942850  + 52.69019803950 *T)
    sumR0  +=     236.639 * cos(0.90802744873  + 1375.77379984580 *T)
    sumR0  +=     171.915 * cos(5.56318632797  + 213.34727954780 *T)
    sumR0  +=     169.865 * cos(2.85667554010  + 99.16062095550 *T)
    sumR0  +=     214.398 * cos(4.20253525974  + 2531.13495725280 *T)
    sumR0  +=     172.010 * cos(2.36537801012  + 213.25091132820 *T)
    sumR0  +=     165.707 * cos(2.63679789706  + 215.74677599280 *T)
    sumR0  +=     230.892 * cos(5.49463421262  + 191.95845443560 *T)
    sumR0  +=     177.585 * cos(0.38155817719  + 430.53034413910 *T)
    sumR0  +=     191.514 * cos(2.95906900704  + 437.64389113990 *T)
    sumR0  +=     163.250 * cos(3.45832517280  + 617.80588578620 *T)
    sumR0  +=     162.305 * cos(5.73050678664  + 203.00415469950 *T)
    sumR0  +=     175.108 * cos(5.71404465044  + 1066.49547719000 *T)
    sumR0  +=     183.041 * cos(5.66851947172  + 2111.65031337760 *T)
    sumR0  +=     150.077 * cos(4.40663921925  + 417.03696332040 *T)
    sumR0  +=     187.935 * cos(6.07916265661  + 563.63121503840 *T)
    sumR0  +=     145.127 * cos(5.08176368814  + 423.41679713830 *T)
    sumR0  +=     137.491 * cos(5.43912787991  + 222.86032299360 *T)
    sumR0  +=     172.824 * cos(1.84920994090  + 1589.07289528380 *T)
    sumR0  +=     165.478 * cos(2.89132196119  + 214.26230328450 *T)
    sumR0  +=     145.727 * cos(1.56565192483  + 831.85574074960 *T)
    sumR0  +=     176.864 * cos(2.30323752987  + 9999.98645077300 *T)
    sumR0  +=     128.877 * cos(2.55338644107  + 414.06801790380 *T)
    sumR0  +=     120.093 * cos(0.04329750542  + 1361.54670584420 *T)
    sumR0  +=     143.441 * cos(0.99817357720  + 76.26607127560 *T)
    sumR0  +=     108.747 * cos(2.09282278191  + 207.67002114550 *T)
    sumR0  +=     132.106 * cos(2.85902597898  + 312.45971639350 *T)
    sumR0  +=     112.238 * cos(0.26221759151  + 2104.53676637680 *T)
    sumR0  +=     125.186 * cos(4.78354048063  + 205.22234059070 *T)
    sumR0  +=     104.427 * cos(3.63671899047  + 65.22037101170 *T)
    sumR0  +=     107.447 * cos(3.67064138701  + 212.77783057620 *T)
    sumR0  +=     108.642 * cos(2.85492389024  + 21.34064100240 *T)
    sumR0  +=      97.743 * cos(5.12231845599  + 2634.22773147140 *T)
    sumR0  +=     109.097 * cos(1.63231061493  + 208.63322899200 *T)
    sumR0  +=      96.852 * cos(4.19928280035  + 305.34616939270 *T)
    sumR0  +=      96.507 * cos(2.56002066845  + 1692.16566950240 *T)
    sumR0  +=      85.829 * cos(4.54545085982  + 210.37833413120 *T)
    sumR0  +=      99.249 * cos(5.13816222131  + 1574.84580128220 *T)
    sumR0  +=     112.532 * cos(5.03109281265  + 703.63318461740 *T)
    sumR0  +=      84.023 * cos(1.18337717265  + 429.77958461370 *T)
    sumR0  +=      89.021 * cos(5.38791571457  + 107.02492748170 *T)
    sumR0  +=     110.191 * cos(2.43656081234  + 355.74874557180 *T)
    sumR0  +=      90.659 * cos(4.20908809746  + 213.82036029980 *T)
    sumR0  +=      95.885 * cos(5.44594259071  + 2428.04218303420 *T)
    sumR0  +=      94.109 * cos(2.39786381418  + 483.22054217860 *T)
    sumR0  +=      85.609 * cos(0.03354346966  + 860.30992875280 *T)
    sumR0  +=      88.796 * cos(4.05766306750  + 128.95626931510 *T)
    sumR0  +=      81.951 * cos(1.66499731549  + 62.25142559510 *T)
    sumR0  +=      91.240 * cos(3.96942332591  + 2847.52682690940 *T)
    sumR0  +=      83.961 * cos(4.60845858022  + 177.87437278590 *T)
    sumR0  +=      88.376 * cos(3.86800515885  + 140.00196957900 *T)
    sumR0  +=      93.308 * cos(0.73846639887  + 831.10498122420 *T)
    sumR0  +=      91.872 * cos(2.94977605320  + 35.42472265210 *T)
    sumR0  +=      87.077 * cos(1.33390590052  + 1905.46476494040 *T)
    sumR0  +=      96.584 * cos(4.84438390997  + 131.40394986990 *T)
    sumR0  +=      71.010 * cos(0.99334817658  + 405.25754987360 *T)
    sumR0  +=      95.266 * cos(2.51506908152  + 2.44768055480 *T)
    sumR0  +=      72.514 * cos(4.63213873657  + 245.54242435240 *T)
    sumR0  +=      82.580 * cos(1.52823217919  + 145.63104387150 *T)
    sumR0  +=      76.693 * cos(3.15240783008  + 767.36908292080 *T)
    sumR0  +=      70.317 * cos(4.04253707270  + 173.94221952280 *T)
    sumR0  +=      86.015 * cos(2.30103727270  + 85.82729883120 *T)
    sumR0  +=      66.529 * cos(4.75053522835  + 70.84944530420 *T)
    sumR0  +=      65.835 * cos(2.46869725001  + 280.96714700450 *T)
    sumR0  +=      64.824 * cos(0.09343869325  + 9.56122755560 *T)
    sumR0  +=      71.557 * cos(0.01212415296  + 565.11568774670 *T)
    sumR0  +=      66.533 * cos(1.08034871114  + 339.28641933650 *T)
    sumR0  +=      63.488 * cos(2.01740971153  + 234.63973644040 *T)
    sumR0  +=      60.786 * cos(5.12026947473  + 756.32338265690 *T)
    sumR0  +=      58.123 * cos(6.05732868566  + 1677.93857550080 *T)
    sumR0  +=      64.236 * cos(1.28586474622  + 1148.24761040620 *T)
    sumR0  +=      73.124 * cos(4.37810889148  + 425.11371816770 *T)
    sumR0  +=      55.012 * cos(3.85865703217  + 342.25536475310 *T)
    sumR0  +=      57.101 * cos(6.26689214029  + 2420.92863603340 *T)
    sumR0  +=      64.090 * cos(4.09854757476  + 327.43756992050 *T)
    sumR0  +=      55.306 * cos(1.60456896521  + 543.02428721890 *T)
    sumR0  +=      57.987 * cos(5.47269124340  + 347.88443904560 *T)
    sumR0  +=      73.581 * cos(3.72292337326  + 92.04707395470 *T)
    sumR0  +=      73.760 * cos(3.57045342615  + 1.48447270830 *T)
    sumR0  +=      64.940 * cos(2.44739629174  + 267.47376618580 *T)
    sumR0  +=      54.414 * cos(3.71479080197  + 344.70304530790 *T)
    sumR0  +=      49.783 * cos(3.93453970179  + 192.69216761850 *T)
    sumR0  +=      49.537 * cos(3.22831070579  + 333.65734504400 *T)
    sumR0  +=      47.539 * cos(3.92925402178  + 199.28444975750 *T)
    sumR0  +=      49.368 * cos(4.90341763553  + 217.49188113200 *T)
    sumR0  +=      62.711 * cos(4.40120079629  + 214.78356814630 *T)
    sumR0  +=      46.359 * cos(2.09430260266  + 212.54833591260 *T)
    sumR0  +=      46.289 * cos(2.64038453480  + 10.29494073850 *T)
    sumR0  +=      54.335 * cos(1.07179534996  + 362.86229257260 *T)
    sumR0  +=      58.742 * cos(2.62270940799  + 225.82926841020 *T)
    sumR0  +=      48.457 * cos(3.15166418511  + 216.21985674480 *T)
    sumR0  +=      46.316 * cos(4.86226642770  + 2950.61960112800 *T)
    sumR0  +=      45.970 * cos(4.97297391881  + 198.32124191100 *T)
    sumR0  +=      46.678 * cos(2.44960215701  + 207.14875628370 *T)
    sumR0  +=      44.905 * cos(1.77616995803  + 223.59403617650 *T)
    sumR0  +=      44.521 * cos(5.55987055442  + 264.50482076920 *T)
    sumR0  +=      55.914 * cos(4.29520232351  + 329.72519178090 *T)
    sumR0  +=      49.643 * cos(5.20789299388  + 2744.43405269080 *T)
    sumR0  +=      58.829 * cos(4.23073947869  + 700.66423920080 *T)
    sumR0  +=      52.629 * cos(3.79230629070  + 343.21857259960 *T)
    sumR0  +=      41.532 * cos(0.74488808688  + 125.98732389850 *T)
    sumR0  +=      47.767 * cos(2.39260015876  + 207.88246946660 *T)
    sumR0  +=      56.157 * cos(2.07214273531  + 124.43341522100 *T)
    sumR0  +=      43.345 * cos(1.83707598036  + 106.27416795630 *T)
    sumR0  +=      39.793 * cos(4.00870764324  + 12.53017297220 *T)
    sumR0  +=      53.882 * cos(4.97905460628  + 134.58534360760 *T)
    sumR0  +=      50.135 * cos(5.75914508514  + 320.32402291970 *T)
    sumR0  +=      44.960 * cos(5.35721924134  + 218.92816973050 *T)
    sumR0  +=      41.089 * cos(4.92252591399  + 1891.23767093880 *T)
    sumR0  +=      46.509 * cos(2.06623129884  + 2008.55753915900 *T)
    sumR0  +=      42.949 * cos(0.39856812529  + 357.44566660120 *T)
    sumR0  +=      37.992 * cos(2.06495914285  + 247.23934538180 *T)
    sumR0  +=      48.733 * cos(5.32762223699  + 3127.31333126180 *T)
    sumR0  +=      34.583 * cos(5.62555932761  + 99.91138048090 *T)
    sumR0  +=      41.092 * cos(2.47264897370  + 237.67811782620 *T)
    sumR0  +=      40.763 * cos(4.08408559215  + 621.73803904930 *T)
    sumR0  +=      34.213 * cos(0.73077393007  + 750.10360753340 *T)
    sumR0  +=      33.967 * cos(5.31264617621  + 206.23373254700 *T)
    sumR0  +=      36.509 * cos(1.68826775750  + 22.09140052780 *T)
    sumR0  +=      39.361 * cos(3.45730719990  + 241.61027108930 *T)
    sumR0  +=      34.796 * cos(2.24780137629  + 487.36514376280 *T)
    sumR0  +=      33.049 * cos(4.86593901955  + 209.10630974400 *T)
    sumR0  +=      32.584 * cos(2.22713131846  + 319.57326339430 *T)
    sumR0  +=      39.035 * cos(3.73870591196  + 3163.91869656600 *T)
    sumR0  +=      32.722 * cos(1.06640549236  + 252.65597135320 *T)
    sumR0  +=      38.671 * cos(4.39617126814  + 18.15924726470 *T)
    sumR0  +=      34.514 * cos(1.82607500690  + 380.12776796000 *T)
    sumR0  +=      41.539 * cos(0.08136234251  + 210.33015002140 *T)
    sumR0  +=      33.527 * cos(5.80475568528  + 251.43213107580 *T)
    sumR0  +=      31.221 * cos(1.96489151107  + 244.31858407500 *T)
    sumR0  +=      30.521 * cos(2.26854188579  + 1169.58825140860 *T)
    sumR0  +=      34.828 * cos(5.96324553131  + 217.96496188400 *T)
    sumR0  +=      38.481 * cos(4.43707551964  + 160.60889739850 *T)
    sumR0  +=      35.998 * cos(3.83262381556  + 56.62235130260 *T)
    sumR0  +=      31.041 * cos(4.89914223233  + 144.14657116320 *T)
    sumR0  +=      32.342 * cos(3.58191018804  + 231.45834270270 *T)
    sumR0  +=      28.838 * cos(5.80081031514  + 1994.33044515740 *T)
    sumR0  +=      32.175 * cos(2.13166877923  + 206.13736432740 *T)
    sumR0  +=      32.643 * cos(1.93131580544  + 98.89998852460 *T)
    sumR0  +=      34.917 * cos(5.65276617691  + 497.44763618020 *T)
    sumR0  +=      28.928 * cos(2.21653288920  + 14.97785352700 *T)
    sumR0  +=      31.569 * cos(3.81846560564  + 73.29712585900 *T)
    sumR0  +=      32.199 * cos(0.99811846290  + 1464.63948006280 *T)
    sumR0  +=      29.153 * cos(5.98414099408  + 2737.32050569000 *T)
    sumR0  +=      36.706 * cos(4.75493516597  + 348.84764689210 *T)
    sumR0  +=      28.665 * cos(1.68732054583  + 78.71375183040 *T)
    sumR0  +=      27.501 * cos(6.12086395418  + 214.04985496340 *T)
    sumR0  +=      28.795 * cos(0.04448605904  + 5.62907429250 *T)
    sumR0  +=      27.205 * cos(0.24587543816  + 313.21047591890 *T)
    sumR0  +=      32.441 * cos(3.77921585847  + 33.94024994380 *T)
    sumR0  +=      27.088 * cos(5.20310098020  + 148.07872442630 *T)
    sumR0  +=      34.956 * cos(3.43886187587  + 273.10284047830 *T)
    sumR0  +=      33.076 * cos(2.44662095168  + 969.62247809490 *T)
    sumR0  +=      27.745 * cos(1.44598606685  + 258.87574647670 *T)
    sumR0  +=      27.178 * cos(4.25918596220  + 179.35884549420 *T)
    sumR0  +=      27.872 * cos(0.78772093522  + 546.95644048200 *T)
    sumR0  +=      29.106 * cos(4.83947711462  + 905.88657979150 *T)
    sumR0  +=      27.417 * cos(2.44930366818  + 254.94359321360 *T)
    sumR0  +=      34.296 * cos(6.00920969644  + 166.82867252200 *T)
    sumR0  +=      28.859 * cos(6.02917249910  + 188.92007304980 *T)
    sumR0  +=      26.001 * cos(0.65046992484  + 654.12438031560 *T)
    sumR0  +=      33.560 * cos(1.23732329127  + 2221.85663459700 *T)
    sumR0  +=      24.356 * cos(0.52248751330  + 894.84087952760 *T)
    sumR0  +=      27.767 * cos(5.17820678484  + 5.41662597140 *T)
    sumR0  +=      25.568 * cos(3.35897159622  + 0.96320784650 *T)
    sumR0  +=      22.879 * cos(3.51293480690  + 458.84151979040 *T)
    sumR0  +=      24.496 * cos(0.00976884124  + 69.15252427480 *T)
    sumR0  +=      28.794 * cos(0.75545700854  + 488.84961647110 *T)
    sumR0  +=      31.228 * cos(2.05299907796  + 282.45161971280 *T)
    sumR0  +=      25.438 * cos(5.29037729250  + 636.71589257630 *T)
    sumR0  +=      25.332 * cos(4.97007969450  + 3060.82592234740 *T)
    sumR0  +=      23.596 * cos(2.54766434769  + 196.62432088160 *T)
    sumR0  +=      29.602 * cos(3.92688207792  + 206.70681329900 *T)
    sumR0  +=      28.255 * cos(2.72125009693  + 32.24332891440 *T)
    sumR0  +=      22.115 * cos(4.75775237642  + 213.18722085340 *T)
    sumR0  +=      22.130 * cos(3.25436709191  + 681.54178408960 *T)
    sumR0  +=      21.675 * cos(4.61403328597  + 3267.01147078460 *T)
    sumR0  +=      22.115 * cos(3.16759500067  + 213.41097002260 *T)
    sumR0  +=      26.912 * cos(2.86269769133  + 24.37902238820 *T)
    sumR0  +=      20.737 * cos(1.66895754198  + 274.06604832480 *T)
    sumR0  +=      28.309 * cos(4.73122154345  + 552.58551477450 *T)
    sumR0  +=      25.252 * cos(5.11986371899  + 168.05251279940 *T)
    sumR0  +=      26.364 * cos(1.59272536419  + 491.81856188770 *T)
    sumR0  +=      21.995 * cos(0.88079009280  + 635.96513305090 *T)
    sumR0  +=      27.076 * cos(5.53694832022  + 555.55446019110 *T)
    sumR0  +=      19.683 * cos(2.14388519695  + 54.17467074780 *T)
    sumR0  +=      27.266 * cos(3.57891326986  + 561.18353448360 *T)
    sumR0  +=      25.162 * cos(1.78070903718  + 182.27960680100 *T)
    sumR0  +=      21.386 * cos(3.86030772476  + 116.42609634290 *T)
    sumR0  +=      25.572 * cos(1.62093861709  + 2324.94940881560 *T)
    sumR0  +=      20.025 * cos(2.90618582553  + 120.35824960600 *T)
    sumR0  +=      19.882 * cos(5.59203696008  + 4.19278569400 *T)
    sumR0  +=      19.454 * cos(0.10623632006  + 218.71572140940 *T)
    sumR0  +=      25.617 * cos(2.09931460158  + 248.72381809010 *T)
    sumR0  +=      19.804 * cos(2.52180124343  + 1485.98012106520 *T)
    sumR0  +=      18.516 * cos(2.54810951896  + 213.51154375910 *T)
    sumR0  +=      19.831 * cos(0.07955320843  + 842.15068148810 *T)
    sumR0  +=      18.516 * cos(5.37755110510  + 213.08664711690 *T)
    sumR0  +=      23.655 * cos(1.59974907716  + 738.79727483860 *T)
    sumR0  +=      20.375 * cos(2.94653107321  + 59.80374504030 *T)
    sumR0  +=      24.247 * cos(3.15387696867  + 240.38643081190 *T)
    sumR0  +=      18.294 * cos(3.18715992969  + 295.05122865420 *T)
    sumR0  +=      17.464 * cos(2.90471803626  + 477.80391620720 *T)
    sumR0  +=      20.698 * cos(1.07232100334  + 494.26624244250 *T)
    sumR0  +=      20.400 * cos(1.83665590916  + 533.62311835770 *T)
    sumR0  +=      21.285 * cos(0.63341794388  + 189.72322220190 *T)
    sumR0  +=      16.116 * cos(0.60069688498  + 746.92221379570 *T)
    sumR0  +=      16.297 * cos(3.98317294128  + 2.92076130680 *T)
    sumR0  +=      16.922 * cos(4.74266972033  + 2207.62954059540 *T)
    sumR0  +=      20.479 * cos(6.05098286202  + 173.68158709190 *T)
    sumR0  +=      15.447 * cos(1.49120311247  + 543.91805909620 *T)
    sumR0  +=      19.944 * cos(4.94086632750  + 121.25202148330 *T)
    sumR0  +=      17.127 * cos(0.71458025372  + 1781.03134971940 *T)
    sumR0  +=      17.240 * cos(0.67749766724  + 151.04766984290 *T)
    sumR0  +=      15.574 * cos(5.70296527381  + 3053.71237534660 *T)
    sumR0  +=      15.036 * cos(5.52770334605  + 2310.72231481400 *T)
    sumR0  +=      15.928 * cos(4.45642717299  + 643.82943957710 *T)
    sumR0  +=      16.165 * cos(0.63286131026  + 358.93013930950 *T)
    sumR0  +=      14.589 * cos(5.26158292613  + 472.17484191470 *T)
    sumR0  +=      16.545 * cos(3.52813228069  + 3480.31056622260 *T)
    sumR0  +=      18.912 * cos(0.55218675639  + 4.66586644600 *T)
    sumR0  +=      17.595 * cos(2.26495491189  + 672.14061522840 *T)
    sumR0  +=      18.104 * cos(2.71285673689  + 181.80652604900 *T)
    sumR0  +=      15.918 * cos(5.23446779429  + 135.54855145410 *T)
    sumR0  +=      13.931 * cos(3.19357128657  + 213.55972786890 *T)
    sumR0  +=      14.058 * cos(0.82375896652  + 221.37585028530 *T)
    sumR0  +=      13.931 * cos(4.73208739639  + 213.03846300710 *T)
    sumR0  +=      14.690 * cos(2.65882838685  + 292.01284726840 *T)
    sumR0  +=      14.454 * cos(0.21819892811  + 235.39049596580 *T)
    sumR0  +=      16.168 * cos(0.91025406068  + 280.00393915800 *T)
    sumR0  +=      13.327 * cos(3.54947442109  + 205.66428357540 *T)
    sumR0  +=      16.104 * cos(0.82547975762  + 176.65053250850 *T)
    sumR0  +=      16.441 * cos(5.39398801335  + 424.15051032120 *T)
    sumR0  +=      12.747 * cos(0.75780958758  + 721.64941953020 *T)
    sumR0  +=      12.754 * cos(3.55466871752  + 153.49535039770 *T)
    sumR0  +=      14.448 * cos(0.12049617049  + 313.68355667090 *T)
    sumR0  +=      16.499 * cos(3.26383140489  + 6283.07584999140 *T)
    sumR0  +=      16.564 * cos(1.62649604519  + 5856.47765911540 *T)
    sumR0  +=      14.950 * cos(1.23923264394  + 2641.34127847220 *T)
    sumR0  +=      15.724 * cos(1.18874754834  + 486.40193591630 *T)
    sumR0  +=      11.893 * cos(0.91693668558  + 416.30325013750 *T)
    sumR0  +=      11.684 * cos(1.11385455828  + 81.75213321620 *T)
    sumR0  +=      12.985 * cos(4.74373293725  + 3377.21779200400 *T)
    sumR0  +=      11.864 * cos(0.64411806416  + 28.31117565130 *T)
    sumR0  +=      13.216 * cos(4.95904024430  + 1279.79457262800 *T)
    sumR0  +=      16.121 * cos(0.98185208328  + 2538.24850425360 *T)
    sumR0  +=      14.900 * cos(1.76649832526  + 569.04784100980 *T)
    sumR0  +=      11.337 * cos(4.36555105334  + 3583.40334044120 *T)
    sumR0  +=      11.253 * cos(5.98638731448  + 193.65537546500 *T)
    sumR0  +=      14.753 * cos(2.92291248767  + 167.08930495290 *T)
    sumR0  +=      13.774 * cos(2.50808183571  + 1802.37199072180 *T)
    sumR0  +=      11.068 * cos(0.00471764868  + 629.60234557550 *T)
    sumR0  +=      12.781 * cos(3.62178749219  + 67.66805156650 *T)
    sumR0  +=      12.238 * cos(0.27163151602  + 1044.40407666220 *T)
    sumR0  +=      11.021 * cos(0.15223056578  + 501.37978944330 *T)
    sumR0  +=      14.206 * cos(2.63254885854  + 618.55664531160 *T)
    sumR0  +=      14.365 * cos(0.37819794671  + 601.76425067620 *T)
    sumR0  +=      15.034 * cos(2.67095006272  + 46.47042291600 *T)
    sumR0  +=      12.248 * cos(2.19751851112  + 650.94298657790 *T)
    sumR0  +=      10.783 * cos(2.86375137884  + 113.38771495710 *T)
    sumR0  +=      11.418 * cos(1.20874560246  + 172.24529849340 *T)
    sumR0  +=      14.613 * cos(6.05645353059  + 468.24268865160 *T)
    sumR0  +=      10.580 * cos(2.05903854864  + 429.04587143080 *T)
    sumR0  +=      13.721 * cos(2.20936291526  + 228.27694896500 *T)
    sumR0  +=      12.180 * cos(1.82585577726  + 241.87090352020 *T)
    sumR0  +=      10.787 * cos(5.06924118186  + 162.89651925890 *T)
    sumR0  +=      12.056 * cos(3.20018724042  + 72.07328558160 *T)
    sumR0  +=      12.233 * cos(4.50741930970  + 425.63498302950 *T)
    sumR0  +=      12.101 * cos(4.14977794161  + 1108.13997496560 *T)
    sumR0  +=       9.843 * cos(1.49451039604  + 226.63241756230 *T)
    sumR0  +=      10.287 * cos(2.10680007784  + 1033.35837639830 *T)
    sumR0  +=       9.975 * cos(2.81640446254  + 518.64526483070 *T)
    sumR0  +=       9.597 * cos(4.80028087522  + 426.64637498580 *T)
    sumR0  +=      10.746 * cos(4.66838299108  + 129.91947716160 *T)
    sumR0  +=      12.961 * cos(5.11568581806  + 219.44943459230 *T)
    sumR0  +=      12.302 * cos(5.33568547700  + 776.93031047640 *T)
    sumR0  +=       9.484 * cos(4.85702954575  + 820.05928096030 *T)
    sumR0  +=      11.441 * cos(3.85769732764  + 405.99126305650 *T)
    sumR0  +=       9.625 * cos(1.60280478656  + 426.55000676620 *T)
    sumR0  +=       9.164 * cos(0.70204567980  + 403.02231763990 *T)
    sumR0  +=      10.112 * cos(2.76486875630  + 210.59078245230 *T)
    sumR0  +=      10.816 * cos(1.36864298163  + 170.76082578510 *T)
    sumR0  +=      10.187 * cos(2.36063948382  + 685.47393735270 *T)
    sumR0  +=      12.397 * cos(6.06349943525  + 875.83029900100 *T)
    sumR0  +=      12.146 * cos(2.04060386262  + 508.35032409220 *T)
    sumR0  +=       9.574 * cos(3.19555214859  + 286.59622129700 *T)
    sumR0  +=      10.193 * cos(4.01123146905  + 381.35160823740 *T)
    sumR0  +=       8.900 * cos(3.63260235880  + 319.31263096340 *T)
    sumR0  +=      10.052 * cos(5.16107251040  + 216.00740842370 *T)
    sumR0  +=       8.528 * cos(3.88076551354  + 630.33605875840 *T)
    sumR0  +=       8.875 * cos(5.46623776078  + 3370.10424500320 *T)
    sumR0  +=       8.401 * cos(5.65557131026  + 213.45915413240 *T)
    sumR0  +=      10.033 * cos(5.97497644283  + 6.15033915430 *T)
    sumR0  +=       8.401 * cos(2.27009862215  + 213.13903674360 *T)
    sumR0  +=      11.661 * cos(0.95163302252  + 694.07195706180 *T)
    sumR0  +=       8.696 * cos(2.33868966556  + 220.36445832900 *T)
    sumR0  +=       8.141 * cos(5.54059747150  + 220.46082654860 *T)
    sumR0  +=       9.615 * cos(2.75755414306  + 556.51766803760 *T)
    sumR0  +=       9.120 * cos(0.44322374149  + 2097.42321937600 *T)
    sumR0  +=       8.109 * cos(5.53989498262  + 181.05576652360 *T)
    sumR0  +=      10.763 * cos(0.05616402982  + 691.10301164520 *T)
    sumR0  +=       9.579 * cos(2.84979792871  + 184.09414790940 *T)
    sumR0  +=       9.958 * cos(2.38581008546  + 945.24345570670 *T)
    sumR0  +=       8.526 * cos(0.17821781104  + 289.56516671360 *T)
    sumR0  +=       7.700 * cos(0.00481375410  + 7.16173111060 *T)
    sumR0  +=       8.613 * cos(0.82900327241  + 2957.73314812880 *T)
    sumR0  +=       9.517 * cos(2.27516458273  + 8.07675484730 *T)
    sumR0  +=       9.049 * cos(3.37335025790  + 731.94436026870 *T)
    sumR0  +=       7.888 * cos(5.78452089815  + 230.82520325630 *T)
    sumR0  +=       7.684 * cos(3.10462250617  + 7.06536289100 *T)
    sumR0  +=       7.391 * cos(5.29648701813  + 2627.11418447060 *T)
    sumR0  +=       9.875 * cos(4.69411059509  + 10213.28554621100 *T)
    sumR0  +=       7.328 * cos(0.09051133382  + 100.64509366380 *T)
    sumR0  +=       7.999 * cos(1.60928374337  + 696.51963761660 *T)
    sumR0  +=       7.600 * cos(4.90078510977  + 51.20572533120 *T)
    sumR0  +=       9.928 * cos(5.25713005643  + 699.70103135430 *T)
    sumR0  +=       7.333 * cos(5.61982406824  + 31.49256938900 *T)
    sumR0  +=       7.376 * cos(4.52737009022  + 616.32141307790 *T)
    sumR0  +=       9.422 * cos(2.44475274779  + 2118.76386037840 *T)
    sumR0  +=       7.300 * cos(4.01885475010  + 212.02707105080 *T)
    sumR0  +=       7.502 * cos(4.71301501745  + 436.15941843160 *T)
    sumR0  +=       9.071 * cos(2.75662160229  + 130.44074202340 *T)
    sumR0  +=       8.913 * cos(2.19608557019  + 427.56139872250 *T)
    sumR0  +=       8.801 * cos(4.26655882704  + 141.22580985640 *T)
    sumR0  +=       6.853 * cos(5.69082635009  + 480.77286162380 *T)
    sumR0  +=       7.765 * cos(3.27218537808  + 3796.70243587920 *T)
    sumR0  +=       9.159 * cos(3.04713671650  + 9786.68735533500 *T)
    sumR0  +=       9.034 * cos(2.04165937353  + 204.70107572890 *T)
    sumR0  +=       6.902 * cos(4.61962635489  + 2524.02141025200 *T)
    sumR0  +=       6.728 * cos(0.58794595002  + 739.80866679490 *T)
    sumR0  +=       6.741 * cos(0.52362906624  + 135.33610313300 *T)
    sumR0  +=       8.201 * cos(5.03994203224  + 411.62033734900 *T)
    sumR0  +=       7.240 * cos(3.90653111215  + 214.57111982520 *T)
    sumR0  +=       6.887 * cos(4.11954799957  + 662.53120356300 *T)
    sumR0  +=       6.566 * cos(2.67659365854  + 194.17664032680 *T)
    sumR0  +=       6.539 * cos(6.25585361704  + 31.01948863700 *T)
    sumR0  +=       7.484 * cos(5.56871021201  + 271.40591944890 *T)
    sumR0  +=       8.078 * cos(3.09955817560  + 353.30106501700 *T)
    sumR0  +=       7.072 * cos(1.10066698352  + 282.66406803390 *T)
    sumR0  +=       6.089 * cos(0.79684364835  + 593.42686339800 *T)
    sumR0  +=       6.140 * cos(3.79672343724  + 180.16199464630 *T)
    sumR0  +=       6.703 * cos(3.82854248620  + 412.58354519550 *T)
    sumR0  +=       6.024 * cos(5.46288776207  + 724.83081326790 *T)
    sumR0  +=       8.102 * cos(4.51051495778  + 268.43697403230 *T)
    sumR0  +=       6.033 * cos(1.24326252021  + 447.93883187840 *T)
    sumR0  +=       7.425 * cos(2.29394888999  + 532.61172640140 *T)
    sumR0  +=       6.007 * cos(2.87215425398  + 426.07692601420 *T)
    sumR0  +=       7.568 * cos(0.79147591036  + 2854.64037391020 *T)
    sumR0  +=       5.816 * cos(1.70824982811  + 50.40257617910 *T)
    sumR0  +=       7.534 * cos(5.38598292680  + 953.10776223290 *T)
    sumR0  +=       5.863 * cos(2.04201456623  + 454.90936652730 *T)
    sumR0  +=       7.291 * cos(0.88044346877  + 457.61767951300 *T)
    sumR0  +=       6.235 * cos(4.51960341418  + 3693.60966166060 *T)
    sumR0  +=       6.359 * cos(6.27053660532  + 313.94418910180 *T)
    sumR0  +=       5.722 * cos(0.47241118592  + 610.69233878540 *T)
    sumR0  +=       5.704 * cos(0.45599464136  + 643.07868005170 *T)
    sumR0  +=       6.176 * cos(3.98739420856  + 835.03713448730 *T)
    sumR0  +=       5.730 * cos(0.50814242490  + 1038.04128918680 *T)
    sumR0  +=       6.812 * cos(4.20463385690  + 938.12990870590 *T)
    sumR0  +=       5.620 * cos(4.08049141112  + 3899.79521009780 *T)
    sumR0  +=       6.770 * cos(4.22172125738  + 916.93228005540 *T)
    sumR0  +=       6.069 * cos(3.46762401122  + 278.25883401880 *T)
    sumR0  +=       5.875 * cos(5.51773010551  + 1073.60902419080 *T)
    sumR0  +=       5.558 * cos(3.29478679376  + 20.60692781950 *T)
    sumR0  +=       6.274 * cos(4.88767368263  + 0.52126486180 *T)
    sumR0  +=       5.794 * cos(4.09991767938  + 391.17346822390 *T)
    sumR0  +=       5.442 * cos(2.79802608247  + 397.39324334740 *T)
    sumR0  +=       5.754 * cos(4.42718264879  + 165.60483224460 *T)
    sumR0  +=       5.879 * cos(6.17871525366  + 291.26208774300 *T)
    sumR0  +=       6.716 * cos(2.18663847730  + 627.36711334180 *T)
    sumR0  +=       5.761 * cos(0.62536160332  + 114.39910691340 *T)
    sumR0  +=       5.359 * cos(2.29390692216  + 331.20966448920 *T)
    sumR0  +=       6.210 * cos(4.92273259045  + 450.97721326420 *T)
    sumR0  +=       6.686 * cos(2.13438181268  + 285.63301345050 *T)
    sumR0  +=       5.173 * cos(4.10128119721  + 6.59228213900 *T)
    sumR0  +=       5.707 * cos(3.48716972669  + 230.56457082540 *T)
    sumR0  +=       6.363 * cos(5.64626069194  + 518.38463239980 *T)
    sumR0  +=       5.241 * cos(3.78081098206  + 418.52143602870 *T)
    sumR0  +=       5.191 * cos(4.39595146262  + 84.34282612290 *T)
    sumR0  +=       6.710 * cos(2.96748528229  + 624.91943278700 *T)
    sumR0  +=       4.931 * cos(2.71959451867  + 558.00214074590 *T)
    sumR0  +=       5.225 * cos(4.65463431385  + 310.71461125430 *T)
    sumR0  +=       4.857 * cos(2.61373582429  + 66.70484372000 *T)
    sumR0  +=       4.847 * cos(3.76991373317  + 423.67742956920 *T)
    sumR0  +=       5.284 * cos(1.96024672163  + 1182.92157353290 *T)
    sumR0  +=       5.933 * cos(2.74003948393  + 219.89137757700 *T)
    sumR0  +=       6.373 * cos(1.41296346460  + 606.76018552230 *T)
    sumR0  +=       4.728 * cos(0.23421038001  + 1063.31408345230 *T)
    sumR0  +=       6.408 * cos(1.16687419680  + 268.95823889410 *T)
    sumR0  +=       4.782 * cos(1.56813683227  + 420.96911658350 *T)
    sumR0  +=       5.399 * cos(4.57611449409  + 238.90195810360 *T)
    sumR0  +=       5.161 * cos(0.14436456585  + 2413.81508903260 *T)
    sumR0  +=       4.858 * cos(5.21378840436  + 3686.49611465980 *T)
    sumR0  +=       5.086 * cos(1.73392381835  + 337.73251065900 *T)
    sumR0  +=       4.650 * cos(3.71029121290  + 305.08553696180 *T)
    sumR0  +=       4.896 * cos(3.67786531840  + 240.12579838100 *T)
    sumR0  +=       5.949 * cos(0.29956165181  + 524.01370669230 *T)
    sumR0  +=       4.968 * cos(2.53258931342  + 980.66817835880 *T)
    sumR0  +=       4.944 * cos(2.16189522746  + 104.05598206510 *T)
    sumR0  +=       5.366 * cos(3.54867806985  + 107.28555991260 *T)
    sumR0  +=       4.917 * cos(0.48641512683  + 3274.12501778540 *T)
    sumR0  +=       5.617 * cos(6.27593478237  + 112.65400177420 *T)
    sumR0  +=       4.524 * cos(5.09539085552  + 103.14095832840 *T)
    sumR0  +=       5.643 * cos(1.52724336480  + 105.54045477340 *T)
    sumR0  +=       4.504 * cos(1.68251875362  + 196.03362005060 *T)
    sumR0  +=       4.897 * cos(4.90011892854  + 102.12956637210 *T)
    sumR0  +=       4.525 * cos(1.88735156553  + 103.04459010880 *T)
    sumR0  +=       4.327 * cos(1.45407229380  + 409.92341631960 *T)
    sumR0  +=       5.095 * cos(3.40640608336  + 427.11945573780 *T)
    sumR0  +=       5.782 * cos(3.55197606731  + 25874.60404613620 *T)
    sumR0  +=       4.192 * cos(0.16603430914  + 958.57677783100 *T)
    sumR0  +=       4.976 * cos(0.50639895683  + 511.53171782990 *T)
    sumR0  +=       4.167 * cos(5.94725762070  + 316.44005376640 *T)
    sumR0  +=       4.353 * cos(3.78587101731  + 1171.87587326900 *T)
    sumR0  +=       5.387 * cos(2.03693287651  + 2435.15573003500 *T)
    sumR0  +=       4.067 * cos(4.66592603130  + 106.01353552540 *T)
    sumR0  +=       4.817 * cos(3.53529781673  + 960.22130923370 *T)
    sumR0  +=       4.048 * cos(3.20024146722  + 775.23338944700 *T)
    sumR0  +=       4.016 * cos(6.00569143107  + 945.99421523210 *T)
    sumR0  +=       3.989 * cos(3.15130319196  + 115.62294719080 *T)
    sumR0  +=       4.559 * cos(5.59555355771  + 778.41478318470 *T)
    sumR0  +=       4.153 * cos(2.75042736587  + 316.34368554680 *T)
    sumR0  +=       3.983 * cos(2.00842137744  + 597.35901666110 *T)
    sumR0  +=       4.212 * cos(4.16852690218  + 823.99143422340 *T)
    sumR0  +=       5.193 * cos(0.71717111984  + 810.65811209910 *T)
    sumR0  +=       3.927 * cos(5.04361736754  + 2943.50605412720 *T)
    sumR0  +=       4.225 * cos(0.02571003853  + 0.75075952540 *T)
    sumR0  +=       4.926 * cos(1.12994881124  + 526.98265210890 *T)
    sumR0  +=       4.170 * cos(3.94116290117  + 422.40540518200 *T)
    sumR0  +=       4.432 * cos(3.99599046800  + 393.46109008430 *T)
    sumR0  +=       3.840 * cos(1.21234108241  + 212.07525516060 *T)
    sumR0  +=       3.866 * cos(4.20930793423  + 97.67614824720 *T)
    sumR0  +=       4.440 * cos(1.35536679738  + 211.60217440860 *T)
    sumR0  +=       3.840 * cos(0.43014354282  + 214.52293571540 *T)
    sumR0  +=       4.724 * cos(3.62039208608  + 638.41281360570 *T)
    sumR0  +=       4.870 * cos(5.75874599620  + 1246.65747183630 *T)
    sumR0  +=       4.449 * cos(1.43384065964  + 184.98791978670 *T)
    sumR0  +=       3.931 * cos(2.36660272585  + 909.81873305460 *T)
    sumR0  +=       3.787 * cos(5.98932416906  + 325.95309721220 *T)
    sumR0  +=       3.665 * cos(0.72917314141  + 20.44686912510 *T)
    sumR0  +=       4.243 * cos(2.29103096797  + 453.42489381900 *T)
    sumR0  +=       3.730 * cos(6.24831601183  + 159.12442469020 *T)
    sumR0  +=       3.900 * cos(1.31013240315  + 850.01498801430 *T)
    sumR0  +=       5.134 * cos(1.97348901289  + 526.50957135690 *T)
    sumR0  +=       3.621 * cos(2.77435773661  + 123.53964334370 *T)
    sumR0  +=       3.607 * cos(5.32058842710  + 406.95447090300 *T)
    sumR0  +=       3.802 * cos(1.94444523548  + 421.18156490460 *T)
    sumR0  +=       3.906 * cos(3.02475451573  + 317.35507750310 *T)
    sumR0  +=       4.473 * cos(0.80804073855  + 838.21852822500 *T)
    sumR0  +=       4.081 * cos(2.08732468180  + 988.53248488500 *T)
    sumR0  +=       3.843 * cos(4.99347148246  + 7.63481186260 *T)
    sumR0  +=       4.467 * cos(6.09037793116  + 760.25553592000 *T)
    sumR0  +=       3.514 * cos(3.97285766412  + 426.48631629140 *T)
    sumR0  +=       3.504 * cos(0.85064201666  + 299.12639426920 *T)
    sumR0  +=       4.397 * cos(1.68577228317  + 824.74219374880 *T)
    sumR0  +=       3.581 * cos(2.35235960566  + 337.80194662820 *T)
    sumR0  +=       4.606 * cos(3.48411642192  + 913.96333463880 *T)
    sumR0  +=       3.790 * cos(3.64538213705  + 216.26804085460 *T)
    sumR0  +=       3.496 * cos(0.95035381131  + 436.89313161450 *T)
    sumR0  +=       4.422 * cos(0.82822191292  + 43.28902917830 *T)
    sumR0  +=       3.688 * cos(2.83785443800  + 739.05790726950 *T)
    sumR0  +=       4.439 * cos(1.19409419107  + 421.93232443000 *T)
    sumR0  +=       3.572 * cos(2.77298538478  + 444.75743814070 *T)
    sumR0  +=       4.420 * cos(5.44308967028  + 963.40270297140 *T)
    sumR0  +=       4.443 * cos(3.73070830296  + 37.87240320690 *T)
    sumR0  +=       4.322 * cos(4.75680702521  + 40.84134862350 *T)
    sumR0  +=       3.724 * cos(0.59005210557  + 256.42806592190 *T)
    sumR0  +=       4.471 * cos(2.22367643527  + 318.83955021140 *T)
    sumR0  +=       4.184 * cos(1.52719196640  + 298.23262239190 *T)
    sumR0  +=       3.534 * cos(5.01937599570  + 386.98068252990 *T)
    sumR0  +=       3.400 * cos(3.22663067085  + 4113.09430553580 *T)
    sumR0  +=       4.556 * cos(1.35715974815  + 495.75071515080 *T)
    sumR0  +=       4.453 * cos(1.80417064247  + 829.62050851590 *T)
    sumR0  +=       3.617 * cos(1.51036385224  + 41.64449777560 *T)
    sumR0  +=       3.789 * cos(4.80357656146  + 238.42887735160 *T)
    sumR0  +=       3.514 * cos(2.38272766645  + 426.71006546060 *T)
    sumR0  +=       3.208 * cos(1.74465274123  + 952.35700270750 *T)
    sumR0  +=       4.398 * cos(2.65839000906  + 832.58945393250 *T)
    sumR0  +=       4.092 * cos(3.07954777295  + 60.76695288680 *T)
    sumR0  +=       3.813 * cos(5.63047104819  + 315.42866181010 *T)
    sumR0  +=       3.219 * cos(6.22278803635  + 754.83890994860 *T)
    sumR0  +=       3.102 * cos(2.69222024257  + 343.73983746140 *T)
    sumR0  +=       4.045 * cos(4.02463772100  + 376.19561469690 *T)
    sumR0  +=       3.576 * cos(0.38532787280  + 214.99601646740 *T)
    sumR0  +=       3.291 * cos(5.49542015261  + 143.93412284210 *T)
    sumR0  +=       3.981 * cos(5.75449411958  + 239.16259053450 *T)
    sumR0  +=       3.249 * cos(0.58789568678  + 619.29035849450 *T)
    sumR0  +=       3.114 * cos(0.02831060137  + 221.16340196420 *T)
    sumR0  +=       3.555 * cos(3.12207684735  + 1048.33622992530 *T)
    sumR0  +=       3.010 * cos(1.91180343491  + 93.53154666300 *T)
    sumR0  +=       3.384 * cos(1.81702854004  + 443.86366626340 *T)
    sumR0  +=       3.222 * cos(2.37342117781  + 429.51895218280 *T)
    sumR0  +=       3.431 * cos(5.92099840679  + 570.74476203920 *T)
    sumR0  +=       3.271 * cos(2.04947945059  + 806.72595883600 *T)
    sumR0  +=       3.207 * cos(5.44018976766  + 402.21916848780 *T)
    sumR0  +=       3.091 * cos(0.10717557454  + 3590.51688744200 *T)
    sumR0  +=       2.951 * cos(1.76365810296  + 426.81063919710 *T)
    sumR0  +=       2.958 * cos(0.23653889192  + 1354.43315884340 *T)
    sumR0  +=       2.948 * cos(4.59289832104  + 426.38574255490 *T)
    sumR0  +=       3.506 * cos(5.43222584214  + 84.93352695390 *T)
    sumR0  +=       2.894 * cos(5.69678330542  + 1262.38608488870 *T)
    sumR0  +=       3.682 * cos(1.07122313007  + 395.57870223900 *T)
    sumR0  +=       2.983 * cos(5.25093816048  + 313.47110834980 *T)
    sumR0  +=       3.117 * cos(4.18767239237  + 366.79444583570 *T)
    sumR0  +=       2.873 * cos(4.45472312727  + 361.37781986430 *T)
    sumR0  +=       3.521 * cos(2.05528981993  + 1261.63532536330 *T)
    sumR0  +=       3.496 * cos(1.87950759078  + 439.12836384820 *T)
    sumR0  +=       3.012 * cos(0.64439385874  + 263.02034806090 *T)
    sumR0  +=       2.849 * cos(1.12491777974  + 262.05714021440 *T)
    sumR0  +=       2.910 * cos(2.76192171681  + 541.53981451060 *T)
    sumR0  +=       3.322 * cos(6.08893948791  + 108.72184851110 *T)
    sumR0  +=       3.181 * cos(1.01419299056  + 418.00017116690 *T)
    sumR0  +=       2.793 * cos(0.38781777981  + 211.86280683950 *T)
    sumR0  +=       3.091 * cos(2.18216748751  + 306.83064210100 *T)
    sumR0  +=       3.748 * cos(3.89145855821  + 220.93390730060 *T)
    sumR0  +=       2.982 * cos(1.99831689446  + 117.91056905120 *T)
    sumR0  +=       2.793 * cos(1.25466684542  + 214.73538403650 *T)
    sumR0  +=       3.512 * cos(1.50965040301  + 885.43971066640 *T)
    sumR0  +=       2.716 * cos(0.80710391613  + 757.21715453420 *T)
    sumR0  +=       3.137 * cos(2.09889265033  + 2751.54759969160 *T)
    sumR0  +=       2.738 * cos(4.89270330923  + 464.73122651380 *T)
    sumR0  +=       2.875 * cos(4.28436709414  + 4010.00153131720 *T)
    sumR0  +=       3.313 * cos(3.01452486457  + 336.83873878170 *T)
    sumR0  +=       2.746 * cos(2.69963506928  + 380.38840039090 *T)
    sumR0  +=       3.132 * cos(2.19562786872  + 2.96894541660 *T)
    sumR0  +=       3.233 * cos(0.60809684558  + 3171.03224356680 *T)
    sumR0  +=       3.034 * cos(0.93246285284  + 205.43478891180 *T)
    sumR0  +=       3.006 * cos(5.91067479448  + 2.70831298570 *T)
    sumR0  +=       3.643 * cos(5.58302397259  + 423.62924545940 *T)
    sumR0  +=       2.625 * cos(1.07042050691  + 23.57587323610 *T)
    sumR0  +=       3.495 * cos(0.19887562030  + 576.16138801060 *T)
    sumR0  +=       2.590 * cos(0.21252773750  + 110.25450532920 *T)
    sumR0  +=       2.704 * cos(6.12908233599  + 572.22923474750 *T)
    sumR0  +=       2.568 * cos(0.17571588314  + 1056.20053645150 *T)
    sumR0  +=       2.583 * cos(2.96927378731  + 384.05992122310 *T)
    sumR0  +=       2.555 * cos(3.96441052072  + 430.79097657000 *T)
    sumR0  +=       2.786 * cos(2.54945911818  + 195.89060769870 *T)
    sumR0  +=       2.869 * cos(4.82964665921  + 710.74673161820 *T)
    sumR0  +=       2.534 * cos(5.31005598763  + 427.34895040140 *T)
    sumR0  +=       2.618 * cos(1.22081401503  + 36.64856292950 *T)
    sumR0  +=       3.464 * cos(2.32811328200  + 285.37238101960 *T)
    sumR0  +=       3.374 * cos(3.34109586766  + 162.09337010680 *T)
    sumR0  +=       2.694 * cos(4.64149271687  + 140.96517742550 *T)
    sumR0  +=       2.603 * cos(0.99527295832  + 92.30770638560 *T)
    sumR0  +=       3.140 * cos(5.40790277580  + 328.24071907260 *T)
    sumR0  +=       2.603 * cos(4.31532790880  + 561.93429400900 *T)
    sumR0  +=       2.987 * cos(0.82758128867  + 45.57665103870 *T)
    sumR0  +=       2.959 * cos(3.35623851523  + 273.85360000370 *T)
    sumR0  +=       2.561 * cos(0.54160683162  + 107.75864066460 *T)
    sumR0  +=       2.900 * cos(1.20691455948  + 462.02291352810 *T)
    sumR0  +=       2.648 * cos(1.90547819027  + 88.11492069160 *T)
    sumR0  +=       2.461 * cos(4.68211868869  + 2840.41327990860 *T)
    sumR0  +=       2.534 * cos(5.00311256556  + 431.26405732200 *T)
    sumR0  +=       2.521 * cos(3.32160472310  + 136.06981631590 *T)
    sumR0  +=       2.945 * cos(1.06452531856  + 732.69511979410 *T)
    sumR0  +=       2.654 * cos(1.36744710395  + 460.53844081980 *T)
    sumR0  +=       3.297 * cos(1.33975572602  + 305.60680182360 *T)
    sumR0  +=       2.634 * cos(2.29995533800  + 519.39602435610 *T)
    sumR0  +=       2.805 * cos(5.62255444533  + 1699.27921650320 *T)
    sumR0  +=       2.439 * cos(5.14733660159  + 303.86169668440 *T)
    sumR0  +=       2.434 * cos(3.71460437051  + 4216.18707975440 *T)
    sumR0  +=       2.416 * cos(3.76296045457  + 77.75054398390 *T)
    sumR0  +=       2.803 * cos(2.55280894914  + 505.31194270640 *T)
    sumR0  +=       2.592 * cos(3.32836551071  + 110.15813710960 *T)
    sumR0  +=       3.074 * cos(1.71462387764  + 256.58812461630 *T)
    sumR0  +=       3.295 * cos(0.81766682522  + 705.11765732570 *T)
    sumR0  +=       3.183 * cos(6.15742006608  + 109.24311337290 *T)
    sumR0  +=       2.908 * cos(5.38534195293  + 315.16802937920 *T)
    sumR0  +=       2.326 * cos(1.42604031905  + 131.54696222180 *T)
    sumR0  +=       2.427 * cos(2.04627850740  + 124.50285119020 *T)
    sumR0  +=       2.632 * cos(1.41253794767  + 211.65456403530 *T)
    sumR0  +=       2.297 * cos(1.38016674676  + 425.84743135060 *T)
    sumR0  +=       2.318 * cos(6.27716072818  + 317.14262918200 *T)
    sumR0  +=       2.310 * cos(4.86442292404  + 3259.89792378380 *T)
    sumR0  +=       2.873 * cos(1.10206537875  + 7.86430652620 *T)
    sumR0  +=       2.616 * cos(0.11849676899  + 133.10087089930 *T)
    sumR0  +=       3.213 * cos(4.49320461690  + 432.22726516850 *T)
    sumR0  +=       2.276 * cos(6.04688191978  + 214.10224459010 *T)
    sumR0  +=       2.276 * cos(1.96882478275  + 212.49594628590 *T)
    sumR0  +=       2.917 * cos(0.96774661857  + 100.38446123290 *T)
    sumR0  +=       2.890 * cos(5.72610904534  + 322.02094394910 *T)
    sumR0  +=       2.829 * cos(2.66887892162  + 141.48644228730 *T)
    sumR0  +=       2.695 * cos(1.47488571070  + 42.53826965290 *T)
    sumR0  +=       2.697 * cos(5.34002228297  + 432.01481684740 *T)
    sumR0  +=       2.229 * cos(4.40717937246  + 540.73666535850 *T)
    sumR0  +=       2.214 * cos(2.43714413196  + 426.85882330690 *T)
    sumR0  +=       2.512 * cos(4.68291916658  + 1596.18644228460 *T)
    sumR0  +=       2.202 * cos(5.91122030007  + 867.42347575360 *T)
    sumR0  +=       2.856 * cos(0.94736445171  + 41.05379694460 *T)
    sumR0  +=       2.396 * cos(0.10516628717  + 206.93630796260 *T)
    sumR0  +=       2.345 * cos(1.16685267780  + 640.86049416050 *T)
    sumR0  +=       2.263 * cos(4.62327588198  + 188.02630117250 *T)
    sumR0  +=       2.214 * cos(3.97566024178  + 426.33755844510 *T)
    sumR0  +=       2.274 * cos(4.94055830720  + 4002.88798431640 *T)
    sumR0  +=       2.541 * cos(0.83705615200  + 12352.85260454480 *T)
    sumR0  +=       2.145 * cos(3.40120044084  + 111.16952906590 *T)
    sumR0  +=       2.895 * cos(6.07389082608  + 2914.01423582380 *T)
    sumR0  +=       2.520 * cos(1.05396310009  + 184.72728735580 *T)
    sumR0  +=       2.448 * cos(3.49820841117  + 481.73606947030 *T)
    sumR0  +=       2.343 * cos(2.86472924644  + 207.07932031450 *T)
    sumR0  +=       2.964 * cos(5.96264929181  + 465.95506679120 *T)
    sumR0  +=       2.122 * cos(4.04560058177  + 118.07062774560 *T)
    sumR0  +=       2.452 * cos(4.35251210402  + 1382.88734684660 *T)
    sumR0  +=       2.131 * cos(0.61211416273  + 335.14181775230 *T)
    sumR0  +=       2.116 * cos(4.76181734842  + 765.88461021250 *T)
    sumR0  +=       2.245 * cos(5.67078632283  + 6467.92575796160 *T)
    sumR0  +=       2.425 * cos(0.84789834075  + 550.13783421970 *T)
    sumR0  +=       2.227 * cos(1.15684015463  + 227.31374111850 *T)
    sumR0  +=       2.314 * cos(6.13104971819  + 2730.20695868920 *T)
    sumR0  +=       2.344 * cos(0.35844885568  + 217.44369702220 *T)
    sumR0  +=       2.529 * cos(3.07975959821  + 774.48262992160 *T)
    sumR0  +=       2.523 * cos(1.75026771081  + 1578.02719501990 *T)
    sumR0  +=       2.111 * cos(4.24637589094  + 96.87299909510 *T)
    sumR0  +=       2.826 * cos(1.01974994073  + 87.31177153950 *T)
    sumR0  +=       2.906 * cos(3.75374302356  + 428.08266358430 *T)
    sumR0  +=       2.113 * cos(2.19787343926  + 449.23210812500 *T)
    sumR0  +=       2.142 * cos(1.19671359858  + 209.15449385380 *T)
    sumR0  +=       2.882 * cos(2.59371585952  + 39.61750834610 *T)
    sumR0  +=       2.078 * cos(2.87503604503  + 14.01464568050 *T)
    sumR0  +=       2.090 * cos(1.99032225653  + 441.57604440300 *T)
    sumR0  +=       2.519 * cos(2.99001165551  + 745.27768239300 *T)
    sumR0  +=       2.035 * cos(5.37147785849  + 1041.22268292450 *T)
    sumR0  +=       2.054 * cos(1.11817372961  + 842.90144101350 *T)
    sumR0  +=       2.023 * cos(2.94559148702  + 668.20846196530 *T)
    sumR0  +=       2.529 * cos(4.34280159004  + 221.89711514710 *T)
    sumR0  +=       2.554 * cos(5.56906955622  + 214.19286731530 *T)
    sumR0  +=       2.712 * cos(1.60469055827  + 1050.99635880120 *T)
    sumR0  +=       2.350 * cos(3.36541706919  + 220.20019411770 *T)
    sumR0  +=       2.015 * cos(2.60446576036  + 315.64111013120 *T)
    sumR0  +=       2.158 * cos(1.64945261993  + 219.66188291340 *T)
    sumR0  +=       2.120 * cos(2.93968038721  + 304.12232911530 *T)
    sumR0  +=       2.357 * cos(1.94433441808  + 233.90602325750 *T)
    sumR0  +=       2.579 * cos(4.54124062411  + 484.70501488690 *T)
    sumR0  +=       2.046 * cos(5.45531068264  + 200.55647414470 *T)
    sumR0  +=       2.040 * cos(2.04492641594  + 1097.09427470170 *T)
    sumR0  +=       2.675 * cos(1.20234167733  + 28.57180808220 *T)
    sumR0  +=       2.528 * cos(4.69268465973  + 637.44960575920 *T)
    sumR0  +=       2.314 * cos(2.81193072994  + 25.12978191360 *T)
    sumR0  +=       2.689 * cos(5.03868493349  + 1269.49963188950 *T)
    sumR0  +=       2.115 * cos(3.10772296248  + 1276.61317889030 *T)
    sumR0  +=       2.712 * cos(1.49710379127  + 3340.61242669980 *T)
    sumR0  +=       2.138 * cos(4.52114042624  + 378.90392768260 *T)
    sumR0  +=       2.708 * cos(0.01014338204  + 389.94962794650 *T)
    sumR0  +=       2.560 * cos(5.75783882561  + 544.50875992720 *T)
    sumR0  +=       2.028 * cos(0.24331359951  + 146.59425171800 *T)
    sumR0  +=       2.096 * cos(1.44475430956  + 864.24208201590 *T)
    sumR0  +=       1.897 * cos(4.63194412388  + 220.30076785420 *T)
    sumR0  +=       1.901 * cos(1.78319572727  + 192.85222631290 *T)
    sumR0  +=       2.011 * cos(4.11578325523  + 315.87060479480 *T)
    sumR0  +=       2.014 * cos(2.26726355818  + 198.10879358990 *T)
    sumR0  +=       1.905 * cos(2.15527255015  + 326.68681039510 *T)
    sumR0  +=       1.949 * cos(3.95440811214  + 103.61403908040 *T)
    sumR0  +=       2.098 * cos(5.24613314798  + 175.42669223110 *T)
    sumR0  +=       1.884 * cos(5.66018186202  + 1310.39337013970 *T)
    sumR0  +=       1.911 * cos(2.60665446600  + 301.41401612960 *T)
    sumR0  +=       2.160 * cos(3.42888079793  + 420.00590873700 *T)
    sumR0  +=       2.325 * cos(5.89632178100  + 815.06334611420 *T)
    sumR0  +=       1.914 * cos(0.22451332248  + 171.65459766240 *T)
    sumR0  +=       1.854 * cos(0.04278915026  + 233.74596456310 *T)
    sumR0  +=       1.973 * cos(2.68640259480  + 769.81676347560 *T)
    sumR0  +=       1.941 * cos(0.61369890353  + 3487.42411322340 *T)
    sumR0  +=       1.836 * cos(3.41496980986  + 195.77298761970 *T)
    sumR0  +=       2.554 * cos(2.35660179716  + 212.40532356070 *T)
    sumR0  +=       1.822 * cos(4.05510232882  + 639.94547042380 *T)
    sumR0  +=       1.883 * cos(6.27079329518  + 16.67477455640 *T)
    sumR0  +=       1.865 * cos(0.17460226411  + 244.79166482700 *T)
    sumR0  +=       2.097 * cos(4.58369520569  + 316.91313451840 *T)
    sumR0  +=       1.879 * cos(2.76480572708  + 28.45418800320 *T)
    sumR0  +=       2.111 * cos(2.92457831824  + 328.92204262880 *T)
    sumR0  +=       2.077 * cos(0.35943033580  + 589.49471013490 *T)
    sumR0  +=       1.825 * cos(4.04945265223  + 190.40454575810 *T)
    sumR0  +=       1.895 * cos(3.89414880651  + 334.55111692130 *T)
    sumR0  +=       2.425 * cos(3.76754213762  + 25558.21217647960 *T)
    sumR0  +=       2.218 * cos(1.85341154236  + 635.23141986800 *T)
    sumR0  +=       1.782 * cos(0.86927461254  + 92.79783348010 *T)
    sumR0  +=       2.432 * cos(3.78026263567  + 1254.52177836250 *T)
    sumR0  +=       2.106 * cos(0.17285594964  + 354.26427286350 *T)
    sumR0  +=       1.791 * cos(6.23892012939  + 1670.82502850000 *T)
    sumR0  +=       1.996 * cos(1.40940081042  + 230.70758317730 *T)
    sumR0  +=       1.753 * cos(1.86660297451  + 241.75328344120 *T)
    sumR0  +=       2.093 * cos(2.39366777880  + 187.43560034150 *T)
    sumR0  +=       2.007 * cos(3.54083120293  + 226.79247625670 *T)
    sumR0  +=       1.949 * cos(1.36533052698  + 1385.17496870700 *T)
    sumR0  +=       1.737 * cos(2.67583588366  + 6.36278747540 *T)
    sumR0  +=       1.868 * cos(4.24454204649  + 1119.18567522950 *T)
    sumR0  +=       1.896 * cos(3.81474515719  + 310.97524368520 *T)
    sumR0  +=       1.860 * cos(3.67887919389  + 1321.43907040360 *T)
    sumR0  +=       2.305 * cos(3.53252557028  + 1570.91364801910 *T)
    sumR0  +=       2.008 * cos(3.88504783546  + 638.93407846750 *T)
    sumR0  +=       1.928 * cos(2.64393870433  + 525.75881183150 *T)
    sumR0  +=       1.807 * cos(0.76057354967  + 66.18357885820 *T)
    sumR0  +=       1.824 * cos(0.85822155861  + 639.84910220420 *T)
    sumR0  +=       2.221 * cos(4.82210413830  + 1585.89150154610 *T)
    sumR0  +=       2.227 * cos(4.58488941022  + 271.61836777000 *T)
    sumR0  +=       1.897 * cos(0.64334758250  + 55.65914345610 *T)
    sumR0  +=       1.792 * cos(0.00514800434  + 827.92358748650 *T)
    sumR0  +=       1.831 * cos(3.69768852728  + 172.45774681450 *T)
    sumR0  +=       1.790 * cos(4.80062977720  + 3576.28979344040 *T)
    sumR0  +=       1.698 * cos(0.72475212282  + 295.19424100610 *T)
    sumR0  +=       1.744 * cos(3.45256183290  + 238.57188970350 *T)
    sumR0  +=       2.334 * cos(1.51845210531  + 170.01006625970 *T)
    sumR0  +=       1.669 * cos(4.44620549295  + 4319.27985397300 *T)
    sumR0  +=       1.939 * cos(3.43927826945  + 102.57150935680 *T)
    sumR0  +=       1.733 * cos(5.96815907422  + 837.69726336320 *T)
    sumR0  +=       1.686 * cos(2.12870436615  + 491.55792945680 *T)
    sumR0  +=       1.651 * cos(3.20586624475  + 281.17959532560 *T)
    sumR0  +=       1.884 * cos(4.24447812450  + 13.49338081870 *T)
    sumR0  +=       1.880 * cos(0.33845094634  + 214.94362684070 *T)
    sumR0  +=       1.817 * cos(3.07678560214  + 220.52451702340 *T)
    sumR0  +=       1.872 * cos(3.79328892492  + 392.65794093220 *T)
    sumR0  +=       2.195 * cos(1.93786776664  + 259.76951835400 *T)
    sumR0  +=       1.881 * cos(4.49314034712  + 199.80571461930 *T)
    sumR0  +=       1.662 * cos(0.29659841675  + 314.90739694830 *T)
    sumR0  +=       1.626 * cos(2.29697402942  + 3067.93946934820 *T)
    sumR0  +=       1.927 * cos(1.00706624515  + 26.82670294300 *T)
    sumR0  +=       1.621 * cos(0.01495920679  + 1379.70595310890 *T)
    sumR0  +=       1.655 * cos(4.15494230496  + 4326.39340097380 *T)
    sumR0  +=       1.719 * cos(5.97443860990  + 152.53214255120 *T)
    sumR0  +=       1.653 * cos(5.65633302949  + 448.68959140380 *T)
    sumR0  +=       1.790 * cos(3.73709604091  + 10.03430830760 *T)
    sumR0  +=       1.600 * cos(2.28430251221  + 749.20983565610 *T)
    sumR0  +=       1.770 * cos(0.03256515992  + 364.34676528090 *T)
    sumR0  +=       1.939 * cos(5.93191442558  + 249.94765836750 *T)
    sumR0  +=       1.988 * cos(4.78810872536  + 101.86893394120 *T)
    sumR0  +=       1.667 * cos(5.52149899431  + 229.97386999440 *T)
    sumR0  +=       2.215 * cos(3.55121116323  + 594.65070367540 *T)
    sumR0  +=       1.950 * cos(0.80878923373  + 1049.08698945070 *T)
    sumR0  +=       1.773 * cos(2.55608046714  + 9985.75935677140 *T)
    sumR0  +=       2.088 * cos(2.33165208016  + 420.44785172170 *T)
    sumR0  +=       2.107 * cos(2.43624356568  + 453.68552624990 *T)
    sumR0  +=       1.907 * cos(4.72569972805  + 857.12853501510 *T)
    sumR0  +=       1.609 * cos(4.96540433043  + 285.11174858870 *T)
    sumR0  +=       2.152 * cos(4.87024306306  + 186.21176006410 *T)
    sumR0  +=       1.685 * cos(5.68609178354  + 200.03520928290 *T)
    sumR0  +=       1.752 * cos(5.21532265401  + 25448.00585526019 *T)
    sumR0  +=       1.870 * cos(6.14683645342  + 347.36317418380 *T)
    sumR0  +=       1.731 * cos(1.95944272122  + 934.94851496820 *T)
    sumR0  +=       1.680 * cos(5.56246697700  + 170.97327410620 *T)
    sumR0  +=       1.652 * cos(1.24521351050  + 398.14400287280 *T)
    sumR0  +=       1.548 * cos(0.38524522125  + 17.26547538740 *T)
    sumR0  +=       1.577 * cos(0.23430114545  + 434.67494572330 *T)
    sumR0  +=       1.652 * cos(2.84480428863  + 385.54439393140 *T)
    sumR0  +=       1.770 * cos(1.15057996280  + 199.96577331370 *T)
    sumR0  +=       1.528 * cos(4.02240583348  + 236.19364511790 *T)
    sumR0  +=       2.118 * cos(5.47803429266  + 369.08206769610 *T)
    sumR0  +=       1.543 * cos(2.54353574089  + 632.73555520340 *T)
    sumR0  +=       1.504 * cos(5.94300855424  + 280.21638747910 *T)
    sumR0  +=       1.495 * cos(2.19380669867  + 407.47573576480 *T)
    sumR0  +=       1.894 * cos(1.88797605501  + 598.84348936940 *T)
    sumR0  +=       1.515 * cos(4.70072295492  + 211.29335786790 *T)
    sumR0  +=       1.767 * cos(3.11910667879  + 2921.12778282460 *T)
    sumR0  +=       1.515 * cos(3.31498374761  + 215.30483300810 *T)
    sumR0  +=       1.728 * cos(5.28095966912  + 219.51887056150 *T)
    sumR0  +=       1.906 * cos(5.24236020775  + 248.46318565920 *T)
    sumR0  +=       1.607 * cos(0.80041605735  + 642.34496686880 *T)
    sumR0  +=       1.640 * cos(2.93387205029  + 1364.72809958190 *T)
    sumR0  +=       1.585 * cos(0.77219822539  + 661.23792731640 *T)
    sumR0  +=       1.458 * cos(5.67666822477  + 632.83192342300 *T)
    sumR0  +=       1.866 * cos(4.40562835971  + 971.10695080320 *T)
    sumR0  +=       1.838 * cos(0.48492190760  + 1127.04998175570 *T)
    sumR0  +=       1.903 * cos(5.18692835916  + 2015.67108615980 *T)
    sumR0  +=       1.590 * cos(2.80043208070  + 633.74694715970 *T)
    sumR0  +=       1.489 * cos(4.10155671855  + 77837.11123384659 *T)
    sumR0  +=       1.573 * cos(4.23741356107  + 203.89792657680 *T)
    sumR0  +=       1.690 * cos(0.65475720351  + 2700.71514038580 *T)
    sumR0  +=       1.872 * cos(3.54376036064  + 354.52490529440 *T)
    sumR0  +=       1.419 * cos(4.53129346734  + 224.60542813280 *T)
    sumR0  +=       1.575 * cos(3.52476647615  + 373.90799283650 *T)
    sumR0  +=       1.817 * cos(3.95203885550  + 6076.89030155420 *T)
    sumR0  +=       1.446 * cos(5.41423319377  + 317.87634236490 *T)
    sumR0  +=       1.552 * cos(1.89030720529  + 1304.92435454160 *T)
    sumR0  +=       1.394 * cos(1.86243646383  + 913.75088631770 *T)
    sumR0  +=       1.940 * cos(4.36562864826  + 432.74853003030 *T)
    sumR0  +=       1.655 * cos(1.31748248488  + 25668.41849769900 *T)
    sumR0  +=       1.438 * cos(5.12958189872  + 71.81265315070 *T)
    sumR0  +=       1.929 * cos(4.90721606846  + 206.39799675830 *T)
    sumR0  +=       1.533 * cos(2.10965059985  + 378.64329525170 *T)
    sumR0  +=       1.421 * cos(6.22897936063  + 904.40210708320 *T)
    sumR0  +=       1.450 * cos(1.98065714692  + 205.97310011610 *T)
    sumR0  +=       1.584 * cos(5.94110940455  + 1226.21060271120 *T)
    sumR0  +=       1.510 * cos(3.77771207288  + 222.70026429920 *T)
    sumR0  +=       1.670 * cos(2.51954641624  + 976.73602509570 *T)
    sumR0  +=       1.823 * cos(1.60093991502  + 1141.13406340540 *T)
    sumR0  +=       1.564 * cos(3.94804398924  + 9566.27471289620 *T)
    sumR0  +=       1.406 * cos(3.65940442223  + 316.50374424120 *T)
    sumR0  +=       1.653 * cos(3.62394476466  + 968.13800538660 *T)
    sumR0  +=       1.495 * cos(0.77832279170  + 5959.57043333400 *T)
    sumR0  +=       1.408 * cos(5.24970924980  + 316.27999507200 *T)
    sumR0  +=       1.724 * cos(0.25483952164  + 125.18417474640 *T)
    sumR0  +=       1.426 * cos(2.26871672572  + 17.40848773930 *T)
    sumR0  +=       1.881 * cos(4.12588105003  + 562.14674233010 *T)
    sumR0  +=       1.370 * cos(4.74474866880  + 1357.61455258110 *T)
    sumR0  +=       1.485 * cos(2.18712736768  + 9889.78012955360 *T)
    sumR0  +=       1.369 * cos(1.31846306523  + 1160.83017300510 *T)
    sumR0  +=       1.394 * cos(5.66924517860  + 1736.99156101570 *T)
    sumR0  +=       1.375 * cos(0.16983286236  + 346.39996633730 *T)
    sumR0  +=       1.525 * cos(2.40244831911  + 419.43645976540 *T)
    sumR0  +=       1.359 * cos(2.68668516027  + 310.76279536410 *T)
    sumR0  +=       1.396 * cos(3.67055397940  + 253.45912050530 *T)
    sumR0  +=       1.454 * cos(5.49356262440  + 504.56118318100 *T)
    sumR0  +=       1.813 * cos(0.21882066860  + 263.70167161710 *T)
    sumR0  +=       1.330 * cos(3.01736059716  + 254.14044406150 *T)
    sumR0  +=       1.498 * cos(0.17578085903  + 155.78297225810 *T)
    sumR0  +=       1.509 * cos(5.13129901759  + 768.85355562910 *T)
    sumR0  +=       1.550 * cos(1.44225397189  + 1894.41906467650 *T)
    sumR0  +=       1.447 * cos(1.44933809994  + 893.35640681930 *T)
    sumR0  +=       1.306 * cos(5.76425101758  + 714.67888488130 *T)
    sumR0  +=       1.660 * cos(1.34160526151  + 322.61164478010 *T)
    sumR0  +=       1.347 * cos(3.00388920953  + 843.63515419640 *T)
    sumR0  +=       1.432 * cos(3.43786149731  + 251.17149864490 *T)
    sumR0  +=       1.510 * cos(4.39762427873  + 25.27279426550 *T)
    sumR0  +=       1.587 * cos(5.17106904014  + 141.69889060840 *T)
    sumR0  +=       1.477 * cos(5.47518377610  + 226.04171673130 *T)
    sumR0  +=       1.356 * cos(4.24406292182  + 332.17287233570 *T)
    sumR0  +=       1.644 * cos(1.74367211793  + 67.88049988760 *T)
    sumR0  +=       1.407 * cos(5.79229630947  + 188.16931352440 *T)
    sumR0  +=       1.575 * cos(0.09808372057  + 702.14871190910 *T)
    sumR0  +=       1.765 * cos(4.93410889383  + 201.51968199120 *T)
    sumR0  +=       1.318 * cos(4.91605557404  + 17.52610781830 *T)
    sumR0  +=       1.701 * cos(4.47360878108  + 384.27236954420 *T)
    sumR0  +=       1.304 * cos(6.04155032791  + 25.86349509650 *T)
    sumR0  +=       1.269 * cos(4.91035989349  + 354.99798604640 *T)
    sumR0  +=       1.620 * cos(5.54960841244  + 260.36021918500 *T)
    sumR0  +=       1.263 * cos(1.16521999431  + 255.83736509090 *T)
    sumR0  +=       1.744 * cos(3.70453251764  + 147.11551657980 *T)
    sumR0  +=       1.579 * cos(4.31561365365  + 2228.97018159780 *T)
    sumR0  +=       1.557 * cos(0.57740217353  + 3178.14579056760 *T)
    sumR0  +=       1.302 * cos(0.32055726013  + 119.76754877500 *T)
    sumR0  +=       1.405 * cos(3.20407486040  + 395.10562148700 *T)
    sumR0  +=       1.234 * cos(2.34766954239  + 318.67949151700 *T)
    sumR0  +=       1.519 * cos(4.74629629688  + 100.17201291180 *T)
    sumR0  +=       1.573 * cos(4.00132484524  + 1264.29545423920 *T)
    sumR0  +=       1.222 * cos(0.10709243166  + 1372.59240610810 *T)
    sumR0  +=       1.205 * cos(0.24105241435  + 466.75821594330 *T)
    sumR0  +=       1.188 * cos(1.27112537278  + 1184.40604624120 *T)
    sumR0  +=       1.280 * cos(4.85454052139  + 535.91074021810 *T)
    sumR0  +=       1.482 * cos(4.47166692910  + 763.43692965770 *T)
    sumR0  +=       1.233 * cos(1.64803509193  + 433.66355376700 *T)
    sumR0  +=       1.476 * cos(3.76634399110  + 272.58157561650 *T)
    sumR0  +=       1.199 * cos(1.78020373551  + 102.34201469320 *T)
    sumR0  +=       1.186 * cos(4.72162748523  + 795.68025857210 *T)
    sumR0  +=       1.291 * cos(4.42697938285  + 10220.39909321180 *T)
    sumR0  +=       1.576 * cos(2.40263038916  + 348.63519857100 *T)
    sumR0  +=       1.497 * cos(3.44614317326  + 3024.22055704320 *T)
    sumR0  +=       1.175 * cos(4.85145058205  + 433.75992198660 *T)
    sumR0  +=       1.262 * cos(5.79416346069  + 531.97858695500 *T)
    sumR0  +=       1.486 * cos(4.39599352105  + 1055.44977692610 *T)
    sumR0  +=       1.351 * cos(0.46461977407  + 707.56533788050 *T)
    sumR0  +=       1.230 * cos(5.18147817992  + 752.39122939380 *T)
    sumR0  +=       1.175 * cos(4.35535063059  + 3892.68166309700 *T)
    sumR0  +=       1.581 * cos(5.49361132323  + 419.53282798500 *T)
    sumR0  +=       1.247 * cos(4.22328749428  + 113.12708252620 *T)
    sumR0  +=       1.145 * cos(2.26067253357  + 199.12018554620 *T)
    sumR0  +=       1.385 * cos(0.89711064123  + 6073.70890781650 *T)
    sumR0  +=       1.228 * cos(4.55057016747  + 680.05731138130 *T)
    sumR0  +=       1.470 * cos(1.58708185256  + 409.18970313670 *T)
    sumR0  +=       1.366 * cos(3.99684537321  + 6065.84460129030 *T)
    sumR0  +=       1.131 * cos(1.56474593118  + 196.83676920270 *T)
    sumR0  +=       1.163 * cos(5.75528918663  + 2303.60876781320 *T)
    sumR0  +=       1.142 * cos(4.35845106342  + 1834.61531963620 *T)
    sumR0  +=       1.244 * cos(1.93187654929  + 623.22251175760 *T)
    sumR0  +=       1.136 * cos(3.13253323524  + 611.44309831080 *T)
    sumR0  +=       1.425 * cos(2.65840274172  + 1253.77101883710 *T)
    sumR0  +=       1.114 * cos(3.43048279234  + 771.30123618390 *T)
    sumR0  +=       1.314 * cos(0.15326588489  + 493.30303459600 *T)
    sumR0  +=       1.109 * cos(2.95808402860  + 1091.62525910360 *T)
    sumR0  +=       1.177 * cos(3.88159541809  + 128.36556848410 *T)
    sumR0  +=       1.150 * cos(4.69140569690  + 1.27202438720 *T)
    sumR0  +=       1.527 * cos(1.09983755253  + 683.02625679790 *T)
    sumR0  +=       1.304 * cos(5.24544813643  + 5650.29211067820 *T)
    sumR0  +=       1.347 * cos(4.11616699496  + 97.46369992610 *T)
    sumR0  +=       1.085 * cos(0.70231952018  + 1166.40685767090 *T)
    sumR0  +=       1.224 * cos(4.22994822530  + 827.17282796110 *T)
    sumR0  +=       1.142 * cos(5.36157631813  + 199.02381732660 *T)
    sumR0  +=       1.373 * cos(0.89371361814  + 799.61241183520 *T)
    sumR0  +=       1.085 * cos(1.15969472512  + 398.28701522470 *T)
    sumR0  +=       1.098 * cos(3.75659421786  + 318.39760722670 *T)
    sumR0  +=       1.136 * cos(1.35127769399  + 205.92491600630 *T)
    sumR0  +=       1.108 * cos(6.03201954623  + 206.44618086810 *T)
    sumR0  +=       1.146 * cos(0.18236094571  + 6386.16862421000 *T)
    sumR0  +=       1.216 * cos(0.49809632153  + 1178.98942026980 *T)
    sumR0  +=       1.295 * cos(2.32056477953  + 10003.91860403610 *T)
    sumR0  +=       1.082 * cos(1.96611069200  + 3700.72320866140 *T)
    sumR0  +=       1.094 * cos(5.12244388591  + 314.38613208650 *T)
    sumR0  +=       1.464 * cos(4.54056066665  + 1248.14194454460 *T)
    sumR0  +=       1.277 * cos(5.42029902662  + 9996.05429750990 *T)
    sumR0  +=       1.248 * cos(0.21625135029  + 101.60830151030 *T)
    sumR0  +=       1.055 * cos(5.53725373260  + 1578.77795454530 *T)
    sumR0  +=       1.265 * cos(5.85587479852  + 82.85835341460 *T)
    sumR0  +=       1.058 * cos(0.73824266822  + 670.91677495100 *T)
    sumR0  +=       1.127 * cos(3.63458498010  + 582.38116313410 *T)
    sumR0  +=       1.188 * cos(2.10062104535  + 423.88987789030 *T)
    sumR0  +=       1.217 * cos(2.49656109071  + 311.72600321060 *T)
    sumR0  +=       1.080 * cos(3.96349373526  + 118.87377689770 *T)
    sumR0  +=       1.175 * cos(0.91096377814  + 740.06929922580 *T)
    sumR0  +=       1.087 * cos(5.46774426742  + 494.47869076360 *T)
    sumR0  +=       1.080 * cos(3.49168860514  + 847.04604259770 *T)
    sumR0  +=       1.095 * cos(0.97418295319  + 1159.29331067010 *T)
    sumR0  +=       1.078 * cos(1.75579678521  + 1457.52593306200 *T)
    sumR0  +=       1.087 * cos(4.81206824168  + 109.68505635760 *T)
    sumR0  +=       1.293 * cos(0.02397468965  + 2723.09341168840 *T)
    sumR0  +=       1.190 * cos(4.49552956868  + 429.30650386170 *T)
    sumR0  +=       1.127 * cos(0.84707518843  + 48.75804477640 *T)
    sumR0  +=       1.047 * cos(4.58416926615  + 89.75945209430 *T)
    sumR0  +=       1.251 * cos(1.16783030789  + 455.16999895820 *T)
    sumR0  +=       1.122 * cos(5.72161306640  + 78263.70942472259 *T)
    sumR0  +=       1.027 * cos(0.16330222064  + 229.76142167330 *T)
    sumR0  +=       1.069 * cos(2.37188773221  + 848.53051530600 *T)
    sumR0  +=       1.252 * cos(5.28238896229  + 6080.82245481730 *T)
    sumR0  +=       1.181 * cos(5.22299379363  + 1459.95656727430 *T)
    sumR0  +=       1.382 * cos(0.51603096285  + 774.00954916960 *T)
    sumR0  +=       1.064 * cos(5.95222326171  + 1144.31545714310 *T)
    sumR0  +=       1.083 * cos(5.04862249728  + 629.86297800640 *T)
    sumR0  +=       1.061 * cos(3.38604454777  + 27.08733537390 *T)
    sumR0  +=       1.415 * cos(4.85796248007  + 2332.06295581640 *T)
    sumR0  +=       1.082 * cos(4.07686503205  + 1245.17299912800 *T)
    sumR0  +=       1.112 * cos(6.07617329506  + 870.46185713940 *T)
    sumR0  +=       1.072 * cos(0.73647405514  + 1482.79872732750 *T)
    sumR0  +=       1.322 * cos(2.81015928946  + 223.33340374560 *T)
    sumR0  +=       1.398 * cos(1.55232715558  + 25771.51127191760 *T)
    sumR0  +=       1.065 * cos(3.98884050015  + 683.18631549230 *T)
    sumR0  +=       1.008 * cos(5.19594380826  + 316.13123722570 *T)
    sumR0  +=       1.197 * cos(0.79148395839  + 9580.50180689780 *T)
    sumR0  +=       1.402 * cos(4.82957073563  + 883.79517926370 *T)
    sumR0  +=       1.064 * cos(4.65334775068  + 201.99276274320 *T)
    sumR0  +=       1.108 * cos(5.88857586823  + 657.16276170140 *T)
    sumR0  +=       0.996 * cos(4.99081076034  + 426.75824957040 *T)
    sumR0  +=       0.996 * cos(1.60533807224  + 426.43813218160 *T)
    sumR0  +=       1.028 * cos(5.92128319450  + 108.50940019000 *T)
    sumR0  +=       1.004 * cos(3.54259597860  + 754.03576079650 *T)
    sumR0  +=       1.214 * cos(0.81213286478  + 1773.91780271860 *T)
    sumR0  +=       1.318 * cos(0.60269176130  + 1123.11782849260 *T)
    sumR0  +=       1.327 * cos(4.75885008900  + 321.80849562800 *T)
    sumR0  +=       1.014 * cos(5.79119811472  + 6460.81221096080 *T)
    sumR0  +=       1.178 * cos(0.47169015111  + 495.96316347190 *T)
    sumR0  +=       0.999 * cos(5.95030119388  + 3906.90875709860 *T)
    sumR0  +=       1.033 * cos(3.75433174131  + 414.81877742920 *T)
    sumR0  +=       1.002 * cos(1.39171012432  + 1251.34038462480 *T)
    sumR0  +=       0.990 * cos(1.32621236288  + 1268.74887236410 *T)
    sumR0  +=       1.275 * cos(1.95417923977  + 757.80785536520 *T)
    sumR0  +=       1.174 * cos(6.04352585298  + 225.07850888480 *T)
    sumR0  +=       1.174 * cos(5.43253033568  + 849.26422848890 *T)
    sumR0  +=       1.004 * cos(5.33434806968  + 46.20979048510 *T)
    sumR0  +=       1.108 * cos(1.28177889943  + 294.30046912880 *T)
    sumR0  +=       0.976 * cos(1.83523959034  + 5.88970672340 *T)
    sumR0  +=       0.971 * cos(3.38563950019  + 306.09692891810 *T)
    sumR0  +=       1.050 * cos(3.88449467091  + 632.26247445140 *T)
    sumR0  +=       1.050 * cos(3.26096036982  + 159.71512552120 *T)
    sumR0  +=       1.041 * cos(2.36429894351  + 821.54375366860 *T)
    sumR0  +=       1.218 * cos(4.61739999906  + 990.22940591440 *T)
    sumR0  +=       1.342 * cos(1.55614528399  + 498.93210888850 *T)
    sumR0  +=       0.967 * cos(3.84645372731  + 604.47256366190 *T)
    sumR0  +=       1.171 * cos(0.42265751679  + 10011.03215103690 *T)
    sumR0  +=       0.965 * cos(0.05396772193  + 962.50893109410 *T)
    sumR0  +=       1.096 * cos(3.04685199735  + 608.40471692500 *T)
    sumR0  +=       1.166 * cos(6.14999706886  + 737.31280213030 *T)
    sumR0  +=       0.986 * cos(3.71830385737  + 1235.61177157240 *T)
    sumR0  +=       0.953 * cos(0.79704964354  + 16.46232623530 *T)
    sumR0  +=       0.976 * cos(1.05304205075  + 8.33738727820 *T)
    sumR0  +=       1.142 * cos(1.06057209808  + 369.97583957340 *T)
    sumR0  +=       1.060 * cos(4.36236322604  + 633.30500417500 *T)
    sumR0  +=       1.138 * cos(4.31859245106  + 98.42690777260 *T)
    sumR0  +=       1.006 * cos(5.89037944896  + 10007.09999777380 *T)
    sumR0  +=       0.967 * cos(1.56852913547  + 157.63995198190 *T)
    sumR0  +=       1.082 * cos(0.99548769517  + 4.14460158420 *T)
    sumR0  +=       1.009 * cos(6.15037679495  + 401.32539661050 *T)
    sumR0  +=       0.981 * cos(2.37620383333  + 35.21227433100 *T)
    sumR0  +=       1.011 * cos(5.42685471402  + 110.72758608120 *T)
    sumR0  +=       0.989 * cos(4.11575312649  + 413.85556958270 *T)
    sumR0  +=       1.024 * cos(1.90798238649  + 1175.80802653210 *T)
    sumR0  +=       1.079 * cos(4.95981991427  + 631.82053146670 *T)
    sumR0  +=       0.982 * cos(3.02842195594  + 347.41135829360 *T)
    sumR0  +=       0.979 * cos(3.74615289445  + 700.45179087970 *T)
    sumR0  +=       0.928 * cos(5.22236951137  + 1173.52040467170 *T)
    sumR0  +=       0.912 * cos(4.14451390992  + 469.72716135990 *T)
    sumR0  +=       0.938 * cos(1.18273838991  + 254.35289238260 *T)
    sumR0  +=       0.977 * cos(1.26684849112  + 104.57724692690 *T)
    sumR0  +=       1.134 * cos(5.87478488618  + 6058.73105428950 *T)
    sumR0  +=       1.092 * cos(4.51789158271  + 532.13864564940 *T)
    sumR0  +=       1.132 * cos(4.55420027150  + 1912.57831194120 *T)
    sumR0  +=       0.915 * cos(4.87266214195  + 18.91000679010 *T)
    sumR0  +=       0.915 * cos(5.96624579967  + 1987.21689815660 *T)
    sumR0  +=       1.039 * cos(2.40020830681  + 6475.03930496240 *T)
    sumR0  +=       0.965 * cos(4.98927479154  + 394.35486196160 *T)
    sumR0  +=       0.964 * cos(3.60903715900  + 3281.23856478620 *T)
    sumR0  +=       0.897 * cos(2.81660605059  + 316.23181096220 *T)
    sumR0  +=       1.008 * cos(5.78024010734  + 502.86426215160 *T)
    sumR0  +=       1.005 * cos(6.12431717236  + 6275.96230299060 *T)
    sumR0  +=       0.895 * cos(6.09245508513  + 316.55192835100 *T)
    sumR0  +=       1.224 * cos(1.73535287415  + 5120.60114558360 *T)
    sumR0  +=       1.138 * cos(2.92901543353  + 1037.29052966140 *T)
    sumR0  +=       0.924 * cos(5.70601816488  + 614.83694036960 *T)
    sumR0  +=       0.893 * cos(5.25155704274  + 475.35623565240 *T)
    sumR0  +=       1.042 * cos(1.10027795842  + 1518.22344997960 *T)
    sumR0  +=       0.890 * cos(4.03192782386  + 1314.32552340280 *T)
    sumR0  +=       0.926 * cos(3.35110915055  + 635.70450062000 *T)
    sumR0  +=       1.130 * cos(5.49282680494  + 92.94084583200 *T)
    sumR0  +=       0.890 * cos(2.12933822393  + 3384.33133900480 *T)
    sumR0  +=       0.967 * cos(5.86215202069  + 13.33332212430 *T)
    sumR0  +=       1.004 * cos(1.73116475997  + 10316.37832042960 *T)
    sumR0  +=       1.226 * cos(3.52834223937  + 80.41067285980 *T)
    sumR0  +=       1.029 * cos(4.90620832171  + 19.12245511120 *T)
    sumR0  +=       0.996 * cos(0.76740358631  + 733.42883297700 *T)
    sumR0  +=       0.910 * cos(4.08904906301  + 3333.49887969900 *T)
    sumR0  +=       1.235 * cos(1.23871819142  + 357.23321828010 *T)
    sumR0  +=       1.218 * cos(0.18349810348  + 1090.40141882620 *T)
    sumR0  +=       0.864 * cos(4.71917415767  + 620.25356634100 *T)
    sumR0  +=       0.935 * cos(1.45887009044  + 1042.91960395390 *T)
    sumR0  +=       0.866 * cos(4.04792682992  + 522.52923398400 *T)
    sumR0  +=       1.158 * cos(5.43322209110  + 1089.12939443900 *T)
    sumR0  +=       0.860 * cos(0.49220052417  + 64.95973858080 *T)
    sumR0  +=       1.190 * cos(5.58965369650  + 2810.92146160520 *T)
    sumR0  +=       0.957 * cos(3.12914047010  + 628.59095361920 *T)
    sumR0  +=       0.861 * cos(5.69790389801  + 103.84353374400 *T)
    sumR0  +=       1.037 * cos(5.91424823262  + 11.30633269480 *T)
    sumR0  +=       0.918 * cos(0.21424702155  + 373.01422095920 *T)
    sumR0  +=       0.836 * cos(3.02501867546  + 387.24131496080 *T)
    sumR0  +=       1.158 * cos(3.33343863758  + 6290.18939699220 *T)
    sumR0  +=       0.856 * cos(0.81593288669  + 907.37105249980 *T)
    sumR0  +=       1.036 * cos(3.11936047271  + 5429.87946823940 *T)
    sumR0  +=       0.853 * cos(6.23618175592  + 938.88066823130 *T)
    sumR0  +=       0.982 * cos(3.39082880963  + 521.61421024730 *T)
    sumR0  +=       0.851 * cos(5.14029961564  + 802.79380557290 *T)
    sumR0  +=       0.828 * cos(0.28399876908  + 338.48327018440 *T)
    sumR0  +=       0.868 * cos(1.89151676387  + 627.15466502070 *T)
    sumR0  +=       0.878 * cos(2.67671626912  + 688.65533109040 *T)
    sumR0  +=       0.921 * cos(0.18441593712  + 3803.81598288000 *T)
    sumR0  +=       0.821 * cos(0.74694467095  + 1152.17976366930 *T)
    sumR0  +=       0.841 * cos(2.42616504698  + 625.88264063350 *T)
    sumR0  +=       0.862 * cos(3.35273419872  + 425.32616648880 *T)
    sumR0  +=       0.887 * cos(3.46938383985  + 1748.78802080500 *T)
    sumR0  +=       0.815 * cos(5.95428642326  + 321.05773610260 *T)
    sumR0  +=       0.958 * cos(1.35525670354  + 1201.83158032300 *T)
    sumR0  +=       0.905 * cos(5.43093361027  + 236.87496867410 *T)
    sumR0  +=       0.862 * cos(3.24167644516  + 427.87021526320 *T)
    sumR0  +=       0.793 * cos(5.01131700831  + 109.99387289830 *T)
    sumR0  +=       0.791 * cos(2.21809404489  + 110.41876954050 *T)
    sumR0  +=       0.842 * cos(5.04957483651  + 444.12429869430 *T)
    sumR0  +=       0.887 * cos(4.23752031714  + 3553.91152213780 *T)
    sumR0  +=       0.851 * cos(4.64229745145  + 4105.98075853500 *T)
    sumR0  +=       1.067 * cos(4.69271921916  + 559.69906177530 *T)
    sumR0  +=       0.966 * cos(4.55013458162  + 9360.08916445900 *T)
    sumR0  +=       0.977 * cos(1.50925667031  + 186.47239249500 *T)
    sumR0  +=       1.092 * cos(0.58153747940  + 203.26478713040 *T)
    sumR0  +=       0.997 * cos(0.24589891452  + 439.93151300030 *T)
    sumR0  +=       0.786 * cos(3.84829878956  + 194.38908864790 *T)
    sumR0  +=       0.774 * cos(3.76050639310  + 219.14061805160 *T)
    sumR0  +=       0.845 * cos(4.21602090805  + 2648.45482547300 *T)
    sumR0  +=       0.962 * cos(0.30590569897  + 229.34073054800 *T)
    sumR0  +=       0.763 * cos(0.05577842075  + 846.13101886100 *T)
    sumR0  +=       0.839 * cos(5.68124142701  + 2620.00063746980 *T)
    sumR0  +=       0.818 * cos(2.52609626540  + 26301.20223701220 *T)
    sumR0  +=       0.929 * cos(0.71906225883  + 740.28174754690 *T)
    sumR0  +=       0.895 * cos(5.83218231202  + 4539.69249641180 *T)
    sumR0  +=       0.939 * cos(3.68460642343  + 817.77165909990 *T)
    sumR0  +=       0.755 * cos(4.90024080821  + 532.87235883230 *T)
    sumR0  +=       0.784 * cos(1.14096100609  + 551.03160609700 *T)
    sumR0  +=       0.899 * cos(1.85252071775  + 835.78789401270 *T)
    sumR0  +=       0.753 * cos(0.04254534997  + 1534.73816584160 *T)
    sumR0  +=       1.033 * cos(1.00137993270  + 134.11226285560 *T)
    sumR0  +=       0.851 * cos(6.12272864540  + 1475.68518032670 *T)
    sumR0  +=       0.797 * cos(5.14322789256  + 473.65931462300 *T)
    sumR0  +=       0.729 * cos(4.94931618796  + 476.10699517780 *T)
    sumR0  +=       0.852 * cos(3.11058720799  + 232.42155054920 *T)
    sumR0  +=       0.783 * cos(4.50805467439  + 1151.42900414390 *T)
    sumR0  +=       0.751 * cos(0.92289775523  + 1884.12412393800 *T)
    sumR0  +=       0.832 * cos(4.19686348297  + 29.20494752860 *T)
    sumR0  +=       0.720 * cos(0.40961041030  + 522.62560220360 *T)
    sumR0  +=       0.722 * cos(3.96121088528  + 1474.93442080130 *T)
    sumR0  +=       0.788 * cos(0.99170388242  + 121.84272231430 *T)
    sumR0  +=       0.722 * cos(4.16734185316  + 232.20910222810 *T)
    sumR0  +=       0.970 * cos(2.32204039048  + 566.60016045500 *T)
    sumR0  +=       0.814 * cos(5.19337022083  + 948.21240112330 *T)
    sumR0  +=       0.724 * cos(2.30837674225  + 949.12742486000 *T)
    sumR0  +=       0.770 * cos(5.89605163084  + 156.67674413540 *T)
    sumR0  +=       0.705 * cos(5.40102118863  + 1193.96727379680 *T)
    sumR0  +=       0.809 * cos(3.56474059969  + 845.33207522580 *T)
    sumR0  +=       0.755 * cos(3.94580797273  + 451.72797278960 *T)
    sumR0  +=       0.731 * cos(6.11847213487  + 1239.54392483550 *T)
    sumR0  +=       0.747 * cos(1.40599730465  + 782.34693644780 *T)
    sumR0  +=       0.861 * cos(1.84312374221  + 984.60033162190 *T)
    sumR0  +=       0.695 * cos(4.78088165969  + 10419.47109464820 *T)
    sumR0  +=       0.804 * cos(1.07998437400  + 89.00869256890 *T)
    sumR0  +=       0.828 * cos(5.86080569334  + 845.11962690470 *T)
    sumR0  +=       0.692 * cos(3.38736418117  + 6489.26139842860 *T)
    sumR0  +=       0.694 * cos(3.07863807714  + 316.60431797770 *T)
    sumR0  +=       0.690 * cos(5.86681311380  + 316.17942133550 *T)
    sumR0  +=       0.714 * cos(5.38707933404  + 567.82400073240 *T)
    sumR0  +=       0.767 * cos(0.12081849650  + 485.88067105450 *T)
    sumR0  +=       0.820 * cos(1.87877245664  + 499.89531673500 *T)
    sumR0  +=       0.705 * cos(0.57839934869  + 1053.75285589670 *T)
    sumR0  +=       0.684 * cos(4.88442270630  + 2545.36205125440 *T)
    sumR0  +=       0.689 * cos(6.14296395253  + 622.48879857470 *T)
    sumR0  +=       0.823 * cos(3.50224755884  + 877.57540414020 *T)
    sumR0  +=       0.827 * cos(2.59300433753  + 232.94281541100 *T)
    sumR0  +=       0.735 * cos(3.05650026582  + 66.91729204110 *T)

    # SATURN Radius
    sumR1  = 0
    sumR1  +=  6182981.282 * cos(0.25843515034  + 213.29909543800 *T)
    sumR1  +=  506577.574 * cos(0.71114650941  + 206.18554843720 *T)
    sumR1  +=  341394.136 * cos(5.79635773960  + 426.59819087600 *T)
    sumR1  +=  188491.375 * cos(0.47215719444  + 220.41264243880 *T)
    sumR1  +=  186261.540 * cos(3.14159265359  + 0.00000000000 *T)
    sumR1  +=  143891.176 * cos(1.40744864239  + 7.11354700080 *T)
    sumR1  +=   49621.111 * cos(6.01744469580  + 103.09277421860 *T)
    sumR1  +=   20928.189 * cos(5.09245654470  + 639.89728631400 *T)
    sumR1  +=   19952.612 * cos(1.17560125007  + 419.48464387520 *T)
    sumR1  +=   18839.639 * cos(1.60819563173  + 110.20632121940 *T)
    sumR1  +=   12892.827 * cos(5.94330258435  + 433.71173787680 *T)
    sumR1  +=   13876.565 * cos(0.75886204364  + 199.07200143640 *T)
    sumR1  +=    5396.699 * cos(1.28852405908  + 14.22709400160 *T)
    sumR1  +=    4869.308 * cos(0.86793894213  + 323.50541665740 *T)
    sumR1  +=    4247.455 * cos(0.39299384543  + 227.52618943960 *T)
    sumR1  +=    3252.084 * cos(1.25853470491  + 95.97922721780 *T)
    sumR1  +=    2856.006 * cos(2.16731405366  + 735.87651353180 *T)
    sumR1  +=    2909.411 * cos(4.60679154788  + 202.25339517410 *T)
    sumR1  +=    3081.408 * cos(3.43662557418  + 522.57741809380 *T)
    sumR1  +=    1987.689 * cos(2.45054204795  + 412.37109687440 *T)
    sumR1  +=    1941.309 * cos(6.02393385142  + 209.36694217490 *T)
    sumR1  +=    1581.446 * cos(1.29191789712  + 210.11770170030 *T)
    sumR1  +=    1339.511 * cos(4.30801821806  + 853.19638175200 *T)
    sumR1  +=    1315.590 * cos(1.25296446023  + 117.31986822020 *T)
    sumR1  +=    1203.085 * cos(1.86654673794  + 316.39186965660 *T)
    sumR1  +=    1091.088 * cos(0.07527246854  + 216.48048917570 *T)
    sumR1  +=     954.403 * cos(5.15173410519  + 647.01083331480 *T)
    sumR1  +=     966.012 * cos(0.47991379141  + 632.78373931320 *T)
    sumR1  +=     881.827 * cos(1.88471724478  + 1052.26838318840 *T)
    sumR1  +=     874.215 * cos(1.40224683864  + 224.34479570190 *T)
    sumR1  +=     897.512 * cos(0.98343776092  + 529.69096509460 *T)
    sumR1  +=     784.866 * cos(3.06377517461  + 838.96928775040 *T)
    sumR1  +=     739.892 * cos(1.38225356694  + 625.67019231240 *T)
    sumR1  +=     612.961 * cos(3.03307306767  + 63.73589830340 *T)
    sumR1  +=     658.210 * cos(4.14362930980  + 309.27832265580 *T)
    sumR1  +=     649.600 * cos(1.72489486160  + 742.99006053260 *T)
    sumR1  +=     599.236 * cos(2.54924174765  + 217.23124870110 *T)
    sumR1  +=     502.886 * cos(2.12958819475  + 3.93215326310 *T)
    sumR1  +=     413.017 * cos(4.59334402271  + 415.55249061210 *T)
    sumR1  +=     356.117 * cos(2.30312127651  + 728.76296653100 *T)
    sumR1  +=     344.777 * cos(5.88787577835  + 440.82528487760 *T)
    sumR1  +=     395.004 * cos(0.53349091102  + 956.28915597060 *T)
    sumR1  +=     335.526 * cos(1.61614647174  + 1368.66025284500 *T)
    sumR1  +=     362.772 * cos(4.70691652867  + 302.16477565500 *T)
    sumR1  +=     321.611 * cos(0.97931764923  + 3.18139373770 *T)
    sumR1  +=     277.783 * cos(0.26007031431  + 195.13984817330 *T)
    sumR1  +=     291.173 * cos(2.83129427918  + 1155.36115740700 *T)
    sumR1  +=     264.971 * cos(2.42670902733  + 88.86568021700 *T)
    sumR1  +=     264.864 * cos(5.82860588985  + 149.56319713460 *T)
    sumR1  +=     316.777 * cos(3.58395655749  + 515.46387109300 *T)
    sumR1  +=     294.324 * cos(2.81632778983  + 11.04570026390 *T)
    sumR1  +=     244.864 * cos(1.04493438899  + 942.06206196900 *T)
    sumR1  +=     215.368 * cos(3.56535574833  + 490.33408917940 *T)
    sumR1  +=     264.047 * cos(1.28547685567  + 1059.38193018920 *T)
    sumR1  +=     246.245 * cos(0.90730313861  + 191.95845443560 *T)
    sumR1  +=     222.077 * cos(5.13193212050  + 269.92144674060 *T)
    sumR1  +=     194.973 * cos(4.56665009915  + 846.08283475120 *T)
    sumR1  +=     182.802 * cos(2.67913220473  + 127.47179660680 *T)
    sumR1  +=     181.645 * cos(4.93431600689  + 74.78159856730 *T)
    sumR1  +=     174.651 * cos(3.44560172182  + 137.03302416240 *T)
    sumR1  +=     165.515 * cos(5.99775895715  + 536.80451209540 *T)
    sumR1  +=     154.809 * cos(1.19720845085  + 265.98929347750 *T)
    sumR1  +=     169.743 * cos(4.63464467495  + 284.14854074220 *T)
    sumR1  +=     151.526 * cos(0.52928231044  + 330.61896365820 *T)
    sumR1  +=     152.461 * cos(5.43886711695  + 422.66603761290 *T)
    sumR1  +=     157.687 * cos(2.99559914619  + 340.77089204480 *T)
    sumR1  +=     140.630 * cos(2.02069760726  + 1045.15483618760 *T)
    sumR1  +=     139.834 * cos(1.35282959390  + 1685.05212250160 *T)
    sumR1  +=     140.977 * cos(1.27099900689  + 203.00415469950 *T)
    sumR1  +=     136.013 * cos(5.01678984678  + 351.81659230870 *T)
    sumR1  +=     153.391 * cos(0.26968607873  + 1272.68102562720 *T)
    sumR1  +=     129.476 * cos(1.14344730612  + 21.34064100240 *T)
    sumR1  +=     127.831 * cos(2.53876158952  + 1471.75302706360 *T)
    sumR1  +=     126.538 * cos(3.00310970076  + 277.03499374140 *T)
    sumR1  +=     100.277 * cos(3.61360169153  + 1066.49547719000 *T)
    sumR1  +=     103.169 * cos(0.38175114761  + 203.73786788240 *T)
    sumR1  +=     107.527 * cos(4.31870663477  + 210.85141488320 *T)
    sumR1  +=      95.934 * cos(0.79463744168  + 1258.45393162560 *T)
    sumR1  +=      82.663 * cos(0.28181414606  + 234.63973644040 *T)
    sumR1  +=      97.986 * cos(2.56085956186  + 191.20769491020 *T)
    sumR1  +=      97.389 * cos(3.26245865063  + 831.85574074960 *T)
    sumR1  +=      72.227 * cos(4.37984630380  + 860.30992875280 *T)
    sumR1  +=      70.639 * cos(0.73191513920  + 437.64389113990 *T)
    sumR1  +=      70.447 * cos(0.87698401733  + 423.41679713830 *T)
    sumR1  +=      72.057 * cos(5.58013290518  + 429.77958461370 *T)
    sumR1  +=      73.332 * cos(0.62505906432  + 1375.77379984580 *T)
    sumR1  +=      66.433 * cos(2.68414462465  + 405.25754987360 *T)
    sumR1  +=      63.812 * cos(1.75051498180  + 1361.54670584420 *T)
    sumR1  +=      61.601 * cos(1.09332288242  + 2001.44399215820 *T)
    sumR1  +=      67.006 * cos(0.06872766216  + 408.43894361130 *T)
    sumR1  +=      68.945 * cos(2.47127505057  + 949.17560896980 *T)
    sumR1  +=      60.456 * cos(2.25094790113  + 1788.14489672020 *T)
    sumR1  +=      67.074 * cos(5.45365870159  + 200.76892246580 *T)
    sumR1  +=      65.579 * cos(0.05539079332  + 1589.07289528380 *T)
    sumR1  +=      49.320 * cos(4.17243429807  + 138.51749687070 *T)
    sumR1  +=      50.648 * cos(6.26867505289  + 223.59403617650 *T)
    sumR1  +=      55.166 * cos(4.59491533823  + 628.85158605010 *T)
    sumR1  +=      47.916 * cos(0.83929741626  + 10.29494073850 *T)
    sumR1  +=      46.691 * cos(2.17322569098  + 312.19908396260 *T)
    sumR1  +=      54.179 * cos(0.28360076018  + 124.43341522100 *T)
    sumR1  +=      49.511 * cos(3.79960349195  + 215.74677599280 *T)
    sumR1  +=      40.136 * cos(5.18161452756  + 1478.86657406440 *T)
    sumR1  +=      39.302 * cos(0.56257369109  + 1574.84580128220 *T)
    sumR1  +=      34.962 * cos(4.68487505703  + 38.13303563780 *T)
    sumR1  +=      42.770 * cos(2.98582069454  + 1148.24761040620 *T)
    sumR1  +=      36.521 * cos(0.63453270366  + 52.69019803950 *T)
    sumR1  +=      39.752 * cos(0.28412706854  + 131.40394986990 *T)
    sumR1  +=      31.777 * cos(5.19036499973  + 76.26607127560 *T)
    sumR1  +=      33.041 * cos(1.97964846430  + 142.44965013380 *T)
    sumR1  +=      42.053 * cos(4.83017951800  + 288.08069400530 *T)
    sumR1  +=      30.757 * cos(1.47903923433  + 1677.93857550080 *T)
    sumR1  +=      42.829 * cos(3.38225543528  + 208.63322899200 *T)
    sumR1  +=      29.245 * cos(5.09869866956  + 654.12438031560 *T)
    sumR1  +=      29.165 * cos(4.95664881649  + 1795.25844372100 *T)
    sumR1  +=      29.136 * cos(2.74747553685  + 404.50679034820 *T)
    sumR1  +=      32.689 * cos(6.12099521344  + 145.63104387150 *T)
    sumR1  +=      28.008 * cos(0.83185907283  + 2317.83586181480 *T)
    sumR1  +=      27.725 * cos(2.24364073545  + 430.53034413910 *T)
    sumR1  +=      29.939 * cos(1.96415498448  + 2104.53676637680 *T)
    sumR1  +=      32.982 * cos(3.28236160491  + 222.86032299360 *T)
    sumR1  +=      31.772 * cos(6.02453027348  + 1905.46476494040 *T)
    sumR1  +=      26.959 * cos(5.24308283338  + 388.46515523820 *T)
    sumR1  +=      26.514 * cos(0.99638302878  + 107.02492748170 *T)
    sumR1  +=      25.421 * cos(2.87336642463  + 703.63318461740 *T)
    sumR1  +=      24.908 * cos(1.07713811775  + 99.91138048090 *T)
    sumR1  +=      24.955 * cos(6.23974037842  + 106.27416795630 *T)
    sumR1  +=      24.894 * cos(0.81040976807  + 312.45971639350 *T)
    sumR1  +=      24.340 * cos(0.54867402916  + 214.26230328450 *T)
    sumR1  +=      28.441 * cos(0.82630052794  + 1692.16566950240 *T)
    sumR1  +=      23.219 * cos(5.07995629354  + 479.28838891550 *T)
    sumR1  +=      24.362 * cos(3.10643455533  + 212.33588759150 *T)
    sumR1  +=      21.951 * cos(6.06688237952  + 85.82729883120 *T)
    sumR1  +=      22.046 * cos(3.89863665506  + 563.63121503840 *T)
    sumR1  +=      22.596 * cos(4.86725457223  + 295.05122865420 *T)
    sumR1  +=      21.256 * cos(5.10797617452  + 333.65734504400 *T)
    sumR1  +=      25.985 * cos(2.20813879137  + 1265.56747862640 *T)
    sumR1  +=      20.904 * cos(3.28855303434  + 70.84944530420 *T)
    sumR1  +=      21.505 * cos(3.79541155976  + 347.88443904560 *T)
    sumR1  +=      22.067 * cos(4.22716352578  + 217.96496188400 *T)
    sumR1  +=      20.629 * cos(1.68732248608  + 231.45834270270 *T)
    sumR1  +=      21.429 * cos(3.08914428467  + 554.06998748280 *T)
    sumR1  +=      21.310 * cos(0.38868340861  + 319.57326339430 *T)
    sumR1  +=      20.521 * cos(2.45651851283  + 18.15924726470 *T)
    sumR1  +=      26.055 * cos(4.27554951169  + 483.22054217860 *T)
    sumR1  +=      20.703 * cos(5.12057936320  + 362.86229257260 *T)
    sumR1  +=      22.047 * cos(5.51249354809  + 343.21857259960 *T)
    sumR1  +=      19.443 * cos(2.02441679295  + 313.21047591890 *T)
    sumR1  +=      20.163 * cos(5.08481373110  + 750.10360753340 *T)
    sumR1  +=      20.125 * cos(3.42997916125  + 213.34727954780 *T)
    sumR1  +=      24.196 * cos(0.64787472796  + 207.88246946660 *T)
    sumR1  +=      21.977 * cos(0.72894956852  + 99.16062095550 *T)
    sumR1  +=      21.120 * cos(2.69286728009  + 1464.63948006280 *T)
    sumR1  +=      17.192 * cos(4.71525117969  + 2111.65031337760 *T)
    sumR1  +=      18.540 * cos(0.04817255506  + 245.54242435240 *T)
    sumR1  +=      17.521 * cos(3.83662880684  + 497.44763618020 *T)
    sumR1  +=      16.107 * cos(4.22374822303  + 565.11568774670 *T)
    sumR1  +=      21.607 * cos(4.16647257628  + 2.44768055480 *T)
    sumR1  +=      15.979 * cos(0.27376396113  + 225.82926841020 *T)
    sumR1  +=      16.831 * cos(1.41134653939  + 114.13847448250 *T)
    sumR1  +=      15.626 * cos(2.82768623405  + 81.75213321620 *T)
    sumR1  +=      15.499 * cos(1.20606390539  + 1994.33044515740 *T)
    sumR1  +=      15.168 * cos(3.84591816174  + 1162.47470440780 *T)
    sumR1  +=      16.436 * cos(3.04752365976  + 134.58534360760 *T)
    sumR1  +=      15.870 * cos(0.33026420429  + 1891.23767093880 *T)
    sumR1  +=      20.370 * cos(0.23170286692  + 213.25091132820 *T)
    sumR1  +=      16.291 * cos(1.70643197929  + 2420.92863603340 *T)
    sumR1  +=      16.280 * cos(4.94159427320  + 357.44566660120 *T)
    sumR1  +=      18.076 * cos(5.69515344123  + 56.62235130260 *T)
    sumR1  +=      13.724 * cos(0.57240190030  + 2634.22773147140 *T)
    sumR1  +=      17.355 * cos(3.55311137444  + 218.92816973050 *T)
    sumR1  +=      13.740 * cos(5.70545527289  + 92.04707395470 *T)
    sumR1  +=      15.328 * cos(1.31338692850  + 216.21985674480 *T)
    sumR1  +=      12.538 * cos(5.19222019427  + 635.96513305090 *T)
    sumR1  +=      12.815 * cos(1.60151130870  + 320.32402291970 *T)
    sumR1  +=      13.043 * cos(0.45068441373  + 1169.58825140860 *T)
    sumR1  +=      11.984 * cos(5.94916123570  + 543.91805909620 *T)
    sumR1  +=      11.753 * cos(2.80279347133  + 217.49188113200 *T)
    sumR1  +=      14.746 * cos(5.56520105813  + 344.70304530790 *T)
    sumR1  +=      12.762 * cos(1.63557330778  + 273.10284047830 *T)
    sumR1  +=      11.855 * cos(2.46234840263  + 721.64941953020 *T)
    sumR1  +=      13.309 * cos(5.75641013916  + 2221.85663459700 *T)
    sumR1  +=      14.471 * cos(0.45316163629  + 2008.55753915900 *T)
    sumR1  +=      11.840 * cos(1.75720772380  + 160.60889739850 *T)
    sumR1  +=      12.374 * cos(1.01456317602  + 329.72519178090 *T)
    sumR1  +=      10.747 * cos(1.58065203003  + 212.77783057620 *T)
    sumR1  +=      12.758 * cos(1.91952373240  + 1581.95934828300 *T)
    sumR1  +=      11.944 * cos(4.44720922423  + 32.24332891440 *T)
    sumR1  +=      11.865 * cos(5.10696147162  + 4.66586644600 *T)
    sumR1  +=      11.861 * cos(4.30847607078  + 618.55664531160 *T)
    sumR1  +=      10.036 * cos(0.48709852137  + 305.34616939270 *T)
    sumR1  +=      12.777 * cos(3.74412991331  + 508.35032409220 *T)
    sumR1  +=      10.677 * cos(0.76645916273  + 218.71572140940 *T)
    sumR1  +=      11.351 * cos(3.00009819697  + 198.32124191100 *T)
    sumR1  +=      10.249 * cos(2.40923650192  + 546.95644048200 *T)
    sumR1  +=       9.984 * cos(2.63882014753  + 416.30325013750 *T)
    sumR1  +=       9.345 * cos(5.45917317860  + 414.06801790380 *T)
    sumR1  +=       9.317 * cos(4.46380159546  + 2428.04218303420 *T)
    sumR1  +=       9.928 * cos(4.04821559448  + 62.25142559510 *T)
    sumR1  +=      12.767 * cos(3.43273835457  + 258.87574647670 *T)
    sumR1  +=       9.733 * cos(1.61066324680  + 327.43756992050 *T)
    sumR1  +=      11.163 * cos(2.40665325234  + 1781.03134971940 *T)
    sumR1  +=      10.608 * cos(2.07480020830  + 213.82036029980 *T)
    sumR1  +=       9.125 * cos(2.92369523159  + 1279.79457262800 *T)
    sumR1  +=       9.525 * cos(1.10338403136  + 113.38771495710 *T)
    sumR1  +=       9.805 * cos(3.28427768485  + 275.55052103310 *T)
    sumR1  +=      11.263 * cos(1.89402915826  + 561.18353448360 *T)
    sumR1  +=       8.572 * cos(2.17858055966  + 425.11371816770 *T)
    sumR1  +=       8.577 * cos(1.95484887975  + 35.42472265210 *T)
    sumR1  +=      10.157 * cos(0.09037368733  + 182.27960680100 *T)
    sumR1  +=      11.807 * cos(3.71278037583  + 350.33211960040 *T)
    sumR1  +=       8.595 * cos(1.83382454431  + 629.60234557550 *T)
    sumR1  +=       8.396 * cos(3.76782674303  + 251.43213107580 *T)
    sumR1  +=       8.460 * cos(0.35676476459  + 617.80588578620 *T)
    sumR1  +=       8.250 * cos(5.31140994372  + 65.22037101170 *T)
    sumR1  +=       8.342 * cos(1.38307663880  + 1.48447270830 *T)
    sumR1  +=       7.987 * cos(5.13622898170  + 22.09140052780 *T)
    sumR1  +=       8.377 * cos(0.91817077859  + 1485.98012106520 *T)
    sumR1  +=       7.980 * cos(0.94199750915  + 2310.72231481400 *T)
    sumR1  +=       8.898 * cos(0.54037636841  + 168.05251279940 *T)
    sumR1  +=       8.233 * cos(3.45785310349  + 424.15051032120 *T)
    sumR1  +=       8.034 * cos(3.38451795597  + 144.14657116320 *T)
    sumR1  +=       7.871 * cos(5.14041888473  + 358.93013930950 *T)
    sumR1  +=       8.868 * cos(6.13541788772  + 621.73803904930 *T)
    sumR1  +=       7.523 * cos(5.75475671698  + 447.93883187840 *T)
    sumR1  +=       7.515 * cos(2.18967849979  + 264.50482076920 *T)
    sumR1  +=       8.083 * cos(1.42661116937  + 2737.32050569000 *T)
    sumR1  +=       8.199 * cos(0.96419579079  + 767.36908292080 *T)
    sumR1  +=       8.232 * cos(0.35471613534  + 278.51946644970 *T)
    sumR1  +=       8.226 * cos(5.44467204721  + 254.94359321360 *T)
    sumR1  +=       6.779 * cos(1.19567671732  + 5.41662597140 *T)
    sumR1  +=       8.928 * cos(4.88240256153  + 120.35824960600 *T)
    sumR1  +=       7.845 * cos(4.56376829397  + 280.96714700450 *T)
    sumR1  +=       6.566 * cos(3.50152072308  + 9.56122755560 *T)
    sumR1  +=       6.398 * cos(0.33471834269  + 2950.61960112800 *T)
    sumR1  +=       6.881 * cos(3.39438820076  + 98.89998852460 *T)
    sumR1  +=       7.418 * cos(4.52451404934  + 5.62907429250 *T)
    sumR1  +=       8.021 * cos(0.94470052446  + 636.71589257630 *T)
    sumR1  +=       6.134 * cos(0.18013315689  + 2207.62954059540 *T)
    sumR1  +=       7.153 * cos(3.85218295688  + 214.04985496340 *T)
    sumR1  +=       6.046 * cos(4.66733263196  + 543.02428721890 *T)
    sumR1  +=       6.365 * cos(2.12000811873  + 274.06604832480 *T)
    sumR1  +=       6.481 * cos(5.31032923608  + 6076.89030155420 *T)
    sumR1  +=       5.935 * cos(6.16808119163  + 650.94298657790 *T)
    sumR1  +=       5.752 * cos(3.55773840903  + 1073.60902419080 *T)
    sumR1  +=       6.438 * cos(0.44934410249  + 10007.09999777380 *T)
    sumR1  +=       6.283 * cos(3.20942251433  + 219.44943459230 *T)
    sumR1  +=       5.542 * cos(3.61193204407  + 125.98732389850 *T)
    sumR1  +=       5.522 * cos(3.84217355164  + 181.05576652360 *T)
    sumR1  +=       5.777 * cos(3.00590926498  + 121.25202148330 *T)
    sumR1  +=       6.670 * cos(1.65236689367  + 1898.35121793960 *T)
    sumR1  +=       7.591 * cos(0.10483002359  + 2324.94940881560 *T)
    sumR1  +=       5.881 * cos(1.04006410206  + 9992.87290377220 *T)
    sumR1  +=       5.609 * cos(4.83142709229  + 643.07868005170 *T)
    sumR1  +=       5.569 * cos(2.23863483508  + 1038.04128918680 *T)
    sumR1  +=       5.755 * cos(5.91598458372  + 6062.66320755260 *T)
    sumR1  +=       5.845 * cos(6.10234689502  + 209.10630974400 *T)
    sumR1  +=       5.577 * cos(0.81426649853  + 472.17484191470 *T)
    sumR1  +=       5.247 * cos(0.56496127013  + 192.69216761850 *T)
    sumR1  +=       5.493 * cos(5.81071309534  + 237.67811782620 *T)
    sumR1  +=       5.148 * cos(4.85160826999  + 267.47376618580 *T)
    sumR1  +=       6.122 * cos(2.11480301005  + 2097.42321937600 *T)
    sumR1  +=       6.188 * cos(4.59441762166  + 207.67002114550 *T)
    sumR1  +=       6.303 * cos(0.75806431119  + 210.37833413120 *T)
    sumR1  +=       5.102 * cos(4.01017179605  + 205.22234059070 *T)
    sumR1  +=       6.583 * cos(1.79054357427  + 12.53017297220 *T)
    sumR1  +=       4.902 * cos(0.85099521860  + 247.23934538180 *T)
    sumR1  +=       4.918 * cos(4.03512681632  + 487.36514376280 *T)
    sumR1  +=       5.818 * cos(5.48495503489  + 2538.24850425360 *T)
    sumR1  +=       4.855 * cos(4.18197778083  + 2744.43405269080 *T)
    sumR1  +=       4.885 * cos(0.25103933716  + 129.91947716160 *T)
    sumR1  +=       5.748 * cos(0.55968589618  + 116.42609634290 *T)
    sumR1  +=       4.901 * cos(4.48628916012  + 291.26208774300 *T)
    sumR1  +=       4.720 * cos(5.57686152365  + 342.25536475310 *T)
    sumR1  +=       5.962 * cos(5.12885837444  + 692.58748435350 *T)
    sumR1  +=       5.629 * cos(4.39847572369  + 196.62432088160 *T)
    sumR1  +=       5.596 * cos(0.94874135403  + 1802.37199072180 *T)
    sumR1  +=       6.197 * cos(3.80364010966  + 339.28641933650 *T)
    sumR1  +=       4.668 * cos(3.16816375033  + 148.07872442630 *T)
    sumR1  +=       4.891 * cos(2.67234862638  + 417.03696332040 *T)
    sumR1  +=       4.959 * cos(1.63453587065  + 166.82867252200 *T)
    sumR1  +=       4.408 * cos(4.95179678525  + 184.09414790940 *T)
    sumR1  +=       4.449 * cos(5.69134789394  + 252.65597135320 *T)
    sumR1  +=       4.943 * cos(0.85358212806  + 46.47042291600 *T)
    sumR1  +=       5.153 * cos(3.82176885491  + 842.15068148810 *T)
    sumR1  +=       5.930 * cos(5.95484153666  + 486.40193591630 *T)
    sumR1  +=       4.206 * cos(2.97664198894  + 380.12776796000 *T)
    sumR1  +=       4.467 * cos(0.24914978400  + 128.95626931510 *T)
    sumR1  +=       5.419 * cos(6.19106890918  + 337.73251065900 *T)
    sumR1  +=       4.499 * cos(4.71434958315  + 151.04766984290 *T)
    sumR1  +=       4.233 * cos(4.18702525973  + 685.47393735270 *T)
    sumR1  +=       4.695 * cos(1.54881559549  + 214.78356814630 *T)
    sumR1  +=       4.084 * cos(4.87173226400  + 14.97785352700 *T)
    sumR1  +=       4.321 * cos(5.42615168860  + 436.89313161450 *T)
    sumR1  +=       5.145 * cos(0.49931857511  + 248.72381809010 *T)
    sumR1  +=       3.897 * cos(0.74661138504  + 2627.11418447060 *T)
    sumR1  +=       3.995 * cos(3.07750371135  + 710.74673161820 *T)
    sumR1  +=       4.089 * cos(5.81996977038  + 491.81856188770 *T)
    sumR1  +=       4.532 * cos(3.67494714028  + 189.72322220190 *T)
    sumR1  +=       3.690 * cos(1.26565281569  + 211.81462272970 *T)
    sumR1  +=       4.036 * cos(1.15473702593  + 3053.71237534660 *T)
    sumR1  +=       3.672 * cos(4.52661018437  + 488.84961647110 *T)
    sumR1  +=       3.662 * cos(2.87243745783  + 411.62033734900 *T)
    sumR1  +=       3.653 * cos(3.06205147988  + 409.92341631960 *T)
    sumR1  +=       3.908 * cos(3.45947158106  + 220.46082654860 *T)
    sumR1  +=       4.989 * cos(3.36376245705  + 824.74219374880 *T)
    sumR1  +=       3.677 * cos(3.55713278092  + 244.31858407500 *T)
    sumR1  +=       3.580 * cos(1.57825591891  + 643.82943957710 *T)
    sumR1  +=       3.546 * cos(2.19846245030  + 135.33610313300 *T)
    sumR1  +=       3.560 * cos(4.51362022045  + 601.76425067620 *T)
    sumR1  +=       3.843 * cos(0.98567531677  + 271.40591944890 *T)
    sumR1  +=       3.559 * cos(1.11005765159  + 6283.07584999140 *T)
    sumR1  +=       4.266 * cos(6.19696005871  + 268.43697403230 *T)
    sumR1  +=       3.442 * cos(4.27628882392  + 867.42347575360 *T)
    sumR1  +=       4.844 * cos(3.73706907228  + 235.39049596580 *T)
    sumR1  +=       3.659 * cos(2.21859531609  + 2.92076130680 *T)
    sumR1  +=       3.958 * cos(5.17084570945  + 114.39910691340 *T)
    sumR1  +=       3.609 * cos(5.54387488088  + 458.84151979040 *T)
    sumR1  +=       4.470 * cos(3.74256930900  + 699.70103135430 *T)
    sumR1  +=       3.293 * cos(4.48068043469  + 289.56516671360 *T)
    sumR1  +=       3.240 * cos(5.94728881707  + 131.54696222180 *T)
    sumR1  +=       3.477 * cos(3.54553285172  + 963.40270297140 *T)
    sumR1  +=       3.838 * cos(4.77967877681  + 175.16605980020 *T)
    sumR1  +=       3.223 * cos(1.95410765469  + 212.02707105080 *T)
    sumR1  +=       4.053 * cos(4.19011281964  + 501.37978944330 *T)
    sumR1  +=       3.100 * cos(2.11956558345  + 916.93228005540 *T)
    sumR1  +=       3.183 * cos(1.93201605379  + 1354.43315884340 *T)
    sumR1  +=       3.301 * cos(1.80825506815  + 756.32338265690 *T)
    sumR1  +=       4.187 * cos(5.96622666047  + 212.54833591260 *T)
    sumR1  +=       3.716 * cos(3.70660462807  + 204.70107572890 *T)
    sumR1  +=       3.000 * cos(6.15443664698  + 3267.01147078460 *T)
    sumR1  +=       2.993 * cos(4.20888489881  + 533.62311835770 *T)
    sumR1  +=       4.125 * cos(6.09715151219  + 2641.34127847220 *T)
    sumR1  +=       3.145 * cos(2.55483540896  + 905.88657979150 *T)
    sumR1  +=       2.982 * cos(1.52760656472  + 945.99421523210 *T)
    sumR1  +=       3.015 * cos(1.76012152992  + 28.31117565130 *T)
    sumR1  +=       3.453 * cos(1.42473508236  + 2214.74308759620 *T)
    sumR1  +=       2.926 * cos(5.50138147476  + 24.37902238820 *T)
    sumR1  +=       2.978 * cos(4.27440059910  + 195.89060769870 *T)
    sumR1  +=       3.526 * cos(3.63935401565  + 229.97386999440 *T)
    sumR1  +=       2.860 * cos(4.52551886503  + 241.61027108930 *T)
    sumR1  +=       3.059 * cos(5.68165832697  + 282.66406803390 *T)
    sumR1  +=       3.415 * cos(5.26311934884  + 67.66805156650 *T)
    sumR1  +=       2.819 * cos(5.42053027567  + 305.08553696180 *T)
    sumR1  +=       3.503 * cos(1.31670335802  + 69.15252427480 *T)
    sumR1  +=       2.746 * cos(0.82597971627  + 444.75743814070 *T)
    sumR1  +=       2.796 * cos(0.07021047160  + 681.54178408960 *T)
    sumR1  +=       3.366 * cos(4.03843228994  + 6.15033915430 *T)
    sumR1  +=       3.242 * cos(2.63461047831  + 739.80866679490 *T)
    sumR1  +=       2.718 * cos(3.40899287465  + 188.92007304980 *T)
    sumR1  +=       2.741 * cos(3.22092213412  + 776.93031047640 *T)
    sumR1  +=       2.793 * cos(3.39766347322  + 431.26405732200 *T)
    sumR1  +=       2.966 * cos(3.91429372950  + 526.50957135690 *T)
    sumR1  +=       2.693 * cos(3.38996413068  + 778.41478318470 *T)
    sumR1  +=       2.680 * cos(3.82192393959  + 3060.82592234740 *T)
    sumR1  +=       2.954 * cos(2.69669880207  + 426.64637498580 *T)
    sumR1  +=       2.681 * cos(1.04615621583  + 28.45418800320 *T)
    sumR1  +=       3.182 * cos(2.72333374876  + 432.22726516850 *T)
    sumR1  +=       2.633 * cos(2.55029306465  + 10213.28554621100 *T)
    sumR1  +=       2.923 * cos(0.85695094024  + 2435.15573003500 *T)
    sumR1  +=       2.596 * cos(5.42890752137  + 207.14875628370 *T)
    sumR1  +=       3.225 * cos(0.96538615730  + 2118.76386037840 *T)
    sumR1  +=       2.774 * cos(0.33260844270  + 326.68681039510 *T)
    sumR1  +=       2.550 * cos(5.88893697427  + 439.12836384820 *T)
    sumR1  +=       2.716 * cos(3.15505406487  + 170.76082578510 *T)
    sumR1  +=       2.942 * cos(4.88555233562  + 397.39324334740 *T)
    sumR1  +=       3.121 * cos(1.87815629157  + 2413.81508903260 *T)
    sumR1  +=       3.263 * cos(2.59868619716  + 213.03846300710 *T)
    sumR1  +=       2.518 * cos(0.15471130491  + 945.24345570670 *T)
    sumR1  +=       3.169 * cos(5.70993651497  + 381.35160823740 *T)
    sumR1  +=       2.515 * cos(0.06248441393  + 427.56139872250 *T)
    sumR1  +=       3.279 * cos(4.95751323467  + 313.94418910180 *T)
    sumR1  +=       2.595 * cos(5.13169797457  + 299.12639426920 *T)
    sumR1  +=       2.572 * cos(3.42558391509  + 4.19278569400 *T)
    sumR1  +=       2.580 * cos(2.03280916494  + 319.31263096340 *T)
    sumR1  +=       3.294 * cos(6.24566168486  + 421.18156490460 *T)
    sumR1  +=       2.580 * cos(2.62721090534  + 213.18722085340 *T)
    sumR1  +=       2.879 * cos(0.45679876898  + 285.63301345050 *T)
    sumR1  +=       2.406 * cos(4.57473098758  + 228.27694896500 *T)
    sumR1  +=       2.518 * cos(2.55500085830  + 140.00196957900 *T)
    sumR1  +=       2.422 * cos(2.36310658303  + 84.34282612290 *T)
    sumR1  +=       2.374 * cos(2.25544718932  + 17.26547538740 *T)
    sumR1  +=       2.627 * cos(1.26370339212  + 724.83081326790 *T)
    sumR1  +=       2.346 * cos(3.77641630157  + 206.23373254700 *T)
    sumR1  +=       2.463 * cos(5.42094278240  + 395.57870223900 *T)
    sumR1  +=       2.352 * cos(0.63041319237  + 210.59078245230 *T)
    sumR1  +=       3.166 * cos(0.26273580642  + 201.51968199120 *T)
    sumR1  +=       2.405 * cos(0.78919759458  + 426.07692601420 *T)
    sumR1  +=       2.390 * cos(5.89523812458  + 738.79727483860 *T)
    sumR1  +=       2.515 * cos(0.70044371265  + 2943.50605412720 *T)
    sumR1  +=       2.332 * cos(4.06963624306  + 519.39602435610 *T)
    sumR1  +=       3.132 * cos(2.79331632190  + 732.69511979410 *T)
    sumR1  +=       2.658 * cos(3.34020209714  + 1141.13406340540 *T)
    sumR1  +=       2.258 * cos(0.12403309730  + 2524.02141025200 *T)
    sumR1  +=       2.697 * cos(2.58404587754  + 425.63498302950 *T)
    sumR1  +=       2.416 * cos(3.85003724506  + 696.51963761660 *T)
    sumR1  +=       2.597 * cos(2.54164936697  + 436.15941843160 *T)
    sumR1  +=       2.192 * cos(3.07202313269  + 203.26478713040 *T)
    sumR1  +=       2.424 * cos(2.60715310452  + 511.53171782990 *T)
    sumR1  +=       2.126 * cos(0.14811901148  + 405.99126305650 *T)
    sumR1  +=       2.306 * cos(1.25068142377  + 427.11945573780 *T)
    sumR1  +=       2.121 * cos(0.43505808954  + 184.98791978670 *T)
    sumR1  +=       2.755 * cos(3.02380019321  + 468.24268865160 *T)
    sumR1  +=       2.333 * cos(3.02634928771  + 216.00740842370 *T)
    sumR1  +=       2.182 * cos(4.27912012069  + 7.16173111060 *T)
    sumR1  +=       2.101 * cos(4.31781498012  + 572.22923474750 *T)
    sumR1  +=       2.362 * cos(4.82914341110  + 556.51766803760 *T)
    sumR1  +=       2.218 * cos(0.82936075453  + 3370.10424500320 *T)
    sumR1  +=       2.103 * cos(5.25950154713  + 661.23792731640 *T)
    sumR1  +=       2.580 * cos(1.03705340380  + 213.41097002260 *T)
    sumR1  +=       2.366 * cos(6.14368355608  + 205.43478891180 *T)
    sumR1  +=       2.042 * cos(0.21462094901  + 3259.89792378380 *T)
    sumR1  +=       2.547 * cos(4.69969204009  + 221.37585028530 *T)
    sumR1  +=       1.987 * cos(3.22670561632  + 1382.88734684660 *T)
    sumR1  +=       2.213 * cos(0.89932827487  + 286.59622129700 *T)
    sumR1  +=       2.191 * cos(0.08759174058  + 259.76951835400 *T)
    sumR1  +=       1.968 * cos(0.57824086026  + 180.16199464630 *T)
    sumR1  +=       2.037 * cos(2.35713099759  + 610.69233878540 *T)
    sumR1  +=       1.959 * cos(2.18553775379  + 72.07328558160 *T)
    sumR1  +=       2.061 * cos(1.68041202479  + 1670.82502850000 *T)
    sumR1  +=       1.940 * cos(0.62951066481  + 406.95447090300 *T)
    sumR1  +=       2.043 * cos(4.39130144045  + 576.16138801060 *T)
    sumR1  +=       1.936 * cos(1.05286934530  + 1262.38608488870 *T)
    sumR1  +=       1.975 * cos(0.31945835160  + 938.12990870590 *T)
    sumR1  +=       2.015 * cos(1.66410213484  + 193.65537546500 *T)
    sumR1  +=       1.971 * cos(0.72639439054  + 200.55647414470 *T)
    sumR1  +=       1.952 * cos(6.25320630177  + 241.75328344120 *T)
    sumR1  +=       1.976 * cos(1.31263772699  + 135.54855145410 *T)
    sumR1  +=       2.448 * cos(0.52850194172  + 429.51895218280 *T)
    sumR1  +=       1.977 * cos(3.13944703383  + 421.93232443000 *T)
    sumR1  +=       1.853 * cos(0.17184530353  + 196.03362005060 *T)
    sumR1  +=       2.552 * cos(5.39764879348  + 2854.64037391020 *T)
    sumR1  +=       1.830 * cos(1.47821899466  + 638.41281360570 *T)
    sumR1  +=       2.245 * cos(6.00427164270  + 230.70758317730 *T)
    sumR1  +=       1.822 * cos(6.08626100417  + 1261.63532536330 *T)
    sumR1  +=       2.168 * cos(0.41741136149  + 213.51154375910 *T)
    sumR1  +=       1.869 * cos(3.67791368036  + 403.02231763990 *T)
    sumR1  +=       1.866 * cos(1.59662677545  + 391.17346822390 *T)
    sumR1  +=       2.034 * cos(1.21814866092  + 3046.59882834580 *T)
    sumR1  +=       1.929 * cos(4.93193335031  + 420.96911658350 *T)
    sumR1  +=       1.746 * cos(5.09757251683  + 107.75864066460 *T)
    sumR1  +=       2.168 * cos(3.24685294764  + 213.08664711690 *T)
    sumR1  +=       2.178 * cos(5.09777299346  + 558.00214074590 *T)
    sumR1  +=       1.992 * cos(2.29524873043  + 1773.91780271860 *T)
    sumR1  +=       1.761 * cos(2.88655624670  + 141.69889060840 *T)
    sumR1  +=       1.769 * cos(5.47051542758  + 206.13736432740 *T)
    sumR1  +=       1.734 * cos(2.11941015901  + 430.79097657000 *T)
    sumR1  +=       2.377 * cos(1.07633521570  + 59.80374504030 *T)
    sumR1  +=       1.797 * cos(2.90984583978  + 92.79783348010 *T)
    sumR1  +=       1.725 * cos(5.22827286197  + 757.21715453420 *T)
    sumR1  +=       2.305 * cos(5.88235807192  + 426.55000676620 *T)
    sumR1  +=       1.751 * cos(5.28990803470  + 87.31177153950 *T)
    sumR1  +=       2.202 * cos(1.28096946505  + 624.91943278700 *T)
    sumR1  +=       2.043 * cos(0.46193065602  + 831.10498122420 *T)
    sumR1  +=       1.931 * cos(1.26974971942  + 219.89137757700 *T)
    sumR1  +=       1.953 * cos(2.96900002385  + 398.14400287280 *T)
    sumR1  +=       1.676 * cos(4.81683149512  + 181.80652604900 *T)
    sumR1  +=       1.902 * cos(2.74426125465  + 4952.06359328620 *T)
    sumR1  +=       2.133 * cos(5.37177705284  + 627.36711334180 *T)
    sumR1  +=       1.962 * cos(3.52111949662  + 213.45915413240 *T)
    sumR1  +=       1.709 * cos(6.14073761844  + 952.35700270750 *T)
    sumR1  +=       1.784 * cos(1.05243716682  + 353.30106501700 *T)
    sumR1  +=       1.700 * cos(1.17418864170  + 739.05790726950 *T)
    sumR1  +=       1.609 * cos(1.35009554392  + 84.93352695390 *T)
    sumR1  +=       2.038 * cos(2.47570829812  + 26.82670294300 *T)
    sumR1  +=       1.870 * cos(5.61729116529  + 2957.73314812880 *T)
    sumR1  +=       1.962 * cos(0.13564680851  + 213.13903674360 *T)
    sumR1  +=       2.041 * cos(3.31354526279  + 1596.18644228460 *T)
    sumR1  +=       1.612 * cos(6.19495100885  + 432.01481684740 *T)
    sumR1  +=       1.742 * cos(2.87947098602  + 179.35884549420 *T)
    sumR1  +=       1.964 * cos(2.84253666387  + 429.04587143080 *T)
    sumR1  +=       1.805 * cos(0.60932632638  + 532.61172640140 *T)
    sumR1  +=       1.647 * cos(0.82347900016  + 214.57111982520 *T)
    sumR1  +=       1.893 * cos(4.33962647901  + 173.94221952280 *T)
    sumR1  +=       1.689 * cos(1.13037158144  + 586.31331639720 *T)
    sumR1  +=       1.523 * cos(2.71561930244  + 73.29712585900 *T)
    sumR1  +=       1.524 * cos(5.26558677448  + 5429.87946823940 *T)
    sumR1  +=       1.582 * cos(2.79533721474  + 842.90144101350 *T)
    sumR1  +=       1.608 * cos(2.33230359324  + 418.52143602870 *T)
    sumR1  +=       1.579 * cos(1.15182102801  + 731.94436026870 *T)
    sumR1  +=       1.689 * cos(1.91915438546  + 630.33605875840 *T)
    sumR1  +=       1.990 * cos(5.23790221176  + 550.13783421970 *T)
    sumR1  +=       1.772 * cos(2.95372411478  + 172.24529849340 *T)
    sumR1  +=       1.596 * cos(0.99004701777  + 953.10776223290 *T)
    sumR1  +=       1.784 * cos(3.91391032360  + 159.12442469020 *T)
    sumR1  +=       1.592 * cos(2.99690086808  + 45.57665103870 *T)
    sumR1  +=       1.968 * cos(0.23073879009  + 220.36445832900 *T)
    sumR1  +=       1.549 * cos(5.88699595922  + 60.55450456570 *T)
    sumR1  +=       1.459 * cos(5.51999778036  + 273.85360000370 *T)
    sumR1  +=       1.909 * cos(2.78415262815  + 418.00017116690 *T)
    sumR1  +=       1.445 * cos(3.25530914937  + 453.42489381900 *T)
    sumR1  +=       1.454 * cos(0.16250693313  + 115.62294719080 *T)
    sumR1  +=       1.566 * cos(2.24077018103  + 1056.20053645150 *T)
    sumR1  +=       1.412 * cos(3.45442909885  + 354.99798604640 *T)
    sumR1  +=       1.564 * cos(3.38591337689  + 409.18970313670 *T)
    sumR1  +=       1.631 * cos(1.06286709889  + 213.55972786890 *T)
    sumR1  +=       1.415 * cos(1.32091877590  + 373.90799283650 *T)
    sumR1  +=       1.389 * cos(0.40584159469  + 9360.08916445900 *T)
    sumR1  +=       1.663 * cos(2.33357114562  + 188.02630117250 *T)
    sumR1  +=       1.426 * cos(5.44677783737  + 864.24208201590 *T)
    sumR1  +=       1.716 * cos(3.96056093028  + 1699.27921650320 *T)
    sumR1  +=       1.682 * cos(0.39747670320  + 17.40848773930 *T)
    sumR1  +=       1.368 * cos(5.83289186692  + 569.04784100980 *T)
    sumR1  +=       1.416 * cos(3.65464640816  + 6.85291456990 *T)
    sumR1  +=       1.469 * cos(3.49069193830  + 934.94851496820 *T)
    sumR1  +=       1.330 * cos(4.41794310534  + 3914.02230409940 *T)
    sumR1  +=       1.309 * cos(1.29979865382  + 428.08266358430 *T)
    sumR1  +=       1.300 * cos(1.57748627871  + 238.57188970350 *T)
    sumR1  +=       1.389 * cos(1.31202796503  + 6275.96230299060 *T)
    sumR1  +=       1.384 * cos(0.67585082323  + 2751.54759969160 *T)
    sumR1  +=       1.471 * cos(1.21149871903  + 2531.13495725280 *T)
    sumR1  +=       1.334 * cos(4.11154515525  + 206.93630796260 *T)
    sumR1  +=       1.601 * cos(0.93356520728  + 355.74874557180 *T)
    sumR1  +=       1.259 * cos(2.56678207309  + 850.01498801430 *T)
    sumR1  +=       1.277 * cos(0.41764451386  + 100.64509366380 *T)
    sumR1  +=       1.436 * cos(4.06045514506  + 177.87437278590 *T)
    sumR1  +=       1.308 * cos(1.01324076289  + 423.67742956920 *T)
    sumR1  +=       1.541 * cos(6.03020449918  + 292.01284726840 *T)
    sumR1  +=       1.307 * cos(5.83815678434  + 5863.59120611620 *T)
    sumR1  +=       1.613 * cos(2.45074803642  + 1049.08698945070 *T)
    sumR1  +=       1.249 * cos(3.01518429832  + 464.73122651380 *T)
    sumR1  +=       1.250 * cos(6.23516728885  + 823.99143422340 *T)
    sumR1  +=       1.275 * cos(2.68217384213  + 637.44960575920 *T)
    sumR1  +=       1.249 * cos(2.97028182853  + 51749.20809227239 *T)
    sumR1  +=       1.240 * cos(2.66940683813  + 2700.71514038580 *T)
    sumR1  +=       1.456 * cos(1.85558224828  + 96.87299909510 *T)
    sumR1  +=       1.491 * cos(4.98649587341  + 295.19424100610 *T)
    sumR1  +=       1.230 * cos(4.27283851216  + 12139.55350910680 *T)
    sumR1  +=       1.292 * cos(2.73196017809  + 10206.17199921020 *T)
    sumR1  +=       1.247 * cos(3.77399749791  + 504.56118318100 *T)
    sumR1  +=       1.408 * cos(1.02955773079  + 518.38463239980 *T)
    sumR1  +=       1.223 * cos(1.12202093840  + 221.16340196420 *T)
    sumR1  +=       1.420 * cos(4.39795293289  + 606.76018552230 *T)
    sumR1  +=       1.190 * cos(1.57292553631  + 820.05928096030 *T)
    sumR1  +=       1.247 * cos(0.99102599652  + 9793.80090233580 *T)
    sumR1  +=       1.234 * cos(1.10826361423  + 2303.60876781320 *T)
    sumR1  +=       1.186 * cos(4.55984967028  + 9808.53818466140 *T)
    sumR1  +=       1.346 * cos(4.94456950019  + 384.05992122310 *T)
    sumR1  +=       1.514 * cos(3.60392291730  + 2015.67108615980 *T)
    sumR1  +=       1.432 * cos(2.28704432909  + 525.49817940060 *T)
    sumR1  +=       1.129 * cos(0.87100340620  + 162.09337010680 *T)
    sumR1  +=       1.196 * cos(5.13485214850  + 227.31374111850 *T)
    sumR1  +=       1.339 * cos(2.48923887712  + 206.70681329900 *T)
    sumR1  +=       1.421 * cos(1.65379789078  + 857.12853501510 *T)
    sumR1  +=       1.162 * cos(1.92099315083  + 220.93390730060 *T)
    sumR1  +=       1.277 * cos(4.85435999187  + 54.17467074780 *T)
    sumR1  +=       1.153 * cos(5.33028034679  + 233.90602325750 *T)
    sumR1  +=       1.214 * cos(4.11324721963  + 3377.21779200400 *T)
    sumR1  +=       1.109 * cos(5.68915582674  + 162.89651925890 *T)
    sumR1  +=       1.068 * cos(4.85383480876  + 611.44309831080 *T)
    sumR1  +=       1.119 * cos(1.40805686363  + 1987.21689815660 *T)
    sumR1  +=       1.085 * cos(0.64208148190  + 731.68372783780 *T)
    sumR1  +=       1.435 * cos(3.20880139888  + 835.78789401270 *T)
    sumR1  +=       1.184 * cos(2.99776919968  + 199.28444975750 *T)
    sumR1  +=       1.281 * cos(3.12245339510  + 427.34895040140 *T)
    sumR1  +=       1.058 * cos(5.17851282929  + 306.09692891810 *T)
    sumR1  +=       1.152 * cos(4.39244449554  + 199.96577331370 *T)
    sumR1  +=       1.036 * cos(3.68027119804  + 597.35901666110 *T)
    sumR1  +=       1.055 * cos(3.25561743426  + 394.35486196160 *T)
    sumR1  +=       1.127 * cos(4.33255371960  + 552.58551477450 *T)
    sumR1  +=       1.213 * cos(6.21447612110  + 42.53826965290 *T)
    sumR1  +=       1.117 * cos(3.74367882111  + 214.19286731530 *T)
    sumR1  +=       1.023 * cos(3.84199833949  + 894.84087952760 *T)
    sumR1  +=       1.042 * cos(5.30120078590  + 450.97721326420 *T)
    sumR1  +=       1.290 * cos(3.96221234564  + 318.83955021140 *T)
    sumR1  +=       1.073 * cos(4.10012122884  + 188.16931352440 *T)
    sumR1  +=       1.204 * cos(0.37702365750  + 393.46109008430 *T)
    sumR1  +=       1.214 * cos(2.01826978554  + 401.32539661050 *T)
    sumR1  +=       1.018 * cos(0.02946649279  + 2840.41327990860 *T)
    sumR1  +=       1.237 * cos(5.41088851225  + 425.84743135060 *T)
    sumR1  +=       1.187 * cos(5.16511890602  + 838.21852822500 *T)
    sumR1  +=       1.276 * cos(2.93572146232  + 1457.52593306200 *T)
    sumR1  +=       0.994 * cos(3.40079885702  + 211.60217440860 *T)
    sumR1  +=       1.042 * cos(2.42209320898  + 361.37781986430 *T)
    sumR1  +=       1.093 * cos(3.66289018246  + 226.63241756230 *T)
    sumR1  +=       0.978 * cos(3.76334208607  + 5856.47765911540 *T)
    sumR1  +=       1.263 * cos(2.09195268609  + 78.71375183040 *T)
    sumR1  +=       1.009 * cos(5.85963705048  + 1268.74887236410 *T)
    sumR1  +=       1.148 * cos(4.39543895068  + 570.74476203920 *T)
    sumR1  +=       1.051 * cos(3.27272240682  + 153.49535039770 *T)
    sumR1  +=       0.975 * cos(3.42924642244  + 105.54045477340 *T)
    sumR1  +=       0.997 * cos(4.30943991893  + 212.40532356070 *T)
    sumR1  +=       0.954 * cos(3.88548755058  + 171.65459766240 *T)
    sumR1  +=       0.960 * cos(1.90180005280  + 1159.29331067010 *T)
    sumR1  +=       0.953 * cos(3.40787141587  + 244.79166482700 *T)
    sumR1  +=       0.969 * cos(1.93369993197  + 525.75881183150 *T)
    sumR1  +=       0.918 * cos(1.73738789723  + 223.33340374560 *T)
    sumR1  +=       1.164 * cos(5.05392864346  + 263.70167161710 *T)
    sumR1  +=       0.951 * cos(4.23581224839  + 92.94084583200 *T)
    sumR1  +=       1.160 * cos(5.80630916592  + 460.53844081980 *T)
    sumR1  +=       1.186 * cos(4.46262000755  + 465.95506679120 *T)
    sumR1  +=       0.931 * cos(2.09868057209  + 205.66428357540 *T)
    sumR1  +=       0.942 * cos(3.86810837922  + 238.42887735160 *T)
    sumR1  +=       1.020 * cos(5.53181822898  + 0.04818410980 *T)
    sumR1  +=       0.995 * cos(2.03457885490  + 6290.18939699220 *T)
    sumR1  +=       0.888 * cos(2.60957592990  + 1912.57831194120 *T)
    sumR1  +=       0.873 * cos(5.78433393020  + 480.77286162380 *T)
    sumR1  +=       0.854 * cos(1.63255087291  + 328.24071907260 *T)
    sumR1  +=       1.123 * cos(4.07401922216  + 3693.60966166060 *T)
    sumR1  +=       0.934 * cos(3.52355235083  + 10220.39909321180 *T)
    sumR1  +=       0.833 * cos(3.03302227840  + 532.87235883230 *T)
    sumR1  +=       1.007 * cos(2.73615455688  + 4841.85727206680 *T)
    sumR1  +=       0.870 * cos(1.06968760644  + 51.20572533120 *T)
    sumR1  +=       0.891 * cos(1.36817544763  + 700.45179087970 *T)
    sumR1  +=       0.833 * cos(5.39754715806  + 159.71512552120 *T)
    sumR1  +=       0.854 * cos(1.91765015557  + 622.48879857470 *T)
    sumR1  +=       0.976 * cos(3.09106001923  + 2332.06295581640 *T)
    sumR1  +=       0.819 * cos(5.55683690482  + 462.02291352810 *T)
    sumR1  +=       0.800 * cos(1.47042677460  + 969.62247809490 *T)
    sumR1  +=       0.933 * cos(1.40166917666  + 287.93768165340 *T)
    sumR1  +=       0.784 * cos(1.69235162770  + 477.80391620720 *T)
    sumR1  +=       0.782 * cos(3.98153416585  + 702.14871190910 *T)
    sumR1  +=       0.810 * cos(5.87689161549  + 561.93429400900 *T)
    sumR1  +=       0.858 * cos(4.02964773169  + 41.64449777560 *T)
    sumR1  +=       0.819 * cos(0.98885784755  + 960.22130923370 *T)
    sumR1  +=       0.882 * cos(1.49559638306  + 760.25553592000 *T)
    sumR1  +=       0.767 * cos(2.46787654531  + 402.21916848780 *T)
    sumR1  +=       0.851 * cos(3.64678001195  + 348.63519857100 *T)
    sumR1  +=       0.852 * cos(1.03470840672  + 2620.00063746980 *T)
    sumR1  +=       0.770 * cos(4.67090683753  + 16.67477455640 *T)
    sumR1  +=       0.849 * cos(5.27994730935  + 74.63858621540 *T)
    sumR1  +=       0.854 * cos(2.56811257488  + 432.74853003030 *T)
    sumR1  +=       0.767 * cos(1.11806243753  + 2847.52682690940 *T)
    sumR1  +=       0.729 * cos(0.44171990710  + 898.77303279070 *T)
    sumR1  +=       0.776 * cos(5.54603607568  + 3171.03224356680 *T)
    sumR1  +=       0.721 * cos(6.05392551158  + 91.78644152380 *T)
    sumR1  +=       0.739 * cos(3.57172839746  + 775.23338944700 *T)
    sumR1  +=       0.730 * cos(4.98865345688  + 705.11765732570 *T)
    sumR1  +=       0.705 * cos(0.44942445750  + 219.66188291340 *T)
    sumR1  +=       0.708 * cos(0.69014726046  + 1048.33622992530 *T)
    sumR1  +=       0.706 * cos(2.22974712805  + 29.20494752860 *T)
    sumR1  +=       0.711 * cos(0.00981284716  + 2115.58246664070 *T)
    sumR1  +=       0.722 * cos(4.14075205197  + 225.07850888480 *T)
    sumR1  +=       0.735 * cos(1.54083195463  + 201.99276274320 *T)
    sumR1  +=       0.746 * cos(4.08997409526  + 849.26422848890 *T)
    sumR1  +=       0.729 * cos(2.81779134868  + 419.53282798500 *T)
    sumR1  +=       0.837 * cos(1.00844734156  + 4127.32139953740 *T)
    sumR1  +=       0.939 * cos(0.46037763932  + 5488.86810538160 *T)
    sumR1  +=       0.939 * cos(1.85473712038  + 5062.26991450560 *T)
    sumR1  +=       0.721 * cos(1.62872794201  + 2200.51599359460 *T)

    # SATURN Radius
    sumR2  = 0
    sumR2  +=  436902.464 * cos(4.78671673044  + 213.29909543800 *T)
    sumR2  +=   71922.760 * cos(2.50069994874  + 206.18554843720 *T)
    sumR2  +=   49766.792 * cos(4.97168150870  + 220.41264243880 *T)
    sumR2  +=   43220.894 * cos(3.86940443794  + 426.59819087600 *T)
    sumR2  +=   29645.554 * cos(5.96310264282  + 7.11354700080 *T)
    sumR2  +=    4141.650 * cos(4.10670940823  + 433.71173787680 *T)
    sumR2  +=    4720.909 * cos(2.47527992423  + 199.07200143640 *T)
    sumR2  +=    3789.370 * cos(3.09771025067  + 639.89728631400 *T)
    sumR2  +=    2963.990 * cos(1.37206248846  + 103.09277421860 *T)
    sumR2  +=    2556.363 * cos(2.85065721526  + 419.48464387520 *T)
    sumR2  +=    2208.457 * cos(6.27588858707  + 110.20632121940 *T)
    sumR2  +=    2187.621 * cos(5.85545832218  + 14.22709400160 *T)
    sumR2  +=    1956.896 * cos(4.92448618045  + 227.52618943960 *T)
    sumR2  +=    2326.801 * cos(0.00000000000  + 0.00000000000 *T)
    sumR2  +=     923.840 * cos(5.46392422737  + 323.50541665740 *T)
    sumR2  +=     705.936 * cos(2.97081280098  + 95.97922721780 *T)
    sumR2  +=     546.115 * cos(4.12854181522  + 412.37109687440 *T)
    sumR2  +=     373.838 * cos(5.83435991809  + 117.31986822020 *T)
    sumR2  +=     360.882 * cos(3.27703082368  + 647.01083331480 *T)
    sumR2  +=     356.350 * cos(3.19152043942  + 210.11770170030 *T)
    sumR2  +=     390.627 * cos(4.48106176893  + 216.48048917570 *T)
    sumR2  +=     431.485 * cos(5.17825414612  + 522.57741809380 *T)
    sumR2  +=     325.598 * cos(2.26867601656  + 853.19638175200 *T)
    sumR2  +=     405.018 * cos(4.17294157872  + 209.36694217490 *T)
    sumR2  +=     204.494 * cos(0.08774848590  + 202.25339517410 *T)
    sumR2  +=     206.854 * cos(4.02188336738  + 735.87651353180 *T)
    sumR2  +=     178.474 * cos(4.09716541453  + 440.82528487760 *T)
    sumR2  +=     180.143 * cos(3.59704903955  + 632.78373931320 *T)
    sumR2  +=     153.656 * cos(3.13470530382  + 625.67019231240 *T)
    sumR2  +=     147.779 * cos(0.13614300541  + 302.16477565500 *T)
    sumR2  +=     123.189 * cos(4.18895309647  + 88.86568021700 *T)
    sumR2  +=     133.076 * cos(2.59350469420  + 191.95845443560 *T)
    sumR2  +=     100.367 * cos(5.46056190585  + 3.18139373770 *T)
    sumR2  +=     131.975 * cos(5.93293968941  + 309.27832265580 *T)
    sumR2  +=      97.235 * cos(4.01832604356  + 728.76296653100 *T)
    sumR2  +=     110.709 * cos(4.77853798276  + 838.96928775040 *T)
    sumR2  +=     119.053 * cos(5.55385105975  + 224.34479570190 *T)
    sumR2  +=      93.852 * cos(4.38395529912  + 217.23124870110 *T)
    sumR2  +=     108.701 * cos(5.29310899841  + 515.46387109300 *T)
    sumR2  +=      78.609 * cos(5.72525447528  + 21.34064100240 *T)
    sumR2  +=      81.468 * cos(5.10897365253  + 956.28915597060 *T)
    sumR2  +=      96.412 * cos(6.25859229567  + 742.99006053260 *T)
    sumR2  +=      69.228 * cos(4.04901237761  + 3.93215326310 *T)
    sumR2  +=      65.168 * cos(3.77713343518  + 1052.26838318840 *T)
    sumR2  +=      64.088 * cos(5.81235002453  + 529.69096509460 *T)
    sumR2  +=      62.541 * cos(2.18445116349  + 195.13984817330 *T)
    sumR2  +=      56.987 * cos(3.14666549033  + 203.00415469950 *T)
    sumR2  +=      55.979 * cos(4.84108422860  + 234.63973644040 *T)
    sumR2  +=      52.940 * cos(5.07780548444  + 330.61896365820 *T)
    sumR2  +=      50.635 * cos(2.77318570728  + 942.06206196900 *T)
    sumR2  +=      41.649 * cos(4.79014211005  + 63.73589830340 *T)
    sumR2  +=      44.858 * cos(0.56460613593  + 269.92144674060 *T)
    sumR2  +=      41.357 * cos(3.73496404402  + 316.39186965660 *T)
    sumR2  +=      52.847 * cos(3.92623831484  + 949.17560896980 *T)
    sumR2  +=      38.398 * cos(3.73966157784  + 1045.15483618760 *T)
    sumR2  +=      37.583 * cos(4.18924633757  + 536.80451209540 *T)
    sumR2  +=      35.285 * cos(2.90795856092  + 284.14854074220 *T)
    sumR2  +=      33.576 * cos(3.80465978802  + 149.56319713460 *T)
    sumR2  +=      41.073 * cos(4.57870454147  + 1155.36115740700 *T)
    sumR2  +=      30.412 * cos(2.48140171991  + 860.30992875280 *T)
    sumR2  +=      31.373 * cos(4.84075951849  + 1272.68102562720 *T)
    sumR2  +=      30.218 * cos(4.35186294470  + 405.25754987360 *T)
    sumR2  +=      39.430 * cos(3.50858482049  + 422.66603761290 *T)
    sumR2  +=      29.658 * cos(1.58886982096  + 1066.49547719000 *T)
    sumR2  +=      35.202 * cos(5.94478241578  + 1059.38193018920 *T)
    sumR2  +=      25.829 * cos(3.54946335477  + 1368.66025284500 *T)
    sumR2  +=      26.283 * cos(4.81567477177  + 124.43341522100 *T)
    sumR2  +=      29.963 * cos(3.66312205813  + 429.77958461370 *T)
    sumR2  +=      33.011 * cos(4.96879544579  + 831.85574074960 *T)
    sumR2  +=      24.305 * cos(5.31133255082  + 10.29494073850 *T)
    sumR2  +=      26.332 * cos(4.45253273390  + 223.59403617650 *T)
    sumR2  +=      22.108 * cos(2.76092021113  + 415.55249061210 *T)
    sumR2  +=      27.187 * cos(1.66347897738  + 277.03499374140 *T)
    sumR2  +=      21.639 * cos(1.03836302307  + 11.04570026390 *T)
    sumR2  +=      19.713 * cos(2.52194629263  + 1258.45393162560 *T)
    sumR2  +=      17.062 * cos(3.27669927228  + 654.12438031560 *T)
    sumR2  +=      17.261 * cos(3.49414816663  + 1361.54670584420 *T)
    sumR2  +=      16.097 * cos(1.73396878598  + 490.33408917940 *T)
    sumR2  +=      21.099 * cos(3.62102032955  + 1265.56747862640 *T)
    sumR2  +=      17.692 * cos(4.31141612385  + 1471.75302706360 *T)
    sumR2  +=      13.458 * cos(0.32327889681  + 295.05122865420 *T)
    sumR2  +=      12.586 * cos(3.13794576887  + 74.78159856730 *T)
    sumR2  +=      12.023 * cos(2.32917797741  + 210.85141488320 *T)
    sumR2  +=      15.120 * cos(3.59558424278  + 265.98929347750 *T)
    sumR2  +=      12.959 * cos(4.62359706368  + 1589.07289528380 *T)
    sumR2  +=      15.424 * cos(5.01335704925  + 127.47179660680 *T)
    sumR2  +=      11.193 * cos(4.54981248285  + 81.75213321620 *T)
    sumR2  +=      13.449 * cos(4.88710089777  + 437.64389113990 *T)
    sumR2  +=      10.673 * cos(5.05234757424  + 191.20769491020 *T)
    sumR2  +=      13.963 * cos(3.04990968366  + 423.41679713830 *T)
    sumR2  +=      10.614 * cos(5.02845923229  + 137.03302416240 *T)
    sumR2  +=      14.382 * cos(4.68720080027  + 1148.24761040620 *T)
    sumR2  +=      13.470 * cos(1.90280407135  + 408.43894361130 *T)
    sumR2  +=      10.077 * cos(5.20426583827  + 340.77089204480 *T)
    sumR2  +=      10.323 * cos(3.34460279759  + 1685.05212250160 *T)
    sumR2  +=       9.563 * cos(3.17317920222  + 351.81659230870 *T)
    sumR2  +=      11.295 * cos(5.47808960704  + 1375.77379984580 *T)
    sumR2  +=       8.617 * cos(2.81294528041  + 99.91138048090 *T)
    sumR2  +=       8.460 * cos(3.22691940753  + 1677.93857550080 *T)
    sumR2  +=       7.914 * cos(2.35624291874  + 1574.84580128220 *T)
    sumR2  +=       7.587 * cos(6.08171425316  + 231.45834270270 *T)
    sumR2  +=       9.175 * cos(3.40072244924  + 1581.95934828300 *T)
    sumR2  +=       7.337 * cos(2.00393601815  + 131.40394986990 *T)
    sumR2  +=       8.240 * cos(4.04095881407  + 1788.14489672020 *T)
    sumR2  +=       7.579 * cos(3.68311134272  + 846.08283475120 *T)
    sumR2  +=       6.691 * cos(4.37253800717  + 145.63104387150 *T)
    sumR2  +=       7.539 * cos(3.29482043104  + 750.10360753340 *T)
    sumR2  +=       6.367 * cos(4.00239137708  + 447.93883187840 *T)
    sumR2  +=       6.249 * cos(4.55603671940  + 106.27416795630 *T)
    sumR2  +=       6.489 * cos(1.33782087599  + 215.74677599280 *T)
    sumR2  +=       6.501 * cos(3.78204726337  + 313.21047591890 *T)
    sumR2  +=       5.978 * cos(0.55276980086  + 18.15924726470 *T)
    sumR2  +=       6.171 * cos(2.84712795642  + 138.51749687070 *T)
    sumR2  +=       6.837 * cos(4.83481646949  + 319.57326339430 *T)
    sumR2  +=       6.678 * cos(5.43046031699  + 508.35032409220 *T)
    sumR2  +=       7.175 * cos(4.37855723752  + 1464.63948006280 *T)
    sumR2  +=       5.753 * cos(4.14268749228  + 543.91805909620 *T)
    sumR2  +=       5.727 * cos(4.35383078313  + 1905.46476494040 *T)
    sumR2  +=       5.101 * cos(2.63866058897  + 288.08069400530 *T)
    sumR2  +=       5.311 * cos(3.62520849510  + 6076.89030155420 *T)
    sumR2  +=       5.498 * cos(4.19972735173  + 721.64941953020 *T)
    sumR2  +=       5.089 * cos(5.04845206653  + 10007.09999777380 *T)
    sumR2  +=       5.505 * cos(1.13479635941  + 56.62235130260 *T)
    sumR2  +=       4.820 * cos(3.30043078578  + 76.26607127560 *T)
    sumR2  +=       4.915 * cos(6.17790518458  + 483.22054217860 *T)
    sumR2  +=       5.048 * cos(2.44627820757  + 628.85158605010 *T)
    sumR2  +=       4.534 * cos(1.19648682598  + 200.76892246580 *T)
    sumR2  +=       4.817 * cos(3.11549733365  + 2001.44399215820 *T)
    sumR2  +=       4.712 * cos(1.26507812515  + 6062.66320755260 *T)
    sumR2  +=       4.811 * cos(5.78388270496  + 184.84490743480 *T)
    sumR2  +=       4.775 * cos(0.76197795755  + 333.65734504400 *T)
    sumR2  +=       4.514 * cos(0.95293919611  + 343.21857259960 *T)
    sumR2  +=       4.525 * cos(2.68827745072  + 9992.87290377220 *T)
    sumR2  +=       4.378 * cos(0.80241129896  + 222.86032299360 *T)
    sumR2  +=       4.873 * cos(5.92092913946  + 618.55664531160 *T)
    sumR2  +=       4.142 * cos(1.91878383159  + 497.44763618020 *T)
    sumR2  +=       5.112 * cos(4.50449287745  + 416.30325013750 *T)
    sumR2  +=       4.125 * cos(1.98204847532  + 347.88443904560 *T)
    sumR2  +=       4.045 * cos(2.87666810085  + 38.13303563780 *T)
    sumR2  +=       4.133 * cos(2.90478811425  + 107.02492748170 *T)
    sumR2  +=       4.035 * cos(2.92972681787  + 1994.33044515740 *T)
    sumR2  +=       4.916 * cos(3.12316267561  + 1898.35121793960 *T)
    sumR2  +=       3.657 * cos(3.24680246734  + 362.86229257260 *T)
    sumR2  +=       3.753 * cos(0.87719890943  + 703.63318461740 *T)
    sumR2  +=       3.576 * cos(3.48080143501  + 388.46515523820 *T)
    sumR2  +=       3.555 * cos(4.08436297683  + 430.53034413910 *T)
    sumR2  +=       3.598 * cos(0.05255328597  + 32.24332891440 *T)
    sumR2  +=       3.561 * cos(5.46414552453  + 6283.07584999140 *T)
    sumR2  +=       3.480 * cos(1.81622589595  + 70.84944530420 *T)
    sumR2  +=       3.827 * cos(3.12041228490  + 635.96513305090 *T)
    sumR2  +=       3.399 * cos(0.54882815021  + 10213.28554621100 *T)
    sumR2  +=       3.399 * cos(3.51833356080  + 629.60234557550 *T)
    sumR2  +=       3.364 * cos(3.27821747958  + 357.44566660120 *T)
    sumR2  +=       3.260 * cos(1.97623748027  + 203.73786788240 *T)
    sumR2  +=       3.118 * cos(2.18465627368  + 1891.23767093880 *T)
    sumR2  +=       3.163 * cos(1.26040995242  + 134.58534360760 *T)
    sumR2  +=       4.004 * cos(5.45434102599  + 1692.16566950240 *T)
    sumR2  +=       3.180 * cos(2.46319174788  + 867.42347575360 *T)
    sumR2  +=       3.389 * cos(4.20503159673  + 337.73251065900 *T)
    sumR2  +=       3.026 * cos(2.19331614526  + 217.96496188400 *T)
    sumR2  +=       3.573 * cos(5.55097240810  + 113.38771495710 *T)
    sumR2  +=       3.682 * cos(3.78966280284  + 2104.53676637680 *T)
    sumR2  +=       3.125 * cos(4.09203641264  + 1478.86657406440 *T)
    sumR2  +=       2.881 * cos(3.90810650240  + 312.19908396260 *T)
    sumR2  +=       3.199 * cos(3.92123638342  + 1038.04128918680 *T)
    sumR2  +=       4.014 * cos(5.17826893553  + 404.50679034820 *T)
    sumR2  +=       3.907 * cos(4.11767191780  + 1781.03134971940 *T)
    sumR2  +=       3.144 * cos(1.61185684069  + 1073.60902419080 *T)
    sumR2  +=       3.072 * cos(5.00675625396  + 312.45971639350 *T)
    sumR2  +=       3.034 * cos(5.46288652854  + 258.87574647670 *T)
    sumR2  +=       2.884 * cos(2.38477237305  + 181.05576652360 *T)
    sumR2  +=       2.986 * cos(0.88783591586  + 1279.79457262800 *T)
    sumR2  +=       2.683 * cos(0.00956197492  + 195.89060769870 *T)
    sumR2  +=       3.081 * cos(5.60034737330  + 216.21985674480 *T)
    sumR2  +=       2.626 * cos(6.12701960244  + 273.10284047830 *T)
    sumR2  +=       2.665 * cos(2.31576422128  + 565.11568774670 *T)
    sumR2  +=       3.245 * cos(3.87540558646  + 85.82729883120 *T)
    sumR2  +=       2.740 * cos(5.73784096806  + 160.60889739850 *T)
    sumR2  +=       2.876 * cos(4.74720607366  + 213.25091132820 *T)
    sumR2  +=       2.523 * cos(5.30458920892  + 444.75743814070 *T)
    sumR2  +=       2.752 * cos(5.08984539930  + 1169.58825140860 *T)
    sumR2  +=       2.889 * cos(1.66674437398  + 213.34727954780 *T)
    sumR2  +=       2.923 * cos(4.21481009033  + 650.94298657790 *T)
    sumR2  +=       3.036 * cos(2.55426675350  + 6069.77675455340 *T)
    sumR2  +=       3.116 * cos(2.67220972004  + 52.69019803950 *T)
    sumR2  +=       2.371 * cos(0.89591351822  + 121.25202148330 *T)
    sumR2  +=       2.993 * cos(3.96957827454  + 9999.98645077300 *T)
    sumR2  +=       3.088 * cos(0.40656113014  + 561.18353448360 *T)
    sumR2  +=       2.385 * cos(4.74063881551  + 218.71572140940 *T)
    sumR2  +=       2.632 * cos(1.12706218927  + 344.70304530790 *T)
    sumR2  +=       2.316 * cos(4.08445262041  + 131.54696222180 *T)
    sumR2  +=       2.214 * cos(3.37726228553  + 22.09140052780 *T)
    sumR2  +=       2.129 * cos(3.32497715011  + 358.93013930950 *T)
    sumR2  +=       2.679 * cos(1.68971401870  + 208.63322899200 *T)
    sumR2  +=       2.607 * cos(5.10250482155  + 824.74219374880 *T)
    sumR2  +=       2.250 * cos(2.60474848767  + 305.34616939270 *T)
    sumR2  +=       2.087 * cos(3.37293958793  + 320.32402291970 *T)
    sumR2  +=       2.693 * cos(3.62159456470  + 436.89313161450 *T)
    sumR2  +=       2.492 * cos(2.96129217279  + 2214.74308759620 *T)
    sumR2  +=       2.704 * cos(2.88483697319  + 643.07868005170 *T)
    sumR2  +=       2.124 * cos(1.61210282593  + 218.92816973050 *T)
    sumR2  +=       2.037 * cos(4.63481160778  + 188.02630117250 *T)
    sumR2  +=       2.394 * cos(3.46386258552  + 6275.96230299060 *T)
    sumR2  +=       1.973 * cos(2.28886138203  + 2627.11418447060 *T)
    sumR2  +=       1.937 * cos(5.67082364247  + 28.45418800320 *T)
    sumR2  +=       1.920 * cos(4.25647211328  + 546.95644048200 *T)
    sumR2  +=       2.498 * cos(3.57572154405  + 2420.92863603340 *T)
    sumR2  +=       1.898 * cos(1.30987536388  + 212.33588759150 *T)
    sumR2  +=       1.852 * cos(1.58508015515  + 424.15051032120 *T)
    sumR2  +=       1.850 * cos(3.57830449726  + 329.72519178090 *T)
    sumR2  +=       2.128 * cos(3.95329215734  + 1795.25844372100 *T)
    sumR2  +=       2.236 * cos(4.22073549375  + 2221.85663459700 *T)
    sumR2  +=       1.933 * cos(1.68771499202  + 350.33211960040 *T)
    sumR2  +=       1.799 * cos(2.06541260431  + 144.14657116320 *T)
    sumR2  +=       1.904 * cos(4.60953896857  + 182.27960680100 *T)
    sumR2  +=       2.236 * cos(5.17945392885  + 99.16062095550 *T)
    sumR2  +=       1.755 * cos(2.73425330428  + 291.26208774300 *T)
    sumR2  +=       2.231 * cos(5.42548168745  + 207.88246946660 *T)
    sumR2  +=       1.848 * cos(2.24194286719  + 168.05251279940 *T)
    sumR2  +=       1.726 * cos(1.31878655393  + 219.44943459230 *T)
    sumR2  +=       1.709 * cos(5.55913931846  + 92.79783348010 *T)
    sumR2  +=       1.693 * cos(1.95360003617  + 129.91947716160 *T)
    sumR2  +=       2.064 * cos(4.84900344498  + 1141.13406340540 *T)
    sumR2  +=       1.758 * cos(5.05088656436  + 214.26230328450 *T)
    sumR2  +=       1.781 * cos(2.85880153340  + 636.71589257630 *T)
    sumR2  +=       1.900 * cos(2.90295578617  + 2310.72231481400 *T)
    sumR2  +=       1.759 * cos(5.34657858395  + 45.57665103870 *T)
    sumR2  +=       1.654 * cos(6.14450664508  + 554.06998748280 *T)
    sumR2  +=       1.578 * cos(4.50941374663  + 210.37833413120 *T)
    sumR2  +=       1.681 * cos(3.55136706992  + 1354.43315884340 *T)
    sumR2  +=       1.862 * cos(3.01276783582  + 2317.83586181480 *T)
    sumR2  +=       1.589 * cos(1.15773448350  + 235.39049596580 *T)
    sumR2  +=       1.551 * cos(2.15558953807  + 207.67002114550 *T)
    sumR2  +=       1.874 * cos(4.12861627986  + 225.82926841020 *T)
    sumR2  +=       1.621 * cos(3.29992957653  + 1670.82502850000 *T)
    sumR2  +=       1.911 * cos(0.17724319140  + 12.53017297220 *T)
    sumR2  +=       1.477 * cos(5.90270260570  + 1.48447270830 *T)
    sumR2  +=       1.618 * cos(5.72513459206  + 1485.98012106520 *T)
    sumR2  +=       1.446 * cos(1.78104589920  + 1382.88734684660 *T)
    sumR2  +=       1.683 * cos(3.43534671475  + 2428.04218303420 *T)
    sumR2  +=       1.542 * cos(5.51223038941  + 204.70107572890 *T)
    sumR2  +=       1.420 * cos(2.07339356364  + 198.32124191100 *T)
    sumR2  +=       1.444 * cos(5.56032454849  + 128.36556848410 *T)
    sumR2  +=       1.476 * cos(6.12782257368  + 212.77783057620 *T)
    sumR2  +=       1.474 * cos(0.33626790634  + 213.82036029980 *T)
    sumR2  +=       1.428 * cos(3.25039966249  + 945.99421523210 *T)
    sumR2  +=       1.410 * cos(0.68747644676  + 429.04587143080 *T)
    sumR2  +=       1.752 * cos(2.70090942746  + 12.74262129330 *T)
    sumR2  +=       1.681 * cos(4.97526853273  + 2008.55753915900 *T)
    sumR2  +=       1.408 * cos(0.80461100746  + 1585.14074202070 *T)
    sumR2  +=       1.485 * cos(0.49674043855  + 120.35824960600 *T)
    sumR2  +=       1.490 * cos(2.68459799437  + 207.14875628370 *T)
    sumR2  +=       1.411 * cos(4.36399216092  + 5863.59120611620 *T)
    sumR2  +=       1.315 * cos(4.73430848989  + 241.75328344120 *T)
    sumR2  +=       1.516 * cos(4.99488503706  + 1162.47470440780 *T)
    sumR2  +=       1.310 * cos(1.98714265058  + 563.63121503840 *T)
    sumR2  +=       1.286 * cos(2.12891372062  + 251.43213107580 *T)
    sumR2  +=       1.271 * cos(5.70165238307  + 2.92076130680 *T)
    sumR2  +=       1.312 * cos(1.68811514551  + 2207.62954059540 *T)
    sumR2  +=       1.259 * cos(0.35924965717  + 334.55111692130 *T)
    sumR2  +=       1.252 * cos(2.14513440216  + 1055.44977692610 *T)
    sumR2  +=       1.401 * cos(6.13250261735  + 1802.37199072180 *T)
    sumR2  +=       1.343 * cos(5.79995727295  + 9793.80090233580 *T)
    sumR2  +=       1.228 * cos(3.29059284057  + 661.23792731640 *T)
    sumR2  +=       1.202 * cos(2.88792018909  + 2413.81508903260 *T)
    sumR2  +=       1.286 * cos(5.72360160371  + 298.23262239190 *T)
    sumR2  +=       1.357 * cos(0.93175963411  + 217.49188113200 *T)
    sumR2  +=       1.356 * cos(2.28121627817  + 601.76425067620 *T)
    sumR2  +=       1.190 * cos(1.94993809928  + 501.37978944330 *T)
    sumR2  +=       1.304 * cos(0.37337923280  + 3473.19701922180 *T)
    sumR2  +=       1.350 * cos(2.87235622320  + 142.44965013380 *T)
    sumR2  +=       1.349 * cos(3.21102203937  + 175.16605980020 *T)
    sumR2  +=       1.312 * cos(3.70149813509  + 2111.65031337760 *T)
    sumR2  +=       1.129 * cos(1.08860603834  + 842.15068148810 *T)
    sumR2  +=       1.237 * cos(0.08698781252  + 526.50957135690 *T)
    sumR2  +=       1.217 * cos(3.89835349840  + 209.10630974400 *T)
    sumR2  +=       1.467 * cos(1.16228775027  + 621.73803904930 *T)
    sumR2  +=       1.044 * cos(0.30512759901  + 436.15941843160 *T)
    sumR2  +=       1.140 * cos(5.33720637097  + 114.13847448250 *T)
    sumR2  +=       1.295 * cos(4.70261675421  + 9786.68735533500 *T)
    sumR2  +=       1.037 * cos(4.07846687083  + 156.67674413540 *T)
    sumR2  +=       1.391 * cos(4.73554028436  + 398.14400287280 *T)
    sumR2  +=       1.167 * cos(5.68899703631  + 479.28838891550 *T)
    sumR2  +=       1.035 * cos(5.34279429465  + 327.43756992050 *T)
    sumR2  +=       0.997 * cos(1.19323192891  + 710.74673161820 *T)
    sumR2  +=       1.193 * cos(5.17722376816  + 98.89998852460 *T)
    sumR2  +=       1.165 * cos(4.58588490135  + 732.69511979410 *T)
    sumR2  +=       1.161 * cos(4.90854984994  + 10206.17199921020 *T)
    sumR2  +=       1.144 * cos(0.50394784140  + 3906.90875709860 *T)
    sumR2  +=       1.182 * cos(3.69482624364  + 2854.64037391020 *T)
    sumR2  +=       0.970 * cos(2.89031410383  + 1987.21689815660 *T)
    sumR2  +=       1.039 * cos(0.48694895443  + 525.49817940060 *T)
    sumR2  +=       1.079 * cos(3.61750956217  + 2097.42321937600 *T)
    sumR2  +=       1.148 * cos(3.31015591733  + 5856.47765911540 *T)
    sumR2  +=       1.241 * cos(4.31971543677  + 230.70758317730 *T)
    sumR2  +=       0.910 * cos(4.59825926062  + 380.12776796000 *T)
    sumR2  +=       0.907 * cos(1.34912454077  + 685.47393735270 *T)
    sumR2  +=       1.166 * cos(1.61085609717  + 5849.36411211460 *T)
    sumR2  +=       0.882 * cos(6.12045540405  + 519.39602435610 *T)
    sumR2  +=       0.963 * cos(4.96065454054  + 699.70103135430 *T)
    sumR2  +=       1.062 * cos(5.13323858077  + 2751.54759969160 *T)
    sumR2  +=       0.865 * cos(6.12821112133  + 245.54242435240 *T)
    sumR2  +=       1.100 * cos(2.18435744407  + 1699.27921650320 *T)
    sumR2  +=       0.822 * cos(5.55083534581  + 739.05790726950 *T)
    sumR2  +=       0.926 * cos(2.01158276144  + 417.03696332040 *T)
    sumR2  +=       0.813 * cos(5.18401872205  + 214.78356814630 *T)
    sumR2  +=       1.033 * cos(5.48677848094  + 3995.77443731560 *T)
    sumR2  +=       0.872 * cos(3.02363724703  + 306.09692891810 *T)
    sumR2  +=       0.796 * cos(0.44343664540  + 486.40193591630 *T)
    sumR2  +=       0.878 * cos(1.82164034386  + 135.33610313300 *T)
    sumR2  +=       0.791 * cos(2.14989417962  + 2620.00063746980 *T)
    sumR2  +=       0.881 * cos(2.39697554334  + 289.56516671360 *T)
    sumR2  +=       0.782 * cos(4.50471317138  + 980.66817835880 *T)
    sumR2  +=       0.783 * cos(1.14229319753  + 540.73666535850 *T)
    sumR2  +=       0.831 * cos(0.69937251013  + 421.93232443000 *T)
    sumR2  +=       0.770 * cos(2.40292326155  + 576.16138801060 *T)
    sumR2  +=       0.950 * cos(5.97621460162  + 196.62432088160 *T)
    sumR2  +=       0.814 * cos(4.19303098086  + 831.10498122420 *T)
    sumR2  +=       0.969 * cos(4.78071024754  + 326.68681039510 *T)
    sumR2  +=       0.760 * cos(0.44860533530  + 425.63498302950 *T)
    sumR2  +=       0.907 * cos(0.94781730418  + 525.75881183150 *T)
    sumR2  +=       0.788 * cos(0.14287187051  + 916.93228005540 *T)
    sumR2  +=       0.801 * cos(1.86383119100  + 3039.48528134500 *T)
    sumR2  +=       0.801 * cos(0.46947170994  + 3466.08347222100 *T)
    sumR2  +=       0.747 * cos(6.05374861925  + 211.81462272970 *T)
    sumR2  +=       0.968 * cos(3.02618272726  + 2634.22773147140 *T)
    sumR2  +=       0.739 * cos(2.27110740297  + 2303.60876781320 *T)
    sumR2  +=       0.750 * cos(5.48554383902  + 173.94221952280 *T)
    sumR2  +=       1.024 * cos(1.91564925560  + 229.97386999440 *T)
    sumR2  +=       0.816 * cos(4.98990432666  + 4209.07353275360 *T)
    sumR2  +=       0.728 * cos(1.30997967935  + 511.53171782990 *T)
    sumR2  +=       0.716 * cos(3.74192651696  + 3053.71237534660 *T)
    sumR2  +=       0.727 * cos(0.39191881243  + 1493.09366806600 *T)
    sumR2  +=       0.717 * cos(2.68899513085  + 228.27694896500 *T)
    sumR2  +=       0.739 * cos(2.12749199443  + 1176.70179840940 *T)
    sumR2  +=       0.805 * cos(0.07187193910  + 556.51766803760 *T)
    sumR2  +=       0.835 * cos(3.48287855700  + 84.93352695390 *T)
    sumR2  +=       0.790 * cos(0.48073040004  + 4017.11507831800 *T)
    sumR2  +=       0.725 * cos(1.96643065215  + 220.46082654860 *T)
    sumR2  +=       0.683 * cos(2.68825142163  + 151.04766984290 *T)
    sumR2  +=       0.739 * cos(3.33688408107  + 953.10776223290 *T)
    sumR2  +=       0.745 * cos(6.22304530635  + 1269.49963188950 *T)

    # SATURN Radius
    sumR3  = 0
    sumR3  +=   20315.005 * cos(3.02186626038  + 213.29909543800 *T)
    sumR3  +=    8923.581 * cos(3.19144205755  + 220.41264243880 *T)
    sumR3  +=    6908.677 * cos(4.35174889353  + 206.18554843720 *T)
    sumR3  +=    4087.129 * cos(4.22406927376  + 7.11354700080 *T)
    sumR3  +=    3879.041 * cos(2.01056445995  + 426.59819087600 *T)
    sumR3  +=    1070.788 * cos(4.20360341236  + 199.07200143640 *T)
    sumR3  +=     907.332 * cos(2.28344368029  + 433.71173787680 *T)
    sumR3  +=     606.121 * cos(3.17458570534  + 227.52618943960 *T)
    sumR3  +=     596.639 * cos(4.13455753351  + 14.22709400160 *T)
    sumR3  +=     483.181 * cos(1.17345973258  + 639.89728631400 *T)
    sumR3  +=     393.174 * cos(0.00000000000  + 0.00000000000 *T)
    sumR3  +=     229.472 * cos(4.69838526383  + 419.48464387520 *T)
    sumR3  +=     188.250 * cos(4.59003889007  + 110.20632121940 *T)
    sumR3  +=     149.508 * cos(3.20199444400  + 103.09277421860 *T)
    sumR3  +=     121.442 * cos(3.76831374104  + 323.50541665740 *T)
    sumR3  +=     101.215 * cos(5.81884137755  + 412.37109687440 *T)
    sumR3  +=     102.146 * cos(4.70974422803  + 95.97922721780 *T)
    sumR3  +=      93.078 * cos(1.43531270909  + 647.01083331480 *T)
    sumR3  +=      72.601 * cos(4.15395598507  + 117.31986822020 *T)
    sumR3  +=      84.347 * cos(2.63462379693  + 216.48048917570 *T)
    sumR3  +=      62.198 * cos(2.31239345505  + 440.82528487760 *T)
    sumR3  +=      45.145 * cos(4.37317047297  + 191.95845443560 *T)
    sumR3  +=      49.536 * cos(2.38854232908  + 209.36694217490 *T)
    sumR3  +=      54.829 * cos(0.30526468471  + 853.19638175200 *T)
    sumR3  +=      40.498 * cos(1.83836569765  + 302.16477565500 *T)
    sumR3  +=      38.089 * cos(5.94455115525  + 88.86568021700 *T)
    sumR3  +=      32.243 * cos(4.01146349387  + 21.34064100240 *T)
    sumR3  +=      40.671 * cos(0.68845183210  + 522.57741809380 *T)
    sumR3  +=      28.209 * cos(5.77193013961  + 210.11770170030 *T)
    sumR3  +=      24.976 * cos(3.06249709014  + 234.63973644040 *T)
    sumR3  +=      20.824 * cos(4.92570695678  + 625.67019231240 *T)
    sumR3  +=      25.070 * cos(0.73137425284  + 515.46387109300 *T)
    sumR3  +=      17.485 * cos(5.73135068691  + 728.76296653100 *T)
    sumR3  +=      18.009 * cos(1.45593152612  + 309.27832265580 *T)
    sumR3  +=      16.927 * cos(3.52771580455  + 3.18139373770 *T)
    sumR3  +=      13.437 * cos(3.36479898106  + 330.61896365820 *T)
    sumR3  +=      11.090 * cos(3.37212682914  + 224.34479570190 *T)
    sumR3  +=      11.082 * cos(3.41719974793  + 956.28915597060 *T)
    sumR3  +=       9.978 * cos(1.58791582772  + 202.25339517410 *T)
    sumR3  +=      11.551 * cos(5.99093726182  + 735.87651353180 *T)
    sumR3  +=      10.500 * cos(6.06911092266  + 405.25754987360 *T)
    sumR3  +=       9.144 * cos(2.93557421591  + 124.43341522100 *T)
    sumR3  +=       8.737 * cos(4.65432480769  + 632.78373931320 *T)
    sumR3  +=      10.023 * cos(0.58247011625  + 860.30992875280 *T)
    sumR3  +=       7.482 * cos(4.50669216436  + 942.06206196900 *T)
    sumR3  +=      10.091 * cos(0.28268774007  + 838.96928775040 *T)
    sumR3  +=       9.243 * cos(2.57034547708  + 223.59403617650 *T)
    sumR3  +=       8.652 * cos(1.75808100881  + 429.77958461370 *T)
    sumR3  +=       7.564 * cos(1.45635107202  + 654.12438031560 *T)
    sumR3  +=       7.058 * cos(5.47394786065  + 1045.15483618760 *T)
    sumR3  +=       6.970 * cos(1.51811695028  + 422.66603761290 *T)
    sumR3  +=       8.067 * cos(4.48457709292  + 742.99006053260 *T)
    sumR3  +=       6.817 * cos(4.83084424818  + 316.39186965660 *T)
    sumR3  +=       7.693 * cos(0.43769724671  + 831.85574074960 *T)
    sumR3  +=       7.934 * cos(4.20112367712  + 195.13984817330 *T)
    sumR3  +=       6.119 * cos(2.33960392135  + 269.92144674060 *T)
    sumR3  +=       5.589 * cos(1.14518720694  + 284.14854074220 *T)
    sumR3  +=       5.564 * cos(4.18123189068  + 529.69096509460 *T)
    sumR3  +=       5.034 * cos(2.12020038657  + 295.05122865420 *T)
    sumR3  +=       6.556 * cos(3.42459866876  + 10.29494073850 *T)
    sumR3  +=       5.544 * cos(2.46823271699  + 536.80451209540 *T)
    sumR3  +=       6.189 * cos(6.01433827520  + 1066.49547719000 *T)
    sumR3  +=       5.649 * cos(0.82784598388  + 217.23124870110 *T)
    sumR3  +=       4.264 * cos(3.23245736673  + 1272.68102562720 *T)
    sumR3  +=       4.450 * cos(0.92477808590  + 203.00415469950 *T)
    sumR3  +=       3.268 * cos(4.32777516976  + 1258.45393162560 *T)
    sumR3  +=       3.655 * cos(0.05832123987  + 81.75213321620 *T)
    sumR3  +=       3.951 * cos(0.11124996745  + 1155.36115740700 *T)
    sumR3  +=       3.773 * cos(6.01157059552  + 1052.26838318840 *T)
    sumR3  +=       2.915 * cos(5.64342950039  + 3.93215326310 *T)
    sumR3  +=       3.019 * cos(2.19411778004  + 447.93883187840 *T)
    sumR3  +=       2.977 * cos(1.89387342550  + 149.56319713460 *T)
    sumR3  +=       3.146 * cos(0.19215180096  + 1148.24761040620 *T)
    sumR3  +=       2.763 * cos(0.92363342001  + 508.35032409220 *T)
    sumR3  +=       2.790 * cos(4.97199778427  + 1677.93857550080 *T)
    sumR3  +=       2.608 * cos(2.99591016813  + 1589.07289528380 *T)
    sumR3  +=       2.881 * cos(5.40535671721  + 1361.54670584420 *T)
    sumR3  +=       2.337 * cos(1.30362271569  + 184.84490743480 *T)
    sumR3  +=       2.536 * cos(3.71412120849  + 408.43894361130 *T)
    sumR3  +=       2.450 * cos(3.22118361135  + 319.57326339430 *T)
    sumR3  +=       2.585 * cos(2.31346415454  + 543.91805909620 *T)
    sumR3  +=       2.324 * cos(5.87500715503  + 721.64941953020 *T)
    sumR3  +=       1.990 * cos(0.51565577383  + 416.30325013750 *T)
    sumR3  +=       2.490 * cos(4.24017800021  + 1059.38193018920 *T)
    sumR3  +=       1.935 * cos(2.41463084855  + 337.73251065900 *T)
    sumR3  +=       1.886 * cos(0.53809070779  + 635.96513305090 *T)
    sumR3  +=       1.893 * cos(5.62352727352  + 11.04570026390 *T)
    sumR3  +=       2.389 * cos(5.73399981234  + 313.21047591890 *T)
    sumR3  +=       1.900 * cos(2.41000566465  + 131.54696222180 *T)
    sumR3  +=       1.743 * cos(4.57646237847  + 1994.33044515740 *T)
    sumR3  +=       1.913 * cos(5.17436386408  + 2854.64037391020 *T)
    sumR3  +=       1.946 * cos(6.23355845623  + 1471.75302706360 *T)
    sumR3  +=       1.963 * cos(6.17814558628  + 1464.63948006280 *T)
    sumR3  +=       1.838 * cos(5.59464577559  + 1038.04128918680 *T)
    sumR3  +=       1.541 * cos(0.60765337379  + 210.85141488320 *T)
    sumR3  +=       1.617 * cos(1.75479346067  + 195.89060769870 *T)
    sumR3  +=       1.577 * cos(0.55789908488  + 2324.94940881560 *T)
    sumR3  +=       1.492 * cos(0.26624235633  + 497.44763618020 *T)
    sumR3  +=       1.659 * cos(2.57526072926  + 2090.30967237520 *T)
    sumR3  +=       1.809 * cos(1.82317819973  + 436.89313161450 *T)
    sumR3  +=       1.566 * cos(6.15328100324  + 490.33408917940 *T)
    sumR3  +=       1.771 * cos(6.11741716855  + 1073.60902419080 *T)
    sumR3  +=       1.456 * cos(0.85374460914  + 415.55249061210 *T)
    sumR3  +=       1.645 * cos(2.95335775161  + 437.64389113990 *T)
    sumR3  +=       1.391 * cos(4.12025028560  + 1574.84580128220 *T)
    sumR3  +=       1.585 * cos(5.96841377266  + 1781.03134971940 *T)
    sumR3  +=       1.507 * cos(3.84895122542  + 1251.34038462480 *T)
    sumR3  +=       1.442 * cos(5.32547705924  + 2538.24850425360 *T)
    sumR3  +=       1.805 * cos(1.50973093681  + 750.10360753340 *T)
    sumR3  +=       1.462 * cos(3.28599831588  + 1884.12412393800 *T)
    sumR3  +=       1.482 * cos(0.99340744053  + 643.07868005170 *T)
    sumR3  +=       1.312 * cos(3.79347668996  + 1567.73225428140 *T)
    sumR3  +=       1.665 * cos(0.02551523913  + 423.41679713830 *T)
    sumR3  +=       1.469 * cos(5.35285153471  + 1354.43315884340 *T)
    sumR3  +=       1.352 * cos(0.69945139243  + 867.42347575360 *T)
    sumR3  +=       1.124 * cos(1.79624810407  + 618.55664531160 *T)
    sumR3  +=       1.126 * cos(4.70052329245  + 113.38771495710 *T)
    sumR3  +=       1.122 * cos(3.95537224270  + 1891.23767093880 *T)
    sumR3  +=       1.458 * cos(1.50198846753  + 430.53034413910 *T)
    sumR3  +=       1.145 * cos(5.13093399117  + 25.27279426550 *T)
    sumR3  +=       1.178 * cos(2.97062300389  + 241.75328344120 *T)
    sumR3  +=       1.274 * cos(2.29089799814  + 2420.92863603340 *T)
    sumR3  +=       1.071 * cos(0.04888943982  + 63.73589830340 *T)
    sumR3  +=       1.377 * cos(5.58271514873  + 1382.88734684660 *T)
    sumR3  +=       1.145 * cos(3.10797488346  + 2200.51599359460 *T)
    sumR3  +=       1.076 * cos(0.79465514815  + 127.47179660680 *T)
    sumR3  +=       1.046 * cos(5.85060227045  + 215.74677599280 *T)
    sumR3  +=       1.082 * cos(3.72589445510  + 131.40394986990 *T)
    sumR3  +=       1.218 * cos(0.47504349592  + 824.74219374880 *T)
    sumR3  +=       1.116 * cos(3.78039049056  + 1375.77379984580 *T)
    sumR3  +=       0.969 * cos(5.90752273481  + 265.98929347750 *T)
    sumR3  +=       1.230 * cos(1.41325962069  + 2634.22773147140 *T)
    sumR3  +=       1.070 * cos(4.80334493874  + 1987.21689815660 *T)
    sumR3  +=       0.946 * cos(6.25968535931  + 2015.67108615980 *T)
    sumR3  +=       1.030 * cos(1.08973644893  + 362.86229257260 *T)
    sumR3  +=       1.072 * cos(5.41838042079  + 1279.79457262800 *T)
    sumR3  +=       0.880 * cos(1.92224908504  + 483.22054217860 *T)
    sumR3  +=       0.878 * cos(2.96591300878  + 934.94851496820 *T)
    sumR3  +=       0.879 * cos(2.65659265685  + 145.63104387150 *T)
    sumR3  +=       0.872 * cos(6.26261969664  + 2.44768055480 *T)
    sumR3  +=       1.082 * cos(4.48298283322  + 2214.74308759620 *T)
    sumR3  +=       0.959 * cos(0.74479087918  + 16.67477455640 *T)
    sumR3  +=       1.035 * cos(4.05664979327  + 231.45834270270 *T)
    sumR3  +=       0.851 * cos(0.09360495322  + 628.85158605010 *T)
    sumR3  +=       0.888 * cos(5.98816755324  + 2524.02141025200 *T)
    sumR3  +=       0.866 * cos(3.16259265630  + 2207.62954059540 *T)
    sumR3  +=       0.843 * cos(1.23731248821  + 74.78159856730 *T)
    sumR3  +=       0.809 * cos(2.89742868175  + 2008.55753915900 *T)
    sumR3  +=       0.779 * cos(2.28434811609  + 1478.86657406440 *T)
    sumR3  +=       0.990 * cos(5.32604038017  + 2428.04218303420 *T)
    sumR3  +=       0.795 * cos(2.38178135810  + 2228.97018159780 *T)
    sumR3  +=       0.765 * cos(4.70033674940  + 1670.82502850000 *T)
    sumR3  +=       1.024 * cos(4.23352869513  + 1802.37199072180 *T)
    sumR3  +=       0.831 * cos(5.87457134912  + 1368.66025284500 *T)
    sumR3  +=       0.717 * cos(5.92144324994  + 1685.05212250160 *T)
    sumR3  +=       0.772 * cos(1.15596098579  + 3053.71237534660 *T)
    sumR3  +=       0.691 * cos(3.13193109668  + 56.62235130260 *T)

    # SATURN Radius
    sumR4  = 0
    sumR4  +=    1202.050 * cos(1.41499446465  + 220.41264243880 *T)
    sumR4  +=     707.796 * cos(1.16153570102  + 213.29909543800 *T)
    sumR4  +=     516.121 * cos(6.23973568330  + 206.18554843720 *T)
    sumR4  +=     426.664 * cos(2.46924890293  + 7.11354700080 *T)
    sumR4  +=     267.736 * cos(0.18659206741  + 426.59819087600 *T)
    sumR4  +=     170.171 * cos(5.95926972384  + 199.07200143640 *T)
    sumR4  +=     145.113 * cos(1.44211060143  + 227.52618943960 *T)
    sumR4  +=     150.339 * cos(0.47970167140  + 433.71173787680 *T)
    sumR4  +=     121.033 * cos(2.40527320817  + 14.22709400160 *T)
    sumR4  +=      47.332 * cos(5.56857488676  + 639.89728631400 *T)
    sumR4  +=      15.745 * cos(2.90112466278  + 110.20632121940 *T)
    sumR4  +=      16.668 * cos(0.52920774279  + 440.82528487760 *T)
    sumR4  +=      18.954 * cos(5.85626429118  + 647.01083331480 *T)
    sumR4  +=      14.074 * cos(1.30343550656  + 412.37109687440 *T)
    sumR4  +=      12.708 * cos(2.09349305926  + 323.50541665740 *T)
    sumR4  +=      14.724 * cos(0.29905316786  + 419.48464387520 *T)
    sumR4  +=      11.133 * cos(2.46304825990  + 117.31986822020 *T)
    sumR4  +=      11.320 * cos(0.21785507019  + 95.97922721780 *T)
    sumR4  +=       9.233 * cos(2.28127318068  + 21.34064100240 *T)
    sumR4  +=       9.246 * cos(1.56496312830  + 88.86568021700 *T)
    sumR4  +=       8.970 * cos(0.68301278041  + 216.48048917570 *T)
    sumR4  +=       7.674 * cos(3.59367715368  + 302.16477565500 *T)
    sumR4  +=       7.823 * cos(4.48688804175  + 853.19638175200 *T)
    sumR4  +=       8.360 * cos(1.27239488455  + 234.63973644040 *T)
    sumR4  +=       9.552 * cos(3.14159265359  + 0.00000000000 *T)
    sumR4  +=       4.834 * cos(2.58836294602  + 515.46387109300 *T)
    sumR4  +=       6.059 * cos(5.16774448740  + 103.09277421860 *T)
    sumR4  +=       4.410 * cos(0.02211643085  + 191.95845443560 *T)
    sumR4  +=       4.364 * cos(1.59622746023  + 330.61896365820 *T)
    sumR4  +=       3.676 * cos(3.29899839673  + 210.11770170030 *T)
    sumR4  +=       4.364 * cos(5.97349927933  + 654.12438031560 *T)
    sumR4  +=       4.447 * cos(4.97415112184  + 860.30992875280 *T)
    sumR4  +=       3.220 * cos(2.72684237392  + 522.57741809380 *T)
    sumR4  +=       4.005 * cos(1.59858435636  + 405.25754987360 *T)
    sumR4  +=       3.099 * cos(0.75235436533  + 209.36694217490 *T)
    sumR4  +=       2.464 * cos(1.19167306488  + 124.43341522100 *T)
    sumR4  +=       3.088 * cos(1.32258934286  + 728.76296653100 *T)
    sumR4  +=       2.220 * cos(3.28087994088  + 203.00415469950 *T)
    sumR4  +=       2.127 * cos(6.14648095022  + 429.77958461370 *T)
    sumR4  +=       2.110 * cos(0.75462855247  + 295.05122865420 *T)
    sumR4  +=       2.020 * cos(3.89394929749  + 1066.49547719000 *T)
    sumR4  +=       2.248 * cos(0.49319150178  + 447.93883187840 *T)
    sumR4  +=       2.180 * cos(0.72761059998  + 625.67019231240 *T)
    sumR4  +=       1.809 * cos(0.09057839517  + 942.06206196900 *T)
    sumR4  +=       1.672 * cos(1.39635398184  + 224.34479570190 *T)
    sumR4  +=       1.641 * cos(3.02468307550  + 184.84490743480 *T)
    sumR4  +=       1.772 * cos(0.81879250825  + 223.59403617650 *T)
    sumR4  +=       1.902 * cos(2.00472814984  + 831.85574074960 *T)
    sumR4  +=       1.600 * cos(5.41185167676  + 824.74219374880 *T)
    sumR4  +=       1.505 * cos(5.95520747253  + 422.66603761290 *T)
    sumR4  +=       1.133 * cos(1.11512973946  + 838.96928775040 *T)
    sumR4  +=       1.190 * cos(1.89600567803  + 956.28915597060 *T)
    sumR4  +=       1.487 * cos(2.11906469507  + 529.69096509460 *T)
    sumR4  +=       1.409 * cos(0.72254420236  + 536.80451209540 *T)
    sumR4  +=       1.125 * cos(0.89062692183  + 721.64941953020 *T)
    sumR4  +=       1.301 * cos(1.64867038984  + 17.40848773930 *T)
    sumR4  +=       1.164 * cos(5.96957981840  + 195.13984817330 *T)
    sumR4  +=       0.950 * cos(5.36080713290  + 316.39186965660 *T)
    sumR4  +=       0.985 * cos(3.05768671768  + 1574.84580128220 *T)
    sumR4  +=       1.050 * cos(1.59202481523  + 735.87651353180 *T)
    sumR4  +=       0.817 * cos(4.92838813598  + 56.62235130260 *T)
    sumR4  +=       0.780 * cos(2.72125404102  + 508.35032409220 *T)
    sumR4  +=       0.969 * cos(1.00708261792  + 1045.15483618760 *T)
    sumR4  +=       0.716 * cos(1.11042181341  + 1169.58825140860 *T)

    # SATURN Radius
    sumR5  = 0
    sumR5  +=     128.612 * cos(5.91282565136  + 220.41264243880 *T)
    sumR5  +=      32.273 * cos(0.69256228602  + 7.11354700080 *T)
    sumR5  +=      26.698 * cos(5.91428528629  + 227.52618943960 *T)
    sumR5  +=      19.923 * cos(0.67370653385  + 14.22709400160 *T)
    sumR5  +=      20.223 * cos(4.95136801768  + 433.71173787680 *T)
    sumR5  +=      13.537 * cos(1.45669521408  + 199.07200143640 *T)
    sumR5  +=      14.097 * cos(2.67074280191  + 206.18554843720 *T)
    sumR5  +=      13.364 * cos(4.58826996370  + 426.59819087600 *T)
    sumR5  +=       7.257 * cos(4.62966127155  + 213.29909543800 *T)
    sumR5  +=       4.876 * cos(3.61448275002  + 639.89728631400 *T)
    sumR5  +=       3.136 * cos(4.65661021909  + 191.95845443560 *T)
    sumR5  +=       2.917 * cos(0.48665273315  + 323.50541665740 *T)
    sumR5  +=       3.759 * cos(4.89624165044  + 440.82528487760 *T)
    sumR5  +=       3.303 * cos(4.07190859545  + 647.01083331480 *T)
    sumR5  +=       2.883 * cos(3.18003019204  + 419.48464387520 *T)
    sumR5  +=       2.338 * cos(3.69553554327  + 88.86568021700 *T)
    sumR5  +=       1.950 * cos(5.32729247780  + 302.16477565500 *T)
    sumR5  +=       2.052 * cos(3.31663577368  + 95.97922721780 *T)
    sumR5  +=       1.591 * cos(2.67009215574  + 853.19638175200 *T)
    sumR5  +=       2.028 * cos(0.56025552769  + 117.31986822020 *T)
    sumR5  +=       1.560 * cos(0.85608042681  + 515.46387109300 *T)
    sumR5  +=       1.678 * cos(0.00000000000  + 0.00000000000 *T)
    sumR5  +=       1.102 * cos(5.98011943842  + 3.18139373770 *T)
    sumR5  +=       1.285 * cos(5.82563377753  + 234.63973644040 *T)
    sumR5  +=       0.896 * cos(5.22791858719  + 216.48048917570 *T)
    sumR5  +=       1.141 * cos(0.15741228205  + 412.37109687440 *T)
    sumR5  +=       0.798 * cos(0.37452846153  + 28.45418800320 *T)
    sumR5  +=       0.837 * cos(5.04769794123  + 124.43341522100 *T)

    # calculate and scale the sums
    L = sumL0 + sumL1*T**1 + sumL2*T**2 + sumL3*T**3 + sumL4*T**4 + sumL5*T**5
    L /= 1e8
    L = Mod(degrees(L), 360.)
    B = sumB0 + sumB1*T**1 + sumB2*T**2 + sumB3*T**3 + sumB4*T**4 + sumB5*T**5
    B /= 1e8
    B = degrees(B)
    R = sumR0 + sumR1*T**1 + sumR2*T**2 + sumR3*T**3 + sumR4*T**4 + sumR5*T**5
    R /= 1e8

    # calculate X, Y, Z
    X = R * Cos(B) * Cos(L)
    Y = R * Cos(B) * Sin(L)
    Z = R * Sin(B)

    # create dictionary
    Dict = {  "Object"  : "Saturn",
              "Mode"    : "LBR",
              "Long"    : L,
              "Lat"     : B,
              "Rau"     : R,
              "X"       : X,
              "Y"       : Y,
              "Z"       : Z,
              "Alt"     : -34./60.
           }
    Data = (L, B, R, Dict)


    return Data
#######################################################################



def getSaturnData(dtCivicTime, zone, verbose = False):
    L, B, R, Dict = getSaturnLBRData(dtCivicTime, zone, verbose)
    X = Dict["X"]; Y = Dict["Y"]; Z = Dict["Z"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Le, Be, Re, Dicte = getEarthLBRData(dtCivicTime, zone, verbose)
        Xe = Dicte["X"]; Ye = Dicte["Y"]; Ze = Dicte["Z"]
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getUranusLBRData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Uranus"
#       Long  :  heliocentric Longitude, degrees
#       Lat   :  heliocentric Latitude,degrees
#       Rau   :  Radius, distaance Sun - Object (au)
#
#   Similarly for each planet
#   data from Meeus, Appendix II and VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.2 hvf 2023-04-06 getUranusLBRData.py created by mkPlanetLBRData.py"
VersionList.append(Version)

def getUranusLBRData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. (31.1) Julian Millennia

#######################################################################

    # URANUS Longitude
    sumL0  = 0
    sumL0  +=  548129294.299 * cos(0.00000000000  + 0.00000000000 *T)
    sumL0  +=  9260408.252 * cos(0.89106421530  + 74.78159856730 *T)
    sumL0  +=  1504247.826 * cos(3.62719262195  + 1.48447270830 *T)
    sumL0  +=  365981.718 * cos(1.89962189068  + 73.29712585900 *T)
    sumL0  +=  272328.132 * cos(3.35823710524  + 149.56319713460 *T)
    sumL0  +=   70328.499 * cos(5.39254431993  + 63.73589830340 *T)
    sumL0  +=   68892.609 * cos(6.09292489045  + 76.26607127560 *T)
    sumL0  +=   61998.592 * cos(2.26952040469  + 2.96894541660 *T)
    sumL0  +=   61950.714 * cos(2.85098907565  + 11.04570026390 *T)
    sumL0  +=   26468.869 * cos(3.14152087888  + 71.81265315070 *T)
    sumL0  +=   25710.505 * cos(6.11379842935  + 454.90936652730 *T)
    sumL0  +=   21078.897 * cos(4.36059465144  + 148.07872442630 *T)
    sumL0  +=   17818.665 * cos(1.74436982544  + 36.64856292950 *T)
    sumL0  +=   14613.471 * cos(4.73732047977  + 3.93215326310 *T)
    sumL0  +=   11162.535 * cos(5.82681993692  + 224.34479570190 *T)
    sumL0  +=   10997.934 * cos(0.48865493179  + 138.51749687070 *T)
    sumL0  +=    9527.487 * cos(2.95516893093  + 35.16409022120 *T)
    sumL0  +=    7545.543 * cos(5.23626440666  + 109.94568878850 *T)
    sumL0  +=    4220.170 * cos(3.23328535514  + 70.84944530420 *T)
    sumL0  +=    4051.850 * cos(2.27754158724  + 151.04766984290 *T)
    sumL0  +=    3354.607 * cos(1.06549008887  + 4.45341812490 *T)
    sumL0  +=    2926.671 * cos(4.62903695486  + 9.56122755560 *T)
    sumL0  +=    3490.352 * cos(5.48305567292  + 146.59425171800 *T)
    sumL0  +=    3144.093 * cos(4.75199307603  + 77.75054398390 *T)
    sumL0  +=    2922.410 * cos(5.35236743380  + 85.82729883120 *T)
    sumL0  +=    2272.790 * cos(4.36600802756  + 70.32818044240 *T)
    sumL0  +=    2051.209 * cos(1.51773563459  + 0.11187458460 *T)
    sumL0  +=    2148.599 * cos(0.60745800902  + 38.13303563780 *T)
    sumL0  +=    1991.726 * cos(4.92437290826  + 277.03499374140 *T)
    sumL0  +=    1376.208 * cos(2.04281409054  + 65.22037101170 *T)
    sumL0  +=    1666.910 * cos(3.62744580852  + 380.12776796000 *T)
    sumL0  +=    1284.183 * cos(3.11346336879  + 202.25339517410 *T)
    sumL0  +=    1150.416 * cos(0.93344454002  + 3.18139373770 *T)
    sumL0  +=    1533.223 * cos(2.58593414266  + 52.69019803950 *T)
    sumL0  +=    1281.641 * cos(0.54269869505  + 222.86032299360 *T)
    sumL0  +=    1372.100 * cos(4.19641615561  + 111.43016149680 *T)
    sumL0  +=    1220.998 * cos(0.19901396193  + 108.46121608020 *T)
    sumL0  +=     946.195 * cos(1.19249463066  + 127.47179660680 *T)
    sumL0  +=    1150.993 * cos(4.17898207045  + 33.67961751290 *T)
    sumL0  +=    1244.342 * cos(0.91612680579  + 2.44768055480 *T)
    sumL0  +=    1072.008 * cos(0.23564502877  + 62.25142559510 *T)
    sumL0  +=    1090.461 * cos(1.77501638912  + 12.53017297220 *T)
    sumL0  +=     707.875 * cos(5.18285226584  + 213.29909543800 *T)
    sumL0  +=     653.401 * cos(0.96586909116  + 78.71375183040 *T)
    sumL0  +=     627.562 * cos(0.18210181975  + 984.60033162190 *T)
    sumL0  +=     524.495 * cos(2.01276706996  + 299.12639426920 *T)
    sumL0  +=     559.370 * cos(3.35776737704  + 0.52126486180 *T)
    sumL0  +=     606.827 * cos(5.43209728952  + 529.69096509460 *T)
    sumL0  +=     404.891 * cos(5.98689011389  + 8.07675484730 *T)
    sumL0  +=     467.211 * cos(0.41484068933  + 145.10977900970 *T)
    sumL0  +=     471.288 * cos(1.40664336447  + 184.72728735580 *T)
    sumL0  +=     483.219 * cos(2.10553990154  + 0.96320784650 *T)
    sumL0  +=     395.614 * cos(5.87039580949  + 351.81659230870 *T)
    sumL0  +=     433.532 * cos(5.52142978255  + 183.24281464750 *T)
    sumL0  +=     309.885 * cos(5.83301304674  + 145.63104387150 *T)
    sumL0  +=     378.609 * cos(2.34975805006  + 56.62235130260 *T)
    sumL0  +=     398.996 * cos(0.33810765436  + 415.55249061210 *T)
    sumL0  +=     300.379 * cos(5.64353974146  + 22.09140052780 *T)
    sumL0  +=     249.229 * cos(4.74617120584  + 225.82926841020 *T)
    sumL0  +=     239.334 * cos(2.35045874708  + 137.03302416240 *T)
    sumL0  +=     294.172 * cos(5.83916826225  + 39.61750834610 *T)
    sumL0  +=     216.480 * cos(4.77847481363  + 340.77089204480 *T)
    sumL0  +=     251.792 * cos(1.63696775578  + 221.37585028530 *T)
    sumL0  +=     219.621 * cos(1.92212987979  + 67.66805156650 *T)
    sumL0  +=     201.963 * cos(1.29693040865  + 0.04818410980 *T)
    sumL0  +=     224.097 * cos(0.51574863468  + 84.34282612290 *T)
    sumL0  +=     216.549 * cos(6.14211862702  + 5.93789083320 *T)
    sumL0  +=     222.588 * cos(2.84309380331  + 0.26063243090 *T)
    sumL0  +=     207.828 * cos(5.58020570040  + 68.84370773410 *T)
    sumL0  +=     187.474 * cos(1.31924326253  + 0.16005869440 *T)
    sumL0  +=     158.028 * cos(0.73811997211  + 54.17467074780 *T)
    sumL0  +=     199.146 * cos(0.95634155010  + 152.53214255120 *T)
    sumL0  +=     168.648 * cos(5.87874000882  + 18.15924726470 *T)
    sumL0  +=     170.300 * cos(3.67717520688  + 5.41662597140 *T)
    sumL0  +=     193.652 * cos(1.88800122606  + 456.39383923560 *T)
    sumL0  +=     192.998 * cos(0.91616058506  + 453.42489381900 *T)
    sumL0  +=     181.934 * cos(3.53624029238  + 79.23501669220 *T)
    sumL0  +=     173.145 * cos(1.53860728054  + 160.60889739850 *T)
    sumL0  +=     164.588 * cos(1.42379714838  + 106.97674337190 *T)
    sumL0  +=     171.968 * cos(5.67952685533  + 219.89137757700 *T)
    sumL0  +=     162.792 * cos(3.05029377666  + 112.91463420510 *T)
    sumL0  +=     146.653 * cos(1.26300172265  + 59.80374504030 *T)
    sumL0  +=     139.453 * cos(5.38597723400  + 32.19514480460 *T)
    sumL0  +=     138.585 * cos(4.25994786673  + 909.81873305460 *T)
    sumL0  +=     143.058 * cos(1.29995487555  + 35.42472265210 *T)
    sumL0  +=     123.840 * cos(1.37359990336  + 7.11354700080 *T)
    sumL0  +=     104.414 * cos(5.02820888813  + 0.75075952540 *T)
    sumL0  +=     103.277 * cos(0.68095301267  + 14.97785352700 *T)
    sumL0  +=      94.741 * cos(0.90674090409  + 74.66972398270 *T)
    sumL0  +=      82.978 * cos(2.92828718445  + 265.98929347750 *T)
    sumL0  +=     110.163 * cos(2.02685778976  + 554.06998748280 *T)
    sumL0  +=      94.226 * cos(3.94266328260  + 74.89347315190 *T)
    sumL0  +=      79.858 * cos(1.01446829180  + 6.59228213900 *T)
    sumL0  +=     109.376 * cos(5.70581833286  + 77.96299230500 *T)
    sumL0  +=      85.876 * cos(1.70649435603  + 82.85835341460 *T)
    sumL0  +=     103.562 * cos(1.45770270246  + 24.37902238820 *T)
    sumL0  +=      74.667 * cos(4.63177552576  + 69.36497259590 *T)
    sumL0  +=      79.919 * cos(3.00974084247  + 297.64192156090 *T)
    sumL0  +=      84.502 * cos(0.36887189574  + 186.21176006410 *T)
    sumL0  +=      88.810 * cos(0.52481330563  + 181.75834193920 *T)
    sumL0  +=      70.303 * cos(1.18986880009  + 66.70484372000 *T)
    sumL0  +=      69.965 * cos(0.87476081875  + 305.34616939270 *T)
    sumL0  +=      69.927 * cos(3.76102749315  + 131.40394986990 *T)
    sumL0  +=      84.604 * cos(5.88725183325  + 256.53994050650 *T)
    sumL0  +=      74.341 * cos(6.24271323846  + 447.79581952650 *T)
    sumL0  +=      62.310 * cos(0.16901376623  + 479.28838891550 *T)
    sumL0  +=      72.726 * cos(2.84892775693  + 462.02291352810 *T)
    sumL0  +=      69.060 * cos(4.43934854374  + 39.35687591520 *T)
    sumL0  +=      76.568 * cos(4.58721110340  + 6.21977512350 *T)
    sumL0  +=      73.387 * cos(4.27603448634  + 87.31177153950 *T)
    sumL0  +=      55.307 * cos(1.49636544147  + 71.60020482960 *T)
    sumL0  +=      57.291 * cos(1.63015165542  + 143.62530630140 *T)
    sumL0  +=      61.661 * cos(3.18604743524  + 77.22927912210 *T)
    sumL0  +=      57.634 * cos(3.67180685401  + 51.20572533120 *T)
    sumL0  +=      50.289 * cos(1.12279384633  + 20.60692781950 *T)
    sumL0  +=      53.744 * cos(5.51890986247  + 128.95626931510 *T)
    sumL0  +=      57.894 * cos(2.66877593418  + 381.61224066830 *T)
    sumL0  +=      58.112 * cos(1.58629352171  + 60.76695288680 *T)
    sumL0  +=      45.382 * cos(0.48053933052  + 14.01464568050 *T)
    sumL0  +=      37.581 * cos(6.06822931932  + 211.81462272970 *T)
    sumL0  +=      38.640 * cos(3.43597050177  + 153.49535039770 *T)
    sumL0  +=      46.087 * cos(4.36201639577  + 75.74480641380 *T)
    sumL0  +=      40.088 * cos(4.57333927519  + 46.20979048510 *T)
    sumL0  +=      34.229 * cos(2.93967782207  + 140.00196957900 *T)
    sumL0  +=      38.669 * cos(5.58941074168  + 99.16062095550 *T)
    sumL0  +=      34.827 * cos(1.02792863024  + 203.73786788240 *T)
    sumL0  +=      40.024 * cos(0.69889667397  + 218.40690486870 *T)
    sumL0  +=      32.538 * cos(4.21625657443  + 200.76892246580 *T)
    sumL0  +=      31.865 * cos(5.50961503408  + 72.33391801250 *T)
    sumL0  +=      41.695 * cos(3.82438031124  + 81.00137369080 *T)
    sumL0  +=      34.795 * cos(0.39363490236  + 1.37259812370 *T)
    sumL0  +=      39.775 * cos(6.05600836903  + 293.18850343600 *T)
    sumL0  +=      27.577 * cos(2.18261286374  + 125.98732389850 *T)
    sumL0  +=      36.279 * cos(1.66586085405  + 258.02441321480 *T)
    sumL0  +=      35.442 * cos(1.96652806541  + 835.03713448730 *T)
    sumL0  +=      35.361 * cos(3.72258690030  + 692.58748435350 *T)
    sumL0  +=      27.323 * cos(2.10164372072  + 209.36694217490 *T)
    sumL0  +=      26.530 * cos(4.48265986115  + 373.90799283650 *T)
    sumL0  +=      34.472 * cos(1.07907945481  + 191.20769491020 *T)
    sumL0  +=      29.915 * cos(3.87358632506  + 259.50888592310 *T)
    sumL0  +=      26.233 * cos(3.63172504384  + 490.33408917940 *T)
    sumL0  +=      25.848 * cos(0.54461409359  + 41.64449777560 *T)
    sumL0  +=      26.989 * cos(6.27711247734  + 28.57180808220 *T)
    sumL0  +=      26.391 * cos(5.81110061049  + 75.30286342910 *T)
    sumL0  +=      34.227 * cos(6.05617272657  + 275.55052103310 *T)
    sumL0  +=      24.279 * cos(3.18776564878  + 81.37388070630 *T)
    sumL0  +=      29.937 * cos(1.88789751816  + 269.92144674060 *T)
    sumL0  +=      26.235 * cos(6.20105251336  + 134.58534360760 *T)
    sumL0  +=      22.754 * cos(0.92919725789  + 288.08069400530 *T)
    sumL0  +=      25.180 * cos(5.42547381962  + 116.42609634290 *T)
    sumL0  +=      22.715 * cos(0.53098783687  + 1514.29129671650 *T)
    sumL0  +=      26.485 * cos(4.77176167929  + 284.14854074220 *T)
    sumL0  +=      27.008 * cos(4.75281624832  + 41.10198105440 *T)
    sumL0  +=      21.972 * cos(4.58613057386  + 404.50679034820 *T)
    sumL0  +=      22.012 * cos(1.84389287183  + 617.80588578620 *T)
    sumL0  +=      24.694 * cos(4.70875195490  + 378.64329525170 *T)
    sumL0  +=      28.949 * cos(0.17127584792  + 528.20649238630 *T)
    sumL0  +=      20.492 * cos(0.10285646641  + 195.13984817330 *T)
    sumL0  +=      20.696 * cos(5.62143477633  + 55.65914345610 *T)
    sumL0  +=      25.843 * cos(0.74627159338  + 278.51946644970 *T)
    sumL0  +=      22.990 * cos(3.58378694661  + 1.59634729290 *T)
    sumL0  +=      21.843 * cos(0.05733533568  + 173.94221952280 *T)
    sumL0  +=      19.050 * cos(2.30351091243  + 5.10780943070 *T)
    sumL0  +=      20.675 * cos(2.64113858585  + 105.49227066360 *T)
    sumL0  +=      21.856 * cos(5.87352402691  + 45.57665103870 *T)
    sumL0  +=      21.120 * cos(1.98081790016  + 114.39910691340 *T)
    sumL0  +=      19.279 * cos(2.84304025179  + 159.12442469020 *T)
    sumL0  +=      19.061 * cos(0.50598371738  + 67.35923502580 *T)
    sumL0  +=      20.434 * cos(3.77601951414  + 135.54855145410 *T)
    sumL0  +=      17.326 * cos(4.47793157645  + 120.35824960600 *T)
    sumL0  +=      20.547 * cos(0.88695598555  + 255.05546779820 *T)
    sumL0  +=      19.320 * cos(1.48569290504  + 0.89377187730 *T)
    sumL0  +=      21.331 * cos(2.74470023060  + 28.31117565130 *T)
    sumL0  +=      17.582 * cos(4.09139636700  + 296.15744885260 *T)
    sumL0  +=      15.918 * cos(3.94525074972  + 17.52610781830 *T)
    sumL0  +=      15.562 * cos(0.92748407689  + 300.61086697750 *T)
    sumL0  +=      16.439 * cos(0.30868798605  + 30.71067209630 *T)
    sumL0  +=      15.237 * cos(4.93048601827  + 7.42236354150 *T)
    sumL0  +=      19.284 * cos(6.21950083268  + 329.83706636550 *T)
    sumL0  +=      13.860 * cos(0.56255266406  + 144.14657116320 *T)
    sumL0  +=      16.206 * cos(2.30292598693  + 344.70304530790 *T)
    sumL0  +=      16.041 * cos(0.19723295436  + 103.09277421860 *T)
    sumL0  +=      14.414 * cos(2.57606243208  + 230.56457082540 *T)
    sumL0  +=      16.789 * cos(4.93540052916  + 565.11568774670 *T)
    sumL0  +=      17.052 * cos(1.81844925116  + 294.67297614430 *T)
    sumL0  +=      16.766 * cos(0.27542186330  + 73.81839072080 *T)
    sumL0  +=      15.428 * cos(1.91577056305  + 96.87299909510 *T)
    sumL0  +=      15.718 * cos(3.87095025861  + 98.89998852460 *T)
    sumL0  +=      11.923 * cos(6.17545505441  + 44.72531777680 *T)
    sumL0  +=      12.407 * cos(6.22419970167  + 80.19822453870 *T)
    sumL0  +=      13.040 * cos(1.99652993223  + 27.08733537390 *T)
    sumL0  +=      13.229 * cos(3.43782440072  + 227.31374111850 *T)
    sumL0  +=      11.669 * cos(4.31526860843  + 426.59819087600 *T)
    sumL0  +=      14.378 * cos(5.78353646474  + 1059.38193018920 *T)
    sumL0  +=      15.879 * cos(0.98454960055  + 6208.29425142410 *T)
    sumL0  +=      11.158 * cos(1.74417430690  + 220.41264243880 *T)
    sumL0  +=      11.989 * cos(5.84388657950  + 13.33332212430 *T)
    sumL0  +=      11.386 * cos(2.55925734515  + 19.12245511120 *T)
    sumL0  +=      13.281 * cos(5.39472153462  + 391.17346822390 *T)
    sumL0  +=      12.295 * cos(4.57340278496  + 23.57587323610 *T)
    sumL0  +=      12.827 * cos(1.77410269070  + 180.27386923090 *T)
    sumL0  +=      11.651 * cos(4.29138607818  + 142.44965013380 *T)
    sumL0  +=      12.248 * cos(2.44241346243  + 100.38446123290 *T)
    sumL0  +=      12.421 * cos(2.32591770919  + 80.71948940050 *T)
    sumL0  +=       9.774 * cos(0.39898140151  + 7.86430652620 *T)
    sumL0  +=      13.172 * cos(2.74099358938  + 177.87437278590 *T)
    sumL0  +=      12.262 * cos(5.42795591646  + 831.10498122420 *T)
    sumL0  +=      10.272 * cos(5.90194483926  + 74.52096613640 *T)
    sumL0  +=       9.317 * cos(3.75869700774  + 74.82978267710 *T)
    sumL0  +=      10.701 * cos(4.00709797731  + 235.39049596580 *T)
    sumL0  +=       9.243 * cos(5.38492199672  + 20.44686912510 *T)
    sumL0  +=       9.461 * cos(2.60126707172  + 92.30770638560 *T)
    sumL0  +=      12.066 * cos(5.52163100220  + 74.26033370550 *T)
    sumL0  +=      10.836 * cos(1.88393779293  + 241.61027108930 *T)
    sumL0  +=       9.317 * cos(1.16483658310  + 74.73341445750 *T)
    sumL0  +=      10.718 * cos(5.50310449842  + 187.69623277240 *T)
    sumL0  +=      12.057 * cos(6.02120050390  + 154.01661525950 *T)
    sumL0  +=       9.124 * cos(1.15458738606  + 0.63313944640 *T)
    sumL0  +=      11.526 * cos(6.26425302826  + 155.78297225810 *T)
    sumL0  +=      12.200 * cos(5.79400179483  + 1364.72809958190 *T)
    sumL0  +=      10.979 * cos(5.76614513865  + 628.85158605010 *T)
    sumL0  +=       8.532 * cos(5.18016456150  + 1.64453140270 *T)
    sumL0  +=       8.660 * cos(3.78133822411  + 74.94165726170 *T)
    sumL0  +=      11.227 * cos(1.30788626675  + 604.47256366190 *T)
    sumL0  +=      10.531 * cos(1.05867421534  + 291.70403072770 *T)
    sumL0  +=       8.446 * cos(3.52020067595  + 756.32338265690 *T)
    sumL0  +=      10.291 * cos(5.30493908317  + 75.04223099820 *T)
    sumL0  +=       8.015 * cos(6.24347048958  + 543.02428721890 *T)
    sumL0  +=       7.796 * cos(5.23497582886  + 58.10682401090 *T)
    sumL0  +=       9.310 * cos(1.65210713729  + 24.11838995730 *T)
    sumL0  +=       8.642 * cos(1.14285691458  + 74.62153987290 *T)
    sumL0  +=       7.797 * cos(4.00208030502  + 31.49256938900 *T)
    sumL0  +=       8.915 * cos(3.63129389881  + 408.43894361130 *T)
    sumL0  +=       7.191 * cos(4.24536221306  + 110.20632121940 *T)
    sumL0  +=       9.764 * cos(0.10205649809  + 366.48562929500 *T)
    sumL0  +=       8.710 * cos(2.25910759480  + 451.94042111070 *T)
    sumL0  +=       9.430 * cos(2.00492467431  + 331.32153907380 *T)
    sumL0  +=       9.008 * cos(1.63146330622  + 443.86366626340 *T)
    sumL0  +=       7.247 * cos(4.35313018726  + 88.11492069160 *T)
    sumL0  +=       7.659 * cos(0.61918159043  + 50.40257617910 *T)
    sumL0  +=       6.836 * cos(5.12190844483  + 0.80314915210 *T)
    sumL0  +=       9.367 * cos(1.42664537007  + 414.06801790380 *T)
    sumL0  +=       8.136 * cos(0.45279998999  + 25.60286266560 *T)
    sumL0  +=       9.199 * cos(2.41000352664  + 10138.50394764370 *T)
    sumL0  +=       6.896 * cos(5.85017813531  + 339.28641933650 *T)
    sumL0  +=       6.487 * cos(6.03397141885  + 1.22384027740 *T)
    sumL0  +=       7.186 * cos(4.00480285222  + 157.63995198190 *T)
    sumL0  +=       8.130 * cos(0.21907525983  + 422.66603761290 *T)
    sumL0  +=       8.022 * cos(2.09953974305  + 92.94084583200 *T)
    sumL0  +=       6.371 * cos(4.47820781123  + 79.88940799800 *T)
    sumL0  +=       7.957 * cos(5.86499639179  + 760.25553592000 *T)
    sumL0  +=       8.383 * cos(2.33782809093  + 417.03696332040 *T)
    sumL0  +=       7.931 * cos(3.41952210669  + 7.70424783180 *T)
    sumL0  +=       6.613 * cos(1.39197711439  + 16.67477455640 *T)
    sumL0  +=       7.142 * cos(5.57864813931  + 4.73530241520 *T)
    sumL0  +=       7.441 * cos(0.48121969777  + 68.18931642830 *T)
    sumL0  +=       7.187 * cos(0.50441238918  + 457.87831194390 *T)
    sumL0  +=       6.613 * cos(2.84628770892  + 142.14083359310 *T)
    sumL0  +=       6.399 * cos(3.88681409308  + 74.03083904190 *T)
    sumL0  +=       6.140 * cos(1.65776909220  + 350.33211960040 *T)
    sumL0  +=       7.053 * cos(0.13890020306  + 306.83064210100 *T)
    sumL0  +=       6.010 * cos(5.46434004640  + 48.75804477640 *T)
    sumL0  +=       6.810 * cos(6.15448079403  + 67.88049988760 *T)
    sumL0  +=       5.982 * cos(2.36098472874  + 2.00573757010 *T)
    sumL0  +=       5.709 * cos(1.49928444044  + 206.18554843720 *T)
    sumL0  +=       6.843 * cos(1.08172913275  + 465.95506679120 *T)
    sumL0  +=       6.408 * cos(5.07331258075  + 4.66586644600 *T)
    sumL0  +=       7.805 * cos(3.98866710061  + 3.62333672240 *T)
    sumL0  +=       5.527 * cos(5.57881556653  + 2.92076130680 *T)
    sumL0  +=       5.521 * cos(3.38225987040  + 149.45132255000 *T)
    sumL0  +=       6.107 * cos(1.95504762856  + 216.92243216040 *T)
    sumL0  +=       5.692 * cos(2.83076925167  + 260.99335863140 *T)
    sumL0  +=       6.586 * cos(2.71085048651  + 329.72519178090 *T)
    sumL0  +=       5.826 * cos(3.98800970226  + 347.88443904560 *T)
    sumL0  +=       5.745 * cos(0.49160564101  + 0.37250701550 *T)
    sumL0  +=       5.473 * cos(5.69158856279  + 1.69692102940 *T)
    sumL0  +=       4.982 * cos(2.40870746521  + 342.25536475310 *T)
    sumL0  +=       6.040 * cos(4.78944090986  + 558.00214074590 *T)
    sumL0  +=       5.317 * cos(2.78403764459  + 13.49338081870 *T)
    sumL0  +=       5.090 * cos(5.47747622578  + 372.42352012820 *T)
    sumL0  +=       4.894 * cos(1.77082918618  + 333.65734504400 *T)
    sumL0  +=       5.389 * cos(2.94076732149  + 9.40116886120 *T)
    sumL0  +=       4.757 * cos(5.37129102802  + 61.28821774860 *T)
    sumL0  +=       5.000 * cos(3.43988321744  + 518.64526483070 *T)
    sumL0  +=       5.664 * cos(3.30309284254  + 0.65439130580 *T)
    sumL0  +=       5.530 * cos(0.45092393824  + 162.09337010680 *T)
    sumL0  +=       5.746 * cos(3.45964923866  + 55.13787859430 *T)
    sumL0  +=       5.666 * cos(1.23578675332  + 328.35259365720 *T)
    sumL0  +=       5.071 * cos(5.42033080481  + 977.48678462110 *T)
    sumL0  +=       5.719 * cos(0.66718965817  + 92.04707395470 *T)
    sumL0  +=       4.994 * cos(1.29267872727  + 983.11585891360 *T)
    sumL0  +=       5.570 * cos(2.36255927193  + 6.90109867970 *T)
    sumL0  +=       5.189 * cos(2.40682220291  + 58.31927233200 *T)
    sumL0  +=       5.502 * cos(0.13301359232  + 149.67507171920 *T)
    sumL0  +=       4.706 * cos(1.85473330365  + 119.50691634410 *T)
    sumL0  +=       4.305 * cos(4.18171934306  + 90.82323367730 *T)
    sumL0  +=       5.953 * cos(1.73036741041  + 152.74459087230 *T)
    sumL0  +=       5.607 * cos(5.53187692339  + 1087.69310584050 *T)
    sumL0  +=       5.591 * cos(5.75072223569  + 358.93013930950 *T)
    sumL0  +=       4.441 * cos(0.97726075887  + 4.19278569400 *T)
    sumL0  +=       4.608 * cos(3.31800103668  + 89.75945209430 *T)
    sumL0  +=       4.677 * cos(5.01422713233  + 43.12897048390 *T)
    sumL0  +=       4.034 * cos(1.08242564328  + 75.53235809270 *T)
    sumL0  +=       5.626 * cos(1.10270225604  + 66.91729204110 *T)
    sumL0  +=       4.058 * cos(1.94012136050  + 17.26547538740 *T)
    sumL0  +=       4.770 * cos(2.24207019076  + 986.08480433020 *T)
    sumL0  +=       5.207 * cos(1.36600428910  + 767.36908292080 *T)
    sumL0  +=       4.940 * cos(0.13733547633  + 0.85133326190 *T)
    sumL0  +=       4.339 * cos(5.33814728291  + 152.01087768940 *T)
    sumL0  +=       3.917 * cos(3.85320550575  + 2.28762186040 *T)
    sumL0  +=       3.903 * cos(5.87573410158  + 16.46232623530 *T)
    sumL0  +=       4.655 * cos(0.93665017107  + 267.47376618580 *T)
    sumL0  +=       4.638 * cos(5.42566923517  + 16.04163511000 *T)
    sumL0  +=       5.177 * cos(3.40845690805  + 1289.94650101460 *T)
    sumL0  +=       3.825 * cos(0.59888105730  + 210.33015002140 *T)
    sumL0  +=       5.048 * cos(2.16732539242  + 367.97010200330 *T)
    sumL0  +=       3.819 * cos(1.70901925915  + 5.62907429250 *T)
    sumL0  +=       5.004 * cos(0.26759264038  + 403.13419222450 *T)
    sumL0  +=       4.202 * cos(5.12029089394  + 19.01058052660 *T)
    sumL0  +=       4.472 * cos(2.88978371811  + 59.28248017850 *T)
    sumL0  +=       4.700 * cos(4.17709035394  + 130.44074202340 *T)
    sumL0  +=       4.487 * cos(0.85521839581  + 969.62247809490 *T)
    sumL0  +=       3.802 * cos(4.59721371468  + 25.86349509650 *T)
    sumL0  +=       4.024 * cos(4.98868930941  + 30.05628079050 *T)
    sumL0  +=       4.722 * cos(6.16359211847  + 173.68158709190 *T)
    sumL0  +=       4.732 * cos(3.76697693308  + 373.01422095920 *T)
    sumL0  +=       3.542 * cos(0.76768843819  + 114.13847448250 *T)
    sumL0  +=       3.370 * cos(2.00021522907  + 286.59622129700 *T)
    sumL0  +=       3.626 * cos(3.20240733896  + 991.71387862270 *T)
    sumL0  +=       4.028 * cos(0.46802022168  + 387.24131496080 *T)
    sumL0  +=       3.442 * cos(5.20959733350  + 894.84087952760 *T)
    sumL0  +=       3.273 * cos(5.46374958434  + 192.69216761850 *T)
    sumL0  +=       3.291 * cos(3.97847646998  + 264.50482076920 *T)
    sumL0  +=       3.291 * cos(1.62538722379  + 681.54178408960 *T)
    sumL0  +=       3.198 * cos(3.96580804800  + 146.38180339690 *T)
    sumL0  +=       4.016 * cos(1.40657464840  + 383.09671337660 *T)
    sumL0  +=       3.844 * cos(3.60271287642  + 0.59070083100 *T)
    sumL0  +=       3.123 * cos(4.50215520755  + 1439.50969814920 *T)
    sumL0  +=       3.434 * cos(5.18704419009  + 97.41551581630 *T)
    sumL0  +=       3.123 * cos(4.14198630214  + 76.47851959670 *T)
    sumL0  +=       3.259 * cos(1.65410614252  + 214.78356814630 *T)
    sumL0  +=       3.557 * cos(4.15769848885  + 68.56182344380 *T)
    sumL0  +=       2.934 * cos(2.95575531139  + 120.99138905240 *T)
    sumL0  +=       2.914 * cos(3.69930995976  + 874.39401040250 *T)
    sumL0  +=       3.301 * cos(2.14570582133  + 253.57099508990 *T)
    sumL0  +=       3.362 * cos(4.82277888708  + 19.64371997300 *T)
    sumL0  +=       3.218 * cos(2.56428709831  + 60.55450456570 *T)
    sumL0  +=       3.059 * cos(3.57539890234  + 117.91056905120 *T)
    sumL0  +=       3.073 * cos(3.54739757836  + 95.38852638680 *T)
    sumL0  +=       2.789 * cos(0.65190913388  + 42.53826965290 *T)
    sumL0  +=       3.235 * cos(5.31608088666  + 546.95644048200 *T)
    sumL0  +=       3.657 * cos(5.89905956226  + 16.15350969460 *T)
    sumL0  +=       3.549 * cos(2.76314903735  + 82.48584639910 *T)
    sumL0  +=       3.627 * cos(4.68663059919  + 593.42686339800 *T)
    sumL0  +=       3.306 * cos(1.57486085317  + 312.45971639350 *T)
    sumL0  +=       3.602 * cos(2.51921910142  + 22.89454967990 *T)
    sumL0  +=       3.431 * cos(4.95532928836  + 49.72125262290 *T)
    sumL0  +=       2.675 * cos(2.41314606353  + 29.20494752860 *T)
    sumL0  +=       3.101 * cos(1.42849885249  + 17.63798240290 *T)
    sumL0  +=       3.399 * cos(3.02815712113  + 88.79624424780 *T)
    sumL0  +=       3.379 * cos(2.65894323745  + 771.30123618390 *T)
    sumL0  +=       2.547 * cos(6.10642153361  + 455.87257437380 *T)
    sumL0  +=       2.967 * cos(0.31461418738  + 150.52640498110 *T)
    sumL0  +=       2.681 * cos(1.16839594153  + 477.80391620720 *T)
    sumL0  +=       3.343 * cos(3.09880618811  + 552.58551477450 *T)
    sumL0  +=       2.678 * cos(1.42841144096  + 1.11196569280 *T)
    sumL0  +=       2.540 * cos(1.94053883528  + 6.48040755440 *T)
    sumL0  +=       2.491 * cos(2.97226347939  + 453.94615868080 *T)
    sumL0  +=       2.744 * cos(1.93313970916  + 73.18525127440 *T)
    sumL0  +=       2.935 * cos(4.59394106280  + 167.08930495290 *T)
    sumL0  +=       3.007 * cos(1.38745560615  + 365.00115658670 *T)
    sumL0  +=       3.053 * cos(1.91792962252  + 561.18353448360 *T)
    sumL0  +=       2.496 * cos(5.45540866674  + 66.18357885820 *T)
    sumL0  +=       2.622 * cos(1.46324659292  + 33.13710079170 *T)
    sumL0  +=       3.203 * cos(4.01683757076  + 555.55446019110 *T)
    sumL0  +=       2.317 * cos(1.13727677715  + 43.24084506850 *T)
    sumL0  +=       2.341 * cos(5.87635018071  + 228.27694896500 *T)
    sumL0  +=       2.841 * cos(3.60234459541  + 42.58645376270 *T)
    sumL0  +=       2.858 * cos(1.53714262537  + 353.30106501700 *T)
    sumL0  +=       2.484 * cos(3.85791009894  + 104.00779795530 *T)
    sumL0  +=       2.903 * cos(5.21967656512  + 73.40900044360 *T)
    sumL0  +=       2.760 * cos(1.21343315367  + 32.24332891440 *T)
    sumL0  +=       2.269 * cos(3.31411391807  + 4.14460158420 *T)
    sumL0  +=       2.241 * cos(2.65636547591  + 70.11573212130 *T)
    sumL0  +=       2.246 * cos(4.08175081363  + 123.53964334370 *T)
    sumL0  +=       2.583 * cos(2.38305971478  + 100.64509366380 *T)
    sumL0  +=       2.761 * cos(5.71758409791  + 43.28902917830 *T)
    sumL0  +=       2.827 * cos(6.16734582851  + 101.86893394120 *T)
    sumL0  +=       2.838 * cos(0.54888495490  + 20.49505323490 *T)
    sumL0  +=       2.145 * cos(4.04195315408  + 47.06112374700 *T)
    sumL0  +=       2.810 * cos(1.90169260186  + 273.10284047830 *T)
    sumL0  +=       2.922 * cos(2.79808700183  + 418.26080359780 *T)
    sumL0  +=       2.070 * cos(5.50402718290  + 47.69426319340 *T)
    sumL0  +=       2.071 * cos(2.01973573060  + 316.39186965660 *T)
    sumL0  +=       2.520 * cos(3.12740527423  + 905.88657979150 *T)
    sumL0  +=       2.170 * cos(4.46196560050  + 2.70831298570 *T)
    sumL0  +=       2.399 * cos(3.78849518316  + 75.58474771940 *T)
    sumL0  +=       2.746 * cos(4.82558024832  + 6.85291456990 *T)
    sumL0  +=       2.717 * cos(2.33108458294  + 404.61866493280 *T)
    sumL0  +=       2.416 * cos(4.11932546205  + 332.80601178210 *T)
    sumL0  +=       1.974 * cos(5.79881978458  + 11.15757484850 *T)
    sumL0  +=       1.967 * cos(5.43918682709  + 199.28444975750 *T)
    sumL0  +=       2.282 * cos(2.79530897096  + 22.63391724900 *T)
    sumL0  +=       1.910 * cos(0.27727117649  + 69.15252427480 *T)
    sumL0  +=       2.471 * cos(3.51033894778  + 8.59801970910 *T)
    sumL0  +=       1.904 * cos(3.45282024423  + 472.17484191470 *T)
    sumL0  +=       2.606 * cos(0.43601023323  + 439.78275515400 *T)
    sumL0  +=       2.663 * cos(5.43112910549  + 3265.83082813250 *T)
    sumL0  +=       2.058 * cos(1.69362390174  + 65.87476231750 *T)
    sumL0  +=       2.258 * cos(5.32927779367  + 908.33426034630 *T)
    sumL0  +=       1.873 * cos(5.55399805910  + 175.16605980020 *T)
    sumL0  +=       2.222 * cos(0.96973865202  + 39.09624348430 *T)
    sumL0  +=       2.121 * cos(2.00302088316  + 106.01353552540 *T)
    sumL0  +=       2.010 * cos(1.49945418027  + 29.22619938800 *T)
    sumL0  +=       2.222 * cos(4.36573603431  + 468.24268865160 *T)
    sumL0  +=       2.055 * cos(0.05798973044  + 205.22234059070 *T)
    sumL0  +=       2.086 * cos(0.44287700052  + 10.29494073850 *T)
    sumL0  +=       1.770 * cos(4.32048805830  + 0.45757438700 *T)
    sumL0  +=       2.112 * cos(5.78682409103  + 486.40193591630 *T)
    sumL0  +=       1.909 * cos(0.82888506421  + 254.94359321360 *T)
    sumL0  +=       1.973 * cos(6.05826379648  + 78.40493528970 *T)
    sumL0  +=       1.908 * cos(5.55892482384  + 15.49911838880 *T)
    sumL0  +=       1.888 * cos(6.20874408008  + 198.32124191100 *T)
    sumL0  +=       2.092 * cos(2.55561831566  + 49.50880430180 *T)
    sumL0  +=       1.775 * cos(6.17741514589  + 258.87574647670 *T)
    sumL0  +=       2.146 * cos(1.40508118810  + 526.72201967800 *T)
    sumL0  +=       1.800 * cos(0.04238718337  + 334.29048449040 *T)
    sumL0  +=       1.738 * cos(1.99152421966  + 77.06922042770 *T)
    sumL0  +=       1.641 * cos(3.36410541913  + 118.02244363580 *T)
    sumL0  +=       2.084 * cos(5.21275540105  + 134.06407874580 *T)
    sumL0  +=       1.861 * cos(2.97480744198  + 178.78939652260 *T)
    sumL0  +=       1.750 * cos(2.01731567093  + 142.66209845490 *T)
    sumL0  +=       1.600 * cos(1.60963172329  + 40.16002506730 *T)
    sumL0  +=       1.668 * cos(1.53361997245  + 0.83008140250 *T)
    sumL0  +=       1.769 * cos(4.72689497119  + 32.71640966640 *T)
    sumL0  +=       1.782 * cos(2.60469159465  + 166.82867252200 *T)
    sumL0  +=       1.765 * cos(5.57583636983  + 522.57741809380 *T)
    sumL0  +=       1.877 * cos(1.09044005603  + 274.06604832480 *T)
    sumL0  +=       1.544 * cos(1.91798419140  + 303.86169668440 *T)
    sumL0  +=       1.532 * cos(3.44973397383  + 124.50285119020 *T)
    sumL0  +=       1.658 * cos(5.23946791059  + 233.90602325750 *T)
    sumL0  +=       1.743 * cos(2.32369273283  + 290.21955801940 *T)
    sumL0  +=       1.528 * cos(0.62020771152  + 1033.35837639830 *T)
    sumL0  +=       1.522 * cos(6.11272567668  + 165.60483224460 *T)
    sumL0  +=       1.496 * cos(2.02190170195  + 150.08446199640 *T)
    sumL0  +=       1.490 * cos(3.30997217921  + 820.05928096030 *T)
    sumL0  +=       1.398 * cos(3.41567259878  + 4.99593484610 *T)
    sumL0  +=       1.886 * cos(6.25585539882  + 162.89651925890 *T)
    sumL0  +=       1.388 * cos(0.62745508416  + 448.68959140380 *T)
    sumL0  +=       1.918 * cos(0.91483173263  + 1819.63746610920 *T)
    sumL0  +=       1.682 * cos(2.11545135265  + 189.72322220190 *T)
    sumL0  +=       1.711 * cos(2.55731536599  + 1108.13997496560 *T)
    sumL0  +=       1.598 * cos(0.89607036108  + 115.88357962170 *T)
    sumL0  +=       1.477 * cos(0.22838214106  + 370.93904741990 *T)
    sumL0  +=       1.727 * cos(1.54005322759  + 401.64971951620 *T)
    sumL0  +=       1.432 * cos(3.73381952953  + 8.90683624980 *T)
    sumL0  +=       1.338 * cos(0.19338311739  + 81.89514556810 *T)
    sumL0  +=       1.618 * cos(6.02595306259  + 31.23193695810 *T)
    sumL0  +=       1.452 * cos(0.11827434627  + 72.77586099720 *T)
    sumL0  +=       1.626 * cos(4.26332651029  + 369.45457471160 *T)
    sumL0  +=       1.284 * cos(3.18389639039  + 362.86229257260 *T)
    sumL0  +=       1.755 * cos(5.57436830525  + 344.96367773880 *T)
    sumL0  +=       1.451 * cos(4.42615381750  + 189.18070548070 *T)
    sumL0  +=       1.294 * cos(2.77775613125  + 63.62402371880 *T)
    sumL0  +=       1.712 * cos(2.16785817753  + 536.80451209540 *T)
    sumL0  +=       1.266 * cos(2.78408709208  + 55.77101804070 *T)
    sumL0  +=       1.700 * cos(2.49908604932  + 441.26722786230 *T)
    sumL0  +=       1.645 * cos(2.41257585783  + 10.08249241740 *T)
    sumL0  +=       1.666 * cos(5.00664377609  + 79.51690098250 *T)
    sumL0  +=       1.674 * cos(0.00394784675  + 491.55792945680 *T)
    sumL0  +=       1.334 * cos(5.62972572008  + 129.91947716160 *T)
    sumL0  +=       1.328 * cos(0.25606135840  + 114.94162363460 *T)
    sumL0  +=       1.176 * cos(1.63208422172  + 84.18276742850 *T)
    sumL0  +=       1.177 * cos(4.10449248614  + 103.35340664950 *T)
    sumL0  +=       1.384 * cos(5.73004529870  + 89.33876096900 *T)
    sumL0  +=       1.287 * cos(0.99537181009  + 14.66903698630 *T)
    sumL0  +=       1.171 * cos(5.98078310685  + 57.14361616440 *T)
    sumL0  +=       1.126 * cos(3.39586759408  + 375.39246554480 *T)
    sumL0  +=       1.406 * cos(5.18477940039  + 113.87784205160 *T)
    sumL0  +=       1.397 * cos(1.49751443233  + 14.22709400160 *T)
    sumL0  +=       1.206 * cos(3.60301196272  + 480.77286162380 *T)
    sumL0  +=       1.113 * cos(4.80418427391  + 9.44935297100 *T)
    sumL0  +=       1.434 * cos(1.57158177893  + 419.74527630610 *T)
    sumL0  +=       1.473 * cos(1.03881736383  + 1215.16490244730 *T)
    sumL0  +=       1.311 * cos(2.99704179684  + 458.84151979040 *T)
    sumL0  +=       1.453 * cos(6.10676427884  + 64.69910614990 *T)
    sumL0  +=       1.070 * cos(3.16542344402  + 54.33472944220 *T)
    sumL0  +=       1.389 * cos(2.78512263875  + 26.02355379090 *T)
    sumL0  +=       1.203 * cos(0.20627563214  + 0.56944897160 *T)
    sumL0  +=       1.343 * cos(5.58004577468  + 95.22846769240 *T)
    sumL0  +=       1.062 * cos(2.40616687148  + 154.97982310600 *T)
    sumL0  +=       1.399 * cos(1.66776602336  + 240.38643081190 *T)
    sumL0  +=       1.036 * cos(5.53891715915  + 403.02231763990 *T)
    sumL0  +=       1.269 * cos(2.37684527290  + 37.87240320690 *T)
    sumL0  +=       1.197 * cos(4.87553746725  + 1044.40407666220 *T)
    sumL0  +=       1.009 * cos(2.74619462960  + 80.41067285980 *T)
    sumL0  +=       1.330 * cos(0.99603813295  + 483.22054217860 *T)
    sumL0  +=       1.348 * cos(0.58829539202  + 476.43131808350 *T)
    sumL0  +=       0.989 * cos(3.31666847329  + 18.91000679010 *T)
    sumL0  +=       1.054 * cos(2.85972567059  + 616.32141307790 *T)
    sumL0  +=       1.276 * cos(4.72938141859  + 691.10301164520 *T)
    sumL0  +=       1.219 * cos(3.62909689220  + 106.27416795630 *T)
    sumL0  +=       1.269 * cos(1.53301050628  + 280.96714700450 *T)
    sumL0  +=       0.968 * cos(2.73688433893  + 218.92816973050 *T)
    sumL0  +=       1.330 * cos(5.69234088687  + 694.07195706180 *T)
    sumL0  +=       1.121 * cos(2.80542439790  + 148.59998928810 *T)
    sumL0  +=       0.980 * cos(6.04026702553  + 5.46901559810 *T)
    sumL0  +=       1.235 * cos(0.61136787453  + 237.67811782620 *T)
    sumL0  +=       1.161 * cos(5.32209024820  + 369.08206769610 *T)
    sumL0  +=       0.944 * cos(1.14261861393  + 384.05992122310 *T)
    sumL0  +=       1.017 * cos(2.49896409345  + 147.11551657980 *T)
    sumL0  +=       1.082 * cos(2.46236323548  + 326.86812094890 *T)
    sumL0  +=       1.037 * cos(5.92126063748  + 4.82592514040 *T)
    sumL0  +=       1.232 * cos(5.83725190224  + 63.84777288800 *T)
    sumL0  +=       0.914 * cos(0.36627060914  + 10.93382567930 *T)
    sumL0  +=       1.165 * cos(5.32140830393  + 308.31511480930 *T)
    sumL0  +=       0.952 * cos(4.78982033367  + 93.90405367850 *T)
    sumL0  +=       0.993 * cos(3.39918521663  + 10.78506783300 *T)
    sumL0  +=       1.206 * cos(3.64530500530  + 699.70103135430 *T)
    sumL0  +=       0.893 * cos(2.97140509591  + 248.46318565920 *T)
    sumL0  +=       0.883 * cos(2.85605198817  + 15.19030184810 *T)
    sumL0  +=       0.875 * cos(3.43925709280  + 3.08082000120 *T)
    sumL0  +=       1.103 * cos(4.80576195766  + 6133.51265285680 *T)
    sumL0  +=       1.178 * cos(6.01576659565  + 377.15882254340 *T)
    sumL0  +=       0.974 * cos(2.51206761828  + 121.84272231430 *T)
    sumL0  +=       0.865 * cos(5.68958102479  + 141.69889060840 *T)
    sumL0  +=       0.974 * cos(3.15729174941  + 215.43795945210 *T)
    sumL0  +=       0.847 * cos(0.84854843713  + 2043.98226181110 *T)
    sumL0  +=       0.961 * cos(0.22181374419  + 0.91502373670 *T)
    sumL0  +=       1.167 * cos(2.37544946421  + 33.94024994380 *T)
    sumL0  +=       1.013 * cos(3.43778868786  + 36.90919536040 *T)
    sumL0  +=       0.838 * cos(1.63355479706  + 2.33580597020 *T)
    sumL0  +=       1.113 * cos(2.50989694970  + 405.99126305650 *T)
    sumL0  +=       0.987 * cos(1.14030030863  + 82.20396210880 *T)
    sumL0  +=       1.060 * cos(1.70915765427  + 438.29828244570 *T)
    sumL0  +=       0.829 * cos(2.97491446672  + 62.77269045690 *T)
    sumL0  +=       0.991 * cos(4.44869793177  + 406.10313764110 *T)
    sumL0  +=       0.952 * cos(6.13897716036  + 184.98791978670 *T)
    sumL0  +=       1.033 * cos(4.19932839584  + 141.48644228730 *T)
    sumL0  +=       1.029 * cos(5.13205530996  + 157.26744496640 *T)
    sumL0  +=       0.805 * cos(3.11000318272  + 93.79217909390 *T)
    sumL0  +=       0.865 * cos(4.44578048207  + 295.19424100610 *T)
    sumL0  +=       0.921 * cos(4.88190545687  + 12.00890811040 *T)
    sumL0  +=       0.949 * cos(5.99910796869  + 606.76018552230 *T)
    sumL0  +=       0.991 * cos(1.68012021428  + 40.58071619260 *T)
    sumL0  +=       0.854 * cos(0.37823682862  + 217.23124870110 *T)
    sumL0  +=       0.883 * cos(2.36224385140  + 3.77209456870 *T)
    sumL0  +=       1.016 * cos(3.90745585959  + 194.28851491140 *T)
    sumL0  +=       0.807 * cos(5.93051451738  + 302.09533968580 *T)
    sumL0  +=       0.879 * cos(0.52695940866  + 1057.89745748090 *T)
    sumL0  +=       1.009 * cos(1.19621149495  + 490.07345674850 *T)
    sumL0  +=       0.801 * cos(4.96724351781  + 661.09491496450 *T)
    sumL0  +=       0.843 * cos(0.97496000705  + 73.88782669000 *T)
    sumL0  +=       1.028 * cos(2.65189503651  + 477.91579079180 *T)
    sumL0  +=       0.846 * cos(3.36343733960  + 40.84134862350 *T)
    sumL0  +=       0.772 * cos(4.93551925711  + 425.11371816770 *T)
    sumL0  +=       0.781 * cos(0.59382881638  + 97.67614824720 *T)
    sumL0  +=       0.770 * cos(4.25621058806  + 488.84961647110 *T)
    sumL0  +=       0.935 * cos(1.24971148781  + 624.91943278700 *T)
    sumL0  +=       0.984 * cos(4.44298060183  + 171.65459766240 *T)
    sumL0  +=       0.806 * cos(0.09410536829  + 440.68227252570 *T)
    sumL0  +=       0.769 * cos(4.09296452529  + 140.65636088480 *T)
    sumL0  +=       0.802 * cos(0.78515729603  + 11.84884941600 *T)
    sumL0  +=       1.007 * cos(4.06909438635  + 76.15419669100 *T)
    sumL0  +=       0.935 * cos(0.03808890956  + 156.15547927360 *T)
    sumL0  +=       0.712 * cos(5.35549696452  + 610.69233878540 *T)
    sumL0  +=       0.911 * cos(4.70177653335  + 81.68269724700 *T)
    sumL0  +=       0.926 * cos(3.09765550633  + 833.55266177900 *T)
    sumL0  +=       0.812 * cos(3.54377377085  + 149.40313844020 *T)
    sumL0  +=       0.801 * cos(5.48034970408  + 21.97952594320 *T)
    sumL0  +=       0.900 * cos(4.21739347020  + 778.41478318470 *T)
    sumL0  +=       0.689 * cos(2.56862945159  + 109.31254934210 *T)
    sumL0  +=       0.685 * cos(3.19965828980  + 31.65262808340 *T)
    sumL0  +=       0.952 * cos(4.82579978820  + 1744.85586754190 *T)
    sumL0  +=       0.724 * cos(2.16878848875  + 1171.87587326900 *T)
    sumL0  +=       0.897 * cos(3.94746491183  + 75.67537044460 *T)
    sumL0  +=       0.739 * cos(5.45802693622  + 252.65597135320 *T)
    sumL0  +=       0.821 * cos(4.26046087515  + 1246.65747183630 *T)
    sumL0  +=       0.663 * cos(5.83767831921  + 86.63044798330 *T)
    sumL0  +=       0.664 * cos(1.72432848951  + 216.48048917570 *T)
    sumL0  +=       0.721 * cos(3.98089320988  + 902.70518605380 *T)
    sumL0  +=       0.663 * cos(4.98388191647  + 958.57677783100 *T)
    sumL0  +=       0.750 * cos(2.66119349235  + 363.51668387840 *T)
    sumL0  +=       0.828 * cos(3.62849315181  + 14.81779483260 *T)
    sumL0  +=       0.663 * cos(3.23521229496  + 207.88246946660 *T)
    sumL0  +=       0.681 * cos(3.29667467046  + 25.06034594440 *T)
    sumL0  +=       0.887 * cos(4.81199147907  + 155.50108796780 *T)
    sumL0  +=       0.646 * cos(2.65129054432  + 685.47393735270 *T)
    sumL0  +=       0.808 * cos(5.68764955594  + 280.00393915800 *T)
    sumL0  +=       0.619 * cos(1.32296886424  + 193.65537546500 *T)
    sumL0  +=       0.621 * cos(2.57837200005  + 703.63318461740 *T)
    sumL0  +=       0.854 * cos(3.63173686962  + 411.62033734900 *T)
    sumL0  +=       0.728 * cos(1.05845783695  + 916.93228005540 *T)
    sumL0  +=       0.733 * cos(0.75817076935  + 44.09217833040 *T)
    sumL0  +=       0.629 * cos(1.52747752455  + 397.39324334740 *T)
    sumL0  +=       0.599 * cos(5.30671888409  + 180.16199464630 *T)
    sumL0  +=       0.615 * cos(5.68908944293  + 25.27279426550 *T)
    sumL0  +=       0.709 * cos(0.64922689354  + 14.55716240170 *T)
    sumL0  +=       0.729 * cos(4.79389069212  + 479.40026350010 *T)
    sumL0  +=       0.667 * cos(1.98320895029  + 37.61177077600 *T)
    sumL0  +=       0.586 * cos(1.51157853976  + 668.20846196530 *T)
    sumL0  +=       0.639 * cos(1.69085181319  + 262.47783133970 *T)
    sumL0  +=       0.616 * cos(4.62035066985  + 12.26954054130 *T)
    sumL0  +=       0.687 * cos(2.24368916079  + 228.79821382680 *T)
    sumL0  +=       0.599 * cos(2.08317681783  + 149.30256470370 *T)
    sumL0  +=       0.597 * cos(3.14660293947  + 137.55428902420 *T)
    sumL0  +=       0.581 * cos(2.69049614736  + 823.99143422340 *T)
    sumL0  +=       0.709 * cos(4.38514100216  + 184.09414790940 *T)
    sumL0  +=       0.671 * cos(3.46925958949  + 105.38039607900 *T)
    sumL0  +=       0.619 * cos(2.91325544152  + 236.87496867410 *T)
    sumL0  +=       0.558 * cos(2.96177194880  + 34.20088237470 *T)
    sumL0  +=       0.648 * cos(5.56457931302  + 140.96517742550 *T)
    sumL0  +=       0.581 * cos(0.55427680962  + 331.20966448920 *T)
    sumL0  +=       0.585 * cos(0.15548306049  + 232.42155054920 *T)
    sumL0  +=       0.548 * cos(3.57525860446  + 497.44763618020 *T)
    sumL0  +=       0.574 * cos(5.60908001848  + 118.87377689770 *T)
    sumL0  +=       0.702 * cos(1.74156189506  + 149.04193227280 *T)
    sumL0  +=       0.543 * cos(0.49890445043  + 133.10087089930 *T)
    sumL0  +=       0.716 * cos(3.04149734724  + 131.92521473170 *T)
    sumL0  +=       0.544 * cos(6.22369103738  + 149.61138124440 *T)
    sumL0  +=       0.539 * cos(0.74276113752  + 911.30320576290 *T)
    sumL0  +=       0.614 * cos(5.83710659138  + 181.05576652360 *T)
    sumL0  +=       0.601 * cos(0.30768922616  + 407.58761034940 *T)
    sumL0  +=       0.635 * cos(4.03476459045  + 136.06981631590 *T)
    sumL0  +=       0.526 * cos(5.42874995984  + 450.97721326420 *T)
    sumL0  +=       0.547 * cos(3.07676037032  + 204.70107572890 *T)
    sumL0  +=       0.622 * cos(3.08666523105  + 268.43697403230 *T)
    sumL0  +=       0.537 * cos(4.25467814241  + 217.44369702220 *T)
    sumL0  +=       0.566 * cos(4.84686444604  + 842.15068148810 *T)
    sumL0  +=       0.636 * cos(2.57425168783  + 621.73803904930 *T)
    sumL0  +=       0.623 * cos(5.84341335807  + 52.80207262410 *T)
    sumL0  +=       0.544 * cos(3.62983006500  + 149.51501302480 *T)
    sumL0  +=       0.578 * cos(0.34796271917  + 139.48070471720 *T)
    sumL0  +=       0.537 * cos(5.99181083831  + 246.97871295090 *T)
    sumL0  +=       0.674 * cos(2.57972741298  + 602.98809095360 *T)
    sumL0  +=       0.539 * cos(6.19662961610  + 696.51963761660 *T)
    sumL0  +=       0.516 * cos(2.19916575703  + 458.09076026500 *T)
    sumL0  +=       0.632 * cos(5.26658553640  + 67.07735073550 *T)
    sumL0  +=       0.581 * cos(0.05320320337  + 95.93104310800 *T)
    sumL0  +=       0.504 * cos(6.24600928623  + 149.72325582900 *T)
    sumL0  +=       0.638 * cos(6.23121553223  + 10063.72234907640 *T)
    sumL0  +=       0.528 * cos(0.20662780149  + 310.17209453310 *T)
    sumL0  +=       0.537 * cos(2.96207822442  + 73.13706716460 *T)
    sumL0  +=       0.508 * cos(5.29969144068  + 335.77495719870 *T)
    sumL0  +=       0.487 * cos(2.83772541949  + 143.93412284210 *T)
    sumL0  +=       0.537 * cos(3.36808372143  + 252.08652238160 *T)
    sumL0  +=       0.632 * cos(5.88494938125  + 920.86443331850 *T)
    sumL0  +=       0.523 * cos(6.13183488285  + 1589.07289528380 *T)
    sumL0  +=       0.579 * cos(0.04597861846  + 563.63121503840 *T)
    sumL0  +=       0.613 * cos(0.34938781762  + 343.47920503050 *T)
    sumL0  +=       0.495 * cos(2.89212499810  + 61.44827644300 *T)
    sumL0  +=       0.623 * cos(0.71740350315  + 513.07988101300 *T)
    sumL0  +=       0.498 * cos(2.53375871592  + 41.75637236020 *T)
    sumL0  +=       0.564 * cos(2.01612524784  + 449.28029223480 *T)
    sumL0  +=       0.480 * cos(0.10535023009  + 69.67378913660 *T)
    sumL0  +=       0.494 * cos(3.25187012728  + 428.08266358430 *T)
    sumL0  +=       0.536 * cos(5.81149025999  + 282.66406803390 *T)
    sumL0  +=       0.468 * cos(0.89483830828  + 541.53981451060 *T)
    sumL0  +=       0.533 * cos(2.44239677121  + 393.46109008430 *T)
    sumL0  +=       0.589 * cos(6.24067076234  + 29.79564835960 *T)
    sumL0  +=       0.465 * cos(0.25006743710  + 57.25549074900 *T)
    sumL0  +=       0.622 * cos(3.89339038121  + 416.77633088950 *T)
    sumL0  +=       0.519 * cos(2.72375973888  + 469.13646052890 *T)
    sumL0  +=       0.498 * cos(1.76422801185  + 380.38840039090 *T)
    sumL0  +=       0.596 * cos(0.83642843095  + 98.35747180340 *T)
    sumL0  +=       0.459 * cos(0.39052216206  + 197.79997704920 *T)
    sumL0  +=       0.574 * cos(2.86366933069  + 170.76082578510 *T)
    sumL0  +=       0.518 * cos(3.38058345605  + 535.91074021810 *T)
    sumL0  +=       0.564 * cos(1.20395155832  + 832.58945393250 *T)
    sumL0  +=       0.556 * cos(4.39974034374  + 196.62432088160 *T)
    sumL0  +=       0.537 * cos(3.93637064940  + 460.53844081980 *T)
    sumL0  +=       0.482 * cos(2.41562148830  + 827.92358748650 *T)
    sumL0  +=       0.578 * cos(2.39644032176  + 1670.07426897460 *T)
    sumL0  +=       0.475 * cos(4.19223519775  + 271.40591944890 *T)
    sumL0  +=       0.452 * cos(3.99146251480  + 135.33610313300 *T)
    sumL0  +=       0.514 * cos(6.11377193423  + 1894.41906467650 *T)
    sumL0  +=       0.531 * cos(3.45607724228  + 450.45594840240 *T)
    sumL0  +=       0.492 * cos(5.87591888758  + 170.17012495410 *T)
    sumL0  +=       0.588 * cos(2.66953705406  + 310.97524368520 *T)
    sumL0  +=       0.564 * cos(1.04491117370  + 446.31134681820 *T)
    sumL0  +=       0.440 * cos(5.85084571537  + 224.23292111730 *T)
    sumL0  +=       0.441 * cos(2.19799439287  + 119.39504175950 *T)
    sumL0  +=       0.549 * cos(6.05651523611  + 76.37794586020 *T)
    sumL0  +=       0.573 * cos(2.29898526750  + 122.47586176070 *T)
    sumL0  +=       0.473 * cos(4.61187869812  + 291.26208774300 *T)
    sumL0  +=       0.553 * cos(4.90464199013  + 463.50738623640 *T)
    sumL0  +=       0.599 * cos(1.48666209087  + 149.82382956550 *T)
    sumL0  +=       0.499 * cos(3.72896978991  + 853.19638175200 *T)
    sumL0  +=       0.440 * cos(5.05024690419  + 79.44746501330 *T)
    sumL0  +=       0.558 * cos(0.42332722744  + 283.62727588040 *T)
    sumL0  +=       0.458 * cos(4.49655973916  + 754.83890994860 *T)
    sumL0  +=       0.449 * cos(1.32861330901  + 308.68762182480 *T)
    sumL0  +=       0.565 * cos(0.45628024105  + 241.87090352020 *T)
    sumL0  +=       0.510 * cos(3.70202346104  + 452.46168597250 *T)
    sumL0  +=       0.404 * cos(0.12335821240  + 1097.09427470170 *T)
    sumL0  +=       0.409 * cos(4.02092464698  + 735.87651353180 *T)
    sumL0  +=       0.464 * cos(3.82915608692  + 1094.54602041040 *T)
    sumL0  +=       0.439 * cos(2.72266354653  + 376.19561469690 *T)
    sumL0  +=       0.478 * cos(0.23380952322  + 1182.92157353290 *T)
    sumL0  +=       0.412 * cos(1.21971515436  + 419.48464387520 *T)
    sumL0  +=       0.546 * cos(0.22296640745  + 829.62050851590 *T)
    sumL0  +=       0.510 * cos(2.69499052512  + 412.58354519550 *T)
    sumL0  +=       0.409 * cos(2.51935747849  + 409.07208305770 *T)
    sumL0  +=       0.451 * cos(0.56137272347  + 758.77106321170 *T)
    sumL0  +=       0.399 * cos(0.86037315330  + 337.80194662820 *T)
    sumL0  +=       0.393 * cos(0.40202463200  + 107.49800823370 *T)
    sumL0  +=       0.494 * cos(5.02745190154  + 619.29035849450 *T)
    sumL0  +=       0.404 * cos(0.08539758465  + 18.96239641680 *T)
    sumL0  +=       0.504 * cos(4.18251931021  + 449.49274055590 *T)
    sumL0  +=       0.406 * cos(3.80082989682  + 34.53095077480 *T)
    sumL0  +=       0.545 * cos(2.80919248176  + 514.56435372130 *T)
    sumL0  +=       0.390 * cos(1.65941826256  + 447.20511869550 *T)
    sumL0  +=       0.448 * cos(2.81540452771  + 400.16524680790 *T)
    sumL0  +=       0.405 * cos(4.86073222353  + 1404.08497549710 *T)
    sumL0  +=       0.476 * cos(1.61050626902  + 54.28654533240 *T)
    sumL0  +=       0.406 * cos(1.29798079034  + 226.79247625670 *T)
    sumL0  +=       0.526 * cos(5.35780726572  + 838.21852822500 *T)
    sumL0  +=       0.403 * cos(2.75405589772  + 285.11174858870 *T)
    sumL0  +=       0.381 * cos(4.97702366598  + 309.27832265580 *T)
    sumL0  +=       0.448 * cos(1.37926537411  + 745.27768239300 *T)
    sumL0  +=       0.419 * cos(0.90546724862  + 451.72797278960 *T)
    sumL0  +=       0.450 * cos(1.92391706975  + 474.94684537520 *T)
    sumL0  +=       0.474 * cos(5.54351717465  + 494.26624244250 *T)
    sumL0  +=       0.460 * cos(5.09575399931  + 289.56516671360 *T)
    sumL0  +=       0.455 * cos(3.14755087330  + 168.57377766120 *T)
    sumL0  +=       0.372 * cos(4.87645271422  + 116.53797092750 *T)
    sumL0  +=       0.479 * cos(1.08512503555  + 154.29849954980 *T)
    sumL0  +=       0.374 * cos(0.16389070181  + 1190.78588005910 *T)
    sumL0  +=       0.422 * cos(3.51871257208  + 706.81457835510 *T)
    sumL0  +=       0.354 * cos(4.06405413226  + 124.29040286910 *T)
    sumL0  +=       0.494 * cos(4.57924296149  + 167.72244439930 *T)
    sumL0  +=       0.357 * cos(5.78050145791  + 1265.56747862640 *T)
    sumL0  +=       0.402 * cos(2.67652703260  + 464.47059408290 *T)
    sumL0  +=       0.370 * cos(2.21677703856  + 232.04904353370 *T)
    sumL0  +=       0.462 * cos(4.10424305270  + 27.72047482030 *T)
    sumL0  +=       0.364 * cos(3.07518732480  + 442.63982598600 *T)
    sumL0  +=       0.403 * cos(1.69214233165  + 90.28071695610 *T)
    sumL0  +=       0.348 * cos(1.16647947937  + 357.44566660120 *T)
    sumL0  +=       0.443 * cos(1.32306861852  + 298.23262239190 *T)
    sumL0  +=       0.386 * cos(4.34980428548  + 227.52618943960 *T)
    sumL0  +=       0.337 * cos(5.92030047826  + 445.34813897170 *T)
    sumL0  +=       0.356 * cos(2.38824200660  + 511.59540830470 *T)
    sumL0  +=       0.357 * cos(3.19265737844  + 21.19762865050 *T)
    sumL0  +=       0.406 * cos(3.72223708907  + 3116.26763099790 *T)
    sumL0  +=       0.341 * cos(3.05116722794  + 15.78100267910 *T)
    sumL0  +=       0.379 * cos(4.72472516443  + 30.59879751170 *T)
    sumL0  +=       0.400 * cos(0.71868453904  + 836.52160719560 *T)
    sumL0  +=       0.370 * cos(0.17259001853  + 6531.66165626500 *T)
    sumL0  +=       0.456 * cos(0.83408547295  + 674.80074410430 *T)
    sumL0  +=       0.376 * cos(5.91068811321  + 1617.38407093510 *T)
    sumL0  +=       0.392 * cos(1.90856045571  + 25863.55834587229 *T)
    sumL0  +=       0.333 * cos(4.77074940789  + 76.78733613740 *T)
    sumL0  +=       0.332 * cos(0.85699402720  + 749.20983565610 *T)
    sumL0  +=       0.321 * cos(3.88221470645  + 38.60611638980 *T)
    sumL0  +=       0.321 * cos(2.59404134515  + 1300.99220127850 *T)
    sumL0  +=       0.438 * cos(2.60178805278  + 224.45667028650 *T)
    sumL0  +=       0.371 * cos(3.73501205989  + 328.24071907260 *T)
    sumL0  +=       0.357 * cos(0.03656571669  + 148.81243760920 *T)
    sumL0  +=       0.389 * cos(3.06990362181  + 1012.91150727320 *T)
    sumL0  +=       0.392 * cos(3.14428599675  + 125.18417474640 *T)
    sumL0  +=       0.326 * cos(2.23565995627  + 89.59939339990 *T)
    sumL0  +=       0.314 * cos(4.56810921721  + 1681.11996923850 *T)
    sumL0  +=       0.333 * cos(4.37613329736  + 147.96684984170 *T)
    sumL0  +=       0.326 * cos(4.15448016748  + 21.14944454070 *T)
    sumL0  +=       0.371 * cos(4.09066371754  + 239.16259053450 *T)
    sumL0  +=       0.372 * cos(1.12230345314  + 321.76031151820 *T)
    sumL0  +=       0.305 * cos(6.12924444546  + 19.97378837310 *T)
    sumL0  +=       0.407 * cos(3.65906570714  + 679.25416222920 *T)
    sumL0  +=       0.406 * cos(3.58469900333  + 26013.12154300690 *T)
    sumL0  +=       0.310 * cos(3.92339533494  + 229.08009811710 *T)
    sumL0  +=       0.321 * cos(1.35118535306  + 172.45774681450 *T)
    sumL0  +=       0.332 * cos(3.54513021513  + 288.73508531110 *T)
    sumL0  +=       0.312 * cos(2.87878773413  + 806.72595883600 *T)
    sumL0  +=       0.372 * cos(1.99953045718  + 192.80404220310 *T)
    sumL0  +=       0.302 * cos(0.01867543539  + 501.37978944330 *T)
    sumL0  +=       0.299 * cos(3.96468960950  + 20277.00789528740 *T)
    sumL0  +=       0.308 * cos(2.66235795286  + 248.72381809010 *T)
    sumL0  +=       0.298 * cos(3.52867456736  + 21.45826108140 *T)
    sumL0  +=       0.370 * cos(3.51061046963  + 91.45637312370 *T)
    sumL0  +=       0.297 * cos(0.79872983355  + 742.99006053260 *T)
    sumL0  +=       0.414 * cos(4.81163687199  + 589.49471013490 *T)
    sumL0  +=       0.302 * cos(5.26079338050  + 27.56041612590 *T)
    sumL0  +=       0.314 * cos(3.14643487607  + 361.37781986430 *T)
    sumL0  +=       0.356 * cos(4.63707521448  + 442.75170057060 *T)
    sumL0  +=       0.357 * cos(3.44793069844  + 44.61344319220 *T)
    sumL0  +=       0.370 * cos(4.25032151516  + 304.23420369990 *T)
    sumL0  +=       0.366 * cos(5.43115395433  + 625.99451521810 *T)
    sumL0  +=       0.365 * cos(0.30454498410  + 6283.07584999140 *T)
    sumL0  +=       0.322 * cos(2.32892248876  + 229.45260513260 *T)
    sumL0  +=       0.369 * cos(3.28573365074  + 104.52906281710 *T)
    sumL0  +=       0.327 * cos(2.98588869318  + 348.84764689210 *T)
    sumL0  +=       0.315 * cos(3.66842071994  + 230.93707784090 *T)
    sumL0  +=       0.356 * cos(0.90433599977  + 29.74746424980 *T)
    sumL0  +=       0.384 * cos(0.91820739126  + 549.72844394250 *T)
    sumL0  +=       0.317 * cos(3.86462587284  + 639.89728631400 *T)
    sumL0  +=       0.333 * cos(4.95319798125  + 881.50755740330 *T)
    sumL0  +=       0.285 * cos(4.09883967296  + 904.18965876210 *T)
    sumL0  +=       0.338 * cos(5.65177488491  + 195.77298761970 *T)
    sumL0  +=       0.365 * cos(0.57418860616  + 285.63301345050 *T)
    sumL0  +=       0.363 * cos(0.98999016221  + 839.70300093330 *T)
    sumL0  +=       0.330 * cos(2.26308969695  + 49.17873590170 *T)
    sumL0  +=       0.335 * cos(2.25619157817  + 272.58157561650 *T)
    sumL0  +=       0.376 * cos(5.87496858487  + 268.95823889410 *T)
    sumL0  +=       0.362 * cos(4.94491380965  + 572.22923474750 *T)
    sumL0  +=       0.320 * cos(5.58342880588  + 459.36278465220 *T)
    sumL0  +=       0.299 * cos(2.86286938521  + 883.79517926370 *T)
    sumL0  +=       0.286 * cos(2.49409963193  + 156.04360468900 *T)
    sumL0  +=       0.272 * cos(3.62976505444  + 754.03576079650 *T)
    sumL0  +=       0.349 * cos(2.34615857088  + 6069.77675455340 *T)
    sumL0  +=       0.279 * cos(4.04872155075  + 180.79513409270 *T)
    sumL0  +=       0.328 * cos(1.21350330743  + 148.19059901090 *T)
    sumL0  +=       0.315 * cos(0.74270298817  + 320.27583880990 *T)
    sumL0  +=       0.324 * cos(5.54296698387  + 1507.17774971570 *T)
    sumL0  +=       0.266 * cos(4.36134021576  + 1310.39337013970 *T)
    sumL0  +=       0.262 * cos(5.84934968714  + 450.17406411210 *T)
    sumL0  +=       0.280 * cos(5.05848320657  + 102.52332524700 *T)
    sumL0  +=       0.347 * cos(4.70068639620  + 282.14280317210 *T)
    sumL0  +=       0.342 * cos(5.47365149093  + 163.57784281510 *T)
    sumL0  +=       0.364 * cos(3.29301824378  + 378.90392768260 *T)
    sumL0  +=       0.330 * cos(4.63426494882  + 341.99473232220 *T)
    sumL0  +=       0.259 * cos(2.43682741156  + 170.01006625970 *T)
    sumL0  +=       0.336 * cos(3.79029047358  + 9999.98645077300 *T)
    sumL0  +=       0.284 * cos(2.52583672467  + 266.10116806210 *T)
    sumL0  +=       0.281 * cos(3.93593342516  + 194.17664032680 *T)
    sumL0  +=       0.297 * cos(0.18595848541  + 491.81856188770 *T)
    sumL0  +=       0.290 * cos(4.49575150721  + 151.85081899500 *T)
    sumL0  +=       0.284 * cos(1.59495665161  + 336.83873878170 *T)
    sumL0  +=       0.255 * cos(5.85817353877  + 229.34073054800 *T)
    sumL0  +=       0.260 * cos(5.92834225312  + 455.06942522170 *T)
    sumL0  +=       0.274 * cos(0.53977064975  + 380.23964254460 *T)
    sumL0  +=       0.259 * cos(3.15728797958  + 454.74930783290 *T)
    sumL0  +=       0.303 * cos(0.17851964142  + 384.58118608490 *T)
    sumL0  +=       0.285 * cos(1.29732672572  + 25.12978191360 *T)
    sumL0  +=       0.273 * cos(4.18776699292  + 177.30492381430 *T)
    sumL0  +=       0.326 * cos(1.68159391466  + 161.72086309130 *T)
    sumL0  +=       0.321 * cos(2.36931686576  + 2274.54683263650 *T)
    sumL0  +=       0.241 * cos(3.57660492473  + 150.31395666000 *T)
    sumL0  +=       0.273 * cos(3.78780099400  + 380.01589337540 *T)
    sumL0  +=       0.265 * cos(3.02540120552  + 454.79749194270 *T)
    sumL0  +=       0.265 * cos(6.06087280189  + 455.02124111190 *T)
    sumL0  +=       0.279 * cos(3.82392760479  + 31.54075349880 *T)
    sumL0  +=       0.273 * cos(2.64820862667  + 838.00607990390 *T)
    sumL0  +=       0.238 * cos(5.17487170793  + 263.02034806090 *T)
    sumL0  +=       0.299 * cos(3.94906046599  + 531.97858695500 *T)
    sumL0  +=       0.317 * cos(2.28289083599  + 44.07092647100 *T)
    sumL0  +=       0.250 * cos(1.35950829789  + 304.12232911530 *T)
    sumL0  +=       0.312 * cos(2.73327875294  + 442.37919355510 *T)
    sumL0  +=       0.314 * cos(3.86400459047  + 734.45573129830 *T)
    sumL0  +=       0.251 * cos(0.15148137746  + 221.16340196420 *T)
    sumL0  +=       0.272 * cos(5.71864670101  + 164.54105066160 *T)
    sumL0  +=       0.305 * cos(4.96642198943  + 1140.38330388000 *T)
    sumL0  +=       0.282 * cos(5.46073788901  + 550.13783421970 *T)
    sumL0  +=       0.273 * cos(5.68721459468  + 92.41958097020 *T)
    sumL0  +=       0.256 * cos(0.88787970870  + 418.52143602870 *T)
    sumL0  +=       0.229 * cos(5.02021405557  + 144.89733068860 *T)
    sumL0  +=       0.231 * cos(5.70236752870  + 132.88842257820 *T)
    sumL0  +=       0.218 * cos(2.05623736180  + 303.05854753230 *T)
    sumL0  +=       0.300 * cos(1.76109447754  + 371.52974825090 *T)
    sumL0  +=       0.216 * cos(2.97122807313  + 176.65053250850 *T)
    sumL0  +=       0.224 * cos(3.12798198868  + 188.92007304980 *T)
    sumL0  +=       0.212 * cos(1.30757526083  + 74.14845912090 *T)
    sumL0  +=       0.238 * cos(4.65119406609  + 385.75684225250 *T)
    sumL0  +=       0.216 * cos(0.07086910120  + 893.35640681930 *T)
    sumL0  +=       0.262 * cos(5.78959872639  + 635.96513305090 *T)
    sumL0  +=       0.250 * cos(4.47327859711  + 551.10104206620 *T)
    sumL0  +=       0.222 * cos(1.64692618955  + 76.42612997000 *T)
    sumL0  +=       0.254 * cos(2.61838408005  + 525.23754696970 *T)
    sumL0  +=       0.284 * cos(4.13290731223  + 544.50875992720 *T)
    sumL0  +=       0.262 * cos(2.82476092056  + 971.10695080320 *T)
    sumL0  +=       0.266 * cos(4.12467258610  + 375.67434983510 *T)
    sumL0  +=       0.212 * cos(3.61675003296  + 75.41473801370 *T)
    sumL0  +=       0.253 * cos(3.14224867483  + 270.18207917150 *T)
    sumL0  +=       0.250 * cos(4.32883971376  + 346.44815044710 *T)
    sumL0  +=       0.273 * cos(1.95676918609  + 402.21916848780 *T)
    sumL0  +=       0.259 * cos(1.96371242284  + 968.13800538660 *T)
    sumL0  +=       0.238 * cos(1.29663338057  + 421.18156490460 *T)
    sumL0  +=       0.243 * cos(5.94961177434  + 117.36805233000 *T)
    sumL0  +=       0.241 * cos(4.70849619029  + 406.95447090300 *T)
    sumL0  +=       0.216 * cos(3.31232425021  + 190.66517818900 *T)
    sumL0  +=       0.201 * cos(1.23733749784  + 799.61241183520 *T)
    sumL0  +=       0.223 * cos(0.98087204684  + 627.36711334180 *T)
    sumL0  +=       0.226 * cos(1.66139333004  + 1366.21257229020 *T)
    sumL0  +=       0.205 * cos(0.33839683950  + 143.34342201110 *T)
    sumL0  +=       0.240 * cos(0.71559872262  + 525.75881183150 *T)
    sumL0  +=       0.223 * cos(2.57722930616  + 981.63138620530 *T)
    sumL0  +=       0.193 * cos(4.48974066435  + 172.19711438360 *T)
    sumL0  +=       0.221 * cos(2.88670838151  + 238.90195810360 *T)
    sumL0  +=       0.177 * cos(3.87122035013  + 389.68899551560 *T)
    sumL0  +=       0.175 * cos(5.85737374379  + 170.71264167530 *T)
    sumL0  +=       0.233 * cos(0.63169996424  + 980.66817835880 *T)
    sumL0  +=       0.225 * cos(4.81504648561  + 88.27497938600 *T)
    sumL0  +=       0.210 * cos(2.49819501106  + 128.43500445330 *T)
    sumL0  +=       0.172 * cos(0.91255921858  + 210.85141488320 *T)
    sumL0  +=       0.176 * cos(3.85133296117  + 605.95703637020 *T)
    sumL0  +=       0.211 * cos(1.72999855867  + 10213.28554621100 *T)
    sumL0  +=       0.170 * cos(1.61405340100  + 1512.80682400820 *T)
    sumL0  +=       0.216 * cos(1.54874566441  + 1060.86640289750 *T)
    sumL0  +=       0.194 * cos(6.07381783144  + 520.12973753900 *T)
    sumL0  +=       0.170 * cos(2.58526747515  + 1515.77576942480 *T)
    sumL0  +=       0.172 * cos(1.45073604378  + 995.64603188580 *T)
    sumL0  +=       0.160 * cos(4.20015913513  + 433.71173787680 *T)
    sumL0  +=       0.219 * cos(1.63985385568  + 630.33605875840 *T)
    sumL0  +=       0.219 * cos(1.96273394194  + 313.68355667090 *T)
    sumL0  +=       0.159 * cos(3.62343846627  + 73.97844941520 *T)
    sumL0  +=       0.206 * cos(0.83764718449  + 104.83787935780 *T)
    sumL0  +=       0.203 * cos(0.68701289007  + 1363.24362687360 *T)
    sumL0  +=       0.158 * cos(0.81666479221  + 987.56927703850 *T)
    sumL0  +=       0.172 * cos(2.97938676702  + 327.43756992050 *T)
    sumL0  +=       0.147 * cos(2.15517092627  + 73.24894174920 *T)
    sumL0  +=       0.156 * cos(1.29643469200  + 216.26804085460 *T)
    sumL0  +=       0.153 * cos(3.77492020946  + 768.85355562910 *T)
    sumL0  +=       0.147 * cos(4.74903152642  + 73.34530996880 *T)
    sumL0  +=       0.198 * cos(2.80749830888  + 225.30800354840 *T)
    sumL0  +=       0.179 * cos(0.35410767087  + 421.22974901440 *T)
    sumL0  +=       0.204 * cos(3.47423038287  + 564.85505531580 *T)
    sumL0  +=       0.172 * cos(2.88420642820  + 233.53351624200 *T)
    sumL0  +=       0.195 * cos(3.09114364733  + 294.30046912880 *T)
    sumL0  +=       0.181 * cos(4.75684139861  + 71.15826184490 *T)
    sumL0  +=       0.160 * cos(3.16134209902  + 70.04629615210 *T)
    sumL0  +=       0.155 * cos(2.07144269719  + 91.24392480260 *T)
    sumL0  +=       0.154 * cos(3.37462128317  + 138.40562228610 *T)
    sumL0  +=       0.192 * cos(6.19629054455  + 312.19908396260 *T)
    sumL0  +=       0.162 * cos(0.60896273267  + 73.03649342810 *T)
    sumL0  +=       0.172 * cos(2.01712700688  + 973.55463135800 *T)
    sumL0  +=       0.145 * cos(1.72582731815  + 302.37722397610 *T)
    sumL0  +=       0.170 * cos(2.97437423989  + 3191.04922956520 *T)
    sumL0  +=       0.153 * cos(0.11158566352  + 138.62937145530 *T)
    sumL0  +=       0.164 * cos(5.37129563471  + 457.35704708210 *T)
    sumL0  +=       0.132 * cos(3.15466226029  + 523.47118997110 *T)
    sumL0  +=       0.137 * cos(2.88313323946  + 765.88461021250 *T)
    sumL0  +=       0.128 * cos(1.75773421717  + 77.70235987410 *T)
    sumL0  +=       0.173 * cos(5.03552846066  + 415.29185818120 *T)
    sumL0  +=       0.137 * cos(4.77137510538  + 73.45718455340 *T)
    sumL0  +=       0.163 * cos(0.99316178485  + 94.42531854030 *T)
    sumL0  +=       0.138 * cos(4.68330115148  + 517.16079212240 *T)
    sumL0  +=       0.133 * cos(2.95376828791  + 75.15410558280 *T)
    sumL0  +=       0.134 * cos(1.35779558361  + 249.94765836750 *T)
    sumL0  +=       0.168 * cos(4.73164542970  + 108.72184851110 *T)
    sumL0  +=       0.119 * cos(1.48229220689  + 237.41748539530 *T)
    sumL0  +=       0.134 * cos(1.52277177920  + 154.67100656530 *T)
    sumL0  +=       0.117 * cos(5.48642273610  + 437.64389113990 *T)
    sumL0  +=       0.131 * cos(5.76525548944  + 75.43598987310 *T)
    sumL0  +=       0.127 * cos(2.38309227230  + 208.84567731310 *T)
    sumL0  +=       0.115 * cos(4.38951350436  + 343.21857259960 *T)
    sumL0  +=       0.114 * cos(4.48142283161  + 224.86606056370 *T)
    sumL0  +=       0.117 * cos(5.37802827323  + 293.70976829780 *T)
    sumL0  +=       0.136 * cos(2.80772094137  + 374.49869366750 *T)

    # URANUS Longitude
    sumL1  = 0
    sumL1  +=  7502543121.646 * cos(0.00000000000  + 0.00000000000 *T)
    sumL1  +=  154458.244 * cos(5.24201658072  + 74.78159856730 *T)
    sumL1  +=   24456.413 * cos(1.71255705309  + 1.48447270830 *T)
    sumL1  +=    9257.828 * cos(0.42844639064  + 11.04570026390 *T)
    sumL1  +=    8265.977 * cos(1.50220035110  + 63.73589830340 *T)
    sumL1  +=    7841.715 * cos(1.31983607251  + 149.56319713460 *T)
    sumL1  +=    3899.105 * cos(0.46483574024  + 3.93215326310 *T)
    sumL1  +=    2283.777 * cos(4.17367533997  + 76.26607127560 *T)
    sumL1  +=    1926.600 * cos(0.53013080152  + 2.96894541660 *T)
    sumL1  +=    1232.727 * cos(1.58634458237  + 70.84944530420 *T)
    sumL1  +=     791.206 * cos(5.43641224143  + 3.18139373770 *T)
    sumL1  +=     766.954 * cos(1.99555409575  + 73.29712585900 *T)
    sumL1  +=     481.671 * cos(2.98401996914  + 85.82729883120 *T)
    sumL1  +=     449.798 * cos(4.13826237508  + 138.51749687070 *T)
    sumL1  +=     445.600 * cos(3.72300400331  + 224.34479570190 *T)
    sumL1  +=     426.554 * cos(4.73126059388  + 71.81265315070 *T)
    sumL1  +=     347.735 * cos(2.45372261286  + 9.56122755560 *T)
    sumL1  +=     353.752 * cos(2.58324496886  + 148.07872442630 *T)
    sumL1  +=     317.084 * cos(5.57855232072  + 52.69019803950 *T)
    sumL1  +=     179.920 * cos(5.68367730922  + 12.53017297220 *T)
    sumL1  +=     171.084 * cos(3.00060075287  + 78.71375183040 *T)
    sumL1  +=     205.585 * cos(2.36263144251  + 2.44768055480 *T)
    sumL1  +=     158.029 * cos(2.90931969498  + 0.96320784650 *T)
    sumL1  +=     189.068 * cos(4.20242881378  + 56.62235130260 *T)
    sumL1  +=     154.670 * cos(5.59083925605  + 4.45341812490 *T)
    sumL1  +=     183.762 * cos(0.28371004654  + 151.04766984290 *T)
    sumL1  +=     143.464 * cos(2.59049246726  + 62.25142559510 *T)
    sumL1  +=     151.984 * cos(2.94217326890  + 77.75054398390 *T)
    sumL1  +=     153.515 * cos(4.65186885939  + 35.16409022120 *T)
    sumL1  +=     121.452 * cos(4.14839204920  + 127.47179660680 *T)
    sumL1  +=     115.546 * cos(3.73224603791  + 65.22037101170 *T)
    sumL1  +=     102.022 * cos(4.18754517993  + 145.63104387150 *T)
    sumL1  +=     101.718 * cos(6.03385875009  + 0.11187458460 *T)
    sumL1  +=      88.202 * cos(3.99035787994  + 18.15924726470 *T)
    sumL1  +=      87.549 * cos(6.15520787584  + 202.25339517410 *T)
    sumL1  +=      80.530 * cos(2.64124743934  + 22.09140052780 *T)
    sumL1  +=      72.047 * cos(6.04545933578  + 70.32818044240 *T)
    sumL1  +=      68.570 * cos(4.05071895264  + 77.96299230500 *T)
    sumL1  +=      59.173 * cos(3.70413919082  + 67.66805156650 *T)
    sumL1  +=      47.267 * cos(3.54312460519  + 351.81659230870 *T)
    sumL1  +=      42.534 * cos(5.72357370899  + 5.41662597140 *T)
    sumL1  +=      44.339 * cos(5.90865821911  + 7.11354700080 *T)
    sumL1  +=      35.605 * cos(3.29197259183  + 8.07675484730 *T)
    sumL1  +=      35.524 * cos(3.32784616138  + 71.60020482960 *T)
    sumL1  +=      36.116 * cos(5.89964278801  + 33.67961751290 *T)
    sumL1  +=      30.608 * cos(5.46414592601  + 160.60889739850 *T)
    sumL1  +=      31.454 * cos(5.62015632303  + 984.60033162190 *T)
    sumL1  +=      38.544 * cos(4.91519003848  + 222.86032299360 *T)
    sumL1  +=      34.996 * cos(5.08034112149  + 38.13303563780 *T)
    sumL1  +=      30.811 * cos(5.49591403863  + 59.80374504030 *T)
    sumL1  +=      28.947 * cos(4.51867390414  + 84.34282612290 *T)
    sumL1  +=      26.627 * cos(5.54127301037  + 131.40394986990 *T)
    sumL1  +=      29.866 * cos(1.65980844667  + 447.79581952650 *T)
    sumL1  +=      29.206 * cos(1.14722640419  + 462.02291352810 *T)
    sumL1  +=      25.753 * cos(4.99362028417  + 137.03302416240 *T)
    sumL1  +=      25.373 * cos(5.73584678604  + 380.12776796000 *T)
    sumL1  +=      21.672 * cos(2.80556379586  + 69.36497259590 *T)
    sumL1  +=      26.605 * cos(6.14640604128  + 299.12639426920 *T)
    sumL1  +=      22.995 * cos(2.24925345862  + 111.43016149680 *T)
    sumL1  +=      19.246 * cos(3.55645739672  + 54.17467074780 *T)
    sumL1  +=      21.780 * cos(0.93285892393  + 213.29909543800 *T)
    sumL1  +=      19.338 * cos(1.86249384092  + 108.46121608020 *T)
    sumL1  +=      16.153 * cos(3.10208165842  + 14.97785352700 *T)
    sumL1  +=      13.126 * cos(1.95385539499  + 87.31177153950 *T)
    sumL1  +=      13.907 * cos(1.54149045800  + 340.77089204480 *T)
    sumL1  +=      13.549 * cos(4.38455126720  + 5.93789083320 *T)
    sumL1  +=      13.102 * cos(5.88301410143  + 6.21977512350 *T)
    sumL1  +=      11.810 * cos(0.32615567587  + 35.42472265210 *T)
    sumL1  +=      10.980 * cos(1.69230280951  + 45.57665103870 *T)
    sumL1  +=      12.351 * cos(0.32823896833  + 51.20572533120 *T)
    sumL1  +=      10.906 * cos(5.97068444790  + 265.98929347750 *T)
    sumL1  +=      11.446 * cos(3.37831545858  + 72.33391801250 *T)
    sumL1  +=      12.013 * cos(3.60395709253  + 269.92144674060 *T)
    sumL1  +=      11.662 * cos(1.74504271366  + 79.23501669220 *T)
    sumL1  +=      13.777 * cos(2.69028726334  + 225.82926841020 *T)
    sumL1  +=      12.006 * cos(5.34430562395  + 152.53214255120 *T)
    sumL1  +=       9.866 * cos(5.50316093605  + 153.49535039770 *T)
    sumL1  +=      10.436 * cos(4.16875643286  + 24.37902238820 *T)
    sumL1  +=      10.632 * cos(3.06875158069  + 284.14854074220 *T)
    sumL1  +=       9.613 * cos(0.49590148788  + 209.36694217490 *T)
    sumL1  +=       9.283 * cos(3.54479191952  + 41.64449777560 *T)
    sumL1  +=       9.536 * cos(5.60054956443  + 82.85835341460 *T)
    sumL1  +=       9.740 * cos(1.01087744586  + 68.84370773410 *T)
    sumL1  +=       9.187 * cos(4.49472579228  + 20.60692781950 *T)
    sumL1  +=      10.159 * cos(3.51765739489  + 529.69096509460 *T)
    sumL1  +=       8.612 * cos(3.88869873588  + 60.76695288680 *T)
    sumL1  +=      10.030 * cos(4.64790204580  + 77.22927912210 *T)
    sumL1  +=       8.689 * cos(1.96813580258  + 195.13984817330 *T)
    sumL1  +=       8.370 * cos(4.40914764204  + 134.58534360760 *T)
    sumL1  +=       9.273 * cos(3.93291227900  + 39.61750834610 *T)
    sumL1  +=       7.784 * cos(5.35626068469  + 75.74480641380 *T)
    sumL1  +=       7.724 * cos(5.77176047568  + 73.81839072080 *T)
    sumL1  +=       7.683 * cos(4.44252070929  + 14.01464568050 *T)
    sumL1  +=       8.355 * cos(2.44425910430  + 146.59425171800 *T)
    sumL1  +=       7.954 * cos(5.73093878181  + 184.72728735580 *T)
    sumL1  +=       7.465 * cos(2.18972405572  + 145.10977900970 *T)
    sumL1  +=       6.430 * cos(0.84582374839  + 32.19514480460 *T)
    sumL1  +=       6.257 * cos(2.17085130003  + 74.89347315190 *T)
    sumL1  +=       7.911 * cos(0.17275924476  + 120.35824960600 *T)
    sumL1  +=       7.036 * cos(4.12047266896  + 191.20769491020 *T)
    sumL1  +=       6.860 * cos(2.13462553365  + 116.42609634290 *T)
    sumL1  +=       5.191 * cos(3.11155355454  + 106.97674337190 *T)
    sumL1  +=       4.798 * cos(2.25093144226  + 46.20979048510 *T)
    sumL1  +=       4.566 * cos(3.45427648666  + 0.75075952540 *T)
    sumL1  +=       4.401 * cos(3.94058045671  + 6.59228213900 *T)
    sumL1  +=       4.214 * cos(5.17805765625  + 144.14657116320 *T)
    sumL1  +=       4.409 * cos(0.24427052932  + 92.94084583200 *T)
    sumL1  +=       4.866 * cos(1.15344187054  + 112.91463420510 *T)
    sumL1  +=       4.744 * cos(5.18229292013  + 81.00137369080 *T)
    sumL1  +=       4.332 * cos(2.52429167546  + 99.16062095550 *T)
    sumL1  +=       3.876 * cos(2.78430217652  + 565.11568774670 *T)
    sumL1  +=       3.801 * cos(0.75133837939  + 58.10682401090 *T)
    sumL1  +=       4.146 * cos(5.84943984597  + 221.37585028530 *T)
    sumL1  +=       3.885 * cos(4.95626104286  + 125.98732389850 *T)
    sumL1  +=       3.815 * cos(3.23004401930  + 479.28838891550 *T)
    sumL1  +=       3.679 * cos(5.28098232097  + 66.91729204110 *T)
    sumL1  +=       3.479 * cos(2.95514470947  + 74.66972398270 *T)
    sumL1  +=       3.514 * cos(4.90090391308  + 28.31117565130 *T)
    sumL1  +=       4.515 * cos(4.15474629145  + 344.70304530790 *T)
    sumL1  +=       4.036 * cos(2.28903172191  + 109.94568878850 *T)
    sumL1  +=       4.266 * cos(2.68534591099  + 7.86430652620 *T)
    sumL1  +=       3.428 * cos(0.02846652682  + 140.00196957900 *T)
    sumL1  +=       3.644 * cos(5.32002093810  + 408.43894361130 *T)
    sumL1  +=       3.252 * cos(1.44975192429  + 128.95626931510 *T)
    sumL1  +=       4.143 * cos(1.89070487241  + 277.03499374140 *T)
    sumL1  +=       3.177 * cos(0.04197149544  + 220.41264243880 *T)
    sumL1  +=       3.901 * cos(6.25926496244  + 0.89377187730 *T)
    sumL1  +=       3.787 * cos(0.02516903921  + 152.74459087230 *T)
    sumL1  +=       3.200 * cos(0.52009458683  + 2.28762186040 *T)
    sumL1  +=       2.995 * cos(1.94615440691  + 80.19822453870 *T)
    sumL1  +=       4.029 * cos(5.24603808726  + 96.87299909510 *T)
    sumL1  +=       3.302 * cos(4.81033551060  + 422.66603761290 *T)
    sumL1  +=       3.189 * cos(6.26156603400  + 456.39383923560 *T)
    sumL1  +=       2.804 * cos(1.35626949052  + 404.50679034820 *T)
    sumL1  +=       2.970 * cos(0.54327361056  + 159.12442469020 *T)
    sumL1  +=       3.465 * cos(5.88337990735  + 16.67477455640 *T)
    sumL1  +=       3.518 * cos(4.99649404130  + 36.64856292950 *T)
    sumL1  +=       3.081 * cos(2.82772472086  + 453.42489381900 *T)
    sumL1  +=       3.320 * cos(1.56223495893  + 23.57587323610 *T)
    sumL1  +=       2.573 * cos(6.19617997586  + 135.54855145410 *T)
    sumL1  +=       2.547 * cos(5.19937103778  + 173.94221952280 *T)
    sumL1  +=       2.534 * cos(1.85452635674  + 490.33408917940 *T)
    sumL1  +=       3.106 * cos(6.07067928601  + 142.44965013380 *T)
    sumL1  +=       3.302 * cos(1.02846689671  + 297.64192156090 *T)
    sumL1  +=       2.429 * cos(1.33640100979  + 211.81462272970 *T)
    sumL1  +=       2.792 * cos(3.89897022917  + 358.93013930950 *T)
    sumL1  +=       2.947 * cos(5.31528985588  + 55.13787859430 *T)
    sumL1  +=       2.449 * cos(3.44007536754  + 206.18554843720 *T)
    sumL1  +=       2.407 * cos(4.38551271701  + 60.55450456570 *T)
    sumL1  +=       2.425 * cos(2.22643225523  + 66.70484372000 *T)
    sumL1  +=       2.295 * cos(2.31690029267  + 31.49256938900 *T)
    sumL1  +=       2.225 * cos(0.41365126245  + 81.37388070630 *T)
    sumL1  +=       2.196 * cos(0.76281798713  + 17.52610781830 *T)
    sumL1  +=       2.301 * cos(3.60815987923  + 288.08069400530 *T)
    sumL1  +=       2.557 * cos(0.73679737974  + 200.76892246580 *T)
    sumL1  +=       2.158 * cos(2.61924330277  + 13.33332212430 *T)
    sumL1  +=       2.048 * cos(6.27204714771  + 98.89998852460 *T)
    sumL1  +=       2.054 * cos(3.61072687338  + 333.65734504400 *T)
    sumL1  +=       2.190 * cos(2.49696729700  + 76.47851959670 *T)
    sumL1  +=       2.092 * cos(1.66496421654  + 235.39049596580 *T)
    sumL1  +=       2.206 * cos(2.35938756479  + 347.88443904560 *T)
    sumL1  +=       2.469 * cos(4.70656858928  + 186.21176006410 *T)
    sumL1  +=       2.226 * cos(5.97327738150  + 1514.29129671650 *T)
    sumL1  +=       1.851 * cos(2.19455296942  + 203.73786788240 *T)
    sumL1  +=       1.865 * cos(4.98207204280  + 5.10780943070 *T)
    sumL1  +=       2.171 * cos(5.49034081907  + 373.01422095920 *T)
    sumL1  +=       1.999 * cos(5.80509154216  + 146.38180339690 *T)
    sumL1  +=       1.903 * cos(4.32950489567  + 49.50880430180 *T)
    sumL1  +=       1.732 * cos(3.94794012202  + 24.11838995730 *T)
    sumL1  +=       1.747 * cos(2.46883637489  + 55.65914345610 *T)
    sumL1  +=       1.833 * cos(3.35110048460  + 143.62530630140 *T)
    sumL1  +=       1.686 * cos(1.28621563322  + 103.09277421860 *T)
    sumL1  +=       1.720 * cos(2.35857527806  + 1.64453140270 *T)
    sumL1  +=       1.641 * cos(2.99507314472  + 391.17346822390 *T)
    sumL1  +=       1.610 * cos(0.97420709262  + 977.48678462110 *T)
    sumL1  +=       1.696 * cos(4.98332661473  + 387.24131496080 *T)
    sumL1  +=       1.527 * cos(3.15107379811  + 7.42236354150 *T)
    sumL1  +=       1.570 * cos(1.61119571428  + 991.71387862270 *T)
    sumL1  +=       1.497 * cos(2.89637638984  + 19.64371997300 *T)
    sumL1  +=       1.507 * cos(3.32822127349  + 909.81873305460 *T)
    sumL1  +=       1.375 * cos(5.75263837916  + 19.12245511120 *T)
    sumL1  +=       1.407 * cos(2.20244941425  + 67.35923502580 *T)
    sumL1  +=       1.364 * cos(4.40006421418  + 27.08733537390 *T)
    sumL1  +=       1.357 * cos(4.33780029649  + 70.11573212130 *T)
    sumL1  +=       1.311 * cos(4.62202930578  + 81.89514556810 *T)
    sumL1  +=       1.307 * cos(2.79964247834  + 25.60286266560 *T)
    sumL1  +=       1.312 * cos(3.73623252660  + 628.85158605010 *T)
    sumL1  +=       1.286 * cos(3.96557527092  + 61.28821774860 *T)
    sumL1  +=       1.723 * cos(4.56068809303  + 305.34616939270 *T)
    sumL1  +=       1.313 * cos(4.90611014973  + 617.80588578620 *T)
    sumL1  +=       1.508 * cos(6.25017976193  + 604.47256366190 *T)
    sumL1  +=       1.235 * cos(5.93779486368  + 415.55249061210 *T)
    sumL1  +=       1.278 * cos(3.21119872139  + 92.04707395470 *T)
    sumL1  +=       1.357 * cos(0.72647086107  + 546.95644048200 *T)
    sumL1  +=       1.552 * cos(5.05296247763  + 10.29494073850 *T)
    sumL1  +=       1.230 * cos(1.52077038294  + 157.63995198190 *T)
    sumL1  +=       1.210 * cos(2.63049415027  + 426.59819087600 *T)
    sumL1  +=       1.206 * cos(4.83219370572  + 100.38446123290 *T)
    sumL1  +=       1.234 * cos(4.46203104116  + 162.09337010680 *T)
    sumL1  +=       1.174 * cos(5.32356191090  + 17.26547538740 *T)
    sumL1  +=       1.431 * cos(6.18138614295  + 14.22709400160 *T)
    sumL1  +=       1.244 * cos(0.16929250603  + 29.20494752860 *T)
    sumL1  +=       1.180 * cos(4.09719023908  + 443.86366626340 *T)
    sumL1  +=       1.180 * cos(3.31438239649  + 44.72531777680 *T)
    sumL1  +=       1.259 * cos(1.88793196065  + 0.65439130580 *T)
    sumL1  +=       1.263 * cos(3.49967730885  + 230.56457082540 *T)
    sumL1  +=       1.168 * cos(2.04071854201  + 30.71067209630 *T)
    sumL1  +=       1.523 * cos(2.28101186489  + 373.90799283650 *T)
    sumL1  +=       1.429 * cos(2.05075136274  + 181.75834193920 *T)
    sumL1  +=       1.065 * cos(2.95960854361  + 241.61027108930 *T)
    sumL1  +=       1.253 * cos(0.23639539817  + 561.18353448360 *T)
    sumL1  +=       1.255 * cos(1.25819925760  + 155.78297225810 *T)
    sumL1  +=       1.044 * cos(2.89293032709  + 543.02428721890 *T)
    sumL1  +=       1.062 * cos(3.26314901318  + 28.57180808220 *T)
    sumL1  +=       1.124 * cos(1.06535506684  + 88.11492069160 *T)
    sumL1  +=       1.186 * cos(5.73445278027  + 329.72519178090 *T)
    sumL1  +=       1.190 * cos(2.82438170535  + 41.10198105440 *T)
    sumL1  +=       1.067 * cos(0.27101806190  + 58.31927233200 *T)
    sumL1  +=       1.017 * cos(4.30527610005  + 67.88049988760 *T)
    sumL1  +=       0.959 * cos(5.20504598622  + 42.53826965290 *T)
    sumL1  +=       0.984 * cos(4.90934403664  + 465.95506679120 *T)
    sumL1  +=       0.944 * cos(0.66925769374  + 88.79624424780 *T)
    sumL1  +=       1.017 * cos(4.37095088461  + 13.49338081870 *T)
    sumL1  +=       1.222 * cos(5.13450955699  + 300.61086697750 *T)
    sumL1  +=       0.989 * cos(0.53937909300  + 80.71948940050 *T)
    sumL1  +=       0.890 * cos(3.09802121989  + 110.20632121940 *T)
    sumL1  +=       1.095 * cos(1.70637576740  + 43.12897048390 *T)
    sumL1  +=       0.992 * cos(4.17968869928  + 154.01661525950 *T)
    sumL1  +=       0.950 * cos(0.09841899432  + 273.10284047830 *T)
    sumL1  +=       0.870 * cos(4.77500238443  + 33.13710079170 *T)
    sumL1  +=       0.867 * cos(4.22078052532  + 20.44686912510 *T)
    sumL1  +=       0.830 * cos(5.23785245773  + 472.17484191470 *T)
    sumL1  +=       0.884 * cos(4.34377463442  + 105.49227066360 *T)
    sumL1  +=       0.812 * cos(3.53258780148  + 39.35687591520 *T)
    sumL1  +=       1.055 * cos(1.52219418153  + 227.31374111850 *T)
    sumL1  +=       0.963 * cos(1.87806076896  + 259.50888592310 *T)
    sumL1  +=       0.859 * cos(0.57844232244  + 152.01087768940 *T)
    sumL1  +=       0.861 * cos(4.69213709412  + 1059.38193018920 *T)
    sumL1  +=       1.084 * cos(2.79612346618  + 48.75804477640 *T)
    sumL1  +=       0.994 * cos(2.87052008214  + 454.90936652730 *T)
    sumL1  +=       0.831 * cos(1.62068330602  + 554.06998748280 *T)
    sumL1  +=       0.891 * cos(2.85026036860  + 32.24332891440 *T)
    sumL1  +=       0.876 * cos(0.83921717739  + 4.73530241520 *T)
    sumL1  +=       0.707 * cos(6.16918394997  + 3.62333672240 *T)
    sumL1  +=       0.787 * cos(1.95585343912  + 16.46232623530 *T)
    sumL1  +=       0.702 * cos(5.49557046240  + 558.00214074590 *T)
    sumL1  +=       0.817 * cos(0.38724470336  + 378.64329525170 *T)
    sumL1  +=       0.804 * cos(2.25693582099  + 16.04163511000 *T)
    sumL1  +=       0.866 * cos(1.80814575866  + 258.87574647670 *T)
    sumL1  +=       0.651 * cos(3.72120167607  + 286.59622129700 *T)
    sumL1  +=       0.672 * cos(1.00052727778  + 522.57741809380 *T)
    sumL1  +=       0.631 * cos(4.14839739363  + 141.69889060840 *T)
    sumL1  +=       0.748 * cos(4.19441869839  + 486.40193591630 *T)
    sumL1  +=       0.668 * cos(0.77754011576  + 120.99138905240 *T)
    sumL1  +=       0.619 * cos(4.31040053492  + 455.87257437380 *T)
    sumL1  +=       0.619 * cos(4.77556598202  + 453.94615868080 *T)
    sumL1  +=       0.647 * cos(5.74952736928  + 119.50691634410 *T)
    sumL1  +=       0.609 * cos(0.24149609998  + 117.91056905120 *T)
    sumL1  +=       0.630 * cos(1.79018649942  + 440.68227252570 *T)
    sumL1  +=       0.601 * cos(1.41196883461  + 218.92816973050 *T)
    sumL1  +=       0.719 * cos(4.27398947015  + 50.40257617910 *T)
    sumL1  +=       0.594 * cos(3.92150462249  + 25.27279426550 *T)
    sumL1  +=       0.710 * cos(0.45768559438  + 536.80451209540 *T)
    sumL1  +=       0.706 * cos(6.15599144951  + 258.02441321480 *T)
    sumL1  +=       0.617 * cos(2.80636989892  + 68.56182344380 *T)
    sumL1  +=       0.587 * cos(5.47247350993  + 767.36908292080 *T)
    sumL1  +=       0.690 * cos(3.48978614301  + 835.03713448730 *T)
    sumL1  +=       0.537 * cos(4.06668446648  + 450.97721326420 *T)
    sumL1  +=       0.511 * cos(0.60155300709  + 264.50482076920 *T)
    sumL1  +=       0.694 * cos(1.18127476921  + 129.91947716160 *T)
    sumL1  +=       0.584 * cos(1.94104733057  + 106.27416795630 *T)
    sumL1  +=       0.522 * cos(5.95180617510  + 518.64526483070 *T)
    sumL1  +=       0.507 * cos(4.39658523394  + 121.84272231430 *T)
    sumL1  +=       0.627 * cos(2.24582628581  + 218.40690486870 *T)
    sumL1  +=       0.485 * cos(0.02058107411  + 106.01353552540 *T)
    sumL1  +=       0.592 * cos(2.06072766194  + 296.15744885260 *T)
    sumL1  +=       0.587 * cos(0.18557470860  + 458.09076026500 *T)
    sumL1  +=       0.483 * cos(1.50333774574  + 150.52640498110 *T)
    sumL1  +=       0.474 * cos(4.99848521665  + 458.84151979040 *T)
    sumL1  +=       0.566 * cos(1.94435189030  + 699.70103135430 *T)
    sumL1  +=       0.472 * cos(1.86519720200  + 180.16199464630 *T)
    sumL1  +=       0.472 * cos(0.07145793467  + 216.48048917570 *T)
    sumL1  +=       0.571 * cos(6.01195273302  + 47.06112374700 *T)
    sumL1  +=       0.460 * cos(3.76890954025  + 342.25536475310 *T)
    sumL1  +=       0.489 * cos(2.96084966272  + 385.75684225250 *T)
    sumL1  +=       0.458 * cos(1.99730631732  + 275.55052103310 *T)
    sumL1  +=       0.460 * cos(5.75982407113  + 89.75945209430 *T)
    sumL1  +=       0.549 * cos(1.43219978325  + 171.65459766240 *T)
    sumL1  +=       0.544 * cos(0.04821904056  + 114.39910691340 *T)
    sumL1  +=       0.450 * cos(1.94933296558  + 148.59998928810 *T)
    sumL1  +=       0.444 * cos(2.94093732205  + 692.58748435350 *T)
    sumL1  +=       0.442 * cos(2.15938034999  + 173.68158709190 *T)
    sumL1  +=       0.543 * cos(2.61197342701  + 451.72797278960 *T)
    sumL1  +=       0.465 * cos(0.31777753866  + 756.32338265690 *T)
    sumL1  +=       0.441 * cos(2.82271922049  + 32.71640966640 *T)
    sumL1  +=       0.538 * cos(2.39420182072  + 339.28641933650 *T)
    sumL1  +=       0.569 * cos(0.84686482736  + 260.99335863140 *T)
    sumL1  +=       0.572 * cos(5.40379754526  + 278.51946644970 *T)
    sumL1  +=       0.422 * cos(4.61520857062  + 40.16002506730 *T)
    sumL1  +=       0.451 * cos(4.50911201020  + 142.14083359310 *T)
    sumL1  +=       0.501 * cos(0.18290112601  + 331.32153907380 *T)
    sumL1  +=       0.468 * cos(0.97688759019  + 760.25553592000 *T)
    sumL1  +=       0.443 * cos(4.58896013561  + 149.67507171920 *T)
    sumL1  +=       0.428 * cos(1.02564850231  + 469.13646052890 *T)
    sumL1  +=       0.500 * cos(4.34235307579  + 166.82867252200 *T)
    sumL1  +=       0.412 * cos(5.69502940499  + 92.30770638560 *T)
    sumL1  +=       0.404 * cos(5.18855270166  + 22.63391724900 *T)
    sumL1  +=       0.396 * cos(3.98515136901  + 31.23193695810 *T)
    sumL1  +=       0.421 * cos(5.47567810199  + 104.00779795530 *T)
    sumL1  +=       0.425 * cos(3.50702044406  + 180.27386923090 *T)
    sumL1  +=       0.415 * cos(1.52291071520  + 497.44763618020 *T)
    sumL1  +=       0.430 * cos(2.39159932023  + 39.09624348430 *T)
    sumL1  +=       0.401 * cos(0.55271143649  + 95.38852638680 *T)
    sumL1  +=       0.384 * cos(2.48712922138  + 210.33015002140 *T)
    sumL1  +=       0.422 * cos(1.02056886848  + 468.24268865160 *T)
    sumL1  +=       0.465 * cos(5.72323435231  + 183.24281464750 *T)
    sumL1  +=       0.383 * cos(2.63486938783  + 685.47393735270 *T)
    sumL1  +=       0.367 * cos(5.39331524988  + 874.39401040250 *T)
    sumL1  +=       0.461 * cos(3.57961254790  + 187.69623277240 *T)
    sumL1  +=       0.409 * cos(4.21780704807  + 181.05576652360 *T)
    sumL1  +=       0.440 * cos(0.36380766054  + 367.97010200330 *T)
    sumL1  +=       0.392 * cos(5.44355925956  + 26.02355379090 *T)
    sumL1  +=       0.431 * cos(3.83885208954  + 254.94359321360 *T)
    sumL1  +=       0.366 * cos(2.92275490656  + 291.26208774300 *T)
    sumL1  +=       0.416 * cos(2.54190330826  + 255.05546779820 *T)
    sumL1  +=       0.348 * cos(0.35176743482  + 46.47042291600 *T)
    sumL1  +=       0.413 * cos(2.41518097006  + 483.22054217860 *T)
    sumL1  +=       0.386 * cos(4.76483292968  + 268.43697403230 *T)
    sumL1  +=       0.344 * cos(0.20350283971  + 184.09414790940 *T)
    sumL1  +=       0.350 * cos(1.24205287122  + 97.41551581630 *T)
    sumL1  +=       0.361 * cos(5.68393391400  + 353.30106501700 *T)
    sumL1  +=       0.359 * cos(2.62171903648  + 162.89651925890 *T)
    sumL1  +=       0.381 * cos(3.38777292581  + 114.94162363460 *T)
    sumL1  +=       0.352 * cos(2.76374792259  + 295.19424100610 *T)
    sumL1  +=       0.340 * cos(1.34666360560  + 34.20088237470 *T)
    sumL1  +=       0.433 * cos(1.90504858871  + 123.53964334370 *T)
    sumL1  +=       0.389 * cos(2.41268196916  + 381.61224066830 *T)
    sumL1  +=       0.383 * cos(3.20416581825  + 79.44746501330 *T)
    sumL1  +=       0.369 * cos(2.15185889720  + 555.55446019110 *T)
    sumL1  +=       0.370 * cos(2.19402183275  + 562.66800719190 *T)
    sumL1  +=       0.327 * cos(3.40081544565  + 309.27832265580 *T)
    sumL1  +=       0.378 * cos(5.75737470182  + 916.93228005540 *T)
    sumL1  +=       0.318 * cos(4.53066393124  + 350.33211960040 *T)
    sumL1  +=       0.376 * cos(1.74845257914  + 545.47196777370 *T)
    sumL1  +=       0.346 * cos(4.15815107375  + 282.66406803390 *T)
    sumL1  +=       0.319 * cos(4.38123849114  + 154.97982310600 *T)
    sumL1  +=       0.320 * cos(0.81846631878  + 610.69233878540 *T)
    sumL1  +=       0.327 * cos(5.07873875355  + 189.72322220190 *T)
    sumL1  +=       0.306 * cos(1.71903179875  + 394.35486196160 *T)
    sumL1  +=       0.329 * cos(1.82999432252  + 706.81457835510 *T)
    sumL1  +=       0.335 * cos(4.78622577105  + 109.31254934210 *T)
    sumL1  +=       0.310 * cos(5.08120849869  + 376.19561469690 *T)
    sumL1  +=       0.327 * cos(1.87637598331  + 207.88246946660 *T)
    sumL1  +=       0.323 * cos(1.88845451800  + 192.69216761850 *T)
    sumL1  +=       0.284 * cos(2.88222063053  + 384.05992122310 *T)
    sumL1  +=       0.283 * cos(4.63187254084  + 332.17287233570 *T)
    sumL1  +=       0.294 * cos(2.84554743359  + 267.47376618580 *T)
    sumL1  +=       0.285 * cos(0.97965330777  + 113.87784205160 *T)
    sumL1  +=       0.319 * cos(5.09582764612  + 285.63301345050 *T)
    sumL1  +=       0.280 * cos(0.67871105907  + 312.45971639350 *T)
    sumL1  +=       0.300 * cos(5.93285242876  + 124.29040286910 *T)
    sumL1  +=       0.320 * cos(4.86151247369  + 448.68959140380 *T)
    sumL1  +=       0.310 * cos(3.75000484412  + 253.57099508990 *T)
    sumL1  +=       0.311 * cos(5.59686590720  + 271.40591944890 *T)
    sumL1  +=       0.316 * cos(1.89686876211  + 228.27694896500 *T)
    sumL1  +=       0.269 * cos(0.14585942744  + 142.66209845490 *T)
    sumL1  +=       0.270 * cos(2.12904548682  + 778.41478318470 *T)
    sumL1  +=       0.267 * cos(0.96560769114  + 90.82323367730 *T)
    sumL1  +=       0.308 * cos(1.38454900684  + 375.39246554480 *T)
    sumL1  +=       0.298 * cos(3.99595366039  + 451.94042111070 *T)
    sumL1  +=       0.278 * cos(3.38339026214  + 346.39996633730 *T)
    sumL1  +=       0.287 * cos(1.01918432834  + 905.88657979150 *T)
    sumL1  +=       0.263 * cos(0.16921968622  + 124.50285119020 *T)
    sumL1  +=       0.283 * cos(5.95865378023  + 362.86229257260 *T)
    sumL1  +=       0.266 * cos(3.07331582044  + 193.65537546500 *T)
    sumL1  +=       0.264 * cos(5.47114459575  + 133.10087089930 *T)
    sumL1  +=       0.288 * cos(2.78232740152  + 1812.52391910840 *T)
    sumL1  +=       0.361 * cos(4.30140629884  + 198.32124191100 *T)
    sumL1  +=       0.257 * cos(1.60206491208  + 369.08206769610 *T)
    sumL1  +=       0.274 * cos(2.88347680082  + 233.90602325750 *T)
    sumL1  +=       0.267 * cos(4.90554019072  + 681.54178408960 *T)
    sumL1  +=       0.305 * cos(1.55983861329  + 49.72125262290 *T)
    sumL1  +=       0.253 * cos(0.50457429429  + 316.39186965660 *T)
    sumL1  +=       0.258 * cos(5.81453094409  + 630.33605875840 *T)
    sumL1  +=       0.280 * cos(1.15452517706  + 986.08480433020 *T)
    sumL1  +=       0.265 * cos(4.93584097286  + 831.10498122420 *T)
    sumL1  +=       0.246 * cos(1.25186233620  + 134.06407874580 *T)
    sumL1  +=       0.298 * cos(5.75927878031  + 902.70518605380 *T)
    sumL1  +=       0.240 * cos(2.84888261768  + 44.09217833040 *T)
    sumL1  +=       0.261 * cos(2.20643594285  + 73.08467753790 *T)
    sumL1  +=       0.321 * cos(3.46864827820  + 372.42352012820 *T)
    sumL1  +=       0.234 * cos(6.06783988023  + 147.11551657980 *T)
    sumL1  +=       0.262 * cos(2.69623862046  + 167.72244439930 *T)
    sumL1  +=       0.267 * cos(4.05985113852  + 75.30286342910 *T)
    sumL1  +=       0.240 * cos(0.48471871511  + 172.19711438360 *T)
    sumL1  +=       0.261 * cos(4.64354183979  + 535.32003938710 *T)
    sumL1  +=       0.244 * cos(5.85987959874  + 507.59956456680 *T)
    sumL1  +=       0.237 * cos(4.79666486485  + 377.15882254340 *T)
    sumL1  +=       0.224 * cos(1.94589447357  + 593.42686339800 *T)
    sumL1  +=       0.226 * cos(3.71637531808  + 449.28029223480 *T)
    sumL1  +=       0.233 * cos(5.98739382153  + 219.89137757700 *T)
    sumL1  +=       0.240 * cos(2.71609791875  + 227.52618943960 *T)
    sumL1  +=       0.221 * cos(2.23218400256  + 460.53844081980 *T)
    sumL1  +=       0.226 * cos(2.74516124394  + 446.31134681820 *T)
    sumL1  +=       0.221 * cos(3.20339807670  + 463.50738623640 *T)
    sumL1  +=       0.203 * cos(5.04975483055  + 457.87831194390 *T)
    sumL1  +=       0.191 * cos(4.24841510229  + 4.66586644600 *T)
    sumL1  +=       0.169 * cos(0.59358171769  + 983.11585891360 *T)
    sumL1  +=       0.179 * cos(4.12060524413  + 310.17209453310 *T)
    sumL1  +=       0.187 * cos(6.22165475247  + 294.67297614430 *T)
    sumL1  +=       0.162 * cos(1.30776665222  + 248.72381809010 *T)
    sumL1  +=       0.219 * cos(4.17413407057  + 303.86169668440 *T)
    sumL1  +=       0.193 * cos(1.64715944768  + 91.45637312370 *T)
    sumL1  +=       0.170 * cos(2.18067759964  + 66.18357885820 *T)
    sumL1  +=       0.156 * cos(4.92094728667  + 68.18931642830 *T)
    sumL1  +=       0.145 * cos(5.51404722738  + 280.96714700450 *T)
    sumL1  +=       0.144 * cos(5.81835834612  + 75.53235809270 *T)
    sumL1  +=       0.153 * cos(0.48549989656  + 144.89733068860 *T)
    sumL1  +=       0.168 * cos(5.81402201452  + 149.45132255000 *T)
    sumL1  +=       0.150 * cos(4.66632209585  + 306.83064210100 *T)
    sumL1  +=       0.131 * cos(1.01359934164  + 175.16605980020 *T)
    sumL1  +=       0.174 * cos(3.03279013213  + 298.23262239190 *T)
    sumL1  +=       0.163 * cos(1.97665571311  + 221.16340196420 *T)
    sumL1  +=       0.144 * cos(2.59058085010  + 217.23124870110 *T)
    sumL1  +=       0.167 * cos(2.74604167580  + 69.15252427480 *T)
    sumL1  +=       0.129 * cos(2.87574897902  + 156.15547927360 *T)

    # URANUS Longitude
    sumL2  = 0
    sumL2  +=   53033.277 * cos(0.00000000000  + 0.00000000000 *T)
    sumL2  +=    2357.636 * cos(2.26014661705  + 74.78159856730 *T)
    sumL2  +=     769.129 * cos(4.52561041823  + 11.04570026390 *T)
    sumL2  +=     551.533 * cos(3.25814281023  + 63.73589830340 *T)
    sumL2  +=     541.532 * cos(2.27573907424  + 3.93215326310 *T)
    sumL2  +=     529.473 * cos(4.92348433826  + 1.48447270830 *T)
    sumL2  +=     257.521 * cos(3.69059216858  + 3.18139373770 *T)
    sumL2  +=     238.835 * cos(5.85806638405  + 149.56319713460 *T)
    sumL2  +=     181.904 * cos(6.21763603405  + 70.84944530420 *T)
    sumL2  +=      49.401 * cos(6.03101301723  + 56.62235130260 *T)
    sumL2  +=      53.504 * cos(1.44225240953  + 76.26607127560 *T)
    sumL2  +=      38.222 * cos(1.78467827781  + 52.69019803950 *T)
    sumL2  +=      44.753 * cos(3.90904910523  + 2.44768055480 *T)
    sumL2  +=      44.530 * cos(0.81152639478  + 85.82729883120 *T)
    sumL2  +=      37.403 * cos(4.46228598032  + 2.96894541660 *T)
    sumL2  +=      33.029 * cos(0.86388149962  + 9.56122755560 *T)
    sumL2  +=      24.292 * cos(2.10702559049  + 18.15924726470 *T)
    sumL2  +=      29.423 * cos(5.09818697708  + 73.29712585900 *T)
    sumL2  +=      22.135 * cos(4.81730808582  + 78.71375183040 *T)
    sumL2  +=      22.491 * cos(5.99320728691  + 138.51749687070 *T)
    sumL2  +=      17.226 * cos(2.53537183199  + 145.63104387150 *T)
    sumL2  +=      21.392 * cos(2.39880709309  + 77.96299230500 *T)
    sumL2  +=      20.578 * cos(2.16918786539  + 224.34479570190 *T)
    sumL2  +=      16.777 * cos(3.46631344086  + 12.53017297220 *T)
    sumL2  +=      12.012 * cos(0.01941361902  + 22.09140052780 *T)
    sumL2  +=      10.466 * cos(4.45556032593  + 62.25142559510 *T)
    sumL2  +=      11.010 * cos(0.08496274370  + 127.47179660680 *T)
    sumL2  +=       8.668 * cos(4.25550086984  + 7.11354700080 *T)
    sumL2  +=      10.476 * cos(5.16453084068  + 71.60020482960 *T)
    sumL2  +=       7.160 * cos(1.24903906391  + 5.41662597140 *T)
    sumL2  +=       8.387 * cos(5.50115930045  + 67.66805156650 *T)
    sumL2  +=       6.087 * cos(5.44611674384  + 65.22037101170 *T)
    sumL2  +=       6.013 * cos(4.51836836347  + 151.04766984290 *T)
    sumL2  +=       5.718 * cos(1.82933915340  + 202.25339517410 *T)
    sumL2  +=       6.109 * cos(3.36320161279  + 447.79581952650 *T)
    sumL2  +=       6.003 * cos(5.72500086735  + 462.02291352810 *T)
    sumL2  +=       5.111 * cos(3.52374555791  + 59.80374504030 *T)
    sumL2  +=       5.155 * cos(1.05810305746  + 131.40394986990 *T)
    sumL2  +=       5.969 * cos(5.61147374852  + 148.07872442630 *T)
    sumL2  +=       5.065 * cos(3.36477113418  + 4.45341812490 *T)
    sumL2  +=       4.845 * cos(1.20298837109  + 71.81265315070 *T)
    sumL2  +=       3.979 * cos(0.67629577193  + 77.75054398390 *T)
    sumL2  +=       3.673 * cos(1.76315074166  + 351.81659230870 *T)
    sumL2  +=       3.149 * cos(3.83590892865  + 45.57665103870 *T)
    sumL2  +=       3.036 * cos(3.32062892682  + 160.60889739850 *T)
    sumL2  +=       3.033 * cos(6.14532331482  + 77.22927912210 *T)
    sumL2  +=       3.596 * cos(4.57256025582  + 454.90936652730 *T)
    sumL2  +=       2.664 * cos(5.36121614612  + 269.92144674060 *T)
    sumL2  +=       2.498 * cos(1.04819496324  + 69.36497259590 *T)
    sumL2  +=       2.307 * cos(2.69282373897  + 84.34282612290 *T)
    sumL2  +=       2.249 * cos(5.07693376112  + 14.97785352700 *T)
    sumL2  +=       2.228 * cos(1.38937510191  + 284.14854074220 *T)
    sumL2  +=       2.064 * cos(4.34647674542  + 984.60033162190 *T)
    sumL2  +=       2.105 * cos(2.32047802326  + 120.35824960600 *T)
    sumL2  +=       1.864 * cos(5.70354779393  + 54.17467074780 *T)
    sumL2  +=       2.005 * cos(3.87177765185  + 195.13984817330 *T)
    sumL2  +=       1.622 * cos(5.07964536529  + 209.36694217490 *T)
    sumL2  +=       1.597 * cos(0.48807990368  + 137.03302416240 *T)
    sumL2  +=       1.583 * cos(2.90536212187  + 51.20572533120 *T)
    sumL2  +=       1.725 * cos(6.25703202673  + 41.64449777560 *T)
    sumL2  +=       2.073 * cos(1.24032244487  + 35.16409022120 *T)
    sumL2  +=       1.543 * cos(2.15414338268  + 70.32818044240 *T)
    sumL2  +=       1.671 * cos(6.28283232471  + 277.03499374140 *T)
    sumL2  +=       1.494 * cos(6.04572758571  + 87.31177153950 *T)
    sumL2  +=       1.418 * cos(1.15843502159  + 213.29909543800 *T)
    sumL2  +=       1.239 * cos(4.63223076077  + 92.94084583200 *T)
    sumL2  +=       1.238 * cos(2.65969680342  + 134.58534360760 *T)
    sumL2  +=       1.273 * cos(5.87964059822  + 60.55450456570 *T)
    sumL2  +=       1.160 * cos(1.03320781667  + 153.49535039770 *T)
    sumL2  +=       1.430 * cos(4.68022239016  + 299.12639426920 *T)
    sumL2  +=       1.117 * cos(2.62506108047  + 72.33391801250 *T)
    sumL2  +=       1.142 * cos(4.64615099782  + 152.74459087230 *T)
    sumL2  +=       0.974 * cos(2.85233132493  + 222.86032299360 *T)
    sumL2  +=       1.046 * cos(4.81299135934  + 116.42609634290 *T)
    sumL2  +=       0.872 * cos(3.49659835508  + 340.77089204480 *T)
    sumL2  +=       0.952 * cos(2.10837480840  + 20.60692781950 *T)
    sumL2  +=       0.964 * cos(2.46471453524  + 380.12776796000 *T)
    sumL2  +=       0.843 * cos(6.12869288891  + 49.50880430180 *T)
    sumL2  +=       0.821 * cos(0.27134156683  + 191.20769491020 *T)
    sumL2  +=       0.813 * cos(4.08930465981  + 14.22709400160 *T)
    sumL2  +=       0.796 * cos(6.17066846300  + 344.70304530790 *T)
    sumL2  +=       0.924 * cos(2.11096444289  + 14.01464568050 *T)
    sumL2  +=       0.791 * cos(2.38927423348  + 58.10682401090 *T)
    sumL2  +=       0.781 * cos(0.74223115950  + 408.43894361130 *T)
    sumL2  +=       0.759 * cos(3.77564054479  + 80.19822453870 *T)
    sumL2  +=       0.884 * cos(1.99930014838  + 265.98929347750 *T)
    sumL2  +=       0.722 * cos(3.10001033669  + 422.66603761290 *T)
    sumL2  +=       0.750 * cos(2.33167721991  + 358.93013930950 *T)
    sumL2  +=       0.687 * cos(2.02866342040  + 33.67961751290 *T)
    sumL2  +=       0.603 * cos(1.10391172652  + 55.13787859430 *T)
    sumL2  +=       0.655 * cos(3.85415269764  + 16.67477455640 *T)
    sumL2  +=       0.606 * cos(0.15052747979  + 28.31117565130 *T)
    sumL2  +=       0.639 * cos(5.16714934188  + 23.57587323610 *T)
    sumL2  +=       0.658 * cos(0.75636229109  + 76.47851959670 *T)
    sumL2  +=       0.590 * cos(1.73778850095  + 8.07675484730 *T)
    sumL2  +=       0.565 * cos(4.92645232089  + 35.42472265210 *T)
    sumL2  +=       0.656 * cos(2.34273264083  + 38.13303563780 *T)
    sumL2  +=       0.542 * cos(5.97968975563  + 146.59425171800 *T)
    sumL2  +=       0.518 * cos(3.19086220901  + 152.53214255120 *T)
    sumL2  +=       0.536 * cos(4.52808465499  + 220.41264243880 *T)
    sumL2  +=       0.489 * cos(4.80633294199  + 159.12442469020 *T)
    sumL2  +=       0.491 * cos(0.85765309118  + 565.11568774670 *T)
    sumL2  +=       0.483 * cos(3.52583593251  + 144.14657116320 *T)
    sumL2  +=       0.521 * cos(5.21561656321  + 206.18554843720 *T)
    sumL2  +=       0.477 * cos(4.25420753202  + 365.90067395840 *T)
    sumL2  +=       0.466 * cos(5.13219663072  + 297.64192156090 *T)
    sumL2  +=       0.557 * cos(0.98387565952  + 225.82926841020 *T)
    sumL2  +=       0.531 * cos(4.22534657450  + 29.20494752860 *T)
    sumL2  +=       0.500 * cos(3.49663062387  + 128.95626931510 *T)
    sumL2  +=       0.445 * cos(2.60797570173  + 96.87299909510 *T)
    sumL2  +=       0.466 * cos(6.05585106742  + 70.11573212130 *T)
    sumL2  +=       0.425 * cos(1.04692398351  + 19.64371997300 *T)
    sumL2  +=       0.491 * cos(2.26123398680  + 152.01087768940 *T)
    sumL2  +=       0.455 * cos(5.45520675000  + 333.65734504400 *T)
    sumL2  +=       0.458 * cos(0.91654899383  + 373.01422095920 *T)
    sumL2  +=       0.520 * cos(5.72828536642  + 111.43016149680 *T)
    sumL2  +=       0.432 * cos(1.04604024916  + 125.98732389850 *T)
    sumL2  +=       0.387 * cos(2.82547341355  + 200.76892246580 *T)
    sumL2  +=       0.383 * cos(1.91679738697  + 5.62907429250 *T)
    sumL2  +=       0.504 * cos(1.95816731769  + 415.55249061210 *T)
    sumL2  +=       0.370 * cos(3.21958844151  + 387.24131496080 *T)
    sumL2  +=       0.379 * cos(2.75940848661  + 81.89514556810 *T)
    sumL2  +=       0.345 * cos(2.98021540638  + 429.77958461370 *T)
    sumL2  +=       0.368 * cos(6.20331898497  + 456.39383923560 *T)
    sumL2  +=       0.335 * cos(5.29062985955  + 13.33332212430 *T)
    sumL2  +=       0.320 * cos(0.74685222907  + 347.88443904560 *T)
    sumL2  +=       0.307 * cos(1.65925943351  + 99.16062095550 *T)
    sumL2  +=       0.284 * cos(2.09437476480  + 129.91947716160 *T)
    sumL2  +=       0.275 * cos(0.62680026669  + 31.49256938900 *T)
    sumL2  +=       0.339 * cos(1.65968150805  + 142.44965013380 *T)
    sumL2  +=       0.270 * cos(2.79378345550  + 977.48678462110 *T)
    sumL2  +=       0.284 * cos(2.42261530322  + 546.95644048200 *T)
    sumL2  +=       0.271 * cos(0.59225635449  + 1894.41906467650 *T)
    sumL2  +=       0.263 * cos(3.49309481771  + 440.68227252570 *T)
    sumL2  +=       0.295 * cos(0.43376026627  + 373.90799283650 *T)
    sumL2  +=       0.271 * cos(4.82853730065  + 561.18353448360 *T)
    sumL2  +=       0.248 * cos(5.72379940676  + 79.23501669220 *T)
    sumL2  +=       0.252 * cos(5.76728095309  + 235.39049596580 *T)
    sumL2  +=       0.261 * cos(0.09830739366  + 991.71387862270 *T)
    sumL2  +=       0.238 * cos(1.14634663242  + 288.08069400530 *T)
    sumL2  +=       0.220 * cos(5.12386263728  + 479.28838891550 *T)
    sumL2  +=       0.263 * cos(1.52366362053  + 146.38180339690 *T)
    sumL2  +=       0.230 * cos(0.35983101967  + 109.94568878850 *T)
    sumL2  +=       0.214 * cos(2.94388765580  + 184.72728735580 *T)
    sumL2  +=       0.198 * cos(5.72107161521  + 453.42489381900 *T)
    sumL2  +=       0.185 * cos(4.71600176780  + 108.46121608020 *T)
    sumL2  +=       0.139 * cos(2.99832472431  + 211.81462272970 *T)
    sumL2  +=       0.167 * cos(1.19643522280  + 39.61750834610 *T)
    sumL2  +=       0.180 * cos(0.80198096578  + 183.24281464750 *T)
    sumL2  +=       0.131 * cos(2.73236351123  + 522.57741809380 *T)
    sumL2  +=       0.142 * cos(5.03489222377  + 536.80451209540 *T)

    # URANUS Longitude
    sumL3  = 0
    sumL3  +=     120.936 * cos(0.02418789918  + 74.78159856730 *T)
    sumL3  +=      68.064 * cos(4.12084267733  + 3.93215326310 *T)
    sumL3  +=      52.828 * cos(2.38964061260  + 11.04570026390 *T)
    sumL3  +=      43.754 * cos(2.95965039734  + 1.48447270830 *T)
    sumL3  +=      45.300 * cos(2.04423798410  + 3.18139373770 *T)
    sumL3  +=      45.806 * cos(0.00000000000  + 0.00000000000 *T)
    sumL3  +=      24.969 * cos(4.88741307918  + 63.73589830340 *T)
    sumL3  +=      21.061 * cos(4.54511486862  + 70.84944530420 *T)
    sumL3  +=      19.897 * cos(2.31320314136  + 149.56319713460 *T)
    sumL3  +=       8.901 * cos(1.57548871761  + 56.62235130260 *T)
    sumL3  +=       4.271 * cos(0.22777319552  + 18.15924726470 *T)
    sumL3  +=       3.613 * cos(5.39244611308  + 76.26607127560 *T)
    sumL3  +=       3.488 * cos(4.97622811775  + 85.82729883120 *T)
    sumL3  +=       3.479 * cos(4.12969359977  + 52.69019803950 *T)
    sumL3  +=       3.572 * cos(0.95052448578  + 77.96299230500 *T)
    sumL3  +=       2.328 * cos(0.85770961794  + 145.63104387150 *T)
    sumL3  +=       2.696 * cos(0.37287796344  + 78.71375183040 *T)
    sumL3  +=       1.946 * cos(2.67997393431  + 7.11354700080 *T)
    sumL3  +=       2.156 * cos(5.65647821519  + 9.56122755560 *T)
    sumL3  +=       1.363 * cos(4.86983744746  + 224.34479570190 *T)
    sumL3  +=       1.333 * cos(1.25032115614  + 12.53017297220 *T)
    sumL3  +=       1.613 * cos(0.48764377311  + 71.60020482960 *T)
    sumL3  +=       1.475 * cos(5.19957293069  + 73.29712585900 *T)
    sumL3  +=       1.225 * cos(3.93406822032  + 22.09140052780 *T)
    sumL3  +=       0.911 * cos(2.18921999026  + 127.47179660680 *T)
    sumL3  +=       0.811 * cos(3.98323855938  + 462.02291352810 *T)
    sumL3  +=       0.808 * cos(5.06374463008  + 447.79581952650 *T)
    sumL3  +=       0.718 * cos(0.34600103024  + 5.62907429250 *T)
    sumL3  +=       0.722 * cos(1.05856935832  + 138.51749687070 *T)
    sumL3  +=       0.687 * cos(2.93752748595  + 131.40394986990 *T)
    sumL3  +=       0.463 * cos(1.58927254512  + 151.04766984290 *T)
    sumL3  +=       0.414 * cos(4.32904287620  + 120.35824960600 *T)
    sumL3  +=       0.372 * cos(0.73596518002  + 269.92144674060 *T)
    sumL3  +=       0.345 * cos(3.05968942771  + 561.18353448360 *T)
    sumL3  +=       0.338 * cos(5.94221536204  + 284.14854074220 *T)
    sumL3  +=       0.343 * cos(4.01891371998  + 546.95644048200 *T)
    sumL3  +=       0.382 * cos(5.93515231196  + 45.57665103870 *T)
    sumL3  +=       0.276 * cos(3.44212110991  + 202.25339517410 *T)
    sumL3  +=       0.300 * cos(1.13119175675  + 160.60889739850 *T)
    sumL3  +=       0.268 * cos(3.24615387968  + 536.80451209540 *T)
    sumL3  +=       0.358 * cos(1.10916640253  + 333.65734504400 *T)
    sumL3  +=       0.236 * cos(4.65292396535  + 387.24131496080 *T)
    sumL3  +=       0.204 * cos(5.81663798296  + 373.01422095920 *T)
    sumL3  +=       0.145 * cos(2.75632381347  + 92.94084583200 *T)
    sumL3  +=       0.149 * cos(0.13764106563  + 71.81265315070 *T)
    sumL3  +=       0.156 * cos(2.90936922804  + 153.49535039770 *T)

    # URANUS Longitude
    sumL4  = 0
    sumL4  +=     113.855 * cos(3.14159265359  + 0.00000000000 *T)
    sumL4  +=       5.599 * cos(4.57882424417  + 74.78159856730 *T)
    sumL4  +=       3.203 * cos(0.34623003207  + 11.04570026390 *T)
    sumL4  +=       1.217 * cos(3.42199121826  + 56.62235130260 *T)
    sumL4  +=       0.634 * cos(4.65759668097  + 18.15924726470 *T)
    sumL4  +=       0.171 * cos(3.80393539303  + 149.56319713460 *T)
    sumL4  +=       0.133 * cos(4.35519131657  + 63.73589830340 *T)

    # URANUS Longitude
    sumL5  = 0
    sumL5  +=       0.873 * cos(3.14159265359  + 0.00000000000 *T)

    # URANUS Latitude
    sumB0  = 0
    sumB0  +=  1346277.639 * cos(2.61877810545  + 74.78159856730 *T)
    sumB0  +=   62341.405 * cos(5.08111175856  + 149.56319713460 *T)
    sumB0  +=   61601.203 * cos(3.14159265359  + 0.00000000000 *T)
    sumB0  +=    9963.744 * cos(1.61603876357  + 76.26607127560 *T)
    sumB0  +=    9926.151 * cos(0.57630387917  + 73.29712585900 *T)
    sumB0  +=    3259.455 * cos(1.26119385960  + 224.34479570190 *T)
    sumB0  +=    2972.318 * cos(2.24367035538  + 1.48447270830 *T)
    sumB0  +=    2010.257 * cos(6.05550401088  + 148.07872442630 *T)
    sumB0  +=    1522.172 * cos(0.27960386377  + 63.73589830340 *T)
    sumB0  +=     924.055 * cos(4.03822927853  + 151.04766984290 *T)
    sumB0  +=     760.624 * cos(6.14000431923  + 71.81265315070 *T)
    sumB0  +=     420.265 * cos(5.21279984788  + 11.04570026390 *T)
    sumB0  +=     430.668 * cos(3.55445034854  + 213.29909543800 *T)
    sumB0  +=     436.843 * cos(3.38082524317  + 529.69096509460 *T)
    sumB0  +=     522.309 * cos(3.32085194770  + 138.51749687070 *T)
    sumB0  +=     434.625 * cos(0.34065281858  + 77.75054398390 *T)
    sumB0  +=     462.630 * cos(0.74256727574  + 85.82729883120 *T)
    sumB0  +=     232.649 * cos(2.25716421383  + 222.86032299360 *T)
    sumB0  +=     215.838 * cos(1.59121704940  + 38.13303563780 *T)
    sumB0  +=     244.698 * cos(0.78795150326  + 2.96894541660 *T)
    sumB0  +=     179.935 * cos(3.72487952673  + 299.12639426920 *T)
    sumB0  +=     174.895 * cos(1.23550262213  + 146.59425171800 *T)
    sumB0  +=     173.667 * cos(1.93654269131  + 380.12776796000 *T)
    sumB0  +=     160.368 * cos(5.33635436463  + 111.43016149680 *T)
    sumB0  +=     144.064 * cos(5.96239326415  + 35.16409022120 *T)
    sumB0  +=     102.049 * cos(2.61876256513  + 78.71375183040 *T)
    sumB0  +=     116.363 * cos(5.73877190007  + 70.84944530420 *T)
    sumB0  +=     106.441 * cos(0.94103112994  + 70.32818044240 *T)
    sumB0  +=      86.163 * cos(0.70262506622  + 39.61750834610 *T)
    sumB0  +=      72.617 * cos(0.20564696113  + 225.82926841020 *T)
    sumB0  +=      71.172 * cos(0.83343269975  + 109.94568878850 *T)
    sumB0  +=      57.502 * cos(2.67039425415  + 108.46121608020 *T)
    sumB0  +=      54.255 * cos(3.35166579613  + 184.72728735580 *T)
    sumB0  +=      44.470 * cos(2.74408231138  + 152.53214255120 *T)
    sumB0  +=      38.591 * cos(5.17394663303  + 202.25339517410 *T)
    sumB0  +=      39.157 * cos(2.17108251341  + 351.81659230870 *T)
    sumB0  +=      41.346 * cos(3.22134319551  + 160.60889739850 *T)
    sumB0  +=      35.140 * cos(4.00111634363  + 112.91463420510 *T)
    sumB0  +=      33.073 * cos(3.61378095742  + 221.37585028530 *T)
    sumB0  +=      31.315 * cos(2.71969470781  + 145.10977900970 *T)
    sumB0  +=      37.336 * cos(4.02053241202  + 52.69019803950 *T)
    sumB0  +=      32.028 * cos(1.29160071142  + 145.63104387150 *T)
    sumB0  +=      27.574 * cos(3.70064266960  + 36.64856292950 *T)
    sumB0  +=      24.277 * cos(2.84989187496  + 127.47179660680 *T)
    sumB0  +=      24.635 * cos(1.11645461259  + 3.93215326310 *T)
    sumB0  +=      24.315 * cos(5.48987913644  + 79.23501669220 *T)
    sumB0  +=      21.418 * cos(0.63722900407  + 277.03499374140 *T)
    sumB0  +=      19.826 * cos(2.59334182230  + 84.34282612290 *T)
    sumB0  +=      22.373 * cos(5.73687615457  + 4.45341812490 *T)
    sumB0  +=      19.137 * cos(1.30214105578  + 62.25142559510 *T)
    sumB0  +=      19.789 * cos(4.72260849557  + 297.64192156090 *T)
    sumB0  +=      20.299 * cos(1.06070151806  + 454.90936652730 *T)
    sumB0  +=      19.768 * cos(5.77906142568  + 305.34616939270 *T)
    sumB0  +=      21.348 * cos(1.01350946382  + 33.67961751290 *T)
    sumB0  +=      15.142 * cos(2.91786832554  + 426.59819087600 *T)
    sumB0  +=      16.000 * cos(1.95535748902  + 186.21176006410 *T)
    sumB0  +=      13.819 * cos(2.67163927171  + 74.66972398270 *T)
    sumB0  +=      11.463 * cos(5.73391138419  + 41.10198105440 *T)
    sumB0  +=      10.741 * cos(3.73401569675  + 1059.38193018920 *T)
    sumB0  +=      11.450 * cos(3.98177764866  + 106.97674337190 *T)
    sumB0  +=      10.360 * cos(4.75567608732  + 183.24281464750 *T)
    sumB0  +=      10.232 * cos(6.18772866993  + 373.90799283650 *T)
    sumB0  +=      13.803 * cos(5.70712120608  + 74.89347315190 *T)
    sumB0  +=      10.553 * cos(3.78602881738  + 490.33408917940 *T)
    sumB0  +=      11.838 * cos(5.96756415681  + 87.31177153950 *T)
    sumB0  +=      10.030 * cos(1.74828757238  + 22.09140052780 *T)
    sumB0  +=      10.107 * cos(0.92911975959  + 65.22037101170 *T)
    sumB0  +=       9.127 * cos(5.11093790809  + 153.49535039770 *T)
    sumB0  +=      12.093 * cos(2.53736362742  + 9.56122755560 *T)
    sumB0  +=       8.646 * cos(4.18351923569  + 12.53017297220 *T)
    sumB0  +=       9.978 * cos(5.83600622359  + 604.47256366190 *T)
    sumB0  +=      11.352 * cos(2.12645777694  + 68.84370773410 *T)
    sumB0  +=       8.472 * cos(3.36885457285  + 72.33391801250 *T)
    sumB0  +=       7.797 * cos(5.11771906359  + 77.96299230500 *T)
    sumB0  +=       8.302 * cos(5.19247905162  + 77.22927912210 *T)
    sumB0  +=       7.696 * cos(3.25189037096  + 71.60020482960 *T)
    sumB0  +=       7.513 * cos(2.98265970100  + 114.39910691340 *T)
    sumB0  +=       6.947 * cos(3.31871016057  + 56.62235130260 *T)
    sumB0  +=       6.490 * cos(0.88434578474  + 288.08069400530 *T)
    sumB0  +=       6.394 * cos(3.51142812432  + 220.41264243880 *T)
    sumB0  +=       6.211 * cos(0.58222518453  + 340.77089204480 *T)
    sumB0  +=       6.772 * cos(4.09374798222  + 137.03302416240 *T)
    sumB0  +=       5.595 * cos(5.68643434536  + 259.50888592310 *T)
    sumB0  +=       5.309 * cos(2.65421183211  + 300.61086697750 *T)
    sumB0  +=       4.950 * cos(4.99672086239  + 219.89137757700 *T)
    sumB0  +=       6.419 * cos(0.44895727879  + 140.00196957900 *T)
    sumB0  +=       4.975 * cos(4.06722486039  + 143.62530630140 *T)
    sumB0  +=       5.692 * cos(3.76563463180  + 67.66805156650 *T)
    sumB0  +=       4.853 * cos(2.03383592524  + 415.55249061210 *T)
    sumB0  +=       3.796 * cos(1.26231186682  + 75.30286342910 *T)
    sumB0  +=       3.807 * cos(2.24787582155  + 909.81873305460 *T)
    sumB0  +=       3.812 * cos(3.10475682509  + 181.75834193920 *T)
    sumB0  +=       3.764 * cos(5.20052090560  + 227.31374111850 *T)
    sumB0  +=       3.445 * cos(5.91769433069  + 296.15744885260 *T)
    sumB0  +=       3.517 * cos(5.44397685665  + 628.85158605010 *T)
    sumB0  +=       3.943 * cos(4.95136058926  + 265.98929347750 *T)
    sumB0  +=       3.472 * cos(0.34737998380  + 131.40394986990 *T)
    sumB0  +=       3.390 * cos(0.53497504164  + 206.18554843720 *T)
    sumB0  +=       3.038 * cos(4.68314286209  + 159.12442469020 *T)
    sumB0  +=       3.190 * cos(5.68929316349  + 235.39049596580 *T)
    sumB0  +=       3.303 * cos(2.09359507373  + 73.81839072080 *T)
    sumB0  +=       3.069 * cos(4.49065085092  + 66.70484372000 *T)
    sumB0  +=       3.285 * cos(0.00780313833  + 75.74480641380 *T)
    sumB0  +=       2.917 * cos(4.21615078632  + 258.02441321480 *T)
    sumB0  +=       3.747 * cos(0.82999983666  + 74.26033370550 *T)
    sumB0  +=       2.814 * cos(3.96708337625  + 82.85835341460 *T)
    sumB0  +=       2.474 * cos(3.81319259323  + 7.11354700080 *T)
    sumB0  +=       2.394 * cos(2.21483198491  + 54.17467074780 *T)
    sumB0  +=       2.555 * cos(2.97023907145  + 378.64329525170 *T)
    sumB0  +=       2.631 * cos(1.55153254691  + 154.01661525950 *T)
    sumB0  +=       2.633 * cos(2.28385552693  + 32.19514480460 *T)
    sumB0  +=       2.643 * cos(3.96832729680  + 381.61224066830 *T)
    sumB0  +=       2.206 * cos(3.06995275892  + 59.80374504030 *T)
    sumB0  +=       2.635 * cos(0.53987945692  + 211.81462272970 *T)
    sumB0  +=       2.071 * cos(1.97429082033  + 18.15924726470 *T)
    sumB0  +=       2.485 * cos(3.55433846990  + 96.87299909510 *T)
    sumB0  +=       2.061 * cos(4.50102695788  + 5.93789083320 *T)
    sumB0  +=       1.916 * cos(1.60538526374  + 80.19822453870 *T)
    sumB0  +=       2.480 * cos(0.63321072542  + 187.69623277240 *T)
    sumB0  +=       2.039 * cos(2.97351088965  + 191.20769491020 *T)
    sumB0  +=       1.833 * cos(1.95824865568  + 81.89514556810 *T)
    sumB0  +=       1.719 * cos(2.22526635038  + 479.28838891550 *T)
    sumB0  +=       1.745 * cos(3.22821992592  + 218.40690486870 *T)
    sumB0  +=       1.857 * cos(1.66304484985  + 984.60033162190 *T)
    sumB0  +=       1.766 * cos(5.24239122261  + 105.49227066360 *T)
    sumB0  +=       1.524 * cos(6.05374020264  + 99.16062095550 *T)
    sumB0  +=       1.519 * cos(0.94716867229  + 372.42352012820 *T)
    sumB0  +=       1.614 * cos(3.39986066169  + 230.56457082540 *T)
    sumB0  +=       1.711 * cos(3.44237080993  + 522.57741809380 *T)
    sumB0  +=       1.504 * cos(1.34653259405  + 74.52096613640 *T)
    sumB0  +=       1.577 * cos(4.38020936720  + 80.71948940050 *T)
    sumB0  +=       1.360 * cos(5.48691240270  + 74.82978267710 *T)
    sumB0  +=       1.364 * cos(4.56045715617  + 42.58645376270 *T)
    sumB0  +=       1.398 * cos(0.33827838973  + 142.44965013380 *T)
    sumB0  +=       1.709 * cos(3.63188407264  + 554.06998748280 *T)
    sumB0  +=       1.360 * cos(2.89305157919  + 74.73341445750 *T)
    sumB0  +=       1.260 * cos(5.50922979275  + 74.94165726170 *T)
    sumB0  +=       1.374 * cos(4.39897993200  + 260.99335863140 *T)
    sumB0  +=       1.366 * cos(2.15288773765  + 162.09337010680 *T)
    sumB0  +=       1.244 * cos(4.97789913094  + 149.45132255000 *T)
    sumB0  +=       1.269 * cos(0.84167691738  + 767.36908292080 *T)
    sumB0  +=       1.278 * cos(4.53585702916  + 294.67297614430 *T)
    sumB0  +=       1.342 * cos(5.11117141196  + 51.20572533120 *T)
    sumB0  +=       1.180 * cos(0.79882196802  + 116.42609634290 *T)
    sumB0  +=       1.495 * cos(0.74986873597  + 75.04223099820 *T)
    sumB0  +=       1.207 * cos(3.67288675913  + 20.60692781950 *T)
    sumB0  +=       1.181 * cos(4.39598416757  + 180.27386923090 *T)
    sumB0  +=       1.248 * cos(3.13312504066  + 67.35923502580 *T)
    sumB0  +=       1.263 * cos(2.87116663203  + 74.62153987290 *T)
    sumB0  +=       1.380 * cos(3.76141611602  + 92.94084583200 *T)
    sumB0  +=       1.113 * cos(3.87133607367  + 39.35687591520 *T)
    sumB0  +=       1.018 * cos(6.20393099094  + 835.03713448730 *T)
    sumB0  +=       0.962 * cos(3.30343472839  + 255.05546779820 *T)
    sumB0  +=       1.238 * cos(1.73023505315  + 149.67507171920 *T)
    sumB0  +=       0.970 * cos(1.71236273285  + 115.88357962170 *T)
    sumB0  +=       1.001 * cos(5.49914631698  + 256.53994050650 *T)
    sumB0  +=       0.921 * cos(3.07729879788  + 8.07675484730 *T)
    sumB0  +=       0.914 * cos(0.00764291274  + 536.80451209540 *T)
    sumB0  +=       0.911 * cos(6.23753038018  + 200.76892246580 *T)
    sumB0  +=       0.956 * cos(5.76811833839  + 128.95626931510 *T)
    sumB0  +=       0.999 * cos(0.33400530567  + 404.50679034820 *T)
    sumB0  +=       0.952 * cos(3.00456073496  + 14.97785352700 *T)
    sumB0  +=       0.765 * cos(3.45454533660  + 214.78356814630 *T)
    sumB0  +=       0.800 * cos(3.20912932090  + 28.31117565130 *T)
    sumB0  +=       0.799 * cos(4.11425365829  + 125.98732389850 *T)
    sumB0  +=       1.021 * cos(1.79905869707  + 3.18139373770 *T)
    sumB0  +=       0.706 * cos(5.80210566917  + 157.63995198190 *T)
    sumB0  +=       0.715 * cos(5.56313177065  + 146.38180339690 *T)
    sumB0  +=       0.689 * cos(1.84748347121  + 41.64449777560 *T)
    sumB0  +=       0.682 * cos(5.16479782395  + 74.03083904190 *T)
    sumB0  +=       0.673 * cos(2.65544175682  + 75.53235809270 *T)
    sumB0  +=       0.723 * cos(4.75905991606  + 331.32153907380 *T)
    sumB0  +=       0.730 * cos(2.25510749124  + 453.42489381900 *T)
    sumB0  +=       0.691 * cos(3.58561635364  + 362.86229257260 *T)
    sumB0  +=       0.641 * cos(5.77408891198  + 110.20632121940 *T)
    sumB0  +=       0.671 * cos(5.96862039131  + 135.54855145410 *T)
    sumB0  +=       0.631 * cos(1.97807297205  + 639.89728631400 *T)
    sumB0  +=       0.774 * cos(0.42035450706  + 565.11568774670 *T)
    sumB0  +=       0.705 * cos(2.94649553712  + 60.76695288680 *T)
    sumB0  +=       0.663 * cos(5.32574112049  + 142.14083359310 *T)
    sumB0  +=       0.612 * cos(2.01741578932  + 195.13984817330 *T)
    sumB0  +=       0.749 * cos(5.56828487823  + 2.44768055480 *T)
    sumB0  +=       0.798 * cos(0.90969731665  + 152.01087768940 *T)
    sumB0  +=       0.747 * cos(5.07639466593  + 89.75945209430 *T)
    sumB0  +=       0.650 * cos(4.56215718085  + 216.92243216040 *T)
    sumB0  +=       0.651 * cos(2.80626026285  + 50.40257617910 *T)
    sumB0  +=       0.593 * cos(1.89556258897  + 203.73786788240 *T)
    sumB0  +=       0.550 * cos(2.90625551534  + 68.18931642830 *T)
    sumB0  +=       0.548 * cos(3.75628845322  + 617.80588578620 *T)
    sumB0  +=       0.554 * cos(2.78135114877  + 14.01464568050 *T)
    sumB0  +=       0.530 * cos(3.51385025328  + 291.70403072770 *T)
    sumB0  +=       0.506 * cos(4.94619342366  + 81.37388070630 *T)
    sumB0  +=       0.649 * cos(5.74895589744  + 141.48644228730 *T)
    sumB0  +=       0.593 * cos(5.40734033998  + 692.58748435350 *T)
    sumB0  +=       0.544 * cos(2.99910512780  + 152.74459087230 *T)
    sumB0  +=       0.485 * cos(2.36317665443  + 448.68959140380 *T)
    sumB0  +=       0.481 * cos(5.81647231299  + 134.58534360760 *T)
    sumB0  +=       0.517 * cos(4.97759795528  + 387.24131496080 *T)
    sumB0  +=       0.573 * cos(2.46311368783  + 81.00137369080 *T)
    sumB0  +=       0.470 * cos(1.30184316812  + 228.27694896500 *T)
    sumB0  +=       0.475 * cos(0.53480492526  + 303.86169668440 *T)
    sumB0  +=       0.485 * cos(6.21247575899  + 5.41662597140 *T)
    sumB0  +=       0.468 * cos(0.55881267334  + 23.57587323610 *T)
    sumB0  +=       0.585 * cos(4.67924542643  + 88.79624424780 *T)
    sumB0  +=       0.512 * cos(5.86200059955  + 293.18850343600 *T)
    sumB0  +=       0.445 * cos(2.50076311432  + 905.88657979150 *T)
    sumB0  +=       0.501 * cos(4.79997295899  + 306.83064210100 *T)
    sumB0  +=       0.418 * cos(5.21379084769  + 35.42472265210 *T)
    sumB0  +=       0.408 * cos(5.69107313998  + 284.14854074220 *T)
    sumB0  +=       0.474 * cos(3.03149617428  + 286.59622129700 *T)
    sumB0  +=       0.432 * cos(4.20907682097  + 278.51946644970 *T)
    sumB0  +=       0.506 * cos(2.05348204197  + 373.01422095920 *T)
    sumB0  +=       0.410 * cos(5.30637634877  + 95.38852638680 *T)
    sumB0  +=       0.478 * cos(2.41106642594  + 358.93013930950 *T)
    sumB0  +=       0.536 * cos(1.82614772260  + 114.13847448250 *T)
    sumB0  +=       0.383 * cos(5.54541241459  + 419.48464387520 *T)
    sumB0  +=       0.413 * cos(0.03813081773  + 103.09277421860 *T)
    sumB0  +=       0.368 * cos(4.08526832792  + 1589.07289528380 *T)
    sumB0  +=       0.369 * cos(1.82533858431  + 334.29048449040 *T)
    sumB0  +=       0.491 * cos(5.58913582973  + 68.56182344380 *T)
    sumB0  +=       0.387 * cos(0.56619310480  + 602.98809095360 *T)
    sumB0  +=       0.427 * cos(5.08349119654  + 367.97010200330 *T)
    sumB0  +=       0.475 * cos(0.17507881032  + 120.35824960600 *T)
    sumB0  +=       0.342 * cos(5.27187859255  + 28.57180808220 *T)
    sumB0  +=       0.407 * cos(2.00853504718  + 679.25416222920 *T)
    sumB0  +=       0.381 * cos(4.61366060949  + 329.72519178090 *T)
    sumB0  +=       0.388 * cos(0.88856038803  + 483.22054217860 *T)
    sumB0  +=       0.376 * cos(1.28856348513  + 155.78297225810 *T)
    sumB0  +=       0.326 * cos(6.09140263554  + 456.39383923560 *T)
    sumB0  +=       0.318 * cos(0.09996195379  + 69.36497259590 *T)
    sumB0  +=       0.292 * cos(5.11578046796  + 375.39246554480 *T)
    sumB0  +=       0.299 * cos(6.04121646505  + 332.80601178210 *T)
    sumB0  +=       0.328 * cos(3.47252263966  + 73.40900044360 *T)
    sumB0  +=       0.283 * cos(1.81773059325  + 647.01083331480 *T)
    sumB0  +=       0.308 * cos(3.50154864071  + 30.71067209630 *T)
    sumB0  +=       0.283 * cos(1.88910019831  + 24.37902238820 *T)
    sumB0  +=       0.278 * cos(3.85507901929  + 760.25553592000 *T)
    sumB0  +=       0.273 * cos(4.22941219477  + 391.17346822390 *T)
    sumB0  +=       0.293 * cos(5.44464406156  + 477.91579079180 *T)
    sumB0  +=       0.289 * cos(3.85492516765  + 209.36694217490 *T)
    sumB0  +=       0.309 * cos(1.97045147502  + 543.02428721890 *T)
    sumB0  +=       0.324 * cos(5.57191515554  + 501.37978944330 *T)
    sumB0  +=       0.248 * cos(2.17741598681  + 611.58611066270 *T)
    sumB0  +=       0.245 * cos(1.04008534095  + 1332.48477066750 *T)
    sumB0  +=       0.248 * cos(6.19516038159  + 1134.16352875650 *T)
    sumB0  +=       0.242 * cos(1.37720813333  + 121.25202148330 *T)
    sumB0  +=       0.233 * cos(0.50933224249  + 462.02291352810 *T)
    sumB0  +=       0.319 * cos(4.24219881040  + 328.35259365720 *T)
    sumB0  +=       0.244 * cos(6.00644853640  + 295.19424100610 *T)
    sumB0  +=       0.228 * cos(0.72608678727  + 233.90602325750 *T)
    sumB0  +=       0.223 * cos(5.35785607671  + 983.11585891360 *T)
    sumB0  +=       0.304 * cos(5.68101077712  + 189.18070548070 *T)
    sumB0  +=       0.223 * cos(2.04168197187  + 370.93904741990 *T)
    sumB0  +=       0.245 * cos(4.69742022955  + 316.39186965660 *T)
    sumB0  +=       0.217 * cos(4.33845164683  + 269.92144674060 *T)
    sumB0  +=       0.267 * cos(0.15744446718  + 10213.28554621100 *T)
    sumB0  +=       0.210 * cos(5.84975171904  + 147.11551657980 *T)
    sumB0  +=       0.224 * cos(3.04829549918  + 1439.50969814920 *T)
    sumB0  +=       0.226 * cos(0.72248476212  + 45.57665103870 *T)
    sumB0  +=       0.202 * cos(1.37341689387  + 302.09533968580 *T)
    sumB0  +=       0.207 * cos(6.13936312021  + 344.70304530790 *T)
    sumB0  +=       0.228 * cos(2.33602531859  + 150.52640498110 *T)
    sumB0  +=       0.203 * cos(2.38070591912  + 275.55052103310 *T)
    sumB0  +=       0.194 * cos(5.11344829813  + 1215.16490244730 *T)
    sumB0  +=       0.259 * cos(2.78974616768  + 144.14657116320 *T)
    sumB0  +=       0.199 * cos(3.21010368905  + 7.42236354150 *T)
    sumB0  +=       0.246 * cos(6.15106498377  + 6.21977512350 *T)
    sumB0  +=       0.180 * cos(4.70377036870  + 518.64526483070 *T)
    sumB0  +=       0.186 * cos(3.85070006482  + 46.20979048510 *T)
    sumB0  +=       0.175 * cos(3.72163733058  + 150.08446199640 *T)
    sumB0  +=       0.165 * cos(1.87245210311  + 310.17209453310 *T)
    sumB0  +=       0.166 * cos(3.24028819042  + 173.94221952280 *T)
    sumB0  +=       0.181 * cos(0.39521867351  + 0.96320784650 *T)
    sumB0  +=       0.144 * cos(1.78180338482  + 531.17543780290 *T)
    sumB0  +=       0.137 * cos(6.20635855175  + 369.45457471160 *T)
    sumB0  +=       0.136 * cos(4.00164375048  + 526.72201967800 *T)
    sumB0  +=       0.141 * cos(4.02238453909  + 350.33211960040 *T)
    sumB0  +=       0.125 * cos(5.28865037145  + 329.83706636550 *T)
    sumB0  +=       0.134 * cos(3.97421067761  + 228.79821382680 *T)
    sumB0  +=       0.132 * cos(4.52023845365  + 148.59998928810 *T)
    sumB0  +=       0.125 * cos(4.24724390191  + 248.72381809010 *T)
    sumB0  +=       0.120 * cos(4.08565783859  + 154.97982310600 *T)
    sumB0  +=       0.133 * cos(3.16576403244  + 262.47783133970 *T)

    # URANUS Latitude
    sumB1  = 0
    sumB1  +=  206366.162 * cos(4.12394311407  + 74.78159856730 *T)
    sumB1  +=    8563.230 * cos(0.33819986165  + 149.56319713460 *T)
    sumB1  +=    1725.703 * cos(2.12193159895  + 73.29712585900 *T)
    sumB1  +=    1368.860 * cos(3.06861722047  + 76.26607127560 *T)
    sumB1  +=    1374.449 * cos(0.00000000000  + 0.00000000000 *T)
    sumB1  +=     399.847 * cos(2.84767037795  + 224.34479570190 *T)
    sumB1  +=     450.639 * cos(3.77656180977  + 1.48447270830 *T)
    sumB1  +=     307.214 * cos(1.25456766737  + 148.07872442630 *T)
    sumB1  +=     154.336 * cos(3.78575467747  + 63.73589830340 *T)
    sumB1  +=     110.888 * cos(5.32888676461  + 138.51749687070 *T)
    sumB1  +=     112.432 * cos(5.57299891505  + 151.04766984290 *T)
    sumB1  +=      83.493 * cos(3.59152795558  + 71.81265315070 *T)
    sumB1  +=      55.573 * cos(3.40135416354  + 85.82729883120 *T)
    sumB1  +=      41.377 * cos(4.45476669141  + 78.71375183040 *T)
    sumB1  +=      53.690 * cos(1.70455769943  + 77.75054398390 *T)
    sumB1  +=      41.912 * cos(1.21476607434  + 11.04570026390 *T)
    sumB1  +=      31.959 * cos(3.77446207748  + 222.86032299360 *T)
    sumB1  +=      30.297 * cos(2.56371683644  + 2.96894541660 *T)
    sumB1  +=      26.977 * cos(5.33695500294  + 213.29909543800 *T)
    sumB1  +=      26.222 * cos(0.41620628369  + 380.12776796000 *T)
    sumB1  +=      20.094 * cos(5.93085633510  + 529.69096509460 *T)
    sumB1  +=      22.992 * cos(2.48887389394  + 146.59425171800 *T)
    sumB1  +=      19.590 * cos(5.37213500014  + 299.12639426920 *T)
    sumB1  +=      20.408 * cos(3.70179681605  + 70.84944530420 *T)
    sumB1  +=      19.102 * cos(1.09213276596  + 111.43016149680 *T)
    sumB1  +=      19.411 * cos(3.83015453768  + 38.13303563780 *T)
    sumB1  +=      10.847 * cos(2.66326308043  + 3.93215326310 *T)
    sumB1  +=      10.249 * cos(2.31278807720  + 109.94568878850 *T)
    sumB1  +=       9.405 * cos(2.76950513184  + 39.61750834610 *T)
    sumB1  +=       7.660 * cos(1.81108462850  + 225.82926841020 *T)
    sumB1  +=       8.082 * cos(4.69064168719  + 184.72728735580 *T)
    sumB1  +=       6.584 * cos(5.50417589189  + 35.16409022120 *T)
    sumB1  +=       7.410 * cos(1.17879753422  + 65.22037101170 *T)
    sumB1  +=       6.451 * cos(4.98294064391  + 71.60020482960 *T)
    sumB1  +=       6.089 * cos(1.31830108565  + 52.69019803950 *T)
    sumB1  +=       4.768 * cos(5.90574941745  + 145.63104387150 *T)
    sumB1  +=       4.840 * cos(4.86390682412  + 221.37585028530 *T)
    sumB1  +=       4.192 * cos(3.29643787103  + 77.96299230500 *T)
    sumB1  +=       4.711 * cos(4.24289069791  + 152.53214255120 *T)
    sumB1  +=       4.894 * cos(6.01164167429  + 160.60889739850 *T)
    sumB1  +=       3.738 * cos(4.75287390209  + 70.32818044240 *T)
    sumB1  +=       3.481 * cos(0.64108927026  + 153.49535039770 *T)
    sumB1  +=       3.758 * cos(3.94715595219  + 351.81659230870 *T)
    sumB1  +=       3.114 * cos(0.10537144899  + 112.91463420510 *T)
    sumB1  +=       2.788 * cos(4.24118032837  + 74.66972398270 *T)
    sumB1  +=       2.505 * cos(0.04576283378  + 297.64192156090 *T)
    sumB1  +=       2.563 * cos(4.15665405963  + 305.34616939270 *T)
    sumB1  +=       2.544 * cos(5.25903565788  + 56.62235130260 *T)
    sumB1  +=       2.247 * cos(4.24726481845  + 36.64856292950 *T)
    sumB1  +=       2.541 * cos(0.40106060407  + 77.22927912210 *T)
    sumB1  +=       2.212 * cos(2.88960413468  + 277.03499374140 *T)
    sumB1  +=       2.299 * cos(3.57748029365  + 186.21176006410 *T)
    sumB1  +=       2.661 * cos(0.53230319176  + 79.23501669220 *T)
    sumB1  +=       2.157 * cos(2.10150995852  + 127.47179660680 *T)
    sumB1  +=       2.265 * cos(1.41055702214  + 4.45341812490 *T)
    sumB1  +=       2.103 * cos(4.27438518414  + 22.09140052780 *T)
    sumB1  +=       1.861 * cos(3.75619999278  + 145.10977900970 *T)
    sumB1  +=       1.759 * cos(2.10240976488  + 131.40394986990 *T)
    sumB1  +=       1.661 * cos(4.84483054269  + 62.25142559510 *T)
    sumB1  +=       1.496 * cos(1.72084298116  + 220.41264243880 *T)
    sumB1  +=       1.659 * cos(5.86539712478  + 454.90936652730 *T)
    sumB1  +=       1.428 * cos(0.31508367934  + 137.03302416240 *T)
    sumB1  +=       1.522 * cos(1.00801468633  + 75.74480641380 *T)
    sumB1  +=       1.459 * cos(6.17427145114  + 426.59819087600 *T)
    sumB1  +=       1.463 * cos(5.14953143442  + 84.34282612290 *T)
    sumB1  +=       1.453 * cos(2.22988903923  + 206.18554843720 *T)
    sumB1  +=       1.358 * cos(5.85111427068  + 183.24281464750 *T)
    sumB1  +=       1.405 * cos(2.43582184515  + 87.31177153950 *T)
    sumB1  +=       1.495 * cos(5.55621838458  + 67.66805156650 *T)
    sumB1  +=       1.317 * cos(1.91178535183  + 140.00196957900 *T)
    sumB1  +=       1.068 * cos(1.51430678116  + 373.90799283650 *T)
    sumB1  +=       1.439 * cos(0.99170994448  + 74.89347315190 *T)
    sumB1  +=       1.065 * cos(4.15616015505  + 288.08069400530 *T)
    sumB1  +=       1.096 * cos(1.63909426062  + 41.10198105440 *T)
    sumB1  +=       1.189 * cos(0.90595784409  + 33.67961751290 *T)
    sumB1  +=       0.961 * cos(5.48175535705  + 490.33408917940 *T)
    sumB1  +=       0.851 * cos(0.95029849401  + 909.81873305460 *T)
    sumB1  +=       0.820 * cos(0.78610123063  + 259.50888592310 *T)
    sumB1  +=       0.881 * cos(4.31294603221  + 9.56122755560 *T)
    sumB1  +=       0.708 * cos(0.00007309836  + 81.89514556810 *T)
    sumB1  +=       0.709 * cos(3.18853632737  + 80.19822453870 *T)
    sumB1  +=       0.786 * cos(5.18884635415  + 114.39910691340 *T)
    sumB1  +=       0.822 * cos(0.01949759759  + 18.15924726470 *T)
    sumB1  +=       0.656 * cos(6.16899483115  + 96.87299909510 *T)
    sumB1  +=       0.879 * cos(1.82006610038  + 73.81839072080 *T)
    sumB1  +=       0.872 * cos(0.30134022304  + 12.53017297220 *T)
    sumB1  +=       0.860 * cos(0.21225398802  + 3.18139373770 *T)
    sumB1  +=       0.637 * cos(2.64378420008  + 75.30286342910 *T)
    sumB1  +=       0.727 * cos(0.02846968582  + 66.70484372000 *T)
    sumB1  +=       0.600 * cos(4.42462853209  + 415.55249061210 *T)
    sumB1  +=       0.590 * cos(4.17885957237  + 300.61086697750 *T)
    sumB1  +=       0.610 * cos(6.07202921132  + 219.89137757700 *T)
    sumB1  +=       0.611 * cos(0.97629869063  + 296.15744885260 *T)
    sumB1  +=       0.635 * cos(2.21125075603  + 74.26033370550 *T)
    sumB1  +=       0.529 * cos(2.35940463062  + 7.11354700080 *T)
    sumB1  +=       0.622 * cos(2.21801944850  + 211.81462272970 *T)
    sumB1  +=       0.519 * cos(2.01872911223  + 142.44965013380 *T)
    sumB1  +=       0.489 * cos(1.83419944488  + 92.94084583200 *T)
    sumB1  +=       0.445 * cos(2.51784247184  + 604.47256366190 *T)
    sumB1  +=       0.413 * cos(5.36482818305  + 82.85835341460 *T)
    sumB1  +=       0.445 * cos(0.33164113115  + 227.31374111850 *T)
    sumB1  +=       0.456 * cos(2.44190834824  + 381.61224066830 *T)
    sumB1  +=       0.378 * cos(1.91873737843  + 202.25339517410 *T)
    sumB1  +=       0.509 * cos(5.83556856314  + 191.20769491020 *T)
    sumB1  +=       0.455 * cos(5.15414537021  + 522.57741809380 *T)
    sumB1  +=       0.419 * cos(2.80644155875  + 72.33391801250 *T)
    sumB1  +=       0.333 * cos(0.32014837950  + 2.44768055480 *T)
    sumB1  +=       0.360 * cos(1.47248643716  + 378.64329525170 *T)
    sumB1  +=       0.306 * cos(0.15517399606  + 159.12442469020 *T)
    sumB1  +=       0.301 * cos(4.46417652272  + 536.80451209540 *T)
    sumB1  +=       0.353 * cos(0.48749845867  + 128.95626931510 *T)
    sumB1  +=       0.351 * cos(6.24769322491  + 5.93789083320 *T)
    sumB1  +=       0.298 * cos(1.71815652029  + 235.39049596580 *T)
    sumB1  +=       0.315 * cos(2.44922921309  + 187.69623277240 *T)
    sumB1  +=       0.318 * cos(0.70176359510  + 181.75834193920 *T)
    sumB1  +=       0.314 * cos(4.68400251693  + 14.97785352700 *T)
    sumB1  +=       0.282 * cos(3.70093718573  + 108.46121608020 *T)
    sumB1  +=       0.272 * cos(3.91340553608  + 617.80588578620 *T)
    sumB1  +=       0.273 * cos(3.29483889428  + 387.24131496080 *T)
    sumB1  +=       0.323 * cos(4.90410549341  + 258.02441321480 *T)
    sumB1  +=       0.288 * cos(4.42249612833  + 195.13984817330 *T)
    sumB1  +=       0.250 * cos(1.23231297183  + 703.63318461740 *T)
    sumB1  +=       0.338 * cos(2.84645768890  + 154.01661525950 *T)
    sumB1  +=       0.297 * cos(1.16538119842  + 146.38180339690 *T)
    sumB1  +=       0.248 * cos(4.90614051989  + 41.64449777560 *T)
    sumB1  +=       0.275 * cos(5.35665949805  + 80.71948940050 *T)
    sumB1  +=       0.257 * cos(1.82441994046  + 230.56457082540 *T)
    sumB1  +=       0.234 * cos(0.27679874465  + 33.13710079170 *T)
    sumB1  +=       0.280 * cos(1.73679618032  + 265.98929347750 *T)
    sumB1  +=       0.229 * cos(0.49529839431  + 74.82978267710 *T)
    sumB1  +=       0.229 * cos(4.18462288684  + 74.73341445750 *T)
    sumB1  +=       0.253 * cos(2.63817804331  + 74.52096613640 *T)
    sumB1  +=       0.252 * cos(2.04143912495  + 75.04223099820 *T)
    sumB1  +=       0.213 * cos(4.16218259902  + 74.62153987290 *T)
    sumB1  +=       0.212 * cos(0.51761494342  + 74.94165726170 *T)
    sumB1  +=       0.201 * cos(4.54140547837  + 20.60692781950 *T)
    sumB1  +=       0.194 * cos(4.17282454759  + 116.42609634290 *T)
    sumB1  +=       0.213 * cos(2.29528235429  + 32.19514480460 *T)
    sumB1  +=       0.174 * cos(3.15418942153  + 228.27694896500 *T)
    sumB1  +=       0.194 * cos(0.06960211137  + 42.58645376270 *T)
    sumB1  +=       0.173 * cos(5.59700344643  + 68.84370773410 *T)
    sumB1  +=       0.159 * cos(4.91721631097  + 143.62530630140 *T)
    sumB1  +=       0.150 * cos(0.92771324396  + 404.50679034820 *T)
    sumB1  +=       0.136 * cos(2.51083022906  + 372.42352012820 *T)
    sumB1  +=       0.134 * cos(3.22507836958  + 479.28838891550 *T)
    sumB1  +=       0.124 * cos(0.42063711585  + 149.45132255000 *T)
    sumB1  +=       0.122 * cos(1.20639876458  + 5.41662597140 *T)
    sumB1  +=       0.119 * cos(0.45375065997  + 74.03083904190 *T)
    sumB1  +=       0.146 * cos(5.08207330360  + 294.67297614430 *T)
    sumB1  +=       0.118 * cos(4.22640788890  + 75.53235809270 *T)
    sumB1  +=       0.118 * cos(1.50613822829  + 344.70304530790 *T)
    sumB1  +=       0.121 * cos(2.13544759505  + 209.36694217490 *T)
    sumB1  +=       0.121 * cos(6.08239076370  + 260.99335863140 *T)
    sumB1  +=       0.154 * cos(4.18369977366  + 39.35687591520 *T)

    # URANUS Latitude
    sumB2  = 0
    sumB2  +=    9211.656 * cos(5.80044305785  + 74.78159856730 *T)
    sumB2  +=     556.926 * cos(0.00000000000  + 0.00000000000 *T)
    sumB2  +=     286.265 * cos(2.17729776353  + 149.56319713460 *T)
    sumB2  +=      94.969 * cos(3.84237569809  + 73.29712585900 *T)
    sumB2  +=      45.419 * cos(4.87822046064  + 76.26607127560 *T)
    sumB2  +=      20.107 * cos(5.46264485369  + 1.48447270830 *T)
    sumB2  +=      14.793 * cos(0.87983715652  + 138.51749687070 *T)
    sumB2  +=      13.963 * cos(5.07234043994  + 63.73589830340 *T)
    sumB2  +=      14.261 * cos(2.84517742687  + 148.07872442630 *T)
    sumB2  +=      10.122 * cos(5.00290894862  + 224.34479570190 *T)
    sumB2  +=       8.299 * cos(6.26655615197  + 78.71375183040 *T)
    sumB2  +=       4.729 * cos(5.16274174929  + 71.81265315070 *T)
    sumB2  +=       3.816 * cos(6.28224514574  + 85.82729883120 *T)
    sumB2  +=       3.488 * cos(3.53472172445  + 11.04570026390 *T)
    sumB2  +=       2.555 * cos(1.44444215715  + 151.04766984290 *T)
    sumB2  +=       2.353 * cos(4.23069776466  + 3.93215326310 *T)
    sumB2  +=       2.585 * cos(0.41383633246  + 71.60020482960 *T)
    sumB2  +=       1.394 * cos(4.13126838571  + 146.59425171800 *T)
    sumB2  +=       1.183 * cos(3.68471361409  + 77.75054398390 *T)
    sumB2  +=       1.103 * cos(5.54212014132  + 222.86032299360 *T)
    sumB2  +=       1.205 * cos(5.05109252937  + 380.12776796000 *T)
    sumB2  +=       1.146 * cos(1.95280464754  + 529.69096509460 *T)
    sumB2  +=       0.977 * cos(1.52652616357  + 77.96299230500 *T)
    sumB2  +=       1.025 * cos(4.33698643491  + 2.96894541660 *T)
    sumB2  +=       0.858 * cos(2.78728745263  + 111.43016149680 *T)
    sumB2  +=       0.868 * cos(5.55175791193  + 38.13303563780 *T)
    sumB2  +=       0.633 * cos(0.41074353315  + 213.29909543800 *T)
    sumB2  +=       0.596 * cos(5.39265533517  + 127.47179660680 *T)
    sumB2  +=       0.586 * cos(4.00404667232  + 109.94568878850 *T)
    sumB2  +=       0.543 * cos(2.40369406419  + 153.49535039770 *T)
    sumB2  +=       0.486 * cos(2.05237757516  + 299.12639426920 *T)
    sumB2  +=       0.557 * cos(3.13408880388  + 65.22037101170 *T)
    sumB2  +=       0.457 * cos(3.96543219832  + 454.90936652730 *T)
    sumB2  +=       0.481 * cos(2.81511187371  + 160.60889739850 *T)
    sumB2  +=       0.421 * cos(2.16819778071  + 56.62235130260 *T)
    sumB2  +=       0.326 * cos(4.52920012430  + 39.61750834610 *T)
    sumB2  +=       0.308 * cos(6.26508780547  + 70.32818044240 *T)
    sumB2  +=       0.338 * cos(5.14594268587  + 3.18139373770 *T)
    sumB2  +=       0.288 * cos(0.47061435406  + 22.09140052780 *T)
    sumB2  +=       0.336 * cos(4.23512034174  + 35.16409022120 *T)
    sumB2  +=       0.316 * cos(3.93430525759  + 52.69019803950 *T)
    sumB2  +=       0.306 * cos(1.10359318443  + 70.84944530420 *T)
    sumB2  +=       0.250 * cos(3.58780257084  + 202.25339517410 *T)
    sumB2  +=       0.239 * cos(4.77679306080  + 87.31177153950 *T)
    sumB2  +=       0.227 * cos(2.74138067839  + 12.53017297220 *T)
    sumB2  +=       0.263 * cos(1.59203582407  + 84.34282612290 *T)
    sumB2  +=       0.215 * cos(3.88195737361  + 131.40394986990 *T)
    sumB2  +=       0.216 * cos(5.80700510713  + 74.66972398270 *T)
    sumB2  +=       0.264 * cos(1.82574036051  + 77.22927912210 *T)
    sumB2  +=       0.222 * cos(0.04111883550  + 184.72728735580 *T)
    sumB2  +=       0.197 * cos(1.56602555362  + 9.56122755560 *T)
    sumB2  +=       0.193 * cos(2.26416938160  + 75.74480641380 *T)
    sumB2  +=       0.179 * cos(2.69065316892  + 145.63104387150 *T)
    sumB2  +=       0.170 * cos(2.75844544119  + 73.81839072080 *T)
    sumB2  +=       0.155 * cos(3.55393249110  + 18.15924726470 *T)
    sumB2  +=       0.174 * cos(4.76111441901  + 277.03499374140 *T)
    sumB2  +=       0.140 * cos(0.28714762870  + 221.37585028530 *T)
    sumB2  +=       0.134 * cos(1.29065526326  + 206.18554843720 *T)
    sumB2  +=       0.127 * cos(6.17908901556  + 62.25142559510 *T)
    sumB2  +=       0.116 * cos(6.26646620658  + 220.41264243880 *T)

    # URANUS Latitude
    sumB3  = 0
    sumB3  +=     267.832 * cos(1.25097888291  + 74.78159856730 *T)
    sumB3  +=      11.048 * cos(3.14159265359  + 0.00000000000 *T)
    sumB3  +=       6.154 * cos(4.00663614486  + 149.56319713460 *T)
    sumB3  +=       3.361 * cos(5.77804694935  + 73.29712585900 *T)
    sumB3  +=       1.602 * cos(1.05657834344  + 63.73589830340 *T)
    sumB3  +=       1.265 * cos(1.66795295537  + 78.71375183040 *T)
    sumB3  +=       1.183 * cos(2.58856450374  + 138.51749687070 *T)
    sumB3  +=       1.087 * cos(0.28687213135  + 76.26607127560 *T)
    sumB3  +=       0.640 * cos(1.87238784591  + 71.60020482960 *T)
    sumB3  +=       0.590 * cos(0.80206040001  + 1.48447270830 *T)
    sumB3  +=       0.467 * cos(4.42872012006  + 148.07872442630 *T)
    sumB3  +=       0.272 * cos(4.00684090176  + 85.82729883120 *T)
    sumB3  +=       0.203 * cos(0.60406901282  + 71.81265315070 *T)
    sumB3  +=       0.180 * cos(5.55657564049  + 3.93215326310 *T)
    sumB3  +=       0.168 * cos(4.67745630044  + 70.84944530420 *T)
    sumB3  +=       0.170 * cos(2.93672195979  + 11.04570026390 *T)

    # URANUS Latitude
    sumB4  = 0
    sumB4  +=       5.719 * cos(2.85499529315  + 74.78159856730 *T)
    sumB4  +=       0.300 * cos(3.14159265359  + 0.00000000000 *T)

    # URANUS Radius
    sumR0  = 0
    sumR0  +=  1921264847.881 * cos(0.00000000000  + 0.00000000000 *T)
    sumR0  +=  88784984.055 * cos(5.60377526994  + 74.78159856730 *T)
    sumR0  +=  3440835.545 * cos(0.32836098991  + 73.29712585900 *T)
    sumR0  +=  2055653.495 * cos(1.78295170028  + 149.56319713460 *T)
    sumR0  +=  649321.851 * cos(4.52247298119  + 76.26607127560 *T)
    sumR0  +=  602248.144 * cos(3.86003820462  + 63.73589830340 *T)
    sumR0  +=  496404.171 * cos(1.40139934716  + 454.90936652730 *T)
    sumR0  +=  338525.522 * cos(1.58002682946  + 138.51749687070 *T)
    sumR0  +=  243508.222 * cos(1.57086595074  + 71.81265315070 *T)
    sumR0  +=  190521.915 * cos(1.99809364502  + 1.48447270830 *T)
    sumR0  +=  161858.251 * cos(2.79137863469  + 148.07872442630 *T)
    sumR0  +=  143705.902 * cos(1.38368574483  + 11.04570026390 *T)
    sumR0  +=   93192.359 * cos(0.17437193645  + 36.64856292950 *T)
    sumR0  +=   71424.265 * cos(4.24509327405  + 224.34479570190 *T)
    sumR0  +=   89805.842 * cos(3.66105366329  + 109.94568878850 *T)
    sumR0  +=   39009.624 * cos(1.66971128869  + 70.84944530420 *T)
    sumR0  +=   46677.322 * cos(1.39976563936  + 35.16409022120 *T)
    sumR0  +=   39025.681 * cos(3.36234710692  + 277.03499374140 *T)
    sumR0  +=   36755.160 * cos(3.88648934736  + 146.59425171800 *T)
    sumR0  +=   30348.875 * cos(0.70100446346  + 151.04766984290 *T)
    sumR0  +=   29156.264 * cos(3.18056174556  + 77.75054398390 *T)
    sumR0  +=   20471.584 * cos(1.55588961500  + 202.25339517410 *T)
    sumR0  +=   25620.360 * cos(5.25656292802  + 380.12776796000 *T)
    sumR0  +=   25785.805 * cos(3.78537741503  + 85.82729883120 *T)
    sumR0  +=   22637.152 * cos(0.72519137745  + 529.69096509460 *T)
    sumR0  +=   20473.163 * cos(2.79639811626  + 70.32818044240 *T)
    sumR0  +=   17900.561 * cos(0.55455488605  + 2.96894541660 *T)
    sumR0  +=   12328.151 * cos(5.96039150918  + 127.47179660680 *T)
    sumR0  +=   14701.566 * cos(4.90434406648  + 108.46121608020 *T)
    sumR0  +=   11494.701 * cos(0.43774027872  + 65.22037101170 *T)
    sumR0  +=   15502.809 * cos(5.35405037603  + 38.13303563780 *T)
    sumR0  +=   10792.699 * cos(1.42104858472  + 213.29909543800 *T)
    sumR0  +=   11696.085 * cos(3.29825599114  + 3.93215326310 *T)
    sumR0  +=   11959.355 * cos(1.75044072173  + 984.60033162190 *T)
    sumR0  +=   12896.507 * cos(2.62154018241  + 111.43016149680 *T)
    sumR0  +=   11852.996 * cos(0.99342814582  + 52.69019803950 *T)
    sumR0  +=    9111.446 * cos(4.99638600045  + 62.25142559510 *T)
    sumR0  +=    8420.550 * cos(5.25350716616  + 222.86032299360 *T)
    sumR0  +=    7449.125 * cos(0.79491905956  + 351.81659230870 *T)
    sumR0  +=    8402.147 * cos(5.03877516489  + 415.55249061210 *T)
    sumR0  +=    6046.370 * cos(5.67960948357  + 78.71375183040 *T)
    sumR0  +=    5524.133 * cos(3.11499484161  + 9.56122755560 *T)
    sumR0  +=    7329.454 * cos(3.97277527840  + 183.24281464750 *T)
    sumR0  +=    5444.878 * cos(5.10575635361  + 145.10977900970 *T)
    sumR0  +=    5238.103 * cos(2.62960141797  + 33.67961751290 *T)
    sumR0  +=    4079.167 * cos(3.22064788674  + 340.77089204480 *T)
    sumR0  +=    3801.606 * cos(6.10985558505  + 184.72728735580 *T)
    sumR0  +=    3919.476 * cos(4.25015288873  + 39.61750834610 *T)
    sumR0  +=    2940.492 * cos(2.14637460319  + 137.03302416240 *T)
    sumR0  +=    3781.219 * cos(3.45840272873  + 456.39383923560 *T)
    sumR0  +=    2942.239 * cos(0.42393808854  + 299.12639426920 *T)
    sumR0  +=    3686.787 * cos(2.48718116535  + 453.42489381900 *T)
    sumR0  +=    3101.743 * cos(4.14031063896  + 219.89137757700 *T)
    sumR0  +=    2962.641 * cos(0.82977991995  + 56.62235130260 *T)
    sumR0  +=    2937.799 * cos(3.67657450930  + 140.00196957900 *T)
    sumR0  +=    2865.128 * cos(0.30996903761  + 12.53017297220 *T)
    sumR0  +=    2538.032 * cos(4.85457831993  + 131.40394986990 *T)
    sumR0  +=    1962.510 * cos(5.24342224065  + 84.34282612290 *T)
    sumR0  +=    2363.550 * cos(0.44253328372  + 554.06998748280 *T)
    sumR0  +=    1979.394 * cos(6.12836181686  + 106.97674337190 *T)
    sumR0  +=    2182.572 * cos(2.94040431638  + 305.34616939270 *T)
    sumR0  +=    1962.974 * cos(0.04114739120  + 221.37585028530 *T)
    sumR0  +=    1829.560 * cos(4.01105771632  + 68.84370773410 *T)
    sumR0  +=    1642.920 * cos(0.35564102554  + 67.66805156650 *T)
    sumR0  +=    1584.850 * cos(3.16267171762  + 225.82926841020 *T)
    sumR0  +=    1848.655 * cos(2.91111759376  + 909.81873305460 *T)
    sumR0  +=    1632.430 * cos(4.23061792837  + 22.09140052780 *T)
    sumR0  +=    1401.390 * cos(1.39084023521  + 265.98929347750 *T)
    sumR0  +=    1403.717 * cos(5.63563637532  + 4.45341812490 *T)
    sumR0  +=    1655.866 * cos(1.96431297431  + 79.23501669220 *T)
    sumR0  +=    1248.978 * cos(5.44027380866  + 54.17467074780 *T)
    sumR0  +=    1563.447 * cos(1.47917835549  + 112.91463420510 *T)
    sumR0  +=    1248.054 * cos(4.88984353601  + 479.28838891550 *T)
    sumR0  +=    1197.439 * cos(2.52185744943  + 145.63104387150 *T)
    sumR0  +=    1506.952 * cos(5.24186185583  + 181.75834193920 *T)
    sumR0  +=    1481.746 * cos(5.66203046912  + 152.53214255120 *T)
    sumR0  +=    1438.838 * cos(1.53046287618  + 447.79581952650 *T)
    sumR0  +=    1408.514 * cos(4.41921749601  + 462.02291352810 *T)
    sumR0  +=    1477.112 * cos(4.32214690647  + 256.53994050650 *T)
    sumR0  +=    1228.314 * cos(5.97703331040  + 59.80374504030 *T)
    sumR0  +=    1249.958 * cos(6.24484546141  + 160.60889739850 *T)
    sumR0  +=     906.468 * cos(5.62025869483  + 74.66972398270 *T)
    sumR0  +=    1090.681 * cos(4.15393813845  + 77.96299230500 *T)
    sumR0  +=     844.931 * cos(0.12943398585  + 82.85835341460 *T)
    sumR0  +=     900.363 * cos(2.37315925843  + 74.89347315190 *T)
    sumR0  +=    1071.957 * cos(1.74286714339  + 528.20649238630 *T)
    sumR0  +=     689.708 * cos(3.08097059985  + 69.36497259590 *T)
    sumR0  +=     593.798 * cos(4.50074517056  + 8.07675484730 *T)
    sumR0  +=     718.559 * cos(4.00047509264  + 128.95626931510 *T)
    sumR0  +=     699.574 * cos(0.03987168068  + 143.62530630140 *T)
    sumR0  +=     575.656 * cos(5.89552672641  + 66.70484372000 *T)
    sumR0  +=     759.004 * cos(2.13700057433  + 692.58748435350 *T)
    sumR0  +=     710.449 * cos(5.41605755095  + 218.40690486870 *T)
    sumR0  +=     548.672 * cos(5.62811496970  + 3.18139373770 *T)
    sumR0  +=     651.632 * cos(4.42340061551  + 18.15924726470 *T)
    sumR0  +=     539.825 * cos(6.20788667166  + 71.60020482960 *T)
    sumR0  +=     544.539 * cos(5.69375108253  + 203.73786788240 *T)
    sumR0  +=     710.276 * cos(4.21967260022  + 381.61224066830 *T)
    sumR0  +=     593.819 * cos(3.83805798523  + 32.19514480460 *T)
    sumR0  +=     710.134 * cos(4.48972171999  + 293.18850343600 *T)
    sumR0  +=     705.482 * cos(0.45521177725  + 835.03713448730 *T)
    sumR0  +=     588.000 * cos(5.08252923316  + 186.21176006410 *T)
    sumR0  +=     598.231 * cos(0.35815291076  + 269.92144674060 *T)
    sumR0  +=     641.914 * cos(2.71127457036  + 87.31177153950 *T)
    sumR0  +=     495.621 * cos(2.65094755989  + 200.76892246580 *T)
    sumR0  +=     630.252 * cos(4.46146214548  + 275.55052103310 *T)
    sumR0  +=     575.195 * cos(5.57862480486  + 2.44768055480 *T)
    sumR0  +=     569.870 * cos(1.63930932740  + 77.22927912210 *T)
    sumR0  +=     556.672 * cos(1.07231961344  + 1059.38193018920 *T)
    sumR0  +=     449.439 * cos(0.27981733949  + 617.80588578620 *T)
    sumR0  +=     463.608 * cos(1.43448297993  + 297.64192156090 *T)
    sumR0  +=     436.547 * cos(0.52802035072  + 209.36694217490 *T)
    sumR0  +=     463.938 * cos(2.35443114417  + 211.81462272970 *T)
    sumR0  +=     435.943 * cos(2.10077211065  + 1514.29129671650 *T)
    sumR0  +=     515.534 * cos(3.23274579379  + 284.14854074220 *T)
    sumR0  +=     454.879 * cos(4.08364210459  + 99.16062095550 *T)
    sumR0  +=     477.430 * cos(2.89397217998  + 39.35687591520 *T)
    sumR0  +=     542.331 * cos(5.39481705077  + 278.51946644970 *T)
    sumR0  +=     410.087 * cos(3.04968860441  + 404.50679034820 *T)
    sumR0  +=     367.848 * cos(0.71159607058  + 125.98732389850 *T)
    sumR0  +=     503.096 * cos(5.83931251717  + 191.20769491020 *T)
    sumR0  +=     487.532 * cos(0.06402454583  + 60.76695288680 *T)
    sumR0  +=     455.043 * cos(2.59321186669  + 490.33408917940 *T)
    sumR0  +=     436.291 * cos(2.08183813746  + 51.20572533120 *T)
    sumR0  +=     435.803 * cos(2.79445203085  + 75.74480641380 *T)
    sumR0  +=     323.546 * cos(4.82899980859  + 195.13984817330 *T)
    sumR0  +=     359.363 * cos(0.00868012078  + 35.42472265210 *T)
    sumR0  +=     429.314 * cos(3.08031550488  + 41.10198105440 *T)
    sumR0  +=     320.021 * cos(5.48625497747  + 14.97785352700 *T)
    sumR0  +=     414.331 * cos(0.09012800478  + 258.02441321480 *T)
    sumR0  +=     379.715 * cos(0.05832815311  + 378.64329525170 *T)
    sumR0  +=     420.062 * cos(2.25393983318  + 81.00137369080 *T)
    sumR0  +=     357.721 * cos(4.71414305625  + 173.94221952280 *T)
    sumR0  +=     358.922 * cos(0.35213227553  + 426.59819087600 *T)
    sumR0  +=     405.410 * cos(6.12263257999  + 24.37902238820 *T)
    sumR0  +=     365.158 * cos(5.59483211224  + 255.05546779820 *T)
    sumR0  +=     308.102 * cos(3.92355394354  + 116.42609634290 *T)
    sumR0  +=     325.660 * cos(4.71996698332  + 134.58534360760 *T)
    sumR0  +=     292.781 * cos(3.99521194830  + 72.33391801250 *T)
    sumR0  +=     386.543 * cos(0.68619006966  + 230.56457082540 *T)
    sumR0  +=     305.686 * cos(3.76108783519  + 344.70304530790 *T)
    sumR0  +=     286.972 * cos(1.84990335310  + 153.49535039770 *T)
    sumR0  +=     353.640 * cos(4.65717995107  + 329.83706636550 *T)
    sumR0  +=     302.051 * cos(0.13190003806  + 565.11568774670 *T)
    sumR0  +=     241.128 * cos(1.60454142389  + 81.37388070630 *T)
    sumR0  +=     249.829 * cos(4.24205256241  + 75.30286342910 *T)
    sumR0  +=     245.063 * cos(5.94905404273  + 20.60692781950 *T)
    sumR0  +=     248.277 * cos(1.06282887181  + 105.49227066360 *T)
    sumR0  +=     305.353 * cos(2.55534744586  + 6208.29425142410 *T)
    sumR0  +=     296.328 * cos(4.21100245276  + 1364.72809958190 *T)
    sumR0  +=     219.938 * cos(2.96119055727  + 120.35824960600 *T)
    sumR0  +=     233.564 * cos(2.97074409938  + 46.20979048510 *T)
    sumR0  +=     262.422 * cos(3.83652250971  + 831.10498122420 *T)
    sumR0  +=     233.546 * cos(4.48117006140  + 628.85158605010 *T)
    sumR0  +=     187.432 * cos(3.03529190348  + 135.54855145410 *T)
    sumR0  +=     216.776 * cos(3.42907414802  + 241.61027108930 *T)
    sumR0  +=     255.760 * cos(1.16707893460  + 177.87437278590 *T)
    sumR0  +=     220.458 * cos(0.19633492290  + 180.27386923090 *T)
    sumR0  +=     224.519 * cos(0.40677777819  + 114.39910691340 *T)
    sumR0  +=     205.398 * cos(2.30380942634  + 259.50888592310 *T)
    sumR0  +=     211.106 * cos(4.93079982424  + 103.09277421860 *T)
    sumR0  +=     175.758 * cos(5.50822822216  + 7.11354700080 *T)
    sumR0  +=     188.512 * cos(2.23588941288  + 5.41662597140 *T)
    sumR0  +=     171.718 * cos(5.21730232334  + 41.64449777560 *T)
    sumR0  +=     176.136 * cos(1.95958319897  + 756.32338265690 *T)
    sumR0  +=     170.447 * cos(4.94978757413  + 206.18554843720 *T)
    sumR0  +=     169.454 * cos(4.04319823722  + 55.65914345610 *T)
    sumR0  +=     219.015 * cos(0.24790282027  + 294.67297614430 *T)
    sumR0  +=     187.768 * cos(2.04538775456  + 408.43894361130 *T)
    sumR0  +=     182.258 * cos(0.70728384467  + 391.17346822390 *T)
    sumR0  +=     192.095 * cos(5.76718231319  + 291.70403072770 *T)
    sumR0  +=     153.684 * cos(4.70659406659  + 543.02428721890 *T)
    sumR0  +=     170.043 * cos(4.50995820508  + 288.08069400530 *T)
    sumR0  +=     164.097 * cos(5.22527540372  + 67.35923502580 *T)
    sumR0  +=     194.341 * cos(6.11690364710  + 414.06801790380 *T)
    sumR0  +=     168.027 * cos(5.25810639105  + 518.64526483070 *T)
    sumR0  +=     156.641 * cos(0.66304836778  + 220.41264243880 *T)
    sumR0  +=     182.330 * cos(0.78383856974  + 417.03696332040 *T)
    sumR0  +=     167.462 * cos(4.92241597775  + 422.66603761290 *T)
    sumR0  +=     170.770 * cos(2.30927162659  + 98.89998852460 *T)
    sumR0  +=     161.678 * cos(3.27259601116  + 443.86366626340 *T)
    sumR0  +=     132.763 * cos(2.88875442023  + 373.90799283650 *T)
    sumR0  +=     161.140 * cos(3.82341391177  + 451.94042111070 *T)
    sumR0  +=     179.292 * cos(4.82405681293  + 366.48562929500 *T)
    sumR0  +=     178.153 * cos(3.98026039043  + 10138.50394764370 *T)
    sumR0  +=     141.929 * cos(1.26972581554  + 159.12442469020 *T)
    sumR0  +=     153.750 * cos(4.27847681414  + 45.57665103870 *T)
    sumR0  +=     161.513 * cos(4.99545008738  + 73.81839072080 *T)
    sumR0  +=     146.315 * cos(2.65664902119  + 465.95506679120 *T)
    sumR0  +=     124.875 * cos(4.30470898895  + 339.28641933650 *T)
    sumR0  +=     154.620 * cos(4.32046228120  + 760.25553592000 *T)
    sumR0  +=     142.894 * cos(2.07773752143  + 457.87831194390 *T)
    sumR0  +=     152.408 * cos(4.64742446768  + 155.78297225810 *T)
    sumR0  +=     116.389 * cos(4.43513730944  + 5.93789083320 *T)
    sumR0  +=     113.444 * cos(4.65351596266  + 80.19822453870 *T)
    sumR0  +=     107.611 * cos(3.77290419929  + 142.44965013380 *T)
    sumR0  +=     133.740 * cos(5.30894739047  + 14.01464568050 *T)
    sumR0  +=     116.104 * cos(2.51182725670  + 296.15744885260 *T)
    sumR0  +=     129.106 * cos(0.36277717661  + 96.87299909510 *T)
    sumR0  +=     122.766 * cos(2.38341351026  + 141.48644228730 *T)
    sumR0  +=     101.368 * cos(1.05739625315  + 92.30770638560 *T)
    sumR0  +=     114.669 * cos(6.24863527978  + 767.36908292080 *T)
    sumR0  +=     113.283 * cos(0.83051319425  + 100.38446123290 *T)
    sumR0  +=     107.199 * cos(2.39365512354  + 347.88443904560 *T)
    sumR0  +=      95.443 * cos(0.80094579583  + 342.25536475310 *T)
    sumR0  +=     110.789 * cos(0.38651051525  + 216.92243216040 *T)
    sumR0  +=     126.978 * cos(0.42359358250  + 331.32153907380 *T)
    sumR0  +=     112.635 * cos(0.08107814739  + 558.00214074590 *T)
    sumR0  +=     103.166 * cos(0.69792283389  + 358.93013930950 *T)
    sumR0  +=     111.474 * cos(0.75023459027  + 80.71948940050 *T)
    sumR0  +=      90.902 * cos(5.16530481614  + 144.14657116320 *T)
    sumR0  +=      90.677 * cos(0.22036476597  + 333.65734504400 *T)
    sumR0  +=      98.568 * cos(4.33164222339  + 74.52096613640 *T)
    sumR0  +=      89.306 * cos(2.18851161761  + 74.82978267710 *T)
    sumR0  +=     117.216 * cos(3.94965784596  + 74.26033370550 *T)
    sumR0  +=      89.088 * cos(5.87783179087  + 74.73341445750 *T)
    sumR0  +=      97.316 * cos(0.69429695020  + 977.48678462110 *T)
    sumR0  +=     116.587 * cos(1.83677031994  + 1289.94650101460 *T)
    sumR0  +=      85.449 * cos(5.80255966149  + 6.59228213900 *T)
    sumR0  +=      86.823 * cos(5.61973473261  + 300.61086697750 *T)
    sumR0  +=     105.226 * cos(5.94513614941  + 328.35259365720 *T)
    sumR0  +=     112.117 * cos(1.21168089807  + 329.72519178090 *T)
    sumR0  +=      82.982 * cos(2.20797412496  + 74.94165726170 *T)
    sumR0  +=      94.345 * cos(4.53937998713  + 28.57180808220 *T)
    sumR0  +=     106.847 * cos(1.82071328579  + 306.83064210100 *T)
    sumR0  +=     103.572 * cos(2.99368274596  + 6.21977512350 *T)
    sumR0  +=     106.357 * cos(0.81583874750  + 1087.69310584050 *T)
    sumR0  +=      77.728 * cos(2.73390123734  + 110.20632121940 *T)
    sumR0  +=      98.405 * cos(3.73478182667  + 75.04223099820 *T)
    sumR0  +=      86.231 * cos(2.83316881064  + 983.11585891360 *T)
    sumR0  +=      89.023 * cos(4.73754458960  + 604.47256366190 *T)
    sumR0  +=      83.013 * cos(1.88273535999  + 387.24131496080 *T)
    sumR0  +=      90.227 * cos(3.80367274711  + 986.08480433020 *T)
    sumR0  +=      84.598 * cos(1.25774132938  + 142.14083359310 *T)
    sumR0  +=      74.690 * cos(1.35097482767  + 350.33211960040 *T)
    sumR0  +=      95.770 * cos(5.54845504768  + 969.62247809490 *T)
    sumR0  +=      90.277 * cos(0.36773710508  + 0.96320784650 *T)
    sumR0  +=      82.748 * cos(5.85590525764  + 74.62153987290 *T)
    sumR0  +=      75.828 * cos(2.78019216029  + 88.11492069160 *T)
    sumR0  +=      83.850 * cos(1.84386358668  + 227.31374111850 *T)
    sumR0  +=      70.705 * cos(4.65567024014  + 44.72531777680 *T)
    sumR0  +=      71.322 * cos(3.64963906751  + 894.84087952760 *T)
    sumR0  +=      94.141 * cos(4.98819201726  + 403.13419222450 *T)
    sumR0  +=      88.966 * cos(4.43895583278  + 154.01661525950 *T)
    sumR0  +=      79.436 * cos(5.66662613679  + 267.47376618580 *T)
    sumR0  +=      75.615 * cos(5.40971072536  + 50.40257617910 *T)
    sumR0  +=      68.583 * cos(4.76679841388  + 991.71387862270 *T)
    sumR0  +=      65.256 * cos(0.69286370395  + 152.74459087230 *T)
    sumR0  +=      63.031 * cos(2.89946567712  + 79.88940799800 *T)
    sumR0  +=      63.878 * cos(0.09820555288  + 681.54178408960 *T)
    sumR0  +=      80.101 * cos(2.97520561915  + 526.72201967800 *T)
    sumR0  +=      69.693 * cos(3.95281159807  + 187.69623277240 *T)
    sumR0  +=      59.492 * cos(3.59642351692  + 58.10682401090 *T)
    sumR0  +=      59.273 * cos(0.50930692071  + 28.31117565130 *T)
    sumR0  +=      68.590 * cos(2.41880311530  + 235.39049596580 *T)
    sumR0  +=      66.007 * cos(5.04558399435  + 30.71067209630 *T)
    sumR0  +=      70.223 * cos(3.73647415486  + 546.95644048200 *T)
    sumR0  +=      66.836 * cos(0.85506033017  + 522.57741809380 *T)
    sumR0  +=      63.027 * cos(0.29269109052  + 119.50691634410 *T)
    sumR0  +=      62.023 * cos(2.31557510311  + 74.03083904190 *T)
    sumR0  +=      71.379 * cos(3.16967571102  + 23.57587323610 *T)
    sumR0  +=      74.827 * cos(5.36812537961  + 373.01422095920 *T)
    sumR0  +=      64.204 * cos(2.36817149460  + 157.63995198190 *T)
    sumR0  +=      70.712 * cos(0.55830476304  + 92.94084583200 *T)
    sumR0  +=      55.762 * cos(5.27011035858  + 874.39401040250 *T)
    sumR0  +=      75.638 * cos(4.66344127677  + 101.86893394120 *T)
    sumR0  +=      73.727 * cos(6.20581665991  + 312.45971639350 *T)
    sumR0  +=      72.940 * cos(0.58406607757  + 367.97010200330 *T)
    sumR0  +=      53.230 * cos(2.24728742995  + 17.52610781830 *T)
    sumR0  +=      63.139 * cos(4.59563922296  + 67.88049988760 *T)
    sumR0  +=      60.550 * cos(0.57591315857  + 253.57099508990 *T)
    sumR0  +=      52.946 * cos(2.45947017614  + 264.50482076920 *T)
    sumR0  +=      70.236 * cos(1.51860943454  + 552.58551477450 *T)
    sumR0  +=      68.624 * cos(2.44507780453  + 555.55446019110 *T)
    sumR0  +=      62.796 * cos(0.33786296181  + 561.18353448360 *T)
    sumR0  +=      49.009 * cos(1.09233728279  + 19.12245511120 *T)
    sumR0  +=      64.636 * cos(5.27469970900  + 68.18931642830 *T)
    sumR0  +=      62.957 * cos(5.35891188483  + 92.04707395470 *T)
    sumR0  +=      47.664 * cos(3.90924952181  + 192.69216761850 *T)
    sumR0  +=      65.279 * cos(4.23629510074  + 771.30123618390 *T)
    sumR0  +=      65.190 * cos(3.73942854797  + 536.80451209540 *T)
    sumR0  +=      59.452 * cos(6.10554259948  + 365.00115658670 *T)
    sumR0  +=      52.153 * cos(1.71734604937  + 905.88657979150 *T)
    sumR0  +=      46.035 * cos(3.87093684776  + 210.33015002140 *T)
    sumR0  +=      46.429 * cos(5.97423131576  + 477.80391620720 *T)
    sumR0  +=      62.115 * cos(2.67544358037  + 130.44074202340 *T)
    sumR0  +=      46.038 * cos(3.89378239085  + 48.75804477640 *T)
    sumR0  +=      42.663 * cos(3.81519760715  + 61.28821774860 *T)
    sumR0  +=      53.909 * cos(2.86457147106  + 353.30106501700 *T)
    sumR0  +=      46.936 * cos(1.00011046774  + 166.82867252200 *T)
    sumR0  +=      42.217 * cos(2.61748790314  + 90.82323367730 *T)
    sumR0  +=      43.324 * cos(4.15777895713  + 173.68158709190 *T)
    sumR0  +=      41.296 * cos(1.79930408254  + 149.45132255000 *T)
    sumR0  +=      44.960 * cos(1.76623306927  + 0.52126486180 *T)
    sumR0  +=      51.904 * cos(2.97773319756  + 383.09671337660 *T)
    sumR0  +=      42.931 * cos(1.57416456203  + 120.99138905240 *T)
    sumR0  +=      49.611 * cos(4.03427920470  + 303.86169668440 *T)
    sumR0  +=      45.263 * cos(3.58382163089  + 97.41551581630 *T)
    sumR0  +=      38.695 * cos(2.39404211169  + 31.49256938900 *T)
    sumR0  +=      38.072 * cos(5.79473670350  + 75.53235809270 *T)
    sumR0  +=      50.126 * cos(4.76412907201  + 911.30320576290 *T)
    sumR0  +=      50.884 * cos(5.15513957132  + 439.78275515400 *T)
    sumR0  +=      43.148 * cos(0.84999004804  + 58.31927233200 *T)
    sumR0  +=      42.732 * cos(5.17318058934  + 162.09337010680 *T)
    sumR0  +=      50.298 * cos(5.81603435915  + 66.91729204110 *T)
    sumR0  +=      35.639 * cos(1.87447823723  + 472.17484191470 *T)
    sumR0  +=      49.963 * cos(1.88943490790  + 42.58645376270 *T)
    sumR0  +=      39.974 * cos(1.74262050679  + 89.75945209430 *T)
    sumR0  +=      45.252 * cos(1.92511912328  + 55.13787859430 *T)
    sumR0  +=      44.896 * cos(1.48355901890  + 450.97721326420 *T)
    sumR0  +=      34.297 * cos(5.20257496546  + 316.39186965660 *T)
    sumR0  +=      46.355 * cos(0.33942039181  + 273.10284047830 *T)
    sumR0  +=      37.152 * cos(2.03757941865  + 117.91056905120 *T)
    sumR0  +=      46.106 * cos(5.62315633955  + 1819.63746610920 *T)
    sumR0  +=      39.368 * cos(4.19402806344  + 486.40193591630 *T)
    sumR0  +=      41.039 * cos(4.82994471947  + 149.67507171920 *T)
    sumR0  +=      44.959 * cos(0.72694662195  + 3265.83082813250 *T)
    sumR0  +=      43.617 * cos(0.75332422672  + 404.61866493280 *T)
    sumR0  +=      31.823 * cos(3.84768075667  + 20.44686912510 *T)
    sumR0  +=      44.196 * cos(4.36769721266  + 418.26080359780 *T)
    sumR0  +=      37.900 * cos(3.02928044053  + 167.08930495290 *T)
    sumR0  +=      43.684 * cos(1.57328182739  + 491.55792945680 *T)
    sumR0  +=      34.004 * cos(1.26257052908  + 260.99335863140 *T)
    sumR0  +=      31.276 * cos(4.16123711648  + 13.33332212430 *T)
    sumR0  +=      39.984 * cos(2.86626125620  + 468.24268865160 *T)
    sumR0  +=      36.490 * cos(2.58804294589  + 68.56182344380 *T)
    sumR0  +=      32.364 * cos(3.11577354875  + 103.35340664950 *T)
    sumR0  +=      33.857 * cos(0.15592410716  + 24.11838995730 *T)
    sumR0  +=      35.933 * cos(1.36784550071  + 59.28248017850 *T)
    sumR0  +=      33.633 * cos(0.75501177400  + 290.21955801940 *T)
    sumR0  +=      29.751 * cos(5.33178627038  + 1033.35837639830 *T)
    sumR0  +=      32.036 * cos(4.67549858000  + 205.22234059070 *T)
    sumR0  +=      30.991 * cos(4.62823866461  + 258.87574647670 *T)
    sumR0  +=      35.268 * cos(1.00718464327  + 1108.13997496560 *T)
    sumR0  +=      33.366 * cos(3.40738625377  + 43.12897048390 *T)
    sumR0  +=      32.638 * cos(5.25485850258  + 114.13847448250 *T)
    sumR0  +=      29.825 * cos(5.64157476876  + 254.94359321360 *T)
    sumR0  +=      31.613 * cos(3.78231393110  + 152.01087768940 *T)
    sumR0  +=      30.980 * cos(2.26660677937  + 104.00779795530 *T)
    sumR0  +=      34.591 * cos(5.17326577255  + 25.60286266560 *T)
    sumR0  +=      28.398 * cos(1.76872790446  + 820.05928096030 *T)
    sumR0  +=      27.991 * cos(3.92486885309  + 199.28444975750 *T)
    sumR0  +=      28.986 * cos(2.58171811759  + 76.47851959670 *T)
    sumR0  +=      33.772 * cos(5.79359878723  + 274.06604832480 *T)
    sumR0  +=      29.401 * cos(5.93638676504  + 280.96714700450 *T)
    sumR0  +=      31.094 * cos(1.39352495971  + 178.78939652260 *T)
    sumR0  +=      30.118 * cos(0.44367887423  + 27.08733537390 *T)
    sumR0  +=      33.820 * cos(6.26168443513  + 401.64971951620 *T)
    sumR0  +=      27.513 * cos(2.15194454461  + 480.77286162380 *T)
    sumR0  +=      26.880 * cos(2.51300272780  + 123.53964334370 *T)
    sumR0  +=      26.139 * cos(0.21985367371  + 286.59622129700 *T)
    sumR0  +=      26.455 * cos(3.88229792258  + 372.42352012820 *T)
    sumR0  +=      33.974 * cos(1.44637843871  + 88.79624424780 *T)
    sumR0  +=      30.107 * cos(0.82723915882  + 100.64509366380 *T)
    sumR0  +=      27.715 * cos(4.64827434185  + 198.32124191100 *T)
    sumR0  +=      33.687 * cos(1.14348201049  + 82.48584639910 *T)
    sumR0  +=      26.493 * cos(1.97889544238  + 95.38852638680 *T)
    sumR0  +=      24.355 * cos(2.37839176150  + 146.38180339690 *T)
    sumR0  +=      26.590 * cos(0.39881920389  + 106.01353552540 *T)
    sumR0  +=      27.006 * cos(2.10206230691  + 1057.89745748090 *T)
    sumR0  +=      23.976 * cos(6.21233637686  + 16.67477455640 *T)
    sumR0  +=      30.970 * cos(5.34005431547  + 476.43131808350 *T)
    sumR0  +=      24.073 * cos(3.42953641968  + 1044.40407666220 *T)
    sumR0  +=      27.023 * cos(0.71284764471  + 248.72381809010 *T)
    sumR0  +=      29.098 * cos(3.99184722502  + 908.33426034630 *T)
    sumR0  +=      22.862 * cos(2.26978781393  + 175.16605980020 *T)
    sumR0  +=      24.026 * cos(0.36584131268  + 73.18525127440 *T)
    sumR0  +=      28.024 * cos(3.46485782266  + 1439.50969814920 *T)
    sumR0  +=      22.034 * cos(0.05163807300  + 33.13710079170 *T)
    sumR0  +=      22.185 * cos(5.32252126255  + 483.22054217860 *T)
    sumR0  +=      21.027 * cos(0.37224660652  + 214.78356814630 *T)
    sumR0  +=      20.548 * cos(1.80004483299  + 118.02244363580 *T)
    sumR0  +=      27.835 * cos(4.12412553530  + 694.07195706180 *T)
    sumR0  +=      25.500 * cos(5.49632191634  + 115.88357962170 *T)
    sumR0  +=      21.377 * cos(3.89179204956  + 66.18357885820 *T)
    sumR0  +=      27.201 * cos(5.76148797900  + 1215.16490244730 *T)
    sumR0  +=      24.984 * cos(0.65339418015  + 132.88842257820 *T)
    sumR0  +=      23.976 * cos(4.56161326826  + 458.84151979040 *T)
    sumR0  +=      21.116 * cos(1.13610706250  + 60.55450456570 *T)
    sumR0  +=      26.263 * cos(2.77532723118  + 490.07345674850 *T)
    sumR0  +=      26.369 * cos(3.37120039300  + 49.72125262290 *T)
    sumR0  +=      22.870 * cos(4.53135637620  + 78.40493528970 *T)
    sumR0  +=      26.872 * cos(3.26037129303  + 691.10301164520 *T)
    sumR0  +=      25.004 * cos(3.65018677651  + 73.40900044360 *T)
    sumR0  +=      20.874 * cos(3.92589972978  + 134.06407874580 *T)
    sumR0  +=      20.915 * cos(5.53955400138  + 129.91947716160 *T)
    sumR0  +=      23.067 * cos(2.56806856688  + 332.80601178210 *T)
    sumR0  +=      22.630 * cos(5.02721554401  + 150.52640498110 *T)
    sumR0  +=      19.123 * cos(1.92386327535  + 124.50285119020 *T)
    sumR0  +=      20.678 * cos(0.98302410602  + 29.20494752860 *T)
    sumR0  +=      18.755 * cos(1.07911898422  + 70.11573212130 *T)
    sumR0  +=      19.458 * cos(1.33847349577  + 616.32141307790 *T)
    sumR0  +=      23.071 * cos(3.93152899657  + 43.28902917830 *T)
    sumR0  +=      23.313 * cos(0.61185525008  + 189.72322220190 *T)
    sumR0  +=      19.660 * cos(1.40884902649  + 1589.07289528380 *T)
    sumR0  +=      24.990 * cos(0.91842956919  + 441.26722786230 *T)
    sumR0  +=      23.555 * cos(0.02127675886  + 593.42686339800 *T)
    sumR0  +=      18.288 * cos(4.55111843462  + 165.60483224460 *T)
    sumR0  +=      20.980 * cos(0.88504201898  + 326.86812094890 *T)
    sumR0  +=      24.940 * cos(4.63470443286  + 162.89651925890 *T)
    sumR0  +=      18.941 * cos(5.10763304564  + 81.89514556810 *T)
    sumR0  +=      18.911 * cos(1.23351635328  + 13.49338081870 *T)
    sumR0  +=      17.358 * cos(4.05768226252  + 403.02231763990 *T)
    sumR0  +=      17.362 * cos(5.28607227640  + 7.86430652620 *T)
    sumR0  +=      22.513 * cos(3.15059891398  + 419.74527630610 *T)
    sumR0  +=      21.237 * cos(2.14856256664  + 75.58474771940 *T)
    sumR0  +=      17.845 * cos(2.54349200329  + 47.06112374700 *T)
    sumR0  +=      16.995 * cos(2.48647736969  + 2043.98226181110 *T)
    sumR0  +=      23.676 * cos(5.80355919955  + 232.04904353370 *T)
    sumR0  +=      22.639 * cos(2.07623129509  + 699.70103135430 *T)
    sumR0  +=      19.261 * cos(1.56494156016  + 425.11371816770 *T)
    sumR0  +=      21.067 * cos(5.30844438236  + 237.67811782620 *T)
    sumR0  +=      22.733 * cos(0.28303126440  + 0.11187458460 *T)
    sumR0  +=      16.372 * cos(3.45984005656  + 0.75075952540 *T)
    sumR0  +=      21.213 * cos(0.95828006612  + 405.99126305650 *T)
    sumR0  +=      18.033 * cos(1.60723214246  + 215.43795945210 *T)
    sumR0  +=      16.267 * cos(4.89002016360  + 69.15252427480 *T)
    sumR0  +=      21.738 * cos(3.24738839789  + 1744.85586754190 *T)
    sumR0  +=      16.149 * cos(0.35803995032  + 77.06922042770 *T)
    sumR0  +=      21.710 * cos(0.88800040769  + 344.96367773880 *T)
    sumR0  +=      17.204 * cos(6.04366142241  + 32.24332891440 *T)
    sumR0  +=      17.883 * cos(4.01076173641  + 280.00393915800 *T)
    sumR0  +=      15.918 * cos(2.96623390816  + 25.86349509650 *T)
    sumR0  +=      14.769 * cos(3.73887340623  + 610.69233878540 *T)
    sumR0  +=      15.033 * cos(4.24825484707  + 228.27694896500 *T)
    sumR0  +=      15.586 * cos(5.07987082740  + 114.94162363460 *T)
    sumR0  +=      15.392 * cos(0.22971106129  + 17.26547538740 *T)
    sumR0  +=      15.354 * cos(0.25482391126  + 661.09491496450 *T)
    sumR0  +=      14.617 * cos(1.13349626273  + 823.99143422340 *T)
    sumR0  +=      16.232 * cos(3.43499743810  + 147.11551657980 *T)
    sumR0  +=      14.654 * cos(1.68288566884  + 207.88246946660 *T)
    sumR0  +=      17.682 * cos(5.94376629143  + 624.91943278700 *T)
    sumR0  +=      18.837 * cos(1.38335408070  + 377.15882254340 *T)
    sumR0  +=      15.425 * cos(1.66489033237  + 440.68227252570 *T)
    sumR0  +=      14.764 * cos(4.41710614445  + 16.46232623530 *T)
    sumR0  +=      14.402 * cos(0.41359448817  + 142.66209845490 *T)
    sumR0  +=      16.992 * cos(0.16042368544  + 438.29828244570 *T)
    sumR0  +=      13.268 * cos(3.04634728126  + 668.20846196530 *T)
    sumR0  +=      16.460 * cos(0.92068542861  + 369.08206769610 *T)
    sumR0  +=      17.239 * cos(4.51659246818  + 606.76018552230 *T)
    sumR0  +=      13.238 * cos(0.13650358961  + 216.48048917570 *T)
    sumR0  +=      15.832 * cos(4.94315562971  + 124.29040286910 *T)
    sumR0  +=      14.374 * cos(2.93700606008  + 419.48464387520 *T)
    sumR0  +=      12.927 * cos(1.65950183061  + 54.33472944220 *T)
    sumR0  +=      14.224 * cos(4.42286781619  + 47.69426319340 *T)
    sumR0  +=      12.753 * cos(0.03020931725  + 217.23124870110 *T)
    sumR0  +=      14.792 * cos(1.08447500622  + 49.50880430180 *T)
    sumR0  +=      14.031 * cos(3.68785757687  + 16.04163511000 *T)
    sumR0  +=      13.709 * cos(4.78890618802  + 72.77586099720 *T)
    sumR0  +=      13.073 * cos(1.54064778942  + 218.92816973050 *T)
    sumR0  +=      17.474 * cos(5.05621281434  + 564.85505531580 *T)
    sumR0  +=      12.686 * cos(3.44640888880  + 958.57677783100 *T)
    sumR0  +=      13.035 * cos(0.56445754615  + 1171.87587326900 *T)
    sumR0  +=      12.458 * cos(3.29187197133  + 902.70518605380 *T)
    sumR0  +=      11.893 * cos(1.41294011193  + 55.77101804070 *T)
    sumR0  +=      15.018 * cos(3.43209569509  + 19.01058052660 *T)
    sumR0  +=      16.470 * cos(2.04067754807  + 411.62033734900 *T)
    sumR0  +=      15.619 * cos(1.53464600544  + 833.55266177900 *T)
    sumR0  +=      15.678 * cos(5.92839374034  + 778.41478318470 *T)
    sumR0  +=      12.039 * cos(5.17748353434  + 135.33610313300 *T)
    sumR0  +=      15.523 * cos(3.54656631824  + 113.87784205160 *T)
    sumR0  +=      14.364 * cos(4.19825110964  + 89.33876096900 *T)
    sumR0  +=      15.424 * cos(2.12697366269  + 106.27416795630 *T)
    sumR0  +=      11.957 * cos(1.43314130608  + 455.87257437380 *T)
    sumR0  +=      15.938 * cos(5.49575810978  + 513.07988101300 *T)
    sumR0  +=      13.532 * cos(4.11463529983  + 95.22846769240 *T)
    sumR0  +=      15.105 * cos(1.86350524526  + 7.70424783180 *T)
    sumR0  +=      15.832 * cos(3.42498484109  + 79.51690098250 *T)
    sumR0  +=      11.492 * cos(4.65187455620  + 149.61138124440 *T)
    sumR0  +=      11.406 * cos(1.31085455047  + 63.62402371880 *T)
    sumR0  +=      14.469 * cos(3.35284802718  + 19.64371997300 *T)
    sumR0  +=      11.953 * cos(0.20979051344  + 65.87476231750 *T)
    sumR0  +=      12.039 * cos(0.01423238410  + 397.39324334740 *T)
    sumR0  +=      14.157 * cos(1.87440535404  + 6283.07584999140 *T)
    sumR0  +=      11.357 * cos(0.19079103112  + 5.62907429250 *T)
    sumR0  +=      14.109 * cos(0.09348109701  + 6133.51265285680 *T)
    sumR0  +=      15.322 * cos(3.54468546172  + 252.65597135320 *T)
    sumR0  +=      11.681 * cos(0.85100356112  + 5.10780943070 *T)
    sumR0  +=      14.134 * cos(5.66340426198  + 639.89728631400 *T)
    sumR0  +=      11.052 * cos(0.47607339302  + 150.08446199640 *T)
    sumR0  +=      11.507 * cos(5.19480309409  + 1182.92157353290 *T)
    sumR0  +=      11.492 * cos(2.05801478181  + 149.51501302480 *T)
    sumR0  +=      11.571 * cos(4.78210724970  + 334.29048449040 *T)
    sumR0  +=      10.671 * cos(4.67373109923  + 149.72325582900 *T)
    sumR0  +=      11.651 * cos(3.13272450186  + 93.90405367850 *T)
    sumR0  +=      14.316 * cos(0.08421279341  + 240.38643081190 *T)
    sumR0  +=      10.855 * cos(4.52379396618  + 453.94615868080 *T)
    sumR0  +=      11.900 * cos(1.41784572428  + 26.02355379090 *T)
    sumR0  +=      10.851 * cos(4.40625021974  + 57.14361616440 *T)
    sumR0  +=      13.385 * cos(0.76174742916  + 37.87240320690 *T)
    sumR0  +=      10.664 * cos(5.81644528276  + 193.65537546500 *T)
    sumR0  +=      10.700 * cos(5.34595506070  + 331.20966448920 *T)
    sumR0  +=      10.465 * cos(3.82648204886  + 180.16199464630 *T)
    sumR0  +=      13.350 * cos(0.86920479636  + 22.89454967990 *T)
    sumR0  +=      10.324 * cos(2.99969783109  + 525.75881183150 *T)
    sumR0  +=      14.293 * cos(1.06904465002  + 477.91579079180 *T)
    sumR0  +=      12.341 * cos(4.62813430535  + 1894.41906467650 *T)
    sumR0  +=      12.539 * cos(3.70404881494  + 67.07735073550 *T)
    sumR0  +=      11.771 * cos(1.07971321862  + 363.51668387840 *T)
    sumR0  +=      11.466 * cos(0.93528386040  + 121.84272231430 *T)
    sumR0  +=      12.839 * cos(0.31988787839  + 474.94684537520 *T)
    sumR0  +=      10.194 * cos(6.23976471898  + 84.18276742850 *T)
    sumR0  +=      12.300 * cos(2.85238700423  + 184.09414790940 *T)
    sumR0  +=      13.861 * cos(3.48367688770  + 157.26744496640 *T)
    sumR0  +=      11.395 * cos(4.25533440680  + 181.05576652360 *T)
    sumR0  +=      10.146 * cos(6.01371693363  + 43.24084506850 *T)
    sumR0  +=       9.889 * cos(0.01753733887  + 40.16002506730 *T)
    sumR0  +=      10.798 * cos(2.46364243940  + 140.65636088480 *T)
    sumR0  +=       9.893 * cos(2.72219544398  + 384.05992122310 *T)
    sumR0  +=      11.012 * cos(3.77284307154  + 494.26624244250 *T)
    sumR0  +=      10.226 * cos(1.49169427598  + 80.41067285980 *T)
    sumR0  +=      11.981 * cos(2.69741212203  + 369.45457471160 *T)
    sumR0  +=      10.658 * cos(1.78702880970  + 252.08652238160 *T)
    sumR0  +=      12.506 * cos(4.66852994807  + 64.69910614990 *T)
    sumR0  +=      11.640 * cos(5.70405852396  + 39.09624348430 *T)
    sumR0  +=      12.305 * cos(1.73322306013  + 229.08009811710 *T)
    sumR0  +=       9.954 * cos(3.72074935511  + 233.90602325750 *T)
    sumR0  +=      11.004 * cos(3.58723041577  + 449.28029223480 *T)
    sumR0  +=      10.581 * cos(2.79550711884  + 1246.65747183630 *T)
    sumR0  +=      10.411 * cos(2.86727145530  + 189.18070548070 *T)
    sumR0  +=       9.259 * cos(5.19834823120  + 749.20983565610 *T)
    sumR0  +=      12.468 * cos(0.76477162698  + 122.47586176070 *T)
    sumR0  +=      10.099 * cos(6.06894682979  + 156.15547927360 *T)
    sumR0  +=      12.671 * cos(6.19797171716  + 149.82382956550 *T)
    sumR0  +=       9.053 * cos(0.99447742265  + 109.31254934210 *T)
    sumR0  +=       9.595 * cos(1.00827882958  + 393.46109008430 *T)
    sumR0  +=      10.455 * cos(1.23531019351  + 148.59998928810 *T)
    sumR0  +=      10.645 * cos(5.50399216121  + 460.53844081980 *T)
    sumR0  +=      12.433 * cos(5.29843661600  + 20.49505323490 *T)
    sumR0  +=       9.002 * cos(5.51216892840  + 133.10087089930 *T)
    sumR0  +=       9.882 * cos(5.68987366321  + 42.53826965290 *T)
    sumR0  +=      12.110 * cos(3.14577799081  + 30.05628079050 *T)
    sumR0  +=      10.380 * cos(3.54528360301  + 619.29035849450 *T)
    sumR0  +=      10.139 * cos(1.90528799801  + 25.06034594440 *T)
    sumR0  +=      11.206 * cos(5.87823238990  + 832.58945393250 *T)
    sumR0  +=       9.283 * cos(3.00514787072  + 754.83890994860 *T)
    sumR0  +=      10.994 * cos(0.05721003392  + 54.28654533240 *T)
    sumR0  +=       8.981 * cos(5.82462023723  + 248.46318565920 *T)
    sumR0  +=       8.634 * cos(5.49123270314  + 448.68959140380 *T)
    sumR0  +=       9.390 * cos(1.32674472283  + 9.40116886120 *T)
    sumR0  +=       9.621 * cos(5.69090299390  + 73.88782669000 *T)
    sumR0  +=      10.843 * cos(1.42812648220  + 268.43697403230 *T)
    sumR0  +=      10.874 * cos(2.61361741990  + 446.31134681820 *T)
    sumR0  +=       9.681 * cos(4.27051176079  + 282.66406803390 *T)
    sumR0  +=      10.770 * cos(0.19304986906  + 463.50738623640 *T)
    sumR0  +=       8.223 * cos(3.30114329151  + 172.19711438360 *T)
    sumR0  +=      10.216 * cos(5.04428111805  + 241.87090352020 *T)
    sumR0  +=       8.890 * cos(2.53320903906  + 271.40591944890 *T)
    sumR0  +=      10.621 * cos(4.39013117792  + 63.84777288800 *T)
    sumR0  +=       9.085 * cos(0.99085954236  + 6.90109867970 *T)
    sumR0  +=      10.414 * cos(4.96367476854  + 97.67614824720 *T)
    sumR0  +=       8.382 * cos(5.03764591595  + 370.93904741990 *T)
    sumR0  +=       8.791 * cos(3.05426995163  + 291.26208774300 *T)
    sumR0  +=       8.629 * cos(0.14674938050  + 262.47783133970 *T)
    sumR0  +=       9.673 * cos(4.25597330570  + 602.98809095360 *T)
    sumR0  +=       7.939 * cos(5.71230451368  + 541.53981451060 *T)
    sumR0  +=       7.981 * cos(2.35900017752  + 196.62432088160 *T)
    sumR0  +=       7.941 * cos(0.88239951788  + 154.97982310600 *T)
    sumR0  +=       8.212 * cos(0.00845991197  + 76.42612997000 *T)
    sumR0  +=      10.135 * cos(1.90258069764  + 91.45637312370 *T)
    sumR0  +=       8.948 * cos(4.31891786278  + 469.13646052890 *T)
    sumR0  +=       9.906 * cos(0.61122653279  + 308.31511480930 *T)
    sumR0  +=       8.257 * cos(4.61012292958  + 69.67378913660 *T)
    sumR0  +=      10.291 * cos(3.58217488981  + 842.15068148810 *T)
    sumR0  +=      10.672 * cos(2.28920805112  + 194.28851491140 *T)
    sumR0  +=       9.024 * cos(0.70282370018  + 685.47393735270 *T)
    sumR0  +=       7.552 * cos(4.86800510978  + 1097.09427470170 *T)
    sumR0  +=       9.496 * cos(1.06662350720  + 93.79217909390 *T)
    sumR0  +=       8.413 * cos(3.15290837718  + 32.71640966640 *T)
    sumR0  +=       8.914 * cos(5.03579282562  + 450.45594840240 *T)
    sumR0  +=       8.008 * cos(4.33420849497  + 302.09533968580 *T)
    sumR0  +=      10.397 * cos(4.90564475822  + 829.62050851590 *T)
    sumR0  +=       7.401 * cos(5.67595616187  + 337.80194662820 *T)
    sumR0  +=       7.419 * cos(3.04672439120  + 7.42236354150 *T)
    sumR0  +=       7.936 * cos(4.37116642726  + 464.47059408290 *T)
    sumR0  +=       9.731 * cos(5.70393303400  + 98.35747180340 *T)
    sumR0  +=       9.287 * cos(4.16913084905  + 15.49911838880 *T)
    sumR0  +=      10.072 * cos(1.18963356664  + 621.73803904930 *T)
    sumR0  +=       9.867 * cos(1.34469368253  + 142.97091499560 *T)
    sumR0  +=       8.913 * cos(1.33679256423  + 0.26063243090 *T)
    sumR0  +=       8.734 * cos(2.03651443558  + 149.40313844020 *T)
    sumR0  +=       7.745 * cos(0.14834225031  + 1404.08497549710 *T)
    sumR0  +=       9.037 * cos(3.53203542312  + 636.66770846650 *T)
    sumR0  +=       7.707 * cos(1.87083579542  + 31.65262808340 *T)
    sumR0  +=       8.566 * cos(2.48280357910  + 497.44763618020 *T)
    sumR0  +=      10.074 * cos(2.59066128203  + 711.44930703380 *T)
    sumR0  +=       8.726 * cos(5.88905195509  + 82.20396210880 *T)
    sumR0  +=       9.218 * cos(1.10861658464  + 412.58354519550 *T)
    sumR0  +=       9.483 * cos(6.02304095716  + 916.93228005540 *T)
    sumR0  +=       7.012 * cos(6.27635752748  + 376.19561469690 *T)
    sumR0  +=       9.867 * cos(1.93140204720  + 62.77269045690 *T)
    sumR0  +=       7.121 * cos(1.54192963162  + 679.25416222920 *T)
    sumR0  +=       8.440 * cos(5.06400284478  + 1.37259812370 *T)
    sumR0  +=       8.915 * cos(2.39052372377  + 76.15419669100 *T)
    sumR0  +=       9.637 * cos(3.78128664809  + 838.21852822500 *T)
    sumR0  +=       7.059 * cos(5.91499106809  + 74.14845912090 *T)
    sumR0  +=       9.570 * cos(1.97721363991  + 703.63318461740 *T)
    sumR0  +=       7.516 * cos(4.87738017916  + 310.17209453310 *T)
    sumR0  +=       7.059 * cos(1.94098053303  + 75.41473801370 *T)
    sumR0  +=       8.552 * cos(6.14581140636  + 17.63798240290 *T)
    sumR0  +=       6.859 * cos(5.11679089849  + 107.49800823370 *T)
    sumR0  +=       7.148 * cos(1.73466140387  + 1190.78588005910 *T)
    sumR0  +=       9.112 * cos(0.88368663290  + 362.86229257260 *T)
    sumR0  +=       9.300 * cos(1.44257902224  + 763.43692965770 *T)
    sumR0  +=       9.073 * cos(4.31998777457  + 16.15350969460 *T)
    sumR0  +=       9.126 * cos(3.74417347717  + 4.66586644600 *T)
    sumR0  +=       7.869 * cos(4.65596954763  + 232.42155054920 *T)
    sumR0  +=       9.168 * cos(3.25096522859  + 155.50108796780 *T)
    sumR0  +=       7.624 * cos(0.88232424215  + 459.36278465220 *T)
    sumR0  +=       8.200 * cos(1.51866334747  + 10063.72234907640 *T)
    sumR0  +=       8.579 * cos(2.37726234500  + 75.67537044460 *T)
    sumR0  +=       7.595 * cos(2.63499505823  + 657.16276170140 *T)
    sumR0  +=       8.131 * cos(6.15861249482  + 745.27768239300 *T)
    sumR0  +=       6.398 * cos(0.61376490225  + 73.24894174920 *T)
    sumR0  +=       7.710 * cos(4.02552779925  + 4.73530241520 *T)
    sumR0  +=       6.380 * cos(3.20688120531  + 73.34530996880 *T)
    sumR0  +=       7.505 * cos(0.67397826037  + 228.79821382680 *T)
    sumR0  +=       7.129 * cos(1.33525552417  + 236.87496867410 *T)
    sumR0  +=       7.444 * cos(3.05581518163  + 171.65459766240 *T)
    sumR0  +=       6.361 * cos(6.05108999867  + 95.97922721780 *T)
    sumR0  +=       7.086 * cos(4.88497877319  + 6531.66165626500 *T)
    sumR0  +=       6.527 * cos(4.01380149030  + 118.87377689770 *T)
    sumR0  +=       7.984 * cos(1.70695215254  + 104.52906281710 *T)
    sumR0  +=       6.239 * cos(1.08160874504  + 143.93412284210 *T)
    sumR0  +=       7.200 * cos(1.19830150903  + 1617.38407093510 *T)
    sumR0  +=       6.390 * cos(4.83649966441  + 1072.71525231350 *T)
    sumR0  +=       7.857 * cos(3.06062400692  + 341.99473232220 *T)
    sumR0  +=       6.158 * cos(0.04340975323  + 627.36711334180 *T)
    sumR0  +=       8.664 * cos(5.60425824325  + 2810.92146160520 *T)
    sumR0  +=       6.147 * cos(4.16482048084  + 1300.99220127850 *T)
    sumR0  +=       8.148 * cos(3.11700641910  + 10213.28554621100 *T)
    sumR0  +=       8.603 * cos(3.80404544682  + 25558.21217647960 *T)
    sumR0  +=       7.586 * cos(2.86885781812  + 406.10313764110 *T)
    sumR0  +=       7.925 * cos(3.11650504000  + 81.68269724700 *T)
    sumR0  +=       7.063 * cos(5.35078594952  + 73.03649342810 *T)
    sumR0  +=       6.265 * cos(3.27996815317  + 116.53797092750 *T)
    sumR0  +=       7.345 * cos(0.45699353973  + 192.80404220310 *T)
    sumR0  +=       6.878 * cos(1.29690429239  + 22.63391724900 *T)
    sumR0  +=       6.102 * cos(2.06174377751  + 73.97844941520 *T)
    sumR0  +=       7.867 * cos(1.45455437627  + 131.92521473170 *T)
    sumR0  +=       6.837 * cos(0.07237438837  + 90.28071695610 *T)
    sumR0  +=       7.740 * cos(2.87078307084  + 79.44746501330 *T)
    sumR0  +=       6.744 * cos(0.01897075429  + 572.22923474750 *T)
    sumR0  +=       6.034 * cos(0.96635268225  + 476.31944349890 *T)
    sumR0  +=       8.276 * cos(2.40596529645  + 674.80074410430 *T)
    sumR0  +=       6.002 * cos(3.24986200464  + 76.78733613740 *T)
    sumR0  +=       6.863 * cos(1.24850658686  + 400.16524680790 *T)
    sumR0  +=       6.375 * cos(1.37050567525  + 75.15410558280 *T)
    sumR0  +=       7.445 * cos(5.47946546419  + 50.66320861000 *T)
    sumR0  +=       5.889 * cos(5.83364715391  + 164.12035953630 *T)
    sumR0  +=       7.190 * cos(3.21396813566  + 71.15826184490 *T)
    sumR0  +=       6.305 * cos(1.65443603478  + 70.04629615210 *T)
    sumR0  +=       6.628 * cos(1.99043744200  + 1.59634729290 *T)
    sumR0  +=       6.081 * cos(1.26599890530  + 61.44827644300 *T)
    sumR0  +=       7.795 * cos(0.70956881527  + 44.07092647100 *T)
    sumR0  +=       7.784 * cos(5.47416712189  + 416.77633088950 *T)
    sumR0  +=       5.727 * cos(2.39389303582  + 20277.00789528740 *T)
    sumR0  +=       6.363 * cos(1.96472326808  + 288.73508531110 *T)
    sumR0  +=       5.724 * cos(4.30532783970  + 86.63044798330 *T)
    sumR0  +=       7.348 * cos(5.28008775997  + 285.63301345050 *T)
    sumR0  +=       5.693 * cos(1.13590287643  + 445.34813897170 *T)
    sumR0  +=       5.799 * cos(2.43420435064  + 180.79513409270 *T)
    sumR0  +=       6.927 * cos(4.19297070447  + 525.23754696970 *T)
    sumR0  +=       6.143 * cos(2.23902465258  + 1310.39337013970 *T)
    sumR0  +=       6.680 * cos(5.36285833774  + 452.46168597250 *T)
    sumR0  +=       5.709 * cos(3.65303501856  + 442.37919355510 *T)
    sumR0  +=       6.238 * cos(4.60538870228  + 137.55428902420 *T)
    sumR0  +=       5.932 * cos(3.22993254909  + 73.45718455340 *T)
    sumR0  +=       7.734 * cos(5.72805659987  + 154.29849954980 *T)
    sumR0  +=       5.782 * cos(2.17577477697  + 2.28762186040 *T)
    sumR0  +=       6.016 * cos(3.46585546590  + 346.39996633730 *T)
    sumR0  +=       7.517 * cos(5.63641104589  + 549.72844394250 *T)
    sumR0  +=       6.685 * cos(4.84971092735  + 148.81243760920 *T)
    sumR0  +=       5.797 * cos(0.51568102762  + 149.30256470370 *T)
    sumR0  +=       7.739 * cos(3.23104533276  + 589.49471013490 *T)
    sumR0  +=       7.590 * cos(2.49504939023  + 321.76031151820 *T)
    sumR0  +=       6.706 * cos(5.23367324742  + 769.81676347560 *T)
    sumR0  +=       7.210 * cos(2.56243122515  + 375.67434983510 *T)
    sumR0  +=       5.491 * cos(1.84139760824  + 375.39246554480 *T)
    sumR0  +=       6.592 * cos(3.39703193659  + 389.68899551560 *T)
    sumR0  +=       6.414 * cos(3.56513278405  + 488.84961647110 *T)
    sumR0  +=       6.376 * cos(0.24081237769  + 881.50755740330 *T)
    sumR0  +=       5.785 * cos(3.49508162978  + 102.52332524700 *T)
    sumR0  +=       6.079 * cos(0.84517404881  + 89.59939339990 *T)
    sumR0  +=       5.799 * cos(2.19388408812  + 8.90683624980 *T)
    sumR0  +=       5.305 * cos(1.97785611200  + 150.31395666000 *T)
    sumR0  +=       5.422 * cos(1.21990761182  + 332.17287233570 *T)
    sumR0  +=       7.235 * cos(2.78896876234  + 748.09786996330 *T)
    sumR0  +=       6.360 * cos(0.36304095923  + 74.40909155180 *T)
    sumR0  +=       5.285 * cos(4.41108441808  + 12.00890811040 *T)
    sumR0  +=       6.602 * cos(3.01699401726  + 442.75170057060 *T)
    sumR0  +=       6.509 * cos(0.49819168917  + 1147.49685088080 *T)
    sumR0  +=       5.585 * cos(6.04761249158  + 172.45774681450 *T)
    sumR0  +=       6.673 * cos(0.77536194908  + 6069.77675455340 *T)
    sumR0  +=       7.125 * cos(0.34356180793  + 511.59540830470 *T)
    sumR0  +=       5.288 * cos(4.27462942653  + 11.15757484850 *T)
    sumR0  +=       6.767 * cos(2.00969662613  + 105.38039607900 *T)
    sumR0  +=       5.611 * cos(2.40776057824  + 1286.01434775150 *T)
    sumR0  +=       6.456 * cos(4.48699081452  + 31.23193695810 *T)
    sumR0  +=       5.898 * cos(5.26174074234  + 757.80785536520 *T)
    sumR0  +=       5.153 * cos(2.41386832919  + 742.99006053260 *T)
    sumR0  +=       5.087 * cos(6.03592089039  + 980.66817835880 *T)
    sumR0  +=       6.198 * cos(0.83056505252  + 1507.17774971570 *T)
    sumR0  +=       5.425 * cos(1.93107713343  + 40.84134862350 *T)
    sumR0  +=       6.091 * cos(5.18564204379  + 487.10451133190 *T)
    sumR0  +=       5.715 * cos(1.96802719384  + 394.35486196160 *T)
    sumR0  +=       5.509 * cos(1.31275092080  + 883.79517926370 *T)
    sumR0  +=       5.320 * cos(4.22718652038  + 65.38042970610 *T)
    sumR0  +=       5.357 * cos(1.80483136985  + 139.48070471720 *T)
    sumR0  +=       5.204 * cos(3.39869589191  + 1400.15282223400 *T)
    sumR0  +=       6.537 * cos(2.31923989568  + 328.24071907260 *T)
    sumR0  +=       5.041 * cos(2.94673346440  + 361.37781986430 *T)
    sumR0  +=       4.969 * cos(6.27367198215  + 0.16005869440 *T)
    sumR0  +=       5.334 * cos(2.78985718428  + 217.44369702220 *T)
    sumR0  +=       5.654 * cos(0.51056760715  + 285.11174858870 *T)
    sumR0  +=       6.432 * cos(2.21948959433  + 9999.98645077300 *T)
    sumR0  +=       5.768 * cos(5.10735836078  + 216.26804085460 *T)
    sumR0  +=       4.977 * cos(2.62435916254  + 194.17664032680 *T)
    sumR0  +=       6.932 * cos(1.71722863424  + 378.90392768260 *T)
    sumR0  +=       5.187 * cos(3.04429850681  + 1083.76095257740 *T)
    sumR0  +=       5.791 * cos(3.94061476250  + 550.13783421970 *T)
    sumR0  +=       5.816 * cos(2.24843661305  + 230.93707784090 *T)
    sumR0  +=       5.319 * cos(0.06998825350  + 336.83873878170 *T)
    sumR0  +=       5.427 * cos(0.30577275388  + 40.58071619260 *T)
    sumR0  +=       6.469 * cos(3.02579309025  + 298.23262239190 *T)
    sumR0  +=       4.974 * cos(1.21594265105  + 455.06942522170 *T)
    sumR0  +=       4.960 * cos(4.72806210230  + 454.74930783290 *T)
    sumR0  +=       5.619 * cos(2.27500303004  + 227.52618943960 *T)
    sumR0  +=       6.328 * cos(0.97544932086  + 249.94765836750 *T)
    sumR0  +=       5.319 * cos(4.59867974067  + 454.79749194270 *T)
    sumR0  +=       4.791 * cos(4.40360629153  + 853.19638175200 *T)
    sumR0  +=       6.519 * cos(3.03043401282  + 167.72244439930 *T)
    sumR0  +=       5.450 * cos(6.00971547441  + 25.12978191360 *T)
    sumR0  +=       5.094 * cos(3.96693309189  + 1066.49547719000 *T)
    sumR0  +=       5.779 * cos(0.65954416303  + 272.58157561650 *T)
    sumR0  +=       6.491 * cos(4.68529651540  + 312.19908396260 *T)
    sumR0  +=       5.234 * cos(4.34712255335  + 233.53351624200 *T)
    sumR0  +=       5.396 * cos(5.62885554221  + 418.52143602870 *T)
    sumR0  +=       5.048 * cos(2.46802064424  + 987.56927703850 *T)
    sumR0  +=       6.152 * cos(0.79853332272  + 2274.54683263650 *T)
    sumR0  +=       6.506 * cos(1.72915575120  + 125.18417474640 *T)
    sumR0  +=       4.993 * cos(3.75975860404  + 57.25549074900 *T)
    sumR0  +=       6.295 * cos(0.84778953014  + 10.08249241740 *T)
    sumR0  +=       6.251 * cos(4.78782138567  + 270.18207917150 *T)
    sumR0  +=       5.785 * cos(4.31237764709  + 374.49869366750 *T)
    sumR0  +=       5.406 * cos(5.49902863401  + 632.78373931320 *T)
    sumR0  +=       6.224 * cos(0.12733845417  + 149.04193227280 *T)
    sumR0  +=       4.921 * cos(1.43037646364  + 73.13706716460 *T)
    sumR0  +=       5.076 * cos(1.34845106372  + 455.02124111190 *T)
    sumR0  +=       5.730 * cos(3.30386575867  + 88.27497938600 *T)
    sumR0  +=       4.618 * cos(0.64720625124  + 119.39504175950 *T)
    sumR0  +=       6.213 * cos(2.58827934841  + 544.50875992720 *T)
    sumR0  +=       4.825 * cos(6.08615765986  + 304.12232911530 *T)
    sumR0  +=       4.825 * cos(6.00483903794  + 226.79247625670 *T)
    sumR0  +=       5.528 * cos(5.69752791882  + 548.44091319030 *T)
    sumR0  +=       5.108 * cos(2.74489127167  + 423.62924545940 *T)
    sumR0  +=       5.426 * cos(2.44835106987  + 531.97858695500 *T)
    sumR0  +=       4.573 * cos(1.61098293427  + 357.44566660120 *T)
    sumR0  +=       4.487 * cos(6.09067660554  + 204.70107572890 *T)
    sumR0  +=       5.866 * cos(6.20513223441  + 772.78570889220 *T)
    sumR0  +=       5.334 * cos(2.49860553733  + 1131.19458333990 *T)
    sumR0  +=       5.656 * cos(4.75744184558  + 491.81856188770 *T)
    sumR0  +=       4.441 * cos(0.23590452375  + 35.68535508300 *T)
    sumR0  +=       4.370 * cos(3.81136490830  + 1329.30337692980 *T)
    sumR0  +=       4.406 * cos(3.42865095493  + 144.89733068860 *T)
    sumR0  +=       5.251 * cos(4.72114047741  + 535.32003938710 *T)
    sumR0  +=       5.174 * cos(1.37807596858  + 520.12973753900 *T)
    sumR0  +=       4.331 * cos(2.64717426456  + 1517.26024213310 *T)
    sumR0  +=       4.802 * cos(2.60569463520  + 177.30492381430 *T)
    sumR0  +=       4.368 * cos(3.36272561974  + 1503.24559645260 *T)
    sumR0  +=       4.335 * cos(2.73379207096  + 289.56516671360 *T)
    sumR0  +=       5.198 * cos(0.97116582962  + 128.43500445330 *T)
    sumR0  +=       4.437 * cos(0.56678131875  + 253.45912050530 *T)
    sumR0  +=       5.386 * cos(5.84886051674  + 268.69760646320 *T)
    sumR0  +=       5.376 * cos(1.30096148962  + 436.81380973740 *T)
    sumR0  +=       5.797 * cos(4.33049740199  + 208.84567731310 *T)
    sumR0  +=       4.353 * cos(1.66111524192  + 1261.63532536330 *T)
    sumR0  +=       4.812 * cos(4.95769337401  + 545.47196777370 *T)
    sumR0  +=       5.897 * cos(2.04201205180  + 8.59801970910 *T)
    sumR0  +=       5.032 * cos(2.80550759770  + 360.41461201780 *T)
    sumR0  +=       4.921 * cos(2.55658380096  + 260.36021918500 *T)
    sumR0  +=       4.413 * cos(3.23825819993  + 973.55463135800 *T)
    sumR0  +=       4.506 * cos(0.17509624151  + 380.38840039090 *T)
    sumR0  +=       4.947 * cos(5.50324549675  + 365.90067395840 *T)
    sumR0  +=       4.174 * cos(2.99974290843  + 136.06981631590 *T)
    sumR0  +=       4.553 * cos(2.77416673233  + 147.96684984170 *T)
    sumR0  +=       5.344 * cos(1.81213470593  + 521.09294538550 *T)
    sumR0  +=       5.155 * cos(0.78324341489  + 1670.07426897460 *T)
    sumR0  +=       5.133 * cos(1.37435234967  + 271.61836777000 *T)
    sumR0  +=       5.708 * cos(2.52872222038  + 501.37978944330 *T)
    sumR0  +=       4.933 * cos(1.36454104948  + 238.90195810360 *T)
    sumR0  +=       4.973 * cos(1.92960964594  + 535.91074021810 *T)
    sumR0  +=       4.935 * cos(5.04375067678  + 697.80716836880 *T)
    sumR0  +=       4.129 * cos(1.50064332826  + 71.86083726050 *T)
    sumR0  +=       5.207 * cos(4.01877367340  + 92.41958097020 *T)
    sumR0  +=       4.587 * cos(4.78553156868  + 95.93104310800 *T)
    sumR0  +=       5.050 * cos(5.41251268131  + 758.77106321170 *T)
    sumR0  +=       5.012 * cos(4.50266403888  + 635.96513305090 *T)
    sumR0  +=       4.248 * cos(0.65406962267  + 920.86443331850 *T)
    sumR0  +=       5.150 * cos(1.13490701556  + 310.97524368520 *T)
    sumR0  +=       4.146 * cos(5.54040372231  + 1048.33622992530 *T)
    sumR0  +=       4.256 * cos(4.20942901957  + 25.27279426550 *T)
    sumR0  +=       4.569 * cos(5.19758291396  + 10.29494073850 *T)
    sumR0  +=       4.259 * cos(5.53202386861  + 184.98791978670 *T)
    sumR0  +=       4.315 * cos(2.80569687202  + 213.95348674380 *T)
    sumR0  +=       4.604 * cos(2.51643176466  + 962.50893109410 *T)
    sumR0  +=       5.105 * cos(1.26007002216  + 971.10695080320 *T)
    sumR0  +=       4.500 * cos(6.15796742231  + 1052.26838318840 *T)
    sumR0  +=       4.095 * cos(0.63467124507  + 1321.43907040360 *T)
    sumR0  +=       3.968 * cos(0.07377679014  + 77.70235987410 *T)
    sumR0  +=       4.617 * cos(2.77367751889  + 406.95447090300 *T)
    sumR0  +=       4.899 * cos(4.65767840428  + 305.60680182360 *T)
    sumR0  +=       3.900 * cos(1.66467970991  + 945.24345570670 *T)
    sumR0  +=       4.133 * cos(3.76396043787  + 263.02034806090 *T)
    sumR0  +=       3.899 * cos(4.28677450975  + 224.23292111730 *T)
    sumR0  +=       5.030 * cos(6.24112139981  + 1162.47470440780 *T)
    sumR0  +=       5.024 * cos(0.39738855487  + 968.13800538660 *T)
    sumR0  +=       4.894 * cos(2.50422546622  + 355.74874557180 *T)
    sumR0  +=       4.283 * cos(3.15267059582  + 846.08283475120 *T)
    sumR0  +=       3.941 * cos(0.04342429962  + 1235.61177157240 *T)
    sumR0  +=       4.077 * cos(5.68854469970  + 695.55642977010 *T)
    sumR0  +=       3.842 * cos(0.37429373422  + 774.48262992160 *T)
    sumR0  +=       4.674 * cos(0.08112657673  + 1366.21257229020 *T)
    sumR0  +=       4.671 * cos(4.38923533828  + 117.36805233000 *T)
    sumR0  +=       5.313 * cos(4.38472090135  + 689.61853893690 *T)
    sumR0  +=       3.787 * cos(1.41443617212  + 48.91810347080 *T)
    sumR0  +=       4.236 * cos(1.60316940746  + 367.59759498780 *T)
    sumR0  +=       4.569 * cos(2.88138923862  + 551.10104206620 *T)
    sumR0  +=       4.636 * cos(5.93442268083  + 148.19059901090 *T)
    sumR0  +=       5.128 * cos(0.17600225009  + 433.71173787680 *T)
    sumR0  +=       4.264 * cos(2.08657038625  + 325.38364824060 *T)
    sumR0  +=       3.885 * cos(5.85359840623  + 450.71658083330 *T)
    sumR0  +=       4.753 * cos(2.58442943928  + 358.40887444770 *T)
    sumR0  +=       4.226 * cos(6.24596640453  + 448.97147569410 *T)
    sumR0  +=       3.776 * cos(1.78756451192  + 71.70077856610 *T)
    sumR0  +=       4.912 * cos(4.45665056284  + 51.88704888740 *T)
    sumR0  +=       3.854 * cos(2.72138633161  + 151.85081899500 *T)
    sumR0  +=       4.561 * cos(0.07201979569  + 2349.32843120380 *T)
    sumR0  +=       4.291 * cos(5.39929339966  + 523.75307426140 *T)
    sumR0  +=       4.143 * cos(0.17158866270  + 735.87651353180 *T)
    sumR0  +=       3.806 * cos(1.44358694049  + 138.62937145530 *T)
    sumR0  +=       3.654 * cos(2.41520715554  + 348.84764689210 *T)
    sumR0  +=       3.728 * cos(1.69745141654  + 984.71220620650 *T)
    sumR0  +=       4.176 * cos(4.01139155515  + 195.77298761970 *T)
    sumR0  +=       4.986 * cos(1.03562905920  + 224.45667028650 *T)
    sumR0  +=       4.031 * cos(0.92145122185  + 76.00543884470 *T)
    sumR0  +=       4.098 * cos(3.51214223942  + 72.49397670690 *T)
    sumR0  +=       3.812 * cos(4.41246815759  + 1511.32235129990 *T)
    sumR0  +=       4.098 * cos(2.39702785276  + 239.16259053450 *T)
    sumR0  +=       4.894 * cos(5.26621064696  + 601.50361824530 *T)
    sumR0  +=       4.459 * cos(5.76440378473  + 836.52160719560 *T)
    sumR0  +=       4.373 * cos(4.08948598951  + 75.43598987310 *T)
    sumR0  +=       4.363 * cos(6.01127167247  + 421.18156490460 *T)
    sumR0  +=       4.414 * cos(0.35830661600  + 168.31314523030 *T)
    sumR0  +=       3.700 * cos(2.04103813925  + 63.21463344160 *T)
    sumR0  +=       4.648 * cos(2.07482117651  + 1106.65550225730 *T)
    sumR0  +=       4.304 * cos(3.03122452327  + 1109.62444767390 *T)
    sumR0  +=       3.618 * cos(4.84400847177  + 893.35640681930 *T)
    sumR0  +=       4.023 * cos(0.59685650686  + 91.24392480260 *T)
    sumR0  +=       4.937 * cos(5.49417275871  + 976.73602509570 *T)
    sumR0  +=       4.373 * cos(3.76648561161  + 74.12720726150 *T)
    sumR0  +=       4.310 * cos(3.72983787822  + 673.31627139600 *T)
    sumR0  +=       4.680 * cos(3.89631254150  + 163.57784281510 *T)
    sumR0  +=       4.172 * cos(0.75349427039  + 1500.06420271490 *T)
    sumR0  +=       4.492 * cos(1.87283145714  + 141.17762574660 *T)
    sumR0  +=       4.174 * cos(5.82910805335  + 346.44815044710 *T)
    sumR0  +=       3.775 * cos(0.83052387256  + 827.92358748650 *T)
    sumR0  +=       3.481 * cos(2.70828672792  + 818.57480825200 *T)
    sumR0  +=       4.853 * cos(0.95917381603  + 58.17051448570 *T)
    sumR0  +=       4.021 * cos(3.11274455034  + 377.41945497430 *T)
    sumR0  +=       4.409 * cos(0.16607520728  + 630.33605875840 *T)
    sumR0  +=       3.597 * cos(1.02560564654  + 515.46387109300 *T)
    sumR0  +=       3.444 * cos(1.38488805947  + 117.31986822020 *T)
    sumR0  +=       4.021 * cos(5.68447974866  + 3.49021027840 *T)
    sumR0  +=       4.244 * cos(3.75845344717  + 733.42883297700 *T)
    sumR0  +=       3.933 * cos(4.55157642432  + 240.12579838100 *T)
    sumR0  +=       4.421 * cos(1.51263319894  + 1610.27052393430 *T)
    sumR0  +=       3.390 * cos(4.00215380112  + 74.99404688840 *T)
    sumR0  +=       4.556 * cos(3.41531360529  + 1140.38330388000 *T)
    sumR0  +=       4.171 * cos(0.76417016678  + 623.43496007870 *T)
    sumR0  +=       3.441 * cos(2.56450835637  + 14.81779483260 *T)
    sumR0  +=       3.587 * cos(4.10186965494  + 343.21857259960 *T)
    sumR0  +=       3.997 * cos(5.74857613262  + 6212.22640468720 *T)
    sumR0  +=       4.215 * cos(0.84469743228  + 176.65053250850 *T)
    sumR0  +=       4.098 * cos(1.41920746453  + 559.69906177530 *T)
    sumR0  +=       4.553 * cos(1.09016692751  + 561.88610989920 *T)
    sumR0  +=       3.493 * cos(0.11837510368  + 1031.87390369000 *T)
    sumR0  +=       3.596 * cos(5.33968666729  + 394.94556279260 *T)
    sumR0  +=       3.501 * cos(5.53309359866  + 594.91133610630 *T)
    sumR0  +=       3.564 * cos(1.57868308864  + 354.99798604640 *T)
    sumR0  +=       4.279 * cos(2.35436288262  + 562.66800719190 *T)
    sumR0  +=       3.398 * cos(1.62086348540  + 941.31130244360 *T)
    sumR0  +=       3.803 * cos(1.78948693511  + 251.17149864490 *T)
    sumR0  +=       3.299 * cos(5.13631478768  + 477.00076705510 *T)
    sumR0  +=       3.422 * cos(3.55674695079  + 256.42806592190 *T)
    sumR0  +=       3.802 * cos(4.09287840097  + 268.95823889410 *T)
    sumR0  +=       3.818 * cos(4.64162443046  + 71.92452773530 *T)
    sumR0  +=       4.623 * cos(2.72836206211  + 6244.94281435360 *T)
    sumR0  +=       3.466 * cos(1.91387688001  + 58.73996345730 *T)
    sumR0  +=       3.435 * cos(2.96178782926  + 995.64603188580 *T)
    sumR0  +=       3.626 * cos(5.35614681493  + 57.51612317990 *T)
    sumR0  +=       4.512 * cos(1.18350543284  + 170.76082578510 *T)
    sumR0  +=       4.378 * cos(0.27346037700  + 469.72716135990 *T)
    sumR0  +=       3.323 * cos(4.51516827363  + 454.86118241750 *T)
    sumR0  +=       3.320 * cos(1.42938752044  + 454.95755063710 *T)
    sumR0  +=       4.578 * cos(4.74980514730  + 731.94436026870 *T)
    sumR0  +=       3.249 * cos(0.67719975914  + 74.04788538440 *T)
    sumR0  +=       4.145 * cos(5.58064267022  + 57.79800747020 *T)
    sumR0  +=       3.512 * cos(6.09122971288  + 70.58881287330 *T)
    sumR0  +=       3.814 * cos(2.49565462974  + 6204.36209816100 *T)
    sumR0  +=       3.828 * cos(4.39751907192  + 586.31331639720 *T)
    sumR0  +=       3.639 * cos(4.85097208169  + 138.40562228610 *T)
    sumR0  +=       3.518 * cos(0.52105043625  + 262.80789973980 *T)
    sumR0  +=       3.671 * cos(1.99667387765  + 511.53171782990 *T)
    sumR0  +=       3.215 * cos(0.64628330219  + 887.72733252680 *T)
    sumR0  +=       3.718 * cos(3.27473813045  + 454.64873409640 *T)
    sumR0  +=       3.638 * cos(2.63250736806  + 455.16999895820 *T)
    sumR0  +=       3.772 * cos(0.88810300052  + 10142.43610090680 *T)
    sumR0  +=       3.190 * cos(4.87960158471  + 455.66012605270 *T)
    sumR0  +=       3.669 * cos(4.11456655271  + 409.92341631960 *T)
    sumR0  +=       3.166 * cos(1.24948126394  + 82.64590509350 *T)
    sumR0  +=       3.530 * cos(4.02075420346  + 388.20452280730 *T)
    sumR0  +=       3.163 * cos(1.65294183878  + 765.88461021250 *T)
    sumR0  +=       3.568 * cos(5.95965909592  + 460.84725736050 *T)
    sumR0  +=       3.450 * cos(0.08821515281  + 49.17873590170 *T)
    sumR0  +=       3.270 * cos(1.81146731641  + 34.20088237470 *T)
    sumR0  +=       3.188 * cos(1.74587038709  + 18.91000679010 *T)
    sumR0  +=       3.305 * cos(5.77382040863  + 10.52443540210 *T)
    sumR0  +=       3.345 * cos(4.15802505352  + 1515.77576942480 *T)
    sumR0  +=       3.160 * cos(1.06549289762  + 454.15860700190 *T)
    sumR0  +=       3.525 * cos(2.56091667232  + 78114.14622758799 *T)
    sumR0  +=       4.124 * cos(1.06751791085  + 388.46515523820 *T)
    sumR0  +=       4.016 * cos(2.48751669586  + 531.17543780290 *T)
    sumR0  +=       3.147 * cos(5.05814757549  + 1521.40484371730 *T)
    sumR0  +=       4.180 * cos(1.16833674781  + 514.56435372130 *T)
    sumR0  +=       3.212 * cos(3.18682610058  + 1512.80682400820 *T)
    sumR0  +=       3.486 * cos(0.22630227172  + 36.53668834490 *T)
    sumR0  +=       4.211 * cos(3.21876950029  + 761.74000862830 *T)
    sumR0  +=       3.485 * cos(3.26177495276  + 36.76043751410 *T)
    sumR0  +=       3.506 * cos(6.26633354904  + 545.27502581760 *T)
    sumR0  +=       3.733 * cos(2.55776517455  + 279.48267429620 *T)
    sumR0  +=       4.227 * cos(1.41381723926  + 41.75637236020 *T)
    sumR0  +=       3.187 * cos(1.62296832026  + 138.46931276090 *T)
    sumR0  +=       3.934 * cos(2.86731965547  + 832.06818907070 *T)
    sumR0  +=       3.684 * cos(3.94174060300  + 179.31066138440 *T)
    sumR0  +=       3.115 * cos(5.67364420834  + 73.55775828990 *T)
    sumR0  +=       3.663 * cos(1.80556740809  + 31.54075349880 *T)
    sumR0  +=       3.171 * cos(4.85529878165  + 138.56568098050 *T)
    sumR0  +=       3.777 * cos(5.94890597104  + 873.17017012510 *T)
    sumR0  +=       4.357 * cos(4.15105623366  + 10175.15251057320 *T)
    sumR0  +=       3.951 * cos(1.60185888278  + 576.16138801060 *T)
    sumR0  +=       3.525 * cos(4.80316435970  + 429.77958461370 *T)
    sumR0  +=       3.330 * cos(5.62171319933  + 1116.00428149180 *T)
    sumR0  +=       3.943 * cos(4.62641543020  + 898.77303279070 *T)
    sumR0  +=       3.382 * cos(1.45717307307  + 5983.94945572220 *T)
    sumR0  +=       3.215 * cos(3.73878297941  + 335.77495719870 *T)
    sumR0  +=       3.423 * cos(5.07987216951  + 143.34342201110 *T)
    sumR0  +=       4.161 * cos(5.39091883238  + 1363.24362687360 *T)
    sumR0  +=       3.457 * cos(3.94796907904  + 444.82687410990 *T)
    sumR0  +=       3.593 * cos(3.91831549069  + 10134.57179438060 *T)
    sumR0  +=       3.666 * cos(4.22620338990  + 36.17548217750 *T)
    sumR0  +=       3.580 * cos(3.43130119859  + 912.78767847120 *T)
    sumR0  +=       3.700 * cos(0.67445695843  + 73.93026530540 *T)
    sumR0  +=       3.679 * cos(1.10949079061  + 686.95841006100 *T)
    sumR0  +=       3.547 * cos(5.63096398237  + 440.89472084680 *T)
    sumR0  +=       3.622 * cos(1.10742531477  + 2250.16781024830 *T)
    sumR0  +=       3.562 * cos(3.80604468765  + 1525.33699698040 *T)
    sumR0  +=       3.330 * cos(1.76480149289  + 78.97438426130 *T)
    sumR0  +=       3.738 * cos(1.76754910180  + 384.58118608490 *T)
    sumR0  +=       3.986 * cos(1.06842874470  + 743.79320968470 *T)
    sumR0  +=       3.032 * cos(5.77852412826  + 612.17681149370 *T)
    sumR0  +=       3.008 * cos(0.64086342534  + 210.85141488320 *T)
    sumR0  +=       3.334 * cos(4.81681647959  + 597.35901666110 *T)
    sumR0  +=       3.141 * cos(3.11768616608  + 6607.92772754060 *T)
    sumR0  +=       3.022 * cos(2.08709314702  + 34.53095077480 *T)
    sumR0  +=       3.226 * cos(3.19780030434  + 377.68008740520 *T)
    sumR0  +=       4.065 * cos(3.53637930424  + 402.21916848780 *T)
    sumR0  +=       4.138 * cos(0.28927701421  + 517.16079212240 *T)
    sumR0  +=       3.697 * cos(0.89932694516  + 75.63293182920 *T)
    sumR0  +=       3.918 * cos(5.73859894835  + 94.42531854030 *T)
    sumR0  +=       3.374 * cos(0.40974405580  + 677.76968952090 *T)
    sumR0  +=       3.194 * cos(4.56998602897  + 1385.17496870700 *T)
    sumR0  +=       3.180 * cos(1.03370427552  + 885.43971066640 *T)
    sumR0  +=       3.235 * cos(5.09681747179  + 464.99185894470 *T)
    sumR0  +=       3.798 * cos(5.76464888795  + 586.37700687200 *T)
    sumR0  +=       3.173 * cos(5.68964342749  + 4.19278569400 *T)
    sumR0  +=       3.181 * cos(2.87968862974  + 9914.15915194180 *T)
    sumR0  +=       3.355 * cos(5.45857968674  + 73.08467753790 *T)
    sumR0  +=       2.899 * cos(5.12928266291  + 448.31708438830 *T)
    sumR0  +=       3.706 * cos(2.47342147635  + 64.25716316520 *T)
    sumR0  +=       3.796 * cos(4.29502458131  + 164.54105066160 *T)
    sumR0  +=       3.534 * cos(4.42464754991  + 46.47042291600 *T)
    sumR0  +=       3.488 * cos(5.62713714766  + 3189.56475685690 *T)
    sumR0  +=       2.940 * cos(4.33606107945  + 78263.70942472259 *T)
    sumR0  +=       3.309 * cos(3.10770680369  + 519.60847267720 *T)
    sumR0  +=       4.030 * cos(4.41794838679  + 772.58876693610 *T)
    sumR0  +=       2.867 * cos(5.17129632099  + 346.18751801620 *T)
    sumR0  +=       3.842 * cos(1.87994191354  + 299.71709510020 *T)
    sumR0  +=       3.846 * cos(5.38315286213  + 980.14691349700 *T)
    sumR0  +=       3.724 * cos(4.94511644698  + 984.48845703730 *T)
    sumR0  +=       2.814 * cos(2.17260061398  + 191.31956949480 *T)
    sumR0  +=       3.392 * cos(3.08552116090  + 245.49424024260 *T)
    sumR0  +=       2.923 * cos(0.37358823115  + 6.48040755440 *T)
    sumR0  +=       3.025 * cos(1.19297242418  + 104.83787935780 *T)
    sumR0  +=       3.470 * cos(1.89084704021  + 44.61344319220 *T)
    sumR0  +=       2.931 * cos(0.78809830626  + 540.05534180230 *T)
    sumR0  +=       3.707 * cos(4.14868763219  + 6136.48159827340 *T)
    sumR0  +=       3.931 * cos(5.52289695589  + 6171.64568849460 *T)
    sumR0  +=       3.056 * cos(2.97000936733  + 250.60204967330 *T)
    sumR0  +=       3.117 * cos(0.76332399369  + 229.45260513260 *T)
    sumR0  +=       3.091 * cos(4.91941279780  + 221.16340196420 *T)
    sumR0  +=       3.378 * cos(3.82658652472  + 25936.85547173129 *T)
    sumR0  +=       3.041 * cos(2.14503983522  + 6604.95878212400 *T)
    sumR0  +=       2.865 * cos(0.53734608663  + 273.85360000370 *T)
    sumR0  +=       2.845 * cos(0.34922064899  + 85.93917341580 *T)
    sumR0  +=       3.657 * cos(6.02755271763  + 340.88276662940 *T)
    sumR0  +=       2.818 * cos(4.44508352472  + 369.34270012700 *T)
    sumR0  +=       2.861 * cos(2.19284349075  + 295.19424100610 *T)
    sumR0  +=       2.865 * cos(3.21935127992  + 3.82027867850 *T)
    sumR0  +=       2.797 * cos(5.97725967979  + 2014.02655475710 *T)
    sumR0  +=       3.280 * cos(0.74832416123  + 422.71422172270 *T)
    sumR0  +=       3.509 * cos(1.92501559437  + 343.47920503050 *T)
    sumR0  +=       2.962 * cos(2.29867992492  + 661.15860543930 *T)
    sumR0  +=       3.485 * cos(4.53205302380  + 676.28521681260 *T)
    sumR0  +=       2.836 * cos(1.20779683660  + 1119.18567522950 *T)
    sumR0  +=       3.603 * cos(2.72183511139  + 508.62646288810 *T)
    sumR0  +=       3.620 * cos(5.57691156197  + 10066.69129449300 *T)
    sumR0  +=       2.731 * cos(0.96663411770  + 582.38116313410 *T)
    sumR0  +=       3.306 * cos(6.20620840278  + 11.56696512570 *T)
    sumR0  +=       3.279 * cos(6.13563821647  + 276.07178589490 *T)
    sumR0  +=       2.747 * cos(4.57476263020  + 226.63241756230 *T)
    sumR0  +=       2.825 * cos(1.24120423378  + 989.05374974680 *T)
    sumR0  +=       2.707 * cos(0.46257342768  + 1458.47209456600 *T)
    sumR0  +=       2.755 * cos(5.43548338507  + 246.97871295090 *T)
    sumR0  +=       3.338 * cos(3.98641322371  + 488.58898404020 *T)
    sumR0  +=       2.960 * cos(5.31788128818  + 1467.07011427510 *T)
    sumR0  +=       3.062 * cos(1.93542241882  + 987.78172535960 *T)
    sumR0  +=       3.504 * cos(4.10521239427  + 6280.10690457480 *T)
    sumR0  +=       3.584 * cos(3.12196206517  + 108.72184851110 *T)
    sumR0  +=       2.975 * cos(0.13746189123  + 9987.45627780080 *T)
    sumR0  +=       2.634 * cos(5.61201014857  + 412.37109687440 *T)
    sumR0  +=       3.702 * cos(0.66231252049  + 10101.85538471420 *T)
    sumR0  +=       3.261 * cos(4.39228048501  + 75.37229939830 *T)
    sumR0  +=       2.939 * cos(3.26319979850  + 130.55261660800 *T)
    sumR0  +=       2.742 * cos(6.24317126103  + 447.20511869550 *T)
    sumR0  +=       3.008 * cos(4.24451493185  + 170.17012495410 *T)
    sumR0  +=       3.001 * cos(2.45489658954  + 230.82520325630 *T)
    sumR0  +=       2.722 * cos(4.98348297926  + 754.03576079650 *T)
    sumR0  +=       2.928 * cos(5.73784691627  + 14.66903698630 *T)
    sumR0  +=       2.699 * cos(2.98043546816  + 27.72047482030 *T)
    sumR0  +=       3.678 * cos(3.46436124301  + 26468.03090953420 *T)
    sumR0  +=       3.500 * cos(4.41810854452  + 322.61164478010 *T)
    sumR0  +=       2.943 * cos(3.58544468129  + 12489.88562870720 *T)
    sumR0  +=       2.894 * cos(2.26999120840  + 1615.89959822680 *T)
    sumR0  +=       2.679 * cos(1.32002304425  + 236.19364511790 *T)
    sumR0  +=       2.711 * cos(1.25250599577  + 52250.58788171570 *T)
    sumR0  +=       2.573 * cos(4.41371056719  + 262.05714021440 *T)
    sumR0  +=       3.483 * cos(1.97113718781  + 655.93892142400 *T)
    sumR0  +=       2.600 * cos(3.80107226970  + 70.63699698310 *T)
    sumR0  +=       3.280 * cos(1.75059058234  + 683.98946464440 *T)
    sumR0  +=       3.260 * cos(3.46047977615  + 74.19089773630 *T)
    sumR0  +=       3.097 * cos(5.47452034532  + 302.37722397610 *T)
    sumR0  +=       2.565 * cos(3.47443238116  + 2042.49778910280 *T)
    sumR0  +=       3.158 * cos(4.58160924364  + 12492.85457412380 *T)
    sumR0  +=       3.373 * cos(5.52806001629  + 10210.31660079440 *T)
    sumR0  +=       2.619 * cos(4.43681016753  + 949.17560896980 *T)
    sumR0  +=       3.290 * cos(1.35055242247  + 515.67631941410 *T)
    sumR0  +=       3.180 * cos(4.49938964982  + 694.83822295220 *T)
    sumR0  +=       2.558 * cos(5.07921716363  + 197.79997704920 *T)
    sumR0  +=       2.902 * cos(2.61388996483  + 115.36231475990 *T)
    sumR0  +=       2.586 * cos(4.09771865336  + 1448.91086701040 *T)
    sumR0  +=       3.160 * cos(4.82168102018  + 714.67888488130 *T)
    sumR0  +=       3.016 * cos(3.16691101108  + 385.75684225250 *T)
    sumR0  +=       3.001 * cos(3.24181300229  + 1618.86854364340 *T)
    sumR0  +=       3.319 * cos(5.98200177347  + 533.83556667880 *T)
    sumR0  +=       3.307 * cos(3.31452419197  + 732.97125859000 *T)
    sumR0  +=       2.530 * cos(0.13809025963  + 591.94239068970 *T)
    sumR0  +=       2.975 * cos(5.72770980032  + 1011.42703456490 *T)
    sumR0  +=       2.857 * cos(4.11031053901  + 2267.43328563570 *T)
    sumR0  +=       3.091 * cos(4.87979664891  + 582.64179556500 *T)
    sumR0  +=       3.422 * cos(6.18592254593  + 281.48841186630 *T)
    sumR0  +=       2.501 * cos(0.50645055144  + 29.22619938800 *T)
    sumR0  +=       3.502 * cos(0.17203520151  + 371.52974825090 *T)
    sumR0  +=       2.607 * cos(3.51828908958  + 112.39336934330 *T)
    sumR0  +=       2.547 * cos(4.45612695304  + 901.22071334550 *T)
    sumR0  +=       2.778 * cos(4.97805873371  + 132.57960603750 *T)
    sumR0  +=       2.476 * cos(1.55163657371  + 1234.12729886410 *T)
    sumR0  +=       2.929 * cos(2.34725672182  + 273.15102458810 *T)
    sumR0  +=       2.724 * cos(5.73177362443  + 688.65533109040 *T)
    sumR0  +=       2.667 * cos(6.13733078138  + 161.41204655060 *T)
    sumR0  +=       2.877 * cos(5.47506403197  + 1436.54075273260 *T)
    sumR0  +=       2.779 * cos(1.54196175338  + 680.05731138130 *T)
    sumR0  +=       2.427 * cos(4.64913431310  + 392.65794093220 *T)
    sumR0  +=       3.223 * cos(2.57976952494  + 267.58564077040 *T)
    sumR0  +=       2.750 * cos(1.29662662582  + 108.98248094200 *T)
    sumR0  +=       3.220 * cos(1.06290187940  + 388.72578766910 *T)
    sumR0  +=       3.236 * cos(3.47802643973  + 283.62727588040 *T)
    sumR0  +=       3.027 * cos(5.38233284458  + 44.09217833040 *T)
    sumR0  +=       2.830 * cos(5.70274947128  + 327.43756992050 *T)
    sumR0  +=       2.768 * cos(5.42256168790  + 482.25733433210 *T)
    sumR0  +=       3.230 * cos(5.65661970187  + 134.37289528650 *T)
    sumR0  +=       2.685 * cos(5.03687735302  + 763.22448133660 *T)
    sumR0  +=       2.441 * cos(5.19830386978  + 380.23964254460 *T)
    sumR0  +=       2.608 * cos(3.11386505876  + 578.44900987100 *T)
    sumR0  +=       2.465 * cos(1.49242672830  + 141.69889060840 *T)
    sumR0  +=       2.332 * cos(3.19371804203  + 683.02625679790 *T)
    sumR0  +=       2.307 * cos(4.07685834192  + 78.92620015150 *T)
    sumR0  +=       2.280 * cos(2.52476606956  + 156.67674413540 *T)
    sumR0  +=       2.260 * cos(6.23958645351  + 400.57463708510 *T)
    sumR0  +=       2.301 * cos(2.47762267951  + 107.91869935900 *T)
    sumR0  +=       3.089 * cos(5.69639540598  + 537.39521292640 *T)
    sumR0  +=       3.074 * cos(1.16740737548  + 58.62808887270 *T)
    sumR0  +=       2.846 * cos(0.02430566936  + 563.37058260750 *T)
    sumR0  +=       2.423 * cos(4.24559706691  + 27.74172667970 *T)
    sumR0  +=       2.670 * cos(3.48404619442  + 123.01837848190 *T)
    sumR0  +=       2.228 * cos(3.33159258912  + 1257.70317210020 *T)
    sumR0  +=       2.445 * cos(2.29979337782  + 280.21638747910 *T)
    sumR0  +=       2.260 * cos(3.85026258012  + 753.14198891920 *T)
    sumR0  +=       2.195 * cos(4.94038065567  + 1222.27844944810 *T)
    sumR0  +=       2.579 * cos(1.76841912074  + 710.74673161820 *T)
    sumR0  +=       2.402 * cos(4.47942458428  + 569.04784100980 *T)
    sumR0  +=       2.181 * cos(5.39048760967  + 318.67949151700 *T)
    sumR0  +=       2.317 * cos(6.27946729049  + 493.04240216510 *T)
    sumR0  +=       2.420 * cos(3.66436222896  + 3.62333672240 *T)

    # URANUS Radius
    sumR1  = 0
    sumR1  +=  1479896.370 * cos(3.67205705317  + 74.78159856730 *T)
    sumR1  +=   71212.085 * cos(6.22601006675  + 63.73589830340 *T)
    sumR1  +=   68626.972 * cos(6.13411265052  + 149.56319713460 *T)
    sumR1  +=   20857.262 * cos(5.24625494219  + 11.04570026390 *T)
    sumR1  +=   21468.152 * cos(2.60176704270  + 76.26607127560 *T)
    sumR1  +=   24059.649 * cos(3.14159265359  + 0.00000000000 *T)
    sumR1  +=   11405.346 * cos(0.01848461561  + 70.84944530420 *T)
    sumR1  +=    7496.775 * cos(0.42360033283  + 73.29712585900 *T)
    sumR1  +=    4243.800 * cos(1.41692350371  + 85.82729883120 *T)
    sumR1  +=    3505.936 * cos(2.58354048851  + 138.51749687070 *T)
    sumR1  +=    3228.835 * cos(5.25499602896  + 3.93215326310 *T)
    sumR1  +=    3926.694 * cos(3.15513991323  + 71.81265315070 *T)
    sumR1  +=    3060.010 * cos(0.15321893225  + 1.48447270830 *T)
    sumR1  +=    3578.446 * cos(2.31160668309  + 224.34479570190 *T)
    sumR1  +=    2564.251 * cos(0.98076846352  + 148.07872442630 *T)
    sumR1  +=    2429.445 * cos(3.99440122468  + 52.69019803950 *T)
    sumR1  +=    1644.719 * cos(2.65349313124  + 127.47179660680 *T)
    sumR1  +=    1583.766 * cos(1.43045619196  + 78.71375183040 *T)
    sumR1  +=    1413.112 * cos(4.57461892062  + 202.25339517410 *T)
    sumR1  +=    1489.525 * cos(2.67559167316  + 56.62235130260 *T)
    sumR1  +=    1403.237 * cos(1.36985349744  + 77.75054398390 *T)
    sumR1  +=    1228.220 * cos(1.04703640149  + 62.25142559510 *T)
    sumR1  +=    1508.028 * cos(5.05996325425  + 151.04766984290 *T)
    sumR1  +=     992.085 * cos(2.17168865909  + 65.22037101170 *T)
    sumR1  +=    1032.731 * cos(0.26459059027  + 131.40394986990 *T)
    sumR1  +=     861.867 * cos(5.05530802218  + 351.81659230870 *T)
    sumR1  +=     744.445 * cos(3.07640148939  + 35.16409022120 *T)
    sumR1  +=     604.362 * cos(0.90717667985  + 984.60033162190 *T)
    sumR1  +=     646.851 * cos(4.47290422910  + 70.32818044240 *T)
    sumR1  +=     574.710 * cos(3.23070708457  + 447.79581952650 *T)
    sumR1  +=     687.470 * cos(2.49912565674  + 77.96299230500 *T)
    sumR1  +=     623.602 * cos(0.86253073820  + 9.56122755560 *T)
    sumR1  +=     527.794 * cos(5.15136007084  + 2.96894541660 *T)
    sumR1  +=     561.839 * cos(2.71778158980  + 462.02291352810 *T)
    sumR1  +=     530.364 * cos(5.91655309045  + 213.29909543800 *T)
    sumR1  +=     460.080 * cos(4.22302465979  + 12.53017297220 *T)
    sumR1  +=     494.280 * cos(0.46291078127  + 145.63104387150 *T)
    sumR1  +=     487.336 * cos(0.70614146398  + 380.12776796000 *T)
    sumR1  +=     380.908 * cos(3.85089591694  + 3.18139373770 *T)
    sumR1  +=     444.352 * cos(2.15558291251  + 67.66805156650 *T)
    sumR1  +=     338.800 * cos(2.53820897704  + 18.15924726470 *T)
    sumR1  +=     372.947 * cos(5.05141251694  + 529.69096509460 *T)
    sumR1  +=     348.345 * cos(1.74874852104  + 71.60020482960 *T)
    sumR1  +=     405.881 * cos(1.22961727600  + 22.09140052780 *T)
    sumR1  +=     268.913 * cos(6.24069521597  + 340.77089204480 *T)
    sumR1  +=     255.585 * cos(2.95695013627  + 84.34282612290 *T)
    sumR1  +=     259.465 * cos(3.92053708924  + 59.80374504030 *T)
    sumR1  +=     224.731 * cos(3.90961468562  + 160.60889739850 *T)
    sumR1  +=     221.710 * cos(3.64727173951  + 137.03302416240 *T)
    sumR1  +=     254.591 * cos(3.50411592815  + 38.13303563780 *T)
    sumR1  +=     238.290 * cos(2.04879982674  + 269.92144674060 *T)
    sumR1  +=     272.355 * cos(3.38363105223  + 222.86032299360 *T)
    sumR1  +=     200.648 * cos(1.24861003313  + 69.36497259590 *T)
    sumR1  +=     234.153 * cos(0.27825220612  + 108.46121608020 *T)
    sumR1  +=     188.515 * cos(4.41307507326  + 265.98929347750 *T)
    sumR1  +=     211.691 * cos(0.68027381802  + 111.43016149680 *T)
    sumR1  +=     205.946 * cos(1.53379817229  + 284.14854074220 *T)
    sumR1  +=     196.179 * cos(4.77152996605  + 299.12639426920 *T)
    sumR1  +=     153.102 * cos(5.21761881347  + 209.36694217490 *T)
    sumR1  +=     162.563 * cos(4.34054353610  + 33.67961751290 *T)
    sumR1  +=     150.563 * cos(1.98966326297  + 54.17467074780 *T)
    sumR1  +=     137.012 * cos(0.40323866041  + 195.13984817330 *T)
    sumR1  +=     117.171 * cos(0.39649791652  + 87.31177153950 *T)
    sumR1  +=     127.913 * cos(2.40333045173  + 39.61750834610 *T)
    sumR1  +=     104.218 * cos(2.92152185788  + 134.58534360760 *T)
    sumR1  +=     103.862 * cos(1.81622936156  + 72.33391801250 *T)
    sumR1  +=     105.741 * cos(0.17067407327  + 79.23501669220 *T)
    sumR1  +=     106.419 * cos(0.69799543514  + 2.44768055480 *T)
    sumR1  +=      95.326 * cos(4.02880266738  + 82.85835341460 *T)
    sumR1  +=     104.772 * cos(4.43616414428  + 305.34616939270 *T)
    sumR1  +=      93.825 * cos(5.01823592717  + 51.20572533120 *T)
    sumR1  +=     103.739 * cos(2.57553519741  + 191.20769491020 *T)
    sumR1  +=     106.679 * cos(1.22996874093  + 225.82926841020 *T)
    sumR1  +=      93.452 * cos(3.09274255916  + 77.22927912210 *T)
    sumR1  +=      97.398 * cos(3.81380841075  + 152.53214255120 *T)
    sumR1  +=      84.583 * cos(5.72473747348  + 68.84370773410 *T)
    sumR1  +=      77.395 * cos(0.08281157747  + 45.57665103870 *T)
    sumR1  +=      76.207 * cos(4.20384370842  + 73.81839072080 *T)
    sumR1  +=      86.249 * cos(0.53131085736  + 145.10977900970 *T)
    sumR1  +=      75.795 * cos(3.78559826812  + 75.74480641380 *T)
    sumR1  +=      77.592 * cos(1.63628139623  + 479.28838891550 *T)
    sumR1  +=      84.612 * cos(0.61662456010  + 116.42609634290 *T)
    sumR1  +=     100.209 * cos(4.94084867643  + 120.35824960600 *T)
    sumR1  +=      72.142 * cos(4.30505812564  + 565.11568774670 *T)
    sumR1  +=      70.733 * cos(2.38450718488  + 60.76695288680 *T)
    sumR1  +=      71.585 * cos(3.93906647867  + 153.49535039770 *T)
    sumR1  +=      84.566 * cos(5.56037336584  + 344.70304530790 *T)
    sumR1  +=      63.556 * cos(1.93742986679  + 41.64449777560 *T)
    sumR1  +=      71.619 * cos(3.71213491656  + 408.43894361130 *T)
    sumR1  +=      61.594 * cos(3.90006698249  + 4.45341812490 *T)
    sumR1  +=      64.973 * cos(1.55845503407  + 106.97674337190 *T)
    sumR1  +=      59.913 * cos(0.60110866128  + 74.89347315190 *T)
    sumR1  +=      62.000 * cos(4.39369268007  + 453.42489381900 *T)
    sumR1  +=      63.361 * cos(4.19159979468  + 184.72728735580 *T)
    sumR1  +=      62.301 * cos(3.23773103318  + 422.66603761290 *T)
    sumR1  +=      54.427 * cos(3.72545550857  + 7.11354700080 *T)
    sumR1  +=      52.474 * cos(6.08562717749  + 404.50679034820 *T)
    sumR1  +=      59.073 * cos(1.55568469603  + 456.39383923560 *T)
    sumR1  +=      52.597 * cos(3.50492233970  + 125.98732389850 *T)
    sumR1  +=      52.835 * cos(5.20100035142  + 358.93013930950 *T)
    sumR1  +=      58.123 * cos(5.33480562448  + 220.41264243880 *T)
    sumR1  +=      52.909 * cos(4.44819701196  + 426.59819087600 *T)
    sumR1  +=      50.934 * cos(0.52638534200  + 490.33408917940 *T)
    sumR1  +=      54.968 * cos(1.60146090981  + 14.97785352700 *T)
    sumR1  +=      49.491 * cos(4.25534603275  + 5.41662597140 *T)
    sumR1  +=      51.303 * cos(0.36772379136  + 206.18554843720 *T)
    sumR1  +=      51.821 * cos(1.75832999538  + 8.07675484730 *T)
    sumR1  +=      56.964 * cos(0.84114552694  + 146.59425171800 *T)
    sumR1  +=      49.109 * cos(0.94061875871  + 99.16062095550 *T)
    sumR1  +=      46.361 * cos(5.35115472594  + 152.74459087230 *T)
    sumR1  +=      48.023 * cos(1.97249712347  + 288.08069400530 *T)
    sumR1  +=      43.772 * cos(3.03713403879  + 20.60692781950 *T)
    sumR1  +=      49.493 * cos(5.84619560979  + 112.91463420510 *T)
    sumR1  +=      41.987 * cos(0.04620500196  + 128.95626931510 *T)
    sumR1  +=      48.628 * cos(3.62817742782  + 81.00137369080 *T)
    sumR1  +=      41.472 * cos(2.33730376429  + 277.03499374140 *T)
    sumR1  +=      39.983 * cos(5.09525356576  + 35.42472265210 *T)
    sumR1  +=      41.948 * cos(2.51050760642  + 24.37902238820 *T)
    sumR1  +=      38.325 * cos(3.61946898382  + 173.94221952280 *T)
    sumR1  +=      38.385 * cos(2.06003220130  + 333.65734504400 *T)
    sumR1  +=      42.597 * cos(1.26088737300  + 1514.29129671650 *T)
    sumR1  +=      38.855 * cos(0.74239364306  + 347.88443904560 *T)
    sumR1  +=      38.535 * cos(4.95064283065  + 92.94084583200 *T)
    sumR1  +=      33.234 * cos(1.38358507432  + 74.66972398270 *T)
    sumR1  +=      33.788 * cos(3.68407945156  + 66.91729204110 *T)
    sumR1  +=      38.953 * cos(5.49236040328  + 200.76892246580 *T)
    sumR1  +=      31.850 * cos(0.53990592534  + 203.73786788240 *T)
    sumR1  +=      33.320 * cos(6.26012644668  + 1059.38193018920 *T)
    sumR1  +=      30.806 * cos(2.53797566903  + 977.48678462110 *T)
    sumR1  +=      29.198 * cos(5.43116906000  + 58.10682401090 *T)
    sumR1  +=      30.059 * cos(0.19481555617  + 387.24131496080 *T)
    sumR1  +=      28.997 * cos(3.10546504714  + 991.71387862270 *T)
    sumR1  +=      35.640 * cos(3.72863820177  + 96.87299909510 *T)
    sumR1  +=      27.607 * cos(0.37142052647  + 80.19822453870 *T)
    sumR1  +=      32.492 * cos(4.38403518987  + 221.37585028530 *T)
    sumR1  +=      27.029 * cos(1.35552416596  + 0.96320784650 *T)
    sumR1  +=      31.276 * cos(0.79566430555  + 373.01422095920 *T)
    sumR1  +=      31.122 * cos(2.05381353845  + 230.56457082540 *T)
    sumR1  +=      25.883 * cos(3.46808071409  + 144.14657116320 *T)
    sumR1  +=      30.201 * cos(0.71392007232  + 109.94568878850 *T)
    sumR1  +=      24.688 * cos(3.04162764358  + 14.01464568050 *T)
    sumR1  +=      27.882 * cos(4.76559523368  + 415.55249061210 *T)
    sumR1  +=      25.110 * cos(5.12405829717  + 81.37388070630 *T)
    sumR1  +=      25.582 * cos(2.56904073164  + 522.57741809380 *T)
    sumR1  +=      24.351 * cos(2.20289059750  + 628.85158605010 *T)
    sumR1  +=      25.479 * cos(1.79521877300  + 143.62530630140 *T)
    sumR1  +=      24.182 * cos(5.67160913092  + 443.86366626340 *T)
    sumR1  +=      25.679 * cos(5.43185950751  + 546.95644048200 *T)
    sumR1  +=      24.177 * cos(5.59982039849  + 32.19514480460 *T)
    sumR1  +=      24.428 * cos(3.30271734903  + 617.80588578620 *T)
    sumR1  +=      23.535 * cos(0.65842590604  + 46.20979048510 *T)
    sumR1  +=      22.371 * cos(4.82094751058  + 135.54855145410 *T)
    sumR1  +=      27.179 * cos(2.02720001624  + 536.80451209540 *T)
    sumR1  +=      22.213 * cos(4.61664624220  + 391.17346822390 *T)
    sumR1  +=      21.973 * cos(4.59216260632  + 241.61027108930 *T)
    sumR1  +=      20.813 * cos(0.24392941148  + 465.95506679120 *T)
    sumR1  +=      27.264 * cos(2.15210992383  + 140.00196957900 *T)
    sumR1  +=      21.356 * cos(5.27168432406  + 159.12442469020 *T)
    sumR1  +=      23.632 * cos(4.94972840898  + 561.18353448360 *T)
    sumR1  +=      24.921 * cos(0.54550733267  + 181.75834193920 *T)
    sumR1  +=      23.027 * cos(3.80632203913  + 55.13787859430 *T)
    sumR1  +=      19.799 * cos(1.30259938601  + 518.64526483070 *T)
    sumR1  +=      19.252 * cos(1.31448491434  + 543.02428721890 *T)
    sumR1  +=      19.704 * cos(4.90869636976  + 909.81873305460 *T)
    sumR1  +=      20.801 * cos(0.91178207093  + 76.47851959670 *T)
    sumR1  +=      19.876 * cos(0.66494008343  + 66.70484372000 *T)
    sumR1  +=      18.957 * cos(4.67998817036  + 98.89998852460 *T)
    sumR1  +=      25.913 * cos(4.52903186569  + 454.90936652730 *T)
    sumR1  +=      21.888 * cos(1.23372931740  + 41.10198105440 *T)
    sumR1  +=      18.703 * cos(6.09640927844  + 103.09277421860 *T)
    sumR1  +=      18.207 * cos(0.97283864525  + 55.65914345610 *T)
    sumR1  +=      21.247 * cos(4.19373732137  + 329.72519178090 *T)
    sumR1  +=      19.408 * cos(4.31468230800  + 6.21977512350 *T)
    sumR1  +=      18.497 * cos(5.78624335074  + 142.44965013380 *T)
    sumR1  +=      22.588 * cos(5.84591645052  + 297.64192156090 *T)
    sumR1  +=      16.770 * cos(6.09084656811  + 211.81462272970 *T)
    sumR1  +=      16.432 * cos(2.50008469020  + 61.28821774860 *T)
    sumR1  +=      20.361 * cos(3.16137245375  + 186.21176006410 *T)
    sumR1  +=      15.955 * cos(2.98317221345  + 81.89514556810 *T)
    sumR1  +=      18.953 * cos(6.01226591746  + 155.78297225810 *T)
    sumR1  +=      17.686 * cos(4.82613965176  + 273.10284047830 *T)
    sumR1  +=      15.141 * cos(3.65588411561  + 472.17484191470 *T)
    sumR1  +=      18.440 * cos(3.47582817224  + 36.64856292950 *T)
    sumR1  +=      16.303 * cos(0.13086415177  + 554.06998748280 *T)
    sumR1  +=      18.633 * cos(0.23932740251  + 23.57587323610 *T)
    sumR1  +=      14.352 * cos(2.69389896537  + 70.11573212130 *T)
    sumR1  +=      15.190 * cos(2.43789398875  + 486.40193591630 *T)
    sumR1  +=      14.002 * cos(5.12389205028  + 29.20494752860 *T)
    sumR1  +=      15.758 * cos(4.24947053051  + 146.38180339690 *T)
    sumR1  +=      14.125 * cos(1.55719788547  + 110.20632121940 *T)
    sumR1  +=      17.477 * cos(1.94549668506  + 835.03713448730 *T)
    sumR1  +=      13.691 * cos(1.63831110442  + 92.04707395470 *T)
    sumR1  +=      13.801 * cos(0.13721153975  + 235.39049596580 *T)
    sumR1  +=      13.573 * cos(2.85427895075  + 49.50880430180 *T)
    sumR1  +=      12.563 * cos(3.20921738646  + 100.38446123290 *T)
    sumR1  +=      12.390 * cos(2.88595800082  + 60.55450456570 *T)
    sumR1  +=      14.986 * cos(0.32593957273  + 259.50888592310 *T)
    sumR1  +=      12.922 * cos(2.77565630582  + 105.49227066360 *T)
    sumR1  +=      12.323 * cos(3.36427641421  + 440.68227252570 *T)
    sumR1  +=      15.233 * cos(0.25589845180  + 258.87574647670 *T)
    sumR1  +=      12.106 * cos(0.10857558014  + 157.63995198190 *T)
    sumR1  +=      12.883 * cos(0.30655541587  + 124.29040286910 *T)
    sumR1  +=      10.900 * cos(3.42905554547  + 33.13710079170 *T)
    sumR1  +=      11.206 * cos(4.98840478043  + 604.47256366190 *T)
    sumR1  +=      10.812 * cos(3.86253020441  + 767.36908292080 *T)
    sumR1  +=      11.561 * cos(2.60450144944  + 166.82867252200 *T)
    sumR1  +=      10.200 * cos(5.27810824796  + 264.50482076920 *T)
    sumR1  +=      10.926 * cos(0.64149188846  + 558.00214074590 *T)
    sumR1  +=      12.315 * cos(4.33998516461  + 16.67477455640 *T)
    sumR1  +=       9.946 * cos(0.67298666287  + 31.49256938900 *T)
    sumR1  +=      12.641 * cos(4.83194943583  + 114.39910691340 *T)
    sumR1  +=      10.479 * cos(0.20404797652  + 275.55052103310 *T)
    sumR1  +=      11.291 * cos(0.96120625051  + 373.90799283650 *T)
    sumR1  +=      12.144 * cos(1.91712815063  + 378.64329525170 *T)
    sumR1  +=      12.229 * cos(0.70465454670  + 218.40690486870 *T)
    sumR1  +=      10.753 * cos(5.74480767273  + 88.11492069160 *T)
    sumR1  +=       9.481 * cos(0.65566927406  + 353.30106501700 *T)
    sumR1  +=      11.006 * cos(2.62953946665  + 154.01661525950 *T)
    sumR1  +=       9.113 * cos(2.99457723478  + 681.54178408960 *T)
    sumR1  +=      10.429 * cos(2.33056994007  + 132.88842257820 *T)
    sumR1  +=       9.169 * cos(4.79284571455  + 216.48048917570 *T)
    sumR1  +=       9.341 * cos(0.75923548315  + 129.91947716160 *T)
    sumR1  +=       8.917 * cos(0.78008399009  + 67.35923502580 *T)
    sumR1  +=       8.757 * cos(6.12717748848  + 150.52640498110 *T)
    sumR1  +=       9.637 * cos(2.88664912193  + 67.88049988760 *T)
    sumR1  +=      10.465 * cos(0.36943456465  + 699.70103135430 *T)
    sumR1  +=       9.301 * cos(1.49620591593  + 19.64371997300 *T)
    sumR1  +=       9.367 * cos(5.26481516822  + 80.71948940050 *T)
    sumR1  +=      10.076 * cos(3.56540311122  + 278.51946644970 *T)
    sumR1  +=       9.455 * cos(3.06088968751  + 149.67507171920 *T)
    sumR1  +=       9.168 * cos(3.02528121597  + 162.09337010680 *T)
    sumR1  +=       8.395 * cos(2.18455001650  + 342.25536475310 *T)
    sumR1  +=       9.233 * cos(5.32613442062  + 152.01087768940 *T)
    sumR1  +=       9.786 * cos(2.43713607191  + 75.30286342910 *T)
    sumR1  +=      10.029 * cos(0.81917102953  + 339.28641933650 *T)
    sumR1  +=       9.429 * cos(1.93671715384  + 147.11551657980 *T)
    sumR1  +=       7.861 * cos(4.71717822837  + 106.01353552540 *T)
    sumR1  +=       8.813 * cos(0.01616162513  + 42.58645376270 *T)
    sumR1  +=       7.808 * cos(0.61104170424  + 135.33610313300 *T)
    sumR1  +=       8.193 * cos(2.59644466423  + 469.13646052890 *T)
    sumR1  +=      10.084 * cos(2.58619215129  + 50.40257617910 *T)
    sumR1  +=       8.574 * cos(5.69115937472  + 760.25553592000 *T)
    sumR1  +=       7.525 * cos(2.64764195045  + 5.93789083320 *T)
    sumR1  +=       8.699 * cos(0.54050826161  + 66.18357885820 *T)
    sumR1  +=       8.027 * cos(1.94079002321  + 180.27386923090 *T)
    sumR1  +=       7.547 * cos(5.94593031762  + 97.41551581630 *T)
    sumR1  +=       7.597 * cos(5.80197738402  + 450.97721326420 *T)
    sumR1  +=       8.666 * cos(3.69933873164  + 300.61086697750 *T)
    sumR1  +=       7.685 * cos(1.47377256329  + 32.24332891440 *T)
    sumR1  +=       8.195 * cos(2.30769657654  + 254.94359321360 *T)
    sumR1  +=       8.473 * cos(1.27680705911  + 39.35687591520 *T)
    sumR1  +=       7.026 * cos(0.68091865104  + 874.39401040250 *T)
    sumR1  +=       8.898 * cos(0.16273040357  + 43.12897048390 *T)
    sumR1  +=       7.205 * cos(4.98177531040  + 117.91056905120 *T)
    sumR1  +=       7.389 * cos(4.09295183164  + 92.30770638560 *T)
    sumR1  +=       7.314 * cos(5.04313738379  + 756.32338265690 *T)
    sumR1  +=       8.454 * cos(1.22026161161  + 79.44746501330 *T)
    sumR1  +=       6.925 * cos(6.04100189247  + 350.33211960040 *T)
    sumR1  +=       8.793 * cos(1.33398658801  + 48.75804477640 *T)
    sumR1  +=       7.270 * cos(3.32609286227  + 68.18931642830 *T)
    sumR1  +=       6.825 * cos(4.77832275072  + 142.66209845490 *T)
    sumR1  +=       6.816 * cos(3.90452052962  + 480.77286162380 *T)
    sumR1  +=       7.062 * cos(1.27536949417  + 68.56182344380 *T)
    sumR1  +=       7.947 * cos(4.29940380231  + 624.91943278700 *T)
    sumR1  +=       6.741 * cos(5.43264472273  + 610.69233878540 *T)
    sumR1  +=       6.529 * cos(5.43599941795  + 88.79624424780 *T)
    sumR1  +=       7.635 * cos(4.81180007736  + 312.45971639350 *T)
    sumR1  +=       7.235 * cos(3.18370421558  + 268.43697403230 *T)
    sumR1  +=       8.133 * cos(1.98936178361  + 692.58748435350 *T)
    sumR1  +=       6.477 * cos(1.05238958778  + 685.47393735270 *T)
    sumR1  +=       6.630 * cos(1.37656948077  + 291.26208774300 *T)
    sumR1  +=       6.878 * cos(2.59188446778  + 282.66406803390 *T)
    sumR1  +=       7.123 * cos(5.79744758808  + 468.24268865160 *T)
    sumR1  +=       6.320 * cos(2.58497126634  + 458.09076026500 *T)
    sumR1  +=       6.222 * cos(5.68982546821  + 113.87784205160 *T)
    sumR1  +=       7.635 * cos(0.49482302003  + 296.15744885260 *T)
    sumR1  +=       8.521 * cos(0.00576688485  + 227.31374111850 *T)
    sumR1  +=       6.520 * cos(3.99093726386  + 42.53826965290 *T)
    sumR1  +=       6.435 * cos(1.03721543102  + 365.90067395840 *T)
    sumR1  +=       6.107 * cos(0.35071886662  + 148.59998928810 *T)
    sumR1  +=       8.199 * cos(1.13448902886  + 69.15252427480 *T)
    sumR1  +=       6.102 * cos(0.94101111641  + 13.33332212430 *T)
    sumR1  +=       5.989 * cos(4.98445156102  + 184.09414790940 *T)
    sumR1  +=       6.355 * cos(0.16346166674  + 228.27694896500 *T)
    sumR1  +=       7.955 * cos(4.03567630186  + 183.24281464750 *T)
    sumR1  +=       5.884 * cos(4.40842406038  + 19.12245511120 *T)
    sumR1  +=       5.938 * cos(5.40863870407  + 17.52610781830 *T)
    sumR1  +=       5.869 * cos(5.39494525133  + 95.38852638680 *T)
    sumR1  +=       5.775 * cos(2.81250784939  + 121.84272231430 *T)
    sumR1  +=       6.070 * cos(4.23605170027  + 119.50691634410 *T)
    sumR1  +=       6.349 * cos(3.52304701692  + 285.63301345050 *T)
    sumR1  +=       5.780 * cos(0.17831551537  + 458.84151979040 *T)
    sumR1  +=       5.674 * cos(4.16711163603  + 89.75945209430 *T)
    sumR1  +=       5.534 * cos(4.24741728108  + 75.53235809270 *T)
    sumR1  +=       5.648 * cos(2.81224199321  + 154.97982310600 *T)
    sumR1  +=       6.939 * cos(3.31979953743  + 306.83064210100 *T)
    sumR1  +=       5.682 * cos(4.79764449768  + 248.72381809010 *T)
    sumR1  +=       6.087 * cos(4.04640130992  + 271.40591944890 *T)
    sumR1  +=       6.869 * cos(1.34392408836  + 7.86430652620 *T)
    sumR1  +=       5.611 * cos(5.32955957046  + 920.86443331850 *T)
    sumR1  +=       6.495 * cos(0.45735814276  + 106.27416795630 *T)
    sumR1  +=       5.353 * cos(2.49825965802  + 24.11838995730 *T)
    sumR1  +=       6.612 * cos(5.24626646696  + 58.31927233200 *T)
    sumR1  +=       5.552 * cos(0.24515487696  + 173.68158709190 *T)
    sumR1  +=       5.209 * cos(6.07866998675  + 134.06407874580 *T)
    sumR1  +=       5.176 * cos(3.69984512887  + 778.41478318470 *T)
    sumR1  +=       5.949 * cos(3.63204266272  + 189.72322220190 *T)
    sumR1  +=       6.360 * cos(0.35370738262  + 411.62033734900 *T)
    sumR1  +=       5.147 * cos(1.55440402971  + 193.65537546500 *T)
    sumR1  +=       6.436 * cos(5.18759014405  + 120.99138905240 *T)
    sumR1  +=       6.994 * cos(4.85978914075  + 419.48464387520 *T)
    sumR1  +=       5.323 * cos(0.50787742639  + 16.46232623530 *T)
    sumR1  +=       5.085 * cos(1.28917723765  + 267.47376618580 *T)
    sumR1  +=       5.993 * cos(4.70505267412  + 298.23262239190 *T)
    sumR1  +=       5.507 * cos(2.72405080404  + 986.08480433020 *T)
    sumR1  +=       6.163 * cos(1.87793216012  + 397.39324334740 *T)
    sumR1  +=       4.846 * cos(5.66714115710  + 90.82323367730 *T)
    sumR1  +=       4.875 * cos(1.24385851949  + 25.60286266560 *T)
    sumR1  +=       5.374 * cos(0.31175745933  + 192.69216761850 *T)
    sumR1  +=       5.262 * cos(1.85699096844  + 114.94162363460 *T)
    sumR1  +=       5.373 * cos(6.22242588334  + 91.45637312370 *T)
    sumR1  +=       5.050 * cos(3.39322756907  + 831.10498122420 *T)
    sumR1  +=       4.637 * cos(0.84958882655  + 403.02231763990 *T)
    sumR1  +=       6.382 * cos(2.77560901069  + 198.32124191100 *T)
    sumR1  +=       4.685 * cos(4.94029403928  + 902.70518605380 *T)
    sumR1  +=       5.005 * cos(1.40309022449  + 6.15033915430 *T)
    sumR1  +=       5.014 * cos(5.57665259095  + 451.94042111070 *T)
    sumR1  +=       4.580 * cos(2.47734499363  + 31.23193695810 *T)
    sumR1  +=       5.129 * cos(3.23528704150  + 109.31254934210 *T)
    sumR1  +=       4.459 * cos(6.22635092697  + 207.88246946660 *T)
    sumR1  +=       5.734 * cos(0.96616252776  + 483.22054217860 *T)
    sumR1  +=       4.425 * cos(2.74721673213  + 823.99143422340 *T)
    sumR1  +=       4.575 * cos(1.87994871749  + 44.72531777680 *T)
    sumR1  +=       4.748 * cos(0.34902594832  + 457.87831194390 *T)
    sumR1  +=       4.268 * cos(4.89983575247  + 124.50285119020 *T)
    sumR1  +=       4.709 * cos(5.28612293112  + 449.28029223480 *T)
    sumR1  +=       5.761 * cos(2.09247769051  + 187.69623277240 *T)
    sumR1  +=       4.284 * cos(0.66132439268  + 210.33015002140 *T)
    sumR1  +=       4.318 * cos(1.68857333749  + 309.27832265580 *T)
    sumR1  +=       4.332 * cos(1.41872733238  + 25.12978191360 *T)
    sumR1  +=       4.305 * cos(1.05990546337  + 606.76018552230 *T)
    sumR1  +=       4.519 * cos(5.84384426255  + 905.88657979150 *T)
    sumR1  +=       3.934 * cos(0.41768897300  + 180.16199464630 *T)
    sumR1  +=       3.973 * cos(3.22666150606  + 639.89728631400 *T)
    sumR1  +=       4.871 * cos(4.61331971606  + 258.02441321480 *T)
    sumR1  +=       4.604 * cos(4.77631056831  + 463.50738623640 *T)
    sumR1  +=       3.943 * cos(3.31312639875  + 107.49800823370 *T)
    sumR1  +=       4.217 * cos(0.73383451512  + 497.44763618020 *T)
    sumR1  +=       4.057 * cos(1.67333716577  + 7.42236354150 *T)
    sumR1  +=       3.854 * cos(6.13547145503  + 34.20088237470 *T)
    sumR1  +=       4.617 * cos(5.89829880253  + 303.86169668440 *T)
    sumR1  +=       5.086 * cos(2.85235518740  + 28.31117565130 *T)
    sumR1  +=       5.337 * cos(2.36556705745  + 477.80391620720 *T)
    sumR1  +=       4.456 * cos(1.74674336635  + 95.97922721780 *T)
    sumR1  +=       4.138 * cos(3.80344455465  + 460.53844081980 *T)
    sumR1  +=       3.812 * cos(2.48508006441  + 25.27279426550 *T)
    sumR1  +=       4.732 * cos(0.87519409311  + 255.05546779820 *T)
    sumR1  +=       3.843 * cos(4.02615028031  + 104.00779795530 *T)
    sumR1  +=       3.776 * cos(2.89171052095  + 27.08733537390 *T)
    sumR1  +=       4.932 * cos(0.36238909407  + 123.53964334370 *T)
    sumR1  +=       4.371 * cos(3.74322467592  + 376.19561469690 *T)
    sumR1  +=       3.747 * cos(3.04126115463  + 142.14083359310 *T)
    sumR1  +=       4.232 * cos(4.31629167726  + 446.31134681820 *T)
    sumR1  +=       3.685 * cos(3.26448469664  + 170.76082578510 *T)
    sumR1  +=       3.575 * cos(4.31199276037  + 572.22923474750 *T)
    sumR1  +=       3.567 * cos(4.08542270507  + 433.71173787680 *T)
    sumR1  +=       4.496 * cos(2.10358455875  + 838.21852822500 *T)
    sumR1  +=       3.505 * cos(3.53902384390  + 520.12973753900 *T)
    sumR1  +=       3.524 * cos(3.75716903766  + 473.06861379200 *T)
    sumR1  +=       3.962 * cos(5.33706246667  + 43.28902917830 *T)
    sumR1  +=       3.597 * cos(3.65066955203  + 976.00231191280 *T)
    sumR1  +=       3.487 * cos(2.12239114397  + 316.39186965660 *T)
    sumR1  +=       3.475 * cos(4.44351326599  + 384.05992122310 *T)
    sumR1  +=       3.628 * cos(2.11511417759  + 73.18525127440 *T)
    sumR1  +=       3.702 * cos(3.86923731076  + 981.63138620530 *T)
    sumR1  +=       3.687 * cos(5.18698183343  + 993.19835133100 *T)
    sumR1  +=       3.599 * cos(2.07986409347  + 47.69426319340 *T)
    sumR1  +=       3.807 * cos(4.21821126511  + 196.62432088160 *T)
    sumR1  +=       4.707 * cos(4.56309173897  + 47.06112374700 *T)
    sumR1  +=       4.312 * cos(0.38740046308  + 988.53248488500 *T)
    sumR1  +=       3.867 * cos(2.08559458308  + 457.35704708210 *T)
    sumR1  +=       4.723 * cos(4.16947683948  + 219.89137757700 *T)
    sumR1  +=       3.527 * cos(0.20371576470  + 394.35486196160 *T)
    sumR1  +=       3.644 * cos(5.82023483708  + 586.31331639720 *T)
    sumR1  +=       3.328 * cos(2.93840719007  + 535.91074021810 *T)
    sumR1  +=       3.321 * cos(4.19289134366  + 114.13847448250 *T)
    sumR1  +=       4.128 * cos(3.06165703109  + 377.15882254340 *T)
    sumR1  +=       3.545 * cos(4.41886084391  + 1293.87865427770 *T)
    sumR1  +=       3.295 * cos(2.97049569593  + 15.19030184810 *T)
    sumR1  +=       3.337 * cos(6.23473900765  + 9947.05568153210 *T)
    sumR1  +=       3.253 * cos(5.22412177835  + 425.11371816770 *T)
    sumR1  +=       3.677 * cos(5.31389484415  + 141.69889060840 *T)
    sumR1  +=       3.242 * cos(4.68868636498  + 978.97125732940 *T)
    sumR1  +=       3.266 * cos(3.57072306171  + 17.26547538740 *T)
    sumR1  +=       3.435 * cos(0.52794358986  + 141.48644228730 *T)
    sumR1  +=       3.242 * cos(2.62760698007  + 6.59228213900 *T)
    sumR1  +=       3.613 * cos(1.94737668557  + 661.09491496450 *T)
    sumR1  +=       3.182 * cos(0.36603315110  + 449.49274055590 *T)
    sumR1  +=       3.311 * cos(1.25616383318  + 233.90602325750 *T)
    sumR1  +=       3.403 * cos(6.03792583170  + 199.28444975750 *T)
    sumR1  +=       4.196 * cos(4.26442082589  + 381.61224066830 *T)
    sumR1  +=       3.961 * cos(4.53281422377  + 916.93228005540 *T)
    sumR1  +=       3.846 * cos(3.76849990033  + 8.59801970910 *T)
    sumR1  +=       3.350 * cos(5.63661413371  + 444.82687410990 *T)
    sumR1  +=       3.780 * cos(5.35722293289  + 328.24071907260 *T)
    sumR1  +=       3.166 * cos(2.16351748263  + 983.11585891360 *T)
    sumR1  +=       3.538 * cos(1.89746744103  + 280.96714700450 *T)
    sumR1  +=       3.930 * cos(2.09444900058  + 653.98136796370 *T)
    sumR1  +=       3.282 * cos(1.91872815218  + 2349.32843120380 *T)
    sumR1  +=       3.269 * cos(0.52855777633  + 450.45594840240 *T)
    sumR1  +=       3.582 * cos(1.60170266832  + 1587.58842257550 *T)
    sumR1  +=       3.522 * cos(2.51782036180  + 237.67811782620 *T)
    sumR1  +=       3.024 * cos(3.54567524563  + 94.42531854030 *T)
    sumR1  +=       3.528 * cos(4.79818282081  + 406.95447090300 *T)
    sumR1  +=       2.996 * cos(2.59155293620  + 6133.51265285680 *T)
    sumR1  +=       3.146 * cos(2.18094827839  + 216.92243216040 *T)
    sumR1  +=       3.610 * cos(6.15486273902  + 171.65459766240 *T)
    sumR1  +=       2.977 * cos(0.69478628170  + 294.30046912880 *T)
    sumR1  +=       3.377 * cos(1.21382647091  + 162.89651925890 *T)
    sumR1  +=       3.347 * cos(4.14981703949  + 214.78356814630 *T)
    sumR1  +=       2.953 * cos(2.18721777019  + 597.35901666110 *T)
    sumR1  +=       4.049 * cos(3.15153850922  + 833.55266177900 *T)
    sumR1  +=       3.725 * cos(5.84743216544  + 6058.73105428950 *T)
    sumR1  +=       3.390 * cos(1.18412112871  + 167.72244439930 *T)
    sumR1  +=       3.142 * cos(2.26934209337  + 517.16079212240 *T)
    sumR1  +=       4.077 * cos(0.07273073033  + 1190.03512053370 *T)
    sumR1  +=       3.020 * cos(2.64998251178  + 20.44686912510 *T)
    sumR1  +=       3.926 * cos(1.41612569694  + 346.18751801620 *T)
    sumR1  +=       3.110 * cos(1.11431255827  + 1044.40407666220 *T)
    sumR1  +=       2.836 * cos(0.62522723719  + 749.20983565610 *T)
    sumR1  +=       2.831 * cos(4.78996738581  + 820.05928096030 *T)
    sumR1  +=       2.824 * cos(0.87232289414  + 30.71067209630 *T)
    sumR1  +=       3.114 * cos(1.79734939525  + 414.06801790380 *T)
    sumR1  +=       2.801 * cos(3.99301180541  + 10063.72234907640 *T)
    sumR1  +=       3.489 * cos(1.86982946081  + 371.52974825090 *T)
    sumR1  +=       3.725 * cos(1.68366366742  + 683.98946464440 *T)
    sumR1  +=       3.763 * cos(3.28247771799  + 432.81796599950 *T)
    sumR1  +=       3.493 * cos(0.98765698465  + 9988.94075050910 *T)
    sumR1  +=       3.523 * cos(5.12512607932  + 105.38039607900 *T)
    sumR1  +=       2.839 * cos(4.22662576295  + 990.22940591440 *T)
    sumR1  +=       3.432 * cos(2.80483162230  + 764.18768918310 *T)
    sumR1  +=       2.733 * cos(0.42373696972  + 354.99798604640 *T)
    sumR1  +=       3.146 * cos(5.19208910201  + 417.03696332040 *T)
    sumR1  +=       3.041 * cos(5.75641149588  + 409.92341631960 *T)
    sumR1  +=       3.379 * cos(5.47448876584  + 1396.22066897090 *T)
    sumR1  +=       3.102 * cos(0.41684444780  + 521.09294538550 *T)
    sumR1  +=       2.863 * cos(0.41519700992  + 894.84087952760 *T)
    sumR1  +=       2.707 * cos(3.60084311477  + 621.73803904930 *T)
    sumR1  +=       3.128 * cos(5.23384180625  + 424.15051032120 *T)
    sumR1  +=       3.107 * cos(2.44919355737  + 4.66586644600 *T)
    sumR1  +=       2.683 * cos(3.88682711832  + 133.10087089930 *T)
    sumR1  +=       2.660 * cos(4.78670985324  + 362.86229257260 *T)
    sumR1  +=       3.200 * cos(1.88004939357  + 331.20966448920 *T)
    sumR1  +=       2.730 * cos(4.12217979791  + 600.54041039880 *T)
    sumR1  +=       3.414 * cos(4.93712749827  + 1140.38330388000 *T)
    sumR1  +=       2.653 * cos(5.10283251074  + 118.02244363580 *T)
    sumR1  +=       3.222 * cos(4.76521772319  + 294.67297614430 *T)
    sumR1  +=       3.289 * cos(4.26401031509  + 544.50875992720 *T)
    sumR1  +=       3.100 * cos(5.47928527930  + 701.18550406260 *T)
    sumR1  +=       2.785 * cos(5.19343849039  + 144.89733068860 *T)
    sumR1  +=       2.607 * cos(4.72531286187  + 122.47586176070 *T)
    sumR1  +=       2.581 * cos(6.27329466695  + 908.33426034630 *T)
    sumR1  +=       3.285 * cos(1.95972622670  + 372.42352012820 *T)
    sumR1  +=       2.897 * cos(0.37378090180  + 582.38116313410 *T)
    sumR1  +=       2.615 * cos(2.25516923974  + 74.99404688840 *T)
    sumR1  +=       3.582 * cos(1.27992264402  + 987.56927703850 *T)
    sumR1  +=       3.115 * cos(5.10929689813  + 459.05396811150 *T)
    sumR1  +=       2.857 * cos(5.90256930211  + 525.23754696970 *T)
    sumR1  +=       2.589 * cos(1.83177157032  + 657.16276170140 *T)
    sumR1  +=       2.539 * cos(4.14968938109  + 74.73341445750 *T)
    sumR1  +=       2.797 * cos(2.82242772664  + 2036.86871481030 *T)
    sumR1  +=       2.688 * cos(2.16500211397  + 262.80789973980 *T)
    sumR1  +=       2.744 * cos(1.54445470732  + 28.57180808220 *T)
    sumR1  +=       2.539 * cos(0.46036497385  + 74.82978267710 *T)
    sumR1  +=       3.322 * cos(3.50108539407  + 82.64590509350 *T)
    sumR1  +=       2.810 * cos(6.06709915335  + 374.49869366750 *T)
    sumR1  +=       2.504 * cos(3.52394801700  + 1183.67233305830 *T)
    sumR1  +=       2.565 * cos(1.64023845161  + 73.40900044360 *T)
    sumR1  +=       2.531 * cos(3.50486296784  + 293.18850343600 *T)
    sumR1  +=       2.663 * cos(4.23321349902  + 421.18156490460 *T)
    sumR1  +=       2.793 * cos(2.00644423849  + 75.04223099820 *T)
    sumR1  +=       2.430 * cos(1.56119387576  + 136.06981631590 *T)
    sumR1  +=       2.553 * cos(1.25909246207  + 670.49608382570 *T)
    sumR1  +=       2.604 * cos(3.87350462519  + 74.03083904190 *T)
    sumR1  +=       2.510 * cos(3.35948960782  + 464.99185894470 *T)
    sumR1  +=       3.005 * cos(0.81031349171  + 73.88782669000 *T)
    sumR1  +=       3.110 * cos(6.14956891318  + 118.87377689770 *T)
    sumR1  +=       3.234 * cos(2.45751141361  + 98.35747180340 *T)
    sumR1  +=       2.774 * cos(6.26134027482  + 1022.31267613440 *T)
    sumR1  +=       2.402 * cos(4.38353347008  + 1303.27982313890 *T)
    sumR1  +=       3.296 * cos(3.84350963765  + 511.53171782990 *T)
    sumR1  +=       2.800 * cos(2.60339313269  + 74.52096613640 *T)
    sumR1  +=       3.005 * cos(0.76247280223  + 75.67537044460 *T)
    sumR1  +=       2.434 * cos(4.94784679430  + 969.62247809490 *T)
    sumR1  +=       2.632 * cos(0.63557110200  + 227.52618943960 *T)
    sumR1  +=       2.669 * cos(0.73340228210  + 73.08467753790 *T)
    sumR1  +=       2.465 * cos(1.30648773380  + 77.06922042770 *T)
    sumR1  +=       3.237 * cos(3.19110274211  + 1887.30551767570 *T)
    sumR1  +=       2.395 * cos(2.76580569447  + 768.85355562910 *T)
    sumR1  +=       3.230 * cos(0.01981320255  + 881.50755740330 *T)
    sumR1  +=       2.747 * cos(5.59085990261  + 388.72578766910 *T)
    sumR1  +=       3.008 * cos(5.65955463660  + 1969.20066324380 *T)
    sumR1  +=       3.008 * cos(0.91409756228  + 2118.76386037840 *T)
    sumR1  +=       2.465 * cos(0.26629856014  + 72.49397670690 *T)
    sumR1  +=       2.629 * cos(4.00618677646  + 26.02355379090 *T)
    sumR1  +=       2.272 * cos(2.77069357315  + 515.46387109300 *T)
    sumR1  +=       2.360 * cos(4.12736987374  + 74.62153987290 *T)
    sumR1  +=       2.255 * cos(3.36574443950  + 286.59622129700 *T)
    sumR1  +=       2.211 * cos(5.18239546182  + 59.28248017850 *T)
    sumR1  +=       2.994 * cos(2.83179016989  + 184.98791978670 *T)
    sumR1  +=       2.492 * cos(1.19872353228  + 383.09671337660 *T)
    sumR1  +=       2.355 * cos(0.48259604722  + 74.94165726170 *T)
    sumR1  +=       2.185 * cos(6.07997119980  + 63.62402371880 *T)
    sumR1  +=       2.228 * cos(1.42452148910  + 6219.33995168800 *T)

    # URANUS Radius
    sumR2  = 0
    sumR2  +=   22439.904 * cos(0.69953118760  + 74.78159856730 *T)
    sumR2  +=    4727.037 * cos(1.69901641488  + 63.73589830340 *T)
    sumR2  +=    1681.903 * cos(4.64833551727  + 70.84944530420 *T)
    sumR2  +=    1433.755 * cos(3.52119917947  + 149.56319713460 *T)
    sumR2  +=    1649.559 * cos(3.09660078980  + 11.04570026390 *T)
    sumR2  +=     770.188 * cos(0.00000000000  + 0.00000000000 *T)
    sumR2  +=     461.009 * cos(0.76676632849  + 3.93215326310 *T)
    sumR2  +=     500.429 * cos(6.17229032223  + 76.26607127560 *T)
    sumR2  +=     390.371 * cos(4.49605283502  + 56.62235130260 *T)
    sumR2  +=     389.945 * cos(5.52673426377  + 85.82729883120 *T)
    sumR2  +=     292.097 * cos(0.20389012095  + 52.69019803950 *T)
    sumR2  +=     272.898 * cos(3.84707823651  + 138.51749687070 *T)
    sumR2  +=     286.579 * cos(3.53357683270  + 73.29712585900 *T)
    sumR2  +=     205.449 * cos(3.24758017121  + 78.71375183040 *T)
    sumR2  +=     219.674 * cos(1.96418942891  + 131.40394986990 *T)
    sumR2  +=     215.788 * cos(0.84812474187  + 77.96299230500 *T)
    sumR2  +=     128.834 * cos(2.08146849515  + 3.18139373770 *T)
    sumR2  +=     148.554 * cos(4.89840863841  + 127.47179660680 *T)
    sumR2  +=     117.452 * cos(4.93414907433  + 447.79581952650 *T)
    sumR2  +=     112.690 * cos(1.01361852218  + 462.02291352810 *T)
    sumR2  +=      98.875 * cos(6.15817742611  + 224.34479570190 *T)
    sumR2  +=      91.379 * cos(0.67973399531  + 18.15924726470 *T)
    sumR2  +=      89.217 * cos(0.23425778826  + 202.25339517410 *T)
    sumR2  +=      88.206 * cos(2.93094837724  + 62.25142559510 *T)
    sumR2  +=     114.066 * cos(4.78741873960  + 145.63104387150 *T)
    sumR2  +=     103.858 * cos(3.58561789629  + 71.60020482960 *T)
    sumR2  +=      61.819 * cos(3.29964272893  + 351.81659230870 *T)
    sumR2  +=      57.782 * cos(4.90737420887  + 22.09140052780 *T)
    sumR2  +=      64.369 * cos(3.39006689398  + 1.48447270830 *T)
    sumR2  +=      71.110 * cos(6.10490061068  + 454.90936652730 *T)
    sumR2  +=      50.990 * cos(3.86691997779  + 65.22037101170 *T)
    sumR2  +=      63.537 * cos(3.96202309168  + 67.66805156650 *T)
    sumR2  +=      58.957 * cos(5.55530463687  + 9.56122755560 *T)
    sumR2  +=      48.700 * cos(3.74709235789  + 269.92144674060 *T)
    sumR2  +=      43.584 * cos(1.92568752002  + 59.80374504030 *T)
    sumR2  +=      42.170 * cos(2.61650997054  + 151.04766984290 *T)
    sumR2  +=      42.420 * cos(6.13634453301  + 284.14854074220 *T)
    sumR2  +=      44.340 * cos(5.89997845114  + 71.81265315070 *T)
    sumR2  +=      37.328 * cos(5.91300114911  + 984.60033162190 *T)
    sumR2  +=      36.201 * cos(5.40315761474  + 77.75054398390 *T)
    sumR2  +=      41.989 * cos(2.09071623849  + 12.53017297220 *T)
    sumR2  +=      31.411 * cos(4.59200004835  + 148.07872442630 *T)
    sumR2  +=      31.289 * cos(2.26696307388  + 195.13984817330 *T)
    sumR2  +=      27.150 * cos(3.53242984046  + 209.36694217490 *T)
    sumR2  +=      28.152 * cos(4.57845964163  + 77.22927912210 *T)
    sumR2  +=      26.097 * cos(0.65978256272  + 120.35824960600 *T)
    sumR2  +=      24.372 * cos(5.86680440531  + 69.36497259590 *T)
    sumR2  +=      23.037 * cos(1.03776963677  + 84.34282612290 *T)
    sumR2  +=      22.679 * cos(1.71434243970  + 160.60889739850 *T)
    sumR2  +=      27.650 * cos(4.91488946525  + 277.03499374140 *T)
    sumR2  +=      20.816 * cos(2.19643268155  + 45.57665103870 *T)
    sumR2  +=      19.961 * cos(2.32077356180  + 2.44768055480 *T)
    sumR2  +=      16.584 * cos(4.77529536873  + 213.29909543800 *T)
    sumR2  +=      16.578 * cos(1.85615182154  + 340.77089204480 *T)
    sumR2  +=      17.196 * cos(4.36852462522  + 54.17467074780 *T)
    sumR2  +=      16.053 * cos(3.64619586667  + 152.74459087230 *T)
    sumR2  +=      14.806 * cos(5.43824503068  + 408.43894361130 *T)
    sumR2  +=      13.872 * cos(3.38531100784  + 358.93013930950 *T)
    sumR2  +=      13.328 * cos(5.25179190495  + 137.03302416240 *T)
    sumR2  +=      13.286 * cos(1.26285812368  + 134.58534360760 *T)
    sumR2  +=      12.890 * cos(3.03270380745  + 92.94084583200 *T)
    sumR2  +=      12.467 * cos(1.33213558369  + 51.20572533120 *T)
    sumR2  +=      13.450 * cos(1.53176996919  + 422.66603761290 *T)
    sumR2  +=      16.442 * cos(0.40190549188  + 265.98929347750 *T)
    sumR2  +=      11.996 * cos(5.10426418352  + 191.20769491020 *T)
    sumR2  +=      12.898 * cos(4.43242192513  + 87.31177153950 *T)
    sumR2  +=      11.449 * cos(2.02645622099  + 7.11354700080 *T)
    sumR2  +=      11.826 * cos(4.65645290272  + 41.64449777560 *T)
    sumR2  +=      12.045 * cos(3.23910807852  + 116.42609634290 *T)
    sumR2  +=      11.680 * cos(3.73278249629  + 220.41264243880 *T)
    sumR2  +=      11.573 * cos(4.16500659139  + 60.55450456570 *T)
    sumR2  +=      10.175 * cos(0.32936886913  + 70.32818044240 *T)
    sumR2  +=      11.332 * cos(1.07613885149  + 72.33391801250 *T)
    sumR2  +=       9.655 * cos(3.05950236129  + 2.96894541660 *T)
    sumR2  +=       9.279 * cos(2.43997351068  + 565.11568774670 *T)
    sumR2  +=       8.986 * cos(5.18839740735  + 225.82926841020 *T)
    sumR2  +=      10.284 * cos(1.18602582060  + 344.70304530790 *T)
    sumR2  +=       8.844 * cos(6.00894470528  + 5.41662597140 *T)
    sumR2  +=       8.508 * cos(5.24741470216  + 347.88443904560 *T)
    sumR2  +=       8.319 * cos(3.71723808749  + 14.97785352700 *T)
    sumR2  +=       8.276 * cos(2.27408171672  + 299.12639426920 *T)
    sumR2  +=       8.064 * cos(5.71681525179  + 55.13787859430 *T)
    sumR2  +=       7.830 * cos(0.90313686798  + 222.86032299360 *T)
    sumR2  +=       8.335 * cos(4.48600419464  + 70.11573212130 *T)
    sumR2  +=       8.763 * cos(5.81519440120  + 153.49535039770 *T)
    sumR2  +=       8.472 * cos(3.91387041805  + 333.65734504400 *T)
    sumR2  +=       9.874 * cos(5.96526143660  + 35.16409022120 *T)
    sumR2  +=       9.647 * cos(0.38872626737  + 415.55249061210 *T)
    sumR2  +=       7.106 * cos(1.50598488470  + 991.71387862270 *T)
    sumR2  +=       6.596 * cos(1.18068235818  + 96.87299909510 *T)
    sumR2  +=       8.065 * cos(2.25930653257  + 206.18554843720 *T)
    sumR2  +=       6.479 * cos(2.99461362786  + 380.12776796000 *T)
    sumR2  +=       9.012 * cos(6.05343622530  + 146.38180339690 *T)
    sumR2  +=       6.131 * cos(0.05596259493  + 99.16062095550 *T)
    sumR2  +=       5.799 * cos(0.82465326137  + 142.44965013380 *T)
    sumR2  +=       5.816 * cos(4.63029217647  + 49.50880430180 *T)
    sumR2  +=       5.608 * cos(0.66268449799  + 58.10682401090 *T)
    sumR2  +=       5.966 * cos(2.48916255408  + 373.01422095920 *T)
    sumR2  +=       5.710 * cos(2.23566160404  + 80.19822453870 *T)
    sumR2  +=       5.272 * cos(5.06746739956  + 440.68227252570 *T)
    sumR2  +=       5.162 * cos(4.36457872885  + 977.48678462110 *T)
    sumR2  +=       5.428 * cos(0.85181859845  + 546.95644048200 *T)
    sumR2  +=       5.766 * cos(0.34229025692  + 536.80451209540 *T)
    sumR2  +=       5.924 * cos(5.48443563529  + 76.47851959670 *T)
    sumR2  +=       5.340 * cos(3.73073116400  + 23.57587323610 *T)
    sumR2  +=       5.174 * cos(4.13873402677  + 132.88842257820 *T)
    sumR2  +=       5.310 * cos(6.14059082194  + 39.61750834610 *T)
    sumR2  +=       5.790 * cos(3.39593613152  + 458.09076026500 *T)
    sumR2  +=       5.007 * cos(4.25821412289  + 522.57741809380 *T)
    sumR2  +=       4.967 * cos(4.79184817938  + 387.24131496080 *T)
    sumR2  +=       5.183 * cos(3.25775152471  + 561.18353448360 *T)
    sumR2  +=       4.602 * cos(1.69262282455  + 152.53214255120 *T)
    sumR2  +=       5.302 * cos(1.83522660093  + 124.29040286910 *T)
    sumR2  +=       5.005 * cos(0.36630565950  + 60.76695288680 *T)
    sumR2  +=       4.454 * cos(2.30288945184  + 312.45971639350 *T)
    sumR2  +=       4.457 * cos(0.45775730382  + 33.13710079170 *T)
    sumR2  +=       5.722 * cos(0.89523844278  + 81.89514556810 *T)
    sumR2  +=       5.842 * cos(0.92039543147  + 20.60692781950 *T)
    sumR2  +=       5.743 * cos(0.66226484448  + 38.13303563780 *T)
    sumR2  +=       4.255 * cos(3.55373860346  + 479.28838891550 *T)
    sumR2  +=       4.190 * cos(4.37674804409  + 79.23501669220 *T)
    sumR2  +=       4.194 * cos(1.64986267170  + 128.95626931510 *T)
    sumR2  +=       5.125 * cos(1.40553011416  + 144.14657116320 *T)
    sumR2  +=       4.045 * cos(6.07362424798  + 19.64371997300 *T)
    sumR2  +=       3.984 * cos(5.77178406410  + 288.08069400530 *T)
    sumR2  +=       5.017 * cos(2.99521887648  + 29.20494752860 *T)
    sumR2  +=       3.842 * cos(2.60024827897  + 426.59819087600 *T)
    sumR2  +=       3.861 * cos(3.19886211335  + 159.12442469020 *T)
    sumR2  +=       3.870 * cos(4.43713601497  + 141.69889060840 *T)
    sumR2  +=       5.316 * cos(4.07970979457  + 111.43016149680 *T)
    sumR2  +=       4.553 * cos(0.01384318412  + 298.23262239190 *T)
    sumR2  +=       3.737 * cos(5.28319518103  + 353.30106501700 *T)
    sumR2  +=       3.939 * cos(5.27301148162  + 521.09294538550 *T)
    sumR2  +=       3.710 * cos(5.15385470848  + 490.33408917940 *T)
    sumR2  +=       4.039 * cos(0.60924359087  + 152.01087768940 *T)
    sumR2  +=       3.861 * cos(1.34394383700  + 535.32003938710 *T)
    sumR2  +=       4.385 * cos(0.62057680100  + 827.17282796110 *T)
    sumR2  +=       3.567 * cos(4.71986443303  + 6.90109867970 *T)
    sumR2  +=       3.576 * cos(3.24526237368  + 230.56457082540 *T)
    sumR2  +=       3.469 * cos(0.79054323335  + 983.11585891360 *T)
    sumR2  +=       4.524 * cos(2.86819565712  + 129.91947716160 *T)
    sumR2  +=       3.648 * cos(5.59395544992  + 774.48262992160 *T)
    sumR2  +=       3.513 * cos(4.49630054276  + 376.19561469690 *T)
    sumR2  +=       3.432 * cos(2.55614913808  + 258.87574647670 *T)
    sumR2  +=       4.352 * cos(2.09804374929  + 404.50679034820 *T)
    sumR2  +=       3.336 * cos(0.89628904042  + 469.13646052890 *T)
    sumR2  +=       3.274 * cos(3.86236880159  + 42.53826965290 *T)
    sumR2  +=       3.201 * cos(2.76459652868  + 248.72381809010 *T)
    sumR2  +=       3.184 * cos(0.07709843451  + 1514.29129671650 *T)
    sumR2  +=       3.783 * cos(5.29835962126  + 369.08206769610 *T)
    sumR2  +=       3.266 * cos(2.24754480216  + 73.81839072080 *T)
    sumR2  +=       3.055 * cos(2.60120354415  + 433.71173787680 *T)
    sumR2  +=       3.051 * cos(4.54953369151  + 980.66817835880 *T)
    sumR2  +=       3.062 * cos(1.27089879603  + 200.76892246580 *T)
    sumR2  +=       3.055 * cos(1.70878161343  + 639.89728631400 *T)
    sumR2  +=       3.110 * cos(3.63187411723  + 16.67477455640 *T)
    sumR2  +=       3.472 * cos(4.93521260607  + 411.62033734900 *T)
    sumR2  +=       3.531 * cos(4.49372794858  + 881.50755740330 *T)
    sumR2  +=       3.284 * cos(5.59170577331  + 472.17484191470 *T)
    sumR2  +=       3.015 * cos(6.02967446446  + 291.26208774300 *T)
    sumR2  +=       3.467 * cos(2.17484439267  + 554.06998748280 *T)
    sumR2  +=       3.138 * cos(0.52367930477  + 1094.80665284130 *T)
    sumR2  +=       3.257 * cos(2.49339546514  + 451.72797278960 *T)
    sumR2  +=       2.881 * cos(0.50481204892  + 305.34616939270 *T)
    sumR2  +=       3.082 * cos(4.20145474081  + 146.59425171800 *T)
    sumR2  +=       2.883 * cos(2.44983947531  + 135.33610313300 *T)
    sumR2  +=       2.965 * cos(0.39294995530  + 25.27279426550 *T)
    sumR2  +=       2.831 * cos(2.52728803131  + 867.28046340170 *T)
    sumR2  +=       2.728 * cos(5.29491477549  + 125.98732389850 *T)
    sumR2  +=       2.857 * cos(4.71106805785  + 218.92816973050 *T)
    sumR2  +=       2.763 * cos(4.27510031656  + 350.33211960040 *T)
    sumR2  +=       2.730 * cos(1.98552777251  + 82.85835341460 *T)
    sumR2  +=       2.857 * cos(3.08706426922  + 216.48048917570 *T)
    sumR2  +=       3.365 * cos(3.67691210011  + 661.09491496450 *T)
    sumR2  +=       2.925 * cos(1.43646759644  + 381.61224066830 *T)
    sumR2  +=       2.753 * cos(0.39468041761  + 33.67961751290 *T)
    sumR2  +=       2.756 * cos(4.62672498840  + 1357.61455258110 *T)
    sumR2  +=       3.450 * cos(2.12911756067  + 685.47393735270 *T)
    sumR2  +=       2.571 * cos(5.92862393284  + 89.75945209430 *T)
    sumR2  +=       2.677 * cos(0.76342313946  + 486.40193591630 *T)
    sumR2  +=       2.689 * cos(4.16436463826  + 235.39049596580 *T)
    sumR2  +=       2.646 * cos(3.81808560938  + 550.88859374510 *T)
    sumR2  +=       3.369 * cos(3.17071565345  + 108.46121608020 *T)
    sumR2  +=       2.613 * cos(5.68333838067  + 24.37902238820 *T)
    sumR2  +=       2.736 * cos(1.87107584495  + 529.69096509460 *T)
    sumR2  +=       2.606 * cos(4.36605237304  + 1080.57955883970 *T)
    sumR2  +=       2.407 * cos(3.07343136742  + 391.17346822390 *T)
    sumR2  +=       2.446 * cos(5.73846381540  + 535.91074021810 *T)
    sumR2  +=       2.334 * cos(5.18878243102  + 1059.38193018920 *T)
    sumR2  +=       2.568 * cos(1.09886876369  + 913.00012679230 *T)
    sumR2  +=       2.236 * cos(6.10115874045  + 140.00196957900 *T)
    sumR2  +=       3.053 * cos(5.35047433775  + 681.54178408960 *T)

    # URANUS Radius
    sumR3  = 0
    sumR3  +=    1164.382 * cos(4.73453291602  + 74.78159856730 *T)
    sumR3  +=     212.367 * cos(3.34255734999  + 63.73589830340 *T)
    sumR3  +=     196.408 * cos(2.98004616318  + 70.84944530420 *T)
    sumR3  +=     104.527 * cos(0.95807937648  + 11.04570026390 *T)
    sumR3  +=      71.681 * cos(0.02528455665  + 56.62235130260 *T)
    sumR3  +=      72.540 * cos(0.99701907912  + 149.56319713460 *T)
    sumR3  +=      54.875 * cos(2.59436811267  + 3.93215326310 *T)
    sumR3  +=      34.029 * cos(3.81553325635  + 76.26607127560 *T)
    sumR3  +=      32.081 * cos(3.59825177840  + 131.40394986990 *T)
    sumR3  +=      29.641 * cos(3.44111535957  + 85.82729883120 *T)
    sumR3  +=      36.377 * cos(5.65035573017  + 77.96299230500 *T)
    sumR3  +=      27.663 * cos(0.42836001470  + 3.18139373770 *T)
    sumR3  +=      27.464 * cos(2.55126467481  + 52.69019803950 *T)
    sumR3  +=      24.569 * cos(5.14034173566  + 78.71375183040 *T)
    sumR3  +=      19.390 * cos(5.13477648625  + 18.15924726470 *T)
    sumR3  +=      15.767 * cos(0.37116951743  + 447.79581952650 *T)
    sumR3  +=      15.441 * cos(5.57271837433  + 462.02291352810 *T)
    sumR3  +=      15.232 * cos(3.85998573509  + 73.29712585900 *T)
    sumR3  +=      15.475 * cos(2.97496547327  + 145.63104387150 *T)
    sumR3  +=      17.951 * cos(0.00000000000  + 0.00000000000 *T)
    sumR3  +=      15.958 * cos(5.19915553904  + 71.60020482960 *T)
    sumR3  +=      11.056 * cos(6.03152659562  + 138.51749687070 *T)
    sumR3  +=      10.529 * cos(3.58261852497  + 224.34479570190 *T)
    sumR3  +=       7.606 * cos(1.44542030704  + 1.48447270830 *T)
    sumR3  +=       8.121 * cos(2.61579604319  + 22.09140052780 *T)
    sumR3  +=       7.107 * cos(5.43946774526  + 269.92144674060 *T)
    sumR3  +=       6.459 * cos(4.37142319461  + 284.14854074220 *T)
    sumR3  +=       6.818 * cos(0.01396812984  + 151.04766984290 *T)
    sumR3  +=       8.101 * cos(0.29563819537  + 127.47179660680 *T)
    sumR3  +=       5.768 * cos(4.22672716593  + 373.01422095920 *T)
    sumR3  +=       5.022 * cos(1.84154937974  + 202.25339517410 *T)
    sumR3  +=       4.692 * cos(2.78404575440  + 120.35824960600 *T)
    sumR3  +=       5.087 * cos(0.77745294804  + 62.25142559510 *T)
    sumR3  +=       4.160 * cos(1.83820502779  + 72.33391801250 *T)
    sumR3  +=       3.922 * cos(1.88900691473  + 209.36694217490 *T)
    sumR3  +=       5.201 * cos(4.15791319343  + 195.13984817330 *T)
    sumR3  +=       3.636 * cos(1.99709010456  + 65.22037101170 *T)
    sumR3  +=       3.582 * cos(3.92592140377  + 124.29040286910 *T)
    sumR3  +=       3.808 * cos(1.04818660873  + 92.94084583200 *T)
    sumR3  +=       4.241 * cos(3.95755998904  + 9.56122755560 *T)
    sumR3  +=       3.497 * cos(1.54139696251  + 148.07872442630 *T)
    sumR3  +=       3.195 * cos(2.98608971329  + 387.24131496080 *T)
    sumR3  +=       3.950 * cos(1.85721204932  + 152.74459087230 *T)
    sumR3  +=       3.277 * cos(1.40881404192  + 351.81659230870 *T)
    sumR3  +=       3.605 * cos(1.17366167402  + 153.49535039770 *T)
    sumR3  +=       2.940 * cos(6.03594958459  + 12.53017297220 *T)
    sumR3  +=       2.744 * cos(5.64674283515  + 134.58534360760 *T)
    sumR3  +=       2.800 * cos(0.79480255927  + 572.22923474750 *T)
    sumR3  +=       3.054 * cos(5.84310939105  + 160.60889739850 *T)
    sumR3  +=       2.662 * cos(1.98593312104  + 450.97721326420 *T)
    sumR3  +=       2.700 * cos(2.77036653988  + 213.29909543800 *T)
    sumR3  +=       2.323 * cos(1.67918985468  + 358.93013930950 *T)
    sumR3  +=       2.254 * cos(5.77129530133  + 84.34282612290 *T)
    sumR3  +=       2.291 * cos(4.81424601600  + 536.80451209540 *T)
    sumR3  +=       2.213 * cos(2.20360299816  + 465.95506679120 *T)

    # URANUS Radius
    sumR4  = 0
    sumR4  +=      52.996 * cos(3.00838033088  + 74.78159856730 *T)
    sumR4  +=       9.887 * cos(1.91399083603  + 56.62235130260 *T)
    sumR4  +=       7.008 * cos(5.08677527404  + 11.04570026390 *T)
    sumR4  +=       6.728 * cos(5.42924958121  + 149.56319713460 *T)
    sumR4  +=       3.703 * cos(5.22728581851  + 131.40394986990 *T)
    sumR4  +=       3.361 * cos(1.29695290266  + 85.82729883120 *T)
    sumR4  +=       2.664 * cos(0.44064577837  + 63.73589830340 *T)
    sumR4  +=       2.309 * cos(0.92380720934  + 145.63104387150 *T)
    sumR4  +=       2.383 * cos(6.21390585593  + 358.93013930950 *T)
    sumR4  +=       2.288 * cos(2.23425399117  + 440.68227252570 *T)
    sumR4  +=       3.093 * cos(3.14159265359  + 0.00000000000 *T)

    # calculate and scale the sums
    L = sumL0 + sumL1*T**1 + sumL2*T**2 + sumL3*T**3 + sumL4*T**4 + sumL5*T**5
    L /= 1e8
    L = Mod(degrees(L), 360.)
    B = sumB0 + sumB1*T**1 + sumB2*T**2 + sumB3*T**3 + sumB4*T**4
    B /= 1e8
    B = degrees(B)
    R = sumR0 + sumR1*T**1 + sumR2*T**2 + sumR3*T**3 + sumR4*T**4
    R /= 1e8

    # calculate X, Y, Z
    X = R * Cos(B) * Cos(L)
    Y = R * Cos(B) * Sin(L)
    Z = R * Sin(B)

    # create dictionary
    Dict = {  "Object"  : "Uranus",
              "Mode"    : "LBR",
              "Long"    : L,
              "Lat"     : B,
              "Rau"     : R,
              "X"       : X,
              "Y"       : Y,
              "Z"       : Z,
              "Alt"     : -34./60.
           }
    Data = (L, B, R, Dict)


    return Data
#######################################################################



def getUranusData(dtCivicTime, zone, verbose = False):
    L, B, R, Dict = getUranusLBRData(dtCivicTime, zone, verbose)
    X = Dict["X"]; Y = Dict["Y"]; Z = Dict["Z"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Le, Be, Re, Dicte = getEarthLBRData(dtCivicTime, zone, verbose)
        Xe = Dicte["X"]; Ye = Dicte["Y"]; Ze = Dicte["Z"]
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getVenusLBRData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Venus"
#       Long  :  heliocentric Longitude, degrees
#       Lat   :  heliocentric Latitude,degrees
#       Rau   :  Radius, distaance Sun - Object (au)
#
#   Similarly for each planet
#   data from Meeus, Appendix II and VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.2 hvf 2023-04-06 getVenusLBRData.py created by mkPlanetLBRData.py"
VersionList.append(Version)

def getVenusLBRData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. (31.1) Julian Millennia

#######################################################################

    # VENUS Longitude
    sumL0  = 0
    sumL0  +=  317614666.774 * cos(0.00000000000  + 0.00000000000 *T)
    sumL0  +=  1353968.419 * cos(5.59313319619  + 10213.28554621100 *T)
    sumL0  +=   89891.645 * cos(5.30650048468  + 20426.57109242200 *T)
    sumL0  +=    5477.201 * cos(4.41630652531  + 7860.41939243920 *T)
    sumL0  +=    3455.732 * cos(2.69964470778  + 11790.62908865880 *T)
    sumL0  +=    2372.061 * cos(2.99377539568  + 3930.20969621960 *T)
    sumL0  +=    1317.108 * cos(5.18668219093  + 26.29831979980 *T)
    sumL0  +=    1664.069 * cos(4.25018935030  + 1577.34354244780 *T)
    sumL0  +=    1438.322 * cos(4.15745043958  + 9683.59458111640 *T)
    sumL0  +=    1200.521 * cos(6.15357115319  + 30639.85663863300 *T)
    sumL0  +=     761.380 * cos(1.95014702120  + 529.69096509460 *T)
    sumL0  +=     707.676 * cos(1.06466707214  + 775.52261132400 *T)
    sumL0  +=     584.836 * cos(3.99839884762  + 191.44826611160 *T)
    sumL0  +=     769.314 * cos(0.81629615911  + 9437.76293488700 *T)
    sumL0  +=     499.915 * cos(4.12340210074  + 15720.83878487840 *T)
    sumL0  +=     326.221 * cos(4.59056473097  + 10404.73381232260 *T)
    sumL0  +=     429.498 * cos(3.58642859752  + 19367.18916223280 *T)
    sumL0  +=     326.967 * cos(5.67736583705  + 5507.55323866740 *T)
    sumL0  +=     231.937 * cos(3.16251057072  + 9153.90361602180 *T)
    sumL0  +=     179.695 * cos(4.65337915578  + 1109.37855209340 *T)
    sumL0  +=     128.263 * cos(4.22604493736  + 20.77539549240 *T)
    sumL0  +=     155.464 * cos(5.57043888948  + 19651.04848109800 *T)
    sumL0  +=     127.907 * cos(0.96209822685  + 5661.33204915220 *T)
    sumL0  +=     105.547 * cos(1.53721191253  + 801.82093112380 *T)
    sumL0  +=      85.722 * cos(0.35589249966  + 3154.68708489560 *T)
    sumL0  +=      99.121 * cos(0.83288185132  + 213.29909543800 *T)
    sumL0  +=      98.804 * cos(5.39389655503  + 13367.97263110660 *T)
    sumL0  +=      82.094 * cos(3.21596990826  + 18837.49819713819 *T)
    sumL0  +=      88.031 * cos(3.88868860307  + 9999.98645077300 *T)
    sumL0  +=      71.577 * cos(0.11145739345  + 11015.10647733480 *T)
    sumL0  +=      56.122 * cos(4.24039855475  + 7.11354700080 *T)
    sumL0  +=      70.239 * cos(0.67458813282  + 23581.25817731760 *T)
    sumL0  +=      50.796 * cos(0.24531603049  + 11322.66409830440 *T)
    sumL0  +=      46.111 * cos(5.31576465717  + 18073.70493865020 *T)
    sumL0  +=      44.574 * cos(6.06282201966  + 40853.14218484400 *T)
    sumL0  +=      42.594 * cos(5.32873337210  + 2352.86615377180 *T)
    sumL0  +=      42.635 * cos(1.79955421680  + 7084.89678111520 *T)
    sumL0  +=      41.177 * cos(0.36240972161  + 382.89653222320 *T)
    sumL0  +=      35.749 * cos(2.70448479296  + 10206.17199921020 *T)
    sumL0  +=      33.893 * cos(2.02347322198  + 6283.07584999140 *T)
    sumL0  +=      29.138 * cos(3.59230925768  + 22003.91463486980 *T)
    sumL0  +=      28.479 * cos(2.22375414002  + 1059.38193018920 *T)
    sumL0  +=      29.850 * cos(4.02176977477  + 10239.58386601080 *T)
    sumL0  +=      33.252 * cos(2.10025596509  + 27511.46787353720 *T)
    sumL0  +=      30.172 * cos(4.94191919890  + 13745.34623902240 *T)
    sumL0  +=      29.252 * cos(3.51392387787  + 283.85931886520 *T)
    sumL0  +=      24.424 * cos(2.70177493852  + 8624.21265092720 *T)
    sumL0  +=      20.274 * cos(3.79493637509  + 14143.49524243060 *T)
    sumL0  +=      24.322 * cos(4.27814493315  + 5.52292430740 *T)
    sumL0  +=      26.260 * cos(0.54067587552  + 17298.18232732620 *T)
    sumL0  +=      20.492 * cos(0.58547075036  + 38.02767263580 *T)
    sumL0  +=      18.988 * cos(4.13811500642  + 4551.95349705880 *T)
    sumL0  +=      23.739 * cos(4.82870797552  + 6872.67311951120 *T)
    sumL0  +=      15.885 * cos(1.50067222283  + 8635.94200376320 *T)
    sumL0  +=      19.069 * cos(6.12025580313  + 29050.78374334920 *T)
    sumL0  +=      18.269 * cos(3.04740408477  + 19999.97290154599 *T)
    sumL0  +=      13.656 * cos(4.41336292334  + 3532.06069281140 *T)
    sumL0  +=      17.094 * cos(3.52161526430  + 31441.67756975680 *T)
    sumL0  +=      10.955 * cos(2.84562790076  + 18307.80723204360 *T)
    sumL0  +=      11.048 * cos(2.58361219075  + 9786.68735533500 *T)
    sumL0  +=       9.904 * cos(1.08737710389  + 7064.12138562280 *T)
    sumL0  +=      10.576 * cos(0.85419784436  + 10596.18207843420 *T)
    sumL0  +=       9.231 * cos(5.52471655579  + 12566.15169998280 *T)
    sumL0  +=      11.599 * cos(5.81007422699  + 19896.88012732740 *T)
    sumL0  +=      11.807 * cos(1.91250672543  + 21228.39202354580 *T)
    sumL0  +=      10.105 * cos(2.34270786693  + 10742.97651130560 *T)
    sumL0  +=       8.154 * cos(1.92331359797  + 15.25247118500 *T)
    sumL0  +=       8.893 * cos(1.97291388515  + 10186.98722641120 *T)
    sumL0  +=       9.352 * cos(4.94508904764  + 35371.88726597640 *T)
    sumL0  +=       6.821 * cos(4.39733188968  + 8662.24032356300 *T)
    sumL0  +=       6.688 * cos(1.55310437864  + 14945.31617355440 *T)
    sumL0  +=       6.413 * cos(2.17677652923  + 10988.80815753500 *T)
    sumL0  +=       5.802 * cos(1.93462125906  + 3340.61242669980 *T)
    sumL0  +=       5.950 * cos(2.96578175391  + 4732.03062734340 *T)
    sumL0  +=       5.275 * cos(5.01877102496  + 28286.99048486120 *T)
    sumL0  +=       7.047 * cos(1.00111452053  + 632.78373931320 *T)
    sumL0  +=       5.048 * cos(4.27886209626  + 29580.47470844380 *T)
    sumL0  +=       6.305 * cos(0.35506331180  + 103.09277421860 *T)
    sumL0  +=       5.959 * cos(5.04792949464  + 245.83164622940 *T)
    sumL0  +=       4.651 * cos(0.85218058876  + 6770.71060124560 *T)
    sumL0  +=       5.580 * cos(0.48723384809  + 522.57741809380 *T)
    sumL0  +=       5.327 * cos(3.03115417024  + 10021.83728009940 *T)
    sumL0  +=       5.010 * cos(5.77375166500  + 28521.09277825460 *T)
    sumL0  +=       4.608 * cos(1.93302108394  + 4705.73230754360 *T)
    sumL0  +=       5.526 * cos(3.36797048901  + 25158.60171976540 *T)
    sumL0  +=       3.863 * cos(4.89351531412  + 25934.12433108940 *T)
    sumL0  +=       5.303 * cos(0.08161426841  + 39302.09696219600 *T)
    sumL0  +=       4.254 * cos(5.36046113295  + 21535.94964451540 *T)
    sumL0  +=       3.763 * cos(1.05304597315  + 19.66976089979 *T)
    sumL0  +=       4.407 * cos(4.02575374517  + 74.78159856730 *T)
    sumL0  +=       4.145 * cos(1.14356412295  + 9676.48103411560 *T)
    sumL0  +=       4.318 * cos(4.38289970585  + 316.39186965660 *T)
    sumL0  +=       3.642 * cos(6.11733529325  + 3128.38876509580 *T)
    sumL0  +=       3.238 * cos(5.39551036769  + 419.48464387520 *T)
    sumL0  +=       3.909 * cos(4.05263635330  + 9690.70812811720 *T)
    sumL0  +=       3.152 * cos(0.72553551731  + 16496.36139620240 *T)
    sumL0  +=       3.496 * cos(0.72414615705  + 3723.50895892300 *T)
    sumL0  +=       3.755 * cos(3.80208713127  + 19786.67380610799 *T)
    sumL0  +=       2.891 * cos(3.33782737770  + 32217.20018108080 *T)
    sumL0  +=       3.016 * cos(1.57249112496  + 17277.40693183380 *T)
    sumL0  +=       3.825 * cos(0.19612312903  + 426.59819087600 *T)
    sumL0  +=       3.798 * cos(0.45524571743  + 10316.37832042960 *T)
    sumL0  +=       2.570 * cos(1.20813474107  + 13936.79450513400 *T)
    sumL0  +=       2.796 * cos(3.65128969074  + 206.18554843720 *T)
    sumL0  +=       2.466 * cos(3.61988676373  + 1551.04522264800 *T)
    sumL0  +=       3.108 * cos(1.50325806664  + 43232.30665841560 *T)
    sumL0  +=       2.976 * cos(4.79415001304  + 29088.81141598500 *T)
    sumL0  +=       2.217 * cos(3.59623681714  + 24356.78078864160 *T)
    sumL0  +=       2.227 * cos(4.96059221940  + 536.80451209540 *T)
    sumL0  +=       2.397 * cos(3.45249688427  + 19374.30270923360 *T)
    sumL0  +=       2.462 * cos(0.53295178258  + 19360.07561523199 *T)
    sumL0  +=       2.205 * cos(2.70399309963  + 12592.45001978260 *T)
    sumL0  +=       2.230 * cos(3.01413465913  + 18875.52586977400 *T)
    sumL0  +=       1.858 * cos(4.06129152783  + 2379.16447357160 *T)
    sumL0  +=       1.807 * cos(3.15086214479  + 9573.38825989700 *T)
    sumL0  +=       2.238 * cos(5.52216925076  + 10138.50394764370 *T)
    sumL0  +=       2.195 * cos(2.32046770554  + 8094.52168583260 *T)
    sumL0  +=       2.101 * cos(2.90421302975  + 9967.45389998160 *T)
    sumL0  +=       1.916 * cos(4.56513949099  + 2218.75710418680 *T)
    sumL0  +=       1.467 * cos(2.42640162465  + 10234.06094170340 *T)
    sumL0  +=       1.726 * cos(5.59790693845  + 20452.86941222180 *T)
    sumL0  +=       1.455 * cos(2.44757248737  + 1589.07289528380 *T)
    sumL0  +=       1.991 * cos(4.04623390359  + 31749.23519072640 *T)
    sumL0  +=       1.406 * cos(2.71736996917  + 16983.99614745660 *T)
    sumL0  +=       1.658 * cos(0.11252373292  + 153.77881048480 *T)
    sumL0  +=       1.851 * cos(2.92898027939  + 47162.51635463520 *T)
    sumL0  +=       1.492 * cos(1.07513892753  + 9103.90699411760 *T)
    sumL0  +=       1.247 * cos(2.48433565896  + 17778.11626694899 *T)
    sumL0  +=       1.549 * cos(4.20553654300  + 3442.57494496540 *T)
    sumL0  +=       1.243 * cos(3.95452438599  + 170.67287061920 *T)
    sumL0  +=       1.694 * cos(6.20694480406  + 33019.02111220460 *T)
    sumL0  +=       1.221 * cos(4.77931820602  + 30110.16567353840 *T)
    sumL0  +=       1.206 * cos(0.30531303095  + 29864.33402730900 *T)
    sumL0  +=       1.238 * cos(5.05581820608  + 20213.27199698400 *T)
    sumL0  +=       1.152 * cos(3.26229919481  + 11.72935283600 *T)
    sumL0  +=       1.179 * cos(1.69491074791  + 20400.27277262220 *T)
    sumL0  +=       1.165 * cos(2.88995128147  + 574.34479833480 *T)
    sumL0  +=       1.306 * cos(0.14519588607  + 9146.79006902100 *T)
    sumL0  +=       1.113 * cos(1.52598846804  + 10426.58464164900 *T)
    sumL0  +=       1.104 * cos(2.58791423813  + 18849.22754997420 *T)
    sumL0  +=       1.045 * cos(0.57539216420  + 15874.61759536320 *T)
    sumL0  +=       1.360 * cos(2.41976595457  + 38734.37832446560 *T)
    sumL0  +=       0.981 * cos(4.37930727798  + 110.20632121940 *T)
    sumL0  +=       1.095 * cos(0.49492867814  + 51066.42773105500 *T)
    sumL0  +=       1.146 * cos(4.54241454215  + 10220.39909321180 *T)
    sumL0  +=       0.981 * cos(1.65915064733  + 10103.07922499160 *T)
    sumL0  +=       1.270 * cos(4.69374306132  + 9050.81084180320 *T)
    sumL0  +=       1.065 * cos(4.41645258887  + 22805.73556599360 *T)
    sumL0  +=       0.854 * cos(2.34437926957  + 6681.22485339960 *T)
    sumL0  +=       1.104 * cos(0.49781459714  + 1.48447270830 *T)
    sumL0  +=       1.075 * cos(1.09857593161  + 377.37360791580 *T)
    sumL0  +=       1.114 * cos(4.35024775806  + 51092.72605085480 *T)
    sumL0  +=       0.829 * cos(5.41196274578  + 27991.40181316000 *T)
    sumL0  +=       0.900 * cos(2.74195379617  + 41962.52073693740 *T)
    sumL0  +=       1.010 * cos(2.96092073452  + 135.62532501000 *T)
    sumL0  +=       0.768 * cos(3.98260860494  + 18844.61174413899 *T)
    sumL0  +=       1.018 * cos(1.36891050752  + 36949.23080842420 *T)
    sumL0  +=       0.726 * cos(1.67728773965  + 21202.09370374600 *T)
    sumL0  +=       0.727 * cos(0.89048212541  + 467.96499035440 *T)
    sumL0  +=       0.869 * cos(2.93767679827  + 10192.51015071860 *T)
    sumL0  +=       0.696 * cos(5.35698039414  + 10063.72234907640 *T)
    sumL0  +=       0.920 * cos(4.17128923588  + 18734.40542291960 *T)
    sumL0  +=       0.691 * cos(1.50594097883  + 27197.28169366760 *T)
    sumL0  +=       0.835 * cos(0.48050621092  + 20618.01935853360 *T)
    sumL0  +=       0.711 * cos(0.19750098222  + 18830.38465013739 *T)
    sumL0  +=       0.811 * cos(0.16685071959  + 12432.04265039780 *T)
    sumL0  +=       0.756 * cos(3.79022623226  + 9161.01716302260 *T)
    sumL0  +=       0.622 * cos(5.33659507738  + 9411.46461508720 *T)
    sumL0  +=       0.862 * cos(5.72705356405  + 10175.15251057320 *T)
    sumL0  +=       0.853 * cos(0.10404188453  + 2107.03450754240 *T)
    sumL0  +=       0.742 * cos(3.96365892051  + 813.55028395980 *T)
    sumL0  +=       0.705 * cos(0.71229660616  + 220.41264243880 *T)
    sumL0  +=       0.584 * cos(1.71900692700  + 36.02786667740 *T)
    sumL0  +=       0.612 * cos(0.36418385449  + 949.17560896980 *T)
    sumL0  +=       0.587 * cos(1.58648949290  + 6.62855890001 *T)
    sumL0  +=       0.581 * cos(5.49288908804  + 6309.37416979120 *T)
    sumL0  +=       0.581 * cos(4.80353237853  + 24150.08005134500 *T)
    sumL0  +=       0.516 * cos(6.07328802561  + 38.13303563780 *T)
    sumL0  +=       0.627 * cos(5.47281424954  + 9580.50180689780 *T)
    sumL0  +=       0.601 * cos(1.40500080774  + 1162.47470440780 *T)
    sumL0  +=       0.620 * cos(4.00681042667  + 9992.87290377220 *T)
    sumL0  +=       0.611 * cos(3.62010998629  + 7255.56965173440 *T)
    sumL0  +=       0.697 * cos(2.22359630727  + 348.92442044800 *T)
    sumL0  +=       0.693 * cos(5.77432072851  + 55022.93574707440 *T)
    sumL0  +=       0.494 * cos(0.29761886866  + 7058.59846131540 *T)
    sumL0  +=       0.563 * cos(0.24172140474  + 37410.56723987860 *T)
    sumL0  +=       0.487 * cos(5.86917216517  + 10137.01947493540 *T)
    sumL0  +=       0.493 * cos(2.04534833854  + 735.87651353180 *T)
    sumL0  +=       0.636 * cos(2.79707392326  + 40879.44050464380 *T)
    sumL0  +=       0.519 * cos(4.13945657630  + 16522.65971600220 *T)
    sumL0  +=       0.535 * cos(4.60569597820  + 19573.37471066999 *T)
    sumL0  +=       0.555 * cos(5.88120421263  + 26735.94526221320 *T)
    sumL0  +=       0.440 * cos(5.61490649795  + 23958.63178523340 *T)
    sumL0  +=       0.541 * cos(0.62494922735  + 10007.09999777380 *T)
    sumL0  +=       0.427 * cos(4.02335620501  + 14.22709400160 *T)
    sumL0  +=       0.434 * cos(0.29028429049  + 9264.10993724120 *T)
    sumL0  +=       0.451 * cos(1.66320363626  + 26087.90314157420 *T)
    sumL0  +=       0.422 * cos(3.38413582674  + 10787.63034454580 *T)
    sumL0  +=       0.569 * cos(5.14001758731  + 27490.69247804480 *T)
    sumL0  +=       0.421 * cos(4.23407313457  + 39793.76025465480 *T)
    sumL0  +=       0.458 * cos(5.28786368820  + 49.75702547180 *T)
    sumL0  +=       0.418 * cos(5.69097790790  + 14765.23904326980 *T)
    sumL0  +=       0.475 * cos(0.97544690438  + 1052.26838318840 *T)
    sumL0  +=       0.387 * cos(4.41665162999  + 21.85082932640 *T)
    sumL0  +=       0.523 * cos(2.90512426870  + 20235.12282631040 *T)
    sumL0  +=       0.506 * cos(5.26999314618  + 29999.95935231900 *T)
    sumL0  +=       0.438 * cos(2.49457071132  + 20956.26205751660 *T)
    sumL0  +=       0.530 * cos(0.71368442157  + 33794.54372352860 *T)
    sumL0  +=       0.382 * cos(1.92119365480  + 3.93215326310 *T)
    sumL0  +=       0.365 * cos(3.81715328784  + 20419.45754542119 *T)
    sumL0  +=       0.426 * cos(2.06384083608  + 38204.68735937100 *T)
    sumL0  +=       0.496 * cos(0.44077356179  + 9835.91193829520 *T)
    sumL0  +=       0.410 * cos(4.93346326003  + 19264.09638801420 *T)
    sumL0  +=       0.380 * cos(3.79573358631  + 8521.11987670860 *T)
    sumL0  +=       0.334 * cos(5.51158557799  + 10251.31321884680 *T)
    sumL0  +=       0.412 * cos(2.56129670728  + 77.67377042800 *T)
    sumL0  +=       0.418 * cos(2.37865963521  + 32.53255079140 *T)
    sumL0  +=       0.325 * cos(6.03020523465  + 18947.70451835760 *T)
    sumL0  +=       0.400 * cos(0.91999360201  + 227.47613278900 *T)
    sumL0  +=       0.437 * cos(0.91420135162  + 58953.14544329400 *T)
    sumL0  +=       0.360 * cos(0.82477639126  + 22.77520145080 *T)
    sumL0  +=       0.413 * cos(4.22381905655  + 44809.65020086340 *T)
    sumL0  +=       0.375 * cos(3.15657291896  + 19992.85935454519 *T)
    sumL0  +=       0.371 * cos(6.05370874275  + 20007.08644854680 *T)
    sumL0  +=       0.361 * cos(5.44371227904  + 19470.28193645139 *T)
    sumL0  +=       0.386 * cos(5.28120540405  + 47623.85278608960 *T)
    sumL0  +=       0.389 * cos(0.73216672240  + 19050.79729257620 *T)
    sumL0  +=       0.320 * cos(2.84811591194  + 10199.05845220940 *T)
    sumL0  +=       0.386 * cos(3.88754165531  + 1975.49254585600 *T)
    sumL0  +=       0.276 * cos(4.33979180814  + 20809.46762464520 *T)
    sumL0  +=       0.276 * cos(0.50647429773  + 9830.38901398780 *T)
    sumL0  +=       0.309 * cos(3.79299100668  + 18204.71445782499 *T)
    sumL0  +=       0.377 * cos(0.73768791281  + 11506.76976979360 *T)
    sumL0  +=       0.322 * cos(0.96138442100  + 30666.15495843280 *T)
    sumL0  +=       0.363 * cos(1.30472406690  + 9367.20271145980 *T)
    sumL0  +=       0.366 * cos(2.79972786028  + 11272.66747640020 *T)
    sumL0  +=       0.271 * cos(4.66141338193  + 846.08283475120 *T)
    sumL0  +=       0.259 * cos(0.42031175750  + 39264.06928956020 *T)
    sumL0  +=       0.285 * cos(0.40546033634  + 30.91412563500 *T)
    sumL0  +=       0.247 * cos(4.80676426942  + 36147.40987730040 *T)
    sumL0  +=       0.264 * cos(2.71608177583  + 11.04570026390 *T)
    sumL0  +=       0.233 * cos(2.76423842887  + 187.92514776260 *T)
    sumL0  +=       0.248 * cos(1.60765612338  + 10497.14486507620 *T)
    sumL0  +=       0.271 * cos(0.82348919630  + 19793.78735310880 *T)
    sumL0  +=       0.225 * cos(3.80080957016  + 8631.32619792800 *T)
    sumL0  +=       0.263 * cos(1.92311689852  + 37724.75341974820 *T)
    sumL0  +=       0.214 * cos(5.01663795092  + 639.89728631400 *T)
    sumL0  +=       0.289 * cos(0.12342601246  + 20277.00789528740 *T)
    sumL0  +=       0.210 * cos(0.12771800254  + 29.82143814880 *T)
    sumL0  +=       0.227 * cos(4.18036609801  + 17468.85519794540 *T)
    sumL0  +=       0.274 * cos(2.34929343000  + 62883.35513951360 *T)
    sumL0  +=       0.260 * cos(5.65254501655  + 48739.85989708300 *T)
    sumL0  +=       0.271 * cos(4.95325404028  + 4214.06901508480 *T)
    sumL0  +=       0.219 * cos(2.08775228014  + 194.97138446060 *T)
    sumL0  +=       0.191 * cos(2.49267248333  + 568.82187402740 *T)
    sumL0  +=       0.250 * cos(1.52909737354  + 6037.24420376200 *T)
    sumL0  +=       0.231 * cos(5.23674429498  + 491.66329245880 *T)
    sumL0  +=       0.182 * cos(4.98046042571  + 18418.01355326299 *T)
    sumL0  +=       0.188 * cos(2.82273639603  + 1385.89527633620 *T)
    sumL0  +=       0.204 * cos(4.09939796199  + 14919.01785375460 *T)
    sumL0  +=       0.176 * cos(3.82400982460  + 9360.08916445900 *T)
    sumL0  +=       0.198 * cos(2.76491873243  + 10217.21769947410 *T)
    sumL0  +=       0.168 * cos(5.19268384202  + 1066.49547719000 *T)
    sumL0  +=       0.199 * cos(1.95301487982  + 7564.83072073800 *T)
    sumL0  +=       0.171 * cos(2.59623459612  + 20405.79569692960 *T)
    sumL0  +=       0.172 * cos(5.29332132623  + 11764.33076885900 *T)
    sumL0  +=       0.165 * cos(2.88557908025  + 10207.76262190360 *T)
    sumL0  +=       0.164 * cos(3.25435371801  + 3914.95722503460 *T)
    sumL0  +=       0.200 * cos(3.82443218090  + 18314.92077904440 *T)
    sumL0  +=       0.169 * cos(1.78341902878  + 31022.75317085620 *T)
    sumL0  +=       0.179 * cos(0.90840065587  + 7880.08915333899 *T)
    sumL0  +=       0.163 * cos(2.79665037814  + 41.55079098480 *T)
    sumL0  +=       0.154 * cos(3.90796293476  + 30213.25844775700 *T)
    sumL0  +=       0.153 * cos(0.07463240782  + 28528.20632525540 *T)
    sumL0  +=       0.194 * cos(5.95838706838  + 8617.09910392640 *T)
    sumL0  +=       0.171 * cos(4.58206324409  + 20447.34648791440 *T)
    sumL0  +=       0.150 * cos(2.11647586229  + 17248.42530185440 *T)
    sumL0  +=       0.149 * cos(2.17259986320  + 9929.42622734580 *T)
    sumL0  +=       0.191 * cos(0.82310353823  + 52670.06959330260 *T)
    sumL0  +=       0.148 * cos(2.94315921485  + 41654.96311596780 *T)
    sumL0  +=       0.149 * cos(4.49798039726  + 30831.30490474460 *T)
    sumL0  +=       0.184 * cos(2.46923348701  + 34596.36465465240 *T)
    sumL0  +=       0.146 * cos(2.69452930300  + 43071.89928903080 *T)
    sumL0  +=       0.159 * cos(2.11137713570  + 19317.19254032860 *T)
    sumL0  +=       0.154 * cos(2.76536164654  + 28513.97923125379 *T)
    sumL0  +=       0.140 * cos(4.94595038686  + 9256.99639024040 *T)
    sumL0  +=       0.141 * cos(2.57248458154  + 13553.89797291080 *T)
    sumL0  +=       0.137 * cos(1.66482327575  + 2636.72547263700 *T)
    sumL0  +=       0.140 * cos(5.23039605990  + 22645.32819660879 *T)
    sumL0  +=       0.132 * cos(5.35690599728  + 19624.75016129820 *T)
    sumL0  +=       0.140 * cos(2.90637712614  + 48947.66387067660 *T)
    sumL0  +=       0.129 * cos(3.95303623681  + 32858.61374281979 *T)
    sumL0  +=       0.156 * cos(6.01143316387  + 29057.89729034999 *T)
    sumL0  +=       0.134 * cos(5.75241675118  + 68050.42387851159 *T)
    sumL0  +=       0.154 * cos(3.66827363753  + 276.74577186440 *T)
    sumL0  +=       0.176 * cos(3.77298381177  + 66813.56483573320 *T)
    sumL0  +=       0.126 * cos(5.00217740223  + 27461.71084806540 *T)
    sumL0  +=       0.135 * cos(1.34807013920  + 53285.18483524180 *T)
    sumL0  +=       0.150 * cos(0.25029475344  + 290.97286586600 *T)
    sumL0  +=       0.152 * cos(3.13035670092  + 29043.67019634839 *T)
    sumL0  +=       0.169 * cos(5.04348109430  + 73.29712585900 *T)
    sumL0  +=       0.166 * cos(5.39219948035  + 41236.03871706720 *T)
    sumL0  +=       0.163 * cos(5.59796070987  + 7576.56007357400 *T)
    sumL0  +=       0.126 * cos(0.77391784606  + 49.99662190420 *T)
    sumL0  +=       0.163 * cos(0.44241846674  + 20350.30502114640 *T)
    sumL0  +=       0.136 * cos(3.09066368912  + 418.92439890060 *T)
    sumL0  +=       0.154 * cos(0.47086190960  + 28418.00000403600 *T)
    sumL0  +=       0.120 * cos(0.88536981986  + 29573.36116144300 *T)
    sumL0  +=       0.132 * cos(1.48009769040  + 17085.95866572220 *T)
    sumL0  +=       0.126 * cos(1.39497760964  + 966.97087743560 *T)
    sumL0  +=       0.143 * cos(3.84026797958  + 14128.24277124560 *T)
    sumL0  +=       0.147 * cos(2.11627427804  + 34363.36559755600 *T)
    sumL0  +=       0.106 * cos(2.04696932293  + 37674.99639427640 *T)
    sumL0  +=       0.106 * cos(1.43873202489  + 27682.14074415640 *T)
    sumL0  +=       0.149 * cos(0.09286508794  + 8144.27871130440 *T)
    sumL0  +=       0.103 * cos(0.01992041470  + 18300.69368504279 *T)
    sumL0  +=       0.121 * cos(3.57602835443  + 45.14121963660 *T)
    sumL0  +=       0.125 * cos(0.11630302078  + 149.56319713460 *T)
    sumL0  +=       0.102 * cos(4.17947097730  + 2333.19639287200 *T)
    sumL0  +=       0.099 * cos(1.51324741657  + 10419.47109464820 *T)
    sumL0  +=       0.133 * cos(3.02183293676  + 76251.32777062019 *T)
    sumL0  +=       0.136 * cos(4.17517197268  + 3646.35037735440 *T)
    sumL0  +=       0.123 * cos(0.44045588682  + 515.46387109300 *T)
    sumL0  +=       0.113 * cos(5.69261397718  + 10110.19277199240 *T)
    sumL0  +=       0.098 * cos(6.23797900467  + 202.25339517410 *T)
    sumL0  +=       0.099 * cos(3.75627530197  + 59728.66805461800 *T)
    sumL0  +=       0.101 * cos(4.62832557536  + 65236.22129328540 *T)
    sumL0  +=       0.111 * cos(1.25947267588  + 10846.06928552420 *T)
    sumL0  +=       0.110 * cos(5.87455577536  + 38500.27603107220 *T)
    sumL0  +=       0.128 * cos(6.01024562160  + 90394.82301305080 *T)
    sumL0  +=       0.091 * cos(1.77665981007  + 1539.31586981200 *T)
    sumL0  +=       0.092 * cos(0.99804571578  + 95.97922721780 *T)
    sumL0  +=       0.120 * cos(3.93060866244  + 38526.57435087200 *T)
    sumL0  +=       0.117 * cos(2.24143299549  + 56600.27928952220 *T)
    sumL0  +=       0.118 * cos(6.09121325940  + 29786.66025688100 *T)
    sumL0  +=       0.098 * cos(4.60938156207  + 11787.10597030980 *T)
    sumL0  +=       0.097 * cos(3.92727733144  + 11794.15220700780 *T)
    sumL0  +=       0.093 * cos(5.23395435043  + 14169.79356223040 *T)
    sumL0  +=       0.096 * cos(5.27525709038  + 8734.41897214660 *T)
    sumL0  +=       0.094 * cos(0.18166654805  + 67589.08744705719 *T)
    sumL0  +=       0.110 * cos(4.96279287076  + 48417.97290558199 *T)
    sumL0  +=       0.086 * cos(0.39533409505  + 3956.50801601940 *T)
    sumL0  +=       0.085 * cos(5.69642646462  + 37703.97802425580 *T)
    sumL0  +=       0.114 * cos(5.19676285428  + 70743.77453195279 *T)
    sumL0  +=       0.081 * cos(5.51324815184  + 412.37109687440 *T)
    sumL0  +=       0.089 * cos(2.13409771828  + 44768.09940987860 *T)
    sumL0  +=       0.084 * cos(6.02475904578  + 10632.77019008620 *T)
    sumL0  +=       0.085 * cos(4.60912614442  + 45585.17281218740 *T)
    sumL0  +=       0.078 * cos(4.47358603432  + 114.43928868521 *T)
    sumL0  +=       0.097 * cos(4.02223363535  + 10218.80847051840 *T)
    sumL0  +=       0.081 * cos(1.03870237004  + 9793.80090233580 *T)
    sumL0  +=       0.092 * cos(0.80301220092  + 24383.07910844140 *T)
    sumL0  +=       0.087 * cos(2.15124790938  + 28313.28880466100 *T)
    sumL0  +=       0.075 * cos(5.17868679355  + 63658.87775083760 *T)
    sumL0  +=       0.078 * cos(5.81927313665  + 567.71863773040 *T)
    sumL0  +=       0.075 * cos(1.72618192481  + 19580.48825767080 *T)
    sumL0  +=       0.071 * cos(0.10259261764  + 90695.75207512038 *T)
    sumL0  +=       0.077 * cos(6.16012067704  + 1573.82042409880 *T)
    sumL0  +=       0.076 * cos(5.12884307551  + 49515.38250840700 *T)
    sumL0  +=       0.069 * cos(0.29569499484  + 10175.25787357520 *T)
    sumL0  +=       0.061 * cos(4.80385549281  + 19889.76658032659 *T)
    sumL0  +=       0.060 * cos(4.56685040226  + 30426.55754319500 *T)
    sumL0  +=       0.062 * cos(4.16222812699  + 42430.48572729180 *T)
    sumL0  +=       0.050 * cos(6.17899839001  + 22779.43724619380 *T)
    sumL0  +=       0.048 * cos(1.52546758016  + 20639.87018786000 *T)
    sumL0  +=       0.046 * cos(4.41738494249  + 34570.06633485260 *T)
    sumL0  +=       0.037 * cos(4.69675087759  + 44007.82926973960 *T)

    # VENUS Longitude
    sumL1  = 0
    sumL1  +=  1021352943052.898 * cos(0.00000000000  + 0.00000000000 *T)
    sumL1  +=   95707.712 * cos(2.46424448979  + 10213.28554621100 *T)
    sumL1  +=   14444.977 * cos(0.51624564679  + 20426.57109242200 *T)
    sumL1  +=     213.374 * cos(1.79547929368  + 30639.85663863300 *T)
    sumL1  +=     151.669 * cos(6.10635282369  + 1577.34354244780 *T)
    sumL1  +=     173.904 * cos(2.65535879443  + 26.29831979980 *T)
    sumL1  +=      82.233 * cos(5.70234133730  + 191.44826611160 *T)
    sumL1  +=      69.734 * cos(2.68136034979  + 9437.76293488700 *T)
    sumL1  +=      52.408 * cos(3.60013087656  + 775.52261132400 *T)
    sumL1  +=      38.318 * cos(1.03379038025  + 529.69096509460 *T)
    sumL1  +=      29.633 * cos(1.25056322354  + 5507.55323866740 *T)
    sumL1  +=      25.056 * cos(6.10664792855  + 10404.73381232260 *T)
    sumL1  +=      17.772 * cos(6.19369798901  + 1109.37855209340 *T)
    sumL1  +=      16.510 * cos(2.64330452640  + 7.11354700080 *T)
    sumL1  +=      14.230 * cos(5.45138233941  + 9153.90361602180 *T)
    sumL1  +=      12.607 * cos(1.24464400689  + 40853.14218484400 *T)
    sumL1  +=      11.627 * cos(4.97604495371  + 213.29909543800 *T)
    sumL1  +=      12.563 * cos(1.88122199199  + 382.89653222320 *T)
    sumL1  +=       8.869 * cos(0.95282732248  + 13367.97263110660 *T)
    sumL1  +=       7.374 * cos(4.39476760580  + 10206.17199921020 *T)
    sumL1  +=       6.552 * cos(2.28168808058  + 2352.86615377180 *T)
    sumL1  +=       6.255 * cos(4.08056644034  + 3154.68708489560 *T)
    sumL1  +=       6.697 * cos(5.05673427795  + 801.82093112380 *T)
    sumL1  +=       4.084 * cos(4.12103826030  + 18837.49819713819 *T)
    sumL1  +=       4.882 * cos(3.44515199115  + 11015.10647733480 *T)
    sumL1  +=       3.549 * cos(6.19934345402  + 5.52292430740 *T)
    sumL1  +=       3.448 * cos(1.77405651704  + 11322.66409830440 *T)
    sumL1  +=       4.290 * cos(0.08154809210  + 6283.07584999140 *T)
    sumL1  +=       3.694 * cos(2.48453945256  + 5661.33204915220 *T)
    sumL1  +=       3.555 * cos(1.48036949420  + 1059.38193018920 *T)
    sumL1  +=       3.023 * cos(2.24092938317  + 18073.70493865020 *T)
    sumL1  +=       3.000 * cos(0.39169917698  + 15.25247118500 *T)
    sumL1  +=       2.563 * cos(0.35147506973  + 22003.91463486980 *T)
    sumL1  +=       2.774 * cos(1.45683830639  + 10239.58386601080 *T)
    sumL1  +=       2.951 * cos(5.34618097429  + 7084.89678111520 *T)
    sumL1  +=       2.344 * cos(2.36652432105  + 17298.18232732620 *T)
    sumL1  +=       2.405 * cos(2.36085282088  + 10596.18207843420 *T)
    sumL1  +=       1.720 * cos(4.72129626061  + 10186.98722641120 *T)
    sumL1  +=       2.209 * cos(2.07730338665  + 8635.94200376320 *T)
    sumL1  +=       2.122 * cos(4.47091605309  + 8624.21265092720 *T)
    sumL1  +=       1.527 * cos(0.67146857292  + 14143.49524243060 *T)
    sumL1  +=       1.473 * cos(2.59350470099  + 7064.12138562280 *T)
    sumL1  +=       1.311 * cos(0.90408820221  + 12566.15169998280 *T)
    sumL1  +=       1.474 * cos(5.92236241437  + 9786.68735533500 *T)
    sumL1  +=       1.237 * cos(2.59740787132  + 4551.95349705880 *T)
    sumL1  +=       1.219 * cos(2.83617320088  + 9676.48103411560 *T)
    sumL1  +=       1.116 * cos(3.83715584719  + 21228.39202354580 *T)
    sumL1  +=       1.006 * cos(4.26200749078  + 426.59819087600 *T)
    sumL1  +=       1.150 * cos(2.35531987378  + 9690.70812811720 *T)
    sumL1  +=       1.219 * cos(2.27324315182  + 522.57741809380 *T)
    sumL1  +=       1.150 * cos(0.81088598778  + 10742.97651130560 *T)
    sumL1  +=       1.101 * cos(3.74248783564  + 18307.80723204360 *T)
    sumL1  +=       1.031 * cos(2.03889374176  + 38.02767263580 *T)
    sumL1  +=       0.971 * cos(6.10590045414  + 3532.06069281140 *T)
    sumL1  +=       0.844 * cos(4.75124127613  + 10988.80815753500 *T)
    sumL1  +=       0.908 * cos(1.06613723738  + 10021.83728009940 *T)
    sumL1  +=       0.824 * cos(0.23090829723  + 28286.99048486120 *T)
    sumL1  +=       0.821 * cos(2.60456032773  + 19.66976089979 *T)
    sumL1  +=       0.728 * cos(0.10716917942  + 4705.73230754360 *T)
    sumL1  +=       0.744 * cos(3.33129778857  + 536.80451209540 *T)
    sumL1  +=       0.816 * cos(1.27303930175  + 19896.88012732740 *T)
    sumL1  +=       0.929 * cos(1.08024621325  + 11790.62908865880 *T)
    sumL1  +=       0.797 * cos(2.23891816523  + 3723.50895892300 *T)
    sumL1  +=       0.704 * cos(5.95307260017  + 20.77539549240 *T)
    sumL1  +=       0.665 * cos(0.21346689192  + 7860.41939243920 *T)
    sumL1  +=       0.733 * cos(2.22147883292  + 19360.07561523199 *T)
    sumL1  +=       0.702 * cos(1.76206343944  + 19374.30270923360 *T)
    sumL1  +=       0.575 * cos(2.38792087791  + 6770.71060124560 *T)
    sumL1  +=       0.538 * cos(1.52023264138  + 25934.12433108940 *T)
    sumL1  +=       0.690 * cos(4.01873754171  + 19651.04848109800 *T)
    sumL1  +=       0.532 * cos(4.41576130890  + 574.34479833480 *T)
    sumL1  +=       0.540 * cos(2.15936134728  + 16496.36139620240 *T)
    sumL1  +=       0.576 * cos(5.41170044566  + 206.18554843720 *T)
    sumL1  +=       0.482 * cos(0.40815793538  + 3340.61242669980 *T)
    sumL1  +=       0.501 * cos(3.08578363577  + 245.83164622940 *T)
    sumL1  +=       0.488 * cos(5.22311611589  + 25158.60171976540 *T)
    sumL1  +=       0.450 * cos(0.21279844600  + 11.72935283600 *T)
    sumL1  +=       0.432 * cos(1.32004964493  + 103.09277421860 *T)
    sumL1  +=       0.434 * cos(5.91094755233  + 19786.67380610799 *T)
    sumL1  +=       0.564 * cos(0.38776462529  + 19367.18916223280 *T)
    sumL1  +=       0.421 * cos(2.71057839701  + 13936.79450513400 *T)
    sumL1  +=       0.549 * cos(6.08792865644  + 3930.20969621960 *T)
    sumL1  +=       0.478 * cos(4.70234715828  + 14945.31617355440 *T)
    sumL1  +=       0.408 * cos(4.80890663927  + 32217.20018108080 *T)
    sumL1  +=       0.404 * cos(2.85003595942  + 29864.33402730900 *T)
    sumL1  +=       0.407 * cos(2.94002049006  + 10220.39909321180 *T)
    sumL1  +=       0.359 * cos(0.72354778897  + 419.48464387520 *T)
    sumL1  +=       0.449 * cos(1.44520508753  + 8662.24032356300 *T)
    sumL1  +=       0.353 * cos(2.22195492336  + 51066.42773105500 *T)
    sumL1  +=       0.324 * cos(1.40308439067  + 29580.47470844380 *T)
    sumL1  +=       0.443 * cos(1.93864353398  + 9146.79006902100 *T)
    sumL1  +=       0.314 * cos(0.96837035284  + 20618.01935853360 *T)
    sumL1  +=       0.324 * cos(5.10759068171  + 24356.78078864160 *T)
    sumL1  +=       0.324 * cos(1.80146948625  + 18830.38465013739 *T)
    sumL1  +=       0.370 * cos(6.16895004656  + 2218.75710418680 *T)
    sumL1  +=       0.278 * cos(2.20429108375  + 18844.61174413899 *T)
    sumL1  +=       0.286 * cos(3.08459438435  + 17277.40693183380 *T)
    sumL1  +=       0.383 * cos(0.13890934755  + 4732.03062734340 *T)
    sumL1  +=       0.292 * cos(0.43528982259  + 29088.81141598500 *T)
    sumL1  +=       0.273 * cos(5.84415407168  + 9573.38825989700 *T)
    sumL1  +=       0.324 * cos(2.14144542781  + 9999.98645077300 *T)
    sumL1  +=       0.264 * cos(5.20407029554  + 220.41264243880 *T)
    sumL1  +=       0.254 * cos(0.34411959301  + 28521.09277825460 *T)
    sumL1  +=       0.300 * cos(3.76014360906  + 8094.52168583260 *T)
    sumL1  +=       0.301 * cos(3.64457981649  + 20400.27277262220 *T)
    sumL1  +=       0.287 * cos(1.84003536598  + 1589.07289528380 *T)
    sumL1  +=       0.206 * cos(0.97167234723  + 10234.06094170340 *T)
    sumL1  +=       0.212 * cos(0.24173677600  + 36.02786667740 *T)
    sumL1  +=       0.216 * cos(5.88618923030  + 18875.52586977400 *T)
    sumL1  +=       0.198 * cos(1.89506914939  + 20452.86941222180 *T)
    sumL1  +=       0.258 * cos(6.27611355094  + 1551.04522264800 *T)
    sumL1  +=       0.197 * cos(2.09222675324  + 9683.59458111640 *T)
    sumL1  +=       0.217 * cos(5.79472589364  + 9103.90699411760 *T)
    sumL1  +=       0.188 * cos(0.39123199129  + 19573.37471066999 *T)
    sumL1  +=       0.195 * cos(6.23142464829  + 30110.16567353840 *T)
    sumL1  +=       0.187 * cos(5.49670351645  + 170.67287061920 *T)
    sumL1  +=       0.178 * cos(4.90042854659  + 10787.63034454580 *T)
    sumL1  +=       0.188 * cos(1.62614804098  + 9161.01716302260 *T)
    sumL1  +=       0.211 * cos(2.71884568392  + 15720.83878487840 *T)
    sumL1  +=       0.177 * cos(1.88170417337  + 33019.02111220460 *T)
    sumL1  +=       0.209 * cos(2.66033422116  + 3442.57494496540 *T)
    sumL1  +=       0.164 * cos(4.92240093026  + 10426.58464164900 *T)
    sumL1  +=       0.186 * cos(5.13678812068  + 7255.56965173440 *T)
    sumL1  +=       0.177 * cos(5.70206821967  + 9992.87290377220 *T)
    sumL1  +=       0.214 * cos(2.70027196648  + 3128.38876509580 *T)
    sumL1  +=       0.208 * cos(3.38876526854  + 17778.11626694899 *T)
    sumL1  +=       0.147 * cos(4.25008782855  + 16983.99614745660 *T)
    sumL1  +=       0.148 * cos(3.46404418130  + 21202.09370374600 *T)
    sumL1  +=       0.189 * cos(1.43553862242  + 2379.16447357160 *T)
    sumL1  +=       0.139 * cos(2.99154379541  + 110.20632121940 *T)
    sumL1  +=       0.159 * cos(5.23851679605  + 10007.09999777380 *T)
    sumL1  +=       0.136 * cos(0.88942869764  + 22805.73556599360 *T)
    sumL1  +=       0.155 * cos(5.90500835975  + 12592.45001978260 *T)
    sumL1  +=       0.151 * cos(0.03422618975  + 27991.40181316000 *T)
    sumL1  +=       0.153 * cos(4.01743770323  + 33794.54372352860 *T)
    sumL1  +=       0.121 * cos(0.51392111799  + 21535.94964451540 *T)
    sumL1  +=       0.109 * cos(2.25388616761  + 26735.94526221320 *T)
    sumL1  +=       0.109 * cos(0.78612823474  + 6681.22485339960 *T)
    sumL1  +=       0.122 * cos(4.84805105466  + 19992.85935454519 *T)
    sumL1  +=       0.112 * cos(3.31796669604  + 36949.23080842420 *T)
    sumL1  +=       0.106 * cos(3.34507236765  + 10103.07922499160 *T)
    sumL1  +=       0.114 * cos(4.36384000196  + 20007.08644854680 *T)
    sumL1  +=       0.098 * cos(5.07711736751  + 135.62532501000 *T)
    sumL1  +=       0.120 * cos(5.41870615047  + 37724.75341974820 *T)
    sumL1  +=       0.103 * cos(2.62610244425  + 20213.27199698400 *T)
    sumL1  +=       0.085 * cos(5.04808202087  + 9830.38901398780 *T)
    sumL1  +=       0.103 * cos(2.01549383816  + 45585.17281218740 *T)
    sumL1  +=       0.088 * cos(2.62613816931  + 21.85082932640 *T)
    sumL1  +=       0.084 * cos(3.50355880173  + 639.89728631400 *T)
    sumL1  +=       0.099 * cos(0.61079620895  + 41654.96311596780 *T)
    sumL1  +=       0.088 * cos(3.63836700262  + 49515.38250840700 *T)
    sumL1  +=       0.098 * cos(2.42401801881  + 23581.25817731760 *T)
    sumL1  +=       0.081 * cos(0.46468679835  + 77.67377042800 *T)
    sumL1  +=       0.092 * cos(4.82530051729  + 29043.67019634839 *T)
    sumL1  +=       0.102 * cos(4.27051236894  + 15874.61759536320 *T)
    sumL1  +=       0.090 * cos(4.34075776744  + 29057.89729034999 *T)
    sumL1  +=       0.081 * cos(0.01896422336  + 24150.08005134500 *T)
    sumL1  +=       0.093 * cos(1.79250830018  + 12432.04265039780 *T)
    sumL1  +=       0.087 * cos(5.25157021446  + 14128.24277124560 *T)
    sumL1  +=       0.089 * cos(5.65756996753  + 377.37360791580 *T)
    sumL1  +=       0.097 * cos(5.67942873241  + 227.47613278900 *T)
    sumL1  +=       0.076 * cos(2.93363913259  + 38204.68735937100 *T)
    sumL1  +=       0.091 * cos(2.60544242067  + 1052.26838318840 *T)
    sumL1  +=       0.087 * cos(3.82284200928  + 27511.46787353720 *T)
    sumL1  +=       0.073 * cos(4.75280755154  + 40879.44050464380 *T)
    sumL1  +=       0.067 * cos(3.54815262526  + 30666.15495843280 *T)
    sumL1  +=       0.067 * cos(5.81350818057  + 20809.46762464520 *T)
    sumL1  +=       0.064 * cos(4.24772678145  + 153.77881048480 *T)
    sumL1  +=       0.064 * cos(2.99454749109  + 27197.28169366760 *T)
    sumL1  +=       0.070 * cos(4.03868009742  + 56600.27928952220 *T)
    sumL1  +=       0.071 * cos(4.33628806850  + 39264.06928956020 *T)
    sumL1  +=       0.069 * cos(1.73648747605  + 37410.56723987860 *T)
    sumL1  +=       0.065 * cos(1.08206062736  + 68050.42387851159 *T)
    sumL1  +=       0.062 * cos(4.77698454650  + 3914.95722503460 *T)
    sumL1  +=       0.061 * cos(4.96121014691  + 34596.36465465240 *T)
    sumL1  +=       0.063 * cos(5.04865067599  + 53445.59220462660 *T)
    sumL1  +=       0.058 * cos(3.74010494151  + 1066.49547719000 *T)
    sumL1  +=       0.057 * cos(5.39355890141  + 20419.45754542119 *T)
    sumL1  +=       0.057 * cos(3.59399518494  + 735.87651353180 *T)
    sumL1  +=       0.065 * cos(2.10322000074  + 74.78159856730 *T)
    sumL1  +=       0.073 * cos(1.31083648835  + 11272.66747640020 *T)
    sumL1  +=       0.055 * cos(1.33161298098  + 18300.69368504279 *T)
    sumL1  +=       0.065 * cos(4.21150522641  + 49.75702547180 *T)
    sumL1  +=       0.061 * cos(5.66161679402  + 17468.85519794540 *T)
    sumL1  +=       0.053 * cos(4.30231233835  + 18849.22754997420 *T)
    sumL1  +=       0.055 * cos(2.63906959481  + 52670.06959330260 *T)
    sumL1  +=       0.050 * cos(5.69803054279  + 39793.76025465480 *T)
    sumL1  +=       0.049 * cos(0.77345264124  + 35371.88726597640 *T)
    sumL1  +=       0.048 * cos(6.00565977593  + 283.85931886520 *T)
    sumL1  +=       0.047 * cos(2.63299859494  + 51868.24866217880 *T)
    sumL1  +=       0.046 * cos(0.05105081843  + 38526.57435087200 *T)
    sumL1  +=       0.050 * cos(4.37549274002  + 28513.97923125379 *T)
    sumL1  +=       0.046 * cos(2.93422086586  + 27682.14074415640 *T)
    sumL1  +=       0.051 * cos(5.45979584751  + 60530.48898574180 *T)
    sumL1  +=       0.045 * cos(5.59492908223  + 467.96499035440 *T)
    sumL1  +=       0.045 * cos(2.34680401001  + 9411.46461508720 *T)
    sumL1  +=       0.045 * cos(0.02999265111  + 44809.65020086340 *T)
    sumL1  +=       0.043 * cos(5.62725673544  + 14.22709400160 *T)
    sumL1  +=       0.047 * cos(3.73567275749  + 64460.69868196140 *T)
    sumL1  +=       0.046 * cos(0.12586526756  + 57375.80190084620 *T)
    sumL1  +=       0.044 * cos(2.03114426076  + 18314.92077904440 *T)
    sumL1  +=       0.039 * cos(0.99375127466  + 94138.32702008578 *T)
    sumL1  +=       0.053 * cos(0.41974404621  + 30831.30490474460 *T)
    sumL1  +=       0.055 * cos(1.38351566741  + 38500.27603107220 *T)
    sumL1  +=       0.041 * cos(4.47012768909  + 40077.61957352000 *T)
    sumL1  +=       0.041 * cos(0.36665992484  + 19999.97290154599 *T)
    sumL1  +=       0.040 * cos(3.06358586355  + 813.55028395980 *T)
    sumL1  +=       0.040 * cos(2.16802870803  + 59728.66805461800 *T)
    sumL1  +=       0.037 * cos(1.08739100421  + 17085.95866572220 *T)
    sumL1  +=       0.039 * cos(1.31040309875  + 48739.85989708300 *T)
    sumL1  +=       0.036 * cos(1.43280677914  + 42456.78404709160 *T)
    sumL1  +=       0.037 * cos(0.14190533464  + 29050.78374334920 *T)
    sumL1  +=       0.037 * cos(3.66792179278  + 20956.26205751660 *T)
    sumL1  +=       0.025 * cos(3.38876180652  + 7058.59846131540 *T)
    sumL1  +=       0.031 * cos(6.16829805337  + 10192.51015071860 *T)

    # VENUS Longitude
    sumL2  = 0
    sumL2  +=   54127.076 * cos(0.00000000000  + 0.00000000000 *T)
    sumL2  +=    3891.460 * cos(0.34514360047  + 10213.28554621100 *T)
    sumL2  +=    1337.880 * cos(2.02011286082  + 20426.57109242200 *T)
    sumL2  +=      23.836 * cos(2.04592119012  + 26.29831979980 *T)
    sumL2  +=      19.331 * cos(3.53527371458  + 30639.85663863300 *T)
    sumL2  +=       9.984 * cos(3.97130221102  + 775.52261132400 *T)
    sumL2  +=       7.046 * cos(1.51962593409  + 1577.34354244780 *T)
    sumL2  +=       6.014 * cos(0.99926757893  + 191.44826611160 *T)
    sumL2  +=       3.163 * cos(4.36095475762  + 9437.76293488700 *T)
    sumL2  +=       2.125 * cos(2.65810625752  + 40853.14218484400 *T)
    sumL2  +=       1.934 * cos(3.39287946981  + 382.89653222320 *T)
    sumL2  +=       1.460 * cos(6.04899046273  + 529.69096509460 *T)
    sumL2  +=       1.346 * cos(2.94633106219  + 5507.55323866740 *T)
    sumL2  +=       1.025 * cos(1.40598904981  + 10404.73381232260 *T)
    sumL2  +=       1.221 * cos(3.73339139385  + 3154.68708489560 *T)
    sumL2  +=       1.033 * cos(3.52858472904  + 11015.10647733480 *T)
    sumL2  +=       0.955 * cos(5.11133878923  + 801.82093112380 *T)
    sumL2  +=       0.742 * cos(1.49198584483  + 1109.37855209340 *T)
    sumL2  +=       0.525 * cos(3.32087042103  + 213.29909543800 *T)
    sumL2  +=       0.578 * cos(0.92614279843  + 10239.58386601080 *T)
    sumL2  +=       0.602 * cos(5.19220099775  + 7084.89678111520 *T)
    sumL2  +=       0.431 * cos(2.67159914364  + 13367.97263110660 *T)
    sumL2  +=       0.389 * cos(4.14116341739  + 8635.94200376320 *T)
    sumL2  +=       0.355 * cos(1.12061570874  + 9153.90361602180 *T)
    sumL2  +=       0.301 * cos(3.90047984197  + 10596.18207843420 *T)
    sumL2  +=       0.212 * cos(5.32697688872  + 18837.49819713819 *T)
    sumL2  +=       0.260 * cos(0.22761369281  + 2352.86615377180 *T)
    sumL2  +=       0.243 * cos(4.70747902991  + 6283.07584999140 *T)
    sumL2  +=       0.196 * cos(4.10467294392  + 11790.62908865880 *T)
    sumL2  +=       0.194 * cos(6.01197759470  + 7860.41939243920 *T)
    sumL2  +=       0.140 * cos(4.97015671653  + 14143.49524243060 *T)
    sumL2  +=       0.134 * cos(4.10529011674  + 17298.18232732620 *T)
    sumL2  +=       0.119 * cos(3.39375528828  + 11322.66409830440 *T)
    sumL2  +=       0.126 * cos(0.09854516140  + 18073.70493865020 *T)
    sumL2  +=       0.122 * cos(5.92478855457  + 574.34479833480 *T)
    sumL2  +=       0.107 * cos(0.35660030184  + 1059.38193018920 *T)
    sumL2  +=       0.108 * cos(2.25352052666  + 12566.15169998280 *T)
    sumL2  +=       0.093 * cos(5.48716819776  + 10021.83728009940 *T)
    sumL2  +=       0.084 * cos(4.89744332968  + 18307.80723204360 *T)
    sumL2  +=       0.074 * cos(2.35354025573  + 426.59819087600 *T)
    sumL2  +=       0.093 * cos(4.99316908815  + 14945.31617355440 *T)
    sumL2  +=       0.069 * cos(3.86409065860  + 51066.42773105500 *T)
    sumL2  +=       0.082 * cos(5.36280178643  + 10186.98722641120 *T)
    sumL2  +=       0.077 * cos(3.75728548372  + 3723.50895892300 *T)
    sumL2  +=       0.063 * cos(5.39882267787  + 21228.39202354580 *T)
    sumL2  +=       0.056 * cos(4.11564786973  + 7064.12138562280 *T)
    sumL2  +=       0.056 * cos(6.26920407723  + 32217.20018108080 *T)
    sumL2  +=       0.060 * cos(5.02186497542  + 19367.18916223280 *T)
    sumL2  +=       0.058 * cos(5.13263709670  + 20400.27277262220 *T)
    sumL2  +=       0.051 * cos(4.52870390511  + 22003.91463486980 *T)
    sumL2  +=       0.041 * cos(3.83822107919  + 16496.36139620240 *T)
    sumL2  +=       0.041 * cos(3.36020411807  + 4705.73230754360 *T)
    sumL2  +=       0.043 * cos(5.98371820588  + 15720.83878487840 *T)
    sumL2  +=       0.047 * cos(0.18498155367  + 18875.52586977400 *T)
    sumL2  +=       0.038 * cos(0.52232581277  + 1551.04522264800 *T)
    sumL2  +=       0.039 * cos(5.05391675878  + 10742.97651130560 *T)
    sumL2  +=       0.036 * cos(3.16242472203  + 20452.86941222180 *T)
    sumL2  +=       0.035 * cos(5.17462577483  + 29088.81141598500 *T)
    sumL2  +=       0.035 * cos(3.47325394141  + 24356.78078864160 *T)
    sumL2  +=       0.031 * cos(4.74511706231  + 28521.09277825460 *T)
    sumL2  +=       0.029 * cos(0.19383091192  + 19896.88012732740 *T)
    sumL2  +=       0.033 * cos(1.80059867635  + 20618.01935853360 *T)
    sumL2  +=       0.024 * cos(0.14022912457  + 21202.09370374600 *T)
    sumL2  +=       0.022 * cos(4.73565067573  + 10988.80815753500 *T)
    sumL2  +=       0.018 * cos(0.74039763161  + 25158.60171976540 *T)
    sumL2  +=       0.019 * cos(1.53770232218  + 28286.99048486120 *T)
    sumL2  +=       0.014 * cos(1.49084059765  + 30110.16567353840 *T)
    sumL2  +=       0.013 * cos(4.72171283479  + 29864.33402730900 *T)
    sumL2  +=       0.013 * cos(5.79700427846  + 29580.47470844380 *T)
    sumL2  +=       0.014 * cos(3.69205225010  + 27511.46787353720 *T)

    # VENUS Longitude
    sumL3  = 0
    sumL3  +=     135.742 * cos(4.80389020993  + 10213.28554621100 *T)
    sumL3  +=      77.846 * cos(3.66876371591  + 20426.57109242200 *T)
    sumL3  +=      26.023 * cos(0.00000000000  + 0.00000000000 *T)
    sumL3  +=       1.214 * cos(5.31970006917  + 30639.85663863300 *T)
    sumL3  +=       0.254 * cos(4.15021671822  + 40853.14218484400 *T)
    sumL3  +=       0.008 * cos(5.55523563261  + 51066.42773105500 *T)
    sumL3  +=       0.008 * cos(1.40501229148  + 1577.34354244780 *T)
    sumL3  +=       0.006 * cos(1.27791479726  + 10404.73381232260 *T)
    sumL3  +=       0.006 * cos(5.76447068962  + 10239.58386601080 *T)

    # VENUS Longitude
    sumL4  = 0
    sumL4  +=     114.016 * cos(3.14159265359  + 0.00000000000 *T)
    sumL4  +=       3.209 * cos(5.20514170164  + 20426.57109242200 *T)
    sumL4  +=       1.714 * cos(2.51099591706  + 10213.28554621100 *T)
    sumL4  +=       0.050 * cos(0.71356059861  + 30639.85663863300 *T)
    sumL4  +=       0.023 * cos(5.68127607034  + 40853.14218484400 *T)

    # VENUS Longitude
    sumL5  = 0
    sumL5  +=       0.874 * cos(3.14159265359  + 0.00000000000 *T)
    sumL5  +=       0.117 * cos(0.54643013000  + 20426.57109242200 *T)
    sumL5  +=       0.118 * cos(1.90548541922  + 10213.28554621100 *T)
    sumL5  +=       0.002 * cos(1.07734277826  + 40853.14218484400 *T)
    sumL5  +=       0.002 * cos(1.89460223529  + 30639.85663863300 *T)

    # VENUS Latitude
    sumB0  = 0
    sumB0  +=  5923638.472 * cos(0.26702775813  + 10213.28554621100 *T)
    sumB0  +=   40107.978 * cos(1.14737178106  + 20426.57109242200 *T)
    sumB0  +=   32814.918 * cos(3.14159265359  + 0.00000000000 *T)
    sumB0  +=    1011.392 * cos(1.08946123021  + 30639.85663863300 *T)
    sumB0  +=     149.458 * cos(6.25390296069  + 18073.70493865020 *T)
    sumB0  +=     137.788 * cos(0.86020146523  + 1577.34354244780 *T)
    sumB0  +=     129.973 * cos(3.67152483651  + 9437.76293488700 *T)
    sumB0  +=     119.507 * cos(3.70468812804  + 2352.86615377180 *T)
    sumB0  +=     107.971 * cos(4.53903677647  + 22003.91463486980 *T)
    sumB0  +=      92.029 * cos(1.53954562706  + 9153.90361602180 *T)
    sumB0  +=      52.982 * cos(2.28138172277  + 5507.55323866740 *T)
    sumB0  +=      45.617 * cos(0.72319641722  + 10239.58386601080 *T)
    sumB0  +=      38.855 * cos(2.93437865147  + 10186.98722641120 *T)
    sumB0  +=      43.491 * cos(6.14015776699  + 11790.62908865880 *T)
    sumB0  +=      41.700 * cos(5.99126845246  + 19896.88012732740 *T)
    sumB0  +=      39.644 * cos(3.86842095901  + 8635.94200376320 *T)
    sumB0  +=      39.175 * cos(3.94960351174  + 529.69096509460 *T)
    sumB0  +=      33.320 * cos(4.83194909595  + 14143.49524243060 *T)
    sumB0  +=      23.711 * cos(2.90646621218  + 10988.80815753500 *T)
    sumB0  +=      23.500 * cos(2.00770618322  + 13367.97263110660 *T)
    sumB0  +=      21.809 * cos(2.69701424951  + 19651.04848109800 *T)
    sumB0  +=      20.653 * cos(0.98666685459  + 775.52261132400 *T)
    sumB0  +=      16.976 * cos(4.13711782135  + 10021.83728009940 *T)
    sumB0  +=      17.835 * cos(5.96268643102  + 25934.12433108940 *T)
    sumB0  +=      14.949 * cos(5.61075168206  + 10404.73381232260 *T)
    sumB0  +=      18.579 * cos(1.80529277514  + 40853.14218484400 *T)
    sumB0  +=      15.407 * cos(3.29563855296  + 11015.10647733480 *T)
    sumB0  +=      12.936 * cos(5.42651448496  + 29580.47470844380 *T)
    sumB0  +=      11.962 * cos(3.57604253827  + 10742.97651130560 *T)
    sumB0  +=      11.827 * cos(1.19070919600  + 8624.21265092720 *T)
    sumB0  +=      11.466 * cos(5.12780364967  + 6283.07584999140 *T)
    sumB0  +=       9.484 * cos(2.75167834335  + 191.44826611160 *T)
    sumB0  +=      13.129 * cos(5.70735942511  + 9683.59458111640 *T)
    sumB0  +=       8.583 * cos(0.43182249199  + 9786.68735533500 *T)
    sumB0  +=       9.763 * cos(0.14614896296  + 20618.01935853360 *T)
    sumB0  +=       8.148 * cos(1.30548515603  + 15720.83878487840 *T)
    sumB0  +=       6.050 * cos(6.26541665966  + 11322.66409830440 *T)
    sumB0  +=       5.955 * cos(4.92235372433  + 1059.38193018920 *T)
    sumB0  +=       6.983 * cos(3.44920932146  + 17298.18232732620 *T)
    sumB0  +=       6.228 * cos(1.13312070908  + 29864.33402730900 *T)
    sumB0  +=       6.186 * cos(4.92498052443  + 19367.18916223280 *T)
    sumB0  +=       6.155 * cos(2.42413946900  + 4705.73230754360 *T)
    sumB0  +=       5.204 * cos(3.42528906628  + 9103.90699411760 *T)
    sumB0  +=       6.000 * cos(3.57639095526  + 3154.68708489560 *T)
    sumB0  +=       4.796 * cos(3.86676184909  + 7860.41939243920 *T)
    sumB0  +=       5.289 * cos(4.99182712443  + 7084.89678111520 *T)
    sumB0  +=       4.070 * cos(5.58798144740  + 12566.15169998280 *T)
    sumB0  +=       3.942 * cos(5.68758787835  + 10206.17199921020 *T)
    sumB0  +=       3.797 * cos(3.89520601076  + 10192.51015071860 *T)
    sumB0  +=       3.798 * cos(6.06410995916  + 10234.06094170340 *T)
    sumB0  +=       3.579 * cos(0.73789669235  + 4551.95349705880 *T)
    sumB0  +=       3.641 * cos(2.61501257205  + 15874.61759536320 *T)
    sumB0  +=       3.266 * cos(0.97517223854  + 23581.25817731760 *T)
    sumB0  +=       2.813 * cos(0.29951755546  + 9411.46461508720 *T)
    sumB0  +=       3.048 * cos(2.51085146990  + 33794.54372352860 *T)
    sumB0  +=       2.559 * cos(4.58043833032  + 801.82093112380 *T)
    sumB0  +=       2.462 * cos(5.05790874754  + 29050.78374334920 *T)
    sumB0  +=       2.593 * cos(5.73113176751  + 20213.27199698400 *T)
    sumB0  +=       2.625 * cos(4.24272906574  + 213.29909543800 *T)
    sumB0  +=       2.246 * cos(0.82112963936  + 28286.99048486120 *T)
    sumB0  +=       2.229 * cos(2.22457598233  + 10426.58464164900 *T)
    sumB0  +=       1.742 * cos(1.48394229233  + 7058.59846131540 *T)
    sumB0  +=       1.660 * cos(5.42775825275  + 32217.20018108080 *T)
    sumB0  +=       1.491 * cos(4.64883377941  + 1109.37855209340 *T)
    sumB0  +=       2.010 * cos(0.75702888128  + 9999.98645077300 *T)
    sumB0  +=       1.562 * cos(3.93962080463  + 37724.75341974820 *T)
    sumB0  +=       1.538 * cos(2.17309577080  + 21535.94964451540 *T)
    sumB0  +=       1.546 * cos(4.70759186462  + 14945.31617355440 *T)
    sumB0  +=       1.200 * cos(1.48282382657  + 9830.38901398780 *T)
    sumB0  +=       1.224 * cos(5.55090394449  + 5661.33204915220 *T)
    sumB0  +=       1.111 * cos(1.20276209213  + 9573.38825989700 *T)
    sumB0  +=       1.064 * cos(1.98891375536  + 26.29831979980 *T)
    sumB0  +=       1.041 * cos(5.38535116069  + 7.11354700080 *T)
    sumB0  +=       1.036 * cos(1.16719443387  + 8662.24032356300 *T)
    sumB0  +=       1.143 * cos(3.20596958337  + 3532.06069281140 *T)
    sumB0  +=       1.201 * cos(0.81913312536  + 8094.52168583260 *T)
    sumB0  +=       1.005 * cos(2.38429892132  + 27511.46787353720 *T)
    sumB0  +=       1.047 * cos(4.56525030769  + 20419.45754542119 *T)
    sumB0  +=       0.968 * cos(6.18496721871  + 25158.60171976540 *T)
    sumB0  +=       1.044 * cos(1.98055689074  + 10596.18207843420 *T)
    sumB0  +=       0.962 * cos(0.48573513747  + 23958.63178523340 *T)
    sumB0  +=       0.846 * cos(0.01566400887  + 3128.38876509580 *T)
    sumB0  +=       0.792 * cos(5.39686899735  + 24356.78078864160 *T)
    sumB0  +=       0.858 * cos(5.34692750735  + 41654.96311596780 *T)
    sumB0  +=       0.757 * cos(6.25904553773  + 20452.86941222180 *T)
    sumB0  +=       0.801 * cos(4.62406152514  + 9929.42622734580 *T)
    sumB0  +=       0.802 * cos(5.37234892520  + 10497.14486507620 *T)
    sumB0  +=       0.750 * cos(3.85219782842  + 21228.39202354580 *T)
    sumB0  +=       0.700 * cos(1.98097957188  + 3930.20969621960 *T)
    sumB0  +=       0.719 * cos(6.11596800207  + 10218.80847051840 *T)
    sumB0  +=       0.672 * cos(6.23429601219  + 14765.23904326980 *T)
    sumB0  +=       0.639 * cos(5.37566437358  + 1589.07289528380 *T)
    sumB0  +=       0.605 * cos(2.42330391120  + 10251.31321884680 *T)
    sumB0  +=       0.726 * cos(6.16683781802  + 18875.52586977400 *T)
    sumB0  +=       0.613 * cos(5.99731180690  + 4732.03062734340 *T)
    sumB0  +=       0.720 * cos(3.84286345199  + 10207.76262190360 *T)
    sumB0  +=       0.637 * cos(6.17053891156  + 10220.39909321180 *T)
    sumB0  +=       0.515 * cos(1.03001478293  + 22779.43724619380 *T)
    sumB0  +=       0.574 * cos(0.43813688572  + 17085.95866572220 *T)
    sumB0  +=       0.510 * cos(1.41065159851  + 9161.01716302260 *T)
    sumB0  +=       0.569 * cos(3.34601425125  + 3340.61242669980 *T)
    sumB0  +=       0.608 * cos(1.25236241968  + 10175.25787357520 *T)
    sumB0  +=       0.524 * cos(2.39794248670  + 26087.90314157420 *T)
    sumB0  +=       0.542 * cos(1.34665646732  + 29088.81141598500 *T)
    sumB0  +=       0.527 * cos(4.01994270827  + 18849.22754997420 *T)
    sumB0  +=       0.569 * cos(1.65498800378  + 39264.06928956020 *T)
    sumB0  +=       0.518 * cos(4.96996115446  + 30213.25844775700 *T)
    sumB0  +=       0.514 * cos(5.78413007838  + 12592.45001978260 *T)
    sumB0  +=       0.538 * cos(4.56198493922  + 10063.72234907640 *T)
    sumB0  +=       0.484 * cos(4.18538027381  + 14919.01785375460 *T)
    sumB0  +=       0.493 * cos(4.79939382739  + 9146.79006902100 *T)
    sumB0  +=       0.427 * cos(3.76876868949  + 11272.66747640020 *T)
    sumB0  +=       0.495 * cos(0.49175293655  + 45585.17281218740 *T)
    sumB0  +=       0.494 * cos(3.74345863918  + 31441.67756975680 *T)
    sumB0  +=       0.524 * cos(0.97991794166  + 30110.16567353840 *T)
    sumB0  +=       0.483 * cos(1.87898057316  + 51066.42773105500 *T)
    sumB0  +=       0.505 * cos(3.70047474212  + 20400.27277262220 *T)
    sumB0  +=       0.351 * cos(4.34026574490  + 10137.01947493540 *T)
    sumB0  +=       0.355 * cos(5.56672554631  + 18837.49819713819 *T)
    sumB0  +=       0.328 * cos(3.78427378910  + 6681.22485339960 *T)
    sumB0  +=       0.349 * cos(4.20550749672  + 20956.26205751660 *T)
    sumB0  +=       0.333 * cos(4.44969281739  + 28521.09277825460 *T)
    sumB0  +=       0.296 * cos(2.83205515646  + 17277.40693183380 *T)
    sumB0  +=       0.311 * cos(2.57334132897  + 20809.46762464520 *T)
    sumB0  +=       0.294 * cos(0.75089224483  + 3149.16416058820 *T)
    sumB0  +=       0.377 * cos(3.98143308775  + 21202.09370374600 *T)
    sumB0  +=       0.272 * cos(5.56183082489  + 16496.36139620240 *T)
    sumB0  +=       0.314 * cos(0.02584607093  + 13745.34623902240 *T)
    sumB0  +=       0.263 * cos(0.55328410985  + 36147.40987730040 *T)
    sumB0  +=       0.286 * cos(5.16408902215  + 426.59819087600 *T)
    sumB0  +=       0.279 * cos(4.29871615943  + 19999.97290154599 *T)
    sumB0  +=       0.280 * cos(1.92925047377  + 49515.38250840700 *T)
    sumB0  +=       0.265 * cos(4.81168402147  + 20235.12282631040 *T)
    sumB0  +=       0.273 * cos(5.12740051559  + 35371.88726597640 *T)
    sumB0  +=       0.306 * cos(5.28903794869  + 382.89653222320 *T)
    sumB0  +=       0.223 * cos(2.50591724074  + 26709.64694241340 *T)
    sumB0  +=       0.235 * cos(5.96522395118  + 10198.03307502600 *T)
    sumB0  +=       0.234 * cos(3.52866583267  + 10228.53801739600 *T)
    sumB0  +=       0.224 * cos(6.24561979789  + 7064.12138562280 *T)
    sumB0  +=       0.251 * cos(2.84739274245  + 33019.02111220460 *T)
    sumB0  +=       0.196 * cos(1.50610393790  + 31749.23519072640 *T)
    sumB0  +=       0.192 * cos(1.69321442572  + 13341.67431130680 *T)
    sumB0  +=       0.180 * cos(6.19353087076  + 39793.76025465480 *T)
    sumB0  +=       0.199 * cos(1.16433321880  + 22805.73556599360 *T)
    sumB0  +=       0.180 * cos(3.72646417141  + 1551.04522264800 *T)
    sumB0  +=       0.173 * cos(3.35235705827  + 53445.59220462660 *T)
    sumB0  +=       0.195 * cos(1.51901264131  + 43232.30665841560 *T)
    sumB0  +=       0.174 * cos(2.84049662693  + 9967.45389998160 *T)
    sumB0  +=       0.163 * cos(4.29160537719  + 36949.23080842420 *T)
    sumB0  +=       0.169 * cos(0.37000676558  + 10459.11719244040 *T)
    sumB0  +=       0.137 * cos(5.61149803116  + 10529.67741586760 *T)
    sumB0  +=       0.139 * cos(0.87847805052  + 16522.65971600220 *T)
    sumB0  +=       0.139 * cos(4.12576475427  + 36301.18868778519 *T)
    sumB0  +=       0.127 * cos(5.14447758616  + 5481.25491886760 *T)
    sumB0  +=       0.131 * cos(3.11317801589  + 9896.89367655440 *T)
    sumB0  +=       0.131 * cos(0.89697384735  + 3442.57494496540 *T)
    sumB0  +=       0.121 * cos(1.32802112907  + 38734.37832446560 *T)
    sumB0  +=       0.122 * cos(1.59017183044  + 10110.19277199240 *T)
    sumB0  +=       0.123 * cos(2.33714216061  + 10316.37832042960 *T)
    sumB0  +=       0.133 * cos(2.90682399304  + 9793.80090233580 *T)
    sumB0  +=       0.111 * cos(2.52077634760  + 13936.79450513400 *T)
    sumB0  +=       0.120 * cos(0.36076947165  + 536.80451209540 *T)
    sumB0  +=       0.115 * cos(2.53355582059  + 26735.94526221320 *T)
    sumB0  +=       0.108 * cos(2.65839634325  + 10232.95530711079 *T)
    sumB0  +=       0.108 * cos(0.55230439694  + 10193.61578531121 *T)
    sumB0  +=       0.138 * cos(1.06919239240  + 65236.22129328540 *T)
    sumB0  +=       0.101 * cos(3.17012502017  + 19317.19254032860 *T)
    sumB0  +=       0.127 * cos(5.63110477712  + 10288.06714477830 *T)
    sumB0  +=       0.127 * cos(3.86278127025  + 10138.50394764370 *T)
    sumB0  +=       0.137 * cos(2.93350659460  + 47162.51635463520 *T)
    sumB0  +=       0.095 * cos(5.03917884334  + 52175.80628314840 *T)
    sumB0  +=       0.094 * cos(0.71308489207  + 38500.27603107220 *T)
    sumB0  +=       0.092 * cos(5.46204624886  + 11764.33076885900 *T)
    sumB0  +=       0.096 * cos(1.52914774412  + 9690.70812811720 *T)
    sumB0  +=       0.101 * cos(0.83318284426  + 6489.77658728800 *T)
    sumB0  +=       0.115 * cos(3.76443612245  + 522.57741809380 *T)
    sumB0  +=       0.089 * cos(2.53312656681  + 10735.86296430480 *T)
    sumB0  +=       0.082 * cos(0.85628515615  + 2379.16447357160 *T)
    sumB0  +=       0.103 * cos(5.22683237620  + 103.09277421860 *T)
    sumB0  +=       0.090 * cos(2.12423586627  + 28313.28880466100 *T)
    sumB0  +=       0.090 * cos(0.39668501735  + 9580.50180689780 *T)
    sumB0  +=       0.074 * cos(6.02680095550  + 3723.50895892300 *T)
    sumB0  +=       0.081 * cos(5.25045057985  + 10419.47109464820 *T)
    sumB0  +=       0.080 * cos(4.23724598221  + 10007.09999777380 *T)
    sumB0  +=       0.091 * cos(2.48874147947  + 10846.06928552420 *T)
    sumB0  +=       0.085 * cos(3.82784790321  + 51868.24866217880 *T)
    sumB0  +=       0.081 * cos(2.26235214191  + 3903.91137641980 *T)
    sumB0  +=       0.097 * cos(0.77295091600  + 18307.80723204360 *T)
    sumB0  +=       0.094 * cos(0.17063414792  + 6872.67311951120 *T)
    sumB0  +=       0.080 * cos(5.62254102739  + 29999.95935231900 *T)
    sumB0  +=       0.068 * cos(2.71762936670  + 16983.99614745660 *T)
    sumB0  +=       0.066 * cos(0.76731351736  + 20.77539549240 *T)
    sumB0  +=       0.075 * cos(0.36155638007  + 39302.09696219600 *T)
    sumB0  +=       0.075 * cos(2.27327165974  + 8521.11987670860 *T)
    sumB0  +=       0.058 * cos(2.14482855875  + 8631.32619792800 *T)
    sumB0  +=       0.064 * cos(5.83569051301  + 2118.76386037840 *T)
    sumB0  +=       0.058 * cos(2.98524209824  + 19889.76658032659 *T)
    sumB0  +=       0.054 * cos(1.78260668333  + 40077.61957352000 *T)
    sumB0  +=       0.055 * cos(4.70485939861  + 639.89728631400 *T)
    sumB0  +=       0.060 * cos(5.89661892920  + 41962.52073693740 *T)
    sumB0  +=       0.066 * cos(2.24746237999  + 74.78159856730 *T)
    sumB0  +=       0.061 * cos(3.40726181591  + 27490.69247804480 *T)
    sumB0  +=       0.051 * cos(3.07811180039  + 24150.08005134500 *T)
    sumB0  +=       0.057 * cos(2.30081371235  + 20529.66386664059 *T)
    sumB0  +=       0.052 * cos(2.37192464233  + 29573.36116144300 *T)
    sumB0  +=       0.052 * cos(4.76610409132  + 57375.80190084620 *T)
    sumB0  +=       0.047 * cos(1.61630288856  + 30831.30490474460 *T)
    sumB0  +=       0.054 * cos(5.89684197257  + 19903.99367432819 *T)
    sumB0  +=       0.040 * cos(5.32101847424  + 42430.48572729180 *T)
    sumB0  +=       0.051 * cos(5.29186795569  + 29587.58825544460 *T)

    # VENUS Latitude
    sumB1  = 0
    sumB1  +=  513347.602 * cos(1.80364310797  + 10213.28554621100 *T)
    sumB1  +=    4380.100 * cos(3.38615711591  + 20426.57109242200 *T)
    sumB1  +=     196.586 * cos(2.53001197486  + 30639.85663863300 *T)
    sumB1  +=     199.162 * cos(0.00000000000  + 0.00000000000 *T)
    sumB1  +=      14.031 * cos(2.27087044687  + 9437.76293488700 *T)
    sumB1  +=      12.958 * cos(1.50735622957  + 18073.70493865020 *T)
    sumB1  +=      11.941 * cos(5.60462450426  + 1577.34354244780 *T)
    sumB1  +=      10.324 * cos(5.24224313355  + 2352.86615377180 *T)
    sumB1  +=       9.294 * cos(6.07545631303  + 22003.91463486980 *T)
    sumB1  +=       7.441 * cos(1.50257909439  + 11790.62908865880 *T)
    sumB1  +=       8.031 * cos(0.29371105198  + 9153.90361602180 *T)
    sumB1  +=       7.514 * cos(5.08081885990  + 10186.98722641120 *T)
    sumB1  +=       4.669 * cos(3.87801635015  + 10239.58386601080 *T)
    sumB1  +=       4.399 * cos(3.58872736593  + 40853.14218484400 *T)
    sumB1  +=       3.975 * cos(1.28397121206  + 10404.73381232260 *T)
    sumB1  +=       4.657 * cos(0.75073886819  + 5507.55323866740 *T)
    sumB1  +=       3.783 * cos(4.33004753984  + 19651.04848109800 *T)
    sumB1  +=       3.390 * cos(4.88976070903  + 10988.80815753500 *T)
    sumB1  +=       3.555 * cos(1.25927550356  + 19896.88012732740 *T)
    sumB1  +=       3.479 * cos(5.50797002160  + 529.69096509460 *T)
    sumB1  +=       2.884 * cos(0.08549582037  + 14143.49524243060 *T)
    sumB1  +=       1.786 * cos(0.37134513186  + 13367.97263110660 *T)
    sumB1  +=       1.600 * cos(1.68378002982  + 20618.01935853360 *T)
    sumB1  +=       1.539 * cos(1.21683853657  + 25934.12433108940 *T)
    sumB1  +=       1.341 * cos(2.90077139758  + 15720.83878487840 *T)
    sumB1  +=       0.993 * cos(1.74681248965  + 11322.66409830440 *T)
    sumB1  +=       1.165 * cos(6.13437155401  + 7860.41939243920 *T)
    sumB1  +=       1.115 * cos(0.66743690380  + 29580.47470844380 *T)
    sumB1  +=       0.923 * cos(2.25384969096  + 10021.83728009940 *T)
    sumB1  +=       0.965 * cos(1.36425494833  + 9683.59458111640 *T)
    sumB1  +=       0.973 * cos(0.39071758442  + 6283.07584999140 *T)
    sumB1  +=       0.805 * cos(0.53331923557  + 8624.21265092720 *T)
    sumB1  +=       0.913 * cos(0.76046003719  + 8635.94200376320 *T)
    sumB1  +=       0.991 * cos(0.55319879330  + 19367.18916223280 *T)
    sumB1  +=       0.609 * cos(2.62364470139  + 23581.25817731760 *T)
    sumB1  +=       0.532 * cos(5.10925676528  + 9786.68735533500 *T)
    sumB1  +=       0.476 * cos(6.17672999981  + 11015.10647733480 *T)
    sumB1  +=       0.472 * cos(1.69672629200  + 17298.18232732620 *T)
    sumB1  +=       0.503 * cos(2.65840772485  + 29864.33402730900 *T)
    sumB1  +=       0.456 * cos(5.01205315518  + 10742.97651130560 *T)
    sumB1  +=       0.478 * cos(3.94100005156  + 775.52261132400 *T)
    sumB1  +=       0.477 * cos(3.71554345922  + 10596.18207843420 *T)
    sumB1  +=       0.347 * cos(2.34551062680  + 9411.46461508720 *T)
    sumB1  +=       0.458 * cos(2.31894399069  + 9999.98645077300 *T)
    sumB1  +=       0.374 * cos(3.76878356974  + 21228.39202354580 *T)
    sumB1  +=       0.440 * cos(4.33400244581  + 15874.61759536320 *T)
    sumB1  +=       0.349 * cos(1.31468836511  + 10234.06094170340 *T)
    sumB1  +=       0.310 * cos(5.45422332781  + 10192.51015071860 *T)
    sumB1  +=       0.346 * cos(0.94242286364  + 1059.38193018920 *T)
    sumB1  +=       0.308 * cos(4.90145899142  + 3930.20969621960 *T)
    sumB1  +=       0.331 * cos(4.89498986674  + 10206.17199921020 *T)
    sumB1  +=       0.269 * cos(2.39650266204  + 801.82093112380 *T)
    sumB1  +=       0.269 * cos(0.00589873499  + 9830.38901398780 *T)
    sumB1  +=       0.261 * cos(3.48196147279  + 7058.59846131540 *T)
    sumB1  +=       0.290 * cos(0.10953964861  + 29050.78374334920 *T)
    sumB1  +=       0.283 * cos(6.12133736787  + 20419.45754542119 *T)
    sumB1  +=       0.232 * cos(3.07845850030  + 12566.15169998280 *T)
    sumB1  +=       0.265 * cos(4.02431894973  + 33794.54372352860 *T)
    sumB1  +=       0.220 * cos(2.37315851889  + 4551.95349705880 *T)
    sumB1  +=       0.247 * cos(3.07626728158  + 28286.99048486120 *T)
    sumB1  +=       0.202 * cos(3.56872121409  + 21535.94964451540 *T)
    sumB1  +=       0.225 * cos(5.76888896320  + 213.29909543800 *T)
    sumB1  +=       0.217 * cos(0.88382111135  + 20213.27199698400 *T)
    sumB1  +=       0.172 * cos(6.12653050186  + 9161.01716302260 *T)
    sumB1  +=       0.195 * cos(5.47240855400  + 37724.75341974820 *T)
    sumB1  +=       0.153 * cos(4.07656151671  + 27511.46787353720 *T)
    sumB1  +=       0.174 * cos(1.33676849359  + 32217.20018108080 *T)
    sumB1  +=       0.157 * cos(5.98474214437  + 26.29831979980 *T)
    sumB1  +=       0.163 * cos(5.45519134760  + 10426.58464164900 *T)
    sumB1  +=       0.129 * cos(2.08748660996  + 3128.38876509580 *T)
    sumB1  +=       0.131 * cos(1.51959002513  + 10218.80847051840 *T)
    sumB1  +=       0.139 * cos(4.42330401713  + 10220.39909321180 *T)
    sumB1  +=       0.126 * cos(2.62296638037  + 22779.43724619380 *T)
    sumB1  +=       0.146 * cos(4.69869606856  + 25158.60171976540 *T)
    sumB1  +=       0.172 * cos(6.13435208788  + 18837.49819713819 *T)
    sumB1  +=       0.157 * cos(5.44507403858  + 4705.73230754360 *T)
    sumB1  +=       0.117 * cos(6.18296175153  + 20400.27277262220 *T)
    sumB1  +=       0.164 * cos(3.30849473132  + 51066.42773105500 *T)
    sumB1  +=       0.113 * cos(3.64412860654  + 7.11354700080 *T)
    sumB1  +=       0.109 * cos(5.21220660788  + 8662.24032356300 *T)
    sumB1  +=       0.133 * cos(1.78047296245  + 191.44826611160 *T)
    sumB1  +=       0.117 * cos(0.14681677884  + 9146.79006902100 *T)
    sumB1  +=       0.116 * cos(0.61940521198  + 41654.96311596780 *T)
    sumB1  +=       0.096 * cos(1.49631428731  + 7084.89678111520 *T)
    sumB1  +=       0.096 * cos(1.21744230443  + 10198.03307502600 *T)
    sumB1  +=       0.082 * cos(1.45863866349  + 10207.76262190360 *T)
    sumB1  +=       0.085 * cos(6.04057728058  + 21202.09370374600 *T)
    sumB1  +=       0.083 * cos(0.19985600927  + 14919.01785375460 *T)
    sumB1  +=       0.077 * cos(5.50132310610  + 5661.33204915220 *T)
    sumB1  +=       0.077 * cos(2.00173927326  + 10228.53801739600 *T)
    sumB1  +=       0.093 * cos(1.85466268819  + 45585.17281218740 *T)
    sumB1  +=       0.066 * cos(3.25826124156  + 1109.37855209340 *T)
    sumB1  +=       0.089 * cos(0.64100435648  + 3154.68708489560 *T)
    sumB1  +=       0.061 * cos(3.80043027736  + 11272.66747640020 *T)
    sumB1  +=       0.077 * cos(1.85516358950  + 3532.06069281140 *T)
    sumB1  +=       0.062 * cos(0.81341290651  + 382.89653222320 *T)
    sumB1  +=       0.072 * cos(2.35312965005  + 9103.90699411760 *T)
    sumB1  +=       0.053 * cos(3.21969389511  + 20452.86941222180 *T)
    sumB1  +=       0.067 * cos(1.42090542131  + 24356.78078864160 *T)
    sumB1  +=       0.056 * cos(2.97733070198  + 30110.16567353840 *T)
    sumB1  +=       0.051 * cos(4.22406663447  + 20809.46762464520 *T)
    sumB1  +=       0.058 * cos(6.20761936031  + 29088.81141598500 *T)
    sumB1  +=       0.061 * cos(3.27309494322  + 49515.38250840700 *T)
    sumB1  +=       0.046 * cos(5.49443476235  + 31441.67756975680 *T)
    sumB1  +=       0.050 * cos(4.16651052942  + 13341.67431130680 *T)
    sumB1  +=       0.047 * cos(1.25473247769  + 33019.02111220460 *T)
    sumB1  +=       0.047 * cos(2.03402044389  + 23958.63178523340 *T)
    sumB1  +=       0.036 * cos(5.24409311105  + 3149.16416058820 *T)
    sumB1  +=       0.038 * cos(4.15337829669  + 18849.22754997420 *T)
    sumB1  +=       0.042 * cos(0.43005959574  + 1589.07289528380 *T)
    sumB1  +=       0.041 * cos(1.21289342964  + 12592.45001978260 *T)
    sumB1  +=       0.038 * cos(5.91928287144  + 28521.09277825460 *T)
    sumB1  +=       0.033 * cos(3.98241699279  + 4732.03062734340 *T)
    sumB1  +=       0.035 * cos(2.24417218267  + 16496.36139620240 *T)
    sumB1  +=       0.040 * cos(6.13293942728  + 26087.90314157420 *T)
    sumB1  +=       0.044 * cos(1.78123294860  + 426.59819087600 *T)
    sumB1  +=       0.041 * cos(3.16744909855  + 39264.06928956020 *T)
    sumB1  +=       0.033 * cos(4.96183427323  + 536.80451209540 *T)
    sumB1  +=       0.034 * cos(0.12963030501  + 30213.25844775700 *T)
    sumB1  +=       0.036 * cos(5.41167321573  + 522.57741809380 *T)
    sumB1  +=       0.027 * cos(4.44250239485  + 17277.40693183380 *T)
    sumB1  +=       0.034 * cos(5.94541303751  + 9929.42622734580 *T)
    sumB1  +=       0.033 * cos(0.40689057274  + 10497.14486507620 *T)
    sumB1  +=       0.023 * cos(2.59067946967  + 10175.25787357520 *T)
    sumB1  +=       0.022 * cos(0.69625017371  + 19999.97290154599 *T)
    sumB1  +=       0.023 * cos(3.76162101633  + 10251.31321884680 *T)
    sumB1  +=       0.023 * cos(0.62711494266  + 35371.88726597640 *T)
    sumB1  +=       0.022 * cos(4.64142978776  + 19889.76658032659 *T)
    sumB1  +=       0.020 * cos(4.01315480107  + 26709.64694241340 *T)
    sumB1  +=       0.020 * cos(4.03344400680  + 29573.36116144300 *T)
    sumB1  +=       0.023 * cos(0.90416640595  + 8094.52168583260 *T)
    sumB1  +=       0.022 * cos(1.92092469688  + 17085.95866572220 *T)
    sumB1  +=       0.019 * cos(5.04938942644  + 6681.22485339960 *T)

    # VENUS Latitude
    sumB2  = 0
    sumB2  +=   22377.665 * cos(3.38509143877  + 10213.28554621100 *T)
    sumB2  +=     281.739 * cos(0.00000000000  + 0.00000000000 *T)
    sumB2  +=     173.164 * cos(5.25563766915  + 20426.57109242200 *T)
    sumB2  +=      26.945 * cos(3.87040891568  + 30639.85663863300 *T)
    sumB2  +=       1.174 * cos(0.09768632072  + 10186.98722641120 *T)
    sumB2  +=       0.685 * cos(3.19139067811  + 11790.62908865880 *T)
    sumB2  +=       0.788 * cos(4.36515965295  + 10239.58386601080 *T)
    sumB2  +=       0.592 * cos(5.22270440328  + 40853.14218484400 *T)
    sumB2  +=       0.515 * cos(6.12821215207  + 10988.80815753500 *T)
    sumB2  +=       0.538 * cos(0.57550272342  + 2352.86615377180 *T)
    sumB2  +=       0.540 * cos(3.11657836329  + 18073.70493865020 *T)
    sumB2  +=       0.454 * cos(2.79306867629  + 10404.73381232260 *T)
    sumB2  +=       0.374 * cos(6.10468482446  + 9437.76293488700 *T)
    sumB2  +=       0.431 * cos(4.00778431184  + 1577.34354244780 *T)
    sumB2  +=       0.360 * cos(6.01747842320  + 19651.04848109800 *T)
    sumB2  +=       0.375 * cos(1.31319959789  + 22003.91463486980 *T)
    sumB2  +=       0.354 * cos(5.12509281266  + 9153.90361602180 *T)
    sumB2  +=       0.150 * cos(4.58623687118  + 15720.83878487840 *T)
    sumB2  +=       0.164 * cos(5.41790158607  + 5507.55323866740 *T)
    sumB2  +=       0.159 * cos(2.78191550878  + 19896.88012732740 *T)
    sumB2  +=       0.157 * cos(0.65774905071  + 529.69096509460 *T)
    sumB2  +=       0.155 * cos(2.54824315372  + 9683.59458111640 *T)
    sumB2  +=       0.109 * cos(2.01866665583  + 14143.49524243060 *T)
    sumB2  +=       0.106 * cos(2.28289033017  + 6283.07584999140 *T)
    sumB2  +=       0.115 * cos(3.23636374193  + 20618.01935853360 *T)
    sumB2  +=       0.128 * cos(5.32400510939  + 13367.97263110660 *T)
    sumB2  +=       0.087 * cos(3.28265082435  + 11322.66409830440 *T)
    sumB2  +=       0.090 * cos(5.23585072275  + 10596.18207843420 *T)
    sumB2  +=       0.055 * cos(4.82369879741  + 7058.59846131540 *T)
    sumB2  +=       0.044 * cos(0.58444963462  + 10206.17199921020 *T)
    sumB2  +=       0.044 * cos(2.34401612969  + 19367.18916223280 *T)
    sumB2  +=       0.038 * cos(4.55053233088  + 9999.98645077300 *T)
    sumB2  +=       0.039 * cos(5.84340580032  + 10220.39909321180 *T)
    sumB2  +=       0.036 * cos(4.41006216127  + 51066.42773105500 *T)
    sumB2  +=       0.039 * cos(3.14348236386  + 9411.46461508720 *T)
    sumB2  +=       0.033 * cos(4.55748660340  + 10742.97651130560 *T)
    sumB2  +=       0.037 * cos(2.79630938717  + 25934.12433108940 *T)
    sumB2  +=       0.034 * cos(0.55287110072  + 11015.10647733480 *T)
    sumB2  +=       0.034 * cos(2.25809144959  + 29580.47470844380 *T)
    sumB2  +=       0.038 * cos(1.88638747393  + 801.82093112380 *T)
    sumB2  +=       0.034 * cos(1.22706917271  + 10021.83728009940 *T)
    sumB2  +=       0.027 * cos(4.83867137637  + 9830.38901398780 *T)
    sumB2  +=       0.027 * cos(4.31140179350  + 23581.25817731760 *T)
    sumB2  +=       0.027 * cos(2.17187621336  + 8635.94200376320 *T)
    sumB2  +=       0.020 * cos(5.66581696952  + 21228.39202354580 *T)
    sumB2  +=       0.024 * cos(2.17208107850  + 18849.22754997420 *T)
    sumB2  +=       0.020 * cos(5.29318634138  + 775.52261132400 *T)
    sumB2  +=       0.019 * cos(2.73486845601  + 3128.38876509580 *T)
    sumB2  +=       0.013 * cos(3.40362915274  + 1059.38193018920 *T)
    sumB2  +=       0.014 * cos(0.05074160195  + 7860.41939243920 *T)
    sumB2  +=       0.014 * cos(5.43035907265  + 26.29831979980 *T)
    sumB2  +=       0.012 * cos(3.24834347355  + 9103.90699411760 *T)
    sumB2  +=       0.013 * cos(5.04826725887  + 7.11354700080 *T)
    sumB2  +=       0.015 * cos(1.42027402522  + 29050.78374334920 *T)
    sumB2  +=       0.010 * cos(4.98138067490  + 10426.58464164900 *T)
    sumB2  +=       0.011 * cos(0.85773045784  + 17298.18232732620 *T)
    sumB2  +=       0.011 * cos(4.23048200054  + 29864.33402730900 *T)
    sumB2  +=       0.010 * cos(0.26447399758  + 3930.20969621960 *T)
    sumB2  +=       0.011 * cos(1.46728576671  + 20419.45754542119 *T)

    # VENUS Latitude
    sumB3  = 0
    sumB3  +=     646.671 * cos(4.99166565277  + 10213.28554621100 *T)
    sumB3  +=      19.952 * cos(3.14159265359  + 0.00000000000 *T)
    sumB3  +=       5.540 * cos(0.77376923951  + 20426.57109242200 *T)
    sumB3  +=       2.526 * cos(5.44493763020  + 30639.85663863300 *T)
    sumB3  +=       0.079 * cos(1.51447613604  + 10186.98722641120 *T)
    sumB3  +=       0.056 * cos(0.63647808442  + 40853.14218484400 *T)
    sumB3  +=       0.058 * cos(5.70731176550  + 10239.58386601080 *T)
    sumB3  +=       0.031 * cos(4.72523061067  + 11790.62908865880 *T)
    sumB3  +=       0.026 * cos(1.02068113372  + 10988.80815753500 *T)
    sumB3  +=       0.025 * cos(5.60599130442  + 9437.76293488700 *T)
    sumB3  +=       0.017 * cos(2.05293621864  + 2352.86615377180 *T)
    sumB3  +=       0.011 * cos(4.33056892256  + 10404.73381232260 *T)
    sumB3  +=       0.009 * cos(1.36283915068  + 19651.04848109800 *T)
    sumB3  +=       0.007 * cos(4.69592781899  + 18073.70493865020 *T)
    sumB3  +=       0.006 * cos(2.97926526705  + 22003.91463486980 *T)

    # VENUS Latitude
    sumB4  = 0
    sumB4  +=      14.102 * cos(0.31537190181  + 10213.28554621100 *T)
    sumB4  +=       0.190 * cos(2.35466404492  + 20426.57109242200 *T)
    sumB4  +=       0.164 * cos(0.74476215141  + 30639.85663863300 *T)
    sumB4  +=       0.214 * cos(3.14159265359  + 0.00000000000 *T)
    sumB4  +=       0.004 * cos(2.34190883009  + 40853.14218484400 *T)

    # VENUS Latitude
    sumB5  = 0
    sumB5  +=       0.239 * cos(2.05201727566  + 10213.28554621100 *T)
    sumB5  +=       0.039 * cos(0.00000000000  + 0.00000000000 *T)
    sumB5  +=       0.011 * cos(3.82500275251  + 20426.57109242200 *T)
    sumB5  +=       0.009 * cos(2.32953116868  + 30639.85663863300 *T)

    # VENUS Radius
    sumR0  = 0
    sumR0  +=  72334820.905 * cos(0.00000000000  + 0.00000000000 *T)
    sumR0  +=  489824.185 * cos(4.02151832268  + 10213.28554621100 *T)
    sumR0  +=    1658.058 * cos(4.90206728012  + 20426.57109242200 *T)
    sumR0  +=    1632.093 * cos(2.84548851892  + 7860.41939243920 *T)
    sumR0  +=    1378.048 * cos(1.12846590600  + 11790.62908865880 *T)
    sumR0  +=     498.399 * cos(2.58682187717  + 9683.59458111640 *T)
    sumR0  +=     373.958 * cos(1.42314837063  + 3930.20969621960 *T)
    sumR0  +=     263.616 * cos(5.52938185920  + 9437.76293488700 *T)
    sumR0  +=     237.455 * cos(2.55135903978  + 15720.83878487840 *T)
    sumR0  +=     221.983 * cos(2.01346776772  + 19367.18916223280 *T)
    sumR0  +=     119.467 * cos(3.01975365264  + 10404.73381232260 *T)
    sumR0  +=     125.896 * cos(2.72769833559  + 1577.34354244780 *T)
    sumR0  +=      76.178 * cos(1.59577224486  + 9153.90361602180 *T)
    sumR0  +=      85.336 * cos(3.98607953754  + 19651.04848109800 *T)
    sumR0  +=      74.347 * cos(4.11957854039  + 5507.55323866740 *T)
    sumR0  +=      41.904 * cos(1.64273363458  + 18837.49819713819 *T)
    sumR0  +=      42.493 * cos(3.81864530735  + 13367.97263110660 *T)
    sumR0  +=      39.430 * cos(5.39019422358  + 23581.25817731760 *T)
    sumR0  +=      29.042 * cos(5.67739528728  + 5661.33204915220 *T)
    sumR0  +=      27.555 * cos(5.72392407794  + 775.52261132400 *T)
    sumR0  +=      27.283 * cos(4.82151812709  + 11015.10647733480 *T)
    sumR0  +=      31.274 * cos(2.31806719544  + 9999.98645077300 *T)
    sumR0  +=      19.700 * cos(4.96157560245  + 11322.66409830440 *T)
    sumR0  +=      19.809 * cos(0.53189326492  + 27511.46787353720 *T)
    sumR0  +=      13.567 * cos(3.75530870628  + 18073.70493865020 *T)
    sumR0  +=      12.921 * cos(1.13381083556  + 10206.17199921020 *T)
    sumR0  +=      16.215 * cos(0.56453834290  + 529.69096509460 *T)
    sumR0  +=      11.821 * cos(5.09025877427  + 3154.68708489560 *T)
    sumR0  +=      11.728 * cos(0.23432298744  + 7084.89678111520 *T)
    sumR0  +=      13.079 * cos(5.24353197586  + 17298.18232732620 *T)
    sumR0  +=      13.180 * cos(3.37207825651  + 13745.34623902240 *T)
    sumR0  +=       9.097 * cos(3.07004895769  + 1109.37855209340 *T)
    sumR0  +=      10.818 * cos(2.45024712908  + 10239.58386601080 *T)
    sumR0  +=      11.438 * cos(4.56838894696  + 29050.78374334920 *T)
    sumR0  +=       8.377 * cos(5.78327612352  + 30639.85663863300 *T)
    sumR0  +=       8.193 * cos(1.95023111860  + 22003.91463486980 *T)
    sumR0  +=       9.308 * cos(1.61615909286  + 2352.86615377180 *T)
    sumR0  +=      10.652 * cos(1.95528396140  + 31441.67756975680 *T)
    sumR0  +=      10.357 * cos(1.20234990061  + 15874.61759536320 *T)
    sumR0  +=       9.585 * cos(1.46639856228  + 19999.97290154599 *T)
    sumR0  +=       6.506 * cos(2.17390732263  + 14143.49524243060 *T)
    sumR0  +=       7.562 * cos(1.13789564977  + 8624.21265092720 *T)
    sumR0  +=       6.434 * cos(0.84419623033  + 6283.07584999140 *T)
    sumR0  +=       5.898 * cos(0.01093731110  + 8635.94200376320 *T)
    sumR0  +=       5.632 * cos(3.94956548631  + 12566.15169998280 *T)
    sumR0  +=       5.523 * cos(1.27394296557  + 18307.80723204360 *T)
    sumR0  +=       4.488 * cos(2.47835729057  + 191.44826611160 *T)
    sumR0  +=       4.529 * cos(4.73027770400  + 19896.88012732740 *T)
    sumR0  +=       6.193 * cos(3.25881250939  + 6872.67311951120 *T)
    sumR0  +=       6.070 * cos(0.35337419942  + 21228.39202354580 *T)
    sumR0  +=       4.315 * cos(2.59737099519  + 4551.95349705880 *T)
    sumR0  +=       6.005 * cos(3.37874723475  + 35371.88726597640 *T)
    sumR0  +=       3.852 * cos(1.01162850357  + 9786.68735533500 *T)
    sumR0  +=       4.033 * cos(0.00050855580  + 801.82093112380 *T)
    sumR0  +=       3.920 * cos(5.56542869407  + 10596.18207843420 *T)
    sumR0  +=       2.709 * cos(5.80195530112  + 7064.12138562280 *T)
    sumR0  +=       3.216 * cos(0.39767254848  + 10186.98722641120 *T)
    sumR0  +=       3.089 * cos(6.26174762876  + 14945.31617355440 *T)
    sumR0  +=       2.982 * cos(4.21196716354  + 28521.09277825460 *T)
    sumR0  +=       3.284 * cos(0.70709821006  + 10742.97651130560 *T)
    sumR0  +=       3.484 * cos(4.79878191875  + 39302.09696219600 *T)
    sumR0  +=       3.172 * cos(1.80518954174  + 25158.60171976540 *T)
    sumR0  +=       2.463 * cos(0.68708153678  + 10988.80815753500 *T)
    sumR0  +=       2.374 * cos(3.77948685343  + 21535.94964451540 *T)
    sumR0  +=       2.198 * cos(2.82996372521  + 8662.24032356300 *T)
    sumR0  +=       1.958 * cos(5.41763804167  + 16496.36139620240 *T)
    sumR0  +=       1.876 * cos(2.63426768393  + 29580.47470844380 *T)
    sumR0  +=       1.902 * cos(2.85782199133  + 3532.06069281140 *T)
    sumR0  +=       1.706 * cos(3.67573010379  + 26.29831979980 *T)
    sumR0  +=       1.817 * cos(0.41611036449  + 4705.73230754360 *T)
    sumR0  +=       1.858 * cos(1.50368318296  + 10021.83728009940 *T)
    sumR0  +=       2.087 * cos(6.22112874639  + 43232.30665841560 *T)
    sumR0  +=       1.950 * cos(2.21447019683  + 19786.67380610799 *T)
    sumR0  +=       1.497 * cos(0.00134773824  + 17277.40693183380 *T)
    sumR0  +=       1.819 * cos(3.23144993268  + 29088.81141598500 *T)
    sumR0  +=       1.423 * cos(5.85979618707  + 9676.48103411560 *T)
    sumR0  +=       1.223 * cos(5.55818994329  + 6770.71060124560 *T)
    sumR0  +=       1.140 * cos(5.92088900094  + 13936.79450513400 *T)
    sumR0  +=       1.484 * cos(2.47665429253  + 31749.23519072640 *T)
    sumR0  +=       1.185 * cos(1.42087628351  + 4732.03062734340 *T)
    sumR0  +=       1.323 * cos(2.48821075422  + 9690.70812811720 *T)
    sumR0  +=       1.249 * cos(1.88323673734  + 19374.30270923360 *T)
    sumR0  +=       1.270 * cos(5.24647873116  + 19360.07561523199 *T)
    sumR0  +=       1.402 * cos(5.17536780118  + 10316.37832042960 *T)
    sumR0  +=       1.042 * cos(3.05454698508  + 25934.12433108940 *T)
    sumR0  +=       1.174 * cos(1.42913732999  + 18875.52586977400 *T)
    sumR0  +=       1.278 * cos(1.35747287297  + 47162.51635463520 *T)
    sumR0  +=       0.917 * cos(6.26337648765  + 20618.01935853360 *T)
    sumR0  +=       0.905 * cos(1.12740203561  + 12592.45001978260 *T)
    sumR0  +=       1.093 * cos(4.64451720605  + 33019.02111220460 *T)
    sumR0  +=       1.014 * cos(1.09259406433  + 1059.38193018920 *T)
    sumR0  +=       0.783 * cos(2.02118183873  + 24356.78078864160 *T)
    sumR0  +=       0.779 * cos(0.41585274010  + 3340.61242669980 *T)
    sumR0  +=       0.700 * cos(1.14936815714  + 16983.99614745660 *T)
    sumR0  +=       0.878 * cos(0.87852464964  + 38734.37832446560 *T)
    sumR0  +=       0.623 * cos(0.89976912165  + 17778.11626694899 *T)
    sumR0  +=       0.608 * cos(1.58476225197  + 9573.38825989700 *T)
    sumR0  +=       0.800 * cos(3.94213003073  + 10138.50394764370 *T)
    sumR0  +=       0.760 * cos(1.31851313748  + 9967.45389998160 *T)
    sumR0  +=       0.802 * cos(2.78173370208  + 51092.72605085480 *T)
    sumR0  +=       0.664 * cos(4.45864682400  + 3128.38876509580 *T)
    sumR0  +=       0.674 * cos(5.11214939998  + 382.89653222320 *T)
    sumR0  +=       0.530 * cos(0.85392938403  + 10234.06094170340 *T)
    sumR0  +=       0.509 * cos(3.56809374595  + 28286.99048486120 *T)
    sumR0  +=       0.600 * cos(4.25927726907  + 41962.52073693740 *T)
    sumR0  +=       0.601 * cos(5.78144137895  + 213.29909543800 *T)
    sumR0  +=       0.595 * cos(2.83045104588  + 22805.73556599360 *T)
    sumR0  +=       0.673 * cos(6.06079908421  + 36949.23080842420 *T)
    sumR0  +=       0.535 * cos(5.85422519711  + 9103.90699411760 *T)
    sumR0  +=       0.544 * cos(5.44806074800  + 3723.50895892300 *T)
    sumR0  +=       0.492 * cos(3.83802404893  + 27991.40181316000 *T)
    sumR0  +=       0.635 * cos(0.76494024849  + 8094.52168583260 *T)
    sumR0  +=       0.434 * cos(6.22214487735  + 27197.28169366760 *T)
    sumR0  +=       0.459 * cos(3.55062885479  + 20213.27199698400 *T)
    sumR0  +=       0.398 * cos(6.16269975784  + 10426.58464164900 *T)
    sumR0  +=       0.378 * cos(2.41665947591  + 18844.61174413899 *T)
    sumR0  +=       0.421 * cos(4.86552697954  + 9146.79006902100 *T)
    sumR0  +=       0.500 * cos(4.20351458644  + 55022.93574707440 *T)
    sumR0  +=       0.404 * cos(4.95834410782  + 37410.56723987860 *T)
    sumR0  +=       0.402 * cos(2.97963246945  + 10220.39909321180 *T)
    sumR0  +=       0.464 * cos(2.59869499733  + 18734.40542291960 *T)
    sumR0  +=       0.352 * cos(0.08963076359  + 10103.07922499160 *T)
    sumR0  +=       0.348 * cos(4.90260339364  + 18830.38465013739 *T)
    sumR0  +=       0.338 * cos(3.22520096478  + 24150.08005134500 *T)
    sumR0  +=       0.375 * cos(6.17532088136  + 26087.90314157420 *T)
    sumR0  +=       0.425 * cos(1.20052578280  + 40879.44050464380 *T)
    sumR0  +=       0.408 * cos(3.12833060705  + 9050.81084180320 *T)
    sumR0  +=       0.385 * cos(1.94284690176  + 283.85931886520 *T)
    sumR0  +=       0.337 * cos(4.87838699272  + 12432.04265039780 *T)
    sumR0  +=       0.326 * cos(4.27369741426  + 26735.94526221320 *T)
    sumR0  +=       0.309 * cos(0.50597475053  + 38204.68735937100 *T)
    sumR0  +=       0.329 * cos(3.88430599153  + 29864.33402730900 *T)
    sumR0  +=       0.313 * cos(1.36138752543  + 10192.51015071860 *T)
    sumR0  +=       0.347 * cos(3.58439807209  + 27490.69247804480 *T)
    sumR0  +=       0.251 * cos(3.78618457047  + 10063.72234907640 *T)
    sumR0  +=       0.244 * cos(3.83523342668  + 9411.46461508720 *T)
    sumR0  +=       0.281 * cos(4.50895206233  + 32217.20018108080 *T)
    sumR0  +=       0.237 * cos(0.87748812245  + 6681.22485339960 *T)
    sumR0  +=       0.315 * cos(5.62657778233  + 58953.14544329400 *T)
    sumR0  +=       0.311 * cos(4.15626121491  + 10175.15251057320 *T)
    sumR0  +=       0.247 * cos(2.53637594113  + 16522.65971600220 *T)
    sumR0  +=       0.219 * cos(5.08729383251  + 7058.59846131540 *T)
    sumR0  +=       0.291 * cos(3.72567217056  + 29999.95935231900 *T)
    sumR0  +=       0.267 * cos(2.97685503991  + 19573.37471066999 *T)
    sumR0  +=       0.280 * cos(3.70200084294  + 47623.85278608960 *T)
    sumR0  +=       0.239 * cos(3.94545782067  + 9580.50180689780 *T)
    sumR0  +=       0.246 * cos(2.18244883930  + 9161.01716302260 *T)
    sumR0  +=       0.253 * cos(2.69506547016  + 3442.57494496540 *T)
    sumR0  +=       0.265 * cos(2.62811801237  + 44809.65020086340 *T)
    sumR0  +=       0.194 * cos(4.78926136175  + 33794.54372352860 *T)
    sumR0  +=       0.187 * cos(3.65620881095  + 20452.86941222180 *T)
    sumR0  +=       0.224 * cos(2.43601863127  + 9992.87290377220 *T)
    sumR0  +=       0.193 * cos(2.55112161845  + 2379.16447357160 *T)
    sumR0  +=       0.201 * cos(1.90356905733  + 1551.04522264800 *T)
    sumR0  +=       0.176 * cos(4.29837616553  + 10137.01947493540 *T)
    sumR0  +=       0.184 * cos(6.16061560223  + 36147.40987730040 *T)
    sumR0  +=       0.175 * cos(2.71984797040  + 20809.46762464520 *T)
    sumR0  +=       0.186 * cos(2.55098927966  + 14919.01785375460 *T)
    sumR0  +=       0.161 * cos(4.13272567123  + 23958.63178523340 *T)
    sumR0  +=       0.221 * cos(4.83552377614  + 20277.00789528740 *T)
    sumR0  +=       0.160 * cos(1.81472642729  + 10787.63034454580 *T)
    sumR0  +=       0.199 * cos(5.74259798330  + 30666.15495843280 *T)
    sumR0  +=       0.160 * cos(4.46270605493  + 18947.70451835760 *T)
    sumR0  +=       0.187 * cos(2.98688597588  + 2218.75710418680 *T)
    sumR0  +=       0.189 * cos(5.34607810282  + 10007.09999777380 *T)
    sumR0  +=       0.198 * cos(0.77846666692  + 62883.35513951360 *T)
    sumR0  +=       0.144 * cos(5.00261963924  + 9264.10993724120 *T)
    sumR0  +=       0.171 * cos(2.05212624568  + 7255.56965173440 *T)
    sumR0  +=       0.188 * cos(4.08173534559  + 48739.85989708300 *T)
    sumR0  +=       0.146 * cos(3.94191715702  + 6309.37416979120 *T)
    sumR0  +=       0.146 * cos(5.06313558118  + 39264.06928956020 *T)
    sumR0  +=       0.135 * cos(5.93689169614  + 37724.75341974820 *T)
    sumR0  +=       0.139 * cos(2.81266025896  + 20.77539549240 *T)
    sumR0  +=       0.177 * cos(5.16224804657  + 9835.91193829520 *T)
    sumR0  +=       0.119 * cos(1.37254262864  + 40077.61957352000 *T)
    sumR0  +=       0.120 * cos(0.21443767468  + 31022.75317085620 *T)
    sumR0  +=       0.128 * cos(2.92458887798  + 7.11354700080 *T)
    sumR0  +=       0.150 * cos(5.73646272556  + 632.78373931320 *T)
    sumR0  +=       0.106 * cos(0.62224833817  + 11272.66747640020 *T)
    sumR0  +=       0.114 * cos(2.63301326520  + 17468.85519794540 *T)
    sumR0  +=       0.123 * cos(6.22518843711  + 53285.18483524180 *T)
    sumR0  +=       0.107 * cos(1.17258978900  + 43071.89928903080 *T)
    sumR0  +=       0.103 * cos(1.09613781581  + 41654.96311596780 *T)
    sumR0  +=       0.109 * cos(2.01412667085  + 20419.45754542119 *T)
    sumR0  +=       0.102 * cos(4.23406964348  + 10251.31321884680 *T)
    sumR0  +=       0.116 * cos(1.27731728606  + 10199.05845220940 *T)
    sumR0  +=       0.103 * cos(5.25887538465  + 9830.38901398780 *T)
    sumR0  +=       0.112 * cos(2.24436894064  + 18204.71445782499 *T)
    sumR0  +=       0.111 * cos(2.23547857955  + 8521.11987670860 *T)
    sumR0  +=       0.118 * cos(0.23754207200  + 10497.14486507620 *T)
    sumR0  +=       0.123 * cos(0.88054816668  + 34596.36465465240 *T)
    sumR0  +=       0.102 * cos(4.39438646620  + 18300.69368504279 *T)
    sumR0  +=       0.131 * cos(6.01711652115  + 9367.20271145980 *T)
    sumR0  +=       0.100 * cos(5.00532389609  + 10175.25787357520 *T)
    sumR0  +=       0.107 * cos(0.41270197502  + 40853.14218484400 *T)
    sumR0  +=       0.132 * cos(5.45008342761  + 11506.76976979360 *T)
    sumR0  +=       0.098 * cos(1.07722950958  + 13553.89797291080 *T)
    sumR0  +=       0.094 * cos(2.91720097590  + 44007.82926973960 *T)
    sumR0  +=       0.097 * cos(1.04004223634  + 68050.42387851159 *T)
    sumR0  +=       0.127 * cos(2.20215372683  + 66813.56483573320 *T)
    sumR0  +=       0.111 * cos(1.57823839032  + 29043.67019634839 *T)
    sumR0  +=       0.118 * cos(2.33268176890  + 18314.92077904440 *T)
    sumR0  +=       0.090 * cos(2.42353056125  + 32858.61374281979 *T)
    sumR0  +=       0.109 * cos(3.82796787296  + 19470.28193645139 *T)
    sumR0  +=       0.111 * cos(4.47666957576  + 29057.89729034999 *T)
    sumR0  +=       0.101 * cos(3.41528493660  + 19264.09638801420 *T)
    sumR0  +=       0.092 * cos(3.66289799512  + 22645.32819660879 *T)
    sumR0  +=       0.094 * cos(6.07530805791  + 10846.06928552420 *T)
    sumR0  +=       0.114 * cos(4.02718653431  + 7576.56007357400 *T)
    sumR0  +=       0.087 * cos(6.01842459303  + 17085.95866572220 *T)
    sumR0  +=       0.109 * cos(5.46886607309  + 52670.06959330260 *T)
    sumR0  +=       0.107 * cos(0.54805946713  + 34363.36559755600 *T)
    sumR0  +=       0.108 * cos(5.44460610707  + 19050.79729257620 *T)
    sumR0  +=       0.076 * cos(6.15177368654  + 27682.14074415640 *T)
    sumR0  +=       0.107 * cos(4.80525404063  + 8144.27871130440 *T)
    sumR0  +=       0.073 * cos(1.60549217847  + 20956.26205751660 *T)
    sumR0  +=       0.097 * cos(5.13542051130  + 22779.43724619380 *T)
    sumR0  +=       0.068 * cos(2.31300447144  + 8631.32619792800 *T)
    sumR0  +=       0.091 * cos(4.28652743953  + 10110.19277199240 *T)
    sumR0  +=       0.093 * cos(5.27290609264  + 522.57741809380 *T)
    sumR0  +=       0.071 * cos(3.65565961690  + 11764.33076885900 *T)
    sumR0  +=       0.089 * cos(1.79712963206  + 45585.17281218740 *T)
    sumR0  +=       0.067 * cos(2.25900071584  + 9360.08916445900 *T)
    sumR0  +=       0.085 * cos(0.67062144972  + 56600.27928952220 *T)
    sumR0  +=       0.080 * cos(1.58278081077  + 19992.85935454519 *T)
    sumR0  +=       0.065 * cos(6.23472325597  + 10419.47109464820 *T)
    sumR0  +=       0.064 * cos(0.53356325917  + 17248.42530185440 *T)
    sumR0  +=       0.085 * cos(4.52011215904  + 29786.66025688100 *T)
    sumR0  +=       0.068 * cos(4.48235266554  + 10632.77019008620 *T)
    sumR0  +=       0.064 * cos(4.33495700921  + 47938.03896595920 *T)
    sumR0  +=       0.071 * cos(3.03858484137  + 11787.10597030980 *T)
    sumR0  +=       0.087 * cos(4.81823063172  + 2107.03450754240 *T)
    sumR0  +=       0.070 * cos(2.35648061034  + 11794.15220700780 *T)
    sumR0  +=       0.080 * cos(2.33248094128  + 38526.57435087200 *T)
    sumR0  +=       0.070 * cos(3.70454061100  + 8734.41897214660 *T)
    sumR0  +=       0.077 * cos(4.49569185467  + 20007.08644854680 *T)
    sumR0  +=       0.072 * cos(1.19410424468  + 10217.21769947410 *T)
    sumR0  +=       0.068 * cos(2.01841060183  + 14128.24277124560 *T)
    sumR0  +=       0.064 * cos(5.39293951654  + 7880.08915333899 *T)
    sumR0  +=       0.066 * cos(3.20467071127  + 14765.23904326980 *T)
    sumR0  +=       0.080 * cos(3.41620457770  + 48417.97290558199 *T)
    sumR0  +=       0.080 * cos(3.39651161571  + 245.83164622940 *T)
    sumR0  +=       0.066 * cos(5.85414440204  + 9793.80090233580 *T)
    sumR0  +=       0.082 * cos(3.62592908644  + 70743.77453195279 *T)
    sumR0  +=       0.058 * cos(4.95174942212  + 30110.16567353840 *T)
    sumR0  +=       0.079 * cos(6.24161471033  + 6037.24420376200 *T)
    sumR0  +=       0.069 * cos(5.50183658445  + 19793.78735310880 *T)
    sumR0  +=       0.056 * cos(1.24148350566  + 10207.76262190360 *T)
    sumR0  +=       0.070 * cos(2.45123308846  + 10218.80847051840 *T)
    sumR0  +=       0.064 * cos(5.53983104501  + 10735.86296430480 *T)
    sumR0  +=       0.054 * cos(3.62259713240  + 27461.71084806540 *T)
    sumR0  +=       0.073 * cos(1.75882480924  + 1589.07289528380 *T)
    sumR0  +=       0.075 * cos(3.38244819846  + 4214.06901508480 *T)
    sumR0  +=       0.054 * cos(0.64971567468  + 9929.42622734580 *T)
    sumR0  +=       0.054 * cos(3.40959637230  + 18418.01355326299 *T)
    sumR0  +=       0.056 * cos(3.65815006538  + 14169.79356223040 *T)
    sumR0  +=       0.056 * cos(0.71243223808  + 9896.89367655440 *T)
    sumR0  +=       0.052 * cos(1.33348131940  + 20400.27277262220 *T)
    sumR0  +=       0.067 * cos(3.12806595400  + 5481.25491886760 *T)
    sumR0  +=       0.058 * cos(0.54482893546  + 28313.28880466100 *T)
    sumR0  +=       0.054 * cos(0.15603935681  + 19580.48825767080 *T)
    sumR0  +=       0.051 * cos(3.37515473510  + 9256.99639024040 *T)
    sumR0  +=       0.063 * cos(3.38848970950  + 49515.38250840700 *T)
    sumR0  +=       0.069 * cos(4.90917651401  + 63498.47038145279 *T)
    sumR0  +=       0.057 * cos(5.07437742030  + 18521.10632748160 *T)
    sumR0  +=       0.050 * cos(1.59156823654  + 18631.31264870099 *T)
    sumR0  +=       0.054 * cos(6.25816208666  + 37674.99639427640 *T)
    sumR0  +=       0.057 * cos(5.48065460919  + 24383.07910844140 *T)
    sumR0  +=       0.045 * cos(1.10466490660  + 10408.25693067160 *T)
    sumR0  +=       0.051 * cos(3.61196470313  + 426.59819087600 *T)
    sumR0  +=       0.057 * cos(2.09567711267  + 60530.48898574180 *T)
    sumR0  +=       0.060 * cos(5.94659889997  + 13897.66359620120 *T)
    sumR0  +=       0.051 * cos(5.47238517720  + 57837.13833230060 *T)
    sumR0  +=       0.051 * cos(2.32438478428  + 19779.56025910719 *T)
    sumR0  +=       0.052 * cos(3.23766328818  + 18940.59097135679 *T)
    sumR0  +=       0.043 * cos(5.74921510909  + 51868.24866217880 *T)
    sumR0  +=       0.048 * cos(1.12206254877  + 9779.57380833420 *T)
    sumR0  +=       0.058 * cos(3.08646083897  + 12074.48840752400 *T)
    sumR0  +=       0.046 * cos(4.07536026888  + 7863.94251078820 *T)
    sumR0  +=       0.045 * cos(4.75746520642  + 7856.89627409019 *T)
    sumR0  +=       0.054 * cos(4.43528236634  + 8617.09910392640 *T)
    sumR0  +=       0.050 * cos(3.70569982975  + 42456.78404709160 *T)
    sumR0  +=       0.044 * cos(1.29248911155  + 69166.43098950500 *T)
    sumR0  +=       0.046 * cos(0.41229872114  + 7564.83072073800 *T)
    sumR0  +=       0.044 * cos(6.17937388307  + 13341.67431130680 *T)
    sumR0  +=       0.053 * cos(4.71388531889  + 53445.59220462660 *T)
    sumR0  +=       0.041 * cos(3.48003037828  + 37895.42629036740 *T)
    sumR0  +=       0.040 * cos(1.23305546260  + 10228.53801739600 *T)
    sumR0  +=       0.053 * cos(5.04979874661  + 74673.98422817240 *T)
    sumR0  +=       0.039 * cos(1.36646013032  + 21202.09370374600 *T)
    sumR0  +=       0.039 * cos(2.15376025201  + 8947.71806758460 *T)
    sumR0  +=       0.041 * cos(6.17532984460  + 65236.22129328540 *T)
    sumR0  +=       0.052 * cos(1.29052331493  + 90394.82301305080 *T)
    sumR0  +=       0.039 * cos(0.70253732683  + 18093.37469954999 *T)
    sumR0  +=       0.052 * cos(1.18164377451  + 10211.80107350270 *T)
    sumR0  +=       0.047 * cos(1.78672260794  + 10401.21069397360 *T)
    sumR0  +=       0.040 * cos(3.66961416802  + 10198.03307502600 *T)
    sumR0  +=       0.051 * cos(2.71698589018  + 94325.03270927040 *T)
    sumR0  +=       0.036 * cos(1.25091711620  + 10323.49186743040 *T)
    sumR0  +=       0.049 * cos(1.21335959420  + 9721.62225375220 *T)
    sumR0  +=       0.042 * cos(6.05968230173  + 105460.99111839020 *T)
    sumR0  +=       0.046 * cos(5.06978748275  + 20350.30502114640 *T)
    sumR0  +=       0.040 * cos(1.97645050921  + 32243.49850088060 *T)
    sumR0  +=       0.036 * cos(4.96702216961  + 36301.18868778519 *T)
    sumR0  +=       0.037 * cos(5.29642935562  + 38.02767263580 *T)
    sumR0  +=       0.039 * cos(0.52064327313  + 26709.64694241340 *T)
    sumR0  +=       0.035 * cos(2.34112124655  + 58946.51688439399 *T)
    sumR0  +=       0.034 * cos(1.82989750626  + 17675.02349273040 *T)
    sumR0  +=       0.034 * cos(0.76493664110  + 55798.45835839840 *T)
    sumR0  +=       0.035 * cos(1.09353675147  + 69159.80243060499 *T)
    sumR0  +=       0.031 * cos(5.59148330297  + 10639.88373708700 *T)
    sumR0  +=       0.032 * cos(3.32960781870  + 71519.29714327680 *T)
    sumR0  +=       0.031 * cos(5.98191446392  + 24341.52831745660 *T)
    sumR0  +=       0.031 * cos(0.68615213145  + 10202.23984594710 *T)
    sumR0  +=       0.030 * cos(4.42039942947  + 10459.11719244040 *T)
    sumR0  +=       0.029 * cos(1.30367701539  + 20103.06567576459 *T)
    sumR0  +=       0.031 * cos(4.51793347997  + 2957.71589447660 *T)
    sumR0  +=       0.035 * cos(4.05634321290  + 19903.99367432819 *T)
    sumR0  +=       0.030 * cos(1.32113757427  + 574.34479833480 *T)
    sumR0  +=       0.029 * cos(3.36506645849  + 10288.06714477830 *T)
    sumR0  +=       0.029 * cos(1.40019042576  + 9988.94075050910 *T)
    sumR0  +=       0.032 * cos(0.21932095318  + 24978.52458948080 *T)
    sumR0  +=       0.034 * cos(5.22945947227  + 8673.96967639900 *T)
    sumR0  +=       0.039 * cos(4.50883171158  + 16004.69810374360 *T)
    sumR0  +=       0.028 * cos(2.32945945641  + 11392.48008525060 *T)
    sumR0  +=       0.034 * cos(3.92498967835  + 536.80451209540 *T)
    sumR0  +=       0.032 * cos(5.46972716255  + 64607.84893354619 *T)
    sumR0  +=       0.028 * cos(2.38858990128  + 20235.12282631040 *T)
    sumR0  +=       0.030 * cos(3.34585843979  + 39793.76025465480 *T)
    sumR0  +=       0.026 * cos(5.36096904409  + 1478.86657406440 *T)

    # VENUS Radius
    sumR1  = 0
    sumR1  +=   34551.039 * cos(0.89198710598  + 10213.28554621100 *T)
    sumR1  +=     234.203 * cos(1.77224942714  + 20426.57109242200 *T)
    sumR1  +=     233.998 * cos(3.14159265359  + 0.00000000000 *T)
    sumR1  +=      23.864 * cos(1.11274502648  + 9437.76293488700 *T)
    sumR1  +=      10.568 * cos(4.59168210921  + 1577.34354244780 *T)
    sumR1  +=       9.124 * cos(4.53540907003  + 10404.73381232260 *T)
    sumR1  +=       6.599 * cos(5.97703999838  + 5507.55323866740 *T)
    sumR1  +=       4.667 * cos(3.87683960551  + 9153.90361602180 *T)
    sumR1  +=       3.840 * cos(5.66196924375  + 13367.97263110660 *T)
    sumR1  +=       2.666 * cos(2.82413291285  + 10206.17199921020 *T)
    sumR1  +=       2.194 * cos(2.05314419626  + 775.52261132400 *T)
    sumR1  +=       2.094 * cos(2.55137285015  + 18837.49819713819 *T)
    sumR1  +=       1.782 * cos(2.64808558644  + 30639.85663863300 *T)
    sumR1  +=       1.845 * cos(1.87612936641  + 11015.10647733480 *T)
    sumR1  +=       1.303 * cos(0.20613045603  + 11322.66409830440 *T)
    sumR1  +=       1.169 * cos(0.79431893441  + 17298.18232732620 *T)
    sumR1  +=       1.001 * cos(6.16555101536  + 10239.58386601080 *T)
    sumR1  +=       0.915 * cos(4.59854496966  + 1109.37855209340 *T)
    sumR1  +=       0.884 * cos(0.66706834422  + 18073.70493865020 *T)
    sumR1  +=       0.849 * cos(5.58641571940  + 12566.15169998280 *T)
    sumR1  +=       1.071 * cos(4.94792017474  + 6283.07584999140 *T)
    sumR1  +=       0.887 * cos(2.47785193216  + 3154.68708489560 *T)
    sumR1  +=       0.904 * cos(0.81413053841  + 10596.18207843420 *T)
    sumR1  +=       0.818 * cos(0.90016838097  + 5661.33204915220 *T)
    sumR1  +=       0.845 * cos(5.48504338112  + 529.69096509460 *T)
    sumR1  +=       0.824 * cos(3.74837629121  + 7084.89678111520 *T)
    sumR1  +=       0.652 * cos(5.07444932607  + 22003.91463486980 *T)
    sumR1  +=       0.847 * cos(0.44119876869  + 8635.94200376320 *T)
    sumR1  +=       0.638 * cos(4.10125791268  + 191.44826611160 *T)
    sumR1  +=       0.615 * cos(3.14417599741  + 10186.98722641120 *T)
    sumR1  +=       0.527 * cos(5.86792949279  + 2352.86615377180 *T)
    sumR1  +=       0.520 * cos(5.33201358267  + 14143.49524243060 *T)
    sumR1  +=       0.576 * cos(2.25212731258  + 21228.39202354580 *T)
    sumR1  +=       0.662 * cos(2.86880467345  + 8624.21265092720 *T)
    sumR1  +=       0.554 * cos(2.17186191243  + 18307.80723204360 *T)
    sumR1  +=       0.515 * cos(4.34331395104  + 9786.68735533500 *T)
    sumR1  +=       0.501 * cos(5.56479589366  + 10742.97651130560 *T)
    sumR1  +=       0.426 * cos(1.02161443120  + 7064.12138562280 *T)
    sumR1  +=       0.418 * cos(1.26803034691  + 9676.48103411560 *T)
    sumR1  +=       0.391 * cos(0.78974645621  + 9690.70812811720 *T)
    sumR1  +=       0.334 * cos(3.18175822557  + 10988.80815753500 *T)
    sumR1  +=       0.375 * cos(0.66142254036  + 19360.07561523199 *T)
    sumR1  +=       0.364 * cos(0.19369831864  + 19374.30270923360 *T)
    sumR1  +=       0.313 * cos(1.09734397626  + 4551.95349705880 *T)
    sumR1  +=       0.330 * cos(0.58817502306  + 16496.36139620240 *T)
    sumR1  +=       0.339 * cos(5.76768761396  + 10021.83728009940 *T)
    sumR1  +=       0.291 * cos(3.65846764668  + 25158.60171976540 *T)
    sumR1  +=       0.223 * cos(4.33581625553  + 19786.67380610799 *T)
    sumR1  +=       0.266 * cos(3.57408827667  + 801.82093112380 *T)
    sumR1  +=       0.274 * cos(5.73346687248  + 11790.62908865880 *T)
    sumR1  +=       0.275 * cos(5.65814317085  + 19896.88012732740 *T)
    sumR1  +=       0.212 * cos(4.27038489878  + 4705.73230754360 *T)
    sumR1  +=       0.230 * cos(6.13406345590  + 1059.38193018920 *T)
    sumR1  +=       0.204 * cos(4.87348390351  + 7860.41939243920 *T)
    sumR1  +=       0.241 * cos(1.13551531894  + 26.29831979980 *T)
    sumR1  +=       0.206 * cos(0.31907973682  + 382.89653222320 *T)
    sumR1  +=       0.216 * cos(2.54741101724  + 19651.04848109800 *T)
    sumR1  +=       0.212 * cos(3.15264941106  + 14945.31617355440 *T)
    sumR1  +=       0.163 * cos(1.13604744392  + 13936.79450513400 *T)
    sumR1  +=       0.151 * cos(5.11341268743  + 28521.09277825460 *T)
    sumR1  +=       0.151 * cos(0.81278755582  + 6770.71060124560 *T)
    sumR1  +=       0.150 * cos(5.02227334847  + 29088.81141598500 *T)
    sumR1  +=       0.146 * cos(1.37568138685  + 10220.39909321180 *T)
    sumR1  +=       0.127 * cos(4.49298610074  + 3532.06069281140 *T)
    sumR1  +=       0.121 * cos(6.26589208179  + 29580.47470844380 *T)
    sumR1  +=       0.147 * cos(6.16092774714  + 8662.24032356300 *T)
    sumR1  +=       0.114 * cos(0.00114012635  + 25934.12433108940 *T)
    sumR1  +=       0.115 * cos(3.56897715344  + 24356.78078864160 *T)
    sumR1  +=       0.124 * cos(0.67547060274  + 3723.50895892300 *T)
    sumR1  +=       0.145 * cos(0.36415036222  + 9146.79006902100 *T)
    sumR1  +=       0.104 * cos(4.27865011376  + 9573.38825989700 *T)
    sumR1  +=       0.136 * cos(5.09581116181  + 19367.18916223280 *T)
    sumR1  +=       0.102 * cos(1.53637788668  + 17277.40693183380 *T)
    sumR1  +=       0.117 * cos(0.57543238496  + 9999.98645077300 *T)
    sumR1  +=       0.092 * cos(0.22936081655  + 18830.38465013739 *T)
    sumR1  +=       0.112 * cos(4.04771058036  + 9103.90699411760 *T)
    sumR1  +=       0.098 * cos(3.78447692407  + 213.29909543800 *T)
    sumR1  +=       0.085 * cos(5.84471458481  + 10234.06094170340 *T)
    sumR1  +=       0.079 * cos(0.64440357793  + 18844.61174413899 *T)
    sumR1  +=       0.084 * cos(0.56950139213  + 9683.59458111640 *T)
    sumR1  +=       0.107 * cos(1.77067111589  + 17778.11626694899 *T)
    sumR1  +=       0.081 * cos(6.19048382717  + 20618.01935853360 *T)
    sumR1  +=       0.087 * cos(0.15771136594  + 33019.02111220460 *T)
    sumR1  +=       0.082 * cos(4.80683817059  + 3930.20969621960 *T)
    sumR1  +=       0.086 * cos(2.21505615071  + 8094.52168583260 *T)
    sumR1  +=       0.064 * cos(2.69215119482  + 16983.99614745660 *T)
    sumR1  +=       0.069 * cos(0.83385751986  + 3128.38876509580 *T)
    sumR1  +=       0.081 * cos(4.88025042367  + 4732.03062734340 *T)
    sumR1  +=       0.059 * cos(3.34348033725  + 10787.63034454580 *T)
    sumR1  +=       0.061 * cos(0.04044699966  + 9161.01716302260 *T)
    sumR1  +=       0.064 * cos(4.13127333938  + 9992.87290377220 *T)
    sumR1  +=       0.060 * cos(6.24603986632  + 32217.20018108080 *T)
    sumR1  +=       0.054 * cos(3.38449893196  + 10426.58464164900 *T)
    sumR1  +=       0.054 * cos(5.15939119644  + 28286.99048486120 *T)
    sumR1  +=       0.063 * cos(4.32339245083  + 12592.45001978260 *T)
    sumR1  +=       0.060 * cos(4.48753846170  + 18875.52586977400 *T)
    sumR1  +=       0.057 * cos(3.64912085313  + 10007.09999777380 *T)
    sumR1  +=       0.049 * cos(5.10267262491  + 19573.37471066999 *T)
    sumR1  +=       0.047 * cos(5.79444960738  + 68050.42387851159 *T)
    sumR1  +=       0.052 * cos(3.56658420552  + 7255.56965173440 *T)
    sumR1  +=       0.050 * cos(1.61783309819  + 36949.23080842420 *T)
    sumR1  +=       0.053 * cos(2.64370544855  + 15874.61759536320 *T)
    sumR1  +=       0.040 * cos(3.93466530964  + 20419.45754542119 *T)
    sumR1  +=       0.051 * cos(0.79154899901  + 23581.25817731760 *T)
    sumR1  +=       0.038 * cos(1.77428239418  + 10103.07922499160 *T)
    sumR1  +=       0.049 * cos(1.12423644455  + 3442.57494496540 *T)
    sumR1  +=       0.040 * cos(5.22874487975  + 21535.94964451540 *T)
    sumR1  +=       0.038 * cos(1.12473430132  + 7.11354700080 *T)
    sumR1  +=       0.038 * cos(0.11510547453  + 11272.66747640020 *T)
    sumR1  +=       0.036 * cos(2.02476324983  + 7058.59846131540 *T)
    sumR1  +=       0.047 * cos(0.05589432390  + 12432.04265039780 *T)
    sumR1  +=       0.034 * cos(3.45481114998  + 9830.38901398780 *T)
    sumR1  +=       0.045 * cos(4.59817214088  + 10192.51015071860 *T)
    sumR1  +=       0.037 * cos(4.93959675364  + 3340.61242669980 *T)
    sumR1  +=       0.044 * cos(0.70533027806  + 20213.27199698400 *T)
    sumR1  +=       0.034 * cos(2.16487642765  + 64460.69868196140 *T)
    sumR1  +=       0.031 * cos(1.57612397319  + 36147.40987730040 *T)
    sumR1  +=       0.028 * cos(2.56454760402  + 94138.32702008578 *T)
    sumR1  +=       0.033 * cos(1.08907268562  + 29864.33402730900 *T)
    sumR1  +=       0.029 * cos(0.59718407064  + 59728.66805461800 *T)
    sumR1  +=       0.031 * cos(3.04423979263  + 40879.44050464380 *T)
    sumR1  +=       0.035 * cos(0.32247158762  + 1589.07289528380 *T)
    sumR1  +=       0.031 * cos(3.27727318906  + 19992.85935454519 *T)
    sumR1  +=       0.027 * cos(5.83705748551  + 17085.95866572220 *T)
    sumR1  +=       0.032 * cos(2.64260788260  + 41962.52073693740 *T)
    sumR1  +=       0.028 * cos(4.90613317287  + 29050.78374334920 *T)
    sumR1  +=       0.025 * cos(4.55050389739  + 14919.01785375460 *T)
    sumR1  +=       0.028 * cos(3.58851614957  + 40853.14218484400 *T)
    sumR1  +=       0.029 * cos(2.79705093386  + 20007.08644854680 *T)
    sumR1  +=       0.033 * cos(0.93862065616  + 15720.83878487840 *T)
    sumR1  +=       0.024 * cos(2.74970637101  + 18947.70451835760 *T)
    sumR1  +=       0.024 * cos(4.38966861409  + 46386.99374331120 *T)
    sumR1  +=       0.024 * cos(0.73361964525  + 9411.46461508720 *T)
    sumR1  +=       0.028 * cos(4.19559784013  + 37674.99639427640 *T)
    sumR1  +=       0.023 * cos(1.00023735538  + 22779.43724619380 *T)
    sumR1  +=       0.026 * cos(0.46990555736  + 13745.34623902240 *T)
    sumR1  +=       0.028 * cos(4.65181292126  + 1551.04522264800 *T)
    sumR1  +=       0.025 * cos(4.18690270765  + 44007.82926973960 *T)
    sumR1  +=       0.022 * cos(0.98102807789  + 426.59819087600 *T)
    sumR1  +=       0.030 * cos(1.24986033487  + 27461.71084806540 *T)
    sumR1  +=       0.027 * cos(3.94986823486  + 17468.85519794540 *T)
    sumR1  +=       0.021 * cos(6.09897508157  + 18300.69368504279 *T)
    sumR1  +=       0.025 * cos(4.75875623888  + 27991.40181316000 *T)
    sumR1  +=       0.022 * cos(2.95281481673  + 40077.61957352000 *T)
    sumR1  +=       0.028 * cos(6.12038264955  + 38500.27603107220 *T)
    sumR1  +=       0.022 * cos(4.11184201321  + 19779.56025910719 *T)
    sumR1  +=       0.027 * cos(3.72446446080  + 19793.78735310880 *T)
    sumR1  +=       0.020 * cos(4.27086627368  + 31441.67756975680 *T)
    sumR1  +=       0.022 * cos(4.99040169444  + 31022.75317085620 *T)
    sumR1  +=       0.023 * cos(1.33505132122  + 65236.22129328540 *T)
    sumR1  +=       0.021 * cos(4.46897353468  + 53285.18483524180 *T)
    sumR1  +=       0.020 * cos(4.15140915983  + 2218.75710418680 *T)
    sumR1  +=       0.025 * cos(2.18447182965  + 27511.46787353720 *T)
    sumR1  +=       0.019 * cos(1.43653410349  + 27197.28169366760 *T)
    sumR1  +=       0.027 * cos(1.22555218015  + 42430.48572729180 *T)
    sumR1  +=       0.019 * cos(3.65054338893  + 49515.38250840700 *T)
    sumR1  +=       0.022 * cos(5.88380811711  + 10218.80847051840 *T)
    sumR1  +=       0.018 * cos(2.29853355765  + 19264.09638801420 *T)
    sumR1  +=       0.017 * cos(5.44429906531  + 6681.22485339960 *T)
    sumR1  +=       0.020 * cos(3.68116637773  + 14128.24277124560 *T)
    sumR1  +=       0.021 * cos(4.30316190532  + 44809.65020086340 *T)
    sumR1  +=       0.020 * cos(2.48583613985  + 33794.54372352860 *T)
    sumR1  +=       0.017 * cos(3.02735393984  + 28528.20632525540 *T)
    sumR1  +=       0.019 * cos(5.92656850674  + 22805.73556599360 *T)
    sumR1  +=       0.022 * cos(5.30827572791  + 10207.76262190360 *T)
    sumR1  +=       0.020 * cos(0.75829381378  + 18314.92077904440 *T)
    sumR1  +=       0.017 * cos(5.63315744126  + 16522.65971600220 *T)
    sumR1  +=       0.016 * cos(1.71021408448  + 536.80451209540 *T)
    sumR1  +=       0.015 * cos(5.27016880041  + 53445.59220462660 *T)
    sumR1  +=       0.017 * cos(5.61443395877  + 47938.03896595920 *T)
    sumR1  +=       0.015 * cos(5.81110284451  + 43071.89928903080 *T)
    sumR1  +=       0.015 * cos(4.96237667003  + 19999.97290154599 *T)
    sumR1  +=       0.018 * cos(0.55618686515  + 14765.23904326980 *T)
    sumR1  +=       0.014 * cos(3.48144272414  + 29786.66025688100 *T)
    sumR1  +=       0.015 * cos(5.84132627836  + 10228.53801739600 *T)
    sumR1  +=       0.016 * cos(1.05720065324  + 26735.94526221320 *T)
    sumR1  +=       0.014 * cos(6.08462030302  + 35371.88726597640 *T)
    sumR1  +=       0.014 * cos(2.84532871890  + 574.34479833480 *T)
    sumR1  +=       0.015 * cos(5.34517715140  + 10198.03307502600 *T)
    sumR1  +=       0.013 * cos(0.45004137509  + 20452.86941222180 *T)

    # VENUS Radius
    sumR2  = 0
    sumR2  +=    1406.587 * cos(5.06366395190  + 10213.28554621100 *T)
    sumR2  +=      15.529 * cos(5.47321687981  + 20426.57109242200 *T)
    sumR2  +=      13.059 * cos(0.00000000000  + 0.00000000000 *T)
    sumR2  +=       1.099 * cos(2.78883988292  + 9437.76293488700 *T)
    sumR2  +=       0.488 * cos(6.27806914496  + 1577.34354244780 *T)
    sumR2  +=       0.361 * cos(6.11914188253  + 10404.73381232260 *T)
    sumR2  +=       0.310 * cos(1.38984998403  + 5507.55323866740 *T)
    sumR2  +=       0.389 * cos(1.95017779915  + 11015.10647733480 *T)
    sumR2  +=       0.372 * cos(2.33222828423  + 775.52261132400 *T)
    sumR2  +=       0.207 * cos(5.63406721595  + 10239.58386601080 *T)
    sumR2  +=       0.168 * cos(1.10765197296  + 13367.97263110660 *T)
    sumR2  +=       0.175 * cos(6.16674652950  + 30639.85663863300 *T)
    sumR2  +=       0.168 * cos(3.64495311632  + 7084.89678111520 *T)
    sumR2  +=       0.120 * cos(5.85815843789  + 9153.90361602180 *T)
    sumR2  +=       0.160 * cos(2.21564938463  + 3154.68708489560 *T)
    sumR2  +=       0.118 * cos(2.62358866565  + 8635.94200376320 *T)
    sumR2  +=       0.112 * cos(2.36235956804  + 10596.18207843420 *T)
    sumR2  +=       0.092 * cos(0.72664449269  + 12566.15169998280 *T)
    sumR2  +=       0.067 * cos(3.76089669118  + 18837.49819713819 *T)
    sumR2  +=       0.065 * cos(2.47983709990  + 11790.62908865880 *T)
    sumR2  +=       0.048 * cos(4.26620187144  + 2352.86615377180 *T)
    sumR2  +=       0.048 * cos(5.50898189550  + 191.44826611160 *T)
    sumR2  +=       0.048 * cos(2.54730918293  + 17298.18232732620 *T)
    sumR2  +=       0.046 * cos(3.40293459332  + 14143.49524243060 *T)
    sumR2  +=       0.041 * cos(1.83997113019  + 11322.66409830440 *T)
    sumR2  +=       0.037 * cos(6.17871126027  + 1109.37855209340 *T)
    sumR2  +=       0.039 * cos(4.77190210316  + 18073.70493865020 *T)
    sumR2  +=       0.035 * cos(3.10133256432  + 4705.73230754360 *T)
    sumR2  +=       0.046 * cos(3.30090415967  + 6283.07584999140 *T)
    sumR2  +=       0.034 * cos(3.91721765773  + 10021.83728009940 *T)
    sumR2  +=       0.034 * cos(3.24663787383  + 22003.91463486980 *T)
    sumR2  +=       0.042 * cos(3.39360926939  + 14945.31617355440 *T)
    sumR2  +=       0.044 * cos(4.42979374073  + 7860.41939243920 *T)
    sumR2  +=       0.034 * cos(2.16381407025  + 16496.36139620240 *T)
    sumR2  +=       0.031 * cos(0.45714618479  + 26.29831979980 *T)
    sumR2  +=       0.035 * cos(3.62868651241  + 801.82093112380 *T)
    sumR2  +=       0.032 * cos(1.84138997078  + 382.89653222320 *T)
    sumR2  +=       0.025 * cos(3.32908650295  + 18307.80723204360 *T)
    sumR2  +=       0.026 * cos(3.64313769818  + 29088.81141598500 *T)
    sumR2  +=       0.029 * cos(3.82967178810  + 10186.98722641120 *T)
    sumR2  +=       0.022 * cos(3.17741520378  + 28521.09277825460 *T)
    sumR2  +=       0.021 * cos(2.52643834111  + 529.69096509460 *T)
    sumR2  +=       0.025 * cos(5.71401244457  + 21202.09370374600 *T)
    sumR2  +=       0.021 * cos(3.77813434325  + 21228.39202354580 *T)
    sumR2  +=       0.019 * cos(5.24505118517  + 19896.88012732740 *T)
    sumR2  +=       0.018 * cos(4.62463651925  + 19651.04848109800 *T)
    sumR2  +=       0.016 * cos(3.35893297896  + 28286.99048486120 *T)
    sumR2  +=       0.015 * cos(5.05571633205  + 33019.02111220460 *T)
    sumR2  +=       0.014 * cos(2.83786355803  + 19786.67380610799 *T)
    sumR2  +=       0.014 * cos(1.79922718553  + 9830.38901398780 *T)
    sumR2  +=       0.014 * cos(3.14801263138  + 19367.18916223280 *T)
    sumR2  +=       0.014 * cos(3.57896195191  + 10988.80815753500 *T)
    sumR2  +=       0.013 * cos(3.06303088617  + 10742.97651130560 *T)
    sumR2  +=       0.013 * cos(5.43981998532  + 25158.60171976540 *T)
    sumR2  +=       0.015 * cos(4.83166312889  + 18875.52586977400 *T)
    sumR2  +=       0.012 * cos(2.54141086214  + 7064.12138562280 *T)
    sumR2  +=       0.012 * cos(4.45255110769  + 15720.83878487840 *T)
    sumR2  +=       0.010 * cos(1.87933121728  + 24356.78078864160 *T)
    sumR2  +=       0.011 * cos(2.58708635685  + 9103.90699411760 *T)
    sumR2  +=       0.010 * cos(2.17901309900  + 3723.50895892300 *T)
    sumR2  +=       0.008 * cos(3.63520673832  + 1059.38193018920 *T)
    sumR2  +=       0.008 * cos(4.67523115598  + 25934.12433108940 *T)
    sumR2  +=       0.009 * cos(5.97856553283  + 9683.59458111640 *T)

    # VENUS Radius
    sumR3  = 0
    sumR3  +=      49.582 * cos(3.22263554520  + 10213.28554621100 *T)
    sumR3  +=       0.831 * cos(3.21219077104  + 20426.57109242200 *T)
    sumR3  +=       0.112 * cos(3.14159265359  + 0.00000000000 *T)
    sumR3  +=       0.013 * cos(3.77448689585  + 30639.85663863300 *T)
    sumR3  +=       0.009 * cos(4.19802043629  + 10239.58386601080 *T)
    sumR3  +=       0.006 * cos(0.20714935358  + 10186.98722641120 *T)
    sumR3  +=       0.005 * cos(0.68781956122  + 8635.94200376320 *T)

    # VENUS Radius
    sumR4  = 0
    sumR4  +=       0.573 * cos(0.92229697820  + 10213.28554621100 *T)
    sumR4  +=       0.040 * cos(0.95468912157  + 20426.57109242200 *T)
    sumR4  +=       0.006 * cos(3.14159265359  + 0.00000000000 *T)

    # VENUS Radius
    sumR5  = 0
    sumR5  +=       0.045 * cos(0.30032866722  + 10213.28554621100 *T)
    sumR5  +=       0.002 * cos(5.29627718483  + 20426.57109242200 *T)

    # calculate and scale the sums
    L = sumL0 + sumL1*T**1 + sumL2*T**2 + sumL3*T**3 + sumL4*T**4 + sumL5*T**5
    L /= 1e8
    L = Mod(degrees(L), 360.)
    B = sumB0 + sumB1*T**1 + sumB2*T**2 + sumB3*T**3 + sumB4*T**4 + sumB5*T**5
    B /= 1e8
    B = degrees(B)
    R = sumR0 + sumR1*T**1 + sumR2*T**2 + sumR3*T**3 + sumR4*T**4 + sumR5*T**5
    R /= 1e8

    # calculate X, Y, Z
    X = R * Cos(B) * Cos(L)
    Y = R * Cos(B) * Sin(L)
    Z = R * Sin(B)

    # create dictionary
    Dict = {  "Object"  : "Venus",
              "Mode"    : "LBR",
              "Long"    : L,
              "Lat"     : B,
              "Rau"     : R,
              "X"       : X,
              "Y"       : Y,
              "Z"       : Z,
              "Alt"     : -34./60.
           }
    Data = (L, B, R, Dict)


    return Data
#######################################################################



def getVenusData(dtCivicTime, zone, verbose = False):
    L, B, R, Dict = getVenusLBRData(dtCivicTime, zone, verbose)
    X = Dict["X"]; Y = Dict["Y"]; Z = Dict["Z"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Le, Be, Re, Dicte = getEarthLBRData(dtCivicTime, zone, verbose)
        Xe = Dicte["X"]; Ye = Dicte["Y"]; Ze = Dicte["Z"]
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


# Planetary Phenomena - Jean Meeus Chapter 35
# hvf 08.09.2021 30.09.2022 02.10.2023


from itertools import chain

VersionPlanets = "V2.1 hvf 2023-10-02 Planets.py based on Meeus"

def getPerihelionAphelion(planet, yyyy, verbose=False):
    def getJDE_peri_apo(planet, k):
    # Meeus chapter 37 
    # k integer: perihelion
    # k + 0.5  : aphelion
        if planet == "Mercury":
            JDE = 2451590.257 + 87.96934963 * k
        elif planet == "Venus":
            JDE = 2451738.233 + 224.7008187 * k - 0.0327E-6 * k*k
        elif planet == "Earth":
            JDE = 2451547.507 + 365.2596358 * k + 0.0158E-6 * k*k
        elif planet == "Mars":
            JDE = 2452195.026 + 686.9957843 * k - 0.1187E-6 * k*k
        elif planet == "Jupiter":
            JDE = 2455636.938 + 4332.897090 * k + 0.1368E-3 * k*k
        elif planet == "Saturn":
            JDE = 2452730.11  + 10764.21731 * k + 0.826E-3 * k*k
        elif planet == "Uranus":
            JDE = 2470213.5   + 30694.8767  * k - 0.00541  * k*k
        elif planet == "Neptune":
            JDE = 2468895.7   + 60190.32    * k + 0.03175  * k*k
        return JDE

    def getk(planet, y):
    # Meeus chapter 37 
    # k integer: perihelion
    # k + 0.5  : aphelion
        if planet == "Mercury":
            k = 4.15201 * (y - 2000.12)
        elif planet == "Venus":
            k = 1.62549 * (y - 2000.53)
        elif planet == "Earth":
            k = 0.99997 * (y - 2000.01)
        elif planet == "Mars":
            k = 0.53166 * (y - 2001.78)
        elif planet == "Jupiter":
            k = 0.08430 * (y - 2011.20)
        elif planet == "Saturn":
            k = 0.03393 * (y - 2003.52)
        elif planet == "Uranus":
            k = 0.01190 * (y - 2051.1)
        elif planet == "Neptune":
            k = 0.00607 * (y - 2047.5)
        return k

    # Dictionary with the functions to get planet data
    PlanetGet  = {
                "Venus"     : getVenusData,
                "Mars"      : getMarsData,
                "Jupiter"   : getJupiterData,
                "Saturn"    : getSaturnData,
                "Mercury"   : getMercuryData,
                "Earth"     : getEarthData,
                "Uranus"    : getUranusData,
                "Neptune"   : getNeptuneData,
#                "Sun"       : getSunData,
#                "Moon"      : getMoonData,
                }

        
    PeriAphList = []
    typeText = None
    kfloat = getk(planet, yyyy)
    kstart = floor(kfloat)  # perihelion before yyyy
    for kperi in range(kstart, kstart+5):
        #print ("kperi = ", kperi)
        kaph = kperi + 0.5  # aphelion
        JDEp = getJDE_peri_apo(planet, kperi)
        JDEa = getJDE_peri_apo(planet, kaph)

        DTp = hvfTime(JDEp)
        DTa = hvfTime(JDEa)

        getdata = PlanetGet[planet]
        RA, Decl, D = getdata(DTp, 0., False)
        Raup = D["Rau"]

        Eperi = {   "Object"    : planet,
                "JDE"       : JDEp,
                "DT"        : DTp,
                "Type"      : "Perihel",
                "Rau"       : Raup,
            }

        RA, Decl, D = getdata(DTa, 0., False)
        Raua = D["Rau"]
        Eaph  = {   "Object"    : planet,
                "JDE"       : JDEa,
                "DT"        : DTa,
                "Type"      : "Aphel",
                "Rau"       : Raua,
            }
        (y, m, d) = DTp.getDate()
        if y == yyyy: PeriAphList.append(Eperi)
        (y, m, d) = DTa.getDate()
        if y == yyyy: PeriAphList.append(Eaph )
    return PeriAphList

######################33
def getOppoConjforPlanet(planet, isInfConj, yyyy, verbose = False):
    # inferior conjunction (Mercury, Venus) is Opposition for the other planets
    isOppo = isInfConj
    typeText = None
    if planet == "Mercury" and isInfConj: # Erdnaehe
        typeText = "Inferior Conjunction"
        # Table 35.A
        A = 2451612.023; B = 115.8774771
        M0 = 63.5867;   M1 = 114.2088742
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " inferior conjunction (near Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # corrections of table 35.B
        corr =               0.0545 + 0.0002*T
        corr += Sin(M) *   (-6.2008 + 0.0074*T + 0.00003*T*T)
        corr += Cos(M) *   (-3.2750 - 0.0197*T + 0.00001*T*T)
        corr += Sin(2*M) * ( 0.4737 - 0.0052*T - 0.00001*T*T)
        corr += Cos(2*M) * ( 0.8111 + 0.0033*T - 0.00002*T*T)
        corr += Sin(3*M) * ( 0.0037 + 0.0018*T)
        corr += Cos(3*M) * (-0.1768            + 0.00001*T*T)
        corr += Sin(4*M) * (-0.0211 - 0.0004*T)
        corr += Cos(4*M) * ( 0.0326 - 0.0003*T)
        corr += Sin(5*M) * ( 0.0083 + 0.0001*T)
        corr += Cos(5*M) * (-0.0040 + 0.0001*T)

    elif planet == "Mercury" and not isInfConj: # Erdferne, hinter der Sonne
        typeText = "Superior Conjunction"
        # Table 35.A
        A = 2451554.084; B = 115.8774771
        M0 =  6.4822;   M1 = 114.2088742
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " superior conjunction (far Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # corrections of table 35.B
        corr =              -0.0548 - 0.0002*T
        corr += Sin(M) *   ( 7.3894 - 0.0100*T - 0.00003*T*T)
        corr += Cos(M) *   ( 3.2200 + 0.0197*T - 0.00001*T*T)
        corr += Sin(2*M) * ( 0.8383 - 0.0064*T - 0.00001*T*T)
        corr += Cos(2*M) * ( 0.9666 + 0.0039*T - 0.00003*T*T)
        corr += Sin(3*M) * ( 0.0770 - 0.0026*T)
        corr += Cos(3*M) * ( 0.2758 + 0.0002*T - 0.00002*T*T)
        corr += Sin(4*M) * (-0.0128 - 0.0008*T)
        corr += Cos(4*M) * ( 0.0734 - 0.0004*T - 0.00001*T*T)
        corr += Sin(5*M) * (-0.0122 - 0.0002*T)
        corr += Cos(5*M) * ( 0.0173 - 0.0002*T)

    elif planet == "Venus" and isInfConj:   # Erdnaehe
        typeText = "Inferior Conjunction"
        # Table 35.A
        A = 2451996.706; B = 583.921361
        M0 = 82.7311;   M1 = 215.513058
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " inferior conjunction (near Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # corrections of table 35.B
        corr =              -0.0096 + 0.0002*T - 0.00001*T*T
        corr += Sin(M) *   ( 2.0009 - 0.0033*T - 0.00001*T*T)
        corr += Cos(M) *   ( 0.5980 - 0.0104*T + 0.00001*T*T)
        corr += Sin(2*M) * ( 0.0967 - 0.0018*T - 0.00003*T*T)
        corr += Cos(2*M) * ( 0.0913 + 0.0009*T - 0.00002*T*T)
        corr += Sin(3*M) * ( 0.0046 - 0.0002*T)
        corr += Cos(3*M) * ( 0.0079 + 0.0001*T)

    elif planet == "Venus" and not isInfConj:   # Erdferne, hinter der Sonne
        typeText = "Superior Conjunction"
        # Table 35.A
        A = 2451704.746; B = 583.921361
        M0 = 154.9745;  M1 = 215.513058
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " superior conjunction (far Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # corrections of table 35.B
        corr =               0.0099 - 0.0002*T - 0.00001*T*T
        corr += Sin(M) *   ( 4.1991 - 0.0121*T - 0.00003*T*T)
        corr += Cos(M) *   (-0.6095 + 0.0102*T - 0.00002*T*T)
        corr += Sin(2*M) * ( 0.2500 - 0.0028*T - 0.00003*T*T)
        corr += Cos(2*M) * ( 0.0063 + 0.0025*T - 0.00002*T*T)
        corr += Sin(3*M) * ( 0.0232 - 0.0005*T - 0.00001*T*T)
        corr += Cos(3*M) * ( 0.0031 + 0.0004*T)

    elif planet == "Mars" and isOppo: # Erdnaehe, Opposition
        typeText = "Opposition          "
        # Table 35.A
        A = 2452097.382; B = 779.936104
        M0 = 181.9573;  M1 =  48.705244
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Opposition (near Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # corrections of table 35.B
        corr =               -0.3088            + 0.00002*T*T
        corr += Sin(M) *   (-17.6965 + 0.0363*T + 0.00005*T*T)
        corr += Cos(M) *   (18.3131 + 0.0467*T - 0.00006*T*T)
        corr += Sin(2*M) * (-0.2162 - 0.0198*T - 0.00001*T*T)
        corr += Cos(2*M) * (-4.5028 - 0.0019*T + 0.00007*T*T)
        corr += Sin(3*M) * ( 0.8987 + 0.0058*T - 0.00002*T*T)
        corr += Cos(3*M) * ( 0.7666 - 0.0050*T - 0.00003*T*T)
        corr += Sin(4*M) * (-0.3636 - 0.0001*T + 0.00002*T*T)
        corr += Cos(4*M) * ( 0.0402 + 0.0032*T)
        corr += Sin(5*M) * ( 0.0737 - 0.0008*T)
        corr += Cos(5*M) * (-0.0980 - 0.0011*T)

    elif planet == "Mars" and not isOppo: # Erdferne, Konjunktion, hinter S.
        typeText = "Superior Conjunction"
        # Table 35.A
        A = 2451707.414; B = 779.936104
        M0 = 157.6047;  M1 =  48.705244
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Conjunktion (far Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # corrections of table 35.B
        corr =               0.3102 - 0.0001*T + 0.00001*T*T
        corr += Sin(M) *   ( 9.7273 - 0.0156*T + 0.00001*T*T)
        corr += Cos(M) *  (-18.3195 - 0.0467*T + 0.00009*T*T)
        corr += Sin(2*M) * (-1.6488 - 0.0133*T + 0.00001*T*T)
        corr += Cos(2*M) * (-2.6117 - 0.0020*T + 0.00004*T*T)
        corr += Sin(3*M) * (-0.6827 - 0.0026*T + 0.00001*T*T)
        corr += Cos(3*M) * ( 0.0281 + 0.0035*T + 0.00001*T*T)
        corr += Sin(4*M) * (-0.0823 + 0.0006*T + 0.00001*T*T)
        corr += Cos(4*M) * ( 0.1584 + 0.0013*T)
        corr += Sin(5*M) * ( 0.0270 + 0.0005*T)
        corr += Cos(5*M) * ( 0.0433)

    elif planet == "Jupiter" and isOppo: # Erdnaehe, Opposition
        typeText = "Opposition          "
        # Table 35.A
        A = 2451870.628; B = 398.884046
        M0 = 318.4681;  M1 =  33.140229
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Opposition (near Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # additional angle for Jupiter (p. 251)
        a = 82.74 + 40.76*T

        # corrections of table 35.B
        corr = -0.1029 - 0.00009*T*T
        corr += Sin(M) *   ( -1.9658 - 0.0056*T + 0.00007*T*T)
        corr += Cos(M) *   (  6.1537 + 0.0210*T - 0.00006*T*T)
        corr += Sin(2*M) * ( -0.2081 - 0.0013*T)
        corr += Cos(2*M) * ( -0.1116 - 0.0010*T)
        corr += Sin(3*M) * (  0.0074 + 0.0001*T)
        corr += Cos(3*M) * ( -0.0097 - 0.0001*T)
        corr += Sin(a  ) * (  0.0    + 0.0144*T - 0.00008*T*T)
        corr += Cos(a  ) * (  0.3642 - 0.0019*T - 0.00029*T*T)

    elif planet == "Jupiter" and not isOppo: # Erdferne, Konjunktion
        typeText = "Superior Conjunction"
        # Table 35.A
        A = 2451671.186; B = 398.884046
        M0 = 121.8980;  M1 =  33.140229
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Conjunction (far Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # additional angle for Jupiter (p. 251)
        a = 82.74 + 40.76*T

        # corrections of table 35.B
        corr =               0.1027 + 0.0002*T - 0.00009*T*T
        corr += Sin(M) *   (-2.2637 + 0.0163*T - 0.00003*T*T)
        corr += Cos(M) *   (-6.1540 - 0.0210*T + 0.00008*T*T)
        corr += Sin(2*M) * (-0.2021 - 0.0017*T + 0.00001*T*T)
        corr += Cos(2*M) * ( 0.1310 - 0.0008*T)
        corr += Sin(3*M) * ( 0.0086)
        corr += Cos(3*M) * ( 0.0087 + 0.0002*T)
        corr += Sin(a  ) * ( 0.0    + 0.0144*T - 0.00008*T*T)
        corr += Cos(a  ) * ( 0.3642 - 0.0019*T - 0.00029*T*T)

    elif planet == "Saturn" and isOppo: # Erdnaehe, Opposition
        typeText = "Opposition          "
        # Table 35.A
        A = 2451870.170; B = 378.091904
        M0 = 318.0172;  M1 =  12.647487
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Opposition (near Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # additional angle for Saturn (p. 251)
        a = 82.74 + 40.76*T
        b = 29.86 + 1181.36*T
        c = 14.13 + 590.68*T
        d = 220.02 + 1262.87*T

        # corrections of table 35.B
        corr =               -0.0209 + 0.0006*T + 0.00023*T*T
        corr += Sin(M) *   (  4.5795 - 0.0312*T - 0.00017*T*T)
        corr += Cos(M) *   (  1.1462 - 0.0351*T + 0.00011*T*T)
        corr += Sin(2*M) * (  0.0985 - 0.0015*T)
        corr += Cos(2*M) * (  0.0733 - 0.0031*T + 0.00001*T*T)
        corr += Sin(3*M) * (  0.0025 - 0.0001*T)
        corr += Cos(3*M) * (  0.0050 - 0.0002*T)
        corr += Sin(a  ) * (  0.0    - 0.0337*T + 0.00018*T*T)
        corr += Cos(a  ) * ( -0.8510 + 0.0044*T + 0.00068*T*T)
        corr += Sin(b  ) * (  0.0    - 0.0064*T + 0.00004*T*T)
        corr += Cos(b  ) * (  0.2397 - 0.0012*T - 0.00008*T*T)
        corr += Sin(c  ) * (  0.0    - 0.0010*T)
        corr += Cos(c  ) * (  0.1245 + 0.0006*T)
        corr += Sin(d  ) * (  0.0    + 0.0024*T - 0.00003*T*T)
        corr += Cos(d  ) * (  0.0477 - 0.0005*T - 0.00006*T*T)


    elif planet == "Saturn" and not isOppo: # Erdferne, Konjunktion
        typeText = "Superior Conjunction"
        # Table 35.A
        A = 2451681.124; B = 378.091904
        M0 = 131.6934;  M1 =  12.647487
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Conjunction (far Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # additional angle for Saturn (p. 251)
        a = 82.74 + 40.76*T
        b = 29.86 + 1181.36*T
        c = 14.13 + 590.68*T
        d = 220.02 + 1262.87*T

        # corrections of table 35.B
        corr =               0.0172 - 0.0006*T + 0.00023*T*T
        corr += Sin(M) *   (-8.5885 + 0.0411*T + 0.00020*T*T)
        corr += Cos(M) *   (-1.1470 + 0.0352*T - 0.00011*T*T)
        corr += Sin(2*M) * ( 0.3331 - 0.0034*T - 0.00001*T*T)
        corr += Cos(2*M) * ( 0.1145 - 0.0045*T + 0.00002*T*T)
        corr += Sin(3*M) * (-0.0169 + 0.0002*T)
        corr += Cos(3*M) * (-0.0109 + 0.0004*T)
        corr += Sin(a  ) * ( 0.0    - 0.0337*T + 0.00018*T*T)
        corr += Cos(a  ) * (-0.8510 + 0.0044*T + 0.00068*T*T)
        corr += Sin(b  ) * (  0.0    - 0.0064*T + 0.00004*T*T)
        corr += Cos(b  ) * (  0.2397 - 0.0012*T - 0.00008*T*T)
        corr += Sin(c  ) * (  0.0    - 0.0010*T)
        corr += Cos(c  ) * (  0.1245 + 0.0006*T)
        corr += Sin(d  ) * (  0.0    + 0.0024*T - 0.00003*T*T)
        corr += Cos(d  ) * (  0.0477 - 0.0005*T - 0.00006*T*T)

    elif planet == "Uranus" and isOppo: # Erdnaehe, Opposition
        typeText = "Opposition          "
        # Table 35.A
        A = 2451764.317; B = 369.656035
        M0 = 213.6884;  M1 =   4.333093
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Opposition (near Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # additional angle for Uranus (p. 251)
        e = 207.83 + 8.51*T
        f = 108.84 + 419.96*T

        # corrections of table 35.B
        corr =                0.0844 - 0.0006*T
        corr += Sin(M) *   ( -0.1048 + 0.0246*T)
        corr += Cos(M) *   ( -5.1221 + 0.0104*T + 0.00003*T*T)
        corr += Sin(2*M) * ( -0.1428 + 0.0005*T)
        corr += Cos(2*M) * ( -0.0148 - 0.0013*T)
        corr += Sin(3*M) * (  0.0)
        corr += Cos(3*M) * (  0.0055)
        corr += Cos(e  ) * (  0.8850)
        corr += Cos(f  ) * (  0.2153)


    elif planet == "Uranus" and not isOppo: # Erdferne, Konjunktion
        typeText = "Superior Conjunction"
        # Table 35.A
        A = 2451579.489; B = 369.656035
        M0 =  31.5219;  M1 =   4.333093
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Opposition (near Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # additional angle for Uranus (p. 251)
        e = 207.83 + 8.51*T
        f = 108.84 + 419.96*T

        # corrections of table 35.B
        corr =               -0.0859 + 0.0003*T
        corr += Sin(M) *   ( -3.8179 - 0.0148*T + 0.00003*T*T)
        corr += Cos(M) *   (  5.1228 - 0.0105*T - 0.00002*T*T)
        corr += Sin(2*M) * ( -0.0803 + 0.0011*T)
        corr += Cos(2*M) * ( -0.1905 - 0.0006*T)
        corr += Sin(3*M) * (  0.0088 + 0.0001*T)
        corr += Cos(3*M) * (  0.0)
        corr += Cos(e  ) * (  0.8850)
        corr += Cos(f  ) * (  0.2153)


    elif planet == "Neptune" and isOppo: # Erdnaehe, Opposition
        typeText = "Opposition          "
        # Table 35.A
        A = 2451753.122; B = 367.486703
        M0 = 202.6544;  M1 =   2.194998
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Opposition (near Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # additional angle for Neptune (p. 251)
        e = 207.83 + 8.51*T
        g = 276.74 + 209.98*T

        # corrections of table 35.B
        corr =               -0.0140            + 0.00001*T*T
        corr += Sin(M) *   ( -1.3486 + 0.0010*T + 0.00001*T*T)
        corr += Cos(M) *   (  0.8597 + 0.0037*T)
        corr += Sin(2*M) * ( -0.0082 - 0.0002*T + 0.00001*T*T)
        corr += Cos(2*M) * (  0.0037 - 0.0003*T)
        corr += Cos(e  ) * ( -0.5964)
        corr += Cos(g  ) * (  0.0728)


    elif planet == "Neptune" and not isOppo: # Erdferne, Konjunktion
        typeText = "Superior Conjunction"
        # Table 35.A
        A = 2451569.379; B = 367.486703
        M0 =  21.5569;  M1 =   2.194998
        kfloat = (365.2425 * yyyy + 1721060 - A)/B  # (36.1)
        # nearest integer
        k = int(round(kfloat, 0))
        if verbose: 
            print (planet, " Opposition (near Earth)", yyyy, 
                    "kfloat = %.3f" % kfloat, "k =", k)

        JDE0 = A + k*B; M = M0 + k*M1; M = Mod(M, 360.)
        T = (JDE0 - 2451545)/36525  # centuries since 2000

        # additional angle for Neptune (p. 251)
        e = 207.83 + 8.51*T
        g = 276.74 + 209.98*T

        # corrections of table 35.B
        corr =                0.0168
        corr += Sin(M) *   ( -2.5606 + 0.0088*T + 0.00002*T*T)
        corr += Cos(M) *   ( -0.8611 - 0.0037*T + 0.00002*T*T)
        corr += Sin(2*M) * (  0.0118 - 0.0004*T + 0.00001*T*T)
        corr += Cos(2*M) * (  0.0307 - 0.0003*T)
        corr += Cos(e  ) * ( -0.5964)
        corr += Cos(g  ) * (  0.0728)

    ###################################################################
    # common part for all planets
    # apply corr and convert to JDE
    JDE = JDE0 + corr
    DT = hvfTime(JDE)
    DTrounded = roundhvfTime(DT, 3600)     # round to nearest hour

    Event = {   "Object"    : planet,
                "JDE0"      : JDE0,     # used for elongation
                "JDE"       : JDE,
                "M"         : M, 
                "T"         : T,
                "DT"        : DT,
                "DTrounded" : DTrounded,
                "Type"      : typeText,
                "elon"      : None,     # used for elongation
            }

    return Event

####################################################################### 

# get eastern (visible in the evening) and western (visible in the morning)
# elongations for marcur and Venus
# Meeus table 36C

def getElongationforPlanet(planet, yyyy, verbose = False):
    # get data for planet
    Dict = getOppoConjforPlanet(planet, True, yyyy, verbose)
    JDE0 = Dict["JDE0"]
    T = Dict["T"]
    M = Dict["M"]
    JDEeast = None; JDEwest = None; elonEast = None; elonWest = None

    if planet == "Venus":
        # Eastern Elongation, correction (days) to mean inferior conjunction
        corr = -70.7600 + 0.0002*T - 0.00001*T*T
        corr += Sin(  M) * (1.0282 - 0.0010*T - 0.00001*T*T)
        corr += Cos(  M) * (0.2761 - 0.0060*T)
        corr += Sin(2*M) * (-0.0438 - 0.0023*T + 0.00002*T*T)
        corr += Cos(2*M) * (0.1660 - 0.0037*T - 0.00004*T*T)
        corr += Sin(3*M) * (0.0036 + 0.0001*T)
        corr += Cos(3*M) * (-0.0011 + 0.00001*T*T)

        JDEeast = JDE0 + corr

        elonEast = 46.3173 + 0.0001*T
        elonEast += Sin(  M) * (0.6916 - 0.0024*T)
        elonEast += Cos(  M) * (0.6676 - 0.0045*T)
        elonEast += Sin(2*M) * (0.0309 - 0.0002*T)
        elonEast += Cos(2*M) * (0.0036 - 0.0001*T)

        # Western Elongation, correction (days) to mean inferior conjunction
        corr =  70.7462            - 0.00001*T*T
        corr += Sin(  M) * (1.1218 - 0.0025*T - 0.00001*T*T)
        corr += Cos(  M) * (0.4538 - 0.0066*T)
        corr += Sin(2*M) * (0.1320 + 0.0020*T - 0.00003*T*T)
        corr += Cos(2*M) * (-0.0702 + 0.0022*T + 0.00004*T*T)
        corr += Sin(3*M) * (0.0062 - 0.0001*T)
        corr += Cos(3*M) * (0.0015 - 0.00001*T*T)

        JDEwest = JDE0 + corr

        elonWest = 46.3245
        elonWest += Sin(  M) * (-0.5366 - 0.0003*T + 0.00001*T*T)
        elonWest += Cos(  M) * (0.3097 + 0.0016*T - 0.00001*T*T)
        elonWest += Sin(2*M) * (-0.0163)
        elonWest += Cos(2*M) * (-0.0075 + 0.0001*T)

    elif planet == "Mercury":
        # Eastern Elongation, correction (days) to mean inferior conjunction
        corr = -21.6101 + 0.0002*T
        corr += Sin(  M) * (-1.9803 - 0.0060*T + 0.00001*T*T)
        corr += Cos(  M) * ( 1.4151 - 0.0072*T - 0.00001*T*T)
        corr += Sin(2*M) * ( 0.5528 - 0.0005*T - 0.00001*T*T)
        corr += Cos(2*M) * ( 0.2905 + 0.0034*T + 0.00001*T*T)
        corr += Sin(3*M) * (-0.1121 - 0.0001*T + 0.00001*T*T)
        corr += Cos(3*M) * (-0.0098 - 0.0015*T)
        corr += Sin(4*M) * ( 0.0192)
        corr += Cos(4*M) * ( 0.0111 + 0.0004*T)
        corr += Sin(5*M) * (-0.0061)
        corr += Cos(5*M) * (-0.0032 - 0.0001*T)

        JDEeast = JDE0 + corr

        elonEast = 22.4697
        elonEast += Sin(  M) * (-4.2666 + 0.0054*T + 0.00002*T*T)
        elonEast += Cos(  M) * (-1.8537 - 0.0137*T)
        elonEast += Sin(2*M) * ( 0.3598 + 0.0008*T - 0.00001*T*T)
        elonEast += Cos(2*M) * (-0.0680 + 0.0026*T)
        elonEast += Sin(3*M) * (-0.0524 - 0.0003*T)
        elonEast += Cos(3*M) * ( 0.0052 - 0.0006*T)
        elonEast += Sin(4*M) * ( 0.0107 + 0.0001*T)
        elonEast += Cos(4*M) * (-0.0013 + 0.0001*T)
        elonEast += Sin(5*M) * (-0.0021)
        elonEast += Cos(5*M) * ( 0.0003)

        # Western Elongation, correction (days) to mean inferior conjunction
        corr =  21.6249 - 0.00002*T
        corr += Sin(  M) * ( 0.1306 + 0.0065*T)
        corr += Cos(  M) * (-2.7661 - 0.0011*T + 0.00001*T*T)
        corr += Sin(2*M) * ( 0.2438 - 0.0024*T - 0.00001*T*T)
        corr += Cos(2*M) * ( 0.5767 + 0.0023*T)
        corr += Sin(3*M) * ( 0.1041)
        corr += Cos(3*M) * (-0.0184 + 0.0007*T)
        corr += Sin(4*M) * (-0.0051 - 0.0001*T)
        corr += Cos(4*M) * ( 0.0048 + 0.0001*T)
        corr += Sin(5*M) * ( 0.0026)
        corr += Cos(5*M) * ( 0.0037)

        JDEwest = JDE0 + corr

        elonWest = 22.4143 - 0.0001*T
        elonWest += Sin(  M) * ( 4.3651 - 0.0048*T - 0.00002*T*T)
        elonWest += Cos(  M) * ( 2.3787 + 0.0121*T - 0.00001*T*T)
        elonWest += Sin(2*M) * ( 0.2674 + 0.0022*T)
        elonWest += Cos(2*M) * (-0.3873 + 0.0008*T + 0.00001*T*T)
        elonWest += Sin(3*M) * (-0.0369 - 0.0001*T)
        elonWest += Cos(3*M) * ( 0.0017 - 0.0001*T)
        elonWest += Sin(4*M) * ( 0.0059)
        elonWest += Cos(4*M) * ( 0.0061 + 0.0001*T)
        elonWest += Sin(5*M) * ( 0.0007)
        elonWest += Cos(5*M) * (-0.0011)

    # return Julian Day and Elongation (degrees) East and West and
    # date and time of the lower conjunction between the elongations
    DTeast = hvfTime(JDEeast)
    DTwest = hvfTime(JDEwest)
    EventEast = {"Object"    : planet,
                "JDE0"      : JDE0,     # used for elongation
                "JDE"       : JDEeast,
                "M"         : M, 
                "T"         : T,
                "DT"        : DTeast,
                "DTrounded" : roundhvfTime(DTeast),
                "Type"      : "Eastern Elongation",
                "elon"      : elonEast,     # used for elongation
            }
    EventWest = {"Object"    : planet,
                "JDE0"      : JDE0,     # used for elongation
                "JDE"       : JDEwest,
                "M"         : M, 
                "T"         : T,
                "DT"        : DTwest,
                "DTrounded" : roundhvfTime(DTwest),
                "Type"      : "Western Elongation",
                "elon"      : elonWest,     # used for elongation
            }

    return [EventEast, EventWest]


#######################################################################

from operator import itemgetter
import os

def Usage(progname):
    print ("Usage: %s Planet [Planet ...] year [year ...]" % progname)
    print ("Usage: %s Planet [Planet ...] year-year ..." % progname)
    print ("Planets are Mercury, Venus, Mars, Jupiter, Saturn, Uranus, Neptune")
    print ()
    print ("Usage: %s all year [year ...]" % progname)
    print ("Usage: %s all year-year ..." % progname)
    print ("       produces output for all seven planets")
    print ("       year must be not less than -4713 (start of Julian Day counting)")
    print ()
    print ("Usage: %s verify" % progname)
    print ("       runs Meeus's benchmark cases")
    print ()
    print ("Planet Events are:")
    print ("       Superior Conjunction, planet passes behind the Sun")
    print ("                the angle between Sun and planet has a minimum")
    print ("       Inferior Conjunction, planet passes between Sun and Earth")
    print ("                this is only possible for Mercury and Venus")
    print ("       Opposition, planet is opposite to Sun")
    print ("                this applies to the outer planets")
    print ("       Eastern Elongation (Venus, Mercury): maximum angle between")
    print ("                Sun and planet, planet visible in the evening")
    print ("       Western Elongation (Venus, Mercury): maximum angle between")
    print ("                Sun and planet, planet visible in the morning")
    sys.exit(0)

def Verify():
    
        print ("Verification for each Planet")
        print ("============================")
        print ("All times are TT")
        print ()
    
        Dict = getOppoConjforPlanet("Mercury", True, 1993.75)
        DT = Dict['DT']
        print ("Meeus p.252 ex.36a: Mercury Inferior Conjunction nearest to 1993-10-01")
        print ("Expecting    1993-11-06 03h")
        print ("Calculated ", nicehvfTimehh(DT))
        print ()
    
        print ("Mercury first observed transit over the solar disk by Gassendi in Paris")
        print ("Expecting    1631-11-07 07h")
        # first observed transit of Mercury over the solar disk by Gassendi in paris
        Dict = getOppoConjforPlanet("Mercury", True, 1631.75, False)
        DT = Dict['DT']
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        
        print ("Venus Inferior Conjunction 1874-12-09 04h")
        Dict = getOppoConjforPlanet("Venus", True, 1874.9, False)
        DT = Dict['DT']
        print ("Expecting    1874-12-09 04h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        
        print ("Venus Inferior Conjunction 1882-12-06 17h")
        print ("Last transit of Venus before that of A.D. 2004")
        Dict = getOppoConjforPlanet("Venus", True, 1882.9, False)
        DT = Dict['DT']
        print ("Expecting    1882-12-06 17h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        
        print ("Venus Inferior Conjunction 2004-06-08 08h")
        print ("Last transit of Venus for Europeans")
        Dict = getOppoConjforPlanet("Venus", True, 2004, False)
        DT = Dict['DT']
        print ("Expecting    2004-06-08 08h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        
        print ("Venus Inferior Conjunction 2012-06-06 02h")
        print ("Last transit of Venus in our lifetime")
        Dict = getOppoConjforPlanet("Venus", True, 2012, False)
        DT = Dict['DT']
        print ("Expecting    2012-06-06 02h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        
        print ("Mars Opposition 2729-09-09 3h")
        print ("A perihelic oppostion of Mars")
        Dict = getOppoConjforPlanet("Mars", True, 2729.75)
        DT = Dict['DT']
        print ("Expecting    2729-09-09 03h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
    
        print ("Jupiter Opposition -0006-09-15 7h")
        # we need to set the approx. year 50 -5.75 to get this opposition correctly
        Dict = getOppoConjforPlanet("Jupiter", True, -5.00)
        DT = Dict['DT']
        print ("Expecting   -0006-09-15 07h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        
        print ("Saturn Opposition -0006-09-14 9h")
        Dict = getOppoConjforPlanet("Saturn", True, -5.00)
        DT = Dict['DT']
        print ("Expecting   -0006-09-14 09h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        
        print ("Saturn Conjunction 2125-08-26 10h")
        Dict = getOppoConjforPlanet("Saturn", False, 2125.6)
        DT = Dict['DT']
        print ("Expecting    2125-08-26 10h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        
        print ("Uranus Opposition 1780-12-17 14h")
        print ("Three months before Uranus' discovery by William Herschel")
        Dict = getOppoConjforPlanet("Uranus", True, 1781)
        DT = Dict['DT']
        print ("Expecting    1780-12-17 14h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        
        print ("Neptune Opposition 1846-08-20 4h")
        print ("One month before Neptune's discovery")
        Dict = getOppoConjforPlanet("Neptune", True, 1846.7)
        DT = Dict['DT']
        print ("Expecting    1846-08-20 04h")
        print ("Calculated ", nicehvfTimehh(DT))
        print()
        

if __name__ == "__main__":


    PlanetList = ["Mercury", "Venus", "Mars", "Jupiter", "Saturn", 
                    "Uranus", "Neptune"]
    isInferior = {"Mercury": True, "Venus": True, "Mars": False,
                    "Jupiter": False, "Saturn": False, "Uranus": False,
                    "Neptune": False}
    PeriApoList = ["Mercury", "Venus", "Earth", "Mars", "Jupiter", "Saturn", 
                    "Uranus", "Neptune"]
    Planet = None

    print ("%s Version %s" % (sys.argv[0], VersionPlanets))
    print ()

    # read years
    foundOne = False

    usePlanetList = []
    useYearList = []
    useYearSet = set()

    SynopticEventList = []
    SynopticEventSet = set()
    usedPlanetSet = set()
    WantedYearsSet = set()

    for ia in range(1, len(sys.argv)):
        arg = sys.argv[ia]
        if arg in PlanetList:
            Planet = arg
            usePlanetList.append(Planet)
        elif arg == 'all':
            usePlanetList = PlanetList
        elif arg == 'verify':
            Verify()
            sys.exit(0)
        else:
            yearList = re.split(r'-', arg)
            # we have to consider negative years as well!
            if len(yearList) == 2:  # startYear-lastYear, both positive 
                try:
                    if len(yearList[0]) > 0:
                        startYear = int(yearList[0])
                        lastYear  = int(yearList[1])
                    else:           # -startyear, negative, e.g. -1234
                        startYear = -int(yearList[1])
                        lastYear  = startYear
                except:
                    Usage(sys.argv[0])
            elif len(yearList) == 3:  # -0003-0003 from -3 to +3
                try:
                    startYear = -int(yearList[1])
                    lastYear  = int(yearList[2])
                except:
                    Usage(sys.argv[0])
            elif len(yearList) == 4:  # -0003--0002 from -3 to -2
                try:
                    startYear = -int(yearList[1])
                    lastYear  = -int(yearList[3])
                except:
                    Usage(sys.argv[0])
            else:
                try:
                    startYear = int(yearList[0])
                    lastYear  = startYear
                except:
                    Usage(sys.argv[0])

            for year in range(startYear, lastYear+1):
                WantedYearsSet.add(year)
                # we need to cover a larger span than just the year in question
                # otherwie we may miss some events
                # make the grid sufficiently fine since we have many 
                # mercury events each year
                # here we use 3 months steps
                years = set([year+a/4. for a in range(-3, 8)])
                useYearSet |= years
                #print ("year, useYearSet = ", year, sorted(useYearSet))
    
    # for planets and year
    for p in usePlanetList:
        if p in usedPlanetSet: break
        else: usedPlanetSet.add(p)
        usedTimeSet = set()
        for y in useYearSet:
            foundOne = True
            Dictsup = getOppoConjforPlanet(p, False, y)
            if Dictsup['DT'] not in usedTimeSet:
                SynopticEventList.append(Dictsup)
                usedTimeSet.add(Dictsup['DT'])
            Dictinf = getOppoConjforPlanet(p, True, y)
            if Dictinf['DT'] not in usedTimeSet:
                SynopticEventList.append(Dictinf)
                usedTimeSet.add(Dictinf['DT'])
            if isInferior[p]:
                DictList = getElongationforPlanet(p, y)
                if DictList[0]['DT'] not in usedTimeSet:
                    SynopticEventList.append(DictList[0])
                    usedTimeSet.add(DictList[0]['DT'])
                if DictList[1]['DT'] not in usedTimeSet:
                    SynopticEventList.append(DictList[1])
                    usedTimeSet.add(DictList[1]['DT'])
            PeriDictList = getPerihelionAphelion(p, y)
            #print (PeriDictList)
            #SynopticEventList = chain(SynopticEventList, PeriDictList)
            for dict in PeriDictList:
                SynopticEventList.append(dict)
    '''
                print ("Planet: ", dict["Object"])
                print ("JDE   : ", dict["JDE"])
                print ("DT    : ", dict["DT"])
                print ("Type  : ", dict["Type"])
                print ()
    '''            

    if not foundOne: Usage(sys.argv[0])

    print ("  Planet Events")
    print ("  =============")

    #print (SynopticEventList)
 
    SortedSynopticEventList = sorted(SynopticEventList, key=itemgetter('DT'))
    print (
"  Date       Time Planet   Event                     Elongation")
    for i in range(len(SortedSynopticEventList)):
        Event = SortedSynopticEventList[i]
        DT = Event['DT']
        (yy, mm, dd) = DT.getDate()
        if yy not in WantedYearsSet: continue
        # check if previous event is in another year, then add a blank line
        if i > 0:
            Event1 = SortedSynopticEventList[i-1]
            DT1 = Event1['DT']
            (yy1, mm1, dd1) = DT1.getDate()
            if yy != yy1: print ()   

        #if Event['elon']:
        if 'elon' in Event and Event['elon']:
            print (" %14s  %-7s  %-20s   %8.2f" % (
            nicehvfTimehh(Event['DT']),
            Event['Object'], Event['Type'], 
            Event['elon']))
        elif Event['Type'] == 'Perihel' or Event['Type'] == 'Aphel':
            print (" %14s  %-7s  %-20s  R = %.8f au" % (
            nicehvfTimehh(Event['DT']),
            Event['Object'], Event['Type'], Event['Rau'])) 

        else:
            print (" %14s  %-7s  %-20s" % (
            nicehvfTimehh(Event['DT']),
            Event['Object'], Event['Type'])) 
    print()
    #foundOne = True

    if not foundOne: Usage(sys.argv[0])
    sys.exit()
