#! /usr/bin/env python3
# Julian to Greenwich and vice versa
# with my own hvfTime class, works also for BC dates and leap years
# hvf 16.09.2021 26.12.2021 11.11.2022 13.04.2023
# make hvfTime() hashable -- hvf 02.10.2023

from datetime import datetime, timedelta
from math import floor
import sys
import re
import dateutil.parser

VersionJulian = "V1.8 hvf 2023-10-02 HVFTime.py based on Meeus"


# JulianDay from year, month, day -  hvf 16.11.2021
def getJulianDayfromYMD(year, month, day, hour=0, minute=0, second=0, 
                        verbose = False):
    if verbose: 
        print ("getJulianDayfromYMD: ", year, month, day, hour, minute, second)
    day  += hour/24. + \
            minute/1440. + \
            second/86400.
    isGregor = False    # 04.10.1582 is last day before Gregorian calendar
    if year > 1582 or year == 1582 and (month > 10 or month == 10 and day > 4):
        isGregor = True
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    # JD 0 is -4712-01-01.5 i.e. -4712-01-01 12h
    # N.B. JD 1524.5 is -4708-03-05 00h
    if verbose: print ("JD = ", JD)
    return JD


def roundhvfTime(hvfDT = None, roundTo = 3600):
    # Round a hvfTime object to anytime lapse in seconds
    # hvfDT: hvftime object, default now
    # roundTo: closest number of seconds to round to, default 1 hour.
    # Author: Harald von Fellenberg 2021
    # Problem: 2021-11-21 1223 returns 2021-11-21 12:22:60.000
    # hvf 2022-09-09
    verbose=False
    if not isinstance(hvfDT, hvfTime): hvfDT = hvfTime.now()
    h, m, s = hvfDT.getTime()
    seconds = h*3600 + m*60 + s
    fraction = seconds - floor(seconds)     # fractional seconds
    if verbose:
        print ("seconds, fraction, floor, rounding = ", seconds, fraction, floor(seconds))
    seconds = floor(seconds)
    rounding = (seconds+roundTo/2) // roundTo * roundTo
    if verbose: print (rounding)
    return hvfDT + timedelta(0, rounding-seconds, -fraction*1e6)

# returns yyyymmdd hh for a hvfTime object, rounded
def nicehvfTimehh (hvfDT):
    hvfDTrounded = roundhvfTime(hvfDT)
    yy, mm, dd = hvfDTrounded.getDate()
    hh, dummy, dummy = hvfDTrounded.getTime()

    s = ''
    if yy >= 0:
        s = " %04d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    else:
        s = "%05d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    return s

#######################################################################
# replaces astropy.time.Time since that does not support leap days 
# before 100BC
# values are represented in Julian days
# creator parameter can be:
# - a hvfTime object
# - a JD JulianDay object (float)
# - a datetime object
# - date and time as six values yy, mm, dd, [hh, MM, ss]
#   dd and ss may have fractions
# hvf 26.12.2021

class hvfTime:
    def __init__(self, *arg):
        # print ("__init__: arg = ", arg)
        # argument is a hvfTime object
        if isinstance(arg[0], hvfTime):
            JD = arg[0].getJD()
            self.__JD = JD
        # argument is a JD
        elif isinstance(arg[0], float) and len(arg) == 1:
            JD = arg[0]
            self.__JD = JD
        # argument is a datetime object
        elif isinstance(arg[0], datetime):
            dt = arg[0]
            yy = dt.year
            mm = dt.month
            dd = dt.day
            hh = dt.hour
            MM = dt.minute
            ss = dt.second + dt.microsecond*1e-6
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        # argument is yy mm dd [hh MM ss]
        elif (isinstance(arg[0], int) or isinstance(arg[0], float)) and \
            len(arg) >= 3:
            hh = 0; MM = 0; ss = 0
            yy = arg[0]; mm = arg[1]; dd = arg[2]
            if len(arg) > 3: hh = arg[3]
            if len(arg) > 4: MM = arg[4]
            if len(arg) > 5: ss = arg[5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD

        # argument is (yy mm dd [hh MM ss])
        elif isinstance(arg[0], tuple) or isinstance(arg[0], list):
            hh = 0; MM = 0; ss = 0
            yy = arg[0][0]; mm = arg[0][1]; dd = arg[0][2]
            if len(arg[0]) > 3: hh = arg[0][3]
            if len(arg[0]) > 4: MM = arg[0][4]
            if len(arg[0]) > 5: ss = arg[0][5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        else:
            print ("hvfTime instance not supported: ", arg[0])
            sys.exit(1)
            
    def getJD(self):
        return self.__JD

    def getDate(self, verbose = False):  # returns (yy, mm, dd), 
                                        # where dd may have fraction
        # works also for negative JDs! hvf 29.12.2021
        JD = self.__JD
        jd1 = JD + 0.5
        I = floor(jd1)      # integer part, may be negative
        F = jd1 - I         # fractional part, always positive
        if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
            A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
            B = I + A - int(A/4) + 1
        else:   B = I       # Julian Calendar for earlier dates
        C = B + 1524
        D = floor((C-122.1)/365.25)
        E = floor(365.25 * D)
        G = floor((C-E)/30.6001)
        d = C - E + F - floor(30.6001 * G)
        if G < 13.5:    m = G - 1
        else:           m = G - 13
        if m > 2.5:     y = D - 4716
        else:           y = D - 4715

        if verbose:
            print ("getDate: JD, jd1, I, F = ", JD, jd1, I, F)
            print ("getDate: B, C, D, E, G = ", B, C, D, E, G)
            print ("getDate: y, m, d       = ", y, m, d)
        return (y, m, d)

    def getTime(self):      # seconds have 3 digi fractions (millisec)
        dummy, dummy, t = self.getDate()
        t -= floor(t)
        # t is fraction of day - convert to sec and round
        daysec = round(t * 86400., 3)
        hms = daysec / 3600.
        #hms = t*24.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 3)
        #s = int((ms - m) * 60)
        #print ("daysec, hms, h, ms, m, s = ", daysec, hms, h, ms, m, s)
    
        return (h, m, s)

    def __str__(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, hh, mm, ss)

    def getStr(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        daysec = round(ss + mm*60 + hh*3600, 0)
        hms = daysec / 3600.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 0)

        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, h, m, s)

    # Ernesti, Python3, p. 380
    # hvfTime = hvfTime + timedelta
    # hvfTime = hvfTime + day   (float or int)
    def __add__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__add__: wrong type ", type(td))
            sys.exit(1)
        return hvfTime(self.__JD + sec/86400.)

    # hvfTime += timedelta
    def __iadd__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__iadd__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD += sec/86400.
        return self

    # hvfTime = hvfTime - timedelta
    # and   n = hvfTime - hvfTime       n in days
    def __sub__(self, td):
        # subtract a timedelta object 
        if isinstance (td, timedelta):
            sec = td.total_seconds()
            return hvfTime(self.__JD - sec/86400.)
        elif isinstance(td, hvfTime):
            # return JDs in days as difference between two hvfTime objects
            return self.__JD - td.getJD()
        else:
            print ("hvfTime.__sub__: wrong type ", type(td), td)
            sys.exit(1)

    # hvfTime -= timedelta
    def __isub__(self, td):
        # subtract a timedelta object 
        try:    
            sec = td.total_seconds()
        except:
            print ("hvfTime.__isub__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD -= sec/86400.
        return self

    # hvfTime < hvfTime
    def __lt__(self, other):
        return self.__JD < other.getJD()

    # hvfTime <= hvfTime
    def __le__(self, other):
        return self.__JD <= other.getJD()

    # hvfTime == hvfTime
    def __eq__(self, other):
        return self.__JD == other.getJD()

    # hash - makes hvfTime hashable, i.e. as key in dict or set 
    def __hash__(self):
        return hash(self.__JD)

    # hvfTime != hvfTime
    def __ne__(self, other):
        return self.__JD != other.getJD()

    # hvfTime > hvfTime
    def __gt__(self, other):
        return self.__JD > other.getJD()

    # hvfTime >= hvfTime
    def __ge__(self, other):
        return self.__JD >= other.getJD()

    def now():  # returns current time (h, m, s), uses datetime.now()
                # roundedd to integer seconds
        return hvfTime(datetime.now()).getTime()

    def today():    # returns current date (y, m, d), uses datetime.today()
                    # d has fractions
        return hvfTime(datetime.today()).getDate()
    
    def getDatetime(self):
        (yy, mm, dd) = self.getDate()
        dd = floor(dd)
        (HH, MM, SS) = self.getTime()
        SS = int(round(SS, 0))
        return datetime(yy, mm, dd, HH, MM, SS)
#######################################################################

# parse date time
# returns a hvfTime object
# hvf 19.11.2021 25.12.2021
# allowed patterns:
# 2021-11-21
# 2021-11-21 12:23:45.635
# 2021-11-21T12:23:45.635
# 2021-11-21 122345.635
# 2021-11-21 122345
# 2021-11-21 1223
# 2021-11-21 12
# -4712-1-1.5 
# 25.12.2021 12:34:56.789
# 1.1.2022
# 2451545.0 (a Julian Day)
# works for dates between -9999-01-01 and 9999-12-31

def parseDateTime(s, verbose = False):
    HH = MM = SS = 0
    # check if two parts separated by blank or T
    if verbose: print ("parseDateTime input: ", s)
    s = s.strip()
    dpart = s
    tpart = None
    parts = re.split (r"[\sT]", s)
    if len(parts) >= 2:
        dpart = parts[0]    # date part
        tpart = parts[1]    # time part
        if verbose:
            print ("date part = ", dpart)
            print ("time part = ", tpart)

    # parse date  2021-12-24 or  4712-1-1.5
    if re.fullmatch(r'\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        ymd = re.split(r'-', dpart)
        yy = ymd[0]
        mm = ymd[1]
        dd = ymd[2]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date 24.12.2021 or 1.1.2022
    elif re.fullmatch(r'\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        ymd = re.split(r'\.', dpart)
        yy = ymd[2]
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -4712-01-01.5
    elif re.fullmatch(r'-\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        symd = re.split(r'-', dpart)
        yy = -int(symd[1])
        mm = symd[2]
        dd = symd[3]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -24.12.2021 or -1.1.2022
    elif re.fullmatch(r'-\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        symd = re.split(r'-', dpart)
        ymd = re.split(r'\.', symd[1])
        yy = -int(ymd[2])
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    else:
        print ("parseDateTime format unknown: ", s)
    # parse date  2021-12-24 or  4712-1-1.5
    # parse date 24.12.2021 or 1.1.2022
    # parse date -4712-01-01.5
    # parse date -24.12.2021 or -1.1.2022
        print ("Allowed date formats include 2021-12-24, 4712-1-1.5,")
        print ("    24.12.2021, 1.1.2022, -4712-01-01.5, -24.12.2021, -1.1.2022")
        print ("Allowed range of years is from -9999 to 9999")
        sys.exit(1)

    # parse time using dateutil.parser.isoparse - this is easier
    # than checking all the different possible input formats
    if tpart:
        deltaT = dateutil.parser.isoparse("2021-12-24 " + tpart) \
                - datetime(2021, 12, 24)
        Secs = round(deltaT.total_seconds(), 3)    # avoid rounding error
        HH = Secs // 3600
        Secs -= HH * 3600
        MM = Secs // 60
        Secs -= MM * 60
        SS = Secs
        if verbose:
            print ("HH, MM, SS = ", HH, MM, SS)

    #return hvfTime object
    return hvfTime(int(yy), int(mm), float(dd), int(HH), int(MM), float(SS))

#######################################################################
# catculate weekday for any date from -9999-01-01 to 9999-12-31
# Meeus 1991 p. 65

def getDayofWeek(dt):
    name =["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday",
            "Friday", "Saturday"]
    hvfdt = hvfTime(dt)
    (yy, mm, dd) = hvfdt.getDate()
    hvfdtMidnight = hvfTime(yy, mm, int(dd), 0, 0, 0.0)
    JD0 = hvfdtMidnight.getJD()
    weekday = int(JD0 + 1.51) % 7
    return name[weekday]

#######################################################################

#!/usr/bin/env python3
# -*- coding: iso-8859-1 -*-

# Der folgende Python-Code fuer Sonne und Mond 
# basiert auf der Portierung des java-Scripts von Arnold Barmettler
# http://lexikon.astronomie.info/java/sunmoon/sunmoon.html
#
#     Entfernen Sie folgende Informationen auf keinen Fall: / 
#     Do not remove following text:
#     Source code by Arnold Barmettler, www.astronomie.info / www.CalSky.com
#     and Harald von Fellenberg, www.hvf.ch
#     based on algorithms by Peter Duffett-Smith's great and easy book
#     'Practical Astronomy with your Calculator or Spreadsheet'.
#     and Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Totale Ueberarbeitung durch Harald von Fellenberg, Dezember 2020
#   Portierung auf Python 3
#   Ergaenzung fuer Mitternachtssonne, Polarnacht
#   Ergaenzungen fuer Daemmerung, Hoehe der Sonne beim Transit tags und nachts
#   Ergaenzung mit Zeitgleichung, wahre und mittlere Ortszeit
#   Azimut für Sonnenaufgang, Sonnenuntergang
#   Liste einiger Städte mit ihren Koordinaten
#   Korrekte Berechnung in allen Zeitzonen
#   Behandlung aller Planeten
#
#   Die Behandlung von Sonne und Mond wurde ersetzt durch die Algorithmen in
#   Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Die Daten aller Planeten beruhen auf VSPO87
#   Pierre Bretagnon, Gerard Francou
#   Bureau des Longitudes
#   77, Avenue Denfert-Rochereau
#   F75014, Paris, France
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/
#
#   Die Resultate wurden mit denen auf https://www.timeanddate.com/ verglichen,
#   die Uebereinstimmung ist gut innerhalb von wenigen Minuten.
#   Beim Mond ist stets das Datum zu beachten!
#   Liegt die lokale Zeit zwischen Mitternacht und "zone", so wird für den Mond
#   zum Teil auf den Vortag zurückgegriffen. Dies ist besonders für Orte mit
#   grosser Zeitzonendifferenz wichtig, wie Sydney oder Honolulu.
#   Für Orte in den Polarregionen wird Mitternachtssonne und Polarnacht
#   korrekt ausgegeben, ausser am Tag, wo die Sonne unter- oder aufgeht.
#   Ebenso fuer den Mond.
#
#   Anmerkung: die Webseiten astronomie.info sowie calsky.com existieren
#   seit Oktober 2020 nicht mehr.
#
#   Alle 8 Planeten sind ebenfalls behandelt, gemaess den Algorithmen in
#   Jean Meeus, Astronomical Algorithms.
#   Die Daten zu den 8 Planeten stammen aus den VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Version 1.0
#   Version 2.0
#   Version 3.0
#   Version 3.1
#   Version 3.4
#   Version 4.0 inkl Sun-Events, Moon-Events
#   Version 5.1 
Version = "V5.3 hvf 2023-09-27 SunHead.py using VSOP87 ephemerides and HVFTime"
if 'VersionList' not in globals().keys(): VersionList = []
VersionList.append(Version)

import time
from datetime import datetime, date, timedelta
from math import sin, acos, cos, pi, radians, degrees, atan, atan2
from math import asin, tan, ceil, floor, sqrt, fabs
import argparse
import sys
import dateutil.parser
import os

# alle Konstanten sind auf die Epoche 2010.0, JD=2455196.5 bezogen
# falls nicht anders dokumentiert.

def getEpoch2010():
    return 2455196.5

# https://www.geeksforgeeks.org/how-to-print-to-stderr-and-stdout-in-python/
def print_to_stderr(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stderr) 

def print_to_stdout(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stdout) 

# produces a datetime object output string without microseconds
# seconds get rounded

def nicedatetime(dt):       # yyyy-mm-dd HH:MM:SS
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        string = dt.getStr()
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%Y-%m-%d %H:%M:%S")
    return string

def niceshortdatetime(dt):  # dd.mm.yy HH:MM:SS
    # only meaningful for known century
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        (yy, mm, dd) = dt.getDate()
        yy = int(abs(yy)%100)    # only last 2 ciphers, also for negative yyyy
        (hh, MM, ss) = dt.getTime()
        # round seconds! use %f fmt - hvf 13.04.2023
        #string = "%02d.%02d.%02d %02d:%02d:%02d" % (dd, mm, yy, hh, MM, ss)
        string = "%02d.%02d.%02d %02d:%02d:%02.0f" % (dd, mm, yy, hh, MM, ss)
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%d.%m.%y %H:%M:%S")
    return string

def nicedateshorttime(dt):  # yyyy-mm-dd HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=30)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        (h, m, s) = dt.getTime()
        string = "%4d-%02d-%02d %02d:%02d" % (y, mm, dd, h, m)
    else:
        string = dt.strftime("%Y-%m-%d %H:%M")
    return string

def niceshortdateshorttime(dt):  # dd.mm.yy HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        dtrounded = roundhvfTime(dt, 60)
        #print ("dt, dtrounded = ", dt, dtrounded)
        (y, mm, dd) = dtrounded.getDate()
        yy = abs(y)%100
        dd = floor(dd)
        (h, m, s) = dtrounded.getTime()
        string = "%02d.%02d.%02d %02d:%02d" % (dd, mm, yy, h, m)
    else:
        dt += timedelta(seconds=30)
        string = dt.strftime("%d.%m.%y %H:%M")
    return string

def nicedate(dt):           # yyyy-mm-dd
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=0.5)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        string = "%04d-%02d-%02d" % (y, mm, dd)
    else:
        string = dt.strftime("%Y-%m-%d")
    return string

def nicetime(dt):           # HH:MM:SS
    if isinstance(dt, type(None)): return None
    try:
        if isinstance(dt, hvfTime):
            (h, m, s) = dt.getTime()
            s =round(s, 0)
            string = "%02d:%02d:%02d" % (h, m, s)
        else:
            dt += timedelta(seconds=0.5)
            string = dt.strftime("%H:%M:%S")
    except:
        string = ' '*8
    return string

def nicetimedelta(dt):      # +-HH:MM:SS
    if isinstance(dt, type(None)): return None
    #dt += timedelta(seconds=0.5)
    # treat also -00:00:05 correctly! hvf 16.06.2021
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60

    string = "%1s%2d:%02d:%02d" % (sign, hours, minutes, seconds)
    return string

def nicedeltadayhhmm(dt):   # +-days HH.MM
    if isinstance(dt, type(None)): return None
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    days  = int(totalsecs/86400.); totalsecs -= days*86400.
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60:
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60
    if hours >= 24:
        days += 1; hours -= 24

    string = "%1s%d days %02d:%02d" % (sign, days, hours, minutes)
    return string

def niceminsec(secs):       # +- M:SS  for Time Equation - Zeitgleichung
    if isinstance(secs, type(None)): return None
    sign = ' '
    if secs < 0: sign = '-'; secs = abs(secs)
    minutes = int(secs/60.); secs -= minutes*60.
    seconds = round(secs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60

    string = "%1s%d:%02d" % (sign, minutes, seconds)
    return string

# printZeile(kopf, Liste, format, zeile=79)
# schreibt eine Zeile der max. Laenge zeile mit kopf und den
# Werten aus Liste, formattiert je mit format.
# Padding vor jedem Listenelement
# Die  formattierten Werte muessen gleich lang sein.
# hvf 31.12.2020

def printZeile(kopf, Liste, format, zeile=79):
    rest = zeile
    string = kopf
    rest -= len(string)
    n = len(Liste)      # Anzahl Werte
    test = format % Liste[0]
    testlen = len(test)
    padlen = int((rest - n*testlen)/n)
    padding = ' '*padlen

    for item in Liste:
        string += padding
        string += format % item

    print (string)
    return string

# modulus in Python has always the same sign as b
# Mod(-3, 24) gives 21
def Mod(a, b):
    return (a % b)

# return integer value, closer to 0
def Int(x): return int(x)

# return fractional part, closer to 0 (i.e. may be nagative)
def Frac(x): return (x - int(x))

# Winkelfunktionen fuer Grad Eingaben
def Sin(x):
    return sin(radians(Mod(x, 360.)))

def Cos(x):
    return cos(radians(Mod(x, 360.)))

def Tan(x):
    return tan(radians(Mod(x, 360.)))

def Deg(x):
    return degrees(x)

# calculate angular distance between two objects with (RA1, Decl1), (RA2, Decl2)
# all angles in degrees
# see Meeus, chapter 16, p. 105, Angular Separation
def AngularDistance(star1, star2, verbose=False):
    RA1 = star1[0]; Decl1 = star1[1]
    RA2 = star2[0]; Decl2 = star2[1]
    cosD = Sin(Decl1)*Sin(Decl2) + Cos(Decl1)*Cos(Decl2)*Cos(RA1-RA2)
    D = degrees(acos(cosD))
    # check if angular separation is small
    if D < 0.2:     # below 0.2 degrees, 12'
        Dmean = (Decl1+Decl2)/2.
        Ddelta = Decl1 - Decl2
        D = sqrt(((RA1-RA2)*Cos(Dmean))**2 + Ddelta**2)
    return D


#####################################################################
# Quelle: https://de.wikipedia.org/wiki/Julianisches_Datum
# abgerufen am 20.10.2019
# Name geaendert in getJulianDay (siehe Meeus, chapter 7) - hvf 24.03.2023

# Calculate Julian day for all Gregorian and Julian dates
# Greenwich date is supplied as a datetime object

def getJulianDay(dtGreenwichDate, isGregor=True ):
    '''
    # day may have fractions for h, m, s, us
    # Jan and Feb are months 13 and 14
    year  = dtGreenwichDate.year
    month = dtGreenwichDate.month
    day   = dtGreenwichDate.day
    day  += dtGreenwichDate.hour/24. + dtGreenwichDate.minute/1440. + \
            dtGreenwichDate.second/86400. + \
            dtGreenwichDate.microsecond/86400e6
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    return JD
    '''
    hvfdt = hvfTime(dtGreenwichDate)
    return hvfdt.getJD()

################################################################
# calculate Greenwich date from Julian date
# Ref. Practical Astronomy (see above), p.11
# returns a datetime object for midnight 

def getGreenwichDate(JD):
    '''
    jd1 = JD + 0.5
    I = int(jd1)        # integer part
    F = jd1 - I         # fractional part
    if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
        A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
        B = I + A - int(A/4) + 1
    else:   B = I       # Julian Calendar for earlier dates
    C = B + 1524
    D = int((C-122.1)/365.25)
    E = int(365.25 * D)
    G = int((C-E)/30.6001)
    d = C - E + F - int(30.6001 * G)
    if G < 13.5:    m = G - 1
    else:           m = G - 13
    if m > 2.5:     y = D - 4716
    else:           y = D - 4715

    # add fractional day - must use timedelta for this
    dtGreenwichDate = datetime(y, m, 1) + timedelta(days=d-1)
    return dtGreenwichDate
    '''

    return hvfTime(float(JD))

#   HMSDH convert hours, minutes, seconds to decimal hours
def HMSDH(hh, mm, ss):  # hh may be negative
    habs = abs(hh)
    habs += mm/60. + ss/3600.
    h = habs
    if hh < 0: h = -habs
    return h

#   DHHMS converts decimal hours into hours, minutes, seconds
#   seconds may have fractions, no rounding
def DHHMS(h):   # h may be negative, then only hours will be negative
    sign = 1
    if h < 0: sign = -1
    habs = abs(h)
    hh = int(habs)
    mm = int(Frac(habs)* 60.)
    ss = Frac(habs*60)*60.
    hh = hh * sign
    return (hh, mm, ss)

#   returns HH:MM and a rounded value of h
#   h must be positive or None or ''
def HHMM(hh):
    if (hh == None or hh == ''):
        return('--')
    m = Frac(hh) * 60.0
    h = Int(hh)
    if (m >= 59.5):
        h += 1
        m -= 60.0
    m = int(round(m))
    string = "%02d:%02d = " % (h, m) + str(round(hh, 4))
    return (string)

#   returns HH:MM:SS and a rounded value of h
#   h must be positive or None or ''
def HHMMSS(hh):
    if (hh == 0):
        return('--')
    m = Frac(hh) * 60
    h = Int(hh)
    s = Frac(m) * 60.0
    m = Int(m)
    if (s >= 59.5):
        m += 1
        s -= 60.0
    if (m >= 60):
        h += 1
        m -= 60
    s = int(round(s))
    string = "%02d:%02d:%02d = " % (h, m, s) + str(round(hh, 5))
    return (string)

def hmsstring(hour):
    # hour may also be negative, e.g. if used to convert declination
    (hh, mm, ss) = DHHMS(hour)
    str = "%d:%02d:%02d" % (hh, mm, ss)
    return str


#   local civic time to Universal Time
#   returns a datetime object 

def getUTfromCivicTime(dtLocalCivicTime, zone):    
    # dtLocalCivicTime is a datetime object
    # zone is time zone offset in hours, positive going East
    dtUT = dtLocalCivicTime - timedelta(hours=zone)
    return dtUT

def getCivicTimefromUT(dtUT, zone):
    dtLocalCivicTime = dtUT + timedelta(hours=zone)
    return dtLocalCivicTime

#   UTC to mean local time 
#   dtUT: Universal Time datetime object
#   longitude: Eastern longitude, degrees
#   returns a datetime object, rounded to seconds

def getMeanLocalTime(dtUT, longitude):
    dtmeanLT = dtUT + timedelta(seconds=round(240.*longitude, 0))
    return dtmeanLT

def getUTfromMeanLocalTime(dtmeanLT, longitude):
    dtUT =  dtmeanLT - timedelta(seconds=round(240.*longitude, 0))
    return dtUT

#   datetime.datetime or datetime.time to decimal hours

def getDecimalHour(dttime):
    '''
    hour = dttime.hour + dttime.minute/60. + dttime.second/3600. \
            + dttime.microsecond/3600.E6
    '''
    (hour, min, sec) = hvfTime(dttime).getTime()
    return hour    

# calculate horizontal coordinates (Azimut, Altitude)
# from ecliptic coordinates (Hour angle, Declination)
#   Parameters:
#   Hour    Hour angle counted from meridian to West, degrees
#   Decl    Declination, degrees
#   latitude    degrees
#   Returns:
#   azimut  counted from North, degrees
#   alt     altitude above horizon, degrees
#   x, y, z Cartesian coords in horizontal system, x pointing South, y West
#   https://de.wikipedia.org/wiki/Astronomische_Koordinatensysteme#Umrechnungen

def getAziAltfromHourDecl(Hour, Decl, latitude):
    # Hour angle and Declination into azimut and altitude
    azimut = atan2(Sin(Hour), Cos(Hour)*Sin(latitude)-Tan(Decl)*Cos(latitude))
    alt    = asin(Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour))
    azimut = Mod(degrees(azimut)+180., 360.)
    alt    = degrees(alt)
    # Cartesian coords in horizon system
    x = Cos(alt)*Cos(azimut-180.)
    y = Cos(alt)*Sin(azimut-180.)
    z = Sin(alt)
    # dito calculated from  Hour, Decl, latitude
    xx = -Cos(latitude)*Sin(Decl)+Sin(latitude)*Cos(Decl)*Cos(Hour)
    yy = Cos(Decl)*Sin(Hour)
    zz = Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour)
    xyzDict = {
        "x":    x,
        "y":    y,
        "z":    z, 
        "xx":   xx,
        "yy":   yy,
        "zz":   zz, }
    return (azimut, alt, xyzDict)

# Zodiac signs for Sun and Moon
#   lon ecliptic longitude, degrees
def getSign(lon):
    Signs_de = ("Widder", "Stier", "Zwillinge", "Krebs", "Löwe", "Jungfrau", \
        "Waage", "Skorpion", "Schütze", "Steinbock", "Wassermann", "Fische")
    Signs_en = ("Aries", "Taurus", "Gemini", "Cancer", "Leo", "Virgo", \
        "Libra", "Scorpio", "Sagittarius", "Capricorn", "Aquarius", "Pisces")
    return (Signs_en[int(floor(lon / 30))])

########################################################################
#
# getNutation(dtCivicTime, zone)
# Nutation in ecliptic longitude and ecliptic latitude
# see Meeus, Astronomical Algorithms, chapter 21

def getNutation(dtCivicTime, zone, verbose=False):
    # Epoch is J2000.0
    # JD 2451545
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    T = (JD - 2451545.0)/36525. # Centuries after the Epoch, 21.1

    # Moon Ascending Node
    Omega = 125.04452 - 1934.136261 * T + 0.0020708 * T*T + T*T*T/450000.
    Omega = Mod(Omega, 360.)
    O = radians(Omega)

    # accuracy of 0.5 arcsec in longitude, 0.1 arcsec in latitude
    # Sun Mean Longitude
    LmeanSun = 280.4665 + 36000.7698 * T
    LmeanSun = Mod(LmeanSun, 360.)
    LS = radians(LmeanSun)

    # Moon Mean Longitude
    LmeanMoon = 218.3165 + 481267.8813 * T
    LmeanMoon = Mod(LmeanMoon, 360.)
    LM = radians(LmeanMoon)

    # Effect on ecliptic longitude and latitude, arcsecs
    deltaELong = -17.20*sin(O) -1.32*sin(2.*LS) -0.23*sin(2*LM) +0.21*sin(2*O)
    deltaELat  = 9.20*cos(O) + 0.57*cos(2.*LS) + 0.10*cos(2*LM) -0.09*cos(2*O)

    # obliquity of the ecliptic, 21.2
    eps0 = 23+26/60.+21.448/3600. - 46.8150/3600.*T - 0.00059/3600.*T*T \
            +0.001813/3600*T*T*T
    eps = eps0 + deltaELat/3600.

    if verbose:
        print ("getNutation")
        print ("JD = ", JD)
        print ("T = ", T)
        print ("LmeanSun = ", LmeanSun)
        print ("LmeanMoon = ", LmeanMoon)
        print ("getNutation deltaELong (arcsec) = %.2f" % deltaELong)
        print ("getNutation deltaELat  (arcsec) = %.2f" % deltaELat )
        print ("eps0 (degrees) = %.6f" % eps0, ", ", HHMMSS(eps0))
        print ("eps  (degrees) = %.6f" % eps, ", ", HHMMSS(eps) )

    # return values in degrees
    return (deltaELong/3600., deltaELat/3600., eps)

#################################################################
#   Funktionen für die Berechnung der Zeitgleichung

def getSunStufffromTT(TT, verbose = False):
    # TT Tage nach Epoch2010 inkl. DeltaT Korrektur
    # returns RA, epsilon in degrees

    def atanLage(x, Lage):    # arctangens mit Lageparameter, alle Werte in Grad
        # siehe https://en.wikipedia.org/wiki/Equation_of_time
        # Resultat in Grad
        xrad = radians(x)
        Lagerad = radians(Lage)
        y = atan(xrad)
        valuerad = y + pi * round((Lagerad-y)/pi, 0)
        value = degrees(valuerad)
        return value
    
    def getE(t):    # numerische Exzentrizität zur Zeit t
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        e = 0.01670491 - 4.208E-5*T - 1.26E-7*T*T
        return e
    
    def getEps(t):  # Schiefe der Ekliptik in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
        return eps
    
    def getOmegabar(t):     # Länge des Perihels in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        omegabar = 283.11243767 + 1.720171*T + 0.000452778*T*T
        return omegabar
    
    def getL(t):    # mittlere Länge der Sonne in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        L = 279.557285 + 36000.769585*T + 0.0003025*T*T
        return L
    
    def getC(e, M):    # Mittelpunktsgleichung equation of the center
        # berechnet wahre minus mittlere Anomalie als Fkt der Exzentrizität
        # und der mittleren Anomalie in dritter Näherung in e
        # e numerische Exzentrizität
        # M mittlere Anomalie der Sonne in Grad
        # Resultat in Grad
        # Danby, Fundamentals of Celestial Mechanics, p.438
        Mrad = radians(M)
        Crad = (2*e-e*e*e/4)*sin(Mrad) + (1.25*e*e -11./24.*e**4)*sin(2*Mrad)\
             + 13./12.*e*e*e*sin(3*Mrad) + 103./96.*e**4*sin(4*Mrad)
        C = degrees(Crad)
        return C

    t = TT
    e = getE(t)                 # Exzentrizität der Erdbahn, Zeit t
    eps = getEps(t)             # Schiefe der Ekliptik in Grad, Zeit t
    epsRad = radians(eps)       # Schiefe der Ekliptik in Radians
    omegabar = getOmegabar(t)   # Länge des Perihels in Grad, Zeit t
    L = getL(t)                 # mittlere Länge der Sonne in Grad, Zeit t
    Lhaupt = Mod(L, 360)        # Hauptwert der Länge der Sonne
    M = Lhaupt - omegabar       # Mittlere Anomalie der Sonne in Grad, Zeit t
    Mhaupt = Mod(M, 360)        # Hauptwert der mittleren Anomalie
    C = getC(e, Mhaupt)         # Mittelpunktsgleichung, wahre minus mittlere A
    Lwahr = L + C               # wahre Länge der Sonne in Grad, Zeit t
    LwahrRad = radians(Lwahr)   # wahre Länge der Sonne in Radians
    Lwahrhaupt = Mod(Lwahr, 360)# Hauptwert der wahren Länge
    LwahrHauptRad = radians(Lwahrhaupt)  # dito, in Radians


    # Rektaszension der Sonne in Funktion der wahren Länge der Sonne und der 
    # Schiefe der Ekliptik
    alpha = atanLage(degrees(tan(LwahrRad)*cos(epsRad)), Lwahr)
    alphahaupt = atanLage(degrees(tan(LwahrHauptRad)*cos(epsRad)), Lwahrhaupt)

    Dict = { "Object"   : "SunStuff",
            "TT"        : t,            # Tage nach Epoch2010 inkl. deltaT
            "MPEq"      : C,            # wahre - mittlere Anomalie
            "omegabar"  : omegabar,     # Länge des perihels in Grad, Zeit t
            "M"         : M,            # mittlere Anomalie, Grad
            "nu"        : C+M,          # wahre Anomalie
            "Long"      : Lwahrhaupt,   # wahre Länge der Sonne, Grad
            "eps"       : eps,          # Schiefe der Ekliptik, Grad
            "exc"       : e,            # Exzentrizität der Erdbahn
            }
    if verbose:
        print ("Numerische Exzentrizitaet e = ", "%.6f" % e)
        print ("Omegabar, Laenge des Perihels = ", "%.3f Grad" % omegabar)
        print ("Lhaupt, Hauptwert Mittlere Laenge = ", "%.3f Grad" % Lhaupt)
        print ("M Mittlere Anomalie = ", "%.3f Grad" % M)
        print ("Mhaupt, Hauptwert Mittlere Anomalie = ", "%.3f Grad" % Mhaupt)
        print ("C Mittelpunktsgleichung = ", "%.3f Grad" % C)
        print ("Alpha Wahre Anomalie = ", "%.3f Grad" % alpha)
    return (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict)

#######################################################################    
def Zeitgleichung(JD, verbose=False):
    # berechnet die Zeitgleichung in ganzen Sekunden
    # JD Julianisches Datum
    # verbose aktiviert zusätzlichen Output
    # Resultat in Sekunden ohne Rundung
    # Epoche konsistent 2010.0, 31.12.2009 0h UT, JD 2455196.5
    dtEpoch2010 = datetime(2009, 12, 31)
    tbase = getJulianDay (dtEpoch2010)
    t = JD - tbase
    TT = JD - tbase + getDeltaT(JD)/86400.

    (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict) = getSunStufffromTT(TT, verbose)

    # Zeitgleichung erste Ursache Exzentrizität der Erdbahn (epsilon = 0)
    ZGstar0 = L - Lwahr         # in Grad
    ZG0 = 240. * ZGstar0        # in Sekunden

    # zweite Ursache Anteil Schiefe der Ekliptik
    ZGstareps = Lwahrhaupt - alphahaupt # in Grad
    ZGeps = 240. * ZGstareps            # in Sekunden

    # Summe der beiden Ursachen
    ZGstar = ZGstar0 + ZGstareps        # in Grad
    ZGsecs = ZG0 + ZGeps                # in Sekunden

    # ZGmin, ZGsec for printing purposes 
    ZGstr = niceminsec(ZGsecs)
    if verbose:
        print ("Epoch2010 = ", dtEpoch2010)
        print ("Epoch Julian Date = ", tbase)
        print ("TT = ", TT)
        print ("Schiefe der Ekliptik eps = ", "%.3f Grad" % eps)
        print ("L, Mittlere Laenge = ", "%.3f Grad" % L)
        print ("Lwahr Wahre Sonnenlaenge = ", "%.3f Grad" % Lwahr)
        print ("Lwahrhaupt Hauptwert Wahre Sonnenlaenge = ", "%.3f Grad" \
                    % Lwahrhaupt)
        print ("Alphahaupt Wahre Anomalie = ", "%.3f Grad" % alphahaupt)
        print ("ZGstar0 = ", "%.3f" % ZGstar0, " Grad")
        print ("ZG0 = ", "%.2f" % ZG0, " Sekunden")
        print ("ZGstareps = ", "%.3f" % ZGstareps, " Grad")
        print ("ZGeps = ", "%.2f" % ZGeps, " Sekunden")
        print ("ZGstar = ", "%.3f" % ZGstar, " Grad")
        print ("ZGsecs = ", "%.2f" % ZGsecs, " Sekunden")
        print ("ZG = %s" % ZGstr)

    return (ZGsecs, ZGstr, Dict)

def getZeitgleichung(JD, verbose=False):
    # returns a timedelta object and two int's for printing
    zgsecs, ZGstr, Dict = Zeitgleichung(JD, verbose)
    return (timedelta(seconds=zgsecs), ZGstr, Dict)

# Greenwich Mean Sidereal Time from Julian Date

def getGMSTfromJD(JD):
    '''
    # new Epoch2010, JD=2455196.5
    # JD usually float or int
    # returns a datetime object
    UT = Frac(JD-0.5) * 24.     # UT in hours
    JD = floor(JD-0.5) + 0.5    # JD at 0 hours UT
    # JD 2455196.5 is Epoch 2010.0
    # T is centuries after Epoch
    T  = (JD - 2455196.5) / 36525.
    T0 = 6.636798592 + T * (2400.051341 + T * 0.000025862)
    ST = Mod(T0 + UT * 1.002737909, 24.)
    Gdate = getGreenwichDate(JD)
    return Gdate + timedelta(hours=ST)
    '''

    # Meeus chapter 11 p. 83
    T = (JD - 2451545.0) / 36525.   # (11.1)
    theta0 = 280.46061837 + 360.98564736629 * (JD - 2451545.0) \
            + 0.000387933 * T*T - T*T*T/387110000   # (11.4)
    # normalize to 360 degrees
    theta = Mod(theta0, 360.)
    # add to midnight
    JD0 = floor(JD-0.5) + 0.5
    Gdate = hvfTime(JD0 + theta/360.)
    return Gdate

# Local Mean Sideral Time from GMST, longitude
# returns a datetime object

def getLMSTfromGMST(gmst, longitude):
    # gmst datetime object
    # longitude in degrees
    # returns a datetime object
    lmst = gmst + timedelta(hours=longitude/15.)
    return lmst

# Greenwich Mean Sideral Time from LMST, longitude
# returns a datetime object

def getGMSTfromLMST(lmst, longitude):
    # lmst datetime object
    # longitude in degrees
    # returns a datetime object
    #gmst = datetime.time(hour=lmst) - timedelta(hours=longitude/15.)
    gmst = lmst - timedelta(hours=longitude/15.)
    return gmst
    

# deltaT, the difference between Terrestrial Time TT and
# Universal Time 
# see https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
# for Julian Date JD

def getDeltaT(JD): 
    # convert JD into year, month, day
    '''
    dtymd = getGreenwichDate(JD)
    year = dtymd.year
    month = dtymd.month
    y = year + (month-0.5)/12.
    # print("getDeltaT year, month, y = %d %d %.2f" % (year, month, y))
    '''
    (year, month, day) = hvfTime(JD).getDate()
    # We define the decimal year "y" as follows:
    y = year + (month - 0.5)/12

    if year < -500:
        u = (year-1820.)/100.
        deltaT = -20 + 32*u*u
    elif year < 500:
        u = y/100.
        deltaT = 10583.6 - 1014.41*u + 33.78311*u*u - 5.952053*u**3 \
                - 0.1798452*u**4 + 0.022174192*u**5 + 0.0090316521*u**6
    elif year < 1600:
        u = (y-1000.)/100.
        deltaT = 1574.2 - 556.01*u + 71.23472*u*u + 0.319781*u**3 \
                - 0.8503463*u**4 - 0.005050998*u**5 + 0.0083572073*u**6
    elif year < 1700:
        t = y - 1600.
        deltaT = 120. -0.9808*t - 0.01532*t*t + t**3/7129
    elif year < 1800:
        t = y - 1700.
        deltaT = 8.83 + 0.1603*t - 0.0059285*t*t + 0.00013336*t**3 \
                - t**4/1174000.
    elif year < 1860:
        t = y - 1800.
        deltaT = 13.72 - 0.332447*t + 0.0068612*t*t + 0.0041116*t**3 \
                - 0.00037436*t**4 + 0.0000121272*t**5 \
                - 0.0000001699*t**6 + 0.000000000875*t**7
    elif year < 1900:
        t = y - 1860.
        deltaT = 7.62 + 0.5737*t - 0.251754*t*t + 0.01680668*t**3 \
                - 0.0004473624*t**4 + t**5/233174.
    elif year < 1920:
        t = y - 1900.
        deltaT = -2.79 + 1.494119*t - 0.0598939*t*t + 0.0061966*t**3 \
                -0.000197*t**4
    elif year < 1941:
        t = y - 1920.
        deltaT = 21.20 + 0.84493*t - 0.076100*t*t + 0.0020936*t**3
    elif year < 1961:
        t = y - 1950.
        deltaT = 29.07 + 0.407*t - t*t/233. + t**3/2547.
    elif year < 1986:
        t = y - 1975.
        deltaT = 45.45 + 1.067*t - t*t/260. - t**3/718.
    elif year < 2005:
        t = y - 2000.
        deltaT = 63.86 + 0.3345*t - 0.060374*t*t + 0.0017275*t**3 \
                + 0.000651814*t**4 + 0.00002373599*t**5
    elif year < 2050:
        t = y - 2000.
        deltaT = 62.92 + 0.32217*t + 0.005589*t*t
    elif year < 2150:
        deltaT = -20. + 32*((y-1820.)/100)**2 - 0.5628*(2150.-y)
    else:   # after 2150
        u = (year-1820.)/100.
        deltaT = -20. + 32.*u*u
        # https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
        # small correction beyond 2150 (only needed for Perihelion calc's)
        c = -0.000012932 * (y - 1955.)**2
        deltaT += c

    return deltaT

def getEps(t):  # Schiefe der Ekliptik in Grad
    # t in Tage nach Epoch2010
    # Resultat in Grad
    T = t / 36525
    eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
    return eps


#######################################################################
# Jean Meeus, Astronomical Algorithms, Willmann-Bell, Richmond VA, 1991
#
# hvf 03.12.2020

# quadratic interpolation
# xi 3 Stuetzstellen, yi 3 Funktionswerte
# Meeus p. 25
def interp(xi, yi, x):
    a10 = (yi[1]-yi[0])/(xi[1]-xi[0])
    a21 = (yi[2]-yi[1])/(xi[2]-xi[1])
    a210 = (a21-a10)/(xi[2]-xi[0])
    res = yi[0] + a10*(x-xi[0]) + a210*(x-xi[0])*(x-xi[1])
    return res

#############################################################################

from datetime import datetime
global EarthdtUTimes
EarthdtUTimes = {}  # empty dict of times
#!/usr/bin/env python
# coding: utf-8
#   getEarthLBRData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Earth"
#       Long  :  heliocentric Longitude, degrees
#       Lat   :  heliocentric Latitude,degrees
#       Rau   :  Radius, distaance Sun - Object (au)
#
#   Similarly for each planet
#   data from Meeus, Appendix II and VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.2 hvf 2023-04-06 getEarthLBRData.py created by mkPlanetLBRData.py"
VersionList.append(Version)

def getEarthLBRData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. (31.1) Julian Millennia

#######################################################################


    #strdtUT = dtUT.isoformat(sep=' ')     # convert to string and use as key to dict
    #strdtUT = str(dtUT)     # usually yyyy-mm-dd HH:MM:SS.sss
    strdtUT = dtUT.getStr()     # usually yyyy-mm-dd HH:MM:SS.sss
    # check if we have been here before
    if strdtUT in EarthdtUTimes:
        Data = EarthdtUTimes[strdtUT]
        # print ("getEarthLBRData using stored data %s" % strdtUT)
        return Data # L, B, R, Dict
    # print ("getEarthLBRData calculate fresh: dtUT = %s" % strdtUT)

    # EARTH Longitude
    sumL0  = 0
    sumL0  +=  175347045.673 * cos(0.00000000000  + 0.00000000000 *T)
    sumL0  +=  3341656.456 * cos(4.66925680417  + 6283.07584999140 *T)
    sumL0  +=   34894.275 * cos(4.62610241759  + 12566.15169998280 *T)
    sumL0  +=    3417.571 * cos(2.82886579606  + 3.52311834900 *T)
    sumL0  +=    3497.056 * cos(2.74411800971  + 5753.38488489680 *T)
    sumL0  +=    3135.896 * cos(3.62767041758  + 77713.77146812050 *T)
    sumL0  +=    2676.218 * cos(4.41808351397  + 7860.41939243920 *T)
    sumL0  +=    2342.687 * cos(6.13516237631  + 3930.20969621960 *T)
    sumL0  +=    1273.166 * cos(2.03709655772  + 529.69096509460 *T)
    sumL0  +=    1324.292 * cos(0.74246356352  + 11506.76976979360 *T)
    sumL0  +=     901.855 * cos(2.04505443513  + 26.29831979980 *T)
    sumL0  +=    1199.167 * cos(1.10962944315  + 1577.34354244780 *T)
    sumL0  +=     857.223 * cos(3.50849156957  + 398.14900340820 *T)
    sumL0  +=     779.786 * cos(1.17882652114  + 5223.69391980220 *T)
    sumL0  +=     990.250 * cos(5.23268129594  + 5884.92684658320 *T)
    sumL0  +=     753.141 * cos(2.53339053818  + 5507.55323866740 *T)
    sumL0  +=     505.264 * cos(4.58292563052  + 18849.22754997420 *T)
    sumL0  +=     492.379 * cos(4.20506639861  + 775.52261132400 *T)
    sumL0  +=     356.655 * cos(2.91954116867  + 0.06731030280 *T)
    sumL0  +=     284.125 * cos(1.89869034186  + 796.29800681640 *T)
    sumL0  +=     242.810 * cos(0.34481140906  + 5486.77784317500 *T)
    sumL0  +=     317.087 * cos(5.84901952218  + 11790.62908865880 *T)
    sumL0  +=     271.039 * cos(0.31488607649  + 10977.07880469900 *T)
    sumL0  +=     206.160 * cos(4.80646606059  + 2544.31441988340 *T)
    sumL0  +=     205.385 * cos(1.86947813692  + 5573.14280143310 *T)
    sumL0  +=     202.261 * cos(2.45767795458  + 6069.77675455340 *T)
    sumL0  +=     126.184 * cos(1.08302630210  + 20.77539549240 *T)
    sumL0  +=     155.516 * cos(0.83306073807  + 213.29909543800 *T)
    sumL0  +=     115.132 * cos(0.64544911683  + 0.98032106820 *T)
    sumL0  +=     102.851 * cos(0.63599846727  + 4694.00295470760 *T)
    sumL0  +=     101.724 * cos(4.26679821365  + 7.11354700080 *T)
    sumL0  +=      99.206 * cos(6.20992940258  + 2146.16541647520 *T)
    sumL0  +=     132.212 * cos(3.41118275555  + 2942.46342329160 *T)
    sumL0  +=      97.607 * cos(0.68101272270  + 155.42039943420 *T)
    sumL0  +=      85.128 * cos(1.29870743025  + 6275.96230299060 *T)
    sumL0  +=      74.651 * cos(1.75508916159  + 5088.62883976680 *T)
    sumL0  +=     101.895 * cos(0.97569221824  + 15720.83878487840 *T)
    sumL0  +=      84.711 * cos(3.67080093025  + 71430.69561812909 *T)
    sumL0  +=      73.547 * cos(4.67926565481  + 801.82093112380 *T)
    sumL0  +=      73.874 * cos(3.50319443167  + 3154.68708489560 *T)
    sumL0  +=      78.756 * cos(3.03698313141  + 12036.46073488820 *T)
    sumL0  +=      79.637 * cos(1.80791330700  + 17260.15465469040 *T)
    sumL0  +=      85.803 * cos(5.98322631256  + 161000.68573767410 *T)
    sumL0  +=      56.963 * cos(2.78430398043  + 6286.59896834040 *T)
    sumL0  +=      61.148 * cos(1.81839811024  + 7084.89678111520 *T)
    sumL0  +=      69.627 * cos(0.83297596966  + 9437.76293488700 *T)
    sumL0  +=      56.116 * cos(4.38694880779  + 14143.49524243060 *T)
    sumL0  +=      62.449 * cos(3.97763880587  + 8827.39026987480 *T)
    sumL0  +=      51.145 * cos(0.28306864501  + 5856.47765911540 *T)
    sumL0  +=      55.577 * cos(3.47006009062  + 6279.55273164240 *T)
    sumL0  +=      41.036 * cos(5.36817351402  + 8429.24126646660 *T)
    sumL0  +=      51.605 * cos(1.33282746983  + 1748.01641306700 *T)
    sumL0  +=      51.992 * cos(0.18914945834  + 12139.55350910680 *T)
    sumL0  +=      49.000 * cos(0.48735065033  + 1194.44701022460 *T)
    sumL0  +=      39.200 * cos(6.16832995016  + 10447.38783960440 *T)
    sumL0  +=      35.566 * cos(1.77597314691  + 6812.76681508600 *T)
    sumL0  +=      36.770 * cos(6.04133859347  + 10213.28554621100 *T)
    sumL0  +=      36.596 * cos(2.56955238628  + 1059.38193018920 *T)
    sumL0  +=      33.291 * cos(0.59309499459  + 17789.84561978500 *T)
    sumL0  +=      35.954 * cos(1.70876111898  + 2352.86615377180 *T)
    sumL0  +=      40.938 * cos(2.39850881707  + 19651.04848109800 *T)
    sumL0  +=      30.047 * cos(2.73975123935  + 1349.86740965880 *T)
    sumL0  +=      30.412 * cos(0.44294464135  + 83996.84731811189 *T)
    sumL0  +=      23.663 * cos(0.48473567763  + 8031.09226305840 *T)
    sumL0  +=      23.574 * cos(2.06527720049  + 3340.61242669980 *T)
    sumL0  +=      21.089 * cos(4.14825464101  + 951.71840625060 *T)
    sumL0  +=      24.738 * cos(0.21484762138  + 3.59042865180 *T)
    sumL0  +=      25.352 * cos(3.16470953405  + 4690.47983635860 *T)
    sumL0  +=      22.820 * cos(5.22197888032  + 4705.73230754360 *T)
    sumL0  +=      21.419 * cos(1.42563735525  + 16730.46368959580 *T)
    sumL0  +=      21.891 * cos(5.55594302562  + 553.56940284240 *T)
    sumL0  +=      17.481 * cos(4.56052900359  + 135.06508003540 *T)
    sumL0  +=      19.925 * cos(5.22208471269  + 12168.00269657460 *T)
    sumL0  +=      19.860 * cos(5.77470167653  + 6309.37416979120 *T)
    sumL0  +=      20.300 * cos(0.37133792946  + 283.85931886520 *T)
    sumL0  +=      14.421 * cos(4.19315332546  + 242.72860397400 *T)
    sumL0  +=      16.225 * cos(5.98837722564  + 11769.85369316640 *T)
    sumL0  +=      15.077 * cos(4.19567181073  + 6256.77753019160 *T)
    sumL0  +=      19.124 * cos(3.82219996949  + 23581.25817731760 *T)
    sumL0  +=      18.888 * cos(5.38626880969  + 149854.40013480789 *T)
    sumL0  +=      14.346 * cos(3.72355084422  + 38.02767263580 *T)
    sumL0  +=      17.898 * cos(2.21490735647  + 13367.97263110660 *T)
    sumL0  +=      12.054 * cos(2.62229588349  + 955.59974160860 *T)
    sumL0  +=      11.287 * cos(0.17739328092  + 4164.31198961300 *T)
    sumL0  +=      13.971 * cos(4.40138139996  + 6681.22485339960 *T)
    sumL0  +=      13.621 * cos(1.88934471407  + 7632.94325965020 *T)
    sumL0  +=      12.503 * cos(1.13052412208  + 5.52292430740 *T)
    sumL0  +=      10.498 * cos(5.35909518669  + 1592.59601363280 *T)
    sumL0  +=       9.803 * cos(0.99947478995  + 11371.70468975820 *T)
    sumL0  +=       9.220 * cos(4.57138609781  + 4292.33083295040 *T)
    sumL0  +=      10.327 * cos(6.19982566125  + 6438.49624942560 *T)
    sumL0  +=      12.003 * cos(1.00351456700  + 632.78373931320 *T)
    sumL0  +=      10.827 * cos(0.32734520222  + 103.09277421860 *T)
    sumL0  +=       8.356 * cos(4.53902685948  + 25132.30339996560 *T)
    sumL0  +=      10.005 * cos(6.02914963280  + 5746.27133789600 *T)
    sumL0  +=       8.409 * cos(3.29946744189  + 7234.79425624200 *T)
    sumL0  +=       8.006 * cos(5.82145271907  + 28.44918746780 *T)
    sumL0  +=      10.523 * cos(0.93871805506  + 11926.25441366880 *T)
    sumL0  +=       7.686 * cos(3.12142363172  + 7238.67559160000 *T)
    sumL0  +=       9.378 * cos(2.62414241032  + 5760.49843189760 *T)
    sumL0  +=       8.127 * cos(6.11228001785  + 4732.03062734340 *T)
    sumL0  +=       9.232 * cos(0.48343968736  + 522.57741809380 *T)
    sumL0  +=       9.802 * cos(5.24413991147  + 27511.46787353720 *T)
    sumL0  +=       7.871 * cos(0.99590177926  + 5643.17856367740 *T)
    sumL0  +=       8.123 * cos(6.27053013650  + 426.59819087600 *T)
    sumL0  +=       9.048 * cos(5.33686335897  + 6386.16862421000 *T)
    sumL0  +=       8.620 * cos(4.16538210888  + 7058.59846131540 *T)
    sumL0  +=       6.297 * cos(4.71724819317  + 6836.64525283380 *T)
    sumL0  +=       7.575 * cos(3.97382858911  + 11499.65622279280 *T)
    sumL0  +=       7.756 * cos(2.95729056763  + 23013.53953958720 *T)
    sumL0  +=       7.314 * cos(0.60652505806  + 11513.88331679440 *T)
    sumL0  +=       5.955 * cos(2.87641047971  + 6283.14316029419 *T)
    sumL0  +=       6.534 * cos(5.79072926033  + 18073.70493865020 *T)
    sumL0  +=       7.188 * cos(3.99831508699  + 74.78159856730 *T)
    sumL0  +=       7.346 * cos(4.38582365437  + 316.39186965660 *T)
    sumL0  +=       5.413 * cos(5.39199024641  + 419.48464387520 *T)
    sumL0  +=       5.127 * cos(2.36062848786  + 10973.55568635000 *T)
    sumL0  +=       7.056 * cos(0.32258441903  + 263.08392337280 *T)
    sumL0  +=       6.625 * cos(3.66475158672  + 17298.18232732620 *T)
    sumL0  +=       6.762 * cos(5.91132535899  + 90955.55169449610 *T)
    sumL0  +=       4.938 * cos(5.73672165674  + 9917.69687450980 *T)
    sumL0  +=       5.547 * cos(2.45152597661  + 12352.85260454480 *T)
    sumL0  +=       5.958 * cos(3.32051344676  + 6283.00853968860 *T)
    sumL0  +=       4.471 * cos(2.06385999536  + 7079.37385680780 *T)
    sumL0  +=       6.153 * cos(1.45823331144  + 233141.31440436147 *T)
    sumL0  +=       4.348 * cos(4.42342175480  + 5216.58037280140 *T)
    sumL0  +=       6.123 * cos(1.07494905258  + 19804.82729158280 *T)
    sumL0  +=       4.488 * cos(3.65285037150  + 206.18554843720 *T)
    sumL0  +=       4.020 * cos(0.83995823171  + 20.35531939880 *T)
    sumL0  +=       5.188 * cos(4.06503864016  + 6208.29425142410 *T)
    sumL0  +=       5.307 * cos(0.38217636096  + 31441.67756975680 *T)
    sumL0  +=       3.785 * cos(2.34369213733  + 3.88133535800 *T)
    sumL0  +=       4.497 * cos(3.27230796845  + 11015.10647733480 *T)
    sumL0  +=       4.132 * cos(0.92128915753  + 3738.76143010800 *T)
    sumL0  +=       3.521 * cos(5.97844807108  + 3894.18182954220 *T)
    sumL0  +=       4.215 * cos(1.90601120623  + 245.83164622940 *T)
    sumL0  +=       3.701 * cos(5.03069397926  + 536.80451209540 *T)
    sumL0  +=       3.865 * cos(1.82634360607  + 11856.21865142450 *T)
    sumL0  +=       3.652 * cos(1.01838584934  + 16200.77272450120 *T)
    sumL0  +=       3.390 * cos(0.97785123922  + 8635.94200376320 *T)
    sumL0  +=       3.737 * cos(2.95380107829  + 3128.38876509580 *T)
    sumL0  +=       3.507 * cos(3.71291946325  + 6290.18939699220 *T)
    sumL0  +=       3.086 * cos(3.64646921512  + 10.63666534980 *T)
    sumL0  +=       3.397 * cos(1.10590684017  + 14712.31711645800 *T)
    sumL0  +=       3.334 * cos(0.83684924911  + 6496.37494542940 *T)
    sumL0  +=       2.805 * cos(2.58504514144  + 14314.16811304980 *T)
    sumL0  +=       3.650 * cos(1.08344142571  + 88860.05707098669 *T)
    sumL0  +=       3.388 * cos(3.20185096055  + 5120.60114558360 *T)
    sumL0  +=       3.252 * cos(3.47859752062  + 6133.51265285680 *T)
    sumL0  +=       2.553 * cos(3.94869034189  + 1990.74501704100 *T)
    sumL0  +=       3.520 * cos(2.05559692878  + 244287.60000722768 *T)
    sumL0  +=       2.565 * cos(1.56071784900  + 23543.23050468179 *T)
    sumL0  +=       2.621 * cos(3.85639359951  + 266.60704172180 *T)
    sumL0  +=       2.955 * cos(3.39692949667  + 9225.53927328300 *T)
    sumL0  +=       2.876 * cos(6.02635617464  + 154717.60988768269 *T)
    sumL0  +=       2.395 * cos(1.16131956403  + 10984.19235169980 *T)
    sumL0  +=       3.161 * cos(1.32798718453  + 10873.98603048040 *T)
    sumL0  +=       3.163 * cos(5.08946464629  + 21228.39202354580 *T)
    sumL0  +=       2.361 * cos(4.27212906992  + 6040.34724601740 *T)
    sumL0  +=       3.030 * cos(1.80209931347  + 35371.88726597640 *T)
    sumL0  +=       2.343 * cos(3.57689860500  + 10969.96525769820 *T)
    sumL0  +=       2.618 * cos(2.57870156528  + 22483.84857449259 *T)
    sumL0  +=       2.113 * cos(3.71393780256  + 65147.61976813770 *T)
    sumL0  +=       2.019 * cos(0.81393923319  + 170.67287061920 *T)
    sumL0  +=       2.003 * cos(0.38091017375  + 6172.86952877200 *T)
    sumL0  +=       2.506 * cos(3.74379142438  + 10575.40668294180 *T)
    sumL0  +=       2.381 * cos(0.10581361289  + 7.04623669800 *T)
    sumL0  +=       1.949 * cos(4.86892513469  + 36.02786667740 *T)
    sumL0  +=       2.074 * cos(4.22794774570  + 5650.29211067820 *T)
    sumL0  +=       1.924 * cos(5.59460549860  + 6282.09552892320 *T)
    sumL0  +=       1.949 * cos(1.07002512703  + 5230.80746680300 *T)
    sumL0  +=       1.988 * cos(5.19736046771  + 6262.30045449900 *T)
    sumL0  +=       1.887 * cos(3.74365662683  + 23.87843774780 *T)
    sumL0  +=       1.787 * cos(1.25929682929  + 12559.03815298200 *T)
    sumL0  +=       1.883 * cos(1.90364058477  + 15.25247118500 *T)
    sumL0  +=       1.816 * cos(3.68083868442  + 15110.46611986620 *T)
    sumL0  +=       1.701 * cos(4.41105895380  + 110.20632121940 *T)
    sumL0  +=       1.990 * cos(3.93295788548  + 6206.80977871580 *T)
    sumL0  +=       2.103 * cos(0.75354917468  + 13521.75144159140 *T)
    sumL0  +=       1.774 * cos(0.48747535361  + 1551.04522264800 *T)
    sumL0  +=       1.882 * cos(0.86684493432  + 22003.91463486980 *T)
    sumL0  +=       1.924 * cos(1.22898324132  + 709.93304855830 *T)
    sumL0  +=       2.009 * cos(4.62850921980  + 6037.24420376200 *T)
    sumL0  +=       1.924 * cos(0.60231842508  + 6284.05617105960 *T)
    sumL0  +=       1.596 * cos(3.98332956992  + 13916.01910964160 *T)
    sumL0  +=       1.664 * cos(4.41939715469  + 8662.24032356300 *T)
    sumL0  +=       1.971 * cos(1.04560500503  + 18209.33026366019 *T)
    sumL0  +=       1.942 * cos(4.31335979989  + 6244.94281435360 *T)
    sumL0  +=       1.476 * cos(0.93271367331  + 2379.16447357160 *T)
    sumL0  +=       1.810 * cos(0.49112137707  + 1.48447270830 *T)
    sumL0  +=       1.346 * cos(1.51574702235  + 4136.91043351620 *T)
    sumL0  +=       1.528 * cos(5.61835711404  + 6127.65545055720 *T)
    sumL0  +=       1.791 * cos(3.22187270126  + 39302.09696219600 *T)
    sumL0  +=       1.747 * cos(3.05638656738  + 18319.53658487960 *T)
    sumL0  +=       1.431 * cos(4.51153808594  + 20426.57109242200 *T)
    sumL0  +=       1.695 * cos(0.22047718414  + 25158.60171976540 *T)
    sumL0  +=       1.242 * cos(4.46665769933  + 17256.63153634140 *T)
    sumL0  +=       1.463 * cos(4.69242679213  + 14945.31617355440 *T)
    sumL0  +=       1.205 * cos(1.86912144659  + 4590.91018048900 *T)
    sumL0  +=       1.192 * cos(2.74227166898  + 12569.67481833180 *T)
    sumL0  +=       1.222 * cos(5.18120087482  + 5333.90024102160 *T)
    sumL0  +=       1.390 * cos(5.42894648983  + 143571.32428481648 *T)
    sumL0  +=       1.473 * cos(1.70479245805  + 11712.95531823080 *T)
    sumL0  +=       1.362 * cos(2.61069503292  + 6062.66320755260 *T)
    sumL0  +=       1.148 * cos(6.03001800540  + 3634.62102451840 *T)
    sumL0  +=       1.198 * cos(5.15294130422  + 10177.25767953360 *T)
    sumL0  +=       1.266 * cos(0.11421493643  + 18422.62935909819 *T)
    sumL0  +=       1.411 * cos(1.09908857534  + 3496.03282613400 *T)
    sumL0  +=       1.349 * cos(2.99805109633  + 17654.78053974960 *T)
    sumL0  +=       1.253 * cos(2.79850152848  + 167283.76158766547 *T)
    sumL0  +=       1.311 * cos(1.60942984879  + 5481.25491886760 *T)
    sumL0  +=       1.079 * cos(6.20304501787  + 3.28635741780 *T)
    sumL0  +=       1.181 * cos(1.20653776978  + 131.54196168640 *T)
    sumL0  +=       1.254 * cos(5.45103277798  + 6076.89030155420 *T)
    sumL0  +=       1.035 * cos(2.32142722747  + 7342.45778018060 *T)
    sumL0  +=       1.117 * cos(0.38838354256  + 949.17560896980 *T)
    sumL0  +=       0.966 * cos(3.18341890851  + 11087.28512591840 *T)
    sumL0  +=       1.171 * cos(3.39635049962  + 12562.62858163380 *T)
    sumL0  +=       1.121 * cos(0.72627490378  + 220.41264243880 *T)
    sumL0  +=       1.024 * cos(2.19378315386  + 11403.67699557500 *T)
    sumL0  +=       0.888 * cos(3.91173199285  + 4686.88940770680 *T)
    sumL0  +=       0.910 * cos(1.98802695087  + 735.87651353180 *T)
    sumL0  +=       0.830 * cos(0.48984915507  + 24072.92146977640 *T)
    sumL0  +=       1.096 * cos(6.17377835617  + 5436.99301524020 *T)
    sumL0  +=       0.908 * cos(0.44959639433  + 7477.52286021600 *T)
    sumL0  +=       0.974 * cos(1.52996238356  + 9623.68827669120 *T)
    sumL0  +=       0.840 * cos(1.79543266333  + 5429.87946823940 *T)
    sumL0  +=       0.778 * cos(6.17699177946  + 38.13303563780 *T)
    sumL0  +=       0.776 * cos(4.09855402433  + 14.22709400160 *T)
    sumL0  +=       1.068 * cos(4.64200173735  + 43232.30665841560 *T)
    sumL0  +=       0.954 * cos(1.49988435748  + 1162.47470440780 *T)
    sumL0  +=       0.907 * cos(0.86986870809  + 10344.29506538580 *T)
    sumL0  +=       0.931 * cos(4.06044689031  + 28766.92442448400 *T)
    sumL0  +=       0.739 * cos(5.04368197372  + 639.89728631400 *T)
    sumL0  +=       0.937 * cos(3.46884698960  + 1589.07289528380 *T)
    sumL0  +=       0.763 * cos(5.86304932998  + 16858.48253293320 *T)
    sumL0  +=       0.953 * cos(4.20801492835  + 11190.37790013700 *T)
    sumL0  +=       0.708 * cos(1.72899988940  + 13095.84266507740 *T)
    sumL0  +=       0.969 * cos(1.64439522215  + 29088.81141598500 *T)
    sumL0  +=       0.717 * cos(0.16688678895  + 11.72935283600 *T)
    sumL0  +=       0.962 * cos(3.53092337542  + 12416.58850284820 *T)
    sumL0  +=       0.747 * cos(5.77866940346  + 12592.45001978260 *T)
    sumL0  +=       0.672 * cos(1.91095796194  + 3.93215326310 *T)
    sumL0  +=       0.671 * cos(5.46240843677  + 18052.92954315780 *T)
    sumL0  +=       0.675 * cos(6.28311558823  + 4535.05943692440 *T)
    sumL0  +=       0.684 * cos(0.39975012080  + 5849.36411211460 *T)
    sumL0  +=       0.799 * cos(0.29851185294  + 12132.43996210600 *T)
    sumL0  +=       0.758 * cos(0.96370823331  + 1052.26838318840 *T)
    sumL0  +=       0.782 * cos(5.33878339919  + 13517.87010623340 *T)
    sumL0  +=       0.730 * cos(1.70106160291  + 17267.26820169119 *T)
    sumL0  +=       0.749 * cos(2.59599901875  + 11609.86254401220 *T)
    sumL0  +=       0.734 * cos(2.78417782952  + 640.87760738220 *T)
    sumL0  +=       0.688 * cos(5.15048287468  + 16496.36139620240 *T)
    sumL0  +=       0.770 * cos(1.62469589333  + 4701.11650170840 *T)
    sumL0  +=       0.633 * cos(2.20587893893  + 25934.12433108940 *T)
    sumL0  +=       0.760 * cos(4.21317219403  + 377.37360791580 *T)
    sumL0  +=       0.584 * cos(2.13420121623  + 10557.59416082380 *T)
    sumL0  +=       0.574 * cos(0.24250054587  + 9779.10867612540 *T)
    sumL0  +=       0.573 * cos(3.16435264609  + 533.21408344360 *T)
    sumL0  +=       0.685 * cos(3.19344289472  + 12146.66705610760 *T)
    sumL0  +=       0.675 * cos(0.96179233959  + 10454.50138660520 *T)
    sumL0  +=       0.648 * cos(1.46327342555  + 6268.84875598980 *T)
    sumL0  +=       0.589 * cos(2.50543543638  + 3097.88382272579 *T)
    sumL0  +=       0.551 * cos(5.28099026956  + 9388.00590941520 *T)
    sumL0  +=       0.696 * cos(3.65342150016  + 4804.20927592700 *T)
    sumL0  +=       0.669 * cos(2.51030077026  + 2388.89402044920 *T)
    sumL0  +=       0.550 * cos(0.06883864342  + 20199.09495963300 *T)
    sumL0  +=       0.629 * cos(4.13350995675  + 45892.73043315699 *T)
    sumL0  +=       0.678 * cos(6.09190163533  + 135.62532501000 *T)
    sumL0  +=       0.593 * cos(1.50136257618  + 226858.23855437007 *T)
    sumL0  +=       0.542 * cos(3.58573645173  + 6148.01076995600 *T)
    sumL0  +=       0.682 * cos(5.02203067788  + 17253.04110768959 *T)
    sumL0  +=       0.565 * cos(4.29309238610  + 11933.36796066960 *T)
    sumL0  +=       0.486 * cos(0.77746204893  + 27.40155609680 *T)
    sumL0  +=       0.503 * cos(0.58963565969  + 15671.08175940660 *T)
    sumL0  +=       0.616 * cos(4.06539884128  + 227.47613278900 *T)
    sumL0  +=       0.583 * cos(6.12695541996  + 18875.52586977400 *T)
    sumL0  +=       0.537 * cos(2.15056440980  + 21954.15760939799 *T)
    sumL0  +=       0.669 * cos(6.06986269566  + 47162.51635463520 *T)
    sumL0  +=       0.475 * cos(0.40343842110  + 6915.85958930460 *T)
    sumL0  +=       0.540 * cos(2.83444222174  + 5326.78669402080 *T)
    sumL0  +=       0.530 * cos(5.26359885263  + 10988.80815753500 *T)
    sumL0  +=       0.582 * cos(3.24533095664  + 153.77881048480 *T)
    sumL0  +=       0.641 * cos(3.24711791371  + 2107.03450754240 *T)
    sumL0  +=       0.621 * cos(3.09698523779  + 33019.02111220460 *T)
    sumL0  +=       0.466 * cos(3.14982372198  + 10440.27429260360 *T)
    sumL0  +=       0.466 * cos(0.90708835657  + 5966.68398033480 *T)
    sumL0  +=       0.528 * cos(0.81926454470  + 813.55028395980 *T)
    sumL0  +=       0.603 * cos(3.81378921927  + 316428.22867391503 *T)
    sumL0  +=       0.559 * cos(1.81894804124  + 17996.03116822220 *T)
    sumL0  +=       0.437 * cos(2.28625594435  + 6303.85124548380 *T)
    sumL0  +=       0.518 * cos(4.86069178322  + 20597.24396304120 *T)
    sumL0  +=       0.424 * cos(6.23520018693  + 6489.26139842860 *T)
    sumL0  +=       0.518 * cos(6.17617826756  + 0.24381748350 *T)
    sumL0  +=       0.404 * cos(5.72804304258  + 5642.19824260920 *T)
    sumL0  +=       0.458 * cos(1.34117773915  + 6287.00800325450 *T)
    sumL0  +=       0.548 * cos(5.68454458320  + 155427.54293624099 *T)
    sumL0  +=       0.547 * cos(1.03391472061  + 3646.35037735440 *T)
    sumL0  +=       0.428 * cos(4.69800981138  + 846.08283475120 *T)
    sumL0  +=       0.413 * cos(6.02520699406  + 6279.48542133960 *T)
    sumL0  +=       0.534 * cos(3.03030638223  + 66567.48586525429 *T)
    sumL0  +=       0.383 * cos(1.49056949125  + 19800.94595622480 *T)
    sumL0  +=       0.410 * cos(5.28319622279  + 18451.07854656599 *T)
    sumL0  +=       0.352 * cos(4.68891600359  + 4907.30205014560 *T)
    sumL0  +=       0.480 * cos(5.36572651091  + 348.92442044800 *T)
    sumL0  +=       0.344 * cos(5.89157452896  + 6546.15977336420 *T)
    sumL0  +=       0.340 * cos(0.37557426440  + 13119.72110282519 *T)
    sumL0  +=       0.434 * cos(4.98417785901  + 6702.56049386660 *T)
    sumL0  +=       0.332 * cos(2.68902519126  + 29296.61538957860 *T)
    sumL0  +=       0.448 * cos(2.16478480251  + 5905.70224207560 *T)
    sumL0  +=       0.344 * cos(2.06546633735  + 49.75702547180 *T)
    sumL0  +=       0.315 * cos(1.24023811803  + 4061.21921539440 *T)
    sumL0  +=       0.324 * cos(2.30897526929  + 5017.50837136500 *T)
    sumL0  +=       0.413 * cos(0.17171692962  + 6286.66627864320 *T)
    sumL0  +=       0.431 * cos(3.86601101393  + 12489.88562870720 *T)
    sumL0  +=       0.349 * cos(4.55372342974  + 4933.20844033260 *T)
    sumL0  +=       0.323 * cos(0.41971136084  + 10770.89325626180 *T)
    sumL0  +=       0.341 * cos(2.68612860807  + 11.04570026390 *T)
    sumL0  +=       0.316 * cos(3.52936906658  + 17782.73207278420 *T)
    sumL0  +=       0.315 * cos(5.63357264999  + 568.82187402740 *T)
    sumL0  +=       0.340 * cos(3.83571212349  + 10660.68693504240 *T)
    sumL0  +=       0.297 * cos(0.62691416712  + 20995.39296644940 *T)
    sumL0  +=       0.405 * cos(1.00085779471  + 16460.33352952499 *T)
    sumL0  +=       0.414 * cos(1.21998752076  + 51092.72605085480 *T)
    sumL0  +=       0.336 * cos(4.71465945226  + 6179.98307577280 *T)
    sumL0  +=       0.361 * cos(3.71227508354  + 28237.23345938940 *T)
    sumL0  +=       0.385 * cos(6.21925225757  + 24356.78078864160 *T)
    sumL0  +=       0.327 * cos(1.05606504715  + 11919.14086666800 *T)
    sumL0  +=       0.327 * cos(6.14222420989  + 6254.62666252360 *T)
    sumL0  +=       0.268 * cos(2.47224339737  + 664.75604513000 *T)
    sumL0  +=       0.269 * cos(1.86207884109  + 23141.55838292460 *T)
    sumL0  +=       0.345 * cos(0.93461290184  + 6058.73105428950 *T)
    sumL0  +=       0.296 * cos(4.51687557180  + 6418.14093002680 *T)
    sumL0  +=       0.353 * cos(4.50033653082  + 36949.23080842420 *T)
    sumL0  +=       0.260 * cos(4.04963546305  + 6525.80445396540 *T)
    sumL0  +=       0.298 * cos(2.20046722622  + 156137.47598479927 *T)
    sumL0  +=       0.253 * cos(3.49900838384  + 29864.33402730900 *T)
    sumL0  +=       0.254 * cos(2.44901693835  + 5331.35744374080 *T)
    sumL0  +=       0.296 * cos(0.84347588787  + 5729.50644714900 *T)
    sumL0  +=       0.298 * cos(1.29194706125  + 22805.73556599360 *T)
    sumL0  +=       0.241 * cos(2.00721280805  + 16737.57723659660 *T)
    sumL0  +=       0.311 * cos(1.23668016334  + 6281.59137728310 *T)
    sumL0  +=       0.240 * cos(2.51650377121  + 6245.04817735560 *T)
    sumL0  +=       0.332 * cos(3.55576945724  + 7668.63742494250 *T)
    sumL0  +=       0.264 * cos(4.44052061202  + 12964.30070339100 *T)
    sumL0  +=       0.257 * cos(1.79654471948  + 11080.17157891760 *T)
    sumL0  +=       0.260 * cos(3.33077598420  + 5888.44996493220 *T)
    sumL0  +=       0.285 * cos(0.30886361430  + 11823.16163945020 *T)
    sumL0  +=       0.290 * cos(5.70141882483  + 77.67377042800 *T)
    sumL0  +=       0.255 * cos(4.00939664440  + 5881.40372823420 *T)
    sumL0  +=       0.253 * cos(4.73318493678  + 16723.35014259500 *T)
    sumL0  +=       0.228 * cos(0.95333661324  + 5540.08578945880 *T)
    sumL0  +=       0.319 * cos(1.38633229189  + 163096.18036118348 *T)
    sumL0  +=       0.224 * cos(1.65156322696  + 10027.90319572920 *T)
    sumL0  +=       0.226 * cos(0.34106460604  + 17796.95916678580 *T)
    sumL0  +=       0.236 * cos(4.19817431922  + 19.66976089979 *T)
    sumL0  +=       0.280 * cos(4.14080268970  + 12539.85338018300 *T)
    sumL0  +=       0.275 * cos(5.50306930248  + 32.53255079140 *T)
    sumL0  +=       0.223 * cos(5.23334210294  + 56.89837493560 *T)
    sumL0  +=       0.217 * cos(6.08587881787  + 6805.65326808520 *T)
    sumL0  +=       0.280 * cos(4.52472044653  + 6016.46880826960 *T)
    sumL0  +=       0.227 * cos(5.06509843737  + 6277.55292568400 *T)
    sumL0  +=       0.226 * cos(5.17755154305  + 11720.06886523160 *T)
    sumL0  +=       0.245 * cos(3.96486270306  + 22.77520145080 *T)
    sumL0  +=       0.220 * cos(4.72078081970  + 6.62855890001 *T)
    sumL0  +=       0.207 * cos(5.71701403951  + 41.55079098480 *T)
    sumL0  +=       0.204 * cos(3.91227411250  + 2699.73481931760 *T)
    sumL0  +=       0.209 * cos(0.86881969011  + 6321.10352262720 *T)
    sumL0  +=       0.200 * cos(2.11984445273  + 4274.51831083240 *T)
    sumL0  +=       0.200 * cos(5.39839888163  + 6019.99192661860 *T)
    sumL0  +=       0.209 * cos(5.67606291663  + 11293.47067435560 *T)
    sumL0  +=       0.252 * cos(1.64965729351  + 9380.95967271720 *T)
    sumL0  +=       0.275 * cos(5.04826903506  + 73.29712585900 *T)
    sumL0  +=       0.208 * cos(1.88207277133  + 11300.58422135640 *T)
    sumL0  +=       0.272 * cos(0.74640926842  + 1975.49254585600 *T)
    sumL0  +=       0.199 * cos(3.30836672397  + 22743.40937951640 *T)
    sumL0  +=       0.269 * cos(4.48560812155  + 64471.99124174489 *T)
    sumL0  +=       0.192 * cos(2.17464236325  + 5863.59120611620 *T)
    sumL0  +=       0.228 * cos(5.85373115869  + 128.01884333740 *T)
    sumL0  +=       0.261 * cos(2.64321183295  + 55022.93574707440 *T)
    sumL0  +=       0.220 * cos(5.75012110079  + 29.42950853600 *T)
    sumL0  +=       0.187 * cos(4.03230554718  + 467.96499035440 *T)
    sumL0  +=       0.200 * cos(5.60556112058  + 1066.49547719000 *T)
    sumL0  +=       0.231 * cos(1.09802712785  + 12341.80690428090 *T)
    sumL0  +=       0.199 * cos(0.29500625200  + 149.56319713460 *T)
    sumL0  +=       0.249 * cos(5.10473210814  + 7875.67186362420 *T)
    sumL0  +=       0.208 * cos(0.93013835019  + 14919.01785375460 *T)
    sumL0  +=       0.179 * cos(0.87104393079  + 12721.57209941700 *T)
    sumL0  +=       0.203 * cos(1.56920753653  + 28286.99048486120 *T)
    sumL0  +=       0.179 * cos(2.47036386443  + 16062.18452611680 *T)
    sumL0  +=       0.198 * cos(3.54061588502  + 30.91412563500 *T)
    sumL0  +=       0.171 * cos(3.45356518113  + 5327.47610838280 *T)
    sumL0  +=       0.183 * cos(0.72325421604  + 6272.03014972750 *T)
    sumL0  +=       0.216 * cos(2.97174580686  + 19402.79695281660 *T)
    sumL0  +=       0.168 * cos(2.51550550242  + 23937.85638974100 *T)
    sumL0  +=       0.195 * cos(0.09045393425  + 156.40072050240 *T)
    sumL0  +=       0.179 * cos(4.49471798090  + 31415.37924995700 *T)
    sumL0  +=       0.216 * cos(0.42177594328  + 23539.70738633280 *T)
    sumL0  +=       0.189 * cos(0.37542530191  + 9814.60410029120 *T)
    sumL0  +=       0.218 * cos(2.36835880025  + 16627.37091537720 *T)
    sumL0  +=       0.166 * cos(4.23182968446  + 16840.67001081519 *T)
    sumL0  +=       0.200 * cos(2.02153258098  + 16097.67995028260 *T)
    sumL0  +=       0.169 * cos(0.91318727000  + 95.97922721780 *T)
    sumL0  +=       0.211 * cos(5.73370637657  + 151.89728108520 *T)
    sumL0  +=       0.204 * cos(0.42643085174  + 515.46387109300 *T)
    sumL0  +=       0.212 * cos(3.00233538977  + 12043.57428188900 *T)
    sumL0  +=       0.192 * cos(5.46153589821  + 6379.05507720920 *T)
    sumL0  +=       0.165 * cos(1.38698167064  + 4171.42553661380 *T)
    sumL0  +=       0.160 * cos(6.23798383332  + 202.25339517410 *T)
    sumL0  +=       0.215 * cos(0.20889073407  + 5621.84292321040 *T)
    sumL0  +=       0.181 * cos(4.12439203622  + 13341.67431130680 *T)
    sumL0  +=       0.153 * cos(1.24460848836  + 29826.30635467320 *T)
    sumL0  +=       0.150 * cos(3.12999753018  + 799.82112516540 *T)
    sumL0  +=       0.175 * cos(4.55671604437  + 239424.39025435288 *T)
    sumL0  +=       0.192 * cos(1.33928820063  + 394.62588505920 *T)
    sumL0  +=       0.149 * cos(2.65697593276  + 21.33564046700 *T)
    sumL0  +=       0.146 * cos(5.58021191726  + 412.37109687440 *T)
    sumL0  +=       0.156 * cos(3.75650175503  + 12323.42309600880 *T)
    sumL0  +=       0.143 * cos(3.75708566606  + 58864.54391814630 *T)
    sumL0  +=       0.143 * cos(3.28248547724  + 29.82143814880 *T)
    sumL0  +=       0.144 * cos(1.07862546598  + 1265.56747862640 *T)
    sumL0  +=       0.148 * cos(0.23389236655  + 10021.83728009940 *T)
    sumL0  +=       0.193 * cos(5.92751083086  + 40879.44050464380 *T)
    sumL0  +=       0.140 * cos(4.97612440269  + 158.94351778320 *T)
    sumL0  +=       0.148 * cos(2.61640453469  + 17157.06188047180 *T)
    sumL0  +=       0.141 * cos(3.66871308723  + 26084.02180621620 *T)
    sumL0  +=       0.147 * cos(5.09968173403  + 661.23292678100 *T)
    sumL0  +=       0.146 * cos(4.96885605695  + 57375.80190084620 *T)
    sumL0  +=       0.142 * cos(0.78678347839  + 12779.45079542080 *T)
    sumL0  +=       0.134 * cos(4.79432636012  + 111.18664228760 *T)
    sumL0  +=       0.140 * cos(1.27748013377  + 107.66352393860 *T)
    sumL0  +=       0.169 * cos(2.74893543762  + 26735.94526221320 *T)
    sumL0  +=       0.165 * cos(3.95288000638  + 6357.85744855870 *T)
    sumL0  +=       0.183 * cos(5.43418358741  + 369.69981594040 *T)
    sumL0  +=       0.134 * cos(3.09132862833  + 17.81252211800 *T)
    sumL0  +=       0.132 * cos(3.05633896779  + 22490.96212149340 *T)
    sumL0  +=       0.134 * cos(4.09472795832  + 6599.46771964800 *T)
    sumL0  +=       0.181 * cos(4.22950689891  + 966.97087743560 *T)
    sumL0  +=       0.152 * cos(5.28885894415  + 12669.24447420140 *T)
    sumL0  +=       0.150 * cos(5.86819430908  + 97238.62754448748 *T)
    sumL0  +=       0.142 * cos(5.87266532526  + 22476.73502749179 *T)
    sumL0  +=       0.145 * cos(5.07330784304  + 87.30820453981 *T)
    sumL0  +=       0.133 * cos(5.65471067133  + 31.97230581680 *T)
    sumL0  +=       0.124 * cos(2.83326217072  + 12566.21901028560 *T)
    sumL0  +=       0.135 * cos(3.12861731644  + 32217.20018108080 *T)
    sumL0  +=       0.137 * cos(0.86487461904  + 9924.81042151060 *T)
    sumL0  +=       0.172 * cos(1.98369595114  + 174242.46596404971 *T)
    sumL0  +=       0.170 * cos(4.41115280254  + 327574.51427678124 *T)
    sumL0  +=       0.151 * cos(0.46542099527  + 39609.65458316560 *T)
    sumL0  +=       0.148 * cos(2.13439571118  + 491.66329245880 *T)
    sumL0  +=       0.153 * cos(3.78801830344  + 17363.24742890899 *T)
    sumL0  +=       0.165 * cos(5.31654110459  + 16943.76278503380 *T)
    sumL0  +=       0.165 * cos(4.06747587817  + 58953.14544329400 *T)
    sumL0  +=       0.118 * cos(0.63846333239  + 6.06591562980 *T)
    sumL0  +=       0.159 * cos(0.86086959274  + 221995.02880149524 *T)
    sumL0  +=       0.119 * cos(5.96432932413  + 1385.89527633620 *T)
    sumL0  +=       0.114 * cos(5.16516114595  + 25685.87280280800 *T)
    sumL0  +=       0.112 * cos(3.39403722178  + 21393.54196985760 *T)
    sumL0  +=       0.112 * cos(4.92889233335  + 56.80326216980 *T)
    sumL0  +=       0.119 * cos(2.40637635942  + 18635.92845453620 *T)
    sumL0  +=       0.115 * cos(0.23374479051  + 418.92439890060 *T)
    sumL0  +=       0.122 * cos(0.93575234049  + 24492.40611365159 *T)
    sumL0  +=       0.115 * cos(4.58880032176  + 26709.64694241340 *T)
    sumL0  +=       0.130 * cos(4.85539251000  + 22345.26037610820 *T)
    sumL0  +=       0.140 * cos(1.09413073202  + 44809.65020086340 *T)
    sumL0  +=       0.112 * cos(6.05401806281  + 433.71173787680 *T)
    sumL0  +=       0.104 * cos(1.54931540602  + 127.95153303460 *T)
    sumL0  +=       0.105 * cos(4.82620858888  + 33794.54372352860 *T)
    sumL0  +=       0.102 * cos(4.12448497391  + 15664.03552270859 *T)
    sumL0  +=       0.107 * cos(4.67919356465  + 77690.75950573849 *T)
    sumL0  +=       0.118 * cos(4.52320170120  + 19004.64794940840 *T)
    sumL0  +=       0.107 * cos(5.71774478555  + 77736.78343050249 *T)
    sumL0  +=       0.143 * cos(1.81201813018  + 4214.06901508480 *T)
    sumL0  +=       0.125 * cos(1.14419195615  + 625.67019231240 *T)
    sumL0  +=       0.124 * cos(3.27736514057  + 12566.08438968000 *T)
    sumL0  +=       0.110 * cos(1.08682570828  + 2787.04302385740 *T)
    sumL0  +=       0.105 * cos(1.78318141871  + 18139.29450141590 *T)
    sumL0  +=       0.102 * cos(4.75119578149  + 12242.64628332540 *T)
    sumL0  +=       0.137 * cos(1.43510636754  + 86464.61331683119 *T)
    sumL0  +=       0.101 * cos(4.91289409429  + 401.67212175720 *T)
    sumL0  +=       0.129 * cos(1.23567904485  + 12029.34718788740 *T)
    sumL0  +=       0.138 * cos(2.45654707999  + 7576.56007357400 *T)
    sumL0  +=       0.103 * cos(0.40004073416  + 90279.92316810328 *T)
    sumL0  +=       0.108 * cos(0.98989774940  + 5636.06501667660 *T)
    sumL0  +=       0.117 * cos(5.17362872063  + 34520.30930938080 *T)
    sumL0  +=       0.100 * cos(3.95534628189  + 5547.19933645960 *T)
    sumL0  +=       0.098 * cos(1.28118280598  + 21548.96236929180 *T)
    sumL0  +=       0.097 * cos(3.34717130592  + 16310.97904572060 *T)
    sumL0  +=       0.098 * cos(4.37041908717  + 34513.26307268280 *T)
    sumL0  +=       0.125 * cos(2.72164432960  + 24065.80792277559 *T)
    sumL0  +=       0.102 * cos(0.66938025772  + 10239.58386601080 *T)
    sumL0  +=       0.119 * cos(1.21689479331  + 1478.86657406440 *T)
    sumL0  +=       0.094 * cos(1.99595224256  + 13362.44970679920 *T)
    sumL0  +=       0.094 * cos(4.30965982872  + 26880.31981303260 *T)
    sumL0  +=       0.095 * cos(2.89807657534  + 34911.41207609100 *T)
    sumL0  +=       0.106 * cos(1.00156653590  + 16522.65971600220 *T)
    sumL0  +=       0.097 * cos(0.89642320201  + 71980.63357473118 *T)
    sumL0  +=       0.116 * cos(4.19967201116  + 206.70073729660 *T)
    sumL0  +=       0.099 * cos(1.37437847718  + 1039.02661079040 *T)
    sumL0  +=       0.126 * cos(3.21642544972  + 305281.94307104882 *T)
    sumL0  +=       0.094 * cos(0.68997876060  + 7834.12107263940 *T)
    sumL0  +=       0.094 * cos(5.58132218606  + 3104.93005942380 *T)
    sumL0  +=       0.095 * cos(3.03823741110  + 8982.81066930900 *T)
    sumL0  +=       0.108 * cos(0.52696637156  + 276.74577186440 *T)
    sumL0  +=       0.124 * cos(3.43899862683  + 172146.97134054027 *T)
    sumL0  +=       0.102 * cos(1.04031728553  + 95143.13292097810 *T)
    sumL0  +=       0.104 * cos(3.39218586218  + 290.97286586600 *T)
    sumL0  +=       0.110 * cos(3.68205877433  + 22380.75580027400 *T)
    sumL0  +=       0.117 * cos(0.78475956902  + 83286.91426955358 *T)
    sumL0  +=       0.083 * cos(0.18241793425  + 15141.39079431200 *T)
    sumL0  +=       0.089 * cos(4.45371820659  + 792.77488846740 *T)
    sumL0  +=       0.082 * cos(4.80703651241  + 6819.88036208680 *T)
    sumL0  +=       0.087 * cos(3.43122851097  + 27707.54249429480 *T)
    sumL0  +=       0.101 * cos(5.32081603011  + 2301.58581590939 *T)
    sumL0  +=       0.082 * cos(0.87060089842  + 10241.20229116720 *T)
    sumL0  +=       0.086 * cos(4.61919461931  + 36147.40987730040 *T)
    sumL0  +=       0.095 * cos(2.87032884659  + 23020.65308658799 *T)
    sumL0  +=       0.088 * cos(3.21133165690  + 33326.57873317420 *T)
    sumL0  +=       0.080 * cos(1.84900424847  + 21424.46664430340 *T)
    sumL0  +=       0.101 * cos(4.18796434479  + 30666.15495843280 *T)
    sumL0  +=       0.107 * cos(5.77864921649  + 34115.11406927460 *T)
    sumL0  +=       0.104 * cos(1.08739495962  + 6288.59877429880 *T)
    sumL0  +=       0.110 * cos(3.32898859416  + 72140.62866668739 *T)
    sumL0  +=       0.087 * cos(4.40657711727  + 142.17862703620 *T)
    sumL0  +=       0.109 * cos(1.94546030825  + 24279.10701821359 *T)
    sumL0  +=       0.087 * cos(4.32472045435  + 742.99006053260 *T)
    sumL0  +=       0.107 * cos(4.91580912547  + 277.03499374140 *T)
    sumL0  +=       0.088 * cos(2.10180220766  + 26482.17080962440 *T)
    sumL0  +=       0.086 * cos(4.01887374432  + 12491.37010141550 *T)
    sumL0  +=       0.106 * cos(5.49092372854  + 62883.35513951360 *T)
    sumL0  +=       0.080 * cos(6.19781316983  + 6709.67404086740 *T)
    sumL0  +=       0.088 * cos(2.09872810657  + 238004.52415723627 *T)
    sumL0  +=       0.083 * cos(4.90662164029  + 51.28033786241 *T)
    sumL0  +=       0.095 * cos(4.13387406591  + 18216.44381066100 *T)
    sumL0  +=       0.078 * cos(6.06949391680  + 148434.53403769128 *T)
    sumL0  +=       0.079 * cos(3.03048221644  + 838.96928775040 *T)
    sumL0  +=       0.074 * cos(5.49813051211  + 29026.48522950779 *T)
    sumL0  +=       0.073 * cos(3.05008665738  + 567.71863773040 *T)
    sumL0  +=       0.084 * cos(0.46604373274  + 45.14121963660 *T)
    sumL0  +=       0.093 * cos(2.52267536308  + 48739.85989708300 *T)
    sumL0  +=       0.076 * cos(1.76418124905  + 41654.96311596780 *T)
    sumL0  +=       0.067 * cos(5.77851227793  + 6311.52503745920 *T)
    sumL0  +=       0.062 * cos(3.32967880172  + 15508.61512327440 *T)
    sumL0  +=       0.079 * cos(5.59773841328  + 71960.38658322369 *T)
    sumL0  +=       0.057 * cos(3.90629505268  + 5999.21653112620 *T)
    sumL0  +=       0.061 * cos(0.05695043232  + 7856.89627409019 *T)
    sumL0  +=       0.061 * cos(5.63297958433  + 7863.94251078820 *T)
    sumL0  +=       0.065 * cos(3.72178394016  + 12573.26524698360 *T)
    sumL0  +=       0.057 * cos(4.18217219541  + 26087.90314157420 *T)
    sumL0  +=       0.066 * cos(3.92262333487  + 69853.35207568129 *T)
    sumL0  +=       0.053 * cos(5.51119362045  + 77710.24834977149 *T)
    sumL0  +=       0.053 * cos(4.88573986961  + 77717.29458646949 *T)
    sumL0  +=       0.062 * cos(2.88876342225  + 9411.46461508720 *T)
    sumL0  +=       0.051 * cos(1.12657183874  + 82576.98122099529 *T)
    sumL0  +=       0.045 * cos(2.95671076719  + 24602.61243487099 *T)
    sumL0  +=       0.040 * cos(5.55145719241  + 12565.17137891460 *T)
    sumL0  +=       0.039 * cos(1.20838190039  + 18842.11400297339 *T)
    sumL0  +=       0.045 * cos(3.18590558749  + 45585.17281218740 *T)
    sumL0  +=       0.049 * cos(2.44790934886  + 13613.80427733600 *T)

    # EARTH Longitude
    sumL1  = 0
    sumL1  +=  628331966747.491 * cos(0.00000000000  + 0.00000000000 *T)
    sumL1  +=  206058.863 * cos(2.67823455584  + 6283.07584999140 *T)
    sumL1  +=    4303.430 * cos(2.63512650414  + 12566.15169998280 *T)
    sumL1  +=     425.264 * cos(1.59046980729  + 3.52311834900 *T)
    sumL1  +=     108.977 * cos(2.96618001993  + 1577.34354244780 *T)
    sumL1  +=      93.478 * cos(2.59212835365  + 18849.22754997420 *T)
    sumL1  +=     119.261 * cos(5.79557487799  + 26.29831979980 *T)
    sumL1  +=      72.122 * cos(1.13846158196  + 529.69096509460 *T)
    sumL1  +=      67.768 * cos(1.87472304791  + 398.14900340820 *T)
    sumL1  +=      67.327 * cos(4.40918235168  + 5507.55323866740 *T)
    sumL1  +=      59.027 * cos(2.88797038460  + 5223.69391980220 *T)
    sumL1  +=      55.976 * cos(2.17471680261  + 155.42039943420 *T)
    sumL1  +=      45.407 * cos(0.39803079805  + 796.29800681640 *T)
    sumL1  +=      36.369 * cos(0.46624739835  + 775.52261132400 *T)
    sumL1  +=      28.958 * cos(2.64707383882  + 7.11354700080 *T)
    sumL1  +=      19.097 * cos(1.84628332577  + 5486.77784317500 *T)
    sumL1  +=      20.844 * cos(5.34138275149  + 0.98032106820 *T)
    sumL1  +=      18.508 * cos(4.96855124577  + 213.29909543800 *T)
    sumL1  +=      16.233 * cos(0.03216483047  + 2544.31441988340 *T)
    sumL1  +=      17.293 * cos(2.99116864949  + 6275.96230299060 *T)
    sumL1  +=      15.832 * cos(1.43049285325  + 2146.16541647520 *T)
    sumL1  +=      14.615 * cos(1.20532366323  + 10977.07880469900 *T)
    sumL1  +=      11.877 * cos(3.25804815607  + 5088.62883976680 *T)
    sumL1  +=      11.514 * cos(2.07502418155  + 4694.00295470760 *T)
    sumL1  +=       9.721 * cos(4.23925472239  + 1349.86740965880 *T)
    sumL1  +=       9.969 * cos(1.30262991097  + 6286.59896834040 *T)
    sumL1  +=       9.452 * cos(2.69957062864  + 242.72860397400 *T)
    sumL1  +=      12.461 * cos(2.83432285512  + 1748.01641306700 *T)
    sumL1  +=      11.808 * cos(5.27379790480  + 1194.44701022460 *T)
    sumL1  +=       8.577 * cos(5.64475868067  + 951.71840625060 *T)
    sumL1  +=      10.641 * cos(0.76614199202  + 553.56940284240 *T)
    sumL1  +=       7.576 * cos(5.30062664886  + 2352.86615377180 *T)
    sumL1  +=       5.834 * cos(1.76649917904  + 1059.38193018920 *T)
    sumL1  +=       6.385 * cos(2.65033984967  + 9437.76293488700 *T)
    sumL1  +=       5.223 * cos(5.66135767624  + 71430.69561812909 *T)
    sumL1  +=       5.305 * cos(0.90857521574  + 3154.68708489560 *T)
    sumL1  +=       6.101 * cos(4.66632584188  + 4690.47983635860 *T)
    sumL1  +=       4.330 * cos(0.24102555403  + 6812.76681508600 *T)
    sumL1  +=       5.041 * cos(1.42490103709  + 6438.49624942560 *T)
    sumL1  +=       4.259 * cos(0.77355900599  + 10447.38783960440 *T)
    sumL1  +=       5.198 * cos(1.85353197345  + 801.82093112380 *T)
    sumL1  +=       3.744 * cos(2.00119516488  + 8031.09226305840 *T)
    sumL1  +=       3.558 * cos(2.42901552681  + 14143.49524243060 *T)
    sumL1  +=       3.372 * cos(3.86210700128  + 1592.59601363280 *T)
    sumL1  +=       3.374 * cos(0.88776219727  + 12036.46073488820 *T)
    sumL1  +=       3.175 * cos(3.18785710594  + 4705.73230754360 *T)
    sumL1  +=       3.221 * cos(0.61599835472  + 8429.24126646660 *T)
    sumL1  +=       4.132 * cos(5.23992859705  + 7084.89678111520 *T)
    sumL1  +=       2.970 * cos(6.07026318493  + 4292.33083295040 *T)
    sumL1  +=       2.900 * cos(2.32464208411  + 20.35531939880 *T)
    sumL1  +=       3.504 * cos(4.79975694359  + 6279.55273164240 *T)
    sumL1  +=       2.950 * cos(1.43108874817  + 5746.27133789600 *T)
    sumL1  +=       2.697 * cos(4.80368225199  + 7234.79425624200 *T)
    sumL1  +=       2.531 * cos(6.22290682655  + 6836.64525283380 *T)
    sumL1  +=       2.745 * cos(0.93466065396  + 5760.49843189760 *T)
    sumL1  +=       3.250 * cos(3.39954640038  + 7632.94325965020 *T)
    sumL1  +=       2.277 * cos(5.00277837672  + 17789.84561978500 *T)
    sumL1  +=       2.075 * cos(3.95534978634  + 10213.28554621100 *T)
    sumL1  +=       2.061 * cos(2.22411683077  + 5856.47765911540 *T)
    sumL1  +=       2.252 * cos(5.67166499885  + 11499.65622279280 *T)
    sumL1  +=       2.148 * cos(5.20184578235  + 11513.88331679440 *T)
    sumL1  +=       1.886 * cos(0.53198320577  + 3340.61242669980 *T)
    sumL1  +=       1.875 * cos(4.73511970207  + 83996.84731811189 *T)
    sumL1  +=       2.060 * cos(2.54987293999  + 25132.30339996560 *T)
    sumL1  +=       1.794 * cos(1.47435409831  + 4164.31198961300 *T)
    sumL1  +=       1.778 * cos(3.02473091781  + 5.52292430740 *T)
    sumL1  +=       2.029 * cos(0.90960209983  + 6256.77753019160 *T)
    sumL1  +=       2.075 * cos(2.26767270157  + 522.57741809380 *T)
    sumL1  +=       1.772 * cos(3.02622802353  + 5753.38488489680 *T)
    sumL1  +=       1.569 * cos(6.12410242782  + 5216.58037280140 *T)
    sumL1  +=       1.590 * cos(4.63713748247  + 3.28635741780 *T)
    sumL1  +=       1.542 * cos(4.20004448567  + 13367.97263110660 *T)
    sumL1  +=       1.427 * cos(1.19088061711  + 3894.18182954220 *T)
    sumL1  +=       1.375 * cos(3.09301252193  + 135.06508003540 *T)
    sumL1  +=       1.359 * cos(4.24532506641  + 426.59819087600 *T)
    sumL1  +=       1.340 * cos(5.76511818622  + 6040.34724601740 *T)
    sumL1  +=       1.284 * cos(3.08524663344  + 5643.17856367740 *T)
    sumL1  +=       1.250 * cos(3.07748157144  + 11926.25441366880 *T)
    sumL1  +=       1.551 * cos(3.07665451458  + 6681.22485339960 *T)
    sumL1  +=       1.268 * cos(2.09196018331  + 6290.18939699220 *T)
    sumL1  +=       1.144 * cos(3.24444699514  + 12168.00269657460 *T)
    sumL1  +=       1.248 * cos(3.44504937285  + 536.80451209540 *T)
    sumL1  +=       1.118 * cos(2.31829670425  + 16730.46368959580 *T)
    sumL1  +=       1.105 * cos(5.31966001019  + 23.87843774780 *T)
    sumL1  +=       1.051 * cos(3.75015946014  + 7860.41939243920 *T)
    sumL1  +=       1.025 * cos(2.44688534235  + 1990.74501704100 *T)
    sumL1  +=       0.962 * cos(0.81771017882  + 3.88133535800 *T)
    sumL1  +=       0.910 * cos(0.41727865299  + 7079.37385680780 *T)
    sumL1  +=       0.883 * cos(5.16833917651  + 11790.62908865880 *T)
    sumL1  +=       0.957 * cos(4.07673573735  + 6127.65545055720 *T)
    sumL1  +=       1.110 * cos(3.90096793825  + 11506.76976979360 *T)
    sumL1  +=       0.802 * cos(3.88778875582  + 10973.55568635000 *T)
    sumL1  +=       0.780 * cos(2.39934293755  + 1589.07289528380 *T)
    sumL1  +=       0.758 * cos(1.30034364248  + 103.09277421860 *T)
    sumL1  +=       0.749 * cos(4.96275803300  + 6496.37494542940 *T)
    sumL1  +=       0.765 * cos(3.36312388424  + 36.02786667740 *T)
    sumL1  +=       0.915 * cos(5.41543742089  + 206.18554843720 *T)
    sumL1  +=       0.776 * cos(2.57589093871  + 11371.70468975820 *T)
    sumL1  +=       0.772 * cos(3.98369209464  + 955.59974160860 *T)
    sumL1  +=       0.749 * cos(5.17890001805  + 10969.96525769820 *T)
    sumL1  +=       0.806 * cos(0.34218864254  + 9917.69687450980 *T)
    sumL1  +=       0.728 * cos(5.20962563787  + 38.02767263580 *T)
    sumL1  +=       0.685 * cos(2.77592961854  + 20.77539549240 *T)
    sumL1  +=       0.636 * cos(4.28242193632  + 28.44918746780 *T)
    sumL1  +=       0.608 * cos(5.63278508906  + 10984.19235169980 *T)
    sumL1  +=       0.704 * cos(5.60738823665  + 3738.76143010800 *T)
    sumL1  +=       0.685 * cos(0.38876148682  + 15.25247118500 *T)
    sumL1  +=       0.601 * cos(0.73489602442  + 419.48464387520 *T)
    sumL1  +=       0.716 * cos(2.65279791438  + 6309.37416979120 *T)
    sumL1  +=       0.584 * cos(5.54502568227  + 17298.18232732620 *T)
    sumL1  +=       0.650 * cos(1.13379656406  + 7058.59846131540 *T)
    sumL1  +=       0.688 * cos(2.59683891779  + 3496.03282613400 *T)
    sumL1  +=       0.485 * cos(0.44467180946  + 12352.85260454480 *T)
    sumL1  +=       0.528 * cos(2.74936967681  + 3930.20969621960 *T)
    sumL1  +=       0.597 * cos(5.27668281777  + 10575.40668294180 *T)
    sumL1  +=       0.583 * cos(3.18929067810  + 4732.03062734340 *T)
    sumL1  +=       0.526 * cos(5.01697321546  + 5884.92684658320 *T)
    sumL1  +=       0.540 * cos(1.29175137075  + 640.87760738220 *T)
    sumL1  +=       0.473 * cos(5.49953306970  + 5230.80746680300 *T)
    sumL1  +=       0.406 * cos(5.21248452189  + 220.41264243880 *T)
    sumL1  +=       0.395 * cos(1.87474483222  + 16200.77272450120 *T)
    sumL1  +=       0.370 * cos(3.84921354713  + 18073.70493865020 *T)
    sumL1  +=       0.367 * cos(0.88533542778  + 6283.14316029419 *T)
    sumL1  +=       0.379 * cos(0.37983009325  + 10177.25767953360 *T)
    sumL1  +=       0.356 * cos(3.84145204913  + 11712.95531823080 *T)
    sumL1  +=       0.374 * cos(5.01577520608  + 7.04623669800 *T)
    sumL1  +=       0.381 * cos(4.30250406634  + 6062.66320755260 *T)
    sumL1  +=       0.471 * cos(0.86381834647  + 6069.77675455340 *T)
    sumL1  +=       0.367 * cos(1.32943839763  + 6283.00853968860 *T)
    sumL1  +=       0.460 * cos(5.19667219575  + 6284.05617105960 *T)
    sumL1  +=       0.333 * cos(5.54256205741  + 4686.88940770680 *T)
    sumL1  +=       0.341 * cos(4.36522989934  + 7238.67559160000 *T)
    sumL1  +=       0.336 * cos(4.00205876835  + 3097.88382272579 *T)
    sumL1  +=       0.359 * cos(6.22679790284  + 245.83164622940 *T)
    sumL1  +=       0.307 * cos(2.35299010924  + 170.67287061920 *T)
    sumL1  +=       0.343 * cos(3.77164927143  + 6076.89030155420 *T)
    sumL1  +=       0.296 * cos(5.44152227481  + 17260.15465469040 *T)
    sumL1  +=       0.328 * cos(0.13837875384  + 11015.10647733480 *T)
    sumL1  +=       0.268 * cos(1.13904550630  + 12569.67481833180 *T)
    sumL1  +=       0.263 * cos(0.00538633678  + 4136.91043351620 *T)
    sumL1  +=       0.282 * cos(5.04399837480  + 7477.52286021600 *T)
    sumL1  +=       0.288 * cos(3.13401177517  + 12559.03815298200 *T)
    sumL1  +=       0.259 * cos(0.93882269387  + 5642.19824260920 *T)
    sumL1  +=       0.292 * cos(1.98420020514  + 12132.43996210600 *T)
    sumL1  +=       0.247 * cos(3.84244798532  + 5429.87946823940 *T)
    sumL1  +=       0.245 * cos(5.70467521726  + 65147.61976813770 *T)
    sumL1  +=       0.241 * cos(0.99480969552  + 3634.62102451840 *T)
    sumL1  +=       0.246 * cos(3.06168069935  + 110.20632121940 *T)
    sumL1  +=       0.239 * cos(6.11855909114  + 11856.21865142450 *T)
    sumL1  +=       0.263 * cos(0.66348415419  + 21228.39202354580 *T)
    sumL1  +=       0.262 * cos(1.51070507866  + 12146.66705610760 *T)
    sumL1  +=       0.230 * cos(1.75927314884  + 9779.10867612540 *T)
    sumL1  +=       0.223 * cos(2.00967043606  + 6172.86952877200 *T)
    sumL1  +=       0.246 * cos(1.10411690865  + 6282.09552892320 *T)
    sumL1  +=       0.221 * cos(3.03945240854  + 8635.94200376320 *T)
    sumL1  +=       0.214 * cos(4.03840869663  + 14314.16811304980 *T)
    sumL1  +=       0.236 * cos(5.46915070580  + 13916.01910964160 *T)
    sumL1  +=       0.224 * cos(4.68408089456  + 24072.92146977640 *T)
    sumL1  +=       0.212 * cos(2.13695625494  + 5849.36411211460 *T)
    sumL1  +=       0.207 * cos(3.07724246401  + 11.72935283600 *T)
    sumL1  +=       0.207 * cos(6.10306282747  + 23543.23050468179 *T)
    sumL1  +=       0.266 * cos(1.00709566823  + 2388.89402044920 *T)
    sumL1  +=       0.217 * cos(6.27837036335  + 17267.26820169119 *T)
    sumL1  +=       0.204 * cos(2.34615348695  + 266.60704172180 *T)
    sumL1  +=       0.195 * cos(5.55015549753  + 6133.51265285680 *T)
    sumL1  +=       0.188 * cos(2.52667166175  + 6525.80445396540 *T)
    sumL1  +=       0.185 * cos(0.90960768344  + 18319.53658487960 *T)
    sumL1  +=       0.177 * cos(1.73429218289  + 154717.60988768269 *T)
    sumL1  +=       0.187 * cos(4.76483647432  + 4535.05943692440 *T)
    sumL1  +=       0.186 * cos(4.63080493407  + 10440.27429260360 *T)
    sumL1  +=       0.215 * cos(2.81255454560  + 7342.45778018060 *T)
    sumL1  +=       0.172 * cos(1.45551888559  + 9225.53927328300 *T)
    sumL1  +=       0.162 * cos(3.30661909388  + 639.89728631400 *T)
    sumL1  +=       0.168 * cos(2.17671416605  + 27.40155609680 *T)
    sumL1  +=       0.160 * cos(1.68164180475  + 15110.46611986620 *T)
    sumL1  +=       0.158 * cos(0.13519771874  + 13095.84266507740 *T)
    sumL1  +=       0.183 * cos(0.56281322071  + 13517.87010623340 *T)
    sumL1  +=       0.179 * cos(3.58450811616  + 87.30820453981 *T)
    sumL1  +=       0.152 * cos(2.84070476818  + 5650.29211067820 *T)
    sumL1  +=       0.182 * cos(0.44065530624  + 17253.04110768959 *T)
    sumL1  +=       0.160 * cos(5.95767264171  + 4701.11650170840 *T)
    sumL1  +=       0.142 * cos(1.46290137520  + 11087.28512591840 *T)
    sumL1  +=       0.142 * cos(2.04464036087  + 20426.57109242200 *T)
    sumL1  +=       0.131 * cos(5.40912137746  + 2699.73481931760 *T)
    sumL1  +=       0.144 * cos(2.07312090485  + 25158.60171976540 *T)
    sumL1  +=       0.147 * cos(6.15106982168  + 9623.68827669120 *T)
    sumL1  +=       0.141 * cos(5.55739979498  + 10454.50138660520 *T)
    sumL1  +=       0.135 * cos(0.06098110407  + 16723.35014259500 *T)
    sumL1  +=       0.124 * cos(5.81218025669  + 17256.63153634140 *T)
    sumL1  +=       0.124 * cos(2.36293551623  + 4933.20844033260 *T)
    sumL1  +=       0.126 * cos(3.47435905118  + 22483.84857449259 *T)
    sumL1  +=       0.159 * cos(5.63954754618  + 5729.50644714900 *T)
    sumL1  +=       0.123 * cos(3.92815963256  + 17996.03116822220 *T)
    sumL1  +=       0.148 * cos(3.02509280598  + 1551.04522264800 *T)
    sumL1  +=       0.120 * cos(5.91904349732  + 6206.80977871580 *T)
    sumL1  +=       0.134 * cos(3.11122937825  + 21954.15760939799 *T)
    sumL1  +=       0.119 * cos(5.52141123450  + 709.93304855830 *T)
    sumL1  +=       0.122 * cos(3.00813429479  + 19800.94595622480 *T)
    sumL1  +=       0.127 * cos(1.37618620001  + 14945.31617355440 *T)
    sumL1  +=       0.141 * cos(2.56889468729  + 1052.26838318840 *T)
    sumL1  +=       0.123 * cos(2.83671175442  + 11919.14086666800 *T)
    sumL1  +=       0.118 * cos(0.81934438215  + 5331.35744374080 *T)
    sumL1  +=       0.151 * cos(2.68731829165  + 11769.85369316640 *T)
    sumL1  +=       0.119 * cos(5.08835797638  + 5481.25491886760 *T)
    sumL1  +=       0.153 * cos(2.46021790779  + 11933.36796066960 *T)
    sumL1  +=       0.108 * cos(1.04936452145  + 11403.67699557500 *T)
    sumL1  +=       0.128 * cos(0.99794735107  + 8827.39026987480 *T)
    sumL1  +=       0.144 * cos(2.54869747042  + 227.47613278900 *T)
    sumL1  +=       0.150 * cos(4.50631437136  + 2379.16447357160 *T)
    sumL1  +=       0.107 * cos(1.79272017026  + 13119.72110282519 *T)
    sumL1  +=       0.107 * cos(4.43556814486  + 18422.62935909819 *T)
    sumL1  +=       0.109 * cos(0.29269062317  + 16737.57723659660 *T)
    sumL1  +=       0.141 * cos(3.18979826258  + 6262.30045449900 *T)
    sumL1  +=       0.122 * cos(4.23040027813  + 29.42950853600 *T)
    sumL1  +=       0.111 * cos(5.16954029551  + 17782.73207278420 *T)
    sumL1  +=       0.100 * cos(3.52213872761  + 18052.92954315780 *T)
    sumL1  +=       0.108 * cos(1.08514212991  + 16858.48253293320 *T)
    sumL1  +=       0.106 * cos(1.96085248410  + 74.78159856730 *T)
    sumL1  +=       0.110 * cos(2.30582372873  + 16460.33352952499 *T)
    sumL1  +=       0.097 * cos(3.50918940210  + 5333.90024102160 *T)
    sumL1  +=       0.099 * cos(3.56417337974  + 735.87651353180 *T)
    sumL1  +=       0.094 * cos(5.01857894228  + 3128.38876509580 *T)
    sumL1  +=       0.097 * cos(1.65579893894  + 533.21408344360 *T)
    sumL1  +=       0.092 * cos(0.89217162285  + 29296.61538957860 *T)
    sumL1  +=       0.123 * cos(3.16062050433  + 9380.95967271720 *T)
    sumL1  +=       0.102 * cos(1.20493500565  + 23020.65308658799 *T)
    sumL1  +=       0.088 * cos(2.21296088224  + 12721.57209941700 *T)
    sumL1  +=       0.089 * cos(1.54264720310  + 20199.09495963300 *T)
    sumL1  +=       0.113 * cos(4.83320707870  + 16496.36139620240 *T)
    sumL1  +=       0.121 * cos(6.19860353182  + 9388.00590941520 *T)
    sumL1  +=       0.089 * cos(4.08082274765  + 22805.73556599360 *T)
    sumL1  +=       0.098 * cos(1.09181832830  + 12043.57428188900 *T)
    sumL1  +=       0.086 * cos(1.13655027605  + 143571.32428481648 *T)
    sumL1  +=       0.088 * cos(5.96980472191  + 107.66352393860 *T)
    sumL1  +=       0.082 * cos(5.01340404594  + 22003.91463486980 *T)
    sumL1  +=       0.094 * cos(1.69615700473  + 23006.42599258639 *T)
    sumL1  +=       0.081 * cos(3.00657814365  + 2118.76386037840 *T)
    sumL1  +=       0.098 * cos(1.39215287161  + 8662.24032356300 *T)
    sumL1  +=       0.077 * cos(3.33555190840  + 15720.83878487840 *T)
    sumL1  +=       0.082 * cos(5.86880116464  + 2787.04302385740 *T)
    sumL1  +=       0.076 * cos(5.67183650604  + 14.22709400160 *T)
    sumL1  +=       0.081 * cos(6.16619455699  + 1039.02661079040 *T)
    sumL1  +=       0.076 * cos(3.21449884756  + 111.18664228760 *T)
    sumL1  +=       0.078 * cos(1.37531518377  + 21947.11137270000 *T)
    sumL1  +=       0.074 * cos(3.58814195051  + 11609.86254401220 *T)
    sumL1  +=       0.077 * cos(4.84846488388  + 22743.40937951640 *T)
    sumL1  +=       0.090 * cos(1.48869013606  + 15671.08175940660 *T)
    sumL1  +=       0.082 * cos(3.48618399109  + 29088.81141598500 *T)
    sumL1  +=       0.069 * cos(3.55746476593  + 4590.91018048900 *T)
    sumL1  +=       0.069 * cos(1.93625656075  + 135.62532501000 *T)
    sumL1  +=       0.070 * cos(2.66548322237  + 18875.52586977400 *T)
    sumL1  +=       0.069 * cos(5.41478093731  + 26735.94526221320 *T)
    sumL1  +=       0.079 * cos(5.15154513662  + 12323.42309600880 *T)
    sumL1  +=       0.094 * cos(3.62899392448  + 77713.77146812050 *T)
    sumL1  +=       0.078 * cos(4.17011182047  + 1066.49547719000 *T)
    sumL1  +=       0.071 * cos(3.89435637865  + 22779.43724619380 *T)
    sumL1  +=       0.063 * cos(4.53968787714  + 8982.81066930900 *T)
    sumL1  +=       0.069 * cos(0.96028230548  + 14919.01785375460 *T)
    sumL1  +=       0.076 * cos(3.29092216589  + 2942.46342329160 *T)
    sumL1  +=       0.063 * cos(4.09167842893  + 16062.18452611680 *T)
    sumL1  +=       0.065 * cos(3.34580407184  + 51.28033786241 *T)
    sumL1  +=       0.065 * cos(5.75757544877  + 52670.06959330260 *T)
    sumL1  +=       0.068 * cos(5.75884067555  + 21424.46664430340 *T)
    sumL1  +=       0.057 * cos(5.45122399850  + 12592.45001978260 *T)
    sumL1  +=       0.057 * cos(5.25043362558  + 20995.39296644940 *T)
    sumL1  +=       0.073 * cos(0.53299090807  + 2301.58581590939 *T)
    sumL1  +=       0.070 * cos(4.31243357502  + 19402.79695281660 *T)
    sumL1  +=       0.067 * cos(2.53852336668  + 377.37360791580 *T)
    sumL1  +=       0.056 * cos(3.20816844695  + 24889.57479599160 *T)
    sumL1  +=       0.053 * cos(3.17816599142  + 18451.07854656599 *T)
    sumL1  +=       0.053 * cos(3.61529270216  + 77.67377042800 *T)
    sumL1  +=       0.053 * cos(0.45467549335  + 30666.15495843280 *T)
    sumL1  +=       0.061 * cos(0.14807288453  + 23013.53953958720 *T)
    sumL1  +=       0.051 * cos(3.32803972907  + 56.89837493560 *T)
    sumL1  +=       0.052 * cos(3.41177624177  + 23141.55838292460 *T)
    sumL1  +=       0.058 * cos(3.13638677202  + 309.27832265580 *T)
    sumL1  +=       0.070 * cos(2.50592323465  + 31415.37924995700 *T)
    sumL1  +=       0.052 * cos(5.10673376738  + 17796.95916678580 *T)
    sumL1  +=       0.067 * cos(6.27917920454  + 22345.26037610820 *T)
    sumL1  +=       0.050 * cos(0.42577644151  + 25685.87280280800 *T)
    sumL1  +=       0.048 * cos(0.70204553333  + 1162.47470440780 *T)
    sumL1  +=       0.066 * cos(3.64350022359  + 15265.88651930040 *T)
    sumL1  +=       0.050 * cos(5.74382917440  + 19.66976089979 *T)
    sumL1  +=       0.050 * cos(4.69825387775  + 28237.23345938940 *T)
    sumL1  +=       0.047 * cos(5.74015846442  + 12139.55350910680 *T)
    sumL1  +=       0.054 * cos(1.97301333704  + 23581.25817731760 *T)
    sumL1  +=       0.049 * cos(4.98223579027  + 10021.83728009940 *T)
    sumL1  +=       0.046 * cos(5.41431705539  + 33019.02111220460 *T)
    sumL1  +=       0.051 * cos(1.23882053879  + 12539.85338018300 *T)
    sumL1  +=       0.046 * cos(2.41369976086  + 98068.53671630540 *T)
    sumL1  +=       0.044 * cos(0.80750593746  + 167283.76158766547 *T)
    sumL1  +=       0.045 * cos(4.39613584445  + 433.71173787680 *T)
    sumL1  +=       0.044 * cos(2.57358208785  + 12964.30070339100 *T)
    sumL1  +=       0.046 * cos(0.26142733448  + 11.04570026390 *T)
    sumL1  +=       0.045 * cos(2.46230645202  + 51868.24866217880 *T)
    sumL1  +=       0.048 * cos(0.89551707131  + 56600.27928952220 *T)
    sumL1  +=       0.057 * cos(1.86416707010  + 25287.72379939980 *T)
    sumL1  +=       0.042 * cos(5.26377513431  + 26084.02180621620 *T)
    sumL1  +=       0.049 * cos(3.17757670611  + 6303.85124548380 *T)
    sumL1  +=       0.052 * cos(3.65266055509  + 7872.14874527520 *T)
    sumL1  +=       0.040 * cos(1.81891629936  + 34596.36465465240 *T)
    sumL1  +=       0.043 * cos(1.94164978061  + 1903.43681250120 *T)
    sumL1  +=       0.041 * cos(0.74461854136  + 23937.85638974100 *T)
    sumL1  +=       0.048 * cos(6.26034008181  + 28286.99048486120 *T)
    sumL1  +=       0.045 * cos(5.45575017530  + 60530.48898574180 *T)
    sumL1  +=       0.040 * cos(2.92105728682  + 21548.96236929180 *T)
    sumL1  +=       0.040 * cos(0.04502010161  + 38526.57435087200 *T)
    sumL1  +=       0.053 * cos(3.64791042082  + 11925.27409260060 *T)
    sumL1  +=       0.041 * cos(5.04048954693  + 27832.03821928320 *T)
    sumL1  +=       0.042 * cos(5.19292937193  + 19004.64794940840 *T)
    sumL1  +=       0.040 * cos(2.57120233428  + 24356.78078864160 *T)
    sumL1  +=       0.038 * cos(3.49190341464  + 226858.23855437007 *T)
    sumL1  +=       0.039 * cos(4.61184303844  + 95.97922721780 *T)
    sumL1  +=       0.043 * cos(2.20648228147  + 13521.75144159140 *T)
    sumL1  +=       0.040 * cos(5.83461945819  + 16193.65917750039 *T)
    sumL1  +=       0.045 * cos(3.73714372195  + 7875.67186362420 *T)
    sumL1  +=       0.043 * cos(1.14078465002  + 49.75702547180 *T)
    sumL1  +=       0.037 * cos(1.29390383811  + 310.84079886840 *T)
    sumL1  +=       0.038 * cos(0.95970925950  + 664.75604513000 *T)
    sumL1  +=       0.037 * cos(4.27532649462  + 6709.67404086740 *T)
    sumL1  +=       0.038 * cos(2.20108541046  + 28628.33622609960 *T)
    sumL1  +=       0.039 * cos(0.85957361635  + 16522.65971600220 *T)
    sumL1  +=       0.040 * cos(4.35214003837  + 48739.85989708300 *T)
    sumL1  +=       0.036 * cos(1.68167662194  + 10344.29506538580 *T)
    sumL1  +=       0.040 * cos(5.13217319067  + 15664.03552270859 *T)
    sumL1  +=       0.036 * cos(3.72187132496  + 30774.50164257480 *T)
    sumL1  +=       0.036 * cos(3.32158458257  + 16207.88627150200 *T)
    sumL1  +=       0.045 * cos(3.94202418608  + 10988.80815753500 *T)
    sumL1  +=       0.039 * cos(1.51948786199  + 12029.34718788740 *T)
    sumL1  +=       0.026 * cos(3.87685883180  + 6262.72053059260 *T)
    sumL1  +=       0.024 * cos(4.91804163466  + 19651.04848109800 *T)
    sumL1  +=       0.023 * cos(0.29300197709  + 13362.44970679920 *T)
    sumL1  +=       0.021 * cos(3.18605672363  + 6277.55292568400 *T)
    sumL1  +=       0.021 * cos(6.07546891132  + 18139.29450141590 *T)
    sumL1  +=       0.022 * cos(2.31199937177  + 6303.43116939020 *T)
    sumL1  +=       0.021 * cos(3.58418394393  + 18209.33026366019 *T)
    sumL1  +=       0.026 * cos(2.06801296900  + 12573.26524698360 *T)
    sumL1  +=       0.021 * cos(1.56857722317  + 13341.67431130680 *T)
    sumL1  +=       0.024 * cos(5.72605158675  + 29864.33402730900 *T)
    sumL1  +=       0.024 * cos(1.40237993205  + 14712.31711645800 *T)
    sumL1  +=       0.025 * cos(5.71466092822  + 25934.12433108940 *T)

    # EARTH Longitude
    sumL2  = 0
    sumL2  +=   52918.870 * cos(0.00000000000  + 0.00000000000 *T)
    sumL2  +=    8719.837 * cos(1.07209665242  + 6283.07584999140 *T)
    sumL2  +=     309.125 * cos(0.86728818832  + 12566.15169998280 *T)
    sumL2  +=      27.339 * cos(0.05297871691  + 3.52311834900 *T)
    sumL2  +=      16.334 * cos(5.18826691036  + 26.29831979980 *T)
    sumL2  +=      15.752 * cos(3.68457889430  + 155.42039943420 *T)
    sumL2  +=       9.541 * cos(0.75742297675  + 18849.22754997420 *T)
    sumL2  +=       8.937 * cos(2.05705419118  + 77713.77146812050 *T)
    sumL2  +=       6.952 * cos(0.82673305410  + 775.52261132400 *T)
    sumL2  +=       5.064 * cos(4.66284525271  + 1577.34354244780 *T)
    sumL2  +=       4.061 * cos(1.03057162962  + 7.11354700080 *T)
    sumL2  +=       3.463 * cos(5.14074632811  + 796.29800681640 *T)
    sumL2  +=       3.169 * cos(6.05291851171  + 5507.55323866740 *T)
    sumL2  +=       3.020 * cos(1.19246506441  + 242.72860397400 *T)
    sumL2  +=       2.886 * cos(6.11652627155  + 529.69096509460 *T)
    sumL2  +=       3.810 * cos(3.44050803490  + 5573.14280143310 *T)
    sumL2  +=       2.714 * cos(0.30637881025  + 398.14900340820 *T)
    sumL2  +=       2.371 * cos(4.38118838167  + 5223.69391980220 *T)
    sumL2  +=       2.538 * cos(2.27992810679  + 553.56940284240 *T)
    sumL2  +=       2.079 * cos(3.75435330484  + 0.98032106820 *T)
    sumL2  +=       1.675 * cos(0.90216407959  + 951.71840625060 *T)
    sumL2  +=       1.534 * cos(5.75900462759  + 1349.86740965880 *T)
    sumL2  +=       1.224 * cos(2.97328088405  + 2146.16541647520 *T)
    sumL2  +=       1.449 * cos(4.36415913970  + 1748.01641306700 *T)
    sumL2  +=       1.341 * cos(3.72061130861  + 1194.44701022460 *T)
    sumL2  +=       1.254 * cos(2.94846826628  + 6438.49624942560 *T)
    sumL2  +=       0.999 * cos(5.98640014468  + 6286.59896834040 *T)
    sumL2  +=       0.917 * cos(4.79788687522  + 5088.62883976680 *T)
    sumL2  +=       0.828 * cos(3.31321076572  + 213.29909543800 *T)
    sumL2  +=       1.103 * cos(1.27104454479  + 161000.68573767410 *T)
    sumL2  +=       0.762 * cos(3.41582762988  + 5486.77784317500 *T)
    sumL2  +=       1.044 * cos(0.60409577691  + 3154.68708489560 *T)
    sumL2  +=       0.887 * cos(5.23465144638  + 7084.89678111520 *T)
    sumL2  +=       0.645 * cos(1.60096192515  + 2544.31441988340 *T)
    sumL2  +=       0.681 * cos(3.43155669169  + 4694.00295470760 *T)
    sumL2  +=       0.605 * cos(2.47806340546  + 10977.07880469900 *T)
    sumL2  +=       0.706 * cos(6.19393222575  + 4690.47983635860 *T)
    sumL2  +=       0.643 * cos(1.98042503148  + 801.82093112380 *T)
    sumL2  +=       0.502 * cos(1.44394375363  + 6836.64525283380 *T)
    sumL2  +=       0.490 * cos(2.34129524194  + 1592.59601363280 *T)
    sumL2  +=       0.458 * cos(1.30876448575  + 4292.33083295040 *T)
    sumL2  +=       0.431 * cos(0.03526421494  + 7234.79425624200 *T)
    sumL2  +=       0.379 * cos(3.17030522615  + 6309.37416979120 *T)
    sumL2  +=       0.348 * cos(0.99049550009  + 6040.34724601740 *T)
    sumL2  +=       0.386 * cos(1.57019797263  + 71430.69561812909 *T)
    sumL2  +=       0.347 * cos(0.67013291338  + 1059.38193018920 *T)
    sumL2  +=       0.458 * cos(3.81499443681  + 149854.40013480789 *T)
    sumL2  +=       0.302 * cos(1.91760044838  + 10447.38783960440 *T)
    sumL2  +=       0.307 * cos(3.55343347416  + 8031.09226305840 *T)
    sumL2  +=       0.395 * cos(4.93701776616  + 7632.94325965020 *T)
    sumL2  +=       0.314 * cos(3.18093696547  + 2352.86615377180 *T)
    sumL2  +=       0.282 * cos(4.41936437052  + 9437.76293488700 *T)
    sumL2  +=       0.276 * cos(2.71314254553  + 3894.18182954220 *T)
    sumL2  +=       0.298 * cos(2.52037474210  + 6127.65545055720 *T)
    sumL2  +=       0.230 * cos(1.37790215549  + 4705.73230754360 *T)
    sumL2  +=       0.252 * cos(0.55330133471  + 6279.55273164240 *T)
    sumL2  +=       0.255 * cos(5.26570187369  + 6812.76681508600 *T)
    sumL2  +=       0.275 * cos(0.67264264272  + 25132.30339996560 *T)
    sumL2  +=       0.178 * cos(0.92820785174  + 1990.74501704100 *T)
    sumL2  +=       0.221 * cos(0.63897368842  + 6256.77753019160 *T)
    sumL2  +=       0.155 * cos(0.77319790838  + 14143.49524243060 *T)
    sumL2  +=       0.150 * cos(2.40470465561  + 426.59819087600 *T)
    sumL2  +=       0.196 * cos(6.06877865012  + 640.87760738220 *T)
    sumL2  +=       0.137 * cos(2.21679460145  + 8429.24126646660 *T)
    sumL2  +=       0.127 * cos(3.26094223174  + 17789.84561978500 *T)
    sumL2  +=       0.128 * cos(5.47237279946  + 12036.46073488820 *T)
    sumL2  +=       0.122 * cos(2.16291082757  + 10213.28554621100 *T)
    sumL2  +=       0.118 * cos(0.45789822268  + 7058.59846131540 *T)
    sumL2  +=       0.141 * cos(2.34932647403  + 11506.76976979360 *T)
    sumL2  +=       0.100 * cos(0.85621569847  + 6290.18939699220 *T)
    sumL2  +=       0.092 * cos(5.10587476002  + 7079.37385680780 *T)
    sumL2  +=       0.126 * cos(2.65428307012  + 88860.05707098669 *T)
    sumL2  +=       0.106 * cos(5.85646710022  + 7860.41939243920 *T)
    sumL2  +=       0.084 * cos(3.57457554262  + 16730.46368959580 *T)
    sumL2  +=       0.089 * cos(4.21433259618  + 83996.84731811189 *T)
    sumL2  +=       0.097 * cos(5.57938280855  + 13367.97263110660 *T)
    sumL2  +=       0.102 * cos(2.05853060226  + 87.30820453981 *T)
    sumL2  +=       0.080 * cos(4.73792651816  + 11926.25441366880 *T)
    sumL2  +=       0.080 * cos(5.41418965044  + 10973.55568635000 *T)
    sumL2  +=       0.106 * cos(4.10978997399  + 3496.03282613400 *T)
    sumL2  +=       0.102 * cos(3.62650006043  + 244287.60000722768 *T)
    sumL2  +=       0.075 * cos(4.89483161769  + 5643.17856367740 *T)
    sumL2  +=       0.087 * cos(0.42863750683  + 11015.10647733480 *T)
    sumL2  +=       0.069 * cos(1.88908760720  + 10177.25767953360 *T)
    sumL2  +=       0.089 * cos(1.35567273119  + 6681.22485339960 *T)
    sumL2  +=       0.066 * cos(0.99455837265  + 6525.80445396540 *T)
    sumL2  +=       0.067 * cos(5.51240997070  + 3097.88382272579 *T)
    sumL2  +=       0.076 * cos(2.72016814799  + 4164.31198961300 *T)
    sumL2  +=       0.063 * cos(1.44349902540  + 9917.69687450980 *T)
    sumL2  +=       0.078 * cos(3.51469733747  + 11856.21865142450 *T)
    sumL2  +=       0.085 * cos(0.50956043858  + 10575.40668294180 *T)
    sumL2  +=       0.067 * cos(3.62043033405  + 16496.36139620240 *T)
    sumL2  +=       0.055 * cos(5.24637517308  + 3340.61242669980 *T)
    sumL2  +=       0.048 * cos(5.43966777314  + 20426.57109242200 *T)
    sumL2  +=       0.064 * cos(5.79535817813  + 2388.89402044920 *T)
    sumL2  +=       0.046 * cos(5.43499966519  + 6275.96230299060 *T)
    sumL2  +=       0.050 * cos(3.86263598617  + 5729.50644714900 *T)
    sumL2  +=       0.044 * cos(1.52269529228  + 12168.00269657460 *T)
    sumL2  +=       0.057 * cos(4.96352373486  + 14945.31617355440 *T)
    sumL2  +=       0.045 * cos(1.00861230160  + 8635.94200376320 *T)
    sumL2  +=       0.043 * cos(3.30685683359  + 9779.10867612540 *T)
    sumL2  +=       0.042 * cos(0.63481258930  + 2699.73481931760 *T)
    sumL2  +=       0.041 * cos(5.67996766641  + 11712.95531823080 *T)
    sumL2  +=       0.056 * cos(4.34024451468  + 90955.55169449610 *T)
    sumL2  +=       0.041 * cos(5.81722212845  + 709.93304855830 *T)
    sumL2  +=       0.053 * cos(6.17052087143  + 233141.31440436147 *T)
    sumL2  +=       0.037 * cos(3.12495025087  + 16200.77272450120 *T)
    sumL2  +=       0.035 * cos(5.76973458495  + 12569.67481833180 *T)
    sumL2  +=       0.037 * cos(0.31656444326  + 24356.78078864160 *T)
    sumL2  +=       0.035 * cos(0.96229051027  + 17298.18232732620 *T)
    sumL2  +=       0.033 * cos(5.23130355867  + 5331.35744374080 *T)
    sumL2  +=       0.035 * cos(0.62517020593  + 25158.60171976540 *T)
    sumL2  +=       0.035 * cos(0.80004512129  + 13916.01910964160 *T)
    sumL2  +=       0.037 * cos(2.89336088688  + 12721.57209941700 *T)
    sumL2  +=       0.030 * cos(4.50198402401  + 23543.23050468179 *T)
    sumL2  +=       0.030 * cos(5.31355708693  + 18319.53658487960 *T)
    sumL2  +=       0.029 * cos(3.47275229977  + 13119.72110282519 *T)
    sumL2  +=       0.029 * cos(3.11002782516  + 4136.91043351620 *T)
    sumL2  +=       0.032 * cos(5.52273255667  + 5753.38488489680 *T)
    sumL2  +=       0.035 * cos(3.79699996680  + 143571.32428481648 *T)
    sumL2  +=       0.026 * cos(1.50634201907  + 154717.60988768269 *T)
    sumL2  +=       0.030 * cos(3.53519084118  + 6284.05617105960 *T)
    sumL2  +=       0.023 * cos(4.41808025967  + 5884.92684658320 *T)
    sumL2  +=       0.025 * cos(1.38477355808  + 65147.61976813770 *T)
    sumL2  +=       0.023 * cos(3.49782549797  + 7477.52286021600 *T)
    sumL2  +=       0.019 * cos(3.14329413716  + 6496.37494542940 *T)
    sumL2  +=       0.019 * cos(2.20135125199  + 18073.70493865020 *T)
    sumL2  +=       0.019 * cos(4.95020255309  + 3930.20969621960 *T)
    sumL2  +=       0.019 * cos(0.57998702747  + 31415.37924995700 *T)
    sumL2  +=       0.021 * cos(1.75474323399  + 12139.55350910680 *T)
    sumL2  +=       0.019 * cos(3.92233070499  + 19651.04848109800 *T)
    sumL2  +=       0.014 * cos(0.98131213224  + 12559.03815298200 *T)
    sumL2  +=       0.019 * cos(4.93309333729  + 2942.46342329160 *T)
    sumL2  +=       0.016 * cos(5.55997534558  + 8827.39026987480 *T)
    sumL2  +=       0.013 * cos(1.68808165516  + 4535.05943692440 *T)
    sumL2  +=       0.013 * cos(0.33982116161  + 4933.20844033260 *T)
    sumL2  +=       0.012 * cos(1.85426309994  + 5856.47765911540 *T)
    sumL2  +=       0.010 * cos(4.82763996845  + 13095.84266507740 *T)
    sumL2  +=       0.011 * cos(5.38005490571  + 11790.62908865880 *T)
    sumL2  +=       0.010 * cos(1.40815507226  + 10988.80815753500 *T)
    sumL2  +=       0.011 * cos(3.05005267431  + 17260.15465469040 *T)
    sumL2  +=       0.010 * cos(4.93364992366  + 12352.85260454480 *T)

    # EARTH Longitude
    sumL3  = 0
    sumL3  +=     289.226 * cos(5.84384198723  + 6283.07584999140 *T)
    sumL3  +=      34.955 * cos(0.00000000000  + 0.00000000000 *T)
    sumL3  +=      16.819 * cos(5.48766912348  + 12566.15169998280 *T)
    sumL3  +=       2.962 * cos(5.19577265202  + 155.42039943420 *T)
    sumL3  +=       1.288 * cos(4.72200252235  + 3.52311834900 *T)
    sumL3  +=       0.635 * cos(5.96925937141  + 242.72860397400 *T)
    sumL3  +=       0.714 * cos(5.30045809128  + 18849.22754997420 *T)
    sumL3  +=       0.402 * cos(3.78682982419  + 553.56940284240 *T)
    sumL3  +=       0.072 * cos(4.29768126180  + 6286.59896834040 *T)
    sumL3  +=       0.067 * cos(0.90721687647  + 6127.65545055720 *T)
    sumL3  +=       0.036 * cos(5.24029648014  + 6438.49624942560 *T)
    sumL3  +=       0.024 * cos(5.16003960716  + 25132.30339996560 *T)
    sumL3  +=       0.023 * cos(3.01921570335  + 6309.37416979120 *T)
    sumL3  +=       0.017 * cos(5.82863573502  + 6525.80445396540 *T)
    sumL3  +=       0.017 * cos(3.67772863930  + 71430.69561812909 *T)
    sumL3  +=       0.009 * cos(4.58467294499  + 1577.34354244780 *T)
    sumL3  +=       0.008 * cos(1.40626662824  + 11856.21865142450 *T)
    sumL3  +=       0.008 * cos(5.07561257196  + 6256.77753019160 *T)
    sumL3  +=       0.007 * cos(2.82473374405  + 83996.84731811189 *T)
    sumL3  +=       0.005 * cos(2.71488713339  + 10977.07880469900 *T)
    sumL3  +=       0.005 * cos(3.76879847273  + 12036.46073488820 *T)
    sumL3  +=       0.005 * cos(4.28412873331  + 6275.96230299060 *T)

    # EARTH Longitude
    sumL4  = 0
    sumL4  +=     114.084 * cos(3.14159265359  + 0.00000000000 *T)
    sumL4  +=       7.717 * cos(4.13446589358  + 6283.07584999140 *T)
    sumL4  +=       0.765 * cos(3.83803776214  + 12566.15169998280 *T)
    sumL4  +=       0.420 * cos(0.41925861858  + 155.42039943420 *T)
    sumL4  +=       0.040 * cos(3.59847585840  + 18849.22754997420 *T)
    sumL4  +=       0.041 * cos(3.14398414077  + 3.52311834900 *T)
    sumL4  +=       0.035 * cos(5.00298940826  + 5573.14280143310 *T)
    sumL4  +=       0.013 * cos(0.48794833701  + 77713.77146812050 *T)
    sumL4  +=       0.010 * cos(5.64801766350  + 6127.65545055720 *T)
    sumL4  +=       0.008 * cos(2.84160570605  + 161000.68573767410 *T)
    sumL4  +=       0.002 * cos(0.54912904658  + 6438.49624942560 *T)

    # EARTH Longitude
    sumL5  = 0
    sumL5  +=       0.878 * cos(3.14159265359  + 0.00000000000 *T)
    sumL5  +=       0.172 * cos(2.76579069510  + 6283.07584999140 *T)
    sumL5  +=       0.050 * cos(2.01353298182  + 155.42039943420 *T)
    sumL5  +=       0.028 * cos(2.21496423926  + 12566.15169998280 *T)
    sumL5  +=       0.005 * cos(1.75600058765  + 18849.22754997420 *T)

    # EARTH Latitude
    sumB0  = 0
    sumB0  +=     279.620 * cos(3.19870156017  + 84334.66158130829 *T)
    sumB0  +=     101.643 * cos(5.42248619256  + 5507.55323866740 *T)
    sumB0  +=      80.445 * cos(3.88013204458  + 5223.69391980220 *T)
    sumB0  +=      43.806 * cos(3.70444689758  + 2352.86615377180 *T)
    sumB0  +=      31.933 * cos(4.00026369781  + 1577.34354244780 *T)
    sumB0  +=      22.724 * cos(3.98473831560  + 1047.74731175470 *T)
    sumB0  +=      16.392 * cos(3.56456119782  + 5856.47765911540 *T)
    sumB0  +=      18.141 * cos(4.98367470263  + 6283.07584999140 *T)
    sumB0  +=      14.443 * cos(3.70275614914  + 9437.76293488700 *T)
    sumB0  +=      14.304 * cos(3.41117857525  + 10213.28554621100 *T)
    sumB0  +=      11.246 * cos(4.82820690530  + 14143.49524243060 *T)
    sumB0  +=      10.900 * cos(2.08574562327  + 6812.76681508600 *T)
    sumB0  +=       9.714 * cos(3.47303947752  + 4694.00295470760 *T)
    sumB0  +=      10.367 * cos(4.05663927946  + 71092.88135493269 *T)
    sumB0  +=       8.775 * cos(4.44016515669  + 5753.38488489680 *T)
    sumB0  +=       8.366 * cos(4.99251512180  + 7084.89678111520 *T)
    sumB0  +=       6.921 * cos(4.32559054073  + 6275.96230299060 *T)
    sumB0  +=       9.145 * cos(1.14182646613  + 6620.89011318780 *T)
    sumB0  +=       7.194 * cos(3.60193205752  + 529.69096509460 *T)
    sumB0  +=       7.698 * cos(5.55425745881  + 167621.57585086187 *T)
    sumB0  +=       5.285 * cos(2.48446991566  + 4705.73230754360 *T)
    sumB0  +=       5.208 * cos(6.24992674537  + 18073.70493865020 *T)
    sumB0  +=       4.529 * cos(2.33827747356  + 6309.37416979120 *T)
    sumB0  +=       5.579 * cos(4.41023653738  + 7860.41939243920 *T)
    sumB0  +=       4.743 * cos(0.70995680136  + 5884.92684658320 *T)
    sumB0  +=       4.301 * cos(1.10255777773  + 6681.22485339960 *T)
    sumB0  +=       3.849 * cos(1.82229412531  + 5486.77784317500 *T)
    sumB0  +=       4.093 * cos(5.11700141207  + 13367.97263110660 *T)
    sumB0  +=       3.681 * cos(0.43793170356  + 3154.68708489560 *T)
    sumB0  +=       3.420 * cos(5.42034800952  + 6069.77675455340 *T)
    sumB0  +=       3.617 * cos(6.04641937526  + 3930.20969621960 *T)
    sumB0  +=       3.670 * cos(4.58210192227  + 12194.03291462090 *T)
    sumB0  +=       2.918 * cos(1.95463881126  + 10977.07880469900 *T)
    sumB0  +=       2.797 * cos(5.61259275048  + 11790.62908865880 *T)
    sumB0  +=       2.502 * cos(0.60499729367  + 6496.37494542940 *T)
    sumB0  +=       2.319 * cos(5.01648216014  + 1059.38193018920 *T)
    sumB0  +=       2.684 * cos(1.39470396488  + 22003.91463486980 *T)
    sumB0  +=       2.428 * cos(3.24183056052  + 78051.58573131690 *T)
    sumB0  +=       2.120 * cos(4.30691000285  + 5643.17856367740 *T)
    sumB0  +=       2.257 * cos(3.15557225618  + 90617.73743129970 *T)
    sumB0  +=       1.813 * cos(3.75574218285  + 3340.61242669980 *T)
    sumB0  +=       2.226 * cos(2.79699346659  + 12036.46073488820 *T)
    sumB0  +=       1.888 * cos(0.86991545823  + 8635.94200376320 *T)
    sumB0  +=       1.517 * cos(1.95852055701  + 398.14900340820 *T)
    sumB0  +=       1.581 * cos(3.19976230948  + 5088.62883976680 *T)
    sumB0  +=       1.421 * cos(6.25530883827  + 2544.31441988340 *T)
    sumB0  +=       1.595 * cos(0.25619915135  + 17298.18232732620 *T)
    sumB0  +=       1.391 * cos(4.69964175561  + 7058.59846131540 *T)
    sumB0  +=       1.478 * cos(2.81808207569  + 25934.12433108940 *T)
    sumB0  +=       1.481 * cos(3.65823554806  + 11506.76976979360 *T)
    sumB0  +=       1.693 * cos(4.95689385293  + 156475.29024799570 *T)
    sumB0  +=       1.183 * cos(1.29343061246  + 775.52261132400 *T)
    sumB0  +=       1.114 * cos(2.37889311846  + 3738.76143010800 *T)
    sumB0  +=       0.994 * cos(4.30088900425  + 9225.53927328300 *T)
    sumB0  +=       0.924 * cos(3.06451026812  + 4164.31198961300 *T)
    sumB0  +=       0.867 * cos(0.55606931068  + 8429.24126646660 *T)
    sumB0  +=       0.988 * cos(5.97286104208  + 7079.37385680780 *T)
    sumB0  +=       0.824 * cos(1.50984806173  + 10447.38783960440 *T)
    sumB0  +=       0.915 * cos(0.12635654592  + 11015.10647733480 *T)
    sumB0  +=       0.742 * cos(1.99159139281  + 26087.90314157420 *T)
    sumB0  +=       1.039 * cos(3.14159265359  + 0.00000000000 *T)
    sumB0  +=       0.850 * cos(4.24120016095  + 29864.33402730900 *T)
    sumB0  +=       0.755 * cos(2.89631873320  + 4732.03062734340 *T)
    sumB0  +=       0.714 * cos(1.37548118603  + 2146.16541647520 *T)
    sumB0  +=       0.708 * cos(1.91406542362  + 8031.09226305840 *T)
    sumB0  +=       0.746 * cos(0.57893808616  + 796.29800681640 *T)
    sumB0  +=       0.802 * cos(5.12339137230  + 2942.46342329160 *T)
    sumB0  +=       0.751 * cos(1.67479850166  + 21228.39202354580 *T)
    sumB0  +=       0.602 * cos(4.09976538826  + 64809.80550494129 *T)
    sumB0  +=       0.594 * cos(3.49580704962  + 16496.36139620240 *T)
    sumB0  +=       0.592 * cos(4.59481504319  + 4690.47983635860 *T)
    sumB0  +=       0.530 * cos(5.73979295200  + 8827.39026987480 *T)
    sumB0  +=       0.503 * cos(5.66433137112  + 33794.54372352860 *T)
    sumB0  +=       0.483 * cos(1.57106522411  + 801.82093112380 *T)
    sumB0  +=       0.438 * cos(0.06707733767  + 3128.38876509580 *T)
    sumB0  +=       0.423 * cos(2.86944595927  + 12566.15169998280 *T)
    sumB0  +=       0.504 * cos(3.26207669160  + 7632.94325965020 *T)
    sumB0  +=       0.552 * cos(1.02926440457  + 239762.20451754928 *T)
    sumB0  +=       0.427 * cos(3.67434378210  + 213.29909543800 *T)
    sumB0  +=       0.404 * cos(1.46193297142  + 15720.83878487840 *T)
    sumB0  +=       0.503 * cos(4.85802444134  + 6290.18939699220 *T)
    sumB0  +=       0.417 * cos(0.81920713533  + 5216.58037280140 *T)
    sumB0  +=       0.365 * cos(0.01002966162  + 12168.00269657460 *T)
    sumB0  +=       0.363 * cos(1.28376436579  + 6206.80977871580 *T)
    sumB0  +=       0.353 * cos(4.70059133110  + 7234.79425624200 *T)
    sumB0  +=       0.415 * cos(0.96862624175  + 4136.91043351620 *T)
    sumB0  +=       0.387 * cos(3.09145061418  + 25158.60171976540 *T)
    sumB0  +=       0.373 * cos(2.65119262792  + 7342.45778018060 *T)
    sumB0  +=       0.361 * cos(2.97762937739  + 9623.68827669120 *T)
    sumB0  +=       0.418 * cos(3.75759994446  + 5230.80746680300 *T)
    sumB0  +=       0.396 * cos(1.22507712354  + 6438.49624942560 *T)
    sumB0  +=       0.322 * cos(1.21162178805  + 8662.24032356300 *T)
    sumB0  +=       0.284 * cos(5.64170320068  + 1589.07289528380 *T)
    sumB0  +=       0.379 * cos(1.72248432748  + 14945.31617355440 *T)
    sumB0  +=       0.320 * cos(3.94161159962  + 7330.82316174610 *T)
    sumB0  +=       0.313 * cos(5.47602376446  + 1194.44701022460 *T)
    sumB0  +=       0.292 * cos(1.38971327603  + 11769.85369316640 *T)
    sumB0  +=       0.305 * cos(0.80429352049  + 37724.75341974820 *T)
    sumB0  +=       0.257 * cos(5.81382809757  + 426.59819087600 *T)
    sumB0  +=       0.265 * cos(6.10358507671  + 6836.64525283380 *T)
    sumB0  +=       0.250 * cos(4.56452895547  + 7477.52286021600 *T)
    sumB0  +=       0.266 * cos(2.62926282354  + 7238.67559160000 *T)
    sumB0  +=       0.263 * cos(6.22089501237  + 6133.51265285680 *T)
    sumB0  +=       0.306 * cos(2.79682380531  + 1748.01641306700 *T)
    sumB0  +=       0.236 * cos(2.46093023714  + 11371.70468975820 *T)
    sumB0  +=       0.316 * cos(1.62662805006  + 250908.49012041549 *T)
    sumB0  +=       0.216 * cos(3.68721275185  + 5849.36411211460 *T)
    sumB0  +=       0.230 * cos(0.36165162947  + 5863.59120611620 *T)
    sumB0  +=       0.233 * cos(5.03509933858  + 20426.57109242200 *T)
    sumB0  +=       0.200 * cos(5.86073159059  + 4535.05943692440 *T)
    sumB0  +=       0.277 * cos(4.65400292395  + 82239.16695779889 *T)
    sumB0  +=       0.209 * cos(3.72323200804  + 10973.55568635000 *T)
    sumB0  +=       0.199 * cos(5.05186622555  + 5429.87946823940 *T)
    sumB0  +=       0.256 * cos(2.40923279770  + 19651.04848109800 *T)
    sumB0  +=       0.210 * cos(4.50691909144  + 29088.81141598500 *T)
    sumB0  +=       0.181 * cos(6.00294783127  + 4292.33083295040 *T)
    sumB0  +=       0.249 * cos(0.12900984422  + 154379.79562448629 *T)
    sumB0  +=       0.209 * cos(3.87759458598  + 17789.84561978500 *T)
    sumB0  +=       0.225 * cos(3.18339652605  + 18875.52586977400 *T)
    sumB0  +=       0.191 * cos(4.53897489299  + 18477.10876461230 *T)
    sumB0  +=       0.172 * cos(2.09694183014  + 13095.84266507740 *T)
    sumB0  +=       0.182 * cos(3.16107943500  + 16730.46368959580 *T)
    sumB0  +=       0.188 * cos(2.22746128596  + 41654.96311596780 *T)
    sumB0  +=       0.164 * cos(5.18686275017  + 5481.25491886760 *T)
    sumB0  +=       0.160 * cos(2.49298855159  + 12592.45001978260 *T)
    sumB0  +=       0.155 * cos(1.59595438230  + 10021.83728009940 *T)
    sumB0  +=       0.135 * cos(0.21349051064  + 10988.80815753500 *T)
    sumB0  +=       0.178 * cos(3.80375177970  + 23581.25817731760 *T)
    sumB0  +=       0.123 * cos(1.66800739151  + 15110.46611986620 *T)
    sumB0  +=       0.122 * cos(2.72678272244  + 18849.22754997420 *T)
    sumB0  +=       0.126 * cos(1.17675512910  + 14919.01785375460 *T)
    sumB0  +=       0.142 * cos(3.95053441332  + 337.81426319640 *T)
    sumB0  +=       0.116 * cos(6.06340906229  + 6709.67404086740 *T)
    sumB0  +=       0.137 * cos(3.52143246757  + 12139.55350910680 *T)
    sumB0  +=       0.136 * cos(2.92179113542  + 32217.20018108080 *T)
    sumB0  +=       0.110 * cos(3.51203379263  + 18052.92954315780 *T)
    sumB0  +=       0.147 * cos(4.63371971408  + 22805.73556599360 *T)
    sumB0  +=       0.108 * cos(5.45280814878  + 7.11354700080 *T)
    sumB0  +=       0.148 * cos(0.65447253687  + 95480.94718417450 *T)
    sumB0  +=       0.119 * cos(5.92110458985  + 33019.02111220460 *T)
    sumB0  +=       0.110 * cos(5.34824206306  + 639.89728631400 *T)
    sumB0  +=       0.106 * cos(3.71081682629  + 14314.16811304980 *T)
    sumB0  +=       0.139 * cos(6.17607198418  + 24356.78078864160 *T)
    sumB0  +=       0.118 * cos(5.59738712670  + 161338.50000087050 *T)
    sumB0  +=       0.117 * cos(3.65065271640  + 45585.17281218740 *T)
    sumB0  +=       0.127 * cos(4.74596574209  + 49515.38250840700 *T)
    sumB0  +=       0.120 * cos(1.04211499785  + 6915.85958930460 *T)
    sumB0  +=       0.120 * cos(5.60638811846  + 5650.29211067820 *T)
    sumB0  +=       0.115 * cos(3.10668213289  + 14712.31711645800 *T)
    sumB0  +=       0.099 * cos(0.69018940049  + 12779.45079542080 *T)
    sumB0  +=       0.097 * cos(1.07908724794  + 9917.69687450980 *T)
    sumB0  +=       0.093 * cos(2.62295197319  + 17260.15465469040 *T)
    sumB0  +=       0.099 * cos(4.45774681732  + 4933.20844033260 *T)
    sumB0  +=       0.123 * cos(1.37488922089  + 28286.99048486120 *T)
    sumB0  +=       0.121 * cos(5.19767249813  + 27511.46787353720 *T)
    sumB0  +=       0.105 * cos(0.87192267806  + 77375.95720492408 *T)
    sumB0  +=       0.087 * cos(3.93637812950  + 17654.78053974960 *T)
    sumB0  +=       0.122 * cos(2.23956068680  + 83997.09113559539 *T)
    sumB0  +=       0.087 * cos(4.18201600952  + 22779.43724619380 *T)
    sumB0  +=       0.104 * cos(4.59580877295  + 1349.86740965880 *T)
    sumB0  +=       0.102 * cos(2.83545248411  + 12352.85260454480 *T)
    sumB0  +=       0.102 * cos(3.97386522171  + 10818.13528691580 *T)
    sumB0  +=       0.101 * cos(4.32892825857  + 36147.40987730040 *T)
    sumB0  +=       0.094 * cos(5.00001709261  + 150192.21439800429 *T)
    sumB0  +=       0.077 * cos(3.97199369296  + 1592.59601363280 *T)
    sumB0  +=       0.100 * cos(6.07733097102  + 26735.94526221320 *T)
    sumB0  +=       0.086 * cos(5.26029638250  + 28313.28880466100 *T)
    sumB0  +=       0.093 * cos(4.31900620254  + 44809.65020086340 *T)
    sumB0  +=       0.076 * cos(6.22743405935  + 13521.75144159140 *T)
    sumB0  +=       0.072 * cos(1.55820597747  + 6256.77753019160 *T)
    sumB0  +=       0.082 * cos(4.95202664555  + 10575.40668294180 *T)
    sumB0  +=       0.082 * cos(1.69647647075  + 1990.74501704100 *T)
    sumB0  +=       0.075 * cos(2.29836095644  + 3634.62102451840 *T)
    sumB0  +=       0.075 * cos(2.66367876557  + 16200.77272450120 *T)
    sumB0  +=       0.087 * cos(0.26630214764  + 31441.67756975680 *T)
    sumB0  +=       0.077 * cos(2.25530954137  + 5235.32853823670 *T)
    sumB0  +=       0.076 * cos(1.09869730846  + 12903.96596317920 *T)
    sumB0  +=       0.058 * cos(4.28246138307  + 12559.03815298200 *T)
    sumB0  +=       0.064 * cos(5.51112830114  + 173904.65170085328 *T)
    sumB0  +=       0.056 * cos(2.60133794851  + 73188.37597844210 *T)
    sumB0  +=       0.055 * cos(5.81483150022  + 143233.51002162008 *T)
    sumB0  +=       0.054 * cos(3.38482031504  + 323049.11878710287 *T)
    sumB0  +=       0.039 * cos(3.28500401343  + 71768.50988132549 *T)
    sumB0  +=       0.039 * cos(3.11239910690  + 96900.81328129108 *T)

    # EARTH Latitude
    sumB1  = 0
    sumB1  +=       9.030 * cos(3.89729061890  + 5507.55323866740 *T)
    sumB1  +=       6.177 * cos(1.73038850355  + 5223.69391980220 *T)
    sumB1  +=       3.800 * cos(5.24404145734  + 2352.86615377180 *T)
    sumB1  +=       2.834 * cos(2.47345037450  + 1577.34354244780 *T)
    sumB1  +=       1.817 * cos(0.41874743765  + 6283.07584999140 *T)
    sumB1  +=       1.499 * cos(1.83320979291  + 5856.47765911540 *T)
    sumB1  +=       1.466 * cos(5.69401926017  + 5753.38488489680 *T)
    sumB1  +=       1.301 * cos(2.18890066314  + 9437.76293488700 *T)
    sumB1  +=       1.233 * cos(4.95222451476  + 10213.28554621100 *T)
    sumB1  +=       1.021 * cos(0.12866660208  + 7860.41939243920 *T)
    sumB1  +=       0.982 * cos(0.09005453285  + 14143.49524243060 *T)
    sumB1  +=       0.865 * cos(1.73949953555  + 3930.20969621960 *T)
    sumB1  +=       0.581 * cos(2.26949174067  + 5884.92684658320 *T)
    sumB1  +=       0.524 * cos(5.65662503159  + 529.69096509460 *T)
    sumB1  +=       0.473 * cos(6.22750969242  + 6309.37416979120 *T)
    sumB1  +=       0.451 * cos(1.53288619213  + 18073.70493865020 *T)
    sumB1  +=       0.364 * cos(3.61614477374  + 13367.97263110660 *T)
    sumB1  +=       0.372 * cos(3.22470721320  + 6275.96230299060 *T)
    sumB1  +=       0.268 * cos(2.34341267879  + 11790.62908865880 *T)
    sumB1  +=       0.322 * cos(0.94084045832  + 6069.77675455340 *T)
    sumB1  +=       0.232 * cos(0.26781182579  + 7058.59846131540 *T)
    sumB1  +=       0.216 * cos(6.05952221329  + 10977.07880469900 *T)
    sumB1  +=       0.232 * cos(2.93325646109  + 22003.91463486980 *T)
    sumB1  +=       0.204 * cos(3.86264841382  + 6496.37494542940 *T)
    sumB1  +=       0.202 * cos(2.81892511133  + 15720.83878487840 *T)
    sumB1  +=       0.185 * cos(4.93512381859  + 12036.46073488820 *T)
    sumB1  +=       0.220 * cos(3.99305643742  + 6812.76681508600 *T)
    sumB1  +=       0.166 * cos(1.74970002999  + 11506.76976979360 *T)
    sumB1  +=       0.212 * cos(1.57166285369  + 4694.00295470760 *T)
    sumB1  +=       0.157 * cos(1.08259734788  + 5643.17856367740 *T)
    sumB1  +=       0.154 * cos(5.99434678412  + 5486.77784317500 *T)
    sumB1  +=       0.144 * cos(5.23285656085  + 78051.58573131690 *T)
    sumB1  +=       0.144 * cos(1.16454655948  + 90617.73743129970 *T)
    sumB1  +=       0.137 * cos(2.67760436027  + 6290.18939699220 *T)
    sumB1  +=       0.180 * cos(2.06509026215  + 7084.89678111520 *T)
    sumB1  +=       0.121 * cos(5.90212574947  + 9225.53927328300 *T)
    sumB1  +=       0.150 * cos(2.00175038718  + 5230.80746680300 *T)
    sumB1  +=       0.149 * cos(5.06157254516  + 17298.18232732620 *T)
    sumB1  +=       0.118 * cos(5.39979058038  + 3340.61242669980 *T)
    sumB1  +=       0.161 * cos(3.32421999691  + 6283.31966747490 *T)
    sumB1  +=       0.121 * cos(4.36722193162  + 19651.04848109800 *T)
    sumB1  +=       0.116 * cos(5.83462858507  + 4705.73230754360 *T)
    sumB1  +=       0.128 * cos(4.35489873365  + 25934.12433108940 *T)
    sumB1  +=       0.143 * cos(0.00000000000  + 0.00000000000 *T)
    sumB1  +=       0.109 * cos(2.52157834166  + 6438.49624942560 *T)
    sumB1  +=       0.099 * cos(2.70727488041  + 5216.58037280140 *T)
    sumB1  +=       0.103 * cos(0.93782340879  + 8827.39026987480 *T)
    sumB1  +=       0.082 * cos(4.29214680390  + 8635.94200376320 *T)
    sumB1  +=       0.079 * cos(2.24085737326  + 1059.38193018920 *T)
    sumB1  +=       0.097 * cos(5.50959692365  + 29864.33402730900 *T)
    sumB1  +=       0.072 * cos(0.21891639822  + 21228.39202354580 *T)
    sumB1  +=       0.071 * cos(2.86755026812  + 6681.22485339960 *T)
    sumB1  +=       0.074 * cos(2.20184828895  + 37724.75341974820 *T)
    sumB1  +=       0.063 * cos(4.45586625948  + 7079.37385680780 *T)
    sumB1  +=       0.061 * cos(0.63918772258  + 33794.54372352860 *T)
    sumB1  +=       0.047 * cos(2.09070235724  + 3128.38876509580 *T)
    sumB1  +=       0.047 * cos(3.32543843300  + 26087.90314157420 *T)
    sumB1  +=       0.049 * cos(1.60680905005  + 6702.56049386660 *T)
    sumB1  +=       0.057 * cos(0.11215813438  + 29088.81141598500 *T)
    sumB1  +=       0.056 * cos(5.47982934911  + 775.52261132400 *T)
    sumB1  +=       0.050 * cos(1.89396788463  + 12139.55350910680 *T)
    sumB1  +=       0.047 * cos(2.97214907240  + 20426.57109242200 *T)
    sumB1  +=       0.041 * cos(5.55329394890  + 11015.10647733480 *T)
    sumB1  +=       0.041 * cos(5.91861144924  + 23581.25817731760 *T)
    sumB1  +=       0.045 * cos(4.95273290181  + 5863.59120611620 *T)
    sumB1  +=       0.050 * cos(3.62740835096  + 41654.96311596780 *T)
    sumB1  +=       0.037 * cos(6.09033460601  + 64809.80550494129 *T)
    sumB1  +=       0.037 * cos(5.86153655431  + 12566.15169998280 *T)
    sumB1  +=       0.046 * cos(1.65798680284  + 25158.60171976540 *T)
    sumB1  +=       0.038 * cos(2.00673650251  + 426.59819087600 *T)
    sumB1  +=       0.036 * cos(6.24373396652  + 6283.14316029419 *T)
    sumB1  +=       0.036 * cos(0.40465162918  + 6283.00853968860 *T)
    sumB1  +=       0.032 * cos(6.03707103538  + 2942.46342329160 *T)
    sumB1  +=       0.041 * cos(4.86809570283  + 1592.59601363280 *T)
    sumB1  +=       0.028 * cos(4.38359423735  + 7632.94325965020 *T)
    sumB1  +=       0.028 * cos(6.03334294232  + 17789.84561978500 *T)
    sumB1  +=       0.026 * cos(3.88971333608  + 5331.35744374080 *T)
    sumB1  +=       0.026 * cos(5.94932724051  + 16496.36139620240 *T)
    sumB1  +=       0.031 * cos(1.44666331503  + 16730.46368959580 *T)
    sumB1  +=       0.026 * cos(6.26376705837  + 23543.23050468179 *T)
    sumB1  +=       0.033 * cos(0.93797239147  + 213.29909543800 *T)
    sumB1  +=       0.026 * cos(3.71858432944  + 13095.84266507740 *T)
    sumB1  +=       0.027 * cos(0.60565274405  + 10988.80815753500 *T)
    sumB1  +=       0.023 * cos(4.44388985550  + 18849.22754997420 *T)
    sumB1  +=       0.028 * cos(1.53862289477  + 6279.48542133960 *T)
    sumB1  +=       0.028 * cos(1.96831814872  + 6286.66627864320 *T)
    sumB1  +=       0.028 * cos(5.78094918529  + 15110.46611986620 *T)
    sumB1  +=       0.026 * cos(2.48165809843  + 5729.50644714900 *T)
    sumB1  +=       0.020 * cos(3.85655029499  + 9623.68827669120 *T)
    sumB1  +=       0.021 * cos(5.83006047147  + 7234.79425624200 *T)
    sumB1  +=       0.021 * cos(0.69628570421  + 398.14900340820 *T)
    sumB1  +=       0.022 * cos(5.02222806555  + 6127.65545055720 *T)
    sumB1  +=       0.020 * cos(3.47611265290  + 6148.01076995600 *T)
    sumB1  +=       0.020 * cos(0.90769829044  + 5481.25491886760 *T)
    sumB1  +=       0.020 * cos(0.03081589303  + 6418.14093002680 *T)
    sumB1  +=       0.020 * cos(3.74220084927  + 1589.07289528380 *T)
    sumB1  +=       0.021 * cos(4.00149269576  + 3154.68708489560 *T)
    sumB1  +=       0.018 * cos(1.58348238359  + 2118.76386037840 *T)
    sumB1  +=       0.019 * cos(0.85407021371  + 14712.31711645800 *T)

    # EARTH Latitude
    sumB2  = 0
    sumB2  +=       1.662 * cos(1.62703209173  + 84334.66158130829 *T)
    sumB2  +=       0.492 * cos(2.41382223971  + 1047.74731175470 *T)
    sumB2  +=       0.344 * cos(2.24353004539  + 5507.55323866740 *T)
    sumB2  +=       0.258 * cos(6.00906896311  + 5223.69391980220 *T)
    sumB2  +=       0.131 * cos(0.95447345240  + 6283.07584999140 *T)
    sumB2  +=       0.086 * cos(1.67530247303  + 7860.41939243920 *T)
    sumB2  +=       0.090 * cos(0.97606804452  + 1577.34354244780 *T)
    sumB2  +=       0.090 * cos(0.37899871725  + 2352.86615377180 *T)
    sumB2  +=       0.089 * cos(6.25807507963  + 10213.28554621100 *T)
    sumB2  +=       0.075 * cos(0.84213523741  + 167621.57585086187 *T)
    sumB2  +=       0.052 * cos(1.70501566089  + 14143.49524243060 *T)
    sumB2  +=       0.057 * cos(6.15295833679  + 12194.03291462090 *T)
    sumB2  +=       0.051 * cos(1.27616016740  + 5753.38488489680 *T)
    sumB2  +=       0.051 * cos(5.37229738682  + 6812.76681508600 *T)
    sumB2  +=       0.034 * cos(1.73672994279  + 7058.59846131540 *T)
    sumB2  +=       0.038 * cos(2.77761031485  + 10988.80815753500 *T)
    sumB2  +=       0.046 * cos(3.38617099014  + 156475.29024799570 *T)
    sumB2  +=       0.021 * cos(1.95248349228  + 8827.39026987480 *T)
    sumB2  +=       0.018 * cos(3.33419222028  + 8429.24126646660 *T)
    sumB2  +=       0.019 * cos(4.32945160287  + 17789.84561978500 *T)
    sumB2  +=       0.017 * cos(0.66191210656  + 6283.00853968860 *T)
    sumB2  +=       0.018 * cos(3.74885333072  + 11769.85369316640 *T)
    sumB2  +=       0.017 * cos(4.23058370776  + 10977.07880469900 *T)
    sumB2  +=       0.017 * cos(1.78116162721  + 5486.77784317500 *T)
    sumB2  +=       0.021 * cos(1.36972913918  + 12036.46073488820 *T)
    sumB2  +=       0.017 * cos(2.79601092529  + 796.29800681640 *T)
    sumB2  +=       0.015 * cos(0.43087848850  + 11790.62908865880 *T)
    sumB2  +=       0.017 * cos(1.35132152761  + 78051.58573131690 *T)
    sumB2  +=       0.015 * cos(1.17032155085  + 213.29909543800 *T)
    sumB2  +=       0.018 * cos(2.85221514199  + 5088.62883976680 *T)
    sumB2  +=       0.017 * cos(0.21780913672  + 6283.14316029419 *T)
    sumB2  +=       0.013 * cos(1.21201504386  + 25132.30339996560 *T)
    sumB2  +=       0.012 * cos(1.12953712197  + 90617.73743129970 *T)
    sumB2  +=       0.012 * cos(5.13714452592  + 7079.37385680780 *T)
    sumB2  +=       0.013 * cos(3.79842135217  + 4933.20844033260 *T)
    sumB2  +=       0.012 * cos(4.89407978213  + 3738.76143010800 *T)
    sumB2  +=       0.015 * cos(6.05682328852  + 398.14900340820 *T)
    sumB2  +=       0.014 * cos(4.81029291856  + 4694.00295470760 *T)
    sumB2  +=       0.011 * cos(0.61684523405  + 3128.38876509580 *T)
    sumB2  +=       0.011 * cos(5.32876538500  + 6040.34724601740 *T)
    sumB2  +=       0.014 * cos(5.27227350286  + 4535.05943692440 *T)
    sumB2  +=       0.011 * cos(2.39292099451  + 5331.35744374080 *T)
    sumB2  +=       0.010 * cos(4.45296532710  + 6525.80445396540 *T)
    sumB2  +=       0.014 * cos(4.66400985037  + 8031.09226305840 *T)
    sumB2  +=       0.010 * cos(3.22472385926  + 9437.76293488700 *T)
    sumB2  +=       0.011 * cos(3.80913404437  + 801.82093112380 *T)
    sumB2  +=       0.010 * cos(5.15032130575  + 11371.70468975820 *T)
    sumB2  +=       0.013 * cos(0.98720797401  + 5729.50644714900 *T)
    sumB2  +=       0.009 * cos(5.94191743597  + 7632.94325965020 *T)

    # EARTH Latitude
    sumB3  = 0
    sumB3  +=       0.011 * cos(0.23877262399  + 7860.41939243920 *T)
    sumB3  +=       0.009 * cos(1.16069982609  + 5507.55323866740 *T)
    sumB3  +=       0.008 * cos(1.65357552925  + 5884.92684658320 *T)
    sumB3  +=       0.008 * cos(2.86720038197  + 7058.59846131540 *T)
    sumB3  +=       0.007 * cos(3.04818741666  + 5486.77784317500 *T)
    sumB3  +=       0.007 * cos(2.59437103785  + 529.69096509460 *T)
    sumB3  +=       0.008 * cos(4.02863090524  + 6256.77753019160 *T)
    sumB3  +=       0.008 * cos(2.42003508927  + 5753.38488489680 *T)
    sumB3  +=       0.006 * cos(0.84181087594  + 6275.96230299060 *T)
    sumB3  +=       0.006 * cos(5.40160929468  + 1577.34354244780 *T)
    sumB3  +=       0.007 * cos(2.73399865247  + 6309.37416979120 *T)

    # EARTH Latitude
    sumB4  = 0
    sumB4  +=       0.004 * cos(0.79662198849  + 6438.49624942560 *T)
    sumB4  +=       0.005 * cos(0.84308705203  + 1047.74731175470 *T)
    sumB4  +=       0.005 * cos(0.05711572303  + 84334.66158130829 *T)
    sumB4  +=       0.003 * cos(3.46779895686  + 6279.55273164240 *T)
    sumB4  +=       0.003 * cos(2.89822201212  + 6127.65545055720 *T)

    # EARTH Radius
    sumR0  = 0
    sumR0  +=  100013988.799 * cos(0.00000000000  + 0.00000000000 *T)
    sumR0  +=  1670699.626 * cos(3.09846350771  + 6283.07584999140 *T)
    sumR0  +=   13956.023 * cos(3.05524609620  + 12566.15169998280 *T)
    sumR0  +=    3083.720 * cos(5.19846674381  + 77713.77146812050 *T)
    sumR0  +=    1628.461 * cos(1.17387749012  + 5753.38488489680 *T)
    sumR0  +=    1575.568 * cos(2.84685245825  + 7860.41939243920 *T)
    sumR0  +=     924.799 * cos(5.45292234084  + 11506.76976979360 *T)
    sumR0  +=     542.444 * cos(4.56409149777  + 3930.20969621960 *T)
    sumR0  +=     472.110 * cos(3.66100022149  + 5884.92684658320 *T)
    sumR0  +=     328.780 * cos(5.89983646482  + 5223.69391980220 *T)
    sumR0  +=     345.983 * cos(0.96368617687  + 5507.55323866740 *T)
    sumR0  +=     306.784 * cos(0.29867139512  + 5573.14280143310 *T)
    sumR0  +=     174.844 * cos(3.01193636534  + 18849.22754997420 *T)
    sumR0  +=     243.189 * cos(4.27349536153  + 11790.62908865880 *T)
    sumR0  +=     211.829 * cos(5.84714540314  + 1577.34354244780 *T)
    sumR0  +=     185.752 * cos(5.02194447178  + 10977.07880469900 *T)
    sumR0  +=     109.835 * cos(5.05510636285  + 5486.77784317500 *T)
    sumR0  +=      98.316 * cos(0.88681311277  + 6069.77675455340 *T)
    sumR0  +=      86.499 * cos(5.68959778254  + 15720.83878487840 *T)
    sumR0  +=      85.825 * cos(1.27083733351  + 161000.68573767410 *T)
    sumR0  +=      62.916 * cos(0.92177108832  + 529.69096509460 *T)
    sumR0  +=      57.056 * cos(2.01374292014  + 83996.84731811189 *T)
    sumR0  +=      64.903 * cos(0.27250613787  + 17260.15465469040 *T)
    sumR0  +=      49.384 * cos(3.24501240359  + 2544.31441988340 *T)
    sumR0  +=      55.736 * cos(5.24159798933  + 71430.69561812909 *T)
    sumR0  +=      42.515 * cos(6.01110242003  + 6275.96230299060 *T)
    sumR0  +=      46.963 * cos(2.57805070386  + 775.52261132400 *T)
    sumR0  +=      38.968 * cos(5.36071738169  + 4694.00295470760 *T)
    sumR0  +=      44.661 * cos(5.53715807302  + 9437.76293488700 *T)
    sumR0  +=      35.660 * cos(1.67468058995  + 12036.46073488820 *T)
    sumR0  +=      31.921 * cos(0.18368229781  + 5088.62883976680 *T)
    sumR0  +=      31.846 * cos(1.77775642085  + 398.14900340820 *T)
    sumR0  +=      33.193 * cos(0.24370300098  + 7084.89678111520 *T)
    sumR0  +=      38.245 * cos(2.39255343974  + 8827.39026987480 *T)
    sumR0  +=      28.464 * cos(1.21344868176  + 6286.59896834040 *T)
    sumR0  +=      37.490 * cos(0.82952922332  + 19651.04848109800 *T)
    sumR0  +=      36.957 * cos(4.90107591914  + 12139.55350910680 *T)
    sumR0  +=      34.537 * cos(1.84270693282  + 2942.46342329160 *T)
    sumR0  +=      26.275 * cos(4.58896850401  + 10447.38783960440 *T)
    sumR0  +=      24.596 * cos(3.78660875483  + 8429.24126646660 *T)
    sumR0  +=      23.587 * cos(0.26866117066  + 796.29800681640 *T)
    sumR0  +=      27.793 * cos(1.89934330904  + 6279.55273164240 *T)
    sumR0  +=      23.927 * cos(4.99598548138  + 5856.47765911540 *T)
    sumR0  +=      20.349 * cos(4.65267995431  + 2146.16541647520 *T)
    sumR0  +=      23.287 * cos(2.80783650928  + 14143.49524243060 *T)
    sumR0  +=      22.103 * cos(1.95004702988  + 3154.68708489560 *T)
    sumR0  +=      19.506 * cos(5.38227371393  + 2352.86615377180 *T)
    sumR0  +=      17.958 * cos(0.19871379385  + 6812.76681508600 *T)
    sumR0  +=      17.174 * cos(4.43315560735  + 10213.28554621100 *T)
    sumR0  +=      16.190 * cos(5.23160507859  + 17789.84561978500 *T)
    sumR0  +=      17.314 * cos(6.15200787916  + 16730.46368959580 *T)
    sumR0  +=      13.814 * cos(5.18962074032  + 8031.09226305840 *T)
    sumR0  +=      18.833 * cos(0.67306674027  + 149854.40013480789 *T)
    sumR0  +=      18.331 * cos(2.25348733734  + 23581.25817731760 *T)
    sumR0  +=      13.641 * cos(3.68516118804  + 4705.73230754360 *T)
    sumR0  +=      13.139 * cos(0.65289581324  + 13367.97263110660 *T)
    sumR0  +=      10.414 * cos(4.33285688538  + 11769.85369316640 *T)
    sumR0  +=       9.978 * cos(4.20126336355  + 6309.37416979120 *T)
    sumR0  +=      10.169 * cos(1.59390681369  + 4690.47983635860 *T)
    sumR0  +=       7.564 * cos(2.62560597390  + 6256.77753019160 *T)
    sumR0  +=       9.661 * cos(3.67586791220  + 27511.46787353720 *T)
    sumR0  +=       6.743 * cos(0.56270332741  + 3340.61242669980 *T)
    sumR0  +=       8.743 * cos(6.06359123461  + 1748.01641306700 *T)
    sumR0  +=       7.786 * cos(3.67371235637  + 12168.00269657460 *T)
    sumR0  +=       6.633 * cos(5.66149277792  + 11371.70468975820 *T)
    sumR0  +=       7.712 * cos(0.31242577789  + 7632.94325965020 *T)
    sumR0  +=       6.592 * cos(3.13576266188  + 801.82093112380 *T)
    sumR0  +=       7.460 * cos(5.64757188143  + 11926.25441366880 *T)
    sumR0  +=       6.933 * cos(2.92384586400  + 6681.22485339960 *T)
    sumR0  +=       6.802 * cos(1.42329806420  + 23013.53953958720 *T)
    sumR0  +=       6.115 * cos(5.13393615454  + 1194.44701022460 *T)
    sumR0  +=       6.477 * cos(2.64986648492  + 19804.82729158280 *T)
    sumR0  +=       5.233 * cos(4.62434053374  + 6438.49624942560 *T)
    sumR0  +=       6.147 * cos(3.02863936662  + 233141.31440436147 *T)
    sumR0  +=       4.608 * cos(1.72194702724  + 7234.79425624200 *T)
    sumR0  +=       4.221 * cos(1.55697533729  + 7238.67559160000 *T)
    sumR0  +=       5.314 * cos(2.40716580847  + 11499.65622279280 *T)
    sumR0  +=       5.128 * cos(5.32398965690  + 11513.88331679440 *T)
    sumR0  +=       4.770 * cos(0.25554312006  + 11856.21865142450 *T)
    sumR0  +=       5.519 * cos(2.09089154502  + 17298.18232732620 *T)
    sumR0  +=       5.625 * cos(4.34052903053  + 90955.55169449610 *T)
    sumR0  +=       4.578 * cos(4.46569641570  + 5746.27133789600 *T)
    sumR0  +=       3.788 * cos(4.90729383510  + 4164.31198961300 *T)
    sumR0  +=       5.337 * cos(5.09957905104  + 31441.67756975680 *T)
    sumR0  +=       3.967 * cos(1.20054555174  + 1349.86740965880 *T)
    sumR0  +=       4.008 * cos(3.03007204392  + 1059.38193018920 *T)
    sumR0  +=       3.476 * cos(0.76080277030  + 10973.55568635000 *T)
    sumR0  +=       4.232 * cos(1.05485713117  + 5760.49843189760 *T)
    sumR0  +=       4.582 * cos(3.76570026763  + 6386.16862421000 *T)
    sumR0  +=       3.335 * cos(3.13829943354  + 6836.64525283380 *T)
    sumR0  +=       3.418 * cos(3.00072390334  + 4292.33083295040 *T)
    sumR0  +=       3.598 * cos(5.70718084323  + 5643.17856367740 *T)
    sumR0  +=       3.237 * cos(4.16448773994  + 9917.69687450980 *T)
    sumR0  +=       4.154 * cos(2.59941292162  + 7058.59846131540 *T)
    sumR0  +=       3.362 * cos(4.54577697964  + 4732.03062734340 *T)
    sumR0  +=       2.978 * cos(1.30561268820  + 6283.14316029419 *T)
    sumR0  +=       2.765 * cos(0.51311975679  + 26.29831979980 *T)
    sumR0  +=       2.802 * cos(5.66263240521  + 8635.94200376320 *T)
    sumR0  +=       2.927 * cos(5.73787481548  + 16200.77272450120 *T)
    sumR0  +=       3.164 * cos(1.69140262657  + 11015.10647733480 *T)
    sumR0  +=       2.598 * cos(2.96244118586  + 25132.30339996560 *T)
    sumR0  +=       3.519 * cos(3.62639325753  + 244287.60000722768 *T)
    sumR0  +=       2.676 * cos(4.20725700850  + 18073.70493865020 *T)
    sumR0  +=       2.978 * cos(1.74971565805  + 6283.00853968860 *T)
    sumR0  +=       2.287 * cos(1.06975704977  + 14314.16811304980 *T)
    sumR0  +=       2.863 * cos(5.92838131397  + 14712.31711645800 *T)
    sumR0  +=       3.071 * cos(0.23793217002  + 35371.88726597640 *T)
    sumR0  +=       2.656 * cos(0.89959301780  + 12352.85260454480 *T)
    sumR0  +=       2.415 * cos(2.79975176257  + 709.93304855830 *T)
    sumR0  +=       2.814 * cos(3.51488206882  + 21228.39202354580 *T)
    sumR0  +=       1.977 * cos(2.61358297550  + 951.71840625060 *T)
    sumR0  +=       2.548 * cos(2.47684686575  + 6208.29425142410 *T)
    sumR0  +=       1.999 * cos(0.56090388160  + 7079.37385680780 *T)
    sumR0  +=       2.305 * cos(1.05376461628  + 22483.84857449259 *T)
    sumR0  +=       1.855 * cos(2.86090681163  + 5216.58037280140 *T)
    sumR0  +=       2.157 * cos(1.31396741861  + 154717.60988768269 *T)
    sumR0  +=       1.970 * cos(4.36929875289  + 167283.76158766547 *T)
    sumR0  +=       1.635 * cos(5.85571606764  + 10984.19235169980 *T)
    sumR0  +=       1.754 * cos(2.14452408833  + 6290.18939699220 *T)
    sumR0  +=       2.154 * cos(6.03828341543  + 10873.98603048040 *T)
    sumR0  +=       1.714 * cos(3.70157691113  + 1592.59601363280 *T)
    sumR0  +=       1.541 * cos(6.21598380732  + 23543.23050468179 *T)
    sumR0  +=       1.611 * cos(1.99824499377  + 10969.96525769820 *T)
    sumR0  +=       1.712 * cos(1.34295663542  + 3128.38876509580 *T)
    sumR0  +=       1.642 * cos(5.55026665339  + 6496.37494542940 *T)
    sumR0  +=       1.502 * cos(5.43948825854  + 155.42039943420 *T)
    sumR0  +=       1.827 * cos(5.91227480261  + 3738.76143010800 *T)
    sumR0  +=       1.726 * cos(2.16764983583  + 10575.40668294180 *T)
    sumR0  +=       1.532 * cos(5.35683107070  + 13521.75144159140 *T)
    sumR0  +=       1.829 * cos(1.66006148731  + 39302.09696219600 *T)
    sumR0  +=       1.605 * cos(1.90928637633  + 6133.51265285680 *T)
    sumR0  +=       1.282 * cos(2.46014880418  + 13916.01910964160 *T)
    sumR0  +=       1.211 * cos(4.41360631550  + 3894.18182954220 *T)
    sumR0  +=       1.394 * cos(1.77801929354  + 9225.53927328300 *T)
    sumR0  +=       1.571 * cos(4.95512957592  + 25158.60171976540 *T)
    sumR0  +=       1.205 * cos(1.19212540615  + 3.52311834900 *T)
    sumR0  +=       1.132 * cos(2.69830084955  + 6040.34724601740 *T)
    sumR0  +=       1.504 * cos(5.77002730341  + 18209.33026366019 *T)
    sumR0  +=       1.393 * cos(1.62621805428  + 5120.60114558360 *T)
    sumR0  +=       1.077 * cos(2.93931554233  + 17256.63153634140 *T)
    sumR0  +=       1.232 * cos(0.71655165307  + 143571.32428481648 *T)
    sumR0  +=       1.087 * cos(0.99769687939  + 955.59974160860 *T)
    sumR0  +=       1.068 * cos(5.28472576231  + 65147.61976813770 *T)
    sumR0  +=       0.980 * cos(5.10949204607  + 6172.86952877200 *T)
    sumR0  +=       1.169 * cos(3.11664290862  + 14945.31617355440 *T)
    sumR0  +=       1.202 * cos(4.02992510402  + 553.56940284240 *T)
    sumR0  +=       0.979 * cos(2.00000879212  + 15110.46611986620 *T)
    sumR0  +=       0.962 * cos(4.02380771400  + 6282.09552892320 *T)
    sumR0  +=       0.999 * cos(3.62643002790  + 6262.30045449900 *T)
    sumR0  +=       1.030 * cos(5.84989900289  + 213.29909543800 *T)
    sumR0  +=       1.014 * cos(2.84221578218  + 8662.24032356300 *T)
    sumR0  +=       1.185 * cos(1.51330541132  + 17654.78053974960 *T)
    sumR0  +=       0.967 * cos(2.67081017562  + 5650.29211067820 *T)
    sumR0  +=       1.222 * cos(2.65423784904  + 88860.05707098669 *T)
    sumR0  +=       0.981 * cos(2.36370360283  + 6206.80977871580 *T)
    sumR0  +=       1.033 * cos(0.13874927606  + 11712.95531823080 *T)
    sumR0  +=       1.103 * cos(3.08477302937  + 43232.30665841560 *T)
    sumR0  +=       0.781 * cos(2.53372735932  + 16496.36139620240 *T)
    sumR0  +=       1.019 * cos(3.04569392376  + 6037.24420376200 *T)
    sumR0  +=       0.795 * cos(5.80662989111  + 5230.80746680300 *T)
    sumR0  +=       0.813 * cos(3.57710279439  + 10177.25767953360 *T)
    sumR0  +=       0.962 * cos(5.31470594766  + 6284.05617105960 *T)
    sumR0  +=       0.721 * cos(5.96264301567  + 12559.03815298200 *T)
    sumR0  +=       0.966 * cos(2.74714939953  + 6244.94281435360 *T)
    sumR0  +=       0.921 * cos(0.10155275926  + 29088.81141598500 *T)
    sumR0  +=       0.692 * cos(3.89764447548  + 1589.07289528380 *T)
    sumR0  +=       0.719 * cos(5.91791450402  + 4136.91043351620 *T)
    sumR0  +=       0.772 * cos(4.05505682353  + 6127.65545055720 *T)
    sumR0  +=       0.712 * cos(5.49291532439  + 22003.91463486980 *T)
    sumR0  +=       0.672 * cos(1.60700490811  + 11087.28512591840 *T)
    sumR0  +=       0.690 * cos(4.50539825563  + 426.59819087600 *T)
    sumR0  +=       0.854 * cos(3.26104981596  + 20426.57109242200 *T)
    sumR0  +=       0.656 * cos(4.32410182940  + 16858.48253293320 *T)
    sumR0  +=       0.840 * cos(2.59572585222  + 28766.92442448400 *T)
    sumR0  +=       0.692 * cos(0.61650089011  + 11403.67699557500 *T)
    sumR0  +=       0.700 * cos(3.40901167143  + 7.11354700080 *T)
    sumR0  +=       0.726 * cos(0.04243053594  + 5481.25491886760 *T)
    sumR0  +=       0.557 * cos(4.78317696534  + 20199.09495963300 *T)
    sumR0  +=       0.649 * cos(1.04027912958  + 6062.66320755260 *T)
    sumR0  +=       0.633 * cos(5.70229959167  + 45892.73043315699 *T)
    sumR0  +=       0.592 * cos(6.11836729658  + 9623.68827669120 *T)
    sumR0  +=       0.523 * cos(3.62840021266  + 5333.90024102160 *T)
    sumR0  +=       0.604 * cos(5.57734696185  + 10344.29506538580 *T)
    sumR0  +=       0.496 * cos(2.21023499449  + 1990.74501704100 *T)
    sumR0  +=       0.691 * cos(1.96071732602  + 12416.58850284820 *T)
    sumR0  +=       0.640 * cos(1.59074172032  + 18319.53658487960 *T)
    sumR0  +=       0.625 * cos(3.82362791378  + 13517.87010623340 *T)
    sumR0  +=       0.663 * cos(5.08444996779  + 283.85931886520 *T)
    sumR0  +=       0.475 * cos(1.17025894287  + 12569.67481833180 *T)
    sumR0  +=       0.664 * cos(4.50029469969  + 47162.51635463520 *T)
    sumR0  +=       0.569 * cos(0.16310365162  + 17267.26820169119 *T)
    sumR0  +=       0.568 * cos(3.86100969474  + 6076.89030155420 *T)
    sumR0  +=       0.539 * cos(4.83282276086  + 18422.62935909819 *T)
    sumR0  +=       0.466 * cos(0.75872342878  + 7342.45778018060 *T)
    sumR0  +=       0.541 * cos(3.07212190507  + 226858.23855437007 *T)
    sumR0  +=       0.458 * cos(0.26774483096  + 4590.91018048900 *T)
    sumR0  +=       0.610 * cos(1.53597051291  + 33019.02111220460 *T)
    sumR0  +=       0.617 * cos(2.62356328726  + 11190.37790013700 *T)
    sumR0  +=       0.548 * cos(4.55798855791  + 18875.52586977400 *T)
    sumR0  +=       0.633 * cos(4.60110281228  + 66567.48586525429 *T)
    sumR0  +=       0.596 * cos(5.78202396722  + 632.78373931320 *T)
    sumR0  +=       0.533 * cos(5.01786882904  + 12132.43996210600 *T)
    sumR0  +=       0.603 * cos(5.38458554802  + 316428.22867391503 *T)
    sumR0  +=       0.469 * cos(0.59168241917  + 21954.15760939799 *T)
    sumR0  +=       0.548 * cos(3.50613163558  + 17253.04110768959 *T)
    sumR0  +=       0.502 * cos(0.98804327589  + 11609.86254401220 *T)
    sumR0  +=       0.568 * cos(1.98497313089  + 7668.63742494250 *T)
    sumR0  +=       0.482 * cos(1.62141803864  + 12146.66705610760 *T)
    sumR0  +=       0.391 * cos(3.68718382989  + 18052.92954315780 *T)
    sumR0  +=       0.457 * cos(3.77205737340  + 156137.47598479927 *T)
    sumR0  +=       0.401 * cos(5.28260651958  + 15671.08175940660 *T)
    sumR0  +=       0.469 * cos(1.80963184268  + 12562.62858163380 *T)
    sumR0  +=       0.508 * cos(3.36399024699  + 20597.24396304120 *T)
    sumR0  +=       0.450 * cos(5.66054299250  + 10454.50138660520 *T)
    sumR0  +=       0.375 * cos(4.98534633105  + 9779.10867612540 *T)
    sumR0  +=       0.523 * cos(0.97215560834  + 155427.54293624099 *T)
    sumR0  +=       0.403 * cos(5.13939866506  + 1551.04522264800 *T)
    sumR0  +=       0.372 * cos(3.69883738807  + 9388.00590941520 *T)
    sumR0  +=       0.367 * cos(4.43875659716  + 4535.05943692440 *T)
    sumR0  +=       0.406 * cos(4.20863156600  + 12592.45001978260 *T)
    sumR0  +=       0.360 * cos(2.53924644657  + 242.72860397400 *T)
    sumR0  +=       0.471 * cos(4.61907324819  + 5436.99301524020 *T)
    sumR0  +=       0.441 * cos(5.83872966262  + 3496.03282613400 *T)
    sumR0  +=       0.385 * cos(4.94496680973  + 24356.78078864160 *T)
    sumR0  +=       0.349 * cos(6.15018231784  + 19800.94595622480 *T)
    sumR0  +=       0.355 * cos(0.21895678106  + 5429.87946823940 *T)
    sumR0  +=       0.344 * cos(5.62993724928  + 2379.16447357160 *T)
    sumR0  +=       0.380 * cos(2.72105213143  + 11933.36796066960 *T)
    sumR0  +=       0.432 * cos(0.24221790536  + 17996.03116822220 *T)
    sumR0  +=       0.378 * cos(5.22517556974  + 7477.52286021600 *T)
    sumR0  +=       0.337 * cos(5.10888041439  + 5849.36411211460 *T)
    sumR0  +=       0.315 * cos(0.57827745123  + 10557.59416082380 *T)
    sumR0  +=       0.318 * cos(4.49953141399  + 3634.62102451840 *T)
    sumR0  +=       0.323 * cos(1.54274281393  + 10440.27429260360 *T)
    sumR0  +=       0.309 * cos(5.76839284397  + 20.77539549240 *T)
    sumR0  +=       0.301 * cos(2.34727604008  + 4686.88940770680 *T)
    sumR0  +=       0.414 * cos(5.93237602310  + 51092.72605085480 *T)
    sumR0  +=       0.361 * cos(2.16398609550  + 28237.23345938940 *T)
    sumR0  +=       0.288 * cos(0.18376252189  + 13095.84266507740 *T)
    sumR0  +=       0.277 * cos(5.12952205045  + 13119.72110282519 *T)
    sumR0  +=       0.327 * cos(6.19222146204  + 6268.84875598980 *T)
    sumR0  +=       0.273 * cos(0.30522428863  + 23141.55838292460 *T)
    sumR0  +=       0.267 * cos(5.76152585786  + 5966.68398033480 *T)
    sumR0  +=       0.308 * cos(5.99280509979  + 22805.73556599360 *T)
    sumR0  +=       0.345 * cos(2.92489919444  + 36949.23080842420 *T)
    sumR0  +=       0.253 * cos(5.20995219509  + 24072.92146977640 *T)
    sumR0  +=       0.342 * cos(5.72702586209  + 16460.33352952499 *T)
    sumR0  +=       0.261 * cos(2.00304796059  + 6148.01076995600 *T)
    sumR0  +=       0.238 * cos(5.08264392839  + 6915.85958930460 *T)
    sumR0  +=       0.249 * cos(2.94762789744  + 135.06508003540 *T)
    sumR0  +=       0.306 * cos(3.89764686987  + 10988.80815753500 *T)
    sumR0  +=       0.305 * cos(0.05827812117  + 4701.11650170840 *T)
    sumR0  +=       0.319 * cos(2.95712862064  + 163096.18036118348 *T)
    sumR0  +=       0.209 * cos(4.43768461442  + 6546.15977336420 *T)
    sumR0  +=       0.270 * cos(2.06643178717  + 4804.20927592700 *T)
    sumR0  +=       0.217 * cos(0.73691592312  + 6303.85124548380 *T)
    sumR0  +=       0.206 * cos(0.32075959415  + 25934.12433108940 *T)
    sumR0  +=       0.218 * cos(0.18428135264  + 28286.99048486120 *T)
    sumR0  +=       0.205 * cos(5.21312087405  + 20995.39296644940 *T)
    sumR0  +=       0.199 * cos(0.44384292491  + 16737.57723659660 *T)
    sumR0  +=       0.230 * cos(6.06567392849  + 6287.00800325450 *T)
    sumR0  +=       0.219 * cos(1.29194216300  + 5326.78669402080 *T)
    sumR0  +=       0.201 * cos(1.74700937253  + 22743.40937951640 *T)
    sumR0  +=       0.207 * cos(4.45440927276  + 6279.48542133960 *T)
    sumR0  +=       0.269 * cos(6.05640445030  + 64471.99124174489 *T)
    sumR0  +=       0.190 * cos(0.99256176518  + 29296.61538957860 *T)
    sumR0  +=       0.238 * cos(5.42471431221  + 39609.65458316560 *T)
    sumR0  +=       0.262 * cos(5.26961924198  + 522.57741809380 *T)
    sumR0  +=       0.210 * cos(4.68618183158  + 6254.62666252360 *T)
    sumR0  +=       0.197 * cos(2.80624554080  + 4933.20844033260 *T)
    sumR0  +=       0.252 * cos(4.36220154608  + 40879.44050464380 *T)
    sumR0  +=       0.261 * cos(1.07241516738  + 55022.93574707440 *T)
    sumR0  +=       0.189 * cos(3.82966734476  + 419.48464387520 *T)
    sumR0  +=       0.185 * cos(4.14324541379  + 5642.19824260920 *T)
    sumR0  +=       0.247 * cos(3.44855612987  + 6702.56049386660 *T)
    sumR0  +=       0.205 * cos(4.04424043223  + 536.80451209540 *T)
    sumR0  +=       0.191 * cos(3.14082686083  + 16723.35014259500 *T)
    sumR0  +=       0.222 * cos(5.16263907319  + 23539.70738633280 *T)
    sumR0  +=       0.180 * cos(4.56214752149  + 6489.26139842860 *T)
    sumR0  +=       0.219 * cos(0.80382553358  + 16627.37091537720 *T)
    sumR0  +=       0.227 * cos(0.60156339452  + 5905.70224207560 *T)
    sumR0  +=       0.168 * cos(0.88753528161  + 16062.18452611680 *T)
    sumR0  +=       0.158 * cos(0.92127725775  + 23937.85638974100 *T)
    sumR0  +=       0.157 * cos(4.69607868164  + 6805.65326808520 *T)
    sumR0  +=       0.207 * cos(4.88410451334  + 6286.66627864320 *T)
    sumR0  +=       0.160 * cos(4.95943826846  + 10021.83728009940 *T)
    sumR0  +=       0.166 * cos(0.97126433565  + 3097.88382272579 *T)
    sumR0  +=       0.209 * cos(5.75663411805  + 3646.35037735440 *T)
    sumR0  +=       0.175 * cos(6.12762824412  + 239424.39025435288 *T)
    sumR0  +=       0.173 * cos(3.13887234973  + 6179.98307577280 *T)
    sumR0  +=       0.157 * cos(3.62822058179  + 18451.07854656599 *T)
    sumR0  +=       0.157 * cos(4.67695912235  + 6709.67404086740 *T)
    sumR0  +=       0.146 * cos(3.09506069735  + 4907.30205014560 *T)
    sumR0  +=       0.165 * cos(2.27139128760  + 10660.68693504240 *T)
    sumR0  +=       0.201 * cos(1.67701267433  + 2107.03450754240 *T)
    sumR0  +=       0.144 * cos(3.96947747592  + 6019.99192661860 *T)
    sumR0  +=       0.171 * cos(5.91302216729  + 6058.73105428950 *T)
    sumR0  +=       0.144 * cos(2.13155655120  + 26084.02180621620 *T)
    sumR0  +=       0.151 * cos(0.67417383554  + 2388.89402044920 *T)
    sumR0  +=       0.189 * cos(5.07122281033  + 263.08392337280 *T)
    sumR0  +=       0.146 * cos(5.10373877968  + 10770.89325626180 *T)
    sumR0  +=       0.187 * cos(1.23915444627  + 19402.79695281660 *T)
    sumR0  +=       0.174 * cos(0.08407293391  + 9380.95967271720 *T)
    sumR0  +=       0.137 * cos(1.26247412309  + 12566.21901028560 *T)
    sumR0  +=       0.137 * cos(3.52826010842  + 639.89728631400 *T)
    sumR0  +=       0.148 * cos(1.76124372592  + 5888.44996493220 *T)
    sumR0  +=       0.164 * cos(2.39195095081  + 6357.85744855870 *T)
    sumR0  +=       0.146 * cos(2.43675816553  + 5881.40372823420 *T)
    sumR0  +=       0.161 * cos(1.15721259372  + 26735.94526221320 *T)
    sumR0  +=       0.131 * cos(2.51859277344  + 6599.46771964800 *T)
    sumR0  +=       0.153 * cos(5.85203687779  + 6281.59137728310 *T)
    sumR0  +=       0.151 * cos(3.72338532649  + 12669.24447420140 *T)
    sumR0  +=       0.132 * cos(2.38417741883  + 6525.80445396540 *T)
    sumR0  +=       0.129 * cos(0.75556744143  + 5017.50837136500 *T)
    sumR0  +=       0.127 * cos(0.00254936441  + 10027.90319572920 *T)
    sumR0  +=       0.148 * cos(2.85102145528  + 6418.14093002680 *T)
    sumR0  +=       0.143 * cos(5.74460279367  + 26087.90314157420 *T)
    sumR0  +=       0.172 * cos(0.41289962240  + 174242.46596404971 *T)
    sumR0  +=       0.136 * cos(4.15497742275  + 6311.52503745920 *T)
    sumR0  +=       0.170 * cos(5.98194913129  + 327574.51427678124 *T)
    sumR0  +=       0.124 * cos(1.65497607604  + 32217.20018108080 *T)
    sumR0  +=       0.136 * cos(2.48430783417  + 13341.67431130680 *T)
    sumR0  +=       0.165 * cos(2.49667924600  + 58953.14544329400 *T)
    sumR0  +=       0.123 * cos(3.45660563754  + 6277.55292568400 *T)
    sumR0  +=       0.117 * cos(0.86065134175  + 6245.04817735560 *T)
    sumR0  +=       0.149 * cos(5.61358280963  + 5729.50644714900 *T)
    sumR0  +=       0.153 * cos(0.26860029950  + 245.83164622940 *T)
    sumR0  +=       0.128 * cos(0.71204006588  + 103.09277421860 *T)
    sumR0  +=       0.159 * cos(2.43166592149  + 221995.02880149524 *T)
    sumR0  +=       0.130 * cos(2.80707316718  + 6016.46880826960 *T)
    sumR0  +=       0.137 * cos(1.70657709294  + 12566.08438968000 *T)
    sumR0  +=       0.111 * cos(1.56305648432  + 17782.73207278420 *T)
    sumR0  +=       0.113 * cos(3.58302904101  + 25685.87280280800 *T)
    sumR0  +=       0.109 * cos(3.26403795962  + 6819.88036208680 *T)
    sumR0  +=       0.122 * cos(0.34120688217  + 1162.47470440780 *T)
    sumR0  +=       0.119 * cos(5.84644718278  + 12721.57209941700 *T)
    sumR0  +=       0.144 * cos(2.28899679126  + 12489.88562870720 *T)
    sumR0  +=       0.137 * cos(5.82029768354  + 44809.65020086340 *T)
    sumR0  +=       0.107 * cos(2.42818544140  + 5547.19933645960 *T)
    sumR0  +=       0.134 * cos(1.26539982939  + 5331.35744374080 *T)
    sumR0  +=       0.103 * cos(5.96518130595  + 6321.10352262720 *T)
    sumR0  +=       0.109 * cos(0.33808549034  + 11300.58422135640 *T)
    sumR0  +=       0.129 * cos(5.89187277327  + 12029.34718788740 *T)
    sumR0  +=       0.122 * cos(5.77325634636  + 11919.14086666800 *T)
    sumR0  +=       0.107 * cos(6.24998989350  + 77690.75950573849 *T)
    sumR0  +=       0.107 * cos(1.00535580713  + 77736.78343050249 *T)
    sumR0  +=       0.143 * cos(0.24122178432  + 4214.06901508480 *T)
    sumR0  +=       0.143 * cos(0.88529649733  + 7576.56007357400 *T)
    sumR0  +=       0.107 * cos(2.92124030496  + 31415.37924995700 *T)
    sumR0  +=       0.099 * cos(5.70862227072  + 5540.08578945880 *T)
    sumR0  +=       0.110 * cos(0.37528037383  + 5863.59120611620 *T)
    sumR0  +=       0.104 * cos(4.44107178366  + 2118.76386037840 *T)
    sumR0  +=       0.098 * cos(5.95877916706  + 4061.21921539440 *T)
    sumR0  +=       0.113 * cos(1.24206857385  + 84672.47584450469 *T)
    sumR0  +=       0.124 * cos(2.55619029867  + 12539.85338018300 *T)
    sumR0  +=       0.110 * cos(3.66952094329  + 238004.52415723627 *T)
    sumR0  +=       0.112 * cos(4.32512422943  + 97238.62754448748 *T)
    sumR0  +=       0.097 * cos(3.70151541181  + 11720.06886523160 *T)
    sumR0  +=       0.120 * cos(1.26895630252  + 12043.57428188900 *T)
    sumR0  +=       0.094 * cos(2.56461130309  + 19004.64794940840 *T)
    sumR0  +=       0.117 * cos(3.65425622684  + 34520.30930938080 *T)
    sumR0  +=       0.098 * cos(0.13589994287  + 11080.17157891760 *T)
    sumR0  +=       0.097 * cos(5.38330115253  + 7834.12107263940 *T)
    sumR0  +=       0.097 * cos(2.46722096722  + 71980.63357473118 *T)
    sumR0  +=       0.095 * cos(5.36958330451  + 6288.59877429880 *T)
    sumR0  +=       0.111 * cos(5.01961920313  + 11823.16163945020 *T)
    sumR0  +=       0.090 * cos(2.72299804525  + 26880.31981303260 *T)
    sumR0  +=       0.099 * cos(0.90164266377  + 18635.92845453620 *T)
    sumR0  +=       0.126 * cos(4.78722177847  + 305281.94307104882 *T)
    sumR0  +=       0.093 * cos(0.21240380046  + 18139.29450141590 *T)
    sumR0  +=       0.124 * cos(5.00979495566  + 172146.97134054027 *T)
    sumR0  +=       0.099 * cos(5.67090026475  + 16522.65971600220 *T)
    sumR0  +=       0.092 * cos(2.28180963676  + 12491.37010141550 *T)
    sumR0  +=       0.090 * cos(4.50544881196  + 40077.61957352000 *T)
    sumR0  +=       0.100 * cos(2.00639461612  + 12323.42309600880 *T)
    sumR0  +=       0.095 * cos(5.68801979087  + 14919.01785375460 *T)
    sumR0  +=       0.087 * cos(1.86043406047  + 27707.54249429480 *T)
    sumR0  +=       0.105 * cos(3.02903468417  + 22345.26037610820 *T)
    sumR0  +=       0.087 * cos(5.43970168638  + 6272.03014972750 *T)
    sumR0  +=       0.089 * cos(1.63389387182  + 33326.57873317420 *T)
    sumR0  +=       0.082 * cos(5.58298993353  + 10241.20229116720 *T)
    sumR0  +=       0.094 * cos(5.47749711149  + 9924.81042151060 *T)
    sumR0  +=       0.082 * cos(4.71988314145  + 15141.39079431200 *T)
    sumR0  +=       0.097 * cos(5.61458778738  + 2787.04302385740 *T)
    sumR0  +=       0.096 * cos(3.89073946348  + 6379.05507720920 *T)
    sumR0  +=       0.081 * cos(3.13038482444  + 36147.40987730040 *T)
    sumR0  +=       0.110 * cos(4.89978492291  + 72140.62866668739 *T)
    sumR0  +=       0.097 * cos(5.20764563059  + 6303.43116939020 *T)
    sumR0  +=       0.082 * cos(5.26342716139  + 9814.60410029120 *T)
    sumR0  +=       0.109 * cos(2.35555589770  + 83286.91426955358 *T)
    sumR0  +=       0.097 * cos(2.58492958057  + 30666.15495843280 *T)
    sumR0  +=       0.093 * cos(1.32651591333  + 23020.65308658799 *T)
    sumR0  +=       0.078 * cos(3.99588630754  + 11293.47067435560 *T)
    sumR0  +=       0.090 * cos(0.57771932738  + 26482.17080962440 *T)
    sumR0  +=       0.106 * cos(3.92012705073  + 62883.35513951360 *T)
    sumR0  +=       0.098 * cos(2.94397773524  + 316.39186965660 *T)
    sumR0  +=       0.076 * cos(3.96310417608  + 29026.48522950779 *T)
    sumR0  +=       0.078 * cos(1.97068529306  + 90279.92316810328 *T)
    sumR0  +=       0.076 * cos(0.23027966596  + 21424.46664430340 *T)
    sumR0  +=       0.080 * cos(2.23099742212  + 266.60704172180 *T)
    sumR0  +=       0.079 * cos(1.46227790922  + 8982.81066930900 *T)
    sumR0  +=       0.102 * cos(4.92129953565  + 5621.84292321040 *T)
    sumR0  +=       0.100 * cos(0.39243148321  + 24279.10701821359 *T)
    sumR0  +=       0.071 * cos(1.52014858474  + 33794.54372352860 *T)
    sumR0  +=       0.076 * cos(0.22880641443  + 57375.80190084620 *T)
    sumR0  +=       0.091 * cos(0.96515913904  + 48739.85989708300 *T)
    sumR0  +=       0.075 * cos(2.77638585157  + 12964.30070339100 *T)
    sumR0  +=       0.077 * cos(5.18846946344  + 11520.99686379520 *T)
    sumR0  +=       0.068 * cos(0.50006599129  + 4274.51831083240 *T)
    sumR0  +=       0.075 * cos(2.07323762803  + 15664.03552270859 *T)
    sumR0  +=       0.074 * cos(1.01884134928  + 6393.28217121080 *T)
    sumR0  +=       0.077 * cos(0.46665178780  + 16207.88627150200 *T)
    sumR0  +=       0.081 * cos(4.10452219483  + 161710.61878623240 *T)
    sumR0  +=       0.067 * cos(3.83840630887  + 6262.72053059260 *T)
    sumR0  +=       0.071 * cos(3.91415523291  + 7875.67186362420 *T)
    sumR0  +=       0.081 * cos(0.91938383237  + 74.78159856730 *T)
    sumR0  +=       0.083 * cos(4.69916218791  + 23006.42599258639 *T)
    sumR0  +=       0.063 * cos(2.32556465878  + 6279.19451463340 *T)
    sumR0  +=       0.065 * cos(5.41938745446  + 28628.33622609960 *T)
    sumR0  +=       0.065 * cos(3.02336771694  + 5959.57043333400 *T)
    sumR0  +=       0.064 * cos(3.31033198370  + 2636.72547263700 *T)
    sumR0  +=       0.064 * cos(0.18375587519  + 1066.49547719000 *T)
    sumR0  +=       0.080 * cos(5.81239171612  + 12341.80690428090 *T)
    sumR0  +=       0.066 * cos(2.15105504851  + 38.02767263580 *T)
    sumR0  +=       0.062 * cos(2.43313614978  + 10138.10951694860 *T)
    sumR0  +=       0.060 * cos(3.16153906470  + 5490.30096152400 *T)
    sumR0  +=       0.069 * cos(0.30764736334  + 7018.95236352320 *T)
    sumR0  +=       0.068 * cos(2.24442548639  + 24383.07910844140 *T)
    sumR0  +=       0.078 * cos(1.39649386463  + 9411.46461508720 *T)
    sumR0  +=       0.063 * cos(0.72976362625  + 6286.95718534940 *T)
    sumR0  +=       0.073 * cos(4.95125917731  + 6453.74872061060 *T)
    sumR0  +=       0.078 * cos(0.32736023459  + 6528.90749622080 *T)
    sumR0  +=       0.059 * cos(4.95362151577  + 35707.71008290740 *T)
    sumR0  +=       0.070 * cos(2.37962727525  + 15508.61512327440 *T)
    sumR0  +=       0.073 * cos(1.35229143111  + 5327.47610838280 *T)
    sumR0  +=       0.072 * cos(5.91833527334  + 10881.09957748120 *T)
    sumR0  +=       0.059 * cos(5.36231868425  + 10239.58386601080 *T)
    sumR0  +=       0.059 * cos(1.63156134967  + 61306.01159706580 *T)
    sumR0  +=       0.054 * cos(4.29491690425  + 21947.11137270000 *T)
    sumR0  +=       0.057 * cos(5.89190132575  + 34513.26307268280 *T)
    sumR0  +=       0.074 * cos(1.38235845304  + 9967.45389998160 *T)
    sumR0  +=       0.053 * cos(3.86543309344  + 32370.97899156560 *T)
    sumR0  +=       0.055 * cos(4.51794544854  + 34911.41207609100 *T)
    sumR0  +=       0.063 * cos(5.41479412056  + 11502.83761653050 *T)
    sumR0  +=       0.063 * cos(2.34416220742  + 11510.70192305670 *T)
    sumR0  +=       0.068 * cos(0.77493931112  + 29864.33402730900 *T)
    sumR0  +=       0.060 * cos(5.57024703495  + 5756.90800324580 *T)
    sumR0  +=       0.072 * cos(2.80863088166  + 10866.87248347960 *T)
    sumR0  +=       0.061 * cos(2.69736991384  + 82576.98122099529 *T)
    sumR0  +=       0.063 * cos(5.32068807257  + 3116.65941225980 *T)
    sumR0  +=       0.052 * cos(1.02278758099  + 6272.43918464160 *T)
    sumR0  +=       0.069 * cos(5.00698550308  + 25287.72379939980 *T)
    sumR0  +=       0.066 * cos(6.12047940728  + 12074.48840752400 *T)
    sumR0  +=       0.051 * cos(2.59519527563  + 11396.56344857420 *T)
    sumR0  +=       0.056 * cos(2.57995973521  + 17892.93839400359 *T)
    sumR0  +=       0.059 * cos(0.44167237620  + 250570.67585721909 *T)
    sumR0  +=       0.059 * cos(3.84070143543  + 5483.25472482600 *T)
    sumR0  +=       0.049 * cos(0.54704693048  + 22594.05489571199 *T)
    sumR0  +=       0.065 * cos(2.38423614501  + 52670.06959330260 *T)
    sumR0  +=       0.069 * cos(5.34363738671  + 66813.56483573320 *T)
    sumR0  +=       0.057 * cos(5.42770501007  + 310145.15282392362 *T)
    sumR0  +=       0.053 * cos(1.17760296075  + 149.56319713460 *T)
    sumR0  +=       0.061 * cos(4.02090887211  + 34596.36465465240 *T)
    sumR0  +=       0.049 * cos(4.18361320516  + 18606.49894600020 *T)
    sumR0  +=       0.055 * cos(0.83886167974  + 20452.86941222180 *T)
    sumR0  +=       0.050 * cos(1.46327331958  + 37455.72649597440 *T)
    sumR0  +=       0.048 * cos(4.53854727167  + 29822.78323632420 *T)
    sumR0  +=       0.058 * cos(3.34847975377  + 33990.61834428620 *T)
    sumR0  +=       0.065 * cos(1.45522693982  + 76251.32777062019 *T)
    sumR0  +=       0.056 * cos(2.35650663692  + 37724.75341974820 *T)
    sumR0  +=       0.052 * cos(2.61551081496  + 5999.21653112620 *T)
    sumR0  +=       0.053 * cos(0.17334326094  + 77717.29458646949 *T)
    sumR0  +=       0.053 * cos(0.79879700631  + 77710.24834977149 *T)
    sumR0  +=       0.047 * cos(0.43240779709  + 735.87651353180 *T)
    sumR0  +=       0.053 * cos(4.58763261686  + 11616.97609101300 *T)
    sumR0  +=       0.048 * cos(6.20230111054  + 4171.42553661380 *T)
    sumR0  +=       0.052 * cos(1.09723616404  + 640.87760738220 *T)
    sumR0  +=       0.057 * cos(3.42008310383  + 50317.20343953080 *T)
    sumR0  +=       0.053 * cos(1.01528448581  + 149144.46708624958 *T)
    sumR0  +=       0.047 * cos(3.00924906195  + 52175.80628314840 *T)
    sumR0  +=       0.052 * cos(2.03254070404  + 6293.71251534120 *T)
    sumR0  +=       0.048 * cos(0.12356889734  + 13362.44970679920 *T)
    sumR0  +=       0.045 * cos(3.37963782356  + 10763.77970926100 *T)
    sumR0  +=       0.047 * cos(5.50981287869  + 12779.45079542080 *T)
    sumR0  +=       0.062 * cos(5.45209070099  + 949.17560896980 *T)
    sumR0  +=       0.061 * cos(2.93237974631  + 5791.41255753260 *T)
    sumR0  +=       0.044 * cos(2.87440620802  + 8584.66166590080 *T)
    sumR0  +=       0.046 * cos(4.03141796560  + 10667.80048204320 *T)
    sumR0  +=       0.047 * cos(3.89902931422  + 3903.91137641980 *T)
    sumR0  +=       0.046 * cos(2.75700467329  + 6993.00889854970 *T)
    sumR0  +=       0.045 * cos(1.93386293300  + 206.18554843720 *T)
    sumR0  +=       0.047 * cos(2.57670800912  + 11492.54267579200 *T)
    sumR0  +=       0.044 * cos(3.62570223167  + 63658.87775083760 *T)
    sumR0  +=       0.051 * cos(0.84536826273  + 12345.73905754400 *T)
    sumR0  +=       0.043 * cos(0.01524970172  + 37853.87549938260 *T)
    sumR0  +=       0.041 * cos(3.27146326065  + 8858.31494432060 *T)
    sumR0  +=       0.045 * cos(3.03765521215  + 65236.22129328540 *T)
    sumR0  +=       0.047 * cos(1.44447548944  + 21393.54196985760 *T)
    sumR0  +=       0.058 * cos(5.45843180927  + 1975.49254585600 *T)
    sumR0  +=       0.050 * cos(2.13285524146  + 12573.26524698360 *T)
    sumR0  +=       0.041 * cos(1.32190847146  + 2547.83753823240 *T)
    sumR0  +=       0.047 * cos(3.67579608544  + 28313.28880466100 *T)
    sumR0  +=       0.041 * cos(2.24013475126  + 8273.82086703240 *T)
    sumR0  +=       0.047 * cos(6.21438985953  + 10991.30589870060 *T)
    sumR0  +=       0.042 * cos(3.01631817350  + 853.19638175200 *T)
    sumR0  +=       0.056 * cos(1.09773690181  + 77376.20102240759 *T)
    sumR0  +=       0.040 * cos(2.35698541041  + 2699.73481931760 *T)
    sumR0  +=       0.043 * cos(5.28030898459  + 17796.95916678580 *T)
    sumR0  +=       0.054 * cos(2.59175932091  + 22910.44676536859 *T)
    sumR0  +=       0.054 * cos(0.88027764102  + 71960.38658322369 *T)
    sumR0  +=       0.055 * cos(0.07988899477  + 83467.15635301729 *T)
    sumR0  +=       0.039 * cos(1.12867321442  + 9910.58332750900 *T)
    sumR0  +=       0.040 * cos(1.35670430524  + 27177.85152920020 *T)
    sumR0  +=       0.039 * cos(4.39624220245  + 5618.31980486140 *T)
    sumR0  +=       0.042 * cos(4.78798367468  + 7856.89627409019 *T)
    sumR0  +=       0.047 * cos(2.75482175292  + 18202.21671665939 *T)
    sumR0  +=       0.039 * cos(1.97008298629  + 24491.42579258340 *T)
    sumR0  +=       0.042 * cos(4.04346599946  + 7863.94251078820 *T)
    sumR0  +=       0.038 * cos(0.49178679251  + 38650.17350619900 *T)
    sumR0  +=       0.036 * cos(4.86047906533  + 4157.19844261220 *T)
    sumR0  +=       0.043 * cos(5.64354880978  + 1062.90504853820 *T)
    sumR0  +=       0.036 * cos(3.98066313627  + 12565.17137891460 *T)
    sumR0  +=       0.042 * cos(2.30753932657  + 6549.68289171320 *T)
    sumR0  +=       0.040 * cos(5.39694918320  + 9498.21223063460 *T)
    sumR0  +=       0.040 * cos(3.30603243754  + 23536.11695768099 *T)
    sumR0  +=       0.050 * cos(6.15760345261  + 78051.34191383338 *T)

    # EARTH Radius
    sumR1  = 0
    sumR1  +=  103018.608 * cos(1.10748969588  + 6283.07584999140 *T)
    sumR1  +=    1721.238 * cos(1.06442301418  + 12566.15169998280 *T)
    sumR1  +=     702.215 * cos(3.14159265359  + 0.00000000000 *T)
    sumR1  +=      32.346 * cos(1.02169059149  + 18849.22754997420 *T)
    sumR1  +=      30.799 * cos(2.84353804832  + 5507.55323866740 *T)
    sumR1  +=      24.971 * cos(1.31906709482  + 5223.69391980220 *T)
    sumR1  +=      18.485 * cos(1.42429748614  + 1577.34354244780 *T)
    sumR1  +=      10.078 * cos(5.91378194648  + 10977.07880469900 *T)
    sumR1  +=       8.634 * cos(0.27146150602  + 5486.77784317500 *T)
    sumR1  +=       8.654 * cos(1.42046854427  + 6275.96230299060 *T)
    sumR1  +=       5.069 * cos(1.68613426734  + 5088.62883976680 *T)
    sumR1  +=       4.985 * cos(6.01401770704  + 6286.59896834040 *T)
    sumR1  +=       4.669 * cos(5.98724494073  + 529.69096509460 *T)
    sumR1  +=       4.395 * cos(0.51800238019  + 4694.00295470760 *T)
    sumR1  +=       3.872 * cos(4.74969833437  + 2544.31441988340 *T)
    sumR1  +=       3.750 * cos(5.07097685568  + 796.29800681640 *T)
    sumR1  +=       4.100 * cos(1.08424786092  + 9437.76293488700 *T)
    sumR1  +=       3.518 * cos(0.02290216272  + 83996.84731811189 *T)
    sumR1  +=       3.436 * cos(0.94937019624  + 71430.69561812909 *T)
    sumR1  +=       3.221 * cos(6.15628775313  + 2146.16541647520 *T)
    sumR1  +=       3.414 * cos(5.41218322538  + 775.52261132400 *T)
    sumR1  +=       2.863 * cos(5.48432847146  + 10447.38783960440 *T)
    sumR1  +=       2.520 * cos(0.24276941146  + 398.14900340820 *T)
    sumR1  +=       2.201 * cos(4.95216196651  + 6812.76681508600 *T)
    sumR1  +=       2.186 * cos(0.41991743105  + 8031.09226305840 *T)
    sumR1  +=       2.838 * cos(3.42034351366  + 2352.86615377180 *T)
    sumR1  +=       2.554 * cos(6.13241878525  + 6438.49624942560 *T)
    sumR1  +=       1.932 * cos(5.31374608366  + 8429.24126646660 *T)
    sumR1  +=       2.429 * cos(3.09164528262  + 4690.47983635860 *T)
    sumR1  +=       1.730 * cos(1.53686208550  + 4705.73230754360 *T)
    sumR1  +=       2.250 * cos(3.68863633842  + 7084.89678111520 *T)
    sumR1  +=       2.093 * cos(1.28191783032  + 1748.01641306700 *T)
    sumR1  +=       1.441 * cos(0.81656250862  + 14143.49524243060 *T)
    sumR1  +=       1.483 * cos(3.22225357771  + 7234.79425624200 *T)
    sumR1  +=       1.754 * cos(3.22883705112  + 6279.55273164240 *T)
    sumR1  +=       1.583 * cos(4.09702349428  + 11499.65622279280 *T)
    sumR1  +=       1.575 * cos(5.53890170575  + 3154.68708489560 *T)
    sumR1  +=       1.847 * cos(1.82040335363  + 7632.94325965020 *T)
    sumR1  +=       1.504 * cos(3.63293385726  + 11513.88331679440 *T)
    sumR1  +=       1.337 * cos(4.64440864339  + 6836.64525283380 *T)
    sumR1  +=       1.275 * cos(2.69341415363  + 1349.86740965880 *T)
    sumR1  +=       1.352 * cos(6.15101580257  + 5746.27133789600 *T)
    sumR1  +=       1.125 * cos(3.35673439497  + 17789.84561978500 *T)
    sumR1  +=       1.470 * cos(3.65282991755  + 1194.44701022460 *T)
    sumR1  +=       1.177 * cos(2.57676109092  + 13367.97263110660 *T)
    sumR1  +=       1.101 * cos(4.49748696552  + 4292.33083295040 *T)
    sumR1  +=       1.234 * cos(5.65036509521  + 5760.49843189760 *T)
    sumR1  +=       0.984 * cos(0.65517395136  + 5856.47765911540 *T)
    sumR1  +=       0.928 * cos(2.32420318751  + 10213.28554621100 *T)
    sumR1  +=       1.077 * cos(5.82812169132  + 12036.46073488820 *T)
    sumR1  +=       0.916 * cos(0.76613009583  + 16730.46368959580 *T)
    sumR1  +=       0.877 * cos(1.50137505051  + 11926.25441366880 *T)
    sumR1  +=       1.023 * cos(5.62076589825  + 6256.77753019160 *T)
    sumR1  +=       0.851 * cos(0.65709335533  + 155.42039943420 *T)
    sumR1  +=       0.802 * cos(4.10519132088  + 951.71840625060 *T)
    sumR1  +=       0.857 * cos(1.41661697538  + 5753.38488489680 *T)
    sumR1  +=       0.994 * cos(1.14418521187  + 1059.38193018920 *T)
    sumR1  +=       0.813 * cos(1.63948433322  + 6681.22485339960 *T)
    sumR1  +=       0.662 * cos(4.55200452260  + 5216.58037280140 *T)
    sumR1  +=       0.644 * cos(4.19478168733  + 6040.34724601740 *T)
    sumR1  +=       0.626 * cos(1.50767713598  + 5643.17856367740 *T)
    sumR1  +=       0.590 * cos(6.18277145205  + 4164.31198961300 *T)
    sumR1  +=       0.635 * cos(0.52413263542  + 6290.18939699220 *T)
    sumR1  +=       0.650 * cos(0.97935690350  + 25132.30339996560 *T)
    sumR1  +=       0.568 * cos(2.30125315873  + 10973.55568635000 *T)
    sumR1  +=       0.547 * cos(5.27256412213  + 3340.61242669980 *T)
    sumR1  +=       0.547 * cos(2.20144422886  + 1592.59601363280 *T)
    sumR1  +=       0.526 * cos(0.92464258226  + 11371.70468975820 *T)
    sumR1  +=       0.490 * cos(5.90951388655  + 3894.18182954220 *T)
    sumR1  +=       0.478 * cos(1.66857963179  + 12168.00269657460 *T)
    sumR1  +=       0.516 * cos(3.59803483887  + 10969.96525769820 *T)
    sumR1  +=       0.518 * cos(3.97914412373  + 17298.18232732620 *T)
    sumR1  +=       0.534 * cos(5.03740926442  + 9917.69687450980 *T)
    sumR1  +=       0.487 * cos(2.50545369269  + 6127.65545055720 *T)
    sumR1  +=       0.416 * cos(4.04828175503  + 10984.19235169980 *T)
    sumR1  +=       0.538 * cos(5.54081539805  + 553.56940284240 *T)
    sumR1  +=       0.402 * cos(2.16544019233  + 7860.41939243920 *T)
    sumR1  +=       0.553 * cos(2.32177369366  + 11506.76976979360 *T)
    sumR1  +=       0.367 * cos(3.39152532250  + 6496.37494542940 *T)
    sumR1  +=       0.360 * cos(5.34379853282  + 7079.37385680780 *T)
    sumR1  +=       0.337 * cos(3.61563704045  + 11790.62908865880 *T)
    sumR1  +=       0.456 * cos(0.30754294809  + 801.82093112380 *T)
    sumR1  +=       0.417 * cos(3.70009308674  + 10575.40668294180 *T)
    sumR1  +=       0.381 * cos(5.82033971802  + 7058.59846131540 *T)
    sumR1  +=       0.321 * cos(0.31988767355  + 16200.77272450120 *T)
    sumR1  +=       0.364 * cos(1.08414306177  + 6309.37416979120 *T)
    sumR1  +=       0.294 * cos(4.54798604957  + 11856.21865142450 *T)
    sumR1  +=       0.290 * cos(1.26473978562  + 8635.94200376320 *T)
    sumR1  +=       0.399 * cos(4.16998866302  + 26.29831979980 *T)
    sumR1  +=       0.262 * cos(5.08316906342  + 10177.25767953360 *T)
    sumR1  +=       0.243 * cos(2.25746091190  + 11712.95531823080 *T)
    sumR1  +=       0.237 * cos(1.05070575346  + 242.72860397400 *T)
    sumR1  +=       0.275 * cos(3.45319481756  + 5884.92684658320 *T)
    sumR1  +=       0.255 * cos(5.38496831087  + 21228.39202354580 *T)
    sumR1  +=       0.307 * cos(4.24313526604  + 3738.76143010800 *T)
    sumR1  +=       0.216 * cos(3.46037894728  + 213.29909543800 *T)
    sumR1  +=       0.196 * cos(0.69029243914  + 1990.74501704100 *T)
    sumR1  +=       0.198 * cos(5.16301829964  + 12352.85260454480 *T)
    sumR1  +=       0.214 * cos(3.91876200279  + 13916.01910964160 *T)
    sumR1  +=       0.212 * cos(4.00861198517  + 5230.80746680300 *T)
    sumR1  +=       0.184 * cos(5.59805976614  + 6283.14316029419 *T)
    sumR1  +=       0.184 * cos(2.85275392124  + 7238.67559160000 *T)
    sumR1  +=       0.179 * cos(2.54259058334  + 14314.16811304980 *T)
    sumR1  +=       0.225 * cos(1.64458698399  + 4732.03062734340 *T)
    sumR1  +=       0.236 * cos(5.58826125715  + 6069.77675455340 *T)
    sumR1  +=       0.187 * cos(2.72805985443  + 6062.66320755260 *T)
    sumR1  +=       0.184 * cos(6.04216273598  + 6283.00853968860 *T)
    sumR1  +=       0.230 * cos(3.62591335086  + 6284.05617105960 *T)
    sumR1  +=       0.163 * cos(2.19117396803  + 18073.70493865020 *T)
    sumR1  +=       0.172 * cos(0.97612950740  + 3930.20969621960 *T)
    sumR1  +=       0.215 * cos(1.04672844028  + 3496.03282613400 *T)
    sumR1  +=       0.169 * cos(4.75084479006  + 17267.26820169119 *T)
    sumR1  +=       0.152 * cos(0.19390712179  + 9779.10867612540 *T)
    sumR1  +=       0.182 * cos(5.16288118255  + 17253.04110768959 *T)
    sumR1  +=       0.149 * cos(0.80944184260  + 709.93304855830 *T)
    sumR1  +=       0.163 * cos(2.19209570390  + 6076.89030155420 *T)
    sumR1  +=       0.186 * cos(5.01159497089  + 11015.10647733480 *T)
    sumR1  +=       0.134 * cos(0.97765485759  + 65147.61976813770 *T)
    sumR1  +=       0.141 * cos(4.38421981312  + 4136.91043351620 *T)
    sumR1  +=       0.158 * cos(4.60974280627  + 9623.68827669120 *T)
    sumR1  +=       0.133 * cos(3.30508592837  + 154717.60988768269 *T)
    sumR1  +=       0.163 * cos(6.11782626245  + 3.52311834900 *T)
    sumR1  +=       0.174 * cos(1.58078542187  + 7.11354700080 *T)
    sumR1  +=       0.141 * cos(0.49976927274  + 25158.60171976540 *T)
    sumR1  +=       0.124 * cos(6.03440460031  + 9225.53927328300 *T)
    sumR1  +=       0.150 * cos(5.30166336812  + 13517.87010623340 *T)
    sumR1  +=       0.127 * cos(1.92389511438  + 22483.84857449259 *T)
    sumR1  +=       0.121 * cos(2.37813129011  + 167283.76158766547 *T)
    sumR1  +=       0.120 * cos(3.98423684853  + 4686.88940770680 *T)
    sumR1  +=       0.117 * cos(5.81072642211  + 12569.67481833180 *T)
    sumR1  +=       0.122 * cos(5.60973054224  + 5642.19824260920 *T)
    sumR1  +=       0.157 * cos(3.40236426002  + 16496.36139620240 *T)
    sumR1  +=       0.129 * cos(2.10705116371  + 1589.07289528380 *T)
    sumR1  +=       0.116 * cos(0.55839966736  + 5849.36411211460 *T)
    sumR1  +=       0.123 * cos(1.52961392771  + 12559.03815298200 *T)
    sumR1  +=       0.111 * cos(0.44848279675  + 6172.86952877200 *T)
    sumR1  +=       0.123 * cos(5.81645568991  + 6282.09552892320 *T)
    sumR1  +=       0.150 * cos(4.26278409223  + 3128.38876509580 *T)
    sumR1  +=       0.106 * cos(2.27437761356  + 5429.87946823940 *T)
    sumR1  +=       0.104 * cos(4.42743707728  + 23543.23050468179 *T)
    sumR1  +=       0.121 * cos(0.39459045915  + 12132.43996210600 *T)
    sumR1  +=       0.104 * cos(2.41842602527  + 426.59819087600 *T)
    sumR1  +=       0.110 * cos(5.80381480447  + 16858.48253293320 *T)
    sumR1  +=       0.100 * cos(2.93805577485  + 4535.05943692440 *T)
    sumR1  +=       0.097 * cos(3.97935904984  + 6133.51265285680 *T)
    sumR1  +=       0.110 * cos(6.22339014386  + 12146.66705610760 *T)
    sumR1  +=       0.098 * cos(0.87576563709  + 6525.80445396540 *T)
    sumR1  +=       0.098 * cos(3.15248421301  + 10440.27429260360 *T)
    sumR1  +=       0.095 * cos(2.46168411100  + 3097.88382272579 *T)
    sumR1  +=       0.088 * cos(0.23371480284  + 13119.72110282519 *T)
    sumR1  +=       0.098 * cos(5.77016493489  + 7342.45778018060 *T)
    sumR1  +=       0.092 * cos(6.03915555063  + 20426.57109242200 *T)
    sumR1  +=       0.096 * cos(5.56909292561  + 2388.89402044920 *T)
    sumR1  +=       0.081 * cos(1.32131147691  + 5650.29211067820 *T)
    sumR1  +=       0.086 * cos(3.94529200528  + 10454.50138660520 *T)
    sumR1  +=       0.076 * cos(2.70729716925  + 143571.32428481648 *T)
    sumR1  +=       0.091 * cos(5.64100034152  + 8827.39026987480 *T)
    sumR1  +=       0.076 * cos(1.80783856698  + 28286.99048486120 *T)
    sumR1  +=       0.081 * cos(1.90858992196  + 29088.81141598500 *T)
    sumR1  +=       0.075 * cos(3.40955892978  + 5481.25491886760 *T)
    sumR1  +=       0.069 * cos(4.49936170873  + 17256.63153634140 *T)
    sumR1  +=       0.088 * cos(1.10098454357  + 11769.85369316640 *T)
    sumR1  +=       0.066 * cos(2.78285801977  + 536.80451209540 *T)
    sumR1  +=       0.068 * cos(3.88179770758  + 17260.15465469040 *T)
    sumR1  +=       0.084 * cos(1.59303306354  + 9380.95967271720 *T)
    sumR1  +=       0.088 * cos(3.88076636762  + 7477.52286021600 *T)
    sumR1  +=       0.061 * cos(6.17558202197  + 11087.28512591840 *T)
    sumR1  +=       0.060 * cos(4.34824715818  + 6206.80977871580 *T)
    sumR1  +=       0.082 * cos(4.59843208943  + 9388.00590941520 *T)
    sumR1  +=       0.079 * cos(1.63131230601  + 4933.20844033260 *T)
    sumR1  +=       0.078 * cos(4.20905757484  + 5729.50644714900 *T)
    sumR1  +=       0.057 * cos(5.48157926651  + 18319.53658487960 *T)
    sumR1  +=       0.060 * cos(1.01261781084  + 12721.57209941700 *T)
    sumR1  +=       0.056 * cos(1.63031935692  + 15720.83878487840 *T)
    sumR1  +=       0.055 * cos(0.24926735018  + 15110.46611986620 *T)
    sumR1  +=       0.061 * cos(5.93059279661  + 12539.85338018300 *T)
    sumR1  +=       0.055 * cos(4.84298966314  + 13095.84266507740 *T)
    sumR1  +=       0.067 * cos(6.11690589247  + 8662.24032356300 *T)
    sumR1  +=       0.054 * cos(5.73750638571  + 3634.62102451840 *T)
    sumR1  +=       0.074 * cos(1.05466745829  + 16460.33352952499 *T)
    sumR1  +=       0.053 * cos(2.29084335688  + 16062.18452611680 *T)
    sumR1  +=       0.064 * cos(2.13513767927  + 7875.67186362420 *T)
    sumR1  +=       0.067 * cos(0.07096807518  + 14945.31617355440 *T)
    sumR1  +=       0.051 * cos(2.31511194429  + 6262.72053059260 *T)
    sumR1  +=       0.057 * cos(5.77055471237  + 12043.57428188900 *T)
    sumR1  +=       0.056 * cos(4.41980790431  + 4701.11650170840 *T)
    sumR1  +=       0.059 * cos(5.87963500073  + 5331.35744374080 *T)
    sumR1  +=       0.058 * cos(2.30546168628  + 955.59974160860 *T)
    sumR1  +=       0.049 * cos(1.93839278478  + 5333.90024102160 *T)
    sumR1  +=       0.048 * cos(2.69973662261  + 6709.67404086740 *T)
    sumR1  +=       0.064 * cos(1.64379897981  + 6262.30045449900 *T)
    sumR1  +=       0.046 * cos(3.98449608961  + 98068.53671630540 *T)
    sumR1  +=       0.050 * cos(3.68875893005  + 12323.42309600880 *T)
    sumR1  +=       0.045 * cos(3.30068569697  + 22003.91463486980 *T)
    sumR1  +=       0.047 * cos(1.26317154881  + 11919.14086666800 *T)
    sumR1  +=       0.045 * cos(0.89150445122  + 51868.24866217880 *T)
    sumR1  +=       0.043 * cos(1.61526242998  + 6277.55292568400 *T)
    sumR1  +=       0.043 * cos(5.74295325645  + 11403.67699557500 *T)
    sumR1  +=       0.044 * cos(3.43070646822  + 10021.83728009940 *T)
    sumR1  +=       0.056 * cos(0.02481833774  + 15671.08175940660 *T)
    sumR1  +=       0.055 * cos(3.14274403422  + 33019.02111220460 *T)
    sumR1  +=       0.045 * cos(3.00877289177  + 8982.81066930900 *T)
    sumR1  +=       0.046 * cos(0.73303568429  + 6303.43116939020 *T)
    sumR1  +=       0.049 * cos(1.60455690285  + 6303.85124548380 *T)
    sumR1  +=       0.045 * cos(0.40210030323  + 6805.65326808520 *T)
    sumR1  +=       0.053 * cos(0.94869680175  + 10988.80815753500 *T)
    sumR1  +=       0.041 * cos(1.61122384329  + 6819.88036208680 *T)
    sumR1  +=       0.055 * cos(0.89439119424  + 11933.36796066960 *T)
    sumR1  +=       0.045 * cos(3.88495384656  + 60530.48898574180 *T)
    sumR1  +=       0.040 * cos(4.75740908001  + 38526.57435087200 *T)
    sumR1  +=       0.040 * cos(1.49921251887  + 18451.07854656599 *T)
    sumR1  +=       0.040 * cos(3.77498297228  + 26087.90314157420 *T)
    sumR1  +=       0.051 * cos(1.70258603562  + 1551.04522264800 *T)
    sumR1  +=       0.039 * cos(2.97100699926  + 2118.76386037840 *T)
    sumR1  +=       0.053 * cos(5.19854123078  + 77713.77146812050 *T)
    sumR1  +=       0.047 * cos(4.26356628717  + 21424.46664430340 *T)
    sumR1  +=       0.037 * cos(0.62902722802  + 24356.78078864160 *T)
    sumR1  +=       0.036 * cos(0.11087914947  + 10344.29506538580 *T)
    sumR1  +=       0.036 * cos(0.77037556319  + 12029.34718788740 *T)
    sumR1  +=       0.035 * cos(3.30933994515  + 24072.92146977640 *T)
    sumR1  +=       0.035 * cos(5.93650887012  + 31570.79964939120 *T)
    sumR1  +=       0.036 * cos(2.15108874765  + 30774.50164257480 *T)
    sumR1  +=       0.036 * cos(1.75078825382  + 16207.88627150200 *T)
    sumR1  +=       0.033 * cos(5.06264177921  + 226858.23855437007 *T)
    sumR1  +=       0.034 * cos(6.16891378800  + 24491.42579258340 *T)
    sumR1  +=       0.035 * cos(3.19120695549  + 32217.20018108080 *T)
    sumR1  +=       0.034 * cos(2.31528650443  + 55798.45835839840 *T)
    sumR1  +=       0.032 * cos(4.21446357042  + 15664.03552270859 *T)
    sumR1  +=       0.039 * cos(1.24979117796  + 6418.14093002680 *T)
    sumR1  +=       0.037 * cos(4.11943655770  + 2787.04302385740 *T)
    sumR1  +=       0.032 * cos(1.62887710890  + 639.89728631400 *T)
    sumR1  +=       0.038 * cos(5.89832942685  + 640.87760738220 *T)
    sumR1  +=       0.032 * cos(1.72442327688  + 27433.88921587499 *T)
    sumR1  +=       0.031 * cos(2.78828943753  + 12139.55350910680 *T)
    sumR1  +=       0.035 * cos(4.44608896525  + 18202.21671665939 *T)
    sumR1  +=       0.034 * cos(3.96287980676  + 18216.44381066100 *T)
    sumR1  +=       0.033 * cos(4.73611335874  + 16723.35014259500 *T)
    sumR1  +=       0.034 * cos(1.43910280005  + 49515.38250840700 *T)
    sumR1  +=       0.031 * cos(0.23302920161  + 23581.25817731760 *T)
    sumR1  +=       0.029 * cos(2.02633840220  + 11609.86254401220 *T)
    sumR1  +=       0.030 * cos(2.54923230240  + 9924.81042151060 *T)
    sumR1  +=       0.032 * cos(4.91793198558  + 11300.58422135640 *T)
    sumR1  +=       0.028 * cos(0.26187189577  + 13521.75144159140 *T)
    sumR1  +=       0.028 * cos(3.84568936822  + 2699.73481931760 *T)
    sumR1  +=       0.029 * cos(1.83149729794  + 29822.78323632420 *T)
    sumR1  +=       0.033 * cos(4.60320094415  + 19004.64794940840 *T)
    sumR1  +=       0.027 * cos(4.46183450287  + 6702.56049386660 *T)
    sumR1  +=       0.030 * cos(4.46494072240  + 36147.40987730040 *T)
    sumR1  +=       0.027 * cos(0.03211931363  + 6279.78949257360 *T)
    sumR1  +=       0.026 * cos(5.46497324333  + 6245.04817735560 *T)
    sumR1  +=       0.035 * cos(4.52695674113  + 36949.23080842420 *T)
    sumR1  +=       0.027 * cos(3.52528177609  + 10770.89325626180 *T)
    sumR1  +=       0.026 * cos(1.48499438453  + 11080.17157891760 *T)
    sumR1  +=       0.035 * cos(2.82154380962  + 19402.79695281660 *T)
    sumR1  +=       0.025 * cos(2.46339998836  + 6279.48542133960 *T)
    sumR1  +=       0.026 * cos(4.97688894643  + 16737.57723659660 *T)
    sumR1  +=       0.026 * cos(2.36136541526  + 17996.03116822220 *T)
    sumR1  +=       0.029 * cos(4.15148654061  + 45892.73043315699 *T)
    sumR1  +=       0.026 * cos(4.50714272714  + 17796.95916678580 *T)
    sumR1  +=       0.027 * cos(4.72625223674  + 1066.49547719000 *T)
    sumR1  +=       0.025 * cos(2.89309528854  + 6286.66627864320 *T)
    sumR1  +=       0.027 * cos(0.37462444357  + 12964.30070339100 *T)
    sumR1  +=       0.029 * cos(4.94860010533  + 5863.59120611620 *T)
    sumR1  +=       0.031 * cos(3.93096113577  + 29864.33402730900 *T)
    sumR1  +=       0.024 * cos(6.14987193584  + 18606.49894600020 *T)
    sumR1  +=       0.024 * cos(3.74225964547  + 29026.48522950779 *T)
    sumR1  +=       0.025 * cos(5.70460621565  + 27707.54249429480 *T)
    sumR1  +=       0.025 * cos(5.33928840652  + 15141.39079431200 *T)
    sumR1  +=       0.027 * cos(3.02320897140  + 6286.36220740920 *T)
    sumR1  +=       0.023 * cos(0.28364955406  + 5327.47610838280 *T)
    sumR1  +=       0.026 * cos(1.34240461687  + 18875.52586977400 *T)
    sumR1  +=       0.024 * cos(1.33998410121  + 19800.94595622480 *T)
    sumR1  +=       0.025 * cos(6.00172494004  + 6489.26139842860 *T)
    sumR1  +=       0.022 * cos(1.81777974484  + 6288.59877429880 *T)
    sumR1  +=       0.022 * cos(3.58603606640  + 6915.85958930460 *T)
    sumR1  +=       0.029 * cos(2.09564449439  + 15265.88651930040 *T)
    sumR1  +=       0.022 * cos(1.02173599251  + 11925.27409260060 *T)
    sumR1  +=       0.022 * cos(4.74660932338  + 28230.18722269139 *T)
    sumR1  +=       0.021 * cos(2.30688751432  + 5999.21653112620 *T)
    sumR1  +=       0.021 * cos(3.22654944430  + 25934.12433108940 *T)
    sumR1  +=       0.021 * cos(3.04956726238  + 6566.93516885660 *T)
    sumR1  +=       0.027 * cos(5.35653084499  + 33794.54372352860 *T)
    sumR1  +=       0.028 * cos(3.91168324815  + 18208.34994259200 *T)
    sumR1  +=       0.020 * cos(1.52296293311  + 135.06508003540 *T)
    sumR1  +=       0.022 * cos(4.66462839521  + 13362.44970679920 *T)
    sumR1  +=       0.019 * cos(1.78121167862  + 156137.47598479927 *T)
    sumR1  +=       0.019 * cos(2.99969102221  + 19651.04848109800 *T)
    sumR1  +=       0.019 * cos(2.86664273362  + 18422.62935909819 *T)
    sumR1  +=       0.025 * cos(0.94995632141  + 31415.37924995700 *T)
    sumR1  +=       0.019 * cos(4.71432851499  + 77690.75950573849 *T)
    sumR1  +=       0.019 * cos(2.54227398241  + 77736.78343050249 *T)
    sumR1  +=       0.020 * cos(5.91915117116  + 48739.85989708300 *T)

    # EARTH Radius
    sumR2  = 0
    sumR2  +=    4359.385 * cos(5.78455133738  + 6283.07584999140 *T)
    sumR2  +=     123.633 * cos(5.57934722157  + 12566.15169998280 *T)
    sumR2  +=      12.341 * cos(3.14159265359  + 0.00000000000 *T)
    sumR2  +=       8.792 * cos(3.62777733395  + 77713.77146812050 *T)
    sumR2  +=       5.689 * cos(1.86958905084  + 5573.14280143310 *T)
    sumR2  +=       3.301 * cos(5.47027913302  + 18849.22754997420 *T)
    sumR2  +=       1.471 * cos(4.48028885617  + 5507.55323866740 *T)
    sumR2  +=       1.013 * cos(2.81456417694  + 5223.69391980220 *T)
    sumR2  +=       0.854 * cos(3.10878241236  + 1577.34354244780 *T)
    sumR2  +=       1.102 * cos(2.84173992403  + 161000.68573767410 *T)
    sumR2  +=       0.648 * cos(5.47349498544  + 775.52261132400 *T)
    sumR2  +=       0.609 * cos(1.37969434104  + 6438.49624942560 *T)
    sumR2  +=       0.499 * cos(4.41649242250  + 6286.59896834040 *T)
    sumR2  +=       0.417 * cos(0.90242451175  + 10977.07880469900 *T)
    sumR2  +=       0.402 * cos(3.20376585290  + 5088.62883976680 *T)
    sumR2  +=       0.351 * cos(1.81079227770  + 5486.77784317500 *T)
    sumR2  +=       0.467 * cos(3.65753702738  + 7084.89678111520 *T)
    sumR2  +=       0.458 * cos(5.38585314743  + 149854.40013480789 *T)
    sumR2  +=       0.304 * cos(3.51701098693  + 796.29800681640 *T)
    sumR2  +=       0.266 * cos(6.17413982699  + 6836.64525283380 *T)
    sumR2  +=       0.279 * cos(1.84120501086  + 4694.00295470760 *T)
    sumR2  +=       0.260 * cos(1.41629543251  + 2146.16541647520 *T)
    sumR2  +=       0.266 * cos(3.13832905677  + 71430.69561812909 *T)
    sumR2  +=       0.321 * cos(5.35313367048  + 3154.68708489560 *T)
    sumR2  +=       0.238 * cos(2.17720020018  + 155.42039943420 *T)
    sumR2  +=       0.293 * cos(4.61501268144  + 4690.47983635860 *T)
    sumR2  +=       0.229 * cos(4.75969588070  + 7234.79425624200 *T)
    sumR2  +=       0.211 * cos(0.21868065485  + 4705.73230754360 *T)
    sumR2  +=       0.201 * cos(4.21905743357  + 1349.86740965880 *T)
    sumR2  +=       0.195 * cos(4.57808285364  + 529.69096509460 *T)
    sumR2  +=       0.253 * cos(2.81496293039  + 1748.01641306700 *T)
    sumR2  +=       0.182 * cos(5.70454011389  + 6040.34724601740 *T)
    sumR2  +=       0.179 * cos(6.02897097053  + 4292.33083295040 *T)
    sumR2  +=       0.186 * cos(1.58690991244  + 6309.37416979120 *T)
    sumR2  +=       0.170 * cos(2.90220009715  + 9437.76293488700 *T)
    sumR2  +=       0.166 * cos(1.99984925026  + 8031.09226305840 *T)
    sumR2  +=       0.158 * cos(0.04783713552  + 2544.31441988340 *T)
    sumR2  +=       0.197 * cos(2.01083639502  + 1194.44701022460 *T)
    sumR2  +=       0.165 * cos(5.78372596778  + 83996.84731811189 *T)
    sumR2  +=       0.214 * cos(3.38285934319  + 7632.94325965020 *T)
    sumR2  +=       0.140 * cos(0.36401486094  + 10447.38783960440 *T)
    sumR2  +=       0.151 * cos(0.95153163031  + 6127.65545055720 *T)
    sumR2  +=       0.136 * cos(1.48426306582  + 2352.86615377180 *T)
    sumR2  +=       0.127 * cos(5.48475435134  + 951.71840625060 *T)
    sumR2  +=       0.126 * cos(5.26866506592  + 6279.55273164240 *T)
    sumR2  +=       0.125 * cos(3.75754889288  + 6812.76681508600 *T)
    sumR2  +=       0.101 * cos(4.95015746147  + 398.14900340820 *T)
    sumR2  +=       0.102 * cos(0.68468295277  + 1592.59601363280 *T)
    sumR2  +=       0.100 * cos(1.14568935785  + 3894.18182954220 *T)
    sumR2  +=       0.129 * cos(0.76540016965  + 553.56940284240 *T)
    sumR2  +=       0.109 * cos(5.41063597567  + 6256.77753019160 *T)
    sumR2  +=       0.075 * cos(5.84804322893  + 242.72860397400 *T)
    sumR2  +=       0.095 * cos(1.94452244083  + 11856.21865142450 *T)
    sumR2  +=       0.077 * cos(0.69373708195  + 8429.24126646660 *T)
    sumR2  +=       0.100 * cos(5.19725292131  + 244287.60000722768 *T)
    sumR2  +=       0.080 * cos(6.18440483705  + 1059.38193018920 *T)
    sumR2  +=       0.069 * cos(5.25699888595  + 14143.49524243060 *T)
    sumR2  +=       0.085 * cos(5.39484725499  + 25132.30339996560 *T)
    sumR2  +=       0.066 * cos(0.51779993906  + 801.82093112380 *T)
    sumR2  +=       0.055 * cos(5.16878202461  + 7058.59846131540 *T)
    sumR2  +=       0.051 * cos(3.88759155247  + 12036.46073488820 *T)
    sumR2  +=       0.050 * cos(5.57636570536  + 6290.18939699220 *T)
    sumR2  +=       0.061 * cos(2.24359003264  + 8635.94200376320 *T)
    sumR2  +=       0.050 * cos(5.54441900966  + 1990.74501704100 *T)
    sumR2  +=       0.056 * cos(4.00301078040  + 13367.97263110660 *T)
    sumR2  +=       0.052 * cos(4.13138898038  + 7860.41939243920 *T)
    sumR2  +=       0.052 * cos(3.90943054011  + 26.29831979980 *T)
    sumR2  +=       0.041 * cos(3.57128482780  + 7079.37385680780 *T)
    sumR2  +=       0.056 * cos(2.76959005761  + 90955.55169449610 *T)
    sumR2  +=       0.042 * cos(1.91461189199  + 7477.52286021600 *T)
    sumR2  +=       0.042 * cos(0.42728171713  + 10213.28554621100 *T)
    sumR2  +=       0.042 * cos(1.09413724455  + 709.93304855830 *T)
    sumR2  +=       0.039 * cos(3.93298068961  + 10973.55568635000 *T)
    sumR2  +=       0.038 * cos(6.17935925345  + 9917.69687450980 *T)
    sumR2  +=       0.049 * cos(0.83021145241  + 11506.76976979360 *T)
    sumR2  +=       0.053 * cos(1.45828359397  + 233141.31440436147 *T)
    sumR2  +=       0.047 * cos(6.21568666789  + 6681.22485339960 *T)
    sumR2  +=       0.037 * cos(0.36359309980  + 10177.25767953360 *T)
    sumR2  +=       0.035 * cos(3.33024911524  + 5643.17856367740 *T)
    sumR2  +=       0.034 * cos(5.63446915337  + 6525.80445396540 *T)
    sumR2  +=       0.035 * cos(5.36033855038  + 25158.60171976540 *T)
    sumR2  +=       0.034 * cos(5.36319798321  + 4933.20844033260 *T)
    sumR2  +=       0.033 * cos(4.24722336872  + 12569.67481833180 *T)
    sumR2  +=       0.043 * cos(5.26370903404  + 10575.40668294180 *T)
    sumR2  +=       0.042 * cos(5.08837645072  + 11015.10647733480 *T)
    sumR2  +=       0.040 * cos(1.98334703186  + 6284.05617105960 *T)
    sumR2  +=       0.042 * cos(4.22496037505  + 88860.05707098669 *T)
    sumR2  +=       0.029 * cos(3.19088628170  + 11926.25441366880 *T)
    sumR2  +=       0.029 * cos(0.15217616684  + 12168.00269657460 *T)
    sumR2  +=       0.030 * cos(1.61904744136  + 9779.10867612540 *T)
    sumR2  +=       0.027 * cos(0.76388991416  + 1589.07289528380 *T)
    sumR2  +=       0.036 * cos(2.74712003443  + 3738.76143010800 *T)
    sumR2  +=       0.033 * cos(3.08807829566  + 3930.20969621960 *T)
    sumR2  +=       0.031 * cos(5.34906619513  + 143571.32428481648 *T)
    sumR2  +=       0.025 * cos(0.10240267494  + 22483.84857449259 *T)
    sumR2  +=       0.030 * cos(3.47110495524  + 14945.31617355440 *T)
    sumR2  +=       0.024 * cos(1.10425016019  + 4535.05943692440 *T)
    sumR2  +=       0.024 * cos(1.58037259780  + 6496.37494542940 *T)
    sumR2  +=       0.023 * cos(3.87710321433  + 6275.96230299060 *T)
    sumR2  +=       0.025 * cos(3.94529778970  + 3128.38876509580 *T)
    sumR2  +=       0.023 * cos(3.44685609601  + 4136.91043351620 *T)
    sumR2  +=       0.023 * cos(3.83156029849  + 5753.38488489680 *T)
    sumR2  +=       0.022 * cos(1.86956128067  + 16730.46368959580 *T)
    sumR2  +=       0.025 * cos(2.42188933855  + 5729.50644714900 *T)
    sumR2  +=       0.020 * cos(1.78208352927  + 17789.84561978500 *T)
    sumR2  +=       0.021 * cos(4.30363087400  + 16858.48253293320 *T)
    sumR2  +=       0.021 * cos(0.49258939822  + 29088.81141598500 *T)
    sumR2  +=       0.025 * cos(1.33030250444  + 6282.09552892320 *T)
    sumR2  +=       0.027 * cos(2.54785812264  + 3496.03282613400 *T)
    sumR2  +=       0.022 * cos(1.11232521950  + 12721.57209941700 *T)
    sumR2  +=       0.021 * cos(5.97759081637  + 7.11354700080 *T)
    sumR2  +=       0.019 * cos(0.80292033311  + 16062.18452611680 *T)
    sumR2  +=       0.023 * cos(4.12454848769  + 2388.89402044920 *T)
    sumR2  +=       0.022 * cos(4.92663152168  + 18875.52586977400 *T)
    sumR2  +=       0.023 * cos(5.68902059771  + 16460.33352952499 *T)
    sumR2  +=       0.023 * cos(4.97346265647  + 17260.15465469040 *T)
    sumR2  +=       0.023 * cos(3.03021283729  + 66567.48586525429 *T)
    sumR2  +=       0.016 * cos(3.89740925257  + 5331.35744374080 *T)
    sumR2  +=       0.017 * cos(3.08268671348  + 154717.60988768269 *T)
    sumR2  +=       0.016 * cos(3.95085099736  + 3097.88382272579 *T)
    sumR2  +=       0.016 * cos(3.99041783945  + 6283.14316029419 *T)
    sumR2  +=       0.020 * cos(6.10644140189  + 167283.76158766547 *T)
    sumR2  +=       0.015 * cos(4.09775914607  + 11712.95531823080 *T)
    sumR2  +=       0.016 * cos(5.71769940700  + 17298.18232732620 *T)
    sumR2  +=       0.016 * cos(3.28894009404  + 5884.92684658320 *T)
    sumR2  +=       0.015 * cos(5.64785377164  + 12559.03815298200 *T)
    sumR2  +=       0.016 * cos(4.43452080930  + 6283.00853968860 *T)
    sumR2  +=       0.014 * cos(2.31721603062  + 5481.25491886760 *T)
    sumR2  +=       0.014 * cos(4.43479032305  + 13517.87010623340 *T)
    sumR2  +=       0.014 * cos(4.73209312936  + 7342.45778018060 *T)
    sumR2  +=       0.012 * cos(0.64705975463  + 18073.70493865020 *T)
    sumR2  +=       0.011 * cos(1.51443332200  + 16200.77272450120 *T)
    sumR2  +=       0.011 * cos(0.88708889185  + 21228.39202354580 *T)
    sumR2  +=       0.014 * cos(4.50116508534  + 640.87760738220 *T)
    sumR2  +=       0.011 * cos(4.64339996198  + 11790.62908865880 *T)
    sumR2  +=       0.011 * cos(1.31064298246  + 4164.31198961300 *T)
    sumR2  +=       0.009 * cos(3.02238989305  + 23543.23050468179 *T)
    sumR2  +=       0.009 * cos(2.04999402381  + 22003.91463486980 *T)
    sumR2  +=       0.009 * cos(4.91488110218  + 213.29909543800 *T)

    # EARTH Radius
    sumR3  = 0
    sumR3  +=     144.595 * cos(4.27319435148  + 6283.07584999140 *T)
    sumR3  +=       6.729 * cos(3.91697608662  + 12566.15169998280 *T)
    sumR3  +=       0.774 * cos(0.00000000000  + 0.00000000000 *T)
    sumR3  +=       0.247 * cos(3.73019298781  + 18849.22754997420 *T)
    sumR3  +=       0.036 * cos(2.80081409050  + 6286.59896834040 *T)
    sumR3  +=       0.033 * cos(5.62216602775  + 6127.65545055720 *T)
    sumR3  +=       0.019 * cos(3.71292621802  + 6438.49624942560 *T)
    sumR3  +=       0.016 * cos(4.26011484232  + 6525.80445396540 *T)
    sumR3  +=       0.016 * cos(3.50416887054  + 6256.77753019160 *T)
    sumR3  +=       0.014 * cos(3.62127621114  + 25132.30339996560 *T)
    sumR3  +=       0.011 * cos(4.39200958819  + 4705.73230754360 *T)
    sumR3  +=       0.011 * cos(5.22327127059  + 6040.34724601740 *T)
    sumR3  +=       0.010 * cos(4.28045254647  + 83996.84731811189 *T)
    sumR3  +=       0.009 * cos(1.56864096494  + 5507.55323866740 *T)
    sumR3  +=       0.011 * cos(1.37795688024  + 6309.37416979120 *T)
    sumR3  +=       0.010 * cos(5.19937959068  + 71430.69561812909 *T)
    sumR3  +=       0.009 * cos(0.47275199930  + 6279.55273164240 *T)
    sumR3  +=       0.009 * cos(0.74642756529  + 5729.50644714900 *T)
    sumR3  +=       0.007 * cos(2.97374891560  + 775.52261132400 *T)
    sumR3  +=       0.007 * cos(3.28615691021  + 7058.59846131540 *T)
    sumR3  +=       0.007 * cos(2.19184402142  + 6812.76681508600 *T)
    sumR3  +=       0.005 * cos(3.15419034438  + 529.69096509460 *T)
    sumR3  +=       0.006 * cos(4.54725567047  + 1059.38193018920 *T)
    sumR3  +=       0.005 * cos(1.51104406936  + 7079.37385680780 *T)
    sumR3  +=       0.007 * cos(2.98052059053  + 6681.22485339960 *T)
    sumR3  +=       0.005 * cos(2.30961231391  + 12036.46073488820 *T)
    sumR3  +=       0.005 * cos(3.71102966917  + 6290.18939699220 *T)

    # EARTH Radius
    sumR4  = 0
    sumR4  +=       3.858 * cos(2.56384387339  + 6283.07584999140 *T)
    sumR4  +=       0.306 * cos(2.26769501230  + 12566.15169998280 *T)
    sumR4  +=       0.053 * cos(3.44031471924  + 5573.14280143310 *T)
    sumR4  +=       0.015 * cos(2.04794573436  + 18849.22754997420 *T)
    sumR4  +=       0.013 * cos(2.05688873673  + 77713.77146812050 *T)
    sumR4  +=       0.007 * cos(4.41218854480  + 161000.68573767410 *T)
    sumR4  +=       0.005 * cos(5.26154653107  + 6438.49624942560 *T)
    sumR4  +=       0.005 * cos(4.07695126049  + 6127.65545055720 *T)
    sumR4  +=       0.006 * cos(3.81514213664  + 149854.40013480789 *T)
    sumR4  +=       0.003 * cos(1.28175749811  + 6286.59896834040 *T)

    # EARTH Radius
    sumR5  = 0
    sumR5  +=       0.086 * cos(1.21579741687  + 6283.07584999140 *T)
    sumR5  +=       0.012 * cos(0.65617264033  + 12566.15169998280 *T)
    sumR5  +=       0.001 * cos(0.38068797142  + 18849.22754997420 *T)

    # calculate and scale the sums
    L = sumL0 + sumL1*T**1 + sumL2*T**2 + sumL3*T**3 + sumL4*T**4 + sumL5*T**5
    L /= 1e8
    L = Mod(degrees(L), 360.)
    B = sumB0 + sumB1*T**1 + sumB2*T**2 + sumB3*T**3 + sumB4*T**4
    B /= 1e8
    B = degrees(B)
    R = sumR0 + sumR1*T**1 + sumR2*T**2 + sumR3*T**3 + sumR4*T**4 + sumR5*T**5
    R /= 1e8

    # calculate X, Y, Z
    X = R * Cos(B) * Cos(L)
    Y = R * Cos(B) * Sin(L)
    Z = R * Sin(B)

    # create dictionary
    Dict = {  "Object"  : "Earth",
              "Mode"    : "LBR",
              "Long"    : L,
              "Lat"     : B,
              "Rau"     : R,
              "X"       : X,
              "Y"       : Y,
              "Z"       : Z,
              "Alt"     : -34./60.
           }
    Data = (L, B, R, Dict)


    # save parameters and values
    EarthdtUTimes[strdtUT] = Data


    return Data
#######################################################################



def getEarthData(dtCivicTime, zone, verbose = False):
    L, B, R, Dict = getEarthLBRData(dtCivicTime, zone, verbose)
    X = Dict["X"]; Y = Dict["Y"]; Z = Dict["Z"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Le, Be, Re, Dicte = getEarthLBRData(dtCivicTime, zone, verbose)
        Xe = Dicte["X"]; Ye = Dicte["Y"]; Ze = Dicte["Z"]
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#######################################################################
#
#   getSunData, based on Meeus, chapter 24
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   SunRA       Rektaszension in degrees
#   SunDecl     Declination above equator in degrees
#   SunDict     Dictionary with Sun-specific output values:
#       Object:     "Sun"
#       Ltrue       True (ecliptic) Longitude in degrees
#       Lattrue     True ecliptic latitude, always 0.0
#       R           Distance Earth - Sun in km
#       epsilon     Schiefe der Ekliptik (oblique) in degrees
#       Alt         Altitude at rise, set (degrees)

Version = "V4.1 hvf 2022-05-19 getSunData.py"
VersionList.append(Version)

def getSunData(dtCivicTime, zone, verbose=False):
    # use VSOP87 data for Earth
    # use getNutation to include nutation
    deltaLong, deltaLat, epsilon = getNutation(dtCivicTime, zone, False)    
    # we do not need the nutation effects, just epsilon
    deltaLong = 0; deltaLat = 0

    # ecliptic latitude of Earth, chapter 31 - use negative value for Sun,
    # add 180 to longitude
    # Meeus p. 154
    dummy, dummy, Dict = getEarthData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    SunLat = -B + deltaLat
    Ltrue = Mod(L + deltaLong + 180., 360.)
    Rau = Dict["Rau"]
    Rabs = Rau * 149597870.700    # km

    # Rektaszension, Deklination
    SunRA = atan2(Cos(epsilon)*Sin(Ltrue) - Tan(SunLat)*Sin(epsilon), \
            Cos(Ltrue))
    SunDecl = asin(Sin(SunLat)*Cos(epsilon) \
            + Cos(SunLat)*Sin(epsilon)*Sin(Ltrue))
    SunRA = Mod(degrees(SunRA), 360.)
    SunDecl = degrees(SunDecl)

    # altitude at Sunrise, Sunset (degrees), standard value
    Alt = -50./60.    # degrees, 50'

    if (verbose):
        print ("SunData:")
        print ("Rau = ", Rau)
        print ("Rabs = ", Rabs)
        print ("epsilon = ", epsilon)
        print ("SunRA = ", SunRA)
        print ("SunDecl = ", SunDecl)
        print ("Alt = ", Alt)
#       Object:     "Sun"
#       SunLong     Ltrue   True (ecliptic) Longitude in degrees
#       SunLat      True ecliptic latitude, always 0.0
#       R           Rabs    Distance Earth - Sun in km
#       epsilon     Schiefe der Ekliptik (oblique) in degrees
#       Alt         Altitude at rise, set (degrees)
    SunDict = { "Object" : "Sun",
                "Long"  : Ltrue,
                "SunLong"  : Ltrue,
                "Lat"   : SunLat,
                "SunLat"   : SunLat,
                "RA"    : SunRA,
                "Decl"  : SunDecl,
                "R"       : Rabs,
                "Rau"     : Rau,
                "epsilon" : epsilon,
                "Alt"     : Alt,
                "X"     : 0,
                "Y"     : 0,
                "Z"     : 0,
                "eX"     : -Dict["X"],
                "eY"     : -Dict["Y"],
                "eZ"     : -Dict["Z"],
                "eRau"     : Rau,
                "eLong"  : Ltrue,
                "eLat"   : SunLat,
                }
    return (SunRA, SunDecl, SunDict)

############## End of getSunData() #####################################
########################################################################
#
#   getMoonData(dtCivicTime, zone, verbose=False):
#   Returns:    
#   MoonRA      Rektaszension in degrees
#   MoonDecl    Declination above equator in degrees
#   MoonDict    Dictionary with Moon-specific output values:
#       Object:     "Moon"
#       MoonLong    ecliptic longitude, degrees
#       MoonLat     ecliptic latitude, degrees
#       MoonDist    distance Earth - Moon, km
#       MoonDistAU  distance Earth - Moon, au
#       X           geocentric X, au
#       Y           geocentric Y, au
#       Z           geocentric Z, au
#       MoonHP      horizontal parallax, degrees
#       MoonNode    ecliptic longitude of ascending node, degrees
#       MoonPerigee longitude of perigee og Moon's orbit, degrees
#       Alt         Altitude at rise, set (degrees)

Version = "V4.1 hvf 2022-05-19 getMoonData.py"
VersionList.append(Version)

def getMoonData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.

    T = TT / 36525.    # eq. 21.1

    # obliquity of the ecliptic, eq. 21.2, degrees
    epsilon = 23.+26./60+21.448/3600. - 46.8150/3600.*T \
        - 0.00059/3600.*T*T + 0.001813/3600.*T*T*T

    # Moon mean longitude (mean equinox of the date, incl, effect of light-time)
    # eq. 45.1
    Lprime  = 218.3164591 + T * (481267.88134236 + T * (-0.0013268 \
            + T * (1./538841. - T / 65194000)))

    # Mean elongation of the Moon
    # eq. 45.2
    D       = 297.8502042 + T * (445267.1115168 + T * (-0.0016300 \
            + T * (1./545868.  -T / 113065000.)))

    # Sun's mean anomaly
    # eq. 45.3
    M       = 357.5291092 + T * (35999.0502909 + T * (-0.0001536 \
            + T / 24490000))

    # Moon's mean anomaly
    # eq 45.4
    MM      = 134.9634114 + T * (477198.8676313 + T * (0.0089970 \
            + T * (1./69699.  - T / 14712000.)))

    # Moon's argument of latitude (mean distance of the Moon
    # from its ascending node)
    # eq 45.5
    F       = 93.2720993 + T * (483202.0175273 + T * (-0.0034029 \
            + T * (-1./3526000. + T / 863310000.)))

    # three further arguments are needed
    A1      = 119.75 + 131.849 * T          # action of Venus
    A2      =  53.09 + 479264.290 * T       # action of Jupiter
    A3      = 313.45 + 481266.484 * T

    # effect of decreasing eccentricity of Earth's orbit
    E       = 1. + T * (-0.002516 + T * -0.0000074)
    EE      = E*E

    # Longitude, distance: now calculate sums sigL, sigR
    # see Table 45.A, p.309
    arg = MM;           sigL = 6288774*Sin(arg);    sigR = -20905355*Cos(arg)
    arg = 2*D - MM;     sigL += 1274027*Sin(arg);   sigR += -3699111*Cos(arg)
    arg = 2*D;          sigL += 658314*Sin(arg);    sigR += -2955968*Cos(arg)
    arg = 2*MM;         sigL += 213618*Sin(arg);    sigR += -569925*Cos(arg)  
    arg = M;            sigL += -185116*Sin(arg)*E; sigR += 48888*Cos(arg)*E
    arg = 2*F;          sigL += -114332*Sin(arg);   sigR += -3149*Cos(arg)
    arg = 2*D - 2*MM;   sigL += 58793*Sin(arg);     sigR += 246158*Cos(arg)
    arg = 2*D - M - MM; sigL += 57066*Sin(arg)*E;   sigR += -152138*Cos(arg)*E
    arg = 2*D + MM;     sigL += 53322*Sin(arg);     sigR += -170733*Cos(arg)
    arg = 2*D - M;      sigL += 45758*Sin(arg)*E;   sigR += -204586*Cos(arg)*E
    arg = M - MM;       sigL += -40923*Sin(arg)*E;  sigR += -129620*Cos(arg)*E
    arg = D;            sigL += -34720*Sin(arg);    sigR += 108743*Cos(arg)
    arg = M + MM;       sigL += -30383*Sin(arg)*E;  sigR +=  104755*Cos(arg)*E
    arg = 2*D - 2*F;    sigL += 15327*Sin(arg);     sigR += 10321*Cos(arg)
    arg = MM + 2*F;     sigL += -12528*Sin(arg)    
    arg = MM - 2*F;     sigL += 10980*Sin(arg);     sigR += 79661*Cos(arg)
    arg = 4*D-MM;       sigL += 10675*Sin(arg);     sigR += -34782*Cos(arg)
    arg = 3*MM;         sigL += 10034*Sin(arg);     sigR += -23210*Cos(arg)
    arg = 4*D-2*MM;     sigL +=  8548*Sin(arg);     sigR += -21636*Cos(arg)
    arg = 2*D + M - MM; sigL += -7888*Sin(arg)*E;   sigR += 24208*Cos(arg)*E
    arg = 2*D + M ;     sigL += -6766*Sin(arg)*E;   sigR += 30824*Cos(arg)*E
    arg = D - MM;       sigL += -5163*Sin(arg);     sigR += -8379*Cos(arg)
    arg = D + M;        sigL += 4987*Sin(arg)*E;    sigR += -16675*Cos(arg)*E
    arg = 2*D - M + MM; sigL += 4036*Sin(arg)*E;    sigR += -12831*Cos(arg)*E
    arg = 2*D + 2*MM;   sigL += 3994*Sin(arg);      sigR += -10445*Cos(arg)
    arg = 4*D;          sigL += 3861*Sin(arg);      sigR += -11650*Cos(arg)
    arg = 2*D - 3*MM;   sigL += 3665*Sin(arg);      sigR += 14403*Cos(arg)
    arg = M - 2*MM;     sigL += -2689*Sin(arg)*E;   sigR += -7003*Cos(arg)*E
    arg = 2*D - MM + 2*F; sigL += -2602*Sin(arg)    
    arg = 2*D - M - 2*MM; sigL += 2390*Sin(arg)*E;  sigR += 10056*Cos(arg)*E
    arg = D + MM;       sigL += -2348*Sin(arg);     sigR += 6322*Cos(arg)
    arg = 2*D - 2*M ;   sigL += 2236*Sin(arg)*EE;   sigR += -9884*Cos(arg)*EE
    arg = M + 2*MM;     sigL += -2120*Sin(arg)*E;   sigR += 5751*Cos(arg)*E
    arg = 2*M ;         sigL += -2069*Sin(arg)*EE
    arg = 2*D - 2*M - MM; sigL += 2048*Sin(arg)*EE; sigR += -4950*Cos(arg)*EE
    arg = 2*D + MM - 2*F;   sigL += -1773*Sin(arg); sigR += 4130*Cos(arg)
    arg = 2*D + 2*F;    sigL += -1595*Sin(arg)
    arg = 4*D - M - MM; sigL += 1215*Sin(arg)*E;    sigR += -3958*Cos(arg)*E
    arg = 2*MM + 2*F;   sigL += -1110*Sin(arg)
    arg = 3*D - MM;     sigL += -892*Sin(arg);      sigR += 3258*Cos(arg)
    arg = 2*D + M + MM; sigL += -810*Sin(arg)*E;    sigR += 2616*Cos(arg)*E
    arg = 4*D - M - 2*MM; sigL += 759*Sin(arg)*E;   sigR += -1897*Cos(arg)*E
    arg = 2*M - MM;     sigL += -713*Sin(arg)*EE;   sigR += -2117*Cos(arg)*EE
    arg = 2*D + 2*M - MM; sigL += -700*Sin(arg)*EE; sigR += 2354*Cos(arg)*EE
    arg = 2*D + M - 2*MM;   sigL += 691*Sin(arg)*E
    arg = 2*D - M - 2*F;    sigL += 596*Sin(arg)*E
    arg = 4*D + MM;     sigL += 549*Sin(arg);       sigR += -1423*Cos(arg)
    arg = 4*MM;         sigL += 537*Sin(arg);       sigR += -1117*Cos(arg)
    arg = 4*D - M;      sigL += 520*Sin(arg)*E;     sigR += -1571*Cos(arg)*E
    arg = D - 2*MM;     sigL += -487*Sin(arg);      sigR += -1739*Cos(arg)
    arg = 2*D + M - 2*F; sigL += -399*Sin(arg)*E
    arg = 2*MM - 2*F;   sigL += -381*Sin(arg);      sigR += -4421*Cos(arg)
    arg = D + M + MM ;  sigL += 351*Sin(arg)*E
    arg = 3*D - 2*MM;   sigL += -340*Sin(arg)
    arg = 4*D - 3*MM;   sigL += 330*Sin(arg)
    arg = 2*D - M + 2*MM ;  sigL += 327*Sin(arg)*E
    arg = 2*M + MM;     sigL += -323*Sin(arg)*EE;   sigR += 1165*Cos(arg)*EE
    arg = D + M - MM ;  sigL += 299*Sin(arg)*E
    arg = 2*D + 3*MM;   sigL += 294*Sin(arg)
    arg = 2*D - MM - 2*F;                           sigR += 8752*Cos(arg)

    # Latitude: calculate sumB, see Table 45.B, p.311
    arg = F;                    sigB = 5128122*Sin(arg)
    arg = MM + F;               sigB += 280602*Sin(arg)
    arg = MM - F;               sigB += 277693*Sin(arg)
    arg = 2*D - F;              sigB += 173237*Sin(arg)
    arg = 2*D - MM + F;         sigB += 55413*Sin(arg)
    arg = 2*D - MM - F;         sigB += 46271*Sin(arg)
    arg = 2*D + F;              sigB += 32573*Sin(arg)
    arg = 2*MM + F;             sigB += 17198*Sin(arg)
    arg = 2*D + MM - F;         sigB += 9266*Sin(arg)
    arg = 2*MM - F;             sigB += 8822*Sin(arg)
    arg = 2*D - M - F;          sigB += 8216*Sin(arg)*E
    arg = 2*D - 2*MM - F;       sigB += 4324*Sin(arg)
    arg = 2*D + MM + F;         sigB += 4200*Sin(arg)
    arg = 2*D + M - F;          sigB += -3359*Sin(arg)*E
    arg = 2*D - M - MM + F;     sigB += 2463*Sin(arg)*E
    arg = 2*D - M + F;          sigB += 2211*Sin(arg)*E
    arg = 2*D - M - MM - F;     sigB += 2065*Sin(arg)*E
    arg = M - MM - F;           sigB += -1870*Sin(arg)*E
    arg = 4*D - MM - F;         sigB += 1828*Sin(arg)
    arg = M + F;                sigB += -1794*Sin(arg)*E
    arg = 3*F;                  sigB += -1749*Sin(arg)
    arg = M - MM + F;           sigB += -1565*Sin(arg)*E
    arg = D + F;                sigB += -1491*Sin(arg)
    arg = M + MM + F;           sigB += -1475*Sin(arg)*E
    arg = M + MM - F;           sigB += -1410*Sin(arg)*E
    arg = M - F;                sigB += -1344*Sin(arg)*E
    arg = D - F;                sigB += -1335*Sin(arg)
    arg = 3*MM + F;             sigB += 1107*Sin(arg)
    arg = 4*D - F;              sigB += 1021*Sin(arg)
    arg = 4*D - M + F;          sigB += 833*Sin(arg)
    arg = MM - 3*F;             sigB += 777*Sin(arg)
    arg = 4*D - 2*MM + F;       sigB += 671*Sin(arg)
    arg = 2*D - 3*F;            sigB += 607*Sin(arg)
    arg = 2*D + 2*MM - F;       sigB += 596*Sin(arg)
    arg = 2*D - M + MM - F;     sigB += 491*Sin(arg)*E
    arg = 2*D - 2*MM + F;       sigB += -451*Sin(arg)
    arg = 3*MM - F;             sigB += 439*Sin(arg)
    arg = 2*D + 2*MM + F;       sigB += 422*Sin(arg)
    arg = 2*D - 3*MM - F;       sigB += 421*Sin(arg)
    arg = 2*D + M - MM + F;     sigB += -366*Sin(arg)*E
    arg = 2*D + M + F;          sigB += -351*Sin(arg)*E
    arg = 4*D + F;              sigB += 331*Sin(arg)
    arg = 2*D - M + MM + F;     sigB += 315*Sin(arg)*E
    arg = 2*D - 2*M - F;        sigB += 302*Sin(arg)*EE
    arg = MM + 3*F;             sigB += -283*Sin(arg)
    arg = 2*D + M + MM - F;     sigB += -229*Sin(arg)*E
    arg = D + M - F;            sigB += 223*Sin(arg)*E
    arg = D + M + F;            sigB += 223*Sin(arg)*E
    arg = M - 2*MM - F;         sigB += -220*Sin(arg)*E
    arg = 2*D + M - MM - F;     sigB += -220*Sin(arg)*E
    arg = D + MM + F;           sigB += -185*Sin(arg)
    arg = 2*D - M - 2*MM - F;   sigB += 181*Sin(arg)*E
    arg = M + 2*MM + F;         sigB += -177*Sin(arg)*E
    arg = 4*D - 2*MM - F;       sigB += 176*Sin(arg)
    arg = 4*D - M - MM - F;     sigB += 166*Sin(arg)*E
    arg = D + MM - F;           sigB += -164*Sin(arg)
    arg = 4*D + MM - F;         sigB += 132*Sin(arg)
    arg = D - MM - F;           sigB += -119*Sin(arg)
    arg = 4*D - M - F;          sigB += 115*Sin(arg)*E
    arg = 2*D - 2*M + F;        sigB += 107*Sin(arg)*EE

    # Additives to sigL (p.112)
    sigL +=   3958*Sin(A1) \
            + 1962*Sin(Lprime-F) \
            + 318*Sin(A2)

    # Additives to sigB
    sigB += -2235*Sin(Lprime) \
            + 382*Sin(A3) \
            + 175*Sin(A1-F) \
            + 175*Sin(A1+F) \
            + 127*Sin(Lprime-MM) \
            - 115*Sin(Lprime+MM)

    MoonLong = Lprime + sigL/1e6                # degrees
    MoonLong = Mod(MoonLong, 360.)              # degrees, normnalized
    MoonLat  = sigB/1e6                         # degrees
    MoonDist = 385000.56 + sigR/1e3             # km
    MoonHP   = degrees(asin(6378.14/MoonDist))  # degrees
    MoonDistAU = MoonDist / 149597870.700       # au
    X = MoonDistAU * Cos(MoonLat) * Cos(MoonLong) # au, geocentric X
    Y = MoonDistAU * Cos(MoonLat) * Sin(MoonLong) # au, geocentric y
    Z = MoonDistAU * Sin(MoonLat)               # au, geocentric Z

    # Lunar Node and lunar Perigee
    # p.313, eq. 45.7

    MoonNode = 125.0445550 + T * (-1934.1361849 + T * (0.0020762 \
        + T * (1./467410. - T / 60616000.)))
    MoonNode = Mod(MoonNode, 360.)
    MoonPerigee = 83.3532430 + T * (4069.0137111 + T * (-0.0103238 \
        + T * (-1./80053 + T / 18999000.)))
    MoonPerigee = Mod(MoonPerigee, 360.)

    MoonRA = atan2(Sin(MoonLong)*Cos(epsilon)-Tan(MoonLat)*Sin(epsilon), \
                Cos(MoonLong))
    MoonDecl = asin(Sin(MoonLat)*Cos(epsilon) \
            + Cos(MoonLat)*Sin(epsilon)*Sin(MoonLong))
    MoonRA = Mod(degrees(MoonRA), 360.); MoonDecl = degrees(MoonDecl)

    # Moon altitude at rise, set (see p.98)
    Alt = 0.7275 * MoonHP - 34./60. # degrees

    if verbose:
        print ("JD  = ", JD)
        print ("T   = ", T)
        print ("Lprime = ", Deg(Lprime))
        print ("D   = ", Deg(D))
        print ("M   = ", Deg(M))
        print ("MM  = ", Deg(MM))
        print ("F   = ", Deg(F))
        print ("A1  = ", Deg(A1))
        print ("A2  = ", Deg(A2))
        print ("A3  = ", Deg(A3))
        print ("E   = ", E)
        print ("sigL= %.1f" % sigL)
        print ("sigB= %.1f" % sigB)
        print ("sigR= %.1f" % sigR)
        print ("MoonLong    = %10.6f degrees" % MoonLong)
        print ("MoonLat     = %10.6f degrees" % MoonLat )
        print ("MoonDist    =   %.1f km" % MoonDist)
        print ("MoonHP      = %10.6f degrees" % MoonHP  )
        print ("MoonNode    = %10.6f degrees" % MoonNode)
        print ("MoonPerigee = %10.6f degrees" % MoonPerigee)
        print ("MoonRA      = %10.6f degrees" % MoonRA     )
        print ("MoonDecl    = %10.6f degrees" % MoonDecl)
        print ("Alt         = %10.6f degrees" % Alt     )
#   Returns:    
#   MoonRA      Rektaszension in degrees
#   MoonDecl    Declination above equator in degrees
#   MoonDict    Dictionary with Moon-specific output values:
#       Object:     "Moon"
#       MoonLong    ecliptic longitude, degrees
#       MoonLat     ecliptic latitude, degrees
#       RA          Rektaszension
#       Decl        Deklination
#       MoonDist    distance Earth - Moon, km
#       MoonHP      horizontal parallax, degrees
#       MoonDistAU  distance Earth - Moon, au
#       X           geocentric X, au
#       Y           geocentric Y, au
#       Z           geocentric Z, au
#       MoonNode    ecliptic longitude of ascending node, degrees
#       MoonPerigee longitude of perigee og Moon's orbit, degrees
#       Alt         Altitude at rise, set (degrees)
#       suffix      "++" if MoonDist <= 360000, "--" if MoonDist >= 405000
    suffix = '  '
    if MoonDist <= 360000. : suffix = "++" # Super Moon
    if MoonDist >= 405000. : suffix = "--" # Mini Moon

    MoonDict = {
        "Object"    : "Moon",
        "MoonLong"  : MoonLong,
        "MoonLat"   : MoonLat,
        "RA"        : MoonRA,
        "Decl"      : MoonDecl,
        "MoonDist"  : MoonDist,
        "MoonHP"    : MoonHP,
        "MoonDistAU": MoonDistAU,
        "X"         : X,
        "Y"         : Y,
        "Z"         : Z,
        "MoonNode"  : MoonNode,
        "MoonPerigee": MoonPerigee,
        "Alt"       : Alt,
        "suffix"    : suffix}
    

    return (MoonRA, MoonDecl, MoonDict)

################ End of getMoonData() #################################
#######################################################################
#
#   https://de.wikipedia.org/wiki/Besselsche_Elemente#Berechnung_der_Besselschen_Elemente
#   getBesselElements(JDEeclipse, verbose=False):
#   Returns:
#       tbd
#

Version = "V1.1 hvf 2023-10-20 getBesselElements.py"
VersionList.append(Version)

def getBesselElements(JDEeclipse, verbose=False):
#
#   getMoonData, getSunData need UT
    dtUT = JDEeclipse - getDeltaT(JDEeclipse)/86400.
    deltaT = getDeltaT(JDEeclipse)
#
#   T0 is the reference hour, the nearest integer hour to the maximum
#   of the eclipse in TT
    T0 = roundhvfTime(hvfTime(JDEeclipse), 3600)

# get Sun and Moon Data at time of eclipse
    (MoonRA, MoonDecl, MoonDict) = getMoonData(hvfTime(dtUT), 0, verbose)
    (SunRA, SunDecl, SunDict)    = getSunData(hvfTime(dtUT), 0, verbose)

    # convert RA into hours
    MoonRAh = HHMMSS(MoonRA/15.)
    SunRAh = HHMMSS(SunRA/15.)

    # convert Decl into ddmmss
    MoonDeclh = hmsstring(MoonDecl)
    SunDeclh = hmsstring(SunDecl)

    MoonDist = MoonDict["MoonDist"]     # Moon distance, km
    SunDist  = SunDict["R"]             # Sun distance, km

    EarthAeqRad = 6378.137              # Earth aequatorial radius, km
                                        # Moon radius for umbral cone
    MoonRadUmbra = 0.272274             # in units of Earth, Meeus 1951-2200
    MoonRadPenum = 0.272481             # Moon radius for penumbral cone
    SunRad = 696342 / EarthAeqRad       # Sun radius in units of Earth

    rSun = SunDist / EarthAeqRad      # Sun distance in units of Earth radius
    rMoon = MoonDist / EarthAeqRad    # Moon distance in units of Earth radius

    # Ortsvektoren von Sonne und Mond
    
    MoonXYZ = [rMoon*Cos(MoonRA)*Cos(MoonDecl), rMoon*Sin(MoonRA)*Cos(MoonDecl),
                rMoon*Sin(MoonDecl)]

    SunXYZ = [rSun*Cos(SunRA)*Cos(SunDecl), rSun*Sin(SunRA)*Cos(SunDecl),
                rSun*Sin(SunDecl)]

    # Differenz der Ortsvektoren -> Schattenachse

    Kraw = [SunXYZ[i]-MoonXYZ[i] for i in range(3)]
    Knorm = sqrt(sum([Kraw[i]**2 for i in range(3)]))
    Kvec = [Kraw[i] / Knorm for i in range(3)]

    # RA a and Decl d of Schattenachse

    a = atan2(Kvec[1], Kvec[0])
    d = asin(Kvec[2])

    a = degrees(a)
    a = Mod(a, 360.)
    d = degrees(d)

    # x and y, coords of Shadow Axis on Fundamental Plane
    x = rMoon * Cos(MoonDecl)*Sin(MoonRA-a)
    y = rMoon * (Sin(MoonDecl)*Cos(d)-Cos(MoonDecl)*Sin(d)*Cos(MoonRA-a))
    zMoon = rMoon * (Sin(MoonDecl)*Sin(d)+Cos(MoonDecl)*Cos(d)*Cos(MoonRA-a))
    gamma = sqrt(x*x+y*y)   # distance of eclipse axis from Earth center
    if y < 0: gamma = -gamma    # convention

    # Kegelmantel fuer Halb- und Kernschatten

    sinF1 = (SunRad + MoonRadPenum) / Knorm
    sinF2 = (SunRad - MoonRadUmbra) / Knorm

    # Abstand des Halb- und Kernschattenkegels von der fundamentalebene

    zV1 = zMoon + MoonRadPenum/sinF1   # Halbschatten
    zV2 = zMoon - MoonRadUmbra/sinF2   # Kernschatten

    # Radien der Schattenkegel
    
    tanF1 = tan(asin(sinF1))
    tanF2 = tan(asin(sinF2))

    l1 = zV1 * tanF1    # immper positiv
    l2 = zV2 * tanF2    # negativ fuer totale SoFi

    BesselDict = {  "Object" :  "Bessel",
                    "JDE"   :   nicedatetime(hvfTime(JDEeclipse)),
                    "T0"    :   nicedatetime(T0),
                    "deltaT":   deltaT,
                    "Penum" :   MoonRadPenum,
                    "Umbra" :   MoonRadUmbra,
                    "SunRad":   SunRad,
                    "SunRA" :   SunRA,
                    "SunRAh" :   SunRAh,
                    "SunDecl":  SunDecl,
                    "SunDeclh":  SunDeclh,
                    "MoonRA":   MoonRA,
                    "MoonRAh":   MoonRAh,
                    "MoonDecl": MoonDecl,
                    "MoonDeclh": MoonDeclh,
                    "ERad"  :   6378.137,      # Earth aequatorial radius, km
                    "x"     :   x,
                    "y"     :   y,
                    "gamma" :   gamma,
                    "RAaxis":   a,
                    "Decaxis":  d,
                    "l1"    :   l1,
                    "l2"    :   l2,
                    "sinF1" :   sinF1,
                    "sinF2" :   sinF2,
                    "tanF1" :   tanF1,
                    "tanF2" :   tanF2,
                    "rSun"  :   rSun,
                    "rMoon" :   rMoon,
                    "zMoon" :   zMoon,
                    "zV1"   :   zV1,
                    "zV2"   :   zV2,
                    }

    return BesselDict

#########################################################


def  printBesselElements(BesselDict):
    bd = BesselDict
    print ("Object : ",  bd["Object"])
    print ("JDE    : ",  bd["JDE"])
    print ("T0     : ",  bd["T0"])
    print ("deltaT : %.1f" %  bd["deltaT"], "sec")
    print ("Earth Radius (km) : %.3f" % bd["ERad"], " used as unit below")
    print ("Penum  : %.5f" %  bd["Penum"],  "Moon radius for penumbral cone")
    print ("Umbra  : %.5f" %  bd["Umbra"],  "Moon radius for umbral cone")
    print ("SunRad : %.5f" %  bd["SunRad"], " Sun radius")
    print ("SunRA  : %.5f" %  bd["SunRA"],  " degrees")
    print ("SunRAh : ",  bd["SunRAh"],      " hours")
    print ("SunDecl: %.5f" %  bd["SunDecl"], " degrees")
    print ("SunDeclh: ",  bd["SunDeclh"],   " dmmss")
    print ("MoonRA : %.5f" %  bd["MoonRA"], " degrees")
    print ("MoonRAh: ",  bd["MoonRAh"],     " hours")
    print ("MoonDecl: %.5f" %  bd["MoonDecl"], " degrees")
    print ("MoonDeclh: ",  bd["MoonDeclh"], " dmmss")
    print ("x, y   : %.7f, %.7f" % ( bd["x"], bd["y"]), " coords of Shadow Axis on Fundamental Plane")
    print ("gamma  : %.7f" %  bd["gamma"])
    print ("Moon shadow Axis RA   : %4.5f" %  bd["RAaxis"], " degrees")
    print ("Moon shadow Axis Decl : %4.5f" %  bd["Decaxis"], " degrees")
    print ("sinF1  : %.9f" % bd["sinF1"], " Penumbra Cone Shell Opening")
    print ("sinF2  : %.9f" % bd["sinF2"], " Umbra Cone Shell Opening")
    print ("l1     : %.9f" % bd["l1"], " Penumbra Radius")
    print ("l2     : %.9f" % bd["l2"], " Umbra Radius, negative for total eclipse")
    print ("zV1    : %.7f" % bd["zV1"], " Distance of Penumbra Cone Top from Fundamental Plane")
    print ("zV2    : %.9f" % bd["zV2"], " Distance of Umbra Cone Top from Fundamental Plane")
'''
    # Kegelmantel fuer Halb- und Kernschatten

    sinF1 = (SunRad + MoonRadPenum) / Knorm
    sinF2 = (SunRad - MoonRadUmbra) / Knorm
                    "tanF1" :   tanF1,
                    "tanF2" :   tanF2,
                    "rSun"  :   rSun,
                    "rMoon" :   rMoon,
                    "zMoon" :   zMoon,
    # Abstand des Halb- und Kernschattenkegels von der fundamentalebene

    zV1 = zMoon + MoonRadPenum/sinF1   # Halbschatten
    zV2 = zMoon - MoonRadUmbra/sinF2   # Kernschatten
'''
# Solar Eclipse Stuff, Meeus Chapter 52
# hvf 02.01.2023
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87

import sys

VersionSolarEclipse = "1.3 hvf 20.10.2023 using ELP-2000/82 and VSOP Ephemerides"
def Usage(progname):
    print ("Usage: %s [UTC|TT] year [year ...]" % progname)
    print ("Usage: %s [UTC|TT] fromyear-untilyear ..." % progname)
    print (" year man be between -9999 and 9999")
    sys.exit(0)

def isNewMoon(k):   # k must be integer
    test = (abs(k - int(k)) < 0.001)
    # if verbose: print ("isNewMoon k, test = ", k, test)
    return test     # True or False

def isFullMoon(k):  # k must be half-integer
    test = (abs(abs(k - int(k)) - 0.5) < 0.001)
    # if verbose: print ("isFullMoon k, test = ", k, test)
    return test     # True or False 

def isFirstQuarter(k):  # k must be .25
    test = abs(abs(k - floor(k)) - 0.25) < 0.001
    # if verbose: print ("isFirstQuarter k, test = ", k, test)
    return test     # True or False 

def isLastQuarter(k):  # k must be .75
    test = abs(abs(ceil(k) - k) - 0.25) < 0.001
    # if verbose: print ("isLastQuarter k, test = ", k, test)
    return test     # True or False 

# getNewMoonJDEbyK (k, verbose = False)
# k counts the number of new moons since 2000-01-06
# Meeus chapter 47
# returns the Julian Ephemeris Days (JDE) time of the mean phase of the Moon
# returns also the elements for solar eclipse (for new moon)
# returns JDEeclipse, NewMoonDict
def getNewMoonJDEbyK(k, verbose = True):
    # we need only k integer for New Moon and Solar Eclipse
    # k integer:        New Moon    - Solar eclipse possible
    # k integer + 0.25: First Quarter
    # k integer + 0.5 : Full Moon   - Lunar eclipse possible
    # k integer + 0.75: Last Quarter

    T  = k / 1236.85            # (47.3) Julian centuries after Epoch2000
    E = 1 - 0.002516*T - 0.0000074*T*T      # (45.6)
    EE = E*E

    # Sun's Mean anomaly (47.4)
    M = 2.5534 + 29.10535669*k - 0.0000218*T*T - 0.00000011*T*T*T
    M = Mod(M, 360.)

    # Moon's Mean anomaly (47.5)
    MM = 201.5643 + 385.81693528*k + 0.0107438*T*T \
            + 0.00001239*T**3 - 0.000000058*T**4
    MM = Mod(MM, 360.)

    # Moon's argument of Latitude (47.6)
    F = 160.7108 + 390.67050274*k - 0.0016341*T*T \
            - 0.00000227*T**3 + 0.000000011*T**4
    F = Mod(F, 360.)

    # Longitude of the Ascending Node of the lunar orbit (47.7)
    Omega = 124.7746 - 1.56375580*k \
            + 0.0020691*T*T + 0.00000215*T**3
    Omega = Mod(Omega, 360.)

    # chapter 52 Solar Eclipses  p.350
    F1 = F - 0.02665 * Sin(Omega)

    # Planetary Arguments

    A1  = 299.77 +  0.107408*k - 0.009173*T*T
    A2  = 251.88 +  0.016321*k
    A3  = 251.83 + 26.651886*k
    A4  = 349.42 + 36.412478*k
    A5  =  84.66 + 18.206239*k
    A6  = 141.74 + 53.303771*k
    A7  = 207.14 +  2.453732*k
    A8  = 154.84 +  7.306860*k
    A9  =  34.52 + 27.261239*k
    A10 = 207.19 +  0.121824*k
    A11 = 291.34 +  1.844379*k
    A12 = 161.72 + 24.198154*k
    A13 = 239.56 + 25.513099*k
    A14 = 331.55 +  3.592518*k

    # Corrections (in days) to JDE (p. 321)
    # for New Moon and Full Moon

    arg = MM;       sNew =  -0.40720*Sin(arg);   sFull =  -0.40614*Sin(arg)
    arg = M;        sNew +=  0.17241*E*Sin(arg); sFull +=  0.17302*E*Sin(arg)
    arg = 2*MM;     sNew +=  0.01608*Sin(arg);   sFull +=  0.01614*Sin(arg)
    arg = 2*F;      sNew +=  0.01039*Sin(arg);   sFull +=  0.01043*Sin(arg)
    arg = MM-M;     sNew +=  0.00739*E*Sin(arg); sFull +=  0.00734*E*Sin(arg)
    arg = MM+M;     sNew += -0.00514*E*Sin(arg); sFull += -0.00515*E*Sin(arg)
    arg = 2*M;      sNew +=  0.00208*EE*Sin(arg); sFull += 0.00209*EE*Sin(arg)
    arg = MM-2*F;   sNew += -0.00111*Sin(arg);   sFull += -0.00111*Sin(arg)
    arg = MM+2*F;   sNew += -0.00057*Sin(arg);   sFull += -0.00057*Sin(arg)
    arg = 2*MM+M;   sNew +=  0.00056*E*Sin(arg); sFull +=  0.00056*E*Sin(arg)
    arg = 3*MM;     sNew += -0.00042*Sin(arg);   sFull += -0.00042*Sin(arg)
    arg = M+2*F;    sNew +=  0.00042*E*Sin(arg); sFull +=  0.00042*E*Sin(arg)
    arg = M-2*F;    sNew +=  0.00038*E*Sin(arg); sFull +=  0.00038*E*Sin(arg)
    arg = 2*MM-M;   sNew += -0.00024*E*Sin(arg); sFull += -0.00024*E*Sin(arg)
    arg = Omega;    sNew += -0.00017*Sin(arg);   sFull += -0.00017*Sin(arg)
    arg = MM+2*M;   sNew += -0.00007*Sin(arg);   sFull += -0.00007*Sin(arg)
    arg = 2*MM-2*F; sNew +=  0.00004*Sin(arg);   sFull +=  0.00004*Sin(arg)
    arg = 3*M;      sNew +=  0.00004*Sin(arg);   sFull +=  0.00004*Sin(arg)
    arg = MM+M-2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = 2*MM+2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = MM+M+2*F; sNew += -0.00003*Sin(arg);   sFull += -0.00003*Sin(arg)
    arg = MM-M+2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = MM-M-2*F; sNew += -0.00002*Sin(arg);   sFull += -0.00002*Sin(arg)
    arg = 3*MM+M;   sNew += -0.00002*Sin(arg);   sFull += -0.00002*Sin(arg)
    arg = 4*MM;     sNew +=  0.00002*Sin(arg);   sFull +=  0.00002*Sin(arg)

    # Correction (in days) for First and Last Quarter (p.322)

    arg = MM;       sQ =  -0.62801*Sin(arg); 
    arg = M;        sQ +=  0.17172*E*Sin(arg); 
    arg = MM+M;     sQ += -0.01183*E*Sin(arg); 
    arg = 2*MM;     sQ +=  0.00862*Sin(arg); 
    arg = 2*F;      sQ +=  0.00804*Sin(arg); 
    arg = MM-M;     sQ +=  0.00454*E*Sin(arg); 
    arg = 2*M;      sQ +=  0.00204*EE*Sin(arg); 
    arg = MM-2*F;   sQ += -0.00180*Sin(arg); 
    arg = MM+2*F;   sQ += -0.00070*Sin(arg); 
    arg = 3*MM;     sQ += -0.00040*Sin(arg); 
    arg = 2*MM-M;   sQ += -0.00034*E*Sin(arg); 
    arg = M+2*F;    sQ +=  0.00032*E*Sin(arg); 
    arg = M-2*F;    sQ +=  0.00032*E*Sin(arg); 
    arg = 2*M+MM;   sQ += -0.00028*EE*Sin(arg); 
    arg = 2*MM+M;   sQ +=  0.00027*E*Sin(arg); 
    arg = Omega;    sQ += -0.00017*Sin(arg); 
    arg = MM-M-2*F; sQ += -0.00005*Sin(arg); 
    arg = 2*MM+2*F; sQ +=  0.00004*Sin(arg); 
    arg = MM+M+2*F; sQ += -0.00004*Sin(arg); 
    arg = MM-2*M;   sQ +=  0.00004*Sin(arg); 
    arg = MM+M-2*F; sQ +=  0.00003*Sin(arg); 
    arg = 3*M;      sQ +=  0.00003*Sin(arg); 
    arg = 2*MM-2*F; sQ +=  0.00002*Sin(arg); 
    arg = MM-M+2*F; sQ +=  0.00002*Sin(arg); 
    arg = 3*MM+M;   sQ += -0.00002*Sin(arg); 

    # additional correction for Quarter phases

    W = 0.00306 - 0.00038*E*Cos(M) + 0.00026*Cos(MM) \
        - 0.00002*Cos(MM-M) + 0.00002*Cos(MM+M) + 0.00002*Cos(2*F)
    # for first Quarter: +W
    # for last Quarter:  -W

    # additional correction for all phases

    addCorr = 0.000325*Sin(A1)  + 0.000165*Sin(A2)  + 0.000164*Sin(A3) \
            + 0.000126*Sin(A4)  + 0.000110*Sin(A5)  + 0.000062*Sin(A6) \
            + 0.000060*Sin(A7)  + 0.000056*Sin(A8)  + 0.000047*Sin(A9) \
            + 0.000042*Sin(A10) + 0.000040*Sin(A11) + 0.000037*Sin(A12) \
            + 0.000035*Sin(A13) + 0.000023*Sin(A14)

    # times of the mean phases of the Moon, already affected by the Sun's
    # aberration and by the Moon's light-time (eq. 47.1)

    JDE = 2451550.09765 + 29.530588853*k \
            + 0.0001337*T*T \
            - 0.000000150*T*T*T \
            + 0.00000000073*T**4
    JDEsave = JDE

    # Eclipses correction to time of max eclipse (52.1), in days
    # sN: New Moon; sF: Full Moon
    arg = 2*MM;     sN  = 0.0161*Sin(arg); 
    arg = 2*F1;     sN += -0.0097*Sin(arg)
    arg = MM-M;     sN += 0.0073*E*Sin(arg)
    arg = MM+M;     sN += -0.0050*E*Sin(arg)
    arg = MM-2*F1;  sN += -0.0023*Sin(arg)
    arg = 2*M;      sN += 0.0021*E*Sin(arg)
    arg = MM+2*F1;  sN += 0.0012*Sin(arg)
    arg = 2*MM+M;   sN += 0.0006*E*Sin(arg)
    arg = 3*MM;     sN += -0.0004*Sin(arg)
    arg = M+2*F1;   sN += -0.0003*E*Sin(arg)
    arg = A1;       sN += 0.0003*Sin(arg)
    arg = M-2*F1;   sN += -0.0002*E*Sin(arg)
    arg = 2*MM-M;   sN += -0.0002*E*Sin(arg)
    arg = Omega;    sN += -0.0002*Sin(arg)

    sF = sN

    arg = MM;   sN += -0.4075*Sin(arg);     sF += -0.4065*Sin(arg)
    arg = M;    sN += 0.1721*E*Sin(arg);    sF += 0.1727*E*Sin(arg)

    # P and Q - p. 351
    Q = 5.2207
    arg = M;    P  = 0.2070*E*Sin(arg);     Q += -0.0048*E*Cos(arg)
    arg = 2*M;  P += 0.0024*E*Sin(arg);     Q += 0.0020*E*Cos(arg)
    arg = MM;   P += -0.0392*Sin(arg);      Q += -0.3299*Cos(arg)
    arg = 2*MM; P += 0.0116*Sin(arg)
    arg = MM+M; P += -0.0073*E*Sin(arg);    Q += -0.0060*E*Cos(arg)
    arg = MM-M; P += 0.0067*E*Sin(arg);     Q += 0.0041*E*Cos(arg)
    arg = 2*F1; P += 0.0118*Sin(arg)

    # WW, gamma, u
    # For a Solar eclipse, gamma is the least distance from the axis 
    # of the Moon's shadow to the center of the Earth, 
    # in units of the equatorial radius of the Earth.
    # gamma is positive if the shadow passes North of the center.
    # when gamma is between +0.9972 and -0.9972, the solar eclipse is central:
    # there exists a line of central eclipse on the Earth's surface.
    WW = abs(Cos(F1))
    gamma = (P*Cos(F1) + Q*Sin(F1)) * (1. - 0.0048*WW)

    # u denotes the radius of the Moon's umbral cone in the fundamental plane.
    # The radius of the penumbral cone in the fundamental plane is
    #   u + 0.5461

    u = 0.0059
    u += 0.0046*E*Cos(M)
    u += -0.0182*Cos(MM)
    u += 0.0004*Cos(2*MM)
    u += -0.0005*Cos(M+MM)

    # if abs(gamma) is between 0.9972 and 1.5433+u, the eclipse is not central.
    # If abs(gamma) is between 0.9972 and 1.5433 + u, the eclipse is not
    # central. In most cases, it is then a partial eclipse. However, when
    # abs(gamma) is between 0.9972 and 1.0260, a part of the umbral cone may
    # touch the surface of the Earth (within the polar regions), while the
    # axis of the cone does not touch the Earth. These non-central total
    # or annular eclipses occur when 0.9972 < abs(gamma) < 0.9972 + |u|.



    '''
    def isNewMoon(k):   # k must be integer
        test = (abs(k - int(k)) < 0.001)
        # if verbose: print ("isNewMoon k, test = ", k, test)
        return test     # True or False

    def isFullMoon(k):  # k must be half-integer
        test = (abs(abs(k - int(k)) - 0.5) < 0.001)
        # if verbose: print ("isFullMoon k, test = ", k, test)
        return test     # True or False 

    def isFirstQuarter(k):  # k must be .25
        test = abs(abs(k - floor(k)) - 0.25) < 0.001
        # if verbose: print ("isFirstQuarter k, test = ", k, test)
        return test     # True or False 

    def isLastQuarter(k):  # k must be .75
        test = abs(abs(ceil(k) - k) - 0.25) < 0.001
        # if verbose: print ("isLastQuarter k, test = ", k, test)
        return test     # True or False 
    '''

    if isNewMoon(k):
        JDE += sNew
        JDEeclipse = JDEsave + sN
        # if verbose: print ("New Moon")
    elif isFullMoon(k):
        JDE += sFull
        JDELunarEcl = JDEsave + sF
        # if verbose: print ("Full Moon")
    elif isFirstQuarter(k):
        JDE += sQ
        JDE += W
        # if verbose: print ("First Quarter")
    elif isLastQuarter(k):
        JDE += sQ
        JDE -= W
        # if verbose: print ("Last Quarter")
    else:
        print ("k not for New or Full Moon: ", k)
    JDE += addCorr
    eclipseType = None
    shortEclipseType = None
    LunarEclipseType = None
    shortLunarEclipseType = None
    omega = None

    # JDEeclipse ist Julian Date of maximum of the eclipse
    # JDE is the J.D. of the Moon event (new, half, full)

    # calculate Bessel elements
    if isNewMoon(k): 
        BesselDict = getBesselElements (JDEeclipse, verbose)
    
        if abs(gamma) < 0.9972-0.0010:  # 1957-03-30 is annular not central
        #if abs(gamma) < 0.997:  # Meeus 1951-2200 § 5
            if u < 0:
                eclipseType = "Total Central Solar Eclipse"
                shortEclipseType = "t"
            elif u < 0.0047:
                omega = 0.00464*sqrt(1-gamma**2)
                if u < omega:
                    eclipseType = "Central Annular-Total Solar Eclipse"
                    shortEclipseType = "rt"
                else:
                    eclipseType = "Central Annular Solar Eclipse"
                    shortEclipseType = "r"
            elif u >= 0.0047:
                eclipseType = "Central Annular Solar Eclipse"
                shortEclipseType = "r"
        elif abs(gamma) < 0.9972+abs(u):
            if u < 0:
                eclipseType = "Non-central Total Solar Eclipse"
                shortEclipseType = "(t)"
            else:
                eclipseType = "Non-central Annular Solar Eclipse"
                shortEclipseType = "(t)"
        elif abs(gamma) < 1.5433+abs(u):
            eclipseType = "Partial Solar Eclipse"
            shortEclipseType = "p"
        if verbose and eclipseType is not None: print ("gamma = ", gamma)

        NewMoonDict = { "Object"    :  "Newmoon",
                    "JDE"       :   JDE,
                    "JDEeclipse":   JDEeclipse,
                    "k"         :   k,
                    "gamma"     :   gamma, 
                    "u"         :   u,
                    "omega"     :   omega,
                    "type"      :   eclipseType,
                    "shorttype" :   shortEclipseType,
                    "BesselDict":   BesselDict,
                    }

        return JDEeclipse, NewMoonDict

    elif isFullMoon(k):
        BesselDict = getBesselElements (JDELunarEcl, verbose)
        # Full Moon: Lunar Eclipse, Meeus (52.3) ff
        # penumbra radius rho
        rho = 1.2848 + u
        # umbra radius
        sigma = 0.7403 - u
        # magnitudes for penumbral and umbral eclipses
        magPenum = (1.5573 + u - abs(gamma)) / 0.5450
        magUmbra = (1.0128 - u - abs(gamma)) / 0.5450
        if magUmbra > 0:
            LunarEclipseType = "Umbral Lunar Eclipse"
            shortLunarEclipseType = "Umbral"
        elif magPenum > 0:
            LunarEclipseType = "Penmbral Lunar Eclipse"
            shortLunarEclipseType = "Penumbral"
    
        FullMoonDict = { "Object"   :  "Fullmoon",
                    "JDE"       :   JDE,
                    "JDELunarEcl":  JDELunarEcl,
                    "k"         :   k,
                    "gamma"     :   gamma, 
                    "u"         :   u,
                    "rho"       :   rho,
                    "sigma"     :   sigma,
                    "magPenum"  :   magPenum,
                    "magUmbra"  :   magUmbra,
                    "type"      :   LunarEclipseType,
                    "shorttype" :   shortLunarEclipseType,
                    "BesselDict":   BesselDict,
                    }

   

    verbose = True
    if verbose and isFullMoon(k) and FullMoonDict['type']:
        print ("getNewMoonJDEbyK: k = %.2f" % k)
        print ("JDE = %.6f, %s" % (JDE, nicedatetime(hvfTime(JDE))))
        print ("JDEsave = %.6f, %s" % (JDEsave, nicedatetime(hvfTime(JDEsave))))
        print ("corrections sN, sF = ", sN, sF)
        print ("JDELunarEcl = %.6f, %s" % (JDELunarEcl, nicedatetime(hvfTime(JDELunarEcl))))
        print ("E, M(Sun), MM(Moon), F, Omega = %.7f, %.4f, %.4f, %.4f, %.4f" \
            % (E, M, MM, F, Omega)) 
        print ("F1, WW, gamma, u = %.4f, %.4f, %.4f, %.4f" % (F1, WW, gamma, u))
        print ("gamma = %.4f, closest distance of shadow to Earth center"%gamma)
        print ("Radius of penumbral cone rho = %.4f" % rho)
        print ("Radius of umbral cone sigma = %.4f" % sigma)
        print ("Magnitude of penumbral eclipse = %.4f" % magPenum)
        print ("Magnitude of umbral eclipse = %.4f" % magUmbra)
        print ("Type = ", LunarEclipseType)
        print ()
    return JDELunarEcl, FullMoonDict

# getNewMoons(year, useUT)
#   returns:
#   list of (dtnewMoon, NewMoonDict) for solar eclipses
#   for the calender year 
#   useUT: convert ephemerides TT to UTC
#
def getNewMoons(year, useUT):
    # chapter 47 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-2000.)*12.3685   # (47.2), relative moon phase at new year
    if (k0 == int(k0)): k = k0      
    else: k  = ceil(k0) - 1     # start with new moon
    T  = k / 1236.85            # (47.3) Julian centuries since the Epoch 2000
    y  = year
    EventList = []
    EventDistList = []
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    while y == year:
        #verb = k == int(k)
        #print ("k = ", k)
        verb = True
        verb = False
        # New Moon 
        if isNewMoon(k):
            (JDEeclipse, NewMoonDict) = getNewMoonJDEbyK(k, verb)
            dt = getGreenwichDate(JDEeclipse)
            deltaT = timedelta(seconds=getDeltaT(JDEeclipse))
            if useUT: dt -= deltaT
            #print ("Eclipse Date = ", nicedatetime(dt))
            # guard against first dt is in previous year!
            thisyear, dummy, dummy = dt.getDate()
            if NewMoonDict["type"] is not None and thisyear == year:
            #print ("Eclipse Date = ", nicedatetime(dt), NewMoonDict["type"])
                EventList.append((dt, NewMoonDict))
        k += 0.5
        T = k / 1236.85
        y = floor(T * 100 + 2000)
        #print ("k = ", k)
        if isFullMoon(k):
            (JDELunarEcl, FullMoonDict) = getNewMoonJDEbyK(k, verb)
            dt = getGreenwichDate(JDELunarEcl)
            deltaT = timedelta(seconds=getDeltaT(JDELunarEcl))
            if useUT: dt -= deltaT
            #print ("Eclipse Date = ", nicedatetime(dt))
            # guard against first dt is in previous year!
            thisyear, dummy, dummy = dt.getDate()
            if FullMoonDict["type"] is not None and thisyear == year:
            #print ("Eclipse Date = ", nicedatetime(dt), NewMoonDict["type"])
                EventList.append((dt, FullMoonDict))
            
        #print ("year y = ", y)
    return EventList
#########################################################33      

#print ("###########################################################")
print ()
print ("%s Version %s" % (sys.argv[0], VersionSolarEclipse))
print ()
print ("Explanations:")
print ("k:     Lunation number, counts the number of new moons since 2000-01-06")
print ("gamma: minimum distance from the axis of the Moon's shadow cone")
print ("       to the center of the Earth (in Earth equatorial radius)")
print ("       positive or negative if the axis of the shadow cone")
print ("       passes north or south of the Earth's center")
print ("Type:  t total central Solar eclipse")
print ("       r annular central Solar eclipse")
print ("       rt annular total central Solar eclipse")
print ("       p partial Solar eclipse")
print ("       (t) non-central total Solar eclipse")
print ("       (r) non-central annular Solar eclipse") 
print ("u:     radius of the Moon's umbral cone in the fundamental")
print ("       plane (in equatorial Earth radius)")
print ("       u is negative for total, positive for annular eclipse")
print ()

import re
# year-year or year is possible

# read years
useUT = False       # TT is now the norm
foundOne = False
while len(sys.argv) > 1:
    arg = sys.argv[1]
    if arg.upper() == 'UT' or arg.upper() == 'UTC':
        useUT = True
        del sys.argv[1]
        continue
    elif arg.upper() == 'TT':
        useUT = False
        del sys.argv[1]
        continue
    else:
        yearList = re.split(r'-', arg)
        # we have to consider negative years as well!
        if len(yearList) == 2:  # startYear-lastYear, both positive 
            try:
                if len(yearList[0]) > 0: 
                    startYear = int(yearList[0])
                    lastYear  = int(yearList[1])
                else:           # -startyear, negative, e.g. -1234
                    startYear = -int(yearList[1])
                    lastYear  = startYear
            except:
                Usage(sys.argv[0])
        elif len(yearList) == 3:  # -0003-0003 from -3 to +3
            try:
                startYear = -int(yearList[1])
                lastYear  = int(yearList[2])
            except:
                Usage(sys.argv[0])
        elif len(yearList) == 4:  # -0003--0002 from -3 to -2
            try:
                startYear = -int(yearList[1])
                lastYear  = -int(yearList[3])
            except:
                Usage(sys.argv[0])
        else:
            try:
                startYear = int(yearList[0])
                lastYear  = startYear
            except:
                Usage(sys.argv[0])

        if useUT: comment = "Universal Time Coordinated (UTC)"
        else: comment = "Terrestrial Time (TT)"
        print ("startYear, lastYear = ", startYear, lastYear)
        print ()
        for year in range(startYear, lastYear+1):
            print (" Solar Eclipses for the Year %5d" % year, comment)
            print (" ---------------------------------", "-"*len(comment))
            print ()
            EventList = getNewMoons(year, useUT)
            while len(EventList) > 0:
                dt = EventList[0][0]    # either UTC or TT
                eclType = EventList[0][1]["type"]
                shortEclType = EventList[0][1]["shorttype"]
                BesselDict = EventList[0][1]["BesselDict"]
                k = EventList[0][1]["k"]
    # u denotes the radius of the Moon's umbral cone in the fundamental plane.
    # The radius of the penumbral cone in the fundamental plane is
    #   u + 0.5461
                u       = EventList[0][1]["u"]
                print ("Date ", nicedatetime(dt), "k ", k, "Object ", EventList[0][1]["Object"],
                        "Type ", shortEclType, ", u = %.4f" % u, " umbral cone radius")
                #print ("BesselDict = ", BesselDict)
                #printBesselElements(BesselDict)
                del EventList[0]
                print()
            foundOne = True
    del sys.argv[1]

if not foundOne: Usage(sys.argv[0])
