#! /usr/bin/env python3
# Julian to Greenwich and vice versa
# with my own hvfTime class, works also for BC dates and leap years
# hvf 16.09.2021 26.12.2021 11.11.2022 13.04.2023
# make hvfTime() hashable -- hvf 02.10.2023

from datetime import datetime, timedelta
from math import floor
import sys
import re
import dateutil.parser

VersionJulian = "V1.8 hvf 2023-10-02 HVFTime.py based on Meeus"


# JulianDay from year, month, day -  hvf 16.11.2021
def getJulianDayfromYMD(year, month, day, hour=0, minute=0, second=0, 
                        verbose = False):
    if verbose: 
        print ("getJulianDayfromYMD: ", year, month, day, hour, minute, second)
    day  += hour/24. + \
            minute/1440. + \
            second/86400.
    isGregor = False    # 04.10.1582 is last day before Gregorian calendar
    if year > 1582 or year == 1582 and (month > 10 or month == 10 and day > 4):
        isGregor = True
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    # JD 0 is -4712-01-01.5 i.e. -4712-01-01 12h
    # N.B. JD 1524.5 is -4708-03-05 00h
    if verbose: print ("JD = ", JD)
    return JD


def roundhvfTime(hvfDT = None, roundTo = 3600):
    # Round a hvfTime object to anytime lapse in seconds
    # hvfDT: hvftime object, default now
    # roundTo: closest number of seconds to round to, default 1 hour.
    # Author: Harald von Fellenberg 2021
    # Problem: 2021-11-21 1223 returns 2021-11-21 12:22:60.000
    # hvf 2022-09-09
    verbose=False
    if not isinstance(hvfDT, hvfTime): hvfDT = hvfTime.now()
    h, m, s = hvfDT.getTime()
    seconds = h*3600 + m*60 + s
    fraction = seconds - floor(seconds)     # fractional seconds
    if verbose:
        print ("seconds, fraction, floor, rounding = ", seconds, fraction, floor(seconds))
    seconds = floor(seconds)
    rounding = (seconds+roundTo/2) // roundTo * roundTo
    if verbose: print (rounding)
    return hvfDT + timedelta(0, rounding-seconds, -fraction*1e6)

# returns yyyymmdd hh for a hvfTime object, rounded
def nicehvfTimehh (hvfDT):
    hvfDTrounded = roundhvfTime(hvfDT)
    yy, mm, dd = hvfDTrounded.getDate()
    hh, dummy, dummy = hvfDTrounded.getTime()

    s = ''
    if yy >= 0:
        s = " %04d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    else:
        s = "%05d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    return s

#######################################################################
# replaces astropy.time.Time since that does not support leap days 
# before 100BC
# values are represented in Julian days
# creator parameter can be:
# - a hvfTime object
# - a JD JulianDay object (float)
# - a datetime object
# - date and time as six values yy, mm, dd, [hh, MM, ss]
#   dd and ss may have fractions
# hvf 26.12.2021

class hvfTime:
    def __init__(self, *arg):
        # print ("__init__: arg = ", arg)
        # argument is a hvfTime object
        if isinstance(arg[0], hvfTime):
            JD = arg[0].getJD()
            self.__JD = JD
        # argument is a JD
        elif isinstance(arg[0], float) and len(arg) == 1:
            JD = arg[0]
            self.__JD = JD
        # argument is a datetime object
        elif isinstance(arg[0], datetime):
            dt = arg[0]
            yy = dt.year
            mm = dt.month
            dd = dt.day
            hh = dt.hour
            MM = dt.minute
            ss = dt.second + dt.microsecond*1e-6
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        # argument is yy mm dd [hh MM ss]
        elif (isinstance(arg[0], int) or isinstance(arg[0], float)) and \
            len(arg) >= 3:
            hh = 0; MM = 0; ss = 0
            yy = arg[0]; mm = arg[1]; dd = arg[2]
            if len(arg) > 3: hh = arg[3]
            if len(arg) > 4: MM = arg[4]
            if len(arg) > 5: ss = arg[5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD

        # argument is (yy mm dd [hh MM ss])
        elif isinstance(arg[0], tuple) or isinstance(arg[0], list):
            hh = 0; MM = 0; ss = 0
            yy = arg[0][0]; mm = arg[0][1]; dd = arg[0][2]
            if len(arg[0]) > 3: hh = arg[0][3]
            if len(arg[0]) > 4: MM = arg[0][4]
            if len(arg[0]) > 5: ss = arg[0][5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        else:
            print ("hvfTime instance not supported: ", arg[0])
            sys.exit(1)
            
    def getJD(self):
        return self.__JD

    def getDate(self, verbose = False):  # returns (yy, mm, dd), 
                                        # where dd may have fraction
        # works also for negative JDs! hvf 29.12.2021
        JD = self.__JD
        jd1 = JD + 0.5
        I = floor(jd1)      # integer part, may be negative
        F = jd1 - I         # fractional part, always positive
        if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
            A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
            B = I + A - int(A/4) + 1
        else:   B = I       # Julian Calendar for earlier dates
        C = B + 1524
        D = floor((C-122.1)/365.25)
        E = floor(365.25 * D)
        G = floor((C-E)/30.6001)
        d = C - E + F - floor(30.6001 * G)
        if G < 13.5:    m = G - 1
        else:           m = G - 13
        if m > 2.5:     y = D - 4716
        else:           y = D - 4715

        if verbose:
            print ("getDate: JD, jd1, I, F = ", JD, jd1, I, F)
            print ("getDate: B, C, D, E, G = ", B, C, D, E, G)
            print ("getDate: y, m, d       = ", y, m, d)
        return (y, m, d)

    def getTime(self):      # seconds have 3 digi fractions (millisec)
        dummy, dummy, t = self.getDate()
        t -= floor(t)
        # t is fraction of day - convert to sec and round
        daysec = round(t * 86400., 3)
        hms = daysec / 3600.
        #hms = t*24.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 3)
        #s = int((ms - m) * 60)
        #print ("daysec, hms, h, ms, m, s = ", daysec, hms, h, ms, m, s)
    
        return (h, m, s)

    def __str__(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, hh, mm, ss)

    def getStr(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        daysec = round(ss + mm*60 + hh*3600, 0)
        hms = daysec / 3600.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 0)

        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, h, m, s)

    # Ernesti, Python3, p. 380
    # hvfTime = hvfTime + timedelta
    # hvfTime = hvfTime + day   (float or int)
    def __add__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__add__: wrong type ", type(td))
            sys.exit(1)
        return hvfTime(self.__JD + sec/86400.)

    # hvfTime += timedelta
    def __iadd__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__iadd__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD += sec/86400.
        return self

    # hvfTime = hvfTime - timedelta
    # and   n = hvfTime - hvfTime       n in days
    def __sub__(self, td):
        # subtract a timedelta object 
        if isinstance (td, timedelta):
            sec = td.total_seconds()
            return hvfTime(self.__JD - sec/86400.)
        elif isinstance(td, hvfTime):
            # return JDs in days as difference between two hvfTime objects
            return self.__JD - td.getJD()
        else:
            print ("hvfTime.__sub__: wrong type ", type(td), td)
            sys.exit(1)

    # hvfTime -= timedelta
    def __isub__(self, td):
        # subtract a timedelta object 
        try:    
            sec = td.total_seconds()
        except:
            print ("hvfTime.__isub__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD -= sec/86400.
        return self

    # hvfTime < hvfTime
    def __lt__(self, other):
        return self.__JD < other.getJD()

    # hvfTime <= hvfTime
    def __le__(self, other):
        return self.__JD <= other.getJD()

    # hvfTime == hvfTime
    def __eq__(self, other):
        return self.__JD == other.getJD()

    # hash - makes hvfTime hashable, i.e. as key in dict or set 
    def __hash__(self):
        return hash(self.__JD)

    # hvfTime != hvfTime
    def __ne__(self, other):
        return self.__JD != other.getJD()

    # hvfTime > hvfTime
    def __gt__(self, other):
        return self.__JD > other.getJD()

    # hvfTime >= hvfTime
    def __ge__(self, other):
        return self.__JD >= other.getJD()

    def now():  # returns current time (h, m, s), uses datetime.now()
                # roundedd to integer seconds
        return hvfTime(datetime.now()).getTime()

    def today():    # returns current date (y, m, d), uses datetime.today()
                    # d has fractions
        return hvfTime(datetime.today()).getDate()
    
    def getDatetime(self):
        (yy, mm, dd) = self.getDate()
        dd = floor(dd)
        (HH, MM, SS) = self.getTime()
        SS = int(round(SS, 0))
        return datetime(yy, mm, dd, HH, MM, SS)
#######################################################################

# parse date time
# returns a hvfTime object
# hvf 19.11.2021 25.12.2021
# allowed patterns:
# 2021-11-21
# 2021-11-21 12:23:45.635
# 2021-11-21T12:23:45.635
# 2021-11-21 122345.635
# 2021-11-21 122345
# 2021-11-21 1223
# 2021-11-21 12
# -4712-1-1.5 
# 25.12.2021 12:34:56.789
# 1.1.2022
# 2451545.0 (a Julian Day)
# works for dates between -9999-01-01 and 9999-12-31

def parseDateTime(s, verbose = False):
    HH = MM = SS = 0
    # check if two parts separated by blank or T
    if verbose: print ("parseDateTime input: ", s)
    s = s.strip()
    dpart = s
    tpart = None
    parts = re.split (r"[\sT]", s)
    if len(parts) >= 2:
        dpart = parts[0]    # date part
        tpart = parts[1]    # time part
        if verbose:
            print ("date part = ", dpart)
            print ("time part = ", tpart)

    # parse date  2021-12-24 or  4712-1-1.5
    if re.fullmatch(r'\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        ymd = re.split(r'-', dpart)
        yy = ymd[0]
        mm = ymd[1]
        dd = ymd[2]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date 24.12.2021 or 1.1.2022
    elif re.fullmatch(r'\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        ymd = re.split(r'\.', dpart)
        yy = ymd[2]
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -4712-01-01.5
    elif re.fullmatch(r'-\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        symd = re.split(r'-', dpart)
        yy = -int(symd[1])
        mm = symd[2]
        dd = symd[3]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -24.12.2021 or -1.1.2022
    elif re.fullmatch(r'-\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        symd = re.split(r'-', dpart)
        ymd = re.split(r'\.', symd[1])
        yy = -int(ymd[2])
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    else:
        print ("parseDateTime format unknown: ", s)
    # parse date  2021-12-24 or  4712-1-1.5
    # parse date 24.12.2021 or 1.1.2022
    # parse date -4712-01-01.5
    # parse date -24.12.2021 or -1.1.2022
        print ("Allowed date formats include 2021-12-24, 4712-1-1.5,")
        print ("    24.12.2021, 1.1.2022, -4712-01-01.5, -24.12.2021, -1.1.2022")
        print ("Allowed range of years is from -9999 to 9999")
        sys.exit(1)

    # parse time using dateutil.parser.isoparse - this is easier
    # than checking all the different possible input formats
    if tpart:
        deltaT = dateutil.parser.isoparse("2021-12-24 " + tpart) \
                - datetime(2021, 12, 24)
        Secs = round(deltaT.total_seconds(), 3)    # avoid rounding error
        HH = Secs // 3600
        Secs -= HH * 3600
        MM = Secs // 60
        Secs -= MM * 60
        SS = Secs
        if verbose:
            print ("HH, MM, SS = ", HH, MM, SS)

    #return hvfTime object
    return hvfTime(int(yy), int(mm), float(dd), int(HH), int(MM), float(SS))

#######################################################################
# catculate weekday for any date from -9999-01-01 to 9999-12-31
# Meeus 1991 p. 65

def getDayofWeek(dt):
    name =["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday",
            "Friday", "Saturday"]
    hvfdt = hvfTime(dt)
    (yy, mm, dd) = hvfdt.getDate()
    hvfdtMidnight = hvfTime(yy, mm, int(dd), 0, 0, 0.0)
    JD0 = hvfdtMidnight.getJD()
    weekday = int(JD0 + 1.51) % 7
    return name[weekday]

#######################################################################

#!/usr/bin/env python3
# -*- coding: iso-8859-1 -*-

# Der folgende Python-Code fuer Sonne und Mond 
# basiert auf der Portierung des java-Scripts von Arnold Barmettler
# http://lexikon.astronomie.info/java/sunmoon/sunmoon.html
#
#     Entfernen Sie folgende Informationen auf keinen Fall: / 
#     Do not remove following text:
#     Source code by Arnold Barmettler, www.astronomie.info / www.CalSky.com
#     and Harald von Fellenberg, www.hvf.ch
#     based on algorithms by Peter Duffett-Smith's great and easy book
#     'Practical Astronomy with your Calculator or Spreadsheet'.
#     and Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Totale Ueberarbeitung durch Harald von Fellenberg, Dezember 2020
#   Portierung auf Python 3
#   Ergaenzung fuer Mitternachtssonne, Polarnacht
#   Ergaenzungen fuer Daemmerung, Hoehe der Sonne beim Transit tags und nachts
#   Ergaenzung mit Zeitgleichung, wahre und mittlere Ortszeit
#   Azimut für Sonnenaufgang, Sonnenuntergang
#   Liste einiger Städte mit ihren Koordinaten
#   Korrekte Berechnung in allen Zeitzonen
#   Behandlung aller Planeten
#
#   Die Behandlung von Sonne und Mond wurde ersetzt durch die Algorithmen in
#   Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Die Daten aller Planeten beruhen auf VSPO87
#   Pierre Bretagnon, Gerard Francou
#   Bureau des Longitudes
#   77, Avenue Denfert-Rochereau
#   F75014, Paris, France
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/
#
#   Die Resultate wurden mit denen auf https://www.timeanddate.com/ verglichen,
#   die Uebereinstimmung ist gut innerhalb von wenigen Minuten.
#   Beim Mond ist stets das Datum zu beachten!
#   Liegt die lokale Zeit zwischen Mitternacht und "zone", so wird für den Mond
#   zum Teil auf den Vortag zurückgegriffen. Dies ist besonders für Orte mit
#   grosser Zeitzonendifferenz wichtig, wie Sydney oder Honolulu.
#   Für Orte in den Polarregionen wird Mitternachtssonne und Polarnacht
#   korrekt ausgegeben, ausser am Tag, wo die Sonne unter- oder aufgeht.
#   Ebenso fuer den Mond.
#
#   Anmerkung: die Webseiten astronomie.info sowie calsky.com existieren
#   seit Oktober 2020 nicht mehr.
#
#   Alle 8 Planeten sind ebenfalls behandelt, gemaess den Algorithmen in
#   Jean Meeus, Astronomical Algorithms.
#   Die Daten zu den 8 Planeten stammen aus den VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Version 1.0
#   Version 2.0
#   Version 3.0
#   Version 3.1
#   Version 3.4
#   Version 4.0 inkl Sun-Events, Moon-Events
#   Version 5.1 
Version = "V5.3 hvf 2023-09-27 SunHead.py using VSOP87 ephemerides and HVFTime"
if 'VersionList' not in globals().keys(): VersionList = []
VersionList.append(Version)

import time
from datetime import datetime, date, timedelta
from math import sin, acos, cos, pi, radians, degrees, atan, atan2
from math import asin, tan, ceil, floor, sqrt, fabs
import argparse
import sys
import dateutil.parser
import os

# alle Konstanten sind auf die Epoche 2010.0, JD=2455196.5 bezogen
# falls nicht anders dokumentiert.

def getEpoch2010():
    return 2455196.5

# https://www.geeksforgeeks.org/how-to-print-to-stderr-and-stdout-in-python/
def print_to_stderr(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stderr) 

def print_to_stdout(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stdout) 

# produces a datetime object output string without microseconds
# seconds get rounded

def nicedatetime(dt):       # yyyy-mm-dd HH:MM:SS
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        string = dt.getStr()
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%Y-%m-%d %H:%M:%S")
    return string

def niceshortdatetime(dt):  # dd.mm.yy HH:MM:SS
    # only meaningful for known century
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        (yy, mm, dd) = dt.getDate()
        yy = int(abs(yy)%100)    # only last 2 ciphers, also for negative yyyy
        (hh, MM, ss) = dt.getTime()
        # round seconds! use %f fmt - hvf 13.04.2023
        #string = "%02d.%02d.%02d %02d:%02d:%02d" % (dd, mm, yy, hh, MM, ss)
        string = "%02d.%02d.%02d %02d:%02d:%02.0f" % (dd, mm, yy, hh, MM, ss)
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%d.%m.%y %H:%M:%S")
    return string

def nicedateshorttime(dt):  # yyyy-mm-dd HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=30)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        (h, m, s) = dt.getTime()
        string = "%4d-%02d-%02d %02d:%02d" % (y, mm, dd, h, m)
    else:
        string = dt.strftime("%Y-%m-%d %H:%M")
    return string

def niceshortdateshorttime(dt):  # dd.mm.yy HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        dtrounded = roundhvfTime(dt, 60)
        #print ("dt, dtrounded = ", dt, dtrounded)
        (y, mm, dd) = dtrounded.getDate()
        yy = abs(y)%100
        dd = floor(dd)
        (h, m, s) = dtrounded.getTime()
        string = "%02d.%02d.%02d %02d:%02d" % (dd, mm, yy, h, m)
    else:
        dt += timedelta(seconds=30)
        string = dt.strftime("%d.%m.%y %H:%M")
    return string

def nicedate(dt):           # yyyy-mm-dd
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=0.5)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        string = "%04d-%02d-%02d" % (y, mm, dd)
    else:
        string = dt.strftime("%Y-%m-%d")
    return string

def nicetime(dt):           # HH:MM:SS
    if isinstance(dt, type(None)): return None
    try:
        if isinstance(dt, hvfTime):
            (h, m, s) = dt.getTime()
            s =round(s, 0)
            string = "%02d:%02d:%02d" % (h, m, s)
        else:
            dt += timedelta(seconds=0.5)
            string = dt.strftime("%H:%M:%S")
    except:
        string = ' '*8
    return string

def nicetimedelta(dt):      # +-HH:MM:SS
    if isinstance(dt, type(None)): return None
    #dt += timedelta(seconds=0.5)
    # treat also -00:00:05 correctly! hvf 16.06.2021
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60

    string = "%1s%2d:%02d:%02d" % (sign, hours, minutes, seconds)
    return string

def nicedeltadayhhmm(dt):   # +-days HH.MM
    if isinstance(dt, type(None)): return None
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    days  = int(totalsecs/86400.); totalsecs -= days*86400.
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60:
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60
    if hours >= 24:
        days += 1; hours -= 24

    string = "%1s%d days %02d:%02d" % (sign, days, hours, minutes)
    return string

def niceminsec(secs):       # +- M:SS  for Time Equation - Zeitgleichung
    if isinstance(secs, type(None)): return None
    sign = ' '
    if secs < 0: sign = '-'; secs = abs(secs)
    minutes = int(secs/60.); secs -= minutes*60.
    seconds = round(secs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60

    string = "%1s%d:%02d" % (sign, minutes, seconds)
    return string

# printZeile(kopf, Liste, format, zeile=79)
# schreibt eine Zeile der max. Laenge zeile mit kopf und den
# Werten aus Liste, formattiert je mit format.
# Padding vor jedem Listenelement
# Die  formattierten Werte muessen gleich lang sein.
# hvf 31.12.2020

def printZeile(kopf, Liste, format, zeile=79):
    rest = zeile
    string = kopf
    rest -= len(string)
    n = len(Liste)      # Anzahl Werte
    test = format % Liste[0]
    testlen = len(test)
    padlen = int((rest - n*testlen)/n)
    padding = ' '*padlen

    for item in Liste:
        string += padding
        string += format % item

    print (string)
    return string

# modulus in Python has always the same sign as b
# Mod(-3, 24) gives 21
def Mod(a, b):
    return (a % b)

# return integer value, closer to 0
def Int(x): return int(x)

# return fractional part, closer to 0 (i.e. may be nagative)
def Frac(x): return (x - int(x))

# Winkelfunktionen fuer Grad Eingaben
def Sin(x):
    return sin(radians(Mod(x, 360.)))

def Cos(x):
    return cos(radians(Mod(x, 360.)))

def Tan(x):
    return tan(radians(Mod(x, 360.)))

def Deg(x):
    return degrees(x)

# calculate angular distance between two objects with (RA1, Decl1), (RA2, Decl2)
# all angles in degrees
# see Meeus, chapter 16, p. 105, Angular Separation
def AngularDistance(star1, star2, verbose=False):
    RA1 = star1[0]; Decl1 = star1[1]
    RA2 = star2[0]; Decl2 = star2[1]
    cosD = Sin(Decl1)*Sin(Decl2) + Cos(Decl1)*Cos(Decl2)*Cos(RA1-RA2)
    D = degrees(acos(cosD))
    # check if angular separation is small
    if D < 0.2:     # below 0.2 degrees, 12'
        Dmean = (Decl1+Decl2)/2.
        Ddelta = Decl1 - Decl2
        D = sqrt(((RA1-RA2)*Cos(Dmean))**2 + Ddelta**2)
    return D


#####################################################################
# Quelle: https://de.wikipedia.org/wiki/Julianisches_Datum
# abgerufen am 20.10.2019
# Name geaendert in getJulianDay (siehe Meeus, chapter 7) - hvf 24.03.2023

# Calculate Julian day for all Gregorian and Julian dates
# Greenwich date is supplied as a datetime object

def getJulianDay(dtGreenwichDate, isGregor=True ):
    '''
    # day may have fractions for h, m, s, us
    # Jan and Feb are months 13 and 14
    year  = dtGreenwichDate.year
    month = dtGreenwichDate.month
    day   = dtGreenwichDate.day
    day  += dtGreenwichDate.hour/24. + dtGreenwichDate.minute/1440. + \
            dtGreenwichDate.second/86400. + \
            dtGreenwichDate.microsecond/86400e6
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    return JD
    '''
    hvfdt = hvfTime(dtGreenwichDate)
    return hvfdt.getJD()

################################################################
# calculate Greenwich date from Julian date
# Ref. Practical Astronomy (see above), p.11
# returns a datetime object for midnight 

def getGreenwichDate(JD):
    '''
    jd1 = JD + 0.5
    I = int(jd1)        # integer part
    F = jd1 - I         # fractional part
    if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
        A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
        B = I + A - int(A/4) + 1
    else:   B = I       # Julian Calendar for earlier dates
    C = B + 1524
    D = int((C-122.1)/365.25)
    E = int(365.25 * D)
    G = int((C-E)/30.6001)
    d = C - E + F - int(30.6001 * G)
    if G < 13.5:    m = G - 1
    else:           m = G - 13
    if m > 2.5:     y = D - 4716
    else:           y = D - 4715

    # add fractional day - must use timedelta for this
    dtGreenwichDate = datetime(y, m, 1) + timedelta(days=d-1)
    return dtGreenwichDate
    '''

    return hvfTime(float(JD))

#   HMSDH convert hours, minutes, seconds to decimal hours
def HMSDH(hh, mm, ss):  # hh may be negative
    habs = abs(hh)
    habs += mm/60. + ss/3600.
    h = habs
    if hh < 0: h = -habs
    return h

#   DHHMS converts decimal hours into hours, minutes, seconds
#   seconds may have fractions, no rounding
def DHHMS(h):   # h may be negative, then only hours will be negative
    sign = 1
    if h < 0: sign = -1
    habs = abs(h)
    hh = int(habs)
    mm = int(Frac(habs)* 60.)
    ss = Frac(habs*60)*60.
    hh = hh * sign
    return (hh, mm, ss)

#   returns HH:MM and a rounded value of h
#   h must be positive or None or ''
def HHMM(hh):
    if (hh == None or hh == ''):
        return('--')
    m = Frac(hh) * 60.0
    h = Int(hh)
    if (m >= 59.5):
        h += 1
        m -= 60.0
    m = int(round(m))
    string = "%02d:%02d = " % (h, m) + str(round(hh, 4))
    return (string)

#   returns HH:MM:SS and a rounded value of h
#   h must be positive or None or ''
def HHMMSS(hh):
    if (hh == 0):
        return('--')
    m = Frac(hh) * 60
    h = Int(hh)
    s = Frac(m) * 60.0
    m = Int(m)
    if (s >= 59.5):
        m += 1
        s -= 60.0
    if (m >= 60):
        h += 1
        m -= 60
    s = int(round(s))
    string = "%02d:%02d:%02d = " % (h, m, s) + str(round(hh, 5))
    return (string)

def hmsstring(hour):
    # hour may also be negative, e.g. if used to convert declination
    (hh, mm, ss) = DHHMS(hour)
    str = "%d:%02d:%02d" % (hh, mm, ss)
    return str


#   local civic time to Universal Time
#   returns a datetime object 

def getUTfromCivicTime(dtLocalCivicTime, zone):    
    # dtLocalCivicTime is a datetime object
    # zone is time zone offset in hours, positive going East
    dtUT = dtLocalCivicTime - timedelta(hours=zone)
    return dtUT

def getCivicTimefromUT(dtUT, zone):
    dtLocalCivicTime = dtUT + timedelta(hours=zone)
    return dtLocalCivicTime

#   UTC to mean local time 
#   dtUT: Universal Time datetime object
#   longitude: Eastern longitude, degrees
#   returns a datetime object, rounded to seconds

def getMeanLocalTime(dtUT, longitude):
    dtmeanLT = dtUT + timedelta(seconds=round(240.*longitude, 0))
    return dtmeanLT

def getUTfromMeanLocalTime(dtmeanLT, longitude):
    dtUT =  dtmeanLT - timedelta(seconds=round(240.*longitude, 0))
    return dtUT

#   datetime.datetime or datetime.time to decimal hours

def getDecimalHour(dttime):
    '''
    hour = dttime.hour + dttime.minute/60. + dttime.second/3600. \
            + dttime.microsecond/3600.E6
    '''
    (hour, min, sec) = hvfTime(dttime).getTime()
    return hour    

# calculate horizontal coordinates (Azimut, Altitude)
# from ecliptic coordinates (Hour angle, Declination)
#   Parameters:
#   Hour    Hour angle counted from meridian to West, degrees
#   Decl    Declination, degrees
#   latitude    degrees
#   Returns:
#   azimut  counted from North, degrees
#   alt     altitude above horizon, degrees
#   x, y, z Cartesian coords in horizontal system, x pointing South, y West
#   https://de.wikipedia.org/wiki/Astronomische_Koordinatensysteme#Umrechnungen

def getAziAltfromHourDecl(Hour, Decl, latitude):
    # Hour angle and Declination into azimut and altitude
    azimut = atan2(Sin(Hour), Cos(Hour)*Sin(latitude)-Tan(Decl)*Cos(latitude))
    alt    = asin(Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour))
    azimut = Mod(degrees(azimut)+180., 360.)
    alt    = degrees(alt)
    # Cartesian coords in horizon system
    x = Cos(alt)*Cos(azimut-180.)
    y = Cos(alt)*Sin(azimut-180.)
    z = Sin(alt)
    # dito calculated from  Hour, Decl, latitude
    xx = -Cos(latitude)*Sin(Decl)+Sin(latitude)*Cos(Decl)*Cos(Hour)
    yy = Cos(Decl)*Sin(Hour)
    zz = Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour)
    xyzDict = {
        "x":    x,
        "y":    y,
        "z":    z, 
        "xx":   xx,
        "yy":   yy,
        "zz":   zz, }
    return (azimut, alt, xyzDict)

# Zodiac signs for Sun and Moon
#   lon ecliptic longitude, degrees
def getSign(lon):
    Signs_de = ("Widder", "Stier", "Zwillinge", "Krebs", "Löwe", "Jungfrau", \
        "Waage", "Skorpion", "Schütze", "Steinbock", "Wassermann", "Fische")
    Signs_en = ("Aries", "Taurus", "Gemini", "Cancer", "Leo", "Virgo", \
        "Libra", "Scorpio", "Sagittarius", "Capricorn", "Aquarius", "Pisces")
    return (Signs_en[int(floor(lon / 30))])

########################################################################
#
# getNutation(dtCivicTime, zone)
# Nutation in ecliptic longitude and ecliptic latitude
# see Meeus, Astronomical Algorithms, chapter 21

def getNutation(dtCivicTime, zone, verbose=False):
    # Epoch is J2000.0
    # JD 2451545
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    T = (JD - 2451545.0)/36525. # Centuries after the Epoch, 21.1

    # Moon Ascending Node
    Omega = 125.04452 - 1934.136261 * T + 0.0020708 * T*T + T*T*T/450000.
    Omega = Mod(Omega, 360.)
    O = radians(Omega)

    # accuracy of 0.5 arcsec in longitude, 0.1 arcsec in latitude
    # Sun Mean Longitude
    LmeanSun = 280.4665 + 36000.7698 * T
    LmeanSun = Mod(LmeanSun, 360.)
    LS = radians(LmeanSun)

    # Moon Mean Longitude
    LmeanMoon = 218.3165 + 481267.8813 * T
    LmeanMoon = Mod(LmeanMoon, 360.)
    LM = radians(LmeanMoon)

    # Effect on ecliptic longitude and latitude, arcsecs
    deltaELong = -17.20*sin(O) -1.32*sin(2.*LS) -0.23*sin(2*LM) +0.21*sin(2*O)
    deltaELat  = 9.20*cos(O) + 0.57*cos(2.*LS) + 0.10*cos(2*LM) -0.09*cos(2*O)

    # obliquity of the ecliptic, 21.2
    eps0 = 23+26/60.+21.448/3600. - 46.8150/3600.*T - 0.00059/3600.*T*T \
            +0.001813/3600*T*T*T
    eps = eps0 + deltaELat/3600.

    if verbose:
        print ("getNutation")
        print ("JD = ", JD)
        print ("T = ", T)
        print ("LmeanSun = ", LmeanSun)
        print ("LmeanMoon = ", LmeanMoon)
        print ("getNutation deltaELong (arcsec) = %.2f" % deltaELong)
        print ("getNutation deltaELat  (arcsec) = %.2f" % deltaELat )
        print ("eps0 (degrees) = %.6f" % eps0, ", ", HHMMSS(eps0))
        print ("eps  (degrees) = %.6f" % eps, ", ", HHMMSS(eps) )

    # return values in degrees
    return (deltaELong/3600., deltaELat/3600., eps)

#################################################################
#   Funktionen für die Berechnung der Zeitgleichung

def getSunStufffromTT(TT, verbose = False):
    # TT Tage nach Epoch2010 inkl. DeltaT Korrektur
    # returns RA, epsilon in degrees

    def atanLage(x, Lage):    # arctangens mit Lageparameter, alle Werte in Grad
        # siehe https://en.wikipedia.org/wiki/Equation_of_time
        # Resultat in Grad
        xrad = radians(x)
        Lagerad = radians(Lage)
        y = atan(xrad)
        valuerad = y + pi * round((Lagerad-y)/pi, 0)
        value = degrees(valuerad)
        return value
    
    def getE(t):    # numerische Exzentrizität zur Zeit t
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        e = 0.01670491 - 4.208E-5*T - 1.26E-7*T*T
        return e
    
    def getEps(t):  # Schiefe der Ekliptik in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
        return eps
    
    def getOmegabar(t):     # Länge des Perihels in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        omegabar = 283.11243767 + 1.720171*T + 0.000452778*T*T
        return omegabar
    
    def getL(t):    # mittlere Länge der Sonne in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        L = 279.557285 + 36000.769585*T + 0.0003025*T*T
        return L
    
    def getC(e, M):    # Mittelpunktsgleichung equation of the center
        # berechnet wahre minus mittlere Anomalie als Fkt der Exzentrizität
        # und der mittleren Anomalie in dritter Näherung in e
        # e numerische Exzentrizität
        # M mittlere Anomalie der Sonne in Grad
        # Resultat in Grad
        # Danby, Fundamentals of Celestial Mechanics, p.438
        Mrad = radians(M)
        Crad = (2*e-e*e*e/4)*sin(Mrad) + (1.25*e*e -11./24.*e**4)*sin(2*Mrad)\
             + 13./12.*e*e*e*sin(3*Mrad) + 103./96.*e**4*sin(4*Mrad)
        C = degrees(Crad)
        return C

    t = TT
    e = getE(t)                 # Exzentrizität der Erdbahn, Zeit t
    eps = getEps(t)             # Schiefe der Ekliptik in Grad, Zeit t
    epsRad = radians(eps)       # Schiefe der Ekliptik in Radians
    omegabar = getOmegabar(t)   # Länge des Perihels in Grad, Zeit t
    L = getL(t)                 # mittlere Länge der Sonne in Grad, Zeit t
    Lhaupt = Mod(L, 360)        # Hauptwert der Länge der Sonne
    M = Lhaupt - omegabar       # Mittlere Anomalie der Sonne in Grad, Zeit t
    Mhaupt = Mod(M, 360)        # Hauptwert der mittleren Anomalie
    C = getC(e, Mhaupt)         # Mittelpunktsgleichung, wahre minus mittlere A
    Lwahr = L + C               # wahre Länge der Sonne in Grad, Zeit t
    LwahrRad = radians(Lwahr)   # wahre Länge der Sonne in Radians
    Lwahrhaupt = Mod(Lwahr, 360)# Hauptwert der wahren Länge
    LwahrHauptRad = radians(Lwahrhaupt)  # dito, in Radians


    # Rektaszension der Sonne in Funktion der wahren Länge der Sonne und der 
    # Schiefe der Ekliptik
    alpha = atanLage(degrees(tan(LwahrRad)*cos(epsRad)), Lwahr)
    alphahaupt = atanLage(degrees(tan(LwahrHauptRad)*cos(epsRad)), Lwahrhaupt)

    Dict = { "Object"   : "SunStuff",
            "TT"        : t,            # Tage nach Epoch2010 inkl. deltaT
            "MPEq"      : C,            # wahre - mittlere Anomalie
            "omegabar"  : omegabar,     # Länge des perihels in Grad, Zeit t
            "M"         : M,            # mittlere Anomalie, Grad
            "nu"        : C+M,          # wahre Anomalie
            "Long"      : Lwahrhaupt,   # wahre Länge der Sonne, Grad
            "eps"       : eps,          # Schiefe der Ekliptik, Grad
            "exc"       : e,            # Exzentrizität der Erdbahn
            }
    if verbose:
        print ("Numerische Exzentrizitaet e = ", "%.6f" % e)
        print ("Omegabar, Laenge des Perihels = ", "%.3f Grad" % omegabar)
        print ("Lhaupt, Hauptwert Mittlere Laenge = ", "%.3f Grad" % Lhaupt)
        print ("M Mittlere Anomalie = ", "%.3f Grad" % M)
        print ("Mhaupt, Hauptwert Mittlere Anomalie = ", "%.3f Grad" % Mhaupt)
        print ("C Mittelpunktsgleichung = ", "%.3f Grad" % C)
        print ("Alpha Wahre Anomalie = ", "%.3f Grad" % alpha)
    return (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict)

#######################################################################    
def Zeitgleichung(JD, verbose=False):
    # berechnet die Zeitgleichung in ganzen Sekunden
    # JD Julianisches Datum
    # verbose aktiviert zusätzlichen Output
    # Resultat in Sekunden ohne Rundung
    # Epoche konsistent 2010.0, 31.12.2009 0h UT, JD 2455196.5
    dtEpoch2010 = datetime(2009, 12, 31)
    tbase = getJulianDay (dtEpoch2010)
    t = JD - tbase
    TT = JD - tbase + getDeltaT(JD)/86400.

    (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict) = getSunStufffromTT(TT, verbose)

    # Zeitgleichung erste Ursache Exzentrizität der Erdbahn (epsilon = 0)
    ZGstar0 = L - Lwahr         # in Grad
    ZG0 = 240. * ZGstar0        # in Sekunden

    # zweite Ursache Anteil Schiefe der Ekliptik
    ZGstareps = Lwahrhaupt - alphahaupt # in Grad
    ZGeps = 240. * ZGstareps            # in Sekunden

    # Summe der beiden Ursachen
    ZGstar = ZGstar0 + ZGstareps        # in Grad
    ZGsecs = ZG0 + ZGeps                # in Sekunden

    # ZGmin, ZGsec for printing purposes 
    ZGstr = niceminsec(ZGsecs)
    if verbose:
        print ("Epoch2010 = ", dtEpoch2010)
        print ("Epoch Julian Date = ", tbase)
        print ("TT = ", TT)
        print ("Schiefe der Ekliptik eps = ", "%.3f Grad" % eps)
        print ("L, Mittlere Laenge = ", "%.3f Grad" % L)
        print ("Lwahr Wahre Sonnenlaenge = ", "%.3f Grad" % Lwahr)
        print ("Lwahrhaupt Hauptwert Wahre Sonnenlaenge = ", "%.3f Grad" \
                    % Lwahrhaupt)
        print ("Alphahaupt Wahre Anomalie = ", "%.3f Grad" % alphahaupt)
        print ("ZGstar0 = ", "%.3f" % ZGstar0, " Grad")
        print ("ZG0 = ", "%.2f" % ZG0, " Sekunden")
        print ("ZGstareps = ", "%.3f" % ZGstareps, " Grad")
        print ("ZGeps = ", "%.2f" % ZGeps, " Sekunden")
        print ("ZGstar = ", "%.3f" % ZGstar, " Grad")
        print ("ZGsecs = ", "%.2f" % ZGsecs, " Sekunden")
        print ("ZG = %s" % ZGstr)

    return (ZGsecs, ZGstr, Dict)

def getZeitgleichung(JD, verbose=False):
    # returns a timedelta object and two int's for printing
    zgsecs, ZGstr, Dict = Zeitgleichung(JD, verbose)
    return (timedelta(seconds=zgsecs), ZGstr, Dict)

# Greenwich Mean Sidereal Time from Julian Date

def getGMSTfromJD(JD):
    '''
    # new Epoch2010, JD=2455196.5
    # JD usually float or int
    # returns a datetime object
    UT = Frac(JD-0.5) * 24.     # UT in hours
    JD = floor(JD-0.5) + 0.5    # JD at 0 hours UT
    # JD 2455196.5 is Epoch 2010.0
    # T is centuries after Epoch
    T  = (JD - 2455196.5) / 36525.
    T0 = 6.636798592 + T * (2400.051341 + T * 0.000025862)
    ST = Mod(T0 + UT * 1.002737909, 24.)
    Gdate = getGreenwichDate(JD)
    return Gdate + timedelta(hours=ST)
    '''

    # Meeus chapter 11 p. 83
    T = (JD - 2451545.0) / 36525.   # (11.1)
    theta0 = 280.46061837 + 360.98564736629 * (JD - 2451545.0) \
            + 0.000387933 * T*T - T*T*T/387110000   # (11.4)
    # normalize to 360 degrees
    theta = Mod(theta0, 360.)
    # add to midnight
    JD0 = floor(JD-0.5) + 0.5
    Gdate = hvfTime(JD0 + theta/360.)
    return Gdate

# Local Mean Sideral Time from GMST, longitude
# returns a datetime object

def getLMSTfromGMST(gmst, longitude):
    # gmst datetime object
    # longitude in degrees
    # returns a datetime object
    lmst = gmst + timedelta(hours=longitude/15.)
    return lmst

# Greenwich Mean Sideral Time from LMST, longitude
# returns a datetime object

def getGMSTfromLMST(lmst, longitude):
    # lmst datetime object
    # longitude in degrees
    # returns a datetime object
    #gmst = datetime.time(hour=lmst) - timedelta(hours=longitude/15.)
    gmst = lmst - timedelta(hours=longitude/15.)
    return gmst
    

# deltaT, the difference between Terrestrial Time TT and
# Universal Time 
# see https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
# for Julian Date JD

def getDeltaT(JD): 
    # convert JD into year, month, day
    '''
    dtymd = getGreenwichDate(JD)
    year = dtymd.year
    month = dtymd.month
    y = year + (month-0.5)/12.
    # print("getDeltaT year, month, y = %d %d %.2f" % (year, month, y))
    '''
    (year, month, day) = hvfTime(JD).getDate()
    # We define the decimal year "y" as follows:
    y = year + (month - 0.5)/12

    if year < -500:
        u = (year-1820.)/100.
        deltaT = -20 + 32*u*u
    elif year < 500:
        u = y/100.
        deltaT = 10583.6 - 1014.41*u + 33.78311*u*u - 5.952053*u**3 \
                - 0.1798452*u**4 + 0.022174192*u**5 + 0.0090316521*u**6
    elif year < 1600:
        u = (y-1000.)/100.
        deltaT = 1574.2 - 556.01*u + 71.23472*u*u + 0.319781*u**3 \
                - 0.8503463*u**4 - 0.005050998*u**5 + 0.0083572073*u**6
    elif year < 1700:
        t = y - 1600.
        deltaT = 120. -0.9808*t - 0.01532*t*t + t**3/7129
    elif year < 1800:
        t = y - 1700.
        deltaT = 8.83 + 0.1603*t - 0.0059285*t*t + 0.00013336*t**3 \
                - t**4/1174000.
    elif year < 1860:
        t = y - 1800.
        deltaT = 13.72 - 0.332447*t + 0.0068612*t*t + 0.0041116*t**3 \
                - 0.00037436*t**4 + 0.0000121272*t**5 \
                - 0.0000001699*t**6 + 0.000000000875*t**7
    elif year < 1900:
        t = y - 1860.
        deltaT = 7.62 + 0.5737*t - 0.251754*t*t + 0.01680668*t**3 \
                - 0.0004473624*t**4 + t**5/233174.
    elif year < 1920:
        t = y - 1900.
        deltaT = -2.79 + 1.494119*t - 0.0598939*t*t + 0.0061966*t**3 \
                -0.000197*t**4
    elif year < 1941:
        t = y - 1920.
        deltaT = 21.20 + 0.84493*t - 0.076100*t*t + 0.0020936*t**3
    elif year < 1961:
        t = y - 1950.
        deltaT = 29.07 + 0.407*t - t*t/233. + t**3/2547.
    elif year < 1986:
        t = y - 1975.
        deltaT = 45.45 + 1.067*t - t*t/260. - t**3/718.
    elif year < 2005:
        t = y - 2000.
        deltaT = 63.86 + 0.3345*t - 0.060374*t*t + 0.0017275*t**3 \
                + 0.000651814*t**4 + 0.00002373599*t**5
    elif year < 2050:
        t = y - 2000.
        deltaT = 62.92 + 0.32217*t + 0.005589*t*t
    elif year < 2150:
        deltaT = -20. + 32*((y-1820.)/100)**2 - 0.5628*(2150.-y)
    else:   # after 2150
        u = (year-1820.)/100.
        deltaT = -20. + 32.*u*u
        # https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
        # small correction beyond 2150 (only needed for Perihelion calc's)
        c = -0.000012932 * (y - 1955.)**2
        deltaT += c

    return deltaT

def getEps(t):  # Schiefe der Ekliptik in Grad
    # t in Tage nach Epoch2010
    # Resultat in Grad
    T = t / 36525
    eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
    return eps


#######################################################################
# Jean Meeus, Astronomical Algorithms, Willmann-Bell, Richmond VA, 1991
#
# hvf 03.12.2020

# quadratic interpolation
# xi 3 Stuetzstellen, yi 3 Funktionswerte
# Meeus p. 25
def interp(xi, yi, x):
    a10 = (yi[1]-yi[0])/(xi[1]-xi[0])
    a21 = (yi[2]-yi[1])/(xi[2]-xi[1])
    a210 = (a21-a10)/(xi[2]-xi[0])
    res = yi[0] + a10*(x-xi[0]) + a210*(x-xi[0])*(x-xi[1])
    return res

#############################################################################

from datetime import datetime
global EarthdtUTimes
EarthdtUTimes = {}  # empty dict of times
#!/usr/bin/env python
# coding: utf-8
#   getEarthXYZData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Earth"
#       X     :  X coord (unit: au)
#       Y     :  Y coord (unit: au)
#       Z     :  Z coord (unit: au)
#       Long  :  heliocentric longitude, degrees
#       Lat   :  heliocentric latitude, degrees
#       Rau   :  radius, distance Object - Sun (unit: au)
#
#   Similarly for each planet
#   data from VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.1 hvf 2022-11-03 getEarthXYZData.py created by mkPlanetXYZData.py"
VersionList.append(Version)

import math

def getEarthXYZData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. 31.1 Julian Millennia

#######################################################################


    #strdtUT = dtUT.isoformat(sep=' ')     # convert to string and use as key to dict
    #strdtUT = str(dtUT)     # usually yyyy-mm-dd HH:MM:SS.sss
    strdtUT = dtUT.getStr()     # usually yyyy-mm-dd HH:MM:SS.sss
    # check if we have been here before
    if strdtUT in EarthdtUTimes:
        Data = EarthdtUTimes[strdtUT]
        # print ("getEarthXYZData using stored data %s" % strdtUT)
        return Data # X, Y, Z, Dict
    # print ("getEarthXYZData calculate fresh: dtUT = %s" % strdtUT)

    # EARTH X Coord
    sumX0  = 0
    sumX0  +=  99986069.925 * cos(1.75347045757  + 6283.31966747490 *T)
    sumX0  +=  2506324.281 * cos(4.93819429098  + 0.24381748350 *T)
    sumX0  +=  835274.807 * cos(1.71033525539  + 12566.39551746630 *T)
    sumX0  +=   10466.796 * cos(1.66721984219  + 18849.47136745770 *T)
    sumX0  +=    3490.608 * cos(4.44373803231  + 6282.83203250789 *T)
    sumX0  +=    3110.838 * cos(0.66875189331  + 83997.09113559539 *T)
    sumX0  +=    2561.408 * cos(0.58588607490  + 529.93478257810 *T)
    sumX0  +=    2142.365 * cos(1.09204474884  + 1577.09972496430 *T)
    sumX0  +=    1709.161 * cos(0.49540863237  + 6279.79654912590 *T)
    sumX0  +=    1707.934 * cos(6.15314019418  + 6286.84278582391 *T)
    sumX0  +=    1442.753 * cos(3.47210398336  + 2353.10997125530 *T)
    sumX0  +=    1113.027 * cos(3.69621650479  + 5223.45010231870 *T)
    sumX0  +=     934.484 * cos(6.07385378286  + 12036.70455237170 *T)
    sumX0  +=     899.854 * cos(3.17607463681  + 10213.52936369450 *T)
    sumX0  +=     566.820 * cos(2.15241946891  + 1059.62574767270 *T)
    sumX0  +=     661.858 * cos(1.31175222119  + 5753.62870238030 *T)
    sumX0  +=     739.644 * cos(4.36662524112  + 398.39282089170 *T)
    sumX0  +=     681.381 * cos(2.21815397940  + 4705.97612502710 *T)
    sumX0  +=     611.293 * cos(5.38470180335  + 6813.01063256950 *T)
    sumX0  +=     451.129 * cos(6.09315891204  + 5885.17066406670 *T)
    sumX0  +=     451.985 * cos(1.27931036318  + 6257.02134767510 *T)
    sumX0  +=     449.743 * cos(5.36941929064  + 6309.61798727471 *T)
    sumX0  +=     406.291 * cos(0.54369369385  + 6681.46867088311 *T)
    sumX0  +=     541.115 * cos(0.78670634299  + 775.76642880750 *T)
    sumX0  +=     546.682 * cos(1.46109463961  + 14143.73905991410 *T)
    sumX0  +=     513.780 * cos(4.43695037790  + 7860.66320992270 *T)
    sumX0  +=     263.916 * cos(5.39551752630  + 5507.79705615090 *T)
    sumX0  +=     217.220 * cos(4.51265697015  + 11507.01358727710 *T)
    sumX0  +=     292.146 * cos(2.51372357301  + 5507.30942118390 *T)
    sumX0  +=     227.915 * cos(1.23940916733  + 7058.84227879890 *T)
    sumX0  +=     226.835 * cos(3.27356345008  + 4693.75913722409 *T)
    sumX0  +=     255.833 * cos(2.26545070197  + 12168.24651405810 *T)
    sumX0  +=     256.157 * cos(1.45478864804  + 710.17686604180 *T)
    sumX0  +=     178.679 * cos(2.96645630109  + 796.54182429990 *T)
    sumX0  +=     178.303 * cos(0.40472561766  + 6283.25235717211 *T)
    sumX0  +=     178.303 * cos(6.24380795499  + 6283.38697777770 *T)
    sumX0  +=     155.471 * cos(1.62414733243  + 25132.54721744910 *T)
    sumX0  +=     155.246 * cos(1.41969109555  + 5487.02166065850 *T)
    sumX0  +=     207.640 * cos(5.85326148962  + 11790.87290614230 *T)
    sumX0  +=     199.988 * cos(4.07216830110  + 17790.08943726851 *T)
    sumX0  +=     128.970 * cos(5.21698329136  + 7079.61767429131 *T)
    sumX0  +=     128.155 * cos(4.80258701251  + 3739.00524759150 *T)
    sumX0  +=     152.241 * cos(0.86901699174  + 213.54291292150 *T)
    sumX0  +=      78.109 * cos(1.84890519772  + 8827.63408735830 *T)
    sumX0  +=      73.503 * cos(2.69517697923  + 1589.31671276730 *T)
    sumX0  +=      93.001 * cos(3.93777887053  + 9437.51911740350 *T)
    sumX0  +=      65.526 * cos(3.65568333160  + 11770.09751064990 *T)
    sumX0  +=      63.240 * cos(2.24124127597  + 6262.54427198250 *T)
    sumX0  +=      62.961 * cos(4.40729385308  + 6304.09506296730 *T)
    sumX0  +=      85.870 * cos(3.02430778985  + 167284.00540514899 *T)
    sumX0  +=      83.233 * cos(6.19780778434  + 3340.85624418330 *T)
    sumX0  +=      62.850 * cos(2.03823135444  + 7.35736448430 *T)
    sumX0  +=      79.971 * cos(2.49335782552  + 6070.02057203690 *T)
    sumX0  +=      60.043 * cos(3.39955163445  + 4137.15425099970 *T)
    sumX0  +=      75.613 * cos(4.15948190002  + 6496.61876291290 *T)
    sumX0  +=      53.731 * cos(1.56268892040  + 1194.69082770810 *T)
    sumX0  +=      57.574 * cos(3.96971590315  + 6284.29998854310 *T)
    sumX0  +=      57.574 * cos(2.67881766950  + 6282.33934640670 *T)
    sumX0  +=      72.385 * cos(4.85114866885  + 10976.83498721549 *T)
    sumX0  +=      51.135 * cos(1.31237333336  + 6290.43321447570 *T)
    sumX0  +=      50.604 * cos(5.34520074613  + 6276.20612047410 *T)
    sumX0  +=      49.613 * cos(2.64370047145  + 6127.89926804070 *T)
    sumX0  +=      48.009 * cos(4.00571885519  + 6438.74006690910 *T)
    sumX0  +=      59.233 * cos(3.78322753249  + 3154.44326741210 *T)
    sumX0  +=      48.264 * cos(4.77126143981  + 801.57711364030 *T)
    sumX0  +=      47.368 * cos(6.10696378485  + 3128.63258257930 *T)
    sumX0  +=      39.225 * cos(3.25281924495  + 8429.48508395010 *T)
    sumX0  +=      51.307 * cos(5.78448747277  + 12353.09642202830 *T)
    sumX0  +=      50.639 * cos(2.05215225623  + 11856.46246890800 *T)
    sumX0  +=      39.555 * cos(4.93022654622  + 5481.49873635110 *T)
    sumX0  +=      42.757 * cos(3.64063452002  + 17260.39847217390 *T)
    sumX0  +=      48.941 * cos(0.45355151773  + 9225.78309076650 *T)
    sumX0  +=      48.302 * cos(0.60366334071  + 2544.07060239989 *T)
    sumX0  +=      37.971 * cos(3.04093006964  + 426.84200835950 *T)
    sumX0  +=      43.643 * cos(3.14688065147  + 5856.23384163189 *T)
    sumX0  +=      31.830 * cos(2.00885127406  + 2145.92159899169 *T)
    sumX0  +=      33.959 * cos(1.72230567102  + 7085.14059859870 *T)
    sumX0  +=      30.184 * cos(3.96882361878  + 10977.32262218251 *T)
    sumX0  +=      32.562 * cos(2.84647033523  + 4164.06817212950 *T)
    sumX0  +=      38.887 * cos(5.36022568355  + 13367.72881362310 *T)
    sumX0  +=      28.515 * cos(6.10933021457  + 12569.91863581531 *T)
    sumX0  +=      28.873 * cos(1.01228967614  + 3930.45351370310 *T)
    sumX0  +=      36.633 * cos(2.90798803424  + 18073.94875613370 *T)
    sumX0  +=      24.984 * cos(2.17919076736  + 26.05450231630 *T)
    sumX0  +=      26.364 * cos(0.54621601823  + 9438.00675237050 *T)
    sumX0  +=      30.326 * cos(1.99372292658  + 4535.30325440790 *T)
    sumX0  +=      28.166 * cos(0.48127325062  + 12562.87239911730 *T)
    sumX0  +=      28.001 * cos(2.82603231075  + 5088.87265725030 *T)
    sumX0  +=      21.157 * cos(5.07256589577  + 11371.94850724170 *T)
    sumX0  +=      23.113 * cos(0.85996930170  + 5223.93773728571 *T)
    sumX0  +=      21.642 * cos(4.63524157818  + 12559.28197046550 *T)
    sumX0  +=      19.448 * cos(4.40136974621  + 12565.90788249930 *T)
    sumX0  +=      26.055 * cos(0.30340363466  + 71430.45180064559 *T)
    sumX0  +=      18.232 * cos(3.41362095130  + 1747.77259558350 *T)
    sumX0  +=      23.248 * cos(6.19009992884  + 18319.78040236310 *T)
    sumX0  +=      22.785 * cos(1.16821848960  + 5753.14106741329 *T)
    sumX0  +=      17.091 * cos(0.58584172932  + 4933.45225781610 *T)
    sumX0  +=      19.399 * cos(4.43188457477  + 10447.14402212089 *T)
    sumX0  +=      21.067 * cos(3.80033599994  + 7477.76667769950 *T)
    sumX0  +=      21.310 * cos(4.65919931667  + 8031.33608054190 *T)
    sumX0  +=      14.189 * cos(1.30988546287  + 2942.70724077510 *T)
    sumX0  +=      13.864 * cos(4.59350621319  + 8636.18582124671 *T)
    sumX0  +=      18.884 * cos(2.42694418444  + 156137.71980228281 *T)
    sumX0  +=      17.875 * cos(0.15660631596  + 1592.83983111630 *T)
    sumX0  +=      18.602 * cos(0.50024326019  + 17297.93850984270 *T)
    sumX0  +=      13.605 * cos(5.19720916665  + 13096.08648256090 *T)
    sumX0  +=      15.064 * cos(3.12148038350  + 16496.60521368590 *T)
    sumX0  +=      12.969 * cos(6.06491705231  + 7633.18707713370 *T)
    sumX0  +=      14.403 * cos(1.43725922780  + 20426.81490990550 *T)
    sumX0  +=      13.311 * cos(3.60732173221  + 12139.79732659030 *T)
    sumX0  +=      13.645 * cos(5.14981984839  + 13368.21644859010 *T)
    sumX0  +=      11.583 * cos(5.46063496654  + 5331.60126122430 *T)
    sumX0  +=      11.405 * cos(4.40016025355  + 529.44714761109 *T)
    sumX0  +=      15.057 * cos(5.21434071429  + 7084.65296363170 *T)
    sumX0  +=      14.330 * cos(6.02775140319  + 7342.70159766410 *T)
    sumX0  +=      12.375 * cos(3.10941922715  + 6279.72923882311 *T)
    sumX0  +=      12.366 * cos(3.53911446993  + 6286.91009612670 *T)
    sumX0  +=      10.016 * cos(2.24542867964  + 397.90518592470 *T)
    sumX0  +=      12.435 * cos(1.06626484546  + 15110.70993734970 *T)
    sumX0  +=       9.513 * cos(1.19056065248  + 7235.03807372550 *T)
    sumX0  +=      10.997 * cos(2.19198404583  + 10989.05197501850 *T)
    sumX0  +=      11.583 * cos(4.05250824543  + 5729.75026463250 *T)
    sumX0  +=       9.034 * cos(5.42875830514  + 9623.93209417470 *T)
    sumX0  +=      11.842 * cos(4.16957446176  + 15721.08260236190 *T)
    sumX0  +=       8.869 * cos(5.04690919891  + 6148.25458743950 *T)
    sumX0  +=       8.614 * cos(1.60161199451  + 6418.38474751031 *T)
    sumX0  +=      10.312 * cos(2.59789642414  + 6836.88907031730 *T)
    sumX0  +=       8.411 * cos(2.44368872285  + 14712.56093394150 *T)
    sumX0  +=       7.911 * cos(3.15428722892  + 2119.00767786191 *T)
    sumX0  +=      10.379 * cos(4.81870974023  + 1349.62359217529 *T)
    sumX0  +=      10.374 * cos(2.40569215505  + 5486.53402569149 *T)
    sumX0  +=      10.480 * cos(2.95288921695  + 5999.46034860970 *T)
    sumX0  +=       7.404 * cos(5.41419049870  + 6040.59106350090 *T)
    sumX0  +=       7.357 * cos(3.83057351958  + 5088.38502228330 *T)
    sumX0  +=       9.820 * cos(3.69556241034  + 6567.17898634010 *T)
    sumX0  +=       7.018 * cos(1.23420241153  + 6526.04827144891 *T)
    sumX0  +=       9.668 * cos(1.92238869394  + 21228.14820606230 *T)
    sumX0  +=       7.545 * cos(1.23604547808  + 12540.09719766650 *T)
    sumX0  +=       7.202 * cos(5.88391537700  + 6245.29199483910 *T)
    sumX0  +=       7.144 * cos(0.76464656573  + 6321.34734011070 *T)
    sumX0  +=       8.883 * cos(3.89630774866  + 5642.93474619389 *T)
    sumX0  +=       6.626 * cos(0.70123567856  + 5327.71992586630 *T)
    sumX0  +=       7.015 * cos(4.13622360700  + 7860.17557495569 *T)
    sumX0  +=       6.159 * cos(0.94592300466  + 12964.54452087450 *T)
    sumX0  +=       7.413 * cos(5.03498725246  + 23543.47432216530 *T)
    sumX0  +=       6.355 * cos(5.03541099132  + 1990.98883452450 *T)
    sumX0  +=       7.286 * cos(3.58219614228  + 537.04832957890 *T)
    sumX0  +=       6.314 * cos(6.22485628824  + 951.47458876710 *T)
    sumX0  +=       6.257 * cos(4.23208117686  + 4690.72365384210 *T)
    sumX0  +=       6.484 * cos(3.20813291319  + 16730.70750707931 *T)
    sumX0  +=       5.803 * cos(6.09476150956  + 955.35592412510 *T)
    sumX0  +=       6.391 * cos(3.97601526750  + 24073.16528725990 *T)
    sumX0  +=       7.530 * cos(0.20298795021  + 10973.79950383350 *T)
    sumX0  +=       6.921 * cos(0.69969701816  + 522.82123557730 *T)
    sumX0  +=       7.626 * cos(4.29336813013  + 22004.15845235330 *T)
    sumX0  +=       7.531 * cos(3.51464444079  + 18422.87317658170 *T)
    sumX0  +=       6.902 * cos(2.75534627331  + 155.17658195069 *T)
    sumX0  +=       5.434 * cos(5.94625981048  + 7238.91940908350 *T)
    sumX0  +=       6.223 * cos(2.24396249102  + 18451.32236404950 *T)
    sumX0  +=       7.296 * cos(5.98982040270  + 16730.21987211229 *T)
    sumX0  +=       5.833 * cos(2.32930728499  + 640.14110379750 *T)
    sumX0  +=       7.171 * cos(6.14759594277  + 3929.96587873609 *T)
    sumX0  +=       6.251 * cos(2.19374266424  + 6277.79674316750 *T)
    sumX0  +=       6.251 * cos(4.45479090841  + 6288.84259178230 *T)
    sumX0  +=       5.683 * cos(3.49552871951  + 1551.28904013150 *T)
    sumX0  +=       6.383 * cos(0.64667497225  + 5216.33655531789 *T)
    sumX0  +=       6.194 * cos(3.56522608256  + 5230.56364931950 *T)
    sumX0  +=       5.036 * cos(3.83694551305  + 14314.41193053330 *T)
    sumX0  +=       6.774 * cos(1.99358309023  + 102.84895673509 *T)
    sumX0  +=       6.224 * cos(0.35231413534  + 11014.86265985130 *T)
    sumX0  +=       4.621 * cos(1.34668136864  + 553.32558535889 *T)
    sumX0  +=       6.293 * cos(2.32035231613  + 5650.53592816170 *T)
    sumX0  +=       6.297 * cos(4.40170593853  + 26088.14695905770 *T)
    sumX0  +=       5.186 * cos(0.71187314766  + 77714.01528560399 *T)
    sumX0  +=       6.196 * cos(2.58705857336  + 84672.23202702119 *T)
    sumX0  +=       6.160 * cos(4.78250009298  + 239424.63407183639 *T)
    sumX0  +=       5.414 * cos(2.99961957880  + 6180.22689325630 *T)
    sumX0  +=       5.413 * cos(3.65415218952  + 6386.41244169350 *T)
    sumX0  +=       4.863 * cos(3.76722359858  + 90280.16698558679 *T)
    sumX0  +=       5.711 * cos(4.32712250291  + 6916.10340678810 *T)
    sumX0  +=       5.150 * cos(4.10583483822  + 1577.58735993130 *T)
    sumX0  +=       4.269 * cos(2.38358764863  + 7875.91568110771 *T)
    sumX0  +=       4.017 * cos(3.78580903100  + 6254.87048000710 *T)
    sumX0  +=       4.069 * cos(2.42056660765  + 3634.37720703489 *T)
    sumX0  +=       3.993 * cos(2.86272128776  + 6311.76885494270 *T)
    sumX0  +=       4.016 * cos(5.24041079205  + 4690.23601887509 *T)
    sumX0  +=       5.290 * cos(3.34482936009  + 25158.35790228190 *T)
    sumX0  +=       4.831 * cos(2.83916520281  + 5760.74224938110 *T)
    sumX0  +=       3.487 * cos(3.51068129706  + 10447.63165708790 *T)
    sumX0  +=       4.120 * cos(3.29656037233  + 6709.91785835091 *T)
    sumX0  +=       4.410 * cos(3.80589939396  + 6805.89708556871 *T)
    sumX0  +=       4.023 * cos(6.22474978871  + 4731.78680985990 *T)
    sumX0  +=       3.980 * cos(3.31931976159  + 5856.72147659890 *T)
    sumX0  +=       3.173 * cos(5.19128151293  + 1066.73929467350 *T)
    sumX0  +=       3.293 * cos(4.01637480858  + 9917.45305702629 *T)
    sumX0  +=       2.935 * cos(3.67590980046  + 12592.69383726610 *T)
    sumX0  +=       2.979 * cos(6.20067573402  + 12566.46282776910 *T)
    sumX0  +=       3.980 * cos(1.19681754169  + 13341.91812879030 *T)
    sumX0  +=       3.892 * cos(5.52571426134  + 75.02541605080 *T)
    sumX0  +=       3.042 * cos(3.19425693799  + 18053.17336064130 *T)
    sumX0  +=       3.608 * cos(5.60952778146  + 6208.53806890760 *T)
    sumX0  +=       3.721 * cos(5.22104546686  + 5966.92779781830 *T)
    sumX0  +=       3.581 * cos(1.03995231220  + 6358.10126604220 *T)
    sumX0  +=       2.880 * cos(1.61225556126  + 10575.65050042530 *T)
    sumX0  +=       2.806 * cos(4.21551614220  + 5863.83502359970 *T)
    sumX0  +=       3.626 * cos(1.42632163253  + 6599.71153713150 *T)
    sumX0  +=       2.794 * cos(5.69269191706  + 8030.84844557490 *T)
    sumX0  +=       2.729 * cos(4.17319557124  + 3.76693583251 *T)
    sumX0  +=       3.625 * cos(3.00192313835  + 6020.23574410210 *T)
    sumX0  +=       3.206 * cos(2.15397091751  + 11506.52595231009 *T)
    sumX0  +=       3.556 * cos(4.28947576556  + 8428.99744898310 *T)
    sumX0  +=       2.856 * cos(3.33625143564  + 12721.81591690050 *T)
    sumX0  +=       2.611 * cos(2.43261867894  + 6702.80431135010 *T)
    sumX0  +=       2.535 * cos(1.58310702384  + 31415.62306744050 *T)
    sumX0  +=       3.520 * cos(5.37986371510  + 250570.91967470257 *T)
    sumX0  +=       3.434 * cos(3.64656923781  + 6546.40359084770 *T)
    sumX0  +=       2.572 * cos(4.79831420582  + 5884.68302909969 *T)
    sumX0  +=       2.647 * cos(3.81972272861  + 2352.62233628830 *T)
    sumX0  +=       3.026 * cos(5.24078074980  + 13916.26292712510 *T)
    sumX0  +=       2.378 * cos(3.62833713669  + 2389.13783793270 *T)
    sumX0  +=       2.979 * cos(0.36159339669  + 12566.32820716351 *T)
    sumX0  +=       3.082 * cos(1.77514222265  + 14945.07235607090 *T)
    sumX0  +=       2.716 * cos(3.06935389211  + 12029.59100537090 *T)
    sumX0  +=       2.215 * cos(5.26643739208  + 13362.69352428270 *T)
    sumX0  +=       3.032 * cos(4.76553829426  + 29088.56759850150 *T)
    sumX0  +=       2.421 * cos(3.15326621321  + 11015.35029481830 *T)
    sumX0  +=       2.275 * cos(5.95465292259  + 6077.13411903770 *T)
    sumX0  +=       2.466 * cos(5.59166943387  + 16200.52890701769 *T)
    sumX0  +=       2.529 * cos(5.94709131586  + 12043.81809937250 *T)
    sumX0  +=       2.015 * cos(2.48432239832  + 6262.96434807611 *T)
    sumX0  +=       2.218 * cos(0.69324011482  + 6489.50521591210 *T)
    sumX0  +=       2.005 * cos(4.16423892775  + 6303.67498687370 *T)
    sumX0  +=       2.010 * cos(4.11254992503  + 4700.87268422489 *T)
    sumX0  +=       2.641 * cos(4.28275048815  + 4590.66636300550 *T)
    sumX0  +=       2.438 * cos(6.12910205560  + 149.80701461810 *T)
    sumX0  +=       1.893 * cos(0.98057464900  + 6279.43833211690 *T)
    sumX0  +=       1.893 * cos(5.66795892365  + 6287.20100283290 *T)
    sumX0  +=       2.156 * cos(2.00902236290  + 18139.53831889940 *T)
    sumX0  +=       2.071 * cos(4.32179144410  + 11926.49823115230 *T)
    sumX0  +=       2.395 * cos(0.13148357341  + 1162.71852189130 *T)
    sumX0  +=       1.954 * cos(0.37275781655  + 13518.11392371690 *T)
    sumX0  +=       2.006 * cos(0.31048315304  + 3154.93090237910 *T)
    sumX0  +=       1.776 * cos(0.14855204887  + 13521.99525907490 *T)
    sumX0  +=       1.964 * cos(0.25103153440  + 4686.64559022330 *T)
    sumX0  +=       2.332 * cos(4.40649103371  + 10022.08109758290 *T)
    sumX0  +=       1.863 * cos(5.26030609599  + 242.97242145751 *T)
    sumX0  +=       1.928 * cos(4.59250450967  + 5746.51515537950 *T)
    sumX0  +=       2.436 * cos(4.40770823480  + 95143.37673846159 *T)
    sumX0  +=       2.163 * cos(1.41797907369  + 6037.48802124550 *T)
    sumX0  +=       2.250 * cos(2.39728922664  + 12669.48829168490 *T)
    sumX0  +=       2.052 * cos(5.23006608728  + 6529.15131370430 *T)
    sumX0  +=       1.886 * cos(3.61213584641  + 7238.43177411650 *T)
    sumX0  +=       2.170 * cos(5.49596575387  + 19651.29229858150 *T)
    sumX0  +=       2.058 * cos(4.69821533065  + 23013.78335707070 *T)
    sumX0  +=       1.774 * cos(2.08482289391  + 6820.12417957030 *T)
    sumX0  +=       1.841 * cos(0.34653293659  + 65147.37595065419 *T)
    sumX0  +=       1.578 * cos(5.64311385009  + 4292.57465043390 *T)
    sumX0  +=       1.577 * cos(0.80779559190  + 7632.69944216670 *T)
    sumX0  +=       1.541 * cos(1.80359816798  + 13119.96492030870 *T)
    sumX0  +=       2.012 * cos(0.36995312887  + 4292.08701546690 *T)
    sumX0  +=       1.543 * cos(0.68755069426  + 6293.95633282471 *T)
    sumX0  +=       1.543 * cos(5.96098287839  + 6272.68300212510 *T)
    sumX0  +=       1.513 * cos(0.65061890512  + 20.53157800890 *T)
    sumX0  +=       1.489 * cos(2.95205535091  + 110.45013870291 *T)
    sumX0  +=       1.639 * cos(4.44076102657  + 17655.02435723310 *T)
    sumX0  +=       1.443 * cos(2.26329309493  + 1052.51220067191 *T)
    sumX0  +=       1.337 * cos(3.06401553951  + 2544.55823736690 *T)
    sumX0  +=       1.797 * cos(6.18654785499  + 33018.77729472110 *T)
    sumX0  +=       1.488 * cos(5.67450527487  + 76.50988875911 *T)
    sumX0  +=       1.453 * cos(5.46078654600  + 633.02755679670 *T)
    sumX0  +=       1.349 * cos(5.75097749599  + 6016.71262575310 *T)
    sumX0  +=       1.477 * cos(4.14004774348  + 12779.69461290430 *T)
    sumX0  +=       1.688 * cos(3.20087734666  + 18875.28205229050 *T)
    sumX0  +=       1.491 * cos(6.22974184712  + 9411.70843257070 *T)
    sumX0  +=       1.471 * cos(5.75779240419  + 18636.17227201970 *T)
    sumX0  +=       1.263 * cos(0.91728971350  + 12573.50906446710 *T)
    sumX0  +=       1.685 * cos(5.70877589839  + 25934.36814857290 *T)
    sumX0  +=       1.375 * cos(1.08927160683  + 2378.92065608810 *T)
    sumX0  +=       1.641 * cos(6.27664360507  + 6.86972951729 *T)
    sumX0  +=       1.200 * cos(1.47556274227  + 11499.90004027630 *T)
    sumX0  +=       1.272 * cos(0.89738913067  + 6549.92670919670 *T)
    sumX0  +=       1.214 * cos(1.22660393299  + 10973.31186886650 *T)
    sumX0  +=       1.406 * cos(5.67596795657  + 11790.38527117530 *T)
    sumX0  +=       1.524 * cos(4.27066787572  + 18209.57408114370 *T)
    sumX0  +=       1.149 * cos(3.01961706159  + 10177.50149701710 *T)
    sumX0  +=       1.547 * cos(4.02915828753  + 11926.01059618529 *T)
    sumX0  +=       1.460 * cos(4.97235137781  + 246.07546371290 *T)
    sumX0  +=       1.323 * cos(1.38571450422  + 8661.99650607950 *T)
    sumX0  +=       1.252 * cos(1.41166880118  + 6993.25271603320 *T)
    sumX0  +=       1.459 * cos(5.29807880334  + 38.37685312130 *T)
    sumX0  +=       1.235 * cos(2.84096519816  + 24357.02460612510 *T)
    sumX0  +=       1.028 * cos(2.51029997292  + 6112.64679685570 *T)
    sumX0  +=       1.330 * cos(1.00065727327  + 2146.40923395870 *T)
    sumX0  +=       1.312 * cos(1.10091584033  + 12491.61391889900 *T)
    sumX0  +=       1.246 * cos(4.66430542078  + 5429.63565075589 *T)
    sumX0  +=       0.992 * cos(4.13817744468  + 6453.99253809410 *T)
    sumX0  +=       0.988 * cos(0.97969587833  + 8274.06468451590 *T)
    sumX0  +=       1.365 * cos(6.11146862666  + 11371.46087227470 *T)
    sumX0  +=       1.020 * cos(2.82319817767  + 4732.27444482690 *T)
    sumX0  +=       1.191 * cos(3.43008039922  + 6290.36590417291 *T)
    sumX0  +=       1.191 * cos(3.21845317343  + 6276.27343077690 *T)
    sumX0  +=       0.978 * cos(4.73852758176  + 6247.29180079750 *T)
    sumX0  +=       0.971 * cos(1.91000724147  + 6319.34753415231 *T)
    sumX0  +=       0.962 * cos(2.63568544853  + 12565.41519639810 *T)
    sumX0  +=       1.025 * cos(2.21768401152  + 12545.62012197390 *T)
    sumX0  +=       0.990 * cos(0.39039931315  + 4694.24677219110 *T)
    sumX0  +=       0.964 * cos(1.79192739930  + 3893.93801205869 *T)
    sumX0  +=       0.944 * cos(5.86379546667  + 6259.44122972711 *T)
    sumX0  +=       0.944 * cos(0.78473810598  + 6307.19810522270 *T)
    sumX0  +=       1.092 * cos(3.92065202481  + 17797.20298426930 *T)
    sumX0  +=       1.132 * cos(1.00628522970  + 17782.97589026770 *T)
    sumX0  +=       0.859 * cos(1.48109589441  + 1692.40948698591 *T)
    sumX0  +=       1.213 * cos(4.04235990086  + 82576.73740351178 *T)
    sumX0  +=       0.942 * cos(5.22790737110  + 6298.57213865991 *T)
    sumX0  +=       0.942 * cos(1.42062620155  + 6268.06719628990 *T)
    sumX0  +=       1.045 * cos(0.46195584294  + 15508.85894075790 *T)
    sumX0  +=       0.851 * cos(5.19678367921  + 6173.11334625550 *T)
    sumX0  +=       0.850 * cos(1.45204280805  + 6393.52598869430 *T)
    sumX0  +=       0.895 * cos(2.40089211752  + 3904.15519390330 *T)
    sumX0  +=       1.020 * cos(0.71539416375  + 220.65645992230 *T)
    sumX0  +=       0.845 * cos(5.75499771232  + 17256.87535382490 *T)
    sumX0  +=       0.863 * cos(4.43158774725  + 949.41942645330 *T)
    sumX0  +=       0.853 * cos(2.75614337272  + 16201.01654198470 *T)
    sumX0  +=       0.811 * cos(6.13459778682  + 4803.96545844350 *T)
    sumX0  +=       0.966 * cos(4.14597367445  + 206.42936592071 *T)
    sumX0  +=       0.917 * cos(2.47007329129  + 149854.64395229140 *T)
    sumX0  +=       1.083 * cos(1.32514794865  + 36948.98699094070 *T)
    sumX0  +=       0.779 * cos(3.58339682996  + 2648.69864295650 *T)
    sumX0  +=       0.964 * cos(0.77431665261  + 796.05418933290 *T)
    sumX0  +=       0.995 * cos(0.25423339373  + 11403.92081305850 *T)
    sumX0  +=       0.962 * cos(3.92658368218  + 12567.37583853451 *T)
    sumX0  +=       0.779 * cos(0.45683397157  + 10213.04172872750 *T)
    sumX0  +=       0.980 * cos(4.62590378445  + 22805.49174851010 *T)
    sumX0  +=       0.930 * cos(2.22316684933  + 2787.28684134090 *T)
    sumX0  +=       0.829 * cos(5.09643797682  + 5120.35732810009 *T)
    sumX0  +=       0.733 * cos(2.60949856930  + 10575.16286545830 *T)
    sumX0  +=       0.743 * cos(3.79905998257  + 7834.36489012290 *T)
    sumX0  +=       0.818 * cos(1.64368328490  + 5572.89898394960 *T)
    sumX0  +=       0.905 * cos(3.81538816339  + 6284.80414018320 *T)
    sumX0  +=       0.905 * cos(2.83314540925  + 6281.83519476660 *T)
    sumX0  +=       0.798 * cos(2.62501051906  + 12410.97511803210 *T)
    sumX0  +=       0.817 * cos(0.52083502825  + 12416.83232033170 *T)
    sumX0  +=       0.884 * cos(0.84319899817  + 22483.60475700909 *T)
    sumX0  +=       0.755 * cos(3.82941056281  + 4060.97539791089 *T)
    sumX0  +=       0.651 * cos(3.93683821152  + 17259.91083720689 *T)
    sumX0  +=       0.616 * cos(5.70474451820  + 1596.43025976811 *T)
    sumX0  +=       0.651 * cos(2.41491885723  + 1748.26023055050 *T)
    sumX0  +=       0.723 * cos(3.06743374106  + 161000.92955515761 *T)
    sumX0  +=       0.784 * cos(0.87758060898  + 4907.05823266209 *T)
    sumX0  +=       0.784 * cos(2.20169255256  + 7234.55043875850 *T)
    sumX0  +=       0.785 * cos(3.43654258027  + 846.32665223470 *T)
    sumX0  +=       0.611 * cos(1.53152362816  + 853.44019923550 *T)
    sumX0  +=       0.576 * cos(1.22798260830  + 12323.66691349230 *T)
    sumX0  +=       0.570 * cos(4.54665961586  + 12587.17091295870 *T)
    sumX0  +=       0.576 * cos(3.06908128494  + 13915.77529215809 *T)
    sumX0  +=       0.770 * cos(5.60311452827  + 6069.53293706989 *T)
    sumX0  +=       0.773 * cos(0.21363273558  + 6133.26883537330 *T)
    sumX0  +=       0.581 * cos(1.34025743471  + 11933.61177815310 *T)
    sumX0  +=       0.719 * cos(0.90840576142  + 15720.59496739490 *T)
    sumX0  +=       0.578 * cos(4.26997200274  + 8662.48414104651 *T)
    sumX0  +=       0.536 * cos(6.06660590951  + 18852.99448580670 *T)
    sumX0  +=       0.633 * cos(1.17258360285  + 52176.05010063190 *T)
    sumX0  +=       0.581 * cos(2.92798889289  + 5334.14405850510 *T)
    sumX0  +=       0.532 * cos(4.58762178742  + 18842.35782045690 *T)
    sumX0  +=       0.657 * cos(3.25871870022  + 5849.12029463110 *T)
    sumX0  +=       0.599 * cos(2.11793230261  + 6151.77770578850 *T)
    sumX0  +=       0.540 * cos(3.24412649701  + 6286.60602489270 *T)
    sumX0  +=       0.540 * cos(3.40440707563  + 6280.03331005710 *T)
    sumX0  +=       0.628 * cos(0.31668062965  + 17298.42614480970 *T)
    sumX0  +=       0.532 * cos(4.41105068397  + 11514.12713427790 *T)
    sumX0  +=       0.513 * cos(3.70291605445  + 12456.18919624690 *T)
    sumX0  +=       0.649 * cos(4.90568628758  + 11764.57458634250 *T)
    sumX0  +=       0.582 * cos(4.53092429894  + 6414.86162916130 *T)
    sumX0  +=       0.517 * cos(4.06133972007  + 3340.36860921629 *T)
    sumX0  +=       0.609 * cos(3.94494306228  + 10420.23010099111 *T)
    sumX0  +=       0.615 * cos(4.71270964329  + 10983.94853421629 *T)
    sumX0  +=       0.622 * cos(5.53858241889  + 5326.54287653730 *T)
    sumX0  +=       0.536 * cos(3.70224967645  + 7232.49527644471 *T)
    sumX0  +=       0.516 * cos(2.92986975188  + 433.95555536030 *T)
    sumX0  +=       0.613 * cos(1.79095811234  + 10969.72144021469 *T)
    sumX0  +=       0.583 * cos(6.15311267171  + 5863.34738863270 *T)
    sumX0  +=       0.636 * cos(6.06042539245  + 26735.70144472970 *T)
    sumX0  +=       0.665 * cos(2.74682435671  + 40879.19668716030 *T)
    sumX0  +=       0.583 * cos(2.82325235087  + 12592.20620229910 *T)
    sumX0  +=       0.479 * cos(1.33548820034  + 5547.44315394310 *T)
    sumX0  +=       0.571 * cos(2.58673571504  + 6062.90702503610 *T)
    sumX0  +=       0.565 * cos(3.92092198777  + 4171.18171913030 *T)
    sumX0  +=       0.462 * cos(1.96630641591  + 3104.68624194030 *T)
    sumX0  +=       0.551 * cos(4.03874430699  + 6503.73230991370 *T)
    sumX0  +=       0.504 * cos(5.14547341413  + 15670.83794192309 *T)
    sumX0  +=       0.535 * cos(6.12277243019  + 173567.08125514039 *T)
    sumX0  +=       0.460 * cos(3.18401450277  + 3495.78900865049 *T)
    sumX0  +=       0.449 * cos(5.09833649130  + 4274.27449334889 *T)
    sumX0  +=       0.453 * cos(3.57566005736  + 9387.76209193169 *T)
    sumX0  +=       0.527 * cos(6.24827137997  + 24602.85625235450 *T)
    sumX0  +=       0.502 * cos(0.97407216668  + 12490.12944619070 *T)
    sumX0  +=       0.603 * cos(0.85487069842  + 322711.54834138998 *T)
    sumX0  +=       0.532 * cos(1.83659178149  + 5120.84496306710 *T)
    sumX0  +=       0.530 * cos(0.43491203890  + 18845.94824910870 *T)
    sumX0  +=       0.431 * cos(5.31145947559  + 7019.19618100671 *T)
    sumX0  +=       0.465 * cos(5.83298981980  + 8827.14645239129 *T)
    sumX0  +=       0.559 * cos(1.71560880010  + 1582.20316576650 *T)
    sumX0  +=       0.485 * cos(6.13610027053  + 29296.85920706210 *T)
    sumX0  +=       0.584 * cos(0.07138791653  + 72850.80553272919 *T)
    sumX0  +=       0.550 * cos(3.95560592819  + 213.05527795450 *T)
    sumX0  +=       0.481 * cos(1.86946119801  + 14.47091148511 *T)
    sumX0  +=       0.569 * cos(2.95240683895  + 97238.87136197100 *T)
    sumX0  +=       0.551 * cos(1.64539130884  + 14313.92429556630 *T)
    sumX0  +=       0.389 * cos(3.43047350256  + 6245.18663183710 *T)
    sumX0  +=       0.389 * cos(3.21806007009  + 6321.45270311270 *T)
    sumX0  +=       0.388 * cos(1.13963224225  + 6297.54676147650 *T)
    sumX0  +=       0.388 * cos(5.50890133040  + 6269.09257347330 *T)
    sumX0  +=       0.515 * cos(1.67075264819  + 12320.56387123691 *T)
    sumX0  +=       0.394 * cos(6.10437801091  + 4156.95462512869 *T)
    sumX0  +=       0.491 * cos(5.96788884074  + 1479.11039154791 *T)
    sumX0  +=       0.440 * cos(0.97316688838  + 5650.04829319470 *T)
    sumX0  +=       0.382 * cos(3.07660544855  + 9917.94069199330 *T)
    sumX0  +=       0.503 * cos(4.65767245093  + 17157.30569795530 *T)
    sumX0  +=       0.439 * cos(4.82562869150  + 233141.55822184496 *T)
    sumX0  +=       0.381 * cos(2.02484381212  + 14143.25142494710 *T)
    sumX0  +=       0.371 * cos(4.59067170554  + 5643.42238116090 *T)
    sumX0  +=       0.516 * cos(5.61424407240  + 135.30889751891 *T)
    sumX0  +=       0.417 * cos(3.74461559548  + 13760.84252769090 *T)
    sumX0  +=       0.483 * cos(4.41750912984  + 9779.35249360890 *T)
    sumX0  +=       0.381 * cos(4.36951119319  + 14919.26167123810 *T)
    sumX0  +=       0.378 * cos(4.51952349691  + 17267.51201917470 *T)
    sumX0  +=       0.393 * cos(1.56678612595  + 7872.39256275871 *T)
    sumX0  +=       0.376 * cos(4.45581102621  + 13517.62628874989 *T)
    sumX0  +=       0.365 * cos(2.11020607854  + 6923.21695378890 *T)
    sumX0  +=       0.371 * cos(5.76233679446  + 13625.77744765550 *T)
    sumX0  +=       0.355 * cos(5.24317955258  + 10874.22984796390 *T)
    sumX0  +=       0.499 * cos(2.72562606986  + 161710.86260371591 *T)
    sumX0  +=       0.379 * cos(0.81621597880  + 3185.43584474911 *T)
    sumX0  +=       0.471 * cos(4.78294711940  + 11712.71150074729 *T)
    sumX0  +=       0.417 * cos(3.33073591690  + 22779.68106367730 *T)
    sumX0  +=       0.485 * cos(1.35444198321  + 12528.26248182851 *T)
    sumX0  +=       0.358 * cos(3.49115357527  + 6295.04902031090 *T)
    sumX0  +=       0.358 * cos(3.15737999738  + 6271.59031463890 *T)
    sumX0  +=       0.365 * cos(3.62050447597  + 6836.40143535029 *T)
    sumX0  +=       0.351 * cos(2.21887831196  + 11617.21990849651 *T)
    sumX0  +=       0.351 * cos(2.96356670458  + 205.94173095370 *T)
    sumX0  +=       0.429 * cos(0.79790217165  + 3894.42564702570 *T)
    sumX0  +=       0.388 * cos(0.49080052992  + 956.53297345411 *T)
    sumX0  +=       0.464 * cos(0.69464877419  + 23581.50199480110 *T)
    sumX0  +=       0.396 * cos(2.35908361907  + 5231.05128428650 *T)
    sumX0  +=       0.421 * cos(4.82739444320  + 7445.79437188270 *T)
    sumX0  +=       0.366 * cos(0.62612486089  + 17253.28492517310 *T)
    sumX0  +=       0.360 * cos(0.81896878015  + 21393.78578734110 *T)
    sumX0  +=       0.336 * cos(1.41330667750  + 6279.38751421180 *T)
    sumX0  +=       0.336 * cos(5.23522689515  + 6287.25182073800 *T)
    sumX0  +=       0.326 * cos(4.93781139708  + 1059.13811270570 *T)
    sumX0  +=       0.395 * cos(0.53943839302  + 5642.44206009270 *T)
    sumX0  +=       0.450 * cos(0.23150267191  + 1385.31775746760 *T)
    sumX0  +=       0.363 * cos(4.29615345291  + 22484.09239197610 *T)
    sumX0  +=       0.399 * cos(0.80548840004  + 16858.72635041670 *T)
    sumX0  +=       0.316 * cos(4.03075171777  + 20995.63678393290 *T)
    sumX0  +=       0.430 * cos(2.39021484637  + 19650.80466361450 *T)
    sumX0  +=       0.363 * cos(4.29069778349  + 7335.58805066330 *T)
    sumX0  +=       0.317 * cos(1.52889301942  + 11769.60987568290 *T)
    sumX0  +=       0.395 * cos(5.39427984473  + 5905.94605955911 *T)
    sumX0  +=       0.324 * cos(3.70774573147  + 37.78385515230 *T)
    sumX0  +=       0.299 * cos(3.86306301151  + 641.12142486571 *T)
    sumX0  +=       0.304 * cos(0.15985948407  + 5750.10558403130 *T)
    sumX0  +=       0.298 * cos(3.82763680613  + 1350.11122714230 *T)
    sumX0  +=       0.414 * cos(4.17890604161  + 44809.40638337990 *T)
    sumX0  +=       0.339 * cos(4.29176984719  + 3.68833577960 *T)
    sumX0  +=       0.351 * cos(5.93395494462  + 12345.98287502750 *T)
    sumX0  +=       0.362 * cos(0.42214982034  + 21953.91379191449 *T)
    sumX0  +=       0.304 * cos(4.93815385347  + 29826.55017215670 *T)
    sumX0  +=       0.395 * cos(0.07721971830  + 4176.28515993250 *T)
    sumX0  +=       0.320 * cos(2.49901226571  + 10818.37910439930 *T)
    sumX0  +=       0.384 * cos(4.01872268681  + 10177.01386205009 *T)
    sumX0  +=       0.279 * cos(0.96409855547  + 10970.20907518170 *T)
    sumX0  +=       0.365 * cos(1.25425361388  + 6660.69327539070 *T)
    sumX0  +=       0.377 * cos(0.82514906124  + 29864.57784479250 *T)
    sumX0  +=       0.275 * cos(5.73507538514  + 20597.48778052470 *T)
    sumX0  +=       0.316 * cos(2.59688143484  + 316.63568714010 *T)
    sumX0  +=       0.339 * cos(6.10829032638  + 6924.19727485710 *T)
    sumX0  +=       0.386 * cos(2.29032835664  + 2636.96929012050 *T)
    sumX0  +=       0.273 * cos(1.09519308783  + 26709.89075989690 *T)
    sumX0  +=       0.314 * cos(1.47223989608  + 14945.55999103790 *T)
    sumX0  +=       0.281 * cos(4.86653427016  + 16858.23871544969 *T)
    sumX0  +=       0.287 * cos(3.30150277561  + 18073.46112116669 *T)
    sumX0  +=       0.265 * cos(5.11010310272  + 19379.16233255230 *T)
    sumX0  +=       0.325 * cos(2.49278036693  + 12360.20996902910 *T)
    sumX0  +=       0.362 * cos(1.18157001338  + 30665.91114094930 *T)
    sumX0  +=       0.269 * cos(0.20537228942  + 6816.53375091851 *T)
    sumX0  +=       0.344 * cos(3.51559724211  + 6147.69434246491 *T)
    sumX0  +=       0.255 * cos(4.94623609008  + 1376.01761732930 *T)
    sumX0  +=       0.334 * cos(3.13303124260  + 6418.94499248490 *T)
    sumX0  +=       0.316 * cos(5.54205330844  + 6055.84353468590 *T)
    sumX0  +=       0.283 * cos(4.23095234926  + 28287.23430234470 *T)
    sumX0  +=       0.305 * cos(4.26252488384  + 16522.41589851870 *T)
    sumX0  +=       0.270 * cos(0.05215867170  + 283.61550138170 *T)
    sumX0  +=       0.243 * cos(2.54680473739  + 6255.91811137810 *T)
    sumX0  +=       0.243 * cos(4.10172883526  + 6310.72122357170 *T)
    sumX0  +=       0.296 * cos(0.07899510796  + 6129.54085699010 *T)
    sumX0  +=       0.301 * cos(1.10648037942  + 6510.79580026390 *T)
    sumX0  +=       0.329 * cos(1.15683223901  + 377.61742539930 *T)
    sumX0  +=       0.267 * cos(3.42011505320  + 24705.94902657310 *T)
    sumX0  +=       0.285 * cos(2.50494570372  + 5469.76938351510 *T)
    sumX0  +=       0.286 * cos(0.28647364691  + 6437.09847795970 *T)
    sumX0  +=       0.311 * cos(3.21063889136  + 11720.31268271510 *T)
    sumX0  +=       0.319 * cos(4.71059907821  + 169379.50002865840 *T)
    sumX0  +=       0.241 * cos(2.99439330770  + 632.53992182970 *T)
    sumX0  +=       0.229 * cos(1.01522459538  + 1265.81129610991 *T)
    sumX0  +=       0.235 * cos(3.36978882786  + 4487.57358878689 *T)
    sumX0  +=       0.253 * cos(0.51356087882  + 4377.36726756750 *T)
    sumX0  +=       0.277 * cos(1.58719268637  + 419.24082639170 *T)
    sumX0  +=       0.234 * cos(5.12574585310  + 11713.19913571430 *T)
    sumX0  +=       0.216 * cos(5.01445941939  + 10454.25756912169 *T)
    sumX0  +=       0.241 * cos(5.44670796756  + 103.33659170210 *T)
    sumX0  +=       0.214 * cos(2.09481715597  + 2222.10045208050 *T)
    sumX0  +=       0.213 * cos(3.79438383373  + 30356.24113725130 *T)
    sumX0  +=       0.240 * cos(2.57405501441  + 6309.13035230770 *T)
    sumX0  +=       0.212 * cos(2.36637350989  + 262.84010588929 *T)
    sumX0  +=       0.259 * cos(3.21725974670  + 6283.56348495841 *T)
    sumX0  +=       0.259 * cos(3.43127382595  + 6283.07584999140 *T)
    sumX0  +=       0.222 * cos(1.42789121927  + 10440.03047512009 *T)
    sumX0  +=       0.201 * cos(5.11101601554  + 5746.02752041250 *T)
    sumX0  +=       0.274 * cos(3.85387261790  + 23581.01435983410 *T)
    sumX0  +=       0.243 * cos(4.14346519403  + 7096.86995143470 *T)
    sumX0  +=       0.234 * cos(5.23697266935  + 5573.38661891660 *T)
    sumX0  +=       0.203 * cos(0.00533428175  + 16460.08971204149 *T)
    sumX0  +=       0.246 * cos(5.83654379480  + 8672.21368792411 *T)
    sumX0  +=       0.216 * cos(4.93054195434  + 5437.23683272371 *T)
    sumX0  +=       0.210 * cos(5.82498971470  + 9381.20349020070 *T)
    sumX0  +=       0.199 * cos(0.93528415346  + 11216.52810780750 *T)
    sumX0  +=       0.270 * cos(0.40019527935  + 284.10313634870 *T)
    sumX0  +=       0.207 * cos(3.06628694399  + 12562.80508881451 *T)
    sumX0  +=       0.212 * cos(1.75471606720  + 7129.40250222610 *T)
    sumX0  +=       0.269 * cos(1.52668960070  + 70755.31090921978 *T)
    sumX0  +=       0.221 * cos(3.40184768346  + 77713.52765063698 *T)
    sumX0  +=       0.225 * cos(4.01628844723  + 5635.82119919310 *T)
    sumX0  +=       0.197 * cos(3.36122016267  + 14712.07329897449 *T)
    sumX0  +=       0.192 * cos(2.04581063866  + 17272.12782500990 *T)
    sumX0  +=       0.241 * cos(5.09422826000  + 15907.00794416610 *T)
    sumX0  +=       0.261 * cos(5.60213035381  + 48739.61607959950 *T)
    sumX0  +=       0.255 * cos(0.53150571182  + 6206.56596123230 *T)
    sumX0  +=       0.256 * cos(2.30148639986  + 224.58861318540 *T)
    sumX0  +=       0.182 * cos(4.41917983264  + 18848.98373249069 *T)
    sumX0  +=       0.250 * cos(4.24183231344  + 5934.39524702691 *T)
    sumX0  +=       0.196 * cos(2.21911266897  + 16460.57734700850 *T)
    sumX0  +=       0.187 * cos(4.75761261391  + 22003.67081738630 *T)
    sumX0  +=       0.229 * cos(0.16732487654  + 2942.21960580809 *T)
    sumX0  +=       0.175 * cos(5.41173900219  + 11614.67711121570 *T)
    sumX0  +=       0.187 * cos(5.47425531847  + 9778.86485864190 *T)
    sumX0  +=       0.175 * cos(4.58223305763  + 3744.58352825430 *T)
    sumX0  +=       0.241 * cos(0.28794939753  + 8390.35417501730 *T)
    sumX0  +=       0.236 * cos(2.02485303450  + 1.72829019180 *T)
    sumX0  +=       0.217 * cos(5.02460023035  + 17996.27498570570 *T)
    sumX0  +=       0.173 * cos(1.08399000143  + 6275.71848550709 *T)
    sumX0  +=       0.191 * cos(6.27756282518  + 394.86970254271 *T)
    sumX0  +=       0.231 * cos(2.60472823534  + 34596.12083716890 *T)
    sumX0  +=       0.169 * cos(5.11378923747  + 6438.25243194210 *T)
    sumX0  +=       0.235 * cos(3.47396101480  + 17256.38771885789 *T)
    sumX0  +=       0.188 * cos(5.59662915440  + 401.91593924071 *T)
    sumX0  +=       0.217 * cos(4.97442863783  + 10984.43616918330 *T)
    sumX0  +=       0.230 * cos(2.40692408305  + 6632.24408792290 *T)
    sumX0  +=       0.208 * cos(0.72059454942  + 11087.52894340190 *T)
    sumX0  +=       0.165 * cos(2.37182548424  + 743.23387801611 *T)
    sumX0  +=       0.177 * cos(4.73067975820  + 4796.85191144269 *T)
    sumX0  +=       0.208 * cos(4.59043597135  + 3097.64000524229 *T)
    sumX0  +=       0.167 * cos(3.97387947664  + 5714.49779344750 *T)
    sumX0  +=       0.198 * cos(3.26732644351  + 5539.84197197530 *T)
    sumX0  +=       0.169 * cos(3.73983013574  + 12132.68377958950 *T)
    sumX0  +=       0.214 * cos(4.34216809137  + 24492.64993113510 *T)
    sumX0  +=       0.172 * cos(1.25880044898  + 6233.56264200310 *T)
    sumX0  +=       0.172 * cos(5.38973312367  + 6333.07669294670 *T)
    sumX0  +=       0.223 * cos(4.99990847060  + 266.85085920531 *T)
    sumX0  +=       0.159 * cos(3.90246086071  + 13199.17925677950 *T)
    sumX0  +=       0.219 * cos(5.40805183208  + 10344.05124790229 *T)
    sumX0  +=       0.207 * cos(3.49598218672  + 12569.98594611810 *T)
    sumX0  +=       0.176 * cos(4.05199209274  + 12012.82611462390 *T)
    sumX0  +=       0.170 * cos(6.01039539440  + 6294.36536773881 *T)
    sumX0  +=       0.170 * cos(0.63813817825  + 6272.27396721100 *T)
    sumX0  +=       0.205 * cos(5.87166506948  + 13119.47728534169 *T)
    sumX0  +=       0.167 * cos(5.54385526640  + 17686.99666304990 *T)
    sumX0  +=       0.210 * cos(0.88175440791  + 13521.50762410789 *T)
    sumX0  +=       0.163 * cos(1.46796231870  + 802.06474860730 *T)
    sumX0  +=       0.159 * cos(4.85018558111  + 5017.26455388150 *T)
    sumX0  +=       0.148 * cos(0.75075299347  + 419.72846135871 *T)
    sumX0  +=       0.165 * cos(0.87966880312  + 20199.33877711650 *T)
    sumX0  +=       0.196 * cos(3.57999903444  + 33326.33491569069 *T)
    sumX0  +=       0.146 * cos(0.39484750798  + 19800.70213874130 *T)
    sumX0  +=       0.148 * cos(2.67465416652  + 6852.14154150230 *T)
    sumX0  +=       0.146 * cos(0.23882077577  + 17370.60479339330 *T)
    sumX0  +=       0.143 * cos(0.85202349514  + 5618.56362234490 *T)
    sumX0  +=       0.142 * cos(5.37141800193  + 17654.53672226610 *T)
    sumX0  +=       0.140 * cos(1.22241558577  + 2008.80135664250 *T)
    sumX0  +=       0.153 * cos(1.79747194391  + 5436.74919775670 *T)
    sumX0  +=       0.162 * cos(3.70215026876  + 775.27879384050 *T)
    sumX0  +=       0.162 * cos(4.78754155685  + 12552.16842346470 *T)
    sumX0  +=       0.143 * cos(2.32218951292  + 5010.15100688070 *T)
    sumX0  +=       0.188 * cos(3.68702080121  + 28.69300495130 *T)
    sumX0  +=       0.155 * cos(6.15650399253  + 11610.10636149570 *T)
    sumX0  +=       0.174 * cos(5.71964679538  + 20452.62559473830 *T)
    sumX0  +=       0.179 * cos(5.30407615411  + 27511.22405605370 *T)
    sumX0  +=       0.135 * cos(0.07815962516  + 12431.33043743090 *T)
    sumX0  +=       0.164 * cos(5.60920106543  + 28767.16824196750 *T)
    sumX0  +=       0.134 * cos(5.44099179137  + 16840.91382829870 *T)
    sumX0  +=       0.174 * cos(4.27755511799  + 19805.07110906630 *T)
    sumX0  +=       0.146 * cos(1.55848605346  + 12701.46059750170 *T)
    sumX0  +=       0.137 * cos(2.60101256262  + 11.28951774740 *T)
    sumX0  +=       0.167 * cos(1.27059509188  + 17473.69756761190 *T)
    sumX0  +=       0.152 * cos(4.19104806587  + 16627.61473286070 *T)
    sumX0  +=       0.124 * cos(5.79651030598  + 6948.07571260490 *T)
    sumX0  +=       0.135 * cos(3.38729000615  + 3531.28443281630 *T)
    sumX0  +=       0.172 * cos(4.94261447200  + 167959.14629657479 *T)
    sumX0  +=       0.123 * cos(5.21069118841  + 23013.29572210369 *T)
    sumX0  +=       0.122 * cos(5.69263982892  + 3583.58484815730 *T)
    sumX0  +=       0.170 * cos(1.45223428169  + 333857.83394425618 *T)
    sumX0  +=       0.155 * cos(4.05745332726  + 6058.48723680599 *T)
    sumX0  +=       0.120 * cos(1.19140907580  + 12809.12412144031 *T)
    sumX0  +=       0.141 * cos(5.52535299558  + 162420.79565227419 *T)
    sumX0  +=       0.120 * cos(5.84076902832  + 12528.36784483050 *T)
    sumX0  +=       0.123 * cos(6.20892018031  + 25933.88051360590 *T)
    sumX0  +=       0.144 * cos(2.13726910032  + 95.73540973430 *T)
    sumX0  +=       0.165 * cos(0.74320909184  + 52669.82577581910 *T)
    sumX0  +=       0.119 * cos(1.19695838858  + 19247.62037086590 *T)
    sumX0  +=       0.117 * cos(0.73795876397  + 11610.79577585770 *T)
    sumX0  +=       0.163 * cos(3.11536370558  + 661.47674426450 *T)
    sumX0  +=       0.163 * cos(4.35849093916  + 9929.67004482930 *T)
    sumX0  +=       0.119 * cos(4.28708216617  + 12250.00364780970 *T)
    sumX0  +=       0.145 * cos(3.90603326867  + 6205.64589704690 *T)
    sumX0  +=       0.145 * cos(2.74250030398  + 6360.99343790291 *T)
    sumX0  +=       0.159 * cos(4.18513637906  + 228278.34846897016 *T)
    sumX0  +=       0.114 * cos(1.84089435963  + 19402.55313533309 *T)
    sumX0  +=       0.147 * cos(4.06094470975  + 38526.33053338850 *T)
    sumX0  +=       0.156 * cos(2.57788751226  + 4307.82712161890 *T)
    sumX0  +=       0.131 * cos(1.70667997590  + 21228.63584102931 *T)
    sumX0  +=       0.118 * cos(5.40927777429  + 6263.64990657511 *T)
    sumX0  +=       0.118 * cos(1.23925579836  + 6302.98942837470 *T)
    sumX0  +=       0.137 * cos(2.54415078163  + 6315.85221826630 *T)
    sumX0  +=       0.137 * cos(4.10438279102  + 6250.78711668350 *T)
    sumX0  +=       0.109 * cos(2.76724250809  + 11925.51791008410 *T)
    sumX0  +=       0.111 * cos(4.37410999057  + 6226.42129253930 *T)
    sumX0  +=       0.111 * cos(2.27442358208  + 6340.21804241050 *T)
    sumX0  +=       0.123 * cos(2.37617280556  + 24734.39821404090 *T)
    sumX0  +=       0.107 * cos(2.16622176546  + 12463.30274324771 *T)
    sumX0  +=       0.133 * cos(3.03066152119  + 18875.76968725751 *T)
    sumX0  +=       0.123 * cos(5.64258939044  + 6260.54446602410 *T)
    sumX0  +=       0.123 * cos(1.00594418221  + 6306.09486892571 *T)
    sumX0  +=       0.113 * cos(1.95402250286  + 2111.89413086110 *T)
    sumX0  +=       0.131 * cos(3.61193663544  + 18415.75962958090 *T)
    sumX0  +=       0.110 * cos(1.76186229885  + 6289.94822637491 *T)
    sumX0  +=       0.110 * cos(4.88667127380  + 6276.69110857489 *T)
    sumX0  +=       0.103 * cos(3.89043805400  + 6241.76887649010 *T)
    sumX0  +=       0.103 * cos(2.75809551865  + 6324.87045845970 *T)
    sumX0  +=       0.113 * cos(2.20489389956  + 3496.27664361750 *T)
    sumX0  +=       0.103 * cos(4.57936207476  + 10344.53888286930 *T)
    sumX0  +=       0.115 * cos(2.66876079900  + 24336.24921063270 *T)
    sumX0  +=       0.107 * cos(1.72027504380  + 83974.07917321339 *T)
    sumX0  +=       0.107 * cos(2.75882626470  + 84020.10309797739 *T)
    sumX0  +=       0.106 * cos(3.27680652019  + 12772.58106590351 *T)
    sumX0  +=       0.143 * cos(1.51224865614  + 2069.25065239010 *T)
    sumX0  +=       0.143 * cos(0.91984229557  + 18773.20529618210 *T)
    sumX0  +=       0.115 * cos(3.78326005147  + 3641.49075403570 *T)
    sumX0  +=       0.112 * cos(5.73842498855  + 11499.41240530929 *T)
    sumX0  +=       0.099 * cos(1.79019915134  + 11190.62171762050 *T)
    sumX0  +=       0.126 * cos(1.19294266238  + 18823.17304765790 *T)
    sumX0  +=       0.115 * cos(4.66544469064  + 12570.32767072940 *T)
    sumX0  +=       0.100 * cos(3.52903073097  + 16062.42834360030 *T)
    sumX0  +=       0.107 * cos(4.02566335272  + 5216.82419028490 *T)
    sumX0  +=       0.120 * cos(4.96595619835  + 9814.36028280769 *T)
    sumX0  +=       0.138 * cos(4.55918305844  + 6210.02254161590 *T)
    sumX0  +=       0.138 * cos(2.08935051421  + 6356.61679333390 *T)
    sumX0  +=       0.105 * cos(1.03351724810  + 12721.32828193349 *T)
    sumX0  +=       0.135 * cos(0.57056498904  + 18699.90817032310 *T)
    sumX0  +=       0.104 * cos(2.15061044327  + 12560.87259315890 *T)
    sumX0  +=       0.098 * cos(5.57514638218  + 5815.35467712050 *T)
    sumX0  +=       0.111 * cos(3.97355954616  + 10239.34004852730 *T)
    sumX0  +=       0.100 * cos(4.55824596261  + 263.32774085630 *T)
    sumX0  +=       0.095 * cos(1.74741902247  + 155.66421691770 *T)
    sumX0  +=       0.130 * cos(2.12273849218  + 27511.71169102070 *T)
    sumX0  +=       0.120 * cos(0.27869701115  + 31441.43375227330 *T)
    sumX0  +=       0.114 * cos(3.75372093481  + 6155.30082413750 *T)
    sumX0  +=       0.114 * cos(2.89481263784  + 6411.33851081230 *T)
    sumX0  +=       0.119 * cos(5.21922596573  + 951.96222373410 *T)
    sumX0  +=       0.117 * cos(1.90078630587  + 28236.98964190589 *T)
    sumX0  +=       0.107 * cos(1.90914871846  + 21.01921297590 *T)
    sumX0  +=       0.095 * cos(5.63307991850  + 11300.82803883990 *T)
    sumX0  +=       0.110 * cos(2.79120257993  + 6312.74917601091 *T)
    sumX0  +=       0.110 * cos(3.85733099272  + 6253.89015893890 *T)
    sumX0  +=       0.097 * cos(4.22068998833  + 78263.95324220609 *T)
    sumX0  +=       0.095 * cos(5.98375623165  + 23938.10020722450 *T)
    sumX0  +=       0.091 * cos(2.99457476326  + 12829.47944083910 *T)
    sumX0  +=       0.112 * cos(4.31165167302  + 16737.82105408010 *T)
    sumX0  +=       0.126 * cos(3.10862438843  + 3.27930086550 *T)
    sumX0  +=       0.100 * cos(3.03286435761  + 6133.75647034030 *T)
    sumX0  +=       0.127 * cos(5.41503454763  + 1293.24040609909 *T)
    sumX0  +=       0.119 * cos(5.99165259734  + 17893.18221148710 *T)
    sumX0  +=       0.089 * cos(1.33458093593  + 23539.95120381630 *T)
    sumX0  +=       0.126 * cos(0.25750692887  + 311565.26273852371 *T)
    sumX0  +=       0.093 * cos(5.67599033319  + 736.12033101530 *T)
    sumX0  +=       0.120 * cos(2.08797711521  + 14158.99153109910 *T)
    sumX0  +=       0.118 * cos(1.43212365415  + 16061.94070863329 *T)
    sumX0  +=       0.099 * cos(3.61958161583  + 6432.88286460950 *T)
    sumX0  +=       0.095 * cos(5.99715593195  + 2699.49100183409 *T)
    sumX0  +=       0.089 * cos(2.30094279014  + 15671.32557689010 *T)
    sumX0  +=       0.124 * cos(0.48008010598  + 178430.29100801519 *T)
    sumX0  +=       0.089 * cos(1.07338684504  + 6751.28465782931 *T)
    sumX0  +=       0.096 * cos(2.64494034094  + 7349.81514466491 *T)
    sumX0  +=       0.096 * cos(0.74221743876  + 24066.05174025910 *T)
    sumX0  +=       0.103 * cos(4.37747381017  + 18202.46053414290 *T)
    sumX0  +=       0.099 * cos(6.06683620849  + 6252.40554183991 *T)
    sumX0  +=       0.099 * cos(0.58169736416  + 6314.23379310990 *T)
    sumX0  +=       0.090 * cos(0.38965345976  + 58864.30010066279 *T)
    sumX0  +=       0.086 * cos(0.59831141560  + 9380.71585523369 *T)
    sumX0  +=       0.086 * cos(0.93909542206  + 10557.35034334029 *T)
    sumX0  +=       0.098 * cos(3.41472072058  + 6439.72038797730 *T)
    sumX0  +=       0.098 * cos(3.23381285207  + 6126.91894697251 *T)
    sumX0  +=       0.086 * cos(1.34148300572  + 23123.98967829010 *T)
    sumX0  +=       0.118 * cos(3.73844359380  + 13951.95709241740 *T)
    sumX0  +=       0.117 * cos(4.10902635534  + 89570.23393702849 *T)
    sumX0  +=       0.083 * cos(3.14133645510  + 8858.07112683710 *T)
    sumX0  +=       0.092 * cos(4.01226998426  + 12342.05072176440 *T)
    sumX0  +=       0.085 * cos(2.29415888429  + 5017.75218884850 *T)
    sumX0  +=       0.101 * cos(0.23020335948  + 18429.98672358250 *T)
    sumX0  +=       0.088 * cos(3.76744436736  + 17054.21292373670 *T)
    sumX0  +=       0.113 * cos(2.05049922518  + 12985.88016134151 *T)
    sumX0  +=       0.083 * cos(0.90577628609  + 20597.00014555770 *T)
    sumX0  +=       0.082 * cos(0.19375496576  + 5483.49854230950 *T)
    sumX0  +=       0.087 * cos(0.10696172464  + 21424.22282681990 *T)
    sumX0  +=       0.107 * cos(2.21757059467  + 522.33360061030 *T)
    sumX0  +=       0.084 * cos(2.41107951633  + 6187.34044025710 *T)
    sumX0  +=       0.084 * cos(4.23745405632  + 6379.29889469270 *T)
    sumX0  +=       0.107 * cos(0.88149755532  + 24382.83529095790 *T)
    sumX0  +=       0.082 * cos(0.94957627406  + 8983.05448679250 *T)
    sumX0  +=       0.105 * cos(3.87374571694  + 6131.42238638970 *T)
    sumX0  +=       0.105 * cos(2.77478785571  + 6435.21694856010 *T)
    sumX0  +=       0.114 * cos(4.07080782247  + 8258.81221333091 *T)
    sumX0  +=       0.082 * cos(3.82951941927  + 3957.88262369230 *T)
    sumX0  +=       0.112 * cos(2.70888594040  + 3738.51761262449 *T)
    sumX0  +=       0.102 * cos(2.89783593458  + 5767.85579638190 *T)
    sumX0  +=       0.102 * cos(3.75069763806  + 6798.78353856790 *T)
    sumX0  +=       0.092 * cos(1.35671084183  + 18216.68762814450 *T)
    sumX0  +=       0.083 * cos(1.35246903804  + 29864.09020982550 *T)
    sumX0  +=       0.112 * cos(5.48905138227  + 12189.02190955050 *T)
    sumX0  +=       0.080 * cos(3.59069348639  + 24080.27883426070 *T)
    sumX0  +=       0.080 * cos(4.80792276932  + 15141.14697682849 *T)
    sumX0  +=       0.088 * cos(3.00407783628  + 1573.57660661529 *T)
    sumX0  +=       0.081 * cos(3.72088724498  + 1550.80140516450 *T)
    sumX0  +=       0.092 * cos(4.36458397294  + 101426.45258845300 *T)
    sumX0  +=       0.110 * cos(0.37007007331  + 78423.94833416230 *T)
    sumX0  +=       0.087 * cos(2.32197271603  + 1580.62284331330 *T)
    sumX0  +=       0.085 * cos(6.14240130116  + 27043.25906569930 *T)
    sumX0  +=       0.078 * cos(1.45118645187  + 6812.52299760250 *T)
    sumX0  +=       0.080 * cos(3.36946826019  + 6081.06627230081 *T)
    sumX0  +=       0.080 * cos(3.27906531246  + 6485.57306264900 *T)
    sumX0  +=       0.079 * cos(4.53816488559  + 36109.62602214810 *T)
    sumX0  +=       0.088 * cos(0.87069012239  + 16944.00660251730 *T)
    sumX0  +=       0.081 * cos(1.41865192078  + 5.27910682390 *T)
    sumX0  +=       0.093 * cos(5.85619230131  + 16737.33341911309 *T)
    sumX0  +=       0.088 * cos(3.24901264480  + 12537.94632999850 *T)
    sumX0  +=       0.096 * cos(5.19394745813  + 20198.85114214949 *T)
    sumX0  +=       0.106 * cos(2.16665694222  + 56600.03547203870 *T)
    sumX0  +=       0.097 * cos(0.14089796919  + 6040.10342853389 *T)
    sumX0  +=       0.081 * cos(3.62140277149  + 3956.26419853590 *T)
    sumX0  +=       0.077 * cos(1.39348893935  + 40796.58274015770 *T)
    sumX0  +=       0.076 * cos(2.20964866296  + 22743.16556203289 *T)
    sumX0  +=       0.098 * cos(5.48885750823  + 42456.54022960810 *T)
    sumX0  +=       0.083 * cos(2.20640709608  + 19801.18977370830 *T)
    sumX0  +=       0.079 * cos(4.50777033878  + 5622.08674069391 *T)
    sumX0  +=       0.096 * cos(1.98497858569  + 5888.69378241570 *T)
    sumX0  +=       0.096 * cos(4.66355498696  + 6677.94555253410 *T)
    sumX0  +=       0.075 * cos(6.24025577336  + 41194.73174356590 *T)
    sumX0  +=       0.074 * cos(0.66729085356  + 6261.98402700790 *T)
    sumX0  +=       0.074 * cos(5.98124271908  + 6304.65530794190 *T)
    sumX0  +=       0.073 * cos(4.02724017625  + 5870.94857060050 *T)
    sumX0  +=       0.073 * cos(2.62129339640  + 6695.69076434930 *T)
    sumX0  +=       0.085 * cos(3.65241534932  + 12850.25483633150 *T)
    sumX0  +=       0.072 * cos(0.04178130909  + 6253.49822932610 *T)
    sumX0  +=       0.072 * cos(0.32356695638  + 6313.14110562371 *T)
    sumX0  +=       0.100 * cos(5.37794526764  + 5316.34879003930 *T)
    sumX0  +=       0.085 * cos(2.90983978106  + 12282.53619860110 *T)
    sumX0  +=       0.075 * cos(1.96590105616  + 24422.61416889080 *T)
    sumX0  +=       0.076 * cos(1.98227687206  + 63659.12156832110 *T)
    sumX0  +=       0.072 * cos(0.25938390676  + 16723.59396007850 *T)
    sumX0  +=       0.097 * cos(4.92418361725  + 17995.78735073869 *T)
    sumX0  +=       0.087 * cos(3.63417247018  + 18106.48130692510 *T)
    sumX0  +=       0.080 * cos(5.17587638062  + 17363.49124639250 *T)
    sumX0  +=       0.070 * cos(4.63132769081  + 6124.37614969171 *T)
    sumX0  +=       0.070 * cos(2.01720588184  + 6442.26318525810 *T)
    sumX0  +=       0.070 * cos(2.85493203288  + 4705.48849006010 *T)
    sumX0  +=       0.078 * cos(4.94849285823  + 23550.58786916610 *T)
    sumX0  +=       0.070 * cos(2.27501820444  + 12036.21691740469 *T)
    sumX0  +=       0.077 * cos(3.43499900992  + 5237.67719632029 *T)
    sumX0  +=       0.069 * cos(6.02127545231  + 16207.64245401849 *T)
    sumX0  +=       0.082 * cos(5.48703292615  + 6774.98295993371 *T)
    sumX0  +=       0.067 * cos(0.17045217443  + 7083.14079264031 *T)
    sumX0  +=       0.067 * cos(1.83540783927  + 6394.50630976251 *T)
    sumX0  +=       0.067 * cos(4.81312573338  + 6172.13302518730 *T)
    sumX0  +=       0.077 * cos(4.99636632687  + 9924.56660402710 *T)
    sumX0  +=       0.080 * cos(5.36758481138  + 22380.99961775750 *T)
    sumX0  +=       0.070 * cos(4.60174692010  + 6390.98319141350 *T)
    sumX0  +=       0.070 * cos(2.04678665255  + 6175.65614353630 *T)
    sumX0  +=       0.074 * cos(2.55353968467  + 16193.41536001689 *T)
    sumX0  +=       0.076 * cos(5.64799287134  + 32217.44399856430 *T)
    sumX0  +=       0.091 * cos(2.47526506655  + 6653.01948341530 *T)
    sumX0  +=       0.091 * cos(4.17326850610  + 5913.61985153450 *T)
    sumX0  +=       0.067 * cos(0.23293815799  + 6265.50714535691 *T)
    sumX0  +=       0.067 * cos(0.13241010748  + 6301.13218959290 *T)
    sumX0  +=       0.067 * cos(1.44106493302  + 5959.32661585050 *T)
    sumX0  +=       0.083 * cos(2.94569161083  + 16723.10632511149 *T)
    sumX0  +=       0.080 * cos(0.77941152413  + 23646.56709638390 *T)
    sumX0  +=       0.068 * cos(3.64312239532  + 4897.42439113870 *T)
    sumX0  +=       0.069 * cos(2.14076318935  + 6944.55259425590 *T)
    sumX0  +=       0.089 * cos(1.99980743200  + 10660.44311755889 *T)
    sumX0  +=       0.079 * cos(1.67507258686  + 35371.64344849290 *T)
    sumX0  +=       0.073 * cos(2.11438932219  + 6370.62787201471 *T)
    sumX0  +=       0.073 * cos(4.53414425046  + 6196.01146293510 *T)
    sumX0  +=       0.065 * cos(3.66592557921  + 22345.01655862470 *T)
    sumX0  +=       0.064 * cos(4.23436375372  + 15265.64270181689 *T)
    sumX0  +=       0.066 * cos(2.69579215047  + 6315.29197329170 *T)
    sumX0  +=       0.066 * cos(3.95274142218  + 6251.34736165811 *T)
    sumX0  +=       0.065 * cos(5.01328627787  + 323.74923414091 *T)
    sumX0  +=       0.076 * cos(6.17321822941  + 10873.74221299689 *T)
    sumX0  +=       0.084 * cos(5.21213167261  + 109.96250373590 *T)
    sumX0  +=       0.063 * cos(4.27995662426  + 11823.40545693370 *T)
    sumX0  +=       0.064 * cos(0.09093186873  + 28774.28178896830 *T)
    sumX0  +=       0.072 * cos(0.54185576116  + 18099.75944096650 *T)
    sumX0  +=       0.066 * cos(2.78134607936  + 17576.79034183050 *T)
    sumX0  +=       0.072 * cos(1.59783352501  + 245707.70992182777 *T)
    sumX0  +=       0.061 * cos(5.48721212671  + 10557.83797830730 *T)
    sumX0  +=       0.064 * cos(0.75473231286  + 71430.93943561260 *T)
    sumX0  +=       0.068 * cos(2.90814349886  + 28760.05469496669 *T)
    sumX0  +=       0.062 * cos(0.67966569221  + 3854.78984947370 *T)
    sumX0  +=       0.073 * cos(6.18332621846  + 23440.38154794670 *T)
    sumX0  +=       0.060 * cos(3.22950754418  + 13088.97293556010 *T)
    sumX0  +=       0.078 * cos(4.56094059497  + 12564.91104475801 *T)
    sumX0  +=       0.060 * cos(4.47214922427  + 7548.88714610130 *T)
    sumX0  +=       0.059 * cos(2.55184049976  + 6286.35515085690 *T)
    sumX0  +=       0.075 * cos(4.42124944201  + 18625.12657175580 *T)
    sumX0  +=       0.058 * cos(0.47683719349  + 35050.24409195890 *T)
    sumX0  +=       0.060 * cos(0.20507190989  + 553.81322032590 *T)
    sumX0  +=       0.069 * cos(5.82409350668  + 167993.93845370729 *T)
    sumX0  +=       0.059 * cos(0.42390666617  + 41991.02975038230 *T)
    sumX0  +=       0.062 * cos(2.87556762555  + 15663.79170522509 *T)
    sumX0  +=       0.081 * cos(1.27058846781  + 7250.29054491051 *T)
    sumX0  +=       0.059 * cos(2.68580345393  + 6277.25375184510 *T)
    sumX0  +=       0.059 * cos(3.96273011871  + 6289.38558310470 *T)
    sumX0  +=       0.063 * cos(2.13380171799  + 18003.38853270651 *T)
    sumX0  +=       0.057 * cos(4.87646676285  + 793.01870595090 *T)
    sumX0  +=       0.057 * cos(0.02714811761  + 10027.65937824569 *T)
    sumX0  +=       0.060 * cos(2.20528431296  + 647.25465079831 *T)
    sumX0  +=       0.072 * cos(4.16486482248  + 4597.77991000629 *T)
    sumX0  +=       0.059 * cos(3.82312915284  + 6279.30891415889 *T)
    sumX0  +=       0.068 * cos(2.71579349818  + 3166.66025521511 *T)
    sumX0  +=       0.056 * cos(4.67855976016  + 6226.51640530510 *T)
    sumX0  +=       0.056 * cos(1.96997381249  + 6340.12292964471 *T)
    sumX0  +=       0.056 * cos(2.52728721608  + 12303.31159409350 *T)
    sumX0  +=       0.057 * cos(3.09052199581  + 5864.39526857430 *T)
    sumX0  +=       0.057 * cos(3.55801157684  + 6702.24406637550 *T)
    sumX0  +=       0.069 * cos(1.90302432953  + 23536.36077516450 *T)
    sumX0  +=       0.059 * cos(0.79584524101  + 5760.25461441409 *T)
    sumX0  +=       0.065 * cos(0.39208551086  + 12139.30969162329 *T)
    sumX0  +=       0.059 * cos(3.38503180730  + 67589.33126454070 *T)
    sumX0  +=       0.073 * cos(0.51774809611  + 11079.92776143409 *T)
    sumX0  +=       0.076 * cos(2.34882980338  + 23227.08245250870 *T)
    sumX0  +=       0.067 * cos(2.11713130342  + 4804.45309341050 *T)
    sumX0  +=       0.073 * cos(6.00949211634  + 30349.12759025050 *T)
    sumX0  +=       0.074 * cos(4.74855693255  + 92747.93298430610 *T)
    sumX0  +=       0.070 * cos(1.56579326622  + 12299.78847574450 *T)
    sumX0  +=       0.060 * cos(0.37222072501  + 12171.76963240710 *T)
    sumX0  +=       0.059 * cos(2.60537764043  + 15.49628866851 *T)
    sumX0  +=       0.074 * cos(5.91207332800  + 3684.13423250670 *T)
    sumX0  +=       0.058 * cos(3.15618940951  + 6717.03140535170 *T)
    sumX0  +=       0.053 * cos(2.11594966811  + 26087.65932409069 *T)
    sumX0  +=       0.059 * cos(1.05069506955  + 12164.72339570910 *T)
    sumX0  +=       0.063 * cos(3.66132366701  + 5219.51794905560 *T)
    sumX0  +=       0.056 * cos(4.02457482788  + 3178.38960805111 *T)
    sumX0  +=       0.063 * cos(0.59069175384  + 5227.38225558179 *T)
    sumX0  +=       0.070 * cos(3.26289105386  + 19004.40413192490 *T)
    sumX0  +=       0.072 * cos(1.05516042130  + 4583.55281600469 *T)
    sumX0  +=       0.052 * cos(5.67763480740  + 3627.26366003410 *T)
    sumX0  +=       0.052 * cos(4.87358219235  + 6411.27120050950 *T)
    sumX0  +=       0.052 * cos(1.77495138030  + 6155.36813444030 *T)
    sumX0  +=       0.054 * cos(3.61747287023  + 5849.60792959810 *T)
    sumX0  +=       0.066 * cos(4.36702417592  + 5791.16874004909 *T)
    sumX0  +=       0.066 * cos(1.22543152233  + 5791.65637501610 *T)
    sumX0  +=       0.051 * cos(0.84172627776  + 5113.24378109930 *T)
    sumX0  +=       0.058 * cos(4.29307138428  + 30775.72578112650 *T)
    sumX0  +=       0.061 * cos(5.44071514919  + 14169.54974474690 *T)
    sumX0  +=       0.054 * cos(4.73181482869  + 10454.74520408871 *T)
    sumX0  +=       0.056 * cos(2.77114790322  + 23141.80220040810 *T)
    sumX0  +=       0.067 * cos(2.31997560907  + 28313.04498717750 *T)
    sumX0  +=       0.056 * cos(1.94933767144  + 5244.29305668450 *T)
    sumX0  +=       0.062 * cos(2.18007483018  + 5657.64947516250 *T)
    sumX0  +=       0.062 * cos(4.46845874247  + 6908.98985978730 *T)
    sumX0  +=       0.050 * cos(1.53914825533  + 2574.99527684569 *T)
    sumX0  +=       0.050 * cos(1.32272515661  + 536.56069461190 *T)
    sumX0  +=       0.050 * cos(1.95397557344  + 6684.99178923210 *T)
    sumX0  +=       0.050 * cos(4.69455799921  + 5881.64754571770 *T)
    sumX0  +=       0.049 * cos(5.07674655291  + 16310.73522823709 *T)
    sumX0  +=       0.049 * cos(5.07674655291  + 16311.22286320410 *T)
    sumX0  +=       0.065 * cos(0.63077087865  + 46386.74992582770 *T)
    sumX0  +=       0.069 * cos(3.59016732594  + 60530.24516825830 *T)
    sumX0  +=       0.068 * cos(6.19031950148  + 96678.14268052568 *T)
    sumX0  +=       0.049 * cos(2.44170258748  + 12323.17927852529 *T)
    sumX0  +=       0.050 * cos(3.95175048363  + 21954.40142688150 *T)
    sumX0  +=       0.049 * cos(0.26987394090  + 4164.55580709650 *T)
    sumX0  +=       0.059 * cos(2.11910349639  + 33794.78754101210 *T)
    sumX0  +=       0.050 * cos(3.21674377716  + 43739.04616344930 *T)
    sumX0  +=       0.059 * cos(4.28455887932  + 3981.73385156551 *T)
    sumX0  +=       0.058 * cos(1.59500717257  + 27707.29867681129 *T)
    sumX0  +=       0.051 * cos(3.00541030027  + 7669.21494381111 *T)
    sumX0  +=       0.048 * cos(4.80655958998  + 3646.59419483791 *T)
    sumX0  +=       0.065 * cos(3.20870046710  + 82534.64743809510 *T)
    sumX0  +=       0.056 * cos(4.19440679704  + 800.06494264891 *T)
    sumX0  +=       0.057 * cos(0.99121344005  + 11609.61872652869 *T)
    sumX0  +=       0.048 * cos(5.23363777755  + 26.54213728330 *T)
    sumX0  +=       0.047 * cos(1.10988199958  + 30640.10045611650 *T)
    sumX0  +=       0.060 * cos(1.97427905695  + 8982.56685182550 *T)
    sumX0  +=       0.049 * cos(3.72444308522  + 96563.24283557820 *T)
    sumX0  +=       0.050 * cos(5.17708029533  + 17583.90388883130 *T)
    sumX0  +=       0.046 * cos(2.75949485809  + 23539.46356884929 *T)
    sumX0  +=       0.048 * cos(2.83397317642  + 170.42905313570 *T)
    sumX0  +=       0.055 * cos(4.77764015710  + 735.63269604830 *T)
    sumX0  +=       0.054 * cos(4.33153690679  + 6680.98103591609 *T)
    sumX0  +=       0.060 * cos(1.72047044215  + 44033.88377205590 *T)
    sumX0  +=       0.064 * cos(4.40652103235  + 80181.29364935629 *T)
    sumX0  +=       0.059 * cos(5.18038765510  + 24279.35083569710 *T)
    sumX0  +=       0.049 * cos(5.16566622453  + 5490.54477900750 *T)
    sumX0  +=       0.058 * cos(1.24075349030  + 6490.02040477150 *T)
    sumX0  +=       0.058 * cos(5.40778008235  + 6076.61893017830 *T)
    sumX0  +=       0.046 * cos(6.06699951300  + 18208.10612510850 *T)
    sumX0  +=       0.045 * cos(1.62616739636  + 4480.46004178609 *T)
    sumX0  +=       0.055 * cos(3.56883579861  + 16097.92376776610 *T)
    sumX0  +=       0.055 * cos(0.42724314502  + 16097.43613279909 *T)
    sumX0  +=       0.061 * cos(5.10427594640  + 491.90710994230 *T)
    sumX0  +=       0.054 * cos(2.79730041476  + 6006.57389561050 *T)
    sumX0  +=       0.054 * cos(3.85123315789  + 6560.06543933930 *T)
    sumX0  +=       0.044 * cos(3.20288072834  + 5333.65642353809 *T)
    sumX0  +=       0.052 * cos(5.42298629738  + 244287.84382471119 *T)
    sumX0  +=       0.044 * cos(1.12093494405  + 2301.34199842589 *T)
    sumX0  +=       0.045 * cos(2.78871592108  + 33794.29990604510 *T)
    sumX0  +=       0.046 * cos(2.27794751316  + 4384.48081456829 *T)
    sumX0  +=       0.048 * cos(6.20709507263  + 35579.93505705350 *T)
    sumX0  +=       0.051 * cos(0.20647678216  + 18326.89394936390 *T)
    sumX0  +=       0.045 * cos(4.21002666289  + 533.45790092711 *T)
    sumX0  +=       0.046 * cos(4.72419290173  + 16703.30595098250 *T)
    sumX0  +=       0.047 * cos(0.82323755553  + 5209.22300831710 *T)
    sumX0  +=       0.051 * cos(2.88192111957  + 38650.41732368250 *T)
    sumX0  +=       0.046 * cos(4.04752097938  + 12555.34981720240 *T)
    sumX0  +=       0.052 * cos(6.21526623133  + 5992.34680160890 *T)
    sumX0  +=       0.052 * cos(0.43326734132  + 6574.29253334090 *T)
    sumX0  +=       0.044 * cos(5.37917268925  + 69942.19741831250 *T)
    sumX0  +=       0.049 * cos(5.19212576524  + 12146.91087359110 *T)
    sumX0  +=       0.053 * cos(3.10372910020  + 39301.85314471250 *T)
    sumX0  +=       0.045 * cos(6.22187415000  + 46360.93924099490 *T)
    sumX0  +=       0.043 * cos(3.11607958976  + 8402.08352785330 *T)
    sumX0  +=       0.043 * cos(1.76871622593  + 44137.19516685750 *T)
    sumX0  +=       0.049 * cos(1.97645700808  + 22029.96913718610 *T)
    sumX0  +=       0.044 * cos(3.04557845219  + 31441.92138724031 *T)
    sumX0  +=       0.042 * cos(1.28858480303  + 9225.29545579949 *T)
    sumX0  +=       0.045 * cos(4.79112566977  + 71519.54096076029 *T)
    sumX0  +=       0.051 * cos(4.52703512787  + 7762.18624153930 *T)
    sumX0  +=       0.041 * cos(0.43156198612  + 3735.48212924251 *T)
    sumX0  +=       0.057 * cos(2.81178684757  + 40398.43373674950 *T)
    sumX0  +=       0.047 * cos(4.46091937774  + 4707.98623122570 *T)
    sumX0  +=       0.056 * cos(0.65497165024  + 28664.07546774890 *T)
    sumX0  +=       0.050 * cos(4.15809563419  + 11919.38468415150 *T)
    sumX0  +=       0.047 * cos(0.89798744348  + 316428.47249139857 *T)
    sumX0  +=       0.041 * cos(4.27548568888  + 27278.71263392430 *T)
    sumX0  +=       0.056 * cos(2.85120735939  + 83659.52068988250 *T)
    sumX0  +=       0.048 * cos(2.50261741205  + 12662.37474468410 *T)
    sumX0  +=       0.054 * cos(0.83831240155  + 16627.12709789369 *T)
    sumX0  +=       0.052 * cos(4.41165168727  + 12571.91844177370 *T)
    sumX0  +=       0.043 * cos(4.69793524808  + 7322.34627826531 *T)
    sumX0  +=       0.041 * cos(3.56154893542  + 664.99986261351 *T)
    sumX0  +=       0.042 * cos(1.54738888492  + 5641.95442512570 *T)
    sumX0  +=       0.044 * cos(1.44765859641  + 6425.49829451110 *T)
    sumX0  +=       0.044 * cos(5.20087497624  + 6141.14104043870 *T)
    sumX0  +=       0.039 * cos(0.38205897719  + 22594.29871319550 *T)
    sumX0  +=       0.040 * cos(5.88641926833  + 20894.53186172529 *T)
    sumX0  +=       0.044 * cos(5.28307931868  + 5540.32960694230 *T)
    sumX0  +=       0.050 * cos(5.69294803866  + 17782.48825530069 *T)
    sumX0  +=       0.054 * cos(4.69164296803  + 6006.28467373350 *T)
    sumX0  +=       0.054 * cos(1.95689060462  + 6560.35466121630 *T)
    sumX0  +=       0.045 * cos(5.54499872381  + 142861.14741877469 *T)
    sumX0  +=       0.047 * cos(1.12158335918  + 24499.76347813590 *T)
    sumX0  +=       0.043 * cos(1.36615498859  + 7026.30972800750 *T)
    sumX0  +=       0.047 * cos(1.95434456607  + 955.84355909210 *T)
    sumX0  +=       0.047 * cos(4.50829221159  + 24485.53638413430 *T)
    sumX0  +=       0.040 * cos(2.60727518415  + 9388.24972689870 *T)
    sumX0  +=       0.043 * cos(5.81439105801  + 6359.58573875050 *T)
    sumX0  +=       0.042 * cos(3.96208965107  + 8635.69818627969 *T)
    sumX0  +=       0.051 * cos(2.46325624829  + 27831.79440179969 *T)
    sumX0  +=       0.042 * cos(1.94770311943  + 6334.60000533731 *T)
    sumX0  +=       0.042 * cos(4.70083045322  + 6232.03932961249 *T)
    sumX0  +=       0.052 * cos(3.53311914292  + 11905.16259068530 *T)
    sumX0  +=       0.040 * cos(1.50948151820  + 7076.09455594230 *T)
    sumX0  +=       0.040 * cos(0.29378456988  + 5444.35037972450 *T)
    sumX0  +=       0.040 * cos(0.07156369559  + 7122.28895522530 *T)
    sumX0  +=       0.038 * cos(2.24525725009  + 44933.49317367390 *T)
    sumX0  +=       0.037 * cos(0.27418012894  + 5715.60102974450 *T)
    sumX0  +=       0.037 * cos(0.09116813653  + 6851.03830520530 *T)
    sumX0  +=       0.042 * cos(0.79913949254  + 36147.65369478390 *T)
    sumX0  +=       0.042 * cos(2.85822305358  + 6238.17844783830 *T)
    sumX0  +=       0.042 * cos(3.79031051907  + 6328.46088711150 *T)
    sumX0  +=       0.040 * cos(3.64347872660  + 3214.89256315970 *T)
    sumX0  +=       0.043 * cos(0.65521409425  + 23141.31456544110 *T)
    sumX0  +=       0.041 * cos(3.95515629208  + 19804.58347409930 *T)
    sumX0  +=       0.038 * cos(2.90982798521  + 20.11150191529 *T)
    sumX0  +=       0.050 * cos(1.62788860301  + 84334.66158130829 *T)
    sumX0  +=       0.050 * cos(2.84763289309  + 60284.16619777939 *T)

    # EARTH X Coord
    sumX1  = 0
    sumX1  +=  154550.744 * cos(0.64605836878  + 0.24381748350 *T)
    sumX1  +=   51503.383 * cos(6.00263199393  + 12566.39551746630 *T)
    sumX1  +=    1290.763 * cos(5.95941652859  + 18849.47136745770 *T)
    sumX1  +=     702.576 * cos(1.75347943445  + 6283.31966747490 *T)
    sumX1  +=     430.422 * cos(2.45216492471  + 6282.83203250789 *T)
    sumX1  +=     212.689 * cos(1.73380190045  + 6279.79654912590 *T)
    sumX1  +=     212.524 * cos(4.91484799365  + 6286.84278582391 *T)
    sumX1  +=      62.308 * cos(0.36193297710  + 4705.97612502710 *T)
    sumX1  +=      59.808 * cos(3.81187956678  + 6257.02134767510 *T)
    sumX1  +=      59.474 * cos(2.83659652893  + 6309.61798727471 *T)
    sumX1  +=      48.914 * cos(5.21332670904  + 775.76642880750 *T)
    sumX1  +=      42.814 * cos(0.43864033644  + 1059.62574767270 *T)
    sumX1  +=      46.457 * cos(0.01089905463  + 7860.66320992270 *T)
    sumX1  +=      36.653 * cos(2.19635713914  + 5753.62870238030 *T)
    sumX1  +=      35.649 * cos(1.44752011203  + 5885.17066406670 *T)
    sumX1  +=      35.362 * cos(4.47215447256  + 6813.01063256950 *T)
    sumX1  +=      32.151 * cos(5.19677261546  + 6681.46867088311 *T)
    sumX1  +=      28.763 * cos(5.91618988907  + 25132.54721744910 *T)
    sumX1  +=      28.447 * cos(1.14976229491  + 6127.89926804070 *T)
    sumX1  +=      27.537 * cos(5.49920443604  + 6438.74006690910 *T)
    sumX1  +=      24.819 * cos(2.92230738957  + 5487.02166065850 *T)
    sumX1  +=      20.615 * cos(3.71821460067  + 7079.61767429131 *T)
    sumX1  +=      19.621 * cos(2.88650825392  + 5507.79705615090 *T)
    sumX1  +=      18.378 * cos(1.46521342874  + 11790.87290614230 *T)
    sumX1  +=      16.494 * cos(6.21543454587  + 11507.01358727710 *T)
    sumX1  +=      16.756 * cos(3.81938434057  + 7058.84227879890 *T)
    sumX1  +=      14.558 * cos(5.97530104893  + 6290.43321447570 *T)
    sumX1  +=      14.404 * cos(0.68122522191  + 6276.20612047410 *T)
    sumX1  +=      14.052 * cos(1.45747121073  + 796.54182429990 *T)
    sumX1  +=      12.261 * cos(4.15789187748  + 4693.75913722409 *T)
    sumX1  +=      12.758 * cos(0.34831751321  + 7.35736448430 *T)
    sumX1  +=      10.086 * cos(3.29242550455  + 3739.00524759150 *T)
    sumX1  +=       9.469 * cos(4.63920457850  + 6070.02057203690 *T)
    sumX1  +=      10.425 * cos(2.38242547117  + 6284.29998854310 *T)
    sumX1  +=      10.425 * cos(4.26610810148  + 6282.33934640670 *T)
    sumX1  +=       9.570 * cos(1.89427100067  + 4137.15425099970 *T)
    sumX1  +=       9.044 * cos(2.00972967096  + 6496.61876291290 *T)
    sumX1  +=       8.544 * cos(0.05731253621  + 1194.69082770810 *T)
    sumX1  +=       8.214 * cos(1.25401494002  + 1589.31671276730 *T)
    sumX1  +=       6.157 * cos(3.35676469316  + 8827.63408735830 *T)
    sumX1  +=       6.271 * cos(4.75475820102  + 8429.48508395010 *T)
    sumX1  +=       5.524 * cos(5.36941014190  + 4933.45225781610 *T)
    sumX1  +=       7.314 * cos(0.49102526454  + 4535.30325440790 *T)
    sumX1  +=       5.157 * cos(5.14965163127  + 11770.09751064990 *T)
    sumX1  +=       6.749 * cos(4.32513020462  + 5088.87265725030 *T)
    sumX1  +=       4.853 * cos(0.62462823219  + 6040.59106350090 *T)
    sumX1  +=       5.304 * cos(5.64244320178  + 3154.44326741210 *T)
    sumX1  +=       4.985 * cos(4.62777616911  + 12569.91863581531 *T)
    sumX1  +=       4.709 * cos(3.96386608147  + 5331.60126122430 *T)
    sumX1  +=       4.601 * cos(6.02376124126  + 6526.04827144891 *T)
    sumX1  +=       4.201 * cos(1.28125193997  + 7633.18707713370 *T)
    sumX1  +=       5.607 * cos(2.55800509408  + 5729.75026463250 *T)
    sumX1  +=       4.257 * cos(4.12450123624  + 3930.45351370310 *T)
    sumX1  +=       4.394 * cos(0.04709285339  + 12559.28197046550 *T)
    sumX1  +=       3.870 * cos(2.68681123019  + 7235.03807372550 *T)
    sumX1  +=       5.154 * cos(6.15930214427  + 8031.33608054190 *T)
    sumX1  +=       5.036 * cos(4.09009014859  + 6836.88907031730 *T)
    sumX1  +=       5.076 * cos(2.30606090748  + 7477.76667769950 *T)
    sumX1  +=       3.593 * cos(2.40570544354  + 12565.90788249930 *T)
    sumX1  +=       3.384 * cos(5.40354177904  + 10977.32262218251 *T)
    sumX1  +=       3.371 * cos(0.28961103245  + 11371.94850724170 *T)
    sumX1  +=       3.536 * cos(3.73353119602  + 4164.06817212950 *T)
    sumX1  +=       4.292 * cos(4.93786086114  + 1592.83983111630 *T)
    sumX1  +=       3.407 * cos(2.40599273330  + 3128.63258257930 *T)
    sumX1  +=       3.152 * cos(1.73356525688  + 5223.93773728571 *T)
    sumX1  +=       2.883 * cos(4.92077247219  + 1747.77259558350 *T)
    sumX1  +=       2.751 * cos(5.24449444697  + 7342.70159766410 *T)
    sumX1  +=       3.323 * cos(1.91880015934  + 801.57711364030 *T)
    sumX1  +=       2.999 * cos(2.25355241223  + 8636.18582124671 *T)
    sumX1  +=       2.496 * cos(3.52078273708  + 2145.92159899169 *T)
    sumX1  +=       3.358 * cos(4.25271343539  + 155.17658195069 *T)
    sumX1  +=       2.297 * cos(4.53559364224  + 17260.39847217390 *T)
    sumX1  +=       2.047 * cos(3.53660263792  + 1990.98883452450 *T)
    sumX1  +=       2.778 * cos(1.46152428305  + 5481.49873635110 *T)
    sumX1  +=       2.025 * cos(1.44003078455  + 951.47458876710 *T)
    sumX1  +=       2.663 * cos(5.71398175957  + 26.05450231630 *T)
    sumX1  +=       2.009 * cos(5.73390835065  + 4690.72365384210 *T)
    sumX1  +=       2.148 * cos(1.89688144273  + 537.04832957890 *T)
    sumX1  +=       1.853 * cos(2.84762029989  + 553.32558535889 *T)
    sumX1  +=       2.478 * cos(0.03933370984  + 1349.62359217529 *T)
    sumX1  +=       2.023 * cos(2.39095052485  + 522.82123557730 *T)
    sumX1  +=       1.735 * cos(2.91944410412  + 529.44714761109 *T)
    sumX1  +=       2.417 * cos(5.16772880951  + 7085.14059859870 *T)
    sumX1  +=       1.809 * cos(0.78457177103  + 397.90518592470 *T)
    sumX1  +=       1.893 * cos(4.22746619412  + 9438.00675237050 *T)
    sumX1  +=       1.708 * cos(0.29122424525  + 10989.05197501850 *T)
    sumX1  +=       1.898 * cos(2.34007370780  + 5216.33655531789 *T)
    sumX1  +=       1.820 * cos(1.87728974480  + 5230.56364931950 *T)
    sumX1  +=       1.541 * cos(1.10077450497  + 426.84200835950 *T)
    sumX1  +=       1.555 * cos(3.67004127739  + 13096.08648256090 *T)
    sumX1  +=       1.779 * cos(1.80817581905  + 12562.87239911730 *T)
    sumX1  +=       1.448 * cos(0.18764842398  + 5753.14106741329 *T)
    sumX1  +=       1.450 * cos(0.99962470222  + 6262.96434807611 *T)
    sumX1  +=       1.450 * cos(5.64890887043  + 6303.67498687370 *T)
    sumX1  +=       1.822 * cos(1.70526332053  + 10973.79950383350 *T)
    sumX1  +=       1.372 * cos(0.89281758077  + 7875.91568110771 *T)
    sumX1  +=       1.375 * cos(0.76961577554  + 7084.65296363170 *T)
    sumX1  +=       1.400 * cos(4.85224030747  + 12721.81591690050 *T)
    sumX1  +=       1.247 * cos(1.86305212941  + 2119.00767786191 *T)
    sumX1  +=       1.231 * cos(4.10244270612  + 18319.78040236310 *T)
    sumX1  +=       1.141 * cos(3.49241515175  + 1066.73929467350 *T)
    sumX1  +=       1.138 * cos(2.80935582144  + 2942.70724077510 *T)
    sumX1  +=       1.047 * cos(6.02902034733  + 5642.93474619389 *T)
    sumX1  +=       1.058 * cos(3.76081315695  + 242.97242145751 *T)
    sumX1  +=       1.025 * cos(5.32138364615  + 10447.14402212089 *T)
    sumX1  +=       0.973 * cos(0.24159402695  + 640.14110379750 *T)
    sumX1  +=       1.114 * cos(5.94057908115  + 15721.08260236190 *T)
    sumX1  +=       0.963 * cos(2.13180404708  + 2389.13783793270 *T)
    sumX1  +=       0.941 * cos(5.78706319005  + 20426.81490990550 *T)
    sumX1  +=       1.068 * cos(0.37393702270  + 529.93478257810 *T)
    sumX1  +=       0.929 * cos(3.11135132378  + 10575.65050042530 *T)
    sumX1  +=       0.899 * cos(1.07875107899  + 16496.60521368590 *T)
    sumX1  +=       0.889 * cos(0.29953586851  + 6277.79674316750 *T)
    sumX1  +=       0.889 * cos(0.06581239696  + 6288.84259178230 *T)
    sumX1  +=       1.018 * cos(4.23379398746  + 12540.09719766650 *T)
    sumX1  +=       1.060 * cos(1.05626464973  + 5760.74224938110 *T)
    sumX1  +=       1.017 * cos(5.59172649862  + 6805.89708556871 *T)
    sumX1  +=       0.797 * cos(5.36416276600  + 14314.41193053330 *T)
    sumX1  +=       0.795 * cos(1.67821896162  + 6286.60602489270 *T)
    sumX1  +=       0.795 * cos(4.97031461103  + 6280.03331005710 *T)
    sumX1  +=       0.800 * cos(4.75445763497  + 12029.59100537090 *T)
    sumX1  +=       0.719 * cos(3.86746073982  + 9623.93209417470 *T)
    sumX1  +=       0.698 * cos(0.23124353834  + 6148.25458743950 *T)
    sumX1  +=       0.686 * cos(5.36448356218  + 5856.72147659890 *T)
    sumX1  +=       0.673 * cos(5.56184807840  + 12964.54452087450 *T)
    sumX1  +=       0.902 * cos(2.26778049127  + 13368.21644859010 *T)
    sumX1  +=       0.678 * cos(0.13408295783  + 6418.38474751031 *T)
    sumX1  +=       0.671 * cos(1.28940504615  + 6709.91785835091 *T)
    sumX1  +=       0.744 * cos(4.25820705300  + 12043.81809937250 *T)
    sumX1  +=       0.706 * cos(4.09030374624  + 16730.70750707931 *T)
    sumX1  +=       0.661 * cos(3.99296008106  + 14712.56093394150 *T)
    sumX1  +=       0.633 * cos(0.86914955896  + 4292.57465043390 *T)
    sumX1  +=       0.621 * cos(3.31200729066  + 13119.96492030870 *T)
    sumX1  +=       0.638 * cos(0.47109166303  + 4690.23601887509 *T)
    sumX1  +=       0.625 * cos(1.88693051693  + 13518.11392371690 *T)
    sumX1  +=       0.637 * cos(6.16733887432  + 5746.51515537950 *T)
    sumX1  +=       0.628 * cos(1.85369027207  + 4686.64559022330 *T)
    sumX1  +=       0.653 * cos(4.28657823528  + 3929.96587873609 *T)
    sumX1  +=       0.578 * cos(5.33616352365  + 5088.38502228330 *T)
    sumX1  +=       0.563 * cos(4.82044552939  + 10447.63165708790 *T)
    sumX1  +=       0.610 * cos(0.48716946879  + 6820.12417957030 *T)
    sumX1  +=       0.666 * cos(3.30009380787  + 3634.37720703489 *T)
    sumX1  +=       0.720 * cos(0.47572131456  + 13916.26292712510 *T)
    sumX1  +=       0.620 * cos(5.87390167695  + 31415.62306744050 *T)
    sumX1  +=       0.553 * cos(4.28779208332  + 6259.44122972711 *T)
    sumX1  +=       0.553 * cos(2.36074148933  + 6307.19810522270 *T)
    sumX1  +=       0.523 * cos(0.75768878632  + 14143.73905991410 *T)
    sumX1  +=       0.524 * cos(5.54806074141  + 12139.79732659030 *T)
    sumX1  +=       0.569 * cos(0.12658241682  + 12036.70455237170 *T)
    sumX1  +=       0.514 * cos(2.30528734098  + 4700.87268422489 *T)
    sumX1  +=       0.563 * cos(2.21509710432  + 11014.86265985130 *T)
    sumX1  +=       0.495 * cos(3.58067034949  + 13362.69352428270 *T)
    sumX1  +=       0.481 * cos(2.50655660751  + 6279.43833211690 *T)
    sumX1  +=       0.481 * cos(4.14197696514  + 6287.20100283290 *T)
    sumX1  +=       0.466 * cos(4.51385713822  + 10177.50149701710 *T)
    sumX1  +=       0.440 * cos(0.45716811540  + 1577.58735993130 *T)
    sumX1  +=       0.437 * cos(3.18132586326  + 11499.90004027630 *T)
    sumX1  +=       0.452 * cos(5.55409870988  + 12573.50906446710 *T)
    sumX1  +=       0.499 * cos(1.08746970886  + 12410.97511803210 *T)
    sumX1  +=       0.440 * cos(2.15051971988  + 24073.16528725990 *T)
    sumX1  +=       0.587 * cos(4.71917094180  + 6.86972951729 *T)
    sumX1  +=       0.388 * cos(2.04118602600  + 7860.17557495569 *T)
    sumX1  +=       0.398 * cos(5.77147994683  + 8274.06468451590 *T)
    sumX1  +=       0.423 * cos(1.16599451585  + 4694.24677219110 *T)
    sumX1  +=       0.530 * cos(3.00672799320  + 12592.69383726610 *T)
    sumX1  +=       0.379 * cos(2.02388854283  + 6180.22689325630 *T)
    sumX1  +=       0.379 * cos(4.62464502982  + 6386.41244169350 *T)
    sumX1  +=       0.377 * cos(5.90352173005  + 7872.39256275871 *T)
    sumX1  +=       0.419 * cos(5.62219874401  + 5327.71992586630 *T)
    sumX1  +=       0.383 * cos(6.24432820927  + 6247.29180079750 *T)
    sumX1  +=       0.383 * cos(0.40420536338  + 6319.34753415231 *T)
    sumX1  +=       0.375 * cos(5.62315844522  + 2352.62233628830 *T)
    sumX1  +=       0.457 * cos(4.19200440747  + 6077.13411903770 *T)
    sumX1  +=       0.457 * cos(2.45652916518  + 6489.50521591210 *T)
    sumX1  +=       0.481 * cos(1.89483272077  + 4292.08701546690 *T)
    sumX1  +=       0.344 * cos(0.28668049892  + 18451.32236404950 *T)
    sumX1  +=       0.453 * cos(0.72577236980  + 2787.28684134090 *T)
    sumX1  +=       0.364 * cos(4.39782645564  + 6245.29199483910 *T)
    sumX1  +=       0.364 * cos(2.25070711701  + 6321.34734011070 *T)
    sumX1  +=       0.327 * cos(2.71208539186  + 12323.66691349230 *T)
    sumX1  +=       0.358 * cos(4.87210297876  + 9917.45305702629 *T)
    sumX1  +=       0.342 * cos(0.54833716779  + 6262.54427198250 *T)
    sumX1  +=       0.342 * cos(6.10019640486  + 6304.09506296730 *T)
    sumX1  +=       0.318 * cos(0.13364568371  + 11926.49823115230 *T)
    sumX1  +=       0.320 * cos(2.70295484544  + 77714.01528560399 *T)
    sumX1  +=       0.354 * cos(1.02467818407  + 7238.91940908350 *T)
    sumX1  +=       0.318 * cos(5.32503015719  + 6254.87048000710 *T)
    sumX1  +=       0.318 * cos(1.32350341546  + 6311.76885494270 *T)
    sumX1  +=       0.304 * cos(2.00970548291  + 12779.69461290430 *T)
    sumX1  +=       0.343 * cos(3.71302827314  + 6298.57213865991 *T)
    sumX1  +=       0.343 * cos(2.93550529951  + 6268.06719628990 *T)
    sumX1  +=       0.359 * cos(4.09955266779  + 1052.51220067191 *T)
    sumX1  +=       0.398 * cos(0.12298117861  + 1551.28904013150 *T)
    sumX1  +=       0.306 * cos(2.58873546625  + 5863.83502359970 *T)
    sumX1  +=       0.300 * cos(1.77608786075  + 90280.16698558679 *T)
    sumX1  +=       0.307 * cos(3.29263568042  + 3893.93801205869 *T)
    sumX1  +=       0.297 * cos(0.51221148619  + 5429.63565075589 *T)
    sumX1  +=       0.365 * cos(2.74542548477  + 10022.08109758290 *T)
    sumX1  +=       0.336 * cos(2.72364909606  + 17782.97589026770 *T)
    sumX1  +=       0.295 * cos(4.05850212848  + 6702.80431135010 *T)
    sumX1  +=       0.288 * cos(2.18795828306  + 18073.94875613370 *T)
    sumX1  +=       0.287 * cos(3.33776996109  + 1577.09972496430 *T)
    sumX1  +=       0.378 * cos(0.39946174567  + 2353.10997125530 *T)
    sumX1  +=       0.289 * cos(3.04361347297  + 213.54291292150 *T)
    sumX1  +=       0.347 * cos(0.72838501040  + 5223.45010231870 *T)
    sumX1  +=       0.321 * cos(2.23787969105  + 17797.20298426930 *T)
    sumX1  +=       0.289 * cos(5.30650717041  + 220.65645992230 *T)
    sumX1  +=       0.260 * cos(1.04284839158  + 955.35592412510 *T)
    sumX1  +=       0.268 * cos(3.62847841862  + 14945.07235607090 *T)
    sumX1  +=       0.242 * cos(4.34386674771  + 2544.55823736690 *T)
    sumX1  +=       0.245 * cos(2.31719119888  + 7632.69944216670 *T)
    sumX1  +=       0.234 * cos(4.07393437960  + 1596.43025976811 *T)
    sumX1  +=       0.267 * cos(5.83055956746  + 206.42936592071 *T)
    sumX1  +=       0.282 * cos(4.50260656021  + 13341.91812879030 *T)
    sumX1  +=       0.279 * cos(3.13773429847  + 4731.78680985990 *T)
    sumX1  +=       0.290 * cos(2.03216763842  + 5642.44206009270 *T)
    sumX1  +=       0.298 * cos(0.93822937797  + 17790.08943726851 *T)
    sumX1  +=       0.270 * cos(2.16529088691  + 12168.24651405810 *T)
    sumX1  +=       0.294 * cos(2.52895912407  + 2146.40923395870 *T)
    sumX1  +=       0.214 * cos(0.93543623291  + 8030.84844557490 *T)
    sumX1  +=       0.207 * cos(1.60966002008  + 213.05527795450 *T)
    sumX1  +=       0.216 * cos(5.60316710178  + 3185.43584474911 *T)
    sumX1  +=       0.195 * cos(2.87143133187  + 5884.68302909969 *T)
    sumX1  +=       0.199 * cos(3.93236191602  + 1748.26023055050 *T)
    sumX1  +=       0.250 * cos(4.61596023287  + 6924.19727485710 *T)
    sumX1  +=       0.194 * cos(2.36749003244  + 641.12142486571 *T)
    sumX1  +=       0.203 * cos(2.25356528795  + 6503.73230991370 *T)
    sumX1  +=       0.203 * cos(4.39496828470  + 6062.90702503610 *T)
    sumX1  +=       0.258 * cos(5.63025007570  + 796.05418933290 *T)
    sumX1  +=       0.189 * cos(1.63377883836  + 11506.52595231009 *T)
    sumX1  +=       0.186 * cos(0.12949416246  + 18209.57408114370 *T)
    sumX1  +=       0.184 * cos(4.20955149479  + 12566.46282776910 *T)
    sumX1  +=       0.180 * cos(5.76808682495  + 853.44019923550 *T)
    sumX1  +=       0.185 * cos(4.68506649147  + 3495.78900865049 *T)
    sumX1  +=       0.205 * cos(4.93663483706  + 5849.12029463110 *T)
    sumX1  +=       0.187 * cos(5.91135912176  + 10213.52936369450 *T)
    sumX1  +=       0.187 * cos(2.05685668523  + 6290.36590417291 *T)
    sumX1  +=       0.187 * cos(4.59167688742  + 6276.27343077690 *T)
    sumX1  +=       0.235 * cos(5.91611965384  + 9779.35249360890 *T)
    sumX1  +=       0.173 * cos(1.38702654435  + 19651.29229858150 *T)
    sumX1  +=       0.184 * cos(4.65365446464  + 12566.32820716351 *T)
    sumX1  +=       0.230 * cos(2.23774195109  + 12567.37583853451 *T)
    sumX1  +=       0.166 * cos(1.32597581910  + 110.45013870291 *T)
    sumX1  +=       0.189 * cos(2.84387211432  + 3.76693583251 *T)
    sumX1  +=       0.166 * cos(1.19482070134  + 433.95555536030 *T)
    sumX1  +=       0.186 * cos(4.46969004938  + 5863.34738863270 *T)
    sumX1  +=       0.184 * cos(3.01631602481  + 10983.94853421629 *T)
    sumX1  +=       0.184 * cos(3.38049875173  + 6037.48802124550 *T)
    sumX1  +=       0.156 * cos(2.34433370292  + 2648.69864295650 *T)
    sumX1  +=       0.182 * cos(3.47271515850  + 10969.72144021469 *T)
    sumX1  +=       0.175 * cos(3.26771548230  + 6529.15131370430 *T)
    sumX1  +=       0.154 * cos(5.67725689556  + 6453.99253809410 *T)
    sumX1  +=       0.154 * cos(0.97127667708  + 6112.64679685570 *T)
    sumX1  +=       0.164 * cos(4.52273465985  + 12353.09642202830 *T)
    sumX1  +=       0.146 * cos(4.05729660897  + 149.80701461810 *T)
    sumX1  +=       0.184 * cos(0.20513580951  + 11015.35029481830 *T)
    sumX1  +=       0.191 * cos(2.29402324776  + 3894.42564702570 *T)
    sumX1  +=       0.144 * cos(3.76178854682  + 18636.17227201970 *T)
    sumX1  +=       0.146 * cos(1.99164382254  + 13760.84252769090 *T)
    sumX1  +=       0.142 * cos(1.34565609994  + 4156.95462512869 *T)
    sumX1  +=       0.185 * cos(3.72101619927  + 7234.55043875850 *T)
    sumX1  +=       0.133 * cos(0.01824676794  + 18139.53831889940 *T)
    sumX1  +=       0.131 * cos(4.75893185547  + 5507.30942118390 *T)
    sumX1  +=       0.146 * cos(5.03857686201  + 18875.28205229050 *T)
    sumX1  +=       0.128 * cos(3.41348197806  + 6069.53293706989 *T)
    sumX1  +=       0.128 * cos(0.18086618365  + 16200.52890701769 *T)
    sumX1  +=       0.123 * cos(5.45134858107  + 20.53157800890 *T)
    sumX1  +=       0.121 * cos(0.98195264587  + 17256.87535382490 *T)
    sumX1  +=       0.123 * cos(0.10275700246  + 6393.52598869430 *T)
    sumX1  +=       0.123 * cos(0.26259126301  + 6173.11334625550 *T)
    sumX1  +=       0.117 * cos(4.47711713050  + 18852.99448580670 *T)
    sumX1  +=       0.120 * cos(1.03913122140  + 9381.20349020070 *T)
    sumX1  +=       0.113 * cos(2.33706395068  + 65147.37595065419 *T)
    sumX1  +=       0.137 * cos(3.68199834086  + 16201.01654198470 *T)
    sumX1  +=       0.113 * cos(5.07439785395  + 6836.40143535029 *T)
    sumX1  +=       0.123 * cos(4.42839393502  + 12565.41519639810 *T)
    sumX1  +=       0.132 * cos(6.08414353296  + 17655.02435723310 *T)
    sumX1  +=       0.113 * cos(4.12294240934  + 10575.16286545830 *T)
    sumX1  +=       0.113 * cos(2.09696400140  + 10213.04172872750 *T)
    sumX1  +=       0.116 * cos(2.23828708218  + 17253.28492517310 *T)
    sumX1  +=       0.120 * cos(2.28034205827  + 10420.23010099111 *T)
    sumX1  +=       0.109 * cos(0.50597705280  + 633.02755679670 *T)
    sumX1  +=       0.103 * cos(0.11832394316  + 6295.04902031090 *T)
    sumX1  +=       0.103 * cos(0.24702432231  + 6271.59031463890 *T)
    sumX1  +=       0.104 * cos(5.58102765204  + 23013.78335707070 *T)
    sumX1  +=       0.105 * cos(3.75605193168  + 11790.38527117530 *T)
    sumX1  +=       0.109 * cos(2.65465737803  + 3340.36860921629 *T)
    sumX1  +=       0.102 * cos(2.59681923716  + 10970.20907518170 *T)
    sumX1  +=       0.101 * cos(4.42190708814  + 4803.96545844350 *T)
    sumX1  +=       0.114 * cos(2.21750546661  + 4171.18171913030 *T)
    sumX1  +=       0.111 * cos(3.66232090402  + 1582.20316576650 *T)
    sumX1  +=       0.107 * cos(2.47220154064  + 23543.47432216530 *T)
    sumX1  +=       0.099 * cos(2.67705199399  + 10973.31186886650 *T)
    sumX1  +=       0.102 * cos(5.67042027327  + 6549.92670919670 *T)
    sumX1  +=       0.102 * cos(0.97811329938  + 6016.71262575310 *T)
    sumX1  +=       0.120 * cos(2.49695982999  + 11514.12713427790 *T)
    sumX1  +=       0.095 * cos(2.69903868611  + 17267.51201917470 *T)
    sumX1  +=       0.121 * cos(0.05336928445  + 13625.77744765550 *T)
    sumX1  +=       0.100 * cos(2.04355401267  + 10976.83498721549 *T)
    sumX1  +=       0.090 * cos(3.68840859712  + 76.50988875911 *T)
    sumX1  +=       0.100 * cos(4.37665365923  + 8661.99650607950 *T)
    sumX1  +=       0.089 * cos(5.33255005639  + 1350.11122714230 *T)
    sumX1  +=       0.087 * cos(4.33390332130  + 9917.94069199330 *T)
    sumX1  +=       0.086 * cos(5.93949751968  + 12809.12412144031 *T)
    sumX1  +=       0.095 * cos(1.34358559692  + 12345.98287502750 *T)
    sumX1  +=       0.084 * cos(1.54530867800  + 775.27879384050 *T)
    sumX1  +=       0.106 * cos(5.14280637669  + 11216.52810780750 *T)
    sumX1  +=       0.095 * cos(2.50046187937  + 12012.82611462390 *T)
    sumX1  +=       0.082 * cos(0.01600918310  + 5643.42238116090 *T)
    sumX1  +=       0.081 * cos(3.84772162011  + 11614.67711121570 *T)
    sumX1  +=       0.084 * cos(1.14755262027  + 6255.91811137810 *T)
    sumX1  +=       0.084 * cos(5.50098095238  + 6310.72122357170 *T)
    sumX1  +=       0.080 * cos(0.34711727404  + 6923.21695378890 *T)
    sumX1  +=       0.085 * cos(5.79487336888  + 5635.82119919310 *T)
    sumX1  +=       0.084 * cos(3.04423309467  + 10440.03047512009 *T)
    sumX1  +=       0.080 * cos(4.19702378753  + 3583.58484815730 *T)
    sumX1  +=       0.092 * cos(0.87932993874  + 10818.37910439930 *T)
    sumX1  +=       0.077 * cos(5.70308186355  + 6993.25271603320 *T)
    sumX1  +=       0.080 * cos(1.36958890049  + 13521.99525907490 *T)
    sumX1  +=       0.104 * cos(5.42045444776  + 5650.04829319470 *T)
    sumX1  +=       0.093 * cos(0.29330855376  + 4732.27444482690 *T)
    sumX1  +=       0.085 * cos(0.17489631763  + 22805.49174851010 *T)
    sumX1  +=       0.086 * cos(0.81273058913  + 12360.20996902910 *T)
    sumX1  +=       0.075 * cos(4.00156639703  + 5120.35732810009 *T)
    sumX1  +=       0.090 * cos(0.62558959530  + 6370.62787201471 *T)
    sumX1  +=       0.090 * cos(6.02294397735  + 6196.01146293510 *T)
    sumX1  +=       0.082 * cos(0.19767948100  + 18842.35782045690 *T)
    sumX1  +=       0.101 * cos(6.08995730057  + 3097.64000524229 *T)
    sumX1  +=       0.094 * cos(5.45896547971  + 10177.01386205009 *T)
    sumX1  +=       0.083 * cos(5.31992956371  + 18415.75962958090 *T)
    sumX1  +=       0.073 * cos(6.09826272325  + 949.41942645330 *T)
    sumX1  +=       0.101 * cos(2.86248284119  + 3104.68624194030 *T)
    sumX1  +=       0.074 * cos(5.96635290969  + 13517.62628874989 *T)
    sumX1  +=       0.092 * cos(2.35039346805  + 16858.72635041670 *T)
    sumX1  +=       0.075 * cos(0.00198703030  + 1059.13811270570 *T)
    sumX1  +=       0.085 * cos(6.09242867296  + 398.39282089170 *T)
    sumX1  +=       0.069 * cos(0.89145782253  + 11713.19913571430 *T)
    sumX1  +=       0.084 * cos(4.35477901935  + 2378.92065608810 *T)
    sumX1  +=       0.069 * cos(0.61262618132  + 11712.71150074729 *T)
    sumX1  +=       0.075 * cos(1.62832949134  + 30356.24113725130 *T)
    sumX1  +=       0.067 * cos(4.61214364579  + 3154.93090237910 *T)
    sumX1  +=       0.076 * cos(4.83480996889  + 18429.98672358250 *T)
    sumX1  +=       0.068 * cos(4.26265140929  + 11925.51791008410 *T)
    sumX1  +=       0.067 * cos(3.24609638890  + 10454.25756912169 *T)
    sumX1  +=       0.073 * cos(6.07047724510  + 15670.83794192309 *T)
    sumX1  +=       0.065 * cos(0.20786131115  + 6438.25243194210 *T)
    sumX1  +=       0.069 * cos(3.29499829591  + 17298.42614480970 *T)
    sumX1  +=       0.087 * cos(5.09754892709  + 3904.15519390330 *T)
    sumX1  +=       0.076 * cos(0.73384165663  + 5231.05128428650 *T)
    sumX1  +=       0.064 * cos(3.69394501139  + 3496.27664361750 *T)
    sumX1  +=       0.078 * cos(4.31639568529  + 8672.21368792411 *T)
    sumX1  +=       0.066 * cos(3.41816290372  + 14143.25142494710 *T)
    sumX1  +=       0.063 * cos(0.99537559299  + 24357.02460612510 *T)
    sumX1  +=       0.063 * cos(5.29204726636  + 15720.59496739490 *T)
    sumX1  +=       0.061 * cos(3.72699038600  + 16460.57734700850 *T)
    sumX1  +=       0.073 * cos(4.50408758477  + 9387.76209193169 *T)
    sumX1  +=       0.057 * cos(2.80099560279  + 17259.91083720689 *T)
    sumX1  +=       0.059 * cos(0.69432195316  + 9778.86485864190 *T)
    sumX1  +=       0.058 * cos(3.24953926506  + 155.66421691770 *T)
    sumX1  +=       0.060 * cos(3.51548937348  + 34570.31015233610 *T)
    sumX1  +=       0.057 * cos(4.46074610322  + 149854.64395229140 *T)
    sumX1  +=       0.056 * cos(5.33393809517  + 12456.18919624690 *T)
    sumX1  +=       0.056 * cos(0.91240244730  + 17996.27498570570 *T)
    sumX1  +=       0.056 * cos(2.17966368743  + 11764.57458634250 *T)
    sumX1  +=       0.054 * cos(2.96145051332  + 4705.48849006010 *T)
    sumX1  +=       0.055 * cos(4.53949491324  + 13915.77529215809 *T)
    sumX1  +=       0.061 * cos(0.15191781763  + 28237.47727687290 *T)
    sumX1  +=       0.065 * cos(5.87045161229  + 7335.58805066330 *T)
    sumX1  +=       0.072 * cos(0.77556931591  + 6055.84353468590 *T)
    sumX1  +=       0.072 * cos(5.87296425674  + 6510.79580026390 *T)
    sumX1  +=       0.071 * cos(0.23083621209  + 12545.62012197390 *T)
    sumX1  +=       0.061 * cos(1.27148175728  + 6312.74917601091 *T)
    sumX1  +=       0.061 * cos(5.37705181537  + 6253.89015893890 *T)
    sumX1  +=       0.052 * cos(0.26640147565  + 5326.54287653730 *T)
    sumX1  +=       0.054 * cos(1.21229891621  + 2699.49100183409 *T)
    sumX1  +=       0.052 * cos(2.44818566509  + 8983.05448679250 *T)
    sumX1  +=       0.066 * cos(2.35337977576  + 14919.26167123810 *T)
    sumX1  +=       0.050 * cos(5.93502263343  + 5572.89898394960 *T)
    sumX1  +=       0.051 * cos(0.76420455409  + 16522.41589851870 *T)
    sumX1  +=       0.049 * cos(4.18738074622  + 35579.93505705350 *T)
    sumX1  +=       0.053 * cos(1.36303997543  + 6208.53806890760 *T)
    sumX1  +=       0.053 * cos(5.28549359722  + 6358.10126604220 *T)
    sumX1  +=       0.049 * cos(4.92672897893  + 21393.78578734110 *T)
    sumX1  +=       0.049 * cos(6.04327871377  + 5547.44315394310 *T)
    sumX1  +=       0.049 * cos(0.60525485888  + 7019.19618100671 *T)
    sumX1  +=       0.049 * cos(2.59123701567  + 12416.83232033170 *T)
    sumX1  +=       0.050 * cos(5.65157500725  + 12592.20620229910 *T)
    sumX1  +=       0.050 * cos(0.07520350695  + 26084.26562369970 *T)
    sumX1  +=       0.051 * cos(3.40680645218  + 23006.66981006990 *T)
    sumX1  +=       0.058 * cos(2.46543782438  + 15141.14697682849 *T)
    sumX1  +=       0.047 * cos(0.97665096734  + 24279.35083569710 *T)
    sumX1  +=       0.049 * cos(4.98006572526  + 6816.53375091851 *T)
    sumX1  +=       0.049 * cos(1.66846784739  + 5750.10558403130 *T)
    sumX1  +=       0.046 * cos(3.45134537804  + 19379.16233255230 *T)
    sumX1  +=       0.052 * cos(4.17729143118  + 16737.33341911309 *T)
    sumX1  +=       0.047 * cos(2.09298679166  + 24066.05174025910 *T)
    sumX1  +=       0.046 * cos(5.73796654718  + 104351.85638378028 *T)
    sumX1  +=       0.063 * cos(1.54245276158  + 8662.48414104651 *T)
    sumX1  +=       0.048 * cos(1.71424177408  + 26735.70144472970 *T)
    sumX1  +=       0.046 * cos(5.48854983581  + 12132.68377958950 *T)
    sumX1  +=       0.050 * cos(4.49443808542  + 24602.85625235450 *T)
    sumX1  +=       0.054 * cos(0.03975141461  + 7834.36489012290 *T)
    sumX1  +=       0.045 * cos(5.05887207146  + 161000.92955515761 *T)
    sumX1  +=       0.049 * cos(4.51484833854  + 29303.97275406290 *T)
    sumX1  +=       0.049 * cos(3.00236481482  + 10984.43616918330 *T)
    sumX1  +=       0.045 * cos(5.42122497288  + 45584.92899470390 *T)
    sumX1  +=       0.046 * cos(3.44074843825  + 5244.29305668450 *T)
    sumX1  +=       0.048 * cos(1.96431249949  + 23581.01435983410 *T)
    sumX1  +=       0.049 * cos(0.60097746518  + 19650.80466361450 *T)
    sumX1  +=       0.049 * cos(4.98977220505  + 29289.74566006130 *T)
    sumX1  +=       0.044 * cos(3.01088620106  + 6390.98319141350 *T)
    sumX1  +=       0.044 * cos(3.63764737159  + 6175.65614353630 *T)
    sumX1  +=       0.051 * cos(1.08500510566  + 17789.60180230149 *T)
    sumX1  +=       0.042 * cos(1.47304377234  + 24705.94902657310 *T)
    sumX1  +=       0.042 * cos(4.33726197214  + 15663.79170522509 *T)
    sumX1  +=       0.043 * cos(1.51170891404  + 16460.08971204149 *T)
    sumX1  +=       0.043 * cos(1.85670214007  + 553.81322032590 *T)
    sumX1  +=       0.041 * cos(4.78695133916  + 17370.60479339330 *T)
    sumX1  +=       0.053 * cos(1.02071442021  + 13119.47728534169 *T)
    sumX1  +=       0.042 * cos(0.40756986853  + 4164.55580709650 *T)
    sumX1  +=       0.041 * cos(3.62600931231  + 23020.89690407150 *T)
    sumX1  +=       0.045 * cos(2.13148338898  + 54247.16931826690 *T)
    sumX1  +=       0.040 * cos(3.00393862246  + 32243.25468339710 *T)
    sumX1  +=       0.040 * cos(0.45945430396  + 24336.24921063270 *T)
    sumX1  +=       0.040 * cos(0.14772670564  + 11769.60987568290 *T)
    sumX1  +=       0.045 * cos(4.68905632727  + 16723.10632511149 *T)
    sumX1  +=       0.045 * cos(1.54746367368  + 16723.59396007850 *T)
    sumX1  +=       0.042 * cos(4.32247442790  + 8827.14645239129 *T)
    sumX1  +=       0.040 * cos(0.00254937606  + 8402.08352785330 *T)
    sumX1  +=       0.039 * cos(5.06793905702  + 16062.42834360030 *T)
    sumX1  +=       0.044 * cos(2.87431244066  + 5856.23384163189 *T)
    sumX1  +=       0.038 * cos(2.71291798518  + 6297.54676147650 *T)
    sumX1  +=       0.038 * cos(3.93561558747  + 6269.09257347330 *T)
    sumX1  +=       0.041 * cos(4.95313128130  + 15508.85894075790 *T)
    sumX1  +=       0.038 * cos(0.25558032671  + 6394.50630976251 *T)
    sumX1  +=       0.038 * cos(0.10976793876  + 6172.13302518730 *T)
    sumX1  +=       0.039 * cos(3.31895424299  + 29826.55017215670 *T)
    sumX1  +=       0.046 * cos(5.02561530595  + 17297.93850984270 *T)
    sumX1  +=       0.041 * cos(2.47236149910  + 6309.13035230770 *T)
    sumX1  +=       0.038 * cos(1.11170997081  + 29089.05523346850 *T)
    sumX1  +=       0.038 * cos(6.16488744565  + 11933.61177815310 *T)
    sumX1  +=       0.037 * cos(4.91170884792  + 26482.41462710790 *T)
    sumX1  +=       0.037 * cos(2.07642652738  + 20452.62559473830 *T)
    sumX1  +=       0.037 * cos(5.02138933475  + 18073.46112116669 *T)
    sumX1  +=       0.036 * cos(1.41330554556  + 11499.41240530929 *T)
    sumX1  +=       0.045 * cos(0.04410300421  + 3340.85624418330 *T)
    sumX1  +=       0.049 * cos(5.78167784671  + 18216.68762814450 *T)
    sumX1  +=       0.038 * cos(6.16266694488  + 18202.46053414290 *T)
    sumX1  +=       0.041 * cos(5.48085477947  + 5216.82419028490 *T)
    sumX1  +=       0.036 * cos(4.64059514280  + 4060.97539791089 *T)
    sumX1  +=       0.048 * cos(0.43106257187  + 951.96222373410 *T)
    sumX1  +=       0.041 * cos(2.33852383906  + 18848.98373249069 *T)
    sumX1  +=       0.038 * cos(6.17960213131  + 26709.89075989690 *T)
    sumX1  +=       0.036 * cos(4.69601378640  + 28230.43104017490 *T)
    sumX1  +=       0.036 * cos(0.39760453863  + 24491.18197509989 *T)
    sumX1  +=       0.036 * cos(6.28050310342  + 9924.56660402710 *T)
    sumX1  +=       0.048 * cos(0.32621587842  + 8982.56685182550 *T)
    sumX1  +=       0.037 * cos(5.17748996647  + 22484.09239197610 *T)
    sumX1  +=       0.047 * cos(1.64189885412  + 6040.10342853389 *T)
    sumX1  +=       0.034 * cos(5.26052334707  + 6418.94499248490 *T)
    sumX1  +=       0.034 * cos(1.38801022558  + 6147.69434246491 *T)
    sumX1  +=       0.034 * cos(2.09927479760  + 9380.71585523369 *T)
    sumX1  +=       0.036 * cos(5.01642644412  + 7238.43177411650 *T)
    sumX1  +=       0.034 * cos(1.97967235505  + 17272.12782500990 *T)
    sumX1  +=       0.037 * cos(2.80234472663  + 58953.38926077750 *T)
    sumX1  +=       0.034 * cos(1.90506635542  + 29026.72904699130 *T)
    sumX1  +=       0.033 * cos(2.30422309015  + 12721.32828193349 *T)
    sumX1  +=       0.035 * cos(3.20661909223  + 7322.34627826531 *T)
    sumX1  +=       0.044 * cos(2.79016934175  + 3981.73385156551 *T)
    sumX1  +=       0.033 * cos(2.02741588975  + 19804.58347409930 *T)
    sumX1  +=       0.033 * cos(4.13127522861  + 173567.08125514039 *T)
    sumX1  +=       0.033 * cos(4.59578048276  + 205.94173095370 *T)
    sumX1  +=       0.033 * cos(3.87810188817  + 23013.29572210369 *T)
    sumX1  +=       0.034 * cos(0.56182222293  + 49515.13869092350 *T)
    sumX1  +=       0.034 * cos(4.95130100456  + 38500.51984855570 *T)
    sumX1  +=       0.032 * cos(4.38300866562  + 17686.99666304990 *T)
    sumX1  +=       0.036 * cos(0.10343586811  + 5641.95442512570 *T)
    sumX1  +=       0.032 * cos(4.48170445266  + 21228.63584102931 *T)
    sumX1  +=       0.033 * cos(1.70919210465  + 34520.55312686430 *T)
    sumX1  +=       0.033 * cos(4.08300924175  + 8635.69818627969 *T)
    sumX1  +=       0.031 * cos(0.30962555206  + 5746.02752041250 *T)
    sumX1  +=       0.041 * cos(4.09101330171  + 21202.33752122950 *T)
    sumX1  +=       0.037 * cos(1.21597687953  + 7349.81514466491 *T)
    sumX1  +=       0.034 * cos(0.61019544663  + 16496.11757871889 *T)
    sumX1  +=       0.032 * cos(2.47223060088  + 33019.26492968810 *T)
    sumX1  +=       0.032 * cos(6.25412877228  + 21150.56954840009 *T)
    sumX1  +=       0.037 * cos(3.41910013684  + 29296.85920706210 *T)
    sumX1  +=       0.036 * cos(3.45170661535  + 27511.22405605370 *T)
    sumX1  +=       0.035 * cos(2.20855027459  + 11933.12414318609 *T)
    sumX1  +=       0.035 * cos(2.69261850768  + 11918.89704918449 *T)
    sumX1  +=       0.035 * cos(0.83074174405  + 29062.75691366870 *T)
    sumX1  +=       0.030 * cos(2.96012497927  + 12490.12944619070 *T)
    sumX1  +=       0.032 * cos(0.38688555099  + 6334.60000533731 *T)
    sumX1  +=       0.032 * cos(6.26164802166  + 6232.03932961249 *T)
    sumX1  +=       0.029 * cos(2.48914084100  + 419.24082639170 *T)
    sumX1  +=       0.029 * cos(2.48914084100  + 419.72846135871 *T)
    sumX1  +=       0.029 * cos(0.09098498823  + 16858.23871544969 *T)
    sumX1  +=       0.030 * cos(4.30269873097  + 16208.13008898551 *T)
    sumX1  +=       0.029 * cos(2.26254947152  + 27278.71263392430 *T)
    sumX1  +=       0.032 * cos(3.16446260317  + 5017.26455388150 *T)
    sumX1  +=       0.028 * cos(1.87847577782  + 24080.27883426070 *T)
    sumX1  +=       0.030 * cos(2.57345584103  + 23539.95120381630 *T)
    sumX1  +=       0.029 * cos(6.17121903514  + 18606.25512851669 *T)
    sumX1  +=       0.030 * cos(3.16071544619  + 30665.91114094930 *T)
    sumX1  +=       0.034 * cos(5.86279015300  + 6660.69327539070 *T)
    sumX1  +=       0.034 * cos(0.78574341965  + 5905.94605955911 *T)
    sumX1  +=       0.038 * cos(5.12222686816  + 31571.04346687470 *T)
    sumX1  +=       0.038 * cos(2.95935499043  + 16061.94070863329 *T)
    sumX1  +=       0.029 * cos(0.07805875622  + 23539.46356884929 *T)
    sumX1  +=       0.033 * cos(1.31621330654  + 3738.51761262449 *T)
    sumX1  +=       0.028 * cos(4.45185289388  + 2942.21960580809 *T)
    sumX1  +=       0.032 * cos(4.32330584281  + 15110.70993734970 *T)
    sumX1  +=       0.030 * cos(1.39195112722  + 55023.17956455790 *T)
    sumX1  +=       0.027 * cos(0.53286078729  + 233141.55822184496 *T)
    sumX1  +=       0.030 * cos(2.32405719366  + 12587.17091295870 *T)
    sumX1  +=       0.028 * cos(5.54771281521  + 34911.65589357450 *T)
    sumX1  +=       0.027 * cos(6.17188911074  + 1692.40948698591 *T)
    sumX1  +=       0.027 * cos(4.56663555607  + 2544.07060239989 *T)
    sumX1  +=       0.027 * cos(3.42500443489  + 24382.83529095790 *T)
    sumX1  +=       0.027 * cos(5.99215939135  + 6205.64589704690 *T)
    sumX1  +=       0.027 * cos(0.65637418130  + 6360.99343790291 *T)
    sumX1  +=       0.027 * cos(0.24480889857  + 31172.89446346650 *T)
    sumX1  +=       0.026 * cos(3.76734683925  + 36949.47462590770 *T)
    sumX1  +=       0.034 * cos(6.02674277322  + 18053.17336064130 *T)
    sumX1  +=       0.027 * cos(2.32047706036  + 32367.34147369110 *T)
    sumX1  +=       0.027 * cos(1.77181133426  + 4487.57358878689 *T)
    sumX1  +=       0.028 * cos(3.80331914659  + 10239.34004852730 *T)
    sumX1  +=       0.026 * cos(6.01470923824  + 4796.85191144269 *T)
    sumX1  +=       0.026 * cos(6.27941236443  + 6226.42129253930 *T)
    sumX1  +=       0.026 * cos(0.36912120822  + 6340.21804241050 *T)
    sumX1  +=       0.029 * cos(0.18788001431  + 5974.04134481910 *T)
    sumX1  +=       0.029 * cos(0.17746825116  + 6592.59799013070 *T)
    sumX1  +=       0.026 * cos(3.39870851542  + 19402.55313533309 *T)
    sumX1  +=       0.028 * cos(2.42845493949  + 46386.74992582770 *T)
    sumX1  +=       0.032 * cos(0.33223863826  + 9225.78309076650 *T)
    sumX1  +=       0.026 * cos(2.54665382210  + 9910.33951002549 *T)
    sumX1  +=       0.025 * cos(5.74078884001  + 15265.64270181689 *T)
    sumX1  +=       0.029 * cos(5.90495655038  + 52176.05010063190 *T)
    sumX1  +=       0.027 * cos(4.22338082584  + 62883.59895699710 *T)
    sumX1  +=       0.024 * cos(0.55027097677  + 11617.21990849651 *T)
    sumX1  +=       0.024 * cos(2.62016281496  + 5120.84496306710 *T)
    sumX1  +=       0.028 * cos(3.25064295131  + 6680.98103591609 *T)
    sumX1  +=       0.024 * cos(4.40689701323  + 522.33360061030 *T)
    sumX1  +=       0.025 * cos(3.86362291910  + 6263.64990657511 *T)
    sumX1  +=       0.025 * cos(2.78491065355  + 6302.98942837470 *T)
    sumX1  +=       0.029 * cos(2.44405132476  + 23581.50199480110 *T)
    sumX1  +=       0.024 * cos(0.50709291410  + 29424.87805039950 *T)
    sumX1  +=       0.028 * cos(2.66240986547  + 16737.82105408010 *T)
    sumX1  +=       0.024 * cos(1.98880441412  + 22743.16556203289 *T)
    sumX1  +=       0.024 * cos(5.13039706771  + 22743.65319699990 *T)
    sumX1  +=       0.024 * cos(5.85302121512  + 25158.84553724890 *T)
    sumX1  +=       0.025 * cos(3.95115098429  + 21424.22282681990 *T)
    sumX1  +=       0.025 * cos(0.80955833070  + 21424.71046178690 *T)
    sumX1  +=       0.024 * cos(1.92863541994  + 14712.07329897449 *T)
    sumX1  +=       0.028 * cos(4.50601984314  + 5760.25461441409 *T)
    sumX1  +=       0.024 * cos(3.73455640738  + 31969.19247028290 *T)
    sumX1  +=       0.023 * cos(0.62345777089  + 17893.18221148710 *T)
    sumX1  +=       0.030 * cos(3.85563309519  + 8584.90548338430 *T)
    sumX1  +=       0.024 * cos(3.21746306158  + 23550.58786916610 *T)
    sumX1  +=       0.023 * cos(1.36063534536  + 6717.03140535170 *T)
    sumX1  +=       0.022 * cos(3.99741791962  + 7445.79437188270 *T)
    sumX1  +=       0.023 * cos(5.13650786055  + 5849.60792959810 *T)
    sumX1  +=       0.025 * cos(0.81555419576  + 14155.46841275010 *T)
    sumX1  +=       0.023 * cos(3.58569412080  + 6294.36536773881 *T)
    sumX1  +=       0.023 * cos(3.06283945185  + 6272.27396721100 *T)
    sumX1  +=       0.024 * cos(4.24597953395  + 18208.10612510850 *T)
    sumX1  +=       0.024 * cos(1.10438688036  + 18208.59376007550 *T)
    sumX1  +=       0.021 * cos(5.78251945962  + 22779.68106367730 *T)
    sumX1  +=       0.029 * cos(3.31448732534  + 28628.58004358310 *T)
    sumX1  +=       0.021 * cos(5.13427576878  + 40879.68432212730 *T)
    sumX1  +=       0.021 * cos(4.08575884387  + 30221.17605721590 *T)
    sumX1  +=       0.022 * cos(1.38262024424  + 4379.88285497370 *T)
    sumX1  +=       0.022 * cos(5.26591980546  + 8186.75647997610 *T)
    sumX1  +=       0.022 * cos(4.52225402167  + 28760.05469496669 *T)
    sumX1  +=       0.026 * cos(3.41053115091  + 18422.87317658170 *T)
    sumX1  +=       0.020 * cos(0.80720267360  + 12036.21691740469 *T)
    sumX1  +=       0.020 * cos(4.78724181293  + 28313.04498717750 *T)
    sumX1  +=       0.027 * cos(0.67000770864  + 34513.50689016630 *T)
    sumX1  +=       0.020 * cos(2.07630932980  + 34115.35788675810 *T)
    sumX1  +=       0.020 * cos(1.72148703277  + 21548.71855180830 *T)
    sumX1  +=       0.020 * cos(3.68576430300  + 17654.53672226610 *T)
    sumX1  +=       0.025 * cos(3.53209868087  + 55798.70217588190 *T)
    sumX1  +=       0.020 * cos(0.81961551511  + 20597.48778052470 *T)
    sumX1  +=       0.020 * cos(2.60684350676  + 18875.76968725751 *T)
    sumX1  +=       0.022 * cos(0.44774356245  + 1588.82907780029 *T)
    sumX1  +=       0.024 * cos(1.20852424503  + 9411.70843257070 *T)
    sumX1  +=       0.019 * cos(5.85945140949  + 536.56069461190 *T)
    sumX1  +=       0.020 * cos(4.99560905506  + 6187.34044025710 *T)
    sumX1  +=       0.020 * cos(1.65292451759  + 6379.29889469270 *T)
    sumX1  +=       0.022 * cos(4.44471803531  + 18326.89394936390 *T)
    sumX1  +=       0.022 * cos(2.18348213630  + 6233.56264200310 *T)
    sumX1  +=       0.022 * cos(4.46505143635  + 6333.07669294670 *T)
    sumX1  +=       0.018 * cos(2.03036294822  + 5972.47886860650 *T)
    sumX1  +=       0.018 * cos(4.61817062443  + 6594.16046634330 *T)
    sumX1  +=       0.019 * cos(4.28397604583  + 6948.07571260490 *T)
    sumX1  +=       0.019 * cos(2.36455752682  + 5618.56362234490 *T)
    sumX1  +=       0.019 * cos(0.18461366541  + 83974.07917321339 *T)
    sumX1  +=       0.019 * cos(4.29574443999  + 84020.10309797739 *T)
    sumX1  +=       0.020 * cos(3.84957027230  + 50316.95962204730 *T)
    sumX1  +=       0.020 * cos(2.85391796921  + 9070.36269133230 *T)

    # EARTH X Coord
    sumX2  = 0
    sumX2  +=   52911.498 * cos(3.32403354915  + 6283.31966747490 *T)
    sumX2  +=    6074.441 * cos(2.05485843872  + 0.24381748350 *T)
    sumX2  +=    2406.871 * cos(4.23086027149  + 12566.39551746630 *T)
    sumX2  +=      96.033 * cos(4.14485769516  + 18849.47136745770 *T)
    sumX2  +=      29.888 * cos(0.73436192593  + 6282.83203250789 *T)
    sumX2  +=      14.021 * cos(3.21121684298  + 6279.79654912590 *T)
    sumX2  +=      13.375 * cos(3.31400005464  + 6286.84278582391 *T)
    sumX2  +=       8.148 * cos(2.19949162987  + 6309.61798727471 *T)
    sumX2  +=       8.195 * cos(4.38954108022  + 6257.02134767510 *T)
    sumX2  +=       8.001 * cos(5.91956929382  + 6127.89926804070 *T)
    sumX2  +=       7.754 * cos(0.72231700953  + 6438.74006690910 *T)
    sumX2  +=       7.223 * cos(5.38114271434  + 83997.09113559539 *T)
    sumX2  +=       4.616 * cos(6.16703227370  + 710.17686604180 *T)
    sumX2  +=       3.060 * cos(4.83964573312  + 4705.97612502710 *T)
    sumX2  +=       3.717 * cos(2.45040909403  + 5507.79705615090 *T)
    sumX2  +=       2.991 * cos(4.05993241608  + 25132.54721744910 *T)
    sumX2  +=       3.235 * cos(4.10584554352  + 7058.84227879890 *T)
    sumX2  +=       2.393 * cos(3.40776365917  + 775.76642880750 *T)
    sumX2  +=       2.034 * cos(4.34141229883  + 6290.43321447570 *T)
    sumX2  +=       2.028 * cos(2.28023581144  + 6276.20612047410 *T)
    sumX2  +=       2.064 * cos(1.58527926677  + 7860.66320992270 *T)
    sumX2  +=       1.914 * cos(4.42314112548  + 5487.02166065850 *T)
    sumX2  +=       1.770 * cos(5.05862783743  + 1059.62574767270 *T)
    sumX2  +=       1.761 * cos(3.37902852940  + 5753.62870238030 *T)
    sumX2  +=       1.558 * cos(2.13660228192  + 7079.61767429131 *T)
    sumX2  +=       1.551 * cos(2.12932375831  + 6040.59106350090 *T)
    sumX2  +=       1.436 * cos(2.85243683562  + 5885.17066406670 *T)
    sumX2  +=       1.469 * cos(4.51404554645  + 6526.04827144891 *T)
    sumX2  +=       1.319 * cos(3.46234944603  + 6681.46867088311 *T)
    sumX2  +=       1.180 * cos(2.99170595830  + 6813.01063256950 *T)
    sumX2  +=       1.337 * cos(1.03982518814  + 5729.75026463250 *T)
    sumX2  +=       1.039 * cos(5.82323058792  + 6282.33934640670 *T)
    sumX2  +=       1.313 * cos(2.16977183919  + 529.93478257810 *T)
    sumX2  +=       1.041 * cos(0.76562479642  + 6284.29998854310 *T)
    sumX2  +=       1.202 * cos(5.59909179750  + 6836.88907031730 *T)
    sumX2  +=       0.904 * cos(2.42848578586  + 5331.60126122430 *T)
    sumX2  +=       0.870 * cos(3.83740101245  + 4933.45225781610 *T)
    sumX2  +=       1.149 * cos(4.59530313333  + 167284.00540514899 *T)
    sumX2  +=       1.069 * cos(5.80442130983  + 1577.09972496430 *T)
    sumX2  +=       0.772 * cos(4.23369730112  + 7235.03807372550 *T)
    sumX2  +=       0.846 * cos(2.93878609085  + 11790.87290614230 *T)
    sumX2  +=       0.742 * cos(0.30768536261  + 4137.15425099970 *T)
    sumX2  +=       0.966 * cos(3.62307276164  + 11856.46246890800 *T)
    sumX2  +=       0.664 * cos(2.78855764568  + 7633.18707713370 *T)
    sumX2  +=       0.662 * cos(4.76048207091  + 1194.69082770810 *T)
    sumX2  +=       0.659 * cos(1.23774551244  + 11507.01358727710 *T)
    sumX2  +=       0.853 * cos(5.22418483353  + 4535.30325440790 *T)
    sumX2  +=       0.823 * cos(5.77365627974  + 155.17658195069 *T)
    sumX2  +=       0.783 * cos(5.87375907691  + 5088.87265725030 *T)
    sumX2  +=       0.785 * cos(5.03738396313  + 2353.10997125530 *T)
    sumX2  +=       0.694 * cos(1.93706172082  + 801.57711364030 *T)
    sumX2  +=       0.566 * cos(6.00002295544  + 796.54182429990 *T)
    sumX2  +=       0.684 * cos(2.68347359202  + 3128.63258257930 *T)
    sumX2  +=       0.499 * cos(2.99674498554  + 12569.91863581531 *T)
    sumX2  +=       0.596 * cos(1.38539217884  + 8031.33608054190 *T)
    sumX2  +=       0.486 * cos(6.11316978391  + 1589.31671276730 *T)
    sumX2  +=       0.483 * cos(6.25283705744  + 8429.48508395010 *T)
    sumX2  +=       0.441 * cos(0.27873568624  + 6496.61876291290 *T)
    sumX2  +=       0.412 * cos(5.58840963851  + 4693.75913722409 *T)
    sumX2  +=       0.561 * cos(0.75091675093  + 7477.76667769950 *T)
    sumX2  +=       0.390 * cos(6.20968064366  + 6070.02057203690 *T)
    sumX2  +=       0.405 * cos(1.55299195368  + 3739.00524759150 *T)
    sumX2  +=       0.501 * cos(3.39229516665  + 1592.83983111630 *T)
    sumX2  +=       0.392 * cos(5.39455294636  + 26.05450231630 *T)
    sumX2  +=       0.369 * cos(4.36331143713  + 553.32558535889 *T)
    sumX2  +=       0.483 * cos(1.36538995021  + 12036.70455237170 *T)
    sumX2  +=       0.489 * cos(2.13588971106  + 5223.45010231870 *T)
    sumX2  +=       0.481 * cos(4.74275484789  + 10213.52936369450 *T)
    sumX2  +=       0.448 * cos(0.85608336591  + 156137.71980228281 *T)
    sumX2  +=       0.321 * cos(2.96911879180  + 951.47458876710 *T)
    sumX2  +=       0.320 * cos(2.00218171333  + 1990.98883452450 *T)
    sumX2  +=       0.361 * cos(0.59519501420  + 12565.90788249930 *T)
    sumX2  +=       0.361 * cos(0.08584353459  + 12721.81591690050 *T)
    sumX2  +=       0.412 * cos(5.91338851626  + 398.39282089170 *T)
    sumX2  +=       0.304 * cos(0.96839213314  + 4690.72365384210 *T)
    sumX2  +=       0.356 * cos(1.28855678128  + 5481.49873635110 *T)
    sumX2  +=       0.283 * cos(2.23992178340  + 242.97242145751 *T)
    sumX2  +=       0.362 * cos(3.88633843420  + 9438.00675237050 *T)
    sumX2  +=       0.248 * cos(4.75322854174  + 8827.63408735830 *T)
    sumX2  +=       0.252 * cos(1.16164987849  + 3154.44326741210 *T)
    sumX2  +=       0.257 * cos(1.79452152688  + 11371.94850724170 *T)
    sumX2  +=       0.327 * cos(3.02081233401  + 14143.73905991410 *T)
    sumX2  +=       0.230 * cos(0.23367695814  + 1747.77259558350 *T)
    sumX2  +=       0.289 * cos(5.24697142787  + 7085.14059859870 *T)
    sumX2  +=       0.297 * cos(1.59965866765  + 1349.62359217529 *T)
    sumX2  +=       0.209 * cos(0.27341674720  + 11770.09751064990 *T)
    sumX2  +=       0.207 * cos(4.92862210539  + 4164.06817212950 *T)
    sumX2  +=       0.188 * cos(5.64673454004  + 7875.91568110771 *T)
    sumX2  +=       0.189 * cos(0.60545229996  + 2389.13783793270 *T)
    sumX2  +=       0.194 * cos(0.41046959310  + 10977.32262218251 *T)
    sumX2  +=       0.176 * cos(4.08570919572  + 7342.70159766410 *T)
    sumX2  +=       0.164 * cos(2.79150408226  + 5223.93773728571 *T)
    sumX2  +=       0.205 * cos(3.21370431881  + 10973.79950383350 *T)
    sumX2  +=       0.205 * cos(2.25101785528  + 13368.21644859010 *T)
    sumX2  +=       0.139 * cos(4.61869477858  + 10575.65050042530 *T)
    sumX2  +=       0.155 * cos(5.81412821499  + 12410.97511803210 *T)
    sumX2  +=       0.152 * cos(0.10500467147  + 3930.45351370310 *T)
    sumX2  +=       0.129 * cos(5.52193535090  + 397.90518592470 *T)
    sumX2  +=       0.127 * cos(0.96654930095  + 12592.69383726610 *T)
    sumX2  +=       0.132 * cos(0.16052766322  + 8636.18582124671 *T)
    sumX2  +=       0.122 * cos(4.78358483655  + 13119.96492030870 *T)
    sumX2  +=       0.163 * cos(0.93280092706  + 5507.30942118390 *T)
    sumX2  +=       0.114 * cos(5.70431693679  + 17260.39847217390 *T)
    sumX2  +=       0.118 * cos(2.37854817453  + 4292.57465043390 *T)
    sumX2  +=       0.148 * cos(5.63695738547  + 17790.08943726851 *T)
    sumX2  +=       0.123 * cos(3.75785910050  + 12562.87239911730 *T)
    sumX2  +=       0.103 * cos(3.36994741884  + 13518.11392371690 *T)
    sumX2  +=       0.138 * cos(3.83865460083  + 12168.24651405810 *T)
    sumX2  +=       0.107 * cos(4.33977826153  + 10989.05197501850 *T)
    sumX2  +=       0.104 * cos(5.27997888785  + 2145.92159899169 *T)
    sumX2  +=       0.096 * cos(1.97792917854  + 6283.25235717211 *T)
    sumX2  +=       0.099 * cos(2.27831564394  + 13096.08648256090 *T)
    sumX2  +=       0.096 * cos(1.53382620869  + 6283.38697777770 *T)
    sumX2  +=       0.084 * cos(5.20596620689  + 16730.70750707931 *T)
    sumX2  +=       0.111 * cos(3.96322491245  + 12540.09719766650 *T)
    sumX2  +=       0.087 * cos(6.02515097465  + 10177.50149701710 *T)
    sumX2  +=       0.078 * cos(4.18422621711  + 12323.66691349230 *T)
    sumX2  +=       0.076 * cos(5.70862044549  + 6709.91785835091 *T)
    sumX2  +=       0.105 * cos(3.53845405180  + 5642.44206009270 *T)
    sumX2  +=       0.073 * cos(0.89693480363  + 5856.72147659890 *T)
    sumX2  +=       0.102 * cos(0.66758153957  + 250570.91967470257 *T)
    sumX2  +=       0.070 * cos(5.22505818809  + 5753.14106741329 *T)
    sumX2  +=       0.071 * cos(0.56166089035  + 14314.41193053330 *T)
    sumX2  +=       0.092 * cos(1.96485842296  + 13916.26292712510 *T)
    sumX2  +=       0.070 * cos(2.53450052921  + 7084.65296363170 *T)
    sumX2  +=       0.091 * cos(3.10960565577  + 6924.19727485710 *T)
    sumX2  +=       0.083 * cos(3.98989749432  + 31415.62306744050 *T)
    sumX2  +=       0.088 * cos(5.01539409795  + 71430.45180064559 *T)
    sumX2  +=       0.074 * cos(1.96538390591  + 529.44714761109 *T)
    sumX2  +=       0.085 * cos(5.97852929798  + 95143.37673846159 *T)
    sumX2  +=       0.061 * cos(4.23114381660  + 8274.06468451590 *T)
    sumX2  +=       0.060 * cos(5.52155170015  + 16496.60521368590 *T)
    sumX2  +=       0.058 * cos(1.47891111226  + 5642.93474619389 *T)
    sumX2  +=       0.078 * cos(2.41141378465  + 213.54291292150 *T)
    sumX2  +=       0.059 * cos(4.67526538731  + 640.14110379750 *T)
    sumX2  +=       0.057 * cos(2.06129454632  + 4690.23601887509 *T)
    sumX2  +=       0.056 * cos(3.80170948025  + 13341.91812879030 *T)
    sumX2  +=       0.065 * cos(3.38872777921  + 4731.78680985990 *T)
    sumX2  +=       0.051 * cos(4.81168622601  + 3893.93801205869 *T)
    sumX2  +=       0.050 * cos(4.17369676080  + 12573.50906446710 *T)
    sumX2  +=       0.050 * cos(4.41669813786  + 3634.37720703489 *T)
    sumX2  +=       0.049 * cos(2.27712508812  + 18319.78040236310 *T)
    sumX2  +=       0.066 * cos(5.49585584500  + 2787.28684134090 *T)
    sumX2  +=       0.049 * cos(1.99905022377  + 13362.69352428270 *T)
    sumX2  +=       0.048 * cos(3.76233131596  + 12964.54452087450 *T)
    sumX2  +=       0.053 * cos(4.12374270394  + 14919.26167123810 *T)
    sumX2  +=       0.058 * cos(1.23083144160  + 1577.58735993130 *T)
    sumX2  +=       0.052 * cos(1.27828874996  + 15721.08260236190 *T)
    sumX2  +=       0.061 * cos(3.42876383277  + 4292.08701546690 *T)
    sumX2  +=       0.044 * cos(0.31046230867  + 24073.16528725990 *T)
    sumX2  +=       0.044 * cos(0.38954461131  + 10447.14402212089 *T)
    sumX2  +=       0.042 * cos(4.09986255075  + 20426.81490990550 *T)
    sumX2  +=       0.059 * cos(1.01599200374  + 84672.23202702119 *T)
    sumX2  +=       0.051 * cos(5.38279738858  + 6370.62787201471 *T)
    sumX2  +=       0.051 * cos(1.26573618407  + 6196.01146293510 *T)
    sumX2  +=       0.046 * cos(0.52171400874  + 2119.00767786191 *T)
    sumX2  +=       0.041 * cos(4.09327874029  + 3185.43584474911 *T)
    sumX2  +=       0.056 * cos(3.27189314076  + 10976.83498721549 *T)
    sumX2  +=       0.051 * cos(2.37524851068  + 9437.51911740350 *T)
    sumX2  +=       0.049 * cos(3.21159567950  + 239424.63407183639 *T)
    sumX2  +=       0.036 * cos(6.20393463674  + 3495.78900865049 *T)
    sumX2  +=       0.044 * cos(3.69396549327  + 18139.53831889940 *T)
    sumX2  +=       0.035 * cos(3.61386877088  + 3929.96587873609 *T)
    sumX2  +=       0.044 * cos(1.66014811397  + 8661.99650607950 *T)
    sumX2  +=       0.046 * cos(3.78615165617  + 3894.42564702570 *T)
    sumX2  +=       0.034 * cos(3.81444510796  + 6262.54427198250 *T)
    sumX2  +=       0.034 * cos(5.98049768477  + 6304.09506296730 *T)
    sumX2  +=       0.035 * cos(3.58408872678  + 18875.28205229050 *T)
    sumX2  +=       0.035 * cos(6.09075949155  + 10420.23010099111 *T)
    sumX2  +=       0.032 * cos(0.62535692558  + 23013.78335707070 *T)
    sumX2  +=       0.034 * cos(3.61147349841  + 7.35736448430 *T)
    sumX2  +=       0.045 * cos(1.48783646485  + 3340.85624418330 *T)
    sumX2  +=       0.035 * cos(0.38590076970  + 10213.04172872750 *T)
    sumX2  +=       0.031 * cos(5.38472872675  + 14712.56093394150 *T)
    sumX2  +=       0.030 * cos(4.39715700541  + 12809.12412144031 *T)
    sumX2  +=       0.039 * cos(1.14782571257  + 9779.35249360890 *T)
    sumX2  +=       0.041 * cos(5.61324415849  + 82576.73740351178 *T)
    sumX2  +=       0.031 * cos(0.76672410297  + 22779.68106367730 *T)
    sumX2  +=       0.028 * cos(4.07975589699  + 2942.70724077510 *T)
    sumX2  +=       0.028 * cos(2.36060849192  + 5429.63565075589 *T)
    sumX2  +=       0.027 * cos(4.04275590321  + 11014.86265985130 *T)
    sumX2  +=       0.027 * cos(2.11023646997  + 9623.93209417470 *T)
    sumX2  +=       0.025 * cos(1.72579553961  + 18209.57408114370 *T)
    sumX2  +=       0.026 * cos(2.55064187749  + 9381.20349020070 *T)
    sumX2  +=       0.025 * cos(2.68812943588  + 3583.58484815730 *T)
    sumX2  +=       0.030 * cos(5.98905764002  + 10447.63165708790 *T)
    sumX2  +=       0.032 * cos(4.08880306095  + 796.05418933290 *T)
    sumX2  +=       0.023 * cos(6.09814489767  + 9917.45305702629 *T)
    sumX2  +=       0.026 * cos(0.90493478768  + 12012.82611462390 *T)
    sumX2  +=       0.024 * cos(1.88327707702  + 14143.25142494710 *T)
    sumX2  +=       0.023 * cos(2.26727226510  + 26709.89075989690 *T)
    sumX2  +=       0.021 * cos(3.86895197800  + 7632.69944216670 *T)
    sumX2  +=       0.021 * cos(2.37138453193  + 17256.87535382490 *T)
    sumX2  +=       0.028 * cos(1.07450535792  + 12353.09642202830 *T)
    sumX2  +=       0.022 * cos(4.83463880427  + 77714.01528560399 *T)
    sumX2  +=       0.021 * cos(5.98744320172  + 6993.25271603320 *T)
    sumX2  +=       0.020 * cos(0.31265663265  + 6836.40143535029 *T)
    sumX2  +=       0.020 * cos(5.44332565028  + 1748.26023055050 *T)
    sumX2  +=       0.026 * cos(2.02675372626  + 9225.78309076650 *T)
    sumX2  +=       0.023 * cos(0.80832702606  + 149854.64395229140 *T)
    sumX2  +=       0.026 * cos(5.31358634432  + 2544.07060239989 *T)
    sumX2  +=       0.018 * cos(2.28255799994  + 11614.67711121570 *T)
    sumX2  +=       0.020 * cos(4.61413285151  + 426.84200835950 *T)
    sumX2  +=       0.019 * cos(3.66100662326  + 30640.10045611650 *T)
    sumX2  +=       0.018 * cos(4.77794649120  + 155.66421691770 *T)
    sumX2  +=       0.018 * cos(1.89582517195  + 11926.49823115230 *T)
    sumX2  +=       0.022 * cos(3.74943338657  + 17298.42614480970 *T)
    sumX2  +=       0.019 * cos(3.25511318252  + 18073.46112116669 *T)
    sumX2  +=       0.020 * cos(2.60030261119  + 19651.29229858150 *T)
    sumX2  +=       0.024 * cos(1.57388276311  + 5856.23384163189 *T)
    sumX2  +=       0.023 * cos(4.78368329384  + 72850.80553272919 *T)
    sumX2  +=       0.017 * cos(3.95703430839  + 8983.05448679250 *T)
    sumX2  +=       0.016 * cos(5.17069195690  + 16460.57734700850 *T)
    sumX2  +=       0.016 * cos(1.20099644560  + 17259.91083720689 *T)
    sumX2  +=       0.021 * cos(3.77366558760  + 16858.72635041670 *T)
    sumX2  +=       0.020 * cos(0.39427566601  + 23543.47432216530 *T)
    sumX2  +=       0.021 * cos(3.79445696120  + 13367.72881362310 *T)
    sumX2  +=       0.020 * cos(4.04662298246  + 2146.40923395870 *T)
    sumX2  +=       0.014 * cos(0.14269366851  + 22484.09239197610 *T)
    sumX2  +=       0.014 * cos(4.66600220934  + 16200.52890701769 *T)
    sumX2  +=       0.015 * cos(4.21847977142  + 20199.33877711650 *T)
    sumX2  +=       0.019 * cos(2.81420484440  + 8672.21368792411 *T)
    sumX2  +=       0.019 * cos(4.46915157584  + 18073.94875613370 *T)
    sumX2  +=       0.013 * cos(4.71429939378  + 16201.01654198470 *T)
    sumX2  +=       0.017 * cos(1.38925168648  + 7860.17557495569 *T)
    sumX2  +=       0.013 * cos(4.69427401677  + 18451.32236404950 *T)
    sumX2  +=       0.016 * cos(3.33196861521  + 5572.89898394960 *T)
    sumX2  +=       0.012 * cos(2.70989366066  + 17996.27498570570 *T)
    sumX2  +=       0.012 * cos(0.72221452149  + 19403.04077030010 *T)
    sumX2  +=       0.012 * cos(5.53894834507  + 14945.07235607090 *T)
    sumX2  +=       0.015 * cos(2.04879024946  + 12559.28197046550 *T)
    sumX2  +=       0.012 * cos(4.37879205675  + 10818.37910439930 *T)
    sumX2  +=       0.015 * cos(0.57625721169  + 12567.37583853451 *T)
    sumX2  +=       0.011 * cos(1.78471307843  + 28767.16824196750 *T)
    sumX2  +=       0.013 * cos(1.09547571668  + 90280.16698558679 *T)
    sumX2  +=       0.011 * cos(5.02230424673  + 22805.49174851010 *T)
    sumX2  +=       0.011 * cos(2.24605985689  + 35372.13108345990 *T)
    sumX2  +=       0.010 * cos(6.26009139838  + 11506.52595231009 *T)
    sumX2  +=       0.014 * cos(1.93802583909  + 21228.63584102931 *T)
    sumX2  +=       0.010 * cos(0.13064402023  + 12779.69461290430 *T)
    sumX2  +=       0.010 * cos(6.05309072952  + 23141.80220040810 *T)
    sumX2  +=       0.010 * cos(1.18002886182  + 1350.11122714230 *T)
    sumX2  +=       0.010 * cos(2.50762435685  + 10575.16286545830 *T)
    sumX2  +=       0.010 * cos(2.55643759758  + 22345.50419359170 *T)
    sumX2  +=       0.010 * cos(5.33273079625  + 9778.86485864190 *T)
    sumX2  +=       0.009 * cos(1.49151911226  + 29826.55017215670 *T)
    sumX2  +=       0.010 * cos(4.24000912367  + 23581.50199480110 *T)
    sumX2  +=       0.011 * cos(0.39610887139  + 25158.84553724890 *T)

    # EARTH X Coord
    sumX3  = 0
    sumX3  +=     232.790 * cos(3.40634928966  + 0.24381748350 *T)
    sumX3  +=      76.843 * cos(2.52439403387  + 12566.39551746630 *T)
    sumX3  +=      35.331 * cos(3.34616699853  + 6283.31966747490 *T)
    sumX3  +=       5.282 * cos(2.40489818589  + 18849.47136745770 *T)
    sumX3  +=       1.631 * cos(5.43859760597  + 6282.83203250789 *T)
    sumX3  +=       1.483 * cos(2.22678849266  + 6438.74006690910 *T)
    sumX3  +=       1.479 * cos(4.40151760084  + 6127.89926804070 *T)
    sumX3  +=       0.652 * cos(4.70876083993  + 6279.79654912590 *T)
    sumX3  +=       0.656 * cos(1.58695055624  + 6286.84278582391 *T)
    sumX3  +=       0.317 * cos(3.00237328519  + 6526.04827144891 *T)
    sumX3  +=       0.318 * cos(3.63024148359  + 6040.59106350090 *T)
    sumX3  +=       0.227 * cos(2.27961613476  + 25132.54721744910 *T)
    sumX3  +=       0.201 * cos(0.81376672334  + 6836.88907031730 *T)
    sumX3  +=       0.201 * cos(5.80642900510  + 5729.75026463250 *T)
    sumX3  +=       0.036 * cos(1.33765000112  + 12569.91863581531 *T)
    sumX3  +=       0.034 * cos(1.93992455025  + 4705.97612502710 *T)
    sumX3  +=       0.033 * cos(4.22745351005  + 12410.97511803210 *T)
    sumX3  +=       0.032 * cos(5.38348719867  + 6257.02134767510 *T)
    sumX3  +=       0.032 * cos(4.41212819179  + 6309.61798727471 *T)
    sumX3  +=       0.026 * cos(0.50334838608  + 775.76642880750 *T)
    sumX3  +=       0.023 * cos(2.01776542272  + 1059.62574767270 *T)
    sumX3  +=       0.025 * cos(1.58405990040  + 7860.66320992270 *T)
    sumX3  +=       0.027 * cos(5.17969059866  + 12565.90788249930 *T)
    sumX3  +=       0.020 * cos(3.76292921295  + 5753.62870238030 *T)
    sumX3  +=       0.019 * cos(3.01515389134  + 5885.17066406670 *T)
    sumX3  +=       0.019 * cos(6.05056350877  + 6813.01063256950 *T)
    sumX3  +=       0.018 * cos(2.29290644779  + 12721.81591690050 *T)
    sumX3  +=       0.017 * cos(0.49235818139  + 6681.46867088311 *T)
    sumX3  +=       0.013 * cos(4.49557806784  + 5487.02166065850 *T)
    sumX3  +=       0.011 * cos(5.29141894629  + 7079.61767429131 *T)
    sumX3  +=       0.010 * cos(4.46246654504  + 5507.79705615090 *T)
    sumX3  +=       0.010 * cos(3.03842434664  + 11790.87290614230 *T)
    sumX3  +=       0.009 * cos(1.50545167256  + 11507.01358727710 *T)
    sumX3  +=       0.011 * cos(6.25880735981  + 12592.69383726610 *T)
    sumX3  +=       0.009 * cos(5.39258795934  + 7058.84227879890 *T)
    sumX3  +=       0.008 * cos(1.26531983919  + 6290.43321447570 *T)
    sumX3  +=       0.008 * cos(2.25393064493  + 6276.20612047410 *T)
    sumX3  +=       0.008 * cos(3.03070423945  + 796.54182429990 *T)
    sumX3  +=       0.007 * cos(2.58386169059  + 4693.75913722409 *T)
    sumX3  +=       0.007 * cos(1.92157668691  + 7.35736448430 *T)
    sumX3  +=       0.005 * cos(4.86563371701  + 3739.00524759150 *T)
    sumX3  +=       0.005 * cos(6.21240311845  + 6070.02057203690 *T)
    sumX3  +=       0.006 * cos(3.95562930301  + 6284.29998854310 *T)
    sumX3  +=       0.006 * cos(5.83931193332  + 6282.33934640670 *T)
    sumX3  +=       0.005 * cos(3.46747665337  + 4137.15425099970 *T)
    sumX3  +=       0.005 * cos(3.58293336989  + 6496.61876291290 *T)

    # EARTH X Coord
    sumX4  = 0
    sumX4  +=     114.918 * cos(0.06053023506  + 6283.31966747490 *T)
    sumX4  +=       6.817 * cos(4.47624663983  + 0.24381748350 *T)
    sumX4  +=       3.158 * cos(0.47910545815  + 12566.39551746630 *T)
    sumX4  +=       0.211 * cos(3.72351646714  + 6438.74006690910 *T)
    sumX4  +=       0.210 * cos(2.88489921647  + 6127.89926804070 *T)
    sumX4  +=       0.253 * cos(0.64665875683  + 18849.47136745770 *T)
    sumX4  +=       0.073 * cos(3.92376796583  + 6282.83203250789 *T)
    sumX4  +=       0.042 * cos(4.60322795936  + 710.17686604180 *T)
    sumX4  +=       0.023 * cos(6.04513529243  + 6279.79654912590 *T)
    sumX4  +=       0.022 * cos(6.04140763882  + 6286.84278582391 *T)
    sumX4  +=       0.012 * cos(0.51640078868  + 25132.54721744910 *T)
    sumX4  +=       0.009 * cos(4.21008055560  + 83997.09113559539 *T)
    sumX4  +=       0.009 * cos(5.19392488383  + 11856.46246890800 *T)
    sumX4  +=       0.008 * cos(6.16587249238  + 167284.00540514899 *T)
    sumX4  +=       0.005 * cos(5.97913996910  + 6257.02134767510 *T)
    sumX4  +=       0.005 * cos(3.78878463461  + 6309.61798727471 *T)
    sumX4  +=       0.005 * cos(2.68915538708  + 12410.97511803210 *T)
    sumX4  +=       0.005 * cos(5.56850457487  + 156137.71980228281 *T)
    sumX4  +=       0.004 * cos(3.40118346072  + 12565.90788249930 *T)
    sumX4  +=       0.003 * cos(5.17221494066  + 529.93478257810 *T)

    # EARTH X Coord
    sumX5  = 0
    sumX5  +=       0.877 * cos(0.16136296068  + 6283.31966747490 *T)
    sumX5  +=       0.305 * cos(5.62973274927  + 0.24381748350 *T)
    sumX5  +=       0.101 * cos(4.88076699149  + 12566.39551746630 *T)
    sumX5  +=       0.025 * cos(5.30593958386  + 6438.74006690910 *T)
    sumX5  +=       0.025 * cos(1.27879113841  + 6127.89926804070 *T)
    sumX5  +=       0.010 * cos(5.11399286577  + 18849.47136745770 *T)
    sumX5  +=       0.003 * cos(2.56205735968  + 6282.83203250789 *T)

    # EARTH Y Coord
    sumY0  = 0
    sumY0  +=  99986069.925 * cos(0.18267413078  + 6283.31966747490 *T)
    sumY0  +=  2506324.281 * cos(3.36739796418  + 0.24381748350 *T)
    sumY0  +=  835274.807 * cos(0.13953892859  + 12566.39551746630 *T)
    sumY0  +=   10466.796 * cos(0.09642351540  + 18849.47136745770 *T)
    sumY0  +=    3490.608 * cos(6.01453435910  + 6282.83203250789 *T)
    sumY0  +=    3110.838 * cos(5.38114087369  + 83997.09113559539 *T)
    sumY0  +=    2561.408 * cos(5.29827505528  + 529.93478257810 *T)
    sumY0  +=    2142.365 * cos(2.66284107563  + 1577.09972496430 *T)
    sumY0  +=    1709.161 * cos(5.20779761275  + 6279.79654912590 *T)
    sumY0  +=    1707.934 * cos(4.58234386738  + 6286.84278582391 *T)
    sumY0  +=    1442.753 * cos(1.90130765657  + 2353.10997125530 *T)
    sumY0  +=    1113.027 * cos(5.26701283158  + 5223.45010231870 *T)
    sumY0  +=     934.484 * cos(4.50305745607  + 12036.70455237170 *T)
    sumY0  +=     899.854 * cos(1.60527831001  + 10213.52936369450 *T)
    sumY0  +=     566.820 * cos(0.58162314212  + 1059.62574767270 *T)
    sumY0  +=     661.858 * cos(6.02414120157  + 5753.62870238030 *T)
    sumY0  +=     739.644 * cos(2.79582891432  + 398.39282089170 *T)
    sumY0  +=     681.381 * cos(0.64735765260  + 4705.97612502710 *T)
    sumY0  +=     611.293 * cos(3.81390547656  + 6813.01063256950 *T)
    sumY0  +=     451.129 * cos(4.52236258525  + 5885.17066406670 *T)
    sumY0  +=     451.985 * cos(5.99169934357  + 6257.02134767510 *T)
    sumY0  +=     449.743 * cos(3.79862296384  + 6309.61798727471 *T)
    sumY0  +=     406.291 * cos(5.25608267423  + 6681.46867088311 *T)
    sumY0  +=     541.115 * cos(5.49909532338  + 775.76642880750 *T)
    sumY0  +=     546.682 * cos(6.17348361999  + 14143.73905991410 *T)
    sumY0  +=     513.780 * cos(2.86615405111  + 7860.66320992270 *T)
    sumY0  +=     263.916 * cos(3.82472119950  + 5507.79705615090 *T)
    sumY0  +=     217.220 * cos(2.94186064336  + 11507.01358727710 *T)
    sumY0  +=     292.146 * cos(4.08451989980  + 5507.30942118390 *T)
    sumY0  +=     227.915 * cos(5.95179814771  + 7058.84227879890 *T)
    sumY0  +=     226.835 * cos(4.84435977688  + 4693.75913722409 *T)
    sumY0  +=     255.833 * cos(0.69465437518  + 12168.24651405810 *T)
    sumY0  +=     256.157 * cos(6.16717762843  + 710.17686604180 *T)
    sumY0  +=     178.679 * cos(1.39565997429  + 796.54182429990 *T)
    sumY0  +=     178.303 * cos(5.11711459804  + 6283.25235717211 *T)
    sumY0  +=     178.303 * cos(4.67301162820  + 6283.38697777770 *T)
    sumY0  +=     155.471 * cos(0.05335100564  + 25132.54721744910 *T)
    sumY0  +=     155.246 * cos(6.13208007594  + 5487.02166065850 *T)
    sumY0  +=     207.640 * cos(4.28246516283  + 11790.87290614230 *T)
    sumY0  +=     199.988 * cos(2.50137197430  + 17790.08943726851 *T)
    sumY0  +=     128.970 * cos(3.64618696457  + 7079.61767429131 *T)
    sumY0  +=     128.155 * cos(3.23179068571  + 3739.00524759150 *T)
    sumY0  +=     152.241 * cos(5.58140597212  + 213.54291292150 *T)
    sumY0  +=      78.109 * cos(0.27810887092  + 8827.63408735830 *T)
    sumY0  +=      73.503 * cos(1.12438065243  + 1589.31671276730 *T)
    sumY0  +=      93.001 * cos(5.50857519732  + 9437.51911740350 *T)
    sumY0  +=      65.526 * cos(2.08488700481  + 11770.09751064990 *T)
    sumY0  +=      63.240 * cos(0.67044494917  + 6262.54427198250 *T)
    sumY0  +=      62.961 * cos(2.83649752628  + 6304.09506296730 *T)
    sumY0  +=      85.870 * cos(1.45351146305  + 167284.00540514899 *T)
    sumY0  +=      83.233 * cos(4.62701145755  + 3340.85624418330 *T)
    sumY0  +=      62.850 * cos(0.46743502764  + 7.35736448430 *T)
    sumY0  +=      79.971 * cos(0.92256149872  + 6070.02057203690 *T)
    sumY0  +=      60.043 * cos(1.82875530765  + 4137.15425099970 *T)
    sumY0  +=      75.613 * cos(2.58868557323  + 6496.61876291290 *T)
    sumY0  +=      53.731 * cos(6.27507790078  + 1194.69082770810 *T)
    sumY0  +=      57.574 * cos(2.39891957636  + 6284.29998854310 *T)
    sumY0  +=      57.574 * cos(1.10802134270  + 6282.33934640670 *T)
    sumY0  +=      72.385 * cos(0.13875968847  + 10976.83498721549 *T)
    sumY0  +=      51.135 * cos(6.02476231374  + 6290.43321447570 *T)
    sumY0  +=      50.604 * cos(3.77440441934  + 6276.20612047410 *T)
    sumY0  +=      49.613 * cos(1.07290414466  + 6127.89926804070 *T)
    sumY0  +=      48.009 * cos(2.43492252840  + 6438.74006690910 *T)
    sumY0  +=      59.233 * cos(5.35402385929  + 3154.44326741210 *T)
    sumY0  +=      48.264 * cos(0.05887245943  + 801.57711364030 *T)
    sumY0  +=      47.368 * cos(4.53616745806  + 3128.63258257930 *T)
    sumY0  +=      39.225 * cos(1.68202291816  + 8429.48508395010 *T)
    sumY0  +=      51.307 * cos(4.21369114598  + 12353.09642202830 *T)
    sumY0  +=      50.639 * cos(0.48135592944  + 11856.46246890800 *T)
    sumY0  +=      39.555 * cos(3.35943021943  + 5481.49873635110 *T)
    sumY0  +=      42.757 * cos(2.06983819322  + 17260.39847217390 *T)
    sumY0  +=      48.941 * cos(5.16594049812  + 9225.78309076650 *T)
    sumY0  +=      48.302 * cos(2.17445966751  + 2544.07060239989 *T)
    sumY0  +=      37.971 * cos(1.47013374285  + 426.84200835950 *T)
    sumY0  +=      43.643 * cos(4.71767697827  + 5856.23384163189 *T)
    sumY0  +=      31.830 * cos(3.57964760085  + 2145.92159899169 *T)
    sumY0  +=      33.959 * cos(0.15150934422  + 7085.14059859870 *T)
    sumY0  +=      30.184 * cos(2.39802729198  + 10977.32262218251 *T)
    sumY0  +=      32.562 * cos(4.41726666203  + 4164.06817212950 *T)
    sumY0  +=      38.887 * cos(0.64783670316  + 13367.72881362310 *T)
    sumY0  +=      28.515 * cos(4.53853388778  + 12569.91863581531 *T)
    sumY0  +=      28.873 * cos(5.72467865653  + 3930.45351370310 *T)
    sumY0  +=      36.633 * cos(1.33719170745  + 18073.94875613370 *T)
    sumY0  +=      24.984 * cos(3.74998709416  + 26.05450231630 *T)
    sumY0  +=      26.364 * cos(5.25860499861  + 9438.00675237050 *T)
    sumY0  +=      30.326 * cos(0.42292659979  + 4535.30325440790 *T)
    sumY0  +=      28.166 * cos(5.19366223100  + 12562.87239911730 *T)
    sumY0  +=      28.001 * cos(1.25523598395  + 5088.87265725030 *T)
    sumY0  +=      21.157 * cos(3.50176956898  + 11371.94850724170 *T)
    sumY0  +=      23.113 * cos(5.57235828208  + 5223.93773728571 *T)
    sumY0  +=      21.642 * cos(3.06444525139  + 12559.28197046550 *T)
    sumY0  +=      19.448 * cos(5.97216607301  + 12565.90788249930 *T)
    sumY0  +=      26.055 * cos(1.87419996146  + 71430.45180064559 *T)
    sumY0  +=      18.232 * cos(4.98441727809  + 1747.77259558350 *T)
    sumY0  +=      23.248 * cos(4.61930360205  + 18319.78040236310 *T)
    sumY0  +=      22.785 * cos(2.73901481639  + 5753.14106741329 *T)
    sumY0  +=      17.091 * cos(5.29823070971  + 4933.45225781610 *T)
    sumY0  +=      19.399 * cos(6.00268090156  + 10447.14402212089 *T)
    sumY0  +=      21.067 * cos(2.22953967314  + 7477.76667769950 *T)
    sumY0  +=      21.310 * cos(3.08840298987  + 8031.33608054190 *T)
    sumY0  +=      14.189 * cos(6.02227444325  + 2942.70724077510 *T)
    sumY0  +=      13.864 * cos(3.02270988640  + 8636.18582124671 *T)
    sumY0  +=      18.884 * cos(0.85614785764  + 156137.71980228281 *T)
    sumY0  +=      17.875 * cos(4.86899529635  + 1592.83983111630 *T)
    sumY0  +=      18.602 * cos(2.07103958699  + 17297.93850984270 *T)
    sumY0  +=      13.605 * cos(3.62641283985  + 13096.08648256090 *T)
    sumY0  +=      15.064 * cos(1.55068405670  + 16496.60521368590 *T)
    sumY0  +=      12.969 * cos(4.49412072551  + 7633.18707713370 *T)
    sumY0  +=      14.403 * cos(6.14964820819  + 20426.81490990550 *T)
    sumY0  +=      13.311 * cos(2.03652540541  + 12139.79732659030 *T)
    sumY0  +=      13.645 * cos(3.57902352159  + 13368.21644859010 *T)
    sumY0  +=      11.583 * cos(3.88983863974  + 5331.60126122430 *T)
    sumY0  +=      11.405 * cos(5.97095658034  + 529.44714761109 *T)
    sumY0  +=      15.057 * cos(0.50195173390  + 7084.65296363170 *T)
    sumY0  +=      14.330 * cos(4.45695507640  + 7342.70159766410 *T)
    sumY0  +=      12.375 * cos(1.53862290035  + 6279.72923882311 *T)
    sumY0  +=      12.366 * cos(1.96831814313  + 6286.91009612670 *T)
    sumY0  +=      10.016 * cos(3.81622500644  + 397.90518592470 *T)
    sumY0  +=      12.435 * cos(5.77865382585  + 15110.70993734970 *T)
    sumY0  +=       9.513 * cos(5.90294963286  + 7235.03807372550 *T)
    sumY0  +=      10.997 * cos(0.62118771904  + 10989.05197501850 *T)
    sumY0  +=      11.583 * cos(2.48171191864  + 5729.75026463250 *T)
    sumY0  +=       9.034 * cos(3.85796197835  + 9623.93209417470 *T)
    sumY0  +=      11.842 * cos(2.59877813496  + 15721.08260236190 *T)
    sumY0  +=       8.869 * cos(3.47611287212  + 6148.25458743950 *T)
    sumY0  +=       8.614 * cos(0.03081566771  + 6418.38474751031 *T)
    sumY0  +=      10.312 * cos(1.02710009735  + 6836.88907031730 *T)
    sumY0  +=       8.411 * cos(0.87289239606  + 14712.56093394150 *T)
    sumY0  +=       7.911 * cos(1.58349090213  + 2119.00767786191 *T)
    sumY0  +=      10.379 * cos(0.10632075985  + 1349.62359217529 *T)
    sumY0  +=      10.374 * cos(3.97648848184  + 5486.53402569149 *T)
    sumY0  +=      10.480 * cos(1.38209289015  + 5999.46034860970 *T)
    sumY0  +=       7.404 * cos(3.84339417190  + 6040.59106350090 *T)
    sumY0  +=       7.357 * cos(5.40136984638  + 5088.38502228330 *T)
    sumY0  +=       9.820 * cos(2.12476608355  + 6567.17898634010 *T)
    sumY0  +=       7.018 * cos(5.94659139192  + 6526.04827144891 *T)
    sumY0  +=       9.668 * cos(3.49318502073  + 21228.14820606230 *T)
    sumY0  +=       7.545 * cos(5.94843445846  + 12540.09719766650 *T)
    sumY0  +=       7.202 * cos(4.31311905021  + 6245.29199483910 *T)
    sumY0  +=       7.144 * cos(5.47703554612  + 6321.34734011070 *T)
    sumY0  +=       8.883 * cos(5.46710407546  + 5642.93474619389 *T)
    sumY0  +=       6.626 * cos(5.41362465894  + 5327.71992586630 *T)
    sumY0  +=       7.015 * cos(5.70701993379  + 7860.17557495569 *T)
    sumY0  +=       6.159 * cos(5.65831198505  + 12964.54452087450 *T)
    sumY0  +=       7.413 * cos(3.46419092567  + 23543.47432216530 *T)
    sumY0  +=       6.355 * cos(3.46461466453  + 1990.98883452450 *T)
    sumY0  +=       7.286 * cos(2.01139981548  + 537.04832957890 *T)
    sumY0  +=       6.314 * cos(1.51246730786  + 951.47458876710 *T)
    sumY0  +=       6.257 * cos(2.66128485006  + 4690.72365384210 *T)
    sumY0  +=       6.484 * cos(1.63733658639  + 16730.70750707931 *T)
    sumY0  +=       5.803 * cos(1.38237252918  + 955.35592412510 *T)
    sumY0  +=       6.391 * cos(2.40521894071  + 24073.16528725990 *T)
    sumY0  +=       7.530 * cos(4.91537693059  + 10973.79950383350 *T)
    sumY0  +=       6.921 * cos(5.41208599854  + 522.82123557730 *T)
    sumY0  +=       7.626 * cos(2.72257180334  + 22004.15845235330 *T)
    sumY0  +=       7.531 * cos(1.94384811400  + 18422.87317658170 *T)
    sumY0  +=       6.902 * cos(4.32614260011  + 155.17658195069 *T)
    sumY0  +=       5.434 * cos(4.37546348368  + 7238.91940908350 *T)
    sumY0  +=       6.223 * cos(0.67316616422  + 18451.32236404950 *T)
    sumY0  +=       7.296 * cos(1.27743142232  + 16730.21987211229 *T)
    sumY0  +=       5.833 * cos(0.75851095819  + 640.14110379750 *T)
    sumY0  +=       7.171 * cos(1.43520696239  + 3929.96587873609 *T)
    sumY0  +=       6.251 * cos(0.62294633745  + 6277.79674316750 *T)
    sumY0  +=       6.251 * cos(2.88399458161  + 6288.84259178230 *T)
    sumY0  +=       5.683 * cos(1.92473239272  + 1551.28904013150 *T)
    sumY0  +=       6.383 * cos(2.21747129905  + 5216.33655531789 *T)
    sumY0  +=       6.194 * cos(5.13602240936  + 5230.56364931950 *T)
    sumY0  +=       5.036 * cos(2.26614918626  + 14314.41193053330 *T)
    sumY0  +=       6.774 * cos(3.56437941703  + 102.84895673509 *T)
    sumY0  +=       6.224 * cos(1.92311046214  + 11014.86265985130 *T)
    sumY0  +=       4.621 * cos(2.91747769543  + 553.32558535889 *T)
    sumY0  +=       6.293 * cos(0.74955598933  + 5650.53592816170 *T)
    sumY0  +=       6.297 * cos(2.83090961174  + 26088.14695905770 *T)
    sumY0  +=       5.186 * cos(5.42426212805  + 77714.01528560399 *T)
    sumY0  +=       6.196 * cos(4.15785490016  + 84672.23202702119 *T)
    sumY0  +=       6.160 * cos(3.21170376619  + 239424.63407183639 *T)
    sumY0  +=       5.414 * cos(1.42882325200  + 6180.22689325630 *T)
    sumY0  +=       5.413 * cos(2.08335586272  + 6386.41244169350 *T)
    sumY0  +=       4.863 * cos(2.19642727179  + 90280.16698558679 *T)
    sumY0  +=       5.711 * cos(2.75632617612  + 6916.10340678810 *T)
    sumY0  +=       5.150 * cos(2.53503851142  + 1577.58735993130 *T)
    sumY0  +=       4.269 * cos(0.81279132183  + 7875.91568110771 *T)
    sumY0  +=       4.017 * cos(2.21501270421  + 6254.87048000710 *T)
    sumY0  +=       4.069 * cos(3.99136293444  + 3634.37720703489 *T)
    sumY0  +=       3.993 * cos(1.29192496097  + 6311.76885494270 *T)
    sumY0  +=       4.016 * cos(0.52802181166  + 4690.23601887509 *T)
    sumY0  +=       5.290 * cos(4.91562568689  + 25158.35790228190 *T)
    sumY0  +=       4.831 * cos(1.26836887601  + 5760.74224938110 *T)
    sumY0  +=       3.487 * cos(1.93988497026  + 10447.63165708790 *T)
    sumY0  +=       4.120 * cos(1.72576404553  + 6709.91785835091 *T)
    sumY0  +=       4.410 * cos(2.23510306717  + 6805.89708556871 *T)
    sumY0  +=       4.023 * cos(1.51236080833  + 4731.78680985990 *T)
    sumY0  +=       3.980 * cos(1.74852343480  + 5856.72147659890 *T)
    sumY0  +=       3.173 * cos(3.62048518613  + 1066.73929467350 *T)
    sumY0  +=       3.293 * cos(5.58717113537  + 9917.45305702629 *T)
    sumY0  +=       2.935 * cos(2.10511347366  + 12592.69383726610 *T)
    sumY0  +=       2.979 * cos(4.62987940723  + 12566.46282776910 *T)
    sumY0  +=       3.980 * cos(5.90920652207  + 13341.91812879030 *T)
    sumY0  +=       3.892 * cos(3.95491793455  + 75.02541605080 *T)
    sumY0  +=       3.042 * cos(1.62346061120  + 18053.17336064130 *T)
    sumY0  +=       3.608 * cos(4.03873145467  + 6208.53806890760 *T)
    sumY0  +=       3.721 * cos(3.65024914006  + 5966.92779781830 *T)
    sumY0  +=       3.581 * cos(5.75234129258  + 6358.10126604220 *T)
    sumY0  +=       2.880 * cos(0.04145923446  + 10575.65050042530 *T)
    sumY0  +=       2.806 * cos(2.64471981541  + 5863.83502359970 *T)
    sumY0  +=       3.626 * cos(6.13871061291  + 6599.71153713150 *T)
    sumY0  +=       2.794 * cos(0.98030293667  + 8030.84844557490 *T)
    sumY0  +=       2.729 * cos(2.60239924444  + 3.76693583251 *T)
    sumY0  +=       3.625 * cos(1.43112681156  + 6020.23574410210 *T)
    sumY0  +=       3.206 * cos(3.72476724430  + 11506.52595231009 *T)
    sumY0  +=       3.556 * cos(5.86027209236  + 8428.99744898310 *T)
    sumY0  +=       2.856 * cos(1.76545510884  + 12721.81591690050 *T)
    sumY0  +=       2.611 * cos(0.86182235214  + 6702.80431135010 *T)
    sumY0  +=       2.535 * cos(0.01231069705  + 31415.62306744050 *T)
    sumY0  +=       3.520 * cos(3.80906738831  + 250570.91967470257 *T)
    sumY0  +=       3.434 * cos(2.07577291101  + 6546.40359084770 *T)
    sumY0  +=       2.572 * cos(0.08592522543  + 5884.68302909969 *T)
    sumY0  +=       2.647 * cos(5.39051905541  + 2352.62233628830 *T)
    sumY0  +=       3.026 * cos(3.66998442300  + 13916.26292712510 *T)
    sumY0  +=       2.378 * cos(2.05754080989  + 2389.13783793270 *T)
    sumY0  +=       2.979 * cos(5.07398237707  + 12566.32820716351 *T)
    sumY0  +=       3.082 * cos(3.34593854945  + 14945.07235607090 *T)
    sumY0  +=       2.716 * cos(1.49855756532  + 12029.59100537090 *T)
    sumY0  +=       2.215 * cos(3.69564106529  + 13362.69352428270 *T)
    sumY0  +=       3.032 * cos(0.05314931388  + 29088.56759850150 *T)
    sumY0  +=       2.421 * cos(1.58246988642  + 11015.35029481830 *T)
    sumY0  +=       2.275 * cos(4.38385659579  + 6077.13411903770 *T)
    sumY0  +=       2.466 * cos(0.87928045349  + 16200.52890701769 *T)
    sumY0  +=       2.529 * cos(4.37629498907  + 12043.81809937250 *T)
    sumY0  +=       2.015 * cos(0.91352607152  + 6262.96434807611 *T)
    sumY0  +=       2.218 * cos(5.40562909521  + 6489.50521591210 *T)
    sumY0  +=       2.005 * cos(2.59344260095  + 6303.67498687370 *T)
    sumY0  +=       2.010 * cos(5.68334625182  + 4700.87268422489 *T)
    sumY0  +=       2.641 * cos(5.85354681494  + 4590.66636300550 *T)
    sumY0  +=       2.438 * cos(4.55830572881  + 149.80701461810 *T)
    sumY0  +=       1.893 * cos(5.69296362938  + 6279.43833211690 *T)
    sumY0  +=       1.893 * cos(4.09716259686  + 6287.20100283290 *T)
    sumY0  +=       2.156 * cos(0.43822603611  + 18139.53831889940 *T)
    sumY0  +=       2.071 * cos(2.75099511730  + 11926.49823115230 *T)
    sumY0  +=       2.395 * cos(4.84387255380  + 1162.71852189130 *T)
    sumY0  +=       1.954 * cos(5.08514679694  + 13518.11392371690 *T)
    sumY0  +=       2.006 * cos(5.02287213343  + 3154.93090237910 *T)
    sumY0  +=       1.776 * cos(4.86094102925  + 13521.99525907490 *T)
    sumY0  +=       1.964 * cos(1.82182786120  + 4686.64559022330 *T)
    sumY0  +=       2.332 * cos(2.83569470692  + 10022.08109758290 *T)
    sumY0  +=       1.863 * cos(3.68950976920  + 242.97242145751 *T)
    sumY0  +=       1.928 * cos(3.02170818288  + 5746.51515537950 *T)
    sumY0  +=       2.436 * cos(2.83691190801  + 95143.37673846159 *T)
    sumY0  +=       2.163 * cos(6.13036805408  + 6037.48802124550 *T)
    sumY0  +=       2.250 * cos(0.82649289984  + 12669.48829168490 *T)
    sumY0  +=       2.052 * cos(3.65926976048  + 6529.15131370430 *T)
    sumY0  +=       1.886 * cos(5.18293217320  + 7238.43177411650 *T)
    sumY0  +=       2.170 * cos(3.92516942708  + 19651.29229858150 *T)
    sumY0  +=       2.058 * cos(3.12741900385  + 23013.78335707070 *T)
    sumY0  +=       1.774 * cos(0.51402656711  + 6820.12417957030 *T)
    sumY0  +=       1.841 * cos(1.91732926338  + 65147.37595065419 *T)
    sumY0  +=       1.578 * cos(4.07231752329  + 4292.57465043390 *T)
    sumY0  +=       1.577 * cos(2.37859191870  + 7632.69944216670 *T)
    sumY0  +=       1.541 * cos(0.23280184119  + 13119.96492030870 *T)
    sumY0  +=       2.012 * cos(1.94074945567  + 4292.08701546690 *T)
    sumY0  +=       1.543 * cos(5.39993967465  + 6293.95633282471 *T)
    sumY0  +=       1.543 * cos(4.39018655159  + 6272.68300212510 *T)
    sumY0  +=       1.513 * cos(2.22141523192  + 20.53157800890 *T)
    sumY0  +=       1.489 * cos(1.38125902411  + 110.45013870291 *T)
    sumY0  +=       1.639 * cos(2.86996469977  + 17655.02435723310 *T)
    sumY0  +=       1.443 * cos(0.69249676813  + 1052.51220067191 *T)
    sumY0  +=       1.337 * cos(1.49321921272  + 2544.55823736690 *T)
    sumY0  +=       1.797 * cos(1.47415887460  + 33018.77729472110 *T)
    sumY0  +=       1.488 * cos(4.10370894807  + 76.50988875911 *T)
    sumY0  +=       1.453 * cos(3.88999021921  + 633.02755679670 *T)
    sumY0  +=       1.349 * cos(4.18018116919  + 6016.71262575310 *T)
    sumY0  +=       1.477 * cos(2.56925141669  + 12779.69461290430 *T)
    sumY0  +=       1.688 * cos(4.77167367345  + 18875.28205229050 *T)
    sumY0  +=       1.491 * cos(4.65894552032  + 9411.70843257070 *T)
    sumY0  +=       1.471 * cos(4.18699607739  + 18636.17227201970 *T)
    sumY0  +=       1.263 * cos(5.62967869388  + 12573.50906446710 *T)
    sumY0  +=       1.685 * cos(4.13797957160  + 25934.36814857290 *T)
    sumY0  +=       1.375 * cos(2.66006793362  + 2378.92065608810 *T)
    sumY0  +=       1.641 * cos(1.56425462469  + 6.86972951729 *T)
    sumY0  +=       1.200 * cos(6.18795172265  + 11499.90004027630 *T)
    sumY0  +=       1.272 * cos(5.60977811105  + 6549.92670919670 *T)
    sumY0  +=       1.214 * cos(2.79740025979  + 10973.31186886650 *T)
    sumY0  +=       1.406 * cos(0.96357897619  + 11790.38527117530 *T)
    sumY0  +=       1.524 * cos(2.69987154893  + 18209.57408114370 *T)
    sumY0  +=       1.149 * cos(1.44882073480  + 10177.50149701710 *T)
    sumY0  +=       1.547 * cos(5.59995461433  + 11926.01059618529 *T)
    sumY0  +=       1.460 * cos(3.40155505101  + 246.07546371290 *T)
    sumY0  +=       1.323 * cos(2.95651083102  + 8661.99650607950 *T)
    sumY0  +=       1.252 * cos(6.12405778157  + 6993.25271603320 *T)
    sumY0  +=       1.459 * cos(3.72728247655  + 38.37685312130 *T)
    sumY0  +=       1.235 * cos(1.27016887136  + 24357.02460612510 *T)
    sumY0  +=       1.028 * cos(0.93950364612  + 6112.64679685570 *T)
    sumY0  +=       1.330 * cos(5.71304625366  + 2146.40923395870 *T)
    sumY0  +=       1.312 * cos(5.81330482072  + 12491.61391889900 *T)
    sumY0  +=       1.246 * cos(6.23510174757  + 5429.63565075589 *T)
    sumY0  +=       0.992 * cos(2.56738111789  + 6453.99253809410 *T)
    sumY0  +=       0.988 * cos(5.69208485871  + 8274.06468451590 *T)
    sumY0  +=       1.365 * cos(1.39907964627  + 11371.46087227470 *T)
    sumY0  +=       1.020 * cos(1.25240185088  + 4732.27444482690 *T)
    sumY0  +=       1.191 * cos(1.85928407242  + 6290.36590417291 *T)
    sumY0  +=       1.191 * cos(1.64765684663  + 6276.27343077690 *T)
    sumY0  +=       0.978 * cos(3.16773125497  + 6247.29180079750 *T)
    sumY0  +=       0.971 * cos(0.33921091467  + 6319.34753415231 *T)
    sumY0  +=       0.962 * cos(1.06488912173  + 12565.41519639810 *T)
    sumY0  +=       1.025 * cos(0.64688768472  + 12545.62012197390 *T)
    sumY0  +=       0.990 * cos(5.10278829353  + 4694.24677219110 *T)
    sumY0  +=       0.964 * cos(3.36272372610  + 3893.93801205869 *T)
    sumY0  +=       0.944 * cos(4.29299913988  + 6259.44122972711 *T)
    sumY0  +=       0.944 * cos(5.49712708636  + 6307.19810522270 *T)
    sumY0  +=       1.092 * cos(2.34985569802  + 17797.20298426930 *T)
    sumY0  +=       1.132 * cos(5.71867421008  + 17782.97589026770 *T)
    sumY0  +=       0.859 * cos(6.19348487479  + 1692.40948698591 *T)
    sumY0  +=       1.213 * cos(5.61315622766  + 82576.73740351178 *T)
    sumY0  +=       0.942 * cos(3.65711104430  + 6298.57213865991 *T)
    sumY0  +=       0.942 * cos(6.13301518193  + 6268.06719628990 *T)
    sumY0  +=       1.045 * cos(5.17434482333  + 15508.85894075790 *T)
    sumY0  +=       0.851 * cos(3.62598735241  + 6173.11334625550 *T)
    sumY0  +=       0.850 * cos(6.16443178844  + 6393.52598869430 *T)
    sumY0  +=       0.895 * cos(0.83009579073  + 3904.15519390330 *T)
    sumY0  +=       1.020 * cos(5.42778314413  + 220.65645992230 *T)
    sumY0  +=       0.845 * cos(4.18420138552  + 17256.87535382490 *T)
    sumY0  +=       0.863 * cos(2.86079142045  + 949.41942645330 *T)
    sumY0  +=       0.853 * cos(1.18534704592  + 16201.01654198470 *T)
    sumY0  +=       0.811 * cos(1.42220880644  + 4803.96545844350 *T)
    sumY0  +=       0.966 * cos(2.57517734766  + 206.42936592071 *T)
    sumY0  +=       0.917 * cos(0.89927696450  + 149854.64395229140 *T)
    sumY0  +=       1.083 * cos(2.89594427545  + 36948.98699094070 *T)
    sumY0  +=       0.779 * cos(2.01260050316  + 2648.69864295650 *T)
    sumY0  +=       0.964 * cos(2.34511297941  + 796.05418933290 *T)
    sumY0  +=       0.995 * cos(4.96662237412  + 11403.92081305850 *T)
    sumY0  +=       0.962 * cos(2.35578735539  + 12567.37583853451 *T)
    sumY0  +=       0.779 * cos(2.02763029837  + 10213.04172872750 *T)
    sumY0  +=       0.980 * cos(6.19670011124  + 22805.49174851010 *T)
    sumY0  +=       0.930 * cos(0.65237052253  + 2787.28684134090 *T)
    sumY0  +=       0.829 * cos(0.38404899643  + 5120.35732810009 *T)
    sumY0  +=       0.733 * cos(4.18029489609  + 10575.16286545830 *T)
    sumY0  +=       0.743 * cos(2.22826365578  + 7834.36489012290 *T)
    sumY0  +=       0.818 * cos(3.21447961169  + 5572.89898394960 *T)
    sumY0  +=       0.905 * cos(2.24459183660  + 6284.80414018320 *T)
    sumY0  +=       0.905 * cos(1.26234908246  + 6281.83519476660 *T)
    sumY0  +=       0.798 * cos(1.05421419226  + 12410.97511803210 *T)
    sumY0  +=       0.817 * cos(5.23322400864  + 12416.83232033170 *T)
    sumY0  +=       0.884 * cos(2.41399532496  + 22483.60475700909 *T)
    sumY0  +=       0.755 * cos(5.40020688961  + 4060.97539791089 *T)
    sumY0  +=       0.651 * cos(5.50763453832  + 17259.91083720689 *T)
    sumY0  +=       0.616 * cos(4.13394819140  + 1596.43025976811 *T)
    sumY0  +=       0.651 * cos(0.84412253043  + 1748.26023055050 *T)
    sumY0  +=       0.723 * cos(1.49663741427  + 161000.92955515761 *T)
    sumY0  +=       0.784 * cos(2.44837693578  + 4907.05823266209 *T)
    sumY0  +=       0.784 * cos(3.77248887935  + 7234.55043875850 *T)
    sumY0  +=       0.785 * cos(1.86574625348  + 846.32665223470 *T)
    sumY0  +=       0.611 * cos(6.24391260854  + 853.44019923550 *T)
    sumY0  +=       0.576 * cos(5.94037158869  + 12323.66691349230 *T)
    sumY0  +=       0.570 * cos(2.97586328906  + 12587.17091295870 *T)
    sumY0  +=       0.576 * cos(4.63987761174  + 13915.77529215809 *T)
    sumY0  +=       0.770 * cos(0.89072554788  + 6069.53293706989 *T)
    sumY0  +=       0.773 * cos(1.78442906238  + 6133.26883537330 *T)
    sumY0  +=       0.581 * cos(6.05264641509  + 11933.61177815310 *T)
    sumY0  +=       0.719 * cos(2.47920208822  + 15720.59496739490 *T)
    sumY0  +=       0.578 * cos(2.69917567595  + 8662.48414104651 *T)
    sumY0  +=       0.536 * cos(4.49580958272  + 18852.99448580670 *T)
    sumY0  +=       0.633 * cos(5.88497258324  + 52176.05010063190 *T)
    sumY0  +=       0.581 * cos(1.35719256610  + 5334.14405850510 *T)
    sumY0  +=       0.532 * cos(3.01682546062  + 18842.35782045690 *T)
    sumY0  +=       0.657 * cos(4.82951502701  + 5849.12029463110 *T)
    sumY0  +=       0.599 * cos(0.54713597581  + 6151.77770578850 *T)
    sumY0  +=       0.540 * cos(1.67333017022  + 6286.60602489270 *T)
    sumY0  +=       0.540 * cos(1.83361074884  + 6280.03331005710 *T)
    sumY0  +=       0.628 * cos(5.02906961003  + 17298.42614480970 *T)
    sumY0  +=       0.532 * cos(2.84025435717  + 11514.12713427790 *T)
    sumY0  +=       0.513 * cos(2.13211972766  + 12456.18919624690 *T)
    sumY0  +=       0.649 * cos(3.33488996078  + 11764.57458634250 *T)
    sumY0  +=       0.582 * cos(2.96012797214  + 6414.86162916130 *T)
    sumY0  +=       0.517 * cos(5.63213604686  + 3340.36860921629 *T)
    sumY0  +=       0.609 * cos(2.37414673548  + 10420.23010099111 *T)
    sumY0  +=       0.615 * cos(0.00032066291  + 10983.94853421629 *T)
    sumY0  +=       0.622 * cos(0.82619343851  + 5326.54287653730 *T)
    sumY0  +=       0.536 * cos(2.13145334966  + 7232.49527644471 *T)
    sumY0  +=       0.516 * cos(1.35907342509  + 433.95555536030 *T)
    sumY0  +=       0.613 * cos(3.36175443914  + 10969.72144021469 *T)
    sumY0  +=       0.583 * cos(1.44072369133  + 5863.34738863270 *T)
    sumY0  +=       0.636 * cos(1.34803641207  + 26735.70144472970 *T)
    sumY0  +=       0.665 * cos(4.31762068350  + 40879.19668716030 *T)
    sumY0  +=       0.583 * cos(4.39404867767  + 12592.20620229910 *T)
    sumY0  +=       0.479 * cos(6.04787718072  + 5547.44315394310 *T)
    sumY0  +=       0.571 * cos(1.01593938825  + 6062.90702503610 *T)
    sumY0  +=       0.565 * cos(5.49171831457  + 4171.18171913030 *T)
    sumY0  +=       0.462 * cos(3.53710274271  + 3104.68624194030 *T)
    sumY0  +=       0.551 * cos(2.46794798020  + 6503.73230991370 *T)
    sumY0  +=       0.504 * cos(0.43308443375  + 15670.83794192309 *T)
    sumY0  +=       0.535 * cos(4.55197610339  + 173567.08125514039 *T)
    sumY0  +=       0.460 * cos(4.75481082956  + 3495.78900865049 *T)
    sumY0  +=       0.449 * cos(0.38594751092  + 4274.27449334889 *T)
    sumY0  +=       0.453 * cos(5.14645638415  + 9387.76209193169 *T)
    sumY0  +=       0.527 * cos(4.67747505318  + 24602.85625235450 *T)
    sumY0  +=       0.502 * cos(5.68646114706  + 12490.12944619070 *T)
    sumY0  +=       0.603 * cos(5.56725967880  + 322711.54834138998 *T)
    sumY0  +=       0.532 * cos(0.26579545469  + 5120.84496306710 *T)
    sumY0  +=       0.530 * cos(5.14730101928  + 18845.94824910870 *T)
    sumY0  +=       0.431 * cos(3.74066314880  + 7019.19618100671 *T)
    sumY0  +=       0.465 * cos(1.12060083942  + 8827.14645239129 *T)
    sumY0  +=       0.559 * cos(0.14481247330  + 1582.20316576650 *T)
    sumY0  +=       0.485 * cos(4.56530394374  + 29296.85920706210 *T)
    sumY0  +=       0.584 * cos(4.78377689692  + 72850.80553272919 *T)
    sumY0  +=       0.550 * cos(5.52640225498  + 213.05527795450 *T)
    sumY0  +=       0.481 * cos(0.29866487122  + 14.47091148511 *T)
    sumY0  +=       0.569 * cos(1.38161051216  + 97238.87136197100 *T)
    sumY0  +=       0.551 * cos(3.21618763564  + 14313.92429556630 *T)
    sumY0  +=       0.389 * cos(1.85967717576  + 6245.18663183710 *T)
    sumY0  +=       0.389 * cos(1.64726374330  + 6321.45270311270 *T)
    sumY0  +=       0.388 * cos(5.85202122263  + 6297.54676147650 *T)
    sumY0  +=       0.388 * cos(3.93810500361  + 6269.09257347330 *T)
    sumY0  +=       0.515 * cos(0.09995632139  + 12320.56387123691 *T)
    sumY0  +=       0.394 * cos(1.39198903052  + 4156.95462512869 *T)
    sumY0  +=       0.491 * cos(4.39709251394  + 1479.11039154791 *T)
    sumY0  +=       0.440 * cos(2.54396321518  + 5650.04829319470 *T)
    sumY0  +=       0.382 * cos(1.50580912176  + 9917.94069199330 *T)
    sumY0  +=       0.503 * cos(3.08687612414  + 17157.30569795530 *T)
    sumY0  +=       0.439 * cos(3.25483236471  + 233141.55822184496 *T)
    sumY0  +=       0.381 * cos(3.59564013891  + 14143.25142494710 *T)
    sumY0  +=       0.371 * cos(3.01987537874  + 5643.42238116090 *T)
    sumY0  +=       0.516 * cos(4.04344774561  + 135.30889751891 *T)
    sumY0  +=       0.417 * cos(2.17381926869  + 13760.84252769090 *T)
    sumY0  +=       0.483 * cos(2.84671280305  + 9779.35249360890 *T)
    sumY0  +=       0.381 * cos(2.79871486639  + 14919.26167123810 *T)
    sumY0  +=       0.378 * cos(2.94872717012  + 17267.51201917470 *T)
    sumY0  +=       0.393 * cos(6.27917510634  + 7872.39256275871 *T)
    sumY0  +=       0.376 * cos(6.02660735301  + 13517.62628874989 *T)
    sumY0  +=       0.365 * cos(0.53940975175  + 6923.21695378890 *T)
    sumY0  +=       0.371 * cos(4.19154046767  + 13625.77744765550 *T)
    sumY0  +=       0.355 * cos(3.67238322579  + 10874.22984796390 *T)
    sumY0  +=       0.499 * cos(1.15482974306  + 161710.86260371591 *T)
    sumY0  +=       0.379 * cos(5.52860495919  + 3185.43584474911 *T)
    sumY0  +=       0.471 * cos(0.07055813902  + 11712.71150074729 *T)
    sumY0  +=       0.417 * cos(1.75993959010  + 22779.68106367730 *T)
    sumY0  +=       0.485 * cos(6.06683096360  + 12528.26248182851 *T)
    sumY0  +=       0.358 * cos(1.92035724848  + 6295.04902031090 *T)
    sumY0  +=       0.358 * cos(1.58658367058  + 6271.59031463890 *T)
    sumY0  +=       0.365 * cos(5.19130080276  + 6836.40143535029 *T)
    sumY0  +=       0.351 * cos(0.64808198517  + 11617.21990849651 *T)
    sumY0  +=       0.351 * cos(4.53436303138  + 205.94173095370 *T)
    sumY0  +=       0.429 * cos(5.51029115203  + 3894.42564702570 *T)
    sumY0  +=       0.388 * cos(5.20318951030  + 956.53297345411 *T)
    sumY0  +=       0.464 * cos(5.40703775457  + 23581.50199480110 *T)
    sumY0  +=       0.396 * cos(0.78828729228  + 5231.05128428650 *T)
    sumY0  +=       0.421 * cos(3.25659811640  + 7445.79437188270 *T)
    sumY0  +=       0.366 * cos(5.33851384127  + 17253.28492517310 *T)
    sumY0  +=       0.360 * cos(5.53135776054  + 21393.78578734110 *T)
    sumY0  +=       0.336 * cos(6.12569565789  + 6279.38751421180 *T)
    sumY0  +=       0.336 * cos(3.66443056835  + 6287.25182073800 *T)
    sumY0  +=       0.326 * cos(0.22542241669  + 1059.13811270570 *T)
    sumY0  +=       0.395 * cos(5.25182737341  + 5642.44206009270 *T)
    sumY0  +=       0.450 * cos(1.80229899871  + 1385.31775746760 *T)
    sumY0  +=       0.363 * cos(2.72535712612  + 22484.09239197610 *T)
    sumY0  +=       0.399 * cos(5.51787738042  + 16858.72635041670 *T)
    sumY0  +=       0.316 * cos(2.45995539098  + 20995.63678393290 *T)
    sumY0  +=       0.430 * cos(3.96101117316  + 19650.80466361450 *T)
    sumY0  +=       0.363 * cos(2.71990145670  + 7335.58805066330 *T)
    sumY0  +=       0.317 * cos(3.09968934621  + 11769.60987568290 *T)
    sumY0  +=       0.395 * cos(3.82348351794  + 5905.94605955911 *T)
    sumY0  +=       0.324 * cos(5.27854205827  + 37.78385515230 *T)
    sumY0  +=       0.299 * cos(2.29226668471  + 641.12142486571 *T)
    sumY0  +=       0.304 * cos(4.87224846445  + 5750.10558403130 *T)
    sumY0  +=       0.298 * cos(2.25684047934  + 1350.11122714230 *T)
    sumY0  +=       0.414 * cos(5.74970236841  + 44809.40638337990 *T)
    sumY0  +=       0.339 * cos(5.86256617398  + 3.68833577960 *T)
    sumY0  +=       0.351 * cos(4.36315861783  + 12345.98287502750 *T)
    sumY0  +=       0.362 * cos(1.99294614713  + 21953.91379191449 *T)
    sumY0  +=       0.304 * cos(3.36735752668  + 29826.55017215670 *T)
    sumY0  +=       0.395 * cos(4.78960869869  + 4176.28515993250 *T)
    sumY0  +=       0.320 * cos(0.92821593891  + 10818.37910439930 *T)
    sumY0  +=       0.384 * cos(5.58951901360  + 10177.01386205009 *T)
    sumY0  +=       0.279 * cos(5.67648753585  + 10970.20907518170 *T)
    sumY0  +=       0.365 * cos(5.96664259426  + 6660.69327539070 *T)
    sumY0  +=       0.377 * cos(5.53753804163  + 29864.57784479250 *T)
    sumY0  +=       0.275 * cos(4.16427905835  + 20597.48778052470 *T)
    sumY0  +=       0.316 * cos(1.02608510804  + 316.63568714010 *T)
    sumY0  +=       0.339 * cos(4.53749399958  + 6924.19727485710 *T)
    sumY0  +=       0.386 * cos(0.71953202985  + 2636.96929012050 *T)
    sumY0  +=       0.273 * cos(5.80758206821  + 26709.89075989690 *T)
    sumY0  +=       0.314 * cos(6.18462887647  + 14945.55999103790 *T)
    sumY0  +=       0.281 * cos(0.15414528977  + 16858.23871544969 *T)
    sumY0  +=       0.287 * cos(4.87229910240  + 18073.46112116669 *T)
    sumY0  +=       0.265 * cos(3.53930677592  + 19379.16233255230 *T)
    sumY0  +=       0.325 * cos(0.92198404013  + 12360.20996902910 *T)
    sumY0  +=       0.362 * cos(2.75236634018  + 30665.91114094930 *T)
    sumY0  +=       0.269 * cos(4.91776126981  + 6816.53375091851 *T)
    sumY0  +=       0.344 * cos(1.94480091531  + 6147.69434246491 *T)
    sumY0  +=       0.255 * cos(3.37543976329  + 1376.01761732930 *T)
    sumY0  +=       0.334 * cos(1.56223491581  + 6418.94499248490 *T)
    sumY0  +=       0.316 * cos(3.97125698165  + 6055.84353468590 *T)
    sumY0  +=       0.283 * cos(2.66015602247  + 28287.23430234470 *T)
    sumY0  +=       0.305 * cos(5.83332121064  + 16522.41589851870 *T)
    sumY0  +=       0.270 * cos(1.62295499850  + 283.61550138170 *T)
    sumY0  +=       0.243 * cos(0.97600841060  + 6255.91811137810 *T)
    sumY0  +=       0.243 * cos(2.53093250846  + 6310.72122357170 *T)
    sumY0  +=       0.296 * cos(4.79138408835  + 6129.54085699010 *T)
    sumY0  +=       0.301 * cos(5.81886935980  + 6510.79580026390 *T)
    sumY0  +=       0.329 * cos(5.86922121940  + 377.61742539930 *T)
    sumY0  +=       0.267 * cos(1.84931872641  + 24705.94902657310 *T)
    sumY0  +=       0.285 * cos(0.93414937692  + 5469.76938351510 *T)
    sumY0  +=       0.286 * cos(4.99886262730  + 6437.09847795970 *T)
    sumY0  +=       0.311 * cos(1.63984256457  + 11720.31268271510 *T)
    sumY0  +=       0.319 * cos(3.13980275142  + 169379.50002865840 *T)
    sumY0  +=       0.241 * cos(4.56518963450  + 632.53992182970 *T)
    sumY0  +=       0.229 * cos(5.72761357577  + 1265.81129610991 *T)
    sumY0  +=       0.235 * cos(4.94058515466  + 4487.57358878689 *T)
    sumY0  +=       0.253 * cos(2.08435720561  + 4377.36726756750 *T)
    sumY0  +=       0.277 * cos(3.15798901317  + 419.24082639170 *T)
    sumY0  +=       0.234 * cos(3.55494952630  + 11713.19913571430 *T)
    sumY0  +=       0.216 * cos(0.30207043900  + 10454.25756912169 *T)
    sumY0  +=       0.241 * cos(3.87591164077  + 103.33659170210 *T)
    sumY0  +=       0.214 * cos(0.52402082918  + 2222.10045208050 *T)
    sumY0  +=       0.213 * cos(2.22358750694  + 30356.24113725130 *T)
    sumY0  +=       0.240 * cos(4.14485134121  + 6309.13035230770 *T)
    sumY0  +=       0.212 * cos(3.93716983669  + 262.84010588929 *T)
    sumY0  +=       0.259 * cos(1.64646341991  + 6283.56348495841 *T)
    sumY0  +=       0.259 * cos(1.86047749915  + 6283.07584999140 *T)
    sumY0  +=       0.222 * cos(2.99868754607  + 10440.03047512009 *T)
    sumY0  +=       0.201 * cos(0.39862703515  + 5746.02752041250 *T)
    sumY0  +=       0.274 * cos(5.42466894469  + 23581.01435983410 *T)
    sumY0  +=       0.243 * cos(2.57266886723  + 7096.86995143470 *T)
    sumY0  +=       0.234 * cos(3.66617634256  + 5573.38661891660 *T)
    sumY0  +=       0.203 * cos(1.57613060854  + 16460.08971204149 *T)
    sumY0  +=       0.246 * cos(4.26574746801  + 8672.21368792411 *T)
    sumY0  +=       0.216 * cos(3.35974562754  + 5437.23683272371 *T)
    sumY0  +=       0.210 * cos(4.25419338790  + 9381.20349020070 *T)
    sumY0  +=       0.199 * cos(5.64767313384  + 11216.52810780750 *T)
    sumY0  +=       0.270 * cos(5.11258425974  + 284.10313634870 *T)
    sumY0  +=       0.207 * cos(1.49549061720  + 12562.80508881451 *T)
    sumY0  +=       0.212 * cos(0.18391974040  + 7129.40250222610 *T)
    sumY0  +=       0.269 * cos(6.23907858108  + 70755.31090921978 *T)
    sumY0  +=       0.221 * cos(4.97264401025  + 77713.52765063698 *T)
    sumY0  +=       0.225 * cos(5.58708477402  + 5635.82119919310 *T)
    sumY0  +=       0.197 * cos(4.93201648946  + 14712.07329897449 *T)
    sumY0  +=       0.192 * cos(0.47501431186  + 17272.12782500990 *T)
    sumY0  +=       0.241 * cos(3.52343193321  + 15907.00794416610 *T)
    sumY0  +=       0.261 * cos(0.88974137342  + 48739.61607959950 *T)
    sumY0  +=       0.255 * cos(2.10230203862  + 6206.56596123230 *T)
    sumY0  +=       0.256 * cos(0.73069007306  + 224.58861318540 *T)
    sumY0  +=       0.182 * cos(5.98997615944  + 18848.98373249069 *T)
    sumY0  +=       0.250 * cos(2.67103598665  + 5934.39524702691 *T)
    sumY0  +=       0.196 * cos(0.64831634218  + 16460.57734700850 *T)
    sumY0  +=       0.187 * cos(0.04522363352  + 22003.67081738630 *T)
    sumY0  +=       0.229 * cos(1.73812120333  + 2942.21960580809 *T)
    sumY0  +=       0.175 * cos(3.84094267540  + 11614.67711121570 *T)
    sumY0  +=       0.187 * cos(0.76186633809  + 9778.86485864190 *T)
    sumY0  +=       0.175 * cos(6.15302938442  + 3744.58352825430 *T)
    sumY0  +=       0.241 * cos(5.00033837792  + 8390.35417501730 *T)
    sumY0  +=       0.236 * cos(0.45405670770  + 1.72829019180 *T)
    sumY0  +=       0.217 * cos(3.45380390356  + 17996.27498570570 *T)
    sumY0  +=       0.173 * cos(2.65478632822  + 6275.71848550709 *T)
    sumY0  +=       0.191 * cos(4.70676649838  + 394.86970254271 *T)
    sumY0  +=       0.231 * cos(4.17552456214  + 34596.12083716890 *T)
    sumY0  +=       0.169 * cos(0.40140025709  + 6438.25243194210 *T)
    sumY0  +=       0.235 * cos(5.04475734159  + 17256.38771885789 *T)
    sumY0  +=       0.188 * cos(4.02583282760  + 401.91593924071 *T)
    sumY0  +=       0.217 * cos(3.40363231104  + 10984.43616918330 *T)
    sumY0  +=       0.230 * cos(0.83612775625  + 6632.24408792290 *T)
    sumY0  +=       0.208 * cos(5.43298352980  + 11087.52894340190 *T)
    sumY0  +=       0.165 * cos(0.80102915744  + 743.23387801611 *T)
    sumY0  +=       0.177 * cos(0.01829077781  + 4796.85191144269 *T)
    sumY0  +=       0.208 * cos(6.16123229814  + 3097.64000524229 *T)
    sumY0  +=       0.167 * cos(2.40308314985  + 5714.49779344750 *T)
    sumY0  +=       0.198 * cos(4.83812277030  + 5539.84197197530 *T)
    sumY0  +=       0.169 * cos(2.16903380895  + 12132.68377958950 *T)
    sumY0  +=       0.214 * cos(2.77137176457  + 24492.64993113510 *T)
    sumY0  +=       0.172 * cos(5.97118942936  + 6233.56264200310 *T)
    sumY0  +=       0.172 * cos(3.81893679688  + 6333.07669294670 *T)
    sumY0  +=       0.223 * cos(3.42911214381  + 266.85085920531 *T)
    sumY0  +=       0.159 * cos(2.33166453392  + 13199.17925677950 *T)
    sumY0  +=       0.219 * cos(0.69566285170  + 10344.05124790229 *T)
    sumY0  +=       0.207 * cos(1.92518585993  + 12569.98594611810 *T)
    sumY0  +=       0.176 * cos(2.48119576594  + 12012.82611462390 *T)
    sumY0  +=       0.170 * cos(4.43959906760  + 6294.36536773881 *T)
    sumY0  +=       0.170 * cos(5.35052715864  + 6272.27396721100 *T)
    sumY0  +=       0.205 * cos(1.15927608909  + 13119.47728534169 *T)
    sumY0  +=       0.167 * cos(3.97305893961  + 17686.99666304990 *T)
    sumY0  +=       0.210 * cos(2.45255073470  + 13521.50762410789 *T)
    sumY0  +=       0.163 * cos(6.18035129909  + 802.06474860730 *T)
    sumY0  +=       0.159 * cos(0.13779660073  + 5017.26455388150 *T)
    sumY0  +=       0.148 * cos(5.46314197385  + 419.72846135871 *T)
    sumY0  +=       0.165 * cos(5.59205778351  + 20199.33877711650 *T)
    sumY0  +=       0.196 * cos(5.15079536124  + 33326.33491569069 *T)
    sumY0  +=       0.146 * cos(1.96564383477  + 19800.70213874130 *T)
    sumY0  +=       0.148 * cos(1.10385783973  + 6852.14154150230 *T)
    sumY0  +=       0.146 * cos(4.95120975616  + 17370.60479339330 *T)
    sumY0  +=       0.143 * cos(5.56441247552  + 5618.56362234490 *T)
    sumY0  +=       0.142 * cos(0.65902902155  + 17654.53672226610 *T)
    sumY0  +=       0.140 * cos(5.93480456615  + 2008.80135664250 *T)
    sumY0  +=       0.153 * cos(3.36826827070  + 5436.74919775670 *T)
    sumY0  +=       0.162 * cos(5.27294659555  + 775.27879384050 *T)
    sumY0  +=       0.162 * cos(3.21674523006  + 12552.16842346470 *T)
    sumY0  +=       0.143 * cos(3.89298583971  + 5010.15100688070 *T)
    sumY0  +=       0.188 * cos(2.11622447442  + 28.69300495130 *T)
    sumY0  +=       0.155 * cos(4.58570766573  + 11610.10636149570 *T)
    sumY0  +=       0.174 * cos(1.00725781500  + 20452.62559473830 *T)
    sumY0  +=       0.179 * cos(0.59168717373  + 27511.22405605370 *T)
    sumY0  +=       0.135 * cos(4.79054860554  + 12431.33043743090 *T)
    sumY0  +=       0.164 * cos(4.03840473864  + 28767.16824196750 *T)
    sumY0  +=       0.134 * cos(3.87019546457  + 16840.91382829870 *T)
    sumY0  +=       0.174 * cos(2.70675879120  + 19805.07110906630 *T)
    sumY0  +=       0.146 * cos(6.27087503385  + 12701.46059750170 *T)
    sumY0  +=       0.137 * cos(1.03021623583  + 11.28951774740 *T)
    sumY0  +=       0.167 * cos(5.98298407227  + 17473.69756761190 *T)
    sumY0  +=       0.152 * cos(2.62025173907  + 16627.61473286070 *T)
    sumY0  +=       0.124 * cos(4.22571397919  + 6948.07571260490 *T)
    sumY0  +=       0.135 * cos(4.95808633295  + 3531.28443281630 *T)
    sumY0  +=       0.172 * cos(0.23022549161  + 167959.14629657479 *T)
    sumY0  +=       0.123 * cos(0.49830220802  + 23013.29572210369 *T)
    sumY0  +=       0.122 * cos(4.12184350213  + 3583.58484815730 *T)
    sumY0  +=       0.170 * cos(6.16462326207  + 333857.83394425618 *T)
    sumY0  +=       0.155 * cos(5.62824965405  + 6058.48723680599 *T)
    sumY0  +=       0.120 * cos(5.90379805618  + 12809.12412144031 *T)
    sumY0  +=       0.141 * cos(3.95455666878  + 162420.79565227419 *T)
    sumY0  +=       0.120 * cos(4.26997270152  + 12528.36784483050 *T)
    sumY0  +=       0.123 * cos(1.49653119992  + 25933.88051360590 *T)
    sumY0  +=       0.144 * cos(3.70806542711  + 95.73540973430 *T)
    sumY0  +=       0.165 * cos(2.31400541864  + 52669.82577581910 *T)
    sumY0  +=       0.119 * cos(5.90934736896  + 19247.62037086590 *T)
    sumY0  +=       0.117 * cos(5.45034774436  + 11610.79577585770 *T)
    sumY0  +=       0.163 * cos(1.54456737879  + 661.47674426450 *T)
    sumY0  +=       0.163 * cos(2.78769461236  + 9929.67004482930 *T)
    sumY0  +=       0.119 * cos(2.71628583938  + 12250.00364780970 *T)
    sumY0  +=       0.145 * cos(2.33523694188  + 6205.64589704690 *T)
    sumY0  +=       0.145 * cos(1.17170397718  + 6360.99343790291 *T)
    sumY0  +=       0.159 * cos(2.61434005227  + 228278.34846897016 *T)
    sumY0  +=       0.114 * cos(3.41169068642  + 19402.55313533309 *T)
    sumY0  +=       0.147 * cos(5.63174103655  + 38526.33053338850 *T)
    sumY0  +=       0.156 * cos(1.00709118546  + 4307.82712161890 *T)
    sumY0  +=       0.131 * cos(0.13588364911  + 21228.63584102931 *T)
    sumY0  +=       0.118 * cos(3.83848144749  + 6263.64990657511 *T)
    sumY0  +=       0.118 * cos(5.95164477875  + 6302.98942837470 *T)
    sumY0  +=       0.137 * cos(0.97335445483  + 6315.85221826630 *T)
    sumY0  +=       0.137 * cos(2.53358646423  + 6250.78711668350 *T)
    sumY0  +=       0.109 * cos(1.19644618130  + 11925.51791008410 *T)
    sumY0  +=       0.111 * cos(2.80331366377  + 6226.42129253930 *T)
    sumY0  +=       0.111 * cos(0.70362725529  + 6340.21804241050 *T)
    sumY0  +=       0.123 * cos(0.80537647876  + 24734.39821404090 *T)
    sumY0  +=       0.107 * cos(0.59542543867  + 12463.30274324771 *T)
    sumY0  +=       0.133 * cos(1.45986519440  + 18875.76968725751 *T)
    sumY0  +=       0.123 * cos(4.07179306364  + 6260.54446602410 *T)
    sumY0  +=       0.123 * cos(5.71833316259  + 6306.09486892571 *T)
    sumY0  +=       0.113 * cos(0.38322617607  + 2111.89413086110 *T)
    sumY0  +=       0.131 * cos(2.04114030865  + 18415.75962958090 *T)
    sumY0  +=       0.110 * cos(0.19106597205  + 6289.94822637491 *T)
    sumY0  +=       0.110 * cos(3.31587494701  + 6276.69110857489 *T)
    sumY0  +=       0.103 * cos(2.31964172720  + 6241.76887649010 *T)
    sumY0  +=       0.103 * cos(1.18729919186  + 6324.87045845970 *T)
    sumY0  +=       0.113 * cos(0.63409757277  + 3496.27664361750 *T)
    sumY0  +=       0.103 * cos(3.00856574797  + 10344.53888286930 *T)
    sumY0  +=       0.115 * cos(1.09796447220  + 24336.24921063270 *T)
    sumY0  +=       0.107 * cos(0.14947871700  + 83974.07917321339 *T)
    sumY0  +=       0.107 * cos(1.18802993790  + 84020.10309797739 *T)
    sumY0  +=       0.106 * cos(1.70601019340  + 12772.58106590351 *T)
    sumY0  +=       0.143 * cos(6.22463763653  + 2069.25065239010 *T)
    sumY0  +=       0.143 * cos(5.63223127595  + 18773.20529618210 *T)
    sumY0  +=       0.115 * cos(5.35405637827  + 3641.49075403570 *T)
    sumY0  +=       0.112 * cos(1.02603600816  + 11499.41240530929 *T)
    sumY0  +=       0.099 * cos(0.21940282455  + 11190.62171762050 *T)
    sumY0  +=       0.126 * cos(5.90533164276  + 18823.17304765790 *T)
    sumY0  +=       0.115 * cos(3.09464836385  + 12570.32767072940 *T)
    sumY0  +=       0.100 * cos(1.95823440417  + 16062.42834360030 *T)
    sumY0  +=       0.107 * cos(2.45486702592  + 5216.82419028490 *T)
    sumY0  +=       0.120 * cos(0.25356721797  + 9814.36028280769 *T)
    sumY0  +=       0.138 * cos(2.98838673165  + 6210.02254161590 *T)
    sumY0  +=       0.138 * cos(0.51855418741  + 6356.61679333390 *T)
    sumY0  +=       0.105 * cos(2.60431357489  + 12721.32828193349 *T)
    sumY0  +=       0.135 * cos(5.28295396943  + 18699.90817032310 *T)
    sumY0  +=       0.104 * cos(0.57981411648  + 12560.87259315890 *T)
    sumY0  +=       0.098 * cos(4.00435005538  + 5815.35467712050 *T)
    sumY0  +=       0.111 * cos(5.54435587296  + 10239.34004852730 *T)
    sumY0  +=       0.100 * cos(2.98744963581  + 263.32774085630 *T)
    sumY0  +=       0.095 * cos(0.17662269567  + 155.66421691770 *T)
    sumY0  +=       0.130 * cos(0.55194216539  + 27511.71169102070 *T)
    sumY0  +=       0.120 * cos(1.84949333794  + 31441.43375227330 *T)
    sumY0  +=       0.114 * cos(2.18292460802  + 6155.30082413750 *T)
    sumY0  +=       0.114 * cos(1.32401631104  + 6411.33851081230 *T)
    sumY0  +=       0.119 * cos(3.64842963894  + 951.96222373410 *T)
    sumY0  +=       0.117 * cos(3.47158263266  + 28236.98964190589 *T)
    sumY0  +=       0.107 * cos(0.33835239167  + 21.01921297590 *T)
    sumY0  +=       0.095 * cos(4.06228359170  + 11300.82803883990 *T)
    sumY0  +=       0.110 * cos(1.22040625314  + 6312.74917601091 *T)
    sumY0  +=       0.110 * cos(2.28653466592  + 6253.89015893890 *T)
    sumY0  +=       0.097 * cos(2.64989366154  + 78263.95324220609 *T)
    sumY0  +=       0.095 * cos(4.41295990485  + 23938.10020722450 *T)
    sumY0  +=       0.091 * cos(1.42377843647  + 12829.47944083910 *T)
    sumY0  +=       0.112 * cos(2.74085534623  + 16737.82105408010 *T)
    sumY0  +=       0.126 * cos(4.67942071523  + 3.27930086550 *T)
    sumY0  +=       0.100 * cos(1.46206803081  + 6133.75647034030 *T)
    sumY0  +=       0.127 * cos(0.70264556724  + 1293.24040609909 *T)
    sumY0  +=       0.119 * cos(4.42085627054  + 17893.18221148710 *T)
    sumY0  +=       0.089 * cos(6.04696991631  + 23539.95120381630 *T)
    sumY0  +=       0.126 * cos(4.96989590925  + 311565.26273852371 *T)
    sumY0  +=       0.093 * cos(4.10519400640  + 736.12033101530 *T)
    sumY0  +=       0.120 * cos(0.51718078841  + 14158.99153109910 *T)
    sumY0  +=       0.118 * cos(3.00291998094  + 16061.94070863329 *T)
    sumY0  +=       0.099 * cos(2.04878528903  + 6432.88286460950 *T)
    sumY0  +=       0.095 * cos(1.28476695157  + 2699.49100183409 *T)
    sumY0  +=       0.089 * cos(0.73014646335  + 15671.32557689010 *T)
    sumY0  +=       0.124 * cos(5.19246908636  + 178430.29100801519 *T)
    sumY0  +=       0.089 * cos(5.78577582542  + 6751.28465782931 *T)
    sumY0  +=       0.096 * cos(1.07414401415  + 7349.81514466491 *T)
    sumY0  +=       0.096 * cos(5.45460641914  + 24066.05174025910 *T)
    sumY0  +=       0.103 * cos(2.80667748337  + 18202.46053414290 *T)
    sumY0  +=       0.099 * cos(4.49603988169  + 6252.40554183991 *T)
    sumY0  +=       0.099 * cos(5.29408634455  + 6314.23379310990 *T)
    sumY0  +=       0.090 * cos(1.96044978655  + 58864.30010066279 *T)
    sumY0  +=       0.086 * cos(2.16910774240  + 9380.71585523369 *T)
    sumY0  +=       0.086 * cos(2.50989174885  + 10557.35034334029 *T)
    sumY0  +=       0.098 * cos(1.84392439378  + 6439.72038797730 *T)
    sumY0  +=       0.098 * cos(1.66301652528  + 6126.91894697251 *T)
    sumY0  +=       0.086 * cos(6.05387198611  + 23123.98967829010 *T)
    sumY0  +=       0.118 * cos(2.16764726700  + 13951.95709241740 *T)
    sumY0  +=       0.117 * cos(2.53823002855  + 89570.23393702849 *T)
    sumY0  +=       0.083 * cos(4.71213278190  + 8858.07112683710 *T)
    sumY0  +=       0.092 * cos(2.44147365747  + 12342.05072176440 *T)
    sumY0  +=       0.085 * cos(0.72336255749  + 5017.75218884850 *T)
    sumY0  +=       0.101 * cos(4.94259233986  + 18429.98672358250 *T)
    sumY0  +=       0.088 * cos(2.19664804056  + 17054.21292373670 *T)
    sumY0  +=       0.113 * cos(0.47970289838  + 12985.88016134151 *T)
    sumY0  +=       0.083 * cos(2.47657261289  + 20597.00014555770 *T)
    sumY0  +=       0.082 * cos(4.90614394615  + 5483.49854230950 *T)
    sumY0  +=       0.087 * cos(1.67775805144  + 21424.22282681990 *T)
    sumY0  +=       0.107 * cos(3.78836692146  + 522.33360061030 *T)
    sumY0  +=       0.084 * cos(0.84028318953  + 6187.34044025710 *T)
    sumY0  +=       0.084 * cos(2.66665772953  + 6379.29889469270 *T)
    sumY0  +=       0.107 * cos(2.45229388212  + 24382.83529095790 *T)
    sumY0  +=       0.082 * cos(5.66196525445  + 8983.05448679250 *T)
    sumY0  +=       0.105 * cos(2.30294939014  + 6131.42238638970 *T)
    sumY0  +=       0.105 * cos(1.20399152892  + 6435.21694856010 *T)
    sumY0  +=       0.114 * cos(2.50001149568  + 8258.81221333091 *T)
    sumY0  +=       0.082 * cos(5.40031574607  + 3957.88262369230 *T)
    sumY0  +=       0.112 * cos(4.27968226719  + 3738.51761262449 *T)
    sumY0  +=       0.102 * cos(1.32703960779  + 5767.85579638190 *T)
    sumY0  +=       0.102 * cos(2.17990131127  + 6798.78353856790 *T)
    sumY0  +=       0.092 * cos(6.06909982221  + 18216.68762814450 *T)
    sumY0  +=       0.083 * cos(2.92326536483  + 29864.09020982550 *T)
    sumY0  +=       0.112 * cos(3.91825505548  + 12189.02190955050 *T)
    sumY0  +=       0.080 * cos(2.01989715959  + 24080.27883426070 *T)
    sumY0  +=       0.080 * cos(0.09553378894  + 15141.14697682849 *T)
    sumY0  +=       0.088 * cos(4.57487416308  + 1573.57660661529 *T)
    sumY0  +=       0.081 * cos(5.29168357177  + 1550.80140516450 *T)
    sumY0  +=       0.092 * cos(2.79378764615  + 101426.45258845300 *T)
    sumY0  +=       0.110 * cos(5.08245905369  + 78423.94833416230 *T)
    sumY0  +=       0.087 * cos(3.89276904283  + 1580.62284331330 *T)
    sumY0  +=       0.085 * cos(1.43001232077  + 27043.25906569930 *T)
    sumY0  +=       0.078 * cos(3.02198277867  + 6812.52299760250 *T)
    sumY0  +=       0.080 * cos(1.79867193339  + 6081.06627230081 *T)
    sumY0  +=       0.080 * cos(1.70826898567  + 6485.57306264900 *T)
    sumY0  +=       0.079 * cos(2.96736855880  + 36109.62602214810 *T)
    sumY0  +=       0.088 * cos(5.58307910278  + 16944.00660251730 *T)
    sumY0  +=       0.081 * cos(2.98944824758  + 5.27910682390 *T)
    sumY0  +=       0.093 * cos(1.14380332092  + 16737.33341911309 *T)
    sumY0  +=       0.088 * cos(1.67821631800  + 12537.94632999850 *T)
    sumY0  +=       0.096 * cos(0.48155847774  + 20198.85114214949 *T)
    sumY0  +=       0.106 * cos(3.73745326901  + 56600.03547203870 *T)
    sumY0  +=       0.097 * cos(1.71169429599  + 6040.10342853389 *T)
    sumY0  +=       0.081 * cos(5.19219909829  + 3956.26419853590 *T)
    sumY0  +=       0.077 * cos(6.10587791973  + 40796.58274015770 *T)
    sumY0  +=       0.076 * cos(3.78044498975  + 22743.16556203289 *T)
    sumY0  +=       0.098 * cos(0.77646852784  + 42456.54022960810 *T)
    sumY0  +=       0.083 * cos(0.63561076928  + 19801.18977370830 *T)
    sumY0  +=       0.079 * cos(2.93697401198  + 5622.08674069391 *T)
    sumY0  +=       0.096 * cos(0.41418225890  + 5888.69378241570 *T)
    sumY0  +=       0.096 * cos(3.09275866016  + 6677.94555253410 *T)
    sumY0  +=       0.075 * cos(4.66945944657  + 41194.73174356590 *T)
    sumY0  +=       0.074 * cos(5.37967983395  + 6261.98402700790 *T)
    sumY0  +=       0.074 * cos(4.41044639229  + 6304.65530794190 *T)
    sumY0  +=       0.073 * cos(2.45644384945  + 5870.94857060050 *T)
    sumY0  +=       0.073 * cos(1.05049706961  + 6695.69076434930 *T)
    sumY0  +=       0.085 * cos(2.08161902252  + 12850.25483633150 *T)
    sumY0  +=       0.072 * cos(4.75417028947  + 6253.49822932610 *T)
    sumY0  +=       0.072 * cos(5.03595593677  + 6313.14110562371 *T)
    sumY0  +=       0.100 * cos(3.80714894085  + 5316.34879003930 *T)
    sumY0  +=       0.085 * cos(1.33904345426  + 12282.53619860110 *T)
    sumY0  +=       0.075 * cos(0.39510472937  + 24422.61416889080 *T)
    sumY0  +=       0.076 * cos(0.41148054527  + 63659.12156832110 *T)
    sumY0  +=       0.072 * cos(4.97177288715  + 16723.59396007850 *T)
    sumY0  +=       0.097 * cos(0.21179463686  + 17995.78735073869 *T)
    sumY0  +=       0.087 * cos(2.06337614338  + 18106.48130692510 *T)
    sumY0  +=       0.080 * cos(3.60508005382  + 17363.49124639250 *T)
    sumY0  +=       0.070 * cos(3.06053136401  + 6124.37614969171 *T)
    sumY0  +=       0.070 * cos(0.44640955505  + 6442.26318525810 *T)
    sumY0  +=       0.070 * cos(4.42572835968  + 4705.48849006010 *T)
    sumY0  +=       0.078 * cos(3.37769653143  + 23550.58786916610 *T)
    sumY0  +=       0.070 * cos(3.84581453123  + 12036.21691740469 *T)
    sumY0  +=       0.077 * cos(5.00579533671  + 5237.67719632029 *T)
    sumY0  +=       0.069 * cos(1.30888647192  + 16207.64245401849 *T)
    sumY0  +=       0.082 * cos(3.91623659935  + 6774.98295993371 *T)
    sumY0  +=       0.067 * cos(4.88284115481  + 7083.14079264031 *T)
    sumY0  +=       0.067 * cos(0.26461151247  + 6394.50630976251 *T)
    sumY0  +=       0.067 * cos(3.24232940659  + 6172.13302518730 *T)
    sumY0  +=       0.077 * cos(0.28397734649  + 9924.56660402710 *T)
    sumY0  +=       0.080 * cos(3.79678848458  + 22380.99961775750 *T)
    sumY0  +=       0.070 * cos(3.03095059330  + 6390.98319141350 *T)
    sumY0  +=       0.070 * cos(0.47599032576  + 6175.65614353630 *T)
    sumY0  +=       0.074 * cos(4.12433601146  + 16193.41536001689 *T)
    sumY0  +=       0.076 * cos(4.07719654455  + 32217.44399856430 *T)
    sumY0  +=       0.091 * cos(0.90446873976  + 6653.01948341530 *T)
    sumY0  +=       0.091 * cos(2.60247217930  + 5913.61985153450 *T)
    sumY0  +=       0.067 * cos(4.94532713838  + 6265.50714535691 *T)
    sumY0  +=       0.067 * cos(4.84479908786  + 6301.13218959290 *T)
    sumY0  +=       0.067 * cos(3.01186125982  + 5959.32661585050 *T)
    sumY0  +=       0.083 * cos(4.51648793763  + 16723.10632511149 *T)
    sumY0  +=       0.080 * cos(5.49180050451  + 23646.56709638390 *T)
    sumY0  +=       0.068 * cos(2.07232606852  + 4897.42439113870 *T)
    sumY0  +=       0.069 * cos(0.56996686256  + 6944.55259425590 *T)
    sumY0  +=       0.089 * cos(3.57060375879  + 10660.44311755889 *T)
    sumY0  +=       0.079 * cos(3.24586891365  + 35371.64344849290 *T)
    sumY0  +=       0.073 * cos(0.54359299539  + 6370.62787201471 *T)
    sumY0  +=       0.073 * cos(2.96334792367  + 6196.01146293510 *T)
    sumY0  +=       0.065 * cos(5.23672190600  + 22345.01655862470 *T)
    sumY0  +=       0.064 * cos(5.80516008052  + 15265.64270181689 *T)
    sumY0  +=       0.066 * cos(1.12499582368  + 6315.29197329170 *T)
    sumY0  +=       0.066 * cos(2.38194509538  + 6251.34736165811 *T)
    sumY0  +=       0.065 * cos(3.44248995108  + 323.74923414091 *T)
    sumY0  +=       0.076 * cos(1.46082924902  + 10873.74221299689 *T)
    sumY0  +=       0.084 * cos(0.49974269222  + 109.96250373590 *T)
    sumY0  +=       0.063 * cos(2.70916029747  + 11823.40545693370 *T)
    sumY0  +=       0.064 * cos(4.80332084911  + 28774.28178896830 *T)
    sumY0  +=       0.072 * cos(2.11265208796  + 18099.75944096650 *T)
    sumY0  +=       0.066 * cos(1.21054975257  + 17576.79034183050 *T)
    sumY0  +=       0.072 * cos(0.02703719822  + 245707.70992182777 *T)
    sumY0  +=       0.061 * cos(3.91641579991  + 10557.83797830730 *T)
    sumY0  +=       0.064 * cos(5.46712129325  + 71430.93943561260 *T)
    sumY0  +=       0.068 * cos(1.33734717206  + 28760.05469496669 *T)
    sumY0  +=       0.062 * cos(2.25046201901  + 3854.78984947370 *T)
    sumY0  +=       0.073 * cos(4.61252989167  + 23440.38154794670 *T)
    sumY0  +=       0.060 * cos(1.65871121739  + 13088.97293556010 *T)
    sumY0  +=       0.078 * cos(2.99014426818  + 12564.91104475801 *T)
    sumY0  +=       0.060 * cos(2.90135289748  + 7548.88714610130 *T)
    sumY0  +=       0.059 * cos(4.12263682656  + 6286.35515085690 *T)
    sumY0  +=       0.075 * cos(2.85045311521  + 18625.12657175580 *T)
    sumY0  +=       0.058 * cos(5.18922617388  + 35050.24409195890 *T)
    sumY0  +=       0.060 * cos(4.91746089028  + 553.81322032590 *T)
    sumY0  +=       0.069 * cos(4.25329717989  + 167993.93845370729 *T)
    sumY0  +=       0.059 * cos(5.13629564655  + 41991.02975038230 *T)
    sumY0  +=       0.062 * cos(4.44636395235  + 15663.79170522509 *T)
    sumY0  +=       0.081 * cos(5.98297744820  + 7250.29054491051 *T)
    sumY0  +=       0.059 * cos(1.11500712714  + 6277.25375184510 *T)
    sumY0  +=       0.059 * cos(2.39193379192  + 6289.38558310470 *T)
    sumY0  +=       0.063 * cos(0.56300539120  + 18003.38853270651 *T)
    sumY0  +=       0.057 * cos(3.30567043605  + 793.01870595090 *T)
    sumY0  +=       0.057 * cos(1.59794444440  + 10027.65937824569 *T)
    sumY0  +=       0.060 * cos(0.63448798616  + 647.25465079831 *T)
    sumY0  +=       0.072 * cos(5.73566114928  + 4597.77991000629 *T)
    sumY0  +=       0.059 * cos(5.39392547964  + 6279.30891415889 *T)
    sumY0  +=       0.068 * cos(1.14499717138  + 3166.66025521511 *T)
    sumY0  +=       0.056 * cos(3.10776343336  + 6226.51640530510 *T)
    sumY0  +=       0.056 * cos(0.39917748570  + 6340.12292964471 *T)
    sumY0  +=       0.056 * cos(0.95649088929  + 12303.31159409350 *T)
    sumY0  +=       0.057 * cos(1.51972566902  + 5864.39526857430 *T)
    sumY0  +=       0.057 * cos(1.98721525004  + 6702.24406637550 *T)
    sumY0  +=       0.069 * cos(0.33222800273  + 23536.36077516450 *T)
    sumY0  +=       0.059 * cos(2.36664156781  + 5760.25461441409 *T)
    sumY0  +=       0.065 * cos(1.96288183765  + 12139.30969162329 *T)
    sumY0  +=       0.059 * cos(1.81423548051  + 67589.33126454070 *T)
    sumY0  +=       0.073 * cos(2.08854442290  + 11079.92776143409 *T)
    sumY0  +=       0.076 * cos(0.77803347659  + 23227.08245250870 *T)
    sumY0  +=       0.067 * cos(0.54633497662  + 4804.45309341050 *T)
    sumY0  +=       0.073 * cos(4.43869578954  + 30349.12759025050 *T)
    sumY0  +=       0.074 * cos(3.17776060575  + 92747.93298430610 *T)
    sumY0  +=       0.070 * cos(6.27818224660  + 12299.78847574450 *T)
    sumY0  +=       0.060 * cos(5.08460970539  + 12171.76963240710 *T)
    sumY0  +=       0.059 * cos(1.03458131364  + 15.49628866851 *T)
    sumY0  +=       0.074 * cos(1.19968434762  + 3684.13423250670 *T)
    sumY0  +=       0.058 * cos(1.58539308271  + 6717.03140535170 *T)
    sumY0  +=       0.053 * cos(3.68674599490  + 26087.65932409069 *T)
    sumY0  +=       0.059 * cos(5.76308404993  + 12164.72339570910 *T)
    sumY0  +=       0.063 * cos(5.23211999381  + 5219.51794905560 *T)
    sumY0  +=       0.056 * cos(2.45377850108  + 3178.38960805111 *T)
    sumY0  +=       0.063 * cos(2.16148808063  + 5227.38225558179 *T)
    sumY0  +=       0.070 * cos(4.83368738065  + 19004.40413192490 *T)
    sumY0  +=       0.072 * cos(2.62595674809  + 4583.55281600469 *T)
    sumY0  +=       0.052 * cos(0.96524582701  + 3627.26366003410 *T)
    sumY0  +=       0.052 * cos(3.30278586555  + 6411.27120050950 *T)
    sumY0  +=       0.052 * cos(0.20415505351  + 6155.36813444030 *T)
    sumY0  +=       0.054 * cos(2.04667654344  + 5849.60792959810 *T)
    sumY0  +=       0.066 * cos(5.93782050272  + 5791.16874004909 *T)
    sumY0  +=       0.066 * cos(5.93782050272  + 5791.65637501610 *T)
    sumY0  +=       0.051 * cos(2.41252260456  + 5113.24378109930 *T)
    sumY0  +=       0.058 * cos(2.72227505748  + 30775.72578112650 *T)
    sumY0  +=       0.061 * cos(0.72832616881  + 14169.54974474690 *T)
    sumY0  +=       0.054 * cos(3.16101850189  + 10454.74520408871 *T)
    sumY0  +=       0.056 * cos(1.20035157643  + 23141.80220040810 *T)
    sumY0  +=       0.067 * cos(3.89077193587  + 28313.04498717750 *T)
    sumY0  +=       0.056 * cos(0.37854134464  + 5244.29305668450 *T)
    sumY0  +=       0.062 * cos(0.60927850338  + 5657.64947516250 *T)
    sumY0  +=       0.062 * cos(2.89766241568  + 6908.98985978730 *T)
    sumY0  +=       0.050 * cos(3.10994458213  + 2574.99527684569 *T)
    sumY0  +=       0.050 * cos(2.89352148340  + 536.56069461190 *T)
    sumY0  +=       0.050 * cos(0.38317924664  + 6684.99178923210 *T)
    sumY0  +=       0.050 * cos(3.12376167242  + 5881.64754571770 *T)
    sumY0  +=       0.049 * cos(0.36435757253  + 16310.73522823709 *T)
    sumY0  +=       0.049 * cos(3.50595022612  + 16311.22286320410 *T)
    sumY0  +=       0.065 * cos(2.20156720545  + 46386.74992582770 *T)
    sumY0  +=       0.069 * cos(5.16096365273  + 60530.24516825830 *T)
    sumY0  +=       0.068 * cos(4.61952317468  + 96678.14268052568 *T)
    sumY0  +=       0.049 * cos(4.01249891427  + 12323.17927852529 *T)
    sumY0  +=       0.050 * cos(2.38095415684  + 21954.40142688150 *T)
    sumY0  +=       0.049 * cos(4.98226292129  + 4164.55580709650 *T)
    sumY0  +=       0.059 * cos(0.54830716960  + 33794.78754101210 *T)
    sumY0  +=       0.050 * cos(1.64594745036  + 43739.04616344930 *T)
    sumY0  +=       0.059 * cos(2.71376255252  + 3981.73385156551 *T)
    sumY0  +=       0.058 * cos(3.16580349936  + 27707.29867681129 *T)
    sumY0  +=       0.051 * cos(1.43461397348  + 7669.21494381111 *T)
    sumY0  +=       0.048 * cos(3.23576326319  + 3646.59419483791 *T)
    sumY0  +=       0.065 * cos(1.63790414031  + 82534.64743809510 *T)
    sumY0  +=       0.056 * cos(2.62361047024  + 800.06494264891 *T)
    sumY0  +=       0.057 * cos(2.56200976684  + 11609.61872652869 *T)
    sumY0  +=       0.048 * cos(3.66284145075  + 26.54213728330 *T)
    sumY0  +=       0.047 * cos(5.82227097996  + 30640.10045611650 *T)
    sumY0  +=       0.060 * cos(3.54507538375  + 8982.56685182550 *T)
    sumY0  +=       0.049 * cos(2.15364675842  + 96563.24283557820 *T)
    sumY0  +=       0.050 * cos(3.60628396854  + 17583.90388883130 *T)
    sumY0  +=       0.046 * cos(4.33029118489  + 23539.46356884929 *T)
    sumY0  +=       0.048 * cos(4.40476950322  + 170.42905313570 *T)
    sumY0  +=       0.055 * cos(0.06525117672  + 735.63269604830 *T)
    sumY0  +=       0.054 * cos(5.90233323358  + 6680.98103591609 *T)
    sumY0  +=       0.060 * cos(3.29126676894  + 44033.88377205590 *T)
    sumY0  +=       0.064 * cos(5.97731735914  + 80181.29364935629 *T)
    sumY0  +=       0.059 * cos(3.60959132830  + 24279.35083569710 *T)
    sumY0  +=       0.049 * cos(3.59486989773  + 5490.54477900750 *T)
    sumY0  +=       0.058 * cos(5.95314247069  + 6490.02040477150 *T)
    sumY0  +=       0.058 * cos(3.83698375555  + 6076.61893017830 *T)
    sumY0  +=       0.046 * cos(1.35461053261  + 18208.10612510850 *T)
    sumY0  +=       0.045 * cos(3.19696372316  + 4480.46004178609 *T)
    sumY0  +=       0.055 * cos(1.99803947181  + 16097.92376776610 *T)
    sumY0  +=       0.055 * cos(1.99803947181  + 16097.43613279909 *T)
    sumY0  +=       0.061 * cos(3.53347961961  + 491.90710994230 *T)
    sumY0  +=       0.054 * cos(1.22650408797  + 6006.57389561050 *T)
    sumY0  +=       0.054 * cos(2.28043683109  + 6560.06543933930 *T)
    sumY0  +=       0.044 * cos(4.77367705513  + 5333.65642353809 *T)
    sumY0  +=       0.052 * cos(3.85218997059  + 244287.84382471119 *T)
    sumY0  +=       0.044 * cos(2.69173127084  + 2301.34199842589 *T)
    sumY0  +=       0.045 * cos(4.35951224787  + 33794.29990604510 *T)
    sumY0  +=       0.046 * cos(3.84874383995  + 4384.48081456829 *T)
    sumY0  +=       0.048 * cos(4.63629874583  + 35579.93505705350 *T)
    sumY0  +=       0.051 * cos(4.91886576255  + 18326.89394936390 *T)
    sumY0  +=       0.045 * cos(2.63923033609  + 533.45790092711 *T)
    sumY0  +=       0.046 * cos(3.15339657493  + 16703.30595098250 *T)
    sumY0  +=       0.047 * cos(2.39403388232  + 5209.22300831710 *T)
    sumY0  +=       0.051 * cos(1.31112479278  + 38650.41732368250 *T)
    sumY0  +=       0.046 * cos(2.47672465258  + 12555.34981720240 *T)
    sumY0  +=       0.052 * cos(4.64446990453  + 5992.34680160890 *T)
    sumY0  +=       0.052 * cos(5.14565632171  + 6574.29253334090 *T)
    sumY0  +=       0.044 * cos(3.80837636245  + 69942.19741831250 *T)
    sumY0  +=       0.049 * cos(3.62132943845  + 12146.91087359110 *T)
    sumY0  +=       0.053 * cos(4.67452542700  + 39301.85314471250 *T)
    sumY0  +=       0.045 * cos(4.65107782321  + 46360.93924099490 *T)
    sumY0  +=       0.043 * cos(1.54528326296  + 8402.08352785330 *T)
    sumY0  +=       0.043 * cos(0.19791989913  + 44137.19516685750 *T)
    sumY0  +=       0.049 * cos(3.54725333487  + 22029.96913718610 *T)
    sumY0  +=       0.044 * cos(1.47478212539  + 31441.92138724031 *T)
    sumY0  +=       0.042 * cos(2.85938112982  + 9225.29545579949 *T)
    sumY0  +=       0.045 * cos(3.22032934298  + 71519.54096076029 *T)
    sumY0  +=       0.051 * cos(2.95623880108  + 7762.18624153930 *T)
    sumY0  +=       0.041 * cos(5.14395096650  + 3735.48212924251 *T)
    sumY0  +=       0.057 * cos(1.24099052077  + 40398.43373674950 *T)
    sumY0  +=       0.047 * cos(6.03171570454  + 4707.98623122570 *T)
    sumY0  +=       0.056 * cos(5.36736063063  + 28664.07546774890 *T)
    sumY0  +=       0.050 * cos(2.58729930740  + 11919.38468415150 *T)
    sumY0  +=       0.047 * cos(5.61037642387  + 316428.47249139857 *T)
    sumY0  +=       0.041 * cos(2.70468936208  + 27278.71263392430 *T)
    sumY0  +=       0.056 * cos(1.28041103259  + 83659.52068988250 *T)
    sumY0  +=       0.048 * cos(0.93182108526  + 12662.37474468410 *T)
    sumY0  +=       0.054 * cos(2.40910872834  + 16627.12709789369 *T)
    sumY0  +=       0.052 * cos(2.84085536047  + 12571.91844177370 *T)
    sumY0  +=       0.043 * cos(3.12713892129  + 7322.34627826531 *T)
    sumY0  +=       0.041 * cos(1.99075260862  + 664.99986261351 *T)
    sumY0  +=       0.042 * cos(3.11818521171  + 5641.95442512570 *T)
    sumY0  +=       0.044 * cos(6.16004757680  + 6425.49829451110 *T)
    sumY0  +=       0.044 * cos(3.63007864944  + 6141.14104043870 *T)
    sumY0  +=       0.039 * cos(5.09444795757  + 22594.29871319550 *T)
    sumY0  +=       0.040 * cos(1.17403028794  + 20894.53186172529 *T)
    sumY0  +=       0.044 * cos(3.71228299188  + 5540.32960694230 *T)
    sumY0  +=       0.050 * cos(0.98055905828  + 17782.48825530069 *T)
    sumY0  +=       0.054 * cos(3.12084664124  + 6006.28467373350 *T)
    sumY0  +=       0.054 * cos(0.38609427782  + 6560.35466121630 *T)
    sumY0  +=       0.045 * cos(0.83260974343  + 142861.14741877469 *T)
    sumY0  +=       0.047 * cos(5.83397233956  + 24499.76347813590 *T)
    sumY0  +=       0.043 * cos(6.07854396898  + 7026.30972800750 *T)
    sumY0  +=       0.047 * cos(0.38354823927  + 955.84355909210 *T)
    sumY0  +=       0.047 * cos(2.93749588479  + 24485.53638413430 *T)
    sumY0  +=       0.040 * cos(1.03647885736  + 9388.24972689870 *T)
    sumY0  +=       0.043 * cos(4.24359473122  + 6359.58573875050 *T)
    sumY0  +=       0.042 * cos(5.53288597787  + 8635.69818627969 *T)
    sumY0  +=       0.051 * cos(4.03405257508  + 27831.79440179969 *T)
    sumY0  +=       0.042 * cos(0.37690679264  + 6334.60000533731 *T)
    sumY0  +=       0.042 * cos(3.13003412642  + 6232.03932961249 *T)
    sumY0  +=       0.052 * cos(1.96232281613  + 11905.16259068530 *T)
    sumY0  +=       0.040 * cos(6.22187049858  + 7076.09455594230 *T)
    sumY0  +=       0.040 * cos(5.00617355027  + 5444.35037972450 *T)
    sumY0  +=       0.040 * cos(4.78395267597  + 7122.28895522530 *T)
    sumY0  +=       0.038 * cos(0.67446092329  + 44933.49317367390 *T)
    sumY0  +=       0.037 * cos(4.98656910933  + 5715.60102974450 *T)
    sumY0  +=       0.037 * cos(4.80355711691  + 6851.03830520530 *T)
    sumY0  +=       0.042 * cos(5.51152847292  + 36147.65369478390 *T)
    sumY0  +=       0.042 * cos(1.28742672679  + 6238.17844783830 *T)
    sumY0  +=       0.042 * cos(2.21951419227  + 6328.46088711150 *T)
    sumY0  +=       0.040 * cos(5.21427505339  + 3214.89256315970 *T)
    sumY0  +=       0.043 * cos(2.22601042104  + 23141.31456544110 *T)
    sumY0  +=       0.041 * cos(5.52595261887  + 19804.58347409930 *T)
    sumY0  +=       0.038 * cos(4.48062431201  + 20.11150191529 *T)
    sumY0  +=       0.050 * cos(0.05709227621  + 84334.66158130829 *T)
    sumY0  +=       0.050 * cos(4.41842921988  + 60284.16619777939 *T)

    # EARTH Y Coord
    sumY1  = 0
    sumY1  +=  154550.744 * cos(5.35844734917  + 0.24381748350 *T)
    sumY1  +=   51503.383 * cos(4.43183566713  + 12566.39551746630 *T)
    sumY1  +=    1290.763 * cos(4.38862020180  + 18849.47136745770 *T)
    sumY1  +=     702.576 * cos(0.18268310766  + 6283.31966747490 *T)
    sumY1  +=     430.422 * cos(4.02296125150  + 6282.83203250789 *T)
    sumY1  +=     212.689 * cos(0.16300557365  + 6279.79654912590 *T)
    sumY1  +=     212.524 * cos(3.34405166686  + 6286.84278582391 *T)
    sumY1  +=      62.308 * cos(5.07432195748  + 4705.97612502710 *T)
    sumY1  +=      59.808 * cos(2.24108323998  + 6257.02134767510 *T)
    sumY1  +=      59.474 * cos(1.26580020213  + 6309.61798727471 *T)
    sumY1  +=      48.914 * cos(3.64253038225  + 775.76642880750 *T)
    sumY1  +=      42.814 * cos(5.15102931682  + 1059.62574767270 *T)
    sumY1  +=      46.457 * cos(4.72328803502  + 7860.66320992270 *T)
    sumY1  +=      36.653 * cos(0.62556081234  + 5753.62870238030 *T)
    sumY1  +=      35.649 * cos(6.15990909241  + 5885.17066406670 *T)
    sumY1  +=      35.362 * cos(2.90135814576  + 6813.01063256950 *T)
    sumY1  +=      32.151 * cos(3.62597628866  + 6681.46867088311 *T)
    sumY1  +=      28.763 * cos(4.34539356227  + 25132.54721744910 *T)
    sumY1  +=      28.447 * cos(5.86215127530  + 6127.89926804070 *T)
    sumY1  +=      27.537 * cos(3.92840810925  + 6438.74006690910 *T)
    sumY1  +=      24.819 * cos(1.35151106278  + 5487.02166065850 *T)
    sumY1  +=      20.615 * cos(2.14741827387  + 7079.61767429131 *T)
    sumY1  +=      19.621 * cos(1.31571192713  + 5507.79705615090 *T)
    sumY1  +=      18.378 * cos(6.17760240913  + 11790.87290614230 *T)
    sumY1  +=      16.494 * cos(4.64463821908  + 11507.01358727710 *T)
    sumY1  +=      16.756 * cos(2.24858801378  + 7058.84227879890 *T)
    sumY1  +=      14.558 * cos(4.40450472214  + 6290.43321447570 *T)
    sumY1  +=      14.404 * cos(5.39361420229  + 6276.20612047410 *T)
    sumY1  +=      14.052 * cos(6.16986019111  + 796.54182429990 *T)
    sumY1  +=      12.261 * cos(5.72868820428  + 4693.75913722409 *T)
    sumY1  +=      12.758 * cos(5.06070649359  + 7.35736448430 *T)
    sumY1  +=      10.086 * cos(1.72162917776  + 3739.00524759150 *T)
    sumY1  +=       9.469 * cos(3.06840825171  + 6070.02057203690 *T)
    sumY1  +=      10.425 * cos(0.81162914438  + 6284.29998854310 *T)
    sumY1  +=      10.425 * cos(2.69531177468  + 6282.33934640670 *T)
    sumY1  +=       9.570 * cos(0.32347467387  + 4137.15425099970 *T)
    sumY1  +=       9.044 * cos(0.43893334417  + 6496.61876291290 *T)
    sumY1  +=       8.544 * cos(4.76970151659  + 1194.69082770810 *T)
    sumY1  +=       8.214 * cos(5.96640392040  + 1589.31671276730 *T)
    sumY1  +=       6.157 * cos(1.78596836636  + 8827.63408735830 *T)
    sumY1  +=       6.271 * cos(3.18396187423  + 8429.48508395010 *T)
    sumY1  +=       5.524 * cos(3.79861381511  + 4933.45225781610 *T)
    sumY1  +=       7.314 * cos(5.20341424492  + 4535.30325440790 *T)
    sumY1  +=       5.157 * cos(3.57885530448  + 11770.09751064990 *T)
    sumY1  +=       6.749 * cos(2.75433387783  + 5088.87265725030 *T)
    sumY1  +=       4.853 * cos(5.33701721257  + 6040.59106350090 *T)
    sumY1  +=       5.304 * cos(0.93005422139  + 3154.44326741210 *T)
    sumY1  +=       4.985 * cos(3.05697984231  + 12569.91863581531 *T)
    sumY1  +=       4.709 * cos(2.39306975468  + 5331.60126122430 *T)
    sumY1  +=       4.601 * cos(4.45296491447  + 6526.04827144891 *T)
    sumY1  +=       4.201 * cos(5.99364092036  + 7633.18707713370 *T)
    sumY1  +=       5.607 * cos(0.98720876729  + 5729.75026463250 *T)
    sumY1  +=       4.257 * cos(2.55370490945  + 3930.45351370310 *T)
    sumY1  +=       4.394 * cos(4.75948183378  + 12559.28197046550 *T)
    sumY1  +=       3.870 * cos(1.11601490340  + 7235.03807372550 *T)
    sumY1  +=       5.154 * cos(4.58850581748  + 8031.33608054190 *T)
    sumY1  +=       5.036 * cos(2.51929382179  + 6836.88907031730 *T)
    sumY1  +=       5.076 * cos(0.73526458069  + 7477.76667769950 *T)
    sumY1  +=       3.593 * cos(3.97650177033  + 12565.90788249930 *T)
    sumY1  +=       3.384 * cos(3.83274545225  + 10977.32262218251 *T)
    sumY1  +=       3.371 * cos(5.00200001284  + 11371.94850724170 *T)
    sumY1  +=       3.536 * cos(5.30432752281  + 4164.06817212950 *T)
    sumY1  +=       4.292 * cos(3.36706453434  + 1592.83983111630 *T)
    sumY1  +=       3.407 * cos(0.83519640650  + 3128.63258257930 *T)
    sumY1  +=       3.152 * cos(0.16276893009  + 5223.93773728571 *T)
    sumY1  +=       2.883 * cos(0.20838349180  + 1747.77259558350 *T)
    sumY1  +=       2.751 * cos(3.67369812017  + 7342.70159766410 *T)
    sumY1  +=       3.323 * cos(3.48959648614  + 801.57711364030 *T)
    sumY1  +=       2.999 * cos(0.68275608544  + 8636.18582124671 *T)
    sumY1  +=       2.496 * cos(5.09157906387  + 2145.92159899169 *T)
    sumY1  +=       3.358 * cos(5.82350976219  + 155.17658195069 *T)
    sumY1  +=       2.297 * cos(2.96479731544  + 17260.39847217390 *T)
    sumY1  +=       2.047 * cos(1.96580631113  + 1990.98883452450 *T)
    sumY1  +=       2.778 * cos(6.17391326343  + 5481.49873635110 *T)
    sumY1  +=       2.025 * cos(3.01082711135  + 951.47458876710 *T)
    sumY1  +=       2.663 * cos(1.00159277919  + 26.05450231630 *T)
    sumY1  +=       2.009 * cos(4.16311202386  + 4690.72365384210 *T)
    sumY1  +=       2.148 * cos(0.32608511593  + 537.04832957890 *T)
    sumY1  +=       1.853 * cos(4.41841662668  + 553.32558535889 *T)
    sumY1  +=       2.478 * cos(1.61013003664  + 1349.62359217529 *T)
    sumY1  +=       2.023 * cos(0.82015419806  + 522.82123557730 *T)
    sumY1  +=       1.735 * cos(4.49024043092  + 529.44714761109 *T)
    sumY1  +=       2.417 * cos(3.59693248271  + 7085.14059859870 *T)
    sumY1  +=       1.809 * cos(2.35536809783  + 397.90518592470 *T)
    sumY1  +=       1.893 * cos(2.65666986732  + 9438.00675237050 *T)
    sumY1  +=       1.708 * cos(5.00361322564  + 10989.05197501850 *T)
    sumY1  +=       1.898 * cos(3.91087003459  + 5216.33655531789 *T)
    sumY1  +=       1.820 * cos(3.44808607160  + 5230.56364931950 *T)
    sumY1  +=       1.541 * cos(5.81316348536  + 426.84200835950 *T)
    sumY1  +=       1.555 * cos(2.09924495059  + 13096.08648256090 *T)
    sumY1  +=       1.779 * cos(0.23737949225  + 12562.87239911730 *T)
    sumY1  +=       1.448 * cos(1.75844475078  + 5753.14106741329 *T)
    sumY1  +=       1.450 * cos(5.71201368260  + 6262.96434807611 *T)
    sumY1  +=       1.450 * cos(4.07811254364  + 6303.67498687370 *T)
    sumY1  +=       1.822 * cos(0.13446699374  + 10973.79950383350 *T)
    sumY1  +=       1.372 * cos(5.60520656116  + 7875.91568110771 *T)
    sumY1  +=       1.375 * cos(2.34041210234  + 7084.65296363170 *T)
    sumY1  +=       1.400 * cos(3.28144398068  + 12721.81591690050 *T)
    sumY1  +=       1.247 * cos(0.29225580262  + 2119.00767786191 *T)
    sumY1  +=       1.231 * cos(2.53164637932  + 18319.78040236310 *T)
    sumY1  +=       1.141 * cos(1.92161882496  + 1066.73929467350 *T)
    sumY1  +=       1.138 * cos(1.23855949465  + 2942.70724077510 *T)
    sumY1  +=       1.047 * cos(1.31663136695  + 5642.93474619389 *T)
    sumY1  +=       1.058 * cos(2.19001683016  + 242.97242145751 *T)
    sumY1  +=       1.025 * cos(0.60899466577  + 10447.14402212089 *T)
    sumY1  +=       0.973 * cos(4.95398300734  + 640.14110379750 *T)
    sumY1  +=       1.114 * cos(4.36978275435  + 15721.08260236190 *T)
    sumY1  +=       0.963 * cos(0.56100772029  + 2389.13783793270 *T)
    sumY1  +=       0.941 * cos(4.21626686326  + 20426.81490990550 *T)
    sumY1  +=       1.068 * cos(5.08632600308  + 529.93478257810 *T)
    sumY1  +=       0.929 * cos(1.54055499699  + 10575.65050042530 *T)
    sumY1  +=       0.899 * cos(5.79114005938  + 16496.60521368590 *T)
    sumY1  +=       0.889 * cos(5.01192484889  + 6277.79674316750 *T)
    sumY1  +=       0.889 * cos(4.77820137734  + 6288.84259178230 *T)
    sumY1  +=       1.018 * cos(2.66299766067  + 12540.09719766650 *T)
    sumY1  +=       1.060 * cos(5.76865363011  + 5760.74224938110 *T)
    sumY1  +=       1.017 * cos(4.02093017183  + 6805.89708556871 *T)
    sumY1  +=       0.797 * cos(3.79336643920  + 14314.41193053330 *T)
    sumY1  +=       0.795 * cos(0.10742263482  + 6286.60602489270 *T)
    sumY1  +=       0.795 * cos(3.39951828424  + 6280.03331005710 *T)
    sumY1  +=       0.800 * cos(3.18366130818  + 12029.59100537090 *T)
    sumY1  +=       0.719 * cos(2.29666441302  + 9623.93209417470 *T)
    sumY1  +=       0.698 * cos(4.94363251872  + 6148.25458743950 *T)
    sumY1  +=       0.686 * cos(3.79368723538  + 5856.72147659890 *T)
    sumY1  +=       0.673 * cos(3.99105175160  + 12964.54452087450 *T)
    sumY1  +=       0.902 * cos(0.69698416448  + 13368.21644859010 *T)
    sumY1  +=       0.678 * cos(4.84647193821  + 6418.38474751031 *T)
    sumY1  +=       0.671 * cos(6.00179402654  + 6709.91785835091 *T)
    sumY1  +=       0.744 * cos(2.68741072621  + 12043.81809937250 *T)
    sumY1  +=       0.706 * cos(2.51950741945  + 16730.70750707931 *T)
    sumY1  +=       0.661 * cos(2.42216375426  + 14712.56093394150 *T)
    sumY1  +=       0.633 * cos(5.58153853935  + 4292.57465043390 *T)
    sumY1  +=       0.621 * cos(1.74121096386  + 13119.96492030870 *T)
    sumY1  +=       0.638 * cos(2.04188798982  + 4690.23601887509 *T)
    sumY1  +=       0.625 * cos(0.31613419014  + 13518.11392371690 *T)
    sumY1  +=       0.637 * cos(4.59654254753  + 5746.51515537950 *T)
    sumY1  +=       0.628 * cos(3.42448659886  + 4686.64559022330 *T)
    sumY1  +=       0.653 * cos(5.85737456207  + 3929.96587873609 *T)
    sumY1  +=       0.578 * cos(0.62377454326  + 5088.38502228330 *T)
    sumY1  +=       0.563 * cos(3.24964920260  + 10447.63165708790 *T)
    sumY1  +=       0.610 * cos(5.19955844918  + 6820.12417957030 *T)
    sumY1  +=       0.666 * cos(4.87089013467  + 3634.37720703489 *T)
    sumY1  +=       0.720 * cos(5.18811029494  + 13916.26292712510 *T)
    sumY1  +=       0.620 * cos(4.30310535016  + 31415.62306744050 *T)
    sumY1  +=       0.553 * cos(2.71699575652  + 6259.44122972711 *T)
    sumY1  +=       0.553 * cos(0.78994516254  + 6307.19810522270 *T)
    sumY1  +=       0.523 * cos(5.47007776670  + 14143.73905991410 *T)
    sumY1  +=       0.524 * cos(3.97726441461  + 12139.79732659030 *T)
    sumY1  +=       0.569 * cos(4.83897139720  + 12036.70455237170 *T)
    sumY1  +=       0.514 * cos(3.87608366777  + 4700.87268422489 *T)
    sumY1  +=       0.563 * cos(3.78589343112  + 11014.86265985130 *T)
    sumY1  +=       0.495 * cos(2.00987402269  + 13362.69352428270 *T)
    sumY1  +=       0.481 * cos(0.93576028071  + 6279.43833211690 *T)
    sumY1  +=       0.481 * cos(2.57118063835  + 6287.20100283290 *T)
    sumY1  +=       0.466 * cos(2.94306081142  + 10177.50149701710 *T)
    sumY1  +=       0.440 * cos(5.16955709579  + 1577.58735993130 *T)
    sumY1  +=       0.437 * cos(1.61052953646  + 11499.90004027630 *T)
    sumY1  +=       0.452 * cos(3.98330238308  + 12573.50906446710 *T)
    sumY1  +=       0.499 * cos(5.79985868924  + 12410.97511803210 *T)
    sumY1  +=       0.440 * cos(0.57972339309  + 24073.16528725990 *T)
    sumY1  +=       0.587 * cos(0.00678196141  + 6.86972951729 *T)
    sumY1  +=       0.388 * cos(3.61198235279  + 7860.17557495569 *T)
    sumY1  +=       0.398 * cos(4.20068362004  + 8274.06468451590 *T)
    sumY1  +=       0.423 * cos(5.87838349624  + 4694.24677219110 *T)
    sumY1  +=       0.530 * cos(1.43593166640  + 12592.69383726610 *T)
    sumY1  +=       0.379 * cos(0.45309221604  + 6180.22689325630 *T)
    sumY1  +=       0.379 * cos(3.05384870302  + 6386.41244169350 *T)
    sumY1  +=       0.377 * cos(4.33272540326  + 7872.39256275871 *T)
    sumY1  +=       0.419 * cos(4.05140241721  + 5327.71992586630 *T)
    sumY1  +=       0.383 * cos(4.67353188247  + 6247.29180079750 *T)
    sumY1  +=       0.383 * cos(5.11659434377  + 6319.34753415231 *T)
    sumY1  +=       0.375 * cos(0.91076946484  + 2352.62233628830 *T)
    sumY1  +=       0.457 * cos(2.62120808067  + 6077.13411903770 *T)
    sumY1  +=       0.457 * cos(0.88573283839  + 6489.50521591210 *T)
    sumY1  +=       0.481 * cos(3.46562904756  + 4292.08701546690 *T)
    sumY1  +=       0.344 * cos(4.99906947930  + 18451.32236404950 *T)
    sumY1  +=       0.453 * cos(5.43816135018  + 2787.28684134090 *T)
    sumY1  +=       0.364 * cos(2.82703012884  + 6245.29199483910 *T)
    sumY1  +=       0.364 * cos(0.67991079022  + 6321.34734011070 *T)
    sumY1  +=       0.327 * cos(1.14128906507  + 12323.66691349230 *T)
    sumY1  +=       0.358 * cos(0.15971399837  + 9917.45305702629 *T)
    sumY1  +=       0.342 * cos(5.26072614817  + 6262.54427198250 *T)
    sumY1  +=       0.342 * cos(4.52940007806  + 6304.09506296730 *T)
    sumY1  +=       0.318 * cos(4.84603466410  + 11926.49823115230 *T)
    sumY1  +=       0.320 * cos(1.13215851865  + 77714.01528560399 *T)
    sumY1  +=       0.354 * cos(5.73706716445  + 7238.91940908350 *T)
    sumY1  +=       0.318 * cos(3.75423383039  + 6254.87048000710 *T)
    sumY1  +=       0.318 * cos(6.03589239585  + 6311.76885494270 *T)
    sumY1  +=       0.304 * cos(0.43890915611  + 12779.69461290430 *T)
    sumY1  +=       0.343 * cos(2.14223194635  + 6298.57213865991 *T)
    sumY1  +=       0.343 * cos(1.36470897271  + 6268.06719628990 *T)
    sumY1  +=       0.359 * cos(2.52875634099  + 1052.51220067191 *T)
    sumY1  +=       0.398 * cos(4.83537015899  + 1551.28904013150 *T)
    sumY1  +=       0.306 * cos(1.01793913946  + 5863.83502359970 *T)
    sumY1  +=       0.300 * cos(0.20529153395  + 90280.16698558679 *T)
    sumY1  +=       0.307 * cos(4.86343200722  + 3893.93801205869 *T)
    sumY1  +=       0.297 * cos(2.08300781298  + 5429.63565075589 *T)
    sumY1  +=       0.365 * cos(1.17462915798  + 10022.08109758290 *T)
    sumY1  +=       0.336 * cos(1.15285276926  + 17782.97589026770 *T)
    sumY1  +=       0.295 * cos(2.48770580169  + 6702.80431135010 *T)
    sumY1  +=       0.288 * cos(0.61716195626  + 18073.94875613370 *T)
    sumY1  +=       0.287 * cos(4.90856628789  + 1577.09972496430 *T)
    sumY1  +=       0.378 * cos(5.11185072605  + 2353.10997125530 *T)
    sumY1  +=       0.289 * cos(1.47281714617  + 213.54291292150 *T)
    sumY1  +=       0.347 * cos(2.29918133719  + 5223.45010231870 *T)
    sumY1  +=       0.321 * cos(0.66708336426  + 17797.20298426930 *T)
    sumY1  +=       0.289 * cos(3.73571084362  + 220.65645992230 *T)
    sumY1  +=       0.260 * cos(2.61364471837  + 955.35592412510 *T)
    sumY1  +=       0.268 * cos(5.19927474541  + 14945.07235607090 *T)
    sumY1  +=       0.242 * cos(2.77307042091  + 2544.55823736690 *T)
    sumY1  +=       0.245 * cos(3.88798752568  + 7632.69944216670 *T)
    sumY1  +=       0.234 * cos(2.50313805280  + 1596.43025976811 *T)
    sumY1  +=       0.267 * cos(4.25976324066  + 206.42936592071 *T)
    sumY1  +=       0.282 * cos(2.93181023342  + 13341.91812879030 *T)
    sumY1  +=       0.279 * cos(4.70853062526  + 4731.78680985990 *T)
    sumY1  +=       0.290 * cos(0.46137131163  + 5642.44206009270 *T)
    sumY1  +=       0.298 * cos(5.65061835836  + 17790.08943726851 *T)
    sumY1  +=       0.270 * cos(0.59449456012  + 12168.24651405810 *T)
    sumY1  +=       0.294 * cos(0.95816279728  + 2146.40923395870 *T)
    sumY1  +=       0.214 * cos(2.50623255971  + 8030.84844557490 *T)
    sumY1  +=       0.207 * cos(3.18045634687  + 213.05527795450 *T)
    sumY1  +=       0.216 * cos(4.03237077499  + 3185.43584474911 *T)
    sumY1  +=       0.195 * cos(4.44222765867  + 5884.68302909969 *T)
    sumY1  +=       0.199 * cos(2.36156558922  + 1748.26023055050 *T)
    sumY1  +=       0.250 * cos(3.04516390608  + 6924.19727485710 *T)
    sumY1  +=       0.194 * cos(0.79669370564  + 641.12142486571 *T)
    sumY1  +=       0.203 * cos(0.68276896116  + 6503.73230991370 *T)
    sumY1  +=       0.203 * cos(2.82417195790  + 6062.90702503610 *T)
    sumY1  +=       0.258 * cos(0.91786109532  + 796.05418933290 *T)
    sumY1  +=       0.189 * cos(3.20457516516  + 11506.52595231009 *T)
    sumY1  +=       0.186 * cos(4.84188314285  + 18209.57408114370 *T)
    sumY1  +=       0.184 * cos(2.63875516800  + 12566.46282776910 *T)
    sumY1  +=       0.180 * cos(4.19729049815  + 853.44019923550 *T)
    sumY1  +=       0.185 * cos(6.25586281827  + 3495.78900865049 *T)
    sumY1  +=       0.205 * cos(0.22424585668  + 5849.12029463110 *T)
    sumY1  +=       0.187 * cos(4.34056279496  + 10213.52936369450 *T)
    sumY1  +=       0.187 * cos(0.48606035843  + 6290.36590417291 *T)
    sumY1  +=       0.187 * cos(3.02088056063  + 6276.27343077690 *T)
    sumY1  +=       0.235 * cos(4.34532332705  + 9779.35249360890 *T)
    sumY1  +=       0.173 * cos(6.09941552473  + 19651.29229858150 *T)
    sumY1  +=       0.184 * cos(3.08285813784  + 12566.32820716351 *T)
    sumY1  +=       0.230 * cos(0.66694562429  + 12567.37583853451 *T)
    sumY1  +=       0.166 * cos(6.03836479949  + 110.45013870291 *T)
    sumY1  +=       0.189 * cos(1.27307578752  + 3.76693583251 *T)
    sumY1  +=       0.166 * cos(5.90720968173  + 433.95555536030 *T)
    sumY1  +=       0.186 * cos(6.04048637618  + 5863.34738863270 *T)
    sumY1  +=       0.184 * cos(4.58711235161  + 10983.94853421629 *T)
    sumY1  +=       0.184 * cos(1.80970242494  + 6037.48802124550 *T)
    sumY1  +=       0.156 * cos(0.77353737613  + 2648.69864295650 *T)
    sumY1  +=       0.182 * cos(5.04351148530  + 10969.72144021469 *T)
    sumY1  +=       0.175 * cos(1.69691915551  + 6529.15131370430 *T)
    sumY1  +=       0.154 * cos(4.10646056877  + 6453.99253809410 *T)
    sumY1  +=       0.154 * cos(5.68366565747  + 6112.64679685570 *T)
    sumY1  +=       0.164 * cos(2.95193833305  + 12353.09642202830 *T)
    sumY1  +=       0.146 * cos(2.48650028218  + 149.80701461810 *T)
    sumY1  +=       0.184 * cos(4.91752478989  + 11015.35029481830 *T)
    sumY1  +=       0.191 * cos(0.72322692096  + 3894.42564702570 *T)
    sumY1  +=       0.144 * cos(2.19099222002  + 18636.17227201970 *T)
    sumY1  +=       0.146 * cos(0.42084749575  + 13760.84252769090 *T)
    sumY1  +=       0.142 * cos(2.91645242674  + 4156.95462512869 *T)
    sumY1  +=       0.185 * cos(5.29181252606  + 7234.55043875850 *T)
    sumY1  +=       0.133 * cos(4.73063574833  + 18139.53831889940 *T)
    sumY1  +=       0.131 * cos(0.04654287508  + 5507.30942118390 *T)
    sumY1  +=       0.146 * cos(0.32618788162  + 18875.28205229050 *T)
    sumY1  +=       0.128 * cos(4.98427830485  + 6069.53293706989 *T)
    sumY1  +=       0.128 * cos(1.75166251045  + 16200.52890701769 *T)
    sumY1  +=       0.123 * cos(0.73895960068  + 20.53157800890 *T)
    sumY1  +=       0.121 * cos(5.69434162626  + 17256.87535382490 *T)
    sumY1  +=       0.123 * cos(4.81514598285  + 6393.52598869430 *T)
    sumY1  +=       0.123 * cos(4.97498024339  + 6173.11334625550 *T)
    sumY1  +=       0.117 * cos(2.90632080371  + 18852.99448580670 *T)
    sumY1  +=       0.120 * cos(5.75152020178  + 9381.20349020070 *T)
    sumY1  +=       0.113 * cos(3.90786027747  + 65147.37595065419 *T)
    sumY1  +=       0.137 * cos(2.11120201406  + 16201.01654198470 *T)
    sumY1  +=       0.113 * cos(0.36200887356  + 6836.40143535029 *T)
    sumY1  +=       0.123 * cos(2.85759760823  + 12565.41519639810 *T)
    sumY1  +=       0.132 * cos(4.51334720616  + 17655.02435723310 *T)
    sumY1  +=       0.113 * cos(5.69373873614  + 10575.16286545830 *T)
    sumY1  +=       0.113 * cos(3.66776032820  + 10213.04172872750 *T)
    sumY1  +=       0.116 * cos(0.66749075538  + 17253.28492517310 *T)
    sumY1  +=       0.120 * cos(0.70954573147  + 10420.23010099111 *T)
    sumY1  +=       0.109 * cos(5.21836603318  + 633.02755679670 *T)
    sumY1  +=       0.103 * cos(4.83071292354  + 6295.04902031090 *T)
    sumY1  +=       0.103 * cos(4.95941330270  + 6271.59031463890 *T)
    sumY1  +=       0.104 * cos(4.01023132524  + 23013.78335707070 *T)
    sumY1  +=       0.105 * cos(5.32684825848  + 11790.38527117530 *T)
    sumY1  +=       0.109 * cos(4.22545370482  + 3340.36860921629 *T)
    sumY1  +=       0.102 * cos(1.02602291037  + 10970.20907518170 *T)
    sumY1  +=       0.101 * cos(5.99270341494  + 4803.96545844350 *T)
    sumY1  +=       0.114 * cos(3.78830179341  + 4171.18171913030 *T)
    sumY1  +=       0.111 * cos(2.09152457722  + 1582.20316576650 *T)
    sumY1  +=       0.107 * cos(0.90140521385  + 23543.47432216530 *T)
    sumY1  +=       0.099 * cos(4.24784832079  + 10973.31186886650 *T)
    sumY1  +=       0.102 * cos(4.09962394648  + 6549.92670919670 *T)
    sumY1  +=       0.102 * cos(5.69050227976  + 6016.71262575310 *T)
    sumY1  +=       0.120 * cos(0.92616350319  + 11514.12713427790 *T)
    sumY1  +=       0.095 * cos(1.12824235931  + 17267.51201917470 *T)
    sumY1  +=       0.121 * cos(4.76575826483  + 13625.77744765550 *T)
    sumY1  +=       0.100 * cos(3.61435033946  + 10976.83498721549 *T)
    sumY1  +=       0.090 * cos(2.11761227033  + 76.50988875911 *T)
    sumY1  +=       0.100 * cos(5.94744998603  + 8661.99650607950 *T)
    sumY1  +=       0.089 * cos(3.76175372959  + 1350.11122714230 *T)
    sumY1  +=       0.087 * cos(2.76310699450  + 9917.94069199330 *T)
    sumY1  +=       0.086 * cos(4.36870119289  + 12809.12412144031 *T)
    sumY1  +=       0.095 * cos(6.05597457730  + 12345.98287502750 *T)
    sumY1  +=       0.084 * cos(3.11610500480  + 775.27879384050 *T)
    sumY1  +=       0.106 * cos(3.57201004990  + 11216.52810780750 *T)
    sumY1  +=       0.095 * cos(0.92966555257  + 12012.82611462390 *T)
    sumY1  +=       0.082 * cos(4.72839816349  + 5643.42238116090 *T)
    sumY1  +=       0.081 * cos(2.27692529332  + 11614.67711121570 *T)
    sumY1  +=       0.084 * cos(5.85994160066  + 6255.91811137810 *T)
    sumY1  +=       0.084 * cos(3.93018462558  + 6310.72122357170 *T)
    sumY1  +=       0.080 * cos(5.05950625443  + 6923.21695378890 *T)
    sumY1  +=       0.085 * cos(1.08248438849  + 5635.82119919310 *T)
    sumY1  +=       0.084 * cos(4.61502942146  + 10440.03047512009 *T)
    sumY1  +=       0.080 * cos(2.62622746074  + 3583.58484815730 *T)
    sumY1  +=       0.092 * cos(5.59171891913  + 10818.37910439930 *T)
    sumY1  +=       0.077 * cos(4.13228553675  + 6993.25271603320 *T)
    sumY1  +=       0.080 * cos(6.08197788088  + 13521.99525907490 *T)
    sumY1  +=       0.104 * cos(0.70806546737  + 5650.04829319470 *T)
    sumY1  +=       0.093 * cos(5.00569753414  + 4732.27444482690 *T)
    sumY1  +=       0.085 * cos(1.74569264442  + 22805.49174851010 *T)
    sumY1  +=       0.086 * cos(5.52511956951  + 12360.20996902910 *T)
    sumY1  +=       0.075 * cos(5.57236272383  + 5120.35732810009 *T)
    sumY1  +=       0.090 * cos(5.33797857569  + 6370.62787201471 *T)
    sumY1  +=       0.090 * cos(4.45214765055  + 6196.01146293510 *T)
    sumY1  +=       0.082 * cos(4.91006846138  + 18842.35782045690 *T)
    sumY1  +=       0.101 * cos(1.37756832018  + 3097.64000524229 *T)
    sumY1  +=       0.094 * cos(0.74657649932  + 10177.01386205009 *T)
    sumY1  +=       0.083 * cos(3.74913323692  + 18415.75962958090 *T)
    sumY1  +=       0.073 * cos(4.52746639645  + 949.41942645330 *T)
    sumY1  +=       0.101 * cos(4.43327916798  + 3104.68624194030 *T)
    sumY1  +=       0.074 * cos(1.25396392931  + 13517.62628874989 *T)
    sumY1  +=       0.092 * cos(0.77959714125  + 16858.72635041670 *T)
    sumY1  +=       0.075 * cos(1.57278335709  + 1059.13811270570 *T)
    sumY1  +=       0.085 * cos(4.52163234616  + 398.39282089170 *T)
    sumY1  +=       0.069 * cos(5.60384680292  + 11713.19913571430 *T)
    sumY1  +=       0.084 * cos(5.92557534615  + 2378.92065608810 *T)
    sumY1  +=       0.069 * cos(2.18342250812  + 11712.71150074729 *T)
    sumY1  +=       0.075 * cos(0.05753316454  + 30356.24113725130 *T)
    sumY1  +=       0.067 * cos(3.04134731899  + 3154.93090237910 *T)
    sumY1  +=       0.076 * cos(3.26401364210  + 18429.98672358250 *T)
    sumY1  +=       0.068 * cos(2.69185508250  + 11925.51791008410 *T)
    sumY1  +=       0.067 * cos(4.81689271569  + 10454.25756912169 *T)
    sumY1  +=       0.073 * cos(1.35808826471  + 15670.83794192309 *T)
    sumY1  +=       0.065 * cos(1.77865763794  + 6438.25243194210 *T)
    sumY1  +=       0.069 * cos(1.72420196911  + 17298.42614480970 *T)
    sumY1  +=       0.087 * cos(3.52675260029  + 3904.15519390330 *T)
    sumY1  +=       0.076 * cos(5.44623063701  + 5231.05128428650 *T)
    sumY1  +=       0.064 * cos(2.12314868459  + 3496.27664361750 *T)
    sumY1  +=       0.078 * cos(2.74559935849  + 8672.21368792411 *T)
    sumY1  +=       0.066 * cos(4.98895923052  + 14143.25142494710 *T)
    sumY1  +=       0.063 * cos(5.70776457338  + 24357.02460612510 *T)
    sumY1  +=       0.063 * cos(0.57965828598  + 15720.59496739490 *T)
    sumY1  +=       0.061 * cos(2.15619405920  + 16460.57734700850 *T)
    sumY1  +=       0.073 * cos(6.07488391156  + 9387.76209193169 *T)
    sumY1  +=       0.057 * cos(4.37179192959  + 17259.91083720689 *T)
    sumY1  +=       0.059 * cos(2.26511827996  + 9778.86485864190 *T)
    sumY1  +=       0.058 * cos(1.67874293827  + 155.66421691770 *T)
    sumY1  +=       0.060 * cos(1.94469304669  + 34570.31015233610 *T)
    sumY1  +=       0.057 * cos(2.88994977642  + 149854.64395229140 *T)
    sumY1  +=       0.056 * cos(3.76314176838  + 12456.18919624690 *T)
    sumY1  +=       0.056 * cos(5.62479142768  + 17996.27498570570 *T)
    sumY1  +=       0.056 * cos(0.60886736064  + 11764.57458634250 *T)
    sumY1  +=       0.054 * cos(4.53224684011  + 4705.48849006010 *T)
    sumY1  +=       0.055 * cos(6.11029124003  + 13915.77529215809 *T)
    sumY1  +=       0.061 * cos(4.86430679802  + 28237.47727687290 *T)
    sumY1  +=       0.065 * cos(4.29965528550  + 7335.58805066330 *T)
    sumY1  +=       0.072 * cos(5.48795829629  + 6055.84353468590 *T)
    sumY1  +=       0.072 * cos(4.30216792995  + 6510.79580026390 *T)
    sumY1  +=       0.071 * cos(4.94322519248  + 12545.62012197390 *T)
    sumY1  +=       0.061 * cos(5.98387073766  + 6312.74917601091 *T)
    sumY1  +=       0.061 * cos(3.80625548858  + 6253.89015893890 *T)
    sumY1  +=       0.052 * cos(1.83719780245  + 5326.54287653730 *T)
    sumY1  +=       0.054 * cos(2.78309524301  + 2699.49100183409 *T)
    sumY1  +=       0.052 * cos(0.87738933830  + 8983.05448679250 *T)
    sumY1  +=       0.066 * cos(0.78258344897  + 14919.26167123810 *T)
    sumY1  +=       0.050 * cos(1.22263365304  + 5572.89898394960 *T)
    sumY1  +=       0.051 * cos(2.33500088089  + 16522.41589851870 *T)
    sumY1  +=       0.049 * cos(2.61658441943  + 35579.93505705350 *T)
    sumY1  +=       0.053 * cos(6.07542895582  + 6208.53806890760 *T)
    sumY1  +=       0.053 * cos(3.71469727042  + 6358.10126604220 *T)
    sumY1  +=       0.049 * cos(3.35593265214  + 21393.78578734110 *T)
    sumY1  +=       0.049 * cos(4.47248238697  + 5547.44315394310 *T)
    sumY1  +=       0.049 * cos(5.31764383927  + 7019.19618100671 *T)
    sumY1  +=       0.049 * cos(1.02044068888  + 12416.83232033170 *T)
    sumY1  +=       0.050 * cos(0.93918602686  + 12592.20620229910 *T)
    sumY1  +=       0.050 * cos(4.78759248733  + 26084.26562369970 *T)
    sumY1  +=       0.051 * cos(1.83601012538  + 23006.66981006990 *T)
    sumY1  +=       0.058 * cos(4.03623415117  + 15141.14697682849 *T)
    sumY1  +=       0.047 * cos(5.68903994773  + 24279.35083569710 *T)
    sumY1  +=       0.049 * cos(3.40926939847  + 6816.53375091851 *T)
    sumY1  +=       0.049 * cos(0.09767152059  + 5750.10558403130 *T)
    sumY1  +=       0.046 * cos(1.88054905125  + 19379.16233255230 *T)
    sumY1  +=       0.052 * cos(5.74808775798  + 16737.33341911309 *T)
    sumY1  +=       0.047 * cos(0.52219046486  + 24066.05174025910 *T)
    sumY1  +=       0.046 * cos(4.16717022039  + 104351.85638378028 *T)
    sumY1  +=       0.063 * cos(6.25484174196  + 8662.48414104651 *T)
    sumY1  +=       0.048 * cos(3.28503810088  + 26735.70144472970 *T)
    sumY1  +=       0.046 * cos(3.91775350902  + 12132.68377958950 *T)
    sumY1  +=       0.050 * cos(2.92364175862  + 24602.85625235450 *T)
    sumY1  +=       0.054 * cos(4.75214039499  + 7834.36489012290 *T)
    sumY1  +=       0.045 * cos(3.48807574467  + 161000.92955515761 *T)
    sumY1  +=       0.049 * cos(2.94405201174  + 29303.97275406290 *T)
    sumY1  +=       0.049 * cos(1.43156848803  + 10984.43616918330 *T)
    sumY1  +=       0.045 * cos(0.70883599249  + 45584.92899470390 *T)
    sumY1  +=       0.046 * cos(1.86995211145  + 5244.29305668450 *T)
    sumY1  +=       0.048 * cos(3.53510882628  + 23581.01435983410 *T)
    sumY1  +=       0.049 * cos(2.17177379198  + 19650.80466361450 *T)
    sumY1  +=       0.049 * cos(3.41897587826  + 29289.74566006130 *T)
    sumY1  +=       0.044 * cos(1.44008987426  + 6390.98319141350 *T)
    sumY1  +=       0.044 * cos(2.06685104480  + 6175.65614353630 *T)
    sumY1  +=       0.051 * cos(2.65580143245  + 17789.60180230149 *T)
    sumY1  +=       0.042 * cos(6.18543275273  + 24705.94902657310 *T)
    sumY1  +=       0.042 * cos(5.90805829894  + 15663.79170522509 *T)
    sumY1  +=       0.043 * cos(3.08250524084  + 16460.08971204149 *T)
    sumY1  +=       0.043 * cos(0.28590581328  + 553.81322032590 *T)
    sumY1  +=       0.041 * cos(3.21615501236  + 17370.60479339330 *T)
    sumY1  +=       0.053 * cos(2.59151074700  + 13119.47728534169 *T)
    sumY1  +=       0.042 * cos(5.11995884892  + 4164.55580709650 *T)
    sumY1  +=       0.041 * cos(2.05521298551  + 23020.89690407150 *T)
    sumY1  +=       0.045 * cos(3.70227971577  + 54247.16931826690 *T)
    sumY1  +=       0.040 * cos(4.57473494926  + 32243.25468339710 *T)
    sumY1  +=       0.040 * cos(5.17184328434  + 24336.24921063270 *T)
    sumY1  +=       0.040 * cos(1.71852303243  + 11769.60987568290 *T)
    sumY1  +=       0.045 * cos(6.25985265406  + 16723.10632511149 *T)
    sumY1  +=       0.045 * cos(6.25985265406  + 16723.59396007850 *T)
    sumY1  +=       0.042 * cos(5.89327075469  + 8827.14645239129 *T)
    sumY1  +=       0.040 * cos(4.71493835645  + 8402.08352785330 *T)
    sumY1  +=       0.039 * cos(3.49714273023  + 16062.42834360030 *T)
    sumY1  +=       0.044 * cos(4.44510876745  + 5856.23384163189 *T)
    sumY1  +=       0.038 * cos(1.14212165839  + 6297.54676147650 *T)
    sumY1  +=       0.038 * cos(2.36481926067  + 6269.09257347330 *T)
    sumY1  +=       0.041 * cos(3.38233495451  + 15508.85894075790 *T)
    sumY1  +=       0.038 * cos(4.96796930709  + 6394.50630976251 *T)
    sumY1  +=       0.038 * cos(4.82215691915  + 6172.13302518730 *T)
    sumY1  +=       0.039 * cos(1.74815791619  + 29826.55017215670 *T)
    sumY1  +=       0.046 * cos(0.31322632556  + 17297.93850984270 *T)
    sumY1  +=       0.041 * cos(4.04315782590  + 6309.13035230770 *T)
    sumY1  +=       0.038 * cos(5.82409895120  + 29089.05523346850 *T)
    sumY1  +=       0.038 * cos(4.59409111885  + 11933.61177815310 *T)
    sumY1  +=       0.037 * cos(3.34091252113  + 26482.41462710790 *T)
    sumY1  +=       0.037 * cos(3.64722285417  + 20452.62559473830 *T)
    sumY1  +=       0.037 * cos(0.30900035436  + 18073.46112116669 *T)
    sumY1  +=       0.036 * cos(2.98410187235  + 11499.41240530929 *T)
    sumY1  +=       0.045 * cos(4.75649198460  + 3340.85624418330 *T)
    sumY1  +=       0.049 * cos(4.21088151991  + 18216.68762814450 *T)
    sumY1  +=       0.038 * cos(4.59187061808  + 18202.46053414290 *T)
    sumY1  +=       0.041 * cos(3.91005845267  + 5216.82419028490 *T)
    sumY1  +=       0.036 * cos(6.21139146959  + 4060.97539791089 *T)
    sumY1  +=       0.048 * cos(5.14345155225  + 951.96222373410 *T)
    sumY1  +=       0.041 * cos(3.90932016585  + 18848.98373249069 *T)
    sumY1  +=       0.038 * cos(4.60880580452  + 26709.89075989690 *T)
    sumY1  +=       0.036 * cos(3.12521745961  + 28230.43104017490 *T)
    sumY1  +=       0.036 * cos(1.96840086543  + 24491.18197509989 *T)
    sumY1  +=       0.036 * cos(1.56811412304  + 9924.56660402710 *T)
    sumY1  +=       0.048 * cos(1.89701220521  + 8982.56685182550 *T)
    sumY1  +=       0.037 * cos(3.60669363968  + 22484.09239197610 *T)
    sumY1  +=       0.047 * cos(3.21269518092  + 6040.10342853389 *T)
    sumY1  +=       0.034 * cos(3.68972702028  + 6418.94499248490 *T)
    sumY1  +=       0.034 * cos(6.10039920596  + 6147.69434246491 *T)
    sumY1  +=       0.034 * cos(3.67007112439  + 9380.71585523369 *T)
    sumY1  +=       0.036 * cos(0.30403746373  + 7238.43177411650 *T)
    sumY1  +=       0.034 * cos(0.40887602826  + 17272.12782500990 *T)
    sumY1  +=       0.037 * cos(1.23154839984  + 58953.38926077750 *T)
    sumY1  +=       0.034 * cos(0.33427002863  + 29026.72904699130 *T)
    sumY1  +=       0.033 * cos(3.87501941695  + 12721.32828193349 *T)
    sumY1  +=       0.035 * cos(1.63582276543  + 7322.34627826531 *T)
    sumY1  +=       0.044 * cos(1.21937301495  + 3981.73385156551 *T)
    sumY1  +=       0.033 * cos(3.59821221655  + 19804.58347409930 *T)
    sumY1  +=       0.033 * cos(2.56047890182  + 173567.08125514039 *T)
    sumY1  +=       0.033 * cos(6.16657680956  + 205.94173095370 *T)
    sumY1  +=       0.033 * cos(5.44889821497  + 23013.29572210369 *T)
    sumY1  +=       0.034 * cos(2.13261854972  + 49515.13869092350 *T)
    sumY1  +=       0.034 * cos(3.38050467777  + 38500.51984855570 *T)
    sumY1  +=       0.032 * cos(2.81221233882  + 17686.99666304990 *T)
    sumY1  +=       0.036 * cos(1.67423219491  + 5641.95442512570 *T)
    sumY1  +=       0.032 * cos(2.91090812586  + 21228.63584102931 *T)
    sumY1  +=       0.033 * cos(0.13839577785  + 34520.55312686430 *T)
    sumY1  +=       0.033 * cos(5.65380556855  + 8635.69818627969 *T)
    sumY1  +=       0.031 * cos(1.88042187886  + 5746.02752041250 *T)
    sumY1  +=       0.041 * cos(2.52021697492  + 21202.33752122950 *T)
    sumY1  +=       0.037 * cos(5.92836585991  + 7349.81514466491 *T)
    sumY1  +=       0.034 * cos(2.18099177342  + 16496.11757871889 *T)
    sumY1  +=       0.032 * cos(0.90143427409  + 33019.26492968810 *T)
    sumY1  +=       0.032 * cos(1.54173979189  + 21150.56954840009 *T)
    sumY1  +=       0.037 * cos(1.84830381005  + 29296.85920706210 *T)
    sumY1  +=       0.036 * cos(5.02250294215  + 27511.22405605370 *T)
    sumY1  +=       0.035 * cos(3.77934660139  + 11933.12414318609 *T)
    sumY1  +=       0.035 * cos(4.26341483447  + 11918.89704918449 *T)
    sumY1  +=       0.035 * cos(5.54313072444  + 29062.75691366870 *T)
    sumY1  +=       0.030 * cos(1.38932865248  + 12490.12944619070 *T)
    sumY1  +=       0.032 * cos(5.09927453137  + 6334.60000533731 *T)
    sumY1  +=       0.032 * cos(4.69085169487  + 6232.03932961249 *T)
    sumY1  +=       0.029 * cos(4.05993716780  + 419.24082639170 *T)
    sumY1  +=       0.029 * cos(0.91834451421  + 419.72846135871 *T)
    sumY1  +=       0.029 * cos(1.66178131503  + 16858.23871544969 *T)
    sumY1  +=       0.030 * cos(2.73190240418  + 16208.13008898551 *T)
    sumY1  +=       0.029 * cos(0.69175314473  + 27278.71263392430 *T)
    sumY1  +=       0.032 * cos(4.73525892997  + 5017.26455388150 *T)
    sumY1  +=       0.028 * cos(0.30767945102  + 24080.27883426070 *T)
    sumY1  +=       0.030 * cos(1.00265951424  + 23539.95120381630 *T)
    sumY1  +=       0.029 * cos(1.45883005475  + 18606.25512851669 *T)
    sumY1  +=       0.030 * cos(4.73151177298  + 30665.91114094930 *T)
    sumY1  +=       0.034 * cos(4.29199382621  + 6660.69327539070 *T)
    sumY1  +=       0.034 * cos(5.49813240003  + 5905.94605955911 *T)
    sumY1  +=       0.038 * cos(3.55143054136  + 31571.04346687470 *T)
    sumY1  +=       0.038 * cos(4.53015131722  + 16061.94070863329 *T)
    sumY1  +=       0.029 * cos(1.64885508302  + 23539.46356884929 *T)
    sumY1  +=       0.033 * cos(2.88700963334  + 3738.51761262449 *T)
    sumY1  +=       0.028 * cos(6.02264922068  + 2942.21960580809 *T)
    sumY1  +=       0.032 * cos(2.75250951602  + 15110.70993734970 *T)
    sumY1  +=       0.030 * cos(6.10434010761  + 55023.17956455790 *T)
    sumY1  +=       0.027 * cos(5.24524976767  + 233141.55822184496 *T)
    sumY1  +=       0.030 * cos(0.75326086687  + 12587.17091295870 *T)
    sumY1  +=       0.028 * cos(3.97691648841  + 34911.65589357450 *T)
    sumY1  +=       0.027 * cos(4.60109278395  + 1692.40948698591 *T)
    sumY1  +=       0.027 * cos(6.13743188286  + 2544.07060239989 *T)
    sumY1  +=       0.027 * cos(4.99580076168  + 24382.83529095790 *T)
    sumY1  +=       0.027 * cos(4.42136306455  + 6205.64589704690 *T)
    sumY1  +=       0.027 * cos(5.36876316169  + 6360.99343790291 *T)
    sumY1  +=       0.027 * cos(4.95719787895  + 31172.89446346650 *T)
    sumY1  +=       0.026 * cos(2.19655051246  + 36949.47462590770 *T)
    sumY1  +=       0.034 * cos(4.45594644643  + 18053.17336064130 *T)
    sumY1  +=       0.027 * cos(0.74968073356  + 32367.34147369110 *T)
    sumY1  +=       0.027 * cos(3.34260766105  + 4487.57358878689 *T)
    sumY1  +=       0.028 * cos(5.37411547339  + 10239.34004852730 *T)
    sumY1  +=       0.026 * cos(1.30232025786  + 4796.85191144269 *T)
    sumY1  +=       0.026 * cos(4.70861603764  + 6226.42129253930 *T)
    sumY1  +=       0.026 * cos(5.08151018860  + 6340.21804241050 *T)
    sumY1  +=       0.029 * cos(4.90026899469  + 5974.04134481910 *T)
    sumY1  +=       0.029 * cos(4.88985723155  + 6592.59799013070 *T)
    sumY1  +=       0.026 * cos(4.96950484222  + 19402.55313533309 *T)
    sumY1  +=       0.028 * cos(3.99925126629  + 46386.74992582770 *T)
    sumY1  +=       0.032 * cos(5.04462761864  + 9225.78309076650 *T)
    sumY1  +=       0.026 * cos(4.11745014890  + 9910.33951002549 *T)
    sumY1  +=       0.025 * cos(1.02839985963  + 15265.64270181689 *T)
    sumY1  +=       0.029 * cos(4.33416022359  + 52176.05010063190 *T)
    sumY1  +=       0.027 * cos(2.65258449905  + 62883.59895699710 *T)
    sumY1  +=       0.024 * cos(5.26265995716  + 11617.21990849651 *T)
    sumY1  +=       0.024 * cos(1.04936648816  + 5120.84496306710 *T)
    sumY1  +=       0.028 * cos(4.82143927810  + 6680.98103591609 *T)
    sumY1  +=       0.024 * cos(5.97769334002  + 522.33360061030 *T)
    sumY1  +=       0.025 * cos(2.29282659231  + 6263.64990657511 *T)
    sumY1  +=       0.025 * cos(1.21411432675  + 6302.98942837470 *T)
    sumY1  +=       0.029 * cos(0.87325499796  + 23581.50199480110 *T)
    sumY1  +=       0.024 * cos(5.21948189448  + 29424.87805039950 *T)
    sumY1  +=       0.028 * cos(1.09161353868  + 16737.82105408010 *T)
    sumY1  +=       0.024 * cos(3.55960074091  + 22743.16556203289 *T)
    sumY1  +=       0.024 * cos(3.55960074091  + 22743.65319699990 *T)
    sumY1  +=       0.024 * cos(4.28222488832  + 25158.84553724890 *T)
    sumY1  +=       0.025 * cos(5.52194731109  + 21424.22282681990 *T)
    sumY1  +=       0.025 * cos(5.52194731109  + 21424.71046178690 *T)
    sumY1  +=       0.024 * cos(3.49943174674  + 14712.07329897449 *T)
    sumY1  +=       0.028 * cos(6.07681616993  + 5760.25461441409 *T)
    sumY1  +=       0.024 * cos(2.16376008059  + 31969.19247028290 *T)
    sumY1  +=       0.023 * cos(5.33584675127  + 17893.18221148710 *T)
    sumY1  +=       0.030 * cos(2.28483676839  + 8584.90548338430 *T)
    sumY1  +=       0.024 * cos(1.64666673478  + 23550.58786916610 *T)
    sumY1  +=       0.023 * cos(6.07302432574  + 6717.03140535170 *T)
    sumY1  +=       0.022 * cos(2.42662159283  + 7445.79437188270 *T)
    sumY1  +=       0.023 * cos(3.56571153376  + 5849.60792959810 *T)
    sumY1  +=       0.025 * cos(5.52794317615  + 14155.46841275010 *T)
    sumY1  +=       0.023 * cos(2.01489779401  + 6294.36536773881 *T)
    sumY1  +=       0.023 * cos(1.49204312505  + 6272.27396721100 *T)
    sumY1  +=       0.024 * cos(5.81677586075  + 18208.10612510850 *T)
    sumY1  +=       0.024 * cos(5.81677586075  + 18208.59376007550 *T)
    sumY1  +=       0.021 * cos(4.21172313283  + 22779.68106367730 *T)
    sumY1  +=       0.029 * cos(1.74369099855  + 28628.58004358310 *T)
    sumY1  +=       0.021 * cos(3.56347944199  + 40879.68432212730 *T)
    sumY1  +=       0.021 * cos(2.51496251708  + 30221.17605721590 *T)
    sumY1  +=       0.022 * cos(6.09500922463  + 4379.88285497370 *T)
    sumY1  +=       0.022 * cos(3.69512347867  + 8186.75647997610 *T)
    sumY1  +=       0.022 * cos(2.95145769488  + 28760.05469496669 *T)
    sumY1  +=       0.026 * cos(1.83973482412  + 18422.87317658170 *T)
    sumY1  +=       0.020 * cos(2.37799900039  + 12036.21691740469 *T)
    sumY1  +=       0.020 * cos(0.07485283254  + 28313.04498717750 *T)
    sumY1  +=       0.027 * cos(5.38239668902  + 34513.50689016630 *T)
    sumY1  +=       0.020 * cos(0.50551300300  + 34115.35788675810 *T)
    sumY1  +=       0.020 * cos(3.29228335956  + 21548.71855180830 *T)
    sumY1  +=       0.020 * cos(5.25656062979  + 17654.53672226610 *T)
    sumY1  +=       0.025 * cos(1.96130235408  + 55798.70217588190 *T)
    sumY1  +=       0.020 * cos(5.53200449549  + 20597.48778052470 *T)
    sumY1  +=       0.020 * cos(1.03604717996  + 18875.76968725751 *T)
    sumY1  +=       0.022 * cos(2.01853988925  + 1588.82907780029 *T)
    sumY1  +=       0.024 * cos(5.92091322542  + 9411.70843257070 *T)
    sumY1  +=       0.019 * cos(1.14706242910  + 536.56069461190 *T)
    sumY1  +=       0.020 * cos(3.42481272827  + 6187.34044025710 *T)
    sumY1  +=       0.020 * cos(0.08212819079  + 6379.29889469270 *T)
    sumY1  +=       0.022 * cos(2.87392170851  + 18326.89394936390 *T)
    sumY1  +=       0.022 * cos(0.61268580951  + 6233.56264200310 *T)
    sumY1  +=       0.022 * cos(2.89425510955  + 6333.07669294670 *T)
    sumY1  +=       0.018 * cos(0.45956662142  + 5972.47886860650 *T)
    sumY1  +=       0.018 * cos(3.04737429764  + 6594.16046634330 *T)
    sumY1  +=       0.019 * cos(2.71317971903  + 6948.07571260490 *T)
    sumY1  +=       0.019 * cos(0.79376120003  + 5618.56362234490 *T)
    sumY1  +=       0.019 * cos(4.89700264579  + 83974.07917321339 *T)
    sumY1  +=       0.019 * cos(2.72494811319  + 84020.10309797739 *T)
    sumY1  +=       0.020 * cos(5.42036659910  + 50316.95962204730 *T)
    sumY1  +=       0.020 * cos(1.28312164242  + 9070.36269133230 *T)

    # EARTH Y Coord
    sumY2  = 0
    sumY2  +=   52911.498 * cos(1.75323722235  + 6283.31966747490 *T)
    sumY2  +=    6074.441 * cos(0.48406211192  + 0.24381748350 *T)
    sumY2  +=    2406.871 * cos(2.66006394470  + 12566.39551746630 *T)
    sumY2  +=      96.033 * cos(2.57406136836  + 18849.47136745770 *T)
    sumY2  +=      29.888 * cos(2.30515825272  + 6282.83203250789 *T)
    sumY2  +=      14.021 * cos(1.64042051619  + 6279.79654912590 *T)
    sumY2  +=      13.375 * cos(1.74320372785  + 6286.84278582391 *T)
    sumY2  +=       8.148 * cos(0.62869530307  + 6309.61798727471 *T)
    sumY2  +=       8.195 * cos(2.81874475342  + 6257.02134767510 *T)
    sumY2  +=       8.001 * cos(4.34877296703  + 6127.89926804070 *T)
    sumY2  +=       7.754 * cos(5.43470598991  + 6438.74006690910 *T)
    sumY2  +=       7.223 * cos(3.81034638754  + 83997.09113559539 *T)
    sumY2  +=       4.616 * cos(4.59623594691  + 710.17686604180 *T)
    sumY2  +=       3.060 * cos(3.26884940633  + 4705.97612502710 *T)
    sumY2  +=       3.717 * cos(0.87961276723  + 5507.79705615090 *T)
    sumY2  +=       2.991 * cos(2.48913608928  + 25132.54721744910 *T)
    sumY2  +=       3.235 * cos(2.53504921673  + 7058.84227879890 *T)
    sumY2  +=       2.393 * cos(1.83696733237  + 775.76642880750 *T)
    sumY2  +=       2.034 * cos(2.77061597204  + 6290.43321447570 *T)
    sumY2  +=       2.028 * cos(0.70943948464  + 6276.20612047410 *T)
    sumY2  +=       2.064 * cos(0.01448293997  + 7860.66320992270 *T)
    sumY2  +=       1.914 * cos(2.85234479868  + 5487.02166065850 *T)
    sumY2  +=       1.770 * cos(3.48783151064  + 1059.62574767270 *T)
    sumY2  +=       1.761 * cos(1.80823220260  + 5753.62870238030 *T)
    sumY2  +=       1.558 * cos(0.56580595512  + 7079.61767429131 *T)
    sumY2  +=       1.551 * cos(0.55852743151  + 6040.59106350090 *T)
    sumY2  +=       1.436 * cos(1.28164050883  + 5885.17066406670 *T)
    sumY2  +=       1.469 * cos(2.94324921966  + 6526.04827144891 *T)
    sumY2  +=       1.319 * cos(1.89155311923  + 6681.46867088311 *T)
    sumY2  +=       1.180 * cos(1.42090963151  + 6813.01063256950 *T)
    sumY2  +=       1.337 * cos(5.75221416852  + 5729.75026463250 *T)
    sumY2  +=       1.039 * cos(4.25243426113  + 6282.33934640670 *T)
    sumY2  +=       1.313 * cos(0.59897551240  + 529.93478257810 *T)
    sumY2  +=       1.041 * cos(5.47801377680  + 6284.29998854310 *T)
    sumY2  +=       1.202 * cos(4.02829547071  + 6836.88907031730 *T)
    sumY2  +=       0.904 * cos(0.85768945907  + 5331.60126122430 *T)
    sumY2  +=       0.870 * cos(2.26660468565  + 4933.45225781610 *T)
    sumY2  +=       1.149 * cos(3.02450680654  + 167284.00540514899 *T)
    sumY2  +=       1.069 * cos(1.09203232944  + 1577.09972496430 *T)
    sumY2  +=       0.772 * cos(2.66290097432  + 7235.03807372550 *T)
    sumY2  +=       0.846 * cos(1.36798976406  + 11790.87290614230 *T)
    sumY2  +=       0.742 * cos(5.02007434299  + 4137.15425099970 *T)
    sumY2  +=       0.966 * cos(2.05227643484  + 11856.46246890800 *T)
    sumY2  +=       0.664 * cos(1.21776131888  + 7633.18707713370 *T)
    sumY2  +=       0.662 * cos(3.18968574411  + 1194.69082770810 *T)
    sumY2  +=       0.659 * cos(5.95013449282  + 11507.01358727710 *T)
    sumY2  +=       0.853 * cos(3.65338850674  + 4535.30325440790 *T)
    sumY2  +=       0.823 * cos(1.06126729936  + 155.17658195069 *T)
    sumY2  +=       0.783 * cos(4.30296275012  + 5088.87265725030 *T)
    sumY2  +=       0.785 * cos(3.46658763633  + 2353.10997125530 *T)
    sumY2  +=       0.694 * cos(3.50785804761  + 801.57711364030 *T)
    sumY2  +=       0.566 * cos(4.42922662864  + 796.54182429990 *T)
    sumY2  +=       0.684 * cos(1.11267726523  + 3128.63258257930 *T)
    sumY2  +=       0.499 * cos(1.42594865875  + 12569.91863581531 *T)
    sumY2  +=       0.596 * cos(6.09778115923  + 8031.33608054190 *T)
    sumY2  +=       0.486 * cos(4.54237345712  + 1589.31671276730 *T)
    sumY2  +=       0.483 * cos(4.68204073064  + 8429.48508395010 *T)
    sumY2  +=       0.441 * cos(4.99112466662  + 6496.61876291290 *T)
    sumY2  +=       0.412 * cos(0.87602065813  + 4693.75913722409 *T)
    sumY2  +=       0.561 * cos(5.46330573132  + 7477.76667769950 *T)
    sumY2  +=       0.390 * cos(4.63888431687  + 6070.02057203690 *T)
    sumY2  +=       0.405 * cos(6.26538093406  + 3739.00524759150 *T)
    sumY2  +=       0.501 * cos(1.82149883986  + 1592.83983111630 *T)
    sumY2  +=       0.392 * cos(0.68216396598  + 26.05450231630 *T)
    sumY2  +=       0.369 * cos(5.93410776393  + 553.32558535889 *T)
    sumY2  +=       0.483 * cos(6.07777893059  + 12036.70455237170 *T)
    sumY2  +=       0.489 * cos(3.70668603785  + 5223.45010231870 *T)
    sumY2  +=       0.481 * cos(3.17195852109  + 10213.52936369450 *T)
    sumY2  +=       0.448 * cos(5.56847234629  + 156137.71980228281 *T)
    sumY2  +=       0.321 * cos(4.53991511859  + 951.47458876710 *T)
    sumY2  +=       0.320 * cos(0.43138538654  + 1990.98883452450 *T)
    sumY2  +=       0.361 * cos(2.16599134100  + 12565.90788249930 *T)
    sumY2  +=       0.361 * cos(4.79823251497  + 12721.81591690050 *T)
    sumY2  +=       0.412 * cos(4.34259218946  + 398.39282089170 *T)
    sumY2  +=       0.304 * cos(5.68078111352  + 4690.72365384210 *T)
    sumY2  +=       0.356 * cos(6.00094576166  + 5481.49873635110 *T)
    sumY2  +=       0.283 * cos(0.66912545661  + 242.97242145751 *T)
    sumY2  +=       0.362 * cos(2.31554210740  + 9438.00675237050 *T)
    sumY2  +=       0.248 * cos(3.18243221495  + 8827.63408735830 *T)
    sumY2  +=       0.252 * cos(2.73244620528  + 3154.44326741210 *T)
    sumY2  +=       0.257 * cos(0.22372520009  + 11371.94850724170 *T)
    sumY2  +=       0.327 * cos(1.45001600721  + 14143.73905991410 *T)
    sumY2  +=       0.230 * cos(1.80447328493  + 1747.77259558350 *T)
    sumY2  +=       0.289 * cos(3.67617510107  + 7085.14059859870 *T)
    sumY2  +=       0.297 * cos(3.17045499445  + 1349.62359217529 *T)
    sumY2  +=       0.209 * cos(4.98580572759  + 11770.09751064990 *T)
    sumY2  +=       0.207 * cos(0.21623312501  + 4164.06817212950 *T)
    sumY2  +=       0.188 * cos(4.07593821324  + 7875.91568110771 *T)
    sumY2  +=       0.189 * cos(5.31784128034  + 2389.13783793270 *T)
    sumY2  +=       0.194 * cos(5.12285857348  + 10977.32262218251 *T)
    sumY2  +=       0.176 * cos(2.51491286893  + 7342.70159766410 *T)
    sumY2  +=       0.164 * cos(1.22070775546  + 5223.93773728571 *T)
    sumY2  +=       0.205 * cos(1.64290799201  + 10973.79950383350 *T)
    sumY2  +=       0.205 * cos(0.68022152849  + 13368.21644859010 *T)
    sumY2  +=       0.139 * cos(3.04789845179  + 10575.65050042530 *T)
    sumY2  +=       0.155 * cos(4.24333188820  + 12410.97511803210 *T)
    sumY2  +=       0.152 * cos(4.81739365185  + 3930.45351370310 *T)
    sumY2  +=       0.129 * cos(0.80954637051  + 397.90518592470 *T)
    sumY2  +=       0.127 * cos(5.67893828134  + 12592.69383726610 *T)
    sumY2  +=       0.132 * cos(4.87291664360  + 8636.18582124671 *T)
    sumY2  +=       0.122 * cos(3.21278850975  + 13119.96492030870 *T)
    sumY2  +=       0.163 * cos(2.50359725386  + 5507.30942118390 *T)
    sumY2  +=       0.114 * cos(4.13352060999  + 17260.39847217390 *T)
    sumY2  +=       0.118 * cos(0.80775184774  + 4292.57465043390 *T)
    sumY2  +=       0.148 * cos(4.06616105867  + 17790.08943726851 *T)
    sumY2  +=       0.123 * cos(2.18706277371  + 12562.87239911730 *T)
    sumY2  +=       0.103 * cos(1.79915109205  + 13518.11392371690 *T)
    sumY2  +=       0.138 * cos(2.26785827403  + 12168.24651405810 *T)
    sumY2  +=       0.107 * cos(2.76898193474  + 10989.05197501850 *T)
    sumY2  +=       0.104 * cos(0.56758990747  + 2145.92159899169 *T)
    sumY2  +=       0.096 * cos(0.40713285174  + 6283.25235717211 *T)
    sumY2  +=       0.099 * cos(0.70751931715  + 13096.08648256090 *T)
    sumY2  +=       0.096 * cos(6.24621518908  + 6283.38697777770 *T)
    sumY2  +=       0.084 * cos(3.63516988010  + 16730.70750707931 *T)
    sumY2  +=       0.111 * cos(2.39242858566  + 12540.09719766650 *T)
    sumY2  +=       0.087 * cos(4.45435464785  + 10177.50149701710 *T)
    sumY2  +=       0.078 * cos(2.61342989032  + 12323.66691349230 *T)
    sumY2  +=       0.076 * cos(4.13782411870  + 6709.91785835091 *T)
    sumY2  +=       0.105 * cos(1.96765772501  + 5642.44206009270 *T)
    sumY2  +=       0.073 * cos(5.60932378402  + 5856.72147659890 *T)
    sumY2  +=       0.102 * cos(5.37997051996  + 250570.91967470257 *T)
    sumY2  +=       0.070 * cos(0.51266920771  + 5753.14106741329 *T)
    sumY2  +=       0.071 * cos(5.27404987073  + 14314.41193053330 *T)
    sumY2  +=       0.092 * cos(0.39406209616  + 13916.26292712510 *T)
    sumY2  +=       0.070 * cos(4.10529685600  + 7084.65296363170 *T)
    sumY2  +=       0.091 * cos(1.53880932897  + 6924.19727485710 *T)
    sumY2  +=       0.083 * cos(2.41910116753  + 31415.62306744050 *T)
    sumY2  +=       0.088 * cos(0.30300511757  + 71430.45180064559 *T)
    sumY2  +=       0.074 * cos(3.53618023271  + 529.44714761109 *T)
    sumY2  +=       0.085 * cos(4.40773297118  + 95143.37673846159 *T)
    sumY2  +=       0.061 * cos(2.66034748981  + 8274.06468451590 *T)
    sumY2  +=       0.060 * cos(3.95075537336  + 16496.60521368590 *T)
    sumY2  +=       0.058 * cos(3.04970743906  + 5642.93474619389 *T)
    sumY2  +=       0.078 * cos(0.84061745785  + 213.54291292150 *T)
    sumY2  +=       0.059 * cos(3.10446906052  + 640.14110379750 *T)
    sumY2  +=       0.057 * cos(3.63209087311  + 4690.23601887509 *T)
    sumY2  +=       0.056 * cos(2.23091315346  + 13341.91812879030 *T)
    sumY2  +=       0.065 * cos(4.95952410600  + 4731.78680985990 *T)
    sumY2  +=       0.051 * cos(0.09929724563  + 3893.93801205869 *T)
    sumY2  +=       0.050 * cos(2.60290043400  + 12573.50906446710 *T)
    sumY2  +=       0.050 * cos(5.98749446465  + 3634.37720703489 *T)
    sumY2  +=       0.049 * cos(0.70632876133  + 18319.78040236310 *T)
    sumY2  +=       0.066 * cos(3.92505951821  + 2787.28684134090 *T)
    sumY2  +=       0.049 * cos(0.42825389697  + 13362.69352428270 *T)
    sumY2  +=       0.048 * cos(2.19153498916  + 12964.54452087450 *T)
    sumY2  +=       0.053 * cos(2.55294637714  + 14919.26167123810 *T)
    sumY2  +=       0.058 * cos(5.94322042199  + 1577.58735993130 *T)
    sumY2  +=       0.052 * cos(5.99067773035  + 15721.08260236190 *T)
    sumY2  +=       0.061 * cos(4.99956015956  + 4292.08701546690 *T)
    sumY2  +=       0.044 * cos(5.02285128905  + 24073.16528725990 *T)
    sumY2  +=       0.044 * cos(1.96034093811  + 10447.14402212089 *T)
    sumY2  +=       0.042 * cos(2.52906622396  + 20426.81490990550 *T)
    sumY2  +=       0.059 * cos(2.58678833053  + 84672.23202702119 *T)
    sumY2  +=       0.051 * cos(3.81200106179  + 6370.62787201471 *T)
    sumY2  +=       0.051 * cos(5.97812516445  + 6196.01146293510 *T)
    sumY2  +=       0.046 * cos(5.23410298912  + 2119.00767786191 *T)
    sumY2  +=       0.041 * cos(2.52248241350  + 3185.43584474911 *T)
    sumY2  +=       0.056 * cos(4.84268946755  + 10976.83498721549 *T)
    sumY2  +=       0.051 * cos(3.94604483748  + 9437.51911740350 *T)
    sumY2  +=       0.049 * cos(1.64079935270  + 239424.63407183639 *T)
    sumY2  +=       0.036 * cos(1.49154565635  + 3495.78900865049 *T)
    sumY2  +=       0.044 * cos(2.12316916648  + 18139.53831889940 *T)
    sumY2  +=       0.035 * cos(5.18466509768  + 3929.96587873609 *T)
    sumY2  +=       0.044 * cos(3.23094444076  + 8661.99650607950 *T)
    sumY2  +=       0.046 * cos(2.21535532938  + 3894.42564702570 *T)
    sumY2  +=       0.034 * cos(2.24364878116  + 6262.54427198250 *T)
    sumY2  +=       0.034 * cos(4.40970135798  + 6304.09506296730 *T)
    sumY2  +=       0.035 * cos(5.15488505358  + 18875.28205229050 *T)
    sumY2  +=       0.035 * cos(4.51996316476  + 10420.23010099111 *T)
    sumY2  +=       0.032 * cos(5.33774590597  + 23013.78335707070 *T)
    sumY2  +=       0.034 * cos(2.04067717162  + 7.35736448430 *T)
    sumY2  +=       0.045 * cos(6.20022544523  + 3340.85624418330 *T)
    sumY2  +=       0.035 * cos(1.95669709650  + 10213.04172872750 *T)
    sumY2  +=       0.031 * cos(3.81393239996  + 14712.56093394150 *T)
    sumY2  +=       0.030 * cos(2.82636067861  + 12809.12412144031 *T)
    sumY2  +=       0.039 * cos(5.86021469295  + 9779.35249360890 *T)
    sumY2  +=       0.041 * cos(0.90085517811  + 82576.73740351178 *T)
    sumY2  +=       0.031 * cos(5.47911308336  + 22779.68106367730 *T)
    sumY2  +=       0.028 * cos(2.50895957020  + 2942.70724077510 *T)
    sumY2  +=       0.028 * cos(3.93140481871  + 5429.63565075589 *T)
    sumY2  +=       0.027 * cos(5.61355223000  + 11014.86265985130 *T)
    sumY2  +=       0.027 * cos(0.53944014317  + 9623.93209417470 *T)
    sumY2  +=       0.025 * cos(0.15499921281  + 18209.57408114370 *T)
    sumY2  +=       0.026 * cos(0.97984555070  + 9381.20349020070 *T)
    sumY2  +=       0.025 * cos(1.11733310909  + 3583.58484815730 *T)
    sumY2  +=       0.030 * cos(4.41826131323  + 10447.63165708790 *T)
    sumY2  +=       0.032 * cos(5.65959938774  + 796.05418933290 *T)
    sumY2  +=       0.023 * cos(1.38575591729  + 9917.45305702629 *T)
    sumY2  +=       0.026 * cos(5.61732376806  + 12012.82611462390 *T)
    sumY2  +=       0.024 * cos(3.45407340381  + 14143.25142494710 *T)
    sumY2  +=       0.023 * cos(0.69647593831  + 26709.89075989690 *T)
    sumY2  +=       0.021 * cos(5.43974830479  + 7632.69944216670 *T)
    sumY2  +=       0.021 * cos(0.80058820514  + 17256.87535382490 *T)
    sumY2  +=       0.028 * cos(5.78689433830  + 12353.09642202830 *T)
    sumY2  +=       0.022 * cos(3.26384247748  + 77714.01528560399 *T)
    sumY2  +=       0.021 * cos(4.41664687493  + 6993.25271603320 *T)
    sumY2  +=       0.020 * cos(1.88345295944  + 6836.40143535029 *T)
    sumY2  +=       0.020 * cos(3.87252932348  + 1748.26023055050 *T)
    sumY2  +=       0.026 * cos(0.45595739946  + 9225.78309076650 *T)
    sumY2  +=       0.023 * cos(5.52071600644  + 149854.64395229140 *T)
    sumY2  +=       0.026 * cos(0.60119736394  + 2544.07060239989 *T)
    sumY2  +=       0.018 * cos(0.71176167314  + 11614.67711121570 *T)
    sumY2  +=       0.020 * cos(3.04333652471  + 426.84200835950 *T)
    sumY2  +=       0.019 * cos(2.09021029647  + 30640.10045611650 *T)
    sumY2  +=       0.018 * cos(3.20715016441  + 155.66421691770 *T)
    sumY2  +=       0.018 * cos(0.32502884515  + 11926.49823115230 *T)
    sumY2  +=       0.022 * cos(2.17863705978  + 17298.42614480970 *T)
    sumY2  +=       0.019 * cos(4.82590950931  + 18073.46112116669 *T)
    sumY2  +=       0.020 * cos(1.02950628439  + 19651.29229858150 *T)
    sumY2  +=       0.024 * cos(3.14467908991  + 5856.23384163189 *T)
    sumY2  +=       0.023 * cos(3.21288696705  + 72850.80553272919 *T)
    sumY2  +=       0.017 * cos(2.38623798159  + 8983.05448679250 *T)
    sumY2  +=       0.016 * cos(3.59989563011  + 16460.57734700850 *T)
    sumY2  +=       0.016 * cos(2.77179277240  + 17259.91083720689 *T)
    sumY2  +=       0.021 * cos(2.20286926080  + 16858.72635041670 *T)
    sumY2  +=       0.020 * cos(5.10666464640  + 23543.47432216530 *T)
    sumY2  +=       0.021 * cos(5.36525328799  + 13367.72881362310 *T)
    sumY2  +=       0.020 * cos(2.47582665567  + 2146.40923395870 *T)
    sumY2  +=       0.014 * cos(4.85508264889  + 22484.09239197610 *T)
    sumY2  +=       0.014 * cos(6.23679853613  + 16200.52890701769 *T)
    sumY2  +=       0.015 * cos(2.64768344463  + 20199.33877711650 *T)
    sumY2  +=       0.019 * cos(1.24340851761  + 8672.21368792411 *T)
    sumY2  +=       0.019 * cos(2.89835524904  + 18073.94875613370 *T)
    sumY2  +=       0.013 * cos(3.14350306699  + 16201.01654198470 *T)
    sumY2  +=       0.017 * cos(2.96004801328  + 7860.17557495569 *T)
    sumY2  +=       0.013 * cos(3.12347768998  + 18451.32236404950 *T)
    sumY2  +=       0.016 * cos(4.90276494200  + 5572.89898394960 *T)
    sumY2  +=       0.012 * cos(1.13909733386  + 17996.27498570570 *T)
    sumY2  +=       0.012 * cos(5.43460350187  + 19403.04077030010 *T)
    sumY2  +=       0.012 * cos(0.82655936469  + 14945.07235607090 *T)
    sumY2  +=       0.015 * cos(0.47799392266  + 12559.28197046550 *T)
    sumY2  +=       0.012 * cos(2.80799572996  + 10818.37910439930 *T)
    sumY2  +=       0.015 * cos(5.28864619208  + 12567.37583853451 *T)
    sumY2  +=       0.011 * cos(0.21391675163  + 28767.16824196750 *T)
    sumY2  +=       0.013 * cos(5.80786469706  + 90280.16698558679 *T)
    sumY2  +=       0.011 * cos(0.30991526634  + 22805.49174851010 *T)
    sumY2  +=       0.011 * cos(0.67526353010  + 35372.13108345990 *T)
    sumY2  +=       0.010 * cos(1.54770241799  + 11506.52595231009 *T)
    sumY2  +=       0.014 * cos(0.36722951229  + 21228.63584102931 *T)
    sumY2  +=       0.010 * cos(4.84303300061  + 12779.69461290430 *T)
    sumY2  +=       0.010 * cos(4.48229440272  + 23141.80220040810 *T)
    sumY2  +=       0.010 * cos(5.89241784221  + 1350.11122714230 *T)
    sumY2  +=       0.010 * cos(4.07842068364  + 10575.16286545830 *T)
    sumY2  +=       0.010 * cos(0.98564127078  + 22345.50419359170 *T)
    sumY2  +=       0.010 * cos(0.62034181586  + 9778.86485864190 *T)
    sumY2  +=       0.009 * cos(6.20390809264  + 29826.55017215670 *T)
    sumY2  +=       0.010 * cos(2.66921279687  + 23581.50199480110 *T)
    sumY2  +=       0.011 * cos(5.10849785178  + 25158.84553724890 *T)

    # EARTH Y Coord
    sumY3  = 0
    sumY3  +=     232.790 * cos(1.83555296287  + 0.24381748350 *T)
    sumY3  +=      76.843 * cos(0.95359770708  + 12566.39551746630 *T)
    sumY3  +=      35.331 * cos(1.77537067174  + 6283.31966747490 *T)
    sumY3  +=       5.282 * cos(0.83410185910  + 18849.47136745770 *T)
    sumY3  +=       1.631 * cos(0.72620862558  + 6282.83203250789 *T)
    sumY3  +=       1.483 * cos(0.65599216587  + 6438.74006690910 *T)
    sumY3  +=       1.479 * cos(2.83072127404  + 6127.89926804070 *T)
    sumY3  +=       0.652 * cos(3.13796451313  + 6279.79654912590 *T)
    sumY3  +=       0.656 * cos(0.01615422945  + 6286.84278582391 *T)
    sumY3  +=       0.317 * cos(1.43157695840  + 6526.04827144891 *T)
    sumY3  +=       0.318 * cos(2.05944515679  + 6040.59106350090 *T)
    sumY3  +=       0.227 * cos(0.70881980797  + 25132.54721744910 *T)
    sumY3  +=       0.201 * cos(5.52615570373  + 6836.88907031730 *T)
    sumY3  +=       0.201 * cos(4.23563267830  + 5729.75026463250 *T)
    sumY3  +=       0.036 * cos(6.05003898151  + 12569.91863581531 *T)
    sumY3  +=       0.034 * cos(0.36912822345  + 4705.97612502710 *T)
    sumY3  +=       0.033 * cos(2.65665718326  + 12410.97511803210 *T)
    sumY3  +=       0.032 * cos(3.81269087187  + 6257.02134767510 *T)
    sumY3  +=       0.032 * cos(2.84133186500  + 6309.61798727471 *T)
    sumY3  +=       0.026 * cos(5.21573736647  + 775.76642880750 *T)
    sumY3  +=       0.023 * cos(0.44696909593  + 1059.62574767270 *T)
    sumY3  +=       0.025 * cos(0.01326357360  + 7860.66320992270 *T)
    sumY3  +=       0.027 * cos(0.46730161827  + 12565.90788249930 *T)
    sumY3  +=       0.020 * cos(2.19213288615  + 5753.62870238030 *T)
    sumY3  +=       0.019 * cos(1.44435756454  + 5885.17066406670 *T)
    sumY3  +=       0.019 * cos(4.47976718197  + 6813.01063256950 *T)
    sumY3  +=       0.018 * cos(0.72211012099  + 12721.81591690050 *T)
    sumY3  +=       0.017 * cos(5.20474716177  + 6681.46867088311 *T)
    sumY3  +=       0.013 * cos(2.92478174105  + 5487.02166065850 *T)
    sumY3  +=       0.011 * cos(3.72062261949  + 7079.61767429131 *T)
    sumY3  +=       0.010 * cos(2.89167021825  + 5507.79705615090 *T)
    sumY3  +=       0.010 * cos(1.46762801985  + 11790.87290614230 *T)
    sumY3  +=       0.009 * cos(6.21784065295  + 11507.01358727710 *T)
    sumY3  +=       0.011 * cos(4.68801103302  + 12592.69383726610 *T)
    sumY3  +=       0.009 * cos(3.82179163254  + 7058.84227879890 *T)
    sumY3  +=       0.008 * cos(5.97770881958  + 6290.43321447570 *T)
    sumY3  +=       0.008 * cos(0.68313431813  + 6276.20612047410 *T)
    sumY3  +=       0.008 * cos(1.45990791265  + 796.54182429990 *T)
    sumY3  +=       0.007 * cos(4.15465801739  + 4693.75913722409 *T)
    sumY3  +=       0.007 * cos(0.35078036011  + 7.35736448430 *T)
    sumY3  +=       0.005 * cos(3.29483739022  + 3739.00524759150 *T)
    sumY3  +=       0.005 * cos(4.64160679165  + 6070.02057203690 *T)
    sumY3  +=       0.006 * cos(2.38483297622  + 6284.29998854310 *T)
    sumY3  +=       0.006 * cos(4.26851560652  + 6282.33934640670 *T)
    sumY3  +=       0.005 * cos(1.89668032657  + 4137.15425099970 *T)
    sumY3  +=       0.005 * cos(2.01213704309  + 6496.61876291290 *T)

    # EARTH Y Coord
    sumY4  = 0
    sumY4  +=     114.918 * cos(4.77291921544  + 6283.31966747490 *T)
    sumY4  +=       6.817 * cos(2.90545031303  + 0.24381748350 *T)
    sumY4  +=       3.158 * cos(5.19149443854  + 12566.39551746630 *T)
    sumY4  +=       0.211 * cos(2.15272014034  + 6438.74006690910 *T)
    sumY4  +=       0.210 * cos(1.31410288968  + 6127.89926804070 *T)
    sumY4  +=       0.253 * cos(5.35904773722  + 18849.47136745770 *T)
    sumY4  +=       0.073 * cos(5.49456429263  + 6282.83203250789 *T)
    sumY4  +=       0.042 * cos(3.03243163257  + 710.17686604180 *T)
    sumY4  +=       0.023 * cos(4.47433896564  + 6279.79654912590 *T)
    sumY4  +=       0.022 * cos(4.47061131203  + 6286.84278582391 *T)
    sumY4  +=       0.012 * cos(5.22878976906  + 25132.54721744910 *T)
    sumY4  +=       0.009 * cos(2.63928422881  + 83997.09113559539 *T)
    sumY4  +=       0.009 * cos(3.62312855703  + 11856.46246890800 *T)
    sumY4  +=       0.008 * cos(4.59507616558  + 167284.00540514899 *T)
    sumY4  +=       0.005 * cos(4.40834364231  + 6257.02134767510 *T)
    sumY4  +=       0.005 * cos(2.21798830782  + 6309.61798727471 *T)
    sumY4  +=       0.005 * cos(1.11835906029  + 12410.97511803210 *T)
    sumY4  +=       0.005 * cos(3.99770824807  + 156137.71980228281 *T)
    sumY4  +=       0.004 * cos(4.97197978751  + 12565.90788249930 *T)
    sumY4  +=       0.003 * cos(3.60141861387  + 529.93478257810 *T)

    # EARTH Y Coord
    sumY5  = 0
    sumY5  +=       0.877 * cos(4.87375194107  + 6283.31966747490 *T)
    sumY5  +=       0.305 * cos(4.05893642248  + 0.24381748350 *T)
    sumY5  +=       0.101 * cos(3.30997066470  + 12566.39551746630 *T)
    sumY5  +=       0.025 * cos(3.73514325706  + 6438.74006690910 *T)
    sumY5  +=       0.025 * cos(5.99118011879  + 6127.89926804070 *T)
    sumY5  +=       0.010 * cos(3.54319653898  + 18849.47136745770 *T)
    sumY5  +=       0.003 * cos(4.13285368647  + 6282.83203250789 *T)

    # EARTH Z Coord
    sumZ0  = 0
    sumZ0  +=     279.620 * cos(3.19870156017  + 84334.66158130829 *T)
    sumZ0  +=     101.625 * cos(5.42248110597  + 5507.55323866740 *T)
    sumZ0  +=      80.461 * cos(3.88027157914  + 5223.69391980220 *T)
    sumZ0  +=      43.855 * cos(3.70430347723  + 2352.86615377180 *T)
    sumZ0  +=      31.853 * cos(3.99997840986  + 1577.34354244780 *T)
    sumZ0  +=      22.724 * cos(3.98473831560  + 1047.74731175470 *T)
    sumZ0  +=      16.395 * cos(3.56456162523  + 5856.47765911540 *T)
    sumZ0  +=      18.150 * cos(4.98479613321  + 6283.07584999140 *T)
    sumZ0  +=      14.477 * cos(3.70258423465  + 9437.76293488700 *T)
    sumZ0  +=      14.325 * cos(3.41020246931  + 10213.28554621100 *T)
    sumZ0  +=      11.204 * cos(4.83021499098  + 14143.49524243060 *T)
    sumZ0  +=      10.896 * cos(2.08023708321  + 6812.76681508600 *T)
    sumZ0  +=       9.716 * cos(3.47560205064  + 4694.00295470760 *T)
    sumZ0  +=      10.363 * cos(4.05663928093  + 71092.88135493269 *T)
    sumZ0  +=       8.792 * cos(4.44890729280  + 5753.38488489680 *T)
    sumZ0  +=       8.337 * cos(4.99167706048  + 7084.89678111520 *T)
    sumZ0  +=       6.921 * cos(4.32559054073  + 6275.96230299060 *T)
    sumZ0  +=       9.145 * cos(1.14182646613  + 6620.89011318780 *T)
    sumZ0  +=       7.205 * cos(3.62441599378  + 529.69096509460 *T)
    sumZ0  +=       7.698 * cos(5.55425745881  + 167621.57585086187 *T)
    sumZ0  +=       5.186 * cos(6.25384533126  + 18073.70493865020 *T)
    sumZ0  +=       5.027 * cos(2.49727910749  + 4705.73230754360 *T)
    sumZ0  +=       4.529 * cos(2.33827747356  + 6309.37416979120 *T)
    sumZ0  +=       4.754 * cos(0.71100829534  + 5884.92684658320 *T)
    sumZ0  +=       4.292 * cos(1.10034019036  + 6681.22485339960 *T)
    sumZ0  +=       3.855 * cos(1.82331477760  + 5486.77784317500 *T)
    sumZ0  +=       5.256 * cos(4.42445744523  + 7860.41939243920 *T)
    sumZ0  +=       4.025 * cos(5.11990285618  + 13367.97263110660 *T)
    sumZ0  +=       4.061 * cos(6.02923989849  + 3930.20969621960 *T)
    sumZ0  +=       3.798 * cos(0.44370219832  + 3154.68708489560 *T)
    sumZ0  +=       2.933 * cos(5.12428135363  + 1059.38193018920 *T)
    sumZ0  +=       3.424 * cos(5.42065191700  + 6069.77675455340 *T)
    sumZ0  +=       3.670 * cos(4.58210192227  + 12194.03291462090 *T)
    sumZ0  +=       2.912 * cos(1.92688087702  + 10977.07880469900 *T)
    sumZ0  +=       2.506 * cos(0.60484952637  + 6496.37494542940 *T)
    sumZ0  +=       2.680 * cos(1.39497359287  + 22003.91463486980 *T)
    sumZ0  +=       2.120 * cos(4.30691000285  + 5643.17856367740 *T)
    sumZ0  +=       2.235 * cos(0.81363184041  + 8635.94200376320 *T)
    sumZ0  +=       1.819 * cos(3.75748003107  + 3340.61242669980 *T)
    sumZ0  +=       2.230 * cos(2.76407822446  + 12036.46073488820 *T)
    sumZ0  +=       1.945 * cos(5.69949789177  + 11790.62908865880 *T)
    sumZ0  +=       1.529 * cos(1.98639348122  + 398.14900340820 *T)
    sumZ0  +=       1.581 * cos(3.19976230948  + 5088.62883976680 *T)
    sumZ0  +=       1.424 * cos(6.26116472313  + 2544.31441988340 *T)
    sumZ0  +=       1.405 * cos(4.69394873481  + 7058.59846131540 *T)
    sumZ0  +=       1.588 * cos(0.25704624784  + 17298.18232732620 *T)
    sumZ0  +=       1.478 * cos(2.81808207569  + 25934.12433108940 *T)
    sumZ0  +=       1.693 * cos(4.95689385293  + 156475.29024799570 *T)
    sumZ0  +=       1.123 * cos(2.38605285936  + 3738.76143010800 *T)
    sumZ0  +=       0.990 * cos(4.29600699156  + 9225.53927328300 *T)
    sumZ0  +=       0.923 * cos(3.07195736431  + 4164.31198961300 *T)
    sumZ0  +=       0.863 * cos(0.55119493097  + 8429.24126646660 *T)
    sumZ0  +=       0.984 * cos(5.96784225392  + 7079.37385680780 *T)
    sumZ0  +=       1.086 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ0  +=       0.824 * cos(1.50050055147  + 10447.38783960440 *T)
    sumZ0  +=       0.829 * cos(3.51336978167  + 11506.76976979360 *T)
    sumZ0  +=       0.908 * cos(0.12102885474  + 11015.10647733480 *T)
    sumZ0  +=       0.742 * cos(1.99159139281  + 26087.90314157420 *T)
    sumZ0  +=       0.850 * cos(4.24120016095  + 29864.33402730900 *T)
    sumZ0  +=       0.762 * cos(2.90028034892  + 4732.03062734340 *T)
    sumZ0  +=       0.711 * cos(1.38596151724  + 2146.16541647520 *T)
    sumZ0  +=       0.766 * cos(0.61397570576  + 796.29800681640 *T)
    sumZ0  +=       0.708 * cos(1.91406542362  + 8031.09226305840 *T)
    sumZ0  +=       0.810 * cos(5.13961498599  + 2942.46342329160 *T)
    sumZ0  +=       0.751 * cos(1.67479850166  + 21228.39202354580 *T)
    sumZ0  +=       0.589 * cos(2.02193316047  + 775.52261132400 *T)
    sumZ0  +=       0.513 * cos(2.60734651128  + 12566.15169998280 *T)
    sumZ0  +=       0.550 * cos(1.61119298963  + 801.82093112380 *T)
    sumZ0  +=       0.592 * cos(4.59481504319  + 4690.47983635860 *T)
    sumZ0  +=       0.519 * cos(5.72925039114  + 8827.39026987480 *T)
    sumZ0  +=       0.515 * cos(4.09976487552  + 64809.80550494129 *T)
    sumZ0  +=       0.503 * cos(5.66433137112  + 33794.54372352860 *T)
    sumZ0  +=       0.476 * cos(3.69640472448  + 213.29909543800 *T)
    sumZ0  +=       0.463 * cos(1.24987240483  + 15720.83878487840 *T)
    sumZ0  +=       0.411 * cos(0.10769444525  + 3128.38876509580 *T)
    sumZ0  +=       0.504 * cos(3.26207669160  + 7632.94325965020 *T)
    sumZ0  +=       0.552 * cos(1.02926440457  + 239762.20451754928 *T)
    sumZ0  +=       0.394 * cos(5.81496021156  + 426.59819087600 *T)
    sumZ0  +=       0.476 * cos(3.52810472421  + 16496.36139620240 *T)
    sumZ0  +=       0.503 * cos(4.85802444134  + 6290.18939699220 *T)
    sumZ0  +=       0.417 * cos(0.81920713533  + 5216.58037280140 *T)
    sumZ0  +=       0.363 * cos(5.70015720452  + 1589.07289528380 *T)
    sumZ0  +=       0.363 * cos(1.28376436579  + 6206.80977871580 *T)
    sumZ0  +=       0.353 * cos(4.70059133110  + 7234.79425624200 *T)
    sumZ0  +=       0.381 * cos(2.61890749829  + 7342.45778018060 *T)
    sumZ0  +=       0.420 * cos(0.97821135132  + 4136.91043351620 *T)
    sumZ0  +=       0.335 * cos(6.22057967320  + 12168.00269657460 *T)
    sumZ0  +=       0.387 * cos(3.09145061418  + 25158.60171976540 *T)
    sumZ0  +=       0.349 * cos(2.94840720073  + 9623.68827669120 *T)
    sumZ0  +=       0.418 * cos(3.75759994446  + 5230.80746680300 *T)
    sumZ0  +=       0.396 * cos(1.22507712354  + 6438.49624942560 *T)
    sumZ0  +=       0.322 * cos(1.21162178805  + 8662.24032356300 *T)
    sumZ0  +=       0.323 * cos(5.50254808471  + 1194.44701022460 *T)
    sumZ0  +=       0.372 * cos(1.72256107938  + 14945.31617355440 *T)
    sumZ0  +=       0.305 * cos(0.80429352049  + 37724.75341974820 *T)
    sumZ0  +=       0.265 * cos(6.10358507671  + 6836.64525283380 *T)
    sumZ0  +=       0.250 * cos(4.56452895547  + 7477.52286021600 *T)
    sumZ0  +=       0.266 * cos(2.62926282354  + 7238.67559160000 *T)
    sumZ0  +=       0.262 * cos(1.34297269105  + 11769.85369316640 *T)
    sumZ0  +=       0.263 * cos(6.22089501237  + 6133.51265285680 *T)
    sumZ0  +=       0.302 * cos(2.81139045072  + 1748.01641306700 *T)
    sumZ0  +=       0.316 * cos(1.62662805006  + 250908.49012041549 *T)
    sumZ0  +=       0.226 * cos(2.42346415873  + 11371.70468975820 *T)
    sumZ0  +=       0.216 * cos(3.68721275185  + 5849.36411211460 *T)
    sumZ0  +=       0.276 * cos(2.33399767923  + 19651.04848109800 *T)
    sumZ0  +=       0.230 * cos(0.36165162947  + 5863.59120611620 *T)
    sumZ0  +=       0.200 * cos(5.86073159059  + 4535.05943692440 *T)
    sumZ0  +=       0.277 * cos(4.65400292395  + 82239.16695779889 *T)
    sumZ0  +=       0.199 * cos(5.05186622555  + 5429.87946823940 *T)
    sumZ0  +=       0.206 * cos(3.70556027982  + 10973.55568635000 *T)
    sumZ0  +=       0.210 * cos(4.50691909144  + 29088.81141598500 *T)
    sumZ0  +=       0.181 * cos(6.00294783127  + 4292.33083295040 *T)
    sumZ0  +=       0.249 * cos(0.12900984422  + 154379.79562448629 *T)
    sumZ0  +=       0.167 * cos(6.27348370813  + 10988.80815753500 *T)
    sumZ0  +=       0.183 * cos(3.19836584401  + 16730.46368959580 *T)
    sumZ0  +=       0.225 * cos(3.18339652605  + 18875.52586977400 *T)
    sumZ0  +=       0.192 * cos(0.87192032514  + 77375.95720492408 *T)
    sumZ0  +=       0.188 * cos(2.22746128596  + 41654.96311596780 *T)
    sumZ0  +=       0.160 * cos(5.20207795189  + 5481.25491886760 *T)
    sumZ0  +=       0.192 * cos(3.89678943865  + 17789.84561978500 *T)
    sumZ0  +=       0.159 * cos(5.32590846475  + 7.11354700080 *T)
    sumZ0  +=       0.214 * cos(4.08171403539  + 337.81426319640 *T)
    sumZ0  +=       0.184 * cos(3.75465072369  + 23581.25817731760 *T)
    sumZ0  +=       0.160 * cos(4.53897529970  + 18477.10876461230 *T)
    sumZ0  +=       0.144 * cos(5.19486374726  + 20426.57109242200 *T)
    sumZ0  +=       0.126 * cos(1.69446958184  + 15110.46611986620 *T)
    sumZ0  +=       0.149 * cos(1.55306832217  + 10021.83728009940 *T)
    sumZ0  +=       0.127 * cos(5.31068233515  + 639.89728631400 *T)
    sumZ0  +=       0.118 * cos(2.73268114062  + 18849.22754997420 *T)
    sumZ0  +=       0.123 * cos(2.55361087146  + 12592.45001978260 *T)
    sumZ0  +=       0.113 * cos(6.07178776726  + 6709.67404086740 *T)
    sumZ0  +=       0.130 * cos(3.94161517411  + 7330.82316174610 *T)
    sumZ0  +=       0.110 * cos(3.51203379263  + 18052.92954315780 *T)
    sumZ0  +=       0.147 * cos(4.63371971408  + 22805.73556599360 *T)
    sumZ0  +=       0.113 * cos(1.20711498905  + 14919.01785375460 *T)
    sumZ0  +=       0.108 * cos(3.76359880753  + 14314.16811304980 *T)
    sumZ0  +=       0.148 * cos(0.65447253687  + 95480.94718417450 *T)
    sumZ0  +=       0.121 * cos(3.13748782783  + 14712.31711645800 *T)
    sumZ0  +=       0.119 * cos(5.92110458985  + 33019.02111220460 *T)
    sumZ0  +=       0.124 * cos(2.93638421840  + 32217.20018108080 *T)
    sumZ0  +=       0.117 * cos(3.65065271640  + 45585.17281218740 *T)
    sumZ0  +=       0.127 * cos(4.74596574209  + 49515.38250840700 *T)
    sumZ0  +=       0.120 * cos(1.04211499785  + 6915.85958930460 *T)
    sumZ0  +=       0.120 * cos(5.60638811846  + 5650.29211067820 *T)
    sumZ0  +=       0.127 * cos(2.70766997307  + 12352.85260454480 *T)
    sumZ0  +=       0.113 * cos(5.39691616452  + 5235.32853823670 *T)
    sumZ0  +=       0.097 * cos(1.07908724794  + 9917.69687450980 *T)
    sumZ0  +=       0.123 * cos(5.15992792934  + 27511.46787353720 *T)
    sumZ0  +=       0.099 * cos(4.45774681732  + 4933.20844033260 *T)
    sumZ0  +=       0.087 * cos(3.93637812950  + 17654.78053974960 *T)
    sumZ0  +=       0.122 * cos(2.23956068680  + 83997.09113559539 *T)
    sumZ0  +=       0.102 * cos(3.97386522171  + 10818.13528691580 *T)
    sumZ0  +=       0.084 * cos(4.21241958885  + 22779.43724619380 *T)
    sumZ0  +=       0.103 * cos(4.63519841733  + 1349.86740965880 *T)
    sumZ0  +=       0.081 * cos(4.01655240430  + 1592.59601363280 *T)
    sumZ0  +=       0.092 * cos(3.24182657569  + 78051.58573131690 *T)
    sumZ0  +=       0.094 * cos(4.33873972800  + 36147.40987730040 *T)
    sumZ0  +=       0.076 * cos(2.83897554420  + 17260.15465469040 *T)
    sumZ0  +=       0.100 * cos(6.07733097102  + 26735.94526221320 *T)
    sumZ0  +=       0.078 * cos(0.72462427283  + 12779.45079542080 *T)
    sumZ0  +=       0.086 * cos(5.26029638250  + 28313.28880466100 *T)
    sumZ0  +=       0.093 * cos(4.31900620254  + 44809.65020086340 *T)
    sumZ0  +=       0.076 * cos(6.22743405935  + 13521.75144159140 *T)
    sumZ0  +=       0.081 * cos(2.16420552322  + 13095.84266507740 *T)
    sumZ0  +=       0.100 * cos(1.38002787119  + 28286.99048486120 *T)
    sumZ0  +=       0.072 * cos(1.55820597747  + 6256.77753019160 *T)
    sumZ0  +=       0.082 * cos(4.95202664555  + 10575.40668294180 *T)
    sumZ0  +=       0.082 * cos(1.69647647075  + 1990.74501704100 *T)
    sumZ0  +=       0.096 * cos(6.16038106485  + 24356.78078864160 *T)
    sumZ0  +=       0.075 * cos(2.29836095644  + 3634.62102451840 *T)
    sumZ0  +=       0.075 * cos(2.66367876557  + 16200.77272450120 *T)
    sumZ0  +=       0.087 * cos(0.26630214764  + 31441.67756975680 *T)
    sumZ0  +=       0.080 * cos(5.00001604436  + 150192.21439800429 *T)
    sumZ0  +=       0.079 * cos(0.01398391548  + 90617.73743129970 *T)
    sumZ0  +=       0.054 * cos(5.59738773448  + 161338.50000087050 *T)
    sumZ0  +=       0.056 * cos(2.60133794851  + 73188.37597844210 *T)
    sumZ0  +=       0.055 * cos(5.81483150022  + 143233.51002162008 *T)
    sumZ0  +=       0.054 * cos(3.38482031504  + 323049.11878710287 *T)

    # EARTH Z Coord
    sumZ1  = 0
    sumZ1  +=       9.031 * cos(3.89751156799  + 5507.55323866740 *T)
    sumZ1  +=       6.179 * cos(1.73051337995  + 5223.69391980220 *T)
    sumZ1  +=       3.793 * cos(5.24575814515  + 2352.86615377180 *T)
    sumZ1  +=       2.840 * cos(2.47694599818  + 1577.34354244780 *T)
    sumZ1  +=       1.817 * cos(0.41874743765  + 6283.07584999140 *T)
    sumZ1  +=       1.499 * cos(1.83320979291  + 5856.47765911540 *T)
    sumZ1  +=       1.463 * cos(5.68891324948  + 5753.38488489680 *T)
    sumZ1  +=       1.302 * cos(2.18800611215  + 9437.76293488700 *T)
    sumZ1  +=       1.239 * cos(4.95327097854  + 10213.28554621100 *T)
    sumZ1  +=       1.029 * cos(0.09367831611  + 7860.41939243920 *T)
    sumZ1  +=       0.975 * cos(0.08833256880  + 14143.49524243060 *T)
    sumZ1  +=       0.851 * cos(1.79547916132  + 3930.20969621960 *T)
    sumZ1  +=       0.581 * cos(2.26949174067  + 5884.92684658320 *T)
    sumZ1  +=       0.515 * cos(5.64196681593  + 529.69096509460 *T)
    sumZ1  +=       0.473 * cos(6.22750969242  + 6309.37416979120 *T)
    sumZ1  +=       0.449 * cos(1.52767374606  + 18073.70493865020 *T)
    sumZ1  +=       0.360 * cos(3.62344325932  + 13367.97263110660 *T)
    sumZ1  +=       0.372 * cos(3.22470721320  + 6275.96230299060 *T)
    sumZ1  +=       0.260 * cos(1.87645933242  + 11790.62908865880 *T)
    sumZ1  +=       0.322 * cos(0.94084045832  + 6069.77675455340 *T)
    sumZ1  +=       0.232 * cos(0.27531852596  + 7058.59846131540 *T)
    sumZ1  +=       0.217 * cos(6.03652873142  + 10977.07880469900 *T)
    sumZ1  +=       0.232 * cos(2.93325646109  + 22003.91463486980 *T)
    sumZ1  +=       0.204 * cos(3.86264841382  + 6496.37494542940 *T)
    sumZ1  +=       0.189 * cos(2.88937704419  + 15720.83878487840 *T)
    sumZ1  +=       0.179 * cos(4.90537280911  + 12036.46073488820 *T)
    sumZ1  +=       0.222 * cos(3.98495366315  + 6812.76681508600 *T)
    sumZ1  +=       0.213 * cos(1.57516933652  + 4694.00295470760 *T)
    sumZ1  +=       0.157 * cos(1.08259734788  + 5643.17856367740 *T)
    sumZ1  +=       0.154 * cos(5.99434678412  + 5486.77784317500 *T)
    sumZ1  +=       0.137 * cos(2.67760436027  + 6290.18939699220 *T)
    sumZ1  +=       0.179 * cos(2.05905949693  + 7084.89678111520 *T)
    sumZ1  +=       0.121 * cos(5.90212574947  + 9225.53927328300 *T)
    sumZ1  +=       0.125 * cos(2.24111392416  + 1059.38193018920 *T)
    sumZ1  +=       0.150 * cos(2.00175038718  + 5230.80746680300 *T)
    sumZ1  +=       0.149 * cos(5.06157254516  + 17298.18232732620 *T)
    sumZ1  +=       0.118 * cos(5.39979058038  + 3340.61242669980 *T)
    sumZ1  +=       0.161 * cos(3.32421999691  + 6283.31966747490 *T)
    sumZ1  +=       0.115 * cos(5.92406672373  + 4705.73230754360 *T)
    sumZ1  +=       0.118 * cos(4.40207874911  + 19651.04848109800 *T)
    sumZ1  +=       0.128 * cos(4.35489873365  + 25934.12433108940 *T)
    sumZ1  +=       0.109 * cos(2.52157834166  + 6438.49624942560 *T)
    sumZ1  +=       0.099 * cos(2.70727488041  + 5216.58037280140 *T)
    sumZ1  +=       0.122 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ1  +=       0.103 * cos(0.93782340879  + 8827.39026987480 *T)
    sumZ1  +=       0.083 * cos(4.12473540351  + 8635.94200376320 *T)
    sumZ1  +=       0.097 * cos(5.50959692365  + 29864.33402730900 *T)
    sumZ1  +=       0.073 * cos(1.73905345744  + 11506.76976979360 *T)
    sumZ1  +=       0.083 * cos(5.69169692653  + 775.52261132400 *T)
    sumZ1  +=       0.072 * cos(0.21891639822  + 21228.39202354580 *T)
    sumZ1  +=       0.071 * cos(2.86755026812  + 6681.22485339960 *T)
    sumZ1  +=       0.074 * cos(2.20184828895  + 37724.75341974820 *T)
    sumZ1  +=       0.063 * cos(4.45586625948  + 7079.37385680780 *T)
    sumZ1  +=       0.047 * cos(2.04946724464  + 3128.38876509580 *T)
    sumZ1  +=       0.061 * cos(0.63918772258  + 33794.54372352860 *T)
    sumZ1  +=       0.047 * cos(3.32543843300  + 26087.90314157420 *T)
    sumZ1  +=       0.049 * cos(1.60680905005  + 6702.56049386660 *T)
    sumZ1  +=       0.057 * cos(0.11215813438  + 29088.81141598500 *T)
    sumZ1  +=       0.049 * cos(3.02832204050  + 20426.57109242200 *T)
    sumZ1  +=       0.041 * cos(5.55329394890  + 11015.10647733480 *T)
    sumZ1  +=       0.041 * cos(5.91861144924  + 23581.25817731760 *T)
    sumZ1  +=       0.045 * cos(2.00068583743  + 426.59819087600 *T)
    sumZ1  +=       0.045 * cos(4.95273290181  + 5863.59120611620 *T)
    sumZ1  +=       0.050 * cos(3.62740835096  + 41654.96311596780 *T)
    sumZ1  +=       0.046 * cos(1.65798680284  + 25158.60171976540 *T)
    sumZ1  +=       0.036 * cos(5.61836577943  + 12566.15169998280 *T)
    sumZ1  +=       0.036 * cos(6.24373396652  + 6283.14316029419 *T)
    sumZ1  +=       0.036 * cos(0.40465162918  + 6283.00853968860 *T)
    sumZ1  +=       0.032 * cos(6.09025731476  + 64809.80550494129 *T)
    sumZ1  +=       0.032 * cos(6.03707103538  + 2942.46342329160 *T)
    sumZ1  +=       0.041 * cos(4.86809570283  + 1592.59601363280 *T)
    sumZ1  +=       0.037 * cos(1.04055368426  + 213.29909543800 *T)
    sumZ1  +=       0.031 * cos(3.62641145030  + 13095.84266507740 *T)
    sumZ1  +=       0.028 * cos(4.38359423735  + 7632.94325965020 *T)
    sumZ1  +=       0.030 * cos(2.03616887071  + 12139.55350910680 *T)
    sumZ1  +=       0.028 * cos(6.03334294232  + 17789.84561978500 *T)
    sumZ1  +=       0.026 * cos(3.88971333608  + 5331.35744374080 *T)
    sumZ1  +=       0.031 * cos(1.44666331503  + 16730.46368959580 *T)
    sumZ1  +=       0.026 * cos(6.26376705837  + 23543.23050468179 *T)
    sumZ1  +=       0.023 * cos(4.44388985550  + 18849.22754997420 *T)
    sumZ1  +=       0.025 * cos(4.13395006026  + 3154.68708489560 *T)
    sumZ1  +=       0.028 * cos(1.53862289477  + 6279.48542133960 *T)
    sumZ1  +=       0.028 * cos(1.96831814872  + 6286.66627864320 *T)
    sumZ1  +=       0.028 * cos(5.78094918529  + 15110.46611986620 *T)
    sumZ1  +=       0.025 * cos(0.62040343662  + 10988.80815753500 *T)
    sumZ1  +=       0.022 * cos(6.02390113954  + 16496.36139620240 *T)
    sumZ1  +=       0.026 * cos(2.48165809843  + 5729.50644714900 *T)
    sumZ1  +=       0.020 * cos(3.85655029499  + 9623.68827669120 *T)
    sumZ1  +=       0.021 * cos(5.83006047147  + 7234.79425624200 *T)
    sumZ1  +=       0.021 * cos(0.69628570421  + 398.14900340820 *T)
    sumZ1  +=       0.022 * cos(5.02222806555  + 6127.65545055720 *T)
    sumZ1  +=       0.020 * cos(3.47611265290  + 6148.01076995600 *T)
    sumZ1  +=       0.020 * cos(0.90769829044  + 5481.25491886760 *T)
    sumZ1  +=       0.020 * cos(0.03081589303  + 6418.14093002680 *T)
    sumZ1  +=       0.020 * cos(3.74220084927  + 1589.07289528380 *T)
    sumZ1  +=       0.018 * cos(1.58348238359  + 2118.76386037840 *T)
    sumZ1  +=       0.019 * cos(0.85407021371  + 14712.31711645800 *T)

    # EARTH Z Coord
    sumZ2  = 0
    sumZ2  +=       1.662 * cos(1.62703209173  + 84334.66158130829 *T)
    sumZ2  +=       0.492 * cos(2.41382223971  + 1047.74731175470 *T)
    sumZ2  +=       0.344 * cos(2.24353004539  + 5507.55323866740 *T)
    sumZ2  +=       0.258 * cos(6.00906896311  + 5223.69391980220 *T)
    sumZ2  +=       0.131 * cos(0.95447345240  + 6283.07584999140 *T)
    sumZ2  +=       0.086 * cos(1.69139956945  + 7860.41939243920 *T)
    sumZ2  +=       0.090 * cos(0.97606804452  + 1577.34354244780 *T)
    sumZ2  +=       0.090 * cos(0.37899871725  + 2352.86615377180 *T)
    sumZ2  +=       0.089 * cos(6.25807507963  + 10213.28554621100 *T)
    sumZ2  +=       0.075 * cos(0.84213523741  + 167621.57585086187 *T)
    sumZ2  +=       0.052 * cos(1.70501566089  + 14143.49524243060 *T)
    sumZ2  +=       0.057 * cos(6.15295833679  + 12194.03291462090 *T)
    sumZ2  +=       0.051 * cos(1.27616016740  + 5753.38488489680 *T)
    sumZ2  +=       0.051 * cos(5.37229738682  + 6812.76681508600 *T)
    sumZ2  +=       0.034 * cos(1.73672994279  + 7058.59846131540 *T)
    sumZ2  +=       0.038 * cos(2.77761031485  + 10988.80815753500 *T)
    sumZ2  +=       0.046 * cos(3.38617099014  + 156475.29024799570 *T)
    sumZ2  +=       0.021 * cos(1.95248349228  + 8827.39026987480 *T)
    sumZ2  +=       0.018 * cos(3.33419222028  + 8429.24126646660 *T)
    sumZ2  +=       0.019 * cos(4.32945160287  + 17789.84561978500 *T)
    sumZ2  +=       0.017 * cos(0.66191210656  + 6283.00853968860 *T)
    sumZ2  +=       0.018 * cos(3.74885333072  + 11769.85369316640 *T)
    sumZ2  +=       0.017 * cos(4.23058370776  + 10977.07880469900 *T)
    sumZ2  +=       0.017 * cos(1.78116162721  + 5486.77784317500 *T)
    sumZ2  +=       0.018 * cos(6.14348240716  + 11790.62908865880 *T)
    sumZ2  +=       0.022 * cos(1.34376831204  + 12036.46073488820 *T)
    sumZ2  +=       0.017 * cos(2.79601092529  + 796.29800681640 *T)
    sumZ2  +=       0.015 * cos(1.17032155085  + 213.29909543800 *T)
    sumZ2  +=       0.018 * cos(2.85221514199  + 5088.62883976680 *T)
    sumZ2  +=       0.017 * cos(0.21780913672  + 6283.14316029419 *T)
    sumZ2  +=       0.013 * cos(1.21201504386  + 25132.30339996560 *T)
    sumZ2  +=       0.012 * cos(5.13714452592  + 7079.37385680780 *T)
    sumZ2  +=       0.013 * cos(3.79842135217  + 4933.20844033260 *T)
    sumZ2  +=       0.012 * cos(4.89407978213  + 3738.76143010800 *T)
    sumZ2  +=       0.015 * cos(6.05682328852  + 398.14900340820 *T)
    sumZ2  +=       0.014 * cos(4.81029291856  + 4694.00295470760 *T)
    sumZ2  +=       0.011 * cos(0.61684523405  + 3128.38876509580 *T)
    sumZ2  +=       0.011 * cos(5.32876538500  + 6040.34724601740 *T)
    sumZ2  +=       0.014 * cos(5.27227350286  + 4535.05943692440 *T)
    sumZ2  +=       0.011 * cos(2.39292099451  + 5331.35744374080 *T)
    sumZ2  +=       0.010 * cos(4.45296532710  + 6525.80445396540 *T)
    sumZ2  +=       0.014 * cos(4.66400985037  + 8031.09226305840 *T)
    sumZ2  +=       0.010 * cos(3.22472385926  + 9437.76293488700 *T)
    sumZ2  +=       0.011 * cos(3.80913404437  + 801.82093112380 *T)
    sumZ2  +=       0.010 * cos(5.15032130575  + 11371.70468975820 *T)
    sumZ2  +=       0.013 * cos(0.98720797401  + 5729.50644714900 *T)
    sumZ2  +=       0.009 * cos(5.94191743597  + 7632.94325965020 *T)

    # EARTH Z Coord
    sumZ3  = 0
    sumZ3  +=       0.011 * cos(0.23877262399  + 7860.41939243920 *T)
    sumZ3  +=       0.009 * cos(1.16069982609  + 5507.55323866740 *T)
    sumZ3  +=       0.008 * cos(1.65357552925  + 5884.92684658320 *T)
    sumZ3  +=       0.008 * cos(2.86720038197  + 7058.59846131540 *T)
    sumZ3  +=       0.007 * cos(3.04818741666  + 5486.77784317500 *T)
    sumZ3  +=       0.007 * cos(2.59437103785  + 529.69096509460 *T)
    sumZ3  +=       0.008 * cos(4.02863090524  + 6256.77753019160 *T)
    sumZ3  +=       0.008 * cos(2.42003508927  + 5753.38488489680 *T)
    sumZ3  +=       0.006 * cos(0.84181087594  + 6275.96230299060 *T)
    sumZ3  +=       0.006 * cos(5.40160929468  + 1577.34354244780 *T)
    sumZ3  +=       0.007 * cos(2.73399865247  + 6309.37416979120 *T)

    # EARTH Z Coord
    sumZ4  = 0
    sumZ4  +=       0.004 * cos(0.79662198849  + 6438.49624942560 *T)
    sumZ4  +=       0.005 * cos(0.84308705203  + 1047.74731175470 *T)
    sumZ4  +=       0.005 * cos(0.05711572303  + 84334.66158130829 *T)
    sumZ4  +=       0.003 * cos(3.46779895686  + 6279.55273164240 *T)
    sumZ4  +=       0.003 * cos(2.89822201212  + 6127.65545055720 *T)

    # calculate and scale the sums
    X = sumX0 + sumX1*T**1 + sumX2*T**2 + sumX3*T**3 + sumX4*T**4 + sumX5*T**5
    X /= 1e8
    Y = sumY0 + sumY1*T**1 + sumY2*T**2 + sumY3*T**3 + sumY4*T**4 + sumY5*T**5
    Y /= 1e8
    Z = sumZ0 + sumZ1*T**1 + sumZ2*T**2 + sumZ3*T**3 + sumZ4*T**4
    Z /= 1e8

    # calculate Rau, Long Lat
    Rau = math.sqrt(X*X+Y*Y+Z*Z)
    Lon = Mod(degrees(atan2(Y, X)), 360.)
    Lat = degrees(atan2(Z, math.sqrt(X*X+Y*Y)))
    
    # create dictionary
    Dict = {"Object"  : "Earth",
            "Mode"  : "XYZ",
            "X"     : X,
            "Y"     : Y,
            "Z"     : Z,
            "Long"  : Lon,
            "Lat"   : Lat,
            "Rau"   : Rau,
            "Alt"   : -34./60.
           }
    Data = (X, Y, Z, Dict)


    # save parameters and values
    EarthdtUTimes[strdtUT] = Data


    return Data
#######################################################################



def getEarthData(dtCivicTime, zone, verbose = False):
    X, Y, Z, Dict = getEarthXYZData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Xe, Ye, Ze, Dicte = getEarthXYZData(dtCivicTime, zone, verbose)
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#######################################################################
#
#   getSunData, based on Meeus, chapter 24
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   SunRA       Rektaszension in degrees
#   SunDecl     Declination above equator in degrees
#   SunDict     Dictionary with Sun-specific output values:
#       Object:     "Sun"
#       Ltrue       True (ecliptic) Longitude in degrees
#       Lattrue     True ecliptic latitude, always 0.0
#       R           Distance Earth - Sun in km
#       epsilon     Schiefe der Ekliptik (oblique) in degrees
#       Alt         Altitude at rise, set (degrees)

Version = "V4.1 hvf 2022-05-19 getSunData.py"
VersionList.append(Version)

def getSunData(dtCivicTime, zone, verbose=False):
    # use VSOP87 data for Earth
    # use getNutation to include nutation
    deltaLong, deltaLat, epsilon = getNutation(dtCivicTime, zone, False)    
    # we do not need the nutation effects, just epsilon
    deltaLong = 0; deltaLat = 0

    # ecliptic latitude of Earth, chapter 31 - use negative value for Sun,
    # add 180 to longitude
    # Meeus p. 154
    dummy, dummy, Dict = getEarthData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    SunLat = -B + deltaLat
    Ltrue = Mod(L + deltaLong + 180., 360.)
    Rau = Dict["Rau"]
    Rabs = Rau * 149597870.700    # km

    # Rektaszension, Deklination
    SunRA = atan2(Cos(epsilon)*Sin(Ltrue) - Tan(SunLat)*Sin(epsilon), \
            Cos(Ltrue))
    SunDecl = asin(Sin(SunLat)*Cos(epsilon) \
            + Cos(SunLat)*Sin(epsilon)*Sin(Ltrue))
    SunRA = Mod(degrees(SunRA), 360.)
    SunDecl = degrees(SunDecl)

    # altitude at Sunrise, Sunset (degrees), standard value
    Alt = -50./60.    # degrees, 50'

    if (verbose):
        print ("SunData:")
        print ("Rau = ", Rau)
        print ("Rabs = ", Rabs)
        print ("epsilon = ", epsilon)
        print ("SunRA = ", SunRA)
        print ("SunDecl = ", SunDecl)
        print ("Alt = ", Alt)
#       Object:     "Sun"
#       SunLong     Ltrue   True (ecliptic) Longitude in degrees
#       SunLat      True ecliptic latitude, always 0.0
#       R           Rabs    Distance Earth - Sun in km
#       epsilon     Schiefe der Ekliptik (oblique) in degrees
#       Alt         Altitude at rise, set (degrees)
    SunDict = { "Object" : "Sun",
                "Long"  : Ltrue,
                "SunLong"  : Ltrue,
                "Lat"   : SunLat,
                "SunLat"   : SunLat,
                "RA"    : SunRA,
                "Decl"  : SunDecl,
                "R"       : Rabs,
                "Rau"     : Rau,
                "epsilon" : epsilon,
                "Alt"     : Alt,
                "X"     : 0,
                "Y"     : 0,
                "Z"     : 0,
                "eX"     : -Dict["X"],
                "eY"     : -Dict["Y"],
                "eZ"     : -Dict["Z"],
                "eRau"     : Rau,
                "eLong"  : Ltrue,
                "eLat"   : SunLat,
                }
    return (SunRA, SunDecl, SunDict)

############## End of getSunData() #####################################
########################################################################
#
#   getMoonData(dtCivicTime, zone, verbose=False):
#   Returns:    
#   MoonRA      Rektaszension in degrees
#   MoonDecl    Declination above equator in degrees
#   MoonDict    Dictionary with Moon-specific output values:
#       Object:     "Moon"
#       MoonLong    ecliptic longitude, degrees
#       MoonLat     ecliptic latitude, degrees
#       MoonDist    distance Earth - Moon, km
#       MoonDistAU  distance Earth - Moon, au
#       X           geocentric X, au
#       Y           geocentric Y, au
#       Z           geocentric Z, au
#       MoonHP      horizontal parallax, degrees
#       MoonNode    ecliptic longitude of ascending node, degrees
#       MoonPerigee longitude of perigee og Moon's orbit, degrees
#       Alt         Altitude at rise, set (degrees)

Version = "V4.1 hvf 2022-05-19 getMoonData.py"
VersionList.append(Version)

def getMoonData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.

    T = TT / 36525.    # eq. 21.1

    # obliquity of the ecliptic, eq. 21.2, degrees
    epsilon = 23.+26./60+21.448/3600. - 46.8150/3600.*T \
        - 0.00059/3600.*T*T + 0.001813/3600.*T*T*T

    # Moon mean longitude (mean equinox of the date, incl, effect of light-time)
    # eq. 45.1
    Lprime  = 218.3164591 + T * (481267.88134236 + T * (-0.0013268 \
            + T * (1./538841. - T / 65194000)))

    # Mean elongation of the Moon
    # eq. 45.2
    D       = 297.8502042 + T * (445267.1115168 + T * (-0.0016300 \
            + T * (1./545868.  -T / 113065000.)))

    # Sun's mean anomaly
    # eq. 45.3
    M       = 357.5291092 + T * (35999.0502909 + T * (-0.0001536 \
            + T / 24490000))

    # Moon's mean anomaly
    # eq 45.4
    MM      = 134.9634114 + T * (477198.8676313 + T * (0.0089970 \
            + T * (1./69699.  - T / 14712000.)))

    # Moon's argument of latitude (mean distance of the Moon
    # from its ascending node)
    # eq 45.5
    F       = 93.2720993 + T * (483202.0175273 + T * (-0.0034029 \
            + T * (-1./3526000. + T / 863310000.)))

    # three further arguments are needed
    A1      = 119.75 + 131.849 * T          # action of Venus
    A2      =  53.09 + 479264.290 * T       # action of Jupiter
    A3      = 313.45 + 481266.484 * T

    # effect of decreasing eccentricity of Earth's orbit
    E       = 1. + T * (-0.002516 + T * -0.0000074)
    EE      = E*E

    # Longitude, distance: now calculate sums sigL, sigR
    # see Table 45.A, p.309
    arg = MM;           sigL = 6288774*Sin(arg);    sigR = -20905355*Cos(arg)
    arg = 2*D - MM;     sigL += 1274027*Sin(arg);   sigR += -3699111*Cos(arg)
    arg = 2*D;          sigL += 658314*Sin(arg);    sigR += -2955968*Cos(arg)
    arg = 2*MM;         sigL += 213618*Sin(arg);    sigR += -569925*Cos(arg)  
    arg = M;            sigL += -185116*Sin(arg)*E; sigR += 48888*Cos(arg)*E
    arg = 2*F;          sigL += -114332*Sin(arg);   sigR += -3149*Cos(arg)
    arg = 2*D - 2*MM;   sigL += 58793*Sin(arg);     sigR += 246158*Cos(arg)
    arg = 2*D - M - MM; sigL += 57066*Sin(arg)*E;   sigR += -152138*Cos(arg)*E
    arg = 2*D + MM;     sigL += 53322*Sin(arg);     sigR += -170733*Cos(arg)
    arg = 2*D - M;      sigL += 45758*Sin(arg)*E;   sigR += -204586*Cos(arg)*E
    arg = M - MM;       sigL += -40923*Sin(arg)*E;  sigR += -129620*Cos(arg)*E
    arg = D;            sigL += -34720*Sin(arg);    sigR += 108743*Cos(arg)
    arg = M + MM;       sigL += -30383*Sin(arg)*E;  sigR +=  104755*Cos(arg)*E
    arg = 2*D - 2*F;    sigL += 15327*Sin(arg);     sigR += 10321*Cos(arg)
    arg = MM + 2*F;     sigL += -12528*Sin(arg)    
    arg = MM - 2*F;     sigL += 10980*Sin(arg);     sigR += 79661*Cos(arg)
    arg = 4*D-MM;       sigL += 10675*Sin(arg);     sigR += -34782*Cos(arg)
    arg = 3*MM;         sigL += 10034*Sin(arg);     sigR += -23210*Cos(arg)
    arg = 4*D-2*MM;     sigL +=  8548*Sin(arg);     sigR += -21636*Cos(arg)
    arg = 2*D + M - MM; sigL += -7888*Sin(arg)*E;   sigR += 24208*Cos(arg)*E
    arg = 2*D + M ;     sigL += -6766*Sin(arg)*E;   sigR += 30824*Cos(arg)*E
    arg = D - MM;       sigL += -5163*Sin(arg);     sigR += -8379*Cos(arg)
    arg = D + M;        sigL += 4987*Sin(arg)*E;    sigR += -16675*Cos(arg)*E
    arg = 2*D - M + MM; sigL += 4036*Sin(arg)*E;    sigR += -12831*Cos(arg)*E
    arg = 2*D + 2*MM;   sigL += 3994*Sin(arg);      sigR += -10445*Cos(arg)
    arg = 4*D;          sigL += 3861*Sin(arg);      sigR += -11650*Cos(arg)
    arg = 2*D - 3*MM;   sigL += 3665*Sin(arg);      sigR += 14403*Cos(arg)
    arg = M - 2*MM;     sigL += -2689*Sin(arg)*E;   sigR += -7003*Cos(arg)*E
    arg = 2*D - MM + 2*F; sigL += -2602*Sin(arg)    
    arg = 2*D - M - 2*MM; sigL += 2390*Sin(arg)*E;  sigR += 10056*Cos(arg)*E
    arg = D + MM;       sigL += -2348*Sin(arg);     sigR += 6322*Cos(arg)
    arg = 2*D - 2*M ;   sigL += 2236*Sin(arg)*EE;   sigR += -9884*Cos(arg)*EE
    arg = M + 2*MM;     sigL += -2120*Sin(arg)*E;   sigR += 5751*Cos(arg)*E
    arg = 2*M ;         sigL += -2069*Sin(arg)*EE
    arg = 2*D - 2*M - MM; sigL += 2048*Sin(arg)*EE; sigR += -4950*Cos(arg)*EE
    arg = 2*D + MM - 2*F;   sigL += -1773*Sin(arg); sigR += 4130*Cos(arg)
    arg = 2*D + 2*F;    sigL += -1595*Sin(arg)
    arg = 4*D - M - MM; sigL += 1215*Sin(arg)*E;    sigR += -3958*Cos(arg)*E
    arg = 2*MM + 2*F;   sigL += -1110*Sin(arg)
    arg = 3*D - MM;     sigL += -892*Sin(arg);      sigR += 3258*Cos(arg)
    arg = 2*D + M + MM; sigL += -810*Sin(arg)*E;    sigR += 2616*Cos(arg)*E
    arg = 4*D - M - 2*MM; sigL += 759*Sin(arg)*E;   sigR += -1897*Cos(arg)*E
    arg = 2*M - MM;     sigL += -713*Sin(arg)*EE;   sigR += -2117*Cos(arg)*EE
    arg = 2*D + 2*M - MM; sigL += -700*Sin(arg)*EE; sigR += 2354*Cos(arg)*EE
    arg = 2*D + M - 2*MM;   sigL += 691*Sin(arg)*E
    arg = 2*D - M - 2*F;    sigL += 596*Sin(arg)*E
    arg = 4*D + MM;     sigL += 549*Sin(arg);       sigR += -1423*Cos(arg)
    arg = 4*MM;         sigL += 537*Sin(arg);       sigR += -1117*Cos(arg)
    arg = 4*D - M;      sigL += 520*Sin(arg)*E;     sigR += -1571*Cos(arg)*E
    arg = D - 2*MM;     sigL += -487*Sin(arg);      sigR += -1739*Cos(arg)
    arg = 2*D + M - 2*F; sigL += -399*Sin(arg)*E
    arg = 2*MM - 2*F;   sigL += -381*Sin(arg);      sigR += -4421*Cos(arg)
    arg = D + M + MM ;  sigL += 351*Sin(arg)*E
    arg = 3*D - 2*MM;   sigL += -340*Sin(arg)
    arg = 4*D - 3*MM;   sigL += 330*Sin(arg)
    arg = 2*D - M + 2*MM ;  sigL += 327*Sin(arg)*E
    arg = 2*M + MM;     sigL += -323*Sin(arg)*EE;   sigR += 1165*Cos(arg)*EE
    arg = D + M - MM ;  sigL += 299*Sin(arg)*E
    arg = 2*D + 3*MM;   sigL += 294*Sin(arg)
    arg = 2*D - MM - 2*F;                           sigR += 8752*Cos(arg)

    # Latitude: calculate sumB, see Table 45.B, p.311
    arg = F;                    sigB = 5128122*Sin(arg)
    arg = MM + F;               sigB += 280602*Sin(arg)
    arg = MM - F;               sigB += 277693*Sin(arg)
    arg = 2*D - F;              sigB += 173237*Sin(arg)
    arg = 2*D - MM + F;         sigB += 55413*Sin(arg)
    arg = 2*D - MM - F;         sigB += 46271*Sin(arg)
    arg = 2*D + F;              sigB += 32573*Sin(arg)
    arg = 2*MM + F;             sigB += 17198*Sin(arg)
    arg = 2*D + MM - F;         sigB += 9266*Sin(arg)
    arg = 2*MM - F;             sigB += 8822*Sin(arg)
    arg = 2*D - M - F;          sigB += 8216*Sin(arg)*E
    arg = 2*D - 2*MM - F;       sigB += 4324*Sin(arg)
    arg = 2*D + MM + F;         sigB += 4200*Sin(arg)
    arg = 2*D + M - F;          sigB += -3359*Sin(arg)*E
    arg = 2*D - M - MM + F;     sigB += 2463*Sin(arg)*E
    arg = 2*D - M + F;          sigB += 2211*Sin(arg)*E
    arg = 2*D - M - MM - F;     sigB += 2065*Sin(arg)*E
    arg = M - MM - F;           sigB += -1870*Sin(arg)*E
    arg = 4*D - MM - F;         sigB += 1828*Sin(arg)
    arg = M + F;                sigB += -1794*Sin(arg)*E
    arg = 3*F;                  sigB += -1749*Sin(arg)
    arg = M - MM + F;           sigB += -1565*Sin(arg)*E
    arg = D + F;                sigB += -1491*Sin(arg)
    arg = M + MM + F;           sigB += -1475*Sin(arg)*E
    arg = M + MM - F;           sigB += -1410*Sin(arg)*E
    arg = M - F;                sigB += -1344*Sin(arg)*E
    arg = D - F;                sigB += -1335*Sin(arg)
    arg = 3*MM + F;             sigB += 1107*Sin(arg)
    arg = 4*D - F;              sigB += 1021*Sin(arg)
    arg = 4*D - M + F;          sigB += 833*Sin(arg)
    arg = MM - 3*F;             sigB += 777*Sin(arg)
    arg = 4*D - 2*MM + F;       sigB += 671*Sin(arg)
    arg = 2*D - 3*F;            sigB += 607*Sin(arg)
    arg = 2*D + 2*MM - F;       sigB += 596*Sin(arg)
    arg = 2*D - M + MM - F;     sigB += 491*Sin(arg)*E
    arg = 2*D - 2*MM + F;       sigB += -451*Sin(arg)
    arg = 3*MM - F;             sigB += 439*Sin(arg)
    arg = 2*D + 2*MM + F;       sigB += 422*Sin(arg)
    arg = 2*D - 3*MM - F;       sigB += 421*Sin(arg)
    arg = 2*D + M - MM + F;     sigB += -366*Sin(arg)*E
    arg = 2*D + M + F;          sigB += -351*Sin(arg)*E
    arg = 4*D + F;              sigB += 331*Sin(arg)
    arg = 2*D - M + MM + F;     sigB += 315*Sin(arg)*E
    arg = 2*D - 2*M - F;        sigB += 302*Sin(arg)*EE
    arg = MM + 3*F;             sigB += -283*Sin(arg)
    arg = 2*D + M + MM - F;     sigB += -229*Sin(arg)*E
    arg = D + M - F;            sigB += 223*Sin(arg)*E
    arg = D + M + F;            sigB += 223*Sin(arg)*E
    arg = M - 2*MM - F;         sigB += -220*Sin(arg)*E
    arg = 2*D + M - MM - F;     sigB += -220*Sin(arg)*E
    arg = D + MM + F;           sigB += -185*Sin(arg)
    arg = 2*D - M - 2*MM - F;   sigB += 181*Sin(arg)*E
    arg = M + 2*MM + F;         sigB += -177*Sin(arg)*E
    arg = 4*D - 2*MM - F;       sigB += 176*Sin(arg)
    arg = 4*D - M - MM - F;     sigB += 166*Sin(arg)*E
    arg = D + MM - F;           sigB += -164*Sin(arg)
    arg = 4*D + MM - F;         sigB += 132*Sin(arg)
    arg = D - MM - F;           sigB += -119*Sin(arg)
    arg = 4*D - M - F;          sigB += 115*Sin(arg)*E
    arg = 2*D - 2*M + F;        sigB += 107*Sin(arg)*EE

    # Additives to sigL (p.112)
    sigL +=   3958*Sin(A1) \
            + 1962*Sin(Lprime-F) \
            + 318*Sin(A2)

    # Additives to sigB
    sigB += -2235*Sin(Lprime) \
            + 382*Sin(A3) \
            + 175*Sin(A1-F) \
            + 175*Sin(A1+F) \
            + 127*Sin(Lprime-MM) \
            - 115*Sin(Lprime+MM)

    MoonLong = Lprime + sigL/1e6                # degrees
    MoonLong = Mod(MoonLong, 360.)              # degrees, normnalized
    MoonLat  = sigB/1e6                         # degrees
    MoonDist = 385000.56 + sigR/1e3             # km
    MoonHP   = degrees(asin(6378.14/MoonDist))  # degrees
    MoonDistAU = MoonDist / 149597870.700       # au
    X = MoonDistAU * Cos(MoonLat) * Cos(MoonLong) # au, geocentric X
    Y = MoonDistAU * Cos(MoonLat) * Sin(MoonLong) # au, geocentric y
    Z = MoonDistAU * Sin(MoonLat)               # au, geocentric Z

    # Lunar Node and lunar Perigee
    # p.313, eq. 45.7

    MoonNode = 125.0445550 + T * (-1934.1361849 + T * (0.0020762 \
        + T * (1./467410. - T / 60616000.)))
    MoonNode = Mod(MoonNode, 360.)
    MoonPerigee = 83.3532430 + T * (4069.0137111 + T * (-0.0103238 \
        + T * (-1./80053 + T / 18999000.)))
    MoonPerigee = Mod(MoonPerigee, 360.)

    MoonRA = atan2(Sin(MoonLong)*Cos(epsilon)-Tan(MoonLat)*Sin(epsilon), \
                Cos(MoonLong))
    MoonDecl = asin(Sin(MoonLat)*Cos(epsilon) \
            + Cos(MoonLat)*Sin(epsilon)*Sin(MoonLong))
    MoonRA = Mod(degrees(MoonRA), 360.); MoonDecl = degrees(MoonDecl)

    # Moon altitude at rise, set (see p.98)
    Alt = 0.7275 * MoonHP - 34./60. # degrees

    if verbose:
        print ("JD  = ", JD)
        print ("T   = ", T)
        print ("Lprime = ", Deg(Lprime))
        print ("D   = ", Deg(D))
        print ("M   = ", Deg(M))
        print ("MM  = ", Deg(MM))
        print ("F   = ", Deg(F))
        print ("A1  = ", Deg(A1))
        print ("A2  = ", Deg(A2))
        print ("A3  = ", Deg(A3))
        print ("E   = ", E)
        print ("sigL= %.1f" % sigL)
        print ("sigB= %.1f" % sigB)
        print ("sigR= %.1f" % sigR)
        print ("MoonLong    = %10.6f degrees" % MoonLong)
        print ("MoonLat     = %10.6f degrees" % MoonLat )
        print ("MoonDist    =   %.1f km" % MoonDist)
        print ("MoonHP      = %10.6f degrees" % MoonHP  )
        print ("MoonNode    = %10.6f degrees" % MoonNode)
        print ("MoonPerigee = %10.6f degrees" % MoonPerigee)
        print ("MoonRA      = %10.6f degrees" % MoonRA     )
        print ("MoonDecl    = %10.6f degrees" % MoonDecl)
        print ("Alt         = %10.6f degrees" % Alt     )
#   Returns:    
#   MoonRA      Rektaszension in degrees
#   MoonDecl    Declination above equator in degrees
#   MoonDict    Dictionary with Moon-specific output values:
#       Object:     "Moon"
#       MoonLong    ecliptic longitude, degrees
#       MoonLat     ecliptic latitude, degrees
#       RA          Rektaszension
#       Decl        Deklination
#       MoonDist    distance Earth - Moon, km
#       MoonHP      horizontal parallax, degrees
#       MoonDistAU  distance Earth - Moon, au
#       X           geocentric X, au
#       Y           geocentric Y, au
#       Z           geocentric Z, au
#       MoonNode    ecliptic longitude of ascending node, degrees
#       MoonPerigee longitude of perigee og Moon's orbit, degrees
#       Alt         Altitude at rise, set (degrees)
#       suffix      "++" if MoonDist <= 360000, "--" if MoonDist >= 405000
    suffix = '  '
    if MoonDist <= 360000. : suffix = "++" # Super Moon
    if MoonDist >= 405000. : suffix = "--" # Mini Moon

    MoonDict = {
        "Object"    : "Moon",
        "MoonLong"  : MoonLong,
        "MoonLat"   : MoonLat,
        "RA"        : MoonRA,
        "Decl"      : MoonDecl,
        "MoonDist"  : MoonDist,
        "MoonHP"    : MoonHP,
        "MoonDistAU": MoonDistAU,
        "X"         : X,
        "Y"         : Y,
        "Z"         : Z,
        "MoonNode"  : MoonNode,
        "MoonPerigee": MoonPerigee,
        "Alt"       : Alt,
        "suffix"    : suffix}
    

    return (MoonRA, MoonDecl, MoonDict)

################ End of getMoonData() #################################
Version = "V5.1 hvf 2023-03-13 getObjectTimes.py"
VersionList.append(Version)
# steps of 2 hours except for Sun and Moon
# I realized that Moon south passing on 2022-05-14 for Singapur gave wrong
# azimut values, therefore I change stepsize for Sun and Moon to 15 minutes
# hvf 16.05.2022
# third order Chebysheff polynomial interpolation
# use hvfTime - hvf 09.11.2022
# interpolate x and y instead of azimut, avoids problem near Zenit
# e.g. Singapur 2023-03-09
# hvf 13.03.2023

# important libraries
import sys
# Python Cookbook p.22 sorting a list of dicts by a common key
from operator import itemgetter
from datetime import datetime, date, timedelta
import numpy as np
import scipy.optimize


# some small functions
def isUp(alt, Alt):
    return alt > Alt
def isDown(alt, Alt):
    return alt <= Alt
def isNE(azimut):
    return azimut >= 0 and azimut < 90.
def isSE(azimut):
    return azimut >= 90. and azimut < 180.
def isSW(azimut):
    return azimut >= 180. and azimut < 270.
def isNW(azimut):
    return azimut >= 270. and azimut < 360.
# using x, y, z coords: see https://de.wikipedia.org/wiki/Astronomische_Koordinatensysteme#Umrechnungen
# x points South, y points West, z points Zenit
def isNExy(x, y):
    return (x <= 0. and y <= 0.)
def isSExy(x, y):
    return (x > 0. and y <= 0.)
def isSWxy(x, y):
    return (x > 0. and y > 0.)
def isNWxy(x, y):
    return (x <= 0. and y > 0.)
def isNxy(x, y):
    return (x <= 0.)
def isSxy(x, y):
    return (x > 0.)
def isExy(x, y):
    return (y <= 0.)
def isWxy(x, y):
    return (y > 0.)

# create coefficients for a polynome through (xi, yi)
def getPolyCoeffs(xi, yi):
    # return coeffs of polynom of degree n-1 going through (xi, yi)
    xarr = np.array(xi)
    yarr = np.array(yi)
    n = len(xarr)
    if n != len(yarr):
        print ("arrays xi and yi differ in length: ", n, len(yarr))
        sys.exit(1)
    A = np.ndarray((n, n))
    for row in range(n):
        for col in range(n):
            a = xarr[col]**row
            A[row, col] = a
    Ainv = np.linalg.inv(A)
    coeffs = np.dot(yarr, Ainv)
    check = np.dot(coeffs, A)
    delta = yarr-check
    return coeffs

# return value of polynom with coefficients coeffs for value x
def getPoly(x, coeffs):
    # return f(x), value of polynom with coefficients coeffs at value x
    f = 0.0
    for n in range(len(coeffs)):
        f += x**n * coeffs[n]
    return f

# function getLMST() calculates local sideral time
# for given civic time dt and time zone and longitude
def getLMST(dt, zone, longitude, verbose = False):
    # day at 00:00 civic time
    dtUTC = getUTfromCivicTime(dt, zone)
    JD = getJulianDay(dtUTC)
    GreenwichST = getGMSTfromJD(JD)
    LocalST = getLMSTfromGMST(GreenwichST, longitude)
    if verbose:
        print ("Day            = ", nicedatetime(dt))
        print ("UTC            = ", nicedatetime(dtUTC))
        print ("JD             = ", JD)
        print ("Greenwich Sideral Time = ", nicedatetime (GreenwichST))
        print ("Local     Sideral Time = ", nicedatetime (LocalST))
        print ("Longitude      = ", longitude)
    return LocalST

# get azimut and altitude for given civic time, zone, object data, 
#   longitude, latitude
def getAziAltfromCivicTimeLongLat(dtCivicTime, zone, getData, 
        longitude, latitude, verbose = False):

    (RA, Decl, Dict) = getData(dtCivicTime, zone, verbose)
    localST = getLMST(dtCivicTime, zone, longitude, verbose)
    RAh = RA/15.    # convert to hours
    HourAngle = localST - timedelta(hours=RAh)   # hour angle
    # use hvfTime()
    (h, m, s) = HourAngle.getTime()
    HourAngledecimal = h + m/60. + s/3600.
    HourAngledecimalstring = "%02d:%02d:%02d" % (h, m, s)

    #HourAngledecimal = HourAngle.time().hour + HourAngle.time().minute/60. \
    #    + HourAngle.time().second/3600 + HourAngle.time().microsecond/3600e6
    HourAnglegrad = HourAngledecimal*15.
    # get azimut, altitude from hour angle, declination, latitude
    (azi, alt, xyzDict) = getAziAltfromHourDecl(HourAnglegrad, Decl, latitude)

    verbose = False
    if verbose:
        print ("getAziAltfromCivicTimeLongLat:")
        print ("    dt = ", nicedatetime(dtCivicTime), ", zone = ", zone)
        print ("    Declination, latitude = ", round(Decl, 2),round(latitude,2))
        print ("    Hour Angle (deg) = ", HourAnglegrad)
        print ("    Azi, alt = ", round(azi, 2), round(alt, 2))
        print ("    x, y, z    = ", xyzDict["x"], xyzDict["y"], xyzDict["z"])
        print ("    xx, yy, zz = ", xyzDict["xx"], xyzDict["yy"], xyzDict["zz"])
    verbose = False

    # add xyzDict to Dict
    Dict["xyzDict"] = xyzDict
    return (azi, alt, Dict)

# interpolate altitude between times dt1 and dt2 for zone, use Npoints
# Chebychev points between dt1 and dt2
# only used for rise and set, where Alt is given in the object Dict (Moon)
# Alt is None if Dict["Alt"] value is used, else Alt is used (for twilight)
def interpolateAlt(dt1, dt2, zone, Npoints, getData, Alt,
        longitude, latitude, verbose = False):

    altList = []
    xList = []
    DictList = []
    AltList = []
    dtspan = (dt2-dt1)/2
    dtmean = dt1 + dtspan
    for j in range(Npoints):
        # Tschebyscheff Stuetzstellen zwischen -1 und +1
        xj = cos((2.*j+1.)/(2.*Npoints)*pi)
        dt = dtmean + dtspan * xj       # zwischen dt1 und dt2
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
            getData, longitude, latitude)
        altList.append(alt)
        xList.append(xj)
        if Alt is None:
            AltList.append(Dict["Alt"])
        else:
            AltList.append(Alt)
            

    # create polynom for altitude and target Altitude
    # Altitude varies for Moon only
    coeffs = getPolyCoeffs(xList, altList)
    altCoeffs = getPolyCoeffs(xList, AltList)

    # apply target Alt - Alt varies for Moon only
    coeffs -= altCoeffs

    # get zero of polynom - catch ValueError
    try:
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1., 1.), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
    except ValueError:  # extend interval a little bit
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1.01, 1.01), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
        if verbose:
            fa = getPoly(-1.0, coeffs)
            fb = getPoly(1.0, coeffs)
            print ("f(a), f(b) = ", fa, fb)
            print ("xzero = ", xzero.root)

    dtzero = dtmean + dtspan*xzero.root
    if verbose:
        print ("xList = ", xList)
        print ("AltList = ", AltList)
        print ("coeffs = ", coeffs)
        print ("altCoeffs = ", altCoeffs)
        print ("xzero = ", xzero)
        print ("xzero.root = ", xzero.root)
        print ("dtzero = ", dtzero)

    return dtzero

# interpolate azimut between times dt1 and dt1 for zone, use Npoints
# Chebychev points between dt1 and dt2
def interpolateAzi(dt1, dt2, zone, targetAzi, Npoints, getData,
        longitude, latitude, verbose = False):

    aziList = []
    xList = []
    DictList = []
    dtspan = (dt2-dt1)/2
    #print ("interpolateAzi dt1, dt2 = ", dt1, dt2)
    #print ("interpolateAzi dtspan = ", dtspan, type(dtspan))
    dtmean = dt1 + dtspan
    for j in range(Npoints):
        # Tschebyscheff Stuetzstellen, zwischen -1 und +1
        xj = cos((2.*j+1.)/(2.*Npoints)*pi)
        dt = dtmean + dtspan * xj       # zwischen dt1 und dt2
        #print("j ", j, "dt ", nicedatetime(dt))
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
            getData, longitude, latitude, verbose)
        if targetAzi >= 0. and targetAzi < 45.:
            if azi > 225.: azi -= 360.
        aziList.append(azi)
        xList.append(xj)

    # create polynom
    coeffs = getPolyCoeffs(xList, aziList)

    # apply target Azimut
    coeffs[0] -= targetAzi

    # get zero of polynom, catch ValueError (Clansayes, 2022-06-02)
    try:
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1., 1.), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
    except ValueError:  # we need to enlarge the interval a little bit
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1.01, 1.01), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
        if verbose:
            fa = getPoly(-1.0, coeffs)
            fb = getPoly(1.0, coeffs)
            print ("f(a), f(b) = ", fa, fb)
            print ("xzero = ", xzero.root)

    dtzero = dtmean + dtspan*xzero.root
    if verbose:
        print ("targetAzi = ", targetAzi)
        print ("xList = ", xList)
        print ("aziList = ", aziList)
        print ("coeffs = ", coeffs)
        print ("xzero = ", xzero)
        print ("xzero.root = ", xzero.root)
        print ("dtzero = ", dtzero)

    return dtzero

# interpolate y (transit over meridian)  between times dt1 and dt1 for zone, i
# use Npoints Chebychev points between dt1 and dt2
def interpolateY(dt1, dt2, zone, targetY, Npoints, getData,
        longitude, latitude, verbose = False):

    YList = []
    xList = []  # Stuetzstellen
    DictList = []
    dtspan = (dt2-dt1)/2
    #print ("interpolateY dt1, dt2 = ", dt1, dt2)
    #print ("interpolateY dtspan = ", dtspan, type(dtspan))
    dtmean = dt1 + dtspan
    for j in range(Npoints):
        # Tschebyscheff Stuetzstellen, zwischen -1 und +1
        xj = cos((2.*j+1.)/(2.*Npoints)*pi)
        dt = dtmean + dtspan * xj       # zwischen dt1 und dt2
        #print("j ", j, "dt ", nicedatetime(dt))
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
            getData, longitude, latitude, verbose)
        xtemp = Dict["xyzDict"]["x"]   
        ytemp = Dict["xyzDict"]["y"]   
        YList.append(ytemp)
        xList.append(xj)

    # create polynom
    coeffs = getPolyCoeffs(xList, YList)

    # apply target Azimut
    coeffs[0] -= targetY

    # get zero of polynom, catch ValueError (Clansayes, 2022-06-02)
    try:
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1., 1.), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
    except ValueError:  # we need to enlarge the interval a little bit
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1.01, 1.01), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
        if verbose:
            fa = getPoly(-1.0, coeffs)
            fb = getPoly(1.0, coeffs)
            print ("f(a), f(b) = ", fa, fb)
            print ("xzero = ", xzero.root)

    dtzero = dtmean + dtspan*xzero.root
    if verbose:
        print ("targetY = ", targetY)
        print ("xList = ", xList)
        print ("YList = ", YList)
        print ("coeffs = ", coeffs)
        print ("xzero = ", xzero)
        print ("xzero.root = ", xzero.root)
        print ("dtzero = ", dtzero)

    return dtzero

# interpolate x (transit over East-West)  between times dt1 and dt1 for zone,
# use Npoints Chebychev points between dt1 and dt2
def interpolateX(dt1, dt2, zone, targetX, Npoints, getData,
        longitude, latitude, verbose = False):

    XList = []
    xList = []  # Stuetzstellen
    DictList = []
    dtspan = (dt2-dt1)/2
    #print ("interpolateX dt1, dt2 = ", dt1, dt2)
    #print ("interpolateX dtspan = ", dtspan, type(dtspan))
    dtmean = dt1 + dtspan
    for j in range(Npoints):
        # Tschebyscheff Stuetzstellen, zwischen -1 und +1
        xj = cos((2.*j+1.)/(2.*Npoints)*pi)
        dt = dtmean + dtspan * xj       # zwischen dt1 und dt2
        #print("j ", j, "dt ", nicedatetime(dt))
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
            getData, longitude, latitude, verbose)
        xtemp = Dict["xyzDict"]["x"]   
        ytemp = Dict["xyzDict"]["y"]   
        XList.append(xtemp)
        xList.append(xj)

    # create polynom
    coeffs = getPolyCoeffs(xList, XList)

    # apply target Azimut
    coeffs[0] -= targetX

    # get zero of polynom, catch ValueError (Clansayes, 2022-06-02)
    try:
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1., 1.), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
    except ValueError:  # we need to enlarge the interval a little bit
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1.01, 1.01), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
        if verbose:
            fa = getPoly(-1.0, coeffs)
            fb = getPoly(1.0, coeffs)
            print ("f(a), f(b) = ", fa, fb)
            print ("xzero = ", xzero.root)

    dtzero = dtmean + dtspan*xzero.root
    if verbose:
        print ("targetX = ", targetX)
        print ("xList = ", xList)
        print ("XList = ", XList)
        print ("coeffs = ", coeffs)
        print ("xzero = ", xzero)
        print ("xzero.root = ", xzero.root)
        print ("dtzero = ", dtzero)

    return dtzero


#######################################################################
#   getObjectTimes  Parameters:
#   dtCivicTime hvfTime object containing date and time of request
#   longitude   geographic longitude, degrees, positive if East
#   latitude    geographic latitude, degrees, positive if North
#   zone        time zone, hours East of Greenwich
#   getData     Function that supplies the necessary parameters
#               like getMoonData, getSunData
#   getData Returns:
#       RA       Rektaszension in degrees
#       Decl     Declination above equator in degrees
#       Dict     Dictionary with object-specific output values
#   Alt         altitude above horizon defining rise and set,
#               default value is None, Alt is taken from Object Dict
#   verbose     default False, use if debugging output is required
#
#   getObjectTimes returns a List of dictionaries with events containing: 
#   Object:     Dict["Object"],  # Sun, Moon, planets
#   dt    :     dt,     # hvfTime object with date and time of event
#   azi   :     azi,    # azimut of object at time of event, degrees
#   alt   :     alt,    # altitude above horizon of event, degrees
#   event :             # event type, see below
#   Alt   :     Dict["Alt"],    # altitude at time of event, taken from Dict
#   Dict  :     Dict,   # object dictionary got from getData
#   Event types:                    
#   "Rise"                rise time
#   "Set"                 set time 
#   "East  Passing (N-S)" time at first vertical (East, azimut 90 degrees)
#   "East  Passing (S-N)" time at first vertical (East, azimut 90 degrees)
#   "Transit South (E-W)" time of transit on South meridian, above horizon
#   "Transit South (W-E)" time of transit on South meridian, above horizon
#   "Transit North (E-W)" time of transit on North meridian, above horizon
#   "Transit North (W-E)" time of transit on North meridian, above horizon
#   "West  Passing (S-N)" time at second vertical (West, azimut 270 degrees)
#   "West  Passing (N-S)" time at second vertical (West, azimut 270 degrees)
#   "North Passing (W-E)" time of transit on North meridian, below horizon
#   "North Passing (E-W)" time of transit on North meridian, below horizon
#   "South Passing (E-W)" time of transit on South meridian, below horizon
#   "South Passing (W-E)" time of transit on South meridian, below horizon
#
#   all quantities are not defined (not in the dict) if there is no such event
#   NB. calling sequence is identical to getRTS()
#
#   The returned List of Dictionaries is sorted by the event time
#
#   Method: interpolation on the specified day
#
#   hvf 03.07.2021 09.06.2022
#   interpolate x, y (horizontal coords, x pointing South, y West)
#   avoids inaccuacies near Zenit
#   hvf 13.03.2023§

def getObjectTimes(dtCivicTime, longitude, latitude, zone, getData, 
                    Alt=None, verbose=False):
    
    if verbose: print ("getObjectTimes dt = ", dtCivicTime, type(dtCivicTime))
    # North Pole, South Pole: avoid singularities
    if latitude >=  90: latitude =  89.99
    if latitude <= -90: latitude = -89.99

    # create list of object data during 24 hours, steps+1 elements

    #dtMidnight = datetime(dtCivicTime.year, dtCivicTime.month, dtCivicTime.day)
    (y, m, d) = dtCivicTime.getDate()
    d = floor(d)    # strip fractions, i.e. hours, mins, secs
    dtMidnight = hvfTime(y, m, d, 0, 0, 0)
    if verbose: print ("dtMidnight = ", dtMidnight)
    Steps = 12  # one step every 2 hour
    #Steps = 96  # one step every 15 min, necessary for singapur ?????????
    # we want finer steps for Sun and Moon
    (dummy, dummy, Dict) = getAziAltfromCivicTimeLongLat(dtMidnight, zone, 
            getData, longitude, latitude)
    if (Dict["Object"] == "Sun" or Dict["Object"] == "Moon"):
        Steps = 96  # one step every 15 minutes
    intervalHours = 24. / Steps
    Npoints = 4 # interpolate between 4 points
    statusList = []
    EventList = []
    statusListXY = []
    useAzi = False  # do not use Azimut interpolation

    if verbose: print ("Steps = ", Steps, ", interval = ", intervalHours)

    for i in range(Steps+1):
        hour = intervalHours * i
        dtDayHour = dtMidnight + timedelta(hours=hour)
        if verbose: print ("dtDayHour = ", dtDayHour)
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dtDayHour, zone, 
            getData, longitude, latitude)
        if verbose: print ("azi, alt = ", azi, alt)
        # use Alt if a value is given, not None
        if Alt is None:
            useAlt = Dict["Alt"]
        else:
            useAlt = Alt
        x = Dict["xyzDict"]["x"]
        y = Dict["xyzDict"]["y"]
        # store azi, alt and rest of stuff
        statusList.append((i, hour, isUp(alt, useAlt), 
            isDown(alt, useAlt), isNE(azi), isSE(azi), 
            isSW(azi), isNW(azi)))
        statusListXY.append((i, hour, isUp(alt, useAlt), 
            isDown(alt, useAlt), isNxy(x, y), isSxy(x, y), 
            isWxy(x, y), isExy(x, y), round(x,3), round(y,3),
            round(azi, 1), round(alt, 2)))
    
    if verbose:
        print ("getObjectData stored events:")
        print (" i  hour isUP   isDown isNE  isSE  isSW  isNW")
        for status in statusList:
            print (status)
        print (" i  hour isUP   isDown isN   isS   isW   isE    x       y")
        for status in statusListXY:
            print (status)
    

    for n in range(len(statusList)-1):
        (i, hour, UP, DOWN, NE, SE, SW, NW ) = statusList[n]
        (i1, hour1, UP1, DOWN1, NE1, SE1, SW1, NW1) = statusList[n+1]
        dt1 = dtMidnight + timedelta(hours=hour)
        dt2 = dtMidnight + timedelta(hours=hour1)

        # Set
        if UP and not UP1 or not DOWN and DOWN1:
            if verbose:
                print("Setting between %.1f and %.1f" % (hour, hour1))
            dt = interpolateAlt(dt1, dt2, zone, Npoints, getData, Alt,
                    longitude, latitude, verbose)
            (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
                getData, longitude, latitude)
            Event = {'Object': Dict["Object"],  # Sun, Moon
                    'dt': dt,
                    'azi': azi,
                    'alt': alt,
                    'event': "Set",
                    'Alt': Dict["Alt"],
                    'Dict': Dict,
                    }
            EventList.append(Event)

        # Rise
        if not UP and UP1 or DOWN and not DOWN1:
            if verbose:
                print("Rising  between %.1f and %.1f" % (hour, hour1))
            dt = interpolateAlt(dt1, dt2, zone, Npoints, getData, Alt,
                    longitude, latitude, verbose)
            (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
                getData, longitude, latitude)
            Event = {'Object': Dict["Object"],  # Sun, Moon
                    'dt': dt,
                    'azi': azi,
                    'alt': alt,
                    'event': "Rise",
                    'Alt': Dict["Alt"],
                    'Dict': Dict,
                    }
            EventList.append(Event)
            
    for n in range(len(statusListXY)-1):
        (i, hour, UP, DOWN, N, S, W, E, xx, yy, aa, hh ) = statusListXY[n]
        (i1, hour1, UP1, DOWN1, N1, S1, W1, E1, xx1, yy1, aa1, hh1) = statusListXY[n+1]
        dt1 = dtMidnight + timedelta(hours=hour)
        dt2 = dtMidnight + timedelta(hours=hour1)

        # East or West Passing North to South, South to North: x = 0
        NS1 = N and S1  # from N to S
        SN1 = S and N1  # from S to N
        if NS1 or SN1:
            if NS1 and E:
                sevent = "Passing  East (N-S)"
            elif NS1 and W:
                sevent = "Passing  West (N-S)"
            elif SN1 and E:
                sevent = "East  Passing (S-N)"
            elif SN1 and W:
                sevent = "Passing  West (S-N)"
            if verbose:
                print(sevent, " between %.1f and %.1f" % (hour, hour1))
            dt = interpolateX(dt1, dt2, zone, 0., Npoints, 
                    getData, longitude, latitude, verbose)
            (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
                getData, longitude, latitude)
            Event = {'Object': Dict["Object"],  # Sun, Moon
                    'dt': dt,
                    'azi': azi,
                    'alt': alt,
                    'event': sevent,
                    'Alt': Dict["Alt"],
                    'Dict': Dict,
                    }
            EventList.append(Event)

        # North or South passing or transit Est to West, West to East: y = 0.
        EW1 = E and W1  # from E to W
        WE1 = W and E1  # from W to E
        if EW1 or WE1:
            dt = interpolateY(dt1, dt2, zone, 0., Npoints, 
                    getData, longitude, latitude, verbose)
            (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
                getData, longitude, latitude)
            if alt > Dict["Alt"]:
                if EW1 and S:
                    sevent = "Transit South (E-W)"
                elif EW1 and N:
                    sevent = "Transit North (E-W)"
                elif WE1 and S:
                    sevent = "Transit South (W-E)"
                elif WE1 and N:
                    sevent = "Transit North (W-E)"
            else:
                if EW1 and S:
                    sevent = "Passing South (E-W)"
                elif EW1 and N:
                    sevent = "Passing North (E-W)"
                elif WE1 and S:
                    sevent = "Passing South (W-E)"
                elif WE1 and N:
                    sevent = "Passing North (W-E)"
                    
            Event = {'Object': Dict["Object"],  # Sun, Moon
                    'dt': dt,
                    'azi': azi,
                    'alt': alt,
                    'event': sevent,
                    'Alt': Dict["Alt"],
                    'Dict': Dict,
                    }
            if verbose:
                print(sevent, " between %.1f and %.1f" % (hour, hour1))
            EventList.append(Event)

    # sort events by time dt
    SortedEventList = sorted(EventList, key = itemgetter('dt'))
    if verbose:
        print ("Steps = %d, Npoints = %d" % (Steps, Npoints))
        for e in SortedEventList:
            print ("%-4s %20s %s Azimut %3d Altitude %7.3f" % (e["Object"], e["event"], 
                nicedatetime(e['dt']), round(e["azi"], 0), e["alt"]))


    return SortedEventList


'''
EventList = getObjectTimes(dt, longitude, latitude, Zone, getMoonData, None,False)
print ("%-4s Events:" % EventList[0]["Object"])
for e in EventList:
    print ("%-4s %20s %s Azimut %3d Altitude %7.3f" % (e["Object"], e["event"], 
                nicedatetime(e['dt']), round(e["azi"], 0), e["alt"]))
print ()
EventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, None, False)
print ("%-4s Events:" % EventList[0]["Object"])
for e in EventList:
    print ("%-4s %s %s Azimut %3d Altitude %7.3f" % (e["Object"], e["event"], 
                nicedatetime(e['dt']), round(e["azi"], 0), e["alt"]))

print ()
'''

Version = "V3.1 hvf 2023-11-29 MoonPhasesMain.py using ELP-2000/82 and VSOP Ephemerides"
VersionList.append(Version)
# Phases of the Moon, Meeus Chapter 47
# hvf 08.02.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87
# includes Solar Eclipses - hvf 28.06.2023

# getPhasesJDEbyK (k, verbose = False)
# k counts the number of new moons since 2000-01-06
# Meeus chapter 47
# returns the Julian Ephemeris Days (JDE) time of the mean phase of the Moon
# returns also the elements for solar eclipse (for new moon)

def getPhasesJDEbyK(k, verbose = False):
    # k integer:        New Moon    - Solar eclipse possible
    # k integer + 0.25: First Quarter
    # k integer + 0.5 : Full Moon   - Lunar eclipse possible
    # k integer + 0.75: Last Quarter

    T  = k / 1236.85            # Julian centuries after Epoch2000 (47.3)
    E = 1 - 0.002516*T - 0.0000074*T*T      # (45.6)
    EE = E*E

    # Sun's Mean anomaly at time JDE (47.4)
    M = 2.5534 + 29.10535669*k - 0.0000218*T*T - 0.00000011*T*T*T
    M = Mod(M, 360.)

    # Moon's Mean anomaly (47.5)
    MM = 201.5643 + 385.81693528*k + 0.0107438*T*T \
            + 0.00001239*T**3 - 0.000000058*T**4
    MM = Mod(MM, 360.)

    # Moon's argument of Latitude (47.6)
    F = 160.7108 + 390.67050274*k - 0.0016341*T*T \
            - 0.00000227*T**3 + 0.000000011*T**4
    F = Mod(F, 360.)

    # Longitude of the Ascending Node of the lunar orbit (47.7)
    Omega = 124.7746 - 1.56375580*k \
            + 0.0020691*T*T + 0.00000215*T**3
    Omega = Mod(Omega, 360.)

    # chapter 52 Solar Eclipses  p.350
    F1 = F - 0.02665 * Sin(Omega)

    # Planetary Arguments

    A1  = 299.77 +  0.107408*k - 0.009173*T*T
    A2  = 251.88 +  0.016321*k
    A3  = 251.83 + 26.651886*k
    A4  = 349.42 + 36.412478*k
    A5  =  84.66 + 18.206239*k
    A6  = 141.74 + 53.303771*k
    A7  = 207.14 +  2.453732*k
    A8  = 154.84 +  7.306860*k
    A9  =  34.52 + 27.261239*k
    A10 = 207.19 +  0.121824*k
    A11 = 291.34 +  1.844379*k
    A12 = 161.72 + 24.198154*k
    A13 = 239.56 + 25.513099*k
    A14 = 331.55 +  3.592518*k

    # Corrections (in days) to JDE (p. 321)
    # for New Moon and Full Moon

    arg = MM;       sNew =  -0.40720*Sin(arg);   sFull =  -0.40614*Sin(arg)
    arg = M;        sNew +=  0.17241*E*Sin(arg); sFull +=  0.17302*E*Sin(arg)
    arg = 2*MM;     sNew +=  0.01608*Sin(arg);   sFull +=  0.01614*Sin(arg)
    arg = 2*F;      sNew +=  0.01039*Sin(arg);   sFull +=  0.01043*Sin(arg)
    arg = MM-M;     sNew +=  0.00739*E*Sin(arg); sFull +=  0.00734*E*Sin(arg)
    arg = MM+M;     sNew += -0.00514*E*Sin(arg); sFull += -0.00515*E*Sin(arg)
    arg = 2*M;      sNew +=  0.00208*EE*Sin(arg); sFull += 0.00209*EE*Sin(arg)
    arg = MM-2*F;   sNew += -0.00111*Sin(arg);   sFull += -0.00111*Sin(arg)
    arg = MM+2*F;   sNew += -0.00057*Sin(arg);   sFull += -0.00057*Sin(arg)
    arg = 2*MM+M;   sNew +=  0.00056*E*Sin(arg); sFull +=  0.00056*E*Sin(arg)
    arg = 3*MM;     sNew += -0.00042*Sin(arg);   sFull += -0.00042*Sin(arg)
    arg = M+2*F;    sNew +=  0.00042*E*Sin(arg); sFull +=  0.00042*E*Sin(arg)
    arg = M-2*F;    sNew +=  0.00038*E*Sin(arg); sFull +=  0.00038*E*Sin(arg)
    arg = 2*MM-M;   sNew += -0.00024*E*Sin(arg); sFull += -0.00024*E*Sin(arg)
    arg = Omega;    sNew += -0.00017*Sin(arg);   sFull += -0.00017*Sin(arg)
    arg = MM+2*M;   sNew += -0.00007*Sin(arg);   sFull += -0.00007*Sin(arg)
    arg = 2*MM-2*F; sNew +=  0.00004*Sin(arg);   sFull +=  0.00004*Sin(arg)
    arg = 3*M;      sNew +=  0.00004*Sin(arg);   sFull +=  0.00004*Sin(arg)
    arg = MM+M-2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = 2*MM+2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = MM+M+2*F; sNew += -0.00003*Sin(arg);   sFull += -0.00003*Sin(arg)
    arg = MM-M+2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = MM-M-2*F; sNew += -0.00002*Sin(arg);   sFull += -0.00002*Sin(arg)
    arg = 3*MM+M;   sNew += -0.00002*Sin(arg);   sFull += -0.00002*Sin(arg)
    arg = 4*MM;     sNew +=  0.00002*Sin(arg);   sFull +=  0.00002*Sin(arg)

    # Correction (in days) for First and Last Quarter (p.322)

    arg = MM;       sQ =  -0.62801*Sin(arg); 
    arg = M;        sQ +=  0.17172*E*Sin(arg); 
    arg = MM+M;     sQ += -0.01183*E*Sin(arg); 
    arg = 2*MM;     sQ +=  0.00862*Sin(arg); 
    arg = 2*F;      sQ +=  0.00804*Sin(arg); 
    arg = MM-M;     sQ +=  0.00454*E*Sin(arg); 
    arg = 2*M;      sQ +=  0.00204*EE*Sin(arg); 
    arg = MM-2*F;   sQ += -0.00180*Sin(arg); 
    arg = MM+2*F;   sQ += -0.00070*Sin(arg); 
    arg = 3*MM;     sQ += -0.00040*Sin(arg); 
    arg = 2*MM-M;   sQ += -0.00034*E*Sin(arg); 
    arg = M+2*F;    sQ +=  0.00032*E*Sin(arg); 
    arg = M-2*F;    sQ +=  0.00032*E*Sin(arg); 
    arg = 2*M+MM;   sQ += -0.00028*EE*Sin(arg); 
    arg = 2*MM+M;   sQ +=  0.00027*E*Sin(arg); 
    arg = Omega;    sQ += -0.00017*Sin(arg); 
    arg = MM-M-2*F; sQ += -0.00005*Sin(arg); 
    arg = 2*MM+2*F; sQ +=  0.00004*Sin(arg); 
    arg = MM+M+2*F; sQ += -0.00004*Sin(arg); 
    arg = MM-2*M;   sQ +=  0.00004*Sin(arg); 
    arg = MM+M-2*F; sQ +=  0.00003*Sin(arg); 
    arg = 3*M;      sQ +=  0.00003*Sin(arg); 
    arg = 2*MM-2*F; sQ +=  0.00002*Sin(arg); 
    arg = MM-M+2*F; sQ +=  0.00002*Sin(arg); 
    arg = 3*MM+M;   sQ += -0.00002*Sin(arg); 

    # additional correction for Quarter phases only

    W = 0.00306 - 0.00038*E*Cos(M) + 0.00026*Cos(MM) \
        - 0.00002*Cos(MM-M) + 0.00002*Cos(MM+M) + 0.00002*Cos(2*F)
    # for first Quarter: +W
    # for last Quarter:  -W

    # additional correction for all phases

    addCorr = 0.000325*Sin(A1)  + 0.000165*Sin(A2)  + 0.000164*Sin(A3) \
            + 0.000126*Sin(A4)  + 0.000110*Sin(A5)  + 0.000062*Sin(A6) \
            + 0.000060*Sin(A7)  + 0.000056*Sin(A8)  + 0.000047*Sin(A9) \
            + 0.000042*Sin(A10) + 0.000040*Sin(A11) + 0.000037*Sin(A12) \
            + 0.000035*Sin(A13) + 0.000023*Sin(A14)

    # times of the mean phases of the Moon, already affected by the Sun's
    # aberration and by the Moon's light-time (eq. 47.1)

    JDE = 2451550.09765 + 29.530588853*k \
            + 0.0001337*T*T \
            - 0.000000150*T*T*T \
            + 0.00000000073*T**4
    JDEsave = JDE

    # Eclipses correction to time of max eclipse (52.1)
    # sN: New Moon; sF: Full Moon
    arg = 2*MM;     sN  = 0.0161*Sin(arg); 
    arg = 2*F1;     sN += -0.0097*Sin(arg)
    arg = MM-M;     sN += 0.0073*E*Sin(arg)
    arg = MM+M;     sN += -0.0050*E*Sin(arg)
    arg = MM-2*F1;  sN += -0.0023*Sin(arg)
    arg = 2*M;      sN += 0.0021*E*Sin(arg)
    arg = MM+2*F1;  sN += 0.0012*Sin(arg)
    arg = 2*MM+M;   sN += 0.0006*E*Sin(arg)
    arg = 3*MM;     sN += -0.0004*Sin(arg)
    arg = M+2*F1;   sN += -0.0003*E*Sin(arg)
    arg = A1;       sN += 0.0003*Sin(arg)
    arg = M-2*F1;   sN += -0.0002*E*Sin(arg)
    arg = 2*MM-M;   sN += -0.0002*E*Sin(arg)
    arg = Omega;    sN += -0.0002*Sin(arg)

    sF = sN

    arg = MM;   sN += -0.4075*Sin(arg);     sF += -0.4065*Sin(arg)
    arg = M;    sN += 0.1721*E*Sin(arg);    sF += 0.1727*E*Sin(arg)

    # P and Q - p. 351
    Q = 5.2207
    arg = M;    P  = 0.2070*E*Sin(arg);     Q += -0.0048*E*Cos(arg)
    arg = 2*M;  P += 0.0024*E*Sin(arg);     Q += 0.0020*E*Cos(arg)
    arg = MM;   P += -0.0392*Sin(arg);      Q += -0.3299*Cos(arg)
    arg = 2*MM; P += 0.0116*Sin(arg)
    arg = MM+M; P += -0.0073*E*Sin(arg);    Q += -0.0060*E*Cos(arg)
    arg = MM-M; P += 0.0067*E*Sin(arg);     Q += 0.0041*E*Cos(arg)
    arg = 2*F1; P += 0.0118*Sin(arg)

    # WW, gamma, u
    # For a Solar eclipse, gamma is the least distance from the axis 
    # of the Moon's shadow to the center of the Earth, 
    # in units of the equatorial radius of the Earth.
    # gamma is positive if the shadow passes North of the center.
    # when gamma is between +0.9972 and -0.9972, the solar eclipse is central:
    # there exists a line of central eclipse on the Earth's surface.
    WW = abs(Cos(F1))
    gamma = (P*Cos(F1) + Q*Sin(F1)) * (1. - 0.0048*WW)

    # u denotes the radius of the Moon's umbral cone in the fundamental plane.
    # The radius of the penumbral cone in the fundamental plane is
    #   u + 0.5461

    u = 0.0059
    u += 0.0046*E*Cos(M)
    u += -0.0182*Cos(MM)
    u += 0.0004*Cos(2*MM)
    u += -0.0005*Cos(M+MM)

    # if abs(gamma) is between 0.9972 and 1.5433+u, the eclipse is not central.
    # If abs(gamma) is between 0.9972 and 1.5433 + u, the eclipse is not
    # central. In most cases, it is then a partial eclipse. However, when
    # abs(gamma) is between 0.9972 and 1.0260, a part of the umbral cone may
    # touch the surface of the Earth (within the polar regions), while the
    # axis of the cone does not touch the Earth. These non-central total
    # or annular eclipses occur when 0.9972 < abs(gamma) < 0.9972 + |u|.


    # CAREFUL: these tests must work for negative years as well!

    def isNewMoon(k):   # k must be integer
        test = (abs(k - int(k)) < 0.001)
        # if verbose: print ("isNewMoon k, test = ", k, test)
        return test     # True or False

    def isFullMoon(k):  # k must be half-integer
        test = (abs(abs(k - int(k)) - 0.5) < 0.001)
        # if verbose: print ("isFullMoon k, test = ", k, test)
        return test     # True or False 

    def isFirstQuarter(k):  # k must be .25
        test = abs(abs(k - floor(k)) - 0.25) < 0.001
        # if verbose: print ("isFirstQuarter k, test = ", k, test)
        return test     # True or False 

    def isLastQuarter(k):  # k must be .75
        test = abs(abs(ceil(k) - k) - 0.25) < 0.001
        # if verbose: print ("isLastQuarter k, test = ", k, test)
        return test     # True or False 

    JDEeclipse = None
    if isNewMoon(k):
        JDE += sNew
        JDEeclipse = hvfTime(JDEsave + sN)
        # if verbose: print ("New Moon")
    elif isFullMoon(k):
        JDE += sFull
        # if verbose: print ("Full Moon")
    elif isFirstQuarter(k):
        JDE += sQ
        JDE += W
        # if verbose: print ("First Quarter")
    elif isLastQuarter(k):
        JDE += sQ
        JDE -= W
        # if verbose: print ("Last Quarter")
    else:
        print ("k not for New or Full Moon: ", k)
    JDE += addCorr
    eclType = None
    omega = None
    SEpossible = False
    if isNewMoon(k) and abs(gamma) <= 1.5433+u: 
        SEpossible = True
        if abs(gamma) < 0.9972:
            if u < 0:
                eclType = ("Total Central Solar Eclipse")
            elif u < 0.0047:
                omega = 0.00464*sqrt(1-gamma**2)
                if u < omega:
                    eclType = ("Annular-total Central Solar Eclipse")
                else:
                    eclType = ("Annular Central Solar Eclipse")
            elif u >= 0.0047:
                eclType = ("Annular Central Solar Eclipse")
        elif abs(gamma) < 0.9972+abs(u):
            eclType = ("Non-central total or annular Solar eclipse")
        elif abs(gamma) < 1.5433+u:
            eclType = ("Non-central Solar Eclipse")
        if verbose:
            print ("getPhasesJDEbyK: k = %.2f" % k)
            print ("JDE = %.6f, %s" % (JDE, nicedatetime(hvfTime(JDE))))
            print ("JDEsave = %.6f, %s" % (JDEsave, nicedatetime(hvfTime(JDEsave))))
            print ("corrections sN, sF = ", sN, sF)
            #if JDEeclipse: print ("JDEeclipse = %.6f, %s" % (JDEeclipse, nicedatetime(hvfTime(JDEeclipse))))
            if JDEeclipse: print ("JDEeclipse = %s" % (nicedatetime(hvfTime(JDEeclipse))))
            print ("E, M(Sun), MM(Moon), F, Omega = %.7f, %.4f, %.4f, %.4f, %.4f" \
            % (E, M, MM, F, Omega)) 
            print ("F1, WW, gamma, u = %.4f, %.4f, %.4f, %.4f" % (F1, WW, gamma, u))
            print ("gamma = %.4f, closest distance of shadow to Earth center"%gamma)
            print ("Radius of umbral cone u = %.4f" % u)
            print ()
    
    EclipseDict = { "Object" : "Sun",
            "possible"  : SEpossible,
            "JDE"   : JDE,
            "JDEeclipse"   : JDEeclipse,
            "dt"    : JDEeclipse,
            "u"     : u,
            "gamma" : gamma,
            "omega" : omega,
            "event" : eclType,
            }
    if SEpossible and verbose: 
        print (EclipseDict)
        print ("JDEclipse = ", nicedatetime(hvfTime(JDEeclipse)))
        print ()


    return JDE, EclipseDict

# getMoonPhases(year, useUT)
#   returns:
#   newMoon
#   firstQuarter
#   fullMoon
#   lastQuarter
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
from operator import itemgetter

def getMoonPhases(year, useUT):
    # chapter 47 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-2000.)*12.3685  # relative moon phase at new year
    if (k0 == int(k0)): k = k0
    else: k  = ceil(k0) - 1     # start with new moon
    T  = k / 1236.85            # Julian centuries after Epoch2000
    y  = year
    EventList = []
    EventDistList = []
    eventType = ["New Moon", "First Quarter", "Full Moon", "Last Quarter"]
    EventTypeList = []
    # starting values
    #print ("getMoonPhases start: k, T, y =", k, T, y)

    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    while y == year:
        #print ("y, k, ", y, k)
        dateList = ['', '', '', ''] # new,first q, full, last q
        distList = ['', '']         # new, full moon
        for i in range(4):
            #verb = k == int(k)      # verbose
            verb = False
            JDE, EclipseDict = getPhasesJDEbyK(k, verb)  # TT
            dtE = getGreenwichDate(JDE)      # TT
            deltaT = timedelta(seconds=getDeltaT(JDE))
            dtUTC = dtE - deltaT
            dtDisplay = dtE
            if useUT: dtDisplay = dtUTC

            # getMoonData expects UTC time
            (MoonRA, MoonDecl, MoonDict) = getMoonData(dtUTC, 0, False)
            MoonDist = MoonDict["MoonDist"]
            # Super Moon, Mini Moon
            suffix = "  "
            if MoonDist <= 360000. : suffix = "++" # Super Moon
            if MoonDist >= 405000. : suffix = "--" # Mini Moon
            (yyyy, dummy, dummy) = dtUTC.getDate()

            # Event
            Event = {'Object': MoonDict["Object"],  # Moon
                    'dt'    : dtDisplay,
                    'event' : eventType[i],
                    'mode'  : eventType[i],
                    'Dict'  : MoonDict,
                    }
            EventTypeList.append(Event)
            # just pass eclipse dict, adjust dt for UTC or TT
            if EclipseDict["possible"] :
                JDEeclipseDisplay = EclipseDict["JDEeclipse"]
                if useUT:
                    JDEeclipseDisplay = JDEeclipseDisplay - deltaT    
                EclipseDict["dt"] = JDEeclipseDisplay        

                EventTypeList.append(EclipseDict)
            #if (dt.year == year): 
            if (yyyy == year): 
                dateList[i] = niceshortdatetime(dtDisplay)
                if i == 0:
                    distList[0] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
                if i == 2:
                    distList[1] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
            k += 0.25
            T = k / 1236.85
            #print ("k, T, yyyy =", k, T, yyyy)
        EventList.append(dateList)
        if len(distList[0]+distList[1]) > 0: EventDistList.append(distList)
        T = k / 1236.85
        #y = int(T * 100 + 2000)
        y = floor(T * 100 + 2000)
        #print ("k, T, y =", k, T, y)
   # sort events by time dt in EventTypeList (solar eclipses)
    SortedEventTypeList = sorted(EventTypeList, key = itemgetter('dt'))

    return (EventList, EventDistList, SortedEventTypeList)

#########################################################      

# getMoonPhasesxxx(dtCivicTime, zone, verbose=False)
#   
#   returns:
#   newMoon
#   firstQuarter
#   fullMoon
#   lastQuarter
#   for about 30 days before and after dtCivicTime
#
from operator import itemgetter

def getMoonPhasesxxx(dtCivicTime, zone, verbose=False):
    # chapter 47 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 
    #k0 = (yearxxx-2000.)*12.3685  # relative moon phase at given 
    k0 = Ty * 12.3685  # relative moon phase at given (47.2)
    k  = ceil(k0) - 1     # start with new moon
    T  = k / 1236.85      # Julian centuries after Epoch2000
    kmin = k - 2
    kmax = k + 3
    EventList = []
    EventDistList = []
    eventType = ["New Moon", "First Quarter", "Full Moon", "Last Quarter"]
    EventTypeList = []
    # starting values
    #print ("getMoonPhases start: k, T, y =", k, T, y)
    useUT = True
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    k = kmin
    for ki in range(kmin, kmax+1):
    #while y == year:
        #print ("y, k, ", y, k)
        dateList = ['', '', '', ''] # new,first q, full, last q
        distList = ['', '']         # new, full moon
        for i in range(4):
            #verb = k == int(k)      # verbose
            verb = False
            JDE, EclipseDict = getPhasesJDEbyK(k, verb)  # TT
            # only +- 3 weeks
            
            dtE = getGreenwichDate(JDE)      # TT
            deltaT = timedelta(seconds=getDeltaT(JDE))
            dtUTC = dtE - deltaT
            dtDisplay = dtE
            if useUT: dtDisplay = dtUTC

            # getMoonData expects UTC time
            (MoonRA, MoonDecl, MoonDict) = getMoonData(dtUTC, 0, False)
            MoonDist = MoonDict["MoonDist"]
            # Super Moon, Mini Moon
            suffix = "  "
            if MoonDist <= 360000. : suffix = "++" # Super Moon
            if MoonDist >= 405000. : suffix = "--" # Mini Moon
            (yyyy, dummy, dummy) = dtUTC.getDate()

            dtCT = getCivicTimefromUT(dtUTC, Zone)
            # Event
            #Event = {'Object': MoonDict["Object"],  # Moon
            Event = {'Object': "Moon Phases",
                    'dt'    : dtUTC,
                    'dtCivicTime' : dtCT,
                    'event' : eventType[i],
                    'mode'  : eventType[i],
                    'decl'  : MoonDecl,
                    'dist'  : MoonDist,
                    'suffix': suffix,
   
                    'Dict'  : MoonDict,
                    }
            EventTypeList.append(Event)
            # just pass eclipse dict, adjust dt for UTC or TT
            if EclipseDict["possible"] :
                JDEeclipseDisplay = EclipseDict["JDEeclipse"]
                if useUT:
                    JDEeclipseDisplay = JDEeclipseDisplay - deltaT    
                EclipseDict["dt"] = JDEeclipseDisplay        
                EclipseDict["dtCivicTime"] = \
                    getCivicTimefromUT(JDEeclipseDisplay, Zone)        
                EventTypeList.append(EclipseDict)
            if True:
                dateList[i] = niceshortdatetime(dtDisplay)
                if i == 0:
                    distList[0] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
                if i == 2:
                    distList[1] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
            k += 0.25
            T = k / 1236.85
        EventList.append(dateList)
        if len(distList[0]+distList[1]) > 0: EventDistList.append(distList)
        T = k / 1236.85
        #y = int(T * 100 + 2000)
        y = floor(T * 100 + 2000)
        #print ("k, T, y =", k, T, y)
   # sort events by time dt in EventTypeList (solar eclipses)
    SortedEventTypeList = sorted(EventTypeList, key = itemgetter('dt'))

    return (EventList, EventDistList, SortedEventTypeList)

#########################################################      
# Maximum Declinations of the Moon, Meeus Chapter 50
# hvf 02.05.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87
# replace TDT by TT (terrestrial time) - hvf 06.04.2023

import sys

VersionMoon = "V3.1 hvf 2023-11-29 MoonDecliMain.py using ELP-2000/82 and VSOP Ephemerides"
VersionList.append(VersionMoon)

def Usage(progname):
    print ("Usage: %s [UTC|TT] year [year ...]" % progname)
    sys.exit(0)

# getDecliJDEbyK (k, verbose = False)
# k counts the number of revolutions since 2000-01-01
# Meeus chapter 50
# returns the Julian Ephemeris Days (JDE) time and maximum declination the Moon
# both for Northern and Southern maxima
# return (JDEN, deltaN, JDES, deltaS)
def getDecliJDEbyK(k, verbose = False):
    # k integer:        Maximum Declination N, S since 2000-01-01

    T  = k / 1336.86            # (50.1), Julian centuries after Epoch2000
    E = 1 - 0.002516*T - 0.0000074*T*T      # (45.6)
    EE = E*E

    # DN, DS for North and South extrema
    DN = 152.2029 + 333.0705546*k - 0.0004025*T*T \
        + 0.00000011*T**3
    DN = Mod(DN, 360.)
    DS = 345.6676 + 333.0705546*k - 0.0004025*T*T \
        + 0.00000011*T**3
    DS = Mod(DS, 360.)

    # Sun's Mean anomaly 
    MN = 14.8591 + 26.9281592*k - 0.0000544*T*T - 0.00000010*T*T*T
    MN = Mod(MN, 360.)
    MS =  1.3981 + 26.9281592*k - 0.0000544*T*T - 0.00000010*T*T*T
    MS = Mod(MS, 360.)

    # Moon's Mean anomaly 
    MMN =  4.6881 + 356.9562795*k + 0.0103126*T*T \
            + 0.00001251*T**3
    MMN = Mod(MMN, 360.)
    MMS = 186.2100 + 356.9562795*k + 0.0103126*T*T \
            + 0.00001251*T**3
    MMS = Mod(MMS, 360.)

    FN = 325.8867 +1.4467806*k \
            - 0.0020708*T*T - 0.00000215*T**3 
    FN = Mod(FN, 360.)
    FS = 145.1633 +1.4467806*k \
            - 0.0020708*T*T - 0.00000215*T**3 
    FS = Mod(FS, 360.)

    JDEN = 2451562.5897 + 27.321582241*k \
            + 0.000100695*T*T \
            - 0.000000141*T*T*T 

    JDES = 2451548.9289 + 27.321582241*k \
            + 0.000100695*T*T \
            - 0.000000141*T*T*T 

    # Table 50.A periodic terms for the time of the Moon's maximum declination
    sumTimNorth = 0.8975 * Cos(FN) \
                - 0.4726 * Sin(MMN) \
                - 0.1030 * Sin(2*FN) \
                - 0.0976 * Sin(2*DN-MMN) \
                - 0.0462 * Cos(MMN-FN) \
                - 0.0461 * Cos(MMN+FN) \
                - 0.0438 * Sin(2*DN) \
                + 0.0162*E*Sin(MN) \
                - 0.0157 * Cos(3*FN) \
                + 0.0145 * Sin(MMN+2*FN) \
                + 0.0136 * Cos(2*DN-FN) \
                - 0.0095 * Cos(2*DN-MMN-FN) \
                - 0.0091 * Cos(2*DN-MMN+FN) \
                - 0.0089 * Cos(2*DN+FN) \
                + 0.0075 * Sin(2*MMN) \
                - 0.0068 * Sin(MMN-2*FN) \
                + 0.0061 * Cos(2*MMN-FN) \
                - 0.0047 * Sin(MMN+3*FN) \
                - 0.0043*E*Sin(2*DN-MN-MMN) \
                - 0.0040 * Cos(MMN-2*FN) \
                - 0.0037 * Sin(2*DN-2*MMN) \
                + 0.0031 * Sin(FN) \
                + 0.0030 * Sin(2*DN+MMN) \
                - 0.0029 * Cos(MMN+2*FN) \
                - 0.0029*E*Sin(2*DN-MN) \
                - 0.0027 * Sin(MMN+FN) \
                + 0.0024*E*Sin(MN-MMN) \
                - 0.0021 * Sin(MMN-3*FN) \
                + 0.0019 * Sin(2*MMN+FN) \
                + 0.0018 * Cos(2*DN-2*MMN-FN) \
                + 0.0018 * Sin(3*FN) \
                + 0.0017 * Cos(MMN+3*FN) \
                + 0.0017 * Cos(2*MMN) \
                - 0.0014 * Cos(2*DN-MMN) \
                + 0.0013 * Cos(2*DN+MMN+FN) \
                + 0.0013 * Cos(MMN) \
                + 0.0012 * Sin(3*MMN+FN) \
                + 0.0011 * Sin(2*DN-MMN+FN) \
                - 0.0011 * Cos(2*DN-2*MMN) \
                + 0.0010 * Cos(DN+FN) \
                + 0.0010*E*Sin(MN+MMN) \
                - 0.0009 * Sin(2*DN-2*FN) \
                + 0.0007 * Cos(2*MMN+FN) \
                - 0.0007 * Cos(3*MMN+FN) 
                
    sumTimSouth =-0.8975 * Cos(FS) \
                - 0.4726 * Sin(MMS) \
                - 0.1030 * Sin(2*FS) \
                - 0.0976 * Sin(2*DS-MMS) \
                + 0.0541 * Cos(MMS-FS) \
                + 0.0516 * Cos(MMS+FS) \
                - 0.0438 * Sin(2*DS) \
                + 0.0112*E*Sin(MS) \
                + 0.0157 * Cos(3*FS) \
                + 0.0023 * Sin(MMS+2*FS) \
                - 0.0136 * Cos(2*DS-FS) \
                + 0.0110 * Cos(2*DS-MMS-FS) \
                + 0.0091 * Cos(2*DS-MMS+FS) \
                + 0.0089 * Cos(2*DS+FS) \
                + 0.0075 * Sin(2*MMS) \
                - 0.0030 * Sin(MMS-2*FS) \
                - 0.0061 * Cos(2*MMS-FS) \
                - 0.0047 * Sin(MMS+3*FS) \
                - 0.0043*E*Sin(2*DS-MS-MMS) \
                + 0.0040 * Cos(MMS-2*FS) \
                - 0.0037 * Sin(2*DS-2*MMS) \
                - 0.0031 * Sin(FS) \
                + 0.0030 * Sin(2*DS+MMS) \
                + 0.0029 * Cos(MMS+2*FS) \
                - 0.0029*E*Sin(2*DS-MS) \
                - 0.0027 * Sin(MMS+FS) \
                + 0.0024*E*Sin(MS-MMS) \
                - 0.0021 * Sin(MMS-3*FS) \
                - 0.0019 * Sin(2*MMS+FS) \
                - 0.0006 * Cos(2*DS-2*MMS-FS) \
                - 0.0018 * Sin(3*FS) \
                - 0.0017 * Cos(MMS+3*FS) \
                + 0.0017 * Cos(2*MMS) \
                + 0.0014 * Cos(2*DS-MMS) \
                - 0.0013 * Cos(2*DS+MMS+FS) \
                - 0.0013 * Cos(MMS) \
                + 0.0012 * Sin(3*MMS+FS) \
                + 0.0011 * Sin(2*DS-MMS+FS) \
                + 0.0011 * Cos(2*DS-2*MMS) \
                + 0.0010 * Cos(DS+FS) \
                + 0.0010*E*Sin(MS+MMS) \
                - 0.0009 * Sin(2*DS-2*FS) \
                - 0.0007 * Cos(2*MMS+FS) \
                - 0.0007 * Cos(3*MMS+FS) 
                
    JDEN += sumTimNorth
    JDES += sumTimSouth

    # Table 50.B Periodic terms for the value of the Moon's maximum declination

    delta = 23.6961 - 0.013004*T

    sumDelNorth =  5.1093 * Sin(FN) \
                + 0.2658 * Cos(2*FN) \
                + 0.1448 * Sin(2*DN-FN) \
                - 0.0322 * Sin(3*FN) \
                + 0.0133 * Cos(2*DN-2*FN) \
                + 0.0125 * Cos(2*DN) \
                - 0.0124 * Sin(MMN-FN) \
                - 0.0101 * Sin(MMN+2*FN) \
                + 0.0097 * Cos(FN) \
                - 0.0087*E*Sin(2*DN+MN-FN) \
                + 0.0074 * Sin(MMN+3*FN) \
                + 0.0067 * Sin(DN+FN) \
                + 0.0063 * Sin(MMN-2*FN) \
                + 0.0060*E*Sin(2*DN-MN-FN) \
                - 0.0057 * Sin(2*DN-MMN-FN) \
                - 0.0056 * Cos(MMN+FN) \
                + 0.0052 * Cos(MMN+2*FN) \
                + 0.0041 * Cos(2*MMN+FN) \
                - 0.0040 * Cos(MMN-3*FN) \
                + 0.0038 * Cos(2*MMN-FN) \
                - 0.0034 * Cos(MMN-2*FN) \
                - 0.0029 * Sin(2*MMN) \
                + 0.0029 * Sin(3*MMN+FN) \
                - 0.0028*E*Cos(2*DN+MN-FN) \
                - 0.0028 * Cos(MMN-FN) \
                - 0.0023 * Cos(3*FN) \
                - 0.0021 * Sin(2*DN+FN) \
                + 0.0019 * Cos(MMN+3*FN) \
                + 0.0018 * Cos(DN+FN) \
                + 0.0017 * Sin(2*MMN-FN) \
                + 0.0015 * Cos(3*MMN+FN) \
                + 0.0014 * Cos(2*DN+2*MMN+FN) \
                - 0.0012 * Sin(2*DN-2*MMN-FN) \
                - 0.0012 * Cos(2*MMN) \
                - 0.0010 * Cos(MMN) \
                - 0.0010 * Sin(2*FN) \
                + 0.0006 * Sin(MMN+FN)


    sumDelSouth = -5.1093 * Sin(FS) \
                + 0.2658 * Cos(2*FS) \
                - 0.1448 * Sin(2*DS-FS) \
                + 0.0322 * Sin(3*FS) \
                + 0.0133 * Cos(2*DS-2*FS) \
                + 0.0125 * Cos(2*DS) \
                - 0.0015 * Sin(MMS-FS) \
                + 0.0101 * Sin(MMS+2*FS) \
                - 0.0097 * Cos(FS) \
                + 0.0087*E*Sin(2*DS+MS-FS) \
                + 0.0074 * Sin(MMS+3*FS) \
                + 0.0067 * Sin(DS+FS) \
                - 0.0063 * Sin(MMS-2*FS) \
                - 0.0060*E*Sin(2*DS-MS-FS) \
                + 0.0057 * Sin(2*DS-MMS-FS) \
                - 0.0056 * Cos(MMS+FS) \
                - 0.0052 * Cos(MMS+2*FS) \
                - 0.0041 * Cos(2*MMS+FS) \
                - 0.0040 * Cos(MMS-3*FS) \
                - 0.0038 * Cos(2*MMS-FS) \
                + 0.0034 * Cos(MMS-2*FS) \
                - 0.0029 * Sin(2*MMS) \
                + 0.0029 * Sin(3*MMS+FS) \
                + 0.0028*E*Cos(2*DS+MS-FS) \
                - 0.0028 * Cos(MMS-FS) \
                + 0.0023 * Cos(3*FS) \
                + 0.0021 * Sin(2*DS+FS) \
                + 0.0019 * Cos(MMS+3*FS) \
                + 0.0018 * Cos(DS+FS) \
                - 0.0017 * Sin(2*MMS-FS) \
                + 0.0015 * Cos(3*MMS+FS) \
                + 0.0014 * Cos(2*DS+2*MMS+FS) \
                + 0.0012 * Sin(2*DS-2*MMS-FS) \
                - 0.0012 * Cos(2*MMS) \
                + 0.0010 * Cos(MMS) \
                - 0.0010 * Sin(2*FS) \
                + 0.0037 * Sin(MMS+FS)

    deltaN = delta + sumDelNorth
    deltaS = -(delta + sumDelSouth)

    if verbose:
        print ("getDecliJDEbyK: k = %.2f" % k)
        dtN = getGreenwichDate(JDEN)
        dtS = getGreenwichDate(JDES)
        print ("JDEN = %.6f" % JDEN, nicedatetime(dtN), " deltaN = %.6f"%deltaN)
        print ("JDES = %.6f" % JDES, nicedatetime(dtS), " deltaS = %.6f"%deltaS)

    return (JDEN, deltaN, JDES, deltaS)

# getMoonDecli(year, useUT)
#   returns:
#   Ascending Nodes
#   Descending Nodes
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonDecli(year, useUT):
    # chapter 50 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-2000.03)*13.3686  # k0 sideric months since 2000-01-10
    if (k0 == int(k0)): kstart = k0
    else: kstart  = ceil(k0) - 1     # start with Ascending Node
    y  = year
    EventList = []
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    for k in range(kstart, kstart+16):
        dateList = ['', '']
        foundDate = False
        T  = k / 1336.86            # Julian centuries after Epoch2000
        y = int(T * 100 + 2000)
        (JDEN, deltaN, JDES, deltaS) = getDecliJDEbyK(k, False)
        dtN = getGreenwichDate(JDEN)
        dtS = getGreenwichDate(JDES)
        deltaTN = timedelta(seconds=getDeltaT(JDEN))
        deltaTS = timedelta(seconds=getDeltaT(JDES))
        if useUT: 
            dtN -= deltaTN
            dtS -= deltaTS
        # South is before North for a given k, so we display South first

        (yyyy, dummy, dummy) = dtS.getDate()
        #if dtS.year == year: 
        if yyyy == year: 
            dateList[0] = nicedateshorttime(dtS) + " %9.4f"%deltaS
            foundDate = True

        (yyyy, dummy, dummy) = dtN.getDate()
        #if dtN.year == year: 
        if yyyy == year: 
            dateList[1] = nicedateshorttime(dtN) + " %9.4f"%deltaN
            foundDate = True

        if foundDate: EventList.append(dateList)
    return EventList
#########################################################     

# getMoonDeclixxx(dtCivicTime, zone, verbose=False)
#   returns:
#   Ascending Nodes
#   Descending Nodes
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonDeclixxx(dtCivicTime, zone, verbose=False):
    # chapter 50 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 

    #k0 = (year-2000.03)*13.3686  # k0 sideric months since 2000-01-10
    k0 = (Ty-0.03)*13.3686  # k0 sideric months since 2000-01-10
    k  = ceil(k0) - 2     # start with Ascending Node
    kmin = k 
    kmax = k + 3
    EventList = []
    useUT = True
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    for k in range(kmin, kmax+1):
        dateList = ['', '']
        foundDate = False
        T  = k / 1336.86            # Julian centuries after Epoch2000
        y = int(T * 100 + 2000)
        (JDEN, deltaN, JDES, deltaS) = getDecliJDEbyK(k, False)
        dtN = getGreenwichDate(JDEN)
        dtS = getGreenwichDate(JDES)
        deltaTN = timedelta(seconds=getDeltaT(JDEN))
        deltaTS = timedelta(seconds=getDeltaT(JDES))
        if useUT: 
            dtN -= deltaTN
            dtS -= deltaTS
        dtCTN = getCivicTimefromUT(dtN, zone)
        dtCTS = getCivicTimefromUT(dtS, zone)


        mode = ''

        (yyyy, dummy, dummy) = dtS.getDate()
        #if dtS.year == year: 
        if True:
            #dateList[0] = nicedateshorttime(dtS) + " %9.4f"%deltaS
            #foundDate = True
            (dummy, dummy, MoonDict) = getMoonData(dtS, 0, False)
            MoonDist = MoonDict["MoonDist"]
            suffix = MoonDict["suffix"]
            mode = "South Decl"
            Event = {'Object': 'Declination',
                    'dt': dtS,
                    'dtCivicTime' : dtCTS,
                    'dist': MoonDist,
                    'decl': deltaS,
                    'mode': mode,
                    'suffix': suffix,
                    'Dict': MoonDict}
            EventList.append(Event)

        (yyyy, dummy, dummy) = dtN.getDate()
        #if dtN.year == year: 
        if True:
            #dateList[1] = nicedateshorttime(dtN) + " %9.4f"%deltaN
            (dummy, MoonDecl, MoonDict) = getMoonData(dtN, 0, False)
            MoonDist = MoonDict["MoonDist"]
            suffix = MoonDict["suffix"]
            mode = "North Decl"
            Event = {'Object': 'Declination',
                    'dt': dtN,
                    'dtCivicTime' : dtCTN,
                    'dist': MoonDist,
                    'decl': deltaN,
                    'mode': mode,
                    'suffix': suffix,
                    'Dict': MoonDict}
            EventList.append(Event)

    return EventList
#########################################################     
#! /usr/bin/env python3
# Perigee and Apogee of the Moon, Meeus Chapter 48
# hvf 28.02.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87

from datetime import datetime, date, timedelta
from math import sin, acos, cos, pi, radians, degrees, atan, atan2
from math import asin, tan, ceil, floor, sqrt, fabs
import sys

VersionMoonPeriApo ="V3.1 hvf 2023-11-29 MoonPeriMain.py based on Meeus, Astronomical Algorithms"
VersionList.append(VersionMoonPeriApo)

def Usage(progname):
    print ("Usage: %s [UTC|TT] year [year ...]" % progname)
    print ("Usage: %s [UTC|TT] year-year ...]" % progname)
    sys.exit(0)

def isPerigee(k):   # k must be integer
    test = (abs(k - int(k)) < 0.001)
    #if verbose: print ("isPerigee k, test = ", k, test)
    return test     # True or False

def isApogee(k):  # k must be half-integer
    test = (abs(abs(k - int(k)) - 0.5) < 0.001)
    #if verbose: print ("isApogee k, test = ", k, test)
    return test     # True or False 

# getPeriJDEbyK (k, verbose = False)
# k counts the number of perigees since 1999-12-22
# Meeus chapter 48
# returns the Julian Ephemeris Days (JDE) time of the perigee, apogee
def getPeriJDEbyK(k, verbose = False):
    # k integer:        Perigee
    # k integer + 0.5 : Apogee

    T  = k / 1325.55            # 48.3, Julian centuries after Epoch2000

    # mean perigee, apogee (48.1)
    JDE = 2451534.6698 + 27.55454988 * k \
                        - 0.0006886 * T*T \
                        - 0.000001098 * T**3 \
                        + 0.0000000052 * T**4
    JDE0 = JDE

    # Moon's mean elongation at time JDE 
    D = 171.9179  + 335.9106046 * k \
                    - 0.0100250 * T*T \
                    - 0.00001156 * T**3 \
                    + 0.000000055 * T**4
    D = Mod(D, 360.)

    # Sun's Mean anomaly
    M = 347.3477   + 27.1577721 * k \
                    - 0.0008323 * T*T \
                    - 0.0000010 * T**3 
    M = Mod(M, 360.)

    # Moon's argument of Latitude
    F = 316.6109  + 364.5287911 * k \
                    - 0.0125131 * T*T \
                    - 0.0000148 *T**3 
    F = Mod(F, 360.)

    # sumP: perigee sum (in days)
    # sumA: apogee sum (in days)
    # sum HPP: perigee parallax (in arcsec), table 48.B
    # sum HPA: apogee parallax (in arcsec), table 48.B
    sumHPP = 3629.215;      sumHPA = 3245.251
    sumP = 0.;                                  sumA = 0.
    arg = 2*D;      sumP += -1.6769 * Sin(arg); sumA += 0.4392 * Sin(arg)
    sumHPP +=  63.224 * Cos(arg);    sumHPA += -9.147 * Cos(arg)
    arg = 4*D;      sumP +=  0.4589 * Sin(arg); sumA += 0.0684 * Sin(arg)
    sumHPP +=  -6.990 * Cos(arg);    sumHPA +=  0.355 * Cos(arg)
    arg = 6*D;      sumP += -0.1856 * Sin(arg); sumA += 0.0144 * Sin(arg)
    sumHPP +=   1.927 * Cos(arg);    sumHPA +=  0.052 * Cos(arg)
    arg = 8*D;      sumP +=  0.0883 * Sin(arg); sumA += 0.0035 * Sin(arg)
    sumHPP +=  -0.702 * Cos(arg);    sumHPA +=  0.010 * Cos(arg)
    arg = 2*D-M;    sumP += (-0.0773 + 0.00019 * T) * Sin(arg); \
                                sumA += (0.0426 - 0.00011 * T) * Sin(arg)
    sumHPP += (2.834 - 0.0071 * T) * Cos(arg);    sumHPA +=  0.159 * Cos(arg)
    arg = M;        sumP += ( 0.0502 - 0.00013 * T) * Sin(arg); \
                                sumA += (0.0456 - 0.00011 * T) * Sin(arg)
    sumHPP += ( 0.696 - 0.0017 * T) * Cos(arg);    
    sumHPA += (-0.656 + 0.0016 * T) * Cos(arg)
    arg = 10*D;     sumP += -0.0460 * Sin(arg); sumA +=  0.0009 * Sin(arg)
    sumHPP +=  0.297 * Cos(arg)
    arg = 4*D-M;    sumP += ( 0.0422 - 0.00011 * T) * Sin(arg); \
                                sumA +=  0.0113 * Sin(arg)
    sumHPP += (-0.629 + 0.0016 * T) * Cos(arg);    sumHPA += 0.065 * Cos(arg)
    arg = 6*D-M;    sumP += -0.0256 * Sin(arg); sumA +=  0.0034 * Sin(arg)
    sumHPP +=  0.260 * Cos(arg);    sumHPA +=  0.014 * Cos(arg)
    arg = 12*D;     sumP +=  0.0253 * Sin(arg); sumA +=  0.0003 * Sin(arg)
    sumHPP += -0.138 * Cos(arg)
    arg = D;        sumP +=  0.0237 * Sin(arg); sumA += -0.0189 * Sin(arg)
    sumHPP += -1.263 * Cos(arg);    sumHPA += -0.841 * Cos(arg)
    arg = 8*D-M;    sumP +=  0.0162 * Sin(arg); sumA +=  0.0011 * Sin(arg)
    sumHPP += -0.127 * Cos(arg)
    arg = 14*D;     sumP += -0.0145 * Sin(arg)
    sumHPP +=  0.068 * Cos(arg)
    arg = 2*F;      sumP +=  0.0129 * Sin(arg); sumA +=  0.0212 * Sin(arg)
    sumHPP += -0.690 * Cos(arg);    sumHPA +=  0.697 * Cos(arg)
    arg = 3*D;      sumP += -0.0112 * Sin(arg); sumA += -0.0017 * Sin(arg)
    sumHPP +=  0.201 * Cos(arg)
    arg = 10*D-M;   sumP += -0.0104 * Sin(arg); sumA +=  0.0004 * Sin(arg)
    sumHPP +=  0.067 * Cos(arg)
    arg = 16*D;     sumP +=  0.0086 * Sin(arg)
    sumHPP += -0.035 * Cos(arg)
    arg = 12*D-M;   sumP +=  0.0069 * Sin(arg)
    sumHPP += -0.038 * Cos(arg)
    arg = 5*D;      sumP +=  0.0066 * Sin(arg); sumA += -0.0004 * Sin(arg)
    sumHPP += -0.079 * Cos(arg)
    arg = 2*D+2*F;  sumP += -0.0053 * Sin(arg); sumA +=  0.0047 * Sin(arg)
    sumHPP +=  0.104 * Cos(arg);    sumHPA +=  0.031 * Cos(arg)
    arg = 18*D;     sumP += -0.0052 * Sin(arg)
    sumHPP +=  0.019 * Cos(arg)
    arg = 14*D-M;   sumP += -0.0046 * Sin(arg)
    sumHPP +=  0.023 * Cos(arg)
    arg = 7*D;      sumP += -0.0041 * Sin(arg)
    sumHPP +=  0.037 * Cos(arg)
    arg = 2*D+M;    sumP +=  0.0040 * Sin(arg); sumA +=  0.0005 * Sin(arg)
    sumHPP += -0.161 * Cos(arg);    sumHPA +=  0.043 * Cos(arg)
    arg = 20*D;     sumP +=  0.0032 * Sin(arg)
    sumHPP += -0.010 * Cos(arg)
    arg = D+M;      sumP += -0.0032 * Sin(arg); sumA +=  0.0036 * Sin(arg)
    sumHPP +=  0.157 * Cos(arg);    sumHPA +=  0.127 * Cos(arg)
    arg = 16*D-M;   sumP +=  0.0031 * Sin(arg)
    sumHPP += -0.014 * Cos(arg)
    arg = 4*D+M;    sumP += -0.0029 * Sin(arg)
    sumHPP +=  0.054 * Cos(arg)
    arg = 9*D;      sumP +=  0.0027 * Sin(arg)
    sumHPP += -0.020 * Cos(arg)
    arg = 4*D+2*F;  sumP +=  0.0027 * Sin(arg); sumA +=  0.0013 * Sin(arg)
    sumHPP += -0.037 * Cos(arg)

    arg = 2*D-2*M;  sumP += -0.0027 * Sin(arg); sumA +=  0.0022 * Sin(arg)
    sumHPP +=  0.104 * Cos(arg);    sumHPA +=  0.022 * Cos(arg)
    arg = 4*D-2*M;  sumP +=  0.0024 * Sin(arg); sumA +=  0.0010 * Sin(arg)
    sumHPP += -0.038 * Cos(arg)
    arg = 6*D-2*M;  sumP += -0.0021 * Sin(arg); sumA +=  0.0004 * Sin(arg)
    sumHPP +=  0.022 * Cos(arg)
    arg = 22*D;     sumP += -0.0021 * Sin(arg)
    arg = 18*D-M;   sumP += -0.0021 * Sin(arg)
    arg = 6*D+M;    sumP +=  0.0019 * Sin(arg)
    sumHPP += -0.025 * Cos(arg)
    arg = 11*D;     sumP += -0.0018 * Sin(arg)
    sumHPP +=  0.011 * Cos(arg)
    arg = 8*D+M;    sumP += -0.0014 * Sin(arg)
    sumHPP +=  0.012 * Cos(arg)
    arg = 4*D-2*F;  sumP += -0.0014 * Sin(arg); sumA += -0.0004 * Sin(arg)
    sumHPP +=  0.013 * Cos(arg)
    arg = 6*D+2*F;  sumP += -0.0014 * Sin(arg); sumA +=  0.0004 * Sin(arg)
    sumHPP +=  0.017 * Cos(arg)
    arg = 3*D+M;    sumP +=  0.0014 * Sin(arg); sumA +=  0.0007 * Sin(arg)
    sumHPP += -0.030 * Cos(arg)
    arg = 5*D+M;    sumP += -0.0014 * Sin(arg)
    sumHPP +=  0.010 * Cos(arg)
    arg = 13*D;     sumP +=  0.0013 * Sin(arg)
    arg = 20*D-M;   sumP +=  0.0013 * Sin(arg)
    arg = 3*D+2*M;  sumP +=  0.0011 * Sin(arg)
    arg = 2*D-2*F+M;sumP += -0.0011 * Sin(arg)
    arg = D+2*M;    sumP += -0.0010 * Sin(arg)
    arg = 2*F-M;    sumP +=  0.0008 * Sin(arg)
    sumHPP +=  0.014 * Cos(arg)
    arg = 2*D+4*F;  sumP +=  0.0007 * Sin(arg)
    arg = 2*F-2*M;  sumP += -0.0006 * Sin(arg)
    arg = 2*D-2*F+2*M;sumP += -0.0006 * Sin(arg)
    arg = 24*D;     sumP +=  0.0006 * Sin(arg)
    arg = 4*D-4*F;  sumP +=  0.0005 * Sin(arg)
    arg = 2*D+2*M;  sumP +=  0.0005 * Sin(arg); sumA +=  0.0005 * Sin(arg)
    sumHPP += -0.023 * Cos(arg);    sumHPA +=  0.019 * Cos(arg)
    arg = D-M;      sumP += -0.0004 * Sin(arg); sumA += -0.0003 * Sin(arg)
    sumHPP +=  0.029 * Cos(arg)
    arg = 2*D-2*F;                              sumA += -0.0034 * Sin(arg)
    sumHPP += -0.392 * Cos(arg);    sumHPA += -0.023 * Cos(arg)
    arg = 2*M;                                  sumA +=  0.0006 * Sin(arg)
    sumHPP +=  0.023 * Cos(arg);    sumHPA += -0.016 * Cos(arg)
    arg = 2*F+M;                                sumA +=  0.0003 * Sin(arg)
    arg = 2*D+2*F-M;                            sumA +=  0.0003 * Sin(arg)
    arg = 2*D-2*F-M;sumHPP += -0.021 * Sin(arg)

    MoonDistP = 6378.14 / Sin(sumHPP/3600.)
    MoonDistA = 6378.14 / Sin(sumHPA/3600.)

    if isPerigee(k):
        JDE += sumP
        MoonDist = MoonDistP
        if verbose: print ("Perigee")
    elif isApogee(k):
        JDE += sumA
        MoonDist = MoonDistA
        if verbose: print ("Apogee")
    else:
        print ("k not for Perigee or Apogee: ", k)
    

    if verbose:
        print ("getPeriJDEbyK: k = %.2f" % k)
        print ("JDE0 = %.6f" % JDE0)
        print ("M(Sun), D, F, sumP, sumA = %.7f, %.4f, %.4f, %.4f, %.4f" \
            % (M, D, F, sumP, sumA)) 
        print ("sumP = %.5f, sumA = %.5f" % (sumP, sumA))
        print ("sumHPP = %.5f, sumHPA = %.5f" % (sumHPP, sumHPA))
        print ("JDE = %.5f" % JDE)
    return (JDE, MoonDist)

# getMoonPeriApo(year, useUT)
#   returns:
#   Perigee
#   Apogee
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonPeriApo(year, useUT):
    # chapter 48 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-1999.97)*13.2555  # counting perigee zero 48.2
    if (k0 == int(k0)): kstart = k0
    else: kstart  = ceil(k0) - 1     # start with Perigee
    
    EventList = []
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    print (" Moon Perigees and Apogees for the Year %4d" % year, comment)
    print (" -------------------------------------------", "-"*len(comment))
    print ()
    print ("             Perigee                          Apogee")
    
    for kint in range(kstart, kstart+16):
        dateList = ['', '']
        for k in (kint, kint+0.5):
            T  = k / 1325.55            # Julian centuries
            #y  = year
            y = int(T * 100 + 1999.97)
            (JDE, MoonDist) = getPeriJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT
        
            # call getMoonData for distance of moon
            #(MoonRa, MoonDecl, MoonDict) = getMoonData(dt-deltaT, 0.0, False)
            #MoonDist = MoonDict["MoonDist"]
            # Super Moon, Mini Moon
            suffix = "   "
            if MoonDist <= 360000. : suffix = " ++" # Super Moon
            if MoonDist >= 405000. : suffix = " --" # Mini Moon

    
            (yyyy, dummy, dummy) = dt.getDate()
            if isPerigee(k) and yyyy == year: 
                #print ("Perigee  ", niceshortdatetime(dt))
                dateList[0] = niceshortdatetime(dt) + " %.0f"%MoonDist + suffix
            if isApogee(k): 
                #print ("Apogee   ", niceshortdatetime(dt))
                #if dt.year == year: 
                if yyyy == year: 
                    dateList[1] = niceshortdatetime(dt) + " %.0f"%MoonDist + suffix
                if len (dateList[0]) + len(dateList[1]) > 0: 
                    EventList.append(dateList)
                dateList = ['', '']
    
    return EventList
#########################################################      

# getMoonPeriApoxxx(dtCivicTime, zone, verbose=False)
#   returns:
#   Perigee
#   Apogee
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonPeriApoxxx(dtCivicTime, zone, verbose=False):
    # chapter 48 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 

    #k0 = (year-1999.97)*13.2555  # counting perigee zero 48.2
    k0 = (Ty+0.03)*13.2555  # counting perigee zero 48.2
    k  = ceil(k0) - 2     # start with Perigee
    kmin = k
    kmax = k + 4
    EventList = []
    useUT = True
    
    for kint in range(kmin, kmax):
        for k in (kint, kint+0.5):
            T  = k / 1325.55            # Julian centuries
            y = int(T * 100 + 1999.97)
            (JDE, MoonDist) = getPeriJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT
            dtCT = getCivicTimefromUT(dt, zone)

            # call getMoonData for declination of Moon
            (dummy, dummy, MoonDict) = getMoonData(dt, 0, False)
            MoonDecl = MoonDict["Decl"]
            # Super Moon, Mini Moon
            suffix = MoonDict["suffix"]
            mode = ''
            if isPerigee(k): 
                mode = 'Perigee'
            if isApogee(k): 
                mode = 'Apogee'
            Event = {'Object': 'Perigee',
                    'dt': dt,
                    'dtCivicTime' : dtCT,
                    'dist': MoonDist,
                    'decl': MoonDecl,
                    'mode': mode,
                    'suffix': suffix,
                    'Dict': MoonDict}
            EventList.append(Event)
    return EventList
#########################################################      
# Nodes of the Moon, Meeus Chapter 49
# hvf 02.05.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87

import sys

VersionMoon = "V3.1 hvf 2023-11-29 MoonNodesMain.py using ELP-2000/82 and VSOP Ephemerides"
VersionList.append(VersionMoon)

def Usage(progname):
    print ("Usage: %s [UTC|TT] year [year ...]" % progname)
    sys.exit(0)

def isAscendingNode(k):   # k must be integer
    test = (abs(k - int(k)) < 0.001)
    #if verbose: print ("isAscendingNode k, test = ", k, test)
    return test     # True or False

def isDescendingNode(k):  # k must be half-integer
    test = (abs(abs(k - int(k)) - 0.5) < 0.001)
    #if verbose: print ("isDescendingNode k, test = ", k, test)
    return test     # True or False 

# getNodesJDEbyK (k, verbose = False)
# k counts the number of new moons since 2000-01-06
# Meeus chapter 49
# returns the Julian Ephemeris Days (JDE) time of the node passing  of the Moon
def getNodesJDEbyK(k, verbose = False):
    # k integer:        Ascending Node
    # k integer + 0.5 : Descending Node

    T  = k / 1342.23            # (49.1), Julian centuries after Epoch2000
    E = 1 - 0.002516*T - 0.0000074*T*T      # (45.6)
    EE = E*E

    # D
    D = 183.6380 + 331.73735691*k + 0.0015057*T*T \
        + 0.00000209*T**3 - 0.000000010*T**4
    D = Mod(D, 360.)

    # Sun's Mean anomaly 
    M = 17.4006 + 26.82037250*k - 0.0000999*T*T - 0.00000006*T*T*T
    M = Mod(M, 360.)

    # Moon's Mean anomaly 
    MM = 38.3776 + 355.52747322*k + 0.0123577*T*T \
            + 0.000014628*T**3 - 0.000000016*T**4
    MM = Mod(MM, 360.)

    # Longitude of the Ascending Node of the lunar orbit 
    Omega = 123.9767 - 1.44098949*k \
            + 0.0020625*T*T + 0.00000214*T**3 - 0.000000016*T**4
    Omega = Mod(Omega, 360.)

    V = 299.75 + 132.85*T - 0.009173*T*T

    P = Omega + 272.75 - 2.3*T

    JDE = 2451565.16195 + 27.212220817*k \
            + 0.0002572*T*T \
            + 0.000000021*T*T*T \
            - 0.00000000088*T**4 \
            - 0.4721 * Sin(MM) \
            - 0.1649 * Sin(2*D) \
            - 0.0868 * Sin(2*D-MM) \
            + 0.0084 * Sin(2*D+MM) \
            - 0.0083*E*Sin(2*D-M) \
            - 0.0039*E*Sin(2*D-M-MM) \
            + 0.0034 * Sin(2*MM) \
            - 0.0031 * Sin(2*D-2*MM) \
            + 0.0030*E*Sin(2*D+M) \
            + 0.0028*E*Sin(M-MM) \
            + 0.0026*E*Sin(M) \
            + 0.0025 * Sin(4*D) \
            + 0.0024 * Sin(D) \
            + 0.0022*E*Sin(M+MM) \
            + 0.0017 * Sin(Omega) \
            + 0.0014 * Sin(4*D-MM) \
            + 0.0005*E*Sin(2*D+M-MM) \
            + 0.0004*E*Sin(2*D-M+MM) \
            - 0.0003*E*Sin(2*D-2*M) \
            + 0.0003*E*Sin(4*D-M) \
            + 0.0003 * Sin(V) \
            + 0.0003 * Sin(P)

    if isAscendingNode(k):
        if verbose: print ("Ascending Node")
    elif isDescendingNode(k):
        if verbose: print ("Descending Node")
    else:
        print ("k not for Ascending or descending Node: ", k)

    if verbose:
        print ("getNodesJDEbyK: k = %.2f" % k)
        dt = getGreenwichDate(JDE)
        print ("JDE = %.6f" % JDE, nicedatetime(dt))
        print ("E, M(Sun), MM(Moon), D, Omega = %.7f, %.4f, %.4f, %.4f, %.4f" \
            % (E, M, MM, D, Omega)) 
        print ("V, P = %.7f, %.7f" % (V, P))
    return JDE

# getMoonNodes(year, useUT)
#   returns:
#   Ascending Nodes
#   Descending Nodes
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonNodes(year, useUT):
    # chapter 49 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-2000.05)*13.4223  # k0 = 0 is ascending node on 2000-01-21
    if (k0 == int(k0)): kstart = k0
    else: kstart  = ceil(k0) - 1     # start with Ascending Node
    y  = year
    EventList = []
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    for kint in range(kstart, kstart+16):
        dateList = ['', '']
        foundDate = False
        for k in(kint, kint+0.5):
            T  = k / 1342.23            # Julian centuries after Epoch2000
            y = int(T * 100 + 2000)
            JDE = getNodesJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT

            (yyyy, dummy, dummy) = dt.getDate()
            #if isAscendingNode(k) and dt.year == year: 
            if isAscendingNode(k) and yyyy == year: 
                dateList[0] = nicedatetime(dt)
                foundDate = True

            #if isDescendingNode(k) and dt.year == year: 
            if isDescendingNode(k) and yyyy == year: 
                dateList[1] = nicedatetime(dt)
                foundDate = True

        if foundDate: EventList.append(dateList)
        #print ("k, T, y =", k, T, y)
    return EventList
#########################################################     

# getMoonNodesxxx(dtCivicTime, zone, verbose=False)
#   returns:
#   Ascending Nodes
#   Descending Nodes
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonNodesxxx(dtCivicTime, zone, verbose=False):
    # chapter 49 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 
    k0 = (Ty-0.05)*13.4223  # k0 = 0 is ascending node on 2000-01-21
    kstart  = ceil(k0) - 1     # start with Ascending Node
    kmin = kstart - 2
    kmax = kstart + 3
    EventList = []
    useUT = True
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    for kint in range(kmin, kmax+1):
        dateList = ['', '']
        foundDate = False
        for k in(kint, kint+0.5):
            T  = k / 1342.23            # Julian centuries after Epoch2000
            y = int(T * 100 + 2000)
            JDE = getNodesJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT
            dtCT = getCivicTimefromUT(dt, zone)

           # call getMoonData for declination, distance of Moon
            (dummy, MoonDecl, MoonDict) = getMoonData(dt, 0, False)
            MoonDist = MoonDict['MoonDist']
            # Super Moon, Mini Moon
            suffix = MoonDict["suffix"]
            MoonDecl = MoonDict["Decl"]
            mode = '  '
            if isAscendingNode(k):
                mode = 'Ascending'
            if isDescendingNode(k):
                mode = 'Descending'

            # here we cheat a little bit: since getMoonData does not predict
            # latitude zero for the nodes, we set it here explicitly
            MoonDict['MoonLat'] = 0.0

            Event = {'Object': 'Nodes',
                    'dt': dt,
                    'dtCivicTime': dtCT,
                    'dist': MoonDist,
                    'decl': MoonDecl,
                    'mode': mode,
                    'suffix': suffix,
                    'Dict': MoonDict}

            EventList.append(Event)

    return EventList
#########################################################     

#######################################################################
# uses timezone data from pytz
# hvf 07.04.2021
# replace getRTS() by getObjectTimes() for more precise rise and set times
# hvf 15.06.2022 07.08.2022
# use hvfTime() instead of datetime()
# add Terrestrial Time TT to output - hvf 4.12.2022
# add duration of twilight phases - hvf 19.08.2023
import pytz
from itertools import chain

Version = "V8.8 hvf 2023-11-14 using VSOP87 ephemerides \n  and known timezones, flexible date input format, years from -9999 to 9999"
longVersion = "V8.8 hvf 2023-11-14 SunPlanetsMain-TZ.py using VSOP87 ephemerides \n  and known timezones, flexible date input format, years from -9999 to 9999"
VersionList.append(longVersion)

def shortUsage(progname):
    print ("Run %s --help for more explanations." % progname)

def Usage(progname):
    global Version
    #print ("%s %s" % (progname, Version))
    print ()
    print ("Run %s --help for this explanation." % progname)
    print ("Run %s --list for a list of the predefined locations." % progname)
    print ("Run %s --listtimezones for a list of all known timezones." %
            progname)
    print ("Run %s --check location to check if location is known or not." %
            progname)
    print ("Run %s --new newlocation --lon xx.xx --lat yy.yy for a new location, longitude and latitude must also be specified." % progname)
    print ("Run %s for default location, current date and time." % progname)
    print ("Run %s knownlocation for known location, current date and time." %
            progname)
    print ("  Known locations supply longitude, latitude, and timezone.")
    print ("  Longitude is in degrees East of Greenwich.")
    print ("  Latitude is in degrees, positive for North, negative for South.")
    print ("Additional Options:")
    print ("--date yyyy-mm-dd  Date in ISO format, e.g. 2021-05-19")
    print ("  also dd.mm.yyyy, -dd.mm.yyyy, -yyyy-mm-dd, yyyy between -9999 and 9999")
    print ("--time HH:MM:SS.sss Time in ISO format and variants, e.g. 13:15:23")
    print ("  also HH, HHMM, HHMMSS, HHMMSS:sss")
    print ("--zone hoursEast Timezone in hours East of Greenwich (UTC), e.g. 2")
    print ("--timezone knownTimezone  Timezone, e.g. Europe/Zurich")
    print ("  Date and Time default to current day and local time unless specified.")
    print ()
    return

# getUTfromCivicTimezone (dt, tdUTCOffset)
# dt datetime object
# tdUTCOffset   timedelta object
# returns a datetime object

def getUTfromCivicTimezone (dt, tdUTCOffset):
    return dt - tdUTCOffset

# niceTimeDeltaHHMM() returns a string +-HH:MM
def niceTimeDeltaHHMM(td):  
    # td is a timedelta, positive or negative
    if td.total_seconds() < 0.0:  sign = '-'; td = - td
    else:       sign = '+'
    hours = int(td.total_seconds() / 3600.)
    td -= timedelta(hours=hours)
    mins = int(td.total_seconds() / 60.)
    #secs = int(Mod(td.total_seconds(), 60.))
    tdstr = sign + "%2.2d:%2.2d" % (hours, mins)
    return tdstr

# getTZoffset(Timezone)
# returns the offset in seconds of dtCivicTime for Timezone compared to UTC
def getTZoffset(dtCivicTime, Timezone):
    #print (" dt = ", dtCivicTime, ", Timezone = ", Timezone)
    dt_with_tz = pytz.timezone(Timezone).localize(dtCivicTime)
    tdUTCOffset = dt_with_tz.utcoffset()
    #Zone = tdUTCOffset.total_seconds() / 3600.
    return tdUTCOffset   

def getTimezoneDetails(year):
# prints all timezones with their offsets
    jan = str(year) + "-01-10 1200"
    jul = str(year) + "-07-10 1200"
    print ("Timezones and their Offsets for year ", year)
    nlines = 0
    # timezones per country
    country_timezones_list = [i for i in pytz.country_timezones]
    for c in country_timezones_list:
        #print ("Country ", c, sorted(pytz.country_timezones(c)))
        for ctz in sorted(pytz.country_timezones(c)):
            if Mod(nlines, 50) == 0:
                print ()
                print ("Country   Timezone                    Offset in ", year)
                print ("                                      January  July")
            nlines += 1
            dtjan = dateutil.parser.isoparse(jan)
            dtjul = dateutil.parser.isoparse(jul)
            janOffset = getTZoffset(dtjan, ctz)
            julOffset = getTZoffset(dtjul, ctz)
            janStr = niceTimeDeltaHHMM(janOffset)
            julStr = niceTimeDeltaHHMM(julOffset)
            print ("%2.2s    %-30s  %6.6s   %6.6s" % (c, ctz, janStr,julStr))
    print ("%d timezones found." % nlines)
    nlines = 0

    # timezones not assigned to a country
    tz_notusedList = []
    for tz in pytz.all_timezones:
        country_in_tz = [c for c in pytz.country_timezones if tz in pytz.country_timezones(c)]
        if len(country_in_tz) > 0:
            #print ("Timezone", tz, country_in_tz)
            pass
        else:
            tz_notusedList.append(tz)
    for tz in sorted(tz_notusedList):
        if Mod(nlines, 50) == 0:
            print ()
            print ("Timezone                              Offset in ", year)
            print ("                                      January  July")
        nlines += 1
        dtjan = dateutil.parser.isoparse(jan)
        dtjul = dateutil.parser.isoparse(jul)
        janOffset = getTZoffset(dtjan, tz)
        julOffset = getTZoffset(dtjul, tz)
        janStr = niceTimeDeltaHHMM(janOffset)
        julStr = niceTimeDeltaHHMM(julOffset)
        print ("%-34s    %6.6s   %6.6s" % (tz, janStr,julStr))
    print ("%d timezones found." % nlines)
    return


# Main program
if __name__ == '__main__':
    
    # preset some locations
    Locations = {
        "Zurich": {"lon": 8.55, "lat": 47.3767, "zone": 1, "tz" : "Europe/Zurich"},
        "Gockhausen": {"lon": 8.592861, "lat": 47.381643, "zone": 1, "tz" : "Europe/Zurich"},
        "Lenk": {"lon": 7.44810, "lat": 46.46025, "zone": 1, "tz" : "Europe/Zurich"},
        "Clansayes": {"lon": 4.8273, "lat": 44.3923, "zone": 1, "tz" : "Europe/Paris"},
        "Paris": {"lon": 2.550, "lat": 49.013, "zone": 1, "tz" : "Europe/Paris"},
        "Greenwich": {"lon": 0.0, "lat": 51.48125, "zone": 0, "tz" : "GMT"},
        "Auckland": {"lon": 174.7667, "lat": -36.85, "zone": 12, "tz" : "Pacific/Auckland"},
        "Wellington": {"lon": 174.767, "lat": -41.283, "zone": 12, "tz" : "Pacific/Auckland"},
        "Christchurch": {"lon": 172.532, "lat": -43.489, "zone": 12, "tz" : "Pacific/Auckland"},
        "KapHoorn": {"lon": -67.2892, "lat": -55.9678, "zone": -3, "tz" : "America/Punta_Arenas"},
        "Hammerfest": {"lon": 23.6833, "lat": 70.6667, "zone": 1, "tz" : "Europe/Oslo"},
        "Barentsburg": {"lon": 14.2167, "lat": 78.0667, "zone": 1, "tz" : "Arctic/Longyearbyen"},
        "Nordpol": {"lon": 0.0, "lat": 90, "zone": 0, "tz" : "GMT"},
        "Suedpol": {"lon": 0.0, "lat": -90, "zone": 13, "tz" : "Antarctica/South_Pole"},
        "McMurdoStation": {"lon": 166.6667, "lat": -77.85, "zone": 13, "tz" : "Antarctica/McMurdo"},
        "HongKong": {"lon": 114.1667, "lat": 22.3, "zone": 8, "tz" : "Asia/Hong_Kong"},
        "Singapur": {"lon": 103.8333, "lat": 1.2833, "zone": 8, "tz" : "Asia/Singapore"},
        "Anchorage": {"lon": -149.8833, "lat": 61.1833, "zone": -8, "tz" : "America/Anchorage"},
        "Hawaii": {"lon": -157.85, "lat": 21.3167, "zone": -10, "tz" : "Pacific/Honolulu"},
        "Honolulu": {"lon": -157.85, "lat": 21.3167, "zone": -10, "tz" : "Pacific/Honolulu"},
        "Adelaide": {"lon": 138.600, "lat": -34.917, "zone": 9.5, "tz" : "Australia/Adelaide"},
        "Brisbane": {"lon": 153.033, "lat": -27.467, "zone": 10, "tz" : "Australia/Brisbane"},
        "BuenosAires": {"lon": -58.536, "lat": -34.822, "zone": -3, "tz" : "America/Argentina/Buenos_Aires"},
        "Canberra": {"lon": 149.117, "lat": -35.300, "zone": 10, "tz" : "Australia/Canberra"},
        "Sydney": {"lon": 151.200, "lat": -33.850, "zone": 10, "tz" : "Australia/Sydney"},
        "Chicago": {"lon": -87.905, "lat": 41.979, "zone": -5, "tz" : "America/Chicago"},
        "Doha": {"lon": 51.565, "lat": 25.261, "zone": 3, "tz" : "Asia/Qatar"},
        "Dubai": {"lon": 55.364, "lat": 25.253, "zone": 4, "tz" : "Asia/Dubai"},
        "Hobart": {"lon": 147.317, "lat": -42.867, "zone": 10, "tz" : "Australia/Hobart"},
        "Melbourne": {"lon": 144.950, "lat": -37.800, "zone": 10, "tz" : "Australia/Melbourne"},
        "Perth": {"lon": 115.850, "lat": -31.950, "zone": 8, "tz" : "Australia/Perth"},
        "Johannesburg": {"lon": 28.067, "lat": -26.200, "zone": 2, "tz" : "Africa/Johannesburg"},
        "Kapstadt": {"lon": 18.597, "lat": -33.969, "zone": 2, "tz" : "Africa/Johannesburg"},
        "LosAngeles": {"lon": -118.408, "lat": 33.943, "zone": -7, "tz" : "America/Los_Angeles"},
        "SanFrancisco": {"lon": -122.375, "lat": 37.619, "zone": -7, "tz" : "America/Los_Angeles"},
        "NewYork": {"lon": -73.779, "lat": 40.640, "zone": -4, "tz" : "America/New_York"},
        "Montreal": {"lon": -73.567, "lat": 45.500, "zone": -4, "tz" : "America/Montreal"},
        "Vancouver": {"lon": -123.184, "lat": 49.194, "zone": -7, "tz" : "America/Vancouver"},
        "Santiago": {"lon": -70.667, "lat": -33.450, "zone": -3, "tz" : "America/Santiago"},
        "Ushuaia": {"lon": -68.300, "lat": -54.800, "zone": -3, "tz" : "America/Argentina/Ushuaia"},
        "Tokyo": {"lon": 139.7667, "lat": 35.6833, "zone": 9, "tz" : "Asia/Tokyo"},
        "Peking": {"lon": 116.3833, "lat": 39.9333, "zone": 8, "tz" : "Asia/Shanghai"},
        "Moskau": {"lon": 37.6167, "lat": 55.7500, "zone": 3, "tz" : "Europe/Moscow"},
        "Kiev": {"lon": 30.5164, "lat": 50.4536, "zone": 2, "tz" : "Europe/Kiev"},

    }

############### input reading and parsing ######################    
    print ("###########################################################")
    progname = os.path.basename(sys.argv[0])
    for arg in sys.argv:
        print ("%s " % arg, end='')
    print ()
    print ("%s Version %s" % (progname, Version))
    # use current day and time as starting point unless explicit data are given
    # UTC as default
    locationFlag = False
    newlocFlag = False
    lonFlag = False
    latFlag = False
    Zone = 0
    Timezone = "UTC"   # timezone
    zoneFlag = False
    tzFlag = False
    dateFlag = False
    timeFlag = False
    day = None
    daytime = None
    # read input parameters: location latitude longitude [Zone yyyy-mm-dd HH:MM:SS]
    # --help
    # --list
    # --listtimezones
    # --check location
    # location
    # --new location
    # --lon longitude
    # --lat latitude
    # --zone
    # --timezone
    # --date date
    # --time time

    # set default location and current date and time
    if len(sys.argv) <= 1:
        # Gockhausen (from Google Maps)
        location = "Gockhausen (CH)"
        longitude = 8.592861
        latitude =  47.381643
        Timezone = "Europe/Zurich"
        tzFlag = True
        lonFlag = True
        latFlag = True

    while len(sys.argv) > 1:
        param = sys.argv[1]
        # split into words if necessary
        paramList = param.split()
        # print ("ParamList = ", paramList)
        if len(paramList) > 1:
            sys.argv[2:2] = paramList
            # print ("sys.argv = ", sys.argv)
            param = sys.argv[2]
            # print ("param = ", param)
            del sys.argv[1:2]
            # print ("sys.argv = ", sys.argv)
            # print ("param = ", param)
        del sys.argv[1]

        if param == '--help':
            for v in VersionList:
                print (v)
            Usage(progname)
            sys.exit(0)

        elif param == '--list':
            print ("Location      Latitude  Longitude    Zone Timezone")
            for location in sorted(Locations):
                longitude = Locations[location]["lon"]
                latitude  = Locations[location]["lat"]
                Zone      = Locations[location]["zone"]
                Timezone  = Locations[location]["tz"]
                print ("%-12s  %8.3f   %8.3f   %5.1f %s" % \
                    (location, latitude, longitude, Zone, Timezone))
            sys.exit(0)

        elif param == '--listtimezones':
            dt = dateutil.parser.isoparse(str(date.today()))
            getTimezoneDetails(dt.year)
            sys.exit(0)

        elif param == '--check':
            try:
                location = sys.argv[1]
                del sys.argv[1]
            except Exception as e:
                print ("Missing Parameter: Location\n") 
                shortUsage(progname)
                sys.exit(0)

            if location in Locations:
                longitude = Locations[location]["lon"]
                latitude  = Locations[location]["lat"]
                Zone      = Locations[location]["zone"]
                Timezone  = Locations[location]["tz"]
                print ("%-12s  %8.3f   %8.3f   %5.1f %s" % \
                    (location, latitude, longitude, Zone, Timezone))
                lonFlag = True
                latFlag = True
                tzFlag  = True
            else:
                print ("Location %s is not known.\n" % location)
                shortUsage(progname)
            sys.exit(0)

        elif param == '--new':
            try:
                location = sys.argv[1]
                del sys.argv[1]
                newlocFlag = True
                locationFlag = True
            except Exception as e:
                print ("Missing Parameter: Location\n") 
                shortUsage(progname)
                sys.exit(0)

        elif param == '--lon':
            try:
                longitude = float(sys.argv[1])
                del sys.argv[1]
                lonFlag = True
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)

        elif param == '--lat':
            try:
                latitude = float(sys.argv[1])
                del sys.argv[1]
                latFlag = True
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)
            
        elif param == '--zone':
            try:
                Zone = float(sys.argv[1])
                del sys.argv[1]
                zoneFlag = True
                Timezone = 'GMT'
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)

        elif param == '--timezone':
            try:
                Timezone = sys.argv[1]
                del sys.argv[1]
                if Timezone not in pytz.all_timezones_set:
                    print ("Timezone %s is not valid." % Timezone)
                    Usage(progname)
                    sys.exit(0)
                zoneFlag = False
                tzFlag = True
            except Exception as e:
                print ("Error ", e)
                shortUsage(progname)
                sys.exit(0)

        elif param == '--date':
            try:
                day = sys.argv[1]
                del sys.argv[1]
                dateFlag = True
                # print ("date day = ", day)
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)

        elif param == '--time':
            try:
                daytime = sys.argv[1]
                del sys.argv[1]
                timeFlag = True
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)

        # Lookup known locations
        elif param in Locations:
            location  = param
            longitude = Locations[location]["lon"]
            latitude  = Locations[location]["lat"]
            Zone      = Locations[location]["zone"]
            Timezone  = Locations[location]["tz"]
            tzFlag = True
            lonFlag = True
            latFlag = True
            zoneFlag = False
            locationFlag = True
        else:
        # Unknown location, the whole set
            print ("Parameter %s not known." % param)
            shortUsage(progname)
            sys.exit(0)
    # no location set, use our default
        if not locationFlag:
        # Gockhausen (from Google Maps)
            location = "Gockhausen (CH)"
            longitude = 8.592861
            latitude =  47.381643
            Timezone = "Europe/Zurich"
            tzFlag = True
            lonFlag = True
            latFlag = True
            locationFlag = True

    # yyyy-mm-dd
    today = hvfTime.today() # (y, m, d), d has fractions
    #print ("today = ", today)
    # HH:MM:SS
    now = hvfTime.now()     # (h, m, s)
    if dateFlag: 
        td = day    # string
    else:
        #td = today  # tupel
        td = "%04d-%02d-%02d" % today

    if timeFlag:
        tt = daytime    # string
    else:
        #tt = now        # tupel
        tt = "%02d:%02d:%02d" % now
    t1 = str(td) + " " + tt
    # dt is a hvfTime object
    try:
        #dttemp = dateutil.parser.isoparse(t1)
        dttemp = parseDateTime(t1)
        dt = hvfTime(dttemp)
    except Exception as e:
        print ("Error ", e)
        print ("Cannot parse '%s' into datetime." % t1)
        shortUsage(progname)
        sys.exit(0)

########################################################################        
    # apply timezone
    if zoneFlag:    # offset in hours
        dt_with_tz  = getUTfromCivicTime(dt, Zone)
        dt_with_tz  = dt
        tdUTCOffset = timedelta(hours=Zone)
        tdstring = niceTimeDeltaHHMM(tdUTCOffset)
        Timezone = tdstring
        zonetime_with_tz = str(dt_with_tz) + tdstring
    elif tzFlag:    # offset by name
        try:
            dt_with_tz = pytz.timezone(Timezone).localize(dttemp.getDatetime())
            tzinfo = dt_with_tz.tzinfo
            #print ("tzinfo = ", tzinfo)
            tdUTCOffset = dt_with_tz.utcoffset()
            Zone = tdUTCOffset.total_seconds() / 3600.
            tdstring = niceTimeDeltaHHMM(tdUTCOffset)
            zonetime_with_tz = dt_with_tz
        except:     # year outside datetime range
            Timezone = 'UTC'
            Zone = 0
            dt_with_tz  = dt
            tdUTCOffset = timedelta(hours=Zone)
            tdstring = niceTimeDeltaHHMM(tdUTCOffset)
            zonetime_with_tz = str(dt_with_tz) + tdstring

    #print ("utcoffset = ", tdstring)
    #print ("Timezone = ", Timezone)

    # check longitude and latitude
    if not lonFlag:
        print ("No Longitude specified.\n")
        shortUsage(progname)
        sys.exit(0)
    if not latFlag:
        print ("No Latitude specified.\n")
        shortUsage(progname)
        sys.exit(0)

    if abs(latitude) > 90.0:
        print ("Latitude ", latitude, " must be between -90 and +90 degrees.")
        shortUsage(progname)
        sys.exit(0)
########################################################################
# hier geht es richtig los
    verbose = False
    print ("Location                    =", location)
    print ("Latitude                    =", latitude, " degrees N")
    print ("Longitude                   =", longitude, " degrees E")
    print ("Timezone                    =", Timezone)
    print ("Zonetime, Day of Week       =", dt, getDayofWeek(dt))
    print ("Zonetime with TZ            =", zonetime_with_tz)
    print ("Zone                        =", Zone, "hours East of Greenwich") 
    dtLocalCivicTime = dt
    # convert local time zone to Greenwich mean time (UTC)
    #UTC = getUTfromCivicTime(dt, Zone)
    UTC = getUTfromCivicTimezone(dt, tdUTCOffset)
    # mean local time
    meanLT = getMeanLocalTime(UTC, longitude)
    print ("UTC Universal Time          =", UTC)
    print ("Mean Local Time             =", meanLT)

    # Zeitgleichung, wahre Ortszeit
    # Julian Day für Greenwich mean time
    #dtGMT = dt-timedelta(hours=Zone)
    dtGMT = UTC
    JD0 = getJulianDay(dtGMT)
    # find Julian Date for midnight GMT
    #dtGMTmidnight = datetime(dtGMT.year, dtGMT.month, dtGMT.day)
    (y, m, d) = dtGMT.getDate()
    d = floor(d)
    dtGMTmidnight = hvfTime(y, m, d, 0, 0, 0)
    JD1 = getJulianDay(dtGMTmidnight)
    # td is time eq. as a timedelta object, zgstr is a printable 
    (td, zgstr, SunStuffDict) = getZeitgleichung(JD0, False)
    # wahre Ortszeit
    trueLT = meanLT + td
    # Greenwich date
    greenwichDate = getGreenwichDate(JD0)
    greenwichDateJD1 = getGreenwichDate(JD1)
    # Greenwich mean sideral time
    greenwichST = getGMSTfromJD(JD0)
    localST = getLMSTfromGMST(greenwichST, longitude)
    deltaT = getDeltaT(JD0)
    TT = greenwichDate + timedelta(seconds=deltaT)
    JDE = getJulianDay(TT)    # Julian Ephemeris Day

    print ("Time Equation               = %s" % zgstr, " minutes")
    print ("True Local Time             = %s" % nicedatetime(trueLT))
    print ("Julian Day  at midnight GMT = ", JD1)
    print ("Greenwich Date midnight GMT = ", nicedatetime(greenwichDateJD1))
    print ()
    print ("Julian Day JD               (current time) = ", JD0)
    print ("Greenwich Date              (current time) = ", nicedatetime(greenwichDate))
    print ("Greenwich Mean Sideral Time (current time) = ", nicedatetime(greenwichST))
    print ("Local Mean Sideral Time     (current time) = ", nicedatetime(localST))
    print ("Longitude (hours)           = ", hmsstring(longitude/15.))
    print ("Mean Local Noon             = ", hmsstring(12.0+Zone-longitude/15.))
    print ("deltaT (sec)                = ", round(deltaT, 1))
    print ("Terrestrial Time TT         (current time) = ", nicedatetime(TT))
    print ("Julian Ephemeris Day JDE    (current time) = ", JDE)
    print ()
    # get Sun Decl, RA, hour angle, altitude, azimut
    # at current local civic time
    (RA, Decl, SunDict) = getSunData(dt, Zone, False)
    RAh = RA/15. # convert to hours
    # Stundenwinkel, hour angle
    Hour = localST - timedelta(hours=RAh)
    # use hvfTime()
    (h, m, s) = Hour.getTime()
    Hourdecimal = h + m/60. + s/3600.
    Hourdecimalstring = "%02d:%02d:%02d" % (h, m, s)
    #Hourdecimal = Hour.time().hour + Hour.time().minute/60. \
    #            + Hour.time().second/3600 + Hour.time().microsecond/3600e6
    Hourgrad = Hourdecimal*15.
    # get azimut, altitude from hour angle, declination, latitude
    (azi, alt, xyzDict) = getAziAltfromHourDecl(Hourgrad, Decl, latitude)
    SunSign = getSign(SunDict["SunLong"])

    print ("Sun Longitude     (degrees) = %8.3f" % SunDict["SunLong"])
    print ("Sun Latitude      (degrees) = %8.5f" % SunDict["SunLat"])
    print ("Sun RA            (hours)   = ", hmsstring(RAh))
    print ("Sun RA            (degrees) = %8.3f" % RA)
    print ("Sun Declination   (degrees) = %8.5f" % Decl)
    print ("Sun Hour Angle    (hours)   = ", Hourdecimalstring)
    print ("Sun Hour Angle    (degrees) = ", round(Hourgrad, 3))
    print ("Azimut            (degrees) = %6.1f" % azi)
    print ("Altitude          (degrees) = %6.1f (no refraction)" % alt)
    print ("SunSign                     = ", SunSign)
    print ("Distance Sun-Earth     (km) =  %d" % SunDict["R"])
    print ("Distance Sun-Earth     (au) =  %.8f" % SunDict["Rau"])
    print ()

#############  Sun Rise time, Set time ##############################
    # get transit time, altitude, declination
    SunEventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                None, False)

    #print ("SunEventList = ", SunEventList)
    # Sun  Rise, Set, and twilight events
    TwilightList = []
    RiseTime = None; SetTime = None # initialize for later
    RiseList = [e for e in SunEventList if e["event"] == "Rise"]
    for e in RiseList:
        RiseTime = e["dt"]
        azi = e["azi"]
        alt = e["alt"]
        TwilightList.append((RiseTime, "Sun Rise", azi, alt))
    SetList = [e for e in SunEventList if e["event"] == "Set"]
    for e in SetList:
        SetTime = e["dt"]
        azi = e["azi"]
        alt = e["alt"]
        TwilightList.append((SetTime, "Sun Set", azi, alt))
 
    # find lowest and highest altitude SunMinAlt and SunMaxAlt
    # we will need this to find out if we have twilight, midnight sun, etc.
    Alt = SunEventList[0]["Dict"]["Alt"]
    AltList = []
    # changed search string - hvf 13.03.2023
    for e in SunEventList:
        if "North " in e["event"]:
            AltList.append(e["alt"])
        if "South " in e["event"]:
            AltList.append(e["alt"])
            

    SunMinAlt = min(AltList)
    SunMaxAlt = max(AltList)
    #################3
    #print ("SunMinAlt = ", SunMinAlt)
    #print ("SunMaxAlt = ", SunMaxAlt)

    # civil twilight, 6 degrees below horizon
    Sun06EventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                -6., False)
    # find Rise event
    Rise06List = [e for e in Sun06EventList if e["event"] == "Rise"]
    for e in Rise06List:
        Rise06Time = e["dt"]
        TwilightList.append ((Rise06Time, "Civil Twilight Begin"))

    # find Set event - may have 2 Set events, e.g 2022-06-09 Gockhausen
    Set06List = [e for e in Sun06EventList if e["event"] == "Set"]
    for e in Set06List:
        Set06Time = e["dt"]
        TwilightList.append ((Set06Time, "Civil Twilight End"))

    # nautical twilight, 12 degrees below horizon
    Sun12EventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                -12., False)
    # find Rise event
    Rise12List = [e for e in Sun12EventList if e["event"] == "Rise"]
    for e in Rise12List:
        Rise12Time = e["dt"]
        TwilightList.append ((Rise12Time, "Nautical Twilight Begin"))

    # find Set event - may have 2 Set events, e.g 2022-06-09 Gockhausen
    Set12List = [e for e in Sun12EventList if e["event"] == "Set"]
    for e in Set12List:
        Set12Time = e["dt"]
        TwilightList.append ((Set12Time, "Nautical Twilight End"))


    # astronomical twilight, 18 degrees below horizon
    Sun18EventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                -18., False)
    # find Rise event
    Rise18List = [e for e in Sun18EventList if e["event"] == "Rise"]
    for e in Rise18List:
        Rise18Time = e["dt"]
        TwilightList.append ((Rise18Time, "Astronomical Twilight Begin"))

    # find Set event - may have 2 Set events, e.g 2022-06-09 Gockhausen
    Set18List = [e for e in Sun18EventList if e["event"] == "Set"]
    for e in Set18List:
        Set18Time = e["dt"]
        TwilightList.append ((Set18Time, "Astronomical Twilight End"))

    # find Sun alt at midnight 
    #dtMidnight = datetime(dt.year, dt.month, dt.day)
    (y, m, d) = dt.getDate()
    d = floor(d)
    dtMidnight = hvfTime(y, m, d, 0, 0, 0)
 
    (dummy, altMidnight, MidnightDict) = getAziAltfromCivicTimeLongLat \
            (dtMidnight, Zone, getSunData, longitude, latitude)
    dtMidnight1 = dtMidnight + timedelta(hours=24.)
    (dummy, altMidnight1, MidnightDict1) = getAziAltfromCivicTimeLongLat \
            (dtMidnight1, Zone, getSunData, longitude, latitude)

    Alt = MidnightDict["Alt"]
    Alt1 = MidnightDict1["Alt"]
    if altMidnight < -18.:
        TwilightList.append ((dtMidnight, "Deep Night"))
    elif altMidnight < -12.:
        TwilightList.append ((dtMidnight, "Astronomical Twilight"))
    elif altMidnight < -6.:
        TwilightList.append ((dtMidnight, "Nautical Twilight"))
    elif altMidnight < Alt:
        TwilightList.append ((dtMidnight, "Civil Twilight"))

    if altMidnight1 < -18.:
        TwilightList.append ((dtMidnight1, "Deep Night"))
    elif altMidnight1 < -12.:
        TwilightList.append ((dtMidnight1, "Astronomical Twilight"))
    elif altMidnight1 < -6.:
        TwilightList.append ((dtMidnight1, "Nautical Twilight"))
    elif altMidnight1 < Alt:
        TwilightList.append ((dtMidnight1, "Civil Twilight"))

    # print TwilightList
    #print ("Twilight List = ", TwilightList)
    SortedTwilightList = sorted(TwilightList, key = itemgetter(0))
    #print ("Sorted Twilight List = ", SortedTwilightList)

    # get length of each twilight phase
    TwilightLength = []
    for i in range(1, len(SortedTwilightList)):
        e1 = SortedTwilightList[i-1]
        e = SortedTwilightList[i]
        tt = (e[0] - e1[0])
        nicet = nicetimedelta(tt)
        TwilightLength.append(nicet)
    #print ("TwilightLength = ", TwilightLength)

    # print Twilight
    SunSeen = False
    for e in SortedTwilightList:
        try:
            twiLen = TwilightLength.pop(0)
        except:
            twiLen = ''
        if not SunSeen and "Sun" in e[1]: 
            print ()        # blank line before Sun Rise
            SunSeen = True
        elif SunSeen and "Sun" in e[1]: 
            # keep twiLen
            TwilightLength.insert(0, twiLen)
        if SunSeen and "Sun" not in e[1]: 
            print ()        # blank line after Sun Set
            SunSeen = False
            if RiseTime and SetTime:
                daylength = SetTime - RiseTime
                print ("Day Length (hours)          = ", 
                    nicetimedelta(daylength))
                print ()
        if len(e) > 2:      # Sun gets also Azimut: Sun Rise, Sun Set
            print ("%-27s = %s   Azimut = %3d" % (e[1], nicedatetime(e[0]),
                round(e[2], 0)))
        else:               # twilight
            print ("%-27s = %s   Dauer %s" % (e[1], nicedatetime(e[0]), twiLen))

    # Midnight Sun
    if SunMinAlt >  Alt:
        print ("Sun never sets, Midnight Sun")
    if SunMaxAlt <= Alt:
        print ()
        print ("Sun never rises above horizon, Polar Night")
    print ()
############## SUN EVENTS, ORDERED BY TIME OF DAY ######################
    SunEventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                None, False)
    print ("%-4s Events:" % SunEventList[0]["Object"])
    for e in SunEventList:
        print ("%-4s %-20s %s Azi %3d Alt %7.3f Di %.0f" % (e["Object"], 
                e["event"], nicedatetime(e['dt']), round(e["azi"], 0), 
                e["alt"], e["Dict"]["R"]))

    print ()

############## MOON DATA ####################
    print ("Moon Position")
    (MoonRA, MoonDecl, MoonDict) = getMoonData(dt, Zone, False)
    MoonLong = MoonDict["MoonLong"]
    MoonLat  = MoonDict["MoonLat"]
    MoonDist = MoonDict["MoonDist"]
    MoonHP   = MoonDict["MoonHP"]
    MoonAlt  = MoonDict["Alt"]

    print ("Moon Longitude    (degrees) = %8.3f" % MoonLong)
    print ("Moon Latitude     (degrees) = %8.3f" % MoonLat)
    print ("Moon RA           (hours)   = ", hmsstring(MoonRA/15.))
    print ("Moon RA           (degrees) = %8.3f" % MoonRA)
    print ("Moon Declination  (degrees) = %8.3f" % MoonDecl)

    # hour angle
    MoonHour = localST - timedelta(hours=MoonRA/15.)
    # use hvfTime()
    (h, m, s) = MoonHour.getTime()
    MoonHourdecimal = h + m/60. + s/3600.
    MoonHourdecimalstring = "%02d:%02d:%02d" % (h, m, s)
    print ("Moon Hour Angle   (hours)   = ", MoonHourdecimalstring)
    #MoonHourdecimal = MoonHour.time().hour + MoonHour.time().minute/60. \
    #     + MoonHour.time().second/3600 + MoonHour.time().microsecond/3600e6
    MoonHourgrad = MoonHourdecimal*15.
    print ("Moon Hour Angle   (degrees) = %8.3f" % MoonHourgrad)
    (azi, alt, xyzDict) = getAziAltfromHourDecl(MoonHourgrad, MoonDecl,latitude)
    print ("Azimut            (degrees) = %6.1f" % azi)
    print ("Altitude          (degrees) = %6.1f (no refraction)" % alt)
    
    suffix = ""
    if MoonDist <= 360000. : suffix = " ++"
    if MoonDist >= 405000. : suffix = " --"
    print ("Distance Moon-Earth    (km) =  %.0f" % MoonDist + suffix)
    print ("MoonHP            (degrees) = %8.3f" %  MoonHP)

    #Alt = 0.7275 * MoonHP - 34./60.     # Meeus p. 98
    print ("Altitude Moonrise, Moonset  = %8.3f" %  MoonAlt)
    print()

    # Mondphasen, Sternbilder etc.
    MoonAge = Mod(MoonLong - SunDict["SunLong"], 360.)
    MoonPhase = 0.5 * (1. - Cos(MoonAge))   # between 0 and 1
    Phases = ("Neumond", "Zunehmende Sichel", "Erstes Viertel",\
             "Zunehmender Mond", "Vollmond", "Abnehmender Mond",\
             "Letztes Viertel", "Abnehmende Sichel", "Neumond")
    mainPhase = 360./29.53
    p = Mod(MoonAge, 90.)
    if p<mainPhase or p>(90.-mainPhase):
        p = 2 * round(MoonAge/90.)
    else:
        p = 2 * floor(MoonAge/90.) + 1
    ShowMoonPhase = Phases[int(p)]
    print ("MoonAge           (degrees) = %8.3f" % MoonAge  )
    print ("MoonPhase (between 0 and 1) = %8.3f" % MoonPhase)
    print ("Show MoonPhase              = ", ShowMoonPhase)
    MoonSign = getSign(MoonLong)
    print ("MoonSign                    = ", MoonSign)
    # Distance (angle) between Sun and Moon
    D = AngularDistance(getSunData(dt, Zone,False), getMoonData(dt, Zone,False))
    print ("Angular Distance Sun-Moon   = %8.3f" % D)

    print ()

    # Moon Parallax - Meeus chapter 39
    # (39.2), assume Earth radius is 1
    deltaRA = atan2(-Cos(latitude)*Sin(MoonHP)*Sin(MoonHourgrad), 
                Cos(MoonDecl)-Cos(latitude)*Sin(MoonHP)*Cos(MoonHourgrad))
    if verbose: print ("deltaRA (rad) = ", deltaRA)
    deltaRA = degrees(deltaRA)
    if verbose: print ("deltaRA (deg) = ", deltaRA)
    # (39.3), assume Earth radius is 1
    MDprime = atan2((Sin(MoonDecl)-Sin(latitude)*Sin(MoonHP))*Cos(deltaRA),
                 Cos(MoonDecl)-Cos(latitude)*Sin(MoonHP)*Cos(MoonHourgrad))
    if verbose: print ("MDprime (rad) = ", MDprime)
    MDprime = degrees(MDprime)
    if verbose: print ("MDprime (deg) = ", MDprime)
    RAprime = MoonRA + deltaRA
    if verbose: print ("RAprime (deg) = ", RAprime)
    print ("Moon RA parallax  (degrees) = %8.3f" % RAprime)
    print ("Moon Decl parallx (degrees) = %8.3f" % MDprime)
    print ()
############## MOON EVENTS, ORDERED BY TIME OF DAY ######################
    MoonEventList = getObjectTimes(dt, longitude, latitude, Zone, getMoonData, 
                        None,False)
    print ("%-4s Events:" % MoonEventList[0]["Object"])
    for e in MoonEventList:
        print ("%-4s %-20s %s Azi %3d Alt %7.3f Di %.0f %2s" % (e["Object"], 
            e["event"], nicedatetime(e['dt']), round(e["azi"], 0), e["alt"],
            e["Dict"]["MoonDist"], e["Dict"]["suffix"]))

    print ()

#   sys.exit(0)
#######################################################################
    useUT = True
    PhasesEventList, EDList, SETList = getMoonPhasesxxx(dt, Zone)
    DecliEventList = getMoonDeclixxx(dt, Zone)
    PeriEventList  = getMoonPeriApoxxx(dt, Zone)
    NodesEventList = getMoonNodesxxx(dt, Zone)
    SynopticEventList = chain(SETList, DecliEventList, PeriEventList, 
                            NodesEventList)

    SortedSynopticEventList = sorted(SynopticEventList,key=itemgetter('dt'))

    print ("Synopsis of Moon Events within %2d days:" % 21, \
   "       Distance   Declination  Latitude")
    print (
   "                                               km         degrees      degrees")
    isBefore = True
    for i in range(len(SortedSynopticEventList)):
        Event = SortedSynopticEventList[i]
        # use only events within +- 21 days of dt
        dtCT = Event['dtCivicTime']
        dtdiff = dt - dtCT
        if abs(dtdiff) > 21.: continue
        if isBefore and dtdiff <= 0.:
            print ()
            isBefore = False
        if Event['Object'] == 'Sun':
            print (" %-11s %17s %-36s %s" % (
                    "Sun", 
                    nicedatetime(Event['dtCivicTime']), Event['event'], ''))
        else:
            print (" %-11s %17s %-13s %6d %2s %8.4f    %8.4f" % (
            Event['Object'], 
            nicedatetime(Event['dtCivicTime']), 
                    Event['mode'], Event['dist'], Event['suffix'],
                    Event['decl'], Event['Dict']['MoonLat']))
    print()
